#!/u/joe/perl561/bin/perl -w

=head1 NAME

imsDepValid - interface to connect the DepValid applet to the Web Gateway

# Originally developed for DepValid Applet (WG 2.1 )

# Following steps are executed per request :
#   1. Init
#   2. Handle HTTP Request (Get/Post)
#   3. Parse ODL
#   4. Invoke Action method 
#   5. Build imsdepval input ODL
#   6. Run imsdepval
#   7. Compose Results
#   8. Return Results

# Acceptable Actions :
#       fetchAll -  First action done, returns the QUERY from the 
#          User's Session to the applet , returns the appropiate Valids and 
#          Filters based on that QUERY.
#       fetchMoreInfo - Saves the applet's query into the file imsDepValq.odl
# 	   in the user's session directory.
#       fetchValidsFilters - Returns the appropiate Valids and Filters
#          based on the Applet's query.
#       fetchValids - Returns the appropiate Valids based on the Applet's query.
#       fetchValidatedQuery - Confirms that the attributes are compatable
#          with each other.
#       updateQuery - Updates the Web Gateway's QUERY with the Applet's query.
#

=cut

#------------------------------------------------------------
#
# PHASE 1... SET UP INCLUDE PATH
#
#------------------------------------------------------------

# Ok , Figure out the system_architecture to include ( IRIX, Solaris, etc)
# We need to add this to the @INC so that Perl can find the C modules associated
# with the CPAN modules ( GD)
BEGIN {
    # Determine architecture
    eval "use Config";
    $system_architecture = $Config{archname};
}

# Where are we? and where is everybody else?

BEGIN {
   use Cwd;
   # Figure out top directory:
   my $path = (($0 =~ m|^/|) ? $0 : (cwd()."/$0"));
   ($path =~ m|^(.*)/[^/]+$|) or 
          spit_out("ignored_argument", "can't extract info from $path!");
   my $top = "$1/..";

   # Set up include path:
   unshift @INC, "$top/lib/perl/first", "$top/lib/perl", "$top/lib/perl/plus", "$top/lib/perl/CPAN", "$top/lib/perl/CPAN/$system_architecture";
}

use IW::App;
use IW::DepValid();
use IW::Widget::DepValids;

use IU;
use IU::ODL ();
use IU::Session();
use IU::Query;
use IU::Log;
use IU::DAACExpansions;

# Requirement of Perl 5.004
use IO::Scalar;

# All known attributes (hard coded) 
# Will be a problem when we support extended valids.

my @attributes_list = qw( CAMPAIGN DATA_CENTER_ID PARAMETER DATASET_ID 
                          PROCESSING_LEVEL SENSOR_NAME SOURCE_NAME);

# Special: expansions of data center acronyms.
my $daacs_names = new IU::DAACExpansions;
$daacs_names->parse_ODL;
my %DataCenterCmnts = $daacs_names->get;

# Holds Global instance of IU:Query object.
# Can't be instantiated here because we need the Session.state first.
# or else we'll get a BLANK query even though the session has one.
my $QUERY ;

# Holds a Web Gateway's copy (Global) instance of IW:DepValid object.
# derived from QUERY object.
my $gatewayDepValid ;

# Holds a local copy (Global) instance of IW:DepValid object.
# which contains the applet's request (query)
my $localDepValid ;

# User's Session directory
my $session_dir; 

# User's Session Id (the directory name under imswww-usr)
my $session_id;

# the URL that will get us into the User's Session directory
my $session_url;

#------------------------------------------------------------
#    get_request
#------------------------------------------------------------
#     Parse the HTTP POST request
#       Returns : the string that was in the request

sub get_request {
   my $self = shift;
   my $request_method = $ENV{'REQUEST_METHOD'};
   my $form_info;

   if ($request_method eq "POST" ) {
        $size_of_information = $ENV{'CONTENT_LENGTH'};
        read (STDIN, $form_info, $size_of_information) ;
   }
   else {
       # we don't handle GET/HEAD requests, so die!
       $self->spit_out( "$request_method not supported!");
   }

   return $form_info;
}

#------------------------------------------------------------
#  new ( Constructor)
#------------------------------------------------------------
#
# OO Constructor
# Now we can call methods!

sub new {
   my $self = {};
   bless $self;
}


#------------------------------------------------------------
#   main
#------------------------------------------------------------
#
# Save the HTTP request to a file
# Parse/Extract values from the ODL object.
# Execute the action obtained from the field ACTION.

sub main {

   # Call the Constructor. Now we're OO compliant.
   my $self = main->new;

   # This is a list of ACCEPTED actions to execute.
   my %invoke_method = (
                "fetchAll"  => \&fetchAll,
           "fetchMoreInfo"  => \&fetchMoreInfo,
      "fetchValidsFilters"  => \&fetchValidsFilters,
             "fetchValids"  => \&fetchValids,
     "fetchValidatedQuery"  => \&fetchValidatedQuery,
             "updateQuery"  => \&updateQuery );

   # Read the HTTP Request
   my $request_sent = $self->get_request();

   # We need to parse the Applet's request
   # Creates a Filehandle to a scalar variable (see ma, no tmp files).
   $AH = tie *IN, 'IO::Scalar', \$request_sent;

   # Read/Parse the scalar variable into an ODL object.
   my $request_odl = parse IU::ODL \*IN;

   # Close the filehandle
   $AH->close;

   # Sanity check : Do we have an ODL object ???
   $request_odl or $self->spit_out( "couldn't parse file");
 
   # Start extracting values from the ODL object ( like the REQUEST_FIELD)
   $self->from_ODL($request_odl);

   # Get the User's Session or die !!!
   my  $gaeatmp_dir = $self->{Gaeatmp_Dir} || $self->spit_out( "No Gaeatmp_Dir supplied");

   # Get the Session's Url
   $session_url =  $self->{ParentUrl} || $self->spit_out("No session_url supplied");

   # Get the session directory and session's id 
   ($session_dir, $session_id) = split(/imswww-usr\//,$gaeatmp_dir);

   # Append the "base" tag onto the user's directory.
   $session_dir .= "imswww-usr" ;

   # Remove the duplicate / from the path.
   $session_id =~ s/\///;

   # Set environment variables.
   $ENV{REMOTE_USER}= $session_id;
   $ENV{GAEATMP_DIR} = ($session_dir."/".$session_id);

   # Where are all our support files located (like archive.odl)
   $ENV{GAEADATA_DIR}="${IU::GaeaDataDir}";

   #  What level should the IMS libraries log messages at?
   #         4 = up to WARNING
   #         5 = up to NOTICE
   #         6 = up to INFO
   #         7 = up to DEBUG

   $ENV{SYSLOG_LEVEL}=5;
   # The Applet's request has been parsed , set localDepValid to the Core 
   # attributes.

#   $localDepValid =  $self->{Attributes} || 
#		spit_out ("No DepValid was Retrieved") ;

   # What action needs to be preformed???
   my $action = $self->{Action};

   # Anything we print from here on out , the applet can see.
   print " Content-type: text/plain \n\n";

   # Goto the mode ( same name as the action) 

   ( defined ($invoke_method{$action})  ? &{$invoke_method{$action}}
               ($self) : $self->spit_out("Unsupported Action"));

}

#------------------------------------------------------------
#    spit_out
#------------------------------------------------------------
# Causes an error message that the applet can see
# 
# Name note :  puke/die are pretty well defined, no need to "overload" their
# format.
 
sub spit_out {
    my ($self, $message) = @_;
    print ("<!-".$message.">\n");
    exit (-1);
}

#------------------------------------------------------------
#    from_ODL
#------------------------------------------------------------
# Parse the request and query from the applet's request.
# ( While we're at it, clean up the spaces (if they're there))
#
# The message is formated as so :
#   GROUP = ACTION
#       VALUE  = "fetchMoreInfo"
#       GAEATMP_DIR = "/www/imsdev/jsa/docs/imswww-usr/913820872-18191/"
#       URL = "/~imsdev/jsa/imswww-usr/913820872-18191/"
#       GROUP = VALUES
#            REQUEST_FIELD = ( "PARAMETER" )
#            DATA_CENTER_ID = ( "ASF")
#            ....
#       END_GROUP = VALUES
#   END_GROUP = ACTION

sub from_ODL {
    my $self = shift;
    my $odl = shift || return undef;
    my ($param, $value);

    # Read the ODL stream:
    my $search;

    if ($search = $odl->param('ACTION')) {
        $search->rewind();

        while (($param, $value) = $search->next) {
           # The action that needs performed.
           if ($param eq 'VALUE') {
               $self->{Action} = $value->text_value;
               $self->{Action} =~ s/\s+//;
           }

           # The user's session directory
           if ($param eq 'GAEATMP_DIR') {
               $self->{Gaeatmp_Dir} = $value->text_value;
               $self->{Gaeatmp_Dir} =~ s/\s//;
           }

           # The session directory's URL
           if ($param eq 'URL') {
              $self->{ParentUrl} = $value->text_value;
              $self->{ParentUrl} =~ s/\s//;
           }

           # Parse the Query from the ODL
           if ($param eq 'VALUES') {
	      my $depValid_list = $self->parse_attributes($value);
	      $self->{Attributes} = $depValid_list;
           }
        }
    }

    # Return it:
    $self;
}

#------------------------------------------------------------
#   retrieveSession 
#------------------------------------------------------------
#
# This routine : opens the user's session state, pulls in the QUERY object 
# ( the user's 'present' search) from the gateway  , and  set's some global 
# variables (for later on)
#  
# Required by  fetchAll and updateQuery
#
sub retrieveSession {
   my $self = shift;

   # Figure out if the user is using a Persistent or a Tmp account
   # persistent is a hardcode location of the persistent accounts
   my $type =  (  $session_url =~ /persistent/ ?   "PERSISTENT" : "STANDARD" );   
   # Get the Session's Url
   $session_url or $self->spit_out("No session_url supplied");

   $SESSION = (IU::Session->resume(Parent=>$session_dir,
               ParentURL =>$session_url , Id => $session_id,
                Type => $type )) or $self->spit_out ("couldn't resume session ");

   # Allow us to read/write a state database for this session.
   $SESSION->setup_state(Opts=>'WRCREAT|BLOCK');

   # Pull in the QUERY object from the Web Gateway.
   $QUERY = $SESSION->state->{Query} || IU::Query->new ;

   # Just gives us the core attributes ( it's a DepValid Object)
   $gatewayDepValid = $QUERY->depvalid_interface() || IW::DepValid->new;

   1;
}


#------------------------------------------------------------
#    fetchAll
#------------------------------------------------------------
# Before the applet can be displayed, the applet needs the user's search
#  from the Web Gateway.
# This method gets the user's query (if present) from their session.
# ***Which will overwrite the localQuery in this method.
#
#  Uses the query to get the appropiate Valids.
#  Uses the query to get the appropiate Filters.
#  Returns the User's present Query, Valids and Filters 
#       in Parameter=Valid format.

sub fetchAll {
   my $self = shift;

   # Connect to the user's session state.
   $self->retrieveSession();

   
   # Do we REALLY have a Query to work with?
   my $attribute;
   my $have_query = 'NO';
   foreach  $attribute ($gatewayDepValid->get_categories) {
      my @cat_array = $gatewayDepValid->set($attribute);
      if ( int(@cat_array) > 0 ) {
        $have_query ="YES";
      }
   }

   # Yes, we do.
   if ( $have_query eq "YES") {
      # Create an ODL object
      my $dep_odl = $gatewayDepValid->build_query;

      # Verify the dependencies between the attribute.
      # Creates the file imsDepValq.odl in the user's session directory.
      my $valid_query =  $gatewayDepValid->check_query($dep_odl);

      # Convert the Daac short names to Long names.
      my $updated_valid_query = $self->expandDaacNames($valid_query);

      # Display the results.
      $updated_valid_query->display if (defined ($valid_query));
   }
   else {
      # Problem : If this is the user's second "turn"  using the applet and
      # they cleared their previous QUERY (leaving a blank search form)
      # then re-enter the applet. The file "imsDepValq.odl" still contains
      # the user's last "search".
      # Not a problem if there is a new query, it overwrites imsDepValq.odl.

      my $file = ($self->{Gaeatmp_Dir}."/imsDepValq.odl");
      if ( -e $file) {
         ( unlink($file)) || $self->spit_out("Couldn't delete file") ;
      }
   }

   # if gatewayDepValid has a query, now the local copy needs it so
   # the dependencies are still true when we get the valids and filters.
   $localDepValid = $gatewayDepValid;

   # The  Copying above, caused us to lose the REQUEST_FIELD/FILTERS &
   # PATTERNS, so refill them in.

   $localDepValid->{REQUEST_FIELD} = $self->{Attributes}->{REQUEST_FIELD};
   $localDepValid->{FILTERS} = $self->{Attributes}->{FILTERS};
   $localDepValid->{PATTERNS} = $self->{Attributes}->{PATTERNS};

   # Get the appropiate valids.
   $self->get_Valids();

   # Get the appropiate filters.
   $self->get_Filters();

}

#------------------------------------------------------------
#  fetchMoreInfo
#------------------------------------------------------------
# Saves the applet's query into the file imsDepValq.odl in the user's session
# directory.

sub fetchMoreInfo {
   my $self = shift;

   # Get the Extended Valids too.
   $self->read_extended_valids();

   # Build the Query based on the Applet's valids plus any Extended Valids
   # that maybe around.
   my $query_odl = $localDepValid->build_query;

   # Write the Query to file.
   my $odlfile =  $localDepValid->write_query( $query_odl ,
        File=>'imsDepValq.odl', Type=>'VALIDS');

   # Return some type of acknowledgement.
   ( $odlfile  ? print "OK\n" : print "<!--no output--!>\n"); 
}

#------------------------------------------------------------
#     fetchValidatedQuery
#------------------------------------------------------------
# 
# Takes a query as input , runs imdepval(.c) ((ex) which verifies that a s
# selected parameter is valid with a selected data_center), expands the
# Short data center names to long and displays the results from the
# verification.
#
#   Verify that the dependencies between attributes are true, if
#   not place an "*" in front of invalid entries.

sub fetchValidatedQuery {
   my $self = shift;

   $self->read_extended_valids();
   $localDepValid->{REQUEST_FIELD} = $self->{Attributes}->{REQUEST_FIELD};
   $localDepValid->{FILTERS} = $self->{Attributes}->{FILTERS};
   $localDepValid->{PATTERNS} = $self->{Attributes}->{PATTERNS};

   # Verify that localDepValid has real Query (values) inside  it.
   my $have_query = "NO";
    foreach  $attribute (IU::Category->get_core()) {
      my @cat_array = $localDepValid->set($attribute);
      if ( int(@cat_array) > 0 ) {
        $have_query ="YES";
      }
   }

   my $queryODL = $localDepValid->build_query();


   # Yes, we do.
   # Run imdepval on it.
   if ( $have_query eq "YES") {
       # Convert the request into an ODL object
       my $dep_odl = $localDepValid->build_query;

       # Run imsdepval the validate the dependencies
       my $valid_query =  $localDepValid->check_query($dep_odl);

       # Convert the Daac short names to Long names (Spell out the acronym).
       my $updated_valid_query = $self->expandDaacNames($valid_query);

       # Display results
       $updated_valid_query->display if (defined ($valid_query));
    }

    # To speed up the verification/new valids & filters process, 
    # fetchValidatedQuery now does a fetchValids itself
    # ( it originally did this in 2 steps instead of one)

    $self->get_Valids();
}

#------------------------------------------------------------
#    fetchValidsFilters 
#------------------------------------------------------------
# Routines that return the Valids and Filters in one swoop.

sub fetchValidsFilters {
    my $self = shift;

    $self->read_extended_valids();
    $self->get_Valids(); 
    $self->get_Filters();
}

#------------------------------------------------------------
#  fetchValids
#------------------------------------------------------------
#  Returns the Valids
#  Communication layer between Applet and CGI script

sub fetchValids {
   my $self = shift;
   $self->read_extended_valids();
   $self->get_Valids();
}

#------------------------------------------------------------
#     updateQuery
#------------------------------------------------------------
#    Update the User's Query based on the query from the applet
#    Needs the user's session_id
#  

sub updateQuery {
   my $self = shift;

   # Connect to the user's session state.
   $self->retrieveSession();

   $self->read_extended_valids();

   # Update the Web Gateway's QUERY object  based on the user's 
   # selection from the applet.
   my $attribute;
   # Set the CORE attributes only, don't mess with the Extended.
   foreach  $attribute (IU::Category->get_core) {
        if ( $localDepValid->set($attribute) ) {
          my @attr_values = $localDepValid->set($attribute);
          $QUERY->criteria_strings($attribute, \@attr_values);
        }
        else {
          $QUERY->criteria($attribute, undef);
        }
   }

   # Write it in stone ( to the user's session state)
   $QUERY->install;

   print "OK\n";
}


#------------------------------------------------------------
#     get_Valids
#------------------------------------------------------------
#  Get the valids based on the supplied REQUEST_FIELD
#  Results also based on (any) Patterns/Query supplied.

sub get_Valids {
    my ($self, %opts) = @_;
    my %new_query;
    my $session_query ;
    my @valids;

    my $request_field = $localDepValid->{REQUEST_FIELD};
    my @filters = defined($localDepValid->{FILTERS})
                      ? @{$localDepValid->{FILTERS}}
                      : ();

    #  The valids are "cleaned" of leading "*", etc.
    #  Items consisting of just dashes (separators) are removed.
    @filters = grep {!/^-+$|^\s*$/} @filters;        

    # Items which contain nothing but whitespace are removed.
    @filters =  grep { s/^\s*|\s*$//g} @filters;

    # Clean up the Patterns (same way as the Filters were)
    # Items which contain nothing but whitespace are removed.
    my @patterns;
    ( @patterns  = (grep /\S/, map { s/^\s*|\s*$//g; $_} split /;/,
                 $localDepValid->{PATTERNS} ))
                 if ($localDepValid->{PATTERNS});

    @valids = $localDepValid->get_valids($request_field,
                                  Filters => \@filters,
                                  Patterns => \@patterns);

   # Remove duplicate entries and sort the results. 
   my @sorted_valids = $self->clean_up(\@valids);

   # Did we get any results
   # imsdepval could have crashed???
   my $valids_count = ( @sorted_valids ? int(@sorted_valids) : 
            print "No Valids returned\n");

   # Display the Valids in VALIDS=value format
   # If this list is of Daacs (short) then convert to the long format 
   if (@sorted_valids) {
      my $i;
      my $expansion;
      for ($i=0;$i<$valids_count;$i++) {
           if ($localDepValid->{REQUEST_FIELD} eq 'DATA_CENTER_ID') {
              $expansion = ($DataCenterCmnts{$sorted_valids[$i]})
                 ? "($DataCenterCmnts{$sorted_valids[$i]})" : "";
               print "VALIDS=$sorted_valids[$i] $expansion\n"
          }
          else {
              print "VALIDS=".$sorted_valids[$i]."\n";
          }
      }
   }
}

#------------------------------------------------------------
#     clean_up
#------------------------------------------------------------
#  Remove duplicate and sort the results.

sub clean_up {
   my $self = shift;
   my $dirty_array = shift;

   my %full_hash = ( map { $_=>1 } @$dirty_array );
   my @clean_array = sort keys (%full_hash);

   return @clean_array;
}

#------------------------------------------------------------
#     get_Filters
#------------------------------------------------------------
#  Get the filters based on the supplied REQUEST_FIELD
#  Results also based on (any) Patterns/Query supplied.

sub get_Filters {
   my ($self, %opts) = @_;
   my %new_query;
   my $session_query ;

   my $request_field = $localDepValid->{REQUEST_FIELD};

   @filters = $localDepValid->get_filters($request_field,
                                  Filters => \@filters);

   # Remove duplicate and  Sort the results.
   @$filters = $self->clean_up(\@filters);

   # Display the Filters ( FILTERS=value format )
   my $filters_count =(@filters ? int(@filters) : 0 );
   if ($filters_count > 0) {
      for ($i=0;$i<$filters_count;$i++) {
        print "FILTERS=".$filters[$i]."\n";
      }
   }
   else {
	print "No Filters returned\n";
   }


}

#------------------------------------------------------------
#    expandDaacNames
#------------------------------------------------------------
#
# Spell out the acronyms.
# For instance :
# GSFC = Goddard Space Flight Center

sub expandDaacNames {
   my $self = shift;
   my $query_chunk = shift;
   my ($key, $val, $elem);

   my @long_daacs;
   my $expansion;
   while (($key, $val) = each %$query_chunk) {
      if ($key eq 'DATA_CENTER_ID') {
        foreach $elem (@{$query_chunk->{$key}}) {
           $elem =~ s/\s+//;
           my $clean_elem = $elem;
           $clean_elem  =~ (s/\*// ) if ($clean_elem  =~ /\*/);
           $expansion = ($DataCenterCmnts{$clean_elem})
              ? "($DataCenterCmnts{$clean_elem})" : "";
           push(@long_daacs, "$elem $expansion");
        }
        $query_chunk->set($key,@long_daacs);
      }
   }
   return $query_chunk;
}

#------------------------------------------------------------
#     to_ODL
#------------------------------------------------------------
#     Convert Query into ODL object
#	return the ODL object

sub to_ODL {
   my $self = shift;

   my $field_agg = IU::ODL->new_group;
   my $attribute;
   foreach  $attribute (@attributes_list) {
   if ($localDepValid->{$attribute}) {
         $field_agg->add_param($attribute => 
		IU::ODL->conv($localDepValid->{$attribute}));
      }
   }

   return ($field_agg);
}

#------------------------------------------------------------
#     parse_Values
#------------------------------------------------------------
#  Parse the Query/REQUEST_FIELD/Filters/Patterns from ODL message
#
sub parse_attributes {
    my $self = shift;
    my $odl = shift;

    my $tmpDepValid = IW::DepValid->new( Session_Path => "$self->{Gaeatmp_Dir}/");
    my $value;
    # The selected attribute
    if ($value = $odl->param('REQUEST_FIELD')) {
         $tmpDepValid->{REQUEST_FIELD} = $value->text_value;
         $tmpDepValid->{REQUEST_FIELD} =~ s/\s//;
    }

    # The any filters associated with the selected attribute
    if ($value = $odl->param('FILTERS')) {
        my @filters =  $value->text_values;
        my $i;

        $tmpDepValid->{FILTERS} = \@filters;
        # How do I clean an array???
        # $class->{FILTERS} =~ s/\s//;
    }
    if ($value = $odl->param('PATTERNS')) {
        $tmpDepValid->{PATTERNS} = $value->text_value;
        # collapse sequences of whitespace to one space
        $tmpDepValid->{PATTERNS} =~ s/\s+/ /g;
    }


    # Core dependent valids, and other sequence strings...

    my @clear_array = undef;
    my ($i, $attribute);
       my @attribute_list = IU::Category->get_core();
       foreach  $attribute (@attribute_list) {
           $tmpDepValid->set($attribute, \@{undef});
           my $value;
           if ($value = $odl->param($attribute)) {
             my @attribute_values = sort(map {clean_valid($_)}
                                      $value->text_values);

             # Applet sends long names for the Data Centers, imsdepval can only
             # handle the short. So we remove the extension.
             if ($attribute eq 'DATA_CENTER_ID') {
                my @clean_daacs ;
                for ($i=0;$i<int(@attribute_values);$i++) {
                        my @parts = split(/\s\(/, $attribute_values[$i]);
                        $parts[0] =~ s/\s//;
                        push ( @clean_daacs, $parts[0]);
                }
                @attribute_values = @clean_daacs;
             }

             # Put the parsed values into the self object.
             $tmpDepValid->set($attribute, \@attribute_values);

             # We've have at least one attribute
             $tmpDepValid->{HAVE_QUERY} = "YES";
           }
       }
    return $tmpDepValid;
}

#------------------------------------------------------------
#     read_extended_valids
#------------------------------------------------------------
#  Applet doesn't support extended valids, but we need to use them
#  when a user changes attributes.

sub read_extended_valids {
    my $self = shift;
    my $path = ($self->{Gaeatmp_Dir}."/imsDepValq.odl");

    # The Applet's request has been parsed , set localDepValid to the Core
    # attributes.
    $localDepValid =  $self->{Attributes} ||
                 $self->spit_out ("No DepValid was Retrieved \n") ;

    if ( -s  $path ) {
       open EXTENDED, "$path" || 
          $self->spit_out ("can't open $path for reading: $!");
       my $full_query_odl =  parse IU::ODL \*EXTENDED;
       close EXTENDED;

       my $extendedDepValid = 
        IW::DepValid->new( Session_Path => "$self->{Gaeatmp_Dir}/");
       $extendedDepValid->from_ODL($full_query_odl);
       $extendedDepValid ||  $self->spit_out ("No DepValid was Retrieved") ;
  
       my $attribute;
       foreach  $attribute (IU::Category->get_core) {
         if ( $self->{Attributes}->set($attribute) ) {
          my @attr_values = $self->{Attributes}->set($attribute);
          $extendedDepValid->set($attribute, \@attr_values);
         }
         else {
          $extendedDepValid->set($attribute, undef);
         }
       }

       $localDepValid = $extendedDepValid;
       $localDepValid->{REQUEST_FIELD} = $self->{Attributes}->{REQUEST_FIELD};
       $localDepValid->{FILTERS} = $self->{Attributes}->{FILTERS};
       $localDepValid->{PATTERNS} = $self->{Attributes}->{PATTERNS};
    }
}

###  GO! ###
exit(main() ? 0 : -1);
