#!/u/joe/perl561/bin/perl -w

=head1 NAME

imscleanaccounts.pl
   - periodic clean up of the persistent user accounts.

=head1 DESCRIPTION

Go through the accounts in the persistent user directory and
clean out old results, logout inactive sessions, etc.  See
comments in header of cleanup_accounts below.

Script is run by wwwuser, usually in a cronjob.
Usage:
    ~imswww/pub/lib/bin/imscleanaccounts.pl -once
See print_usage or try it without any arguments.

=head1 WARNING

Assumes that it is being run out of the ./lib/bin directory
BUT use the full-path to this script!  There should only be
one instance of this script running per persistent user directory.

=head1 AUTHOR

Larry P. Rosen, Raytheon STX Corporation
Partly based on imscleanup by Eryq (Erik Dorfman).

=cut

# GLOBAL
my $topDir;  # This is the directory containing all the user accounts.
             # Don't confuse with $IU::TopDir which heads the
             # installation.

use Cwd;

# Set up include path:
BEGIN {
    # Figure out top directory:
    my $path = (($0 =~ m|^/|) ? $0 : (cwd()."/$0"));
    ($path =~ m|/lib/bin/[^/]+$|) or
       die "can't extract info from $path!";   
    my $perltop = "$`/lib/perl";

    # Set up include path:
    unshift @INC, "$perltop/first", "$perltop", "$perltop/plus";
    push @INC, "$perltop/fallback";
}

use IU;
use IU::Log;
use IU::Util qw(units2secs units2bytes nice_time);
use IU::ExclFileLogger;
use IU::Event;
use IU::FileCache;
use IU::Session;
require IU::Mail;

use strict;

#------------------------------------------------------------
#
# GLOBALS
#
#------------------------------------------------------------
# If in loop mode, wait this many minutes between loops.
my $PAUSEMINS = 15;

#------------------------------------------------------------
# cleanup_accounts
#------------------------------------------------------------
# Cleanup all user accounts that need it.
# 1. Logout accounts that have been inactive for LogoutTime.
# 2. Delete results of logged out accounts that have not been accessed
#    for a period of ResultsLifetime.
# 3. Delete all non-critical files, such as logs, RxODL, TxODL, B*.odl,
#    and browse files, from logged out accounts if the account is
#    inactive for more than SessionLife.  They aren't used
#    except for debugging.
# 4. Check sum of all accounts disk space.  If greater than
#    AllAccountsMax, start deleting results from older accounts.
# 5. Check each account disk usage.  If greater than AccountMax, then:
#      If logged out, delete the results.
#      Else if account is logged in, warn the user by placing a warning
#       file in their account.
# 6. Re-check the sum of all accounts disk space.  If greater than
#    CriticalAllAccountsMax then we are in trouble.  E-mail the system
#    administrator with the level of the problem and how far we got
#    here.
#    a. Delete non-critical files of all logged out accounts.
#    Re-check disk usage.  If still greater, then:
#    b. Delete results from all logged out accounts.
#    Re-check disk usage.  If still greater, then:
#    c. For logged in accounts, delete any guide files that haven't
#       been accessed for a period of GuideKillTime.
#    Re-check disk usage.  If still greater, then:
#    d. Warn all logged in users to logout, explaining why.

sub cleanup_accounts {
    inform "Cleaning user accounts:";

    #   1. Logout accounts that have been inactive for LogoutTime.
    #      This function will also get a hash of all user accounts
    #      and their last access time.
    inform "Logout accounts inactive for $IU::LogoutTime:";
    my %accountAges = logout_inactive_accounts ();

    #   2. Delete results of logged out accounts that have not been
    #      accessed for a period of ResultsLifetime.
    inform "Deleting results older than $IU::ResultsLifetime:";
    delete_old_results (%accountAges);

    #   3. Remove non-critical files from logged out accounts:
    #      RxODL, TxODL, B*.odl, logs, and browse.
    inform "Purging non-critical files from logged out accounts.";
    delete_all_noncritical (%accountAges);

    #   4. Check sum of all accounts disk space.  If greater than
    #      AllAccountsMax, then delete results from logged out
    #      accounts.  E-mail the site administrator.
    #      Set flag ($diskUsageExceededFlag) if disk usage is too big.
    inform "Checking if total account disk space < " . 
	"$IU::AllAccountsMax.";

    my $diskUsageExceededFlag = check_accounts_total (%accountAges);

    #   5. Check each account disk usage.  If greater than AccountMax:
    #      a. If logged out, delete the results.
    #      c. If logged in, warn the user by placing a warning
    #         file in their account.
    inform "Checking if individual accounts disk usage < " .
	"$IU::AccountMax.";

    check_accounts (%accountAges);

    #   Even if $diskUsageExceededFlag is false, call the next
    #   function, because it needs to first purge any previous
    #   warning files.
    #   6. Re-check the sum of all accounts disk space.  If greater
    #      than CriticalAllAccountsMax then we are in trouble.  E-mail
    #      the system administrator with the level of the problem and
    #      how far we got here.
    #      a. Delete non-critical files of all logged out accounts.
    #      Re-check disk usage.  If still greater then:
    #      b. Delete results from all logged out accounts.
    #      Re-check disk usage.  If still greater then:
    #      c. For logged in accounts, delete any guide files that
    #         haven't been accessed for a period of GuideKillTime.
    #      Re-check disk usage.  If still greater then:
    #      d. Warn all logged in users to logout, explaining why.
    #
    check_critical_usage ($diskUsageExceededFlag, %accountAges);

    1;
}

#------------------------------------------------------------
# logout_inactive_accounts
#------------------------------------------------------------
# Logout accounts that have been inactive for LogoutTime.
# Global: $topDir - scalar - full path of directory of accounts.
# Returns: %accountAges - hash - table of age in seconds since last
#                                access for each account.  Account
#                                names are keys.
#------------------------------------------------------------
sub logout_inactive_accounts {
    my (%accountAges, $age);

    # Open directory stream on user-account directories...
    opendir DIR, $topDir or puke "couldn't open $topDir directory: $!";

    # Check each directory:
    my $account;
    my $path_account;
    while (defined($account = readdir(DIR))) {
        $path_account = $topDir . '/' . $account;
	# Skip the "." and ".." files (and any that begin in ".",
        # skip links, and non-directories.
	if (($account =~ /^\./) || (-l $path_account) ||
	    !(-d $path_account)) {
	    next;
	}
	# Get the age of the session by checking the age of .kilroy.
        # -M returns float number of days, so multiply to get secs.
	$age = (-M "$path_account/.kilroy") * 86400;

	# Store the age in the hash of accounts.  Name is key.
	$accountAges{$account} = $age;

	# Skip accounts that are already logged out.
	unless (-e "$path_account/persistent_user.lock") {
	    next;
	}
	# If the account hasn't been accessed in a while, LogoutTime,
        # log it out (by deleting the lock file).
	if ($age > units2secs($IU::LogoutTime)) {
	    if (unlink ("$path_account/persistent_user.lock") == 1) {
		inform "Logged out user $account";
	    } else {
		inform "Problem logging out $account. $!";
	    }
	}
    }
    return %accountAges;
}

#------------------------------------------------------------
# delete_old_results  ACCOUNT_AGES_HASH
#------------------------------------------------------------
# Delete results of logged out accounts that have not been
# accessed for a period of ResultsLifetime.
# Global: $topDir - scalar - full path of directory of accounts.
# Input:  %accountAges - hash - table of age in seconds since last
#                                access for each account.  Account
#                                names are keys.
#------------------------------------------------------------
sub delete_old_results {
    my (%accountAges) = @_;

    my @accounts = sort (keys (%accountAges));

    # Check each directory:
    my $account;
    foreach $account (@accounts) {
	# Skip accounts that are logged in.
	if (-e "$topDir/$account/persistent_user.lock") {
	    next;
	}
	# Ready for deletion?
	if ($accountAges{$account} >
            (units2secs($IU::ResultsLifetime))) {
	    delete_results ($account, 'TIME') or
		error "cleanup of $account failed; continuing..."; 
	}
    }
}

#------------------------------------------------------------
# delete_results  ACCOUNT  REASON
#------------------------------------------------------------
# If the account has been purged, just return.
# Delete results and other state files from a user account that
# is logged out.  This code actually deletes everything except
# what we save here.  Two temporary directories are used.  The
# first will hold what we want to save.  The second is what we
# rename the existing user account directory.  Then we rename the
# first temporary directory to the user account directory.
# Then we delete the second temporary directory.
# Global: $topDir - scalar - full path of directory of accounts.
# Input:  $account - scalar - account name = directory name
#         $reason - scalar - determines the message that appears on
#                            the user's next page access.  This is a
#                            temporary message because the account is
#                            logged out.
#         $reason is one of:
#            'TIME' for exceeding results limit.
#            'SPACE' for exceeding individual space limit.
#            'ALLSPACE' for all accounts exceeding space limit.
# Returns: 1 if results were deleted from account.
#          0 if error - failed to delete results.
#          3 if account had already been purged.
#------------------------------------------------------------
sub delete_results {
    my ($account, $reason) = @_;

    # If account already has been purged (there won't be a state file)
    # just return.
    unless (-e "$topDir/$account/session.state") {
	return 3;
    }
    # Files we want to keep. (We delete all others!)
    my @keepers = qw (.htaccess .kilroy prefs.dat query.html
		      dynamic.odl pub/query.dat pub/.htaccess
                      profile2.odl);

    my $status;

    # If a temporary directory to hold the stuff we'll keep already
    # exists, try to remove it now.
    if (-e "$topDir/cleaning_$account") {
        $status = system ("/bin/rm -rf $topDir/cleaning_$account");
	if ($status >> 8) {
	    return error "Couldn't remove cleanup directory, " .
		"go on. $!"; 
	}
    }
    # If a temporary directory for deletion already exists,
    # try to remove it now.
    if (-e "$topDir/temp_$account") {
	$status = system ("/bin/rm -rf $topDir/temp_$account");
	if ($status >> 8) {
	    return error "Couldn't remove cleanup temp directory, " .
		"go on. $!"; 
	}
    }
    # Make a temporary directory to hold the stuff we'll keep.
    $status = mkdir "$topDir/cleaning_$account", 0755;
    unless ($status) {
	return error "Couldn't create directory cleaning_$account " .
	    "for cleanup! $!";
    }
    # Need a "pub" subdirectory.
    $status = mkdir "$topDir/cleaning_$account/pub", 0755;
    unless ($status) {
	return error "Couldn't create cleaning_$account/pub for " .
	    "cleanup! $!";
    }
    # Copy each file we want to keep.
    my ($keep, $file);
    foreach $keep (@keepers) {
	$file = "$topDir/$account/$keep";
	# If the file is in the old dir:
	if (-e $file) {
	    # Copy the file to the temp. directory.
	    $status =
		system (
		   "/bin/cp -R $file $topDir/cleaning_$account/$keep");
	    if ($status >> 8) {
		return error "Couldn't copy $keep. $!";
	    }
	}
    }
    # Create a notification message in user's account, based on REASON.
    if (open MSG, ">$topDir/cleaning_$account/temp_msg.txt") {
      SWITCH: {
	  if ($reason eq 'TIME') {  # Deleted old results.
	      print MSG "Because you have not accessed this ",
	          "account in at least ";
	      print MSG 
		  nice_time (units2secs ($IU::ResultsLifetime));
	      print MSG <<EOTIMEMSG;
, your previous results were removed.  This is done
to keep your results from becoming unreliable (out of
date) and to free up system computer disk space.
Your last search query is still available and can be loaded
into the search form by clicking the "Last search done" link 
below this message. 
EOTIMEMSG
	      last SWITCH;
	  }
	  if ($reason eq 'SPACE') {
	      # Indidvidual account exceeded limit.
	      print MSG <<EOSPACEMSG;
Because your account was taking up too much
disk space on this system, your previous results were
removed.  In the future, you can reduce space usage
by often emptying your shopping cart and folder,
and by performing searches that return fewer results.
Your last search query is still available and can be loaded
into the search form by clicking the "Last search done" link
below this message.
We apologize for any inconvenience this removal may have caused.
EOSPACEMSG
	      last SWITCH;
	  }
	  if ($reason eq 'ALLSPACE') {
	      # Total of accounts exceeded space limit
	      print MSG <<EOALLMSG;
WARNING!!!  There was a shortage of computer
disk space on this system.  The system administrator
will be taking steps to correct the problem.  Your
previous results were removed.  In the future, you
can help reduce space usage by often emptying your
shopping cart and folder, and by performing
searches that return fewer results. 
Your last search query is still available and can be loaded
into the search form by clicking the "Last search done" link
below this message.
We apologize for any inconvenience this removal may have caused.
EOALLMSG
	      last SWITCH;
	  }
        }  # end SWITCH
	close MSG;
    }
    # Check one last time to make sure no one has just logged in.
    if (-e "$topDir/$account/persistent_user.lock") {
	inform "Aborting cleanup of account $account; just logged in.";
	# Try to remove the temporary directory.
	# We don't care now about status.
	system ("/bin/rm -rf $topDir/cleaning_$account");
	return 1;
    }
    # Rename the user account directory to a temporary one.
    unless (rename ("$topDir/$account", "$topDir/temp_$account")) {
	return error "Can't rename user account $account to " .
	    "temp_$account! $!";
    }
    # Rename newly created directory of kept stuff to the user account.
    unless (rename ("$topDir/cleaning_$account", "$topDir/$account")) {
	inform "Aborting - Can't rename kept stuff back to user " .
	    "account directory! Now attempting to restore their old " .
		"directory back.";
	unless (rename ("$topDir/temp_$account", "$topDir/$account")) {
	    return error "WE CANNOT RESTORE THE USERS ACCOUNT!!! " .
		"Can't rename temporary directory temp_$account to " .
		    "user account directory $account! $!";
	}
    }
    # Now we can delete the old stuff.
    $status = system ("/bin/rm -rf $topDir/temp_$account");
    if ($status >> 8) {
	return error "Couldn't remove old stuff. $!"; 
    }
    inform "Deleted results from account $account.";
    # Return true, all ok.
    1;
}

#------------------------------------------------------------
# check_accounts_total  ACCOUNT_AGES_HASH
#------------------------------------------------------------
# For user accounts: 
#    Check sum of all accounts disk space.  If greater than
#    AllAccountsMax, then delete results files from logged out
#    accounts.  Start with accounts that haven't been touched in the
#    longest time and work forwards until the limit is not exceeded.
#    E-mail site administrator.
# Global: $topDir - scalar - full path of directory of accounts.
# Input:  %accountAges - hash - table of age in seconds since last
#                                access for each account.  Account
#                                names are keys.
# Returns 1 if the accounts total still exceeds the limit, else 0.
#------------------------------------------------------------
sub check_accounts_total {
    my (%accountAges) = @_;

    my ($startSize, $endSize); # Current all accounts disk usage in kb.

    if (too_big ($topDir, $IU::AllAccountsMax, \$startSize)) {

	my $count = 0;             # Count of accounts purged.
        my $still = 1;             # True if limit still exceeded.

	# Get list of accounts, sorted from oldest to newest access.
	my @accounts = sort { $accountAges{$b} <=> $accountAges{$a} }
	               keys %accountAges;
	my $account;

	# While the limit is still exceeded, purge next oldest account.
	foreach $account (@accounts) {

	    # Skip accounts that are logged in.
	    if (-e "$topDir/$account/persistent_user.lock") {
		next;
	    }
	    # Purge the account.
	    if (delete_results ($account, 'ALLSPACE') == 1) {
		# Increment count of directories purged.
		$count++;
	    }
	    # Recheck the limit.  If we're ok, break the loop.
	    unless (too_big ($topDir, $IU::AllAccountsMax, \$endSize)){
		$still = 0;
		last;
	    }
	}
	# E-mail ims administrator, warning of the disk limit.
	my $message =
"WARNING!!! The IMS Web Gateway running on your system
has run out of disk space for user accounts and has
attempted to reduce disk usage by removing results
from logged out users.

The current limit AllAccountsMax is $IU::AllAccountsMax
and is set in the file SETUP in the installation
directory ($IU::TopDir).  Results were
deleted from $count logged out accounts.
The original disk usage was $startSize kb.
After purging, the usage is now $endSize kb.\n";

	# If limit still exceeded (still == 1) then big trouble ahead.
	if ($still) {
	    $message .= "The limit is still exceeded!\n";
	} else {
	    $message .= "The disk usage is now less than the limit,
but may be exceeded in the future.\n";
	}
	$message .= "If possible, you should increase the disk
space allocated for all user accounts in SETUP (and re-run 
\"setup -apply\" as user imswww.  Let the Web Gateway developers
know if you need help.\n";

	# Create mail object:
	my $mail = new IU::Mail;
	$mail->to($IU::SiteAdminEmail || 'imswww');   # Who to send to.
	$mail->to('lrosen@harp.gsfc.nasa.gov');   # Who to send to.
	$mail->subject(
           "IMS Web Gateway Running Out of Total Account Disk Space");
	$mail->set('Content-Type', 'text/html');
        # Write message:
	my $FH = $mail->open || puke "mail open failed!";
	print $FH $message;
	$mail->send;
	return 1 if $still;
    }
    # Else total of accounts is now ok.  Return 0.
    return 0;
}

#------------------------------------------------------------
# too_big FULLPATH_DIRECTORY LIMIT SIZE_REF
#------------------------------------------------------------
# If the total disk usage for the specified directory is less
# than the specified limit, return 0 (false), else return true (1).
# Also sets variable reference points to the disk usage of the
# specified directory.
# Input: $fulldir - scalar - full directory path of a directory
#        $limit - scalar - disk space limit not to be exceeded in Units
#                          such as "5mb" or "25 GB".
#        $refsize - scalar reference - reference to scalar to hold
#                                      disk usage of specified
#                                      directory in KiloBytes.
# Returns: 0 if total recursive disk usage of the specified
#               directory is less than the specified limit.
#          Else: 1 == true.
#          And also sets variable to which the reference points to
#              the disk usage of the specified directory.
#------------------------------------------------------------
sub too_big {
    my ($fulldir, $limit, $refsize) = @_;
    
    # Create a command to get the total disk usage in KB.
    my $cmd = "du -ks $fulldir";
    # Open pipe to disk usage command.
    open (DISKUSAGE, "$cmd |") or
        puke "Cannot determine disk usage: Open $fulldir";
    my $sizestring = <DISKUSAGE>;
    close (DISKUSAGE) or
	puke "Cannot determine disk usage: Close  $fulldir";
    my ($size, $junk) = split ' ', $sizestring;

    # Store directory size in the variable pointed to by the reference.
    $$refsize = $size;

    # Convert size in KB to bytes.
    $size *= 1024;

    # If disk usage less than limit return false (0);
    if ($size < units2bytes ($limit)) {
#	inform "$fulldir usage: $size bytes < limit $limit";
	return 0;
    } else {
#	inform "$fulldir usage: $size bytes exceeds limit $limit";
	return 1;
    }
}

#------------------------------------------------------------
# delete_all_noncritical  ACCOUNT_AGES_HASH
#------------------------------------------------------------
# For all user accounts that are logged out, delete non-critical
# files (such as logs, ODL, browse).
# Global: $topDir - scalar - full path of directory of accounts.
# Input:  %accountAges - hash - table of age in seconds since last
#                                access for each account.  Account
#                                names are keys.
#------------------------------------------------------------
sub delete_all_noncritical {
    my (%accountAges) = @_;

    my @accounts = sort (keys (%accountAges));
    my ($account, $fulldir);

    # Check each directory:
    foreach $account (@accounts) {
	# Skip accounts that are logged in.
	if (-e "$topDir/$account/persistent_user.lock") {
	    next;
	}
        # Skip account if accessed < SessionLife ago.
        if ($accountAges{$account} < units2secs ($IU::SessionLife)) {
	    next;
	}
	# Delete the noncritical files in the account:
	$fulldir = "$topDir/$account";
	delete_noncritical ($fulldir);
    }
}

#------------------------------------------------------------
# delete_noncritical FULLPATHNAME_OF_ACCOUNT
#------------------------------------------------------------
# For the given full path to a user account directory, delete
# the "non-critical" files there.
# Be careful editing this routine; we're using "rm -rf" !!!
# Input: $fulldir - scalar - full pathname of account directory
#------------------------------------------------------------
sub delete_noncritical {
    my ($fulldir) = shift;

    # List of files to delete.  Would do *.odl except keep dynamic.odl.
    # Use glob on patterns with "*" to get matching directory contents.
    # "browse" will delete entire subdirectory.
    my @non_crit = ("$fulldir/browse",       
                    "$fulldir/depvalid.odl",
                    "$fulldir/filterq.odl",
                    "$fulldir/search.odl");
    push (@non_crit, glob ("$fulldir/*log"));
    push (@non_crit, glob ("$fulldir/*TmpFile.*"));
    push (@non_crit, glob ("$fulldir/B*.odl"));

    # If the search subdirectory exists, add its non-critical
    # files to list of files to delete.
    my $subdir = 'search';
    my $fullsubdir = $fulldir . '/' . $subdir;
    if (-e $fullsubdir) {
       push (@non_crit, glob ("$fullsubdir/*log"));
       push (@non_crit, glob ("$fullsubdir/*TmpFile.*"));
       push (@non_crit, glob ("$fullsubdir/B*.odl"));
       push (@non_crit, "$fullsubdir/search.odl");
       push (@non_crit, "$fullsubdir/client.obj");
    }
    my $count = 0;   # Number of files deleted (browse files count as 1).

    my $file;

    # Delete the list of noncritical files.
    foreach $file (@non_crit) {
        $count += unlink ($file)  if (-e $file);
    }
    # Delete browse directory and contents if it exists.
    my $delbrowse = '';
    $file = $fulldir . '/browse';
    if (-e $file) {
       unless (system ("/bin/rm -rf $file")) {
	  $count++;
	  $delbrowse = 'at least';
       }
    }
    # Open the "pub" directory and remove gif files.
    if (opendir (PUB, "$fulldir/pub")) {
        while (defined ($file = readdir (PUB))) {
	    # If file ends in .gif, remove it.
	    if ($file =~ /\.gif$/) {
	        $count += unlink ("$fulldir/pub/$file");
	    }
	}
	closedir (PUB);
    }
    if ($count) {
	inform "Deleted $delbrowse $count non-critical files from account: " .
	    "$fulldir.";
    }
}

#------------------------------------------------------------
# check_accounts    ACCOUNT_AGES_HASH
#------------------------------------------------------------
# For user accounts:
#    Check each account disk usage.  If greater than AccountMax, then
#    a. If logged out, delete the results.
#    b. Else if logged in, warn the user by placing a warning
#       file in their account.
# Global: $topDir - scalar - full path of directory of accounts.
# Input:  %accountAges - hash - table of age in seconds since last
#                                access for each account.  Account
#                                names are keys.
#------------------------------------------------------------
sub check_accounts {
    my (%accountAges) = @_;

    my @accounts = sort (keys (%accountAges));
    my ($account, $fulldir, %biggies);
    my $size = 0;

    # Check each directory:
    foreach $account (@accounts) {
	$fulldir = "$topDir/$account";

	if (too_big ($fulldir, $IU::AccountMax, \$size)) {
	    $biggies {$account} = $size;

	    # Account too big.
	    # If the account is logged out, delete results.

	    if (!(-e "$fulldir/persistent_user.lock")) {

		delete_results ($account, 'SPACE');
	    
	    } else {
                # If the account is logged in, warn the user by placing
		# a warning file in their account.
		inform "$account too big but logged in, warning user.";
		if (open MSG, ">$fulldir/temp_msg.txt") {
		    print MSG <<EOMSG;
Your account is taking up too much computer disk
space on this system.  You can reduce space usage
by emptying your shopping cart and folder,
and by performing a search that returns fewer results.
We apologize for any inconvenience this may cause.


If you get this message repeatedly, Exit your account
and don't sign in for an hour or so.  Your current
results will be purged, so place any orders first.
You can do searches and orders as a guest while
you wait.
EOMSG
                    close MSG;
 		}
	    }
	}  # If account too big.
	# Else check next account.
    }   # foreach loop continues.
    # If any accounts were too big, send e-mail to the admin so 
    # he/she can keep track of typical usage.
    if (%biggies) {
	my @bigs = sort (keys (%biggies));
	my $message =
"There were " . scalar (@bigs) . " accounts that exceeded the disk 
usage limit of $IU::AccountMax.  They were:\n\tAccount   \tDisk Usage
\t-------   \t----------\n";
	foreach $account (@bigs) {
	    $message .= "\t$account \t" . $biggies{$account} . "\n";
	}
	$message .=
"\nYou might consider increasing the usage limit 'AccountMax' in the
SETUP file in the installation directory ($IU::TopDir).
Contact the Web Gateway development team if you have questions.";

	# Create mail object:
	my $mail = new IU::Mail;

	$mail->to($IU::SiteAdminEmail || 'imswww');   # Who to send to.
	$mail->subject(
"IMS Web Gateway Running Out of Individual Account Disk Space");
	$mail->set('Content-Type', 'text/html');
        # Write message:
	my $FH = $mail->open || puke "mail open failed!";
	print $FH $message;
	$mail->send;
    }
}

#------------------------------------------------------------
# check_critical_usage    DISK_USAGE_FLAG  ACCOUNT_AGES_HASH
#------------------------------------------------------------
# For user accounts:
# Re-check the sum of all accounts disk space.  If greater than
#    CriticalAllAccountsMax then we are in trouble.  E-mail the system
#    administrator with the level of problem and how far we got here.
#    a. Delete non-critical files of all logged out accounts.
#    Re-check.  If still greater then:
#    b. Delete results from all logged out accounts.
#    Re-check.  If still greater then:
#    c. For logged in accounts, delete any guide files that haven't
#       been accessed for a period of GuideKillTime.
#    Re-check.  If still greater then:
#    d. Warn all logged in users to logout, explaining why.
# Global: $topDir - scalar - full path of directory of accounts.
# Input:  $DiskUsage - scalar - true if too much disk space was being
#                            used at the end of check_accounts_total().
#         %accountAges - hash - table of age in seconds since last
#                            access for each account.  Account
#                            names are keys.
#------------------------------------------------------------
sub check_critical_usage {
    my ($DiskUsage, %accountAges) = @_;

    # First remove any old LOGOUT Warning messages that were created
    # for a critical disk usage exceeded for all accounts (logged in
    # or out).  The messages will be recreated if there is still a 
    # disk shortage.
    my @accounts = sort (keys (%accountAges));
    my ($account, $level, $size1, $size2, $size3, $size4, $message);
    foreach $account (@accounts) {
	system ("/bin/rm -rf $topDir/$account/space_msg.txt");
    }
    # Only bother doing the checks if disk usage was too big at the end
    # of check_accounts_total() above.

    if (!($DiskUsage)) {
	return;
    }
    inform "Checking if disk usage < critical limit " .
	"$IU::CriticalAllAccountsMax";
    $level = 0;
    # If the critical limit is not exceeded, return.
    if (too_big ($topDir, $IU::CriticalAllAccountsMax, \$size1)) {
	# Delete all noncritical files from logged out accounts.
        # This step was done earlier, but since we are in a critical
        # condition, we may as well try again (someone may have just
        # logged out).
	delete_all_noncritical (%accountAges);
	$level = 1;
	if (too_big ($topDir, $IU::CriticalAllAccountsMax, \$size2)) {
	    # Delete all results files from logged out accounts.
	    delete_all_results (%accountAges);
	    $level = 2;
	    if (too_big ($topDir, $IU::CriticalAllAccountsMax, 
                         \$size3)) {
		# Delete old guide results in active sessions.
		delete_old_guides (%accountAges);
		$level = 3;
		if (too_big ($topDir, $IU::CriticalAllAccountsMax,
                             \$size4)) {
		    $level = 4;
		    inform "Warning users to exit their accounts.";
		    # Warn each account:
		    foreach $account (@accounts) {
			# Warn all users to logout for a while.
			if (open MSG,
                            ">$topDir/$account/space_msg.txt") {
			    print MSG <<EOMSG;
WARNING!!!  There is a CRITICAL shortage of
computer disk space on this system.  The system
administrator will be taking steps to correct
the problem.  You should save any search queries you have
constructed, finish placing any orders you have
started, and EXIT AS SOON AS POSSIBLE.  You may try
back in a few hours, but this problem may reoccur
until the management fixes it.  We apologize for any
inconvenience this may cause.
EOMSG
			    close MSG;
			}
		    }
		}
	    }
	}
    } else {
	# Return if critical limit was not exceeded.
	return;
    }
    $message =
"URGENT WARNING!!! The IMS Web Gateway running on your system
has run out of disk space for user accounts.  The usage
has exceeded the CRITICAL limit set in the file SETUP
in the installation directory ($IU::TopDir).  The limit
CriticalAllAccountsMax is set to $IU::CriticalAllAccountsMax.
You probably need much more disk space for the number of
Web Gateway user accounts you have: " . scalar (@accounts) . "

The accounts disk usage was: $size1 kb.
After deleting non-critical files from logged out accounts,
the accounts disk usage was: $size2 kb\n";
    if ($level >= 2) {
	$message .= "
After results were deleted from logged out accounts;
the accounts disk usage was: $size3 kb\n";
	if ($level >= 3) {
	    $message .= "
After guide files (older than " . 
nice_time (units2secs ($IU::GuideKillTime)) . 
") were deleted from accounts of users (including those who are
currently SIGNED-IN), the accounts disk usage was: $size4 kb\n";
	    if ($level >= 4) {
		$message .= "
The accounts disk usage still exceeds the limit.
Users were encouraged to exit their accounts.\n";
	    }
	}
	$message .= "
You should increase the critical disk space limit in the SETUP file
and re-run 'setup -apply' as user imswww.
Contact the Web Gateway development team if you have questions.";

	# Create mail object:
	my $mail = new IU::Mail;

	$mail->to($IU::SiteAdminEmail || 'imswww');   # Who to send to.
	$mail->subject("IMS Web Gateway Accounts Exceeded CRITICAL " .
		       "Disk Space");
	$mail->set('Content-Type', 'text/html');
        # Write message:
	my $FH = $mail->open || puke "mail open failed!";
	print $FH $message;
	$mail->send;
    }
}

#------------------------------------------------------------
# delete_all_results  TOP_DIRECTORY    ACCOUNT_AGES_HASH
#------------------------------------------------------------
# For all user accounts that are logged out, delete all
# files except account info and last search.
# Global: $topDir - scalar - full path of directory of accounts.
# Input:  %accountAges - hash - table of age in seconds since last
#                                access for each account.  Account
#                                names are keys.
#------------------------------------------------------------
sub delete_all_results {
    my (%accountAges) = @_;

    my @accounts = sort (keys (%accountAges));
    my ($account, $fulldir);

    inform "Deleting results files from logged out user accounts.";

    # Check each directory:
    foreach $account (@accounts) {
	# Skip accounts that are logged in.
	if (-e "$topDir/$account/persistent_user.lock") {
	    next;
	}
	# Delete the noncritical files in the account:
	delete_results ($account, 'ALLSPACE');
    }
}

#------------------------------------------------------------
# delete_old_guides  ACCOUNT_AGES_HASH
#------------------------------------------------------------
# For all user accounts (even logged in), delete all
# guide files that haven't been accessed in a period of
# time >= GuideKillTime.
# Global: $topDir - scalar - full path of directory of accounts.
# Input:  %accountAges - hash - table of age in seconds since last
#                                access for each account.  Account
#                                names are keys.
#------------------------------------------------------------
sub delete_old_guides {
    my (%accountAges) = @_;

    my @accounts = sort (keys (%accountAges));
    my ($account, $fulldir);

    inform "Deleting old guides.";
    my (@files, $file, $age, $deleted);

    # Check each directory:
    foreach $account (@accounts) {
	# Open directory stream on account directory...
	opendir ACCT, "$topDir/$account" or
	    puke "couldn't open $topDir/$account directory: $!";

	# Get list of all text files.
	@files = grep -T, map "$topDir/$account/$_", readdir ACCT;

	# Get list of text files begining with GuideSearch.
	@files = grep /GuideSearch.+/, @files;

	if (@files) {
	    $deleted = 0;
	    foreach $file (@files) {
		$age = (-M $file)*86400;     # convert days to seconds

		# If guide file is older than GuideKillTime, delete.
		if ($age > units2secs ($IU::GuideKillTime)) {
		    system ("/bin/rm -f $file");
		    $deleted = 1;
		}
	    }
	    # If any were deleted, put a warning in their directory.
	    if ($deleted) {
		if (open MSG, ">>$topDir/$account/temp_msg.txt") {
		    print MSG <<EOMSG;
Because of a critical shortage of computer disk
space on this system, your most recent Detailed
Document Search (Guide) Results have been removed.
You can help reduce space usage by emptying your
shopping cart and folder, and by performing a
search that returns fewer results.  The system
administrator has been notified.
We apologize for any inconvenience this may cause.
EOMSG
		}
		close MSG;
	    }
	}
	closedir DIR;
    }
}

#------------------------------------------------------------
# once CYCLE        -- do once:
#------------------------------------------------------------
sub once {
    my ($cycle) = @_;

    inform("BEGUN cycle $cycle at ", scalar(localtime));    
    cleanup_accounts();
    inform("COMPLETED cycle $cycle at ", scalar(localtime));
    1;
}

#------------------------------------------------------------
# print_setup       -- output the setup
#------------------------------------------------------------
sub print_setup {

    # Print it:
    print STDERR <<EOF;

Setup for imscleanaccounts.pl:

    Manage files under $topDir.

    In loop mode, cycle every $PAUSEMINS minutes.

EOF
    exit(-1);
}

#------------------------------------------------------------
# print_usage        -- usage
#------------------------------------------------------------
sub print_usage {

    # Print it:
    print STDERR <<EOF;

Usage: 
 1. Standalone.  Puts process in infinite loop, $PAUSEMINS min. cycle.
    Less reliable, since signals may kill the process.

    ~imswww/pub/lib/bin/imscleanaccounts.pl -loop

 2. Run once.  Suitable command for "crontab" entry.  Best bet.

    ~imswww/pub/lib/bin/imscleanaccounts.pl -cron
 or
    ~imswww/pub/lib/bin/imscleanaccounts.pl -once
   
 3. Current setup.
    ~imswww/pub/lib/bin/imscleanaccounts.pl -setup

** Remember to run me with the same uid as the owner of the session
files, which is the same uid that CGI scripts run with
(usually "wwwuser")!

EOF
    exit(-1);
}

#------------------------------------------------------------
# init               -- init the application
#------------------------------------------------------------
sub init {

    # Set for debugging:
    $IU::MsgLevel = $IU::MsgLevel = 2;

    # Stop immediately if we look like the wrong user:
    if ((-e "$IU::DocsDir/imswww-usr/cache") &&
	(! -w "$IU::DocsDir/imswww-usr/cache")) {
	puke 554, "No permission to do my job... ",
	          "am I running as the wrong user?";
    }

    # Open a tty log stream:
    IU::Log::init_term(\*STDOUT);
    
    # Open a file log stream:
    my %now = IU::Util::timeparts(localtime());
    my $logpath = sprintf(
        "${IU::TopDir}/log/cleanacts-%04d-%02d-%02d.log",
			  $now{Yr}, $now{Mon}, $now{Day});
    my $logger = IU::ExclFileLogger->open(">>$logpath", 'NONBLOCK') 
	or puke "I can't log to $logpath, so I'm not running: $!.";
    $logger->autoflush();
    IU::Log::set_handler('CLEANUP', $logger);
    1;
}

#------------------------------------------------------------
# main
#------------------------------------------------------------
sub main {
    $ARGV[0] or ($ARGV[0] = '');

    # If the persistent directory doesn't end in a / add one.
    if ("${IU::PersistentDir}" !~/\/$/) {
	$topDir = "${IU::PersistentDir}/imswww-usr";
    } else {
	$topDir = "${IU::PersistentDir}imswww-usr";
    }

    # Check invocation:
    if ($ARGV[0] =~ /^-loop$/) {       	
	eval "use sigtrap";
	init() or die "couldn't init";
	my $cycle = 0;
	while (1) {
	    once(++$cycle);
	    sleep($PAUSEMINS * 60);     # 15 min cycle
	}
    }
    elsif ($ARGV[0] =~ /^-(once|cron)$/) {
	init() or die "couldn't init";
	once(0);
    }
    elsif ($ARGV[0] =~ /^-(setup)$/) {
	print_setup();
    }
    else {
	print_usage();
    }
    exit(0);
}
exit(main() ? 0 : -1);

#------------------------------------------------------------
1;
