#!/u/joe/perl561/bin/perl -w

=head1 NAME

imsfixaccounts.pl
   - periodic repair of the persistent user accounts.

=head1 DESCRIPTION

Go through the accounts in the persistent user directory and

=over 4 

=item 1.

Delete any empty ("::") entries in the password file.

=item 2.

Delete any entries in the password file that do not correspond to 
an account.

=item 3.

Fix any mislinked renamed directories.

=item 4.

Update the email addresses associated with each remaining valid user.

=back

Script must be run by wwwuser, usually in a cronjob.

Usage:
    ~imswww/pub/lib/bin/imscleanaccounts.pl

See print_usage or try it without any arguments.

=head1 WARNING

Assumes that it is being run out of the ./lib/bin directory
BUT use the full-path to this script!  There should only be
one instance of this script running per persistent user directory.

The update process is lengthy; it should be run no more than once or
twice in a day. 

The program automatically takes the system down via "down.txt"; 
running it during working hours could make you very unpopular.

=head1 AUTHOR

Joe McMahon, Raytheon
Based on imscleanacct by Larry Rosen.

=cut

# GLOBAL
my $topDir;  # This is the directory containing all the user accounts.
             # Don't confuse with $IU::TopDir which heads the
             # installation.

use Cwd;

# Set up include path:
BEGIN {
    # Figure out top directory:
    my $path = (($0 =~ m|^/|) ? $0 : (cwd()."/$0"));
    ($path =~ m|/lib/bin/[^/]+$|) or
       die "can't extract info from $path!";   
    my $perltop = "$`/lib/perl";

    # Set up include path:
    unshift @INC, "$perltop/first", "$perltop", "$perltop/plus",
                  "$perltop/CPAN";
    push @INC, "$perltop/fallback";
}

use IU;
use IU::Log;
use IU::Util qw(units2secs units2bytes nice_time);
use IU::ExclFileLogger;
use IU::UserAdmin;

use strict;

#------------------------------------------------------------
# fix_accounts
#------------------------------------------------------------
# Fix all user accounts that need it.

sub cleanup_accounts {
    inform "Cleaning user accounts:";

    my $pw = IU::UserAdmin->new("$IU::PersistentDir/.htpasswd");
    $pw -> attach_emails(Verbose=>1);

    1;
}

#------------------------------------------------------------
# print_setup       -- output the setup
#------------------------------------------------------------
sub print_setup {

    # Print it:
    print STDERR <<EOF;

Setup for imscleanaccounts.pl:

    Manage files under $topDir.

EOF
    exit(-1);
}

#------------------------------------------------------------
# print_usage        -- usage
#------------------------------------------------------------
sub print_usage {

    # Print it:
    print STDERR <<EOF;

Usage: 

 1. Run once.  Suitable command for "crontab" entry.  Best bet.

    ~imswww/pub/lib/bin/imscleanaccounts.pl -cron
 or
    ~imswww/pub/lib/bin/imscleanaccounts.pl -once
   
 2. Current setup.
    ~imswww/pub/lib/bin/imscleanaccounts.pl -setup

** Remember to run me with the same uid as the owner of the session
files, which is the same uid that CGI scripts run with
(usually "wwwuser")!

EOF
    exit(-1);
}

#------------------------------------------------------------
# init               -- init the application
#------------------------------------------------------------
sub init {

    # Set for debugging:
    $IU::MsgLevel = $IU::MsgLevel = 2;

    # Stop immediately if we look like the wrong user:
    if ((-e "$IU::PersistentDir") && (! -w _)) {
	puke 554, "No permission to do my job... ",
	          "am I running as the wrong user?";
    }

    # Open a tty log stream:
    IU::Log::init_term(\*STDOUT);
    
    # Open a file log stream:
    my %now = IU::Util::timeparts(localtime());
    my $logpath = sprintf(
        "${IU::TopDir}/log/fixacts-%04d-%02d-%02d.log",
			  $now{Yr}, $now{Mon}, $now{Day});
    my $logger = IU::ExclFileLogger->open(">>$logpath", 'NONBLOCK') 
	or puke "I can't log to $logpath, so I'm not running: $!.";
    $logger->autoflush();
    IU::Log::set_handler('CLEANUP', $logger);
    1;
}

#------------------------------------------------------------
# main
#------------------------------------------------------------
sub main {
    $ARGV[0] or ($ARGV[0] = '');

    # If the persistent directory doesn't end in a / add one.
    if ("${IU::PersistentDir}" !~/\/$/) {
	$topDir = "${IU::PersistentDir}/imswww-usr";
    } else {
	$topDir = "${IU::PersistentDir}imswww-usr";
    }

    # Check invocation:
    if ($ARGV[0] =~ /^-(once|cron)$/) {
	init() or die "couldn't init";
	cleanup_accounts(0);
    }
    elsif ($ARGV[0] =~ /^-(setup)$/) {
	print_setup();
    }
    else {
	print_usage();
    }
    exit(0);
}
exit(main() ? 0 : -1);

#------------------------------------------------------------
1;

