package IU::AccountStatusRequest;


=head1 NAME

IU::AccountStatusRequest - request object for account status.


=head1 SYNOPSIS

Search properties:
    
    # What kind of search is this?  Can return any of...
    #     ACCOUNT_STATUS_REQUEST
    $q->search_type;
    
    # What kind of ODL search message is this *really*?  Can return any of...
    #     ACCOUNT_STATUS_REQUEST
    $q->ODL_search_type;

Search categories/criteria:

    # What categories has the user given constraints in?
    @categories = $q->categories;
    
    # Get user's choices for this criterion, as "criterion" objects:
    @criteria = $q->criteria("PARAMETER");
    
    # Set criteria information, using an array of "criterion" objects:
    $q->criteria("PARAMETER", \@params);

Dependent valids:

    # What DAACS will we search on, based on dependent valids?
    $q->determine_daacs_to_search;
    $q->daacs_to_search;

ODL interface:

    # ODL interface:
    $q = IU::AccountStatusRequest->from_ODL($odl);
    $odl = $q->to_ODL;
     
Output as HTML:

    # Dump pretty text (note: takes an HTML stream!!!):
    $q->print_HTML($HTML)


=head1 DESCRIPTION

This object encapsulates all kinds of queries: inventory, directory,
guide, etc.


=head1 VERSION


=cut

use IU;
use IU::Log;
use IU::Category;
use IU::Criterion;
use IU::IW;
use IU::ArchiveOdl;
use IU::WWWValids;

use IW::App;
use IW::User;
use IW::DepValid;
use IW::Widget::DepValids;

use strict;
use vars qw(@ISA @EXPORT
            $QUERY
            $MAX_TIME_DEFAULT);
@ISA = qw(Exporter);
@EXPORT = qw($QUERY);

#------------------------------
#
# Globals...
#
#------------------------------

# Default time limit for searches
$MAX_TIME_DEFAULT = $IU::SETUP::MaxSearchTime;
# The current query (EXPORTED for the convenience of other modules):
$QUERY = undef;

#------------------------------
#
# Private utilities...
#
#------------------------------

#------------------------------------------------------------
# _one VARNAME,SELF,[VALUE]
#------------------------------------------------------------
# I<Private utility.>  Get/set a single-valued instance variable.
# Return undef if it's undefined.
sub _one {
    my ($varname, $self, $value) = @_;
    $self->{$varname} = $value if (@_ > 2);
    $self->{$varname};   
}

#------------------------------------------------------------
# _many VARNAME,SELF,[ARRAYREF]
#------------------------------------------------------------
# I<Private utility.>  Get/set a multi-valued instance variable.
# Set it by providing an arrayref: the array will be duplicated.
#
# If invoked with no args, just returns info.
# Returns the array of all values in an array context, and just 
# the array size in a scalar context.

sub _many {
    my ($varname, $self, $values) = @_;
    if (@_ > 2) {
	my @array = @$values;
	$self->{$varname} = \@array;
    }
    (wantarray ? @{$self->{$varname}} : int(@{$self->{$varname}}));
}

#
# Construction...
#
#------------------------------

#------------------------------------------------------------
# new 
#------------------------------------------------------------
# Class method.  Constructor.

sub new {    
    my $class = shift;
    my $self = {};
    $self->{SearchType} = 'ACCOUNT_STATUS_REQUEST';
    $self->{Categories}    = {};
    bless $self, $class;
}

#------------------------------
#
# Known categories...
#
#------------------------------



#------------------------------------------------------------
# Instance methods.
# Convenient ways to get/set the array of valids for core      .
# All of these accept an optional arrayref of strings to set, and 
# return an array of strings.
#
sub data_center_ids   { my $self = shift; 
			$self->criteria_strings('DATA_CENTER_ID',@_)}

#------------------------------
#
# Search options/properties...
#
#------------------------------

#------------------------------------------------------------
# max_time
#------------------------------------------------------------

sub max_time {
    my ($self, $value) = @_;
    $self->{MaxTime} = $value if @_ > 1;
    $self->{MaxTime} || $MAX_TIME_DEFAULT;
}

#------------------------------------------------------------
# name [NAME]
#------------------------------------------------------------
# Set/get the name of the query.
#
sub name  { _one('Name', @_) }

#------------------------------------------------------------
# filename
#------------------------------------------------------------
# Get a good filename for the query.
#
sub filename {
    my $self = shift;
    my $name = $self->name || 'untitled';
    $name =~ s/^\s+|\s+$//g;        # trim
    $name =~ s/[^a-z0-9_+-]/_/gi;   # turn illegals into underscores
    $name =~ s/_+/_/g;              # collapse underscores
    "$name.query";
}

#------------------------------------------------------------
# ODL_search_type
#------------------------------------------------------------
# Return the search type, as it would appear in the ODL.
#
sub ODL_search_type { 
    my $self = shift;
    return $self->search_type;
}

#------------------------------------------------------------
# search_type [VALUE]
#------------------------------------------------------------
# Set/get the search type, in the most-detailed sense.
# Compare with ODL_search_type.
#
sub search_type { _one('SearchType', @_) }

#------------------------------------------------------------
# criteria CATEGORY, [CRITERIA]
#------------------------------------------------------------
# Instance method, for convenience.
#
# With just CATEGORY, returns array of criteria (or empty array if none).
#
# With CATEGORY and CRITERIA, sets CATEGORY's criteria to be the
# given CRITERIA, which should be a ref to an array of Criterion objects.
# If you supply a non-arrayref, it will be assumed to be the sole criterion.  
#
# Use a CRITERIA of undef (or a ref to the empty array) to erase all 
# criteria for this category (and the category as well).

sub criteria {
    my ($self, $cat, $values) = @_;
    if (@_ > 2) {             # set new values
	my @safe;             # safe place for new values

	# Coerce values to be either empty, or array of one criterion:
	$values ||= [];                          # undef/false to empty array
	@safe = ((ref($values) eq 'ARRAY') ? @$values : ($values));
	
	# Set:
	(@safe ? $self->{Categories}{$cat} = \@safe 
	       : delete $self->{Categories}{$cat});
    }
    @{$self->{Categories}{$cat} || []};         # return
}

#------------------------------------------------------------
# criteria_strings CATEGORY, [STRINGS]
#------------------------------------------------------------
# Instance method, for convenience.
#
# With just CATEGORY, returns array of values (or empty array)
# held in the first criterion, which is assumed to be a "SelectMany".
#
# With CATEGORY and STRINGS, where STRINGS is a ref to an array of strings,
# creates a new "SelectMany" criterion with the given values, and stores 
# it as the only criterion for this category.

sub criteria_strings {
    my ($self, $cat, $strings) = @_;
    my $crit;            # note: initially undefined!

    # Set new values?
    if (@_ > 2) {    
	if (@$strings) {       # some... create and set criteria!
	    $crit = IU::Criterion::SelectMany->new;
	    $crit->field('VALUES', $strings);
	}
	$self->criteria($cat, $crit);
    }

    # Return them:
    $crit = ($self->criteria($cat))[0];          # get first criterion
    return ($crit ? @{$crit->field('VALUES') || []} : ()); 
}

#------------------------------------------------------------
# fixup
#------------------------------------------------------------
# After reading in a query from a foreign source - a file, a set
# of CGI tags, etc. - we may need to adjust some things:
#
#    * Default values may need to be filled in
#    * A fixup() message is sent to the spatial search criterion
#    * etc.
#
# Returns the "self" object on success.

sub fixup { 
    my $self = shift;

    puke "fixup is no longer used";

    # Done!
    $self;
}

#------------------------------------------------------------    
# what_now
#------------------------------------------------------------    
# I<Instance method.>
# Based on the search type, are we missing some required fields?
# Returns an array of items still needed.

sub what_now {
    my $self = shift;
    my @needed = ();
    push @needed, 'a data center' unless $self->data_center_ids;
    return @needed;
}

#------------------------------------------------------------    
# incomplete
#------------------------------------------------------------    
# I<Instance method.>
# Based on the search type, are we missing some required fields?
# If so, returns a string (true) describing what's missing.
# If not, returns 0 (false).

sub incomplete {
    my $self = shift;

    unless ($self->data_center_ids) {
	    return "You must supply at least one Data Center.";
    }
    return 0;
}

#------------------------------------------------------------    
# install
#------------------------------------------------------------    
# I<Instance method.>  WG-specific.
# Re-install this as the current account status request in the session state.

sub install {
    my $self = shift;
    $SESSION->state->{AccountStatusRequest} = $self;
}

#------------------------------------------------------------    
# lookup
#------------------------------------------------------------ 
# I<Class method.  WG-specific.>
# Lookup current account status request in the session state.  
# Returns undef if none so far.

sub lookup {
    my $class = shift;
     
    return $SESSION->state->{AccountStatusRequest};
}


#------------------------------------------------------------
# reload PATH
#------------------------------------------------------------
# recreate query using the supplied ODL file

sub reload {

    my ($self, $path) = @_;
    if (! -s $path) {
       puke "no account status request file: $path";
    }
    # Parse it into an ODL object:
    open ODL, $path or puke "open $path: $!";
    my $odl = eval { parse IU::ODL \*ODL };
    close ODL;
    $odl or puke "couldn't parse ODL in $path (line $.)";

    # Convert the ODL into a query object:
    $self->from_ODL($odl,KeepType=>'1')
       or puke "couldn't extract valid user data from ODL";

    # Done!
    return $self;
}

#------------------------------------------------------------
# uses CATEGORY, [SEARCHTYPE]
#------------------------------------------------------------
# I<Class/instance method.>
# Returns true if the given CATEGORY would be a functional part of the 
# query under the given SEARCHTYPE, and false otherwise.   With
# no SEARCHTYPE, will use the query's search type.

sub uses {
   my ($self, $cat) = @_;
   return 1 if ($cat eq 'DATA_CENTER_ID');
   return 0;
}

#------------------------------------------------------------
# uses_opt OPTION, [SEARCHTYPE]
#------------------------------------------------------------
# I<Class/instance method.>
# Much like uses(), but for miscellaneous query options that aren't
# really categories.  Valid OPTIONs are:
#
#     QUERY_NAME       Query name
#
# Returns true if the given OPTION would be a functional part of the 
# query under the given SEARCHTYPE, and false otherwise.   With
# no SEARCHTYPE, will use the query's search type.
# Always show guide limit since it will take effect if the user selects
# Detailed Info while doing any kind of search.

sub uses_opt {
    my ($self, $opt, $type) = @_;
    
    return 1 if ($opt eq 'QUERY_NAME');
}

#------------------------------------------------------------    
# validate
#------------------------------------------------------------    
# I<Instance method.>
# Based on the search type, are we ready to go?
# This is an assertion: it returns the self object on success, but on failure
# it raises a fatal exception which describes why validation failed. 

sub validate {
    my $self = shift;

    # Is user not done?
    my $problem = $self->incomplete;
    puke 516, $problem if $problem;
    $self;
}

#------------------------------
#
# ODL interface
#
#------------------------------

#------------------------------------------------------------
# from_ODL ODLOBJECT, PARAMHASH
#------------------------------------------------------------
# I<Class or instance method.>
# Initiallize the address from a standard IMS ODL object.
#
# The PARAMHASH can contain the following:
#

sub from_ODL {
    my ($self, $odl, %params) = @_;
    $odl || return undef;
    my ($param, $value);
    my $search;

    # Be a class method, too!
    ref($self) or $self = $self->new;
    
    $self->search_type('ACCOUNT_STATUS_REQUEST');
    # 
    # EXTRACT THE SEARCH CONSTRAINTS FROM THE ODL...
    #
    $search->rewind;

    while (($param, $value) = $search->next) {
       $self->criteria_strings($param, [$value->text_values])
          if ($param eq 'DATA_CENTER_ID');
    }    

    $self;
}

#------------------------------------------------------------
# to_ODL PARAMHASH
#------------------------------------------------------------
# Convert query to an ODL object, and return it.
# The ODL object is a group *containing* an ACCOUNT_STATUS_REQUEST group
#
# No authentication information is placed in the ODL!!!!
#
# PARAMHASH may contain the following:
#
sub to_ODL {
    my ($self, %params) = @_;
    my $no_daacs         = $params{NoDaacs};

    require IU::ODL;

    # Begin:
    my $root_odl = IU::ODL->new_group;    # outermost group
    my $srch_odl = IU::ODL->new_group;

    # Start off the file with the search type:
    $root_odl->add_param($self->ODL_search_type => $srch_odl);

    # The message id (do we REALLY need this?)
    $srch_odl->add_param(MESSAGE_ID    => '0'); 

    my $persistent_user = $ENV{REMOTE_USER};
    $srch_odl->add_param(DORRAN_USERNAME => IU::ODL->quoted($persistent_user))
       if($persistent_user ne 'guest');

    my $user = $SESSION->state->{User};
    if ($user) {
       $srch_odl->add_param(CONTACT_ADDRESS  => $user->req_addr->to_ODL)
          if $user->req_addr;
    }

    $srch_odl->add_param(DATA_CENTER_ID=>IU::ODL->conv($self->data_center_ids));

    # Max search time.
    $srch_odl->add_param(TIME_LIMIT => IU::ODL->number($self->max_time));

    # Done!
    $root_odl;
}

#------------------------------------------------------------
# to_search_ODL
#------------------------------------------------------------
# Convert query to an ODL object suitable for input to the client, 
# and return it. 
#

sub to_search_ODL {
    my $self = shift;

    # Start with the basics, *without* the DAACs:
    my $root_odl = $self->to_ODL || return undef;

    # Get the actual search group, checking all standard types:
    my $srch_odl = $root_odl->param('ACCOUNT_STATUS_REQUEST') ||
		   puke "No recognized search group in ODL!"; 

    # Add the authenticator, if any:
 
    # Add V0 authenticator, if any:
    my $authenticator = 
       eval { $SESSION->state->{User}->authenticator };
    $srch_odl->add_param(AUTHENTICATOR => 
       IU::ODL->conv($authenticator))
       if ($authenticator);

    # Add the NASDA authenticator, if any:
    my $nasda_authenticator = 
       eval {$SESSION->state->{User}->nasda_authenticator};
    $srch_odl->add_param(NASDA_AUTHENTICATOR => 
       IU::ODL->conv($nasda_authenticator))
       if ($nasda_authenticator);

    # Add the ecs authenticator if "ECS Profile SMC" is defined in setup.
    my $ecs_authenticator = 
       eval { $SESSION->state->{User}->ecs_authenticator };
    $srch_odl->add_param(ECS_AUTHENTICATOR => 
       IU::ODL->conv($ecs_authenticator))
       if ($ecs_authenticator && $IU::EcsProfileSmc ne '');

    # Done!
    return $root_odl;
}

#------------------------------------------------------------
# print_HTML HTMLSTREAM
#------------------------------------------------------------
# Output summary as pretty HTML text.

sub print_HTML {
    my $self = shift;
    my $HH = shift;    # HTML stream

    $HH -> UL;
    my ($cat, $val, $crit);
    foreach $cat (sort {category2pretty($a) cmp category2pretty($b) 
			} $self->categories) {
	# Skip this category if not used by current search type:
	$self->uses($cat) or next;

        #Skip time range, geo region or extended category
        #if this is a directory search
        #Can't base it on uses because need that for query.dat to be correct
        if (($self->search_type eq 'DIRECTORY_SEARCH')&&
            ($cat eq 'GEOGRAPHIC_REGION' || $cat eq 'TIME_RANGE'
             || (IU::Category->is_extended($cat)))){next;} 
	# Get criteria:
	my @criteria = $self->criteria($cat); @criteria or next;
	my $pretty_cat = category2pretty($cat);

	$HH -> LI -> t_B($pretty_cat, ": ");
	if (IU::Category->is_singleton($cat)) {   # has one criterion/variant
	    $crit = $criteria[0];
	    $HH -> t($crit->as_string);
	}
        else {                                    # has multiple criteria
	    $HH -> UL;
	    foreach $crit (@criteria) {
	        $HH -> LI -> t_B($crit->name, ": ") -> t($crit->as_string);
	     }
	    $HH -> _UL;
        }
        
    }

    1;
}


#------------------------------------------------------------    
1;
