package IU::Auth;

=head1 NAME

IU::Auth - parallels IK_Auth.c

=head1 DESCRIPTION

Generate authentication strings for outgoing ODL messages.

=head1 AUTHOR

Eryq

=cut

use IU;
use IU::Chunk ();
use IU::Log;
use strict;

#------------------------------------------------------------
#
# Globals
#
#------------------------------------------------------------

# The salt characters:
my $SALT = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789/.';

#------------------------------------------------------------
# make_ecs_authenticator
#------------------------------------------------------------
# PARAMETERS:
# Currently: the first string is the user name and the second
# string is the user key or password, however this may change
# because this is using the first algorithm ECS gave us, which
# maybe replaced with a new algorithm.
# Calls a C program which uses the ECS provided algorithm.
# RETURNS:
# the ECS authenticator

sub make_ecs_authenticator {

    my ($string1, $auth_key) = @_;

    my $authenticator;
    my $chunk;
    my @chunks;

    # Run program, get output:
    my $cmd = "${IU::LibDir}/bin/authenticator".
              " -u ". $string1 . 
              " -p ". $auth_key;

    open RESULT, "$cmd |" or puke 453, "couldn't run ecs authenticator: $!";
   
    while ($chunk = IU::Chunk->input(\*RESULT)) {
        push @chunks, $chunk;
    }

    close RESULT;

    defined($chunks[0]) && ref($chunks[0]) or
        puke 453, "ECS authenticator program didn't work. (It returned " .
            "nothing.  string = $string1, auth key = $auth_key): $!";

    return ($chunks[0]->values('ECS_AUTHENTICATOR'));
}

#------------------------------------------------------------
# make_authenticator
#------------------------------------------------------------
# PARAMETERS:
#   These are 3 arbitrary strings.  Usually, however, string1 is
#   the user's first name, string2 is the last name, and
#   authentication_key is the user key.
#
# NOTES:
#   Authenticator sites require that the first and last name used to generate 
#   the authenticator be the same as the first and last name in the 
#   contact address.  If they are not the same, the authentication will
#   fail.
#
#   The algorithm for creating the authenticator is very simple.  It
#   makes use of the Unix crypt(3) function.  The string1 and string2
#   parameters are used to pick the two salt characters.  This is
#   achieved via a checksum of sorts. The salt characters and the
#   authentication_key parameters are passed to crypt() and the
#   result from crypt is returned by the function.  When string1 and
#   string2 are traversed to create the salt characters, only
#   alphanumeric characters are used.  Additionally, it changes all
#   alphabetic characters to uppercase.  This is done in order to
#   avoid problems with people typing in their names in different
#   ways.
#
# RETURNS
#   The authenticator.
    
sub make_authenticator {
    my ($string1, $string2, $auth_key) = @_;
    my $i;
    my $salt = '..';

    # If the user does not have an authentication key, then simply
    #   zero out the IK_authenticator string.  This will result in a
    #   zero-length string being sent out as the AUTHENTICATOR parameter
    #   in ODL messages:
    return '' if (!defined($auth_key) || ($auth_key eq ''));
    
    # First, we want to strip out any non-alphanumerics in the
    #   authentication key and change everything to the same case:
    my $auth_key_tmp = lc($auth_key);
    $auth_key_tmp =~ s/[^A-Z0-9]//gi;
 
    # Calculate the first salt character:
    $i = 0;
    while ($string1 =~ m/([a-z0-9])/gi) {    # for each alnum in string...
	$i += ord(uc($1));
    }
    substr($salt,0,1) = substr($SALT, ($i % length($SALT)), 1);

    # Calculate the second salt character:
    $i = 0;
    while ($string2 =~ m/([a-z0-9])/gi) {    # for each alnum in string...
	$i += ord(uc($1));
    }
    substr($salt,1,1) = substr($SALT, ($i % length($SALT)), 1);
    
    # Encrypt the key, and return:
    return crypt($auth_key_tmp, $salt);
}

#--------------------------------------------------------------------
# Mode independent methods for checking daacs needing authentication.
#--------------------------------------------------------------------

#--------------------------------------------------------------------
# daacs_needing_v0_auth   \@DAAC_ARRAY
#--------------------------------------------------------------------
# From the list of given daacs, return list of those that require V0
# authenticator.  If no daacs given, return list of all that do.
# Input: ref to array of daacs to check, or none.
# Output: array of daacs that require V0 authenticator.

sub daacs_needing_v0_auth {
   my %daacs = daacs_needing_auth (shift());

   # Return array of daacs (may be undef!)
   return (@{$daacs{V0}});
}

#--------------------------------------------------------------------
# daacs_needing_nasda_auth   \@DAAC_ARRAY
#--------------------------------------------------------------------
# From the list of given daacs, return list of those that require NASDA
# authenticator.  If no daacs given, return list of all that do.
# Input: ref to array of daacs to check, or none.
# Output: array of daacs that require NASDA authenticator.

sub daacs_needing_nasda_auth {
   my %daacs = daacs_needing_auth (shift());

   # Return array of daacs (may be undef!)
   return (@{$daacs{NASDA}});
}

#--------------------------------------------------------------------
# daacs_needing_ecs_auth   \@DAAC_ARRAY
#--------------------------------------------------------------------
# From the list of given daacs, return list of those that require ECS
# authenticator.  If no daacs given, return list of all that do.
# Input: ref to array of daacs to check, or none.
# Output: array of daacs that require ECS authenticator.

sub daacs_needing_ecs_auth {
   my %daacs = daacs_needing_auth (shift());

   # Return array of daacs (may be undef!)
   return (@{$daacs{ECS}});
}

#--------------------------------------------------------------------
# daacs_needing_any_auth   \@DAAC_ARRAY
#--------------------------------------------------------------------
# From the list of given daacs, return list of those that require V0
# or ECS authenticator.  If no daacs given, return list of all that
# do.
# Input: ref to array of daacs to check, or none.
# Output: array of daacs that require V0 or ECS authenticator.

sub daacs_needing_any_auth {
   my %daac_hash = daacs_needing_auth (shift());

   # Note that the two arrays returned may contain the same daac.
   # We only want to list it once, so lets merge as hashes.
   # (Don't merge key 'undef' so only add array if it is defined.
   my %merge;
   map { $merge{$_}++ } @{$daac_hash{V0}};
   map { $merge{$_}++ } @{$daac_hash{ECS}};
   map { $merge{$_}++ } @{$daac_hash{NASDA}};

   my @daacs = keys (%merge);

   # Return array of daacs (may be undef!)
   return (@daacs);
}

#--------------------------------------------------------------------
# daacs_needing_auth   [\@DAAC_ARRAY]
#--------------------------------------------------------------------
# From the list of given daacs, returns two lists.  One of just those
# that need ECS authentication, the other is just those that need V0
# authentication.  Note that a daac can end up in both lists.  If no
# daacs are given, the list of all daacs in wwwvalids.conf will be
# checked.  The two lists are returned via an array of 2 array refs.
# Input: ref to array of daacs to check, or none.
# Output: hash of 2 array references.  Those refer to arrays of daac
#   ID's where the daac needs ECS authentication and the other array
#   is for those that need V0 authentication.  Return looks like:
#   ('V0'=>\@daacsV0, 'ECS'=>\@daacsECS, 'NASDA'=>\@needsNASDA).
#   Any array may be empty (undefined). A daac ID may be in both
#   arrays (if daac needs both).

sub daacs_needing_auth {
   my $daac_ref = shift;

   # Open static valids database:
   my %WWWVALIDS;
   if (-e "${IU::LibDir}/conf/wwwvalids.gdb") {
      eval {tie %WWWVALIDS, 'IU::Store',
            "${IU::LibDir}/conf/wwwvalids.gdb", 'READER'};
      $@ and return error "tie failed: $@";
   }
   # Get list of daacs to check.
   my @daacs;
   if ($daac_ref) {           # List given. Dereference.
      @daacs = @$daac_ref;
   } else {
      # Not given list of daacs.  Go through wwwvalids to make list.
      # Keys of WWWVALIDS are things like "daac:GSFC" and
      # "dataset:AVHRR".  Use "grep" to find all keys that begin with
      # the string 'daac:'.  Take that array and use "map" to apply
      # "substr" to each element, which will return the DAAC ID parts.
      # The following line could've been written the long/slow way:
      #    my @keys = keys (%WWWVALIDS);
      #    my ($key, @things);
      #    foreach $key (@keys) {
      #       push (@things, $key)   if ($key =~ /^daac:/);
      #    }
      #    my $thing;
      #    foreach $thing (@things) {
      #       $daac = substr ($thing, 5);
      #       push (@daacs, $daac);
      #    }

      @daacs = map {substr ($_, 5)}
         grep { $_ =~ /^daac:/} keys (%WWWVALIDS);
   }
   # For each data center we're considering...
   my $archid;       # Each daac.
   my (@needsV0, @needsECS);     # Daacs needing V0,ECS authenticator.
   my @needsNASDA;               # DAACs needing NASDA authenticator.
   foreach $archid (@daacs) {
      # Is there an entry for it in the static valids?
      my $valids;
      if ($valids = $WWWVALIDS{"daac:$archid"}) {
	 push (@needsV0, $archid) if ($valids->{REQUIRES_V0_AUTH});
	 push (@needsNASDA, $archid) if ($valids->{REQUIRES_NASDA_AUTH});
	 push (@needsECS, $archid) if ($valids->{REQUIRES_ECS_AUTH});
      }
   }                                    # end: foreach daac 

   # Close static valids database:
   untie %WWWVALIDS;

   # Return arrays of daacs.
   return ('V0'=>\@needsV0, 'ECS'=>\@needsECS, 'NASDA'=>\@needsNASDA);
}

#--------------------------------------------------------------------
# End of module.
#------------------------------------------------------------
# Execute simple test if run as a script...
{ 
  package main; no strict;
  $INC{'IU/Auth.pm'} = 1;
  eval join('',<main::DATA>) || die "$@ $main::DATA" unless caller();
}
1;           # end the module
__END__

my $lastname = "Ouithaqueue";
my $firstname = "Eryq";
my $authkey = "foobar";

my $auth = IU::Auth::make_authenticator($firstname, $lastname, $authkey);
print "auth = <$auth>\n";

1;
