package IU::Category;


=head1 NAME
    
IU::Category - manage "database" of category names.


=head1 DESCRIPTION

Categories are the different components of a query... PARAMETER,
SOURCE_NAME, etc.  This module tracks information about certain 
special (known) categories.


=head1 VALID CATEGORIES

The standard 7 dependent-valids categories (all of which are lists
of strings, and which have the "standard" ODL names):

    CAMPAIGN
    DATASET_ID
    DATA_CENTER_ID
    PARAMETER
    PROCESSING_LEVEL
    SENSOR_NAME
    SOURCE_NAME

The following non-dependent categories, which nonetheless use 
"standard" ODL names:

    GRANULE_ID_REQ        String list; used in granule id searches.
    TEXT_STRING           String; used in guide searches.

The additional special categories, with new names:

    GEOGRAPHIC_REGION     Geographic constraint: range, polygon, point,
                          path-row, or "global".
    TIME_RANGE            Temporal range: start date and/or stop date.

Finally, any category name in the extended search criteria is viable.
Such names may not conflict with any category name listed above.
    

=head NON-CATEGORIES

The following are "non-categories", but they are used to denote destinations
in the interface, and thus may not be used as a valid category name:

    MAIN_SEARCH_FORM      The main search form
    SEARCH_OPTIONS        The search options
    EXTENDED_SEARCH       Extended search criteria

=cut

use strict;
use vars qw(@ISA @EXPORT);
use IU;
use Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
	     pretty2category
	     category2pretty
	     extended2pretty
	     valid2pretty
	     extvalid2pretty
             category2context
	     category2upperpretty
	     upperpretty2category
	     );


#------------------------------
#
# GLOBALS...
#
#------------------------------

# The category database:
my %Category = 
    (
     'CAMPAIGN' => {
	 Context   =>'campaignsel',
	 Defn      => "The project or study that obtained the data.",
	 DepValid  => 1,
	 HasDI     => 1,
	 HasDefn   => 1,
	 Pretty    => 'Campaign',
	 UpperPretty    => 'CAMPAIGN',
         Display   => 6,     # Order in which to display 
         Name      => 'CAMPAIGN',
	 Sig       =>'pcampaign',
         GuideSummCore => 1,
     },
     'DATA_CENTER_ID' => {
	 Context   =>'daacsel',
	 Defn      => "The place where the data are archived.",
	 DepValid  => 1,
	 HasDI     => 1,
	 HasDefn   => 1,
	 Pretty    => 'Data Center',
	 UpperPretty    => 'DATA CENTER',
         Display   => 4,     # Order in which to display
         Name      => 'DATA_CENTER_ID',
	 Sig       =>'pdaac',
         GuideSummCore => 1,
     },
     'DATASET_ID' => {
	 Context   =>'dsetsel',    # help context
	 Defn      => "A named collection of data/observations.",
	 DepValid  => 1,           # is this one of the 7 core dependent valid?
	 HasDI     => 1,
	 HasDefn   => 1,
	 Pretty    => 'Data Set',  # pretty way to print name
	 UpperPretty    => 'DATA SET',
         Display   => 2,     # Order in which to display
         Name      => 'DATASET_ID',
	 Sig       =>'pdset',      # signature
         NewCore   => 1,
         GuideSummCore => 1,
     },
     'GEOGRAPHIC_REGION' => {
	 Context   =>'geosel',
	 Pretty    => 'Geographic Region',
     },
     'GRANULE_ID_REQ' => {
	 Context   =>'granuleid',
	 Defn      => "Patterns which match the IDs of desired data granules.",
	 Pretty    => 'Data Granule ID',
     },
     'REQUEST_ID' => {
	 Context   =>'requestid',
	 Defn      => "Order Tracking Number provided by the Data Center.",
	 Pretty    => 'Order Tracking Number',#for improved order tracking
     },
     'PARAMETER' => {	 
	 Context   =>'paramsel',
	 Defn      => ("A physical property being measured in the data ".
		       "(e.g., humidity)."),
	 DepValid  => 1,
	 HasDI     => 0,
	 HasDefn   => 1,
	 Pretty    => 'Parameter',
	 UpperPretty    => 'PARAMETER',
         Display   => 1,     # Order in which to display
         Name      => 'PARAMETER',
	 Sig       =>'pparam',      
         NewCore   => 1,
         GuideSummCore => 1,
     },
     'PROCESSING_LEVEL' => {
	 Context   =>'proclevelsel',
	 Defn      => ("Amount of processing performed on the data ".
		       "(e.g., 0 = raw)."),
	 DepValid  => 1,
	 HasDI     => 0,
	 HasDefn   => 1,
	 Pretty    => 'Processing Level',
	 UpperPretty    => 'PROCESSING LEVEL',
         Display   => 7,     # Order in which to display
         Name      => 'PROCESSING_LEVEL',
	 Sig       =>'pproclevel',      
     },
     'PRODUCT_STATUS_REQUEST' => {
	 Context   =>'prodstatreq',
	 Pretty    => 'Order Status Request',
     },
     'SENSOR_NAME' => {
	 Context   =>'sensorsel',
	 Defn      => "An instrument used in gathering the data.",
	 DepValid  => 1,
	 HasDI     => 1,
	 HasDefn   => 1,
	 Pretty    => 'Sensor',
	 UpperPretty    => 'SENSOR',
         Display   => 3,     # Order in which to display
         Name      => 'SENSOR_NAME',
	 Sig       =>'psensor',      
         NewCore   => 1,
         GuideSummCore => 1,
     },
     'SOURCE_NAME' => {
	 Context   =>'sourcesel',
	 Defn      => ("The spacecraft, airplane, etc. the sensor was ".
		       "located on."),
	 DepValid  => 1,
	 HasDI     => 1,
	 HasDefn   => 1,
	 Pretty    => 'Source',
	 UpperPretty    => 'SOURCE',
         Display   => 5,     # Order in which to display
         Name      => 'SOURCE_NAME',
	 Sig       =>'psource',
         GuideSummCore => 1,      
     },
     'TEXT_STRING' => {
	 Context   => 'guidetext',
	 Defn      => "Free text which is used in guide searches.",
	 Pretty    => 'Guide Text',
     },
     'TIME_RANGE' => {
	 Context   =>'datesel',
	 Defn      => "The time period covered by the data.",
     },
     'INVENTORY_SEARCH' => {
         Pretty         => 'Primary Data Search',
     },
     'DIRECTORY_SEARCH' => {
         Pretty         => 'Summary Document Search',
     },
     'GUIDE_SEARCH'     => {
         Pretty         => 'Detailed Document Search',
     },
     'GRANULE_SEARCH'     => {
         Pretty         => 'Data Granule ID Search',
     },
     'BROWSE_SEARCH'     => {
         Pretty         => "$IU::AirsBrowseSearchTitle",
     },
     'LOCAL_GRANULE_SEARCH'     => {
         Pretty         => 'Local Granule ID Search',
     }
     );

# Map pretty names back to category names:
my %CategoryNamed = map {
    ($Category{$_}->{Pretty} || '') => $_
} keys %Category;
delete $CategoryNamed{''};

# Map upper pretty names back to category names:
my %CategoryNamedUpper = map {
    ($Category{$_}->{UpperPretty} || '') => $_
} keys %Category;
delete $CategoryNamedUpper{''};

#------------------------------------------------------------
# category2context VALID
#------------------------------------------------------------
# I<Function, exported.>
# Turn a valid into a selection help context.
#
sub category2context {
   my $cat = shift;

   # Extract the appropriate context from the database.
   my $context = IU::Category->lookup($cat)->{Context};

   # Add 'pg' and return it if we founf a context; else return
   # undef to ensure no false context gets constructed.
   return ($context ? $context . 'pg' : undef);
} 

#------------------------------------------------------------
# category2pretty VALID
#------------------------------------------------------------
# I<Function, exported.>
# Turn a valid into "Mixed Case Like This", for display or menus.
# The mapping MUST be invertable by pretty2category()!
#
sub category2pretty {
    my $cat = shift;

    # Try to find an explicit name:
    my $pretty = IU::Category->lookup($cat)->{Pretty};
    return $pretty if $pretty;
    
    # Try to clean up the category name:
    $cat =~ tr/A-Z_/a-z /;        # downcase
    $cat =~ s/\b(\w)/\u$1/g;      # capitalize leading letters
    $cat;
}
*valid2pretty = \&category2pretty;

#------------------------------------------------------------
# category2upperpretty VALID
#------------------------------------------------------------
# I<Function, exported.>
# Turn a valid into "UPPER CASE LIKE THIS", for display or menus.
# The mapping MUST be invertable by upperpretty2category()!
#
sub category2upperpretty {
    my $cat = shift;

    # Try to find an explicit name:
    my $pretty = IU::Category->lookup($cat)->{UpperPretty};
    return $pretty if $pretty;
}
#------------------------------------------------------------

#------------------------------------------------------------
# extended2pretty VALID
#------------------------------------------------------------
# I<Function, exported.>
# Turn an extended valid into "Mixed Case Like_this", leaving
# underscores in the value, for display.
#
sub extended2pretty {
    my $cat = shift;

    # Try to clean up the category name:
    $cat =~ tr/A-Z/a-z/;        # downcase
    $cat =~ s/\b(\w)/\u$1/g;    # capitalize leading letters
    $cat;
}
*extvalid2pretty = \&extended2pretty;

#------------------------------------------------------------
# get_core
#------------------------------------------------------------
# I<Class method.>
# Return array of all core dependent valids categories.

sub get_core {
    my $class = shift;
    grep { $class->lookup($_)->{DepValid} } (keys %Category);
}

#------------------------------------------------------------
# get_core_in_display_order
#------------------------------------------------------------
# I<Class method.>
# Return array of all core dependent valids categories in the order
# we want to display them

sub get_core_in_display_order {
    my $class = shift;
    my @hat;
    my @ordered_cats;
    my @final_cats;

    @hat = grep {$_->{DepValid} } (values %Category);
    
    sub the_right_way {

        $a->{Display} <=> $b->{Display};

    }

    # get the cats from the hat
    # sorted by display order. 

    @ordered_cats = sort the_right_way @hat;
    @final_cats = map { $_->{Name} } (@ordered_cats);

}


#------------------------------------------------------------
# is_core      CATEGORY
# is_dependent CATEGORY
# is_extended  CATEGORY
# is_singleton CATEGORY
#------------------------------------------------------------
# I<Class methods.>
# Test a category name for membership in certain groups:
#
#     core      == one of the 7 core dependent valids
#     dependent == any kind of dependent valid: core + extended
#     extended  == one of the extended search categories
#     singleton == special categories with one variant
#
# B<WARNING:> Currently, you MUST supply a real query category name!  
# Don't just hand it any old text string to see if you've got a 
# category, because if it doesn't recognize it it'll always assume
# that it's looking at an "extended" category!!!

# Is this one of the 7 core dependent valid categories?
sub is_core { 
    my ($self, $cat) = @_;
    $self->lookup($cat)->{DepValid};
}

# Is this one of the 3 core dependent valid categories?
sub is_new_core { 
    my ($self, $cat) = @_;
    $self->lookup($cat)->{NewCore};
}

# Is this one of the 6 core dependent valid categories needed by 
# guide or directory search?
sub is_guidesumm_core { 
    my ($self, $cat) = @_;
    $self->lookup($cat)->{GuideSummCore};
}

# Is this any kind of dependent valid?
sub is_dependent { 
    my ($self, $cat) = @_;
    $self->is_core($cat) || $self->is_extended($cat);
}

# Is this POSSIBLY one of the extended search criteria?
# It is if it's TRUE and yet UNKNOWN!
sub is_extended {
    my ($self, $cat) = @_;
    $cat and !$Category{$cat};    # KLUDGE
}

# Is this a special category that normally has only one variant?
sub is_singleton { 
    my ($self, $cat) = @_;
    return ($self->is_core($cat) || 
	    ($cat eq 'GEOGRAPHIC_REGION') || 
	    ($cat eq 'TIME_RANGE') ||
	    ($cat eq 'GRANULE_ID_REQ') ||
	    ($cat eq 'TEXT_STRING'));
}

#------------------------------------------------------------
# lookup CATEGORY
#------------------------------------------------------------
# I<Class method.>
# Return info for a dependent-valids category, as a hash reference.
# Returns ref to empty hash if lookup fails, since this might simply
# be an unknown category.

sub lookup {
    my ($class, $cat) = @_;
    $Category{$cat} || {}
}

#------------------------------------------------------------
# pretty2category PRETTY
#------------------------------------------------------------
# I<Function, exported.>
# Turn a prettified valid into "UPPERCASE_LIKE_THIS", for use.
#
sub pretty2category {
    my $pretty = shift;

    # Try to find an explicit inversion:
    my $cat = $CategoryNamed{$pretty};
    return $cat if $cat;
    $pretty;
}
#------------------------------------------------------------

#------------------------------------------------------------
# upperpretty2category UPPERPRETTY
#------------------------------------------------------------
# I<Function, exported.>
# Turn an upper prettified valid into "UPPERCASE_LIKE_THIS", for use.
#
sub upperpretty2category {
    my $pretty = shift;

    # Try to find an explicit inversion:
    my $cat = $CategoryNamedUpper{$pretty};
    return $cat if $cat;
    $pretty;
}

#------------------------------------------------------------
1;
