package IU::Chunk;

=head1 NAME

IU::Chunk - read/write RFC822-style "chunks"


=head1 DESCRIPTION

For reading/parsing C<"parameter:value"> pairs).  
We use these extensively in the IMS/WWW system, since they are easy
to generate and parse.

This package assumes that each chunk is in either this RFC-822-style format
(IMS/WWW pre-1.6, or "flush-left"):

    FIRST_PARAM: line 1 of first value
    SECOND_PARAM: line 1 of first value
     line 2 of first value
     line 3 of first value
    SECOND_PARAM: line 1 of second value
    THIRD_PARAM: line 1 of first value

Or this format (IMS/WWW 1.7, or "indented"):

    --
    FIRST_PARAM: line 1 of first value
    SECOND_PARAM: line 1 of first value
          : line 2 of first value
          : line 3 of first value
    SECOND_PARAM: line 1 of second value
    THIRD_PARAM: line 1 of first value

Basically, if the chunk begins with a line matching /^--/ (illegal in 
flush-left), it indicates that leading whitespace up to and including an 
optional ":" is to be discarded.  We then go on to process the remaining 
string, which will then be in "flush-left" format.  
I<This allows us to be nicely backwards-compatible.>

As you can see, newlines are preserved in values via the special 
"continuation" line syntax: a flush-left line which begins with a 
space (or tab) is regarded as continuing the value in the previous line.

As of IMS/WWW 1.7, tags are converted to all-uppercase-and-underscores before 
storing them.  This is consistent with IMS data dictionary conventions.

Multiple chunks in the same stream are separated by one or more
EMPTY or COMMENT lines, where an EMPTY line has nothing in it
but the carriage return, and a COMMENT line begins with a '#'
in the first column.


=head2 Substructures

You can nest groups inside of chunks.  This is most easily shown with
"indented" syntax (though it works with flush-left as well):

    --
    NAME: John 
    AGE: 21
    WEIGHT: 17.5
    BEGIN ADDRESS
       STREET: 123 Main
             : Apt. 1
       CITY: Chicago
       STATE: IL
    END ADDRESS
    BEGIN ADDRESS
       STREET: 456 Claremont
       CITY: Greenbelt
       STATE: Maryland
    END ADDRESS

This creates chunks inside the chunk.  They are parsed recursively.
Use it like this:

   my @addrs = $chunk->values('address');
   my $street = $addrs[0]->value('street');     # 123 Main

I<Do not put blank lines after the END... you'll break up the chunk!>


=head2 Sequences

To allow representation of an odd feature in ODL, we support "sequences":
basically, this is a special kid of scalar value whose components are
separated by the string "\001\n".  Notice that it is possible for any
given parameter to have multiple values, each of which might in turn
be a "sequence".


=head2 Case- and punctuation-folding

Storage/access of attributes is done in a case-folded and 
hyphen/underscore-folded way; the following are all equivalent:

   $chunk->value('Entry-Id');
   $chunk->value('entry-id');
   $chunk->value('ENTRY-ID');
   $chunk->value('ENTRY_ID');

This should eliminate a few accidents from your code.


=head2 Indentation

All leading whitespace is stripped from a chunk line before any
processing is done.  That means you can indent as you like.


=head1 NOTES

Each parameter's value is stored as an anonymous array.

To allow arbitrary indentation, we are no longer RFC-822 
compliant as of RCS version 1.13.  This is not expected to be a
major problem.


=head1 AUTHOR

Eryq / Hughes STX Corporation


=head1 RCS INFO

$Id: Chunk.pm,v 1.17 2002/09/12 17:06:25 joe Exp $

=cut

use IU::Log;

# To make Perl happy...
DESTROY { 1; }

use strict;



#------------------------------------------------------------
# new
#------------------------------------------------------------
# Constructor; create a new, empty record.

sub new {
    my $class = shift;
    bless {}, $class;
}

#------------------------------------------------------------
# input FILEHANDLE, [OPTS]
#------------------------------------------------------------
# I<Class or instance method>.
# Skip comments and empty lines, and read next record from FILEHANDLE
# into the internal data structure.
#
# The OPTS are for internal use only.
#
# WARNING: Testing this function interactively may yield stange results.
# Give it a test files.
# 
# RETURNS: the object on success, 0 for EOF, undef for failure.

sub input {
    my ($self, $in, $opts) = @_;
    my ($key, $value);
    local($/) = "\n";		# just to be safe
    my $line = '';
    my $values;
    $opts ||= {};               # default
    
    # Construct if necessary:
    ref($self) or $self = $self->new;
    
    # Skip comments and EMPTY lines until next record (these SHOULD be white):
    while (1) {
	defined($line = <$in>) or return 0;      # expected EOF
	chomp $line;
	next if ($line =~ /^\#/);      # comment
	next if ($line eq '');         # blank line
	last;
    }
    
    # We have a chomped line sitting in queue; the first line of a chunk
    # or a group.  Swallow up an optional syntax specifier: 
    if ($line =~ /^--/) {
	$opts->{Indented} = 1;
	defined($line = <$in>) or return 0;      # unexpected EOF
	chomp $line;
    }

    # We have a chomped line sitting in queue; the first REAL line of a 
    # chunk or a group.  Process until END or blank line.
    while (1) {     

	# Remove "indented" prefix (then we can process as though flush-left):
	$line =~ s/^\s*:?// if $opts->{Indented};

	# Check for END: 
	last if ($line =~ /^END(\s|\Z)/);    # handles empty BEGIN/END groups

	# Extract parts of line:
	if ($line =~ /^([A-Za-z0-9-_]+)[ \t]*:[ \t]*(.*)/) {     
                                              # a "key:" line...
	    ($key, $value) = ($1, $2);              # extract key and value
	    $key =~ tr/a-z-/A-Z_/;                  # fold key
	    ### $value =~ s/^\\n//;                 ### KLUDGE: PATCH ODL BUG
	    $self->{$key} or $self->{$key} = [];    # create array if needed
	    $values = $self->{$key};
	    push(@$values, $value);                 # start storing
	}
	elsif ($line =~ /^[ \t](.*)/) {           # a continuation line...
	    $values or next;
	    $values->[-1] .= ("\n" . $1);
	}
	elsif ($line =~ /^BEGIN\s+(\S+)/) {   # begin sub-structure
	    my $subkey = $1;                        # extract subkey
	    $subkey =~ tr/a-z-/A-Z_/;               # fold subkey
	    my $value = ref($self)->input($in, $opts);
	    $self->{$subkey} or $self->{$subkey} = [];
	    push @{$self->{$subkey}}, $value;
	}
	else {			              # yow!
	    return error "bad record line: <$line>";
	}
    }
    continue {      # next line!
	defined($line = <$in>) or return 0;
	# return error "EOF before blank-line end of record";
	chomp $line;
	last if (($line eq '') || ($line =~ /^END(\s|\Z)/));
    }

    # Done!
    return $self;
}

#------------------------------------------------------------
# keys 
#------------------------------------------------------------
# Return the keys, in no particular order.
#
# The keys will be all-lowercase, with hyphens and underscores
# folded to hyphens; e.g., "like-this".  However, you can still
# access the chunk with keys "LIKE_THIS", "Like-This", etc.

sub keys {
    my $self = shift;
    return (keys(%$self));
}

#------------------------------------------------------------
# value KEY
#------------------------------------------------------------
# Given a key, return the first value, or undef if not present.

sub value {
    my ($self, $key) = @_;
    $key =~ tr/a-z-/A-Z_/;                  # fold key    
    ($self->{$key} ? $self->{$key}[0] : undef);
}

#------------------------------------------------------------
# values KEY
#------------------------------------------------------------
# Given a key, return the LIST of values.

sub values {
    my ($self, $key) = @_;
    $key =~ tr/a-z-/A-Z_/;                  # fold key    
    if (wantarray) {     # array context: return values, or ()
	return ($self->{$key} ? @{$self->{$key}} : ());
    }
    else {               # scalar context: return first value, or undef
	return $self->value($key);
    }
}

#------------------------------------------------------------
# values_ref KEY
#------------------------------------------------------------
# Given a key, return a REFERENCE TO the array of values, or []

sub values_ref {
    my ($self, $key) = @_;
    $key =~ tr/a-z-/A-Z_/;                  # fold key    
    return ($self->{$key} ? $self->{$key} : []);
}

#------------------------------------------------------------
# add KEY, VALUES...
#------------------------------------------------------------
# Add to a KEY's value list the given VALUES.

sub add {
    my ($self, $key) = (shift, shift);
    $key =~ tr/a-z-/A-Z_/;                  # fold key    
    $self->{$key} or $self->{$key} = [];
    push(@{$self->{$key}}, @_);
}

#------------------------------------------------------------
# set KEY, VALUES...
#------------------------------------------------------------
# Set a KEY to a (list of) VALUES.

sub set {
    my ($self, $key) = (shift, shift);
    $key =~ tr/a-z-/A-Z_/;                  # fold key    
    $self->{$key} = [];
    push(@{$self->{$key}}, @_);
}

#------------------------------------------------------------
# unset KEY
#------------------------------------------------------------
# Unset a key, completely removing all values..

sub unset {
    my ($self, $key) = @_;
    $key =~ tr/a-z-/A-Z_/;                  # fold key    
    delete($self->{$key});
}

#------------------------------------------------------------
# print [FILEHANDLE], [LEVEL] 
#------------------------------------------------------------
# Output contents in a readable form.  
# FILEHANDLE defaults to the currently-selected one. 
#
# The LEVEL argument is used to handle indentation.

sub print {
    my ($self, $FH, $level) = @_;
    $FH    ||= select;
    $level ||= 0;    
    my $in   = ('  ' x $level);
    my $cont = '  :';
    no strict 'refs';

    # The syntax specifier:
    print $FH "--\n" unless $level;

    # The data:
    my ($key, $elem);
    foreach $key (sort CORE::keys %$self) {
	foreach $elem (@{$self->{$key}}) {
	    if (ref($elem)) {
		print $FH $in, "BEGIN \U$key\E\n";
		$elem->print($FH, $level+1);
		print $FH $in, "END \U$key\E\n";
	    }
	    else {
		my $temp = $elem;
		$temp =~ s/\n/\n$in$cont /g;
		print $FH $in, "\U$key\E: $temp\n";
	    }
	}
    }
    print $FH "\n" unless $level;
}

#------------------------------------------------------------
# dump [FILEHANDLE] 
#------------------------------------------------------------
# Dump contents, for debugging only!  FILEHANDLE defaults to the 
# currently-selected one. 

sub dump {
    my ($self, $fh, $level) = @_;
    $fh   ||= select;
    $level ||= 0;
    my ($key, $val, $elem);
    my $in = '   ' x $level;

    while (($key, $val) = each %$self) {
	foreach $elem (@{$self->{$key}}) {
	    if (ref($elem)) {
		print "${in}BEGIN $key\n";
		$elem->dump($fh, $level+1);
		print "${in}END\n";
	    }
	    else {
		my $temp = $elem;
		$temp =~ s/\n/\n$in /;
		print "${in}$key= $temp\n";
	    }
	}
    }
    print "\n" unless $level;
}

#------------------------------------------------------------
#
#------------------------------------------------------------
# Milked down version of dump. Used by imsDepVal.cgi to display
# Parameter=value.
# Tried print but didn't work.

sub display {
    my ($self, $fh, $level) = @_;
    $fh   ||= select;
    $level ||= 0;
    my ($key, $val, $elem);
    my $in = '   ' x $level;

    while (($key, $val) = each %$self) {
        foreach $elem (@{$self->{$key}}) {
                print "${in}$key=$elem\n";
            }
        }
}

#------------------------------------------------------------
# first_value CHUNKS, ATTR
#------------------------------------------------------------
# I<Class method.>
# Return first defined value of ATTR in CHUNKS, or undef.

sub first_value {
    my ($self, $chunks, $attr) = @_;
    my $chunk;
    my $value;
    foreach $chunk (@$chunks) {
	last if defined($value = $chunk->value($attr));
    }
    return $value;
}

#------------------------------------------------------------
# seq_value ATTR
#------------------------------------------------------------
# I<Instance method.>
# We have an attribute whose (first) value may have come from either
# a "sequence string" or a plain string.  We want to treat it as though
# it had come from the latter.
#
# Remove any separators ("\001\n") from the first value, and return that. 
# Return undef if not present.

sub seq_value {
    my ($self, $attr) = @_;
    my $value;
    defined($value = defined($self->value($attr))) or return undef;
    $value =~ s/\001\n//g;
    return $value;
}

#------------------------------------------------------------
# seq_values ATTR
#------------------------------------------------------------
# I<Instance method.>
# We have an attribute whose value(s) may have come from a "sequence".
#
# Split all values on the separator ("\001\n")
# and return all the resulting elements.  Return the empty array
# if not present.

sub seq_values {
    my ($self, $attr) = @_;
    return (map { split(/\001\n/, $_) } $self->values($attr));
}

#------------------------------------------------------------
1;
