package IU::ClientChunks;


=head1 NAME

IU::ClientChunks - composite a stream of IMS/WWW application chunks


=head1 DESCRIPTION

We plan to read chunks (see IU::Chunk) in which might be partial in
nature; that is, the information about a dataset might be spread out
over several "dataset" chunks, and some information about a granule
might have to come from the dataset that the granule belongs to (e.g.,
SOURCE_NAME, DATASET_ID).

This module provides some methods for doing such compositing.
It uses a "hold queue" (a hash of chunks by some unique key)
in which the fully-composited chunks are accumulated.


=head1 NOTES

Yes, by putting things in a hash rather than an array, we lose 
the "order" the chunks originally came in... but the whole idea 
behind compositing is to duplicate the important information so 
that "order" no longer matters.  The hash gives us efficient lookup.

Some simplifying assumptions are made because we know something about
the chunks we'll be seeing: they are identical in order and
content to the ODL messages that L<imsclient> will receive.
Thus, one reason we composite is to "de-chunk" the information
from ODL messages.


=head1 AUTHOR

Eryq, Hughes STX Corporation

=cut


use IU::Log;
use IU::Granule;
use IU::Dataset;
use IU::MediaHandlingFee;
use IU::MediaMaximumCapacity;
use IU::Chunk;


#------------------------------------------------------------
# expand_chunk CHUNK, [HELD]
#------------------------------------------------------------
# I<Class method.>
# Expand a single CHUNK, tearing it apart if need be, and putting
# the expanded parts into the HELD hash. 
#
#    Granule chunks are hashed by "g:$gkey".
#    Dataset chunks are hashed by "ds:$dskey".
#
# WARNING: the chunks will be destructively modified, both to
# composite information together, and to tear it apart.
#
# Returns ref to hash of held chunks.

sub expand_chunk {
    my ($self, $chunk, $held) = @_;
    $held ||= {};      # hold queue

    # What is it?
    my $type = (split(/\s+/, ($chunk->value('X-TYPE') || '')))[0];

    # Case: GRANULE chunk...
    if ($type eq 'GRANULE') { 

	# Get unique object key:
	my $gkey    = $chunk->value('GRANULE_WEB_KEY');
	if (!$gkey) {
	    warn "can't expand a granule chunk without a GRANULE_WEB_KEY!";
	    goto DONE;
	}
	#debug "Read granule $gkey from .obj";

	# Some granule-level information may be stored in the dataset:
	my $dskey = IU::Granule::GkeyToDskey($gkey);
	my $dset_chunk = $held->{"ds:$dskey"};
	if ($dset_chunk) {    # we have a dataset chunk: default the info!
	    #debug "Defaulting info for $gkey from $dskey";
	    my $attr;
	    foreach $attr (qw(BROWSE_PRODUCT_DESCRIPTION
			      CAMPAIGN
			      COMMENT
			      DAY_NIGHT
			      DATASET_ID
			      DATA_CENTER_ID
			      GLOBAL_GRANULE
			      MD_ENTRY_ID
                              MISC_URL
			      PACKAGE_ID			      
			      PARAMETER
			      PROCESSING_LEVEL
			      SENSOR_NAME
			      SOURCE_NAME)) {
		if (!defined($chunk->value($attr))) {
		    $chunk->set($attr, $dset_chunk->values($attr));
		}
	    }
	}

	# Store:
	$held->{"g:$gkey"} = $chunk;
    }

    # Case: DATASET chunk...
    elsif ($type eq 'DATASET') { 

	# Get unique object key:
	my $dskey = $chunk->value('DATASET_WEB_KEY');
	if (!$dskey) {
	    warn "can't expand a data set chunk without a DATASET_WEB_KEY!";
	    goto DONE;
	}
	my $key = "ds:$dskey";

	# Extract any granule chunks NOW (we don't want to save these):
	my @gran_chunks = $chunk->values('GRANULE');
	$chunk->unset('GRANULE');

	# Some dataset-level information may be stored in the message:
	my $msg_chunk = $held->{'lastmsg'};
	if ($msg_chunk) {    # we have a dataset chunk: default the info!
	    my $attr;
	    #debug "Defaulting info for $dskey from msg";
	    foreach $attr (qw(DATA_CENTER_ID)) {
		if (!defined($chunk->value($attr))) {
		    $chunk->set($attr, $msg_chunk->values($attr));
		}
	    }
	}

	# Due to ODL-chunking effects, we might have seen this dataset before.
	# If so, just add in the appropriate information to the OLD object:
	if ($held->{$key}) {            # already seen
	    #debug "Read additional data set chunk for $dskey from .obj";
	    foreach $attr (qw(
			      DATASET_HOME_PAGE
			      DATA_URL
			      MD_INFO
			      MISC_URL
			      NUMBER_OF_GRANULE_HITS
			      SPECIALIZED_SEARCH_URL
			      BROWSE_URL
			      STATUS_CODE
			      STATUS_CODE_COMMENT
			      )) {
		my @values = $chunk->values($attr);
		$held->{$key}->set($attr, @values) if @values;
	    }
	}
	else {                          # first time!
	    #debug "Read data set $dskey from .obj";
	    $held->{$key} = $chunk;
	}
	
	# Process each granule chunk:
	my $gran_chunk;
	foreach $gran_chunk (@gran_chunks) {
	    $gran_chunk->set('X-TYPE', 'GRANULE');
	    $self->expand_chunk($gran_chunk, $held);
	}
    }

    # Case: MEDIA chunk...
    elsif ($type eq 'MEDIA_HANDLING_FEE') {

        # Get unique object key:
        my $mkey = $chunk->value('TYPE_ID');
        if (!$mkey) {
            warn "can't expand a media chunk without a TYPE_ID!";
            goto DONE;
        }
        # DATA_SET_ID is stored at the message, get it and set in the 
        # media chunk.
        my $msg_chunk = $held->{'lastmsg'};
        my $archive_id = $msg_chunk->values('DATA_CENTER_ID');
        if ($msg_chunk) { 
           $chunk->set('DATA_CENTER_ID', $msg_chunk->values('DATA_CENTER_ID'));
        }
        my $key = "media:$mkey$archive_id";
        $held->{$key} = $chunk;
    }
    # Case: MEDIA SIZE chunk...
    elsif ($type eq 'MEDIA_MAXIMUM_CAPACITY') {

        # Get unique object key:
        my $mskey = $chunk->value('TYPE_ID');
        if (!$mskey) {
            warn "can't expand a media size chunk without a TYPE_ID!";
            goto DONE;
        }
        # DATA_SET_ID is stored at the message, get it and set in the
        # media size chunk.
        my $msg_chunk = $held->{'lastmsg'};
        my $archive_id = $msg_chunk->values('DATA_CENTER_ID');
        if ($msg_chunk) {
           $chunk->set('DATA_CENTER_ID', $msg_chunk->values('DATA_CENTER_ID'));
        }
        my $key = "mediasize:$mskey$archive_id";
        $held->{$key} = $chunk;
    }
    # Case: INVENTORY or DIRECTORY RESULTS chunk...
    elsif (($type eq 'INVENTORY_RESULT') ||  
	   ($type eq 'DIRECTORY_RESULT')) { 

	# Extract any dataset chunks:
	my @dset_chunks = $chunk->values('DATASET');
	$chunk->unset('DATASET');

        # Extract any media fee chunks:
        my @media_chunks = $chunk->values('MEDIA_HANDLING_FEE');
        $chunk->unset('MEDIA_HANDLING_FEE');

        # Extract any media size chunks:
        my @mediasize_chunks = $chunk->values('MEDIA_MAXIMUM_CAPACITY');
        $chunk->unset('MEDIA_MAXIMUM_CAPACITY');

	# Hold this chunk:
	$held->{"lastmsg"} = $chunk;

	# Process each dataset chunk:
	my $dset_chunk;
	foreach $dset_chunk (@dset_chunks) {
	    $dset_chunk->set('X-TYPE', 'DATASET');
	    $self->expand_chunk($dset_chunk, $held);
	}
        # Process each media fee chunk:
        my $media_chunk;
        foreach $media_chunk (@media_chunks) {
           $media_chunk->set('X-TYPE','MEDIA_HANDLING_FEE');
           $self->expand_chunk($media_chunk,$held);
        }       
        # Process each media size chunk:
        my $mediasize_chunk;
        foreach $mediasize_chunk (@mediasize_chunks) {
           $mediasize_chunk->set('X-TYPE','MEDIA_MAXIMUM_CAPACITY');
           $self->expand_chunk($mediasize_chunk,$held);
        }
    }

    # Case: ROOT chunk...
    elsif (($type eq 'ROOT')) { 
	
	# Determine the kind of search results in this root chunk.
	# Then expand that subchunk appropriately:
	foreach (qw(INVENTORY_RESULT 
		    DIRECTORY_RESULT
		    FTP_BROWSE_RESULT
		    INTEGRATED_BROWSE_RESULT
		    PRODUCT_RESULT)) {
	    my $msg_chunk;
	    if ($msg_chunk = $chunk->value($_)) {
		$msg_chunk->set('X-TYPE', $_);
		$self->expand_chunk($msg_chunk, $held);
		last;
	    }
	}
    }

    # Otherwise: skip it:
    else {
	warn "ignoring strange .obj chunk (type = $type)!";
    }

    # Done!
  DONE:
    return $held;
}

#------------------------------------------------------------
# expand_chunks CHUNKS
#------------------------------------------------------------
# I<Class method.>
# Expand the given array of chunks.  Return expanded chunks
# on a hash:
#
#    Granule chunks are hashed by "g:$gkey".
#    Dataset chunks are hashed by "ds:$dskey".
#
# WARNING: some of the hunks may be destructively modified.

sub expand_chunks {
    my ($self, $chunks) = @_;
    my %held;
    my $chunk;
    foreach $chunk (@$chunks) {
	$self->expand_chunk($chunk, \%held);
    }    
    return \%held;
}

#------------------------------------------------------------
1;
