package IU::Coord;

=head1 NAME

IU::Coord - manage the parsing of global coordinates (degress-minutes-seconds)

=head1 SYNOPSIS

    use IU::Coord;
     
    $degs = str2deg("30 d, 30 m");   # $degs is now 30.5
   
    $dms = deg2dms(45.5002);         # $dms is now "45:30:01"


=head1 AUTHOR

Original code by John Kodis, Hughes STX Corporation.

Organized into modular form by Eryq, Hughes STX Corporation.

=head1 RCS INFO

$Id: Coord.pm,v 1.4 2002/09/12 17:04:55 joe Exp $

=cut

use Exporter;
use IU;
use IU::Log;

use strict;
use vars qw(@ISA @EXPORT);

@ISA = qw(Exporter);
@EXPORT = qw(str2deg deg2dms str2dms);


#------------------------------------------------------------
#
# Globals...
#
#------------------------------------------------------------

# Should we debug?  If debugging, error messages are sent to stdout
# when errors are detected.  If debugging is turned off, an undef
# result is returned to indicate that an error has been detected, and
# the calling routine gets to handle it.

my $DEBUG = 0;


#------------------------------------------------------------
# str2deg STRING
#------------------------------------------------------------
# Converts numerous types of angular position strings into decimal degrees.
#
sub str2deg {
    my ($str) = @_;
    my ($hemi, $deg, $min, $sec, $angle);

    # Strip leading blanks, then look for a leading hemisphere
    # indicator.

    defined($str) or $str = '';    # prevent those warnings
    $str =~ s/^\s+//;
    if ($str =~ /^([nsew+-])(.*)/i) {
	$hemi = $1;
	($str = $2) =~ s/^\s+//;
    }

    # Look for an angle specification, starting with the most
    # difficult to match (e.g., least likely to be matched
    # inadvertently) pattern first.  The regex's must be set up to
    # leave the degree, minute, and seconds strings in $1-$3.
    my $leftovers = "";
    if ($str =~ /^(\d{2,3})\.(\d{2})\.(\d{2})(.*)/) {
	$leftovers = $4; # Matches ``123 34 56''
    } elsif ($str =~ /^(\d{2,3})\s+(\d{2})(.*)/) {
	$leftovers = $3; # Matches ``123 34''
    } elsif ($str =~ /^(\d+)[:dDo*](\d+)[:''](\d+[.,]?\d*)[""]?(.*)/) {
	$leftovers = $4; # Matches ``123:34:56.78'' and related patterns
    } elsif ($str =~ /^(\d+)[:dDo*](\d+[.,]?\d*)[:'']?(.*)/) {
	$leftovers = $3; # Matches ``123:34.56'' and related patterns
    } elsif ($str =~ /^(\d+[.,]?\d*)(.*)/) {
	$leftovers = $2; # Matches ``123.45'' 
                         #and other plain old decimal degree patterns
    }

    # Omitting the angle specification is an error.

    if (!defined($1)) {
	$DEBUG && print "*** Missing angle specification\n";
	return undef;
    }

    # Piece together the final angle from the deg, min, sec components..

    $angle = $1;
    if (defined($2)) { $angle += $2/60; }
    if (defined($3)) { $angle += $3/60/60; }

    # If no hemisphere indication yet, try for one at the end of the
    # string.

    ($str = $leftovers) =~ s/^\s+//;
    if (!defined($hemi) && $str =~ /^([nsew+-])(.*)/i) {
	$hemi = $1;
	($str = $2) =~ s/^\s+//;
    }

    # Extra stuff left at the end of the string is an error.

    if ($str ne "") {
	$DEBUG && print "*** Found extra stuff ($str) at end of string\n";
	return undef;
    }

    # Now adjust the sign to match the hemisphere and return it.

    if (defined($hemi) && $hemi =~ /[sw-]/i) { $angle *= -1; }
    return $angle;
}


#------------------------------------------------------------
# deg2dms DEGS, [HEMISPHERE]
#------------------------------------------------------------
# Converts degrees to a dd:mm:ss string
#
sub deg2dms {
    my ($angle, $hemi) = @_;

    # Compute a deg:min:sec string for the absolute value of the
    # angle.  This must be padded up by half of the value of the least
    # significant character to be displayed to prevent, e.g., 30:00
    # from being displayed as 29:60, which while arguably correct is
    # not standard form.

    my $rem = abs($angle) + (0.5/60/60);
    my $deg = int($rem); $rem = ($rem-$deg) * 60;
    my $min = int($rem); $rem = ($rem-$min) * 60;
    $rem = $rem - 0.5;
    my $dms;
    if($rem > -0.00001 && $rem < 0.00001) {
       $dms = sprintf("%d:%02d:00.0", $deg, $min);
    } elsif($rem < 10.0) {
       $dms = sprintf("%d:%02d:0%.5f", $deg, $min, $rem);
    } else {
       $dms = sprintf("%d:%02d:%.5f", $deg, $min, $rem);
    }

    # Compute the hemisphere indicator character.  This defaults to
    # +/-, but can be set to n/s by passing an "n" or "s" string, or
    # to e/w by passing an "e" or "w" string in the second parameter
    # to the call.  Passing a "+", "n", or "e" places the hemisphere
    # indicator character in front of the deg:min:sec string, while a
    # "-", "s", or "w" places the hemisphere indicator character after
    # the deg:min:sec string.  Passing an upper case character causes
    # an upper case character to be used in the returned string.

    my $hc; # the hemisphere indicator character.
    $hemi = $hemi || "+";
    if ($hemi !~ /\s*[nsew]/i) {
	$hc = ($angle >= 0)? "+": "-";
    } elsif ($hemi =~ /\s*[ns]/i) {
	$hc = ($angle >= 0)? "n": "s";
    } else {
	$hc = ($angle >= 0)? "e": "w";
    }
    if ($hemi =~ /\s*[NSEW]/) {
	$hc = uc($hc);
    }

    return ($hemi =~ /[-]/i) ? ($hemi . $dms) : ($dms);
}

sub str2dms {
   my ($str) = @_;
   my ($hemi, $deg, $min, $sec, $angle);
   my $newstr;
   defined($str) or $str = '';    # prevent those warnings
   $str =~ s/^\s+//;
   if ($str =~ /^([+-])(.*)/i) {
       $hemi = $1;
       ($str = $2) =~ s/^\s+//;
   }
   if (!$hemi && $str =~ /[sw]$/i) {
      $hemi = '-';
   }
   # Look for an angle specification, starting with the most
   # difficult to match (e.g., least likely to be matched
   # inadvertantly) pattern first.  The regex's must be set up to
   # leave the degree, minute, and seconds strings in $1-$3.

   if ($str =~ /^(\d{2,3})\.(\d{2})\.(\d{2})/) {
      # Matches ``123.34.56''
      $newstr = "$hemi"."$1".':'."$2".':'."$3.0";
   } elsif ($str =~ /^(\d{2,3})\s+(\d{2})\s+(\d{2})/) {
      # Matches ``123 34 56''
      $newstr = "$hemi"."$1".':'."$2".':'."$3.0";
   } elsif ($str =~ /^(\d{2,3})\s+(\d{2})/) {
      # Matches ``123 34''
      $newstr = "$hemi"."$1".':'."$2".':'.'00.0';
   } elsif ($str =~ /^(\d+)[:dDo*](\d+)[:''](\d+[.,]?\d*)[""]?/) {
      # Matches ``123:34:56.78'' and related patterns
      if($3 > -0.00001 && $3 < 0.00001) {
         $newstr = "$hemi"."$1".':'."$2".':'.'00.0';
      } elsif($3 < 10.0) {
         $newstr = "$hemi".sprintf("%d:%02d:0%.5f", $1, $2, $3);
      } else {
         $newstr = "$hemi".sprintf("%d:%02d:%.5f", $1, $2, $3);
      }
   } elsif ($str =~ /^(\d+)[:dDo*](\d+[.,]?\d*)[:'']?/) {
      # Matches ``123:34.56'' and related patterns
      $angle = $1;
      if (defined($2)) { $angle += $2/60; }
      $newstr = deg2dms($angle, $hemi);
   } elsif ($str =~ /^(\d+[.,]?\d*)/) {
      # Matches ``123.45'' and other plain old decimal degree patterns
      $angle = $1;
      $newstr = deg2dms($angle, $hemi);
   }
   return $newstr;

}
#------------------------------------------------------------
1;

