package IU::Coverage;

=head1 NAME

IU::Coverage - a granule/dataset coverage

=head1 DESCRIPTION

This parallels the IK_Coverage struct in the IK layer.

=head1 INSTANCE VARIABLES

The instance variables are:

    Area    The coverage area, as a string.  
            Format, POLYGON: "Pole LAT1 LON1 LAT2 LON2 LAT3 LON3 LAT4 LON4"
            Format, RANGE:   "LATn LONw LATs LONe"
            Format, POINT:   "LAT  LON"
            Format, G_RING:  "LAT1 LON1 LAT2 LON2 LAT3 LON3 LAT4 \
                              LON4 LAT5 LON5"
            Format, EMPTY:   ""

    Center  The center point, as a string.
            Format: "LAT LON"
    
    Type    The coverage type: POINT, RANGE, POLYGON, G_RING,
            GLOBAL or EMPTY.
   
    Pole    The "pole inside" value for polygons: "N", "S", "B",or"X".

The syntax of Area and Center come originally from imsgdbdump; they 
may be obtained from the following instance methods:

=over

=item area

This string is formatted based on the type; the possible forms are:

     Type     Area-format
     ------------------------------------------------------------
     POLYGON  "POLE LAT1 LON1 LAT2 LON2 LAT3 LON3 LAT4 LON4"
     GLOBAL   undef                   ## for backwards compatibility
     POINT    "LAT LON"
     RANGE    "NLAT WLON SLAT ELON"   ## coords of top-left, bot-right
     EMPTY    ''

The idea is that points are always in the form "LAT LON", where
LAT is -90.0 to 90.0, and LON is -180.0 to 180.0.

=item center

This string is also formatted based on type:

     Type     Center-format
     ------------------------------------------------------------
     GLOBAL   undef                   ## for backwards compatibility
     AREA     "LAT LON"
     POINT    "LAT LON"
     RANGE    "LAT LON"
     EMPTY    ''

=back

=head1 RCS INFO

$Id: Coverage.pm,v 1.31 2003/09/04 17:26:46 seufert Exp $

=cut

use IU;
use IU::Log;
use IU::ODL;
use IU::ResultGroup;


use strict;
#------------------------------
#
# GLOBALS
#
#------------------------------

# Some HTML:
my $DEG = '&#176;';

#------------------------------------------------------------
# new PARAMHASH
#------------------------------------------------------------
# Create a new object.  The paramhash contains the following:
#
#    Area   => The coverage area, as a string
#    Center => The center point
#    Type   => The coverage type: POINT, RANGE, POLYGON, GLOBAL,
#              or EMPTY.

sub new   { 
   my $class = shift;
   my %params = @_;
   my $self = \%params;
   exists ($params{'Type'}) or $self->{Type} = 'EMPTY';
   bless $self, $class;
}

#------------------------------------------------------------
# area
#------------------------------------------------------------
# Returns the coverage area, as a string.
# See the top of this file for the formats.

sub area { 
   my $self = shift;
   if ($self->type eq 'POLYGON') {
      return "$self->{Pole} $self->{Area}";
   }
   else {
      return ($self->{Area} || '');
   }
}

#------------------------------------------------------------
# area_pretty
#------------------------------------------------------------
# Returns the area, as a pretty string of the form.
#
#    "00.00^Lat, 00.00^Lon;\n
#     ...
#     00.00^Lat, 00.00^Lon"...
#
# Returns "Global" for global coverage.
# Returns empty if the area is unknown, and undef if the type is
# completely unrecognized.

sub area_pretty {
   my $self = shift;
   my $txt_only = shift;
   my $i;

   # Quit early if global:
   return 'Global' if ($self->type eq 'GLOBAL');

   my $str = $self->{Area} || return '';

   if ($self->type eq 'G_RING'){
      # Picture of the GRing stucture :
      # Area (Hashtable when G_RING is concerned)
      #  entries : Outer - (Array of Arrays)
      #  entries : Inner - (Array of Arrays)

      # Initialize some values before we need them.
      my @outer = {$self->{Area}->{Outer}};
      my $str = $self->{Area} || return '';
      my $outer_label = ($txt_only) ? 'Outer Ring:' : "Outer Ring:\n"; 
      my $outer_pretty;
      my $outer_elements = int (@{$str->{Outer}});
      my ($o,$lat_long_values);

      # This loop handles the NUMBER of Outer Rings

      for ( $o = 0; $o < $outer_elements; $o++) {
         my @lat_long = @{${$str->{Outer}}[$o]};
         my $no_elements = int(@lat_long);   
         my ($i,$elements_pretty);  
         # We have ONE Outer Ring, add it onto "elements" heap.
         for ( $i = 0; $i < $no_elements; $i++) {
            ($elements_pretty .= ($lat_long[$i])." ")
               if ($lat_long[$i]);
         }
         # We have all of the lat/longs , make 'em pretty
         if ($txt_only) {
            $elements_pretty =~ s{([-+\d\.]+)\s+([-+\d\.]+)}
                              {$1 Lat,$2 Lon;}gx;
         } else {
            $elements_pretty =~ s{([-+\d\.]+)\s+([-+\d\.]+)} 
                              {$1\xB0\xA0Lat,\xA0$2\xB0\xA0Lon;\n}gx;
            $elements_pretty =~ s{-}{-\xA0}gx;
            $elements_pretty =~ s{;\n\Z}{};
         } 

         # Master list of ALL the Outer Rings, use a "ruler" to
         # divide'em
         $outer_pretty = $outer_pretty.$outer_label.$elements_pretty;
      }
      my $inner_elements = int (@{$str->{Inner}});

      # Look at Outer ring for description of code.
      my $inner_label = ($txt_only) ? 'Inner Ring:' : "Inner Ring:\n";
      my $inner_pretty; 
      for ( $o = 0; $o < $inner_elements; $o++) {
         my @lat_long = @{${$str->{Inner}}[$o]};
         my $no_elements = int(@lat_long);
         my $elements_pretty;
         for ($i = 0; $i < $no_elements; $i++) {
            $elements_pretty .= ($lat_long[$i])." ";
         }
         if ($txt_only) {
            $elements_pretty =~ s{([-+\d\.]+)\s+([-+\d\.]+)}
                              {$1 Lat,$2 Lon;}gx;
         } else {
            $elements_pretty =~ s{([-+\d\.]+)\s+([-+\d\.]+)}
                              {$1\xB0\xA0Lat,\xA0$2\xB0\xA0Lon;\n}gx;
            $elements_pretty =~ s{-}{-\xA0}gx;
            $elements_pretty =~ s{;\n\Z}{};
         }
         $inner_pretty = $inner_pretty.$inner_label.$elements_pretty
            if ($elements_pretty);
      }
      # Since G_RING is done, return both halves.
      my $label = ($txt_only) ? 'G-ring: ' : "G-ring: \n\n";
      return ($label.$outer_pretty.$inner_pretty);
   }
   # Walk down points, mapping text as we go:
   if ($txt_only) {
      $str =~ s{([-+\d\.]+)\s+([-+\d\.]+)} {$1 Lat, $2 Lon;}gx;
   } else {
      $str =~ s{([-+\d\.]+)\s+([-+\d\.]+)} {$1\xB0\xA0Lat,\xA0$2\xB0\xA0Lon;\n}gx;
      $str =~ s{-}{-\xA0}gx;
      $str =~ s{;\n\Z}{};
   }
   $str;
}

#------------------------------------------------------------
# area_as_HTML
#------------------------------------------------------------
# Returns the area, as HTML.  
# Each lat/long pair is NOtBReakable.
#
# Returns "Global" for global coverage.
# Returns empty if the area is unknown.

sub area_as_HTML {
    my $self = shift;

    return 'Global' if ($self->type eq 'GLOBAL');

    # Quit early if none:
    return '' unless ($self->{Area});

    # If G-ring, just return pretty area for now.
    if ($self->type eq 'G_RING'){
       return $self->area_pretty;
    }

    # Walk down points, mapping text as we go:
    my $str = $self->{Area} || return '';
    $str =~ s{([-+\d\.]+)\s+([-+\d\.]+)} 
             {$1$DEG\xA0Lat,\xA0$2$DEG\xA0Lon;\n}gx;
    $str =~ s{-}{-\xA0}gx;
    $str =~ s{;\n\Z}{};
    $str;
}

#------------------------------------------------------------
# area_as_range_HTML
#------------------------------------------------------------
# Returns the area, as an HTML string.  The format here is:
#
#    34^ to -12^ Lat, -160^ to -91^ Lon
#
# Each lat-lat or lon-lon range is NOtBReakable.
#
# Returns "Global" for global coverage.
# Returns empty if the area is unknown, and undef if the type is
# completely unrecognized.
#
# NOTE: This is intended for the display of datasets in dataset lists;
# its presence here is a slight KLUDGE. 

sub area_as_range_HTML {
    my $self = shift;

    # Quit early if empty:
    return '' unless ($self->{Area});

    return 'Global' if ($self->type eq 'GLOBAL');

    # Get the info:
    my @pts = split /\s+/, $self->{Area};
    my @nwse;                         # nLat, wLon, sLat, eLon
    if ($self->type eq 'POINT') {
       @nwse = @pts[0,1,0,1];
    }
    elsif ($self->type eq 'RANGE') {
       @nwse = @pts;
    }
    elsif ($self->type eq 'POLYGON') {
       my @lats = sort { $b <=> $a } @pts[0,2,4,6];  # north to south
       my @lons = sort { $a <=> $b } @pts[1,3,5,7];  # west to east
       @nwse = ($lats[0], $lons[0], $lats[3], $lons[3]);
    }
    else {                 # Hunh?
       return undef;
    }
    # Map all values to integers:
    my ($nLat, $wLon, $sLat, $eLon) = map { sprintf("%d", $_) } @nwse;

    # Format:
    return "<NOBR>$nLat$DEG to $sLat$DEG Lat,</NOBR> " .
           "<NOBR>$wLon$DEG to $eLon$DEG Lon</NOBR>";
}

#------------------------------------------------------------
# center
#------------------------------------------------------------
# Returns the area centroid, as a string of the form "LAT LON",
# or undef if none.
#
# Generally this always exists (except for GLOBAL coverage),
# although it might not hurt someday to compute it from the area.

sub center { 
    $_[0]->{Center} || '';
}

#------------------------------------------------------------
# center_pretty  FLAG
#------------------------------------------------------------
# Returns the area centroid, as a string of the form "LAT LON".
# Returns "Global" for global coverage.
# Returns empty string if none.

sub center_pretty { 
    my $self = shift;
    my $txt_only = shift || '0'; 

    # Quit early if global:
    return 'Global' if ($self->type eq 'GLOBAL');

    # Walk down points, mapping text as we go:
    my $str = $self->{Center} || return '';
    if ($txt_only) {
       $str =~ s{([-+\d\.]+)\s+([-+\d\.]+)} {$1 Lat, $2 Lon; }gx;
    } else {
       $str =~ s{([-+\d\.]+)\s+([-+\d\.]+)} {$1\xB0 Lat, $2\xB0 Lon;\n}gx;
       $str =~ s{;\n\Z}{};
    }
    $str;
}

#------------------------------
# center_as_HTML
#------------------------------
# Return the area center, as an HTML string.  The format here is:
#
#    34.20^ Lat, -160.51^ Lon
#
# The lat and the lon are independently NOtBReakable.
#
# This is intended for the display of granules in granule lists;
# its presence here is a slight KLUDGE.
#
# Returns empty if the center is unknown, and undef if the type is
# completely unrecognized.

sub center_as_HTML {
   my $self = shift;
    
   # Output:
   if ($self->type eq 'GLOBAL') {
      return 'Global';
   }
   elsif (defined($self->{Center})) {
      my @c = map { sprintf("%.2f", $_) } 
      split(/\s+/, $self->{Center});
      my $str = "$c[0]$DEG\xA0Lat,\xA0$c[1]$DEG\xA0Lon";
      $str =~ s{-}{-\xA0}gx;
      return $str;
   }
   else {
      return '';
   }          
}

#------------------------------------------------------------
# pole
#------------------------------------------------------------
# Which pole is inside?
# Returns false if not a polygon.

sub pole { 
    $_[0]->{Pole};
}

#------------------------------------------------------------
# type
#------------------------------------------------------------
# Returns the area type: POINT, RANGE, POLYGON, GLOBAL, or EMPTY.

sub type   { 
    ($_[0]->{Type} || 'EMPTY');
}

# paths -- return a reference to the array of path names, or []
sub paths { $_[0]->{Paths} || []; }

# rows -- return a reference to the array of row names, or []
sub rows { $_[0]->{Rows} || []; }

# wrs -- return the WRS type
sub wrs   { 
    ($_[0]->{WRS} || '');
}

#------------------------------------------------------------
# from_chunk CHUNK
#------------------------------------------------------------
# Initiallize from a chunk which *contains* any of...
#
#    GLOBAL_GRANULE
#    POINT_LOC
#    RANGE_LOC
#    POLYGON_LOC
#    PATH_ROW_LOC 

sub from_chunk   { 
    my ($self, $chunk) = @_;
    my $info;

    # Create:
    ref($self) or $self = $self->new;
    $self->{Area} = $self->{Center} = $self->{Pole} = undef;

    # It's a GLOBAL...
    $info = uc($chunk->value('GLOBAL_GRANULE'));
    if ($info && $info eq 'Y') {
       $self->{Type} = 'GLOBAL';
    }
    # It's a POINT...
    elsif ($info = $chunk->value('POINT_LOC')) {
       $self->{Type} = 'POINT';

       # Get coords/center: 
       my @coords = map {$_ = sprintf("%.2f", $_)
                          }($info->seq_values('LATITUDE'), 
                            $info->seq_values('LONGITUDE'));
       # Store:
       $self->{Area}   = join ' ', @coords;
       $self->{Center} = join ' ', @coords;
    }
    # It's a G_RING...
    elsif ($info = $chunk->value('G_RING_LOC')) {

        $self->{Type} = 'G_RING';
        # Get center:
        my @center = map {$_ = sprintf("%.2f", $_)
                          }($info->value('CENTROID_LAT'),
                            $info->value('CENTROID_LON'));
        my (@outer_array, @inner_array);
        my ($url_chunk, my $outer_counter);
        $outer_counter = 0;

        # Extract outer Values from chunk
        foreach $url_chunk ($info->values('OUTER_RING')) {
            my @outer_coords = ();

            my @outer_lats = grep (defined $_, map {$_ = sprintf("%.2f", $_)
                        }($url_chunk->seq_values('LATITUDE')));
            my @outer_lons = grep (defined $_, map {$_ = sprintf("%.2f", $_)
                        }($url_chunk->seq_values('LONGITUDE')));
            #debug "outer_lons: '".join("' '",@outer_lats),"'";

            # Merge the Lat/Lon arrays into an array of points.
            for (0..$#outer_lats) {
               push @outer_coords, $outer_lats[$_], $outer_lons[$_];
            };
            #debug "outer_coords: '".join("' '",@outer_coords),"'";

            # Store reference of array in array outer_array
            $outer_array[$outer_counter] = \@outer_coords;

            my $inner_chunk;
            my $inner_counter = 0;
            foreach $inner_chunk ($url_chunk->values('INNER_RING')) {
               my @inner_lats = map {$_ = sprintf("%.2f", $_)
                        }($inner_chunk->seq_values('LATITUDE'));
               #debug "inner_lats: '".join("' '",@inner_lats),"'";
               my @inner_lons = map {$_ = sprintf("%.2f", $_)
                        }($inner_chunk->seq_values('LONGITUDE'));
               #debug "inner_lons: '".join("' '",@inner_lons),"'";
               my @inner_coords = ();
               for (0..$#inner_lats) {
                  push @inner_coords, $inner_lats[$_], $inner_lons[$_];
               };
               #debug "inner_coords: '".join("' '",@inner_coords),"'";
               $inner_array[$inner_counter] = \@inner_coords;
               $inner_counter +=  1;
            }
            $outer_counter +=1;
        }
        # Store:
        $self->{Area}->{Outer}   = \@outer_array;
        $self->{Area}->{Inner}   = \@inner_array;
        $self->{Center} = join ' ', @center;
    }
    # It's a RANGE...
    elsif ($info = $chunk->value('RANGE_LOC')) {
       $self->{Type} = 'RANGE';

       # Get coords:
       my @coords = map {$_ = sprintf("%.2f", $_)}
          ($info->value('NORTH_LATITUDE'),
           $info->value('WEST_LONGITUDE'),
           $info->value('SOUTH_LATITUDE'),
           $info->value('EAST_LONGITUDE'));

       # Get center:
       # Need more complicated way of calculating center longitude 
       # for ranges that cross the dateline.

       my $lonCenter;

       # convert to 0 to 360 degree system.

       my $lonEast = ( ($coords[3] < 0) ? $coords[3] + 360 :
                                          $coords[3] );
       my $lonWest = ( ($coords[1] < 0) ? $coords[1] + 360 :
                                          $coords[1] );

       # calculated the center longitude.

       $lonEast += 360 if ($lonWest > $lonEast);
       $lonCenter = ($lonEast + $lonWest)/2.0;

       # convert back to -180 to 180 degree system.

       $lonCenter -= 360 if ($lonCenter > 180);

#  my @center = map {$_ = sprintf("%.2f", $_)
#          }(($coords[0]+$coords[2])/2.0,
#            ($coords[1]+$coords[3])/2.0);

       my @center = map {$_ = sprintf("%.2f", $_)}
          (($coords[0]+$coords[2])/2.0,$lonCenter);

       # Store:
       $self->{Area}   = join ' ', @coords;
       $self->{Center} = join ' ', @center;
    }
    # It's a POLYGON...
    elsif ($info = $chunk->value('POLYGON_LOC')) {
       $self->{Type} = 'POLYGON';

       # Get coords: 
       my @lats = map {$_ = sprintf("%.2f", $_)}
          ($info->seq_values('LATITUDE'));
       my @lons = map {$_ = sprintf("%.2f", $_)}
          ($info->seq_values('LONGITUDE'));
       my @coords;
       for (0..3) { push @coords, $lats[$_], $lons[$_] };

       # Get center:
       my @center = map {$_ = sprintf("%.2f", $_)}
          ($info->value('CENTROID_LAT'),
           $info->value('CENTROID_LON'));
       # Store:
       $self->{Area}   = join ' ', @coords;
       $self->{Center} = join ' ', @center;
       $self->{Pole}   = $info->value('POLE_INCLUDED') || 'X';
    }
#    else {
#       return undef;
#    }
    # The following checks for presence of a PATH_ROW_LOC group
    if($info = $chunk->value('PATH_ROW_LOC')) {
       $self->{Paths} = [$info->seq_values('PATH')];
       $self->{Rows}  = [$info->seq_values('ROW')];
       $self->{WRS}   = $info->value('WRS_TYPE');
       # The following code checks for path row in the specialized results and
       # any specialized results in Result Groups.
       # if any path/row info found, it over writes that from the PATH_ROW_LOC 
       my $results_chunk;
       my $name;
       my $value;
       my @paths;
       my @rows;
       foreach $results_chunk ($chunk->values('SPECIALIZED_RESULTS')) {
          $name = $results_chunk->value('RESULT_NAME');
          $value  = $results_chunk->value('RESULT_VALUE');
          if ($name eq 'SP_STARTING_PATH') {
	     $paths[0] = $value;
          } elsif ($name eq 'SP_STARTING_ROW') {
	     $rows[0] = $value; #unless ($value eq $rows[1]);
          } elsif ($name eq 'SP_ENDING_ROW') {
	     $rows[1] = $value; #unless ($value eq $rows[0]);
          }
       }
       my $result_group_chunk;
       foreach $result_group_chunk ($chunk->values('RESULT_GROUP')) {

          my $resultGroup = new IU::ResultGroup;
          $resultGroup->from_chunk($result_group_chunk);

          my $start_path_ref =  
         ( $resultGroup->findSpecializedResults('SP_STARTING_PATH') || undef);
          $paths[0]  = $start_path_ref->[0]->resultValueAsString 
                       if (defined($start_path_ref->[0]));

          my $row_one_ref = 
             ($resultGroup->findSpecializedResults('SP_STARTING_ROW') || undef);

          $rows[0]  = $row_one_ref->[0]->resultValueAsString 
                       if (defined($row_one_ref->[0]));

          my $row_two_ref = 
               ($resultGroup->findSpecializedResults('SP_ENDING_ROW') || undef);

          $rows[1]  = $row_two_ref->[0]->resultValueAsString 
                if (defined($row_two_ref->[0]));
       }

       $self->{Paths} = \@paths if (int(@paths) > 0);
       $self->{Rows}  = \@rows if (int(@rows) > 0);

    }
    # Done!
    $self;
}

#------------------------------------------------------------
# to_ODL 
#------------------------------------------------------------
# I<Instance method.> 
# Method to convert coverage into an ODL.
# Depending on the geo type, it will be one of POINT_LOC,
# POLYGON_LOC, RANGE_LOC, GLOBAL_GRANULE, or G_RING_LOC ODL group.
# Unlike other to_ODL methods, this one will return two things,
# the ODL parameter name (the group name one of above) and
# the value (an ODL group of the right type, or string in the
# case of global.  Use this method this way:
#      $odl->add_param($coverage->to_ODL);

sub to_ODL {
   my $self = shift;
   # The coverage type: POINT, RANGE, POLYGON, G_RING, GLOBAL, or EMPTY.

   # If empty return nothing.
   return ('') if (!defined ($self->{Type}) or
                   ($self->{Type} eq 'EMPTY'));

   # If it is global, return the parameter, value.
   if ($self->{Type} eq 'GLOBAL') {   # It's a GLOBAL...
      return ('GLOBAL_GRANULE' => 'Y');
   }
   # Else, create the geo group and get the center values.
   # Create new ODL group for geo.
   my $loc_odl = IU::ODL->new_group;
 
   # Split the Center into lat and lon (LAT, LON).
   my ($center_lat, $center_lon);
   if ($self->{Center}) {
      ($center_lat, $center_lon) = split (/\s/, $self->{Center});
   }

   # Depending on the geo type, create the right ODL.

   # If type polygon:
   if ($self->{Type} eq 'POLYGON') {
      # Split the Area into arrays of lats and lons.
      # Format, "LAT1 LON1 LAT2 LON2 LAT3 LON3 LAT4 LON4"
      my (@lat, @lon);
      ($lat[0], $lon[0], $lat[1], $lon[1],
       $lat[2], $lon[2], $lat[3], $lon[3]) =
         split (/\s/, $self->{Area});
      # Make ODL arrays.
      my @odl_real_lats = map {IU::ODL->real($_)} @lat;
      my @odl_real_lons = map {IU::ODL->real($_)} @lon;
      # Add the lats and lons to the local ODL group.
      $loc_odl->add_param('LATITUDE' => 
                          IU::ODL->sequence(@odl_real_lats));
      $loc_odl->add_param('LONGITUDE' => 
                          IU::ODL->sequence(@odl_real_lons));
      # Add center to local group.
      if ($self->{Center}) {
         $loc_odl->add_param('CENTROID_LAT' => IU::ODL->real($center_lat));
         $loc_odl->add_param('CENTROID_LON' => IU::ODL->real($center_lon));
      }
      # Add the pole.
      $loc_odl->add_param('POLE_INCLUDED' => $self->{Pole});
      # Add the local group to the ODL as POLYGON_LOC.
      return ('POLYGON_LOC' => $loc_odl);
   }
   # If type point:
   elsif ($self->{Type} eq 'POINT') {
      # For POINT, Center is point so we already have values, but,
      # for some reason, they are sequences so have to be arrays.
      my @odl_real_lat = (IU::ODL->real($center_lat));
      my @odl_real_lon = (IU::ODL->real($center_lon));
      # Add lat & lon to POINT_LOC group.
      $loc_odl->add_param('LATITUDE' => 
                          IU::ODL->sequence(@odl_real_lat));
      $loc_odl->add_param('LONGITUDE' => 
                          IU::ODL->sequence(@odl_real_lon));
      # Add the local group to the odl as POINT_LOC.
      return ('POINT_LOC' => $loc_odl);
   }
   # If type g-ring:
   elsif ($self->{Type} eq 'G_RING') {
      # Since we didn't store which inner rings belong to
      # which outer rings, we'll put all inner rings inside
      # the first outer ring ODL group.
      my $first = 1;   # Flag if first outer ring, so add inner.

      # outer is array of outer rings (which are refs to arrays).
      my @outer = @{$self->{Area}->{Outer}};
      # inner is array of inner rings (which are refs to arrays).
      my @inner = @{$self->{Area}->{Inner}};

      my ($num, $i);
      my $outer_coords_ref;   # reference to outer coordinates array
      # For every outer ring:
      foreach $outer_coords_ref (@outer) {
         my (@lats, @lons);
         my (@lat_lons) = @{$outer_coords_ref};
         $num = int(@lat_lons);
         for ($i=0; $i<$num; $i++) {
            push(@lats, $lat_lons[$i]);   # Extract next lat
            $i++;
            push(@lons, $lat_lons[$i]);   # Extract next lon
         }
         # Create new ODL group for this OUTER_RING.
         my $out_odl = IU::ODL->new_group;
         # Create list of ODL reals for Lat and Lon.
         my @odl_real_lat = map {IU::ODL->real($_)} @lats;
         my @odl_real_lon = map {IU::ODL->real($_)} @lons;
         # Add lat & lon to OUTER_RING group.
         $out_odl->add_param('LATITUDE' => 
                             IU::ODL->sequence(@odl_real_lat));
         $out_odl->add_param('LONGITUDE' => 
                             IU::ODL->sequence(@odl_real_lon));
         # If first outer ring, add inner rings.
         if ($first) {
            $first = 0;   # Only one first outer ring.
            my ($num, $i);
            my $inner_coords_ref;   # reference to inner coords array
            # For each inner ring:
            foreach $inner_coords_ref (@inner) {
               my (@lats, @lons);
               my (@lat_lons) = @{$inner_coords_ref};
               $num = int(@lat_lons);
               for ($i=0; $i<$num; $i++) {
                  push(@lats, $lat_lons[$i]);   # Extract next lat
                  $i++;
                  push(@lons, $lat_lons[$i]);   # Extract next lon
               }
               # Create new ODL group for this INNER_RING.
               my $in_odl = IU::ODL->new_group;
               # Create list of ODL reals for Lat and Lon.
               my @odl_real_lat = map {IU::ODL->real($_)} @lats;
               my @odl_real_lon = map {IU::ODL->real($_)} @lons;
               # Add lat & lon to INNER_RING group.
               $in_odl->add_param('LATITUDE' => 
                                  IU::ODL->sequence(@odl_real_lat));
               $in_odl->add_param('LONGITUDE' => 
                                  IU::ODL->sequence(@odl_real_lon));
               # Add this INNER_RING to the first OUTER_RING group.
               $out_odl->add_param('INNER_RING' => $in_odl);
            }     # End For every inner ring.
         }     # End If first outer ring.
         # Add the outer ring group to the local group.
         $loc_odl->add_param('OUTER_RING' => $out_odl);
      }     # End For each outer ring.
      # Add center to local group.
      if ($self->{Center}) {
         $loc_odl->add_param('CENTROID_LAT' => IU::ODL->real($center_lat));
         $loc_odl->add_param('CENTROID_LON' => IU::ODL->real($center_lon));
      }
      # Add the local group to ODL as a G_RING_LOC.
      return ('G_RING_LOC' => $loc_odl);
   }     # End If type g-ring.
   # If it is a range:
   elsif ($self->{Type} eq 'RANGE') {
      # Split the Area into array of coords
      # Format:   "LATn LONw LATs LONe"
      my ($nlat, $wlon, $slat, $elon) = split (/\s/, $self->area);
      # Add lat & lon to local group.
      $loc_odl->add_param('NORTH_LATITUDE' => IU::ODL->real($nlat));
      $loc_odl->add_param('WEST_LONGITUDE' => IU::ODL->real($wlon));
      $loc_odl->add_param('SOUTH_LATITUDE' => IU::ODL->real($slat));
      $loc_odl->add_param('EAST_LONGITUDE' => IU::ODL->real($elon));
      # Add the local group to the ODL as RANGE_LOC.
      return ('RANGE_LOC' => $loc_odl);
   }
   # Done!
   $loc_odl;
}
#------------------------------------------------------------
1;
