package IU::Criterion;


=head1 NAME

IU::Criterion - specialized criteria


=head1 DESCRIPTION

Instance of this class are analogous to the SPECIALIZED_CRITERIA group 
as described by the IMS V0 Data Dictionary.

Most of the work is done by the subclasses of this class, one subclass
for each "kind of interface" being presented to the user.


=head1 READING FROM CHUNKS

Input chunks are presumed to come
from direct conversion of the ODL, and look like the "value" of
the SPECIALIZED_CRITERIA "parameter" in the following chunk:

      BEGIN SPECIALIZED_CRITERIA
         CATEGORY_NAME: Radiation
         VARIANT: MAX GAMMA
 	 CRITERIA_NAME: Maximum Gamma
 	 CRITERIA_TYPE: REAL
	 COMMENT: This is the maximum gamma you can take before sunburning.
         CRITERIA_MIN: 0.0
         CRITERIA_MAX: 10.0
         SELECT_NUM: ONE
      END SPECIALIZED_CRITERIA


=head1 CLASS HIERARCHY

One a criterion is read in, it is reblessed to an appropriate subclass:

      Criterion
        |
        |-- Geo
        |-- Time
        |-- InputNum
        |-- InputRange
        |-- InputText
        |-- SelectMany
        |-- FixedField
        |-- FloatingScene
        `-- SelectOne


=head1 INSTANCE VARIABLES

More or less as described by the data dictionary.


=head1 RCS INFO

$Id: Criterion.pm,v 1.28 2003/08/01 19:10:46 solomon Exp $


=head1 SUBCLASSES

=cut

use IU;
use IU::Chunk;
use IU::Log;
use IU::IW;
use IU::ODL;
use IU::FancyURL;

use IU::Criterion::Geo;
use IU::Criterion::Time;
use IU::Criterion::InputNum;
use IU::Criterion::InputRange;
use IU::Criterion::InputText;
use IU::Criterion::InputDate;
use IU::Criterion::SelectOne;
use IU::Criterion::SelectMany;
use IU::Criterion::FixedField;
use IU::Criterion::FloatingScene;
use IU::Criterion::Latitude;
use IU::Criterion::Longitude;
use IU::Criterion::SubPage;
use IW::Date qw(TZ_to_datetime);
use IW::Util (qw(:url));

#------------------------------
#
# VARIOUS ACCESS METHODS...
#
#------------------------------

# Get the cost, optional

sub cost {
   shift->{Cost}
}

# Get the comment, as a string with newlines:
sub comment { 
    shift->{COMMENT} 
}

# Is this a range?
sub is_range { 
    $_[0]->{RANGE} && ($_[0]->{RANGE} eq 'Y') 
}

# Is it obscured?  Returns true or false.  Argument sets it.
sub is_obscured { 
    my ($self, $val) = @_;
    $self->{OBSCURED} = ($val ? 'Y' : undef)   if @_ > 1;
    ($self->{OBSCURED} || '') eq 'Y';
}

# Is it required?  Returns true or false.  Argument sets it.
sub is_required { 
    my ($self, $val) = @_;
    $self->{REQUIRED} = ($val ? 'Y' : undef)   if @_ > 1;
    ($self->{REQUIRED} || '') eq 'Y';
}

# Get value to look for in granule PSA.
sub match {
    @{$_[0]->{CRITERIA_VALUE_MATCH} || []}
}

# Get criteria value from value of the PSA.
sub value_from_psa {
    @{$_[0]->{CRITERIA_VALUE_FROM_PSA} || []}
}

# Get criteria default from value of the PSA.
sub default_from_psa {
    @{$_[0]->{CRITERIA_DEFAULT_FROM_PSA} || []}
}

# Get min from psa, for ranges:
sub min_from_psa  {
    $_[0]->{CRITERIA_MIN_FROM_PSA}
}

# Get min from core attribute, for ranges:
sub min_from_core  {
    $_[0]->{CRITERIA_MIN_FROM_CORE}
}

# Get min allowed value, for ranges:
sub min  { 
    $_[0]->{CRITERIA_MIN}
}

# Get max from psa, for ranges:
sub max_from_psa {
    $_[0]->{CRITERIA_MAX_FROM_PSA}
}

# Get max from core attribute, for ranges:
sub max_from_core {
    $_[0]->{CRITERIA_MAX_FROM_CORE}
}

# Get max allowed value, for ranges:
sub max { 
    $_[0]->{CRITERIA_MAX}
}

# Get the maximum length, for text input:
sub max_len { 
    shift->{MAX_LEN} 
}

# Get whether or not criteria should be validated:
sub do_validation {
   $_[0]->{VALIDATE} || 'Y';
}

# Get the name of this criterion:
sub name { 
    shift->{CRITERIA_NAME} 
}

# Get the search category:
sub category { 
    shift->{CATEGORY_NAME} 
}

# Get the variant.  
#    WARNING: because of case-sensitivity problems (i.e.,
#    imsdepval sometimes returns mixed case as uppercase, sometimes not),
#    it is returned all-uppercase.  Variants are case-folded for WG.
sub variant { 
    defined($_[0]) || return undef;      # prevent warnings
    $_[0]->{VARIANT};
}

# How many selections are allowed, ONE or MANY?
# Default is ONE.
sub select_num {
    $_[0]->{SELECT_NUM} || 'ONE';
}

# Get the ODL type: STRING, REAL, FLOAT, INTEGER, or DATE.
sub type { 
    shift->{CRITERIA_TYPE} 
}
# Get sub criteria:
sub subcriteria {
    @{$_[0]->{SUBCRITERIA} || undef} 
}

# Get allowed values for valids lists:
sub valids { 
    @{$_[0]->{CRITERIA_VALUE} || []} 
}

# Get allowed values for valids lists:
sub defaults {
    @{$_[0]->{CRITERIA_DEFAULT} || undef}
}




#------------------------------
#
# MISC METHODS...
#
#------------------------------

#------------------------------
# ident [IDENT]
#    Set/get the identifier:
#
sub ident {
    my ($self, $ident) = @_;
    $self->{Ident} = $ident if (@_ > 1);
    $self->{Ident};
}

#------------------------------
# user_made_selections
#     Are there user selections?  Return true or false.
#     A selection must be defined for it to count.
#
sub user_made_selections { 
    my $self = shift;
    $self->{UserSlct} ||= {}; 
    grep {defined($self->{UserSlct}{$_})} keys %{$self->{UserSlct}};
} 

#------------------------------
# user_selections
#     Get a ref to the hash of user selections.
#
sub user_selections { $_[0]->{UserSlct} ||= {} }

#------------------------------
# get_user_selections
# return array of user selections for this criterion.
sub get_user_selections {

   @{ $_[0]->{UserSlct}{VALUES} || undef};
}

#------------------------------
# value_to_CGI VALUE
#     Make value nice for CGI (strips T and Z from date):
#
sub value_to_CGI {
    my ($self, $value) = @_;
    $value =~ s/[TZ]/ /g  if $self->type eq 'DATE';
    $value;
}

#------------------------------
# whine ERRNO, PROBLEM
#     Private: the user didn't supply a value I wanted: puke!
#
sub whine {
    my ($self, $eno, $problem) = (shift, shift, join('',@_));
    my $ident = $self->ident || '';
    puke $eno, "$ident criterion `", $self->name, "': $problem";
}



#------------------------------
#
# MAIN PUBLIC INTERFACE...
#
#------------------------------

#------------------------------------------------------------
# new PARAMHASH
#------------------------------------------------------------
# I<Constructor.>

sub new {
    my ($class, %params) = @_;
    my $self = bless \%params, $class;
    $self->{UserSlct} = {};
    $self;
}

#---------------------------------------------------------------------
# copy IU::Criterion
#---------------------------------------------------------------------
# Make your Criterion object a deep copy of another Criterion.  Usage:
#    my $new = new IU::Criterion;
#    $new->copy($old);
# Input: a criterion to copy.  Can be object of any subclass too.
# Output: modifies your existing criterion to be a copy of the old.

sub copy {
   my ($self, $old) = @_;
   my $key;
   my $subcrit;
   my %newcrit;
   my @subcriteria;
   foreach $key (qw(CATEGORY_NAME
                    CRITERIA_MAX
                    CRITERIA_MAX_FROM_PSA
                    CRITERIA_MAX_FROM_CORE
                    CRITERIA_MIN
                    CRITERIA_MIN_FROM_PSA
                    CRITERIA_MIN_FROM_CORE
                    CRITERIA_NAME
                    CRITERIA_TYPE    
                    MAX_LEN
                    RANGE
                    REQUIRED
                    OBSCURED
                    SELECT_NUM
                    VALIDATE
                    COMMENT
                    CRITERIA_VALUE
                    CRITERIA_DEFAULT
                    CRITERIA_VALUE_MATCH
                    CRITERIA_VALUE_FROM_PSA
                    CRITERIA_DEFAULT_FROM_PSA
                    SUBCRITERIA
                    VARIANT)) {
      $self->{$key} = $old->{$key} if (defined ($old->{$key}));
   }

   # Get subclass, and rebless:
   my $subclass = $self->subclass || return error "no appropriate subclass!";
   bless $self, "IU::Criterion::$subclass";
   return $self;
}

#------------------------------------------------------------
# field FIELD, [VALUE]
#------------------------------------------------------------
# Instance method.
# Get/set part of the user selection.  
# The FIELD is specific to the given subclass.
# Returns the selection, or undef if none.
#
sub field {
    my ($self, $field, $value) = @_;
    $self->{UserSlct}{$field} = $value if (@_ > 2);
    $self->{UserSlct}{$field};
}

#------------------------------------------------------------
# from_chunk CHUNK
#------------------------------------------------------------
# I<Constructor.>  Create from a SPECIALIZED_CRITERIA chunk.

sub from_chunk {
    my ($self, $chunk, $granule) = @_;
    # Make safe for instances or classes:
    ref($self) or $self = $self->new;

    # Get basic scalar instance vars:
    my $key;
    foreach $key (qw(
		     CATEGORY_NAME
		     CRITERIA_MAX
                     CRITERIA_MAX_FROM_PSA
                     CRITERIA_MAX_FROM_CORE
		     CRITERIA_MIN
                     CRITERIA_MIN_FROM_PSA
                     CRITERIA_MIN_FROM_CORE
		     CRITERIA_NAME
		     CRITERIA_TYPE		     
		     MAX_LEN
		     RANGE
		     REQUIRED
                     OBSCURED
		     SELECT_NUM
                     VALIDATE
		     VARIANT
		     )) {
	$self->{$key} = $chunk->value($key);
    }

    # Add any MISC_URL groups.
    $key = 'MISC_URL';

    # Allows multiple URLs:
    $self->{$key} = [];

    # Look for MISC_URL groups if there are any
    my ($url_chunk, $furl);
    foreach $url_chunk ($chunk->values($key)) {
       my $furl = IU::FancyURL->from_chunk($url_chunk, $key);
       # save the MISC_URL groups as they are found
       push @{$self->{$key}}, $furl;

    }

    # Handle nested SPECIALIZED-CRITERIA, will handle as many levels
    # of nesting as is present in the ODL.
    # Store them in {SUBCRITERIA} array.
    my $crit_chunk;
    my @subcriteria;
    foreach $crit_chunk ($chunk->values('SPECIALIZED_CRITERIA')) {

        # Parse it, and save it:
        my $crit = IU::Criterion->from_chunk($crit_chunk,$granule) || next;
        push @subcriteria, $crit;
    }
    $self->{SUBCRITERIA} = \@subcriteria;
    # Fixup precision on reals (sigh) so that they don't display with
    # a ridiculous number of decimal places...
    if (($self->{CRITERIA_TYPE} eq 'REAL') 
        or ($self->{CRITERIA_TYPE} eq 'FLOAT')) {
	$self->{CRITERIA_MIN} =~ s{(\.\d+?)0+\Z}{$1} if ($self->{CRITERIA_MIN});
	$self->{CRITERIA_MAX} =~ s{(\.\d+?)0+\Z}{$1} if ($self->{CRITERIA_MAX});
    }

    # Get sequence strings:
    foreach $key (qw(COMMENT)) {
	$self->{$key} = join "\n", $chunk->seq_values($key);
    }

    # Get sequences:
    foreach $key (qw(CRITERIA_VALUE CRITERIA_DEFAULT CRITERIA_VALUE_MATCH
        CRITERIA_VALUE_FROM_PSA CRITERIA_DEFAULT_FROM_PSA)) {
	$self->{$key} = [$chunk->seq_values($key)];
    }

    # If "FROM_PSA" or "MATCH" keywords in SPECIALIZED_CRITERIA group 
    # then replace standard values with values from the specificed granule
    # product specific attributes (PSAs).

    if (($self->value_from_psa || $self->default_from_psa ||
         $self->min_from_psa || $self->max_from_psa || $self->match) 
        && defined($granule)) {
       my @new_values;
       my @new_defaults;
       my @new_matches;

       # Get top level SPECIALIZED_RESULTS for the granule.
       my @results = @{$granule->results};

       # Get all the SPECIALIZED_RESULTS inside RESULT_GROUPS for 
       # the granule and flatten in with the top level SPECIALIZED_RESULTS.
       foreach my $resultGroup (@{$granule->resultGroup}) {
          push @results, @{$resultGroup->specializedResults};
       } 

       foreach my $result (@results) {

          push @new_values, ${$result->resultValue}[0]
             if (grep {$_ eq $result->resultName} $self->value_from_psa);
  
          push @new_defaults, ${$result->resultValue}[0]
             if (grep {$_ eq $result->resultName} $self->default_from_psa);

          push @new_matches, $result->resultName
             if ( (grep {$_ eq $result->resultName} $self->valids) &&
                  (grep {$_ eq ${$result->resultValue}[0]} $self->match));

          $self->{CRITERIA_MIN} = ${$result->resultValue}[0]
             if ($self->min_from_psa && 
                 ($result->resultName eq $self->min_from_psa));
          
          $self->{CRITERIA_MAX} = ${$result->resultValue}[0]
             if ($self->max_from_psa &&
                 ($result->resultName eq $self->max_from_psa));

       } 
       # replace CRITERIA_VALUE from package with psa values.
       $self->{CRITERIA_VALUE} = \@new_values if (int(@new_values) > 0);

       # replace CRITERIA_VALUE from package with match psa names.
       $self->{CRITERIA_VALUE} = \@new_matches if (int(@new_matches) > 0);

       # replace CRITERIA_DEFAULT from package with psa values.
       $self->{CRITERIA_DEFAULT} = \@new_defaults if (int(@new_defaults) > 0);
    }

    # Fixup the comment:
    if ($self->{COMMENT}) {
	$self->{COMMENT} .= '.' unless ($self->{COMMENT} =~ /[\.\?\!]$/);
    }
 
    # Get subclass, and rebless:
    my $subclass = $self->subclass || return error "no appropriate subclass!";
    bless $self, "IU::Criterion::$subclass";

    # Get GEO values from the granule coverage.
    if ($self->type eq 'GEO' && defined($granule)) {
       $self->coverage($granule->coverage);
    }

    # Get Date/Time values from the granule coverage if not supplied
    # by min and max. 
    if ($self->type eq 'DATE' && defined($granule)) {
       $self->{CRITERIA_MIN} = $granule->start_date 
          if ($self->min_from_core eq 'START_DATE');
       $self->{CRITERIA_MAX} = $granule->stop_date 
          if ($self->max_from_core eq 'STOP_DATE');
    }

    return $self;
}

#------------------------------------------------------------
# subclass
#------------------------------------------------------------
# I<Instance method, private.>
# Return the subclass we belong to.  This is used by the base class
# constructor to re-bless the instance into something more specific.

sub subclass {
    my $self = shift;

    # Use cached value if available:
    return $self->{FormType} if defined($self->{FormType});

    # Get the generic directions:
    my $case;
    if ($self->type eq 'GEO') {           # geographic area
	return 'Geo';
    }
    elsif ($self->type eq 'TIME') {           # geographic area
	return 'Time';
    }
    elsif (($self->type eq 'STRING') &&   # fill-in-the-blank
	(defined($self->max_len))) {     
	return 'InputText';
    }
    elsif ($self->is_range) {             # range
	return 'InputRange';
    }
    elsif ($self->type eq 'DATE' && !$self->is_range) {
        return 'InputDate';
    }
    elsif (($self->select_num eq 'ONE') &&  
	   defined($self->min) &&
	   defined($self->max) &&
	   !($self->valids)) {            # enter number with a range
	return 'InputNum';	
    }
    elsif ($self->type eq 'FIXED') {       # fixed field
           return 'FixedField';
    }
    elsif ($self->type eq 'FLOATINGSCENE') {       # FloatingScene
           return 'FloatingScene';
    }
    elsif ($self->type eq 'SUBPAGE') {       # SubPage
           return 'SubPage';
    }
    elsif ($self->type eq 'LATITUDE') {       # Latitude
           return 'Latitude';
    }
    elsif ($self->type eq 'LONGITUDE') {       # Longitude
           return 'Longitude';
    }
    elsif ($self->valids) {               # select from list of valids
        return (($self->select_num eq 'ONE') ? 'SelectOne' : 'SelectMany');
    }
    else {
	return error "Bad criterion: ", $self->name, ": can't be classified";
    }
}

#------------------------------------------------------------
# valid_type INPUT
#------------------------------------------------------------
# I<Instance method.>
# Asserts that INPUT matches type $self->type.
# Fatal exception if not.
# Alters argument in-place.

sub valid_type {
    my ($self, $input) = @_;
    my $ok;
    my $clean;         # a clean version of the input
    if (($self->type eq 'INTEGER') or ($self->type eq 'INT')) {
	# The following code was modified to handle negative integers
	$ok = ($input =~ /^\d+$|^-\d+$/);
        $clean = $input;
    }
    elsif (($self->type eq 'REAL') or ($self->type eq 'FLOAT')) {
	# The following code was modified to handle negative reals
        $ok = ($input =~ /^\d+$|^\d*\.(\d*)$|^-\d+$|^-\d*\.(\d*)$/);
	# The check for length($2) was added to handle negative reals
	my $prec = length($1) || length($2) || 1;  # keep precision intact
	$clean = sprintf("%.${prec}f", $input);
    }
    elsif ($self->type eq 'DATE') {
	my $date = (IW::Date->parse_TZ($input) ||
		    eval {IW::Date->parse_two(split(/\s+/, $input))});
	$ok = ref($date);                         # make sure it's an object
	$clean = $date->to_ODL->text if $ok;
    }    
    elsif ($self->type eq 'STRING') {
	$ok = 1;
	$clean = $input;
    }
    $ok or $self->whine(511, "input '$input' is not a good ", $self->type);
    $clean;
}

#------------------------------------------------------------
# valid_range INPUT
#------------------------------------------------------------
# I<Instance method.>
# Asserts that INPUT is in the range [$self->min, $self->max].
# Fatal exception if not.

sub valid_range {
    my ($self, $input) = @_;
    my $ok;

    return 1 if ($self->do_validation eq 'N');

    if (($self->type eq 'INTEGER') or ($self->type eq 'REAL') or
        ($self->type eq 'INT') or ($self->type eq 'FLOAT')) {
	$ok = 1;
	$ok &&= ($self->min <= $input) if defined($self->min);
	$ok &&= ($self->max >= $input) if defined($self->max);
    }
    elsif ($self->type eq 'DATE') {
	$ok = 1;
	$ok &&= ($self->min le $input) if defined($self->min);
	$ok &&= ($self->max ge $input) if defined($self->max);
    }
    $ok or $self->whine(513, "input '$input' not in range of ", 
			$self->min, " to ", $self->max);
    $clean;
}





#------------------------------
#
# CGI INTERFACE...
#
#------------------------------

#------------------------------------------------------------
# all_blank PREFIX, FIELDS...
#------------------------------------------------------------
# I<Instance method, utility.>
# Are the given CGI fields blank?
#
sub all_blank { 
    my ($self, $prefix, @fields) = @_;
    my $field;
    foreach $field (@fields) {
        IW::Util::blank($REQ->param("${prefix}$field")) or return 0;
    }
    1;
}

#------------------------------------------------------------
# print_form PARAMHASH
#------------------------------------------------------------
# I<Instance method.>
# Output form as HTML fields.  Params are:
#
#     Prefix  => the prefix to use
#     Invalid => output so as to indicate that this variant is invalid
#     Action  => Reset/Clear ( used by the "Reset"/"Clear" buttons

sub print_form {
    my $self = shift;
    my %params = @_;
    my $prefix = $params{Prefix} || '';
    my $action = $params{Action} || '';
    my $preset = $params{Preset} || '';
    
    # Adjust prefix, so that the name of this criterion is tacked on the end:
    $prefix .= $self->name . '.';

    if ($self->type eq 'FIXED') {
        $HTML -> TD;
        $self->print_fields(Prefix=>$prefix, Action=>$action, Preset=>$preset);
        $HTML ->_TD;
        $HTML -> TD ->t($self->comment)->_TD;
    } else {

        my $name = $self->variant || $self->name;
        # Table cell 1a: the name:
        $HTML -> TD;
    	$HTML -> t_B("*")          if $params{Invalid};
        $HTML -> t_B($name, ": ");
        $HTML -> t("(required)") if $self->is_required;

    	# Table cell 1b: the comment and directions:
   	$HTML -> BR -> small;
    	$HTML -> t($self->comment);
    	$HTML -> t_B("\nDirections:\n") -> t($self->directions);
        $self->info;
    	$HTML -> _small -> _TD;
    
    	# Table cell 2: the fill-in fields:
    	$HTML -> TD(NOWRAP=>undef);
    	$self->print_fields(Prefix=>$prefix, Action=>$action, Preset=>$preset);
        $HTML -> _TD;
    }
    my $misc_URLs_ref = $self->{'MISC_URL'};
    my $furl;
    # If MISC_URLs are present
    if (int (@{$misc_URLs_ref})) {
       $HTML -> TD;
       # Print a link for each MISC_URL
       foreach $furl (@$misc_URLs_ref) {
            # Use the URL comment as the text to be linked.
            $HTML -> hypertext($furl->comment,
                               logged_url($furl->URL,'type'=>$furl->type),
                               TARGET=>'support');
            $HTML -> BR;
        }

       $HTML -> _TD; 
    }
    1;
}

#------------------------------------------------------------
# sub info
#------------------------------------------------------------
# Info to help user in filling the option
# Override in subclass if needed.

sub info { undef }


#------------------------------
#
# ODL INTERFACE...
#
#------------------------------

#------------------------------------------------------------
# read_ODL ODL
#------------------------------------------------------------
# I<Instance method.>  
# Given a criterion object and values embedded in ODL, set up the object.

sub read_ODL {
    my ($self, $odl) = @_;
    warn "UNFINISHED!";
}

#------------------------------------------------------------
# to_request_ODL ODL
#------------------------------------------------------------
# I<Instance method.>  
# Return an ODL group for this criterion with the contents of 
# a SPECIALIZED_CRITERIA group, such as one which might appear 
# in a PRODUCT_REQUEST.
#
# Raises a FATAL EXCEPTION if the criterion appears to not have any 
# user input assigned to it; to prevent this from happening (it's possible
# that an optional criterion won't have user input) invoke 
# user_made_selections() first to test if this criterion has been assigned 
# a user value or not.

sub to_request_ODL {
    my ($self, $passed_odl) = @_;
    
    # Start with the basics:
    my $crit_odl = IU::ODL->new_group;    
    my $param;
    foreach $param (qw(CATEGORY_NAME
		       VARIANT
		       CRITERIA_NAME
		       CRITERIA_TYPE)) {
	defined($self->{$param}) or next;
	$crit_odl->add_param($param => IU::ODL->quoted($self->{$param}));
    }

    # Add user selections:
    my $slct  = $self->user_selections;
    my $kind = (($self->type eq 'STRING' or $self->type eq 'FIXED') 
                ? 'quoted' : 'bare');
    ($self->add_selections_to_ODL($crit_odl, $kind)) 
	or $self->whine('000', "can't convert criterion to request ODL ",
			       "if no user-selections!");
    # If an odl was passed in then add the specialized criteria to that
    # and return the modified odl.
    if ($passed_odl) {
       $passed_odl->add_param(SPECIALIZED_CRITERIA=> $crit_odl);
       return $passed_odl;
    }
#    $special_odl->add_param(SPECIALIZED_CRITERIA=>$crit_odl);
    # No odl was passed in, so just return an criteria odl object.
    return $crit_odl;
}



#------------------------------
#
# ABSTRACT METHODS...
#
#------------------------------

#------------------------------------------------------------
# as_string PARAMHASH
#------------------------------------------------------------
# I<Instance method.>
# Return object as a string compact enough for printing on a single line.
# If PARAMHASH contains a true "Verbose" option, then a multiline
# version is okay (if it will be more informative).  
#
sub as_string { 
    puke 600, "you must override this abstract method";
}
sub to_string { 
    my $str = shift->as_string(@_);
    warn "to_string() is deprecated; please use as_string instead";
    $str;
}

#------------------------------------------------------------
# directions
#------------------------------------------------------------
# I<Instance method.>  
# Return generic input directions.
#
sub directions { "" };

#------------------------------------------------------------
# print_fields PARAMHASH
#------------------------------------------------------------
# I<Instance method.>
# Print just the HTML input fields.
#
#     Prefix =>  The prefix (like "ship.").  Optional; defaults to "".
#
sub print_fields { puke 600, "you must override this abstract method!" };

#------------------------------------------------------------
# read_CGI PARAMHASH
#------------------------------------------------------------
# I<Instance method.>
# Initiallize the user selections from CGI tags.  The PARAMHASH contains
# the following:
#
#     Prefix =>  The prefix (like "ship.").  Optional; defaults to "".
#
# Returns true on success.
#
# Raises a fatal exception if the user input is missing, or if it
# doesn't match the criteria's specifications.
#
sub read_CGI { puke 600, "you must override this abstract method!" };

#------------------------------------------------------------
# set_CGI PARAMHASH
#------------------------------------------------------------
# I<Instance method.>
# Initiallize the CGI tags from this object's settings.  The PARAMHASH contains
# the following:
#
#     Prefix =>  The prefix (like "ship.").  Optional; defaults to "".
#
sub set_CGI { puke 600, "you must override this abstract method!" };

#------------------------------------------------------------
# validate
#------------------------------------------------------------
# I<Instance method.>
# Fix up the object, and validate it, returning the self object.
# Raises a fatal exception if validation fails.
#
sub validate {
    $_[0];
}


# DEPRECATED METHODS:
sub fill_in_from_CGI      { shift->read_CGI(@_)  }
sub selections_to_string  { shift->as_string(@_) }

#------------------------------------------------------------
1;
