package IU::Criterion::Geo;

=head1 NAME

IU::Criterion::Geo - a geographic region

=head1 DESCRIPTION

A geographic region that is part of a query.

Although instances of this class can print out a form capable
of inputting any of the different classes of geoselection objects (GeoPoint,
GeoRange, etc.), there is only ever assumed to be ONE geoselection.
These are the possibilities (* means that it was not supported
at the time of this writing):

       Geoselection type:        Geo object is:
       ----------------------------------------------------------
       POINT_LOC                 IU::Geo::Point
       RANGE_LOC                 IU::Geo::Range
     * POLYGON_LOC               IU::Geo::Polygon
     * PATH_ROW_LOC              IU::Geo::PathRow
     * GLOBAL_GRANULES_ONLY      IU::Geo::Global

=cut

use IU;
use IU::Log;
use IU::IW;
use IU::Criterion;

use strict;
use vars qw(@ISA);

use IU::Geo::Global;
use IU::Geo::Range;
use IU::Geo::Polygon;
use IU::Geo::PathRow;

@ISA = qw(IU::Criterion);




#------------------------------------------------------------
# directions
#------------------------------------------------------------
# Instance method, concrete override.
# Return the generic input directions.  See superclass for details.
#
sub directions {
    my $self = shift;
    my $dir = "";
    my %allowed = map {$_=>1} $self->valids;
    my @choices = 
     (($allowed{BY_POINT_LOC} ? "a single latitude/longitude point with " : ()),
      ($allowed{BY_POLYGON_LOC} ? 
       "four latitude/longitude corner points with" : ()),
      ($allowed{BY_RANGE_LOC} ? "a range of " : ()));
    $dir .= "Enter " . join(" OR ", @choices) . " \n\n";
    $dir .= "latitudes [-90 south to 90 north] and ".
            "longitudes [-180 west to 180 east]. \n\n ".
            "Format as either DD.DDDD or DD:MM:SS. \n\n";
    $dir;
}
#------------------------------------------------------------
# info 
#------------------------------------------------------------
# Display granule coverage:

sub info {
   my $self = shift;

   if ($self->coverage) {
      $HTML -> P -> t_B('Geographic Coordinates of data granule are: ');
      my @coords = split (/;/, $self->coverage->area_pretty);
      if (int(@coords) == 1) {
         print $coords[0], "\n";
         $HTML -> BR;
      } else {
         $HTML -> BR;
         print '&nbsp';     # Need a leading blank for first coord.
         my $coord;
         foreach $coord (@coords) {
            $HTML -> hspace(20);
            print $coord, "\n";
            $HTML -> BR;
         } # end foreach $coord
      } # end if (int(@coords) == 1)
   } # end if ($coverage->area)
}
#------------------------------------------------------------
# geo GEO 
#------------------------------------------------------------
# Instance method.
# Get/set the geoselection.  The GEO is an IU::Geo::* object.
# Returns the current geoselection. or undef on failure.
#
sub geo {
    shift->field('GEO', @_);
}

#------------------------------------------------------------
# coverage COVERAGE 
#------------------------------------------------------------
# Instance method.
# Get/set the coverage.  The COVERAGE is an IU::Coverage object.
# Returns the granule coverage  or undef on failure.
# Note: the coverage type may not be the same as the GEO type.

sub coverage {
    my ($self, $value) = @_;
    $self->{COVERAGE} = $value if (@_ > 1);
    $self->{COVERAGE};
}

#------------------------------------------------------------
# feature
#------------------------------------------------------------
# name of the geo point or region, e.g. Atlanta, Grand Canyon.

sub feature {
    my ($self) = @_;
    my $geo = $self->geo || return undef;
    my $feature = $geo->feature || undef;
    return $feature;
}

#------------------------------------------------------------
# as_string PARAMHASH
#------------------------------------------------------------
# Instance method, concrete override.
# Return selected values as a string.  See superclass for details.
#
sub as_string {
    my ($self, %params) = @_;
    my $geo = $self->geo || return undef;       # get geo or return

    # Get basic string:
    my $str = $geo->as_string(%params);
    return $str unless $params{Verbose};
    
    # Make it wordy:
    $str =~ s{;}{\n}g;
    if    ($geo->ODL_type eq 'RANGE_LOC') {
	$str = "RANGE:\n$str";
    }
    elsif ($geo->ODL_type eq 'POLYGON_LOC') {
	$str = "POLYGON:\n$str";
    }
    elsif ($geo->ODL_type eq 'PATH_ROW_LOC') {
	$str = "PATH/ROW:\n$str";
    }
    elsif ($geo->ODL_type eq 'POINT_LOC') {
        $str = "POINT:\n$str";
    }
    return $str;
}




#------------------------------
#
# CGI/ODL interface...
#
#------------------------------

#------------------------------------------------------------
# print_fields PARAMHASH
#------------------------------------------------------------
# Instance method, concrete override.
# Output basic CGI input fields.  See superclass for details.
#
sub print_fields {
    my ($self, %params) = @_;
    my $prefix = $params{Prefix};
    my %allowed = map {$_=>1} $self->valids;
    my $output;

    # Point:
    if ($allowed{BY_POINT_LOC}) {
	$HTML -> t("Lat: ") -> input_text(Name=>"${prefix}lat", Size=>10);
	$HTML -> BR;
	$HTML -> t("Lon: ") -> input_text(Name=>"${prefix}lon", Size=>10);
	++$output;
    }

    # Range:
    if ($allowed{BY_RANGE_LOC}) {
        # Set CGI tags so Range form is initialized with the granule's coverage.
#        if ($coverage->type eq 'RANGE') {
#           # create geo object from the coverage object:
#           my $geo_coverage = new IU::Geo::Range; 
#           my $odl = $coverage->to_ODL;
#           $geo_coverage->from_ODL($odl->param('RANGE_LOC'));
#           # tell geo object to output CGI tags:
#           $geo_coverage->set_CGI(${prefix})
#        } 
#        $REQ->param("${prefix}north"=>'180.0');

        $HTML -> TABLE;
        # Row 1: north lat...
        $HTML -> TR;
        $HTML -> TD(COLSPAN=>3, NOWRAP=>undef, ALIGN=>'CENTER')
               -> field_cell(Title=>"North latitude", Stack=>1)
               -> input_text(Size=>10, Name=>"${prefix}north")
               -> _field_cell;
        $HTML  -> _TD;
        $HTML  -> _TR;

        # Row 2: west and east lon...
        $HTML -> TR;
        $HTML -> TD(NOWRAP=>undef, ALIGN=>'CENTER')
                     -> field_cell(Title=>"West longitude", Stack=>1)
                     -> input_text(Size=>10, Name=>"${prefix}west")
                     -> _field_cell
                     -> _TD
              -> TD(NOWRAP=>undef, ALIGN=>'CENTER')
                     -> field_cell(Title=>"East longitude", Stack=>1)
                     -> input_text(Size=>10, Name=>"${prefix}east")
                     -> _field_cell
                     -> _TD
                     -> _TR;

        # Row 3: south lat...
        $HTML -> TR;
        $HTML -> TD(COLSPAN=>3, NOWRAP=>undef, ALIGN=>'CENTER')
              -> field_cell(Title=>"South latitude", Stack=>1)
              -> input_text(Size=>10, Name=>"${prefix}south")
              -> _field_cell
              -> _TD
              -> _TR;
        $HTML -> _TABLE;
	++$output;
    }

    # Polygon:
    if ($allowed{BY_POLYGON_LOC}) {
      $HTML -> HR         if $output;      # need separator
      $HTML -> t("Upper Left Corner:")
            -> t("Lat:") ->input_text(Name=>"${prefix}0.lat", Size=>10)
            -> t("Lon:") ->input_text(Name=>"${prefix}0.lon", Size=>10)
            -> BR;
      $HTML -> t("Upper Right Corner:")
            -> t("Lat:") ->input_text(Name=>"${prefix}1.lat", Size=>10)
            -> t("Lon:") ->input_text(Name=>"${prefix}1.lon", Size=>10)
            -> BR;
      $HTML -> t("Lower Right Corner:")
            -> t("Lat:") ->input_text(Name=>"${prefix}2.lat", Size=>10)
            -> t("Lon:") ->input_text(Name=>"${prefix}2.lon", Size=>10)
            -> BR;
      $HTML -> t("Lower Left Corner:")
            -> t("Lat:") ->input_text(Name=>"${prefix}3.lat", Size=>10)
            -> t("Lon:") ->input_text(Name=>"${prefix}3.lon", Size=>10)
            -> BR;
      # hardcode projection to be "Orthograhic".
      $HTML -> hidden("${prefix}projection",'ORTHOGRAPHIC');
    }
}

#------------------------------------------------------------
# read_CGI PARAMHASH
#------------------------------------------------------------
# Instance method, concrete override.
# Fill in geoselection from CGI params.  See superclass for details.
#
sub read_CGI {
    my ($self, %params) = @_;
    my $prefix = $params{Prefix};

    # Clear any existing entry:
    $self->geo(undef);
    foreach (1) {
	# If there was an attempt at entering a point...    
        if (! $self->all_blank($prefix, qw(lat lon))) {
	    $self->geo(IU::Geo::Point->from_CGI($prefix)->fixup);
	    last;
	}

	# If there was an attempt at entering a range...
	if (! $self->all_blank($prefix, qw(north south east west))) {
	    $self->geo(IU::Geo::Range->from_CGI($prefix)->fixup);
	    last;
	}

        # If there was an attempt at entering a polygon 
        if (! $self->all_blank($prefix, qw(1.lat 1.lon 2.lat 2.lon 
                                           3.lat 3.lon 4.lat 4.lon))) {
            $self->geo(IU::Geo::Polygon->from_CGI($prefix)->fixup);
            last;
        }

	# Nothing!
	$self->is_required ? $self->whine(516, "required, but empty")
	                   : return undef; 
    };
    1;  # done!
}

#------------------------------------------------------------
# set_CGI PARAMHASH
#------------------------------------------------------------
# Instance method, concrete override.
# Set CGI tags from the object.  See superclass for details.
#
sub set_CGI {
    my ($self, %params) = @_;    
    my $geo = $self->geo;
    $geo->set_CGI($params{Prefix}) if $geo;
    1;
}

#------------------------------------------------------------
# add_selections_to_ODL ODL,KIND
#------------------------------------------------------------
# Instance method, concrete override.
# Add selections to ODL group.  See superclass for details.
#
sub add_selections_to_ODL {
    my ($self, $odl, $kind) = @_;
    my $geo = $self->geo || return undef;       # get geo object or fail
    $odl->add_param('CRITERIA_VALUE' => IU::ODL->quoted('BY_'.$geo->ODL_type));
    $odl->add_param($geo->ODL_type => $geo->to_ODL);
    1;
}

#------------------------------------------------------------
1;
