package IU::Criterion::InputRange;

=head1 NAME

IU::Criterion::InputRange - a range in a query

=head1 DESCRIPTION

Input a range of two numbers/dates.

=cut

use IU;
use IU::Log;
use IU::IW;
use IU::Criterion;

use IW::Date;       # KLUDGE, but IW::Date should be changed to IU::Date!

use strict;
use vars qw(@ISA);

@ISA = qw(IU::Criterion);


#------------------------------------------------------------
# as_string
#------------------------------------------------------------
# Instance method, concrete override.
# Return selected values as a string.  See superclass for details.
#
sub as_string {
    my ($self, %params) = @_;
    my $min = $self->{UserSlct}{MIN}; 
    my $max = $self->{UserSlct}{MAX}; 
    return undef if (!defined($min) and !defined($max));  # nothing selected

    # Clean up:
    my %words;
    if ($self->type eq 'DATE') {
	$min = IW::Date::TZ_to_datetime($min) if $min;
	$max = IW::Date::TZ_to_datetime($max) if $max;
#	%words = (Min=>'EARLIEST AVAILABLE', Max=>'LATEST AVAILABLE', 
	%words = (Min  =>'01 Jan 1950, 00:00:00 (default)', 
		  Max  =>'Current Date (default)',
		  From =>'Start', 
		  To   =>'Stop');
    }
    else {
	%words = (Min=>'LOWEST', Max=>'HIGHEST', 
		  From=>'Min', To=>'Max');
    }
    $min = defined($min) ? $min : $words{Min};
    $max = defined($max) ? $max : $words{Max};
    
    # Return:
    if ($params{Verbose})  { return "$words{From}: $min\n" .
				    "$words{To}: $max" }
    else                   { return "$min to $max"  }
}

#------------------------------------------------------------
# directions
#------------------------------------------------------------
# Instance method, concrete override.
# Return the generic input directions.  See superclass for details.
#
sub directions {
    my $self = shift;
    my $min = $self->value_to_CGI($self->min);
    my $max = $self->value_to_CGI($self->max);
    my $dir = "Enter a range of ".$self->type."(s), from $min to $max. " .
    "Supplying only one value causes the other to default to the min/max.";
    if ($self->type eq 'DATE') {
       $dir .= " Leave both values blank to get entire time range.".
               " Format: YYYY-MM-DD HH:MM:SS"; 
    }
    return $dir;
}

#------------------------------------------------------------
# validate
#------------------------------------------------------------
# Instance method, concrete override.
# Fixup and validate the user input...
#
#   1. Raises a fatal exception if the types of MIN or MAX are wrong.
#   2. If both MIN and MAX are missing, raises a fatal exception if this 
#      field is required, and returns undef if it is not.
#   3. If just one of MIN or MAX is missing, the other value will be filled 
#      in using the defined range boundaries (if they exist).
#
sub validate {
    my $self = shift;
    my $user = $self->{UserSlct};
    my $field;

    # Check types and ranges, fixing up if needed:
    foreach $field (qw(MIN MAX)) {
	next if (!defined($user->{$field}));    # no need to check undefs
	$user->{$field} = $self->valid_type($user->{$field});
	$self->valid_range($user->{$field});
    }
    # ...at this point, any defined values are clean!
   
    # If we got nothing (possible if this was optional?) then return:
    if (!defined($user->{'MIN'} and !defined($user->{'MAX'}))) {
	$self->is_required ? $self->whine(516,"both ends of range are missing")
	                   : return undef;
    }

    # Only one given?  Default other field...
    defined($user->{'MIN'}) or $user->{'MIN'} = $self->min;
    defined($user->{'MAX'}) or $user->{'MAX'} = $self->max;

    # If we have two at this point, swap if needed:
    if (defined($user->{MIN}) && defined($user->{MAX})) { 
	if (($self->type eq 'DATE')                # swap if...
	    ? ($user->{MAX} lt $user->{MIN})       # maxdate < mindate, or...
	    : ($user->{MAX} < $user->{MIN})) {     # maxnum < minnum
	    ($user->{MIN}, $user->{MAX}) = ($user->{MAX}, $user->{MIN});   
	}
    }
    return 1;
}


#------------------------------
#
# CGI/ODL interface...
#
#------------------------------

#------------------------------------------------------------
# print_fields PARAMHASH
#------------------------------------------------------------
# Instance method, concrete override.
# Output basic CGI input fields.  See superclass for details.
#
# Action => Either Reset/Clear
# Prefix => Name of field

sub print_fields {
    my ($self, %params) = @_;
    my $prefix = $params{Prefix};
    my $action = $params{Action};

    # Get min and max:
    my $min = $self->value_to_CGI($self->min);
    my $max = $self->value_to_CGI($self->max);;
    
    # Get size:
    my $size = (($self->type eq 'DATE') ? 20 : 11);    

    # If the valids have default values we need to set'em?
    my @default = $self->defaults;

    # As of 2.2 we've added a Clear button to the Extended Valids
    # If the Clear button should remove all values from the fields.
    # If the page is first load/Reseted then it should display the FCFS value :
    #           selected value | default value 
    #
    # Reset only displays the Default values

    if ( $action eq "Clear" ) {
        $HTML -> input_text(Name=>"${prefix}min", 
                            Value =>"",
                            Default=>$default[0],  
                            Size=>$size);
        $HTML -> t("min to ");
        $HTML -> BR if ($size > 15);
        $HTML -> input_text(Name=>"${prefix}max", 
                            Value =>"", 
                            Default=>$default[1], 
                            Size=>$size);
        $HTML -> t("max");
    }
    elsif ($action eq "Reset" ) {
        $HTML -> input_text(Name=>"${prefix}min", 
                            Value=> (defined($default[0]) ? $default[0] : ""),
                            Size=>$size);
        $HTML -> t("min to ");
        $HTML -> BR if ($size > 15);
        $HTML -> input_text(Name=>"${prefix}max", 
                            Value=> ( defined($default[1]) ? $default[1] : ""),
                            Size=>$size);
        $HTML -> t("max");
    }
    else {
        $HTML -> input_text(Name=>"${prefix}min", 
                            Value=>($REQ->param("${prefix}min") || $default[0]),
                            Size=>$size);
        $HTML -> t("min to ");
        $HTML -> BR if ($size > 15);
        $HTML -> input_text(Name=>"${prefix}max", 
                            Value=>($REQ->param("${prefix}max") || $default[1]),
                            Size=>$size);
        $HTML -> t("max");
    }
}

#------------------------------------------------------------
# read_CGI PARAMHASH
#------------------------------------------------------------
# Instance method, concrete override.
# Fill in geoselection from CGI params.  See superclass for details.
#
sub read_CGI {
    my ($self, %params) = @_;
    my $prefix = $params{Prefix};
    my $user = ($self->{UserSlct} = {});

    # Get min/max values:
    my $box;
    foreach $box (qw(min max)) {
	my $val = IW::Util::trim($REQ->param("${prefix}${box}"));
	next if IW::Util::blank($val);
	$user->{uc($box)} = $val;            # store it!
    }

    # Validate and return:
    return $self->validate;
}

#------------------------------------------------------------
# set_CGI PARAMHASH
#------------------------------------------------------------
# Instance method, concrete override.
# Set CGI tags from the object.  See superclass for details.
#
sub set_CGI {
    my ($self, %params) = @_;
    my $prefix = $params{Prefix};

    # Get it:
    my $tag;
    foreach $tag (qw(min max)) {
	my $val = $self->value_to_CGI($self->{UserSlct}{uc($tag)});
	$REQ->param("${prefix}$tag" => $val);
    }
}

#------------------------------------------------------------
# read_ODL ODL
#------------------------------------------------------------
# Instance method, concrete override.
# Fill in selections from ODL.
#
sub read_ODL {
    my ($self, $odl) = @_;
    my $value_odl;
    $self->{UserSlct} = {};
    if ($value_odl = $odl->param('CRITERIA_MIN')) {
	$self->{UserSlct}{MIN} = $value_odl->text;
    }
    if ($value_odl = $odl->param('CRITERIA_MAX')) {
	$self->{UserSlct}{MAX} = $value_odl->text;
    }
    1;
}

#------------------------------------------------------------
# add_selections_to_ODL ODL,KIND
#------------------------------------------------------------
# Instance method, concrete override.
# Add selections to ODL group.  See superclass for details.
#
sub add_selections_to_ODL {
    my ($self, $odl, $kind) = @_;
    my $min = $self->{UserSlct}{MIN};
    my $max = $self->{UserSlct}{MAX};
    (defined($min) || defined($max)) or return undef;   # either alone is ok

    $odl->add_param(CRITERIA_MIN => IU::ODL->$kind($min))  if defined($min);
    $odl->add_param(CRITERIA_MAX => IU::ODL->$kind($max))  if defined($max);
    1;
}

#------------------------------------------------------------
1;
