package IU::Criterion::Latitude;

=head1 NAME

IU::Criterion::Latitude - a latitude in degrees:minutes:seconds

=cut

use IU;
use IU::Log;
use IU::Coord;
use IU::IW;
use IU::Criterion;

use strict;
use vars qw(@ISA);

@ISA = qw(IU::Criterion);


#------------------------------------------------------------
# as_string
#------------------------------------------------------------
# Instance method, concrete override.
# Return selected values as a string.  See superclass for details.
#
sub as_string {
        my($self,%params) = @_;
        if ($self->is_obscured()) {
            "(not shown)";
        }
        else {
            $self->{UserSlct}{VALUE};
        }
}

#------------------------------------------------------------
# directions
#------------------------------------------------------------
# Instance method, concrete override.
# Return the generic input directions.  See superclass for details.
#
sub directions {
    my $self = shift;
    "Enter a string of characters that represents Latitude."; 
}



#------------------------------
#
# CGI/ODL interface...
#
#------------------------------

#------------------------------------------------------------
# print_fields PARAMHASH
#------------------------------------------------------------
# Instance method, concrete override.
# Output basic CGI input fields.  See superclass for details.
#
sub print_fields {
    my ($self, %params) = @_;
    my $prefix = $params{Prefix};
    my $preset = $params{Preset};

    my @default = $self->defaults;
    my $newdefault = $preset || $default[0];

    my $maxlen = $self->max_len;
    my $obscured = $self->is_obscured();
    if ($obscured) {
        $HTML -> input_text(Name=>"${prefix}val", Default=>$newdefault, 
                            Size=>$maxlen, Type=>"password");
        debug "Field ${prefix}val was obscured";
    }
    elsif ($maxlen <= 25) {
	$HTML -> input_text(Name=>"${prefix}val", Default=>$newdefault, 
                            Size=>$maxlen);
    }
    else {
	$HTML -> input_textarea(Name=>"${prefix}val", Default=>$newdefault,
                                Cols=>25, Rows=>3);
    }
}

#------------------------------------------------------------
# read_CGI PARAMHASH
#------------------------------------------------------------
# Instance method, concrete override.
# Fill in geoselection from CGI params.  See superclass for details.
#
sub read_CGI {
    my ($self, %params) = @_;
    my $prefix = $params{Prefix};

    debug "prefix is $prefix";
    # Get value, and check it:
    my $val = $REQ->param("${prefix}val");
    $val = str2dms($val);
    if (IW::Util::blank($val)) {     # KLUDGE
        $self->is_required ? $self->whine(516, "required, but no value given")
                           : return undef;
    }
    my $ok = $self->valid_type($val);     # check against type, and clean up
    $val = undef if(!$ok);
    $REQ->delete("${prefix}val") if(!$ok);
    

    # Okay!
    $self->{UserSlct} = {VALUE=>$val};
    return 1;
}

#------------------------------------------------------------
# set_CGI PARAMHASH
#------------------------------------------------------------
# Instance method, concrete override.
# Set CGI tags from the object.  See superclass for details.
#
sub set_CGI {
    my ($self, %params) = @_;
    my $prefix = $params{Prefix};

    # Get it:
    $REQ->param("${prefix}val" => $self->{UserSlct}{VALUE});
}

#------------------------------------------------------------
# read_ODL ODL
#------------------------------------------------------------
# Instance method, concrete override.
# Fill in selections from ODL.
#
sub read_ODL {
    my ($self, $odl) = @_;
    my $value_odl;
    $self->{UserSlct} = {};
    if ($value_odl = $odl->param('CRITERIA_VALUE')) {
	$self->{UserSlct}{VALUE} = $value_odl->text;
    }
}

#------------------------------------------------------------
# add_selections_to_ODL ODL,KIND
#------------------------------------------------------------
# Instance method, concrete override.
# Add selections to ODL group.  See superclass for details.
#
sub add_selections_to_ODL {
    my ($self, $odl, $kind) = @_;
    my $val = $self->{UserSlct}{VALUE}; defined($val) or return undef;

    $odl->add_param(CRITERIA_VALUE => IU::ODL->quoted($val));
    1;
}

#------------------------------------------------------------
# valid_type INPUT
#------------------------------------------------------------
# I<Instance method.>
# Asserts that INPUT matches type $self->type.
# Fatal exception if not.
# Alters argument in-place.

sub valid_type {
   my ($self, $str) = @_;
   my $input = $str;
   my $hemi;
   my $angle;
   my $ok;
   defined($str) or $str = '';    # prevent those warnings
   $str =~ s/^\s+//;
   if ($str =~ /^([+-])(.*)$/i) {
       $hemi = $1;
       ($str = $2) =~ s/^\s+//;
   }
   if ($str =~ /^(\d+)[:dDo*](\d+)[:''](\d+[.,]?\d*)[""]?/) {
      # Matches ``123:34:56.78'' and related patterns
   }
   $ok = 1 if($1 <= 90.0 && $2 >= 0.0 && $2 < 60.0 && $3 >= 0.0 && $3 < 60.0); 

   $angle = $1;
   if (defined($2)) { $angle += $2/60; }
   if (defined($3)) { $angle += $3/60/60; }
   $ok = 0 if($angle > 90.0);
   $ok or $self->whine(511, "input '$input' is not a good ", $self->type);
   return $ok;
}

#------------------------------------------------------------
1;

