package IU::Criterion::Time;

=head1 NAME

IU::Criterion::Time - a time range in a query


=head1 DESCRIPTION

Input a time range of two dates.  Optionally, include a START_DAY_OF_YEAR
and/or STOP_DAY_OF_YEAR

=cut

use IU;
use IU::Log;
use IU::IW;
use IU::Criterion;

use IW::Date;       # KLUDGE, but IW::Date should be changed to IU::Date!

use strict;
use vars qw(@ISA);

@ISA = qw(IU::Criterion);

#------------------------------------------------------------
# new []
#------------------------------------------------------------
# I<Class method, constructor.>
#
sub new {
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    my @values = @{$self->{VALUES}};
    my $start_tz = $values[0];
    my $stop_tz = $values[1];
    my $start_doy = $values[2];
    my $stop_doy = $values[3];
    debug "Start=".$start_tz." end=".$stop_tz." sdoy=".$start_doy." stop=".$stop_doy;
    $self->{Date} = (new IU::Criterion::InputRange
             CRITERIA_NAME => 'TIME_RANGE',
             CRITERIA_TYPE => 'DATE');
    $self->{Date}->field('MIN', $start_tz);
    $self->{Date}->field('MAX', $stop_tz);
    $self->{DOY} = (new IU::Criterion::InputRange
             CRITERIA_NAME => 'DAY_RANGE',
             CRITERIA_TYPE => 'INTEGER');
    $self->{DOY}->field('MIN', $start_doy);
    $self->{DOY}->field('MAX', $stop_doy);
    debug "In time start_doy=", $self->startDayOfYear;
    $self->validate;

    debug "In time";
    debug "In time start_doy=", $self->startDayOfYear;
    $self->{UserSlct} = {};
    bless $self, $class;
    $self;
}

#------------------------------------------------------------
# _one VARNAME,SELF,[VALUE]
#------------------------------------------------------------
# I<Private utility.>  Get/set a single-valued instance variable.
sub _one {
    my ($varname, $self, $value) = @_;
    $self->{$varname} = $value if (@_ > 2);
    $self->{$varname};
}

# startDate - returns the start date
sub startDate {
    my $self = shift;
    $self->{Date}->field('MIN');
}

# stopDate - returns the stop date
sub stopDate {
    my $self = shift;
    $self->{Date}->field('MAX');
}

# startDayOfYear - returns the start day of year
sub startDayOfYear {
    my $self = shift;
    $self->{DOY}->field('MIN');
}

# stopDayOfYear - returns the stop day of year
sub stopDayOfYear {
    my $self = shift;
    $self->{DOY}->field('MAX');
}

#------------------------------------------------------------
#------------------------------------------------------------
# as_string
#------------------------------------------------------------
# Instance method, concrete override.
# Return selected values as a string.  See superclass for details.
#
sub as_string {
    my $self = shift;
    my $min;
    my $max;
    my %words;
    $min = $self->{DOY}->{UserSlct}{MIN};
    $max = $self->{DOY}->{UserSlct}{MAX};
    debug "in as_string";
    if (defined($self->{Date}) && (defined($min) || defined($max))) {
	%words = (Min=>'1', Max=>'366',
	 From=>'Start day of year', To=>'End day of year');
	$min = defined($min) ? $min : $words{Min};
	$max = defined($max) ? $max : $words{Max};
        $self->{Date}->as_string(@_)."\n"."$words{From}: $min\n" .
	 "$words{To}: $max";
    } elsif (defined($self->{Date})) {
        $self->{Date}->as_string(@_);
    } else {
	return "(The time period covered by the data.)";
    }

}

#------------------------------------------------------------
# validate
#------------------------------------------------------------
# Instance method, concrete override.
# Fixup and validate the user input...
#
#   1. Raises a fatal exception if the types of MIN or MAX are wrong.
#   2. If both MIN and MAX are missing, raises a fatal exception if this 
#      field is required, and returns undef if it is not.
#   3. If just one of MIN or MAX is missing, the other value will be filled 
#      in using the defined range boundaries (if they exist).
#
sub validate {
    my $self = shift;
    my $min;
    my $max;
    $min = $self->{DOY}->{UserSlct}{MIN};
    $max = $self->{DOY}->{UserSlct}{MAX};
    debug "in validate";
    $self->{Date}->validate();
    if (defined($min) && ($min < 1 || $min > 366)) {
	$self->whine(511,"Start day of year not valid!");
    }
    if (defined($max) && ($max < 1 || $max > 366)) {
	$self->whine(511,"End day of year not valid!");
    }
    if (!defined($min) && defined($max)) {
	$self->{DOY}->{UserSlct}{MIN} = 1;
	$min = 1;
    }
    if (defined($min) && !defined($max)) {
        $self->{DOY}->{UserSlct}{MAX} = 366;
	$max = 366;
    }
    if ($max < $min) {
	$self->{DOY}->{UserSlct}{MIN} = $max;
	$self->{DOY}->{UserSlct}{MAX} = $min;
    }
    return 1;
}

#------------------------------------------------------------
1;
