package IU::DAACExpansions;

=head1 NAME

IU::DAACExpansions - DAAC short-name to long-name utility by 
                     parsing the archive.odl

=head1 DESCRIPTION

This routine centralizes the mapping of DAAC short names to long names.
We intend (in a future release) to pick up all this from the archive.odl.

=head1 SYNOPSIS

    # Create empty DAACExpansions
    my $daacAddress = new IU::DAACExpansions;

    # Initialize mapping hash with DAACs ids and 
    # DAACs long names read from archive.odl
    $daacAddress->parse_ODL();

    # Fetch a reference to a copy of the mapping hash
    my $hashref = DAACExpansions->get();

    # Fetch the hash contents
    my %hash    = DAACExpansions->get();

=cut

use IU;
use IU::ODL;
use IU::Log;
use strict;

#---------------------------------------------------------
# Class variables
#---------------------------------------------------------
#
# Mapping hash to store DataCenterId => DataCenterName
# example: "ASF" => "Alaska SAR Facility"

my %DataCenterCmnts;

#------------------------------------------------------------
# new PARAMHASH
#------------------------------------------------------------
sub new {
    my ($class, %params) = @_;
    bless \%params, $class;
}

#------------------------------------------------------------
# Get the path for the archive.odl file, then parse it.
#------------------------------------------------------------
sub parse_ODL  {
my $self = shift;

    # Read file 
    my $path_to_archiveODL  = "$IU::GaeaDataDir/archive.odl";
    open ODL, $path_to_archiveODL or puke "open $path_to_archiveODL: $!";

    # Parse it into an ODL object:
    my $odl = eval { parse IU::ODL \*ODL };
    close ODL;
    debug "parsed archive.odl";

    # Make sure we have the file parsed into an odl object.
    $odl or puke "couldn't parse ODL in $path_to_archiveODL (line $.)";

    # Convert the ODL into a query object and get the DAACs Names:
    $self->get_names_from_ODL($odl);

    return;
}

#------------------------------------------------------------
# Parse the contents of DATA_CENTER_INFO and initialize the 
# mapping hash with data center short and long names. 
#------------------------------------------------------------
sub get_names_from_ODL  {
    my ($self, $odl, %params) = @_;
    $odl || return undef;
    my ($param, $value);
    my $search;
 
    if ($search = $odl->param('DATA_CENTER_INFO')) {
	while (($param, $value) = $search->next) {
		if ($param eq 'DATA_CENTER') {
				my $tmp_daac = $self->parse_content($value);	
                                $DataCenterCmnts{$tmp_daac->{DataCenterId}} =
                                            $tmp_daac->{DataCenterName};
			 # debug "$tmp_daac->{DataCenterId}} = $tmp_daac";
		}
	}
    }

    return;
}

#------------------------------------------------------------
# Parse the contents of short (DATA_CENTER_ID) and 
# long (DATA_CENTER_NAME) names.
#------------------------------------------------------------
sub parse_content {
    my $class = shift;
    my $odl = shift;   

    my $self = $class->new;
    my $value;
    if ($value = $odl->param('DATA_CENTER_ID')) {
         $self->{DataCenterId} = $value->text_value;
#	 debug "Daac : ".$self->{DataCenterId};
    }
    if ($value = $odl->param('DATA_CENTER_NAME')) {
        $self->{DataCenterName} = $value->text_value;
    }
     $self;
}

#------------------------------------------------------------
#
# Access methods...
#
#------------------------------------------------------------
#---------------------------------------------------------
# get
#---------------------------------------------------------
#    Make a copy of the mapping hash.
#    Return either it or a pointer to it, depending on
#        what the caller wants.

sub get {
   my $class = shift;
   my %hashcopy = %DataCenterCmnts;
   return (wantarray ? %hashcopy : \%hashcopy);
}

#------------------------------------------------------------
1;
