package IU::Dataset;


=head1 NAME

IU::Dataset -- objects representing datasets returned from a search


=head1 AUTHOR

Eryq (Erik Dorfman), Hughes STX Corporation


=head1 INSTANCE VARIABLES

The basic ones are:

    ArchiveId   => archive id (e.g., GSFC)
    BrowseDesc  => browse description
    Comment     => comment
    Coverage    => spatial coverage; an IU::Coverage
    Id          => dataset id
    Key         => dataset key (e.g., 3/245)
    NGranules   => number of granules returned by the search
    NGranuleHits=> number of granules actually hit by the search
    OrigCenter  => originating center
    Restriction => restrictions

The MD-oriented ones (normally only asigned after directory search) are
as follows (these probably belong in a sub-object):

    MDArchInfo  => Master Directory archive info
    MDAttribute => Master Directory attribute
    MDEntry     => Master Directory entry id
    MDPersonnel => Master Directory personnel
    MDReference => Master Directory reference
    MDStartTime => Master Directory start time
    MDStopTime  => Master Directory stop time
    MDTitle     => Master Directory title

The URL-oriented ones are:

    DATA_URL               => array of IU::FancyURLs for any on-line data
    DATASET_HOME_PAGE      => array of IU::FancyURLs for any home pages
    SPECIALIZED_SEARCH_URL => array of IU::FancyURLs for special pages
    MISC_URL               => array of IU::FancyURLs for misc links
    BROWSE_URL             => array of IU::FancyURLs for browse links
    GUIDE_URL		   => array of IU::FancyURLs for ECS to send guide 
                              documents to the EDG

The following are actually dependent on the granules returned by the search:

    Sensors     => all sensors covered by returned granules
    Sources     => all sources covered by returned granules
    StartDate   => earliest start date of returned granules
    StopDate    => latest stop date of returned granules

The following are produced at a TBD stage in the dataset object's 
lifetime (most likely during imsgdbdump):

    AllowsEmailOrder => if a granule has no packages, can it be email ordered?
    ContactAddr      => contact address for this dataset


These are special:
    PkgCombos  => the package combos currently selected for this dataset
    The Key is special; all others come directly from the search results.

=head1 RCS INFO

$Id: Dataset.pm,v 1.42 2002/09/24 14:49:05 solomon Exp $

=cut

use IU::Log;
use IW::DaacAddress;
use IU::Chunk;
use IU::Storable;
use IU::Account;
use IU::Coverage;
use IU::FancyURL;
use IU::HandlingFee;
use IU::SpecializedResults;
use IU::Interface::Orderable;
use strict;

@IU::Dataset::ISA = qw(IU::Storable IU::Interface::Orderable);

#------------------------------------------------------------
#
# Globals...
#
#------------------------------------------------------------

# URL instance vars:
my @URLInstVars = qw(
    DATA_URL  
    DATASET_HOME_PAGE       
    SPECIALIZED_SEARCH_URL 
    MISC_URL               
    BROWSE_URL               
    GUIDE_URL
);

#------------------------------------------------------------
#
# Construction/destruction methods...
#
#------------------------------------------------------------

# For the AutoLoader:
DESTROY { 1 }

#------------------------------------------------------------
# new
#------------------------------------------------------------
# Create a new, empty dataset object.
# Returns the dataset object for success, false for failure (incluing EOF).

sub new {
    my $class = shift;
    my $self = bless {}, $class;
    $self->{Accounts} = [];
    $self;
}

#------------------------------------------------------------
# input FILEHANDLE
#------------------------------------------------------------
# Read a dataset chunk from FILEHANDLE into the internal data structure.
#
# Returns the dataset object for success, false for failure (incluing EOF).

sub input {
    my $class = shift;

    # Create new object, initiallizing from chunk:
    $class->from_chunk(IU::Chunk->input(@_));
}

#------------------------------------------------------------
# from_chunk CHUNK
#------------------------------------------------------------
# Read an old-style dataset chunk into the internal data structure.
# A dataset chunk containing "x-error" is considered null and void.
#
# Returns the dataset object for success, false for failure (incluing EOF).

sub from_chunk {
    my ($class, $chunk) = @_;
    my $errmsg;

    # Input chunk, checking for errors:
    defined($chunk) or return undef;
    defined($errmsg = $chunk->value('x-error'))
	and return error 454, $errmsg;

    # Create new object, initiallizing from chunk:
    my $self = {};

    # Set simple values:
    $self->{ArchiveId}   = $chunk->value('DATA_CENTER_ID');
    $self->{BrowseDesc}  = $chunk->value('BROWSE_PRODUCT_DESCRIPTION');
    $self->{Comment}     = $chunk->value('COMMENT');
    $self->{DisclaimerComment}  = $chunk->value('DISCLAIMER_COMMENT');
    $self->{DisclaimerUrl}      = $chunk->value('DISCLAIMER_URL');
    $self->{Comment}     = $chunk->value('COMMENT');
    $self->{Id}          = $chunk->value('DATASET_ID');
    $self->{Key}         = $chunk->value('DATASET_WEB_KEY');
    $self->{MDArchInfo}  = $chunk->value('MD_ARCHIVE_INFO');
    $self->{MDAttribute} = $chunk->value('MD_ATTRIBUTE');
    $self->{MDEntry}     = $chunk->value('MD_ENTRY') ||
	                   $chunk->value('MD_ENTRY_ID');
    $self->{MDPersonnel} = $chunk->value('MD_PERSONNEL');
    $self->{MDReference} = $chunk->value('MD_REFERENCE');
    $self->{MDStartTime} = $chunk->value('MD_START_TIME');
    $self->{MDStopTime}  = $chunk->value('MD_STOP_TIME');
    $self->{MDTitle}     = $chunk->value('MD_TITLE');
    $self->{NGranules}   = $chunk->value('NUMBER_OF_GRANULES');
    $self->{NGranuleHits}= $chunk->value('NUMBER_OF_GRANULE_HITS');
    $self->{OrigCenter}  = $chunk->value('ORG_CENTER');
    $self->{Restriction} = $chunk->value('RESTRICTION');
    $self->{StartDate}   = $chunk->value('START_DATE');
    $self->{StatusCodeCmnt} = $chunk->value('STATUS_CODE_COMMENT');
    $self->{StopDate}    = $chunk->value('STOP_DATE');

    # Multi-valued fields:
    $self->{PackageIds} = [ $chunk->seq_values('PACKAGE_ID')];
    $self->{Campaigns}  = [ $chunk->seq_values('CAMPAIGN')    ];
    $self->{Params}     = [ $chunk->seq_values('PARAMETER')   ];
    $self->{ProcLevels} = [ $chunk->seq_values('PROCESSING_LEVEL') ];
    $self->{Sensors}    = [ $chunk->seq_values('SENSOR_NAME') ];
    $self->{Sources}    = [ $chunk->seq_values('SOURCE_NAME') ];

    # Coverage:
    $self->{Coverage} = IU::Coverage->from_chunk($chunk) ||
	                (new IU::Coverage Type=>'EMPTY');

    my $results_chunk;
    foreach $results_chunk ($chunk->values('SPECIALIZED_RESULTS')) {
        my $result = IU::SpecializedResults->from_chunk($results_chunk);
        push @{$self->{Results}}, $result;
    }


    # Read in HANDLING_FEE group(s)
    my $fee_chunk;
    foreach $fee_chunk  ($chunk->values('HANDLING_FEE')) {
        my $fee = IU::HandlingFee->from_chunk($fee_chunk);
        push @{$self->{HandlingFee}}, $fee;
    }

    # Order-related stuff:
    $self->{AllowsEmailOrder} =	$chunk->value('allows-email-order'); 

    # Account info:
    $self->{Accounts} = [];
    my $account_chunk;
    foreach $account_chunk ($chunk->values('VALID_ACCOUNT'),
			    $chunk->values('VALID_ACCOUNTS')) {
	my $account = IU::Account->from_chunk($account_chunk) || next;
	push @{$self->{Accounts}}, $account;
    }    

    # Contact info:
    my $contact_chunk = $chunk->value('contact-addr');
    if ($contact_chunk) {
        $self->{ContactAddr} =
           IW::DaacAddress->from_chunk($contact_chunk);
    }

    # URLs (we tolerate more than one of each type):
    my $key;
    foreach $key (@URLInstVars) {

	# Allow many URLs of the current flavor... 
	$self->{$key} = [];
	my ($url_chunk, $furl);
	foreach $url_chunk ($chunk->values($key)) {
	    my $furl = IU::FancyURL->from_chunk($url_chunk, $key);
	    push @{$self->{$key}}, $furl;
	}
    }
    
    # Re-bless and return:
    bless $self, $class;
}

#------------------------------------------------------------
#
# Access methods...
#
#------------------------------------------------------------

#----------
# _one_scalar -- GENERIC, scalar valued, undef allowed...
#----------
sub _one_scalar { 
    my ($self, $ivar, $value) = @_;
    (@_ > 2) and $self->{$ivar} = $value;
    $self->{$ivar};
}

#----------
# _string_set -- GENERIC, array-valued, arrayrefs returned
#----------
sub _string_set { 
    my ($self, $ivar, $value) = @_;
    if (@_ > 2) {
	if(defined(@$value)) {
	    my %uniq;
	    local $_;
	    foreach (@$value) {
		next unless defined $_;
		$uniq{$_}++;
	    }
	    $self->{$ivar} = [keys %uniq];   # fresh copy, no dups
	}
    }
    $self->{$ivar} ||= [];
}

#----------
# Single-valued...
#----------

# archive_id   -- return the archive id
sub archive_id    { shift->_one_scalar('ArchiveId',@_) }

# browse_desc  -- what is the dataset's browse description?
sub browse_desc   { shift->_one_scalar('BrowseDesc',@_) }

# comment      -- return the dataset comment
sub comment       { shift->_one_scalar('Comment',@_) }

# coverage     -- return the coverage object
sub coverage      { shift->_one_scalar('Coverage',@_) }

# disclaimer_comment   -- return the dataset disclaimer/comment
sub disclaimer_comment { shift->_one_scalar('DisclaimerComment',@_) }

# disclaimer_url      -- return the URL pointing to the dataset disclaimer
sub disclaimer_url  { shift->_one_scalar('DisclaimerUrl',@_) }

# id           -- return the dataset id
sub id            { shift->_one_scalar('Id',@_) }
sub dataset_id    { shift->_one_scalar('Id',@_) }    ### to be "orderable"

# key          -- return the dataset key
sub key           { shift->_one_scalar('Key',@_) }
sub dataset_key   { shift->_one_scalar('Key',@_) }   ### to be "orderable"

# MD_arch_info -- return the MD archive info
sub MD_arch_info  { shift->_one_scalar('MDArchInfo',@_) }

# MD_attribute -- return the MD attribute
sub MD_attribute  { shift->_one_scalar('MDAttribute',@_) }

# MD_entry     -- return the MD entry id
sub MD_entry      { 
    my $self = shift;
    $self->{MDEntry} ? IW::Util::trim($self->{MDEntry}) : undef;
}

# MD_personnel -- return the MD personnel
sub MD_personnel  { shift->_one_scalar('MDPersonnel',@_) }

# MD_reference -- return the MD reference
sub MD_reference  { shift->_one_scalar('MDReference',@_) }

# MD_start_time -- return the MD start time
sub MD_start_time { shift->_one_scalar('MDStartTime',@_) }

# MD_stop_time -- return the MD stop time
sub MD_stop_time  { shift->_one_scalar('MDStopTime',@_) }

# MD_title     -- return the MD title
sub MD_title      { shift->_one_scalar('MDTitle',@_) }

# n_granules   -- return the number of granules
sub n_granules    { (shift->_one_scalar('NGranules',@_) || 0) }

# n_granule_hits  -- return the number of granules actually hit
sub n_granule_hits { (shift->_one_scalar('NGranuleHits',@_) || 0) }

# orig_center  -- return the dataset's originating center
sub orig_center   { shift->_one_scalar('OrigCenter',@_) }

# pkg_set_for  -- return package selection for dataset or granule
sub pkg_set_for   { shift->_one_scalar('PkgSetFor',@_) }

# restriction  -- return the dataset restriction
sub restriction   { shift->_one_scalar('Restriction',@_) }

# start_date   -- return the start date
sub start_date    { shift->_one_scalar('StartDate',@_) }

# status_code_comment  -- return the start date
sub status_code_comment { shift->_one_scalar('StatusCodeCmnt',@_) }

# stop_date    -- return the stop date
sub stop_date     { shift->_one_scalar('StopDate',@_) }

# Would you believe these are ALL DUMMY variables!!
# Since we're messing with ObjectList, we have to modify Granule AND Dataset
# to get on thing into a Granule List column.
#
# The alternative is to override print with a local copy in GranList, but
# which is the GREATER of 2 evils? JSA

sub cloud_cover { return undef;}
sub local_granule_ID { return undef;}
sub automatic_quality_flag { return undef;}
sub operational_quality_flag { return undef;}
sub science_quality_flag { return undef;}
sub size_mbecs_granule { return undef;}

#----------
# Multi-valued, string sets...
#----------

# PackageIds   -- return a reference to the array of PackageIds, or []
sub package_ids    { shift->_string_set('PackageIds',@_) }

# results -- return a reference to the array of specialized results, or []
sub results { $_[0]->{Results} || []; }

# campaigns   -- return a reference to the array of campaigns, or []
sub campaigns    { shift->_string_set('Campaigns',@_) }

# parameters  -- return a reference to the array of parameters, or []
sub parameters   { shift->_string_set('Params',@_) }

# proc_levels -- return a reference to the array of processing levels, or []
sub proc_levels  { shift->_string_set('ProcLevels',@_) }

# sensors    -- return a reference to the array of sensor names, or []
sub sensors      { shift->_string_set('Sensors',@_) }

# sources    -- return a reference to the array of source names, or []
sub sources      { shift->_string_set('Sources',@_) }

#----------
# Multi-valued, URLs...
#----------

# data_URLs    -- return ref to array of data URLs (as IU::FancyURLs), or []
sub data_URLs    { $_[0]->{DATA_URL} }

# home_URLs    -- return ref to array of data URLs (as IU::FancyURLs), or []
sub home_URLs    { $_[0]->{DATASET_HOME_PAGE} }

# misc_URLs    -- return ref to array of misc URLs (as IU::FancyURLs), or []
sub misc_URLs    { $_[0]->{MISC_URL} }

# search_URLs  -- return ref to array of search URLs (as IU::FancyURLs), or []
sub search_URLs  { $_[0]->{SPECIALIZED_SEARCH_URL} }

# browse_URLs  -- return ref to array of browse URLs (as IU::FancyURLs), or []
sub browse_URLs  { $_[0]->{BROWSE_URL} }

# guide_URLs  -- If GUIDE_URL is undefined,  [] is returned.
sub guide_URLs  { $_[0]->{GUIDE_URL} || [] }

# has_links     -- return whether or not there are external links
sub has_links {
    my $self = shift;
    @{$self->data_URLs} ||
	@{$self->home_URLs} ||
	    @{$self->misc_URLs} ||
		@{$self->search_URLs} ||
		    @{$self->browse_URLs} ||
                        @{$self->guide_URLs};
} 

#----------
# Multi-valued, misc...
#----------

# handling_fee     -- return Handling Fee group(s)
sub  handling_fee  { $_[0]->{HandlingFee}}

# REF 
#
# mediahandlingfee -- set/return ref to array of MediaHandlingFee objects
#
sub mediahandlingfee { 
    my ($self, $ref) = @_;
    if (@_ > 1) {$self->{MediaHandlingFee} = $ref;} 
    $self->{MediaHandlingFee} ||= [];
}

# REF
#
# mediamaximumcapacity -- set/return ref to array of MediaMBECSData objects
#
sub mediamaximumcapacity {
    my ($self, $ref) = @_;
    if (@_ > 1) {$self->{MediaMaximumCapacity} = $ref;}
    $self->{MediaMaximumCapacity} ||= [];
}

# accounts     -- return account information, or []
sub accounts     { $_[0]->{Accounts} || [] }

#------------------------------
#
# ORDERABLE INTERFACE...
#
#------------------------------
# These methods are necessary for IU::Interface::Orderable.

#------------------------------
# order_status
#------------------------------
# B<DANGER! HIGHLY FRAGILE!>
# Returns...
#
#    'NEED_PKGS'  if the item just needs packages selected
#    'GO_PKGS'    if the item is ready for a package order
#    'GO_EMAIL'   if the item has no packages, but can be email-ordered
#     0           (false, defined) if the item can't be ordered

sub order_status {
    my $self = shift;

    # It is if it has no packages, and the dataset CAN email order:
    if (@{$self->package_ids}) {         # try for packages...
        return (int(@{$self->package_combos}) ? 'GO_PKGS' : 'NEED_PKGS');
    }
    return 0;                         # Not supporting e-mail ordering yet!
}

#------------------------------
# order_thingy
#------------------------------
sub order_thingy { "data set" }

#------------------------------
# order_type
#------------------------------
sub order_type   { "DATASET" }

#------------------------------
# dataset
#------------------------------
sub dataset { $_[0] } 

#------------------------------------------------------------
# chosen_account [ACCTNAME]
#------------------------------------------------------------
# Choose the given account, or return the chosen one.

sub chosen_account {
    my ($self, $acctname) = @_;
    (@_ > 1) and $self->{ChosenAcct} = $acctname;
    ($self->{ChosenAcct} || '');
}

#------------------------------------------------------------
# contact_addr [ADDRESS]
#------------------------------------------------------------
# Get/set the contact address.

sub contact_addr {
    my ($self, $addr) = @_;
    #(@_ > 1) and $self->{ContactAddr} = $addr;
    if (defined $addr) {
#        debug "Setting contact address to $addr";
        $self->{ContactAddr} = $addr;
    }
    $self->{ContactAddr};
}

#------------------------------------------------------------
# allows_email_order [BOOL]
#------------------------------------------------------------
# 1 = true
# 0 = false
# undef defaults to 1.

sub allows_email_order {
    my ($self, $val) = @_;
    (@_ > 1) and $self->{AllowsEmailOrder} = $val;
    defined($self->{AllowsEmailOrder}) ? $self->{AllowsEmailOrder} : 1
}

#------------------------------------------------------------
# can_email_order [VALUE]
#------------------------------------------------------------
# If a granule in this dataset has no packaging info, are we allowed to
# order the granule by email?  Returns true if yes, false if no.
#
# Always returns false if there is no contact address, or if the 
# contact address is missing the email address!

sub can_email_order {
    my $self = shift;
#    debug "checking if dataset can email order";
    my $contact = $self->contact_addr;
    return $self->allows_email_order && $contact && $contact->email;
}

#------------------------------------------------------------
# some_options [OPTSTYPE, [OPTIONS]]
#------------------------------------------------------------
# Set/get the options named WHICH.  Possibilities for OPTSTYPE
# include "SUBSET", "ORDER", and "DATASET_ORDER".
#
# If no arguments are given, returns a list of OPTSTYPEs which
# have values in this object.
#
sub some_options {
    my ($self, $type, $opts) = @_;
    $type or return grep {$self->{Options}{$_}}(keys %{$self->{Options}});
    $self->{SomeOptions}{$type} = $opts if @_ > 2;
    $self->{SomeOptions}{$type};
}

#------------------------------------------------------------
# orderable [BOOL]
#------------------------------------------------------------
# 1 = true
# 0 = false
# undef defaults to 1.

sub orderable {
    my $self = shift; 
    return (@{$self->package_ids} ? 1 : 0); 
}

# carted      -- is the dataset in the shopping cart?
sub carted    { 
    my ($self, $value) = @_;
    $self->{Carted} = $value if (@_ > 1);
    $self->{Carted};
}

#------------------------------------------------------------
#
# INVENTORY-DEPENDENT ACTIVITIES...
#
#------------------------------------------------------------

#------------------------------
# inventory [INVSTORE]
#------------------------------
# B<DANGER! HIGHLY FRAGILE!>
#
# With argument, associates this dataset with an inventory storage object, 
# returning the previous association.  With no arg, returns the current 
# association (or undef if none).
#
# WARNING: ALL ACCESSES TO AN INVENTORY SHOULD BE EVAL-PROTECTED!!!!

sub inventory {
    my ($self, $inv) = @_;
    my $oldinv = $self->{INV};
    $self->{INV} = $inv if @_ > 1;
    $oldinv;
}

#------------------------------------------------------------
#
# DEPRECATED COVERAGE METHODS: use the coverage object directly!
#
#------------------------------------------------------------

# area             -- return the area
sub area { $_[0]->{Coverage}->area; }

# area_center      -- return the area center
sub area_center { $_[0]->{Coverage}->center; }

# area_type        -- return the area type
sub area_type { $_[0]->{Coverage}->type; }

#------------------------------------------------------------
1;

