package IU::Diag;


=head1 NAME

IU::Diag -- diagnosis of errors


=head1 DESCRIPTION
    
Loaded only when errors occur.  Used in outputting the dying gasp.
Give it an error number, and you get back a suggestion in text format.
The diagnoses come from a file, probably

    $TopDir/lib/conf/diag.dat

The file consists of separate paragraphs (blocks of text separated
by two or more newlines), where each paragraph should begin with either
a # (a comment) or an C<=nnn>, or an C<=nnn-mmm>.  Like this:

    # A comment.
    
    =452
    IMS/WWW wanted to either open a file to read it (in which
    case, the file is probably not present) or to write it
    (in which case, IMS/WWW probably does not have permission
    to write in that particular directory).  Examine the 
    name of the file, and try and determine what happened.
    
    =454-459
    A helper application launched by IMS/WWW has failed.  IMS/WWW can't 
    get you the exact details, but maybe the error message made it to 
    the IMS/WWW log file.  Contact your IMS/WWW administrator.
    
    =501
    This next line has a single space in it...
      
    ...so this is all REALLY one "paragraph", though it looks like two.


=head1 SYNOPSIS

    $text = IW::Diag::help($errno)


=head1 AUTHOR
    
Eryq (Erik Dorfman), Hughes STX Corporation

=cut


use IU;
use IU::Log;
use FileHandle ();

use strict;

#------------------------------------------------------------
# help ERRNO
#------------------------------------------------------------
# Given an IMS/WWW ERRNO, return the diagnosis, or undef on error.

sub help {
    my $code = shift;      # error code
    my $msg;

    # Open the file:
    my $diagfile = "$IU::TopDir/lib/conf/diag.dat";
    open DIAG, $diagfile or return error "couldn't open $diagfile: $!";
    DIAG->input_record_separator('');     # read paragraphs

    # Scan:
    while (<DIAG>) {
	next if /^#/;                         # skip if comment
	next if !/^=(\d+)(-(\d+))?/;          # skip if not a =ddd line
       	
	# Got one!  Are we in its range?
	my ($lo, $hi) = ($1, (defined($3) ? $3 : $1));
	next unless ($code >= $lo) and ($code <= $hi);
	
	# In range!  Chop off first line, and return the rest:
	s/\A.*$//m;
	s/^=?[ \t]*$//mg;        # replace = or blank lines with nothing
	$msg = $_; 
	last;
    }
    close DIAG;

    # Return whatever we got:
    return $msg;
}

#------------------------------------------------------------
1;
