package IU::ECSProfileRequest;

=head1 NAME

IU::ECSProfileRequest - user request for ECS Profile retrieval, update,
submit or password change.


=head1 SYNOPSIS

Search options/properties:
    
    # What kind of search is this?  Can return any of...
    #     PROFILE_RETRIEVAL_REQUEST
    #     PROFILE_UPDATE_REQUEST
    #     PROFILE_SUBMIT_REQUEST
    #     USER_PASSWORD_CHANGE_REQUEST
    $q->search_type;
    
    # What kind of ODL search message is this *really*?  Can return any of...
    #     PROFILE_RETRIEVAL_REQUEST
    #     PROFILE_UPDATE_REQUEST
    #     PROFILE_SUBMIT_REQUEST
    #     USER_PASSWORD_CHANGE_REQUEST
    $q->ODL_search_type;

ODL interface:

    # ODL interface:
    $q = IU::ECSProfileRequest->from_ODL($odl);
    $odl = $q->to_ODL;
     

=head1 DESCRIPTION

This object encapsulates ECS profile requests.

=cut

use IU;
use IU::Log;
use IU::IW;

use IW::App;
use IW::User;

use strict;
use vars qw(@ISA @EXPORT
	    $MAX_TIME_DEFAULT
            $ECS_PROFILE_REQUEST);
@ISA = qw(Exporter);
@EXPORT = qw($ECS_PROFILE_REQUEST);

#------------------------------
#
# Globals...
#
#------------------------------

# The current query (EXPORTED for the convenience of other modules):
$ECS_PROFILE_REQUEST = undef;

# Default time limit for searches
$MAX_TIME_DEFAULT = $IU::SETUP::MaxSearchTime;

#------------------------------
#
# Private utilities...
#
#------------------------------

#------------------------------------------------------------
# _one VARNAME,SELF,[VALUE]
#------------------------------------------------------------
# I<Private utility.>  Get/set a single-valued instance variable.
# Return undef if it's undefined.
sub _one {
    my ($varname, $self, $value) = @_;
    $self->{$varname} = $value if (@_ > 2);
    $self->{$varname};   
}

#------------------------------------------------------------
# _many VARNAME,SELF,[ARRAYREF]
#------------------------------------------------------------
# I<Private utility.>  Get/set a multi-valued instance variable.
# Set it by providing an arrayref: the array will be duplicated.
#
# If invoked with no args, just returns info.
# Returns the array of all values in an array context, and just 
# the array size in a scalar context.

sub _many {
    my ($varname, $self, $values) = @_;
    if (@_ > 2) {
	my @array = @$values;
	$self->{$varname} = \@array;
    }
    (wantarray ? @{$self->{$varname}} : int(@{$self->{$varname}}));
}

#------------------------------
#
# Construction...
#
#------------------------------

#------------------------------------------------------------
# new 
#------------------------------------------------------------
# Class method.  Constructor.

sub new {    
    my $class = shift;
    my $searchType = shift;
    my $reset = shift;  # Optional flag to indicate password has been reset
    my $self = {};
    $self->{SearchType} = $searchType;
    $self->{Reset} = $reset;
    $self->{DaacsToSearch} = [];
    $self->{DaacsNeedingECSAuth} = [];
    bless $self, $class;
}

#------------------------------------------------------------
# daacs_to_search
#------------------------------------------------------------
# I<Instance method.>
# Return the list of data centers to search.
# This will be empty until set up by determine_daacs_to_search().

sub daacs_to_search { _many('DaacsToSearch', @_) }

#------------------------------------------------------------
# daacs_needing_ecs_auth
#------------------------------------------------------------
# I<Instance method.>
# Return the list of data centers which according the wwwvalids 
# need the ECS authenticator.
# This will be empty until set up by determine_daacs_to_search().

sub daacs_needing_ecs_auth { _many('DaacsNeedingECSAuth', @_) }

#------------------------------
#
# Search options/properties...
#
#------------------------------

#------------------------------------------------------------
# max_time [VALUE]
#------------------------------------------------------------
# Set/get maximum number of minutes that a search can run.
#

sub max_time {
    my ($self, $value) = @_;
    $self->{MaxTime} = $value if @_ > 1;
    $self->{MaxTime} || $MAX_TIME_DEFAULT;
}

#------------------------------------------------------------
# name [NAME]
#------------------------------------------------------------
# Set/get the name of the query.
#
sub name  { _one('Name', @_) }

#------------------------------------------------------------
# daac [Daac]
#------------------------------------------------------------
# Set/get the daac of the query.
#
sub daac  { _one('Daac', @_) }

#------------------------------------------------------------
# homeDaac [Daac]
#------------------------------------------------------------
# Set/get the homeDaac of the query.
#
sub homeDaac  { _one('HomeDaac', @_) }

#------------------------------------------------------------
# filename
#------------------------------------------------------------
# Get a good filename for the query.
#
sub filename {
    my $self = shift;
    my $name = $self->name || 'untitled';
    $name =~ s/^\s+|\s+$//g;        # trim
    $name =~ s/[^a-z0-9_+-]/_/gi;   # turn illegals into underscores
    $name =~ s/_+/_/g;              # collapse underscores
    "$name.query";
}

#------------------------------------------------------------
# ODL_search_type
#------------------------------------------------------------
# Return the search type, as it would appear in the ODL.
#
sub ODL_search_type { 
    my $self = shift;
    my $type = $self->search_type;
    return $type;
}

#------------------------------------------------------------
# search_type [VALUE]
#------------------------------------------------------------
# Set/get the search type, in the most-detailed sense.
# Compare with ODL_search_type.
#
sub search_type { _one('SearchType', @_) }

#------------------------------------------------------------
# reset [VALUE]
#------------------------------------------------------------
# Set/get the reset flag.  This flag indicates that the password
# has been reset without using the original password.
#
sub reset { _one('Reset', @_) }

#------------------------------------------------------------
# origin [VALUE]
#------------------------------------------------------------
# Set/get the origin flag.  
# This indicates the mode that this request came from.
#
sub origin { _one('Origin', @_) }

#------------------------------------------------------------
# which_daacs_need_authenticator  
#------------------------------------------------------------
# returns array of daacs

sub which_daacs_need_authenticator {

    my $self = shift;
    $self->determine_daacs_to_search;
    my @initialDaacs = $self->daacs_to_search;

    # Open static valids database:
    my %WWWVALIDS;
    if (-e "${IU::LibDir}/conf/wwwvalids.gdb") {
        eval {tie %WWWVALIDS, 'IU::Store', "${IU::LibDir}/conf/wwwvalids.gdb",
              'READER'};
        $@ and return error "tie failed: $@";
    }

    # For each data center we're considering...
    my %needs;        # Daacs needing authenticator. 
    my $archid;
    foreach $archid (@initialDaacs) {
        # Is there an entry for it in the static valids?
        my $valids;
        if ($valids = $WWWVALIDS{"daac:$archid"}) {
            $needs{$archid} = 'ECS' if ($valids->{REQUIRES_ECS_AUTH}); 
        }
    }                                    # end: foreach daac 

    # Close static valids database:
    untie %WWWVALIDS;

    return %needs;
}

#------------------------------------------------------------
# filter_daacs_to_search
#------------------------------------------------------------
# I<Instance method, private.>
# Take the list of data centers to actually query, and pare them down.
# based on the static valids file.  See determine_daacs_to_search().
#
# Raises a fatal exception if no DAACs remain.

sub filter_daacs_to_search {
    my $self = shift;
    my @initialDaacs = $self->daacs_to_search;
    my @filteredDaacs;

    # Open static valids database:
    my %WWWVALIDS;
    if (-e "${IU::LibDir}/conf/wwwvalids.gdb") {
        eval {tie %WWWVALIDS, 'IU::Store', "${IU::LibDir}/conf/wwwvalids.gdb",
              'READER'};
        $@ and return error "tie failed: $@";
    } 

    #--------------------
    # create list of data centers needing ECS 
    #--------------------

    # For each data center we're considering...
    my %needsECS;       # Daacs needing ecs authenticator.
    my $daac;

    foreach $daac (@initialDaacs) {
        # Is there an entry for it in the static valids?
        my $valids;
        if ($valids = $WWWVALIDS{"daac:$daac"}) {
            $needsECS{$daac} = $valids->{REQUIRES_ECS_AUTH};
        }                       # end if
    }                                    # end: foreach daac

    my @daacsECS = grep { $needsECS{$_}} (keys %needsECS);

    $self->daacs_needing_ecs_auth(\@daacsECS);

    # Close static valids database:
    untie %WWWVALIDS;

    # Set new list of daacs to search:
    @filteredDaacs = @initialDaacs;
    $self->daacs_to_search(\@filteredDaacs);
    1;
}

#------------------------------------------------------------
# determine_daacs_to_search
#------------------------------------------------------------
# I<Instance method.>
# Determine list of data centers to actually query, using dependent valids.
# Fatal exception if no DAACs can be set.

sub determine_daacs_to_search {
    my $self = shift;
    my @finalDaacs = ();
    $finalDaacs[0] = $self->daac;
    $self->daacs_to_search(\@finalDaacs);
    $self->filter_daacs_to_search;
    # Done!
    1;
}

#------------------------------------------------------------    
# install
#------------------------------------------------------------    
# I<Instance method.>  WG-specific.
# Re-install this as the current query in the session state.

sub install {
    my $self = shift;
    $SESSION->state->{ECSProfileRequest} = $self;
}

#------------------------------------------------------------    
# lookup
#------------------------------------------------------------ 
# I<Class method.  WG-specific.>
# Lookup current query in the session state.  Returns undef if none so far.

sub lookup {
    my $self = shift;
    my $query = $SESSION->state->{ECSProfileRequest};
    return $query;
}

#------------------------------------------------------------
# uses CATEGORY, [SEARCHTYPE]
#------------------------------------------------------------
# I<Class/instance method.>
# Returns true if the given CATEGORY would be a functional part of the 
# query under the given SEARCHTYPE, and false otherwise.   With
# no SEARCHTYPE, will use the query's search type.

sub uses {
    my ($self, $cat, $type) = @_;
    $type ||= $self->search_type;  
    if ($type eq 'PROFILE_RETRIEVAL_REQUEST') { 
        return 1 if ($cat eq 'DATA_CENTER_ID');
	return 0;
    }
    elsif ($type eq 'PROFILE_UPDATE_REQUEST') { 
        return 1 if ($cat eq 'DATA_CENTER_ID');
	return 0;
    }
    elsif ($type eq 'PROFILE_SUBMIT_REQUEST') { 
        return 1 if ($cat eq 'DATA_CENTER_ID');
	return 0;
    }
    elsif ($type eq 'USER_PASSWORD_CHANGE_REQUEST') { 
        return 1 if ($cat eq 'DATA_CENTER_ID');
	return 0;
    }
    else {         # any other kind of search
	return 1;        
    }
}

#------------------------------------------------------------
# uses_opt OPTION, [SEARCHTYPE]
#------------------------------------------------------------
# I<Class/instance method.>
# Much like uses(), but for miscellaneous query options that aren't
# really categories.  Valid OPTIONs are:
#
#     QUERY_NAME       Query name
#
# Returns true if the given OPTION would be a functional part of the 
# query under the given SEARCHTYPE, and false otherwise.   With
# no SEARCHTYPE, will use the query's search type.
# Always show guide limit since it will take effect if the user selects
# Detailed Info while doing any kind of search.

sub uses_opt {
    my ($self, $opt, $type) = @_;

    $type ||= $self->search_type;
    if (($type eq 'PROFILE_RETRIEVAL_REQUEST') ||
        ($type eq 'PROFILE_UPDATE_REQUEST') ||
        ($type eq 'USER_PASSWORD_CHANGE_REQUEST') ||
        ($type eq 'PROFILE_SUBMIT_REQUEST')) {
	return 1 if ($opt eq 'QUERY_NAME');
	return 0;
    }
    else {         # any other kind of search
	return 1;        
    }
}

#------------------------------
#
# ODL interface
#
#------------------------------

#------------------------------------------------------------
# from_ODL ODLOBJECT, PARAMHASH
#------------------------------------------------------------
# I<Class or instance method.>
# Initiallize the address from a standard IMS ODL object.
#
# The PARAMHASH can contain the following:
#
#     KeepType => If true, the search type will be discarded in 
#                 favor of the type of any CURRENTLY-INSTALLED query.
#                 (See lookup() and install() for details).
#                 If false (default), the ODL alone is used in determining
#                 the search type. 

sub from_ODL {
    my ($self, $odl, %params) = @_;
    $odl || return undef;
    my ($param, $value);
    my $search;
    my $keep_type = $params{KeepType};

    # Be a class method, too!
    ref($self) or $self = $self->new;
    #
    # DETERMINE THE SEARCH TYPE...
    #
    #    If "keep type" is true, determine the search type as given by 
    #    any currently-installed query, and prefer that value to the 
    #    actual one given by the ODL.
    #
    #    This is done so that when the user is UPLOADING an ODL search file,
    #    we automatically coerce it into the type of search form 
    #    (INVENTORY, DIRECTORY, GRANULE, etc.) that they are uploading into.

    # Determine the search type as given by the currently-installed query:
    my $prev_search_type;
    if ($keep_type) {
	my $prev_query = ref($self)->lookup;        # don't put it in $QUERY!
	$prev_search_type = $prev_query && $prev_query->search_type;
    }

    # Determine the search type as given by the ODL itself.  
    #    NOTE: this is not necessarily the actual search type of the
    #    top-level message... rather, it is the search type as derived
    #    from the ODL itself.  Notice the case for GRANULE_SEARCH.
    my $new_search_type;
    if ($search = $odl->param('PROFILE_RETRIEVAL_REQUEST')) {
	$new_search_type = 'PROFILE_RETRIEVAL_REQUEST';
    }
    elsif ($search = $odl->param('PROFILE_UPDATE_REQUEST')) {
	$new_search_type = 'PROFILE_UPDATE_REQUEST';
    }
    elsif ($search = $odl->param('PROFILE_SUBMIT_REQUEST')) {
	$new_search_type = 'PROFILE_SUBMIT_REQUEST';
    }
    elsif ($search = $odl->param('USER_PASSWORD_CHANGE_REQUEST')) {
	$new_search_type = 'USER_PASSWORD_CHANGE_REQUEST';
    }
    else {
        puke "ODL doesn't seem to hold  PROFILE_RETRIEVAL_REQUEST ",
             "PROFILE_SUBMIT_REQUEST, USER_PASSWORD_CHANGE_REQUEST ",
             "or PROFILE_UPDATE_REQUEST.";
    }

    # Now, resolve the three search types, preferring the CGI input:
    $self->search_type($prev_search_type || $new_search_type);

    # 
    # EXTRACT THE SEARCH CONSTRAINTS FROM THE ODL...
    #
    $search->rewind;
    my %queue;

    while (($param, $value) = $search->next) {

	if ($param eq 'TIME_LIMIT') {
	    $self->max_time($value->text_value);
	}
	elsif ($param eq 'SEARCH_NAME') {
	    $self->name($value->text_value);
	}
	else {
	    warn "skipping unrecognized search parameter: '$param'";
	}
    }    
    $self;
}

#------------------------------------------------------------
# to_ODL PARAMHASH
#------------------------------------------------------------
# Convert query to an ODL object, and return it.
# The ODL object is a group *containing* an INVENTORY_SEARCH group
# (or DIRECTORY_SEARCH, etc.).
#
# No authentication information is placed in the ODL!!!!
#
# PARAMHASH may contain the following:
#
#    NoDaacs   => If true, then the DATA_CENTER_ID group (user-specified
#                 data centers) will *not* be output in the ODL.
#                 Generally, we handle these separately.
#
#    Everything=> By default, ODL is only output for the current search type,
#                 meaning that if an attribute is not used in the current
#                 search type, then it will not appear in the ODL.
#                 If you set this option to true, then *all* criteria will 
#                 be included in the ODL.
#
#    ForSearch => If true, ODL is tweaked to default/modify certain fields
#                 (e.g., stop date) to make DAACs happy.  Normally,
#                 we don't want the user's query criteria mucked with,
#                 so the default is false.
#
# NOTE: for downloading, use the Everything option to output ALL the 
# categories to the ODL, even ones which are unused for this type of search.  
# This is so an inventory search may be uploaded, turned into a granule-id 
# search, and downloaded, but still retain all the inventory search 
# characteristics.
#
# NOTE: the ForSearch option produces *slightly* different ODL.  Normally,
# we return an accurate, uploadable representation of the query.  With
# ForSearch, we ensure that certain ODL fields are defaulted, to deal 
# with the less-tolerant DAACs.

sub to_ODL {
    my ($self, %params) = @_;
    my $no_daacs         = $params{NoDaacs};
    my $write_everything = $params{Everything};
    my $for_search       = $params{ForSearch};

    require IU::ODL;

    # Begin:
    my $root_odl = IU::ODL->new_group;    # outermost group
    my $srch_odl = IU::ODL->new_group;
    my $extsrch_odl;                      # will be created only if needed

    # Start off the file with the search type:
    $root_odl->add_param($self->ODL_search_type => $srch_odl);

    # The message id (do we REALLY need this?)
    $srch_odl->add_param(MESSAGE_ID    => '0'); 

    # When a user has a password change either by admin mode or by forgot
    # your password link, the session is not the same as the session of
    # the account being changed.  That is why we use a second user object
    # for this purpose, when the reset flag is set.
    my $reset = $self->reset;
    my $user;
    my $origin = $self->origin;
    if($reset || $origin eq 'ADMINECSACCOUNT') {
       $user = $SESSION->state->{User2};
    } else {
       $user = $SESSION->state->{User};
    }
    # The options...
    if ($self->ODL_search_type eq 'PROFILE_SUBMIT_REQUEST') {
       my $home_daac = $self->homeDaac;
       $srch_odl->add_param(HOME_DAAC => $home_daac) if($home_daac);
    }
    if (($self->ODL_search_type eq 'PROFILE_UPDATE_REQUEST') ||
        ($self->ODL_search_type eq 'PROFILE_SUBMIT_REQUEST')) {
        if ($user) {
            $srch_odl->add_param(CONTACT_ADDRESS  => $user->req_addr->to_ODL)
                if $user->req_addr;
        }
        if ($user) {
            $srch_odl->add_param(SHIPPING_ADDRESS  => $user->ship_addr->to_ODL)
                if $user->ship_addr;
        }
        if ($user) {
            $srch_odl->add_param(BILLING_ADDRESS  => $user->bill_addr->to_ODL)
                if $user->bill_addr;
        }
        if ($user) {
            $srch_odl->add_param(USER_AFFILIATION => $user->affil->to_ODL)
                if $user->affil;
        }
    }
    if ($user) {
       # Add the USER_ID for these search types:
       if (($self->ODL_search_type eq 'PROFILE_SUBMIT_REQUEST') ||
           ($self->ODL_search_type eq 'USER_PASSWORD_CHANGE_REQUEST')) {
          $srch_odl->add_param(USER_ID => $user->ecs_username)
           if $user->ecs_username;
       }
       # Add the PASSWORD_RESET flag if the $reset flag is set for
       # the 'USER_PASSWORD_CHANGE_REQUEST'. 
       if ($self->ODL_search_type eq 'USER_PASSWORD_CHANGE_REQUEST') {
          $srch_odl->add_param (PASSWORD_RESET => IU::ODL->symbol('Y')) 
           if $reset;
       }
    }

    # Completely non-standard stuff:
    if ($self->name) {
	$srch_odl->add_param(SEARCH_NAME => IU::ODL->quoted($self->name));
    }

    # Done!
    $root_odl;
}

#------------------------------------------------------------
# to_search_ODL
#------------------------------------------------------------
# Convert query to an ODL object suitable for input to the client, 
# and return it. The ODL object is a group containing an INVENTORY_SEARCH 
# parameter.
#
# Unlike a simple to_ODL, this method uses the dependent valids mechanism
# to determine which DAACs to search

sub to_search_ODL {
    my $self = shift;

    # Start by figuring out which DAACS to search, if needed:
    $self->daacs_to_search
	or $self->determine_daacs_to_search
	    or puke "no data centers to search!";  

    # Start with the basics, *without* the DAACs:
    my $root_odl = $self->to_ODL(NoDaacs=>1, ForSearch=>1) 
	|| return undef;

    # Get the actual search group, checking all standard types:
    my $srch_odl = $root_odl->param('PROFILE_RETRIEVAL_REQUEST') ||
		   $root_odl->param('PROFILE_UPDATE_REQUEST') ||
		   $root_odl->param('PROFILE_SUBMIT_REQUEST') ||
		   $root_odl->param('USER_PASSWORD_CHANGE_REQUEST') ||
		   puke "No recognized search group in ODL!"; 

    # Add the data centers (notice that we use the QUERY daacs!):
    if ($self->daacs_to_search) {
	$srch_odl->add_param(DATA_CENTER_ID => 
  			     IU::ODL->conv($self->{DaacsToSearch}));
    }

    my $reset = $self->reset;
    my $user;
    my $origin = $self->origin;
    if($reset || $origin eq 'ADMINECSACCOUNT') {
       $user = $SESSION->state->{User2};
    } else {
       $user = $SESSION->state->{User};
    }
    # Add the ecs authenticator, if any:
    my $ecs_authenticator = $user->ecs_authenticator;
    my $old_ecs_authenticator = $user->old_ecs_authenticator;
    $old_ecs_authenticator = $ecs_authenticator if (!$old_ecs_authenticator);
    if ($ecs_authenticator) {
       # For 'USER_PASSWORD_CHANGE_REQUEST' NEW_ECS_AUTHENTICATOR is
       # the current ecs authenticator. ECS_AUTHENTICATOR is the previous
       # ecs authenticator.  This will change in the future.
       if ($self->ODL_search_type eq 'USER_PASSWORD_CHANGE_REQUEST') {
          $srch_odl->add_param(NEW_ECS_AUTHENTICATOR => 
                               IU::ODL->conv($ecs_authenticator));
          $srch_odl->add_param(ECS_AUTHENTICATOR => 
                               IU::ODL->conv($old_ecs_authenticator));
       } else {
	  # All other search types just add the current ecs authenticator.
          $srch_odl->add_param(ECS_AUTHENTICATOR => 
                               IU::ODL->conv($ecs_authenticator));
       }
    } else {
       puke 513, "No ECS authenticator";
    }

    # Done!
    return $root_odl;
}

#------------------------------------------------------------    
1;
