package IU::Errors;

=head1 NAME

IU::Errors - a simple object holding some error strings.  Used in
User and Address objects.

=head1 DESCRIPTION

Any object that reads data and finds errors in it, used to puke.
Instead, the object can contain an IU::Errors object and fill
it with error messages.  Then, any calling Modes can check for
errors and print out messages and redraw screens without
necessarily puking.

Just add a new IU::Errors to your $self.
Then when an error happens, add an error message to it
using the set() method.
The calling mode can check if there were errors by asking
the errors object with the any_errors() method.
The mode can get the array of error strings with the get() method.

=head2 DATA

=begin list

=item

$self

This is the usual hash reference.  It contains one key: STRINGS
whos value is a reference to an array of strings.  This object could
be simpler, just using an array reference instead of the hash
reference, but in the future someone may want to add more keys.

=end list

= head2 METHODS

=begin list

=item

new ( ) . . . Constructor.

Create an empty hash reference $self.  Add an empty array reference
with key = 'STRINGS'.
Usage:  my $Errors = new IU::Errors;

=item

get ( )

Returns the array of error strings (if there are any, else undef).
Usage:  my @errors = $Errors->get;

=item

set ($STRING, [,$STRING...])

Add the given string or strings to the 'STRINGS' array in the
IU::Errors object.
Usage:  $Errors->set('City name too long...');
Usage:  $Errors->set('City name too long...',
                     'State name has illegal character @');

=item

clear ( )

Delete all strings in the Errors object.
Usage:  $Errors->clear;

=item

any_errors ( )

Test if any errors set.
Usage:  if ($Errors->any_errors) { ... }

=end list

=head1 AUTHOR

Larry P. Rosen, EOS Data Gateway Team, Raytheon Systems Corp.

=head1 RCS INFO

$Id: Errors.pm,v 1.1 1999/03/31 23:35:00 lrosen Exp $

=cut

#=====================================================================

use strict;

#---------------------------------------------------------------------
# new 
#---------------------------------------------------------------------
# Constructor class method.  Creates a new, empty Errors object.
# Usage:  my $Errors = new IU::Errors;
# Input: none.   Output: new IU::Errors object.

sub new {
   my $class = shift;                   # Get class.
   my $self = {};                       # Start with empty hash.
   $self->{STRINGS} = [];               # Add empty array ref.
   bless $self, $class;                 # Create the object.
   return ($self);                      # Return object.
}

#---------------------------------------------------------------------
# get
#---------------------------------------------------------------------
# Returns the array of error strings (if there are any, else undef).
# Usage:  my @errors = $Errors->get;
# Input: none.   Output: array of error strings or undef if none.

sub get {
   my $self = shift;
   return (@{$self->{STRINGS}});
}

#---------------------------------------------------------------------
# set  $STRING [,$STRING...]
#---------------------------------------------------------------------
# Add the given string or strings to the 'STRINGS' array in the
# IU::Errors object.
# Usage:  $Errors->set('City name too long...');
# Usage:  $Errors->set('City name too long...',
#                      'State name has illegal character @');
# Input: A string or a list of strings.
# Output: Total number of error strings in object after call.

sub set {
   my $self = shift;
   push (@{$self->{STRINGS}}, @_);
}

#---------------------------------------------------------------------
# clear
#---------------------------------------------------------------------
# Delete all strings in the Errors object.
# Usage:  $Errors->clear;
# Input: None.  Output. Empty array reference.

sub clear {
   my $self = shift;
   $self->{STRINGS} = [];
}

#---------------------------------------------------------------------
# any_errors
#---------------------------------------------------------------------
# Test if any errors set.
# Usage:  if ($Errors->any_errors) { ... }
# Input: None.  Output: Number or error strings.

sub any_errors {
   my $self = shift;
   return (scalar (@{$self->{STRINGS}}));
}

#---------------------------------------------------------------------
# equals IU::Errors
#---------------------------------------------------------------------
# Test if two Errors objects are equal (have the same strings).
# Usage:  if ($err1->equals($err2)) { ...
# Input: An IU::Errors object.  Output 1 or 0.  1 if equal.

sub equals {
   my ($self, $err2) = @_;
   # Get arrays.
   my @arr1 = @{$self->{STRINGS}};
   my @arr2 = @{$err2->{STRINGS}};
   # Get number of error strings in each.
   my $num_err1 = scalar (@arr1);
   my $num_err2 = scalar (@arr2);
   # Return false unless they have the same number of error strings.
   return (0) unless ($num_err1 == $num_err2);
   # Return true if they have no errors.
   return (1) if ($num_err1 == 0);
   # Must check each string.  Orders may be scrambled, so sort.
   @arr1 = sort (@arr1);
   @arr2 = sort (@arr2);
   my $i;
   for ($i=0; $i<$num_err1; $i++) {
      # If any string doesn't match return false.
      return (0) unless ($arr1[$i] eq $arr2[$i]);
   }
   # All matched, return true.
   return (1);
}

#=====================================================================
1;
