package IU::Event;


=head1 NAME

IU::Event - log an event using the IMS stats format


=head1 DESCRIPTION

Log a special event to this day's event log file, for statistics
gathering.  Events are logged in standard IMS/WWW stats format,
which is basically param=value.  Log file names look like this:

       ${IU::LogDir}/1996-12-09.statlog

It is a fatal error if the message cannot be logged.


=head1 SYNOPSIS

=head2 Normal usage

Normally, you'd invoke the logging function like this:

       ims_event 'LOGIN', 
                  PARAM1 => value1, 
                  PARAM2 => value2, 
                  ...;

Each value may be either a simple scalar, or a ref to an array;
in the latter case, one entry is made under that parameter for
each element of the array.  So this:

       ims_event 'INVENTORY', 
                  MAXGRANS=>10, 
                  ARCHID=>['GSFC','MSFC','EDC'];

Results in a log entry like this:

       daac=v0ieww00
       event=inventory
       sessid=841943137-2498
       sid=v0ieww00
       event_date=19960905
       event_time=121105
       addr=fake.name@user.host.name
       maxgrans=10   
       archid=GSFC
       archid=MSFC
       archid=EDC


=head2 The "external" values

Normally, certain "external" values will be used in generating 
the message; these are:

       TIME      The time of the event, in seconds-since-epoch.
                 Logged under "event_date" and "event_time".
                 NOTE: this still goes to the log file for the REAL time!
       
       SESSID    The gateway session id (if available).
                 Logged as "sessid" (no, that's not a mistake).
               
       SYSID     The system id (defaults to IU::StatsId).
                 Logged as "sid" (no, that's not a mistake).

However, you can put an optional hashref after the event name; 
such a hash can contain overrides for these values that will be used
in generating the message.  These take precedence over any defaults:

       ims_event 'LOGIN' { SESSID=>'foobar' } 
                  PARAM1 => value1, 
                  PARAM2 => value2, 
                  ...;


=head1 NOTES

A costly way of making log entries, but REAL simple.

In the following table, each entry lists the record type and a short
description, followed by the parameters for each different class or mode
it is called from (all-uppercase means it's a mode: (FOO) would mean C<ModeFOO.pm>).

Events we support I<(* means that it's a standard IMS event, and that we
include it in our stats reports)>:

         accountcreated ------------- new persistent account
                                      account => account name (CREATEACCOUNT)

         account_status ------------- check account with daac
                                      archid=> DAAC list (ACCOUNTSTATUS) 
                                      -----------------------------------
                                      ARCHID=> DAAC list (ORDERSTATUS)

         badadmin ------------------- failed login attempt to admin scripts
                                      (ADMINLOGIN)
                                      
         bad_login ------------------ bad login to persistent account
                                      ACCOUNT=> account (CREATEACCOUNT)
                                      -------------------------------------
                                      no extra parameters (ModeLOGIN - bug?)
 
         browsertest ---------------- user wants to test out their browser
                                      (BROWSERTEST)

         CHANGE_ECS_PASSWORD -------- seamless ECS password change
                                      no extra parameters (ECSPROFILE)

         comments ------------------- user comments submitted successfully
                                      no extra parameters (CMNTFORM)

         definition ----------------- definition request (DEFN)
                                      defn_type => type
                                      defn_item => item
                                      defn_okay => true/false

         datagranule ---------------- granule id search
                                      ARCHID => DAAC list (SEARCH)
                                      DATASET => data set list 
                                      GRANULEID => granule ID list

     (*) directory ------------------ directory request
                                      ARCHID=> DAAC list (ACCOUNTSTATUS)
                                      -----------------------------------
                                      ARCHID=> DAAC list (ORDERSTATUS)
                                      -----------------------------------
                                      ARCHID=> DAAC list (SEARCH)

         dyinggasp ------------------ dying gasp occurred (App.pm)
                                      ERRNO => error number
                                      ERRSTR => error string
                                      ERRMSG => error message
 
         geo ------------------------ selection type and value
                                       (not yet implemented)

     (*) guide ---------------------- guide request
                                      ARCHID => DAAC list (ACCOUNTSTATUS)
                                      ------------------------------------
                                      guideitem => doc title (GUIDE)
                                      -----------------------------------
                                      ARCHID=> DAAC list (ORDERSTATUS)
                                      -----------------------------------
                                      ARCHID=> DAAC list (SEARCH)

         help ----------------------- help request
                                      HELPFILE => page (HELP)

         intbrowse ------------------ integrated browse
                                      ARCHID => DAAC (VIEWBROWSE)
                                      DSETID => data set
                                      GRANID => granule

     (*) inventory ------------------ inventory search
                                      ARCHID => DAAC list (ACCOUNTSTATUS)
                                      -----------------------------------
                                      ARCHID => DAAC list (ORDERSTATUS)
                                      -----------------------------------
                                      ARCHID => DAAC list (SEARCH)
                                      GEO => geo region
                                      DATES_ARE => Julian Date Range or 
                                                   Standard Date Range
                                      TIME => start date, stop date,
                                              start time, stop time

         localgranule --------------- local granule ID search
                                      DATASET => data set list (SEARCH)
                                      LOCALGRANULEID => local granue ID list

     (*) login ---------------------- new session
                                      AGENT=> user agent string

     (*) logout --------------------- session deleted (by cleanup routine)
         lookup --------------------- record of d/s lookup and its results

         matchbyterm ---------------- results for a term being matched 
                                      (SSRCHFORM)
                                      TERM => word used for match
                                      HITS => number of data sets for term

         matchall ------------------- assembled results for a "match all"
                                      (SSRCHFORM)
                                      TERM => list of terms
                                      HITS => number of combined hits
                                      
         matchany ------------------- assembled results for a "match any"
                                      (SSRCHFORM)
                                      TERM => list of terms
                                      HITS => number of combined hits

         matchexact ----------------- assembled results for a "match exactly"
                                      (SSRCHFORM)
                                      TERM => list of terms
                                      HITS => number of combined hits

         mode ----------------------- a new mode is being handled for a client
                                      SESSID => session ID (Mode.pm)

         namechanged ---------------- user changed name of persistent account
                                      OLDACCOUNT => name (CREATEACCOUNT)
                                      NEWACCOUNT => name 

         orderesult ----------------- request completed (ORDER)
                                      daac => DAAC 
                                      requestid => EDG request ID
                                      daacorderid => DAAC order ID

         passwordchanged ------------ user changed password on account
                                      ACCOUNT => session name (CREATEACCOUNT)

         PRICE_ESTIMATE_REQUEST ----- made a price estimate request
                                      none (JPRICEESTIMATE)
                                      -----------------------------------
                                      none (PRICEESTIMATE)

     (*) productreq ----------------- product request 
                                      requestid => request ID (ORDER)
                                      lastname => last name
                                      firstname => first name
                                      street => street addres
                                      city => city
                                      state => if any
                                      country => country
                                      zip => if any
                                      phone => phone #
                                      email => email address
                                      otype => "Normal" or "Email"
                                      -------------------------------------
                                      OTYPE => "FtpBrowse" (FTPBROWSE) 
                                      ARCHID => DAAC
                                      DSETID => Data set
                                      GRANID => Granule ID
                                      
         product_status ------------- product status check with daac
                                      ARCHID => DAAC list (ACCOUNTSTATUS)
                                      -----------------------------------
                                      ARCHID=> DAAC list (ORDERSTATUS)

         RETRIEVE_ECS_USER_PROFILE -- get profile for seamless registration
                                      no extra parameters (ECSPROFILE)

         searchpass ----------------- search passed in (SRCHPASS)
                                      referer => HTTP_REFERER
                                      datasets => data set list
                                      action => requested action 

         searchquery ---------------- record of query just prior to search

         SUBMIT_ECS_USER_PROFILE ---- send profile for seamless registration
                                      no extra parameters (ECSPROFILE)

         time ----------------------- selection type and value 
                                      (not yet implemented)

         UPDATE_ECS_USER_PROFILE ---- update profile for seamless registration
                                      no extra parameters (ECSPROFILE)

         url ------------------------ user was redirected (ENERGIZE)
                                      URL => URL
                                      TYPE => type or "UNKNOWN"

=head1 BUGS

Does not require any IW modules (which is good), B<but>, it I<does> 
peek at $IW::App::SESSION to get the default session id... which is evil.


=head1 AUTHOR

Eryq, Hughes STX Corporation

=head1 VERSION

$Id: Event.pm,v 1.20 2001/04/06 19:07:25 joe Exp $

=cut

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(ims_event);

use IU;
use IU::Log;
use IU::WWWValids;
use strict;
use vars qw($WWWVALIDS);

# Flock stuff:
my %LOCK = (SH=>1, EX=>2, NB=>4, UN=>8);


#------------------------------------------------------------
# ims_event 
#------------------------------------------------------------
# Log an important event.  Fatal error if log cannot be opened.

sub ims_event {
    my $event = shift;

    my $activity_key = $IU::LogLevel ? lc($IU::LogLevel) : "maximal";

    # Grab the list of records that are to be skipped (if any exists).
    $WWWVALIDS = IU::WWWValids->new if (!defined($WWWVALIDS));
    my @record_list = $WWWVALIDS->activeStatRecords(uc($activity_key));

    # If there was a stat record in wwwvalids.conf for the current
    # loglevel, set up the local data structures to *not* record those records.
    if (@record_list) {
       # Maximal always means everything; don't let this be overridden.
       if ($activity_key eq "maximal") {
          debug qq(Attempt to override "maximal" ignored);
          @record_list = ();
       }
       else {
          debug "LogLevel $activity_key; not recording: @record_list";
       }
    }

    # There were no entries for the selected LogLevel. Assume "maximal". 
    else {
       debug 'no record list, recording everything ("maximal")';
       @record_list = ();
    }

    # Skip events that shouldn't be logged.
    local $_;
    if (grep {/^$event$/i} @record_list) {
      debug "Event '$event' not logged - log level is '$activity_key'";
      return;
    }
    else {
      debug "Recording $event at loglevel $activity_key";
    }
    my $env = {};
    $env = shift if (ref($_[0]) && (ref($_[0]) eq 'HASH'));   # opt next arg
    my @pairs = @_;         # for safety


    #
    # KLUDGE to get info specific to Web Gateway!!!!
    # Try and grab session id and user email.
    #
    my ($WG_addr, $WG_sessid);
    if ($IW::App::SESSION) {
	require IW::User;
	$WG_sessid = $IW::App::SESSION->id;
	my $state  = $IW::App::SESSION->state;
	my $user   = $state ? $state->{User} : undef;
	$WG_addr   = $user && $user->req_addr && $user->req_addr->email;
    }
    
    # Setup info:
    my $time   = ($env->{TIME} || 
		  time());
    my $sessid = ($env->{SESSID} || 
		  $WG_sessid ||
		  'unknown');
    my $sysid  = ($env->{SYSID} || 
		  $IU::StatsId || 
		  'unknown');
    my $host   = ($ENV{REMOTE_HOST} || 
		  $ENV{REMOTE_ADDR} || 
		  'fake.host');
    my $addr   = ($env->{ADDR} || 
		  $WG_addr ||
		  "fake.name\@$host");

    # Get the REAL date and time (using in figuring out which log to open):    
    my ($se, $mi, $hr, $dy, $mo, $yr) = localtime(time());
    ++$mo;     # 0-11 => 1-12
    #($yr < 100) && ($yr += (($yr < 50) ? 2000 : 1900));

    $yr += 1900;
    my $datestr = sprintf("%04d%02d%02d", $yr, $mo, $dy);
    my $timestr = sprintf("%02d%02d%02d", $hr, $mi, $se);

    # Open log file, negotiating lock:
    my $fname = sprintf("%04d-%02d-%02d.statlog", $yr, $mo, $dy);
#    debug "logging $event to $fname";
    open(EVENTS, ">>${IU::LogDir}/$fname")
	or puke 452, "couldn't open '${IU::LogDir}/$fname' for append: $!";
    flock(EVENTS, $LOCK{EX}) 
	or puke 490, "couldn't lock log file";
    seek(EVENTS, 0, 2);      # in case someone appended while we were waiting

    # Get the EVENT date and time: 
    ($se, $mi, $hr, $dy, $mo, $yr) = localtime($time);
    ++$mo;     # 0-11 => 1-12
    #($yr < 100) && ($yr += (($yr < 50) ? 2000 : 1900));
    my $edatestr = sprintf("%04d%02d%02d", $yr+1900, $mo, $dy);
    my $etimestr = sprintf("%02d%02d%02d", $hr, $mi, $se);

    # Write standard message header:
    print EVENTS 
	"daac=$sysid\n",              # no, that's not a mistake
	"event=\L$event\E\n",
	"sessid=$sessid\n",           # ...neither is this
	"sid=$sysid\n",               # ...and neither is this
	"event_date=$edatestr\n",
	"event_time=$etimestr\n",
	"pid=$$\n",
	"addr=$addr\n";
    
    # Get other message info:
    my ($v, @vs);
    while (@pairs) {
	my $key = shift @pairs;
	my $value = shift @pairs;

	# Get list of all actual values under this key:
	if (!ref($value))              { @vs = ($value) }
	elsif (ref($value) eq 'ARRAY') { @vs = @$value }
	else                           { @vs = () }
	
	# Output each:
	foreach $v (@vs) {
	    $v =~ s|[,\s]+| |g;       # turn bad chars to single spaces
	    print EVENTS "\L$key\E=$v\n";
	}
    }
    print EVENTS "\n";

    # Close!
    EVENTS->flush();
    flock(EVENTS, $LOCK{UN});
    close(EVENTS);    
    1;
}

#------------------------------------------------------------
1;
