package IU::EventEntry;

=head1 NAME

IU::EventEntry -- read event log entries

=head1 SYNOPSIS

    while ($entry = input IU::EventEntry \*STDIN) {

    }

=head1 DESCRIPTION

For reading/parsing event log entries, as created by IU::Event.
This package assumed that each event is logged in this form:

    daac=v0ieww00
    event=foo
    sessid=847747507-1022
    sid=v0ieww00
    event_date=19961111
    event_time=152531
    addr=fake.name@fake.host
    param=value
    param=value
    otherparam=value
    otherparam=value
    
Multiple entries in the same stream are separated by one or more
empty lines.

There's even an output routine, which outputs in a pretty and compact way:

   FOO         1996/11/11 15:25:31         v0ieww00 
               847747507-1022   fake.name@fake.host
   param:      value
   param:      value
   otherparam: value
   otherparam: value 

=head1 NOTES

Tags are converted to all-lowercase before storing them.

The implementation gets tricky since we need to be able to efficiently
retrieve values in two different ways:

=over

=item *

In the order they were parsed.  This is because some standard
messages depend on the order of the elements to define "group".
Product requests are logged this way.

=item *

By attribute (e.g., 'sessid').

=back





=head1 AUTHOR

Eryq / Hughes STX Corporation

=head1 RCSID

$Id: EventEntry.pm,v 1.3 1997/01/09 20:51:44 eryq Exp $

=head1 PUBLIC INTERFACE

=cut

use IU::Log;

# To make Perl happy...
DESTROY { 1; }

use strict;



#------------------------------------------------------------
# do_files 
#------------------------------------------------------------
=item do_files %params

Class method.  Named params are:

   Files     Files to scan
   Action    Call &$Action($entry) for each entry in Files.
             Return trace char to display, or undef for none.
   Trace     Where to print trace info

Prints nice trace info to STDOUT.

=cut

sub do_files {
    my ($class, %params) = @_;
    my $file;
    my $TRACE = $params{Trace};

    # Scan all log files:
    foreach $file (@{$params{Files}}) {
	$TRACE and print $TRACE "Scanning $file...\n" if $TRACE;

	# Open it:
	if (! open LOG, "<$file") {  
	    warn "Open of $file failed: $!\n";
	    $TRACE and print $TRACE "Open failed: $!\n"; 
	    next;
	}

	# Scan it:
	my $entry;
	my $count = 0;
	while ($entry = input IU::EventEntry \*LOG) {

	    # Tally and mark:
	    ++$count;

	    # Do stuff:
	    my $tracechar = &{$params{Action}}($entry);
	    if ($TRACE && defined($tracechar) && (length($tracechar) == 1)) {
		print $TRACE $tracechar;
		print $TRACE "\n" if ($count % 50 == 0);
		
	    }
	}
	print $TRACE "\n" if $TRACE;
	
	# Done!
	close LOG;
    }
    1;
}



#------------------------------------------------------------
# new
#------------------------------------------------------------
=item new

Constructor; create a new, empty record.

=cut

sub new {
    my $class = shift;
    bless {Pairs=>[], Hash=>{}}, $class;
}

#------------------------------------------------------------
# input
#------------------------------------------------------------
=item input FILEHANDLE

I<Class or instance method>.
Read next record from FILEHANDLE into the internal data structure.

WARNING: Testing this function interactively may yield stange results.
Give it a test files.
 
RETURNS: the object on success, 0 for EOF, undef for failure.

=cut

sub input {
    my ($self, $in) = @_;
    my ($key, $value);
    local($/) = "";		# paragraph-at-a-time
    my $line = '';
    my $values;
    my $entry;
   
    # Construct if necessary:
    ref($self) or $self = $self->new;
    defined($entry = <$in>) or return 0;
    my @pairs = split(/\n/, $entry);
    @pairs or return undef;

    # Assign:
    $self->{Pairs} = [];
    $self->{Hash} = {};
    my $pair;
    foreach $pair (@pairs) {
	my ($key, $val) = split '=', $pair, 2;
	$key = lc($key);
	push @{$self->{Pairs}}, [$key, $val];
	push @{($self->{Hash}{$key} ||= [])}, $val;
    }

    # Done!
    return $self;
}

#------------------------------------------------------------
# keys 
#------------------------------------------------------------
=item keys

Return the keys, in no particular order.

=cut

sub keys {
    my $self = shift;
    return (keys %{$self->{Hash}});
}

#------------------------------------------------------------
# value
#------------------------------------------------------------
=item value KEY

Given a key, return the first value, or undef if not present.

=cut

sub value {
    my ($self, $key) = (shift, lc(shift));
    ($self->{Hash}{$key} ? $self->{Hash}{$key}[0] : undef);
}

#------------------------------------------------------------
# values 
#------------------------------------------------------------
=item values KEY

Given a key, return the array of values, or empty if none.
In a scalar context, returns the number of values.

=cut

sub values {
    my ($self, $key) = (shift, lc(shift));
    return ($self->{Hash}{$key} ? @{$self->{Hash}{$key}} : ());
}

#------------------------------------------------------------
# add
#------------------------------------------------------------
=item add KEY, VALUES...

Add values to a key's list of values.
Appends them to the end of the virtual linear record; e.g.:

     ...existing...
     key=value1
     key=value2
     ...
     key=valuen

=cut

sub add {
    my ($self, $key, @values) = (shift, lc(shift), @_);
    push @{$self->{Pairs}}, map {[$key => $_]} @values;
    push @{$self->{Hash}{$key} ||= []}, @values;
}

#------------------------------------------------------------
# datetime_str
#------------------------------------------------------------
=item datetime_str

Return date and time, as sortable string

=cut

sub datetime_str {
    my $self = shift;
    $self->value('event_date') . $self->value('event_time');
}

#------------------------------------------------------------
# Access methods
#------------------------------------------------------------
=item addr

=item event

=item event_date

=item event_time

=item session_id

=item system_id

Return standard fields.

=cut

sub addr       {  $_[0]->value('addr') }
sub event      {  $_[0]->value('event') }
sub event_date {  $_[0]->value('event_date') }
sub event_time {  $_[0]->value('event_time') }
sub session_id {  $_[0]->value('sessid') }
sub system_id  {  $_[0]->value('sid') }            # :-(


#------------------------------------------------------------
# print
#------------------------------------------------------------
=item print [FILEHANDLE]

Output contents in standard form.
FILEHANDLE defaults to the currently-selected one. 

=cut

sub print {
    my ($self, $fh) = @_;
    $fh or $fh = select;
    no strict 'refs';
    my $pair;

    foreach $pair (@{$self->{Pairs}}) {
	print $fh $pair->[0], '=', $pair->[1], "\n";
    }   
    print $fh "\n";
}

#------------------------------------------------------------
# dump
#------------------------------------------------------------
=item dump [FILEHANDLE]

Output contents in compact form.
FILEHANDLE defaults to the currently-selected one. 

=cut

sub dump {
    my ($self, $fh) = @_;
    $fh or $fh = select(); 
    no strict 'refs';

    my ($yr, $mo, $dy) = ($self->event_date =~ /^(\d{4})(\d{2})(\d{2})$/);
    my ($hr, $mi, $se) = ($self->event_time =~ /^(\d{2})(\d{2})(\d{2})$/);    

    # Line 1: event (modes) date time:
    my $event = uc($self->event);
    if ($event eq 'MODE') {
	my $mode    = $self->value('mode')    || 'none';
	my $submode = $self->value('submode') || 'none';
	my $pages = 
	     "[" . ($self->value('ufrom')||'000')
	    . " -> " . ($self->value('upage')||'000'). "]";
	$event = sprintf("%s: %s/%s", $event, $mode, $submode);
	$event = sprintf("%-30s $pages", $event);
    }
    elsif ($event eq 'DYINGGASP') {
	$event .= ': ' . $self->value('eno') . ' ' . $self->value('estr');
    }
    elsif ($event eq 'HELP') {
	$event .= ': ' . $self->value('helpfile');
    }
    print $fh "$event\n";

    # Line 1a: error message:
    if ($self->value('emsg')) {
	print $fh "\t", substr($self->value('emsg'), 0, 70), "\n";
    }

    # Line 2: date time sessid user:
    my $addr = $self->addr;
    $addr =~ s/^fake\.name\@//;
    print $fh "\t$yr/$mo/$dy $hr:$mi:$se ", $self->session_id, " ",
              $addr, "\n";

    my $pair;
    foreach $pair (@{$self->{Pairs}}) {
	next if ($pair->[0] =~ /^(event|event_date|event_time|sid|sessid)$/);
    }
}

#------------------------------------------------------------
1;


