package IU::FancyURL;

=head1 NAME

IU::FancyURL - a URL that links to on-line data

=head1 DESCRIPTION

This class models a new entity that is being put in the ODL
extensions.  The corresponding ODL looks like any of...

       GROUP = DATA_URL
          URL = 
          URL_COMMENT =
       END_GROUP = DATA_URL

       GROUP = DATASET_HOME_PAGE
          URL = 
          URL_COMMENT =
       END_GROUP = DATASET_HOME_PAGE

In the future, other things (like icon specifications) could be added
to the ODL, and hence to the object's attributes.

=head1 SYNOPSIS

    $furl = new IU::FancyURL URL     =>"http://some.url.com/",
                             Comment =>"A lovely little URL";
    
    # Output it:
    print '<A HREF="', $furl->URL, '">';
    print $furl->comment;      # unsafe! comment may contain bad characters!
    print '</A>';
    
    # The same, using HTML streams (much safer):
    $HTML -> A(HREF=>$furl->URL) -> t($furl->comment) -> _A;

=head1 AUTHOR

Eryq, Hughes STX Corporation

=head1 RCS INFO

$Id: FancyURL.pm,v 1.5 1997/09/16 15:35:16 eryq Exp $

=cut

use IU;
use IU::Chunk;
use IU::ODL ();
use IU::Log;


#------------------------------------------------------------
# new
#------------------------------------------------------------
sub new {
    my $class = shift;
    my %params = @_;
    bless \%params, $class;
}

#------------------------------------------------------------
# URL
#------------------------------------------------------------
# Return the URL, or undef if there is none.

sub URL {
    $_[0]->{'URL'};
}

#------------------------------------------------------------
# type [VALUE]
#------------------------------------------------------------
# Get/set the type.  Returns false if type unknown/
# This is the name of the ODL group; e.g., "MISC_URL".

sub type {
    my ($self, $value) = @_;
    $self->{Type} = $value if @_ > 1;
    $self->{Type};
}

#------------------------------------------------------------
# comment
#------------------------------------------------------------
# Return the comment, or the empty string if there is none.

sub comment {
    defined($_[0]->{Comment}) ? $_[0]->{Comment} : '';
}

#------------------------------------------------------------
# from_chunk CHUNK, [TYPE]
#------------------------------------------------------------
# I<Class or instance method.>
# Create a new instance from a chunk.

sub from_chunk {
    my ($self, $chunk, $type) = @_;
    ref($self) or $self = $self->new;
    $self->{'URL'}   = $chunk->value('URL');
    $self->{Comment} = $chunk->value('URL_COMMENT');
    $self->{Type}    = $type if $type;
    $self;
}

#------------------------------------------------------------
# from_ODL ODL
#------------------------------------------------------------
# I<Class or instance method.>
# Create a new instance from an ODL group.

sub from_ODL {
    my ($self, $odl) = @_;
    my $url_odl;
    my $urlcmnt_odl;

    # Be a class or instance method:
    ref($self) or $self = $self->new;

    # Extract URL:
    if ($url_odl = $odl->param('URL')) {
	my $value = $url_odl->text;
#	$value =~ s/^\\n//;
        $self->{'URL'} =  $value ;
#	$self->{'URL'} = $url_odl->text;
	}

    $self->{'URL'} or puke "missing URL!";
    
    # Extract comment:
    if ($urlcmnt_odl = $odl->param('URL_COMMENT')) {
	# String, or sequence?
	$urlcmnt_odl || puke " no url comment";

	if ($urlcmnt_odl->type eq 'ARRAY') {
	    $self->{Comment} = 
		join('', map { $_->text } $urlcmnt_odl->as_array); 
	} 
	else {
	    $self->{Comment} = $urlcmnt_odl->text;
	}
    }
    
    # Done!
    $self;    
}

#------------------------------------------------------------
1;

