package IU::FileCache;

=head1 NAME

IU::FileCache - manage cache of files


=head1 DESCRIPTION

This module allows you to store/retrieve one or more related files
under a single identifier.

It uses blocking locks, so updates are easy to synchronize.


=head1 SYNOPSIS

Normal usage:

   # Open the cache (creating it if necessary):
   my $cache = open IU::FileCache Path=>"/home/mycache";
   
   # Store some related files to cache (copies files into cache directory):
   $cache->insert_files($key, \@paths);
   
   # Lookup any (active) entry from cache:
   my $entry = $cache->lookup($key);
   
   # Use that entry:
   my $inserted = $entry->insert_time;  # date entered into cache
   my $lookedup = $entry->lookup_time;  # date of last lookup in cache
   my $accessed = $entry->access_time;  # last time any *file* was accessed
   foreach $path ($entry->paths) {
       debug "file: $path"
   }

Cleanup of individual entries:

   # Mark an entry for deletion from cache (removes normal record only):
   $cache->doom($key);
   
   # Lookup an entry from cache, even if "doomed":
   my $entry = $cache->lookup($key, 'ALL');
   if ($entry->doom_time) {    # when was it doomed?
       print "doomed ", (time() - $entry->doom_time), " seconds ago.";
   }
   
   # Remove an entry from cache NOW (physically deletes the files as well!):
   $cache->delete($key);

Cleanup of the cache as a whole (expensive operations!):

   # Reap all doomed cache files (remove physical files if no logical record):
   $cache->reap('2d');               # reap if doomed over 2 days ago
   
   # Purge the cache:
   $cache->purge(MaxSize     => '10mb',
		 ComfortSize => '5mb',
		 ShelfLife   => '30m',
		 LingerTime  => '1m');

Misc:

   # Get all entry keys in the cache (active or inactive):
   my @keys = $cache->keys;

   # Compute total cache size:
   my ($active, $inactive) = $cache->size;


=head1 NOTES

=head2 Cache creation

The cache is automatically created by any process that attempts to use it.
This may seem a little odd, but it makes it possible to do an emergency
purge of the cache by simply deleteing the full directory structure,
without affecting running code.

Some methods, however, should only be invoked by the cleanup process,
or by some centralized, reliable constructor for an application-specific
cache; these are:

     reap()
     purge()


=head2 Directory structure

Creates the following directory structure:

     Path
        |--- index.db
        |--- e-a1232cc9-af26-2-0.abc
        |--- e-a1232cc9-af26-2-1.def.Z
        |--- e-a1232c23-3232-4-0.xyz
        :

This is very similar to the Netscape cache structure.  Note that 
the extension of the cached file is preserved; this allows an HTTPD
to serve files directly out of the cache.  Even multiple extensions
are preserved, so that things like C<".hdf.Z"> will be remembered.
You can put a .htaccess file in the cache directory, if you like.

Each cache filename is itself of the form:

     e-(time)-(pid)-(seq)-(file).ext

Where:

     time    The time the entry object was created, as 8-digit hex
     pid     Pid of process that created the entry object, as 4-digit hex
     seq     Sequence number within this pid, as a decimal
     file    File number within this collection, as a decimal 


=head2 Access wrappers

It is vital that no one accidentally grab the cache for longer than
they need to, so all method calls that involve the database used tied
hashes that go out of scope when the method ends.


=head1 CACHING ALGORITHM

=head2 Introduction

There are two times we need to concern ourselves with:

=over

=item shelf-life

How long after a cached object was stored/last-accessed before it becomes
eligible for deletion from cache?  Normally, cached objects persist about
30 minutes, after which they are "stale" and ready for deletion.

=item linger-time

How longer after a cached object is I<logically> deleted from cache before it
may safely be I<physically> deleted?  This is usually a very small 
time: 2 minutes or less.
It ensures that "phase 2" (size-based) deletion does not yank the rug out
from under a process: processes are guaranteed the linger-time, even
if the cache has exploded.  Otherwise, in practice, it is unused.

B<Warning:> currently, it is assumed that linger time is SHORTER than 
shelf life!

=back


=head2 The cleanup algorithm

The following algorithm was I<adapted> from:

    A Simple Yet Robust Caching Algorithm
    Based on Dynamic Access Patterns

    James E. Pitkow & Margaret M. Recker
    Graphics, Visualization, & Usability Center
    College of Computing
    Georgia Institute of Technology
    Atlanta, GA 30332-0280 U.S.A.

Here is our modified version:

    let shelf-life = 30 minutes (for example)
    let linger-time = 2 minutes (for example)
         
    if (cache is full) or (just for the heck of it) then 
        
        # Phase 0: purge inactive files...
        reap any inactive files that have lingered more than linger-time
        
        # Phase 1: purge, from least to most fresh, until comfortable...
        foreach (entry, from least-fresh to most-fresh) do
            end if (active cache is at/below comfort level)
            if (entry is not fresh) then
	        delete entry                   # no need for linger
            end
	end
        
        # Phase 2: if phase 1 failed, emergency... be not-so-nice!
        foreach (entry, from largest to smallest) do
            end if (active cache is at/below comfort level)
            if (entry's last-access is before (now - linger-time)) then
                delete entry                  # no need for linger 
	    else
                doom entry                    # mark it: pick it up later
            end
	end
    end 

=head1 WARNINGS

No two cache entries must I<ever> contain pointers to the same cache file.
This would cause tremendous problems when reaping files!


=head1 BUGS

If more than 65535 cache entries are created by the same process within 
a single second, this will fail.  For our purposes, that will suffice.


=cut

use IU;
use IU::Log;
use IU::Store;
use IU::Util qw(units2secs units2bytes);

use strict;
use FileHandle;


#------------------------------
#
# GLOBALS
#
#------------------------------

# Sequence no (so two operations in the same second don't have identical ids):
my $SeqNo = 0;


#------------------------------------------------------------
# new
#------------------------------------------------------------
sub new {
    bless {}, shift;
}

#------------------------------------------------------------
# open PARAMHASH
#------------------------------------------------------------
# Params are...
#
#    Path   => the path to the cache directory

sub open {
    my $self = shift;
    my %params = @_;

    # Create:
    ref($self) or $self = $self->new(@_);

    # Init:
    $self->{Dir} = $params{Dir} or puke "missing directory";
    
    # If the cache directory doesn't exist, create it:
    unless (-d $self->{Dir}) {
	mkdir $self->{Dir}, 0755 or puke "mkdir $self->{Dir}: $!";
    }
    (-w $self->{Dir}) or puke 554, "No privs to write to the cache directory!";

    # Set the name of the index:
    $self->{Index} = "$self->{Dir}/index.db";
    $self->{EntryClass} = 'IU::FileCache::Entry';
    $self;
}



#------------------------------------------------------------
# lookup KEY, [OPTS]
#------------------------------------------------------------
# Get file from cache.  Returns a "cache file" object.
#
# If OPTS contains 'ALL', even inactive records are returned... however,
# the "lookup time" is only recorded for active files!
#
# If OPTS contains 'NOTOUCH', no attempt will be modify the "lookup time".  
# This is to allow admins to look-but-not-touch.

sub lookup {
    my ($self, $key, $opts) = @_;
    inform "Attempting to get $key";

    # Return immediately if there's no cache!
    (-e $self->{Index}) or return undef;

    # Extract opts:
    my $ALL     = ($opts and $opts =~ /\bALL\b/i);
    my $NOTOUCH = ($opts and $opts =~ /\bNOTOUCH\b/i);

    # Login (if needed): 
    my $login = $self->login('WRITER|BLOCK');

    # Get record:
    my $entry = $self->{CACHE}{"entry:$key"};      # look up entry
    if ($entry && $entry->active && !$NOTOUCH) {   # if found and active...
	$entry->lookup_time(time());                   # "touch" it
	$self->{CACHE}{"entry:$key"} = $entry;         # re-store
    } 

    # Logout (if needed) and return record:
    $self->logout if ($login ne 'ALREADY');
    return (($entry && ($entry->active || $ALL)) ? $entry : undef);
}

#------------------------------------------------------------
# insert KEY,ENTRY
#------------------------------------------------------------
# Store a logical file to cache.  Creates the cache index if necessary.
# This will also set the storage time of the object!
#
# Returns the ENTRY argument.

sub insert {
    my ($self, $key, $entry) = @_;
    inform "Inserting $key";

    # Login (if needed): 
    my $login = $self->login('WRCREAT|BLOCK');

    # First, lookup record: are we replacing an existing one?
    my $existing = $self->{CACHE}{"entry:$key"};
    if ($existing) {
	warning "insert $key: overwriting existing entry";
	$existing->unlink;          # get rid of physical files!
    }

    # Store record:
    $entry->insert_time(time());
    $self->{CACHE}{"entry:$key"} = $entry;

    # Logout (if needed) and return:
    $self->logout if ($login ne 'ALREADY');     # logout
    $entry;
}

#------------------------------------------------------------
# update KEY,ENTRY
#------------------------------------------------------------
# Update a logical file's information.  Does not create the cache index.
# This will NOT set the storage time of the object!
#
# Returns the ENTRY argument.

sub update {
    my ($self, $key, $entry) = @_;

    # Login (if needed): 
    my $login = $self->login('WRCREAT|BLOCK');

    # Store record:
    $self->{CACHE}{"entry:$key"} = $entry;

    # Logout (if needed) and return:
    $self->logout if ($login ne 'ALREADY');     # logout
    $entry;
}

#------------------------------------------------------------
# insert_files KEY, \@PATHS
#------------------------------------------------------------
# Store a collection of physical files to cache.
# Creates the cache index if necessary.
#
# Returns the created entry object.

sub insert_files {
    my ($self, $key, $paths) = @_;
    inform "Inserting files under $key";
    my @tgtpaths;

    # Get the time, pid, and sequence:
    my ($time, $pid, $seqno) = (time(), $$, $SeqNo++);

    # Copy the files:
    my $path;
    my $i;
    for ($i = 0; $i < int(@$paths); $i++) {
	my $srcpath = $paths->[$i];        # get path;
	my $srcext  = ext($srcpath);
	my $tgtpath = sprintf("$self->{Dir}/e-%08x-%04x-%d-%d.$srcext",
			      $time, $pid, $seqno, $i);
	cp($srcpath, $tgtpath) or puke "cp: $!";
	push @tgtpaths, $tgtpath;
    }

    # Create a new object:
    my $entry = $self->{EntryClass}->new(Paths=>\@tgtpaths, Key=>$key);

    # Store record, and return it:
    $self->insert($key, $entry);
    $entry;
}

#------------------------------------------------------------
# doom KEY
#------------------------------------------------------------
# Doom an entry.  This deactivates it.
# This will also set the doom time of the object!
#
# Returns false if entry is not there.

sub doom {
    my ($self, $key) = @_;
    inform "Dooming $key (logical delete)";

    # Login (if needed): 
    my $login = $self->login('WRCREAT|BLOCK');

    # Get entry:
    my $entry = $self->{CACHE}{"entry:$key"};
    if ($entry) {
	$entry->doom_time(time);
	$self->{CACHE}{"entry:$key"} = $entry;
    }

    # Logout (if needed) and return:
    $self->logout if ($login ne 'ALREADY');     # logout
    $entry;
}

#------------------------------------------------------------
# delete KEY
#------------------------------------------------------------
# Delete an entry.  This deactivates it and removes all component files.

sub delete {
    my ($self, $key) = @_;
    inform "Deleting $key";

    # Login (if needed): 
    my $login = $self->login('WRCREAT|BLOCK');

    # Get entry:
    my $entry = $self->{CACHE}{"entry:$key"};
    if ($entry) {
	$entry->unlink(time);
	delete $self->{CACHE}{"entry:$key"};
    }

    # Logout (if needed) and return:
    $self->logout if ($login ne 'ALREADY');     # logout
    1;
}

#------------------------------------------------------------
# reap LINGER
#------------------------------------------------------------
# Reap all inactive entries which haven't been used since LINGER ago,
# and were doomed at least that far back as well.   
#
# LINGER is a unit'ed timestring of the form '2s', '4m', '6h', '1d', etc.
# No units means seconds.
#
# WARNING: this is an expensive operation, and may tie up a large cache
# for several seconds or more.

sub reap {
    my ($self, $since) = @_;

    # How long ago:
    my $linger = units2secs($since);

    # Login (if needed): 
    my $login = $self->login('WRCREAT|BLOCK');

    # Get entry:
    my ($ckey, $entry);
    my @ckeys = CORE::keys %{$self->{CACHE}};
    foreach $ckey (@ckeys) {
	next if $ckey !~ /^entry:/;     # not an entry object

	# Get object, skipping if not a candidate:
	$entry = $self->{CACHE}{$ckey};
	next if $entry->active;                          # not dead
	next if (time - $entry->doom_time) < $linger;    # not decayed
	next if (time - $entry->access_time) < $linger;  # not forgotten
       
	# Kill it!
	inform "deleting cache entry: $ckey";
	$entry->unlink;	
	delete $self->{CACHE}{$ckey};
    }
    
    # Logout (if needed):
    $self->logout if ($login ne 'ALREADY');     # logout
    1;
}

#------------------------------------------------------------
# purge PARAMHASH
#------------------------------------------------------------
# Purge the cache.
#
#    MaxSize       unit'ed size: maximum cache size
#    ComfortSize   unit'ed size: comfortable cache size
#    ShelfLife     unit'ed time: how long since last usage before stale? 
#    LingerTime    unit'ed time: how long to persist after logical deletion? 
#    Force         if true, does a purge even if cache isn't full
#
# WARNING: this is an expensive operation, and may tie up a large cache
# for several seconds or more.

sub purge {
    my ($self, %params) = @_;
    my @entries;
    my @active;
    my $entry;
    my $key;

    # Sanity:
    $params{MaxSize} or puke "no max size!";
    $params{ComfortSize} or puke "no comfort size!";

    # Get args:
    my $max_size     = units2bytes($params{MaxSize});
    my $comfort_size = units2bytes($params{ComfortSize});
    my $shelf_life   = units2secs($params{ShelfLife}  || '30m');
    my $linger_time  = units2secs($params{LingerTime} || '2m');


    # Login (if needed): 
    my $login = $self->login('WRCREAT|BLOCK');

    # Get current size:
    my ($active_size, $inactive_size) = $self->size;
    (($active_size >= $max_size) || $params{Force}) or return 0;   # no-op
        

    # 
    # Phase 0: purge inactive files (after linger), regardless of cache size...
    #
    debug "purge: phase 0";
    $self->reap($linger_time);
    
        
    # 
    # Phase 1: purge, from least to most fresh, until comfortable...
    #
    debug "purge: phase 1";
    
    # Get all active entries, sorted from LRU to MRU:
    @entries = map { $self->{CACHE}{"entry:$_"} } $self->keys;
    @active = grep { $_->active } @entries;
    @active = sort { 
	$IU::FileCache::a->usage_time <=> $IU::FileCache::b->usage_time
	} @active;
    
    # Delete candidates from LRU to MRU, until happy:
    foreach $entry (@active) {
	debug "active size now: $active_size bytes";
	last if ($active_size <= $comfort_size);
	if ((time - $entry->usage_time) > $shelf_life) {   # stale!
	    $self->delete($entry->key);
	    $active_size -= $entry->size;
	}
    }
    
    
    # 
    # Phase 2: if phase 1 failed, emergency... be not-so-nice!
    #
    debug "purge: phase 2";
    
    # Get all active entries, sorted from LARGEST to SMALLEST:
    @active = grep { $_->active } @active;
    @active = sort { 
	$IU::FileCache::b->size <=> $IU::FileCache::a->size
	} @active;
    
    # Delete candidates from largest to smallest, until happy:
    foreach $entry (@active) {
	debug "active size now: $active_size bytes";
	last if ($active_size <= $comfort_size);
	if ((time - $entry->access_time) > $linger_time) {   # forsaken!
	    $self->delete($entry->key);
	    $active_size -= $entry->size;
	}
	else {           # used too recently... just doom it:
	    $self->doom($entry->key);
	}
    }
    
    # DONE!
    1;
}

#------------------------------------------------------------
# keys
#------------------------------------------------------------
# Return an array of keys, in no particular order.
# Some keys may be inactive!!!!

sub keys {
    my $self = shift;
    my $login = $self->login('READER|BLOCK');      # login, if needed

    debug "got login: tie to $self->{Index}";
    my @ckeys =  CORE::keys %{$self->{CACHE}};

    $self->logout if ($login ne 'ALREADY');        # logout, if needed
    map { /^entry:(.*)/ ? $1 : () } @ckeys;
}

#------------------------------------------------------------
# size
#------------------------------------------------------------
# Return the total size of the cache.
# In a scalar context, returns the active+inactive size.
# In a list context, returns (active, inactive).

sub size {
    my $self = shift;
    my $login = $self->login('READER|BLOCK');      # login, if needed

    my ($ckey, $cval);
    my ($active, $inactive) = (0, 0);
    while (($ckey, $cval) = each %{$self->{CACHE}}) {
	next if $ckey !~ /^entry:/;     # not an entry object
	($cval->active ? $active : $inactive) += $cval->size;
    }

    $self->logout if ($login ne 'ALREADY');        # logout, if needed
    return (wantarray ? ($active, $inactive) : $active+$inactive);
}

#------------------------------------------------------------
# dump
#------------------------------------------------------------
sub dump {
    my $self = shift;

    print "------------------------\n";
    my ($a, $ia) = $self->size;
    print "active   = $a bytes\n";
    print "inactive = $ia bytes\n";
    my $login = $self->login('READER');
    my @ckeys = CORE::keys %{$self->{CACHE}};
    my $ckey;
    foreach $ckey (sort @ckeys) {
	next if $ckey !~ /^entry:/;     # not an entry object
	my $entry = $self->{CACHE}{$ckey};
	print "$ckey = \n";
	$entry->dump;
    }
    $self->logout if ($login ne 'ALREADY');
}



#------------------------------
#
# EXPERT STUFF
#
#------------------------------

#------------------------------------------------------------
# login OPTS...
#------------------------------------------------------------
# Manually tie the underlying database.  Returns a ref to a tied hash.

sub login {
    my ($self, @opts) = @_;
    return 'ALREADY' if $self->{CACHE};
    $self->{CACHE} = {};
    tie(%{$self->{CACHE}}, 'IU::Store', $self->{Index}, @opts)
	or puke "tie to $self->{Index} failed: $!";
    'DONE';
}

#------------------------------------------------------------
# logout LOGIN
#------------------------------------------------------------
# Manually untie the underlying database.

sub logout {
    my ($self, $login) = @_;
    untie %{$self->{CACHE}};
    delete $self->{CACHE};
}



#------------------------------
#
# UTILITIES
#
#------------------------------

#------------------------------------------------------------
# cp SOURCEPATH, TARGETPATH
#------------------------------------------------------------
sub cp {
    my ($spath, $tpath) = @_;        
    my $out = `cp $spath $tpath 2>&1`;
    (($? >> 8) == 0) 
	or puke(454, "cp $spath $tpath ($out: exit status ", $?>>8, ")");
    1;
}

#------------------------------------------------------------
# ext PATH
#------------------------------------------------------------
# Return the extension.  Things like "foo.hdf.Z" return "hdf.Z"
# The default is "dat".

sub ext {
    my $path = shift;
    $path =~ s|^.*/||;            # remove any directory information
    ($path =~ /\.(.+)$/) ? $1 : 'dat';
}

#------------------------------------------------------------
# max NUMBERS
#------------------------------------------------------------
sub max {
    my $max;
    for ($max = shift; @_; shift) { $max = $_[0] if ($_[0] > $max) }
    $max;
}

#------------------------------------------------------------
# min NUMBERS
#------------------------------------------------------------
sub min {
    my $min;
    for ($min = shift; @_; shift) { $min = $_[0] if ($_[0] < $min) }
    $min;
}


#============================================================
#============================================================

package IU::FileCache::Entry;

=head2 IU::FileCache::Entry

A logical entry in the file cache.  This is what a "lookup" returns.

    # Get last time any *file* in the entry was read:
    my $mra = $entry->lookup_time('MOST_RECENT');

=cut

use IU;
use IU::Log;
use IU::Store;

use strict;

#------------------------------------------------------------
# new PARAMHASH
#------------------------------------------------------------
sub new {
    my ($class, %params) = @_;
    my $self = bless {}, $class;
    $self->{LookupTime} = 0;
    $self->{StorageTime} = 0;
    $self->{DoomTime} = 0;
    $self->{Paths} = ($params{Paths} || []);
    $self->{Key}   = $params{Key};
    $self;
}

#------------------------------------------------------------
# paths
#------------------------------------------------------------
# Get paths. Returns LIST of paths, or empty.

sub paths {
    @{$_[0]->{Paths} || []};
}

#------------------------------------------------------------
# size
#------------------------------------------------------------
# Get total size of all files, in bytes.
# Computes it once per instance, and caches it!

sub size {
    my $self = shift;

    unless (defined($self->{Size})) {
	my $path;
	$self->{Size} = 0;
	foreach $path ($self->paths) { $self->{Size} += (-s $path) }
    }
    $self->{Size};
}

#------------------------------------------------------------
# key
#------------------------------------------------------------
# Get the key.

sub key {
    $_[0]->{Key};
}

#------------------------------------------------------------
# access_time 
#------------------------------------------------------------
# Last time any of the files were physically accessed.

sub access_time {
    my $self = shift;
    my $path;
    my $atime = 0;
    foreach $path ($self->paths) {
	my $path_atime = (stat($path))[8];
	$atime = ($path_atime > $atime) ? $path_atime : $atime;
    }
    $atime;
}

#------------------------------------------------------------
# insert_time [TIME]
#------------------------------------------------------------
# Time this record was stored to cache.

sub insert_time {
    my ($self, $time) = @_;
    @_ > 1 and $self->{StorageTime} = $time;
    $self->{StorageTime};
}

#------------------------------------------------------------
# lookup_time [TIME]
#------------------------------------------------------------
# Last time this record was looked up in cache.

sub lookup_time {
    my ($self, $time) = @_;
    @_ > 1 and $self->{AccessTime} = $time;
    $self->{AccessTime};
}

#------------------------------------------------------------
# doom_time [TIME]
#------------------------------------------------------------
# Time this record was DOOMED (logically deleted).
# Setting this to a non-zero value causes "active" to return false.

sub doom_time {
    my ($self, $time) = @_;
    @_ > 1 and $self->{DoomTime} = $time;
    $self->{DoomTime};
}

#------------------------------------------------------------
# usage_time
#------------------------------------------------------------
# For determining freshness... how long ago was it last used?

sub usage_time {
    my $self = shift;
    $self->lookup_time || $self->insert_time;
}

#------------------------------------------------------------
# active
#------------------------------------------------------------
# Is this record active?

sub active {
    my $self = shift;
    !$self->{DoomTime};
}

#------------------------------------------------------------
# unlink
#------------------------------------------------------------
# Unlink all the component paths.

sub unlink {
    my $self = shift;
    my $path;
    foreach $path ($self->paths) {
	inform "unlinking $path";
	unlink $path;
    }
    1;
}

#------------------------------------------------------------
# dump
#------------------------------------------------------------
# Dump the object.

sub dump {
    my $self = shift;
    my $path;
    print "  Inserted: ", scalar(localtime($self->insert_time)), "\n" 
	if $self->insert_time;
    print "  Lookup: ", scalar(localtime($self->lookup_time)), "\n" 
	if $self->lookup_time;
    print "  Doomed: ", scalar(localtime($self->doom_time)), "\n" 
	if $self->doom_time;
    print "  Access: ", scalar(localtime($self->access_time)), "\n" 
	if $self->access_time;
    print "  Size: ", $self->size, " bytes\n";
    foreach $path ($self->paths) {
	print "    Path: $path\n";
    }
    1;
}

#------------------------------------------------------------
1;
