=head1 NAME

IU::FileLogger -- log to a standard-format logfile 

=head1 SYNOPSIS

    $logger = open IU::FileLogger 'filename'
    $logger = open IU::FileLogger \*STDOUT;

=cut


package IU::FileLogger;

use IU::Logger;
use IU::Util;

@IU::FileLogger::ISA = qw(IU::Logger);

use strict;
no strict 'refs';


#------------------------------------------------------------
#
# Globals
#
#------------------------------------------------------------

# The next filehandle:
my $GENSYM = 'FH00';


#------------------------------------------------------------
# open OPENEXPR
#------------------------------------------------------------
# Open a file logger stream.
#
# The OPENEXPR may be anything you could give to open(), plus a
# FileHandle reference or typeglob reference.

sub open {
    my ($type, $openexpr) = @_;

    # Create the object:
    my $new = IU::Logger->open;

    # Open the file, if necessary:    
    if (ref($openexpr)) {        # probably a FileHandle
	$new->{FH} = $openexpr;
    }
    else {                       # need to open the file
	++$GENSYM;
	open($GENSYM, $openexpr) or 
	    return IU::Log::error("open $openexpr: $!");
	$new->{FH} = "IU::FileLogger::$GENSYM";
    }
    $new->{Opened} = $openexpr;

    # Bless, and return:
    bless $new, $type;
}

#------------------------------------------------------------
# fh -- return the currently-open filehandle
#------------------------------------------------------------
sub fh {
    my $self = shift;
    $self->{FH};
}

#------------------------------------------------------------
# autoflush -- convenience: turn on autoflush for the filehandle
#------------------------------------------------------------
sub autoflush {
    my $self = shift;
    $self->{FH} or return undef;
    my $oldfh;
    $oldfh = select($self->{FH}); $| = 1; select($oldfh);
    1;
}

#------------------------------------------------------------
# close -- close the stream (by closing the filehandle)
#------------------------------------------------------------
sub close {
    my $self = shift;
    if ($self->{FH}) {
	close $self->{FH};
	$self->{FH} = undef;
    }
    1;
}

#------------------------------------------------------------
# msg -- log the given message
#------------------------------------------------------------
sub msg {
    my ($self, $msgtype, $msgno, $caller, @msg) = @_;

    # Format the message, replacing all illegal whitespace with spaces:
    my $msg = join '', @msg;
    $msg =~ s/\s/ /g;

    # Get the current date and time, as YYYY/MM/DD HH:MM:SS
    my @now = localtime;
    #($now[5] < 100) and $now[5] += (($now[5] < 50) ? 2000 : 1900);
    my $timestr = sprintf("%04d/%02d/%02d %02d:%02d:%02d", 
			  1900+$now[5], 1+$now[4], $now[3], 
			  $now[2], $now[1], $now[0]);

    # Get the filename, erasing everything before last two path elems:
#    my $file = (($caller->[1] =~ m|/perl/(.+)$|) ? $1 : 'eval');
    my $file = (($caller->[1] =~ m|/(([^/]+)/([^/]+)$)|) ? $1 : $caller->[1]);

    # First line: type num datetime file line
    my $FH = $self->{FH};
    print $FH $IU::Log::Glyphs{$msgtype},  # type
              sprintf(" %03d ", $msgno),   # num
              $timestr, ' ',               # date time
              "$file $caller->[2] ",       # file line
              "\n";

    # Second line: the message:
    print $FH '     ', $msg, "\n"; 
}

#------------------------- end of file -------------------------
1;
