package IU::Geo::PathRow;

=head1 NAME

IU::Geo::PathRow - a path-row-based geographic search, like a PATH_ROW_LOC

=head1 SYNOPSIS

    my $pr = new IW:::GeoPathRow;
    
    # Paths...
    $pr->path(12);               # set the path
    $pr->path(12, 34);           # set the path, as a range
    $path = $pr->path;           # get the minimum path  
    ($min, $max) = $pr->path;    # get the path range

    # Rows...
    $pr->row(56);                # set the row
    $pr->row(56, 78);            # set the row, as a range
    $row = $pr->row;             # get the minimum row  
    ($min, $max) = $pr->row;     # get the row range  
    
    # Scheme:
    $pr->scheme("WRS-1");        # set scheme
    $pr->scheme;                 # get scheme
    
    # Get the bounding rectangle, as a IU::Geo::Range:
    $georect = $pr->bound_rect;

=head1 INSTANCE VARIABLES

    Paths  => array ref, [min, max]
    Rows   => array ref, [min, max]
    Scheme => WRS scheme name, like "WRS-1"

=head1 AUTHOR

Eryq (Erik Dorfman), Hughes STX Corporation

=head1 VERSION

$Id: PathRow.pm,v 1.4 2002/01/14 20:21:11 seufert Exp $

=cut

use IU;
use IU::Coord;
use IU::Log;
use IU::Geo::WRS;
use IU::IW;

use IW::Util (qw(blank));          # KLUDGE

use strict;
use vars qw(@ISA);
@ISA = qw(IU::Geo::Base);


#------------------------------------------------------------
# new PARAMHASH
#------------------------------------------------------------
# I<Constructor.>
#
sub new {        
    my $class = shift;
    my %params = @_;
    my $self = \%params;
    $self->{Paths} ||= [];
    $self->{Rows} ||= [];
    bless $self, $class;	
}

#------------------------------------------------------------
# as_string
#------------------------------------------------------------
# I<Instance method.>  Return the path-row as a string, of the form:
#
#    "SCHEMENAME: Paths 1 to 5, Rows 10 to 12
#
sub as_string {
    my $self = shift;
    
    return ($self->scheme . ": " .
	    "Paths " . join(" to ", $self->path) . ", " .
	    "Rows "  . join(" to ", $self->row));
}

#------------------------------------------------------------
# path [MINPATH],[MAXPATH]
#------------------------------------------------------------
# I<Instance method.>  Set/get the path.
# In scalar context, returns min path; else, returns path range.
#
sub path {
    my ($self, $min, $max) = @_;
    if (@_ > 1) {
	$max = $min if blank($max); $min = $max if blank($min);
	$self->{Paths} = [$min, $max];
    }
    (wantarray ? @{$self->{Paths}} : $self->{Paths}[0]);
}

#------------------------------------------------------------
# row [MINROW],[MAXROW]
#------------------------------------------------------------
# I<Instance method.>  Set/get the row.
# In scalar context, returns min row; else, returns row range.
#
sub row {
    my ($self, $min, $max) = @_;
    if (@_ > 1) {
	$max = $min if blank($max); $min = $max if blank($min);
	$self->{Rows} = [$min, $max];
    }
    (wantarray ? @{$self->{Rows}} : $self->{Rows}[0]);
}

#------------------------------------------------------------
# scheme [SCHEME]
#------------------------------------------------------------
# I<Instance method.>  Set/get the scheme.
#
sub scheme {
    my ($self, $scheme) = @_;
    $self->{Scheme} = $scheme if (@_ > 1);
    $self->{Scheme};
}

#------------------------------------------------------------
# bound_rect
#------------------------------------------------------------
# I<Instance method.>  Convert to an IU::Geo::Range.
#
sub bound_rect {
    my $self = shift;

    # Get WRS interface:
    my $wrs = IU::Geo::WRS->new($self->scheme);
    $wrs or return error("No such WRS scheme: ", $self->scheme);
    
    # Get the bounding box:
    return $wrs->bound_rect([$self->path], [$self->row]);
}

#------------------------------------------------------------
# fixup 
#------------------------------------------------------------
# Given a path-row, fix it so that...
#      
#   * Everything has a numeric value.
#   * The min path/row precedes the max
#   * Everything is clipped to (-180, 90, 180, -90).
#
# Return the self object on success.
#
sub fixup {
    my $self = shift;

    # Paths...
    if ($self->{Paths}[0] > $self->{Paths}[1]) {
	@{$self->{Paths}} = ($self->{Paths}[1], $self->{Paths}[0]);
    }
    # Rows...
    if ($self->{Rows}[0] > $self->{Rows}[1]) {
	@{$self->{Rows}} = ($self->{Rows}[1], $self->{Rows}[0]);
    }
    $self;
}

#------------------------------------------------------------
# validate
#------------------------------------------------------------
sub validate {
    my $self = shift;
    $self->bound_rect;
    $self;
}


#------------------------------
#
# CGI interface
#
#------------------------------

#------------------------------------------------------------
# check_CGI PREFIX
#------------------------------------------------------------
# I<Class/instance method.>  Check the CGI tags.
# Returns 0, 'SOME', or 'ALL', based on whether the user seemed to
# enter nothing, some things, or everything that was needed.
#
# NOTE: the scheme is NOT checked, since it is assumed to come from a
# popup.  This REALLY should be changed to check for the special '(none)'
# scheme.
# 
# The PREFIX is taken to precede the tag names.
#
sub check_CGI {
    my $class = shift;
    my $prefix = shift;    # followed by standard rectangle element names
    my $field;
    my %got;

    # Sum up:
    foreach $field (qw(start-path end-path start-row end-row)) {
	$got{$field} = 1 if (!blank($REQ->param("${prefix}${field}")));
    }

    # How'd they do?
    if (int(keys %got) == 0) {
	return 0;       # all blank;
    }
    if (($got{'start-path'} || $got{'end-path'}) &&
	($got{'start-row'}  || $got{'end-row'})) {
	return 'ALL';
    }
    return 'SOME';
}

#------------------------------------------------------------
# from_CGI PREFIX
#------------------------------------------------------------
# I<Constructor.>  Read from CGI tags.
# The PREFIX is taken to precede the tag names.
#
sub from_CGI {
    my $class = shift;
    my $prefix = shift;    # followed by standard rectangle element names
    my (@paths, @rows);

    # Check to make sure nothing's blank:
    my $status = $class->check_CGI($prefix);
  
    if ($status eq 'SOME') {

       puke 516, "missing path/row information";
    } 
    elsif ($status eq 'ALL') {

       # Create the object:
       my $self = $class->new;
       $self->scheme($REQ->param("${prefix}scheme"));
       $self->path($REQ->param("${prefix}start-path"),
                   $REQ->param("${prefix}end-path"));
       $self->row($REQ->param("${prefix}start-row"),
                   $REQ->param("${prefix}end-row"));   
       return $self;
    }
    else {

       return undef;
    } 
}

#------------------------------------------------------------
# list_CGI PREFIX
#------------------------------------------------------------
# I<Instance method.>
# Return an array of CGI [tag=>value] pairs which can be used to init 
# this object.  The PREFIX precedes all tags, as it does in from_CGI.
#
sub list_CGI {
    my ($self, $p) = @_;
    $p ||= '';
    my @tags = (["${p}scheme"     => $self->scheme],
		["${p}start-path" => ($self->path)[0]],
		["${p}end-path"   => ($self->path)[1]],
		["${p}start-row"  => ($self->row)[0]],
		["${p}end-row"    => ($self->row)[1]]
		);
    @tags;
}



#------------------------------
#
# ODL interface
#
#------------------------------

#------------------------------------------------------------
# from_ODL ODL
#------------------------------------------------------------
# I<Constructor.>  Read from an ODL object that has the same structure
# as a "PATH_ROW_LOC" group.
#
sub from_ODL {
    my $class = shift;
    my $odl = shift;

    my $self = $class->new;
    my $value;
    if ($value = $odl->param("WRS_TYPE")) {
	$self->scheme($value->text_value);	
    }
    if ($value = $odl->param("PATH")) {
	$self->path($value->text_values);     # expect 12 or (12,34)
    }
    if ($value = $odl->param("ROW")) {
	$self->row($value->text_values);     # expect 56 or (56,78)
    }
    $self;
}

#------------------------------------------------------------
# to_ODL 
#------------------------------------------------------------
# I<Instance method.>  Return as an ODL group, identical in structure
# to a "PATH_ROW_LOC" group.
#
sub to_ODL {
    my $self = shift;

    my $odl = IU::ODL->new_group;
    $odl->add_param(WRS_TYPE => IU::ODL->quoted($self->scheme));
    $odl->add_param(PATH     => IU::ODL->conv([$self->path]));
    $odl->add_param(ROW      => IU::ODL->conv([$self->row]));
    $odl;
}

#------------------------------------------------------------
# ODL_type
#------------------------------------------------------------
sub ODL_type { "PATH_ROW_LOC" }

#------------------------------------------------------------
1;
