package IU::Geo::Point;

=head1 NAME

IU::Geo::Point - a point search region, like a POINT_LOC 

=head1 SYNOPSIS

    my $point = new IW:::GeoPoint;
    $point->lat(45.0);           # set latitude
    $point->lon(-20.0);          # set longitude

=head1 INSTANCE VARIABLES

    Lat => latitude
    Lon => longitude

=head1 AUTHOR

Eryq (Erik Dorfman), Hughes STX Corporation

=head1 VERSION

$Id: Point.pm,v 1.6 2003/10/06 20:49:50 seufert Exp $

=cut

use IU::Coord;
use IU::IW;        # for CGI methods
use IU::Geo::Base;

use strict;
use vars qw(@ISA);
@ISA = qw(IU::Geo::Base);


#------------------------------------------------------------
# new [LAT,LON]
#------------------------------------------------------------
# I<Class method, constructor.>
#
sub new {        
    my $class = shift;
    my $self = {};
    $self->{Lat} = shift;
    $self->{Lon} = shift;
    bless $self, $class;	
}

#------------------------------------------------------------
# ODL_type
#------------------------------------------------------------
sub ODL_type { "POINT_LOC" }

#------------------------------------------------------------
# _one VARNAME,SELF,[VALUE]
#------------------------------------------------------------
# I<Private utility.>  Get/set a single-valued instance variable.
sub _one {
    my ($varname, $self, $value) = @_;
    $self->{$varname} = $value if (@_ > 2);
    $self->{$varname};
}

#------------------------------------------------------------
# lat [LATITUDE]
#------------------------------------------------------------
# I<Instance method.>  Set/get latitude.
#
sub lat {
    my $self = shift;
    $self->{Lat} = shift if (@_);
    $self->{Lat};
}

#------------------------------------------------------------
# lon [LONGITUDE]
#------------------------------------------------------------
# I<Instance method.>  Set/get longitude.
#
sub lon {
    my $self = shift;
    $self->{Lon} = shift if (@_);
    $self->{Lon};
}

# zoom [INTEGER]        -- get/set the WG zoom factor
sub zoom       { _one('Zoom', @_) }

# feature [STRING]   -- get/set the feature name associated with 
# the lat/lon point e.g. "Atlanta GA". Optional

sub feature { _one('Feature', @_) }

#------------------------------------------------------------
# as_string
#------------------------------------------------------------
# I<Instance method.>  Return the point as a string, of the form:
#
#    "30.23^ Lat, -12.1^ Lon"
#
# Each "^" is a Latin-1 degree mark, "\xB0".

sub as_string {
    my ($self, %params) = @_;
    my @lines;
    if ($params{Verbose} > 1) {
        push @lines, "Feature Name: ".$self->feature if $self->feature;
    }
    push @lines, $self->{Lat}."\xB0 Lat, ".$self->{Lon}."\xB0 Lon";
    join(($params{Verbose} ? "\n" : ", "), @lines);
}

#------------------------------------------------------------
# fixup 
#------------------------------------------------------------
# I<Instance method, override.>  
# Given a point, fix it so that...
#      
#   * Everything has a numeric value, defaults (0, 0).
#   * Everything is clipped to (-180, 90, 180, -90).
#   * All numbers are returned as floats, with $precision-digit precision.
#
# Return the self object on success.

sub fixup {
    my $self = shift;
    my $prec = shift;        # precision, in digits
    defined($prec) || ($prec = 4);

    # For quicker manipulation:
    my $lat = $self->{Lat};
    my $lon = $self->{Lon};

    # Force all to be numbers:
    (defined($lat) && $lat =~ /\d/) || ($lat =  0.0);
    (defined($lon) && $lon =~ /\d/) || ($lon =  0.0);

    # Clip:
    ($lat <=  90.0)  || ($lat =  90.0);
    ($lat >= -90.0)  || ($lat = -90.0);
    ($lon <= 180.0)  || ($lon =  180.0);
    ($lon >= -180.0) || ($lon = -180.0);

    # Adjust and save:
    $self->{Lat} = sprintf("%0.${prec}f", $lat);
    $self->{Lon} = sprintf("%0.${prec}f", $lon);

    # Done!
    $self;
}




#------------------------------
#
# CGI interface
#
#------------------------------

#------------------------------------------------------------
# from_CGI PREFIX
#------------------------------------------------------------
# I<Class method, override.>  
# Read from CGI tags.  The PREFIX is taken to precede the tag names.

sub from_CGI {
    my $class = shift;
    my $p = shift;    # followed by standard rectangle element names

    return undef unless defined($REQ->param("${p}lat"));
 
    my $new = $class->new;
    $new->{Lat} = str2deg($REQ->param("${p}lat"));
    $new->{Lon} = str2deg($REQ->param("${p}lon"));
    $new->zoom($REQ->param("${p}zoom") || '1');
    $new->feature($REQ->param("${p}feature") || undef);
    $new;
}

#------------------------------------------------------------
# list_CGI PREFIX
#------------------------------------------------------------
# Return an array of CGI [tag=>value] pairs which can be used to init 
# this object.  The PREFIX precedes all tags, as it does in from_CGI.
#
sub list_CGI {
    my ($self, $p) = @_;
    $p ||= '';
    my @tags;
    push @tags, ["${p}lat" => $self->lat];
    push @tags, ["${p}lon" => $self->lon];
    push @tags, ["${p}feature" => $self->feature] if $self->feature;
    return @tags;
}



#------------------------------
#
# ODL interface
#
#------------------------------

#------------------------------------------------------------
# from_ODL ODL
#------------------------------------------------------------
# I<Constructor.>  
# Read from ODL group with the same structure as a "POINT_LOC" group.

sub from_ODL {
    my $class = shift;
    my $odl = shift;    # followed by standard rectangle element names

    my $self = $class->new;
    my $value;
    if ($value = $odl->param('LATITUDE')) {
	$self->{Lat} = $value->text_value;
    }
    if ($value = $odl->param('LONGITUDE')) {
	$self->{Lon} = $value->text_value;
    }
    if ($value = $odl->param('WG_FEATURE')) {
        $self->feature($value->text_value);
    }
    if ($value = $odl->param('WG_ZOOM')) {
        $self->zoom($value->text_value);
    }
    $self;
}

#------------------------------------------------------------
# to_ODL 
#------------------------------------------------------------
# I<Instance method.>  
# Return as an ODL group, identical in structure to a "POINT_LOC" group.

sub to_ODL {
    my $self = shift;

    my $odl = IU::ODL->new_group;
    $odl->add_param(LATITUDE  => IU::ODL->real($self->lat));
    $odl->add_param(LONGITUDE => IU::ODL->real($self->lon));

    #------------------------------
    # Web-gateway specific...

    if ($self->zoom) {
        $odl->add_param(WG_ZOOM => IU::ODL->int($self->zoom));
    }
    if ($self->feature) {
        $odl->add_param(WG_FEATURE =>
                      IU::ODL->string($self->feature));
    }

    $odl;
}

#------------------------------------------------------------
1;

