package IU::Geo::Polygon;

=head1 NAME

IU::Geo::Polygon - a polygonal region, like a POLYGON_LOC 

=head1 SYNOPSIS

    my $poly = new IW:::GeoPolygon;
    
    $poly->tangent;            returns an IU::Geo::Point
    $poly->points->[0];        get the first point, an IU::Geo::Point
    $poly->projection;         get the map projection 
    $poly->pole;               pole included? N, S, B, or X

=head1 AUTHOR

Eryq (Erik Dorfman), Hughes STX Corporation

=head1 VERSION

$Id: Polygon.pm,v 1.8 2003/09/04 17:36:22 seufert Exp $

=cut

use IU::Coord;
use IU::ODL;
use IU::Geo::Point;
use IU::IW;

use strict;
use vars qw(@ISA);
@ISA = qw(IU::Geo::Base);


#------------------------------------------------------------
# new
#------------------------------------------------------------
sub new {        
    my $class = shift;
    my $self = {};
    $self->{Pole} = 'X';
    $self->{Points} = [];
    bless $self, $class;	
}

#------------------------------------------------------------
# _one VARNAME,SELF,[VALUE]
#------------------------------------------------------------
# I<Private utility.>  Get/set a single-valued instance variable.
sub _one {
    my ($varname, $self, $value) = @_;
    $self->{$varname} = $value if (@_ > 2);
    $self->{$varname};   
}

# tangent [POINT]       --  get/set the tangent point
sub tangent    { _one('Tangent', @_) }

# center [POINT]       --  get/set the tangent point
sub center    { _one('Center', @_) }


# projection [STRING]   --  get/set the map projection
sub projection { _one('Projection', @_) }

# points [\@POINTS]     -- get/set a reference to the array of 4 points
sub points     { _one('Points', @_) }

# pole [STRING]         -- get/set the pole included (N, S, B, or X)
sub pole       { _one('Pole', @_) }

# zoom [INTEGER]        -- get/set the WG zoom factor
sub zoom       { _one('Zoom', @_) }

# feature [STRING]   -- get/set the feature name associated with
# the lat/lon point e.g. "Atlanta GA". Optional

sub feature { _one('Feature', @_) }




#------------------------------------------------------------
# as_string PARAMHASH
#------------------------------------------------------------
# I<Instance method, override.>  
# Return the polygon as a string, of the form:
#
#    "POINT; POINT; POINT; POINT"
#
# Each POINT is the stringification of one of the component points.

sub as_string {
    my ($self, %params) = @_;
    if ($params{Verbose}) {
	my @lines;
	if ($params{Verbose} > 1) {
            push @lines, "Feature Name: ".$self->feature
                if $self->feature;
	    push @lines, "Projection: ".$self->projection 
		if $self->projection;
	    push @lines, "Tangent: ".$self->tangent->as_string 
		if $self->tangent;
            push @lines, "Center: ".$self->center->as_string
                if $self->center;

	}
	push @lines, "Points in clockwise order are:";
	push @lines, map { $_->as_string } @{$self->points};
	return join "\n", @lines;
    }
    else {
	return join "; ", map { $_->as_string } @{$self->points};
    }
}

#------------------------------------------------------------
# fixup [PREC]
#------------------------------------------------------------
# I<Instance method, override.>  
# Given a polygon, fix it up.  Return the self object on success.

sub fixup {
    my $self = shift;
    my $prec = shift;        # precision, in digits
    defined($prec) || ($prec = 4);

    # Pole:
    $self->{Pole} ||= 'X';

    # Tangent:
    $self->{Tangent} or die "I need a tangent!";
    $self->{Tangent}->fixup;

    # Projection:
    $self->{Projection} or die "I need a projection!";
    
    # Points:
    (int(@{$self->{Points}}) == 4) or die "wrong number of points!";
    my $point;
    foreach $point (@{$self->{Points}}) {
	$point->fixup($prec);
    }

    # Done!
    $self;
}


#------------------------------------------------------------
# tangent_lat           -- get the tangent lat, as an integer
# tangent_lon           -- get the tangent lon, as an integer
#------------------------------------------------------------
sub tangent_lat {
    my $self = shift;
    my $t = $self->tangent || return undef;
    int($t->lat);
}
sub tangent_lon {
    my $self = shift;
    my $t = $self->tangent || return undef;
    int($t->lon);
}




#------------------------------
#
# CGI interface
#
#------------------------------

#------------------------------------------------------------
# from_CGI PREFIX
#------------------------------------------------------------
# I<Constructor.>  Read from CGI tags.
# The PREFIX is taken to precede the tag names.
#
# The pole is called           ${prefix}pole
# The projection is called     ${prefix}projection
# The tangent has a prefix of  ${prefix}tangent.
# Each point has a prefix of   ${prefix}0.   ${prefix}1.   etc.

sub from_CGI {
    my $class = shift;
    my $prefix = shift || '';    # followed by standard polygon element names
    my $self = $class->new;

    # Simple stuff:
    $self->projection($REQ->param("${prefix}projection"));
    $self->pole($REQ->param("${prefix}pole") || 'X');
    $self->zoom($REQ->param("${prefix}zoom") || '1');
    $self->feature($REQ->param("${prefix}feature") || undef);

    # The points:
    $self->{Tangent} = IU::Geo::Point->from_CGI("${prefix}tangent.");

    $self->{Center} = IU::Geo::Point->from_CGI("${prefix}4.");

    my $i;
    for $i (0..3) {
	$self->{Points}[$i] = IU::Geo::Point->from_CGI("${prefix}$i.");
    }
    $self;
}

#------------------------------------------------------------
# list_CGI PREFIX
#------------------------------------------------------------
# Return an array of CGI [tag=>value] pairs which can be used to init 
# this object.  The PREFIX precedes all tags, as it does in from_CGI.
#
sub list_CGI {
    my ($self, $p) = @_;
    $p ||= '';
    my @tags;
    
    # Simple stuff:
    push @tags, ["${p}projection" => $self->projection];
    push @tags, ["${p}zoom"       => $self->zoom];
    push @tags, ["${p}pole"       => $self->pole];
    push @tags, ["${p}feature"       => $self->feature] 
       if $self->feature;

    # The tangent:
    if ($self->tangent) {
	push @tags, $self->tangent->list_CGI("${p}tangent.");
    }

    # The Center:
    if ($self->center) {
        push @tags, $self->center->list_CGI("${p}4.");
    }


    # The points:
    my $i;
    for $i (0..3) { push @tags, $self->points->[$i]->list_CGI("${p}$i.") }
    
    # Done!
    return @tags;
}


#------------------------------
#
# ODL interface
#
#------------------------------

#------------------------------------------------------------
# from_ODL ODL
#------------------------------------------------------------
# I<Constructor.>  Read from ODL group with the same structure
# as a "RANGE_LOC" group.

sub from_ODL {
    my $class = shift;
    my $odl = shift;    # followed by standard rectangle element names
    my $value;
    my ($tlat, $tlon);
    my (@lats, @lons);

    my $self = $class->new;

    # Get basic attributes:
    if ($value = $odl->param('POLE_INCLUDED')) {
	$self->pole($value->text_value);
    }
    if ($value = $odl->param('MAP_PROJECTION_TYPE')) {
	$self->projection($value->text_value);
    }
    if ($value = $odl->param('LATITUDE')) {
	@lats = $value->text_values;
    }
    if ($value = $odl->param('LONGITUDE')) {
	@lons = $value->text_values;
    }
    if ($value = $odl->param('TANGENT_LATITUDE')) {
	$tlat = $value->text_value;
    }
    if ($value = $odl->param('TANGENT_LONGITUDE')) {
	$tlon = $value->text_value;
    }
    if ($value = $odl->param('WG_ZOOM')) {
	$self->zoom($value->text_value);
    }
    if ($value = $odl->param('WG_FEATURE')) {
        $self->feature($value->text_value);
    }

    # Convert tangent:
    (defined($tlat) && defined($tlon)) or die "ODL: missing tangent point";
    $self->{Tangent} = new IU::Geo::Point $tlat, $tlon;

    # Convert lats/lons:
    my $i;
    for $i (0..3) {
	$self->{Points}[$i] = new IU::Geo::Point $lats[$i], $lons[$i];
    }
    
    # Done!
    $self;
}

#------------------------------------------------------------
# to_ODL 
#------------------------------------------------------------
# I<Instance method.>  Return as an ODL group, identical in structure
# to a "RANGE_LOC" group.

sub to_ODL {
    my $self = shift;

    my $odl = IU::ODL->new_group;
    if ($self->pole ne 'X') {
	$odl->add_param(POLE_INCLUDED => IU::ODL->symbol($self->pole));
    }

    # Tangent:
    $odl->add_param(TANGENT_LATITUDE  => IU::ODL->real($self->tangent->lat));
    $odl->add_param(TANGENT_LONGITUDE => IU::ODL->real($self->tangent->lon));
    my @pts = @{$self->points};

    # Projection:
    if ($self->projection) {
	$odl->add_param(MAP_PROJECTION_TYPE => 
		      IU::ODL->string($self->projection));
    }

    # Lats:
    my @lat_odls = map {IU::ODL->real($_->lat)} @pts;
    $odl->add_param(LATITUDE => IU::ODL->sequence(@lat_odls));

    # Lons:
    my @lon_odls = map {IU::ODL->real($_->lon)} @pts;
    $odl->add_param(LONGITUDE => IU::ODL->sequence(@lon_odls));


    #------------------------------
    # Web-gateway specific...
    
    if ($self->zoom) {
	$odl->add_param(WG_ZOOM => IU::ODL->int($self->zoom));
    }
    if ($self->feature) {
        $odl->add_param(WG_FEATURE =>
                      IU::ODL->string($self->feature));
    }


    # Done!
    $odl;
}

#------------------------------------------------------------
# ODL_type
#------------------------------------------------------------
sub ODL_type { "POLYGON_LOC" }

#------------------------------------------------------------
1;
