package IU::Geo::Range;

=head1 NAME

IU::Geo::Range - a rectangular search region, like a RANGE_LOC 

=head1 SYNOPSIS

    my $range = new IU::Geo::Range;
    $range->north(90.0);           # set north latitude
    $range->south(-90.0);          # set south latitude
    $range->east(-180.0);          # set east longitude
    $range->west(-180.0);          # set west longitude

=head1 INSTANCE VARIABLES

    NorthLat => latitude of northern edge, in degrees (-90.0 to 90.0)
    SouthLat => latitude of southern edge, in degrees (-90.0 to 90.0)
    EastLong => longitude of eastern edge, in degrees (-180.0 to 180.0)
    WestLong => longitude of western edge, in degrees (-180.0 to 180.0)

=head1 AUTHOR

Eryq (Erik Dorfman), Hughes STX Corporation

=head1 VERSION

$Id: Range.pm,v 1.6 2001/02/27 22:03:05 seufert Exp $

=cut

use IU::Coord;
use IU::IW;
use IU::Geo::Base;
use IU::Log;

use strict;
use vars qw(@ISA);
@ISA = qw(IU::Geo::Base);


# Private functions:
sub _MIN { $_[0] < $_[1] ? $_[0] : $_[1] }
sub _MAX { $_[0] > $_[1] ? $_[0] : $_[1] } 


#------------------------------------------------------------
# new [NORTH,WEST,SOUTH,EAST]
#------------------------------------------------------------
# I<Class method, constructor.>
#
sub new {        
    my $class = shift;
    my $self = {};
    $self->{NorthLat} = shift;
    $self->{WestLong} = shift;
    $self->{SouthLat} = shift;
    $self->{EastLong} = shift;
    bless $self, $class;	
}

#------------------------------------------------------------
# _one VARNAME,SELF,[VALUE]
#------------------------------------------------------------
# I<Private utility.>  Get/set a single-valued instance variable.
sub _one {
    my ($varname, $self, $value) = @_;
    $self->{$varname} = $value if (@_ > 2);
    $self->{$varname};   
}

# north [COORD]      --  get/set the north latitude
sub north { _one('NorthLat', @_) }

# south [COORD]      --  get/set the south latitude
sub south { _one('SouthLat', @_) }

# east [COORD]       --  get/set the east longitude
sub east  { _one('EastLong', @_) }

# west [COORD]       --  get/set the west longitude
sub west  { _one('WestLong', @_) }

# feature [STRING]   -- get/set the feature name associated with
# the lat/lon point e.g. "Atlanta GA". Optional

sub feature { _one('Feature', @_) }


#------------------------------------------------------------
# as_string PARAMHASH
#------------------------------------------------------------
# I<Instance method, override.>  
# Return the rectangle as a string, of the form:
#
#    "30.23^ to 34^ Lat, -12.1^ to -45^ Lon"
#
# Each "^" is a Latin-1 degree mark, "\xB0".
# If Verbose, use newline instead of comma.

sub as_string {
    my ($self, %params) = @_;
    my @lines;  
    if ($params{Verbose} > 1) {
        push @lines, "Feature Name: ".$self->feature if $self->feature;
    }
    push @lines, $self->north . "\xB0 to " . $self->south . "\xB0 Lat";
    push @lines, $self->west  . "\xB0 to " . $self->east  . "\xB0 Lon";
    join(($params{Verbose} ? "\n" : ", "), @lines);
}

#------------------------------------------------------------
# fixup 
#------------------------------------------------------------
# I<Instance method, override.>  
# Given a rectangle, fix it so that...
#      
#   * Everything has a numeric value, defaults (-180, 90, 180, -90).
#   * West is wester than east, north is norther than south.
#   * Everything is clipped to (-180, 90, 180, -90).
#   * All numbers are returned as floats, with $precision-digit precision.
#
# Return the self object on success.

sub fixup {
    my $self = shift;
    my $prec = shift;        # precision, in digits
    defined($prec) || ($prec = 4);

    # For quicker manipulation:
    my $northLat = $self->{NorthLat};
    my $southLat = $self->{SouthLat};
    my $eastLong = $self->{EastLong};
    my $westLong = $self->{WestLong};



    # Force all to be numbers:
    (defined($northLat) && $northLat =~ /\d/) || ($northLat =  90.0);
    (defined($southLat) && $southLat =~ /\d/) || ($southLat = -90.0); 
    (defined($eastLong) && $eastLong =~ /\d/) || ($eastLong =  180.0); 
    (defined($westLong) && $westLong =~ /\d/) || ($westLong = -180.0); 



    # East is east, and west is west:
    my $tmp;
    if ($northLat < $southLat) {
	$tmp = $northLat; $northLat = $southLat; $southLat = $tmp;
    }

#   Allow ranges to cross the dateline.
#
#    if ($eastLong < $westLong) {
#	$tmp = $eastLong; $eastLong = $westLong; $westLong = $tmp;
#    }

    # Clip:
    $northLat = _MIN(_MAX($northLat, -90.0), 90.0);
    $southLat = _MIN(_MAX($southLat, -90.0), 90.0);
    $eastLong = _MIN(_MAX($eastLong, -180.0), 180.0);
    $westLong = _MIN(_MAX($westLong, -180.0), 180.0);

    # Adjust and save:
    $self->{WestLong} = sprintf("%0.${prec}f", $westLong);
    $self->{EastLong} = sprintf("%0.${prec}f", $eastLong);
    $self->{NorthLat} = sprintf("%0.${prec}f", $northLat);
    $self->{SouthLat} = sprintf("%0.${prec}f", $southLat);

    # Done!
    $self;
}



#------------------------------
#
# CGI interface
#
#------------------------------

#------------------------------------------------------------
# from_CGI PREFIX
#------------------------------------------------------------
# I<Constructor.>  Read from CGI tags.
# The PREFIX is taken to precede the tag names.

sub from_CGI {
    my $class = shift;
    my $prefix = shift;    # followed by standard rectangle element names

    my $new = $class->new;
    $new->{NorthLat} = str2deg($REQ->param("${prefix}north"));
    $new->{SouthLat} = str2deg($REQ->param("${prefix}south"));
    $new->{EastLong} = str2deg($REQ->param("${prefix}east"));
    $new->{WestLong} = str2deg($REQ->param("${prefix}west"));

    # Optional:
    $new->feature($REQ->param("${prefix}feature") || undef);


    # For quicker manipulation:
    my $northLat = $new->{NorthLat};
    my $southLat = $new->{SouthLat};
    my $eastLong = $new->{EastLong};
    my $westLong = $new->{WestLong};


    # if nothing was typed in, return undefined for geo object.

    if ( (!defined($northLat) || $northLat !~ /\d/) &&
         (!defined($southLat) || $southLat !~ /\d/) &&
         (!defined($eastLong) || $eastLong !~ /\d/) &&
         (!defined($westLong) || $westLong !~ /\d/) ) {

         return (undef);

    } else {

         return ($new);
    }
}

#------------------------------------------------------------
# list_CGI PREFIX
#------------------------------------------------------------
# I<Instance method.>
# Return an array of CGI [tag=>value] pairs which can be used to init 
# this object.  The PREFIX precedes all tags, as it does in from_CGI.
#
sub list_CGI {
    my ($self, $p) = @_;
    $p ||= '';
    my @tags;
    push @tags, ["${p}north" => $self->north];
    push @tags, ["${p}south" => $self->south];
    push @tags, ["${p}east"  => $self->east];
    push @tags, ["${p}west"  => $self->west];
    push @tags, ["${p}feature" => $self->feature] if $self->feature;

    return @tags;
}


#------------------------------
#
# ODL interface
#
#------------------------------

#------------------------------------------------------------
# from_ODL ODL
#------------------------------------------------------------
# I<Constructor.>  Read from ODL group with the same structure
# as a "RANGE_LOC" group.

sub from_ODL {
    my $class = shift;
    my $odl = shift;    # followed by standard rectangle element names

    my $self = $class->new;
    my $value;
    if ($value = $odl->param('NORTH_LATITUDE')) {
	$self->{NorthLat} = $value->text_value;
    }
    if ($value = $odl->param('SOUTH_LATITUDE')) {
	$self->{SouthLat} = $value->text_value;
    }
    if ($value = $odl->param('EAST_LONGITUDE')) {
	$self->{EastLong} = $value->text_value;
    }
    if ($value = $odl->param('WEST_LONGITUDE')) {
	$self->{WestLong} = $value->text_value;
    }
    if ($value = $odl->param('WG_FEATURE')) {
        $self->feature($value->text_value);
    }

    $self;
}

#------------------------------------------------------------
# to_ODL 
#------------------------------------------------------------
# I<Instance method.>  Return as an ODL group, identical in structure
# to a "RANGE_LOC" group.

sub to_ODL {
    my $self = shift;

    my $odl = IU::ODL->new_group;
    $odl->add_param(NORTH_LATITUDE => IU::ODL->number($self->north));
    $odl->add_param(SOUTH_LATITUDE => IU::ODL->number($self->south));
    $odl->add_param(WEST_LONGITUDE => IU::ODL->number($self->west));
    $odl->add_param(EAST_LONGITUDE => IU::ODL->number($self->east));
    if ($self->feature) {
        $odl->add_param(WG_FEATURE =>
                      IU::ODL->string($self->feature));
    }

    $odl;
}

#------------------------------------------------------------
# ODL_type
#------------------------------------------------------------
sub ODL_type { "RANGE_LOC" }

#------------------------------------------------------------
1;

