package IU::Geo::WRS;


=head1 NAME

IU::Geo::WRS - interface to the World Reference System mapping mechanisms


=head1 SYNOPSIS

    # Get a list of all supported schemes:
    @wrs_schemes = IU::Geo::WRS->schemes;
    
    # Get a WRS interface object for a given scheme, like "WRS-1":
    $wrs = new IU::Geo::WRS "WRS-1";
    
    # Get the bounding box for a single path/row, as an IU::Geo::Range:
    $range = $wrs->bound_rect($path, $row);
     
    # Get the bounding box for a range of path/rows, as an IU::Geo::Range:
    $range = $wrs->bound_rect([$minpath, $maxpath], [$minrow, $maxrow]);


=head1 DESCRIPTION

This interface class handles the mapping from WRS path-row coordinates
of a given scheme, like "WRS-1".

In order for a scheme XYZ to be supported, an executable program
named "imspathrow_XYZ" must reside in TOP/lib/bin.  This program is 
expected to be invoked as:

    imspathrow_XYZ -p 15 -r 20             # path 15, row 20

and is expected to output a chunk to its STDOUT, minimally containing:

    POINT: lon1 lat1
    POINT: lon2 lat2
    POINT: lon3 lat3
    POINT: lon4 lat4

If a range of path-rows are given to be mapped, only the four
endpoints of those ranges are mapped, and the bounding box is
returned around that:

    +---------------+     So if path = (33..42) and row = (5..10),
    |   [B]   .     |     then we find the 4 boxes...
    |   .    .   [D]|    
    |  .    .    .  |         A  =  boundbox(path=33, row=5)
    | .    .    .   |         B  =  boundbox(path=33, row=10)
    |[A]  .    .    |         C  =  boundbox(path=42, row=5)   
    |        [C]    |         D  =  boundbox(path=42, row=10)
    +---------------+
            Z             and then find the bounding box Z of THOSE boxes.



=head1 AUTHOR

Eryq, Hughes STX Corporation


=head1 RCS INFO

$Id: WRS.pm,v 1.8 2003/02/19 15:48:17 seufert Exp $

=cut

use IU;
use IU::Chunk;
use IU::ODL;
use IU::Log;
use IU::Geo::Range;
use IU::Geo::Polygon;

use strict;


#------------------------------------------------------------
# min A,B
# max A,B
#------------------------------------------------------------
# I<Utilities.>  Guess what these do.  :-)

sub min { ($_[0] < $_[1]) ? $_[0] : $_[1] }
sub max { ($_[0] > $_[1]) ? $_[0] : $_[1] }

#------------------------------------------------------------
# schemes
#------------------------------------------------------------
# I<Class/instance method.>
# Return a list of all supported schemes, or empty if none.
# The list is pre-sorted.

sub schemes {
    my $self = shift;


    my @schemes = ('WRS-1','WRS-2');
    return @schemes;
}

#------------------------------------------------------------
# new SCHEME
#------------------------------------------------------------
# I<Constructor.>
# Raises a fatal exception if the given SCHEME is not supported.

sub new {
    my $class = shift;
    my $scheme = shift || puke("missing scheme name");

    # Verify scheme:
    my $path = "${IU::TopDir}/lib/bin/imspathrow_$scheme";
    (-e $path) or puke "WRS scheme '$scheme' not supported: missing $path";
    
    # Build object:
    my $self = {
	Scheme=>$scheme,
	Path=>$path,
    };
    bless $self, $class;
}

#------------------------------------------------------------
# _simple_rect PATH,ROW
#------------------------------------------------------------
# I<Instance method.>
# Return the IU::Geo::Range that bounds the given PATH and ROW, or
# undef on error.  Both PATH and ROW must be integers.

sub _simple_rect {
    my $self = shift;
    my ($path, $row) = @_;
    my $result;

    # Run program, get output:
    $path = int($path);     # untaint!!!
    $row = int($row);       # untaint!!! 
    inform "Invoking $self->{Scheme} converter with path=$path, row=$row";
    open RESULT, "$self->{Path} -p $path -r $row |" or
	puke 453, "couldn't run $self->{Scheme} converter $self->{Path}: $!";
    my $odl = parse IU::ODL \*RESULT;
    close RESULT;
    $odl or puke 453, "couldn't get $self->{Path} results";
    
    # Get info:
    if ($odl->param('PATH_ROW_USAGE')) {
	my @paths = $odl->param('PATH_ROW_USAGE')->param('PATH')->text_values;
	my @rows  = $odl->param('PATH_ROW_USAGE')->param('ROW')->text_values;
	puke 513, "Values out of range: ",
	    "paths = ", join(" - ", @paths), ", rows = ",  join(" - ", @rows);
    }
    my $polygon_odl = $odl->param('POLYGON_LOC') ||
	puke 453, "couldn't get $self->{Path} results";

    # Create polygon object:
    $polygon_odl->param("TANGENT_LATITUDE" => IU::ODL->real("0.0"));
    $polygon_odl->param("TANGENT_LONGITUDE" => IU::ODL->real("0.0"));
    my $polygon = IU::Geo::Polygon->from_ODL($polygon_odl) ||
	die "couldn't create polygon from ODL!";
    
    
    # Get 4 points:
    my @points = @{$polygon->points};
    (int(@points) == 4) or return error "expected 4 points; got ",int(@points);
    
    # Get min and max lat and long:
    my ($n, $s, $e, $w);
    my $p;
    foreach $p (@points) {
	my ($lon, $lat) = ($p->lon, $p->lat);

	if (!defined($n)) {     # first time through...
	    $n = $s = $lat; 
	    $e = $w = $lon;
	}
	else {                  # need to min/mix with existing coords...
	    $n = max($n, $lat);
	    $w = min($w, $lon);
	    $s = min($s, $lat);
	    $e = max($e, $lon);
	}
        debug "lat = $lat, lon = $lon";
    }
    # The following code handles a path-row box that crosses the date-line.
    my $temp;
    if($w < -170.0 and $e > 170.0) {
	$temp = $w;
	$w = $e;
	$e = $temp;
    }

    debug "w = $w, e = $e";

    # Now we have the bounding rectangle... build the range:
    my $gr = IU::Geo::Range->new($n, $w, $s, $e);
# Removed the fixup for now
#    $gr->fixup(6);
    return $gr;
}

#------------------------------------------------------------
# _simple_polygon PATH,ROW
#------------------------------------------------------------
# I<Instance method.>
# Return the IU::Geo::Range that bounds the given PATH and ROW, or
# undef on error.  Both PATH and ROW must be integers.

sub _simple_polygon {
    my $self = shift;
    my ($path, $row) = @_;
    my $result;

    # Run program, get output:
    $path = int($path);     # untaint!!!
    $row = int($row);       # untaint!!!
    inform "Invoking $self->{Scheme} converter with path=$path, row=$row";
    open RESULT, "$self->{Path} -p $path -r $row |" or
        puke 453, "couldn't run $self->{Scheme} converter $self->{Path}: $!";
    my $odl = parse IU::ODL \*RESULT;
    close RESULT;
    $odl or puke 453, "couldn't get $self->{Path} results";
    # Get info:
    if ($odl->param('PATH_ROW_USAGE')) {
        my @paths = $odl->param('PATH_ROW_USAGE')->param('PATH')->text_values;
        my @rows  = $odl->param('PATH_ROW_USAGE')->param('ROW')->text_values;
        puke 513, "Values out of range: ",
            "paths = ", join(" - ", @paths), ", rows = ",  join(" - ", @rows);
    }
    my $polygon_odl = $odl->param('POLYGON_LOC') ||
        puke 453, "couldn't get $self->{Path} results";

    # Create polygon object:
    $polygon_odl->param("TANGENT_LATITUDE" => IU::ODL->real("0.0"));
    $polygon_odl->param("TANGENT_LONGITUDE" => IU::ODL->real("0.0"));
    my $polygon = IU::Geo::Polygon->from_ODL($polygon_odl) ||
        die "couldn't create polygon from ODL!";


    # Get 4 points:
    my @points = @{$polygon->points};
    (int(@points) == 4) or return error "expected 4 points; got ",int(@points);
    return @points;
}

#------------------------------------------------------------
# bound_rect PATH,ROW
#------------------------------------------------------------
# I<Instance method.>
# Return the IU::Geo::Range that bounds the given PATH and ROW, or
# undef on error.
#
# The PATH may be a scalar number, or a range as an arrayref [min, max].
# Ditto for the ROW.

sub bound_rect {
    my ($self, $path, $row) = @_;
    (defined($path) and defined($row)) or puke "undefined path or row!";

    # Turn everything into a range, for consistency:
    my @paths = (ref($path) ? @$path : ($path));
    my @rows  = (ref($row) ? @$row : ($row));
    
    # Collapse degenerate ranges:
    shift @paths if (@paths > 1 and $paths[0] == $paths[1]);
    shift @rows  if (@rows  > 1 and $rows[0] == $rows[1]);
    # Handle simple case, where both are single integers:
    if ((int(@paths) == 1) and (int(@rows) == 1)) {
	return $self->_simple_rect($paths[0], $rows[0]);
    }
    
    # Handle multiple paths and/or rows:
    my ($p, $r, $rect, @rects);
    my $southRow = 122;
    my $northRow = 246;
    my $deltaSouth = 0;
    my $deltaNorth = 0;
    foreach $p (@paths) {
	foreach $r (@rows) {
	    # Map rectangle and save it:
	    ($rect = $self->_simple_rect($p, $r)) 
		or return error "couldn't map $p,$r";
	    push @rects, $rect;
	}
	# The following 4 lines checks for Row 122 the farthest south possible
	if (defined $rows[1] and $rows[0] < $southRow and $rows[1] > $southRow){
	    puke 513, "Values out of range: Rows must be either descending,",
	    " 1-122 or ascending 122 - 248";
	}
	# The following 7 lines checks for Row 246 the farthest north possible
	if (defined $rows[1] and $rows[0] < $northRow and $rows[1] > $northRow){
	    # Map rectangle and save it:
	    ($rect = $self->_simple_rect($p, $northRow)) 
		or return error "couldn't map $p,$northRow";
	    $deltaNorth = 1;
	    push @rects, $rect;
	}
    }

    # Merge results:
    # $i represents the last path-row not including the ones added for
    # southRow and northRow 
    # We are using the last path-row to represent the farthest west the
    # box should be.
    my $i = int(@rects) - $deltaSouth - $deltaNorth - 2;
    my $j = 0;
    debug "i = $i";
    # We are using the first path-row for the eastern most point
    $rect->east($rects[0]->east);
    for ($rect = shift(@rects);
	 @rects;
	 shift(@rects)) {
	$rect->south(min($rect->south, $rects[0]->south));
	$rect->north(max($rect->north, $rects[0]->north));
	debug "i = $i, j = $j";
	if($i == $j) {
	    $rect->west($rects[0]->west);
	}
	++$j;
    }
    if(($rect->west < $rect->east) 
     and (($rect->east - $rect->west) > 180.0)) {
        puke 513, "Values out of range: Extent must",
         " be < 180.0 degress longitude. Narrow the path or row range.";
    }
    if(($rect->west > $rect->east) 
     and (($rect->east - $rect->west + 360.0) > 180.0)) {
        puke 513, "Values out of range: Extent must",
         " be < 180.0 degress longitude. Narrow the path or row range.";
    }
    if (defined $paths[1] and (($paths[1] - $paths[0]) > 88)) {
        puke 513, "Values out of range: Extent must",
         " be < 180.0 degress longitude. Narrow the path range.";
    }
    return $rect;
}

#------------------------------------------------------------
1;
