package IU::Granule;

=head1 NAME

IU::Granule -- granules returned from the last search


=head1 AUTHOR

Eryq (Erik Dorfman), Hughes STX Corporation


=head1 INSTANCE VARIABLES

The basic ones are:

    ArchiveId  => the archive id (e.g., GSFC)
    BrowseDesc => the browse description
    BrowseType => the browse type, 'Y', 'N', or 'FTP_ONLY'
    Campaign   => the campaign
    Comment    => the comment
    Coverage   => the spatial coverage
    DatasetId  => the dataset id
    DayNight   => the day/night flag
    Id         => the granule id
    ImageId    => the image id
    MDEntry    => the MD entry id
    PackageIds => array of package ids
    Parameters => array of parameter names
    ProcLevel  => the processing level
    Sensors    => array of sensor names
    Sources    => array of source names
    StartDate  => the start date, as an ODL date string
    StopDate   => the stop date, as an ODL date string

The URL-oriented ones are:

    DATA_URL               => array of IU::FancyURLs for any on-line data
    MISC_URL               => array of IU::FancyURLs for misc links
    BROWSE_URL             => array of IU::FancyURLs for browse links
    METADATA_URL           => array of IU::FancyURLs for metadata links
    GUIDE_URL              => array of IU::FancyURLs for ECS to pass Guide
                               documents to the EDG

These are special:

    Carted     => the order status
    Key        => the granule key, archtag/dstag/gid
    PkgCombos  => the package combos currently selected for this granule

The Key, ReadFrom, Carted, and PkgCombo are special; all others come
directly from the search results.


=head1 RCS INFO

$Id: Granule.pm,v 1.65 2003/05/21 20:29:01 solomon Exp $

=cut

require GDBM_File;
require IU::Storable;

use IU;
use IU::Coverage;
use IU::Dataset;
use IU::Log;
use IU::FancyURL;
use IU::SpecializedResults;
use IU::ResultGroup;
use IU::Util ();
use IU::Interface::Orderable;
use strict;

@IU::Granule::ISA = qw(IU::Storable IU::Interface::Orderable);

#------------------------------------------------------------
#
# Globals...
#
#------------------------------------------------------------

# URL instance vars:
my @URLInstVars = qw(DATA_URL MISC_URL BROWSE_URL GUIDE_URL METADATA_URL);

#------------------------------------------------------------
#
# Methods...
#
#------------------------------------------------------------

# To make Perl happy...
DESTROY { 1 }

#------------------------------------------------------------
# input FILEHANDLE
#------------------------------------------------------------
# Read a granule chunk from FILEHANDLE into the internal data structure.
#
# The granule object for success, false for failure (incluing EOF).

sub input {
    my $class = shift;

    # Create new object, initiallizing from chunk:
    $class->from_chunk(IU::Chunk->input(@_));
}

# Format for time fields is:  "yyyy-mm-ddThh:mm:ssZ" (Z is optional)
my $DEFAULT_START_TIME = '1990-01-01T00:00:00Z';  # start of the 90's
#my $DEFAULT_STOP_TIME = '2000-12-31T23:59:59Z';   # end of millenium
my $DEFAULT_STOP_TIME  = '2038-01-18T22:14:07Z';  # max possible date (08/1998)

#------------------------------------------------------------
# from_chunk CHUNK
#------------------------------------------------------------
# Read a granule from a chunk.
# A granule chunk containing "x-error" is considered null and void.
#
# The granule object for success, false for failure (incluing EOF).

sub from_chunk {
    my ($class, $chunk) = @_;
    # my $self;
    my $errmsg;

    # Input chunk, and check for errors:
    defined($chunk) or return undef;
    defined($errmsg = $chunk->value('x-error')) 
      and return error 454, $errmsg;

    # Create new object, initiallizing from chunk:
    my $self = {};
    # Get the simple-valued attributes:
    $self->{ArchiveId}   = $chunk->value('DATA_CENTER_ID');
    $self->{BrowseDesc}  = $chunk->value('BROWSE_PRODUCT_DESCRIPTION');
    my $type = $chunk->value('BROWSE_TYPE');
    my $only = $chunk->value('INTEGRATED_BROWSE_ONLY');
    if(defined($type) && defined($only) && $type eq 'Y' && $only eq 'Y') {
        $type = 'INTEGRATED';
    }
    $self->{BrowseType}  = $type;
    $self->{Campaign}    = $chunk->value('CAMPAIGN');
    $self->{Comment}     = $chunk->value('COMMENT');
    $self->{DatasetId}   = $chunk->value('DATASET_ID');
    $self->{DayNight}    = $chunk->value('DAY_NIGHT');
    $self->{Id}          = $chunk->value('GRANULE_ID');
    $self->{ImageId}     = $chunk->value('IMAGE_ID');
    $self->{Key}         = $chunk->value('GRANULE_WEB_KEY');
    $self->{MDEntry}     = $chunk->value('MD_ENTRY_ID');
    $self->{ProcLevel}   = $chunk->value('PROCESSING_LEVEL');

    my $parsestr;                        # string of date to parse.
    $parsestr = $chunk->value('START_DATE');
    if (!defined($parsestr)) {
        $self->{StartDate} = '(none)';
    } elsif ($parsestr =~ /\d+-\d+-\d+T\d+:\d+:\d+/) {
        $self->{StartDate} = $parsestr;
    } else {
        &IU::Util::date_pad(\$parsestr, $DEFAULT_START_TIME);
        $self->{StartDate} = $parsestr;
    }
    $parsestr = $chunk->value('STOP_DATE');
    if (!defined($parsestr)) {
       $self->{StopDate} = '(none)';
    } elsif ($parsestr =~ /\d+-\d+-\d+T\d+:\d+:\d+/) {
        $self->{StopDate} = $parsestr;
    } else {
        &IU::Util::date_pad(\$parsestr, $DEFAULT_STOP_TIME);
        $self->{StopDate} = $parsestr;
    }
    # Get multi-valued attributes:
    $self->{PackageIds}    = [$chunk->seq_values('VALID_PACKAGE_ID')];
    $self->{RawPackageIds} = [$chunk->seq_values('PACKAGE_ID')];
    $self->{Parameters}    = [$chunk->seq_values('PARAMETER')];
    $self->{Sensors}       = [$chunk->seq_values('SENSOR_NAME')];
    $self->{Sources}       = [$chunk->seq_values('SOURCE_NAME')];

    # Coverage:
    $self->{Coverage} = IU::Coverage->from_chunk($chunk);
    my $name;
    my $value;
    my $results_chunk;
    my $granuleId;
    my $nodeType;
    my $numberOfScenes;
    my $numberOfPartials;
    my $newvalue;
    my @listParentDsetId;

    my ($qf_operational, $qf_automatic , $qf_science);
 
    foreach $results_chunk ($chunk->values('SPECIALIZED_RESULTS')) {
        $name = $results_chunk->value('RESULT_NAME');
        $value  = $results_chunk->value('RESULT_VALUE');
        my $result = IU::SpecializedResults->from_chunk($results_chunk);
        if($name eq 'ParentGranCollectionName'){
           $self->{BrowseType}  = 'INTEGRATED';
           my $arrayRef = $result->resultValue;
           my @DsetArray;
           if (ref($arrayRef)) {
              @DsetArray = @$arrayRef;
           } else {
              $DsetArray[0] = $arrayRef;
           }
           push @listParentDsetId, @DsetArray;
           $self->{ParentDatasetId} = \@listParentDsetId if(defined($arrayRef));
        }

        #Why isn't this a HASH? Instead of a looking for each of these values
        #individually? Build a hash with the ODL value we're looking for and the
        # stored value name??? JSA

        $newvalue =  $result->check('QAPercentCloudCover') 
          if(defined($result->check('QAPercentCloudCover'))) ;
        $newvalue = $result->check('QA_SCENE_CCA') 
          if(defined($result->check('QA_SCENE_CCA')));
        $self->{CloudCover} = $newvalue if(defined($newvalue));

        $granuleId = $result->check('LocalGranuleID');
        $self->{LocalGranuleId} = $granuleId if($granuleId);

        my $granuleSize = $result->check('SizeMBECSDataGranule');
        $self->{SizeEcsGranule} = $granuleSize if($granuleSize);

        $nodeType = $result->check('NodeType');
        $self->{NodeType} = $nodeType if($nodeType);

        $numberOfScenes = $result->check('ORIG_TOTAL_WRS_SCENES');
        $self->{NumberOfScenes} = $numberOfScenes if($numberOfScenes);

        $numberOfPartials = $result->check('ORIG_PARTIAL_WRS_SCENES');
        $self->{NumberOfPartials} = $numberOfPartials if($numberOfPartials);

        push @{$self->{Results}}, $result;
    }
    my $result_group_chunk;
    my (@listOperational, @listScience, @listAutomatic);

    my $array_ref;
    foreach $result_group_chunk ($chunk->values('RESULT_GROUP')) {

       my $resultGroup = new IU::ResultGroup;
       $resultGroup->from_chunk($result_group_chunk);
       my $result = $resultGroup->findSpecializedResults('LocalGranuleID');
       if(@$result) {
        $self->{LocalGranuleId} = $result->[0]->resultValueAsString;
       }
       my $size = $resultGroup->findSpecializedResults('SizeMBECSDataGranule');
       if(@$size) {
          $self->{SizeEcsGranule} = $size->[0]->resultValueAsString;
       }
       $result = $resultGroup->findSpecializedResults('NodeType');
       if(@$result) {
        $self->{NodeType} = $result->[0]->resultValueAsString;
       }
       $array_ref = 
         $resultGroup->findSpecializedResults('QAPercentCloudCover');
       if(@$array_ref) {
          $newvalue = $array_ref->[0]->resultValueAsString;
          $self->{CloudCover} = $newvalue if(defined($newvalue));
       }
       $array_ref = 
          $resultGroup->findSpecializedResults('QA_SCENE_CCA');
       if(@$array_ref) {
          $newvalue = $array_ref->[0]->resultValueAsString;
          $self->{CloudCover} = $newvalue if(defined($newvalue));
       }
       # Find the 3 Quality Flags : Operational, Science, and Automatic.
       #  

       my $operational = 
          ($resultGroup->findSpecializedResults('^OperationalQualityFlag$') 
          || undef);
       my $op_value;
       if (defined ($operational)) {
          foreach $op_value (@$operational) {
             push @listOperational , $op_value->resultValueAsString;
          }
       }
       $self->{OperationalQualityFlag} = \@listOperational; 

       my $automatic = 
          ($resultGroup->findSpecializedResults('^AutomaticQualityFlag$') 
          || undef);
       if (defined ($automatic)) {
          foreach $op_value (@$automatic) {
             push @listAutomatic , $op_value->resultValueAsString; 
          }
       }
       $self->{AutomaticQualityFlag} = \@listAutomatic;

       my $science = 
          ($resultGroup->findSpecializedResults('^ScienceQualityFlag$') 
          || undef);
       if (defined ($science)) {
          foreach $op_value (@$science) {
             push @listScience , $op_value->resultValueAsString;    
          }
       }
       $self->{ScienceQualityFlag} = \@listScience;

       my $parentDsetIdRef = 
        ($resultGroup->findSpecializedResults('^ParentGranCollectionName$') ||
         undef);
       my @parentDsetIdArray = @$parentDsetIdRef if defined($parentDsetIdRef);
       my $parentDsetId = $parentDsetIdArray[0] if defined($parentDsetIdRef); 
       if (defined($parentDsetId)) {
           $self->{BrowseType}  = 'INTEGRATED';
           foreach $op_value (@parentDsetIdArray) {
             my $value = $op_value->resultValue;
             my @DsetArray;
             if (ref($value)) {
                @DsetArray = @$value;
             } else {
                $DsetArray[0] = $value;
             }
             push @listParentDsetId, @DsetArray;
             $self->{ParentDatasetId} = \@listParentDsetId if(defined($value));
           }
       }

       push @{$self->{ResultGroup}}, $resultGroup;
   }

   # URLs (we tolerate more than one of each type):
   my $key;
   foreach $key (@URLInstVars) {

      # Allow many URLs of the current flavor...
      $self->{$key} = [];
      my ($url_chunk, $furl);
      foreach $url_chunk ($chunk->values($key)) {
          my $furl = IU::FancyURL->from_chunk($url_chunk, $key);
          push @{$self->{$key}}, $furl;
      }
   }    
   foreach $results_chunk ($chunk->values('DIRECT_ACCESS')) {
      # URLs (we tolerate more than one of each type):
      my $key;
      foreach $key (@URLInstVars) {

         # Allow many URLs of the current flavor...
         my ($url_chunk, $furl);
         foreach $url_chunk ($results_chunk->values($key)) {
            my $furl = IU::FancyURL->from_chunk($url_chunk, $key);
            push @{$self->{$key}}, $furl;
         }
      }    
   }    
   # Re-bless and return:
   bless($self, $class);
}

#------------------------------------------------------------
#
# Access methods...
#
#------------------------------------------------------------

#----------
#
# Single-valued...
#
#----------

sub mapped { undef }

# archive_id  -- return the archive id
sub archive_id  { $_[0]->{ArchiveId}; }

# One of the Quality Flags ( Science,Operational,Automatic)
# Used to determine whether a Quality Flag Disclaimer should be shown
sub automatic_quality_flag   { $_[0]->{AutomaticQualityFlag}; }

# browse_desc -- what is the granule's browse description?
sub browse_desc { $_[0]->{BrowseDesc}; }

# browse_type -- what is the granule's browse type? Y, N, or FTP_ONLY?
sub browse_type { uc($_[0]->{BrowseType} || 'N') }

# campaign   -- what is the granule's campaign?
sub campaign  { $_[0]->{Campaign}; }

# cloud cover   -- what is the granule's cloud cover?
sub cloud_cover  { $_[0]->{CloudCover}; }

# node type   -- what is the granule's node type?
sub node_type  { $_[0]->{NodeType}; }

# local granule ID   -- what is the granule's local granule ID?
sub local_granule_ID  { 
   my $self = shift;

   if ( defined($self->{LocalGranuleId})) {
      if ( ref($self->{LocalGranuleId}) eq 'ARRAY') {
            if (defined(${$self->{LocalGranuleId}}[0])) {
               ${$self->{LocalGranuleId}}[0]->resultValueAsString;
            }
      }
      elsif ( ref($self->{LocalGranuleId}) eq 'IU::SpecializedResults' ) {
          $self->{LocalGranuleId}->resultValueAsString;
      }
      else {
           ($self->{LocalGranuleId});
      }
   } 
 }

# Size_MBECS_Data_Granule   -- what is the granule's size?
sub size_mbecs_granule  {
   my $self = shift;

   if ( defined($self->{SizeEcsGranule})) {
      if ( ref($self->{SizeEcsGranule}) eq 'ARRAY') {
            if (defined(${$self->{SizeEcsGranule}}[0])) {
               ${$self->{SizeEcsGranule}}[0]->resultValueAsString;
            }
      }
      elsif ( ref($self->{SizeEcsGranule}) eq 'IU::SpecializedResults' ) {
          $self->{SizeEcsGranule}->resultValueAsString;
      }
      else {
           ($self->{SizeEcsGranule});
      }
   }
}

# number of scenes   -- what is the granule's number of scenes?
sub number_of_scenes  { $_[0]->{NumberOfScenes}; }

# number of partials -- what is the granule's number of partial scenes?
sub number_of_partials { $_[0]->{NumberOfPartials}; }

# comment    -- what is the granule's comment?
sub comment   { $_[0]->{Comment}; }

# coverage    -- return the coverage object
sub coverage    { $_[0]->{Coverage}; }

# dataset_id  -- return the dataset id
sub dataset_id  { $_[0]->{DatasetId}; }

# dataset_key -- return the dataset key
sub dataset_key { 
    ($_[0]->{Key} =~ m|^([^/]+)/([^/]+)|) || return undef;
    return "$1/$2";
}

# day_night   -- return the day/night flag
sub day_night  { $_[0]->{DayNight}; }

# id         -- return the granule id
sub id         { $_[0]->{Id} }
sub granule_id { $_[0]->{Id} }      ### to be "orderable" 

# image_id   -- return the image id
sub image_id   { $_[0]->{ImageId}; }

# key        -- return the granule key
sub key           { $_[0]->{Key} }
sub granule_key   { $_[0]->{Key} }  ### to be "orderable" 

# MD_entry     -- return the MD entry id
sub MD_entry      { 
    my $self = shift;
    $self->{MDEntry} ? IW::Util::trim($self->{MDEntry}) : undef;
}

# One of the Quality Flags ( Science,Operational,Automatic)
# Used to determine whether a Quality Flag Disclaimer should be shown
sub operational_quality_flag   {
    $_[0]->{OperationalQualityFlag}; 
}

sub parent_dset_id  { $_[0]->{ParentDatasetId}; }

# proc_level  -- what is the granule's processing level
sub proc_level  { $_[0]->{ProcLevel}; }


# One of the Quality Flags ( Science,Operational,Automatic)
# Used to determine whether a Quality Flag Disclaimer should be shown
sub science_quality_flag   {
    $_[0]->{ScienceQualityFlag};
}

# start_date  -- return the start date
sub start_date  { $_[0]->{StartDate}; }

# stop_date   -- return the stop date
sub stop_date   { $_[0]->{StopDate}; }

# carted      -- is the granule in the shopping cart?
sub carted    { 
    my ($self, $value) = @_;
    $self->{Carted} = $value if (@_ > 1);
    $self->{Carted};
}

# workbenched -- indicates granule is in the workbench.
sub workbenched {
    my ($self, $value) = @_;
    $self->{Workbenched} = $value if (@_ > 1);
    $self->{Workbenched};
}

# graphed     -- set if granule is displayed on temporal coverage graph. 
sub graphed {
    my ($self, $value) = @_;
    $self->{Graphed} = $value if (@_ > 1);
    $self->{Graphed};
}

#----------
#
# Multi-valued...
#
#----------

# parameters -- return a reference to the array of parameter names, or []
sub parameters { $_[0]->{Parameters} || []; }

# sensors    -- return a reference to the array of sensor names, or []
sub sensors    { $_[0]->{Sensors} || []; }

# sources    -- return a reference to the array of source names, or []
sub sources    { $_[0]->{Sources} || []; }

# data_URLs    -- return ref to array of data URLs (as IU::FancyURLs), or []
sub data_URLs    { $_[0]->{DATA_URL} }

# misc_URLs    -- return the to array of misc URLs (as IU::FancyURLs), or [] 
sub misc_URLs    { $_[0]->{MISC_URL} }

# metadata_URLs -- return the to array of metadata URLs (as IU::FancyURLs)
# or [] 
sub metadata_URLs    { $_[0]->{METADATA_URL} }

# browse_URLs    -- return the to array of browse URLs (as IU::FancyURLs), or []
sub browse_URLs    { $_[0]->{BROWSE_URL} }

# guide_URLs -- return the to array of guide URLs (as IU::FancyURLs), or []
# Just making sure it'll return [] is undef.
sub guide_URLs  { $_[0]->{GUIDE_URL} || [] }

# has_links     -- return whether or not there are external links
sub has_links {
    my $self = shift;
    @{$self->data_URLs} || @{$self->misc_URLs} || @{$self->browse_URLs} 
     || @{$self->guide_URLs} || @{$self->metadata_URLs};
} 

# has_browse     -- return whether or not there are Browse URLs
sub has_browse {
    my $self = shift;
    @{$self->browse_URLs};
} 

# has_data     -- return whether or not there are Data URLs
sub has_data {
    my $self = shift;
    @{$self->data_URLs};
} 

# has_metadata     -- return whether or not there are Metadata URLs
sub has_metadata {
    my $self = shift;
    @{$self->metadata_URLs};
} 

# Should the user see the QA disclaimer or not?
# Parse the ResultGroup and send the values of ScienceQualityFlag , Operational,
#  and Automatic flag to determine if the user should see the QA disclaimer or
# not ?
sub quality_assurance_status {
    my $self = shift;

         if ( (defined ($self->automatic_quality_flag) &&
                 (int(@{$self->automatic_quality_flag})) > 0) ||
               (defined ($self->operational_quality_flag) &&
                 (int(@{$self->operational_quality_flag})) > 0) ||
               (defined ($self->science_quality_flag) &&
                 (int(@{$self->science_quality_flag})) > 0)) {

            my $quality_flag_status = IW::Disclaimers->verdict_quality_flag($self->automatic_quality_flag, $self->operational_quality_flag, $self->science_quality_flag);

            return ( $quality_flag_status);
         }
         else { return 99 };
}

#------------------------------
# package_ids [VALUE]
#------------------------------
# Return a reference to the array of filtered package ids, or [].
# These are the packages that HAVE PACKAGING INFO.

sub package_ids { 
    $_[0]->{PackageIds} || []; 
}

#------------------------------
# unfiltered_package_ids [VALUE]
#------------------------------
# Return a reference to the array of unfiltered package ids, or [].
# These are *all* packages, WHETHER OR NOT THEY HAVE PACKAGING INFO.

sub unfiltered_package_ids { 
    $_[0]->{RawPackageIds} || []; 
}

# results -- return a reference to the array of specialized results, or []
sub results { $_[0]->{Results} || []; }

# resultGroup -- return a reference to the array of result groups, or []
sub resultGroup { $_[0]->{ResultGroup} || []; }

#------------------------------
#
# ORDERABLE INTERFACE...
#
#------------------------------
# These methods are necessary for IU::Interface::Orderable.

#------------------------------
# order_thingy
#------------------------------
sub order_thingy { "data granule" }

#------------------------------
# order_type
#------------------------------
sub order_type { "GRANULE" }

#------------------------------------------------------------
#
# INVENTORY-DEPENDENT ACTIVITIES...
#
#------------------------------------------------------------

#------------------------------
# inventory [INVSTORE]
#------------------------------
# B<DANGER! HIGHLY FRAGILE!>
#
# With argument, associates this granule with an inventory storage object, 
# returning the previous association.  With no arg, returns the current 
# association (or undef if none).
#
# WARNING: ALL ACCESSES TO AN INVENTORY SHOULD BE EVAL-PROTECTED!!!!

sub inventory {
    my ($self, $inv) = @_;
    my $oldinv = $self->{INV};
    $self->{INV} = $inv if @_ > 1;
    $oldinv;
}

#------------------------------
# dataset
#------------------------------
# B<DANGER! HIGHLY FRAGILE!>
# Return the granule's dataset, as extracted from the inventory.

sub dataset {
    my $self = shift;
    return undef if !$self->{INV};
    my $dset = eval { $self->{INV}->dataset($self->dataset_key) };
    $dset or error "couldn't get data set: $@: $!";
    $dset;
}

#------------------------------------------------------------
# options [OPTSTYPE, [OPTIONS]]
#------------------------------------------------------------
# Set/get the options named WHICH.  Possibilities for OPTSTYPE
# include "SUBSET", "ORDER", and "DATASET_ORDER".
#
# If no arguments are given, returns a list of OPTSTYPEs which
# have values in this object.
#
sub options {
    my ($self, $type, $opts) = @_;
    $type or return grep {$self->{Options}{$_}}(keys %{$self->{Options}});
    $self->{Options}{$type} = $opts if @_ > 2;
    $self->{Options}{$type};
}

#------------------------------------------------------------
# some_options [OPTSTYPE, [OPTIONS]]
#------------------------------------------------------------
# Set/get the options named WHICH.  Possibilities for OPTSTYPE
# include "SUBSET", "ORDER", and "DATASET_ORDER".
#
# If no arguments are given, returns a list of OPTSTYPEs which
# have values in this object.
#
sub some_options {
    my ($self, $type, $opts) = @_;
    $type or return grep {$self->{Options}{$_}}(keys %{$self->{Options}});
    $self->{SomeOptions}{$type} = $opts if @_ > 2;
    $self->{SomeOptions}{$type};
}

#------------------------------
# order_status
#------------------------------
# B<DANGER! HIGHLY FRAGILE!>
# Returns...
#
#    'NEED_PKGS'  if the granule just needs packages selected
#    'GO_PKGS'    if the granule is ready for a package order
#    'GO_EMAIL'   if the granule has no packages, but can be email-ordered
#    'GO_EMAIL_PKGS'   if it has packages selected and will be email-ordered
#     0           (false, defined) if the granule can't be ordered

sub order_status {
   my $self = shift;

   # Can this granule be email-ordered?  If we don't know yet,
   # check the dataset.
   unless (defined($self->{CanEmail})) {
      my $dset = $self->dataset or return 0;     # YOW!! NO INFO!!!!
      $self->{CanEmail} = $dset->can_email_order;
   }
   # If granule has packages:
   if (@{$self->package_ids}) {
      # If granule has packages selected (combos):
      if (int(@{$self->package_combos})) {
         my @comboarray = @{$self->package_combos};
         my $combo = $comboarray[0];
         my $order_opts = $self->options('ORDER');
         my $subset_opts = $self->options('SUBSET');
         my $subset_opts_required = $subset_opts->is_required if($subset_opts);
         if($subset_opts && $subset_opts_required) {
            my $subset_opts_given = $subset_opts->user_made_selections;
            return ('NEED_PKGS') if(!$subset_opts_given);
         }
         my $order_opts_required = $order_opts->is_required if($order_opts);
         if($order_opts && $order_opts_required) {
            my $order_opts_given = $order_opts->user_made_selections;
            return ('NEED_PKGS') if(!$order_opts_given);
         }
         # Ready to order by e-mail or imsclient.
         return ( $self->{CanEmail} ? 'GO_EMAIL_PKGS' : 'GO_PKGS');
      } else {
         # Else, user needs to select some packages.
         return ('NEED_PKGS');
      }
   } else {
      # Else, order by e-mail or failure.
      return ($self->{CanEmail} ? 'GO_EMAIL' : 0);
   }
}

#------------------------------
# possible_to_order
#------------------------------
# B<DANGER! HIGHLY FRAGILE!>
# Return whether or not the granule can conceivably be ordered.

sub possible_to_order {
    my $self = shift;    
    $self->order_status;
}

#------------------------------
# ready_for_order
#------------------------------
# B<DANGER! HIGHLY FRAGILE!>
# Return whether or not the granule can be ordered (by either email or
# by packages) AT THIS MOMENT.  Returns true for statuses:
# GO_PKG, GO_EMAIL, and GO_EMAIL_PKGS.

sub ready_for_order {
    my $self = shift;    
    my $status = $self->order_status;
    return ($status =~ /^GO/);
}

#------------------------------------------------------------
#
# DEPRECATED COVERAGE METHODS: use the coverage object directly!
#
#------------------------------------------------------------

# area             -- return the area
sub area { $_[0]->{Coverage}->area; }

# area_center      -- return the area center
sub area_center { $_[0]->{Coverage}->center; }

# area_center_HTML -- return the area center, pretty (KLUDGE!)
sub area_center_HTML { $_[0]->{Coverage}->center_as_HTML; }

# area_type        -- return the area type
sub area_type { $_[0]->{Coverage}->type; }

#------------------------------------------------------------
#
# UTILITIES...
#
#------------------------------------------------------------

#------------------------------------------------------------
# SplitGkey GKEY
#------------------------------------------------------------
# Split a granule key into its 3 components.

sub SplitGkey {
    ($_[0] =~ m|^([^/]+)/([^/]+)/(.+)$|) || return ();
    ($1, $2, $3);    # (archtag, dstag, gid)
}

#------------------------------------------------------------
# GkeyToDskey GKEY
#------------------------------------------------------------
# Extract the dskey from the GKEY.

sub GkeyToDskey {
    ($_[0] =~ m|^([^/]+)/([^/]+)|) or return undef;
    "$1/$2";    # (archtag, dstag, gid)
}

#------------------------------------------------------------
1;
