package IU::HTAccess;

=head1 NAME

IU::HTAccess - create a very simple .htaccess file

=head1 SYNOPSIS

    # Create a virtual file:
    $access = new IU::HTAccess;
    
    # Start a limit:
    $access->start_limit('GET', 'POST', 'PUT');    # that's the default, BTW
    
    # Set attributes for the current limit:
    $access->order('ALLOW', 'DENY');         # that's the default, BTW
    $access->allow_from('host.one', 'host.two');
    $access->deny_from('host.three', 'host.four');
    
    # End the current limit:
    $access->end_limit;
    
    # Set error documents for this .htaccess:
    $access->error_doc(404,"lame-excuse.html");
    $access->error_doc(500,'"Oops! one of our scripts died."');
    $access->error_doc(401,
                       IW::Util::script_url('mode'=>'FORGOT',
					    "submode"=>'MODAL');
    # Write the file out to disk:
    $access->write_in("/some/http/directory");

=head1 BUGS

Currently, only one "limit" is allowed.  But the interface
is built to allow more in the future.

=head1 AUTHOR

Eryq, Hughes STX Corporation

=head1 VERSION

$Id: HTAccess.pm,v 1.11 2001/01/11 20:14:04 joe Exp $

=cut

use IU;
use IU::Log;
use strict;

#------------------------------------------------------------
# new
#------------------------------------------------------------
# Create a new .htaccess object.  This does I<not> create the file yet!

sub new {
    my $class = shift;
    my $self = shift;
    $self->{Limit} = ['GET', 'POST', 'PUT'];
    $self->{Order} = ['ALLOW', 'DENY'];
    $self->{AuthName} = [];
    $self->{AuthType} = [];
    $self->{AuthUserFile} = [];

    $self->{'allow'} = [];
    $self->{'deny'} = [];
    $self->{'require_type'} = ['valid-user'];

    $self->{ErrorDoc} = {};   # Yes, this is a hash.

    bless $self, $class;
}    

#------------------------------------------------------------
# allow_from HOST, ..., HOST
#------------------------------------------------------------
# Set the allowed hosts for the current limit.

sub allow_from {
    my $self = shift;
    $self->{'allow'} = [@_];    
}    

#------------------------------------------------------------
# deny_from HOST, ..., HOST
#------------------------------------------------------------
# Set the denied hosts for the current limit.

sub deny_from {
    my $self = shift;
    $self->{'deny'} = [@_];    
}    

#------------------------------------------------------------
# end_limit
#------------------------------------------------------------
# End the current limit.

sub end_limit {
    1;
}    

#------------------------------------------------------------
# order CLASS, ..., CLASS
#------------------------------------------------------------
# Set evaluation order for the current limit.  
# Each CLASS should be one of "ALLOW" or "DENY".

sub order {
    my $self = shift;
    $self->{Order} = [map {lc($_)} @_];
}    

#------------------------------------------------------------
# require 
#------------------------------------------------------------
# Set evaluation order for the current limit.
# Each CLASS should be one of "ALLOW" or "DENY".

sub require {
    my $self = shift;
    $self->{require_type} = [map {lc($_)} @_];
}

sub auth_type {
    my $self = shift;
    $self->{AuthType} = [@_] ;
}

sub auth_name {
    my $self = shift;
    $self->{AuthName} = [@_];
}

sub auth_user_file {
    my $self = shift;
    $self->{AuthUserFile} = [@_];
}

#------------------------------------------------------------
# start_limit METHOD, ..., METHOD
#------------------------------------------------------------
# Begin a new limit.  Each METHOD should be one of "GET", "POST", "PUT", etc.

sub start_limit {
    my $self = shift;
    $self->{Limit} = [@_];
}

#------------------------------------------------------------
# error_doc NUMBER, RESPONSE
#------------------------------------------------------------
# Define an ErrorDocument for this .htaccess (Override FileInfo
# must be on for this directory).
#
# NUMBER is an HTTP error number.
# RESPONSE is a URL (relative or absolute), a double-quoted
#          string (INCLUDING the quotes!), or nothing (to 
#          delete the ErrorDocument).
#
# Calling the method twice for the same error number 
# will replace the previous value.
sub error_doc {
  my ($self,$error_number,$what_to_do) = @_;
  $self->{ErrorDoc}->{$error_number} = $what_to_do;
}

#------------------------------------------------------------
# add_type MIMETYPE, SUFFIXES
#------------------------------------------------------------
# Define a MIME type that can be served out of this directory
# and the ones below it.
#
# MIMETYPE is any MIME type, including user-defined ones.
# SUFFIXES is a list of file suffixes that should be served 
#          using this MIME type.
# 
# You may define as many MIME types as you like; calling
# add_type() with the same MIME type will replace the previous
# value.
sub add_type {
   my($self,$mime_type,@suffixes) = @_;
   $self->{MimeTypes} = {} unless defined $self->{MimeTypes};
   $self->{MimeTypes}->{$mime_type} = [@suffixes];
}

#-----------------------------------------------------------
# add_handler HANDLER, SUFFIXES
#-----------------------------------------------------------
# Define a handler for the given suffixes.
#
# HANDLER is a serer-supplied handler name.
# SUFFIXES is a list of suffixes to be processed by this handler.
#
# You may define as many handlers as you like; calling
# add_handler() with the same handler name will replace the previous
# value.
sub add_handler {
   my($self,$handler,@suffixes) = @_;
   $self->{Handlers} = {} unless defined $self->{Handlers};
   $self->{Handlers}->{$handler} = [@suffixes];
}

#------------------------------------------------------------
# options OPTIONLIST
#------------------------------------------------------------
# Add an "Options" line with the supplied options.
#
# OPTIONLIST is one or more server-supported options.
#
# Calling options() with no arguments deletes all options;
# calling multiple times adds arguments to the list.
sub options {
   my($self,@options) = @_;
   $self->{Options} = [] unless defined $self->{Options};
   if (@options) {
      push @{$self->{Options}}, @options;
   }
   else {
      $self->{Options} = [];
   }
}

#-----------------------------------------------------------
# xbithack SETTING
#-----------------------------------------------------------
# Turns on/off the XBitHack option
#
# XBitHack enables SSI for any HTML file with the execute bit on.
#
# SETTING = "full" - inserts XBitHack Full (server uses last-modified header)
#           false value - XBitHack Off
#           true value  - XBitHack On (Enables SSI) 
# Repeated calls change the setting to whatever the last call set it to.
sub xbithack {
   my ($self,$setting) = @_;
    if (lc($setting) eq "full") {
        $self->{XBitHack} = "Full";
    }
    elsif (lc($setting) eq "on" or $setting) {
        $self->{XBitHack} = "On";
    }
    else {
        $self->{XBitHack} = "Off";
    }
}

#------------------------------------------------------------
# no_limit SETTING
#------------------------------------------------------------
# turns on/off printing of the <LIMIT> info.
# default is on.
#
# SETTING true - turn off the limits
#        false - print the limits
sub no_limit {
   my ($self,$setting) = @_;
   if (defined $setting and $setting) {
      $self->{NoLimit} = 1;
   }
   else {
      $self->{NoLimit} = 0;
   }
}
#------------------------------------------------------------
# write_in DIRECTORY
#------------------------------------------------------------
# Write the object to an C<.htaccess> file in the given directory.
# Returns truth on success, falsity on failure.

sub write_in {
    my ($self, $dir) = @_;

    (-d $dir) or return error "directory '$dir' does not exist";
    my $path = "$dir/.htaccess";
    open HTA, ">$path" or return error "open $path: $!";

    # ErrorDocuments, if any.
    foreach my $doc (keys %{$self->{ErrorDoc}}) {
      if (defined $self->{ErrorDoc}->{$doc} and $self->{ErrorDoc}->{$doc}) {
	print HTA "ErrorDocument $doc $self->{ErrorDoc}->{$doc}\n";
      }
    }
    # AddType, if any.
    if (defined $self->{MimeTypes}) {
       my %types = %{$self->{MimeTypes}};
       local $_;
       foreach (keys %types) {
          print HTA "AddType $_ " . join(' ',@{$types{$_}});
       }
       print HTA "\n";
    }
    # AddHandler, if any.
    if (defined $self->{Handlers}) {
       my %handlers = %{$self->{Handlers}};
       local $_;
       foreach (keys %handlers) {
          print HTA "AddHandler $_ " . join(' ',@{$handlers{$_}});
       }
       print HTA "\n";
    }
    # Options, if any.
    if (defined $self->{Options}) {
       print HTA "Options " . join(' ',@{$self->{Options}}) ."\n";
    }
    # XBitHack, if any.
    if (defined $self->{XBitHack}) {
       print HTA "XBitHack $self->{XBitHack}\n";
    }
    # Auth info, if any.
    print HTA "AuthName \"", join(" ",@{$self->{AuthName}}), "\"\n"
      if defined $self->{AuthName} && @{$self->{AuthName}};
    print HTA "AuthType ", join(" ",@{$self->{AuthType}}), "\n"
      if defined $self->{AuthType} && @{$self->{AuthType}};
    print HTA "AuthUserFile ", join(" ",@{$self->{AuthUserFile}}), "\n"
      if defined $self->{AuthUserFile} && @{$self->{AuthUserFile}};
    print HTA "\n";

    unless ($self->{NoLimit}) {
       # Print the limit:
       print HTA "<Limit ", join(' ', @{$self->{Limit}}), ">\n";

       # Allow and deny. Don't print if there are none.
       my($allowdeny,@allowdenys);
       foreach $allowdeny (@{$self->{Order}}) {
   	   my $host;
	   foreach $host (@{$self->{$allowdeny}}) {
	       push @allowdenys, "   $allowdeny from $host\n";
	   }
       }

       if (int @allowdenys) {
         print HTA "   order ", join(',', @{$self->{Order}}),"\n", 
                   @allowdenys;
       }
       else {
         # Print the require:
         my $require;
         foreach $require (@{$self->{require_type}}) {
	   print HTA "   require $require\n";
         }
       }

       # Close the limit:
       print HTA "</Limit>\n";   
    }
    close HTA;
    1;
} 

#------------------------------------------------------------
1;
