package IU::HTTP;

=head1 NAME

IU::HTTP - HTTP client... a lightweight LWP!

=head1 SYNOPSIS

Here's the simple interface:

    # GET a URL... returns a filehandle to read from, or undef on error:
    my $DOC = IU::HTTP->get("http://host.name:port/path/to/doc");
    $DOC or puke "couldn't get URL!";
    while (<$DOC>) {
	print $_;          # output document to STDOUT
    }

Here's the full-fledged interface:

    # Create a new HTTP client object:
    my $httpc = new IU::HTTP;
    
    # Open an HTTP connection:
    $httpc->open($host, $port) or die "couldn't open $host:$port";
    
    # Send a GET command:
    $httpc->send_GET("/path/to/doc") || die "can't get document";
    
    # Swallow the HTTP reply header:
    $httpc->read_header or die "can't read header";
    
    # Now, read the data!
    my $doc = $httpc->fh;
    while (<$doc>) {
	print $_;          # output document to STDOUT
    }
     
    # Done! Sever the connection:
    $httpc->close;

The filehandle returned by 'get' will automatically be closed when
the HTTP object is destroyed.

=cut


use IU;
use IU::Log;
use FileHandle;
use Socket;

use strict;

#------------------------------------------------------------
# new
#------------------------------------------------------------
sub new {
    bless {}, shift;
}

#------------------------------------------------------------
# open HOST,[PORT]
#------------------------------------------------------------
# I<Class or instance method>.
# Opens a connection to the HTTP daemon at HOST (and, optionally, at PORT).
# Returns a truth on success, falsity on error.
#
# The return value is I<actually> a filehandle which may be used to 
# communicate with the daemon directly.

sub open {
    my ($self, $host, $port) = @_;
    $port or $port = 80;
    my ($hostaddr, $sockname);

    # Create new object if needed:
    ref($self) or $self = $self->new;

    # Create new filehandle:
    my $FH = new FileHandle;

    # Get host address:
    if ($host =~ /^\d+\.\d+\.\d+\.\d+$/) {
        $hostaddr = pack('c4', split(/\./, $host));
    }
    else {
	my ($fqdn) = inet_ntoa(inet_aton($host));
	$hostaddr = pack('c4',split(/\./,$fqdn));
        $fqdn or return error 452, "Cannot find hostname $host";
    }
    
    # Create local socket:
    socket($FH, AF_INET, SOCK_STREAM, 0) 
        or return error 490, "Failed to create a local socket: $!";

    # Connect:
    $sockname = pack('S n a4 x8', AF_INET, $port, $hostaddr);    
    connect($FH, $sockname) 
	or return error 490, "Couldn't connect to $host:$port: $!"; 

    # Okay!
    ($self->{FH} = $FH);
}

#------------------------------------------------------------
# close
#------------------------------------------------------------
# Sever the connection.

sub close {
    my $self = shift;
    close($self->{FH}) if $self->{FH};
}

#------------------------------------------------------------
# fh
#------------------------------------------------------------
# Get the filehandle being used for communication.  False if not connected.

sub fh {
    $_[0]->{FH};
}

#------------------------------------------------------------
# send_GET URI
#------------------------------------------------------------
# Send an HTTP GET request.  No usable return value.

sub send_GET {
    my ($self, $uri) = @_;

    # Print the request, followed by the data:				
    print {$self->{FH}} "GET $uri HTTP/1.0\r\n\r\n";
    $self->{FH}->flush();
}

#------------------------------------------------------------
# read_header 
#------------------------------------------------------------
# Read the HTTP response header.  Returns true if the header
# was ready and is a 200-series reply, false otherwise.
#
# WARNING: old-style servers are not supported.

sub read_header {
    my $self = shift;
    my $FH = $self->{FH};

    # Retrieve the initial response line:    
    defined($_ = <$FH>)
	or return error 454, "No response from server (possibly a bad URL)";
    m|^HTTP/\S+\s+(\d+)\s(.*)$|     # HTTP/1.0 or better
	or return error 501, "Old-style HTTP reply ($_) not supported";
    
    # Check status:
    my ($status, $message) = ($1, $2);
    ($status =~ /^2/) or return error 454, "Reply $status: $message";
   
    # Gobble MIME header:
    while (<$FH>) {
	last if /^[\r\n]+$/;       # end of header
    }
    1;
}

#------------------------------------------------------------
# get URL
#------------------------------------------------------------
# I<Class or instance method>.
# Get the given URL.  Returns a filehandle for reading the document
# from, or error.

sub get {
    my ($self, $url) = @_;

    # Create new object if needed:
    ref($self) or $self = $self->new;

    # Parse the URL:
    $url =~ m|^http://([^/:]+)(:([0-9]+))?(/.*)$|i 
	or return error "Bad URL: $url";
    my ($host, $port, $uri) = ($1, $3, $4);
    ($host && $uri) or return error "Unsupported URL: $url";
    $port = 80 unless defined($port);

    # Get it!
    $self->open($host, $port) or return undef;
    $self->send_GET($uri)     or return undef;
    $self->read_header        or return undef;
    $self->{FH};     
}

#------------------------------------------------------------
1;


