package IU::InvStore;


=head1 NAME

IU::InvStore - storage for datasets/granules/etc. from inventory search 


=head1 DESCRIPTION

A dispatcher to IU::Store.

Encapsulates the database used by either the search results or
the shopping cart.

Contains high-level routines specific to storing/retrieving
datasets, granules, etc.


=head1 SYNOPSIS

    # Open for writing/creating/reading:
    my $path = $SESSION->path_to('results.gdb');
    my $RESULTS = IU::Results->open($path, 'READER') || die;

    # Get granule keys, by dataset keys:
    my @gkeys    = $RESULTS->gkeys_for_datasets(\@dskeys);

    # Get granule keys, by package keys:
    my @gkeys    = $RESULTS->gkeys_for_packages(\@dskeys);

    # Add granules:
    $RESULTS->add_granules(\@granules);

    # Get granules, by granule keys:
    my @granules = $RESULTS->granules(\@gkeys);

    # Add datasets:
    $RESULTS->add_datasets(\@datasets);

    # Get datasets, by dataset keys:
    my @datasets = $RESULTS->datasets(\@dskeys);

    # Save gdb results to text file: (see IW::ModeDSETLIST.pm)
        # Get the path to the results gdb.
        my $gdb_path = $SESSION->path_to('results.gdb');
        # Get the type of the last search performed.
        my $last_search = $SESSION->state->{LastSearch};
        # Set "data set only" flag to false if you want granules too.
        my $ds_only = 0;
        # Use the InvStore routine to read the gdb and dump the datasets.
        IU::InvStore->save_db_txt($gdb_path, $last_search, $ds_only);

    # Close:
    close $RESULTS;


=head1 DATABASE SCHEMA

If the key is of the form C<gkey:GKEY>, 
then the value is the IU::Granule specified by C<GKEY>.

If the key is of the form C<dskey:DSKEY>, 
then the value is the IU::Dataset specified by C<DSKEY>.

If the key is of the form C<ds2gkeys:DSKEY>, 
then the value is an array of GKEYS for the dataset specified by C<DSKEY>.

If the key is of the form C<pk2gkeys:U_PKKEY>, 
then the value is an array of GKEYS for the dataset specified by 
the unique package key C<U_PKKEY> (of the form C<archtag/M/pkid>).


=head1 RCS INFO

$Id: InvStore.pm,v 1.28 2002/09/19 14:03:52 lorena Exp $

=cut

use IU;
use IU::Log;
use IU::Store ();
use IU::Dataset;
use IU::Granule;
use strict;

#------------------------------------------------------------
#
# Construction/destruction
#
#------------------------------------------------------------

#------------------------------------------------------------
# DESTROY 
#------------------------------------------------------------
sub DESTROY {
    $_[0]->close;
}

#------------------------------------------------------------
# open PATH,MODESTR
#------------------------------------------------------------
# I<Constructor.>  Open a "results" database. 

sub open {
    my ($class, $path, $modestr) = @_;
    $modestr or puke 600, "open: missing mode string!";

    # Create and init:
    my $self = bless {}, $class;
    $self->{Path} = $path;
    $self->{Store} = {};
    $self->{Stats} = {};
    $self->reopen($modestr);
}

#------------------------------------------------------------
# reopen NEWMODE
#------------------------------------------------------------
# Set the mode to NEWMODE.  Re-ties the underlying hash, if needed.
# Returns the self object, or false on failure.

sub reopen {
    my ($self, $newmode) = @_;

    if ($self->{Mode} and ($self->{Mode} ne $newmode)) {
	untie %{$self->{STORE}};
    }
    tie %{$self->{STORE}}, 'IU::Store', $self->{Path}, $newmode
	or puke "couldn't reopen $self->{Path} for $newmode";
    $self->{Mode} = $newmode;
    $self;
}

#------------------------------------------------------------
# close
#------------------------------------------------------------
# Close the underlying storage (untie it).

sub close {
    my $self = shift;
    if ($self->{Mode}) {
	untie $self->{STORE};
	$self->{Mode} = undef;
    }
}

#------------------------------------------------------------
# path
#------------------------------------------------------------
# Return the path we're opened on.

sub path {
    $_[0]->{Path};
}

#------------------------------------------------------------
# stats
#------------------------------------------------------------
# Get some information about the last operation.  Each operation
# may set some hash elements in a different way.

sub stats {
    shift->{Stats};
}

#------------------------------
#
# GKEYS
#
#------------------------------

#------------------------------------------------------------
# gkeys
#------------------------------------------------------------
# Return the array of all granule keys in storage.

sub gkeys {
    my ($self) = @_;
    map { /^gkey:(.*)/ ? ($1) : () } (keys %{$self->{STORE}});
}

#------------------------------------------------------------
# gkeys_for_datasets \@DSKEYS 
#------------------------------------------------------------
# Given a list of dataset keys, return the array of all granules
# (from the most recent search results) in the given datasets.
#
# INTENDED FOR SEARCH RESULTS, NOT FOR CART!
#
# \@DSKEYS is a (possibly undefined) reference to an array of dataset keys.

sub gkeys_for_datasets {
    my ($self, $dskeys) = @_;
    my $ds_gkeys;             # ref to array of gkeys for this dset

    # KLUDGE: sanity check (for now):
    $self->{Path} =~ /results\.gdb$/ 
	or puke 600, "can't do that for this database!";
    
    # Load all dataset-to-granule mappings, and accumulate array of gkeys:
    my @gkeys = ();
    my $dskey;
    foreach $dskey (@$dskeys) {       
#	debug "getting mapping for '$dskey'";
	($ds_gkeys = $self->{STORE}{"ds2gkeys:$dskey"}) or next;
#	debug "got mapping for '$dskey': " . $ds_gkeys->[0];
	push(@gkeys, @$ds_gkeys);
    }
    @gkeys;
}

#------------------------------------------------------------
# gkeys_for_packages \@PKKEYS
#------------------------------------------------------------
# Given a list of package keys, get all granules keys.
#
# INTENDED FOR SEARCH RESULTS, NOT FOR CART!
#
# This is used to get the "search result" granules from a particular
# package into the shopping cart.
#
# \@PKKEYS is a (possibly undefined) reference to an array of 
# package keys (it is an error if it has no package keys).

sub get_granules_in_packages {
    my ($self, $pkkeys) = @_;

    # KLUDGE: sanity check (for now):
    $self->{Path} =~ /results\.gdb$/ 
	or puke 600, "can't do that for this database!";
    
    # Make sure we have packages:
    (!defined($pkkeys) || (@$pkkeys == 0)) and
	puke 513, "no packages were selected!";


    # Get the array of granule keys for each package, and concat them:
    my @gkeys = ();
    my $pkkey;
    my ($flat, $fluffy);
    my ($archtag, $dstag, $gid, $pkid);
    foreach $pkkey (@$pkkeys) {

	# Handle this pkkey by case: 
	if ($pkkey =~ m|^(\d+)/(\d+)/\*/(.+)$|) {        # GRAN: A/DDD/*/GID
	    ($archtag, $dstag, $gid) = ($1, $2, $3);

	    # Just push this granule key:
	    push(@gkeys, "$archtag/$dstag/$gid");
	}
	elsif ($pkkey =~ m|^(\d+)/(\d+)/\*$|) {          # DSET: A/DDD/*
	    ($archtag, $dstag) = ($1, $2);

	    # Create dataset key:
	    my $dskey = "$archtag/$dstag";

	    # Get all gkeys in this dataset:
	    my $ds_gkeys = $self->{STORE}{"ds2gkeys:$dskey"} || next;

	    # Push them:
	    push(@gkeys, @$ds_gkeys);
	}
	elsif ($pkkey =~ m|^(\d+)/([^/]+)/(.+)$|) {      # PKID: A/DDD/PKID
	    ($archtag, $dstag, $pkid) = ($1, $2, $3);

	    # Create UNIQUE pkkey:
	    my $u_pkkey = "$archtag/M/$pkid";

	    # Get all gkeys in this package:
	    my $pk_gkeys = $self->{STORE}{"pk2gkeys:$u_pkkey"} || next;

	    # Push them:
	    push(@gkeys, @$pk_gkeys);
	}
	else {   # bogus
	    error "Skipping bad package id, `$pkid'";
	    next;
	}
    }
    @gkeys;
}

#------------------------------
#
# GRANULES
#
#------------------------------

#------------------------------------------------------------
# has_granules \@GKEYS
#------------------------------------------------------------
# Do the given granules exist?
# Returns a list that is subset of GKEYS, one for each granule that exists.
# It is an error to give this function an undefined value for \@GKEYS.

sub has_granules {
    my ($self, $gkeys) = @_;
    grep { exists($self->{STORE}{"gkey:$_"}) } @$gkeys;    
}

#------------------------------------------------------------
# granules \@GKEYS
#------------------------------------------------------------
# Get the given granules, by their granule keys.
#
# \@GKEYS is a reference to an array of granule keys:  
#
#    If undefined, it is assumed that ALL granules are desired.  
#    If a ref to the empty list, NO granules will be returned.
#    Otherwise, only the granules with those keys will be returned.
#    A benign FATAL EXCEPTION if any of the granules can't be got.
#
# Storage must at least be open as a READER.
#
# B<WARNING:> THIS SHOULD BE THE ONLY METHOD THAT RETRIEVES GRANULES DIRECTLY!
# All other methods should invoke this one as a back end!

sub granules {
    my ($self, $gkeys) = @_;
    my ($granule, @granules);
    my ($skey, @skeys);

    # Build list of actual storage keys:
    # my @skeys;
    if ($gkeys) {    # given explicity
	@skeys = map { "gkey:$_" } @$gkeys;
    }
    else {           # filter out of storage
	@skeys = grep { /^gkey:/ } (keys %{$self->{STORE}});
    }

    # Get the selected granules:
    foreach $skey (@skeys) {
	$granule = $self->{STORE}{$skey} ||
	    puke(533, "couldn't get data granule ($skey) from database ",
		      "($self->{Path})");
	$granule->inventory($self);              # remember us!
	push(@granules, $granule);
    }
    @granules;
}

#------------------------------------------------------------
# add_granules \@GRANULES
#------------------------------------------------------------
# Add the given granules to the storage.  They are added by GKEY
# existing, matching granules will NOT be overwritten.
#
# Storage must at least be open as a WRITER.
#
# Stats added are:
#
#     AlreadyThere => number of added granules already present in database
#
# B<WARNING:> THIS SHOULD BE THE ONLY METHOD THAT ADDS GRANULES DIRECTLY!
# All other methods should invoke this one as a back end!

sub add_granules {
    my ($self, $granules, $overwrite) = @_;
    # $overwrite true (1) forces granule to be added even if it is already 
    # there.
    my $gran;
    $self->{Stats}{AlreadyThere} = 0;
    foreach $gran (@$granules) {
	my $skey = "gkey:" . $gran->key;
        # skip adding if it is already there
        if (exists($self->{STORE}{$skey}) && !$overwrite) {
           ++($self->{Stats}{AlreadyThere});
           next;
        }
	my $inv = $gran->inventory(undef);     # break association
	$self->{STORE}{$skey} = $gran;         # store!
	$gran->inventory($inv);                # restore association
    }
    1;
}

#------------------------------------------------------------
# remove_granules \@GKEYS
#------------------------------------------------------------
# Remove the given granules from storage.
# Storage must at least be open as a WRITER.
#
# Stats added are:
#
#     AlreadyThere => number of removed granules which were present in database

sub remove_granules {
    my ($self, $gkeys) = @_;
    my $gkey;
    $self->{Stats}{AlreadyThere} = 0;

    # Build list of actual granules keys, if necessary:
    if (!defined($gkeys)) {               # filter out of storage...
        my @gkeys = map {/^gkey:(.*)/ ? $1 : ()} (keys %{$self->{STORE}});
        $gkeys = \@gkeys;
    }
    foreach $gkey (@$gkeys) {
	my $skey = "gkey:$gkey";
	++($self->{Stats}{AlreadyThere}) if exists($self->{STORE}{$skey});
	delete $self->{STORE}{$skey};
    }
    return @$gkeys; # return list of gkeys removed 
}

#------------------------------
#
# DSKEYS
#
#------------------------------

#------------------------------------------------------------
# sorted_dskeys [\@DSKEYS]
#------------------------------------------------------------
# Get/set the sorted order, as a ref to an array of dskeys.

sub sorted_dskeys {
    my ($self, $dskeys) = @_;
    
    if ($dskeys) {
#	debug "reassigning sorted order for dataset keys";
	$self->{STORE}{'dssorted'} = $dskeys;
    }
    else {	
	$dskeys = $self->{STORE}{'dssorted'};
    }
    $dskeys;
}

#------------------------------------------------------------
# sorted_gkeys [\@GKEYS]
#------------------------------------------------------------
# Get/set the sorted order, as a ref to an array of gkeys.

sub sorted_gkeys {
    my ($self, $gkeys) = @_;

    if ($gkeys) {
        debug "reassigning sorted order for data granules keys";
        $self->{STORE}{'gsorted'} = $gkeys;
    }
    else {
        $gkeys = $self->{STORE}{'gsorted'};
    }
    $gkeys;
}

#------------------------------
#
# DATASETS
#
#------------------------------

#------------------------------------------------------------
# dataset DSKEY
#------------------------------------------------------------
# Get the given dataset, by its dataset key.
# Benign fatal exception if not found.
#
# Storage must at least be open as a READER.
#
# B<WARNING:> THIS SHOULD BE THE ONLY METHOD THAT RETRIEVES DATASETS DIRECTLY!
# All other methods should invoke this one as a back end.

sub dataset {
    my ($self, $dskey) = @_;
    
    my $dset = $self->{STORE}{"dskey:$dskey"} ||
	puke(533, "couldn't get data set ($dskey) from database ",
	          "($self->{Path})");
    $dset->inventory($self);                 # remember our source
    $dset;
}

#------------------------------------------------------------
# datasets \@DSKEYS
#------------------------------------------------------------
# Get the given datasets, by their dataset keys.
#
# \@DSKEYS is a reference to an array of dataset keys:  
#
#    If undefined, it is assumed that ALL datasets are desired.  
#    If a ref to the empty list, NO datasets will be returned.
#    Otherwise, only the datasets with those keys will be returned.
#
# Storage must at least be open as a READER.

sub datasets {
    my ($self, $dskeys) = @_;
    my ($dataset, @datasets);
    my ($skey, @skeys);

    # Build list of actual dataset keys, if necessary:
    if (!defined($dskeys)) {               # filter out of storage...
	my @dskeys = map {/^dskey:(.*)/ ? $1 : ()} (keys %{$self->{STORE}});
	$dskeys = \@dskeys;
    }

    # Get the selected datasets:
    map { $self->dataset($_) } @$dskeys;
}

#------------------------------------------------------------
# just_datasets 
#------------------------------------------------------------
# Get the given "orderable" datasets, by their dataset keys.
#
#    If undefined, it is assumed that ALL datasets are desired.
#    If a ref to the empty list, NO datasets will be returned.
#    Otherwise, only the datasets with those keys will be returned.
#
# Storage must at least be open as a READER.

sub just_datasets {
    my $self = shift;
    my $set = $self->{STORE}{JustDatasets} || {}; 

    my @dskeys = keys %$set;

    $self->datasets(\@dskeys);

}

#------------------------------------------------------------
# selected_list_keys
#------------------------------------------------------------
sub selected_list_keys {

   my ($self, $type, $keys) = @_;


   $self->{STORE}{"Selected$type"} = $keys if ($keys);

   if (defined $self->{STORE}{"Selected$type"}) {
     return @{$self->{STORE}{"Selected$type"}};
   } else {
     return (wantarray ? () : undef);
   }
}

#------------------------------------------------------------
# page
#------------------------------------------------------------
# Store which page was last viewed. 

sub page {
   my ($self, $type, $nextpage) = @_; 

   $self->{STORE}{"Page$type"} = $nextpage if ($nextpage);
   if (defined $self->{STORE}{"Page$type"}) {
      return $self->{STORE}{"Page$type"};
   } else {
      return 1;
   }
}

#------------------------------------------------------------
# list_keys
#------------------------------------------------------------
sub list_keys {

   my ($self, $type, $keys) = @_;

   if ($keys) {
      $self->{STORE}{"List$type"} = $keys;
      # If creating a new list then reset page back to first.
      $self->{STORE}{"Page$type"} = 1;
   }
   if (defined $self->{STORE}{"List$type"}) {
      return @{$self->{STORE}{"List$type"}};
   } else {
     return (wantarray ? () : undef);
   }
}


#------------------------------------------------------------
# add_datasets \@DATASETS
#------------------------------------------------------------
# Add the given datasets to the storage.  They are added by DSKEY,
# so existing, matching datasets will be overwritten.
#
# Storage must at least be open as a WRITER.
#
# B<WARNING:> THIS SHOULD BE THE ONLY METHOD THAT ADDS DATASETS DIRECTLY!
# All other methods should invoke this one as a back end.

sub add_datasets {
    my ($self, $datasets) = @_;
    my $dset;
    foreach $dset (@$datasets) {

	my $inv = $dset->inventory(undef);               # break association
	$self->{STORE}{"dskey:" . $dset->key} = $dset;   # store
	$dset->inventory($inv);                          # restore association
    }
    1;
}

#------------------------------------------------------------
# add_just_datasets \@DATASETS
#------------------------------------------------------------
# Add the given "orderable" datasets to the storage.  They are added by DSKEY,
# so existing, matching datasets will be overwritten.
# Create a hash list of them.
#

sub add_just_datasets {
    my ($self, $datasets) = @_;
    my $dset;

    my $set = $self->{STORE}{JustDatasets} || {};

    $self->add_datasets($datasets);

    # create a hash to store the list of datasets put in just for themselves.

    foreach $dset (@$datasets) {

        $set->{$dset->key} = 1;
    }
    $self->{STORE}{JustDatasets} = $set; 
}

#------------------------------------------------------------
# has_datasets \@DSKEYS
#------------------------------------------------------------
# Do the given datasets exist?
# Returns a list that is subset of DSKEYS, one for each datasets that exists.
# It is an error to give this function an undefined value for \@DSKEYS.

sub has_datasets {

    my ($self, $dskeys) = @_;
    my $set = $self->{STORE}{JustDatasets} || {};
    grep { exists($set->{"$_"}) } @$dskeys;


}

#------------------------------------------------------------
# remove_just_datasets \@DSKEYS
#------------------------------------------------------------
# Remove from hash references for datasets to the storage.  
#
#

sub remove_just_datasets {
    my ($self, $dskeys) = @_;
    my $dskey;

    my $set = $self->{STORE}{JustDatasets} || {}; 

    # Build list of actual dataset keys, if necessary:
    if (!defined($dskeys)) {               # filter out of storage...
        my @dskeys =  (keys %{$set});
        $dskeys = \@dskeys;
    }
    # remove dataset keys from hash.

    foreach $dskey (@$dskeys) {

        delete $set->{$dskey};
        ++($self->{Stats}{AlreadyThere}) 
    }
    $self->{STORE}{JustDatasets} = $set;  
    return @$dskeys; # return list of dskeys removed.
}

#------------------------------------------------------------
# cart_medialimit_info \$MEDIAORDERLIMIT
#------------------------------------------------------------
# Get/set selected media order limit information from/to storage.
# It is set by key MediaOrderLimit so existing media limit info
# will be overwritten.  Only used by CART.
#
# Storage must at least be open as a WRITER.
#
# B<WARNING:> THIS SHOULD BE THE ONLY METHOD THAT ADDS DATASETS DIRECTLY!
# All other methods should invoke this one as a back end.

sub cart_medialimit_info {
   my ($self, $mediaorderlimit) = @_;

   $mediaorderlimit and $self->{STORE}{MediaOrderLimit} = $mediaorderlimit; 
   $self->{STORE}{MediaOrderLimit};
}

#------------------------------------------------------------
# remove_cart_medialimit_info 
#------------------------------------------------------------
# Remove existing media order limit information from storage.
# Storage must at least be open as a WRITER.
 
sub remove_cart_medialimit_info {
   my $self = shift;
   delete $self->{STORE}{MediaOrderLimit}; 
   1;
}

#------------------------------------------------------------
# remove_newlines
#------------------------------------------------------------
# Remove new line character "\n" out of text fields.
#------------------------------------------------------------
sub remove_newlines {
   my $value = shift;

   if (defined $value) {
      #$value  =~ s/\cA//g;
      $value  =~ s/\n/ /g;
   }

   return $value;
}

#------------------------------------------------------------
# save_db_txt
#------------------------------------------------------------
# Save results in a database type ascii text format.
# Output to the user's browser as a special MIME document.
#------------------------------------------------------------
sub save_db_txt {
    # Some mappings between web gateway and X client values.
    my %day_night = ('D'    => 'day',
                     'N'    => 'night',
                     'BOTH' => 'day and night'
       	            );
    my %browse_type = ('N'        => 'N',
                       'Y'        => 'IB/FB',
                       'FTP_ONLY' => 'FB'
                      );
    # These separators should match those found in IK_SysConfig.c.
    my $IMS_FIELD_SEPARATOR = ' | ';
    my $IMS_SUBFIELD_SEPARATOR = ' : ';
    # Some variables:
    my @datasets = ();            # array of data set objects
    my $dataset;                  # a data set object
    my @dskeys = ();              # array of data set keys
    my @gkeys = ();               # array of granule keys for a data set
    my @granules = ();            # array of granule objects
    my $granule;                  # granule object
    my $ds_tag;                   # data set tag
    my $coverage;                 # coverage area object

    # Get the passed variables.
    my $self = shift;
    my ($gdb_path, $last_search, $ds_only) = @_;
    # Verify we got the right stuff.

    # Log into results database:
    my $results = $self->open($gdb_path, 'READER')
        || puke "couldn't open results";

    # Get all datasets, and filter out invalid ones:
    @datasets = $results->datasets;

#    if ($SESSION->state->{LastSearch} eq 'INVENTORY_SEARCH') {
    if ($last_search ne 'DIRECTORY_SEARCH') {
	@datasets = grep { $_->n_granules > 0 } @datasets;
    }
    my $numdatasets = $#datasets+1;
    # Trap for no datasets!
    if ($numdatasets <= 0) {
        print STDOUT "No data sets.\n";
        return 1;
    }
    # If data set only flag is not set, print the granule header and granules.
    if (!$ds_only) {
        # Print the granule elements header.
        print STDOUT 'data granule id'                . $IMS_FIELD_SEPARATOR .
                     'archive'                   . $IMS_FIELD_SEPARATOR .
                     'data set'                  . $IMS_FIELD_SEPARATOR .
                     'data set key'              . $IMS_FIELD_SEPARATOR .
                     'campaign'                  . $IMS_FIELD_SEPARATOR .
                     'data granule source'            . $IMS_FIELD_SEPARATOR .
                     'sensor'                    . $IMS_FIELD_SEPARATOR .
                     'granual parameters'        . $IMS_FIELD_SEPARATOR .
                     'processing level'          . $IMS_FIELD_SEPARATOR .
                     'day/night'                 . $IMS_FIELD_SEPARATOR .
                     'browse type'               . $IMS_FIELD_SEPARATOR .
                     'centroid of coverage area' . $IMS_FIELD_SEPARATOR .
                     'start date'                . $IMS_FIELD_SEPARATOR .
                     'stop date'                 . $IMS_FIELD_SEPARATOR .
                     'package id'                . $IMS_FIELD_SEPARATOR .
                     'image id'                  . $IMS_FIELD_SEPARATOR .
                     'browse description'        . $IMS_FIELD_SEPARATOR .
                     "comments\n";
        # For each dataset, get the list of granule keys.
        foreach $dataset (@datasets) {
            push @dskeys, $dataset->key;
            # Extract the GUI's dataset_tag from the dataset key.
            $dskeys[0] =~ /^\d+\/(\d+)/;
            $ds_tag = $1;
            # Get the granule keys for this dataset.
            @gkeys = $results->gkeys_for_datasets(\@dskeys);
            # Get the granules.
            @granules = $results->granules(\@gkeys);
            # Print one line of info per granule.
            foreach $granule (@granules) {
                print STDOUT $granule->id || 'data granule ID missing';
                print STDOUT $IMS_FIELD_SEPARATOR;
                print STDOUT $granule->archive_id || 'Archive name is missing';
                print STDOUT $IMS_FIELD_SEPARATOR;
                print STDOUT $granule->dataset_id || 'Data set name is missing';
                print STDOUT $IMS_FIELD_SEPARATOR;
                print STDOUT $ds_tag;
                print STDOUT $IMS_FIELD_SEPARATOR;
                print STDOUT $granule->campaign;
                print STDOUT $IMS_FIELD_SEPARATOR;
                # Sources, sensors, parameters separated by subfield separator.
                print STDOUT join ($IMS_SUBFIELD_SEPARATOR, @{$granule->sources})
                         || 'no source names';
                print STDOUT $IMS_FIELD_SEPARATOR;
                print STDOUT join ($IMS_SUBFIELD_SEPARATOR, @{$granule->sensors})
                         || 'no sensor names';
                print STDOUT $IMS_FIELD_SEPARATOR;
                print STDOUT join ($IMS_SUBFIELD_SEPARATOR,
                                   @{$granule->parameters}) ||
                             'no data granule parameters';
                print STDOUT $IMS_FIELD_SEPARATOR;
                print STDOUT $granule->proc_level;
                print STDOUT $IMS_FIELD_SEPARATOR;
                print STDOUT $day_night{$granule->day_night};
                print STDOUT $IMS_FIELD_SEPARATOR;
                print STDOUT $browse_type{$granule->browse_type};
                print STDOUT $IMS_FIELD_SEPARATOR;
                # Get the coverage oboject of the granule.
                $coverage = $granule->coverage;
                if ($coverage->type eq 'GLOBAL') {
                    print STDOUT '0.00' . $IMS_SUBFIELD_SEPARATOR . '0.00';
                } else {
                    # Get the coverage area center.  Extract LAT and LON.
                    $coverage->center =~ /(\S+)\s+(\S+)/;
                    print STDOUT $1 . $IMS_SUBFIELD_SEPARATOR . $2;
                }
                print STDOUT $IMS_FIELD_SEPARATOR;
                print STDOUT $granule->start_date;
                print STDOUT $IMS_FIELD_SEPARATOR;
                print STDOUT $granule->stop_date;
                print STDOUT $IMS_FIELD_SEPARATOR;
                print STDOUT join ($IMS_SUBFIELD_SEPARATOR,
                               @{$granule->package_ids}) || "no package id's";
                print STDOUT $IMS_FIELD_SEPARATOR;
                print STDOUT $granule->image_id || 'no image id';
                print STDOUT $IMS_FIELD_SEPARATOR;
                print STDOUT remove_newlines($granule->browse_desc) ||
                                              'no browse descriptions';
                print STDOUT $IMS_FIELD_SEPARATOR;
                print STDOUT remove_newlines($granule->comment) ||
                                              'no data granule comments';
                print STDOUT "\n";
            }
            pop @dskeys;
        }
        # Print a blank line between the granules and the datasets.
        print STDOUT "\n";
    }
    # Print the data set elements header.
    print STDOUT 'data set key'              . $IMS_FIELD_SEPARATOR .
                 'data set'                  . $IMS_FIELD_SEPARATOR .
                 'archive'                   . $IMS_FIELD_SEPARATOR .
                 'GCMD entry tag'            . $IMS_FIELD_SEPARATOR .
                 'data set restrictions'     . $IMS_FIELD_SEPARATOR .
                 'unmapped fields'           . $IMS_FIELD_SEPARATOR .
                 'originating data center'   . $IMS_FIELD_SEPARATOR .
                 'GCMD start time'           . $IMS_FIELD_SEPARATOR .
                 'GCMD stop time'            . $IMS_FIELD_SEPARATOR .
                 'GCMD title'                . $IMS_FIELD_SEPARATOR .
                 'GCMD brief'                . $IMS_FIELD_SEPARATOR .
                 'GCMD attribute'            . $IMS_FIELD_SEPARATOR .
                 'GCMD reference'            . $IMS_FIELD_SEPARATOR .
                 'GCMD personnel'            . $IMS_FIELD_SEPARATOR .
                 'GCMD archive info'         . $IMS_FIELD_SEPARATOR .
                 'browse descriptions'       . $IMS_FIELD_SEPARATOR .
                 "data set comments\n";
    # For each dataset, print one line of info.
    foreach $dataset (@datasets) {
        # Extract the GUI's dataset_tag from the dataset key.
        $dataset->key =~ /^\d+\/(\d+)/;
        $ds_tag = $1;
        print STDOUT $ds_tag;
        print STDOUT $IMS_FIELD_SEPARATOR;
        print STDOUT $dataset->id || 'Data set name is missing';
        print STDOUT $IMS_FIELD_SEPARATOR;
        print STDOUT $dataset->archive_id || 'Archive name is missing';
        print STDOUT $IMS_FIELD_SEPARATOR;
        print STDOUT $dataset->MD_entry || 'no GCMD entry tag';
        print STDOUT $IMS_FIELD_SEPARATOR;
        print STDOUT remove_newlines($dataset->restriction) || 
                                      'no data set restrictions listed';
        print STDOUT $IMS_FIELD_SEPARATOR;
        print STDOUT 'no unmapped fields listed';  # unmapped_field
        print STDOUT $IMS_FIELD_SEPARATOR;
        print STDOUT $dataset->orig_center ||
           'no originating data center listed';
        print STDOUT $IMS_FIELD_SEPARATOR;
        print STDOUT $dataset->MD_start_time || '(none)';
        print STDOUT $IMS_FIELD_SEPARATOR;
        print STDOUT $dataset->MD_stop_time || '(none)';
        print STDOUT $IMS_FIELD_SEPARATOR;
        print STDOUT $dataset->MD_title || '(none)';
        print STDOUT $IMS_FIELD_SEPARATOR;
        print STDOUT '(none)';                 # md_brief
        print STDOUT $IMS_FIELD_SEPARATOR;
        print STDOUT $dataset->MD_attribute || '(none)';
        print STDOUT $IMS_FIELD_SEPARATOR;
        print STDOUT $dataset->MD_reference || '(none)';
        print STDOUT $IMS_FIELD_SEPARATOR;
        print STDOUT $dataset->MD_personnel || '(none)';
        print STDOUT $IMS_FIELD_SEPARATOR;
        print STDOUT $dataset->MD_arch_info || '(none)';
        print STDOUT $IMS_FIELD_SEPARATOR;
        print STDOUT remove_newlines($dataset->browse_desc) ||
                                          'no browse descriptions';
        print STDOUT $IMS_FIELD_SEPARATOR;
        print STDOUT remove_newlines($dataset->comment) ||
                                          'no data set comments';
        print STDOUT "\n";
    }
    1;
}
#------------------------------------------------------------
1;
