package IU::Log;


=head1 NAME

IU::Log - message logging


=head1 DESCRIPTION

Manages the logging of messages to log files or any other
facilities, via exported functions debug(), inform(), warning(),
and error().  

If you init() this module, logging is directed to the standard
facility, and the Perl warn() output is redirected through the
warning() function as well.


=head1 IMPLEMENTATION

Basically, this module keeps a set of one or more "log streams",
each indexed by name.  To install a new log stream, simply provide
the following to the set_handler() function...

=over

=item *

A msg_xxx() function to call whenever a message of any sort is  
logged: that msg_xxx() function will determine how to deal with the 
current message.

=item *

Some optional data (e.g., a filehandle name) which is made
available to the msg_xxx() function via the local variable $DATA.
This allows you to use the same msg_xxx() function for different
log streams.

=back

For example: whenever an error() message is logged, the msg_xxx() 
function is called for I<each> installed log stream... with the 
associated $DATA being made available to the msg_xxx() function.

You can also provide an instance of some subclass of L<IU::Logger>.


=head1  NOTES

It is imperative that the various logging functions in this
file do SOMETHING once the file is loaded, even if it has not
been init'ed!

The message logging functions are not idempotent.  Therefore, various
global sentinel flags have been placed in the code to prevent recusrion.
Note that these flags are reset after an eval() which could cause
a jump back to top level... since we desperately want an unwind() 
function!


=head1 AUTHORS

Eryq (Erik Dorfman) -- Hughes STX Corporation

Joe McMahon -- Hughes STX Corporation

=cut

use IU;
use FileHandle ();

# NOTE: can't "use IU"... must use IU:: functions with scope operator!

require Exporter;


@ISA       = qw(Exporter);
@EXPORT    = qw(DELOUSE debug inform warning error extramsg);
@EXPORT_OK = ();

use strict;
use vars qw(%Glyphs);


#------------------------------------------------------------
#
# Globals, public:
#
#------------------------------------------------------------

# The table of glyphs, for logging:
%Glyphs = (
    'DEBUG'   => 'D   ',
    'INFORM'  => 'II  ',
    'WARNING' => 'WWW ',
    'ERROR'   => 'EEEE',
    'EXTRA'   => '(X) ',
);

#------------------------------------------------------------
#
# Globals, private:
#
#------------------------------------------------------------

# Init state:
my $Init = 0;

# Can we log?
my $LogOpen = 0;

# The message handlers to use:
my %MsgHandlers = ();      

# Saved SIG{__WARN__} handler...
my $OldSigWarn;

# Is a message in progress?  Make message handler idempotent:
my $MsgSentinel = 0;

# Is a warning handler in progress?  Make warning handler idempotent:
my $WarnInProgress = 0;

# For saving the last error logged (sort of like an "errno" variable):
my $LastErrNo  = 0;       # message number of last error() message
my $LastErrMsg = '';      # message text of last error() message



#------------------------------------------------------------
# msg -- private: front end to the current msg_xxx() function
#------------------------------------------------------------
#
# DESCRIPTION
#    Back-end to debug(), error(), etc., and front-end to the
#    message handlers.
#
# SYNOPSIS
#    This subroutine is called like this:
#
#|         msg($msgtype, $msgno, $caller, @msg...);
#
#    Where...
#          $msgtype -- one of: DEBUG, INFORM, WARNING, ERROR, EXTRA
#          $msgno   -- the message number, or 0 if none
#          $caller  -- ref to an array of ($package, $filename, $line)
#          @msg...  -- remaining elements are actual message as logged
#
#    The information is passed on as-is to the actual message handler.
#
# NOTES 
#    This has an additional sentinel to protect against accidental
#    recursive invocations (since logging is not reentrant).
#
#    Remember that even non-IU functions can trigger IU logging, simply
#    by raising a fatal exception (dying).  This can lead to uncleared
#    sentinels if the fatal exception happened *during* message logging.
#    That's why we wrap the critical code in an eval() as added
#    protection.
    
sub msg {                      ## @_ is ($msgtype, $msgno, $caller, @msg)
    return undef if $MsgSentinel;   # recursive call attempted!
    $MsgSentinel = 1;               # begin critical section...
    local $^W = 0;                  # prevents core dumps from recursive warns

    # Call the normal logging handlers, with the same arguments:
    my ($name, $handler);

    #DELOUSE("log handlers...");
    #while (($name, $handler) = each %MsgHandlers) {
    foreach $handler (values %MsgHandlers) {
	#DELOUSE("    Removing handler...");
	$handler->msg(@_) if $handler;
    }
    #DELOUSE("log handlers... done");

    $MsgSentinel = 0;               # end critical section
    1;
}

#------------------------------------------------------------
# is_int -- private: return integer value of string, *without* a warning
#------------------------------------------------------------
sub is_int {                      ## @_ is ($string)
    local $^W = 0;
    int($_[0]);
}

#------------------------------------------------------------
# set_msg_sentinel -- gurutility: set the sentinel that prevents recursive msgs
#------------------------------------------------------------
# WARNING
#    DO NOT USE THIS UNLESS YOU KNOW WHAT YOU'RE DOING!  This function
#    should only be used to set the sentinel to 0 (clear) in response to an
#    eval-caught exception that could unwind the stack past the normal
#    sentinel clearing at the bottom of msg().  
#
#    The symptom of an uncleared sentinel is that the logging of messages
#    mysteriously stops, but execution continues as normal.
#
#    Okay... YOU HAVE BEEN WARNED.

sub set_msg_sentinel {
    my $value = shift;
    $MsgSentinel = $value;
}

#------------------------------------------------------------
# DELOUSE -- EXPORTED: we're debugging the logger: just print to the STDERR
#------------------------------------------------------------
# When stuff goes awry inside the logging package, the only way to debug
# it is to print to the STDERR.  That's what this does.
# 
# WARNING: All DELOUSE() calls MUST be removed in the production system!!!

sub DELOUSE {                      ## @_ is (@list)
    print STDERR "D> ", @_, "\n";
}
#---------------------------------------------------------------------------
# extramsg -- EXPORTED: debug only if forced to.
#---------------------------------------------------------------------------
#
# USAGE
#|    extramsg("Your message", " here"); 
#|    extramsg(123, "Your message", " here"); 
#    
sub extramsg (@) {
    return unless $IU::ForceDebug;
    my @caller = caller;                       # who called us?
    my $msgno = (is_int($_[0]) ? shift : 0);   # optional message number
    msg('EXTRA', $msgno, \@caller, @_);
}

#------------------------------------------------------------
# debug -- EXPORTED: log a debugging message
#------------------------------------------------------------
#
# USAGE
#|    debug("Your message", " here"); 
#|    debug(123, "Your message", " here"); 
#    
sub debug {                      ## @_ is (@list)
    my @caller = caller;                       # who called us?
    my $msgno = (is_int($_[0]) ? shift : 0);   # optional message number
    msg('DEBUG', $msgno, \@caller, @_);
}

#------------------------------------------------------------
# inform -- EXPORTED: log a informative message
#------------------------------------------------------------
#
# USAGE
#|    inform("Your message", " here"); 
#|    inform(123, "Your message", " here"); 
#    
sub inform {                      ## @_ is (@list)
    my @caller = caller;                      # who called us?
    my $msgno = (is_int($_[0]) ? shift : 0);  # optional message number
    msg('INFORM', $msgno, \@caller, @_);
}

#------------------------------------------------------------
# warning -- EXPORTED: log a warning message
#------------------------------------------------------------
#
# USAGE
#|    warning("Your message", " here"); 
#|    warning(123, "Your message", " here"); 
#    
# NOTES
#     The standard warn() function is routed through here.  We
#     needed a separate function in case the user wants to pass in
#     the optional message number, since warn() collapses its argument
#     list into a scalar before giving it to $SIG{__WARN__}  :-(
#
sub warning {                      ## @_ is (@list)
    my @caller = caller;                       # who called us?
    my $msgno = (is_int($_[0]) ? shift : 0);   # optional message number
    msg('WARNING', $msgno, \@caller, @_);
}

#------------------------------------------------------------
# error -- EXPORTED: log an error message
#------------------------------------------------------------
#
# USAGE
#|    error("Your message", " here"); 
#|    error(123, "Your message", " here"); 
#
# NOTES
#     This function always returns undef in a scalar context and
#     the empty list in a list context, so you can use it like this:
#
#|         open FILE, $filespec or return error 123, "open failed: $!";
#

sub error {                      ## @_ is (@list)
    my @caller = caller;                       # who called us?
    my $msgno = (is_int($_[0]) ? shift : 0);   # optional message number
    msg('ERROR', $msgno, \@caller, @_);

    # Because error messages are special, we provide these...
    ($LastErrNo, $LastErrMsg) = ($msgno, join('', @_));
    return (wantarray ? () : undef);        # for nested calls
}

#------------------------------------------------------------
# last_error -- return last error logged via this facility
#------------------------------------------------------------
# RETURNS
#    In a scalar context, returns the last error message logged.
#
#    In an array context, returns the list of two elements:
#    the last error message number, and the last error message.

sub last_error {
    return (wantarray ? ($LastErrNo, $LastErrMsg) : $LastErrMsg);
}


#------------------------------------------------------------
#
# Setup...
#
#------------------------------------------------------------

#------------------------------------------------------------
# init_sigwarn -- install the warning handler
#------------------------------------------------------------
sub init_sigwarn {

    # Install the warn() handler (de-installed on quit());
    #    WARNING: USE THE FULLY-QUALIFIED GLOBAL VARS TO AVOID CV_CLONE PANICS!
    $OldSigWarn = $SIG{'__WARN__'};
    $SIG{'__WARN__'} = sub {
	return undef if $IU::Log::WarnInProgress;
	$IU::Log::WarnInProgress = 1;         # begin critical section...
	 
	my @caller = caller;                       # who called us?
	my $msg = $_[0];
	msg('WARNING', 0, \@caller, $msg);

	$IU::Log::WarnInProgress = 0;         # end critical section...
    };
    1;
}

#------------------------------------------------------------
# quit_sigwarn -- deinstall the warning handler
#------------------------------------------------------------
sub quit_sigwarn {

    # De-install the warn() handler:
    $SIG{'__WARN__'} = $OldSigWarn;
}

#------------------------------------------------------------
# init -- init the logging package
#------------------------------------------------------------
#
# RETURNS
#    True for success, or false for failure.

sub init {
    my $file;

    # Initiallized already?
    return 1 if $Init;
    init_sigwarn();

    # Done!
    $Init = 1;
}

#------------------------------------------------------------
# init_term -- convenience: init and install term logger
#------------------------------------------------------------
sub init_term {
    my $fh = shift;

    require IU::TermLogger;    

    init();    
    set_handler('TERM', IU::TermLogger->open($fh));
    1;
}

#------------------------------------------------------------
# quit -- quit logging (even to the terminal!)
#------------------------------------------------------------
#
# DESCRIPTION
#    Stop logging...really.  Subsequent messages are re-routed to Tumbolia.
#    All currently-open log handlers are closed.
#
#    In other words, we put all our toys back where we found them,
#    and shut everybody up for good.
#
# WARNING
#    Yes, this undoes a little more than init() does, but it's more in the
#    spirit of a quit().

sub quit {
    return 1 if !$Init;         # no-op if not init'ed

    # Don't handle warnings anymore:
    quit_sigwarn();

    # Reset the message handlers to be silent!
    my %savedMsgHandlers = %MsgHandlers;
    %MsgHandlers = ();   # be silent!
    
    # Close any log facilities that might be open:    
    my $handler;
    foreach $handler (values %savedMsgHandlers) {
	$handler->close() if $handler;
    }

    $Init = 0;
    1;
}

#------------------------------------------------------------
# set_handler -- utility: add a msg_xxx-style message handler
#------------------------------------------------------------
# 
# DESCRIPTION
#    You can have one or more "message handler" functions installed,
#    as described by the msg() function.  Each of these is invoked 
#    whenever a message is logged, so it is really best to not
#    add more than one or two.
#
#    There are two ways to use this:
#
#         1.  Provide this function with the "name" of the handler, 
#             followed by instance of IU::Logger (or some subclass thereof).
#             This is the recommended way.
#
#|            set_handler(MYHANDLER, $syslogger);       # a IU::SysLogger;
#
#         2.  Provide this function with the "name" of the handler, and a
#             pointer to its subroutine.  An optional $data argument is 
#             passed into the message handler via the local variable $DATA.
#             This is the deprecated way, and it uses IU::SubLogger.
#
#|            set_handler(MYHANDLER, \&myhandler, $data);
#
#    Message handlers are added to the logger by name, so they
#    are easy to delete: just set the handler at that name to 
#    be the undefined value:
#
#|        set_handler(MYHANDLER, undef);
#
# RETURNS
#    The previous handler, as a IU::Logger.

sub set_handler {
    my ($name, $handler, $data) = @_;
    my $oldhandler = $MsgHandlers{$name};

    if ($handler) {    # install...
	if (ref($handler) eq 'CODE') {  # handler is a subroutine (deprecated)
	    require IU::SubLogger;
	    $MsgHandlers{$name} = IU::SubLogger->open($handler, $data);
	}
	else {                          # handler is a logger... good!
	    $MsgHandlers{$name} = $handler;
	}
    }
    else {             # de-install (but don't close!)
	delete $MsgHandlers{$name};
    }
    $oldhandler;
}

#-------------------- end of Log --------------------
1;
