package IU::Mail;


=head1 NAME

IU::Mail -- send mail


=head1 DESCRIPTION

Send email.  I would have used Mail::Send, except I can't figure
out how to get it to like a "-t" argument in "sendmail", and Mail::Internet
keeps all messages in core.


=head1 SYNOPSIS

   # Create object:
   $mail = new IU::Mail;
   
   # Set up header:
   $mail->from('me@myhost');
   $mail->to('you@yourhost', 'them@theirhost');
   $mail->subject('A very quick note');
   $mail->cc('you@yourhost', 'them@theirhost');
   $mail->bcc('you@yourhost', 'them@theirhost');
   
   # Output mail message:
   $FH = $mail->open() || die "can't open mail stream";
   print $FH "This is the message\n";
   $mail->insert("/home/user/.signature");
   
   # Send it!
   $mail->send() || die "couldn't send message";


=head1 AUTHOR

Eryq, Hughes STX Corporation


=head1 VERSION

$Id: Mail.pm,v 1.12 2000/05/09 16:31:00 joe Exp $

=cut

use IU;
use IU::Log;
use FileHandle ();

use strict;


#------------------------------------------------------------
#
# Globals
#
#------------------------------------------------------------

# For generating filehandles (code predates FileHandle :-) ):
my $GENSYM = 'FH0000';


#------------------------------------------------------------
# new -- create the new mailer object
#------------------------------------------------------------
sub new {
    my $type = shift;
    my $new = {};    
    $new->{Header} = {};
    $new->{Header}{'To'}   = [];
    $new->{Header}{'From'} = [];
    $new->{Header}{'Return-Path'} = [];
    bless $new, $type;
}

#------------------------------------------------------------
# DESTROY -- the object has gone out of scope
#------------------------------------------------------------
DESTROY {
    my $self = shift;
    if ($self->{FH}) {
	close($self->{FH});
	$self->{FH} = undef;
    }
}

#------------------------------------------------------------
# set -- set a header to a list of values
#------------------------------------------------------------
sub set {
    my $self = shift;
    my $hdr = shift;

    # Forbid this if the mail stream is currently open... it's too late!
    $self->{FH} and puke 600, "can't set fields if mail stream is open!";

    # Set it:
    $self->{Header}{$hdr} = [@_];
    1;
}

#------------------------------------------------------------
# add -- add a list of values to the header 
#------------------------------------------------------------
sub add {
    my $self = shift;
    my $hdr = shift;

    # Forbid this if the mail stream is currently open... it's too late!
    $self->{FH} and puke 600, "can't set fields if mail stream is open!";

    # Add values:
    $self->{Header}{$hdr} || ($self->{Header}{$hdr} = []);
    push(@{$self->{Header}{$hdr}}, @_);    
    1;
}

#------------------------------------------------------------
# get -- return a header field's values, as an array ref
#------------------------------------------------------------
sub get {
    my ($self, $hdr) = @_;
    return ($self->{Header}{$hdr} ? $self->{Header}{$hdr} : []);
}

#------------------------------------------------------------
# from -- who is the message to be from?
#------------------------------------------------------------
sub from {
    my $self = shift;
    $self->set('From', @_);
}

#------------------------------------------------------------
# to -- who is the message to?
#------------------------------------------------------------
sub to {
    my $self = shift;
    $self->set('To', @_);
}

#------------------------------------------------------------
# cc -- who is the message to be CC'ed to?
#------------------------------------------------------------
sub cc {
    my $self = shift;
    $self->set('Cc', @_);
}

#------------------------------------------------------------
# bcc -- who is the message to be blind-CC'ed to?
#------------------------------------------------------------
sub bcc {
    my $self = shift;
    $self->set('Bcc', @_);
}

#------------------------------------------------------------
# subject -- what is the subject of the message?
#------------------------------------------------------------
sub subject {
    my $self = shift;
    $self->set('Subject', @_);
}

#------------------------------------------------------------
# returnPath -- what is the return path?
#------------------------------------------------------------
sub returnPath {
    my $self = shift;
    $self->set('Return-Path', "<$_[0]>");
}

#------------------------------------------------------------
# open -- open a "sendmail" stream, returning the filehandle
#------------------------------------------------------------
sub open {    
    my $self = shift;
    no strict 'refs';

    # Fail immedaitely if no destination:
    (@{$self->{Header}{To}}) 
	or puke 600, "can't open mail stream if no 'to' field!";

    # Fill in defaults for 'From' and 'Return-path':
    #    NOTE: WE USE A BOGUS RETURN PATH, TO AVOID AUTO-REPLY CASCADES!!!!!
    unless ($self->get('From')->[0]) {
	# debug "Filling in default 'from'";
	my $hostname = ($IU::HttpdHost || 
			$ENV{SERVER_NAME} ||
			error "can't get hostname: $!: $?");
	$self->from("wwwuser\@$hostname")   if $hostname;
    }
    unless ($self->get('Return-Path')->[0]) {
	# debug "Filling in default 'return-path'";
	$self->returnPath(@{$self->get('From')});
    }

    # Create a new filehandle, and open a sendmail stream:
    my $fh = 'IU::Mail::' . ++$GENSYM;
    open($fh, "| $IU::SendmailCmd ") || 
	return error 453, "couldn't open sendmail stream: $!";
    $self->{FH} = \*$fh;
    
    # Send the mail header:
    my @keys = sort keys %{$self->{Header}};
    my $key;
    foreach $key (@keys) {
	my $val = join(', ', @{$self->{Header}{$key}});
	print $fh "$key: $val\n";
    }
    print $fh "\n";
    
    # Ready:
    return \*$fh;
}

#------------------------------------------------------------
# insert -- insert contents of $openexpr into currently-open mail stream
#------------------------------------------------------------
sub insert {
    my $self = shift;
    my $openexpr = shift;    

    # It is open?
    $self->{FH} or puke 600, "mail can't be sent if not open!";
    my $fh = $self->{FH};

    # Open, and do it:
    CORE::open(INSERT, $openexpr) or return error "can't open $openexpr";
    while (<INSERT>) { print $fh $_ }
    close(INSERT);
    1;
}

#------------------------------------------------------------
# send -- close and send the mail message
#------------------------------------------------------------
sub send {
    my $self = shift;
  
    # It is open?
    $self->{FH} or puke 600, "mail can't be sent if not open!";

    # Close and send:
    close($self->{FH});
    $self->{FH} = undef;

    # Return final result:
    (!($? >> 8)) or return error 454, "sendmail failed: ", ($? >> 8);    
    1;
}    

#------------------------------------------------------------
1;

