package IU::MediaMaximumCapacity;

=head1 NAME

IU::MediaMaximumCapacity - a group that allows DAACs to assign a maximum
                  capacity per media type per order.

=head1 DESCRIPTION

This group  can be found in the top level of the INVENTORY_RESULT message
with the following structure :

    GROUP                  = MEDIA_MAXIMUM_CAPACITY
      MEDIA_MB_CAPACITY = 105.50 
      TYPE_ID   = "8MM"
      COMMENT  = "EDC-ECS Daac media mb capacity for 8mm tapes"
      GROUP = MISC_URL
         URL         = "http://edc.usgs.gov/daacmediacapacity.html"
         URL_COMMENT = "EDC-ECS Daac media Maximum MB Capacity"
      END_GROUP = MISC_URL
    END_GROUP              = MEDIA_MAXIMUM_CAPACITY

The value for DATA_CENTER_ID comes from the message level and is set in the 
Media chunck by ClientChunk.pm when the chunks are parsed.

=cut

use IU;
use IU::Chunk;
use IU::FancyURL;
use IU::ODL ();
use IU::Log;
use IU::IW ;

#------------------------------------------------------------
# new
#------------------------------------------------------------
sub new {
    my $class = shift;
    bless {}, $class;
}

#------------------------------------------------------------
# capacity 
#------------------------------------------------------------
# Return the media capacity 

sub capacity {
    my ($self, $value) = @_;
    $self->{Capacity} = $value if @_ > 1;
    $self->{Capacity};
}

#------------------------------------------------------------
# size_mb 
#------------------------------------------------------------
# Return the request size mb 

sub size_mb {
    my ($self, $value) = @_;
    $self->{SizeMb} = $value if @_ > 1;
    $self->{SizeMb};
}

#------------------------------------------------------------
# size_granules 
#------------------------------------------------------------
# Return the request size granules 

sub size_granules {
    my ($self, $value) = @_;
    $self->{SizeGranules} = $value if @_ > 1;
    $self->{SizeGranules};
}

#------------------------------------------------------------
# archive_id 
#------------------------------------------------------------
# Return the archive id for which this media capacity applies 

sub archive_id {
    my ($self, $value) = @_;
    $self->{ArchiveId} = $value if @_ > 1;
    $self->{ArchiveId};
}

#------------------------------------------------------------
# comment
#------------------------------------------------------------
# Return the comment text associated with the capacity.

sub comment {
    my ($self, $value) = @_;
    $self->{Comment} = $value if @_ > 1;
    $self->{Comment};
}

#------------------------------------------------------------
# type
#------------------------------------------------------------
# Return the comment text associated with the capacity.

sub type {
    my ($self, $value) = @_;
    $self->{Type} = $value if @_ > 1;
    $self->{Type};
}


#-----------------------------------------------------------
# url 
#------------------------------------------------------------
# return url object
sub url {
   my $self = shift;
   $self->{MISC_URL};
}


#------------------------------------------------------------
# from_chunk CHUNK, [TYPE]
#------------------------------------------------------------
# I<Class or instance method.>
# Create a new instance from a chunk.

sub from_chunk {
    my ($self, $chunk, $type) = @_;
    ref($self) or $self = $self->new;
    $self->{Capacity}  = $chunk->value('MEDIA_MB_CAPACITY');
    $self->{Comment} = $chunk->value('COMMENT');
    $self->{ArchiveId} = $chunk->value('DATA_CENTER_ID');
    $self->{Type} = $chunk->value('TYPE_ID');
    $self->{SizeMb} = $chunk->value('REQUEST_SIZE_MB');
    $self->{SizeGranules} = $chunk->value('REQUEST_SIZE_GRANULES');
    my $misc_chunk = $chunk->value('MISC_URL');
    if ($misc_chunk) {
        $self->{MISC_URL} =
	   IU::FancyURL->from_chunk($misc_chunk, 'MISC_URL');
    }
    $self;
}

#------------------------------------------------------------
# from_ODL ODL, DAAC
#------------------------------------------------------------
# I<Class or instance method.>
# Create a new instance from ODL.
# this value in.
sub from_ODL {
    my ($self, $odl, $daac) = @_;
    ref($self) or $self = $self->new;
    $odl->rewind;
    my ($param, $value, $val);
    while (($param, $value) = $odl->next) {
        local $_;
        for ($param) {
            /MEDIA_MB_CAPACITY/ and do {
                $self->{Capacity}  = $value->text;
                next;
            };
            /TYPE_ID/           and do {
                $self->{Type}     = $value->text;
                next;
            };
            /REQUEST_SIZE_MB/ and do {
                $self->{SizeMb} = $value->text;
                next;
            };
            /REQUEST_SIZE_GRANULES/ and do {
                $self->{SizeGranules} = $value->text;
                next;
            };
            /COMMENT/           and do {
               $self->{Comment}   = $value->text;
                next;
            };
            /MISC_URL/           and do {
                $self->{MISC_URL} = IU::FancyURL->from_ODL($value);
                next;
            };
        }
    }
    $self->{ArchiveId} = $daac; 
    $self;
}

#------------------------------------------------------------
1;
