package IU::ODL;

=head1 NAME

IU::ODL - parse and represent a syntactic subset of ODL

=head1 SYNOPSIS

    use IU::ODL (); 

Input:

    # Input an ODL tree
    open IN, "somefile.odl";
    my $odl = parse IU::ODL \*IN;
    close IN;
    $odl or die "couldn't parse file";

Output:

    # Output an ODL tree to the currently-selected filehandle:
    $odl->output;
    
    # Output an ODL tree to a given filehandle:
    $odl->output(\*STDOUT);
    
    # Output a group to a given filehandle:
    $group->output(\*STDOUT, 'PARAMNAME');

Iterate over parameters/elements:
    
    # Iterate through elements of an array (sequence).
    #    Note: each element is an IU::ODL!
    $array->rewind;
    while ($value = $array->next) {
	...
    }
    
    # Iterate through parameter/value pairs of a group/object:
    #    Note: each value is an IU::ODL!
    $group->rewind;
    while (($param, $value) = $group->next) {
	...
    }
    
    # Iterate through particular parameters of a group/object:
    #    Note: each value is an IU::ODL!
    $group->rewind;
    while ($value = $group->next_param('PARAMNAME')) {
	...
    }

Get information as simple text strings:

    # Get first value from a sequence (array) or scalar, as a string:
    $value = $odl->text_value;
   
    # Get array of values from a sequence (array) or scalar, as strings:
    @values = $odl->text_values;
    
    # Get a scalar as a string, and a sequence as strings joined by "\n":
    $value = $odl->text;


=head1 ODL SYNTAX

Here's a sample ODL file:

    GROUP                = KEYS
      INITIAL_USER_KEY     = "13:56:21 199"
      AUTHENTICATION_KEY   = "hughes"
      AUTHENTICATOR        = "2-340923049-23c"
    END_GROUP            = KEYS

    GROUP                = KEYS
      INITIAL_USER_KEY     = "13:56:21 202"
      AUTHENTICATION_KEY   = "hstx"
      AUTHENTICATOR        = "2-340923049-23c"
    END_GROUP            = KEYS

    OBJECT               = REQUEST
      FIRST_NAME           = "James"
      LAST_NAME            = "Amrhein"
      MIDDLE_INITIAL       = ""
      ADDRESS              = ("7701 Greenbelt Road", "Suite 400", "")
      CITY                 = "Greenbelt"
      PHONE                = "301-441-4178"
      EMAIL                = "solomon@harp.gsfc.nasa.gov"
      STATE                = "Maryland"
      COUNTRY              = "USA"
      ZIP                  = "20770"
      FAX                  = "301-441-2392"
      TITLE                = ""
      ORGANIZATION         = "Hughes STX Corporation"
    END_OBJECT           = REQUEST

    END

Here's more-or-less how we represent it (parts of it, anyway).  Notice 
that everything currently uses arrays, with the instance variables stored 
in an anonymous hash in the 0th element.

     [
     {Type=>GROUP},
      KEYS => [ 
	 {Type=>GROUP},
	  INITIAL_USER_KEY   => \"13:56:21 199",
	  AUTHENTICATION_KEY => \"hughes",
	  AUTHENTICATOR      => \"2-340923049-23c",
	  ],
      KEYS => [ 
	 {Type=>GROUP},
	  INITIAL_USER_KEY   => \"13:56:21 202",
	  AUTHENTICATION_KEY => \"hstx",
	  AUTHENTICATOR      => \"2-340923049-23c",
	  ],
      REQUEST => [ 
         {Type=>OBJECT},
	  FIRST_NAME   => \"James",
	  LAST_NAME    => \"Amrhein",
	  ...
	  ADDRESS      => [ 
	       {Type=>ARRAY},
	       \"7701 Greenbelt Road", 
	       \"Suite 400"
	       ],
	  ORGANIZATION => \"Hughes STX Corporation"
	  ]
      ]


=head1 WARNINGS

Here's the nonsense we're up against when trying to parse ODL.
Notice that ODL breaks lines without syntactic annotations
or simple rules to allow us to put them back together:

    SOURCE_NAME          = ("METEOSAT-3", "GMS-3", "GMS-4", "
        GOES-6", "GOES-7", "METEOSAT-2", "METEOSAT-4", "
        METEOSAT-5", "NOAA-10", "NOAA-11", "NOAA-12", "NOAA-9")
    SENSOR_NAME          = ("AVHRR", "MIR", "TOVS", "VISSR")
    PARAMETER            = ("CEILING", "RADIANCE", "OZONE", "
        CLOUDS", "ICE", "OZONE", "SNOW", "CLOUD AMOUNT", "
        PRESSURE", "PRECIPITABLE WATER", "REFLECTANCE", "CLOUD TOP
        PRESSURE", "CLOUD TOP TEMPERATURE", "TEMPERATURE", "CLOUD 
        FORCING", "CLOUD HEIGHT", "CLOUD OPTICAL THICKNESS", "
        CLOUD PRECIPITABLE WATER", "CLOUD TYPES", "CLOUD VERTICAL 
        DISTRIBUTION", "ICE EXTENT", "REFLECTANCE")

It seems like we can assume that if the line is 66 characters long,
then it's been broken right on the whitespace; else, it's been
broken either before a whitespace character, or at the beginning 
of a string.

Here's how we deal with it:

=over

=item 1.

When parsing a string or array, we first try to glue all the 
pertinent lines together, discarding any horizontal whitespace 
after a newline.

=item 2.

If a break occurs past column 66, we insert a single space before the 
newline.

=item 3.

All horizontal space after newlines is deleted.

=item 4.

To patch an infamous ODL line-wrapping bug, any string beginning 
with a literal "\n" followed by a carriage return has both the "\n" and 
the newline removed.  Thus, this:

    DATA_URL = "\n
	ftp://foo.bar.com"

is ultimately treated as this:
		  
    DATA_URL = "ftp://foo.bar.com"

=item 5.

Finally, all remaining newlines are deleted.

=back


=head1 RCS INFO

$Id: ODL.pm,v 1.21 2000/05/09 16:27:13 joe Exp $

=cut

use strict;
use Carp;

#------------------------------------------------------------
#
# GLOBALS
#
#------------------------------------------------------------

# For each parse:
my @STACK = ();

# How to parse:
my $JustOne;


#------------------------------------------------------------
#
# CONSTRUCTORS AND PARSING
#
#------------------------------------------------------------

#------------------------------------------------------------
# new
#------------------------------------------------------------
sub new {
    my $class = shift;
    my $self = [];
    bless $self, $class;
}

#------------------------------------------------------------
# conv VALUE
#------------------------------------------------------------
# I<Constructor.>
# Make a reasonable guess about converting the given Perl data
# structure to an ODL value:
#
#   * Scalars become QUOTED STRINGS
#   * Arrays (of scalars!) become ARRAYS of quoted strings.
#   * Hashes become GROUPS (conv is applied to the values recursively).
#
sub conv {
    my ($class, $value) = @_;
    my $ref = ref($value);
    my $self;

    if (!$ref) {      # scalar
	$self = IU::ODL::Scalar->new($value, 'q');
    }
    elsif ($ref eq 'ARRAY') {
	my $e;
	$self = $class->new_array;
	foreach $e (@$value) { $self->add_elem($class->conv($e)) }
    }
    elsif ($ref eq 'HASH') {
	my ($k, $v);
	my $self = $class->new_group;
	while (($k, $v) = each %$value) {
	    $self->add_param($k, $class->conv($v));
	}
    }
    else {     # assume it's already ODL!
	$self = $value;
    }
    $self;
}


#------------------------------
#
# OTHER CONSTRUCTORS...
#
#------------------------------

# Return a new array (sequence).
# DEPRECATED.  Use sequence.
sub new_array {
    shift->sequence(@_);
}
# Return a new sequence.  
# Any arguments are assumed to be elements to add to the sequence.
sub sequence {
    my $class = shift;
    my $self = bless [{Type=>'ARRAY'}], $class;
    $self->add_elem(@_);
    $self;
}
# Return a new group:
sub new_group {
    bless [{Type=>'GROUP'}], $_[0];
}
# Return a new object:
sub new_object {
    bless [{Type=>'OBJECT'}], $_[0];
}

# Return a new datetime:
sub datetime {
    my ($class, $value) = @_;
    IU::ODL::Scalar->new($value, 'd');
}

# Return a new int:
sub int {
    my ($class, $value) = @_;
    IU::ODL::Scalar->new($value, 'i');
}

# Return a new real:
sub real {
    my ($class, $value) = @_;
    IU::ODL::Scalar->new($value, 'r');
}

# Return a new string:
sub string {
    my ($class, $value) = @_;
    IU::ODL::Scalar->new($value, 'q');
}

# Return a new symbol:
sub symbol {
    my ($class, $value) = @_;
    IU::ODL::Scalar->new($value, 'y');
}


#------------------------------------------------------------
# new_scalar VALUE,TYPE
#------------------------------------------------------------
# Return a new scalar. 
# DEPRECATED: use datetime(), int(), real(), string(), or symbol().
sub new_scalar {
    my ($class, $value, $type) = @_;
    warn "deprecated ${class}::new_scalar method invoked";
    IU::ODL::Scalar->new($value, ($type || 'q'));
}

#------------------------------------------------------------
# number VALUE
#------------------------------------------------------------
# Return new numeric scalar.  
# DEPRECATED: see new_scalar() for alternatives.
sub number {
    my ($class, $value) = @_;
    IU::ODL::Scalar->new($value, 'b');
}

#------------------------------------------------------------
# quoted VALUE
#------------------------------------------------------------
# Constructor.  Return new quoted scalar.
# DEPRECATED: see new_scalar() for alternatives.
sub quoted {
    my ($class, $value) = @_;
    IU::ODL::Scalar->new($value, 'q');
}

#------------------------------------------------------------
# bare VALUE
#------------------------------------------------------------
# Constructor.  Return new bare scalar.
# DEPRECATED: see new_scalar() for alternatives.
sub bare {
    my ($class, $value) = @_;
    IU::ODL::Scalar->new($value, 'b');
}









#------------------------------------------------------------
# add_elem VALUES...
#------------------------------------------------------------
# Add elements to a sequence.
#
# The elements should be either ODL scalar objects, or simple scalars:
# in the latter case, they are automatically converted into a quoted-string.

sub add_elem {
    my ($self, @elems) = @_;
    my $value;
    foreach $value (@elems) {
	ref($value) or ($value = IU::ODL::Scalar->new($value, 'q'));
	push(@$self, $value);
    }
    $self;
}

#------------------------------------------------------------
# add_param PARAM,VALUE
#------------------------------------------------------------
# Add a parameter to a group or object.
#
# The value should be either an ODL object, or a simple scalar:
# in the latter case, it is automatically converted into a quoted-string
# ODL scalar.

sub add_param {
    my ($self, $param, $value) = @_;
    ref($value) or ($value = IU::ODL::Scalar->new($value, 'q'));
    $value->name($param); 
    push(@$self, $param, $value);
}

#------------------------------------------------------------
# lex_array
#------------------------------------------------------------
# I<Private.>
# Lex an array in a string, of the form "SCALAR, SCALAR, ...";

sub lex_array {
    my ($class, $arraytext, $ontoken) = @_;

    while ($arraytext =~ /\G\s*("([^"]*)"|(\S+?))\s*\,/og) {     # "s are fun!
        if (defined($2)) {      # quoted
	    &$ontoken($class, 'AS', $2, 'q');
	}
        else {
	    &$ontoken($class, 'AS', $3, 'b');
	}
    }
}

#------------------------------------------------------------
# kill_newlines STRING
#------------------------------------------------------------
# See the documentation at the top of this module on broken
# ODL line splitting, and how we compensate for it.
# The string is assumed to have no opening quote.

sub kill_newlines {
    my ($text) = @_;

    # Add single space before newlines past column 66, PROVIDED that 
    # there isn't whitespace or a string boundary there already.  
    # DO THIS FIRST!!!
    $text =~ s/([^\n]{65,}[^ \t\"\'])\n/$1 \n/g;

    # Kill all horizontal whitespace at start or after newlines:
    $text =~ s/(\A|\n)[ \t]+/$1/g;

    # Kill literal '\n' at start of strings, if followed by newline: 
    $text =~ s/(\A|[\"\'])\\n\n?/$1/g;

    # Kill all other newlines:
    $text =~ s/\n//g;

    # Done!
    return $text;
}

#------------------------------------------------------------
# lex
#------------------------------------------------------------
# I<Private.>
# Get next token from input.  Input filehandle is in $FH.
# The tokens are:
# 
#    G= type paramname          GROUP = paramname
#    GE type [paramname]        END_GROUP = paramname
#    GP paramname scalar fmt    paramname = "scalar"
#    A= paraname                paramname = (        (start array)
#    AE paraname                )                    (end array)
#    AS scalar fmt              scalar               (in array)
#    ZZ                         EOF
#
# Returns: 1 on success, 0 on benign EOF.

sub lex {
    my ($class, $fh, $ontoken) = @_;
    $ontoken or ($ontoken = \&printtoken);
    while (<$fh>) {
	chomp;

	# Skip completely empty lines...
	next if (/^\s*$/);

	# GROUP or OBJECT...
	if (/^\s*(GROUP|OBJECT)\s*=\s*(\S+)\s*$/i) {
	    &$ontoken($class, 'G=', $1, $2);
	}

	# END_GROUP or END_OBJECT...
	elsif (/^\s*END_(GROUP|OBJECT)\s*(=\s*(\S+)\s*)?$/i) {
	    &$ontoken($class, 'GE', $1, $3);
	    return 1 if ($JustOne && (CORE::int(@STACK) == 2));       # done!
	}

	# Start of quoted string...
	elsif (/^\s*(\S+)\s*=\s*([\"\'])/) { 
            my $aname = $1;
	    my $q     = $2;    # remember the quote char we saw
#           $_ = $';

	    # Accumulate text until closing quote is seen:
            my $add;
            until ((($q eq "'") && (y/'// == 2)) or
                   (($q eq '"') && (y/"// == 2))) {
	       defined($add = <$fh>) or return undef;    # parse err!
               chomp $add;
               $_ .= "\n$add";
            } 
	    
	    # Clean up text:
	    my $clean = kill_newlines($_);
	    $clean =~ s/\A\s*(\S+)\s*=\s*//;     # kill prefix

	    # Output token as double-quoted string (KLUDGE):
	    $clean =~ s/(\A\'|\'\Z)/\"/g;
            &$ontoken($class, 'GP', $aname, qq{"$clean}, 'q');
	}

	# Start of 1-d array...
	elsif (/^\s*(\S+)\s*=\s*\(/) { 
            my $aname = $1;
#           $_ = $';

	    # Accumulate text until ")" is seen:
            my $add;
            until (/\)\s*\Z/) {              # until ) is last thing on line
               defined($add = <$fh>) or return undef;    # parse err!
               chomp $add;
               $_ .= "\n$add";
            } 

	    # Clean up text:
	    my $clean = kill_newlines($_);
	    $clean =~ s/\A\s*(\S+)\s*=\s*\(//; # kill prefix, including "("
            $clean =~ s/\)\s*$/,/;             # replace ")" with a ","

	    # Output tokens:
            &$ontoken($class, 'A=', $aname);
            lex_array($class, $clean, $ontoken);
            &$ontoken($class, 'AE', $aname);
	}

	# Bareword...
	elsif (/^\s*(\S+)\s*=\s*(\S+)\s*$/) {       
            &$ontoken($class, 'GP', $1, $2, 'b');      # param, scalar, fmt
	}

	# END keyword...
        elsif (/^\s*END\s*$/i) {
            &$ontoken($class, 'ZZ');
            return 0;        # benign EOF
        }

	# Hunh?
        else {
            # silently ignore, for now..
        }

    }

    # Be nice.  Finish for them:
    &$ontoken($class, 'ZZ');
    0;       # benign EOF
}

#------------------------------------------------------------
# parser_token
#------------------------------------------------------------
sub parser_token {
    my $class = shift;
    my $tok = shift;
    my $context = $STACK[-1];
    # print join(' | ', $tok, @_), "\n";

    if ($tok eq 'G=') {          # start new group/object...
	my ($type, $param) = @_;
	my $newgroup = bless [{Type=>$type}], $class;
	push @$context, $param, $newgroup;
	push @STACK, $newgroup;
    }
    elsif ($tok eq 'GP') {       # add scalar param to current group/object...
	my ($param, $scalar, $format) = @_;
	push @$context, $param => IU::ODL::Scalar->new($scalar, $format);
    }
    elsif ($tok eq 'GE') {       # end current group/object...
	pop @STACK; 
    }
    elsif ($tok eq 'A=') {       # start new array...
	my $param = shift @_;
	my $newarray = bless [{Type=>'ARRAY'}], $class;
	push @$context, $param, $newarray;
	push @STACK, $newarray;
    }
    elsif ($tok eq 'AS') {       # add scalar element to current array...
	my ($scalar, $format) = @_;
	push @$context, IU::ODL::Scalar->new($scalar, $format);
    }
    elsif ($tok eq 'AE') {       # end current array...
	pop @STACK;
    }
    1;
}

#------------------------------------------------------------
# parse
#------------------------------------------------------------
# Parse in THE ENTIRE FILE, as though there were a huge GROUP 
# surrounding all the items in the file.  Returns that group on success,
# undef on error.

sub parse {
    my ($class, $fh) = @_;

    $JustOne = 0;
    my $self = bless [{Type=>'GROUP'}], $class;
    @STACK = ('' => $self);
    my $status = eval { $class->lex($fh, \&parser_token) };    # just in case
    (defined($status) && !$@) ? $self : undef;
}

#------------------------------------------------------------
# parse_one
#------------------------------------------------------------
# Parse in ONE LABEL from the file.
# Returns the label on success, 0 on END or good end of file, undef on error.

sub parse_one {
    my ($class, $fh) = @_;

    $JustOne = 1;
    my $self = bless [{Type=>'GROUP'}], $class;
    @STACK = ('' => $self);
    my $status = eval { $class->lex($fh, \&parser_token) };    # just in case
    return undef if ($@ || !defined($status));     # error!
    return ($status ? $self : 0);                  # object or benign EOF
}


#------------------------------------------------------------
#
# INSTANCE METHODS
#
#------------------------------------------------------------

#------------------------------------------------------------
# as_array
#------------------------------------------------------------
# Return an Array, Group, or Object as an array of ODL objects.
# In the case of Groups and Objects, the array will be of the
# form (param, value, ..., param, value): perfect for stuffing
# into a hash, but remember that values with duplicate parameter
# names will be lost!

sub as_array {
    my $self = shift;
    my @copy = @$self;
    shift @copy;
    @copy;
}

#------------------------------------------------------------
# is_grouplike
#------------------------------------------------------------
# Is this a Group/Object?

sub is_grouplike {
    ($_[0]->type eq 'GROUP') || ($_[0]->type eq 'OBJECT');
}

#------------------------------------------------------------
# name
#------------------------------------------------------------
# Return the name of a group/object.
# This is set by add_param(), and is the same as its parameter name.

sub name {
    my ($self, $name) = @_;
    $self->[0]{Name} = $name if (@_ > 1);
    $self->[0]{Name};
}

#------------------------------------------------------------
# next
#------------------------------------------------------------
# Get next item in sequence.
#
# Returns ($param, $value) if sent to a Group/Object, and just ($value)
# if sent to an Array.  If no more items remain, returns empty list
# in array context, and undef in a scalar content.

sub next {
    my $self = shift;

    # If a group, return the next param/value pair:
    if ($self->is_grouplike) {
        my $oldwarn;
        ($oldwarn, $^W) = ($^W, 0);
	my $i = $self->[0]{Iter};
	my ($p, $v) = ($self->[2*$i + 1], $self->[2*$i + 2]);
	$v or return (wantarray ? () : undef);        # end!
	++($self->[0]{Iter});
        $^W = $oldwarn;
	return ($p, $v);
    }
    # If an array, return the next value:
    elsif ($self->type eq 'ARRAY') {
	my $i = $self->[0]{Iter};
	my $v = $self->[$i + 1];
	$v or return (wantarray ? () : undef);        # end!
	++($self->[0]{Iter});
	return $v;
    }
}

#------------------------------------------------------------
# next_param PARAMNAME
#------------------------------------------------------------
# Given a Group/Object, get next parameter of the given PARAMNAME.
#
# Returns the value.  If no more items remain, returns empty list
# in array context, and undef in a scalar content.

sub next_param {
    my ($self, $paramname) = @_;
    my ($p, $v);

    while (($p, $v) = $self->next) {
	return $v if ($p eq $paramname);
    }
    return (wantarray ? () : undef);
}

#------------------------------------------------------------
# output [FH, [PARAM, [INDENT]]]
#------------------------------------------------------------
# Output an ODL structure.

sub output {
    my ($self, $FH, $param, $indent) = @_;    
    $FH or $FH = select;
    defined($indent) or ($indent = 0);
    my $i = '  ' x $indent;

    # Get type, and handle:
    my $type = $self->type;
    if ($type eq 'SCALAR') {     # Scalar...
	printf $FH "$i%-20s = %s\n", $param, $self->format;
    }
    elsif (($type eq 'GROUP') || ($type eq 'OBJECT')) {    # group...
	my @copy = @$self; shift @copy;

	printf $FH "$i%-20s = $param\n", $type if $param;
	while (@copy) {
	    my ($p, $v) = (shift @copy, shift @copy);
	    $v->output($FH, $p, $indent+1);
	}	
	printf $FH "$i%-20s = $param\n\n", "END_$type" if $param;
	# printf "${i}END\n" if (!$param);   # removed cuz it created problems
    }    
    elsif ($type eq 'ARRAY') {    # Array...
	my @copy = @$self; shift @copy;

	printf $FH "$i%-20s = (\n", $param;
	while ($_ = shift @copy) {
	    printf $FH "$i%-20s     %s%s\n", '', $_->format($_), 
	    (@copy ? ',' : '');
	}
	printf $FH "$i%-20s   )\n", '';
    }
    else {
	Carp::confess "can't output structure of type $type!";
    }
}

#------------------------------------------------------------
# end [FH]
#------------------------------------------------------------
# I<Class/instance method>.  Just outputs the "END" to end a file.

sub end {
    my ($self, $FH) = @_;
    $FH or $FH = select;
    print $FH "END\n";
}

#------------------------------------------------------------
# param PARAMNAME
#------------------------------------------------------------
# Given a Group/Object, get first parameter of the given PARAMNAME.
# This resets the iterator, so you can use next_param() after it.
#
# If param not found, returns empty list in array context, and undef in 
# a scalar content.

sub param {
    my ($self, $paramname) = @_;
    $self->rewind->next_param($paramname);
}

#------------------------------------------------------------
# rewind
#------------------------------------------------------------
sub rewind {
    my $self = shift;
    $self->[0]{Iter} = 0;
    $self;
}

#------------------------------------------------------------
# text
#------------------------------------------------------------
# Simply join everything in text_values() together by "\n",
# and return the resulting string.  Returns the empty string on error.

sub text {
    join "\n", shift->text_values;
}

#------------------------------------------------------------
# text_value 
#------------------------------------------------------------
# If this is an array (sequence), return the text of the first element.
# Else, return undef.

sub text_value {
    my $self = shift;
    if ($self->type eq 'ARRAY') {
	my $first = $self->[1];
	return ($first ? $first->text_value : undef);
    }
    return undef;
}

#------------------------------------------------------------
# text_values
#------------------------------------------------------------
# If this is an array (sequence), map "text_value" over all the elements.
# Else, return the empty list.

sub text_values {
    my $self = shift;
    if ($self->type eq 'ARRAY') {
	my @elems = @$self; shift @elems;
	return (map {$_->text_value} @elems);
    }
    return ();
}

#------------------------------------------------------------
# type
#------------------------------------------------------------
sub type {
    my $self = shift;
    ref($self) ? $self->[0]{Type} : 'SCALAR';
}




#= = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = 
#= = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = 

package IU::ODL::Scalar;

@IU::ODL::Scalar::ISA = qw(IU::ODL);


#------------------------------------------------------------
# new TEXT,FORMAT
#------------------------------------------------------------
# Create a new scalar.  Format is one of:
#
#      b   unclassified bare text
#
#      q   quoted string
#      y   symbol
#      i   integer
#      r   real
#      d   datetime

sub new {
    my ($class, $text, $format) = @_;
    unless (defined($text)) {
       warn "Attempt to turn undefined value into ODL scalar!";
       # Build and return object:
       my $self = "" . $format;
       return (bless (\$self, $class));
    }
    $format or $format = 'q';

    # Clean it up:
    local $_ = $text;
    if ($format eq 'q') {
	s/\"//g;         # "s again!
    }
    elsif ($format eq 'b') {
	s/\s//g;           # get rid of spaces
	/./ or Carp::confess "can't have empty string be bare!";
    }
    elsif ($format eq 'y') {
	s/[\'\s]//g;           # get rid of ' and spaces
	if (m{\A[a-z_][a-z_0-9]*\Z}i) {
	    # ok
	}
	else {
	    Carp::confess "bad/unsupported symbol ($text)";
	}
    }
    elsif ($format eq 'i') {
	s/\<[^\>]+\>\s*\Z//g;  # kill units
	s/\s//g;               # kill whitespace
	s/^0*(.)/$1/;          # kill leading zeros
	s/\.0*$//;             # kill trailing .0000
	if (/\A[+-]?\d+\Z/) {
	    # ok
	}
	else {
	    Carp::confess "bad/unsupported int ($text)";
	}
	$_ = sprintf("%d", $_);     # don't use int()!!!!!
	s/^\+//;                     # get rid of leading '+'
    }
    elsif ($format eq 'r') {
	s/\<[^\>]+\>\s*\Z//g;  # kill units
	s/\s//g;               # kill whitespace
	s/^0*(.)/$1/;          # kill leading zeros

	if (/\A[+-]?\d+\Z/) {                                 # integer
	    $_ .= '.0';
	}
	elsif (/\A[+-]?((\d+\.\d*)|(\d*\.\d+))\Z/) {          # decimal
	    # ok; perl uses the same syntax!
	    s/(\.[1-9]+)0*$/$1/g;        # get rid of trailing zeros
	}
	elsif (/\A[+-]?(\d+\.\d*|\d*\.\d+)E([+-]?\d+)\Z/i) {  # scientific
	    # ok; perl uses the same syntax!
	}
	else {
	    Carp::confess "bad real ($text)";
	}
	s/^\+//;                     # get rid of leading '+'
    }
    elsif ($format eq 'd') {
	s/\s//g;               # kill whitespace
	my ($date, $time) = split(/[Tt]/, $_, 2);

	# Date:
	($date =~ m{\A(\d+)-(\d{1,2})-(\d{1,2})\Z})
	    or Carp::confess "bad datetime date ($date) in ($text)";
	my ($y, $m, $d) = ($1, $2, $3);
	$date = sprintf("%04d-%02d-%02d", $y, $m, $d);

	# Time:
	$time or $time = "00:00:00Z";
	($time =~ m{\A(\d{1,2}):(\d{2})(:[\d\.]+)?(Z|[+-]?\d+)?\Z}i)
	    or Carp::confess "bad datetime time ($time) in ($text)";
	defined($4) or $time .= 'Z';
	$_ = ($date . 'T' . $time);
    }
    else {
	Carp::confess "bad format ($format)";
    }
    
    # Build and return object:
    my $self = $_ . $format;
    bless \$self, $class;
}

#------------------------------------------------------------
# type        -- return the type:
#------------------------------------------------------------
sub type {
    'SCALAR';
}

#------------------------------------------------------------
# name        -- return the parameter name:
#------------------------------------------------------------
sub name {
    undef;
}

#------------------------------------------------------------
# text        -- return just the text:
#------------------------------------------------------------
sub text {
    my $self = shift;
    my $text = $$self;
    chop $text;
    $text;
}

#------------------------------------------------------------
# text_value  -- return just the text:
#------------------------------------------------------------
sub text_value {
    shift->text;
}

#------------------------------------------------------------
# text_values  -- return just the text:
#------------------------------------------------------------
sub text_values {
    shift->text;
}

#------------------------------------------------------------
# format      -- format for output:
#------------------------------------------------------------
sub format {
    my $self = shift;
    my $text = $$self;

    # Last char indicates format:    
    my $fmt = substr($text, length($text)-1, 1); 
    chop $text; 
    
    # Clothe in quotes?
    ($fmt eq 'q') ? qq{"$text"} : $text;
}

#------------------------------------------------------------

__END__

sub make {
    my ($kind, $raw) = @_;
    my $o = eval { IU::ODL->$kind($raw) };
    print(($@ ? "ERROR: $@" : "VALUE: $raw -> (".$o->format.")\n"));
}

# ints:
my ($raw, $o);

my @test = ("", 
	 "0", 
	 "1", 
	 "-1", 
	 "+1", 
	 "0034", 
	 "-0034", 
	 34.45,
	 -234.45,
	 "12.", 
	 "-12.", 
	 "+12.", 
	 "12.0", 
	 "-12.0", 
	 "12.0000", 
	 "-12.0000", 
	 ".12", 
	 "-.12", 
	 "+.12", 
	 ".", 
	 "1.14E01", 
	 "2.14E-2",
	 "3.14E+3",
	 "4.14E-04",
	 "5.14E+05",
	 "-5.14E+05",
	 "+5.14E+05",
	 "foobar", 
	 "Y", 
	 "23 <secs>", 
	 "1994-12-32",
	 "1994-1-1",
	 "1994-1-1T12:34",
	 "1994-1-1T12:34Z",
	 "1994-1-1T12:34:56Z",
	 "1994-1-1T12:34:56.789Z",
	 'HEY',
	 '_HEY_HEY_001'
);

print "\nINT\n";
foreach $raw (@test) {
    make('int', $raw);
}

print "\nREAL\n";
foreach $raw (@test) {
    make('real', $raw);
}

print "\nDATETIME\n";
foreach $raw (@test) {
    make('datetime', $raw);
}

print "\nSYMBOL\n";
foreach $raw (@test) {
    make('symbol', $raw);
}

print "\nBARE\n";
foreach $raw (@test) {
    make('bare', $raw);
}

print "\nSTRING\n";
foreach $raw (@test) {
    make('string', $raw);
}

1;

