package IU::Options;

=head1 NAME

IU::Options - a collection of criteria (e.g., for subsetting)

=head1 DESCRIPTION

Analogous to the contents of a SUBSET_OPTIONS or SUBSET_SPEC group.
Contains one or more IU::Criterion objects, one for each 
subsetting "criterion".

=head1 RCS INFO

$Id: Options.pm,v 1.12 2003/09/09 14:28:18 solomon Exp $

=cut

use IU;
use IU::Chunk;
use IU::Log;
use IU::Criterion;
use IU::ODL;

use IU::IW;


#------------------------------------------------------------
# new
#------------------------------------------------------------
sub new {
    my $class = shift;
    my $self = bless {}, $class;
    $self->{Criteria} = [];
    $self;
}

#------------------------------------------------------------
# criteria
#------------------------------------------------------------
# Return criteria, as an array:
#
sub criteria { @{$_[0]->{Criteria}} }

#------------------------------------------------------------
# is_required
#------------------------------------------------------------
# Is subsetting required for order?
#
sub is_required { ($_[0]->{REQUIRED} || '') eq 'Y' }

#------------------------------------------------------------
# type [TYPE]
#------------------------------------------------------------
# Get/set the type of these options.
# This should be the same as the prefix of the ODL group name; 
# e.g., "SUBSET" for SUBSET_OPTIONS.  
#
sub type {
    my ($self, $val) = @_;
    $self->{Type} = $val if @_ > 1; 
    $self->{Type};
}

#------------------------------------------------------------
# cost [COST]
#------------------------------------------------------------
# Get /set the cost of these options

sub cost { $_[1] ? $_[0]->{Cost} = $_[1] : $_[0]->{Cost} }

# name -- Product Name of the option
sub name {
    my ($self, $value) = @_;
    $self->{Name} = $value if (@_ > 1);
    $self->{Name};
}

#------------------------------------------------------------
# pretty_type
#------------------------------------------------------------
# Output type in a pretty way: DATASET_ORDER => "Dataset Order".
#
sub pretty_type {
    my $type = uc(shift->type);
    return 'Apply the following processing options to the data:'
     if $type eq 'SUBSET'; 
    return 'Apply the following processing options to the data:'
     if $type eq 'ORDER';
    return 'Give me all the granules for this data set which meet the following restrictions:' if $type eq 'DATASET_ORDER';
    $type;      # just return self
}

#------------------------------------------------------------
# user_made_selections
#------------------------------------------------------------
# I<Instance method.>
# Return true if user selected at least 1 subsetting criterion, 
# false otherwise.   Returns false if there are no criteria at all 
# (degenerate case).

sub user_made_selections {
    my $self = shift;
    my $crit;
    foreach $crit ($self->criteria) { 
	return 1 if $crit->user_made_selections;
    }
    0;
}

#------------------------------------------------------------
# verify_required_selections
#------------------------------------------------------------
# I<Instance method.>
# Assertion.  Return true if user selected all required subsetting criterion, 
# raise fatal exception otherwise.   Returns true if there are no 
# criteria at all (degenerate case).

sub verify_required_selections {
    my $self = shift;
    my $crit;
    foreach $crit ($self->criteria) { 
	### debug "checking crit ", $crit->name, " which ",
	###   ($crit->is_required ? "is" : "is not"), " required";
	if ($crit->is_required and !$crit->user_made_selections) {
	    $crit->whine(516, "required, but no value given");
	}
        my $subcrit;
        foreach $subcrit ($crit->subcriteria) {
           if ($subcrit->is_required and !$subcrit->user_made_selections) {
              $subcrit->whine(516, "required, but no value given");
           }
        }
    }
    1;
}





#------------------------------
#
# Chunk interface
#
#------------------------------

#------------------------------------------------------------
# from_chunk CHUNK, TYPE, [GRANULE]
#------------------------------------------------------------
# Create from a chunk.
#
sub from_chunk {
    my ($self, $chunk, $type, $granule) = @_;
    ### my $dataset = $IU::Options::DatasetContext;    # KLUDGE
    ### my $granule = $IU::Options::GranuleContext;    # KLUDGE

    # Make safe for instances or classes:
    ref($self) or $self = $self->new;
    $self->type($type) if $type;

    # Get basic scalar instance vars:
    my $key;
    foreach $key (qw(REQUIRED)) { $self->{$key} = $chunk->value($key) }
    # Add any SPECIALIZED_CRITERIA...
    my $crit_chunk;
    foreach $crit_chunk ($chunk->values('SPECIALIZED_CRITERIA')) {
	# Parse it, and save it:
	my $crit = IU::Criterion->from_chunk($crit_chunk,$granule) || next;
	push @{$self->{Criteria}}, $crit;
    }

    # Done:
    $self;
}



#------------------------------
#
# CGI interface
#
#------------------------------

#------------------------------------------------------------
# _ident PREFIX
#------------------------------------------------------------
# I<Instance method, private.>  Hack up an ident, for criteria.

sub _ident {
    my ($self, $prefix) = @_;
    my ($ident) = ($prefix =~ /([\d]+)/);
    "Package $ident, " . $self->pretty_type . " options, ";
}

#------------------------------------------------------------
# read_CGI PARAMHASH
#------------------------------------------------------------
# I<Instance method.>
# Initiallize the user selections from CGI tags.  The PARAMHASH contains
# the following:
#
#     Prefix   => The prefix (like "ship.").  Optional; defaults to "".
#
# Normally, a fatal exception is raised if any criterion labelled as 
# "required" is missing in the CGI input.  This is undesirable if the 
# options themselves are not required; in such cases, we prefer to raise 
# exceptions ONLY IF IT LOOKS LIKE THE USER IS SUPPLYING VALUES.
#
# If the options are not required, we relax requirements
# on criteria, THEN read them in; then, if any are given,
# we assume that the user did in fact want to supply these options,
# and demand that all required ones be supplied.    
#
# A fatal exception will be raised if the user input is bad, or if
# any required criteria are missing.  

sub read_CGI {
    my ($self, %params) = @_;
    ref($self) or puke "instance method only!";
 
    # Get args:
    my $prefix = $params{Prefix} || '';
    my $page = $params{Page} || '';
    my $min = $params{Min} || '';

    # Hack up an ident: (KLUDGE)
    my $ident = $self->_ident($prefix);

    # For each criterion:
    my $crit;
    my $crit_req;
    my $cost;
    my $i;
    foreach $crit ($self->criteria) {
	$crit->ident($ident);

	# Read in the criterion, protecting the value if necessary: 
	if (!$self->is_required) {

	    # Protect value.  KLUDGE! FRAGILE CODE! KEEP THAT eval() THERE!
	    $crit_req = $crit->is_required;         # save
	    $crit->is_required(0);                  # fake unrequired
            if($crit->type eq 'SUBPAGE') {
                  eval{$crit->read_CGI(Page=> $page, Min=> $min, Prefix=>
                                       ($prefix . $crit->name))};
            } else {
               eval { $crit->read_CGI(Prefix=>($prefix . $crit->name . ".")) };
            }
	    $crit->is_required($crit_req);          # restore
	    die $@ if $@;      # if we did die, it's not because of req'ment
	} 
	else {
           if($crit->type eq 'SUBPAGE') {
                 $crit->read_CGI(Page=> $page, Min=>$min, Prefix=>
                                      ($prefix . $crit->name));
           } else {
              $crit->read_CGI(Prefix=>($prefix . $crit->name . "."));
           }
	}
        $cost += $crit->cost if $crit->cost;
        # Loop over any criteria within the criteria (nested
        # SPECIALIZED CRITERIA. 
        my $subcrit;
        my $crittype = $crit->type;
        foreach $subcrit ($crit->subcriteria) {
           $subcrit->ident($ident);
#          debug "about to fill ", $subcrit->name, " in from CGI";

           # Read in the subcriterion, protecting the value if necessary:
           if (!$self->is_required) {

              # Protect value.  KLUDGE! FRAGILE CODE! KEEP THAT eval() THERE!
              $subcrit_req = $subcrit->is_required;         # save
              $subcrit->is_required(0);                  # fake unrequired
              unless ($crittype eq 'SUBPAGE') {
                 eval { $subcrit->read_CGI(Prefix=>($prefix . $crit->name.".".
                                                    $subcrit->name . ".")) };
              }
              $subcrit->is_required($subcrit_req);          # restore
              die $@ if $@;      # if we did die, it's not because of req'ment
           }
           else {
              unless ($crittype eq 'SUBPAGE') {
                 $subcrit->read_CGI(Prefix=>($prefix . $crit->name . "." .  
                                          $subcrit->name . "."));
              }
           }
           $cost += $subcrit->cost if $subcrit->cost;
        }
        my $name = $crit->name;
        my $value = $crit->as_string;
        $self->name($value) if($name eq 'Product Name');
    }

    # If we relaxed requirements, we must enforce them again...
    # Assert that all required criteria have been supplied:
    $self->verify_required_selections if (!$self->is_required and 
					   $self->user_made_selections);
    # total cost of this option.
    $self->{Cost} = $cost if $cost;
    1;
}

#------------------------------------------------------------
# set_CGI PARAMHASH
#------------------------------------------------------------
# I<Instance method.>
# Initiallize the CGI tags from the object.  The PARAMHASH contains
# the following:
#
#     Prefix =>  The prefix (like "ship.").  Optional; defaults to "".
#
# A fatal exception will be raised if the user input is bad.

sub set_CGI {
    my ($self, %params) = @_;
    ref($self) or puke "instance method only!";

    # Get args:
    my $prefix = $params{Prefix} || '';

    # Hack up an ident: (KLUDGE)
    my $ident = $self->_ident($prefix);

    # For each criterion:
    my $crit;
    my $crittype;
    foreach $crit ($self->criteria) {
       $crit->ident($ident);
       $crit->set_CGI(Prefix=>($prefix . $crit->name . "."));
       $crittype = $crit->type;
       my $subcrit;
       if ($crittype eq 'SUBPAGE') {
          my $page = $crit->{UserSlct}{VALUE} || 1;
          my @subarray = ($crit->subcriteria);
          my $total = int(@subarray);
          my $number = $total/$page;
          my $j;
          my $i;
          for($i = 1; $i <=$page; $i++) {
             my $begin = ($i-1) * $number;
             my $end = $i*$number;
             for($j = $begin; $j < $end; $j++) {
                $subcrit = $subarray[$j];
                $subcrit->ident($ident);
                $subcrit->set_CGI(Prefix=>($prefix . $crit->name . "page$i." . 
                                  $subcrit->name . "."));
             }
          } 
       } else {
          foreach $subcrit ($crit->subcriteria) {
             $subcrit->ident($ident);
             $subcrit->set_CGI(Prefix=>($prefix . $crit->name . "." . 
                $subcrit->name . "."));
          }
       }
    }
    1;
}



#------------------------------
#
# ODL interface
#
#------------------------------

#------------------------------------------------------------
# to_request_ODL
#------------------------------------------------------------
# I<Instance method.>  Return an ODL group for this criterion
# with the contents of SUBSET_SPEC group, such as one 
# which might appear in a PRODUCT_REQUEST.

sub to_request_ODL {
    my $self = shift;

    # Start with a group:
    my $opts_odl = IU::ODL->new_group;

    # For each criterion...
    my $crit;
    ### debug "converting options to ODL...";
    foreach $crit ($self->criteria) {
	

        if ($crit->type eq 'SUBPAGE') { # special for SUBPAGE
           my $page = $crit->{UserSlct}{VALUE} || 1;
           my $i;
           for ($i = 1; $i <= $page; $i++) {
              $opts_odl = $crit->to_request_ODL($i, $opts_odl);
           }
        }
	# Skip if no user selections for this (presumably optional) criterion.
	#    Note that we catch the required-but-missing cases in from_CGI().
	$crit->user_made_selections or next;

        # I can't figure out why this is here.  It adds an extra
        # geo group to to the options ODL (without SPECIALIZED GROUP wrapper).
        # Also it doesn't work for POINT_LOC, only RANGE_LOC.
        # I'd take this code out but it has been here for over two years
        # without causing trouble and there is a remote possibility that
        # some daac maybe relying on this????
        # Ed Seufert - Sept. 30, 1999 
	if ($crit->type eq 'GEO') {                   # special for GEO
            my $crit_odl = $crit->to_request_ODL;
	    my $loc;
	    foreach $loc (qw(POINT_LOC RANGE_LOC POLYGON_LOC)) {
		if ($crit_odl->param($loc)) {
		    $opts_odl->add_param($loc, $crit_odl->param($loc));
		    last;
		}
	    }
	}

	# Output the criterion if it is not a SUBPAGE type.:
        $opts_odl = $crit->to_request_ODL($opts_odl) 
         if($crit->type ne 'SUBPAGE');
    }

    # Done!
    $opts_odl;
}
    

#------------------------------------------------------------
1;
