package IU::OrderStatusRequest;

=head1 NAME

IU::OrderStatusRequest - user request for Order Status retrieval.


=head1 SYNOPSIS

Search options/properties:
    
    # What kind of search is this?  Can return any of...
    #     
    #     PRODUCT_STATUS_REQUEST
    $q->search_type;
    
    # What kind of ODL search message is this *really*?  Can return any of...
    #     
    #     PRODUCT_STATUS_REQUEST
    $q->ODL_search_type;

ODL interface:

    # ODL interface:
    $q = IU::OrderStatusRequest->from_ODL($odl);
    $odl = $q->to_ODL;
     

=head1 DESCRIPTION

This object encapsulates order status requests.

=cut

use IU;
use IU::Log;
use IU::IW;
use IU::Category;
use IW::App;
use IW::User;

use strict;
use vars qw(@ISA @EXPORT
	    $MAX_TIME_DEFAULT
            $ORDER_STATUS_REQUEST);
@ISA = qw(Exporter);
@EXPORT = qw($ORDER_STATUS_REQUEST);

#------------------------------
#
# Globals...
#
#------------------------------

# The current query (EXPORTED for the convenience of other modules):
$ORDER_STATUS_REQUEST = undef;

# Default time limit for searches
$MAX_TIME_DEFAULT = $IU::SETUP::MaxSearchTime;
# Categories which are output in the ODL as top-level sequence strings
# under the category name (except for DATA_CENTER_ID, which is handled 
# separately):
my %SeqStringCategories = map {$_ => 1}
 qw(
   DATA_CENTER_ID
   REQUEST_ID
);
# Categories which are output in the ODL as top-level simple strings
# under the category name:
my %StringCategories = map {$_ => 1}
 qw(
   TEXT_STRING
);
#------------------------------
#
# Construction...
#
#------------------------------

#------------------------------------------------------------
# new 
#------------------------------------------------------------
# Class method.  Constructor.

sub new {    
    my $class = shift;
    my $searchType = shift;
    my $self = {};
    $self->{SearchType} = $searchType;
    $self->{DaacsToSearch} = [];
    $self->{DaacsNeedingECSAuth} = [];
    $self->{DaacsNeedingV0Auth}  = [];
    $self->{DaacsNeedingNASDAAuth}  = [];
    bless $self, $class;
}

#------------------------------
#
# Private utilities...
#
#------------------------------

#------------------------------------------------------------
# _one VARNAME,SELF,[VALUE]
#------------------------------------------------------------
# I<Private utility.>  Get/set a single-valued instance variable.
# Return undef if it's undefined.
sub _one {
    my ($varname, $self, $value) = @_;

    $self->{$varname} = $value if (@_ > 2) ;
    $self->{$varname};   
}

#------------------------------------------------------------
# _many VARNAME,SELF,[ARRAYREF]
#------------------------------------------------------------
# I<Private utility.>  Get/set a multi-valued instance variable.
# Set it by providing an arrayref: the array will be duplicated.
#
# If invoked with no args, just returns info.
# Returns the array of all values in an array context, and just 
# the array size in a scalar context.

sub _many {
    my ($varname, $self, $values) = @_;
    if (@_ > 2) {
	my @array = @$values;
	$self->{$varname} = \@array;
        

    }
    if (defined @{$self->{$varname}} ){
     (wantarray ? @{$self->{$varname}} : int(@{$self->{$varname}}));
    } else {
      return undef;
    } 
}
#------------------------------------------------------------
# intersect ARRAYREF, ARRAYREF
#------------------------------------------------------------
# Utility: intersect two lists of scalars, and return the result.

sub intersect {
    my ($a, $b) = @_;
    my %in_a = map {$_=>1} @$a;
    return sort(grep {$in_a{$_}} @$b);
}

#------------------------------------------------------------
# daacs_to_search
#------------------------------------------------------------
# I<Instance method.>
# Return the list of data centers to search.
# This will be empty until set up by determine_daacs_to_search().

sub daacs_to_search { _many('DaacsToSearch', @_) }

#------------------------------------------------------------
# daacs_needing_ecs_auth
#------------------------------------------------------------
# I<Instance method.>
# Return the list of data centers which according the wwwvalids 
# need the ECS authenticator.
# This will be empty until set up by determine_daacs_to_search().

sub daacs_needing_ecs_auth { _many('DaacsNeedingECSAuth', @_)}

#------------------------------
#
# Search options/properties...
#
#------------------------------

#------------------------------------------------------------
# max_time [VALUE]
#------------------------------------------------------------
# Set/get maximum number of minutes that a search can run.
#

sub max_time {
    my ($self, $value) = @_;
    $self->{MaxTime} = $value if @_ > 1;
    $self->{MaxTime} || $MAX_TIME_DEFAULT;
}

#------------------------------------------------------------
# name [NAME]
#------------------------------------------------------------
# Set/get the name of the query.
#
sub name  { _one('Name', @_) }

#------------------------------------------------------------
# ODL_search_type
#------------------------------------------------------------
# Return the search type, as it would appear in the ODL.
#
sub ODL_search_type { 
    my $self = shift;
    my $type = $self->search_type;
    return $type;
}

#------------------------------------------------------------
# search_type [VALUE]
#------------------------------------------------------------
# Set/get the search type, in the most-detailed sense.
# Compare with ODL_search_type.
#
sub search_type { return 'PRODUCT_STATUS_REQUEST' }

#-------------------------------------------------------------
# request_ids [VALIDS]
# data_center_ids [VALIDS]
#-------------------------------------------------------------
# Convenient ways to get/set the array of valids for a core category.
# Accepts an optional arrayref of strings to set, and 
# returns an array of strings.
#
sub request_ids       { my $self = shift; 
			$self->criteria_strings('REQUEST_ID',@_)}

sub data_center_ids   { my $self = shift; 
			$self->criteria_strings('DATA_CENTER_ID',@_)}
#------------------------------------------------------------
# criteria CATEGORY, [CRITERIA]
#------------------------------------------------------------
# Instance method, for convenience.
#
# With just CATEGORY, returns array of criteria (or empty array if none).
#
# With CATEGORY and CRITERIA, sets CATEGORY's criteria to be the
# given CRITERIA, which should be a ref to an array of Criterion objects.
# If you supply a non-arrayref, it will be assumed to be the sole criterion.  
#
# Use a CRITERIA of undef (or a ref to the empty array) to erase all 
# criteria for this category (and the category as well).

sub criteria {
    my ($self, $cat, $values) = @_;
    if (@_ > 2) {             # set new values
	my @safe;             # safe place for new values

	# Coerce values to be either empty, or array of one criterion:
	$values ||= [];                          # undef/false to empty array
	@safe = ((ref($values) eq 'ARRAY') ? @$values : ($values));
	
	# Set:
	(@safe ? $self->{Categories}{$cat} = \@safe 
	       : delete $self->{Categories}{$cat});
    }
    @{$self->{Categories}{$cat} || []};         # return
}
#------------------------------------------------------------
# criteria_strings CATEGORY, [STRINGS]
#------------------------------------------------------------
# Instance method, for convenience.
#
# With just CATEGORY, returns array of values (or empty array)
# held in the first criterion, which is assumed to be a "SelectMany".
#
# With CATEGORY and STRINGS, where STRINGS is a ref to an array of strings,
# creates a new "SelectMany" criterion with the given values, and stores 
# it as the only criterion for this category.

sub criteria_strings {
    my ($self, $cat, $strings) = @_;
    my $crit;            # note: initially undefined!

    # Set new values?
    if (@_ > 2) {    
	if (@$strings) {       # some... create and set criteria!
	
	    $crit = IU::Criterion::SelectMany->new;
	    $crit->field('VALUES', $strings);
	}
	$self->criteria($cat, $crit);
    }

    # Return them:
    $crit = ($self->criteria($cat))[0];          # get first criterion
    return ($crit ? @{$crit->field('VALUES') || []} : ()); 

}

#------------------------------------------------------------
# filter_daacs_to_search
#------------------------------------------------------------
# I<Instance method, private.>
# Take the list of data centers to actually query, and pare them down.
# based on the static valids file.  See determine_daacs_to_search().
#
# Raises a fatal exception if no DAACs remain.

sub filter_daacs_to_search {
    my $self = shift;

    my @initialDaacs = $self->daacs_to_search;

    my @filteredDaacs;

    # Open static valids database:
    my %WWWVALIDS;
    if (-e "${IU::LibDir}/conf/wwwvalids.gdb") {
        eval {tie %WWWVALIDS, 'IU::Store', "${IU::LibDir}/conf/wwwvalids.gdb",
              'READER'};
        $@ and return error "tie failed: $@";
    } 

 
    my %needsECS;         # Daacs needing ecs authenticator.
    my %needsV0;          # Daacs needing V0 authenticator.
    my %needsNASDA;       # Daacs needing NASDA authenticator.

    my $daac;
    foreach $daac (@initialDaacs) {
        # Is there an entry for it in the static valids?
        my $valids;


        if ($valids = $WWWVALIDS{"daac:$daac"}) {

            $needsECS{$daac} = $valids->{REQUIRES_ECS_AUTH};
            $needsV0{$daac} = $valids->{REQUIRES_V0_AUTH};
            $needsNASDA{$daac} = $valids->{REQUIRES_NASDA_AUTH};
        }                       # end if
    }                                    # end: foreach daac

    my @daacsECS = grep { $needsECS{$_}} (keys %needsECS);
    my @daacsV0 = grep { $needsV0{$_}} (keys %needsV0);
    my @daacsNASDA = grep { $needsNASDA{$_}} (keys %needsNASDA);
    $self->daacs_needing_ecs_auth(\@daacsECS);
    $self->daacs_needing_v0_auth(\@daacsV0);
    $self->daacs_needing_nasda_auth(\@daacsNASDA);
    # Close static valids database:
    untie %WWWVALIDS;

    # Set new list of daacs to search:
    @filteredDaacs = @initialDaacs;


    $self->daacs_to_search(\@filteredDaacs);
}
#------------------------------------------------------------
# depvalid_interface [OPTIONS]
#------------------------------------------------------------
# I<Instance method.>
# Return a dependent-valids interface object preset with the 
# query criteria 
#
# The OPTIONS is a hash which can contain the following:
#
#

sub depvalid_interface {
    my ($self) = @_;
   require IW::DepValid;  
    # Create a new dependent valids interface:
    my $depvalid = new IW::DepValid ;

    # Set info:
    my $cat = "DATA_CENTER_ID";

    # Get the valids, as appropriate:
    my @valids = ();
    
   @valids = $self->criteria_strings($cat); 

    # Set them:
    $depvalid->set($cat, \@valids);              # set them
    
    $depvalid;
}

#------------------------------------------------------------
# determine_daacs_to_search
#------------------------------------------------------------
# I<Instance method.>
# Determine list of data centers to actually query, using dependent valids.
# Fatal exception if no DAACs can be set.

sub determine_daacs_to_search {
    my $self = shift;

    # Get a dependent-valids interface object for this query:
    require IU::Store;
    my $depval = $self->depvalid_interface(Exclude=>'DATA_CENTER_ID');

    # STEP 1: Set the data centers, as dictated solely by dependent valids:
    my @validDaacs = $depval->get_valids('DATA_CENTER_ID');

    @validDaacs or 
	puke 517, "Your search attributes are all incompatible!",
        " Go back to the Search Form and make compatible selections!";

    $self->daacs_to_search(\@validDaacs);

    # STEP 2: Intersect with the list of *desired* daacs, if any were given:
    if ($self->data_center_ids) {      # user specified some daacs...

	my @finalDaacs = intersect([$self->data_center_ids], 
				   [$self->daacs_to_search]);
	@finalDaacs or puke 517, "You left me with no data centers to search!";

	$self->daacs_to_search(\@finalDaacs);

    }

   # STEP 3: Filter the data centers, based on the static valids:
    $self->filter_daacs_to_search;

    ($self->daacs_to_search) or
        puke 460, "after filtering, no searchable data centers remained!";
}

#------------------------------------------------------------    
# incomplete
#------------------------------------------------------------    
# I<Instance method.>
# Based on the search type, are we missing some required fields?
# If so, returns a string (true) describing what's missing.
# If not, returns 0 (false).

sub incomplete {
    my $self = shift;
    
    if ($self->search_type eq 'PRODUCT_STATUS_REQUEST') {
	unless ($self->data_center_ids) {
	    unless ($self->request_ids) {
		return "You must supply at least one Data Center and one ".
		       "Order Tracking Number.";
	    }
	    return "You must supply at least one Data Center.";
	}
	($self->request_ids)
	    or return "You must supply an Order Tracking Number.";
    }
   else {
	puke "Unknown search type (", $self->search_type, ")!";
    }
    return 0;
}

#------------------------------------------------------------    
# install
#------------------------------------------------------------    
# I<Instance method.>  WG-specific.
# Re-install this as the current query in the session state.

sub install {
    my $self = shift;
    $SESSION->state->{OrderStatusRequest} = $self;
}

#------------------------------------------------------------    
# lookup
#------------------------------------------------------------ 
# I<Class method.  WG-specific.>
# Lookup current query in the session state.  Returns undef if none so far.

sub lookup {
    my $self = shift;
    my $query = $SESSION->state->{OrderStatusRequest};
    return $query;
}

#------------------------------------------------------------
# uses CATEGORY, [SEARCHTYPE]
#------------------------------------------------------------
# I<Class/instance method.>
# Returns true if the given CATEGORY would be a functional part of the 
# query under the given SEARCHTYPE, and false otherwise.   With
# no SEARCHTYPE, will use the query's search type.

sub uses {
    my ($self, $cat, $type) = @_;
    $type ||= $self->search_type;

    if ($type eq 'PRODUCT_STATUS_REQUEST') {
	return 1 if ($cat eq 'DATA_CENTER_ID');
	return 1 if ($cat eq 'REQUEST_ID');
	return 0;
    }  
    else {         # any other kind of search
	return 1;        
    }
}

#------------------------------------------------------------
# uses_opt OPTION, [SEARCHTYPE]
#------------------------------------------------------------
# I<Class/instance method.>
# Much like uses(), but for miscellaneous query options that aren't
# really categories.  Valid OPTIONs are:
#
#     QUERY_NAME       Query name
#
# Returns true if the given OPTION would be a functional part of the 
# query under the given SEARCHTYPE, and false otherwise.   With
# no SEARCHTYPE, will use the query's search type.
# Always show guide limit since it will take effect if the user selects
# Detailed Info while doing any kind of search.

sub uses_opt {
    my ($self, $opt, $type) = @_;

    $type ||= $self->search_type;
    if (($type eq 'PRODUCT_STATUS_REQUEST')) {
	return 1 if ($opt eq 'QUERY_NAME');
	return 0;
    }
    else {         # any other kind of search
	return 1;        
    }
}
#------------------------------------------------------------    
# validate
#------------------------------------------------------------    
# I<Instance method.>
# Based on the search type, are we ready to go?
# This is an assertion: it returns the self object on success, but on failure
# it raises a fatal exception which describes why validation failed. 

sub validate {
    my $self = shift;

    # Is user not done?
    my $problem = $self->incomplete;
    puke 516, $problem if $problem;
    $self;
}
#------------------------------
#
# ODL interface
#
#------------------------------

#------------------------------------------------------------
# from_ODL ODLOBJECT, PARAMHASH
#------------------------------------------------------------
# I<Class or instance method.>
# Initiallize the address from a standard IMS ODL object.
#
# The PARAMHASH can contain the following:
#
#     KeepType => If true, the search type will be discarded in 
#                 favor of the type of any CURRENTLY-INSTALLED query.
#                 (See lookup() and install() for details).
#                 If false (default), the ODL alone is used in determining
#                 the search type. 

sub from_ODL {
    my ($self, $odl, %params) = @_;
    $odl || return undef;
    my ($param, $value);
    my $search;
    my $keep_type = $params{KeepType};

    # Be a class method, too!
    ref($self) or $self = $self->new;
    #
    # DETERMINE THE SEARCH TYPE...
    #
    #    If "keep type" is true, determine the search type as given by 
    #    any currently-installed query, and prefer that value to the 
    #    actual one given by the ODL.
    #
    #    This is done so that when the user is UPLOADING an ODL search file,
    #    we automatically coerce it into the type of search form 
    #    (INVENTORY, DIRECTORY, GRANULE, etc.) that they are uploading into.

    # Determine the search type as given by the currently-installed query:
    my $prev_search_type;
    if ($keep_type) {
	my $prev_query = ref($self)->lookup;        # don't put it in $QUERY!
	$prev_search_type = $prev_query && $prev_query->search_type;
    }

    # Determine the search type as given by the ODL itself.  
    #    NOTE: this is not necessarily the actual search type of the
    #    top-level message... rather, it is the search type as derived
    #    from the ODL itself.  Notice the case for GRANULE_SEARCH.
    my $new_search_type;
    if ($search = $odl->param('PRODUCT_STATUS_REQUEST')) {
	$new_search_type = 'PRODUCT_STATUS_REQUEST';
    }
    else {
        puke "ODL doesn't seem to hold  PRODUCT_STATUS_REQUEST ";
    }

    # Now, resolve the two search types, preferring the CGI input:
    $self->search_type($prev_search_type || $new_search_type);

    # 
    # EXTRACT THE SEARCH CONSTRAINTS FROM THE ODL...
    #
    $search->rewind;
    my %queue;

    while (($param, $value) = $search->next) {

	if ($param eq 'TIME_LIMIT') {
	    $self->max_time($value->text_value);
	}
	elsif ($param eq 'SEARCH_NAME') {
	    $self->name($value->text_value);
	}
	else {
	    warn "skipping unrecognized search parameter: '$param'";
	}
    }    
    $self;
}
#------------------------------------------------------------
# categories
#------------------------------------------------------------
# I<Instance method.>  Return all categories in query.

sub categories {
    my $self = shift;
    keys %{$self->{Categories}};
}
#------------------------------------------------------------
# daacs_needing_v0_auth
#------------------------------------------------------------
# I<Instance method.>
# Return the list of data centers which according the wwwvalids 
# need the V0 authenticator.
# This will be empty until set up by determine_daacs_to_search().

sub daacs_needing_v0_auth { _many('DaacsNeedingV0Auth', @_)}

#------------------------------------------------------------
# daacs_needing_nasda_auth
#------------------------------------------------------------
# I<Instance method.>
# Return the list of data centers which according the wwwvalids 
# need the NASDA authenticator.
# This will be empty until set up by determine_daacs_to_search().

sub daacs_needing_nasda_auth { _many('DaacsNeedingNASDAAuth', @_)}

#------------------------------------------------------------
# print_HTML HTMLSTREAM
#------------------------------------------------------------
# Output summary as pretty HTML text.

sub print_HTML {
    my $self = shift;
    my $HH = shift;    # HTML stream

    $HH -> UL;
    my ($cat, $val, $crit);
    foreach $cat (sort {category2pretty($a) cmp category2pretty($b) 
			} $self->categories) {

	# Skip this category if not used by current search type:
	$self->uses($cat) or next;

	# Get criteria:
	my @criteria = $self->criteria($cat); @criteria or next;
	my $pretty_cat = category2pretty($cat);

	$HH -> LI -> t_B($pretty_cat, ": ");
	if (IU::Category->is_singleton($cat)) {   # has one criterion/variant
	    $crit = $criteria[0];
	    $HH -> t($crit->as_string);
	}
        else {                                    # has multiple criteria
	    $HH -> UL;
	    foreach $crit (@criteria) {
		$HH -> LI -> t_B($crit->name, ": ") -> t($crit->as_string);
	    }
	    $HH -> _UL;
	}
    }
    $HH -> _UL;
    1;
}
#------------------------------------------------------------
# to_ODL PARAMHASH
#------------------------------------------------------------
# Convert query to an ODL object, and return it.
# The ODL object is a group *containing* an INVENTORY_SEARCH group
# (or DIRECTORY_SEARCH, etc.).
#
# No authentication information is placed in the ODL!!!!
#
# PARAMHASH may contain the following:
#
#    NoDaacs   => If true, then the DATA_CENTER_ID group (user-specified
#                 data centers) will *not* be output in the ODL.
#                 Generally, we handle these separately.
#
#    Everything=> By default, ODL is only output for the current search type,
#                 meaning that if an attribute is not used in the current
#                 search type, then it will not appear in the ODL.
#                 If you set this option to true, then *all* criteria will 
#                 be included in the ODL.
#
#    ForSearch => If true, ODL is tweaked to default/modify certain fields
#                 (e.g., stop date) to make DAACs happy.  Normally,
#                 we don't want the user's query criteria mucked with,
#                 so the default is false.
#
# NOTE: for downloading, use the Everything option to output ALL the 
# categories to the ODL, even ones which are unused for this type of search.  
# This is so an inventory search may be uploaded, turned into a granule-id 
# search, and downloaded, but still retain all the inventory search 
# characteristics.
#
# NOTE: the ForSearch option produces *slightly* different ODL.  Normally,
# we return an accurate, uploadable representation of the query.  With
# ForSearch, we ensure that certain ODL fields are defaulted, to deal 
# with the less-tolerant DAACs.

sub to_ODL {
    my ($self, %params) = @_;
    my $no_daacs         = $params{NoDaacs};
    my $write_everything = $params{Everything};
    my $for_search       = $params{ForSearch};

    require IU::ODL;

    # Begin:
    my $root_odl = IU::ODL->new_group;    # outermost group
    my $srch_odl = IU::ODL->new_group;
    my $extsrch_odl;                      # will be created only if needed

    # Start off the file with the search type:
    $root_odl->add_param($self->ODL_search_type => $srch_odl);

    # The message id (do we REALLY need this?)
    $srch_odl->add_param(MESSAGE_ID    => '0'); 

    # The criteria:
 
    # The user's desired request_ids...
    my @rids = $self->request_ids;

    $srch_odl->add_param('REQUEST_ID' => IU::ODL->conv(\@rids));

    # The user's desired data centers...
    my @values = $self->data_center_ids;

    $srch_odl->add_param('DATA_CENTER_ID' => IU::ODL->conv(\@values));

    # Done!
    $root_odl;
}

#------------------------------------------------------------
# to_search_ODL
#------------------------------------------------------------
# Convert query to an ODL object suitable for input to the client, 
# and return it. The ODL object is a group containing an INVENTORY_SEARCH 
# parameter.
#
# Unlike a simple to_ODL, this method uses the dependent valids mechanism
# to determine which DAACs to search

sub to_search_ODL {
    my $self = shift;

    # Start by figuring out which DAACS to search, if needed:

    $self->daacs_to_search
	or $self->determine_daacs_to_search
	    or puke "no data centers to search!";  
    # Start with the basics, *without* the DAACs:
    my $root_odl = $self->to_ODL(NoDaacs=>1, ForSearch=>1) 
	|| return undef;

    # Get the actual search group, checking all standard types:
    my $srch_odl = $root_odl->param('DIRECTORY_SEARCH') ||
	           $root_odl->param('INVENTORY_SEARCH') ||
		   $root_odl->param('GUIDE_SEARCH') ||
		   $root_odl->param('PRODUCT_STATUS_REQUEST') ||
		   $root_odl->param('ACCOUNT_STATUS_REQUEST') ||
		   puke "No recognized search group in ODL!"; 

    # Add the authenticator, if any:
    my $authenticator = eval { $SESSION->state->{User}->authenticator };
    if ($authenticator) {
	$srch_odl->add_param(AUTHENTICATOR => IU::ODL->conv($authenticator));
    }

    # Add the NASDA authenticator, if any:
    my $nasda_authenticator =
     eval {$SESSION->state->{User}->nasda_authenticator};
    if ($nasda_authenticator) {
        $srch_odl->add_param(NASDA_AUTHENTICATOR =>
                             IU::ODL->conv($nasda_authenticator));
    }

    # Check for ECS fields, if we are using the ecs authenticator:
    # Modified to go to SMC instead of home DAAC.
    unless ($IU::EcsProfileSmc eq '') {
       # Add the ecs authenticator, if any:
       my $ecs_authenticator = eval { $SESSION->state->{User}->ecs_authenticator };
       if ($ecs_authenticator) {
           $srch_odl->add_param(ECS_AUTHENTICATOR => IU::ODL->conv($ecs_authenticator));
       }
    }


    # Done!
    return $root_odl;
}

#------------------------------------------------------------    
1;
