package IU::Package;

=head1 NAME

IU::Package - available package for a granule


=head1 SYNOPSIS

    # Lookup packages for this granule:
    @packages = IU::Package->lookup($gkey);
   
    # Get the IU::PkgCombo objects (kind of like LINE_ITEMs) for a package.
    #   Basically, this produces all possible combinations of processing
    #   option, media type, and media format in the PACKAGE "tree":
    @combos = $package->combos;

=head1 DESCRIPTION

Each instance represents roughly the information in a PACKAGE group,
including the contents of any *_OPTIONS groups (represented in this
object via IU::Options objects).

=head1 DATA STRUCTURES

Basically, during package selection, we have this:

     .-----------.
     | Package 1 |------------------.
     `-+---------'                  V
       |                          .----------.----------.----------.
       |                          | Package 1| Package 1| Package 1|
       |    .--------------.      | Subset   | Order    | DSOrder  |
       |--->| PkgCombo 1.1 |      | Options  | Options  | Options  |
       |    |--------------|   	  `-+--------'-+--------'-+--------'
       |--->| PkgCombo 1.2 |        |    .-------------.
       |    |--------------|	    |--->| Criterion   |
       `--->| PkgCombo 1.3 |        |    |-------------|
            `--------------' 	    |--->| Criterion   |
				    |    |-------------|
				    `--->| Criterion   |
				         `-------------'

After packaging options have been chosen, though, only the
IU::PkgCombo objects are retained with the orderable item (e.g.,
the granule), and we have the following tree structure:

     .---------.
     | Granule |
     `-+-------'
       |                          .----------.----------.----------.
       |                          | Package 1| Package 1| Package 1|
       |    .--------------.      | Subset   | Order    | DSOrder  |
       |--->| PkgCombo 1.1 |----->| Options  | Options  | Options  |
       |    |--------------|      `-+--------'-+--------'-+--------'
       |--->| PkgCombo 2.4 |- -     |    .-------------.
       |    |--------------|	    |--->| Criterion   |
       `--->| PkgCombo 2.5 |- -     |    |-------------|
            `--------------'  	    |--->| Criterion   |
				    |    |-------------|
				    `--->| Criterion   |
				         `-------------'

The latter configuration accurately reflects the fact that different
packaging combinations for a granule may have completely different
user-selections for subsetting criteria, since the combos we store for
a granule may come from completely different packages (in the above
case, pkgs 1 and 2).

=head1 RCS INFO

$Id: Package.pm,v 1.19 2002/10/15 21:00:10 seufert Exp $

=cut

use IU;
use IU::Log;
use IU::PkgCombo;
use IU::Chunk;
use IU::FancyURL;
use IU::Granule;
use IU::Dataset;
use IU::Options;
use IU::Session;
use IW::App;
use IU::Util qw(shellquote);

#------------------------------
# Simple access methods...
#------------------------------

# Return the archive id:
sub archive_id { shift->{Chunk}->value('DATA_CENTER_ID') || '' }

# Return the package id, as in PACKAGE_ID.
sub id      { shift->{Id} }

# Return the COMMENT, or the empty string:
sub comment     { shift->{Chunk}->value('COMMENT') || '' }

# Return the TITLE, or the empty string;
sub title       { shift->{Title} }

# return the number of line item allows per granule or the empty string;
sub line_items_per_item { shift->{LineItemsPerItem} }

# Return the INFO_PROMPT, or the empty string:
sub info_prompt     { shift->{Chunk}->value('INFO_PROMPT') || '' }

# Return the number of granules, as in NUMBER_OF_GRANULES
sub n_granules  { shift->{Chunk}->value('NUMBER_OF_GRANULES') || '' }

# Can user select multiple media types for this package?
sub allows_multiple_media {
    my ($self, $value) = @_;
    $self->{AllowMultiMedia} = $value if @_ > 1;
    $self->{AllowMultiMedia};
}
# Return ref to array of misc URLs (as IU::FancyURLs), or []
sub misc_URLs    { $_[0]->{MISC_URL} }

#------------------------------------------------------------
# new
#------------------------------------------------------------
sub new {
    my $class = shift;
    my $self = {
	Options=>{},
    };
    bless $self, $class;
}

#------------------------------------------------------------
# combos 
#------------------------------------------------------------
# I<Instance method.>
# Return the components of this package as an array of package combos.

sub combos {
    my ($self) = @_;
    my @combos;
    my $key;

    # Iterate through the chunk components:
    my ($package_id, $package_chunk);
    my ($option_id,  $option_chunk);
    my $package_size;
    my ($type_id,    $type_chunk);

    # Switch to turn off Handling Fee.
    my $apply_handling_fee;
    my ($format_id,  $format_chunk);
    $package_chunk = $self->{Chunk};
    $package_id    = $self->{Id};

    ### debug "PI $package_id";
    foreach $option_chunk ($package_chunk->values('PROCESSING_OPTION')) {
	$option_id = $option_chunk->value('OPTION_ID');
	### debug "PO $option_id";
	$package_size = $option_chunk->value('PACKAGE_SIZE');
	### debug "PO $package_size";
	foreach $type_chunk ($option_chunk->values('MEDIA_TYPE')) {
	    $type_id = $type_chunk->value('TYPE_ID');
	    ### debug "MT $type_id";

             
            $apply_handling_fee = $type_chunk->value('APPLY_HANDLING_FEE');

	    foreach $format_chunk ($type_chunk->values('MEDIA_FORMAT')) {
		$format_id = $format_chunk->value('FORMAT_ID');
		### debug "MF $format_id";

		# Make a shallow copy:
		my $combo_chunk = new IU::Chunk;
		foreach $key ($format_chunk->keys) {
		    $combo_chunk->set($key => $format_chunk->values($key));
		}

		# Annotate it with "inherited" info:
		$combo_chunk->set(PACKAGE_ID => $package_id);
		$combo_chunk->set(OPTION_ID  => $option_id);
                $combo_chunk->set(PACKAGE_SIZE => $package_size);
		$combo_chunk->set(TYPE_ID    => $type_id);
		$combo_chunk->set(FORMAT_ID  => $format_id);
                $combo_chunk->set(APPLY_HANDLING_FEE => $apply_handling_fee);

		# Create the combo:
		my $combo = IU::PkgCombo->from_chunk($combo_chunk);
		$combo or do {error("unable to create combo!"); next};
	
                #Put in the package title 
                $combo->package_title($self->{Title});

		# If the package has been tied to a granule, tie the combo:
		$combo->granule_key($self->granule_key) if $self->granule_key;

		# Add it:
		push @combos, $combo;
	    }
	}
    }
    # Done!
    @combos;
}

#------------------------------------------------------------
# granule_key [VALUE]
#------------------------------------------------------------
# Get/set the granule key for this package, effectively "tying" this
# package to a granule.  Packages need not be tied; a package like 
# "1/234/*" is totally legal.
#
# WARNING: don't even THINK about storing a pointer to the granule!
# You'll create a pointer cycle, and the granule won't be storable.

sub granule_key {
    my ($self, $gkey) = @_;
    $self->{GKey} = $gkey if (@_ > 1);
    $self->{GKey};
}

#------------------------------------------------------------
# addn_info 
#------------------------------------------------------------
# Get/set the Addition information field.

sub addn_info {
    my ($self, $addn_info) = @_;
    $self->{Addn_Info} = $addn_info if (@_ > 1);
    $self->{Addn_Info};
}

#------------------------------------------------------------
# options OPTSTYPE, [OPTIONS]
#------------------------------------------------------------
# Set/get the options named WHICH.  Possibilities for OPTSTYPE
# are "SUBSET", "ORDER", and "DATASET_ORDER".
#
# If no arguments are given, returns a list of OPTSTYPEs which
# have values in this object.

sub options {
    my ($self, $type, $opts) = @_;
    $type or return grep {$self->{Options}{$_}}(keys %{$self->{Options}});
    $self->{Options}{$type} = $opts if @_ > 2; 
    $self->{Options}{$type};
}

#------------------------------
#
# CHUNK INTERFACE...
#
#------------------------------

#------------------------------------------------------------
# _add_info
#------------------------------------------------------------
# I<Instance method, private.>
# Add info to the package, from the WWWVALIDS file.

sub _add_info {
    my $self = shift;

    # Open file:
    if (-e "${IU::LibDir}/conf/wwwvalids.gdb") {
	eval {tie %WWWVALIDS, 'IU::Store',
	      "${IU::LibDir}/conf/wwwvalids.gdb", 'READER'};
	$@ and return error "tie failed: $@";
    }
    # Get entry:
    my $archid = $self->archive_id; 
    $valids = $WWWVALIDS{"daac:$archid"} or return undef;

    # Media:
    unless (defined($self->allows_multiple_media)) {
	my $amm = $valids->{ALLOWS_MULTIPLE_MEDIA};
#	debug("WWWVALIDS: multi media = ", ($amm ? 'yes' : 'no'));
	$self->allows_multiple_media($amm);
    }
    untie %WWWVALIDS;
} 

#------------------------------------------------------------
# from_chunk CHUNK
#------------------------------------------------------------
# I<Class/instance method.>
# Read from a chunk.

sub from_chunk {
    my ($self, $chunk, $granule) = @_;
    my $subchunk;

    # Allow instance or class method:
    ref($self) or $self = $self->new;

    # Init:
    $self->{Chunk} = ($chunk->value('PACKAGE') ? $chunk->value('PACKAGE') 
		                               : $chunk);
    
    # Simple attributes:
    $self->{Id}              = $self->{Chunk}->value('PACKAGE_ID');
    $self->{AllowMultiMedia} = $self->{Chunk}->value('ALLOWS_MULTIPLE_MEDIA');

    # Get title or make it null string.
    $self->{Title} = $self->{Chunk}->value('TITLE') || '';
    $self->{LineItemsPerItem} = 
       $self->{Chunk}->value('LINE_ITEMS_PER_GRANULE') || '';

    # Options:
    my $opts_type;
    foreach $opts_type (qw(SUBSET
			   ORDER
			   DATASET_ORDER)) {
	if ($subchunk = $self->{Chunk}->value("${opts_type}_OPTIONS")) {
#	    debug "Found options $opts_type in package $self->{Id}";
	    $self->options($opts_type, 
	       IU::Options->from_chunk($subchunk, $opts_type,$granule));
	}
    }
    # Add any MISC_URL groups.
    my $key = 'MISC_URL';

    # Allows multiple URLs:
    $self->{$key} = [];

    my ($url_chunk, $furl);
    foreach $url_chunk ($self->{Chunk}->values($key)) {
       my $furl = IU::FancyURL->from_chunk($url_chunk, $key);
       push @{$self->{$key}}, $furl;

    }
    # Add other info, from WWWVALIDS file:
    $self->_add_info;
    $self;
}

#------------------------------------------------------------
# lookup GKEY, [PARAMHASH]
#------------------------------------------------------------
# I<Constructor, of sorts.>
# Get and return a list of packages for this granule, or the empty
# list if none were found.
#
# Since some info (e.g., min and max range values) should be init'ed
# from the current dataset/granule, as a KLUDGE we take the "context"
# in which this call is being made:
#
#     Granule => the granule object whose GKEY this is
#     Dataset => its dataset object

sub lookup {
    my $class = shift;
    my $gkey = shift;
    my %params = @_;
    
    # Set context (KLUDGE):
    ### $IU::Options::DatasetContext = $params{Dataset}; 
    ### $IU::Options::GranuleContext = $params{Granule};

    my $subdir = $params{SubDir};
    my $file = "package_ids";
    $file = $subdir . '/' . $file if($subdir);
    my $path = $SESSION->path_to($file);
    open OUT, ">$path" or puke 452, "can't open $path for writing: $!";
    my $granule = $params{Granule};
    my $packageIdRef = $params{PackageIds};
    my @package_ids = @{$granule->package_ids};
    @package_ids = @$packageIdRef if($packageIdRef);
    my $pkid;
    foreach $pkid (@package_ids) {
#      debug "package Id = $pkid";
      print OUT "$pkid\n";
    }
    close OUT;
    # Open "imsgdbdump" stream:
    my $command = "${IU::LibDir}/bin/imsgdbdump ptg " . shellquote($gkey);
    $command .= " " . $subdir if($subdir);
    inform "lookup packages; command = $command";	
    open(IN, "$command |") or puke 453, "couldn't open pipe on $command";

    # Read chunks:
    my $chunk;
    my @packages;
    while ($chunk = IU::Chunk->input(\*IN)) {
        # get or skip
	my $package = $class->from_chunk($chunk,$granule) || next;
	$package->granule_key($gkey);              # tie to granule
	push @packages, $package;                  # save it!
    }

    # Close stream and return:
    close IN;           # UNFINISHED: check to see if command failed!

    # Sort them by package id, and return them:
    @packages = sort { $a->{Id} cmp $b->{Id} } @packages;
    return @packages;
}

#------------------------------------------------------------
# lookup_dataset DSKEY, [PARAMHASH]
#------------------------------------------------------------
# I<Constructor, of sorts.>
# Get and return a list of packages for this data set for data set ordering,
# or the empty
# list if none were found.
#
# Since some info (e.g., min and max range values) should be init'ed
# from the current dataset/granule, as a KLUDGE we take the "context"
# in which this call is being made:
#
#     Dataset => its dataset object

sub lookup_dataset {
    my $class = shift;
    my $dskey = shift;
    my %params = @_;
    my @packages;

    my $dset = $params{Dataset};
    my $subdir = $params{SubDir};
    
    # Set context (KLUDGE):
    ### $IU::Options::DatasetContext = $params{Dataset}; 
    my @package_ids = @{$dset->package_ids};

    my $pkid;
    foreach $pkid (@package_ids) {
        # Open "imsgdbdump" stream:
        my $command = join(" ", 
                           "${IU::LibDir}/bin/imsgdbdump", 
                           "ptd", shellquote($dskey),
                           shellquote($pkid));
        $command .= " " . $subdir if($subdir);
        inform "lookup packages; command = $command";	
        open(IN, "$command |") or
            puke 453, "couldn't open pipe on $command";

        # Read chunks:
        my $chunk;
        while ($chunk = IU::Chunk->input(\*IN)) {
            # get or skip
            my $package = $class->from_chunk($chunk,$granule) || next;
            push @packages, $package;                      # save it!
        }

        # Close stream and return:
        close IN;        # UNFINISHED: check to see if command failed!
    }

    # Sort them by package id, and return them:
    @packages = sort { $a->{Id} cmp $b->{Id} } @packages;
    return @packages;
}

#------------------------------------------------------------
1;
