package IU::PkgCombo;

=head1 NAME

IU::PkgCombo - a tuple of package, processing option, media type, and
               media format, possibly with other info.

=head1 SYNOPSIS

   # Read in from a chunk:
   my $combo = IU::PkgCombo->from_chunk($chunk);
   
   # Associate it with a granule:
   $combo->granule_key($gran->key);

=head1 PACKAGE IDS

To understand this module, you must first know that a granule can have
two kinds of "package id" associated with it:

=over *

If the package id is a "*" or contains a "*", that means that all granules 
in this dataset
are individually orderable.  For example, a granule with a package "*" and:

       ARCHIVE_ID = "ARID"                      (archive tag = 2)
       DATASET_ID = "DELIGHTFUL DATA 444"       (dataset tag = 345)
       GRANULE_ID = "geo120966"

is orderable from archive "ABCD", individually, as:

       DATASET_ID = "DELIGHTFUL DATA 444"
       PACKAGE_ID = "geo120966"

In this case, we have the following if ordering this package:

       Package id           =  *
       Package suffix       =  */geo120966    
       
       Package key          =  2/345/*/geo120966  (archtag / dstag / suffix)
       Merged package key   =  2/345/*/geo120966  (archtag / dstag / suffix)
      
       Line item package    =  geo120966

=over *

If the package id is any other string, that means that one or more granules 
in this dataset are orderable as a collection under this id.  
For example, a granule with a package "ABBRVPKGNM" and:

       ARCHIVE_ID = "ARID"                      (archive tag = 2)
       DATASET_ID = "DELIGHTFUL DATA 444"       (dataset tag = 345)
       GRANULE_ID = "geo120966"

is part of a collection of granules that you'll get if you order:

       DATASET_ID = "DELIGHTFUL DATA 444"
       PACKAGE_ID = "ABBRVPKGNM"

In this case, we have the following if ordering this package:

       Package id           =  ABBRVPKGNM
       Package suffix       =  ABBRVPKGNM
       
       Package key          =  2/345/ABBRVPKGNM   (archtag / dstag / suffix)
       Merged package key   =  2/M/ABBRVPKGNM     (archtag /  "M"  / suffix)
       
       Line item package    =  ABBRVPKGNM

=head2 Merged (a.k.a. "unique") package keys

To begin with, you must note that named (non-granule) packages can span
data sets, so the dataaset identifier is ignored for such packages.
In short, there is really no difference between these packages... they
are all "really" the same package:

        "2/1/ABBRVPKGNM"   and
       "2/23/ABBRVPKGNM"   and
      "2/456/ABBRVPKGNM"

There is a one-to-one correspondence with merged package keys and the 
I<real> packages in the system.  If a granule is individually orderable, 
then it has its own merged package key (e.g., "2/345/*/geo120966").  
There is also I<exactly one> merged package key for each "named" package 
available at a DAAC; e.g., "2/M/ABBRVPKGNM".

Thus, to reflect the fact that all these are "really" the same, we "merge"
them by converting the dataset to "M" (for "merged"):

      "2/M/ABBRVPKGNM"

=head2 Merged package-combos keys

This is a unique representation of the combo that can be hashed upon.
It consists of:

     merged-package-key + proc-opt + media-type + media-fmt

Conceptually, this indicates a single unique combo chosen by the
user... if you hash ALL the user's chosen combos by this kind
of key, you'll get one entry per line-item.

=head1 INSTANCE VARIABLES

These must always be available:

    Id        => package id, either "*" or a real "pkid".
    ProcOpt   => processing option
    MediaType => media type
    MediaFmt  => media format

These are available for packages read from the support files:

    Cost      => approx cost
    NGrans    => number of granules

=head1 READING IN FROM CHUNKS

Each chunk is expected to contain the following attributes:

    PACKAGE_ID:          defaultable, see below
    NUMBER_OF_GRANULES:  defaultable, see below
    OPTION_ID:           defaultable, see below
    PACKAGE_SIZE:	 defaultable, see below
    TYPE_ID:             defaultable, see below
    FORMAT_ID:           mandatory
    APPROX_COST:         optional

Other attributes may appear, including DATA_URL groups.

=head1 RCS INFO

$Id: PkgCombo.pm,v 1.23 2003/09/09 14:25:47 solomon Exp $

=cut

use IU;
use IU::Log;
use IU::Chunk;
use IU::FancyURL;

require IU::Storable;
require IU::Granule;

use strict;

@IU::PkgCombo::ISA = qw(IU::Storable);

#------------------------------------------------------------
# new PARAMHASH
#------------------------------------------------------------
# Constructor: create new combo, from params.  The params are
# the instance vars.

sub new {
    my $class = shift;
    my %params = @_;
    my $self = \%params;
    $self->{Options} ||= {};
#    $self->{Copies} ||= {};
    bless $self, $class;
}

#------------------------------------------------------------
# new_fake_for_email 
#------------------------------------------------------------
# Constructor: create fake package combo, for email.
# Package id will be '*', and there will be no other info.

sub new_fake_for_email {
    my $class = shift;
    $class->new(Id=>'*',       # this granule!
		NGrans=>1,
		ProcOpt=>'', MediaType=>'', MediaFmt=>'');   # must define!
}

#------------------------------------------------------------
# copy 
#------------------------------------------------------------
# Copy constructor.
# WARNING: this does a shallow copy!

sub copy {
    my $old = shift;
    my $class = ref($old);
    $class->new(%{$old});
}

#------------------------------------------------------------
# as_hashkey
#------------------------------------------------------------
# Return a unique representation of the combo that can be hashed upon.
# It's contents are unimportant; it need only be unique (with regard
# to all the combos for the current granule) and reproducible.

sub as_hashkey {
    my $self = shift;
    join('/', $self->{Id}, $self->{ProcOpt}, 
	      $self->{MediaType}, $self->{MediaFmt});
}

#------------------------------------------------------------
# merged_key
#------------------------------------------------------------
# Return a unique representation of the combo that can be hashed upon.
# It consists of:
#
#    merged-package-key + proc-opt + media-type + media-fmt
#
# Conceptually, this indicates a single unique combo chosen by the
# user... if you hash ALL the user's chosen combos by this kind
# of key, you'll get one entry per uniquely-orderable thing.

sub merged_key {
    my $self = shift;
    join('/', 
	 $self->merged_package_key,
	 $self->{ProcOpt}, $self->{MediaType}, $self->{MediaFmt});
}

#------------------------------------------------------------
# Misc access methods...
#------------------------------------------------------------

# Get the PACKAGE_ID:
sub package_id  { $_[0]->{Id} }           # "*" or "pkid"

# Get the TITLE:
# if a value is passed in, we're setting, otherwise we're getting
sub package_title {
    if(defined($_[1])) { #setting
        $_[0]->{Title} = $_[1];
    }
    else { #getting
        $_[0]->{Title} ;
    }
}

# Get the processing OPTION_ID:
sub proc_opt    { $_[0]->{ProcOpt} }

# Get the processing PACKAGE_SIZE:
sub pkg_size    {
 my $pksize = $_[0]->{PackageSize};
 if ((!defined($pksize)) || ($pksize eq "0") || ($pksize eq " ") || ($pksize eq "")){
   $pksize = "Unknown";
  }
return $pksize;
 }

# Get the media TYPE_ID:
sub media_type  { $_[0]->{MediaType} }

# Get whether the HANDLING FEE should be applied?
sub apply_handling_fee  { $_[0]->{ApplyHandlingFee} }

# Get the media FORMAT_ID:
sub media_fmt   { $_[0]->{MediaFmt} }

# Get the number of granules in this package:
sub n_grans     { $_[0]->{NGrans} }

# Get the INFO_PROMPT:
sub info_prompt { $_[0]->{INFO_PROMPT} }

# Get the COMMENT:
sub comment     { $_[0]->{COMMENT} }

# Get the APPROX_COST, in decimal dollars:
sub cost   { $_[1] ? $_[0]->{Cost} = $_[1] : $_[0]->{Cost} } 

# Get the APPROX_COST, made pretty for display:
sub pretty_cost { PrettyCost($_[0]->{Cost} || 0) }

# Return ref to array of data URLs (as IU::FancyURLs), or [] if none:
sub data_URLs   { $_[0]->{DATA_URL} }

# Get/Set any additional info entered by user (such as account #)
# If called: $self->addn_info, it returns $self->{AddnInfo}
# If called: $self->addn_info($value), it sets $self->{AddnInfo} to $value.
sub addn_info   { $_[1] ? $_[0]->{AddnInfo} = $_[1] : $_[0]->{AddnInfo} }

# Get/Set the package number.  This is the number in the array of
# packages for this combo.
# If called: $self->pkg_no, it returns $self->{PkgNumber}
# If called: $self->pkg_no($value), it sets $self->{PkgNumber} to $value.
sub pkg_no   { $_[1] ? $_[0]->{PkgNumber} = $_[1] : $_[0]->{PkgNumber} }

#------------------------------------------------------------
# Since the following methods depend on the archive, datasets, and granule
# ids being set, they will only work after the combo has been tied to a 
# granule (by sending it # a granule_key() message with a GKEY argument)...
#------------------------------------------------------------

# Return the web archive key; e.g., "1".
sub archive_key { $_[0]->{ArchTag} }

# Return the web dataset key; e.g., "1/234".
sub dataset_key { $_[0]->{ArchTag} . '/' . $_[0]->{DsTag} }

# Return the granule id; e.g., "G00567.
sub granule_id  { $_[0]->{GId} }

# Return the package key: "archtag/dstag/suffix".
sub key { 
    $_[0]->dataset_key .'/'. $_[0]->suffix;
}

# Return the package id, as used in a LINE_ITEM group:
sub line_item_package { 
#    ($_[0]->{Id} eq '*') ? $_[0]->{GId} : $_[0]->{Id};
#   Allow more that one '*' package per data set.
#   If the package ID contains an '*' then use the granule
#   ID for the package ID when sending out the order. 
# 
    ($_[0]->{Id} =~ /\*/ ) ? $_[0]->{GId} : $_[0]->{Id};
}

# Return the suffix: "*/gid" if a single granule, or "pkid" if not.
sub suffix { 
#    ($_[0]->{Id} eq '*') ? "*/$_[0]->{GId}" : $_[0]->{Id};
#    ($_[0]->{Id} =~ /\*/ ) ? "*/$_[0]->{GId}" : $_[0]->{Id};
    ($_[0]->{Id} =~ /\*/ ) ? "$_[0]->{Id}/$_[0]->{GId}" : $_[0]->{Id};
}

#------------------------------------------------------------
# granule_key [GKEY]
#------------------------------------------------------------
# Get/set the granule key.
# Setting the key "ties" the combo to a granule.

sub granule_key {
    my ($self, $gkey) = @_;
    if (@_ > 1) {    # set it:
	($self->{ArchTag}, $self->{DsTag}, $self->{GId}) = 
	  IU::Granule::SplitGkey($gkey);
    }
    return join '/', $self->{ArchTag}, $self->{DsTag}, $self->{GId};
}

#------------------------------------------------------------
# merged_package_key
#------------------------------------------------------------
# There is a one-to-one correspondence with merged keys and
# smallest-orderable-entities in the system.  Thus, if a granule
# is individually orderable, then it has its own merged key.
# There is also one merged key for each package available at a DAAC.

sub merged_package_key { 
#    if ($_[0]->{Id} eq '*') {    # package of one granule
    if ($_[0]->{Id} =~ /\*/) {
       $_[0]->key;
    }
    else {                       # other package:
       $_[0]->archive_key . "/M/" . $_[0]->suffix;
    }
}

#------------------------------------------------------------
# options [OPTSTYPE, [OPTIONS]]
#------------------------------------------------------------
# Set/get the options named WHICH.  Possibilities for OPTSTYPE
# include "SUBSET", "ORDER", and "DATASET_ORDER".
#
# If no arguments are given, returns a list of OPTSTYPEs which
# have values in this object.
#
sub options {
    my ($self, $type, $opts) = @_;
    $type or return grep {$self->{Options}{$_}}(keys %{$self->{Options}});
    $self->{Options}{$type} = $opts if @_ > 2; 
    $self->{Options}{$type};
}

#------------------------------------------------------------
# copies
#------------------------------------------------------------
# Set/get the number of copies of the combo the user wants.

sub copies {

    my ($self, $copies) = @_;

    $self->{Copies} = $copies if @_ > 1;
    my $new_copies = $self->{Copies};
    $self->{Copies} || 1; 
}

#------------------------------
#
# CHUNK INTERFACE
#
#------------------------------

#------------------------------------------------------------
# from_chunk CHUNK
#------------------------------------------------------------
# Constructor: parse a chunk-encoded combo.

sub from_chunk {
    my ($self, $chunk) = @_;
    ref($self) or $self = $self->new;   
    
    # Trap for error chunk:
    my $errmsg;
    ($errmsg = $chunk->value('x-error')) 
	and return error(450, "Error chunk: $errmsg");
    
    # Simple attributes:
    $self->{Id}          = $chunk->value('PACKAGE_ID');
    $self->{NGrans}      = $chunk->value('NUMBER_OF_GRANULES');
    $self->{ProcOpt}     = $chunk->value('OPTION_ID');
    $self->{PackageSize} = $chunk->value('PACKAGE_SIZE');
    $self->{MediaType}   = $chunk->value('TYPE_ID');
    $self->{MediaFmt}    = $chunk->value('FORMAT_ID');
    $self->{Cost}        = $chunk->value('APPROX_COST');
    $self->{INFO_PROMPT} = $chunk->seq_value('INFO_PROMPT');
    $self->{COMMENT}     = $chunk->seq_value('COMMENT');   
    $self->{ApplyHandlingFee} = $chunk->value('APPLY_HANDLING_FEE');

    # URLs (we tolerate more than one of each type):
    my $key;
    foreach $key (qw(DATA_URL)) {

	# Allow many URLs of the current flavor...
	$self->{$key} = [];
	my ($url_chunk, $furl);
	foreach $url_chunk ($chunk->values($key)) {
	    my $furl = IU::FancyURL->from_chunk($url_chunk, $key);
	    push @{$self->{$key}}, $furl if $furl;
	}
    }
    
    # Validate:
    foreach (qw(Id ProcOpt MediaType MediaFmt)) {
	defined($self->{$_}) or return error "package is missing $_: skipping";
    }
    $self;
}

# name -- Product Name of the option
sub name {
    my ($self, $value) = @_;
    $self->{Name} = $value if (@_ > 1);
    $self->{Name};
}

#------------------------------
#
# UTILITIES
#
#------------------------------

#------------------------------------------------------------
# compare A,B
#------------------------------------------------------------
# Sort comparison routine.

sub compare {
    my ($oa, $ob) = @_;
    my $cmp;
    return $cmp if ($cmp = ($oa->{Id}        cmp $ob->{Id}));
    return $cmp if ($cmp = ($oa->{ProcOpt}   cmp $ob->{ProcOpt}));
    return $cmp if ($cmp = ($oa->{MediaType} cmp $ob->{MediaType}));
    return $cmp if ($cmp = ($oa->{MediaFmt}  cmp $ob->{MediaFmt}));
    0;
}

#------------------------------------------------------------
# PrettyCost COST
#------------------------------------------------------------
# Utility: convert an approximate cost.

sub PrettyCost {
    sprintf("US \$%0.2f", $_[0]);
}

#------------------------------------------------------------
1;


