package IU::PriceEstimate;

=head1 NAME

IU::PriceEstimate - user request for Price Estimate.


=head1 SYNOPSIS

Search options/properties:
    
    # What kind of search is this?  Can return any of...
    #     PRICE_ESTIMATE_REQUEST
    $q->search_type;
    
    # What kind of ODL search message is this *really*?  Can return any of...
    #     PRICE_ESTIMATE_REQUEST
    $q->ODL_search_type;

ODL interface:

    # ODL interface:
    $q = IU::PriceEstimate->from_ODL($odl);
    $odl = $q->to_ODL;
     

=head1 DESCRIPTION

This object encapsulates Price Estimate.

=cut

use IU;
use IU::Log;
use IU::IW;

use IW::App;

use strict;
use vars qw(@ISA @EXPORT
	    $MAX_TIME_DEFAULT
            $PRICE_ESTIMATE_REQUEST);
@ISA = qw(Exporter);
@EXPORT = qw($PRICE_ESTIMATE_REQUEST);

#------------------------------
#
# Globals...
#
#------------------------------

# The current query (EXPORTED for the convenience of other modules):
$PRICE_ESTIMATE_REQUEST = undef;

# Default time limit for searches
$MAX_TIME_DEFAULT = $IU::SETUP::MaxSearchTime;

#------------------------------
#
# Private utilities...
#
#------------------------------

#------------------------------------------------------------
# _one VARNAME,SELF,[VALUE]
#------------------------------------------------------------
# I<Private utility.>  Get/set a single-valued instance variable.
# Return undef if it's undefined.
sub _one {
    my ($varname, $self, $value) = @_;
    $self->{$varname} = $value if (@_ > 2);
    $self->{$varname};   
}

#------------------------------------------------------------
# _many VARNAME,SELF,[ARRAYREF]
#------------------------------------------------------------
# I<Private utility.>  Get/set a multi-valued instance variable.
# Set it by providing an arrayref: the array will be duplicated.
#
# If invoked with no args, just returns info.
# Returns the array of all values in an array context, and just 
# the array size in a scalar context.

sub _many {
    my ($varname, $self, $values) = @_;
    if (@_ > 2) {
	my @array = @$values;
	$self->{$varname} = \@array;
    }
    (wantarray ? @{$self->{$varname}} : int(@{$self->{$varname}}));
}

#------------------------------
#
# Construction...
#
#------------------------------

#------------------------------------------------------------
# new 
#------------------------------------------------------------
# Class method.  Constructor.

sub new {    
    my $class = shift;
    my $searchType = shift;
    my $self = {};
    $self->{SearchType} = $searchType;
    $self->{DaacsToSearch} = [];
    $self->{DaacsNeedingECSAuth} = [];
    bless $self, $class;
}

#------------------------------------------------------------
# daacs_to_search
#------------------------------------------------------------
# I<Instance method.>
# Return the list of data centers to search.
# This will be empty until set up by determine_daacs_to_search().

sub daacs_to_search { _many('DaacsToSearch', @_) }

#------------------------------------------------------------
# latitudes
#------------------------------------------------------------
# I<Instance method.>
# Return the list of latitudes for the POLYGON_LOC

sub latitudes { _many('Latitudes', @_) }

#------------------------------------------------------------
# longitudes
#------------------------------------------------------------
# I<Instance method.>
# Return the list of longitudes for the POLYGON_LOC

sub longitudes { _many('Longitudes', @_) }

#------------------------------------------------------------
# daacs_needing_ecs_auth
#------------------------------------------------------------
# I<Instance method.>
# Return the list of data centers which according the wwwvalids 
# need the ECS authenticator.
# This will be empty until set up by determine_daacs_to_search().

sub daacs_needing_ecs_auth { _many('DaacsNeedingECSAuth', @_) }

#------------------------------------------------------------
# bands
#------------------------------------------------------------
# I<Instance method.>
# Return the list of bands.

sub bands { _many('Bands', @_) }

#------------------------------
#
# Search options/properties...
#
#------------------------------

#------------------------------------------------------------
# max_time [VALUE]
#------------------------------------------------------------
# Set/get maximum number of minutes that a search can run.
#

sub max_time {
    my ($self, $value) = @_;
    $self->{MaxTime} = $value if @_ > 1;
    $self->{MaxTime} || $MAX_TIME_DEFAULT;
}

#------------------------------------------------------------
# name [NAME]
#------------------------------------------------------------
# Set/get the name of the query.
#
sub name  { _one('Name', @_) }

#------------------------------------------------------------
# dataset_id [VALUE]
#------------------------------------------------------------
# Set/get the dataset_id.
#
sub dataset_id  { _one('DatasetID', @_) }

#------------------------------------------------------------
# fullsub [VALUE]
#------------------------------------------------------------
# Set/get the fullsub 
#
sub fullsub  { _one('FullSub', @_) }

#------------------------------------------------------------
# package_id [VALUE]
#------------------------------------------------------------
# Set/get the package_id.
#
sub package_id  { _one('PackageID', @_) }

#------------------------------------------------------------
# processing_options [VALUE]
#------------------------------------------------------------
# Set/get the processing_options.
#
sub processing_options  { _one('ProcessingOptions', @_) }

#------------------------------------------------------------
# media_type [VALUE]
#------------------------------------------------------------
# Set/get the media_type.
#
sub media_type  { _one('MediaType', @_) }

#------------------------------------------------------------
# media_format [VALUE]
#------------------------------------------------------------
# Set/get the media_format.
#
sub media_format  { _one('MediaFormat', @_) }

#------------------------------------------------------------
# billing_id [VALUE]
#------------------------------------------------------------
# Set/get the billing_id.
#
sub billing_id  { _one('BillingID', @_) }

#------------------------------------------------------------
# filename
#------------------------------------------------------------
# Get a good filename for the query.
#
sub filename {
    my $self = shift;
    my $name = $self->name || 'untitled';
    $name =~ s/^\s+|\s+$//g;        # trim
    $name =~ s/[^a-z0-9_+-]/_/gi;   # turn illegals into underscores
    $name =~ s/_+/_/g;              # collapse underscores
    "$name.query";
}

#------------------------------------------------------------
# ODL_search_type
#------------------------------------------------------------
# Return the search type, as it would appear in the ODL.
#
sub ODL_search_type { 
    my $self = shift;
    my $type = $self->search_type;
    return $type;
}

#------------------------------------------------------------
# search_type [VALUE]
#------------------------------------------------------------
# get the search type, in the most-detailed sense.
# Compare with ODL_search_type.
#
sub search_type { 'PRICE_ESTIMATE_REQUEST'  }

#------------------------------------------------------------
# which_daacs_need_authenticator  
#------------------------------------------------------------
# returns array of daacs

sub which_daacs_need_authenticator {

    my $self = shift;
    $self->determine_daacs_to_search;
    my @initialDaacs = $self->daacs_to_search;

    # Open static valids database:
    my %WWWVALIDS;
    if (-e "${IU::LibDir}/conf/wwwvalids.gdb") {
        eval {tie %WWWVALIDS, 'IU::Store', "${IU::LibDir}/conf/wwwvalids.gdb",
              'READER'};
        $@ and return error "tie failed: $@";
    }

    # For each data center we're considering...
    my %needs;        # Daacs needing authenticator. 
    my $archid;
    foreach $archid (@initialDaacs) {
        # Is there an entry for it in the static valids?
        my $valids;
        if ($valids = $WWWVALIDS{"daac:$archid"}) {
            $needs{$archid} = 'ECS' if ($valids->{REQUIRES_ECS_AUTH}); 
        }
    }                                    # end: foreach daac 

    # Close static valids database:
    untie %WWWVALIDS;

    return %needs;
}

#------------------------------------------------------------
# filter_daacs_to_search
#------------------------------------------------------------
# I<Instance method, private.>
# Take the list of data centers to actually query, and pare them down.
# based on the static valids file.  See determine_daacs_to_search().
#
# Raises a fatal exception if no DAACs remain.

sub filter_daacs_to_search {
    my $self = shift;
    my @initialDaacs = $self->daacs_to_search;
    my @filteredDaacs;

    # Open static valids database:
    my %WWWVALIDS;
    if (-e "${IU::LibDir}/conf/wwwvalids.gdb") {
        eval {tie %WWWVALIDS, 'IU::Store', "${IU::LibDir}/conf/wwwvalids.gdb",
              'READER'};
        $@ and return error "tie failed: $@";
    } 

    #--------------------
    # create list of data centers needing ECS 
    #--------------------

    # For each data center we're considering...
    my %needsECS;       # Daacs needing ecs authenticator.
    my $daac;

    foreach $daac (@initialDaacs) {
        # Is there an entry for it in the static valids?
        my $valids;
        if ($valids = $WWWVALIDS{"daac:$daac"}) {
            $needsECS{$daac} = $valids->{REQUIRES_ECS_AUTH};
        }                       # end if
    }                                    # end: foreach daac

    my @daacsECS = grep { $needsECS{$_}} (keys %needsECS);

    $self->daacs_needing_ecs_auth(\@daacsECS);

    # Close static valids database:
    untie %WWWVALIDS;

    # Set new list of daacs to search:
    @filteredDaacs = @initialDaacs;
    $self->daacs_to_search(\@filteredDaacs);
    1;
}

#------------------------------------------------------------
# determine_daacs_to_search
#------------------------------------------------------------
# I<Instance method.>
# Determine list of data centers to actually query, using dependent valids.
# Fatal exception if no DAACs can be set.

sub determine_daacs_to_search {
    my $self = shift;
    my @finalDaacs = ();
    $self->daacs_to_search(\@finalDaacs);
    $self->filter_daacs_to_search;
    # Done!
    1;
}

#------------------------------------------------------------    
# install
#------------------------------------------------------------    
# I<Instance method.>  WG-specific.
# Re-install this as the current query in the session state.

sub install {
    my $self = shift;
    $SESSION->state->{PriceEstimateRequest} = $self;
}

#------------------------------------------------------------    
# lookup
#------------------------------------------------------------ 
# I<Class method.  WG-specific.>
# Lookup current query in the session state.  Returns undef if none so far.

sub lookup {
    my $self = shift;
    my $query = $SESSION->state->{PriceEstimateRequest};
    return $query;
}

#------------------------------------------------------------
# uses CATEGORY, [SEARCHTYPE]
#------------------------------------------------------------
# I<Class/instance method.>
# Returns true if the given CATEGORY would be a functional part of the 
# query under the given SEARCHTYPE, and false otherwise.   With
# no SEARCHTYPE, will use the query's search type.

sub uses {
    my ($self, $cat, $type) = @_;
    $type ||= $self->search_type;  
    if ($type eq 'PRICE_ESTIMATE_REQUEST') { 
        return 1 if ($cat eq 'DATA_CENTER_ID');
	return 0;
    }
    else {         # any other kind of search
	return 1;        
    }
}

#------------------------------------------------------------
# uses_opt OPTION, [SEARCHTYPE]
#------------------------------------------------------------
# I<Class/instance method.>
# Much like uses(), but for miscellaneous query options that aren't
# really categories.  Valid OPTIONs are:
#
#     QUERY_NAME       Query name
#
# Returns true if the given OPTION would be a functional part of the 
# query under the given SEARCHTYPE, and false otherwise.   With
# no SEARCHTYPE, will use the query's search type.
# Always show guide limit since it will take effect if the user selects
# Detailed Info while doing any kind of search.

sub uses_opt {
    my ($self, $opt, $type) = @_;

    $type ||= $self->search_type;
    if ($type eq 'PRICE_ESTIMATE_REQUEST') {
	return 1 if ($opt eq 'QUERY_NAME');
	return 0;
    }
    else {         # any other kind of search
	return 1;        
    }
}

#------------------------------
#
# ODL interface
#
#------------------------------

#------------------------------------------------------------
# from_ODL ODLOBJECT, PARAMHASH
#------------------------------------------------------------
# I<Class or instance method.>
# Initiallize the address from a standard IMS ODL object.
#
# The PARAMHASH can contain the following:
#
#     KeepType => If true, the search type will be discarded in 
#                 favor of the type of any CURRENTLY-INSTALLED query.
#                 (See lookup() and install() for details).
#                 If false (default), the ODL alone is used in determining
#                 the search type. 

sub from_ODL {
    my ($self, $odl, %params) = @_;
    $odl || return undef;
    my ($param, $value);
    my $search;
    my $keep_type = $params{KeepType};

    # Be a class method, too!
    ref($self) or $self = $self->new;
    #
    # DETERMINE THE SEARCH TYPE...
    #
    #    If "keep type" is true, determine the search type as given by 
    #    any currently-installed query, and prefer that value to the 
    #    actual one given by the ODL.
    #
    #    This is done so that when the user is UPLOADING an ODL search file,
    #    we automatically coerce it into the type of search form 
    #    (INVENTORY, DIRECTORY, GRANULE, etc.) that they are uploading into.

    # Determine the search type as given by the currently-installed query:
    my $prev_search_type;
    if ($keep_type) {
	my $prev_query = ref($self)->lookup;        # don't put it in $QUERY!
	$prev_search_type = $prev_query && $prev_query->search_type;
    }

    # Determine the search type as given by the ODL itself.  
    #    NOTE: this is not necessarily the actual search type of the
    #    top-level message... rather, it is the search type as derived
    #    from the ODL itself.  Notice the case for GRANULE_SEARCH.
    my $new_search_type;
    if ($search = $odl->param('PRICE_ESTIMATE_REQUEST')) {
	$new_search_type = 'PRICE_ESTIMATE_REQUEST';
    }
    else {
        puke "ODL doesn't seem to hold PRICE_ESTIMATE_REQUEST ";
    }

    # Now, resolve the two search types, preferring the CGI input:
    $self->search_type($prev_search_type || $new_search_type);

    # 
    # EXTRACT THE SEARCH CONSTRAINTS FROM THE ODL...
    #
    $search->rewind;
    my %queue;

    while (($param, $value) = $search->next) {

	if ($param eq 'TIME_LIMIT') {
	    $self->max_time($value->text_value);
	}
	elsif ($param eq 'SEARCH_NAME') {
	    $self->name($value->text_value);
	}
	else {
	    warn "skipping unrecognized search parameter: '$param'";
	}
    }    
    $self;
}

#------------------------------------------------------------
# to_ODL PARAMHASH
#------------------------------------------------------------
# Convert query to an ODL object, and return it.
# The ODL object is a group *containing* an INVENTORY_SEARCH group
# (or DIRECTORY_SEARCH, etc.).
#
# No authentication information is placed in the ODL!!!!
#
# PARAMHASH may contain the following:
#
#    NoDaacs   => If true, then the DATA_CENTER_ID group (user-specified
#                 data centers) will *not* be output in the ODL.
#                 Generally, we handle these separately.
#
#    Everything=> By default, ODL is only output for the current search type,
#                 meaning that if an attribute is not used in the current
#                 search type, then it will not appear in the ODL.
#                 If you set this option to true, then *all* criteria will 
#                 be included in the ODL.
#
#    ForSearch => If true, ODL is tweaked to default/modify certain fields
#                 (e.g., stop date) to make DAACs happy.  Normally,
#                 we don't want the user's query criteria mucked with,
#                 so the default is false.
#
# NOTE: for downloading, use the Everything option to output ALL the 
# categories to the ODL, even ones which are unused for this type of search.  
# This is so an inventory search may be uploaded, turned into a granule-id 
# search, and downloaded, but still retain all the inventory search 
# characteristics.
#
# NOTE: the ForSearch option produces *slightly* different ODL.  Normally,
# we return an accurate, uploadable representation of the query.  With
# ForSearch, we ensure that certain ODL fields are defaulted, to deal 
# with the less-tolerant DAACs.

sub to_ODL {
    my ($self, %params) = @_;
    my $no_daacs         = $params{NoDaacs};
    my $write_everything = $params{Everything};
    my $for_search       = $params{ForSearch};

    require IU::ODL;

    # Begin:
    my $root_odl = IU::ODL->new_group;    # outermost group
    my $srch_odl = IU::ODL->new_group;

    my $subset_odl = IU::ODL->new_group; # SUBSET_SPEC
    my $crit_odl  = IU::ODL->new_group; # SPECIALIZED_CRITERIA
    my $band_odl  = IU::ODL->new_group; # SPECIALIZED_CRITERIA
    my $line_odl  = IU::ODL->new_group; # LINE_ITEM
    my $polygon_odl = IU::ODL->new_group; # POLYGON_LOC

    # Start off the file with the search type:
    $root_odl->add_param($self->ODL_search_type => $srch_odl);

    # The message id (do we REALLY need this?)
    $srch_odl->add_param(MESSAGE_ID    => '0'); 

    # Completely non-standard stuff:
    if ($self->name) {
	$srch_odl->add_param(SEARCH_NAME => IU::ODL->quoted($self->name));
    }
    # Max guides should be saved for all search types.
#   $srch_odl->add_param(GUIDE_LIMIT =>
#                        IU::ODL->number('40'));

    # Max time should be saved for all search types.
#   $srch_odl->add_param(TIME_LIMIT =>
#                        IU::ODL->number('45'));

    # Add the DORRAN_USERNAME if the user is a persistent user.
    my $persistent_user = $ENV{REMOTE_USER};
    $srch_odl->add_param(DORRAN_USERNAME => IU::ODL->quoted($persistent_user))
     if($persistent_user ne 'guest');

    $srch_odl->add_param(LINE_ITEM => $line_odl);
    my $dataset = $self->dataset_id;
    $line_odl->add_param(DATASET_ID => IU::ODL->conv($dataset)) if $dataset;
    my $package = $self->package_id;
    $line_odl->add_param(PACKAGE_ID => IU::ODL->conv($package)) if $package;
    my $processing = $self->processing_options;
    $line_odl->add_param(PROCESSING_OPTIONS => IU::ODL->conv($processing)) 
     if $processing;
    my $media_type = $self->media_type;
    $line_odl->add_param(MEDIA_TYPE => IU::ODL->conv($media_type)) 
     if $media_type;
    my $media_format = $self->media_format;
    $line_odl->add_param(MEDIA_FORMAT => IU::ODL->conv($media_format)) 
     if $media_format;
    my $billing = $self->billing_id;
    $line_odl->add_param(BILLING_ID => IU::ODL->conv($billing)) if $billing;
    $line_odl->add_param(SUBSET_SPEC => $subset_odl);
    $subset_odl->add_param(FULL_SUBINTERVAL => IU::ODL->quoted($self->fullsub))
       if ($self->fullsub);
    $subset_odl->add_param(SPECIALIZED_CRITERIA => $crit_odl);
    $crit_odl->add_param(CRITERIA_NAME => 'Spatial Subsetting');
    $crit_odl->add_param(CRITERIA_TYPE => 'GEO');
    $crit_odl->add_param(POLYGON_LOC => $polygon_odl);
    my @lat_odls = map {IU::ODL->real($_)} $self->latitudes;
    my @lon_odls = map {IU::ODL->real($_)} $self->longitudes;
    $polygon_odl->add_param(LATITUDE => IU::ODL->sequence(@lat_odls));
    $polygon_odl->add_param(LONGITUDE => IU::ODL->sequence(@lon_odls));
    $subset_odl->add_param(SPECIALIZED_CRITERIA => $band_odl);
    $band_odl->add_param(CRITERIA_NAME => 'Band Subsetting');
    $band_odl->add_param(CRITERIA_TYPE => 'STRING');
    my @bands = $self->bands;
    $band_odl->add_param(CRITERIA_VALUE => IU::ODL->conv(\@bands)) if @bands;
    # Done!
    $root_odl;
}

#------------------------------------------------------------
# to_search_ODL
#------------------------------------------------------------
# Convert query to an ODL object suitable for input to the client, 
# and return it. The ODL object is a group containing an INVENTORY_SEARCH 
# parameter.
#
# Unlike a simple to_ODL, this method uses the dependent valids mechanism
# to determine which DAACs to search

sub to_search_ODL {
    my $self = shift;

    # Start by figuring out which DAACS to search, if needed:
    $self->daacs_to_search
	    or puke "no data centers to search!";  

    # Start with the basics, *without* the DAACs:
    my $root_odl = $self->to_ODL(NoDaacs=>1, ForSearch=>1) 
	|| return undef;

    # Get the actual search group, checking all standard types:
    my $srch_odl = $root_odl->param('PRICE_ESTIMATE_REQUEST') ||
		   puke "No recognized search group in ODL!"; 

    # Add the data centers (notice that we use the QUERY daacs!):
    if ($self->daacs_to_search) {
	$srch_odl->add_param(DATA_CENTER_ID => 
 			     IU::ODL->conv($self->{DaacsToSearch}));
    }

    # Add the ecs authenticator, if any:
    my $ecs_authenticator = eval { $SESSION->state->{User}->ecs_authenticator };
    if ($ecs_authenticator) {
       $srch_odl->add_param(ECS_AUTHENTICATOR => 
                            IU::ODL->conv($ecs_authenticator));
    }

    # Done!
    return $root_odl;
}

#------------------------------------------------------------    
1;
