package IU::Query;


=head1 NAME

IU::Query - user query for data on the "search" form


=head1 SYNOPSIS

Search properties:
    
    # What kind of search is this?  Can return any of...
    #     INVENTORY_SEARCH
    #     DIRECTORY_SEARCH
    #     GUIDE_SEARCH
    #     GRANULE_SEARCH     (a special kind of INVENTORY_SEARCH)
    #     LOCAL_GRANULE_SEARCH     (a special kind of INVENTORY_SEARCH)
    #     BROWSE_SEARCH (a special kind of INVENTORY_SEARCH)
    #     PRODUCT_STATUS_REQUEST
    #     ACCOUNT_STATUS_REQUEST
    $q->search_type;
    
    # What kind of ODL search message is this *really*?  Can return any of...
    #     INVENTORY_SEARCH
    #     DIRECTORY_SEARCH
    #     GUIDE_SEARCH
    #     PRODUCT_STATUS_REQUEST
    #     ACCOUNT_STATUS_REQUEST
    $q->ODL_search_type;

Search categories/criteria:

    # What categories has the user given constraints in?
    @categories = $q->categories;
    
    # Get user's choices for this criterion, as "criterion" objects:
    @criteria = $q->criteria("PARAMETER");
    
    # Set criteria information, using an array of "criterion" objects:
    $q->criteria("PARAMETER", \@params);

Dependent valids:

    # What DAACS will we search on, based on dependent valids?
    $q->determine_daacs_to_search;
    $q->daacs_to_search;

ODL interface:

    # ODL interface:
    $q = IU::Query->from_ODL($odl);
    $odl = $q->to_ODL;
     
Output as HTML:

    # Dump pretty text (note: takes an HTML stream!!!):
    $q->print_HTML($HTML)


=head1 DESCRIPTION

This object encapsulates all kinds of queries: inventory, directory,
guide, etc.


=head1 VERSION

$Id: Query.pm,v 1.99 2003/05/21 14:18:38 solomon Exp $

=cut

use IU;
use IU::Log;
use IU::Category;
use IU::Criterion;
use IU::IW;
use IU::ArchiveOdl;
use IU::Geo::Range;
use IU::WWWValids;

use IW::App;
use IW::User;
use IW::DepValid;
use IW::Widget::DepValids;

use strict;
use vars qw(@ISA @EXPORT
            $QUERY
            $MAX_GRANS_MIN
	    $MAX_GRANS_DEFAULT
	    $MAX_GRANS_MAX
            $MAX_TIME_DEFAULT
	    $MAX_GUIDES_DEFAULT);
@ISA = qw(Exporter);
@EXPORT = qw($QUERY);

#------------------------------
#
# Globals...
#
#------------------------------

# Limits on max number of granules to return:
$MAX_GRANS_MIN     = 1;   # const: min number of grans that can be returned
# Default number of granules to return per dataset
$MAX_GRANS_DEFAULT = $IU::MaxGransDefault || 100;
$MAX_GRANS_MAX     = 1000; # const: max number of granules that can be returned
# Default limit on maximum number of guide documents returned on a guide search
$MAX_GUIDES_DEFAULT = $IU::GuideLimit || 100; 
# Default time limit for searches
$MAX_TIME_DEFAULT = $IU::SETUP::MaxSearchTime;
# The current query (EXPORTED for the convenience of other modules):
$QUERY = undef;

# Categories which are output in the ODL as top-level sequence strings
# under the category name (except for DATA_CENTER_ID, which is handled 
# separately):
my %SeqStringCore = map {$_ => 1}
 qw(
   CAMPAIGN
   DATA_CENTER_ID
   DATASET_ID
   PARAMETER
   PROCESSING_LEVEL
   SENSOR_NAME
   SOURCE_NAME
);

my %SeqStringNewCore = map {$_ => 1}
 qw(
   CAMPAIGN
   PARAMETER
   PROCESSING_LEVEL
   SENSOR_NAME
   SOURCE_NAME
);

my %SeqStringCategories = map {$_ => 1}
 qw(
   GRANULE_ID_REQ
   REQUEST_ID
   RESULT_ATTRIBUTES
);

# Categories which are output in the ODL as top-level simple strings
# under the category name:
my %StringCategories = map {$_ => 1}
 qw(
   TEXT_STRING
);

#------------------------------
#
# Private utilities...
#
#------------------------------

#------------------------------------------------------------
# _one VARNAME,SELF,[VALUE]
#------------------------------------------------------------
# I<Private utility.>  Get/set a single-valued instance variable.
# Return undef if it's undefined.
sub _one {
    my ($varname, $self, $value) = @_;
    $self->{$varname} = $value if (@_ > 2);
    $self->{$varname};   
}

#------------------------------------------------------------
# _many VARNAME,SELF,[ARRAYREF]
#------------------------------------------------------------
# I<Private utility.>  Get/set a multi-valued instance variable.
# Set it by providing an arrayref: the array will be duplicated.
#
# If invoked with no args, just returns info.
# Returns the array of all values in an array context, and just 
# the array size in a scalar context.

sub _many {
    my ($varname, $self, $values) = @_;
    if (@_ > 2) {
	my @array = @$values;
	$self->{$varname} = \@array;
    }
    (wantarray ? @{$self->{$varname}} : int(@{$self->{$varname}}));
}

#------------------------------------------------------------
# intersect ARRAYREF, ARRAYREF
#------------------------------------------------------------
# Utility: intersect two lists of scalars, and return the result.

sub intersect {
    my ($a, $b) = @_;
    my %in_a = map {$_=>1} @$a;
    return sort(grep {$in_a{$_}} @$b);
}

#------------------------------
#
# Construction...
#
#------------------------------

#------------------------------------------------------------
# new 
#------------------------------------------------------------
# Class method.  Constructor.

sub new {    
    my $class = shift;
    my $self = {};
    $self->{SearchType} = 'INVENTORY_SEARCH';
    $self->{Categories}    = {};
    $self->{DaacsToSearch} = [];
    $self->{DaacsNeedingECSAuth} = [];
    $self->{DaacsNeedingV0Auth}  = [];
    $self->{DaacsNeedingNASDAAuth}  = [];
    $self->{ExtendedSearch} = {};
    bless $self, $class;
}

#------------------------------
#
# Known categories...
#
#------------------------------

#------------------------------------------------------------
# datasets_to_search
#------------------------------------------------------------
# I<Instance method.>
# Return the list of data sets to search.
# This will be empty until set up by determine_datasets_to_search().

sub datasets_to_search { _many('DatasetsToSearch', @_) }

#------------------------------------------------------------
# daacs_to_search
#------------------------------------------------------------
# I<Instance method.>
# Return the list of data centers to search.
# This will be empty until set up by determine_daacs_to_search().

sub daacs_to_search { _many('DaacsToSearch', @_) }

#------------------------------------------------------------
# daacs_needing_ecs_auth
#------------------------------------------------------------
# I<Instance method.>
# Return the list of data centers which according the wwwvalids 
# need the ECS authenticator.
# This will be empty until set up by determine_daacs_to_search().

sub daacs_needing_ecs_auth { _many('DaacsNeedingECSAuth', @_) }

#------------------------------------------------------------
# daacs_needing_v0_auth
#------------------------------------------------------------
# I<Instance method.>
# Return the list of data centers which according the wwwvalids 
# need the V0 authenticator.
# This will be empty until set up by determine_daacs_to_search().

sub daacs_needing_v0_auth { _many('DaacsNeedingV0Auth', @_) }

#------------------------------------------------------------
# daacs_needing_nasda_auth
#------------------------------------------------------------
# I<Instance method.>
# Return the list of data centers which according the wwwvalids 
# need the NASDA authenticator.
# This will be empty until set up by determine_daacs_to_search().

sub daacs_needing_nasda_auth { _many('DaacsNeedingNASDAAuth', @_) }

#------------------------------------------------------------
# optstype [TYPE]
#------------------------------------------------------------
sub optstype {
   my ($self, $value) = @_;
   $self->{OptsType} = $value if @_ > 1;
   $self->{OptsType} || '';
}

#------------------------------------------------------------
# depvalidtype [TYPE]
#------------------------------------------------------------
sub depvalidtype {
   my ($self, $value) = @_;
   $self->{DepValidType} = $value if @_ > 1;
   $self->{DepValidType} || '';
}
#------------------------------------------------------------
# geotype [TYPE]
#------------------------------------------------------------
sub geotype {
   my ($self, $value) = @_;
   $self->{GeoType} = $value if @_ > 1;
   $self->{GeoType};
}

#------------------------------------------------------------
# timetype [TYPE]
#------------------------------------------------------------
sub timetype {
   my ($self, $value) = @_;
   $self->{TimeType} = $value if @_ > 1;
   $self->{TimeType} || '';
}


#------------------------------------------------------------
# campaigns       [VALIDS]
# data_center_ids [VALIDS]
# ...
#------------------------------------------------------------
# Instance methods.
# Convenient ways to get/set the array of valids for core      .
# All of these accept an optional arrayref of strings to set, and 
# return an array of strings.
#
sub campaigns         { my $self = shift; 
			$self->criteria_strings('CAMPAIGN',@_)}
sub data_center_ids   { my $self = shift; 
			$self->criteria_strings('DATA_CENTER_ID',@_)}
sub dataset_ids       { my $self = shift; 
			$self->criteria_strings('DATASET_ID',@_)}
sub granule_ids       { my $self = shift; 
			$self->criteria_strings('GRANULE_ID_REQ',@_)}
sub local_granule_id { my $self = shift; 
			$self->criteria('GRANULE DETAILS',@_)}

sub node_type { 
   my $self = shift; 
   $self->criteria('ORBIT',@_)
}

sub missions {
   my ($self, $values) = @_;
   $self->{'WG_MISSIONS'} = $values if (@_ > 1); # set values
   defined($self->{'WG_MISSIONS'}) ? @{$self->{'WG_MISSIONS'}} : undef; 
}

#   _many('WG_MISSIONS', @_) }

sub request_ids       { my $self = shift; 
			$self->criteria_strings('REQUEST_ID',@_)}
sub parameters        { my $self = shift; 
			$self->criteria_strings('PARAMETER',@_)}
sub processing_levels { my $self = shift; 
			$self->criteria_strings('PROCESSING_LEVEL',@_)}
sub sensor_names      { my $self = shift; 
			$self->criteria_strings('SENSOR_NAME',@_)}
sub source_names      { my $self = shift; 
			$self->criteria_strings('SOURCE_NAME',@_)}
sub result_attributes { my $self = shift; 
			$self->criteria_strings('RESULT_ATTRIBUTES',@_)}
sub text_string       { 
    my ($self, $str) = @_;
    if (@_ > 1) {
	$self->criteria_strings('TEXT_STRING', (defined($str) ? [$str] : []));
    }
    ($self->criteria_strings('TEXT_STRING'))[0];
}

### Old-style, DEPRECATED!
*daacs      = \&data_center_ids;
*datasets   = \&dataset_ids;
*geo        = \&geo_region;
*granule_id = \&granule_ids;
*max_grans  = \&max_granules;
*params     = \&parameters;
*proclevels = \&processing_levels;
*sensors    = \&sensor_names;
*sources    = \&source_names;
*textstring = \&text_string;

#------------------------------------------------------------
# geo_region [GEO]
#------------------------------------------------------------
# I<Instance method.>
# Convenient way to get/set the sole geoselection criterion.
# Returns current geo object, or undef if none.
#
sub geo_region {
    my ($self, $geo) = @_;
    my $crit;            # note: initially undefined!
    
    # Set new value?
    if (@_ > 1) {    
	if ($geo) {      # actual value
	    $crit= IU::Criterion::Geo->new(CRITERIA_NAME=>'GEOGRAPHIC_REGION');
	    $crit->geo($geo);
	}
	$self->criteria('GEOGRAPHIC_REGION', $crit);
    }
    
    # Return it:
    $crit = ($self->criteria('GEOGRAPHIC_REGION'))[0];   # get first criterion
    return ($crit and $crit->geo);
}

#------------------------------------------------------------
# time_range_TZ [TZSTART, TZSTOP]
#------------------------------------------------------------
# I<Instance method.>
# Set/get the time range, from the TIME_RANGE category.
# Returns the current time range as an array (TZSTART, TZSTOP).
# TZSTART and TZSTOP are TZ-strings (e.g., "1992-09-12T12:34:56").
#
sub time_range_TZ {
    my ($self, $start_tz, $stop_tz, $start_doy, $stop_doy) = @_;
    my $crit;            # note: initially undefined!
    my @values;
    $values[0] = $start_tz;
    $values[1] = $stop_tz;
    $values[2] = $start_doy;
    $values[3] = $stop_doy;
  
    
    # Set new value?
    if (@_ > 1) {    
	if ($start_tz || $stop_tz) {      # actual value
	    $crit = (new IU::Criterion::Time
		     CRITERIA_NAME => 'TIME_RANGE',
		     CRITERIA_TYPE => 'DATE',
		     VALUES => \@values);	     
	}
	$self->criteria('TIME_RANGE', $crit);
    }
    
    # Return it:
    $crit = ($self->criteria('TIME_RANGE'))[0] || return (); # get 1st crit
    $start_tz = $crit->startDate;
    $stop_tz =  $crit->stopDate;
    $start_doy = $crit->startDayOfYear;
    $stop_doy =  $crit->stopDayOfYear;
    return ($start_tz, $stop_tz, $start_doy, $stop_doy);
}

#------------------------------------------------------------
# include_nontime [VALUE]
#------------------------------------------------------------
# Include Non-Temporal Granules in search results
#
sub include_nontime  { 
   _one('Include_NonTemporal', @_) ;
}


#------------------------------------------------------------
# show_nontime 
#------------------------------------------------------------
# If SETUP flag is not set to 'Y', return undef
#
sub show_nontime  {
   return ( ($IU::SETUP::UseNonTemporal eq 'Y') ? '1' : undef);
}


#------------------------------------------------------------
# include_nongeo [VALUE]
#------------------------------------------------------------
# Include Non-Spatial Granules in search results
#
sub include_nongeo  { 
   _one('Include_NonSpatial', @_) ;
}

#------------------------------------------------------------
# show_nongeo
#------------------------------------------------------------
# If SETUP flag is not set to 'Y', return undef
#
sub show_nongeo  {
   return ( ($IU::SETUP::UseNonSpatial eq 'Y') ? '1' : undef);
}


#------------------------------
#
# Search options/properties...
#
#------------------------------

#------------------------------------------------------------
# max_granules [VALUE]
#------------------------------------------------------------
# Set/get maximum number of granules that a DAAC should return.
#
sub max_granules { 
    my ($self, $value) = @_;
    $self->{MaxGrans} = $value if @_ > 1;
    $self->{MaxGrans} || $MAX_GRANS_DEFAULT;
}

#------------------------------------------------------------
# max_guides [VALUE]
#------------------------------------------------------------
# Set/get maximum number of guide documents that a DAAC should return.
#
sub max_guides { 
    my ($self, $value) = @_;
    $self->{MaxGuides} = $value if @_ > 1;
    $self->{MaxGuides} || $MAX_GUIDES_DEFAULT;
}

#------------------------------------------------------------
# max_time [VALUE]
#------------------------------------------------------------
# Set/get maximum number of minutes that a search can run.
#

sub max_time {
    my ($self, $value) = @_;
    $self->{MaxTime} = $value if @_ > 1;
    $self->{MaxTime} || $MAX_TIME_DEFAULT;
}

#------------------------------------------------------------
# browse_only [VALUE]
#------------------------------------------------------------
# Only return granules with browse products?
#
sub browse_only  { _one('BrowseOnly', @_) }

#------------------------------------------------------------
# all_metadata [VALUE]
#------------------------------------------------------------
# Return all metadata in search results?
#
sub all_metadata  { 

    my ($self, $value) = @_;

    $self->{AllMetadata} = $value if (@_ > 1);    
    $self->{AllMetadata};
 }

#------------------------------------------------------------
# day_night [VALUE]
#------------------------------------------------------------
# Data gathered during Daylight only or Nighttime only?
#
sub day_night  { _one('DayNightFlag', @_) }

#------------------------------------------------------------
# name [NAME]
#------------------------------------------------------------
# Set/get the name of the query.
#
sub name  { _one('Name', @_) }

#------------------------------------------------------------
# filename
#------------------------------------------------------------
# Get a good filename for the query.
#
sub filename {
    my $self = shift;
    my $name = $self->name || 'untitled';
    $name =~ s/^\s+|\s+$//g;        # trim
    $name =~ s/[^a-z0-9_+-]/_/gi;   # turn illegals into underscores
    $name =~ s/_+/_/g;              # collapse underscores
    "$name.query";
}

#------------------------------------------------------------
# ODL_search_type
#------------------------------------------------------------
# Return the search type, as it would appear in the ODL.
#
sub ODL_search_type { 
    my $self = shift;
    my $type = $self->search_type;
    return ((($type eq 'GRANULE_SEARCH') || ($type eq 'BROWSE_SEARCH')
    || ($type eq 'LOCAL_GRANULE_SEARCH')) ? 'INVENTORY_SEARCH' : $type);
}

#------------------------------------------------------------
# search_type [VALUE]
#------------------------------------------------------------
# Set/get the search type, in the most-detailed sense.
# Compare with ODL_search_type.
#
sub search_type { _one('SearchType', @_) }

#------------------------------
#
# Dependent valids...
#
#------------------------------

#------------------------------------------------------------
# depvalid_interface [OPTIONS]
#------------------------------------------------------------
# I<Instance method.>
# Return a dependent-valids interface object preset with all the 
# query criteria (both core and extended).
#
# The OPTIONS is a hash which can contain the following:
#
#     Exclude => If set to a (true) category name, skip this category
#                name when setting up the interface object.

sub depvalid_interface {
    my ($self, %opts) = @_;
    my $exclude_category = $opts{Exclude} || '';

    # Create a new dependent valids interface:
    my $depvalid = new IW::DepValid ( defined($opts{Session_Path}) ? (Session_Path =>$opts{Session_Path}) : ()  );

    # Set all info we might depend on:
    my $cat;
    foreach $cat ($self->categories) {      # for each category in query

	next if ($cat eq $exclude_category);       # skip this category?
	$self->uses($cat) or next;                 # only if using it!

	# Get the valids, as appropriate:
	my @valids = ();
	if (IU::Category->is_core($cat)) {         # core dependent valids
	    @valids = $self->criteria_strings($cat); 
	}
	elsif (IU::Category->is_extended($cat)) {  # extended dependent valids
	    @valids = map {$_->variant} $self->criteria($cat);
	}
	@valids or next;

	# Set them:
	$depvalid->set($cat, \@valids);              # set them
    }

    $depvalid;
}

#------------------------------------------------------------
# valids_for CATEGORY, [OPTIONS]
#------------------------------------------------------------
# I<Instance method.>
# Given a dependent-valids CATEGORY (either a core category like
# PARAMETER, or an extended category like CLOUD_COVER), return
# a list of strings which are the valids (or variant names) for
# that category.
#
#     CATEGORY is...    Returns array of...
#     -----------------------------------------------
#     Core depval       valids
#     Extended          variant names   
#
# Note that we have to use the dependent valids interface for
# each category we test.
#
# The OPTIONS is TBD.

sub valids_for {
    my ($self, $category, %opts) = @_;
    $self->depvalid_interface->get_valids($category);
}


#------------------------------------------------------------
# which_daacs_need_authenticator  
#------------------------------------------------------------
#
# returns array of daacs

sub which_daacs_need_authenticator {

    my $self = shift;
    $self->determine_daacs_to_search;
    my @initialDaacs = $self->daacs_to_search;

    # Open static valids database:
    my %WWWVALIDS;
    if (-e "${IU::LibDir}/conf/wwwvalids.gdb") {
        eval {tie %WWWVALIDS, 'IU::Store', "${IU::LibDir}/conf/wwwvalids.gdb",
              'READER'};
        $@ and return error "tie failed: $@";
    }


    # For each data center we're considering...
    my %needs;        # Daacs needing authenticator. 
    my $archid;
    foreach $archid (@initialDaacs) {
        # Is there an entry for it in the static valids?
        my $valids;
        if ($valids = $WWWVALIDS{"daac:$archid"}) {
            $needs{$archid} = 'V0' if ($valids->{REQUIRES_V0_AUTH}); 
            $needs{$archid} = 'ECS' if ($valids->{REQUIRES_ECS_AUTH}); 
            $needs{$archid} = 'NASDA' if ($valids->{REQUIRES_NASDA_AUTH}); 
        }
    }                                    # end: foreach daac 

    # Close static valids database:
    untie %WWWVALIDS;

    return %needs;
}




#------------------------------------------------------------
# filter_daacs_to_search
#------------------------------------------------------------
# I<Instance method, private.>
# Take the list of data centers to actually query, and pare them down.
# based on the static valids file.  See determine_daacs_to_search().
#
# Raises a fatal exception if no DAACs remain.

sub filter_daacs_to_search {
    my $self = shift;
    my @initialDaacs = $self->daacs_to_search;
    my @filteredDaacs;

    # Open static valids database:
    my %WWWVALIDS;
    if (-e "${IU::LibDir}/conf/wwwvalids.gdb") {
	eval {tie %WWWVALIDS, 'IU::Store', "${IU::LibDir}/conf/wwwvalids.gdb",
	      'READER'};
	$@ and do {
                    # We're outside of the EDG's logging now; just print it
                    # to STDERR and Apache will get it at least.
                    print STDERR "tie failed to wwwvalids.gdb: $@";
                    return error "tie failed: $@";
        }
    }  

    #--------------------
    # create list of data centers needing ECS and V0 authenticator
    #--------------------

    # For each data center we're considering...
    my %needsV0;        # Daacs needing V0 authenticator.
    my %needsECS;       # Daacs needing ecs authenticator.
    my %needsNASDA;     # Daacs needing NASDA authenticator.
    my $daac;

    foreach $daac (@initialDaacs) {
        # Is there an entry for it in the static valids?
        my $valids;
        if ($valids = $WWWVALIDS{"daac:$daac"}) {
            $needsV0{$daac}  = $valids->{REQUIRES_V0_AUTH};
            $needsECS{$daac} = $valids->{REQUIRES_ECS_AUTH};
            $needsNASDA{$daac}  = $valids->{REQUIRES_NASDA_AUTH};
        } 			# end if
    }                                    # end: foreach daac

    my @daacsV0 = grep { $needsV0{$_}} (keys %needsV0);
    my @daacsECS = grep { $needsECS{$_}} (keys %needsECS);
    my @daacsNASDA = grep { $needsNASDA{$_}} (keys %needsNASDA);

    $self->daacs_needing_v0_auth(\@daacsV0);
    $self->daacs_needing_ecs_auth(\@daacsECS);
    $self->daacs_needing_nasda_auth(\@daacsNASDA);

    my %allowed;        # allows GID search? true or false, indexed by daac
    my $archid;
    my $valids;
    my @badDaacs;
    #--------------------
    # GRANULE SEARCH...
    #--------------------
    if ($self->search_type eq 'GRANULE_SEARCH') {

	# Support for search-by-granule-id is determined by the DAAC
	# server: either a server DOES recognize the GRANULE_ID_REQ in the
	# ODL and handles the request properly (failing gracefully
	# if the feature isn't supported), or it DOESN'T.  The older code 
	# which did a dataset-level interpretation of this feature was 
	# EXTREMELY time-consuming, and quite unnecessary.  SOME FEATURES
	# HAVE NOTHING TO DO WITH DATASETS: THEY ARE RELATED DIRECTLY TO DAACS.
	
	# For each data center we're considering...
	foreach $archid (@initialDaacs) {

	    # Is there an entry for it in the static valids?
	    if ($valids = $WWWVALIDS{"daac:$archid"}) {
		$allowed{$archid} = $valids->{ALLOWS_GRANULE_ID};
	    }
	}                                    # end: foreach dataset id

	# Get final list of daacs:
	@filteredDaacs = grep { $allowed{$_}} (keys %allowed); 
	@badDaacs   = grep {!$allowed{$_}} (keys %allowed);
	@filteredDaacs or 
	    puke 544, "search by Data Granule ID not supported for these data ",
	              "centers: ", join(", ", @badDaacs);

    }                                         # end: if a granule id search
    elsif ($self->search_type eq 'PRODUCT_STATUS_REQUEST') {
	# For each data center we're considering...
	foreach $archid (@initialDaacs) {

	    # Is there an entry for it in the static valids?
	    my $valids;
	    if ($valids = $WWWVALIDS{"daac:$archid"}) {
		$allowed{$archid} = $valids->{ALLOWS_ORDER_STATUS};
	    }
	}                                    # end: foreach dataset id

	# Get final list of daacs:
	@filteredDaacs = grep { $allowed{$_}} (keys %allowed); 
	@badDaacs   = grep {!$allowed{$_}} (keys %allowed);
#	@filteredDaacs = @initialDaacs;
	@filteredDaacs or 
	    puke 544, "Product Status Request not supported for these data ",
	              "centers: ", join(", ", @badDaacs);

    }
    #--------------------
    # NOT A GRANULE SEARCH...
    #--------------------
    else {
	# no-op
	@filteredDaacs = @initialDaacs;
    }

    # Close static valids database:
    untie %WWWVALIDS;

    # Set new list of daacs to search:
    $self->data_center_ids(\@filteredDaacs);
    $self->daacs_to_search(\@filteredDaacs);
    1;
}

#------------------------------------------------------------
# determine_datasets_to_search
#------------------------------------------------------------
# I<Instance method.>
# Determine list of data sets to actually query, using dependent valids.
# Fatal exception if no data sets can be set.

sub determine_datasets_to_search {
    my $self = shift;

    # Get a dependent-valids interface object for this query:
    require IU::Store;
    my $depval = $self->depvalid_interface(Exclude=>'DATASET_ID');

    # STEP 1: Set the data sets, as dictated solely by dependent valids:
    my @validDatasets = $depval->get_valids('DATASET_ID');
    @validDatasets or
        puke 517, "All of your search attributes are incompatible.",
        " Please return to the search form and make compatible selections.";
    $self->datasets_to_search(\@validDatasets);

    # STEP 2: Intersect with the list of *desired* datasets, if any were given:
    if ($self->dataset_ids) {      # user specified some datasets...
        my @finalDatasets = intersect([$self->dataset_ids],
                                   [$self->datasets_to_search]);
        @finalDatasets or puke 517, "You left me with no datasets to search!";
        $self->datasets_to_search(\@finalDatasets);
    }

    # Done!
    1;
}

#------------------------------------------------------------
# determine_browse_datasets_to_search
#------------------------------------------------------------
# I<Instance method.>
# Determine list of browse data sets to actually query, using dependent valids.

sub determine_browse_datasets_to_search {
    my $self = shift;
    my $datasetref = shift;
    my @oldDatasets = @$datasetref;

    # Get a dependent-valids interface object for this query:
    require IU::Store;
    my $depval = $self->depvalid_interface(Exclude=>'DATASET_ID');

    # STEP 1: Set the data sets, as dictated solely by dependent valids:
    my @validDatasets = $depval->get_valids('DATASET_ID');
    # STEP 2: Intersect with the list of *desired* datasets, if any were given:
    if (@oldDatasets && @validDatasets) {      
    # user specified some datasets...
        my @finalDatasets = intersect(\@oldDatasets, \@validDatasets);
        $self->dataset_ids(\@finalDatasets);
    }
    # Done!
    return $self;
}

#------------------------------------------------------------
# determine_daacs_to_search
#------------------------------------------------------------
# I<Instance method.>
# Determine list of data centers to actually query, using dependent valids.
# Fatal exception if no DAACs can be set.

sub determine_daacs_to_search {
    my $self = shift;

    # Get a dependent-valids interface object for this query:
    require IU::Store;
    my $depval = $self->depvalid_interface(Exclude=>'DATA_CENTER_ID');

    # STEP 1: Set the data centers, as dictated solely by dependent valids:
    my @validDaacs = $depval->get_valids('DATA_CENTER_ID');
    @validDaacs or 
        puke 517, "All of your search attributes are incompatible.",
        " Please return to the search form and make compatible selections.";
    $self->daacs_to_search(\@validDaacs);

    # STEP 2: Intersect with the list of *desired* daacs, if any were given:
    if ($self->data_center_ids) {      # user specified some daacs...
	my @finalDaacs = intersect([$self->data_center_ids], 
				   [$self->daacs_to_search]);
	@finalDaacs or puke 517, "You left me with no data centers to search!";
	$self->daacs_to_search(\@finalDaacs);


    }

   # STEP 3: Filter the data centers, based on the static valids:
    $self->filter_daacs_to_search;
    ($self->daacs_to_search) or
        puke 460, "after filtering, no searchable data centers remained!";

    # Done!
    1;
}

#------------------------------------------------------------
# validate_daacs_to_search
#------------------------------------------------------------
# I<Instance method.>
# Determines if the Data Centers listed in the QUERY object are valid
# sites to search
#
# INPUT : nothing , becisdes the QUERY Object
# OUTPUT : The self object with a ref of array in {Good} and {Bad}
# to list the good/bad sites

sub validate_daacs_to_search {
    my $self = shift;

    # Get the list of known Daacs from the archive.odl
    my @valid_daacs = @{parse_archive_odl()};

    # List of Daacs located in the QUERY object
    my @Daac_criteria = $self->data_center_ids;

    # Ripped from the Perl Cookbook #4.7
    # Finding Elements in One Array but Not Another

    my %seen ; # lookup table
    my (@good_daacs,@bad_daacs);

    @seen{@valid_daacs} = ();
   
    my ($is_valid_daac,$valid_daac_list);
    foreach $is_valid_daac ( @Daac_criteria) { 
      # Find Daacs which shouldn't be searched on :
      push ( @bad_daacs, $is_valid_daac) unless exists $seen{$is_valid_daac};

      # Find Daacs which should be searched on :
      push ( @good_daacs, $is_valid_daac) if exists $seen{$is_valid_daac};
    }

    $self->{Good} = \@good_daacs;    
    $self->{Bad} = \@bad_daacs;
    return ($self);
    # return (  \@good_daacs, \@bad_daacs);
}

#------------------------------------------------------------
# parse_archive_odl
#------------------------------------------------------------
# Get the list of Daacs presently located in the archive.odl
#
sub parse_archive_odl {
    my $archive_odl = new IU::ArchiveOdl;
    my %archive_odl_parsed = $archive_odl->parse_ODL;
    my @valid_daacs = keys %archive_odl_parsed;
    my $strinme = (join(' ',@valid_daacs));
    return ( \@valid_daacs);
}


#------------------------------
#
# Categories and criteria...
#
#------------------------------

#------------------------------------------------------------
# categories
#------------------------------------------------------------
# I<Instance method.>  Return all categories in query.

sub categories {
    my $self = shift;
    keys %{$self->{Categories}};
}

#------------------------------------------------------------
# criteria CATEGORY, [CRITERIA]
#------------------------------------------------------------
# Instance method, for convenience.
#
# With just CATEGORY, returns array of criteria (or empty array if none).
#
# With CATEGORY and CRITERIA, sets CATEGORY's criteria to be the
# given CRITERIA, which should be a ref to an array of Criterion objects.
# If you supply a non-arrayref, it will be assumed to be the sole criterion.  
#
# Use a CRITERIA of undef (or a ref to the empty array) to erase all 
# criteria for this category (and the category as well).

sub criteria {
    my ($self, $cat, $values) = @_;
    if (@_ > 2) {             # set new values
	my @safe;             # safe place for new values

	# Coerce values to be either empty, or array of one criterion:
	$values ||= [];                          # undef/false to empty array
	@safe = ((ref($values) eq 'ARRAY') ? @$values : ($values));
	
	# Set:
	(@safe ? $self->{Categories}{$cat} = \@safe 
	       : delete $self->{Categories}{$cat});
    }
    @{$self->{Categories}{$cat} || []};         # return
}

#------------------------------------------------------------
# criteria_strings CATEGORY, [STRINGS]
#------------------------------------------------------------
# Instance method, for convenience.
#
# With just CATEGORY, returns array of values (or empty array)
# held in the first criterion, which is assumed to be a "SelectMany".
#
# With CATEGORY and STRINGS, where STRINGS is a ref to an array of strings,
# creates a new "SelectMany" criterion with the given values, and stores 
# it as the only criterion for this category.

sub criteria_strings {
    my ($self, $cat, $strings) = @_;
    my $crit;            # note: initially undefined!

    # Set new values?
    if (@_ > 2) {    
	if (@$strings) {       # some... create and set criteria!
	    $crit = IU::Criterion::SelectMany->new;
	    $crit->field('VALUES', $strings);
	}
	$self->criteria($cat, $crit);
    }

    # Return them:
    $crit = ($self->criteria($cat))[0];          # get first criterion
    return ($crit ? @{$crit->field('VALUES') || []} : ()); 
}

#------------------------------------------------------------
# fixup
#------------------------------------------------------------
# After reading in a query from a foreign source - a file, a set
# of CGI tags, etc. - we may need to adjust some things:
#
#    * Default values may need to be filled in
#    * A fixup() message is sent to the spatial search criterion
#    * etc.
#
# Returns the "self" object on success.

sub fixup { 
    my $self = shift;

    # Fix up the number of granules:
    defined($self->{MaxGrans}) or $self->{MaxGrans} = $MAX_GRANS_DEFAULT;
    $self->{MaxGrans} = $MAX_GRANS_MIN if ($self->{MaxGrans} < $MAX_GRANS_MIN);
    $self->{MaxGrans} = $MAX_GRANS_MAX if ($self->{MaxGrans} > $MAX_GRANS_MAX);

    # Fix up the number of granules:
    defined($self->{MaxGuides}) or $self->{MaxGuides} = $MAX_GUIDES_DEFAULT;

    # Fix up the search region (if any):
    my $geo = $self->geo_region; $geo and $geo->fixup;

    # Done!
    $self;
}

#------------------------------------------------------------
# simple_incomplete
#------------------------------------------------------------
# I<Instance method.>
# Only need data set valids for simple search.
# Based on the search type, are we missing some required fields?
# If so, returns a string (true) describing what's missing.
# If not, returns 0 (false).

sub simple_incomplete {
    my $self = shift;
    if ($self->search_type eq 'INVENTORY_SEARCH') {
        unless ($self->dataset_ids) {
            unless ($self->geo_region) {
                return "You must supply at least one Data Set ".
                       "AND you must select a Geographic ".
                       "Region.  Try to select as small a region as ".
                       "possible.";
            }
            return "You must supply at least one Data Set. ";
        }
        ($self->geo_region)
            or return "You need to select a Geographic Region. ".
                      "Try to select as small a region as possible.";
    }
    return 0;
}

#------------------------------------------------------------    
# what_now
#------------------------------------------------------------    
# I<Instance method.>
# Based on the search type, are we missing some required fields?
# Returns an array of items still needed.

sub what_now {
    my $self = shift;
    my @needed = ();
    if ($self->search_type eq 'INVENTORY_SEARCH') {
	push @needed, 'a parameter, sensor, or data set' 
	    unless ($self->parameters   or
		    $self->sensor_names or
		    $self->dataset_ids);
	push @needed, 'a geographic region' unless $self->geo_region;
    }
    elsif ($self->search_type eq 'GRANULE_SEARCH') {
	push @needed, 'a data set'   unless $self->dataset_ids;
	push @needed, 'a data granule ID' unless $self->granule_ids;
    }
    elsif ($self->search_type eq 'LOCAL_GRANULE_SEARCH') {
	push @needed, 'a data set'   unless $self->dataset_ids;
	push @needed, 'a data granule ID' unless $self->local_granule_id;
    }
    elsif ($self->search_type eq 'BROWSE_SEARCH') {
        push @needed, 'a data set'   unless $self->dataset_ids;
        my ($start_tz, $stop_tz, $start_doy, $stop_doy) = $self->time_range_TZ;
        if(!$start_tz || !$stop_tz) {
          push @needed, 'a time range';
        }
    }
    elsif ($self->search_type eq 'PRODUCT_STATUS_REQUEST') {
	push @needed, 'a data center' unless $self->data_center_ids;
	push @needed, 'an order tracking number', unless $self->request_ids;
    }
    elsif ($self->search_type eq 'ACCOUNT_STATUS_REQUEST') {
	push @needed, 'a data center' unless $self->data_center_ids;
    }
    elsif ($self->search_type eq 'GUIDE_SEARCH') {
	push @needed, 'a parameter, data set, sensor, data center, source, '.
	    'campaign, or text string' unless
		($self->campaigns       or
		 $self->data_center_ids or
		 $self->dataset_ids     or
		 $self->parameters      or
		 $self->sensor_names    or
		 $self->source_names    or
		 $self->text_string);
    }
    elsif ($self->search_type eq 'DIRECTORY_SEARCH') {
	# Nothing required
    }
    else {
	puke "Unknown search type (", $self->search_type, ")!";
    }

    # Return the results
    return @needed;
}

#------------------------------------------------------------    
# incomplete
#------------------------------------------------------------    
# I<Instance method.>
# Based on the search type, are we missing some required fields?
# If so, returns a string (true) describing what's missing.
# If not, returns 0 (false).

sub incomplete {
    my $self = shift;

    if ($self->search_type eq 'INVENTORY_SEARCH') {
	my @cats = $self->categories;
	if (($cats[0] eq 'TIME_RANGE' && !$cats[2]) ||
            ($cats[0] eq 'TIME_RANGE' && !$cats[1]) ||
            ($cats[0] eq 'GEOGRAPHIC_REGION' && !$cats[1]) || !$cats[0]) {
	    unless ($self->geo_region) {
		return "You must supply at least one search keyword, ".
		       "AND you must select a Geographic ".
		       "Region.  Try to select as small a region as ".
		       "possible.";
	    }
	    return "You must supply at least one search keyword.";
	}
	($self->geo_region)
	    or return "You need to select a Geographic Region. ".
		      "Try to select as small a region as possible.";
    }
    elsif ($self->search_type eq 'GRANULE_SEARCH' ) {
	unless ($self->dataset_ids) {
	    unless ($self->granule_ids) {
		return "You must supply at least one Data Set and a ".
		       "Data Granule ID.";
	    }
	    return "You must supply at least one Data Set.";
	}
	($self->granule_ids)
	    or return "You must supply a Data Granule ID.";
    }
    elsif ($self->search_type eq 'BROWSE_SEARCH') {
        my ($start_tz, $stop_tz, $start_doy, $stop_doy) = $self->time_range_TZ;

        unless ($self->dataset_ids) {
          unless ( $start_tz || $stop_tz ) {
                return "You must supply at least one Data Set and a ".
                       "Date range.";
            }
            return "You must supply at least one Data Set.";
        }
        unless ( $start_tz || $stop_tz ) { return "You must supply a Date range."; }
    }
    elsif ($self->search_type eq 'LOCAL_GRANULE_SEARCH') {
        
        my $local = $self->get_local_granule_id;
        
	unless ($self->dataset_ids) {
	    if ($self->local_granule_id) {
                #there is a local granule id in the query, but is
                # it the fake lgid of none selected?or is it blanks?
                if (($local eq "None selected") ||(IW::Util::blank($local))){
		    return "You must supply at least one Data Set and a ".
		       "Local Granule ID.";
                }
            }
            else { # there is no local granule id
          
    
		    return "You must supply at least one Data Set and a ".
		       "Local Granule ID.";                
	    }
	    return "You must supply at least one Data Set.";
	}

        if ($self->local_granule_id){
            #there is a local granule id in the query, but is
            # it the fake lgid of none selected? is it blank?
            if (($local eq "None selected" )|| (IW::Util::blank($local))){
            
               return "You must supply a Local Granule ID.";
            }
        }else { #there is no local granule id
               return "You must supply a Local Granule ID.";
        }

    }
    elsif ($self->search_type eq 'PRODUCT_STATUS_REQUEST') {
	unless ($self->data_center_ids) {
	    unless ($self->request_ids) {
		return "You must supply at least one Data Center and one ".
		       "Order Tracking Number.";
	    }
	    return "You must supply at least one Data Center.";
	}
	($self->request_ids)
	    or return "You must supply an Order Tracking Number.";
    }
    elsif ($self->search_type eq 'ACCOUNT_STATUS_REQUEST') {
	unless ($self->data_center_ids) {
	    return "You must supply at least one Data Center.";
	}
    }
    elsif ($self->search_type eq 'GUIDE_SEARCH') {
	($self->campaigns       ||
	 $self->data_center_ids ||
	 $self->dataset_ids     ||
	 $self->parameters      ||
	 $self->sensor_names    ||
	 $self->source_names    ||
	 $self->text_string)
	    or return "You must supply at least one of these: Campaign, ".
		      "Data Center, Data Set, Parameter, Sensor, Source, ".
		      "or Text String.";
    }
    elsif ($self->search_type eq 'DIRECTORY_SEARCH') {
	# no-op
    }
    else {
	puke "Unknown search type (", $self->search_type, ")!";
    }
    return 0;
}

#------------------------------------------------------------    
# install
#------------------------------------------------------------    
# I<Instance method.>  WG-specific.
# Re-install this as the current query in the session state.

sub install {
    my $self = shift;
    $SESSION->state->{Query} = $self;
    my $srchtype = $self->search_type;
    # Also store a copy of the query by search type so when we toggle we
    # can restore the query for that type:
    $SESSION->state->{'Query'.$srchtype} = $self;
}

#------------------------------------------------------------    
# lookup
#------------------------------------------------------------ 
# I<Class method.  WG-specific.>
# Lookup current query in the session state.  Returns undef if none so far.

sub lookup {
    my $class = shift;
     
    return $SESSION->state->{Query};
}

#------------------------------------------------------------
# default
#------------------------------------------------------------
# I<Class method.  WG-specific.>
# build a default query from template in lib/conf

sub default {
    my ($self,$type) = @_;

    # list of templates
    my %templates = ('INVENTORY_SEARCH' => 'query_inventory.template',
                     'BROWSE_SEARCH'    => 'query_browse.template');

    # Deterine which template to use based on search type requested.
    # Default is INVENTORY.
    my $template = $templates{$type} || $templates{'INVENTORY_SEARCH'};
    $self->reload("${IU::LibDir}/conf/$template");
}

#------------------------------------------------------------
# change_type 
#------------------------------------------------------------
# Get query from the SESSION->state based on the New search type
# if there is none then, morph the current query object into the
# new type using set_dummy_query method. 


sub change_type {

    my ($self,%params) = @_;

    my $newsrchtype = $params{NewSrchType};
    my $prevsrchtype = $params{PrevSrchType};

    my $storedquery = $SESSION->state->{'Query'.$newsrchtype} || undef;
    if (defined($storedquery)) {

       return $storedquery;

    } else {

       return $self->set_dummy_query(NewSrchType=>$newsrchtype,
                                     PrevSrchType=>$prevsrchtype);
    }
}


#------------------------------------------------------------
# reload PATH
#------------------------------------------------------------
# recreate query using the supplied ODL file

sub reload {

    my ($self, $path) = @_;

    # Be a class method, too!
    ref($self) or $self = $self->new;

    if (! -s $path) {
       puke "no query file: $path";
    }
    # Parse it into an ODL object:
    open ODL, $path or puke "open $path: $!";
    my $odl = eval { parse IU::ODL \*ODL };
    close ODL;
    $odl or puke "couldn't parse ODL in $path (line $.)";

    # Convert the ODL into a query object:
    $self->from_ODL($odl,KeepType=>'1')
       or puke "couldn't extract valid user data from ODL";

    # Done!
    return $self;
}

#------------------------------------------------------------
# uses CATEGORY, [SEARCHTYPE]
#------------------------------------------------------------
# I<Class/instance method.>
# Returns true if the given CATEGORY would be a functional part of the 
# query under the given SEARCHTYPE, and false otherwise.   With
# no SEARCHTYPE, will use the query's search type.

sub uses {
    my ($self, $cat, $type) = @_;
    $type ||= $self->search_type;  
    if ($type eq 'GRANULE_SEARCH') {
	return 1 if ($cat eq 'DATASET_ID');
	return 1 if ($cat eq 'DATA_CENTER_ID');
	return 1 if ($cat eq 'GRANULE_ID_REQ');
	return 1 if ($cat eq 'TIME_RANGE');
	return 1 if ($cat eq 'GEOGRAPHIC_REGION');
	return 0;
    }
    elsif ($type eq 'PRODUCT_STATUS_REQUEST') {
	return 1 if ($cat eq 'DATA_CENTER_ID');
	return 1 if ($cat eq 'REQUEST_ID');
	return 0;
    }
    elsif ($type eq 'ACCOUNT_STATUS_REQUEST') {
	return 1 if ($cat eq 'DATA_CENTER_ID');
	return 0;
    }
    elsif ($type eq 'GUIDE_SEARCH') {
	return 1 if (IU::Category->is_core($cat) and 
		     ($cat ne 'PROCESSING_LEVEL'));
	return 1 if ($cat eq 'TEXT_STRING');
	return 0;
    }
    elsif ($type eq 'DIRECTORY_SEARCH') {

        return 0 if ($cat eq 'TEXT_STRING');
        return 0 if ($cat eq 'GRANULE_ID_REQ');
        return 0 if ($cat eq 'EXTENDED_SEARCH');
        return 0 if ($cat eq 'PROCESSING_LEVEL');
	return 0 if ($cat eq 'REQUEST_ID');
        return 1;
 
    }
    elsif ($type eq 'LOCAL_GRANULE_SEARCH') {
	return 0 if ($cat eq 'DATA_CENTER_ID');
	return 0 if ($cat eq 'TIME_RANGE');
	return 1 if ($cat eq 'DATASET_ID');
	return 1 if ($cat eq 'EXTENDED_SEARCH');
        return 1 if ($cat eq 'LOCAL GRANULE ID');
        return 1 if ($cat eq 'GRANULE DETAILS');
	return 1 if ($cat eq 'GEOGRAPHIC_REGION');
	return 0;
    }
    elsif ($type eq 'INVENTORY_SEARCH') {
	return 0 if ($cat eq 'TEXT_STRING');
	return 0 if ($cat eq 'GRANULE_ID_REQ');
	return 0 if ($cat eq 'REQUEST_ID');
	return 1 if ($cat eq 'EXTENDED_SEARCH');
	return 1;
    }
    else {         # any other kind of search
	return 1;        
    }
}

#------------------------------------------------------------
# uses_opt OPTION, [SEARCHTYPE]
#------------------------------------------------------------
# I<Class/instance method.>
# Much like uses(), but for miscellaneous query options that aren't
# really categories.  Valid OPTIONs are:
#
#     QUERY_NAME       Query name
#     GRANULE_LIMIT    Maximum number of granules per dataset
#     BROWSE_ONLY      Only return granules with browse products?
#     GUIDE_LIMIT      Maximum number of guide documents per data center.
#
# Returns true if the given OPTION would be a functional part of the 
# query under the given SEARCHTYPE, and false otherwise.   With
# no SEARCHTYPE, will use the query's search type.
# Always show guide limit since it will take effect if the user selects
# Detailed Info while doing any kind of search.

sub uses_opt {
    my ($self, $opt, $type) = @_;
    if ($opt eq 'GUIDE_LIMIT') {
	return 1;
    }
    $type ||= $self->search_type;
    if ( $type eq 'INVENTORY_SEARCH') {
	return 1;          # all options
    }
    elsif (($type eq 'GUIDE_SEARCH') ||
	   ($type eq 'DIRECTORY_SEARCH') ||
	   ($type eq 'PRODUCT_STATUS_REQUEST') ||
	   ($type eq 'ACCOUNT_STATUS_REQUEST')) {
	return 1 if ($opt eq 'QUERY_NAME');
	return 0;
    }
    elsif ($type eq 'GRANULE_SEARCH') {
        return 1 if ($opt eq 'GRANULE_LIMIT');
        return 1 if ($opt eq 'TIME_LIMIT');
        return 1 if ($opt eq 'QUERY_NAME');
        return 0;
    }
    elsif($type eq 'LOCAL_GRANULE_SEARCH') {
        return 1 if ($opt eq 'GRANULE_LIMIT');
        return 1 if ($opt eq 'TIME_LIMIT');
        return 1 if ($opt eq 'QUERY_NAME');
        return 0 ;
    }
    elsif ($type eq 'BROWSE_SEARCH') {
        return 1 if ($opt eq 'GRANULE_LIMIT');
        return 1 if ($opt eq 'TIME_LIMIT');
        return 1 if ($opt eq 'QUERY_NAME');
        return 0;
    }
    else {         # any other kind of search
	return 1;        
    }
}

#------------------------------------------------------------    
# validate
#------------------------------------------------------------    
# I<Instance method.>
# Based on the search type, are we ready to go?
# This is an assertion: it returns the self object on success, but on failure
# it raises a fatal exception which describes why validation failed. 

sub validate {
    my $self = shift;

    # Is user not done?
    my $problem = $self->incomplete;
    puke 516, $problem if $problem;
    $self;
}

#------------------------------
#
# ODL interface
#
#------------------------------

#------------------------------------------------------------
# from_ODL ODLOBJECT, PARAMHASH
#------------------------------------------------------------
# I<Class or instance method.>
# Initiallize the address from a standard IMS ODL object.
#
# The PARAMHASH can contain the following:
#
#     KeepType => If true, the search type will be discarded in 
#                 favor of the type of any CURRENTLY-INSTALLED query.
#                 (See lookup() and install() for details).
#                 If false (default), the ODL alone is used in determining
#                 the search type. 

sub from_ODL {
    my ($self, $odl, %params) = @_;
    $odl || return undef;
    my ($param, $value);
    my $search;
    my $keep_type = $params{KeepType};

    # Be a class method, too!
    ref($self) or $self = $self->new;
    #
    # DETERMINE THE SEARCH TYPE...
    #
    #    If "keep type" is true, determine the search type as given by 
    #    any currently-installed query, and prefer that value to the 
    #    actual one given by the ODL.
    #
    #    This is done so that when the user is UPLOADING an ODL search file,
    #    we automatically coerce it into the type of search form 
    #    (INVENTORY, DIRECTORY, GRANULE, etc.) that they are uploading into.

    # Determine the search type as given by the currently-installed query:
    my $prev_search_type;
    if ($keep_type) {
	my $prev_query = ref($self)->lookup;        # don't put it in $QUERY!
	$prev_search_type = $prev_query && $prev_query->search_type;
    }

    # Determine the search type as given by the ODL itself.  
    #    NOTE: this is not necessarily the actual search type of the
    #    top-level message... rather, it is the search type as derived
    #    from the ODL itself.  Notice the case for GRANULE_SEARCH.
    my $new_search_type;
    my $special_crit;
    my $cat_name;
    my $local_gid;
    if ($search = $odl->param('DIRECTORY_SEARCH')) {
	$new_search_type = 'DIRECTORY_SEARCH';
    }
    elsif ($search = $odl->param('INVENTORY_SEARCH')) {
    # since the odl for local granule id is identical no matter if coming
    # from the main search page or the local granule id page, don't use
    # the odl to determine the search type, use the search type saved in
    # the session.
        $new_search_type = $SESSION->state->{LastSearchType};
    }
    elsif ($search = $odl->param('GUIDE_SEARCH')) {
	$new_search_type = 'GUIDE_SEARCH';
    }
    elsif ($search = $odl->param('PRODUCT_STATUS_REQUEST')) {
	$new_search_type = 'PRODUCT_STATUS_REQUEST';
    }
    elsif ($search = $odl->param('ACCOUNT_STATUS_REQUEST')) {
	$new_search_type = 'ACCOUNT_STATUS_REQUEST';
    }
    else {
        puke "ODL doesn't seem to hold INVENTORY_SEARCH, DIRECTORY_SEARCH, ",
	     "GUIDE_SEARCH, PRODUCT_STATUS_REQUEST or ACCOUNT_STATUS_REQUEST.";
    }

    # Now, resolve the two search types, preferring the CGI input:
    $self->search_type($prev_search_type || $new_search_type);
    # 
    # EXTRACT THE SEARCH CONSTRAINTS FROM THE ODL...
    #
    ### UNFINISHED... the original code filtered out certain categories
    ### if this search was for granule ids... should we do the same?
    $search->rewind;
    my %queue;
    my $found = 0;  # flag for a path-row value was found

    while (($param, $value) = $search->next) {

	# Core dependent valids.
	if ($SeqStringCore{$param}) {
	    $self->criteria_strings($param, [$value->text_values]);
	}
        # Other sequence strings. 
        elsif ($SeqStringCategories{$param}) {
            $self->criteria_strings($param, [$value->text_values]);
        }

	# Simple strings...
	elsif ($StringCategories{$param}) {
	    $self->criteria_strings($param, [join '', $value->text_values]);
	}

	# Geographic region...
	elsif ($param eq 'GLOBAL_GRANULES_ONLY') {
	    $self->geo_region(IU::Geo::Global->new);
	}
	elsif ($param eq 'RANGE_LOC' && !$found) {
	    $self->geo_region(IU::Geo::Range->from_ODL($value));
	}
	elsif ($param eq 'POLYGON_LOC') {
	    $self->geo_region(IU::Geo::Polygon->from_ODL($value));
	}
	elsif ($param eq 'POINT_LOC') {
	    $self->geo_region(IU::Geo::Point->from_ODL($value));
	}
	elsif ($param eq 'PATH_ROW_LOC') {
	    $self->geo_region(IU::Geo::PathRow->from_ODL($value));
            $found = 1; # flag for path-row item was found
	}
	elsif ($param eq 'INCLUDE_NON_SPATIAL') {
            $self->include_nongeo(uc($value->text_value) eq 'Y');
        }

	# Time range...
	elsif ($param eq 'START_DATE') {
	    $queue{$param} = $value->text_value;
	}
	elsif ($param eq 'STOP_DATE') {
	    $queue{$param} = $value->text_value;
	}
	elsif ($param eq 'START_DAY_OF_YEAR') {
	    $queue{$param} = $value->text_value;
	}
	elsif ($param eq 'STOP_DAY_OF_YEAR') {
	    $queue{$param} = $value->text_value;
	}
        elsif ($param eq 'INCLUDE_NON_TEMPORAL') {
            $self->include_nontime(uc($value->text_value) eq 'Y');
        }

	# Options...
	elsif ($param eq 'BROWSE_ONLY') {
	    $self->browse_only(uc($value->text_value) eq 'Y');
	}
	elsif ($param eq 'ALL_METADATA') {
	    $self->all_metadata($value->text_value);
	}
        elsif ($param eq 'DAY_NIGHT') {
            $self->day_night($value->text_value);
        }
	elsif ($param eq 'GRANULE_LIMIT') {
	    $self->max_granules($value->text_value);
	}
	elsif ($param eq 'GUIDE_LIMIT') {
	    $self->max_guides($value->text_value);
	}
	elsif ($param eq 'TIME_LIMIT') {
	    $self->max_time($value->text_value);
	}
	elsif ($param eq 'SEARCH_NAME') {
	    $self->name($value->text_value);
	}

        # Misc...
        elsif ($param eq 'MESSAGE_ID') {
            # just skip it quietly
        }

        # Web Gateway specific...

        elsif ($param eq 'WG_MISSIONS') {
            $self->missions([$value->text_values]);
        }
        elsif ($param eq 'WG_SRCHTYPE') {
           $self->search_type($value->text_value);
        }
        elsif ($param eq 'WG_OPTSTYPE') {
           $self->optstype($value->text_value);
        }
        elsif ($param eq 'WG_GEOTYPE') {
           $self->geotype($value->text_value);
        }
        elsif ($param eq 'WG_TIMETYPE') {
           $self->timetype($value->text_value);
        }
        elsif ($param eq 'WG_DEPVALIDTYPE') {
           $self->depvalidtype($value->text_value);
        }

	elsif ($param eq 'EXTENDED_SEARCH') {

	    my $crit_odl;
	    $value->rewind;
	    while ($crit_odl = $value->next_param('SPECIALIZED_CRITERIA')) {

		# Get category and variant:
		my $crit_cat = $crit_odl->param('CATEGORY_NAME')->text;

		my $crit_var = $crit_odl->param('VARIANT')->text;
                
		# if uploaded old query change to currently used category
                if(($crit_cat eq 'Local Granule ID') && ($crit_var eq 'LocalGranuleID')){
                     $crit_var = 'LOCALGRANULEID';  
                     $crit_cat = 'GRANULE DETAILS';  
                }
                
		# Find this criterion's spec in the global database:
		my $crit = IW::DepValid->get_variant($crit_cat, $crit_var);

		if (!$crit) {
		    warn "I've never heard of criterion $crit_cat/$crit_var!";
		    next;
		}

		# Take user's value(s), and set them in criterion object:
		$crit->read_ODL($crit_odl);

		# Add the criterion object to the appropriate category:

                if ($new_search_type eq 'LOCAL_GRANULE_SEARCH') {
		    my $crit_value = $crit_odl->param('CRITERIA_VALUE')->text;
                    
                    $self->set_local_granule_id($crit_value);
                }else {
		    $self->{Categories}{$crit_cat} ||= [];
		    push @{$self->{Categories}{$crit_cat}}, $crit;
                }
	     }
	}
	else {
	    warn "skipping unrecognized search parameter: '$param'";
	}
    }    

    # Set time range, if one was given:
    if ($queue{START_DATE} || $queue{STOP_DATE}) {
	$self->time_range_TZ($queue{START_DATE}, $queue{STOP_DATE},
                             $queue{START_DAY_OF_YEAR},
                             $queue{STOP_DAY_OF_YEAR});
    }
    # Fixup and return.  
    #
    # NOTE THAT WE DO NOT VALIDATE: the reason for this
    # is that we want to allow users to upload searches of one type into
    # another type.  For example, uploading an INVENTORY_SEARCH into the
    # GRANULE_ID_SEARCH form, where it is not *yet* valid (until they choose
    # a granule id).
    #
    # This won't be a real weakness, since we always validate when the
    # search is submitted.

    $self->fixup;
    $self;
}

#------------------------------------------------------------
# to_ODL PARAMHASH
#------------------------------------------------------------
# Convert query to an ODL object, and return it.
# The ODL object is a group *containing* an INVENTORY_SEARCH group
# (or DIRECTORY_SEARCH, etc.).
#
# No authentication information is placed in the ODL!!!!
#
# PARAMHASH may contain the following:
#
#    NoDaacs   => If true, then the DATA_CENTER_ID group (user-specified
#                 data centers) will *not* be output in the ODL.
#                 Generally, we handle these separately.
#
#    Everything=> By default, ODL is only output for the current search type,
#                 meaning that if an attribute is not used in the current
#                 search type, then it will not appear in the ODL.
#                 If you set this option to true, then *all* criteria will 
#                 be included in the ODL.
#
#    ForSearch => If true, ODL is tweaked to default/modify certain fields
#                 (e.g., stop date) to make DAACs happy.  Normally,
#                 we don't want the user's query criteria mucked with,
#                 so the default is false.
#
# NOTE: for downloading, use the Everything option to output ALL the 
# categories to the ODL, even ones which are unused for this type of search.  
# This is so an inventory search may be uploaded, turned into a granule-id 
# search, and downloaded, but still retain all the inventory search 
# characteristics.
#
# NOTE: the ForSearch option produces *slightly* different ODL.  Normally,
# we return an accurate, uploadable representation of the query.  With
# ForSearch, we ensure that certain ODL fields are defaulted, to deal 
# with the less-tolerant DAACs.
# If the search type is inventory, the query is converted to a data set 
# based query (i.e. based on the valids selected by the user find all
# the compatable data sets and set those in the ODL instead of the selected
# core valids). This is done because a search, for example a SENSOR, to an
# ECS daac is returning more datasets than are in the valids. These extra
# data sets are filtered out by the client but this makes the
# searching much slowing. By setting datasets via the valids we can avoid
# this problem with ECS servers.

sub to_ODL {
    my ($self, %params) = @_;
    my $no_daacs         = $params{NoDaacs};
    my $write_everything = $params{Everything};
    my $for_search       = $params{ForSearch};

    require IU::ODL;

    # Is this a request to generate ODL for doing an inventory search. 
    my $inv = ($self->search_type eq 'INVENTORY_SEARCH') ? $for_search : 0;

    # Is this a request to generate ODL for doing a directory search.
    my $dir = ($self->search_type eq 'DIRECTORY_SEARCH') ? $for_search : 0;

    # Is this a request to generate ODL for doing a browse search.
    my $browse = ($self->search_type eq 'BROWSE_SEARCH') ? 1 : 0;

    # Begin:
    my $root_odl = IU::ODL->new_group;    # outermost group
    my $srch_odl = IU::ODL->new_group;
    my $extsrch_odl;                      # will be created only if needed

    # Start off the file with the search type:
    $root_odl->add_param($self->ODL_search_type => $srch_odl);

    # The message id (do we REALLY need this?)
    $srch_odl->add_param(MESSAGE_ID    => '0'); 

    my $persistent_user = $ENV{REMOTE_USER};
    $srch_odl->add_param(DORRAN_USERNAME => IU::ODL->quoted($persistent_user))
     if($persistent_user ne 'guest');
    # The criteria:
    my $category;
    foreach $category (sort $self->categories) {

	# Skip if we shouldn't be showing this criterion:
	$write_everything or $self->uses($category) or next;

	# The user's desired data centers...
	if ($category eq 'DATA_CENTER_ID') {	    
	    if (not $no_daacs) {        # yes, they're wanted!
		my @values = $self->data_center_ids;
		@values or next;
		$srch_odl->add_param($category => IU::ODL->conv(\@values));
	    }
	}
	
        
	# Parameter, source, sensor, etc... for non inventory
	elsif ($SeqStringCore{$category} && not $inv) {
	    my @values = $self->criteria_strings($category); 
	    @values or next;
	    $srch_odl->add_param($category => IU::ODL->conv(\@values));
	}
	# Parameter, source, sensor, etc... for inventory
        # Do not write DATASET twice
	elsif ($SeqStringNewCore{$category} && $inv) {
	    my @values = $self->criteria_strings($category); 
	    @values or next;
	    $srch_odl->add_param($category => IU::ODL->conv(\@values));
	}
        # Other sequence string categories...
        elsif ($SeqStringCategories{$category}) {
            my @values = $self->criteria_strings($category);
            @values or next;
            $srch_odl->add_param($category => IU::ODL->conv(\@values));
        }

	# Simple strings...
	elsif ($StringCategories{$category}) {
	    my $value = ($self->criteria_strings($category))[0];
	    defined($value) or next;
	    $srch_odl->add_param($category => IU::ODL->quoted($value));
	}
		
	# The geographic region...
	elsif ($category eq 'GEOGRAPHIC_REGION' && not $dir) {

	    my $geo = $self->geo_region || next;
	    my $type = $geo->ODL_type;
     
            # Output basic geo object:
	    $srch_odl->add_param($type => $geo->to_ODL);	    
	    
	    # Output backup region:
	    if ($type eq 'PATH_ROW_LOC') {
		my $range = $geo->bound_rect || puke "no bounding rectangle!";
		$srch_odl->add_param(RANGE_LOC => $range->to_ODL);	    
	    }
            
	}
	# The time range...
	elsif ($category eq 'TIME_RANGE' && not $dir) {
	    my ($start_tz, $stop_tz, $start_doy, $stop_doy) = 
             $self->time_range_TZ;

	    # Create default values, if exactly one endpoint is missing.
	    # ONLY do this for searches: don't muck with user's ACTUAL dates!
	    if ($for_search) {
		if ($stop_tz && !$start_tz) {   # stop given, but no start
		    $start_tz = '1950-01-01T00:00:00Z';
		    warn "defaulting start date to $start_tz";
		}
		if ($start_tz && !$stop_tz) {   # start given, but no stop
		    require 'POSIX.pm';
		    $stop_tz = POSIX::strftime("%Y-%m-%dT%H:%M:%SZ", gmtime());
		    warn "defaulting stop date to $stop_tz";
		}
		if ($stop_doy && !$start_doy) {   # stop given, but no start
		    $start_doy = 1;
		}
		if ($start_doy && !$stop_doy) {   # start given, but no stop
		    $stop_doy = 366;
		}
	    }

            
	    # Add info:
	    $srch_odl->add_param(START_DATE => IU::ODL->bare($start_tz))  
	    if $start_tz;
	    $srch_odl->add_param(STOP_DATE  =>  IU::ODL->bare($stop_tz))   
	    if $stop_tz;
	    $srch_odl->add_param(START_DAY_OF_YEAR => IU::ODL->bare($start_doy))
	    if $start_doy;
	    $srch_odl->add_param(STOP_DAY_OF_YEAR  =>  IU::ODL->bare($stop_doy))
	    if $stop_doy;
            
	}

	# Extended...
	elsif (IU::Category->is_extended($category)&& not $dir && not $browse){
	    my $crit;
	    foreach $crit ($self->criteria($category)) {
		$extsrch_odl ||= IU::ODL->new_group;    # create if needed
		$extsrch_odl->add_param(SPECIALIZED_CRITERIA => 
					$crit->to_request_ODL);
	    }
        }
	# Hunh?
	else {
	    warn "Unknown category '$category': ignored";
	}
    }

    # We skipped writing the core valids to the ODL, now write the 
    # datasets corresponding to the query to the ODL. 
    if ($inv) {
       $self->determine_datasets_to_search;
       if ($self->datasets_to_search) {
          $srch_odl->add_param(DATASET_ID =>
                             IU::ODL->conv($self->{DatasetsToSearch}));
       }
    }

    # Extended odl when coming from local granule id search page
    if($self->search_type eq 'LOCAL_GRANULE_SEARCH') {
        my $crit;


        foreach $crit($self->criteria("GRANULE DETAILS")) {
            $extsrch_odl = IU::ODL->new_group; 
            my $crit_odl = $crit->to_request_ODL;
            $extsrch_odl->add_param(SPECIALIZED_CRITERIA =>$crit_odl);        
       }
    }
    # The options...

    # For directory searches and local granule id searches
    # we don't need GRANULE_LIMIT!
    if (($self->ODL_search_type eq 'INVENTORY_SEARCH') && (not $dir)) {
       $srch_odl->add_param(GRANULE_LIMIT => 
          IU::ODL->number($self->max_granules));
    } 
    if ($self->ODL_search_type eq 'ACCOUNT_STATUS_REQUEST') {
        my $user = $SESSION->state->{User};
        if ($user) {
            $srch_odl->add_param(CONTACT_ADDRESS  => $user->req_addr->to_ODL)
                if $user->req_addr;
        }
    }
    # Max guides should be saved for all search types except directory.
    if ( not $dir){
        $srch_odl->add_param(GUIDE_LIMIT =>
		             IU::ODL->number($self->max_guides));
    }
    # Max time should be saved for all search types except directory.
    if ( not $dir){
        $srch_odl->add_param(TIME_LIMIT =>
		             IU::ODL->number($self->max_time));
    }

    # Include Non-Temporal granules 
    if ($self->include_nontime &&  not $dir) {
        $srch_odl->add_param(INCLUDE_NON_TEMPORAL => IU::ODL->quoted('Y'));
    }

    # Include Non-Spatial granules 
    if ($self->include_nongeo && not $dir) {
	$srch_odl->add_param(INCLUDE_NON_SPATIAL => IU::ODL->quoted('Y'));

    }

    # The "browse-only":
    if ($self->browse_only && not $dir) {
	$srch_odl->add_param(BROWSE_ONLY => IU::ODL->bare('Y'));
    }

    if ($self->all_metadata && not $dir) {
	$srch_odl->add_param(ALL_METADATA => IU::ODL->quoted($self->all_metadata));
    }

    if ($self->day_night) {
        $srch_odl->add_param(DAY_NIGHT => IU::ODL->quoted($self->day_night));
    }

    # Any extended search stuff:
    if ($extsrch_odl && not $browse) {
	$srch_odl->add_param(EXTENDED_SEARCH => $extsrch_odl);
    }

    # Completely non-standard stuff:
    if (not $for_search) {

       $srch_odl->add_param(SEARCH_NAME => IU::ODL->quoted($self->name))
          if ($self->name);
       $srch_odl->add_param(WG_MISSIONS =>IU::ODL->conv($self->{'WG_MISSIONS'}))
          if (int($self->missions));
       $srch_odl->add_param(WG_SRCHTYPE => IU::ODL->quoted($self->search_type))
          if ($self->search_type);
       $srch_odl->add_param(WG_OPTSTYPE => IU::ODL->quoted($self->optstype))
          if ($self->optstype);
       $srch_odl->add_param(WG_GEOTYPE => IU::ODL->quoted($self->geotype))
          if ($self->geotype);
       $srch_odl->add_param(WG_TIMETYPE => IU::ODL->quoted($self->timetype))
          if ($self->timetype);
       $srch_odl->add_param(WG_DEPVALIDTYPE =>
          IU::ODL->quoted($self->depvalidtype)) if ($self->depvalidtype);
    }

    # Done!
    $root_odl;
}

#------------------------------------------------------------
# to_search_ODL
#------------------------------------------------------------
# Convert query to an ODL object suitable for input to the client, 
# and return it. The ODL object is a group containing an INVENTORY_SEARCH 
# parameter.
#
# Unlike a simple to_ODL, this method uses the dependent valids mechanism
# to determine which DAACs to search

sub to_search_ODL {
    my $self = shift;
    # Start by figuring out which DAACS to search, if needed:
    $self->daacs_to_search
	or $self->determine_daacs_to_search
	    or puke "no data centers to search!";  

    # Start with the basics, *without* the DAACs:
    my $root_odl = $self->to_ODL(NoDaacs=>1, ForSearch=>1) 
	|| return undef;

    # Get the actual search group, checking all standard types:
    my $srch_odl = $root_odl->param('DIRECTORY_SEARCH') ||
	           $root_odl->param('INVENTORY_SEARCH') ||
		   $root_odl->param('GUIDE_SEARCH') ||
		   $root_odl->param('PRODUCT_STATUS_REQUEST') ||
		   $root_odl->param('ACCOUNT_STATUS_REQUEST') ||
		   puke "No recognized search group in ODL!"; 

    # Add the data centers (notice that we use the QUERY daacs!):
    if ($self->daacs_to_search) {
	$srch_odl->add_param(DATA_CENTER_ID => 
  			     IU::ODL->conv($self->{DaacsToSearch}));
    }

    # Add the authenticator, if any:
    my $authenticator = eval { $SESSION->state->{User}->authenticator };
    if ($authenticator) {
	$srch_odl->add_param(AUTHENTICATOR => IU::ODL->conv($authenticator));
    }

    # Add the NASDA authenticator, if any:
    my $nasda_authenticator = 
     eval {$SESSION->state->{User}->nasda_authenticator};
    if ($nasda_authenticator) {
	$srch_odl->add_param(NASDA_AUTHENTICATOR => 
                             IU::ODL->conv($nasda_authenticator));
    }

    # Check for ECS fields, if we are using the ecs authenticator:
    # Modified to check for "ECS Profile SMC" instead of home DAAC.
    unless ($IU::EcsProfileSmc eq '') {
       # Add the ecs authenticator, if any:
       my $ecs_authenticator = eval { $SESSION->state->{User}->ecs_authenticator };
       if ($ecs_authenticator) {
           $srch_odl->add_param(ECS_AUTHENTICATOR => IU::ODL->conv($ecs_authenticator));
       }
    }


    # Done!
    return $root_odl;
}

#------------------------------------------------------------
# print_HTML HTMLSTREAM
#------------------------------------------------------------
# Output summary as pretty HTML text.

sub print_HTML {
    my $self = shift;
    my $HH = shift;    # HTML stream
    my $nogran = shift;

    $HH -> UL;
    my ($cat, $val, $crit);
    foreach $cat (sort {category2pretty($a) cmp category2pretty($b) 
			} $self->categories) {
	# Skip this category if not used by current search type:
	$self->uses($cat) or next;

        #Skip time range, geo region or extended category
        #if this is a directory search
        #Can't base it on uses because need that for query.dat to be correct
        if (($self->search_type eq 'DIRECTORY_SEARCH')&&
            ($cat eq 'GEOGRAPHIC_REGION' || $cat eq 'TIME_RANGE'
             || (IU::Category->is_extended($cat)))){next;} 
	# Get criteria:
	my @criteria = $self->criteria($cat); @criteria or next;
	my $pretty_cat = category2pretty($cat);

	$HH -> LI -> t_B($pretty_cat, ": ");
	if (IU::Category->is_singleton($cat)) {   # has one criterion/variant
	    $crit = $criteria[0];
	    $HH -> t($crit->as_string);
	}
        else {                                    # has multiple criteria
	    $HH -> UL;
	    foreach $crit (@criteria) {
	        $HH -> LI -> t_B($crit->name, ": ") -> t($crit->as_string);
	     }
	    $HH -> _UL;
        }
        
    }

    # Print search options if set:

  
    my $inv = ($self->search_type eq 'INVENTORY_SEARCH') ? '1' : '0';
    my $gran = ($self->search_type eq 'GRANULE_SEARCH') ? '1' : '0';
    my $browse = ($self->search_type eq 'BROWSE_SEARCH') ? '1' : '0';
 
    if (!$nogran && ($inv || $gran || $browse)) {
       $HH -> LI ->t('Return a maximum of '.$self->max_granules.
                     ' data granules per data set.');
    }

    if ($inv && $self->include_nontime) {
       $HH -> LI ->t('Include non-temporal data granules.');
    }

    if ($inv && $self->include_nongeo) {
       $HH -> LI ->t('Include non-spatial data granules.');
    }

    if ($inv && $self->browse_only) {
       $HH -> LI ->t('Only return data granules which have browse products.');
    }

    if ($inv && $self->all_metadata) {
       $HH -> LI ->t('Return '.$self->all_metadata.
                     ' metadata in search results.');
    }

    if ($inv && $self->day_night) {
       my $txt = ($self->day_night eq 'D' ) ? 'day only' : 'night only'; 
       $HH -> LI ->t('Only return data granules which were retrieved '.
                     'during the '.$txt.'.');
    }

    $HH ->_UL;
    1;
}

#------------------------------------------------------------
# set_local_granule_id SELF, TEXT
#------------------------------------------------------------
# When search type is local granule id, need to set in query
# set using input text
sub set_local_granule_id {
    my ($self, $text) = @_;
    

        my @criteria;
        my $cats = "GRANULE DETAILS";


            # Get the criterion objects:
            
            @criteria = $self->depvalid_interface->get_criteria($cats);
            my @selected;

            my $crit;
            foreach $crit (@criteria) {

	        # Read input for this criterion:
	        
                if ($crit->name eq "LocalGranuleID") {
                    
                # remove any end of line chars.
                    chomp($text);

                    if (IW::Util::blank($text)) {     # KLUDGE
                        $text = "None selected"; 
                    }elsif($crit->{UserSlct}{VALUE} ne $text) {
                        
                        
                        $self->criteria('GRANULE DETAILS',undef);

                    } 
                    (length($text) <= $crit->max_len)
	                or $crit->whine(513, "exceeds maximum length of ", $crit->max_len);
    
                    $crit->{UserSlct} = {VALUE=>$text};    
	            push @selected, $crit;
                    $self->local_granule_id(\@selected);
                }
            }

            # Store the new criteria in the query
            

            $self->criteria('GRANULE DETAILS', \@selected);

            my $cat;
            my @variants;
            my $total;
            my @xs_cats = grep {IU::Category->is_extended($_)} $self->categories;
            foreach $cat (@xs_cats) {

                @variants = map {$_->variant} $self->criteria($cat);
	        $total = $total + int(@variants) if(@variants);
	    }
            puke 517, "Your total number of" 
             . " attributes chosen now exceeds the system limit of ${IU::PSAMax}."
             . " Please use the back button of your browser to return to the"
             . " previous page and reduce the number of attributes chosen."
             if($total > ${IU::PSAMax});

    

      
 
}

#------------------------------------------------------------
# set_node_type SELF, TEXT
#------------------------------------------------------------
# When search type is local granule id, need to set in query
# set using input text
sub set_node_type {
    my ($self, $text) = @_;

        my @criteria;
        my @values; 
        my $valueref;
        my $cats = "ORBIT";

            # Get the criterion objects:
            
            @criteria = $self->depvalid_interface->get_criteria($cats);
            my @selected;
            my $critname;
            my $crit;
            foreach $crit (@criteria) {
                             
                $critname = $crit->name;
	        # Read input for this criterion:
                if ($crit->name eq 'NodeType') {
                # remove any end of line chars.
                    chomp($text);

                    if ($text eq 'All') {    
                        $self->criteria('ORBIT',undef);
                        $self->node_type(undef);
                        return 1;
                    }elsif($crit->{UserSlct}{VALUE} ne $text) {
                        $self->criteria('ORBIT',undef);
                    }
                    if($text eq 'NotDescending') {
                       @values = ('Ascending','NorthPole','SouthPole');
                       $valueref = \@values;
                       $crit->{UserSlct} = {VALUES=>$valueref};
                       push @selected, $crit;
                    }elsif($text eq 'NotAscending') {
                       @values = ('Descending','NorthPole','SouthPole');
                       $valueref = \@values;
                       $crit->{UserSlct} = {VALUES=>$valueref};
                       push @selected, $crit;
                    } else { 
                       $values[0] = $text;
                       $valueref = \@values;
                       $crit->{UserSlct} = {VALUES=>$valueref};    
                       push @selected, $crit;
                    }
                    $self->node_type(\@selected);
                }
            }

            # Store the new criteria in the query
            $self->criteria('ORBIT', \@selected);

            my $cat;
            my @variants;
            my $total;
            my @xs_cats = grep {IU::Category->is_extended($_)} $self->categories;
            foreach $cat (@xs_cats) {
                @variants = map {$_->variant} $self->criteria($cat);
	        $total = $total + int(@variants) if(@variants);
	    }

            puke 517, "Your total number of" 
             . " attributes chosen now exceeds the system limit of ${IU::PSAMax}."
             . " Please use the back button of your browser to return to the"
             . " previous page and reduce the number of attributes chosen."
             if($total > ${IU::PSAMax});
}

#----------------------------------------------------
# get_local_granule_id
#----------------------------------------------------
sub get_local_granule_id {
    my $self = shift;
    my $value;
    # Get existing criteria, if any:


   my @criteria = $self->criteria('GRANULE DETAILS');


    @criteria or return;

    my $crit;

    foreach $crit (@criteria) {

        $value =  $crit->{UserSlct}{VALUE};
        
        
        if ($value eq "None selected") {
             $value = '                       ';
        }
    }
    $value;
}

#------------------------------------------------------------
# set_dummy_query
#------------------------------------------------------------
# Based on the search type save or remove various categories
# 1. For 'BROWSE_SEARCH' and 'INVENTORY_SEARCH' save and restore time
#    information in the query.
# 2. For 'INVENTORY_SEARCH' save and restore everything but extended criteria.
# 3. For 'GUIDE_SEARCH', and 'DIRECTORY_SEARCH' save and 
#    restore: datasets, datacenters, sources, sensors, and campaigns.
# 4. For 'GRANULE_SEARCH', and 'LOCAL_GRANULE_SEARCH' save and restore datasets.
# 5. For 'BROWSE_SEARCH', 'GRANULE_SEARCH', and 'LOCAL_GRANULE_SEARCH' reset geo
#    to global search.
# 6. For 'BROWSE_SEARCH' set the 'ParentGranCollectionName' if it is in
#    the valids and set $SESSION->state->{Found} = 1.
sub set_dummy_query {
   my ($self, %params) = @_;
   my $newSrchType = $params{NewSrchType};
   my $prevSrchType = $params{PrevSrchType} || $self->search_type;
   my $geo;
   my $local;
   my @datasets;
   my @datacenters;
   my @parameters;
   my @sources;
   my @sensors;
   my @campaigns;
   my @processing_levels;
   my @granule_ids;
   my ($start_tz, $stop_tz, $start_doy, $stop_doy);

   # Set depvalid widget to be subtype "Valids" for all search types
   # except when search type is INVENTORY and the user has a previously
   # selected value:
   $self->depvalidtype('Valids') unless
      ( ($self->depvalidtype) && ($newSrchType eq 'INVENTORY_SEARCH') );

   if ($newSrchType eq 'BROWSE_SEARCH' || $newSrchType eq 'INVENTORY_SEARCH') {
      # Get the original temporal information
      ($start_tz, $stop_tz, $start_doy, $stop_doy) = $self->time_range_TZ;
   }
   if ($newSrchType eq 'INVENTORY_SEARCH') {
      $geo = $self->geo_region || IU::Geo::Range->new(90.0,-180.0,-90.0,180.0);
      @processing_levels = $self->processing_levels;
   }
   if ($newSrchType eq 'INVENTORY_SEARCH' || $newSrchType eq 'GUIDE_SEARCH' ||
       $newSrchType eq 'GRANULE_SEARCH' || $newSrchType eq 'DIRECTORY_SEARCH' ||
       $newSrchType eq 'LOCAL_GRANULE_SEARCH' || $newSrchType eq 'BROWSE_SEARCH'
) {
      @datasets = $self->dataset_ids;
   }
   if ($newSrchType eq 'INVENTORY_SEARCH' || $newSrchType eq 'GUIDE_SEARCH' ||
       $newSrchType eq 'DIRECTORY_SEARCH') {
      @datacenters = $self->data_center_ids;
      @parameters = $self->parameters;
      @campaigns = $self->campaigns;
      @sensors = $self->sensor_names;
      @sources = $self->source_names;   
   }
   if ($newSrchType eq 'LOCAL_GRANULE_SEARCH') {
      $local = $self->get_local_granule_id;
   }
   if ($newSrchType eq 'GRANULE_SEARCH') {
      @granule_ids = $self->granule_ids;
   }

   # Clear the category information
   $self->{Categories}    = {};

   # Set search type:

   $self->search_type($newSrchType);

   if ($newSrchType eq 'BROWSE_SEARCH' || $newSrchType eq 'INVENTORY_SEARCH') {
      # Restore the time information
      $self->time_range_TZ($start_tz, $stop_tz, $start_doy, $stop_doy);
   }
   if ($newSrchType eq 'INVENTORY_SEARCH') {
      # Restore the geo region
      $self->geo_region($geo);
      $self->processing_levels(\@processing_levels);
   }
   if ($newSrchType eq 'LOCAL_GRANULE_SEARCH') {
      $self->set_local_granule_id($local);
   }
   if ($newSrchType eq 'GRANULE_SEARCH') {
      $self->granule_ids(\@granule_ids);
   }
   if ($newSrchType eq 'INVENTORY_SEARCH' || $newSrchType eq 'GUIDE_SEARCH' ||
       $newSrchType eq 'GRANULE_SEARCH' || $newSrchType eq 'DIRECTORY_SEARCH' ||
       $newSrchType eq 'LOCAL_GRANULE_SEARCH') {
      # Restore the datasets
      $self->dataset_ids(\@datasets);
   }
   if($newSrchType eq 'GRANULE_SEARCH') {
      # Use wwwvalids.conf to determine which DAACS allow search by granule id
      $WWWVALIDS = IU::WWWValids->new if (!defined($WWWVALIDS));
      my $daacarrayref = $WWWVALIDS->daac_array();
      # Set in the query the data centers that allow search by granule id
      $self->data_center_ids($daacarrayref);
   }
   if ($newSrchType eq 'INVENTORY_SEARCH' || $newSrchType eq 'GUIDE_SEARCH' ||
       $newSrchType eq 'DIRECTORY_SEARCH') {
      # If the previous search type was 'GRANULE_SEARCH', clear the
      # data_center_ids from the query.
      $self->data_center_ids(\@datacenters)
         unless ($prevSrchType eq 'GRANULE_SEARCH');
      $self->campaigns(\@campaigns);
      $self->parameters(\@parameters);
      $self->sensor_names(\@sensors);
      $self->source_names(\@sources);
   }
   if ($newSrchType eq 'BROWSE_SEARCH' || $newSrchType eq 'GRANULE_SEARCH' ||
       $newSrchType eq 'LOCAL_GRANULE_SEARCH') {
      # Set the geo to Global
      $geo = IU::Geo::Range->new(90.0,-180.0,-90.0,180.0);
      $geo->fixup->validate;
      $self->geo_region($geo);
      $SESSION->state->{LastMapname} = 'I want to do a GLOBAL SEARCH';
   }
   if ($newSrchType eq 'BROWSE_SEARCH') {
      # Make sure that the valid exists
      # If the 'ParentGranCollectionName' valid is not compatible
      # there will be no data sets available for Browse Search.
      my @valids = 
         $self->depvalid_interface->get_valids("GRANULE DETAILS");
      my $val;
      my $found = 0;
      foreach $val (@valids) {
         $found = 1 if (lc($val) eq lc('ParentGranCollectionName')); 
      }
      $SESSION->state->{Found} = $found;
      if ($found) {
         # Get the criterion objects:
         my @criteria =
            $self->depvalid_interface->get_criteria("GRANULE DETAILS");
         my (@selected,  $crit);
         foreach $crit (@criteria) {

            # Read input for this criterion:
       
            if ($crit->name eq "ParentGranCollectionName") {
                # Store the new criteria in the query
                my @values =  $crit->valids;
                $crit->{UserSlct} = {VALUES=>\@values};

                push @selected, $crit;
                $self->criteria('GRANULE DETAILS', \@selected);
            }
         }
      }
      $self->determine_browse_datasets_to_search(\@datasets);
   }
   return $self;
}
#------------------------------------------------------------    
1;
