package IU::RequestResult;

=head1 NAME

IU::RequestResult - represents the info in an ODL REQUEST_RESULT group
which can be returned as part of PRODUCT_RESULT message. 

=head1 DESCRIPTION

This class models a new entity that is being put in the ODL
to support subsetting. The REQUEST_RESULT allows a status to be returned
for each line_item in the order instead of just one status for the entire
order.  The corresponding ODL looks like ...

       GROUP = REQUEST_RESULT 
          STATUS_CODE
          [STATUS_CODE_COMMENT]
          (MISC_URL)*    see IU:FancyURL 
          DAAC_REQUEST_ID
          LINE_ITEM group
       END_GROUP = REQUEST_RESULT 



=head1 SYNOPSIS

    # create it from an RCF822 chunck.
    my $requestresult = IU::RequestResult->from_chunk($requestresultchunk);
    
    # Output it as text: 
    $requestresult->as_text;
 

=head1 AUTHOR

VirtualEd, GST.

=head1 RCS INFO

=cut

#use IU;
use IU::Chunk;
use IU::FancyURL;
#use IU::Log;


#------------------------------------------------------------
# new
#------------------------------------------------------------
sub new {
    my $class = shift;
    my %params = @_;
    bless \%params, $class;
}

#------------------------------------------------------------
# from_chunk CHUNK, [TYPE]
#------------------------------------------------------------
# I<Class or instance method.>
# Create a new instance from a chunk.

sub from_chunk {

    my ($self, $chunk) = @_;
    ref($self) or $self = $self->new;
    $self->{StatusCode}   = $chunk->value('STATUS_CODE');
    $self->{StatusCodeComment} = $chunk->value('STATUS_CODE_COMMENT');
    $self->{DaacRequestID}    = $chunk->value('DAAC_REQUEST_ID');

    # URLs The ODL spec says just MISC_URL, but I've also added DATA_URL
    # just in case, can have more than one of each type):
    my $url_type;
    foreach $url_type (qw(DATA_URL MISC_URL)) {
        $self->{URLs}{$url_type} ||= [];
        my $url_chunk;
        foreach $url_chunk ($chunk->values($url_type)) {
            push @{$self->{URLs}{$url_type}},
               IU::FancyURL->from_chunk($url_chunk, $url_type);
        }
    }

    # Only interested in the PACKAGE_ID so ignore the rest of the line item.
    my $lineitem_chunk = $chunk->value('LINE_ITEM');
    $self->{PackageID} = $lineitem_chunk->value('PACKAGE_ID');

return $self;

}

#------------------------------------------------------------
# URLs [TYPE]
#------------------------------------------------------------
# Return the array of URLs of the given TYPE (as IU::FancyURLs),
# or empty if none.
#
# If TYPE is not given, returns an array of available types.

sub URLs {
    my ($self, $type) = @_;
    @_ > 1 or return (keys %{$self->{URLs}});
    @{$self->{URLs}{$type} || []};
}

#------------------------------------------------------------
# status_code
#------------------------------------------------------------
# Return the status code

sub status_code          { $_[0]->{StatusCode} }

#------------------------------------------------------------
# package_ID 
#------------------------------------------------------------
# Return the Package ID for the line item 

sub package_ID       { $_[0]->{PackageID} }

#------------------------------------------------------------
# status_code_comment
#------------------------------------------------------------
# Return the status code comment, or the empty string if there is none.

sub status_code_comment  { $_[0]->{StatusCodeComment} || '' }

#------------------------------------------------------------
# daac_request_ID 
#------------------------------------------------------------
# Return the daac request ID 

sub daac_request_ID          { $_[0]->{DaacRequestID} }


#------------------------------------------------------------
# as_text
#------------------------------------------------------------

sub as_text {
1;
}

#------------------------------------------------------------
1;

