package IU::ResultGroup;

=head1 NAME

IU::ResultGroup - handle the RESULT_GROUP group

=head1 DESCRIPTION

This class models a new entity that is being put in the ODL
extensions.  The corresponding ODL looks like:

       GROUP = RESULT_GROUP
          RESULT_GROUP_NAME = "Name"
          GROUP = SPECIALIZED_RESULTS
          END_GROUP = SPECIALIZED_RESULTS
          GROUP = RESULT_GROUP
          END_GROUP = RESULT_GROUP
       END_GROUP = RESULT_GROUP


=head1 SYNOPSIS

    $result = new IU::ResultGroup resultGroupName =>"SUN_AZIMUTH_ANGLE";
    
    # The same, using HTML streams (much safer):
    $HTML -> t($result->resultGroupName) -> t($result->resultValue) -> _A;

=head DATA STRUCTURE

I<This explanation is derived from the code. It may not be correct.>

The data structure created is a tree structure whose nodes contain two lists:
one list of specialized results, and a second list of result groups. This 
allows us to nest result groups as deeply as we care to.

Note that we don't care about multiple independent result groups - that has to
be handled by the caller.

In addition, we have an "index" that allows us to grab a specialized result
directly, without having to crawl through the tree. To avoid possible 
problems with duplicate names, we do this as a bi-level structure: a hash
contains all of the names. The value associated with each name is a reference
to an anonymous array of specialized results with this name. This may have to 
be made more sophisticated.

=head1 AUTHOR

Mark Solomon, Raytheon ITSS

=head1 RCS INFO

$Id: ResultGroup.pm,v 1.9 2001/03/22 21:22:12 seufert Exp $

=cut

use IU;
use IU::Chunk;
use IU::ODL ();
use IU::Log;

#------------------------------------------------------------
# new
#------------------------------------------------------------
sub new {
    my $class = shift;
    $self = {@_};
    $self->{Index} = {};
    bless $self, $class;
}

#------------------------------------------------------------
# resultGroupName
#------------------------------------------------------------
# Return the resultGroupName, or undef if there is none.

sub resultGroupName {
    $_[0]->{ResultGroupName};
}


#------------------------------------------------------------
# specializedResults
#------------------------------------------------------------
# Return a reference to the array of specialized results,
# or []
sub specializedResults { $_[0]->{SpecializedResults} || []; }

#------------------------------------------------------------
# resultGroup
#------------------------------------------------------------
# Return a reference to the array of result groups,
# or []
sub resultGroup { $_[0]->{ResultGroup} || []; }


#------------------------------------------------------------
# from_chunk CHUNK
#------------------------------------------------------------
# I<Class or instance method.>
# Create a new instance from a chunk.

sub from_chunk {
    my ($self, $chunk, $index) = @_;
    ref($self) or $self = $self->new;
    my $top = !defined $index;
    $index = defined $index ? $index : $self -> {Index};
    

    $self->{ResultGroupName}  = $chunk->value('RESULT_GROUP_NAME');
    my $results_chunk;
    my %search_results;

    # Process specialized results on this level. They get added to this
    # level and to the global index.
    foreach $results_chunk ($chunk->values('SPECIALIZED_RESULTS')) {
        my $result = IU::SpecializedResults->from_chunk($results_chunk);
        if (defined $index->{$result->resultName}) {
           # already there, add to existing list
           push @{$index->{$result->resultName}}, $result;
        }
        else {
           $index->{$result->resultName} = [$result];
        }
        # This code is outmoded and will be removed as soon as I
        # scan the rest of the system for code that uses the Check
        # field or field access methods.
        my $search_for;
        my @tmp_var;
        if(defined($self->{Check})) {
           @tmp_var = @{$self->{Check}}; 
        } else {
           @tmp_var = [];
        }
        if (int(@tmp_var) > 0) {
           foreach  $search_for (@tmp_var) {
             my $newvalue = $result->check($search_for);
             $search_results{$search_for} = $newvalue if($newvalue);
           }
        }

        # Shove the specialized result in at this level.
        push @{$self->{SpecializedResults}}, $result;
    }

    # Add all of the result groups under this one to this object.
    # We do a recursive call and let the lower levels sort themselves out.
    my $result_group_chunk;
    foreach $result_group_chunk ($chunk->values('RESULT_GROUP')) {
        my $result_group = 
            IU::ResultGroup->from_chunk($result_group_chunk,$index);
        push @{$self->{ResultGroup}}, $result_group;
    }

    # This saves the old-style index (I think). Again, this is code to be
    # removed after the check for use of the Check field/access methods.
    $self->{CheckValue} = \%search_results;
    $self->{Index} = $index if $top;
    $self;
}
# sub checkName Return the checkName, or undef if there is none.
sub checkName {
   $_[0]->{Check};
}
# sub checkValue Return the checkValue, or undef if there is none.
sub checkValue {
   my $self = shift;
   my $looking_for = shift;
   return ($self->{CheckValue}{$looking_for});
}

# Now handles multiple items to search for.
# Pass it an Array Ref 
sub setCheck {
   my $self = shift;
   $self->{Check} = shift;
}

sub getCheck {
   my $self = shift;
   my $looking_for = shift;
   return ($self->{CheckValue}{$looking_for});
}

#---------------------------------------------------------
# findSpecializedResults PATTERN
#---------------------------------------------------------
# I<Instance method>
# Find and return all specialized results that match the pattern.
# Note that pattern match is case-insensitive.

sub findSpecializedResults {
   my $self = shift;
   my $predicate = shift || ".*";
   my $modifier = shift || "";
   my @results = ();
   foreach my $k (keys %{$self->{Index}}) {
      # See if this key meets the requirements.
      if ($k =~ /$predicate/i) {
         push @results, @{$self->{Index}->{$k}};
      }
      else {
      #  debug "Miss for $predicate: $k";
      }
   }
   \@results;
}

#------------------------------------------------------------
1;
