package IU::Session;

=head1 NAME

IU::Session - a session directory in the IMS/WWW interface

=head1 INSTANCE VARIABLES (PRIVATE)

=over

=item Dir

The session directory.

=item Id

The session id.

=item Parent

The session parent directory.

=item ParentURL

The session parent directory URL, I<with> a trailing slash.
If explicitly set to false on construction, then the C<url()> 
method will raise a fatal exception.

=item State

The session state database.  Since this is opened for read/write
access and not closed until the session object is destroyed, you
can only have one session object attached to the state at a time.

=back

=head1 NOTES

When created, the session directory is given a C<.htaccess> file that
prevents any files under it from being downloaded.  Then, a subdirectory
called C<"pub"> is created with its own C<.htaccess> file that allows
all files under it to be downloaded.  Currently, this C<"pub"> directory 
allows access to all; however, it should limit access to the requester's
known host (taken from the REMOTE_HOST environment variable) in the future.

This module is in IU, since it was suspected that it might be useful
to give utility programs access to session directories from a
high-level perspective.

=head1 AUTHOR

Eryq / Hughes STX Corporation

=head1 VERSION

$Id: Session.pm,v 1.43 2002/08/08 18:25:35 joe Exp $

=cut

use IU;
use IU::Log;

use IU::Store ();
use IU::ODL;
use IU::Util;
use URI::Escape;

use strict;

#------------------------------------------------------------
#
# PRIVATE UTILITIES
#
#------------------------------------------------------------

#------------------------------------------------------------
# good_id ID
#------------------------------------------------------------
# Is this a permissible session id?

sub good_id {
    # The "%" is for the uri escaped characters.
    $_[0] =~ /^[a-z0-9-_\%]+$/i;
}

#------------------------------------------------------------
#
# PRIVATE METHODS
#
#------------------------------------------------------------

#------------------------------------------------------------
# setup
#------------------------------------------------------------
# Common setup.
# Currently a no-op.

sub setup {
    my $self = shift;
}

#------------------------------------------------------------
# setup_files
#------------------------------------------------------------
# We've just created a new session directory.  Fill it with joy and files.

sub setup_files {
    my $self = shift;
    
    # Looks like we'll need to do .htaccess stuff!
    require IU::HTAccess;
    my $access;

    # THE HTACCESS FILE FOR THE TOP DIRECTORY (DENIES EVERYONE)...

    # Build a virtual .htaccess file for this session directory:
    $access = new IU::HTAccess;
    $access->start_limit('GET', 'POST', 'PUT');
    $access->order('ALLOW', 'DENY');
    $access->deny_from('all');
    $access->end_limit;

    # Write it out:
    $access->write_in($self->{Dir}) 
	or return error "couldn't write .htaccess";

    # THE HTACCESS FILE FOR THE "PUB" DIRECTORY (ALLOWS EVERYONE)...

    # Build a virtual .htaccess file for this session directory:
    $access = new IU::HTAccess;
    $access->start_limit('GET', 'POST', 'PUT');
    $access->order('ALLOW', 'DENY');
    $access->allow_from('all');
    $access->end_limit;

    # Write it out:
    $access->write_in("$self->{Dir}/pub") 
	or return error "couldn't write pub/.htaccess";
    1;
}

#------------------------------------------------------------
# write_state
#------------------------------------------------------------
# Write out the state database in the newly-created session.
# This also ends the session link to the state - you have
# to redo a setup_state.

sub write_state {
   my $self = shift;
   untie %{$self->state};
   1;
}

#------------------------------------------------------------
#
# CONSTRUCTING AND INITIALLIZING
#
#------------------------------------------------------------

#------------------------------------------------------------
# DESTROY
#------------------------------------------------------------
# I<Destructor:> writes the database file out to disk.

sub DESTROY {  
    my $self = shift;
    $self->write_state;
}

#------------------------------------------------------------
# new PARAMHASH
#------------------------------------------------------------
# I<Private internal constructor.>  Create a new object, setting 
# up the Parent and ParentURL attributes.
#
# The PARAMHASH contains the same information given to the create() 
# and resume() constructors.

sub new {
    my $class = shift;
    my %params = @_;
    
    # Check usage:
    (exists($params{Parent}) == exists($params{ParentURL}))
	or puke 600, "need to give BOTH Parent and ParentURL, or give neither";

    # Create the object:
    my $self = {};
    bless $self, $class;

    # Setup the parent info, using defaults if necessary:
    $self->{Parent} = $params{Parent} || "${IU::DocsDir}/imswww-usr";
    $self->{ParentURL} = $params{ParentURL} || "${IU::DocsUrl}imswww-usr/";

    # debug "ParentURL = ".$params{ParentURL}."  ".$self->{ParentURL};
    # Return the object:
    $self;
}

#------------------------------------------------------------
# create PARAMHASH
#------------------------------------------------------------
# I<Constructor.>  Create a new session directory, and return 
# an object representing an existing session.  The PARAMHASH
# contains...
#
#      Parent    => parent directory (defaulted to standard location)
#      ParentURL => parent directory URL (defaulted to standard location;
#                   you MUST supply this if supplying the Parent, though
#                   you can supply a false value)
#      Id        => session id (generated automatically if not given)
#      Return    => If 1 means instead of puking here, our parent wants to
#		    handle it themselves.
#
# It is an error if the session directory already exists.
#
# B<NOTE:> all errors in this constructor are handled via a thrown exception.

sub create {
    my $class = shift;
    my %params = @_;

    # Create the object, setting up the parent information:
    my $self = $class->new(@_);

    # Check the parent directory:
    (-d $self->{Parent} && -w $self->{Parent}) 
	or puke "missing/unwritable session parent: '$self->{Parent}'";

    # Get the id, defaulting to time + pid:
    $self->{Id} = uri_escape($params{Id}) || (time . "-$$");
    good_id($self->{Id}) or puke "bad session id: '$self->{Id}'";
    
    # Create the session directory tree:
    $self->{Dir} = "$self->{Parent}/$self->{Id}";
    (! (-d $self->{Dir})) 
	or puke 535, "That User Account already exists on this server!!!";

    mkdir $self->{Dir}, 0755 
	or puke 455, "couldn't create session dir $self->{Dir}: $!";
    mkdir "$self->{Dir}/pub", 0755 
	or puke 455, "couldn't create session dir $self->{Dir}/pub: $!";
    
    # Populate it full of goodies:
    $self->setup_files;

    # Common setup:
    $self->setup;

    # Done!
    $self;
}

#------------------------------------------------------------
# resume PARAMHASH
#------------------------------------------------------------
# I<Constructor.>  Create an object representing an existing
# session.  The PARAMHASH should contain...
#
#      Parent    => parent directory (see create() method for details)
#      ParentURL => parent directory URL (see create() method for details)
#      Id        => session id (mandatory)
#
# It is an error if the session directory does not exist.
#
# B<NOTE:> all errors in this constructor are handled via a thrown exception.

sub resume {
    my $class = shift;
    my %params = @_;

    my $self;
    # Create the object, setting up the parent information:
    trace_to_log("remote user = ".$ENV{REMOTE_USER});

    my $persistent_dir = ${IU::PersistentDir}."imswww-usr/";
    my $persistent_url = ${IU::DocsUrl}."persistent/";

    my $admin = $params{Admin};
    my $the_type = $params{Type} || 'STANDARD';
    if ( (!($params{Type})) && ($ENV{REMOTE_USER}) && 
         (!($ENV{REMOTE_USER} eq "guest")) ) {
	$the_type = 'PERSISTENT';
    }
    trace_to_log("the_type = $the_type");
    $the_type = 'STANDARD' if($admin);
	
    if ( $the_type eq 'PERSISTENT' ) {
#   if ( ($ENV{REMOTE_USER}) && (!($ENV{REMOTE_USER} eq "guest")) ) {

        #  debug "Persistent Constructor";
        my $encoded_ID = uri_escape($ENV{REMOTE_USER});
	(-d $persistent_dir."$encoded_ID") 
	  or puke 508, "User Account $ENV{REMOTE_USER} doesn't exist";

	$self = $class->new(Parent=>$persistent_dir,ParentURL=>$persistent_url);
	$params{Id} = $encoded_ID;
    } 
    else {
	# debug "Standard constructor";  
        $self = $class->new(@_);
    }

    # debug " My Parent = ".$self->{Parent};
    # debug " Id  = ".$params{Id};

    # Check the id:
    ($self->{Id} = $params{Id}) if ($params{Id}) ;
    good_id($self->{Id}) or puke "bad session id: '$self->{Id}'";

    # Check the directory:
     $self->{Dir} = "$self->{Parent}/$self->{Id}";
    (-d $self->{Dir}) or puke 521, "no such session directory, '$self->{Dir}'";

    # Common setup:
    $self->setup;

    # Done!
    $self;
}

#------------------------------------------------------------
# persistent_dir ACCOUNTNAME
#------------------------------------------------------------
# function
# Given an account name return the persistent directory

sub persistent_dir {
   my ($class,$name) = @_;
   "${IU::PersistentDir}/imswww-usr/".uri_escape($name);
}

sub trace_to_log {
   shift if $_[0] eq "IU::Session";
   print STDERR "($$) ",(join "",@_),"\n" if lc($IU::ForceDebug) eq 'yes';
}

#------------------------------------------------------------
#
# PUBLIC INTERFACE
#
#------------------------------------------------------------

#------------------------------------------------------------
# session_exists ACCOUNTNAME
#------------------------------------------------------------
# instance or function 
# Check for the existance of persistent session directory

sub session_exists {

   my $self = shift;
   my $name = shift;

   # check if directory exists. Could be a link, that's why I'm using 
   # -e instead of -d option.
   my $newdir = (-e IU::Session->persistent_dir($name)); 
   return $newdir;
}

#------------------------------------------------------------
# rename OLDNAME, NEWNAME
#------------------------------------------------------------
# instance or function 
# Rename (link to new name) persistent session directory

sub rename {
   my ($class,$oldname,$newname) = @_;
   my $newdir = IU::Session->persistent_dir($newname);
   my $olddir = IU::Session->persistent_dir($oldname);

   # Make sure new name isn't already being used.
   puke "Session directory $newdir already exists" if ( -e $newdir);

   if ( -e $olddir ) {
      symlink ($olddir,$newdir) 
         or puke "Failure in creating link $newdir to $olddir";
   } else {
      puke "Session directory $olddir requested for rename does not exist";
   }
   1;
}

#------------------------------------------------------------
# delete ACCOUNTNAME
#------------------------------------------------------------
# instance or function
# Physically delete the persistent session directory

sub delete {
   my ($class,$name) = @_;
   my $dir = IU::Session->persistent_dir($name);

   # Check if session directory is a sym link from a rename.
   # There may have been more than one rename, resulting in
   # multiple layers of links.
   my $realdir = $dir;
   while ( -l $dir ) {
      $realdir = readlink $dir;
      # Delete the link.
      unlink $dir;
      $dir = $realdir;
   }
   if ( -d $realdir ) {
      system("rm -rf " . IU::Util::shellquote($realdir) );
   } else {
      puke "Session directory $realdir requested for deletion does not exist";
   }
   1;
}

#------------------------------------------------------------
# dir
#------------------------------------------------------------
# Return the session directory.

sub dir {
    $_[0]->{Dir};
}

#------------------------------------------------------------
# id
#------------------------------------------------------------
# Return the session id.

sub id {
    $_[0]->{Id};
}

sub set_CGI_id {
   my $self = shift;
   my $new_value = shift;
   # debug " set_CGI_id = $new_value";
   $self->{Id} = $new_value;
}

#------------------------------------------------------------
# last_touched
#------------------------------------------------------------
# What time (as a time()-style number) was the session last touched?

sub last_touched {
    my $self = shift;
    my $kilroy = $self->dir . "/.kilroy";

    (stat($kilroy))[9];     # last-modified time
}

#------------------------------------------------------------
# path_to RELPATH
#------------------------------------------------------------
# Return the fully-qualified path to file RELPATH, which is a pathname
# relative to the top-level directory of the session.  The file need not 
# exist.  For example:
#
#       open IN, $session->path_to("in.dat");
#       open OUT, ('>' . $session->path_to("outputs/out.dat"));
#
# This method wasn't necessary, but it's sometimes more readable than saying:
#
#       open OUT, ('>' . $session->dir . "/outputs/out.dat");

sub path_to {
    my ($self, $relpath) = @_;
  
    # debug " self->{Dir} =".$self->{Dir};  
    # debug " relpath =".$relpath; 
    
    $relpath = "" unless defined $relpath;

    "$self->{Dir}/$relpath";
}

#------------------------------------------------------------
# setup_env
#------------------------------------------------------------
# Set up environment variables for this session, so that subprocesses
# know which session they're running in.  Intended more for CGI scripts,
# but what the hay: don't call it if you don't need it...
#
# B<WARNING:> this changes %ENV!

sub setup_env {
    my $self = shift;

    # For various IMS/WWW executables:
    $ENV{'IW_TOP_DIR'}  = $IU::TopDir;
    $ENV{'IW_SESSION_ID'} = $self->id;
    $ENV{'IW_SESSION_DIR'} = $self->dir;
    $ENV{'PATH'} .= ':' . $IU::MorePaths;
    
    # For the standalone client:
    $ENV{'GAEADATA_DIR'} = $IU::GaeaDataDir;
    $ENV{'GAEATMP_DIR'} = $self->dir;
    $ENV{'USRDATA_DIR'} = $self->dir; # so the client doesn't complain. sheesh.
    $ENV{'SYSLOG_LEVEL'} = $IU::SyslogLevel || '6';

    # For the IK Libraries :
    $ENV{'SERVTCPPORT'} = $IU::GcmdPort;
    $ENV{'SERVHOSTADDR'} = $IU::GcmdHostAddr;

    # Tell IK that we're the developers, if necessary:    (KLUDGE?)
    $ENV{'IMS_STAFF'} = '1'     if ($IU::StatsId eq 'v0ieww00');

    # Who are we ? a "guest" or a persistent user?
    $ENV{REMOTE_USER} = "guest" unless $ENV{REMOTE_USER};
    1;
}

#------------------------------------------------------------
# setup_state PARAMHASH
#------------------------------------------------------------
# Setup to create/read/write a state database for this session.  
# The CGI scripts will generally do this always, whereas the other
# programs will not.   Params are:
#
#   Filename  => DB, relative to the session dir (defaults to 'session.state')
#   Opts      => those given to IU::Store (default to 'WRCREAT|BLOCK');
#
# B<WARNING:> as a default, it is impossible to have two processes using the
# state database at the same time... so if you grab it, release it quickly!
# Also, be aware that a search may grab the session state for tens of
# minutes, so ADMIN SCRIPTS SHOULD NOT DEPEND ON GETTING IT!

sub setup_state {
    my ($self, %params) = @_;
    my $filename = $params{Filename} || 'session.state';
    my $opts     = $params{Opts}     || 'WRCREAT|BLOCK';

    my %STATE;
    my $path = $self->path_to($filename);    
    tie %STATE, 'IU::Store', $path, $opts, 0600    
	or puke "couldn't tie to $path (maybe: $!)";
    $self->{State} = \%STATE;
}

#------------------------------------------------------------
# state
#------------------------------------------------------------
# Return the database object for the session state, if any.
# The object responds to FETCH(), STORE(), etc. messages.

sub state {
    $_[0]->{State};
}

#------------------------------------------------------------
# touch
#------------------------------------------------------------
# Touch the session, to indicate that we were here.

sub touch {
    my $self = shift;
    my $kilroy = $self->dir . "/.kilroy";

    open KILROY, ">$kilroy" or puke "can't open $kilroy: $!";
    print KILROY time();
    close KILROY;
    1;
}

#------------------------------------------------------------
# url [RELATIVE]
#------------------------------------------------------------
# Assuming the session directory is in the standard place,
# return the URL for the session directory, with a trailing slash.
#
# With a RELATIVE argument, just appends the argument to the session URL,
# and returns that.
#
# If the ParentURL was specified as false (empty, 0, undef) in
# the create() or resume() constructor, then there is no way to
# determine the URL, and a fatal exception is raised.

sub url {
    my $self = shift;
    my $relative = shift || '';
    $self->{ParentURL} or puke "no URL: session created without a parent URL";
    "$self->{ParentURL}".uri_escape($self->{Id})."/$relative";
}

sub copy_directory {
    my $self = shift;
    my $source_dir = shift ;
    my $destination_dir = shift ;

    #debug "source_dir = $source_dir";
    # debug "destination_dir = $destination_dir ";
    system "cp --recursive $source_dir/* $destination_dir";
    # cp --recursive source destination
}

sub javascript_capable {
   my $self = shift;
   $self->state->{JavascriptCapable} = shift if defined $_[0];
   $self->state->{JavascriptCapable};
}

#------------------------------------------------------------
1;
