package IU::SetupFile;

=head1 NAME

IU::SetupFile - parse the SETUP file


=head1 DESCRIPTION

Parse the SETUP file.  Use it like this:

=begin pre
    require IU::SetupFile;

    $setup = IU::SetupFile->read("/home/imswww/v.1.4/SETUP");
    my @vars = $setup->vars;
    foreach $var (sort @vars) {
        print "$var = ", $setup->get($var), "\n";
    }    
=end pre

The configuration variables are also guaranteed to be 
stored under instance variables of the same name, 
so you can also write that C<print> statement like this:

=begin pre
        print "$var = $setup->{$var}\n";
=end pre


=head AUTHOR

Eryq / Hughes STX Corporation

=cut


use IU;
use IU::Log;

use strict;


#------------------------------------------------------------
# new -- create a new object
#------------------------------------------------------------
sub new {
    my $class = shift;
    my $new = {};
    bless $new, $class;
}

#------------------------------------------------------------
# read -- create and read in an object
#------------------------------------------------------------
sub read {
    my ($class, $path) = @_;
    my $new = $class->new;
    $new->read_in($path) or return undef;
    $new;
}

#------------------------------------------------------------
# read_in -- read in a SETUP file
#------------------------------------------------------------
sub read_in {
    my ($self, $path) = @_;
    my $line;

    open SETUP, $path or return error("couldn't open $path: $!");
    while (<SETUP>) {
	chomp;

	next if /^\s*#/;  # comment, possibly with whitespace before
	next if /^\s*$/;  # blank line
	if (/^\s*([A-Z][A-Za-z0-9]*)
	      \s+((\"([^\"]*)\")|(\S+))/x) {
            my $varname = $1;
	    my $value = (($4 ne '') ? $4 : $5);
	    $self->set($varname, $value);
	    debug "read '$varname' = '$value'";
        }
        else {
	    close SETUP;
            return error "syntax error, $path line $.";
        }
    }
    close SETUP;
    1;
}

#------------------------------------------------------------
# vars -- return the vars
#------------------------------------------------------------
sub vars {
    my $self = shift;
    grep {/^[A-Z]/} (keys %$self);
}

#------------------------------------------------------------
# set -- set the given value
#------------------------------------------------------------
sub set {
    my ($self, $var, $value) = @_;
    return undef if $var !~ /^[A-Z]/;
    $self->{$var} = $value;
}

#------------------------------------------------------------
# get -- return the given value
#------------------------------------------------------------
sub get {
    my ($self, $var) = @_;
    return undef if $var !~ /^[A-Z]/;
    $self->{$var};
}

#------------------------------------------------------------
1;
