package IU::SharedStore;


=head1 NAME

IU::SharedStore - a place to put IU::Storable objects

=head1 DESCRIPTION

This is basically a cache-less variant on IU::Store, for use
with blocking locks in a shared environment.  It uses an additional
file to manage synchronization.  I<The underlying database is only
tied when a read/write is requested, which is probably idiotic>.

Use this as a high-level DBMS interface.  It takes care of 
flattening/inflating objects for you.

B<WARNING:> I<you cannot store objects with internal pointer cycles!>

=head1 SYNOPSIS

To open the database for reading and writing, and create it if 
necessary, do this:

    # Open for on-demand creating/writing/reading:
    tie %DB, 'IU::SharedStore', "/tmp/file.gdb";
    
    # Write some info:
    $DB{'Liquid'}    = ['WATER!', 'H', [2], 'Oh!'];
    $DB{'Mr. Clean'} = 'NH_3';
    
    # Read some info:    
    print "Liquid is: ", $DB{'Liquid'}->[0], "\n";

    # Close:
    untie %DB;

=head1 WARNINGS

GDBM files are hardcoded to use non-blocking locks, which is ridiculous,
since blocking locks can be extremely useful in multiprocess environments.  
I<We> want blocking locks... so we have to simulate them by using 
a C<"..LCK"> file in tandem with our database file.  Grrrrrrr.....

=head1 BUGS

Inefficient, but this is unavoidable, given that we are trying to 
minimize the amount of time a process has the database file locked.

Very, very, disk intensive.  Even removing GDBM's sync() calls doesn't
improve this much.

Probably should not have been done with tied hashes, although it is possible
in the future to allow users to invoke methods that will give them
the power to decide when the underlying database should be tied/untied.
For our applications, this won't be needed.

The FIRSTKEY/NEXTKEY had to be implemented in a manner which demands
that FIRSTKEY do the tie, but I<not> do an untie... the untie is done
when NEXTKEY returns undef.  This means that you 
I<absolutely, positively MUST> iterate through I<all> of the keys!
If you stop in the middle, the database will remain tied and read-locked!

=cut

use GDBM_File;

use IU;
use IU::Log;
use IU::Storable;

use strict;

# Map names to modes:
my %ModeMap =  (  
       'READER'  => 0,
       'WRITER'  => 1,
       'WRCREAT' => 2,
       'NEWDB'   => 3
);

# Locks:
my %Lock = (SH=>1,
	    EX=>2,
	    NB=>4,
	    UN=>8);

#------------------------------------------------------------
# open FILE
#------------------------------------------------------------
# Open the database on the given FILE.  
# Beware: FILE..LCK will also be created if it doesn't exist!

sub open {
    my ($class, $file) = @_;

    # Create object:
    my $self = {
	Path => $file,
	GDBM => undef,    # will hold a GDBM_File, *not* a ref to a tied hash!
	Tied => 0,        # defined, but false for now 
    };
    bless $self, $class;

    # Open lockfile, creating if necessary:
    my $lockexpr = ((-e "$file..LCK") ? "+<$file..LCK" : "+>$file..LCK");
    $self->{Lockfile} = FileHandle->new($lockexpr)
	|| puke "couldn't open $lockexpr: $!";

    # Done!
    $self;
}

#------------------------------------------------------------
# close
#------------------------------------------------------------
# Close the database.

sub close {
    my $self = shift;
    $self->tie_un;
    1;
}

#------------------------------------------------------------
# tie_read
#------------------------------------------------------------
sub tie_read {
    my $self = shift;
    my $GDBM = {};

    # Return if already tied right; else, untie:
    return 1 if ($self->{Tied} eq 'READER');
    $self->tie_un if $self->{GDBM};

    # Lock and tie:
    flock($self->{Lockfile}, $Lock{SH}) 
	or puke "can't get shared lock for $self->{Path}: $!";
    tie(%$GDBM, 'GDBM_File', $self->{Path}, $ModeMap{'READER'}, 0600) 
	or puke "read-tie failed to $self->{Path} (maybe: $!)";
    $self->{GDBM} = tied(%$GDBM);
    $self->{Tied} = 'READER';
}

#------------------------------------------------------------
# tie_write
#------------------------------------------------------------
sub tie_write {
    my $self = shift;
    my $GDBM = {};

    # Return if already tied right; else, untie:
    return 1 if ($self->{Tied} eq 'WRCREAT');
    $self->tie_un if $self->{GDBM};

    # Lock and tie:
    flock($self->{Lockfile}, $Lock{EX}) 
	or puke "can't get exclusive lock for $self->{Path}: $!";
    tie(%$GDBM, 'GDBM_File', $self->{Path}, $ModeMap{'WRCREAT'}, 0600) 
	or puke "write-tie failed to $self->{Path} (maybe: $!)";
    $self->{GDBM} = tied(%$GDBM);
    $self->{Tied} = 'WRCREAT';
}

#------------------------------------------------------------
# tie_un
#------------------------------------------------------------
sub tie_un {
    my $self = shift;

    # Return if no tie:
    $self->{GDBM} or return 1;
    
    # Untie and unlock (yes, in that order!):
    $self->{GDBM} = undef;    # KLUDGE: will this untie????
    flock($self->{Lockfile}, $Lock{UN});
    $self->{Tied} = 0;        # defined and false
}

#------------------------------------------------------------
# TIEHASH CLASS,FILE,MODE,[FILEMODE]
#------------------------------------------------------------
# Tie to this class.
#
#      tie(%STORAGE, 'IU::SharedStore', "filename.gdb");

sub TIEHASH {      ## @_ is ($class, $path, $flags, $mode)
    my $class = shift;
    $class->open(@_);
}

#------------------------------------------------------------
# DESTROY
#------------------------------------------------------------
sub DESTROY {
    $_[0]->close;
}

#------------------------------------------------------------
# FETCH KEY
#------------------------------------------------------------
# Fetch from the database. 

sub FETCH {
    my ($self, $key) = @_;

    $self->tie_read;

    # Get from database, and inflate:
    my $flat = $self->{GDBM}->FETCH($key);
    my $value = (defined($flat) ? scalar(inflate IU::Storable \$flat) : undef);
    
    $self->tie_un;
    $value;
}

#------------------------------------------------------------
# STORE KEY,VALUE
#------------------------------------------------------------
# Store a key in the database.  Assumes that we're open for writing.

sub STORE {
    my ($self, $key, $fluffy) = @_;

    $self->tie_write;

    # Flatten, and store in database:
    my $flat = flatten IU::Storable $fluffy;    
    $self->{GDBM}->STORE($key, $flat);    # store flat on disk

    $self->tie_un;
    $fluffy;
}

#------------------------------------------------------------
# CLEAR -- standard method to clear hash
#------------------------------------------------------------
sub CLEAR {
    my $self = shift;

    $self->tie_write;
    $self->{GDBM}->CLEAR;
    $self->tie_un;
}

#------------------------------------------------------------
# DELETE -- standard method to delete key
#------------------------------------------------------------
sub DELETE {
    my ($self, $key) = @_;

    $self->tie_write;
    $self->{GDBM}->DELETE($key);
    $self->tie_un;
}

#------------------------------------------------------------
# EXISTS -- standard method: does key exist?
#------------------------------------------------------------
sub EXISTS {
    my ($self, $key) = @_;

    $self->tie_read;
    $self->{GDBM}->EXISTS($key);
    $self->tie_un;
}

#------------------------------------------------------------
# FIRSTKEY -- standard method to get first key
#------------------------------------------------------------
sub FIRSTKEY {
    my $self = shift;

    $self->tie_read;
    $self->{GDBM}->FIRSTKEY();
    # that's right... don't untie!
}

#------------------------------------------------------------
# NEXTKEY -- standard method to get next key
#------------------------------------------------------------
sub NEXTKEY {
    my $self = shift;

    my $nextkey = $self->{GDBM}->NEXTKEY(@_);
    $nextkey or $self->tie_un;
    $nextkey;
}

#------------------------------------------------------------
1;

