package IU::Spawn;

=head1 NAME

IU::Spawn - spawn a detached grandchild process to run some code

=head1 SYNOPSIS

Here's how to spawn a process and check on it yourself:

    # The subroutine we want to run in the background:
    sub spawned {
	my $spawn = shift;    # (ignored in this function)
         
	print "I'm the baby: gotta love me!\n";
	sleep 10;
	print "Not the mama!\n";
	1;
    }
    
    
    # Spawn the grandchild process:  
    my $tries = 0;
    my $spawn = start IU::Spawn Lockfile=>"/tmp/my..LCK", Action=>\&spawned; 
    
    # Get info:
    my $pid = $spawn->pid;
    my $lockfile = $spawn->lockfile;
    
    # Loop while it's still running...
    while ($spawn->running) {
	print "Try $tries; still going...\n";
	$spawn->kill if (++$tries > 10); 
	sleep 2;
    }
    print "Done after $tries tries!\n";


Here's how to check on something spawned by someone else... all you
need to know is the lock file name:

    my $other = attach IU::Spawn Lockfile=>"/tmp/my..LCK";
    while ($other->running) {
        ...
        $other->kill if ...
    }


=head1 NOTES

The algorithm was developed with several goals in mind. These were: 
preventing fork-based race conditions (via the pipes), 
being able to tell at any point whether the child has exitted without
fear of missing a signal (via the advisory lock), and the ability
to spawn code that would run with the image of the parent.

The lockfile contains the pid of the spawned process.  This is so
someone can attach an IU::Spawn to it just by specifying the lockfile name.

The pid is written by the grandchild into the pipe to the parent.
There is perhaps no urgent reason for this; I could just as easily have
written a single byte to the pipe, and had the parent read the pid from
the lockfile; but, since the lockfile was opened for writing...

We have to have the I<child> obtain an I<exclusive> lock, but the
parent/attaching processes attempt to obtain a I<shared> lock...
otherwise, a testing process that I<did> obtain the lock might appear
(for that brief time) to be the running process itself to I<another> 
testing process. 


=head1 FILE LOCKING

Solaris is apparently fussy about locks... in order to establish an
exclusive (write) lock, our file descriptor needs to be opened for
writing.  Hence, the lock file is opened for both reading and writing
("+<" or "+>") where appropriate.


=head1 BUGS

Should the spawned process be detached from a controlling tty to prevent 
accidents?

There is an I<infinitessimal> chance that the code:

    $spawn->kill;

could kill the wrong process under extremely high system loads, since
checking if the spawned process is running and performing the kill are
not combined into a single atomic operation.  But we're talking one in 
billions.

Since the spawned child becomes an orphan that is adopted by the "init"
process, there is no way to get its exit status.  This is reasonably fair,
since attaching processes don't have access to this information anyway.

=cut

use IU;
use IU::Log;
use IU::Util qw(safe_kill);
use FileHandle ();
use Fcntl;
use strict;

use Exporter;

@IU::Spawn::EXPORT_OK = qw(process_debug);
@IU::Spawn::ISA       = qw(Exporter);

# For flock()...
my %LOCK = (SH=>1, EX=>2, NB=>4, UN=>8);

my $DEBUG = uc($IU::ForceDebug) eq 'YES';

#------------------------------------------------------------
# new 
#------------------------------------------------------------
sub new {
    my $class = shift;
    bless {}, $class;
}

#------------------------------------------------------------
# safe_die
#------------------------------------------------------------
# Die as safely as possible.  We try not to do anything that
# would allocate memory or demand system resources.

sub safe_die {
    syswrite(STDERR, $_[0], length($_[0]));
    syswrite(STDERR, "\n", 1);
    exit -9;
}

#------------------------------------------------------------
# start PARAMHASH
#------------------------------------------------------------
# I<Class or instance method.> Spawn a grandchild.  Params are:
#
#    Lockfile => the lockfile to create
#    Action   => the subroutine to run
#    Session  => the session we're running in
#
# Fatal error if it can't be started.

sub start {
    my $self = shift;
    my %params = @_;
    my $buf = '';
    my $nread;

    # Create self object, if necessary:
    ref($self) or $self = $self->new;
    $self->{Origin} = 'START';
    process_debug("P: REMOTE_USER was $ENV{REMOTE_USER}");
    $ENV{REMOTE_USER} = $params{Session} unless $ENV{REMOTE_USER};
    process_debug("P: REMOTE_USER set to $ENV{REMOTE_USER}");
    process_debug("P: Lockfile: $params{Lockfile}");
    process_debug("P: Session: $params{Session}");

    # Init:
    $self->{Lockfile} = $params{Lockfile} or die "missing lockfile";
    $params{Action} or die "missing action";

    # Open file to be used as lockfile (unlink it first to force new inode):
    ### debug "unlink $self->{Lockfile}, open lockfile";
    unlink $self->{Lockfile};
    ($self->{LOCK} = FileHandle->new("+>$self->{Lockfile}"))
	or die "open $self->{Lockfile}: $!";
    $self->{LOCK}->autoflush(1);

    # Get local filehandles:
    ### debug "get FH objects";
    my $READER = new FileHandle;
    my $WRITER = new FileHandle;

    # Create pipe filehandles:
    pipe $READER, $WRITER or die "pipe: $!";
    $WRITER->autoflush(1);

    # Fork off a child:
    process_debug("P: about to fork");
    my $pid = fork();
    die "fork failed: $!" if (!defined($pid) || ($pid < 0));
    unless ($pid) {             # child
        process_debug("C: REMOTE_USER set to $ENV{REMOTE_USER}");

	# Reset the die handler with something safe:
	local(%SIG);
	$SIG{'__DIE__'} = \&safe_die;   # VITAL!

	# We don't need the reader:
	close $READER;

	# Close stdin and stdout here:
	process_debug("C: redirecting stdio");
	# $self->redirect_stdio;

	# Fork off a grandchild:
	my $cpid = fork();
	if (!defined($cpid) or ($cpid < 0)) {
	    print $WRITER "-00001 unable to fork grandchild: $! (",int($!),")";
            process_debug("C: fork failed: $!");
	    exit -2;
	}
	
	# Handle the grandchild:
	### debug "handling grandchild";
	unless ($cpid) {              # grandchild
            process_debug("GC: REMOTE_USER set to $ENV{REMOTE_USER}");

	    # Lock, and keep it open across exec's:
	    #    *** KLUDGE: SHOULD THAT BE '0' instead of pack()?
	    flock($self->{LOCK}, $LOCK{EX}) or die "flock: $!";
	    fcntl($self->{LOCK}, F_SETFD(), pack("L",0)) or die "fcntl: $!";

	    # Output pid to file:
	    process_debug("GC: output pid to file");
	    $self->{LOCK}->printf("%6d", $$);
	    
	    # We're ready for the parent to start polling us!
	    process_debug("GC: print and close");
	    print $WRITER sprintf("%6d", $$);     # send our pid
	    close $WRITER;

	    process_debug("GC: Running action");
	    my $state = &{$params{Action}}($self);
            process_debug("GC: done");
            close STDERR;                         # get errors out to log
            exit ($state ? 0 : 1);
	} 

	# We're the child, and the grandchild has just been forked...
        process_debug("C: done");
        close STDERR;                             # get errors out to log
	exit 0;
    } 

    # We're the parent, and our child has just been forked:
    process_debug("P: after fork, about to wait for pid $pid");

    # Reap child:  
    waitpid($pid, 0);

    # At this point, child has exitted, so the grandchild process image 
    #   must have been established, or there was a fatal error.
    #   Close other end of pipe:
    close $WRITER; 

    # Synchronize: read until data (grandchild) or EOF (dead):
    process_debug("P: about to read pipe...");
    $nread = read($READER, $buf, 6);
    ($nread == 6) or die "incomplete pid (perhaps fork did not complete?): $!";
    (int($buf) == -1) and die ("can't spawn process: ", <$READER>, "\n");
    close $READER;

    # Store info about child, and return! 
    $self->{KidPid} = $buf;       # happens to be its pid
    process_debug("P: done");
    $self;
}

#------------------------------------------------------------
# attach PARAMHASH
#------------------------------------------------------------
# I<Constructor:> attach to a process spawned by someone else.  Params are:
#
#    Lockfile => the lockfile to create
#
# Just returns undef if the lockfile doesn't exist.

sub attach {
    my $self = shift;
    my %params = @_;

    # Create self object, if necessary:
    ref($self) or $self = $self->new;
    $self->{Origin} = 'ATTACH';

    # Init:
    $self->{Lockfile} = $params{Lockfile} or die "missing lockfile arg!";
    (-e $self->{Lockfile}) or return undef;

    # Open lockfile for reading:
    $self->{LOCK} = FileHandle->new("+<$self->{Lockfile}");
    $self->{LOCK} or die "open $self->{Lockfile}: $!";
    
    # Get pid... this will enable us to kill():
    my $lock = $self->{LOCK};
    my $pid = <$lock>;
    $self->{KidPid} = $pid;

    # Ready:
    $self;
}

#------------------------------------------------------------
# DESTROY
#------------------------------------------------------------
# When a spawn goes out of scope, close our files, but DON'T KILL THE PROCESS!
sub DESTROY {
    my $self = shift;
    close $self->{LOCK} if $self->{LOCK}; 
}


#------------------------------------------------------------
# lockfile
#------------------------------------------------------------
sub lockfile { $_[0]->{Lockfile} }

#------------------------------------------------------------
# kill [SIGNAME]
#------------------------------------------------------------
# Send a signal to the process.  SIGNAME should be 'INT', 'KILL', etc.
# Default is 'TERM'.
#
# This exits with a no-op if the process is not running, to prevent
# accidents resulting from bad code.

sub kill {
    my $self = shift;
    my $sig = shift || 'TERM';
    $self->running or return undef;
    safe_kill $sig, $self->{KidPid};
} 

#------------------------------------------------------------
# origin 
#------------------------------------------------------------
# Was this object created via a START or an ATTACH?
sub origin {
    $_[0]->{Origin};
}

#------------------------------------------------------------
# pid
#------------------------------------------------------------
# What's the pid of our spawned process?

sub pid { $_[0]->{KidPid} }

#------------------------------------------------------------
# redirect_stdio
#------------------------------------------------------------
sub redirect_stdio {
    1;
}

#------------------------------------------------------------
# running
#------------------------------------------------------------
# Are they running?

sub running {
    my $self = shift;
    my $alive = !flock($self->{LOCK}, $LOCK{SH}|$LOCK{NB});
    $alive;
}

sub process_debug {
  shift @_ if $_[0] eq "IU::Spawn";
  print STDERR "($$) ",(join "",@_),"\n" if $DEBUG;
}

#------------------------------------------------------------
1;
