package IU::SpecializedCriteria;

=head1 NAME

IU::SpecializedCriteria - handle the SPECIALIZED_CRITERIA group

=head1 DESCRIPTION

This class models a new entity that is being put in the ODL
extensions.  The corresponding ODL looks like:

       GROUP = SPECIALIZED_CRITERIA
          CRITERIA_NAME = 
          CRITERIA_TYPE =
          CRITERIA_VALUE =
       END_GROUP = SPECIALIZED_CRITERIA


=head1 SYNOPSIS

    $result = new IU::SpecializedCriteria criteriaName =>"SUN_AZIMUTH_ANGLE",
                                          criteriaType => "REAL", 
                                          criteriaValue =>"142.7";
    

=head1 AUTHOR

Mark Solomon, Raytheon ITSS

=head1 RCS INFO

$Id: SpecializedCriteria.pm,v 1.3 1999/08/17 15:38:28 solomon Exp $

=cut

use IU;
use IU::Chunk;
use IU::ODL;
use IU::Log;
use strict;


#------------------------------------------------------------
# new
#------------------------------------------------------------
sub new {
    my $class = shift;
    my %params = @_;
    bless \%params, $class;
}

#------------------------------------------------------------
# criteriaName
#------------------------------------------------------------
# Return the criteriaName, or undef if there is none.

sub criteriaName {
    $_[0]->{CriteriaName};
}

#------------------------------------------------------------
# criteriaType
#------------------------------------------------------------
# Return the criteriaType, or undef if there is none.

sub criteriaType {
    $_[0]->{CriteriaType};
}

#------------------------------------------------------------
# criteriaValue
#------------------------------------------------------------
# Return return a reference to the array of criteriaValue, 
# or undef if there is none.

sub criteriaValue {
    $_[0]->{CriteriaValue} || [];
}

#------------------------------------------------------------
# from_chunk CHUNK
#------------------------------------------------------------
# I<Class or instance method.>
# Create a new instance from a chunk.

sub from_chunk {
    my ($self, $chunk) = @_;
    ref($self) or $self = $self->new;
    $self->{CriteriaName}  = $chunk->value('CRITERIA_NAME');
    $self->{CriteriaType}  = $chunk->value('CRITERIA_TYPE');
    $self->{CriteriaValue} = [$chunk->seq_values('CRITERIA_VALUE')];
    $self;
}

#------------------------------------------------------------
# from_ODL ODL
#------------------------------------------------------------
# I<Class or instance method.>
# Create a new instance from an ODL group.

sub from_ODL {
    my ($self, $odl) = @_;
    my $name_odl;
    my $type_odl;
    my $value_odl;

    # Be a class or instance method:
    ref($self) or $self = $self->new;

    # Extract CRITERIA_NAME:
    if ($name_odl = $odl->param('CRITERIA_NAME')) {
	my $name_value = $name_odl->text;
        $self->{CriteriaName} =  $name_value ;
	}

    $self->{CriteriaName} or puke "missing CRITERIA_NAME!";
    
    # Extract CRITERIA_TYPE:
    if ($type_odl = $odl->param('CRITERIA_TYPE')) {
	my $type_value = $type_odl->text;
        $self->{CriteriaType} =  $type_value ;
	}

    # Extract CRITERIA_VALUE:
    if ($value_odl = $odl->param('CRITERIA_VALUE')) { 
        # String, or sequence?
	$value_odl || puke " no criteria value";

	if ($value_odl->type eq 'ARRAY') {
	    $self->{CriteriaValue} = 
		join('', map { $_->text } $value_odl->as_array); 
	} 
	else {
	    $self->{CriteriaValue} = $value_odl->text;
	}
    }
    
    # Done!
    $self;    
}

#------------------------------------------------------------
1;
