package IU::SpecializedResults;

=head1 NAME

IU::SpecializedResults - handle the SPECIALIZED_RESULTS group

=head1 DESCRIPTION

This class models a new entity that is being put in the ODL
extensions.  The corresponding ODL looks like:

       GROUP = SPECIALIZED_RESULTS
          RESULT_NAME = 
          RESULT_VALUE =
       END_GROUP = SPECIALIZED_RESULTS


=head1 SYNOPSIS

    $result = new IU::SpecializedResults resultName =>"SUN_AZIMUTH_ANGLE",
                                         resultValue =>"142.7";
    
    # The same, using HTML streams (much safer):
    $HTML -> t($result->resultName) -> t($result->resultValue) -> _A;

=head1 AUTHOR

Mark Solomon, Raytheon ITSS

=head1 RCS INFO

$Id: SpecializedResults.pm,v 1.14 2003/11/14 15:02:01 amrhein Exp $

=cut

use IU;
use IU::Chunk;
use IU::ODL ();
use IU::Log;
use strict;

#------------------------------------------------------------
# new
#------------------------------------------------------------
sub new {
    my $class = shift;
    my %params = @_;
    bless \%params, $class;
}

#------------------------------------------------------------
# resultName
#------------------------------------------------------------
# Return the resultName, or undef if there is none.

sub resultName {
    $_[0]->{ResultName};
}

#------------------------------------------------------------
# resultValue
#------------------------------------------------------------
# Return return a reference to the array of resultValue, 
# or undef if there is none.

sub resultValue {
    $_[0]->{ResultValue} || [];
}

sub check {
   my ($self, $checkName) = @_;
   my $newName = $self->resultName;
   if($checkName eq $newName) {
      my $arrayRef = $self->resultValue;
      my @valueArray;
      my $newvalue;
      if (ref($arrayRef)) {
         @valueArray = @$arrayRef;
         if(int(@valueArray) > 1) {
            $newvalue = join(',', @valueArray);
         } else {
            $newvalue = $valueArray[0];
         }
      } else {
         $newvalue = $arrayRef;
      }
      return $newvalue;

    } else {
      return undef;
    }
}

#------------------------------------------------------------
# resultValueAsString
#------------------------------------------------------------
# I<Instance method.>
# Return the value as a string, or a comma-separated list, as appropriate.

sub resultValueAsString {
   my $self = shift;
   return join ", ", @{$self->resultValue};
}

#------------------------------------------------------------
# from_chunk CHUNK
#------------------------------------------------------------
# I<Class or instance method.>
# Create a new instance from a chunk.

sub from_chunk {
    my ($self, $chunk) = @_;
    ref($self) or $self = $self->new;
    $self->{ResultName}  = $chunk->value('RESULT_NAME');
    $self->{ResultValue} = [$chunk->seq_values('RESULT_VALUE')];
    $self;
}

#------------------------------------------------------------
# from_ODL ODL
#------------------------------------------------------------
# I<Class or instance method.>
# Create a new instance from an ODL group.

sub from_ODL {
    my ($self, $odl) = @_;
    my $name_odl;
    my $value_odl;

    # Be a class or instance method:
    ref($self) or $self = $self->new;

    # Extract RESULT_NAME:
    if ($name_odl = $odl->param('RESULT_NAME')) {
	my $value = $name_odl->text;
        $self->{ResultName} =  $value ;
	}

    $self->{ResultName} or puke "missing RESULT_NAME!";
    
    # Extract RESULT_VALUE:
    if ($value_odl = $odl->param('RESULT_VALUE')) { 
        # String, or sequence?
	$value_odl || puke " no RESULT_VALUE";

	if ($value_odl->type eq 'ARRAY') {
	    $self->{ResultValue} = 
		join('', map { $_->text } $value_odl->as_array); 
	} 
	else {
	    $self->{ResultValue} = $value_odl->text;
	}
    }
    
    # Done!
    $self;    
}

#------------------------------------

# Determines whether the granule is MISR/AMSRE/GLAS granule or not
# This method is used to determine whether a granule has a SpecialResult that
# would determine it to be either a MISR/AMSR-E/GLAS granule.

# The method returns 2 values for each sensor, if it IS, ISN'T a 
# MISR/AMSR-E/GLAS/NOSE granules.

sub are_special_granules {
    my $gran_ref = shift;

    my ($is_misr, $isnt_misr) =  IU::SpecializedResults::return_SR_count($gran_ref,"misr");
    my ($is_glas, $isnt_glas) =  IU::SpecializedResults::return_SR_count($gran_ref,"glas");
    my ($is_amsre, $isnt_amsre) =  IU::SpecializedResults::return_SR_count($gran_ref,"amsre");
    my ($is_nose, $isnt_nose) =  IU::SpecializedResults::return_SR_count($gran_ref,"omi");


   return ( $is_misr, $isnt_misr, $is_amsre, $isnt_amsre, $is_glas, $isnt_glas,$is_nose, $isnt_nose);
}

# Ok, based on WHICH instrument is passed, look for the specific PSAs to get the
# PathRow so we can display the Blocks on imscovermap.
#
# Wanna support another instrument? Added to the _list ;)
# Added Nose to the list 4/02/03 JSA


sub return_instrument_pathrow {
        my $specializedResultsArray = shift;
        my $instrument_name = uc(shift);
        my (@start_path, @start_row, @end_row);
        my @misr_list = ("SP_AM_PATH_NO", "SP_AM_MISR_StartBlock", "SP_AM_MISR_EndBlock");
        my @amsre_list = ("NominalPassIndex", "StartingPolygonNumber", "EndingPolygonNumber");
        my @glas_list = ("SP_ICE_PATH_NO", "SP_ICE_GLAS_StartBlock", "SP_ICE_GLAS_EndBlock");
        my @nose_list = ("PathNr", "StartBlockNr","EndBlockNr");

        my @generic_array;
        if ( $instrument_name eq "MISR") {
          @generic_array = @misr_list;
        }
        elsif ( $instrument_name eq "AMSRE") {
         @generic_array = @amsre_list;
        }
        elsif ( $instrument_name =~ "GLAS" || $instrument_name eq "GPS" ) {
         @generic_array = @glas_list;
        }
        elsif ( $instrument_name eq "OMI") {
         @generic_array = @nose_list;
        }

        my ($singleSpecializeResults);
        foreach $singleSpecializeResults (@$specializedResultsArray ) {
           if ( $singleSpecializeResults->resultName() eq $generic_array[0]) {
                 @start_path = @{$singleSpecializeResults->resultValue()};
           }
           if ( $singleSpecializeResults->resultName() eq $generic_array[1]  ) {
                 @start_row = @{$singleSpecializeResults->resultValue()};
           }
           if ( $singleSpecializeResults->resultName() eq $generic_array[2] ){
                 @end_row = @{$singleSpecializeResults->resultValue()};
           }
        }

   return ($start_path[0],$start_row[0], $end_row[0]);
}

sub return_SR_count {
        my $granule_ref = shift;
        my $granule_type = shift;
        my $num_GranType_gran =0;
        my $num_non_GranType_gran = 0;

         # Extract all of the Granule's Specialized Results
         # Determine whether it is a MISR granule or Not.
         my @specializedResultsArray = @{$granule_ref->results()};
         my ($start_path, $start_row, $end_row);
         if ( int(@specializedResultsArray) > 0) {
            ($start_path, $start_row, $end_row) =
              return_instrument_pathrow (\@specializedResultsArray,$granule_type); }
           if(defined($start_row) && defined($end_row) && defined($start_path)){
            $num_GranType_gran += 1;
           }
           else {
            $num_non_GranType_gran += 1 ;
           }
 return ( $num_GranType_gran , $num_non_GranType_gran );
}


#------------------------------------------------------------
1;
