package IU::Storable;

=head1 NAME

IU::Storable -- manage storable objects

=head1 DESCRIPTION

A pretty good way to "flatten" most objects into scalars,
for ultimate storage into a DBM-like database.  The following 
objects may be flattened:

=over

=item *

The undefined value.

=item *

Simple scalars.

=item *

Reference to unblessed arrays, whose elements are themselves storable objects.

=item *

References (blessed or unblessed) to hashes, whose keys are printable 
scalars, and whose instance variables are themselves storable objects.

=back

Classes which wish to have their objects be storable may inherit 
from this class to use the OO-flavored interface; namely, the
flatten() instance method, and the inflate() class method (which
is essentially a constructor).

=head1 SYNOPSIS

You can use it in a non-OO way...

    # Flatten an object:
    my $fluffy = ['nice', 'and', 'fluffy'];
    my $flat = IU::Storable->flatten($fluffy);
    
    # Inflate it again (note that we pass in a reference for efficiency):
    my $fluffy = IU::Storable->inflate(\$flat);

Or you can inherit from it...

    package Stuff;
    @ISA = qw(IU::Storable);
    ... 
    my $flat = $fluffy->flatten;
    my $fluffy = Stuff->inflate(\$flat);

=head1 WARNING

Storable objects may *not* contain circular reference chains: that is, 
they cannot contain references to objects which refer back to themselves.
Circular reference chains will result in an infinite loop.

The new Perl Storable base for this module supposedly corrects this problem,
but self-referential objects aren't that great an idea anyway.

=head1 NOTES

The Perl Storable class on which this one is based does all of its operations
at the C/Perl internals level. This provides much greater speed at the expense
of requiring that the Storable module be installed (approximately a 60-to-1
speedup, based on benchmarks).

Since Storable is most used in the "back-end" (i.e., the parts of the EDG
following the search form), you will see performance gains mostly in the
comm stat, granule list, data set list, cart, and folder.

=head1 VERSION

$Id: Storable.pm,v 1.11 2002/10/08 17:39:00 joe Exp $

=cut

use IU;
use IU::Log;

use Storable qw(freeze thaw);
use strict;


#------------------------------------------------------------
#
# UTILITIES
#
#------------------------------------------------------------

#------------------------------------------------------------
# flatten_it OBJECT,DEPTH,FLATREF
#------------------------------------------------------------
# I<Utility>.  Turn OBJECT into a scalar, adding it into FLATREF.

sub flatten_it {
    my ($o, $depth, $flatref) = @_;
    # Necessary to conform to the old IU::Storable interface.
    # We freeze a reference to the data, and put it back in the
    # proper place for dereferencing during inflate_it().
    $$flatref = freeze \$o;
}

#------------------------------------------------------------
# inflate_it FLATREF,DEPTH
#------------------------------------------------------------
# Turn scalar ref back into an object.

sub inflate_it {
    my ($flatref, $depth) = @_;    # used to separate values
    return undef unless defined $flatref;
    # Because we handled all the nasty multiple references in
    # flatten_it, we can just thaw the reference and all is good.
    thaw $flatref;
}

#------------------------------------------------------------
#
# PUBLIC INTERFACE
#
#------------------------------------------------------------

#------------------------------------------------------------
# flatten [OBJECTREF]
#------------------------------------------------------------
# I<As instance method:> turn the "self" object into a scalar.
# I<As class method:> turn OBJECTREF argument into a scalar.

sub flatten {
    my $self = shift;

    # What should we flatten?
    my $fluffy = (ref($self) ? $self : shift); 
    my $flat = '';
    $flat = flatten_it($fluffy, 0, \$flat);
}

#------------------------------------------------------------
# inflate SCALARREF
#------------------------------------------------------------
# I<Class method:> turn scalar back into an object.

sub inflate {
    my $type = shift;
    my $flatref = shift;  
    ref($flatref)        or puke 600, "please supply a SCALAREF as argument!";
    defined($$flatref)   or return undef;      # harmless kludge 
    my $foo = inflate_it($$flatref, 0);
    $$foo;
}

#------------------------------------------------------------

1;

