package IU::Store;


=head1 NAME

IU::Store - a place to put IU::Storable objects


=head1 DESCRIPTION

Use this as a high-level DBMS interface.  It takes care of 
flattening/inflating objects for you.  
B<WARNING:> I<you cannot store objects with internal pointer cycles!>

An IU::Store also maintains a cache so that if you fetch or store an
object, then subsequent fetches of that object are almost instantaneous.

Finally, you can use either non-blocking locks (the default), or
blocking locks.


=head1 SYNOPSIS

To open the database for reading and writing, and create it if 
necessary, do this:

    # Open for writing/creating/reading:
    tie %DB, 'IU::Store', "/tmp/file.gdb", 'WRCREAT';
    
    # Write some info:
    $DB{'Liquid'}    = ['WATER!', 'H', [2], 'Oh!'];
    $DB{'Mr. Clean'} = 'NH_3';
    
    # Read some info:    
    print "Liquid is: ", $DB{'Liquid'}->[0], "\n";

    # Close:
    untie %DB;

To open it in a read-only fashion, do this:
     
    # Open just for reading:
    tie %DB, 'IU::Store', "/tmp/file.gdb", 'READER';
    print "Liquid is: ", $DB{'Liquid'}->[0], "\n";
    untie %DB;

To open it in a read-only fashion, using blocking locks, do this:
     
    # Open just for reading:
    tie %DB, 'IU::Store', "/tmp/file.gdb", 'READER|BLOCK';
    print "Liquid is: ", $DB{'Liquid'}->[0], "\n";
    untie %DB;


=head1 WARNING

GDBM files are hardcoded to use non-blocking locks, which is ridiculous,
since blocking locks can be extremely useful in multiprocess environments.  
I<We> want blocking locks... so we have to simulate them by grabbing our
own lock on the file before GDBM does... grrr...

Remember that upgrading a lock temporarily releases it, so that other
processes can obtain and release the lock in that timeslice!


=head1 LOCKFILE MECHANISM

How is the ..LCK file opened, and what kind of lock is obtained?
This is based on the operation specified in the OPTS...

     Open as:    flock type   fcntl type   Open ..LCK file:
     ------------------------------------------------------------------
     READER      LOCK_SH      F_RDLCK      read-only
     WRITER      LOCK_EX      F_WRLCK      write, plus read
     WRCREAT        "            "                "
     NEWDB          "            "                "

Note that we I<cannot> depend on getting an exclusive lock on a file unless 
the lockfile can be opened for writing!  Although that works on IRIX, it
does I<not> work on Solaris.

Note also that attempting to open a non-existent database as a READER will 
fail, not at the tie(), but at the point where an attempt is made to
read-open the lockfile.  Opening WRCREAT does not have this problem.

Finally, note also that blocking vs. non-blocking does not matter.

=cut

use GDBM_File;
use FileHandle ();

use IU;
use IU::Log;
use IU::Storable;

use strict;

# Locks:
my %LOCK = (SH=>1, EX=>2, NB=>4, UN=>8);

# Map mode names to GDBM mode constants:
my %ModeToGdbm =  (  
       'READER'  => 0,
       'WRITER'  => 1,
       'WRCREAT' => 2,
       'NEWDB'   => 3
);

# Map mode names to flock locks:
my %ModeToFlock =  (  
       'READER'  => $LOCK{SH},
       'WRITER'  => $LOCK{EX},
       'WRCREAT' => $LOCK{EX},
       'NEWDB'   => $LOCK{EX},
);


#------------------------------------------------------------
# open FILE,OPTS,[PERMS]
#------------------------------------------------------------
# Open the database on the given FILE.
#
# OPTS contains any of: READER, WRITER, WRCREAT, or NEWDB.  It can also
# contain "BLOCK", which indicates that blocking locks are desired:
# e.g., "WRCREAT|BLOCK".  Use "NONBLOCK" to explicitly say you want 
# nonblocking locks; it's okay.
#
# PERMS is just for creating: defaults to 0644.
#
# If you ask for a nonblocking lock, and it can't be achieved, 
# simply returns undef.

sub open {
    my ($class, $file, $opts, $perms) = @_;
    defined($perms) or ($perms = 0644);

    # Get info from opts:
    my $modestr;
    ($modestr) = ($opts =~ /\b(READER|WRITER|WRCREAT|NEWDB)\b/);
    $modestr or puke "missing mode in options ($opts).";
    my $blocking = ($opts =~ /\bBLOCK\b/);

    # Get gdbm mode from mode string:
    my $gdbm_mode = $ModeToGdbm{$modestr};
    my $flock_mode = ($ModeToFlock{$modestr} | ($blocking ? 0 : $LOCK{NB}));

    # What kind of lock do we get, and how do we open the descriptor?
    #    READER  = flock(LOCK_SH) = fcntl(F_RDLCK) = open lockfile read-only
    #    WRITER  = flock(LOCK_EX) = fcntl(F_WRLCK) = open lockfile write+read
    #    WRCREAT = ditto
    #    NEWDB   = ditto
    my $lockpath = "$file..LCK";
    my $lockexpr;
    if ($modestr eq 'READER') {     # open read-only:
#	$lockexpr = "<$lockpath";
#       open read/write anyway for Solaris
	$lockexpr = ((-e $lockpath) ? "+<$lockpath" : "+>$lockpath");
    }
    else {                          # open read/write:
	$lockexpr = ((-e $lockpath) ? "+<$lockpath" : "+>$lockpath");
    }

    # PHASE 1: open a descriptor on file, and get a lock:    
    debug "obtaining lock $flock_mode on $lockexpr";
    my $lockfh = FileHandle->new($lockexpr) || puke "open $lockexpr: $!";
    flock($lockfh, $flock_mode) or 
	return error "flock of $lockexpr (mode $flock_mode): $!";

    # PHASE 2: tie... note that this may also attempt to get a lock! (NB lock)
    my $GDBM = {};
    die "Your database ($file) doesn't exist" 
      if (!-e $file && $modestr eq 'READER' );
    eval { tie(%$GDBM, 'GDBM_File', $file, $gdbm_mode, $perms) };
    puke "tie failed to $file, $gdbm_mode, $perms ($@)" if $@;

    # Create object:
    my $self = {
	Path => $file,
	Mode => $modestr,
	GDBM => tied(%{$GDBM}),    # a GDBM_File, *not* a ref to a tied hash!
	Cache => {},
	Lockfile => $lockfh,       # yes, we need to keep it open!!!!!
    };
    bless $self, $class;
}

#------------------------------------------------------------
# close
#------------------------------------------------------------
# Close the database.  Must untie first, THEN close any lockfile!

sub close {
    my $self = shift;
    $self->{GDBM} = undef;       # IS THIS ENOUGH TO UNTIE? 
    $self->{Lockfile}->close if $self->{Lockfile};
    1;
}



#------------------------------------------------------------
# TIEHASH CLASS,FILE,MODE,[PERMS]
#------------------------------------------------------------
# Tie to this class much the same as to a GDBM_FILE.   
# To open a database %DATABASE for reading, do this:
#
#      tie(%STORAGE, 'IU::Store', "filename.gdb", 'READER');
#
# To open a database %DATABASE for writing/creating with mode 0644, do this:
#
#      tie(%STORAGE, 'IU::Store', "filename.gdb", 'WRCREAT', 0644);
#
# The mode will default to 0644 if left off.

sub TIEHASH {      ## @_ is ($class, $path, $flags, $mode)
    my $class = shift;
    $class->open(@_);
}

#------------------------------------------------------------
# DESTROY
#------------------------------------------------------------
sub DESTROY {
    $_[0]->close;
}

#------------------------------------------------------------
# FETCH KEY
#------------------------------------------------------------
# Fetch from the database.  Assumes that we're open for reading.

sub FETCH {
    my ($self, $key) = @_;

    # Return from cache, if there:
    exists($self->{Cache}{$key}) and return $self->{Cache}{$key};

    # Get from database:
    my $flat = $self->{GDBM}->FETCH($key);
    defined($flat) or return($self->{Cache}{$key} = undef);   # not there!

    # Inflate, cache and return:
    $self->{Cache}{$key} = scalar(inflate IU::Storable \$flat);  
}

#------------------------------------------------------------
# STORE KEY,VALUE
#------------------------------------------------------------
# Store a key in the database.  Assumes that we're open for writing.

sub STORE {
    my ($self, $key, $fluffy) = @_;

    # Check for accidents:
    ($self->{Mode} ne 'READER') 
	or puke 600, "Can't store: database $self->{Path} opened readonly!";

    # Do it:
    my $flat = flatten IU::Storable $fluffy;    
    $self->{GDBM}->STORE($key, $flat);    # store flat on disk
    $self->{Cache}{$key} = $fluffy;       # save fluffy in cache    
    $fluffy;
}

#------------------------------------------------------------
# CLEAR -- standard method to clear hash
#------------------------------------------------------------
sub CLEAR {
    my $self = shift;
    $self->{Cache} = {};
    $self->{GDBM}->CLEAR;
}

#------------------------------------------------------------
# DELETE -- standard method to delete key
#------------------------------------------------------------
sub DELETE {
    my ($self, $key) = @_;
    $self->{GDBM}->DELETE($key);
}

#------------------------------------------------------------
# EXISTS -- standard method: does key exist?
#------------------------------------------------------------
sub EXISTS {
    my ($self, $key) = @_;
    exists($self->{Cache}{$key}) ||      # try cache first
	$self->{GDBM}->EXISTS($key);     # try db second
}

#------------------------------------------------------------
# FIRSTKEY -- standard method to get first key
#------------------------------------------------------------
sub FIRSTKEY {
    my $self = shift;
    $self->{GDBM}->FIRSTKEY();
}

#------------------------------------------------------------
# NEXTKEY -- standard method to get next key
#------------------------------------------------------------
sub NEXTKEY {
    my $self = shift;
    $self->{GDBM}->NEXTKEY(@_);
}

#------------------------------------------------------------
1;
