=head1 NAME

IU::SysLogger -- a syslog() logger

=cut


package IU::SysLogger;

use IU::Logger;
use IU::Syslog;

@ISA = qw(IU::Logger);

use strict;

#------------------------------------------------------------
# open -- open a syslog stream
#------------------------------------------------------------
# DESCRIPTION
#    This opens the syslog stream for logging messages.
#    It must protect against Sys::Syslog's eagerness to die on failure.
#
# CHANGE LOG
#    03/24/96 JDM Changed log level from 'user' to 'local0' to allow
#                 isolation of GLOBE log messages.

sub open {
    my ($type) = @_;

    # Get a new object:
    my $new = IU::Logger->open();

    # Open the syslog(), and verify that it's there without delay!
    my $evalreturn = eval { openlog("$::ProgName $$", 'ndelay', 'local0') };
    IU::Log::set_msg_sentinel(0);
    $@ and 
	return IU::Log::error("openlog() died: $@");    # eval okay?
    $evalreturn or 
	return IU::Log::error("openlog() failed: $!"); # openlog okay?

    bless $new, $type;
}

#------------------------------------------------------------
# msg -- the message function
#------------------------------------------------------------
#
# DESCRIPTION
#    Log the message via the syslog() facility.

sub msg {
    my ($self, $msgtype, $msgno, $caller, @msg) = @_;
    my $facility;

    # Convert message number to log info:
    if ($msgtype eq 'DEBUG') {
	$facility = 'debug';
	return undef;                    # don't syslog these
    }
    elsif ($msgtype eq 'INFORM') { 
	$facility = 'info';
    }
    elsif ($msgtype eq 'WARNING') { 
	$facility = 'warning'; 
    }
    else {
	$facility = 'err';
    }

    # Format the message, replacing all illegal whitespace with spaces:
    my $msg = join '', @msg;
    $msg =~ s/\s/ /g;

    # Get the filename, erasing everything before directory "perl":
    my $file = (($caller->[1] =~ m|/perl/(.+)$|) ? $1 : 'eval');
    my $line = $caller->[2];

    # Log it:
    $msgno or $msgno = '000';
    eval { syslog($facility, "$file $line $msgno $msg") };    
}

#------------------------- end of file -------------------------
1;
