package IU::TermLogger;

=head1 NAME

IU::TermLogger -- log to a terminal

=head1 DESCRIPTION

This logging stream manages standard GLOBE terminal output.
Normally, everything but a DEBUG message is output to the
terminal: however, this can be altered at any time by changing 
the $IU::MsgLevel variable.

When running as a CRON job, the recommendation is to set
$IU::MsgLevel to 0 (quiet) at the very start: that way, only serious
errors will be output, and these will be emailed to the account
owner automatically.

=head1 AUTHOR

Eryq -- Hughes STX Corporation

=cut


use IU::Logger;

@ISA = qw(IU::Logger);

use strict;
use vars qw(%TermGlyphs);

#------------------------------------------------------------
#
# Globals...
#
#------------------------------------------------------------

# The table of glyphs, for logging to terminals:
%TermGlyphs = (
	      DEBUG   => '(dbg) ',
	      INFORM  => '',
	      WARNING => 'Warning: ',
	      ERROR   => 'ERROR: ',
	      EXTRA   => '(!): ',
	   );

#------------------------------------------------------------
# open -- open on either STDOUT or STDERR
#------------------------------------------------------------
sub open {
    my ($type, $which) = @_;
 
    my $new = IU::Logger->open();
    $new->{FH} = $which;
    bless $new, $type;
}

#------------------------------------------------------------
# msg -- log the message
#------------------------------------------------------------
# NOTE: Uses $IU::MsgLevel to determine whether or not to output
#       the message.

sub msg {
    no strict 'refs';
    my ($self, $msgtype, $msgno, $caller, @msg) = @_;
    my $msg = join('', @msg);
    $msg =~ s/\n+$//g;

    # If no explicit message level, assume normal mode:
    my $msglevel = (defined($IU::MsgLevel) ? $IU::MsgLevel : 1);  

    # Should we output the message?
    my $output;
    if ($msglevel == 0) {      # quiet: only severe errors go out
	$output = ($msgno >= 700);          # only classified, serious errors
    }
    elsif ($msglevel == 1) {   # normal:
	$output = ($msgtype ne 'DEBUG');    # anything but a debug
    }
    else {                     # chatty:
	$output = 1;                        # anything goes
    }
    $output or return;

    # Output!
    my $FH = $self->{FH};
    print $FH $TermGlyphs{$msgtype}, $msg, "\n";
}

#------------------------- end of file -------------------------
1;
