package IU::TrackedOrder;


=head1 NAME

IU::TrackedOrder - orders returned in product results.

=head1 SYNOPSIS

ODL interface:

    # ODL interface:
    $q = IU::Query->from_ODL($odl);
    $odl = $q->to_ODL;

=head1 DESCRIPTION

This object encapsulates orders returned from Data centers 

=cut

use IU;
use IU::Log;
use IU::IW;

use IW::Util qw(:string);    # For trim() and blank();
use strict;

use vars qw(@ISA @EXPORT
            $TRACKED_ORDER);
@ISA = qw(Exporter);
@EXPORT = qw($TRACKED_ORDER);
#------------------------------
#
# Globals...
#
#------------------------------

$TRACKED_ORDER = undef;
 
#------------------------------------------------------------
# new PARAMHASH
#------------------------------------------------------------
# I<Constructor.>  Params are:
#
#     ArchiveId => archive id
#     OrderId => order id
#     DateSubmitted => date order submitted
#     NumDatasets => number of datasets ordered
#     DataSets => ordered data sets
#     NumGranules => number of granules ordered
#     Granules => ordered granules   
#
sub new {
    my ($class, %params) = @_;
    my $self = \%params;
    bless $self, $class;
 
}
#------------------------------------------------------------
# save_to_session ODL, OPTIONS
#------------------------------------------------------------
# I<Instance method, private.>
# Output the query ODL to a file in the session directory.
# File is named ArchiveId.OrderId.order
# Returns the path to the file, or undef/exception on error.

sub save_to_session {
    my ($self) = @_;
   #If order is from ASF, strip off the characters "OrderID="
   $self->check_order_id;
    my $extension = ".order";
    my $name = $self->{ArchiveId}.".".$self->{OrderId}.$extension;

    # If a subdirectory for tracked orders doesn't exist, create it.
    my $orderdir = $SESSION->path_to('orders');
    unless (-d $orderdir) {
      mkdir ($orderdir, 0755) or puke "mkdir $orderdir: $!"; }
    my $path = $orderdir."/".$name;

    # open file, convert order to odl, write it into subdirectory
    open ORDER, ">$path" or puke 452, "can't open $path for writing: $!";
    my $odl = $self->to_ODL;
    $odl->output(\*ORDER);
    $odl->end(\*ORDER);
    close ORDER;
    1;
}
#---------------------------------------------------------
# check_order_id
#---------------------------------------------------------
# ASF puts the string "OrderID" before the actual order id
# this needs to be stripped off before saving the orderid
sub check_order_id {
    my ($self) = @_;
    my $order_id = $self->{OrderId};

    if ($self->{ArchiveId} =~ /asf/i ) {
        #yes, order is from ASF, strip off the characters "OrderID="
        $order_id =~ s/OrderID=//;
        $self->{OrderId} = $order_id;
    }
}

#------------------------------
#
# ODL interface
#
#------------------------------
#------------------------------------------------------------
# from_ODL $odl
#------------------------------------------------------------
# Convert  ODL to a tracked order object, and return it.
# The ODL object is a group *containing* an order id, 

# data center id, date order submitted, number of datasets,
# datasets, number of granules, granules

sub from_ODL {
   my ($self, $odl) = @_;
   $odl || return undef;
   # Be a class method, too!
   ref($self) or $self = $self->new;
   my $daac_odl;
   my $trackno_odl;
   my $date_odl;
   my $numdsets_odl;
   my $numdsets;
   my $dataset_odl;
   my $numgranules;
   my $numgranules_odl;
   my $granule_odl;

   my ($param, $value);
   my $result;

   # look for the tracked order odl group and use it to create
   # the tracked order, setting the tracked order with the data 
   # center id, order tracking number, date submitted, number of
   # data sets, data sets, number of granules and granules
   if ($result = $odl->param('TRACKED_ORDER')) {
      if ($daac_odl = $result->param('DATA_CENTER_ID')) {
         my $daac_value = $daac_odl->text;
         $self->{ArchiveId} = $daac_value;
      }
      if ($trackno_odl = $result->param('ORDER_TRACKING_NUMBER')) {
         my $trackno_value = $trackno_odl->text;
         $self->{OrderId} = $trackno_value;
      }
      if ($date_odl = $result->param('DATE_SUBMITTED')) {
         my $date_value = $date_odl->text;
         $self->{DateSubmitted} =  $date_value;
      }
      if ($numdsets_odl = $result->param('NUM_DATASETS')) {
          $numdsets = $numdsets_odl->text;
          $self->{NumDatasets} = $numdsets;

      }
 
      if ($dataset_odl = $result->param('DATASET_IDS')) {
          my @datasets = $dataset_odl->text_values;
          $self->{DataSets} = \@datasets;
      }
      if ($numgranules_odl = $result->param('NUM_GRANULES')) {
          $numgranules = $numgranules_odl->text;
          $self->{NumGranules} = $numgranules;

      }
 
      if ($granule_odl = $result->param('GRANULE_IDS')) {
          my @granules = $granule_odl->text_values;
          $self->{Granule} = \@granules;
 
       }
    
   }
   return $self;
}
#------------------------------------------------------------
# to_ODL PARAMHASH
#------------------------------------------------------------
# Convert tracked order to an ODL object, and return it.
# The ODL object is a group *containing* an order id, 
# data center id, date order submitted, number of data sets,
# data sets, number of granules and granules
#
sub to_ODL {
    my $self = shift;

    my $TO;
    my @datasets;
    my @granules;

    # The data sets and granules in the list from the shopping 
    # cart may be from more than one data center. 
    # Find the data sets and granules in the list
    # that are from the data center where this order was placed
    @datasets = $self->datasets_for_this_daac;
    @granules = $self->granules_for_this_daac;

    my $num_dsets = int(@datasets);
    my $num_grans = int(@granules);

    require IU::ODL;    
    my $file = IU::ODL->new_group;
    $file->add_param(TRACKED_ORDER=>($TO=IU::ODL->new_group));

    $TO->add_param(DATA_CENTER_ID      => $self->{ArchiveId}); 
    $TO->add_param(ORDER_TRACKING_NUMBER => $self->{OrderId});
    $TO->add_param(DATE_SUBMITTED => $self->{DateSubmitted});

    $TO->add_param(NUM_DATASETS => $num_dsets);    
    $TO->add_param(DATASET_IDS => IU::ODL->sequence(@datasets));

    $TO->add_param(NUM_GRANULES => $num_grans);    
    $TO->add_param(GRANULE_IDS => IU::ODL->sequence(@granules));

    $file;
}
#------------------------------------------------------------
# datasets_for_this_daac
#------------------------------------------------------------
# Find the datasets in this list that are from this data center.

sub datasets_for_this_daac {
    my $self = shift;
    my  $dsets = $self->{DataSets};
    my $daac = $self->{ArchiveId};
    my $num_dsets = int(@$dsets);
    my @datasets;

    # Compare the data center id in the data set with the 
    # tracked order archive id. If there is a match, push
    # the data set on the list.
    for (my $i = 0; $i < $num_dsets; $i++) { 
       if($daac eq @$dsets[$i]->archive_id) {
           
           my $dsetid = @$dsets[$i]->dataset_id;
           $dsetid = trim($dsetid);
           push @datasets,$dsetid;
       }
    }
    @datasets;
}
#------------------------------------------------------------
# granules_for_this_daac
#------------------------------------------------------------
# Find the granules in this list that are from this data center.

sub granules_for_this_daac {
    my $self = shift;
    my  $grans = $self->{Granules};
    my $daac = $self->{ArchiveId};
    my $num_grans = int(@$grans);
    my @granules;

    # Compare the data center id in the granule with the 
    # tracked order archive id. If there is a match, push
    # the granule on the list.
    for (my $i = 0; $i < $num_grans; $i++) { 
       if($daac eq @$grans[$i]->archive_id) {
           my $granid = @$grans[$i]->granule_id ;
           $granid = trim($granid);
           my $dsetid = @$grans[$i]->dataset_id ;
           $dsetid = trim ($dsetid);
           push @granules,$granid.' from data set '.$dsetid;
       }
    }
    @granules;
}

1;
