package IU::UserAdmin;

=head1 NAME

IU::UserAdmin - Wrapper class for doing administration of users.  Add, update
and delete user accounts

=head1 DESCRIPTION

external API:

exists($account_name) - check if name is already in use.

update($account_name,$password) - change a password for an account.

add($account_name,$password) - add a new user.

delete($account_name) - delete a user.

check_name_syntax($account_name) - check syntax of account name (and die). 

check_password_syntax($password) - check syntax of account password. 

invalid_name($account_name) - check syntax of account name (without dying).

login($account_name,$password) - verify that account and password are good.

valid-email($email) - check to see if the supplied address is RFC822-valid.

users() - return array list of users.

lookup_sessions_by_email($email) - return list of sessions with this email.

lookup_email_by_session($session) - return email associated with this session.

assign_random_password($session) - generate/change password for session

get_random_password - generate a random password without setting it in the 
session

attach_emails - adds/updates email addresses to all sessions

=cut
use IU;
use IU::Log;
use IU::Session;
use IU::Util qw(chown_by_name);

use IW::Util qw(:url :output :string);
use IW::ModePREFS;  
use strict;


#-------------------------------------------------------------------------
# new - constructor
#-------------------------------------------------------------------------

sub new {
  my $class = shift;
  
  # Create a new instance:
  my $self = {};
  bless $self, $class;
  
  $self->{passwdfile} = ${IU::PersistentDir}.".htpasswd";
$self;
}

#-------------------------------------------------------------------------
# add 
#-------------------------------------------------------------------------

sub add {
  
  my $self = shift;
  my $name = shift;
  my $passwd = shift;
  my $email  = shift || $self->lookup_email_from_session($name);
  
  my $passwdfile = $self->{passwdfile};
  
  my $encrypted = $self->encrypt($passwd);
  
  # lock password file.
  $self->lock;
  
  # Add new user and passwd to passwd file.
  open HTPASSWD, ">>$passwdfile"
    or puke "couldn't open $passwdfile: $!" ;
  
  print HTPASSWD join(":",$name, $encrypted,$email),"\n";
  close HTPASSWD;
  
  # unlock password file
  $self->unlock;
  
  1;
}

#-------------------------------------------------------------------------
# exists - check if account exists already 
#-------------------------------------------------------------------------

sub exists {
  
  my $self = shift;
  my $username = shift;
  
  my $passwdfile = $self->{passwdfile};
  
  if (-e "$passwdfile") {
    open HTPASSWD, "$passwdfile" or
      puke "couldn't open $passwdfile: $!" ;
  } 
  else {
    return 0;  # no .htpasswd means user doesn't exist.
  }
  
  my $name;
  my $password;
  my $email;
  my $found = 0;
  while (<HTPASSWD>) {
    ($name,$password,$email) = split /:/, $_;
    if ($name eq $username) {
      $found = 1;
      last;
    }
  }
  close HTPASSWD;
  return $found;
}

#-------------------------------------------------------------------------
# update - change password 
#-------------------------------------------------------------------------

sub update {
  
  my $self = shift;
  my $username = shift;
  my $newpasswd = shift || "";
  my $newemail  = shift;
  $newemail = $self->lookup_email_from_session($username) unless $newemail;
  debug "user '$username', email '$newemail'";
  
  my $encrypted = $self->encrypt($newpasswd) if $newpasswd;

  my $name;
  my $passwd;
  my $email;
 
  my $passwdfile = $self->{passwdfile};
  
  if (-e "$passwdfile") {
    open HTPASSWD, "$passwdfile" or
      puke "couldn't open $passwdfile: $!" ;
  } 

  local $_; 
  while (<HTPASSWD>) {
    chomp;
    if (/^$username:/) {
       ($name,$passwd,$email) = split /:/;
      last;
    } 
  }
  close HTPASSWD;

  $passwd = $newpasswd if $newpasswd;
  $email  = $newemail  if $newemail;

  $self->delete($name);
  $self->add($name,$passwd,$email);
  1;
}

#-------------------------------------------------------------------------
# delete a user 
#-------------------------------------------------------------------------

sub delete {
  
  my $self = shift;
  my $username = shift;
  
  $self->_ready_to_copy();
  
  my $line;
  my $name;
  my $passwd;
  my $email;
  
  while (<HTPASSWD>) {
    $line = $_;
    chomp $line;
    ($name,$passwd,$email) = split /:/, $line;
    
    next if $name eq $username;
    
    print TMP $line,"\n";
  }
  
  $self->_update_done;
  1;
}

#-------------------------------------------------------------------------
# check_name_syntax
#-------------------------------------------------------------------------

sub check_name_syntax {
  
  my $self = shift;
  my $account = shift;
  my $diagnosis;
  
  if ($diagnosis = $self->invalid_name($account)) {
    return $diagnosis;
  }
  
  # Othewise, we liked it.
  1;
}

#-------------------------------------------------------------------------
# invalid_name
#-------------------------------------------------------------------------
sub invalid_name {
  my $self = shift;
  my $account = shift;
  my $diagnosis = "";
  
  # Make sure we only have .htpasswd safe characters.
  (length($account) < 15) or $diagnosis = 
    "Sorry! Account name is too long, must be 14 or less characters.";
  !($account =~ /^guest$/ ) or $diagnosis = 
    "Sorry! Can not use guest as an account name.";
  ($account =~ /^[a-z0-9_]+$/i )  or $diagnosis =
    "Sorry! Account name can only have characters: ".
      "abcdefghijklmnopqrstuvwxyz, ABCDEFGHIJKLMNOPQRSTUVWXYZ, ".
        "0123456789 or underscores \"_\". No spaces or tabs.";
  $diagnosis; 
}

#-------------------------------------------------------------------------
# check_password_syntax
#-------------------------------------------------------------------------

sub check_password_syntax {  
  # No restrictions on password
  1;
}

#-------------------------------------------------------------------------
# login 
#-------------------------------------------------------------------------
# Check if valid username and password

sub login {
  my $self = shift;
  my $username = shift;
  my $testpass = shift;
  
  # get encoded password from accesslist
  my $htpasswd_file = $self->{passwdfile};
  open HTPASSWD, "$htpasswd_file" or
    puke "couldn't open $htpasswd_file: $!" ;
  
  my $name;
  my $password;
  my $email; 
  my $found = 0;
  while (<HTPASSWD>) {
    ($name,$password,$email) = split /:/, $_;
    if ( $name eq $username) {
      $found = 1;
      last;
    }
  }
  close HTPASSWD;
  
  return 0 if ($found eq 0);
  
  my $realenc = trim($password);
  my $salt = substr($realenc, 0, 2);
  my $testenc = $self->encrypt($testpass, $salt);
  if ($testenc ne $realenc) {
    return 0;
  } 
  1;
}

#-------------------------------------------------------------------------
# encrypt
#-------------------------------------------------------------------------
# encrypt a password

sub encrypt {
  
  my $self = shift;
  my $password = shift;
  
  # if salt wasn't supplied then generate it.
  # 62 is the size of the array ('0'..'9','a'..'z','A'..'Z'), so
  # [int(62 * rand)] randomly picks one element from the array.
  srand(time);
  my $salt = shift || (('0'..'9','a'..'z','A'..'Z')[int(62 * rand)] .
                       ('0'..'9','a'..'z','A'..'Z')[int(62 * rand)]);
  
  crypt($password,$salt);
  
}

#-------------------------------------------------------------------------
# users 
#-------------------------------------------------------------------------
# return list of users from password file

sub users {
  
  my $self  = shift;
  my @users;
  
  my $htpasswd_file = $self->{passwdfile};
  open HTPASSWD, "$htpasswd_file" or
    puke "couldn't open $htpasswd_file: $!" ;
  
  my $name;
  my $password;
  
  while (<HTPASSWD>) {
    my ($name,$password,$email) = split /:/, $_;
    push @users, $name;
  }
  close HTPASSWD;
  return @users;
}

#-------------------------------------------------------------------------
# assign_random_password
#-------------------------------------------------------------------------
# Change password for a user to a random string and return the string.

sub assign_random_password {
  my ($self, $user) = @_;
  
  # Number of probes to make to determine possible following char
  my $CONTEXT = $IU::RandomContext;
  
  # Length of word to generate
  my $LENGTH  = $IU::RandomLength;
  
  # Read the dictionary and clean it up. Remove newlines and 
  # translate all characters to lower case.
  my $dict = ( ($IU::DictFile =~ m|/|) 
               ? $IU::DictFile 
               :"$IU::DocsDir/dict/$IU::DictFile" );
  open (Z,$dict) or puke "Can't open dictionary $dict: $!";
  
  # Note: normally I would just localize $/, but that does NOT
  # seem to work properly. We save and restore it instead.
  my $prev_dslash = $/;
  $/ = "";
  my $corpus = <Z>;
  close Z;
  $/ = $prev_dslash;
  $corpus =~ s/\n/ /g;
  $corpus =~ tr/A-Z/a-z/;
  
  # All 'my' variables are defined outside the loop for maximum speed.
  my $c;
  my @choices;
  my $count;
  my $scan = " ";
  my $out = "";
  
  # Keep going until we have enough characters.
  while (length($out) < $LENGTH) {
    # Do $CONTEXT number of probes to get context for the next letter.
    # We initially start out with no results from probes.
    $count = $CONTEXT;
    @choices = ();
    while ($count--) {
      # Select random start point inside dictionary.
      my $start_pos = int rand length $corpus;
      # select a character by scanning for the current character,
      # starting at the random point we just calculated, and extracting
      # the character right after that. 
      #
      # We do this with index and substring because they are considerably 
      # faster in this application than pattern matching is. (Believe me,
      # I tried pattern matching first!)
      $c = substr($corpus,(index $corpus,$scan,$start_pos)+1, 1);
      # If the character isn't useful, try another random probe.
      redo if grep /$c/,("'"," ");
      # Save it if it is useful.
      push @choices,$c;
    }
    # Choose one of the characters we pulled out at random, and add it to
    # the word. It becomes the context for the next probe.
    $scan = $choices[int(rand($CONTEXT))];
    $out .= $scan;
  }
  
  # Word is full up. Use it.
  $self -> update($user, $out);
  # Return the word so we can mail it out.
  $out;
}

#-------------------------------------------------------------------------
# get_random_password
#-------------------------------------------------------------------------
# Get a password for a user to a random string and return the string.
# This gets a password before setting it so that we can send a message to
# ECS first before we actually set the password.
# This is to support seamless user registration.

sub get_random_password {
  my $self = shift;
  
  # Number of probes to make to determine possible following char
  my $CONTEXT = $IU::RandomContext;
  
  # Length of word to generate
  my $LENGTH  = $IU::RandomLength;
  
  # Read the dictionary and clean it up. Remove newlines and 
  # translate all characters to lower case.
  my $dict = ( ($IU::DictFile =~ m|/|) 
               ? $IU::DictFile 
               :"$IU::DocsDir/dict/$IU::DictFile" );
  open (Z,$dict) or puke "Can't open dictionary $dict: $!";
  
  # Note: normally I would just localize $/, but that does NOT
  # seem to work properly. We save and restore it instead.
  my $prev_dslash = $/;
  $/ = "";
  my $corpus = <Z>;
  close Z;
  $/ = $prev_dslash;
  $corpus =~ s/\n/ /g;
  $corpus =~ tr/A-Z/a-z/;
  
  # All 'my' variables are defined outside the loop for maximum speed.
  my $c;
  my @choices;
  my $count;
  my $scan = " ";
  my $out = "";
  
  # Keep going until we have enough characters.
  while (length($out) < $LENGTH) {
    # Do $CONTEXT number of probes to get context for the next letter.
    # We initially start out with no results from probes.
    $count = $CONTEXT;
    @choices = ();
    while ($count--) {
      # Select random start point inside dictionary.
      my $start_pos = int rand length $corpus;
      # select a character by scanning for the current character,
      # starting at the random point we just calculated, and extracting
      # the character right after that. 
      #
      # We do this with index and substring because they are considerably 
      # faster in this application than pattern matching is. (Believe me,
      # I tried pattern matching first!)
      $c = substr($corpus,(index $corpus,$scan,$start_pos)+1, 1);
      # If the character isn't useful, try another random probe.
      redo if grep /$c/,("'"," ");
      # Save it if it is useful.
      push @choices,$c;
    }
    # Choose one of the characters we pulled out at random, and add it to
    # the word. It becomes the context for the next probe.
    $scan = $choices[int(rand($CONTEXT))];
    $out .= $scan;
  }
  
  # Word is full up. Use it.
  # Return the word so we can mail it out.
  $out;
}
#-------------------------------------------------------------------------
# lookup_sessions_by_email($email)
#-------------------------------------------------------------------------
sub lookup_sessions_by_email {
  my $self = shift;
  my $sought_email = shift;
  
  my $passwdfile = $self->{passwdfile};
  $self->lock;
  
  open HTPASSWD, "$passwdfile" or
    puke "couldn't open $passwdfile: $!" ;
  
  my $line;
  my $name;
  my $passwd;
  my $email;
  my @matches = ();
  
  local $_;
  while (<HTPASSWD>) {
    $line = $_;
    chomp $line;
    ($name,$passwd,$email) = split /:/, $line;
    push @matches, $name if ($email eq $sought_email);
  }
  close HTPASSWD;
  
  # unlock the password file.
  $self->unlock;
  
  # Return matches, if any.
  @matches;
}

#-------------------------------------------------------------------------
# lookup_email_from_session($name)
#-------------------------------------------------------------------------
# If the user exists in the password file, load the user's preferences
# and extract the email from there (done because the password file entry
# may not be "fresh" - the user could have updated the address since the
# session was created).

sub lookup_email_from_session {
  my $self = shift;
  my $user = shift;
  my $email = undef;
  my $dont_install = 1;
  if ($self->exists($user)) {
    debug "$user is in the password file";
    my $dir = IU::Session->persistent_dir($user);
    debug "user's directory is $dir";
    my $prefs_file = ($dir."/prefs.dat");
    debug "looking for prefs file $prefs_file";
    if (-e $prefs_file) {
      debug "prefs file found";
      my $user_info  = IW::ModePREFS::load_prefs($prefs_file, $dont_install);
      $email = $user_info->req_addr->email;
    }
  }
  $email;
}

#-------------------------------------------------------------------------
# attach_emails - update all email addresses
#-------------------------------------------------------------------------
# Do not use from CGI!

sub attach_emails {
  
  my $self = shift;
  my %param = @_;
  my $verbose = exists $param{Verbose};
  
  # Mark the system down, unless it's already down.
  my $remove_down = 0;
  unless (-e "$IU::LibDir/conf/down.txt") {
    open DOWN, ">$IU::LibDir/conf/down.txt" or
      puke "Can't create down.txt: $!";
    print DOWN <<EOF;
System down for password file maintenance. Back in a few moments.
EOF
    close DOWN;
    $remove_down++;
  }
  
  $self->_ready_to_copy;
  
  my $line;
  my $name;
  my $passwd;
  my $email;
  
  inform "Updating email addresses in password file ...";  
  while (<HTPASSWD>) {
    $line = $_;
    chomp $line;
    ($name,$passwd,$email) = split /:/, $line;

    # Skip null entries, however they got there.
    unless ($name) {
       inform "Deleting null entry" if $verbose;
       next;
    }
    
    # We can't use lookup_email_from_session, because we have no session.
    # We have to duplicate a little of the User.pm code to get around that.
    my $dir = IU::Session->persistent_dir($name);
    my $original_dir = $dir;
    my $new_email = "";

    # Handle links (for renamed accounts).
    if (-l $dir) {
      # chase down to the far end to see what's there.
      # The first link is the name that has to be in the password file.
      # The last link is the one we need to check out.
      my @link_chain = ();
	while (-l $dir) {
	  push @link_chain, $dir;
	  $dir = readlink $dir;
	}
      push @link_chain,$dir; 
      
      # If the persistent users have been moved, the links may be wrong.
      # Fix them if you can.
      my $test_dir = IU::Session->persistent_dir((split /\//,$dir)[-1]);
      if ($test_dir ne $dir) {   # not a good local link
	if (-e $test_dir && -d $test_dir) {    # but it's there
	  warn "Account for $name mislinked to $dir; fixed\n";
	  unlink $link_chain[-2];
          symlink $test_dir, $link_chain[-2];
	  $dir = $test_dir;
	}
	else {                                 # Not there at all
	  warn "Dangling symlink for $name; link and entry deleted\n";
	  unlink $original_dir;
	  next;
	}
      }
    }
    if (-e $dir && -d _) {
      my $prefs_file = ($dir."/prefs.dat");
      if (-s $prefs_file) {
	# Open and read prefs file.
	if (open ODL, $prefs_file){
	  # Extract and parse ODL.
	  my $odl =  parse IU::ODL \*ODL;
	  close ODL;
	  if ($odl) {
	    # Find the request field.
	    my %fields = $odl->as_array;
	    if (exists $fields{REQUEST}) {
	      # Extract email address.
	      $new_email = IW::UserAddr->from_ODL($fields{REQUEST})->email;
              inform "Email " .
                     ($email eq $new_email ? "verified" : "updated") .
                     " for $name";
	    }
	  }
	}
      }
    }
    else {
      warn "No directory for $name, removing entry\n";
      next;
    }
    $email = $new_email;
    print TMP join(":",$name, $passwd, $email),"\n";
  }
  inform "done\n";

  $self->_update_done;
  
  # Let it come up again, if it should:
  unlink "$IU::LibDir/conf/down.txt" if $remove_down;
  
  1;
}

#-------------------------------------------------------------------------
#  lock 
#-------------------------------------------------------------------------
# Loosely based on example from the Perl Cookbook.
# Use directory as a lock to avoid a race condition for the lock.

sub lock {
  
  my $self = shift;
  my $passwdfile = $self->{passwdfile};
  my $lockfile = $passwdfile."_lock";
  my $timer = 10;
  
  # Try creating lock directory, if successfull then we got a lock,
  # otherwise try again after waiting 1 second - given up after 10 seconds.
  while (1) {
    last if mkdir($lockfile, 0775);
    puke "Can't get lock on $lockfile: $!" if (--$timer < 1) ;
    sleep 1;
  }
  
}

#-------------------------------------------------------------------------
# unlock
#-------------------------------------------------------------------------

sub unlock {
  my $self = shift;
  system ("rm -rf ".$self->{passwdfile}."_lock");
}

#-------------------------------------------------------------------------

sub _backup_password_file {
  my $passwdfile = shift;
  my @wwwuser = split(/\./, $IU::WwwUser);
  chmod 0664, "$passwdfile.old";
  unlink "$passwdfile.old";
  system "cp $passwdfile $passwdfile.old";
  chown_by_name(@wwwuser,"$passwdfile.old");
  chmod 0644,"$passwdfile.old";
}

#-------------------------------------------------------------------------

sub _rename_password_file {
  my $passwdfile = shift;
  my $tmp = $IU::PersistentDir .".htpasswdtmp";
  my @wwwuser = split(/\./, $IU::WwwUser);
  chown_by_name(@wwwuser, $passwdfile);
  chmod 0664, $passwdfile;
  system("mv $tmp $passwdfile"); 
  chown_by_name(@wwwuser, $passwdfile);
}

#-------------------------------------------------------------------------

sub _ready_to_copy {
  my $self = shift;
  my $passwdfile = $self->{passwdfile};
  
  _backup_password_file($passwdfile);
  
  # lock the password file;
  $self->lock;
  
  my $tmp = $IU::PersistentDir .".htpasswdtmp";
  
  open TMP, ">".$tmp or
    puke "couldn't open $tmp: $!" ;;
  
  open HTPASSWD, "$passwdfile" or
    puke "couldn't open $passwdfile: $!" ;
}

#-------------------------------------------------------------------------

sub _update_done {
  
  my $self = shift;
  
  close HTPASSWD;
  close TMP;
  
  # Finished updating TMP file, now rename tmp to .htpasswd.
  _rename_password_file($self->{passwdfile});
  
  # unlock the passwd file.
  $self->unlock;
}

#-------------------------------------------------------------------------

1;

