package IU::Util;

=head1 NAME

IU::Util - general-purpose utility functions

=cut

require Exporter;
@ISA = qw(Exporter);
@EXPORT_OK = qw(
                date_pad
		units2secs
		units2bytes
		path_filename
		recent_time
		nice_time
		safe_kill
		shellquote 
		touch
		which 
                chown_by_name
		);
use strict;

#------------------------------------------------------------
#
# Globals...
# 
#------------------------------------------------------------

# 3-letter month abbreviations, for prettifying dates and times:
my @MonthAbbrev = 
    ('', 
     'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
     'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'
     );

# Long month names, for prettifying dates and times:
my @MonthName = 
    ('', 
     'January', 'February', 'March', 
     'April',   'May',      'June',
     'July',    'August',   'September', 
     'October', 'November', 'December'
     );

#------------------------------------------------------------
# banner MESSAGE
#------------------------------------------------------------
# Output a banner to the STDOUT.
sub banner {
    my ($msg) = @_;
    print STDOUT "\n$msg\n\n";
}

#------------------------------------------------------------
# date_pad (\date_string, default_string)
#------------------------------------------------------------
# This will pad a date string using the supplied default.
# Thus, if a date string is 1994-03-21, and the default
# is 2000-01-01T10:30:17, then the padded string is:
# 1994-03-21T10:30:17.
# date_string must be a predefined date string but could be empty ''.
# Default string should be in one of two standard forms:
#   yyyy-mm-ddThh:mm:ss  or yyyy-mm-ddThh:mm:ssZ
# where the Z in the second implies GMT time zone (Z for "zulu").
# Returns 1 if success or 0 if used incorrectly.
# Author: Larry Rosen, Hughes STX Corp., June 1997
sub date_pad {
    # Extract passed variables.
    my ($date_string_ref, $default_string) = @_;
    # Check for valid date string reference.
    if (!defined($date_string_ref) || (ref($date_string_ref) ne 'SCALAR')) {
        return 0;
    }
    # Check for valid default date string.
    if (!defined($default_string) || (length($default_string) < 19)) {
        return 0;
    }
    # C Format for time fields is:  "%d-%d-%dT%d:%d:%dZ" (Z is optional)
    my @date_fields;
    my $num_fields;
    my $z='';                        # 'Z' if date has Z in it.
    @date_fields = $$date_string_ref =~ /(\d+)/g;
    $num_fields = scalar(@date_fields);
    if (($$date_string_ref =~ /Z/i) || ($default_string =~ /Z/i)) {
        $z = 'Z';
    } else {
        $z = '';
    }
    if ($num_fields < 1) {
        $$date_string_ref = $default_string;
    } elsif ($num_fields < 2) {
        $$date_string_ref = $date_fields[0] . substr($default_string, 4);
    } elsif ($num_fields < 3) {
        $$date_string_ref = $date_fields[0] . '-' . $date_fields[1] . 
            substr($default_string, 7);
    } elsif ($num_fields < 4) {
        $$date_string_ref = $date_fields[0] . '-' . $date_fields[1] . '-' .
            $date_fields[2] . substr($default_string, 10);
    } elsif ($num_fields < 5) {
        $$date_string_ref = $date_fields[0] . '-' . $date_fields[1] . '-' .
        $date_fields[2] . 'T' . $date_fields[3] . substr($default_string, 13);
    } elsif ($num_fields < 6) {
        $$date_string_ref = $date_fields[0] . '-' . $date_fields[1] . '-' .
        $date_fields[2] . 'T' . $date_fields[3] . ':' . $date_fields[4] .
        substr($default_string, 16);
    } elsif ($num_fields < 7) {
        $$date_string_ref = $date_fields[0] . '-' . $date_fields[1] . '-' .
        $date_fields[2] . 'T' . $date_fields[3] . ':' . $date_fields[4] .
        ':' . $date_fields[5];
    }
    $$date_string_ref .= $z;
    1;
}

#------------------------------------------------------------
# mixed_to_caps STRING
#------------------------------------------------------------
# Given "MixedCase", return "MIXED_CASE".
sub mixed_to_caps {
    my ($str) = @_;

    $str =~ s/([A-Z])/_$1/g;
    $str =~ tr/a-z/A-Z/;    
    return substr($str, 1);
}

#------------------------------------------------------------
# month_abbrev MONTHNO
#------------------------------------------------------------
# Return the 3-letter month abbreviation.
sub month_abbrev {
    my $no = shift;
    $MonthAbbrev[$no];
}

#------------------------------------------------------------
# month_name MONTHNO
#------------------------------------------------------------
# Return the long month name.
sub month_name {
    my $no = shift;
    $MonthName[$no];
}

#------------------------------------------------------------
# recent_time EPOCHTIME
#------------------------------------------------------------
# Format a recent time into a nice-looking localtime.  
# Good for logging, etc.

sub recent_time {
    my $time = shift;

    my $real = (($time =~ /^\d+$/) ? $time : 0);
    my $abs = ($real ? localtime($real) : 'unknown');
    $abs =~ s/\s+\d+$//;        # get rid of trailing year

    my $secs_ago = ($real ? (time - $real) : 0);
    my $mins_ago = int($secs_ago / 60);
    my $hrs_ago  = int($mins_ago / 60);
    my $hr_mins_ago = ($mins_ago - (60 * $hrs_ago));
    my $hr_secs_ago = ($secs_ago - (60 * $mins_ago));
    
    "$abs (${hrs_ago}h ${hr_mins_ago}m ${hr_secs_ago}s ago)";
}

#------------------------------------------------------------
# nice_time SECONDS
#------------------------------------------------------------
# Format a time into a nice-looking set of days, hours, minutes, seconds.
# Used for informing time diffences in other than seconds.
# Input: $time - scalar - a time in seconds.
# If input is not composed of strictly digits, nice_time returns undef!

sub nice_time {
    my $time = shift;

    if (!$time) {
	return "0 seconds";
    }
    # Make sure digits or set to 0.
    my $seconds = (($time =~ /^\d+$/) ? $time : 0);

    my $days = int ($seconds / 86400);
    $seconds -= $days * 86400;
    my $hours = int ($seconds / 3600);
    $seconds -= $hours * 3600;
    my $minutes = int ($seconds / 60);
    $seconds -= $minutes * 60;
    my $string = '';
    $string .= "$days day"       . ($days    == 1 ? " " : "s ") if ($days);
    $string .= "$hours hour"     . ($hours   == 1 ? " " : "s ") if ($hours);
    $string .= "$minutes minute" . ($minutes == 1 ? " " : "s ") if ($minutes);
    $string .= "$seconds second" . ($seconds == 1 ? " " : "s ") if ($seconds);
    return ($string);
}

#------------------------------------------------------------
# safe_kill SIG, PID
#------------------------------------------------------------
# Like kill, but stops you from specifying empty, zero, or negative pids.
# Also stops you from killing yourself or your parent by number: 
# specify a PID of 'SELF' or 'PARENT' to kill those explicitly. 
#
# Illegal kill attempts result in fatal exceptions, since they constitute
# coding errors.

sub safe_kill {
    my ($sig, $pid) = @_;

    # remove any whitespace.
    $pid =~ s/\s//g;

    # Trap for accidents:
    die "safe_kill: empty/zero/negative pid ($pid)"  if (!$pid || ($pid < 0));
    die "safe_kill: illegal to kill self by number"   if ($pid == $$);
    die "safe_kill: illegal to kill parent by number" if ($pid == getppid());

    # remove any whitespace.
    $pid =~ s/\s//g; 

    # Map names to actual pids:
    if    ($pid eq 'SELF')   { $pid = $$ }
    elsif ($pid eq 'PARENT') { $pid = getppid() }
    elsif ($pid =~ /^\d+$/ ) {1}  

    else { die "safe_kill: unknown process name ($pid)" }

    # Do it!
    kill $sig, $pid;
}

#------------------------------------------------------------
# units2secs UNITSTR
#------------------------------------------------------------
# Convert a "unit'ed time string" to seconds.  Such a time string
# is of the form: "2s", "4m", "6h", "8d".  Case does not matter.
#
# No units assumes secs, so units2secs(units2secs($x)) = units2secs($x);

sub units2secs {
    my $units = shift;
    my %mult = ('s'=>1, 'm'=>60, 'h'=>(60*60), 'd'=>(60*60*24));
    $units =~ /^(-?\d+)\s*([smhd]?)/i or die "bad unit-time string: '$units";
    $1 * $mult{$2 ? lc($2) : 's'};
}

#------------------------------------------------------------
# units2bytes UNITSTR
#------------------------------------------------------------
# Convert a "unit'ed byte count" to bytes.  Such a byte count 
# is of the form: "3b", "6k", "6kb", "9m", "9mb", "2g", or "4gb".
# Case does not matter. All K's are 1024.
#
# No units assumes bytes, so units2bytes(units2bytes($x)) = units2bytes($x);

sub units2bytes {
    my $units = shift;
    my %mult = ('b'=>1,
		'k'=>1024, 'kb'=>1024,
		'm'=>(1024*1024), 'mb'=>(1024*1024),
		'g'=>(1024*1024*1024), 'gb'=>(1024*1024*1024));
    $units =~ /^(-?\d+)\s*([bkmg]+)/i or die "bad unit-byte string: '$units";
    $1 * $mult{$2 ? lc($2) : 'b'};
}

#------------------------------------------------------------
# path_filename PATH
#------------------------------------------------------------
# Extract filename from path.
sub path_filename {
    my $path = shift;
    $path =~ m|([^/]+)$|;
    $1;
}

#------------------------------------------------------------
# shellquote STRING
#------------------------------------------------------------
# Sometimes we need to put strings into open() pipe calls,
# but we don't know whether they're safe for /bin/sh or not.
# (for example, the string:
#
#        hi ; > $there
#
# is most definitely NOT safe.  This routine clothes its argument 
# in double quotes, and escapes any problematic characters.  Thus, 
# it would return the above as:
#
#        "hi ; > \$there"
#
# INCLUDING the double quotes.

sub shellquote {
    my $str = shift;

    # Protect $, `, and " from translation, and fold in double-quotes:
    $str =~ s/\$/\\\$/g;
    $str =~ s/\`/\\`/g;
    $str =~ s/\"/\\"/g;
    return "\"$str\"";
}

#------------------------------------------------------------
# timeparts
#------------------------------------------------------------
# Return a hash that encodes the localtime/gmtime
sub timeparts {
    #($_[5] >= 100) or $_[5] += (($_[5] < 50) ? 2000 : 1900);   # year
    (Yr => $_[5] + 1900, 
     Mon => $_[4] + 1, 
     Day => $_[3], 
     Hr  => $_[2], 
     Min => $_[1], 
     Sec => $_[0]);
}

#------------------------------------------------------------
# touch PATH [,OPTS]
#------------------------------------------------------------
# Act like the "touch" program.
# If OPTS contains '-a', only the access time is changed.
# If OPTS contains '-m', only the modification time is changed.

sub touch {
    my $path = shift;
    my $opts = shift || '';
    system("touch $opts " . shellquote($path));
    (($? >> 8) == 0);
}

#------------------------------------------------------------
# which PROGNAME
#------------------------------------------------------------
# Callable version of the "which" program.
sub which {
    my ($progname) = @_;
    my (@path, $dir);

    # Try to find it in path:
    @path = split(/:/, $ENV{'PATH'});
    foreach $dir (@path) {
	if (-x "$dir/$progname") {
	    return "$dir/$progname";
	}
    }
    return undef;
}

#------------------------------------------------------------
# chown_by_name owner, group, file
#------------------------------------------------------------
# change ownership/group of a file by name instead of UID/GID.
#------------------------------------------------------------
sub chown_by_name {
   my ($user, $group, $file) = @_;
   chown( (getpwnam($user))[2], (getgrnam($group))[2], glob($file));
}
1;

