package IU::Vec;

=head1 NAME

IU::Vec - XYZ vector, for map operations
 
=cut

#------------------------------------------------------------
# Vec -- an XYZ vector, for map operations
#------------------------------------------------------------
#
# INSTANCE VARIABLES
#    The object is an array, whose elements are [x, y, z].


#------------------------------------------------------------
#
# Globals
#
#------------------------------------------------------------

$Pi = 3.141592654;    # Pi (of course)


#------------------------------------------------------------
# New -- create a new vector object
#------------------------------------------------------------

sub New {
    my $type = shift;
    my @new;
    if (@_) {
	@new = ($_[0] + 0.0, $_[1] + 0.0, $_[2] + 0.0);
    }
    else {
	@new = (0.0, 0.0, 0.0);
    }
    my $new = \@new;
    bless $new, $type;
    $new;
}

#------------------------------------------------------------
# copy -- return a copy
#------------------------------------------------------------

sub copy {
    my $self = shift;
    IU::Vec->New(@$self);
}

#------------------------------------------------------------
# thetaphi -- set from theta and phi coords (in radians), and return self
#------------------------------------------------------------

sub thetaphi {
    my $self = shift;
    my ($theta, $phi) = @_;
    $self->x(cos($theta));
    $self->y(sin($theta));
    $self->z(sin($phi));
    $self;
}

#------------------------------------------------------------
# lonlat -- set from lon and lat coords (in degrees), and return self
#------------------------------------------------------------

sub lonlat {
    my $self = shift;
    my ($lon, $lat) = @_;
    my $theta = ($lon * $Pi) / 180.0;
    my $phi   = ($lat * $Pi) / 180.0;
    $self->thetaphi($theta, $phi);
}

#------------------------------------------------------------
# x -- set or get x value
#------------------------------------------------------------

sub x {
    my $self = shift;
    my $val = shift;
    (defined($val) ? ($self->[0] = $val) : $self->[0]);
}

#------------------------------------------------------------
# y -- set or get y value
#------------------------------------------------------------

sub y {
    my $self = shift;
    my $val = shift;
    (defined($val) ? ($self->[1] = $val) : $self->[1]);
}

#------------------------------------------------------------
# z -- set or get z value
#------------------------------------------------------------

sub z {
    my $self = shift;
    my $val = shift;
    (defined($val) ? ($self->[2] = $val) : $self->[2]);
}

#------------------------------------------------------------
# theta -- get theta value, -pi...+pi
#------------------------------------------------------------

sub theta {
    my $self = shift;
    atan2($self->y, $self->x);
}

#------------------------------------------------------------
# phi -- get phi value
#------------------------------------------------------------

sub phi {
    my $self = shift;
    my $proj = sqrt(($self->x * $self->x) + ($self->y * $self->y));
    atan2($self->z, $proj);
}

#------------------------------------------------------------
# lon -- get lon value, -180...+180
#------------------------------------------------------------

sub lon {
    my $self = shift;
    ($self->theta * 180) / $Pi;
}

#------------------------------------------------------------
# lat -- get lat value
#------------------------------------------------------------

sub lat {
    my $self = shift;
    ($self->phi * 180) / $Pi;
}

#------------------------------------------------------------
# Ry -- rotate around Y axis by $theta, return new object
#------------------------------------------------------------

sub Ry {
    my $self = shift;
    my $theta = shift;
    my $r = IU::Vec->New;
    $r->x(($self->x * cos($theta)) + ($self->z * -sin($theta))); 
    $r->y($self->y);
    $r->z(($self->x * sin($theta)) + ($self->z * cos($theta))); 
    $r;
}

#------------------------------------------------------------
# Rz -- rotate around Z axis by $theta, return new object
#------------------------------------------------------------

sub Rz {
    my $self = shift;
    my $theta = shift;
    my $r = IU::Vec->New;
    $r->x(($self->x * cos($theta))  + ($self->y * sin($theta))); 
    $r->y(($self->x * -sin($theta)) + ($self->y * cos($theta))); 
    $r->z($self->z);
    $r;
}


#------------------------------------------------------------
1;
