package IU::WWWValids;

=head1 NAME

IU::WWWValids - EDG specific valids information. 

=head1 DESCRIPTION

This class encapsulates the access to the wwwvalids.conf file which
is part of the Gateway's support files.  This replaces direct access
to the file with a standard interface. The wwwvalids.conf file is in
ODL format. Whenever new keywords and groups are added new routines
to provide access to the information should be added to this module.

=head1 SPEED HACKS

Because the parsing of ODL is so awfully slow, we have a utility in
the installation's lib/perl called C<compwwwvalids>. This utility
does a C<compile()> method-call and then stores the output (using Perl's
C<Storable> module) in the C<lib/conf> directory. 

The C<new()> method can then simply reload this stored copy, at a savings
of more than 99% of the CPU time (Storable's deserialization of the saved
object runs approximately 66 times faster than the IU::WWWValids ODL parsing).

=head2 USAGE

# Build a new wwwvalids if we don't have one yet.
$WWWVALIDS = IU::WWWValids->new if (!defined($WWWVALIDS));

=head2 METHOD INTERFACE

=head1 AUTHOR

Ed Seufert - January 27, 2000
Most of the code has been copied from the mkvalids script written by Eryq.

=head1 VERSION

$Id: WWWValids.pm,v 1.32 2003/06/16 19:01:26 lorena Exp $

=cut

#=====================================================================

use IU;
use IU::ODL ();
use IU::Errors;
use IU::Log;
use IW::DaacAddress;
use IU::MediaMaximumCapacity;

use strict;
use Storable;

use vars qw(@ISA @EXPORT
            $WWWVALIDS
            );
@ISA = qw(Exporter);
@EXPORT = qw($WWWVALIDS);

#---------
# Globals
#--------

$WWWVALIDS = undef;

#---------------------------------------------------------------------
# compile
#---------------------------------------------------------------------
# Constructor. Loads the wwwvalids ODL, parses it, and creates an object.

sub compile {
   my $class = shift;                   # Get class.
   my $self = {};                       # Start with empty hash.
   bless $self, $class;                 # Create the object.
   debug "Compiling wwwvalids";

   # for backward compatability with old wwwvalids.conf, set some
   # fallback values:

   my @default_gran = (
              "LocalGranuleID",
              "MISC_URL",
              "START_DATE",
              "STOP_DATE",
              "GEO_CENTER",
              "PATH",
              "ROW",
              "QA_SCENE_CCA",
              "QA_FLAG",
              "DATASET_ID"
             );
   my @all_gran = (
              "GEO_COVERAGE",
              "CAMPAIGN",
              "GEO_CENTER",
              "QA_SCENE_CCA",
              "DATA_CENTER_ID",
              "DATASET_ID",
              "LocalGranuleID",
              "PARAMETER",
              "ORIG_PARTIAL_WRS_SCENES",
              "PROCESSING_LEVEL",
              "QA_FLAG",
              "SENSOR_NAMES",
              "SIZEMBECSDATAGRANULE",
              "SOURCE_NAMES",
              "MISC_URL",
              "START_DATE",
              "STOP_DATE",
              "ORIG_TOTAL_WRS_SCENES",
              "SP_ENDING_ROW",
              "PATH",
              "ROW",
              "SP_STARTING_ROW"
              );
   my @pretty_gran = (
              "Area Coverage",
              "Campaign",
              "Center Point",
              "Cloud Cover",
              "Data Center",
              "Data Set",
              "Local Id",
              "Parameters",
              "Partial WRS Scenes",
              "Processing Level",
              "Quality Flag",
              "Sensor",
              "Size (MB)",
              "Source",
              "Special Processing Links",
              "Start Date",
              "Stop Date",
              "Total WRS Scenes",
              "WRS Ending Row",
              "WRS Path",
              "WRS Row",
              "WRS Starting Row"
              ); 
   my @sort_gran = (
              "GRANULE_ID",
              "START_DATE",
              "NOTHING"
             );

   $self->{"LIST:GRANULE"} = {DEFAULT=>\@default_gran,
                            ALL=>\@all_gran,
                            PRETTY=>\@pretty_gran,
                            SORT=>\@sort_gran };

   my @default_dset = (
              "NUM_GRANULES",
              "SOURCE_NAMES",
              "SENSOR_NAMES"
             );
   my @all_dset = (
              "SOURCE_NAMES",
              "NUM_GRANULES",
              "DATA_CENTER_ID",
              "MISC_URL",
              "START_DATE",
              "STOP_DATE",
              "PARAMETER",
              "CAMPAIGN",
              "PROCESSING_LEVEL",
              "SENSOR_NAMES"
              );
   my @pretty_dset = (
              "Sources",
              "Number of Granules",
              "Data Center",
              "Special Processing Links",
              "Earliest Returned Granule",
              "Latest Returned Granule",
              "Parameters",
              "Campaigns",
              "Processing Levels",
              "Sensors"
              ); 
   my @sort_dset = (
              "DATASET_ID",
              "DATA_CENTER_ID",
              "NOTHING"
              );

   $self->{"LIST:DATASET"} = {DEFAULT=>\@default_dset,
                            ALL=>\@all_dset,
                            PRETTY=>\@pretty_dset,
                            SORT=>\@sort_dset };


	   # Read the wwwvalids.conf from an ODL file:
   open CONF, "${IU::GaeaDataDir}/wwwvalids.conf" 
       or die "open ${IU::GaeaDataDir}/wwwvalids.conf:: $!";

   my $odl;
   # Parse:
   while ($odl = IU::ODL->parse_one(\*CONF)) {
       $self->digest($odl);
   }
   close CONF;
   debug "Compile complete";
   return ($self);                      # Return object.
}

#------------------------------------------------------------
# new
#------------------------------------------------------------
# Singleton constructor. 
#
# Returns either a copy of the current WWWValids object, or,
# if there is none, reloads it from wwwvalids.stored. If this 
# fails, it runs compile() to create the object, but outputs
# a warning to the debug log.

sub new {
  my $class = shift;
  my $CONFIG_DIRECTORY    = "$IU::LibDir/conf";
  my $input_file          = "$CONFIG_DIRECTORY/wwwvalids.stored";

  # If there's already a $WWWVALIDS, we'll skip this.
  unless (defined $WWWVALIDS and ref($WWWVALIDS) eq $class) {
    # No wwwvalids, try reloading it.
    debug "Reloading wwwvalids";
    $WWWVALIDS = Storable::retrieve($input_file);
    unless (defined $WWWVALIDS) {
      $WWWVALIDS = $class->compile();
      warn "Forced to compile WWWValids - was compwwwvalids not run?";
    }
  }
  else {
    debug "Reusing singleton";
  }
  bless $WWWVALIDS, $class;
  $WWWVALIDS;
}
  
#------------------------------------------------------------
# digest
#------------------------------------------------------------
sub digest {
    my $self = shift; 
    my $odl = shift;

    $odl->rewind;
    (my ($param, $value) = $odl->next);
    if ($param eq 'DATA_CENTER') {
       $self->digest_daac($value);
    }
    elsif ($param eq 'DATASET') {
       $self->digest_dataset($value);
    }
    elsif ($param eq 'DISCLAIMER') {
       $self->digest_disclaimer($value);
    }
    elsif ($param eq 'MAP_LIMITS') {
       $self->digest_maplimits($value);
    }
    elsif ($param eq 'RESULTS_LIST_COLUMNS') {
       $self->digest_list($value);
    }
    elsif ($param eq 'STATS_LEVEL') {
       $self->digest_stats($value);
    }
    elsif ($param eq 'DISCIPLINE') {
       $self->digest_discipline($value);
    }

}

#------------------------------------------------------------
# digest_daac
#------------------------------------------------------------
sub digest_daac {
    my $self = shift;
    my $odl = shift;
    my %boolean;

    my $archid = $odl->param('DATA_CENTER_ID')->text;

    # Boolean flags:
    my ($info, $name, $default);
    foreach $info (['ALLOWS_EMAIL',          0],
                   ['ALLOWS_GRANULE_ID',     0],
                   ['REQUIRES_ECS_AUTH',     0],
                   ['REQUIRES_V0_AUTH',      0],
                   ['REQUIRES_NASDA_AUTH',   0],
                   ['V0_DAAC',               0],
                   ['ALLOWS_ORDER_STATUS',   0],
                   ['ALLOWS_ACCOUNT_STATUS', 0],
                   ['ALLOWS_EMAIL_STATUS',   0],
                   ['ALLOWS_MULTIPLE_MEDIA', 0]) {
        my ($name, $default) = @$info;
        my $val = $default;
        my $val_odl;
        if ($val_odl = $odl->param($name)) {     # given
            $val = (uc($val_odl->text) eq 'Y') ? 1 : 0;
            if ( ($name eq 'ALLOWS_GRANULE_ID') && ($val == 1) ) {
               $self->daac_array($archid);
            }
            if ( ($name eq 'ALLOWS_ACCOUNT_STATUS') && ($val == 1) ) {
               $self->account_array($archid);
            }
            if ( ($name eq 'ALLOWS_ORDER_STATUS') && ($val == 1) ) {
               $self->order_status_array($archid);
            }
            if ( ($name eq 'ALLOWS_EMAIL_STATUS') && ($val == 1) ) {
               $self->email_status_array($archid);
            }
            if ( ($name eq 'V0_DAAC') && ($val == 1) ) {
               $self->v0_daacs_array($archid);
            }
        }
        $boolean{$name} = $val;     # 1, 0, or default
    }

    # Contact address:
    my $addr =
       IW::DaacAddress->from_ODL($odl->param('CONTACT_ADDRESS'));
    my $status_addr =
     IW::DaacAddress->from_ODL($odl->param('STATUS_CONTACT_ADDRESS'));
    my $security_settings = parse_security_group($odl->param('SECURITY'));
    my $capacities = $self->parse_capacity_group($odl,$archid);

    $self->{"daac:$archid"} = {
        Allows => $boolean{ALLOWS_EMAIL},  # backward compat
        Addr   => $addr,
        StatusAddr   => $status_addr,
        Security   => $security_settings,
        Capacities => $capacities,
        %boolean
        };
}

#-----------------------------------------------------------------------------
# daac_contact_addr archid
#-----------------------------------------------------------------------------
# Return data center contact object for requested archive id.

sub daac_contact_addr {
   my $self = shift;
   my $archid = shift;

   return $self->{"daac:$archid"}{StatusAddr};
}
#-----------------------------------------------------------------------------
# daac_array.
# store all daacs which allow GRANULE_ID in an array.
#-----------------------------------------------------------------------------

sub daac_array {
   my $self = shift;
   my $value = shift;
   my @array = ();

   my $arrayref = $self->{DaacArray};
   @array = @$arrayref if($arrayref);

  if ($value) {
     @array = (@array, $value);
   }
  $self->{DaacArray} = \@array;
  return \@array;
}

#-----------------------------------------------------------------------------
# account_array.
# store all daacs which allow ACCOUNT_STATUS in an array.
#-----------------------------------------------------------------------------

sub account_array {
   my $self = shift;
   my $value = shift;
   my @array = ();

   my $arrayref = $self->{AccountArray};
   @array = @$arrayref if($arrayref);

  if ($value) {
     @array = (@array, $value);
   }
  $self->{AccountArray} = \@array;
  return \@array;
}

#-----------------------------------------------------------------------------
# order_status_array.
# store all daacs which allow ORDER_STATUS in an array.
#-----------------------------------------------------------------------------

sub order_status_array {
   my $self = shift;
   my $value = shift;
   my @array = ();

   my $arrayref = $self->{OrderStatusArray};
   @array = @$arrayref if($arrayref);

  if ($value) {
     @array = (@array, $value);
   }
  $self->{OrderStatusArray} = \@array;
  return \@array;
}

#-----------------------------------------------------------------------------
# email_status_array.
# store all daacs which allow EMAIL_STATUS in an array.
#-----------------------------------------------------------------------------

sub email_status_array {
   my $self = shift;
   my $value = shift;
   my @array = ();

   my $arrayref = $self->{EmailStatusArray};
   @array = @$arrayref if($arrayref);

  if ($value) {
     @array = (@array, $value);
   }
  $self->{EmailStatusArray} = \@array;
  return \@array;
}


#-----------------------------------------------------------------------------
# v0_daacs_array.
# store all daacs which are V0_DAACs in an array.
#-----------------------------------------------------------------------------

sub v0_daacs_array {
   my $self = shift;
   my $value = shift;
   my @array = ();

   my $arrayref = $self->{V0DaacsArray};
   @array = @$arrayref if($arrayref);

   if ($value) {
      @array = (@array, $value);
   }
   $self->{V0DaacsArray} = \@array;
   return (wantarray ? @array : \@array);
}


#------------------------------------------------------------
# digest_disclaimer
#------------------------------------------------------------
sub digest_disclaimer {
    my $self = shift;
    my $odl = shift;

    my $source = ( $odl->param('SOURCE') ? $odl->param('SOURCE')->text : undef);
    my $sensor = ( $odl->param('SENSOR') ? $odl->param('SENSOR')->text : undef);
    my $text   = $odl->param('TEXT')->text;
    if (defined($source)) {
       $self->{"disclaimer:$source"} = { Text=>$text };
    } 
    elsif (defined($sensor)) {
       $self->{"disclaimer:$sensor"} = { Text=>$text };
    }
}

#------------------------------------------------------------
# digest_list
#------------------------------------------------------------
sub digest_list {
    my $self = shift;
    my $odl = shift;

    my $type = $odl->param('TYPE')->text;
    my @default = $odl->param('DEFAULT')->text_values;
    my @all     = $odl->param('ALL')->text_values;
    my @sort    = $odl->param('SORT')->text_values;
    my @pretty  = $odl->param('PRETTY')->text_values;
    $self->{"LIST:$type"} = {DEFAULT=>\@default,
                             ALL=>\@all,
                             PRETTY=>\@pretty,
                             SORT=>\@sort };

}

#------------------------------------------------------------
# digest_discipline
#------------------------------------------------------------
sub digest_discipline {
    my $self = shift;
    my $odl = shift;

    my $heading  = $odl->param('HEADING')->text;
    my $maxrow    = $odl->param('MAXROW')->text if $odl->param('MAXROW');
    my @missions = $odl->param('MISSIONS')->text_values;
    my @pretty   = $odl->param('PRETTY')->text_values;
    $self->{DISCIPLINES}{$heading}{MISSIONS} = \@missions;
    $self->{DISCIPLINES}{$heading}{PRETTY} = \@pretty;
    $self->{DISCIPLINES}{$heading}{MAXROW} = $maxrow;
}

#------------------------------------------------------------
# digest_stats
#------------------------------------------------------------
sub digest_stats {
    my $self = shift;
    my $odl = shift;

    my $type = $odl->param('TYPE')->text;
    my @records = $odl->param('DONT_RECORD')->text_values;
    $self->{uc("STATS:$type")} = \@records;
}


#------------------------------------------------------------
# digest_dataset
#------------------------------------------------------------
sub digest_dataset {
    my $self = shift;
    my $odl = shift;
    my %boolean;

    my $archid = $odl->param('DATA_CENTER_ID')->text;
    my $dsid   = $odl->param('DATASET_ID')->text;

    # Boolean flags:
    my ($info, $name, $default);
    foreach $info (['ALLOWS_EMAIL',          0],
                   ['ALLOWS_GRANULE_ID',     0],
                   ['ALLOWS_MULTIPLE_MEDIA', 0]) {
        my ($name, $default) = @$info;
        my $val = $default;
        my $val_odl;
        if ($val_odl = $odl->param($name)) {     # given
            $val = (uc($val_odl->text) eq 'Y') ? 1 : 0;
        }
        $boolean{$name} = $val;     # 1, 0, or default
    }

    # Contact address:
    my $addr = IW::DaacAddress->from_ODL($odl->param('CONTACT_ADDRESS'));

    $self->{"dset:$archid:$dsid"} = { Allows=>$boolean{ALLOWS_EMAIL}, 
                                      Addr=>$addr,
                                      %boolean, };

}

sub digest_maplimits {
    my $self = shift;
    my $odl = shift;

    my $swath_name = $odl->param('SWATH_NAME')->text;
    debug "swath_name = $swath_name"; 
    my $path_min = $odl->param('PATH_MIN')->text;
    my $path_max   = $odl->param('PATH_MAX')->text;
    my $block_max   = $odl->param('BLOCK_MAX')->text;
    my $block_min   = $odl->param('BLOCK_MIN')->text;

    $self->{"Maplimits:$swath_name"} = {PathMin=>$path_min,
                             PathMax=>$path_max,
                             BlockMin=>$block_min,
                             BlockMax=>$block_max };
}

#-------------------------------------------------------
# parse_security_group
#-------------------------------------------------------
# duplicate code of from_ODL for the purpose to handle a new
# subgroup SECURITY
# Didn't see the need to create a seperate package since the SECURITY
# group is limited to wwwvalids.conf JSA

sub parse_security_group {
    my $odl = shift || return undef;
    my ($param, $value);
    my $self = bless {};

    my %SimpleVar = map {($_ => 1)}
       qw(SERVER_ADDR SERVER_PORT NON_DEFAULT_PORT LOGIN_AS WWWUSER_HOME);
    # Read the ODL stream:
    $odl->rewind;
    while (($param, $value) = $odl->next) {
        if (($value->type eq 'SCALAR') && $SimpleVar{$param}) {
            $self->{$param} = $value->text;
        }
    }

    # Return it:
    $self;
}

#-------------------------------------------------------
# parse_capacity_group ODL, DAAC
#-------------------------------------------------------
# Wrapper code to parse the CAPACITY ODL group. This group
# is simply a wrapper for one or more MEDIA_MAXIMUM_CAPACITY
# groups. The ODL should look something like this:
#
# GROUP             = CAPACITIES
#  GROUP               = MEDIA_MAXIMUM_CAPACITY
#      MEDIA_MB_CAPACITY = 500
#      TYPE_ID           = "9TRK"
#      COMMENT           = "Media Maximum Capacity"      
#      GROUP         = MISC_URL
#        URL         = "http://edc-ecs/mediacapacitypolicy.html"
#        URL_COMMENT = "Media Maximum Capacity Policy"
#      END_GROUP     = MISC_URL
#    END_GROUP           = MEDIA_MAXIMUM_CAPACITY
# END_GROUP         = CAPACITIES
#
# The wrapper is used to keep the parsing code in MediaMaximumCapacity
# simpler; this way it doesn't have to turn handsprings and act completely
# different from any other from_ODL() method.
sub parse_capacity_group {
    my $self = shift;
    my $odl = shift || return undef;
    my $daac = shift || return undef;
    my ($param, $value);
    my $entries = [];
    $odl->rewind;
    while (($param, $value) = $odl->next) {
        if ($param eq 'MEDIA_MAXIMUM_CAPACITY') {
            my $capacity = IU::MediaMaximumCapacity->from_ODL($value,$daac);
            local $_;
            push @$entries, $capacity;
        }
    }
    if (defined $self->{Capacities}) {
       push @{$self->{Capacities}}, @$entries;
    }
    else {
       $self->{Capacities} = [@$entries];
    }
    $entries;
}

#-----------------------------------------------------------------------------
#  Accessor methods
#-----------------------------------------------------------------------------

#---------------------------
# getDisciplineHeadings
#---------------------------

sub getDisciplineHeadings {

   my $self = shift;
   my @headings = sort keys %{$self->{DISCIPLINES}}; 
   return \@headings;
}

#---------------------------
# SCALAR getDisciplineMaxrow HEADING
#---------------------------

sub getDisciplineMaxrow {

   my $self = shift;
   my $heading = shift;
   return $self->{DISCIPLINES}{$heading}{MAXROW};
}


#---------------------------
# getDisciplineMissions HEADING
#---------------------------

sub getDisciplineMissions {

   my $self = shift;
   my $heading = shift;
   return $self->{DISCIPLINES}{$heading}{MISSIONS};
}

#---------------------------
# getDisciplinePretty HEADING
#---------------------------

sub getDisciplinePretty {

   my $self = shift;
   my $heading = shift;
   return $self->{DISCIPLINES}{$heading}{PRETTY};
}



#---------------------------
# granListDefaults ARRAY REF
#---------------------------

sub granListDefaults {

   my $self = shift; 
   $self->{'LIST:GRANULE'}->{'DEFAULT'};
}
#------------------------
# dsetListUgly
#------------------------

sub dsetListUgly {

   my $self = shift;
   my $prettyList = shift;
   my @uglyList;
   my @prettyAll = @{$self->{'LIST:DATASET'}->{'PRETTY'}};
   my @uglyAll = @{$self->{'LIST:DATASET'}->{'ALL'}};
   my $index=0;
   foreach my $prettyOne (@{$prettyList}) {
      my $allIndex = 0;
      foreach my $allPrettyOne (@prettyAll) {
         if ($prettyOne eq $allPrettyOne) {
           $uglyList[$index++] = $uglyAll[$allIndex];
         }
      $allIndex++;
      }
   }
   return @uglyList;
}

#------------------------
# granListUgly
#------------------------

sub granListUgly {

   my $self = shift;
   my $prettyList = shift;
   my @uglyList;
   my @prettyAll = @{$self->{'LIST:GRANULE'}->{'PRETTY'}};
   my @uglyAll = @{$self->{'LIST:GRANULE'}->{'ALL'}};
   my $index = 0;
   my $match = 0;
   my $allIndex = 0;
   foreach my $prettyOne (@{$prettyList}) {
      $allIndex = 0;
      $match = 0;
      foreach my $allPrettyOne (@prettyAll) {
         if ($prettyOne eq $allPrettyOne) {
           $uglyList[$index++] = $uglyAll[$allIndex];
           $match = 1;
         }
         $allIndex++;
      }
      $uglyList[$index++] = $prettyOne if (!$match);
   }
   return @uglyList;
}
#------------------------
# granListPretty HASH REF
#------------------------

sub granListPretty {

   my $self = shift;
   my %ugly2pretty;
   my $key;
   my $index = 0;
   my @pretty = @{$self->{'LIST:GRANULE'}->{'PRETTY'}};
   foreach $key (@{$self->{'LIST:GRANULE'}->{'ALL'}}) {
      $ugly2pretty{$key} = $pretty[$index];  
      $index++;
   }  
   return %ugly2pretty;
}


#--------------------
# granListAll
#--------------------

sub granListAll {

   my $self = shift;
   $self->{'LIST:GRANULE'}->{'ALL'};
}

#--------------------
# granSortDefaults
#--------------------

sub granSortDefaults {

   my $self = shift;
   $self->{'LIST:GRANULE'}->{'SORT'};
}

#--------------------
# dsetListDefault
#--------------------

sub dsetListDefaults {

   my $self = shift;
   $self->{'LIST:DATASET'}->{'DEFAULT'};
}

#--------------------
# dsetListAll
#--------------------

sub dsetListAll {

   my $self = shift;
   $self->{'LIST:DATASET'}->{'ALL'};
}

#--------------------
# dsetSortDefaults
#--------------------

sub dsetSortDefaults {

   my $self = shift;
   $self->{'LIST:DATASET'}->{'SORT'};
}

#------------------------
# dsetListPretty HASH REF
#------------------------

sub dsetListPretty {

   my $self = shift;
   my %ugly2pretty;
   my $key;
   my $index = 0;
   my @pretty = @{$self->{'LIST:DATASET'}->{'PRETTY'}};
   foreach $key (@{$self->{'LIST:DATASET'}->{'ALL'}}) {
      $ugly2pretty{$key} = $pretty[$index];
      $index++;
   }
   return %ugly2pretty;
}

#------------------------
# activeStatRecords LEVEL
#------------------------
sub activeStatRecords {
   my $self = shift;
   my $level = shift;
   my $records_list; 
   if (defined($records_list = $self->{uc("STATS:$level")})) {
     return @$records_list; 
   }
   else {
     return;
   }
}

#-----------------------
# capacities
#-----------------------
sub capacities {
   my $self = shift;
   defined $self->{Capacities}  
    ? (wantarray ? @{$self->{Capacities}} : $self->{Capacities})
    : return;
}

#-----------------------
# getMapLimits
#
# The limits placed on the MISR/GLAS/AMSRE Map data
#-----------------------
sub getMapLimits {
   my $self = shift;
   my $source = shift;
   my ($map_limit,$path_min, $path_max, $block_min, $block_max);
   if ($map_limit =  $self->{"Maplimits:$source"}) {
        $path_min = $map_limit->{'PathMin'};
        $path_max = $map_limit->{'PathMax'};
        $block_min = $map_limit->{'BlockMin'};
        $block_max = $map_limit->{'BlockMax'};
   }
   debug "getMapLimits $path_max,$block_min,$block_max, $path_min";
   return ($path_max,$block_min,$block_max, $path_min);
}

#---------------------------------------------------------------------
1;
