package IW::Address;

=head1 NAME

IW::Address - represents a general address

=head1 DESCRIPTION

This class represents a general address.  This is an abstract class;
no objects in the system are of type IW::Address.  Use child classes
for objects: IW::UserAddr or IW::DAACAddress.

The new address classes will have the following inheritance
structure:

IW::Address
    |
    +---- IW::DaacAddress
    |
    +---- IW::UserAddr

IW::Address contains generic routines (set/get for all shared fields,
etc.).  Subclasses will contain only new/changed fields and methods.

IW::DaacAddress replaces the IU::Addr and IW::Addr classes.  It has 2
additional fields not found in user address objects: CONTACT_NAME and
DAAC_ORDER_ID.  It doesn't have the from_CGI method that IW::UserAddr
does.

IW::UserAddr is modified to use the new hierarchy and design.
It will add the from_CGI method to the base IW::Address class.

There is a CommentsAddress object used in ModeCMNTFORM and perhaps
ModeFTPBROWSE.  But, since this object is just a UserAddr object with
no postal information, it can use the UserAddr class and just have
different required and print lists.  The CommentsAddress class
had a print_hidden_for_auth method, but this is no longer needed with
the use of required and print lists.

=head2 DATA

IW::Address will contain the common elements of user, daac, and
comments address objects.  It will have the get/set methods for these
elements, and other common methods.

=begin list

=item

%SimpleVar

This is a class variable (not an instance variable).  It is a hash of
keys in the Address object that are simple scalars.  The values are 1.
Here is the list:
      TITLE, FIRST_NAME, MIDDLE_INITIAL, LAST_NAME, ORGANIZATION,
      CITY, STATE, ZIP, COUNTRY, PHONE, FAX, EMAIL.

=item

%SequenceVar

This is a class variable (not an instance variable).  It is a hash of
keys in the Address object that are sequences.  The values are 1.
Here is the list:

     ADDRESS

=item

%MaxLength

This is a class variable (not an instance variable).  It is a hash
that stores the maximum acceptable lengths for the fields above, plus
the non-simple ADDRESS field and the maximum number of lines for the
ADDRESS.

=item

$self

This is the the usual hash reference.  It will contain keys listed
above in %SimpleVars and these additional ones:

=begin list



=item

ERRORS . . an IU::Errors object.  Input validation will store error 
strings in it.

=end list

=end list

=head2 METHODS

=begin list

=item

new ( ) . . . Constructor.

Create an empty hash reference $self.  Add an empty street address
array reference.  Add an empty IU::Errors object reference to the
$self.  Bless the $self hash reference into this class.

=item

_simple_vars ( )

This will return the list of keys of %SimpleVar. Used in from_chunk.

=item


=item

_sequence_vars ( )

This will return the list of keys of %SequenceVar. Used in from_chunk.

_is_simple ($FIELDNAME)

This returns true or false depending on whether the fieldname is in
the list of simple scalar fields, the keys of %SimpleVar.  from_ODL
will use this instead of a direct access to %SimpleVar so that
subclasses can check their list of simple vars first.

=item

_simple_ivar ($FIELDNAME, $MAX, $SELF, $NEWVALUE])

General get/set method for all simple variables above.  Internal
method; not called from outside.  If setting a variable (there is a
$NEWVALUE), then if value exceeds the length $MAX, instead of dying
set an error string in the Errors object, and returns the Errors
object.  $MAX is passed in so that subclasses can use this base class
method and pass in subclass data knowledge.

=item

_initial_ivar ($FIELDNAME, $MAX, $SELF, [$NEWVALUE])

Same as _simple_ivar except this is used for middle initial and
allows the setter to use a period as a second character.  Eg. "M."
will pass but "MP" will not.

=item

These 11 methods are externally available get/set methods
that call _simple_ivar:

title ([$NEWVALUE])
firstname ([$NEWVALUE])
lastname ([$NEWVALUE])
organization ([$NEWVALUE])
city ([$NEWVALUE])
state ([$NEWVALUE])
country ([$NEWVALUE])
zip ([$NEWVALUE])
phone ([$NEWVALUE])
fax ([$NEWVALUE])
email ([$NEWVALUE])

=item

initial ([$NEWVALUE]) . . . This get/set method for middle initial
calls _initial_ivar.

=end list

The following three methods are aliases of some above so that the
from_ODL and from_chunk methods can simply use a lowercase of the
parameter name to get the function name.  Example:
   $self->middle_initial($value);

=begin list

=item

first_name ([$NEWVALUE]) . . . Calls firstname([$NEWVALUE]).
last_name ([$NEWVALUE]) . . . Calls lastname([$NEWVALUE]).
middle_initial ([$NEWVALUE]) . . . Calls initial([$NEWVALUE]).

=item

name ( ) . . . Returns complete title and name as one string.

=item

street ([\@$NEWVALUE])

Get/set method for street address.  If setting and there are too many
lines (as defined in %MaxLength) or any line exceeds the length in
%MaxLength, instead of dying set an error string in Errors, and
return the Errors object.  Note that data is an array reference.

=item

as_string ( )

Method returns the address as a string with newlines.  It is useful
for debuging and for comparing two address objects.

=item

copy ( )

Create and returns a complete deep copy of this object.  Use it like
this:
   $new_addr = new IW::UserAddr;
   $new_addr->copy($old_addr);

=item

equal (IW::Address)

Check if two address objects have textually the same data.  Use it
like this:

   if ($addr1->equal($addr2)) { ...

=item

errors ( ) . . . Public method to access Errors object in the
UserAddr object.

=item

print_receipt_addr ( )

Prints address for order receipts.  Used for both UserAddr and
DaacaAddress objects.

=end list

=head3 ODL INTERFACE

=begin list

=item

from_ODL (IU::ODL)

Set the values in this address object (or create one) from a standard
IMS ODL object.  The ODL parameters have the same names as the
instance variables (list at top).  Some special handling of the
ADDRESS element.  Any errors cause strings to be set in Errors object.
If there were any errors, the User object should generate a form and
indicate which fields need to be fixed.

=item

to_ODL ( )

=end list

Examples:

   # Create address from ODL:
   $addr_odl = parse IU::ODL \*STDIN;
   $addr = IW::Address->from_ODL($addr_odl);

   # Convert address to ODL:
   $addr_odl = $addr->to_ODL;

=head3 CHUNK INTERFACE

=begin list

=item

from_chunk (IU::Chunk)

Creates a new Address object and fills it with values from the given
Chunk.  Any errors cause strings to be set in the Errors object.  If
there were any errors, the User object should generate a form and
indicate which fields need to be fixed.

=end list

=head3 CGI INTERFACE

Since DaacAddress doesn't have a CGI interface, the
from_CGI method will be in the UserAddr subclass.

=head3 HTML INTERFACE

Because print for DaacAddress is quite different from
print for UserAddr, these methods will be in the subclasses.

=head1 AUTHOR

Larry P. Rosen, EOS Data Gateway Team, Raytheon Systems Corp.
Based on old IW::UserAddr.pm by Eryq (Erik Dorfman).

=head1 RCS INFO

$Id: Address.pm,v 1.6 2002/08/07 13:35:59 seufert Exp $

=cut

#=====================================================================

use IU;                      # For puke() 
use IU::Errors;              # For Errors object.
use IU::Log;                 # For error()
use IU::IW;                  # For $HTML, $REQ, and $SESSION.
use Socket;                  # For inet_aton().
use strict;

# Mapping to check membership for SIMPLE (scalar) vars:
my %SimpleVar = map {($_ => 1)} 
   qw (TITLE FIRST_NAME MIDDLE_INITIAL LAST_NAME ORGANIZATION
       CITY STATE ZIP COUNTRY PHONE FAX EMAIL);

# Mapping to check membership for sequence vars:
my %SequenceVar = map {($_ => 1)}
   qw (ADDRESS);

# Valid maximum length for various things:
use vars qw ( %MaxLength );    # Make accessible in subclass.
%MaxLength = (TITLE          => 5,  FIRST_NAME     => 20,
              MIDDLE_INITIAL => 1,  LAST_NAME      => 20,
              ORGANIZATION   => 60, CITY           => 30,
              STATE          => 20, ZIP            => 15,
              COUNTRY        => 30, PHONE          => 22,
              FAX            => 22, EMAIL          => 128,      
              STREET         => 32, NUMSTREETLINES => 3,
             );

#---------------------------------------------------------------------
# new 
#---------------------------------------------------------------------
# Constructor class method.  Creates a new, empty address.
# Invoked when create user address:  $var = new IW::UserAddr;

sub new {
   my $class = shift;                   # Get class.
   my $self = {};                       # Start with empty hash.
   $self->{ADDRESS} = [];               # Add empty address array ref.
   $self->{ADD_PHONE} = [];             # Add empty additional phone array ref.
   $self->{COMMENT} = [];               # Add empty comment array ref.
   $self->{ERRORS} = new IU::Errors;    # Add empty errors object.
   bless $self, $class;                 # Create the object.
   return ($self);                      # Return object.
}

#---------------------------------------------------------------------
# _simple_vars
#---------------------------------------------------------------------
# Private class method to return the list of keys of %SimpleVar.
# Used in from_chunk().  Need a method because child IW::DaacAddress
# will have a _simple_vars() too and that calls this one.
# Input: none.
# Output: array of strings: $self hash keys that are simple scalars.

sub _simple_vars {
   return (keys (%SimpleVar));
}

#---------------------------------------------------------------------
# _sequence_vars
#---------------------------------------------------------------------
# Private class method to return the list of keys of %SequenceVar.
# Used in from_chunk().  Need a method because child IW::DaacAddress
# will have a _sequence_vars() too and that calls this one.
# Input: none.
# Output: array of strings: $self hash keys that are simple scalars.

sub _sequence_vars {
   return (keys (%SequenceVar));
}


#---------------------------------------------------------------------
# _is_simple $FIELDNAME
#---------------------------------------------------------------------
# Private class method that returns true or false depending on
# whether the given field name is in the list of simple scalar fields
# (the keys of %SimpleVar).  from_ODL() will use this instead of a
# direct access to %SimpleVar so that subclasses can check their list
# of simple variables first.
# Input: simple scalar string: one of the keys of $self.
# Output: 1 (true) or 0 (false).

sub _is_simple {
   # Note that $self is $_[0] and fieldname is $_[1]
   return (defined ($SimpleVar{$_[1]}));
}

#---------------------------------------------------------------------
# _simple_ivar $FIELDNAME, $MAX, $SELF, [$NEWVALUE]
#---------------------------------------------------------------------
# Private class method for generically getting and setting instance
# member variables - stored in the $self hash.
# Input: 1. scalar string: name of field to get or set value.
#        2. scalar integer: maximum number of characters for field.
#           This is passed in so subclasses can use this method too.
#        3. IW::Address object.
#        4. (optional) simple scalar string: value to store in object.
# Output:
#    If setting a variable:
#       It returns the new value if everything OK.
#       If value is too long, it sets an error message in the
#       IU::Errors object and returns the IU::Errors object.
#    If getting a variable:
#       It returns the current value of the variable, if it has one.
#       It returns the empty string if the variable is undefined.

sub _simple_ivar {
   my ($field, $max, $self, $value) = @_;

   # If setting:
   if (defined ($value)) {
      # If value string too long, set an error string.
      if (length ($value) > $max) {
         $self->{ERRORS}->set ("$field ($value) is too long: " .
                              length ($value) . ". Maximum is $max.");
         return ($self->{ERRORS});
      } else {
         # Set new value and return it.
         $self->{$field} = $value;
         return ($value);
      }
   } else {
      # Getting.  Return value if defined or empty string.
      return (defined ($self->{$field}) ? $self->{$field} : '');
   }
}

#---------------------------------------------------------------------
# _initial_ivar $FIELDNAME, $MAX, $SELF, [$NEWVALUE]
#---------------------------------------------------------------------
# Same as _simple_ivar except that this specifically allows
# the second character to be a '.' for middle initial.
#---------------------------------------------------------------------

sub _initial_ivar {
   my ($field, $max, $self, $value) = @_;

   # If setting:
   if (defined ($value)) {
      # If value string too long, set an error string.
      if (length ($value) > $max) {
         # If value is a character followed by a dot, remove dot.
         if ($value =~ /^(.)\.$/) {
            $value = $1;
            # Set value and return it.
            $self->{$field} = $value;
            return ($value);
         } else {
            # Else error.
            $self->{ERRORS}->set ("$field ($value) is too long: " .
                              length ($value) . ". Maximum is $max.");
            return ($self->{ERRORS});
         }
      } else {
         # Set new value and return it.
         $self->{$field} = $value;
         return ($value);
      }
   } else {
      # Getting.  Return value if defined or empty string.
      return (defined ($self->{$field}) ? $self->{$field} : '');
   }
}

#---------------------------------------------------------------------
# The simple get/set access methods.  All use _simple_ivar method
# above (except initial uses _initial_ivar).  Note that these return
# whatever _simple_ivar returns.  We must pass the max length so that
# subclasses can also call _simple_ivar with their max lengths.
# Note that the name of the method MUST be the lower case of the
# parameter name (the few that are not have an alias that is).  This
# requirement is for use in the from_ODL and from_CHUNK methods.
# Invocation:    Setting:   $addr->state("Massachusetts");
#                Getting:   $phone = $addr->phone;
# Input: simple scalar string value if setting.  None if getting.
# Output: see _simple_ivar notes above.
#---------------------------------------------------------------------

sub title   { _simple_ivar ('TITLE',   $MaxLength{TITLE},   @_) }
sub city    { _simple_ivar ('CITY',    $MaxLength{CITY},    @_) }
sub state   { _simple_ivar ('STATE',   $MaxLength{STATE},   @_) }
sub country { _simple_ivar ('COUNTRY', $MaxLength{COUNTRY}, @_) }
sub zip     { _simple_ivar ('ZIP',     $MaxLength{ZIP},     @_) }
sub phone   { _simple_ivar ('PHONE',   $MaxLength{PHONE},   @_) }
sub fax     { _simple_ivar ('FAX',     $MaxLength{FAX},     @_) }
sub firstname    { _simple_ivar ('FIRST_NAME',
                                 $MaxLength{FIRST_NAME}, @_) }
sub lastname     { _simple_ivar ('LAST_NAME', 
                                 $MaxLength{LAST_NAME}, @_) }
sub organization { _simple_ivar ('ORGANIZATION',
                                 $MaxLength{ORGANIZATION}, @_) }
sub email {
   # If setting e-mail, do extra validation.
   if (@_ > 1 and $_[1]) {         # Setting to a value.
      my $self = $_[0];
      my $value = $_[1];
      # Make sure address has a '@'.
      if ($value =~ /@/) {
         # Check if host can be found.
# The following block has been commented out because some hosts
# fail the check, for example, id@stx.com and id@erols.com.
# We need to do a different type of checking if any.
#        my ($name, $host) = split (/@/, $value);
#        my $ip = inet_aton ($host);
#        unless (defined ($ip)) {
#           $self->{ERRORS}->set ("Error. E-mail address '$value' " .
#                          "invalid. Couldn't resolve host: '$host'");
#           return ($self->{ERRORS});
#        }
      } else {
         $self->{ERRORS}->set ("Error. E-mail address '$value' " .
                               'invalid. Must contain @.');
         return ($self->{ERRORS});
      }
   }
   # Call get/set method.
   _simple_ivar ('EMAIL',   $MaxLength{EMAIL},   @_);
}

sub initial      { _initial_ivar ('MIDDLE_INITIAL',
                                  $MaxLength{MIDDLE_INITIAL}, @_) }

# Three alias methods used in "from_ODL" and "from_chunk":
sub first_name     { my $self = shift;  $self->firstname(@_); }
sub last_name      { my $self = shift;  $self->lastname(@_);  }
sub middle_initial { my $self = shift;  $self->initial(@_);   }

#---------------------------------------------------------------------
# name
#---------------------------------------------------------------------
# Get the user's complete title and name as one string.
# Input: none.
# Output: scalar string.

sub name {
   my $self = shift;
   my $s = '';
   $s .= $self->title     . ' '      if $self->title;
   $s .= $self->firstname . ' '      if $self->firstname;
   $s .= $self->initial   . '. '     if $self->initial;
   $s .= $self->lastname             if $self->lastname;
   return ($s);
}

#---------------------------------------------------------------------
# street [\@STREETARRAY]
#---------------------------------------------------------------------
# Get/set the user's street address.
# If getting:
#    In an array context, the street address is returned as an array
#       of lines (or empty array if no address).
#    In a scalar context, the street address is returned as a
#       \n-joined string (or empty if no address).
# If setting:
#    Return array of street address.
#    If too many lines or any line too long, set error strings in
#    ERRORS and return ERRORS object.
# Input: reference to street address array of 3 lines max, or none.
# Output: if wantarray: address array
#         else : address as single string with \n inserted.
#         if error: IU::Errors object.

sub street {
   my ($self, $addr) = @_;

   # If setting:
   if (defined ($addr)) {
      my $many = 0;                   # Flag if too many lines.
      my $long = 0;                   # Flag if any line too long.
      my @addr = grep {/./} @$addr;   # Get rid of empty lines.
      # If too many lines, complain.
      if (scalar (@addr) > $MaxLength{NUMSTREETLINES}) {
         $many = 1;
         $self->{ERRORS}->set ("Too many lines in address (" . 
                               scalar (@addr) . ") must be " .
                               $MaxLength{NUMSTREETLINES} .
                               " or less.");
      }
      # If any line too long, complain.
      foreach (@addr) {
         if (length() > $MaxLength{STREET}) {
            $long = 1;
            $self->{ERRORS}-> set ("Street line \"" . $_ .
                                   "\" is too long.  It is " .
                                   length() . ", maximum is " .
                                   $MaxLength{STREET} . ".");
         }
      }
      # If any error return errors object.
      if ($many or $long) {
         return ($self->{ERRORS});
      }
      # Otherwise, set the address and return it.
      $self->{ADDRESS} = \@addr;
      return (@addr);
   } else {
   # Getting.  Return address array if wanted, else string.
   return wantarray ? @{$self->{ADDRESS}} : join ("\n", @{$self->{ADDRESS}});
   }
   1;
}

#------------------------------------------------------------
# blank STRING
#------------------------------------------------------------
# Is this string blank?  It is if it's either undefined, or
# empty, or all-whitespace:  Stolen from IW::Util because we
# can't require IW::App here.

sub blank {
    my $str = shift;
    !defined($str) or ($str eq '') or ($str !~ /\S/);
}

#---------------------------------------------------------------------
# as_string
#---------------------------------------------------------------------
# Return address as a string (with newlines).  Good for debugging.
# Input: none.  Output: string with newlines.

sub as_string {
   my $self = shift;
   return (join ("\n",
                 join (" / ", $self->name, $self->organization),
                 join (" / ", $self->street),
                 join (", ",$self->city, $self->state, $self->country,
                       $self->zip),
                 join (" / ",  ("Phone: " . $self->phone),
                       ("Fax: ".$self->fax)),
                 "Email: " . $self->email  . "\n")
           );
}

#---------------------------------------------------------------------
# copy IW::Address
#---------------------------------------------------------------------
# Make your Address object a deep copy of another Address.  Usage:
#    my $new = new IW::UserAddr;
#    $new->copy($old);
# Input: an address to copy.  Can be object of any subclass too.
# Output: modifies your existing address to be a copy of the old.

sub copy {
   my ($self, $old) = @_;
   foreach ($old->_simple_vars) {
      $self->{$_} = $old->{$_}  if (defined ($old->{$_}));
   }
   if (defined ($old->{ADDRESS})) {
      my @street = @{$old->{ADDRESS}};
      $self->{ADDRESS} = \@street;
   }
   $self->{ERRORS}->set($old->{ERRORS}->get()) if ($old->{ERRORS});
   return ($self);
}

#---------------------------------------------------------------------
# equal IW::Address
#---------------------------------------------------------------------
# Check if two address objects contain the same textual data.  Usage:
#    if ($addr1->equal($addr2)) { ...
# Input: IW::Address object for comparison.  Can be subclass object.
# Output: 1 (true) if data equal. 0 (false) if not.
# Note: Doesn't compare the Errors component.  Should be no need.

sub equal {
   my ($self, $addr2) = @_;
   # Check each simple variable.
   my @simple = $self->_simple_vars;
   # Return 0 if value of simple variables are not equal.
   foreach (@simple) {
      # Return 0 if bothe defined but not equal.
      return (0) if (defined ($self->{$_}) && defined ($addr2->{$_})
                     && $self->{$_} ne $addr2->{$_});
      # Return 0 if one defined and other not.
      return (0) if ((!defined ($self->{$_}) &&
                      defined ($addr2->{$_})) or
                     (defined ($self->{$_}) &&
                      !defined ($addr2->{$_})));
      # Else check next.
   }
   # Check address.
   my @street1 = @{$self->{ADDRESS}};
   my @street2 = @{$addr2->{ADDRESS}};
   # Return false if number of street lines differ.
   my $num;
   return (0) if (($num = scalar (@street1)) != scalar (@street2));
   my $i;
   for ($i=0; $i< $num; $i++) {
      return (0) if ($street1[$i] ne $street2[$i]);
   }
   # All scalar data equal.  Return true.
   return (1);
}

#---------------------------------------------------------------------
# errors
#---------------------------------------------------------------------
# Returns Errors object stored in Address object.
# This could have been: sub errors { return ($_[0]->{ERRORS}); }
# EXCEPT that for backwards compatibility with persistent sessions,
# we need to see if the Errors object exists, and if not create one.

sub errors {
   my $self = shift;
   unless ($self->{ERRORS}) {
      $self->{ERRORS} = new IU::Errors;    # Add empty errors object.
   }
   return $self->{ERRORS};
}

#------------------------------------------------------------
# print_receipt_addr FILEHANDLE
#------------------------------------------------------------
# I<Instance method, private.>
# Write out an address, formatted for an end-user receipt.

sub print_receipt_addr {
   my ($self, $FH) = @_;

   # Name, organization, email:
   print $FH "     Name:         ", $self->name, "\n";
   if ($self->{ORGANIZATION}) {
      print $FH "     Organization: ", $self->{ORGANIZATION}, "\n";
   }
   print $FH "     Email:        ", $self->{EMAIL}, "\n";

   # Street address:
   if ($self->street) {
      print $FH "     Address:      ";
      my $street;
      foreach $street (@{$self->{ADDRESS}}) {
         print $FH $street, "\n                   ";
      }
      print $FH $self->{CITY}, " ", $self->{STATE}, " ", $self->{ZIP},
                "    ", $self->{COUNTRY}, "\n";
   }
   # Phone and fax:
   print $FH "     Phone:        ", $self->{PHONE}, "\n"
      if $self->{PHONE};
   print $FH "     Fax:          ", $self->{FAX}, "\n" 
      if  $self->{FAX};
   1;
}

#---------------------------------------------------------------------
#
# ODL INTERFACE
#
#---------------------------------------------------------------------

#---------------------------------------------------------------------
# from_ODL ODLOBJECT
#---------------------------------------------------------------------
# I<Class or instance method.>
# Initialize the address from a standard IMS ODL object.  The instance
# variables have the same names as the associated parameters.

sub from_ODL {
   my $self = shift;
   my $odl = shift || return (undef);

   # Be a class method, too!
   ref ($self) or $self = $self->new;

   # Read the ODL stream:
   $odl->rewind;
   my ($param, $value, $val);
   while (($param, $value) = $odl->next) {
      if ($param eq 'ADDRESS') {
         # If address, get array of strings and pass to set function.
         # This should be an array but might be a scalar value.
         if ($value->type eq 'ARRAY') {
            $self->street ([ map {$_->text} $value->as_array ]);
         } else {
            $self->street ([ $value->text ]);
         }
      } elsif (($value->type eq 'SCALAR') &&
               $self->_is_simple ($param)) {

         $val = $value->text;    # Get text from ODL value.

         # Call set function for parameter.  This will work for
         # subclasses with additional parameters too, because they
         # have set functions there.  All set functions have names
         # that are the lowercase version of the parameter.
         # Warning: make sure the functions we call don't die!

         my $cmd = '$self->' . lc($param) . '($val)';
         eval ($cmd);
      } else {
         warn "skipping unrecognized address parameter: '$param'";
      }
   }
   # Return it:
   return ($self);
}

#---------------------------------------------------------------------
# to_ODL
#---------------------------------------------------------------------
# I<Instance method.>
# Convert the address into a standard IMS ODL object.

sub to_ODL {

   my $self = shift;
   # Kludge! How long to make ORGANIZATION, because orders to GSFC-ECS hang
   # if it is too long.
   my $organization_len = shift || '70'; # how long to make ORGANIZATION
   my ($ivar, $line);
   my $street_odl;
   # Create street:
   if ($self->street) {
      $street_odl = IU::ODL->new_array;       
      foreach $line ($self->street) {
         next if ($line eq "");
         $street_odl->add_elem (IU::ODL->quoted ($line));
      }
   }
   # Create group, and add stuff in standard order:
   my $odl = IU::ODL->new_group;
   foreach $ivar (qw(TITLE FIRST_NAME MIDDLE_INITIAL LAST_NAME
                     ORGANIZATION ADDRESS CITY STATE ZIP COUNTRY
                     PHONE FAX EMAIL)) {
      if ($ivar eq 'ADDRESS') {
         $odl->add_param ('ADDRESS', $street_odl)     if $street_odl;
      } elsif ($ivar eq 'ORGANIZATION' && defined ($self->{$ivar})) {
         my $organization = substr $self->{$ivar},0,$organization_len;
         $odl->add_param ($ivar, IU::ODL->quoted ($organization)); 
      } else {
         if (defined ($self->{$ivar})) {
            $odl->add_param ($ivar, IU::ODL->quoted ($self->{$ivar}));
         } else {
            $odl->add_param ($ivar, IU::ODL->quoted ("") ) ;
         }
      }
   }
   return ($odl);
}

#---------------------------------------------------------------------
#
# CHUNK INTERFACE
#
#---------------------------------------------------------------------

#---------------------------------------------------------------------
# from_chunk IU::CHUNK
#---------------------------------------------------------------------
# Read a from a chunk and store into the internal data structure.
# A chunk containing "x-error" is considered null and void.
# By using the set functions, any vars that are too long will be
# noted in the object's ERRORS object.

sub from_chunk {
   my ($class, $chunk) = @_;
   my $errmsg;

   # Input chunk, checking for errors:
   defined ($chunk) or return undef;
   defined ($errmsg = $chunk->value ('x-error')) and
      return error 454, $errmsg;

   # Create new object.
   my $self = $class->new;

   # Initialize from chunk:
   my ($var, $val, $cmd);
   foreach $var ($self->_simple_vars) {
      # Call set function for parameter.  This will work for
      # subclasses with additional parameters too, because they
      # have set functions there.  All set functions have names
      # that are the lowercase version of the parameter.
      # Warning: make sure the functions we call don't die!
      $val = $chunk->value ($var);
      if (defined ($val)) {
         $cmd = '$self->' . lc($var).'("'. $val .'")';
         # If an e-mail address, escape the @.
         $cmd =~ s/\@/\\\@/  if ($var eq 'EMAIL');
         eval ($cmd);
      }
   }

   # Initialize from chunk:
   my ($seqvar, @seqarray, $seqcmd);
   foreach $seqvar ($self->_sequence_vars) {
      # Call set function for parameter.  This will work for
      # subclasses with additional parameters too, because they
      # have set functions there.  All set functions have names
      # that are the lowercase version of the parameter.
      # Warning: make sure the functions we call don't die!
      my @seqarray = $chunk->seq_values ($seqvar);

      # The following code compensates for error in getting sequence
      # string into an array.  We got an extra element containing
      # just \001.  Probably an error in IW_Chunk or IW_ODL (c code),
      # or possibly Chunk.pm::input().

      # Strip error character and throw out empty lines.
      my @clean;
      foreach (@seqarray) {
         # strip \001
         my $char = pack ("c", 1);
         my $cleanit = $_;
         $cleanit =~ s/$char//o;
         if ($cleanit =~ /\S/) {     # if any non-whitespace
            push (@clean, $cleanit);
         }
      }
      my $lcseqvar = lc($seqvar);
      if ($lcseqvar eq 'address') {
         $seqcmd = '$self->street(\@clean)';
      } else {
         $seqcmd = '$self->'. $lcseqvar.'(\@clean)';
      }
      eval ($seqcmd);
   }
   return ($self);
}
#=====================================================================
1;
