package IW::App;

=head1 NAME

IW::App - routines for an EOS Data Gateway application

=head1 DESCRIPTION

This module's routines take care of a lot of the dirty work for you: 
creating the CGI object, initiallizing the logs, etc.  Use it like this:

      use IW::App;
      
      ...
      
      # Your main subroutine:
      sub main {
          ...
      }
      
      # Should be the last statement in your file:
      go IW::App \&main;

The static C<go()> method causes an IW::App object to be built and placed 
in $APP (exported by default).  This application object then
calls your designated C<main()> function inside of an eval-wrapper,
so that fatal exceptions which are not otherwise caught propagate 
up to the top level where they can be handled gracefully.

Note that the C<go()> method I<does not return>.  If the main routine
fails, the application will exit with a negative status; otherwise,
it will exit with a zero status.

This is a far better way of dealing with errors than the older
$SIG{__DIE__} approach, since it allows us to catch exceptions
without nasty side-effects.


=head1 NOTES

Why the OO interface?  In the hopes that refinements of EOS Data Gateway CGI
scripts can be made by subclassing.  The use of exported globals may
complicate this, but at least the framework is there...

=head2 The cgi..LCK file

Because the session state is exclusive-locked, special provisions had to 
be made to ensure that repeated requests under the same session id don't
choke the system up with Perl images (each one waiting for the previous
one to quit, and the first one being caught in a loop due to a bug
or to search.browse/order activity).  Hence, the C<"cgi..LCK"> scheme,
whereby each new CGI image kills the previous image.

=head2 A Tale of Two Daemons

To allow the use of two HTTPDs (the normal one for serving documents
and standard CGI scripts, and the custom one for serving our Perl scripts),
we can't just assume that a URL of C<"/"> will get us where we want.
However, accidents will happen...

IN THE FUTURE, 
to prevent nasty accidents that cause the Perl daemon to be invoked when
not needed, all generated HTML pages could a BASE of the C<DocsUrl>
(which explicitly names the site's vanilla HTTPD).  That means that 
badly-formed URLs like C<"/foo"> would be handled by the standard HTTPD,
and not the custom one. 

=head1 WARNING

If you plan to do any forking, you need to close logs or set them
to autoflush beforehand; otherwise, output will be duplicated!


=head1 AUTHOR

Eryq, Hughes STX Corporation

=head RCS INFO

$Id: App.pm,v 1.198 2003/08/26 16:26:21 solomon Exp $

=cut


require Exporter;

@IW::App::ISA    = qw(Exporter);
@IW::App::EXPORT = qw($APP $CGI $HTML $REPLY $REQ $SESSION);

# Pragmatic modules:
use strict qw(refs vars subs);
use vars qw($APP $CGI $HTML $LOG $REPLY $REQ $SESSION);

# Standard/extension modules:
use FileHandle ();
use Fcntl;
use HTML::Stream qw(:funcs);
use Text::Template;
use URI::Escape;

# General EOS Data Gateway utilities:
use IU;
use IU::Log;
use IU::FileLogger ();
use IU::Session ();
use IU::Event;
use IU::Util qw(safe_kill);

# EOS Data Gateway CGI-specific:
use IW::HTML ();
use IW::Navigator;
use IW::Navigator::Link;
use IW::Reply ();
use IW::ReplyHead ();
use IW::Util qw(:taint :url nobreak);

require IW::Note;                # yes, that's right: require

use Data::Dumper;

#------------------------------------------------------------
#
# EXPORTED GLOBALS
#
#------------------------------------------------------------

# The application object.  Only one allowed:
$APP = undef;

# The CGI interface object:
$CGI = undef;

# The CGI request object:
$REQ = undef;

# The CGI reply object:
$REPLY = undef;

# The current session, or undef:
$SESSION = undef;

# An HTML stream, opened on the stdout:
$HTML = undef;

#------------------------------------------------------------
#
# PRIVATE GLOBALS
#
#------------------------------------------------------------

# Private: the log stream's filehandle:
$LOG = undef;

# Our external interface...
my $CGI_REQUEST_CLASS;
my $CGI_BASE_CLASS;
if (0) {
    require CGI::Request;
    require CGI::BasePlus;
    $CGI_REQUEST_CLASS = 'CGI::Request';  
    $CGI_BASE_CLASS    = 'CGI::BasePlus'; 
}
else {
    require IW::CGI;
    $CGI_REQUEST_CLASS = 'IW::CGI::Request';  
    $CGI_BASE_CLASS    = 'IW::CGI::BasePlus';
}

# File locking:
my %LOCK = (SH=>1, EX=>2, NB=>4, UN=>8);

# Fix for broken URI::Escape 3.6 (grrrrr):
my $UESC = '^A-Za-z0-9';

# Raw CGI query:
my %RawQuery;

# Color constants. Scaled to nearest 0x33 because of netscape color cube.
my $light_yellow = "#FFFFCC";
my $brilliant_yellow = "#FFFF00";
my $standard_blue = "#0000FF";
my $deep_navy = "#330099";
 my $nice_aqua = '#CCFFFF';
 my $almost_white = '#EEEEEE';

# Layout constants.
# The navigator column is 80 pixels minimum, and the body column
# is 560 pixels minimum. The full table should be as wide as it needs to be.
 my $NAVIGATOR_COL_CELLSIZE = "13%";
 my $BODY_COL_CELLSIZE      = "87%";
 my $FULL_BODY_TABLESIZE    = "100%";

# The header columns are sized slightly differently for prettier layout:
# The icon column is 130, so the title is far enough to the right, but the
# title column is narrowed slightly to 400 to keep it closer to the
# left column "border" in the background GIF.
 my $HEADER_ICON_CELLSIZE  = 80;
 my $HEADER_TITLE_CELLSIZE = 511;

#------------------------------------------------------------
#
# UTILITY FUNCTIONS
#
#------------------------------------------------------------

#------------------------------------------------------------
# sig_hup SIGNAL
#------------------------------------------------------------
# Signal handler: client has hung up.  Just exit.

sub sig_hup {
    exit -1;
}

#------------------------------------------------------------
# sig_bad SIGNAL
#------------------------------------------------------------
# Signal handler: we dumped core.  Try to do a puke():

sub sig_bad {    

    # Sentinel, to prevent recursive calls:
    return if $IW::App::sig_bad_in_progress;
    $IW::App::sig_bad_in_progress = 1;

    # Die!
    die "Caught a SIG$_[0]: shutting down!";
    exit -1;      # just in case
}

#------------------------------------------------------------
#
# CONSTRUCTING AND INITIALLIZING
#
#------------------------------------------------------------

#------------------------------------------------------------
# go MAIN
#------------------------------------------------------------
# I<Constructor.>
# Create a new application object, and start executing the
# given MAIN routine, which should be a subroutine reference.  
#
# Why the OO interface?  For the same reason as the CGI::
# OO interface: so you can create subclasses with slightly 
# different behaviour.

sub go {
    my ($class, $main) = @_;


    # Sentinel:
    die "sorry; only one $class instance allowed" if $APP;    

    # Create a new instance:
    my $self = {};
    bless $self, $class;

    # Initialize it, or die:
    $self->init;

    # Begin the managed-event section:    
    my $status = eval { &$main };    
    if ($@ or !$status) {               # main failed!
	
	# Get the message together:
	my $gasp = ($@ || 'main routine returned false');

	# Invoke the final exit:
	eval { $self->final_exit($gasp) };

	if ($@) {                 # WHOOOOOOOOOOOOOOOA! a bug in final_exit!!!
	    select $REPLY->CLIENT;          # important! restore old STDOUT
	    die "final_exit() bug!\n$@";    # error will propagate to default
	}
    }
    exit 0;
}

#------------------------------------------------------------
# DESTROY
#------------------------------------------------------------
# I<Destructor.>
# This can do misc cleanup before the application exits.

sub DESTROY { 1 }


#------------------------------------------------------------
#
# PRIVATE METHODS
#
#------------------------------------------------------------

#------------------------------------------------------------
# init_multisubmit_tags
#------------------------------------------------------------
# I<Private initiallizer>.
# If CGI parameter "msubmit" is defined, look for a tag with the name
# of the form "string.x", then look for a corresponding CGI tag matching
# "string" which in its value has encoded in it the multisubmit tags.
# Parse the endcoded multisubmit tags and set them in $REQ->param.
#
# The special tags 'msubmit.x' and 'msubmit.y' are set to the x and y 
# values of the submitted tag.
    
sub init_multisubmit_tags {
    my ($msubmit, $tag, $tagroot, $value, $wasclick);

    $REQ->param('msubmit') or return 1;    # no multisubmit buttons
    foreach $tag ($REQ->params) {

        # Oh! Great Goddess of regular expressions, explain thy 
        # secrets to me...
        # 
        # Look for a tag ending in ".x" or ".X", if you find a .x replace
        # it with nothing e.g. "msubmit7.x" gets turned into "msubmit7"
        # If there is no match then false is returned and the next is 
        # is executed taking us up to the next iteration of the foreach loop. 
        #
        # Break down of the regexp s/\.[xX]$//:
        #    s/     - substitution
        #    \.[xX] - look for .x or .X, escaped ".".
        #    $      - anchor to the end of the string.
        #    //     - replace with nothing.

        next unless ($tag =~ s/\.[xX]$//);

        # Look for the corresponding hidden tag which has encoded in it
        # the CGI tags for this button: 
        $msubmit = $REQ->param($tag) || undef; 
        # Not found so keep looking.
        next unless (defined($msubmit) );

	# Parse and load the encoded query tags if the value begins with
        # "msubmit?":
        #
        # Another regular expression.
        # If the string begins with "msubmit?" then remove it 
        # leaving just the encoded CGI tags.
        #
        # Break down of the regexp s/^msubmit\?//:
        #    s/        - substitution
        #    ^         - archor match to the beginning of the string.
        #    msubmit\? - look for "msubmit?" (escaped "?").
        #    //        - replace with nothing.

        if ($msubmit =~ s/^msubmit\?//) { 
           $REQ->extract_values($msubmit);
        
	   # Set the special 'submit.[xy]' tags.
	   $REQ->extract_values('msubmit.x=' .
              uri_escape($REQ->param("$tag.x"), $UESC));
	   $REQ->extract_values('msubmit.y=' . 
              uri_escape($REQ->param("$tag.y"), $UESC));

           # Found one, don't need to continue looking.
           return 1;
        }
    }
    1;
}

#------------------------------------------------------------
# init
#------------------------------------------------------------
# Initiallize the application object.
#
# This starts everything off: a CGI object is created, a log 
# to a tempfile is opened, signal handlers are installed, etc.
#
# We use confess() instead of puke(), because we need to rely on it strongly.

sub init {
    my $self = shift;

    # Any notes we accumulate:
    $self->{Notes} = [];

    # Install signal handlers for client hangups 
    # Note that we can't intercept SIGKILL or SIGSTOP:
    $SIG{HUP} =  \&sig_hup;
    $SIG{INT} =  \&sig_hup;
    # $SIG{PIPE} =  \&sig_hup;      # is this right?

    # Install *temp* signal handlers for internal errors (core dumps, etc.):
    $SIG{SEGV} = \&sig_hup;                        # not a typo
    defined($SIG{BUS}) && ($SIG{BUS} = \&sig_hup); # not a typo
    $SIG{ILL} =  \&sig_hup;                        # not a typo    

    ### NOTE: we do NOT catch SIGTERM!!! It is vital that SIGTERM 
    ### actually stop the gateway code, so that the "preemptive strike"
    ### will work!
    ###    
    ### $SIG{TERM} = \&sig_hup;

    # Open a per-invocation log, and start logging:
    $LOG = FileHandle->new_tmpfile;
    &IU::Log::init() or Carp::confess "yow! couldn't init logger!";
    my $logger = IU::FileLogger->open($LOG) 
	or Carp::confess "yow! couldn't open logger: " . IU::Log::last_error();
    $logger->autoflush(1);        # necessary because of forking; see above
    &IU::Log::set_handler('TEMPLOG', $logger);
    inform "Logging begun";

    # Create an HTML stream, opened on the standard output:
    $HTML = new IW::HTML \*STDOUT;

    my $w = $^W; $^W = 0;         # turn off warnings until we get past CGI...



    # Create a CGI interface object, which will read in the CGI info:
    $CGI = $CGI_BASE_CLASS->new()
	or Carp::confess "yow! no CGI interface!";

    # Create an object to represent the current CGI request:
#   eval "\&${CGI_REQUEST_CLASS}::Interface(\$CGI)";
    $REQ = $CGI_REQUEST_CLASS->new()
	or Carp::confess "yow! no CGI request interface!";

    $^W = $w;                     # okay; we can start warning agin!

    # QUICK: save the current CGI query!
    foreach ($REQ->params) { $RawQuery{$_} = [$REQ->param($_)] }

    # Create an object to represent the planned reply:
    $REPLY = new IW::Reply
	or Carp::confess "yow! no CGI reply interface!";
    $REPLY->use_pagebuffer;      # we want pagebuffering!

    # ...AT THIS POINT, WE'VE GOT AN APP!
    $APP = $self;
    if ($IU::ForceDebug) {       # explicit setting in configuration
	$self->debug_mode(1) if (uc($IU::ForceDebug) eq 'YES');
    }
    
    # Add the special "multisubmit" tags to the request:
    $self->init_multisubmit_tags;

    # Install *real* signal handlers for internal errors (core dumps, etc.):
    $SIG{SEGV} = \&sig_bad;
    defined($SIG{BUS}) && ($SIG{BUS} = \&sig_bad);
    $SIG{ILL} =  \&sig_bad;

    # Let our HTML stream know about the current CGI request:
    $HTML->req($REQ);

    # Check for "down.txt" file:
    if (-e "${IU::LibDir}/conf/down.txt") {      # we're down!!!!
	$self->final_exit(
               "(Error 901) This EOS Data Gateway is temporarily unavailable.");
	exit 0;
    }

    # Done!
    $self;
}


#------------------------------
#
# OUTPUT
#
#------------------------------

#------------------------------------------------------------
# end_page
#------------------------------------------------------------
# Prints out a standard page footer, and sends the page to the client.
# Also turns off suppress_links(), so we don't trounce on next page
# of a multipart document.

sub end_page {
    my $self = shift;

    my $links = !$self->suppress_links;      # for efficiency

    # PRINT ANY ACCUMULATED NOTES:

    if (@{$self->{Notes}}) {	
	my $note;
	$HTML -> P -> HR -> TABLE;
	foreach $note (@{$self->{Notes}}) {  $note->print_row }
	$HTML -> _TABLE;
    }

    # PRINT THE KEY...
    # Note that no page uses the key anymore, so we just skip it.
    # $self->print_button_key;

    # PRINT THE SIGNATURE...
    
    $HTML -> HR;
    $HTML -> t("Also, please read the NASA ");
    $HTML -> hypertext('Privacy, Security, Notices ',
                       "$IU::DocsUrl"."imswelcome/privacy_policy.html"); 
    $HTML -> t("and the EOS Data Gateway ");
    $HTML -> hypertext('accessibility policy.',
                       "$IU::DocsUrl"."imswelcome/privacy_policy.html#508"); 
    $HTML -> HR -> ADDRESS;
    $HTML -> small;

    unless ($self->skip_email_link()) {
       $HTML -> t("Comments, Questions, or Problems? ");
       $HTML -> hypertext('Email us',
		          IW::Util::nav_comments_url(),
		          TARGET => 'comment');
    }

    $HTML -> BR;
    $HTML -> t("Created by ");
    $HTML -> t("$IU::SystemName");
    output $HTML " version $IU::Version";
    $HTML -> BR;

    # Our Webmaster:
    output $HTML "Webmaster: $IU::WebmasterName ";
    $HTML -> t("($IU::WebmasterEmail)");
    $HTML -> BR;

    # Our NASA task rep:
    output $HTML "Responsible NASA Official: $IU::OfficialName ";
    $HTML -> t("($IU::OfficialEmail)");
    $HTML -> BR;

    # Finish up:
    $HTML -> _small -> _ADDRESS;

    # Close the main table body now.
    $HTML -> _TD -> _TR -> _TABLE;

    # DEBUGGING INFO...
    
    if ($self->{Debug})  {

	# Separator:
	for (1..2) { $HTML -> P -> t('.') }
	$HTML -> P -> IMG(SRC=>doc_url('icons/waveline.gif'),
                          ALT=>'Wavy separator line');
	for (1..2) { $HTML -> P -> t('.') }

        my $inline = 1;

	# Debugging info:
	$HTML -> H1 -> t("Debugging information") -> _H1;
	$self->print_debugging_info(Traceback=>'No traceback: just debugging',
				    Inline=>$inline);

	# Sign it:
	$HTML -> HR -> ADDRESS -> t("$IU::SystemName $IU::Version live trace") 
              -> _ADDRESS;
    }

    # End it and send it:
    $HTML -> _BODY -> _HTML;
    $REPLY->flush_output;   

    # Reset page options, so we don't step on next page of a multipart/replace:
    $self->{Message} = undef;
    $self->{Notes} = [];
    $self->suppress_key(0);
    $self->suppress_links(0);

    1;
}

# include_and_fill front-ends Text::Template's fill_in_file, passing along
# the configuration variables from makewelcome automatically. 
#
# Text::Template::fill_in_file shifts off the first parameter (the file
# name), and then passes all the leftover parameters to 
# Text::Template::fill_in, so we can just tack on the extra parameters here
# and then call the thin we really wanted in the first place.
sub include_and_fill {
  # %config_variables already exists outside, so we can pass it along.
  Text::Template::fill_in_file(shift, @_, PACKAGE => "IW::App",
                                          HASH    => $APP->{ConfigVars}); 
}

#------------------------------------------------------------
# print_last_query  SUBDIR
#------------------------------------------------------------
# Output the last query that was done.
# KLUDGE: this is a rotten place to put this function, but then,
# IW::Util was worse.  I just wanted to get this DONE.

sub print_last_query {
    my %params = @_;
    my $searchType = $params{Type};

    my $self = shift;
    my $subdir = shift;    # may be undef if not specified.

    # Get path to query file, and leave quietly if there was none:
    $SESSION or return undef;
    my $query;
    if (defined ($subdir)) {
        $query = $SESSION->path_to("$subdir/query.html");
    } else {
        $query = $SESSION->path_to('query.html');
    }
    (-e $query) or return undef;

    # Determine the url for downloading the last query:
    my $url = script_url('mode'=>'DNLOAD', 'submode'=>'QUERY');
    
    # Title, summary, and directions:
    $HTML-> TABLE -> TR -> TD(WIDTH=>"100%");
    print "<hr><h3>The last search you submitted was for...</h3>\n";
    $HTML -> small;
    IW::Util::interpolate($query);
    print "If you ";  
    $HTML->hypertext("save this query to your machine", $url);
    print <<EOF;
           , you can restore it later into 
           the search form by entering its URL 
           into the special field at the bottom of the search form.
EOF
    $HTML -> P;
  
    # If it is a search request for browse, print the last link
    # to point to the browse search status page.
    #if (defined ($searchType) and ($searchType eq "BROWSE")) {
    #   $self->print_last_commstat_link ();
    #}
    #else {
    #   $self->print_last_commstat_link ($subdir);
    #}
    $HTML -> _small;
    $HTML -> P;
    $HTML -> _TD -> _TR -> _TABLE;
    1;
}

#------------------------------------------------------------
# print_last_commstat_link  SUBDIR
#------------------------------------------------------------
sub print_last_commstat_link {
    my $self = shift;
    my $subdir = shift;    # May be undef if not given
    my $last_comm_mode;
    if (defined ($subdir) and $subdir) {
        $last_comm_mode = $SESSION->state->{'Last'.$subdir.'CommMode'} ||
                          'SEARCH';
    } else {
        $last_comm_mode = $SESSION->state->{LastCommMode} || 'SEARCH';
    }
    $HTML -> t("See the ");
    $HTML -> hypertext("final communication status screen", 
                       script_url('mode'=>$last_comm_mode,
					   'submode'=>'COMMFINAL')); 
    $HTML -> t(" of the last search performed. ");
}

#------------------------------------------------------------
# string_messages
#------------------------------------------------------------

sub string_messages {
    my $self = shift;
    my $string;
    my $msg = $self->{Message};
    $string = $msg->as_string if $msg;
    return ($string);

}
#------------------------------------------------------------
# print_messages
#------------------------------------------------------------
# Output the raised message "window" above the main body of the page.

sub print_messages {
    my $self = shift;

    my $warn = (-e "${IU::LibDir}/conf/warn.txt");
    my $msg = $self->{Message};
    # Check for a special warn.txt file in the user's session.
    my $special = undef;
    my $specialfile;
    if ($SESSION) {
	$specialfile = $SESSION->path_to('warn.txt');
	$special = (-e $specialfile);
    }
    # Check for a temp_msg.txt file in the user's session.
    my $temp = undef;
    my $tempfile;
    if ($SESSION) {
	$tempfile = $SESSION->path_to('temp_msg.txt');
	$temp = (-e $tempfile);
    }
    # Check for a space_msg.txt file in the user's session.
    my $space = undef;
    my $spacefile;
    if ($SESSION) {
	$spacefile = $SESSION->path_to('space_msg.txt');
	$space = (-e $spacefile);
    }
    # Is there anything to do?
    ($warn || $msg || $special || $temp || $space) or return 1;


    # Site-wide warning messages go right below the title ...
    if ($warn) {
	if (open(WARN, "${IU::LibDir}/conf/warn.txt")) {

	    # Start a new table:
	    $HTML -> TABLE(BORDER=>0, CELLPADDING=>4, CELLSPACING=>2);

	    # Start a new row, with a "warning" icon:
	    $HTML -> TR
		  -> TD(BGCOLOR=>$light_yellow, WIDTH=>'4%')       # yellow
		  -> IMG(SRC   => doc_url('icons/note.gif'),
			 ALT   => '[** WARNING **]')
		  -> _TD
		  -> TD(BGCOLOR=>$light_yellow, WIDTH=>'96%');      # yellow
	    
	    # Slurp in the file, and display it:	    
	    $HTML -> B;
	    output $HTML (WARN->getlines());
	    $HTML -> _B -> _TD -> _TR;
	    close(WARN);
            $HTML -> P;
#            $HTML -> hypertext("List of Public EDG sites",${IU::ImsSitesUrl});
	    $HTML -> _TABLE;
	}
    }
    # Session specific warning messages go next...
    if ($special) {
	if (open(SPECIAL, $specialfile)) {
	    # Start a new row, with a "warning" icon:
	    $HTML -> TABLE(BORDER=>0, CELLPADDING=>4, CELLSPACING=>2);
	    $HTML -> TR 
		  -> TD(BGCOLOR=>$light_yellow, WIDTH=>'4%')       # yellow
		      -> IMG(SRC   => doc_url('icons/note.gif'),
			     ALT   => '[** WARNING **]')
                  -> _TD
		  -> TD(BGCOLOR=>$light_yellow, WIDTH=>'96%');      # yellow
	    
	    # Slurp in the file, and display it:	    
	    $HTML -> B;
	    output $HTML (SPECIAL->getlines());
	    $HTML -> _B -> _TD -> _TR;
	    close(SPECIAL);
	    $HTML -> _TABLE;

	}
    }
    # Session specific temporary messages go next...
    if ($temp) {
	if (open(TEMP, $tempfile)) {
	    # Start a new row, with a "warning" icon:
	    $HTML -> TABLE(BORDER=>0, CELLPADDING=>4, CELLSPACING=>2);
	    $HTML -> TR 
		  -> TD(BGCOLOR=>$light_yellow, WIDTH=>'4%')       # yellow
		      -> IMG(SRC   => doc_url('icons/note.gif'),
			     ALT   => '[** WARNING **]')
		  -> _TD
		  -> TD(BGCOLOR=>$light_yellow, WIDTH=>'96%');      # yellow
	    
	    # Slurp in the file, and display it:	    
	    $HTML -> B;
	    output $HTML (TEMP->getlines());
	    $HTML -> _B -> _TD -> _TR;
	    close(TEMP);
	    $HTML -> _TABLE;

	}
	# Delete the temporary message.
        system ("/bin/rm", "$tempfile");
    }
    # Session specific space messages go next...
    if ($space) {
	if (open(SPACE, $spacefile)) {
	    # Start a new row, with a "warning" icon:
	    $HTML -> TABLE(BORDER=>0, CELLPADDING=>4, CELLSPACING=>2);
	    $HTML -> TR 
		  -> TD(BGCOLOR=>$light_yellow, WIDTH=>'4%')       # yellow
		      -> IMG(SRC   => doc_url('icons/note.gif'),
			     ALT   => '[** WARNING **]')
		  -> _TD
		  -> TD(BGCOLOR=>$light_yellow, WIDTH=>'96%');      # yellow
	    
	    # Slurp in the file, and display it:	    
	    $HTML -> B;
	    output $HTML (SPACE->getlines());
	    $HTML -> _B -> _TD -> _TR;

	    close(SPACE);
	    $HTML -> _TABLE;
	}
    }
    # Now output whatever our short status message was...
    $HTML -> TABLE(BORDER=>0, CELLPADDING=>4, CELLSPACING=>2);
    $msg->print_row if $msg;
    $HTML -> _TABLE;
}

#------------------------------------------------------------
# start_page PARAMHASH
#------------------------------------------------------------
# Prints out a page header for a standard page, complete with the 
# MIME header, the document head, the icons, and the navigator.
#
# Parameters affecting the title are...
#
#    Title      => The title of the page.
#
#    SuperTitle => Optional.  A title above the title in the body.
#
#    BlinkTitle => Optional.  Blink the title.
#
# Parameters affecting the appearance in other ways are...
#
#    Icon       => Optional.  The "src" to use for the page icon.
#
#    Context    => The "context" token.  This token, which is all lowercase
#                  letters and very terse, uniquely identifies this page
#                  for the "Help" subsystem.
#
#    LinkColors => Optional.  Array of colors: [NORMAL, ACTIVE, VISITED].
#
# Parameters affecting the HEAD sent with this page are...
#
#    Base       => Optional.  If given, means that a BASE tag should be
#                  output in the HEAD, with the argument as the HREF.
#
#    HeadTitle  => Optional.  The title of the page as it will appear in
#                  the <HEAD>...</HEAD> portion.  Defaults to the Title.
#
#    Refresh    => Optional.  Provide a hash ref which gives the URL
#                  and seconds to wait; e.g.:
#
#                      Refresh => {Wait=>15, URL=>$url}
#
#                  If URL is omitted, it defaults to that of the current page.
#                  If Wait is omitted, it defaults to 15.
#
#                      Refresh => {}          # wait 15, then reload me 
#
# Parameters affecting the HTTP reply header are...
#
#    ReplyHead  => Optional.  If given, then *this* is the HTTP reply header
#                  to use; otherwise, a default one is generated.
#
#    Expires    => Optional.  Used to set expiration in defaulted reply header.
#                  USE WITH EXTREME CAUTION, AND TEST THOROUGHLY!!!!!
#
#    NoLogOut   => Optional.  Used to suppress the Logout button

sub start_page {
    my $self = shift;
    my %params = @_;

    # Get some params:
    my $bodytitle  = $params{Title} || 'untitled';
    my $headtitle  = $params{HeadTitle} || $bodytitle;
    my $supertitle = $params{SuperTitle};
    my $icon       = $params{Icon} || doc_url('icons/nasalogo.gif');
    my $searchlink = $params{Search} || 'NO';

    my $otherlinks = doc_url('icons/otherlinks.gif');
    my $background = doc_url('icons/bg.gif');

    my $link_colors = $params{LinkColors} || 
	[$standard_blue, $brilliant_yellow,$deep_navy];

    my $nologout = $params{NoLogOut} || 'NO';
    my $java_url = $params{AllowJava} || undef;

    # The HTTP-reply header:
    my $head = $params{ReplyHead};
    if (!$head) {
	$head = new IW::ReplyHead;
	$head->okay if $REPLY->need_status_line;
	$head->content_type('text/html');
	$head->expires($params{Expires}) if $params{Expires};

	# DO NOT ADD THIS LINE BACK IN!!! It confuses Netscape mightily:
	# $head->expires('+2d');     # try to force browser to keep it around
    }
    $head->send(\*STDOUT);

    # Record whatever it was the searchtype was set to (if it was set).
    $SESSION->state->{LastSrchType} = $SESSION->{Navigator}->{SrchType}
      if $SESSION->{Navigator}->{SrchType};

    # If it wasn't, try restoring the last setting (if there was one).
    $SESSION->{Navigator}->{SrchType} =  $SESSION->state->{LastSrchType}
      if  $SESSION->state->{LastSrchType};

    # Figure out refresh, if there is any.
    my $refresh = defined $params{Refresh} && do {
                 my %r = %{$params{Refresh}};
                 defined($r{Wait}) or $r{Wait} = 15;     # default wait
                 $r{Wait} .  ($r{URL} ? "; URL=$r{URL}" : '');
              };

    my $toggle = script_url('mode' => 'MAINSRCH',
                            'SrchTypeevent'  => 'TOGGLE',
                            'SrchTypenewtype' => 'INVENTORY_SEARCH');

    my $help = help_url($params{Context});

    $self->{ConfigVars} = { 
                            'link_colors' => $link_colors,
                            'body_color'  => $almost_white,
                            'bar_color'   => $nice_aqua,
                            'highlight'   => $light_yellow,

                            'title'      => "${IU::SystemName}: $headtitle",
                            'supertitle' => $supertitle,
                            'bodytitle'  => $bodytitle,

                            'logo'      => $icon,

                            'searchlink'        => ($searchlink eq 'YES'),
                            'toggle_to_primary' => $toggle,

                            'comments_url' => IW::Util::nav_comments_url(),
                            'help_url'     => $help,

                            'page'      => ($REQ->param('page') || ""),

                            'refresh' => $refresh,
                            'need_scrollpoint' => $HTML->need_scrollpoint,
                            'links_suppressed' => $self->suppress_links(),

                            'template_dir' => $IU::Templates,

                            'user'         => $ENV{REMOTE_USER},
                            'marks'        => 
                              [$SESSION->{Navigator}->{Indicator},
                               $SESSION->{Navigator}->{SrchType}],
                            'mode'      => $REQ->param('mode'),
                          };

    # Expand the start-page template.
    my $template =
       Text::Template->new(TYPE => 'FILE',
                           SOURCE => "$IU::Templates/app_start.template");

    print
  $template->fill_in(PACKAGE => "IW::App",
                     HASH    => $self->{ConfigVars}
                    );

    # The messages:
    $self->print_messages;
 
    1;
}

#------------------------------
#
# SESSION MANAGEMENT
#
#------------------------------

#------------------------------------------------------------
# get_session [OPTIONS]
#------------------------------------------------------------
# Resume the session given by the 'sid' param, creating a session 
# interface object in $SESSION.  This also "touches" the session.
#
# The OPTIONS are as follows:
#
#      'CREATE'      If there is no 'sid', we I<must> try and create a 
#                    I<brand new> session.  Default setting is off.
#
#      'NOBIGGIE'    It's okay if there's no 'sid': just return undef.  
#                    Default setting is off.
#
#      'REINSTATE'  "Resume" a tempory account even though we're in a persistent
#	             session. Used to 'CANCEL' from a persistent Session.' 
#
#
# Normally, failure to create/resume a session is a fatal error.  
# The only exception is the case where there is no 'sid' AND we shouldn't 
# create a new session AND 'NOBIGGIE' is specified.
# 
# Otherwise, the session object is returned.

sub get_session {
   my ($self, $options, $passed_sid) = @_;

   # If a mode did a switch to another mode, there was no need
   # to get the session again, so just return the current session.
   if (defined ($REQ->param('x-switch'))
      and defined ($SESSION)) {
      return $SESSION;
   }

   # Extract the options:
   $options or $options = '';
    
   my $create    = $options =~ /\bCREATE\b/i;
   my $no_biggie = $options =~ /\bNOBIGGIE\b/i;
   my $reinstate = $options =~ /\bREINSTATE\b/i;
   my $just_created = 0;  
    
   # Get id, and act appropriately:
   my $sid = $REQ->param('sid');
      $sid = $ENV{REMOTE_USER} if $ENV{REMOTE_USER} ne 'guest';

   if ($reinstate && $passed_sid) {  # connect to specified sid.

      $SESSION = IU::Session->resume(Id => $passed_sid, Type=>'STANDARD')
         or puke "couldn't resume session $passed_sid";

   }
   elsif ($sid) {                     # we were given an sid: resume!

      $SESSION = IU::Session->resume(Id => $sid)
         or puke "couldn't resume session $sid";

   }
   elsif ($create) {              # we should create a new session...

      $SESSION = IU::Session->create or puke "couldn't create session";
      ims_event('LOGIN',
                'AGENT' => ($ENV{HTTP_USER_AGENT}||''));     # log it!
      $just_created = 1;

   }
   elsif ($no_biggie) {        # it's okay to not have a session...

      return undef;

   }
   else {                      # ack! we need a session!

      puke 600, "need session, but no sid and creation is forbidden!";

   } 
	
   # Seize the session:
   seize_session();
    
   # Do remaining setup:
   $SESSION->setup_state(Opts=>'WRCREAT|BLOCK');
   $SESSION->setup_env;
   $SESSION->touch;               # kilroy was here

   # If we just created it, store some one-time info:
   if ($just_created) {
      $SESSION->state->{Created} = time;
      $SESSION->state->{RemoteHost} = ($ENV{REMOTE_HOST}||$ENV{REMOTE_ADDR});

      # Set up defaults for User Preferences:
      $SESSION->state->{AllowJava} = $IU::AllowJava || 'Y';
      $SESSION->state->{AllowSmallFont} = 'Y';
      $SESSION->state->{AllowDynamicValids} = 'N'; 
      $SESSION->state->{SRCHFORM_DirectionsSeen} = 0;
   }

   # Handle request to turn java on or off.
   my $java = $REQ->param('java');
   if (defined $java) {
      $SESSION->state->{AllowJava} = $java;
   }

   # Set up a navigator for this new session.
   $SESSION->{Navigator} = new IW::Navigator;

   # Added user and query object to the session state 
   # from prefs.dat and query.dat if needed.

   if (!defined($SESSION->state->{User})) {

      my $prefs_file = $SESSION->path_to('prefs.dat');
      require IW::ModePREFS;
      $SESSION->state->{User} = IW::ModePREFS::load_prefs($prefs_file)
         if (-s $prefs_file); 

   }

   # Make sure we have a query object in the session state.
   # If not, load user's previous query or install a default query.
   if (!defined($SESSION->state->{Query})) {

      my $query_file = $SESSION->path_to('pub/query.dat');
      if (-s $query_file) {
         $SESSION->state->{Query} = IU::Query->new->reload($query_file);
      } else {
         $SESSION->state->{Query} = IU::Query->new->default;
      }
   }
    
   # Regardless of site's debugging state (forced, etc.), honor any 
   # *explicit* wishes of the user:
   my $debug = $SESSION->state->{Debug};
   defined($debug) and $self->debug_mode($debug || 0);
    
   # Set up HTML output:
   # debug "setting up HTML output";
   my $allow_small = $SESSION->state->{AllowSmallFont};
   defined($allow_small) and $HTML->allow_small($allow_small);

   # Done!
   $SESSION;
}

#------------------------------------------------------------
# seize_session
#------------------------------------------------------------
# We've got a session!  But we've got to be careful... a rapid
# click on a hyperlink could mean that the last CGI script is still
# running under this sid... and that can cause problems!  If so, kill it!  

sub seize_session {

    check_lock();

    # Seize and lock the PIDFILE:
    my $lockfile = $SESSION->path_to('cgi..LCK');
    ### DELOUSE "seizing $lockfile";
    if (! -e $lockfile) {       # we're the first!
	### DELOUSE "we're the first to ask for $lockfile";
	open PIDFILE, ">$lockfile"           or puke "open $lockfile: $!";
	print PIDFILE $$;                    # print *before* flock!!!!
	PIDFILE->flush;                      # get that info to disk!!!!
	flock PIDFILE, $LOCK{EX}|$LOCK{NB}   or puke "flock failed: $!";
    }
    else {                      # we're not the first!	
	### DELOUSE "we're NOT the first";
	open PIDFILE, "+<$lockfile"          or puke "open $lockfile: $!";
	PIDFILE->autoflush(1);
	if (!flock PIDFILE, $LOCK{EX}|$LOCK{NB}) {  # eek! they're still here!
	    ### DELOUSE "couldn't flock: $! (are they still here?)";
	    my $otherpid = 0 + <PIDFILE>;
	    ### DELOUSE "their pid is $otherpid, mine is $$!";

	    # Note: MUST do a fatal signal here!!!!
	    #    We use TERM, since we don't trap it, but wghttpd does, so
	    #    it's a little safer.
	    warn "$$ is killing $otherpid";
	    safe_kill 'TERM', $otherpid;    # benign failure on race condition!
	    flock(PIDFILE, $LOCK{EX}) or puke 600, "can't get CGI lock: $!";
	    seek(PIDFILE, 0, 0);
	}
	print PIDFILE $$;                            # mine!
    }
    
    # Make sure the file descriptor does NOT remain open across an exec:
    fcntl(PIDFILE, F_SETFD(), 1) or die "fcntl: $!";
    ### DELOUSE "seized!";
    1;
}

#------------------------------------------------------------
# check_lock 
#------------------------------------------------------------
# If "secured" session and not ModeLOGIN.pm then check for
# persistent_user.lock file, if not there then force the 
# user to a login screen.

sub check_lock {

 
     my $url = script_url();

     if ( ($url =~ /$IU::SecuredDir/) && 
          ($REQ->param('mode') ne 'LOGIN') && 
          ($REQ->param('mode') ne 'PREPARE') && 
          ($REQ->param('mode') ne 'SRCHPASS') &&
          ($REQ->param('mode') ne 'CREATEACCOUNT') )  {

          my $lock_file = $SESSION->path_to('persistent_user.lock');

          if (!(-s $lock_file)) {
    
               my $url = script_url('mode'     => 'LOGIN',
                                    'submode'  => 'RESUME',
                                    'nolock'   => 'Y');

               $REPLY->redirect($url, 'PERMANENT'); 

           }

     }	

1;

}



#------------------------------
#
# OTHER METHODS
#
#------------------------------

#------------------------------------------------------------
# debug_mode YESNO
#------------------------------------------------------------
# Turn debugging mode on/off for output.  No arguments returns current mode.

sub debug_mode {
    my ($self, $yesno) = @_;    
    defined($yesno) and $self->{Debug} = $yesno;
    $self->{Debug};
}

#------------------------------------------------------------
# suppress_key
#------------------------------------------------------------
# Dummy routine; retained to track down all leftover references
# to this routine.

sub suppress_key {
    extramsg "Called by " , join(" ",caller());
    1;
}

#------------------------------------------------------------
# skip_email_link YESNO
#------------------------------------------------------------
# Suppress the normal printing of the "email us" link. 
#
# You I<must> supply an argument to have this function do anything;
# otherwise, it just returns the current state.

sub skip_email_link {
    my ($self, $yesno) = @_;    
    defined($yesno) and $self->{SkipEmailLink} = $yesno;
    extramsg "Skip email is " . $self->{SkipEmailLink} 
      if defined $self->{SkipEmailLink};
    $self->{SkipEmailLink};
}

#------------------------------------------------------------
# suppress_links YESNO
#------------------------------------------------------------
# Suppress the normal printing of certain hyperlinks.  This is for those 
# pages where we don't want the user leaving via the navigator or the 
# 'other links'.
#
# You I<must> supply an argument to have this function do anything;
# otherwise, it just returns the current state.

sub suppress_links {
    my ($self, $yesno) = @_;    
    defined($yesno) and $self->{SuppressLinks} = $yesno;
    $self->{SuppressLinks};
}

#------------------------------------------------------------
# message [MESSAGE]
#------------------------------------------------------------
# Set contents of message window.  Each call overwrites the previous one.
# Returns true if there is a message, false otherwise.

sub message {
    my $self = shift;
    @_ or return $self->{Message};     # no args? just return if message set
    my $note = $_[0];
    unless (ref($note)) {              # if not an object...
	$note = new IW::Note Text=>join('',@_);
    }
    $self->{Message} = $note;
}

#------------------------------------------------------------
# note NOTE
# note CLASS,TEXT
#------------------------------------------------------------
# Set a note.

sub note {
    my $self = shift;
    my $note = $_[0];
    unless (ref($note)) {
	$note = new IW::Note Class=>shift, Text=>(join('',@_));
    }
    push @{$self->{Notes}}, $note;
}




#------------------------------
#
# DEATH AND DYING...
#
#------------------------------

#------------------------------------------------------------
# final_exit REASON
#------------------------------------------------------------
# A special method invoked by the "go" method, for when an application's
# main routine fails.  Cancels prior output, prints the dying gasp,
# and exits.
#
# B<Do not invoke this directly> if you are simply I<using> this
# module.  Just do a die(), croak(), or confess() at any point.

sub final_exit {
    my $self = shift;
    my $fullmsg = shift;


    #
    # Clean up death reason, and extract some info from it...
    #

    # Try to load the error and diagnostics modules:
    eval { require IU::Err   } or error("IU::Err not loaded: $@");
    eval { require IU::Diag; } or error("IU::Diag not loaded: $@");

    # Extract the error number from the death string, if any:
    $fullmsg =~ s{\[[^]]{24}\][ \t][^:]+:}{}mg;   # kill times and files
    my $errno;
    if (($errno) = ($fullmsg =~ /^ *\(Error ([-\d]+)\) /m)) {
	$fullmsg =~ s/^ *\(Error [-\d]+\) //;
    }       
    $errno or $errno = '000';
    
    # Get the basic error string:
    my $errstr = (defined(&IU::Err::str) ? &IU::Err::str($errno) : 
		  '(sorry: no error code description)');
    
    # Whittle down the full error message into more-printable alternatives:
    my ($errmsg) = ($fullmsg =~ m{^\s*(.*)$}m);   # just the top line
    $errmsg =~ s/\s*at \S+ line \S+$//mg;         # chop off location
    $errmsg = ucfirst($errmsg);                   # capitalize
    $errmsg .= '.' if ($errmsg !~ /[\.!?]$/);     # punctuate


    #			
    # Stop the presses!  Stop the presses!
    #

    $REPLY->abort_output;
    error("Fatal error: sending dying gasp...");

    #
    # Output the document...
    #

    # Output the HTTP header:
    my $head = new IW::ReplyHead;
    $head->okay if $REPLY->need_status_line;        # important!
    $head->content_type('text/html');
    print $head->as_string;
    
    # Establish an HTML stream, if none exists:
    $HTML or $HTML = new IW::HTML \*STDOUT;

    # Get a FULLY-QUALIFIED docs URL, since we may be emailing this file!
    my $fullDocsUrl = $IU::DocsUrl;
    unless ($fullDocsUrl =~ m{^http://}i) {

        # Check to see if we're playing with a SECURE server.
        my $s = ($IU::SecureServer ? "s" : "");
	$fullDocsUrl ="http$s://$ENV{SERVER_NAME}:$ENV{SERVER_PORT}$IU::DocsUrl";
    }

    # The head...
    $HTML -> HTML;
    $HTML  -> comment("$errno  $errstr");
    $HTML -> HEAD -> TITLE -> t("$IU::SystemName Error $errno") 
          -> _TITLE -> _HEAD;
    $HTML -> BODY(LINK=>'#E00000', ALINK=>'#FFFF00', VLINK=>'#800000');


    # The start of the body...		    
    $HTML -> TABLE -> TR;
    if (defined($IU::DocsUrl)) {
	$HTML -> TD -> IMG(SRC=>"${fullDocsUrl}icons/nasalogo.gif",
			   ALT=>'NASA logo');
    }
    $HTML -> TD -> color('IMSERROR') 
	  -> H1 
	  -> A(NAME=>'SCROLL')       # set scroll point to text
	  -> t("Error $errno:")
	  -> _A
	  -> BR 
	  -> t("\u$errstr.") 
	  -> _H1 -> _color;
    $HTML -> _TABLE;

    # The summary, of... 
    $HTML -> DL;		    

    #    ...the reason we died:
    my $errmsg_html = &html_escape($errmsg);
    $errmsg_html =~ s/\n/<BR>\n/g;     # format nice
    $HTML -> P -> DT -> 
           B -> color('IMSERROR') ->
           t("Problem:") -> 
           _color -> _B -> DD;
    print $errmsg_html;

    #    ...diagnostics:
    $HTML -> P -> DT 
          -> B -> color('IMSERROR') 
	  -> t("Details, and recommended action to correct the situation:") 
	  -> _color -> _B -> DD;
    if (defined(&IU::Diag::help)) {   # we got 'em
	my $help = IU::Diag::help($errno) || "Sorry... no diagnosis!";
	foreach (split /\n\n+/m, $help) { $HTML -> t($_) -> P }
	$HTML -> DD;
    }
    else {                            # whoa!
	print "None... sorry, I couldn't even load my diagnostics library!";
    }

    #    ...new session hyperlink (if necessary):
    if ($errno == 521) {
	$HTML -> B 
	      -> hypertext("Start a new session.", 
			   script_url('SESSION'=>'', 'mode'=>'MAINSRCH'))
	      -> _B;
    }
    $HTML -> _DL;

    #   ...directions (go "back", unless down for maintenance):
    if ($errno == 901) {      # down for maintenance...
	$HTML -> P -> PRE(STYLE=>"margin-left:5%");
	$HTML -> t(`cat ${IU::LibDir}/conf/down.txt`);
	$HTML -> _PRE -> _P;
        $HTML -> P;
#        $HTML -> hypertext("List of Public EDG sites",${IU::ImsSitesUrl});
    }
    else {                    # other error...		   
	$HTML -> B -> color('IMSERROR');
	output $HTML <<EOF;
	      When you are ready, hit your browser's "back" button to proceed.
EOF
        $HTML -> _color -> _B -> small;
        output $HTML <<EOF; 
              If you get a "Repost Form Data?" message, just click "OK".  
              If you get that message a lot, you need to expand your memory 
              and/or disk cache sizes; it's nothing we can control.
EOF
        $HTML -> _small;
    }

    # Short or long form?
    my $benign_error =
	($errno >= 510 and $errno < 550) ||   # client-input
        ($errno >= 900 and $errno < 950);     # planned/known
    my $short_form = ($benign_error && !$self->debug_mode);  # force long form!

    # The short form should tuncate at this point. If we decide to
    # go long form below. we'll reset the truncation point appropriately.
    $REPLY->set_truncation_point();
    debug "truncation point set to ". $REPLY->{TruncationPoint};
		    
    # If long form, output contacts and debugging info...
    if (!$short_form) {

	# Output contact info:
	$HTML -> P -> HR 
              -> H2 -> t("Contact information") -> _H2
	      -> UL
#             -> LI -> t_B('Site administrator: ') 
#		    -> mailto($IU::SiteAdminEmail)
              -> LI -> t_B("$IU::SystemName Developers: ") 
		    -> mailto($IU::HomeEmail)
              -> LI -> t_B('EOSDIS V0 Home Page: ')  
		    -> url($IU::HomeUrl)
              -> _UL;

	# Output debugging info:
	$self->print_debugging_info(Traceback=>$fullmsg, Inline=>1, 
                                    Errno=>$errno) 
	    unless ($short_form);
    }
		    
    # The signature...
    $HTML -> HR -> ADDRESS -> t("$IU::SystemName error trace") -> _ADDRESS;
    $HTML -> _FONT -> _BODY -> _HTML;

    # At this point, we've prepared all the output to the pagebuffer.
    # But we may want to email the output in addition to just sending it
    # to the client! 
    if (!$benign_error && (uc($IU::EmailGasp) eq 'YES')) {
	eval {
	    require IU::Mail;
	    local($/) = "\n";       # KLUDGE: needed for $GASP... but why?

	    # Create mail object:
	    my $mail = new IU::Mail;
	    $mail->to($IU::SiteAdminEmail || 'imswww');   # try hard
	    $mail->subject("Dying gasp: " . 
			   ($SESSION ? $SESSION->id : '(no session)'));
	    $mail->set('Content-Type', 'text/html');
	    
	    # Rewind pagebuffer, and skip contents until first blank line:
	    my $GASP = $REPLY->rewind_pagebuffer || die "no pagebuffer!";
	    while (<$GASP>) { last if /^\s*$/; }
	    
	    # Write contents:
	    my $FH = $mail->open || die "mail open failed!";	
	    while (<$GASP>) { print $FH $_ };
	    $mail->send;

	    # Log, if we can (limit line length to about 80 characters):
	    my $sid = ($SESSION ? $SESSION->id : $REQ->param('sid'));
	    require IU::Event;
	    &IU::Event::ims_event('DYINGGASP', {SESSID=>$sid},
				  'ERRNO'  => $errno,
				  'ERRSTR' => $errstr,
				  'ERRMSG' => substr($errmsg, 0, 80));
	};
    }

    # If we're not in debug mode, trim off all the logs and stuff we
    # don't want just anybody seeing.
    if (!$self->debug_mode) {
      $REPLY->truncate_pagebuffer();
      # The signature...
      $HTML -> HR -> ADDRESS -> t("$IU::SystemName secured error trace") 
                  -> _ADDRESS;
      $HTML -> _FONT -> _BODY -> _HTML;
    }

    # DONE!
    $REPLY->all_done;
    exit(-1);
}

#------------------------------------------------------------
# print_debugging_info PARAMHASH
#------------------------------------------------------------
# Just output the death rattle without a header... but I<don't>
# do any death processing.
#
# This is to allow the trace information be output after the main
# document in debugging mode.
#
# Params are:
#
#     Traceback => the traceback message, explaining the error in detail
#     Inline    => show files inline (default is just to provide hyperlinks)
#     Errno     => error number; used to include the debug log for 
#                  certain errors (see below)

sub print_debugging_info {
    my $self = shift;
    my %params = @_;
    my $traceback = $params{Traceback};
    my $inline    = $params{Inline};
    my $errno     = $params{Errno};

    # Add any error numbers for which the debug log should still be 
    # printed in "secured" mode. IWe'll index this by the error number
    # below to reset the truncation point.
    my %print_log_for = ( 
                          602 => 1 
                        );

    # Start small font:
    $HTML -> FONT(SIZE=>-1);

    # The error...
    $HTML -> HR -> H2 -> t("Traceback") -> _H2;
    $HTML -> PRE -> t($traceback) -> _PRE;

    # The CGI query, if available...
    if ($REQ) {	
	$HTML -> HR -> H2 -> t("CGI interface") -> _H2;
	my ($param, $val);

	# Raw Query...
	$HTML -> H3 -> t("Original Query") -> _H3;
	$HTML -> PRE;
	foreach $param (sort keys %RawQuery) {
	    my @vals = sort @{$RawQuery{$param} || []};
	    foreach $val (@vals) {
		if ($param =~ /^_/) {    # leading underscore means sensitive!
		    $HTML -> t("$param = ") -> B -> t('CONCEALED') -> _B ->nl;
		}
		else {	                 # okay, print it:
		    $HTML -> t(qq{$param = "$val"\n}); 
		}
	    }
	}
	print "None." if (!(keys %RawQuery));
	$HTML -> _PRE;

	# Diddled query...
	$HTML -> H3 -> t("Virtual Query") -> _H3;
	$HTML -> PRE;
	foreach $param (sort $REQ->params) {
           if (defined  $REQ->param($param)) {
	      my @vals = sort $REQ->param($param);
	      foreach $val (@vals) {
	         if ($param =~ /^_/) {    # leading underscore means sensitive!
	            $HTML -> t("$param = ") -> B -> t('CONCEALED') 
                                               -> _B ->nl;
		 }
		 else {	                 # okay, print it:
		    $HTML -> t(qq{$param = "$val"\n}); 
                 }
              }
	   }
           else {
              $HTML -> t(qq($param = undef\n));
           }
	}
	print "None." if (!($REQ->params));
	$HTML -> _PRE;

        $REPLY->set_truncation_point();
        debug "truncation point set to ". $REPLY->{TruncationPoint};

	# Interface variables...
	$HTML -> H3 -> t("Interface variables") -> _H3;
	$HTML -> PRE;
	foreach $param (sort $REQ->cgi->vars) {
	    defined($val = $REQ->cgi->var($param)) or $val = '';
	    output $HTML "$param = ", '"', $val, '"', "\n";
	}

#         my($key1,$value1);
#	 foreach $key1 (sort (keys %ENV)) {
# defined($value1 = %ENV{$key1}) or $value1 = '';
#	($value1 = $ENV{$key1}) ;
#            output $HTML qq{$key1 = "$value1"\n};
#	}


	$HTML -> _PRE;
    }

    # The log...
    $HTML -> HR -> H2 -> t("Log file: this request") -> _H2;
    $HTML -> PRE;
    print "Load begin: ", ($::TIMES{'Load begin'} || ''), "\n";
    print "Load end:   ", ($::TIMES{'Load end'} || ''), "\n";
    print "Time now:   ", scalar(localtime), "\n";
    print "\n";
    if ($LOG) {
	$LOG->flush;            # flush any pending output
	$LOG->seek(0, 0);       # seek to start
	while (<$LOG>) { 
      	    output $HTML $_; 
	}
    }
    $HTML -> _PRE;

    if ($print_log_for{$errno}) {
        $REPLY->set_truncation_point();
        debug "truncation point REset to ". $REPLY->{TruncationPoint};
    }

    # Other logs...
    #    Notice that we only output the text if this was a DYING GASP:
    #    if we're just debugging, we provide a HYPERLINK instead.  This makes
    #    debugging less painful on a slow connection...
    if (ref($SESSION) eq 'IU::Session') {
	my $file;	
	foreach $file ('search/runclient.log', 'search/syslog', 'gdbdump.syslog', 'pub/query.dat') {

	    # Only if the file exists...
	    my $path = $SESSION->path_to($file);
	    (-e $path) or next;

	    # Output it!
	    if ($inline) {     # In-line the file...
		open LOGFILE, $path or next;
		$HTML -> HR -> H3 -> t("Log file: $file") -> _H3;
		print "Last modified: ", scalar(localtime((stat(LOGFILE))[9]));
		$HTML -> PRE;
		while (<LOGFILE>) { output $HTML $_; }
		$HTML -> _PRE;
		close LOGFILE;
	    }
	    else {             # Just provide hyperlink to the file...
		my $symlink = $SESSION->path_to("pub/$file");
		(-l $symlink) or (symlink $path, $symlink) or next;
		$HTML -> HR -> H3 -> hypertext("Log file: $file",
					       $SESSION->url("pub/$file"))
			    -> _H3;
	    }
	}
    }

    # The setup...
    $HTML -> HR -> H2 -> t("$IU::SystemName setup") -> _H2;
    $HTML -> PRE;
    if (defined(@IU::SETUP::EXPORT)) {
	foreach (@IU::SETUP::EXPORT) {
	    next if /^\$AdminCrypt$/;            # KLUDGE!!!
	    my ($var) = ($_ =~ /^\$(.*)/);       # chop off leading '$'
	    output $HTML qq{$_ = "}, eval "\$IU::SETUP::$var", qq{"\n};
	}
    }
    $HTML -> _PRE;

    # The run-time environment...
    $HTML -> HR -> H2 -> t("Run-time environment") -> _H2;

    # The session info, if available...
    if ( (ref($SESSION) eq 'IU::Session') && $SESSION->state) {
	$HTML -> H3 -> t("Session state") -> _H3;
	$HTML -> PRE;
	my $key;
	foreach $key (sort keys %{$SESSION->state}) {
	    my $value = $SESSION->state->{$key};
	    $value = defined($value) ? qq{"$value"} : 'undef';
	    output $HTML qq{$key = $value\n}; 
	}
	$HTML -> _PRE;
    }

    $HTML -> H3 -> t("Environment variables") -> _H3;
    $HTML -> PRE;
    foreach (sort keys %ENV) { output $HTML qq{$_ = "$ENV{$_}"\n} }    
    $HTML -> _PRE;

    $HTML -> H3 -> t("Permissions") -> _H3;
    $HTML -> PRE;
    output $HTML "Real user id = ", $<, "\n";
    output $HTML "Real group ids = ", $(, "\n";
    output $HTML "Effective user id = ", $>, "\n";
    output $HTML "Effective group ids = ", $), "\n";
    $HTML -> _PRE;

    $HTML -> H3 -> t("Included modules") -> _H3;
    $HTML -> t("A leading 'P' means that the module was pre-loaded by ",
	       "the custom HTTPD (and incurred no request-time overhead ",
	       "for compilation).");
    $HTML -> PRE;
    my $bytes = 0;
    my $inc;
    foreach $inc (sort keys %INC) { 
	($inc && $INC{$inc}) or next;
	$bytes += (-s $INC{$inc});
	output $HTML (sprintf("%s   %-20s => %6d bytes: %s\n", 
			     ($::PRELOADED_INC{$inc} ? 'P' : ' '), 
			     $inc, (-s $INC{$inc}), $INC{$inc}));
    }
    output $HTML '-' x 70, "\n";
    output $HTML sprintf("%-20s => %6d bytes, %s files\n",
			 'TOTAL', $bytes, int(keys(%INC)));
    $HTML -> _PRE;

    # The perl setup...
    $HTML -> HR -> H2 -> t("Perl setup") -> _H2;
    $HTML -> PRE;
    output $HTML "Script path = $0\n";
    output $HTML "Perl version = $]\n";
    $HTML -> _PRE;

    $HTML -> H3 -> t("Include path") -> _H3;
    $HTML -> PRE;
    foreach (@INC) { output $HTML "$_\n" }
    $HTML -> _PRE;

    $HTML -> H3 -> t("Configuration") -> _H3;
    $HTML -> PRE;
    output $HTML `$^X -V`;
    $HTML -> _PRE;

    # The filesystem...
    $HTML -> HR -> H2 -> t("Operating system") -> _H2;

    $HTML -> H3 -> t("Disk usage") -> _H3;
    $HTML -> PRE;
    output $HTML `df`;
    $HTML -> _PRE;

    # Enough for now!
    1;
}

#------------------------------------------------------------
1;

