package IW::CGI;

=head1 NAME

IW::CGI - CGI interface for the EOSDIS V0 scripts

=head1 DESCRIPTION

Provides almost-drop-in replacements for CGI::BasePlus and CGI::Request,
called IW::CGI::BasePlus and IW::CGI::Request.  Only a few methods are 
supported.

In IW::CGI::BasePlus...

    $cgi = new IW::CGI::BasePlus; 
    
    @vars = $cgi->vars;
    $var = $cgi->var('HTTP_USER_AGENT');

In IW::CGI::Request...

    $req = new IW::CGI::Request;
    
    $cgi = $req->cgi;
    
    @params = $req->params;
    @values = $req->param('paramname');
    $value  = $req->param('paramname');
    
    $req->extract_values('param1=value1&param2=value2&param3=value3');

=head1 NOTES

Derived from CGI_Lite, but with even fewer calories.

=cut


# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =

package IW::CGI::BasePlus;

# Fix for broken URI::Escape 3.6 (grrrrr):
my $UESC = '^A-Za-z0-9';

use FileHandle ();
use strict qw(vars subs);        # no refs! it breaks file upload!

# Stuff for parsing multipart postings...
my @Buffer = ();     # array of data read, each ending in nl except maybe last
my $BytesLeft = 0;
my $CHUNKSIZE = 4096;   # how much to read at each go
my $TIMEOUT = 10;       # seconds to read
my $LINELIMIT = 4096;   # lines longer than this are broken 

# My self (only one allowed!)
my $SELF;

#------------------------------
# new
#------------------------------
sub new {
    my $class = shift;
    return $SELF if $SELF;

    # Create interface object:
    $SELF = {};
    bless $SELF, $class;
    
    # Read in data:
    $SELF->parse_form_data;
    $SELF;
}

#------------------------------
# parse_form_data
#------------------------------
# Just puts the stuff into "QS"!
sub parse_form_data {
    my ($self) = @_;
    
    my $request_method = $ENV{'REQUEST_METHOD'};
    my $content_length = $ENV{'CONTENT_LENGTH'};
    my $content_type   = $ENV{'CONTENT_TYPE'};
    my $query_string = '';

    $content_type   = lc($content_type);

    # For get and head, just get and store the query string:
    if ($request_method =~ /^(get|head)$/i) {
	$self->{QS} = $ENV{'QUERY_STRING'};
    } 

    # Post. Could be long, or could be multipart. Decide.
    elsif ($request_method =~ /^post$/i) {

        # Just a regular old post. Read it, set the query string.
	if ($content_type =~ m{^application/x-www-form-urlencoded\b}) {
	    read STDIN, $query_string, $content_length;
	    $self->{QS} = $query_string;
	} 

        # A multipart post. We'll have to parse it.
	elsif ($content_type =~ m{^multipart/form-data\b}) {
	    my $boundary;
	    $content_type =~ /boundary=((\"([^\"]+)\")|(\S+))$/;
            $boundary = ($3 ? $3 : $1);	 
	    $self->parse_multipart($content_length, $boundary);
	} 
	else {
	    die "MIME type '$content_type' unsupported for POST.";
	}
    }
    else {
	die "method '$request_method' unsupported.";
    }
    1;
}

#------------------------------
# vars
#------------------------------
sub vars {
    keys %ENV;
}

#------------------------------
# var
#------------------------------
sub var {
    my ($self, $var) = @_;
    $ENV{$var};
}

#------------------------------
# parse_multipart TOTALBYTES,BOUNDARY
#------------------------------

sub parse_multipart {

    # Setup:
    require URI::Escape;
    import URI::Escape;

    my($self,$length,$boundary) = @_;
    my($buffer) = IW::CGI::MultipartBuffer->new($boundary,$length);
    my(%header,$body);
    while (!$buffer->eof) {
	%header = $buffer->readHeader;
	# In beta1 it was "Content-disposition".  
        # In beta2 it's "Content-Disposition"
	# Sheesh.
	my($key) = $header{'Content-disposition'} ? 
                       'Content-disposition' : 'Content-Disposition';

	my($param) =    $header{$key}=~/ name="(.*?)"/;
	my($filename) = $header{$key}=~/ filename="(.*?)"/;

	my($value);

	if ($filename) {
	    # If we get here, then we are dealing with a potentially large
	    # uploaded file.  Save the data to a temporary file, then open
	    # the file for reading, and stash the filehandle name inside
	    # the query string.
	    my($tmpfile) = IW::CGI::TempFile->new();
	    my $tmp = $tmpfile->as_string;

	    open (OUT,">$tmp") || die "CGI open of $tmpfile: $!\n";
	    chmod 0666,$tmp;	# make sure anyone can delete it.
	    binmode(OUT);

	    my $data;
	    while ($data = $buffer->read) {
                # Clean up possible DOSisms
                $data =~ s/\cM\cJ/\n/g;
                $data =~ s/\cZ//;
                # Clean up possible Macisms
                $data =~ tr/\cM/\n/;	
		print OUT $data;
	    }
	    close OUT;

	    # Now create a new filehandle in App's namespace.
	    my($filehandle);
	    my $newFH = "IW::App::$filename";
	    open($newFH,$tmp) || die "CGI open of $tmpfile: $!\n";
	    binmode($newFH);
	    $value = $filename;

	    # Under Unix, it is safe to let the temporary file be deleted
	    # when it goes out of scope.  The storage is not deallocated
	    # until the last file descriptor is closed.  So we do nothing
	    # special here.
	}

	# If we get here then we're dealing a non-file form field, which we
	# will assume can fit into memory OK.
	else {
	    $value = $buffer->readBody;
	}

	# Now we store the parameter name and the value into our
	# query string for later retrieval
	$CGI::Base::QUERY_STRING .= '&' if $CGI::Base::QUERY_STRING;
	$CGI::Base::QUERY_STRING .= uri_escape($param) . '=' . uri_escape($value);
	$self->{QS} .= uri_escape($param, $UESC) . '=' 
	             . uri_escape($value, $UESC) . '&';
    }
    1;
}


# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =

package IW::CGI::Request;

use URI::Escape;

# The interface object:
my $CGI;

#------------------------------
# Interface
#------------------------------
# Set the interface object to use.  Do this before calling new()!
sub Interface {
    $CGI = shift;
}

#------------------------------
# new
#------------------------------
# Create a new request object, and parse the user's request.
sub new {
    my $class = shift;

    # Create interface, if necessary:
    unless ($CGI) {
	$CGI = new IW::CGI::BasePlus or return undef;
    }

    # Create request object:
    my $self = { 
	CGI => $CGI,
	Params => {},
    };
    bless $self, $class;    

    # Extract values:
    $self->extract_values($CGI->{QS});
    $self;
}

#------------------------------
# cgi
#------------------------------
# Return the CGI interface object.
sub cgi {
    $_[0]->{CGI};
}

#------------------------------
# delete
#------------------------------
# Remove the specified parameter.  We need this since param() (as defined
# by CGI::Request) doesn't allow setting to the empty list.  >:-(
sub delete {
    my ($self, $param) = @_;
    return $self if(!defined($param));
    delete $self->{Params}{$param};
}

#------------------------------
# params
#------------------------------
# Return array of all query params.
sub params {
    keys %{$_[0]->{Params}};
}

#------------------------------
# param PARAMNAME VALUES...
#------------------------------
# Return the value of parameter PARAMNAME.
# If one or more VALUES are given, sets those values first.
sub param {
    my ($self, $param, @vals) = @_;

    # Set:
    if (@vals) {
	$self->{Params}{$param} = \@vals;
    }

    # Return value(s):
    my $vals = [];
    $vals = ($self->{Params}{$param} || []) if(defined($param));
    return (wantarray ? @$vals : $vals->[0]);
}

#------------------------------
# extract_values
#------------------------------
# Extract values from a query string.  Lifted from CGI::Request.
sub extract_values {
    my ($self, $query) = @_;
    
    return () unless defined $query;

    my (@parts, $key, $val);
    my $cgi = $self->cgi;
   
#   IU::Log::inform("Caller: ". join(" ",caller()));
    IU::Log::inform("query = $query");


    # Some times the msubmit query string from Netscape users on NT machines 
    # have questions marks "?" as delimeters instead of "&".
    # Don't know why this is nor have we been able to reproduce it, so we 
    # are assuming that it is bug in the browser that is causing this to happen.
    # Fix up query, if there are any ? in the query string replace them with &s.
    $query =~ s/\?/&/g;
   
    if ($query =~ m/=/ or $query =~ m/&/) {
	$query =~ tr/+/ /;	# RFC1630

        # Handle Mac IE 5 problem:
        #
        # Someone at Microsoft, with a remarkable (OK, not so remarkable, with
        # Microsoft involved) lack of insight, decided that the kind of stuff 
        # we were exploiting to make the msubmit tags work wasn't covered by 
        # the HTTP or HTML specs, so they removed it in Mac IE5.
        #
        # The result was that msubmit got well and truly broken. Instead of
        # getting the nice "button.x" and "button.y" stuff, we now get two
        # NULL (yep, no variable name) variables with a numeric value. And no
        # X or Y either. How nice.
        # 
        # However, the tags DO happen to occur in the outgoing stream exactly
        # at the point where the button that was clicked was in the input 
        # stream. This allows us to put a hidden field directly after each 
        # button and to be able to associate the click (the two null 
        # parameters) with the hidden field immediately following the button.
        #
        # So, in CGI.pm, we remove the two null parameters (but hang onto 
        # them), and reinsert them based on the name of the next field coming
        # along (which is the hidden field we inserted when the button was
        # created).
        #
        # It is NOT pretty, and it is certainly a hack, but msubmits were 
        # pretty much a hack in the first place. we can only hope that 
        # Netscape doesn't decide to break things in a worse way.
        #
        # We use a very funky substitution (note the /x modifier that
        # allows us to spread it out and make it more understandable!).
        # The second part of the substitution is really a string, so
        # we can't spread it out pretty like the first part.
        #
        # Note that the suffixes must be .x and .y, NOT ".X" or ".Y",
        # or a number of other modules will break.

        $query =~ 
        s{
          (&|^)         # Either in the middle or at the beginning,
          =(.*?)        # catch a null parameter, and save its value.
          &=(.*?)       # Do it twice.
          &(.*?)=       # Next parameter (hidden)
         }
         {$1$4.x=$2&$4.y=$3&$4=}x;
         IU::Log::inform("$1$4.x=$2&$4.y=$3&$4=") if $1;

        # The substituted string decodes to:
        #  - the first field separator (or beginning of line)
        #  - the hidden field's name.x = the first null parameter's value
        #  - the hidden field's name.y = the second null parameter's value
        #  - the hidden field and its value (the msubmit string)

        # We now return you to your regularly scheduled string parsing.
	@parts = split(/&/, $query);
	foreach (@parts) {      # Extract into key and value.
	    ($key, $val) = m/^(.*?)=(.*)/;
	    $val = (defined $val) ? uri_unescape($val) : '';
	    $key = uri_unescape($key);

	    # Store as a list of values. Push new value on the end.
	    $self->{Params}{$key} or $self->{Params}{$key} = [];
	    push @{$self->{Params}{$key}}, $val;
	}
    } 
    else {	# no '=' or '&' implies ISINDEX so split on +'s
	@parts = split(/\+/, $query);
	grep { $_ = uri_unescape($_) } @parts;
	$self->param('KEYWORDS', @parts);

	# spot image maps and extract X and Y coords
	if (@parts == 1 and $parts[0] =~ m/^(\d+),(\d+)$/) {
	    $self->param('X', $1);
	    $self->param('Y', $2);
	}
    }
    @parts;
}

# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =

package IW::CGI::MultipartBuffer;

use vars qw($FILLUNIT $TIMEOUT $SPIN_LOOP_MAX $CRLF);

# How many bytes to read at a time.  We use  a 5K buffer by default.
$FILLUNIT = 1024 * 5;		
$TIMEOUT = 10*60;       # 10 minute timeout
$SPIN_LOOP_MAX = 1000;  # bug fix for some Netscape servers
$CRLF = "\015\012";

#------------------------------
# new
#------------------------------
# Record total length of buffer, determine the boundary, and set up an empty
# buffer.

sub new {
    my($package,$boundary,$length,$filehandle) = @_;
    my $IN;
    if ($filehandle) {
	my($package) = caller;
	# force into caller's package if necessary
	$IN = $filehandle=~/[':]/ ? $filehandle : "$package\:\:$filehandle"; 
    }
    $IN = "main::STDIN" unless $IN;
    binmode($IN);

    # Netscape seems to be a little bit unreliable about providing boundary 
    # strings.
    if ($boundary) {
	# Under the MIME spec, the boundary consists of the 
	# characters "--" PLUS the Boundary string
	$boundary = "--$boundary";
	# Read the topmost (boundary) line plus the CRLF
	my($null) = '';
	$length -= read($IN,$null,length($boundary)+2,0);
    } else { # otherwise we find it ourselves
	my($old);
	($old,$/) = ($/,$CRLF); # read a CRLF-delimited line
	$boundary = <$IN>;      # BUG: This won't work correctly under mod_perl
	$length -= length($boundary);
	chomp($boundary);               # remove the CRLF
	$/ = $old;                      # restore old line separator
    }

    my $self = {LENGTH=>$length,
		BOUNDARY=>$boundary,
		IN=>$IN,
		BUFFER=>'',
	    };

    $FILLUNIT = length($boundary) if length($boundary) > $FILLUNIT;

    return bless $self,$package;
}

#------------------------------
# readHeader
#------------------------------
# This reads and returns the header as an associative array.
# It looks for the pattern CRLF/CRLF to terminate the header.
sub readHeader {
    my($self) = @_;
    my($end);
    my($ok) = 0;
    do {
	$self->fillBuffer($FILLUNIT);
	$ok++ if ($end = index($self->{BUFFER},"${CRLF}${CRLF}")) >= 0;
	$ok++ if $self->{BUFFER} eq '';
	$FILLUNIT *= 2 if length($self->{BUFFER}) >= $FILLUNIT; 
    } until $ok;

    my($header) = substr($self->{BUFFER},0,$end+2);
    substr($self->{BUFFER},0,$end+4) = '';
    my %return;
    while ($header=~/^([\w-]+): (.*)$CRLF/mog) {
	$return{$1}=$2;
    }
    return %return;
}

#------------------------------
# readBody
#------------------------------
# This reads and returns the body as a single scalar value. Used when we 
# determine that we're not getting a file download.
sub readBody {
    my($self) = @_;
    my($data);
    my($returnval)='';
    while (defined($data = $self->read)) {
	$returnval .= $data;
    }
    return $returnval;
}

#------------------------------
# read
#------------------------------
# This will read $bytes or until the boundary is hit, whichever happens
# first.  After the boundary is hit, we return undef.  The next read will
# skip over the boundary and begin reading again.
sub read {
    my($self,$bytes) = @_;

    # default number of bytes to read
    $bytes = $bytes || $FILLUNIT;       

    # Fill up our internal buffer in such a way that the boundary
    # is never split between reads.
    $self->fillBuffer($bytes);

    # Find the boundary in the buffer (it may not be there).
    my $start = index($self->{BUFFER},$self->{BOUNDARY});

    # If the boundary begins the data, then skip past it
    # and return undef.  The +2 here is a fiendish plot to
    # remove the CR/LF pair at the end of the boundary.
    if ($start == 0) {

	# clear us out completely if we've hit the last boundary.
	if (index($self->{BUFFER},"$self->{BOUNDARY}--")==0) {
	    $self->{BUFFER}='';
	    $self->{LENGTH}=0;
	    return undef;
	}

	# just remove the boundary.
	substr($self->{BUFFER},0,length($self->{BOUNDARY})+2)='';
	return undef;
    }

    my $bytesToReturn;    
    if ($start > 0) {           # read up to the boundary
	$bytesToReturn = $start > $bytes ? $bytes : $start;
    } else {    # read the requested number of bytes
	# leave enough bytes in the buffer to allow us to read
	# the boundary.  Thanks to Kevin Hendrick for finding
	# this one.
	$bytesToReturn = $bytes - (length($self->{BOUNDARY})+1);
    }

    my $returnval=substr($self->{BUFFER},0,$bytesToReturn);
    substr($self->{BUFFER},0,$bytesToReturn)='';
    
    # If we hit the boundary, remove the CRLF from the end.
    return ($start > 0) ? substr($returnval,0,-2) : $returnval;
}

#------------------------------
# fillBuffer
#------------------------------
# This fills up our internal buffer in such a way that the boundary is never 
# split between reads.
sub fillBuffer {
    my($self,$bytes) = @_;
    return unless $self->{LENGTH};

    my($boundaryLength) = length($self->{BOUNDARY});
    my($bufferLength) = length($self->{BUFFER});
    my($bytesToRead) = $bytes - $bufferLength + $boundaryLength + 2;
    $bytesToRead = $self->{LENGTH} if $self->{LENGTH} < $bytesToRead;

    # Try to read some data.  We may hang here if the browser is screwed up.  
    my $bytesRead = CORE::read($self->{IN},$self->{BUFFER},
                               $bytesToRead,$bufferLength);

    # An apparent bug in the Netscape Commerce server causes the read()
    # to return zero bytes repeatedly without blocking if the
    # remote user aborts during a file transfer.  I don't know how
    # they manage this, but the workaround is to abort if we get
    # more than SPIN_LOOP_MAX consecutive zero reads.
    if ($bytesRead == 0) {
	die  "CGI::BasePlus: Server closed socket during " .
             "multipart read (client aborted?).\n"
	         if ($self->{ZERO_LOOP_COUNTER}++ >= $SPIN_LOOP_MAX);
    } else {
	$self->{ZERO_LOOP_COUNTER}=0;
    }

    $self->{LENGTH} -= $bytesRead;
}

#------------------------------
# eof
#------------------------------
# Return true when we've finished reading - that is, that we really have 
# nothing left to read.
sub eof {
    my($self) = @_;
    return 1 if (length($self->{BUFFER}) == 0)
		 && ($self->{LENGTH} <= 0);
}

# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =

# This is a temporary fix; we really want to use FileHandle here.
# De-hack this so it uses the session directory as it should.

package IW::CGI::TempFile;

use vars qw(@TEMP $TMPDIRECTORY $SEQUENCE);

@TEMP=('/usr/tmp','/var/tmp','/tmp',);
unshift(@TEMP,$ENV{TMPDIR}) if defined($ENV{TMPDIR});

foreach (@TEMP) {
    do {$TMPDIRECTORY = $_; last} if -w $_;
}
$TMPDIRECTORY  = "." unless $TMPDIRECTORY;
$SEQUENCE="CGItemp${$}0000";

# Create a temporary file that will be automatically
# unlinked when finished.
sub new {
    my($package) = @_;
    $SEQUENCE++;
    my $directory = "${TMPDIRECTORY}/${SEQUENCE}";
    return bless \$directory;
}

sub DESTROY {
    my($self) = @_;
    unlink $$self;		# get rid of the file
}

sub as_string {
    my($self) = @_;
    return $$self;
}

#------------------------------------------------------------
1;

