package IW::CartList;

=head1 NAME

IW::CartList - utilities for outputting an orderable item list in shopping cart.


=head1 DESCRIPTION


See IW::ObjectList for details.


=head1 RCS INFO

$Id: CartList.pm,v 1.34 2003/09/09 18:18:42 solomon Exp $

=cut

use IU;
use IU::Log;
use IU::Granule;
use IU::PkgCombo;
use IU::WWWValids;

use IW::User;
use IW::Date qw(TZ_to_datetime);
use IW::ObjectList;
use IW::Util qw(:url);
use IW::App;
use IW::Sort;
#use IW::Babel;

@ISA = qw(IW::ObjectList);

use strict;

#------------------------------------------------------------
#
# GLOBALS
#
#------------------------------------------------------------

# Order to present things:
my @Order;

my $someDemand; #Flag for presence of ASTER On-Demand product

# Translating a granule's 'carted' attribute to a status message:
my %CartStatusMsg = (
    'CAN'    => 'Can be ordered',
    'CANNOT' => "Can't be ordered here",
    'YES'    => 'In shopping cart',
    'JUST'   => 'Just added to cart'
);

# Translating a granule's 'carted' attribute to a status message:
my %CartStatusIcon = (
    'CAN'    => 'icons/order-can.gif',
    'CANNOT' => 'icons/order-cannot.gif',
    'YES'    => 'icons/order-yes.gif',
    'JUST'   => 'icons/order-just.gif'
);

#------------------------------------------------------------
# new PARAMHASH
#------------------------------------------------------------
# Create a new cart list object.
# PARAMHASH contains everything normal to superclass, plus:
#
#     Origin   => origin of these objects: 'CART' or 'RESULTS'.

sub new {
    my $class = shift;
    my %params = @_;

    # Construct:
    my $self = $class->SUPER::new(@_);    

    # Cart is currently using the same attribute columns as the 
    # granule list.
 
    @Order = @{$self->{Display}->granule_summary};

    # Set up defaults, etc:
    $self->{Origin}   or $self->{Origin} = '';
    $self->{NoteFunc} or $self->{NoteFunc} = \&note;
    $self->{Note2Func} or $self->{Note2Func} = \&note2;
    $self->{GraphFunc} or $self->{GraphFunc} = \&graph;
    $self->{Graph2Func} or $self->{Graph2Func} = \&graph2;
    $self;
}

#------------------------------------------------------------
# origin
#------------------------------------------------------------
# Get/set the origin of these granules.

sub origin {
    my ($self, $value) = @_;
    $value and $self->{Origin} = $value;
    $self->{Origin};
}

#------------------------------------------------------------
# sort [SORTREF]
#------------------------------------------------------------
# Sort our list of granules.  Returns a reference to the newly-sorted array.
#
# B<SIDE EFFECTS:> This sets the session-state variable 'LastGranuleSort' 
# to the sort option used, so that it is used/presented as a default for 
# subsequent granule lists.

sub sort {
    my ($self, $sortref) = @_;

    # Figure out the sort algorithm...
    my $sorter = IW::Sort::Granule->appropriate;
    
    # Construct a nice string to describe this:
    $$sortref = $sorter->as_string      if $sortref;
    
    # Do it!    
    my @sorted = $sorter->sort($self->{Elems});
    $self->{Elems} = \@sorted;

    # Make a note in the session state:
    $SESSION->state->{LastGranuleSort} = $sorter;
    $self->{Elems};
}

#------------------------------------------------------------
#
# PRIVATE METHODS/UTILITIES
#
#------------------------------------------------------------

#------------------------------------------------------------
# check_for_some HASH
#------------------------------------------------------------

sub check_for_some {

   my $self = shift;
   my %some;

   %some = map {$_=>0} @Order;
   $some{'SIZEMBECSDATAGRANULE'} = '1';
   $some{'DATA_CENTER_ID'} = '1';

   my $title;
   my $item;
   foreach $item (@{$self->{Elems}}) {
      # As data is found for a column it is removed from the loop,
      # gradually the loop becomes shorter and shorter.
      # If the first item has values for all the columns
      # then this will run only once.
      foreach $title (grep {!$some{$_}} (keys %some) ) {
         if ($title eq 'PATH' and defined($item->coverage) and
            scalar (@{$item->coverage->rows})) {
            $some{$title} = '1';
         }
         elsif ($title eq 'ROW' and defined($item->coverage) and
                scalar (@{$item->coverage->rows})) {
            $some{$title} = '1';
         }
         elsif ($title eq 'START_DATE') {

            $some{$title} = '1' if (defined ($item->start_date)
                                    and $item->start_date ne '(none)');
         }
         elsif ($title eq 'STOP_DATE') {

            $some{$title} = '1' if (defined ($item->stop_date)
                                    and $item->stop_date ne '(none)');
         }
      } # foreach $title.

      # On-Demand is present
      $someDemand = '1' if ($item->name);
   } # foreach $item

   return %some;
}

#------------------------------------------------------------
# print_elems();
#------------------------------------------------------------

sub print_elems {
   my ($self, $start_num, $end_num, $some_notes, $some_pkgs_chosen,
        $some_orderable, $some_comments, $some) = @_;

   my $i = 0;                 # Counter of items. 0 based.

   # First print the items that have order selection.
   for ($i = $start_num-1; $i < ($end_num); $i++) {
      next if ($self->{Elems}[$i]->order_status eq "NEED_PKGS");
      $self->print_row ($i, $some_notes, $some_pkgs_chosen,
                        $some_orderable, $some_comments, $some);
   }
   $i = 0;
   # Second print the items that do not have order selection.
   for ($i = $start_num-1; $i < ($end_num); $i++) {
      next if ($self->{Elems}[$i]->order_status ne "NEED_PKGS");
      $self->print_row ($i, $some_notes, $some_pkgs_chosen,
                        $some_orderable, $some_comments, $some);
   }
}

#------------------------------------------------------------
# print_row();
#------------------------------------------------------------

sub print_row {
   my ($self, $i, $some_notes, $some_pkgs_chosen,
        $some_orderable, $some_comments, $some) = @_;

   print '<tr align=left>';

   # Print the menu buttons in table cells:
   $self->print_buttons($i, $some_notes, $some_pkgs_chosen,
                           $some_orderable, $some_comments);
   $self->print_data($some, $i);

   print "</tr>\n";

   # put listed items as hidden tags so we can figure out what the users
   # selected or unselected.
   $HTML->hidden('keys',$self->{Elems}[$i]->key);
}

#------------------------------------------------------------
# print_data
#------------------------------------------------------------

sub print_data {
   my $self = shift;
   my $some = shift;
   my $i = shift;

   my $item = $self->{Elems}[$i];
   my $type = ($item->order_type eq 'GRANULE') ? 'gkey' : 'dskey';
   my $key  = $item->key;
   my $status = $item->order_status;
   my @key_array = ($type, $key);

   #
   # Added different background color to know which items have order
   # package selected.  If item needs a package, the backgroud is
   # bright yellow, otherwise the bacground will be the default.
   #
   my $lite_gray = "#EEEEEE";
   my $bgcolor = $lite_gray;
   
   # Print subsetting link if available or print "Subsetting unavailable".
   my $order_opts = $item->some_options('ORDER');
   my $subset_opts = $item->some_options('SUBSET');
   my $dataset_opts = $item->some_options('DATASET_ORDER');
   $HTML -> TD(BGCOLOR=>$bgcolor);
   if($order_opts || $subset_opts || $dataset_opts) {
      $HTML -> t("Available");
   } else {
      $HTML -> t("Unavailable");
   }
   $HTML -> _TD;

   # ---------------------------------------------------------
   # Always print item column:
   # ---------------------------------------------------------
   # Size will always be printed, we want to make sure that if the
   # size is zero or we don't have it, we print unavailable.
   #
   my $item_size = $self->set_item_size ($item);
   $item_size = (0+$item_size) ? $item_size : "Unavailable";

   $HTML->TD(BGCOLOR=>$bgcolor)->CENTER;
   # Print granule ID:
   my $id = $item->id || "\xA0";
   $HTML->t_B($id)->BR;
   # Print local ID if present:
   my $local = $item->local_granule_ID;
   print "($local)" if (defined($local) and ($local ne ''));
   $HTML->_CENTER->_TD;
#   print "<td bgcolor=$bgcolor>", $item->id || "\xA0";
#   print '</td>';

   # Print the on demand title
   print "<td bgcolor=$bgcolor>", $item->name || "\xA0", '</td>'
    if($someDemand);
   print "<td bgcolor=$bgcolor>", $item_size || "Unavailable", '</td>';
#   print "<td bgcolor=$bgcolor>", $item->archive_id || "\xA0", '</td>';

   # Only do those columns which have some=1;
   foreach my $title (grep {$some->{$_}} (@Order) ) {

      if ($title eq 'PATH') {

         print "<td bgcolor=$bgcolor>", 
            join ('-',@{$item->coverage->paths}) || "\xA0", '</td>';
      }
      elsif ($title eq 'ROW') {

         my @rows = @{$item->coverage->rows};
         my $string = ($rows[0] eq $rows[1]) ? $rows[0] : join ('-',@rows);
         print "<td bgcolor=$bgcolor>", $string || "\xA0", '</td>';
      }
      elsif ($title eq 'START_DATE') {

         print "<td bgcolor=$bgcolor>", 
            TZ_to_datetime($item->start_date) || "\xA0", '</td>';
      }
      elsif ($title eq 'STOP_DATE') {

         print "<td bgcolor=$bgcolor>", 
            TZ_to_datetime($item->stop_date) || "\xA0", '</td>';
      }
   }

   # "Selection" checkbox... notice, we check for "select all"!
   print "<td align=center bgcolor=$bgcolor>";
   $HTML -> checkbox(Name=>"$type", Choice=>$item->key, Value=>'');
   print "</td>";

}

#------------------------------------------------------------
# set_item_size 
#------------------------------------------------------------
# Private function to set the item size for size display.
# Most granules will have its size info in SPECIALIZED_RESULTS
# SizeMBECSDataGranule attribute in the inventory result. 
#
# For V0 DAACs, a kludge is done:
# Because SPECIALIZED_RESULTS will only be returned by ECS daacs, 
# V0 DAACs should never send the size of the granules.  For V0
# items however, the closes thing to size is package size.
#

sub set_item_size {
   my ($self, $item) = @_;

   my $item_size = $item->size_mbecs_granule;
   return $item_size if ($item_size);

   # Use wwwvalids.conf to determine which DAACS are V0 daacs
   $WWWVALIDS = IU::WWWValids->new if (!defined($WWWVALIDS));
   my @V0_daacs = $WWWVALIDS->v0_daacs_array();

   my $daac;

   # Only return package size if it is a V0 item.
   foreach $daac (@V0_daacs) {
      if ($item->archive_id eq $daac) {
         $item_size = $item->selected_package_size;

         # Return KB converted into MB value without text
         # Pattern matching is done, looking for
         # digit(s), followed by 0+spaces, followed by KB or kb characters
         return ($item_size / 1024) if ($item_size =~ /\d+\s*KB/i );

         # Return value but remove MB text from value
         # Pattern matching is done, looking for
         # 1+digit(s), followed by 0+spaces, followed by MB or mb characters
         return ($item_size / 1) if ($item_size =~ /\d+\s*MB/i );

         # Unknown size type, return as it is.
         return $item_size;
      }
   }

   return undef;
}

#------------------------------------------------------------
# sorted_list
#------------------------------------------------------------
# Private function.  Take a ref to an array of strings, and
# return a single scalar consisting of the elements sorted and joined
# by commas.

sub sorted_list {
    my $unsorted = shift;
    join ', ', (sort @$unsorted);
}

#------------------------------------------------------------
# _xxx_url  -- URLs for menu buttons
#------------------------------------------------------------

# Return the URL for the Choose Package button:
sub _pkg_url {
    my ($self, $item) = @_;
    if ($item->order_type eq 'DATASET') {

       if (($self->origin eq 'CART') && (@{$item->package_ids})) {
       return script_url('mode'=>'PKGSLCT',
                         'submode' => 'DATASET',
                         'dskey'=>$item->key,
                         'origin'=>($self->origin || 'RESULTS'));
       }

    } else {   # must be granule
       if (@{$item->package_ids}) {
          return script_url('mode'=>'PKGSLCT', 
                            'gkey'=>$item->key,
			    'origin'=>($self->origin || 'RESULTS'));
       }
    }
    undef;
}


#------------------------------------------------------------
#
# PRINT...
#

#------------------------------------------------------------
# print_buttons I
#------------------------------------------------------------
# Output the menu buttons for the i'th (0-based) item.

sub print_buttons {
   # Some basic info:
   my ($self, $num, $some_notes, $some_pkgs_chosen, $some_orderable) = @_;
   # Get the individual element.
   my $item = $self->{Elems}[$num];
   my $type = ($item->order_type eq 'GRANULE') ? 'gkey' : 'dskey';
   my $key  = $item->key;

   # Cart is a special case.
   my $status = $item->order_status;
   my $prim_prefix = 'item';

   # Added different background color to know which items have order
   # package selected.  If item needs a package, the backgroud is
   # bright yellow, otherwise the bacground will be the default.
   my $light_yellow = "#FFFFCC";
   my $lite_gray = "#EEEEEE";

   # "Choose Package" button
   for ($status) {                    # switch statement in perl
      /NEED_PKGS/ and do {
 	 # Set the primary action button filename based on user pref.
         # Item needs a package, the backgroud is bright yellow
	 print "<td nowrap bgcolor=$light_yellow>";
         $HTML -> hypertext("Choose Options", $self->_pkg_url($item));
         print '</td>';
	 last;
      };
      /GO_PKGS/ and do {
         my $type = $item->selected_media_type;
###         debug "GO_PKGS media type selected is $type";
	 print "<td bgcolor=$lite_gray align=center> $type <br>";
         $HTML -> hypertext("Change", $self->_pkg_url($item));
         print '</td>';
	 last;
      };
      /GO_EMAIL_PKGS/ and do {
         my $type = $item->selected_media_type;
###         debug "GO_EMAIL_PKGS media type selected is $type";
	 print "<td bgcolor=$lite_gray align=center> $type <br>";
         $HTML -> hypertext("Change", $self->_pkg_url($item));
         print '</td>';
         last;
      };
      /GO_EMAIL/ and do {
         print "<td bgcolor=$lite_gray align=center>Ready</td>";
	 last;
      };
      # else default action (bad. failed.)

         print '<td bgcolor=$lite_gray>Packaging Error</td>';
         print "<td bgcolor=$lite_gray>\xA0</td>" if ($some_pkgs_chosen);
   }
	
   1;
}

#------------------------------------------------------------
# print_head
#------------------------------------------------------------
# Output the column headers for the 1 line per item table.

sub print_head {
   my ($self, $some_notes, $some_pkgs_chosen, $some_orderable,
        $some_comments, $some) = @_;

   my $bgcolor = "#CCCCCC";
   # Print the column headers for the table.
   print '<tr>';

   print "<td bgcolor=$bgcolor nowrap><b>Order Options</b></td>";
   print "<td bgcolor=$bgcolor><b>", "Subsetting</b></td>";
   print "<td bgcolor=$bgcolor><CENTER><b>Data Granule ID</b>".
         "<BR>(Local Granule ID)</CENTER></td>";
   $WWWVALIDS = IU::WWWValids->new() if (!defined($WWWVALIDS));
   my %pretty = $WWWVALIDS->granListPretty;
   my $title;

   # Print the "On-Demand" heading if there are any "On-Demand" items.
   print "<td bgcolor=$bgcolor nowrap><b>", "On-Demand</b></td>" 
    if($someDemand);

   print "<td bgcolor=$bgcolor nowrap><b>", "Size (MB)</b></td>";
#   print "<td bgcolor=$bgcolor nowrap><b>", "Data Center</b></td>";
   my $pretty_title;
   foreach $title (@Order) {                 # order of presentation!
      $pretty_title = ($pretty{$title}) ? $pretty{$title} : $title;
      next if($pretty_title eq 'Size MB');
      next if($pretty_title eq 'Data Center');
      print "<td bgcolor=$bgcolor nowrap><b>", $pretty_title, '</b></td>'
         if ($some->{$title});
   }
   print "<td bgcolor=$bgcolor>";
   $HTML -> B;
   $HTML -> msubmit_button(CAPTION =>  'Delete',
                           QUERY   => ['mode'    => 'CART',
                                       'submode' => 'REM']);
   $HTML -> _B;
   print "</td>";
   print "</tr>";

   1;
}

#------------------------------------------------------------
# print_customize_button
#------------------------------------------------------------
# noop - override of method in ObjectList
# This will suppress the display of the button.

sub print_customize_button {

1;
}

#------------------------------------------------------------
# print_report_button
#------------------------------------------------------------
# noop - override of method in ObjectList
# This will suppress the display of the button.

sub print_report_button {

1;
}

#------------------------------------------------------------
# any_notes
#------------------------------------------------------------
# Return false if there are no notes in the whole table.

sub any_notes {
   0;   # No longer show notes in the cart.
}

#------------------------------------------------------------
# any_orderable
#------------------------------------------------------------
# Return false for all grans.  This method is only used for data sets.

sub any_orderable {
    0;
}

#------------------------------------------------------------
1;
