package IW::Client;

=head1 NAME

IW::Client - spawn/poll the standalone client

=head1 SYNOPSIS

=head2 Usage example

   # Start it:
   my $client = start IW::Client Activity=>'INVENTORY_SEARCH', 
                                 Timeout=>$query->max_time(),
                                 Infile=>$infile;
         
   # Poll every 10 seconds; allow 10 minutes before we force a quit: 
   while ($client->running) {
      if ($client->run_time > (60 * 10)) {
         $client->stop;    # They have 5 seconds to stop, or
         # we kill them 
         last;
      }
      sleep 10;
   }
   if ($client->status eq 'KILLED') {
      ...
   }

=head2 Constructors

   # Start the client, forcibly killing any existing client:
   my $client = start IW::Client Activity=>'INVENTORY_SEARCH', 
                                 Infile=>$infile;
    
   # Attach to an existing client (if any):
   my $client = attach IW::Client;
   $client or debug("no client to attach to!");

   # Attach to an existing search client (if any):
   # (granule ID, Inventory, or Directory search)
   my $client = attach IW::Client 'search';
   $client or debug("no client to attach to!");

=head2 Halting

   # Attempt graceful stop (will kill if running past allowed time):
   $client->stop;
    
   # Kill outright (optional arg is name of signal to use):
   $client->kill;

=head2 Getting the current status

The following are available whether we B<started> or B<attached> to 
the process:

   # How was this object obtained?  Via START or ATTACH?
   my $origin = $client->origin;

   # Are we running?
   if ($client->running) { ... }

   # When did we start/stop (in epoch time)?
   my $start = $client->start_time;
   my $stop = $client->stop_time;    # only valid if *we* detected it!
    
   # How long (did we run / have we been running)?
   # Only valid if *we* detected the stop!
   my $runtime = $client->run_time;

   # Get the current status: 
   #    undef   == unknown
   #    STOPPED == stopped gracefully (only known if we did it)
   #    KILLED  == killed (only known if we did it)
   #    CRASHED == crashed (only known after a poll)
   my $status = $client->status;    

=head2 Polling/using detailed status

The following are available whether we B<started> or B<attached> to 
the process:

   # Get current statuses of connections, activities, etc.
   #    This is expensive; do only when needed (e.g., once every
   #    few seconds)!
   $client->poll;
    
   # Examine the connections:
   foreach $conn ($client->conn_statuses) {
      # $conn is an IW::Conn object
   }
   # Examine the post-processing:
   foreach $proc ($client->proc_statuses) {
      # $proc is an IW::PostProcess object
   }

=head2 Miscellaneous

As of 1.6, the last client that was started is stored in the session
state, under LastClient.  So the following can now be used:

   # Get the activity we were engaged in:
   $client->activity;         # 'INVENTORY_SEARCH', etc.
    
   # A crash/kill was detected: remove crud left behind from this
   # activity:
   $client->remove_wreckage;

=head1 WARNINGS

Since this object requires an open file descriptor, you cannot store
it as-is in the session state.  A safe copy must be stored instead.

=head1 BUGS

There should be a way of preserving start and stop times across an
"attach".  Better yet, there should be a way of thawing a frozen copy
of a "started" client, and then "attaching" to it so all the original
information would be preserved.

=cut

# HACK: import of these variables had to be made explicit.
BEGIN {
  *IW::Client::APP = \$IW::App::APP;
  *IW::Client::SESSION = \$IW::App::SESSION;
}

use vars qw($APP $SESSION);
use IU;
use IU::Log;
use IU::Spawn ();
use IW::Conn;
use IW::PostProcess;
use strict;

@IW::Client::ISA = qw(IU::Spawn);

# Map the various activity names to command-line args for imsclient.
# This is also used to validate an activity.
my %ActivityNameToId = (
   DIRECTORY_SEARCH  => 'ds',
   FTP_BROWSE        => 'fb',
   INTEGRATED_BROWSE => 'ib',
   INVENTORY_SEARCH  => 'is',
   PRODUCT_REQUEST   => 'pr',
   PRODUCT_STATUS_REQUEST => 'os',
   ACCOUNT_STATUS_REQUEST => 'as',
   PROFILE_RETRIEVAL_REQUEST => 'rr',
   PROFILE_UPDATE_REQUEST    => 'ur',
   PRICE_ESTIMATE_REQUEST    => 'er',
   PROFILE_SUBMIT_REQUEST       => 'sr',
   USER_PASSWORD_CHANGE_REQUEST => 'cr',
);

# The following activities do post-processing:
my %ActivityDoesPostProcessing = (
   'INVENTORY_SEARCH' => 1,     # yes
   'DIRECTORY_SEARCH' => 1,     # yes
);

my %SearchDirectoryHash = (
   'INVENTORY_SEARCH' => 'search/', 
   'DIRECTORY_SEARCH' => 'search/',
   'ACCOUNT_STATUS_REQUEST' => 'account_status/',
);

#------------------------------------------------------------
# action
#------------------------------------------------------------
# I<Private utility.>  The action we execute when spawning.
# I know it looks like an instance method, but it's really
# not.  See the IU::Spawn documentation to see how it's used.

sub action {
   my $client = shift;

   inform "starting imsrunclient: ", @{$client->{Args}};
   exec "${IU::LibDir}/bin/imsrunclient", @{$client->{Args}};
}

#------------------------------------------------------------
# nuke_existing SUBDIR
#------------------------------------------------------------
# I<Class method.>  Kill any existing search.  Returns the pid
# killed, or 0.
# SUBDIR is the subdirectory of the search, if any.  undef ok.

sub nuke_existing {
   my $class = shift;

   my $subdir = shift;
   defined ($subdir) or $subdir = '';

   # Attach:
   my $other = $class->attach ($subdir);
   $other or return 0;

   # Kill (if running):
   $other->running or return 0;
   inform "Killing client (pid... ", $other->pid, ")";
   $other->kill;
   $APP->note('WARN', "Killed existing client (pid ", 
              $other->pid, ")");
   $other->pid;
}

#------------------------------
#
# CLASS METHODS
#
#------------------------------

#------------------------------------------------------------
# attach  SUBDIR
#------------------------------------------------------------
# I<Class/instance constructor/initiallizer.>  Attach to
# the last client that was started.
#
# If used as a class method, assumes that an object representing the 
# last client started will be stored in the session state.  If *that*
# fails, a default one is used.
#
# In the event of a benign error, simply returns undef.
#
# If optional SUBDIR is specified, try attach to LastSUBDIRClient.
# SUBDIR should not end in "/".

sub attach { 
   my $self = shift;
   my $subdir = shift;
   defined ($subdir) or $subdir = '';

   # Create new instance, if necessary:
   my $class = ref($self) || $self;
   if (ref($self)) {
      debug "Client::attach: we were already an instance";
   }
   elsif ($self = $SESSION->state->{'Last' . $subdir . 'Client'}) {
      debug "we got back last client run";
   }
   else {
      debug "we're creating a new client";
      $self = $class->new;              # just create a new instance
   }
   # If subdirectory is specified, use lock file in that subdir.
   my $Lockfile = $subdir ? "$subdir".'/client..LCK' : 'client..LCK';

   # Attach, or return false!
   $self->SUPER::attach(Lockfile=>$SESSION->path_to($Lockfile),
                        Action=>\&action) or return undef;
   $self;
}

#------------------------------------------------------------
# start PARAMHASH
#------------------------------------------------------------
# I<Class/instance constructor/initiallizer.>  Params are:
#
#    Activity => Activity to perform.  This is one of
#                INVENTORY_SEARCH, DIRECTORY_SEARCH, etc.
#    Infile   => Path to input file, for the given activity.
#    Timeout  => For informational purposes only, the time limit
#                for the search.
#
# Unsupported params are:
#
#    Args     => Command line args to 'imsclient'

sub start { 
   my $self = shift;
   my %params = @_;

   # Sanity check:
   (-x "${IU::LibDir}/bin/imsrunclient") or
      puke 560, "${IU::LibDir}/bin/imsrunclient is not executable!";
    
   # Create new instance, if necessary:
   ref($self) or $self = $self->new;

   # Check for old-style invocation, and die:
   !exists($params{Args}) or 
      puke 600, "Use of Args is now unsupported! Re-code!";

   # Set up the activity:
   $self->{Activity} = $params{Activity};
   my $activity_id = $ActivityNameToId{$self->{Activity}} ||
      puke 600, "bad/missing activity!";

   # Save the information for the timeout message, if supplied, set to
   # system default max if not:

   $self->{Timeout} = $params{Timeout};
   $self->{Timeout} = $IU::SETUP::MaxSearchTime 
      if $self->{Timeout} eq undef;

   # Set up the args that will go to imsrunclient:
   $self->{Args} = [$activity_id, $params{Infile}, $ENV{REMOTE_USER}];

   # If this is an granule, inventory, or directory search, then
   # use the lock file in the search subdirectory.
   # And set the subdirectory to run in to "search".
   my $rundir = $SearchDirectoryHash{$self->activity} ? 
    $SearchDirectoryHash{$self->activity} : '';
   my $Lockfile = $rundir . "client..LCK";
   # Store the run subdirectory for the client in the client object.
   chop $rundir;
   $self->{RunDir} = $rundir;
   # Create a directory to new_directory stuff in:
   my $new_dir = $SESSION->path_to("$rundir");
   (-d $new_dir) or mkdir($new_dir, 0755) or puke "mkdir $new_dir: $!";

   # We only allow ONE to run at a time:
   if ($SESSION->state->{'Last' . $rundir . "Client"}) {
      ref($self)->nuke_existing($rundir);
   }
   # Prepare to run:
   $self->preflight();

   # Spawn or die!
   $self->SUPER::start(Lockfile=>$SESSION->path_to($Lockfile),
                       Action=>\&action)
      or puke "couldn't spawn client!";

   # Save harmless copy in the session state.  Note that we have to
   # create a copy which does NOT contain the FileHandle.  
   # The copy keeps track of our start time, among other things.
   $self->start_time(time());
   $SESSION->state->{'Last'.$rundir.'Client'} = $self->storable_copy;

   # Done! Return ourself!
   $self;
}

#------------------------------------------------------------
# storable_copy
#------------------------------------------------------------
# I<Constructor>.  Create a storable copy of ourselves.
#
# Normally, we cannot be stored in the session state, since we
# contain an open filehandle for the lock file.

sub storable_copy {
   my $self = shift;

   my %copy = %$self;
   $copy{LOCK} = undef;         # KLUDGE
   bless \%copy, ref($self);    
}

#------------------------------
#
# INSTANCE METHODS
#
#------------------------------

#------------------------------------------------------------
# activity
#------------------------------------------------------------
# The name of the activity we invoked: "INVENTORY_SEARCH", etc.

sub activity {
   uc($_[0]->{Activity} || '');
}

#------------------------------------------------------------
# conn_statuses
#------------------------------------------------------------
# Return the array of DAAC connection statuses, or empty.
# You'll need to poll() for these first!

sub conn_statuses {
   ($_[0]->{Conns} ? @{$_[0]->{Conns}} : ());
}

#------------------------------------------------------------
# _check_for_crash
#------------------------------------------------------------
# I<Private:> does it look like the client crashed by itself?
# The value returned is based on information from the last poll().
# If we did crash, this sets the status (if necessary).

sub _check_for_crash {
   my $self = shift;

   # Return true if known:
   return 1 if ($self->{Status} && $self->{Status} eq 'CRASHED');

   # Return false if we're still running:
   return 0 if $self->running;    

   # Return false if the client was killed by outside forces:
   return 0 if ($self->{Status} && 
                (($self->{Status} eq 'STOPPED') ||
                 ($self->{Status} eq 'KILLED')));

   # We crashed if no status information could be retrieved:
   debug "checking status info";
   @{$self->{Conns}} or return ($self->{Status} = 'CRASHED');

   # If search, we crashed if no post-processing info retrieved:
   debug "checking post-processing info";

   my $rundir = $SearchDirectoryHash{$self->activity} ? 
    $SearchDirectoryHash{$self->activity} : '';
   my $Stopfile = $rundir . "client.stop";
   if ($self->requires_post_processing) {
      @{$self->{Procs}} or return ($self->{Status} = 'CRASHED');
   }
   # Return false right away if we stopped it (since no way to
   # really tell).  This is a light KLUDGE to prevent a fuzzy
   # case from appearing as a bug. 
   if (-e $SESSION->path_to($Stopfile)) {
      debug "client.stop files means crash condition undetectable";
      return 0;
   }
   # We crashed if not everybody finished AND if we didn't stop it:
   my $status;
   foreach $status (@{$self->{Conns}}) {
      if ($status->still_running()) {
         warn "A connection was still running, and we didn't ",
              "stop the search... that means that the client crashed!";

         return ($self->{Status} = 'CRASHED');
      }
   }
   0;     # we're okay!
}

#------------------------------------------------------------
# kill [ARGS]
#------------------------------------------------------------
# Kill the client.  Automatically removes the wreckage.
# Any args are passed to inherited method.

sub kill {
   my $self = shift;
   warn "killing client";
   my $result = $self->SUPER::kill(@_);
   $self->{Status} = 'KILLED';
   $self->remove_wreckage;
   $result;
}

#------------------------------------------------------------
# poll
#------------------------------------------------------------
# Poll the connections.  Since these are written to disk, it is
# valid to do it both before and after the client has completed
# its run.
#
# If the current status is unknown, this method also attempts
# to ascertain it.
#
# Finally: crash detection is done here (since it requires the detailed
# info), and wreckage is removed if a crash is detected.
#
# This is an expensive operation: just do it when you need to!
# For example, once every few seconds.  It's not really useful to do
# it more frequently.

sub poll {
   my $self = shift;
   my ($conn, @conns);
   my ($proc, @procs);

   # Reset everything first:
   $self->{Conns} = [];
   $self->{Procs} = [];

   # If this is an granule, inventory, or directory search, then
   # check the stat file in the search subdirectory.
   
   my $rundir = $SearchDirectoryHash{$self->activity} ? 
    $SearchDirectoryHash{$self->activity} : '';
   my $statfile = $rundir . "client.stat";
   # Parse the client status, and just be silent if not available:
   if (!open(CONNS, $SESSION->path_to($statfile))) {
      warn "I couldn't get the current 'imsclient' communication".
           "status.";
   } else {
      while ($conn = input IW::Conn \*CONNS) {
         # what activity are we doing?
         $conn->activity($self->activity);
         push(@{$self->{Conns}}, $conn);
      }
      close(CONNS);
   }
   # If a search...
   # Parse the post-processing status, and be silent if not available:
   if ($self->requires_post_processing) {
      my $rundir = $SearchDirectoryHash{$self->activity};
      if (!open(PROCS, $SESSION->path_to($rundir . "runclient.stat"))) {
         warn "I couldn't get the current 'imsrunclient' status.";
      } else {
         while ($proc = input IW::PostProcess \*PROCS) {
            push(@{$self->{Procs}}, $proc);
         }
         close(PROCS);
      }
   }
   # If the current status is unknown, try to know it:
   if (!$self->{Status}) {
      # Test whether we crashed, and set status:
      if ($self->_check_for_crash) {
         inform "crash detected: removing wreckage";
         $self->remove_wreckage;
      }
   }
   # The current status is now known to the best of our ability.
   # Refresh the stored copy, so that subsequent reloadings of a comm
   # status page will cause it to start out with an up-to-date status.
   #   (This is a light KLUDGE to prevent a bug that occurred when we
   #    reloaded the comm status screen after a STOP... it had no
   #    record of the comm status that we detected.)
   if ($self->requires_post_processing) {
      $SESSION->state->{LastsearchClient} = $self->storable_copy;
   } else {
      $SESSION->state->{LastClient} = $self->storable_copy;
   }
   # If client crashed, do a puke.  Tell user system error and
   # developers have been notified.  If any results, suggest user
   # go to View Results.  If no results, suggest user go create
   # new search.  Send a dying gasp to developers.
   if ($self->{Status} && $self->{Status} eq 'CRASHED') {
      # If we already puked for this crash, then we're done here.
      # If not, set a puked flag so we only puke once!  This is
      # necessary because the Navigator calls poll.  Then puke.
      unless ($self->{Puked}) {
         $self->{Puked} = 1;    # Set puke flag.
         # Have to save client with puke flag in session state.
         if ($self->requires_post_processing) {
            $SESSION->state->{LastsearchClient} =
               $self->storable_copy;
         } else {
            $SESSION->state->{LastClient} = $self->storable_copy;
         }
         # Check if any data sets received.
         my $num_datasets = 0;
         my $proc;
         foreach $proc ($self->proc_statuses){
            $num_datasets += $proc->n_datasets;
         }
         # Begin error message.
         my $err_string = "There was a system error during your ".
                          "search.  The system administrator has ".
       	                  "been notified and will investigate.  ";
         # Tailor error message based on if we got any data sets
         # and puke.
         if ($num_datasets > 0) {
            puke 456, $err_string . "However, some results were ".
                      "received.  Select your browser's 'Reload' ".
                      "or 'Refresh' button to view these results.";
         } else {
            puke 456, $err_string . "No results were received.  You ".
                      "might want to try your search again, try it ".
                      "later, or try altering your query and search ".
                      "for something else.";
         }
      }
   }
   1;
}

#------------------------------------------------------------
# preflight
#------------------------------------------------------------
# We're just about to spawn.  Prepare the directory.
# WARNING: removes results of previous search/browse!

sub preflight {
   my $self = shift;

   # Subdirectory to run in. Could be blank.
   my $subdir = $self->{RunDir};
   # If subdir given, append slash, else set to empty string.
   defined ($subdir) and $subdir .= '/' or $subdir = '';

   # Create a directory to hide stuff in:
   my $hide = $SESSION->path_to("clienthide");
   (-d $hide) or mkdir($hide, 0755) or puke "mkdir $hide: $!";

   # Delete files from a previous run.  ok if not there.
   unlink $SESSION->path_to("$subdir".'client.obj');
   unlink $SESSION->path_to("$subdir".'client.stat');
   unlink $SESSION->path_to("$subdir".'client.stop');
   unlink glob $SESSION->path_to("$subdir".'runclient.*');

   # Remove activity-specific stuff:
   if ($self->activity eq 'INVENTORY_SEARCH' ||
       $self->activity eq 'DIRECTORY_SEARCH') { 

      # Remove the results of the last search...
      warn "Removing results of previous search!";
      unlink $SESSION->path_to('results.gdb');
      unlink $SESSION->path_to("$subdir".'dataset.gdb'); # not there
      unlink $SESSION->path_to("$subdir".'granule.gdb'); # not there
   }
   elsif ($self->activity eq 'INTEGRATED_BROWSE') {
      warn "Removing results of previous browse!";
      unlink $SESSION->path_to('hdfnames.gdb');
   }
   # Remove other junk:
   my $direntry;
   opendir(DIR, $SESSION->dir) or
       puke 452, "opendir ",$SESSION->dir,": $!";
   while ($direntry = defined(readdir(DIR))) {
      if ($direntry =~ /(\.msg$|TmpFile)/) {
         unlink $SESSION->path_to($direntry);
      }
   }
   closedir(DIR);
   1;
}

#------------------------------------------------------------
# proc_statuses
#------------------------------------------------------------
# Return the array of post-processing statuses, or empty.
# You'll need to poll() for these first!

sub proc_statuses {
   ($_[0]->{Procs} ? @{$_[0]->{Procs}} : ());
}

#------------------------------------------------------------
# redirect_stdio
#------------------------------------------------------------
# I<Override:> close stdin/stdout/stderr in the child when starting.

sub redirect_stdio {

   # Redirecting is safer than an actual close...
   open STDIN, "</dev/null";
   open STDOUT, ">/dev/null";
   open STDERR, ">/dev/null";
}

#------------------------------------------------------------
# remove_wreckage
#------------------------------------------------------------
# We have detected a crash or a kill.  Delete all the incompleted
# stuff that might be lying around from this last request.  
#
# This does I<not> delete logs and status files!
# IMPORTANT: IT MUST NOT DELETE THE client.stop FILE!!!!

sub remove_wreckage {
   my $self = shift;
   ($self->activity) or return undef;    
   if ($ActivityDoesPostProcessing{$self->activity}) {
      warn "Removing wreckage from failed search";
      my $rundir = $SearchDirectoryHash{$self->activity};
      unlink $SESSION->path_to($rundir . "dataset.gdb");
      unlink $SESSION->path_to($rundir . "granule.gdb");
   }
   elsif ($self->activity eq 'INTEGRATED_BROWSE') {
      warn "Removing wreckage from failed browse";
      unlink $SESSION->path_to('hdfnames.gdb');
   }
   1;
}

#------------------------------------------------------------
# requires_post_processing
#------------------------------------------------------------
# Return whether or not post-processing is generally done for this 
# activity.

sub requires_post_processing {
   my $self = shift;
   $ActivityDoesPostProcessing{$self->activity};
}

#------------------------------------------------------------
# running
#------------------------------------------------------------
# Check to see if we're running.  
#
# If we're not, and the stop time hasn't been set, then set it to
# be the current time.  Assuming that we hang onto this object in
# the session state, we should get decent behaviour out of this.

sub running {
   my $self = shift;
   my $running;

   # Get subdirectory that client is running in, if any.
   my $subdir = $self->{RunDir};
   $subdir .= '/' if ($subdir);    # append a / if a subdirectory
   # Find files in run directory of the form runclient.pid.*
   # There should just be one with the * being the process id.
   # Use glob to get files of the form and a regexp to extract the PID.
   my $globfile = $SESSION->path_to($subdir . 'runclient.pid.*');
   my @pidfiles = glob $globfile;

   # If we didn't get the pid file, maybe the client didn't
   # start yet.  Loop/wait up to 10 seconds before returning fail.
   if ($#pidfiles < 0) {
      my $i;
      for ($i=0; $i<10; $i++) {
         debug "Client::running called and no runclient.pid yet. Try ",$i+1;
         sleep(1);  # Wait a second and try again.
         @pidfiles = glob $globfile;
         last if ($#pidfiles >= 0);
      }
      if ($#pidfiles < 0) {
         return 0;   # Never got a runclient.pid file!
      }
   }
   my $pidfile = $pidfiles[$#pidfiles];
   $pidfile =~ /runclient\.pid\.(\d+)$/;
   my $pid = $1;

   # Catch case where imsrunclient hasn't created files.
   if (!defined($pid) or $pid <= 0) {
      return (0);
   }
   # If a "done" file exists for the PID, then imsrunclient is done.

   if (-e $SESSION->path_to("$subdir"."runclient.$pid.done")) {
      $running = 0;
   } else {
      # Else explicitly test for imsrunclient process with "kill".
      $running = CORE::kill 0, $pid;
   }
   # If not running and no stop time set, store the current time.
   if (!$running && !($self->{StopTime})) {
### my $statfile = $SESSION->path_to("runclient.stat");
### $self->{StopTime} =  (-e $statfile) ? (stat($statfile))[10] : time;
      $self->{StopTime} = time;
   }
   $running;
}

#------------------------------------------------------------
# run_time
#------------------------------------------------------------
# If still running, how long have they been running (in seconds)?
# If not running, how long *did* they run?

sub run_time {
   my $self = shift;

   debug "started = ", $self->start_time;
   debug "stopped = ", $self->stop_time;
   my $started = $self->start_time || time;
   my $stopped = $self->stop_time || time;
   debug "time = ", time(), "started = $started; stopped = $stopped";
   $stopped - $started;
}

#------------------------------------------------------------
# start_time
#------------------------------------------------------------
# Return the start time.  This is the time we STARTED the operation.
# With argument, sets it.

sub start_time {
   my ($self, $time) = @_;
   if (@_ > 1) {
      debug "setting start time to ", $time;
      $self->{StartTime} = $time;
   }
   debug "start time is ", $self->{StartTime};
   $self->{StartTime};
}

#------------------------------------------------------------
# status
#------------------------------------------------------------
sub status {
   $_[0]->{Status};
}

#------------------------------------------------------------
# stop [TIMEOUT]
#------------------------------------------------------------
# Stop the client.  Wait TIMEOUT seconds (there's a default); 
# if it's still running, then kill it.  
#
# Returns 'NOOP' (true) if client wasn't running in the first 
# place, 'STOPPED' if it could be stopped, and 'KILLED' otherwise.  
# Sets the status.

sub stop {
   my ($self, $timeout) = @_;
   defined($timeout) or $timeout = (5 * 60);

   # Do nothing if not running:
   return $self->{Status} if $self->{Status};
   $self->running or return 'NOOP';
   warn "stopping the client!";

   # Create the stop file:
   my $rundir = $SearchDirectoryHash{$self->activity} ? 
    $SearchDirectoryHash{$self->activity} : '';
   my $Stopfile = $rundir . "client.stop";
   my $stop = $SESSION->path_to($Stopfile);
   open STOP, ">$stop" or puke "couldn't create $stop: $!";
   print STOP $$;
   close STOP;

   # Wait, checking frequently:
   while ($self->running and ($timeout > 0)) {
      sleep 1;
      $timeout -= 1;
   }
   # Still here?
   if ($self->running) {
      $self->kill;
      $APP->note('ERROR', "The client did not halt normally; " .
                          "results are invalid.");
   } else {
      $APP->note('INFO',  "The client was halted gracefully.");
      $self->{Status} = 'STOPPED';
   }
   return $self->{Status};
}

#------------------------------------------------------------
# stop_time
#------------------------------------------------------------
# Get the stop time.  This is really the time we first detected
# that the client had stopped.

sub stop_time {
   $_[0]->{StopTime};
}

#------------------------------------------------------------
# max_search__time
#------------------------------------------------------------
# Get the maximum time searches should run (in minutes).

sub max_search_time {
   $_[0]->{Timeout};
}

#------------------------------------------------------------
# stopping
#------------------------------------------------------------
# Check to see if we're in the process of stopping (or if we're
# stopped).

sub stopping {
   my $self = shift;
   my $rundir = $SearchDirectoryHash{$self->activity} ? 
    $SearchDirectoryHash{$self->activity} : '';
   my $Stopfile = $rundir . "client.stop";
   return (-e $SESSION->path_to($Stopfile));
}

#------------------------------------------------------------
1;
