package IW::Comm::BrowseReq;

=head1 NAME

IW::Comm::BrowseReq - integrated/FTP browse request

=head1 NOTES

To generate the ODL, loads in the user preferences to get
the contact address.

=cut

use IU::ODL ();
use IU::Granule ();
use IU::Log;

use IW::User;
use IW::UserAddr;
use IW::App;

use strict;

#------------------------------------------------------------
# new PARAMHASH
#------------------------------------------------------------
# I<Constructor.>  Params are:
#
#     ArchId => archive id
#     DsetId => dataset id
#     GranId => granule id
#
sub new {
    my ($class, %params) = @_;
    my $self = \%params;
    bless $self, $class;
}

#------------------------------------------------------------
# to_ODL
#------------------------------------------------------------
# Convert to an ODL object suitable for input to the client, 
# and return it. The ODL object is a group containing a BROWSE_REQUEST
# parameter.
#
# WARNING: the authentication information I<is> placed in the ODL!!!!

sub to_ODL {
    my $self = shift;
    my ($BR, $BG, $CA);

    # Begin:
    my $file = IU::ODL->new_group;    
    $file->add_param(BROWSE_REQUEST => ($BR = IU::ODL->new_group));

    # Browse type:
    $BR->add_param(MESSAGE_ID      => '0'); 
    if ($SESSION->state->{BrowseType} eq 'FTP' ) {
	$BR->add_param(BROWSE_TYPE     => 'FTP_ONLY'); 
    }
    else {
    	$BR->add_param(BROWSE_TYPE     => 'Y'); }

    $BR->add_param(DATA_CENTER_ID  => $self->{ArchId});
    $BR->add_param(BROWSE_GRANULES => ($BG = IU::ODL->new_group));
    
    # Browse granules...
    $BG->add_param(DATASET_ID => $self->{DsetId});
    $BG->add_param(GRANULE_ID => $self->{GranId});

    # Get the user prefs or make new.
    my $user = $SESSION->state->{User} or new IW::User;

    # Add the authenticator, if any:
    my $authenticator = $user->authenticator;
    if ($authenticator) {
	$BR->add_param(AUTHENTICATOR => IU::ODL->conv($authenticator));
    }

    # Add the NASDA authenticator, if any:
    my $nasda_authenticator = $user->nasda_authenticator;
    if ($nasda_authenticator) {
        $BR->add_param(NASDA_AUTHENTICATOR =>
                             IU::ODL->conv($nasda_authenticator));
    }

    # Use ecs authenticator if we need to:
    # Modified to go to "ECS Profile SMC" instead of home DAAC.
    unless ($IU::EcsProfileSmc eq '') {
       # Add the ecs authenticator, if any:
       my $ecs_authenticator = $user->ecs_authenticator;
       if ($ecs_authenticator) {
           $BR->add_param(ECS_AUTHENTICATOR =>
                             IU::ODL->conv($ecs_authenticator));
       }
    }

    # Add the contact information if known, or else fake it:
    my $contact = ($user->req_addr) || (bogus IW::UserAddr);

    # Need this, if user hasn't filled out user prefs!
    $BR->add_param('CONTACT_ADDRESS', $contact->to_ODL);

    # Done!
    $file;
}

#------------------------------------------------------------
1;
