package IW::Comm::FTPBrowseResult;

=head1 NAME

IW::Comm::FTPBrowseResult - object representing the result of an FTP browse 

=head1 NOTES

Parses the object file returned after an FTP Browse Request has been
submitted.  Also looks for the DAAC_CONTACT and DATA_URL extentions.

=cut

use IU;
use IU::Chunk;
use IU::Log;
use IU::FancyURL;
use IW::App;
use IW::DaacAddress;

use strict;

#------------------------------------------------------------
# new
#------------------------------------------------------------
# I<Class method, constructor.> 

sub new {
    my $class = shift;
    my %params = @_;
    bless \%params, $class;
}

#------------------------------------------------------------
# from_chunk CHUNK
#------------------------------------------------------------
# I<Class method, constructor.> 
# Read a FTP_BROWSE_RESULT message from a chunk.
# A browse chunk containing "x-error" is considered null and void.
#
# The browse object for success, false for failure (incluing EOF).

sub from_chunk {
    my ($self, $chunk) = @_;
    my $errmsg;

    # Input chunk, checking for errors:
    defined($chunk) or return undef;
    defined($errmsg = $chunk->value('x-error')) and return error 454, $errmsg;

    # Create new object, initiallizing from chunk:
    ref($self) or $self = $self->new;

    # Status code comment:
    $self->{StatusCodeComment} = $chunk->value('STATUS_CODE_COMMENT');

    # Contact info:
    my $contact_chunk = $chunk->value('DAAC_CONTACT_ADDRESS');
    if ($contact_chunk) {
        $self->{DaacContact} =
          IW::DaacAddress->from_chunk($contact_chunk);
    }

    # URLs (we tolerate more than one of each type):
    my $key;
    foreach $key (qw(DATA_URL MISC_URL)) {
        # Allow many URLs of the current flavor...
        $self->{$key} = [];

        my $url_chunk;
        foreach $url_chunk ($chunk->values($key)) {
	    push @{$self->{$key}}, IU::FancyURL->from_chunk($url_chunk, $key);
        }
    }

    # Done!
    $self;
}

#------------------------------------------------------------
# read_last
#------------------------------------------------------------
# I<Class method, constructor.>
# Read the last set of FTP browse results from the standard location 
# in the standard way.
#
# WARNING: this is only going to work if the last search was 
# an FTP browse!

sub read_last {
    my $class = shift;
    my $objpath = $SESSION->path_to('client.obj');
    my $chunk;   # the current chunk
    my $self;    # new object we'll create
    
    # Read and queue chunks:
    debug "reading raw chunks from $objpath";
    open(OBJ, "<$objpath") or puke "couldn't read $objpath: $!";
    $chunk = input IU::Chunk \*OBJ;
    close OBJ;

    # Check to see if client.obj is empty or not:
    $chunk or puke 451, "client.obj is empty";
    if ($chunk->value('X_TYPE') eq 'ROOT') {
	my $result_chunk = $chunk->value('FTP_BROWSE_RESULT') || 
	    puke 513, "your last search was not an FTP browse";
	$self = $class->from_chunk($result_chunk);
    }

    return $self;
}

#------------------------------------------------------------
# comment
#------------------------------------------------------------
# Return the status code comment, or the empty string if there is none.

sub comment      { $_[0]->{StatusCodeComment} || '' }

#------------------------------------------------------------
# contact
#------------------------------------------------------------
# Return the contact address, an IW::DaacAddress.

sub contact      { $_[0]->{DaacContact} }

#------------------------------------------------------------
# data_URLs
#------------------------------------------------------------
# Return ref to array of data URLs (as IU::FancyURLs), or to [] if none.

sub data_URLs    { $_[0]->{DATA_URL} }

#------------------------------------------------------------
# misc_URLs
#------------------------------------------------------------
# Return ref to array of misc URLs (as IU::FancyURLs), or to [] if none.

sub misc_URLs    { $_[0]->{MISC_URL} }

#------------------------------------------------------------
# archive_id
#------------------------------------------------------------
# Return the archive id.

sub archive_id   { $_[0]->{ArchiveId}; }

#------------------------------------------------------------
1;
