package IW::Comm::Manager;

=head1 NAME

IW::Comm::Manager
    - help a mode manage interaction with comm status screen

=head1 DESCRIPTION

Mix this into any mode to enable the mode to perform complex
comm status screen management.

    package IW::ModeYOURMODE;
    
    use IW::Mode;
    use IW::Comm::Manager;
    
    @ISA = qw(IW::Mode IW::Comm::Manager);

This package provides several abstract submodes and other
handlers which will reduce your job in writing modes which
need to do comm status stuff.


=head1 WHAT YOU NEED TO PROVIDE

Simply provide concrete overrides for the following abstract
methods (see the methods below for more details)...

    comm_preflight
        We want to show the desired page.  Must we do a search first?
        Here's where things like the browse cache come in.
         
    comm_start_search
        Start the search.  You must start the IW::Client.

    comm_finish_up
        A search completed.  Check to see if any post-processing
        needs to be done; if so, do it.  The default is to do nothing.

    comm_ok_page
        Draw the page which would be shown if a search completed okay.
    
    comm_fail_page
        Draw the page which would be shown if a search failed.

You may also want to override these methods as well:

    comm_wait
        How long to wait between auto-refreshes of page, in seconds.
        The default of 15 seconds should generally suffice.
         
    comm_stop_gracefully
        Return true if the STOP button should attempt a graceful stop,
        and false if it should simply do a kill of the client.     

=head1 NOTES

All submodes begin with "COMM", to reduce risk of namespace
conflicts.

All other methods begin with "comm_", to reduce risk of namespace
conflics.

The session state is augmented with the following information:

        LastCommMode      Last comm mode we started; e.g., VIEWBROWSE
    or
        LastsearchCommMode  Last comm mode we started; e.g., SEARCH

=head1 RCS INFO

$Id: Manager.pm,v 1.17 2002/04/17 16:23:17 rrau Exp $

=cut

use IU;
use IU::Log;

use IW::App;
use IW::Client;
use IW::CommStat;
use IW::Util qw(:url);

#------------------------------
#
# MAIN LOGIC
#
#------------------------------
# You will probably NOT want/need to override these...

#------------------------------------------------------------
# comm_start
#------------------------------------------------------------
# Backend for submode COMMSTART.
# If we have what we need to output the ultimate "ok" page, do so now.
# Else, start the search, and redirect to the COMMTRACE mode.  
#
sub comm_start {
   my $self = shift;
    
   my $preflight = $self->comm_preflight;
   if ($preflight ne 'GO') {   # hey, we've already got what we need!
      debug "CommMgr: preflight ($preflight); we don't need ".
            "to search...";
      ($preflight eq 'OK') ? $self->comm_ok_page : 
                             $self->comm_fail_page;
   }
   else {                            # we must do a search...
      # Build the search ODL, and start the client:
      my $started = $self->comm_start_search;
      puke "couldn't start search!"   if not $started; 
      return 1                        if ($started eq 'DONE');   

      my $mode = $REQ->param('mode');
      my $subdir = '';
      $subdir = 'search' if ($mode eq 'SEARCH');

      # Normal search in progress at this point!
      my $client = (attach IW::Client $subdir);

      # Make a note of what mode/operation this "last search" was: 
      $SESSION->state->{'Last'.$subdir.'CommMode'}  =
          $self->comm_mode;

      # Redirect the user's browser to trace it:
      my $url = script_url('mode'   =>$self->comm_mode,
                           'submode'=>'COMMTRACE');
      $REPLY->redirect($url, 'TEMPORARY');
   }
}

#------------------------------------------------------------
# comm_trace COMMSTATUS
#------------------------------------------------------------
# Backend for submode COMMTRACE.
# If the client is still running, send a comm status screen.
# Else, output the appropriate results page.
 #
sub comm_trace {
   my ($self, $commstat) = @_;

   # Get client:
   my $client = $commstat->client;

   # Output only if client is running:
   if ($client->running) {                  # non-final page...

      # Have we been running too long?

#  Do not use comm_timeout to get the max search time.
#  ModeSEARCH->comm_timeout - opens the query.dat file,
#  parses the ODL, creates a query object and then gets the
#  max search time from the query. We do not want to be repeating
#  this expensive operation on every comm stat page reload.
#  The client was started with max search time as a parameter
#  so it knows it.  Use max_search_time stored in the client.
#  if ($client->run_time >= $self->comm_timeout) {

      if ($client->run_time >= $client->max_search_time*60) {
         $self->comm_stop_it($commstat);       # do a stop or kill
      } else {                                 # comm status screen
         $commstat->send_doc(Wait=>$self->comm_wait);   
      }
   } else {                                    # final-page...
      $self->comm_client_stopped(CommStat=>$commstat);
   }
   1;
}

#------------------------------------------------------------
# comm_kill COMMSTATUS
#------------------------------------------------------------
# Backend for submode COMMSTOP.
#
sub comm_kill {
   my ($self, $commstat) = @_;

   # Get client, and stop it if it's running:
   my $client = $commstat->client;
   if ($client->running) {
      debug "CommMgr: comm_kill: client still running; killing it";
      $client->kill;
   } else {
      debug "CommMgr: comm_kill: client already stopped";
   }
   # Output final page:
   $self->comm_client_stopped(CommStat=>$commstat); 
}

#------------------------------------------------------------
# comm_stop COMMSTATUS
#------------------------------------------------------------
# Backend for submode COMMSTOP.
# If the client is still running, output a "stopping" comm status 
# screen, and (if necessary) issue the "stop" command to the client.
# Else, output the appropriate results page.
#
sub comm_stop {
   my ($self, $commstat) = @_;

   # Get client:
   my $client = $commstat->client;

   # Stop it only if necessary...
   if ($client->running) {           # still running...
      debug "CommMgr: comm_stop: client still running";

      # Continue to monitor the stoppage:
      $APP->message(new IW::Note 
                    Class=>'WARN',
                    Text=>"Trying to stop the search gracefully... ".
                          "Please be patient!");
      $commstat->send_doc(Wait=>$self->comm_wait, 
                          Stopping=>1);

      # If client is not in the process of stopping, then STOP IT!
      if (!$client->stopping) {
         debug "I think the client is running; I'm stopping it";
         $client->stop;    # enough already!
      }
   } else {                            # already stopped
      debug "CommMgr: comm_stop: client already stopped";
      $self->comm_client_stopped(CommStat=>$commstat); 
   }
   1;
}

#------------------------------------------------------------
# comm_mode
#------------------------------------------------------------
# What mode will take us back here?
#
sub comm_mode {
   my $self = shift;
   $REQ->param('mode');
}

#------------------------------------------------------------
# comm_stop_it COMMSTAT
#------------------------------------------------------------
# Either stop or kill, whatever's appropriate.

sub comm_stop_it {
   my ($self, $commstat) = @_;
   ($self->comm_stop_gracefully ?  $self->comm_stop($commstat) 
                                :  $self->comm_kill($commstat));    
} 

#------------------------------------------------------------
# comm_stop_url
#------------------------------------------------------------
# What URL will stop us?
#
sub comm_stop_url {
   my $self = shift;
   return script_url(mode    => $self->comm_mode, 
                     submode =>'COMMSTOP');    
}

#------------------------------------------------------------
# comm_client_stopped PARAMHASH
#------------------------------------------------------------
# Either COMMTRACE or COMMSTOP has detected the client being stopped.
# If the client finished okay, do any finishing up and then
# output the "ok" page.  Else, output the "fail" page. 
#
# Params are:
#     CommStat => the comm status object

sub comm_client_stopped {
   my $self = shift;
   my %params = @_;

   # Get comm status object, and poll client for each connection's
   # status:
   my $commstat = $params{CommStat};
   $commstat->client->poll;

   # How'd we do?
   if ($commstat->finished_ok) {
      debug "CommMgr: outputting OK page...";
      $self->comm_finish_up(@_, Status=>'OK'); 
      $self->comm_ok_page(@_);
   } else {
      debug "CommMgr: outputting FAIL page...";
      $self->comm_finish_up(@_, Status=>'FAIL'); 
      $self->comm_fail_page(@_);
   }
}

#------------------------------------------------------------
# comm_final_status_page PARAMHASH
#------------------------------------------------------------
# Draw the final comm status page which would be shown.
#
# WARNING: if you make this your failure page, then beware
# that it expects there to be a comm status object!!!  You
# cannot have comm_preflight() return 'FAIL'; you must die/puke
# in your comm_preflight() if it fails instead.
#
sub comm_final_status_page {
   my ($self, %params) = @_;
   $params{CommStat}->send_doc(IsFinal=>1); 
}

#------------------------------
#
# SUBMODE HANDLERS
#
#------------------------------
# You will probably NOT want/need to override these...

#------------------------------------------------------------
# submode_COMMSTART
#------------------------------------------------------------
# "Show" the page we *really* want.
# E.g., the search results, the browse results, etc.
# If the page can't be shown until a search is done,
# the search is started.
#
sub submode_COMMSTART {
   shift->comm_start(@_);
}

#------------------------------------------------------------
# submode_COMMTRACE
#------------------------------------------------------------
# Draw a single "frame" of the comm status screen (might be the
# final frame).
#
sub submode_COMMTRACE {
   my $self = shift;
   my $mode = $REQ->param('mode');
   my $subdir = '';
   $subdir = 'search' if ($mode eq 'SEARCH');

   # Get a comm status generator, complete with stop button:
   my $stop     = $self->comm_stop_url;
   my $client   = (attach IW::Client $subdir) ||
       die "couldn't attach to client!";
   my $commstat = new IW::CommStat Client=>$client, Stop=>$stop;

   $self->comm_trace($commstat);
}

#------------------------------------------------------------
# submode_COMMSTOP
#------------------------------------------------------------
# Stop the client, and draw the next "frame" of the comm status screen 
# (might be the final frame).
#
sub submode_COMMSTOP {
   my $self = shift;
   my $mode = $REQ->param('mode');
   my $subdir = '';
   $subdir = 'search' if ($mode eq 'SEARCH');

   # Get a comm status generator, with NO stop button:
   my $commstat = new IW::CommStat Client=>(attach IW::Client $subdir);
   $self->comm_stop_it($commstat);
}

#------------------------------------------------------------
# submode_COMMFINAL
#------------------------------------------------------------
# Draw the final "frame" of the comm status screen.
#
sub submode_COMMFINAL {
   my $self = shift;
   my $mode = $REQ->param('mode');
   my $subdir = '';
   $subdir = 'search' if ($mode eq 'SEARCH');

   my $client = attach IW::Client $subdir;

   if ($client->activity) {

      # Get a comm status generator, with NO stop button:
      $APP->message("This is the final communication status page ".
                 "from your last action.  Use the navigator ".
                 "links at left to continue.");
      my $commstat = new IW::CommStat Client=>$client;
      $commstat->send_doc(IsFinal=>1);
   } else {
      $APP->message("Could not find the status page from your last action. If".
                    " you have not used your account in several weeks it may ".
                    "have been deleted by the account clean up routine.");
      $self->submode_COMMINSTRUCTION;
   } 
}

#------------------------------------------------------------
# submode_COMMINSTRUCTION
#------------------------------------------------------------
# Page to describe what is required for this page to be active.
#
sub submode_COMMINSTRUCTION {

      $SESSION->{Navigator}->{Indicator} = "ViewSearch";

      $APP->start_page(Title=>"Search Status",
                       Context =>'comminstruction');

      print "<P>\nPlease use <B>Search Creation</B> Navigation link to create and start a Search ";

     $APP->end_page;
}


#------------------------------------------------------------
# contextual_help
#------------------------------------------------------------
# We actually don't want any contextualat all inside help.

sub contextual_help {
    return ();
}

#------------------------------
#
# ABSTRACT METHODS
#
#------------------------------
# You probably WILL want to override these!

#------------------------------------------------------------
# comm_preflight
#------------------------------------------------------------
# We want to show some desired page.  Must we do a search first?
# Here's where things like the browse cache come in.  
# Your override should return:
#
#      GO     if a search is necessary first
#      OK     if we can just show the 'ok' page
#      FAIL   if we should just show the 'failed' page
#
# The default is to return 'GO', which throws us into the search. 

sub comm_preflight {
   my $self = shift;
   'GO';
}

#------------------------------------------------------------
# comm_ok_page
#------------------------------------------------------------
# Draw the desired page which would be shown if a search completed
# okay.
#
sub comm_ok_page {
   my ($self, %params) = @_;

   $APP->start_page(Title=>"OKAY!");
   $HTML->t("Generic OKAY page. Override comm_ok_page() to change."); 
   $APP->end_page;
}

#------------------------------------------------------------
# comm_fail_page
#------------------------------------------------------------
# Draw the page which would be shown if a search failed.
#
sub comm_fail_page {
   my ($self, %params) = @_;

   $APP->start_page(Title=>"FAIL!");
   $HTML->t("Generic FAIL page. Override comm_fail_page() to change");
   $APP->end_page;
}

#------------------------------------------------------------
# comm_start_search
#------------------------------------------------------------
# Start the search.
#
# Your override should do whatever is necessary to set up and
# start an IW::Client.  It should return the following:
#
#      undef          Failure (you can also puke).
#
#     'STARTED'       A client was started, and should be monitored.
#
#     'DONE'          There was no need to start an IW::Client  (e.g., 
#                     a GUIDE search is done without one), so don't do
#                     any monitoring; we're done!
#
sub comm_start_search {
   my $self = shift; 

   # Because of the new navigation and partial results, granule, 
   # inventory, and directory searches run in a subdirectory.
   # If a subdirectory for searching doesn't exist, create it.
   my $searchdir = $SESSION->path_to('search');
   unless (-d $searchdir) {
      mkdir ($searchdir, 0755) or puke "mkdir $searchdir: $!";

      # HTACCESS FILE FOR THE SEARCH DIRECTORY (DENIES EVERYONE)...
      # Build a virtual .htaccess file for this session directory:
      require IU::HTAccess;
      my $access = new IU::HTAccess;
      $access->start_limit('GET', 'POST', 'PUT');
      $access->order('ALLOW', 'DENY');
      $access->deny_from('all');
      $access->end_limit;
      # Write it out:
      $access->write_in($searchdir) 
          or warn "Couldn't write .htaccess in search subdirectory.";
   }
   # Build search file:
   my $infile = $SESSION->path_to('search/search.odl');
   open SEARCH, ">$infile" or puke "open $infile: $!";
   print SEARCH <<EOF;

GROUP = DIRECTORY_SEARCH
    MESSAGE_ID = "0"
    DATA_CENTER_ID = ("STUB", "EDC", "GSFC")
END_GROUP = DIRECTORY_SEARCH

END
EOF
   close SEARCH;
    
   # Build the "query.html" file (for summarizing the search):
   my $summary_file = $SESSION->path_to('search/query.html');
   open SUMMARY, ">$summary_file" or puke "open $summary_file: $!";
   print SUMMARY "<I>Summary of last search goes here</I>\n";
   close SUMMARY;
    
   # Start the client:
   my $client = IW::Client->start(Activity => 'DIRECTORY_SEARCH',
                                  Infile   => $infile);
   $client or puke "couldn't start imsclient";

   # Now that client is started, log the last search type:
   $SESSION->state->{'LastsearchSearch'} = 'DIRECTORY_SEARCH';
    
   # Done!
   return 'STARTED';
}

#------------------------------------------------------------
# comm_finish_up STATUS
#------------------------------------------------------------
# A search completed.  Check to see if any post-processing
# needs to be done; if so, do it.   The default method does nothing.
#
# The STATUS will be "OK" or "FAIL", based on whether the search 
# finished ok or not.
#
# WARNING: this method gets invoked WHENEVER A COMMTRACE OR COMMSTOP 
# HAS DETERMINED THAT THE CLIENT IS NO LONGER RUNNING.  There is NO
# mechanism in place to ensure that it only gets run once per search!
# THEREFORE, you should test whether or not the activity needs doing
# before you actually do it!!!! 
#
sub comm_finish_up {
   my $self = shift;
   # no-op
   1;
}

#------------------------------------------------------------
# comm_stop_gracefully
#------------------------------------------------------------
sub comm_stop_gracefully {
   0;        # no! just kill it!
}

#------------------------------------------------------------
# comm_timeout
#------------------------------------------------------------
# If a search runs beyond this time (in seconds), stop it.
#
sub comm_timeout {
   my $self = shift;
   ($IU::SETUP::MaxSearchTime * 60); 
}

#------------------------------------------------------------
# comm_wait
#------------------------------------------------------------
# How long to wait between auto-refreshes of the page, in seconds.
# Get from value set in user prefs or default from SETUP.
#
sub comm_wait {
   my $self = shift;

   my $refresh_rate = eval {$SESSION->state->{RefreshRate}};
   if (defined ($refresh_rate)) {
      return $refresh_rate;
   } else {
      return $IU::UpdateEvery;
   }
}

#------------------------------------------------------------
1;
