package IW::Comm::PriceEstimateResult;


=head1 NAME

IW::Comm::PriceEstimateResult - object representing the result of a prod. req. 


=head1 NOTES

Parses the object file returned after a Price Estimate has 
been submitted. 

=cut

use IU;
use IU::Chunk;
use IU::Log;
use IU::ODL;
use IU::LineItemResult;
use IU::StatusCodes;

use IW::App;

use strict;



#------------------------------------------------------------
# new
#------------------------------------------------------------
# I<Class method, constructor.> 
#
sub new {
    my $class = shift;
    my %params = @_;
    my $self = bless \%params, $class;
    $self;
}

#------------------------------------------------------------
# from_chunk CHUNK
#------------------------------------------------------------
# I<Class method, constructor.> 
# Read a PRICE_ESTIMATE_RESULT message from a chunk.
# A browse chunk containing "x-error" is considered null and void.
#
# The browse object for success, false for failure (incluing EOF).
#
sub from_chunk {
    my ($self, $chunk) = @_;
    my $errmsg;

    # Input chunk, checking for errors:
    defined($chunk) or return undef;
    defined($errmsg = $chunk->value('x-error')) and return error 454, $errmsg;

    # Create new object, initiallizing from chunk:
    ref($self) or $self = $self->new;
    $self->{ArchiveId} = $chunk->value('DATA_CENTER_ID');

    # Status code comment:
    $self->{StatusCode}        = $chunk->value('STATUS_CODE');
    $self->{StatusCodeComment} = [$chunk->seq_values('STATUS_CODE_COMMENT')]; 

    # LineItemResult info:
    $self->{Result} = [];
    my $line_item_chunk;
    foreach $line_item_chunk ($chunk->values('LINE_ITEM_RESULT')) {
       my $line_item = IU::LineItemResult->from_chunk($line_item_chunk) || next;
       push @{$self->{Result}}, $line_item;
    }

    # Done!
    $self;
}

sub from_ODL {
   my ($self, $odl) = @_;
   $odl || return undef;
   # Be a class method, too!
   ref($self) or $self = $self->new;
   my $daac_odl;
   my $status_odl;
   my $comment_odl;
   my $line_odl;

   my ($param, $value);
   my $result;
   if ($result = $odl->param('PRICE_ESTIMATE_RESULT')) {
      if ($daac_odl = $result->param('DATA_CENTER_ID')) {
         my $daac_value = $daac_odl->text;
         $self->{ArchiveId} = $daac_value;
      }
      if ($status_odl = $result->param('STATUS_CODE')) {
         my $status_value = $status_odl->text;
         $self->{StatusCode} = $status_value;
      }


      if ($comment_odl = $result->param('STATUS_CODE_COMMENT')) {
         if ($comment_odl->type eq 'ARRAY') {
            $self->{StatusCodeComment} =
                join('?', map { $_->text } $comment_odl->as_array);
         }
         else {
            $self->{StatusCodeComment} = $comment_odl->text;
         }
      }
      if ($line_odl = $result->param('LINE_ITEM_RESULT')) {
         my $line_item = IU::LineItemResult->from_ODL($line_odl);
         push @{$self->{Result}}, $line_item;
      }
   }
   return $self;
}
#------------------------------------------------------------
# read_last
#------------------------------------------------------------
# I<Class method.>
# Read the last set of product results from the standard location 
# in the standard way.  They are returned as several PriceEstimateResult
# objects, hashed by DAAC name.
#
# WARNING: this is only going to work if the last search was 
# a Price Estimate Request!
#
sub read_last {
   my $class = shift;
   my $objpath = $SESSION->path_to('client.obj');
   my $chunk;   # the current chunk
   my $new;     # new object we'll create
   my %last;

   # Read and queue chunks:
#    debug "reading raw chunks from $objpath";
   open(OBJ, "<$objpath") or return %last;
   while ($chunk = input IU::Chunk \*OBJ) {
#	debug "Read PR chunk!";
	($chunk->value('X_TYPE') eq 'ROOT') or next;
	my $result_chunk = $chunk->value('PRICE_ESTIMATE_RESULT');
	$new = $class->from_chunk($result_chunk);
	$last{$new->archive_id} = $new;
   }
   close OBJ;
   $new->write_ODL;
   return %last;
}

sub read_ODL_hash {
   my $self = shift;
   my $path = $SESSION->path_to('estimate.odl');
   my $new;     # new object we'll create
   my %last;
   my $openstatus = open (ODL, $path);
   if ($openstatus) { 
      my $odl = eval { parse IU::ODL \*ODL };
      close ODL;
      $new = $self->from_ODL($odl);
      $last{$new->archive_id} = $new;
   # No results 
   } else {
      ref($self) or $new = $self->new;
      $new->{StatusCode} = '99';
      $new->{StatusCodeComment} = 
         'Unable to get an estimate from the data center';
      $last{'DUMMY'} = $new;
   }
   return %last;
}

sub read_ODL {
   my $self = shift;
   my $path = $SESSION->path_to('estimate.odl');
   my $new;     # new object we'll create
   open (ODL, $path) or puke "could not open $path";
   my $odl = eval { parse IU::ODL \*ODL };
   close ODL;
   $new = $self->from_ODL($odl);
   return $new;
}

sub to_ODL {
   my $self = shift;
   my $root_odl = IU::ODL->new_group;
   my $srch_odl = IU::ODL->new_group;
   $root_odl->add_param('PRICE_ESTIMATE_RESULT' => $srch_odl); 
   $srch_odl->add_param('DATA_CENTER_ID' => $self->archive_id);
   $srch_odl->add_param('STATUS_CODE' => $self->status_code);
   if (int@{$self->status_code_comment} > 0) {
      $srch_odl->add_param('STATUS_CODE_COMMENT' =>
      IU::ODL->conv($self->status_code_comment)) 
   } 
   else { 
      my $text = IU::StatusCodes::str($self->status_code) || 'Unknown';
      $srch_odl->add_param('STATUS_CODE_COMMENT' => $text);
   } 
   my $result;
   my $crit;
   foreach $result (@{$self->line_item_result}) {
      my $result_odl  = IU::ODL->new_group;
      $srch_odl->add_param(LINE_ITEM_RESULT => $result_odl);
      my $package_id = $result->package_id;
      $result_odl->add_param(PACKAGE_ID => $package_id);
      my $cost = $result->cost;
      $result_odl->add_param(EST_COST => $cost);
      foreach $crit (@{$result->specializedCriteria}) {
         my $crit_odl  = IU::ODL->new_group;
	 $result_odl->add_param(SPECIALIZED_CRITERIA => $crit_odl);
	 $crit_odl->add_param(CRITERIA_NAME => $crit->criteriaName);
	 $crit_odl->add_param(CRITERIA_TYPE => $crit->criteriaType);
	 $crit_odl->add_param(CRITERIA_VALUE => 
			      IU::ODL->conv($crit->criteriaValue));
      }
   }
   $root_odl;
}

sub write_ODL {
   my $self = shift;
   my $odl = $self->to_ODL;
   my $path = $SESSION->path_to('estimate.odl');
   open QUERY, ">$path" or puke 452, "can't open $path for writing: $!";
   $odl->output(\*QUERY);
   $odl->end(\*QUERY);
   close QUERY;
   $path;

}
#------------------------------------------------------------
# status_code
#------------------------------------------------------------
# Return the status code
#
sub status_code          { $_[0]->{StatusCode} }

#------------------------------------------------------------
# status_code_comment
#------------------------------------------------------------
# Return the status code comment, or the empty array if there is none.
#
sub status_code_comment  { $_[0]->{StatusCodeComment} || undef }

#------------------------------------------------------------
# archive_id
#------------------------------------------------------------
# Return the archive id.
#
sub archive_id   { $_[0]->{ArchiveId}; }

# line_item_result     -- return line_item_result information, or []
sub line_item_result     { $_[0]->{Result} || [] }

#------------------------------------------------------------
#------------------------------------------------------------
1;
