package IW::Comm::ProductStatusInfo;

=head1 NAME

IW::Comm::ProductStatusInfo - object representing the result of a prod. req. 

=head1 NOTES

Parses the object file returned after a Product Request has 
been submitted. 

=cut

use IU;
use IU::Chunk;
use IU::Log;
use IU::FancyURL;
use IU::OrderStatus;
use IU::OrderStatusInfo;
use IW::App;
use IW::DaacAddress;

use strict;

#------------------------------------------------------------
# new
#------------------------------------------------------------
# I<Class method, constructor.> 

sub new {
    my $class = shift;
    my %params = @_;
    my $self = bless \%params, $class;
    $self->{URLs} ||= {};
    $self;
}

#------------------------------------------------------------
# from_chunk CHUNK
#------------------------------------------------------------
# I<Class method, constructor.> 
# Read a PRODUCT_STATUS_INFO message from a chunk.
# A browse chunk containing "x-error" is considered null and void.
#
# The browse object for success, false for failure (incluing EOF).

sub from_chunk {
    my ($self, $chunk) = @_;
    my $errmsg;

    # Input chunk, checking for errors:
    defined($chunk) or return undef;
    defined($errmsg = $chunk->value('x-error')) and return error 454, $errmsg;

    # Create new object, initiallizing from chunk:
    ref($self) or $self = $self->new;
    $self->{ArchiveId} = $chunk->value('DATA_CENTER_ID');

    # Status code comment:
    $self->{StatusCode}        = $chunk->value('STATUS_CODE');
    $self->{StatusCodeComment} = $chunk->value('STATUS_CODE_COMMENT');

    # Contact info:
    my $contact_chunk = $chunk->value('DAAC_CONTACT_ADDRESS');
    if ($contact_chunk) {
        $self->{DaacContact} =
          IW::DaacAddress->from_chunk($contact_chunk);
    }
    # User Contact info:
    $self->{OrderInfo} = [];
    my $order_status_chunk;
    foreach $order_status_chunk ($chunk->values('ORDER_STATUS_INFO')){
        my $orders = 
          IU::OrderStatusInfo->from_chunk($order_status_chunk) || next;
        push @{$self->{OrderInfo}}, $orders;
    }

    # Done!
    $self;
}

#------------------------------------------------------------
# read_last
#------------------------------------------------------------
# I<Class method.>
# Read the last set of product results from the standard location 
# in the standard way.  They are returned as several ProductStatusInfo
# objects, hashed by DAAC name.
#
# WARNING: this is only going to work if the last search was 
# a product request!

sub read_last {
    my $class = shift;
    my $objpath = $SESSION->path_to('client.obj');
    my $chunk;   # the current chunk
    my $new;     # new object we'll create
    my %last;

    # Read and queue chunks:
    debug "reading raw chunks from $objpath";
    open(OBJ, "<$objpath") or puke "couldn't read $objpath: $!";
    while ($chunk = input IU::Chunk \*OBJ) {
	debug "Read PR chunk!";
	($chunk->value('X_TYPE') eq 'ROOT') or next;
	my $result_chunk = $chunk->value('PRODUCT_STATUS_INFO') || 
	    puke 513, "your last search was not a product request";
	$new = $class->from_chunk($result_chunk);
	$last{$new->archive_id} = $new;
    }
    close OBJ;
    (keys %last) or puke "no product results could be obtained!";
    return %last;
}

#------------------------------------------------------------
# status_code
#------------------------------------------------------------
# Return the status code

sub status_code          { $_[0]->{StatusCode} }

#------------------------------------------------------------
# status_code_comment
#------------------------------------------------------------
# Return the status code comment, or the empty string if there is none.

sub status_code_comment  { $_[0]->{StatusCodeComment} || '' }

#------------------------------------------------------------
# contact
#------------------------------------------------------------
# Return the contact address, an IW::DaacAddress.

sub contact      { $_[0]->{DaacContact} }

#------------------------------------------------------------
# archive_id
#------------------------------------------------------------
# Return the archive id.

sub archive_id   { $_[0]->{ArchiveId}; }

#------------------------------------------------------------
# URLs [TYPE]
#------------------------------------------------------------
# Return the array of URLs of the given TYPE (as IU::FancyURLs), 
# or empty if none.
#
# If TYPE is not given, returns an array of available types.

sub URLs { 
    my ($self, $type) = @_;
    @_ > 1 or return (keys %{$self->{URLs}});  
    @{$self->{URLs}{$type} || []};
}

# orderInfo     -- return orderInfo information, or []
sub orderInfo     { $_[0]->{OrderInfo} || [] }

#------------------------------------------------------------
1;





