package IW::Comm::ProfileUpdateResult;


=head1 NAME

IW::Comm::ProfileUpdateResult - object representing the result of a 
profile retrieval request.


=head1 NOTES

Parses the object file returned after a Profile Update Request has 
been submitted. 

=cut

use IU;
use IU::Chunk;
use IU::Log;
use IU::Addr;
use IW::App;

use strict;



#------------------------------------------------------------
# new
#------------------------------------------------------------
# I<Class method, constructor.> 
#
sub new {
    my $class = shift;
    my %params = @_;
    my $self = bless \%params, $class;
    $self;
}

#------------------------------------------------------------
# from_chunk CHUNK
#------------------------------------------------------------
# I<Class method, constructor.> 
# Read a PROFILE_UPDATE_RESULT message from a chunk.
# A browse chunk containing "x-error" is considered null and void.
#
# The browse object for success, false for failure (incluing EOF).
#
sub from_chunk {
    my ($self, $chunk) = @_;
    my $errmsg;

    # Input chunk, checking for errors:
    defined($chunk) or return undef;
    defined($errmsg = $chunk->value('x-error')) and return error 454, $errmsg;

    # Create new object, initiallizing from chunk:
    ref($self) or $self = $self->new;
    $self->{ArchiveId} = $chunk->value('DATA_CENTER_ID');

    # Status code comment:
    $self->{StatusCode}        = $chunk->value('STATUS_CODE');
    $self->{StatusCodeComment} = $chunk->value('STATUS_CODE_COMMENT');

    # Done!
    $self;
}

#------------------------------------------------------------
# read_last
#------------------------------------------------------------
# I<Class method.>
# Read the last set of profile update results from the standard location 
# in the standard way.  They are handled as a single ProfileUpdateRequest.
# 
#
# WARNING: this is only going to work if the last search was 
# a profile update request!
#
sub read_last {
    my $class = shift;
    my $objpath = $SESSION->path_to('client.obj');
    my $chunk;   # the current chunk
    my $result;     # new object we'll create

    # Read and queue chunks:
    debug "reading raw chunks from $objpath";
    open(OBJ, "<$objpath") or puke "couldn't read $objpath: $!";
    while ($chunk = input IU::Chunk \*OBJ) {
	debug "Read PR chunk!";
	($chunk->value('X_TYPE') eq 'ROOT') or next;
	my $result_chunk = $chunk->value('PROFILE_UPDATE_RESULT') || 
	    puke 513, "Your search did not contain a profile update result.";
	$result = $class->from_chunk($result_chunk);
    }
    close OBJ;
    return ($result);
}


#------------------------------------------------------------
# status_code
#------------------------------------------------------------
# Return the status code
#
sub status_code          { $_[0]->{StatusCode} }

#------------------------------------------------------------
# status_code_comment
#------------------------------------------------------------
# Return the status code comment, or the empty string if there is none.
#
sub status_code_comment  { $_[0]->{StatusCodeComment} || '' }

#------------------------------------------------------------
# archive_id
#------------------------------------------------------------
# Return the archive id.
#
sub archive_id   { $_[0]->{ArchiveId}; }

#------------------------------------------------------------
#------------------------------------------------------------
1;
