package IW::Comm::UserPasswordChangeResult;

=head1 NAME

IW::Comm::UserPasswordChangeResult - object representing the result of
    a user password change request.  This is known as a User Password 
    Change Result message.

=head1 NOTES

Parses the object file returned after a User Password Change Request has 
been submitted.  

=cut

use IU;
use IU::Chunk;
use IU::ODL;
use IU::Log;
use IU::Addr;
use IW::App;

use strict;

#------------------------------------------------------------
# new
#------------------------------------------------------------
# I<Class method, constructor.> 

sub new {
    my $class = shift;
    my %params = @_;
    my $self = bless \%params, $class;
    $self;
}

#------------------------------------------------------------
# from_chunk CHUNK
#------------------------------------------------------------
# I<Class method, constructor.>
# Read a USER_PASSWORD_CHANGE_RESULT message from a chunk.
# A browse chunk containing "x-error" is considered null and void.
#
# The browse object for success, false for failure (incluing EOF).
#
sub from_chunk {
    my ($self, $chunk) = @_;
    my $errmsg;

    # Input chunk, checking for errors:
    defined($chunk) or return undef;
    defined($errmsg = $chunk->value('x-error')) and return error 454, $errmsg;

    # Create new object, initiallizing from chunk:
    ref($self) or $self = $self->new;
    $self->{ArchiveId} = $chunk->value('DATA_CENTER_ID');

    # Status code comment:
    $self->{StatusCode}        = $chunk->value('STATUS_CODE');
    $self->{StatusCodeComment} = $chunk->value('STATUS_CODE_COMMENT');

    # Done!
    $self;
}

#------------------------------------------------------------
# read_last
#------------------------------------------------------------
# I<Class method.>
# Read the last set of product results from the standard location 
# in the standard way.  They are returned as several
# UserPasswordChangeRequest objects, hashed by DAAC name.
#
# WARNING: this is only going to work if the last search was 
# a product request!

sub read_last {
    my $class = shift;
    my $objpath = $SESSION->path_to('client.obj');
    my $chunk;   # the current chunk
    my $result;    # Create new object for result.

    # Read and queue chunks:
    open(OBJ, "<$objpath") or puke "couldn't read $objpath: $!";
    while ($chunk = input IU::Chunk \*OBJ) {
       ($chunk->value('X_TYPE') eq 'ROOT') or next;
       my $result_chunk = $chunk->value('USER_PASSWORD_CHANGE_RESULT') ||
           puke 513, 'Failed to change your password. ',
                     'Please try again later.';
       $result = $class->from_chunk ($result_chunk);
    }
    close OBJ;
    return ($result);
}

#------------------------------------------------------------
# status_code
#------------------------------------------------------------
# Return the status code
#
sub status_code          { $_[0]->{StatusCode} }

#------------------------------------------------------------
# status_code_comment
#------------------------------------------------------------
# Return the status code comment, or the empty string if there is none.
#
sub status_code_comment  { $_[0]->{StatusCodeComment} || '' }

#------------------------------------------------------------
# archive_id
#------------------------------------------------------------
# Return the archive id.
#
sub archive_id   { $_[0]->{ArchiveId}; }

#------------------------------------------------------------
1;
