package IW::CommStat;

=head1 NAME

IW::CommStat - output the communication status screens

=head1 DESCRIPTION

Object responsible for generating the communication status screens
for either the currently-active or the most-recent communication
with the client.

I<This does not track communication, it just reports it.>
To track status, use IW::Client.

This is an ABSTRACT CLASS; you will actually be working with
instances of one of the subclasses named for the type of
transaction taking place; e.g., IW::CommStat::INVENTORY_SEARCH. 
These are primarily used to determine whether the operation *really*
succeeded or not.

=head1 CLIENT/OPERATION STATUSES

Upon completion, the client itself can have any one of a number of
possible B<client statuses>... from these, and some additional 
criteria (e.g., the number of granules recovered), we get a number
of possible B<operation statuses>:

    Client     Operation
    Status     Status         Description
    ------------------------------------------------------------------
    STOPPED    STOPPED        Gracefully stopped.
    
    KILLED     KILLED         Forcibly killed (possibly after a failed 
                              attempt to stop it).
    
    CRASHED    CRASHED        The imsclient/imsrunclient halted
                              prematurely: either they crashed, or they
                              failed to run in the first place.
        
    (other)    OP_FAILED      Operation failed.  This is operation-
                              dependent: For example, a search where no
                              connection returns any granules.
     
    (other)    OP_SUCCEEDED   Operation succeeded.  Like OP_FAILED, 
                              this is operation dependent.

=head1 INSTANCE VARIABLES

    Client            Client object 
    OpName            Operation name: "search", "order", etc.
    Stop              URL for the "STOP" button
    ReportGranules    Report granules received?


=head1 RCSID

$Id: CommStat.pm,v 1.88 2002/11/13 16:27:49 solomon Exp $

=cut

use IU;
use IU::Log;
use IU::Dataset;
use IU::InvStore;
use IU::Event;
use IW::App;
use IW::Client;
use IW::Util qw(:url);

use Text::Wrapper;

use strict;

# Colors to use:
my %Color = (TITLE=>'#D0D0D0',
             ICON =>'#D0D0D0',
             ROW  =>'#FFFFCC');

#------------------------------------------------------------
# new PARAMHASH
#------------------------------------------------------------
# Create a new comm status object, for monitoring a client.  
# The params are:
#
#     Client   => The IW::Client interface object (obtained via either
#                 a "start" or an "attach").  The client's activity
#                 must be known!
#     OpName   => Printable name of the current operation.
#     Stop     => URL for the "stop" button, if any.
#
# If activity is known in the client, an instance of the appropriate 
# subclass of this will be returned; otherwise, you'll get a fatal 
# error.

sub new {
   my ($class, %params) = @_;
   
   # Get the client:
   my $client = $params{Client} ||
       puke 600, "missing client argument!";
   
   # Get the activity, so we know what to bless ourselves as:
   my $activity = $client->activity ||
       puke "client is missing activity!";
   debug "new comm stat: activity = $activity";
   
   # Construct!
   # If activity was given, bless as THAT SUBCLASS;
   # else, two-arg bless:
   my $pkg = "IW::CommStat::\U$activity\E";
   eval "require $pkg"; die $@ if $@;
   my $self = bless {}, $pkg;
   debug "self = $self";
   
   # Init:
   $self->{Client} = $client;
   $self->{OpName} = $params{OpName} || 'operation';
   $self->{Stop}   = $params{Stop};
   $self->{ReportGranules} = 0;
   $self->{Activity} = $activity;
   $self->init(@_);      # subclasses should override this
}

#------------------------------------------------------------
# init PARAMHASH
#------------------------------------------------------------
# Init a new comm status object.

sub init {
   $_[0];       # no-op
}

#------------------------------------------------------------
# as_HTML
#------------------------------------------------------------
# Return a HTML version of the status screen.
# Does not show successful entries, only those still running or
# with bad status.

sub as_HTML {
   my ($self, %params) = @_;

   # Get connections:
   my @conns = sort { $a->daac cmp $b->daac } $self->client->conn_statuses;

   my $table_open = 0;
   foreach my $conn (@conns) {
      # Check the status and if = 1, skip printing entry.
      next if ((($conn->status == 1) || ($conn->status == 21)) && 
       ($conn->stage == 4));
      # Print the connection statuses header:
      unless ($table_open) {
         $HTML->BR->TABLE(CELLSPACING=>2, CELLPADDING=>3, BORDER=>0)
              ->TR
                ->TD(ALIGN=>"LEFT",VALIGN=>"TOP",COLSPAN=>2,
                     BGCOLOR=>$Color{ICON})
                  ->FONT(SIZE=>'+1')
                  ->t_B("Note: The following data center messages " .
                        "were returned:")
                  ->_FONT
                ->_TD
              ->_TR;
         $table_open = 1;
      }
      $HTML->TR
             ->TD(BGCOLOR=>$Color{ICON})
               ->t($conn->daac)
             ->_TD
             ->TD(BGCOLOR=>$Color{ICON})
               ->t($conn->status_cmnt)
             ->_TD
           ->TR;
   }
   $HTML->_TABLE ->BR if $table_open;
}


#------------------------------------------------------------
# as_string
#------------------------------------------------------------
# Return a stringified version of the status screen.
# Nice for putting into email messages.

sub as_string {
   my $self = shift;
   
   # Get connections:
   my @conns = sort { $a->daac cmp $b->daac } 
   $self->client->conn_statuses;
   
   # Print the connection statuses:
   my $conn;

   my $str;
   foreach $conn (@conns) {
# check the status and if = 1 then stage_cmt = "Order Received".  If query was
# unsuccessful, the status will be > 1. See IW::Conn. 

      my $stage_cmnt;
      my $status_cmnt;
      if($conn->status == 1){
        $stage_cmnt = "Order Received";
       
      }
      else {
        $stage_cmnt = $conn->stage_cmnt;

      }
      $status_cmnt = $conn->status_cmnt;
#      $status_cmnt =~ s/[\r\n]/ /g;
# as_string is not used by CommStat. It is used in order receipt, email, and
# product status request. Therefore the stage_cmnt is needed 
      $str .= sprintf("%-12s %-14s \n  %s\n\n", "Data Center: ",
                      $conn->daac, $status_cmnt);

   }
   $str;
}
#------------------------------------------------------------
# as_string_by_daac
#------------------------------------------------------------
# Return a stringified version of the status for one data center.
# Nice for putting into order status and email messages.

sub as_string_by_daac {
   my ($self, $daac) = @_;
 
   # Get connections:
   my @conns = sort { $a->daac cmp $b->daac } 
   $self->client->conn_statuses;
   
   # Print the connection statuses:
   my $conn;
   my $str = "No Connection";
   my $status_cmnt;
   foreach $conn (@conns) {
     if ($conn->daac eq $daac) {
         $status_cmnt = $conn->status_cmnt;
         $status_cmnt =~ s/[\r\n]/ /g;
         if($conn->status == 1){
            
            $str = sprintf("%s %s %s",
                       "Order Received"," / ", $status_cmnt);
         }
         else {
            
            $str = sprintf("%s", $status_cmnt);
      
         }
     }
   }

   $str;
}
#------------------------------------------------------------
# client
#------------------------------------------------------------
# Return the underlying client interface.

sub client {
   my $self = shift;
   $self->{Client};
}

#------------------------------------------------------------
# finished_ok
#------------------------------------------------------------
# We're finished... but did we really finish okay?
# The default test is that EVERY CONNECTION must have finished okay,
# AND we must have at least ONE CONNECTION!!!!!!
#
# Subclasses will generally override this (although they may employ
# the default behavior PLUS some additional tests). 

sub finished_ok { 
   my $self = shift;
   my $conn;
   
   debug "CommStat: using default finished_ok test";
   
   # Get connections:
   my @conns = $self->client->conn_statuses;
   @conns or return error "couldn't have finished ok: no connections!";
   
   # Check em all:
   foreach $conn (@conns) {
      $conn->finished_ok or return 0;
   }
   # To prevent a rare-but-observed race condition, the default
   # behavior is that KILLED clients are untrustworthy, and did
   # NOT finish okay.  Inventory and directory searches override
   # this method, so we're okay. [Added for 1.9 by eryq]
   if ($self->client->status eq 'KILLED') {
      inform "Hmmm... a killed client means we didn't finish okay";
      return 0;
   }
   1;
}

#------------------------------------------------------------
# get_final_op_status
#------------------------------------------------------------
# The operation is done.  But how did we do?  Returns the
# operation status (as described at the top of this file).

sub get_final_op_status {
   my ($self) = @_;
   
   # Get client and client status:
   my $client = $self->client;
   my $client_status = $client->status || '';
   
   # Sum up the number of granules and datasets (important!):
   $self->n_granules;
   $self->n_datasets;
   
   # Determine the operation status:
   my $op_status;
   
   # If client is toast:
   if ($client_status =~ /^(STOPPED|KILLED|CRASHED)$/) {
      $op_status = $client_status;             
   }
   elsif ($self->finished_ok) {                    # OP_SUCCEEDED
      $op_status = 'OP_SUCCEEDED';
   }   
   else {                                          # OP_FAILED
      $op_status = 'OP_FAILED';
   }
   return $op_status;
}

#------------------------------------------------------------
# n_datasets
#------------------------------------------------------------
# Calculate and return the number of datasets received, totalled 
# across all connections.

sub n_datasets { 
   my $self = shift;    
   if($SESSION->state->{Debug}) {
      my $conn;
      my $n = 0;
      foreach $conn ($self->client->conn_statuses) { 
         $n += $conn->n_datasets;
      }
      debug "sum datasets = $n"; 
      return $n;
   } else {
      my $results;
      my $datasets;

      # Attempt to connect to the results database.
      eval {
          $results = undef;
          return unless -e $SESSION->path_to('results.gdb');
           $results = IU::InvStore->
               open($SESSION->path_to('results.gdb'), 'READER');
      };
      unless (defined $results) {
         return 0;
      }
      # It's there. Read it and report on the count.
      my @datasets = $results->datasets;
      foreach (@datasets) {
         $datasets++  # count datasets with zero granules as well.
      }
   return $datasets;
   }
}

#------------------------------------------------------------
# n_granules
#------------------------------------------------------------
# Calculate and return the number of granules received, totalled 
# across all connections.

sub n_granules { 
   my $self = shift;
   if($SESSION->state->{Debug}) {
      my $conn;
      my $n = 0;
      foreach $conn ($self->client->conn_statuses) {
         $n += $conn->n_grans;
      }
      debug "sum granules = $n"; 
      return $n;
   } else {
      my $results;
      my $granules;
      my $datasets;

      # Attempt to connect to the results database.
      eval {
          $results = undef;
          return unless -e $SESSION->path_to('results.gdb');
           $results = IU::InvStore->
               open($SESSION->path_to('results.gdb'), 'READER');
      };
      unless (defined $results) {
	 return 0;
      }
      # It's there. Read it and report on the count.
      my @datasets = $results->datasets;
 
      # Give us the count of Datasets and Granules.
      if ($SESSION->state->{LastSearchType} ne 'DIRECTORY_SEARCH') {
          foreach (@datasets) {
              $granules += $_->n_granules;
              $datasets++  # count datasets with zero granules as well.
          }
      }
      elsif ($SESSION->state->{LastSearchType} eq 'DIRECTORY_SEARCH') {
          foreach (@datasets) {
              $datasets++ ;
          }
          $granules = 0;
      }
   return $granules;
   }
}

#------------------------------------------------------------
# print_directions OPSTATUS
#------------------------------------------------------------
# What should the user do now?

sub print_directions {  
   1;
}

#------------------------------------------------------------
# print_generic_directions OPSTATUS
#------------------------------------------------------------
# What *can* the user do now?

sub print_generic_directions {  
   print "If you have questions regarding a particular data center's ".
         "results, you may locate user support for that data center ".
         "by clicking on the center's name below.";
   1;
}

#------------------------------------------------------------
# print_stop_button 
#------------------------------------------------------------
# Print a substitute for the normal navigator.
# If Wait is given, the user is alerted of an automatic reload
# in Wait seconds.

sub print_stop_button {
   my ($self, %params) = @_;
   
   # Can't do it if no STOP url!
   $self->{Stop} or return 0;
   my $name = $self->{OpName};
   $name =~ tr/a-z/A-Z/;
   $HTML -> BR;
   if ($self->client->activity =~ /^(INVENTORY_SEARCH)/) {
   $HTML -> A(HREF=>IW::Util::script_url('mode'=>'DSETLIST'))
         -> t("VIEW PARTIAL SEARCH RESULTS NOW")
         ->_A;
   $HTML-> HR;
   }
   $HTML -> hypertext(" END $name NOW",$self->{Stop});
   $HTML -> BR;
   $HTML -> BR;
   # The rule and the button:

   if($SESSION->state->{Debug}) {
      if ($params{Wait}) {
         $HTML -> B
               -> t("This page will reload automatically every ")
               -> t("$params{Wait} seconds until the search is done. ")
               -> _B;
         $HTML -> BR;
         $HTML -> BR;
      }
   }
   1;
}

#------------------------------------------------------------
# print_status_comments
#------------------------------------------------------------
# Output the dataset comments.  Each must be preformatted, and 
# organized by DAAC.
#
# We only do this as the final page of an INVENTORY or DIRECTORY
# search, so there's code here of a questionable nature.

sub print_status_comments {
   my $self = shift;
   
   # GET CONNECTIONS...
   
   # Get the conn statuses, sorted by DAAC:
   my @conns = sort { $a->daac cmp $b->daac }
   $self->client->conn_statuses;
   
   # GET DATASETS (IF NECESSARY)...

   my %dsets;
   my $dset;
   my @datasets;
   my $report_datasets = ($self->client->activity =~
                          /^(INVENTORY_SEARCH|DIRECTORY_SEARCH)$/);
   if ($report_datasets) {
      
      # Log into RESULTS database, and get all datasets:
      my $RESULTS = eval {
         return undef unless $SESSION->path_to('results.gdb');
       IU::InvStore->open($SESSION->path_to('results.gdb'),
                          'READER');
      };
      $RESULTS or return undef;
      @datasets = $RESULTS->datasets;
      $RESULTS->close;
      ### @datasets or return undef;          # nothing to say
      
      # Hash the datasets by DAAC:
      foreach $dset (@datasets) {
         my $daac = $dset->archive_id;
         $dsets{$daac} ||= [];
         push @{$dsets{$daac}}, $dset;
      }
   }
   
   # Done!
   1;
}

#------------------------------------------------------------
# print_tracking_info 
#------------------------------------------------------------
# If a tracking number was attached to this request, output it.

sub print_tracking_info {  
   my $self = shift;
   if ($self->tracking_no) {
      $HTML -> HR;
      print "<B>If you want to follow up on problems, your ",
            "tracking number for this request is: </B><BR>";
      print "<FONT SIZE=+2>";
      print "<B>", $self->tracking_no, "</B> ($ENV{SERVER_NAME})\n";
      print "</FONT>\n";
   }
}

#------------------------------------------------------------
# send_doc OPTIONHASH
#------------------------------------------------------------
# Send a single "page" of the communication status document.
# Options are:
#
#   IsFinal      => indicates whether this is the last page
#   KilledByUser => the user just elected to kill the search themselves
#   Stopping     => if true, a halt is in progress
#   Wait         => pause this long, then reload 

sub send_doc {
   my ($self, %options) = @_;
   my ($conn, $null_conn);
   my ($proc, $null_proc);
   my $final_page;
   # Get the granule key:
   my $gkey = $REQ->param('gkey');
   my $newmode = $REQ->param('newmode');
   my $newsubmode = $REQ->param('newsubmode');

   # Poll for the current statuses:
   my $client = $self->client;
   $client->poll;    

   # Get the conn statuses:
   my @conns = sort { $a->daac cmp $b->daac } $client->conn_statuses;
   @conns or $null_conn = IW::Conn->new->null;     # no info available

   # Get the post-processing statuses:
   my @procs = $client->proc_statuses;
   @procs or $null_proc = IW::PostProcess->new->null;
             # no info available

   # Start the page.
   # Are we running, or are we done?  Print header appropriately:
   # if ($final_page = ($options{KilledByUser} || $options{IsFinal})) {
   if ($final_page = $options{IsFinal}) {
      $self->start_final_page(\%options);
   }
   else {
      $self->start_nonfinal_page(\%options);
   }
   if($gkey) {
      $HTML -> hidden('gkey'    => $gkey);
   }
   if($newmode) {
      $HTML -> hidden('newmode'    => $newmode);
   }
   if($newsubmode) {
      $HTML -> hidden('newsubmode'    => $newsubmode);
   }

   # CONNECTION STATUSES... [ICON DAAC NUMGRANS POT STATUS]

   # Print the titles for the connection statuses:
   my $infocols = 0;
   $HTML -> TABLE(CELLSPACING=>0, CELLPADDING=>3, BORDER=>0);
   $HTML -> TR(BGCOLOR=>$Color{TITLE})
         -> TH -> _TH        # icon
         -> TH -> t('Data')->BR->t('Center') -> _TH;    
   if ($self->{ReportDatasets}) {
      ++$infocols;
      $HTML -> TH -> small -> t('Data Sets')->BR->t('Returned') 
                  -> _small
            -> _TH;
   }
   if ($self->{ReportGranules}) {
      ++$infocols;
      $HTML -> TH -> small -> t('Data Granules')->BR->t('Returned') 
            -> _small
            -> _TH;
   }
   $HTML -> TH -> e('nbsp') -> _TH        # pot
         -> TH -> t('Status')-> _TH
         -> TH(ALIGN=>'CENTER') -> t('Data Center')->BR->t('Comments')
         -> _TH
         -> _TR;

   # Print the connection statuses:
   foreach $conn (@conns ? @conns : $null_conn) {
      my %connstate = ();
      $HTML -> TR(VALIGN=>'TOP',BGCOLOR=>$Color{ROW});

      # Icon:
      $HTML -> TD(BGCOLOR=>$Color{ICON});
      print $conn->icon;
      $HTML -> t("\xA0"); 
      $HTML -> _TD;

      # DAAC:
      $HTML -> TD;    
      my $contact_url = ("$IU::ImsUserUrl#" . $conn->daac)
         if $final_page;
      $HTML -> hypertext($conn->daac||"\xA0", $contact_url);
      $connstate{DAAC} = $conn->daac;
      $HTML -> _TD;

      # Datasets, if reporting:
      if ($self->{ReportDatasets}) {
         my $ngot = $conn->n_datasets;
         $connstate{Datasets} = $ngot;
         $HTML -> TD(NOWRAP=>undef, ALIGN=>'CENTER')-> t($ngot);
         $HTML -> _TD;
      }

      # Granules, if reporting:
      if ($self->{ReportGranules}) {
         my $ngot = $conn->n_grans;
         my $nhit = $conn->n_gran_hits;
         $connstate{Granules} = $ngot;
         $connstate{Hits} = $nhit;
         my $grans = $ngot . ($nhit ? " (of $nhit)" : '');
         $HTML -> TD(NOWRAP=>undef, ALIGN=>'CENTER') -> t($grans);
         $HTML -> _TD;
      }

      # Potentiometer and stage comment:
      ### WARNING: DO NOT PRESENT COMMENTS AS A FORM ELEMENT IN A
      ### MULTIPART DOCUMENT, BEFORE THE LAST PAGE!!!!!  
      ### NETSCAPE WILL SCREW IT UP!!!!
      $HTML -> TD(NOWRAP=>undef);
      $HTML -> t("\xA0"); 
      $HTML -> _TD;
      $HTML -> TD;
      $conn->print_pot();
      $HTML -> _TD;
      #$HTML -> TD(NOWRAP=>undef);
      #show_comment($conn->stage_cmnt,'TEXTSIZE'=>'small',
      #             'BREAK'=>'yes');
      #$HTML -> t("\xA0");
      #$HTML -> _TD;

      # Status comment:
      $HTML -> TD(NOWRAP=>undef);
      show_comment($conn->status_cmnt,'BREAK'=>'yes');
      $HTML -> _TD;
      $connstate{StatusComment} = $conn->status_cmnt;
      $connstate{Status} = $conn->status;
      ims_event('COMMSTAT_CONNS',%connstate);
   }

   foreach $proc (@procs) {
     my %procstate = ();
     $procstate{Activity} = $proc->activity;
     $procstate{Stage}    = $proc->stage;
     $procstate{Granules} = $proc->n_grans;
     $procstate{Datasets} = $proc->n_datasets;
     ims_event('COMMSTAT_PROCS',%procstate);
   }

   $HTML -> _TABLE;

   # Tracking info:
   $self->print_tracking_info;

   # DAAC and dataset status code comments...
   $self->print_status_comments if $final_page;

   # End it and send it:
   my $subdir = "search";
   $APP->print_last_query($subdir) if $final_page;
   $APP->suppress_key(1);  
   $APP->end_page;
   1;
}

#------------------------------------------------------------
# show_comment
#------------------------------------------------------------
# Private.  Format a comment nicely.

sub show_comment {
   my ($text, %param) = @_;
   $text or return;     # no-op if no text
   my $line;

   # Set up defaults
   my $linesize = ($param{'LINESIZE'} or '60');
   $param{'TEXTSIZE'} = "normal" unless defined $param{'TEXTSIZE'};
   $param{'BREAK'} = 0 unless defined $param{'BREAK'};

   # Create text formatting object and format text.
   my $w = new Text::Wrapper('columns'=>$linesize);
   $text = $w->wrap($text);

   # Output:
   $HTML -> small  if $param{'TEXTSIZE'} eq 'small';
   foreach $line (split /\n/, $text) {
      # Give us a little breathing room ... 
      $HTML -> e('nbsp') -> e('nbsp') ->t($line);
      if ($param{BREAK}) {
         $HTML -> BR;
      }
      else {
         $HTML -> t(" ");
      }
   }
   $HTML -> _small if $param{'TEXTSIZE'} eq 'small';
   1;
}

#------------------------------------------------------------
# tracking_no
#------------------------------------------------------------
# Get/set the tracking number.

sub tracking_no {
   my $self = shift;
   $self->{TrackingNo} = shift if (@_);
   $self->{TrackingNo};
}

#------------------------------------------------------------
# start_final_page OPTS
#------------------------------------------------------------
# Operation is done.  Print the bit above the DAAC list.
#
# If this function is called (because the standalone client stopped), 
# and not all the final stages have been reached, then we assume the
# standalone has crashed.  Warn the user, but continue.

sub start_final_page {
   my ($self, $options) = @_;

   # Get the final operation status:
   my $op_status = $self->get_final_op_status;

   # Detect crash of standalone client, and report as a note.
   # (We must do this first, before starting the page):
   if ($op_status eq 'CRASHED') {
      $APP->note('ERROR', 
                 "The standalone client has apparently crashed, and ".
                 "so none of the results may be trusted. PLEASE ".
                 "ALERT THE DEVELOPERS.");
   }
   # Set the navigation indicator to view search tab if inventory or
   # directory search; to order data tab if product request.
   if ($self->{Activity} eq 'INVENTORY_SEARCH' or
       $self->{Activity} eq 'DIRECTORY_SEARCH') {
      $SESSION->{Navigator}->{Indicator} = "ViewSearch";
   } elsif ($self->{Activity} eq 'PRODUCT_REQUEST') {
      $SESSION->{Navigator}->{Indicator} = "OrderData";
   }
   # Start the page:
   if (($op_status eq 'STOPPED') || ($op_status eq 'KILLED')) {
                                                 # STOPPED/KILLED

      $APP->start_page(SuperTitle => 'Search Status',
                       Title=>"\u$self->{OpName} Stopped",
                       Context =>'srchfail');
   }
   elsif ($op_status eq 'CRASHED') {            # CRASHED 
      $APP->start_page(SuperTitle => 'Search Status',
                       Title=>"\u$self->{OpName} Crashed",
                       Context =>'srchfail');
   }
   elsif ($op_status eq 'OP_FAILED' || $op_status eq 'OP_SUCCEEDED') {     
      $APP->start_page(SuperTitle => 'Search Status',
                       Title=>"\u$self->{OpName} Completed",
                       Context =>'srchdone');
   }
   else {
      puke "Bad final operation status ($op_status): were we done???";
   }
   # What should they do now?
   $self->print_directions($op_status);

   # How did it go?    
   my $summary;
   if ($summary = $self->summary($op_status)) {
      print "<P><hr>\n";
      $summary->print_table;
   }
   1;
}

#------------------------------------------------------------
# start_nonfinal_page OPTS
#------------------------------------------------------------
# Search is not yet done.  Print the bit above the DAAC list.

sub start_nonfinal_page {
   my ($self, $options) = @_;

   my $client = $self->client;

   # Create a reply head:
   my $head = new IW::ReplyHead;
   $head->okay if $REPLY->need_status_line;
   $head->content_type('text/html');
### $head->no_cache;     ### LEAVE THIS OUT!

   # Pick a good title:
   my $title = ($options->{Stopping} 
                ? "Stopping $self->{OpName}..."
                : "\u$self->{OpName} in progress...");    

   my $wait = $options->{Wait};

   $self->start_page(IsFinal    => $options->{IsFinal},
                     SuperTitle => 'Search Status',
                     Title      => $title,
                     ReplyHead  => $head,
                     NoLogOut   => 'YES',
                     Context    => 'srchrunning',
                     ($wait ? (Refresh=>{Wait=>$wait}) : ())
                    );
   # The "STOP" button:
#   $self->print_stop_button(Wait=>$wait) unless $options->{Stopping};
#   print "<hr>\n";

   # Get approximate start time of operation, and time now:
   my $time_sec = $client->run_time;
   my $time_str = sprintf("%02dm %02ds", ($time_sec/60),
                          ($time_sec%60));
   my $drop_dead = $client->max_search_time();
   my $drop_dead_sec = $drop_dead*60;
   my $how_long  = sprintf("%02dm %02ds", 
                           ((($drop_dead*60) - $time_sec)/60),
                           ((($drop_dead*60) - $time_sec)%60));

   my $mins = ($drop_dead == 1 ? "minute" : "minutes");
   my $html;
   if ($drop_dead_sec - $time_sec < 0) {       # time is up!
      $html = "Your $self->{OpName} has exceeded the requested ".
              "maximum search time of <b>$drop_dead</b> minutes ".
              "and is being cancelled.";
   }
   else {                      # show how long is left
      $html = "Your $self->{OpName} has been running for ".
              "<b>$time_str</b> of real time.";
   }
   $HTML -> BR;
   IW::Note->new(Class => 'RUNNING',
               HTML  => $html)
         ->print_table;
   # The "STOP" button:
   $self->print_stop_button(Wait=>$wait) unless $options->{Stopping};
   1;
}

#------------------------------------------------------------
# start_page PARAMHASH
#------------------------------------------------------------
# Set up navigator and call App's start_page.

sub start_page {
   my $self = shift;
   my %params = @_;

   # Suppress the navigator links (browse, account status, etc).
   $APP->suppress_links(1) unless $params{IsFinal};

   # Set the navigation indicator to order data tab if product request.
   if ($self->{Activity} eq 'PRODUCT_REQUEST') {
      $SESSION->{Navigator}->{Indicator} = "OrderData";
   }
   $APP->start_page(%params);
}

#------------------------------------------------------------
# summary OPSTATUS
#------------------------------------------------------------
# What's the bottom line?  Return an IW::Note.

sub summary { 
   my ($self, $op_status) = @_;
 
   # Get approximate start and stop time of operation:
   my $time_sec = $self->client->run_time;
   my $time_str = sprintf("%02dm %02ds", ($time_sec/60),
                          ($time_sec%60));

   # Set up a summary of how things went:
   my $summary;
   if ($op_status eq 'STOPPED') {
      $summary = new IW::Note 
          Class=>($self->finished_ok ? 'OK' : 'FAILED'),
          HTML=>"Your $self->{OpName} was stopped ".
                "after <b>$time_str</b> of real time.";
   }
   elsif ($op_status eq 'KILLED') {        # STOPPED BAD, OR KILLED
      $summary = new IW::Note 
          Class=>'FAILED',
          HTML=>"Your $self->{OpName} was killed (no results ".
                "obtainable) after <b>$time_str</b> of real time.";
   }
   elsif ($op_status eq 'CRASHED') {            # CRASHED
      $summary = new IW::Note
          Class=>'FAILED',
          HTML=>"Your $self->{OpName} <b>halted prematurely</b> ".
                "after <b>$time_str</b> of real time.";
   }
   elsif ($op_status eq 'OP_FAILED') {          # FAILED
          $HTML->P; 
   }
   else {                                       # SUCCESS!
      $summary = new IW::Note 
          Class=>'OK',
          HTML=>"Your $self->{OpName} finished successfully ".
                "after <b>$time_str</b> of real time.";
   }
   $summary;
}

#------------------------------------------------------------

### Navigator support

#------------------------------------------------------------
# These two routines are required for any class that will be
# called back to by Navigator. Note that most screens in the
# system get along OK because they inherit from Mode, allowing 
# us to use inheritance to handle the absence of these routines
# by letting Mode's AUTOLOAD routine trap them.
#
# CommStat is not a Mode, and rather than suffer possible 
# trouble from making it one, we'll simply implement stub
# routines here for now, to be replaced by final versions
# later. (And of course, we'll update these comments, too.)
#------------------------------------------------------------

#------------------------------------------------------------
# nav_status
#------------------------------------------------------------
# Returns a very short message summarizing the current status.
# Sets links/greyness/etc. depending on whether or not a search
# is running.
#
# Do *not* set the indicator in this method! That should be
# done in the specific method that starts the search. Otherwise,
# the navigator would indicate "View Search" when we did an order
# or a browse, and that would be wrong.

sub nav_status {

   # Determine status.
   my $status = "";

   # A default "page" needs to shown if no search is running/has ran.
   unless ($SESSION->state->{LastsearchClient}) {
     return ("Message" => "No search yet",
             "Mode"    => "SEARCH",
             "Submode" => "COMMINSTRUCTION",
             "Link"    => "",
             );
    }

   # Tell the Navigator what it wants to know.
   if ($SESSION->state->{LastsearchClient}->running) {
      return ("Message" => "Search is running",
              "Mode"    => "SEARCH",
              "Submode" => "COMMTRACE"
              );
   }
   else {
      return ("Message" => "Search completed",
              "Mode"    => "SEARCH",
              "Submode" => "COMMFINAL"
              );
   }
}

#------------------------------------------------------------
1;
