package IW::CommStat::DIRECTORY_SEARCH;

=head1 IW::CommStat::DIRECTORY_SEARCH;

Track communication status for directory searches.  
Post-processing is reported, as are the number of granules.

An B<individual connection> has I<really> finished okay iff the connection
has reported at least one dataset.

The B<entire operation> has I<really> finished okay iff at least one
datasets was recovered from all the connections.

=cut

use strict;
use vars qw(@ISA);

use IW::App;
use IU::Log;

@ISA = qw(IW::CommStat);

#------------------------------------------------------------
# init
#------------------------------------------------------------
sub init {
    my $self = shift;
    $self->{OpName} = 'search';
    $self->{ReportDatasets} = 1;
    $self->{StopNote} = 
	"Be patient: it may take a half a minute to do a graceful stop.";
    debug "init as a directpry search";
    $self;
}

#------------------------------------------------------------
# finished_ok
#------------------------------------------------------------
# Did the directory search finish okay?  It did if we got so much
# as a single dataset.

sub finished_ok {
    my $self = shift;
    debug "datasets = ", $self->n_datasets;
    debug "granules = ", $self->n_granules;
    $self->n_datasets > 0;
}

#------------------------------------------------------------
# print_directions OPSTATUS
#------------------------------------------------------------
sub print_directions {
    my ($self, $op_status) = @_;

    my $numDatasets = $self->n_datasets;

    # What should they do now?
    print "<hr>\n";
    $HTML -> directions;

    my $num_grans = 0;          # Number of granules post-processed.
    my $num_datasets = 0;       # Number of data sets post-processed.
    if ($self->client->requires_post_processing) {
    # Get the post-processing statuses:
        my @procs = $self->client->proc_statuses;
        my $proc;
        if (defined (@procs))  {
	   foreach $proc (@procs) {
              $num_grans += $proc->n_grans;
              $num_datasets += $proc->n_datasets;
           }
        }
    }
    if ($op_status eq 'CRASHED') {      # can't do anything, because of crash
	print "A system error during your ",
              "search and it crashed.  ";
        if ($num_grans || $num_datasets) {
            print "There were some results retrieved and you may ",
                  "view them by selecting the ";
            $HTML -> A(HREF=>IW::Util::script_url('mode'=>'DSETLIST'));
            print "<B>Results: Data Sets</B>";
            $HTML -> _A;
            print " navigation link. You may want to try a different set of ",
                  "search criteria and search again.\n";
        } else {
            print "No results were retrieved.  You may want to try ",
                  "your search again later, or try modifying your ",
                  "search criteria and search again.";
        }
    }
    elsif ($op_status eq 'KILLED') {    # can't do anything, because of kill
	print "The search had to be terminated, and so no results were ";
	print "retrieved. ";
	print "You may want to try a different set of search criteria.\n";
    }
    elsif ($numDatasets > 0) {          # can do something
       print "A total of <b>$numDatasets</b> data sets satisfied your search.";
	print " Go to the ";
	$HTML -> A(HREF=>IW::Util::script_url('mode'=>'DSETLIST'));
	print "<B>Results: Data Sets</B>";
	$HTML -> _A;
	print " to review your results.\n";
    }
    else {                              # can't do anything
	print "No data sets satisfied your search.\n";
	print "Please read the comments below and try again, adjusting ";
	print "your search criteria as needed.";
    }
    $self->print_generic_directions;
    $HTML -> _directions;
}

#------------------------------------------------------------
# start_page PARAMHASH
#------------------------------------------------------------
# Set up navigator and call App's start_page.

sub start_page {
   my $self = shift;
   my %params = @_;

   # Set the navigation indicator to view search tab.
   $SESSION->{Navigator}->{Indicator} = "ViewSearch";
   $APP->start_page(%params);
}

#------------------------------------------------------------
# contextual_help
#------------------------------------------------------------
# Returns a message indicating what to do now.
# If user is not a guest, offer to sign out.

sub contextual_help {

   if ($ENV{REMOTE_USER} =~ /^guest$/i) {
      return ("You can use the 'Results:Data set' Navigation link to switch 
               to viewing results as they arrive."
             );
   } else {
      return ("You can use the 'Results:Data set' Navigation link to switch 
               to viewing results as they arrive.",
              "You can use the 'Exit' Navigation link to leave your 
               account now and sign in later to view results 
               or the search status."
             );
   }
}

1;

