package IW::CommStat::INVENTORY_SEARCH;

=head1 IW::CommStat::INVENTORY_SEARCH

Track communication status for searches.  Post-processing is
reported, as are the number of granules.

An B<individual connection> has I<really> finished okay iff the connection
has reported at least one granule.  

The B<entire operation> has I<really> finished okay iff at least one
granule was recovered from all the connections.

=cut

use strict;
use vars qw(@ISA);

use IW::App;

@ISA = qw(IW::CommStat);

#------------------------------------------------------------
# init
#------------------------------------------------------------
sub init {
    my $self = shift;
    $self->{OpName} = 'search';
    $self->{ReportDatasets} = 1;
    $self->{ReportGranules} = 1;
    $self->{StopNote} = 
	"Be patient: it may take a half a minute to do a graceful stop.";
    $self;
}

#------------------------------------------------------------
# finished_ok
#------------------------------------------------------------
# Did the inventory search finish okay?  It did if we got so much
# as a single dataset, zero granules is okay.

sub finished_ok {
    my $self = shift;
    # $self->n_granules > 0;
    $self->n_datasets > 0;
}

#------------------------------------------------------------
# print_directions OPSTATUS
#------------------------------------------------------------
sub print_directions {
    my ($self, $op_status) = @_;

    my $num_grans = 0;          # Number of granules post-processed.
    my $num_datasets = 0;       # Number of data sets post-processed.
    if ($self->client->requires_post_processing) {
    # Get the post-processing statuses:
        my @procs = $self->client->proc_statuses;
        my $proc;
        if (@procs)  {
	   foreach $proc (@procs) {
              $num_grans += $proc->n_grans if(defined($proc->n_grans));
              $num_datasets += $proc->n_datasets if(defined($proc->n_datasets));
           }
        }
    }
    # What should they do now?
    print "<hr>\n";
    $HTML -> directions;
    if ($op_status eq 'CRASHED') {      # can't do anything, because of crash
	print "A system error occurred during your ",
              "search and it crashed.  ";
        if ($num_grans || $num_datasets) {
            print "There were some results retrieved and you may ",
                  "view them by selecting the ";
            $HTML -> A(HREF=>IW::Util::script_url('mode'=>'DSETLIST'));
            print "<B>Results: Data Set</B>";
            $HTML -> _A;
            print " navigation link.  You may want to try a different set of ",
                  "search criteria and search again.\n";
        } else {
            print "No results were retrieved.  You may want to try ",
                  "your search again later, or try modifying your ",
                  "search criteria and search again.";
        }
    }
    elsif ($op_status eq 'KILLED') {    # can't do anything, because of kill
	print "The search had to be terminated, so no results were ";
	print "retrieved. ";
	print "You may want to try a different set of search criteria.\n";
    }
    elsif ($num_grans > 0) {          # can do something
	print "A total of <b>$num_grans</b> data granules were returned
        for this search.";
	print " Go to ";
	$HTML -> A(HREF=>IW::Util::script_url('mode'=>'DSETLIST'));
	print "<B>Results: Data Set</B>";
	$HTML -> _A;
	print " to review your results and browse/order the data.\n";
    }
    else {                              # can't do anything
	print "No data granules satisfied your search.\n";
	print "Please read the comments below and try again, adjusting ";
	print "your search criteria as needed.\n";
    }
    $self->print_generic_directions;
    $HTML -> _directions;
}

#------------------------------------------------------------
# start_page PARAMHASH
#------------------------------------------------------------
# Set up navigator and call App's start_page.

sub start_page {
   my $self = shift;
   my %params = @_;

   # Set the navigation indicator to view search tab.
   $SESSION->{Navigator}->{Indicator} = "ViewSearch";
   $APP->start_page(%params);
}

#------------------------------------------------------------
# contextual_help
#------------------------------------------------------------
# Returns a message indicating what to do now.
# If user is not a guest, offer to sign out.

sub contextual_help {

   if ($ENV{REMOTE_USER} =~ /^guest$/i) {
      return ("You can use the 'Results:Data set' Navigation link to switch 
               to viewing results as they arrive."
             );
   } else {
      return ("You can use the 'Results:Data set' Navigation link to switch 
               to viewing results as they arrive.",
              "You can use the 'Exit' Navigation link to exit your 
               account now and sign in later to view results 
               or the search status."
             );
   }
}

1;

