package IW::Conn;

=head1 NAME

IW::Conn - a connection status object for a single DAAC


=head1 SYNOPSIS

    # Read from a chunk stream:
    $conn = input IW::Conn \*STDIN;
    
    # What kind of activity are we engaged in?  See IW::Client for the list.
    # Always returns results ALL UPPERCASE.
    $conn->activity('INVENTORY_SEARCH');       # inventory search
    
    # Get stage info:
    $conn->stage;          # numeric stage
    $conn->stage_cmnt;     # printable string
   
    # Get status info:
    $conn->status;         # numeric stage
    $conn->status_cmnt;    # printable string
    
    # Other info:
    $conn->daac;           # data center id
    $conn->n_grans;        # number of granules  
    $conn->n_gran_hits;    # number of granules actually hit
    $conn->n_datasets;     # number of datasets
    
    # Output:
    $conn->print_pot;
     
    # Still running?  
    if ($conn->still_running) {
         # still running
    } 
    elsif ($conn->finished_ok) {
         # not still running, and finished ok
    }
    else {
         # not still running, and didn't finish okay
    }        


=head1 DESCRIPTION

The following are the things we need to deal with, listed from 
highest to lowest level.  There are two things we need to
deal with: B<stage> and B<status>:

=over

=item Stage

This is the high-level stage of processing that is indicated by the
LEDs.  They are:

   0  Pending
   1  Connection open          B 
   2  Request sent             B B
   3  Receiving results        B B B
   4  Success                  B B B G
   5  Aborted                  B B B R
   6  DAAC is down             B B B R
   7  Failed                   B B B R
   8  Failure inferred         B B B R   

=item Status

High-level operation status, as given by the C<STATUS_CODE> and
C<STATUS_CODE_COMMENT> ODL.  This is of interest to the end-user,
so we will report it to them if it is given.  Most importantly,
it contains information as to why a request failed.  For example...

   1  Query success - results returned 
   2  Unsuccessful query
   3  Unsuccessful query; data for source not found 
   4  Unsuccessful query; data for sensor not found
   ...

Comes from the "imsclient" program, which obtains it by calling 
C<IK_Nmstatus> and reporting both the 
C<STATUS_CODE> and C<STATUS_CODE_COMMENT>.

=back

=head1 INSTANCE VARIABLES

Most names are chosen to parallel the data dictionary entries,
and are identical to the values given in the imsclient's 
status file.  They are:
   
    DATA_CENTER_ID           the data center being contacted (e.g., "ORNL")
    MESSAGE_ID               the message id (e.g. "M814136141")
    LED_STAGE                processing stage (e.g. "4")
    STATUS_CODE              high-level operation status (e.g., "1")
    STATUS_CODE_COMMENT      high-level message (e.g., "successful query")
    NUMBER_OF_GRANULES       number of granules returned so far (e.g., 0)
    NUMBER_OF_DATASETS       number of datasets returned so far (e.g., 0)

Additional ones are:

    Activity                 the activity we monitor (INVENTORY_SEARCH, etc.)

=head1 RCSID

$Id: Conn.pm,v 1.24 2002/09/17 18:29:26 joe Exp $

=cut

# HACK: explicit import from App. 
BEGIN {
  *IW::Conn::HTML = \$IW::App::HTML;
}

use vars qw($HTML);
use IU;
use IU::Log;
use IU::Chunk ();

use IW::Util qw(:url);
use IW::App qw($HTML);
use IW::StatusBar;

@ISA = qw(IW::StatusBar);

use strict;


#------------------------------------------------------------
#
# Globals:
#
#------------------------------------------------------------

# Stages of processing, 0-7:
my @Stages = (
	   'Pending',		# 0
	   'Connection open',	# 1
	   'Request sent',	# 2
	   'Receiving results',	# 3
	   'Success',           # 4
	   'Aborted',         	# 5
	   'DAAC is down',	# 6
	   'Failed',            # 7
           'Apparent failure',  # 8
);

#------------------------------------------------------------
# from_chunk CHUNK
#------------------------------------------------------------
# I<Constructor.  Concrete override.> 
# Init from a chunk.

sub from_chunk {
    my ($self, $chunk) = @_;
    my $key;

    # Init:
    foreach $key ($chunk->keys) {
	$key =~ tr/a-z-/A-Z_/;       # convert Like-This to LIKE_THIS
	$self->{$key} = $chunk->value($key);
    }

    # Validate and adjust:
    defined($self->{DATA_CENTER_ID}) or puke 600, "illegal connection object";
    defined($self->{LED_STAGE}) || ($self->{LED_STAGE} = 0);
    
    $self;
}

#------------------------------------------------------------
# null
#------------------------------------------------------------
# I<Constructor:> no info is available.

sub null {
    my $self = shift;

    # Init:
    $self->{DATA_CENTER_ID}  = '';
    $self->{MESSAGE_ID}  = '';
    $self->{LED_STAGE}   = 0;
    $self->{LED_STAGE_COMMENT} = 'NO INFO AVAILABLE YET';
    $self->{STATUS_CODE} = 0;
    $self->{NUMBER_OF_GRANULES}     = 0;
    $self->{NUMBER_OF_GRANULE_HITS} = 0;
    $self->{NUMBER_OF_DATASETS}     = 0;
    $self;
}


#------------------------------------------------------------
#
# Various access methods
#
#------------------------------------------------------------

sub daac        { $_[0]->{DATA_CENTER_ID}}        # deprecated
sub msg_id      { $_[0]->{MESSAGE_ID}}
sub n_grans     { $_[0]->{NUMBER_OF_GRANULES}     || 0}
sub n_gran_hits { $_[0]->{NUMBER_OF_GRANULE_HITS} || 0}
sub n_datasets  { $_[0]->{NUMBER_OF_DATASETS}     || 0}
sub status      { $_[0]->{STATUS_CODE} }
sub status_cmnt { 
    my $stat = $_[0]->{STATUS_CODE_COMMENT};
    "\u$stat";
}

#------------------------------------------------------------
# _fixup
#------------------------------------------------------------
# Fixup status_code_comment to inform the user what has happened.
#

sub _fixup {
   my $self = shift; 

   # No need to do it twice:
   return if $self->{FixedUp}; $self->{FixedUp} = 1;
    
   # Fixups if not running...
   if (!$self->still_running) {     # no longer running:

       # fixup for inventory searches:
       if ($self->activity eq 'INVENTORY_SEARCH') {

          # Got some data:
          if ($self->n_grans > 0) {

             # Change "daac is down" (6) to "Terminated" (7)
             if ($self->stage == 6) {

                 $self->{LED_STAGE} = 7;
             }


             # There was an error reported by imsclient (LED_STAGE > 4, such  
             # as a dropped connection), but we got some data.  Fixup the 
             # status_code_comment to let the user know something went wrong.
             if ($self->stage > 4) {

                $self->{STATUS_CODE_COMMENT} =
                   "Received some granules, but the search stopped ".
                   "unexpectedly. There may be more matching granules ".
                   "than those returned.";
             }
 
          # Got NO data: 
          } else {
    
             # The daac is reporting a success (status_code == 1) but we
             # did not get any data. Report the apparent failure to the user. 
             if ($self->status == 1) {

                $self->{LED_STAGE} = 8;
                $self->{STATUS_CODE_COMMENT} = "No data granules returned."; 
             }

          } # end else no data.

       } # end if inventory search.


       # fixup for directory searches:
       if ($self->activity eq 'DIRECTORY_SEARCH') {

          # Got some data:
          if ($self->n_datasets > 0) {

             # There was an error reported by imsclient (LED_STAGE > 4, such
             # as a dropped connection), but we got some data.  
             # Hide error from user as we don't care for a directory search.
             if ($self->stage > 4) {

                $self->{LED_STAGE} = 4;
             }

          # Got NO data:
          } else {

             # The daac is reporting a success (status_code == 1) but we
             # did not get any data. Report the apparent failure to the user.
             if ($self->status == 1) {

                $self->{LED_STAGE} = 8;
                $self->{STATUS_CODE_COMMENT} = "No data sets returned.";
             }

          } # end else no data.

       } # end if directoy search.

   } # end if running.
   1;
}

#------------------------------------------------------------
# activity 
#------------------------------------------------------------
# What kind of activity are we involved in?  
# Returns 'INVENTORY_SEARCH', 'DIRECTORY_SEARCH', etc., or empty.
#
# NOTE: this affects how finished_ok() is computed!

sub activity {
    my ($self, $act) = @_;
    if (defined($act)) {                 # setting value...
	$self->{Activity} = $act;
	$self->_fixup;
    }
    uc($self->{Activity} || '');
}

#------------------------------------------------------------
# finished_ok
#------------------------------------------------------------
# I<Concrete override.>
# The operation on this DAAC is I<not> running... but did it finish okay?
#
# If this was an inventory search, then a positive number of granules
# forces an "ok".  (KLUDGE).
#
# If this was an directory search, then a positive number of datasets
# forces an "ok".  (KLUDGE).
#
# Otherwise, we must have a successful status code.

sub finished_ok {
    my $self = shift;

    if ($self->activity eq 'INVENTORY_SEARCH') {        # inventory search...
	debug "IS and ngrans = ", $self->n_grans;
	return ($self->n_datasets > 0);
    }
    elsif ($self->activity eq 'DIRECTORY_SEARCH') {     # directory search...
	debug "DS and ndsets = ", $self->n_datasets;
	return ($self->n_datasets > 0);
    }
    else {                               # other: just go by the status code...
	return ($self->status == 1); 
    }
}

#------------------------------------------------------------
# stage
#------------------------------------------------------------
# Return the current stage.

sub stage  { 
    my $self = shift;
    $self->{LED_STAGE};
}

#------------------------------------------------------------
# stage_cmnt
#------------------------------------------------------------
# Return a SHORT comment for this stage.

sub stage_cmnt  { 
    my $self = shift;
    return $self->{LED_STAGE_COMMENT} if $self->{LED_STAGE_COMMENT};
    ($Stages[$self->stage] || 'Unknown');
}

#------------------------------------------------------------
# print_pot 
#------------------------------------------------------------
# I<Concrete override.>
# Print the "LED display" showing our status.

sub print_pot {
    my $self = shift;

    # Stage 1: connection open:
    if ($self->stage < 1) {

       $HTML ->t('Connecting');
    } 
    # Stage 2: sent query:
    elsif ($self->stage == 2) {

       $HTML ->t('Searching');
    }
    # Stage 3: receiving results:
    elsif ($self->stage == 3) {

       $HTML ->t('Receiving Results');
    }
    # Stage 7: aborted search: 
    elsif ($self->stage == 7) {

       $HTML ->t('Terminated');
    }
    # Stage 8: apparent failure:
    elsif  ($self->stage == 8) {

       $HTML->t('Failed');
    }
    # Stages 4,5,& 6: Everything else
    else {

       $HTML ->t('Completed');
    }
    1;
}

#------------------------------------------------------------
# still_running
#------------------------------------------------------------
# I<Concrete override.>
# Is the request on this DAAC still running?

sub still_running {
    my $self = shift;
    ($self->stage < 4);
}

#------------------------------------------------------------
1;

