package IW::DatasetList;

=head1 NAME

IW::DatasetList - utilities for outputting a dataset list


=head1 DESCRIPTION

Several modes currently output a dataset list.  
Here's the common code.

See IW::ObjectList for details.


=head1 HIDDEN FIELDS

Hidden fields are embedded in the output (which of course must be
enclosed in a form, since it contains checkboxes!). 
These fields actually originate from the last request; they are:

     dskey     One for each dskey in the last request
     pkkey     One for each pkkey in the last request
     gkey!     Just generated once: indicates presence of dataset checkboxes

In addition, the checkbox state is sensitive to the following param:

     select    If set, then checkbox state is determined by whether the
               value is "ALL" or "NONE".


=head1 SORTING

A request to sort the datasets uses the most "appropriate" dataset sort.
As a side effect, it sets the LastDatasetSort in the session state.

=cut

use IU;
use IU::Log;
use IU::Dataset;
use IU::WWWValids;

use IW::Date qw(TZ_to_date TZ_to_datetime);
use IW::ObjectList;
use IW::Util qw(:url);
use IW::App;
use IW::Sort;

@ISA = qw(IW::ObjectList);

use strict;

#------------------------------------------------------------
#
# GLOBALS
#
#------------------------------------------------------------

# Order to present things:
my @Order;

#------------------------------------------------------------
# new PARAMHASH
#------------------------------------------------------------
# Create a new dataset-list object.
# PARAMHASH contains everything normal to the superclass.

sub new {
    my $class = shift;
    my %params = @_;

    # Construct:
    my $self = $class->SUPER::new(@_);
    $self->has_summaries(1);
    $self->{SearchType} = $SESSION->state->{LastSearchType};

    my @Summary = @{$self->{Display}->dataset_summary};
    # hard code Data set ID as first column.
    @Order = ('DATASET_ID');

    # If search results are directory, switch off columns that apply
    # to inventory results only.
    if ($self->{SearchType} eq 'DIRECTORY_SEARCH') {
       my $title;
       my %inv_only = (NUM_GRANULES => '1',
                       START_DATE => '1',
                       STOP_DATE =>'1',
                       CAMPAIGN => '1',
                       PARAMETER => '1',
                       PROCESSING_LEVEL => '1',
                       SENSOR_NAMES => '1',
                       SOURCE_NAMES => '1'
                      );
       foreach $title (@Summary) {
          push @Order, $title unless ($inv_only{$title});
       }
    } else {
       push @Order,@Summary;
    }

    $self->{NoteFunc} or $self->{NoteFunc} = \&note;
    $self->{Note2Func} or $self->{Note2Func} = \&note2;

    $self;
}

#------------------------------------------------------------
# origin
#------------------------------------------------------------
# Get/set the origin of these datasets.

sub origin {
    my ($self, $value) = @_;
    $value and $self->{Origin} = $value;
    $self->{Origin};
}

#------------------------------------------------------------
# sort [SORTREF]
#------------------------------------------------------------
# Sort our list of datasets.  Returns a reference to the newly-sorted array.
#
# B<SIDE EFFECTS:> This sets the session-state variable 'LastDatasetSort' 
# to the sort option used, so that it is used/presented as a default for 
# subsequent dataset lists.

sub sort {
    my ($self, $sortref) = @_;

    # Figure out the sort algorithm...
    debug "figuring out the sorting algorithm";
    my $sorter = IW::Sort::Dataset->appropriate;
    
    # Construct a nice string to describe this:
    $$sortref = $sorter->as_string      if $sortref;
    
    # Do it!    
    my @sorted = $sorter->sort($self->{Elems});
    $self->{Elems} = \@sorted;

    # Make a note in the session state:
    $SESSION->state->{LastDatasetSort} = $sorter;
    $self->{Elems};
}

#------------------------------------------------------------
#
# PRIVATE METHODS/UTILITIES
#
#------------------------------------------------------------


#------------------------------------------------------------
# sorted_list
#------------------------------------------------------------
# Private function.  Take a ref to an array of strings, and
# return a single scalar consisting of the elements sorted and joined
# by commas.

sub sorted_list {
    my $unsorted = shift;
    join ', ', (sort @$unsorted);
}

#------------------------------------------------------------
# check_for_some_txt
#------------------------------------------------------------
# returns HASH

sub check_for_some_txt {

   my $self = shift;

   # if user wants to see empty columns then start with an initial value
   # of '1' (true) for the value of the %some hash.
   my $show_empty = ($self->{Display}->show_empty_columns_dset eq 'Y')
       ? '1' : '0';
   my %some = map {$_=>$show_empty} @Order;

   # Determine the width to make each column.
   # Loop over every row (Elem), updating values in %some.

   my $num = int(@{$self->{Elems}});
   my $i=0;
   for ($i=0; $i < $num; $i++ ) {
      $self->print_data_txt(\%some,$i,'1');
   }

   # Update column width using column headers as they maybe wider than
   # the data.
   $self->print_head_txt(\%some,'1');

   return %some;

}

#------------------------------------------------------------
# check_for_some HASH
#------------------------------------------------------------

sub check_for_some {

   my $self = shift;

   # if user wants to see empty columns then start with an initial value
   # of '1' (true) for the value of the %some hash.
   my $show_empty = ($self->{Display}->show_empty_columns_dset eq 'Y')
       ? '1' : '0';
   my %some = map {$_=>$show_empty} @Order;

   # Hard code these as aways being displayed
   $some{'DATASET_ID'} = '1';
   $some{'DATA_CENTER_ID'} = '1';
   $some{'CAMPAIGN'} = '1';
   $some{'PARAMETER'} = '1';
   $some{'PROCESSING_LEVEL'} = '1';
   $some{'SENSOR_NAMES'} = '1';
   $some{'SOURCE_NAMES'} = '1';

   my $title;
   my $item;
   foreach $item (@{$self->{Elems}}) {
      # As data is found for a column it is removed from the loop,
      # gradually the loop becomes shorter and shorter.
      # If the first item has values for all the columns
      # then this will run only once.
      foreach $title (grep {!$some{$_}} (keys %some) ) {
         if ($title eq 'START_DATE') {

            $some{$title} = '1' if (defined ($item->start_date)
                                    and $item->start_date ne '(none)');
         }
         elsif ($title eq 'STOP_DATE') {

            $some{$title} = '1' if (defined ($item->stop_date)
                                    and $item->stop_date ne '(none)');
         }
         elsif ($title eq 'MISC_URL') {
            $some{$title} = '1' if (defined ($item->misc_URLs));
         }
         elsif ($title eq 'NUM_GRANULES') {

            $some{$title} = '1' if (defined($item->n_granules));

         } # end elsif
      }
   }
   return %some;
}
#------------------------------------------------------------
# print_data_txt
#------------------------------------------------------------
# print data for each row for the text only report.

sub print_data_txt {
   my $self = shift;
   my $some = shift;
   my $i = shift;
   my $set_width_only = shift || 0;

   my $width = 1;
   my $title;
   my $item = $self->{Elems}[$i];


   my $field = '';
   # Only do those columns which have some>0;
   my @NewOrder = ($set_width_only)? @Order :  (grep {$some->{$_}} (@Order));
   foreach $title (@NewOrder) {

      $field = '';

      if ($title eq 'DATASET_ID') {
         $field = $item->id;
      } 
      elsif ($title eq 'DATA_CENTER_ID') {

         $field = $item->archive_id;
      }
      elsif ($title eq 'NUM_GRANULES') {
         my $ngot  = defined($item->n_granules) ? $item->n_granules : '';
         my $nhits = $item->n_granule_hits;
         my $note = '';
         if ($nhits) {
            $note = (($nhits == ($ngot+1))
                     ? " (more are known)"
                     : " (of $nhits)");
         }
         $field =  $ngot.$note;
      }
      elsif ($title eq 'CAMPAIGN') {

         $field = sorted_list($item->campaigns);
      }
      elsif ($title eq 'PARAMETER') {

         $field = sorted_list($item->parameters);
      }
      elsif ($title eq 'PROCESSING_LEVEL') {

         $field = sorted_list($item->proc_levels);
      }
      elsif ($title eq 'SENSOR_NAMES') {

         $field = sorted_list($item->sensors);
      }
      elsif ($title eq 'SOURCE_NAMES') {

         $field = sorted_list($item->sources);
      }
      elsif ($title eq 'START_DATE') {

         $field = TZ_to_datetime($item->start_date);
      }
      elsif ($title eq 'MISC_URL') {

         my $num_misc_urls = ($item->misc_URLs ? int(@{$item->misc_URLs}) : 0);
         my ($i, $furl, @list);
         for ($i=0; $i<$num_misc_urls; $i++) {
            $furl = $item->misc_URLs->[$i];
            push (@list, $furl->URL);
         }
         $field = join (' , ', @list);
      }
      elsif ($title eq 'STOP_DATE') {

         $field = TZ_to_datetime($item->stop_date);
      } # end elsif

      if ($set_width_only) {
         $some->{$title} = length($field) if (length($field) > $some->{$title});
      } else {
         $width = $some->{$title};
         printf "%-${width}s",$field;
         print '|';
      }
   } # end foreach $title
}


#------------------------------------------------------------
# print_data
#------------------------------------------------------------

sub print_data {
   my $self = shift;
   my $some = shift;
   my $i = shift;

   my $title;
   my $item = $self->{Elems}[$i];


   # Only do those columns which have some=1;
   foreach $title (grep {$some->{$_}} (@Order) ) {

      if ($title eq 'DATASET_ID') {

         my $id = $item->id || "\xA0";
         print '<td>';
         # Are there any homepage URLs for this dataset?
         my $furl;
         if ($furl = $item->home_URLs->[0]) {    # We have hyperlinks!
            my $url = logged_url($furl->URL, 'type'=>$furl->type);
            $HTML -> A(HREF=>$url) -> t_B($id) -> _A;
         }
         else {                                  # No hyperlinks...
            $HTML -> t_B($id);
         }
         print '</td>';
      }
      elsif ($title eq 'DATA_CENTER_ID') {

         print '<td>', $item->archive_id || "\xA0", '</td>';
      }
      elsif ($title eq 'NUM_GRANULES') {
         my $ngot  = defined($item->n_granules) ? $item->n_granules : '';
         my $nhits = $item->n_granule_hits;
         my $note = '';
         if ($nhits) {
            $ngot .= (($nhits == ($ngot+1))
                     ? " (more are known)"
                     : " (of $nhits)");
         }
         
        print '<td>', $ngot || "\xA0", '</td>';
      }
      elsif ($title eq 'CAMPAIGN') {

         print '<td>', sorted_list($item->campaigns) || "\xA0", '</td>';
      }
      elsif ($title eq 'PARAMETER') {

         print '<td>', sorted_list($item->parameters) || "\xA0", '</td>';
      }
      elsif ($title eq 'PROCESSING_LEVEL') {

         print '<td>', sorted_list($item->proc_levels) || "\xA0", '</td>';
      }
      elsif ($title eq 'SENSOR_NAMES') {

         print '<td>', sorted_list($item->sensors) || "\xA0", '</td>';
      }
      elsif ($title eq 'SOURCE_NAMES') {

         print '<td>', sorted_list($item->sources) || "\xA0", '</td>';
      }
      elsif ($title eq 'MISC_URL') {
         my $num_misc_urls = ($item->misc_URLs ? int(@{$item->misc_URLs}) : 0);
         if ($num_misc_urls == 0) {
            print '<td>', "\xA0",'</td>';
         } else {
            my ($i, $furl, $url, $comment);
            print '<td>';
            for ($i=0; $i<$num_misc_urls; $i++) {
               $furl = $item->misc_URLs->[$i];
               $url = logged_url($furl->URL, 'type'=>$furl->type);
               $comment = $furl->comment;
               print "<a href=$url target=ext-url-window>$comment</a>", '<br>';
            }
            print '</td>';
         }
      }
      elsif ($title eq 'START_DATE') {

         print '<td>', TZ_to_datetime($item->start_date) || "\xA0", '</td>';
      }
      elsif ($title eq 'STOP_DATE') {

         print '<td>', TZ_to_datetime($item->stop_date) || "\xA0", '</td>';

      } # end elsif

   } # end foreach $title
}

#------------------------------------------------------------
#
# PRINT...
#
#------------------------------------------------------------

#------------------------------------------------------------
# _xxx_url DSET  -- URLs for menu buttons
#------------------------------------------------------------

# Return the URL for the Attributes button.
#
# Currently, we force this to undef (i.e., we grey out the button)
# if we are displaying the results of a directory search.  
# KLUDGE! REMOVE THIS WHEN WE GET OTHER STUFF BACK FROM DIRECTORY SEARCHES!

sub _attrs_url {
    my ($self, $dset) = @_;
    return undef if ($SESSION->state->{LastSearch} eq 'DIRECTORY_SEARCH');  
    script_url('mode'=>'DSETATTRS', 
               'origin' => $self->origin,
	       'dskey'=>$dset->key);
}

# Return the URL for the Brief Info button:
sub _binfo_url {
    my ($self, $dset) = @_;
#some data centers are sending back Dummy for mdentry which causes an
#endless search at GCMD
   if($dset->MD_entry eq "Dummy") {
      dif_url(undef);
   }else {
    dif_url($dset->MD_entry);
   }
}

# Return the URL for the Detailed Info button:
sub _dinfo_url {
    my ($self, $dset) = @_;
    guide_url('DATASET_ID', $dset->id);
}

# Return the URL for the Granule List button:
sub _glist_url {
    my ($self, $dset) = @_;
    ($dset->n_granules <= 0) ? undef :
	script_url('mode'=>'GRANLIST', 
                   'submode' => 'NEW',
                   'dskey'=>$dset->key);
}

# Return the URL for the Links button:
sub _links_url {
    my ($self, $dset) = @_;
    $dset->has_links or return undef;
    script_url('mode'=>'DSETATTRS', 
	       'dskey'=>$dset->key) . '#LINKS';
}

# Return the URL for the Choose Package button:
sub _pkg_url {
    my ($self, $dset) = @_;
    if (($self->origin eq 'CART') && (@{$dset->package_ids})) {
        return script_url('mode'=>'PKGSLCT',
                          'submode' => 'DATASET',
                          'dskey'=>$dset->key,
                          'origin'=>($self->origin || 'RESULTS'));
    }
    undef;
}

#------------------------------------------------------------
# _checkbox DSET
#------------------------------------------------------------
sub _checkbox {
    my $dset = shift;

    # The value for the selection checkbox:
    my @checkval = ();
    if ($REQ->param('select')) {       # explicit value...
	@checkval = (Value => ($REQ->param('select') eq 'ALL'));
    }
    $HTML -> checkbox(Name=>"dskey", Choice=>$dset->key, @checkval);
}

#------------------------------------------------------------
# note2 DATASET
#------------------------------------------------------------

sub note2 {
    1;
}

#------------------------------------------------------------
# note DATASET
#------------------------------------------------------------

sub note {
    my $dset = shift;

    if ($dset->carted) {
        $HTML -> IMG(SRC=>doc_url('icons/carted.gif'),
                     ALT=>'In Shopping Cart');
    } else {
       print "\xA0";
    } 
}

#------------------------------------------------------------
# print_buttons I
#------------------------------------------------------------
# Output the menu buttons for the i'th (0-based) data set.

sub print_buttons {
   # Some basic info:
   my ($self, $ds_number, $some_notes, $some_pkgs_chosen, 
       $some_orderable, $some_comments) = @_;

   # Get the individual element.
   my $item = $self->{Elems}[$ds_number];

   my $button_prefix =  'item';
   my $prim_prefix = 'item';

   # "Selection" checkbox... notice, we check for "select all":
   print '<td align=center>';
   _checkbox($item);
   print '</td>';
   # If this data set orderable in whole, put button in table.
   if ($item->orderable) {
      # Add selected dataset to cart:
      my @dskey_array = ('dsaddkey', $item->key);
      print '<td align=center>';
      $HTML -> cartdsetadd_menu_btn(\@dskey_array);
      print '</td>';
   }
   # else if some data sets in table are orderable, put space.
   elsif ($some_orderable) {
      print "<td>\xA0</td>";
   }
   print '<td>';

   # "Links" button...
    $HTML -> hypertext("Other links",
                   $self->_links_url($item)) if ($self->_links_url($item));
    $HTML -> t("Other links") if (!$self->_links_url($item));
   print '</td><td>';

   # "Summary Document" button...
    $HTML -> hypertext("Summary document",
                   $self->_binfo_url($item),
                   TARGET=>'support') if ($self->_binfo_url($item));
    $HTML -> t("Summary document") if (!$self->_binfo_url($item)); 
   print '</td><td>';

   # "Detailed Document" button...
    $HTML -> hypertext("Detailed document",
                   $self->_dinfo_url($item)) if ($self->_dinfo_url($item));
    $HTML -> t("Detailed document") if (!$self->_dinfo_url($item));
   print '</td>';

    # Print any cart note if in cart:
    if ($some_notes) {
	print '<td>';
	$self->print_note($ds_number);
	print '</td>';
    }

    # Print comment link if there are any dataset comments
    if ($some_comments) {
       print '<td>';
       $self->print_comment($item);
       print '</td>';
    }
    1;
}

#------------------------------------------------------------
# print_head_txt
#------------------------------------------------------------
# Output the column headers for the 1 line per item table.

sub print_head_txt {
   my $self = shift;
   my $some = shift;
   my $set_width_only = shift || 0;

   $WWWVALIDS = IU::WWWValids->new() if (!defined($WWWVALIDS));
   my %pretty = $WWWVALIDS->dsetListPretty;
   $pretty{'DATASET_ID'} = 'Data Set';

   my $width = 0;
   my $title;
   foreach $title (@Order) { # order of presentation!
      if ($some->{$title} ) {
       # Just set the column width
       if ($set_width_only) {
         $some->{$title} = length($pretty{$title})
            if (length($pretty{$title}) > $some->{$title});
       }
       # Print column headers
       else {
         $width = $some->{$title};
         printf "%-${width}s",$pretty{$title};
         print '|';
       }
      }
   }
   print "\n" unless $set_width_only;
   foreach $title (@Order) {
      $width = $some->{$title};
      if ($width && !$set_width_only) {
         my $i;
         for ($i=0; $i < $width; $i++) {
            print '-';
         }
         print '|';
      }
   }
}

#------------------------------------------------------------
# print_head
#------------------------------------------------------------
# Output the column headers for the 1 line per item table.

sub print_head {
   my ($self, $some_notes, $some_pkgs_chosen, $some_orderable, 
       $some_comments, $some) = @_;

   print '<tr bgcolor="#CCCCCC">';

   if ($self->{SearchType} ne 'DIRECTORY_SEARCH') {
      print '<td><b>Select</b>';
      $HTML ->BR;
      $self->print_all_none;
      print '</td>';
      if ($some_orderable) {
         print '<td><b>Place Entire Data Set in Cart</b></td>';
      }
      print '<td colspan=3><b>Options...</b></td>';

   } else { # Inventory results

      print '<td><b>Select</b><br>';
      $self->print_all_none;
      print '</td>';
      print '<td><b>Place Entire Data Set in Cart</b></td>'
         if ($some_orderable);
      print '<td colspan=3><b>Menu of Other Actions</b></td>';
   }
   print '<td><b>Notes</b></td>' if ($some_notes);
   print '<td><b>  Comments...  </b></td>' if ($some_comments);

   my $title;
   $WWWVALIDS = IU::WWWValids->new() if (!defined($WWWVALIDS));
   my %pretty = $WWWVALIDS->dsetListPretty;

   # hard code Data set ID.
   $pretty{'DATASET_ID'} = 'Data Set';

   foreach $title (@Order) {                 # order of presentation!
      print '<td nowrap><b>', $pretty{$title}, '</b></td>'
         if ($some->{$title});
   }
   print "</tr>\n";
   1;
}

#------------------------------------------------------------
# print_comment
#------------------------------------------------------------

sub print_comment {

   my $self = shift;
   my $item = shift;
  
   my $comment = $item->status_code_comment;
   my $string = substr($comment,0,17).($comment ? "..." : "");
   my $url = $self->_attrs_url($item).'#SCROLL';
   if ($string) {
      $HTML -> small 
            -> A(HREF=>$url) 
            -> t($string) 
            -> _A
            -> _small;
   } else {
       print "\xA0";
   }
   1;
}


#------------------------------------------------------------
# any_comments
#------------------------------------------------------------
# return trun if there are any data set comments
# hardcoded true for now.

sub any_comments {

    my $self = shift;
    my $num = shift;
    my $ds_number;
    my $dset;

    for ($ds_number = 0; $ds_number < $num; $ds_number++) {
        # Get the individual element.
        $dset = $self->{Elems}[$ds_number];
        # If this data set has a comment, return true.
        if ($dset->status_code_comment) {
            return (1);
        }
    }
    return (0);

}

#------------------------------------------------------------
# any_notes
#------------------------------------------------------------
# Return true if there are any notes in the whole table of all data sets.
# Return false if there are no notes in the whole table.

sub any_notes {
    my $self = shift;
    my $num = shift;

    if (($self->{Origin}) eq 'CART') {
	return (0);
    }

    my $ds_number;
    my $dset;

    for ($ds_number = 0; $ds_number < $num; $ds_number++) {
	# Get the individual element.
	$dset = $self->{Elems}[$ds_number];
	# If data set is in cart, return true.
	if ($dset->carted) {
	    return (1);
	}
    }
    return (0);
}

#------------------------------------------------------------
# any_orderable
#------------------------------------------------------------
# Return true if any data sets are "orderable" in the whole table.
# Return false otherwise.

sub any_orderable {
    my $self = shift;
    my $num = shift;
    my $ds_number;
    my $dset;

    for ($ds_number = 0; $ds_number < $num; $ds_number++) {
	# Get the individual element.
	$dset = $self->{Elems}[$ds_number];
	# If this data set is orderable, return true.
	if ($dset->orderable) {
	    return (1);
	}
    }
    return (0);
}

#------------------------------------------------------------
1;
