package IW::Date;

=head1 NAME

IW::Date -- date/time objects

=head1 AUTHOR

Eryq, Hughes STX Corporation

=head1 INSTANCE VARIABLES

    Year => 4 digit, 0 if never set
    Mon  => 1-12, 1 if never set
    Day  => 1-31, 1 if never set
    Hour => 0-23, 0 if never set
    Min  => 0-59, 0 if never set
    Sec  => 0-59, 0 if never set

=cut

use IU;
use IU::Log;
use IU::Session;
use IW::App;
require Exporter;

@ISA = qw(Exporter);
@EXPORT_OK = qw(TZ_to_date TZ_to_datetime);

use strict;

#------------------------------------------------------------
#
# Globals
#
#------------------------------------------------------------

# 3-letter abbreviations for months of the year:
my @GMon = ('', 
	    'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
	    'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec');

# Quick mapping of 3-letter month abbreviations to month number:
my %GMonNum = (JAN=>1, FEB=>2, MAR=>3, APR=>4,  MAY=>5,  JUN=>6,
	       JUL=>7, AUG=>8, SEP=>9, OCT=>10, NOV=>11, DEC=>12);



#------------------------------------------------------------
# new PARAMHASH
#------------------------------------------------------------
# I<Constructor.>

sub new {    
    my $type = shift;
    my %params = @_;
    my $self = {};
    $self->{Year} = (defined($params{Year}) ? $params{Year} : 0);
    $self->{Mon} =  (defined($params{Mon})  ? $params{Mon}  : 1);
    $self->{Day} =  (defined($params{Day})  ? $params{Day}  : 1);
    $self->{Hour} = (defined($params{Hour}) ? $params{Hour} : 0);
    $self->{Min} =  (defined($params{Min})  ? $params{Min}  : 0);
    $self->{Sec} =  (defined($params{Sec})  ? $params{Sec}  : 0);
    bless $self, $type;
}

#------------------------------------------------------------
# parse_two DATESTR,TIMESTR,[OPT]
#------------------------------------------------------------
# I<Constructor.>  Assuming this object was created with form input, 
# this routine takes that form input and parses it to give us the
# year, month, day, etc. instance vars.  Fatal error if unparseable.
#
# Returns:
#   'BLANK' (true) if original inputs were both blank.
#    The object itself (true) on success.
#    The undef value (false) on error.

sub parse_two {
    my ($self, $datestr, $timestr, $opt) = @_;
               # HIGH opt means coerce to high time, LOW to low time

    # Create object if needed:
    ref($self) or $self = $self->new;

    # Strip leading/trailing blank space:
    $datestr =~ s/^\s+|\s+$//g;
    $timestr =~ s/^\s+|\s+$//g; 
    
    # Extract as much info as possible:
    if ($datestr ne '') {        # date given...

	# Get date:
	$self->parse_date($datestr) 
           or puke 511, "Bad date, '$datestr', (year must be 4 digits, month 2 digits and day 2 digits).";


	# Get any given time, or just use default:
	$timestr = ($timestr ne '') ? $timestr :
	    (($opt eq 'HIGH') ? '23:59:59' : '00:00:00');
	$self->parse_time($timestr) 
           or puke 511, "Bad time, '$timestr', (hours, minutes and seconds must each be 2 digits)";
    }
    else {                       # no date given...
	if ($timestr ne '') {            # ...but time given:
	    puke 516, "You gave a time ($timestr), but no date!";
	}
	else {                           # both blank:
	    return 'BLANK';
	}
    }

    # Still here?  Then we got something!
    return $self;
}

#------------------------------------------------------------
# parse_date DATESTR
#------------------------------------------------------------
# Set elements of the datetime object based on the input string.  The 
# string must contain only the date, with possible leading/trailing whitespace.
# Returns self object, or undef on error.

sub parse_date {    
    my $self = shift;
    my $str = shift; 
    defined($str) || ($str = '');
    my ($day, $month, $year);


    # DD-Mon-YYYY or DD-Monthname-YYYY:
    if ($str =~ m|^\s*(\d{1,2})-([A-Za-z]{3,})-(\d{4})\s*$|) {
	my $mon = $2;
	$mon =~ tr/a-z/A-Z/;
        my $year = $3;
        puke 511, "Please use 4 digits for year $year"  if (length($year)!=4);
	$self->{Year} = $year;
        $month = $GMonNum{substr($mon, 0, 3)} ;
	if ( ($month < 1) || ( $month > 12) ) {
           puke 511, "The Month $month  is invalid. Please select a number between 1 - 12 "; }
	$self->{Mon} = $month;
	$self->{Day} = $self->test_day($1, $month  , $3);  
	return $self;
    }  
    # MM/DD/YYYY: 
    elsif ($str =~ m|^\s*(\d{1,2})/(\d{1,2})/(\d{4})\s*$|) {
        $month = $1 ;
        if ( ($month < 1) || ( $month > 12) ) {
           puke 511, "The Month $month  is invalid. Please select a number between 1 - 12 "; }
	$day = $self->test_day($2, $month, $3);
	($self->{Mon}, $self->{Day}, $self->{Year}) = ($month, $day, $3);
	return $self;
    }  
    # YYYY-MM-DD:
    elsif ($str =~ m|^\s*(\d{3,4})-(\d{1,2})-(\d{1,2})\s*$|) {
        $month = $2 ;
        if ( ($month < 1) || ( $month > 12) ) {
           puke 511, "The Month $month  is invalid. Please select a number between 1 - 12 "; }
	$day = $self->test_day($3,$month,$1);
	($self->{Year}, $self->{Mon}, $self->{Day}) = ($1, $month, $day);
	return $self;
    }  
    else {
	return undef;
    }
}

#------------------------------------------------------------
# parse_two_julian DATESTR,TIMESTR,[OPT]
#------------------------------------------------------------
# I<Constructor.>  Assuming this object was created with form input, 
# this routine takes that form input, parses and converts julian day to
# year, month, day, etc. instance vars.  Fatal error if unparseable.
#
# Returns:
#   'BLANK' (true) if original inputs were both blank.
#    The object itself (true) on success.
#    The undef value (false) on error.

sub parse_two_julian {
    my ($self, $datestr, $timestr, $opt) = @_;
               # HIGH opt means coerce to high time, LOW to low time

    # Create object if needed:
    ref($self) or $self = $self->new;

    # Strip leading/trailing blank space:
    $datestr =~ s/^\s+|\s+$//g if defined $datestr;
    $timestr =~ s/^\s+|\s+$//g if defined $timestr; 
    
    # Extract as much info as possible:
    if (defined $datestr and $datestr ne '') {        # date given...

	# Get date:
	$self->parse_date_julian($datestr) 
           or puke 511, "Bad date, '$datestr', (year must be 4 digits and Julian day 3 digits).";


	# Get any given time, or just use default:
	$timestr = ($timestr ne '') ? $timestr :
	    (($opt eq 'HIGH') ? '23:59:59' : '00:00:00');
	$self->parse_time($timestr) 
           or puke 511, "Bad time, '$timestr', (hours, minutes and seconds must each be 2 digits)";
    }
    else {                       # no date given...
	if (defined $timestr and 
            $timestr ne '') {            # ...but time given:
	    puke 516, "You gave a time ($timestr), but no date!";
	}
	else {                           # both blank:
	    return 'BLANK';
	}
    }

    # Still here?  Then we got something!
    return $self;
}

#------------------------------------------------------------
# parse_date_julian DATESTR
#------------------------------------------------------------
# Set elements of the datetime object based on the input string.  The 
# string must contain only the date, with possible leading/trailing whitespace.
# Returns self object, or undef on error.

sub parse_date_julian {    
    my $self = shift;
    my $str = shift; 
    defined($str) || ($str = '');
    my ($jday, $year);

    # must be in YYYY-DDD format:

    if ($str =~ m|^\s*(\d{4})-(\d{1,3})\s*$|) {
        my $year = $1;
        puke 511, "Please use 4 digits for year $year"  if (length($year)!=4);
	$self->{Year} = $year;
        my $jday = $2;

        my ($month, $day) = $self->Date_JulianToMonthDay($year, $jday);
	$self->{Mon} = $month;
	$self->{Day} = $day;  
	return $self;
    }  
    else {
	return undef;
    }
}

#------------------------------------------------------------
# test_day DAY MONTH YEAR
#------------------------------------------------------------
#  Figures out how many days are in the selected month then 
# compare that number against the user supplied day.
# Puke if necessary.

sub test_day {
     my ($self, $day, $month, $year) = @_;

     # Find out how many days are in $month
     my $no_of_days =  &Date_DaysInMonth ($month, $year);

     # report an error if the user supplied day is < 1 or greater
     # then the number of days allowed for that month.
     if ( $day > $no_of_days) {
        # $day = $no_of_days ; }
        puke 511, "The Day $day is invalid. Please supply a date between ( 1 - $no_of_days )";
     }
     elsif ( $day < 1 ) {
        # $day = 1 ;
	puke 511, "The Day $day is invalid. Please supply a date between 
           ( 1 - $no_of_days )";

     }
     return $day;
}

#------------------------------------------------------------
# parse_time TIMESTR
#------------------------------------------------------------
# Set elements of the datetime object based on the input string.  The 
# string must contain only the time, with possible leading/trailing whitespace.
# Returns self object, or undef on error.

sub parse_time {    
    my $self = shift;
    my $str = shift; 
    defined($str) || ($str = '');

    # (H)H:MM(:SS(.SSS))
    if ($str =~ /^\s*(\d\d?):(\d\d)(:(\d\d(\.\d*)?))?\s*$/) {
  
       # Check validity ( 0-23 )
       if ( $1 > 23 ) {
        puke 511, "The Hour $1 is invalid. Please supply an hour  between "
          ."( 0 - 23 )";
       } 
       $self->{Hour} = $1;
       if ( $2 > 59 ) {
        puke 511, "The Minute $2 is invalid. Please allow minutes to be between "
           ."( 00 - 59 )";
       }
       $self->{Min} = $2;
       my $seconds = ($4 ? $4 : 0);
       if ( $seconds > 59 ) {
        puke 511, "The Second $seconds is invalid. Please allow seconds to be between "
           ."( 00 - 59 )";
       }
       $self->{Sec} = $seconds;
       return $self;
    }  
    return undef;
}

#------------------------------------------------------------
# date_string
#------------------------------------------------------------
# Return date portion as a pretty, I<parseable> string.
sub date_string {
    sprintf "%04d-%02d-%02d", $_[0]->{Year}, $_[0]->{Mon}, $_[0]->{Day};
}

#------------------------------------------------------------
# date_string_julian
#------------------------------------------------------------
# Return date portion as a pretty, I<parseable> string.
sub date_string_julian {
    my $year = $_[0]->{Year};
    my $jday = &Date_MonthDayToJulian($year, $_[0]->{Mon}, $_[0]->{Day});
    sprintf "%04d-%03d", $year, $jday;
}
#------------------------------------------------------------
# time_string
#------------------------------------------------------------
# Return time portion as a pretty, I<parseable> string.
sub time_string {
    sprintf "%02d:%02d:%02d", $_[0]->{Hour}, $_[0]->{Min}, $_[0]->{Sec};
}

#------------------------------------------------------------
# as_string
#------------------------------------------------------------
# Return date and time as a pretty, I<but not parseable>, string.
sub as_string {
    sprintf("%02d %s %4d %02d:%02d:%02d (UTC)",
	    $_[0]->{Day}, $GMon[$_[0]->{Mon}], $_[0]->{Year},
	    $_[0]->{Hour}, $_[0]->{Min}, $_[0]->{Sec});
}

#------------------------------------------------------------
# as_parsable_string
#------------------------------------------------------------
# Return date and time as a pretty, but parseable string.
sub as_parsable_string {
    sprintf("%4d-%02d-%02d %02d:%02d:%02d",
            $_[0]->{Year}, $_[0]->{Mon}, $_[0]->{Day}, 
            $_[0]->{Hour}, $_[0]->{Min}, $_[0]->{Sec});
}


#------------------------------
#
# ODL interface
#
#------------------------------

#------------------------------------------------------------
# parse_TZ TZSTRING
#------------------------------------------------------------
# I<Constructor/initiallizer>.  Extract date/time info from an 
# ODL "TZ" string.  Returns self object, or undef on error.

sub parse_TZ {
    my $self = shift;
    my $tz = shift || '';

    ref($self) or $self = $self->new;     # doubles as a constructor
    defined($tz) || ($tz = '');

    ($tz =~ /^\s*(\d+)-(\d+)-(\d+)[Tt](\d+):(\d+):(\d+)[Zz]\s*$/)
	or return undef;
    ($self->{Year}, $self->{Mon}, $self->{Day},
     $self->{Hour}, $self->{Min}, $self->{Sec}) = ($1, $2, $3, $4, $5, $6);
    $self;
}

#------------------------------------------------------------
# from_ODL ODLDATE
#------------------------------------------------------------
# Create from an ODL "TZ" date object.

sub from_ODL {
    my ($self, $odl) = @_;

    require IU::ODL;
    $self->parse_TZ($odl->text);
}

#------------------------------------------------------------
# to_ODL
#------------------------------------------------------------
# Return date and time as an ODL "TZ" date object.

sub to_ODL {    
    my $self = shift;

    require IU::ODL;
    IU::ODL->bare(sprintf("%04d-%02d-%02dT%02d:%02d:%02dZ",
			  $self->{Year}, $self->{Mon}, $self->{Day},
			  $self->{Hour}, $self->{Min}, $self->{Sec}));
}



#------------------------------------------------------------
#
# Miscellaneous non-OO utilities...
#
#------------------------------------------------------------

#------------------------------------------------------------
# TZ_to_date TZSTRING
#------------------------------------------------------------
# I<Utility>.  Parse a TZ string, and return a prettified version of the date.
# Returns the argument itself if it can't be parsed.

sub TZ_to_date {
    ($_[0] =~ /^(\d+)-(\d+)-(\d+)/) ? sprintf("$3 $GMon[$2] $1") : $_[0];
}

#------------------------------------------------------------
# TZ_to_datetime TZSTRING
#------------------------------------------------------------
# I<Utility>.  Parse a TZ string, and return a prettified version of the date
# and time.  Returns the argument itself if it can't be parsed.

sub TZ_to_datetime {
   my $tz = shift;

   my $dt = $SESSION->state->{DateType} || "Standard Date Range";

   if ($dt eq "Standard Date Range") {
      ($tz =~ /^(\d+)-(\d+)-(\d+)[TZ](\d+:\d+(:[\d.]+)?)/i) ? 
	sprintf("$3 $GMon[$2] $1, $4") : $tz;
   }
   else {    
      # must be Julian Date
      if ( ($tz =~ /^(\d+)-(\d+)-(\d+)[TZ](\d+:\d+(:[\d.]+)?)/i) ) {
         my $jday = &Date_MonthDayToJulian($1, $2, $3);
         sprintf("$1-$jday, $4") ;
      }
      else {
         $tz;
      }
   }
}

#-----------------------------------------------------------#
# Date::Manip - date manipulation routines
#-----------------------------------------------------------#
# The following routines were extracted from 
# Date::Manip - date manipulation routines
# Copyright (c) 1995-1998 Sullivan Beck. All rights reserved.
# This program is free software; you can redistribute it and/or modify it
# under the same terms as Perl itself.
#
# There wasn't a need to install the complete package, so I extracted
# those components that we needed. JSA - 10/7/1998

#------------------------------------------------------------
# Date_DaysInMonth
#------------------------------------------------------------
#  $days=&Date_DaysInMonth($m,$y);
# Returns the number of days in the month.

sub Date_DaysInMonth {
  my($m,$y)=@_;
  puke 511, "Date_DaysInMonth: Please use 4 digits for year, $y"  if (length($y)!=4);
  my(@d_in_m)=(0,31,28,31,30,31,30,31,31,30,31,30,31);
  $d_in_m[2]=29  if (&Date_LeapYear($y));
  return $d_in_m[$m];
}

#------------------------------------------------------------
# Date_LeapYear
#------------------------------------------------------------
# g=&Date_LeapYear($y);
#
# Returns 1 if the argument is a leap year
# Written by David Muir Sharnoff <muir@idiom.com>

sub Date_LeapYear {
  my($y)=@_;
  puke 511, "Please use 4 digits for year" if (length($y)!=4);
  return 0 unless $y % 4 == 0;
  return 1 unless $y % 100 == 0;
  return 0 unless $y % 400 == 0;
  return 1;
}

#------------------------------------------------------------
# Date_DaysInYear
#------------------------------------------------------------
#  $days=&Date_DaysInYear($y);
# Returns the number of days in the year (365 or 366)

sub Date_DaysInYear {
  my($y)=@_;
  puke 511, "Please use 4 digits for year" if (length($y)!=4);
  return 366  if (&Date_LeapYear($y));
  return 365;
}

#------------------------------------------------------------
# Date_JulianToMonthDay
#------------------------------------------------------------
#  ($month, $day)=&Date_JulianToMonthDay($year,$jday);
# Returns the number of days in the month.
# Written by Cbrown July, 2000
sub Date_JulianToMonthDay {
  my ($self, $year, $jday) = @_;
  if ( ($jday < 1) || ( $jday > &Date_DaysInYear($year)) ) {
     puke 511, "The Julian Day $jday  is invalid. Please select a number between 1 - 366 "; }
  my $numDaysTotal = 0;
  my $numDaysInMonth;
  my $month=1;
  while ($numDaysTotal < $jday) {  
     $numDaysInMonth = &Date_DaysInMonth($month, $year);
     $month++;
     $numDaysTotal += $numDaysInMonth;
  }
  $month--;
  $numDaysTotal -= $numDaysInMonth;
  my $day = $jday - $numDaysTotal;
  return ($month, $day);
}

#------------------------------------------------------------
# Date_MonthDayTJulian
#------------------------------------------------------------
#   $jday=&Date_MonthDayToJulian($year,$month,$day);
# Returns the number of days in the month.
# Written by Cbrown July, 2000
sub Date_MonthDayToJulian {
  my ($year, $month, $day) = @_;
  my $numDaysInMonth = &Date_DaysInMonth($month, $year);
  if ( ($day < 1) || ( $day > $numDaysInMonth) ) {
     puke 511, "Date_MonthDayToJulian: The Day $day is invalid. Please select a number between 1 - $numDaysInMonth "; }
  if ( ($month< 1) || ( $month > 12) ) {
     puke 511, "Date_MonthDayToJulian: The Month $month is invalid. Please select a number between 1 - 12 "; }
  my $numDaysTotal = 0;
  my $monthNum=1;
  $month--;
  while ($monthNum <= $month) {  
     $numDaysInMonth = &Date_DaysInMonth($monthNum, $year);
     $monthNum++;
     $numDaysTotal += $numDaysInMonth;
  }
  my $jday = $day + $numDaysTotal;
  return ($jday);
}
   
#------------------------------------------------------------
1;

