package IW::DepValid;

=head1 NAME

IW::DepValid - interface to the dependent valids mechanism

=head1 SYNOPSIS

    # Get a dependent-valids interface object:
    $depval = new IW::DepValid;
     
    # Set up the values currently chosen:
    $depval->set('PARAMETER',   \@params);
    $depval->set('SOURCE_NAME', \@sources);
    $depval->set('SENSOR_NAME', \@sensors);
    $depval->set('CLOUD COVER', \@variants_chosen);
     
    # Get the valid categories to choose from:
    @categories = $depval->get_categories;
           
    # Get the filters for the desired category:
    @filters = $depval->get_filters('DATA_CENTER_ID');

    # Get the valids for the desired category:
    @valids = $depval->get_valids('DATA_CENTER_ID');
    
    # Same, but incorporate any filters/patters:
    @valids = $depval->get_valids('DATA_CENTER_ID',
				  Filters  => \@filters,
				  Patterns => \@patterns);

    # Get the valids for the desired category checking the pattern
    # against all categories.
    @valids = $depval->get_valids('DATA_CENTER_ID',
                                  Filters  => \@filters,
                                  Patterns => \@patterns);

    # Get criterion objects for extended search category:
    @crits = $depval->get_criteria("Cloud cover");

    # Translate aliases into their corresponding valids
    @valids = $depval->get_aliases('DATASET_ID');


=head1 VERSION

$Id: DepValid.pm,v 1.37 2002/09/17 19:58:03 joe Exp $

=cut

use IU::Chunk ();
use IU::Log;
use IU::ODL ();
use IU;

use IW::App;

use strict;


my %Variant;
my %SeqStringCategories = map {$_ => 1}
 qw(
   CAMPAIGN
   DATA_CENTER_ID
   DATASET_ID
   PARAMETER
   PROCESSING_LEVEL
   SENSOR_NAME
   SOURCE_NAME
   GRANULE_ID_REQ
   REQUEST_ID
);

my %StringCategories = map {$_ => 1}
 qw(
   TEXT_STRING
);

#------------------------------------------------------------
# new PARAMHASH
#------------------------------------------------------------
# I<Constructor.>

sub new {
    my ($class, %params) = @_;
    my $self = \%params;
    $self->{Fields} = {};
    $self->{Session_Path} = ( defined($params{Session_Path})  
	? $params{Session_Path} : $SESSION->path_to() );
    bless $self, $class;
}
#------------------------------------------------------------
# build_allquery [FIELD]
#------------------------------------------------------------
# I<Instance method, private.>
# Create an ODL object representing the fields in the current query.
# If optional FIELD is given, then that is the request field.
#
# Returns the ODL "group" object.

sub build_allquery {
    my ($self, $field) = @_;

    # Known values:
    my ($key, $vals);
    my $queryODL = IU::ODL->new_group;
    foreach $key (keys %{$self->{Fields}}) {
        $vals = $self->{Fields}{$key} or next;

        # old-style, "category = (values)"
        #if ($key =~ /^[A-Z_]+$/) {
        #    $queryODL->add_param('ATTRIBUTES' => IU::ODL->conv($vals));
        #}

        # new-style, "GROUP = FIELD"
        my $field_agg = IU::ODL->new_group;
        $field_agg->add_param('NAME'   => $key);
        $field_agg->add_param('VALUES' => IU::ODL->conv($vals));
        $queryODL->add_param('FIELD' => $field_agg);
    }

    # Request field:
    $queryODL->add_param('REQUEST_FIELD' => IU::ODL->quoted($field))
        if ($field);

    $queryODL;
}

#------------------------------------------------------------
# build_query [FIELD]
#------------------------------------------------------------
# I<Instance method, private.>
# Create an ODL object representing the fields in the current query.
# If optional FIELD is given, then that is the request field.
#
# Returns the ODL "group" object.

sub build_query {
    my ($self, $field) = @_;

    # Known values:
    my ($key, $vals);
    my $queryODL = IU::ODL->new_group;    
    foreach $key (keys %{$self->{Fields}}) {
	$vals = $self->{Fields}{$key} or next;

	# old-style, "category = (values)"
	#if ($key =~ /^[A-Z_]+$/) {
	#    $queryODL->add_param($key => IU::ODL->conv($vals));
	#}

	# new-style, "GROUP = FIELD"
	my $field_agg = IU::ODL->new_group;
	$field_agg->add_param('NAME'   => $key);
	$field_agg->add_param('VALUES' => IU::ODL->conv($vals));
	$queryODL->add_param('FIELD' => $field_agg);
    }

    # Request field:
    $queryODL->add_param('REQUEST_FIELD' => IU::ODL->quoted($field))
	if ($field);

    $queryODL;
}


#------------------------------------------------------------
# build_criteria [CRITERIA_NAME]
#------------------------------------------------------------
# I<Instance method, private.>
# Create an ODL object representing the fields in the current query.
# If optional CRITERIA_NAME is given, then that is the request field.
#
# Returns the ODL "group" object.

sub build_criteria {
    my ($self, $criteria_name) = @_;

    # Known values:
    my ($key, $vals);
    my $queryODL = IU::ODL->new_group;    

    # Request field:
    $queryODL->add_param('REQUEST_FIELD' => IU::ODL->quoted($criteria_name))
	if ($criteria_name);

    $queryODL;
}

#------------------------------------------------------------
# write_query ODL, OPTIONS
#------------------------------------------------------------
# I<Instance method, private.>
# Output the query ODL to a file in the session directory.
# The OPTIONS are...
#
#     File => Gives the relative name of the file.
#             The default is "depvalid.odl".
#     Type => Name of the top-level group, 'VALIDS' or 'FILTERS'.
#             The default is 'VALIDS'.
#
# Returns the path to the file, or undef/exception on error.

sub write_query {
    my ($self, $odl, %opts) = @_;
    my $name = $opts{File} || 'depvalid.odl';
    my $type = $opts{Type} || 'VALIDS';

    my $path = $self->{Session_Path}."/".$name;

    open QUERY, ">$path" or puke 452, "can't open $path for writing: $!";
    $odl->output(\*QUERY, $type);
    $odl->end(\*QUERY);
    close QUERY;
    $path;
}

#------------------------------------------------------------
# imsdepval ARGS...
#------------------------------------------------------------
# I<Instance method, private.>
# Run the imsdepval program, with the given ARGS.
# Returns array of all chunks that were read from it, empty on failure.

sub imsdepval {
    my ($self, @args) = @_;
    my @chunks;
    my $chunk;

    # Run program, get output:
    my $cmd = join ' ', "${IU::LibDir}/bin/imsdepval", @args;
    my $err = $self->{Session_Path}."last.err";

    debug "running: $cmd";
    open RESULT, "$cmd 2>$err |" or puke 453, "couldn't run imsdepval: $!";
    while ($chunk = IU::Chunk->input(\*RESULT)) {
	push @chunks, $chunk;
    }
    close RESULT;

    # Grab stderr output, iof necessary:
    @chunks or warn "no chunks in imsdepval results: \n" . `cat $err` . "\n";
    @chunks;
}



#------------------------------
#
# Deprecated interface...
#
#------------------------------

#------------------------------------------------------------
# get FIELD, [OPTIONS]
#------------------------------------------------------------
# I<Instance method, DEPRECATED by get_valids().>
# Compute the dependent valids for the given FIELD. 
#
# Returns an array of all valids in the requested field.
# Returns the empty array if results held no values.
# Returns a "There are no..." string if results were not obtained.
#
# B<Note:> Each invocation of this function runs a subprocess.

sub get {
    my ($self, $field) = @_;

    # Create an ODL object with the query included:
    my $queryODL = $self->build_query($field);
    
    # Create ODL input file in User's Session Directory:
    my $odlfile = $self->write_query($queryODL,
				     File=>'depvalidq.odl', Type=>'VALIDS');

    # Runs imsdepval in "give me valids mode"
    # Results are in Chunk ( Parameter=Value) format
    my @chunks = $self->imsdepval("-l -i $odlfile -mv");
    $chunks[0] or return "There are no compatible valids for this field!";
    return ($chunks[0]->values($field),       # old CGI style
	    $chunks[0]->values('RESULT'),     # new CGI style
            $chunks[0]->values('VALIDS'));    # APPLET style
}


#------------------------------
#
# Setting attributes and getting valids...
#
#------------------------------

#------------------------------------------------------------
# set FIELD, [VALUES]
#------------------------------------------------------------
# I<Instance method.>
# Set the FIELD to the given VALUES (a ref to an array).
# Setting a field to a ref to the empty array (or to undef) effectively 
# deletes it.
#
# With no VALUES argument, just returns the array of current settings.

sub set {
    my ($self, $field, $values) = @_;   
    if (@_ > 2) {      # set...
	my $clean = [ map {uc($_)} @{$values || []} ];
	if (@$clean) { $self->{Fields}{$field} = $clean }
	else         { delete $self->{Fields}{$field} }
    }
    return @{$self->{Fields}{$field} || []};      # get
}

#------------------------------------------------------------
# get_categories [OPTS]
#------------------------------------------------------------
# I<Instance method.>
# Given the current set of choices, compute all the valid categories
# (core and extended).
# OPTS is a hash, currently unused.
#
# Returns the list of categories, or empty/exception on failure.
# Now-incompatible categories are also returned, each preceded by a "*".
#
# B<Note:> Each invocation of this function runs a subprocess.

sub get_categories {
    my $self = shift;
    
    # Create an ODL object with the query:
    my $queryODL = $self->build_query;

    # Create ODL input file:
    my $odlfile = $self->write_query($queryODL, Type=>'VALIDS');
    
    # Run program, get output, and extract results:
    my @chunks = $self->imsdepval("-l -i $odlfile -mc");
    $chunks[0] or return ();
    return ($chunks[0]->values('CATEGORY'),
	    $chunks[0]->values('RESULT'));
}

#------------------------------------------------------------
# get_criteria FIELD
#------------------------------------------------------------
# I<Instance method.>
# Get the user-interface specs for each criterion [variant] in
# the given FIELD (an extended search category).
#
# WARNING:
# I believe that "imsdepval -mo" gets ALL the criteria chunks... not just 
# the chunks for valid criteria.  THIS IS OKAY!  However, users of this 
# method should be careful to determine which variant ARE valid, and which
# ones ARE NOT, before displaying the criteria objects in forms.

sub get_criteria {
    my ($self, $field) = @_;
    require IU::Criterion;

    # Create an ODL object with the query:
    my $queryODL = $self->build_query($field);

    # Create ODL input file:
    my $odlfile = $self->write_query($queryODL, Type=>'VALIDS');
    
    # Run program, get output, and extract results:
    my @chunks = $self->imsdepval("-l -i $odlfile -mo");
    @chunks or return ();

    # Convert result chunks to criteria:
    my $chunk;
    my @criteria;
    foreach $chunk (@chunks) {
	debug "Got criterion chunk";
	my $crit = IU::Criterion->from_chunk($chunk) || next;
	push @criteria, $crit;
    }

    # Done:
    return @criteria;
}


#------------------------------------------------------------
# get_criteria_name CRITERIA_NAME
#------------------------------------------------------------
# I<Instance method.>
# Get the user-interface specs for each criterion in
# that corresponds to the CRITERIA_NAME (within an extended search category).

sub get_criteria_name {
    my ($self, $field) = @_;
    require IU::Criterion;

    # Create an ODL object with the query:
    my $queryODL = $self->build_criteria($field);

    # Create ODL input file:
    my $odlfile = $self->write_query($queryODL, Type=>'VALIDS');
    
    # Run program, get output, and extract results:
    my @chunks = $self->imsdepval("-l -i $odlfile -mz");
    @chunks or return ();

    # Convert result chunks to criteria:
    my $chunk;
    my @criteria;
    foreach $chunk (@chunks) {
	debug "Got criterion chunk";
	my $crit = IU::Criterion->from_chunk($chunk) || next;
	push @criteria, $crit;
    }

    # Done:
    return @criteria;
}

#------------------------------------------------------------
# get_filters FIELD, [OPTS]
#------------------------------------------------------------
# I<Instance method. SUPERCEDES get_filter().>
# Compute the filters for the given FIELD.
# OPTS is a hash, which may contain...
#
# Returns an array of all valid filters of the requested field.
# Now-incompatible filters are also returned, each preceded by a "*".
# Returns the empty array if results held no values.
#
# B<Note:> Each invocation of this function runs a subprocess.

sub get_filters {
    my ($self, $field, %opts) = @_;

    # Create an ODL object with the query:
    my $queryODL = $self->build_query($field);

    # Create ODL input file in the user's session directory :
    my $odlfile = $self->write_query($queryODL, 
				     File=>'filterq.odl', 
				     Type=>'FILTERS');

    # Runs imsdepval in "give me valid filters" mode
    # Results are in Chunk ( Parameter=Value) format
    my @chunks = $self->imsdepval("-l -i $odlfile -mf");

    # Sanity Check 
    $chunks[0] or return ();

    return ($chunks[0]->values($field),       # old CGI style
	    $chunks[0]->values('RESULT'),     # new CGI style
            $chunks[0]->values('FILTERS'));    # APPLET style
}
#------------------------------------------------------------
# get_allvalids FIELD, [OPTS]
#------------------------------------------------------------
# I<Instance method.>
# Compute the valids for the given FIELD.
# Returns the valids that are compatible with the REQUEST_FIELD
# that are listed with the keyword "ATTRIBUTES", where
# the strings that follow the keyword can be any of the 7
# core attributes.
#
# OPTS is a hash, which may contain...
#
#      Patterns => This should be a ref to an array of patterns.
#
#      Filters  => This should be a ref to an array of filters.
#                  If it contains the special ">All compatible..." filter,
#                  then all filters and patterns will be ignored.
#
# Returns the list of compatible valids, or empty/exception on failure.
# Now-incompatible valids are also returned, each preceded by a "*".
#
# B<Note:> Each invocation of this function runs a subprocess.

sub get_allvalids {
    my ($self, $field, %opts) = @_;

    # Get the patterns:
    my @patterns = ($opts{Patterns} ? @{$opts{Patterns}} : ());

    # Get the filters:
    my @filters = sort ($opts{Filters} ? @{$opts{Filters}} : ());


    # Create an ODL object with the basic query:
    my $queryODL = $self->build_allquery($field);

    # The all-compatible case negates filters/patterns:
    if ($filters[0] and
        ($filters[0] =~ /^>All Compatible/i)) {   # if filtering is "ALL"...
        debug "get_valids: >All selected: ignoring filters/patterns";
    }
    else {        # "All" wasn't selected; just use what was there...
        # Add patterns:
        if (@patterns) {         # they want some patterns
            debug "get_allvalids: adding ", int(@patterns), " patterns";
            $queryODL->add_param('ATTRIBUTES' => IU::ODL->conv(\@patterns));
        }
    }

    # Create ODL input file:
    my $odlfile = $self->write_query($queryODL, Type=>'VALIDS');

    # Run program, get output, and extract results:
    my @chunks = $self->imsdepval("-l -i $odlfile -ma");
    $chunks[0] or return ();
    return ($chunks[0]->values($field),       # old-style
            $chunks[0]->values('RESULT'));    # new-style
}

#------------------------------------------------------------
# get_allkeywords FIELD, [OPTS]
#------------------------------------------------------------


sub get_allkeywords {
    my ($self,  %opts) = @_;

    # Create an ODL object with the basic query:
    my $queryODL = $self->build_allquery();

    # Create ODL input file:
    my $odlfile = $self->write_query($queryODL, Type=>'VALIDS');

    # Run program, get output, and extract results:
    my @chunks = $self->imsdepval("-l -i $odlfile -mb");
    $chunks[0] or return ();
    return ($chunks[0]->values('RESULT'));    # new-style
}

#------------------------------------------------------------
# get_allmatches FIELD, [OPTS]
#------------------------------------------------------------


sub get_allmatches {
    my ($self,  %opts) = @_;

    # Get the patterns:
    my @patterns = ($opts{Patterns} ? @{$opts{Patterns}} : ());

    # Create an ODL object with the basic query:
    my $queryODL = $self->build_allquery();

    # The all-compatible case negates filters/patterns:
        # Add patterns:
        if (@patterns) {         # they want some patterns
            debug "get_allvalids: adding ", int(@patterns), " patterns";
            $queryODL->add_param('ATTRIBUTES' => IU::ODL->conv(\@patterns));
        }

    # Create ODL input file:
    my $odlfile = $self->write_query($queryODL, Type=>'VALIDS');

    # Run program, get output, and extract results:
    my @chunks = $self->imsdepval("-l -i $odlfile -md");
    $chunks[0] or return ();
    return ($chunks[0]->values('RESULT'));    # new-style
}



#------------------------------------------------------------
# get_valids FIELD, [OPTS]
#------------------------------------------------------------
# I<Instance method.>
# Compute the valids for the given FIELD.  
# OPTS is a hash, which may contain...
#
#      Patterns => This should be a ref to an array of patterns.
#
#      Filters  => This should be a ref to an array of filters.
#                  If it contains the special ">All compatible..." filter,
#                  then all filters and patterns will be ignored.
#
#
# Returns the list of compatible valids, or empty/exception on failure.
# Now-incompatible valids are also returned, each preceded by a "*".
#
# B<Note:> Each invocation of this function runs a subprocess.

sub get_valids {
    my ($self, $field, %opts) = @_;

    # Get the patterns:
    my @patterns = ($opts{Patterns} ? @{$opts{Patterns}} : ());

    # Get the filters:
    my @filters = sort ($opts{Filters} ? @{$opts{Filters}} : ());

    ### NOTE: this "coercing" of the filter was in the old code.  
    ### It didn't seem right, and things work well without it, so I elided
    ### it when putting together the new code -- [Eryq]
    # defined($filters[0]) or $filters[0] = '';
    
    # Create an ODL object with the basic query:
    my $queryODL = $self->build_query($field);

    # The all-compatible case negates filters/patterns:
    if ($filters[0] and 
	($filters[0] =~ /^>All Compatible/i)) {   # if filtering is "ALL"...
	debug "get_valids: >All selected: ignoring filters/patterns";
    }
    else {        # "All" wasn't selected; just use what was there...
	# Add filters:
	if (@filters) {	    
	    debug "get_valids: adding ", int(@filters), " filters";
	    $queryODL->add_param('FILTER' => IU::ODL->conv(\@filters));
	}

	# Add patterns:
	if (@patterns) {         # they want some patterns
	    debug "get_valids: adding ", int(@patterns), " patterns";
	    $queryODL->add_param('PATTERN' => IU::ODL->conv(\@patterns));
	}
    }

    # Create ODL input file:
    my $odlfile = $self->write_query($queryODL, Type=>'VALIDS');
    
    # Run program, get output, and extract results:
    my @chunks = $self->imsdepval("-l -i $odlfile -mv");
    $chunks[0] or return ();
    return ($chunks[0]->values($field),       # old-style
	    $chunks[0]->values('RESULT'),     # new CGI Style
            $chunks[0]->values('VALIDS'));    # APPLET style
}

#------------------------------------------------------------
# get_field [OPTS]
#------------------------------------------------------------
# I<Instance method.>
# Compute the fields for the given PATTERN
# OPTS is a hash, which may contain...
#
#      Patterns => This should be a ref to an array of patterns.
#
# B<Note:> Each invocation of this function runs a subprocess.

sub get_field {
    my ($self, %opts) = @_;

    # Get the patterns:
    my @patterns = ($opts{Patterns} ? @{$opts{Patterns}} : ());

    # Create an ODL object with the basic query:
    my $queryODL = $self->build_query;

    # Add patterns:
    if (@patterns) {         # they want some patterns
       $queryODL->add_param('PATTERN' => IU::ODL->conv(\@patterns));
    }

    # Create ODL input file:
    my $odlfile = $self->write_query($queryODL, Type=>'VALIDS');
    
    # Run program, get output, and extract results:
    my @chunks = $self->imsdepval("-l -i $odlfile -mg");
    $chunks[0] or return ();
    return ($chunks[0]->values('RESULT'),     # new CGI Style
            $chunks[0]->values('VALIDS'));    # APPLET style
}

#------------------------------------------------------------
# get_aliases FIELD
#------------------------------------------------------------
# I<Instance method.>
# Translate aliases to valids for the given FIELD.
#
# Returns the list of compatible valids, or empty/exception on failure.
# Now-incompatible valids are also returned, each preceded by a "*".
#
# B<Note:> Each invocation of this function runs a subprocess.

sub get_aliases {
    my ($self, $field) = @_;


    # Create an ODL object with the basic query:
    my $queryODL = $self->build_query($field);

    # Create ODL input file:
    my $odlfile = $self->write_query($queryODL, Type=>'VALIDS');

    # Run program, get output, and extract results:
    my @chunks = $self->imsdepval("-l -i $odlfile -ms");
    $chunks[0] or return ();
    return ($chunks[0]->values($field),       # old-style
            $chunks[0]->values('RESULT'));    # new-style
}

#------------------------------------------------------------
# get_selectable_criteria 
#------------------------------------------------------------
# I<Instance method.>
# Translate aliases to valids for the given FIELD.
#
# Returns the list of compatible valids, or empty/exception on failure.
# Now-incompatible valids are also returned, each preceded by a "*".
#
# B<Note:> Each invocation of this function runs a subprocess.

sub get_selectable_criteria {
    my ($self, $field) = @_;

    # Create an ODL object with the basic query:
    my $queryODL = $self->build_query($field);

    # Create ODL input file:
    my $odlfile = $self->write_query($queryODL, Type=>'VALIDS');

    # Run program, get output, and extract results:
    my @chunks = $self->imsdepval("-l -i $odlfile -mr");
    $chunks[0] or return ();
    return ($chunks[0]->values($field),       # old-style
            $chunks[0]->values('RESULT'));    # new-style
}

#------------------------------------------------------------
# get_selectable_default 
#------------------------------------------------------------
# I<Instance method.>
# Translate aliases to valids for the given FIELD.
#
# Returns the list of compatible valids, or empty/exception on failure.
# Now-incompatible valids are also returned, each preceded by a "*".
#
# B<Note:> Each invocation of this function runs a subprocess.

sub get_selectable_default {
    my ($self, $field) = @_;

    # Create an ODL object with the basic query:
    my $queryODL = $self->build_query($field);

    # Create ODL input file:
    my $odlfile = $self->write_query($queryODL, Type=>'VALIDS');

    # Run program, get output, and extract results:
    my @chunks = $self->imsdepval("-l -i $odlfile -me");
    $chunks[0] or return ();
    return ($chunks[0]->values($field),       # old-style
            $chunks[0]->values('RESULT'));    # new-style
}


#------------------------------------------------------------
# get_variant CATEGORY, VARIANT
#------------------------------------------------------------
# I<Class method.>
# Get the criterion object for the given variant, or false if unknown.
# Uses internal cache to eliminate re-gets.
#
# B<Note:> Each invocation of this function will run a subprocess IF
# AND ONLY IF the variant has neither been cached or marked as unavailable
# in the cache.

sub get_variant {
    my ($class, $cat, $var) = @_;
    my $crit;
    $var = uc($var);       ### KLUDGE to case-fold variant names

    # Cached?
    defined($crit = $Variant{$cat}{$var}) and return $crit;

    # Get it:
    debug "get_variant: getting criteria for $cat and $var";
    my @crits = $class->new->get_criteria($cat);
    $Variant{$cat} ||= {};
    foreach $crit (@crits) {
	debug "get_variant: got variant ", uc($crit->variant);
	$Variant{$cat}{uc($crit->variant)} = $crit;
    }
    $Variant{$cat}{$var} ||= 0;       # we tried; don't try again
    $Variant{$cat}{$var};
}

#------------------------------------------------------------
# check_query
#------------------------------------------------------------
#
# Writes ODL to User's Session (file : imsDepValq.odl);
# runs imsdepval
# Return results

sub check_query {
    my $self = shift;

    # Build the ODL object (will include the user's query).
    my $query_odl = $self->build_query;

    # Save the ODL file into the user's session directory
    # FYI: If you change the name in the  'File' parameter, ModeMOREINFO must 
    # be  updated as well.
    my $odlfile = $self->write_query( $query_odl , File=>'imsDepValq.odl', 
                                     Type=>'VALIDS');

    # Runs imsdepval in "validate query mode"
    # Results are in Chunk ( Parameter=Value) format 
    my @chunks = $self->imsdepval("-l -i $odlfile -mq");

    # Sanity Check  
    $chunks[0] or print "No chunk to return \n" ;

    # Return results
    return ( $chunks[0]);
}


#------------------------------------------------------------
# valid_in ALL
#------------------------------------------------------------
# I<Utility function.>
# Given an array of strings, return the subset which has no "*" in front.

sub valid_in {
    my $all = shift;
    grep {!/^\*/} @$all;
}

#------------------------------------------------------------
# invalid_in ALL
#------------------------------------------------------------
# I<Utility function.>
# Given an array of strings, return the subset which has a "*" in front.
# Return the elements stripped of the "*".

sub invalid_in {
    my $all = shift;
    map { s/^\*//, $_ } (grep {/^\*/} @$all);
}


#------------------------------------------------------------
# from_ODL ODLOBJECT
#------------------------------------------------------------
# I<instance method.>
# Initiallize a DepValid object from a standard IMS ODL object.
#
# FYI : Routine created for imsDepVal.cgi, the WG normally creates a Dep
# Valid object after creating/from_ODLing a QUery object. JSA

sub from_ODL {
    my $self = shift;
    my $odl  = shift || return undef;
    my ($search, $param, $value, $field_odl);

    $odl->rewind();
    if ($search = $odl->param('VALIDS')) {
           while ($field_odl = $search->next_param('FIELD')) {
	      my $field_name = $field_odl->param('NAME')->text;
              my @field_values = $field_odl->param('VALUES')->text_values;
              $self->set($field_name, \@field_values);
           }
    }
    return $self;
}

#------------------------------------------------------------
1;
