package IW::GeoWidget::Base;

=head1 NAME

IW::GeoWidget::Base - abstract class for geoselection "widgets"

=cut

use strict;

use IU; # For puke.

use IW::GeoWidget::Mapfile; # List of maps

use IW::GeoWidget::RangeTypein;
use IW::GeoWidget::PointTypein;
use IW::GeoWidget::PathRowTypein;
use IW::GeoWidget::GlobalMap;
use IW::GeoWidget::GlobalGranOnly;
use IW::GeoWidget::JavaMap;
use IW::GeoWidget::PolarMap;
use IW::GeoWidget::RangeMap;

#------------------------------------------------------------
# new
#------------------------------------------------------------
# Constructor.

sub new {
    my ($class, %params) = @_;
    # Only return object if the Geotype requested is
    # supported by the subclass
    return undef unless $class->is_supported($params{Geotype});
    # Create the object
    my $self = bless \%params, $class;
    # Do subclass specific init:
    $self->init;
    $self;
}

#------------------------------------------------------------
# lookup PARAMHASH 
#------------------------------------------------------------
# Class/instance method, constructor.
# Given the type of geowidget, return
# an instance of some subclass capable of dealing with it.

sub lookup {
   my $class = shift;

   # If a supported geowidget type then one of these geowidgets will 
   # volunteer by returning an object. First one to volunteer gets the job.
   my $self = undef;
   $self = IW::GeoWidget::RangeTypein->new(@_) 
        || IW::GeoWidget::PointTypein->new(@_) 
        || IW::GeoWidget::PathRowTypein->new(@_) 
        || IW::GeoWidget::GlobalGranOnly->new(@_)
        || IW::GeoWidget::JavaMap->new(@_)
        || IW::GeoWidget::PolarMap->new(@_) 
        || IW::GeoWidget::RangeMap->new(@_) 
        || IW::GeoWidget::GlobalMap->new(@_);
    puke "unsupported geowidget type" unless ($self);
    return $self;
}

#------------------------------------------------------------
# determine_geotype GEO
#------------------------------------------------------------
# Class method
# Return geotype string that is compatable with the passed in geo object. 

sub determine_geotype  {
   my ($class,$geo) = @_;

   # See if any of the geo widgets support this geo object.
   # If none step up to the job (by returning geotype string), 
   # then default to 'Type in Lat/Lon Range'. 
   # Note: all RANGE_LOCs map to 'Type in Lat/Lon Range'. 
   return IW::GeoWidget::PointTypein->is_compatable($geo)
       || IW::GeoWidget::PathRowTypein->is_compatable($geo)
       || IW::GeoWidget::GlobalGranOnly->is_compatable($geo)
       || IW::GeoWidget::JavaMap->is_compatable($geo)
       || IW::GeoWidget::PolarMap->is_compatable($geo)
       || 'Type in Lat/Lon Range';
}

#------------------------------------------------------------
#------------------------------------------------------------
# Abstract methods
#------------------------------------------------------------
#------------------------------------------------------------

#------------------------------------------------------------
# init
#------------------------------------------------------------
# Instance method.  Initiallize.

sub init { 1 }

#------------------------------------------------------------
# draw
#------------------------------------------------------------
# Instance method, abstract.
# Draw the HTML form widget (map, etc.)

sub draw { 1 }

#------------------------------------------------------------
# update GEO 
#------------------------------------------------------------
# Instance method, abstract.
# Override in subclasses.
# Update current state of geo

sub update { 1 }

#------------------------------------------------------------
# summary 
#------------------------------------------------------------
# Instance method, abstract.
# Override in subclasses.
# Print HTML summary of current state of geo 

sub summary { 1 }

#------------------------------------------------------------
# is_supported GEOTYPE
#------------------------------------------------------------
# Instance method, abstract.
# Override in subclasses.
# Return false if GEOTYPE is not supported by the class
# Return true if GEOTYPE is supported by the class.

sub is_supported { die "Abstract method must override" }

#------------------------------------------------------------
# is_compatable GEO
#------------------------------------------------------------
# Instance method, abstract.
# Override in subclasses.
# Return undef if not compatable GEO
# or return compatable geotype string. 

sub is_compatable { die "Abstract method must override" }

#------------------------------------------------------------
1;
