package IW::GeoWidget::JavaMap;

=head1 NAME

IW::GeoWidget::JavaMap - do geoselection using the Java Applet

=cut

use strict;

use HTML::Stream qw(:funcs); # For html_escape function

use IU::Query; # for $QUERY

use IW::App; # For $HTML.
use IW::GranMap;
use IW::Util qw(:url);
use IW::Java; # for applet code.

use IW::GeoWidget::Base;
use vars qw(@ISA);
@ISA = qw(IW::GeoWidget::Base);

#------------------------------------------------------------
# update GEO
#------------------------------------------------------------
# Update current state of geo

sub update {

   my ($self,$geo) = @_;

   # Check if geo projection is compatable with the geo widget that is going
   # to displayed or set to default value (undef).
   $geo = undef unless $self->is_compatable($geo);

   # Capture selection region from CGI tags:
   $geo = IU::Geo::Polygon->from_CGI("p.")
      if ($REQ->param('p.projection') eq 'ORTHOGRAPHIC');

   return $geo;
}

#------------------------------------------------------------
# summary
#------------------------------------------------------------
sub summary {

   my $self = shift;

   my $geo= $QUERY->geo_region;

   print join("<BR>",split /\n/, html_escape($geo->as_string(Verbose=>2)))
      if (defined($geo));
   1;
}

#------------------------------------------------------------
# draw_popup
#------------------------------------------------------------
# Instance method.
# Output the java mapplet on popup page.

sub draw_popup {

   my $self = shift;

   # Print the "you might not have Java" message.
   $HTML->small;
   $HTML->t('The Orthographic Map (Java) tool will be displayed in a new '.
            'window when it is done loading - this may take a minute or two.');
   $HTML->P->CENTER;
   # Java is enabled applet:
   $HTML -> applet_code(ClassName  => 'IMSWG.IW.JavaEnabled.JavaEnabled.class',
                     CodeBase  => ${IU::DocsUrl}.'java/' ,
                 Alt => 'Simple java applet that indicates java is working', 
                     Width     => 150,
                     Height    => 30 );
   $HTML -> _applet_code;
   $HTML->_CENTER;
   print <<EOF;
          <P>
          <B>Note: if you don't see "Java is enabled" just above  
          this message</B>, either your browser doesn't support Java or you've
          turned Java off in your preferences.
          Click the <B>Cancel</B> button below to return to the Main
          Search Form and use a different Geographic selection if the Java Tool
          is not working for you.
EOF
   $HTML->_small;
   $HTML -> P;

   # The Mapplet:
   $HTML -> applet_code(ClassName  => 'IMSWG.IW.Mapplet.StartMapplet.class' ,
                     CodeBase  => ${IU::DocsUrl}.'java/',
         Alt => 'Java based tool for selecting a geographic region on a map', 
                     Width     => 2,
                     Height    => 2 );
   IW::Java::create_param();
   $HTML -> _applet_code;
   1;
}

#------------------------------------------------------------
# draw   
#------------------------------------------------------------
# Instance method.
# Output the summary map.

sub draw {
   my $self = shift;

   my $geo = $QUERY->geo_region;

   if (defined($geo)) {
      my $url = undef;
      if ($geo->ODL_type eq 'POLYGON_LOC') {

         my $orthogif = undef;
         my $zoom = $geo->zoom || 1;

         # hack to handle zooms greater than 16x
         # The java map goes to zoom higher than 16x, but the summary map
         # (imscovermap) only goes to zoom 16x, so we force higher zooms
         # to 16x.
         $zoom = 16 if ($zoom > 16);

         my $centerLat = $geo->tangent_lat || 0;
         my $centerLon = $geo->tangent_lon || 0;

         my @dummy = ();
         my $some_granules = \@dummy;
         my $newstamps = 0;
         my $xsize = 200;
         my $ysize = 150;

         # Generate the GIF:
         my @filenames = ();
         IW::GranMap::MakeMapGifs(Granules  => $some_granules,
                                Zoom      => $zoom,
                                CenterLat => $centerLat,
                                CenterLon => $centerLon,
                                NewStamps => $newstamps,
                                XSize     => $xsize,
                                YSize     => $ysize,
                                Gifnames  => \@filenames) || return undef;

         # Rename the main output file:
         my $path = $filenames[0];
         my $session_dir = $SESSION->dir;
         rename($path, "$session_dir/pub/geo_query.gif")
            or die "couldn't rename $path: $!";

         $url = $SESSION->url("pub/geo_query.gif?u=".scalar(time));
      } 
      print "<IMG NAME=map SRC=$url ALT=\'Map with selected geographic region'>"
         if $url;

      # Set geo region as CGI tags.
      $geo->set_CGI("p.");

      # Save complete state of geo map as hidden tags on the page.
      $HTML -> hidden('p.0.lat' => $REQ->param('p.0.lat'));
      $HTML -> hidden('p.0.lon' => $REQ->param('p.0.lon'));
      $HTML -> hidden('p.1.lat' => $REQ->param('p.1.lat'));
      $HTML -> hidden('p.1.lon' => $REQ->param('p.1.lon'));
      $HTML -> hidden('p.2.lat' => $REQ->param('p.2.lat'));
      $HTML -> hidden('p.2.lon' => $REQ->param('p.2.lon'));
      $HTML -> hidden('p.3.lat' => $REQ->param('p.3.lat'));
      $HTML -> hidden('p.3.lon' => $REQ->param('p.3.lon'));
      $HTML -> hidden('p.pole'  => $REQ->param('p.pole'));
      $HTML -> hidden('p.zoom'  => $REQ->param('p.zoom'));
      $HTML -> hidden('p.tangent.lat' => $REQ->param('p.tangent.lat'));
      $HTML -> hidden('p.tangent.lon' => $REQ->param('p.tangent.lon'));
      $HTML -> hidden('p.projection'  => $REQ->param('p.projection'));

   } else {

      $HTML->t('Currently no selection.')->BR;

   }

   $HTML->BR;
   $HTML->msubmit_button(QUERY => ['mode'=>$self->{ParentMode},
                                   'submode'=>'UPDATE',
                                   'Geoevent'=>'TOGGLE'],
                         CAPTION =>'Change Orthographic Map (Java) selection');
   1;   
}

#------------------------------------------------------------
# is_supported STRING
#------------------------------------------------------------
# class method
# Return true is this class supported the passed in geotype.

sub is_supported {
   my ($class,$geotype) = @_;
   return 1 if ($geotype eq 'Orthographic Map (Java)');
   0;
}

#------------------------------------------------------------
# is_compatable GEO
#------------------------------------------------------------
# class method
# Return geotype if this class supports the passed in geo object.

sub is_compatable {
   my ($class,$geo) = @_;

   my $ODL_type = $geo ? $geo->ODL_type : undef;
   my $projection = ($ODL_type eq 'POLYGON_LOC') ? $geo->projection : undef;
   return 'Orthographic Map (Java)' if ($projection eq 'ORTHOGRAPHIC');
   undef;
}

#------------------------------------------------------------
1;
