package IW::GeoWidget::Mapfile;

=head1 NAME

IW::GeoWidget::Mapfile - a GIF file for a map

=head1 DESCRIPTION

An instance of this class represents a single GIF file used
for rendering a map in the geoselection screen.

=head1 SYNOPSIS

    # Get all known map names:
    @names = IW::GeoWidget::Mapfile->names;
    
    # Lookup a mapfile by name:
    $map = lookup IW::GeoWidget::Mapfile $names[0];
    
    # Get the GIF filename:
    $filename = $map->filename;

=cut

use IU::Log;
use IU::Vec;

use strict;
use POSIX qw(:math_h);

#------------------------------------------------------------
#
# GLOBALS
#
#------------------------------------------------------------

# The widths of the "axis" borders on the image generated by "imsmap".
# KLUDGE!  These should be taken from "imsmap.h" (or the reverse)!
# my $XBORDER = 26; 52; 28
# my $YBORDER = 17; 34; 62
my $XBORDER = 26;
my $YBORDER = 17;

# All known map data:
my @AllMaps = 
    (

     {   Name   => 'Equatorial Map (small)',
         File   => 'world-blue-sm',
         Height => 135,
         Width  => 180,
         Projection => 'RECTANGULAR',
         Location => 'map'         # 'mapfull'-full path or 'map'-relative
     },     

     {   Name   => 'Equatorial Map',
         File   => 'world-atlas',      # GIF name in ./docs/pics, w/o ".gif"
         Height => 270,                # GIF height, in pixels
         Width  => 360,                # GIF width, in pixels
         Projection => 'RECTANGULAR',  # projection type
         Location => 'map'         # 'mapfull'-full path or 'map'-relative
     },

     # The actual AIRs images are 180x360, but for some reason our formula 
     # doesn't like it when the image size is the same size as the degrees.
     # So we Kludge it by making the formula think the latitude pixals are 
     # bigger. Simpler then rewriting the code to handle cylindrical projection
     # Which I think we are "in sense" doing.  JSA - 04'03
     {   Name   => 'Browse',
         File   => 'world-atlas',      # GIF name in ./docs/pics, w/o ".gif"
         Height => 270,                # Kludge
         Width  => 360,                # GIF width, in pixels
         Projection => 'RECTANGULAR',  # projection type
         Location => 'mapfull'         # 'mapfull'-full path or 'map'-relative
     },


     {   Name   => 'Stereographic N-pole Map',
         File   => 'stereo-north2',
         Height => 350,
         Width  => 350,
         Radius        => 175,
         CenterLat     => 90,
         CenterLon     => 0,
	 Projection    => 'STEREO', 
	 ODLProjection => 'NORTH_POLAR_STEREOGRAPHIC', 
     },

     {   Name   => 'Stereographic S-pole Map',
         File   => 'stereo-south2',
         Height => 350,
         Width  => 350,
         Radius => 175,
         CenterLat     => -90,
         CenterLon     => 0,
         Projection    => 'STEREO', 
         ODLProjection => 'SOUTH_POLAR_STEREOGRAPHIC', 
     },

    );

# All known map data, hashed by name:
my %MapsByName = map {($_->{Name}, $_)} @AllMaps;

     
#------------------------------------------------------------
# names
#------------------------------------------------------------
# Class method. 
# Return array of all known mapfile names.

sub names {
    my $class = shift;
    keys %MapsByName;
}

#------------------------------------------------------------
# lookup NAME
#------------------------------------------------------------
# Constructor.
# Lookup (and return) a mapfile, by name.
# Returns undefined (NOT a fatal exception!) if no such map exists.

sub lookup {
    my ($class, $name) = @_;
 
    $MapsByName{$name} or return undef;
    bless $MapsByName{$name}, $class;       # bless and return in-place!
}

#------------------------------------------------------------
# xy_to_latlon_RECTANGULAR X,Y
#------------------------------------------------------------
# Instance method, private.
# Convert coords (of a mouse click) to a floating-point lat/lon.

sub xy_to_latlon_RECTANGULAR {
    my ($self, $x, $y) = @_;
    my ($lat, $lon);

    # Latitude:
    $lat = 90 - (($y - $YBORDER) * (180.0 / $self->{Height}));
    $lat = sprintf("%0.3f", $lat);
    $lat = -90.0 if ($lat < -90.0);
    $lat =  90.0 if ($lat >  90.0);

    # Longitude:
    $lon = -180.0 + (($x - $XBORDER) * (360.0 / $self->{Width}));
    $lon = sprintf("%0.3f", $lon);   
    $lon = -180.0 if ($lon < -180.0);
    $lon =  180.0 if ($lon >  180.0);

    debug "x= $x, y = $y, lat = $lat, lon = $lon \n";

    return ($lat, $lon);
}

#------------------------------------------------------------
# xy_to_latlon_STEREO X,Y
#------------------------------------------------------------
# Instance method, private.
# Convert y-coord (of a mouse click) to a floating-point latitude.
#

sub xy_to_latlon_STEREO {
    my ($self, $acx, $acy) = @_;
    my ($c_lat, $c_lon);
    my ($phi, $lam);
    my ($lat, $lon);
    
    # Convert click point which is in terms of the entire gif to
    #    a coordinate system centered on the map. The assumption
    #    made here is that CenterLon and CenterLat are in the 
    #    exact center of the gif.    
    my $cx = ($acx - ($self->{Width} / 2));
    my $cy = -($acy - ($self->{Height} / 2));
    my $scaled_radius = ($self->{Radius}) / 2;
    
    my $sqr_x = $cx * $cx;
    my $sqr_y = $cy * $cy;
    my $rho = sqrt($sqr_x + $sqr_y);

    # Which hemispehere are we in?
    if ($self->{CenterLat} >= 0.0) {  
	if ($rho > 0.0) {
	    my $q = $scaled_radius * (1 + sin(($IU::Vec::Pi)/2));
	    my $c = 2 * atan2($rho,$q);
	    $phi = asin(cos($c));
	    $lam = atan2($cx, -$cy);
	} 
	else {
	    $phi = ($IU::Vec::Pi)/2;
	    $lam = 0.0;
	}
    } 
    else {                # $self->{CenterLat} < 0.0
	if ($rho > 0.0) {
	    my $q = $scaled_radius * (1 - sin(-($IU::Vec::Pi)/2));
	    my $c = 2 * atan2($rho,$q);
	    $phi = asin(-cos($c));
	    $lam = atan2($cx, $cy);
	} 
	else {
	    $phi = -($IU::Vec::Pi)/2;
	    $lam = 0.0;
	}
    }
    # Latitude:
    $lat = $phi *  180.0 / ($IU::Vec::Pi);

    # Longitude:
    $lon = $lam *  180.0 / ($IU::Vec::Pi);
    $lon += 360.0 if ($lon < -180.0);
    $lon -= 360.0 if ($lon > 180.0);

    # Covert to lat/lon!
    ($c_lat, $c_lon) = (int($lat), int($lon));
    return ($c_lat, $c_lon);
}

#-----------------------------------------------------------
# xy_to_latlon X,Y
#-----------------------------------------------------------
# Instance method.
# Given a click at X,Y (on the GIF), return the corresponding array (lat,lon).
# Uses the map's projection to determine who conversion should be done.
# Automatically does a "fixup" of x,y.

sub xy_to_latlon {
    my ($self, $x, $y) = @_;
    ($x, $y) = $self->fixup($x, $y);
    my $method = "xy_to_latlon_$self->{Projection}";
    return $self->$method($x, $y);
}

#-----------------------------------------------------------
# filename
#------------------------------------------------------------
# Instance method.
# Return just the filename for the map, complete with .gif extension.

sub filename {
    my ($self, $value) = @_;
    $self->{File} = $value if (@_ > 1);
    $self->{File}.".gif";
#    "$_[0]->{File}.gif";
}

#------------------------------------------------------------
# Other access methods...
#------------------------------------------------------------
sub name       { $_[0]->{Name} }
sub projection     { $_[0]->{Projection} }
sub location     { $_[0]->{Location} }
sub ODL_projection { $_[0]->{ODLProjection} }
sub height     { 
    my ($self, $value) = @_;
    $self->{Height} = $value if (@_ > 1);
    $self->{Height};
}
sub width      { 
    my ($self, $value) = @_;
    $self->{Width} = $value if (@_ > 1);
    $self->{Width}; 
}
sub radius     { $_[0]->{Radius}}
sub center_lat { $_[0]->{CenterLat}}
sub center_lon { $_[0]->{CenterLon}}

#-----------------------------------------------------------
# script
#------------------------------------------------------------
# Instance method.
# Return the name of the script to use for rendering this map.
# Uses the projection.

sub script {
    my $self = shift;
    return 'imsmap.cgi'      if ($self->{Projection} eq 'RECTANGULAR');
    return 'imsorthomap.cgi' if ($self->{Projection} eq 'STEREO');

return  if ($self->{Projection} eq 'JAVA');

    undef;
}

#------------------------------------------------------------
# fixup X,Y
#-------------------------------------------------------------
# Instance method.
# Fixup click so clicks off map are projected to map edge.
# Used for projections which show sphere as a circle.
#
# Returns fixed (X,Y).
# Returns immediately if projection is rectangular.

sub fixup {    
    my ($self, $x, $y) = @_;

    # Does fixup NEED to be done?    
    return ($x, $y) if ($self->{Projection} !~ /^(ORTHO|STEREO)$/);
    
    # Convert click point which is in terms of the entire gif to
    #   a coordinate system centered on the map. The assumption
    #   made here is that CenterLon and CenterLat are in the
    #   exact center of the gif.
    my $cx =  ($x - ($self->{Width}  / 2));
    my $cy = -($y - ($self->{Height} / 2));

    # Compute the distance of the point in pixels from center of map:
    my $distance = sqrt(($cx * $cx) + ($cy * $cy));
    return ($x, $y) if ($distance == 0);        # don't risk div-by-0 error!

    # If click point is outside the edge, assume that user is a sloppy 
    # clicker, and gravitate the point to the edge:
    if ($distance > $self->{Radius}) {       
	$cx *= ($self->{Radius} / $distance);  # normalize x
	$cy *= ($self->{Radius} / $distance);  # normalize y
	$x =  $cx + ($self->{Width}  / 2);
	$y = -$cy + ($self->{Height} / 2);
    }	
    return ($x, $y);
}

#------------------------------------------------------------
1;



