#============================================================
#
package IW::GeoWidget::PolarMap;
#
#============================================================
use strict;

use HTML::Stream qw(:funcs); # for html_escape function.

use IU::Geo::Point;
use IU::Geo::Polygon;
use IU::Query; # For $QUERY

use IW::App; # For $HTML.
use IW::Util qw(:url);

use IW::GeoWidget::Base;
use vars qw(@ISA);
@ISA = qw(IW::GeoWidget::Base);

#------------------------------------------------------------
# init
#------------------------------------------------------------
# Override of parent init
# Called by new constructor
sub init {
   my $self = shift;
   $self->{Mapfile} = lookup IW::GeoWidget::Mapfile $self->{Geotype};
   1;
}

#------------------------------------------------------------
# update GEO
#------------------------------------------------------------
# Update current state of geo

sub update {

   my ($self,$geo) = @_;

   my $mapfile = $self->{Mapfile};

   # Convert x/y clicks to lat/lon points:
   $self->_get_input;

   # If mapstate is GOTPOINT user is in the middle of selecting a geo so set to 
   # to undef.
   if ($REQ->param('p.mapstate') eq 'GOTPOINT') {
      $geo = undef;
   } 
   # Capture selection region from CGI tags:
   elsif (($REQ->param('p.mapstate') eq 'GOTRECT') &&
          ($REQ->param('p.projection') eq $mapfile->ODL_projection)) {
      $geo = IU::Geo::Polygon->from_CGI("p.");
   } 
   # Check if geo projection is compatable with the geo widget that is going
   # to displayed or set to default value (undef).
   elsif ($self->is_compatable($geo) ne $mapfile->name) {
      $REQ->param('p.mapstate'=>'START');
      $geo = undef; 
   }

   return $geo;
}

#------------------------------------------------------------
# _get_input
#------------------------------------------------------------
# Process a user's x/y click on a map, translate into lat/lon
# and then set appropriate CGI tags for use by draw method.

sub _get_input {

   my $self = shift;
   my @coords;
   my ($x0, $y0, $x2, $y2);  # mouse click, raw
   my ($lat, $lon);          # mouse click, translated into coords

   my $mapfile = $self->{Mapfile};
   my $mapclick   = $REQ->param('p.mapclick') || 'START';

   # Handle first map click: 
   if ($mapclick eq 'GOTPOINT')  {    
	
      # Get coordinates of first point (point 0 in 0..3):
      if (defined($REQ->param('msubmit.x'))) {

         # Get coordinates from most-recent mouse click (x,y):
         $x0 = $REQ->param('msubmit.x');
         $y0 = $REQ->param('msubmit.y');

         # Re-enter coordinates as named fields (to init any fill-ins):
         ($lat, $lon) = $mapfile->xy_to_latlon($x0, $y0);
         IU::Geo::Point->new($lat, $lon)->set_CGI("p.0.");

         $REQ->param('p.mapstate', 'GOTPOINT');

         # Add x,y coords, so they can be used for seoncd click:
         $REQ->param("p.0.x" => $x0);
         $REQ->param("p.0.y" => $y0);
      }
      # ...at this point, coords of 1st point in REQ.
   }

   # Handle second map click:
   elsif ($mapclick eq 'GOTRECT') {

      # Get coordinates of second point (point 2 in 0..3):
      if (defined($REQ->param('msubmit.x'))) {

         #--------------------
         # GET ALL FOUR POINTS FROM TWO CLICKS...
         #--------------------
         # This next part is slightly tricky, since we want to ensure
         # that points are in clockwise order as seen from above...

         # Get original click (point 0 in 0..3):
         $x0 = $REQ->param('p.0.x');
         $y0 = $REQ->param('p.0.y');

         # Get coordinates from most-recent mouse click (x,y):
         $x2 = $REQ->param('msubmit.x');
         $y2 = $REQ->param('msubmit.y');
	    
         # Muck with points so that point 0 is top-left, and point 2
         # is bottom-right:
         if ($x0 > $x2) { ($x0, $x2) = ($x2, $x0) }
         if ($y0 > $y2) { ($y0, $y2) = ($y2, $y0) }

         # Re-enter coordinates as named fields (to init any fill-ins):
         ($lat, $lon) = $mapfile->xy_to_latlon($x0, $y0);
         IU::Geo::Point->new($lat, $lon)->set_CGI("p.0.");
         #
         ($lat, $lon) = $mapfile->xy_to_latlon($x2, $y0);
         IU::Geo::Point->new($lat, $lon)->set_CGI("p.1.");
         #
         ($lat, $lon) = $mapfile->xy_to_latlon($x2, $y2);
         IU::Geo::Point->new($lat, $lon)->set_CGI("p.2.");
         #
         ($lat, $lon) = $mapfile->xy_to_latlon($x0, $y2);
         IU::Geo::Point->new($lat, $lon)->set_CGI("p.3.");


         #--------------------
         # GET POLE INFO...
         #--------------------

         # Is the pole included?  Store that information as well:
         my $cx   = $mapfile->width / 2;
         my $cy   = $mapfile->height / 2;
         my $pole = 'X';       # assume none
         if (($x0 <= $cx) && ($cx <= $x2) &&
             ($y0 <= $cy) && ($cy <= $y2)) {	    # box surrounds map center

            $pole = (($mapfile->center_lat > 0) ? 'N' : 'S');
	 }
	 $REQ->param("p.pole" => $pole);
         $REQ->param('p.mapstate', 'GOTRECT');
      # ...at this point, coords of all four points in REQ.
      }
   }
   1;
}

#------------------------------------------------------------
# draw
#------------------------------------------------------------
# By the time draw is called it is assumed that any map clicks
# have already been handled by calling _get_input.
# _get_input is responsible for turning x/y clicks into lat/lon points
# stored in CGI param tags which are used here.
#
# The map cycles through three states: 'START', 'GOTPOINT' and 'GOTRECT'.
# Note that these states exist only in CGI, not with the session.
# If a user back buttons to a previous form the state from that form
# is used and the geo stored in the session is discarded, so what the use
# sees is consistent with what happens.
#
# START:
# This is the default state when there is no state, either because the form is
# just being displayed or the geo type has been switched.
#
# GOTPOINT:
# This is the state after the user has select one corner on the map
#
# GOTRECT:
# This is the state after the user has completed selecting a box on the map
# by clicking a second corner.
#
# If user clicks the map again the state will cycle back to GOTPOINT:


sub draw {
   my $self = shift;

   my $mapfile = $self->{Mapfile};

   # Current state of map:
   my $mapstate = $REQ->param('p.mapstate') || 'START';

   my $directions;
   my $footnote;
   my $mapclick = 'START';
   my @coords;

   # Get points:
   my @p;
   $p[0] = IU::Geo::Point->from_CGI("p.0.");
   $p[2] = IU::Geo::Point->from_CGI("p.2.");

   # mapstate = START:
   # Show initial map state. If geo is undefined then map is blank,
   # or if there is a geo that is displayed.
   if ($mapstate eq 'START') {

      # The directions and the next map state:
      $directions = ("<FONT SIZE=5>Step 1: </FONT>".
		     "<B>Select the first point</B> of your ".
		     "search rectangle:");
      my $geo = $QUERY->geo_region;

      if ($geo and ($geo->ODL_type eq 'POLYGON_LOC') and 
          ($geo->projection eq $mapfile->ODL_projection) ) {

         # initialize with previous selected range

         $geo->set_CGI("p.");

         $p[0] = IU::Geo::Point->from_CGI("p.0.");
         $p[2] = IU::Geo::Point->from_CGI("p.2.");

	 @coords = ('srect' => (join ",",
                               $p[0]->lon, $p[0]->lat,
                               $p[2]->lon, $p[2]->lat));
      }
      # What will be the next state if the map is clicked.
      $mapclick = 'GOTPOINT';
   }

   ## STATE = GOTPOINT: display map with one point...
   elsif ( $mapstate eq 'GOTPOINT') {

      # The directions and the next map state:
      $directions = ("<FONT SIZE=5>Step 2: </FONT>".
                     "Your first selected point is shown. ".
	             "<B>Select the opposite corner </B> ".
	             "of your search rectangle:");

      # The footnote below the map:
      my $precision = 3; # The precision (significant number of digits).
      $footnote = ("<B>Selected point:</B> ".
         IU::Geo::Point->from_CGI("p.0.")->fixup($precision)->as_string);

      # Get coords to draw:
      @coords = ('spoint' => (join ",", $p[0]->lon, $p[0]->lat));

      # What will be the next state if the map is clicked.
      $mapclick = 'GOTRECT';

      # Remember first x/y click, so it can be used in calculating box
      # after second click: 
      $HTML->hidden('p.0.x'   => $REQ->param('p.0.x'));
      $HTML->hidden('p.0.y'  => $REQ->param('p.0.y'));

   }

   ## STATE = GOTRECT: display map with two points...
   elsif ($mapstate eq 'GOTRECT') {

      # The directions and the next map state:
      $directions = ("<FONT SIZE=5>Step 3: </FONT>".
                     "This is your search rectangle. ".  
	             "Select a new first ".
	             "corner on the map to start over:");

      # Get coords to draw:
      @coords = ('srect' => (join ",", 
			       $p[0]->lon, $p[0]->lat,
			       $p[2]->lon, $p[2]->lat));
 
      # What will be the next state if the map is clicked.
      # We have a box, so next click will cycle back to first point.
      $mapclick = 'GOTPOINT';

   }

   # Get source for GIF... this is the NEXT map they say they want:
   my $map_gif = query_url($mapfile->script,
			    'map'        => $mapfile->filename,
                            @coords,
                            'projection' => lc($mapfile->projection),
                            'height'     => $mapfile->height,
                            'width'      => $mapfile->width,
                            'radius'     => $mapfile->radius,
                            'centerlat'  => $mapfile->center_lat,
                            'centerlon'  => $mapfile->center_lon,
			    'sess_dir'   => $SESSION->dir);

   # Output map directions for STEP 1, STEP 2, or STEP 3
   print $directions if ($directions);

   # Output map:
   $HTML -> CENTER;
   $HTML -> msubmit_button
                   (SRC   =>$map_gif,
                    ALT   =>$mapfile->name,
		    QUERY =>['mode'    => $self->{ParentMode},
                             'submode' => 'UPDATE',
                             'Geoevent' => 'CLICK-ON-MAP',
                             'p.mapclick' => $mapclick,
                             'map'     => 'YES']);
 
   # Output footnote (if any):
   $HTML->BR;
   print $footnote if ($footnote);
   $HTML -> _CENTER;

   # Save complete state of geo map as hidden tags on the page. 
   $HTML -> hidden('p.mapstate' => $mapstate);
   $HTML -> hidden('p.0.lat' => $REQ->param('p.0.lat'));
   $HTML -> hidden('p.0.lon' => $REQ->param('p.0.lon'));
   $HTML -> hidden('p.1.lat' => $REQ->param('p.1.lat'));
   $HTML -> hidden('p.1.lon' => $REQ->param('p.1.lon'));
   $HTML -> hidden('p.2.lat' => $REQ->param('p.2.lat'));
   $HTML -> hidden('p.2.lon' => $REQ->param('p.2.lon'));
   $HTML -> hidden('p.3.lat' => $REQ->param('p.3.lat'));
   $HTML -> hidden('p.3.lon' => $REQ->param('p.3.lon'));
   $HTML -> hidden('p.pole' => $REQ->param('p.pole'));
   $HTML -> hidden('p.tangent.lat'   => $mapfile->center_lat);
   $HTML -> hidden('p.tangent.lon'  => $mapfile->center_lon);
   $HTML -> hidden('p.projection'  => $mapfile->ODL_projection);
   1;
}

#------------------------------------------------------------
# summary
#------------------------------------------------------------
sub summary {

   my $self = shift;

   my $geo= $QUERY->geo_region;

   print join("<BR>",split /\n/, html_escape($geo->as_string(Verbose=>2)))
      if (defined($geo));
   1;
} 

#------------------------------------------------------------
# is_supported STRING
#------------------------------------------------------------
# class method
# Return true is this class supports the passed in geotype.

sub is_supported {
   my ($class,$geotype) = @_;
   return 1 if (($geotype eq 'Stereographic N-pole Map') ||
                ($geotype eq 'Stereographic S-pole Map') );
   0;
}

#------------------------------------------------------------
# is_compatable GEO
#------------------------------------------------------------
# class method
# Return geotype if this class supports the passed in geo object.

sub is_compatable {
   my ($class,$geo) = @_;

   my $ODL_type = $geo ? $geo->ODL_type : undef;
   my $projection = ($ODL_type eq 'POLYGON_LOC') ? $geo->projection : undef;
   return 'Stereographic N-pole Map' 
      if ($projection eq 'NORTH_POLAR_STEREOGRAPHIC');
   return 'Stereographic S-pole Map' 
      if ($projection eq 'SOUTH_POLAR_STEREOGRAPHIC');
   undef;
}

#------------------------------------------------------------
1;

