package IW::GeoWidget::RangeMap;

=head1 NAME

IW::GeoWidget::RangeMap - do geoselection from just a range map

Even though this geo widget doesn't use javascript it shares cgi tags
with RangeTypein, so this applies here also:

While the convention on the search form is to use 'p.' for  prefix of
cgi tags, I am using 'p_' here because the 'p.' confuses javascript.
For example: if you had a type-in box named 'p.lat' javascript complains
there is no 'p' object when you try to read window.form.p.lat, whereas
window.form.p_lat works fine. - ERS

=cut

use strict;

use HTML::Stream qw(:funcs); # For html_escape function.

use IU::Coord; # For str2deg function.
use IU::Query; # For $QUERY.

use IW::App; # For $HTML
use IW::Util qw(:url);

use IW::GeoWidget::Base;
use vars qw(@ISA);
@ISA = qw(IW::GeoWidget::Base);

#------------------------------------------------------------
# init
#------------------------------------------------------------
# Override of parent init
# Called by new constructor
sub init {
   my $self = shift;
   # if AIRS browse, use the browse file (as indicated by
   # 'p.browse_location') for the map, otherwise use standard map:
   my $replace_map =  $REQ->param('p.browse_location');
   if (defined($replace_map)) {
      $self->{Mapfile} = lookup IW::GeoWidget::Mapfile 'Browse';
      # Remove .gif extension to be consistent with how the other
      # map names are stored in Mapfile:
      $replace_map =~ s/.gif//; # Remove .gif extension 
      $self->{Mapfile}->filename($replace_map);
   } else { 
      $self->{Mapfile} = lookup IW::GeoWidget::Mapfile 'Equatorial Map';
   }
   1;
}

#------------------------------------------------------------
# update GEO
#------------------------------------------------------------
# Update current state of geo

sub update {

   my ($self,$geo) = @_;

   # Convert x/y clicks on map to $REQ params p_north, p_south, p_west, p_east:
   $self->_get_input;

   # GOTPOINT means user is in the middle of selecting geo so set to undef: 
   if ($REQ->param('p_mapstate') eq 'GOTPOINT') {
      $geo = undef;
   }
   # Capture range related input from form:
   elsif (defined($REQ->param('p_north'))) {
      $geo = IU::Geo::Range->from_CGI("p_");
   }
   # Check if geo ODL type is compatable with the geo widget that is going
   # to displayed or set to default values (Global).
   elsif (!($self->is_compatable($geo))) {
      $REQ->param('p_mapstate'=>'START');
      $geo = IU::Geo::Range->new(90.0, -180.0, -90.0, 180.0);
   }

   # Clear RANGE_LOC CGI tags so they don't carry over to draw:
   $REQ->param('p_north'=>undef);
   $REQ->param('p_west' =>undef);
   $REQ->param('p_south'=>undef);
   $REQ->param('p_east' =>undef);

   return $geo;
}

#------------------------------------------------------------
# _get_input
#------------------------------------------------------------
# Process a user's x/y click on a map, translate into lat/lon
# and then set appropriate CGI tags for use by draw method.


sub _get_input {
   my $self = shift;
   my %params = @_;
   my @coords;
   my ($n, $w, $s, $e);

   my $mapfile = $self->{Mapfile};
   my $mapclick   = $REQ->param('p_mapclick') || 'START';

   # Handle first map click: 
   if ($mapclick eq 'GOTPOINT') {    
      # Get coordinates of 1st point (N-W):
      if (defined($REQ->param('msubmit.x'))) {   # got here via a map click

         # Get coordinates from most-recent mouse click (x,y):
	 ($n, $w) = $mapfile->xy_to_latlon($REQ->param('msubmit.x'),
					   $REQ->param('msubmit.y'));

         # Save coordinates of first click as CGI tags
         # These will be used by draw
         $REQ->param('mapclick_north',$n);
         $REQ->param('mapclick_west' ,$w);
         $REQ->param('p_mapstate', 'GOTPOINT');
      }
      # ...at this point, coords of 1st point in REQ.
   }

   # handle second map click
   elsif ($mapclick eq 'GOTRECT') {

	# Get coordinates of 2nd point (S-E):
	if (defined($REQ->param('msubmit.x'))) {

	    # Get coordinates from most-recent mouse click (x,y):
	    ($s, $e) = $mapfile->xy_to_latlon($REQ->param('msubmit.x'),
					      $REQ->param('msubmit.y'));
	    
	    # Re-enter coordinates as CGI tags: 
            $REQ->param('p_north',$REQ->param('mapclick_north'));
            $REQ->param('p_west',$REQ->param('mapclick_west'));
	    $REQ->param('p_south', $s); 
	    $REQ->param('p_east', $e);
            $REQ->param('p_mapstate', 'GOTRECT');
	}
	# ...at this point, coords of 2nd point in REQ.
   }
   1;
}

#------------------------------------------------------------
# draw
#------------------------------------------------------------
#
# By the time draw is called it is assumed that any map clicks
# have already been handled by calling _get_input.
# _get_input is responsible for turning x/y clicks into lat/lon points
# which are stored in QUERY->geo_region.
#
# The map cycles through three states: 'START', 'GOTPOINT' and 'GOTRECT'.
# Note that these states exist only in CGI, not with the session.
# If a user back buttons to a previous form the state from that form
# is used and the geo stored in the session is discarded, so what the use 
# sees is consistent with what happens.
# 
# START:
# This is the default state when there is no state, either because the form is  
# just being displayed or the geo type has been switched.  
#
# GOTPOINT:
# This is the state after the user has select one corner on the map
#
# GOTRECT:
# This is the state after the user has completed selecting a box on the map
# by clicking a second corner. 
#
# If user clicks the map again the state will cycle back to GOTPOINT: 

sub draw {
   my $self = shift;

   # Current state of map:
   my $mapstate = $REQ->param('p_mapstate') || 'START';

   my $directions;
   my $footnote;
   my $mapclick = 'START';
   my @coords;

   my $geo = $QUERY->geo_region;

   $geo->set_CGI("p_") if ($geo and ($geo->ODL_type eq 'RANGE_LOC'));


   # mapstate = START:
   # Show initial map state. If geo is undefined then map is blank,
   # or if there is a geo that is displayed. 
   if ($mapstate eq 'START') {          

      # The directions and the next map state:
      $directions = ("<FONT SIZE=5>Step 1: </FONT>".
		     "<B>Select a western corner</B> of your ".
		     "desired search rectangle:");

      my $geo = $QUERY->geo_region;

      if ($REQ->param('p_north')) { 

         my $dn = str2deg($REQ->param('p_north'));
         my $dw = str2deg($REQ->param('p_west'));
         my $ds = str2deg($REQ->param('p_south'));
         my $de = str2deg($REQ->param('p_east'));

         @coords = ('srect'=>"$dw,$dn,$de,$ds"); 
      }
      # What will be the next state if the map is clicked.
      $mapclick = 'GOTPOINT';
   }

   ## STATE = GOTPOINT: display map with one point...
   elsif ($mapstate eq 'GOTPOINT') {    

      # The directions and the next map state:
      $directions = ("<FONT SIZE=5>Step 2: </FONT>".
		     "Your first selected point is shown. ".
		     "<B>Now select an opposite eastern corner </B> ".
		     "of your search rectangle:");

      # Get coordinates, in decimal form:
      my $dn = str2deg($REQ->param('mapclick_north'));
      my $dw = str2deg($REQ->param('mapclick_west'));

      # Save coordinates of first click as hidden tags:
      $HTML -> hidden('mapclick_north' => $REQ->param('mapclick_north'));
      $HTML -> hidden('mapclick_west'  => $REQ->param('mapclick_west')); 

      # The footnote below the map:
      my $precision = 3; # The precision (significant number of digits).
      $footnote = ("<B>Selected point:</B> ".
         IU::Geo::Point->new($dn,$dw)->fixup($precision)->as_string);

      # Get coords to draw:
      @coords = ('spoint'=>"$dw,$dn");

      # What will be the next state if the map is clicked.
      $mapclick = 'GOTRECT';
   }

   ## STATE = GOTRECT: display map with two points...
   elsif ($mapstate eq 'GOTRECT') {

      # The directions and the next map state:
      $directions = ("<FONT SIZE=5>Step 3: </FONT>".
		     "This is your search rectangle. ". 
                     "Select a <B> new western  ".
		     "corner </B> on the map to start over:");

      # Get coordinates, in decimal form:
      my $dn = str2deg($REQ->param('p_north'));
      my $dw = str2deg($REQ->param('p_west'));
      my $ds = str2deg($REQ->param('p_south'));
      my $de = str2deg($REQ->param('p_east'));

      # Get coords to draw:
      @coords = ('srect'=>"$dw,$dn,$de,$ds");

      # What will be the next state if the map is clicked.
      # We have a box, so next click will cycle back to first point.
      $mapclick = 'GOTPOINT';
   }

   # Build URL for the map:
   my $mapfile = $self->{Mapfile};
   my $map_gif;
   $map_gif  = query_url($mapfile->script,
                         $mapfile->location => $mapfile->filename,
                         'grid' => 'full',
                         @coords);

   # Output map directions for STEP 1, STEP 2, or STEP 3
   print $directions if ($directions && $self->show_directions);

   # Output click-able-map:
   $HTML -> CENTER;
   $HTML -> msubmit_button(SRC=>$map_gif,
                           ALT=>$mapfile->name,
		QUERY=>['mode'    => $self->{ParentMode},
                        'Geoevent' => 'CLICK-ON-MAP',
                        'p_mapclick' => $mapclick,
                        'submode' => 'UPDATE',
                       ],
                HEIGHT=>304, WIDTH=>412);

   # Output footnote (if any):
   $HTML->BR;
   print $footnote if ($footnote);
   $HTML->_CENTER;

   # Output hidden info which maintains state of current map:
   $HTML -> hidden('p_mapstate' => $mapstate);
   $HTML -> hidden('p_feature' => $REQ->param('p_feature'));
   $HTML -> hidden('p_north' => $REQ->param('p_north'));
   $HTML -> hidden('p_south' => $REQ->param('p_south'));
   $HTML -> hidden('p_east'  => $REQ->param('p_east'));
   $HTML -> hidden('p_west'  => $REQ->param('p_west'));
   $HTML -> hidden('p.browse_location' => $REQ->param('p.browse_location'))
      if $REQ->param('p.browse_location');
   1;
}

#------------------------------------------------------------
# summary
#------------------------------------------------------------
sub summary {

   my $self = shift;

   my $geo= $QUERY->geo_region;

   print join("<BR>",split /\n/, html_escape($geo->as_string(Verbose=>2)))
      if (defined($geo));
1;
}
#------------------------------------------------------------
# show_directions
#------------------------------------------------------------
# Override in subclass if you don't want map directions.

sub show_directions { return 1 }

#------------------------------------------------------------
# is_supported STRING
#------------------------------------------------------------
# class method
# Return true is this class supported the passed in geotype.

sub is_supported {
   my ($class,$geotype) = @_;
   return 1 if ($geotype eq 'Equatorial Map');
   0;
}

#------------------------------------------------------------
# is_compatable GEO
#------------------------------------------------------------
# class method
# Return geotype if this class supports the passed in geo object.

sub is_compatable {
   my ($class,$geo) = @_;

   my $ODL_type = $geo ? $geo->ODL_type : undef;
   return 'Equatorial Map' if ($ODL_type eq 'RANGE_LOC');
   undef;
}

#------------------------------------------------------------
1;
