package IW::GranList;


=head1 NAME

IW::GranList - utilities for outputting a granule list


=head1 DESCRIPTION

Several modes currently output a granule list.  Here's the common code.

See IW::ObjectList for details.


=head1 HIDDEN FIELDS

Hidden fields are embedded in the output (which of course must be
enclosed in a form, since it contains checkboxes!). 
These fields actually originate from the last request; they are:

     dskey     One for each dskey in the last request
     pkkey     One for each pkkey in the last request
     gkey!     Just generated once: indicates presence of granule checkboxes

In addition, the checkbox state is sensitive to the following param:

     select    If set, then checkbox state is determined by whether the
               value is "ALL" or "NONE".

=head1 SORTING

A request to sort the granules uses the most "appropriate" granule sort.
As a side effect, it sets the LastGranuleSort in the session state.

=head1 RCS INFO

$Id: GranList.pm,v 1.92 2003/05/21 20:27:30 solomon Exp $

=cut

use IU;
use IU::Log;
use IU::Granule;
use IU::PkgCombo;

use IW::User;
use IW::Date qw(TZ_to_datetime);
use IW::ObjectList;
use IW::Util qw(:url);
use IW::App;
use IW::Sort;
use IW::Babel;
use IU::WWWValids;  
use IW::Disclaimers;
use IW::Widget::URLTable;

@ISA = qw(IW::ObjectList);

use strict;

#------------------------------------------------------------
#
# GLOBALS
#
#------------------------------------------------------------

# Order to present things:
my @Order;

# Translating a granule's 'carted' attribute to a status message:
my %CartStatusMsg = (
    'CAN'    => 'Can be ordered',
    'CANNOT' => "Can't be ordered here",
    'YES'    => 'In shopping cart',
    'JUST'   => 'Just added to cart'
);

# Translating a granule's 'carted' attribute to a status message:
my %CartStatusIcon = (
    'CAN'    => 'icons/order-can.gif',
    'CANNOT' => 'icons/order-cannot.gif',
    'YES'    => 'icons/order-yes.gif',
    'JUST'   => 'icons/order-just.gif'
);

#------------------------------------------------------------
# new PARAMHASH
#------------------------------------------------------------
# Create a new granule-list object.
# PARAMHASH contains everything normal to superclass, plus:
#
#     Origin   => origin of these objects:  'RESULTS'.

sub new {
    my $class = shift;
    my %params = @_;

    # Construct:
    my $self = $class->SUPER::new(@_);    

    $self->{Type} = 'gkey';

    my @Summary = @{$self->{Display}->granule_summary};
    # hard code Granule ID as first column.
    @Order = ('GRANULE_ID');
    push @Order,@Summary; 

    # Set up defaults, etc:
    $self->{Origin}   or $self->{Origin} = '';
    $self->{NoteFunc} or $self->{NoteFunc} = \&note;
    $self->{Note2Func} or $self->{Note2Func} = \&note2;
    $self->{GraphFunc} or $self->{GraphFunc} = \&graph;
    $self->{Graph2Func} or $self->{Graph2Func} = \&graph2;
    $self;
}

#------------------------------------------------------------
# origin
#------------------------------------------------------------
# Get/set the origin of these granules.

sub origin {
    my ($self, $value) = @_;
    $value and $self->{Origin} = $value;
    $self->{Origin};
}

#------------------------------------------------------------
# sort [SORTREF]
#------------------------------------------------------------
# Sort our list of granules.  Returns a reference to the newly-sorted array.
#
# B<SIDE EFFECTS:> This sets the session-state variable 'LastGranuleSort' 
# to the sort option used, so that it is used/presented as a default for 
# subsequent granule lists.

sub sort {
    my ($self, $sortref) = @_;

    # Figure out the sort algorithm...
    my $sorter = IW::Sort::Granule->appropriate;
    
    # Construct a nice string to describe this:
    $$sortref = $sorter->as_string      if $sortref;
    
    # Do it!    
    my @sorted = $sorter->sort($self->{Elems});
    $self->{Elems} = \@sorted;

    # Make a note in the session state:
    $SESSION->state->{LastGranuleSort} = $sorter;
    $self->{Elems};
}

#------------------------------------------------------------
#
# PRIVATE METHODS/UTILITIES
#
#------------------------------------------------------------


#------------------------------------------------------------
# sorted_list
#------------------------------------------------------------
# Private function.  Take a ref to an array of strings, and
# return a single scalar consisting of the elements sorted and joined
# by commas.

sub sorted_list {
    my $unsorted = shift;
    join ', ', (sort @$unsorted );
}

#------------------------------------------------------------
# check_for_some_txt 
#------------------------------------------------------------
# returns HASH

sub check_for_some_txt {

   my $self = shift;
   my %some;

   # if user wants to see empty columns then start with an initial value
   # of '1' (true) for the value of the %some hash.
   my $show_empty = ($self->{Display}->show_empty_columns_gran eq 'Y') 
       ? '1' : '0';
   %some = map {$_=>$show_empty} @Order;

   # Determine the width to make each column.
   # Loop over every row (Elem), updating values in %some.

   my $num = int(@{$self->{Elems}});
   my $i=0;
   for ($i=0; $i < $num; $i++ ) {
      $self->print_data_txt(\%some,$i,'1');
   }

   # Update column width using column headers as they maybe wider than 
   # the data.
   $self->print_head_txt(\%some,'1');

   return %some;

}

#------------------------------------------------------------
# check_for_some
#------------------------------------------------------------
# returns HASH

sub check_for_some {

   my $self = shift;

   # If user wants to see empty columns then start with an initial value
   # of '1' (true) for the value of the %some hash.
   my $show_empty = ($self->{Display}->show_empty_columns_gran eq 'Y')
       ? '1' : '0';
   my %some = map {$_=>$show_empty} @Order;

   # Hard code these as always being displayed, regardless of whether 
   # or not there are values in the column.
   $some{'DATASET_ID'} = '1'; 
   $some{'DATA_CENTER_ID'} = '1';
   $some{'CAMPAIGN'} = '1'; 
   $some{'PARAMETER'} = '1';
   $some{'PROCESSING_LEVEL'} = '1';
   $some{'SENSOR_NAMES'} = '1';
   $some{'SOURCE_NAMES'} = '1';

   my $title;
   my $item;
   foreach $item (@{$self->{Elems}}) {
      # As data is found for a column it is removed from the loop,
      # gradually the loop becomes shorter and shorter.
      # If the first item has values for all the columns
      # then this will run only once. 
      foreach $title (grep {!$some{$_}} (keys %some) ) {
         if ($title eq 'PATH' and defined($item->coverage) and
            scalar (@{$item->coverage->rows})) {
            $some{$title} = '1';
         }
         elsif ($title eq 'ROW' and defined($item->coverage) and
                scalar (@{$item->coverage->rows})) {
            $some{$title} = '1';
         }
         elsif ($title eq 'QA_FLAG') {  

            $some{$title} = '1' if (
               (defined ($item->automatic_quality_flag) &&
                 (int(@{$item->automatic_quality_flag})) > 0) ||
               (defined ($item->operational_quality_flag) &&
                 (int(@{$item->operational_quality_flag})) > 0) ||
               (defined ($item->science_quality_flag) &&
                 (int(@{$item->science_quality_flag})) > 0) 
            ) ;
         }
         elsif ($title eq 'START_DATE') {

            $some{$title} = '1' if (defined ($item->start_date)
                                    and $item->start_date ne '(none)');
         }
         elsif ($title eq 'STOP_DATE') {

            $some{$title} = '1' if (defined ($item->stop_date) 
                                    and $item->stop_date ne '(none)');
         }
         elsif ($title eq 'GEO_CENTER') {

            $some{$title} = '1' if (defined ($item->area_center));
         }
         elsif ($title eq 'GEO_COVERAGE') {

            $some{$title} = '1' if (defined ($item->coverage->area));
         }
         elsif ($title eq 'QA_SCENE_CCA') {
            $some{$title} = '1' if (defined ($item->cloud_cover));
         }
         elsif ($title eq 'MISC_URL') {
            $some{$title} = '1'
               if (defined($item->misc_URLs) && (int(@{$item->misc_URLs})> 0));
         }
         else { # Not a standard attribute, try specialized results.

            my $result;
            foreach $result(@{$item->results}) {
               $some{$title} = '1' if (uc($result->resultName) eq uc($title));
            } # foreach result.

            # While we're at it,look for the specialized results.
	    # inside the Result Group too. Have any? Print a title for it.

            foreach $result(@{$item->resultGroup}) {
               $some{$title} = '1' if ((int(@{$result->findSpecializedResults("\^$title\$") }) > 0) ||
                    ($self->{Display}->show_empty_columns_gran eq 'Y'));
            } # foreach result Group.
         } # elseif
      } # foreach $title.
   } # foreach $item. 

   $some{'GRANULE_ID'} = '0';
   $some{'LocalGranuleID'} = '0';

   return %some;
}

#------------------------------------------------------------
# print_data
#------------------------------------------------------------

sub print_data {
   my $self = shift;
   my $some = shift;
   my $i = shift;

   my $title;
   my $item = $self->{Elems}[$i];

   # Only do those columns which have some=1;
   foreach $title (grep {$some->{$_}} (@Order) ) {

      if ($title eq 'PATH') {

         print '<td>', join ('-',@{$item->coverage->paths}) || "\xA0", '</td>';
      }
      elsif ($title eq 'ROW') {
  
         my @rows = @{$item->coverage->rows};
         my $string = ($rows[0] eq $rows[1]) ? $rows[0] : join ('-',@rows); 
         print '<td>', $string || "\xA0", '</td>';
      }
      elsif ($title eq 'QA_FLAG') {
         $HTML->TD;

         # quality_assurance_status will return one of the following :
         # 99 - No Quality Flags, so don't show an image.
         # 1  - Quality Flags Passed
         # 0 -  Quality Flags Failed
         #
         my $quality_flag_status = $item->quality_assurance_status;
         if ( $quality_flag_status != 99 ) {
	    my $href = $self->_attrs_url($item) . '#ADD_ATTR_INFO';
            if ($quality_flag_status) {
                $HTML -> hypertext("Pass", $href);
            }
            else {
                $HTML -> hypertext("Fail", $href);
            }
          } else {
             print "\xA0";
          }
         $HTML->_TD;
      }
      elsif ($title eq 'DATASET_ID') {

         print '<td>', $item->dataset_id || "\xA0", '</td>';
      }
      elsif ($title eq 'DATA_CENTER_ID') {

         print '<td>', $item->archive_id || "\xA0", '</td>';
      }
      elsif ($title eq 'CAMPAIGN') {

         print '<td>', $item->campaign || "\xA0", '</td>';
      }
      elsif ($title eq 'PARAMETER') {

         print '<td>', sorted_list($item->parameters) || "\xA0", '</td>';
      }
      elsif ($title eq 'PROCESSING_LEVEL') {

         print '<td>', $item->proc_level || "\xA0", '</td>';
      }
      elsif ($title eq 'SENSOR_NAMES') {

         print '<td>', sorted_list($item->sensors) || "\xA0", '</td>';
      }
      elsif ($title eq 'SOURCE_NAMES') {

         print '<td>', sorted_list($item->sources) || "\xA0", '</td>';
      }
      elsif ($title eq 'START_DATE') {

         print '<td>', TZ_to_datetime($item->start_date) || "\xA0", '</td>';
      }
      elsif ($title eq 'STOP_DATE') {

         print '<td>', TZ_to_datetime($item->stop_date) || "\xA0", '</td>';
      }
      elsif ($title eq 'GEO_CENTER') {
         print '<td>',$item->area_center_HTML || "\xA0" ,'</td>';
      }      
      elsif ($title eq 'GEO_COVERAGE') {
         print '<td>',$item->coverage->area_as_HTML || "\xA0",'</td>';
      }
      elsif ($title eq 'SIZEMBECSDATAGRANULE') {
         my $item_size = (0+$item->size_mbecs_granule)  ?
                              $item->size_mbecs_granule : "";
         print '<td>', $item_size || "\xA0", '</td>';
      }
      elsif ($title eq 'MISC_URL') {
         my $num_misc_urls = ($item->misc_URLs ? int(@{$item->misc_URLs}) : 0);
         if (defined($item->misc_URLs) && $num_misc_urls > 0) {
            my ($i, $furl, $url, $comment);
            print '<td>';
            for ($i=0; $i<$num_misc_urls; $i++) {
               $furl = $item->misc_URLs->[$i];
               $url = logged_url($furl->URL, 'type'=>$furl->type);
               $comment = $furl->comment;
               print "<a href=$url target=ext-url-window>$comment</a>", '<br>';
            }
            print '</td>'; 
         } else {
            print '<td>', "\xA0",'</td>';
         }
      }
      elsif ($title eq 'QA_SCENE_CCA') {
	 if(defined($item->cloud_cover)) {
            print '<td>', $item->cloud_cover, '</td>';
	 } else {
            print '<td>', "\xA0", '</td>';
	 }
      }
      else {
        # Ok, now try to dynamically display the ones that reside inside
        # Specialized Results and Result Group

         my $result;
         my $found = 0;
         my $foundmore = 0;
         my @printed_list;


         # First floor, Specialized Results
         foreach $result(@{$item->results}) {
            if (uc($result->resultName) eq uc($title)) {
               push  @printed_list , sorted_list($result->resultValue); 
               $found++;
            } # endif resultName.
         } # end foreach result.
         my $temp_variable;
 
        # Second floor and higher, Specialized Results can reside in multiple layers of
        # of Results Groups and then come in different types. 
         foreach $result(@{$item->resultGroup}) {
            if ( $temp_variable =  $result->findSpecializedResults("\^$title\$",) ) {
                # temp_variable can be :
                   # a scalar : actual value 
                   # A Specialized Result object : use resultValueAsString to
                   #      display real value.
                   #  An Array of Specialized Result :
                   #            (which can also be a NULL array too) 
                if ( ref($temp_variable) eq 'IU::SpecializedResults' ) {
                      push  @printed_list , sorted_list($result->resultValue);
                }
                elsif ( ref($temp_variable) eq 'ARRAY') {
                  if (defined($temp_variable->[0])) {
                     my $a_tmp;
	             foreach  $a_tmp ( @{$temp_variable} ) {
                           push  @printed_list , $a_tmp->resultValueAsString if 
 (defined($a_tmp->resultValueAsString));
		     } 
                  }
                }
                else {
                    push  @printed_list , sorted_list($temp_variable);
                }
               $foundmore++;

            } # endif resultGroup.
         } # end foreach result.
         print '<td>', join(',',@printed_list ),"\xA0 </td>";
      } # end elsif 
   }
}

#------------------------------------------------------------
# print_data_txt HASH, INT
#------------------------------------------------------------

sub print_data_txt {

   my $self = shift;
   my $some = shift;
   my $i = shift;
   my $set_width_only = shift || 0;

   my $width = 1;
   my $title;
   my $item = $self->{Elems}[$i];

   my $field = '';
   # Only do those columns which have some>0;
   my @NewOrder = ($set_width_only)? @Order :  (grep {$some->{$_}} (@Order));
   foreach $title (@NewOrder) {

      $field = '';

      if ($title eq 'GRANULE_ID') {
         $field = $item->id;
      } 
      elsif ($title eq 'PATH') {
         $field = join ('-',@{$item->coverage->paths});
      }
      elsif ($title eq 'ROW') {
         my @rows = @{$item->coverage->rows};
         $field = ($rows[0] eq $rows[1]) ? $rows[0] : join ('-',@rows);
      }
      elsif ($title eq 'LocalGranuleID') {
         $field = $item->local_granule_ID;
      }
      elsif ($title eq 'QA_FLAG') {
         my $quality_flag_status = $item->quality_assurance_status;
         ($field = ( $quality_flag_status ? 'Quality Flag OK ' :  'Quality Flag FAILED  ')) if ( $quality_flag_status != 99);
      }
      elsif ($title eq 'DATASET_ID') {
         $field = $item->dataset_id;
      }
      elsif ($title eq 'DATA_CENTER_ID') {
         $field = $item->archive_id;
      }
      elsif ($title eq 'CAMPAIGN') {
         $field = $item->campaign;
      }
      elsif ($title eq 'PARAMETER') {
         $field = sorted_list($item->parameters);
      }
      elsif ($title eq 'PROCESSING_LEVEL') {
         $field = $item->proc_level;
      }
      elsif ($title eq 'SENSOR_NAMES') {
         $field = sorted_list($item->sensors);
      }
      elsif ($title eq 'SOURCE_NAMES') {
         $field = sorted_list($item->sources);
      }
      elsif ($title eq 'START_DATE') {

         $field = TZ_to_datetime($item->start_date);
      }
      elsif ($title eq 'STOP_DATE') {

         $field = TZ_to_datetime($item->stop_date);
      }
      elsif ($title eq 'GEO_CENTER') {
         $field = $item->coverage->center_pretty('1');
      }      
      elsif ($title eq 'GEO_COVERAGE') {
         $field = $item->coverage->area_pretty('1');
      }
      elsif ($title eq 'MISC_URL') {
         my $num_misc_urls = ($item->misc_URLs ? int(@{$item->misc_URLs}) : 0);
         if (defined($item->misc_URLs) && $num_misc_urls > 0) {
            my ($i, $furl, @list);
            for ($i=0; $i<$num_misc_urls; $i++) {
               $furl = $item->misc_URLs->[$i];
               push (@list, $furl->URL);
            }
            $field = join (' , ', @list);
         }
      }
      else {
         # Ok, now try to dynamically display the ones that reside inside
         # Specialized Results and Result Group

         my $result;
         my $found = 0;
         my $foundmore = 0;
         my @printed_list;


         # First floor, Specialized Results
         foreach $result(@{$item->results}) {
            if (uc($result->resultName) eq uc($title) ) {
               push  @printed_list , sorted_list($result->resultValue); 
               $found++;
            } # endif resultName.
         } # end foreach result.
         my $temp_variable;
                
         # Second floor +, Specialized Results can reside in multiple layers of
         # of Results Groups and then come in different types. 
         foreach $result(@{$item->resultGroup}) {
            if ( $temp_variable =  
               $result->findSpecializedResults("\^$title\$") ) {

               # temp_variable can be :
               # a scalar : actual value 
               # A Specialized Result object : use resultValueAsString to
               #      display real value.
               #  An Array of Specialized Result (  which can also be null too) 
               if ( ref($temp_variable) eq 'IU::SpecializedResults' ) {
                  push  @printed_list, sorted_list($temp_variable->resultValue);
               }
               elsif ( ref($temp_variable) eq 'ARRAY') {
                  if (defined($temp_variable->[0])) {
                     my ($a_tmp);
                     foreach  $a_tmp ( @{$temp_variable} ) {
                       (push  @printed_list , $a_tmp->resultValueAsString) 
                           if (defined($a_tmp->resultValueAsString));
                     }
                  }
               }
               else {
                  push  @printed_list, sorted_list($temp_variable);
               }
               $foundmore++;
            } # endif resultGroup.
         } # end foreach result.

         if ( $foundmore || $found ) {
             $field = join(",",@printed_list);
         }
         else {  
             # If we don't have any values for the table, fill it with a blank box.
             $field = '' ; }
      } # end elsif 
  
      if ($set_width_only) { 
         $some->{$title} = length($field) if (length($field) > $some->{$title});
      } else {
         $width = $some->{$title};
         printf "%-${width}s",$field;
         print '|';
      }
   } # end foreach 
}

#------------------------------------------------------------
# print_head_txt
#------------------------------------------------------------
# Output the column headers for the 1 line per item table.

sub print_head_txt {
   my $self = shift; 
   my $some = shift;
   my $set_width_only = shift || 0;

   $WWWVALIDS = IU::WWWValids->new() if (!defined($WWWVALIDS));
   my %pretty = $WWWVALIDS->granListPretty;
   $pretty{'GRANULE_ID'} = 'Granule'; 

   my $width = 0;
   my $title;
   my $pretty_title;
   foreach $title (@Order) { # order of presentation!
      $pretty_title = ($pretty{$title}) ? $pretty{$title} : $title;
      if ($some->{$title} ) {
       # Just set the column width
       if ($set_width_only) {
         $some->{$title} = length($pretty_title) 
            if (length($pretty_title) > $some->{$title});
       }
       # Print column headers
       else { 
         $width = $some->{$title};
         printf "%-${width}s",$pretty_title;
         print '|';
       }
      }
   }
   print "\n" unless $set_width_only;
   foreach $title (@Order) {
      $width = $some->{$title};
      if ($width && !$set_width_only) {
         my $i;
         for ($i=0; $i < $width; $i++) { 
            print '-'; 
         }
         print '|';
      }
   }
}

#------------------------------------------------------------
# _xxx_url GRAN -- URLs for menu buttons
#------------------------------------------------------------

# Return the URL for the Attributes button:
sub _attrs_url {
    my ($self, $gran) = @_;
    script_url('mode'=>'GRANATTRS', 
	       'gkey'=>$gran->key,
	       'origin'=>($self->origin || 'RESULTS'));
}

# Return the URL for the Granule Pricing button:
sub _price_url {
   my ($self, $gran) = @_;
   return undef if (!($gran->possible_to_order));
   $self->_attrs_url($gran) . '#PRICE_INFO';
}


# Return the URL for the Browse link:
sub _browse_url {
    my ($self, $gran) = @_;
    my $num_browse_urls = ($gran->has_browse ? int(@{$gran->browse_URLs}) : 0);
    return undef if(!$num_browse_urls);
    if($num_browse_urls == 1) {
       my $tbl = new IW::Widget::URLTable "Browse" => $gran->browse_URLs;
       $tbl->num_URLs or return undef;
       $tbl->print_short;
    } else {
       return script_url('mode'    => 'MULTIBROWSE',
                         'submode' => 'SHOW',
                         'gkey'    => $gran->key,
                         'type' => 'browse',
                         'origin'=>($self->origin || 'RESULTS'));
    }
}
# Return the URL for the View Browse button:
sub _view_url {
    my ($self, $gran) = @_;
    my $href = undef;

    # Determine what kinds of browse this granule has:
    my $has_int_browse  = ($gran->browse_type =~ /^(Y|INTEGRATED)$/i);
    my $num_browse_urls = ($gran->has_browse ? int(@{$gran->browse_URLs}) : 0);

    # If granule has more than one browse source, output multibrowse page.
    # Else, switch to the desired page:
    if (($has_int_browse and $num_browse_urls) or     # VB + >0 BrowseURLs, or
	($num_browse_urls > 1)) {                     # >1 BrowseURLs
	$href = script_url('mode'    => 'MULTIBROWSE',
			   'submode' => 'SHOW',
			   'gkey'    => $gran->key,
			   'origin'=>($self->origin || 'RESULTS'));
    }
    elsif ($has_int_browse) {          # just int browse
	$href = script_url('mode' => 'VIEWBROWSE',
                           'gkey' => $gran->key,
			   'origin'=>($self->origin || 'RESULTS'));
    }
    elsif ($num_browse_urls == 1) {    # exactly one browse url
	my $furl = $gran->browse_URLs->[0];
	$href = logged_url($furl->URL, 'type'=>$furl->type);
    }
    return $href;
}

# Return the URL for the FTP Browse button:
sub _ftp_url {
    my ($self, $gran) = @_;
###    return undef if ($self->origin ne 'RESULTS');      # for now
    if ($gran->browse_type =~ /^(Y|FTP|FTP_ONLY)$/i) {
       return script_url('mode'   =>'FTPBROWSE', 
			 'gkey'   =>$gran->key,
			 'origin'=>($self->origin || 'RESULTS'));
    }
    undef;                # always, for now.
}



# Return the URL for the Direct Access Links:
sub _access_url {
    my ($self, $gran, $type) = @_;
    return undef if(!$gran->has_data && $type eq 'data');
    return undef if(!$gran->has_metadata && $type eq 'metadata');
    my $num_data_urls = int(@{$gran->data_URLs});
    my $num_metadata_urls = int(@{$gran->metadata_URLs});
    my $tbl;
    if($num_data_urls == 1 && $type eq 'data') {
       $tbl = new IW::Widget::URLTable "Data" => $gran->data_URLs;
       $tbl->num_URLs or return undef;
       $tbl->print_short;
    } elsif ($num_metadata_urls == 1 && $type eq 'metadata') {
       $tbl = new IW::Widget::URLTable "Metadata" => $gran->metadata_URLs;
       $tbl->num_URLs or return undef;
       $tbl->print_short;
    } else {
       return script_url('mode'   =>'LINKS', 
                         'gkey'   =>$gran->key,
                         'type' => $type,
                         'origin'=>($self->origin || 'RESULTS'));
    }
}



#------------------------------------------------------------
#
# PRINT...
#
#------------------------------------------------------------

#------------------------------------------------------------
# _checkbox GRAN
#------------------------------------------------------------
sub _checkbox {
    my $gran = shift;

    # The value for the selection checkbox:
    my @checkval = ();
    if ($REQ->param('select')) {       # explicit value...
	@checkval = (Value => ($REQ->param('select') eq 'ALL'));
    }
    $HTML -> checkbox(Name=>"gkey", Choice=>$gran->key, @checkval);
}

#------------------------------------------------------------
# print_title I
#------------------------------------------------------------
# Output the title for the i'th (0-based) granule.

sub print_title {
    my ($self, $gran_num) = @_;
    my $gran = $self->{Elems}[$gran_num];
    my @colors =  split ' ', $gran->graphed;

	print '<td nowrap>';
	if ($gran->graphed) {$HTML -> color($colors[0],$colors[1],$colors[2]);}
	print $gran->id;
	$self->print_graph($gran_num);
	if ($gran->graphed) {$HTML -> _color;}
	print '</td>';
}


#------------------------------------------------------------
# print_buttons I
#------------------------------------------------------------
# Output the menu buttons for the i'th (0-based) granule.
# Only used when displaying one line per item table listing.

sub print_buttons {
    # Some basic info:
    my ($self, $gran_number, $some_notes, $some_pkgs_chosen, $some_orderable) =     @_;
    # Get the individual element.
    my $item = $self->{Elems}[$gran_number];
    my $gkey        = $item->key;
    my $browse_type = $item->browse_type || '';
    my $button_prefix = 'item';         # For name of button file

	# "Selection" checkbox... notice, we check for "select all"!
        # Color the background of this table cell to match the map graph
        if ($item->graphed eq 'red') {
            print "<td bgcolor=red align=center>";
        }
        elsif ($item->graphed) {
            my @color =  split ' ', $item->graphed;
            my $color = sprintf("#%02X%02X%02X", @color);
            print "<td bgcolor=$color align=center>";
        } else {
           print '<td align=center>';
        }

	_checkbox($item);
	print '</td>';
	# Print granule ids and menu of other options buttons.
        $HTML->TD->CENTER; 
        # Print granule ID:
        my $id = $item->id || "\xA0";
        $HTML->t_B($id)->BR;
        # Print local ID if present:
        my $local = $item->local_granule_ID;
        print "($local)" if (defined($local) and ($local ne ''));
        $HTML->_CENTER;
        $HTML->_TD->TD;
        $HTML->CENTER;
        # "Attributes" button...
        $HTML -> hypertext("Attributes",
           $self->_attrs_url($item)) if ($self->_attrs_url($item)); 
        $HTML ->BR; 
        $HTML -> hypertext("Pricing",
           $self->_price_url($item)) if ($self->_price_url($item)); 
        $HTML ->_CENTER;
	print '</td><td>';
        $HTML -> t("Access Unavailable") if (!$item->has_data && 
	  !$item->has_metadata && !$item->has_browse);
        if ($self->_access_url($item,'data')) {
           $HTML -> hypertext("Data", $self->_access_url($item,'data'));
           $HTML->BR;
	}
        if ($self->_access_url($item,'metadata')) {
           $HTML -> hypertext("Metadata", $self->_access_url($item,'metadata'));
           $HTML->BR;
	}
        $HTML -> hypertext("Browse", $self->_browse_url($item)) 
         if ($self->_browse_url($item));
        print '</td><td>';
	# "View Browse" Link...
        $HTML -> hypertext("Image",
                          $self->_view_url($item)) if ($self->_view_url($item));
        $HTML -> t("Image Unavailable") if (!$self->_view_url($item));
	print '</td><td>';
	# "FTP Browse" Link...
        $HTML -> hypertext("Sample",
                          $self->_ftp_url($item)) if ($self->_ftp_url($item));
        $HTML -> t("Sample Unavailable") if (!$self->_ftp_url($item));
	print '</td>';

    # Print any notes (in cart, on workbench, package choice, etc):
    if ($some_notes) {
	print '<td>';
	$self->print_note($gran_number);
	print '</td><td>';
	$self->print_note2($gran_number);
	print '</td>';
    }
    1;
}

#------------------------------------------------------------
# print_head
#------------------------------------------------------------
# Output the column headers for the 1 line per item table.

sub print_head {
   my ($self, $some_notes, $some_pkgs_chosen, $some_orderable,
       $some_comments, $some) = @_;

   # Print the column headers for the table.
   print '<tr bgcolor="#CCCCCC">';
   print '<td>';
   print '<b>Select </b>';
   # Select all/none:
   $HTML ->BR;
   $self->print_all_none;
   print '</td>';
   print '<td colspan=1><b>Data Granule ID</b><BR>(Local Granule ID)</td>';
   print '<td colspan=1><b>Granule Information</b></td>';
   print '<td colspan=1><b>On-line Access</b></td>';
   print '<td colspan=1><b>Image Quicklook</b></td>';
   print '<td colspan=1><b>Request Sample</b></td>';
   print '<td colspan=2><b>Notes</b></td>' if ($some_notes);

   $WWWVALIDS = IU::WWWValids->new() if (!defined($WWWVALIDS));
   my %pretty = $WWWVALIDS->granListPretty;
   $pretty{'GRANULE_ID'} = 'Granule';
   my $title;
   my $pretty_title;
   foreach $title (@Order) {                 # order of presentation!
      $pretty_title = ($pretty{$title}) ? $pretty{$title} : $title;   
      print '<td nowrap><b>', $pretty_title, '</b></td>' 
         if ($some->{$title});
   }
   1;
}


#------------------------------------------------------------
# pass_params
#------------------------------------------------------------
# Pass dataset/granule keys from the last request as hidden fields.
# This is invoked automatically on output!

#sub pass_params {
#    my $self = shift;
#
#    # This listing came from selected datasets or packages: we need to 
#    # preserve that information if sorting!
#    my ($dskey, $pkg);
#    foreach $dskey ($REQ->param('dskey')) {
#	$HTML->hidden('dskey', $dskey);
#    }
#    foreach $pkg ($REQ->param('pkkey')) {
#	$HTML->hidden('pkkey', $pkg);
#    }
#    $HTML->hidden('gkey!', 1);  # granule selection boxes were present 
#
#    # Complete the form:
#    $HTML->hidden('sid', $SESSION->id);
#}

#------------------------------------------------------------
#
# Utilities
#
#------------------------------------------------------------

#------------------------------------------------------------
# note GRANULE
#------------------------------------------------------------
# The default "note" function for granules.

sub note {
    my $gran = shift;
    if ($gran->carted) {
	$HTML->IMG(SRC=>doc_url('icons/carted.gif'),
                   ALT=>'In Shopping Cart');
    } elsif (!($gran->possible_to_order)) {
        my ($go_submode, $title);
        if (defined($gran->misc_URLs) && int(@{$gran->misc_URLs}) > 0) {
           $go_submode = "MISCURL_ORDER";
           $title   = "To Order";
        } else {
           $go_submode = "CANT_ORDER";
           $title   = "Cannot Order";
        }
	my $href = script_url('mode'      => 'GRANLIST',
	         	      'submode'   => $go_submode,
			      'daac'      => $gran->archive_id,
			      'granule'   => $gran->id,
			      'dataset'   => $gran->dataset_id);
	$HTML -> A(HREF=>$href)
	      -> small
	      -> t("$title")
	      -> _small
	      -> _A;

    } else {

       print "\xA0";
    }
}

#------------------------------------------------------------
# note2 GRANULE
#------------------------------------------------------------
# The default "note2" function for granules.

sub note2 {
    my $gran = shift;
    if ($gran->workbenched) {
        $HTML->IMG(SRC=>doc_url('icons/workbenched.gif'),
                   ALT=>'In Folder');
    } else {
       print "\xA0";
    }
}

#------------------------------------------------------------
# graph GRANULE
#------------------------------------------------------------
# The default "graph" function for granules.

sub graph {
   my $gran  = shift; 
   my $color = $gran->graphed;
#   if ($gran->graphed) {
#             $HTML->IMG(SRC=>doc_url("icons/$color.gif"), ALT=>'Graphed');
#    }
}

#------------------------------------------------------------
# any_notes
#------------------------------------------------------------
# Return true if there are any notes in the whole table of all granules.
# Return false if there are no notes in the whole table.

sub any_notes {
    my $self = shift;
    my $num = shift;

#    if (($self->{Origin}) eq 'CART') {
#	return (0);   # No longer show notes in the cart.
#    }
    my $gran_number;
    my $gran;

    for ($gran_number = 0; $gran_number < $num; $gran_number++) {
	# Get the individual element.
	$gran = $self->{Elems}[$gran_number];
	# If there is a cart or workbench note, return true.
	if ($gran->carted or $gran->workbenched or !$gran->possible_to_order) {
	    return (1);
	}
    }
    return (0);
}

#------------------------------------------------------------
# any_orderable
#------------------------------------------------------------
# Return false for all grans.  This method is only used for data sets.

sub any_orderable {
    0;
}

#------------------------------------------------------------
#
# Backwards compatibility:
#
#------------------------------------------------------------
# IF YOU USE THESE, STOP!!!!  CHANGE YOUR CODE!!!!

#------------------------------------------------------------
# sort_granules GRANULES,SORTREF
#------------------------------------------------------------
sub sort_granules {
    my ($granules,$sortref) = @_;
    my $new = IW::GranList->new(Elems=>$granules);
    $new->sort($sortref);
    @{$new->elems};
}

#------------------------------------------------------------
# print_list GRANULES,OPTIONS
#------------------------------------------------------------
# Print a standard granule listing.
#
# GRANULES is a reference to an array of granule, which should already 
# be sorted.
#
# OPTIONS is a hash of options:
#
#     Type => whose granules we are: DATASET, CART or PACKAGE
#     Icon => call this subroutine to generate a LHS icon   

sub print_list {
    my ($granules, $options) = @_;
    my $origin   = $options->{Type} || undef;
    my $notefunc = $options->{Icon} || undef;

    my $self = IW::GranList->new(Elems=>$granules,
				 Origin=>(($origin eq 'DATASET') 
					  ? 'RESULTS' : 'CART'),
				 NoteFunc=>$notefunc,
                                 Note2Func=>'print_note2',
                                 GraphFunc=>'print_graph') ;
    $self->print;
}

#------------------------------------------------------------
1;
