package IW::HTML;

=head1 NAME

IW::HTML - HTML stream subclass, for EOSDIS VO Web Gateway applications


=head1 DESCRIPTION

Identical to a base HTML::Stream, but adds a few methods
useful for EOSDIS VO Web Gateway CGI scripts.


=head1 USING FORM WIDGETS

Some of the methods in this class are documented as creating 
form "widgets".  These are more than just ordinary form elements!

When you create a form widget, there are a few values you may need
to concern yourself with:

=over

=item Name

The field name.

=item Choice(s)

This is the I<complete> set of possible values.
Usually just for single- or multi-selection lists, and for radio buttons.

=item Value(s)

Explicitly sets the value(s) for the widget.
This could be the contents of a text field, or the selected items
in a multiple-selection list.  For backwards-compatibility,
it may also be called "Current", but don't use this.

I<This is the first value we try when determining what to put in the widget:>
if it is undefined/empty, then we look for...

=item I<Existing>

Whatever values under this "Name" were passed in to this invocation of 
the CGI script.  This is not a parameter: it is determined by querying the
CGI interface.  If it is undefined/empty, then we look for...

=item Default(s)

Whatever value(s) we should initialize ourselves to.

=back

By default, widgets are "sticky": a sticky widget is simply one where 
the "Existing" values are consulted.  A "slippery" widget does I<not>
consult the "Existing" values: it goes right from "Values" to "Defaults".

Notice that if "Values" is given and defined, it overshadows any
sticky value.


=head1 AUTHOR

Eryq, Hughes STX Corporation

=head1 RCS INFO

$Id: HTML.pm,v 1.99 2003/05/16 14:38:53 solomon Exp $

=cut

use HTML::Stream qw(:funcs);
@ISA = qw(HTML::Stream);

use strict; 
use vars qw(%StdColors);

#------------------------------------------------------------
#
# GLOBALS
#
#------------------------------------------------------------

# New standard colors:
%StdColors = ( 
	      'IMSERROR'   => '#800010',
	      'IMSBLUE'    => '#1000CC',
	      'REQFIELD'   => '#D00000',
	      
	      'PageBack'   => '#999999',
	      'HiBack'     => '#D0D0D0',
	      'BriteBack'  => '#FFFFFF',
	      'TableCaptionBack'  => '#CCFFFF',
	      'TableHeaderBack'   => '#FFFFCC',
	      
	      'PkgItemsTblHeader' => '#FFFFCC',
	      'PkgOptsTblHeader'  => '#CCFFCC',
	      
	      'StdBlue'    => '#1000CC',
	      
	      'ReqField'   => '#D00000',
	      'ErrorText'  => '#800010',
	      'TitleText'  => '#1000CC',
	      
	      'AlertBack'  => '#FFD0D0',
	      'AlertText'  => '#800010'  
	     );

# KLUDGE: hack to work around Netscape <INPUT TYPE=IMAGE> cache bug.
# How often (in seconds) to refresh images?
my $ImageRefresh = 10; 

#------------------------------------------------------------
# new
#------------------------------------------------------------
sub new {
  my $class = shift;
  my $self = $class->SUPER::new(@_);   # inherited init
  $self->{AllowSmall} = 1;    
  $self->tr_color($self->hue('HiBack'));
  $self->{MsubmitCounter} = 1;
  $self;
}

#------------------------------------------------------------
# allow_small
#------------------------------------------------------------
# Re-set the flag that allows "small" markup:

sub allow_small {
  my ($self, $value) = @_;
  $self->{AllowSmall} = $value;
}

#------------------------------------------------------------
# need_scrollpoint [YESNO]
#------------------------------------------------------------
# Instance method.
# Does this page have an internal anchor point named SCROLL for scrolling
# which it still needs to output?  Set it to any true value you want if yes, 
# false for no.
#
# It is used by applications as follows:
#
# If false when page generation starts (the default), then a SCROLL
# anchor will be output at the top of the document, to catch any
# errant #SCROLL URLs and prevent them from zipping to the bottom.
#
# If true when page generation starts, then scrollpoint() ... _scrollpoint() 
# will work exactly once: the _scrollpoint() method will reset the
# need_scrollpoint to false, so future invocations of scrollpoint() do 
# nothing.

sub need_scrollpoint {
  my ($self, $yesno) = @_;    
  $self->{NeedScrollPoint} = $yesno if @_ > 1;
  $self->{NeedScrollPoint} || '';
}

#------------------------------------------------------------
# req
#------------------------------------------------------------
# Set/get the CGI request object.  This can be any object that responds 
# appropriately to the "param" method; its intent is to allow some
# of the "widget" methods to make use of CGI params as defaults.

sub req {
  my ($self, $req) = @_;
  (@_ >= 2) and $self->{Req} = $req;
  $self->{Req};
}

#------------------------------------------------------------
#
# OUTPUT METHODS
#
#------------------------------------------------------------

#------------------------------------------------------------
# applet_code PARAMHASH
#------------------------------------------------------------
# Loads an Applet.  The PARAMHASH contains:
#
#    ClassName => the applet's filename ( xxx.class)
#    CodeBase  => the directory where the applet resides
#    Width     => width of the applet ( in pixels)
#    Height    => height of the applet ( in pixels)
#    Alt       => alt text *required*

sub applet_code {
  my $self = shift;
  my %params = @_;
  
  # Get the value to use:
  my $classname = $params{ClassName};
  my $code_base = $params{CodeBase};
  my $width     = $params{Width};
  my $height    = $params{Height};
  my $alt       = $params{Alt};
  
  # Output tag:
  $self->tag('APPLET', ALT=>$alt, 
                       CODE=>$classname, CODEBASE=>$code_base, 
                       WIDTH=>$width, HEIGHT=>$height
            );
}

sub _applet_code {
  my $self = shift;
  HTML::Stream->accept_tag('APPLET');
  $self->_APPLET;
}

#------------------------------------------------------------
# applet_param PARAMHASH
#------------------------------------------------------------
# Loads an Applet.  The PARAMHASH contains:
#
#    Name => Parameter Name
#    Value  => Parameter's Value

sub applet_param  {
  my $self = shift;
  my %params = @_;
  
  # Get the value to use:
  my $name = $params{Name};
  my $value = $params{Value};
  
  # Output tag:
  $self->tag('PARAM', NAME=>$name, VALUE=>$value );
}

#------------------------------------------------------------
# button %PARAMS
#------------------------------------------------------------
# Print a possibly-hyperlinked image.
# If there is an HREF, uses SRC for the image.
# If there's no HREF, uses GREY for SRC if given: useful to "grey out" things.
#
# Params are sent in by hash:
#        
#     SRC      The SRC param of the IMG tag.
#     GREY     The SRC param of the IMG tag if no HREF.  Defaults to SRC.
#     ALT      The ALT param of the IMG tag.  Defaults to 'button'.
#     HREF     The HREF of the A tag.  If not given, no hyperlink is output.
#     BORDER   The BORDER param of the IMG tag.  Defaults to 0.
#
# Note that param keys are all-uppercase, to be identical to HTML::Stream
# conventions for specifying HTML tag parameters.

sub button {
  my $self = shift;    
  my %params = @_;
  
  my $src  = $params{SRC};
  my $grey = $params{GREY};
  my $alt  = $params{ALT} || 'button';
  my $href = $params{HREF};
  my $target = $params{TARGET};
  my $border = $params{BORDER} || 0;

  # For NASA 508, add "opens in new window" to alternate text if the
  # TARGET (directs the browser to open a new window) is set:
  $alt .= ' (opens in new window)' if $target;
  
  # Default GREY to SRC:
  $grey = $src unless defined($grey);
  
  # Print:
  if ($href && $target) {
    $self -> A(HREF=>$href, TARGET=>$target)
  }
  elsif ($href) {
    $self -> A(HREF=>$href)
  }
  $self -> IMG(SRC=>($href ? $src : $grey),
	       ALT=>"[$alt]",
	       BORDER=>$border);
  $self -> _A                  if $href;
  $self;
}

#------------------------------------------------------------
# color RED,GREEN,BLUE
# color COLORNAME
#------------------------------------------------------------
# Set the text color. All color values are 0..255.
# Use the C<_color> method to end.

sub color {
  my ($self, @color) = @_;
  my $hue;
  if (@color == 1) {     # a color name: map to RGB array...
    $hue = $StdColors{$color[0]} || die("unknown color: $color[0]");
  }
  else {
    $hue = sprintf("#%02X%02X%02X", @color);
  }
  $self->FONT(COLOR=>$hue);
}
sub _color {
  $_[0]->_FONT;
}

#------------------------------------------------------------
# hue NAME
#------------------------------------------------------------
# Lookup and return a standard color, as '#FFFFFF'.

sub hue {
  my ($self, $hue) = @_;
  $StdColors{$hue}
}

#------------------------------------------------------------
# degs
#------------------------------------------------------------
# Output a "degrees" symbol
sub degs {
  $_[0]->e('#176');
}

#------------------------------------------------------------
# directions 
#------------------------------------------------------------
# Begin a short burst of directions.  Use the C<_directions> method to end.
# Depends on the doc_url().

sub directions {    
  $_[0] -> P; 
#       -> IMG(SRC=>IW::Util::doc_url('icons/hand.right.gif'),
#	   ALT=>'Instructions',
#	   HSPACE=>10,
#	   ALIGN=>'LEFT');
}

sub _directions {
  $_[0] -> BR(CLEAR=>'ALL');
}

#------------------------------------------------------------
# field_cell PARAMHASH
#------------------------------------------------------------
# Start a compact form-input cell in a table.  Does NOT
# output the TD directive.  Use it like this:
#
#     $HTML -> TD(NOWRAP=>undef);
#     $HTML -> field_cell(Title=>"Email") 
#           -> INPUT(SIZE=>20, NAME=>"email", VALUE=>$email)
#           -> _field_cell;
#
# Params are:
#
#     Small  =>  Optional.  Wrap everything in small font?  Default is true.
#     Stack  =>  Stack vertically?  Default is false.
#     Req    =>  Is field required?  Default is false.
#     Title  =>  The title.
#     Subtitle => Subtitle (in smaller font)
#     
# WARNING: the options are remembered until the next _field_cell,
# so you CANNOT nest calls!

sub field_cell {
  my $self = shift;
  my %params = @_;
  $self->{field_cell_params} = \%params;     # remember them
  
  # Adjust field with defaults:
  $params{Small} = 1 unless (exists($params{Small}) && !$params{Small});
  
  # Output:
  $self->small if $params{Small};
  $self->field_title($params{Req}, $params{Title},);
  $self->BR    if $params{Stack};
  if ($params{Subtitle}) {
    $self->t($params{Subtitle});
    $self->BR if $params{Stack};
  }
  $self;
}

sub _field_cell {
  my $self = shift;
  $self->_small if $self->{field_cell_params}{Small};
}

#------------------------------------------------------------
# field_title REQ,TITLE
#------------------------------------------------------------
# Input field title.  If REQ is true, the field is required.

sub field_title {
  my ($self, $req, $title) = @_;
  if ($req) {
     $self -> B -> t("$title") -> _B  -> t(' (required)');
  }
  else {
    $self -> B -> t($title) -> _B;
  }
}

#------------------------------------------------------------
# head N, TEXT...
#------------------------------------------------------------
# Print a simple <Hn>....</Hn> heading.

sub head {
  my $self = shift;
  my $n = shift;
  $self -> tag("H$n") -> t(@_) -> tag("_H$n");
}

#------------------------------------------------------------
# hidden NAME=>VALUE, ..., NAME=>VALUE
#------------------------------------------------------------
# Output a whole lot of hidden fields to a form.
# Newlines are output, so do this I<outside> of any "PRE" environments!

sub hidden {
  my $self = shift;
  while (@_) {
    my ($n, $v) = (shift, shift);
    $self -> tag('INPUT', TYPE=>'hidden', NAME=>$n, VALUE=>$v) -> nl;
  }
  $self;
}

#------------------------------------------------------------
# hspace NSPACES
#------------------------------------------------------------
# Print a small amount of horizontal whitespace.
# NSPACES is the amount of space.  It defaults to 6.

sub hspace {
  my ($self, $n) = @_;
  defined($n) or $n = 6;
  $self->io->print('&nbsp;' x $n);
  $self;
}

#------------------------------------------------------------
# hypertext TEXT, HREF
#------------------------------------------------------------
# Output a simple hypertext link.  If HREF is false, just outputs the TEXT.

sub hypertext {
  my ($self, $text, $href, %params) = @_;
  
  $self->A(HREF=>$href, %params)   if $href;

  # For NASA 508, add "*" to the end of the text if the
  # TARGET (directs the browser to open a new window) is set:
  $text .= '*' if $params{TARGET};

  $self->text($text);
  $self->_A                        if $href;
  $self;
}

#------------------------------------------------------------
# input_hidden NAME,VALUE
#------------------------------------------------------------
# Print a hidden-form field.
#
# You should I<always> use this routine to print hidden fields!
# On some browsers (early Mosaic), all form elements types
# (like "hidden") must be I<lowercase>.

sub input_hidden {
  my ($self, $name, $value) = @_;
  $self->tag('INPUT', TYPE=>'hidden', NAME=>$name, VALUE=>$value);
}

#------------------------------------------------------------
# mailto ADDRESS
#------------------------------------------------------------
# Typeset an email address as a C<mailto:> URL.

sub mailto {
  my ($self, $addr) = @_;
  $self -> A(HREF=>"mailto:$addr") -> t($addr) -> _A;
}

#------------------------------------------------------------
# msubmit_flag 
#------------------------------------------------------------
# Call this anywhere inside your form.  It creates a hidden field of 
# the form 'msubmit=yes' for when "multisubmit buttons" are used.

sub msubmit_flag {
  $_[0]->input_hidden('msubmit', 'yes');
}

#------------------------------------------------------------
# msubmit_name 
#------------------------------------------------------------
# Used to create an unique name for the msubmit hidden tags.
# Starts with "msubmit1" and increments up (e.g. "msubmit2", "msubmit3" 
# for each msubmit button added to the page. 

sub msubmit_name {

   my $self = shift;
   my $name = 'msubmit'.$self->{MsubmitCounter};
   $self->{MsubmitCounter}++;
   return $name;

}

#------------------------------------------------------------
# msubmit_button PARAMHASH
#------------------------------------------------------------
# Outputs special imagemap "button", with the desired (but fake)
# set of query tags encoded and stored in a hidden tag which 
# corresponds to the imagemap's name: the imagemap is named "msubmit7"
# and the corresponding hidden tag is: 
#      msubmit7 = "msubmit?name=value&name=value&"
# The params are:
#
#    SRC     => the source (mutex with CAPTION; see below)
#    CAPTION => the caption (mutex with SRC; see below)
#    QUERY   => a *small* array of the form [name=>value, ..., name=>value]
#    BORDER  => explicit border value (default 0, undef=leave out border tag)
#    ALT     => Alternate name for button.
#    PRETTY  => This is used as the name of the imagemap 
#               It must be unique for the HTML page being generated.
#               If two buttons have the same PRETTY and thus the same name 
#               their actions could be confused.  
#               If PRETTY is not given then a unique name is generated  
#               by sub msubmit_name.
#
# If the SRC is omitted, and a CAPTION is supplied instead, the button
# will be a standard C<"INPUT TYPE=SUBMIT">, with the CAPTION used as the
# value, and the QUERY encoded in the name.
#
# WARNING: there is a Netscape 2.0 cacheing bug involving <INPUT TYPE=IMAGE>.

sub msubmit_button {
  my $self = shift;
  my %params = @_;
  $params{QUERY} or $params{QUERY} = [];
  
  # Build tag which encodes the query:
  my $tag = IW::Util::msubmit_tag($params{QUERY});
  
  my $name = $self->msubmit_name;

  my @options;
  # button options:    
  if ($params{SRC}) {

    # Use PRETTY to make image buttons have pretty names.
    $name = $params{PRETTY} if (defined($params{PRETTY})) ;
                                    
    # Figure out the BORDER to supply:
    exists($params{BORDER})  or $params{BORDER} = 0;   # default if absent
    push (@options,'BORDER'=>$params{BORDER}) if (defined($params{BORDER}));
    push (@options,'onClick'=>$params{onClick}) if (defined($params{onClick}));
    push (@options,'TYPE'=>'IMAGE');
    push (@options,'SRC'=>$params{SRC});
    push (@options,'ALT'=>$params{ALT}) if (defined($params{ALT}));
    push (@options,'NAME'=>$name); 
    push (@options,'HEIGHT'=>$params{HEIGHT}) if $params{HEIGHT};
    push (@options,'WIDTH'=>$params{WIDTH}) if $params{WIDTH};

  }
  else {
    push (@options,'TYPE'=>'submit');
    push (@options,'NAME'=>"$name.x");
    push (@options,'onClick'=>$params{onClick}) if (defined($params{onClick}));
    push (@options,'VALUE'=>$params{CAPTION});
    push (@options,'ALT'=>$params{ALT}) if (defined($params{ALT})); 
  }

  # Output button
  push (@options,'CLASS'=>($params{CLASS} || 'submitbutton'));
  $self->INPUT(@options);

  # Output hidden msubmit tag:
  $self->input_hidden($name,$tag);
}

#------------------------------------------------------------
#  script_FORM PARAMHASH
# _script_FORM 
#------------------------------------------------------------
# Start a form that is submitted back to this script.
# All arguments are passed into FORM.  
#
#   * METHOD defaults to POST.  
#   * ACTION defaults to the script action (which contains the uniquifier 
#     in the PATH_INFO portion, to force a page reload).
#   * The sid is automatically embedded as a hidden field, if present.

# Begin...
sub script_FORM {
  my ($self, %args) = @_;
  $args{METHOD} or $args{METHOD} = 'POST';
  $args{ACTION} or $args{ACTION} = IW::Util::script_action();
  $self->FORM(%args);
  $self->hidden('sid', $IW::App::SESSION->id) if $IW::App::SESSION;
  $self;
}

# End...
sub _script_FORM {
  my ($self) = @_;
  $self->_FORM;
}

#------------------------------------------------------------
#  scrollpoint [MATCHTEXT]
# _scrollpoint
#------------------------------------------------------------
# Output the standard scrollpoint, <A NAME="SCROLL">, if it's
# needed.  See need_scrollpoint() for details.
#
# Normally, the scrollpoint is output simply if need_scrollpoint()
# returns true.  However, if the optional MATCHTEXT is given, then 
# the scrollpoint is only output if it is true AND it matches that
# text.  If the match fails, _scrollpoint() will NOT clear the 
# need_scrollpoint() value.   
#
sub scrollpoint {
  my ($self, $match_ns) = @_;
  my $ns = $self->need_scrollpoint;
  if ($ns and (!$match_ns or ($match_ns eq $ns))) {
    $self->A(NAME=>'SCROLL');  
    $self->{ScrollPointBegun} = 1;
  }
  $self;
}

sub _scrollpoint {
  my $self = shift;
  if ($self->{ScrollPointBegun}) {
    $self->_A;
    $self->{ScrollPointBegun} = 0;
    $self->need_scrollpoint(undef);    # no future ones can be begun
  }
  $self;
}

#------------------------------------------------------------
# _FORM
#------------------------------------------------------------
# Override _FORM from HTML::Stream to auto-output the hidden ENDFORM tag.

sub _FORM {
  my ($self) = @_;
  $self->hidden('endform','1');
  $self->io->print("</FORM>\n");
}

#------------------------------------------------------------
#  small
# _small
#------------------------------------------------------------
# Start small font.  Use _small to end.
# NOTE: the session state tell us whether or not to do this!

sub small {
  my $self = shift;
  $self->FONT(SIZE=>-1)    if $self->{AllowSmall};
  $self;
}

sub _small {
  my $self = shift;
  $self->_FONT             if $self->{AllowSmall};
  $self;
}

#------------------------------------------------------------
# t_B TEXT
#------------------------------------------------------------
# Embolden some text.

sub t_B {
  my $self = shift;
  $self -> B -> text(@_) -> _B;
}

#------------------------------------------------------------
# t_EM TEXT
#------------------------------------------------------------
# Emphasize some text.

sub t_EM {
  my $self = shift;
  $self -> EM -> text(@_) -> _EM;
}

#------------------------------------------------------------
# url URL
#------------------------------------------------------------
# Typeset a URL as a hyperlink to itself.

sub url {
  my ($self, $url) = @_;
  $self -> A(HREF=>$url) -> t($url) -> _A;
}

#------------------------------------------------------------
# warning 
#------------------------------------------------------------
# Begin a short warning.  Use the C<_warning> method to end.
# Depends on the doc_url().

sub warning {    
  $_[0] 
    -> P
    -> IMG(SRC=>IW::Util::doc_url('icons/note-warning.gif'),
	   ALT=>'Warning',
	   HSPACE=>10,
	   ALIGN=>'LEFT');
}

sub _warning {
  $_[0] -> BR(CLEAR=>'ALL');
}

#------------------------------------------------------------
#
# WIDGETS
#
#------------------------------------------------------------

#------------------------------------------------------------
# checkbox PARAMHASH
#------------------------------------------------------------
# Print a checkbox "widget".  The PARAMHASH contains...
#
#    Name      => the name
#    Choice    => the value of this button, if chosen: should evaluate to true!
#
# And, optionally, one of the following:
#
#    Value     => current choice (true or false), given explicitly
#    Default   => default setting (true or false)
#
# CGI PARAMS:
#    If checked, will produce the CGI param "Name = Choice".  
#    If not checked, will not produce anything.
#
# KLUDGE: this would be grossly inefficient if called multiple times with a
# large array of checkboxes, and looking for exact matches.  In other
# words, if given this as our CGI input:
#
#       gkey = 1/2/345
#       gkey = 3/4/567
#       ...
#       gkey = 7/8/999
#
# then finding out whether a checkbox with name "gkey" and value "5/6/789"
# should be check or not requires a search through all n items... and if
# we needed to do that n times, we'd have n^2 complexity: horrible for large 
# values of n (e.g, 100, which is very possible).
#
# Unfortunately, since CGI params may be "tweaked" by our code, we can't
# *always* pre-hash it and save it.  HOWEVER, we'll take that risk (for now),
# and maybe provide an escape route later for ultrahypercomplex scripts. :-( 
# The important thing is that checkboxes be able to tell, efficiently,
# whether or not they were checked the last time through...

sub checkbox {
  my $self = shift;
  my %params = @_;
  
  # Get the checkbox name:
  my $name = $params{Name};
  
  # Check for a checkbox hash under this name; if none, create it!
  $self->{CheckboxHash} or $self->{CheckboxHash} = {};
  unless ($self->{CheckboxHash}{$name}) {
    $self->{CheckboxHash}{$name} = {};	
    my $value;
    foreach $value ($self->{Req}->param($name)) {
      $value = 0 if !defined $value;
      $self->{CheckboxHash}{$name}{$value} = 1;
    }
  }
  
  # Get the value to use:
  my $value = $params{Value};
  my $choice = $params{Choice} || 'yes';
  defined($value) or $value = ($self->{CheckboxHash}{$name}{$choice});
  defined($value) or $value = $params{Default};
  
  # Output tag:
  $self->tag('INPUT', TYPE=>'checkbox', CLASS=>($params{CLASS} || 'checkbox'),
              NAME=>$name, VALUE=>$choice,
	     ($value ? (CHECKED=>undef) : ()));
}

#------------------------------------------------------------
# concealed PARAMHASH
#------------------------------------------------------------
# Print one or more hidden field "widgets".  The PARAMHASH contains...
#
#    Name   => the name (same for all buttons in a radio group)
#
# And, optionally, one of the following:
#
#    Values   => current values, given explicitly
#    Defaults => values, if no existing or explicit values defined.
#
# CGI PARAMS:
#    Will produce the CGI param "Name = value" for each value in the 
#    values used (either Values or Defaults).

sub concealed {
  my $self = shift;
  my %params = @_;
  my @sticky;
  
  # Get the values:
  my $name = $params{Name};       # get the name
  my $values = $params{Values};
  defined($values) or $values = [ $self->{Req}->param($name) ];
  int(@$values)    or $values = $params{Defaults};
  
  # Output tag:
  my $value;
  foreach $value (@$values) {
    $self->tag('INPUT', TYPE=>'hidden', NAME=>$name, VALUE=>$value);
  }
  $self;
}

#------------------------------------------------------------
# input_text PARAMHASH
#------------------------------------------------------------
# Print a text-input "widget".  The PARAMHASH contains...
#
#    Name      => the name
#
# And, optionally, one of the following:
#
#    Value     => current value, given explicitly
#    Default   => default value, if no existing or explicit values defined.
#    Type      => Supply as 'password'
#    Size      => The number of rows
#
# CGI PARAMS:
#    Will produce the CGI param "Name = value" for the value used
#    (either Value or Default).

sub input_text {
  my $self = shift;
  my %params = @_;
  
  # Get the value to use:
  my $name = $params{Name};       # get the name
  my $value = $params{Value};
  defined($value) or $value = ($self->{Req}->param($name))[0];
  defined($value) or $value = $params{Default};
  
  # Output tag:
  $self->tag('INPUT', NAME=>$name, VALUE=>$value, CLASS=>($params{CLASS} | 'textinput'),
   ($params{Type} ? (TYPE=>$params{Type}) : ()),
   ($params{Size} ? (SIZE=>$params{Size}) : ()),
   ($params{Maxlength} ? (MAXLENGTH=>$params{Maxlength}) : ()));
}

#------------------------------------------------------------
# input_textarea PARAMHASH
#------------------------------------------------------------
# Print a text-input "widget".  The PARAMHASH contains...
#
#    Name      => the name
#    Rows      => the rows
#    Cols      => the columns
#
# And, optionally, one of the following:
#
#    Value     => current value, given explicitly
#    Default   => default value, if no existing or explicit values defined.
#
# CGI PARAMS:
#    Will produce the CGI param "Name = value" for the value used
#    (either Value or Default).

sub input_textarea {
  my $self = shift;
  my %params = @_;
  
  # Get the value to use:
  my $name = $params{Name};       # get the name
  my $value = $params{Value};
  defined($value) or $value = ($self->{Req}->param($name))[0];
  defined($value) or $value = $params{Default};
  
  # Output tag:
  $self -> TEXTAREA(NAME=>$name, CLASS=>($params{CLASS} || 'textarea'), ROWS=>$params{Rows}, COLS=>$params{Cols});
  $self -> t($value);
  $self -> _TEXTAREA;
}

#------------------------------------------------------------
# radio_button PARAMHASH
#------------------------------------------------------------
# Print a radio button "widget".  The PARAMHASH contains...
#
#    Name      => the name (same for all buttons in a radio group)
#    Choice    => the value of this button, if chosen
#
# And, optionally, one of the following:
#
#    Value     => current selection of this radio group, given explicitly
#    Default   => default selection of this radio group, if no existing 
#                 or explicit values defined.
#    IsDefault => A nicer "Default": if true, this is the default button.
#
# CGI PARAMS:
#    If this radio button is selected, will produce the CGI param 
#    "Name = Choice".  If not selected, will produce nothing.

sub radio_button {
  my $self = shift;
  my %params = @_;
  
  # Get the value to use:
  my $name = $params{Name};       # get the name
  my $value = $params{Value};
  defined($value) or $value = ($self->{Req}->param($name))[0];
  defined($value) or $value = $params{Default};
  defined($value) or $value = ($params{IsDefault} ? $params{Choice} :undef); 
  
  # On or off?
  my $choice = $params{Choice};
  my $is_selected = (defined($value) && ($choice eq $value));
  
  # Output tag:
  $self->tag('INPUT', TYPE=>'radio', NAME=>$name, VALUE=>$choice, 
             CLASS=>($params{CLASS} || 'radiobutton'),
	     ($is_selected ? (CHECKED=>undef) : ()));
}

#------------------------------------------------------------
# select PARAMHASH
#------------------------------------------------------------
# Print a multi-selection "widget".  The PARAMHASH contains:
#
#    Name    => the widget name
#    Size    => the widget size
#    Choices   => ref to an array of choices
#    ChoiceMap => ref to an array of (text => value) (undef value means text)
#    Multi   => allow many to be selected?
#
# And, optionally, one of the following:
#
#    Values   => ref to array of selections, given explicitly
#    Defaults => ref to array of selections, if no existing or explicit values
#
# CGI PARAMS:
#    Will produce the CGI param "Name = Choice" for each Choice selected.

sub select {
  my ($self, %params) = @_;
  
  # Basic attributes:
  my $name = $params{Name};
  my $size = $params{Size} || 5;
  my $none = $params{None} || 0;
  
  # Get the values which are already selected:
  my $values = $params{Values} || $params{Current};
  # Kludge to add None button to dep valids
  # Skip the following lines to keep values from being picked up from the
  # tags.
  if (!$none) {
    ($values && @$values) or $values = [$self->{Req}->param($name)];
    ($values && @$values) or $values = $params{Defaults};
  }
  
  # Create lookup from values:
  my %selected = map {$_=>1} @{$values || []};
  
  # Start:
  my @multiple = ($params{Multi} ? (MULTIPLE=>undef) : ());
  $self->tag('SELECT', @multiple, NAME=>$name, SIZE=>$size, 
             CLASS=>($params{CLASS} || 'multiselect')) -> nl;
  
  # For each choice...
  my @map = ($params{Choices} 
	     ? map {$_=>undef} @{$params{Choices}}    # create map
	     : @{$params{ChoiceMap}});                # got it already
  my $nselected = 0;
  while (@map) {
    # Get choice and value-if-chosen:
    my $choice = shift @map;
    my $cvalue = shift @map;
    my $svalue = (defined($cvalue) ? $cvalue : $choice);	
    
    # Get OPTION's VALUE:
    my @value = defined($cvalue) ? (VALUE=>$cvalue) : ();	
    
    # Should we label OPTION as SELECTED?  Careful with single-select!
    my @selected;
    if ($selected{$svalue} && ($params{Multi} || !$nselected)) {   # yes!
      @selected = (SELECTED=>undef);
      ++$nselected;
    }
    
    # Output:
    $self -> tag('OPTION', @value, @selected) -> t($choice) -> nl;
  }    
  $self->tag('/SELECT') -> nl;
  $self;
}

#------------------------------------------------------------
# select_many PARAMHASH
#------------------------------------------------------------
# Print a multi-selection "widget".  
# Obsolete front-end onto select().
# The PARAMHASH contains:
#
#    Name    => the widget name
#    Size    => the widget size
#    Choices => ref to an array of choices
#
# And, optionally, one of the following:
#
#    Values   => ref to array of selections, given explicitly
#    Defaults => ref to array of selections, if no existing or explicit values
#
# CGI PARAMS:
#    Will produce the CGI param "Name = Choice" for each Choice selected.

sub select_many {
  my $self = shift;
  $self->select(@_, Multi=>1);
}

#------------------------------------------------------------
# select_one PARAMHASH
#------------------------------------------------------------
# Print a single-selection "widget".  
# Obsolete front-end onto select().
# The PARAMHASH contains:
#
#    Name    => the widget name
#    Choices => ref to an array of choices
#
# And, optionally, one of the following:
#
#    Value   => current selection, given explicitly
#    Default => selection, if no existing or explicit values defined.
#
# CGI PARAMS:
#    Will produce the CGI param "Name = Choice" for the Choice selected.

sub select_one {
  my $self = shift;
  my %params = @_;
  
  # Get explicitly-selected value(s):
  my $value = defined($params{Value}) ? $params{Value} : $params{Current};
  my @values = (defined($value) ? ($value) : ());
  my $size = defined($params{Size}) ? $params{Size} : 1;
  
  # Get fallback value(s):
  my $default = $params{Default};
  my @defaults = (defined($default) ? ($default) : ());
  
  # Call:
  $self->select(@_,
		Size  => $size,
		Multi => 0,
		Values   => \@values,
		Defaults => \@defaults);
}

#------------------------------------------------------------
# sort_menu PARAMHASH
#------------------------------------------------------------
# Print a complex sort menu. The PARAMHASH contains:
#
#    Name    => the widget name (name of the CGI field before the ".0", etc.)
#    Choices => ref to an array of possible sort keys
#
# And, optionally, one of the following:
#
#    Values   => vector of current selections, given explicitly
#    Defaults => vector of selections to fall back on
#
# NOTE: uses select_one(), so it will pick up on CGI params as the
# current default.

sub sort_menu {
  my ($self, %params) = @_;
  my $name     = $params{Name};     # name of CGI field, before ".0", ".1"...
  my $choices  = $params{Choices};  # possible sort keys
  my $defaults = $params{Defaults}; # vector of current choices
  my $nkeys    = $params{NumKeys} || 2;   # number of sort keys
  my $what     = $params{What};
  
  # What can we sort on?
  my @allchoices = ('nothing', @$choices);
  
  $self -> tag('NOBR');
  my $i;
  for ($i = 0; $i < $nkeys; $i++) {
    $self -> t(" then: ")      if $i;
    $self -> select_one(Name   =>"${name}.$i",
			Choices=>($i ? \@allchoices : $choices),
			Default=>($defaults->[$i] || 'nothing'));
  }
  $self -> tag('_NOBR');
}

#------------------------------------------------------------
#
# APPLICATION BUTTONS
#
#------------------------------------------------------------
# WARNING: all of the following buttons assume that they are
# being served by the main script.

#------------------------------------------------------------
# binfo_btn_to HREF
#------------------------------------------------------------
# Print a teeny brief-info (DIF) button, to the given HREF.
# Brief Info is now called Summary Document.  From GCMD.

sub binfo_btn_to {    
  my ($self, $href) = @_;
  $self->button(SRC =>IW::Util::doc_url('btns/item-binfo.gif'),
		GREY=>IW::Util::doc_url('btns/item-binfo-grey.gif'),
		ALT =>'Summary Document',
                TARGET=>'support',
		HREF=>$href);
}

#------------------------------------------------------------
# dinfo_btn_to HREF
#------------------------------------------------------------
# Print a teeny detailed-info (Guide) button, to the given HREF.
# Now called Detailed Document.  From data center.

sub dinfo_btn_to {    
  my ($self, $href) = @_;
  $self->button(SRC =>IW::Util::doc_url('btns/item-dinfo.gif'),
		GREY=>IW::Util::doc_url('btns/item-dinfo-grey.gif'),
		ALT =>'Detailed Document',
		HREF=>$href);
}

#------------------------------------------------------------
# defn_btn TYPE,ITEM
#------------------------------------------------------------
# Print a teeny definition button.  Arguments are the same as defn_url().

sub defn_btn {    
  my $self = shift;
  $self->button(SRC =>IW::Util::doc_url('btns/item-defn.gif'),
		GREY=>IW::Util::doc_url('btns/item-defn-grey.gif'),
		ALT =>'Definition',
                TARGET=>'support',
		HREF=>IW::Util::defn_url(@_));
}

#------------------------------------------------------------
# help_btn CONTEXT
#------------------------------------------------------------
# Print a teeny help button.  Arguments are the same as help_url().

sub help_btn {    
  my ($self, $context) = @_;
  
# Not ready to use Helpet yet, causes some browsers to crash.
#
#    # If java can be used, put the Help Applet (Helpet), if
#    # java can't be used then put the CGI image button.
#
#    if ($IW::App::SESSION->state->{AllowJava} eq 'Y') {
#        $self->applet_code(ClassName => 'IMSWG.IW.HTMLet.HTMLet.class',
#                           CodeBase  => ${IU::DocsUrl}.'java/',
#                           Width     => 35,
#                           Height    => 30 );
#        my $hidden_string .= "doc=http://".$ENV{SERVER_NAME}.${IU::DocsUrl}."manuals/imsuser_".$context.".html&name=Hint"; 
#        $self->applet_param ( Name => 'HIDDEN', Value => $hidden_string);
#        $self->_applet_code; 
#    } else {
  $self->button(SRC=>IW::Util::doc_url('btns/tiny-helpword.gif'),
		ALT=>'Help',
		HREF=>IW::Util::help_url($context));
  #    }
  
}

#------------------------------------------------------------
# dset_attr_btn DSKEY
# dset_attr_btn PARAMHASH
#------------------------------------------------------------
# Print a button for dataset attributes.
#
# DSKey is the dataset key.
# Origin defaults to the current origin tag if unspecified.
# Size is either BIG or SMALL.

sub dset_attr_btn {    
  my $self = shift;
  my %params = ((@_ == 1) ? (DSKey=>$_[0]) : @_); 
  my $dskey  = $params{DSKey};
  my $origin = $params{Origin} || $self->req->param('origin');
  my $size   = $params{Size}   || 'SMALL';
  my $href;
  
  # What's the URL?
  $dskey and $href = IW::Util::script_url('mode'=>'DSETATTRS', 
					  'dskey'=>$dskey,
					  'origin'=>$origin);
  
  # Output !
  $self->button(SRC  =>"${IU::DocsUrl}btns/item-dattr.gif",
		GREY =>"${IU::DocsUrl}btns/item-dattr-grey.gif",
		ALT  =>'Data set Attributes',
		HREF => $href);
}

#------------------------------------------------------------
# gran_attr_btn GKEY
# gran_attr_btn PARAMHASH
#------------------------------------------------------------
# Print a teeny button for granule attributes.
#
# GKey is the dataset key.
# Origin defaults to the current origin tag if unspecified.
# Size is either BIG or SMALL (default).
# GreyedOut means to force the button to be greyed out.

sub gran_attr_btn {    
  my $self = shift;
  my %params = ((@_ == 1) ? (GKey=>$_[0]) : @_); 
  my $gkey   = $params{GKey};
  my $origin = $params{Origin} || $self->req->param('origin');
  my $size   = $params{Size}   || 'SMALL';
  my $greyed = $params{GreyedOut};
  my $href;
  
  # What's the URL?
  if ($gkey and !$greyed) {
    $href = IW::Util::script_url('mode'=>'GRANATTRS', 
				 'gkey'=>$gkey,
				 'origin'=>$origin);
  }
  
  # Output!
  $self->button(SRC  =>"${IU::DocsUrl}btns/item-gattr.gif",
		GREY =>"${IU::DocsUrl}btns/item-gattr-grey.gif",
		ALT  =>'Data Granule Attributes',
		HREF => $href);
}

#------------------------------------------------------------
# gran_attr_link TEXT,GKEY
# gran_attr_link TEXT,PARAMHASH
#------------------------------------------------------------
# Print a link for granule attributes.
#
# GKey is the dataset key.
# Origin defaults to the current origin tag if unspecified.
# Size is either BIG or SMALL (default).
# GreyedOut means to force the button to be greyed out.

sub gran_attr_link {    
  my $self = shift;
  my $text = shift;
  my %params = ((@_ == 1) ? (GKey=>$_[0]) : @_); 
  my $gkey   = $params{GKey};
  my $origin = $params{Origin} || $self->req->param('origin');
  my $size   = $params{Size}   || 'SMALL';
  my $greyed = $params{GreyedOut};
  my $href;
  
  # What's the URL?
  if ($gkey and !$greyed) {
    $href = IW::Util::script_url('mode'=>'GRANATTRS', 
				 'gkey'=>$gkey,
				 'origin'=>$origin);
  }

  # Output!
  $self->hypertext($text,$href);
}

#------------------------------------------------------------
# ...menu_btn 
#------------------------------------------------------------
# Print various top-level "multisubmit" navigation buttons that
# appear at the tops/bottoms of listings.

# Print the button for the granule listing:
sub granlist_menu_btn {
  $_[0] -> msubmit_button(QUERY =>['mode' => 'GRANLIST',
				   'submode'=>'LIST'],
			  SRC   =>"${IU::DocsUrl}btns/sub-glist.gif",
			  ALT   =>'List Data Granules');
}

# Print the button for creating a granule listing from a list of datasets:
sub dset_to_granlist_menu_btn {
  $_[0] -> msubmit_button(QUERY =>['mode' => 'GRANLIST',
				   'submode'=>'NEW'],
			  SRC   =>"${IU::DocsUrl}btns/sub-ds2glist.gif",
			  ALT   =>'List Data Granules');
}

# Print the button for the granule listing:
sub granlistall_menu_btn {
  $_[0] -> msubmit_button(QUERY =>['mode' => 'GRANLIST',
                                   'submode'=>'LIST',
                                   'select' => 'ALL'],
			  SRC   =>"${IU::DocsUrl}btns/sub-glist.gif",
			  ALT   =>'List All Data Granules');
}

# Print the button for displaying the map submode of the granlist.
# Display all granules. 

sub granlistmapall_menu_btn {
  $_[0] -> msubmit_button(QUERY =>['mode' => 'GRANLIST', 
                                   'submode'=>'MAP',
                                   'select' => 'ALL'],
			  SRC   =>"${IU::DocsUrl}btns/sub-gmap.gif",
			  ALT   =>'Show All Data Granules on a Map');
}

# Print the button for displaying the map submode of the granlist :

sub granlistmap_menu_btn {
  $_[0] -> msubmit_button(QUERY =>['mode' => 'GRANLIST', 
                                   'submode'=>'MAP'],
			  SRC   =>"${IU::DocsUrl}btns/sub-gmap.gif",
			  ALT   =>'Show Data Granules on a Map');
}

sub showMisr_toggle_btn {
  $_[0] -> msubmit_button(QUERY =>['mode' => 'GRANLIST',
                                   'submode'=>'MAPTOGGLE_ON',
                                   'mapmode'=>'TOGGLE'],
			  CAPTION =>'Show Block Numbering',
			  ALT =>'Toggle block numbering option on');
}

sub hideMisr_toggle_btn {
  $_[0] -> msubmit_button(QUERY =>['mode' => 'GRANLIST',
                                   'submode'=>'MAPTOGGLE_OFF',
                                   'mapmode'=>'TOGGLE'],
			  CAPTION =>'Hide Block Numbering',
			  ALT =>'Toggle block numbering option off');
}

sub bench_showMisr_toggle_btn {
  $_[0] -> msubmit_button(QUERY =>['mode' => 'WORKBENCH',
				   'submode'=>'MAPTOGGLE_ON',
				   'mapmode'=>'TOGGLE' ],
			  CAPTION =>'Show Block Numbering',
			  ALT =>'Toggle block numbering option on');
}

sub bench_hideMisr_toggle_btn {
  $_[0] -> msubmit_button(QUERY =>['mode' => 'WORKBENCH',
                                   'submode'=>'MAPTOGGLE_OFF',
				   'mapmode'=>'TOGGLE' ],
			  CAPTION =>'Hide Block Numbering',
			  ALT =>'Toggle block numbering option off');
}

# Print the button for  displaying the temporal submode of the workbench:
sub granlisttime_menu_btn {
  $_[0] -> msubmit_button(QUERY =>['mode' => 'GRANLIST', 'submode'=>'TIME'],
			  SRC   =>"${IU::DocsUrl}btns/sub-gtime.gif",
			  ALT   =>'Show Data Granules on a Time Line');
}

# Print the button for  displaying the temporal submode of the workbench:
# Display all granules.

sub granlisttimeall_menu_btn {
  $_[0] -> msubmit_button(QUERY =>['mode' => 'GRANLIST', 
                                   'submode'=>'TIME',
                                   'select' => 'ALL'],
			  SRC   =>"${IU::DocsUrl}btns/sub-gtime.gif",
			  ALT   =>'Show All Data Granules on a Time Line');
}

# Print the button for the granule spatial coverage:
sub covermap_menu_btn {
  $_[0] -> msubmit_button(QUERY =>['mode' => 'COVERMAP'],
			  SRC   =>"${IU::DocsUrl}btns/sub-gmap.gif",
			  ALT   =>'Show Data Granules on a Map');
}

# Print the button for adding to the workbench:
sub workbenchadd_menu_btn {
  $_[0] -> msubmit_button(QUERY =>['mode' => 'WORKBENCH', 'submode'=>'ADD'],
			  SRC   =>"${IU::DocsUrl}btns/sub-wbadd.gif",
			  ALT   =>'Add Data Granules to My Folder');
}

# Print the button for removing from the workbench:
sub workbenchrem_menu_btn {
  $_[0] -> msubmit_button(QUERY =>['mode' => 'WORKBENCH', 'submode'=>'REM'],
			  SRC   =>"${IU::DocsUrl}btns/sub-wbrem.gif",
			  ALT   =>'Remove Data Granule from My Folder');
}

# Print the button for adding to the workbench:
sub workbenchaddall_menu_btn {
  $_[0] -> msubmit_button(QUERY =>['mode' => 'WORKBENCH', 
				   'submode'=>'ADD',
				   'select' => 'ALL'],
			  SRC   =>"${IU::DocsUrl}btns/sub-wbadd.gif",
			  ALT   =>'Add All Data Granules to My Folder');
}

# Print the button for removing from the workbench:
sub workbenchremall_menu_btn {
  $_[0] -> msubmit_button(QUERY =>['mode' => 'WORKBENCH', 
				   'submode'=>'REM',
				   'select' => 'ALL'],
			  SRC   =>"${IU::DocsUrl}btns/sub-wbrem.gif",
			  ALT   =>'Remove All Data Granules from My Folder');
}

# Print the button for displaying the list submode of the workbench:
sub workbenchlist_menu_btn {
  $_[0] -> msubmit_button(QUERY =>['mode' => 'WORKBENCH', 'submode'=>'LIST'],
			  SRC   =>"${IU::DocsUrl}btns/sub-glist.gif",
			  ALT   =>'List Selected Data Granules in My Folder');
}

# Print the button for displaying the list submode of the workbench:
# List all granules.
sub workbenchlistall_menu_btn {
  $_[0] -> msubmit_button(QUERY =>['mode' => 'WORKBENCH',
				   'submode'=>'LIST',
				   'select' => 'ALL'],
			  SRC   =>"${IU::DocsUrl}btns/sub-glist.gif",
			  ALT   =>'List All Data Granules in My Folder');
}

# Print the button for  displaying the map submode of the workbench:
sub workbenchmap_menu_btn {
  $_[0] -> msubmit_button(QUERY =>['mode' => 'WORKBENCH', 'submode'=>'MAP'],
			  SRC   =>"${IU::DocsUrl}btns/sub-gmap.gif",
			  ALT   =>'Show Selected Data Granules on a Map');
}

# Print the button for  displaying the map submode of the workbench:
sub workbenchmapall_menu_btn {
  $_[0] -> msubmit_button(QUERY =>['mode' => 'WORKBENCH',
				   'submode'=>'MAP',
				   'select' => 'ALL'],
			  SRC   =>"${IU::DocsUrl}btns/sub-gmap.gif",
			  ALT   =>'Show All Data Granules on a Map');
}

# Print the button for  displaying the temporal submode of the workbench:
sub workbenchtime_menu_btn {
  $_[0] -> msubmit_button(QUERY =>['mode' => 'WORKBENCH', 'submode'=>'TIME'],
			  SRC   =>"${IU::DocsUrl}btns/sub-gtime.gif",
			  ALT   =>'Show Selected Data Granules on a Time Line');
}

# Print the button for  displaying the temporal submode of the workbench:
sub workbenchtimeall_menu_btn {
  $_[0] -> msubmit_button(QUERY =>['mode' => 'WORKBENCH',
				   'submode'=>'TIME',
				   'select' => 'ALL'],
			  SRC   =>"${IU::DocsUrl}btns/sub-gtime.gif",
			  ALT   =>'Show All Data Granules on a Time Line');
  
}

# Print the button for the granule temporal coverage:
sub covertime_menu_btn {
  $_[0] -> msubmit_button(QUERY =>['mode' => 'COVERTIME'],
			  SRC   =>"${IU::DocsUrl}btns/sub-gtime.gif",
			  ALT   =>'Show Data Granules on a Time Line');
}

# Print the button for adding to the shopping cart:
sub cartadd_menu_btn {
  $_[0] -> msubmit_button(QUERY =>['mode'=>'CART', 'submode'=>'ADD'],
			  SRC   =>"${IU::DocsUrl}btns/sub-cartadd.gif",
			  ALT   =>'Add to Shopping Cart');
}

# Print the button for adding granules from data sets to the shopping cart:
sub dset_to_cartlist_menu_btn {
  $_[0] -> msubmit_button(QUERY =>['mode'=>'CART', 'submode'=>'ADD',
				   'addfrom'=>'DSET'],
			  SRC   =>"${IU::DocsUrl}btns/sub-cartaddfrom.gif",
			  ALT   =>'Add to Shopping Cart');
}

# Print the button for adding granules from data sets to the folder:
sub dset_to_folderlist_menu_btn {
  $_[0] -> msubmit_button(QUERY =>['mode'=>'WORKBENCH', 'submode'=>'ADD',
				   'addfrom'=>'DSET'],
			  SRC   =>"${IU::DocsUrl}btns/sub-folderaddfrom.gif",
			  ALT   =>'Add to Folder');
}

# Print the button for adding to the shopping cart:
sub cartdsetadd_menu_btn {
  my $self = shift;
  my $key_value_list_ref = shift;
  
  if ($key_value_list_ref) {
    $self -> msubmit_button(QUERY =>['mode'=>'CART',
				     'submode'=>'DSETADD',
				     @$key_value_list_ref],
			    SRC   =>"${IU::DocsUrl}btns/sub-cartadd.gif",
			    ALT   =>'Add Data Set to Shopping Cart');
  } else {
    
    $self -> msubmit_button(QUERY =>['mode'=>'CART', 'submode'=>'DSETADD'],
			    SRC   =>"${IU::DocsUrl}btns/sub-cartadd.gif",
			    ALT   =>'Add Data Set to Shopping Cart');
  }
}

# Print the button for creating an Inventory Search:
sub create_search_menu_btn {
  $_[0] -> msubmit_button(QUERY =>['mode'=>'MAINSRCH', 
                                   'submode'=>'UPDATE',
                                   'SrchTypeevent'=>'DSET2INV'],
			  SRC   =>"${IU::DocsUrl}btns/sub-search.gif",
			  ALT   =>'Create Primary Search');
}

# Print the button for removing from the shopping cart:
sub cartrem_menu_btn {
  my $self = shift;
  my $key_value_list_ref = shift;
  
  if ($key_value_list_ref) {
    $self -> msubmit_button(QUERY =>['mode'=>'CART',
				     'submode'=>'REM',
				     @$key_value_list_ref],
			    SRC   =>"${IU::DocsUrl}btns/sub-cartrem.gif",
			    ALT   =>'Remove from Shopping Cart');
  } else {
    $self -> msubmit_button(QUERY =>['mode'=>'CART', 'submode'=>'REM'],
			    SRC   =>"${IU::DocsUrl}btns/sub-cartrem.gif",
			    ALT   =>'Remove from Shopping Cart');
  }
}

# Print the button for removing *everything* from the shopping cart:
sub cartempty_menu_btn {
  $_[0] -> msubmit_button(QUERY =>['mode'=>'CART', 'submode'=>'REM',
				   'remove'=>'ALL'],
			  SRC   =>"${IU::DocsUrl}btns/sub-cartempty.gif",
			  ALT   =>'Empty Shopping Cart');
}

# Print the button for ordering:
sub order_menu_btn {
  $_[0] -> msubmit_button(QUERY =>['mode'=>'ORDER'],
			  SRC   =>"${IU::DocsUrl}btns/sub-order.gif",
			  ALT   =>'Order');
}

# Print the button for adding to the shopping cart:
sub save_results_menu_btn {
  $_[0] -> msubmit_button(QUERY => ['mode'    => 'DSETLIST',
				    'submode' => 'SAVE_DB_TXT'],
			  SRC   =>"${IU::DocsUrl}btns/sub-savedisk.gif",
			  ALT   =>'Save Data as Text File');
}
#------------------------------------------------------------
# forgotten_password [Modal=>1]
#------------------------------------------------------------
# Print a "Forgot your password?" link.
# Specify "Modal=>1" to turn off the links in the navigator.
sub forgotten_password {
  my ($self,%params) = @_;
  $params{'mode'} = 'FORGOT';

  # This looks a bit peculiar; it's because we have to invert 
  # the condition. The INLINE thing has to be supplied to turn
  # the navigator on.
  if (defined $params{'Modal'} && $params{'Modal'}) {
    delete $params{'Modal'};
  }
  else {
    $params{'submode'} = 'INLINE';
  }

  # Generate the URL.
  $self -> A(HREF=>IW::Util::script_url(%params))
    -> t_B('(Forgotten your password?)')
      -> _A;
}

#------------------------------------------------------------
# tr_color [COLOR]
#------------------------------------------------------------
# Set the <TR BGCOLOR> attribute for table rows created with the "tabulate"
# methods.  No other TR tags will be affected.  
#
sub tr_color {
  my ($self, $color) = @_;
  $self->{TRColor} = $color if (@_ > 1);
  ($self->{TRColor} ? (BGCOLOR=>$self->{TRColor}) : ()); 
}

#------------------------------------------------------------
# tabulate_one PARAMHASH
#------------------------------------------------------------
# Print a simple definition list entry.
#
# NOT CHAINABLE!  Returns whether or not something was actually printed!
#
# ARGUMENTS
#    Attr      => the attribute to be displayed.
#    Value     => the value, *without* markup (i.e., the raw text).
#    ValueHTML => the value, *with* HTML markup (optional)
#    Type      => the type for brief/detailed info and defn bts (Value req'd!)
#    Entry     => the MD entry
#    DIF       => create a link to the DIF for the value (Value req'd!)
#    PRE       => Puts <PRE> tags around the text
#    BR        => Puts <BR> tags wherever there's a newline in the next

sub tabulate_one {
  my $self = shift;
  my %params = @_;
  my $attr      = $params{Attr};
  my $value     = $params{Value};
  my $entry     = $params{Entry};
  
  # Return immediately if no work to be done:
  (defined($value) && ($value ne '')) || return 0;
  
  # The attribute:
  $self -> TR($self->tr_color) -> TD(VALIGN=>'TOP') -> t_B($attr) 
    -> _TD -> TD(VALIGN=>'TOP');
  
  # The (optional) definition, brief info, and detailed info links:
  if ($params{Type}) { 
    $self -> defn_btn($params{Type}, $value);	
    #	$self -> binfo_btn_to(IW::Util::supp_url($params{Type}, $value, $entry));
    $self -> binfo_btn_to(IW::Util::gcmd_url($params{Type}, $value, $entry));
    $self -> dinfo_btn_to(IW::Util::guide_url($params{Type}, $value));
  }
  if ($params{DIF}) {
    $self -> binfo_btn_to(IW::Util::dif_url($value));
  }
  $self -> hspace(1) if ($params{Type}||$params{DIF});   # little space
  
  # Create the HTML for the value:
  my $valueHTML = $params{ValueHTML} || HTML::Stream::html_escape($value);
  $valueHTML =~ s/\n/<BR>/g  if $params{BR};
  
  # The value (possibly PREformatted text):
  $self->PRE  if $params{PRE};
  $self->io->print($valueHTML);
  $self->_PRE if $params{PRE};
  $self->_TD;
  1;
}

#------------------------------------------------------------
# tabulate_comments PARAMHASH
#------------------------------------------------------------
# Print comments in a table row for display on the gran or data sets 
# attributes.
# Each value in the array is output as separate line.
# Embedded newlines are removed. 
#  
# Returns whether or not something was actually printed.
#
#     Attr   => Label or keyword diplayed to the left of the text area.
#     Values => a reference to an array of values, each *without* markup
#
sub tabulate_comments {
  my $self = shift;
  my %params = @_;

  my $values = ($params{Values} || []);
  my $wrap = ($params{Wrap} || '80');
  # Return immediately if nothing to be done:
  @{$values} or return 0;

  # Output attribute:
  $self -> TR($self->tr_color)
    -> TD(VALIGN=>'TOP') -> t_B($params{Attr})
      -> _TD -> TD(VALIGN=>'TOP');

  # Output each item:
  my $n = int(@{$values});
  my $i;
  # 
  for ($i = 0; $i < $n; $i++) {
    $self->tag('NOBR');
    my $value = $values->[$i];
    $value =~ s/(\\n)+//g;                 # remove embedded newlines
    $value =~ s/\x1//; # remove the character which prints as a little box in 
                       # HTML. It is ASCII 01 or hex 1 (start of heading).    
    $self ->hspace(1) -> t($value);
    $self->tag('/NOBR');
    $self -> BR if ($i < ($n-1));
  }
  $self->_TD;
  1;
}
#------------------------------------------------------------
# tabulate_many PARAMHASH
#------------------------------------------------------------
# Print a definition list entry, whose value is a sequence.
# Returns whether or not something was actually printed.
#
# Arguments are the same as in &PrintDefn, except that Value is
# replaced by...
#
#     Values => a reference to an array of values, each *without* markup 
#     NoSort => suppresses the automatic sorting of elements
#
sub tabulate_many {
  my $self = shift;
  my %params = @_;
  
  my $values = ($params{Values} || []);
  my $entry     = $params{Entry};
  my ($href, $i);
  
  # Return immediately if nothing to be done:
  @{$values} or return 0;
  
  # Sort values:
  $values = [ sort @$values ] unless $params{NoSort};
  
  # Output attribute:
  $self -> TR($self->tr_color)
    -> TD(VALIGN=>'TOP') -> t_B($params{Attr})
      -> _TD -> TD(VALIGN=>'TOP');
  
  # Output each item:
  my $n = int(@{$values});
  for ($i = 0; $i < $n; $i++) {
    $self->tag('NOBR');
    my $value = $values->[$i];
    $value =~ s/(\\n)+//g;                 # remove embedded newlines
    my $len = length($value);
    if ($len != 0) {
      # The (optional) definition, brief info, and detailed info links:
      if ($params{Type}) { 
	$self -> defn_btn($params{Type}, $value);	
	# $self -> binfo_btn_to(IW::Util::supp_url($params{Type},$value));
	$self -> binfo_btn_to(IW::Util::gcmd_url($params{Type}, 
						 $value, $entry));
	$self -> dinfo_btn_to(IW::Util::guide_url($params{Type}, $value));
      }
    }
    else {
      $self -> defn_btn(undef, $value);	
      $self -> binfo_btn_to(undef, $value);
      $self -> dinfo_btn_to(undef, $value);       
    }
    if ($params{DIF}) {
      $self -> binfo_btn_to(IW::Util::dif_url($value));
    }
    
    # The value:
    $self ->hspace(1) -> t($value);
    $self->tag('/NOBR');
    $self -> BR if ($i < ($n-1));
  }
  $self->_TD;
  1;
}

#------------------------------------------------------------
# tabulate_text ATTR,VALUE,[PREFORMATTED]
#------------------------------------------------------------
# Print a definition list entry, whose value is multiline text.
#
# NOT CHAINABLE!  Returns whether or not something was actually printed!
# If a 3rd argument is present, then the text will be displayed as 
# preformatted text.

sub tabulate_text {
  my ($self, $attr, $value, $preformatted) = @_;
  defined($value) || ($value = '');
  
  # First, get rid of leading whitespace:
  $value =~ s/^\s*//g;
  if ($value ne '') {
    $self -> TR($self->tr_color)
      -> TD(VALIGN=>'TOP') -> t_B($attr) -> _TD -> TD(VALIGN=>'TOP');
    
    $value = HTML::Stream::html_escape($value);
    
    if ($preformatted) {
      $self-> PRE -> t($value) -> _PRE;
    }
    else {
      # Put <br>s at newlines *only* if... neither the preceding nor
      # following character is lowercase or whitespace:
      $value =~ s/([^\sa-z])\n([^\sa-z])/$1<br>\n$2/g;
      
      # Convert multiple embedded newline escapes to paragraphs:
      $value =~ s/(\\n){2}/<p>/g;
      
      # Convert single embedded newline escapes to breaks:
      $value =~ s/(\\n)/<br>/g;
      
      $self->io->print($value);
    }
    return 1;
  }
  0;
}

#------------------------------------------------------------
1;
