package IW::MediaOrderLimit;

=head1 NAME

IW::MediaOrderLimit - check media order limits

=head1 DESCRIPTION

This class evaluates the total order size with the media limit.

=head2 DATA

=begin list

=item

%UserOptionSelections

This is a class variable (not an instance variable).  It is a hash of hashes.
To make the key unique, the key for this hash will be the combination of 
Granule Id/Daac. The value of the hash will be a reference to an array 
containing the size of the granule and the media type selected for the 
granule in this option.

Here is an example of how the UserOptionSelections hash might look:

%UserOptionSelections =
   (
     G1/STUB => [ GranSize => 1.9M,  TypeId => 8MM],
     G2/GSFC => [ GranSize => 2.0M,  TypeId => 8MM],
     G3/STUB => [ GranSize => 18.7M, TypeId => 8MM],
     G4/LARC => [ GranSize => 200M,  TypeId => CDROM]
   );

=item

%DaacMediaSizes

This is a class variable (not an instance variable).  It is a hash of hashes.
The key to the hash will be the combination of Daac/TypeId, the value will 
be a reference to a hash containing either key Maximum Size or Maximum 
Granules. The values for the keys will be either Maximum Size that can 
fit in the media or the Maximum Granules limit for that media. 

Here is an example of how the DaacMediaSizes hash might look:

%DaacMediaSizes =
 ( STUB/8MM   =>  [ MaxSize  => 700MB ],
   GSFC/8MM   =>  [ MaxGrans =>   999 ],
   LARC/DLT   =>  [ MaxGrans =>   876 ],
   LARC/CDROM =>  [ MaxSize  => 600MB ]
 );

=item

%MediaPackagingTable

This is a class variable (not an instance variable).  It is a hash of arrays.

The hash will have as keys the combination of Daac/TypeId, using the keys 
is how the software will match the values of this hash with the values in 
the DaacMediaSizes hash. The values of the hash will be a references to 
the arrays containing the user options selections for the media.

Here is an example of how the DaacMediaSizes hash might look:

%MediaPackagingTable =      
  (  GSFC/8MM   => [G2/GSFC, G5/GSFC, G6/GSFC],
     STUB/8MM   => [G1/STUB, G3/STUB],
     LARC/CDROM => [G4/LARC]
  );

=item

$self

This is the the usual class reference.

=end list

=head2 METHODS

=begin list

=item

new ( ) . . . Constructor.

Create reference $self.  
Bless the $self reference into this class.
Clear all hashes in this class.

=end list

=cut

#=====================================================================

use strict;
use IU;
use IU::Dataset;
use IU::Granule;
use IU::Log;
use IU::MediaMaximumCapacity;
use IW::App;

#---------------------------------------------------------------------
# new
#---------------------------------------------------------------------
# Constructor class method.  Creates a new MediaOrderLimit.
# Invoked when create media order limit:  $var = new IW::MediaOrderLimit;
# 
# Before the constructor is invoked, we need to try to get the MediaOrderLimit
# information out of the CART database, if it does not exist, then we create
# a new class.  Eventually this class has to be stored in the CART.
#

sub new {
   my ($class, %params) = @_;           # Get class.

   my $self = {};
   bless $self, $class;       # Create the object.

   #-----------------------------------------------
   # Call method clear to create and initialize the 
   # hashes in this class, then return itself.
   #-----------------------------------------------
   $self->clear();
   return $self;
}

#---------------------------------------------------------------------
# clear
#---------------------------------------------------------------------
# Subroutine clear sets all of the hashes in this class to null.
#
# The constructor uses this method to create the needed hashes 
# for the class.
#
# This method should also be called everytime the cart is emptied.
#

sub clear {
   my $self = shift;

   # Create and/or initialize the three hashes in this class.
   $self->{DaacMediaSizes} = {};
   $self->{UserOptionSelections} = {};
   $self->{MediaPackagingTable} = {};
}

#---------------------------------------------------------------------
# add
#---------------------------------------------------------------------
# Subroutine add updates the three hashes in this class. 
# It adds elements if they have not been added.  
#
# This subroutine is called when a user selects or changes
# ordering options for items inside the cart.
# Existing values for the same keys will be overritten in 
# all three hashes, we want to do this in case users change
# the ordering options.
# 
# Input Parameters:
#    $dset:      Dataset object
#    $item:      Granule or Dataset object
#    $mediatype: Selected media type for ordering

sub add {
   my ($self, $dset, $item, $mediatype) = @_;

   my $mediamaximumcapacity;
   my $maxvalue;
   my $maxtype;
   my $limits;

   #------------------------------------
   # Get the dataset id and the archive id 
   # from the granule/dataset object
   #------------------------------------
   my $dsetname = $item->dataset_id;
   my $daac = $item->archive_id;

   #debug "Subroutine add, mediatype is $mediatype, dataset name is $dsetname, daac is $daac.\n";

   # Flag to keep track that the media has a limit.
   my $found_flag = 0;

   #------------------------------------
   # If the media has a limit, get mediatype, maxtype, and maxvalue.
   #
   # Get all the media types with limits for the dataset to see
   # if the selected media has a limit.
   #
   # Only continue processing if the selected media for the
   # item has a limit associated in the dataset. 
   #
   # If the selected media has no limit, skip it.  
   # It will be considered unlimited.
   #------------------------------------
   foreach $mediamaximumcapacity (@{$dset->mediamaximumcapacity}) {
      next if ($mediatype ne $mediamaximumcapacity->type);   

      # The selected media has a limit, set the found flag on.
      $found_flag = 1;

      # Selected media has a limit, find out the type of limit it has.
      #------------------------------------
      # If the limit is in size (REQUEST_SIZE_MB), get the 
      # size limit and set the type to MaxSize.
      #------------------------------------
      if (defined ($mediamaximumcapacity->size_mb)) {
         $maxvalue = $mediamaximumcapacity->size_mb;
         $maxtype = "MaxSize";
         #debug "size_mb: maxvalue is $maxvalue, maxtype is $maxtype.\n";

         #------------------------------------
         # Check for the media limit to see if the granule will fit.
         # If limits are founded, return the limit information.
         #------------------------------------
         $limits = $self->check_limit ($item, $mediamaximumcapacity);
         return $limits if ($limits);
         $self->update_daac_media ($mediatype, $item, $maxtype, $maxvalue);
      } 

      #------------------------------------
      # If the limit is in number of granules (REQUEST_SIZE_GRANULES), 
      # get the granule limit and set the type to MaxGrans.
      #------------------------------------
      if (defined ($mediamaximumcapacity->size_granules)) {
         $maxvalue = $mediamaximumcapacity->size_granules;
         $maxtype = "MaxGrans";
         #debug "size_granules: maxvalue is $maxvalue, maxtype is $maxtype.\n";

         #------------------------------------
         # Check for the media limit to see if the granule will fit.
         # If limits are founded, return the limit information.
         #------------------------------------
         $limits = $self->check_limit ($item, $mediamaximumcapacity);
         return $limits if ($limits);
         $self->update_daac_media ($mediatype, $item, $maxtype, $maxvalue);
      }

      #------------------------------------
      # If the limit is the default capacity (MEDIA_MB_CAPACITY), get the 
      # capacity limit and set the type to MaxSize.
      #------------------------------------
      # Commented out, this limit should only be use per individual granules,
      # we do not want to apply limit MEDIA_MB_CAPACITY to the complete order.
      #------------------------------------
      #elsif (defined ($mediamaximumcapacity->capacity)) {
      #   $maxvalue = $mediamaximumcapacity->capacity;   
      #   $maxtype = "MaxSize";
      #   #debug "capacity: maxvalue is $maxvalue, maxtype is $maxtype.\n";
      #}
      #------------------------------------

      #------------------------------------
      # We have all the needed information to update the hashes,
      # call the methods to update the tables providing the 
      # needed information for each table.
      #------------------------------------
      $self->update_user_option ($mediatype, $item);
      $self->update_media_packaging ($mediatype, $item);
   }

   #------------------------------------
   # If the found flag is not set, the media has no limit.  
   # If the media has no limit, we need to remove any 
   # existing settings for this item.
   #------------------------------------
   if ($found_flag == 0) {
      my @items = ($item);
      $self->remove (\@items);
      return;
   }
}

#---------------------------------------------------------------------
# remove
#---------------------------------------------------------------------
# Subroutine remove deletes entries from two hashes in this class,
# UserOptionSelections and MediaPackagingTable.
#
# This subroutine is called when a user removes items with
# selected ordering options from the cart.
#
# Input Parameters:
#    $items: Array of Granules or Datasets objects

sub remove {
   my ($self, $items) = @_;

   my $item;
   my $value;

   # Get the needed information from every item we want to remove.
   # Once we have the needed info, go ahead and update the hashes.
   foreach $item (@$items) {
      #------------------------------------
      # Need to get the key for hash UserOptionSelections.
      # The value of the key is the combination granid/daac
      #------------------------------------
      $value = $item->id . "/" . $item->archive_id;

      # update UserOptionSelections hash
      delete $self->{UserOptionSelections}->{$value};

      # Update MediaPackagingTable hash.
      $self->remove_from_media_packaging ($value);
   }
}

#---------------------------------------------------------------------
# remove
#---------------------------------------------------------------------
# Subroutine remove_from_media_packaging deletes values
# from MediaPackagingTable hash.
#
# This subroutine is called during two scenarios:
# 1. When a user removes items with selected ordering options from the cart.
# 2. When a user adds items, the reasons we call it for adding items are:
#    a) If the value exists in the table it needs to be removed before
#       it is added to avoid duplicate values.
#    b) If the user selected a different media for the item, the current
#       value needs to be deleted to add value containing the new media.
#
# Input Parameters:
#    $value: Value to be removed from the hash

sub remove_from_media_packaging {
   my ($self, $value) = @_;

   my $key;
   my $listref;
   my @newlist;
   my $listvalue;

   #------------------------------------
   # Update MediaPackagingTable hash.
   # The combination granid/daac is a value existing
   # within each key in the MediaPackagingTable hash.
   # We need to remove it out of the hash.
   #------------------------------------
   # To update the MediaPackagingTable hash do the following:
   # 1. Get each key in the hash.
   # 2. Get the reference to the list associated with the key.
   # 3. Filter out current value from list associated with the key.
   # 4. Update each key with the new filtered list.
   #------------------------------------
   foreach $key (keys %{$self->{MediaPackagingTable}}) {
      # Get the reference to the array.
      $listref = $self->{MediaPackagingTable}->{$key};

      # If the reference is not null continue.
      if ($listref) {
         # Clear the temp work array.
         @newlist = ();

         # Set the work array with the current list, use
         # grep to remove the item the user wants to delete
         @newlist = ( grep { $_ ne $value } @$listref );

         # Clear the array stored in the hash.
         @$listref = ();

         # Set the array in the hash with the new values.
         @$listref = @newlist;
         $self->{MediaPackagingTable}->{$key} = $listref;

         #------------------------------------
         #debug "Values for MediaPackagingTable key $key are: \n";
         #my $temp;
         #foreach $temp (@newlist) {
         #   debug "$temp \n";
         #}
         #------------------------------------
      }
   }
}

#---------------------------------------------------------------------
# check_limit
#---------------------------------------------------------------------
# Given a MediaMaximumCapacity, and a Granule or Dataset object, 
# subroutine check_limit figures out if the selected media type
# is big enough to be used for ordering the granule.
# If a limit is encountered, return the information in a hash.
#
# Input Parameters:
#    $item:                 Granule or Dataset object
#    $mediamaximumcapacity: MediaMaximumCapacity object
#
# Output:
#    undef:   If no limits are found it must return undef
#      or
#    $limits: Reference to a hash containing media limit information
#    Returned hash should like like this:
#        Fail       -> 1  (Currently not used, added in case we need a flag)
#        ItemId     -> "granule GRANULE_12345"  or  "dataset DATASET_12345"
#        DataCenter -> "DAAC"
#        MediaLimit -> "000 granules"  or  "000 MB"
#        MediaType  -> "CDROM"
#

sub check_limit {
   my ($self, $item, $mediamaximumcapacity) = @_;

   #------------------------------------
   # Get the information we need from the input
   # parameters to create a string that might be
   # a MediaPackagingTable key.
   #------------------------------------
   my $archive = $item->archive_id;
   my $mediatype = $mediamaximumcapacity->type;
   my $selection = $archive . "/" . $mediatype;
   
   #debug "Selected media type is $selection"; 

   my $type;
   my $typesize;
   my $typegrans;
   my $limits;

   #------------------------------------
   # Check every key from the MediaPackagingTable hash.
   #------------------------------------
   foreach $type (keys %{$self->{MediaPackagingTable}}) {

      #------------------------------------
      # Found one that matches current selection.
      # Need to check if new item will fit.
      #------------------------------------
      if ($type eq $selection) {

         #------------------------------------
         # Get the limit type from the DaacMediaSizes.
         #------------------------------------
         $typesize = defined($self->{DaacMediaSizes}->{$type}->{MaxSize}) ?
                   $self->{DaacMediaSizes}->{$type}->{MaxSize} : "";
         $typegrans = defined($self->{DaacMediaSizes}->{$type}->{MaxGrans}) ?
                   $self->{DaacMediaSizes}->{$type}->{MaxGrans} : "";
         #debug "typesize is $typesize, typegrans is $typegrans";

         #------------------------------------
         # Check for limit type, if MaxGrans value is not undef
         # we have a value, therefore is a granule limit.
         #------------------------------------
         if ($typegrans ne "") {
            my $listref = $self->{MediaPackagingTable}->{$type};

            #------------------------------------
            # Count how many granules in the MediaPackagingTable
            # are selected for this media type.  If the number is
            # grater or equual to the media granule limit, the new granule
            # granule does fits so we need to set the limit info.
            #------------------------------------
            if (scalar(@$listref) >= $typegrans) {
               #------------------------------------
               # We found a limit, no more granules can be added
               # to this media type.  Create a hash containing
               # the granule number limit information.
               #------------------------------------

               # Need item type (granule or dataset) for the id string.
               my $itemtype = 
                  ($item->id eq $item->dataset_id) ? "dataset " : "granule "; 

               # Create the hash with the granule limit information.
               $limits->{Fail} = 1;
               $limits->{ItemId} = $itemtype . $item->id;
               $limits->{DataCenter} = $archive;
               $limits->{MediaLimit} = $typegrans . " granules";
               $limits->{MediaType} = $mediatype;

               # return reference to hash containing the limit information
               return $limits;
            }
         }  # End if ($typegrans ne "")

         #------------------------------------
         # Before checking for a size limit, check for granule
         # size. If no granule size is found, return because
         # the granule size cannot be processed.
         #------------------------------------
         return undef unless $item->size_mbecs_granule;

         #------------------------------------
         # Check for limit type, if MaxSize value is not undef
         # we have a value, therefore is a MB size limit.
         #------------------------------------
         if ($typesize ne "") {
            #------------------------------------
            # Get the granule size to make sure that it will fit
            # with other granules already selected for the media.
            # Start cumulative size with granule size.
            #------------------------------------
            my $item_size = $item->size_mbecs_granule;
            my $cumulative_size = $item_size;
            my $selected;

            #------------------------------------
            # To check the media limit:
            # 1. Get every item in the MediaPackagingTable with
            #    the option of the media type being measured.
            # 2. Get the size of every granule currently selected
            #    from the UserOptionSelections.
            # 3. Add the granule size to the cumulative size.
            # 4. Check total size for limit treshold.
            #------------------------------------
            my $listref = $self->{MediaPackagingTable}->{$type};
            foreach $selected (@$listref) {
               $item_size = 
                  $self->{UserOptionSelections}->{$selected}->{GranSize};
               $cumulative_size = $item_size + $cumulative_size;
            }
 
            # Check if cumulative size is greater than media limit.
            if ($cumulative_size > $typesize) {
               #------------------------------------
               # Found a limit, this granule can not be added
               # to this media type.  Create a hash containing
               # the MB size limit information.
               #------------------------------------
   
               # Need item type (granule or dataset) for the id string.
               my $itemtype = 
                  ($item->id eq $item->dataset_id) ? "dataset " : "granule "; 

               # Create the hash with the MB size limit information.
               $limits->{Fail} = 1;
               $limits->{ItemId} = $itemtype . $item->id;
               $limits->{DataCenter} = $archive;
               $limits->{MediaLimit} = $typesize . " MB";
               $limits->{MediaType} = $mediatype;

               # return reference to hash containing the limit information
               return $limits;
            }
         }

      }  # End if ($type eq $selection)

   }  # End foreach key in MediaPackagingTable

   # Fell out of the for loop, no limits found, must return undef.
   return undef;
}

#---------------------------------------------------------------------
# update_daac_media
#---------------------------------------------------------------------
# Subroutine update_daac_media update hash DaacMediaSizes.
# It adds elements if they have not been added, like this:
# 
# %DaacMediaSizes = ( Daac/TypeId => [ MaxType => MaxValue]  );
#
# This subroutine is called when a user selects or changes
# ordering options for items inside the cart.
# Existing values for the same keys will be overritten in
# all three hashes, we want to do this in case users change
# the ordering options.
#
# Input Parameters:
#    $mediatype: String, type id of media selected for ordering
#    $item:      Granule or Dataset object
#    $maxtype:   String, limit type, values are "MaxSize" or "MaxGrans"
#    $maxvalue:  Media limit, it could be be size or number of granules.

sub update_daac_media {
   my ($self, $mediatype, $item, $maxtype, $maxvalue) = @_;

   #----------------------------------------------
   # Get the daac from the granule or dataset object,
   # concatenate it with the media type to get the 
   # key for the hash.
   #----------------------------------------------
   my $key = $item->archive_id . "/" . $mediatype;

   # update DaacMediaSizes hash
   $self->{DaacMediaSizes}->{$key}->{$maxtype} = $maxvalue; 

   #debug "DaacMediaSizes table with key $key $maxtype is $maxvalue\n";

}

#---------------------------------------------------------------------
# update_user_option
#---------------------------------------------------------------------
# Subroutine update_user_option update hash UserOptionSelections.
# It adds elements if they have not been added, like this:
#
# %UserOptionSelections = 
#    ( GranuleId/Daac => [ "GranSize" => Size,  "TypeId" => MediaType ]);
#
# This subroutine is called when a user selects or changes
# ordering options for items inside the cart.
# Existing values for the same keys will be overritten in
# all three hashes, we want to do this in case users change
# the ordering options.
#
# Input Parameters:
#    $mediatype: String, type id of media selected for ordering
#    $item:      Granule or Dataset object

sub update_user_option {
   my ($self, $mediatype, $item) = @_;

   #----------------------------------------------
   # Get the id and daac from the granule or dataset object,
   # concatenate them to get the key for the hash.
   #----------------------------------------------
   my $key = $item->id . "/" . $item->archive_id;

   # update UserOptionSelections hash 
   # Set "GranSize" using the granule size.
   $self->{UserOptionSelections}->{$key}->{"GranSize"} = 
                                                $item->size_mbecs_granule; 
   # Set "TypeId" using the media type.
   $self->{UserOptionSelections}->{$key}->{"TypeId"} = $mediatype; 
}

#---------------------------------------------------------------------
# update_media_packaging
#---------------------------------------------------------------------
# Subroutine update_media_packaging updates hash MediaPackagingTable.
# It adds elements if they have not been added, like this:
#
# %MediaPackagingTable = 
#         ( Daac/TypeId => [GranId/Daac, GranId/Daac, GranId/Daac]  );
#
# This subroutine is called when a user selects or changes
# ordering options for items inside the cart.
# Existing values for the same keys will be overritten in
# all three hashes, we want to do this in case users change
# the ordering options.
#
# Input Parameters:
#    $mediatype: String, type id of media selected for ordering
#    $item:      Granule or Dataset object

sub update_media_packaging {
   my ($self, $mediatype, $item) = @_;

   #----------------------------------------------
   # Get the daac from the granule or dataset object,
   # concatenate it with the media type to get the
   # key for the hash.
   #----------------------------------------------
   my $key = $item->archive_id . "/" . $mediatype;

   #----------------------------------------------
   # Get the id and daac from the granule or dataset object,
   # concatenate them to get the value to be added.
   #----------------------------------------------
   my $value = $item->id . "/" . $item->archive_id;
   
   #----------------------------------------------
   # Get the reference to the array stored with the new key.
   # If this is a new key and there is no array reference, it
   # will create a new one. 
   #----------------------------------------------
   my $newlist = $self->{MediaPackagingTable}->{$key};

   #----------------------------------------------
   # Add the new value to the array referenced by the key.
   # If the value already exists in the array, remove it
   # before adding it to avoid duplicates.
   #----------------------------------------------
   $self->remove_from_media_packaging ($value);
   push (@$newlist, $value);

   #----------------------------------------------
   # Update MediaPackagingTable hash.
   # Making sure the hash with the key points to the
   # reference of the array with the new value added.
   #----------------------------------------------
   $self->{MediaPackagingTable}->{$key} = $newlist;

   #----------------------------------------------
   # Debug:
   #debug "MediaPackagingTable table has key $key with values\n";
   #foreach $value (@$newlist) {
   #   debug "$value\n";
   #}
   #----------------------------------------------
}

#---------------------------------------------------------------------
# print_debug
#---------------------------------------------------------------------
# Subroutine print_debug is only used for debugging purpose.
# All it does is print the contents of the three main hash tables
# for this class.
#
sub print_debug {
   my $self = shift;

   debug "PRINT_DEBUG: Entering MediaOrderLimit print_debug method.";

   my $mainkey;
   my $key;
   my $value;
   my $valuesref;
   my @values;
   my $i;

   debug "PRINT_DEBUG: Contents of UserOptionSelections Hash:";
   foreach $mainkey ( keys %{$self->{UserOptionSelections}} ) {
      debug "For Main key: $mainkey";
      $key = "GranSize";
      $value = $self->{UserOptionSelections}->{$mainkey}->{$key};
      debug "Key is $key, Value is $value\n";
      $key = "TypeId";
      $value = $self->{UserOptionSelections}->{$mainkey}->{$key};
      debug "Key is $key, Value is $value\n";
      debug "\n";
   }
   debug "PRINT_DEBUG: Finish contents of UserOptionSelections Hash:";
   debug "\n";

   debug "PRINT_DEBUG: Contents of DaacMediaSizes Hash:";
   foreach $mainkey ( keys %{$self->{DaacMediaSizes}} ) {
      debug "For Main key: $mainkey";
      foreach $key (keys %{$self->{DaacMediaSizes}->{$mainkey}}) {
         $value = $self->{DaacMediaSizes}->{$mainkey}->{$key};
         debug "Key/Value: $key / $value";
      }
      debug "\n";
   }
   debug "PRINT_DEBUG: Finish contents of DaacMediaSizes Hash:";
   debug "\n";

   debug "PRINT_DEBUG: Contents of MediaPackagingTable Hash:";
   foreach $mainkey ( keys %{$self->{MediaPackagingTable}} ) {
      debug "For Main key: $mainkey";
      debug "The content is:";
      $valuesref = $self->{MediaPackagingTable}->{$mainkey};
      if ($valuesref) {
         foreach $value (@$valuesref) {
            debug "$value";
         }
      }
      debug "\n";
   }
   debug "PRINT_DEBUG: Finish contents of MediaPackagingTable Hash:";
   debug "\n";

   debug "PRINT_DEBUG: Exiting MediaOrderLimit print_debug method.";
}

#-------------------- end of MediaOrderLimit --------------------
1;
