package IW::Minitable;

=head1 NAME

IW::Minitable - tabulated information in dataset/granule list

=cut

use IW::App;
use HTML::Stream qw(:funcs);


#------------------------------------------------------------
#
# Class (static) methods
#
#------------------------------------------------------------

#------------------------------------------------------------
# new
#------------------------------------------------------------
sub new {
    my $class = shift;
    my $new = {};
  
    $new->{HdrList} = [];     # header info for columns, in order
    $new->{Hdrs} = {};        # header info, indexed by name
    $new->{Data} = {};        # data for a single row, indexed by header name
    $new->{Small} = 1;        # default is small    
    $new->{Type} = 'HORIZ';
    bless $new, $class;
}


#------------------------------------------------------------
#
# Instance (virtual) methods
#
#------------------------------------------------------------

#------------------------------------------------------------
# num_headers -- return number of headers
#------------------------------------------------------------

sub num_headers {
    my $self = shift;
    int(@{$self->{HdrList}});
}


#------------------------------------------------------------
# set_small -- declare whether you want this small or not
#------------------------------------------------------------

sub set_small {
    my $self = shift;
    my $setting = shift;
    $self->{Small} = $setting;
}

#------------------------------------------------------------
# add_header -- add a new header
#------------------------------------------------------------
#
# ARGUMENTS
#    Name     => the name, as bare ASCII text (for <pre> tables & lookup)
#    NameHtml => optional: the name as HTML markup.  Use this only if we can.
#    Href     => the only reliable way to make this name a hyperlink,
#                across all types of table output.  Use this in preference
#                to embedding a hyperlink in the NameHtml markup.
#
sub add_header {
    my $self = shift;
    my %params = @_;

    # Generate the HTML-safe name, if needed:
    defined($params{NameHtml}) ||   
	($params{NameHtml} = html_escape($params{Name}));
    
    # Wrap it in a hyperlink, if desired:
    if (defined($params{Href})) {
	$params{NameHtml} = 
	    qq!<a href="$params{Href}">$params{NameHtml}</a>!;
    }
    
    # Save it:
    $self->{Hdrs}{$params{Name}} = \%params;
    push(@{$self->{HdrList}}, \%params);
    1;
}

#------------------------------------------------------------
# clear -- wipe existing data values, and begin new ones
#------------------------------------------------------------

sub clear {
    my $self = shift;
    $self->{Data} = {};
    1;
}

#------------------------------------------------------------
# set -- set the value for a column
#------------------------------------------------------------
#
# ARGUMENTS
#    Name    => the name of the header shose value we are setting
#    Val     => the value, as bare ASCII text
#    ValHtml => optional: the value as HTML markup

sub set {
    my $self = shift;
    my %params = @_;
    my $name = $params{Name};
    defined($params{Val}) || ($params{Val} = '');

    # Create HTML-safe version of value:
    defined($params{ValHtml}) ||    # use ValHtml if given
	($params{ValHtml} = html_escape($params{Val}));

    # Store it:
    $self->{Data}{$name} = \%params;
}


#------------------------------------------------------------
# print_headers -- print the (horizontal) table headers only
#------------------------------------------------------------
# NOTE: prints ONLY the <th> fields (no <table>...</table>).

sub print_headers {
    my $self = shift;

    foreach $hdr (@{$self->{HdrList}}) {
	print "<TH ALIGN=LEFT>";
	$HTML -> small      if $self->{Small};
	print "<B>", $hdr->{NameHtml}, "</B>";
	$HTML -> _small     if $self->{Small};
	print "</TH>\n";
    }
    1;
}

#------------------------------------------------------------
# print_data -- print the (horizontal) table data only
#------------------------------------------------------------
# NOTE: if HTML3 tables are supported, prints ONLY the <td> fields 
# (no <table>...</table>).

sub print_data {
    my $self = shift;

    foreach $hdr (@{$self->{HdrList}}) {
	print '<TD ALIGN=LEFT>';
	$HTML -> small      if $self->{Small};
	print $self->{Data}{$hdr->{Name}}{ValHtml}, 
	$HTML -> _small     if $self->{Small};
	print "</TD>\n";
    }
    1;
}

#------------------------------------------------------------
# print_full -- print the table row, as used by dataset/granule lists
#------------------------------------------------------------

sub print_full {
    my $self = shift;
    my $i;
     
    # VERTICAL HTML3 TABLE...       
    if ($self->{Type} eq 'VERT') {

	# Keys and values (single cell, one line for each key/value pair):
	print "<td nowrap colspan=10>\n";
	$HTML -> small      if $self->{Small};
	foreach $hdr (@{$self->{HdrList}}) {
	    print '<i>', $hdr->{NameHtml}, ': </i>';
	    $HTML -> hspace(2);
	    print $self->{Data}{$hdr->{Name}}{ValHtml}, "</br>\n";
	}
	$HTML -> _small     if $self->{Small};
    }

    # HORIZONTAL HTML3 TABLE...
    else {

	# Keys & values (two lines in one cell for each key/value pair):
	foreach $hdr (@{$self->{HdrList}}) {
	    print '<td>';
	    $HTML -> small      if $self->{Small};
	    print '<b>', $hdr->{NameHtml}, '</b><br>';
	    print $self->{Data}{$hdr->{Name}}{ValHtml}, "\n";
	    $HTML -> _small     if $self->{Small};
	}
    }
    1;
}

#------------------------------------------------------------
1;

