package IW::Mode;


=head1 NAME

IW::Mode - dispatch activity based on the "mode" CGI tag.


=head1 DESCRIPTION

To support a new mode, give it its own subclass of this class.
The subclass for mode C<MODENAME> should be called C<IW::ModeMODENAME>: 
for example, to support mode C<"WELCOME">, provide an C<IW::ModeWELCOME> 
module.  This module I<must> have a static method C<go()> implemented.

To use this module, just do this:

    use IW::Mode;    
    
    $mode = $REQ->param('mode');     # get the CGI tag
    handle IW::Mode $mode;

An unsupported mode is a fatal error, as is a C<go()> method
that returns a non-true value.


=head1 INSTANCE VARIABLES

None (subclasses will very likely add their own, though).


=head1 NOTES

We use a separate module for each mode so that it has its own
namespace for mode-specific functions.


=head1 AUTHOR

Eryq, Hughes STX Corporation


=head1 RCS INFO

$Id: Mode.pm,v 1.29 2000/02/18 00:13:51 joe Exp $

=cut

use IU;
use IU::Event;
use IU::Log;
use IW::App;

use strict;

#------------------------------------------------------------
# new
#------------------------------------------------------------
# I<Constructor>.

sub new {
   bless {}, shift;
}

#------------------------------------------------------------
# handle MODE
#------------------------------------------------------------
# I<Class method>.  Dispatch the current mode.

sub handle {
   my ($class, $mode, %opts) = @_;
   
   # Extract untainted "substring", demanding all-caps or else:
   $mode = (($mode =~ /^([A-Z]+)$/) ? $1 : undef);  # THIS SYNTAX IMPORTANT!!
   $mode or puke 600, "illegal mode: '$mode' invoked for $class";
   
   # Is this mode supported?
   my $modeclass = "IW::Mode$mode";
   if (!(eval("require $modeclass;"))) {
      puke 602, "module $modeclass not loaded: ", ($@ ? $@ : ''), 
      "is mode '$mode' unsupported?";
   }
   inform "Loaded $modeclass";
   
   my $sid     = $REQ->param('sid');  # session not resumed at this point!
   my $remote_user     = $ENV{REMOTE_USER};
   
   #     debug "MODE.sid == ".$REQ->param('sid');
   # debug "MODE.remote == ".$ENV{REMOTE_USER};
   
   #  Confirm the sid ISN'T blank and REMOTE_USER ISN'T "guest".
   #  Check to see the sid AND the remote_user are the same.
   #  They should be since the session directory CAME from REMOTE_USER.
   #  If NOT go ahead and PUKE. JSA 2-16-98
   # Ps. HATE to do this EVERY time, but the NET is stateless.
   
   # NEED/ FIXNOW! : a new ERROR number in diag.dat
   # Commented out for now !!! - JSA 2-23-98
  
   $sid = "" unless defined $sid; 
   debug "sid = $sid";
   debug "remote_user = $remote_user";
   
   #    if (($sid) && (!($remote_user eq "guest")) ) {
   #                debug "Testing for illegal accounts";
   #   debug "sid = $sid";
   #        debug "remote_user = $remote_user";
   #            if  ($sid ne $remote_user)  {
   #                        puke "Can't use another user's bookmark";
   #                }
   #    }
   
   # Create a new instance:
   my $new = $modeclass->new;
   
   # Log it (a slight kludge; we log tags whose meaning is well-defined):
   
##    We do want to log the switches, otherwise you can't tell what mode
##    the user went to next.
##    unless ($opts{Switch}) {    # don't re-log it!
   
   my $submode = $REQ->param('submode');
   # Don't log commtraces, because it fills up the log with many useless
   # entries. 
   unless ($submode && $submode eq 'COMMTRACE') { 
      # my $sid     = $REQ->param('sid');  # session not resumed at this point!
      my $u       = $REQ->param('u') || 
                          (defined $ENV{PATH_INFO} && 
                              substr($ENV{PATH_INFO}, 2) ) ||
                    '';
      ims_event 'MODE', { SESSID=>$sid },
      MODE=>   ($mode||''),
      SUBMODE=>($submode||''),
      UFROM=>  ($u || ''),
      UPAGE=>  (IW::Util::uniq());
   }
   
   # Call it:
   $new->go() or puke "mode handler for $mode returned failure status";
   1;
}

#------------------------------------------------------------
# switch PARAMHASH
#------------------------------------------------------------
# I<Class method.>
# Wipe the CGI tags (except for certain special ones), 
# replace them with the given tags, and handle the 
# desired mode as though we had received a request to do so:
#
#     switch IW::Mode Tags => {mode=>'NEWMODE', 
#                              tag1=>value1, 
#                              tag2=>[val21, val22]};

sub switch {
   my ($self, %params) = @_;
   my ($t, $v); 
   
   # Clear out CGI params:
   foreach $t ($REQ->params) {
      next if ($t =~ /^(sid|u)$/);
      $REQ->delete($t);
   }
   
   # Set new params:
   while (($t, $v) = each %{$params{Tags}}) {
      $v = [$v] unless ref($v) eq 'ARRAY';     # coerce singletons to arrays
      $REQ->param($t => @$v);
   }
   $REQ->param('x-switch' => 1);    # cheapo flag
   
   # Dispatch!
   my $newmode = $REQ->param('mode') || puke "missing mode tag in switch!"; 
   $self->handle($newmode, Switch=>1);
}

#------------------------------------------------------------
# go_submode
#------------------------------------------------------------
# I<Instance method>.  Get the 'submode' CGI tag (call it SUBMODE), 
# and use it to invoke method submode_SUBMODE() on the self object.
# If no submode exists, invoke method submode_DEFAULT().
#
# Your go() method should use this once it's done all the necessary setup:
#
#     sub go {
#         my $self = shift;
#         $APP->get_session;
#         $self->go_submode;
#     }
#
# Returns true if the submode method returns true.  Pukes if it returns false.

sub go_submode {
   my $self = shift;
   
   # Dispatch!
   my $submode = ($REQ->param('submode') || 'DEFAULT');
   my $method = "submode_$submode";
   $self->$method() or puke "submode $submode failed";
   1;
}

#------------------------------------------------------------
# go
#------------------------------------------------------------
# I<Instance method>.  Abstract main for a mode class.

sub go {
   my $self = shift;
   puke 600, "abstract go() method unimplemented in ", ref($self), "!";
}

#------------------------------------------------------------
# The following routines are generic versions that will be
# called if the subclass doesn't implement them. These are 
# part of Navigator support.
#------------------------------------------------------------
# nav_status
#------------------------------------------------------------
# Returns a dummy status message for undefined callbacks.

sub nav_status {
   my ($class,$step) = @_;
   
   defined $SESSION or 
       puke "App's import of SESSION failed";
   defined $SESSION->{Navigator} or
       puke "Navigator never initialized";
   defined $SESSION->{Navigator}->{$step} or
       puke "Nothing in Navigator for $step";
   
   return ("Message" => "",
           "Link"    => "",
           "Mode"    => "",
           "Submode" => "",
           );
   
}

#------------------------------------------------------------
1;
