package IW::ModeABSTRACTSRCH;

=head1 NAME

IW::ModeABSTRACTSRCH - Abstract class that provides basic frame work for 
                       search forms.  Creates the widget box and tells the 
                       widget box when to update and draw. Also provides 
                       legacy support for SLCT popups.

                submodes (CGI interface):

                  submode_DEFAULT - maps to UPDATE.

                  submode_UPDATE - updates widget box and does a _show.

                  submode_PEEKODL - debug mode to display the Query
                     object as ODL to the screen.

                abstract methods (overridden by child classes):

                  _get_widgetbox - name of widget box to use.

                  _mode - name of this mode.

                internal methods:

                  _init - get session and create the widget box.

                  _go_popup - handling switching to SLCT popups.

                  _show - call _go_popup and if return value is true
                      (no popup) then draw the widget box:
                      $self->_go_popup or $self->{WidgetBox}->draw

                  _show_ODL - display Query as ODL. See submode_PEEKODL.

                public methods:

                  request_popup - used as a callback in widgetbox to request 
                      that a SLCT popup be displayed (see _go_popup).

                  go - all modes have a go method, calls _init and
                      then the appropriate submode.

                legacy methods (needed for SLCT popups):

                  submode_OK - maps to UPDATE.

                  submode_GOTO - maps to UPDATE.

                  submode_SHOW - maps to UPDATE.

                  sub goto - calls method go.     

=cut

use IU;  # For puke and error handling
use IU::Query; # For $QUERY

use IW::App; # For $HTML

use strict;

use vars qw(@ISA);
@ISA = qw(IW::Mode);

#------------------------------------------------------------
#------------------------------------------------------------
# Abstract methods: Child classes must implement these:
#------------------------------------------------------------
#------------------------------------------------------------


#------------------------------------------------------------
# _get_widgetbox
#------------------------------------------------------------
# Abstract
# Return the widgetbox class name to use in building the form:

sub _get_widgetbox { puke "Abstract method must override in child class" }

#------------------------------------------------------------
# _mode
#------------------------------------------------------------
# Abstract
# Return Mode name of the child class:

sub _mode { puke "Abstract method must override in child class" }

#------------------------------------------------------------
#------------------------------------------------------------
# The Submodes (public or external interface).
#------------------------------------------------------------
#------------------------------------------------------------

#------------------------------------------------------------
# submode_DEFAULT 
#------------------------------------------------------------

sub submode_DEFAULT { shift->submode_UPDATE(@_) }

#------------------------------------------------------------
# submode_UPDATE
#------------------------------------------------------------

sub submode_UPDATE {

   my $self = shift;
   
   #
   # Tell widgets to update:
   #
   
   $self->{WidgetBox}->update;

   #
   # Draw search form
   #

   $self->_show;

}

#------------------------------------------------------------
# submode_PEEKODL
#------------------------------------------------------------
# Debug tool to display the query as ODL on the screen. 
# There are no buttons or links in the UI to access this submode.
# You must manually type a URL with mode=MAINSRCH&submode=PEEKODL. 


sub submode_PEEKODL {

   my $self = shift;

   $self->_show_ODL;

}

#------------------------------------------------------------
#------------------------------------------------------------
# The methods (private or internal).
#------------------------------------------------------------
#------------------------------------------------------------

#------------------------------------------------------------
# _init 
#------------------------------------------------------------
#
# Create objects needed by search form.
#

sub _init {

   my $self = shift;

   #
   # Get session (creating it if necessary):
   #

   $APP->get_session('CREATE');

   #
   # Create widget box:
   #
   
   # _get_widgetbox is child method which returns a string that is 
   # name of the widget box module to create. 
   my $widgetboxname = "IW::SrchForm::".$self->_get_widgetbox;
   $self->{WidgetBox} = $widgetboxname->new('ParentMode'=>$self->_mode,
       Callback2request_popup=>sub {$self->request_popup(@_)});

   # Save mode in the session state so popups know where to return to:
   # ERS - not my first choice to do, but this is needed by the popup 
   # pages. So until we rewrite them...

   $SESSION->state->{SearchFormMode}   = $self->_mode;

1; 
}

#------------------------------------------------------------
# _go_popup
#------------------------------------------------------------

sub _go_popup {

   my $self = shift;

   # No popup requested, return 0 (false) value to cause the 
   # search form page to be redisplayed. 

   return 0 unless (defined($self->{POPUP}));

   # There was problem found by one of the widgets during update
   # all popup requests have been canceled.

   return 0 if ($self->{POPUP} eq 'CANCEL'); 

   # No popup but don't want the main search form redisplayed.
   # Return 1 (true) to skip redisplay of main search form. 
   # Used when SAVE to disk button is clicked. 

   return 1 if ($self->{POPUP} eq 'NONE');

   # Add "popup" mode to CGI tags: 

   $self->{TAGS}->{'mode'} = $self->{POPUP};

   # If switch works return value is true '1' and the main search form
   # is not draw.

   return IW::Mode->switch(Tags=>$self->{TAGS});

0;
}

#------------------------------------------------------------
# _show
#------------------------------------------------------------

sub _show {

   my $self = shift;
   $self->_go_popup or $self->{WidgetBox}->draw;

1;
}

#------------------------------------------------------------
# _show_ODL
#------------------------------------------------------------

sub _show_ODL {

   my $self = shift;

   $APP->start_page(SuperTitle=>'Search Form', Title=>'Search ODL');

   my $odl = $QUERY->to_ODL;

   if ($odl) {
      $HTML->t_B("Full QUERY object as ODL:");
      my $TMP = FileHandle->new_tmpfile;
      $HTML -> PRE;
      $odl->output($TMP);
      $TMP->seek(0,0); output $HTML $_ while <$TMP>; close $TMP;
      $HTML -> _PRE;
   }
   else {
      $HTML -> t_EM("Couldn't convert query to ODL");
   }

   my $search_odl = $QUERY->to_ODL('ForSearch'=>'1');

   if ($search_odl) {
      $HTML->t_B("QUERY as ODL sent in search:");
      my $TMP = FileHandle->new_tmpfile;
      $HTML -> PRE;
      $search_odl->output($TMP);
      $TMP->seek(0,0); output $HTML $_ while <$TMP>; close $TMP;
      $HTML -> _PRE;
   }
   else {
      $HTML -> t_EM("Couldn't convert query to search ODL");
   }

   $APP->end_page;

1;
}

#------------------------------------------------------------
#------------------------------------------------------------
# Public methods
#------------------------------------------------------------
#------------------------------------------------------------

#------------------------------------------------------------
# request_popup 
#------------------------------------------------------------

sub request_popup {

    my ($self,$popup,%params) = @_;
    # Once a CANCEL has been set, all further popup requests are
    # ignored.
    $self->{POPUP} = $popup 
       if (defined($popup) and ($self->{POPUP} ne 'CANCEL'));
    # Get any additional CGI tags requested:
    $self->{TAGS} = $params{Tags} || {};
    $self->{POPUP};
}

#------------------------------------------------------------
# go
#------------------------------------------------------------
# Main entry point for mode.

sub go {
    my $self = shift;

    #
    # Create objects needed by the search form;
    #

    $self->_init;

    #
    # Standard submode entry.
    # 

    $self->go_submode;
1;
}

#------------------------------------------------------------
#------------------------------------------------------------
# Some legacy stuff.  These are needed to make the new search
# form work with the SLCT popups (SLCTDEPVAL, SLCTALL, 
# MAPPLET). 
#------------------------------------------------------------
#------------------------------------------------------------ 

#------------------------------------------------------------
# submode_OK
#------------------------------------------------------------
# Popups need for MAIN search form to have an OK submode:

sub submode_OK { shift->submode_UPDATE(@_) }

#------------------------------------------------------------
# submode_CANCEL
#------------------------------------------------------------
# Popups need for MAIN search form to have an CANCEL submode:

sub submode_CANCEL { shift->submode_UPDATE(@_) }

#------------------------------------------------------------
# submode_GOTO
#------------------------------------------------------------
# Popups need for MAIN search form to have an GOTO submode:

sub submode_GOTO { shift->submode_UPDATE(@_) }

#------------------------------------------------------------
# submode_SHOW
#------------------------------------------------------------
# Popups need for MAIN search form to have an SHOW submode:

sub submode_SHOW { shift->submode_UPDATE(@_) }


#------------------------------------------------------------
# goto
#------------------------------------------------------------
# The popups need for the MAIN search form to have a goto
# method.

sub goto {

   my $self = shift;
   $self->go;
1;
}

#------------------------------------------------------------
1;
