package IW::ModeADMINHIST;

=head1 NAME

IW::ModeADMINHIST - show sessions in stats logs

=head1 DESCRIPTION

A subclass of IW::ModeADMINPopup, for viewing/deleting currently-active
user sessions.

=head1 AUTHOR

Eryq

=head1 RCS INFO

$Id: ModeADMINHIST.pm,v 1.3 1999/04/16 15:00:52 lrosen Exp $

=cut

use IU;
use IU::Log;
use IU::Util qw(path_filename recent_time units2secs units2bytes shellquote);
use IU::EventEntry;
use IW::App;
use IW::Util qw(:url);
use IW::Note;
use IW::ModeADMINPopup;

@ISA = qw(IW::ModeADMINPopup);

use strict;

#------------------------------------------------------------
# 
# Globals
#
#------------------------------------------------------------

# Page titles:
my $Title = 'Stat logs report';

# Requests we're interested in:
my %EventWanted = (
    'directory'  => 1,
    'inventory'  => 1,
    'productreq' => 1,
);

# Ignore addresses matching any of these patterns:
my %IgnoreAddr = map {$_=>1}
    qw(
       killians.gsfc.nasa.gov
       killians-e.gsfc.nasa.gov
       odouls.stx.com
       batcave.stx.com
       icehouse.stx.com
       harp.gsfc.nasa.gov
       harp-lite.gsfc.nasa.gov
       sprecher.gsfc.nasa.gov
       selsvr.stx.com
       daacdev1.stx.com
       rhine.gsfc.nasa.gov
       fake.host
       stx.com
       );

my %Session;
my %ByEvent;
my %ByEventAndDaac;
my %ByNumProducts;
my %BySession;
my %DaacSeen;

#------------------------------------------------------------
# ignore_addr
#------------------------------------------------------------
# Ignore these addresses when compiling stats.

sub ignore_addr {
    my $addr = shift;
    $addr =~ s/^[^@]*@//;       # get rid of user name
    return $IgnoreAddr{$addr};
}

#------------------------------------------------------------
# nice_datetime DATETIME
#------------------------------------------------------------
# I<Utility.>  Change a stat log datetime into a pretty datetime.

sub nice_datetime {
    my $dt = shift;
    my ($yr, $mo, $dy, $hr, $mi, $se) = 
	($dt =~ /^\d{2}(\d{2})(\d{2})(\d{2})(\d{2})(\d{2})(\d{2})/);
    "$yr-$mo-$dy $hr:$mi:$se";
}

#------------------------------------------------------------
# stat_datetime DATETIME
#------------------------------------------------------------
# I<Utility.>  Change a stat log datetime into a statfile datetime.

sub stat_datetime {
    my $dt = shift;
    my ($yr, $mo, $dy, $hr, $mi, $se) = 
	($dt =~ /^(\d{4})(\d{2})(\d{2})(\d{2})(\d{2})(\d{2})/);
    "$yr-$mo-$dy";
}

#------------------------------------------------------------
# stat_files [START,STOP]
#------------------------------------------------------------
# I<Instance method.>
# Output a list of paths to stat files, in the given range.
# The START and STOP should each be given as YYYY, YYYY-MM, or YYYY-MM-DD.

sub stat_files {
    my ($self, $start, $stop) = @_;
    my $file;

    # Pad start:
    if ($start) {
	($start =~ /^\d{4}$/)       and ($start .= '-01');
	($start =~ /^\d{4}-\d{2}$/) and ($start .= '-01');
    }

    # Pad stop:
    if ($stop) {
	($stop =~ /^\d{4}$/)       and ($stop .= '-12');
	($stop =~ /^\d{4}-\d{2}$/) and ($stop .= '-31');
    }

    # Scan:
    my @got;
    my $dir = "$IU::TopDir/log";
    opendir LOGS, $dir or die "opendir: $!";
    while ($file = defined(readdir(LOGS))) {
	next if ($file !~ /^(\d{4}-\d{2}-\d{2})\.statlog(\.done)?$/);
	my $date = $1;

	next if ($start and ($date lt $start));
	next if ($stop  and ($date gt $stop));
	
	debug "got $file";
	push @got, "$dir/$file"; 
    }
    closedir LOGS;
    return @got;
}

#------------------------------------------------------------
# do_entry
#------------------------------------------------------------
# I<Instance method.>

sub do_entry {
    my $entry = shift;
    my $event;
    my $daac;
    

    # Skip if one of ours:
    return '.' if ($REQ->param('filter') && ignore_addr($entry->addr));
    
    # Skip if not the right kind of event:
    return '-' if !$EventWanted{$event = lc($entry->event)};
    
    # Get session id:
    my $sessid = $entry->session_id;
    
    # Count the DAACs:
    my @daacs = $entry->values('archid');
    my $nprods = 0;
    if (!@daacs) {
	my @prods = $entry->values('prodname');
	$nprods = @prods;
	foreach (@prods) { s{/.*\Z}{}g }
	my %daacsh = map {$_=>1} @prods;
	@daacs = keys(%daacsh);
	### print STDERR "\nPRODUCT DAACS: ", join(',',@daacs), "\n";
    }
    my $daac_slot = int(@daacs);
    $daac_slot = 10 if $daac_slot > 10;

    # TALLY, by event and DAAC (2-d table):
    foreach $daac (@daacs) {
	$DaacSeen{$daac} = 1;
	$ByEventAndDaac{$event} ||= {};
	$ByEventAndDaac{$event}{$daac} ||= 0;
	++$ByEventAndDaac{$event}{$daac};
    }

    # TALLY number of products, by DAAC (1-d table):
    $ByNumProducts{Products} ||= {};
    $ByNumProducts{Requests} ||= {};
    foreach $daac (@daacs) {
	$ByNumProducts{Products}{$daac} ||= 0;
	$ByNumProducts{Products}{$daac} += $nprods;
	$ByNumProducts{Requests}{$daac} = $ByEventAndDaac{'productreq'}{$daac};
    }

    # TALLY, by event (histogram):
    $ByEvent{$event} ||= [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0];
    ++$ByEvent{$event}[$daac_slot];
        
    # TALLY, by session (histogram):
    $Session{$sessid} ||= {};
    $Session{$sessid}{$event} ||= {};
    foreach $daac (@daacs) {
	$Session{$sessid}{$event}{$daac} = 1;
    }
    return '.';
}

#------------------------------------------------------------
# scan_stats \@FILES
#------------------------------------------------------------
# I<Instance method.>
# Scan the logs, accumulating desired statistics. 

sub scan_stats {
    my ($self, $files) = @_;

    ($files and @$files) or puke 549, "no files to scan!";

    # Scan logs:
    do_files IU::EventEntry 
	Files => $files,
	Trace => undef,
	Action => \&do_entry;
    
    # Second pass, by session:
    my ($sid, $sid_events);
    while (($sid, $sid_events) = each %Session) {
	($sid_events && (keys %$sid_events)) or next;  # skip if no events

	my ($event, $event_daacs);
	while (($event, $event_daacs) = each %$sid_events) {
	    $event_daacs ||= {};
	    my $count = keys %$event_daacs;    # count of DAACs

	    $BySession{$event} ||= [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0];
	    ++$BySession{$event}[$count];
	}
    }
    1;
}

#------------------------------------------------------------
# print_histogram
#------------------------------------------------------------
# I<Instance method.>
#       Slot  => bins 0-9 and 10+

sub print_histogram {
    my $self = shift;
    my %params = @_;
    my @slots = @{$params{Slots} || []};
    my $i;

    # Total stuff up:
    my ($total, $single, $multi) = (0, 0, 0);
    for ($i = 0; $i < @slots; $i++) {
	$total += $slots[$i];
	$multi += $slots[$i] if ($i > 1);
    }
    $single = $slots[1];

    # Output:
    $HTML -> P -> HR -> H3 -> t($params{Title}) -> _H3;
    $HTML -> TABLE(CELLPADDING=>1, CELLSPACING=>1);
    $HTML -> TR(VALIGN=>'BOTTOM')
	  -> TD -> t("# DAACS")
	  -> TD -> t("Count")
	  -> TD -> t("%")
	  -> TD(NOWRAP=>undef, ALIGN=>'LEFT')
	      -> TT 
	      -> t("0%-----------------50%----------------100%") -> BR
	      -> t("|...|...|...|...|...|...|...|...|...|...|")
	      -> _TT
		  ;

    my $i;
    for ($i = 0; $i <= 10; $i++) {
	my $n      = $slots[$i] || 0;
	my $n_frac = ($total ? $n/$total : 0);
	my $n_pct  = int($n_frac * 100) || 0;
	my $line = '|' . (("#" x int($n_frac * 40)) || "");
	$HTML -> TR
	      -> TD(ALIGN=>'RIGHT') -> t(($i == 10) ? "10+" : $i)
	      -> TD(ALIGN=>'RIGHT') -> t($n)
	      -> TD(ALIGN=>'RIGHT') -> t($n_pct)
	      -> TD(NOWRAP=>undef, ALIGN=>'LEFT') -> TT -> t($line) -> _TT
		  ;
    }
    $HTML -> _TABLE;

    $HTML -> TABLE(BORDER=>1, CELLSPACING=>1, WIDTH=>'50%');
    $HTML -> TR -> TD -> t("Total count:") 
	        -> TD -> t($total);
    if ($total > 0) {
	$HTML -> TR -> TD -> t("Single-DAAC:") 
	            -> TD -> t($single)
	            -> TD -> t(int(100 * ($single/$total)),'%')
			;
	$HTML -> TR -> TD -> t("Multi-DAAC:") 
	            -> TD -> t($multi)
	            -> TD -> t(int(100 * ($multi/$total)),'%')
			;
    }
    $HTML -> _TABLE;
    1;
}

#------------------------------------------------------------
# print_table
#------------------------------------------------------------
# I<Instance method.>
#    Title => title
#    X     => x-axis labels
#    Y     => y-axis labels
#    Data  => hash, {x}{y}

sub print_table {
    my $self = shift;
    my %params = @_;
    my $xs = $params{X};
    my $ys = $params{Y};
    my $data = $params{Data};
    my ($x, $y);
    my $i;

    # Output:
    $HTML -> P -> HR -> H3 -> t($params{Title}) -> _H3;
    $HTML -> TABLE(CELLPADDING=>1, CELLSPACING=>1, BORDER=>1);

    # Headers:
    $HTML -> TR(VALIGN=>'BOTTOM') -> TH;
    foreach $x (@$xs) { $HTML -> TH -> t($x) }

    # Data:
    foreach $y (@$ys) {
	$HTML -> TR(VALIGN=>'BOTTOM');
	$HTML -> TH -> t($y);
	foreach $x (@$xs) { 
	    my $d = $data->{$x}{$y}; defined($d) or $d = '';
	    $HTML -> TD(ALIGN=>'MIDDLE') -> t($d); 
	}
    }
    $HTML -> _TABLE;
    1;
}

#------------------------------------------------------------
# print_stats
#------------------------------------------------------------
# I<Instance method.>
# Print the statistics we just scanned.

sub print_stats {
    my $self = shift;

    # Overview:
    $self->print_table(Title=>"Requests by event and DAAC",
		       X    => [sort keys %EventWanted],
		       Y    => [sort keys %DaacSeen],
		       Data =>\%ByEventAndDaac);

    # Number of products requested:
    $self->print_table(Title=>"Number of products requested by DAAC",
		       X    => [sort keys %ByNumProducts],
		       Y    => [sort keys %DaacSeen],
		       Data =>\%ByNumProducts);

    # Print histograms for each event:
    my $event;
    foreach $event (sort keys %EventWanted) {
	$self->print_histogram(Title=>"$event (by request)",
			       Slots=>$ByEvent{$event});
	$self->print_histogram(Title=>"$event (by session)",
			       Slots=>$BySession{$event});
    }
    1;
}

#------------------------------------------------------------
# send_stats_doc PARAMHASH
#------------------------------------------------------------
# I<Instance method.>
# Output the document showing the stats.

sub send_stats_doc {
    my ($self, %params) = @_;
    
    # Start page:
    $self->start_page(Title => $Title);

    # Start form:
    $HTML -> HR -> script_FORM;
    $HTML -> msubmit_flag;
    $HTML -> hidden('mode'  =>$REQ->param('mode'));

    # Report:
    $HTML -> t("Here are the statistics I was able to gather from the ".
	       "stat logs ");
    if ($REQ->param('specify') eq 'FILES') {
	$HTML -> t("which matched ")
	      -> t_B($REQ->param('files'))
	      -> t(".\n")
		  ;
    }
    else {
	$HTML -> t("in the selected time range ")
	      -> t_B("[", $REQ->param('start')," - ",$REQ->param('stop'), "]")
	      -> t(".\n")
		  ;
    }
    $HTML -> small;
    print <<EOF;

<P>The "by-request" stat works very simply: 
<UL>
<LI> We scan the files looking for events of interest ("inventory", etc.).  
     If such an event involves 5 DAACs, we increment the count [of requests] 
     in the 5 DAAC bin for that event.  

<LI> <B>Bottom line:</B> 
      For a given event type (like "inventory"), 
      each log entry with that event type yields a number of DAACS...
      and that number is tallied exactly once, in one of the 10 bins.
</UL>

<P>The "by-session" stat works this way:
<UL>
<LI> <B>If</B> within a session a user does an inventory search to ASF, 
        EDC and JPL, and then later on in the same session does an another 
        inventory search to NSIDC and DLR, 
     <B>then</B> we count the session as spanning 5 unique DAACs
        for inventory search... 
     <B>so</B> we increment the count [of sessions] 
        in the 5 DAAC bin for event "inventory".

<LI> <B>If</B> within a session a user does an product order to EDC 
        and JPL, and then later on within the same session places another 
        order to just EDC,
     <B>then</B> we count the session as spanning 2 unique DAACs
        (not 3!) for product order... 
     <B>so</B> we increment the count [of sessions] 
        in the 2 DAAC bin for event "productreq".

<LI> <B>If</B> within a session a user does 6 directory searches
        to GSFC (and no other directory searches), 
     <B>then</B> we count the session as spanning 1 unique DAAC
        for directory search...
     <B>so</B> we increment the count [of sessions] in the 1 DAAC bin 
        for event "directory".

<LI> <B>Bottom line:</B> 
      For a given event type (like "inventory"), 
      each session yields a total number of unique DAACS involved in 
      events of that type... and that number is tallied exactly once, 
      in one of the 10 bins.
</UL>
EOF
    $HTML -> _small;

    # Stats:
    $self->print_stats;
    $HTML -> _script_FORM;

    # Selection form:
    $self->print_logselect_form;

    # Other info...
    $HTML -> HR;
 
    # Files:
    $HTML -> H3 -> t("The following files were scanned:") -> _H3;
    my $file;
    $HTML -> small;
    foreach $file (@{$params{Files}}) {
	$HTML -> t($file, "  (", -s $file, " bytes)" ) -> BR;
    }
    $HTML -> _small;

    # Filters:
    if ($REQ->param('filter')) {
	$HTML -> H3 -> t("The following addresses were filtered:") -> _H3;
	my $addr;
	foreach $addr (sort keys %IgnoreAddr) {
	    $HTML -> t($addr) -> BR;
	}
    }
    # The signature:
    $self->end_page();
    1;    
}

#------------------------------------------------------------
# print_logselect_form
#------------------------------------------------------------
sub print_logselect_form {
    my $self = shift;

    $HTML -> HR;
    $HTML -> script_FORM;
    $HTML -> hidden('mode'    => $REQ->param('mode'),
		    'submode' => 'ALL');
    $HTML -> msubmit_flag;

    # Directions:
    $HTML -> directions;
    print <<EOF;
    Select the stats options, then choose the dates (or files) you want
    scanned and hit the submit button next to that choice.
EOF
    $HTML -> _directions;

    # Filter?
    $HTML -> P;
    $HTML -> checkbox(Name=>'filter', Choice=>1)
	  -> t_B(" Filter out developers when scanning logs?");

    # Enter as date range:
    $HTML -> P;
    $HTML -> TABLE(WIDTH=>'100%');
    $HTML -> TR
	  -> TD(COLSPAN=>3)
	  -> t("You can (1) ".
	       "enter the earliest start and latest stop dates for stat ".
	       "files to scan (the format is YYYY-MM or YYYY-MM-DD, ".
	       "and it's okay to leave either or both blank)...")
	      ;
    $HTML -> TR
	  -> TD(NOWRAP=>undef) 
	      -> t_B("Start: ")
	      ->input_text(Name=>'start', Size=>11)
	  -> TD(NOWRAP=>undef) 
	      -> t_B("Stop: ")
	      ->input_text(Name=>'stop', Size=>11)
	  -> TD(ALIGN=>'RIGHT')
	      -> msubmit_button(QUERY =>['specify'=>'DATES'],
				CAPTION=>'Compute statistics');
    $HTML -> _TABLE;

    # Enter as files:
    $HTML -> TABLE(WIDTH=>'100%');
    $HTML -> TR
	  -> TD(COLSPAN=>2)
	  -> t("Or, you can (2) ".
	       "enter a shell pattern for the .statlog files ".
	       "you want to scan (it can contain metacharacters like ".
	       "*, ~, etc.)... ")
	      ;
    $HTML -> TR
	  -> TD(NOWRAP=>undef) 
	      -> t_B("Scan files: ")
	      -> input_text(Name=>'files', Size=>30)
	  -> TD(ALIGN=>'RIGHT')
	      -> msubmit_button(QUERY =>['specify'=>'FILES'],
				CAPTION=>'Compute statistics');
    $HTML -> _TABLE;

    $HTML -> _script_FORM; 
}

#------------------------------------------------------------
#
# Submode handlers...
#
#------------------------------------------------------------

#------------------------------------------------------------
# submode_DEFAULT
#------------------------------------------------------------
# What to do if invoked with no explicit submode...

sub submode_DEFAULT {
    shift->submode_CHOOSE;
}

#------------------------------------------------------------
# submode_CHOOSE
#------------------------------------------------------------
# What to do if invoked with no explicit submode...

sub submode_CHOOSE {
    my $self = shift;

    $self->start_page(Title => $Title);
    $self->print_logselect_form();   
    $self->end_page();
    1;    
}

#------------------------------------------------------------
# submode_ALL
#------------------------------------------------------------
# Show all the known sessions.

sub submode_ALL {
    my $self = shift;

    # Get files to scan:
    my @files;
    if ($REQ->param('specify') eq 'FILES') {
	(@files = glob($REQ->param('files')))
	    or puke 513, "Sorry, no match found for ", $REQ->param('files');
    }
    else {
	(@files = $self->stat_files($REQ->param('start'), $REQ->param('stop')))
	    or puke 513, "no stat files to scan!";
    }

    # Scan files for all sessions in them:
    my $list = $self->scan_stats(\@files) || puke 549, "no sessions!";

    # Output:
    $self->send_stats_doc(Files=>\@files);
}

#------------------------------------------------------------
1;
