package IW::ModeADMINPopup;

=head1 NAME

IW::ModeADMINPopup - stub superclass class for various ADMIN* modes

=head1 DESCRIPTION

Since all ADMIN modes have a similar look, feel, and behavior, they
are all subclasses of this one class.

Basically, we have this class hierarchy:

  IW::Mode
     |
     |--- IW::ModeADMIN
     |--- IW::ModeADMINPopup
     :       |
             |--- IW::ModeADMINACTIVE
             |--- IW::ModeADMINCACHE
             |--- IW::ModeADMINLOGIN...
             :

A lot of generic stuff is done in ModeADMINPopup; you you need
to do is have your new ADMIN classes inherit from it.  

=head1 SECURITY

All the real activities have been shunted off into subclasses:
ADMINCACHE, ADMINHTTPD, etc.   Most of them require that you
log into the system first as an administrator.  

To make this more reliable, the "go" method that they inherit
from I<this> class actually does a check first to see if the 
class (e.g., ADMINCACHE) requires a login: since this information
is stored in the session state, a user can't simply "bypass" it
by entering a URL which contains the correct mode and submode.

Note that stealing an admin's session, however, is another matter 
entirely: it is expected that there will be a cookie- or IP-based
authentication step on each request to prevent such "hijacking".

=head1 AUTHOR

Eryq

=head1 RCS INFO

$Id: ModeADMINPopup.pm,v 1.2 2001/11/13 21:57:48 joe Exp $

=cut

use IU;
use IU::Log;

use IW::App;
use IW::Mode;
use IW::Util qw(:url);

use strict;
use vars qw(@ISA);

@ISA = qw(IW::Mode);



#------------------------------------------------------------
#
# Globals
#
#------------------------------------------------------------

# Super title (appears atop all pages):
my $SuperTitle = 'ADMIN Utilities';

#------------------------------------------------------------
#
# Utility methods...
#
#------------------------------------------------------------

#------------------------------------------------------------
#
# Standard output methods...
#
#------------------------------------------------------------

#------------------------------------------------------------
# main_menu_button [LOCATION]
#------------------------------------------------------------
# I<Instance method.>  
# Output a small form for returning to the main ADMIN menu.
# Location may be 'TOP', 'BOTTOM', or not given.

sub main_menu_button {
    my ($self, $location) = @_;
    $location or $location = '';

    $HTML -> B -> FONT(SIZE=>'+2') -> CENTER;
    $HTML -> HR;
    $HTML -> A(HREF=>script_url('mode'=>'ADMIN')) 
	  -> t('Return to main ADMIN menu')
          -> _A;
    $HTML -> _CENTER -> _FONT -> _B;
}

#------------------------------------------------------------
# start_page
#------------------------------------------------------------
# I<Instance method.>  Start an ADMIN page.  Parameters are the
# same as for IW::App::start_page, except that the super title
# is always the same.  A "return to main menu" link is always
# placed at the top.

sub start_page {
    my ($self, %params) = @_;
    $params{SuperTitle} = $SuperTitle;

    $APP->start_page(%params);
    $self->main_menu_button('TOP');
    1;
}

#------------------------------------------------------------
# end_page
#------------------------------------------------------------
# I<Instance method.>  End an ADMIN page.  Suppresses the normal
# key.

sub end_page {
    my $self = shift;

    # The signature:
    $HTML -> P;
    $self->main_menu_button('BOTTOM');
    $APP->suppress_key(1);
    $APP->end_page();    
}


#------------------------------------------------------------
#
# Mode handling...
#
#------------------------------------------------------------

#------------------------------------------------------------
# require_login
#------------------------------------------------------------
# I<Abstract instance method.>  
# Require a login to be established?  Default is yes.

sub require_login {
    1;
}

#------------------------------------------------------------
# go
#------------------------------------------------------------
# I<Instance method.>
# The standard main entry point for modes.  This generic form...
#
#    (1) always requires a session to be established, 
#    (2) always requires a login if $self->require_login returns true.
#    (3) handles submode "FOO" by attempting to invoke method submode_FOO()
#        on the current object (with an empty submode defaulted to
#        submode_DEFAULT().
#
# You can of course override it in your mode handler if you need to.

sub go {
    my $self = shift;

    # Get session or die:
    $APP->get_session();

    # Require login?
    if ($self->require_login) {
	($SESSION->state->{AdminLoggedIn}) or puke 520, "Please log in first";
    }

    # Dispatch!
    my $submode = ($REQ->param('submode') || 'DEFAULT');
    my $method = "submode_$submode";
    $self->$method() or puke "submode $submode failed";
    1;
}

#------------------------------------------------------------
1;
