package IW::ModeADMINSTATS;

=head1 NAME

IW::ModeADMINSTATS - show sessions in stats logs

=head1 DESCRIPTION

A subclass of IW::ModeADMINPopup, for viewing/deleting currently-active
user sessions.

=head1 AUTHOR

Eryq

=head1 RCS INFO

$Id: ModeADMINSTATS.pm,v 1.8 2002/10/17 20:12:43 joe Exp $

=cut

use IU;
use IU::Log;
use IU::Util qw(path_filename recent_time units2secs units2bytes shellquote);
use IU::EventEntry;

use IW::App;
use IW::Util qw(:url);
use IW::Note;
use IW::ModeADMINPopup;
use IW::User;

@ISA = qw(IW::ModeADMINPopup);

use strict;

#------------------------------------------------------------
# 
# Globals
#
#------------------------------------------------------------

# Page titles:
my $Title = 'Stat logs';

# A scratch list of sessions and basic session info, used by scan_sessions.
my %List;

# Scratch data for the chosen session's log entries:
my @Entries;


#------------------------------------------------------------
# nice_datetime DATETIME
#------------------------------------------------------------
# I<Utility.>  Change a stat log datetime into a pretty datetime.

sub nice_datetime {
    my $dt = shift;
    my ($yr, $mo, $dy, $hr, $mi, $se) = 
	($dt =~ /^(\d{4})(\d{2})(\d{2})(\d{2})(\d{2})(\d{2})/);
    "$yr-$mo-$dy $hr:$mi:$se";
}

#------------------------------------------------------------
# stat_datetime DATETIME
#------------------------------------------------------------
# I<Utility.>  Change a stat log datetime into a statfile datetime.

sub stat_datetime {
    my $dt = shift;
    my ($yr, $mo, $dy, $hr, $mi, $se) = 
	($dt =~ /^(\d{4})(\d{2})(\d{2})(\d{2})(\d{2})(\d{2})/);
    "$yr-$mo-$dy";
}

#------------------------------------------------------------
# stat_files [START,STOP]
#------------------------------------------------------------
# I<Instance method.>
# Output a list of paths to stat files, in the given range.
# The START and STOP should each be given as YYYY, YYYY-MM, or YYYY-MM-DD.

sub stat_files {
    my ($self, $start, $stop) = @_;
    my $file;

    # Pad start:
    if ($start) {
	($start =~ /^\d{4}$/)       and ($start .= '-01');
	($start =~ /^\d{4}-\d{2}$/) and ($start .= '-01');
    }

    # Pad stop:
    if ($stop) {
	($stop =~ /^\d{4}$/)       and ($stop .= '-12');
	($stop =~ /^\d{4}-\d{2}$/) and ($stop .= '-31');
    }

    # Scan:
    my @got;
    my $dir = "$IU::TopDir/log";
    opendir LOGS, $dir or die "opendir: $!";
    while (defined($file = readdir(LOGS))) {
        next if ($file !~ /^(\d{4}-\d{2}-\d{2})\.statlog(\.done)?$/);
        my $date = $1;

        next if ($start and ($date lt $start));
        next if ($stop  and ($date gt $stop));

        debug "got $file";
        push @got, "$dir/$file"; 
    }
    closedir LOGS;
    return @got;
}

#------------------------------------------------------------
# scan_sessions \@FILES
#------------------------------------------------------------
# I<Instance method.>
# Construct a list of all sessions in the logs.  The return value, 
# if defined, is a hashref whose keys are session ids, and whose values 
# are themselves hashrefs of:
#
#     Start        start date
#     Stop         stop date
#     Addr         user address

sub scan_sessions {
    my ($self, $files) = @_;

    ($files and @$files) or puke 549, "no files to scan!";

    # Scan logs:
    do_files IU::EventEntry 
	Files => $files,
	Trace => undef,
	Action => sub {
	    my $entry = shift;
	    
	    # Note session id, and force list entry:
	    my $sid = $entry->session_id;
	    $List{$sid} ||= {};
	    
	    # Get earliest time:
	    if (!$List{$sid}{Start} ||
		$List{$sid}{Start} > $entry->datetime_str) {
		$List{$sid}{Start} = $entry->datetime_str;
	    }

	    # Get latest time:
	    if (!$List{$sid}{Stop} ||
		$List{$sid}{Stop} < $entry->datetime_str) {
		$List{$sid}{Stop} = $entry->datetime_str;
	    }

	    # Get user:
	    unless ($entry->event eq 'logout') {
		$List{$sid}{Addr} = $entry->addr;
	    }

	    return '.';
	};
    return \%List;
}

#------------------------------------------------------------
# print_sessions \%LIST 
#------------------------------------------------------------
# I<Instance method.>
# Display a table of all sessions in the logs.  LIST is of the
# same type as the return value of scan_sessions().
#
# The assumption is that this is being output inside a form 
# which will be submitted back to this mode.

sub print_sessions {
    my ($self, $list) = @_;

    # Display:
    my $sid;
    $HTML -> TABLE(BORDER=>1, WIDTH=>'100%');
    $HTML -> TR 
	  -> TH -> t('Session id') -> _TH
	  -> TH -> t('Start time') -> _TH
	  -> TH -> t('Stop time') -> _TH
          -> TH -> t('User') -> _TH;
    $HTML -> _TR;
    foreach $sid (sort keys %$list) {
	next if ($sid eq 'unknown');

	# Get information in a pretty form for printing:
	my $start = nice_datetime($list->{$sid}{Start});
	my $stop  = nice_datetime($list->{$sid}{Stop});
	my $addr  = $list->{$sid}{Addr}; 
	$addr =~ s/^fake.name\@//;

	# Get stats span of this sid:
	my $s_start = stat_datetime($list->{$sid}{Start});
	my $s_stop  = stat_datetime($list->{$sid}{Stop});

	# Output!
	$HTML -> TR
	      -> TD -> B 
		    -> hypertext($sid,
				 script_url('mode'   =>$REQ->param('mode'),
					    'start'  =>$s_start,
					    'stop'   =>$s_stop,
					    'submode'=>'ONE',
					    'sessid' =>$sid))
		-> _B -> _TD
	      -> TD -> t($start) -> _TD
	      -> TD -> t($stop) -> _TD
	      -> TD -> t($addr) -> _TD;
        $HTML -> _TR;
    }
    $HTML -> _TABLE;
    1;
}

#------------------------------------------------------------
# print_session \@ENTRIES
#------------------------------------------------------------
# I<Instance method.>
# Display a table of all activity in the session. 
#
# The assumption is that this is being output inside a form 
# which will be submitted back to this mode.

sub print_session {
    my ($self, $entries) = @_;
    my $entry;
    my %requests;

    # Sanity:
    @$entries or puke 549, "no log entries for session!";

    # Main title:
    $HTML -> H2 -> t("Session: ", $entries->[0]->session_id) -> _H2;

    # Display pretty:
    $HTML -> P -> HR -> H3 -> t("Formatted log entries") -> _H3;

    # Key...
    $HTML -> small -> UL;
    $HTML -> LI;
    output $HTML "The request ID is the value assigned to the uniquifier ",
                 "for all hyperlinks/forms generated on that request. ",
                 "Thus, request ID '1234' means that all hyperlinks on the ",
                 "generated HTML look like 'script?mode=SRCH&u=1234'";
    $HTML -> LI;
    output $HTML "A '<' next to a request ID indicates that the back button ",
                 "was hit.  The page the user went back to is given below ",
                 "by log entry number, for convenience.";
    $HTML -> LI;
    output $HTML "Important events shown in BOLD.";
    $HTML -> _UL -> _small;

    $HTML -> P -> TABLE(BORDER=>1, WIDTH=>'100%');
    $HTML-> TR
	 -> TH -> t("#") -> _TH
	 -> TH -> t("Event") -> _TH
	 -> TH -> t("Date /") -> BR -> t("Time") -> _TH
	 -> TH -> t("Mode /") -> BR -> t("Submode") -> _TH
	 -> TH -> t("Request ID") -> _TH
	 -> TH -> t("Came from") -> BR -> t("Request ID") -> _TH;
    $HTML-> _TR;
    my $eventno = 0;
    my $requestno  = 0;
    foreach $entry (@$entries) {
	++$eventno;
	$HTML-> TR(VALIGN=>'TOP');
	
	# Event number:
	$HTML -> TD -> B -> t($eventno) -> _B -> _TD;

	# Event:
	my $event = uc($entry->event);
	if ($event !~ /^(MODE)$/) {
	    $HTML -> TD -> t_B($event) -> _TD;
	}
	else {
	    $HTML -> TD -> t($event) -> _TD;
	}
	if (uc($entry->event) eq 'MODE') {    # new request!
	    ++$requestno; 
	}

	# Date and time:
	my @dt = split(/\s+/, nice_datetime($entry->datetime_str));
	$HTML -> TD -> t($dt[0]) -> BR -> t($dt[1]) -> _TD;

	# Mode and submode:
	$HTML -> TD
	      -> t($entry->value('mode')) -> BR 
	      -> t($entry->value('submode')) -> _TD;
	
	# Request:
	my $upage = $entry->value('upage');
	$HTML -> TD -> t($upage) -> _TD;

	# Came from:
	my $ufrom = $entry->value('ufrom');
	my $uline = '';
	if ($ufrom > 1000000) {
	    $ufrom = '';
	}
	elsif ($ufrom and ($requests{$ufrom}[0] != ($requestno - 1))) {
	    $uline = $requests{$ufrom}[1];    # index by event no
	    $ufrom .= '<';
	}
	$HTML -> TD -> t($ufrom);
	$HTML -> BR -> small -> t("(Went back to #$uline)") -> _small 
	    if $uline;
	$HTML -> _TD;
        $HTML -> _TR;

	# Remember both the event and request number for this page:
	$requests{$upage} = [$requestno, $eventno];
    }
    $HTML -> _TABLE;

    # Display raw:
    $HTML -> P -> HR -> H3 -> t("Raw log entries") -> _H3;
    $HTML -> PRE -> small;
    $HTML -> TABLE;
    $eventno = 0;
    foreach $entry (@$entries) {
	++$eventno;
	$HTML -> TR
	      -> FONT(SIZE=>'+1') -> B -> t("$eventno:\n") -> _B -> _FONT;
	$entry->print(\*STDOUT);
	$HTML -> _TR; 
    }
    $HTML -> _TABLE;
    $HTML -> _small -> _PRE;
    1;
}

#------------------------------------------------------------
# send_sessions_doc PARAMHASH
#------------------------------------------------------------
# I<Instance method.>
# Output the document showing the sessions.  Params are:
#
#     List => the sessions to list (see show sessions); ,mutex with
#     Entries => the log entries from the chosen session

sub send_sessions_doc {
    my ($self, %params) = @_;
    
    # Start page:
    $self->start_page(Title => $Title);

    # Start main form...
    $HTML -> HR -> script_FORM;
    $HTML -> msubmit_flag;
    $HTML -> hidden('mode'  =>$REQ->param('mode'),
		    'start' =>$REQ->param('start'),
		    'stop'  =>$REQ->param('stop'));

    # Show sessions...
    if ($params{List}) {
	$self->print_sessions($params{List});
    }
    elsif ($params{Entries}) {
	$self->print_session($params{Entries});
    }

    # End form...
    $HTML -> _script_FORM;


    # Date selection form (only if listing)...
    if ($params{List}) {
	$self->print_dateselect_form;
    }

    # The signature:
    $self->end_page();
    1;    
}

#------------------------------------------------------------
# print_dateselect_form
#------------------------------------------------------------
sub print_dateselect_form {
    my $self = shift;

    $HTML -> HR;
    $HTML -> script_FORM;
    $HTML -> hidden('mode'    => $REQ->param('mode'),
		    'submode' => 'ALL');

    # The directions:
    $HTML -> directions
	  -> t("Enter the earliest start and latest stop dates for stat ".
	       "files to scan. Format is YYYY or YYYY-MM or YYYY-MM-DD. ".
	       "It's okay to leave either or both blank.")
	  -> _directions;

    $HTML -> TABLE
	  -> TR
	  -> TD(NOWRAP=>undef) 
	      -> t_B("Start: ")
	      ->input_text(Name=>'start', Size=>11)
	  -> _TD
	  -> TD(NOWRAP=>undef) 
	      -> t_B("Stop: ")
	      ->input_text(Name=>'stop', Size=>11)
	  -> _TD
	  -> TD
	      -> INPUT(TYPE=>'submit', 
		       VALUE=>'List the sessions in this period');
    $HTML -> _TD -> _TR -> _TABLE;
    $HTML -> _script_FORM;       
}


#------------------------------------------------------------
#
# Submode handlers...
#
#------------------------------------------------------------

#------------------------------------------------------------
# submode_DEFAULT
#------------------------------------------------------------
# What to do if invoked with no explicit submode...

sub submode_DEFAULT {
    shift->submode_CHOOSE;
}

#------------------------------------------------------------
# submode_CHOOSE
#------------------------------------------------------------
# What to do if invoked with no explicit submode...

sub submode_CHOOSE {
    my $self = shift;

    $self->start_page(Title => $Title);
    $self->print_dateselect_form();   
    $self->end_page();
    1;    
}

#------------------------------------------------------------
# submode_ALL
#------------------------------------------------------------
# Show all the known sessions.

sub submode_ALL {
    my $self = shift;

    # Get files to scan:
    my @files = $self->stat_files($REQ->param('start'), $REQ->param('stop'));
    @files or puke 549, "no stat files to scan!";

    # Scan files for all sessions in them:
    my $list = $self->scan_sessions(\@files) || puke 549, "no sessions!";

    # Output:
    $self->send_sessions_doc(List=>$list);
}

#------------------------------------------------------------
# submode_ONE
#------------------------------------------------------------
# Show a single session:

sub submode_ONE {
    my $self = shift;

    # Get files to scan:
    my @files = $self->stat_files($REQ->param('start'), $REQ->param('stop'));
    @files or puke 549, "no stat files to scan!";

    # Scan files for that one session in them:
    do_files IU::EventEntry 
	Files => \@files,
	Action => sub {
	    my $entry = shift;    
	    my $sid = $REQ->param('sessid');
	    push @Entries, $entry if ($entry->value('sessid') eq $sid);
	};

    # Output:
    $self->send_sessions_doc(Entries=>\@Entries);
}

#------------------------------------------------------------
1;
