package IW::ModeCART;

=head1 NAME

IW::ModeCART - display/manage the shopping cart

=head1 NOTES

The "CHANGES" hash has the following fields to describe how the
cart's  contents were just changed:

    WhatSelected   What kinds of things ('package', 'data set') user
                      selected.
    NumSelected    Number of things that user selected.
    NumGranules    Number of granules in across all their selections.
    OutsideCart    How many selected granules were *not* in the cart?
    InsideCart     How many selected granules were *already* in the
                      cart?
    Rejected       How many selected granules could not be added to 
                      the cart?
    Change         'ADD' or 'REMOVE'.

Internally, these additional fields are set:

    Sort http://lyta.gsfc.nasa.gov:80/~joe/edg-shadow          The sorting algorithm in use.

=head1 RCS INFO

$Id: ModeCART.pm,v 1.175 2003/09/24 14:55:00 solomon Exp $

=cut

use IU;
use IU::Auth;
use IU::Log;
use IU::Granule;
use IU::Dataset;
use IU::HTAccess;
use IU::InvStore;
use IU::Package;
use IU::MediaMaximumCapacity;

use IW::App;
use IW::Note;
use IW::User;
use IW::Disclaimers;
use IW::Util qw(:url);
use IW::CartList;
use IW::MediaOrderLimit;
use IW::Order;

# use Data::Dumper;  # for debugging

@ISA = qw(IW::Mode);

use strict;

#------------------------------------------------------------
#
# Globals
#
#------------------------------------------------------------

# The databases:
my $RESULTS;
my $RESULTSLIST;
my $CART;
my $WORKBENCH;
my $INVSTORE;
my $INVSTORELIST;
my $PARTIALCART;   # Used to save partial cart when submitting partial order.

# Class to evaluate order-media limits.
my $MEDIAORDERLIMIT; 

# Hash containing medias too small for this granule.
my %SmallMedias;

# These globals and the subroutine p make code for printing
# sentences using the correct plurality nicer to look at and
# easier to maintain.  Format of each pair in %plurals is:
# item name => ['singular form','plural form']
# $plural is 0 or 1, the index for the array. So 0 = singular,
# 1 = plural.  Note this means we can check: if ($plural).

my %plurals = (item => ['item', 'items'],
               has  => ['has',  'have'],
               is   => ['is',   'are'],
               it   => ['it',   'they'],
               was  => ['was',  'were'],
               's'  => ['',     's'],
              );
my $plural;

# Hash containing the names of the $user subroutines
my %user_subs =
   ( 'FTPHOST' => 'ftp_host',
     'FTPPASSWORD' => 'ftp_password',
     'FTPPUSHDEST' => 'ftp_pushdest',
     'FTPUSER' => 'ftp_user',
     'USERSTRING' => 'ftp_userstring');
my @user_subs_keys = keys %user_subs;
my $sub_key;

my $pageNumber; # Sub page number for order or subset options
my $min; # Minimum Sub page number required for order or subset options

#------------------------------------------------------------
# set_small_media DATASET GRANULE PACKAGES
#------------------------------------------------------------ 
sub set_small_media {

   my $dset = shift;
   my $gran = shift;
   my $pkgref = shift;

   %SmallMedias = ();

   # loop through all the package combos and build a hash of
   # unique media types;
   my $package;
   my %combo_types;
   my @combos;
   my $combo;
   my $media_type;
   foreach $package (@{$pkgref}) {
      foreach $combo ($package->combos) {
         $media_type = $combo->media_type;
         $combo_types{"$media_type"} ='1';
      }
   }

   # Check to see if any of the media capacity limit applies to this package.
   # If some do then output the table.
   my $granule_size = $gran->size_mbecs_granule;
   my $mediamaximumcapacity;

   foreach $mediamaximumcapacity (@{$dset->mediamaximumcapacity}) {
      if ($granule_size > $mediamaximumcapacity->capacity) {
         $SmallMedias{$mediamaximumcapacity->type} =
                             $mediamaximumcapacity->capacity;
      }
   }
   return 1;
}

#------------------------------------------------------------
# print_limit_msg
#------------------------------------------------------------
# If the user tried to select a media type that is too small to
# fit the order, print a message at the top of the shopping cart
# that informs the user of the limits with the ordering media.
#

sub print_limit_msg {
  my ($ready, $limits) = @_;

  # If limits are not found, do not print any message, return.
  defined ($limits) or return;

  # -----------------------------------------------------
  # If limits were found, get the information out of the
  # limits hash.  The limits hash information should include:
  #    ItemId:     granule id that reached the limit.
  #    DataCenter: data center where the order is going.
  #    MediaLimit: capacity limit for the media type,
  #                it could be in MB or in number of granules.
  #    MediaType:  media type with limitation. 
  # -----------------------------------------------------
  my $item  = $limits->{ItemId};
  my $daac  = $limits->{DataCenter};
  my $limit = $limits->{MediaLimit};
  my $type  = $limits->{MediaType};

  $HTML -> TABLE(BORDER=>0, CELLPADDING=>0, CELLSPACING=>0 );
  $HTML -> TR -> TD(VALIGN=>'CENTER', ALIGN=>'CENTER');
  # Print the warning icon, and the message.
  $HTML -> warning;

  $HTML -> _TD -> TD;

  $HTML ->t("Your selected order options were applied to a 
             limited number of granules because of a size 
             limitation of $limit per order imposed by data 
             center $daac for media type $type.");

  if (!$ready) {
     $HTML -> P
        -> t("You have two options:")
        -> BR
        -> t(" 1) Submit a partial order for those granules in the cart
                  which are ready to order and then submit a second order
                  for the remaining granules or");
     $HTML -> BR 
        -> t(" 2) Select a different media type for all data granules or
                  for individual data granules in the shopping cart.");
  }

  $HTML -> _TD -> _TR;
  $HTML -> _TABLE;
}

#------------------------------------------------------------
# print_partial_order_link
#------------------------------------------------------------
# If the user tried to select a media type that is too small to
# fit the order, the system will allow the user to submit a partial
# just for the items with ordering options chosen.  To allow this
# a link is printed on top of the cart table.  If there are no
# limitations, the partial order link is not printed.
#

sub print_partial_order_link {
  my ($limits) = @_;

  # If limits are not found, do not print the link, return.
  defined ($limits) or return 1;

  my $url = script_url('mode'=>'CART', 'submode'=>'PARTIAL');

  # Use the same color background as the cart table.
  $HTML -> TR(BGCOLOR=>"#EEEEEE") -> TD -> CENTER;
  $HTML -> hypertext("Submit a Partial Order", $url);
  $HTML -> _CENTER -> _TD -> _TR;
}

#------------------------------------------------------------
# print_change_msg CHANGES SUMMARY_STYLE
#------------------------------------------------------------
# Print the message at the top of the shopping cart document
# describing recent changes, as given by CHANGES.
# If SUMMARY_STYLE is ONE-LINE-PER-ITEM, and what they selected
# is 'just data set', then the messages must reflect the fact
# that only one data set is added at a time in this format.
#
# Most of the if-else structures are just to get the wording
# to match what is really happening and to get the plurality
# correct (I hated seeing "1 of your selections were ..." LPR).
# If you're wondering what p('was') etc. does, you didn't read
# the documentation at the top on %plurals.  Note that there
# are two levels of plurality: how many items the user selected
# and how many items are being added, removed, or rejected.

sub print_change_msg {
  my ($changes) = @_;
  
  # Was there in fact a change?
  defined($changes->{WhatSelected}) or return 1;
  
  # Number selected is used so much lets save it in a variable.
  my $num = $changes->{NumSelected};
  
  # $single is either true if the number selected equals 1, or if the
  # number of granules in the data sets selected equals 1.
  # $single is NOT usually the opposite of $plural (which may be
  # based on the number outside, inside, or rejected).
  my $single = ($num == 1);
  
  # Set the correct plurality of some phrases based on whether
  # selected single.
  my $S = ($single) ? '' : 's';
  
  # Tell them how many granules or data sets they selected.
  # (Note perl's obtuse switch statement looks like 'for' loop,
  #  It only loops once using the value, looking for a match).
  for ($changes->{WhatSelected}) {
    # If adding orderable data sets to cart.
    /^just data set$/ and do {
      if ($single) {
        print "<P> You chose to add <B>1</B> data set.\n";
      } else {
        print "<P> Even though you selected <B>$num</B> data ",
        "sets,\n only the data set whose add-to-cart icon ",
        "you pressed was added to the shopping cart.\n";
        $single = 1;     # Act like they selected one.
      }
      last;
    };

    # Else If adding granules from selected data sets to cart.
    /^data set$/ and do {
      # Set plurality based on number of granules being moved.
      $plural = (($changes->{NumGranules}==1) ? 0 : 1);
      print '<P> There ', p('was'), ' <B>',
      $changes->{NumGranules}, '</B> data granule', p('s'),
      " in the <B>$num</B> data set$S you selected.\n";
      $single = !$plural;  # Act like they selected the granules.
      last;
    };

# We don't want to print the long message any more, commented
# out in case we change our mind later.
#    # Else If selected granules to add to cart.
#    /^granule/ and do {    # Also catches 'granules'
#      print '<P> You selected a total of <B>',
#      "$num</B> data granule$S.\n";
#      last;
#    };
#    # Else default "items".
#    print "<P> You selected a total of <B>$num</B> item$S.\n";
#    # Just in case someone changed the meaning of WhatSelected.
#    last;
  }

  #
  # Total number of acceptable orders = $changes->{OutsideCart}
  #                                   + $changes->{InsideCart});  
  # Print "your selection" or "x of your selections".
  my $selection;
  
  # Tell them what we did with those selections...
  if ($changes->{OutsideCart} || 
      $changes->{InsideCart} || 
      $changes->{Rejected}) {
    
# We don't want to print the long message any more, commented
# out in case we change our mind later.
#    print 'Of them ...' if (!$single);
#    print "<UL>\n";
    
    # Within each check for changes we check for number of
    # changes and number selected so we get the grammar right.
    # Note that if single item selected, OutsideCart is <= 1.
    
    # Report selections that were not already in the cart:
    if ($changes->{OutsideCart}) {
      # Set plurality based on the number outside.
      $plural = (($changes->{OutsideCart}==1) ? 0 : 1);
      
      # Print "your selection was" or "x of your selections were".
      $selection = ($single ? 'Your selection was ' :
                    '<B>'. $changes->{OutsideCart} .
                    '</B> of your selections '. p('was'). ' ');
    }
    # Report selections that were already in the cart: 
    if ($changes->{InsideCart}) {
      # Set plurality based on number that were in the cart.
      $plural = (($changes->{InsideCart}==1) ? 0 : 1);
      
      # Print "your selection was" or "x of your selections were".
      $selection = ($single ? 'Your selection was ' :
                    '<B>'. $changes->{InsideCart} .
                    '</B> of your selections '. p('was'). ' ');
      # If we are adding them:
      if ($changes->{Change} eq 'ADD') {
        print '<LI> ', $selection, '<B>added to your shopping ',
        'cart previously</B>, so ', p('it'), ' ', p('was'),
        " already there.</LI>\n";
      } 
    }
    # Report selections that were rejected:
    if ($changes->{Rejected}) {
      
      # Set plurality based on number that were rejected.
      $plural = (($changes->{Rejected}==1) ? 0 : 1);
      # If we are adding them:
      if ($changes->{Change} eq 'ADD') {
        # If user only had one selected:
        if ($single) {
          print "<LI> Your selection ";
        } 
        else {
          print "<LI> <B>$changes->{Rejected}</B> of your ",
          "selections ";
        }
         $HTML ->t_B("cannot be ordered, ");
    print 'so ', p('it'), ' ', p('was'), " not added to your shopping cart.\n";
      }
    }
    print "</UL>\n";
  }
  1;                 
}

#------------------------------------------------------------
# print_status_msg 
#------------------------------------------------------------
# Print the message at the top of the shopping cart document
# describing current cart contents.

sub print_status_msg {
  my ($ready, $granules) = @_;

  my $items = 0;
  $items = int(@$granules);

  # Nothing there?
  if ($items <=0) {
    print "<hr><H2> " .
      "There are no data sets or data granules in the Shopping Cart! </H2>";
    return;
  }

  # Print the overall summary:
  # Set plurality based on number of items.
  $plural = (($items == 1) ? 0 : 1);
  print '<P> <B>', $items,
  '</B> ', p('item'), ' in your shopping cart';

  if ($ready == 1) {
    print ", 1 item is ready to be ordered.<BR>\n";
  }
  else {
    print ", <B>$ready</B> items ready to be ordered.<BR>\n";
  }

  # If not all have been assigned packages, prompt the user:
  if ($ready < $items) {
    print '<P> Before you can go to <B> Step 2: Order Form </B> ',
    "you must choose ordering options <BR>\n",
    'using the <B>Choose Options</B> link next to each ',
    "item that needs it.\n";
  }
}

#------------------------------------------------------------
# check_items_ready CHANGES,GRANULES,DATASETS
#------------------------------------------------------------
# Figure out the status of the cart, find out how many items
# are ready to be ordered.  The output of this subroutine
# will be used to inform the user with a status message.

sub check_items_ready {
  my ($changes, $granules) = @_;
  
  my $items = 0;
  my $ready = 0;
  $items = int(@$granules);
  
  # Nothing there?
  return (1, $ready) if ($items <=0);

  # Something there!
  # Figure out how many items have either (1) been assigned 
  # packages, or (2) do not need packages assigned to them:
  my $gran;
  my $dset;
  my ($package_ready, $email_ready, $cost) = (0, 0, 0.0);
  my %combos;
  my $combo;
  foreach $gran (@$granules) {
    
    # Readiness:
    my $status = $gran->order_status;
    if    ($status eq 'GO_PKGS')  { ++$package_ready }
    elsif ($status eq 'GO_EMAIL') { ++$email_ready }
    elsif ($status eq 'GO_EMAIL_PKGS') { ++$package_ready }
    
    # Add package combo uniquely:
    foreach $combo (@{$gran->package_combos}) {
      $combos{$combo->merged_key} = $combo;
    }
  }

  $ready = $package_ready + $email_ready;
  
  # Figure out the *real* total cost:
  foreach $combo (values %combos) {
    $cost += ($combo->cost * $combo->copies || 0);
  }
  
  # Let the user know that the cost for e-mail items is unknown.
  if ($email_ready > 0) {
    # Set plurality based on e-mail items without packages.
    $plural = (($email_ready == 1) ? 0 : 1);
    print 'The cost for the ', $email_ready, ' ', p('item'),
    ' that ', p('has'), ' no ordering options, is ',
    "unknown.<BR>\n";
  }

  ($ready < $items) ? return (0, $ready) : return (1, $ready);
}

#------------------------------------------------------------
# print_order_cost -- print the cost if there are any fees.
#------------------------------------------------------------
sub print_order_cost {
  my ($changes, $granules) = @_;

  my $gran;
  my ($package_ready, $cost) = (0, 0.0);
  my %combos;
  my $combo;

  foreach $gran (@$granules) {
        # Readiness:
    my $status = $gran->order_status;
    if    ($status eq 'GO_PKGS')  { ++$package_ready }
    elsif ($status eq 'GO_EMAIL_PKGS') { ++$package_ready }

    # Add package combo uniquely:
    foreach $combo (@{$gran->package_combos}) {
      $combos{$combo->merged_key} = $combo;
    }
  }

  # Figure out the *real* total cost:
  foreach $combo (values %combos) {
    $cost += ($combo->cost * $combo->copies || 0);
  }

  #
  # Display break down of costs with total.
  #
  if ($package_ready > 0) { 
     print '<tr bgcolor="#EEEEEE">';
     print '<td>';
     $HTML -> CENTER -> TABLE;
     #
     # Sum of all combos selected. 
     #

     if ($cost > 0) {
        my $coststring = ($package_ready == 1) ? 'cost' : 'costs';
        print "<TR><TD COLSPAN=2>Subtotal of per item $coststring:</TD>";
        $cost = sprintf("%0.2f",$cost);
        print "<TD ALIGN=\"right\">\$$cost</TD></TR>";

     }
     #
     # Dataset handling fee:
     #
     my %HandlingFee = %{$changes->{HandlingFee}};
     if (int(keys %HandlingFee) > 0) {
        my $subhandlingfee = 0;
        foreach my $key (keys %HandlingFee) {
           my $handlingfee = $HandlingFee{$key}; 
           $handlingfee = sprintf("%0.2f",$handlingfee);
           print "<TR><TD>Handling fee for </TD><TD> $key</TD>";
           print "<TD ALIGN=\"right\">\$$handlingfee</TD></TR>";
           $subhandlingfee += $handlingfee; 
        }
        $cost += $subhandlingfee;
     }
     #
     # Media handling fee
     #
     my %MediaHandlingFee = %{$changes->{MediaHandlingFee}};
     if (int(keys %MediaHandlingFee) > 0) {
        my $submediafee = 0;
        foreach my $key (keys %MediaHandlingFee) {
           my $mediafee = $MediaHandlingFee{$key};
           $mediafee = sprintf("%0.2f",$mediafee);
           print "<TR><TD>Handling fee for</TD><TD> $key</TD>";
           print "<TD ALIGN=\"right\">\$$mediafee</TD></TR>";
           $submediafee += $mediafee;
        }
        $cost += $submediafee;
     }
     # 
     # Display total cost:
     #
     if ($cost > 0) {
        print '<TR><TD COLSPAN=2><B>Total cost (US):</B></TD>';
        $cost = sprintf("%0.2f",$cost);
        print "<TD ALIGN=\"right\"><B>\$$cost</B></TD></TR>";
        $HTML -> TR -> TD(COLSPAN=>3);
        $HTML -> t("Please do not enter credit card or other payment ".
                   "information on this web site. If there are charges ".
                   "associated with your order, an email will be sent to ".
                   "you with payment information, if applicable."); 
        $HTML ->_TD->_TR;
     }
     $HTML -> _TABLE; 
     $HTML -> _CENTER;
     $HTML -> _TD -> _TR;
  } else {

     $HTML -> TR -> TD(bgcolor=>"#EEEEEE");
     $HTML -> t("Please do not enter credit card or other payment ".
                "information on this web site. If there are charges ".
                "associated with your order, an email will be sent to ".
                "you with payment information, if applicable.");
     $HTML ->_TD->_TR;
  }
}

#------------------------------------------------------------
# print_menu -- print the main menu
#------------------------------------------------------------
sub print_menu {
  my ($ready, $gkeys) = @_;
  
  $HTML -> msubmit_flag;
  
  # Only print "go to next step" button if all items are ready 
  # to order:
  if ($ready) {
    # Changed the submode of ORDER to go to NEWFORM so that we
    # can send an ECS Profile Request if the ECSProfileGateway is set.
    $HTML -> FONT(SIZE=>'+3') -> B
          -> msubmit_button(CAPTION => '  Go to Step 2: Order Form ',
                            QUERY =>['mode' => 'ORDER',
                                     'submode' => 'NEWFORM']);
    $HTML -> _B -> _FONT;
    $HTML -> P;
  }

  my $url = script_url('mode'=>'CART', 'submode'=>'REM', 'remove'=>'ALL');
  $HTML -> hypertext("Empty Entire Shopping Cart", $url);
  
  1;
}

#------------------------------------------------------------
# send_doc PARAMHASH
#------------------------------------------------------------
# Output the granule listing document.  Arguments are...
#
#    Changes     => the changes we just made (or undef)
#    ListMode    => mode to use when re-sorting.
#
# NOTE: Because it is possible to create a cart and then remove
# everything from it, leaving a GDBM file, it is perfectly legal
# to have an empty cart.  We should not have a failure in this case.

sub send_doc {
  my %params = @_;
  my $listmode = $params{ListMode};
  my $changes  = $params{Changes} || {};
  my $limits   = $changes->{Limits};
  my $gran;
  my $ready;
  my $itemsready;
  
  # Get display preferences:
  my ($user, $display);
  unless (($user = $SESSION->state->{User}) and
          ($display = $user->display)) {
    $display = new IW::UserDisplay;
  }
  
  IW::CartList->update_virtual_keys('ckey',$CART);
  
  #------------------------------
  # SETUP GRANULE LIST...
  #------------------------------
  
  # Create cart list object:
  my $items = [$CART->granules,$CART->just_datasets()];
  my $nextpage = $REQ->param('nextpage');
  my $page = $CART->page('ckey',$nextpage);
  my $cartlist = IW::CartList->new(Elems=>$items,
                                   Page=>$page);
  $cartlist->origin('CART');
  $cartlist->{DontPassParams} = 1;
  # Sort all items using (and returning) most appropriate sort key:
  my $sort = '';
  $cartlist->sort(\$sort);
  my $sorted_keys = [map {$_->key}  @{$cartlist->elems}];
  #------------------------------
  # OTHER SETUP...
  #------------------------------
  
  # Compute total cost: and mark if on the workbench.
  my $totalcost = 0;
  
  my ($gran_dataset_id, $selected_dataset);
  my %list_fees;
  my %media_fee_hash; 
 
  # either dataset or granule
  foreach my $item (@$items) {
    # Get the total Package cost WITHOUT the Handling Fee.
    $totalcost += ($item->total_cost || 0);
    
    # Now to add the Handling Fee:
    # What do we have? A Dataset or Granule?
    # If it's a granule, get its parent Dataset.
    if ($item->order_type eq 'GRANULE') {
      $gran_dataset_id =  $item->dataset_key;
      $selected_dataset = $CART->dataset( $gran_dataset_id );
    }
    else { $selected_dataset = $item; }
    
    # Check whether the Granule/Dataset HAS package information
    # (meaning it's ready to be ordered) 
    if ($item->order_status eq 'GO_PKGS')  {
      my $combo;
      
      # Look at the various packages that were selected :
      foreach $combo (@{$item->order_pkgcombos}) {
               
         # Check combo to see if media handling fee applies.
         # Add to media_fee hash if there is a match.
         my $mediahandlingfee;
         foreach $mediahandlingfee (@{$selected_dataset->mediahandlingfee}) {
            my $type = $mediahandlingfee->type;
            my $archive_id = $selected_dataset->archive_id;
            #
            # How the hash key is constructed determines at which level
            # media handling fee is calculated.
            #
            # A hash key of $archive_id:$type means the fee is calculated
            # once per daac for that media type.
            #
            # If the fee needs to be calculated at the dataset level
            # then add dataset id as part of the hash key, e.g.
            # my $dataset_id = $selected_dataset->id;
            # $media_fee_hash{"$archive_id:$dataset_id:$type"} =
            #     $mediahandlingfee->fee;
            #
            if ($combo->media_type eq $type) {
               $media_fee_hash{"$archive_id:$type"} = $mediahandlingfee->fee;
               last;
            }
         }
 
        # There is a handling fee  in the DATASET Group
        # Comment/Fee hasn't already been added to the hash?
        if (!($combo->apply_handling_fee eq 'N') && 
            $selected_dataset->handling_fee      && 
            !$list_fees{$selected_dataset->id}) {
          
          # We may have multiple Handling Fees, so parse through 'em. 
          my @multiple_handling_fee = (@{$selected_dataset->handling_fee});
          my $single_handling_fee;
          
          foreach $single_handling_fee ( @multiple_handling_fee) {
            # If there is only one, then the Handling Fee price
            # is the Hash value 
            # ELSE add this price to the previous value.
            if (int(@multiple_handling_fee) > 1) {
              $list_fees{$selected_dataset->id} +=
                $single_handling_fee->price;
            } else {
              $list_fees{$selected_dataset->id} =
                $single_handling_fee->price;
            }
          }
        }
      }
    }
  }
  
  # Pass dataset handling fee in as a hash ref to $changes.
  $changes->{HandlingFee} = \%list_fees;
  # Pass media handling fee in as a hash ref to $changes.
  $changes->{MediaHandlingFee} = \%media_fee_hash;
  $changes->{Cost} = sprintf("%0.2f", $totalcost);
  
  #------------------------------
  # OUTPUT!
  #------------------------------
  # Set the indicator. We can do this any time, but this is as
  # good a time as any other. 
  $SESSION->{Navigator}->{Indicator} = "OrderData";
  
  # Print page header:
  $APP->start_page(SuperTitle=>'Shopping Cart',
                   Title=>"Step 1: Choose Ordering Options",
                   Context =>'shopcart');
  
  # Adjust cart summary, and print message:
  $changes->{Sort} = $sort;

  print_change_msg($changes);

  #-------------------------
  # Figure out the status of the cart.
  # 
  # Value $ready is a flag that let us know if
  # the cart is ready to continue with the order.
  #
  # Value $itemsready is the number of items
  # that have ordering options selected.
  #-------------------------
  ($ready, $itemsready) = check_items_ready($changes, $items);

  #-------------------------
  # If any media limits are reached, notify the user of
  # the problem, otherwise notify order status.
  #-------------------------
  print_limit_msg($ready, $limits) if ($limits);
  print_status_msg($itemsready, $items) unless ($limits);
  
  $HTML -> HR;

  #-------------------------
  # Add button to reach the Customize Order Options page.
  #
  # * Note: this button needs to have a url as a ref.
  # When this button is part of a submit form, it behaves
  # unpredictable when the cart is empty.
  #
  #-------------------------
  my $href = script_url ('mode' => 'PREFSPKG', 'origin' => 'CART');
  $HTML -> TABLE(SIZE=>"75%",BORDER=>0,CELLSPACING=>3)
        -> TR ->TD(VALIGN=>'TOP',ALIGN=>'RIGHT');
  $HTML -> button (SRC  => "${IU::DocsUrl}btns/customize_order.gif",
                   ALT  => 'Customize Order Options',
                   HREF => $href);
  $HTML -> _TD -> TD(VALIGN=>'TOP');
  $HTML -> t("to set order options automatically for future orders ");
  $HTML -> t("from this guest session or your registered user account.");
  $HTML-> _TD -> _TR -> _TABLE;
  $HTML -> BR;

  if (int(@{$cartlist->elems}) > 0) {        # we have items in cart
    $HTML -> script_FORM;
    # Print menu above
    print_menu($ready, $sorted_keys);

    # Print outer cart table in case there is a cost row to be printed.
    $HTML->TABLE(BORDER=>0, CELLSPACING=>0, CELLPADDING=>0);

    #----------------------
    # Only print the partial order link if: 
    # - Limits were reached with any media type.
    # - And there are items in the cart without
    #   ordering options selected.
    #
    # If all the items have selections set, the
    # cart should display the continue with Step 2
    # button.
    #----------------------
    print_partial_order_link($limits) if ($limits && !$ready); 

    print_order_cost($changes, $items);
    #
    # Display payment warning: 
    #
    $HTML -> TR ->TD;
    # Print cart table
    $cartlist->print(NoRule=>1);
    $HTML  -> _TD -> _TR;
    print "</table>\n";
    $HTML -> _script_FORM;                        # end form

  } else {
    $HTML -> P -> HR;
    $HTML -> UL
          -> LI  
          ->t("Use the ") 
          -> t_B("Results:Granule") 
          ->t(" navigation link to go to the Data Granule Listing Page ")
          -> LI 
          ->t(" Select some granules ") 
          ->LI 
          ->t(" Then use the \"Add to Order\" button " .
              "to put them into your Shopping Cart . ")
          -> _UL
          -> t_B ("OR");
    
    $HTML -> UL
          -> LI
          ->t("Use the ")
          -> t_B("Results:Data Set") 
          ->t(" navigation link to go to the " .
              "Results Page ") 
          -> LI 
          ->t(" Select some data sets ")   
          ->LI 
          ->t(" Then use the \"Add data granules to order\" button " .
              "to put them into your Shopping Cart . ")
          -> _UL;
  }
  
  $APP->end_page;
  1;
}

#------------------------------------------------------------
# The Quality Flag Disclaimer gets displayed AFTER the standard Source/Dataset
# Disclaimer
#------------------------------------------------------------

sub print_QAFlag_disclaimer {
 
  my   $hold_status = shift;
 
  # Open static valids database:
  my %WWWVALIDS;
  if (-e "${IU::LibDir}/conf/wwwvalids.gdb") {
    eval {tie %WWWVALIDS, 'IU::Store', 
            "${IU::LibDir}/conf/wwwvalids.gdb",
            'READER'};
    $@ and return error "tie failed: $@";
  }
 
  my $valids = $WWWVALIDS{"disclaimer:QAFLAGS"};
  my $QA_text = $valids->{Text};

  # Add more text to the Disclaimer if the Sci Flag = hold
  if ($hold_status eq 'yes') { 
     my $hold_valids = $WWWVALIDS{"disclaimer:HOLD"};
     my $Hold_text = $hold_valids->{Text};
     $QA_text = $QA_text."<P> <P>".$Hold_text  ;
  }
 
  $APP->suppress_links(1);
  
  # Start page:
  $APP->start_page(SuperTitle=>'Shopping Cart',
                   Title=>"Quality Assurance Disclaimer :");
  $HTML->PRE;
  print <<EOF;
$QA_text
EOF
  
  $HTML->_PRE ->P;
  
  
  # Start form:
  $HTML->script_FORM;
 
  $HTML -> msubmit_flag; 
  # Copy the CGI parameters for the next mode to handle.
  # Skip tag if it has ? in it (ie. the "msubmit?mode=cart&" tag)
  # and the sid and endform since APP will do those for us.
  # Also skip mode and submode since we'll get those from the button.
  my $key;
  foreach $key ($REQ->params) {
    next if ($key =~ /^msubmit\?/ or $key =~ /^msubmit/ or 
             $REQ->param($key) =~ /^msubmit/ or  
             grep /^$key$/, qw(sid endform mode submode gkey dskey keys));
    $HTML->hidden($key => $REQ->param($key));
  }
  
  # Since there can be more then one, we have to handle it ourselves.
  # Create a hidden tag for each dskey/gkey.
  my @dskeys = $REQ->param('dskey');
  my @gkeys = $REQ->param('gkey');
  my ($dskey, $gkey);
  
  my $return_to_sender = $REQ->param('prevmode') || 'GRANLIST';
  
  $HTML->CENTER
    ->msubmit_button(CAPTION => 'Accept',
                     QUERY =>['mode'   => 'CART',
                              'submode' => 'CONT_TO_ADD' ])
    ->hspace(2)
    ->msubmit_button(CAPTION => 'Reject',
                    QUERY =>['mode'   => 'CART',
                              'submode' => 'CLEANUP'])
    ->_CENTER;
  
  # End the form.
  $HTML->_script_FORM;
  $HTML->_B;
  
  $APP->suppress_key(1);
  $APP->end_page;
  1;
}

#------------------------------------------------------------
# print_disclaimer HASH
#------------------------------------------------------------
# Certain SOURCEs/DATASETs have been tagged, so that a disclaimer will be shown
# to the user for their confirmation of terms.
# Offer them the choice to continue or to cancel.
#
# Uses lib/conf/wwwvalids database file.

sub print_disclaimer {
  my %disclaimers = @_;
  
  # Get the list of sources.
  my @sources = keys %disclaimers;
  
  $APP->suppress_links(1);
  
  # Start page:
  $APP->start_page(SuperTitle=>'Shopping Cart',
                   Title=>"Data Quality Summary:");
  
  # One at a time.
  my $each_source ;
  
  # Print disclaimers for EACH of the tagged sources/datasets.
  foreach $each_source (@sources) {
    
    # Put the Disclaimer structured values into simple variables.
    my $the_msg = $disclaimers{$each_source}->{Text};
    my $the_url = $disclaimers{$each_source}->{URL} || undef; 
#       if (defined($disclaimers{$each_source}->{URL}));
    
    # If there is a DISCLAIMER_COMMENT, display it.
    $HTML ->HR
          ->B
          ->t($each_source." :")
          ->_B
          ->P
          ->PRE;
    print <<EOF;
$the_msg
EOF
    
    $HTML ->_PRE;

    # If there is a DISCLAIMER_URL, display it.
    $HTML ->hypertext("Disclaimer for $each_source.",
                      $the_url, TARGET =>'support')
          ->BR
        if ($the_url);
  }
  
  # Start form:
  $HTML->script_FORM;
  $HTML -> msubmit_flag; 

  # Copy the CGI parameters for the next mode to handle.
  # Skip tag if it has ? in it (ie. the "msubmit?mode=cart&" tag)
  # and the sid and endform since APP will do those for us.
  # Also skip mode and submode since we'll get those from the button.
  my $key;
  foreach $key ($REQ->params) {
    next if ($key =~ /^msubmit\?/ or $key =~ /^msubmit/ or
             $REQ->param($key) =~ /^msubmit/ or
             grep /^$key$/, qw(sid endform mode submode gkey dskey keys));
    $HTML->hidden($key => $REQ->param($key));
  }
  
  # Since there can be more then one, we have to handle it ourselves.
  # Create a hidden tag for each dskey/gkey.
  my @dskeys = $REQ->param('dskey');
  my @gkeys = $REQ->param('gkey');
  my ($dskey, $gkey);
  
  # What sources are we working with?
  foreach $each_source (@sources) {
    $HTML->hidden('disclaimed_source' => $each_source);
  }
  
  # We want the user to be able to be sent BACK to their previous screen.
  # Can either be 'WORKBENCH' , 'DSETLIST' or 'GRANLIST'
  #
  # WARNING: The ORIGIN tag refers to the DATABASE, NOT the previous Mode
  # Created PREVMODE tag to handle the previous mode.
  my $return_to_sender = $REQ->param('prevmode') || 'GRANLIST';
  
  # The submit buttons:
   
  $HTML ->CENTER
        ->msubmit_button(CAPTION => 'Accept - Continue to Shopping Cart',
                         QUERY =>['mode'   => 'CART',
                                  'submode' => 'SHOW_QA_FLAGS' ])
        ->hspace(2)
        ->msubmit_button(CAPTION => 'Reject - Return to Previous Page',
                    QUERY =>['mode'   => 'CART',
                              'submode' => 'CLEANUP'])

        ->_CENTER;
  
  # End the form.
  $HTML->_script_FORM;
  $HTML->_B;
  
  $APP->end_page;
  1;
}

#------------------------------------------------------------
# send_warn_pkg_apply GRAN DSET
#------------------------------------------------------------
# Warn the user that they are changing the apply package select
# from affecting only single granules to affecting all granules
# in the data set, which will override any previously selected
# options for other granules in this data set.  Offer them the
# choice to continue or to cancel.

sub send_warn_pkg_apply {
  my $gran = shift;
  my $dset = shift;
  
  $APP->suppress_links(1);
  
  # Start page:
  $APP->start_page(SuperTitle=>'Shopping Cart',
                   Title=>'Warning: Applying to Data Set');
  
  $HTML ->HR
        ->B
        ->warning
        ->H2
        ->color('ErrorText')
        ->t("Warning!")
        ->_color
        ->_H2
        ->_warning
        ->t("You have changed the choice from:")
        ->BR
        ->hspace(3);
  print "&#149";   # Bullet
  $HTML ->t(" Applying the ordering option selected to " .
            "just the one data granule: ");
  print $gran->id;
  $HTML ->BR
        ->t("to:")
        ->BR
        ->hspace(3);
  print "&#149";   # Bullet
  $HTML ->t(" Applying it to all data granules in the data set: ");
  print $dset->id;
  $HTML ->P
        ->t("This will replace your previously selected ordering options 
             for any data granules in this data set.")
        ->P;
  
  # Start form:
  $HTML->script_FORM;
  $HTML -> msubmit_flag; 
  # Copy the CGI parameters for the next mode to handle.
  # Skip tag if it has ? in it (ie. the "msubmit?mode=cart&" tag)
  # and the sid and endform since APP will do those for us.
  # Also skip mode and submode since we'll get those from the button.
  my @tags;
  my $key;
  foreach $key ($REQ->params) {
    next if ($key =~ /^msubmit\?/ or $key =~ /^msubmit/ or
             $REQ->param($key) =~ /^msubmit/ or
             grep /^$key$/, qw(sid endform mode submode));
    # handle multi-valued param tags.
    my $value;
    foreach $value ($REQ->param($key)) {
      $HTML->hidden( $key, $value);
    }
  }
  # Add tag saying the user has been warned.
  $HTML->hidden('warned' => '1');
  
  # The submit buttons:
  $HTML ->CENTER
        ->msubmit_button(CAPTION => 'Continue',
                         QUERY =>['mode'   => 'CART',
                                  'submode' => 'SETPKGS'])
        ->hspace(2)
        ->msubmit_button(CAPTION => 'Cancel',
                         QUERY =>['mode'   => 'PKGSLCT',
                                  'submode' => 'SELECT'])
        ->_CENTER;
  
  # End the form.
  $HTML->_script_FORM;
  $HTML->_B;
  
  $APP->end_page;
  1;
}

#------------------------------------------------------------
# subroutine p
#------------------------------------------------------------
# Note:
#    This routine has a 1 character name so that when used it won't
#    obscure the reading of a print string for you or me.  For the
#    same reason, $plural is global and not passed in.  Code-wise
#    it would make more sense to pass it, but that would make reading
#    the code much harder on maintainers.
# Description:
#    Takes a word and returns the singular or plural form of it
#    based on the current value of the global variable $plural.
#    The word must also be a key in the global hash %plurals.
# How it works:
#    1. The global variable $plural will be 0 if we want singular forms
#       or 1 if we want plural forms.
#    2. $_[0] is the passed word as a string.
#    3. $plurals{$_[0]} is the value in the hash for the passed word.
#       It is an anonymous array reference, like ['has','have'].
#    4. Pretend that the anonymous array reference were called A.
#       @A would be the array, and $A[0] would be 'has', $A[1] would
#       be 'have'.  Since the array reference is anonymous we use the
#       form ${ ... }[$index].
# Author: Larry Rosen
sub p {
  return (${$plurals{$_[0]}}[$plural]);
}

#------------------------------------------------------------
# login
#------------------------------------------------------------
sub login {
  
  # Log into CART database (creating it if necessary):
  $CART = IU::InvStore->open($SESSION->path_to('cart.gdb'), 'WRCREAT')
    || puke "couldn't open cart"; 

  # Log into WORKBENCH database (okay if not there):
  
  # eval {$WORKBENCH = 
  #    IU::InvStore->open($SESSION->path_to('workbench.gdb'), 'READER')};
  
  eval {
    $WORKBENCH = undef;
    return unless -e $SESSION->path_to('workbench.gdb');
    $WORKBENCH = 
      IU::InvStore->open($SESSION->path_to('workbench.gdb'),
                         'WRCREAT');
  };
  
  # Log into RESULTS database (okay if we can't, since someone might
  # have started and then aborted a search!!):
  
  eval {
    $RESULTS = undef;
    return unless -e $SESSION->path_to('results.gdb');
    $RESULTS =
      IU::InvStore->open($SESSION->path_to('results.gdb'),
                         'WRCREAT');
  };
  
  eval {
    $RESULTSLIST = undef;
    return unless -e $SESSION->path_to('resultslist.gdb');
    $RESULTSLIST =
      IU::InvStore->open($SESSION->path_to('resultslist.gdb'),
                         'WRCREAT');
  };
  
  # Figure out the one we'll use for the inventory: 
  my $z = $REQ->param('origin');
  $z = "" unless defined $z;
  $INVSTORE = ($z eq 'WORKBENCH') ? $WORKBENCH : $RESULTS;
  $INVSTORELIST = ($z eq 'WORKBENCH') ? $WORKBENCH : $RESULTSLIST;
  1;
}

#------------------------------------------------------------
# update_packages PARAMHASH
#------------------------------------------------------------
# We just came from the package-selection screen.
# Set packages for the given items.
# They may be datasets, granules, whatever.
# Params are:
#
#      ModifyItems => the items in the cart to change (e.g., all 
#                     grans in a dataset)
#      ChosenItem  => the item we chose packages for
#      Packages    => the packages for the item

sub update_packages {
  my %params = @_;
  my @modify_items = @{$params{ModifyItems} || []};
  my $chosen_item  = $params{ChosenItem};
  my @packages     = @{$params{Packages} || []};
  my $pkno;         # used in iteration
  my $dset;
  my $limits;  # used as reference to a hash when checking media limits.
  
  #------------------------------------------------------------
  # Get the media order limit object stored in the cart,
  # if it does not exist, create a new one.
  #------------------------------------------------------------
  $MEDIAORDERLIMIT = $CART->cart_medialimit_info() || new IW::MediaOrderLimit;

  my $floatscene = 0;  # are there any floating scene criteria.
  # Verify the packages:
  @packages or puke "no ordering options for chosen item!";    
  
  # Build a list of all combos chosen across packages:
  # E.g., ("1.1", "1.2", "2.2") 
  my @all_combo_ids_chosen;
  my @valid_combo_ids_chosen;
  foreach $pkno (1..int(@packages)) { 
    push @all_combo_ids_chosen, $REQ->param("pk-choice.$pkno");

    # --------
    # Filter out anything that is not a valid combo selection.
    # For some options, the default selection is (none) or
    # "I want no items from this option", we need to filter it
    # out from the list of valid options.
    # The cart now displays the media type selected, if there are
    # multiple medias selected the cart notifies so.  The default 
    # "I want no items from this option" options is treated as
    # a valid selection, therefore it needs to be filtered out 
    # from the selected medias to be displayed in the cart.
    # --------
    push @valid_combo_ids_chosen, $REQ->param("pk-choice.$pkno")
                       unless ($REQ->param("pk-choice.$pkno") eq "(none)");
  }

  # Get the list of all combos the user chose, as a lookup table.
  # The keys are of the form "1.1", "1.2", "2.1", etc...
  my %combo_ids_chosen = map {$_=>1} @all_combo_ids_chosen;
  
  # Out of these, get the combos the user selected, annotated with 
  # any options (subset, order, etc.) they provided:
  my @chosen_combos;
  for $pkno (1..int(@packages)) {       # for each package, 1..p
    my $package = $packages[$pkno-1];     # get 0-based package
    my @combos = $package->combos;        # get combos in packages
    my $combono;
    
    # Was any combo from this package chosen?  If not, skip it now:
    (grep { /^\d+/ && ($& eq $pkno) } @all_combo_ids_chosen) or next;
    
    # If there were subsetting/order options, and they are wanted,
    # get 'em:
    my $opts_type;
    my %opts_assigned;
    my $criteria;
    foreach $opts_type (sort $package->options) {
      my $opts = $package->options($opts_type);
      # Check to see if this item is a floating scene granule
      if (($opts_type eq 'SUBSET') && defined($opts)) {
         foreach $criteria ($opts->criteria) {
            $floatscene += 1 if ($criteria->type eq 'FLOATINGSCENE');
         }
      }

      # Are they required?
      my $opts_required = $opts->is_required;
      
      # Were they given?  Fill them in, then check for no 
      # selections:
      $opts->read_CGI(Prefix=>"pk-opts.$pkno.$opts_type.",Page=>$pageNumber,
                      Min=>$min);
      my $opts_given = $opts->user_made_selections;
      
      # Zero it if necessary:
      $opts_assigned{$opts_type} = $opts     if $opts_given;
    }          
    # ...at this point, %opts_assigned contains only those options
    # which the user gave input for, hashed by option type. 
    
    # For each combo in this package...
    for $combono (1..int(@combos)) {
      
      # Did the user select this combo?  If not, skip it:
      $combo_ids_chosen{"$pkno.$combono"} or next;
      
      # They wanted this combo (and hence, the package it's in);
      #    Associate any subset/order options with this combo,
      #    and push it on the list of combos:
      my $combo = $combos[$combono-1];      # get 0-based combo
      my ($opts_type, $opts);
      my $cost;
      my $name;
      while (($opts_type, $opts) = each %opts_assigned) {
        $combo->options($opts_type, $opts);
        # sum up the cost of the options.
        $cost += $opts->cost if $opts->cost;
        # Set the option name if it is present
        $name = $opts->name if $opts->name;
      }
      # Add orignal cost of the combo to the cost of the options and
      # store the total as the new cost of the combo.
      
      $cost += $combo->cost if $combo->cost;
      $combo->cost($cost) if $cost; 
      # Set the combo name if there is one
      $combo->name($name) if $name; 
      
      # get the number of copies the user wants and
      # add that to the combo
      
      if ( $REQ->param("copies.$pkno.$combono") ) {
        my $copies = $REQ->param("copies.$pkno.$combono");
        
        ($copies =~ /^\d+$/) or puke 511,
        "copies '$copies' must be an integer 1 to 99";
        
        ($copies < 100 and $copies > 0) or puke 511,
        "copies '$copies' must be an integer 1 to 99";
        
        $combo->copies($copies);
      }
      # Add any additional info entered for the package to the combo.
      my $addn_info = $REQ->param("additional_info.$pkno");
      puke 513, " Additional information is limited to 80 characters. "
        if (length($addn_info) > 80);
      $combo->addn_info($addn_info);
      
      # Set the package number in the combo (for printing).
      $combo->pkg_no($pkno);
      
      # Add the combo to the list of combos they want.
      push @chosen_combos, $combo;
    }
  }
  #-----------------------
  # MODIFY THE ITEMS...
  
  # Modify all items who need modifying:
  my $item;
  my $msg;

  # --------------------------------------------------
  # Added reference MODIFY to the for loop, this loop will
  # be exited prematuredly if any media limit is reached.
  # --------------------------------------------------
  MODIFY: foreach $item (@modify_items) {
    my $package;
    my $combo;
    
    # For each combo chosen...
    my @item_combos;                 # combos FOR THIS ITEM
    my $skip_flag = 0;
    foreach $combo (@chosen_combos) {
      
      # Skip the selected combo if item doesn't have this package!
      # [Fix for bug #3720]
      my $has_package = 
        grep {$_ eq $combo->package_id} @{$item->package_ids};
      if (!$has_package) {
        inform "item ", $item->key, " does not have package ", 
        $combo->package_id, " and so was skipped"; 
        $skip_flag = 1;
        next;
      }
      my $small_media;
      # Check to make sure the granule has a media type that is large enough
      # If the item is a floating scene granule the size check is skipped.
      if ($item->isa("IU::Granule") && (!$floatscene)) {
         my $granule_size = $item->size_mbecs_granule;
         my $dskey = $item->dataset_key;
         ($dset) = $CART->datasets([$dskey]);

         # ---------------------
         # Check for media type limits before setting any options,
         # if limits are encountered exit the for loop to continue.
         # ---------------------
         $limits = $MEDIAORDERLIMIT->add($dset, $item, $combo->media_type);
         last MODIFY if ($limits);

         # ---------------------
         # Before we assigned the media type selected,
         # check if there has been multiple medias selected
         # so we can inform the user when displaying the cart.
         # ---------------------
         if (int(@valid_combo_ids_chosen) > 1 ) {
            $item->selected_media_type ("Multiple Selection");
         } else {
            $item->selected_media_type ($combo->media_type);
         }
         # Initialize the item to have the name unset
         $item->name('');
         # Set the item name if the combo has a name
         $item->name($combo->name) if($combo->name);
         # Set the hash of %SmallMedias for this granule
         my $mediamaximumcapacity;

         foreach $mediamaximumcapacity (@{$dset->mediamaximumcapacity}) {
            # Make sure media maximum capacity is greater than zero,
            # if is not don't bother adding it to the small media hash.
            my $test_size = ($mediamaximumcapacity->capacity > 0) ?
                                        $mediamaximumcapacity->capacity : '';
            next if ($test_size == '');
            if ($granule_size > $mediamaximumcapacity->capacity) {
               $SmallMedias{$mediamaximumcapacity->type} =
                $mediamaximumcapacity->capacity;
            }
         }
         foreach $small_media (keys %SmallMedias) {
            # If the current selected media_type is too small for this
            # granule, this granule must be skipped.
            if($combo->media_type eq $small_media) {
               my $id = $item->id;
               $msg .= "Granule: $id does not fit media type $small_media".
                " and so was skipped. "; 
               $APP->message (new IW::Note Class=>'WARN', Text => $msg);
               $skip_flag = 1;
               next;
            }
         }

      
      } elsif ($item->isa("IU::Granule") && ($floatscene)) {
         # Set the selected media type if it is a floating scene granule
         $item->selected_media_type ($combo->media_type);
      }
      # Make a duplicate combo FOR THIS ITEM, since we'll be tying 
      # to different items...
      #
      #     ***NOTE*** The shallow copy should ultiamtely be a deep
      #     copy.  Once the new Storable module is in place, you
      #     may be able to easily deep-copy an object $o by saying 
      #     $o2 = thaw(freeze($o));
      #
      my $item_combo = $combo->copy; ### WARNING: SHALLOW COPY!!!!!
      $item_combo->granule_key($item->granule_key);
      push @item_combos, $item_combo;
    }
    # If the granule item does not have combos that 
    # match all of the chosen combos, this granule will be skipped
    # for modifying its package selections.
    # If the "set-for" flag changes for the data set, there may be a dataset 
    # object in the item list which does not have any combos associated 
    # with it.  In this case the dataset item should still be added when 
    # we get to the "add_granules" method.
    next if($skip_flag and $item->order_type eq 'GRANULE' and 
            $item->isa("IU::Granule"));
    
    # REPLACE any item package info with new info.
    #  We replace, since that allows user to CLEAR selected packages.
    $item->order_pkgcombos(\@item_combos);    
    
    # Re-store the item(s).
    # Note that all we are doing is updating items which are already
    # in the cart, so we can just invoke the basic add_xxx() method:
    my $overwrite = '1'; # force granule to be overwritten.
    if ($item->order_type eq 'GRANULE') {
      $CART->add_granules([$item],$overwrite);
    }
    elsif ($item->order_type eq 'DATASET') {
      $CART->add_datasets([$item]);
    }
    else {
      die "bad item type: $item!";
    }
  }
  #-----------------------
  # UPDATE ACCOUNT INFO...
  update_acct_info(%params);
  
  # update cart media limit info in the cart.
  $CART->cart_medialimit_info ($MEDIAORDERLIMIT);

  # Return any limits founded.
  return ($limits);
}

#------------------------------------------------------------
# update_acct_info PARAMHASH
#------------------------------------------------------------
# We just came from the package-selection screen.
# We have also potentially selected an account for a single dataset.
# Get and modify that dataset:
# Params are:
#
#      ChosenItem  => the item we chose packages for

sub update_acct_info {
  my %params = @_;
  my $chosen_item  = $params{ChosenItem};
  my $item_type = $params{ItemType};
  
  # Get the dataset associated with the chosen item:
  my ($chosen_dset) = $CART->datasets([$chosen_item->dataset_key]);
  
  # Was account info given? 
  my $acctname;
  if ($acctname = $REQ->param('account')) {
    
    # Choose the account:
    $chosen_dset->chosen_account($acctname);
    
    # Re-store the dataset object.  Note that it must
    #   -- by definition -- already be in the cart, so
    #   it's sufficient just to update the dataset object itself:
    $CART->add_datasets([$chosen_dset]);        
  }
  # Done!
  1;
}

#------------------------------
#
# SUBMODE HANDLERS 
#
#------------------------------

#------------------------------------------------------------
# submode_DEFAULT
#------------------------------------------------------------
sub submode_DEFAULT {
  shift->submode_LIST;
}

#------------------------------------------------------------
# submode_LIST
#------------------------------------------------------------
# I<Top level:> simply list all granules in the cart
# (there might be none!)

sub submode_LIST {
  # Send the results:
  send_doc() or puke "couldn't send listing!";
  1;
}

#------------------------------------------------------------
# submode_DSETADD
#------------------------------------------------------------
# I<Top level:> add the selected datasets to the shopping cart.
#
# Add dskeys to cart if they are orderable.

sub submode_DSETADD {
  my %changes;
  
  #-----------------
  # GET SELECTIONS...
  #-----------------
  
  # Add from the inventory store:
  $INVSTORE or
    puke "Ack! Couldn't open up results! Were they deleted?";
  
  my @dskeys = ();
  my $opts;
  
  @dskeys = $REQ->param('dsaddkey');
  if (@dskeys) {
    $changes{WhatSelected} = 'just data set';
    # Got it, now get count the data sets actually select'ed.
    my @selkeys = $REQ->param('dskey');
    if (@selkeys) {
      $changes{NumSelected} = int (@selkeys);
    } else {
      $changes{NumSelected} = 1;
    }
  }
  
  #-----------------
  # ADD DATASETS TO CART...
  #-----------------
  
  # Get *selected* datasets from *results*:
  my @selected_dsets = $INVSTORE->datasets(\@dskeys);
  my $i = 0;
  foreach (@selected_dsets) {
     # Get all the packages the user might have selected from:
     my @packages = IU::Package->lookup_dataset($dskeys[$i],
                                                Dataset => $selected_dsets[$i],
                                                SubDir => 'search');
     @packages or puke 560,
      "Data set claims to have ordering options, but none were supplied.";
     $opts = $packages[0]->options('DATASET_ORDER');
     if (defined($opts)) {
        $selected_dsets[$i]->some_options('DATASET_ORDER', $opts);
     }
     $opts = $packages[0]->options('SUBSET');
     if (defined($opts)) {
        $selected_dsets[$i]->some_options('SUBSET', $opts);
     }
     $i++;
  } 
  # Filter out datasets which cannot be ordered for some reason:
  my @addable_dsets = grep { $_->orderable } @selected_dsets;
  
  $CART->add_just_datasets(\@addable_dsets);
  
  # Add datasets to shopping cart, and pick up changes:
  # Always 0 for DS.
  $changes{InsideCart}  = $CART->stats->{AlreadyThere};
  $changes{OutsideCart} = (int(@addable_dsets) -
                           $changes{InsideCart});
  $changes{Rejected}    = (int(@selected_dsets) -
                           int(@addable_dsets));
  $changes{Change}      = 'ADD';
  
  #-----------------
  # DISPLAY CART...
  #-----------------
  # Send the results:
  send_doc(Changes => \%changes)
    or puke "couldn't send listing!";
  1;
}

#------------------------------------------------------------
# submode_ADD
#------------------------------------------------------------

sub submode_ADD {
  my %changes;
  
  #-----------------
  # GET SELECTIONS...
  #-----------------
  
  my @gkeys = IW::CartList->add_virtual_ckeys($INVSTORE,$INVSTORELIST,$CART);
  
  #-----------------
  # OPEN wwwvalids.conf
  #-----------------
  
  # Open static valids database:
  my %WWWVALIDS;
  if (-e "${IU::LibDir}/conf/wwwvalids.gdb") {
    eval {tie %WWWVALIDS, 'IU::Store', "${IU::LibDir}/conf/wwwvalids.gdb",
            'READER'};
    $@ and return error "tie failed: $@";
  }
  
  #-----------------
  # CHECK FOR DISCLAIMER
  #             The lawyers have been here!
  #-----------------
  
  # As of 3.0, a Quality Flag disclaimer will be shown AFTER the Dataset/
  # Source disclaimer page.
  
  my $display_QA_disclaimer = 'no';
  $SESSION->state->{'DisplayHoldMessage'} =  1;
 
  # As of 2.3, we're now required to check whether there is text that the
  # MUST read and ACCEPT prior to ordering the data (DISCLAIMER)
  #
  # The Disclaimer comes in two flavors :
  # Source: The EDG compares the 'selected' granules to see if their SOURCE
  # has been tagged for a Disclaimer. If it is, then the Disclaimer text
  # is extracted from wwwvalids.conf.
  #
  # Dataset: The DAAC sends back a new PARAMETER called DISCLAIMER_COMMENT.
  # Inside this COMMENT is the text that the user needs to READ/ACCEPT.
  # If the text is to LARGE then the DAAC can send back a DISCLAIMER_URL
  # tag which is a FULL URL pointing to the Disclaimer Text on another Server.
  
  
  # Hash that will contain the source=>disclaimer message.
  my %list_disclaimers ;
  
  # Get *selected* granules from *results*:
  my @selected_grans = $INVSTORE->granules(\@gkeys);
  
  # Filter out granules which cannot be ordered for some reason:
  my @addable_grans = grep { $_->possible_to_order } @selected_grans;
  
  # Get a single granule.
  my $single_gran;
  
  # Cycle through the selected granule's Result Group,source and Sensor type to
  # determine whether there should be a disclaimer displayed. If so, get
  # the text associated with the disclaimer.
  
  foreach $single_gran ( @addable_grans) {
    
    # Does this granule have a Specialized Results at all ???
    # If the QA Failed then set a flag in the Session state 
    # So that a QA disclaimer will appear.
    
    if ($single_gran->resultGroup) {
      if ( $single_gran->quality_assurance_status )
        { debug "Granules PASSED"}
      else {
        debug " Granule FAILED";
        my $is_science_hold = IW::Disclaimers->display_hold_text($single_gran->science_quality_flag);

        if ( $is_science_hold) {
           $SESSION->state->{'DisplayHoldMessage'}  = 'yes';
	}

        $SESSION->state->{'DisclaimerQAFlag'} =  "displayDisclaimer"; 
      }
    }
    
    # Find/Get the Granule's Dataset Parent
    my $gran_dataset_id =  $single_gran->dataset_key;
    
    # May either use the RESULTS/FOLDER database.
    my $selected_dataset = $INVSTORE->dataset( $gran_dataset_id );
    
    # Get the array OF arrays of sources and sensors.
    my  @list_gran_sensors = $single_gran->sensors;
    my  @list_gran_sources = $single_gran->sources;
    
    # We only want one of the arrays.
    # Probably should use swift here Hm...
    my $single_gran_sensor_set =  pop @list_gran_sensors  
      if (defined(int(@list_gran_sensors)>0)) ;
    my $single_gran_source_set = pop @list_gran_sources 
      if (defined(int(@list_gran_sources)>0)) ;
    
    # Entry from the wwwvalids database.
    my $valids;
    # Holds the Dataset Disclaimers URL/Disclaimer Text.
    my %sub_disclaimer_hash;
    # If we find a sensor disclaimer, we don't want to show a Source
    # Disclaimer.
    my $found_sensor_disclaimer = 0;
    $SESSION->state->{'found_sensor_disclaimer'} = '0' if (!defined( $SESSION->state->{'found_sensor_disclaimer'}));;   
 
    # Check the Dataset, to see if there is a DATASET Disclaimer.
    # If so, extract the _COMMENT/_URL if present.
    if ($selected_dataset->disclaimer_comment) {
      if(!defined($SESSION->state->{'Disclaimer'.$selected_dataset->id} )) {
        $sub_disclaimer_hash{Text} = $selected_dataset->disclaimer_comment;
        $sub_disclaimer_hash{URL} = $selected_dataset->disclaimer_url;
        $list_disclaimers{$selected_dataset->id} = \%sub_disclaimer_hash;
      } 
    }
    else {
      my ($single_gran_sensor, $single_gran_source);
      
      # Strike One, no Dataset Disclaimer
      # Then check for a  Sensor disclaimer from all the granule's
      # available sensors.
      if ($single_gran_sensor_set){
        foreach  $single_gran_sensor (@$single_gran_sensor_set) {
          # Ok, first see if we've already shown this disclaimer,
          # the sensor name would have been stored in the Session State.
          # No? also make sure there is a disclaimer to show.
          if (!($SESSION->state->{'Disclaimer'.$single_gran_sensor}) && 
              ($valids = $WWWVALIDS{"disclaimer:$single_gran_sensor"})) {
            # Add to the list of disclaimers that should be displayed..
            $sub_disclaimer_hash{Text} = $valids->{Text};
            $list_disclaimers{$single_gran_sensor} = \%sub_disclaimer_hash;
            # $found_sensor_disclaimer = 1;
             $SESSION->state->{'found_sensor_disclaimer'} = '1';
          }
        } 
      }
      # Strike Two, No Sensor, Source?
      if ($single_gran_source_set && ( $SESSION->state->{'found_sensor_disclaimer'} eq '0' )) { 
        my $sensor_name = ${@$single_gran_sensor_set}[0] if ($single_gran_sensor_set); 
      
      foreach  $single_gran_source (@$single_gran_source_set) {
        my $valids;   
        # Same comments as Sensor PLUS check to see if the granule's
        # sensor is already in the Session State.
        # Why? Becuase even though we have a Source disclaimer that
        # should be shown, there might have been a Sensor disclaimer
        # That has ALREADY been shown.
        # Example : Source Terra has a disclaimer, but if the
        # Sensor was MODIS, then we show the Sensor and not the Source
        # Disclaimer.
        if (!($SESSION->state->{'Disclaimer'.$single_gran_source}) && 
            ($valids = $WWWVALIDS{"disclaimer:$single_gran_source"}) &&
            ((!($SESSION->state->{'Disclaimer'.$sensor_name}) ))) {
          # Add to the list of disclaimers that should be displayed..
          $sub_disclaimer_hash{Text} = $valids->{Text};
          $list_disclaimers{$single_gran_source} = \%sub_disclaimer_hash;
        }
      }
    }
  }
}

# Close static valids database:
untie %WWWVALIDS;

if (%list_disclaimers) {
  print_disclaimer(%list_disclaimers) ;
}
else {
  # They've already seen A Disclaimer, so just add the datasets/granules.
  submode_SHOW_QA_FLAGS();
}

1;
}

sub submode_SHOW_QA_FLAGS {
  
  # What SOURCE/DATASET disclaimer did we display?
  my @disclaimed_source = $REQ->param('disclaimed_source');
  
  # just one entry.
   my $single_source;
  
  # Add the source(s) to the list of disclaimers already shown.
  foreach $single_source ( @disclaimed_source ) {
    $SESSION->state->{'Disclaimer'.$single_source} =  1;
#    debug "DisSources : $single_source";
  }
  

  my $flag = $SESSION->state->{'DisclaimerQAFlag'};
  my $hold_status =  $SESSION->state->{'DisplayHoldMessage'} ;
#  debug "Hold status = $hold_status";
  if (defined($flag) && ($flag eq "displayDisclaimer")) {
    print_QAFlag_disclaimer($hold_status) ;
  }
  else {
    # They've already seen A Disclaimer, so just add the datasets/granules.
    submode_CONT_TO_ADD();
  }
  
  1;
}


#------------------------------------------------------------
# submode_CONT_TO_ADD
#------------------------------------------------------------
# The user may/may not have seen a disclaimer.
# Previously was submode_ADD
# I<Top level:> add the selected granules to the shopping cart.
# Granules may be specified either by GKEYs, or (if none) by DSKEYs:
#
#   * If one or more 'gkey's are given, add those granules.
#   * Else if one or more 'dskey's are given, and 'gkey!' is not
#     insisted upon, add all granules in those datasets.
#   * Else, fail.
sub submode_CONT_TO_ADD {
  my %changes;
  
  $SESSION->state->{'DisclaimerQAFlag'} =  1;
 
 
  #-----------------
  # ADD GRANULES TO CART...
  #-----------------
  
  my $chosen_gkey;
  my $chosen_gran;
  my $chosen_dskey;
  my $chosen_dset;
  my $dset_object;
  my @combos;
  my $item;
  my $combo;
  my $pkno;
  my $package;
  my $item_combo;
#  my @item_combos;
  my $newsub; # sub of $user that handles FTP Push.
  my $small_media; # hash key for %SmallMedias
  my $user = $SESSION->state->{User} || new IW::User;
  # Get granules to add:
  my @gkeys = IW::CartList->cont_to_add_virtual_ckeys($CART); 
  $changes{NumSelected} = (int(@gkeys)); 
  my @selected_grans = $INVSTORE->granules(\@gkeys);

  
  # Filter out granules which cannot be ordered for some reason:
  my @addable_grans = grep { $_->possible_to_order } @selected_grans;
  my @packages;
  my $previous_dset;
  my $criteria;
  my $opts;
  my $id_ref;
  my @id_array;
  my @previous_id_array;
  my $i;
  foreach $chosen_gran (@addable_grans) {
     my @newcriteria; # New version of the criteria array
     my $match = 1;
     $chosen_gran->some_options('SUBSET', 0);
     # Flag for looping
     my $userflag = 0;

     # Flag for FTP push
     my $pushflag = 0;

     $id_ref = $chosen_gran->package_ids; 
     @id_array = sort @$id_ref;

     # The following checks for match between the current granule package_ids
     # and the last granule's package_ids.
     $match = 0 if(int(@id_array) != int(@previous_id_array));
     if(int(@id_array) == int(@previous_id_array)) {
        for($i = 0, $i < int(@id_array), $i++) {
           if($previous_id_array[$i] && $id_array[$i] && 
              ($previous_id_array[$i] ne $id_array[$i])) {
              $match = 0;
              last;
           }
        }
     }

     $chosen_gkey = $chosen_gran->key;
     $chosen_dskey = $chosen_gran->dataset_key;
     $chosen_dset = $chosen_gran->dataset_id;
     $dset_object = $chosen_gran->dataset;
     # Get all the packages the user might have selected from:
     # if it is a new data set or if the granule package ids are different
     # from the previous granule
     if((!$previous_dset) || ($chosen_dset ne $previous_dset) || !$match) {
        @packages = IU::Package->lookup($chosen_gkey,
                                        Dataset => $chosen_dset,
                                        Granule => $chosen_gran,
                                        SubDir => 'search');
	$previous_dset = $chosen_dset;
        @previous_id_array = @id_array;
     }
     set_small_media($dset_object, $chosen_gran, \@packages);
     my $floatscene = 0;  # are there any floating scene criteria.
     foreach $package (@packages) {
        $opts = $package->options('SUBSET');
        if (defined($opts)) {
           $chosen_gran->some_options('SUBSET', $opts);
           foreach $criteria ($opts->criteria) {
              $floatscene += 1 if ($criteria->type eq 'FLOATINGSCENE');
           }
        }
        unless($floatscene) {
           $opts = $package->options('ORDER');
           if (defined($opts)) {
              foreach $criteria ($opts->criteria) {
                 last if($pushflag > 4);
                 foreach $sub_key (@user_subs_keys) {
                    # When we get 5 FTP push strings we are finished
                    last if($pushflag > 4);
                    if ($criteria->name eq $sub_key) {
                       my $newsub = $user_subs{$sub_key};
                       if($user->$newsub()) {
                          $criteria->{UserSlct} = {VALUE=>$user->$newsub()};
                          push @newcriteria, $criteria;
                          $pushflag++;
                       } elsif ($sub_key eq 'USERSTRING') {
                          # Because 'USERSTRING' is not required count it
                          # as the fifth string anyway even if it is not 
                          # used.
                          $pushflag++;
                       }
                    }
                 }
              }
              $opts->{Criteria} = \@newcriteria;
           }
           @combos = $package->combos;
           $pkno = 0;
           COMBOTEST: foreach $combo (@combos) {
              $pkno++;
              $opts = $package->options('ORDER');
              $chosen_gran->some_options('ORDER', $opts) if(defined($opts) &&
               ($combo->media_type ne 'FtpPush'));
              $chosen_gran->selected_package_size ($combo->pkg_size);
  
              # Check for small media, if the media is too small for
              # the granule, skip the combo so it will not be displayed
              # in the selection page for this granule.
              foreach $small_media (keys %SmallMedias) {
                 next COMBOTEST if $combo->media_type eq $small_media;
              }
              if(($user->data_format) && ($user->media_type) && 
                 ($user->media_format) && 
                 ($user->data_format eq $combo->proc_opt) &&
                 ($user->media_type eq $combo->media_type) &&
                 ($user->media_format eq $combo->media_fmt)) {
                 last if($userflag);
                 $opts = $package->options('ORDER');
                 $chosen_gran->options('ORDER', $opts) if(defined($opts));
                 $combo->options('ORDER', $opts) if(defined($opts));
                 $opts = $package->options('SUBSET');
                 $chosen_gran->options('SUBSET', $opts) if(defined($opts));
                 $combo->options('SUBSET', $opts) if(defined($opts));
                 $combo->pkg_no($pkno);
                 my $item_combo = $combo->copy; ### WARNING: SHALLOW COPY!!!!!
                 $item_combo->granule_key($chosen_gran->granule_key);
                 my @item_combos;
                 push @item_combos, $item_combo;
                 $chosen_gran->selected_media_type($combo->media_type);
                 $chosen_gran->order_pkgcombos(\@item_combos);
                 $userflag = 1;
              }
           }
        }
     }
     # Add granules to shopping cart, and pick up changes:
     $CART->add_granules([$chosen_gran]);
  }
  $changes{InsideCart}  = $CART->stats->{AlreadyThere};
  $changes{OutsideCart} = (int(@addable_grans) -
                           $changes{InsideCart});
  $changes{Rejected}    = int(@selected_grans) - int(@addable_grans);
  $changes{Change}      = 'ADD';
  $changes{WhatSelected} = 'granules';
  
  #-----------------
  # ADD DATASETS TO CART...
  #-----------------
  
  # Don't add dataset to the cart if it is already there as an
  # orderable (just_datasets) dataset.
  
  my @addable_dskeys;
  my %just_datasets = map {(($_->key)=>1)} $CART->just_datasets;
  my $gran;
  
  foreach $gran (@addable_grans) {
    my $granule_dskey = $gran->dataset_key;
    if (!$just_datasets{$granule_dskey}) {
      
      push @addable_dskeys, $granule_dskey;
    }
  }
  
  # Get these datasets from *results*:
  my @addable_dsets = $INVSTORE->datasets(\@addable_dskeys);
  
  # Add datasets to shopping cart (it's okay if we add more than
  # we'll use):
  $CART->add_datasets(\@addable_dsets);
  
  
  #-----------------
  # DISPLAY CART...
  #-----------------
  # Send the results:
  send_doc(Changes => \%changes)
    or puke "couldn't send listing!";
  1;
}

#------------------------------------------------------------
# submode_MANIFEST
#------------------------------------------------------------
# I<Top level:> Peek at the printed summary

sub submode_MANIFEST {
  
  my $self = shift;
  my %params = @_;
  my $listmode = $params{ListMode};
  my $changes  = $params{Changes} || {};
  my $gran;
  my $ordernow = $REQ->param('ordernow');
  
  # Get all granules and datasets in the shopping cart:
  my @granules = $CART->granules;
  my @datasets = $CART->datasets;
  my @just_datasets = $CART->just_datasets;
  
  # Get list of unique daacs in cart.  (foreach daac, create key)
  my %merge;
  map { $merge{($_->archive_id)}++ } (@datasets, @just_datasets);
  # Keys of %merge are daacs in cart. Check if any need V0 auth.
  my @daacs = keys (%merge);
  my @daacsV0 = IU::Auth::daacs_needing_v0_auth (\@daacs);
  
  # If yes, pass true Required flag to from_CGI method.
  my $auth_required = (@daacsV0 ? 1 : 0);
  
  # Check to see if any DAACs need NASDA authenticator
  my @daacsNASDA = IU::Auth::daacs_needing_nasda_auth (\@daacs);

  # If yes, pass true Required flag to from_CGI method.
  my $nasda_required = (@daacsNASDA ? 1 : 0);
  
  # Get input values from the form.  Updates User object in session.
  #   User info (keys, addresses, affil).  All required except keys.
  my $user = IW::User->from_CGI ('FORM'         => 'ORDER',
                                 'REQUIRED'     => 1,
                                 'AUTHREQUIRED' => $auth_required,
                                 'NASDAREQUIRED' => $nasda_required);
  
  # If any required fields are missing, show the order form again,
  # instead of continuing.
  if ($user->errors->any_errors) {
    # Set error messages into top of page note.
    $APP->message (new IW::Note Class=>'ERROR',
                   HTML =>join ("<BR>\n", $user->errors->get));
    # Switch back to Mode ORDER.
    switch IW::Mode Tags=>{'mode'    => 'ORDER',
                           'u'       => time,
                           'sid'     => $SESSION->id,
                           'bogus'   =>'Fixes a CGI::BasePlus bug',
                          };
    return (1);
  }
  
  # Create order object:
  my ($req_order, $email_order) =
    IW::Order->make(Granules=>\@granules,
                    Datasets=>\@datasets,
                    Just_Datasets=>\@just_datasets,
                    SkipIncomplete=>1);
  
  # Set the indicator. We can do this any time, but this is as good 
  # a time as any other.
  $SESSION->{Navigator}->{Indicator} = "OrderData";
  
  if(!$ordernow) {
     # Print page header:
     $APP->start_page(SuperTitle=>'Shopping Cart',
                      Title=>"Step 3: Order Summary",
                      Context =>'ordersumm');
  
     # The directions:
     #   $HTML -> directions;
  
     # What can they do?
     if ($req_order->need_packages) {   # need packages:
       $HTML -> B
         -> t("You must ")
           -> A(HREF=>script_url('mode'=>'CART'))
             -> t("go back to the shopping cart and choose ordering ".
                  "options")
               ->_A
                 -> t(" for ALL of the items before you can place an ".
                      "order.")
                   -> _B;
       #      $HTML -> _directions;
     }
     else {                             # we can do it!
    
       #      $HTML -> B 
       #            -> t("If you would like you can now go to Step 4.")
       #            -> _B;
       #      $HTML -> _directions;
       $HTML -> script_FORM;
       # Shopping cart:
       $HTML -> P;
       $HTML -> select_one (Name=>'empty-cart',
                            Choices=>['Empty shopping cart if no errors',
                                      'DO NOT empty shopping cart']);
       print "after order has been submitted.<BR>\n";
    
       $HTML -> msubmit_flag;
       $HTML -> CENTER -> FONT(SIZE=>'+3') -> B
         -> msubmit_button(CAPTION =>
                           '  Submit Order!  ',
                           QUERY =>['mode'    => 'ORDER',
                                    'submode' => 'SUBMIT']);
       $HTML -> _CENTER -> _FONT -> _B;
     }

     $HTML -> BR;

     $HTML -> B  
           -> msubmit_button(CAPTION =>
                        'Change Address Entries', 
                        QUERY =>['mode'    => 'ORDER',
                                 'submode' => 'NEWFORM']);
     $HTML -> _B; 
     $HTML -> _script_FORM;

  } # end if !$ordernow
  #commented this out because now we want to create a file that stays
  #in the user's session dir 
  # The summary:
  #   my $TMP = FileHandle->new_tmpfile;
  
  # Get unique request id (session id + "-P" + session prod. req. no):
  #Moved this code from ModeOrder to ModeCART submode MANIFEST because reqid
  # needed to create manifest file for email
  
  my $reqno = 1 + ($SESSION->state->{OrderReqNo} || 0);
  my $reqid = $SESSION->id . "-P$reqno";
  $SESSION->state->{OrderReqNo} = $reqno;
  $SESSION->state->{OrderReqId} = $reqid;
  
  # Create a file to hold the product summary in the users session dir.
  # This file will be appended to the confirmation email.
  # We want this to be hidden from casual prying by mail harvesters
  # and the like, so we create a subdirectory that allows no one to peek in.

  my $manifest_dir = $SESSION->path_to("pub/manifest");
  unless (-e $manifest_dir) {
    # Create directory, or die.
    mkdir $manifest_dir,0775 or puke "Can't create manifest directory: $!";
    # Fix it so no one can see in.
    my $access = new IU::HTAccess;
    $access->start_limit('GET', 'POST');
    $access->order('DENY', 'ALLOW');
    $access->deny_from("all");
    $access->end_limit;
    $access->error_doc(403,"${IU::DocsUrl}/imswelcome/denied.html");
    $access->write_in($manifest_dir);
  }
  my $filename = "Manifest.$reqid";
  my $fullname = "$manifest_dir/$filename";
  open (MANI, "+>$fullname") or puke 452, "couldn't open $fullname";
  
  # Print Order receipt info.  Note: {User} need not be defined.
  IW::Order->print_receipt(\*MANI,
                           User      => $SESSION->state->{User},
                           Requested => $req_order,
                           Emailed   => $email_order);
  
  if($ordernow) {
     switch IW::Mode Tags=>{'mode'    => 'ORDER',
                            'submode' => 'SUBMIT',
                            'u'       => time,
                            'sid'     => $SESSION->id};
  } else {
     $HTML -> P -> HR -> PRE;
     MANI->seek(0,0); output $HTML $_ while <MANI>;
     $HTML -> _PRE;
  
     # The signature:
     $APP->end_page;
  }
  1;
}

#------------------------------------------------------------
# submode_PEEKODL
#------------------------------------------------------------
# I<Top level:> Peek at the ODL for the cart 

sub submode_PEEKODL {
  my %params = @_;
  my $listmode = $params{ListMode};
  my $changes  = $params{Changes} || {};
  my $gran;
  
  # Get all granules and datasets in the shopping cart:
  my @granules = $CART->granules;
  my @datasets = $CART->datasets;
  my @just_datasets = $CART->just_datasets;
  
  # Create order object:
  my ($req_order, $email_order) = 
    IW::Order->make(Granules=>\@granules,
                    Datasets=>\@datasets,
                    Just_Datasets=>\@just_datasets,
                    SkipIncomplete=>1);
  
  # Set the indicator. We can do this any time, but this is as good a time
  # as any other.
  $SESSION->{Navigator}->{Indicator} = "OrderData";
  
  # Print page header:
  $APP->start_page(SuperTitle=>'Shopping Cart',
                   Title=>"ODL order messages");
  
  # Output the PRODUCT_REQUEST order that goes to the DAACs:
  if ($req_order) {
    
    # Add some info to keep it happy:
    $req_order->id("PENDING");
    
    # Convert to ODL, and output:
    my $odl = $req_order->to_request_ODL(NoValidate=>1);
    $HTML -> P -> HR -> H2 -> t("IMS product request") -> _H2;
    if ($odl) {
      my $TMP = FileHandle->new_tmpfile;
      $HTML -> PRE;
      $odl->output($TMP);
      $TMP->seek(0,0); output $HTML $_ while <$TMP>; close $TMP;
      $HTML -> _PRE;
    }
    else {
      $HTML -> t_EM("Couldn't convert order to ODL");
    }
  }
  
  # Output the PRODUCT_REQUEST order that goes to the DAACs:
  if ($email_order) {
    
    # Add some info to keep it happy:
    $email_order->id("PENDING");
    $HTML -> P -> HR -> H2 -> t("E-mail order") -> _H2;
    
    # Create ODL to mail to a particular dataset's contact:
    my $dsid;
    foreach $dsid ($email_order->dataset_ids) {
      my $odl = $email_order->to_email_ODL($dsid, NoValidate=>1);
      if ($odl) {
        my $TMP = FileHandle->new_tmpfile;
        $HTML -> P -> H3 -> t($dsid) -> _H3;
        $HTML -> PRE;
        $odl->output($TMP);
        $TMP->seek(0,0); output $HTML $_ while <$TMP>; close($TMP);
        $HTML -> _PRE;
      }
      else {
        $HTML -> t_EM("Couldn't convert order to ODL");
      }
    }
  }
  # The signature:
  $APP->end_page;
  1;
}

#------------------------------------------------------------
# submode_REM
#------------------------------------------------------------
# I<Top level:> remove the selected items from the shopping cart.
#

sub submode_REM {
  my %changes;
  
  my $remove = $REQ->param('remove');
  
  my @dskeys;
  my @gkeys;
  my @itemkeys; # To remove size media options
  
  # Get media order limit stored in cart, if any.
  $MEDIAORDERLIMIT = $CART->cart_medialimit_info();

  if ($remove eq 'ALL') {
    
    @gkeys = $CART->remove_granules();
    @dskeys = $CART->remove_just_datasets();

    # Clear any information existing in the media
    # order limit object, only if it exists.
    $MEDIAORDERLIMIT->clear() if (defined($MEDIAORDERLIMIT));

    # Remove any order media object from the cart.
    $CART->remove_cart_medialimit_info();

    # Ok, since all the granules are going, clean
    # out any disclaimers that were associated with them.
    my @list_disclaimers =  keys %{$SESSION->state};
    my ($search_disclaimer,@dis_list ) ;
    foreach $search_disclaimer (@list_disclaimers) {
       if ( $search_disclaimer =~ /Disclaimer/ )
          { push @dis_list , $search_disclaimer; }
    }

    foreach $search_disclaimer (@dis_list) {
      delete ( $SESSION->state->{$search_disclaimer});
    }
    delete ( $SESSION->state->{'found_sensor_disclaimer'}) 
         if ($SESSION->state->{'found_sensor_disclaimer'});
  } 
  else {
    @dskeys = $REQ->param('dskey');
    @gkeys  = $REQ->param('gkey');

    # ---------------------------------------------------------------
    # Before we remove the items from the cart, we need to remove the
    # media information out of the media order limit if any exists.
    # ---------------------------------------------------------------
    if (defined($MEDIAORDERLIMIT)) {
       # Get the granules and datasets from the cart using the keys.
       my @grans = $CART->granules(\@gkeys) if (@gkeys);
       my @dsets = $CART->datasets(\@dskeys) if (@dskeys);
       my @items = (@grans, @dsets);

       # Remove them from the media order limit and update the cart.
       $MEDIAORDERLIMIT->remove(\@items) if (@items);

       # Update the media order limit information inside the cart.
       $CART->cart_medialimit_info ($MEDIAORDERLIMIT);
    }

    # Remove them from the cart.
    $CART->remove_granules(\@gkeys) if (@gkeys);
    $CART->remove_just_datasets(\@dskeys) if (@dskeys);
  }
  
  my $items = 0;
  $items = int(@dskeys) + int(@gkeys);
  $changes{WhatSelected} = 'item';
  $changes{NumSelected}  = $items;
  
  # remove items from shopping cart:
  $changes{InsideCart}  = $CART->stats->{AlreadyThere};
  $changes{OutsideCart} = ($items - $changes{InsideCart});
  $changes{Change}      = 'REMOVE';
  
  # Send the results:
  send_doc(Changes => \%changes)
    or puke "couldn't send listing!";
  1;
}

#------------------------------------------------------------
# submode_SETPKGS_DATASET
#------------------------------------------------------------
# I<Top level:> Set the packages for datasets

sub submode_SETPKGS_DATASET {
  my %changes;
  my $chosen_dskey;    # dskey of chosen dataset
  my $chosen_dset;     # chosen dataset
  my $dset;
  
  #--------------------
  # GET THE DATASETS WE'LL BE MODIFYING...
  #--------------------
  
  # Get granule we just assigned packages to:
  $chosen_dskey = $REQ->param('dskey');
  ($chosen_dset) = $CART->datasets([$chosen_dskey]);
  $chosen_dset or puke 533,
  "data set $chosen_dskey isn't in the shopping cart!";
  
  #--------------------
  # GET THE PACKAGES, AND UPDATE THEM...
  #--------------------
  
  # Get all the packages the user might have selected from:
  my @packages = IU::Package->lookup_dataset($chosen_dskey,
                                             Dataset => $chosen_dset,
                                             SubDir => 'search');
  @packages or puke 560,
  "Data set claims to have ordering options, but none were supplied.";
  
  #--------------------
  # Update packages and account info:
  # If any media limits are encountered during the update,
  # add them to the changes hash, they will be needed to
  # print the form.
  #--------------------
  $changes{Limits} = update_packages(ModifyItems => [$chosen_dset],
                                     ChosenItem  => $chosen_dset,
                                     Packages    => \@packages);
  
  #--------------------
  # SEND THE RESULTS...
  #--------------------
  # Send the results:
  send_doc(Changes => \%changes)
    or puke "couldn't send listing!";
  1;
}

#------------------------------------------------------------
# submode_SETPKGS
#------------------------------------------------------------
# I<Top level:> Set the packages

sub submode_SETPKGS {
  my %changes;
  my $chosen_gkey;     # gkey of granule we did "Choose Package" for
  my $chosen_gran;     # granule we did "Choose Package" for
  my $chosen_dskey;    # dskey of chosen granule's dataset
  my $chosen_dset;     # chosen granule's dataset
  my $gran;            # a granule
  
  $pageNumber = $REQ->param('page');
  $min = $REQ->param('min');
  my $page = $REQ->param('page'); # page is the next page
  my $notdone = $REQ->param('notdone'); 
  $pageNumber--; # pageNumber is the current page
  my $opt_num = $REQ->param('optnum') || 1;
  my $gkey = $REQ->param('gkey');
  #--------------------
  # GET THE GRANULES WE'LL BE MODIFYING...
  #--------------------
  
  # Get granule we just assigned packages to:
  $chosen_gkey = $REQ->param('gkey');
  ($chosen_gran) = $CART->granules([$chosen_gkey]);
  $chosen_gran or puke 533,
     "data granule $chosen_gkey not in the shopping cart!";
  
  # Get the dataset which that granule belongs to:
  $chosen_dskey = $chosen_gran->dataset_key;
  ($chosen_dset) = $CART->datasets([$chosen_dskey]);
  $chosen_dset or puke "couldn't get cart data set $chosen_dskey!";
  
  # Get the set-for flag from the data set and the CGI form.        
  my $orig_set_for = $chosen_dset->pkg_set_for;
  my $new_set_for = $REQ->param('set-for');
  my $update_dataset = 0;       # flag whether to update data set.
  
  # If the user changed the set-for flag (or wasn't set before,
  # but is set now):
  if ((!defined($orig_set_for) and defined($new_set_for)) or
      defined($orig_set_for) and defined($new_set_for) and
      $orig_set_for ne $new_set_for) {
    # If user changed from apply-only-to-granule to
    # apply-to-all-in-dataset, then put up a warning page.
    # (if we didn't warn before).
    if ($orig_set_for eq 'granule' and $new_set_for eq 'dataset'
        and $REQ->param('warned') ne '1') {
      send_warn_pkg_apply($chosen_gran, $chosen_dset);
      return (1);
    }
    # Store the new set-for flag in the data set object.
    $chosen_dset->pkg_set_for($new_set_for);
    
    # Set a flag to add the data set to the update list.
    $update_dataset = 1;
  }
  #   # Get all granules currently in shopping cart:
  #   my @granules = $CART->granules();
  
  # We may have chosen one or more packages that the user wants
  # applied to ALL granules in the selected gran's dataset...
  my @modify_grans = ();         # granules to modify
  if ($new_set_for eq 'dataset') {
    # Get all granules currently in shopping cart:
    my @granules = $CART->granules();
    @modify_grans = grep { 
      $_->dataset_key eq $chosen_gran->dataset_key
    } @granules;
    
  }
  else {        # granule
    @modify_grans = ($chosen_gran);
  }
  
  # If data set is to be updated, add it to list of modify items.
  push @modify_grans, $chosen_dset  if ($update_dataset);
  
  #--------------------
  # GET THE PACKAGES, AND UPDATE THEM...
  #--------------------
  
  # Get all the packages the user might have selected from:
  my @packages = IU::Package->lookup($chosen_gkey,
                                     Dataset => $chosen_dset,
                                     Granule => $chosen_gran,
                                     SubDir => 'search');
  @packages or puke "no ordering options for $chosen_gkey!";    
  
  #--------------------
  # Update packages and account info:
  # If any media limits are encountered during the update,
  # add them to the changes hash, they will be needed to
  # print the form.
  #--------------------
  $changes{Limits} = update_packages(ModifyItems => \@modify_grans,
                                     ChosenItem  => $chosen_gran,
                                     Packages    => \@packages);
  #--------------------
  # SEND THE RESULTS...
  #--------------------
  if((defined($page) && $page && defined($notdone) && $notdone) || 
      defined($APP->message)) {
    my @optionarray;
    foreach my $p ($REQ->params()) {
      next if($p !~ /^pk/);
      # Some parameters may have multiple values so loop over them.
      # In an array context $REQ->param() always returns an array
      # even if there is only one value. So if param is single value
      # then this will loop just once, but if it is multi-valued then
      # then it will loop until all values have been stored.
      foreach my $v ($REQ->param($p)) {
         next if(!defined($v) || ($v eq ''));
         push (@optionarray , "$p" , "$v");
      }
    }

    $page-- if ($APP->message && $APP->message->as_string);
    switch IW::Mode Tags=>{'mode'    => 'PKGSLCT',
                           'submode' => 'SUBPAGE',
                           'origin' => 'PKGSLCT',
                           'sid'     => $SESSION->id,
                           'page' => $page,
                           'gkey' => $gkey,
                           'optnum' => $opt_num,
                           @optionarray,
                           'endform' => "1"
                          };

  } elsif ($REQ->param('type') eq "JAVA") {
    
    $APP->start_page(Title=>"Short page for Java post response");
    $APP->end_page();
  } 
  else {
    send_doc(Changes => \%changes)
      or puke "couldn't send listing!";
  }
  1;
}

#------------------------------------------------------------
# go
#------------------------------------------------------------
# Main entry point.

sub go {
  my $self = shift;
  # Get session if we need it.
  $APP->get_session();
  login() or puke "can't log in!";     # log in to databases

  #--------------------
  # Need to know the submode, if it is submode MANIFEST
  # the order process started, in this case skip restoring 
  # the saved partial cart.
  #--------------------
  my $submode = ($REQ->param('submode') || 'DEFAULT');
  $self->restore_partial_cart () unless ($submode eq 'MANIFEST');

  $self->go_submode;                   # dispatch submode
  1;
}

#------------------------------------------------------------
# nav_status
#------------------------------------------------------------
# Returns a very short message summarizing the current status.
# Sets links/greyness/etc. depending on whether or not a search
# is running.
# 
# No matter what provide a link to the Shopping Cart.
#
# Do *not* set the indicator in this method! That should be
# done in the specific method that starts the search. Otherwise,
# the navigator would indicate "View Search" when we did an order
# or a browse, and that would be wrong.
#

sub nav_status {
  return ("Message" => "Nothing in cart",
          "Link"    => "",
          "Mode"    => "CART",
          "Submode" => "LIST",
         );
}

#------------------------------------------------------------
# submode_CLEANUP
#------------------------------------------------------------
# Ok, so the user has REJECTED the Disclaimer.
# What do we do now?
# Reset the disclaimer variables, so that when GOOD
# granules get added the Disclaimer doesn't appear
# ( Which is why this method exists, to fix that bug)

sub submode_CLEANUP {
  # Send the results:
  my $return_to_sender = $REQ->param('prevmode') || 'GRANLIST';

  # Hope I didn't forget anybody?
  $SESSION->state->{'DisclaimerQAFlag'} =  0;
  $SESSION->state->{'DisplayHoldMessage'}  = 'no';

  # Return to wince you came from!
  switch IW::Mode Tags=>{'mode'    => $return_to_sender,
                           'submode' => 'LIST',
                           'sid'     => $SESSION->id,
                           'bogus'   =>'Fixes a CGI::BasePlus bug',
                           'endform' => "1"
                          };
  1;
}

#------------------------------------------------------------
# submode_PARTIAL
#------------------------------------------------------------
# Submode that allows users to submit a partial order.
# When users select a media type with limited capacity,
# instead of forcing the user to delete previously selected
# items from the cart, now we allow them to submit a partial
# order only for the items with options selected.
#
# Submode PARTIAL saves the items that need options to be selected
# into a partial cart database "partialcart.gdb".
# 
# After saving the items into the partial cart, the items are removed
# from the actual cart.  The reason is that orders cannot be submitted
# if any item in the cart is missing ordering options.
# Once all the items are saved and the cart is ready for order, we
# continue with the ordering process as usual.
#
sub submode_PARTIAL {
   my $self = shift;

   # Log into PARTIALCART database (creating it if necessary):
   $PARTIALCART = undef;
   $PARTIALCART = IU::InvStore->open($SESSION->path_to('partialcart.gdb'),
                                   'WRCREAT')
       || puke "couldn't open partialcart database";

   my $item;
   my @items;

   #------------------------------------------------------------
   # Save all the granules in the cart with no ordering options.
   # For every granule that needs to be saved, need to get the
   # reference to it and its granule_key.
   #
   # Add the granules to the partial cart, 
   # then remove them from the cart.
   #------------------------------------------------------------
   my @granules = ();
   my @granules_keys = ();

   @items = $CART->granules;
   foreach $item (@items) {
      if ($item->order_status eq "NEED_PKGS") {
         push (@granules, $item);
         push (@granules_keys, $item->granule_key);
      }
   }
   $PARTIALCART->add_granules(\@granules) if (@granules);
   $CART->remove_granules(\@granules_keys) if (@granules_keys);

   #------------------------------------------------------------
   # Save all the datasets in the cart with no ordering options.
   # For every dataset that needs to be saved, need to get the
   # reference to it and its dataset_key.
   #
   # Add the datasets to the partial cart, 
   # then remove them from the cart.
   #------------------------------------------------------------
   my @datasets = ();
   my @datasets_keys = ();

   # Get all the datasets currently in the cart.
   @items = $CART->datasets;
   foreach $item (@items) {
      # Only select the ones with no ordering options.
      if ($item->order_status eq "NEED_PKGS") {
         push (@datasets, $item);
         push (@datasets_keys, $item->dataset_key);
      }
   }
   $PARTIALCART->add_datasets(\@datasets) if (@datasets);
   $CART->remove_just_datasets(\@datasets_keys) if (@datasets_keys);


   #------------------------------------------------------------
   # Continue with the order process as usual.
   #------------------------------------------------------------
   switch IW::Mode Tags => {'mode'    => 'ORDER',
                            'submode' => 'NEWFORM'};

   1;
}

#------------------------------------------------------------
# restore_partial_cart
#------------------------------------------------------------
# Subroutine restore_partial_cart restores any previously saved
# cart if a partial order was submitted.
#
# To restore the partial cart, a saved cart must exist, and the
# inventory results database must be available.  If any of these
# databases is missing, skip it.
#

sub restore_partial_cart {
   my $self = shift;

   # Log into PARTIALCART database (okay if not there):
   eval {
      $PARTIALCART = undef;
      return unless -e $SESSION->path_to('partialcart.gdb');
      $PARTIALCART =
        IU::InvStore->open($SESSION->path_to('partialcart.gdb'), 'WRCREAT');
   };

   # Return if there is nothing saved or if the results are gone.
   return unless ($PARTIALCART && $INVSTORE);

   my $item;
   my @items;

   #------------------------------------------------------------
   # Restore all the granules saved in the partial cart.
   # Since granule information is more complex, the parent dataset 
   # information is also needed in the cart.
   #
   # For every granule that needs to be restored:
   #    1. Get its granule_key from the partial cart file.
   #    2. Get its dataset_key from the partial cart file.
   #    3. Get the granule from the results database file using its key.
   #    4. Get its dataset from the results database file using its key.
   #    5. Add the granule to the cart.
   #    6. Add the granule dataset parent to the cart.
   #------------------------------------------------------------
   my @granules = ();
   my @granules_keys = ();
   my @datasets = ();
   my @datasets_keys = ();

   @items = $PARTIALCART->granules;
   foreach $item (@items) {
      push (@granules_keys, $item->granule_key);
      push (@datasets_keys, $item->dataset_key);
   }
   if (@granules_keys) {
      @granules = $INVSTORE->granules(\@granules_keys);
      @datasets = $INVSTORE->datasets(\@datasets_keys);
      $CART->add_granules(\@granules);
      $CART->add_datasets(\@datasets);
      $PARTIALCART->remove_granules(\@granules_keys);
   }

   #------------------------------------------------------------
   # Restore all the datasets saved in the partial cart.
   # For every dataset that needs to be restored:
   #    1. Get its dataset_key from the partial cart file.
   #    2. Get the dataset from the results database file using its key.
   #    3. Then add the dataset to the cart.
   #------------------------------------------------------------
   @datasets = ();
   @datasets_keys = ();
   @items = $PARTIALCART->datasets;
   foreach $item (@items) {
      push (@datasets_keys, $item->dataset_key);
   }
   if (@datasets_keys) {
      @datasets = $INVSTORE->datasets(\@datasets_keys);
      $CART->add_datasets(\@datasets);
      $PARTIALCART->remove_just_datasets(\@datasets_keys);
   }

   # Once restored, empty partial cart to avoid reloading it.
   unlink $SESSION->path_to('partialcart.gdb');

   1;
}

#------------------------------------------------------------
1;
