=head1 NAME

IW::ModeDEFN -- an imswww mode module to serve up definition pages

=head2 Parameters

The ModeDEFN module is passed three arguments via the $REQ mechanism
in addition to all the standard stuff that is available through global
parameters and other such mechanisms:

mode -- unused, but presumably always 'DEFN'

type -- the class of thing for which a definition is to be provided.
One of 'campaign', 'dataset', 'data_center', 'parameter', 'sensor',
'source', 'processing_level'.

item -- the specific thing for which a definition is to be provided
This varies greatly with the type, but really does not matter to the code.

=cut

package IW::ModeDEFN;

use IU;
use IU::Event;
use IU::HTTP;
use IU::Log;

use IW::App;
use IW::Util qw(:url :output);

@ISA = qw(IW::Mode);


#
# no_defn -- pukes (but nicely) when unable to get a definition page.
#
sub no_defn {
    my ($type, $item) = @_;

    my %type_text = 
	('campaign'    => 'Campaign', 
	 'dataset'     => 'Dataset Identifier',
	 'data_center' => 'Data Center',
	 'parameter'   => 'parameter',
	 'sensor'      => 'Sensor',
	 'source'      => 'Source',
	 'processing_level'      => 'Processing Level' );

    puke(540, "Unable to obtain a definition for the " .
	 lc($type_text{$type}) . ' "' . lc($item) . '"' );
}


#
# serve_defn -- serves up a definition file for inclusion in mid-page.
#
# There are two problems with including the definition files directoy
# in a web page: they contain a <TITLE> tag, which by has no business
# occuring in the body of a document; and they reference a
# "thin_blue_line" .gif file, which sounds too much like a cop show,
# and which isn't in the relative location specified in the definition
# files.  
#
# The serve_defn routine makes a few simple-minded assumptions about
# the format of the definition files in order to avoid having to parse
# these files.  The body is assumed to begin with the first <H[1-6]>
# header tag, and all <IMG> tags are assumed to be horizontal lines
# that get replaced with a <HR> horizontal rule tag.
#
sub serve_defn {
    my ($url) = @_;
    my ($body);

    my $SRC = IU::HTTP->get($url) || return 0;
    while (<$SRC>) {
	$body = $body || m/<h[1-6]>/i;
	if ($body) {
	    print ((m/<img /i)? "<hr><p>\n": $_);
	}
    }
    return 1;
}

#
# standard go() and submode handler.
#
sub go {
    my $self = shift;
    $APP->get_session();
    $self->go_submode();
}
    
#
# submode_DEFAULT
#
sub submode_DEFAULT {
    my $self = shift;
    $self->submode_GETDEFN;
}

#
# submode_OUT_OF_LINE
#
sub submode_OUT_OF_LINE {
    my $self = shift;
    $self->suppress_links(1);
    $self->submode_DEFAULT;
}

#
# -- main entry point for the definition mode module
# 
sub submode_GETDEFN {

    my $self = shift;

    # Setup stuff -- retrieve the session info, put up a header and
    # fetch mode parameters.

    $APP->get_session();
    $APP->suppress_links($self->suppress_links);
    $APP->start_page(Title=>'Definition', Context=>'definition');

    my $mode = $REQ->param('mode') || 'DEFN'; # N/A?
    my $type = $REQ->param('type');
    my $item = $REQ->param('item');
    $item = 0 if(!$item and $type eq 'processing_level');

    # Assemble the filenames for the various places to look for a
    # definition file matching this type and item.

    my $munged_item = $item;
    $munged_item =~ s/[^A-Z0-9\-]/_/ig;
    $munged_item =~ s/_+/_/g;
    $defn_suffix = "$type/$munged_item.html";
    my $master_url = "${IU::ImsDefnUrl}$defn_suffix";
   debug "master_url = ", $master_url;

    my $success = serve_defn($master_url);

    # Record the definition event, along with the parameters used and
    # the status of the definition page search.

    ims_event('definition', 
	      'defn_type'=>$type, 
	      'defn_item'=>$item, 
	      'defn_okay'=>$success );

    # Either serve the definition page or puke, based on the success
    # of the definition page search.  Return the appropriate status
    # code, in either case.

    if ($success) {
	$APP->end_page();
    } else {
	no_defn($type, $item);
    }

    return $success;
}

#-----------------------------------------------------------
# suppress_links
#-----------------------------------------------------------
# Set/get link suppression.

sub suppress_links {
    my ($self, $parm) = @_;
    $self->{SuppressLinks} = $parm if defined $parm;
    $self->{SuppressLinks};
}

1;
