package IW::ModeECSPROFILE;

=head1 NAME

IW::ModeECSPROFILE - Retrieve or Update ECS user profile.

=head1 NOTES

Invoked when updating or retrieving ECS user profile.
This mode is invoked from either the user prefs page
or from the order form.

If update ECS profile successfully, displays an acknowledgement
page with an "OK" button to go back to where the user came.
If update ECS profile fails, display the failure page.
If retrieve ECS profile successfully, display originating
page (prefs or order) with the updated values.
If retrieve ECS profile fails, display the failure page.

=head2 ASSUMPTIONS

CGI tag "submode" must be either 'UPDATE', 'RETRIEVE', 'ESTABLISH' or 'CHANGE'.
If coming from user prefs page, CGI tag "origin"='PREFS'.
If coming from order form, CGI tag "origin"='ORDER'.

=head1 RCS INFO

$Id: ModeECSPROFILE.pm,v 1.34 2003/03/19 15:13:12 seufert Exp $

=cut

use IU;
use IU::Log;
use IU::Session;
use IU::ODL;
use IU::Event;
use IU::ECSProfileRequest;

use IW::App;
use IW::User;
use IW::Client;
use IW::CommStat;
use IW::Comm::Manager;
use IW::Comm::ProfileRetrievalResult;
use IW::Comm::ProfileSubmitResult;
use IW::Comm::ProfileUpdateResult;
use IW::Comm::UserPasswordChangeResult;
use strict;
use vars qw(@ISA);
@ISA = qw(IW::Mode IW::Comm::Manager);

# use Data::Dumper;  # For debugging

#------------------------------------------------------------
#
# Globals
#
#------------------------------------------------------------
# Map a search type to a log event name:
my %ActivityToEvent = (
   PROFILE_RETRIEVAL_REQUEST => 'RETRIEVE_ECS_USER_PROFILE',
   PROFILE_UPDATE_REQUEST    => 'UPDATE_ECS_USER_PROFILE',
   PROFILE_SUBMIT_REQUEST    => 'SUBMIT_ECS_USER_PROFILE',
   USER_PASSWORD_CHANGE_REQUEST    => 'CHANGE_ECS_PASSWORD',
);

#------------------------------------------------------------
# mode
#------------------------------------------------------------
# I<Instance method, override.>
# Return the mode that comes back here.

sub mode { 'ECSPROFILE' }

#------------------------------------------------------------
# daac [Daac]
#------------------------------------------------------------
# I<Instance method>
# Get/set the desired daac for the profile retreive/update

sub daac {
    my ($self, $value) = @_;
    $self->{Daac} = $value if @_ > 1;
    $self->{Daac};
}

#------------------------------------------------------------
# homeDaac [Daac]
#------------------------------------------------------------
# I<Instance method>
# Get/set the desired homeDaac for the profile retreive/update

sub homeDaac {
    my ($self, $value) = @_;
    $self->{HomeDaac} = $value if @_ > 1;
    $self->{HomeDaac};
}

#------------------------------------------------------------
# reset [VALUE]
#------------------------------------------------------------
# Set/get the reset flag
# The reset flag is used when the password is being reset by admin mode
# or by "forgot your password".
sub reset { _one('Reset', @_) }

#------------------------------------------------------------
# _one VARNAME,SELF,[VALUE]
#------------------------------------------------------------
# I<Private utility.>  Get/set a single-valued instance variable.
# Return undef if it's undefined.
sub _one {
    my ($varname, $self, $value) = @_;
    $self->{$varname} = $value if (@_ > 2);
    $self->{$varname};
}



#------------------------------------------------------------
# retrieving
#------------------------------------------------------------
# I<Instance method.>
# Return true if submode was RETRIEVE.  Assumes that submode
# UPDATE or RETRIEVE was called first.

sub retrieving {
  my $self = shift;
  return (exists($SESSION->state->{SearchType}) and
	  $SESSION->state->{SearchType} eq 'PROFILE_RETRIEVAL_REQUEST');
}

#------------------------------------------------------------
# updating
#------------------------------------------------------------
# I<Instance method.>
# Return true if submode was UPDATE.  Assumes that submode
# UPDATE or RETRIEVE was called first.

sub updating {
  my $self = shift;
  my $return  = (exists($SESSION->state->{SearchType}) and
		 $SESSION->state->{SearchType} eq 'PROFILE_UPDATE_REQUEST');
  return $return;
}

#------------------------------------------------------------
# creating
#------------------------------------------------------------
# I<Instance method.>
# Return true if submode was ESTABLISH.

sub creating {
  my $self = shift;
  my $return  = (exists($SESSION->state->{SearchType}) and
		 $SESSION->state->{SearchType} eq 'PROFILE_SUBMIT_REQUEST');
  return $return;
}

#------------------------------------------------------------
# changing
#------------------------------------------------------------
# I<Instance method.>
# Return true if submode was CHANGE.

sub changing {
  my $self = shift;
  my $return  = (exists($SESSION->state->{SearchType}) and
	 $SESSION->state->{SearchType} eq 'USER_PASSWORD_CHANGE_REQUEST');
  return $return;
}

#------------------------------------------------------------
# send_page
#------------------------------------------------------------
# I<Instance method>.

sub send_page {
  my $self = shift;
  
  $APP->suppress_links(1);
  
  # Start the page:
  $APP->start_page(Title=>"Coding Error");
  
  $HTML->t("There is an error in the coding of ModeECSPROFILE.pm  ")
       ->t("or in the way it is being called.  It should be called ")
       ->t("with submode='UPDATE' or 'RETRIEVE'.");
  
  #--------------------
  # End it and send it:
  #--------------------
  $APP->suppress_key(1);
  $APP->end_page;
  1;
}

#------------------------------------------------------------
# update_ok
#------------------------------------------------------------
# I<Instance method>.
# Set a message announcing successful profile update, and
# switch back to original mode.

sub update_ok {
  my $self = shift;
  my $id = $SESSION->id;

  $APP->message("Your profile has been successfully updated ".
		"with your current session values.");
  switch IW::Mode Tags=>{'mode'    => $self->{Origin},
			 'u'       => time,
			 'sid'     => $id,
			 'bogus'   =>'Fixes a CGI::BasePlus bug',
			};
}

#------------------------------------------------------------
# create_ok
#------------------------------------------------------------
# I<Instance method>.
# Set a message announcing successful profile submit, and
# switch back to original mode.

sub create_ok {
  my $self = shift;
  my $id = $SESSION->id;
  my $origin = $self->{Origin};

  my $submode = ($origin eq 'ADMINECSACCOUNT') ? 'FINAL' : undef;
  $APP->message("Your account has been created.");
  switch IW::Mode Tags=>{'mode'    => $origin,
			 'submode' => $submode,
			 'u'       => time,
			 'sid'     => $id,
			 'bogus'   =>'Fixes a CGI::BasePlus bug',
			};
}

#------------------------------------------------------------
# change_ok
#------------------------------------------------------------
# I<Instance method>.
# Set a message announcing successful user password change, and
# switch back to original mode.

sub change_ok {
  my $self = shift;
  my $id = $SESSION->id;

  $APP->message("Your password has been changed.");
  switch IW::Mode Tags=>{'mode'    => $self->{Origin},
			 'u'       => time,
                         'submode' => 'PASSWDFINAL',
			 'sid'     => $id,
			 'bogus'   =>'Fixes a CGI::BasePlus bug',
			};
}

#------------------------------------------------------------
# install_prefs
#------------------------------------------------------------
# I<Instance method>.
# We have retrieved user profile info from ECS.
# Use it to update the WG user prefs.

sub install_prefs {
   my $self = shift;

   my $newfile = 'newprefs.dat';
   my $result;
   eval { $result =
          IW::Comm::ProfileRetrievalResult->read_last ($newfile) }
     or puke "Didn't successfully retrieve ECS profile. ", $@;

   # Note: $result is an IW::User object that may have Errors in it
   # or may have missing fields.  Note that if a field was in error
   # (too long) it will be blank now and an error message stored.
   # Add the pieces found in the Profile to the current session
   # User object.  Request address, billing address, shipping address
   # and affiliation will be replaced with info from ECS. The rest
   # of the user preferences such as display options, V0 access key
   # are kept the same.

   my @updates;
   if ($result->req_addr) {
      push @updates, 'req_addr' => $result->req_addr;
   }
   if ($result->ship_addr) {
      push @updates, 'ship_addr' => $result->ship_addr;
   }
   if ($result->bill_addr) {
      push @updates, 'bill_addr' => $result->bill_addr;
   }
   if ($result->affil) {
      push @updates, 'affil' => $result->affil;
   }
   # Update and install the session User object if we got any pieces.
   # If no current User object, this creates a new one (but, of course
   # a User object was created from the CGI when the retrieve button
   # was pressed).
   IW::User->update (@updates) if (scalar (@updates));

   # If any errors in retrieved profile, set as APP message.
   my $err;
   if (($err = $result->errors) && $err->any_errors) {
      $APP->message (new IW::Note Class=>'ERROR',
                     HTML =>join ("<BR>\n", $err->get));
   } else {
      # Change message at page top to success.
      $APP->message ('Your profile has been successfully
                      retrieved and installed into your current
                      session.  Please check to make sure that all
                      of the required fields have been filled in.');
   }
   # Determine where we came from. Removed assumption of 'PREFS' if not 'ORDER'.
   my $origin = $SESSION->state->{ProfileFormMode};

   switch IW::Mode Tags=>{'mode'    => $origin,
                          'u'       => time,
                          'sid'     => $SESSION->id,
                          'bogus'   =>'Fixes a CGI::BasePlus bug',
                          };
   1;
}

#------------------------------
#
# COMM HANDLERS...
#
#------------------------------

#------------------------------------------------------------
# comm_preflight
#------------------------------------------------------------
# Do some pre-client run preparations:
#   Create a request object for retrieve or update.
#   Write it to an ODL file: profile_request.odl
#   Go back to comm::Manager to start().

sub comm_preflight {
   my $self = shift;

   my $reset = $self->reset;
   my $request;

   # Create a request object for the type of request.
   # If the reset flag is on, pass it as an additional parameter to
   # ECSProfileRequest object.
   $reset ? ($request = new IU::ECSProfileRequest($self->{SearchType},$reset)) :
            ($request = new IU::ECSProfileRequest($self->{SearchType}));
   my $daac = $self->daac;
   my $home_daac = $self->homeDaac;
   my $origin = $self->origin;
   $request->daac($daac);
   $request->homeDaac($home_daac);
   $request->origin($origin);
   # Create a file to hold the request ODL for input to imsclient.
   my $reqfile = $SESSION->path_to('profile_request.odl');
   # Open the file or die.
   open REQUEST, ">$reqfile" or puke "open $reqfile: $!";
   $request->to_search_ODL->output(\*REQUEST);
   IU::ODL->end(\*REQUEST);
   close REQUEST;

   # Hold the filename.
   $self->{ReqFile} = $reqfile;
   return ('GO');
}

#------------------------------------------------------------
# comm_start_search
#------------------------------------------------------------
# Start the profile request.  Create client object and start it.

sub comm_start_search {
   my $self = shift;

   # Start the client:
   my $client = start IW::Client Activity=>$self->{SearchType},
                      Infile=>$self->{ReqFile};
   $client or puke "couldn't start imsclient:" .
      $self->{SearchType} . ":" . $self->{ReqFile};

   # Done!
   return 'STARTED';
}

#------------------------------------------------------------
# comm_finish_up
#------------------------------------------------------------
# Use default IW::Comm::Manager::comm_finish_up()

#------------------------------------------------------------
# comm_ok_page
#------------------------------------------------------------
# Finished OK.
# If update - switch to origin with success message.
# If retrieve - install prefs and switch to origin.

sub comm_ok_page {
   my ($self, %params) = @_;
   if ($self->updating) {
      $self->update_ok();
   } elsif ($self->retrieving) {
      $self->install_prefs();
   } elsif ($self->creating) { # Creating an ECS account
      $self->create_ok();
   } elsif ($self->changing) { # Changing a password for ECS
      $self->change_ok();
   } else {
      # Don't know - probably coding error.  Switch to prefs.
      $self->update_ok();
   }
   1;
}

#------------------------------------------------------------
# comm_fail_page
#------------------------------------------------------------
# Failure!  Set App message and go back to where you came from.

sub comm_fail_page {
   my $self = shift;
   my $id = $SESSION->id;

   my $origin = $self->{Origin};
   # Failure while creating an ECS account
   if ($self->creating) {
      my $pr =  IW::Comm::ProfileSubmitResult->read_last;
      if(ref($pr)) {
         my $status_code_comment = $pr->status_code_comment;
         my $status_code = $pr->status_code;
         if ($status_code_comment) {
            $APP->message
             ("$status_code_comment, Status Code: $status_code");
         } else {
            $APP->message("Your account was not established.");
         }
      } else {
         $APP->message("The ECS system was down.  Please try again later.");
      }
   } elsif ($self->changing) {
      # Failure while changing an ECS password.
      my $pr =  IW::Comm::UserPasswordChangeResult->read_last;
      if(ref($pr)) {
         my $status_code_comment = $pr->status_code_comment;
         my $status_code = $pr->status_code;
         if ($status_code_comment) {
            $APP->message
             ("$status_code_comment, Status Code: $status_code");
         } else {
            $APP->message("Your password was not changed.");
         }
      } else {
	 # Failure on the other ECS profile messages
         $APP->message("The ECS system was down.  Please try again later.");
      }
      switch IW::Mode Tags=>{'mode'    => $origin,
                             'u'       => time,
                             'submode' => 'PASSWDFINAL',
                             'fail'       => '1',
                             'sid'     => $id};
      return 1;
   } elsif ($self->updating) {
      # Failure while updating an ECS profile.
      my $pr =  IW::Comm::ProfileUpdateResult->read_last;
      if(ref($pr)) {
         my $status_code_comment = $pr->status_code_comment;
         my $status_code = $pr->status_code;
         if ($status_code_comment) {
            $APP->message
             ("$status_code_comment, Status Code: $status_code");
         } elsif($status_code != 1) {
            $APP->message
             ("Your profile was not updated, please contact User Services.");
         }
      } else {
	 # Failure on the ECS profile messages
         $APP->message("The ECS system was down.  Please try again later.");
      }
   } elsif ($self->retrieving) {
      # Failure while retrieving an ECS profile.
      my $pr =  IW::Comm::ProfileRetrievalResult->read_last;
      if(ref($pr)) {
         my $status_code_comment = $pr->status_code_comment;
         my $status_code = $pr->status_code;
         if ($status_code_comment) {
            $APP->message
             ("$status_code_comment, Status Code: $status_code");
         } elsif($status_code != 1) {
            $APP->message
             ("Your profile was not retrieved, please contact User Services.");
         }
      } else {
	 # Failure on the ECS profile messages
         $APP->message("The ECS system was down.  Please try again later.");
      }
   } else {
      $APP->message("Your profile was NOT retrieved.");
   }
   my $submode = ($origin eq 'ADMINECSACCOUNT') ? 'FINAL' : undef;
   switch IW::Mode Tags=>{'mode'    => $origin,
                          'submode' => $submode,
                          'u'       => time,
                          'fail'       => '1',
                          'sid'     => $id};
   1;
}

#------------------------------------------------------------
# comm_trace COMMSTATUS
#------------------------------------------------------------
# Backend for submode COMMTRACE.
# If the client is still running, send a comm status screen.
# Else, output the appropriate results page.

sub comm_trace {
   my ($self, $commstat) = @_;

   # Get client:
   my $client = $commstat->client;

   # Output only if client is running:
   if ($client->running) {                  # non-final page...

      # Have we been running too long?
      if ($client->run_time >= $self->comm_timeout) {
         $self->comm_stop_it($commstat);       # do a stop or kill
      } else {                                 # comm status screen
         $commstat->send_doc(Wait=>$self->comm_wait);
      }
   } else {                                    # final-page...
      $self->comm_client_stopped(CommStat=>$commstat);
   }
   1;
}

#------------------------------
#
# SUBMODE HANDLERS...
#
#------------------------------

#------------------------------------------------------------ 
# submode_DEFAULT 
#------------------------------------------------------------ 
# I<Instance method>.  What to do if there's no explicit submode.
# submode should be RETRIEVE or UPDATE.

sub submode_DEFAULT {
   shift->submode_SHOW; 
}

#------------------------------------------------------------
# submode_SHOW
#------------------------------------------------------------
# I<Instance method>.
# Just show the search form page, without taking any action.

sub submode_SHOW {
   my $self = shift;
   # Output the search form page:
   $self->send_page;
}

#------------------------------------------------------------
# submode_SUBMIT
#------------------------------------------------------------
# Submit the Request.

sub submode_SUBMIT { 
   my $self = shift;

   ims_event($ActivityToEvent{$self->{SearchType}});
   $self->submode_COMMSTART;           # go!
   1;
}

#------------------------------------------------------------
# submode_COMMTRACE
#------------------------------------------------------------
# Draw a single "frame" of the comm status screen (might be
# the final frame).

sub submode_COMMTRACE {
   my $self = shift;

   # Get a comm status generator, complete with stop button:
   my $stop     = $self->comm_stop_url;
   my $client   = (attach IW::Client) ||
      die "couldn't attach to client!";
   my $commstat = new IW::CommStat Client=>$client, Stop=>$stop;

   $self->comm_trace($commstat);
}

#------------------------------------------------------------
# submode_RETRIEVE
#------------------------------------------------------------
# Set up flags for ECS profile retrieval request and start by
# switching to common submode SUBMIT.

sub submode_RETRIEVE {
   my $self = shift;
   $self->{SearchType} = 'PROFILE_RETRIEVAL_REQUEST';
   $SESSION->state->{SearchType} = 'PROFILE_RETRIEVAL_REQUEST';
   # If coming from prefs and retrieving, set next submode to UPLOAD.
   $self->{OriginSub} = 'UPLOAD' if ($self->{Origin} eq 'PREFS');
   my $gateway = $IU::EcsProfileSmc;
   my @selection = split(/;/, $gateway);
   my $daac = $selection[0];
   $self->daac($daac);
   my $user = $SESSION->state->{User} || new IW::User;
   my $keys = $user->keys || new IW::UserKeys;
   $user->install;
   $self->submode_SUBMIT();
}

#------------------------------------------------------------
# submode_UPDATE
#------------------------------------------------------------
# Set up flags for ECS profile update request and start by
# switching to common submode SUBMIT.

sub submode_UPDATE {
   my $self = shift;
   $self->{SearchType} = 'PROFILE_UPDATE_REQUEST';
   $SESSION->state->{SearchType} = 'PROFILE_UPDATE_REQUEST';
   my $gateway = $IU::EcsProfileSmc;
   my @selection = split(/;/, $gateway);
   my $daac = $selection[0];
   $self->daac($daac);
   my $user = $SESSION->state->{User} || new IW::User;
   my $keys = $user->keys || new IW::UserKeys;
   $user->install;
   $self->submode_SUBMIT();
}

#------------------------------------------------------------
# submode_ESTABLISH
#------------------------------------------------------------
# Set up flags for ECS profile update request and start by
# switching to common submode SUBMIT.

sub submode_ESTABLISH {
   my $self = shift;
   my $origin = $REQ->param('origin');
   $self->{SearchType} = 'PROFILE_SUBMIT_REQUEST';
   $SESSION->state->{SearchType} = 'PROFILE_SUBMIT_REQUEST';
   my $home_daac = $REQ->param('home_daac');
   $self->homeDaac($home_daac);
   my $gateway = $IU::EcsProfileSmc;
   my @selection = split(/;/, $gateway);
   my $daac = $selection[0];
   $self->daac($daac);
   my $user;
   if($origin eq 'ADMINECSACCOUNT') {
      $user = $SESSION->state->{User2} || new IW::User;
   } else {
      $user = $SESSION->state->{User} || new IW::User;
   }
   my $keys = $user->keys || new IW::UserKeys;
   $keys->home_daac($home_daac);
   $self->submode_SUBMIT();
}

#------------------------------------------------------------
# submode_CHANGE
#------------------------------------------------------------
# Set up flags for ECS profile update request and start by
# switching to common submode SUBMIT.

sub submode_CHANGE {
   my $self = shift;
   $self->{SearchType} = 'USER_PASSWORD_CHANGE_REQUEST';
   $SESSION->state->{SearchType} = 'USER_PASSWORD_CHANGE_REQUEST';
   my $gateway = $IU::EcsProfileSmc;
   my @selection = split(/;/, $gateway);
   my $daac = $selection[0];
   $self->daac($daac);
   my $user = $SESSION->state->{User2} || new IW::User;
   my $keys = $user->keys || new IW::UserKeys;
   $self->submode_SUBMIT();
}

#------------------------------------------------------------
# submode_TEST
#------------------------------------------------------------
# I<Instance method>.
# Used for manually testing mode, not by the system.

sub submode_TEST {
   my $self = shift;
   $self->{SearchType} = 'PROFILE_RETRIEVAL_REQUEST';
   # If coming from prefs and retrieving, set next submode to UPLOAD.
   $self->{OriginSub} = 'COPY' if ($self->{Origin} eq 'PREFS');
   $self->install_prefs();
}

sub origin {
   my $self = shift;
   return $self->{Origin};
}
#------------------------------------------------------------
# go
#------------------------------------------------------------
# Main entry point for mode.
# submode should be RETRIEVE, UPDATE, ESTABLISH, or CHANGE.

sub go {
   my $self = shift;
   $self->init;
   $self->go_submode;
}

#------------------------------------------------------------
# init
#------------------------------------------------------------
# Init this mode:
#   Get session if not switched here.
#   Store the originating mode and submode names.

sub init {
   my $self = shift;

   # Get session if we need it.
   $APP->get_session();

   # Where did we come from?
   my $origin = $REQ->param('origin');

   # If the reset flag is set in the params, set it in the reset method
   # of this object.
   my $reset = $REQ->param('reset') || undef;
   $self->reset($reset);

   if (defined ($origin)) {
      $SESSION->state->{ProfileFormMode} = $origin;
   } else {
      $origin = $SESSION->state->{ProfileFormMode};
   }
   $self->{Origin} = $origin;
   $self->{OriginSub} = $REQ->param('origin_sub');
   1;
}
1;
