package IW::ModeFORGOT;

=head1 NAME

IW::ModeFORGOT - Allows the user to get a new password for a persistent 
session.

=head1 DESCRIPTION

This mode allows the user to get a new password when they've forgotten it.
It's modeled on similar facilities found on other websites, such as MP3.com
and eBay. The difference here is that instead of keeping the password around
in cleartext, we generate a new random password and e-mail it to them, along 
with a URL which will allow them to get back in and change their password
immediately.

There are a couple of possibilities:

=over 4

=item *

They know that they've forgotten it, but remember their login name.
In this case, they can voluntarily go to the new password page, enter 
their session name, and we can e-mail them a new password.

=item *

They don't know they've forgotten it until they try, but they still remember
their login name. They are sent to a "have you forgotten your password?" page
which allows them either to try entering the system again, or to ask for a 
new password.

=item *

They've forgotten everything, but know they have. They can voluntarily go to
the "forgot my password" link and get a new password by entering their email 
address. We look it up, figure out who they are, and send the new password
accordingly.

=item *

They've forgotten everything and don't know it until they get to the "have you
forgotten your password?" page and discover they can't remember the account
either. They can enter the email, have us look it up, and then get the password
e-mailed to them.

=back

=head1 OO CONSIDERATIONS

ModeFORGOT is designed to be both a standalone class, and a class which can
be inherited from. Subclasses have access to all of this class's screen
management, e-mail, etc.

However, subclasses (e.g., ModeADMINGENPWD) will want the messages to be in
third person, rather than in second person ("the user" rather than "you").
It is necessary for us to be able to track this somehow so that ModeFORGOT
can check it and act accordingly (in e-mail messages, etc.). 

The EmailOrAccount widget has to be given this information so it can set 
everything properly inside itself, and all subclasses will use the 
EmailOrAccount widget to get their input, so it's simplest to store the
Admin mode information in the widget, and ask it to decide what format 
the messages should be in. 

This way, ModeFORGOT can tell it to use second person mode, and then
have all of its output routines be generic. These output routines will 
call the widget's person() method with the alternate texts, and the 
widget will select the proper text for ModeFORGOT without the code 
actually having to know anything other than "the widget knows what to do".

The class hierarchy looks like this:

   ModeFORGOT
      submode_DEFAULT 
        (calls START or INLINE)
        (START calls widget, admin=>0)
      submode_LOOKUP
        (POSTed back to by widget)
        (calls appropriate output routines)
      output routines
        (calls $self->{Widget}->person to arbitrate message text)

   Admin-style subclasses of ModeFORGOT
      submode_DEFAULT 
         (overrides ModeFORGOT)
         (calls widget, admin=>1)
      submode_LOOKUP 
         (POSTed back to by widget)
         (override, calling superclass submode_LOOKUP for functionality,
          or is not present, so superclass submode_LOOKUP is called)
         (calls appropriate output routines)
      output routines
         (inherited)

=cut

use strict;

use IU;
use IU::UserAdmin;
use IU::Log;
use IU::Mail;
use IU::Event;
use IU::Session;

use IW::App;
use IW::Util qw(:url :output :string);
use IW::ModePREFS;
use IW::Note;

use IW::Widget::EmailOrAccount;
use URI::Escape;
use vars qw($User @ISA);
@ISA = qw(IW::Mode);

#-------------------------------------------------------------------------
# Utilities.
#-------------------------------------------------------------------------
# Generic message routine. Shows message of appropriate class
# as an App message at the top of the page. Retains modality.
sub _show_msg {
    my $self = shift;
    my ($class, $msg, $modal) = @_;
    $APP->message (new IW::Note Class=>$class, Text => $msg);
    $self->submode_START($modal);
}

# Error message.
sub show_error {
   my $self = shift;
   my $modal = pop(@_);
   my $msg = join("",@_);
   $self -> _show_msg('ERROR',$msg, $modal);
}

# Info message.
sub show_info {
   my $self = shift;
   my $modal = pop(@_);
   my $msg = join("",@_);
   $self -> _show_msg('INFO',$msg, $modal);
}


# Mail password back.
sub send_done_mail {
    my ($self,%params) = @_;
    local $_;
    foreach (qw(To Password Account)) {
       puke 600, "Implementation error - no '$_' supplied."
          unless $params{$_};
    }
    my $mail = new IU::Mail;
    $mail->from($IU::SiteAdminEmail);
    $mail->to($params{To});
    $mail->subject("Password Change");
    
    my $FH = $mail->open() || puke 600,"Can't open outgoing e-mail: $!";
    my $request = $self->person(", per your request,"," by the administrator");
    print $FH <<EOS;
Your password has been changed$request for your registered user
account at this URL: http://$IU::HttpdHost$IU::WelcomeUrl.

   New password: $params{Password}

It is highly recommended that you change your password again as soon as you 
can, as the security of e-mail cannot be guaranteed. To do so, click on the
"User Preferences" link (located on the left side of the screen in the
navigation area) after you have signed in using the password we've e-mailed
to you. You can then use the "Change Password" button on this page to do it.

EOS
   $mail->send;
}

# Mail list of accounts back.
sub send_list_mail {
    my ($self,%params) = @_;
    local $_;
    foreach (qw(To List)) {
       puke 600, "Implementation error - no '$_' supplied."
          unless $params{$_};
    }
    my $mail = new IU::Mail;
    my @accounts = @{$params{List}};

    $mail->from($IU::SiteAdminEmail);
    $mail->to($params{To});
    $mail->subject("EDG Accounts List");
    
    my $FH = $mail->open() || puke 600,"Can't open outgoing e-mail: $!";
    my $request = $self->person("You requested",
				"The system administrator has e-mailed you");
    print $FH <<EOS;
$request a list of EDG users registered at $IU::HttpdHost
who have your e-mail address as the owner. The following registered users 
fall into that category:

EOS

    foreach (@accounts) {
      print $FH "   * $_\n";
    }

    print $FH <<EOS;

You may select from this list and return to the EDG at $IU::HttpdHost 
to either change your password, or sign in using one of these accounts.

If you cannot remember the password for the account you want to sign in to,
you can go to the EDG at $IU::HttpdHost and get a new one e-mailed to you.
EOS
   $mail->send;
}

# Success message.
sub show_okay {
   my $self = shift;
   my $modal = pop(@_);
   my $msg = join("",@_);
   $self -> _show_msg('OK',$msg, $modal);
}

# Second-person/third-person checking.
#
# The widget is realized by either this class or a subclass of it,
# and is realized by whichever class, which chooses whether or not
# it was realized in Admin mode or not. Rather than track this all 
# over the place, it's easier to just ask the widget, especially in
# code which is inherited by the subclass.

sub person {
   my ($self, $second, $third) = @_;
   my $widget = $self -> {Widget};
   debug "Widget: ".ref($widget);
   $widget->person($second,$third);
}
   
#---------------------------------------------------------------------
# send_ecs_change_req  IW::User
#---------------------------------------------------------------------
# Send the ecs user password change request.
# 1. Make sure we have ECS name & password.
# 2. Set the correct parameters for ModeECSPROFILE.
# 3. Switch control to ModeECSPROFILE.
# Input: The current session User object.  Output: none.

sub send_ecs_change_req {
   my $self = shift;
   my $user = shift;
   my $widget = $self->{Widget};

   # Check if we have ECS name and password keys.
   my $keys = $user->keys;
   unless (defined ($keys) and $keys->ecs_username and
           $keys->ecs_authentication_key and
           $keys->old_ecs_authentication_key) {

      my $new_message = 'Error: You must supply BOTH your User
                         Name and Password.';
      # Required keys are missing.  Warn user and redraw form.
      # If the current APP message contains "were accepted!" then no
      # errors before.  Replace message with ECS name/passwd error.
      # If there were already error messages, append new message.
      if ($APP->message &&
          $APP->message->as_string =~ /were accepted!/) {
         $APP->message (new IW::Note Class=>'ERROR',
                        Text => $new_message);
      } elsif ($APP->message) {
      # There was an error message so append new message.
         my $text = $APP->message->as_string;
         $APP->message (new IW::Note Class=>'ERROR',
                        HTML => "$new_message<BR>\n$text");
      }
      return (1);
   }
   if ($APP->message) {
      unless($APP->message->as_string =~ /were accepted!/) {
         puke " failed here";
         return (1);
      }
   }

   #  Set the parameters for ModeECSPROFILE.
   my $next_submode = 'CHANGE';

   # Verify next submode.
   unless ($next_submode eq 'UPDATE' or $next_submode eq 'RETRIEVE' or
           $next_submode eq 'ESTABLISH' or $next_submode eq 'CHANGE') {
      puke 600, "Invalid submode for ModeECSPROFILE: $next_submode";
   }
   # Switch control to ModeECSPROFILE with correct tags.
   # This was modified to send this message to SMC instead of the home DAAC.
   my $admin = $widget->admin;
   my $origin = $admin ? 'ADMINGENPWD' : 'FORGOT';
   switch IW::Mode Tags=>{'mode'    =>'ECSPROFILE',
                          'submode' =>$next_submode,
                          'origin'  =>$origin,
                          'reset' =>'1'};
   1;
}

#-------------------------------------------------------------------------
# OUTPUT TO SCREEN
#-------------------------------------------------------------------------

#------------------------------
#
# Submode handlers
#
#------------------------------

#------------------------------------------------------------
# submode_DEFAULT
#------------------------------------------------------------
# instance method
#
# Defaults to just showing the page, with the navigator still on.

sub submode_DEFAULT {
    my $self = shift;
    $self -> submode_START;
}

#------------------------------------------------------------
# submode_INLINE
#------------------------------------------------------------
# instance method
#
# Turns on the navigator before showing the page.

sub submode_INLINE {
   my $self = shift;
  $self->submode_START(0);
}



#---------------------------------------------------------------------
# submode_PASSWDFINAL
#---------------------------------------------------------------------
# instance method
#
# Handle changing a password.

sub submode_PASSWDFINAL {
   my $self = shift;
   my $fail = $REQ->param('fail');
   $self->{Widget} = IW::Widget::EmailOrAccount->excyst;
   my $widget = $self->{Widget};
   my $modal = $widget->modality;
   # If the ECS password change failed, display the "Failed" page.
   if($fail eq '1') {
      # Start the page:
      $APP->start_page(Title=>"Password Change Failed",
                       Context=>'setup');
 
      $HTML -> P
            -> t("Your password was not changed." );
      $HTML -> P;
      # End it and send it:
      $APP->end_page;
      return (1);
   }
   my $userAdmin = IU::UserAdmin->new;

   # Mode FORGOT assumes that the password is either being changed from a
   # guest session or an admin session.  That is why we use user object
   # "User2" instead of "User".
   my $user = $SESSION->state->{User2};

   # Retrieve the ECS name and password from the user object.
   my $persist_usr = $user->keys->ecs_username;
   my $pass1 = $user->keys->ecs_authentication_key;
   my $prefs_dir =${IU::PersistentDir}."imswww-usr/".$persist_usr;
   my $prefs_file = $prefs_dir."/prefs.dat";
   my $dont_install = 1;
   my $user_info = IW::ModePREFS::load_prefs($prefs_file, $dont_install);

   # If the account name is not valid show the fail page and the reason.
   if(!$userAdmin->exists($persist_usr)) {
       $APP->start_page(Title => "Password Not Changed");
       $HTML->t("$persist_usr doesn't exist");
   } else {
      debug "$persist_usr exists";
      # Go ahead and update the user password with the password that had
      # been passed in the session under User2.
      if($userAdmin->update($persist_usr, $pass1)) {
        my $keys = ($user_info->keys || IW::UserKeys->new);
        $keys->ecs_username($persist_usr);
        $keys->ecs_authentication_key($pass1);
        $user_info->install(To=>$prefs_dir);
	my $email = $userAdmin->lookup_email_from_session($persist_usr);
        # Send off the email.
        $self->send_done_mail(To       => $email,
                              Password => $pass1,
                              Account  => $persist_usr);
        # Say we did it.
        $self->show_okay("The account '$persist_usr' has a new " .
                         "password, which has been e-mailed to ".
			 $self->person("you","the user") . " at: " .
                         $email, $modal);
      } else {
        # The update failed because of permission problems.
	$HTML->t("Couldn't touch $persist_usr");
      }
   }
   delete $SESSION->state->{User2};

   $HTML -> script_FORM;
   $HTML -> hidden('mode'=>'FORGOT');
   $HTML -> BR -> INPUT(TYPE=>'submit', 
                        VALUE=>'Change another user\'s password');

   # End the form:
   $HTML -> _script_FORM;

   # End the page:
   $APP->end_page();

   1;
}

#------------------------------------------------------------
# submode_START
#------------------------------------------------------------
# instance method
#
# Show the page, turning off the navigator if requested.
# We don't use the person() method here, because other modes
# which inherit from this code will override this method or
# call something else instead.

sub submode_START {
   my $self = shift;
   my $modal = shift;
   $modal = 1 unless defined $modal;

   # Start the page:
   $APP->suppress_links($modal);
   $APP->start_page(Title=>"Get a New Password",
                    Context=>'new_passwd');

   my $directions = <<EOF;
<font SIZE="+1">
You can use this form to change your password and have the new password
e-mailed to you.
</FONT>
EOF

   my $notes = <<EOF;
<B>Note: This form is <I>not</i> for Data Access Key changes.<BR>Please contact
the appropriate User Services Office listed below to obtain a new Data Access
Key.
EOF

   my %sites = split /,/,  $IU::SitesWithKey;
   my $notes2 = "";
   foreach my $site (keys %sites) {
      $notes2 = "<ul>" unless $notes2;
      $notes2 .=<<EOF;
<li>$site - <a href="mailto:$sites{$site}">$sites{$site}</a></li>
EOF
   }
   $notes2 .="</ul>" if $notes2;

   $self->{Widget} = new IW::Widget::EmailOrAccount(
			SubHead    => "Forgotten your password?",
                        Directions => $directions,
                        Notes      => "$notes$notes2",
                        Modal      => $modal);
   $self->{Widget}->print;
   $APP->end_page;
}

#------------------------------------------------------------
# submode_LOOKUP
#------------------------------------------------------------
# instance method
#
# We do use person() here because this code will be shared among
# other classes that want to do the lookup and e-mail.

sub submode_LOOKUP {
   my $self = shift;

   # Set things up for the lookup. Use the default password file.
   my $admin = new IU::UserAdmin;

   # Recreate the widget from the information passed through the page.
   # Why? because we need the widget to call person(), and person()
   # expects a widget which knows what's going on!
 
   $self->{Widget} = IW::Widget::EmailOrAccount->excyst;

   # Determine whether we have an email address or a session name.
   # Try the valid session name check first.
   my $old_acct = $REQ->param('p.emailoracct');
   my $old_val  = $REQ->param('p.emailoracctv');
   my $modal    = $self->{Widget}->modality;
 
   local $_;
   foreach ($old_acct,$old_val) {
      s/^\s+//;   # Kill leading spaces
      s/\s+$//;   # Kill trailing spaces
   }

   # First validation: the two have a value
   if(!$old_acct || !$old_val) {
      $self -> show_error("You must supply a value for both of these fields.",
                          $modal);
      return 1;
   }
   # Second validation: the two have to match.
   if ($old_acct ne $old_val) {
       $self -> show_error("The values you entered do not match. ".
                           "Please check your input and correct it " .
                           "as necessary.",$modal);
      return 1;
   }

   # If it isn't a valid account name, we assume it was an email address.
   my $diagnosis;
   if ($diagnosis = $admin->invalid_name($old_acct)) {
      my @possibles = $admin->lookup_sessions_by_email($old_acct);
      if (int(@possibles)) { # at least one
         if (int(@possibles) == 1) { # in fact, only one!
            # Make up a new password and change the user's password.
	    my $persist_usr = $possibles[0];

	    # This was changed to get a random password instead of setting one.
            my $pwd = $admin->get_random_password($persist_usr);

	    # Create a new user object to store the ECS name and password.
            my $user = new IW::User;
            $user->keys(new IW::UserKeys);
            $user->keys->ecs_username($persist_usr);

	    # Set the new and old ECS password to be the same.
            $user->keys->ecs_authentication_key($pwd);
            $user->keys->old_ecs_authentication_key($pwd);

	    # This user object is stored as "User2" because the user is 
	    # either a guest or is using Admin mode.  This means that 
	    # the account password that is being changed is not the 
	    # same account as the current session.
            $SESSION->state->{User2} = $user;

	    # Only send to ECS if EcsProfileSmc is set in SETUP.
            if ($user && $IU::EcsProfileSmc) {      
               # Additional check for errors
               # was added to "send_ecs_change_req"
               # Send ECS profile change request.
               $self->send_ecs_change_req ($user); 
            } else {
	       # ECS is not involved, go ahead and change the password.
	       $self->submode_PASSWDFINAL;
	    }
         }
         else { # Nope, multiples.
            # Send off the list of possibles.
            $self->send_list_mail(To      => $old_acct,
                                  List    => \@possibles);
            # Say we did that.
            $self->show_info("The e-mail address you supplied matches " .
                             "multiple accounts. A list of those accounts " .
                             "has been e-mailed to the address you searched " .
                             "for.", $modal);
         }
      }
      else { # nobody home
         $self -> show_error("No accounts match that e-mail address. " .
                             "Please check your input.", $modal);
      }
   } else {
   # A valid session name. Treat it as such.
      my $email = $admin->lookup_email_from_session($old_acct);
      debug "session $old_acct belongs to $email";
      unless ($email) { # oops, no such account
          $self -> show_error("There is no account by that name. " .
                              "Please check your input.", $modal);
      } else { # okay, got it
         # Make up a new password and set a temporary new one
         my $pwd = $admin->get_random_password($old_acct);

	 # Create a new user object to store the ECS name and password.
         my $user = new IW::User;
         $user->keys(new IW::UserKeys);

         # Set the new and old ECS password to be the same because this
	 # is a password reset.
         $user->keys->ecs_username($old_acct);
         $user->keys->ecs_authentication_key($pwd);
         $user->keys->old_ecs_authentication_key($pwd);

         # This user object is stored as "User2" because the user is 
         # either a guest or is using Admin mode.  This means that 
         # the account password that is being changed is not the 
         # same account as the current session.
         $SESSION->state->{User2} = $user;
         # This was modified to send to the SMC instead of the home DAAC.
         # Only send to ECS if EcsProfileSmc is set in SETUP.
         if ($user && $IU::EcsProfileSmc) {      
            # Additional check for errors
            # was added to "send_ecs_change_req"
            # Send ECS profile change request.
            $self->send_ecs_change_req ($user); 
         } else {
            # ECS is not involved, go ahead and change the password.
	    $self->submode_PASSWDFINAL;
	 }
      }
   }
}

#------------------------------------------------------------
# go
#------------------------------------------------------------
# instance method
# Entry point.

sub go {
   my $self = shift;

   # Get session
   
   $APP->get_session('CREATE') ;
   
   if (!$REQ->param('sid')) {     # we had to create it.. use redirect trick!
      $REPLY->redirect(script_url('mode'=>'FORGOT',
                                  'submode'=>$REQ->param('submode')));
      return 1;
   }

   # Go!
   $self->go_submode;
}

#------------------------------------------------------------
1;
