package IW::ModeGRANLIST;

=head1 NAME

IW::ModeGRANLIST - granule listing

=head1 SORTING (FUTURE PLANS)

To allow sophisticated navigation, sorted-order will be maintained
I<in the results database>, as an array of dskeys in the current sorted order,
I<possibly with duplicates> when sorting on multiply-values keys.
    
    "dssort" => ["1/231", "4/345", "4/633", "1/231", "2/1021", "2/98"]
    
It is easy to construct the forward mapping (index to dataset), but there is
of course no inverse mapping (dataset to index).

Sorting of the current dataset objects (and re-storing of the new sorted
order) will be done whenever we are about to list.  Since a dataset list 
must always precede a request for dataset attributes, we're safe.

=cut

use IU;
use IU::Log;
use IU::InvStore ();
use IU::Dataset ();
use IU::SpecializedResults;

use IW::App;
use IW::Client;
use IW::CommStat;
use IW::Mode;
use IW::Util qw(:url :output);
use IW::GranList;
use IW::GranTime;
use IW::GranMap;


@ISA = qw(IW::Mode);

use strict;


#------------------------------------------------------------
#
# Globals...
#
#------------------------------------------------------------

# Inventory store:
my $RESULTS;
my $RESULTSLIST;
my $CART;
my $WORKBENCH;


#------------------------------------------------------------
# send_doc PARAMHASH
#------------------------------------------------------------
# Output the granule listing document.  Arguments are..
#
#    Granules    => a ref to an array of granules to list.

sub send_doc {

   my $map_lat = $REQ->param('map_lat');
   my $map_lon = $REQ->param('map_lon');
   my $msg;
   if ( ($map_lat < -90) || ($map_lat > 90) ) {
       $msg="Latitude must be between -90 and +90 degrees.";
   }
   if ( ($map_lon < -180) || ($map_lon > 180) ) {
       $msg = $msg . " Longitude must be between -180 and +180 degrees.";
   }
   if($msg) {
       $APP->message (new IW::Note Class=>'ERROR', Text => $msg);
   }


   # Log into RESULTS database:
   $RESULTS = IU::InvStore->open($SESSION->path_to('results.gdb'), 'WRCREAT')
      || puke "couldn't open results";

   # Log into RESULTSLIST database.
   $RESULTSLIST = IU::InvStore->open($SESSION->path_to('resultslist.gdb'), 'WRCREAT')
      || puke "couldn't open results list database"; 

   # Log into CART database (okay if not there):
   eval{
      $CART = undef;
      return unless -e $SESSION->path_to('cart.gdb');
      $CART = IU::InvStore->open($SESSION->path_to('cart.gdb'), 'READER')
   };

   # Log into WORKBENCH database (okay if not there):
   eval{
      $WORKBENCH = undef;
      return unless -e $SESSION->path_to('workbench.gdb');
      $WORKBENCH = IU::InvStore->open($SESSION->path_to('workbench.gdb'), 
                          'READER')
   };
   my $submode = ($REQ->param('submode') || 'LIST');
   my $select_flag = $REQ->param('select');
   my @dskeys = $REQ->param('dskey');



   # Variables used to keep track how many MISR/NonMISR granules are in the
   # Granule List.
   my $num_misr_gran = 0;
   my $num_non_misr_gran = 0;
   my $num_amsre_gran = 0;
   my $num_non_amsre_gran = 0;
   my $num_glas_gran = 0;
   my $num_non_glas_gran = 0;
   my $num_nose_gran = 0;
   my $num_non_nose_gran = 0;

   my @gkeys;
   if ($submode eq 'NEW') {
      @dskeys = IW::GranList->update_virtual_keys('dskey',$RESULTSLIST,\@dskeys);

      my @ds_gkeys =  $RESULTS->gkeys_for_datasets(\@dskeys);
      # Create new list of unselected granule keys:
      # Test if total number of granules exceeds the list limit.
      my $list_max = ${IU::ListMax} ||  1000;
      if (int(@ds_gkeys) > $list_max) {
      puke 513, "Exceeded $list_max maximum granule list limit, please ",
                "select fewer data sets for which to list granules.";
      }
      # Create new list of unselected granule keys:
      $RESULTSLIST->list_keys('gkey',\@ds_gkeys);
      # Clear selected granules.
      my @empty;
      $RESULTSLIST->selected_list_keys('gkey',\@empty);
      $SESSION->state->{'found_sensor_disclaimer'} = '0';
   } elsif ($submode eq 'SHOW') {
      # Clear selections if user enters page via navigtor link (submode SHOW). 
      my @empty;
      $RESULTSLIST->selected_list_keys('gkey',\@empty);
   } else {
      @gkeys = $REQ->param('gkey');
      @gkeys = IW::GranList->update_virtual_keys('gkey',$RESULTSLIST,\@gkeys);
   } 

   # This inflates all of the granules in the database.
   # This needs to be changed so it only inflates those granules
   # which are going to be displayed on the page, however need 
   # to be smarter about sorting (which needs all the granules). 

   my @allkeys = $RESULTSLIST->list_keys('gkey');
   my @all_granules = $RESULTS->granules(\@allkeys);
   my @selections = $RESULTSLIST->selected_list_keys('gkey');
   my $nextpage = $REQ->param('nextpage');
   my $page = $RESULTSLIST->page('gkey',$nextpage);
   my $all_granlist = IW::GranList->new(Elems=>\@all_granules,
                                        Origin=>'RESULTS',
                                        Type=>'gkey',
                                        Page=>$page,
                                        Selections=>\@selections);

   # Update each granule with it's cart and workbench status.
   my $gran;


   foreach $gran (@all_granules) {
        $gran->carted(1) 
           if ($CART && $CART->has_granules([$gran->key]));
        $gran->workbenched(1) 
           if ($WORKBENCH && $WORKBENCH->has_granules([$gran->key]));

        my ($is_misr, $isnt_misr, $is_amsre, $isnt_amsre, $is_glas, $isnt_glas,$is_nose, $isnt_nose)=  IU::SpecializedResults::are_special_granules($gran);
               

        # Number of MISR granules and NON MISR. 
        $num_misr_gran += $is_misr;
        $num_non_misr_gran += $isnt_misr;
        $num_amsre_gran += $is_amsre;
        $num_non_amsre_gran += $isnt_amsre;
        $num_glas_gran += $is_glas;
        $num_non_glas_gran += $isnt_glas;
        $num_nose_gran += $is_nose;
        $num_non_nose_gran += $isnt_nose;

   }  
   my $sort = '';
   $all_granlist->sort(\$sort);

   # Store sorted granules keys for granule attributes step through.
   my $sorted_gkeys = [map {$_->key}  @{$all_granlist->elems}];
   $RESULTS->sorted_gkeys($sorted_gkeys);

   # Handle text report mode:

   if ($REQ->param('report') eq 'Y') {

      $all_granlist->print_as_text;
      return 1;
   }

   # Set the navigator indicator.
   $SESSION->{Navigator}->{Indicator} = "ViewGranule";

   # start the page and form:

   if ($submode eq 'MAP' || $submode eq 'MAPTOGGLE_ON' || $submode eq 'MAPTOGGLE_OFF' ) {

      $APP->start_page(SuperTitle => "Results: Granule",
                       Title=> "Spatial Coverage" ,
                       Context =>'granmap');

   } elsif ($submode eq 'TIME') {

      $APP->start_page(SuperTitle => "Results: Granule",
                   Title=>"Temporal Coverage",
                   Context =>'grantime');
   } else {

      $APP->start_page(SuperTitle => "Results: Granule",
                   Title=>"Listing",
                   Context =>'granlist');
   }


   # If there is only ONE Dataset, the user is being shown the Granule list
   # instead. So give them some idea what is happening to them.
   if ( $REQ->param('AutoDisplay')) {
        $HTML -> directions
              -> t( 'You are automatically being shown the Granule list since there was only one data set returned.') 
              -> _directions -> BR; 
   }

   # Get client so we can display a summary of the last search:
   my $client = $SESSION->state->{LastsearchClient};
   # Do poll to update the status in the client
   if(defined($client)) {
      $client->poll;
      my $commstat = IW::CommStat->new(Client=>$client);
      $commstat->as_HTML;
   }


   $HTML -> script_FORM;                      # begin form
   $HTML -> msubmit_flag;




   # Forward the Number of MISR Granules to GranMap so it knows to print 
   # the toggle button.
   $HTML -> hidden('NumberMISR', $num_misr_gran);
   $HTML -> hidden('NumberAMSRE', $num_amsre_gran);
   $HTML -> hidden('NumberGLAS', $num_glas_gran);
   $HTML -> hidden('NumberNOSE', $num_nose_gran);
   $HTML -> BR;


   # Output the menu:
   # If there are granules then display the function buttons.

   if ( int(@allkeys) > 0 ) {
      # Show Buttons:
      $HTML -> cartadd_menu_btn;
      $HTML -> granlistmap_menu_btn;
      $HTML -> granlisttime_menu_btn;
      $HTML -> workbenchadd_menu_btn;
      $HTML -> t(" ");
      $HTML -> select_one(Name    => "selected_on",
                        Choices =>['items selected on all pages',
                                   'items selected on this page only']);

      # Only show the list only button if showing the spatial (MAP) 
      # or temporal (TIME) coverage.

      if ($submode eq 'MAP' or $submode eq 'MAPTOGGLE_ON' or $submode eq 'MAPTOGGLE_OFF' or $submode eq 'TIME') {
         $HTML -> BR;
         $HTML -> granlist_menu_btn;
         $HTML -> t(' without MAP or TIME coverage graphic.');
      }
   }
   $HTML -> BR;

   # Output spatial (MAP) or temporal (TIME) coverage.

   if ($submode eq 'MAP' or $submode eq 'MAPTOGGLE_ON' or $submode eq 'MAPTOGGLE_OFF') {
       
      my $mapmode = $REQ->param('mapmode');
 
      # MAPTOGGLE is executed when the User wants the labels turned ON/OFF.
      if ($submode eq 'MAPTOGGLE_ON')  {
               $SESSION->state->{displayLabels}='yes';
      }
      elsif ($submode eq 'MAPTOGGLE_OFF')  {
               $SESSION->state->{displayLabels} = 'no';
      }
      elsif ($mapmode =~ /^STAMPCLICK-(\d+)$/)  {
       # no-op
      }
      elsif ($mapmode eq 'MAPCLICK') {
        # no-op
      }
      # We only have MISR granules, Labels ON
      elsif ( (!defined($SESSION->state->{displayLabels})) && 
              ($num_misr_gran > 0 || $num_amsre_gran > 0 || $num_glas_gran > 0 || $num_nose_gran > 0 ))
             { $SESSION->state->{displayLabels} = 'yes';}
      # No MISR granules, Labels OFF, Daniel-son.
      elsif ( (!defined($SESSION->state->{displayLabels})) &&  
        ($num_non_misr_gran > 0 || $num_non_amsre_gran > 0 || $num_non_glas_gran > 0 || $num_non_nose_gran > 0))
               { $SESSION->state->{displayLabels} = 'no'};

      my @selected_granules = $all_granlist->granules_to_display();

#  Limit was removed at the request of ASF
#      # Limit the number of data granules which can be displayed on the   
#      # coverage map.
#      if (int(@selected_granules) > 20) {
#         print "<H2> Too many data granules selected for display on map, ";
#         print "please limit the number of selections to 20 or less. </H2>";
#      } elsif(@selected_granules) {


      my $no_MAPPED_granules_selected  = 0;
      my $temp_var;

      # How many of the SELECTED granules are MISR?
      foreach  $temp_var ( @selected_granules) {
     my ($is_misr, $isnt_misr, $is_amsre, $isnt_amsre, $is_glas, $isnt_glas,$is_nose, $isnt_nose) =  IU::SpecializedResults::are_special_granules($temp_var);
           $no_MAPPED_granules_selected  += $is_misr+$is_amsre+$is_glas+$is_nose;
      }

      # If there are more then 15 MISR granules, we don't want to overload
      # covermap, so ask the user to select less then 15.
      my $misr_limit = ( $IU::SETUP::MisrLimit || 15 ); 

      if ($no_MAPPED_granules_selected > $misr_limit ) {
         print "<H2> Too many MISR/AMSRE/GLAS granules selected for display on map, ";
         print "please limit the number of selections to ".($misr_limit)." or less. </H2>";
      } elsif(@selected_granules) {
         IW::GranMap->display_map(\@selected_granules)
            or puke "couldn't send map!";
      } else {
         print "<H2> No data granules selected for display on map.</H2>";
      }

   } elsif ($submode eq 'TIME') {

      my @selected_granules = $all_granlist->granules_to_display();
      if (@selected_granules) {
         IW::GranTime->display_time(\@selected_granules)
            or puke "couldn't send time display!";

      } else {
         print "<hr><H2> No data granules selected ";
         print "for display on temporal graph!</H2>";
      }
   }

   if (@allkeys) {
      $all_granlist->print;
   } else {
      print "<hr><H2> No data granules to list! </H2>";
      $HTML->UL
           ->LI->t("Use the ") 
               ->t_B("Results:Data Set") 
               ->t(" navigation link to go to the Results: Data Set page.")
           ->LI->t("Select some data sets.")
           ->LI->t("Then use the \"List Data Granules\" button to ",
                   "display them on the Data Granule Listing Page. ")
           ->_UL;
   }

   # Finish the page:

#   $HTML -> hidden('origin', 'RESULTS');    # this is our origin

   # Thought I could use the ORIGIN tag here, but I can't. 
   # Created a new hidden tag to handle the PREVIOUS Mode problem
   # that can happen if the user CANCEL's a DISCLAIMER notice.
   $HTML -> hidden('prevmode', 'GRANLIST');    

   $HTML -> _script_FORM;                     # end form
   $APP->end_page;
   1;
}

#------------------------------------------------------------
# misc_url_order
#------------------------------------------------------------
# Print a page explaining why can't order granule with link
# to user services at data center.
sub misc_url_order {

    # Set the navigator indicator.
    $SESSION->{Navigator}->{Indicator} = "ViewGranule";

    $APP->start_page(SuperTitle => "Data Granule",
                     Title=> "Cannot Order" ,
		     Context =>'cant_order_gran');
    $HTML -> HR
	  -> t("The data granule: ", $REQ->param('granule'))
	  -> BR
	  -> t("from the data set: ", $REQ->param('dataset'))
	  -> BR
	  -> t("at the data center: ", $REQ->param('daac'))
	  -> BR
	  -> t("cannot be ordered through the EOS Data Gateway.")
	  -> BR
	  -> P
          -> t("To order this granule directly from the data provider, use \
the Special Processing Links on the Results: Granule page.");

    $APP->suppress_key(1);
    $APP->end_page();    

    1;
}

#------------------------------------------------------------
# cant_order
#------------------------------------------------------------
# Print a page explaining why can't order granule with link
# to user services at data center.
sub cant_order {
    my $daac = $REQ->param('daac');
    my $href = "${IU::ImsUserUrl}#$daac";

    # Set the navigator indicator.
    $SESSION->{Navigator}->{Indicator} = "ViewGranule";

    $APP->start_page(SuperTitle => "Data Granule",
                     Title=> "Cannot Order" ,
		     Context =>'cant_order_gran');
    $HTML -> HR
	  -> t("The data granule: ", $REQ->param('granule'))
	  -> BR
	  -> t("from the data set: ", $REQ->param('dataset'))
	  -> BR
	  -> t("at the data center: ", $daac)
	  -> BR
	  -> t("cannot be ordered through this system.")
	  -> t("This may be because the data center hasn't supplied\
packaging information or doesn't have a server that can process\
orders from this system.")
	  -> BR
          -> t("You may be able to order it directly from the data center.  \
Try contacting the user services group at the data center.")
	  -> P;
    print "\n<center>";
    $HTML -> A(HREF=>$href)
	  -> t("Here is their contact information.")
	  -> _A;
    print "</center>\n";
    $HTML ->P;

    $APP->suppress_key(1);
    $APP->end_page();    

    1;
}

sub nav_status {
         return ("Message" => "",
                     "Link"    => "",
                     "Mode"    => "GRANLIST",
                     "Submode" => "SHOW",
         );
}



#------------------------------------------------------------
# go
#------------------------------------------------------------
# Main entry point.

sub go {
    my $self = shift;
    
    # Get session or die:
    $APP->get_session();

#    my $endform = $REQ->param('endform');
#    debug "endform = $endform";

    $REQ->param('endform') ||
        puke 543, "The Data Granule Listing did not receive from the previous ",
	          "page all the information it needs.";

    # What do we do?
    my $submode = ($REQ->param('submode') || 'LIST');
  SWITCH: for ($submode) {
      /LIST/ and do {
	  send_doc()       or puke "couldn't send list";
	  last;
      };
      /SHOW/ and do {
          send_doc()       or puke "couldn't send list";
          last;
      };
      /NEW/ and do {
          send_doc()       or puke "couldn't send list";
          last;
      };
      /MAP/ and do {
          send_doc()       or puke "couldn't send list";
          last;
      };
      /MAPTOGGLE_ON/ and do {
          send_doc()       or puke "couldn't send list";
          last;
      };
      /MAPTOGGLE_OFF/ and do {
          send_doc()       or puke "couldn't send list";
          last;
      };
      /TIME/ and do {
          send_doc()       or puke "couldn't send list";
          last;
      };
      /CANT_ORDER/ and do {
          cant_order()       or puke "couldn't send list";
          last;
      };
      /MISCURL_ORDER/ and do {
          misc_url_order()       or puke "couldn't send list";
          last;
      };
      puke 600, "bad submode: $submode!";
  }    
    1;
}

#------------------------------------------------------------
1;
