#------------------------------------------------------------
# ModeGUIDE.pm
#------------------------------------------------------------
package IW::ModeGUIDE;

=head1 NAME

IW::ModeGUIDE - guide search unifier.

=head1 DESCRIPTION

Unifies the XML and WAIS-based guide searches into one (supposedly)
seamless entity.

This module presents the same external interface to all callers, but
internally dispatches searches to the proper sources. C<wwwvalids.conf>
is used to determine which DAACs are using WAIS and which Spidex (XML).

This module calls either/both as necessary and combines their output
for presentation.

=head1 AUTHOR

This code was written by Joe McMahon.

=head1 RCS INFO

$Id: ModeGUIDE.pm,v 1.53 2003/06/11 19:39:48 joe Exp $

=cut

use IU;
use IU::Log;
use IW::App;
use IW::Util qw(:url :output);

use IW::ModeGUIDEOLD;
use IW::ModeGUIDEXML;

use strict;
use IU::Storable;
use IU::Spawn;

@IW::ModeGUIDE::ISA = qw(IW::Mode);

#--------------------------------------------------------------
# Class methods
#--------------------------------------------------------------
# new()
# This constructs an object which holds references to objects for
# the two possible guide types. 
sub new {
  my ($class) = @_;
  my $self;
  local $_;
  foreach ($IU::GuideType) {
    # If TRANSITIONAL, create an object that can do both kinds of search.
    /TRANSITIONAL/i and do {
     $self = {};
     $self->{Wais} = IW::ModeGUIDEOLD->new;
     $self->{XML} = IW::ModeGUIDEXML->new;
     bless $self,$class;
     };
    /NEW|OLD|BOTH/i and do {
     $self = IW::ModeGUIDEOLD->new;
     };
    /XML/i and do {
     $self = IW::ModeGUIDEXML->new;
     };
  }
  $self;
}

#-------------------------------------------------------------
# shared instance methods
#------------------------------------------------------------

# results_file
#------------------------------------------------------------
# Save results file name in object.
sub results_file {
    my ($self,$file) = @_;
    $self->{ResultsFile} = $file if defined $file;
    $self->{ResultsFile};
}

# andiamo
#------------------------------------------------------------
# "let's go."
#
# Record current status and/or error message info for this subprocess.
# Puke with the error message if there was one.

sub andiamo {
   my ($self,$msg,$hashref) = @_;
   my $file;
   if (defined ($file = $self->results_file())) {
      $hashref->{Error} = $msg if $msg;
      if (open(RESULTS,">$file")) {
         my $frozen = IU::Storable->flatten($hashref);
         print RESULTS $frozen;
         close RESULTS;
         $self->process_debug("subprocess results written");
         puke $msg if $msg;
      }
      else {
         $self->process_debug("Cannot open subprocess results file $file: $!");
      }
   }
   else {
      $self->process_debug("Process has no results_file, but andiamo called");
   } 
}

# suppress_relevance
#------------------------------------------------------------
# Whether or not to show relevance in output.
sub suppress_relevance {
    my ($self,$which) = @_;
    $self->{SuppressRelevance} = $which if defined $which;
    $self->{SuppressRelevance};
}

# suppress_daac
#------------------------------------------------------------
# Whether or not to show daac in output.
sub suppress_daac {
    my ($self,$which) = @_;
    $self->{SuppressDaac} = $which if defined $which;
    $self->{SuppressDaac};
}

# process_log
#------------------------------------------------------------
# set/get the name of the process log.
sub process_log {
   my ($self,$file) = @_;
   $self->{LogFile} = $file if defined $file;
   $self->{LogFile};
}

# process_debug
#------------------------------------------------------------
# Print the message to the debug log.
sub process_debug {
   my ($self,@msg) = @_;
   # Flock stuff:
   my %LOCK = (SH=>1, EX=>2, NB=>4, UN=>8);

   my $msg = join("",@msg,"\n");
   if (defined $self->{LogFile}) {
     my $lf = $self->{LogFile};
     if (open PLOG,">>$lf") {
         if (flock(PLOG, $LOCK{EX})) {
             seek(PLOG, 0, 2); # in case someone appended during the wait
             print PLOG $msg;
         }
         else {
             print STDERR "Can't flock $lf: $!\n";
         }
     }
     else {
        print STDERR "Can't open process log $lf: $!\n";
     }
   }
}
#------------------------------------------------------------
# unshared instance methods
#------------------------------------------------------------

# discard_history
#-------------------------------------------------------------
# Throw away any previously-gotten hits and process states.
sub discard_history {
  my $self = shift;
  my $history = $SESSION->path_to() . 'guide_history';
  unlink $history;
}

# preserve_history
#-------------------------------------------------------------
# Save the current ModeGUIDE object in a file. Used to restore
# the hits for other calls to ModeGUIDE for the same query.
sub preserve_history {
  my $self = shift;
  my $chilly = IU::Storable->flatten($self);
  my $history = $SESSION->path_to() . 'guide_history';
  open HISTORY,">$history" or puke "Can't save history: $!";
  print HISTORY $chilly;
  close HISTORY; 
  chmod 0600, $history;
}

# restore_history
#-------------------------------------------------------------
# Restore the old ModeGUIDE object and extract the last set of
# history from it. 
sub restore_history {
  my $self = shift;
  my $history = $SESSION->path_to() . 'guide_history';
  open HISTORY,"$history" or puke "Can't load history: $!";
  my @chilly = <HISTORY>;
  my $chilly = join("",@chilly);
  my $restore = IU::Storable->inflate(\$chilly);
  close HISTORY; 
  $self->{SearchStart} = $restore->{SearchStart};
  $self->{Serial} = $restore->{Serial};
  $self->{WaisStat} = $SESSION->path_to() . $self->{Serial} . "w_status";
  $self->{XMLStat}  = $SESSION->path_to() . $self->{Serial} . "x_status";
}

# load_results
#---------------------------------------------------------------------
# Called when subprocesses have quit running.
#
# Reads the results files that the processes have written and stuffs them
# into the current object, then shoves that into the history file.
sub load_results {
   my $self = shift;

   # Load wais results.
   # debug "loading WAIS results";
   open WAIS_RESULTS,$SESSION->path_to() . $self->{Serial} . "w_results"
      or $self->{WaisFailed} = 1;
   unless ($self->{WaisFailed}) {
     my @chilly = <WAIS_RESULTS>;
     my $chilly = join("",@chilly);
     my $restore = IU::Storable->inflate(\$chilly);
     close WAIS_RESULTS;
     $self->{WaisQ} = $restore->{Query};
     $self->{WaisQFile} = $restore->{Queryfile};
     $self->{WaisP} = $restore->{Display};
     $self->{WaisHits} = $restore->{Hits};
     $self->{WaisCats} = $restore->{Cats};
     $self->{WaisMsg} = $restore->{Error};
     # debug "Got wais results '$restore->{Query}' '$restore->{Print}'";
   }
   else {
     debug "WAIS results unavailable";
     $self->{WaisQ} = "";
     $self->{WaisQFile} = "";
     $self->{WaisP} = "";
     $self->{WaisHits} = [];
     $self->{WaisCats} = [];
     $self->{WaisMsg} = "WAIS results unavailable";
   }

   # Load XML results.
   # debug "loading XML results";
   open XML_RESULTS,$SESSION->path_to() . $self->{Serial} . "x_results"
      or $self->{XMLFailed} = 1;
   unless ($self->{XMLFailed}) {
     my @chilly = <XML_RESULTS>;
     my $chilly = join("",@chilly);
     my $restore = IU::Storable->inflate(\$chilly);
     close XML_RESULTS;
     $self->{XMLQ} = $restore->{Query};
     $self->{XMLP} = $restore->{Print};
     $self->{XMLHits} = $restore->{Hits};
     $self->{XMLCats} = $restore->{Cats};
     $self->{XMLMsg} = $restore->{Error};
     # debug "Got XML results '$restore->{Query}' '$restore->{Print}'";
   }
   else {
     debug "XML results unavailable";
     $self->{XMLQ} = "";
     $self->{XMLP} = "";
     $self->{XMLHits} = [];
     $self->{XMLCats} = [];
     $self->{XMLMsg} = "XML results unavailable";
   }

   # Save all this.
   $self->preserve_history;
}

# debug_structure REF
#--------------------------------
# Dump out the contents of whatever is pointed to by
# $ref using Data::Dumper.
#
#sub debug_structure {
#   my $ref = shift;
#   local $_;
#   foreach (split(/\n/,Dumper($ref))) {debug $_}
#}

# reconcile_hits
#-------------------------------------------------------------
# reconcile_hits extracts the two sets of hits from the ModeBROWSE
# object and:
# 1) Extracts all "error" hits into a separate list
# 2) Reconciles all the scores to a single scale
# 3) Rescales all the scores to a max of 100, no matter what the 
#    highest score really was.
sub reconcile_hits {
  local $_; 
  my $self = shift;  debug "XML hits: ",    ((ref($self->{XMLHits}) and 
                           int(@{$self->{XMLHits}})) or "(missing)"),
        ", XML errors: ", ((ref($self->{XMLErrs}) and
                           int(@{$self->{XMLErrs}})) or "(missing)"),
        ", XML cats: ",  ((ref($self->{XMLCats}) and 
                           int(@{$self->{XMLCats}})) or "(missing)"),
        ", XML msg: ",  ($self->{XMLMsg} or "(missing)"),
        ", WAIS hits: ", ((ref($self->{WaisHits}) and 
                           int(@{$self->{WaisHits}})) or "(missing)"),
        ", WAIS cats: ", ((ref($self->{WaisCats}) and 
                           int(@{$self->{WaisCats}})) or "(missing)"),
        ", WAIS msg: ", ($self->{XMLMsg} or "(missing)");

  $self->{XMLErrs} = [];
  my @hits = ();
  my @errs = ();
  my $high = 0;
  foreach (@{$self->{XMLHits}}) {
    # Discard 'no results' errors.
    next if $_->{'code'} eq '202';
    # Keep good hits.
    if ($_->{HitType} eq 'doc' and $_->{'status'} eq 'OK') { 
      $_->{HitType} = "XML";
      push @hits, $_;
    }
    # May be an error we want to keep.
    else {
      next unless $_->{'source'};
      next if $_->{'msg'} eq 'Record not found.';
      push @errs, $_;
    }
    $high = $high>$_->{'score'} ? $high : $_->{'score'};
  }

  # rescale scores.
  foreach (@{$self->{XMLHits}}) {
     $_->{'score'} += (100-$high);
  }

  $self->{XMLHits} = \@hits;
  $self->{XMLErrs} = \@errs;
  
  # debug "Errors found: " . int( @{$self->{XMLErrs}});

  $high = 0;  
  foreach (@{$self->{WaisHits}}) {
    $_->{HitType} = "Wais"; 
    # Rescale the WAIS scores to 0-100.
    $_->{'score'} = int($_->{Score}/10);
    $high = $high>$_->{'score'} ? $high : $_->{'score'};
  }

  # rescale scores.
  foreach (@{$self->{WaisHits}}) {
     $_->{'score'} += (100-$high);
  }

  debug "XML hits: ",    ((ref($self->{XMLHits}) and 
                           int(@{$self->{XMLHits}})) or "(missing)"),
        ", XML errors: ", ((ref($self->{XMLErrs}) and
                           int(@{$self->{XMLErrs}})) or "(missing)"),
        ", XML cats: ",  ((ref($self->{XMLCats}) and 
                           int(@{$self->{XMLCats}})) or "(missing)"),
        ", WAIS hits: ", ((ref($self->{WaisHits}) and 
                           int(@{$self->{WaisHits}})) or "(missing)"),
        ", WAIS cats: ", ((ref($self->{WaisCats}) and 
                           int(@{$self->{WaisCats}})) or "(missing)");
}

# finish_display()
#----------------------------------------------------------------------------
# Shows the "pretty" query to the user.
#
# finish_display() prints a link to the raw XML page and shows the query.
sub finish_display {
    my ($self) = @_;
    if ($APP->debug_mode()){
       # Output links to display the subprocess logs and the raw XML.
       $HTML->P->SMALL
            ->hypertext('View subprocess logs',
                        script_url('mode'=>'GUIDE','submode'=>'SHOWLOG'),
                        TARGET =>'support')
            ->_SMALL->BR; 
       $self->{XML}->raw_xml_link;

       # Show any diagnostics from the subprocesses.
       if ($self->{WaisMsg}) {
         $HTML->H2->t("WAIS Errors")->_H2
              ->P(STYLE=>"margin-left:5%")->t($self->{WaisMsg})->_P;
       }

       my $xml_head_out = 0;
       if ($self->{XMLMsg} or $self->{XMLErrs}) {
         if ($self->{XMLMsg}) {
           unless ($xml_head_out) {
             $HTML->H2->t("XML Errors")->_H2
               ->P(STYLE=>"margin-left:5%");
             $xml_head_out++;
           }
           $HTML->t($self->{XMLMsg})->P;
         }

         if (ref($self->{XMLErrs}) eq 'ARRAY' and @{$self->{XMLErrs}} > 0) {
           unless ($xml_head_out) {
             $HTML->H2->t("XML Errors")->_H2
               ->P(STYLE=>"margin-left:5%");
             $xml_head_out++;
           }
           local $_;
           $HTML->UL;
           foreach (@{$self->{XMLErrs}}) {
             $HTML->LI->t($_->{'msg'})->t(" (Datacenter: $_->{'source'})");
           }
           $HTML->_UL;
         }
         $HTML->_P if $xml_head_out;
       }
     }
    # Show the query.
    if ($self->{WaisP}) {
       $HTML->P->t("Your query was:")->UL->LI->t($self->{WaisP})->_UL;
    }
    $APP->end_page;
  }

# status_page
#---------------------------------------------------------------------------
# output the search status page.

sub status_page {
  my ($self,$wais,$xml) = @_;
  $APP->suppress_links(1);
  $APP->start_page(Supertitle => 'Detailed document search',
                   Title=>'Search in progress',
                   Refresh => {Wait=>comm_wait()*5,
                               URL=>script_url('mode'=>'GUIDE', 
                                               'submode'=>'SHOWSTATE')});
  my $elapsed = time - $self->{SearchStart};
  my $how_long  = sprintf("%02dm %02ds", 
                          ($elapsed/60),
                          ($elapsed%60));
  $HTML->H1->t("Detailed document search running")->_H1;
  $HTML->t("Your search is now running. ($how_long)")->P;
  $HTML->P(STYLE=>"margin-left:5%")->B->t("Note: ")->_B
                      ->t("You may not see this page update if your " .
                          "connection is slow, or if you are using Internet " .
                          "Explorer.")
       ->_P; 
  if ($APP->debug_mode()) {
     $HTML->P->t("The WAIS search is " . 
                 ($wais->running ? "running" : "complete"))
             ->t(", and the XML search is "  . 
                 ($xml->running  ? "running" : "complete") .".");
  } 
  $APP->end_page;
}

# comm_wait
#------------------------------------------------------------
# How long to wait for status update?

sub comm_wait {
   my $refresh_rate;
   defined ($refresh_rate = eval {$SESSION->state->{RefreshRate}}) or 
      $IU::UpdateEvery;
}

# there_are_docs HITSREF
#------------------------------------------------------------------------
# Handles figuring out what to say about how many docs were returned,
# limits them if necessary, and lets the caller know how many there are.
#
# HITSREF is a *reference* to a list of hits, assumed to be presorted in 
# order of importance (we throw away the less-important stuff).
# A reference is required because we want to actually modify the caller's copy.
#
# Returns the document count of non-catalog, non-error documents after limiting
# (if any).
sub there_are_docs {
  my ($self,$hits) = @_;
  # limit the number of hits based on the guide limit.
  my $limit = $SESSION->state->{Query}->max_guides || $IU::GuideLimit || 100;

  # Determine how many non-catalog, non-error documents there are.
  my $total = 0; 
  local $_;
  foreach (@$hits) {
    next if $_->{HitType} eq 'err' or 
            $_->{HitType} eq 'cat' or
            $_->{'code'} eq '202';
    $total++;
  }

  # Say what's up.
  my $msg;
  if ($total > $limit) {
    # Cut them down to the limit.
    my @trim = @$hits;
    @$hits = @trim[0..($limit-1)];
    my $href = script_url('mode'    => 'MAINSRCH',
                          'SrchTypeevent'  =>'TOGGLE',
                          'SrchTypenewtype' =>'GUIDE_SEARCH');
    $msg = <<EOF;
At least <b><big>$limit</big></b> documents matched your query; we have discarded the least-relevant documents beyond the first $limit.

<ul>\n<li><i><b>

There may be more documents available, but the current document limit is set to $limit.  You can increase this limit by going to 
<A HREF="$href">Additional Options</A>
(on the detailed document search form).  After you change
the limit you can resubmit your detailed document search.
</b></i></li></ul><P>
EOF
  }
  elsif ($total > 1) {
    $msg = "<b><big>$total</big></b> documents matched your query.";
  }
  elsif ($total == 0) {
    $msg = "<b><big>No</big></b> documents matched your query.\n<P>";
  }
  else {
    $msg = "<b><BIG>1</big></b> document matched your query.\n<P>";
  }  
  $self->{DocLimitMsg} = $msg;
  return $total;
}

# show_errors_if_any
#--------------------------------------------------------------
# Display any error messages.
sub show_errors_if_any {
   my $self = shift;
   my $doc_count = shift;
   my $were_errors = 0;

   # Errors from everybody.
   $were_errors++ if $self->{WaisMsg};
   debug "wais msg $self->{WaisMsg}" if $self->{WaisMsg};
   $were_errors++ if $self->{XMLMsg} or 
                    ($self->{XMLErrs} and 
                     ref($self->{XMLErrs}) eq "ARRAY" and 
                     @{$self->{XMLErrs}});
   local $_;
   my $msg = "";
   foreach ($were_errors) {
      /1/ and do {
         $msg = "Some of the datacenters could not be reached; " .
                "there may be other documents that match your query " .
                "at those datcenters.";
         $msg .= "The ones found are shown." if $doc_count;
         next;
      };
      /2/ and do {
         $msg = "None of the datacenters could be searched; " .
                "the document catalogs are shown below.";
         next;
      };
   }
   $APP->message(IW::Note->new(Class=>'WARN',Text=> "$msg")) if $were_errors;
   $were_errors;
}

# show_cats
#--------------------------------------------------------------
# show document catalog lists.
sub show_cats {
   my $self = shift;
   $self->{XML}->suppress_relevance(1);
   $self->{Wais}->suppress_relevance(1);

   $HTML->B->t("Document catalogs:")->_B->UL;
   if ($self->{XMLCats}) {
      my @cats = @{$self->{XMLCats}};
      @cats = map { $_->[1] }
              sort { $a->[0] cmp $b->[0] } 
              map { [$_->{'title'}, $_] }
              @cats;
      my $hit;
      foreach $hit (@cats) {
        $self->{XML}->OutputDoc($hit, undef, $hit->{DaacFile});
      }
   }
   
   if ($self->{WaisCats}) {
      my @cats = @{$self->{WaisCats}};
      @cats = map { $_->[1] }
              sort { $a->[0] cmp $b->[0] } 
              map { [$_->{'Headline'}, $_] }
              @cats;
      foreach my $hit (@cats) {
        $self->{Wais}->OutputDoc($hit, undef, $self->{DaacFile});
      }
   }
   $HTML->_UL;
}

#-------------------------------------------------------------
# submodes
#-------------------------------------------------------------

# submode_GUIDESRCH
#-------------------------------------------------------------
# Run the searches in parallel and start up the process monitoring.
sub submode_GUIDESRCH {
  my $self = shift;
  
  # Discard any old guide search results.
  $self->discard_history();
  
  # Create a per-search unique number.
  $self->{Serial} = (int(time) % 1000000);     

  # Set the results file names for both processes.
  $self->{Wais}->results_file($SESSION->path_to() .  
                              $self->{Serial} . "w_results");
  $self->{XML}->results_file($SESSION->path_to() .  
                              $self->{Serial} . "x_results");

  # Set up process logs for the subprocesses.
  $self->{Wais}->process_log($SESSION->path_to() .  
                              $self->{Serial} . "w_log");
  $self->{XML}->process_log($SESSION->path_to() .  
                              $self->{Serial} . "x_log");

  # Tell the WAIS side which DAACs to skip (all of the XML DAACs).
  $self->{Wais}->skip_daacs($self->{XML}->extract_daacs());

  # Create closures to be used as references to the proper method calls.
  # Needed because IU::Spawn insists that it get a reference to a sub 
  # to indicate what it should spawn, and this is the only correct way to
  # do that.
  #  \$self->foo       gives you a reference to the result, not the sub.
  #  $obj->can("foo")  gives you a reference to the sub, but loses the object
  #                    that needs to be associated with it.
  #
  # The sub{$obj->method()} business creates an anonymous sub which carries
  # the object and the method both along with it. Even if $obj goes out of
  # scope, this anonymous sub still has an implicit reference to it and
  # can call the method with the proper $self.
  my $waisref = sub {$self->{Wais}->submode_GUIDESRCH()};
  my $xmlref  = sub {$self->{XML}->submode_GUIDESRCH()};

  # Fire off both search types as subprocesses. This way, guide searches
  # won't time out if things are slow. 
  #
  # The subprocesses use IU::Storable and dump their results into
  # files called "${serial}w_results" and "${serial}x_results", respectively.
  IU::Spawn->start(Lockfile=>$SESSION->path_to() . 
                             $self->{Serial} . "w_status",
                   Action=>$waisref);

  IU::Spawn->start(Lockfile=>$SESSION->path_to() . 
                             $self->{Serial} . "x_status",
                   Action=>$xmlref);

  # Record the current lock files in the ModeGUIDE object.
  $self->{WaisStat} = $SESSION->path_to() . $self->{Serial} . "w_status";
  $self->{XMLStat}  = $SESSION->path_to() . $self->{Serial} . "x_status";

  # Record the start time.
  $self->{SearchStart} = time;

  # Remember all this by storing the ModeGUIDE object.
  $self->preserve_history;

  # Start checking the status.
  $REPLY->redirect(script_url('mode'=>'GUIDE',
                              'submode'=>'SHOWSTATE'));
  return 1;
}

# submode SHOWSTATE
#------------------------------------------------------------------------
# Load up last state of search, update the status,and then check to see
# if everything's done. If so, go display the results. Otherwise, output
# the self-updating page that will bring us back here again.
#
# I'm not doing partial results; this is fairly fast, but can sometimes 
# be slow enough that the connection times out before the searches finish.
sub submode_SHOWSTATE {
  my $self = shift;

  # Restore the appropriate fields from the stored ModeGUIDE object.
  $self->restore_history;

  # All we care about is the status of the subprocesses. Whether or not they
  # worked is the display routine's problem.
  #
  # Hook up to the subprocesses.
  my $wais = attach IU::Spawn Lockfile=>$self->{WaisStat};
  my $xml  = attach IU::Spawn Lockfile=>$self->{XMLStat};

  # If things are done, go show them.
  if (!$wais->running && !$xml->running) {
    $self->submode_SHOWBYRELEVANCE();
  }

  # Else not done; output the self-updating status page.
  else {
     $self->status_page($wais,$xml);
  }
  1;
}

# submode SHOWBYDAAC
#---------------------------------------------------------------------
# Sort the results by DAAC and then by relevance within DAAC.
sub submode_SHOWBYDAAC {
  my $self = shift;
  $self->restore_history;

  # If someone was attempting to be clever, send them back to the 
  # status page until everything's done.
  my $wais = IU::Spawn->attach(Lockfile=>$self->{WaisStat});
  my $xml  = IU::Spawn->attach(Lockfile=>$self->{XMLStat});
  $self->submode_SHOWSTATE if $wais->running or $xml->running;

  # All done. Let's show it.
  $self->load_results;
  $self->reconcile_hits;

  # This is by DAAC, so don't show DAAC names in each list element.
  $self->{XML}->suppress_daac(1);
  $self->{Wais}->suppress_daac(1);

  # And they don't want relevance, so suppress that too.
  $self->{XML}->suppress_relevance(1);
  $self->{Wais}->suppress_relevance(1);

  # Tell the WAIS stuff to suppress the links (makes the documents go
  # out-of-line).
  $self->{Wais}->suppress_links(1);

  # Sort hits into descending order. Schwartzian transform.
  #
  # from bottom to top:
  #  - create a list of all the hits.
  #  - create a new array of pairs of entries; the score and the object.
  #  - sort this array based on the scores in each element.
  #  - create yet another array by extracting just the objects out of the 
  #    sorted list.
  # Voila - a list of objects, sorted by score.
  my @hits = map {$_->[1]}
             sort {$b->[0] <=> $a->[0]}
             map {[$_->{'score'},$_]}
             (@{$self->{XMLHits}}, @{$self->{WaisHits}});

  # See if there were documents at all.
  my $any_docs = $self->there_are_docs(\@hits);

  # Output any error notes if we need to.
  my $were_errors = $self->show_errors_if_any($any_docs);

  # Start the page.
  $APP->start_page(Title=>'Search results',
                   Context=>'guide_results_bydatacenter');

  # Link to switch to the other flavor of display.
  $self->{XML}->by_relevance_link();

  # Say how it all went.
  print $self->{DocLimitMsg};

  # Print the document list, if needed.
  if ($any_docs) {
    # Header hasn't been output.
    my $head_out = 0;

    # For 508 compliance, let them know the docs pop up in a new window.
    $HTML -> P(STYLE=>"margin-left:5%")
          ->t("Documents will open in a new window.")
          ->_P;

    # Partition the results by the DAAC they came from this time.
    # We make the hash be indexed by the DAAC names, with the hits
    # in anonymous lists again.
    my %per_daac = ();
    local $_;
    # Stuff all the hits into the hash:
    foreach (@hits) {
      $_->{'source'} = $_->{'Daac'} || 'Undetermined archive' 
         unless $_->{'source'};

      # Add to list if list is there.
      if ($per_daac{$_->{'source'}}) {
        push @{$per_daac{$_->{'source'}}}, $_;
      }
      # Create new list if no list is there.
      else {
        $per_daac{$_->{'source'}} = [$_];
      }
    }

    # Grab each set of hits and print them out.
    foreach my $daac (sort keys %per_daac) {
      my @local_hits;
      @local_hits = @{$per_daac{$daac}};
      # Sort the items by rank; this works because we're within
      # a single DAAC.
      @local_hits = map { $_->[0] }
      sort {$a->[1] <=> $b->[1]}
      map { [$_, $_->{'rank'}] }
      @local_hits;
      # Current relevance is undef; this way, OutputDoc() will not bother
      # to compute the relevance and handle the fields.
      my $current_relevance = undef;
      # Do the hits for this DAAC. OutputDoc() handles printing
      # of the relevance headers.
      foreach my $hit (@local_hits){
        #debug "$daac hit";
        #local $_; foreach (split(/\n/,Dumper($hit))) {debug $_}
        
        # Skip non-document hits.
        next if $hit->{HitType} eq 'err' or 
          $hit->{HitType} eq 'cat' or
            $hit->{'code'} eq '202';
        # Print the DAAC header the first time we get here, and
        # skip it every time thereafter.
        $HTML->H3->t("Results for $daac (",
                     int @local_hits,
                     (int @local_hits eq 1 ? " hit)" : " hits)"))
          ->_H3 -> OL
              unless $head_out++;;
        # We send in whatever the relevance is now; OutputDoc()
        # calculates the relevance for this hit and returns it.
        # Since we've sorted by rank, the relevances will always
        # be decreasing,so each relevance category header only
        # gets output once.
        #
        # Since the relevance comes back each time, we can maintain
        # state for OutputDoc() and it doesn't have to fiddle with
        # the object to maintain state.
        $self->{$hit->{HitType}}
             ->OutputDoc($hit,$current_relevance,
                         $hit->{DaacFile});
      }
      $HTML -> _OL if $head_out;
      $head_out = 0;
    }
  }

  # Print the catalogs if there were errors or there were not any docs,
  # displaying the errors first if there were any.
  $self->show_cats() if $were_errors or !$any_docs;

  # Link to raw XML and the user's query.
  $self->finish_display;
}

# submode SHOWBYRELEVANCE
#---------------------------------------------------------------------
# Sort the results by relevance and then by DAAC within relevance.
sub submode_SHOWBYRELEVANCE {
  my $self = shift;
  $self->restore_history;

  # If someone was attempting to be clever, send them back to the 
  # status page if nothing's done.
  my $wais = IU::Spawn->attach(Lockfile=>$self->{WaisStat});
  my $xml  = IU::Spawn->attach(Lockfile=>$self->{XMLStat});
  $self->submode_SHOWSTATE if $wais->running or $xml->running;

  # All done. Let's show it.
  $self->load_results;
  $self->reconcile_hits;

  # Show by relevance, so don't list relevance in each hit.
  $self->{XML}->suppress_relevance(1);
  $self->{Wais}->suppress_relevance(1);

  # Sort items into descending order by score; we'll partition
  # them into relevance categories next.
  my @hits = map {$_->[1]}
             sort {$b->[0] <=> $a->[0]}
             map {[$_->{'score'},$_]}
             (@{$self->{XMLHits}}, @{$self->{WaisHits}});

  # See if there were documents at all.
  my $any_docs = $self->there_are_docs(\@hits);

  # Output any error notes if we need to.
  my $were_errors = $self->show_errors_if_any($any_docs);

  # start me up.
  $APP->start_page(Title=>'Search results',
                   Context=>'guide_results_byrelevance');
  
  # link to switch views.
  $self->{XML}->by_daac_link();

  if ($any_docs) {
    # For 508 compliance, let them know the docs pop up in a new window.
    $HTML -> P(STYLE=>"margin-left:5%")
          ->t("Documents will open in a new window.")
          ->_P;
    # No header yet.
    my $head_out = 0;

    # Partition the results. We divide the items into 6 bins by score.
    # We build a hash of lists, with the relevance as the hash key, and
    # the hits in the values. We use references to anonymous arrays to
    # handle the lists of values.
    my %per_relevance = ();
    my $relevance =
      ['Very low', 'Low', 'Medium', 'High', 'Very high', 'Highest'];
    foreach (@hits) {
      next unless $_->{'score'};
   
      # Calculate the relevance for this item.
      my $this_rel = $relevance->[int($_->{'score'}/20)];

      # If there is already a list of items here, add this one
      # on to the end.
      if ($per_relevance{$this_rel}) {
        push @{$per_relevance{$this_rel}}, $_;
      }
      
      # If this is the first item, add it as a one-item list.
      else {
        $per_relevance{$this_rel} = [$_];
      }
    }

    # Output the data. Skip any relevances that we don't have anything for.
    # For the ones we actually do have, dump out a header and a list of
    # the items.
    my $rel_head_out;
    foreach my $rel (reverse @$relevance) {
      my @local_hits;
 
      # Skip if this slot is empty.
      next unless ref $per_relevance{$rel} eq "ARRAY";
      
      # Sort the items in this slot by score, descending.
      # This way the highest-relevance stuff occurs toward
      # the top of each list.
      @local_hits = map { $_->[0] }
                    sort {$b->[1] <=> $a->[1]}
                    map { [$_, $_->{'score'}] }
                    @{$per_relevance{$rel}};
      
      # Run through the list of items in this relevance.
      foreach my $hit (@local_hits){
      
        # Skip error and catalog entries.
        next if $hit->{HitType} eq 'err' or 
                $hit->{HitType} eq 'cat' or
                $hit->{'code'} eq '202';
     
        # Output the header the first time we get here; subsequent
        # times, $head_out will be non-zero, so we'll skip printing.
        #$HTML->H3->t("$rel relevance (", int @local_hits,
        #             (int @local_hits eq 1 ? " hit)" : " hits)"))
        #     ->_H3
        #     ->P(STYLE=>"margin-left:5%")->OL
        #       unless $head_out++;
        $HTML->OL unless $rel_head_out++;
        
        # Note that we pass an undefined last_relevance, so
        # OutputDoc won't try to handle things itself. We add
        # the Source parameter to force the datacenter ID to be
        # added to each entry.
        $self->{$hit->{HitType}}
             ->OutputDoc($hit,
                         undef,
                         $hit->{DaacFile},
                         Source=>1);
        $HTML->P;
      }
      # If we actually output the header, we have to close it now.
    }
    $HTML->_OL if $rel_head_out;
  }

  # Done with the list of items. Show the catalog if either there were no 
  # items or there were errors.
  $HTML->_UL;
  $self->show_cats() if $were_errors or !$any_docs;
  
  # Print the user's query and a link to the raw XML display page.
  $self->finish_display;
  1;
}

# submode SHOWLOG
#-------------------------------------------------------------------
# Display the subprocess logs.
sub submode_SHOWLOG {
  my $self = shift;
  $self->restore_history;
  local $_;

  # Get process log names for the subprocesses.
  my $w_log = $self->{Wais}->process_log($SESSION->path_to() .
                                         $self->{Serial} . "w_log");
  my $x_log = $self->{XML}->process_log($SESSION->path_to() .
                                        $self->{Serial} . "x_log");
  $APP->suppress_links(1);
  $APP->start_page;
  $HTML->H2->t("WAIS log")->_H2->PRE;
  open (SRUNLOG,"<$w_log") or print "WAIS log $w_log unreadable: $!";
  print while(<SRUNLOG>);
  close SRUNLOG;
  $HTML->_PRE->HR->H2->t("XML log")->_H2->PRE; 
  open (SRUNLOG,"<$x_log") or print "XML log $x_log unreadable: $!";
  print while(<SRUNLOG>);
  close SRUNLOG;
  $HTML->_PRE;
  $APP->end_page;
}

# Alias submode_SHOWXML to submode_SHOWDEBUG.
*submode_SHOWXML = \&submode_SHOWDEBUG;

sub submode_SHOWDEBUG {
   my $self = shift;
   $self->{XML}->submode_SHOWXML();
}

sub submode_DEFAULT {
   my $self = shift;
   $self->submode_GUIDESRCH();
}

sub submode_OUT_OF_LINE_GET {
   my $self=shift;
   $self->{Wais}->submode_OUT_OF_LINE_GET(@_);
}

#-------------------------------------------------------------
# Mode infrastructure support
#-------------------------------------------------------------
# go
#
# DESCRIPTION
#
#    Determine the submode. The first time through it should be GUIDESRCH.
#    once you've gotten a list of docs then its calls itself with submode
#    GUIDEGET to retrieve the docs.
#
sub go {
    my $self = shift;
    $APP->get_session();
    $self->go_submode();
    1;
}

1;
