#------------------------------------------------------------
# ModeGUIDEOLD.pm
#------------------------------------------------------------
package IW::ModeGUIDEOLD;

=head1 NAME

IW::ModeGUIDEOLD - wais guide search.

=head1 DESCRIPTION

Searches for and returns guide documents.
Performs all the guide searches:

=begin enum

=item

Composed Guide Search using guide search form.

=item

Selecting "Get Detailed Information" for an earth science attribute
(valid) from the bottom of the attribute selection pages.

=item

"Detailed Information" buttons next to data sets in data set listings.

=end enum

=head1 CGI TAGS for SEARCH

=begin list

=item

B<"submode"> - gives the mode of operation:

=begin list

=item

B<GUIDESRCH> (default) - to initiate search for guide documents.

=item

B<GUIDEGET> - to retrieve and display the document selected or found.

=item

B<OUT_OF_LINE_GET> - to retrieve and display the document selected or found,
sans navigator.

=end list

=item

B<"sid"> - gives the session id.

=item

B<"doctype">  gives the document type.  It should be one of:

=begin list

=item

Campaign

=item

Data_Center

=item

Dataset

=item

Parameter

=item

Sensor

=item

Source

=end list

=item

B<"item"> - if searching, gives the item we are searching for.
This form may contain spaces: they will be turned to underscores
automatically by this script when generating the wais query.

=end list

=head2 ADDITIONAL CGI TAGS for RETRIEVAL

=begin list

=item

B<"daac"> - short string for data center from which to retrieve document.

=item

B<"doctype"> - same as in search.

=item

B<"srchfile"> - name of the query file produced (in the session directory)
when a search was done.

=item

B<"catalog"> - a flag indicating that the document is not a guide document
but is a data center's catalog of documents.

=end list

=head1 GuideType

The SETUP file contains a configuration variable for guide searches
called "GuideType".  It is one of 'OLD', 'NEW', or 'BOTH'.  This type
determines the form of the search string.

=begin list

=item

OLD is the search string like: "Parameter=CARBON or Parameter=OZONE".

=item

NEW is the search string like: "parm928 or parm2039".

=item

BOTH is the combined search string like:

"(parm928 or parm2039) or (Parameter=CARBON or Parameter=OZONE)".

=end list

=head1 INVOCATIONS

Guide searches are currently (WG 2.1, July 1998) invoked in the
following ways:

=begin enum

=item

Composed Guide Search using guide search form.

=begin list

=item

When you fill in a Guide search form and click the Start Search button,
nph-cgi invokes IW::ModeSEARCH.pm's go(), which invokes start_search(),
which invokes start_guide_search(), which switchs to Mode GUIDE,
submode GUIDESRCH, with the query object constructed from the
form as a CGI parameter.

=end list

=item

Selecting "Detailed Document" for an earth science attribute
(valid) from the CGI attribute selection pages.

or

Using "Detailed Document" buttons next to data sets in data set listings.

or

Clicking the "Detailed Document" button in the dependent valids applet.

=begin list

=item

All of these will invoke IW::ModeGUIDEOLD::go() without a query object.
The cgi form will pass the appropriate parameters to search.  go()
invokes submode_GUIDESRCH() which invokes DoSearch().

If the invocation is via the applet, a special submode should be supplied 
(OUT_OF_LINE) which 

=begin list

=item 

causes the first page shown to suppress the navigator

=item

causes all links pointing off this page to use the OUT_OF_LINE_GET rather than 
GUIDEGET submode in retrieving the pages (the navigator will be
suppressed here too).

=end list

=end list

=item

Selecting a document link from the list of returned matching guide
documents from a guide search.

=begin list

=item

This invokes IW::ModeGUIDEOLD::go(), which invokes submode_GUIDEGET(),
which invokes DoRetrieve().  Notice that a guide search does NOT
return the actual URL's of the documents, so DoRetrieve()
actually uses the WAIS server to serve up the document.

submode_OUT_OF_LINE_GET() is used if the original page had its navigator 
suppressed.

=end list

=item

As of WG 1.9, if a search finds just one document, DoSearch() will
directly call DoRetrieve() to get the document.

=item

If a search finds no matching documents, the WAIS servers at the
data centers return a link to their catalog of Guide documents.
Prior to WG 1.9 we simply displayed the message: "No relevant
guide documents were found at this DAAC".  As of WG 1.9, we display
the catalog link, so the user can retrieve the catalog.

=end enum

=head1 WARNING

WE ASSUME THE CHARACTERS '=', '?', AND '|' WILL NEVER BE IN ANY VALID!

=head1 AUTHORS

Portions of this code were retrieved from (wais.pl) or son-of-wais.
Eryq adapted the code to fit in with the EOSDIS V0 Gateway distribution.

=begin list

=item

Douglas J. Hoffman, Hughes STX, 10/9/95

=item

Eryq (Erik Dorfman), Hughes STX

=item

Ed Seufert, Hughes STX, Aug 14, 1996

Extensively rewritten for web gateway release 1.5 to use the new
object oriented scheme and libraries created by Eryq.

=item

Larry Rosen, Hughes STX, 1997

=item

Joe McMahon, Raytheon STX, 1998

Added support for link suppression; several subroutine were turned into
methods to allow the link suppression to be "carried down" into the routines
in a relatively transparent way.

=item 

Joe McMahon, GST, 2001

Added support for combined XML/WAIS searching.

=end list

=head1 COPYRIGHT

 h / Eryq (Erik Dorfman) 
NO WARRANTY; see TERMS AND CONDITIONS in distribution for details.

=head1 RCS INFO

$Id: ModeGUIDEOLD.pm,v 1.10 2002/04/02 20:05:40 seufert Exp $

=cut

use IU;
use IU::Log;
use IU::Session;
use IU::Event;
use IU::ODL;
use IU::Query;
use IU::Storable;
 
use IW::App;

use URI::Escape;
use IW::Util qw(:url :output);
use IU::Util qw(shellquote);
use strict; # best for the novice perl user.
use Data::Dumper;
@IW::ModeGUIDEOLD::ISA = qw(IW::Mode IW::ModeGUIDE);
#------------------------------------------------------------
#
# Globals
#
#------------------------------------------------------------
my $Waisq = "waisq";     # The "waisq" program
my @SCenterList = ();

# Map 'doctype' tag values to human-readable description:
my %DoctypeToDesc = ('Campaign'    => 'campaign',
		  'Data_Center' => 'data center',
		  'Dataset'     => 'data set',
		  'Parameter'   => 'parameter',
		  'Sensor'	=> 'sensor/instrument',
		  'Source'      => 'source/platform');		  

# Map the 'doctype' tag to the waisq value used when specifying the 
# $qdoctype in "Document_Type=$qdoctype":
my %QueryDoctype =  ('Campaign'    => 'Campaign',
		  'Data_Center' => 'Data_Center',
		  'Dataset'     => 'dataset',
		  'Parameter'   => 'parameter',
		  'Sensor'	=> 'sensor',
		  'Source'      => 'source');
# Map the 'doctype' tag to the new DocType*.
my %NewDoctype = ('Campaign'    => 'Camp',
                  'Data_Center' => 'Arch',
                  'Dataset'     => 'DS',
                  'Parameter'   => 'Parm',
                  'Sensor'	=> 'Sensor',
                  'Source'      => 'Source');

my $sSysWais;
my $nDataCenterHits;

#------------------------------------------------------------
# new 
#------------------------------------------------------------
# I<Constructor.>
sub new {
    bless {}, shift;
}

#------------------------------------------------------------
# skip_daacs NAMES
#------------------------------------------------------------
# Set the list of names that WAIS searching should skip.
sub skip_daacs {
    my ($self,@which) = @_;
    $self->{SkipThese} = \@which if @which;
    $self->{SkipThese};
}

#------------------------------------------------------------
# SearchString -- build waisq search string 
#------------------------------------------------------------
#
# DESCRIPTION
#    Given the document type and the item (for a guide search)
#    or a constructed query, build the actual query sent to waisq.
#
# ARGUMENTS
#    If two arguments:
#       $doctype is the document type (as described up top)
#       $item is the item (as described up top)
#    Else one argument:
#       $query is a query object.
#
# RETURNS
#    Returns two strings.  The first is the query string to send
#    to waisq.  The second is the string to print on the results
#    page.  They are similar but different, especially once we
#    begin using the new coded strings.
#
sub SearchString {
    my $self = shift;
    # Command to imsmapguide (fullpath)
    my $map_cmd = "${IU::LibDir}/bin/imsmapguide";

    my $the_string = '';    # The old style search string, or string to print.
    my $new_string = '';    # The search string converted to new codes.
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # If two arguments, item is specified.  This is a simple search
    # for a guide document for a valid from an attribute screen or
    # dataset from a dataset listing from an inventory search.
    if (@_ == 2) {
        my ($doctype, $item) = @_;
        # If OLD guide type, return simple string.
        if (${IU::GuideType} =~ /OLD/i) {
            $item =~ tr/ /_/;    # convert spaces to underscores
            $the_string =
               "$doctype=$item and Document_Type=$QueryDoctype{$doctype}";
            return ($the_string, $the_string);
        } else {
            # We must get the new guide string.
            # Open a read pipe to the imsmapguide program, using
            # the requested doctype and item as arguments.
            # First shellquote the item for command line to imsmapguide
            # to protect agains any special characters.
            $new_string = shellquote($item);
            open (MAPGUIDE, "$map_cmd $doctype $new_string |") or
                $self->andiamo("Cannot execute search - " .
                               "cannot open imsmapguide.");
            $new_string = <MAPGUIDE>;
            close (MAPGUIDE);
            # Convert spaces in item to underscores
            $item =~ tr/ /_/;
            $the_string =
               "$doctype=$item and Document_Type=$QueryDoctype{$doctype}";
            $new_string .= " and DocType$NewDoctype{$doctype}";
            # If NEW or TRANSITIONAL guide type, return imsmapguide output.
            if (${IU::GuideType} =~ /NEW|TRANSITIONAL/i) {
               return ($new_string, $the_string);
            } else {
                # Guide type is BOTH. Merge old search string then return.
                $new_string = '(' . $new_string . ') or (' . $the_string . ')';
                return ($new_string, $the_string);
            }
        }
    }
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Else if one argument, a query object is specified, this is a more 
    # involved search. The search string must be constructed from the 
    # elements of the query object. To generate a special character safe
    # string, use '|' for ' '.
    # WE ASSUME '=', '?', AND '|' WILL NEVER BE IN ANY VALID!
    #
    # A string is initially constructed to look like this:
    #    "(Parameter=CARBON|or|Parameter=NITRIC ACID)|and|(Data_Center=GSFC)"
    #    (Notice the space in the item and the | around booleans)
    #
    # To generate the old style search string, translate ' ' to '_' and
    # '|' to ' '.
    #    "(Parameter=CARBON or Parameter=NITRIC ACID) and (Data_Center=GSFC)"
    #
    # To generate the new style search string, create a specially formatted
    # string to write to a file for input to imsmapguide.  Modify the string:
    #    Prepend string with number of attribute=value pairs and a newline.
    #    Replace ' ' with '?' and '=' with ' '.
    #    Eliminate first and last parenthesis.
    #    Then replace '|' before a boolean with a newline '\n'
    #    and replace '|' after boolean with a space and eliminate
    #    parentheses around the and's.
    # Send this string to imsmapguide to generate new coded string.
    #    "3
    #     Parameter CARBON
    #     or Parameter NITRIC?ACID
    #     and Data_Center GSFC"
    #
    elsif (@_ == 1) {
        my $query = shift;
        my @items;                   # Items for which to search.
        my $use_and = 0;             # 0, False if 1st attribute.
                                     # 1, True, prepend " and ".
        my $text_string = '';        # Entered free text string if any.
        my $param_only = 0;          # 1=True=User composed a search with only
                                     # parameters selected. Add Document_Type
                                     # =Dataset to only get dataset documents.

        # Note to Programmers: The code for Parameters is slightly different
        # than subsequent attributes, because it is first (no prefix " and ").

        # Parameters
        #-----------
        @items = $query->params;
        if (@items) {
           $the_string = '(';
           $the_string .= join ('|or|',
              map {"Parameter=$items[$_]"} (0..$#items));
           $the_string .= ')';
           $use_and = 1;
        }
        # Sensors
        #-----------
        @items = $query->sensors;
        if (@items) {
           if ($use_and) {
              $the_string .= '|and|';
           } else {
              $use_and = 1;
           }
           $the_string .= '(';
           $the_string .= join ('|or|',
              map {"Sensor=$items[$_]"} (0..$#items));
           $the_string .= ')';
        }
        # Sources
        #-----------
        @items = $query->sources;
        if (@items) {
           if ($use_and) {
              $the_string .= '|and|';
           } else {
              $use_and = 1;
           }
           $the_string .= '(';
           $the_string .= join ('|or|',
              map {"Source=$items[$_]"} (0..$#items));
           $the_string .= ')';
        }
        # Daacs
        #-----------
        @items = $query->daacs;
        if (@items) {
           if ($use_and) {
              $the_string .= '|and|';
           } else {
              $use_and = 1;
           }
           $the_string .= '(';
           $the_string .= join ('|or|',
              map {"Data_Center=$items[$_]"} (0..$#items));
           $the_string .= ')';
        }
        # Datasets
        #-----------
        @items = $query->datasets;
        if (@items) {
           if ($use_and) {
              $the_string .= '|and|';
           } else {
              $use_and = 1;
           }
           $the_string .= '(';
           $the_string .= join ('|or|',
              map {"Dataset=$items[$_]"} (0..$#items));
           $the_string .= ')';
        }
        # Campaigns
        #-----------
        @items = $query->campaigns;
        if (@items) {
           if ($use_and) {
              $the_string .= '|and|';
           } else {
              $use_and = 1;
           }
           $the_string .= '(';
           $the_string .= join ('|or|',
              map {"Campaign=$items[$_]"} (0..$#items));
           $the_string .= ')';
        }
        # If user entered a free text string, get it but don't append
        # it yet.  We won't try to encode it with imsmapguide.
        $text_string = $query->textstring;

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # If $the_string eq '', then no valids selected.
        if ($the_string eq '') {
           # If no text string, query was empty.  Problem.
           if ($text_string eq '') {
              $self->andiamo("You must supply at least one parameter, sensor, ".
	                     "data set, source, campaign, data center, or text ".
                        "string.");
           }
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
           # Text string provided. Return it as string.
           else {
              return ($text_string, $text_string);
           }
        }
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # Else valids were selected.
        # Determine if user ONLY selected parameters.  If yes, then set flag
        # to later add Document_Type=Dataset to only get dataset documents.
        if ($query->params && !$query->sensors && !$query->sources &&
            !$query->daacs && !$query->datasets && !$query->campaigns &&
            $text_string eq '') {
           $param_only = 1;
        }
        # If guide type is NEW or BOTH, then we need to get the encoded
        # search string from imsmapguide.
        if (${IU::GuideType} !~ /OLD/i) {
           # Count the number of attribute=value pairs.
           my $pair_count = $the_string =~ tr/=/=/;

           # Make a new style string that begins with pair count, a newline,
           # then the_string with: replace ' ' with '?' and '=' with ' ',
           # then replace pipe before a boolean with a newline '\n'
           # and replace pipe after boolean with a space.

           # Begin with a copy of the string.
           $new_string = $the_string;
           # Remove the final closing parenthesis.
           chop $new_string;
           # Remove the first parenthesis.
           substr ($new_string, 0, 1) = '';
           # Prepend the pair count and a newline to the string. Append a \n.
           $new_string = "$pair_count\n" . $new_string . "\n";
           # Replace ' ' with '?' and '=' with ' '.
           $new_string =~ tr/ =/? /;
           # Replace '|' before a boolean with a newline '\n'
           # and replace '|' after boolean with a space and eliminate
           # parentheses around the and's.
           $new_string =~ s/\|or\|/\nor /g;
           $new_string =~ s/\)\|and\|\(/\nand /g;
           # Write the new style string to a NewGuide file in the session dir.
           my $stringfile = $SESSION->path_to() . 'NewGuideString';
           open (NEWGUIDEFILE, ">$stringfile") or
              $self->andiamo("Cannot execute wais search (cannot open string file $stringfile).");
           print NEWGUIDEFILE $new_string;
           close (NEWGUIDEFILE);
           # Make the file read/write only by wwwuser.
           chmod 0600, $stringfile;

           # Use imsmapguide to get the new guide coded string.
           # Open a read pipe to the imsmapguide program, telling it the file
           # containing the new formed string above.
           open (MAPGUIDE, "$map_cmd $stringfile |") or
               $self->andiamo("Cannot execute search - cannot open imsmapguide.");
           # Read the new coded string from the imsmapguide program.
           $new_string = <MAPGUIDE>;
           close (MAPGUIDE);
           # Prepend and append parentheses.
           $new_string = '(' . $new_string . ')';
           # Put parentheses around ' and ' -> ') and ('
           $new_string =~ s/ and /\) and \(/g;

           # If parameter only search, add Document_Type=Dataset.
           if ($param_only) {
              $the_string .= '|and|(Document_Type=Dataset)';
              $new_string .= ' and (' . "DocType$NewDoctype{Dataset}" . ')';
           }
           # If a text string was provided, append it with 'and' to the
           # search string ($new_string) and the print string ($the_string)
           if ($text_string) {
              $the_string .= '|and|';
              $new_string .= ' and (' . $text_string . ')';
              $the_string =~ tr/ \|/_ /;
              $the_string .= '(' . $text_string . ')';
           } else {
              $the_string =~ tr/ \|/_ /;
           }
           if (${IU::GuideType} =~ /NEW/i) {
              # Return the new string.
              return ($new_string, $the_string);
           } else {
              # Return BOTH the new string and the old string, 'or'
              # together with parentheses.
              $new_string = '(' . $new_string . ') or (' . $the_string . ')';
              return ($new_string, $the_string);
           }
        }
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # Else Guide Type OLD.
        # If parameter only search, add Document_Type=Dataset.
        if ($param_only) {
           $the_string .= '|and|(Document_Type=Dataset)';
        }
        # Translate ' ' to '_' and '|' to ' '.
        $the_string =~ tr/ \|/_ /;
        # If a text string was provided, append it with 'and' to the
        # search string.
        if ($text_string) {
           $the_string .= ' and ';
           $the_string .= '(' . $text_string . ')';
        }
        return ($the_string, $the_string);
    }
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Else this function has been misused.
    else {
        puke "SearchString() function used incorrectly. Wrong number args.";
        return ("Unknown", "Unknown");
    }
}

#------------------------------------------------------------
# DataCenterList 
#------------------------------------------------------------
#
# RETURNS
#    The number of data centers on success, undef on error.
sub DataCenterList 
{
    my $query = shift;
    my($sDepOdl,$sDsInput, $nRet);
    my $nCenterCount;

    # Get some of the basic info:
    my $doctype = $REQ->param('doctype');
    my @item    = ($REQ->param('item'));

    require IW::DepValid;   
    my $depval = new IW::DepValid;

    # Get the data centers for this search:
    # If query supplied, set dep valids from query.
    if ($query)  {
        $query->determine_daacs_to_search;
        # The compatible list of daacs is:
        @SCenterList = $query->daacs_to_search;
    }
    # Else set dep valids from doctype & item.
    else {
        if ($doctype =~ /data_center/i) {
            push(@SCenterList, $item[0]);
        } else {
            if ($doctype =~ /source/i) {
                $depval->set('SOURCE_NAME',\@item);
            }
            elsif ($doctype =~ /sensor/i) {
                $depval->set('SENSOR_NAME',\@item);
            }
            elsif ($doctype =~ /dataset/i) {
                $depval->set('DATASET_ID',\@item);
            }
            elsif ($doctype =~ /parameter/i) {
                $depval->set('PARAMETER',\@item);
            }
            elsif ($doctype =~ /campaign/i) {
                $depval->set('CAMPAIGN',\@item);
            }
            # The compatible list of daacs is:
            @SCenterList = $depval->get('DATA_CENTER_ID');
        }
    }
    if (!@SCenterList || !defined($SCenterList[0])) {
        return &Iw::Err(460, "no data centers for this guide query!");
    }
    # Set up program data structures, and return:
    $nCenterCount = int(@SCenterList);
}

#------------------------------------------------------------
# DocUrlForHit -- given a "hit" object, return the document URL
#------------------------------------------------------------

sub DocUrlForHit {
    my $self = shift;
    my $hit = shift;
    my $daacfile = uri_escape (shift);

    my $daac = uri_escape ($hit->{Daac});
    my $docnum = $hit->{DocNum};

    # Determine if we are in suppressed-links land and set the GET submode.
    my $submode = ($self->suppress_links ? 'OUT_OF_LINE_GET' : 'GUIDEGET');
    $submode = (uc($IU::GuideType) eq "TRANSITIONAL") ?
                   "OUT_OF_LINE_GET" : $submode;
    # Create the hyperlink to the document itself:    
    my $link = script_url ('mode'=>'GUIDE', 'submode'=>$submode,
			   'daac'=>$daac, 'docnum'=>$docnum,
			   'srchfile'=>$daacfile);

    # If the link is a catalog link instead of a guide document, put in a flag.
    if ($hit->{Catalog}) {
        $link .= "&catalog=1";
    }
    return($link);
}

#------------------------------------------------------------
# OutputDoc -- we've got all the info we need... output the document summary
#------------------------------------------------------------
#
# ARGUMENTS
#    $hit is a ref to hash of the document info extracted from the
#    waisq output.  It contains the following tags:
#
#       Daac     => data center where this hit came from
#       DocNum   => document number
#
#       Date     => document date
#       Headline => headline info, or undef
#       Type     => document type, or undef
#       Score    => WAIS file score, 0-1000
#       Lines    => number of lines in file
#       Bytes    => number of bytes in file
#
# NOTES
#    The document is output as a list item.  It is the caller's
#    responsibility to ensure that there is a list environment surrounding
#    it.

sub OutputDoc {
    my ($self,$hit,$junk,$daacfile) = @_;
    if ($IU::GuideType eq 'TRANSITIONAL') {
       ($self,$hit,$junk,$daacfile) = @_;
    }
    else {
       ($self,$hit,$daacfile) = @_;
    }

    my $title = $hit->{Headline};

    # Create the hyperlink to the document itself:    
    my $href = $self->DocUrlForHit($hit, $daacfile);
    print "<LI><B>";	  
    if ($IU::GuideType eq "TRANSITIONAL") {
       $HTML->hypertext($title,$href,TARGET=>'support');
    }
    else {
       $HTML->hypertext($title,$href);
    }
    print "</B><BR>\n";
    if (uc($IU::GuideType) eq "TRANSITIONAL") {
      print "Datacenter: $hit->{Daac}, " unless $self->suppress_daac();
      print "Relevance: ", 
            ['Very low','Low','Medium','High','Very high','Highest'] ->
              [int($hit->{'Score'}/200)],
            ", " unless $self->suppress_relevance();
    }
    else {
      print "Score: $hit->{Score}, ";
    }
    print "Lines: $hit->{Lines}, ";
    print "Bytes: $hit->{Bytes}\n";
    1;
}

#------------------------------------------------------------
# SendGuideListing -- output a list of the guide results
#------------------------------------------------------------
#
# DESCRIPTION
#    We just did a guide search that returned either 0 hits,
#    or more than one hit.  Output a listing.

sub SendGuideListing {
    my $self = shift;
    my ($byDaac, $sSearchString, $queryfile, $maxdocs) = @_;
    #
    # Put an entry in the ims event log that a guide search was done
    # and include the list of docs returned.
    #
    my @guideitems = ();

    # Get some basic info:
    my $item = $REQ->param('item');
    my $doctype = $REQ->param('doctype');

    # Tell App we want the link suppressed if so noted.
    $APP->suppress_links($self->suppress_links);

    # Print the page header (with HTTP header):
    $APP->start_page(Title  =>'Detailed Document Search', 
		     Context =>'guidesrch');

    # Output the summary:
    if ($sSearchString) {
        $sSearchString =~ s/ or / or <br>\n/g;
        $sSearchString =~ s/ and / and <br>\n/g;
#        $sSearchString =~ s/_/ /g;
        print "<hr><h2>Search String:</h2><p>\n<h3>$sSearchString\n</h3><hr>";
    }
#    else {
#        print "<hr><h2>\u$DoctypeToDesc{$doctype} ";
#        print "=&quot;$item&quot;:</h2>";
#    }

    # Output the "hit" documents, with the DAACs sorted alphabetically...
    my $daac;
    my $totalHits = 0;
    my $daacfile;
    
    foreach $daac (sort @SCenterList) {	

	print "<big><b>Data Center: $daac </b>&nbsp&nbsp&nbsp&nbsp \n";

	# Tally the hits at this DAAC:
	my $hits_at_daac = int(@{$byDaac->{$daac} || []});
	$totalHits += $hits_at_daac;
	if ($hits_at_daac >= $maxdocs) {
	    print "</big>At least <big>$maxdocs</big> documents ";
	    print "matched your query\n<ul>\n<li><i><b>\n";
	    print "There may be more than these documents, but the current ";
	    print "document limit is set to $maxdocs.  You can increase this ";
	    print "limit by going to ";

	    my $href = script_url('mode'    => 'MAINSRCH',
                                  'SrchTypeevent'  =>'TOGGLE',
                                  'SrchTypenewtype' =>'GUIDE_SEARCH');
	    $HTML -> A(HREF=>$href)
		  -> t("Additional Options")
		  -> _A;

	    print " (on the  ";
	    print " detailed document search form).  After you change ";
	    print "the limit you can resubmit your guide document search.";
	    print "</b></i>\n";
	} elsif ($hits_at_daac != 1) {
	    print "$hits_at_daac </big> documents matched your query\n<ul>\n";
	} else {
	    print "1 </big> document matched your query\n<ul>\n";
	}

	# Output the hit documents:
	if ($hits_at_daac == 0) {      # we got nothing...
	    print "<li><i>\n";
	    print "No relevant guide documents were found at this DAAC.\n";
	    print "</i>\n";
	}
	else {                         # we got something!
            my $hit;
            $daacfile = $queryfile . '.' . $daac;
	    foreach $hit (@{$byDaac->{$daac}}) {
	        $self->OutputDoc($hit, $daacfile);
                push(@guideitems,'guideitem'=>$hit->{Headline});
            }
	}
	print "</ul>\n";
    }
    # Summary:
    print "<hr>\n";
    print "A total of <b>$totalHits</b> documents were found, in the\n";
    print int(@SCenterList), " data center(s) that were searched.\n";
    print "<P>\n";

    ## debug "@guideitems";

    # Print the page footer:
    $APP->end_page;
 
    ims_event('GUIDE',
               @guideitems);
    1;
}

#------------------------------------------------------------
# DoSearch -- do the wais search
#------------------------------------------------------------

# NOTES
#    [EQ] The special document name beginning with "Search produced no result."
#    is not a real document.  This routine now contains a switch to ignore
#    such a document, instead of adding it to the results.

sub DoSearch {
    my $self = shift;
    $self->process_debug("Arrived in DoSearch; results file $self->{ResultsFile}");
    my $uniqnum = (int(time) % 1000000);     # Hold unique number per search.
    my $sSearchString;                       # Query string for waisq.
    my $print_string;                 # String to print on results page.

    # If query is present, then use it to construct the search string
    my $query = $REQ->param('query');

    if (defined ($query)) {
        ($sSearchString, $print_string) = $self->SearchString($query);
    }
    # Else get doctype and item from CGI.
    else {
        # Get the necessary CGI values:
        my $doctype = $REQ->param('doctype');
        my $item    = $REQ->param('item');
        ($sSearchString, $print_string) = $self->SearchString($doctype, $item);
    }
    # Search strings longer than 254 characters causes waisq in freeWAIS 0.5
    # to fail.  We warn user.
    if (length($sSearchString) > 254) {
        $self->andiamo("Detailed document search query too large. ".
                       "Please select fewer items to search.");
        return (1);
    }
    $self->process_debug("print string is $print_string");
    # Create a file to store the search string.
    my $queryfile = $SESSION->path_to() . 'GuideSearch.' . $uniqnum;
    open (SEARCHFILE, ">$queryfile") or
        $self->andiamo("Cannot execute wais search (cannot open query file).");
    print SEARCHFILE $sSearchString;
    close (SEARCHFILE);
    # Make the file read/write only by wwwuser.
    chmod 0600, $queryfile;
    $self->process_debug("query file $queryfile built");

    # Create the array of all "hit" documents:
    my @hits = ();
    my $nHits = 0;
    # Get the wais sources directory (which == the valids directory).
    my $Waisd  = "$IU::GaeaDataDir";
    # Get the list of data centers to query:
    if ($query) {
        &DataCenterList($query) || return undef;
    } else { 
        &DataCenterList || return undef;
	# Since we were not passed the query object (because user chose
	# detailed info), get the query object from disk.  We need it
	# soon to get maximum number of guide documents to retrieve.
	$query = IU::Query->lookup();
    }
    #
    # SEARCH...
    #
    # Search each data center in turn:
    my $sDataCenter;
    my $daacfile;
    # Set the maximum number of guide documents to retrieve.
    my $maxdocs = $query->max_guides;

    foreach $sDataCenter (@SCenterList) {

        # Skip DAACs that are under Spidex control.
        $self->process_debug("Checking $sDataCenter");
        next if uc($IU::GuideType) eq "TRANSITIONAL" and
                grep /^$sDataCenter/,@{$self->{SkipThese}};
                
        $self->process_debug("searching $sDataCenter");
	$nDataCenterHits = 0;
        # Copy the Search String file to a Daac file for output.
        $daacfile = $queryfile . '.' . $sDataCenter;
        if (`cp $queryfile $daacfile 2>>/dev/null`) {
            $self->andiamo("Cannot execute wais search " .
                           "(cannot copy queryfile).");
	}
	# Construct the WAISQ query:
	$sSysWais= 
	    "$Waisq -c $Waisd -m $maxdocs -f $daacfile -S $sDataCenter.src " . 
		"-g " . shellquote($sSearchString) .
		" 2>>/dev/null";
        if (`$sSysWais`) {
            $self->andiamo("Error executing wais search.");
	}
        #debug "The wais command is $sSysWais ";
	# Execute the query, grabbing the results:
	open(WAISQ,"<$daacfile") or
           $self->andiamo("Cannot execute wais search " .
                          "(cannot open daac file).");
        ## debug "Show hits here if any";
	my $hits_rec = {};   # this will hold the document info
 
	while (<WAISQ>) {
        ##    /:score\s+(\d+)/ && (debug "$1");
	    /:score\s+(\d+)/ &&           ($hits_rec->{Score} = $1);
	    /:number-of-lines\s+(\d+)/ && ($hits_rec->{Lines} = $1);
	    /:number-of-bytes\s+(\d+)/ && ($hits_rec->{Bytes} = $1);
	    /:type "(.*)"/ &&             ($hits_rec->{Type} = $1);
	    /:headline "(.*)"/ &&         ($hits_rec->{Headline} = $1); 
	    /:date "(\d+)"/ && do {

	    # Finish up storing document info:
	    $hits_rec->{Date}   = $1;
	    $hits_rec->{Daac}   = $sDataCenter;
	    $hits_rec->{DocNum} = 1 + $nDataCenterHits;
            $hits_rec->{DaacFile} = $daacfile;
	    # Push document onto list, and get ready for next one:
            ## debug "Got a hit $hits_rec";
	    push(@hits, $hits_rec);   
	    $hits_rec = {};
	    # Update the stats:
	    $nDataCenterHits++;
	    $nHits++;
	    }  # do close.
        } # while WAISQ.
	close(WAISQ);
        $self->process_debug("hits now $nHits");
    }  # each data center
    #
    # DO SOME POST-PROCESSING ON RESULTS...
    #
    # Organize results by data center:
    my %byDaac = ();    # arrays of hits, hashed by daac name
    my @cats = ();      # list of catalogs
    my @real = ();      # list of real hits
    my $realNumHits = 0;
    my $lastRealHit;
    my $hit;
    foreach $hit (@hits) {
        # We have not yet chosen where the entry will finally be stacked.
        my $stack_loc = undef;
	# Mark any documents which are actually catalogs.
	if ($hit->{Headline} =~ /^Search produced no result/) {
            $hit->{Catalog} = 1;   # true
            $stack_loc = \@cats;   # stack it on the catalog list
            
	}
        else {
            $hit->{Catalog} = 0;   # false
            $stack_loc = \@real;   # stack it on the real hits list
        }
	# Add hit to the daac's list.  Note there can never be an empty list:
	my $daac = $hit->{Daac};
	defined($byDaac{$daac}) || ($byDaac{$daac} = []);
	push(@{$byDaac{$daac}}, $hit);

	# Remember the hit (in case it's our only one), and tally it:
	$lastRealHit = $hit;
	++$realNumHits;

        # Stack it appropriately for TRANSITIONAL searches.
        push @$stack_loc,$hit;
    }

    # Force no output and record results if we are in TRANSITIONAL mode.
    if (uc($IU::GuideType) eq "TRANSITIONAL") {
       $self->process_debug("Ready to try saving results");
       $self->process_debug("Query=>$sSearchString\nDisplay=>$print_string");
       $self->andiamo("",{
                          Query=>$sSearchString,
                          Display=>$print_string,
                          Hits=>\@real,
                          Cats=>\@cats
                         });
       $self->process_debug("I saved 'em! I saved 'em!");
       return 1;
    }

    # Really do the output if we're not in TRANSITIONAL mode.
    else {
      #
      # OUTPUT...
      #
      if ($realNumHits < 1) {
        # Output the list of NO hits:
        ## debug "No Hits";
        $self->SendGuideListing(\%byDaac, $print_string, 
                                $queryfile, $maxdocs) ||
                                  return undef;
      }
      # If only one hit, and it isn't a "no result - here's the catalog"
      # link, then retrieve it.
      elsif ($realNumHits == 1 && $lastRealHit->{Catalog} == 0) {
        #debug "One guide search hit.  Retrieve it.";
        $daacfile = $queryfile . '.' . $lastRealHit->{Daac};
        $self->DoRetrieve(uri_escape($lastRealHit->{Daac}),   # daac
			  $lastRealHit->{DocNum},             # docnum
			  uri_escape($daacfile),              # srchfile
			  $lastRealHit->{Catalog});           # catalog
      }
      else {
	# Output the list:
	$self->SendGuideListing(\%byDaac, $print_string, 
				$queryfile, $maxdocs) ||
                                  return undef;
      }
    }
    # Done!
    1;
}

#------------------------------------------------------------
# DoRetrieve -- retrieve and return a single guide document
#------------------------------------------------------------
#
# WARNINGS
#    Unlike many of the other functions, this one actually
#    outputs a foreign document.  We assume HTML for now.

sub DoRetrieve {

    # Get needed parameters:
    my $self     = shift;
    my $daac     = shift;
    my $docnum   = shift;
    my $srchfile = shift;
    my $catalog  = shift;

    # Get the wais sources directory (which == the valids directory).
    my $Waisd  = "$IU::GaeaDataDir";

    # Generate the system command:
    my $sys_str;
    $sys_str = "$Waisq -c $Waisd -f $srchfile -S ${daac}.src -v $docnum";

    # Tell App we want the link suppressed if so noted.
    $APP->suppress_links($self->suppress_links);

    # Output (NOTE: can't just dump to "stdout", since we redirect into file!)
    #debug "command is $sys_str";
    $APP->start_page(Title  =>'Detailed Document Retrieval', Context =>'guideretrieval');
    open(DOC, "$sys_str |") or $self->andiamo("couldn't run waisq");
    my $buf = '';

    # If this link is a catalog, read and print a line at a time.
    local $_;
    if ($catalog) {
	$HTML -> head (2, "Catalog of Detailed Documents at $daac") -> P;
    }
    while (<DOC>) {
        print;
        $HTML -> BR if $catalog;
    }
    close(DOC);
    $APP->end_page;
    1;
}
#------------------------------
#
# SUBMODE HANDLERS 
#
#------------------------------

#------------------------------------------------------------
# submode_DEFAULT
#------------------------------------------------------------
sub submode_DEFAULT {
    my $self = shift;
    $self->submode_GUIDESRCH;
}

#------------------------------------------------------------
# submode_OUT_OF_LINE
#------------------------------------------------------------
sub submode_OUT_OF_LINE {
    my $self = shift;
    $self->suppress_links(1);
    $self->submode_GUIDESRCH;
}

#------------------------------------------------------------
# submode_OUT_OF_LINE_GET
#------------------------------------------------------------
sub submode_OUT_OF_LINE_GET {
    my $self = shift;
    $self->suppress_links(1);
    $self->submode_GUIDEGET;
}

#------------------------------------------------------------
# submode_GUIDESRCH
#------------------------------------------------------------
#
sub submode_GUIDESRCH {
    my ($self,$results_file) = shift;
    $self->DoSearch($results_file) or 
       $self->andiamo("Can't start guide search");
}
#------------------------------------------------------------
# submode_GUIDEGET
#------------------------------------------------------------
#
sub submode_GUIDEGET {
    my $self = shift;

    # Get needed CGI tags:
    my $daac     = $REQ->param('daac');
    my $docnum   = $REQ->param('docnum');
    my $srchfile = $REQ->param('srchfile');
    my $catalog  = $REQ->param('catalog');

    $self->DoRetrieve($daac, $docnum, $srchfile, $catalog)
        or puke "'Can't start guide doc retrieval!";
}
#------------------------------------------------------------
# go
# ------------------------------------------------------------
#
# DESCRIPTION
#
#    Determine the submode, the first time through it should be GUIDESRCH
#    once you've gotten a list of docs then its calls itself with submode
#    GUIDEGET to retrieve the docs.
#
sub go {
    my $self = shift;

    # Get session or die:
    $APP->get_session();
    # Dispatch submode
    $self->go_submode();
    1;
}

#-----------------------------------------------------------
# suppress_links
#-----------------------------------------------------------
# Set/get link suppression.

sub suppress_links {
    my ($self, $parm) = @_;
    $self->{SuppressLinks} = $parm if defined $parm;
    $self->{SuppressLinks};
}

sub proxy_graphics {
    # if this is a secure server, the browser won't show any
    # graphics that come from a non-secure source. Fool the
    # browser by rewriting the real URL to a fake URL pointing 
    # to our secure server. Our secure Apache server will rewrite 
    # this URL back again to the real location, but will then proxy 
    # it so the browser doesn't know it's showing an "insecure"
    # graphic. We check for graphics specifically since the browser'
    # doesn't mind showing off-the-page hyperlinks. 
    #
    # Note that if the guide documents get more complex (like
    # Javascript, or forms, or applets), we'll have to change this
    # code to handle that, but that's not likely to happen; we'll
    # bypass that for now.
    my $port = ($IU::SecurePort ? $IU::SecurePort : 443);
    local $_;
    $_ = shift;
    s{
      http://(.*)/                      # the real server
             (.*)                       # the rest of it	
             (.gif|.jpg|.jpeg)          # and it's a graphic
     }
     # prepend secure server name, port, and "proxy."
    {https://$ENV{SERVER_NAME}:$port/proxy.$1/$2$3}gx
          if $IU::SecureServer;         # only if it's a secure server
    return $_;
}
1;
#-------------------- End of file --------------------
