#------------------------------------------------------------
# ModeGUIDEXML.pm
#------------------------------------------------------------
package IW::ModeGUIDEXML;

=head1 NAME

IW::ModeGUIDEXML - XML-powered guide search.

=head1 DESCRIPTION

Searches for and returns guide documents.
Performs all the guide searches:

=begin enum

=item

Composed Guide Search using guide search form.

=item

Selecting "Get Detailed Information" for an earth science attribute
(valid) from the bottom of the attribute selection pages.

=item

"Detailed Information" buttons next to data sets in data set listings.

=end enum

=head1 CGI TAGS for SEARCH

=begin list

=item

B<"submode"> - gives the mode of operation:

=begin list

=item

B<GUIDESRCH> (default) - to initiate search for guide documents.

=item

B<GUIDEGET> - to retrieve and display the document selected or found.

=item

B<OUT_OF_LINE_GET> - to retrieve and display the document selected or found,
sans navigator.

=end list

=item

B<"sid"> - gives the session id.

=item

B<"doctype">  gives the document type.  It should be one of:

=begin list

=item

Campaign

=item

Data_Center

=item

Dataset

=item

Parameter

=item

Sensor

=item

Source

=end list

=item

B<"item"> - if searching, gives the item we are searching for.
This form may contain spaces: they will be turned to underscores
automatically by this script when generating the wais query.

=end list

=head2 ADDITIONAL CGI TAGS for RETRIEVAL

=begin list

=item

B<"daac"> - short string for data center from which to retrieve document.

=item

B<"doctype"> - same as in search.

=item

B<"srchfile"> - name of the query file produced (in the session directory)
when a search was done.

=item

B<"catalog"> - a flag indicating that the document is not a guide document
but is a data center's catalog of documents.

=end list

=head1 GuideType

The SETUP file contains a configuration variable for guide searches
called "GuideType".  It is one of 'NEW', or 'XML'. If it is 'XML',
ModeGUIDE automatically delegates the request to this module. 

=begin list

=item

XML search strings may be either B<text> or B<fielded> searches.
A fielded search uses a search string like 
C<parameter/"carbon" or platform/"landsat 7">.
A text search uses a search string like C<rain and snow>.

=end list

=head1 INVOCATIONS

Guide searches are currently (WG 2.1, July 1998) invoked in the
following ways:

=begin enum

=item

Composed Guide Search using guide search form.

=begin list

=item

When you fill in a Guide search form and click the Start Search button,
nph-cgi invokes IW::ModeSEARCH.pm's go(), which invokes start_search(),
which invokes start_guide_search(), which switchs to ModeGUIDEXML,
submode GUIDESRCH, with the query object constructed from the
form as a CGI parameter.

=end list

=item

Selecting "Detailed Document" for an earth science attribute
(valid) from the CGI attribute selection pages.

or

Using "Detailed Document" buttons next to data sets in data set listings.

or

Clicking the "Detailed Document" button in the dependent valids applet.

=begin list

=item

All of these will invoke IW::ModeGUIDEXML::go() without a query object.
The cgi form will pass the appropriate parameters to search.  go()
invokes submode_GUIDESRCH() which invokes DoSearch().

If the invocation is via the applet, a special submode should be supplied 
(OUT_OF_LINE) which 

=begin list

=item 

causes the first page shown to suppress the navigator

=item

causes all links pointing off this page to use the OUT_OF_LINE_GET rather than 
GUIDEGET submode in retrieving the pages (the navigator will be
suppressed here too).

=end list

=end list

=item

Selecting a document link from the list of returned matching guide
documents from a guide search.

=begin list

=item

This invokes IW::ModeGUIDEXML::go(), which invokes submode_GUIDEGET(),
which invokes DoRetrieve(). An XML guide search returns actual links, 
DoRetrieve is less complex now.

submode_OUT_OF_LINE_GET() is used if the original page had its navigator 
suppressed.

=end list

=item

As of WG 1.9, if a search finds just one document, DoSearch() will
directly call DoRetrieve() to get the document.

=item

If a search finds no matching documents, the XML server return a
"no hits found message". There is no longer a catalog link available.

=end enum

=head1 WARNING

WE ASSUME THE CHARACTERS '=', '?', AND '|' WILL NEVER BE IN ANY VALID!

=head1 AUTHORS

Portions of this code were retrieved from (wais.pl) or son-of-wais.
Eryq adapted the code to fit in with the EOSDIS V0 Gateway distribution.

=begin list

=item

Douglas J. Hoffman, Hughes STX, 10/9/95

=item

Eryq (Erik Dorfman), Hughes STX

=item

Ed Seufert, Hughes STX, Aug 14, 1996

Extensively rewritten for web gateway release 1.5 to use the new
object oriented scheme and libraries created by Eryq.

=item

Larry Rosen, Hughes STX, 1997

=item

Joe McMahon, Raytheon STX, 1998

Added support for link suppression; several subroutine were turned into
methods to allow the link suppression to be "carried down" into the routines
in a relatively transparent way.

=item 

Joe McMahon, Raytheon ITSS, 2000

Cloned ModeGUIDE and rewrote to use SPIDEX searches. 
Old ModeGUIDE will go away after XML searches are implemented.

=item 

Joe McMahon, GST, 2001

Devised ModeGUIDE dispatching scheme; allows old guide code to remain
(relatively) untouched -- just module renaming. Both modes can coexist,
and new modes can be added easily.

=end list

=head1 COPYRIGHT

 h / Eryq (Erik Dorfman) 
NO WARRANTY; see TERMS AND CONDITIONS in distribution for details.

=head1 RCS INFO

$Id: ModeGUIDEXML.pm,v 1.19 2002/08/01 19:18:05 joe Exp $

=cut

use IU;
use IU::Log;
use IU::Session;
use IU::Event;
use IU::ODL;
use IU::Query;
use IU::Storable;
 
use IW::App;

use URI::Escape;
use IW::Util qw(:url :output);
use IU::Util qw(shellquote);

use XML::Parser;
use HTTP::Request::Common qw(GET HEAD POST);
use LWP::UserAgent;

use Data::Dumper;

use vars qw(@hits $results $state $debug $trace $n_hits $n_errs);

use strict; 

@IW::ModeGUIDEXML::ISA = qw(IW::Mode IW::ModeGUIDE);
#------------------------------------------------------------
#
# Globals
#
#------------------------------------------------------------

### The following globals are all required for the XML parser
### support. It might be cleaner-looking to make them package
### globals in a separate package and then import them, but 
### leaving them here seems to communicate what's happening better.

# List of hits, which is an array of hashes; each hash element represents
# the required data for a hit, an error, or a catalog entry.
@hits = ();

# Index into @hits, representing the result currently being processed.
$results = undef;

# Current parser state. The XML parser calls the characters() routine
# whenever it finds information between tags; we use $state to be able 
# to know exactly what it is we should do with the between-tags info
# when characters() is called. Initial state is "start" (we haven't
# done anything yet).
my $state = "start";

# Debugging. Prints out debugging info at each significant state.
# Initially off (and should stay off for prooduction use).
$debug = 0;

# Tracing. Prints out "we got in"/"we got out" info for each pair 
# of tags. Initially off (and should stay off for production use).
$trace = 0;

#------------------------------------------------------------
# new 
#------------------------------------------------------------
# I<Constructor.>
# Just builds an empty hash and blesses it into this class.
sub new {
    bless {}, shift;
}

#------------------------------------------------------------
# SearchString -- build XML search string 
#------------------------------------------------------------
#
# DESCRIPTION
#    Given the document type and the item (for a guide search)
#    or a constructed query, build the actual query sent to isearch.
#
# ARGUMENTS
#    If two arguments:
#       $doctype is the document type (as described up top)
#       $item is the item (as described up top)
#    Else one argument:
#       $query is a query object.
#
# RETURNS
#    Returns two strings.  The first is the query string to send
#    to isearch.  The second is the string to print on the results
#    page.  
#
sub SearchString {
    my $self=shift;
    my $the_string = '';    # The old style search string, or string to print.
    my $new_string = '';    # The search string converted to new codes.

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # If two arguments, item is specified.  This is a simple search
    # for a guide document for a valid from an attribute screen or
    # dataset from a dataset listing from an inventory search.
    # Just lower-case the item and send it back for a free-text search.
    if (@_ == 2) {
        my ($doctype, $item) = @_;
        $the_string = lc($item);
        return $self->validate_length($the_string, $the_string);
    }
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Else if one argument, a query object is specified; this is a more 
    # involved search. The search string must be constructed from the 
    # elements of the query object. 
    #
    elsif (@_ == 1) {
        my $query = shift;
        my @items = ();              # Items for which to search.
        my $text_string = '';        # Entered free text string if any.
        my $pretty_string = '';      # pretty search string to show user

        my @all_items = ();
        my @pretty_items = ();
        local $_;

        # Make up a fielded search. Note that we use an underscore-delimited
        # version of the valid as well as a space-delimited one; this is 
        # because the DAACs didn't always get it consistent in their META
        # tags.
        #
        # The new logic here builds lists of lists of items, to be 
        # assembled later into a query. This gets rid of special-case
        # code for the first item. 
        #
        # The "pretty" items simply get the items as they are; the
        # "all" items are coerced into a proper fielded search (both
        # one with the spaces between words and one without). If we don't
        # have any items for a given category in the query, we don't stack it.

        @items = $query->params;
        push @pretty_items, [map {qq("$_")} @items] if @items;
        push @all_items, [map {my $s = $_; 
                                  (!s/ /_/g 
                                     ? qq(parameter/"$s")
                                     : qq(parameter/"$s" or parameter/$_))
                              } @items]
           if @items;

        @items = $query->sensors;
        push @pretty_items, [map {qq("$_")} @items] if @items;
        push @all_items, [map {my $s = $_; 
                                  (!s/ /_/g 
                                     ? qq(sensor/"$s")
                                     : qq(sensor/"$s" or sensor/$_))
                              } @items]
           if @items;

        @items = $query->sources;
        push @pretty_items, [map {qq("$_")} @items] if @items;
        push @all_items, [map {my $s = $_; 
                                  (!s/ /_/g 
                                     ? qq(source/"$s")
                                     : qq(source/"$s" or source/$_))
                              } @items]
           if @items;

        @items = $query->datasets;
        push @pretty_items, [map {qq("$_")} @items] if @items;
        push @all_items, [map {my $s = $_; 
                                  (!s/ /_/g 
                                     ? qq(dataset/"$s")
                                     : qq(dataset/"$s" or dataset/$_))
                              } @items]
           if @items;

        @items = $query->campaigns;
        push @pretty_items, [map {qq("$_")} @items] if @items;
        push @all_items, [map {my $s = $_; 
                                  (!s/ /_/g 
                                     ? qq(campaign/"$s")
                                     : qq(campaign/"$s" or campaign/$_))
                              } @items]
           if @items;

        @items = $query->daacs;
        push @pretty_items, [map {qq("$_")} @items] if @items;
        push @all_items, [map {my $s = $_; 
                                  (!s/ /_/g 
                                     ? qq(data_center/"$s")
                                     : qq(data_center/"$s" or data_center/$_))
                              } @items]
           if @items;

        # If user entered a free text string as well, get it but don't do
        # anything with it yet. 
        $text_string = fixup_syntax($query->textstring);

        # If there was nothing in the array, no valids were selected.
        unless (@all_items) {
           # If no text string either, the query was empty.  Problem.
           if ($text_string eq '') {
              $self->andiamo("You must supply at least one of parameter, sensor, ".
	                "data set, source, campaign, data center, or a free ".
                        "text string.");
           }
           # Text string provided. Return it as the query string (this will
           # be just a free text search).
           else {
              return $self->validate_length($text_string, $text_string);
           }
        }
        # One or more valids were selected. Build a query out of them.
        else {
           # Run through the list of lists. Items in the sublists are
           # 'or'ed together; subarrays themselves are 'and'ed. This way
           # you get a query which will look for any of the items requested
           # within a category, but which will require that something be
           # found in each category requested.

           # Build the query string. We always parenthesize each category
           # to make sure that the query is unambiguous; since the only
           # thing that will see this string is the search engine, it doesn't
           # have to be pretty. It does, however, have to be rigorously
           # correct. The (possibly extra) parens around each category
           # string ensure this is true.
           $the_string = "";
           foreach my $list (@all_items) {
              $the_string .= " and " if $the_string;
              $the_string .= "( " . join(" or ", @$list) . " )";
           }
           # Now build the string that will be shown to the user. This one
           # should be built with the minimum number of parentheses because
           # it's best to eliminate as much visual noise as possible. We
           # still parenthesize the category lists if they contain more than
           # one item, because the query definition does not specifically
           # provide operator precedence. 
           $pretty_string = "";
           foreach my $list (@pretty_items) {
              $pretty_string .= " and " if $pretty_string;
              # More than one, parenthesize.
              if (@$list > 1) {
                 $pretty_string .= "( " . join(" or ", @$list) . " )";
              }
              # Only one, don't parenthesize.
              else {
                 $pretty_string .= $list->[0];
              }
           }
        }
        # If a free text string was provided too, append it with 'and' to 
        # both the search string and the string we'll show the user.
        $the_string .= ' and ( ' . $text_string . ' )' if $text_string;
        $pretty_string .= ' and ( ' . $text_string . ' )' if $text_string;
        #debug $the_string;
        #debug $pretty_string;
        
        return $self->validate_length($the_string, $pretty_string);
    }

    # Neither one argument nor two. Caller has messed up.
    else {
        puke "SearchString() function used incorrectly. Wrong number args.";
        return ("Unknown", "Unknown");
    }
}

#---------------------------------------------------------------------------
# validate_length(UGLY,PRETTY)
#---------------------------------------------------------------------------
# Verify that the search string isn't too long.
#
#   UGLY is the string that will be passed to Spidex.
# PRETTY is the user's string.
sub validate_length {
   my ($self,$ugly,$pretty) = @_;
   $self->andiamo("Your query is too complex. Please search on fewer items.")
      if length($ugly) > 512;
   return($ugly,$pretty);
}

#---------------------------------------------------------------------------
# fixup_syntax(STRING)
#---------------------------------------------------------------------------
# Fix the search string up so that its syntax matches what Isearch 
# will accept.
# - insert OR between words
# - strip quotes off quoted strings
sub fixup_syntax {
   my $text = shift;
   my @new  = ();
   # Break it up at word boundaries, honoring quoted strings
   push(@new,$+) while 
     $text =~ m{
      "([^\"\\]*(?:\\.[^\"\\]*)*)"\ ? # not a quote or escape, zero or more,
                                      # followed by zero or more escaped
                                      # strings and then not quote or escapes,
                                      # then at least one space, or
    | ([^\ ]+)\ ?                     # not a space, then a space, or
    | \                               # just a space
               }gx;
   # The array now has the string broken at proper space boundaries. We have
   # to extract the operators and add extra "and"s as needed to make this a
   # query that Spidex and Isearch will like.
   my @final = ();
   my $paren_count = 0;
   my $last_was = "";
   local $_;
   while (@new) {
      $_ = shift @new;
     /^(.+?)([\(\)])(.+)$/ and do {
         # Trapped paren (either flavor).
         # This is a word, one or the other paren, and "other stuff" (parens,
         # words, more trapped parens...
         # We split it all up at the paren, and restack everything.  
         unshift @new, $1, $2, $3;
         # We don't set last_was because the output stack is unaltered.
         next;
         };

      /^\($/ and do {
         # solo open paren. Stack and count.
         # If the last thing was a weord or close paren, we need an "or", too.
         $paren_count++;
         my %operation = ("word"  => qq(push \@final,"or",\$_),
                          "close" => qq(push \@final,"or",\$_),
                          "open"  => qq(push \@final,\$_),
                          "and"   => qq(push \@final,\$_),
                          "or"    => qq(push \@final,\$_),
                         );
         my $op = $operation{$last_was} or die "invalid token type: $last_was\n";
         eval $op;
         $last_was = "open";
         next;
         };

      /^\)$/ and do {
         # solo close paren. Stack and count.
         $paren_count--;
         push @final,$_;
         $last_was = "close";
         next;
         };

     /^\((.*)$/ and do {
         # Nestled open paren. Remove it, stack the paren, and
         # put the word back. Either there are more open parens,
         # in which case we come back here, or there aren't, in
         # which case we just process the word below.
         $paren_count++;
         push @final,"(";
         unshift @new,$1;
         $last_was = "open";
         next;
         };

     /^(.*)\)$/  and do {
         # Nestled close paren.
         # Since we're actually splitting off on the *right* this time,
         # we don't stack anything, but just put both tokens back on
         # the incoming stream. if there are still more close parens, 
         # We'll keep coming back here until they are all gone, at which
         # point we'll process the word normally and continue.
         unshift @new, $1, ")";
         # We don't set $last_was because we didn't touch the output stack.
         next;
         };
     
     /^and$/ and do {
         # explicit "and" operator. Stack.
         # Current final token is "and".
         push @final,"and";
         $last_was = "and";
         next;
         };

     /^or$/ and do {
         # explicit "or" operator. Stack.
         # Current final token is "or".
         push @final,"or";
         $last_was = "or";
         next;
         };

     /^.*$/ and do {
         # default case. Check last item stacked, and execute accordingly.
         my %operation = ("word"  => qq(push \@final,"or",\$_),
                          "open"  => qq(push \@final,\$_),
                          "close" => qq(push \@final,"and",\$_),
                          "and"   => qq(push \@final,\$_),
                          "or"    => qq(push \@final,\$_),
                          ""      => qq(push \@final,\$_),
                         );
         my $op = $operation{$last_was} or die "invalid token type: $last_was\n";
         eval $op;
         $last_was = "word";
         next;
         };
   }
   return wantarray ? @final : join(" ",@final);
}

#------------------------------------------------------------
# OutputDoc -- we've got all the info we need... output a document summary
#------------------------------------------------------------
#
# ARGUMENTS
#    $hit is a ref to hash of the document info extracted from the
#    waisq output.  It contains the following tags:
#
#       source   => data center where this hit came from
#       docid    => document number
#       score    => file score, 0-100
#
#    $last_relevance is a string or undef. If a string is provided,
#    OutputDoc will attempt to handle dividing the output into relevances.
#    if it's undef, OutputDoc assumes the caller is doing something else
#    and simply prints the information standalone as a list item.
#
#    %params is a list of optional parameters. Supported parameters:
#       Source: causes OutputDoc to print an extra "Datacenter:" line
#               for the entry.
#
# NOTES
#    The document is output as a list item.  It is the caller's
#    responsibility to ensure that there is a list environment surrounding
#    it.

sub OutputDoc {
    my ($self, $hit, $last_relevance, $daacfile, %params);
    if ($IU::GuideType eq 'TRANSITIONAL') {
       ($self, $hit, $last_relevance, $daacfile, %params) = @_;
    }
    else {
       ($self, $hit, $last_relevance, %params) = @_;
    }

    my $relevance = "";
    my $title = $hit->{'title'};

    # Spidex was returning URLs without the proper port; we try to fix this.
    # This code and the subroutine it calls cam be eliminated when this is
    # fixed.
    $hit = fix_url($hit);

    # If the last relevance is undefined, don't try to figure out
    # whether or not to print a header.
    if (defined $last_relevance) {   
        # Get relevance for the item we're processing now.
        $relevance = 
            ['Very low','Low','Medium','High','Very high','Highest'] ->
                [int($hit->{'score'}/20)];

        # If previous relevance was null, start a new list.
        unless ($last_relevance) {
            $HTML->H4->t("$relevance relevance")->_H4->OL;
        }
        # If relevance has changed between the last relevance and the
        # one we just calculated, close this list and open a new one.
        elsif ($last_relevance ne $relevance) {
            $HTML->_OL->H4->t("$relevance relevance")->_H4->OL;
        }
        # Same relevance as before; just output the item.
        else { 
            # No special HTML required.
        }
    }

    # Output an item. This will be:
    #  - a link to the document, opening in a new window specifically
    #    for guide documents (gets around problems with embedding a guide
    #    document inside an EDG window)
    #  - the document information, if we can get it.

    # Try to get the basic document info. We use HEAD because GET made the
    # script horrendously slow.
    my $ua = LWP::UserAgent->new();
    my $req = HEAD fix_url($hit)->{'documenturl'};
    my $response = $ua->request($req);
    my ($lines,$bytes);

    # If we got it, try to extract the data and output it.
    if ($response->is_success) {
      # Output the link.
      print "<LI><B>";
      $HTML->hypertext($title,$hit->{'documenturl'},TARGET=>'support');
      print "</A></B><BR>\n";
      
      # Print the datacenter name if it was requested.
      print "Datacenter: $hit->{'source'}" if $params{Source} and
                                                  !$self->suppress_daac();
      # Print relevance, if requested.
      print "Relevance: $relevance" unless $self->suppress_relevance();

      # If we got useful data, show it.
      if ($response->content_length > 0) {
        print ", " unless $self->suppress_daac();
        $hit->{'bytes'} = $response->content_length;
        print ", " unless $self->suppress_relevance;
        printf ("Size: %8.1d K", $hit->{'bytes'}/1024);
      }
      if ($response->last_modified > 0) {
        $hit->{'date'}  = scalar localtime $response->last_modified;
        print ", Last updated: $hit->{'date'}";
      }
    }

    # Return the current relevance; this allows the caller to maintain
    # state for us. The relevance can be passed back in on the next call
    # so that this routine can manage the lists for the caller.
    $relevance;
}

#------------------------------------------------------------------------
# ask_spidex(QUERY,DAACSREF,LIMIT)
#------------------------------------------------------------------------
#
# ask_spidex() passes the query off to the XML query engine.
#
#    QUERY is the query string, built by SearchString().
# DAACSREF is a reference to an array of datacenter names; used to
#             restrict the search to a set of daacs.
#    LIMIT is the guide document limit; Spidex is to return no more
#             than this number of documents.
#
# We use LWP to talk to Spidex  and get the XML back. We just return 
# what we get back to the caller.
sub ask_spidex {
    my ($self, $string,$daacs_ref,$limit) = @_;
    my $ua = LWP::UserAgent->new();
    $ua->timeout($IU::SpidexTimeout) if $IU::SpidexTimeout;
    my $request = POST $IU::SpidexUrl,
                      ['criterion'=>$string, 'daac'=>$daacs_ref, 
                       'hitlimit' =>$limit];
    $request->header('Content-Type' => 'text/xml');
    my $response = $ua->request($request);
    return ($response->is_success(), $response);
}

#---------------------------------------------------------------------
# display_by_relevance(XML,SEARCH_STRING,RAW_STRING)
#---------------------------------------------------------------------
# Displays the results by relevance.
#
#           XML is theXML returned by Spidex.
# SEARCH_STRING is the string that should be shown to the user.
#    RAW_STRING is search string that was sent to Spidex.
#
# display_by_relevance() shows the returned items using headers that
# tell the user what the relevance is. This routine handles the headers
# itself.

sub display_by_relevance {
    my ($self, $xml_string, $search_string, $raw_string) = @_;
    # output the results if the search worked.
    if($self->search_good($xml_string, $raw_string,$search_string)) {
        $APP->start_page(Title=>'Search results',
                         Context=>'guide_results');
        $self->by_daac_link();
        $HTML->t('The following documents were found:') -> P;
        # Sort items into descending order by score; we'll partition
        # them into relevance categories next.
        @hits = map {$_->[1]}
                sort {$b->[0] <=> $a->[0]}
                map {[$_->{'score'},$_]}
                @hits;
        
        # Partition the results. We divide the items into 6 bins by score.
        # We build a hash of lists, with the relevance as the hash key, and
        # the hits in the values. We use references to anonymous arrays to
        # handle the lists of values.
        my %per_relevance = ();
        my $relevance = 
            ['Very low', 'Low', 'Medium', 'High', 'Very high', 'Highest'];
        foreach (@hits) {
            next unless $_->{'score'};
            # Calculate the relevance for this item.
            my $this_rel = $relevance->[int($_->{'score'}/20)];
            # If there is already a list of items here, add this one
            # on to the end.
            if ($per_relevance{$this_rel}) {
                push @{$per_relevance{$this_rel}}, $_;
            }
            # If this is the first item, add it as a one-item list.
            else {
                $per_relevance{$this_rel} = [$_];
            }
        }
        # Output the data. Skip any relevances that we don't have anything for.
        # For the ones we actually do have, dump out a header and a list of
        # the items.
        my $head_out;
        foreach my $rel (reverse @$relevance) {
            my @local_hits;
            # Skip if this slot is empty.
            next unless ref $per_relevance{$rel} eq "ARRAY";
            # Sort the items in this slot by score, descending.
            # This way the highest-relevance stuff occurs toward
            # the top of each list.
            @local_hits = map { $_->[0] }
                          sort {$b->[1] <=> $a->[1]}
                          map { [$_, $_->{'score'}] }
                          @{$per_relevance{$rel}};
            # Note that we haven't output the header yet.
            $head_out = 0;
            # Run through the list of items in this relevance.
            foreach my $hit (@local_hits){
                # Skip error and catalog entries.
                next if $hit->{HitType} eq 'err' or $hit->{HitType} eq 'cat';
                # Output the header the first time we get here; subsequent
                # times, $head_out will be non-zero, so we'll skip printing.
                $HTML->H3->t("$rel relevance (", 
                             int @local_hits,
                             (int @local_hits eq 1 ? " hit)" : " hits)"))
                     ->_H3
                     ->P(STYLE=>"margin-left:5%")->OL
                       unless $head_out++;
                # Again, when Spidex is returning URLs with the port, 
                # we can take this out.
                $hit = fix_url($hit);
                # Note that we pass an undefined last_relevance, so
                # OutputDoc won't try to handle things itself. We add
                # the Source parameter to force the datacenter ID to be
                # added to each entry.
                $self->OutputDoc($hit,undef,Source=>1);
            }
            # If we actually output the header, we have to close it now.
            $HTML->_OL->_P if $head_out;
        }
        # Done with the list of items. Show the errors if there are any.
        $HTML->_UL;
        $self->show_errors_if_any;
    }
    # Print the user's query and a link to the raw XML display page.
    $self->finish_display($search_string);
}

#---------------------------------------------------------------------
# display_by_daac(XML,SEARCH_STRING,RAW_STRING)
#---------------------------------------------------------------------
# Displays the results by DAAC.
#
#           XML is the XML returned by Spidex.
# SEARCH_STRING is the string that should be shown to the user.
#    RAW_STRING is search string that was sent to Spidex.
#
# display_by_daac() is a little less complex that display_by_relevance()
# because OutputDoc() handles a lot of the list management for us.
sub display_by_daac {
    my ($self, $xml_string, $search_string, $raw_string) = @_;
    # Show stuff if the search worked.
    if($self->search_good($xml_string, $raw_string,$search_string)) {
        $APP->start_page(Title=>'Search results',
                         Context=>'guide_results');
        $self->by_relevance_link();
        $HTML->t('The following documents were found:') -> P;
        # Sort into descending order.
        @hits = map {$_->[1]}
                sort {$b->[0] <=> $a->[0]}
                map {[$_->{'score'},$_]}
                @hits;

        # Partition the results by the DAAC they came from this time.
        # We make the hash be indexed by the DAAC names, with the hits
        # in anonymous lists again.
        my %per_daac = ();
        local $_;
        # Stuff all the hits into the hash:
        foreach (@hits) {
            # Ignore anything that didn't come from anywhere.
            next unless $_->{'source'};
            # Add to list if list is there.
            if ($per_daac{$_->{'source'}}) {
                push @{$per_daac{$_->{'source'}}}, $_;
            }
            # Create new list if no list is there.
            else {
                $per_daac{$_->{'source'}} = [$_];
            }
        }
        # Grab each set of hits and print them out.
        foreach my $daac (keys %per_daac) {
            my @local_hits;
            @local_hits = @{$per_daac{$daac}};
            # Sort the items by rank; this works because we're within 
            # a single DAAC.
            @local_hits = map { $_->[0] }
                          sort {$a->[1] <=> $b->[1]}
                          map { [$_, $_->{'rank'}] }
                          @local_hits;
            # Header hasn't been output.
            my $head_out = 0;
            # Current relevance is null, NOT undef; this way, it will
            # always be different from the first computed relevance, but
            # is a defined value so that OutputDoc() will actually bother
            # to compute the relevance and handle the fields.
            my $current_relevance = "";
            # Do the hits for this DAAC. OutputDoc() handles printing
            # of the relevance headers.
            foreach my $hit (@local_hits){
                # Skip non-document hits.
                next if $hit->{HitType} eq 'err' or $hit->{HitType} eq 'cat';
                # Print the DAAC header the first time we get here, and
                # skip it every time thereafter.
                $HTML->H3->t("Results for $daac (", 
                             int @local_hits, 
                             (int @local_hits eq 1 ? " hit)" : " hits)"))
                     ->_H3
                     ->P(STYLE=>"margin-left:5%")
                         unless $head_out++;;
                # More code that will go away eventually.
                $hit = fix_url($hit);
                # We send in whatever the relevance is now; OutputDoc()
                # calculates the relevance for this hit and returns it.
                # Since we've sorted by rank, the relevances will always
                # be decreasing,so each relevance category header only
                # gets output once.
                #
                # Since the relevance comes back each time, we can maintain
                # state for OutputDoc() and it doesn't have to fiddle with
                # the object to maintain state.
                $current_relevance = $self->OutputDoc($hit,$current_relevance);
            }
            # Close up the item if it got opened.
            $HTML->_OL->_P if $head_out;
        }
        # Print the errors if we got any.
        $self->show_errors_if_any;
    }
    # Link to raw XML and the user's query.
    $self->finish_display($search_string);
}

#---------------------------------------------------------------------------
# search_good(XML,RAWQUERY,USERQUERY)
#---------------------------------------------------------------------------
#       XML is the XML returned by Spidex.
#  RAWQUERY is the query just as it was passed to the XML search engine.
# USERQUERY is the query as it is shown to the user.
#
# search_good() cleans up the returned XML and archives it for the 
# raw XML display page. It also parses the XML and builds the @hits array.
# If this works, it returns true. If not, it prints what diagnostics it can
# and returns false.

sub search_good {
   my ($self, $xml_string, $raw_string, $user_string) = @_;
   my ($state, $reason) = 
      $self->parse_results($xml_string,$raw_string,$user_string);

   # Parser failed (code problem, or bad XML). We output some diagnostic
   # info and give up.
   if ($reason =~ /XML parsing failed/) { 
      $APP->start_page(Title=>'XML parsing failed',
                       Context=>'guide_parse_failed');
      $HTML->t('The XML parsing process failed; it is impossible to show meaningful results.')
           ->P
           ->t('The XML was:') 
           ->BR 
           ->PRE ->t($xml_string) ->_PRE
           ->P 
           ->t('The error reported was:')
           ->BR
           ->PRE ->t($reason) -> _PRE;
    }

    # Parsing worked, but Spidex reported nothing but errors.
    # Print out a list of the document catalogs. 
    elsif ($n_errs and !$n_hits) {
       $APP->start_page(Title=>'No documents found',
                        Context=>'guide_none_found');
       $HTML->t('No valid results were returned from this guide search.')
            ->P;
       my @cats = map {$_->{'source'}} @hits;
       # debug "Catalogs: ",@cats;
       $self->show_cat('You might want to try the document catalogs:',@cats);
       $self->show_errors_if_any;
       # Do not try to print results.
       return 0;
    }

    # Nothing at all. Fetch the document catalog(s).
    elsif (!$n_hits) {
       $APP->start_page(Title=>'No documents found',
                        Context=>'guide_no_dox');
       my @daacs = map {$_->{'source'}} @hits;
       $self->show_cat('No documents matched your search criteria. Here are the document catalogs:',@daacs);
       # Don't try to show results.
       return 0;
    }
    # Successful parsing, and non-error results. Tell caller to show them.
    return 1;
}

#----------------------------------------------------------------------------
# parse_results XML
#----------------------------------------------------------------------------
# parse the results from Spidex.
#       XML is the XML returned by Spidex.
#
# parse_results() cleans up the returned XML and archives it for the 
# raw XML display page. It also parses the XML and builds the @hits array.
# If this works, it returns true. If not, it returns false and a diagnostic.

sub parse_results {
   my ($self, $xml_string,$raw_string,$user_string) = @_;
   $self->process_debug("entered parse_results");
   ### HACK/BUG: Spidex returns an "isearch:search" where it should return
   #             isearch:present. Fix it for now.
   $xml_string =~ s{/isearch:search}{/isearch:present}g;

   ### HACK/BUG: Some returned XML still has "bad" characters.
   #             We'll try to clean it up.
   $xml_string =~ s{&(.*?[^;])\b}{&amp;$1}gs;

   # Save the raw XML so it can be perused.
   my $xml_file = $SESSION->path_to() . 'guide_xml';
   open XMLOUT, ">$xml_file" or $self->andiamo("Can't save XML in $xml_file: $!");
   $raw_string =~ s/\n//g;
   $user_string =~ s/\n//g;
   print XMLOUT "User query: $user_string\n";
   print XMLOUT "Actual query: $raw_string\n\n";
   print XMLOUT $xml_string;
   close XMLOUT;
   chmod 0600, $xml_file;
   $self->process_debug("saved the XML");

   # Set things up to parse. It's not very neat, but neither is XML parsing.
   my $parser = new XML::Parser('Style'=>'Subs',
                                'ErrorContext'=>2,
                                'Namespaces'=>1);
   $parser->setHandlers('Char',\&characters);

   # List of hits, which is an array of arrays of hashes
   @hits = ();

   # Index of result currently being processed.
   $results = undef;

   # current state
   $state = "start";

   # debugging
   $debug = 0;
   $trace = 0;

   # result counting
   $n_hits = 0;
   $n_errs = 0;

   # Run the parser inside eval() because XML parsers die when the XML is bad.
   # We localize %SIG because (since XML::Parser 2.26) having a __DIE__ 
   # handler in place when we invoke parser() causes XML::Parser to error out
   # without attempting to parse the input. Grump.
   $self->process_debug("beginning parse");
   eval {local %SIG; $SIG{__DIE__} = undef; $parser->parse($xml_string)};

   # __DIE__ handler now back in force.
   #debug "Hits: $n_hits; errors: $n_errs";

   # Parser failed (code problem, or bad XML). We output some diagnostic
   # info and give up.
   if ($@) {
      $self->process_debug("oops - parse failed: $@");
      return (0,"Parsing failed: $@");
   }

   # Successful parsing, and non-error results. Tell caller to show them.
   else {
      $self->process_debug("parse finished OK");
      return 1;
   }
}
#----------------------------------------------------------------------------
# finish_display(QUERY)
#----------------------------------------------------------------------------
# QUERY is the "pretty" query to be shown to the user.
#
# finish_display() prints a link to the raw XML page and shows the query.
sub finish_display {
    my ($self,$search_string) = @_;
    $self->raw_xml_link;
    $HTML->P->t("Your query was:")->UL->LI->t($search_string)->_UL;
    $APP->end_page;
}

#----------------------------------------------------------------------------
# raw_xml_link()
#----------------------------------------------------------------------------
# Print a link to the raw XML page.
sub raw_xml_link {
   my $self = shift;
   $HTML->P->SMALL
        ->hypertext('View raw XML',
                    script_url('mode'=>'GUIDE','submode'=>'SHOWXML'),
                    TARGET=>'support')
        ->_SMALL;
}

#---------------------------------------------------------------------------
# by_daac_link()
#---------------------------------------------------------------------------
# Print a link to the "by daac" page.
sub by_daac_link {
   my $self = shift;
   my $link = script_url('mode'=>'GUIDE', 'submode'=>'SHOWBYDAAC');
   $HTML->P->A(HREF=>$link)->t('View by datacenter')->_A->BR;
}

#---------------------------------------------------------------------------
# by_relevance_link()
#---------------------------------------------------------------------------
# Print a link to the "by relevance" page.
sub by_relevance_link {
   my $self = shift;
   my $link = script_url('mode'=>'GUIDE', 'submode'=>'SHOWBYRELEVANCE');
   $HTML->P->A(HREF=>$link)->t('View plain list')->_A->BR;
}

#--------------------------------------------------------------------------
# extract_cats DAACS
#--------------------------------------------------------------------------
# Fetch the document catalog.
# DAACS is the list of DAACs to select; all are returned if no list is given.
# Returns a list of fake hits that can be displayed or otherwise used.
sub extract_cats {
  my ($self,@daacs) = @_;
   # Grab the Spidex config file; it has the document catalog links in it.
   my $ua = LWP::UserAgent->new();
   # debug "getting catalogs from $IU::SpidexConf";
   my $request = GET $IU::SpidexConf;
   my $response = $ua->request($request);
   # We got the catalog. Split it into lines and extract the URLs we want.
   # Save these in a hash.
   if ($response->is_success) {
      # debug "fetched catalogs successfully";
      my @a = split /\n/,$response->content;
      my %spidex_cat;
      foreach my $a (@a) {
         my ($k,$junk,$v,$junk2) = split /\s+/,$a;
         $spidex_cat{$k} = $v;
      }
      my @hits = ();
      my $n = 0;
      foreach my $daac (sort keys %spidex_cat) {
         # Create a fake hit with the catalog data in it.
         my $h = {'docid' => 'catalog',
                  'rank'  => $n++,
                  'score' => '-1',
                  'source' => 'Spidex',
                  'documenturl' => $spidex_cat{$daac},
                  'title' => "$daac guide catalog",
                 };
         push @hits,$h;
       }
      return @hits;
   }
   # No catalog. Return an empty list.
   else {
     # debug "failed to get catalogs: " . $response->status_line;
     return ();
   }
}

#--------------------------------------------------------------------------
# extract_daacs
#--------------------------------------------------------------------------
# Fetch the list of DAACs under spidex management.
sub extract_daacs {
  my ($self) = @_;
   # Grab the Spidex config file; it has the document catalog links in it.
   my $ua = LWP::UserAgent->new();
   # debug "getting catalogs from $IU::SpidexConf";
   my $request = GET $IU::SpidexConf;
   my $response = $ua->request($request);
   # We got the catalog. Split it into lines and extract the URLs we want.
   # Save these in a hash.
   if ($response->is_success) {
      # debug "fetched catalogs successfully";
      my @a = split /\n/,$response->content;
      my @spidex_daacs = ();
      foreach my $a (@a) {
         my ($k,$junk,$v,$junk2) = split /\s+/,$a;
         push @spidex_daacs,$k;
      }
      return @spidex_daacs;
   }
   # No catalog. Return an empty list.
   else {
     debug "failed to get daac list: " . $response->status_line;
     return ();
   }
}

#--------------------------------------------------------------------------
# show_cat(MESSAGE,DAACS)
#--------------------------------------------------------------------------
# Show the document catalog, printing the appropriate message.
#
# MESSAGE is the message to be displayed before the list of documents.
#   DAACS is the list of DAACs to fetch catalogs for.

sub show_cat {
   my ($self,$string,@daacs) = @_;
   # Extract the catalog as a list of fake hits.
   my @hits = $self->extract_cats(@daacs);

   # We got the catalog. Split it into lines and extract the URLs we want.
   # Save these in a hash.
   if (@hits) {
      # Format and print.
      $HTML->B->t($string)->_B->P;
      my $n = 0;
      # Haven't opened the list of DAACs yet.
      my $list_open = 0;
      foreach my $h (@hits) {
         # output the <UL> the first time we get here; skip thereafter.
         $HTML->UL unless $list_open++;
         # Print the entry and don't do anything fancy with it.
         $self->OutputDoc($h,undef);
      }
      # Close the list if it was opened.
      $HTML->_UL if $list_open;
   }
   # Couldn't fetch the Spidex config file. Temporize.
   else {
      $HTML-> B -> T("Guide catalog(s) temporarily unavailable") -> _B;
   }
}

#-----------------------------------------------------------------------------
# fix_url(URL)
#-----------------------------------------------------------------------------
# Clean up URLs if needed. This is temporary until all the DAACs have given
# us document collections.
#
# URL is, well, a URL.

sub fix_url {
   my $hit = shift;
   ### HACK: the documenturl we get back lacks the port in some cases.
   ###       we use the following hash to clean those up as best we can 
   ###       for now.
   my %url_remap = ( 'http://scrapple.gsfc.nasa.gov/' =>
                     'http://scrapple.gsfc.nasa.gov:81/',
                     'http://dosxx.gsfc.nasa.gov/'    =>
                     'http://dosxx.gsfc.nasa.gov:81/',
                   );
   my ($old,$new);
   while (($old,$new) = each(%url_remap)) {
      next if $hit->{'documenturl'} =~ s/$old/$new/;
   }
   $hit;
}

#-----------------------------------------------------------------------------
# show_errors_if_any()
#-----------------------------------------------------------------------------
# This method scans through the hits and prints out any errors that were 
# encountered. 
#
# Error 202 is actually not an error; it's "I found nothing for your search."
# We skip that one.

sub show_errors_if_any {
    my $self = shift;
    my $head_out = 0;
    # Some errors were found.
    if ($n_errs) {
       # Scan over all the hits.
       foreach my $hit (@hits){
         # Skip real hits and 'nothing found' messages.
         next if $hit->{HitType} eq 'doc' or $hit->{'code'} eq '202';
         # Don't print sourceless errors.
         next unless $hit->{'source'};
         # print the header the first time only.
         $HTML->H2->t('Errors')->_H2
              ->t('The following errors were reported:') -> P
              ->UL
           unless $head_out++;
         # print the error.
         my $db = "(Database: " . $hit->{'source'} . ")";
         print <<EOF;
<li><B>$hit->{'msg'}</B> $db
<P>
EOF
       }
    }
    # Close the list if it got opened.
    $HTML->_UL if $head_out;
} 

#------------------------------------------------------------------------------
# query_these(QUERY)
#------------------------------------------------------------------------------
# If the query has data center IDs selected, return them; else return the
# fake datacenter ALL to force a search everywhere.
sub query_these {
   my ($self,$query) = @_;
   my @items = $query->data_center_ids;
   @items = qw(ALL) unless @items;
   @items;
}

#------------------------------
#
# SUBMODE HANDLERS 
#
#------------------------------

#------------------------------------------------------------
# submode_DEFAULT
#------------------------------------------------------------
sub submode_DEFAULT {
    my ($self, $which, %params) = @_;

    my ($worked,$response) = ();
    my $access;
    my ($send_string,$show_string) = ();
    my ($content,$status) = ();

    # If REREAD parameter is defined, read the raw XML back out of the
    # file where we saved it and skip running the search.
    if ($params{Reread}) {
       # debug "trying to reread the data";
       my $access = "retrieval";
       my $xml_file = $SESSION->path_to() . 'guide_xml';
       unless (open XMLIN, "$xml_file") {
          $status = "Can't read XML in $xml_file: $!";
          $content = "";
          $worked = 0;
       }
       else {
          my @xml = <XMLIN>;
          debug "$#xml lines read";
          $show_string = $xml[0];
          $show_string =~ s/^.*:\s+//;
          $send_string = $xml[1];
          $send_string =~ s/^.*:\s+//;
          $status = "You should not see this.";
          $content = join "",@xml[3..$#xml];
          $worked = 1;
          close XMLIN;
       }
    }
    # This is the first time around. Run the search.
    else {
       $self->process_debug("First time in, running search");
       my $access = "search";
       # If query is present, then use it to construct the search string
       my $query = $REQ->param('query');
       my @daacs = ();
       my $limit = (defined $query && $query->max_guides) || 
                   $IU::GuideLimit || 100;

       # Else get doctype and item from CGI.
       unless (defined $query) {
           # Get the necessary CGI values:
           my $doctype = $REQ->param('doctype');
           my $item    = $REQ->param('item');
           $query = IU::Query->new();
           $query->textstring(lc $item);
       }
       ($send_string, $show_string) = $self->SearchString($query);
       @daacs = $self->query_these($query);
       debug "search string was: '$send_string'";

       $self->process_debug("ready to XML search");
       ($worked, $response) = $self->ask_spidex($send_string,\@daacs,$limit);
       $self->andiamo($response->status_line) unless $worked;
       $content = $response->content;
       $status  = $response->status_line;
       debug "status from Spidex was $status";
       $self->process_debug("spidex said $status");
    }

    # If we're running TRANSITIONAL mode, just pass back the query, hits, 
    # and document catalogs.
    if (uc($IU::GuideType) eq "TRANSITIONAL"){
      $self->process_debug("about to parse XML");
      my($state,$reason) = 
          $self->parse_results($content,$send_string,$show_string);
      my @cats = $self->extract_cats();
      debug $show_string . " state=$state, reason=$reason; ". int(@hits) . 
            " hits, " . int(@cats) . " cats";
      $self->process_debug("building XML results");
      $self->andiamo(undef, {Query => $send_string,
                             Display => $show_string,
                             Hits  => \@hits,
                             Cats  => \@cats
                            });
      return 1;
    }

    # Not TRANSITIONAL; actually show the results.
    else {
      # If the search (or reread) worked, display the results.
      if ($worked) {
        $self->$which($content, $show_string, $send_string);
      }
      else {
        $APP->start_page;
        puke "XML $access failed: ", $status;
        $APP->end_page;
      }
    }
    $APP->end_page;
    1;
}

#------------------------------------------------------------
# submode_SHOWXML
#------------------------------------------------------------
#
sub submode_SHOWXML {
   my $self = shift;
   my $xml_file = $SESSION->path_to() . 'guide_xml';
   open XMLIN, "$xml_file" or puke "Can't read XML in $xml_file: $!";
   my @xml = <XMLIN>;
   shift @xml; # drop user query
   # join all the xml up into a string; split it on ><, and insert
   # a newline between pairs of >< only.
   @xml = map {/></?">\n<":$_} split(/(><)/,join("",@xml));
   close XMLIN;
   $APP->suppress_links(1);
   $APP->start_page(Title=>"Raw XML from search",
                    Context=>'guide_raw_xml');
   $HTML->PRE->t(@xml)->_PRE;
   $APP->end_page;
}
#------------------------------------------------------------
# submode_GUIDESRCH
#------------------------------------------------------------
#
sub submode_GUIDESRCH {
    my $self = shift;
    $self->submode_DEFAULT("display_by_daac")
       or $self->andiamo("Can't start guide search");    
}

sub submode_SHOWBYDAAC {
    my $self = shift;
    $REQ->param('reread',1); 
    $self->submode_DEFAULT("display_by_daac",Reread=>1);
}

sub submode_SHOWBYRELEVANCE {
    my $self = shift;
    $REQ->param('reread',1); 
    $self->submode_DEFAULT("display_by_relevance",Reread=>1);
}


#------------------------------------------------------------
# go
#------------------------------------------------------------
#
# DESCRIPTION
#
#    Determine the submode, the first time through it should be GUIDESRCH
#    once you've gotten a list of docs then its calls itself with submode
#    GUIDEGET to retrieve the docs.
#
sub go {
    my $self = shift;

    # Get session or die:
    $APP->get_session();
    # Dispatch submode
    $self->go_submode();
    1;
}

#-----------------------------------------------------------
# suppress_links
#-----------------------------------------------------------
# Set/get link suppression.

sub suppress_links {
    my ($self, $parm) = @_;
    $self->{SuppressLinks} = $parm if defined $parm;
    $self->{SuppressLinks};
}

sub proxy_graphics {
    # if this is a secure server, the browser won't show any
    # graphics that come from a non-secure source. Fool the
    # browser by rewriting the real URL to a fake URL pointing 
    # to our secure server. Our secure Apache server will rewrite 
    # this URL back again to the real location, but will then proxy 
    # it so the browser doesn't know it's showing an "insecure"
    # graphic. We check for graphics specifically since the browser'
    # doesn't mind showing off-the-page hyperlinks. 
    #
    # Note that if the guide documents get more complex (like
    # Javascript, or forms, or applets), we'll have to change this
    # code to handle that, but that's not likely to happen; we'll
    # bypass that for now.
    my $port = ($IU::SecurePort ? $IU::SecurePort : 443);
    local $_;
    $_ = shift;
    s{
      http://(.*)/                      # the real server
             (.*)                       # the rest of it	
             (.gif|.jpg|.jpeg)          # and it's a graphic
     }
     # prepend secure server name, port, and "proxy."
    {https://$ENV{SERVER_NAME}:$port/proxy.$1/$2$3}gx
          if $IU::SecureServer;         # only if it's a secure server
    return $_;
}

#--------------------------------------------------------------------------
# Note:
# 
# The following subs are the XML::Parser callbacks used to actually 
# parse the XML. 
#
# These MUST NOT be renamed unless the XML DTD changes; if the XML
# changes, these routines will have to change also.

# New 'zpresent'. Nothing to do here.
sub zpresent { 
   debug "<zpresent>\n" if $trace;
   $state = "zpresent"; 
}

# New 'results'. Nothing to do here.
sub results  { 
   debug "<results>\n" if $trace;
   $state = "result";
}

# New 'result'. 
# If status = "OK", a hit:
#    Increment hit count.
#    Add space for the new 'field' values.
#    Grab all of the info about this field.
# If status = "Error", an error:
#    Increment error count.
sub result   { 
   debug "<result>\n" if $trace;
   $state = "result";
   my ($expat,$element,%attrs) = @_;
   push @hits,{};
   $results = $#hits;
   if ($attrs{'status'} eq 'OK') {
      $n_hits++;
      debug "Stacking new result ($n_hits)\n" if $debug;
      $hits[$results]->{HitType} = 'doc';
      local $_;
      foreach (keys %attrs) {
         $hits[$results]->{$_} = $attrs{$_};
      }
   }
   elsif ($attrs{'status'} eq 'Error') {
      $n_errs++;
      debug "Stacking new error ($n_errs)\n" if $debug;
      $hits[$results]->{HitType} = 'err';
      $hits[$results]->{'rank'}  = -1;
      local $_;
      foreach (keys %attrs) {
         $hits[$results]->{$_} = $attrs{$_};
      }
   }
}

# database-level error.
sub present {
   debug "<present>\n" if $trace;
   $state = "present";
   my ($expat,$element,%attrs) = @_;
   push @hits,{};
   $results = $#hits;
   $hits[$results]->{HitType} = 'err';
   $hits[$results]->{'source'} = $attrs{'dbname'};
   $hits[$results]->{'rank'}  = -2;
   # no DB doesn't count
   $n_errs++ if $attrs{'dbname'};  
}

# Note what kind of field this is so the value can be inserted properly.
sub field    { 
   debug "<field>\n" if $trace; 
   $state = "field";
   my ($expat,$element,%attrs) = @_;
   $hits[$results]->{Active} = $attrs{'name'};
}

# Set the 'field name' to capture the error text.
# We localize $^W to avoid a warning everytime this module is loaded.
do {
   local $^W = 0;
   sub error {
      debug "<error>\n" if $trace;
      $state = "error";
      my ($expat,$element,%attrs) = @_;
      $hits[$results]->{'code'} = $attrs{'code'};
      if ($attrs{'code'} eq '202') {
         $hits[$results]->{HitType} = 'cat';
         $n_errs--;
      }
   }
};



# No-op.
sub error_block {
   debug "<error_block>\n" if $trace;
   $state = "error_block";
}

# Capture error code and text.
sub error_text {
   debug "<error_text>\n" if $trace;
   $state = "error_text";
   my($expat,$element,%attrs) = @_;
   $hits[$results]->{'code'} = $attrs{'code'};
}

# Xlinks have the URL in them; we capture this into the appropriate field.
sub xlink    {
   debug "<xlink>\n" if $trace;
   $state = "xlink";
   my ($expat,$element,%attrs) = @_;
   # hits->current->current->current($attrs{'ref'}
   # $attrs{'ref'} =~ s/htmlgd$/html/;
   $hits[$results]->{$hits[$results]->{Active}} = $attrs{'ref'};
}

# End of the field. Delete the "Active" field name so it's not hanging around.
sub field_   { 
   debug "</field>\n" if $trace;
   $state = "_field";
   delete $hits[$results]->{Active};
}

# Nothing to do here.
sub error_block_ {
   debug "</error_block>\n" if $trace;
   $state = "_error_block";
}

# Nothing to do here.
sub error_text_ {
   debug "</error_text>\n" if $trace;
   $state = "_error_text";
}

# Nothing to do here.
sub error_ {
   debug "</error>\n" if $trace;
   $state = "_error";
}

# Nothing to do here.
sub result_  { 
   debug "</result>\n" if $trace;
   $state = "_result";
   my ($expat,$element,%attrs) = @_;
   if ($debug) {
     local $_;
     foreach (keys %{$hits[$results]}) {
       debug "$_ -> $hits[$results]->{$_}\n";
     }
   }
}

# Nothing to do here. Print some debug at most.
sub present_ {
   debug "</present>" if $trace;
   $state = "_present";
   my ($expat,$element,%attrs) = @_;
   if ($debug) {
     local $_;
     foreach (keys %{$hits[$results]}) {
       debug "$_ -> $hits[$results]->{$_}\n";
     }  
   }  
}

# Nothing to do here.
sub results_  { 
   debug "</results>\n" if $trace;
   $state = "_results";
}

# Sort all the hits by DAAC and pigeonhole.
sub zpresent_  { 
   debug "</zpresent>\n" if $trace;
   $state = "_zpresent";
}

# Capture the text if needed.
sub characters {
   my($xp,$txt) = @_;
   # Skip all-blank text
   return if $txt =~ /^ *$/;
   local $_;
   for ($state) {
      # Capture the field contents into the 'Active' field.
      (/field/) and
         do {
           # hits->current->current->current($txt)
           $hits[$results]->{$hits[$results]->{Active}} = $txt;
           next;
         };
      (/error_text/ or /error/) and
         do {
           $hits[$results]->{'msg'} = $txt;
           next;
         };
      # No other cases need text capture.
   }
}

1;
