package IW::ModeORDER;

=head1 NAME

IW::ModeORDER - order granules

=head1 RCS INFO

$Id: ModeORDER.pm,v 1.155 2003/09/08 17:28:49 solomon Exp $

=cut
use File::Copy 'cp';
use File::stat;

use IU;
use IU::Auth;
use IU::Util qw(touch);
use IU::Log;
use IU::InvStore ();
use IU::Granule ();
use IU::Dataset ();
use IU::Mail;
use IU::Event;
use IU::Package;
use IU::StatusCodes;
use IU::TrackedOrder;

use IW::App;
use IW::Mode;
use IW::Util qw(:url);
use IW::Order;
use IW::User;
use IW::ModePREFS;
use IW::Widget::URLTable;
use IW::Client;
use IW::CommStat;
use IW::Comm::Manager;
use IW::Comm::ProductResult;

@ISA = qw(IW::Mode IW::Comm::Manager);

use strict;

#------------------------------------------------------------
#
# Globals...
#
#------------------------------------------------------------

# Inventory store:
my $CART;

#------------------------------------------------------------
# login
#------------------------------------------------------------
# Need to log into the cart database to access the shopping cart

sub login {
    # Start by checking to see if the cart exists.
    # If it does not, then the user deleted it and then reloaded this
    # page. Duh.  Try and give them a helpful message.
    if (! -e $SESSION->path_to('cart.gdb')) {
        puke 533, "There's no shopping cart... did you attempt to ".
                  "reload this page after (1) emptying your shopping".
                  " cart, or (2) submitting an order which did so?";
    }

    # Log into CART database: We need write access since we may need 
    # to delete items from the shopping cart if the user wants the 
    # shopping cart deleted but some orders were not sucessful

    $CART = IU::InvStore->open($SESSION->path_to('cart.gdb'),'WRCREAT')
            || puke "Couldn't open shopping cart... was it deleted?";
    1;
}

#------------------------------------------------------------
# last_request_id
#------------------------------------------------------------
sub last_request_id {
    # Get the id of the last request:
    my $reqno = $SESSION->state->{OrderReqNo};
    return $SESSION->id . "-P$reqno"; 
}

#------------------------------------------------------------
# order_by_email
#------------------------------------------------------------
# Mail the orders we need to mail.

sub order_by_email {
    my $order = shift;

    # For each dataset...
    my $dsetid;
    my $email_no = 0;
    foreach $dsetid ($order->dataset_ids) {
        inform "sending email order to dataset $dsetid";
        ++$email_no;
        eval {        # try hard each time...
            # Get the dataset, using the granule of any item:
            my @items = $order->dataset_items($dsetid);
            my $dset   = $items[0]->orderable->dataset;
            my $archid = $items[0]->orderable->archive_id;

            # Get email addresses for DAAC and user:
            my $contact_addr = $dset->contact_addr ||
                puke "no contact address!";
            my $user_addr    = $order->user->req_addr->email;
            
            # Build a good ID for the subject line, so that requests
            # for data from different data sets don't have identical
            # ids:
            my $email_id = $order->id . "-E$email_no";

            # Convert this dataset's order to ODL:
            my $odl = $order->to_email_ODL($dsetid);

            # Create mail object:
            my $mail = new IU::Mail;
            $mail->to($contact_addr->email);
            $mail->from($user_addr);
            $mail->subject("$IU::SystemName email product request: $email_id");
            $mail->cc($user_addr);
            my $FH = $mail->open();
            $FH || puke "can't open mail stream";

            # Output a greeting:
            print $FH "Dear ", ($contact_addr->name || 'DAAC Contact'), ",\n";
            print $FH <<EOF;

This is a data request generated by the CINTEX Information Management System 
/ $IU::SystemName, with which you are a participating data center.

The information below describes the nature of the data request.  Specific
data ordering information may not be available to the IMS for the data
granules listed below, so if you need additional information on the
requested data media, format, or other order details, please contact
the user directly at (users's email address).

The data ordering information below is in Object Descriptor Language (ODL),
organized into various Groups.  The LINE_ITEM Group towards the bottom
contains the data set name (DATASET_ID) and the data Granules (PACKAGE_ID)
that the user would like to order from your data center.  If data granules
from this data set are not available, or if for any reason you are unable
to process this order, please notify the IMS development team at
${IU::SiteAdminEmail} so that the situation can be corrected.

Thank you for your time,
CINTEX Information Management System / $IU::SystemName Team

------------------------------------------------------------

EOF
            # Output the ODL:
            $odl->output($FH);

            # Send it!
            inform "sending mail to ", $contact_addr->email;
            $mail->send() || puke "couldn't send message";     
        };      
        $@ and $APP->note('WARN', "Couldn't email order: $@");
    }
    1;
}

#------------------------------
#
# USER RECEIPT...
#
#------------------------------

#------------------------------------------------------------
# print_receipt REQID,REQUESTED,EMAILED
#------------------------------------------------------------
# Make a single, merged shipping receipt for the user.
# Put in pub/order.txt.
# As of 2.2 we don't send this .

sub print_receipt {
    my ($reqid, $req_order, $email_order) = @_;
    my $totalCost = 0.0;

    # Open file:
    my $path = $SESSION->path_to("pub/order.txt");
    open RCPT, ">$path" or puke "couldn't open $path: $!";

    # Greeting:

    if (int($req_order->all_items)) {

    print RCPT <<EOF;
This is the FIRST of two electronic mail messages you will receive
from the $IU::SystemName. This message contains a list of the data 
products you requested from the $IU::SystemName.  Your order is 
currently being sent to the appropriate data centers for processing. 
 
After your order has been sent to the appropriate data centers I 
will send to you a SECOND (FOLLOWUP) message with the response from 
each data center. If you do not receive the FOLLOWUP message from 
me, then something went wrong with your order: please contact the 
developers for help at imswww-dev\@harp.gsfc.nasa.gov.


EOF

    } else {

    print RCPT <<EOF;
This message contains a list of the data products you requested from 
the $IU::SystemName.  Your order has been sent to the appropriate 
data centers for processing.


EOF
    }
    # Get user info:
    my $user = $SESSION->state->{User} || puke "no user info!";    

    # The summary:
    IW::Order->print_receipt(\*RCPT,
                             User      => $user,
                             RequestId => $reqid,
                             Requested => $req_order,
                             Emailed   => $email_order);
    # Done!
    close RCPT;
    1;
}

#------------------------------------------------------------
# email_receipt REQID
#------------------------------------------------------------
# Send the receipt for the current order.
# This is the FIRST email message the user gets from us.
# As of 2.2 we don't send this 
sub email_receipt {
    my $reqid = shift;
    
    # Get the current user preferences:
    my $user = $SESSION->state->{User} || puke "no user info!";
    
    # Create mail object:
    my $mail = new IU::Mail;
    $mail->to ($user->req_addr->email);
    $mail->subject ("$IU::SystemName product request: $reqid");
    my $MAIL = $mail->open();
    $MAIL || puke "can't open mail stream";
    
    # Print message body:
    $mail->insert ($SESSION->path_to("pub/order.txt")) ||
        puke "couldn't insert shipping receipt!";
    
    # Send it:
    $mail->send() || return undef;
    1;
}

#------------------------------------------------------------
# email_followup REQID, COMMSTAT
#------------------------------------------------------------
# Send the followup receipt for the current order, based on the comm 
# status screen.
# This is the ONLY email message the user gets from us.

sub email_followup {
    my $reqid = shift;
    my $commstat = shift;
   
    # Return right away if no COMMSTAT:
    $commstat or return undef;

    # Return right away if sent already:
    return 1 if $SESSION->state->{"EmailedFollowup:$reqid"};
    $SESSION->state->{"EmailedFollowup:$reqid"} = 1;

    # Get the client results:
    my %prs;
    eval {%prs =  IW::Comm::ProductResult->read_last } or
        do {%prs = ()};

    # Get the current user preferences:
    my $user = $SESSION->state->{User} || puke "no user info!";

    # Create mail object:
    my $mail = new IU::Mail;
    $mail->to ($user->req_addr->email);
    $mail->from ($IU::HomeEmail);
    $mail->subject ("$IU::SystemName product request");
    my $MAIL = $mail->open();
    $MAIL || puke "can't open mail stream";

    print $MAIL <<EOF;
This is a system generated email.  Please do not reply to it.  Contact
the appropriate Data Center User Services for correspondence.

This is your order receipt. After your request was submitted,
the following responses were returned from the various data centers:

EOF

    my $daac;
    my $line;
    if(defined %prs) {
        foreach $daac (sort keys %prs) {
            my $pr = $prs{$daac};
            my @order_ids;
            print $MAIL "\tDATA CENTER: $daac\n";
            #add daac order id if available, req id otherwise
            my $order_id = undef;
            if($pr->contact) {
               foreach my $prcontact ($pr->contact) {
                  if ($prcontact->daac_order_id) {
                     $order_id = $prcontact->daac_order_id ;
                     push(@order_ids, $order_id);
                  }
               }
            }
            if(int(@order_ids) < 1) {
               push(@order_ids, $reqid);
               print $MAIL "\t\tOrder ID Number: $reqid\n";
            }
            print $MAIL "\tSTATUS/COMMENTS: \n";
            my $commstat_str = $commstat->as_string_by_daac($daac);
            # Create text formatting object and format text.
            my $w = new Text::Wrapper('columns'=>60);
            $commstat_str = $w->wrap($commstat_str);
            foreach $line (split /\n/, $commstat_str) {
              print $MAIL "\t\t$line\n";
      
            }
            foreach my $reqResult ($pr->requestResults) {

               print $MAIL "\n";
               my $packageID = $reqResult->package_ID;
               print $MAIL "\t\tItem: $packageID\n";
               my $daacID = $reqResult->daac_request_ID;
               print $MAIL "\t\t\tData Center Request ID: $daacID\n";
               my $status = $reqResult->status_code eq '1' ? 'Success':'Failed';
               my $status_comment = $reqResult->status_code_comment ||
                   IU::StatusCodes::str($reqResult->status_code);
               my $comment = $status_comment ? ", $status_comment" : '';
               print $MAIL "\t\t\tStatus: $status$comment\n";
            }

            print $MAIL "\tCONTACT POINTS:\n";
            # DAAC contact info (if any):
            my $contact_str = undef;
            foreach my $contact ($pr->contact) {
               $contact_str = ($contact 
                              ? $contact->as_string
                              : "Contact info unavailable");  # if none
               $contact_str =~ s{^}{\t\t}mg;         # indent
               print $MAIL "$contact_str\n";


               # Blank line:
               print $MAIL "\n";
            }
        }
 
        # Sign the message:
        print $MAIL <<EOF;

If your order status shows "failed", please contact the data center immediately.
 
YOUR ORDER SUMMARY
------------------
EOF
        #add manifest to the end of the email
        my $filename = "pub/manifest/Manifest.$reqid";
        my $fullname = $SESSION->path_to($filename);

        open (MANIFEST,$fullname) 
           || puke 452, "can't open manifest $fullname: $!";
        while (<MANIFEST>) {      # read a line from MANIFEST into $_
           print $MAIL $_;        # print that line to MAIL
       }
        close (MANIFEST) || puke "can't close manifest $fullname: $!";
    } else {
       if ($commstat) {
            my $commstat_str = $commstat->as_string;
            print $MAIL "\tSTATUS/COMMENTS: \n";
            # Create text formatting object and format text.
            my $w = new Text::Wrapper('columns'=>60);
            $commstat_str = $w->wrap($commstat_str);
            foreach $line (split /\n/, $commstat_str) {
              print $MAIL "\t\t$line\n";
      
            }
       }
       print $MAIL "\n\tUser Support Contact Information URL:\n";
       print $MAIL "\t\t${IU::ImsUserUrl}";

    }
    # Send it:
    $mail->send() || return undef;
    1;
}
#------------------------------
#
# PAGE OUTPUT
#
#------------------------------

#------------------------------------------------------------
# print_section_head NUMBER, TITLEHTML
#------------------------------------------------------------
sub print_section_head {
    my ($no, $text) = @_;

    # Strip out number from title:
    $HTML -> P -> HR;
    $HTML -> FONT(SIZE=>7, COLOR=>'#006633') -> t_B("$no. ") -> _FONT if ($no);
    $HTML -> FONT(SIZE=>6);
    print $text, " ";
    $HTML -> _FONT -> BR;
}

#------------------------------------------------------------
# send_receipt_doc COMMSTAT
#------------------------------------------------------------
# Send the receipt for the current order.

sub send_receipt_doc {
    my ($commstat) = @_;
#    $SESSION -> state ->{timesordered} = 0;

    #------------------------------
    # LOAD INFO FROM LAST PRODUCT REQUEST...
    #------------------------------

    # Get the id of the last request:
    my $reqid = last_request_id();

    # Read the product request results:
    my %prs;
    eval {%prs =  IW::Comm::ProductResult->read_last } or
        do {%prs = ()};

    my $stat_check = 0;  #we assume all orders successful
    my (@succeeded_list);

#   If the user wants to delete the cart, call daac_status to see 
#   if orders to all of the daacs succeeded.

    if ($SESSION->state->{EmptyCart}) { # user wants to delete cart
       $stat_check = daac_status(%prs);
    }
    # Get the current user preferences:
    my $user = $SESSION->state->{User};
    
    #------------------------------
    # OUTPUT THE RECEIPT PAGE...
    #------------------------------
    # Set the indicator. We can do this any time, but this is as good
    # a time as any other.
    $SESSION->{Navigator}->{Indicator} = "OrderData";
   # Get all granules (and datasets) from the shopping cart:
    my @granules = $CART->granules;
    my @datasets = $CART->datasets;
    my @just_datasets = $CART->just_datasets; #orderable datasets

    # Quietly get rid of anything that's not orderable (KLUDGE):
    @granules = grep { $_->possible_to_order } @granules;

 
    # Construct the orders:
    my ($req_order, $email_order) = 
      IW::Order->make(Granules      => \@granules, 
                      Datasets      => \@datasets,
                      Just_Datasets => \@just_datasets);
    ($req_order || $email_order) or puke "couldn't make order!";

    # Print the page header (with HTTP header):
    $APP->start_page(Title      => 'Order Submitted',
                     SuperTitle => 'Shopping Cart',
                     Context    => 'ordersubm');

    # Current section (we number them manually):
    my $secnno = 0;

 

    # Message:
    ++$secnno;


     if (keys %prs) {
        print "Your order has been submitted to the appropriate data centers, ";
        print "and a copy of the ";

        $HTML -> hypertext("shipping receipt", 
                           script_url('mode'    => 'ORDER',
                                      'submode' => 'RECEIPT'));

        print " has been emailed\n";
        print "to the account <b>", $user->req_addr->email, "</b>.";
        $HTML -> P;
        print "NOTE: If you are ordering HDF and/or HDF-EOS formatted ";
        print " data (e.g. ASTER, MODIS, MISR, MOPITT, CERES), ";
        print " you may need HDF and HDF-EOS tools to read, view and ";
        print " manipulate the data. Please see NASA's ";
       # emit link for HDF viewing tools
        $HTML -> hypertext("Viewing HDF and HDF-EOS files","$IU::HdfTools");
        print " page for more information.";
        my $daac;

        foreach $daac (sort keys %prs) {
            my @order_ids;
            my $pr = $prs{$daac};
            my $order_id = undef;
   #------------------------------ 
   # TRACKING NUMBER
   #------------------------------
            if ($pr->contact) {
               #if a daac order id  is returned use it instead of the reqid
               foreach my $prcontact ($pr->contact) {
                  if ($prcontact->daac_order_id) {
                     $order_id = $prcontact->daac_order_id ;
                     push(@order_ids, $order_id);
                  }
               }
               my $archid = $pr->archive_id;
               $HTML -> P; 
               $HTML -> t_B("Data Center:  $archid ");
               $HTML -> BR;         
               if(int(@order_ids) < 1) {
                  push(@order_ids, $reqid);
                  $HTML -> t_B("Order ID Number: $reqid");
                  $HTML -> BR;
               }

             #save order in tracked order file for order status if order did not
             #fail
               my $num_dsets = int(@just_datasets);
               my $num_granules = int(@granules);
             
               if($pr->status_code == 1) {
                   my $tracked_order;
                   foreach $order_id (@order_ids) {
                      $tracked_order = IU::TrackedOrder->new(ArchiveId =>$archid,
					  OrderId =>$order_id,
					  DateSubmitted =>scalar(localtime),
                                          NumDatasets => $num_dsets,
                                          DataSets => \@just_datasets,
                                          NumGranules => $num_granules,
                                          Granules => \@granules);
                      $tracked_order->save_to_session;
                   }
               }

            }
   #------------------------------ 
   # INITIAL RESPONSES
   #------------------------------
            if ($commstat) {
               my $commstat_str = $commstat->as_string_by_daac($daac); 
               $HTML -> t_B("Status/Comments:   ");
               $HTML->BR;
               show_comment($commstat_str,'LINESIZE'=>75,'BREAK'=>'yes');
               $HTML -> BR;
             }
   #
   # Responses REQUEST_RESULT for each line_item, used by subsetting 
   # to comunicate partial sucess/failure of subsetting on a per line_item
   # basis.
   #

   if (int($pr->requestResults) > 0) {

     $HTML->TABLE(BORDER=>'1');
     
      foreach my $reqResult ($pr->requestResults) {
   
         $HTML->TR->TD; 
         my $packageID = $reqResult->package_ID;
         $HTML -> B ->t("Item: $packageID")->_B->BR; 
         $HTML->_TR->_TD;
         $HTML->TR->TD;
         my $status = $reqResult->status_code eq '1' ? 'Success':'Failed';
         my $status_comment = $reqResult->status_code_comment ||
            IU::StatusCodes::str($reqResult->status_code);
         my $comment = $status_comment ? ", $status_comment" : '';
         $HTML -> t("Status: $status$comment")->BR;
         my $daacID = $reqResult->daac_request_ID;
         $HTML -> t("Data Center Request ID: $daacID")->BR;
         if ( ($reqResult->URLs('DATA_URL')) 
            || ($reqResult->URLs('MISC_URL')) ) {
            my $tbl = new IW::Widget::URLTable
               "Data"              => [$reqResult->URLs('DATA_URL')],
               "Related site/page" => [$reqResult->URLs('MISC_URL')];
            $tbl->num_URLs;
            $tbl->print;
         } 
      $HTML->_TD->_TR;
      } 
   $HTML->_TABLE;
   $HTML->P;
   }

   #------------------------------ 
   # LINKS TO ACTUAL DATA
   #------------------------------             
            if ( ($pr->URLs('DATA_URL')) || ($pr->URLs('MISC_URL')) ) {
               $HTML -> B
                     -> t("Data Links/URLS: ")
                    -> _B;
                    my $tbl = new IW::Widget::URLTable
                    "Data"              => [$pr->URLs('DATA_URL')],
                    "Related site/page" => [$pr->URLs('MISC_URL')];
                    $tbl->num_URLs;
                    $tbl->print;
 
              }
              
    #------------------------------ 
    # CONTACT POINTS
    #------------------------------
             if($pr->contact) {     
                foreach my $prcontact ($pr->contact) {
                   $prcontact->print;
                }
              }
              else {
                   $HTML -> t_B("No contact information for $daac");
              }

              if ($SESSION->state->{EmptyCart}) { # user wants to delete cart
                  if($stat_check == 2){# some orders failed, some orders succeeded
#                 we can delete this one from the cart if order succeeded

                     if($pr->status_code == 1) { # order succeeded
                     # save archive id so we can delete order from cart
                        push(@succeeded_list, $pr->archive_id);
                     
                     }
                  }

              }

#             
#                     
# record the daac_order_id and the request_id in the same log file, if there
# is no daac_order_id the request_id will be entered as daac_order_id.

             foreach $order_id (@order_ids) {
                ims_event('ORDERESULT',
                          'daac'=>$pr->archive_id, 
                          'requestid'=>$reqid,
                          'daacorderid'=>$order_id);
             }
          } #end of for loop
           $HTML -> P; 
           print "If your order status shows \"failed\", please contact the data center 
     immediately.\n";
     

#        } end of if %pr


#-------------------------
# PRINT OUT EMAIL ORDERS
#-------------------------

         my $dsetid;
         my $email_no = 0;
         my $first_time = 1;
         my @email_dskeys;
         my @email_gkeys;

         foreach $dsetid ($email_order->dataset_ids){
           ++$email_no;
           # Get the dataset, using the granule of any item:
           if($first_time) {
               $HTML -> P; 
               $HTML -> FONT(SIZE=>'+1', COLOR=>'#003300')
                     -> t_B("E-mail Orders   ") -> _FONT; 
               $first_time = 0;
           }
 
           my @items = $email_order->dataset_items($dsetid);
           my $dset   = $items[0]->orderable->dataset;
           my $archid = $items[0]->orderable->archive_id;
           @email_gkeys = map{if($_->orderable->order_type eq "GRANULE"){$_->orderable->granule_key}} @items;

           push(@email_dskeys, $dset->dataset_key);

   #------------------------------ 
   # DATA CENTER AND TRACKING NUMBER
   #------------------------------
           my $email_id = $email_order->id;
           $HTML -> P; 
           $HTML -> t_B("Data Center   ")
                 -> t($archid);  
           $HTML -> BR;
           if($email_id) {         
               $HTML -> t_B("Order ID Number   ")
                     -> FONT(SIZE=>'+1', COLOR=>'#003300')
                     -> t($email_id) -> _FONT;
           }
  #------------------------------ 
  # STATUS/COMMENTS
  #------------------------------
           $HTML -> BR;
           $HTML -> t_B("Status/Comments:   ")
                 -> t("Order Sent") ;
           $HTML -> BR;
   #------------------------------ 
   # CONTACT POINTS
   #------------------------------               
           if ($dset->contact_addr ){     
                     $dset->contact_addr->print;
           }
           else {
                     $HTML -> t_B("No contact information");
           }
           $HTML -> BR;
           if ($SESSION->state->{EmptyCart}) { # user wants to delete cart
           #remove e-mail granules from the shopping cart
               if(@email_gkeys){
                   my @temp = $CART->remove_granules(\@email_gkeys);

               }
           }
         }

       
       # delete items from the shopping cart if order was successful
       if ($SESSION->state->{EmptyCart}) { # user wants to delete cart

       # go ahead and delete e-mail order datasets from cart
           if(@email_dskeys){
              my @temp = $CART->remove_just_datasets(\@email_dskeys);

            }

           if($stat_check == 2){# some orders failed, some orders succeeded

             # we can delete this one from the cart if order succeeded using
             # list of daacs where orders succeeded, granules in the order,
             # and orderable datasets (just_datasets).
             delete_items_by_daac(\@succeeded_list,\@granules,\@just_datasets);
           }
       } 
      } else { #prs is undefined
         print "An attempt was made to submit your order to the appropriate ";
         print "data centers, however ";
         print "no acknowledgements were received. You need to ";
         print "contact the data center directly to see if they got the order.";
 
           if ($commstat) {
               $HTML -> P;
               my $commstat_str = $commstat->as_string; 
               $HTML -> t_B("Status/Comments:   ");
               $HTML->BR;
               show_comment($commstat_str,'LINESIZE'=>75,'BREAK'=>'yes');
               $HTML -> BR;
               $HTML -> hypertext(" User Support Contact Information", 
                                                        ${IU::ImsUserUrl});
               $HTML -> BR;
             }

        
     }
    # The signature:
    $APP->end_page;
    return $stat_check;
}
#------------------------------------------------------------
# show_comment
#------------------------------------------------------------
# Format a comment nicely.
sub show_comment {
   my ($text, %param) = @_;
   $text or return;     # no-op if no text
   my $line;

   # Set up defaults
   my $linesize = ($param{'LINESIZE'} or '60');
   $param{'TEXTSIZE'} = "normal" unless defined $param{'TEXTSIZE'};
   $param{'BREAK'} = 0 unless defined $param{'BREAK'};

   # Create text formatting object and format text.
   my $w = new Text::Wrapper('columns'=>$linesize);
   $text = $w->wrap($text);

   # Output:
   $HTML -> small  if $param{'TEXTSIZE'} eq 'small';
   foreach $line (split /\n/, $text) {
      $HTML -> t($line);
      if ($param{BREAK}) {
         $HTML -> BR;
      }
      else {
         $HTML -> t(" ");
      }
   }
   $HTML -> _small if $param{'TEXTSIZE'} eq 'small';
   1;
}

#------------------------------------------------------------
# daac_status
#------------------------------------------------------------
# check the status of the orders. If orders to all daacs were successful,
# return 0 If orders to all daacs failed, return 1. If some orders succeeded
# and some orders failed, return 2.

sub daac_status {

     my %prs = @_;

    my $failed = 0;
    my $success = 0;
    my $stat_check = undef;
    if (keys %prs) {
        my $daac;
        foreach $daac (sort keys %prs) {
            my $pr = $prs{$daac};
           if($pr->status_code == 1){
              $success ++;
           }
           else {
              $failed ++;
           }
       }

       if ($failed == 0 && $success > 0) {
           $stat_check = 0;  # all successful
       }
       elsif ($success == 0 && $failed > 0) {
           $stat_check = 1; # all failed
       }
       elsif ($success > 0 && $failed > 0){
           $stat_check = 2; #mixed results
       }
   }

   return $stat_check;
}

#------------------------------------------------------------
# delete_items_by_daac
#------------------------------------------------------------
# go through list of daacs with successful orders, check each orderable 
# dataset and granule in the cart to see if it came from a 
# successful order. If so, we can delete it from the cart.

sub delete_items_by_daac {
    my($successful_daac, $granules, $just_datasets) = @_;


    my @jdkeys;
    my @gkeys;
    my $daac;
    my %datasets_hash;
    my $gran;
    my $dskey;
    my @dataset_keys;
    my $num_daacs = int(@$successful_daac);
    my $num_grans = int(@$granules);
    my $num_dsets = int(@$just_datasets);


    foreach $daac (@$successful_daac) {

# loop through each granule in the list and if the daac for
# this granule matches, add this granule to the list so it can be deleted
# from the shopping cart, then add to a hash the
# the dataset key for the dataset this granule belongs to 

        for (my $i = 0; $i < $num_grans; $i++) {
           $gran = $$granules[$i];
           if($daac eq $gran->archive_id) { 
               push(@gkeys, $gran->granule_key);
               $dskey = $gran->dataset_key;
               $datasets_hash{$dskey} = 1;
              
            }
        }

# turn the dataset keys into a list so they can be removed from the cart
       
         @dataset_keys = keys(%datasets_hash);

# loop through each orderable dataset and if the data center matches, 
# add the key for this dataset to a list so it can be 
# deleted from the shopping cart.

        for (my $i = 0; $i < $num_dsets; $i++) {
           if($daac eq $$just_datasets[$i]->archive_id) { 
               push(@jdkeys,$$just_datasets[$i]->dataset_key);  
            }
        }

        #remove granules from the shopping cart
        if(@gkeys){
            my @remgkeys = $CART->remove_granules(\@gkeys);
        }

        #remove datasets associated with the granules from the shopping cart
        if(@dataset_keys){
            my @remdkeys = $CART->remove_just_datasets(\@dataset_keys);
            
        }
        #remove orderable datasets from the shopping cart
        if(@jdkeys){
            my @remdkeys = $CART->remove_just_datasets(\@jdkeys);
            
        }
    }
}

#---------------------------------------------------------------------
# send_ecs_profile_req  IW::User
#---------------------------------------------------------------------
# Send the ecs profile request.
# Note that we already updated session User with CGI info.
# 1. Make sure we have ECS name & password.
# 2. Set the correct parameters for ModeECSPROFILE.
# 3. Switch control to ModeECSPROFILE.
# Input: The current session User object.  Output: none.

sub send_ecs_profile_req {
   my $user = shift;

   # Added an optional parameter to indicate next submode.
   my $next = shift;

   # Check if we have ECS name and password keys.
   my $keys = $user->keys;
   unless (defined ($keys) and $keys->ecs_username and
           $keys->ecs_authentication_key) {

      my $new_message = 'Notice: You did not supply BOTH your User
                Name and Password; your order will be submitted as a guest.';
      # Required keys are missing.  Warn user and redraw form.
      # If the current APP message contains "were accepted!" then no
      # errors before.  Replace message with ECS name/passwd error.
      # If there were already error messages, append new message.
      if ($APP->message && 
          $APP->message->as_string =~ /were accepted!/) {
         $APP->message (new IW::Note Class=>'WARN',
                        Text => $new_message);
      } elsif ($APP->message) {
      # There was an error message so append new message.
         my $text = $APP->message->as_string;
         $APP->message (new IW::Note Class=>'WARN',
                        HTML => "$new_message<BR>\n$text");
      } else {   # No previous error message
         $APP->message (new IW::Note Class=>'WARN',
                        HTML => "$new_message<BR>\n");
      }
      # Use the FORM submode to reprint the form.
      submode_FORM();

      return (1);
   }
   if ($APP->message) {
      unless($APP->message->as_string =~ /were accepted!/) {
         submode_FORM();
         return (1);
      }
   }
   # Set the parameters for ModeECSPROFILE.
   # If the next_submode is not in the param tags, perhaps it was passed
   # in as sub parameter "$next".
   my $next_submode = $REQ->param ('nextsubmode') || $next || '';

   # Verify next submode.
   unless ($next_submode eq 'UPDATE' or $next_submode eq 'RETRIEVE') {
      puke 600, "Invalid submode for ModeECSPROFILE: $next_submode";
   }
   # Switch control to ModeECSPROFILE with correct tags.
   my $daac = $IU::EcsProfileSmc;
   unless ($daac) { 
      my $new_message = 'Error: There is no ECS Profile SMC set.';
      if ($APP->message) {
         # There was an error message so append new message.
            my $text = $APP->message->as_string;
            $APP->message (new IW::Note Class=>'ERROR',
                           HTML => "$new_message<BR>\n$text");
      } else {   # No previous error message
            $APP->message (new IW::Note Class=>'ERROR',
                           HTML => "$new_message<BR>\n");
      }
      submode_FORM();
      return (1);
   } else {
      switch IW::Mode Tags=>{'mode'    =>'ECSPROFILE',
                             'submode' =>$next_submode,
                             'origin'  =>'ORDER'};
   }
   1;
}

#------------------------------
#
# COMM HANDLERS...
#
#------------------------------

#------------------------------------------------------------
# comm_preflight
#------------------------------------------------------------
# Do we need to use the client to submit the IMS order?  
#
# If we just had email orders, there's no need to search (we dispatched
# them already): we can move onto the 'OK' page.  
#
# Otherwise, we need to start an IMS order!
#
sub comm_preflight {
    my $self = shift;
    my $req_order   = $self->{ReqOrder};
    my $email_order = $self->{EmailOrder};
    return ($req_order->all_items ? 'GO' : 'OK');
}

#------------------------------------------------------------
# comm_start_search
#------------------------------------------------------------
# Start the IMS order.
# The order object is contained in the ORDER inst. var. of the mode.

sub comm_start_search {
    my $self = shift;
    my $req_order = $self->{ReqOrder};
    my $infile = $SESSION->path_to('order.odl');

    # Build the "order.odl" file, for input to the imsclient:
    open ORDER, ">$infile" or puke "open $infile: $!";
    $req_order->to_request_ODL->output(\*ORDER);
    IU::ODL->end(\*ORDER);
    close ORDER;

    # Start the client:
    my $client = start IW::Client Activity=>'PRODUCT_REQUEST', Infile=>$infile;
    $client or puke "couldn't start imsclient";

    # Redirect the "Order Data" button to here until COMMFINAL is reached.
    $SESSION->{Navigator}->field("OrderData",
                                 'Mode'=>'ORDER','Submode'=>'COMMSTAT');
    # Done!
    return 'STARTED';
}

#------------------------------------------------------------
# comm_finish_up
#------------------------------------------------------------
# Perform any desired post-processing after IMS order completes.
#
# NOTE: email_followup() contains a sentinel which prevents it
# from being invoked multiple times by accident.
#
sub comm_finish_up {
    my ($self, %params) = @_;

    # Send the followup mail message:
    email_followup(last_request_id(), $params{CommStat});

#    # Did they want the cart removed? If so, do it now!
#    if ($params{Status} eq 'OK') {
#       if ($SESSION->state->{EmptyCart}) {
#           unlink $SESSION->path_to("cart.gdb");
#           $APP->message("As you requested, the shopping cart HAS ".
#                         "been emptied out.");
#       }
#       else {
#           $APP->message("As you requested, the shopping cart HAS NOT ".
#                         "been emptied out.");
#       }    
#    }
#    else {
#       $APP->message("Since the request failed, the shopping cart ".
#                     "HAS NOT been emptied out.");
#    }

   # Redirect things BACK to the cart again now that the order's over.
   $SESSION->{Navigator}->field('OrderData',
                                'Mode'=>'CART','Submode'=>'');
    1;
}

#------------------------------------------------------------
# comm_ok_page
#------------------------------------------------------------
# IMS order succeeded.  Output the receipt.
#
sub comm_ok_page {
    my ($self, %params) = @_;

    my $stat_check = send_receipt_doc($params{CommStat});

    # Did they want the cart removed? If so, do it now!
    if ($stat_check == 0) {
        # all orders were successful
        if ($SESSION->state->{EmptyCart}) {

           unlink $SESSION->path_to('cart.gdb');
           $APP->message ('The shopping cart HAS '.
                          'been emptied.');
        }
        else {
           $APP->message ('The shopping cart HAS '.
                          'NOT been emptied.');
        }
    }
    1;
}

#------------------------------------------------------------
# comm_fail_page
#------------------------------------------------------------
# IMS order failed.  Output the receipt (yes, that's right!).
#
sub comm_fail_page {
    my ($self, %params) = @_;

    send_receipt_doc($params{CommStat});
    1;
}

#------------------------------
#
# SUBMODE HANDLERS...
#
#------------------------------

#------------------------------------------------------------
# submode_DEFAULT
#------------------------------------------------------------
sub submode_DEFAULT {

    shift->submode_FORM;
}

#------------------------------------------------------------
# submode_FORM
#------------------------------------------------------------
# I<Top level handler for submode "FORM".>
# Assemble the order, and output the final completion form!

sub submode_FORM {

    # Get all granules from the shopping cart:
    my @granules = $CART->granules;
    my @datasets = $CART->datasets;
    @datasets or
        puke 531, 'there are no datasets in the shopping cart!';
    my @just_datasets = $CART->just_datasets;

    # Quietly get rid of anything that's not orderable (KLUDGE):
    @granules = grep { $_->possible_to_order } @granules;

    # Construct the order:
    my ($req_order, $email_order) =
       IW::Order->make (Granules      => \@granules,
                        Datasets      => \@datasets,
                        Just_Datasets => \@just_datasets);
    ($req_order || $email_order) or puke "couldn't make order!";

    # Set the indicator. We can do this any time, but this is as good
    # a time as any other.
    $SESSION->{Navigator}->{Indicator} = 'OrderData';

    # Start page:
    $APP->start_page (Title      => 'Step 2: Order Form',
                      SuperTitle => 'Shopping Cart',
                      Context    => 'orderform');

    $HTML -> script_FORM;
    # There's a multi-submit button here...
    $HTML -> msubmit_flag;


    my $fail = $REQ->param ('fail');
    if($fail) {
       $HTML -> directions;
       $HTML -> t("You will need to fill in your user profile.");
       $HTML -> _directions;
    }

    # Directions:
    my $ReqFieldColor = '#D00000';
    $HTML -> BR;
    $HTML -> directions;
    print <<EOF;
       Please fill out your <b>Contact, Shipping, and Billing</b>                      information below. You can then review your entire order or                     immediately submit your order.  
EOF
    $HTML -> _directions;

    #------------------------------
    #------------------------------
    # MENU...

    # Start form:
    $HTML -> script_FORM;
    $HTML -> P -> HR (ALIGN=>'CENTER');

    # Get ids of ECS daacs orders are going to. 
    my @req_ecsdaacs = $req_order->ecs_archive_ids;
    my @email_ecsdaacs = $email_order->ecs_archive_ids;
  
    # Get number of ECS daacs orders are going to. 
    my $req_ecsdaacs_no = scalar (@req_ecsdaacs);
    my $email_ecsdaacs_no = scalar (@email_ecsdaacs);

    # Get current user info or make new.
    my $user = $SESSION->state->{User} || new IW::User;

    # Determine if order is going to ASF.  If yes, print message.
    my @daacs = ($req_order->archive_ids, $email_order->archive_ids);
    my %daac_check = map { $_ => 1 } @daacs;
    if (exists ($daac_check{ASF})) {
        $HTML -> directions;
        print <<EOF2;
          <P><B>IMPORTANT:</B> 
             Because your order is going to Alaska SAR Facility,
	     <B>ALL</B> fields must be filled in for a successful
	     order.
EOF2
        $HTML -> _directions;
    }
    # Check if any daacs require V0 authenticator.
    my @daacsV0 = IU::Auth::daacs_needing_v0_auth (\@daacs);

    # If yes, pass true Required flag to print method.
    my $auth_required = (@daacsV0 ? 1 : 0);

    # Check to see if any DAACs need NASDA authenticator
    my @daacsNASDA = IU::Auth::daacs_needing_nasda_auth (\@daacs);

    # If yes, pass true Required flag to from_CGI method.
    my $nasda_required = (@daacsNASDA ? 1 : 0);

    my $get_contact = $REQ->param ('ShowAddress');
    if($get_contact){
         set_prefs_from_edit();
    }
    # Print the user info part of form.
    $user->print ('FORM'=>'ORDER', 'REQUIRED'=>1,
                  'AUTHREQUIRED'=>$auth_required,
                  'NASDAREQUIRED'=>$nasda_required);

    submit_order_button ();
    $HTML -> HR -> H3
          -> t('Optional: Retrieve or Update User Profile...') -> _H3;
    $HTML -> t('You may want to retrieve your profile to restore it to');
    $HTML -> t(' to its original state after you have made changes.');
    $HTML -> t(' You may want to update your profile ');
    $HTML -> t('if you want to make the changes on this page permanent.');
    # Have user object print ECS fields and buttons.
    $user->print('FORM'=>'ECS', 'REQUIRED'=>0, 'ORIGIN'=>'ORDER');
    $HTML -> HR -> H3
         -> t('Save user preferences to a file...') -> _H3;
    print_download_form ();

    #-------------------------------
    # End form:
    #-------------------------------
    $HTML -> _script_FORM;
    
   # The upload form...
   $HTML -> HR -> H3
         -> t('Restore user preferences from a file ...') -> _H3;
   $HTML -> t("This may useful if the profile retrieved from EOS", 
              " is not the order profile you want for the current order.");
   print_upload_form ();
    #-------------------------------
    # End it and send it:
    #-------------------------------
    $APP->end_page;
    1;
}


#---------------------------------------------------------------------
# print_download_form
#---------------------------------------------------------------------
# Print the meat of the download form.

sub print_download_form {

   $HTML -> P(STYLE=>"margin-left:5%") -> TABLE ;
   $HTML -> TD;
   $HTML -> msubmit_button(CAPTION => 'Save Prefs',
                           QUERY   => [u       => time,
                                       sid     => $SESSION->id,
                                       mode    => 'ORDER',
                                       submode => 'DOWNLOAD']);
   $HTML -> _B -> _TD -> TD
         -> small
         -> t('You can save your user preferences ',
              'to your local machine. Click on the button ',
              'at the left .')
         -> _small;

   $HTML -> _TABLE -> _P;

   # End form:
   $HTML -> _FORM;
   1;
}

#---------------------------------------------------------------------
# print_upload_form
#---------------------------------------------------------------------
# Print the meat of the upload form.
# NOTE: unlike normal forms, this has an ENCTYPE of
# 'multipart/form-data'.
# DO NOT USE THIS ENCTYPE UNLESS YOU KNOW WHAT YOU ARE DOING!!!!!!

sub print_upload_form {
   # Start form:
   $HTML -> FORM (ENCTYPE => 'multipart/form-data',
                  ACTION  => IW::Util::script_action(),
                  METHOD  => 'POST')
         -> hidden (u       => time,
                    sid     => $SESSION->id,
                    mode    => 'ORDER',
                    submode => 'UPLOAD');

   # The directions:
   $HTML -> directions
         -> t("Enter the name of an $IU::SystemName user preferences",
              ' file to restore.')
         -> _directions;

   # The fields:
   $HTML -> INPUT (TYPE=>'file',
                   NAME=>'userfile',
                   SIZE=>45) -> BR
         -> INPUT (TYPE=>'submit',
                   VALUE=>
                   'Restore this file to the EOS Data Gateway now!')
         -> hidden('bogus' => 'This field fixes a CGI::BasePlus bug');

   # End form:
   $HTML -> _FORM;
   1;
}

#---------------------------------------------------------------------
# set_prefs_from_edit
#---------------------------------------------------------------------
# Set user preferences from an edit submission.  Gets all data
# from the CGI form and updates and installs User object in session.
# Sets error or success message to be printed.
# Input: none. Output: Return the session User object on success.
#                      Returns the 0 if error.

sub set_prefs_from_edit {

   my $require = shift;
   # Get user info from CGI form and install in session. 

   # If updating ECS profile, we require the standard fields.
   my $submode = ($REQ->param('submode') || 'SHOW');
   my $nextsubmode = $REQ->param('nextsubmode') || '';
   my $required = (($require || ($submode eq 'ECSPROFILE' &&
                    $nextsubmode eq 'UPDATE')) ? 1 : 0);

   #  User info (keys, addresses, affil).
   my $user = IW::User->from_CGI ('FORM'=>'ALL', 
                                  'REQUIRED'=>$required);

   # If any errors, set as APP message.
   my $err;
   if (($err = $user->errors) && $err->any_errors) {
      $APP->message (new IW::Note Class=>'ERROR',
                                  HTML =>join ("<BR>\n", $err->get));
      return (0);
   } else {
      $APP->message ('Your preferences ',
                     ($user->devel ? 
                      '(including developer preferences)' : ''),
                     ' were accepted! Scroll down to see/edit them.');
      return ($user);
   }
}

#---------------------------------------------------------------------
# download_prefs
#---------------------------------------------------------------------
# Download the user preferences file.

sub download_prefs {

   # The HTTP-reply header:
   my $head = new IW::ReplyHead;
   $head->okay if $REPLY->need_status_line;
   $head->content_type ('text/x-ims-prefs');
   $head->filename ('imsprefs.dat');
   $head->send (\*STDOUT);

   # Does it exist?
   (-e $SESSION->path_to ('prefs.dat')) or
      puke 524, "No preferences to save!";

   # The document:
   open PREFS, $SESSION->path_to('prefs.dat') or
      puke 452, "couldn't open prefs.dat: $!";
   print while <PREFS>;
   close PREFS;

   # Done!
   $REPLY->all_done;
   1;
}

#---------------------------------------------------------------------
# set_prefs_from_upload
#---------------------------------------------------------------------
# Set user preferences from an uploaded file.
# Updates the "User" object in the session state, *and* the session
# "prefs.dat" file.

sub set_prefs_from_upload {
   no strict 'refs';               # breaks file upload

   # Get userfile name.
   my $userfile = $REQ->param ('userfile');
   my $FH = "IW::App::$userfile";
   # Save the user file data to the session.

   my $uploaded = $SESSION->path_to ('prefs.dat.new');
   open (PREFS, ">$uploaded") or
      # Failed to open output file.  System error - Puke!
      puke 452, "Couldn't open new output $uploaded: $!";
   while (<$FH>) {
      print PREFS $_;
   }
   # If line number is >= 1000, warn user and don't use it.  Prefs
   # files should be small.  Reprint form.
   unless ($. < 1000) {
      $APP->message (new IW::Note Class=>'ERROR',
                     Text => "The file $userfile looks FAR too big to
                              be correct.  Please choose another file
                              to upload.");
      close PREFS;
      return (1);
   }
   close PREFS;

   # Did we get anything?  If not, warn and reprint form.
   unless (-s $uploaded) {
      $APP->message (new IW::Note Class=>'ERROR',
                     Text => "The file $userfile was empty or doesn't
                              exist.  Please choose another file to
                              upload.");
      return (1);
   }
   # Read the ODL in from the uploaded file:
   open (ODL, $uploaded) or puke 452, "Open $uploaded failed: $!";
   my $odl = parse IU::ODL \*ODL;
   close ODL;
   # Was ODL ok?  If not, warn and reprint form.
   unless ($odl) {
      $APP->message (new IW::Note Class=>'ERROR',
                     Text => "There was an error trying to parse the
                              file $userfile (near line $.).  Please
                              choose another file to upload.");
      return (1);
   }
   $SESSION->state->{Upload} = 1;
   # Create a user preferences object from the ODL:
   my $user = IW::User->from_ODL ($odl);

   # Check if we got a user object.
   unless ($user) {
      $APP->message (new IW::Note Class=>'ERROR',
                     Text => "The file $userfile does not appear to
                              be a user preferences file.  Please
                              choose another file to upload.");
      return (1);
   }
   # Check the user object for error messages.
   if ($user->errors->any_errors) {
      $APP->message (new IW::Note Class=>'ERROR',
                     HTML => "The file $userfile has the following
                              errors:<BR>" .
                     join ("<BR>\n", $user->errors->get));
   } else {
      # Done!
      $APP->message('Your preferences ',
            ($user->devel ? '(including developer preferences)' : ''),
                    ' were restored and accepted! ',
                    'Scroll down to see/edit them.');
   }
   1;
}

#--------------------------------------------------------------------
# Print submit order button.
#--------------------------------------------------------------------
sub submit_order_button {

    # Add select for keep or empty cart here in case user submits the
    # order from this page.
    $HTML -> P;
    $HTML -> select_one (Name=>'empty-cart',
                         Choices=>['Empty shopping cart if no errors',
                                   'DO NOT empty shopping cart']);
    print "after order has been submitted.<BR>\n";

    $HTML -> msubmit_flag;
    $HTML -> CENTER -> FONT (SIZE=>'+3') -> B
          -> msubmit_button (CAPTION =>
            'Go to Step 3: Review Order Summary Before Submitting Order',
                             QUERY =>['mode'    => 'CART',
                                      'submode' => 'MANIFEST']);

    # Add a button to submit order now from here before reviewing summary
    $HTML -> BR;
    $HTML -> t("OR");
    $HTML -> BR;
    $HTML -> msubmit_button(CAPTION =>
                        'Submit Order Now!',
                        QUERY =>['mode'    => 'CART',
                                 'submode' => 'MANIFEST',
                                 'ordernow' => '1']);
    $HTML -> _CENTER -> _FONT -> _B ;
    1;
}

#------------------------------------------------------------
# submode_DOWNLOAD
#------------------------------------------------------------
# Download the preferences from this page
#
sub submode_DOWNLOAD {
   my $self = shift;
   set_prefs_from_edit(1);
   download_prefs ();
}
#------------------------------------------------------------
# submode_SUBMIT
#------------------------------------------------------------
# Submit the order!!!!!
#
sub submode_SUBMIT {
    my $self = shift;
    my $client = attach IW::Client;

    if ( (defined ($client)) && ($client->running) 
        && ($client->activity eq 'PRODUCT_REQUEST') ) {
       $REPLY->redirect(script_url( 'mode'    => 'ORDER',
                                    'submode' => 'COMMTRACE',
                                    'sid'     => $SESSION->id,
                                    'endform' => 1 ));
       return 1;  
    } 

    #   Empty cart flag.
    my $empty_cart = ($REQ->param('empty-cart') !~ /DO NOT/i);
    $SESSION->state->{EmptyCart} = ($empty_cart ? 1 : 0);
    #--------------------
    # BUILD ORDER...

    # Get all granules (and datasets) from the shopping cart:
    my @granules = $CART->granules;
    my @datasets = $CART->datasets;
    my @just_datasets = $CART->just_datasets;

    # Quietly get rid of anything that's not orderable (KLUDGE):
    @granules = grep { $_->possible_to_order } @granules;

    # Construct the orders:
    my ($req_order, $email_order) = 
      IW::Order->make(Granules      => \@granules, 
                      Datasets      => \@datasets,
                      Just_Datasets => \@just_datasets);
    ($req_order || $email_order) or puke "couldn't make order!";

    # Get user info:
    my $user = $SESSION->state->{User} || puke "no user info!";    

# Get unique request id (session id + "-P" + session prod. req. no):
# Move this code to ModeCART submode MANIFEST because reqid needed
# to create manifest file for email
#    my $reqno = 1 + ($SESSION->state->{OrderReqNo} || 0);
#    my $reqid = $SESSION->id . "-P$reqno";
#    $SESSION->state->{OrderReqNo} = $reqno;
#    $SESSION->state->{OrderReqId} = $reqid;
    my $reqid = last_request_id();
    # Add common info:
    my $order;
    foreach $order ($req_order, $email_order) {
        $order->user ($user);
        $order->id ($reqid);
    }
    #--------------------
    # LOG ORDER...

    # KLUDGE: currently, we log both normal and email orders, and we 
    # might not fill in the mediareq and volreq fields...
    my @prods = ();   
    my %prod_hash = ();
    my $item;
#    foreach $item ($req_order->all_items, $email_order->all_items) {
#       my $mediareq = $item->mediareq;
#       my $volreq   = $item->volreq;
#       push(@prods, 'prodname'=>$item->prodname);
#       push(@prods, 'mediareq'=>$mediareq)  if $mediareq;
#       push(@prods, 'volreq'=>$volreq)      if $volreq;
#    }

    my $prodname;
    foreach $item ($req_order->all_items, $email_order->all_items) {

        $prodname = $item->prodname;
        if ($prod_hash{$prodname}) {
            ++$prod_hash{$prodname}  
        } else {
            $prod_hash{$prodname} = 1;
        } # end if 
    } # end foreach $item

    my $key;
    foreach $key (keys %prod_hash) {
        push (@prods, 'prodname'=>$key);
        push (@prods, 'copiesreq'=>$prod_hash{$key});
    }
    ims_event('PRODUCTREQ', 
              'requestid'=>$reqid, # as desired, we log the REQUEST_ID
              'lastname'=>$user->req_addr->lastname,
              'firstname'=>$user->req_addr->firstname,
              'street'=>scalar($user->req_addr->street),
              'city'=>$user->req_addr->city,
              'state'=>$user->req_addr->state,
              'country'=>$user->req_addr->country,
              'zip'=>$user->req_addr->zip,
              'phone'=>$user->req_addr->phone,
              'email'=>$user->req_addr->email,
              'otype'=>'Normal',    # KLUDGE: different for emails?
              @prods);              # the products
    
    #--------------------
    # EMAIL COPY OF ORDER TO USER... THIS IS EMAIL #1
    # As of 2.2 we do not send this email
    
    # Print the receipt to the text file:
#    print_receipt($reqid, $req_order, $email_order);

    # Email it:
#    email_receipt($reqid);

    #--------------------
    # DISPATCH EMAIL-BASED ORDERS...

    # Send out the email orders first:
    inform "about to order by email...";
    order_by_email ($email_order);

    #--------------------
    # DISPATCH REQUEST-BASED ORDERS...
    #
    # If anything was requested, fall into the request-handling loop.
    # THIS WILL ISSUE A REDIRECT TO A COMM STATUS PAGE!!!
    # The redirect is a workaround for a Netscape bug, AND, it keeps
    # the user from accidentally reloading the same page and
    # resubmitting an order.
    #
    $self->{ReqOrder}   = $req_order;
    $self->{EmailOrder} = $email_order;
    $self->submode_COMMSTART;           # go!
    1;
}

#------------------------------------------------------------
# submode_RECEIPT
#------------------------------------------------------------ 
# We need the Receipt to be it's own page.
# or I need an easier way for hyperlinks to WORK!

sub submode_RECEIPT {
    my $self = shift;

    $APP->start_page(Title      => 'Order Receipt',
                     SuperTitle => 'Shopping Cart',
                     Context    => 'orderecpt');
    # Because we are guarding the privacy of the receipt data, we
    # no longer allow direct access via URL. 

    # Open up the receipt.
    my $reqid = last_request_id();
    my $filename = "pub/manifest/Manifest.$reqid";
    my $fullname = $SESSION->path_to($filename);

    if (open RECPT,$fullname) {
        local $_;
        $HTML->H3
             ->t("Your shipping receipt:")
             ->_H3;
        $HTML->PRE;
        while(<RECPT>) { print; }
        $HTML->_PRE;
        close RECPT;
    }
    else {
        print "You currently have no shipping receipt.";
    }
    $APP->end_page; 
    1;
}

#------------------------------------------------------------
# submode_ECSPROFILE
#------------------------------------------------------------ 
# Retrieve or update ECS profile.

sub submode_ECSPROFILE {
    my $self = shift;

    # Get input values from the form.  Updates User object in session.

    # If updating ECS profile, we require the standard fields.
    my $nextsubmode = $REQ->param('nextsubmode') || '';
    my $required = ($nextsubmode eq 'UPDATE' ? 1 : 0);

    #   User info (keys, addresses, affil).  All required except keys.
    my $user = IW::User->from_CGI ('FORM'=>'ORDER',
                                   'REQUIRED'=>$required);

    # If any required fields are missing, show the order form again,
    # instead of continuing.
    if ($user->errors->any_errors) {
        # Set error messages into top of page note.
        $APP->message (new IW::Note Class=>'ERROR',
                       HTML =>join ("<BR>\n", $user->errors->get));
        # Redisplay the form has been moved to "send_ecs_profile_req"
    }

    # This code was added to insure that the ECS password would be
    # retrieved from the stored user prefs and not from the screen.
    my $persist_usr = undef;
    $persist_usr = $user->keys->ecs_username if (defined($user->keys));
    if(defined($persist_usr) && $persist_usr &&
       $ENV{REMOTE_USER} eq $persist_usr) {
       my $prefs_dir =${IU::PersistentDir}."imswww-usr/".$persist_usr;
       my $prefs_file = $prefs_dir."/prefs.dat";
       $user = IW::ModePREFS::load_prefs($prefs_file);
       $user->install if($user);
    }

    # Send the ECS profile request. 
    send_ecs_profile_req ($user);
    1;
}

#------------------------------------------------------------
# submode_NEWFORM 
#------------------------------------------------------------ 
# Retrieve the ECS profile before displaying the ORDER_FORM.
# This only attempts to retrieve the ECS profile if the user is a permanent
# user and he is ordering to an ECS DAAC.  This is determined if the
# following is set in "wwwvalids.conf" for the DAAC:
# REQUIRES_ECS_AUTH       = "Y". 


sub submode_NEWFORM {
    my $self = shift;
    # Get all granules from the shopping cart:
    my @granules = $CART->granules;
    my @datasets = $CART->datasets;
    @datasets or
        puke 531, 'there are no datasets in the shopping cart!';
    my @just_datasets = $CART->just_datasets;

    # Quietly get rid of anything that's not orderable (KLUDGE):
    @granules = grep { $_->possible_to_order } @granules;

    # Construct the order:
    my ($req_order, $email_order) =
       IW::Order->make (Granules      => \@granules,
                        Datasets      => \@datasets,
                        Just_Datasets => \@just_datasets);
    ($req_order || $email_order) or puke "couldn't make order!";
    # Get ids of ECS daacs orders are going to.
    my @req_ecsdaacs = $req_order->ecs_archive_ids;
    my @email_ecsdaacs = $email_order->ecs_archive_ids;
 
    # Get number of ECS daacs orders are going to.
    my $req_ecsdaacs_no = scalar (@req_ecsdaacs);
    my $email_ecsdaacs_no = scalar (@email_ecsdaacs);


    # This only attempts to retrieve the ECS profile if the user is a permanent
    # user and he is ordering to an ECS DAAC.  This is determined if the
    # following is set in "wwwvalids.conf" for the DAAC:
    # REQUIRES_ECS_AUTH       = "Y". 
    # If order is going to ECS, prompt for ECS Access fields.
    if (($ENV{REMOTE_USER} ne 'guest') && 
        (($req_ecsdaacs_no > 0) || ($email_ecsdaacs_no > 0)) ) {
       my $nextsubmode = 'RETRIEVE';

       # Get input values from the user object in session.
       my $user = $SESSION->state->{User};

       # This code was added to insure that the ECS password would be
       # retrieved from the stored user prefs and not from the screen.
       my $persist_usr = undef;
       $persist_usr = $user->keys->ecs_username if (defined($user->keys));
       if(defined($persist_usr) && $persist_usr &&
          $ENV{REMOTE_USER} eq $persist_usr) {
          my $prefs_dir =${IU::PersistentDir}."imswww-usr/".$persist_usr;
          my $prefs_file = $prefs_dir."/prefs.dat";
          $user = IW::ModePREFS::load_prefs($prefs_file);
          $user->install if($user);
       }

       # Send the ECS profile request. 
       send_ecs_profile_req ($user, $nextsubmode);
    } else {
       # If we are not doing an ECS Profile Request go straight to the
       # order form.
       submode_FORM();
    }
    1;
}


#------------------------------------------------------------
# submode_UPLOAD 
#------------------------------------------------------------ 
# Upload preferences from a file

sub submode_UPLOAD {
   my $self = shift;
   set_prefs_from_upload ();
   $self->submode_FORM;
   1;
}
#------------------------------------------------------------
# go
#------------------------------------------------------------
# Main entry point.

sub go {
    my $self = shift;
    
    # Get session if we need it.
    $APP->get_session();
    unless ($REQ->param('submode') eq 'COMMFINAL' or
            $REQ->param('submode') eq 'RECEIPT') {
        login() or puke "can't log in!";    # log into cart database
    }
    $self->go_submode();                # go to submode or die
}

#------------------------------------------------------------
1;
