package IW::ModePKGSLCT;

=head1 NAME

IW::ModePKGSLCT - package selection

=head1 AUTHOR

Eryq, Hughes STX Corporation

=head1 RCS INFO

$Id: ModePKGSLCT.pm,v 1.88 2003/08/15 18:03:30 solomon Exp $

=cut

use IU;
use IU::Coverage;
use IU::Dataset;
use IU::Granule;
use IU::InvStore;
use IU::Log;
use IU::MediaHandlingFee;
use IU::MediaMaximumCapacity;
use IU::Package;
use IU::PkgCombo;
use IU::Session;
use IU::Util qw(shellquote);

use IW::App;
use IW::Comm::AccountStatusResult;
use IW::User;
use IW::Util qw(:url :string);
use IW::Widget::URLTable;

@ISA = qw(IW::Mode);

use strict;

#------------------------------------------------------------
#
# Globals
#
#------------------------------------------------------------

# The databases:
my $CART;

# Hash containing medias too small for this granule.
my %SmallMedias;

# The daac for contact information purposes
my $DAAC;

# Hash containing the names of the $user subroutines
my %user_subs =
   ( 'FTPHOST' => 'ftp_host',
     'FTPPASSWORD' => 'ftp_password',
     'FTPPUSHDEST' => 'ftp_pushdest',
     'FTPUSER' => 'ftp_user',
     'USERSTRING' => 'ftp_userstring');
my @user_subs_keys = keys %user_subs;
my $sub_key;

#------------------------------------------------------------
# print_small_media_table 
#------------------------------------------------------------
sub print_small_media_table {

   my $dset = shift;
   my $gran = shift;
   my $pkgref = shift;

   %SmallMedias = ();

   # loop through all the package combos and build a hash of
   # unique media types;
   my $package;
   my %combo_types;
   my @combos;
   my $combo;
   my $media_type;

   foreach $package (@{$pkgref}) {
      foreach $combo ($package->combos) {
         $media_type = $combo->media_type;
         $combo_types{"$media_type"} ='1';
      }
   }

   # Check to see if any of the media capacity limit applies to this package.
   # If some do then output the table.
   my $granule_size = $gran->size_mbecs_granule;
   my $mediamaximumcapacity;

   CHECK: foreach $mediamaximumcapacity (@{$dset->mediamaximumcapacity}) {
      # -------------------------------------------------------------
      # We need to check for three things:
      #   1. Is the media type in the package combo?
      #   2. Does the media type have a maximum capacity defined?
      #   3. Is the granule too big to fit in the media?
      #
      # If all of these scenarios are true, then we should not allow
      # the granule to be ordered in the small media, add the media
      # type to the SmallMedias table so we can keep track.
      # -------------------------------------------------------------
      if ( ($combo_types{$mediamaximumcapacity->type}) && defined($mediamaximumcapacity->capacity) && ($granule_size > $mediamaximumcapacity->capacity) ) {
         $SmallMedias{$mediamaximumcapacity->type} =
                             $mediamaximumcapacity->capacity;
         next CHECK;
      }
      # -------------------------------------------------------------
      # We are doing similar check for media and granule limits,
      # the difference is that we are checking for size_mb because 
      # some daacs send the media limit as size_mb, not as capacity.
      # -------------------------------------------------------------
      if ( ($combo_types{$mediamaximumcapacity->type}) && defined($mediamaximumcapacity->size_mb) && ($granule_size > $mediamaximumcapacity->size_mb) ) {
         $SmallMedias{$mediamaximumcapacity->type} =
                             $mediamaximumcapacity->size_mb;
      }
   }

   my @types = sort keys %SmallMedias;
   my $capacity;
   if (scalar(@types) > 0) {
      print '<P>The following media is not available for this granule<BR>';
      print 'because the media is too small: <BR>';
      print '<TABLE BORDER=1>';
      print '<TR><TD><B>Media type</B></TD>';
      print '<TD><B>Media Size (MB)</B></TD>';
      print '<TD><B>Granule Size (MB)</B></TD></TR>';
      foreach $media_type (@types) {
         $capacity = $SmallMedias{$media_type};
         print '<TR>';
         print '<TD>',"$media_type",'</TD>';
         print '<TD>',"$capacity",'</TD>';
         print '<TD>',"$granule_size",'</TD>';
         print '</TR>';
      }
      print '</TABLE><P>';
   }
}

#------------------------------------------------------------
# print_media_fee_table DSET 
#------------------------------------------------------------
sub print_media_fee_table {

   my $dset = shift;
   my $pkgref = shift;

   # loop through all the package combos and build a hash of 
   # unique media types;
   my $package;
   my %combo_types;
   my @combos;
   my $combo;
   my $media_type;
   foreach $package (@{$pkgref}) {
      foreach $combo ($package->combos) {
         $media_type = $combo->media_type; 
         $combo_types{"$media_type"} ='1'; 
      }    
   }
   # Check to see if any of the media handling fees apply to this package.
   # If some do then output the table.
   my $count = 0;
   my ($mediahandlingfee,$type,$fee,$comment,$url);
   foreach $mediahandlingfee (@{$dset->mediahandlingfee}) {
      ++$count if ($combo_types{$mediahandlingfee->type});
   }
   if ($count > 0) {
      print '<TABLE BORDER=1>';
      print '<TR><TD><B>Media type</B></TD><TD><B>Fee (US)</B></TD>';
      print '<TD><B>Comments</B></TD></TR>';
      foreach $mediahandlingfee (@{$dset->mediahandlingfee}) {
         $type = $mediahandlingfee->type; 
         # Filter out those media handling fees which don't apply for this 
         # set of package combos. I.e. Only display it if it is one of the 
         # package combos.
         if ($combo_types{"$type"}) {
            $fee = $mediahandlingfee->fee;
            $fee = sprintf("%0.2f",$fee);
            $comment = $mediahandlingfee->comment; 
            $url = $mediahandlingfee->url;
            print '<TR>';
            print '<TD>',"$type",'</TD>';
            print '<TD> $',"$fee",'</TD>';
            print '<TD>';
            print "$comment " if ($comment);
            $HTML -> A(HREF=>$url->URL) -> t($url->comment) -> _A if ($url);
            print '</TD>';
            print '</TR>';
         }
      }
      print '</TABLE><P>';
   }
}

#------------------------------------------------------------
# print_header ITEM
#------------------------------------------------------------
# Print the header at the top of the page, showing the dataset
# and granule id.  Also include geographic info if known,
# area and center.
# ITEM is any orderable item.
sub print_header {
   my $item = shift;
   # The title and links (note: NO dataset attribute button for now!):
   if ($item->dataset_key) {       # the item "has a relevant dataset"
       $HTML -> BR;
       $HTML -> FONT(SIZE=>'+1') 
             -> t_B($item->dataset_id) -> _FONT 
             -> BR;
   }
   if ($item->granule_key) {       # the item "has a relevant granule"
       $HTML  -> FONT(SIZE=>'+1') -> B
            -> I -> t(' Data Granule ID: ') -> _I 
            -> t($item->granule_id) -> _B -> _FONT
            -> BR;
   }
     my $subset_opts = $item->some_options('SUBSET');
    if ( $subset_opts) {
   # If center exists, print it.
   my $coverage = $item->coverage;
   if ($coverage->center) {
      $HTML -> t('Geographic Center: ');
      print $coverage->center_as_HTML;
      $HTML -> BR;
   }
   # If lat lon coords exist, print them.
  if ($coverage->area) {
      $HTML -> t('Geographic Coordinates: ');
      my @coords = split (/;/, $coverage->area_pretty);
      if (int(@coords) == 1) {
         print $coords[0], "\n";
         $HTML -> BR;
      } else {
         $HTML -> BR;
         print '&nbsp';     # Need a leading blank for first coord.
         my $coord;
         foreach $coord (@coords) {
            $HTML -> hspace(20);
            print $coord, "\n";
            $HTML -> BR;
          }
       }

    }
  }
   1;
}

#------------------------------------------------------------
# check_number_of_packages PACKAGEREF
#------------------------------------------------------------
sub check_number_of_packages {
   my $packageref = shift;
   my $combo;
   my $choice;
   my $small_media;
   my $package;
   my $packageNumber = 0;

   foreach $package (@{$packageref}) {
      my @combos = $package->combos;
      @combos or 
       puke 457, "Ordering information not available";
      $choice = 0;
      COMBOTEST: foreach $combo (@combos) {
         ++$choice;
         foreach $small_media (keys %SmallMedias) {
            next COMBOTEST if $combo->media_type eq $small_media;
         }
      }
      ++$packageNumber if($choice);
   }
   return $packageNumber;
}
#------------------------------------------------------------
# print_package_menu ITEM, PACKAGE, PKGNO
#------------------------------------------------------------
# Print the menu of packaging "combos" for a single package.
# The ITEM is an orderable item (dataset, granule, etc.)
# If there is a title in the package, print the title in
# the first cell and the option number above.  Otherwise,
# just print option number in the first cell.
# Note PKGNO here is 1 based, not 0 based.
#
sub print_package_menu {
    my ($item, $package, $pkgno,$floatscene) = @_;
    my $item_thingy = $item->order_thingy;

    # Get all currently-selected package combos for this granule,
    # in a hash:
    my %selected = map {($_->as_hashkey => 1)}
                       @{$item->order_pkgcombos};

    my @combos = $package->combos;
    @combos or puke 457, "Ordering information for $item_thingy not available";

    my $combo;
    my $choice = 0;
    my $small_media;
    my $user = $SESSION->state->{User} || new IW::User;
    my $anything_chosen = 0;
    my $onoff = 0;
    
    COMBOTEST: foreach $combo (@combos) {
       ++$choice;
       foreach $small_media (keys %SmallMedias) {
           next COMBOTEST if $combo->media_type eq $small_media;
       }
       $onoff = $selected{$combo->as_hashkey};
       $anything_chosen ||= $onoff;
    }
    if($choice > 0) {
    # Print header, with package name:
    $HTML -> TABLE(BORDER=>3, CELLPADDING=>3, CELLSPACING=>0,
                   WIDTH=>'75%', 
                   BGCOLOR=>$HTML->hue('PkgItemsTblHeader'));
    $HTML -> TR(VALIGN=>'TOP');
    $HTML -> TD;
    $HTML -> color(0, 40, 80);

    my $title = $package->title;   # May be '' == false.
    if ($title) {
        $HTML -> H2 -> t("Ordering Option $pkgno:") -> BR
              -> t($title) -> _H2;
    } else {
        $HTML -> H2 -> t("Ordering Option: ", $pkgno) -> _H2;
    }
    $HTML -> _color;
    $HTML -> _TD;

    # Print comment, if any:
    $HTML -> TD -> small;

    if ($package->comment) {
        $HTML -> t($package->comment) ->BR;
    }
    # If there are any miscellaneous URLs in the package print them.
    # Note that misc_URLs() returns ref to array that might be [].
    my $misc_URLs_ref = $package->misc_URLs;
    my $ref = $package->{'MISC_URL'};
    my @furls = @$ref;
    if (int (@{$misc_URLs_ref})) {
       my $tbl = new IW::Widget::URLTable 
          "More info:" => $misc_URLs_ref;
       $tbl->print;
    }
    if ($package->id =~ /\*/ ) {
        $HTML -> t("(Contains just this $item_thingy.)");
    }
    else {
        my $ngran = $package->n_granules;
        if ($ngran) {
           $HTML-> t("Contains ", $ngran, " ", $item_thingy);
           ($ngran != 1) ? print 's.' : print '.';
        }
    }
    $HTML -> _small -> _TD;
    $HTML -> _TR;

    # Get the information prompt or default it.
    my $info_prompt;
    if ($package->info_prompt) {
       $info_prompt = $package->info_prompt;
    } elsif ($item->archive_id =~ /^NOAA/i) {
       $info_prompt = 'Please enter your NOAA account number: ';
    } else {
       $info_prompt = '';
    }
    my $addn_info = '';
    # Find if this package has a previously set additional info.
    my $comb;
    # For each combo previously selected for this item:
    foreach $comb (@{$item->package_combos}) {
       # If this selected combo is the current package:
       if ($package->id eq $comb->package_id) {
          $addn_info = $comb->addn_info;
          last;                       # break if we find it.
       }
    }
#    unless ($floatscene) {
       $HTML -> TR -> TD(COLSPAN=>2);
       # If the package has an info prompt, print it.
       if ($info_prompt) {
          $HTML -> small -> t($info_prompt) -> _small -> BR;
       }
       # Print additional info box.
       $HTML -> field_cell(Title=>'Additional Info: ', Stack=>0)
             -> INPUT(TYPE=>'text', SIZE=>80, MAXLENGTH=>80,
                   NAME=>"additional_info.$pkgno",
                   VALUE=>$addn_info)
             -> _field_cell;
       $HTML -> small -> t("") -> _small;
       $HTML -> _TD -> _TR;
#    }
    $HTML -> _TABLE;

    # Setup table and header:
    $HTML -> TABLE(BORDER=>3, CELLPADDING=>2, CELLSPACING=>0,
                   WIDTH=>'75%',
                   BGCOLOR=>$HTML->hue('HiBack'))
          -> TR(BGCOLOR=>$HTML->hue('PkgItemsTblHeader'))
          -> TH -> FONT(SIZE=>'-1') -> t_B('Select'); 
    $HTML -> BR;
    $HTML -> t_B('One') -> _FONT -> _TH;
    $HTML -> FONT(SIZE=>'+1');
    $HTML -> TH -> t_B('Data Format') -> _TH
          -> TH -> t_B('Media type') -> _TH
          -> TH -> t_B('Media format') -> _TH
          -> TH -> t_B('Package Size') -> _TH
          -> TH -> t_B('Cost (US$)') -> _TH
          -> TR;
    $HTML -> _FONT;
    # Output each row of the packages menu:
    $choice = 0;
    
    COMBO: foreach $combo (@combos) {
        ++$choice;
        foreach $small_media (keys %SmallMedias) {
           next COMBO if $combo->media_type eq $small_media;
        }
        
        # Should this choice be on?
        $onoff = $selected{$combo->as_hashkey};
        $onoff = 1 if(($user->data_format eq $combo->proc_opt) &&
                      ($user->media_type eq $combo->media_type) &&
                      ($user->media_format eq $combo->media_fmt) &&
                      (!$anything_chosen));

        $anything_chosen ||= $onoff;

        # Precompute some values:
        my $cost = ($floatscene) ? 'Pending' : sprintf("\$%.02f", $combo->cost);

        # Leave space for URLs?
        my $has_URLs = @{$combo->data_URLs};

        # OUTPUT THE ROW...
        
        $HTML -> TR;

        # Checkbox...
        $HTML -> TD(ALIGN=>'LEFT', ROWSPAN=>($has_URLs ? 2 : 1))
              -> FONT(SIZE=>'+2') -> B;
        if ($package->allows_multiple_media) {

            $HTML -> checkbox    (Name  =>"pk-choice.$pkgno",  # "pk-choice.1"
                                  Choice=>"$pkgno.$choice",    # "1.2"
                                  Value =>$onoff);
            $HTML -> t(" ");

           # determine if the user already selected "number of copies"
           # for this combo, if so then initialize the form.

           my $copies = '1';
               
           my $my_combo; 
           my $pkg_array;
           foreach $pkg_array ($item->order_pkgcombos) {

               foreach $my_combo (@$pkg_array) {

                   if ($my_combo->merged_key eq $combo->merged_key) { 
                       $copies = $my_combo->copies;
                   } # end if

               } # combo foreach                  
           } # end pkg foreach 

           $HTML -> field_cell(Title=>'Copies: ', Stack=>0)
                 -> INPUT(TYPE=>'text', SIZE=>3, MAXLENGTH=>3,
                           NAME=>"copies.$pkgno.$choice",
                           VALUE=>"$copies")      
                 -> _field_cell;
        }
        else {
            $HTML -> radio_button(Name   =>"pk-choice.$pkgno", # "pk-choice.1"
                                  Choice =>"$pkgno.$choice",   # "1.2"
                                  IsDefault =>$onoff);
        }
#       $HTML -> I -> t(" $pkgno.$choice") -> _I;       # "1.2"
        $HTML -> _B -> _FONT;
        $HTML -> _TD;

        # Package id...
        $HTML -> TD(ALIGN=>'CENTER') 
              -> small -> t($combo->proc_opt)   -> _small -> _TD;
        $HTML -> TD(ALIGN=>'CENTER') 
              -> small -> t($combo->media_type) -> _small -> _TD;
        $HTML -> TD(ALIGN=>'CENTER') 
              -> small -> t($combo->media_fmt)  -> _small -> _TD;
        $HTML -> TD(ALIGN=>'CENTER') 
              -> small -> t($combo->pkg_size)  -> _small -> _TD;
        $HTML -> TD(ALIGN=>'CENTER') 
              -> small -> t($cost)              -> _small -> _TD;
        $HTML -> _TR;
        # URL row?
        $has_URLs or next;
        $HTML -> TR;
        $HTML -> TD(ALIGN=>'LEFT', VALIGN=>'TOP', COLSPAN=>5);  
        my $furl;
        $HTML -> small -> UL;
        foreach $furl (@{$combo->data_URLs}) {
            $HTML -> LI
                  -> hypertext("[DATA]", 
                               logged_url($furl->URL, 'type'=>'DATA'))
                  -> t(" ", $furl->comment);
        }
        $HTML -> _UL -> _small;
        $HTML -> _TD -> _TR;
    }

    # Final row, if needed:
    if (!$package->allows_multiple_media) {
        $HTML -> TR -> TD(COLSPAN=>6);
        $HTML -> radio_button(Name   =>"pk-choice.$pkgno", # "pk-choice.1"
                              Choice =>"(none)",
                              IsDefault => !$anything_chosen);
        $HTML -> t_EM(" I want no items from this option.");
        $HTML -> _TD -> _TR;
    }

    # Footer:
    $HTML -> _TABLE;
    } else {
       return 0;
    }
    1;
}

#------------------------------------------------------------
# print_options_menus ITEM,PACKAGE,PKGNO
#------------------------------------------------------------
# Print the menu of subsetting options for a single package.
# Each menu's CGI tags will begin with "pk-opts.PKGNO."
# where PKGNO is 1, 2, etc.
#
# ITEM is any orderable item.

sub print_options_menus {
    my ($item, $package, $pkgno) = @_;
    
    # Stop immediately if no subsetting criteria:
    my $opts_type;
    foreach $opts_type (sort $package->options) {
        my $opts = $package->options($opts_type) || die "hunh?";

        $HTML -> P;
        $HTML -> IMG(SRC=>doc_url("icons/arrowup.gif"), 
                     ALIGN=>'LEFT',
                     HSPACE=>4);
        print_options_menu($item, $package, $pkgno, Opts=>$opts);
    }
}

#------------------------------------------------------------
# print_options_menu ITEM,PACKAGE,PKGNO,OPTIONS
#------------------------------------------------------------
# Print the menu of subset/order options for a single package.
# Each entry's CGI tags will begin with "pk-opts.PKGNO.OPTSTYPE",
# where PKGNO is 1, 2, etc, and OPTSTYPE is the options type for
# this options menu (e.g., "SUBSET", "ORDER", etc.).
#
# ITEM is any orderable item.

sub print_options_menu {
    my ($item, $package, $pkgno, %params) = @_;
    my $opts     = $params{Opts};
    my $optstype = $params{OptsType}  || $opts->type;
    my $title    = $params{OptsTitle} || $opts->pretty_type;
    my $user = $SESSION->state->{User} || new IW::User;

    # Do we have any subsetting options CURRENTLY SELECTED for this package?
    # If so, set up the CGI tags:
    if ($item->order_pkgcombos) {
#       debug "scanning combos to see if package $pkgno currently selected";

        # Find first combo with this package id:
        my $combo = (grep {$_->package_id eq $package->id
                           } @{$item->order_pkgcombos})[0];

        # If one was found, and if it has subset criteria, turn on the values!
        if ($combo and $combo->options($optstype)) {
#           debug "package $pkgno selected, and it has crit'a for $optstype! ".
#                 "setting CGI tags for $optstype!";
            $combo->options($optstype)->set_CGI(Prefix=>
                                                "pk-opts.$pkgno.$optstype.");
        }
    }

    # Start the table of criteria:
    my @criteria = $opts->criteria;
    my $crit;
    $HTML -> TABLE(BORDER=>3, 
                   ###  WIDTH=>'75%', 
                   CELLPADDING=>3, CELLSPACING=>0,
                   BGCOLOR=>$HTML->hue('HiBack'));
    # Title:
    $HTML -> TR(VALIGN=>'TOP');
    $HTML -> TD(COLSPAN=>2, BGCOLOR=>$HTML->hue('PkgOptsTblHeader'))
          -> FONT(SIZE=>'+1') 
          -> t_B("\u$title (for Ordering Options $pkgno)") 
          -> _TD
          -> _FONT;

    # Directions:
    $HTML -> BR;
    if ($opts->is_required) {
        $HTML -> I -> t_B("Mandatory: ") -> _I -> small;
        print <<EOF;
              If you selected any of the items in ordering options $pkgno (above),
              you must give values for 
              any required Options below.  
              You can also give values for the other options, if you desire.
EOF
        $HTML -> _small;
    }
    else {
        $HTML -> I -> t_B("Optional: ") -> _I -> small;
        print <<EOF;
              If you selected any of the items in ordering options $pkgno (above),
              you can provide values for any of the options below.
              <B>Note:</B> if you supply values for <I>any</I> of these
              options, then you must be sure to supply values for
              <I>all</I> of the required options.
EOF
        $HTML -> _small;
    }   
    $HTML -> _TR;
    # Output each special criterion:
    my $type;
    foreach $crit (@criteria) {
        my $type = $crit->type;
        next if($type eq 'SUBPAGE');
        my $preset = undef;
        $HTML -> TR;
        foreach $sub_key (@user_subs_keys) {
           if($sub_key eq $crit->name) {
              my $newsub = $user_subs{$sub_key};
              $preset = $user->$newsub();
           }
        }
        $crit->print_form(Prefix=>"pk-opts.$pkgno.$optstype.", HTML=>$HTML,
                          Preset=>$preset);
        $HTML -> _TR;
    }    
    $HTML -> _TABLE;
    1;
}

sub print_subpage_menu {
    my ($item, $package, $pkgno, %params) = @_;
    my $opts     = $params{Opts};
    my $page     = $params{Page};
    my $optstype = $params{OptsType}  || $opts->type;
    my $title    = $params{OptsTitle} || $opts->pretty_type;
    my $user = $SESSION->state->{User} || new IW::User;

    # Do we have any subsetting options CURRENTLY SELECTED for this package?
    # If so, set up the CGI tags:
    if ($item->order_pkgcombos) {
        # Find first combo with this package id:
        my $combo = (grep {$_->package_id eq $package->id
                           } @{$item->order_pkgcombos})[0];

        # If one was found, and if it has subset criteria, turn on the values!
        if ($combo and $combo->options($optstype)) {
            $combo->options($optstype)->set_CGI(Prefix=>
                                        "pk-opts.$pkgno.$optstype.");
        }
    }

    # Start the table of criteria:
    my @criteria = $opts->criteria;
    my $crit;
    $HTML -> TABLE(BORDER=>1, 
                   WIDTH=>'75%', 
                   CELLPADDING=>1, CELLSPACING=>0,
                   BGCOLOR=>$HTML->hue('HiBack'));
    # Title:
    $HTML -> TR(VALIGN=>'TOP');
    $HTML -> TD(VALIGN=>'TOP', COLSPAN=>2, 
                BGCOLOR=>$HTML->hue('PkgOptsTblHeader'))
          -> FONT(SIZE=>'+1') 
          -> t_B("\u$title (for Ordering Options $pkgno)") 
          -> _TD
          -> _FONT;

    $HTML -> BR;
    # Output each special criterion:
    my $type;
    my $min;
    my $max;
    foreach $crit (@criteria) {
        my $type = $crit->type;
        next if($type ne 'SUBPAGE');
        $max = $crit->max;
        $min = $crit->min;
        my $preset = undef;
        $HTML -> TR;
        $crit->print_form(Prefix=>"pk-opts.$pkgno.$optstype.", 
                          HTML=>$HTML, Preset=>$preset, Page=>$page);
        $HTML -> _TR;
    }    
    $HTML -> _TABLE;
    return $max, $min;
}

#------------------------------------------------------------
# max_min OPTIONS
#------------------------------------------------------------
# Return the max and min for the SUB_PAGE criteria
sub max_min {
   my $opts = shift;
   my @criteria = $opts->criteria;
   my $crit;
   my $type;
   my $min;
   my $max;
   foreach $crit (@criteria) {
      my $type = $crit->type;
      next if($type ne 'SUBPAGE');
      $max = $crit->max;
      $min = $crit->min;
   }    
   return $max, $min;
}

#------------------------------------------------------------
# print_all_or_one_menu GRAN DSET
#------------------------------------------------------------
# Print the menu which asks if they want these options to be
# applied to all possible granules in the shopping cart, or just

#------------------------------------------------------------
# print_all_or_one_menu GRAN DSET
#------------------------------------------------------------
# Print the menu which asks if they want these options to be
# applied to all possible granules in the shopping cart, or just
# to the granule that was selected.

sub print_all_or_one_menu {
    my $gran = shift;
    my $dset = shift;

    # Print the all-or-one option:
    $HTML -> TABLE;

    # Default is originally set for all in data set: 'dataset'.
    # However, if the user set it to 'granule' for a previous
    # granule in the data set, then use that setting.  The setting
    # is stored in the data set object.  In the code, if the
    # setting is in the data set object, use that, else default.

    my $pkg_set_for = $dset->pkg_set_for;
    my $pkg_set_for_dataset = 1;              # default 'dataset'
    my $pkg_set_for_granule = 0;              # opposite

    # If not defined, use default above, else:
    if (defined($pkg_set_for)) {
       if ($pkg_set_for eq 'granule') {
          $pkg_set_for_dataset = 0;
          $pkg_set_for_granule = 1;
       }
    }
    # All...
    $HTML -> TR 
          -> TD -> radio_button(Name      =>'set-for',
                                Choice    =>'dataset',
                                IsDefault =>$pkg_set_for_dataset)
          -> _TD;
    $HTML -> TD;
    print "I want these ordering options for <I>every</I> data granule"; 
    print " that applies in data set "; 
    $HTML -> t_B($gran->dataset_id);
    print " in the shopping cart.";
    $HTML -> _TD -> _TR;

    # One...
    $HTML -> TR 
          -> TD -> radio_button(Name      =>'set-for',
                                Choice    =>'granule',
                                IsDefault =>$pkg_set_for_granule)
          -> _TD;
    $HTML -> TD;
    print "I want these ordering options <I>only</I> for data granule ";
    $HTML -> t_B($gran->id);
    print ".";
    
    $HTML -> _TD -> _TR;
    $HTML -> _TABLE;
}

#------------------------------------------------------------
# print_account_menu DSET
#------------------------------------------------------------
# Print the menu asking the user for their accounting information.

sub print_account_menu {
    my ($dset) = @_;
    my $acct;
    # Get the granule key:
    my $gkey = $REQ->param('gkey');
    my $dskey = $REQ->param('dskey');
    if($gkey) {
        $SESSION->state->{PackageGkey} = $gkey;
        $SESSION->state->{PackageDskey} = undef;
    }
    if($dskey) {
        $SESSION->state->{PackageGkey} = undef;
        $SESSION->state->{PackageDskey} = $dskey;
    }
    if(!$gkey) {
        $gkey = $SESSION->state->{PackageGkey};
    }
    if(!$dskey) {
        $dskey = $SESSION->state->{PackageDskey};
    }
    my $origin = $REQ->param('origin');

    # Count the number of valid accounts:
    my $good = 0; 
    foreach $acct (@{$dset->accounts}) {
        ++$good unless $acct->error_msg;
    }

    # Simple message if no accounts, or no usable ones:
    if (!@{$dset->accounts}) {
        return 1;
    }    
    # Read the account status results:
    my %prs;
    eval {%prs =  IW::Comm::AccountStatusResult->read_last } or
        do {%prs = ()};
    my $daac;
    my $pr;
    foreach $daac (sort keys %prs) {
        $pr = $prs{$daac};
        foreach $acct (@{$pr->accounts}) {
            ++$good unless $acct->error_msg;
        }
    }
    # Output directions:    
    if ($good >= 1) {
        print "Please select one of the following user accounts to ",
              "order data from this data set:";
    }
    else {
        print "<b>A valid user account is needed to order data, yet ",
              "you do not seem to have one.  Read the comments ",
              "below.\n";
    }
    # Output menu:
    $HTML -> P;
    $HTML -> TABLE(BORDER=>3, CELLPADDING=>2, WIDTH=>'90%');
    $HTML -> TR 
          -> TH -> color('REQFIELD') -> t('CHOICE') -> _color -> _TH
          -> TH -> t('Account Name') -> _TH
          -> TH -> t('Balance') -> _TH
          -> TH -> t('Notes') -> _TH -> _TR;
    if(!%prs) {
        foreach $acct (@{$dset->accounts}) {
            $HTML -> TR;
            $HTML -> TD(ALIGN=>'CENTER');       
    
            # If it's a good account, allow user to select it:
            unless ($acct->error_msg) {
                $HTML -> radio_button(Name =>'account',
                  Choice =>$acct->name,
                  IsDefault=>(($acct->name eq $dset->chosen_account)
                              || ($good == 1)));
            }
            $HTML -> _TD;
            $HTML -> TD -> t(defined($acct->name) ? $acct->name : '') -> _TD;
            $HTML -> TD -> t(sprintf("US\$ %0.2f", $acct->value||0)) -> _TD;
            $HTML -> TD -> t($acct->error_msg || '') -> _TD -> _TR;
        }
    }
    foreach $daac (sort keys %prs) {
        $pr = $prs{$daac};
        foreach $acct (@{$pr->accounts}) {
            $HTML -> TR;
            $HTML -> TD(ALIGN=>'CENTER');       

            # If it's a good account, allow user to select it:
            $HTML -> radio_button(Name =>'account',
                  Choice =>$acct->name) -> _TD;
            $HTML -> 
               TD -> t(defined($acct->name) ? $acct->name : '') -> _TD;
            $HTML -> TD -> t(sprintf("US\$ %0.2f", $acct->value||0)) -> _TD;
            $HTML -> TD -> t($acct->dataset, 
                $acct->error_msg ? 
                ("\nError: ", $acct->error_msg) : '' || '') -> _TD -> _TR;
        }
    }
    $HTML -> _TABLE;
    if (@{$dset->accounts}) {
        $HTML -> BR;
        print "<big><b>Optional:</b></big>";
        $HTML -> hspace(4);
        $HTML -> msubmit_button(CAPTION => 'Get Account Status',
                                QUERY =>['mode'   => 'ACCOUNTSTATUS',
                                         'submode' => 'SHOW',
                                         'gkey' => $gkey,
                                         'origin' => $origin]);
        $HTML -> BR;
    }
    1;
}

#------------------------------------------------------------
# print_handling_fee DSET
#------------------------------------------------------------
# Decides whether to print out any Handling Fees associated with 
# the Dataset on the Package Selection Page.

sub print_handling_fee {
   # We've been 'handed' the dataset
   my ($dset) = @_;

   # Is there a Handling Fee within the Dataset?
   # If so, create a BLANK HandlingFee Object 
   # and pass it the DATASET object
   # It'll print ALL the Handling Fee(s) for you.

   if (defined($dset->handling_fee)) {
       my $new_fee_object = IU::HandlingFee->new();
       $new_fee_object->print_handling_fees($dset);
   }
1;
}

#------------------------------------------------------------
# print_form ITEM, PACKAGES, DATASET [, OPTION_NUMBER ]
#------------------------------------------------------------
# Print the meat of the page.
# If optional OPTION_NUMBER is present, it means that
# PACKAGES will only contain one package whose number is
# actually the one given (1 based).
#
sub print_form {
   my ($item, $packagesref, $dset, $opt_num) = @_;
   my @packages = @$packagesref;
   my $goodOptions = 1;

   $HTML -> msubmit_flag;
   # If item is the data set, then next submode is SETPKGS_DATASET.
   # If it is a granule, then next submode is SETPKGS.
   my $next_submode;
   if ($item == $dset) {
      $next_submode = 'SETPKGS_DATASET';
   } else {
      $next_submode = 'SETPKGS';
   }
    $HTML -> HR;

   # The dataset/granule header:
   print_header($item);

   # the media fee summary table:
   print_media_fee_table($dset,\@packages); 

   my $floatscene = 0;  # are there any floating scene criteria. 
   my $subpage = 0;  # are there any subpage criteria. 
   # If optional OPTION_NUMBER is present, it means that
   # PACKAGES will only contain one package whose number is
   # actually the one given.
   if (defined ($opt_num)) {
      # Check if package has any floating scene or subpage criteria.
      my $criteria;
      my $opts = $packages[0]->options('SUBSET');
      if (defined($opts)) {
         foreach $criteria ($opts->criteria) {
            $floatscene += 1 if ($criteria->type eq 'FLOATINGSCENE'); 
            $subpage += 1 if ($criteria->type eq 'SUBPAGE'); 
         } 
      }
      $opts = $packages[0]->options('ORDER');
      if (defined($opts)) {
         foreach $criteria ($opts->criteria) {
            $subpage += 1 if ($criteria->type eq 'SUBPAGE'); 
         } 
      }
      # print the media capacity limit table:
      print_small_media_table($dset, $item, \@packages) if(!$floatscene);

      # Print the account menu if needed even for a single package
      print_account_menu($dset);
      print "\n";
      $goodOptions = 
       print_package_menu($item, $packages[0], $opt_num,$floatscene);
      $HTML -> P; 
      #Print only the ORDER options if there are floating scene criteria. 
      if ($floatscene) {
         my $order_opts = $packages[0]->options('ORDER');
         if (defined($order_opts)) {
            $HTML -> P;
            $HTML -> IMG(SRC=>doc_url("icons/arrowup.gif"),
                     ALIGN=>'LEFT',
                     HSPACE=>4);
            print_options_menu($item, $packages[0],$opt_num,Opts=>$order_opts);
         }
      } elsif($goodOptions) {
         print_options_menus($item, $packages[0], $opt_num);
         $HTML ->BR;
      }
   } else {
      # print the media capacity limit table:
      print_small_media_table($dset, $item, \@packages);
      my $number = check_number_of_packages(\@packages);
      # Menu of available accounts to choose from:
      print_account_menu($dset) if($number);

      # Print message:
      print "There are <B>$number</B> ",
          "different Ordering Options available for this data ",
          "granule.<BR>\n" if($number);

      my $pkgno;
      # The package menu:
      for $pkgno (0..$#packages) {
         my $package = $packages[$pkgno];
         print "\n<HR>" if $goodOptions;
         $goodOptions = print_package_menu($item, $package, $pkgno+1,"0");
         if($goodOptions) {
            $HTML -> P;
            print_options_menus($item, $package, $pkgno+1);
            $HTML ->BR;
         }
      }
   }
   $HTML -> HR if $goodOptions;
 # If package has a floating scene, print button to floating scene page.
   if ($floatscene) {
      $HTML -> FONT(SIZE=>'+2') -> B
            -> msubmit_button(CAPTION =>
                   'Ok! Accept my choice & go on to subsetting page!',
                              QUERY =>['mode'    => 'FLOATINGSCENE',
                                       'pkg_id'  => $packages[0]->id ]);
      $HTML -> _B -> _FONT;

   } elsif ($subpage) {
      $HTML -> FONT(SIZE=>'+2') -> B
            -> msubmit_button(CAPTION =>
               'Continue',
                              QUERY =>['mode'    => 'PKGSLCT',
                                       'submode' => 'SUBPAGE',
                                       'origin' => 'PKGSLCT',
                                       'optnum' => $opt_num,
                                       'page' => 1]);
      $HTML -> _B -> _FONT;
   } else {
   # if no floating scene subsetting then display button to save options
   # and return to the shopping cart.

      if(!$goodOptions) {
         $HTML -> t("Sorry there is no media type that is large enough to");
         $HTML -> t(" handle your granule. ");
         $HTML -> hypertext(nobreak("Contact User Support at $DAAC."),
                            "$IU::ImsUserUrl"."#$DAAC",
                            TARGET => "support");
         $HTML -> t(" You will need to remove this granule from the ");
         $HTML -> t("shopping cart, before you can complete this order.");
         $HTML -> HR;
         $HTML -> _B -> _FONT;
      }
      # The "all-or-one" option:
      if ($goodOptions && ($item->order_thingy =~ /granule/i)) {
         print_all_or_one_menu($item, $dset);
      }

      $HTML -> FONT(SIZE=>'+2') -> B;
      if ($goodOptions) { 
         $HTML -> msubmit_button(CAPTION =>
                  'Ok! Accept my choice & return to the shopping cart!',
                                 QUERY => ['mode' => 'CART',
                                           'submode' => $next_submode]);
      
      } else {
         $HTML -> msubmit_button(CAPTION =>
                  'Cancel & return to the shopping cart!',
                                 QUERY => ['mode' => 'CART',
                                           'submode' => $next_submode]);
      }
      $HTML -> _B -> _FONT;
   }
   1;
}

#------------------------------------------------------------
# print_index
#------------------------------------------------------------
sub print_index {
   my ($item, $packagesref, $dset) = @_;
   my @packages = @$packagesref;

   # Get the granule or data set key and origin tag to pass along:
   my $gkey = $REQ->param('gkey');
   my $dskey = $REQ->param('dskey');   
   my $origin = $REQ->param('origin');
   my @tags = ('origin' => $origin);
   push (@tags, 'gkey' => $gkey) if (defined ($gkey));
   push (@tags, 'dskey' => $dskey) if (defined ($dskey));

   # If item is the data set, then next submode is DATASET_ONE.
   # If it is a granule, then next submode is SELECT_ONE.
   my $next_submode;
   if ($item == $dset) {
      $next_submode = 'DATASET_ONE';
   } else {
      $next_submode = 'SELECT_ONE';
   }
   $HTML -> msubmit_flag;

   # Print directions:
   $HTML -> HR;

   # Menu of available accounts to choose from:
   print_account_menu($dset);

   # The dataset/granule header:
   print_header($item);


   # If you want to display the Handling Fee(s) for a dataset 
   # uncomment this line out :
   #   print_handling_fee($dset);


   # If the item already has any packages selected
   # put a note indicating the selected one(s).

   # (This gets the selected package combos for this item
   #  and gets the package IDs from them)
   my @selected_package_ids = map {$_->package_id}
                              @{$item->order_pkgcombos};
   my $pkgno;
   my $pkgid;
   if (defined (@selected_package_ids) &&
       int(@selected_package_ids)) {

      # List of selected option numbers (not IDs).
      my @selected = ();

      # Find the option number(s)...  Check each id stored in item.
      for $pkgid (@selected_package_ids) {
         for $pkgno (0..$#packages) {
            if ($pkgid eq $packages[$pkgno]->id) {
               push (@selected, $pkgno+1);  # Add to list.
               last;     # Check next pkgid in item.
            }
         }
      }

      $HTML -> HR;

      # If no matches, something wrong...
      unless (int(@selected)) {
         puke 600, "Package ID ", $selected_package_ids[0], " for ",
                   "item isn't in the list of options for this item!";
      }
      print "This ", $item->order_thingy, " currently has Option ";
      print join (", option ", @selected), " selected.<BR>\n",
            "Any selections you make will <B>replace</B> your ",
            "previous choices.<BR>\n";
   }
   # Package index menu:
   $HTML -> H3 -> t("The ordering options are: ") -> _H3;
   my $opt;
   my $title;
   for $pkgno (0..$#packages) {
      $opt = 'Option ' . ($pkgno + 1);
      $title = $packages[$pkgno]->title;
      $HTML -> hspace(4)
            -> msubmit_button(CAPTION => $opt,
                              QUERY =>['mode'    => 'PKGSLCT',
                                       'submode' => $next_submode,
                                       @tags,
                                       'pkg_id'  =>
                                                $packages[$pkgno]->id,
                                      ],
                              ALT => $title)
            -> hspace(3);
      # If there are any miscellaneous URLs in the package print them.
      # Note that misc_URLs() returns ref to array that might be [].
      my $misc_URLs_ref = $packages[$pkgno]->misc_URLs;
      if (int (@{$misc_URLs_ref})) {
         my $tbl = new IW::Widget::URLTable
            "$title" => $misc_URLs_ref;
         $tbl->print_short;
      } else {
         $HTML -> t("$title");
      }
      $HTML -> BR;
   }
   1;
}

#------------------------------------------------------------
# any_titles  ARRAYREF
#------------------------------------------------------------
# Input: reference to array of packages.
# Output: Returns true if any package has a title or if any package
# has flag LINE_ITEMS_PER_GRANULE. Else false.

sub any_titles {
   my $pkgref = shift;
   my @packages = @$pkgref;
   my $package;
   foreach $package (@packages) {
      return 1 if ($package->title or $package->line_items_per_item);
   }
   return 0;
}

#------------------------------
#
# SUBMODE HANDLERS...
#
#------------------------------

#------------------------------------------------------------
# submode_DEFAULT
#------------------------------------------------------------
sub submode_DEFAULT {
    shift->submode_SELECT;
}

#------------------------------------------------------------
# submode_SUBPAGE
#------------------------------------------------------------
sub submode_SUBPAGE {
   my $self = shift;
   my $gran;
   my $dset;
   my $opt_num = $REQ->param('optnum') || 1;
   my $page = $REQ->param('page');

   # GET...

   # Get the granule key:
   my $gkey = $REQ->param('gkey') || 
      puke 600, "no data granule key information";
   my $origin = $REQ->param('origin');

   # Log into CART database (creating it if necessary):
   $CART = IU::InvStore->open($SESSION->path_to('cart.gdb'),
                              'WRCREAT')
       || puke "couldn't open cart"; 

   # Get the granule (from the shopping cart):
   ($gran) = $CART->granules([$gkey]);
   $gran or puke 533, "data granule $gkey is not in the shopping cart!";

   $DAAC = $gran->archive_id;
   # Get associated dataset (from the shopping cart):
   my $dskey = $gran->dataset_key;
   ($dset) = $CART->datasets([$dskey]);
   $dset or puke 600, "data set $dskey isn't in the shopping cart!";

   # Get package info:
   my @packages = IU::Package->lookup($gran->key,
                                       Dataset => $dset,
                                       Granule => $gran,
                                       SubDir => 'search');

   @packages or puke 560, "Data Granule claims to have ordering ",
      "options, but none were supplied.";

   # Set the navigation indicator.
   $SESSION->{Navigator}->{Indicator} = "OrderData";

   # Output the page header (with HTTP header):
   $APP->start_page(Title      =>"Choose Ordering Options",
                    SuperTitle =>'Shopping Cart',
                    Context    =>'choosepkg');
   # Begin the form
   $HTML -> script_FORM;

   foreach my $p ($REQ->params()) {
      next if($p !~ /^pk/);
      # Some parameters may have multiple values so loop over them.
      # In an array context $REQ->param() always returns an array
      # even if there is only one value. So if param is single value
      # then this will loop just once, but if it is multi-valued then
      # then it will loop until all values have been stored.
      foreach my $v ($REQ->param($p)) {
         next if(!defined($v) || ($v eq ''));
         $HTML->hidden($p => $v);
      }
   }

   # Submode for form depends on whether we have the title list
   # or the packages on this page.
   $HTML -> hidden('gkey'    => $gkey,
                   'origin'  => $origin);
   $HTML -> msubmit_flag;
   # If item is the data set, then next submode is SETPKGS_DATASET.
   # If it is a granule, then next submode is SETPKGS.
   my $next_submode;
   $next_submode = 'SETPKGS';
   $HTML -> HR;

   # The dataset/granule header:
#   print_header($gran);

   my $order_opts = $packages[$opt_num-1]->options('ORDER');
   my $subset_opts = $packages[$opt_num-1]->options('SUBSET');
   my $max;
   my $min;
   ($max, $min) = max_min($order_opts) if $order_opts;
   ($max, $min) = max_min($subset_opts) if $subset_opts;
   $HTML ->t("The required minimum number of Ground Control Points is $min.");
   $HTML -> BR;
   $HTML ->t("The maximum number of Ground Control Points is $max.");
   $HTML -> BR;
   ($max, $min) = print_subpage_menu($gran, $packages[$opt_num-1],$opt_num,
    Opts=>$order_opts, Page=>$page) if $order_opts;
   ($max, $min) = print_subpage_menu($gran, $packages[$opt_num-1],$opt_num,
    Opts=>$subset_opts, Page=>$page) if $subset_opts;
   $HTML -> HR;
   $HTML -> FONT(SIZE=>'+2') -> B;
   $page = $max + 1;
   $HTML -> msubmit_button(CAPTION =>
            'Ok! Accept my choice & return to the shopping cart!',
            QUERY => ['mode' => 'CART',
                      'submode' => $next_submode,
                      'optnum' => $opt_num,
                      'page' => $page,
                      'min' => $min]);
   # The "reset" button:
   $HTML -> INPUT(TYPE=>'RESET',
                  VALUE=>'Reset this form');

   $HTML -> _B -> _FONT;
   $HTML -> _script_FORM;

   # The signature:
   $APP->end_page();
   1;
}


#------------------------------------------------------------
# submode_SELECT
#------------------------------------------------------------
sub submode_SELECT { 
    my $self = shift; 
    my $gran;
    my $dset;

    # GET...

    # Get the granule key:
    my $gkey = $REQ->param('gkey') || 
       puke 600, "no data granule key information";
    my $origin = $REQ->param('origin');

    # Log into CART database (creating it if necessary):
    $CART = IU::InvStore->open($SESSION->path_to('cart.gdb'),
                               'WRCREAT')
        || puke "couldn't open cart"; 

    # Get the granule (from the shopping cart):
    ($gran) = $CART->granules([$gkey]);
    $gran or puke 533, "data granule $gkey is not in the shopping cart!";

    $DAAC = $gran->archive_id;
    # Get associated dataset (from the shopping cart):
    my $dskey = $gran->dataset_key;
    ($dset) = $CART->datasets([$dskey]);
    $dset or puke 600, "data set $dskey isn't in the shopping cart!";

    # Get package info:
    my @packages = IU::Package->lookup($gran->key,
                                       Dataset => $dset,
                                       Granule => $gran,
                                       SubDir => 'search');
    # my @package_ids = @{$gran->package_ids};

    @packages or puke 560, "Data Granule claims to have ordering ",
       "options, but none were supplied.";
    # Set the navigation indicator.
    $SESSION->{Navigator}->{Indicator} = "OrderData";

    # If there are TITLE's in the packages, print them as a
    # selection index instead of printing the packages on this
    # page.  However, if there is only one package, print it here.
    # If there are no titles, print the packages here.

    # Coding trick: if one package, $#packages is 0 so the following
    # "&&" never gets evaluated and $anytitles is set to 0.
    # If there are multiple packages, then clause after "&&" gets
    # evaluated and then $anytitles is set to its return value.
    my $anytitles = $#packages && any_titles(\@packages);

    # PRINT

    # Output the page header (with HTTP header):
    $APP->start_page(Title      =>'Choose Ordering Options',
                     SuperTitle =>'Shopping Cart',
                     Context    =>'choosepkg');

    # Begin the form
    $HTML -> script_FORM;

    # Submode for form depends on whether we have the title list
    # or the packages on this page.
    $HTML -> hidden('gkey'    => $gkey,
                    'origin'  => $origin);
    if ($anytitles) {
       # If titles, print index.
       print_index($gran, \@packages, $dset);
    } elsif ($#packages eq 0) {
       # Just one package so print it.
       print_form($gran,\@packages, $dset, '1'); 

    } else {
       # Else print packages. Print form.
       print_form($gran, \@packages, $dset);
    }
    $HTML -> _script_FORM;

    # The signature:
    $APP->end_page();
    1;
}

#------------------------------------------------------------
# submode_SELECT_ONE
#------------------------------------------------------------
sub submode_SELECT_ONE {
    my $self = shift;
    my $gran;
    my $dset;

    # GET...

    # Get the granule key:
    my $gkey = $REQ->param('gkey') || 
        puke 600, "no data granule key information";
    my $origin = $REQ->param('origin');

    my $pkg_id = $REQ->param('pkg_id') ||
        puke 600, "no package id from index of options";

    # Log into CART database (creating it if necessary):
    $CART = IU::InvStore->open($SESSION->path_to('cart.gdb'),
                               'WRCREAT')
        || puke "couldn't open cart"; 

    # Get the granule (from the shopping cart):
    ($gran) = $CART->granules([$gkey]);
    $gran or puke 533,
        "data granule $gkey is not in the shopping cart!";

    # Get associated dataset (from the shopping cart):
    my $dskey = $gran->dataset_key;
    ($dset) = $CART->datasets([$dskey]);
    $dset or puke 600, "data set $dskey isn't in the shopping cart!";

    # Get package info:
    my @packages = IU::Package->lookup($gran->key,
                                       Dataset => $dset,
                                       Granule => $gran,
                                       SubDir => 'search');
    @packages or puke 560, "Data Granule claims to have ordering ",
        "options, but none were supplied.";

    # Find the selected package.
    my $pkgno;
    for ($pkgno = 0; $pkgno <= $#packages; $pkgno++) {
       last if ($pkg_id eq $packages[$pkgno]->id);
    }
    if ($pkgno > $#packages) {
       puke 600, "Package ID $pkg_id for selected option isn't in ",
                 "the list of options for this data granule.";
    }
    # Reduce array of packages to just the one package.
    @packages = ($packages[$pkgno]);

    # Set the navigation indicator.
    $SESSION->{Navigator}->{Indicator} = "OrderData";

    # PRINT...

    # Output the page header (with HTTP header):
    $APP->start_page(Title      =>'Choose Ordering Options',
                     SuperTitle =>'Shopping Cart',
                     Context    =>'choosepkg');


    # Begin the form, and add the hidden tags:
    $HTML -> script_FORM;
    $HTML -> hidden('gkey'    => $gkey,
                    'origin'  => $origin);
    # Print form:
    $HTML -> HR;
    print_form($gran, \@packages, $dset, $pkgno+1);
    $HTML -> _script_FORM;

    # The signature:
    $APP->end_page();
    1;
}

#------------------------------------------------------------
# submode_DATASET 
#------------------------------------------------------------
# Output the menu of packaging options for a dataset.
#
sub submode_DATASET {
    my $self = shift;
    my $dset;

    # GET...

    # Get the data set key:
    my $dskey = $REQ->param('dskey') 
        || puke 600, "no data set key information";

    my $origin = $REQ->param('origin');

    # Log into CART database (creating it if necessary):
    $CART = IU::InvStore->open($SESSION->path_to('cart.gdb'), 'WRCREAT')
        || puke "couldn't open cart"; 

    # Get associated dataset (from the shopping cart):
    ($dset) = $CART->datasets([$dskey]);
    $dset or puke 600, "data set $dskey isn't in the shopping cart!";

    # Get package info:
    my @package_ids = @{$dset->package_ids};
    @package_ids or puke 560, "Data set is not orderable - " .
                              "no ordering options were supplied.";
    my @packages = IU::Package->lookup_dataset($dskey,
                                               Dataset => $dset,
                                               SubDir => 'search');
    @packages or puke 560, "Data set claims to have ordering options, ",
                           "but none were supplied.";

    # Set the navigation indicator.
    $SESSION->{Navigator}->{Indicator} = "OrderData";

    # If there are TITLE's in the packages, print them as a
    # selection index instead of printing the packages on this
    # page.  However, if there is only one package, print it here.
    # If there are no titles, print the packages here.

    # Coding trick: if one package, $#packages is 0 so the following
    # "&&" never gets evaluated and $anytitles is set to 0.
    # If there are multiple packages, then clause after "&&" gets
    # evaluated and then $anytitles is set to its return value.
    my $anytitles = $#packages && any_titles(\@packages);

    # PRINT...

    # Output the page header (with HTTP header):
    $APP->start_page(Title      =>'Choose Ordering Options',
                     SuperTitle =>'Shopping Cart',
                     Context    =>'choosepkg');

    # Begin the form, and add the hidden tags:
    $HTML -> script_FORM;

    # Print message:
    print "There are a total of <B>", int(@packages),"</B> ",
          "different Ordering Options available for this data ",
          "set.<BR>\n";

    # Submode for form depends on whether we have the title list
    # or the packages on this page.
    if ($anytitles) {
       $HTML -> hidden('dskey'    => $dskey,
                       'origin'  => $origin);
       # If titles, print index.
       print_index($dset, \@packages, $dset);
    } else {
       $HTML -> hidden('dskey'    => $dskey,
                       'origin'  => $origin);
       # Else print packages.  Print form.
       print_form($dset, \@packages, $dset);
    }
    $HTML -> _script_FORM;

    # The signature:
    $APP->end_page();
    1;
}

#------------------------------------------------------------
# submode_DATASET_ONE
#------------------------------------------------------------
sub submode_DATASET_ONE {
   my $self = shift;
   my $dset;

   # Get the data set key:
   my $dskey = $REQ->param('dskey') 
      || puke 600, "no data set key information";

   my $origin = $REQ->param('origin');

   # Log into CART database (creating it if necessary):
   $CART = IU::InvStore->open($SESSION->path_to('cart.gdb'),'WRCREAT')
      || puke "couldn't open cart"; 

   # Get associated dataset (from the shopping cart):
   ($dset) = $CART->datasets([$dskey]);
   $dset or puke 600, "data set $dskey isn't in the shopping cart!";

   # Get package info:
   my @package_ids = @{$dset->package_ids};
   @package_ids or puke 560, "Data set is not orderable - ",
                      "no ordering options were supplied.";
   my @packages = IU::Package->lookup_dataset($dskey,
                                              Dataset => $dset,
                                              SubDir => 'search');
   @packages or puke 560, "Data set claims to have ordering ",
                   "options, but none were supplied.";

   my $pkg_id = $REQ->param('pkg_id') ||
       puke 600, "no package id from index of options.";

   # Find the selected package.
   my $pkgno;
   for ($pkgno = 0; $pkgno <= $#packages; $pkgno++) {
      last if ($pkg_id eq $packages[$pkgno]->id);
   }
   if ($pkgno > $#packages) {
      puke 600, "Package ID $pkg_id for selected option isn't in ",
         "the list of options for this data set.";
   }
   # Reduce array of packages to just the one package.
   @packages = ($packages[$pkgno]);

   # Set the navigation indicator.
   $SESSION->{Navigator}->{Indicator} = "OrderData";

   # PRINT...

   # Output the page header (with HTTP header):
   $APP->start_page(Title      =>'Choose Ordering Options',
                    SuperTitle =>'Shopping Cart',
                    Context    =>'choosepkg');

   # Begin the form, and add the hidden tags:
   $HTML -> script_FORM;
   $HTML -> hidden('dskey'    => $dskey,
                   'origin'  => $origin);
   # Print form:
   $HTML -> HR;
   print_form($dset, \@packages, $dset, $pkgno+1);
   $HTML -> _script_FORM;

   # The signature:
   $APP->end_page();
   1;
}

#------------------------------------------------------------
# go
#------------------------------------------------------------
# Main entry point.

sub go {
    my $self = shift;   
    $APP->get_session();     # get session or die
    my $submode = ($REQ->param('submode') || 'DEFAULT');
    my $method = "submode_\U$submode\E";
    $self->$method() or puke "failed while handling submode $submode!";
}

#------------------------------------------------------------
1;
