package IW::ModePREFSPKG;

=head1 NAME

IW::ModePREFSPKG - package selection from the User Preferences page

=head1 AUTHOR

Mark Solomon, GST, based on ModePKGSLCT.

=head1 RCS INFO

$Id: ModePREFSPKG.pm,v 1.13 2003/02/14 16:53:52 solomon Exp $

=cut

use IU;
use IU::Coverage;
use IU::Dataset;
use IU::Granule;
use IU::InvStore;
use IU::Log;
use IU::MediaHandlingFee;
use IU::MediaMaximumCapacity;
use IU::Package;
use IU::PkgCombo;
use IU::Session;
use IU::Util qw(shellquote);

use IW::App;
use IW::Comm::AccountStatusResult;
use IW::Mode;
use IW::User;
use IW::Util qw(:url :string);
use IW::Widget::URLTable;

@ISA = qw(IW::Mode);

use strict;

#------------------------------------------------------------
#
# Globals
#
#------------------------------------------------------------

# The databases:
my $CART;

# Hash containing medias too small for this granule.
my %SmallMedias;

# Hash containing the names of the $user subroutines
my %user_subs = 
   ( 'FTPHOST' => 'ftp_host',
     'FTPPASSWORD' => 'ftp_password',
     'FTPPUSHDEST' => 'ftp_pushdest',
     'FTPUSER' => 'ftp_user',
     'USERSTRING' => 'ftp_userstring');
my @user_subs_keys = keys %user_subs;
my $sub_key; 
my $origin;
#------------------------------------------------------------
# print_small_media_table 
#------------------------------------------------------------
sub print_small_media_table {

   my $dset = shift;
   my $gran = shift;
   my $pkgref = shift;

   %SmallMedias = ();

   # loop through all the package combos and build a hash of
   # unique media types;
   my $package;
   my %combo_types;
   my @combos;
   my $combo;
   my $media_type;
   foreach $package (@{$pkgref}) {
      foreach $combo ($package->combos) {
         $media_type = $combo->media_type;
         $combo_types{"$media_type"} ='1';
      }
   }

   # Check to see if any of the media capacity limit applies to this package.
   # If some do then output the table.
   my $granule_size = $gran->size_mbecs_granule;
   my $mediamaximumcapacity;

   foreach $mediamaximumcapacity (@{$dset->mediamaximumcapacity}) {
      if ($granule_size > $mediamaximumcapacity->capacity) {
         $SmallMedias{$mediamaximumcapacity->type} =
                             $mediamaximumcapacity->capacity;
      }
   }

   my @types = sort keys %SmallMedias;
   my $capacity;
   if (scalar(@types) > 0) {
      print '<P>The following media are not available for this granule <BR>';
      print 'due to insufficient capacity: <BR>';
      print '<TABLE BORDER=1>';
      print '<TR><TD><B>Media type</B></TD>';
      print '<TD><B>Media Capacity (MB)</B></TD>';
      print '<TD><B>Granule Size (MB)</B></TD></TR>';
      foreach $media_type (@types) {
         $capacity = $SmallMedias{$media_type};
         print '<TR>';
         print '<TD>',"$media_type",'</TD>';
         print '<TD>',"$capacity",'</TD>';
         print '<TD>',"$granule_size",'</TD>';
         print '</TR>';
      }
      print '</TABLE><P>';
   }
}

#------------------------------------------------------------
# print_media_fee_table DSET 
#------------------------------------------------------------
sub print_media_fee_table {

   my $dset = shift;
   my $pkgref = shift;

   # loop through all the package combos and build a hash of 
   # unique media types;
   my $package;
   my %combo_types;
   my @combos;
   my $combo;
   my $media_type;
   foreach $package (@{$pkgref}) {
      foreach $combo ($package->combos) {
         $media_type = $combo->media_type; 
         $combo_types{"$media_type"} ='1'; 
      }    
   }
   # Check to see if any of the media handling fees apply to this package.
   # If some do then output the table.
   my $count = 0;
   my ($mediahandlingfee,$type,$fee,$comment,$url);
   foreach $mediahandlingfee (@{$dset->mediahandlingfee}) {
      ++$count if ($combo_types{$mediahandlingfee->type});
   }
   if ($count > 0) {
      print '<TABLE BORDER=1>';
      print '<TR><TD><B>Media type</B></TD><TD><B>Fee (US)</B></TD>';
      print '<TD><B>Comments</B></TD></TR>';
      foreach $mediahandlingfee (@{$dset->mediahandlingfee}) {
         $type = $mediahandlingfee->type; 
         # Filter out those media handling fees which don't apply for this 
         # set of package combos. I.e. Only display it if it is one of the 
         # package combos.
         if ($combo_types{"$type"}) {
            $fee = $mediahandlingfee->fee;
            $fee = sprintf("%0.2f",$fee);
            $comment = $mediahandlingfee->comment; 
            $url = $mediahandlingfee->url;
            print '<TR>';
            print '<TD>',"$type",'</TD>';
            print '<TD> $',"$fee",'</TD>';
            print '<TD>';
            print "$comment " if ($comment);
            $HTML -> A(HREF=>$url->URL) -> t($url->comment) -> _A if ($url);
            print '</TD>';
            print '</TR>';
         }
      }
      print '</TABLE><P>';
   }
}

#------------------------------------------------------------
# print_header ITEM
#------------------------------------------------------------
# Print the header at the top of the page, showing the dataset
# and granule id.  Also include geographic info if known,
# area and center.
# ITEM is any orderable item.
sub print_header {
   my $item = shift;
   # The title and links (note: NO dataset attribute button for now!):
   if ($item->dataset_key) {       # the item "has a relevant dataset"
       $HTML -> BR;
       $HTML -> FONT(SIZE=>'+1') 
             -> t_B($item->dataset_id) -> _FONT 
             -> BR;
   }
       $HTML -> BR;
   if ($item->granule_key) {       # the item "has a relevant granule"
      $HTML -> gran_attr_btn($item->granule_key)
            -> hspace(2)
            -> FONT(SIZE=>'+1') -> B
            -> I -> t(' Data Granule ID: ') -> _I 
            -> t($item->granule_id) -> _B -> _FONT
            -> BR;
   }
   1;
}

#------------------------------------------------------------
# print_package_menu ITEM, PACKAGE, PKGNO
#------------------------------------------------------------
# Print the menu of packaging "combos" for a single package.
# The ITEM is an orderable item (dataset, granule, etc.)
# If there is a title in the package, print the title in
# the first cell and the option number above.  Otherwise,
# just print option number in the first cell.
# Note PKGNO here is 1 based, not 0 based.
#
sub print_package_menu {
    my ($item, $package, $pkgno,$floatscene) = @_;
    my $item_thingy = $item->order_thingy;

    # Get all currently-selected package combos for this granule,
    # in a hash:
    my %selected = map {($_->as_hashkey => 1)}
                       @{$item->order_pkgcombos};

    # Print header, with package name:
    $HTML -> TABLE(BORDER=>3, CELLPADDING=>3, CELLSPACING=>0,
                   WIDTH=>'100%', 
                   BGCOLOR=>$HTML->hue('PkgItemsTblHeader'));
    $HTML -> TR(VALIGN=>'TOP');
    $HTML -> TD;
    $HTML -> color(0, 40, 80);

    my $title = $package->title;   # May be '' == false.
    if ($title) {
        $HTML -> H2 -> t("Ordering Option $pkgno:") -> BR
              -> t($title) -> _H2;
    } else {
        $HTML -> H2 -> t("Ordering Option: ", $pkgno) -> _H2;
    }
    $HTML -> _color;
    $HTML -> _TD;

    # Print comment, if any:
    $HTML -> TD -> small;

    if ($package->comment) {
        $HTML -> t($package->comment) ->BR;
    }
    # If there are any miscellaneous URLs in the package print them.
    # Note that misc_URLs() returns ref to array that might be [].
    my $misc_URLs_ref = $package->misc_URLs;
    my $ref = $package->{'MISC_URL'};
    my @furls = @$ref;
    if (int (@{$misc_URLs_ref})) {
       my $tbl = new IW::Widget::URLTable 
          "More info:" => $misc_URLs_ref;
       $tbl->print;
    }
    $HTML -> _small -> _TD;
    $HTML -> _TR;

    # Get the information prompt or default it.
    my $info_prompt;
    if ($package->info_prompt) {
       $info_prompt = $package->info_prompt;
    } elsif ($item->archive_id =~ /^NOAA/i) {
       $info_prompt = 'Please enter your NOAA account number: ';
    } else {
       $info_prompt = '';
    }
    my $addn_info = '';
    # Find if this package has a previously set additional info.
    my $comb;
    # For each combo previously selected for this item:
    foreach $comb (@{$item->package_combos}) {
       # If this selected combo is the current package:
       if ($package->id eq $comb->package_id) {
          $addn_info = $comb->addn_info;
          last;                       # break if we find it.
       }
    }
#    unless ($floatscene) {
       $HTML -> TR -> TD(COLSPAN=>2);
       # If the package has an info prompt, print it.
       if ($info_prompt) {
          $HTML -> small -> t($info_prompt) -> _small -> BR;
       }
       # Print additional info box.
       $HTML -> field_cell(Title=>'Additional Info: ', Stack=>0)
             -> INPUT(TYPE=>'text', SIZE=>80, MAXLENGTH=>80,
                   NAME=>"additional_info.$pkgno",
                   VALUE=>$addn_info)
             -> _field_cell;
       $HTML -> small -> t("") -> _small;
       $HTML -> _TD -> _TR;
#    }
    $HTML -> _TABLE;

    # Setup table and header:
    $HTML -> TABLE(BORDER=>3, CELLPADDING=>2, CELLSPACING=>0,
                   WIDTH=>'100%',
                   BGCOLOR=>$HTML->hue('HiBack'))
          -> TR(BGCOLOR=>$HTML->hue('PkgItemsTblHeader'))
          -> TH -> FONT(SIZE=>'-1') -> t_B('Select'); 
    $HTML -> BR;
    $HTML -> t_B('One') -> _FONT -> _TH;
    $HTML -> FONT(SIZE=>'+1');
    $HTML -> TH -> t_B('Data Format') -> _TH
          -> TH -> t_B('Media type') -> _TH
          -> TH -> t_B('Media format') -> _TH
          -> TR;
    $HTML -> _FONT;
    # Output each row of the packages menu:
    my @combos = $package->combos;
    @combos or puke 457, "Ordering information for $item_thingy not available";

    my $anything_chosen = 0;
    my $combo;
    my $choice = 0;
    my $small_media;
    my $user = $SESSION->state->{User} || new IW::User;
    
    COMBO: foreach $combo (@combos) {
        foreach $small_media (keys %SmallMedias) {
           next COMBO if $combo->media_type == $small_media;
        }
        ++$choice;
        
        # Should this choice be on?
        my $onoff = $selected{$combo->as_hashkey};
	$onoff = 1 if(($user->data_format eq $combo->proc_opt) &&
                      ($user->media_type eq $combo->media_type) &&
                      ($user->media_format eq $combo->media_fmt));
        $anything_chosen ||= $onoff;

        # Precompute some values:
        my $cost = sprintf("\$%.02f", $combo->cost);

        # Leave space for URLs?
        my $has_URLs = @{$combo->data_URLs};

        # OUTPUT THE ROW...
        
        $HTML -> TR;

        # Checkbox...
        $HTML -> TD(ALIGN=>'LEFT', ROWSPAN=>($has_URLs ? 2 : 1))
              -> FONT(SIZE=>'+2') -> B;
        if ($package->allows_multiple_media) {

            $HTML -> checkbox    (Name  =>"pk-choice.$pkgno",  # "pk-choice.1"
                                  Choice=>"$pkgno.$choice",    # "1.2"
                                  Value =>$onoff);
            $HTML -> t(" ");

           # determine if the user already selected "number of copies"
           # for this combo, if so then initialize the form.

           my $copies = '1';
               
           my $my_combo; 
           my $pkg_array;
           foreach $pkg_array ($item->order_pkgcombos) {

               foreach $my_combo (@$pkg_array) {

                   if ($my_combo->merged_key eq $combo->merged_key) { 
                       $copies = $my_combo->copies;
                   } # end if

               } # combo foreach                  
           } # end pkg foreach 

           $HTML -> field_cell(Title=>'Copies: ', Stack=>0)
                 -> INPUT(TYPE=>'text', SIZE=>3, MAXLENGTH=>3,
                           NAME=>"copies.$pkgno.$choice",
                           VALUE=>"$copies")      
                 -> _field_cell;
        }
        else {
            $HTML -> radio_button(Name   =>"pk-choice.$pkgno", # "pk-choice.1"
                                  Choice =>"$pkgno.$choice",   # "1.2"
                                  IsDefault =>$onoff);
        }
#       $HTML -> I -> t(" $pkgno.$choice") -> _I;       # "1.2"
        $HTML -> _B -> _FONT;
        $HTML -> _TD;

        # Package id...
        $HTML -> TD(ALIGN=>'CENTER') 
              -> small -> t($combo->proc_opt)   -> _small -> _TD;
        $HTML -> TD(ALIGN=>'CENTER') 
              -> small -> t($combo->media_type) -> _small -> _TD;
        $HTML -> TD(ALIGN=>'CENTER') 
              -> small -> t($combo->media_fmt)  -> _small -> _TD;
        $HTML -> _TR;
        # URL row?
        $has_URLs or next;
        $HTML -> TR;
        $HTML -> TD(ALIGN=>'LEFT', VALIGN=>'TOP', COLSPAN=>5);  
        my $furl;
        $HTML -> small -> UL;
        foreach $furl (@{$combo->data_URLs}) {
            $HTML -> LI
                  -> hypertext("[DATA]", 
                               logged_url($furl->URL, 'type'=>'DATA'))
                  -> t(" ", $furl->comment);
        }
        $HTML -> _UL -> _small;
        $HTML -> _TD -> _TR;
    }

    # Final row, if needed:
    if (!$package->allows_multiple_media) {
        $HTML -> TR -> TD(COLSPAN=>6);
        $HTML -> radio_button(Name   =>"pk-choice.$pkgno", # "pk-choice.1"
                              Choice =>"(none)",
                              IsDefault => !$anything_chosen);
        $HTML -> t_EM(" I want no items from this option.");
        $HTML -> _TD -> _TR;
    }

    # Footer:
    $HTML -> _TABLE;
    1;
}

#------------------------------------------------------------
# print_options_menus ITEM,PACKAGE,PKGNO
#------------------------------------------------------------
# Print the menu of subsetting options for a single package.
# Each menu's CGI tags will begin with "pk-opts.PKGNO."
# where PKGNO is 1, 2, etc.
#
# ITEM is any orderable item.

sub print_options_menus {
    my ($item, $package, $pkgno) = @_;
    
    # Stop immediately if no subsetting criteria:
    my $opts_type;
    foreach $opts_type (sort $package->options) {
        my $opts = $package->options($opts_type) || die "hunh?";

        $HTML -> P;
        $HTML -> IMG(SRC=>doc_url("icons/arrowup.gif"), 
                     ALIGN=>'LEFT',
                     HSPACE=>4);
        print_options_menu($item, $package, $pkgno, Opts=>$opts);
    }
}

#------------------------------------------------------------
# print_options_menu ITEM,PACKAGE,PKGNO,OPTIONS
#------------------------------------------------------------
# Print the menu of subset/order options for a single package.
# Each entry's CGI tags will begin with "pk-opts.PKGNO.OPTSTYPE",
# where PKGNO is 1, 2, etc, and OPTSTYPE is the options type for
# this options menu (e.g., "SUBSET", "ORDER", etc.).
#
# ITEM is any orderable item.

sub print_options_menu {
    my ($item, $package, $pkgno, %params) = @_;
    my $opts     = $params{Opts};
    my $optstype = $params{OptsType}  || $opts->type;
    my $title    = $params{OptsTitle} || $opts->pretty_type;
    my $user = $SESSION->state->{User} || new IW::User;

    # Do we have any subsetting options CURRENTLY SELECTED for this package?
    # If so, set up the CGI tags:
    if ($item->order_pkgcombos) {

        # Find first combo with this package id:
        my $combo = (grep {$_->package_id eq $package->id
                           } @{$item->order_pkgcombos})[0];

        # If one was found, and if it has subset criteria, turn on the values!
        if ($combo and $combo->options($optstype)) {
            $combo->options($optstype)->set_CGI(Prefix=>
                                                "pk-opts.$pkgno.$optstype.");
        }
    }

    # Start the table of criteria:
    my @criteria = $opts->criteria;
    my $crit;
    $HTML -> TABLE(BORDER=>3, 
                   ###  WIDTH=>'100%', 
                   CELLPADDING=>3, CELLSPACING=>0,
                   BGCOLOR=>$HTML->hue('HiBack'));
    # Title:
    $HTML -> TR;
    $HTML -> TD(COLSPAN=>2, BGCOLOR=>$HTML->hue('PkgOptsTblHeader'))
          -> FONT(SIZE=>'+1') 
          -> t_B("\u$title (for Ordering Options $pkgno)") 
          -> _TD
          -> _FONT;

    # Directions:
    $HTML -> BR;
    if ($opts->is_required) {
        $HTML -> I -> t_B("Mandatory: ") -> _I -> small;
        print <<EOF;
              If you selected any of the items in ordering options $pkgno (above),
              you must
              give values for 
              any required
              Options below.  
              You can also give values for the other options, if you desire.
EOF
        $HTML -> _small;
    }
    else {
        $HTML -> I -> t_B("Optional: ") -> _I -> small;
        print <<EOF;
              If you selected any of the items in ordering options $pkgno (above),
              you can provide values for any of the options below.
              <B>Note:</B> if you supply values for <I>any</I> of these
              options, then you must be sure to supply values for
              <I>all</I> of the required 
              options.
EOF
        $HTML -> _small;
    }   
    $HTML -> _TR;
    # Output each special criterion:
    foreach $crit (@criteria) {
        my $preset = undef;
        $HTML -> TR;
        foreach $sub_key (@user_subs_keys) {
           if($sub_key eq $crit->name) {
              my $newsub = $user_subs{$sub_key};
              $preset = $user->$newsub();
           }
        }
        $crit->print_form(Prefix=>"pk-opts.$pkgno.$optstype.", HTML=>$HTML,
                          Preset=>$preset);
        $HTML -> _TR;
    }    
    $HTML -> _TABLE;
    1;
}

#------------------------------------------------------------
# print_all_or_one_menu GRAN DSET
#------------------------------------------------------------
# Print the menu which asks if they want these options to be
# applied to all possible granules in the shopping cart, or just
# to the granule that was selected.

sub print_all_or_one_menu {
    my $gran = shift;
    my $dset = shift;

    # Print the all-or-one option:
    $HTML -> TABLE;

    # Default is originally set for all in data set: 'dataset'.
    # However, if the user set it to 'granule' for a previous
    # granule in the data set, then use that setting.  The setting
    # is stored in the data set object.  In the code, if the
    # setting is in the data set object, use that, else default.

    my $pkg_set_for = $dset->pkg_set_for;
    my $pkg_set_for_dataset = 1;              # default 'dataset'
    my $pkg_set_for_granule = 0;              # opposite

    # If not defined, use default above, else:
    if (defined($pkg_set_for)) {
       if ($pkg_set_for eq 'granule') {
          $pkg_set_for_dataset = 0;
          $pkg_set_for_granule = 1;
       }
    }
    # All...
    $HTML -> TR 
          -> TD -> radio_button(Name      =>'set-for',
                                Choice    =>'dataset',
                                IsDefault =>$pkg_set_for_dataset)
          -> _TD;
    $HTML -> TD;
    print "I want these ordering options for <I>every</I> data granule"; 
    print " that applies in data set "; 
    $HTML -> t_B($gran->dataset_id);
    print " in the shopping cart.";
    $HTML -> _TD -> _TR;

    # One...
    $HTML -> TR 
          -> TD -> radio_button(Name      =>'set-for',
                                Choice    =>'granule',
                                IsDefault =>$pkg_set_for_granule)
          -> _TD;
    $HTML -> TD;
    print "I want these ordering options <I>only</I> for data granule ";
    $HTML -> t_B($gran->id);
    print ".";
    
    $HTML -> _TD -> _TR;
    $HTML -> _TABLE;
}

#------------------------------------------------------------
# print_account_menu DSET
#------------------------------------------------------------
# Print the menu asking the user for their accounting information.

sub print_account_menu {
    my ($dset) = @_;
    my $acct;
    # Get the granule key:
    my $gkey = $REQ->param('gkey');
    my $dskey = $REQ->param('dskey');
    if($gkey) {
        $SESSION->state->{PackageGkey} = $gkey;
        $SESSION->state->{PackageDskey} = undef;
    }
    if($dskey) {
        $SESSION->state->{PackageGkey} = undef;
        $SESSION->state->{PackageDskey} = $dskey;
    }
    if(!$gkey) {
        $gkey = $SESSION->state->{PackageGkey};
    }
    if(!$dskey) {
        $dskey = $SESSION->state->{PackageDskey};
    }
    my $origin = $REQ->param('origin');
    1;
}

#------------------------------------------------------------
# print_handling_fee DSET
#------------------------------------------------------------
# Decides whether to print out any Handling Fees associated with 
# the Dataset on the Package Selection Page.

sub print_handling_fee {
   # We've been 'handed' the dataset
   my ($dset) = @_;

   # Is there a Handling Fee within the Dataset?
   # If so, create a BLANK HandlingFee Object 
   # and pass it the DATASET object
   # It'll print ALL the Handling Fee(s) for you.

   if (defined($dset->handling_fee)) {
       my $new_fee_object = IU::HandlingFee->new();
       $new_fee_object->print_handling_fees($dset);
   }
1;
}

#------------------------------------------------------------
# print_form ITEM, PACKAGES, DATASET [, OPTION_NUMBER ]
#------------------------------------------------------------
# Print the meat of the page.
# If optional OPTION_NUMBER is present, it means that
# PACKAGES will only contain one package whose number is
# actually the one given (1 based).
#
sub print_form {
   my ($item, $packagesref, $dset, $opt_num) = @_;
   my @packages = @$packagesref;

   $HTML -> msubmit_flag;
   # If item is the data set, then next submode is SETPKGS_DATASET.
   # If it is a granule, then next submode is SETPKGS.
   my $next_submode;
   if ($item == $dset) {
      $next_submode = 'SETPKGS_DATASET';
   } else {
      $next_submode = 'SETPKGS';
   }

   # Menu of available accounts to choose from:
   print_account_menu($dset);

   # The dataset/granule header:
   print_header($item);

   # the media fee summary table:
   print_media_fee_table($dset,\@packages); 

   my $floatscene = 0;  # are there any floating scene criteria. 
   # If optional OPTION_NUMBER is present, it means that
   # PACKAGES will only contain one package whose number is
   # actually the one given.
   if (defined ($opt_num)) {
      # Check if package has any floating scene criteria. 
      my $criteria;
      my $opts = $packages[0]->options('SUBSET');
      if (defined($opts)) {
         foreach $criteria ($opts->criteria) {
            $floatscene += 1 if ($criteria->type eq 'FLOATINGSCENE'); 
         } 
      }
      print "\n";
#      print_package_menu($item, $packages[0], $opt_num,$floatscene);
      $HTML -> P; 
      #Print only the ORDER options if there are floating scene criteria. 
      if ($floatscene) {
         my $order_opts = $packages[0]->options('ORDER');
         if (defined($order_opts)) {
            $HTML -> P;
            $HTML -> IMG(SRC=>doc_url("icons/arrowup.gif"),
                     ALIGN=>'LEFT',
                     HSPACE=>4);
            print_options_menu($item, $packages[0],$opt_num,Opts=>$order_opts);
         }
      } else {
         print_options_menus($item, $packages[0], $opt_num);
         $HTML ->BR;
      }
   } else {
      # print the media capacity limit table:
      print_small_media_table($dset, $item, \@packages);
      my $pkgno;
      # The package menu:
      for $pkgno (0..$#packages) {
         my $package = $packages[$pkgno];
         print "\n<HR>";
         print_package_menu($item, $package, $pkgno+1,"0");
         $HTML -> P;
         print_options_menus($item, $package, $pkgno+1);
         $HTML ->BR;
      }
   }
   $HTML -> HR;
   # If package has a floating scene, print button to floating scene page.
   if ($floatscene) {
      $HTML -> FONT(SIZE=>'+2') -> B
            -> msubmit_button(CAPTION =>
                   'Ok! Accept my choice & go on to subsetting page!',
                              QUERY =>['mode'    => 'FLOATINGSCENE',
                                       'pkg_id'  => $packages[0]->id ]);
      $HTML -> _B -> _FONT;
   } else {
   # if no floating scene subsetting then display button to save options
   # and return to the shopping cart.

      # The "all-or-one" option:
      if ($item->order_thingy =~ /granule/i) {
         print_all_or_one_menu($item, $dset);
      }

      if($origin eq 'CART') {
         $HTML -> FONT(SIZE=>'+2') -> B
               -> msubmit_button(CAPTION =>
                  'Ok! Accept my choice & return to the Shopping Cart!',
                  QUERY => ['mode' => 'PREFSPKG',
                            'submode' => 'CHOSEN',
                            'origin' => 'CART']);
         $HTML -> _B -> _FONT;
      } else {
         $HTML -> FONT(SIZE=>'+2') -> B
               -> msubmit_button(CAPTION =>
                  'Ok! Accept my choice & return to User Preferences!',
                  QUERY => ['mode' => 'PREFSPKG',
                            'submode' => 'CHOSEN',
                            'origin' => 'PREFS']);
         $HTML -> _B -> _FONT;
      }
   }
   1;
}

#------------------------------------------------------------
# print_index
#------------------------------------------------------------
sub print_index {
   my ($item, $packagesref, $dset) = @_;
   my @packages = @$packagesref;

   # Get the granule or data set key and origin tag to pass along:
   my $gkey = $REQ->param('gkey');
   my $dskey = $REQ->param('dskey');   
   my $origin = $REQ->param('origin');
   my @tags = ('origin' => $origin);
   push (@tags, 'gkey' => $gkey) if (defined ($gkey));
   push (@tags, 'dskey' => $dskey) if (defined ($dskey));

   # If item is the data set, then next submode is DATASET_ONE.
   # If it is a granule, then next submode is SELECT_ONE.
   my $next_submode;
   if ($item == $dset) {
      $next_submode = 'DATASET_ONE';
   } else {
      $next_submode = 'SELECT_ONE';
   }
   $HTML -> msubmit_flag;

   # Print directions:
   $HTML -> HR;

   # Menu of available accounts to choose from:
   print_account_menu($dset);

   # The dataset/granule header:
   print_header($item);


   # If you want to display the Handling Fee(s) for a dataset 
   # uncomment this line out :
   #   print_handling_fee($dset);


   # If the item already has any packages selected
   # put a note indicating the selected one(s).

   # (This gets the selected package combos for this item
   #  and gets the package IDs from them)
   my @selected_package_ids = map {$_->package_id}
                              @{$item->order_pkgcombos};
   my $pkgno;
   my $pkgid;
   if (defined (@selected_package_ids) &&
       int(@selected_package_ids)) {

      # List of selected option numbers (not IDs).
      my @selected = ();

      # Find the option number(s)...  Check each id stored in item.
      for $pkgid (@selected_package_ids) {
         for $pkgno (0..$#packages) {
            if ($pkgid eq $packages[$pkgno]->id) {
               push (@selected, $pkgno+1);  # Add to list.
               last;     # Check next pkgid in item.
            }
         }
      }

      $HTML -> HR;

      # If no matches, something wrong...
      unless (int(@selected)) {
         puke 600, "Package ID ", $selected_package_ids[0], " for ",
                   "item isn't in the list of options for this item!";
      }
      print "This ", $item->order_thingy, " currently has Option ";
      print join (", option ", @selected), " selected.<BR>\n",
            "Any selections you make will <B>replace</B> your ",
            "previous choices.<BR>\n";
   }
   # Package index menu:
   $HTML -> H3 -> t("The ordering options are: ") -> _H3;
   my $opt;
   my $title;
   for $pkgno (0..$#packages) {
      $opt = 'Option ' . ($pkgno + 1);
      $title = $packages[$pkgno]->title;
      $HTML -> hspace(4)
            -> msubmit_button(CAPTION => $opt,
                              QUERY =>['mode'    => 'PKGSLCT',
                                       'submode' => $next_submode,
                                       @tags,
                                       'pkg_id'  =>
                                                $packages[$pkgno]->id,
                                      ],
                              ALT => $title)
            -> hspace(3);
      # If there are any miscellaneous URLs in the package print them.
      # Note that misc_URLs() returns ref to array that might be [].
      my $misc_URLs_ref = $packages[$pkgno]->misc_URLs;
      if (int (@{$misc_URLs_ref})) {
         my $tbl = new IW::Widget::URLTable
            "$title" => $misc_URLs_ref;
         $tbl->print_short;
      } else {
         $HTML -> t("$title");
      }
      $HTML -> BR;
   }
   1;
}

#------------------------------------------------------------
# any_titles  ARRAYREF
#------------------------------------------------------------
# Input: reference to array of packages.
# Output: Returns true if any package has a title or if any package
# has flag LINE_ITEMS_PER_GRANULE. Else false.

sub any_titles {
   my $pkgref = shift;
   my @packages = @$pkgref;
   my $package;
   foreach $package (@packages) {
      return 1 if ($package->title or $package->line_items_per_item);
   }
   return 0;
}

#------------------------------
#
# SUBMODE HANDLERS...
#
#------------------------------

#------------------------------------------------------------
# submode_DEFAULT
#------------------------------------------------------------
sub submode_DEFAULT {
    shift->submode_SELECT;
}

#------------------------------------------------------------
# submode_SELECT
#------------------------------------------------------------
sub submode_SELECT {
    my $self = shift;
    my $gran = IU::Dataset->new;
    my $dset = IU::Dataset->new;

    $origin = $REQ->param('origin') || 'PREFS';


    # Get package info:
    my $grankey = '2702/3465/*';
    my @packageIds;
    $packageIds[0] = '*ABC';
    $packageIds[1] = '*DEF';
    my @packages = IU::Package->lookup($grankey,
                                       Dataset => $dset,
                                       Granule => $gran,
                                       PackageIds => \@packageIds);

    @packages or puke 560, "Data Granule claims to have ordering ",
       "options, but none were supplied.";
    # Set the navigation indicator.
    $SESSION->{Navigator}->{Indicator} = "OrderData";

    # If there are TITLE's in the packages, print them as a
    # selection index instead of printing the packages on this
    # page.  However, if there is only one package, print it here.
    # If there are no titles, print the packages here.

    # Coding trick: if one package, $#packages is 0 so the following
    # "&&" never gets evaluated and $anytitles is set to 0.
    # If there are multiple packages, then clause after "&&" gets
    # evaluated and then $anytitles is set to its return value.
    my $anytitles = $#packages && any_titles(\@packages);

    # PRINT

    # Output the page header (with HTTP header):
    $APP->start_page(Title      =>'Customize Ordering Options',
                     Context    =>'customorder');

    # Begin the form
    my $scroll_to = '#CUSTOM';
    $HTML -> script_FORM(ACTION => script_action().$scroll_to);

    $HTML -> HR;
    $HTML -> TABLE(BORDER=>"0", CELLPADDING=>"4", CELLSPACING=>"2");
    $HTML -> TR;
    $HTML -> TD(WIDTH=>"4%");
    $HTML -> IMG(SRC=>IW::Util::doc_url('icons/note-info.gif'),
                 ALT=>'INFO',
                 ALIGN=>'LEFT');
    $HTML -> _TD;
    $HTML -> TD(BGCOLOR=>"#FFFFCC", WIDTH=>"96%", ALIGN=>"LEFT");
    $HTML -> t("The order options below can be selected and saved for ");
    $HTML -> t("automatic use in your product ordering.  The media type ");
    $HTML -> t("chosen will be applied to the granules selected for ");
    $HTML -> t("ordering, unless the media type is not applicable for ");
    $HTML -> t("the granule(s) selected.");
    $HTML -> _TD;
    $HTML -> _TR;
    $HTML -> _TABLE;
    $HTML -> HR;
    # Print message:
    $HTML -> t( "There are ");
    my $numberOfPackages = int(@packages);
    $HTML -> t_B("$numberOfPackages");
    $HTML -> t(" different Ordering Options available.");
    $HTML -> BR;

    # Submode for form depends on whether we have the title list
    # or the packages on this page.
    $HTML -> hidden('origin'  => $origin);
    if ($anytitles) {
       # If titles, print index.
       print_index($gran, \@packages, $dset);
    } elsif ($#packages eq 0) {
       # Just one package so print it.
       print_form($gran,\@packages, $dset, '1'); 

    } else {
       # Else print packages. Print form.
       print_form($gran, \@packages, $dset);
    }
    $HTML -> _script_FORM;

    # The signature:
    $APP->end_page();
    1;
}

#------------------------------------------------------------
# submode_CHOSEN
#------------------------------------------------------------
sub submode_CHOSEN {
   my $self = shift;   
   my $gran = IU::Dataset->new;
   my $dset = IU::Dataset->new;
   $origin = $REQ->param('origin') || 'PREFS';

   my $number;

   # Get package info:
   my $grankey = '2702/3465/*';
   my @packageIds;
   $packageIds[0] = '*ABC';
   $packageIds[1] = '*DEF';
   my @packages = IU::Package->lookup($grankey,
                                      Dataset => $dset,
                                      Granule => $gran,
                                      PackageIds => \@packageIds);

   @packages or puke 560, "Data Granule claims to have ordering ",
      "options, but none were supplied.";
   my $pkno;         # used in iteration

   # Build a list of all combos chosen across packages:
   # E.g., ("1.1", "1.2", "2.2")
   my @all_combo_ids_chosen;
   foreach $pkno (1..int(@packages)) {
      push @all_combo_ids_chosen, $REQ->param("pk-choice.$pkno");
   }
   my $user = $SESSION->state->{User} || new IW::User;
   # Get the list of all combos the user chose, as a lookup table.
   # The keys are of the form "1.1", "1.2", "2.1", etc...
   my %combo_ids_chosen = map {$_=>1} @all_combo_ids_chosen;
   my @chosen_combos;
   for $pkno (1..int(@packages)) {       # for each package, 1..p
      my $package = $packages[$pkno-1];     # get 0-based package
      my @combos = $package->combos;        # get combos in packages
      my $combono;
   
      # Was any combo from this package chosen?  If not, skip it now:
      (grep { /^\d+/ && ($& eq $pkno) } @all_combo_ids_chosen) or next;

      # If there were subsetting/order options, and they are wanted,
      # get 'em:
      my $opts_type;
      my %opts_assigned;
      foreach $opts_type (sort $package->options) {
         my $opts = $package->options($opts_type);

         # Are they required?
         my $opts_required = $opts->is_required;

         # Were they given?  Fill them in, then check for no
         # selections:
         $opts->read_CGI(Prefix=>"pk-opts.$pkno.$opts_type.");
         my $opts_given = $opts->user_made_selections;

         # Zero it if necessary:
         $opts_assigned{$opts_type} = $opts     if $opts_given;
      }
      # ...at this point, %opts_assigned contains only those options
      # which the user gave input for, hashed by option type.

      # For each combo in this package...
      for $combono (1..int(@combos)) {

         # Did the user select this combo?  If not, skip it:
         $combo_ids_chosen{"$pkno.$combono"} or next;

         # Clear the user object of the FTP items
         foreach $sub_key (@user_subs_keys) {
            my $newsub = $user_subs{$sub_key};
            $user->$newsub(undef);
         }
         # They wanted this combo (and hence, the package it's in);
         #    Associate any subset/order options with this combo,
         #    and push it on the list of combos:
         my $combo = $combos[$combono-1];      # get 0-based combo
      my ($opts_type, $opts);
      while (($opts_type, $opts) = each %opts_assigned) {
         my @optskeys = values %$opts;
         my @newarray = $optskeys[2];

         my $newitem = $newarray[0];
         my @newitems = @$newitem;
         my $newitem2;
         my $critname;
         foreach $newitem2 (@newitems) {
            my @newkeys = keys %$newitem2;
            my @newvalues = values %$newitem2;
            my $i = 0;
            # Need to loop through and find the criteria name first
            # because it can occur after the UserSlct item
            foreach (@newkeys) {
               if($newkeys[$i] eq 'CRITERIA_NAME') {
                  $critname = $newvalues[$i];
               }
               $i++;
            }
            $i = 0;
            foreach (@newkeys) {
               if($newkeys[$i] eq 'UserSlct') {
                  my $hashref = $newvalues[$i];
                  my @keyshash = keys %$hashref;
                  my @valueshash = values %$hashref;
                  foreach $sub_key (@user_subs_keys) {
                     if($sub_key eq $critname) {
                        my $newsub = $user_subs{$sub_key};
                        $user->$newsub($valueshash[0]);
                     }
                  }
               }
               $i++;
            }
         }
         }
         my $proc_opt = $combo->proc_opt;
         my $media_type = $combo->media_type;
         my $media_fmt = $combo->media_fmt;
	 if(defined($proc_opt) && defined($media_type) && defined($media_fmt)) {
	    $user->data_format($proc_opt);
	    $user->media_type($media_type);
	    $user->media_format($media_fmt);
	    $user->install;
	    switch IW::Mode Tags=>{'mode' => $origin};
	    return 1;
	 }
      }
   }
   switch IW::Mode Tags=>{'mode' => $origin};
   return 1;
}
#------------------------------------------------------------
# go
#------------------------------------------------------------
# Main entry point.

sub go {
   my $self = shift;   
   $APP->get_session();     # get session or die
   my $submode = ($REQ->param('submode') || 'DEFAULT');
   my $method = "submode_\U$submode\E";
   $self->$method() or puke "failed while handling submode $submode!";
}

#------------------------------------------------------------
1;
