package IW::ModeSEARCH;

=head1 NAME

IW::ModeSEARCH - do the search

=head1 NOTES

The normal scheme of dealing with a search is...

     1. Post the query to SEARCH/START, which starts the search and
        issues a redirect to a GET request for SEARCH/TRACE mode.
     2. Monitor it in SEARCH/TRACE mode.
     3. Stop it via SEARCH/STOP mode.

This is partly to work around a Netscape 2.x bug, where POSTs requests
returning multipart/x-mixed-replace documents I<fail> to replace the
page in the history list!

Interestingly enough, this workaround solves a thorny problem: namely,
how to prevent someone from accidentally reloading a document that will
kick off a search?  The answer is I<we no longer have to>... since the 
document returned by SEARCH/START is a redirect request, such a
document never exists on the user's history list!  Thanks, Netscape!

=cut

use File::Copy;
use IU;
use IU::Log;
use IU::Errors;
use IU::Event;
use IU::Query;
use IU::Category;
use IW::App;
use IW::Util qw(:url :output);
use IW::ModeGUIDE;
use IW::Client;
use IW::CommStat;
use IW::Comm::Manager;
use IW::User;

@ISA = qw(IW::Mode IW::Comm::Manager);

use strict;

#------------------------------------------------------------
#
# Globals...
#
#------------------------------------------------------------

# Map a search type to a log event name:
my %ActivityToEvent = (
   DIRECTORY_SEARCH     => 'DIRECTORY',
   GRANULE_SEARCH       => 'DATAGRANULE',   # it's an inventory search
   LOCAL_GRANULE_SEARCH => 'LOCALGRANULE',
   GUIDE_SEARCH         => 'GUIDE',
   INVENTORY_SEARCH     => 'INVENTORY',
);

#------------------------------------------------------------
# start_guide_search QUERY
#------------------------------------------------------------
# I<Instance method.>
# Guide search is not done through the imsclient, but through waisq,
# so we handle it here as a special case.

sub start_guide_search {
   my ($self, $query) = @_;

   # Log the search:
   ims_event($ActivityToEvent{'GUIDE_SEARCH'},
             ARCHID   =>[$query->daacs_to_search]);

   # Build a hash of CGI parameters for ModeGUIDE.
   # Invoke the mode which actually does the guide search:
   IW::Mode->switch(Tags => {mode    => 'GUIDE',
                             submode => 'GUIDESRCH',
                             sid     => ($SESSION->id),
                             query   => $query} );
   # Done!
   return 1;
}

#------------------------------------------------------------
# start_search QUERY
#------------------------------------------------------------
# I<Instance method.>
# Start the search.

sub start_search {
   my ($self, $query) = @_;

   #------------------------------
   # SETUP FOR THE SEARCH:
   #------------------------------

   # Clear this variable out so we can automatically display the Granule
   # list IF a single Dataset is returned.
   $SESSION->state->{ShownSingleDataset} = undef ;

   # Figure out the kind of search we're doing:
   my $srchtype = $query->search_type || 
      puke 600, "Ack! No idea what type of search to run!";   
    
   # Build the "query.dat" file, for downloading the query...
   #   This ODL should be created FOR THE USER (e.g., DON'T put in a 
   #   default stop date if no start date was given).  It should only
   #   contain attributes relevant to the current search type.

   my $queryfile = $SESSION->path_to('pub/query.dat');
   open QUERY, ">$queryfile" or puke "open $queryfile: $!";
   $query->to_ODL->output(\*QUERY);
   IU::ODL->end(\*QUERY);
   close QUERY;
#puke;
   # If the search type is GUIDE, we handle it separately:
   if ($srchtype eq 'GUIDE_SEARCH') {
      return $self->start_guide_search($query) ? 'DONE' : undef;
   }
   my $user;    # User preferences object.

   # Handle any inputs the user may have entered on the warning form.
   my $warned = $REQ->param('warned') || 0;
   if ($warned eq '1') {
      # Get ECS keys from warning page.  Updates current user object
      # and installs in session.
      $user = IW::User->from_CGI ('FORM' => 'WARNING');

      # If errors, set APP message and set to re-send warning page.
      if ($user->errors->any_errors) {
         $APP->message (new IW::Note Class=>'ERROR',
                        HTML => join ("<BR>\n", $user->errors->get));
         $warned = '0';
      }
   } else {  # Else not warned. Get the current user object or create.

      # Get existing user prefs.
      $user = $SESSION->state->{User};

      # If none, then force user info to exist and install it.
      $user or ($user = new IW::User)->install;
   }
   # Build the "search.odl" file, for input to the imsclient...
   #   This ODL should be created FOR THE SEARCH (e.g., put in a
   #   default stop date if no start date was given).  It should only
   #   only contain attributes relevant to the current search type.

   my $infile = $SESSION->path_to ('search/search.odl');
   open SEARCH, ">$infile" or puke "open $infile: $!";
   $query->to_search_ODL->output (\*SEARCH);
   IU::ODL->end (\*SEARCH);
   close SEARCH;

   # Need to pop up warning window if a previous search is running.
   # Parameter "killing" is set only if user got a warning pop-up
   # and just hit the button to kill the previous search.  Otherwise
   # it is undefined.  If a search client is currently running, and
   # the user hasn't hit the kill previous search button in the pop-up
   # then we call warn_killing_client which makes the pop-up.

   my $last_client = $SESSION->state->{LastsearchClient};
   if (defined ($last_client) && $last_client->running &&
       ($REQ->param ('killing') ne '1')) {
      $self->warn_killing_client();
      return 'DONE';
   }
   # Authenticator isn't needed or the user has already been warned.
   # Build the "query.html" file (for summarizing the search)...
   # This file gets spliced into the bottom of "result" pages.  It
   # should only contain attributes relevant to the current search
   # type.
   my $summary_file = $SESSION->path_to ('search/query.html');
   open SUMMARY, ">$summary_file" or puke "open $summary_file: $!";
   my $SUMMARY = IW::HTML->new (\*SUMMARY) or puke "no HTML stream!";
   $query->print_HTML ($SUMMARY);
   close SUMMARY;

   #------------------------------
   # SEARCH!
   #------------------------------

   # Log the start of it:
   if ($srchtype eq 'GRANULE_SEARCH' ) {
      ims_event($ActivityToEvent{$srchtype}, 
                ARCHID    => [$query->daacs_to_search],
                DATASET   => [$query->datasets], 
                GRANULEID => [$query->granule_ids]);
   }elsif($srchtype eq 'LOCAL_GRANULE_SEARCH'){
      ims_event($ActivityToEvent{$srchtype}, 
                DATASET   => [$query->datasets], 
                LOCALGRANULEID => [$query->get_local_granule_id]);

   }else {
      my @query = ();
      my $cat;
      foreach $cat ($query->categories) {      # for each category in query
         $query->uses($cat) or next;                 # only if using it!

         # Get the valids, as appropriate:
         if (IU::Category->is_core($cat)) {         # core dependent valids
            push (@query, map {$cat => $_} $query->criteria_strings($cat) );
         }
         elsif (IU::Category->is_extended($cat)) {  # extended dependent valids
            push (@query, map {$cat => $_->variant.':'.$_->as_string} $query->criteria($cat) );
         }
      }

      my $geo = $query->geo_region;
      push (@query, 'GEO' => $query->geo_region->as_string) if $geo;

      ims_event($ActivityToEvent{$srchtype}, 
                DATES_ARE => $SESSION->state->{DateType},
                TIME      => [$query->time_range_TZ],
                ARCHID    => [$query->daacs_to_search], @query);

   }
   # Start the client:
   my $client = IW::Client->start(Activity => $query->ODL_search_type,
                                  Timeout  => $query->max_time(),
                                  Infile   => $infile);
   $client or puke "couldn't start imsclient";

   # Now that the client has been started, save the last 
   # inventory/directory search type in the session state.
   # The only legal values are INVENTORY_SEARCH and DIRECTORY_SEARCH.
   #
   # THIS EXACT LOCATION IS THE ONLY, REPEAT, *ONLY* PLACE THAT THIS 
   # SHOULD BE DONE.  By WAITING UNTIL THE CLIENT STARTS (remember,
   # the client will clear out all previous search results as its
   # first order of business), it guarantees that the session state is
   # always in synch with any current set of search results -- or, if
   # there are none, with the last attempt at GETTING search results.
   #
   # NOTE: technically, we could (almost) get this information by
   # parsing the ODL of the "query.dat" file.  However:
   #
   #   1. That's an expensive operation, and the "last search" info
   #      is used in a number of places when showing results.
   #
   #   2. It is theoretically possible for a search request to create 
   #      the "query.dat" file, but fail to start the client; in such
   #      a case, any results lying around from a previous search
   #      would not have been deleted, and their type might not agree
   #      with the type in the new (but unused) "query.dat"!  
   #
   # So, we continue to use the session state.
   #
   $SESSION->state->{LastSearchType} = $query->search_type;

   # Done!
   return 'STARTED';
}

#------------------------------
#
# Comm manager methods...
#
#------------------------------

#------------------------------------------------------------
# comm_preflight
#------------------------------------------------------------
# Kill persistent counts in session state, then go.
sub comm_preflight {
   my $self = shift;
   $SESSION->state->{OldGranuleCount} = 0;
   $SESSION->state->{OldDatasetCount} = 0;
   $self->SUPER::comm_preflight(@_);
}

#------------------------------------------------------------
# comm_start_search
#------------------------------------------------------------
# Start a search.
# Must return the client.
#
sub comm_start_search {
   my $self = shift;

   # Get current query, and validate:
   my $query = IU::Query->lookup || puke "ack! no query!";

   $query->validate;

   my $cat;
   my @variants;
   my $total;

   my @xs_cats = grep {IU::Category->is_extended($_)} $query->categories;
       foreach $cat (@xs_cats) {

           @variants = map {$_->variant} $query->criteria($cat);
           $total = $total + int(@variants) if(@variants);
       }
   puke 517, "Your total number of"
    . " attributes chosen now exceeds the system limit of ${IU::PSAMax}."
    . " Please use the back button of your browser to return to the"
    . " previous page and reduce the number of"
    . " attributes chosen."
    if($total > ${IU::PSAMax});

   # Because of the new navigation and partial results, granule, 
   # inventory, and directory searches run in a subdirectory.
   # If a subdirectory for searching doesn't exist, create it.
   my $searchdir = $SESSION->path_to('search');
   unless (-d $searchdir) {
      mkdir ($searchdir, 0755) or puke "mkdir $searchdir: $!";

      # HTACCESS FILE FOR THE SEARCH DIRECTORY (DENIES EVERYONE)...
      # Build a virtual .htaccess file for this session directory:
      require IU::HTAccess;
      my $access = new IU::HTAccess;
      $access->start_limit('GET', 'POST', 'PUT');
      $access->order('ALLOW', 'DENY');
      $access->deny_from('all');
      $access->end_limit;
      # Write it out:
      $access->write_in($searchdir) 
         or warn "Couldn't write .htaccess in search subdirectory.";
   }
   # If there is a dynamic valids ODL file, make sure a link to it
   # is in the search subdirectory.  It needs to be in both places
   # because IK needs it in search directory but composing search
   # needs it in session directory.
   my $dfile = $SESSION->path_to('dynamic.odl');   # file
   my $lfile = $searchdir . '/dynamic.odl';        # link
   if (-e $dfile &&                       # If file exists.
       (! -e $lfile)) {                   # If link doesn't exist...
      # Try to create the symbolic link.
      unless (eval { symlink ($dfile, $lfile); }, $@ eq "") {
         debug "Failed to create symbolic link for dynamic.odl";
         # Try copy instead;
         copy ($dfile, $lfile); 
      }
   }
   # Go!
   return $self->start_search($query);
}

#------------------------------------------------------------
# comm_ok_page
#------------------------------------------------------------
# Everything went fine!  Output the page:

sub comm_ok_page {
   my ($self, %params) = @_;
   # Linux needs for the Switch to be replaced w/ a Redirect.
   $REPLY->redirect(script_url('mode'=>'DSETLIST'));
} 

#------------------------------------------------------------
# comm_fail_page
#------------------------------------------------------------
# Our search failed.  Output the final comm status page.

sub comm_fail_page {
   shift->comm_final_status_page(@_);
} 

#------------------------------------------------------------
# comm_stop_gracefully
#------------------------------------------------------------
# Stop gracefully?  Yes; we want those results!

sub comm_stop_gracefully {
   1;
}

#------------------------------
#
# Submodes...
#
#------------------------------
#------------------------------------------------------------
# submode_DEFAULT
#------------------------------------------------------------
# Default submode.

sub submode_DEFAULT {
   shift->submode_COMMSTART(@_);
}

#------------------------------------------------------------
# submode_STATUS
#------------------------------------------------------------
# Figure out the proper commstat page to show, and go there.

sub submode_STATUS {
  my $self = shift;

  # Get the current client, if any.
  my $client = attach IW::Client 'search';

  # The redirection is apparently needed to make the "rollover" happen.
  if ($client) {
    # If the client exists and is running, show running status.
    if ($client->running) {
     $REPLY->redirect(script_url('mode'=>'SEARCH','submode'=>'COMMTRACE'));
    }
    else {
      # Not running; show final status.
     $REPLY->redirect(script_url('mode'=>'SEARCH','submode'=>'COMMFINAL'));
    }
  }
  else {
    # Never run. Do a search, please.
    $self->submode_COMMINSTRUCTION;
  }
}

#------------------------------------------------------------
# warn_killing_client
#------------------------------------------------------------
# Warn user that to a start a new search, the previous running search
# will be killed.

sub warn_killing_client {
   unless ($APP->message) {
      $APP->message ("You are ready to submit a new search, but an".
                     " earlier search is still running. Submitting ".
                     "a new search now will override your previous search.".
                     " To submit this new search press ".
                     "the 'Start New Search' button, otherwise use ".
                     "the 'Cancel' button to go back to the Search ".
                     "Form.  If you want to see the status of your ".
                     "current search, select Search Status from the ".
                     "Navigation bar.");
   }
   # Start page:
   $APP->start_page(SuperTitle=>'Start Search',
                    Title=>'Override Current Search ?',
                    Context=>'abort');

   $HTML -> script_FORM;

   # There's a multi-submit button here...
   $HTML -> msubmit_flag;

   # Start New Search/Cancel buttons.
   start_newsearch_button ();
   cancel_newsearch_button ();

   $HTML -> _script_FORM;

   # Done:
   $APP->end_page;
   1;
}

#--------------------------------------------------------------------
# start_newsearch_button 
#--------------------------------------------------------------------
# Print Start New Search button.
# This subroutine creates the Start New Search button for the aborting 
# previous search warning form.

sub start_newsearch_button {

   $HTML -> msubmit_flag;
   $HTML -> CENTER -> FONT(SIZE=>'+4') -> B
         -> msubmit_button(CAPTION => 'Start New Search',
                           QUERY =>['mode'    => 'SEARCH',
                                    'submode' => 'START',
                                    'warned' => '1',
                                    'killing' => '1']);
   $HTML -> _CENTER -> _FONT -> _B ;
   1;
}

#--------------------------------------------------------------------
# cancel_newsearch_button
#--------------------------------------------------------------------
# Print Cancel new search button.
# This subroutines creates the cancel button for the aborting previous
# search warning form.

sub cancel_newsearch_button {

   $HTML -> msubmit_flag;
   $HTML -> CENTER -> FONT(SIZE=>'+4') -> B
         -> msubmit_button(CAPTION => 'Cancel',
                           QUERY =>['mode'    => 'MAINSRCH',
                                    'warned' => '1',
                                    'killing' => '1']);
   $HTML -> _CENTER -> _FONT -> _B ;
   1;
}

#------------------------------------------------------------
# submode_START
#------------------------------------------------------------
# Backwards-compatibility.

sub submode_START {
   shift->submode_DEFAULT(@_);
#   shift->submode_COMMSTART(@_);
}

#------------------------------------------------------------
# submode_TEST
#------------------------------------------------------------
sub submode_TEST {
   my $self = shift;

   # Build test search file:
   my $infile = $SESSION->path_to('search/search.odl');
   open SEARCH, ">$infile" or puke "open $infile: $!";
   print SEARCH <<EOF;

GROUP = DIRECTORY_SEARCH
    MESSAGE_ID = "0"
    DATA_CENTER_ID = ("STUB", "EDC", "GSFC")
END_GROUP = DIRECTORY_SEARCH

END
EOF
   close SEARCH;
    
   # Input an ODL tree
   open IN, "$infile";
   my $odl = parse IU::ODL \*IN;
   close IN;
   $odl or die "couldn't parse file";

   # Create and install query object:
   my $query = IU::Query->from_ODL($odl);
   $query->install;
    
   # Do it:
   $self->submode_COMMSTART;
}

#------------------------------------------------------------
# go
#------------------------------------------------------------
# Main entry point for the mode.

sub go {
   my $self = shift;    
   $APP->get_session;
   $self->go_submode;
}

#------------------------------------------------------------
1;
