package IW::ModeSLCT;

=head1 NAME

IW::ModeSLCT - abstract search form "popup"

=head1 DESCRIPTION

There is a common "look and feel" to the "popups" which are used
to provide the query criteria.  Therefore, all the concrete mode classes 
which support the search form should be named IW::ModeSLCT*** and 
inherit from this abstract class (formerly IW::Popup).

=cut

use IU;
use IU::Log;
use IU::Query;
use IU::Category;

use IW::App;
use IW::Util qw(:url);
use IW::DepValid;
use IW::Widget::QueryEdit;    
use IW::Widget::DepValids;    

@ISA = qw(IW::Mode);

use vars qw($User);
use strict;


#------------------------------------------------------------
# new PARAMHASH
#------------------------------------------------------------
# I<Constructor.>
# PARAMHASH may contain:
#
#     Category => The actual category we will be selecting for (some
#                 modes, like SLCTDEPVAL, support more than one category).
#                 If CATEGORY is not given, it is taken from the CGI 
#                 environment tag 'category'.
#     Context  => The help context for this page.
#
sub new {
    my ($class, %params) = @_;
    my $self = \%params;
    $self->{Category} ||= $REQ->param('category');

    # Set the help context if it is not set.
    $self->{Context} = category2context($self->{Category})
           unless $self->{Context};
    bless $self, $class;
}

#------------------------------------------------------------
# category
#------------------------------------------------------------
# I<Instance method.>
# Return the category we are selecting for.
#
sub category { 
    my $self = shift;
    $self->{Category} = $_[0] if @_;
    $self->{Category} 
}

#------------------------------------------------------------
# context
#------------------------------------------------------------
# I<Instance method.>
# Return our help context.
#
sub context { 
    my $self = shift;
    $self->{Context} = $_[0] if @_;
    $self->{Context} 
}

#------------------------------------------------------------
# thingy
# thingys
# title
#------------------------------------------------------------
# I<Instance methods.>
# A nice name for our category, both in singular and plural form.  
# Taken directly from the category name.

sub thingy  { 
    category2pretty(shift->category);
}
sub thingys {
    my $self = shift;
    my $plural = $self->title . 's';
    $plural =~ s/ys$/ies/;
    $plural;
}
sub title { 
    shift->thingy;
}

#------------------------------------------------------------
# start_page PARAMHASH
#------------------------------------------------------------
# I<Instance method.>
# Start the page.  Identical to $APP->start_page, except that the
# title and supertitle have been automatically supplied (although
# you can override them if you wish).

sub start_page {
    my $self = shift;

    # Get pretty name for search type:
    my $pretty_type = $QUERY->search_type;
    $pretty_type =~ tr/A-Z_/a-z /; 
    $pretty_type =~ s{\b([a-z])}{\u$1}g;
    
    # Start page:
    $APP->suppress_key(1);    
    $APP->start_page(SuperTitle => "Search Form: " 
                                . valid2pretty($QUERY->search_type),
		     Title      => ("Select: " . $self->title),
		     Context    => $self->context,
		     @_);        # allow overriding
}

#------------------------------------------------------------
# end_page
#------------------------------------------------------------
# I<Instance method.>
# Start the page.  Identical to $APP->end_page.

sub end_page {
    shift;
    $APP->end_page(@_);
}

#------------------------------------------------------------
# print_form
#------------------------------------------------------------
# I<Instance method, abstract.>
# Print the form.
# You can override this in your subclasses.

sub print_form {
    $HTML -> t("Override print_form() to alter the form.");
    undef;
}

#------------------------------------------------------------
# send_page
#------------------------------------------------------------
# I<Instance method, abstract.>
# Output a basic page, invoking print_form() to output the form.  
# You can override this in your subclasses.

sub send_page {
    my ($self, %target) = @_;

    # The grey navigator:
    my $edit_w = IW::Widget::QueryEdit->new(Popup=>$self->category);

    # Have the widget set its return point. Does nothing if no ReturnTo
    # was passed in.
    $edit_w->set_return(%target);

    # Start page:
    $self->start_page;

    # Start form:
    #    Note that submitting the form on a "popup" always tries to get the
    #    browser to do a "submit and scroll" to the #SCROLL anchor.  
    #    By default, that anchor is placed at the top of each page unless
    #    you tell the $APP object not to do so (which the main search form
    #    can do, "preempting" the default SCROLL point to place it lower).
    #
    #    If you want to disable submit-and-scroll for popups, DO IT HERE
    #    by removing setting $scroll_to to empty.    
    #
    my $scroll_to = '#SCROLL';
    $HTML -> HR;
    $HTML -> script_FORM(ACTION => script_action().$scroll_to);

    # Embed standard hidden info for forms on popups:
    $HTML -> hidden('category' => $self->category);
    $HTML -> hidden('mode'     => $edit_w->mode_for($self->category));
    $HTML -> hidden('sc'       => 1);       # "Scroll to Category"
    $HTML -> msubmit_flag;

    
    # Main form:
    # If someone pressed the Reset/Clear button in the Extended Search page
    # pass the action "Reset" or "Clear" value .
    $self->print_form($target{Action});
 
    # Grey navigator: Don't print jump menu if either type of granule id search
    if (!(($QUERY->search_type eq 'GRANULE_SEARCH') || 
           ($QUERY->search_type eq 'LOCAL_GRANULE_SEARCH'))) {
        if ($self->jumpmenu_is_okay()) {
            $HTML -> HR;
            $edit_w->print;
        }
    }

    # End form and page:
    $HTML -> _script_FORM;
    $self->end_page;    

}

#------------------------------------------------------------
# jumpmenu_is_okay
#------------------------------------------------------------
# I<Instance method, abstract.>
# Return true if the jumpmenu should be shown, undef if not.
# Default method assumes jumpmenu should be shown.
# You can override this in your subclasses.
sub jumpmenu_is_okay {1};
   
#------------------------------
#
# Setting query values...
#
#------------------------------

#------------------------------------------------------------
# ok
#------------------------------------------------------------
# I<Instance method, abstract.>
# User hit an "OK" button.  Use CGI input to set valids in query object.
# You must override this in your subclasses.
#
sub ok {
    puke "abstract method ok() not implemented";
}




#------------------------------
#
# Submode handlers...
#
#------------------------------

#------------------------------------------------------------
# submode_DEFAULT
#------------------------------------------------------------
# Entry point.

sub submode_DEFAULT {
    shift->submode_SHOW;
}

#------------------------------------------------------------
# submode_CANCEL
#------------------------------------------------------------
# I<Instance method>.
# Handle a "CANCEL-and-go-to" request submitted to this mode.

sub submode_CANCEL {
    go IW::Widget::QueryEdit;
}

#------------------------------------------------------------
# submode_DELETE
#------------------------------------------------------------
# I<Instance method>.
# Cause the deletion of the current category, followed by a goto.
#
# NOTE: at the time of this writing, mode SRCHFORM also has a DELETECAT
# mode which functions in a similar fashion.  They're small and simple
# enough to coexist peaceably.

sub submode_DELETE {
    my $category = $REQ->param('category');
    if ($QUERY and $category) {
	$QUERY->criteria($category, undef);   
	$APP->message("All criteria for " . category2pretty($category) . 
		      " have been deleted");
    }
    go IW::Widget::QueryEdit;
}

#------------------------------------------------------------
# submode_GOTO
#------------------------------------------------------------
# I<Instance method>.
# For ModeSLCT subclasses, this is just a synonym for submode CANCEL (q.v.)
# since the semantics are "forget what I was just doing, and go elsewhere".

sub submode_GOTO {
    shift->submode_CANCEL(@_);
}

#------------------------------------------------------------
# submode_OK
#------------------------------------------------------------
# The "OK" button next to the valids was just pressed.
#
#    1. Get the user's selection
#    2. Set the selections in the current query object.
#    3. Go to the desired "goto" location

sub submode_OK {
    my $self = shift;

    $self->ok;
    go IW::Widget::QueryEdit;
}

#------------------------------------------------------------
# submode_SHOW
#------------------------------------------------------------
# Entry point.

sub submode_SHOW {
    my $self = shift;
    $self->send_page;
}

#------------------------------------------------------------
# go
#------------------------------------------------------------
# Main entry point for mode.

sub go {
    my $self = shift;   
    $self->init;
    $self->go_submode;
}

#------------------------------------------------------------
# goto
#------------------------------------------------------------
# Standard main entry point for transferring control to this mode 
# from another mode.

sub goto {
    my $self = shift;   
    $self->init;
    $self->submode_DEFAULT;
}

#------------------------------------------------------------
# init
#------------------------------------------------------------
# Init this mode.

sub init {
    $APP->get_session;    
    $QUERY ||= IU::Query->lookup;
    1;
}

#------------------------------------------------------------
1;
