package IW::ModeSLCTALL;

=head1 NAME

IW::ModeSLCTALL - mode to manage selection of extended search criteria

=cut

use IU;
use IU::Log;
use IU::Coord;
use IU::Query;
use IU::Session;
use IU::Category;

use IW::App;
use IW::Mode;
use IW::ModeSLCT;
use IW::Util qw(:string);
use IW::Widget::DepValids;

use strict;
use vars qw(@ISA);

@ISA = qw(IW::ModeSLCT);



#------------------------------------------------------------
# new PARAMHASH
#------------------------------------------------------------
sub new {
    my $self = shift->SUPER::new(Context=>'othercritsel',
                                 @_);      # inherited constructor
    $self;
}

#------------------------------------------------------------
# title
#------------------------------------------------------------
sub title {
    my $self = shift;
    return (($self->category eq 'EXTENDED_SEARCH') ?
	    'Other Criteria' :
            $self->category);
}


#------------------------------
#
# Output...
#
#------------------------------

#------------------------------------------------------------
# set_CGI
#------------------------------------------------------------
# I<Instance method.>
# Use existing query to lie about our input for the category's criteria.

sub set_CGI {
    my $self = shift;
    
    # Get existing criteria, if any:
    debug "set_CGI: getting criteria for ", $self->category;
    my @criteria = $QUERY->criteria($self->category);
    debug "set_CGI: got ", scalar(@criteria), " entries.";
    @criteria or return;

    # Set CGI tags:
    my $crit;
    foreach $crit (@criteria) {
	debug "set_CGI: setting for ", $crit->name;
	$crit->set_CGI(Prefix=>'p.'.$crit->name.'.');   # must use name!
    }
}

#------------------------------------------------------------
# start_page PARAMHASH
#------------------------------------------------------------
# I<Instance method>
# Override the default ModeSLCT start_page to get a whack at
# setting the context right first.
#

sub start_page {

   my $self = $_[0];

   # Determine the context at this point. If the category is 
   # still extended search, then we're on the first sub page,
   # where we select the criterion to update. Otherwise,
   # we're on the criterion entry page and we set the context
   # accordingly.

   debug "We get our shot at the context and category:\n",
         "Context: ", $self->context(), " Category: ", $self->category();

   $self->context('otherentry') if $self->category() ne 'EXTENDED_SEARCH';   

   # Now just do what we used to do. (I love OOP.)

   $self->SUPER::start_page(@_);

}
#------------------------------------------------------------
# print_form
#------------------------------------------------------------
# I<Instance method, override.>
# Output the main body of the page.
#
sub print_form {
    my $self = shift;
    # Value can either be Reset/Clear, used by the Reset & Clear buttons.
    my $action_to_do = shift;

    # Has a category been chosen?
    if ($self->category ne 'EXTENDED_SEARCH') {

	# Output menu for category:
        debug "trying to force context switch";
        $self->context('otherentry');
	$self->print_form_category($self->category, $action_to_do);
    }
    else {
	# Get compatible categories... some may have leading "*":
	my @categories = $QUERY->depvalid_interface->get_categories;

	# Reduce to just extended criteria:
	@categories = grep {!IU::Category->is_core(clean_valid($_))
			    } @categories;
	
	# Anything?
	if (@categories) {   $self->print_form_categories(\@categories) }
	else             {   $self->print_form_nothing }
    }
}

#------------------------------------------------------------
# print_form_categories CATEGORIES
#------------------------------------------------------------
# I<Instance method, override.>
# Output the main body of the page.
#
sub print_form_categories {
    my ($self, $cats) = @_;

    # Directions:
    $HTML -> directions;
    $HTML -> t("The following are the additional search criteria which ",
	       "are available, based on your current query. ")
	  -> t_B("Please pick the one you want, and hit EDIT.");
    $HTML -> _directions;
    
    # Start table:
    $HTML -> CENTER -> TABLE -> TR;
    $HTML -> TD(VALIGN=>'MIDDLE');
    $HTML -> msubmit_button(QUERY => ['submode'=>'CRITFORM'], 
                            CAPTION => "EDIT criteria for...");

    # Choices...
    $HTML -> TD(VALIGN=>'MIDDLE');
    $HTML -> SELECT(NAME=>"p.category", SIZE=>5);
    my $cat;
    foreach $cat (sort @$cats) { $HTML -> tag('OPTION') -> t($cat) -> nl }
    $HTML -> _SELECT;

    $HTML -> _TD -> _TR;
    # End table:
    $HTML -> _TABLE -> _CENTER;
    $HTML -> HR;
    $HTML -> CENTER;
    $HTML -> FONT(SIZE=>'+2');
    # Cancel button
    $HTML -> msubmit_button(CAPTION=>'Return',
                             QUERY=>['submode' => 'GOTO']);
    $HTML -> _FONT;
    $HTML -> _CENTER;
}

#------------------------------------------------------------
# get_valid_variants
#------------------------------------------------------------
# Get valid variants, as a lookup table keyed on the variant name:

sub get_valid_variants {
    my $self = shift;

    # Get the variants (forcing upcase, just in case):
    my @all =  map{uc($_)
		  } $QUERY->valids_for($self->category);  ### runs subprocess
    IU::Log::debug("Variants for ",$self->category," = ",join("|", @all));
    my %is_valid = map {$_=>1} IW::DepValid::valid_in(\@all);
    return \%is_valid;
}

#------------------------------------------------------------
# print_form_category CATEGORY
#------------------------------------------------------------
# I<Instance method, override.>
# Output a single category.
#
sub print_form_category {
    # $action_to_execute can either be Reset/Clear
    my ($self, $cat, $action_to_execute ) = @_;
    
    # Lie to ourselves about the form input:
    $self->set_CGI;

    #------------------------------
    # Get the criterion objects, and a naughty/nice list of variants...
    #------------------------------
    
    # Get ALL the criterion objects FOR THIS CATEGORY:
    debug "Getting criteria for category: ", $self->category;
    my @all_criteria = 
	$QUERY->depvalid_interface->get_criteria($self->category);
    debug "all criteria = ", join(", ", map {$_->variant} @all_criteria);

    # Get CURRENTLY-VALID variants (as hashref keyed on variant name):
    my %variant_is_valid = %{$self->get_valid_variants};

    # Get CURRENTLY-SELECTED variants (as hashref keyed on variant name):
    my %variant_is_selected = map {$_->variant => 1
				   } $QUERY->criteria($self->category);
   
    # NOTE: we display a form for a variant ONLY IF
    #   (1) that variant is currently valid, OR ELSE IF
    #   (2) that variant currently has a value in the query.
    # So let's filter out the bad criteria right now!
    my @criteria = grep {
	($variant_is_valid{uc($_->variant)} ||
	 $variant_is_selected{uc($_->variant)})
	} @all_criteria;
    debug "final criteria = ", join(", ", map {$_->variant} @criteria);

    

    #------------------------------
    # Display them!
    #------------------------------

    # Criteria:
    my $crit;
    $HTML -> CENTER;
    $HTML -> TABLE(BORDER=>4, CELLPADDING=>4);
    foreach $crit (@criteria) {
	$HTML -> TR;
	$crit->print_form(Prefix  => "p.",
			  Invalid => !($variant_is_valid{uc($crit->variant)}),
			  Action  => $action_to_execute);
    }
    $HTML -> _TABLE;
    $HTML -> _CENTER;

    # Submit:
    $HTML -> CENTER;
    $HTML -> FONT(SIZE=>'+2');
    $HTML -> msubmit_button(QUERY => ['submode'=>'OK'],
			    CAPTION => "OK!");
    # Cancel button
    my $mode = $SESSION->state->{SearchFormMode};
    $HTML -> msubmit_button(CAPTION=>'Cancel',
                             QUERY=>['submode' => 'GOTO']);

    # Reset Button
    $HTML -> msubmit_button(CAPTION=>'Reset defaults',
                             QUERY=>['submode' => 'RESET_PAGE']);

    # Clear Button
    $HTML -> msubmit_button(CAPTION=>'Clear form',
                              QUERY=>['submode' => 'CLEAR']);

    $HTML -> _FONT;
    $HTML -> _CENTER;
}

#------------------------------------------------------------
# print_form_nothing CATEGORIES
#------------------------------------------------------------
# I<Instance method, override.>
# Output the main body of the page.
#
sub print_form_nothing {
    my ($self, $cats) = @_;

    # Directions:
    $HTML -> directions;
    $HTML -> t("No additional search categories are compatible ",
	       "with your current query. ")
	  -> t_B("Please hit OK to continue.");
    $HTML -> _directions;
    
    # Start table:
    $HTML -> CENTER;
    $HTML -> FONT(SIZE=>'+2');
    $HTML -> msubmit_button(QUERY   => ['submode'=>'GOTO'], 
			    CAPTION => "OK");
    $HTML -> _FONT;
    $HTML -> _CENTER;
}



#------------------------------
#
# Setting values...
#
#------------------------------

#------------------------------------------------------------
# ok
#------------------------------------------------------------
# I<Instance method, concrete override.>
# User hit an "OK" button.  Use CGI input to set valids in query object.

sub ok {
    my $self = shift;

    # Get the criterion objects:
    debug "ok: Getting criteria for category: ", $self->category;
    my @criteria = $QUERY->depvalid_interface->get_criteria($self->category);
    my @selected;

    # Fill in from CGI, and store in query:
    my $crit;
    foreach $crit (@criteria) {

	# Read CGI input for this criterion:
	debug "ok: reading CGI input for ", $crit->name;
	$crit->read_CGI(Prefix=>('p.'.$crit->name.'.'));

	# Was it not blank?  If so, add it to the selected criteria:
	if ($crit->user_made_selections) {      # not blank!
	    debug "ok: user made selections for ", $crit->name;
	    push @selected, $crit;
	}
    }

    # Store the new criteria in the query, and install!
    debug "ok: Storing selections in: ", $self->category;
    $QUERY->criteria($self->category, \@selected);
    $QUERY->install;     
    my $cat;
    my @variants;
    my $total;
    my @xs_cats = grep {IU::Category->is_extended($_)} $QUERY->categories;
        foreach $cat (@xs_cats) {
            @variants = map {$_->variant} $QUERY->criteria($cat);
	    $total = $total + int(@variants) if(@variants);
	}
    puke 517, "Your total number of" 
     . " attributes chosen now exceeds the system limit of ${IU::PSAMax}."
     . " Please use the back button of your browser to return to the"
     . " previous page and reduce the number of attributes chosen."
     if($total > ${IU::PSAMax});
    return $QUERY;
}


#------------------------------
#
# Submode handlers
#
#------------------------------
# Inherited handlers are sufficient.


#------------------------------------------------------------
# submode_CRITFORM
#------------------------------------------------------------
sub submode_CRITFORM {
    my $self = shift;

    # Set chosen category (if any), and resend page:
    my $chosen_cat = $REQ->param('p.category');
    my $newcat = upperpretty2category($chosen_cat);
    $chosen_cat = $newcat if(IU::Category->is_core($newcat));
    $self->category($chosen_cat) if $chosen_cat;
    my $modename = $SESSION->state->{SearchFormMode};
    if(!$chosen_cat) {

        puke "Please select an attribute";

        return 1; 
    }
    if (IU::Category->is_core($chosen_cat)) {

        use IW::ModeSLCTDEPVAL;
        my $mode = IW::ModeSLCTDEPVAL->new(Category=>$chosen_cat);
        $mode->goto;

        return 1;
    }
    $self->send_page(ReturnTo=>'MAIN_SEARCH_FORM');
}

#------------------------------------------------------------
# submode_CLEAR
#------------------------------------------------------------
# The user wants to clear out the values completely.
# New values : Type = "", Pulldown  = "(none)", SelectMany = deselected.

sub submode_CLEAR {
    my $self = shift;
    # Set chosen category (if any), and resend page:
    my $chosen_cat = $REQ->param('p.category');
    $self->category($chosen_cat) if $chosen_cat;
    $self->send_page(ReturnTo=>'MAIN_SEARCH_FORM', Action => 'Clear');
}

#------------------------------------------------------------
# submode_RESET_PAGE
#------------------------------------------------------------
# The user wants the "reset" the values to it's defaults ( if available) or
# it'll set'em to Type = "", Pulldown  = "(none)", SelectMany = deselected.

sub submode_RESET_PAGE {
    my $self = shift;

    # Set chosen category (if any), and resend page:
    my $chosen_cat = $REQ->param('p.category');
    $self->category($chosen_cat) if $chosen_cat;
    $self->send_page(ReturnTo=>'MAIN_SEARCH_FORM', Action => 'Reset');
}

#------------------------------------------------------------
1;
