package IW::ModeSLCTDEPVAL;

=head1 NAME

IW::ModeSLCTDEPVAL - core dependent valids selections

=cut

use IU;
use IU::Log;
use IU::Session;
use IU::ODL;
use IU::Query;
use IU::Category;
use IU::DAACExpansions;
#use IU::DAACNames;

use IW::App;
use IW::Util qw(:url);
use IW::DepValid;
use IW::Widget::DepValids;    
use IW::ModeSLCT;
use IW::ModeMOREINFO;

@ISA = qw(IW::ModeSLCT);

use vars qw($User);
use strict;


#------------------------------
#
# Globals...
#
#------------------------------

# Special: get expansions of data center acronyms.
my $daacs_names = new IU::DAACExpansions;
$daacs_names->parse_ODL;
my %DataCenterCmnts = $daacs_names->get;

my $Selected = undef;
my $oldchoices = undef;

#------------------------------------------------------------
# new
#------------------------------------------------------------
# I<Class method, constructor.>
# If CATEGORY was not obtained, puke.
#
sub new {
    my $self = shift->SUPER::new(@_);      # inherited constructor
    $self->category or puke 600, "can't make object without category!";
    $self;
}

#------------------------------------------------------------
# comment_table
#------------------------------------------------------------
# I<Instance method.>
# Get a ref to a hash which maps clean valids to their comments.

sub comment_table {
    my $self = shift;
    if ($self->category eq 'DATA_CENTER_ID') {
	return \%DataCenterCmnts;
    }
    {};
}

#------------------------------------------------------------
# get_filters
#------------------------------------------------------------
# I<Instance method.>
# Get the filters for the given search form context. 
# This is used by the search form's popups to generate filters lists.

sub get_filters {
    my $self = shift;

    # Create a new dependent valids interface, based on current query:
    my $depval = $QUERY->depvalid_interface;

    # Get our new filters, and return 'em!
    my @filters = $depval->get_filters($self->category);
    debug "get_filters got ", int(@filters), " filters";
    @filters;
}

#------------------------------------------------------------
# get_valids HASHREF
#------------------------------------------------------------
# I<Instance method.>
# Get the valids for the given search form context, under the
# given filters, and matching the given pattern.
# The valids will be annotated with "*"s when appropriate.
#
# This is used by the search form's popups to generate valids lists.

sub get_valids {
    my $self = shift;
    my %params = @_;
    my @filters  = @{$params{Filters} || []};
    my @patterns = @{$params{Patterns} || []};
    debug "SLCTDEPVAL: get_valids: ", int(@filters), " filters";
    debug "SLCTDEPVAL: get_valids: ", int(@patterns), " patterns";

    # Create a new dependent valids interface, based on current query:
    my $depval = $QUERY->depvalid_interface;

    ### KLUDGE: If we have no filters or patterns, we once needed to use the 
    ### universal pattern or else we'd get nothing back.  This no longer
    ### appears to be the case -- [Eryq]
    # @patterns = ("*") if (!@filters && !@patterns);

    # Get our new valids:
    my @valids = $depval->get_valids($self->category,
				     Patterns => \@patterns,
				     Filters  => \@filters);
    debug "get_valids got ", int(@valids), " valids";

    # Done!
    @valids;
}

#------------------------------------------------------------
# REQ_filters
#------------------------------------------------------------
# I<Instance method.>
# Get the CGI filter args, removing horizontal lines and other garbage.
#
#    * The valids are "cleaned" of leading "*", etc.  
#    * Items consisting of just dashes (separators) are removed.
#    * Items which contain nothing but whitespace are removed.

sub REQ_filters {
    my $self = shift;
    my @items = map { clean_valid($_) } $REQ->param('p.filter');
    grep {!/^-+$|^\s*$/} @items;               # return good stuff
}

#------------------------------------------------------------
# REQ_patterns
#------------------------------------------------------------
# I<Instance method.>
# Get the CGI patterns arg, cleaning it up a bit.

sub REQ_patterns {
    my $self = shift;
    $_ = $REQ->param('p.pattern'); defined($_) or $_ = '';
    s/\s+/ /g;              # collapse sequences of whitespace to one space
    return (grep /\S/, map { s/^\s*|\s*$//g; $_ } split /;/, $_);
}

#------------------------------------------------------------
# REQ_quick_pattern
#------------------------------------------------------------
# I<Instance method.>
# Get the CGI quick_pattern arg, cleaning it up a bit.

sub REQ_quick_pattern {
    my $self = shift;
    $_ = $REQ->param('p.quick_pattern'); defined($_) or $_ = '';
    s/\s+/ /g;              # collapse sequences of whitespace to one space
    return (grep /\S/, map { s/^\s*|\s*$//g; $_ } split /;/, $_);
}

#------------------------------------------------------------
# REQ_add
#------------------------------------------------------------
# I<Instance method.>
# Get the CGI add arg, cleaning it up a bit.

sub REQ_add {
    my $self = shift;
    $_ = $REQ->param('p.add'); defined($_) or $_ = '';
    s/\s+/ /g;              # collapse sequences of whitespace to one space
    return (grep /\S/, map { s/^\s*|\s*$//g; $_ } split /;/, $_);
}


#------------------------------------------------------------
# REQ_valids
#------------------------------------------------------------
# I<Instance method.>
# Get the CGI valids (choices) args, removing horizontal lines and other 
# garbage.
#
#    * The valids are "cleaned" of leading "*", etc.  
#    * Items consisting of just dashes (separators) are removed.
#    * Items which contain nothing but whitespace are removed.

sub REQ_valids {
    my $self = shift;
    my @items = map { clean_valid($_) } $REQ->param('p.valid');
    grep {!/^-+$|^\s*$/} @items;            # return good stuff
}



#------------------------------
#
# Output...
#
#------------------------------

#------------------------------------------------------------
# print_selection_form HASHREF
#------------------------------------------------------------
# I<Instance method.>
# Output the menu.
#
#    Valids  => The valids.  Invalid ones are preceded by a *.
#    Filters => The filters.  Invalid ones are preceded by a *.

sub print_selection_form {
    my $self = shift;
    my %params = @_;

    # Valids and filters:
    my $valids  = $params{Valids};
    my $filters = $params{Filters};

    # Were filters/patterns used in generating this list?
    my $filters_used   = $REQ->param('p.filter');
    my $patterns_used  = $REQ->param('p.pattern');
    my $narrowing_used = ($filters_used || $patterns_used);
    my $select_flag = $REQ->param('select');
    my $valids_menu;
    my $cat = $self->thingy;
    my @empty;

    # Get valids (all, selected, unselected), and create a menu:
    my @lastrequest = $self->REQ_valids;
    my @newrequest = (@$valids,@lastrequest);
    my $request = \@newrequest;

    if ($select_flag eq 'NONE') {
        debug " in NONE";
        $valids_menu =
	    new IW::Widget::DepValids
	        All      => $valids,
	        Selected => \@empty,
	        Comments => $self->comment_table;
    } elsif ($select_flag eq 'ALL') {
        $valids_menu =
	    new IW::Widget::DepValids
	        All      => $request,
	        Selected => $request,
	        Comments => $self->comment_table;
    } else {
        $valids_menu =
	    new IW::Widget::DepValids
	        All      => $valids,
	        Selected => [$REQ->param('p.samepage') 
			     ? $self->REQ_valids 
			     : $QUERY->criteria_strings($self->category)],
	        Comments => $self->comment_table;
    }
    # Get filters (all, selected, unselected), and create a menu:
    my $filters_menu = 
	new IW::Widget::DepValids
	    All      => $filters;
    
    # Set up selections:
    $REQ->param('p.valid' => (keys %{$valids_menu->selected}));

    # KLUDGE: Make sure subsequent pages know that we came from this one.
    # This allows us to detect the case of the user de-selecting all choices
    # and then coming back here.
    $HTML->hidden("p.samepage"=>1);

    $HTML->hidden("p.add"=>$patterns_used) if defined($patterns_used);

    # Patterns and filters section:

    $HTML -> head(2, "Filters for $cat:");
    $HTML -> small;
    print "Type in patterns ",
           qq!(where ";" separates patterns and "*" is a wildcard) !,
          " or select predefined filters from the list below ",
          "and hit <B> APPLY </B>. ",
          "<B>Example: </B>for choices beginning with A or B, type: ",
          "A* ; B*.  ";
    print "<B>Note: </B>",
          "<I>the type-in field will be used in conjunction",
          " with the filters!</I>";
    $HTML -> _small -> BR;

    # Apply button.

    # $HTML -> FONT(SIZE=>'+2', COLOR=>'#0000E0')
    $HTML -> msubmit_button(QUERY => ['submode'=>'UPDATE'],
                            CAPTION => "Apply");
    #      -> _FONT;

    # Pattern type-in box.

    $HTML -> input_text(Name=>"p.pattern",
                    Size=>30);

    #--------------------
    # FILTERS...
    #--------------------

    $HTML->BR;
    ### Are there filters?
    if (@$filters) {                # WE HAVE FILTERS!

        $HTML -> select_many(Name=>"p.filter",
                             Size=>5,
                             Choices=>[$filters_menu->items]);
    }
    else {                          # NO FILTERS!

        $HTML -> BR;
        $HTML -> small;
        $HTML -> t_B("No compatible filters for ",
                     $self->thingy, ".\n");
        $HTML -> _small -> BR;
    }


    ### Are there choices?
       
    $HTML -> head(2,$cat."s:");

    if (@$valids) {                 # WE HAVE VALIDS!
        
	$HTML -> small -> B;
	$HTML -> t("Select the ", $self->thingys, " which you would like to ",
		   "be a part of your query, and hit OK. ");
	$HTML -> _small -> _B; 

        # Definition and Detailed info buttons.
      
        $HTML -> BR; 
        $self->print_moreinfo_form;
        $HTML -> BR;
        $HTML -> small
              -> t("There are a total of ", $valids_menu->count->{Valid},
                   " compatible choices for ", $self->thingy, ".")
              -> _small -> BR;

	$valids_menu->print(Name=>'p.valid', Size=>10, 
                            Title => $self->thingy . " list");

        ### All/None Buttons:
        $HTML ->BR;
        $HTML -> msubmit_button(QUERY =>['submode'=>'UPDATE', 
					 'select' =>'ALL'],
                                SRC   =>doc_url("btns/sub-slctall.gif"));
        $HTML -> t("   ");
        $HTML -> msubmit_button(QUERY =>['submode'=>'UPDATE', 
					 'select' =>'NONE'],
				SRC   =>doc_url("btns/sub-slctnone.gif"));
	$HTML -> small; 
        $HTML -> t_B("\nShortcut: these buttons select all/none from the ",
		     "above list. ");
	$HTML -> _small; 
    }
    else {                          # NO VALIDS
       
        $HTML -> BR; 
	$HTML -> small;
	$HTML -> t_B("No compatible ", 
		     $self->thingy, "s for:\n") -> BR;
        $HTML -> t_B($patterns_used,"\n") -> BR;
	if ($narrowing_used) {
	    $HTML -> t("Perhaps your choice of filters/patterns was too ",
		       "limiting... try changing them and hitting ");
            $HTML -> t_B("Apply.");
	    if($SESSION->state->{AllowDynamicValids} eq 'Y' and
               $self->thingy ne 'Data Center') {
                $HTML -> BR 
                      ->t("Or you can add your selections to dynamic valids ",
                           "by hitting ");
                $HTML -> t_B("ADD.");
	    }
            $HTML -> BR;
	}
	else {
	    $HTML -> t("Apparently, there are no choices for ",
		       $self->thingy, " which would be compatible with the ",
		       "other criteria in your query.  Please hit OK to ",
		       "continue.");
	}
	$HTML -> _small -> BR;	
    }
   
    $HTML -> P; 
    $HTML -> CENTER;
    $HTML -> FONT(SIZE=>'+2');

    # Add button

    if ($SESSION->state->{AllowDynamicValids} eq 'Y' and
         $narrowing_used and $self->thingy ne 'Data Center' and !(@$valids)) {
         $HTML -> msubmit_button(QUERY => ['submode'=>'ADD'],
                                    CAPTION => "ADD...");

        $HTML->t_B(" ".$patterns_used)
             ->t(" to the ".$self->thingy."s")
             ->BR; 

   }
    
   # OK button
    
   if (@$valids) { 
        $HTML -> msubmit_button(QUERY => ['submode'=>'OK'], CAPTION => "OK!");
   }

   # Cancel button

   $HTML -> msubmit_button(CAPTION=>'Cancel',
                             QUERY=>['submode' => 'GOTO']);
   $HTML  -> _FONT;
   $HTML -> _CENTER;
    
   1;
}


#------------------------------------------------------------
# print_moreinfo_form HASHREF
#------------------------------------------------------------
# I<Instance method.>
# Output the form for getting more info.  HASHREF contains.
#
#    Valids  => The valids.  Invalid ones may be preceded by a *.

sub print_moreinfo_form {
    my ($self, %params) = @_;

    debug "inside print_moreinfo_form";
    my $cat = $self->category;
    debug "cat is $cat";
    if (IU::Category->lookup($self->category)->{HasDefn}) {

         $HTML -> msubmit_button(QUERY => ['submode'=>'INFO',
                                             'category' => $self->category,
                                             'info' => 'DEF'],
                               CAPTION => "Definition");
    }

    if (IU::Category->lookup($self->category)->{HasDI}) {

         $HTML -> msubmit_button(QUERY => ['submode'=>'INFO',
                                             'category' => $self->category,
                                             'info' => 'GUIDE'],
                               CAPTION => "Detailed Document");
    }
    $HTML -> t(" for ...");
1;
}

#------------------------------------------------------------
# print_form
#------------------------------------------------------------
# I<Instance method, override.>
# Output the main body of the page.
#
sub print_form {
    my $self = shift;

    # Get currently-valid valids (based on CGI-specified filters and patterns):
    my @valids  = $self->get_valids(Patterns =>[$self->REQ_patterns],
				    Filters  =>[$self->REQ_filters]);

    # Get currently-valid filters:
    my @filters = $self->get_filters();       # currently-valid filters


    # Selection form:
    $self->print_selection_form(Valids  => \@valids,
    				Filters => \@filters);

    1;
}

#------------------------------------------------------------
# start_page PARAMHASH
#------------------------------------------------------------
# I<Instance method.>
# Overriding start_page in ModeSLCT.pm

sub start_page {
    my $self = shift;

    # Get pretty name for search type:
    my $pretty_type = $QUERY->search_type;
    $pretty_type =~ tr/A-Z_/a-z /;
    $pretty_type =~ s{\b([a-z])}{\u$1}g;

    # Start page:
    $APP->suppress_key(1);


    $APP->start_page(SuperTitle => "Search Form: "
                                . valid2pretty($QUERY->search_type),
                     Title      => ("Keyword Selection: " . $self->title),
                     Context    => $self->context,
                     @_);        # allow overriding

}





#------------------------------
#
# Setting values...
#
#------------------------------

#------------------------------------------------------------
# ok
#------------------------------------------------------------
# I<Instance method, concrete override.>
# User hit an "OK" button.  Use CGI input to set valids in query object.

sub ok {
    my $self = shift;

    # Get any selected valids (removing interface garbage):
    my @valids = sort $self->REQ_valids;

    # Update the query object with them (for this category), and re-store it:
    $QUERY->criteria_strings($self->category, \@valids);
    $QUERY->install;
    1;
}




#------------------------------
#
# Submode handlers...
#
#------------------------------


#------------------------------------------------------------
# submode_INFO
#------------------------------------------------------------
# A request for more info on a valid has been received.
# Determine the best way to service that request, then transfer
# to that mode.
#
sub submode_INFO {
    my $self = shift;

    $REQ->param('submode'=>'CGI');
    $REQ->param('mode'=>'MOREINFO');
    my $newmode = IW::ModeMOREINFO->new;
    $newmode->go;

    1;
}

#------------------------------------------------------------
# submode_UPDATE
#------------------------------------------------------------
# Update the choices, based on selected filters/patterns.

sub submode_UPDATE {
    my $self = shift;

    # Re-show:
    $self->send_page;
}


#------------------------------------------------------------
# submode_ADD
#------------------------------------------------------------
# Update the choices, based on selected filters/patterns.

sub submode_ADD {
    my $self = shift;
#    my @patterns = $self->REQ_patterns;
    my @patterns = $self->REQ_add;
    my @oldvalids = [];
    my $queryODL;
    my $inputODL;
    my $existed;
    my $query;
    my $param = $self->category;
    my $path = $SESSION->path_to('dynamic.odl');
    my $vals;
    my $depvalid;
    my $odlfile;
    if (! -s $path) {
	$existed = 0;
        $queryODL = IU::ODL->new_group;
        $vals = \@patterns;
        $queryODL->add_param($param => IU::ODL->conv($vals));

        # Create a new dependent valids interface:
        $depvalid = new IW::DepValid;

        # Create ODL input file:

        $odlfile = $depvalid->write_query($queryODL,
           File=>'dynamic.odl', Type=>'INVENTORY_SEARCH');

    } else {
        # Parse it into an ODL object:
        debug "reloading query in $path";
        open ODL, $path or puke "open $path: $!";
        $inputODL = eval { parse IU::ODL \*ODL };
        close ODL;
        $inputODL or puke "couldn't parse ODL in $path (line $.)";
	$existed = 1;
	$query = IU::Query->from_ODL($inputODL)
		or puke "couldn't extract valid user data from ODL";
	@oldvalids = $query->criteria_strings($param);
	@patterns = (@oldvalids,@patterns);
        debug "patterns[0] = ", $patterns[0];
        debug "param = ", $param;
	$query->criteria_strings($param,\@patterns);
	$queryODL = $query->to_ODL(NoDaacs=>1) || 
	  puke "couldn't convert query to ODL";
	unlink($path);
        open QUERY, ">$path" or puke 452, "can't open $path for writing: $!";
        $queryODL->output(\*QUERY);
	$queryODL->end(\*QUERY);
        close QUERY;

    }
    # Re-show:
    $self->send_page;
}
#------------------------------------------------------------
1;
