package IW::ModeSLCTGEO;

=head1 NAME

IW::ModeSLCTGEO - mode to manage geographic selection

=head1 NOTES

A session state variable is used to hold the last geoselection
map/method name, so that the user goes right to that map/method
the next time they come to this screen.

=head1 CGI TAGS

=over

=item p.geoopt

Describes what geographic region is to be used when the user hits OK.
Normally, this will be either none, global granules, or the current
map selection.

=item p.geotype

Encodes how the geoselection is embedded in the CGI parameters.
Values are RANGE_LOC, POLYGON_LOC, PATH_ROW_LOC, etc.

=item p.mapname

The name of the geoselection method used in the page we just came from.
Values are "Atlas", "Equatorial", etc.

=item p.mapname-next

The name of the geoselection method desired for the next page we're 
generating (shown in the pulldown menu).  Values are "Atlas", "Equatorial", 
etc.

=item p.state

The current state of the page.  Values are dependent on the
geoselection method currently being used.

=back

=cut

use IU;
use IU::Log;
use IU::Coord;
use IU::Query;
require IU::Setup;

use IW::App;
use IW::GeoSelect::Base;
use IW::ModeSLCT;
use IW::Util qw(:url blank);

use strict;
use vars qw(@ISA);

@ISA = qw(IW::ModeSLCT);

#------------------------------------------------------------
# new PARAMHASH
#------------------------------------------------------------
sub new {
    shift->SUPER::new(Category=>'GEOGRAPHIC_REGION', @_);
}

#------------------------------------------------------------
#
#------------------------------------------------------------
sub menu {

   my $self = shift;

   # Menu choices:
   my @SwitchMenu = (
                    "Orthographic Map (Java)",
                    "------------------------------",
                    "Equatorial Map",
                    "Stereographic N-pole Map",
                    "Stereographic S-pole Map",
                    "------------------------------",
                    "Type in Lat/Lon Range",
                    "Type in Path/Row",
                    "------------------------------",
                    "I want to do a GLOBAL SEARCH",
                    "I want GLOBAL GRANULES ONLY",
                  # "I want NO GEOGRAPHIC REGION in my query",
                    ); 
   return \@SwitchMenu;
}

#------------------------------
#
# Output...
#
#------------------------------

#------------------------------------------------------------
# req_mapname TAG
#------------------------------------------------------------
# I<Instance method.>
# Get the requested mapname in tag TAG.  If there appears to
# be none, or if they chose a separator or parenthetic note,
# return undef.

sub req_mapname {
    my ($self, $tag) = @_;
    my $mapname = $REQ->param($tag);
    # If no map chosen, return undef.
    return undef unless (defined($mapname));
    # Map was good, return it's name.
    return $mapname if ($mapname !~ /^(--|\()/);
    $self->{Warn} = 1;
    return undef;
}

#------------------------------------------------------------
# print_form_to_select_widget
#------------------------------------------------------------
# I<Instance method, override.>
# Output the main body of the page.
#
sub print_form_to_select_widget {
    my $self = shift;

    # If they chose a separator or parenthetic note, warn them.
    if ($self->{Warn}) {
       $HTML-> directions;
       print "<B>Do not choose the line separators or ",
             "the parenthetic descriptions in the list.</B>\n";
    }
    # Directions:
    $HTML -> directions;
    print <<EOF;
    <B> Choose a map/method from the list below, and hit the OK button. </B> 
    <p style="margin-left:5%">
    <B>Note:</B> If you are about to use the Java enabled map for the first
    time, we recommend that you read the <a href="$IU::JavaFAQUrl">
    Java FAQ</a> section concerning possible
    <a href="$IU::JavaFAQUrl#problems">problems using the applet.</a>
    </p>
EOF
    $HTML -> _directions;

#    $HTML -> P -> small;
#    print <<EOF;
#       Choices beginning with "I want..." will require no 
#       further action in your part; the others will take you to additional
#       screens where you'll use map-selection or type-in boxes. <P>
# EOF
#    $HTML -> _small;

    # Get map to use, and set as default:

    # "Java enabled Map" changed to "Orthographic Map (Java)", 
    # fix up LastMapname from old session states, so we
    # are backwards compatible.


    $SESSION->state->{LastMapname} = 'Orthographic Map (Java)' 
       if ($SESSION->state->{LastMapname} eq 'Java enabled Map');

    my $draw_mapname = $self->req_mapname('p.mapname') || # explicit selection
	               $SESSION->state->{LastMapname}  || # last map used 
                       'Orthographic Map (Java)';         # fallback

    $REQ->param('p.mapname-next'=>$draw_mapname);

    # Output widget:
    $HTML -> CENTER;
    $HTML ->t("Select region using...");

#    $HTML -> msubmit_button(CAPTION=>"SELECT region using...",
#			    QUERY=>[### 'mode'    => 'SLCTGEO',
#				    'submode' => 'SWITCH']);

    $HTML -> select_one(Name=>'p.mapname-next',
			Choices=>$self->menu() );   
    $HTML -> _CENTER -> P;

    $HTML -> CENTER;

    $HTML -> FONT(SIZE=>'+2');
    $HTML -> msubmit_button(CAPTION=>"OK!",
                            QUERY=>[### 'mode'    => 'SLCTGEO',
                                    'submode' => 'SWITCH']);

    $HTML -> msubmit_button(CAPTION=>'Cancel',
                            QUERY=>[### 'mode'    => 'SLCTGEO',
                                    'submode' => 'GOTO']);
    $HTML -> _FONT;
    $HTML -> _CENTER -> P;
}

#------------------------------------------------------------
# print_form_for_widget
#------------------------------------------------------------
# I<Instance method, override.>
# Output the main body of the page.
#
sub print_form_for_widget {
    my $self = shift;

    #--------------------
    # PROCESS INPUT FROM LAST MAP....
    #--------------------

    # Get the "last" map (the one the user would have clicked on).
    # We'll need this to decode their map click:
    my $mapname = $self->req_mapname('p.mapname') ||   # explicit selection
	          $SESSION->state->{LastMapname}  ||   # last map used
		  'Orthographic Map (Java)';           # 1st time: default map 

    my $geoslct = lookup IW::GeoSelect::Base $mapname;
    $geoslct->state($REQ->param('p.state'));          # set state

    # Get input from user's last action (tweaking x,y clicks, etc.):
    $geoslct->get_input;

    #--------------------
    # OUTPUT USING NEXT MAP....
    #--------------------

    # Get the "next" map (the one the user wants to see next, from pulldown).
    # Usually, this will be identical to the "last" map.
    # If none selected, default to current map.
    my $draw_mapname = $self->req_mapname('p.mapname-next') || $mapname;
    my $draw_mapslct = $geoslct->switch_to($draw_mapname);
    $REQ->param('p.mapname-next' => $draw_mapname);      # init widget

    # Remember this:
    $SESSION->state->{LastMapname} = $draw_mapname;

    # The map/form:
    $HTML -> hidden('p.mapname' => $draw_mapname);
    $draw_mapslct->draw;

    #--------------------
    # SUBMISSION BUTTONS...
    #--------------------

    # The submission buttons:
    $self->print_ok($draw_mapslct->ok_is_okay);
    $self->print_switcher;
    1;
}

#------------------------------------------------------------
# print_form
#------------------------------------------------------------
# I<Instance method, override.>
# Output the main body of the page.
#
sub print_form {
    my $self = shift;

    if (!$self->req_mapname("p.mapname-next")) {       # came from page
	$self->print_form_to_select_widget;
    }
    else {
	$self->print_form_for_widget;
    }
}

#------------------------------------------------------------
# print_ok ISOKAY
#------------------------------------------------------------
# Print the buttons which submit the form.

sub print_ok {
    my ($self, $is_ok) = @_;

    $HTML -> CENTER;
    $HTML -> FONT(SIZE=>'+2');

    # OK:       
    if ($is_ok) {
	$HTML -> msubmit_button(CAPTION=>'OK!',
				QUERY=>[### 'mode'    => 'SLCTGEO',
					'submode' => 'OK']);
    }

    # CANCEL:
    $HTML -> msubmit_button(CAPTION=>'Cancel',
			    QUERY=>[### 'mode'    => 'SLCTGEO',
				    'submode' => 'GOTO']);
    $HTML -> _FONT;
    $HTML -> _CENTER;
}

#------------------------------------------------------------
# print_switcher
#------------------------------------------------------------
# Print the buttons which switch the map type.

sub print_switcher {
    my $self = shift;

    $HTML -> HR;

    # Directions:
    my $this_map = $REQ->param('p.mapname-next');
    $HTML -> P -> small;
    print <<EOF;
          <B>Optional: if you don't want to select your region via
          the means shown above,</B> simply select the map/method
          you <I>want</I> to use below, and hit the SELECT button.

EOF
    $HTML -> _small;
    $HTML -> CENTER;
    $HTML -> TABLE(CELLPADDING=>2, CELLSPACING=>0, BORDER=>0);
    $HTML -> TR;
    
    # Pulldown:
    $HTML -> TD;
    $HTML -> msubmit_button(CAPTION=>'SELECT',
			    QUERY=>[### 'mode'   => 'SLCTGEO',
                                    'submode'=>'SWITCH',
				    'p.state'=> $REQ->param('p.state')]);
    $HTML -> t(" region using... ");
    $HTML -> select_one(Name=>'p.mapname-next',
			Choices=>$self->menu() );
    # Done:
    $HTML -> _TD -> _TR -> _TABLE -> _CENTER;
}

#------------------------------------------------------------
# context
#------------------------------------------------------------
# I<Instance method, abstract.>
# Return the proper help context for this mode/submode combo.
sub context {
    my $self = shift;
    my $mapname = $self->req_mapname('p.mapname-next');
#    debug "context: map is '$mapname'";
    return "geoselpg" unless $mapname;

    my $geoslct = lookup IW::GeoSelect::Base $mapname;
#    debug "Geoselector is a $geoslct";
    my $draw_mapslct = $geoslct->switch_to($mapname);
#    debug "Map to switch to is a $draw_mapslct";
    return $draw_mapslct->context();
}

#------------------------------------------------------------
# jumpmenu_is_okay
#------------------------------------------------------------
# I<Instance method, abstract.>
# Return true if the jumpmenu should be shown, undef if not.
# Default method assumes jumpmenu should be shown.
# You can override this in your subclasses.
#sub jumpmenu_is_okay {
#    my $self = shift;
#    my $mapname = $self->req_mapname('p.mapname-next');
#    debug "Jumpmenu: map is '$mapname'";
#    my $geoslct = lookup IW::GeoSelect::Base $mapname;
#    debug "Geoselector is a $geoslct";
#    my $draw_mapslct = $geoslct->switch_to($mapname);
#    debug "Map to switch to is a $draw_mapslct";
#    return $draw_mapslct->jumpmenu_is_okay();
#}

#------------------------------
#
# Setting values...
#
#------------------------------

#------------------------------------------------------------
# ok
#------------------------------------------------------------
# I<Instance method, concrete override.>
# User hit an "OK" button.  Use CGI input to set valids in query object.

sub ok {
    my $self = shift;

    # Get the geo option:
    my $ok_opt = $REQ->param("p.geoopt");
    my ($geo_type, $geo);
    if ($ok_opt =~ /^NO/) {           # no geo option...
	# geo is undefined
    }
    elsif ($ok_opt =~ /GLOBAL/) {     # global granules only...
	$geo_type = 'GLOBAL_GRANULES_ONLY';
	$geo      = IU::Geo::Global->new;
    }
    else {                            # use selection (map or fill-in)...
	$geo_type = $REQ->param("p.geotype");
	if ($geo_type eq 'RANGE_LOC') {          # rectangle...
	    $geo = IU::Geo::Range->from_CGI("p.");
	}
	elsif ($geo_type eq 'POLYGON_LOC') {     # polygon...
	    $geo = IU::Geo::Polygon->from_CGI("p.");
	}
	elsif ($geo_type eq 'PATH_ROW_LOC') {    # path-row...
	    $geo = IU::Geo::PathRow->from_CGI("p.");
	}
	else {
	    puke "Unknown geoselection option ($geo_type)!";
	}
    }
    # Update the query object with the new geo option, and re-store it:
    $self->set_in_query($geo);
}

#------------------------------------------------------------
# set_in_query GEO
#------------------------------------------------------------
# I<Instance method.>
# Update the query object with the new geo option, and re-store it.
# The GEO can be undefined; this will delete the geo region from the query.

sub set_in_query {
    my ($self, $geo) = @_;

    # Update the query object with the new geo option, and re-store it:
    if ($geo) {
	$geo->fixup->validate;         # fix up and validate geo
    }
    $QUERY->geo_region($geo);          # store in query object
    $QUERY->install;                   # install query object
}

#------------------------------
#
# Submode handlers
#
#------------------------------
# Inherited handlers are sufficient.

#------------------------------------------------------------
# submode_DEFAULT
#------------------------------------------------------------
# Entry point.

sub submode_DEFAULT {
    shift->submode_SWITCH;
}

#------------------------------------------------------------
# submode_SWITCHMAP
#------------------------------------------------------------
# Entry point.

sub submode_SWITCHMAP {
    shift->submode_SWITCH;
}

#------------------------------------------------------------
# submode_SWITCH
#------------------------------------------------------------
# Switch to another type of geoselection.

sub submode_SWITCH {
    my $self = shift;
    my $geo;
    my $done = 1;

    # If switching to an option which requires no further action,
    # behave like an OK button...
    my $mapname = $REQ->param('p.mapname-next');
    if (defined($mapname)) {
       if ($mapname =~ /GLOBAL GRANULES/i) {
	  $self->set_in_query(IU::Geo::Global->new);
	  # Remember this:
	  $SESSION->state->{LastMapname} = $mapname;
       }
       elsif ($mapname =~ /NO GEOGRAPHIC REGION/i) {
	  $self->set_in_query(undef);
       }
       elsif ($mapname =~ m{GLOBAL SEARCH}i) {
	  $self->set_in_query(IU::Geo::Range->new(90.0, -180.0, -90.0, 180.0));
	  # Remember this:
	  $SESSION->state->{LastMapname} = $mapname;
       }
       elsif (!$self->req_mapname) {     # nothing was selected!
	  $done = 0;
       }
       else {                            # something else...
	  $done = 0;
       }
    } else {
       $done = 0;
    }
    # Done?
    if ($done) {
	$self->submode_GOTO;
    }
    else {
	$self->submode_SHOW;
    }
}

#------------------------------------------------------------
1;
