package IW::Navigator;

=head1 NAME
    
  IW::Navigator - Replacement for the Blue Menu Bars

=head1 DESCRIPTION

The C<Navigator> module is simply a data source which organizes all the links
that the EDG will want to show on the left-hand side of the page.

The Navigator class creates a singleton object which is used to both track
the state of the application as a whole (individual modes can "mark" themselves
as active), and to record all the links that are to be provided in the 
navigation bar, along with their organization and order.

=head1 LINK SETS

Links are divided into I<sets>; these sets have both a distinct I<order> and
can be either I<markable> or not.

Link order defines the order in which the individual Navigator::Link objects
are returned to the caller when the set is processed as an array. 

=head1 NOTES

If you subclass this class, you must make sure to clone the object and resave 
it in your subclass! Otherwise, any alteration or reblessing of the object will
be reflected to the copy of the object controlled by this class. This could be
a bad thing.

=cut

use IU;
use IU::Log;
use IU::Session;

use Storable;

use IW::Navigator::Link;

#---------------------------------------------------------------------------
# Class variables.

# The variable holding the singleton Navigator object. Note that no matter
# how many times Navigator->new() is called, the same object will always be
# returned.
my $singleton;

#------------------------------------------------------------------------
# Class methods.
#
# new() returns the contents of $singleton if it is defined; if not, a
# new Navigator object is created and returned; a copy of that new object
# is saved in $singleton for any later calls to new().

sub new {
  my $class = shift;
  return $singleton if defined $singleton;
  $singleton = $class->_create;
  $singleton;
} 

# _create (private) - create a Navigator object

sub _create {
  my $class = shift;
  # Check to see if we can load the saved binary version of the object. We
  # can if the creation dates match on the saved file and the source.
  my $source     = "$IU::LibDir/conf/navigator.conf";
  my $binary     = "$IU::LibDir/conf/navigator.saved";
  my $msg        = "ok";

  # Handle getting the defaults from the lib/conf directory.
  if (-s $binary and (-M $source) > (-M $binary)) {
    # Source older than binary; binary usable.
    eval { $self = retrieve $binary };
    $msg = "reload failed: $@" if $@;
  }
  else {
    # Source newer than binary; recompile and save.
    my $return = do $source;
    $msg = "navigator.conf syntax error: $@" if $@;
    $msg = "couldn't reload navigator: $!"   unless defined $return;
    $msg = "couldn't run $source"            unless $return;
    if ($msg eq "ok") {
      # Good reload, serialize it.
      $self = $IW::Navigator::VAR1;
      eval { store $self, $binary };
      $msg = "Couldn't save new binary: $@" if $@;
    }
    elsif (-s $binary) {
      # bad recompile, load the old one.
      eval { $self = retrieve $binary };
      $msg = "reload failed: $@" if $@;
    }
    else {
      # Messed up even before we built it the first time!
      $msg = "can't compile or reload"; 
    }
  }

  # If all this finagling failed, create a dummy object to save the
  # message in.
  $self = {} unless (defined $self);

  # Set the message, in case we're debugging.
  # If the desired action worked, we don't need to show the message.
  $self->{Msg} = $msg if $IU::ForceDebug eq "YES";

  # Make sure each link can find the navigator.
  foreach $key (keys %$self) {
    next unless defined $self->{$key};
    foreach $entry (@{$self->{$key}}) {
      $entry->navigator($self);
    }
  }
  bless $self, $class;
}

#------------------------------------------------------------------------------
# Instance methods

sub AUTOLOAD {
  use vars qw($AUTOLOAD);
  my($qualifier, $accessor) = ($AUTOLOAD =~ /(.*::)(.*)/);
  {
    no strict 'refs';
    *$AUTOLOAD = sub {
                  my ($self) = shift;
                  $self->{ucfirst($accessor)};
                };
  goto &$AUTOLOAD;
  }
}

sub mark {
   my ($self, $to_mark) = @_;
   $self->{Indicator} = $to_mark if defined $to_mark;
   $self->{Indicator};
}

sub link {
   my ($self, $linkname) = @_;
   foreach my $k (keys %$self) {
     next unless ref $self->{$k} eq "ARRAY";
     foreach my $link (@{$self->{$k}}) {
       return $link if $link->name() eq $linkname;
     }
   }
   undef;
}

#------------------------------------------------------------
1;
