package IW::Note;

=head1 NAME

IW::Note - an admonishment

=head1 DESCRIPTION

Admonishments are single-paragraphs of HTML text with an icon on the left 
side.  They are used so much that there is now a way to build them and pass 
them around.

=head1 SYNOPSIS

    # A note with this message and a default icon:
    my $note = new IW::Note Text=>'Hello, world!');
    
    # A note with this message and an appropriate icon.
    # Known classes are: INFO, WARN, ERROR, RUNNING, OK, FAILED, DIRECTIONS.
    my $note = new IW::Note Text=>'Danger, Will Robinson!',
		            Class=>'WARN');
    
    # Same as above, but supply the HTML text right away:
    my $note = new IW::Note HTML=>'<B>Danger,</B> Will Robinson!', 
                            Class=>'WARN';
    
    # Generate and use, all in one shot:
    $APP->message(new IW::Note Text=>'Hello, world!'); 
    $APP->note(new IW::Note Class=>'WARN', Text=>'Watch out!'); 
     
    # Output a note in a generic way (icon is just left-aligned):
    $note->print;
    
    # Output a note as a row in a table (bounded by <TR> and </TR>):
    $note->print_row;
    
    # Output a note as a complete table (bounded by <TABLE> and </TABLE>):
    $note->print_table;

    # Get just the text of the note (without the icon) for processing.
    $note->as_string;

=cut

use IU;
use IW::App;

use strict;

# Map classes to icons (relative to docs URL):
my %ClassIcon = 
    (
     INFO    => 'icons/note-info.gif',
     WARN    => 'icons/note.gif',     
     ERROR   => 'icons/warning.gif',
     RUNNING => 'icons/status-run.gif',
     OK      => 'icons/status-ok.gif',
     FAILED  => 'icons/status-fail.gif',     
     );

my $NOTE_WIDTH = "96%";
my $LIGHT_WIDTH = "4%";

#------------------------------------------------------------
# new PARAMHASH
#------------------------------------------------------------
sub new {
    my $class = shift;
    my %params = @_;
    my $self = {};
    $self->{HTML} = $params{HTML} || HTML::Stream::html_escape($params{Text});
    $self->{Class} = $params{Class} || 'INFO';
    $ClassIcon{$self->{Class}} or $self->{Class} = 'INFO';
    bless $self, $class;
}

#------------------------------------------------------------
# print
#------------------------------------------------------------
sub print {
    my $self = shift;
    $HTML -> IMG(SRC=>"${IU::DocsUrl}$ClassIcon{$self->{Class}}",
		 ALT=>"[$self->{Class}]", 
		 ALIGN=>'LEFT');
    print $self->{HTML};
    print "<BR CLEAR=ALL>\n";
}

#------------------------------------------------------------
# print_row
#------------------------------------------------------------
sub print_row {
    my $self = shift;
    my $light_yellow = "#FFFFCC";
    my $light_red = "#FF8296";
    my $class = $self->{Class};
    if ($class =~ /^ERR/i) {
       $HTML -> TR -> TD (WIDTH=>$LIGHT_WIDTH)
             -> IMG (SRC=>"${IU::DocsUrl}$ClassIcon{$class}",
                     ALT=>"[$class]")
             -> _TD -> TD (BGCOLOR=>$light_red, WIDTH=>$NOTE_WIDTH,
                           ALIGN=>"LEFT");
       print "<B>", $self->{HTML}, "</B></TD></TR>\n";
    } elsif ($class =~ /^WARN/i) {
       $HTML -> TR -> TD (WIDTH=>$LIGHT_WIDTH)
             -> IMG (SRC=>"${IU::DocsUrl}$ClassIcon{$class}",
                     ALT=>"[$class]")
             -> _TD -> TD (BGCOLOR=>$light_yellow, WIDTH=>$NOTE_WIDTH,
                           ALIGN=>"LEFT");
       print "<B>", $self->{HTML}, "</B></TD></TR>\n";
    } else {
       $HTML -> TR -> TD
             -> IMG (SRC=>"${IU::DocsUrl}$ClassIcon{$self->{Class}}",
                     ALT=>"[$self->{Class}]");
       $HTML -> _TD;
       print "<TD ALIGN=LEFT WIDTH=$NOTE_WIDTH>", $self->{HTML},
             "</TD></TR>\n";
    }
}

#------------------------------------------------------------
# print_table
#------------------------------------------------------------
sub print_table {
    my $self = shift;
    print "<TABLE>"; 
    $self->print_row;
    print "</TABLE>";
}

#------------------------------------------------------------
# as_string
#------------------------------------------------------------
# Just return the HTML string.

sub as_string {
    my $self = shift;
    return ($self->{HTML});
}

#------------------------------------------------------------
1;


