package IW::ObjectList;

=head1 NAME

IW::ObjectList - abstract class for outputting a dataset/granule list


=head1 OUTPUT FORMAT

New default ultra-compact format is entire data set or granule list is
one table, with one line for each item.

=over

=item flavor

A short piece of text that says what kind of object this is.  
It is may be supplied via a small GIF.

=item number

The 1-based number in the list.

=item note

Any optional information significant enough that it be presented here.

=item title

The title of this object. 

=item data

The object data itself may be summarized across I<r> rows 
and I<c> columns, where I<r> and I<c> must be known prior to doing
any printing.

=back


=head1 NOTES

=cut

use IU;
use IU::Log;
use HTML::Stream qw(:funcs);
use IW::App;
use IW::Disclaimers;
use IW::Util qw(:url);

use strict;

#------------------------------------------------------------
# new PARAMHASH
#------------------------------------------------------------
# Create a new object-list object.
#
#     Elems     => ref to array of elements
#     NoteFunc  => optional: called to output notation

sub new {
   my ($class, %params) = @_;
   my $self = bless \%params, $class;
   $self->{HasSummaries} = 0;
   $self->{Elems}        or $self->{Elems} = [];

   # Get display preferences:
   $self->{Display} = eval{$SESSION->state->{User}->display}
                      || new IW::UserDisplay;

   
   # length of a page.
   my $display = $self->{Display};
   my $pagesize = $display->list_length;
   $self->{PageSize} = $pagesize;

   # how many pages
   my $num = int(@{$self->{Elems}});
   my $num_of_pages = $num/$pagesize;
   $self->{NumOfPages} = $num_of_pages;

   # which page
   my $page = $self->{Page} or 1;  # default to first page.
   # If stored page number is larger than the number of pages for this
   # display, then reset page to 1.
   $page = 1 if ($self->{Page} >= ($num_of_pages+1)); 
   $self->{Page} = $page;

   # first item of the page.
   $self->{StartItem} = ($pagesize*($page-1)) + 1;
   
   # last item of the page.
   $self->{EndItem}  = ( ($pagesize*$page) < $num)
                  ? $pagesize*$page : $num;
   $self;
}

#------------------------------------------------------------
# update_virtual_keys
#------------------------------------------------------------
# Class method

sub update_virtual_keys {
   my ($class, $type, $database) = @_;

   my @new_keys = $REQ->param($type); 
   my @new_selected_keys;
   my $select_flag = $REQ->param('select') || ''; 
   my $selected_on = $REQ->param('selected_on') || '';

   if ($select_flag eq 'NONE') {
      # Set to no items selected. 
      @new_selected_keys = ();

   } elsif ($select_flag eq 'ALL') {
      # Set to all items selected.
      @new_selected_keys = $database->list_keys($type);
   } else {
      # Merge selections/deselections from the page with 
      # previous list.
      @new_selected_keys = merge_selections($type,$database,\@new_keys);
   }

   $database->selected_list_keys($type,\@new_selected_keys);
   if ($selected_on eq 'items selected on all pages') {
      return @new_selected_keys;
   } else {
      return @new_keys;
   }
   1;
}

#------------------------------------------------------------
# merge_selections
#------------------------------------------------------------
# internal function

sub merge_selections {

    my ($type, $database, $new_keys) = @_; 
    my @new_selected_keys = ();
    # Get list of items that were on the page.
    my %on_page = map {$_ => 1} $REQ->param('keys');
    # Remove (unselect) all items that were on the page.
    my $selected_key;
    my @selected_list_keys = ();
    @selected_list_keys = $database->selected_list_keys($type);
    foreach $selected_key (@selected_list_keys) {
       push @new_selected_keys, $selected_key if (! $on_page{$selected_key});
    }
    # now add back new selections.
    my $key;
    foreach $key (@$new_keys) {
       push @new_selected_keys, $key;
    }
    return @new_selected_keys;
}

#------------------------------------------------------------
# has_summaries
#------------------------------------------------------------
# Instance method
sub has_summaries {
  my ($self,$arg) = @_;
  $self->{HasSummaries} = $arg if defined $arg;
  $self->{HasSummaries};
}

#------------------------------------------------------------
# granules_to_display 
#------------------------------------------------------------
# object method.
# return list of granules to be displayed on the coverage time/map.

sub granules_to_display {

   my $self = shift;
   my $select_flag = $REQ->param('select') || '';
   my $selected_on = $REQ->param('selected_on') || '';

   my $start_item = $self->{StartItem};
   my $end_item = $self->{EndItem};
   my @granules_to_display = ();

   if ($select_flag eq 'NONE') {
      # Set to no items selected, so none to display. 
      @granules_to_display = ();

   } elsif ($select_flag eq 'ALL') {
      if ($selected_on eq 'items selected on all pages') {
         # Display all items on all pages. 
         @granules_to_display = @{$self->{Elems}}; 
      } else {
         # Display all items on current page only.
         @granules_to_display = @{$self->{Elems}}[$start_item-1..$end_item-1];
      }
   } else {
      my @granules;
      if ($selected_on eq 'items selected on all pages') {
         # Display all items on all pages.
         @granules = @{$self->{Elems}};
      } else {
         # Display all items on current page only.
         @granules = @{$self->{Elems}}[$start_item-1..$end_item-1];
      }
      # filter out unselected items.
      my %selections = map {$_ => 1} @{$self->{Selections}};
      foreach my $granule (@granules) {
         push @granules_to_display, $granule if ($selections{$granule->key});
      }
   }
   return @granules_to_display;
}

#------------------------------------------------------------
# add_virtual_ckeys
#------------------------------------------------------------

sub add_virtual_ckeys {
   my ($class, $odatabase, $odatabaselist, $database) = @_;

   my @gkeys;

   if (!defined($odatabase) || !defined($odatabaselist)) {
      puke 533, "The database you are trying to add from no longer exists.";
   }
   my $addfrom = $REQ->param('addfrom') || 'GRAN';
   if ($addfrom eq 'DSET') {
      my @dskeys = update_virtual_keys('class','dskey',$odatabaselist);
      puke 513, "You must select one or more data sets." unless (@dskeys);
      @gkeys = $odatabase->gkeys_for_datasets(\@dskeys);
   } else {
      @gkeys = update_virtual_keys('class','gkey',$odatabaselist);
      puke 513, "You must select one or more granules." unless (@gkeys);
   }
   # Test if total number of granules being added plus what is currently
   # in the list exceed the list limit.
   my @db_ckeys = $database->list_keys('ckey');
   my %unique_keys_hash = map {$_ => 1} @db_ckeys,@gkeys;
   my @unique_keys_array = keys %unique_keys_hash;
   my $list_max = ${IU::ListMax} ||  1000;
   if (int(@unique_keys_array) > $list_max) {
      if ($addfrom eq 'DSET') {
         puke 513, "Exceeded $list_max maximum item limit for the shopping ",
                   "cart, please select fewer data sets from which to add ",
                   "granules or remove some items from the shopping cart.";
      } else {
         puke 513, "Exceeded $list_max maximum item limit for the shopping ",
                   "cart, please select fewer granules to add or remove some ",
                   "items from the shopping cart.";
      }
   }
  
   $database->list_keys('addckey',\@gkeys);
   return @gkeys;
}
#------------------------------------------------------------
# cont_to_add_virtual_ckeys
#------------------------------------------------------------

sub cont_to_add_virtual_ckeys {

   my ($class,$database) = @_;
   my @ckeys =  $database->list_keys('addckey');
   my @db_ckeys = $database->list_keys('ckey');
   my %unique_keys_hash = map {$_ => 1} @db_ckeys,@ckeys;
   my @unique_keys_array = keys %unique_keys_hash;
   $database->list_keys('ckey',\@unique_keys_array);
   return @ckeys;
}


#------------------------------------------------------------
# add_virtual_gkeys
#------------------------------------------------------------

sub add_virtual_gkeys {
   my ($class, $odatabase, $odatabaselist, $database) = @_;

   if (!defined($odatabase) || !defined($odatabaselist)) {
      puke 533, "The database you are trying to add from no longer exists.";
   }

   my @gkeys;
   my $addfrom = $REQ->param('addfrom');
   if ($addfrom eq 'DSET') {
      my @dskeys = update_virtual_keys('class','dskey',$odatabaselist);
      puke 513, "You must select one or more data sets." unless (@dskeys);
      @gkeys = $odatabase->gkeys_for_datasets(\@dskeys);
   } else {
      @gkeys = update_virtual_keys('class','gkey',$odatabaselist); 
   }
   my @db_gkeys = $database->list_keys('gkey'); 
   my %unique_keys_hash = map {$_ => 1} @db_gkeys,@gkeys;
   my @unique_keys_array = keys %unique_keys_hash;
   my $list_max = ${IU::ListMax} ||  1000;
   if (int(@unique_keys_array) > $list_max) {
      if ($addfrom eq 'DSET') {
         puke 513, "Exceeded $list_max maximum data granule limit, please ",
                   "select fewer data sets from which to add granules.";
      } else {
         puke 513, "Exceeded $list_max maximum data granule limit, please ",
                   "select fewer granules to add.";
      } 
   }
   $database->list_keys('gkey',\@unique_keys_array);
   return @gkeys;
}

#------------------------------------------------------------
# remove_virtual_keys
#------------------------------------------------------------

sub remove_virtual_keys {
    my ($class, $type, $database) = @_;

    my $selected_on = $REQ->param('selected_on') || '';
    my @keys = update_virtual_keys('class',$type,$database);

    my $key;
    my %remove_keys = map {$_ => 1} @keys;
    my @keep_keys_array = ();
    foreach $key ($database->list_keys($type)) { 
       # add to list of keys to keep if not in the remove list.
       push @keep_keys_array, $key unless ($remove_keys{$key});
    }
    $database->list_keys($type,\@keep_keys_array);

    my @selected_keys_array = ();
    foreach $key ($database->selected_list_keys($type)) {
       # add to list of selected keys if not in the remove list.
       push @selected_keys_array, $key unless ($remove_keys{$key});
    }
    $database->selected_list_keys($type,\@selected_keys_array);
    @keys;
}


#------------------------------------------------------------
# elems
#------------------------------------------------------------
# Return a reference to the list of granules.

sub elems {
    $_[0]->{Elems};
}

#------------------------------------------------------------
# origin
#------------------------------------------------------------
# Return the origin of these granules.

sub origin {
    $_[0]->{Origin};
}

#------------------------------------------------------------
# sort [SORTREF]
#------------------------------------------------------------
# Sort our list of elements.  Returns a reference to the newly-sorted array.
# A description of what was done is placed in $$SORTREF. 
#
# I<This is an abstract method:> you must override it in your subclasses.

sub sort {
    my ($self, $sortref) = @_;
    die "abstract method: sort($self)";
}

#------------------------------------------------------------
#
# PRIVATE METHODS/UTILITIES
#
#------------------------------------------------------------


#------------------------------------------------------------
#
# PRINT...
#
#------------------------------------------------------------


#------------------------------------------------------------
# print_note I
#------------------------------------------------------------
# Output the note for the i'th (0-based) element.

sub print_note {
    my ($self, $i) = @_;
    $self->{NoteFunc} and &{$self->{NoteFunc}}($self->{Elems}[$i]);
}

#------------------------------------------------------------
# print_note2 I
#------------------------------------------------------------
# Output the note for the i'th (0-based) element.

sub print_note2 {
    my ($self, $i) = @_;
    $self->{Note2Func} and &{$self->{Note2Func}}($self->{Elems}[$i]);
}

#------------------------------------------------------------
# print_graph I
#------------------------------------------------------------
# Output the note for the i'th (0-based) element.

sub print_graph {
    my ($self, $i) = @_;
    $self->{GraphFunc} and &{$self->{GraphFunc}}($self->{Elems}[$i]);
}

#------------------------------------------------------------
# print_graph2 I
#------------------------------------------------------------
# Output the note for the i'th (0-based) element.

sub print_graph2 {
    my ($self, $i) = @_;
    $self->{Graph2Func} and &{$self->{Graph2Func}}($self->{Elems}[$i]);
}


#------------------------------------------------------------
# print_title I
#------------------------------------------------------------
# Output the note for the i'th (0-based) element.
# I<This is an abstract method:> you must override it in your subclasses.

sub print_title {
    my ($self, $i) = @_;
    die "abstract method: print_title($self)";
}

#------------------------------------------------------------
# print_endtitle I
#------------------------------------------------------------
# Output the last title for the i'th (0-based) element.
# If granule list then final column == data set id.
# This default will do nothing unless overridden in subclass

sub print_endtitle {
    1;
}

#------------------------------------------------------------
# pass_params
#------------------------------------------------------------
# Pass dataset/granule keys from the last request as hidden fields.
# This is invoked automatically on output!

sub pass_params {
    my $self = shift;
    # Abstract method is a no-op:
    1;
}

#------------------------------------------------------------
# print_buttons I
#------------------------------------------------------------
# Output the menu buttons for the i'th (0-based) item.

sub print_buttons {
    my ($self, $i, $pkgs_chosen) = @_;
    die "abstract method: print_buttons($self)";
}

#------------------------------------------------------------
# print_head
#------------------------------------------------------------
# Output the column headers for the 1 line per item table.

sub print_head {
    my $self = shift;
    die "abstract method: print_head($self)";
}


#------------------------------------------------------------
# any_comments
#------------------------------------------------------------
# Abstract. Develop in GranList or DatasetList.

sub any_comments {
    0;
}

#------------------------------------------------------------
# any_notes
#------------------------------------------------------------
# Abstract. Develop in GranList or DatasetList.

sub any_notes {
    1;
}

#------------------------------------------------------------
# any_orderable
#------------------------------------------------------------
# Abstract. Develop in GranList or DatasetList.

sub any_orderable {
    1;
}


#------------------------------------------------------------
# any_pkgs_chosen
#------------------------------------------------------------
# Return true if there are any pkgs chosen in the whole table of all items.
# Return false if there are no packages chosen yet in the whole table.

sub any_pkgs_chosen {
    my $self = shift;
    my $num = shift;
    my $i;
    my $item;

    for ($i = 0; $i < $num; $i++) {
	# Get the individual element.
	$item = $self->{Elems}[$i];
	# If this data set has pkgs selected, return true.
	if ($item->order_status eq 'GO_PKGS' ||
            $item->order_status eq 'GO_EMAIL_PKGS') {
	    return (1);
	}
    }
    return (0);
}

#------------------------------------------------------------
# print_as_text
#------------------------------------------------------------
# print the listing in a text only format without any HTML.

sub print_as_text {
   my $self = shift;

   my %some = $self->check_for_some_txt();
   # Send the HTTP-reply header:
   my $head = new IW::ReplyHead;
   $head->okay if $REPLY->need_status_line;
   $head->content_type('text/plain');
   $head->send(\*STDOUT);

   # Print the column headers for the table:
   $self->print_head_txt(\%some);
   print "\n";
   # Print the column data for the table:
   my $num = int(@{$self->{Elems}});
   my $i=0;
   for ($i=0; $i < $num; $i++ ) { 
      $self->print_data_txt(\%some,$i);
      print "\n";
   }
   # Done!
   $REPLY->all_done;
}

#------------------------------------------------------------
# print OPTIONS
#------------------------------------------------------------
# Print the listing.  OPTIONS is a hash, which may contain...
#
#     NoRule  => suppress the leading horizontal rule

sub print {
   my ($self, %opts) = @_;
   my $i = 0;                    # Counter of items. 0 based.
   # Number of items.
   my $num = int(@{$self->{Elems}});

   # clear keys so we don't have any carryover of selections from
   # previous screen.
   $REQ->delete($self->{Type});
   # Set keys as param tags so checkboxes will set selections.
   $REQ->param($self->{Type}, @{$self->{Selections}});

   my $mode = $REQ->param('mode');
   my $submode = $REQ->param('submode') || 'LIST';
   $submode = 'LIST' unless ($submode eq 'MAP' or $submode eq 'TIME');

   my $buttons = 7;  # number of buttons (2x) to show on the page menu.
   my $page         = $self->{Page};
   my $pagesize     = $self->{PageSize};
   my $start_num    = $self->{StartItem};
   my $end_num      = $self->{EndItem};
   my $num_of_pages = $self->{NumOfPages};

   my $end_page   = ($buttons*2+2);
   my $start_page = ((($page-$buttons) < 1) || ($num_of_pages < $buttons*2+2))
                     ? 1 : ($page-$buttons);

   if ($num_of_pages < ($buttons*2+2)) {
      $end_page = $num_of_pages;
   } elsif ( $page+$buttons > ($buttons*2+2)) {
      $end_page   = ($num_of_pages < ($page+$buttons) )
      ? $num_of_pages : $page+$buttons;
   }

   my $title;
   my $pstring;
   my $some_notes;        # True if any item in whole table has a note.
   my $some_pkgs_chosen;  # True if any item has packages chosen.
   my $some_orderable;    # True if any DATA SET is orderable as whole.
   my $some_comments;     # True if any DATA SET has a comment.
   my $item;              # The granule or data set.

   # Print customize and text only buttons at the top of the table. 
   print '<small>';
   $self->print_customize_button();
   $self->print_report_button();
   print "<TABLE BORDER=0 CELLSPACING=3 CELLPADDING=3>\n";
   print "<BR>Summary documents will open in a new window." 
       if $self->has_summaries;
   print '</small>'; 

   # Set a flag if there are any notes in the table:
   $some_notes = $self->any_notes($num);

   # Set a flag if any packages were chosen:
   $some_pkgs_chosen = $self->any_pkgs_chosen($num);

   # Set a flag if any data set is wholly orderable.  (false for grans).
   $some_orderable = $self->any_orderable($num);

   # Set a flag if any data set has comments. (false for grans).
   $some_comments = $self->any_comments($num);

   my %some = $self->check_for_some();

   # Print the column headers for the table:
   $self->print_head ($some_notes, $some_pkgs_chosen, 
                      $some_orderable, $some_comments, \%some);

   # Print elements in rows.
   $self->print_elems($start_num, $end_num, $some_notes, $some_pkgs_chosen,
                      $some_orderable, $some_comments, \%some);

   print "</table>\n";

   if ($num_of_pages > 1) {

      my $last = $end_num;
      $last = $num if ($last > $num);

      $HTML -> t("You are currently viewing items $start_num through $last ",
                 "out of $num total."); 
  
      $HTML->TABLE(BORDER=>0);
      $HTML->TR;
      if ($start_page > 1) {
         $HTML->TD;
         $HTML-> msubmit_button(QUERY => ['submode' =>$submode,
                                          'mode'    =>$mode,
                                          'nextpage'=>$start_page-1],
                                CAPTION => " << ");
         $HTML->_TD;
      }

      for ($i = $start_page; $i < ($end_page + 1); $i++) {
         my $start_item = ($i == '1') ? '1' : (($i-1)*$pagesize)+1;
         my $end_item = ($i > $end_page) ? $num : (($i-1)*$pagesize + $pagesize);
         $HTML->TD;
         $HTML->B->FONT(SIZE=>'+2') if ($i eq $page);
         $HTML-> msubmit_button(QUERY => ['submode' =>$submode,
                                          'mode'    =>$mode,
                                          'nextpage'=>"$i"],
                                CAPTION => " $start_item - $end_item ");
         $HTML->_B->_FONT if ($i eq $page);
         $HTML->_TD;
      }
      if ($end_page < $num_of_pages) {
         $HTML->TD;
         $HTML-> msubmit_button(QUERY => ['submode' =>$submode,
                                          'mode'    =>$mode,
                                          'nextpage'=>$end_page+1],
                                CAPTION => " >> ");
         $HTML->_TD;
      }
      $HTML->_TR;
      $HTML->_TABLE;
   }

   1;
}

#------------------------------------------------------------
# print_report_button();
#------------------------------------------------------------
# method - override in child class with noop method if not wanted.

sub print_report_button {
    $HTML -> FONT(SIZE=>"+1");
    $HTML -> hypertext('Text-only version',
                       script_url('mode' =>$REQ->param('mode'),'report'=>'Y'),
                       TARGET => 'support')
          -> t(' - for printing or import into a spread sheet.'); 
    $HTML -> _FONT;
}


#------------------------------------------------------------
# print_elems();
#------------------------------------------------------------
# method -  override in child class if wanted different.

sub print_elems {
   my ($self, $start_num, $end_num, $some_notes, $some_pkgs_chosen, 
        $some_orderable, $some_comments, $some) = @_;

   my $i = 0;                 # Counter of items. 0 based.
   for ($i = $start_num-1; $i < ($end_num); $i++) {
      print '<tr bgcolor="#EEEEEE" align=left>';

      # Print the menu buttons in table cells:
      $self->print_buttons($i, $some_notes, $some_pkgs_chosen,
                           $some_orderable, $some_comments);

      $self->print_data($some, $i);

      print "</tr>\n";
      # put listed items as hidden tags so we can figure out what the users
      # selected or unselected.
      $HTML->hidden('keys',$self->{Elems}[$i]->key);
   }
}

#------------------------------------------------------------
# print_customize_button();
#------------------------------------------------------------
# method -  override in child class with noop method if not wanted.

sub print_customize_button {

   my $self = shift;
   
   my $url = script_url('mode'=>'CUSTOMIZE',
                        'type'=>$self->{Type},
                        'prevmode'=>$REQ->param('mode'));
   $HTML -> FONT(SIZE=>"+1");
   $HTML-> A(HREF=>$url)
        -> t('Customize this table')
        -> _A
        -> t(' - add additional information, change columns/number of rows, sort                                       order, etc.');
   $HTML -> _FONT;
   print '<BR>';
   print "\xA0";
}

#------------------------------------------------------------
# print_all_none
#------------------------------------------------------------
# Print the all/none selection buttons are shown

sub print_all_none {

   my $mode = $REQ->param('mode');
   my $submode = $REQ->param('submode') || 'LIST';
   $submode = 'LIST' unless ($submode eq 'MAP' or $submode eq 'TIME');

   $HTML ->FONT(SIZE=>'-1');
   $HTML -> msubmit_button(QUERY =>['mode'=>$mode,
                                   'submode'=>$submode,
                                   'select'=>'ALL'],
                          CAPTION =>'All',
                          ALT   =>'Select All');
   $HTML -> msubmit_button(QUERY =>['mode'=>$mode,
                                   'submode'=>$submode,
                                   'select'=>'NONE'],
                          CAPTION =>'None',
                          ALT   =>'Select None');
   $HTML ->_FONT;
   1;
}

#------------------------------------------------------------
1;
