package IW::Order;

=head1 NAME

IW::Order - a user's order

=head1 DESCRIPTION

This abstract class represents a collection of items in the shopping cart.
The real work is done by the subclasses.

=head1 SYNOPSIS

This abstract class defines the following interface:

    # Make orders from granules/datasets (in shopping cart):
    $order = make IW::Order Granules=>\@granules,
                            Datasets=>\@datasets;

    # Associate this order with user information:
    $order->user($user);
    $user = $order->user;
     
    # Get/set the order id:
    $order->id($id);
    $id = $order->id;
    
    # Get the total cost:
    $cost = $order->cost;
    
    # Return a sorted list of archive ids we're ordering from:
    my @archids = $order->archive_ids;

    # Return the number of ECS archive ids we're ordering from:
    my @ecsdaacs = $order->ecs_archive_ids;

The following are not defined here, but should be:

    # Add an item to the order, in an appropriate fashion
    $order->add_item($order_item);

    # Return all items in the order:
    @items = $order->all_items;

=head1 AUTHOR

Eryq, Hughes STX Corporation

=head1 RCS INFO

$Id: Order.pm,v 1.55 2003/01/21 20:04:53 seufert Exp $

=cut

use IU;
use IU::Addr;         # Needed for back compatibility for DaacAddress
use IU::HandlingFee;
use IU::MediaHandlingFee;  
use IU::InvStore ();
use IU::Log;
use IU::PkgCombo;

use IW::App;
use IW::User;           # DaacAddress included here too.

use strict;

#------------------------------------------------------------
# new
#------------------------------------------------------------
sub new {
    my $self = bless {}, shift;
    $self->{Archives} = {};   # combos by archive, for package ordering
    $self;
}

#------------------------------
#
# Other attributes...
#
#------------------------------

#------------------------------------------------------------
# user [VALUE]
#------------------------------------------------------------
# Get/set the user-info object.

sub user {
    my ($self, $user) = @_;
    @_ > 1 and $self->{User} = $user;
    $self->{User};
}

#------------------------------------------------------------
# id [VALUE]
#------------------------------------------------------------
# Get/set the request id.

sub id {
    my ($self, $id) = @_;
    @_ > 1 and $self->{Id} = $id;
    $self->{Id};
}

#------------------------------------------------------------
# archive_ids
#------------------------------------------------------------
# Return a sorted list of archive ids to order from.
# In a scalar context, returns the number of archive ids.

sub archive_ids {
    my ($self) = @_;
    wantarray ? (sort keys %{$self->{Archives}}) 
	: int(keys %{$self->{Archives}});
}

#------------------------------------------------------------
# ecs_archive_ids
#------------------------------------------------------------
# Return the number of ecs archive ids we're ordering from.

sub ecs_archive_ids {

    my $self = shift;
    my @initialDaacs = $self->archive_ids;

    # Open static valids database:
    my %WWWVALIDS;
    if (-e "${IU::LibDir}/conf/wwwvalids.gdb") {
        eval {tie %WWWVALIDS, 'IU::Store', "${IU::LibDir}/conf/wwwvalids.gdb",
              'READER'};
        $@ and return error "tie failed: $@";
    }

    # For each data center we're ordering from...
    my @needs;        # Daacs needing authenticator.
    my $archid;
    foreach $archid (@initialDaacs) {
        # Is there an entry for it in the static valids?
        my $valids;
        if ($valids = $WWWVALIDS{"daac:$archid"}) {
            push(@needs, $archid)   if ($valids->{REQUIRES_ECS_AUTH});
        }
    }                                    # end: foreach daac

    # Close static valids database:
    untie %WWWVALIDS;

    return @needs;
}

#------------------------------------------------------------
# cost
#------------------------------------------------------------
sub cost {
    my $self = shift;

    die "unimplemented abstract method";
}

#------------------------------------------------------------
# make PARAMHASH
#------------------------------------------------------------
# I<Class method.>
# Build orders from a collection of granules/datasets.
#
#     Granules   => ref to array of granule objects
#     Datasets   => ref to array of dataset objects
#     SkipIncomplete => if granule has no packages, skip silently
#
# Returns two orders, (REQUESTS, EMAIL)

sub make {
    my ($class, %params) = @_;
    my $granules  = $params{Granules};
    my $datasets  = $params{Datasets};
    my $just_datasets = $params{Just_Datasets};

    # Create new order objects:
    my $requested = IW::Order::Request->new;
    my $emailed   = IW::Order::Email->new;
    my $order_object;  # Handle to $requested or $emailed depending on granule.

    # Hash datasets by key:
    my %dataset_hash = map {($_->key)=>$_} @$datasets;

    # Assemble the list of merged package combos
    my $gran;
    my $dset;

    foreach $gran (@$granules) {
	### debug "order/make: adding granule ", $gran->key;

	# Get the dataset for this granule:
	$dset = $dataset_hash{$gran->dataset_key};

	# What is granule's order status?
	my $status = $gran->order_status;

	# If the granule has packages selected:
	if ($status eq 'GO_PKGS' || $status eq 'GO_EMAIL_PKGS') { 

	    # Set order object based on whether ordering by e-mail or not.
	    $order_object = ($status eq 'GO_PKGS') ? $requested : $emailed;

	    # Add each package combo:
	    my $combo;
	    foreach $combo (@{$gran->package_combos}) {
		### debug "order/make: adding combo";
                my $tempmediafee;
                my $mediahandlingfee;
                foreach $tempmediafee (@{$dset->mediahandlingfee}) {
                   if ($combo->media_type eq $tempmediafee->type) {
                      $mediahandlingfee = $tempmediafee;
                      last;
                   }
                }
		# repeat the line item for the number of 
		# copies requested by the user for that combo.
		my $copies_req = $combo->copies;
		my $i;
		for ($i = 0; $i < $copies_req; $i++) {

		   # Create a line item for this combo:
		   my $order_item = 
		      IW::Order::Item->new(Orderable =>$gran,
				       PkgCombo  =>$combo,
				       AcctName  =>$dset->chosen_account,
				       HandlingFee =>$dset->handling_fee,
                                       MediaHandlingFee =>$mediahandlingfee);
		
		   ### debug "adding line item for the $i time";
		   # Add it to the order (identical merged combos are merged):
		   $order_object->add_item($order_item, $i);
		}
	    }
	}
	elsif ($status eq 'GO_EMAIL')  { 

	    # Create a fake package combo (KLUDGE).
	    #    Package id will be "*", 1 granule in size, no other opts.
	    my $fake_combo = IU::PkgCombo->new_fake_for_email;
	    $fake_combo->granule_key($gran->key);           # tie
	   
            # Media handling fee does not apply to email orders
            my $mediahandlingfee = undef; 

	    # Create a line item:
            my $order_item = 
	      IW::Order::Item->new(Orderable =>$gran,
				   PkgCombo  =>$fake_combo,
				   AcctName  =>$dset->chosen_account,
				   HandlingFee =>$dset->handling_fee,
                                   MediaHandlingFee =>$mediahandlingfee);

	    # Add it to the order:
	    $emailed->add_item($order_item);

            # Set the flag indicating at least one order has unknown cost.
            $emailed->unknown_cost('yes');
	}
	elsif ($status eq 'NEED_PKGS') { 
	    $requested->{NeedPkgs} = 1;
	    puke 532, "you need to select ordering options for some of the items!"
		unless ($params{SkipIncomplete});
	}
	else { 
	    warn "skipping unorderable entity (status = $status)!";
	    next;
	}
    }

    foreach $dset (@$just_datasets) {
	### debug "order/make: adding data set ", $dset->key;
	
	
	# Readiness:
	my $status = $dset->order_status;
	if ($status eq 'GO_PKGS') { 
	    
	    # Add each package combo:
	    my $combo;
            my $mediahandlingfee = undef;
	    foreach $combo (@{$dset->package_combos}) {
                my $tempmediafee;
                foreach $tempmediafee (@{$dset->mediahandlingfee}) {
                   if ($combo->media_type eq $tempmediafee->type) {
                      $mediahandlingfee = $tempmediafee;
                      last;
                   }
                }
                my $copies_req = $combo->copies;
                my $i;
                for ($i = 0; $i < $copies_req; $i++) {

		    # Create a line item for this combo:
		    my $order_item = 
		      IW::Order::Item->new(Orderable =>$dset,
				       PkgCombo  =>$combo,
				       AcctName  =>$dset->chosen_account,
				       HandlingFee => $dset->handling_fee,
		                       MediaHandlingFee => $mediahandlingfee);	
		    # Add it to the order (identical merged combos are merged):
		    $requested->add_item($order_item,$i); 
               }
	    }
	}
# Not allowing email ordering for data sets now
#	elsif ($status eq 'GO_EMAIL')  { 
#
#	    # Create a fake package combo (KLUDGE).
#	    #    Package id will be "*", 1 granule in size, no other opts.
#	    my $fake_combo = IU::PkgCombo->new_fake_for_email;
#	    $fake_combo->granule_key($dset->key);           # tie
#	    
#	    # Create a line item:
#	    my $order_item = 
#	      IW::Order::Item->new(Orderable =>$dset,
#				   PkgCombo  =>$fake_combo,
#				   AcctName  =>$dset->chosen_account,
#				   AddnInfo  =>$addn_info);
#	    
#	    # Add it to the order:
#	    $emailed->add_item($order_item); 
#
#	}
	elsif ($status eq 'NEED_PKGS') { 
	    $requested->{NeedPkgs} = 1;
	    puke 532, "you need to select ordering options for some of the items!"
		unless ($params{SkipIncomplete});
	}
	else { 
	    warn "skipping unorderable entity (status = $status)!";
	    next;
	}
    }

    # Done! 
    ($requested, $emailed);
}

#------------------------------------------------------------
# need_packages
#------------------------------------------------------------
# Do we need to assign packages to some of these items before
# an order can be done?
sub need_packages {
    shift->{NeedPkgs};
}

#------------------------------------------------------------
# print_receipt_msg FILEHANDLE,MESSAGE
#------------------------------------------------------------
# I<Instance method, private.>
# Print a message

sub print_receipt_msg {
    my ($self, $FH, @msg) = @_;
    my $ind = " " x 5;
    @msg = split /\n/, join('', @msg);   
    foreach (@msg) {	
	print $FH $ind, "###  ", $_, "\n";
    }
    print $FH "\n";
}

#------------------------------------------------------------
# print_receipt FILEHANDLE, PARAMHASH
#------------------------------------------------------------
# I<Class method.>
# Output a textual human-readable receipt.  PARAMHASH contains:
#
#      User      => User info.  If defined, oit will be printed; else, it'll 
#                   be skipped.
#      RequestId => request id; else, it'll be skipped
#      Requested => an IW::Order::Request for an IMS product request
#      Emailed   => an IW::Order::Request for an emailed request

sub print_receipt {
    my ($self, $FH, %params) = @_;
    my $user      = $params{User};
    my $reqid     = $params{RequestId};
    my $requested = $params{Requested};
    my $emailed   = $params{Emailed};
    my $totalCost = 0;
    my %handling_fee_hash;
    my %media_fee_hash; 
    # Tracking information:

   if ($reqid) {
        print $FH "TRACKING INFO\n";
	print $FH "     Order tracking no:     $reqid\n";
	print $FH "     Time submitted:        ", scalar(localtime), "\n";
    }

    print $FH "\n";

    # User info:
    if ($user) {
       $user->print ('FORM'=>'RECEIPT', 'FILEHANDLE'=>$FH);
    }
    else {
	puke "No user information in print_receipt!";
    }
    # Orders...
    if (!$requested) {                     # nothing requested
	$self->print_receipt_msg($FH,
	       "NO PRODUCT-REQUEST ORDER INFORMATION GIVEN");
    }
    else {                                 # something requested...
	# Do we need packages?
	if ($requested->need_packages) {  
	    $self->print_receipt_msg($FH,
		    "YOU NEED TO ASSIGN PACKAGES TO SOME OF YOUR\n",
		    "DATA GRANULES BEFORE THEY CAN BE ADDED TO THE ORDER!");
	}
	# Output the actual items:
	if (int($requested->archive_ids)) {    # we've got SOMEthing at least!
	    $totalCost += $requested->cost;
            %handling_fee_hash = %{$requested->handling_fee};
            %media_fee_hash = %{$requested->media_handling_fees};
	    $requested->print_receipt_data($FH);
	}
    }
    # Orders by email...    
    if (!$emailed) {                       # nothing requested
	$self->print_receipt_msg($FH,
	       "NO EMAIL ORDER INFORMATION GIVEN");
    }
    else {                                 # something requested...
	# Output the actual items:
	if (int($emailed->dataset_ids)) {      # we've got SOMEthing at least!
	    $totalCost += $emailed->cost;
	    $emailed->print_receipt_data($FH);
	}
    }

    #
    # Print subtotal for line items in the order.
    #

    print $FH ('-' x 60), "\n";
    printf $FH "%-62s \$%.2f\n", "Subtotal of per item costs (US):", $totalCost;

    #
    # Media Handling Fee:
    #

    my $totalMediaHandlingFee = 0;
    my @media_fee_keys = keys %media_fee_hash;

    if ( int(@media_fee_keys) >0 ) {
       print $FH ('-' x 60), "\n", "Media Handling Fee  (US) :", "\n";

       # Display the Media Handling Fee(s)/ their Prices.
       my $single_fee_key;
       foreach $single_fee_key (@media_fee_keys) {
         printf $FH ("  %-60s \$%.02f\n", $single_fee_key, 
                     $media_fee_hash{$single_fee_key});
         $totalMediaHandlingFee += $media_fee_hash{$single_fee_key};
       }
    }

    # Add Media Handling Fee amount to total Cost of the user's order.
    $totalCost += $totalMediaHandlingFee;

    # 
    # Dataset Handling Fee
    #

    my $totalHandlingFee = 0;
    my @fee_keys = keys %handling_fee_hash;

    if ( int(@fee_keys) >0 ) {
       print $FH ('-' x 60), "\n", "Handling Fee  (US) :", "\n";

       # Display the Handling Fee(s)/ their Prices.
       my $single_fee_key;
       foreach $single_fee_key (@fee_keys) {
         printf $FH ("  %-60s \$%.02f\n", $single_fee_key, $handling_fee_hash{$single_fee_key});
         $totalHandlingFee += $handling_fee_hash{$single_fee_key};
       }
    }

    # Add ANY Handling Fee amount to total Cost of the user's order.
    $totalCost += $totalHandlingFee; 

    # Bottom line:    
    print $FH ('-' x 60), "\n";
    printf $FH "%-62s \$%.2f\n", "TOTAL KNOWN COST (US):", $totalCost;

    # Disclaimer:
    print $FH <<EOF;

Please note that this is the MINIMUM ESTIMATED COST for your order.  
If cost information was unavailable for some of the products you requested 
(as for some e-mail orders), OR if there are per-order shipping charges, 
THE ACTUAL COST MAY BE HIGHER.
EOF
    print $FH ('-' x 60), "\n";
    print $FH "End of receipt.\n";
}
#------------------------------------------------------------
# print_line_items FILEHANDLE ARRAYREF
#------------------------------------------------------------
# Output order option line items.
# Parameters: 1. filehandle to open file
#             2. reference to array of items.

sub print_line_items {
   my ($self, $FH, $item_arrayref) = @_;

   # Output the line items:
   my $count = 0;             # tally
   my $order_item;
   foreach $order_item (@$item_arrayref) {
      my $combo = $order_item->package_combo;
      my $dsid  = $order_item->dataset_id;
      my $lgid  = $order_item->orderable->local_granule_ID || undef;
      my $id;
      # Basic options:

      if ($combo->package_id =~ /\*/) {
         $id = $order_item->orderable->id;

         if ($order_item->orderable->order_type eq "GRANULE") {
            printf $FH "Item %03d  Data granule: %s\n", ++$count, $id;
            if (defined $lgid) {
               print $FH  "     Local Granule ID: ", $lgid, "\n";
            }
            print $FH  "     Data set: ", $dsid, "\n";
         }
         else {
            printf $FH "Item %03d  Data set: %s\n", ++$count, $dsid;
         }
      }
      else {
         $id = $combo->line_item_package;
         printf $FH "Item %03d  Package: %s\n", ++$count, $id;   
         print $FH  "     Data set: ", $dsid, "\n"; 
      }      
#      my $title = $combo->package_title;
#      print $FH ": ", $title if (defined ($title));
#      print $FH "\n";
      
      printf $FH  "     Cost:  US \$%.2f\n", $combo->cost;
      print  $FH  "     Format/Media: ", $combo->proc_opt, ": ",
             $combo->media_type, ": ",$combo->media_fmt, "\n";
      print  $FH  "     Additional Information: ",
             $combo->addn_info, "\n" if ($combo->addn_info);

      # Subset/order options:
      my $opts_type;
      foreach $opts_type (sort $combo->options) {
         my $opts = $combo->options($opts_type);
         my $crit;
         printf $FH  "     $opts_type Options:\n";
         foreach $crit ($opts->criteria) {
            # Loop over any nested criteria.
            my $subcrit;
            foreach $subcrit ($crit->subcriteria) {
               my $selected = $subcrit->selections_to_string(Verbose=>1);
               defined($selected) or next;     # no value(s) given
               printf $FH ("        %-12s %s\n",
                        ($subcrit->name.":"), $selected);
            }
            my $selected = $crit->selections_to_string(Verbose=>1);
            defined($selected) or next;     # no value(s) given
            printf $FH ("        %-12s %s\n", ($crit->name.":"), $selected);
         }
      }
      print $FH "\n";
   }
}

#============================================================
#
package IW::Order::Email;
#
#============================================================

=head2 IW::Order::Email

An email order.  This concrete class adds the following functionality:

     # Add an item to the order:
     $order->add_item($order_item);
     
     # Get all items:
     @items = $order->all_items;
         
     # Return a sorted list of dataset ids we're ordering from:
     @dsids = $order->dataset_ids;
      
     # Get the items under a particular dataset id:
     foreach $dsid ($order->dataset_ids) {
	 @items = $order->dataset_items($dsid);
         ...
     }
      
     # Create ODL to mail to a particular dataset's contact:
     foreach $dsid ($order->dataset_ids) {
	 $odl = $order->to_email_ODL($dsid);
     }
     # Return true if any e-mail order items didn't have a
     # known cost in it's packaging. (most don't have packaging).
     $order->unknown_cost;
     # Set the flag indicating at least one order has unknown cost.
     $order->unknown_cost('yes');

=cut

use IU;
use IU::Log;
use IU::PkgCombo;
use IW::App;

@IW::Order::Email::ISA = qw(IW::Order);
use strict;

#------------------------------------------------------------
# new
#------------------------------------------------------------
# Create a new email item.
#
sub new {
    my $self = shift->SUPER::new();
    $self->{EmailByDset} = {};   # combos by dataset, for email ordering
    $self;
}

#------------------------------------------------------------
# add_item ITEM
#------------------------------------------------------------
# Add an item for email order.

sub add_item {
    my ($self, $order_item) = @_;

    # Tally archive:
    $self->{Archives}{$order_item->archive_id} = 1;

    # Get dataset id:
    my $dsetid = $order_item->dataset_id || puke "no data set id!";    

    # Store:
    $self->{EmailByDset}{$dsetid} or $self->{EmailByDset}{$dsetid} = [];
    push @{$self->{EmailByDset}{$dsetid}}, $order_item;
}

#------------------------------------------------------------
# dataset_ids
#------------------------------------------------------------
# Return sorted list of datasets to email-order from.
# In a scalar context, just returns the number of datasets.

sub dataset_ids {
    my ($self) = @_;
    wantarray ? (sort keys %{$self->{EmailByDset}})  
	: int(keys %{$self->{EmailByDset}});
}

#------------------------------------------------------------
# cost
#------------------------------------------------------------
sub cost {
    my $self = shift;
    my ($dsetid, $order_item);
    my $cost = 0.0;

    foreach $dsetid ($self->dataset_ids) {
       foreach $order_item (@{$self->{EmailByDset}{$dsetid}}) {
          $cost += ($order_item->cost || 0);
       }
    }
    $cost = sprintf("%0.2f",$cost);
}

#------------------------------------------------------------
# unknown_cost
#------------------------------------------------------------
# Return true if any e-mail order items didn't have a
# known cost in it's packaging. (most don't have packaging).
# $order->unknown_cost;
# Set the flag indicating at least one order has unknown cost.
# $order->unknown_cost('yes');

sub unknown_cost {
    my $self = shift;
    # If anything TRUE was pased, set the flag.
    if ($_[0]) {
       $self->{CostUnknown} = 1;
    } else {  # Else return current flag if any.
       $self->{CostUnknown};       # may be undef.
    }
}
#------------------------------------------------------------
# dataset_items [DSETID]
#------------------------------------------------------------
# Return emailable line items, in no particular order.
# No dataset id is a fatal error.

sub dataset_items {
    my ($self, $dsetid) = @_;
    defined($dsetid) or puke 600, "undef'ed dset id!";
    @{$self->{EmailByDset}{$dsetid} || []};
}

#------------------------------------------------------------
# all_items
#------------------------------------------------------------
# In array context, returns list of all items.
# In scalar context, returns number of items.

sub all_items {
    my $self = shift;
    
    if (wantarray) {
	return map {$self->dataset_items($_)} $self->dataset_ids;
    }
    else {
	my $total = 0;
	my $dsetid;
	foreach $dsetid ($self->dataset_ids) {	
	    $total += int($self->dataset_items($dsetid));
	}
	return $total;
    }
}

#------------------------------------------------------------
# to_email_ODL DATASETID, [PARAMHASH]
#------------------------------------------------------------
# Convert one particular archive/dataset's request to ODL, for
# an email request.

sub to_email_ODL {
    my ($self, $dsetid, %params) = @_;
    my $no_validate = $params{NoValidate};


    # Begin:
    my $file = IU::ODL->new_group;    
    my $odl = IU::ODL->new_group;
    $file->add_param(PRODUCT_REQUEST => $odl);

    # Get archive id:
    my $any_item = ($self->dataset_items($dsetid))[0];
    my $archid = $any_item->archive_id;

    # Basic stuff:
    $odl->add_param(REQUEST_ID => $self->id);
    $odl->add_param(DATA_CENTER_ID => $archid) if $archid;

    # Add the affiliation and addresses:
    my $user = $self->{User};
    if ($user) {
	$odl->add_param(USER_AFFILIATION => $user->affil->to_ODL)
	    if $user->affil;
	$odl->add_param(CONTACT_ADDRESS  => $user->req_addr->to_ODL)
	    if $user->req_addr;
	$odl->add_param(SHIPPING_ADDRESS => $user->ship_addr->to_ODL)
	    if $user->ship_addr;
	$odl->add_param(BILLING_ADDRESS  => $user->bill_addr->to_ODL)
	    if $user->bill_addr;
    }
    else {
	puke "no user information!" unless $no_validate;
    }

    # The line items:
    my $order_item;
    my @sorted_items = sort { 
	$IW::Order::Email::a->sort_cmp($IW::Order::Email::b) 
	} $self->dataset_items($dsetid);
    foreach $order_item (@sorted_items) {
	$odl->add_param(LINE_ITEM => $order_item->to_ODL);
    }

    $file;   # done!
}

#------------------------------------------------------------
# print_receipt_data FH
#------------------------------------------------------------
# I<Instance method.>
# Print the portion of the receipt for an email order.

sub print_receipt_data {
    my ($self, $FH) = @_;

    my $dsid;
    foreach $dsid ($self->dataset_ids) {
	my $count = 0;                         # tally for each dataset
	
	# Get the sorted array of items:
	my @sorted_items = sort { 
	    $IW::Order::Email::a->sort_cmp($IW::Order::Email::b) 
	    } $self->dataset_items($dsid);
	next if (!@sorted_items);              # what the heck
	
	# Get the contact address and other info, using any item:
	my $archid = $sorted_items[0]->orderable->archive_id || next;
	my $dset   = $sorted_items[0]->orderable->dataset || next;
	my $contact_addr = $dset->contact_addr || next;

	# Header:
	print $FH "-" x 60, "\n";
	print $FH "Ordered by:  ELECTRONIC MAIL TO DATA SET CONTACT\n";
	print $FH "Data center: $archid\n";
	print $FH "Data set:    $dsid\n";
	print $FH "\n";

	# Contact information:
	print $FH "DATA SET CONTACT\n";
        $contact_addr->print_receipt_addr ($FH);
	print $FH "\n";

	# Output the line items:
	$self->print_line_items($FH, \@sorted_items);
    }
    1;
}

#============================================================
#
package IW::Order::Request;
#
#============================================================

=head2 IW::Order::Request

An order that can be handled via the standard IMS product-request
mechanism.  This concrete class adds the following functionality:

     # Add an item to the order:
     $order->add_item($order_item, $copynumber);
     
     # Get all items:
     @items = $order->all_items;
         
     # Get the items under a particular archive:
     foreach $archid ($order->archive_ids) {
	 @items = $order->archive_items($archid);
         ...
     }
      
     # Create ODL to request a particular archive's products:
     $odl = $order->to_request_ODL;

=cut

use IU;
use IU::Log;
use IU::PkgCombo;
use IW::App;

@IW::Order::Request::ISA = qw(IW::Order);
use strict;

#------------------------------------------------------------
# new
#------------------------------------------------------------
# Create a new request item.

sub new {
    my $self = shift->SUPER::new();
    $self->{ReqByArch} = {};   # combos by archive, for package ordering
    $self;
}

#------------------------------------------------------------
# add_item ITEM  COPYNUMBER
#------------------------------------------------------------
# Add an item for a requested order.

sub add_item {
    my ($self, $order_item, $copy) = @_;

    # Get archive id:
    my $archid = $order_item->archive_id || puke "no archive id!";

    # Tally archive:
    $self->{Archives}{$order_item->archive_id} = 1;

    # Get merged key:
    my $key = $order_item->package_combo->merged_key . $copy;
    ### debug "key is $key";
    # Store:
    $self->{ReqByArch}{$archid} or $self->{ReqByArch}{$archid} = {};
    $self->{ReqByArch}{$archid}{$key} = $order_item;
}

#------------------------------------------------------------
# archive_items ARCHID
#------------------------------------------------------------
# Return requestable line items for archive, in no particular order.

sub archive_items {
    my ($self, $archid) = @_;
    defined($archid) or puke 600,"undef'ed arch id!";
    values %{$self->{ReqByArch}{$archid} || {}};
}

#------------------------------------------------------------
# all_items
#------------------------------------------------------------
# In array context, returns list of all items.
# In scalar context, returns number of items.

sub all_items {
    my $self = shift;

    if (wantarray) {
	return map {$self->archive_items($_)} $self->archive_ids;
    }
    else {
	my $total = 0;
	my $archid;
	foreach $archid ($self->archive_ids) {
	    $total += int($self->archive_items($archid));
	}
	return $total;
    }
}

#------------------------------------------------------------
# cost
#------------------------------------------------------------
sub cost {
    my $self = shift;

    # Figure out the total cost of all packages on hand:
    my ($archid, $order_item);
    my $cost = 0.0;

    foreach $archid ($self->archive_ids) {
	foreach $order_item ($self->archive_items($archid)) {
	    $cost += ($order_item->cost || 0);
	}
    }
    $cost = sprintf("%0.2f",$cost);
}

#------------------------------------------------------------
# media_handling_fees
#------------------------------------------------------------

sub media_handling_fees {
    my $self = shift;
    my %media_fee_hash;
    my $request_obj;

    foreach $request_obj ($self->all_items) {
       my $media_handling_fee = $request_obj->media_handling_fee;
       if (defined($media_handling_fee) ) {
          my $archive_id = $request_obj->archive_id;
          my $type = $media_handling_fee->type;
          # How the hash key is constructed determines at which level
          # media handling fee is calculated.
          #
          # A hash key of $archive_id:$type means the fee is calculated
          # once per daac for that media type. 
          #
          # If the fee needs to be calculated at the dataset level
          # then add dataset id as part of the hash key, e.g.
          # my $dataset_id = $request_odj->dataset_id;
          # $media_fee_hash{"$archive_id:$dataset_id:$type"} = 
          #     $media_handling_fee->fee;   
          #
          $media_fee_hash{"$archive_id:$type"} = $media_handling_fee->fee;
       }
    }
   return (\%media_fee_hash);
}

#------------------------------------------------------------
# handling_fee
#------------------------------------------------------------

sub handling_fee {
    my $self = shift;
    my %handling_fee_hash;
    my $request_obj;

    foreach $request_obj ($self->all_items) {

           # Do we have :
           #   Any HANDLING_FEE Group(s)?
           #   Has the HANDLING_FEE already been added to the Hash?
           #   Are we allowed to charge the Handling Fee.
           if ($request_obj->handling_fee_ref && !$handling_fee_hash{$request_obj->dataset_id} && !($request_obj->package_combo->apply_handling_fee eq 'N')) {
              my $single_handling_fee;
              my $fee = IU::HandlingFee->new();

              my $tot_price;
              foreach $single_handling_fee (@{$request_obj->handling_fee_ref}) {
                  $handling_fee_hash{$single_handling_fee->price_comment} = $single_handling_fee->price;
              }
           }
    }
   return (\%handling_fee_hash);
}


#------------------------------------------------------------
# to_archive_request_ODL PARAMHASH
#------------------------------------------------------------
# I<Private.> 
# Convert to ODL suitable for a product request to a single archive.
# This is a group that normally is filed under a PRODUCT_REQUEST tag.

sub to_archive_request_ODL {
    my ($self, %params) = @_;
    my $archid = $params{ArchId} || puke 600, "no archive id!";
    my $no_validate = $params{NoValidate};

    # Begin:
    my $odl = IU::ODL->new_group;

    # Basic stuff:
    $odl->add_param(MESSAGE_ID => '0'); 
    $odl->add_param(REQUEST_ID => $self->id);
    $odl->add_param(DATA_CENTER_ID => $archid);
    # Add the DORRAN_USERNAME if the user is a persistent user
    my $persistent_user = $ENV{REMOTE_USER};
    $odl->add_param(DORRAN_USERNAME => IU::ODL->quoted($persistent_user))
     if($persistent_user ne 'guest');

    my $user = $self->{User};
    if ($user) {
       # Add the authenticator, if any:
       my $authenticator = $user->authenticator;
       if ($authenticator) {
          $odl->add_param(AUTHENTICATOR => $authenticator);
       }
       # Add the nasda_authenticator, if any:
       my $nasda_authenticator = $user->nasda_authenticator;
       if ($nasda_authenticator) {
          $odl->add_param(NASDA_AUTHENTICATOR => $nasda_authenticator);
       }
       # Add the ecs authenticator, if any:
       my $ecs_authenticator = $user->ecs_authenticator;
       if ($ecs_authenticator) {
          $odl->add_param(ECS_AUTHENTICATOR => $ecs_authenticator);
       }
       # Add the affiliation and addresses:
	$odl->add_param(USER_AFFILIATION => $user->affil->to_ODL)
	    if $user->affil;

        # Kludge! to "fix" problem with GSFC-ECS orders.
        # If ORGANIZATION more than 30 chars the order hangs.
        # "Fix" it here until GSFC-ECS can get their server fixed.
        # $len sets the maximum length of ORGANIZATION in the ODL.
        # If longer than $len ORGANIZATION is truncated.
        # LARC-ECS also has the same problem.
        my $len = ($archid eq 'GSFC-ECS' ||
                   $archid eq 'LARC-ECS' ) ? '30':'70';
   
        $odl->add_param(CONTACT_ADDRESS  => $user->req_addr->to_ODL($len))
            if $user->req_addr;
        $odl->add_param(SHIPPING_ADDRESS  => $user->ship_addr->to_ODL($len))
            if $user->ship_addr;
	$odl->add_param(BILLING_ADDRESS  => $user->bill_addr->to_ODL($len))
	    if $user->bill_addr;
    }
    else {       # no user info!
	$no_validate or puke "no user information!";
    }
    # The line items:    
    my $order_item;
    my @sorted_items = sort { 
	$IW::Order::Request::a->sort_cmp($IW::Order::Request::b) 
	} $self->archive_items($archid);
    foreach $order_item (@sorted_items) {
	$odl->add_param(LINE_ITEM => $order_item->to_ODL);
    }
    $odl; # done!
}

#------------------------------------------------------------
# to_request_ODL PARAMHASH
#------------------------------------------------------------
# Convert to ODL suitable for a product request, as given to the client.

sub to_request_ODL {
    my ($self, %params) = @_;
    my $no_validate = $params{NoValidate};

    # Begin:
    my $file = IU::ODL->new_group;    
    my $odl = IU::ODL->new_group;
    $file->add_param('root' => $odl);

    # Basic stuff:
    my $archid;
    foreach $archid ($self->archive_ids) {
	### debug "converting order for $archid to ODL...";
	my $arch_odl = $self->to_archive_request_ODL(ArchId=>$archid,
						     NoValidate=>$no_validate);
	$odl->add_param(PRODUCT_REQUEST => $arch_odl);
    }
    $file;
}

#------------------------------------------------------------
# print_receipt_data FH
#------------------------------------------------------------
# I<Instance method.>
# Print the portion of the receipt for a standard IMS product request order.

sub print_receipt_data {
    my ($self, $FH) = @_;

    my $archid;
    foreach $archid ($self->archive_ids) {	
	my $count = 0;                         # tally for each archive
	
	# Get the sorted array of items:
        my @sorted_items = sort { 
	   $IW::Order::Request::a->sort_cmp($IW::Order::Request::b) 
	    } $self->archive_items($archid);	
        my $number_items = @sorted_items;
        ### debug "Number of items is $number_items";
	next if (!@sorted_items);              # what the heck

        # Header:
	print $FH "-" x 60, "\n";
	print $FH "Ordered by:  STANDARD MEANS (IMS PRODUCT REQUEST)\n";
	print $FH "Data center: $archid\n";
	print $FH "\n";

	# Output the line items:
	$self->print_line_items($FH, \@sorted_items);
    }
    1;
}

#============================================================
#
package IW::Order::Item;
#
#============================================================

=head2 IW::Order::Item

A "line item".  Basically, a PkgCombo plus some other information.
Real order items will have:

     ArchiveId        [AcctName]
     DatasetId
     PkgCombo 

B<WARNING:> this structure is NOT storable!

=cut

use IU;
use IU::Log;
use IU::PkgCombo;

#------------------------------------------------------------
# new PARAMHASH
#------------------------------------------------------------
# Create a new (abstract)item.
# Params are:
#      
#     Item      => an item (dataset/granule) from the shopping cart
#     PkgCombo  => a package combo (package id, processing option,
#                  media type, media format).  If empty or undefined,
#                  this must be ordered by email.
#     AcctName  => billing id chosen by user, if any
#
# WARNING: structure is NOT storable!

sub new {
    my ($class, %params) = @_;
    my $self = bless \%params, $class;
    $self;
}

#------------------------------------------------------------
# Access methods...
#------------------------------------------------------------
sub orderable     { $_[0]->{Orderable} }
sub archive_id    { $_[0]->{Orderable}->archive_id }
sub dataset_id    { $_[0]->{Orderable}->dataset_id }
sub package_combo { $_[0]->{PkgCombo} }
sub account_name  { $_[0]->{AcctName} }
sub handling_fee_ref  { $_[0]->{HandlingFee} }
sub media_handling_fee { $_[0]->{MediaHandlingFee} }

#------------------------------------------------------------
# name
#------------------------------------------------------------
# Name of this line item, for logging purposes.  
#
# Basically, this is "archid/dsetid/pkname", where "pkname" is the line 
# item package name (either a granule id or a real package id).

sub name {
    my $self = shift;
    join('/', 
	 $self->archive_id,
	 $self->dataset_id,
	 $self->package_combo->line_item_package);
}

#------------------------------------------------------------
# cost
#------------------------------------------------------------
# Return cost if known, undef if not.

sub cost {
    my $self = shift;
    $self->{PkgCombo}->cost;
}

#------------------------------------------------------------
# sort_cmp
#------------------------------------------------------------
# I<Instance method, or function.>
# Compare, for the purposes of sorting.
# Items are sorted on:  
#
#       1. dataset id
#       2. line item package id
#       3. processing option
#       4. media type
#       5. media format

sub sort_cmp {
    my ($i, $j) = @_;
    my $comp;

    $comp = ($i->{Orderable}->dataset_id cmp $j->{Orderable}->dataset_id)
	and return $comp;

    $comp = ($i->{PkgCombo}->line_item_package cmp 
	     $j->{PkgCombo}->line_item_package)
	and return $comp;

    $comp = ($i->{PkgCombo}->proc_opt    cmp  $j->{PkgCombo}->proc_opt)
	and return $comp;

    $comp = ($i->{PkgCombo}->media_type  cmp  $j->{PkgCombo}->media_type)
	and return $comp;

    return ($i->{PkgCombo}->media_fmt cmp $j->{PkgCombo}->media_fmt);    
}


#------------------------------------------------------------
# to_ODL
#------------------------------------------------------------
sub to_ODL {
    my $self = shift;
    my $combo = $self->package_combo;

    my $odl = IU::ODL->new_group;
#    $odl->add_param(WG_CONTENTS  => $self->orderable->order_type);
    $odl->add_param(DATASET_ID   => $self->dataset_id);
    $odl->add_param(PACKAGE_ID   => $combo->line_item_package);

    # These should always have string values...
    if ($combo->proc_opt ne "") {
	$odl->add_param(PROCESSING_OPTIONS => $combo->proc_opt);
    }
    if ($combo->media_type ne "") {
	$odl->add_param(MEDIA_TYPE   => $combo->media_type);
    }
    if ($combo->media_fmt ne "") {
	$odl->add_param(MEDIA_FORMAT => $combo->media_fmt);
    }
    # This should be a real number, if it exists:
    my $cost = $combo->cost;
    if (defined($cost) && ($cost ne '')) {
	$odl->add_param(EST_COST => 
			(IU::ODL->number(sprintf("%.02f", $cost))));
    }
    # These might very well be undefined...
    if (defined($combo->addn_info) and ($combo->addn_info ne "")) {
	$odl->add_param(ADDITIONAL_INFO => $combo->addn_info);
    }
    if (defined($self->account_name) and ($self->account_name ne "")) {
	$odl->add_param(BILLING_ID => $self->account_name);
    }
    # Subset/order options:
    my $opts_type;
    foreach $opts_type (sort $combo->options) {
	my $opts = $combo->options($opts_type);
	$odl->add_param("${opts_type}_SPEC" => $opts->to_request_ODL);
    }
    # If there is a coverage:
    my $coverage = $self->orderable->coverage;
    if (defined ($coverage) and $coverage->type ne 'EMPTY') {
       my @items = %{$coverage};
       if (scalar(@items)) {
          # Add Path & Row, if available.  They are sequences of int or empty.
          # Get paths from data item's coverage object.
          my @paths = @{$coverage->paths};
          # Get rows from data item's coverage object.
          my @rows = @{$coverage->rows};
          # If there are paths and rows given:
          if (int(@paths) && int(@rows)) {
             # Create new ODL group for PATH_ROW_LOC.
             my $path_row_odl = IU::ODL->new_group;
             # Create list of ODL integers for paths.
             my @odl_int_paths = map {IU::ODL->int($_)} @paths;
             # Add paths to PATH_ROW_LOC group.
             $path_row_odl->add_param('PATH' => 
                                       IU::ODL->sequence(@odl_int_paths));
             # Create list of ODL integers for rows.
             my @odl_int_rows = map {IU::ODL->int($_)} @rows;
             # Add rows to PATH_ROW_LOC group.
             $path_row_odl->add_param('ROW' =>
                                       IU::ODL->sequence(@odl_int_rows));
             # Add WRS type to PATH_ROW_LOC group, if it is known.
             $path_row_odl->add_param('WRS_TYPE' => $coverage->wrs)
                if ($coverage->wrs);
             # Add the newly made PATH_ROW_LOC group to the line item ODL.
             $odl->add_param('PATH_ROW_LOC' => $path_row_odl);
          }
          # Add Geo coordinates.  Note coverage->to_ODL returns
          # (<ODL group name>, <ODL tree>)
          $odl->add_param($coverage->to_ODL);
       } # End if the coverage is non-empty.
    }     # End If there is a coverage.
    # Return odl object.
    $odl;
}

#--------------------
#
# Event-logging...
#
#--------------------

#------------------------------------------------------------
# prodname
#------------------------------------------------------------
# Name of this line item, for event-logging (the "prodname" field).  
#
# Basically, this is "archid/pkname", where "pkname" is the line 
# item package name (either a granule id or a real package id).

sub prodname {
    my $self = shift;
    join('/',
	 $self->archive_id,
	 $self->package_combo->line_item_package);
}

#------------------------------------------------------------
# mediareq
#------------------------------------------------------------
# Media info for this line item, for event-logging (the "mediareq" field). 
# As per the ICD, this is one of:
#
#     ftp 8mm 4mm 9track 3280 3290 CDROM CDRec floppy MO paper film video
#
# Returns the media type, or empty if none could be determined.

sub mediareq {
    my $self = shift;
    my $mediatype = $self->package_combo->media_type;
    
    if ($mediatype =~ /\b(ftp|8[\s-]?mm|4[\s-]?mm|9[\s-]?track|3280|3290|cd[\s-]?rom|cd[\s-]?rec|floppy|mo|paper|film|video)\b/i) {
	my $raw = lc($1);
	$raw =~ s/[\s-]//g;       # strip whitespace and puncts
	return $raw;
    }
    return '';        # no idea
}

#------------------------------------------------------------
# volreq
#------------------------------------------------------------
# Volume info for this line item, for event-logging (the "volreq" field). 
# Returns the empty if none could be determined.

sub volreq {
    return '';        # no idea
}

#------------------------------------------------------------
1;
