package IW::Popup;


=head1 NAME

IW::Popup - abstract class for "popup" forms


=head1 DESCRIPTION

This is an abstract subclass of IW::Mode which represents a "popup"
form which is invoked as part of another form.  An example would be
the map-selection popup, invoked from the search form.

All the CGI variables provided by a given Popup begin with a unique
"signature".  This may be accessed via the C<sig()> method.  For example,
if C<$popup->sig()> returns "pmap", then possible names for CGI tags 
containing the relevant information might be C<"pmap.north">, C<"pmap.south">,
etc.

Popup forms may consist of a complicated sequence of pages, like the
map selection widget: thus, they must provide a state-transition
method, C<next_state()>, which tells what the next state is.

The current state is kept in the C<{sig}.state> CGI variable.


=head1 SYNOPSIS

    # Create a new popup object:
    my $popup = new YourPopupClass;    # no args: your new should provide
     
    # Output a summary of the popup's contents (inside another page);
    $popup->summary;
    
    # Get the signature of the popup:
    my $sig = $popup->sig;    


=head1 INSTANCE VARIABLES

=over 

=item Sig

The "signature", or prefix of all our CGI variables.

=item OkButtonStr

The message inside a "submit" button.

=back


=head1 NOTES

All CGI fields in the "parent" form (the one who is popping us up)
must be propagated into the popup as hidden fields; this is how (for
example) we keep track of which "parameters" were selected when we are
selecting "sources".

The reason we do not use the session state to keep ahold of these
variables is the "back button": we want to ensure WYSIWYG when users
are looking at a summary of the search about to be made.

Concrete subclasses must not require arguments to C<new()>, since
they are loaded by IW::Mode, which cannot know the appropriate arguments.


=head VERSION

$Id: Popup.pm,v 1.7 1997/03/14 17:14:18 eryq Exp $

=cut

use IU;
use IW::App;
use IW::Mode ();

use strict;

@IW::Popup::ISA = qw(IW::Mode);


#------------------------------------------------------------
# new PARAMHASH
#------------------------------------------------------------
# I<Constructor.>  The PARAMHASH may contain the following:
#
#     Sig => the prefix of the fields: 'pparam', 'pmap', etc.

sub new {
    my $class = shift;
    my %params = @_;

    my $self = {};
    bless $self, $class;

    # The signature:
    $self->{Sig} = $params{Sig} || puke 600, "no popup signature given!";

    # What to print in OK buttons:
    $self->{OkButtonStr} = 'Ok!  Accept my input & return to the search form!';
    $self;
}

#------------------------------------------------------------
# DESTROY 
#------------------------------------------------------------
sub DESTROY { 1 }


#------------------------------------------------------------
# ok_str
#------------------------------------------------------------
# Get the string for the "OK" button.

sub ok_str {
    $_[0]->{OkButtonStr};
}

#------------------------------------------------------------
# param PARAMNAME, [VALUES...]
#------------------------------------------------------------
# Get the given CGI parameter for this popup.  This just prepends
# the signature to the name.

sub param {
    my $self = shift;
    my $tag = shift;
    $REQ->param($self->tag($tag), @_);
}

#------------------------------------------------------------
# pass_foreign
#------------------------------------------------------------
# Dump all the existing 'pxxxx.yyy' CGI params as hidden fields, 
# so we maintain our state.  Also passes the sid, but NOT the mode!
#
# HOWEVER: if a field name begins with our Sig, it is I<not> printed,
# since it is assumed that WE will be outputting it ourselves.

sub pass_foreign {
    my $self = shift;

    # Pass foreign tags:
    my ($tag, $val);
    foreach $tag ($REQ->params) {

	# Skip if ours:
	next if ($self->{Sig} &&
		 substr($tag, 0, length($self->{Sig})) eq $self->{Sig});

	# Skip if not a popup field:
	next if $tag !~ /^p[a-z]+\.[a-z]*/;
	
	# Okay: output the params!
	foreach $val ($REQ->param($tag)) {
	    $HTML -> input_hidden($tag, $val);
	}
    }

    # Pass special tags:
    $HTML -> input_hidden('sid', scalar($REQ->param('sid')));
    1;
}

#------------------------------------------------------------
# sig
#------------------------------------------------------------
# Get the signature (the CGI tag prefix) of this popup.

sub sig {
    $_[0]->{Sig};
}

#------------------------------------------------------------
# tag TAG
#------------------------------------------------------------
# Get the name of a tag with "{Sig}.".  For example, if the signature
# of this popup is C<"foo">, then C<tag("bar")> is C<"foo.bar">.

sub tag {
    my ($self, $tag) = @_;
    "$self->{Sig}.$tag";
}

#------------------------------------------------------------
1;
