package IW::Reply;


=head1 NAME

IW::Reply - CGI reply interface for the EOSDIS V0 Web Gateway scripts


=head1 DESCRIPTION

Routines that manage some of the dirty work of formulating 
replies to the client... in particular, the buffering of output.

You should only ever create one instance of this class.

In order to understand this class, you should acquaint yourself
with the Grand CGI Scheme of CGI:: and IW::

   CLASS             DESCRIPTION
    ============================================================
    CGI::Base         Interfacing with only the most basic CGI mechanisms
                      Query string is constructed, but not parsed.
    
    CGI::BasePlus     Subclass of CGI::Base.
                      Just adds ability to deal with file-upload postings.    
    ------------------------------------------------------------       
    CGI::Request      Parses query into individual parameters.
                      Has pointer to an associated CGI::Base object.
    
    CGI::Form         Subclass of CGI::Request.
                      Contains a lot of HTML-generation methods.    
    ------------------------------------------------------------       
    CGI::Response     For constructing HTTP reply headers.  
                      Currently broken.  Depends on HTTP:: hierarchy.     
    ------------------------------------------------------------       
    IW::Reply         For optional transparent STDOUT buffering.
    ------------------------------------------------------------   
    IW::ReplyHead     HTTP reply header object.
                      Basically, a dumb, lightweight CGI::Response.
    ------------------------------------------------------------   


=head1 PAGE BUFFERING

If the C<use_pagebuffer()> method is invoked at the start, then
this module will invisibly make use of an anonymous temporary file
(a tmpfile), called a I<pagebuffer>, to hold all output for the current 
page until that output is fully prepared.  STDOUT is redirected 
through to that pagebuffer; you can use C<print>, etc. as if nothing
had changed.

For example, the code:

    print "Content-type: text/html\n\n";
    print "<H1>Hello!</H1>\n";

does not actually cause the document to be sent to the client at that point.
It will not be sent until:

    $Reply->flush_output() 

is called.

The page buffer is a true C<tmpfile>, and so is deleted automatically 
on exit.


=head2 Single-part documents

There is nothing special about single-part (e.g., C<"text/html">, 
C<"text/plain">) documents.  Their I<entire> content, including the
MIME header, is pagebuffered.


=head2 Dynamically-updated documents

The use of C<multipart/x-mixed-replace> documents for the connection
status page was the implementation up through WG version 1.8.
However, since MSIE does not support them, they have been
removed in favor of a "client pull" implementation.

To avoid tempting the naive, all the multipart support has been removed
from this module.



=head1 ISSUING REDIRECTS

Redirects may be issued easily via the redirect() method:

    $Reply->redirect($url);                 # temporary redirect
    
    $Reply->redirect($url, 'PERMANENT');    # permanent redirect

This sends the redirect directly to the client: page-buffering is not
used (normally, the HTTP header is printed to STDOUT, which is
page-buffered so that non-HTML documents may be retracted if an error
document needs to be generated).

You should probably not attempt to output anything else after calling
this method.

=cut


require 5.002;      # because of SUPER, FileHandle, etc.

use FileHandle ();
use IW::ReplyHead;
use IU::Log;
use OHTTPD;	    # for is_nph check
use strict;



#------------------------------------------------------------
#
# Globals, private:
#
#------------------------------------------------------------

# If set true, outputs debugging information at the tail of a document:
my $DEBUG_SIGNATURE = 0;

# If defined, then this (and not STDOUT) is the real stdout:
my $ClientFH = undef;

# The current filehandle where buffered output has been redirected:
# WARNING: DO NOT OUTPUT TO THIS DIRECTLY! USE STDOUT!
my $BufferFH = undef;    

# The current (and only) reply object:
my $Reply = undef;


#------------------------------------------------------------
#
# PRIVATE METHODS
#
#------------------------------------------------------------

#------------------------------------------------------------
# CLIENT
#------------------------------------------------------------
# Return the filehandle that I<really> goes to the client.
# Depending on when this is called, this could be either the
# original STDOUT, or the saved STDOUT.

sub CLIENT {    
    ($ClientFH ? $ClientFH : \*STDOUT);
}

#------------------------------------------------------------
# flush_pagebuffer
#------------------------------------------------------------
# Take the temporary file in which we have been accumulating our output
# (the pagebuffer), rewind it, and flush its entire contents to the 
# I<real> STDOUT (connected to the client).  Then discard the temporary 
# file, and reopen a new temporary file.  
#
# B<NOTE:> This does not I<flush> the filehandle-to-the-client: it just
# dumps the pagebuffer's contents to that filehandle.  If you need to 
# get the accumulated output to the client I<now>, use C<flush_output()>.
#
# B<NOTE:> Output is still buffered, but all previous buffer 
# contents are lost.

sub flush_pagebuffer {
    my $self = shift;

    # Flush pending output, rewind buffer, and copy contents to client:
    $self->rewind_pagebuffer;
    $self->print_fh_to_fh($BufferFH, $self->CLIENT);

    # Rebuffer:
    $self->new_pagebuffer;
}

#------------------------------------------------------------
# is_nph_script
#------------------------------------------------------------
# Is this an "nph-" script?

sub is_nph_script {
    my $self = shift;

    # By definition, anything running under the custom HTTPD
    # *must be an NPH script, since no one's parsing the header!
    return 1 if $ENV{'SERVER_SOFTWARE'} eq OHTTPD::software();

    # Otherwise, look at the script name for the telltale "nph-": 
    ### my $name = $CGI::Base::SCRIPT_NAME;
    return ($ENV{'SCRIPT_NAME'} =~ m|/nph-[^/]+$|);
}

#------------------------------------------------------------
# need_status_line
#------------------------------------------------------------
# We need to know this to determine whether or not we should
# preface our header with an HTTP status line.

sub need_status_line {
    shift->is_nph_script;
}

#------------------------------------------------------------
# new_pagebuffer
#------------------------------------------------------------
# Redirects STDOUT so all writes to STDOUT (which remains the
# default filehandle) really go to an anonymous temporary file.
#
# You can invoke this as many times as you like: each invocation 
# will delete the existing temporary file before opening a new one.
# Anything sitting in that file will be lost.
#
# Note that we can't just do a select() of the temp file, since child 
# processes we spawn will inherit our stdout... which we want to be 
# the I<temporary> file...

sub new_pagebuffer {
    my $self = shift;

    # Close any existing buffer:
    if ($BufferFH) {
	close $BufferFH; 
	$BufferFH = undef;
    }

    # Get a new anonymous temp file:
    $BufferFH = FileHandle->new_tmpfile or die "new_tmpfile: $!";

    # Dup STDOUT to go into the tmpfile:
    my $bufno = $BufferFH->fileno;
    open(STDOUT, ">&$bufno") or die "open: $!";
    1;
}

#------------------------------------------------------------
# print_fh_to_fh FH
#------------------------------------------------------------
# Dump contents of FH to the client, line by line.

sub print_fh_to_fh {
    my ($self, $in, $out) = @_;
    print $out $_ while (<$in>);
}

#------------------------------------------------------------
# save_stdout
#------------------------------------------------------------
# Save STDOUT to a new filehandle.  Do this I<once>, at the start.

sub save_stdout {
    my $self = shift;

    open(SAVEOUT, ">&STDOUT") or die "redirect open: $!";  
    $ClientFH = \*SAVEOUT;
}


#------------------------------------------------------------
# 
# PUBLIC INTERFACE
#
#------------------------------------------------------------

#------------------------------------------------------------
# new
#------------------------------------------------------------
# Constructor.  It stores the newly-created instance in the
# global C<$Reply>.  Currently, only one instance is allowed.

sub new {
    my $class = shift;
    die "sorry; only one $class instance currently allowed" if $Reply;
    my $self = {};
    bless $self, $class;
    $Reply = $self;
}

#------------------------------------------------------------
# abort_output
#------------------------------------------------------------
# An error has occurred.  Discard any pending buffered output,
# and leave us in a state to prepare a new (buffered) document.
#
# We just dump the existing buffer and open a new one.

sub abort_output {
    my $self = shift;

    # Can't do it if output isn't buffered!!!!!
    $self->using_pagebuffer or die "can't abort if not using pagebuffering!";
    error "output aborted!";

    # Get a new pagebuffer:
    $self->new_pagebuffer;    
}

#------------------------------------------------------------
# all_done 
#------------------------------------------------------------
# We're done... flush any accumulated output, and disconnect from client.

sub all_done {
    my $self = shift;

    $self->flush_output;
    close($self->CLIENT);
    1;
}

#------------------------------------------------------------
# flush_output
#------------------------------------------------------------
# Whereever we have been putting our output, make sure it gets flushed to 
# the client. 

sub flush_output {
    my $self = shift;

    # If buffered, then write buffered info to client:
    $self->flush_pagebuffer if $BufferFH;

    # Okay, we can flush the filehandle to the client now!    
    FileHandle::flush($self->CLIENT);
}

#------------------------------------------------------------
# redirect URL[,DURATION][,ANNOTATE];
#------------------------------------------------------------
# Issue a redirect to another URL.  Redirects are temporary by default,
# but a DURATION of 'PERMANENT' overrides this.  See documentation above.

sub redirect {
    my ($self, $url, $duration, $annotate) = @_;
    $duration or $duration = 'TEMPORARY';
    $annotate = 0;

    # Construct the header:
    my $head = new IW::ReplyHead;
    $head->redirect($url, $duration) if $self->need_status_line;
    $head->location($url);
    $head->content_type('text/html') if $annotate;
    $head->send($self->CLIENT);  # send it NOW (writes to and flushes CLIENT).
    return 1;

    # Output a redirect document:
    if ($annotate) {
	my $FH = $self->CLIENT;
	print $FH <<EOF;
	  <HTML>
	  <HEAD><TITLE>Redirect!</TITLE></HEAD>
          <BODY>If you see this, please <A HREF="$url">CLICK HERE</A>.</BODY>
	  </HTML>     
EOF
    }
    1;
}

#------------------------------------------------------------
# rewind_pagebuffer 
#------------------------------------------------------------
# Flush and rewind the pagebuffer filehandle, and return it (or
# undef if output is not pagebuffered).
#
# B<WARNING: NOT FOR THE SQUEAMISH!>
# This is used to take a snapshot of the dying gasp.  

sub rewind_pagebuffer {
    my $self = shift;
    
    # No-op if not buffered:
    return undef if !$self->using_pagebuffer;

    # Flush pending output to temp file.
    # KLUDGE: I don't know WHY we need to flush both of them...
    STDOUT->flush();  
    $BufferFH->flush();

    # Rewind the buffer, and write its contents:
    $BufferFH->seek(0, 0);
    $BufferFH;
}

#------------------------------------------------------------
# use_pagebuffer
#------------------------------------------------------------
# Use buffering (with temporary files).
# Do this I<once>, before I<anything> has been output.
# After this is called, using_pagebuffer() is true.

sub use_pagebuffer {
    my $self = shift;

    $self->save_stdout or die "couldn't save stdout";
    $self->new_pagebuffer or die "couldn't buffer stdout";
    1;
}

#------------------------------------------------------------
# using_pagebuffer
#------------------------------------------------------------
# Is output buffered?

sub using_pagebuffer {
    return $BufferFH;
}

#------------------------------------------------------------
# set truncation point
#------------------------------------------------------------
# Record current file position so we can truncate it later

sub set_truncation_point {
  my $self = shift;
  STDOUT->flush();  
  $BufferFH->flush();
  $self->{TruncationPoint} = sysseek($BufferFH, 0, 1);
}

#------------------------------------------------------------
# truncate_output
#------------------------------------------------------------
# Truncate file to the previously-determined truncation point.

sub truncate_pagebuffer {
  my $self = shift;
  STDOUT->flush();  
  $BufferFH->flush();
  truncate $BufferFH, $self->{TruncationPoint}
    or die "Couldn't truncate: $!";
}

1;



