package IW::ReplyHead;

=head1 NAME

IW::ReplyHead - lightweight replacement for broken CGI::Response

=head1 DESCRIPTION

CGI::Response has the right idea, but is broken because of
HTTP:: incompatibilities.

CGI::Base isn't broken, but is tacky and difficult to use.

So, we roll our own.  Sigh.  

I would have called this IW::Response, except it's too similar 
in connotation to IW::Reply... and, it would imply that this
module either mimics or inherits from CGI::Response, which it
does not.

=cut

use FileHandle ();
use strict;


#------------------------------------------------------------
#
# PRIVATE
#
#------------------------------------------------------------

#------------------------------------------------------------
# delta_to_date
#------------------------------------------------------------
# Lifted from CGI.pm's expires().
#
# This internal routine creates an expires string exactly some number of
# hours from the current time in GMT.  This is the format
# required by Netscape cookies, and I think it works for the HTTP
# Expires: header as well.
#
# Format for time can be in any of the forms...
#
#    "now" -- expire immediately
#    "+180s" -- in 180 seconds
#    "+2m" -- in 2 minutes
#    "+12h" -- in 12 hours
#    "+1d"  -- in 1 day
#    "+3M"  -- in 3 months
#    "+2y"  -- in 2 years
#    "-3m"  -- 3 minutes ago(!)
#
# If you don't supply one of these forms, we assume you are
# specifying the date yourself
#
# Note: Netscape's specification for this date requires a four-digit date;
# this code has been updated to conform to the spec.

sub delta_to_date {
    my($time) = @_;
    my(@MON)=qw/Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec/;
    my(@WDAY) = qw/Sunday Monday Tuesday Wednesday Thursday Friday Saturday/;
    my(%mult) = ('s'=>1,
		 'm'=>60,
		 'h'=>60*60,
		 'd'=>60*60*24,
		 'M'=>60*60*24*30,
		 'y'=>60*60*24*365);
    my($offset);
    if (!$time || ($time eq 'now')) {
	$offset = 0;
    } elsif ($time=~/^([+-]?\d+)([mhdMy]?)/) {
	$offset = ($mult{$2} || 1)*$1;
    } else {
	return $time;
    }
    my($sec,$min,$hour,$mday,$mon,$year,$wday) = gmtime(time+$offset);
    return sprintf("%s, %02d-%s-%04d %02d:%02d:%02d GMT",
		   $WDAY[$wday],$mday,$MON[$mon],$year+100,$hour,$min,$sec);
}


#------------------------------------------------------------
#
# PUBLIC INTERFACE
#
#------------------------------------------------------------

#------------------------------------------------------------
# new 
#------------------------------------------------------------
# Create a new reply object

sub new {
    my $class = shift;
    my $self = {};
    $self->{Content_Type} = 'text/html';
    bless $self, $class;
}

#------------------------------------------------------------
# as_string
#------------------------------------------------------------
# Format the entire header as a string, and return it.
# The string is terminated with "\r\n".

sub as_string {
    my $self = shift;
    my $out = '';

    # Status (only if there):
    if (defined($self->{Status})) {
	$out .= "HTTP/1.0 $self->{Status}\r\n";
    }

    # Location (only if there):
    if (defined($self->{Location})) {
	$out .= "Location: $self->{Location}\r\n";
    }

    # URI (only if there):
    if (defined($self->{Location})) {
	$out .= "URI: <$self->{Location}>\r\n";
    }

    # Content-type (only if there):
    if (defined($self->{Content_Type})) {
	$out .= "Content-type: $self->{Content_Type}";
	($self->{Boundary}) and $out .= qq!;boundary="$self->{Boundary}"!;
	($self->{Filename}) and $out .= qq!;name="$self->{Filename}"!;
	$out .= "\r\n";
    }

    # Content-disposition (only if needed):
    if (defined($self->{Filename})) {
	$out .= qq!Content-disposition: inline ; ! .
	        qq!filename="$self->{Filename}"\r\n!;
    }

    # Title (only if needed):
    if (defined($self->{Filename})) {
	$out .= "Title: $self->{Filename}\r\n";
    }

    # No-cache pragma (only if there):
    if (defined($self->{Pragma_NoCache})) {
	$out .= "Pragma: no-cache\r\n";
    }

    # Expires (only if there):
    if (defined($self->{Expires})) {
	$out .= "Expires: $self->{Expires}\r\n";
    }

    # Done!
    $out .= "\r\n";
    $out;
}

#------------------------------------------------------------
# boundary [VALUE]
#------------------------------------------------------------
# Get/set the boundary, for multipart documents:
#
#      $reply->boundary('X------Y');
#
# No arguments means do nothing.  The new value is always returned.

sub boundary {
    my ($self, $value) = @_;
    $self->{Boundary} = $value if (@_ > 1);
    $self->{Boundary};
}

#------------------------------------------------------------
# content_type [VALUE]
#------------------------------------------------------------
# Get/set the content type:
#
#      $reply->content_type('text/html');
#
# No arguments means do nothing.  The new value is always returned.

sub content_type {
    my ($self, $value) = @_;
    $self->{Content_Type} = $value if (@_ > 1);
    $self->{Content_Type};
}

#------------------------------------------------------------
# expires [DELTA]
#------------------------------------------------------------
# Get/set the expiration date.
#
#      $reply->expires('+3h');     # expires 3 hours from now
#
# No arguments means do nothing.  The new value as a full HTTP date
# is always returned.

sub expires {
    my ($self, $delta) = @_;
    if (@_ == 2) {
	$self->{Expires} = &delta_to_date($delta);
    }
    return $self->{Expires};
}

#------------------------------------------------------------
# filename [FILENAME]
#------------------------------------------------------------
# Get/set the recommended filename:
#
#      $reply->filename('foo.dat');
#
# No arguments means do nothing.  The new value is always returned.

sub filename {
    my ($self, $value) = @_;
    $self->{Filename} = $value if (@_ > 1);
    $self->{Filename};
}

#------------------------------------------------------------
# location [URL]
#------------------------------------------------------------
# Get/set the desired redirect location:
#
#      $reply->location($url);
#
# No arguments means do nothing.  The new value is always returned.

sub location {
    my ($self, $value) = @_;
    $self->{Location} = $value if (@_ > 1);
    $self->{Location};
}

#------------------------------------------------------------
# no_cache
#------------------------------------------------------------
# Set the no-cache pragma, and set the expiration date to now.
#
#      $reply->no_cache;
#
# No usable value is returned.

sub no_cache {
    my $self = shift;
    $self->{Pragma_NoCache} = 1;
    $self->expires(0);
    1;
}

#------------------------------------------------------------
# send [FILEHANDLE]
#------------------------------------------------------------
# Write to the given FILEHANDLE (or the default filehandle if
# none given), and then flush the filehandle.  The latter stage is
# very important if this process spawns child processes which might
# write to the same output stream; for example, C<"system('ls -l')">.

sub send {
    my ($self, $FH) = @_;
    $FH or $FH = select;
    print $FH $self->as_string;
    FileHandle::flush($FH);
}

#------------------------------------------------------------
# status [STATUS,MESSAGE]
#------------------------------------------------------------
# Get/set the server header:
#
#      $reply->status(200, "OK");
#
# No arguments means do nothing.  The new value is always returned.

sub status {
    my ($self, $status, $message) = @_;
    if (@_ == 3) {
	$message or ($message = 'No message');
	$self->{Status} = "$status $message";
    }
    return $self->{Status};
}

#------------------------------------------------------------
#
# INITIALLIZERS: DO THESE FIRST!
#
#------------------------------------------------------------

#------------------------------------------------------------
# okay
#------------------------------------------------------------
# Set up an "okay" status header.

sub okay {
    my $self = shift;
    $self->status(200, "OK");
}

#------------------------------------------------------------
# redirect URL[,PERMANENT]
#------------------------------------------------------------
# Set up a "redirect" status header.

sub redirect {
    my ($self, $url, $perm) = @_; 
    defined($perm) or ($perm = '');

    # Set up status line:
    if ($perm =~ /^permanent/i) {
	$self->status(301, "Moved Permanently");
    }
    else {
	$self->status(302, "Moved Temporarily");
    }
}

#------------------------------------------------------------
1;



