package IW::Sort;

=head1 NAME

IW::Sort - create multi-key sort menus, and do the sorting

=head2 SYNOPSIS

Here's the basic public interface:

    # Create a new sort object (subclasses may provide some of these args):
    my $sorter = new IW::Sort
                      KeyMap => {'Start date' => '$a->{SD} cmp $b->{SD}',
                                 'Stop date'  => '$a->{ED} cmp $b->{ED}',
                                 'Id'         => '$a->id <=> $b->id']},
                      SeqLen => 2,
                      Seq    => ['Id', 'Stop date'];
     
    # Output a nice description of the keys: 
    $sorter->as_string;                # "Id and then Start date"
    
    # Set array of sort keys:
    $sorter->sequence('Id', 'Stop date');
    
    # Get effective array of sort keys (filters out 'nothing' entries):
    my @keys = $sorter->sequence;

Here are the routines that deal with actual sorting:    
    
    # Method returns the "test" for a given key, or undef if not handled.
    #    This is Perl code of the form "$a->{Attr} cmp $b->{Attr}".
    #    Subclasses must override this! 
    my $test = $sorter->test_for_key('Start date');

    # Return sort function (for current keys) as a text string:
    my $sort_func = $sorter->sort_func;
    
    # Apply sort function (for current keys) to an array, or die:
    my @sorted = $sorter->sort(\@unsorted);

Here's the CGI interface:    
    
    # CGI interface... output form fields (foo.sortkey.*):
    $sorter->print_fields(Prefix=>'foo');
   
    # CGI interface... read in from form input (foo.sortkey.*):
    $sorter = IW::Sort->from_CGI(Prefix=>'foo');
    

=head1 DESCRIPTION

=head2 CGI interface

Every entity with a CGI interface (from_CGI(), print_fields())
takes a "prefix" in those methods.  This prefix tells it what,
if anything, to stick in front of the CGI tags.

For a normal "sort" object, the tags are:

     sortkey.0
     sortkey.1
     sortkey.2 ...

So if you gave a "prefix" of "gran." to any of those methods, the
CGI tags would (for example) be:

     gran.sortkey.0
     gran.sortkey.1
     gran.sortkey.2 ...

Subclasses of IW::Sort may override the CGI interface so that
a special prefix (e.g., "gran") is always appended after the
caller-specified prefix.


=head1 AUTHOR

Eryq, Hughes STX Corporation.


=head1 RCS INFO

$Id: Sort.pm,v 1.21 2002/09/17 20:02:57 joe Exp $

=cut

use IU;
use IU::Log;
use IW::App;
use IW::Babel;
use strict;


#------------------------------------------------------------
# new PARAMHASH
#------------------------------------------------------------
# Create a complex sort menu. The PARAMHASH contains:
#
#    KeyMap  => hashref of (key=>test) pairs.
#    Keys    => optional: the keys, in the order they should be presented.
#    Seq     => optional: initial/default sequence of sort keys.
#    SeqLen  => optional: max number of sort keys in sequence (default = 2);
#    Prefix  => optional: defaults to 'sortkey'.

sub new {
    my ($class, %params) = @_;
    my $self = bless \%params, $class;
    $self->{KeyMap}  or puke "you need to provide me with the sort keys!";
    $self->{Keys}    or $self->{Keys} = [sort (keys %{$self->{KeyMap}})];
    $self->{SeqLen}  or $self->{SeqLen} = 2;
    $self->{Seq}     or $self->{Seq} = [];
    $self->{Prefix}  or $self->{Prefix} = 'sortkey';
    $self;
}

#------------------------------------------------------------
# keys
#------------------------------------------------------------
# Return the keys, in the order they should be presented.
sub keys {
    my $self = shift;
    @{$self->{Keys}};
}

#------------------------------------------------------------
# as_string
#------------------------------------------------------------
sub as_string {
    my $self = shift;
    join ' and then ', $self->sequence;
}

#------------------------------------------------------------
# sequence
#------------------------------------------------------------
# Set/get the sequence of keys to sort on.
# This is a vector of sort keys, as they appear in the menus, EXCEPT
# that when returning the value, the 'nothing' entries are filtered out.

sub sequence {
    my $self = shift;
    if (@_ > 0) {          # set...
	my @seq = grep { !/^nothing$/ } @_;
	$self->{Seq} = \@seq;
    }
    grep { !/^nothing$/ } @{$self->{Seq} || []};    
}


#------------------------------------------------------------
# test_for_key
#------------------------------------------------------------
sub test_for_key {
    my ($self, $key) = @_;
    $self->{KeyMap}{$key};
}

#------------------------------------------------------------
# sort_func 
#------------------------------------------------------------
# Maps 'keys' entries to tests.
sub sort_func {
    my $self = shift;
    my $key;


    # Old session states will have keymaps with bad names. Fix them.
    foreach (%{$self->{KeyMap}}) {
      my $var = short2var($_);
      if(defined($var) && ($var)) {
         my $newkey = var2short($var);
         $self->{KeyMap}->{$newkey} = $self->{KeyMap}->{$_};
      }
    }

    # Map the keys to comparisons. Note that old session states will
    # have bad sequences. Fix those too.
    #
    # Session states which were created by uploading the example user
    # prefs from the POD in User.pm will have things way in the wrong places.
    # If test_for_key says "no such", just completely ignore the test.
    my @tests;
    foreach $key ($self->remap_names($self->sequence)) {
	my $test = $self->test_for_key($key) || '';
	push @tests, $test if $test;
    }

    # Build the desired sort keys into a single routine:
    my $subt = "sub SORTER {\n   my \$comp;\n";
    my $test;
    foreach $test (@tests) {
	$subt .=  "   \$comp = $test;\n\$comp && return \$comp;\n\n";
    }
    $subt .= "   0;\n}";
    
    # Compile it:
    $subt;
}

#------------------------------------------------------------
# sort ARRAYREF
#------------------------------------------------------------
# Sort elements of ARRAYREF.
# Returns the sorted array (possibly empty if original array was empty!), 
# and dies on failure.  

sub sort {
    my ($self, $unsorted) = @_;

    # Compile sort func:
    my $sort_func = ($self->sort_func || puke "no sort function!");
    eval "$sort_func";
    $@ and puke "couldn't compile sort func <$sort_func>: $@ !";
    
    # Run it!
    no strict 'refs';
    sort SORTER @$unsorted; 
}

#------------------------------------------------------------
#
# CGI interface
#
#------------------------------------------------------------

#------------------------------------------------------------
# print_fields PARAMHASH
#------------------------------------------------------------
# Output CGI widgets. PARAMHASH contains:
#
#    Prefix => the prefix.

sub print_fields {
    my ($self, %params) = @_;
    my $prefix = $params{Prefix} || '';

    # For convenience:
    my $keys = $self->{Keys};

    # Old session states need to have the sequence names fixed.
    my @seq = $self->remap_names($self->sequence);

    # What can we sort on? (And fix the key names while we're finding out.)
    @$keys = $self->remap_names(@$keys);
    my @allkeys = ('nothing', @$keys);

    # Print!
    $HTML -> tag('NOBR');
    my $i;
    for ($i = 0; $i < $self->{SeqLen}; $i++) {
	$HTML -> t(" then: ")      if $i;
	# Compensate for possible old session states by translating
	# from var to short and back again. This cleans up any superseded
	# sort keys.
	$HTML -> select_one(Name   =>"${prefix}$self->{Prefix}.$i",
			    Choices=>($i ? \@allkeys : $keys),
			    Default=>($seq[$i] || 'nothing'));
    }
    $HTML -> tag('_NOBR');
}

#------------------------------------------------------------
# clear_CGI 
#------------------------------------------------------------
# clear CGI tags
#

sub clear_CGI {
   my $self = shift;
   my %params = @_;
   my @seq;
   my $i;

   # Create new instance, if needed:
   ref($self) or ($self = $self->new(@_));
   my $prefix = $params{Prefix};

   # clear CGI tags so they aren't 
   for ($i = 0; $i < $self->{SeqLen}; $i++) {
      $REQ->delete("${prefix}$self->{Prefix}.$i");
   }
}

#------------------------------------------------------------
# from_CGI PARAMHASH
#------------------------------------------------------------
# Create from CGI input. PARAMHASH contains:
#
#    Prefix => the prefix.
#
# Plus any args that would normally be passed to new().
# If no sequence was obtained, returns undef.

sub from_CGI {
    my $self = shift;
    my %params = @_;
    my @seq;
    my $i;

    # Create new instance, if needed:
    ref($self) or ($self = $self->new(@_));
    my $prefix = $params{Prefix};
 
    for ($i = 0; $i < $self->{SeqLen}; $i++) {
	my $key = $REQ->param("${prefix}$self->{Prefix}.$i");
	push @seq, $key    if $key;
    }
    @seq or return undef;       # nothing to read!
    $self->sequence(@seq);
    $self;
}

#------------------------------------------------------------
# remap_names
#------------------------------------------------------------
# Takes a list of key names and remaps them to currently-supported keys.

sub remap_names {
  my ($self,@list) = @_;
  my $key;
  foreach $key (@list) {
    $key = var2short(short2var($key));
  }
  @list;
}


#============================================================
#============================================================

package IW::Sort::Dataset;

=head2  IW::Sort::Dataset

A subclass of IW::Sort, specifically for sorting dataset objects.
The prefix is appended with "dset."; hence, the CGI tags are:

    dset.sortkey.0
    dset.sortkey.1
    dset.sortkey.2

=cut

use IU;
use IU::Log;
use IW::App;
use IW::Babel;

use strict;

@IW::Sort::Dataset::ISA = qw(IW::Sort);

# Map sort keys to tests:
# WARNING!!! THE "NAMES" MUST AGREE WITH THOSE IN IW::UserDisplay!!!
my %Keys = 
    (
     var2short('DATASET_ID')     => '($a->{Id}        cmp $b->{Id})',
     var2short('DATA_CENTER_ID') => '($a->{ArchiveId} cmp $b->{ArchiveId})',
     var2short('START_DATE')     => '($a->{StartDate} cmp $b->{StartDate})',
     var2short('STOP_DATE')      => '($a->{StopDate}  cmp $b->{StopDate})',
     var2short('NUM_GRANULES')   => '($a->{NGranules} <=> $b->{NGranules})',
     );

#------------------------------------------------------------
# new
#------------------------------------------------------------
# I<Constructor.>

sub new {
    my ($class) = @_;
    my $self = $class->SUPER::new(Prefix =>"dset.sortkey",
				  SeqLen => 3,
				  KeyMap => \%Keys);
    $self->sequence(var2short('DATA_CENTER_ID'),
		    var2short('DATASET_ID'));
    $self;
}

#------------------------------------------------------------
# appropriate PREFIX
#------------------------------------------------------------
# I<Constructor.>
#
#    1. Try to get from current set of CGI tags.  If none...
#    2. Try to load "last sort" from session state.  If none...
#    3. Try to load user prefs from session state.  If none...
#    4. Create empty object
#
# The PREFIX is the bit before the ".0", etc.

sub appropriate {
    my $class = shift;
    my $prefix = (shift||'');
    my $self;
    my $case;
        
    if ($REQ->param("${prefix}dset.sortkey.0")) {                    # case 1
	$self = $class->from_CGI(Prefix=>$prefix);	
#	debug "case = 1";
    }
    if (!$self && 
	($self = $SESSION->state->{"LastDatasetSort"})) {            # case 2
#	debug "case = 2";
    }
    if (!$self &&                                                    # case 3
	($self = eval {$SESSION->state->{"User"}->display->dataset_sort})) {
#	debug "case = 3";
    }
    if (!$self) {
	$self = $class->new;
#	debug "case = 4";
    }
    $self;
}


#============================================================
#============================================================

package IW::Sort::Granule;

=head2  IW::Sort::Granule

A subclass of IW::Sort, specifically for sorting granule objects.
The prefix is appended with "gran."; hence, the CGI tags are:

    gran.sortkey.0
    gran.sortkey.1
    gran.sortkey.2

=cut

use IU;
use IU::Log;
use IW::App;
use IW::Babel;
use strict;

@IW::Sort::Granule::ISA = qw(IW::Sort);

# Map sort keys to tests:
# WARNING!!! THE "NAMES" MUST AGREE WITH THOSE IN IW::UserDisplay!!!
my %Keys2 = 
    (
     var2short('DATA_CENTER_ID') => '($a->{ArchiveId} cmp $b->{ArchiveId})',
     var2short('DATASET_ID')     => '($a->{DatasetId} cmp $b->{DatasetId})',
     var2short('GRANULE_ID')     => '($a->{Id}        cmp $b->{Id})',
     var2short('START_DATE')     => '($a->{StartDate} cmp $b->{StartDate})',
     var2short('STOP_DATE')      => '($a->{StopDate}  cmp $b->{StopDate})',
     );

#------------------------------------------------------------
# new
#------------------------------------------------------------
# I<Constructor.>

sub new {
    my ($class) = @_;
    my $self = $class->SUPER::new(Prefix =>"gran.sortkey",
				  SeqLen => 3,
				  KeyMap => \%Keys2);
    $self->sequence(var2short('DATASET_ID'),
		    var2short('START_DATE'));
    $self;
}

#------------------------------------------------------------
# appropriate PREFIX
#------------------------------------------------------------
# I<Constructor.>
#
#    1. Try to get from current set of CGI tags.  If none...
#    2. Try to load "last sort" from session state.  If none...
#    3. Try to load user prefs from session state.  If none...
#    4. Create empty object
#
# The PREFIX is the bit before the ".0", etc.

sub appropriate {
    my $class = shift;
    my $prefix = (shift||'');
    my $self;
    my $case;
        
    if ($REQ->param("${prefix}gran.sortkey.0")) {                    # case 1
	$self = $class->from_CGI(Prefix=>$prefix);	
#	debug "case = 1";
    }
    if (!$self && 
	($self = $SESSION->state->{"LastGranuleSort"})) {            # case 2
#	debug "case = 2";
    }
    if (!$self &&                                                    # case 3
	($self = eval {$SESSION->state->{"User"}->display->granule_sort})) {
#	debug "case = 3";
    }
    if (!$self) {
	$self = $class->new;
#	debug "case = 4";
    }
    $self;
}

#------------------------------------------------------------
1;

