package IW::SrchForm::AbstractWidget;

=head1 NAME

IW::SrchForm::AbstractWidget - Abstact class for search form widgets.

=head1 DESCRIPTION

Outline for interaction of search form widgets with the Search form and query
object.

Subclasses must override the following methods:

   sub name (returns STRING) - the name of the widget.

Subclasses may override the following methods:

   sub types (returns ARRAYREF of STRINGS) - defines the sub types of 
      the widget.

   sub update_widget - do any widget specific updates.

   sub draw_widget - do display for the widget, if any.

=cut

use strict;

use IU;

use IW::App; # For $HTML.

#---------------------------------------------------------------------
#---------------------------------------------------------------------
# Abstract methods: Child classes must implement.
#---------------------------------------------------------------------
#---------------------------------------------------------------------

#------------------------------------------------------------
# name
#------------------------------------------------------------
# Abstract:
# Unique name for this widget.

sub name { puke "abstract method must override" }

#---------------------------------------------------------------------
#---------------------------------------------------------------------
# Generic methods: Child classes may override these:
#---------------------------------------------------------------------
#---------------------------------------------------------------------

#------------------------------------------------------------
# type
#------------------------------------------------------------
# Get/set sub type of widget, override if your widget has 
# more than one subtype and which is stored in the QUERY object:

sub type {

   my ($self,$value) = @_;

   $self->{Type} = $value if (defined($value));
   my $type = $self->{Type} || ${$self->types}[0];
   return $type;
}

#---------------------------------------------------------------------
# types
#---------------------------------------------------------------------
# Generic: child class may override. 
# List of widget sub types

sub types { 

   my @types = ('ABC');
   return \@types;
}

#------------------------------------------------------------
# update_widget EVENT
#------------------------------------------------------------
# Generic widget, override it in your subclasses,
# Do any widget specific updates i.e. handling of Events
# and processing of CGI tags.

sub update_widget { 1 }

#-----------------------------------------------------------
# draw_widget
#-----------------------------------------------------------
# Draw relevant part of the search form.
# Generic widget, override it in your subclasses.

sub draw_widget { 1 }

#------------------------------------------------------------
#------------------------------------------------------------
# Public or external methods:
#------------------------------------------------------------
#------------------------------------------------------------

#------------------------------------------------------------
# new REFWIDGETBOX
#------------------------------------------------------------
sub new {

   my $class = shift;                   # Get class.
   my $self = {};                       # Start with empty hash.

   my $widgetbox = shift;          # Ref to the Widget Box
   $self->{StickyNote} = $widgetbox->{StickyNote}; # Ref to StickyNote Widget. 
   $self->{ParentMode} = $widgetbox->{ParentMode}; # Mode that created widget. 
   # ref to the request_popup method in ModeABSTRACTSRCH.pm 
   $self->{Callback2request_popup} = $widgetbox->{Callback2request_popup}; 
   bless $self,$class;
   return ($self);
}

#------------------------------------------------------------
# update
#------------------------------------------------------------
# Update relevant part of the query object.

sub update {

   my $self = shift;
   my $event = $REQ->param($self->name.'event') || ''; 

   $self->{StickyNote}->pushed_widget($self->name) if $event;
  
   # Enable scrolling; 
   $HTML->need_scrollpoint($self->name) if ($event);

   if ($event eq 'TOGGLE') {

      my $newtype = $REQ->param($self->name.'newtype');
      $self->prev_type($self->type); # save previous type.
      $self->type($newtype);

   } else {

      $self->type($REQ->param($self->name.'type'));

   }

   $self->update_widget($event);
1;
}

#------------------------------------------------------------
# draw 
#------------------------------------------------------------
# draw code common to all widgets.

sub draw {
   my ($self) = @_;

   # print sticky notes:
   $self->{StickyNote}->draw($self->name);
   # Save widget type on the form:
   $HTML->hidden($self->name."type" => $self->type);
   # Draw the widget:
   $self->draw_widget;
}

#------------------------------------------------------------
# prev_type 
#------------------------------------------------------------

sub prev_type {

   my ($self,$prev) = @_;
   $self->{PREVTYPE} = $prev if (defined($prev));
   $self->{PREVTYPE}; 
}

#------------------------------------------------------------
# msg
#------------------------------------------------------------
# add message to the StickNote:

sub msg {

   my ($self,$msg) = @_;

   # Setting a message in the stickynote will also cancel
   # all popup requests.  This is so the user gets
   # a chance to see the message - if we didn't cancel 
   # the user would  be shown a new page and never 
   # have a chance to see the message.
   &{$self->{Callback2request_popup}}('CANCEL');
   $self->{StickyNote}->add_message($self->name,$msg);
1;
}

#------------------------------------------------------------
# err 
#------------------------------------------------------------
# add error to the StickNote:

sub err {

   my ($self,$err) = @_;

   # Setting a error in the stickynote will also cancel
   # all popup requests.  This is so the user gets
   # a chance to see the error - if we didn't cancel
   # the user would  be shown a new page and never
   # have a chance to see the error:
   &{$self->{Callback2request_popup}}('CANCEL');
   $self->{StickyNote}->add_error($self->name,$err);
1;
}


#------------------------------------------------------------
# topmsg
#------------------------------------------------------------
# add message to the StickNote to be displayed at top of page:

sub topmsg {

   my ($self,$msg) = @_;
   # Setting a message in the stickynote will also cancel
   # all popup requests.  This is so the user gets
   # a chance to see the message - if we didn't cancel
   # the user would  be shown a new page and never
   # have a chance to see the message.
   &{$self->{Callback2request_popup}}('CANCEL');
   $self->{StickyNote}->add_message_to_top($msg);
1;
}

#------------------------------------------------------------
# toperr
#------------------------------------------------------------
# add error to the StickNote to be displayed at top of page:

sub toperr {

   my ($self,$err) = @_;
   # Setting a error in the stickynote will also cancel
   # all popup requests.  This is so the user gets
   # a chance to see the error - if we didn't cancel
   # the user would  be shown a new page and never
   # have a chance to see the error.
   &{$self->{Callback2request_popup}}('CANCEL');
   $self->{StickyNote}->add_error_to_top($err);
1;
}

#------------------------------------------------------------
# print_section_title TITLEHTML
#------------------------------------------------------------

sub print_section_title {
    my ($self, $title) = @_;

    $HTML -> FONT(SIZE=>5) -> B;
    print $title;
    $HTML -> _B -> _FONT;
    $HTML -> BR;
}

#------------------------------------------------------------
# pretty_types
#------------------------------------------------------------
# Override in the child widgets if you want the pretty names
# (i.e. what the user sees to be different from the internal 
# stored value.

sub pretty_types {

   my $self = shift;
   my $type = shift; 

   return $type;
}

#------------------------------------------------------------
# print_toggle_buttons
#
# Arguments to print_toggle_buttons are:
# Array of options types for the widget.
# Number of columns to output.
#------------------------------------------------------------
sub print_toggle_buttons {

   my $self = shift;
   my $opt_types = shift;
   my $num_cols = shift || '3';

   # Divide up into rows:
   for (my $i = 0; $i < int(@$opt_types); $i += $num_cols) {
      $HTML -> TR;

      # Fill in row with columns
      for (my $j = 0; $j < $num_cols; $j++) {

         $HTML -> TD(NOWRAP=>undef, WIDTH=>'1%');

         # Don't print anything if just padding out the row.
         if ($i+$j < int(@$opt_types)) {

            my $typeinfo = $$opt_types[$i+$j];

            # Get URL for switching option type (if any):
            my $src = undef;
            my $alt = undef; 
            # Get the source for toggle button, on or off.
            my $pretty = $self->pretty_types($$opt_types[$i+$j]);
            if ($$opt_types[$i+$j] ne $self->type) {
               $src = "${IU::DocsUrl}icons/toggle-off.gif";
               $alt = "Switch to $pretty"; 
            }
            else {
               $src = "${IU::DocsUrl}icons/toggle-on.gif";
               $alt = "Currently selected: $pretty";
            }
            # Button:
            $HTML-> msubmit_button(QUERY =>
                                   ['mode'  => $self->{ParentMode},
                                    'submode'      => 'UPDATE',
                                    $self->name.'event'  => 'TOGGLE',
                                    $self->name.'newtype'=>$$opt_types[$i+$j]
                                   ],
                                   ALT => $alt,
                                   SRC => $src);
            # Label:
            # Print a space between the toggle button and label.
            print " ";
            $HTML -> B -> t($pretty) -> _B;
         }
         $HTML -> _TD;
      }
      $HTML -> _TR;
   }
}

#------------------------------------------------------------
1;

