package IW::SrchForm::AbstractWidgetBox;

=head1 NAME

IW::SrchForm::AbstractWidgetBox - abstract container for widgets.
              Creates the widgets and sticky note, handles
              updating each widget, and draws the search form,
              sticky note and widgets.

                abstract methods (overridden by child classes):

                  widgets - list of widgets used by this widget box.

                override methods (can be overridden by child if needed)
                 
                  init - do child specific setup.

                public methods:

                  new - constructs the widget box and the widgets
                     it uses.

                  update - tells each widget in the widget box to 
                     update and installs the updated QUERY object 
                     into the session.

                  draw - start search form, draw top sticky note,
                     tell each widget to draw, end search page.

=cut

use strict;
use IU; # For puke.
use IU::Query; # For $QUERY.
use IW::App; # For $HTML.
use IW::Util qw(:url :output);

#------------------------------------------------------------
#------------------------------------------------------------
# Abstract methods: child classes must implement these: 
#------------------------------------------------------------
#------------------------------------------------------------

#------------------------------------------------------------
# widgets SRCHTYPE
#------------------------------------------------------------
# Abstract
# Return array of widget names to be used on the form 
# based on the search type arg passed in.
# Note, if no search type is specified then array all widgets
# used in widget box should be returned.

sub widgets { puke "Abstract method must override in child class" }

#------------------------------------------------------------
#------------------------------------------------------------
# Public or external methods:
#------------------------------------------------------------
#------------------------------------------------------------

#------------------------------------------------------------
# new
#------------------------------------------------------------
# Create and return and instance of this class

sub new {
   my ($class, %params) = @_;

   # Basic initialization; make it an object so we can call widgets() later.
   my $self = {};
   $self->{ParentMode} = $params{ParentMode};
   $self->{Callback2request_popup}   = $params{Callback2request_popup};
   bless $self, $class;

   # Do any subclass specific setup:
   $self->init;

   # Initialize the query. We do this now, and no one else has to.
   $self->lookup_query;

   # Create sticky note 
   $self->{StickyNote} = IW::SrchForm::StickyNote->new;

   # Get list of all widgets used and then create and store them in the object. 
   foreach my $widget ($self->widgets) {

      # Create the widget and store it in the object:
      my $module = "IW::SrchForm::$widget";
      $self->{$widget} = $module->new($self);
      # We shouldn't have problems creating the widgets, but
      # we need to know if we did.
      puke "Error creating widget $widget" unless defined($self->{$widget});
   }
   $SESSION->{Navigator}->{SrchType} = 
     $SESSION->state->{LastSrchType} =
       $self->{SrchTypeWidget}->type
         if defined $self->{SrchTypeWidget};
   $SESSION->{Navigator}->{SrchType} = 
     $SESSION->state->{LastSrchType} =
       "INVENTORY_SEARCH"
         unless defined $SESSION->{Navigator}->{SrchType};

   $self;
}

#---------------------------------------------------------------------
# update 
#---------------------------------------------------------------------

sub update {

   my $self = shift;
   $self->{SrchTypeWidget}->update; 

   foreach my $widget ($self->widgets($self->{SrchTypeWidget}->type)) {
      $self->{$widget}->update;
   }

   # install query into the session.
   # Note that this should be the only install done in the search form.
   $self->install_query;
}

#---------------------------------------------------------------------
# draw
#---------------------------------------------------------------------

sub draw {

   my $self = shift;
   
   $APP->start_page(SuperTitle => 'Search Creation',
                    Title      => $self->{SrchTypeWidget}->title,
                    Context    => $self->{SrchTypeWidget}->context,
                    Search     => $self->{SrchTypeWidget}->link);

   $HTML -> script_FORM(ACTION  => script_action().'#SCROLL');

   $HTML -> msubmit_flag;

   # Print message at top of page:
   $self->{StickyNote}->draw_top;

   foreach my $widget ('SrchTypeWidget',
                       $self->widgets($self->{SrchTypeWidget}->type) ) {
      $self->{$widget}->draw;
   }

   # End the form
   $HTML -> _script_FORM;

   # End page
   $APP->end_page;
   1;
}

sub init {1}

sub lookup_query {$QUERY = IU::Query->lookup}

sub install_query {$QUERY->install}

1;
