package IW::SrchForm::DepValidWidget;

=head1 NAME

IW::SrchForm::DepValidWidget - layout dependent valids widget and handle updates.

=head1 DESCRIPTION

This displays the dependent valids selection widget and handle update events.

=cut

use strict;

use IU;
use IU::Category;
use IU::Query;
use IW::DepValidWidget::Base; # Depvalid subtype widgets
use IW::App;

use IW::SrchForm::AbstractWidget;
use vars qw(@ISA);
@ISA = qw(IW::SrchForm::AbstractWidget);

#---------------------------------------------------------------------
# name
#---------------------------------------------------------------------

sub name {

   return 'DepValid';

}

#------------------------------------------------------------
# type
#------------------------------------------------------------
# subtype of the widget.

sub type {

   my ($self,$value) = @_;

   $QUERY->depvalidtype($value) if (defined($value));

   # for backwards compatability convert old type of "Lookup" to "Valids".
   $QUERY->depvalidtype('Valids') if ($QUERY->depvalidtype eq 'Lookup');

   my $type = $QUERY->depvalidtype || 'Discipline';
   $type = 'Valids' if (${IU::DisableByDiscipline} eq 'YES');

   return $type

}

#---------------------------------------------------------------------
# types
#---------------------------------------------------------------------

sub types {

   my @types = ('Discipline',
                'Valids');

   # Disable "By Discipline" option if the flag is set in SETUP:
   if (${IU::DisableByDiscipline} eq 'YES') {
      @types = ('Valids');
   }

   return \@types;
}

#------------------------------------------------------------
# pretty_types
#------------------------------------------------------------
# Override of the parent widget to allow the pretty names
# (i.e. what the user sees to be different from the internal
# stored value.

sub pretty_types {

   my $self = shift;
   my $type = shift;

   my %prettytypes = ('Valids'     => 'By Categories/Attributes',
                      'Discipline' => 'By Discipline');

   return $prettytypes{$type};
}

#------------------------------------------------------------
# update_widget
#------------------------------------------------------------

sub update_widget {

   my $self = shift;
   my $event = shift;

   $self->type('Valids') 
      unless ($QUERY->search_type eq 'INVENTORY_SEARCH');

   # Only process p_valid tags if p.datasetlist is present.
   # p.datasetlist tells us that a dataset menu was submitted.
   if (defined($REQ->param('p.datasetlist'))) {;
      my @items = map { _clean_valid($_) } $REQ->param('p_valid');
      my @cleaned_items = (grep {!/^_+$|^\s*$/} @items) ;
      $QUERY->dataset_ids(\@cleaned_items);
   }

   if ($event eq 'EDIT') {

      # Edit is handled by pop Modes so the only processing done
      # here is to set the NEXT mode to go to and CGI tags:
      # Note: While this is almost exactly like a SELECT event
      # the category param tag is slightly different to prevent
      # cross talk between this and the SELECT element. 
      &{$self->{Callback2request_popup}}('SLCTALL',
                           Tags=>{'submode'    => 'CRITFORM',
                                  'p.category' => $REQ->param('category')});
   }
   elsif ($event eq 'DELETE') {

      my $cat = $REQ->param('category');
      $QUERY->criteria($cat, undef);
      my $prettycat = category2pretty($cat);
      $self->msg("All criteria for $prettycat have been deleted.");
   }
   elsif ($event eq 'DEFINITION') {

      # A request for info on a valid has been received
      my @selections = $REQ->param('p_valid');
      &{$self->{Callback2request_popup}}('MOREINFO',
                            Tags=>{'submode'    => 'CGI',
                                   'category' =>'DATASET_ID',
                                   'info' => 'DEF',
                                   'p.valid' => \@selections });
   } elsif ($event eq 'SELECT') {

      # Select is handled by pop Modes so the only processing done
      # here is to set the NEXT mode to go to and CGI tags:
      my $cat = $REQ->param('p.category') || undef;
      if ($cat) {
         &{$self->{Callback2request_popup}}('SLCTALL',
                           Tags=>{'submode'    => 'CRITFORM',
                                  'p.category' => $REQ->param('p.category')});
      }
      else {

         $self->err("please choose a category then click the ".
                    "'Choose Keywords for >>' button.");
      }
   } elsif ($event eq 'ALLDSETS') {

      # Switch to Valids widget and popup the data set list.
      &{$self->{Callback2request_popup}}('SLCTALL',
                         Tags=>{'submode'    => 'CRITFORM',
                                'p.category' => 'DATA SET'});
   }

   if (defined($REQ->param('discipline'))) {
    
      my @missions = map {$_} $REQ->param('mission');
      # Replace with newmission if user made new selection via the
      # non-javascript CGI form: 
      @missions = ($REQ->param('newmission')) if $REQ->param('newmission');
      $QUERY->missions(\@missions);
   }

   # Process data set lookup requests:

   my $msg = IW::DepValidWidget::Base->lookup(type=>'Lookup',
      ParentMode=>$self->{ParentMode})->update($event);
   $self->msg($msg) if defined($msg);

   # Get and update the DepValidWidget sub type:
   $msg = IW::DepValidWidget::Base->lookup(type=>$self->type,
      ParentMode=>$self->{ParentMode})->update($event);

   $self->msg($msg) if defined($msg);

   # set scroll point when coming from a depvalids popup page:
   if (($REQ->param('mode') eq 'SLCTDEPVAL') || 
       ($REQ->param('mode') eq 'SLCTALL'   )) {

      $HTML->need_scrollpoint($self->name);
      $self->{StickyNote}->pushed_widget($self->name);
   }

1;
}

#------------------------------------------------------------
# _clean_valid
#------------------------------------------------------------
#
sub _clean_valid {
    my $str = shift;
    $str =~ s/^\*\s?//;
    $str;
}


#-----------------------------------------------------------------------
# draw_widget
#------------------------------------------------------------
#

sub draw_widget {

   my $self = shift;

   my $C_HI     = '#D0D0D0';

   #
   # change the header text according to search type
   #
   my %headers = (
      'GUIDE_SEARCH'     => 'Choose Keywords for One or More Categories',
      'GRANULE_SEARCH' => 'Choose Data Set and Data Granule ID',
      'LOCAL_GRANULE_SEARCH' => 'Choose Data Set and Local Granule ID',
      'DIRECTORY_SEARCH' => 'Choose Keywords for One or More Categories',
      'INVENTORY_SEARCH' => 'Choose Keywords for One or More Categories',
      'BROWSE_SEARCH' =>'Choose at Least One of these Browse Data Sets');

   my $header_text = ($self->type eq 'Valids') ? 
                     $headers{$QUERY->search_type} : 'Choose Data Sets';

   # Put some space between the widgets.

   $HTML->P;

   # Start the table.

   $HTML -> TABLE(WIDTH=>'100%', CELLPADDING=>4, CELLSPACING=>1,
                  CLASS=>'srchwidgethdr');

   # First row, title:
   $HTML-> TR -> TD(ALIGN=>'LEFT');
   $HTML-> FONT(SIZE=>'+2')-> t_B($header_text)-> _FONT;
   $HTML->_TD ->TD(ALIGN=>'RIGHT');
   if ($QUERY->search_type eq 'INVENTORY_SEARCH') {
      IW::DepValidWidget::Base->lookup(type=>'Lookup',
         ParentMode=>$self->{ParentMode})->draw;
   } 
   $HTML->_TD ->_TR;

   # Second row, the subwidget area:
   $HTML->TR->TD(CLASS=>'srchwidgetbody',COLSPAN=>2);
   IW::DepValidWidget::Base->lookup(type=>$self->type,
      ParentMode=>$self->{ParentMode})->draw;
   $HTML->_TD->_TR;

   # Third row, the toggle buttons:
   $self->print_toggle_buttons($self->types,'2')
      if (($QUERY->search_type eq 'INVENTORY_SEARCH') &&
          (${IU::DisableByDiscipline} ne 'YES'));

   $HTML -> _TABLE;
   1;
}



#-----------------------------------------------------------------
1;
