package IW::SrchForm::OptionsWidget;

=head1 NAME

IW::SrchForm::OptionsWidget - layout options widget and handle updates.

=head1 DESCRIPTION

Displays the options selection widget and handles update events.

=cut

use strict;

use IU::Event; # For ims_event logging
use IU::Query; # For $QUERY.

use IW::App; # For $HTML
use IW::Util qw(:url); # For gloss_url function.

use IW::SrchForm::AbstractWidget;
use vars qw(@ISA);
@ISA = qw(IW::SrchForm::AbstractWidget);

#---------------------------------------------------------------------
# name
#---------------------------------------------------------------------

sub name {

   return 'Options';
}
#------------------------------------------------------------
# type
#------------------------------------------------------------
# subtype of the widget.

sub type {

   my ($self,$value) = @_;

   $QUERY->optstype($value) if (defined($value));

   my $type = 'All Options';

   return $type
}

#---------------------------------------------------------------------
# types 
#---------------------------------------------------------------------

sub types {

   my @types = ('All Options');
   return \@types;

}

#------------------------------------------------------------
# update_widget
#------------------------------------------------------------
sub update_widget {

   my $self = shift;
   my $event = shift;

   # set next mode to go to after update.
   &{$self->{Callback2request_popup}}('METADATA') if ($event eq 'CUSTOMIZE'); 

   # Set this if the browse checkbox present flag 'p.browsebox' is on
   $QUERY->browse_only($REQ->param('p.browseonly') ? 1 : 0)
      if ($REQ->param('p.browsebox'));

   # ----------------------------------------
   # If 'p.day_night' is set to 'Day/Night' then don't add it to the Query
   # object, but set it to 'D' for 'Day Only' and 'N' for 'Night Only'
   # ----------------------------------------
   my $day_night_flag = $REQ->param('p.day_night');
   if ($day_night_flag) {
      if ($day_night_flag eq 'Day Only') { $QUERY->day_night ('D')}
      elsif ($day_night_flag eq 'Night Only') { $QUERY->day_night ('N')}
      else { $QUERY->day_night (undef)}
   }

   # ----------------------------------------
   # read and set query name
   # ----------------------------------------
   my $name = $REQ->param('p.name');
   if ($name) {
      $QUERY->name($name);
   }
   # ----------------------------------------
   # read and validate max # of granules
   # ----------------------------------------
   my $maxgrans  = $REQ->param('p.maxgrans');
   if ($maxgrans) {
      $maxgrans =~ s/\s//g;
      if ($maxgrans =~ /\D/) {
         $self->err("Maximum number of data granules must be a positive ".
                    "whole number, you entered: \"$maxgrans\".");
      }
      elsif ($maxgrans > $IU::Query::MAX_GRANS_MAX) {
         $self->err("Maximum number of data granules must be less than ".
                    "$IU::Query::MAX_GRANS_MAX, you entered: \"$maxgrans\".");
      }
      else {
         $QUERY->max_granules($maxgrans);
      }
   }

   # ----------------------------------------
   # read and validate max # of guide docs
   # ----------------------------------------
   my $maxguides = $REQ->param('p.maxguides');
   if ($maxguides) {
      $maxguides =~ s/\s//g;
      if ($maxguides =~ /\D/) {
         $self->err("Maximum number of guide documents must be a positive ".
                    "whole number, you entered: \"$maxguides\".");
      }
      else {
         $QUERY->max_guides($maxguides);
      }
   }
   # ----------------------------------------
   # read and validate time limit
   # ----------------------------------------
   my $maxtime   = $REQ->param('p.maxtime');
   if ($maxtime) {
      $maxtime =~ s/\s//g;
      if ($maxtime =~ /\D/) {
         $self->err("Time limit must be a positive whole number of minutes, ".
                    "you entered: \"$maxtime\".");
      }
      else {
         $QUERY->max_time($maxtime);
      }
   }
   # Set this if the include_nongeo checkbox present flag 'p.nongeobox' is on
   $QUERY->include_nongeo($REQ->param('p.include_nongeo') ? 1 : 0)
      if ($REQ->param('p.nongeobox'));
   # Set this if the include_nontime checkbox present flag 'p.nontimebox' is on
   $QUERY->include_nontime($REQ->param('p.include_nontime') ? 1 : 0)
      if ($REQ->param('p.nontimebox'));
}


#------------------------------------------------------------
# draw_widget
#------------------------------------------------------------

sub draw_widget {

   my $self = shift;

   # Start by lying to ourselves about what was posted:
   $self->_set_CGI;

   # Put some space between the widgets.
   $HTML->P;

   $HTML -> TABLE(WIDTH=>'100%', BORDER=>0, CELLPADDING=>4, 
                  CELLSPACING=>1,CLASS=>'srchwidgethdr');
   #
   # ROW 1...
   #

   $HTML -> TR ->TD;

   $self->print_section_title("Choose Additional Options (not required)");  

   $HTML ->_TD->_TR; 

   #
   # ROW 2...
   #

   $HTML -> TR -> TD(CLASS=>'srchwidgetbody');
   $HTML -> UL;
   $self->_print_GRANULE_LIMIT;
   $self->_print_BROWSE_ONLY;
   $self->_print_TIME_LIMIT;
   $self->_print_GUIDE_LIMIT;
   $self->_print_ALL_METADATA;
   $self->_print_include_nongeo;
   $self->_print_include_nontime;
   $self->_print_DAY_NIGHT;
   $self->_print_QUERY_NAME;
   $HTML -> _UL;
   $HTML -> _TD -> _TR;
   $HTML -> _TABLE;
   1;
}

#------------------------------------------------------------
# _set_CGI
#------------------------------------------------------------
# Get any existing query values, and use them to fake our "input".
#
sub _set_CGI {
   my $self = shift;

   $REQ->param('p.browseonly'=> $QUERY->browse_only);
       
   # Day/Night Flag - If no day_night Query object then it's assumed to be 
   #  DAY/NIGHT, else compare it against 'D' ( Day Only) or 'N' ( Night Only)
   $REQ->param('p.day_night' =>((defined($QUERY->day_night)) 
       ? (($QUERY->day_night eq 'D' ) 
       ? 'Day Only' : 'Night Only' )   : 'DAY/NIGHT' ) );
    
   $REQ->param('p.maxgrans'  => $QUERY->max_granules);
   $REQ->param('p.maxguides' => $QUERY->max_guides);
   $REQ->param('p.maxtime'   => $QUERY->max_time);
   $REQ->param('p.name'      => $QUERY->name);
   1;
}
#------------------------------------------------------------
# _print_GRANULE_LIMIT
#------------------------------------------------------------
# I<Instance method.>
# Subform for maximum number of granules.
#
sub _print_GRANULE_LIMIT {
   my $self = shift;

   return unless $QUERY->uses_opt('GRANULE_LIMIT');

   $HTML -> LI;
   $HTML -> t("Return a maximum of ")
         -> input_text(Size    =>5, 
                       Name    =>"p.maxgrans",
                       Default =>$IU::Query::MAX_GRANS_DEFAULT)
         -> t(' ') -> hypertext("data granules", gloss_url("GRANULE"))
         -> t(" per data set ",
            "(Range: $IU::Query::MAX_GRANS_MIN - $IU::Query::MAX_GRANS_MAX).");
   1;
}
#------------------------------------------------------------
# _print_BROWSE_ONLY
#------------------------------------------------------------
# I<Instance method.>
# Subform for "granules with browse products only".
#
sub _print_BROWSE_ONLY {
   my $self = shift;

   return unless $QUERY->uses_opt('BROWSE_ONLY');

   $HTML -> LI;
   # Indicate that the check box was on the page with this hidden tag
   $HTML->hidden('p.browsebox',1);
   $HTML -> checkbox(Name=>"p.browseonly", Choice=>1);

   $HTML -> t(" Only return data granules which have ")
         -> hypertext("browse", gloss_url("BROWSE"))
         -> t(" products.");
   1;
}
#------------------------------------------------------------
# _print_TIME_LIMIT
#------------------------------------------------------------
# I<Instance method.>
# Subform for maximum number of minutes to allow searches to run.
#
sub _print_TIME_LIMIT {
   my $self = shift;

   return unless $QUERY->uses_opt('TIME_LIMIT');

   $HTML -> LI;
   $HTML -> t("Allow searches to run for a maximum of ")
         -> input_text(Size    =>5, 
                       Name    =>"p.maxtime",
                       Default =>$IU::Query::MAX_TIME_DEFAULT)
         -> t(' ') -> t("minute(s) ");
   1;
}
#------------------------------------------------------------
# _print_include_nongeo
#------------------------------------------------------------
# I<Instance method.>
# Subform for printing the Include Non-Spatial checkbox.
#
sub _print_include_nongeo {
   my $self = shift;

   return unless ($QUERY->uses_opt('INCLUDE_NONGEO') && $QUERY->show_nongeo); 
   # For June 2000 testing by ECS, a requirement is to test if the
   # Setup allows this button to be displayed.  Therefore, if some
   # installations desire, this flag can be turned off.
   # Using Query object to encapsulate the SETUP check in one place.
   # _include_nongeo method returns undef if SETUP flag is not set to 'Y'

   my $queryNonGeo = $QUERY->include_nongeo();
   # Indicate that the check box was on the page with this hidden tag
   $HTML->hidden('p.nongeobox',1);
   $HTML -> LI;
   $HTML -> checkbox(Name=>"p.include_nongeo",
                     Value=>$queryNonGeo,
                     Choice=>1)
           -> t(" Include Non-Spatial granules in search results "); 
   1;
}

#------------------------------------------------------------
# _print_include_nontime
#------------------------------------------------------------
# I<Instance method.>
# Subform for printing the Include Non-Temporal checkbox.
#
sub _print_include_nontime {
   my $self = shift;

   return unless ($QUERY->uses_opt('INCLUDE_NONTIME') && $QUERY->show_nontime);
   # For June 2000 testing by ECS, a requirement is to test if the
   # Setup allows this button to be displayed.  Therefore, if some
   # installations desire, this flag can be turned off.
   # Using Query object to encapsulate the SETUP check in one place.
   # include_nontime method returns undef if SETUP flag is not set to 'Y'

   my $queryNonTime = $QUERY->include_nontime();
   # Indicate that the check box was on the page with this hidden tag
   $HTML->hidden('p.nontimebox',1);
   $HTML -> LI;
   $HTML -> checkbox(Name=>"p.include_nontime",
                     Value=>$queryNonTime,
                     Choice=>1)
         -> t(" Include Non-Temporal granules in search results ");
   1;
}
#------------------------------------------------------------
# _print_DAY_NIGHT
#------------------------------------------------------------
# I<Instance method.>
# Subform for Day/Night Flag.
#

sub _print_DAY_NIGHT {
   my $self = shift;

   return unless $QUERY->uses_opt('DAY_NIGHT');

   $HTML -> LI;
   $HTML  -> t(" Only return data granules which were retrieved during the ")
           -> select_one(Name   =>"p.day_night",
                        Choices=>['Day/Night', 'Day Only', 'Night Only'],
                        Default=> 'Day/Night');
   1;
}

#------------------------------------------------------------
# _print_ALL_METADATA
#------------------------------------------------------------
# I<Instance method.>
# Subform for all metadata checkbox.
#
sub _print_ALL_METADATA {
   my $self = shift;

   return unless ((uc($IU::CustomizeMetadata) eq 'YES') &&
      ($QUERY->search_type eq 'INVENTORY_SEARCH')); 
   my $selection;
   my @metadata;

   # -----------------------------------------------------------------
   # Check on metadata flag to display the appropriate message.
   # -----------------------------------------------------------------
   my $all_metadata = $QUERY->all_metadata ();

   # User has customized results metadata, display list of attributes.
   if (defined $QUERY->{Categories}) { 
      if (defined $QUERY->{Categories}{'RESULT_ATTRIBUTES'}) {
        $selection = "SELECTED";
        @metadata = $QUERY->result_attributes ();
        push (@metadata, "STANDARD CORE ATTRIBUTES");
      }
      else {
        $selection = "DEFAULT";
      }
   }
   # User selected all metadata.
   elsif ($all_metadata eq "Y") {
      $selection = "ALL";
   }
   # User selected default metadata.
   else {
      $selection = "DEFAULT";
   }
   ims_event 'METADATA', Selection=>$selection, Items=>[@metadata];
    
   $HTML -> LI;
   print " Return <B>$selection</B> metadata in search results ";
   # Customize button:
   $HTML -> msubmit_button(CAPTION => "Customize",
                            ALT  =>"Customize",
                            QUERY => ['mode'   =>$self->{ParentMode},
                                      'submode'=>'UPDATE',
                                      $self->name.'event'=>'CUSTOMIZE']);
   $HTML ->BR if (int(@metadata));
   foreach my $name (@metadata) {
      print "$name";
      $HTML ->BR;
   }
   1;
}

#------------------------------------------------------------
# _print_GUIDE_LIMIT
#------------------------------------------------------------
# I<Instance method.>
# Subform for maximum number of guide documents.
#
sub _print_GUIDE_LIMIT {
   my $self = shift;

   return unless ($QUERY->search_type eq 'GUIDE_SEARCH');
   $HTML -> LI;
   $HTML -> t("Return a maximum number of ")
         -> input_text(Size    =>5, 
			Name    =>"p.maxguides",
                        Default =>$IU::Query::MAX_GUIDES_DEFAULT)
         -> t(' ') -> t("detailed documents ");
   1;
}
#------------------------------------------------------------
# _print_QUERY_NAME
#------------------------------------------------------------
# I<Instance method.>
# Subform for query name.

sub _print_QUERY_NAME {
   my $self = shift;

   return unless $QUERY->uses_opt('QUERY_NAME');

   $HTML-> LI;
   $HTML-> tag('NOBR')
        -> t("Name this query: ")
        -> input_text(Name=>"p.name", Size=>30)
        -> tag('_NOBR');
   $HTML-> BR;
   $HTML->t_EM('(will be used in creating a file name when saving the query)');

   1;
}

#---------------------------------------------------------------------
1;
