package IW::StatusBar;

=head1 NAME

IW::StatusBar - a status object with an LED display

=head1 DESCRIPTION

Any activities that want to display a status bar can inherit from 
this class.

=head1 SYNOPSIS

    # Input from a stream:
    my $sb = input $class \*FH;
    
    # Get the HTML to use to indicate a particular current status:
    my $mark = $sb->mark('OK');
    
    # Is it still running?
    $sb->still_running;
    
    # If not still running, did it finish okay?
    $sb->finished_ok;
     
    # Get stuff for displaying it:
    my $icon    = $sb->icon;         # HTML for the icon
     
    # Display it:
    $sb->print_pot;

=cut

use IU;
use IW::App;
use IW::Util qw(:url);
use strict;

use HTML::Stream qw(:funcs);

#------------------------------------------------------------
#
# Globals:
#
#------------------------------------------------------------

# Marks to put at the start of each row in the list:
my $Mark = undef;


#------------------------------------------------------------
# new 
#------------------------------------------------------------
sub new {
    bless {}, shift;
}

#------------------------------------------------------------
# input FILEHANDLE
#------------------------------------------------------------
# I<Constructor:> read from a chunk stream.

sub input {
    my ($class, $fh) = @_;

    # Read the raw data:
    my $chunk = (IU::Chunk->input($fh) || return undef);
    
    # Create the new object, and init it:
    my $self = $class->new->from_chunk($chunk);
    $self;
}

#------------------------------------------------------------
# finished_ok
#------------------------------------------------------------
# The activity is I<not> "still running"... but did it "finish okay"?
# Override this in your subclass!

sub finished_ok {
    my $self = shift;
    1;      # assume okay
}

#------------------------------------------------------------
# from_chunk CHUNK
#------------------------------------------------------------
# Init from a chunk.  Override this in your subclasses!
sub from_chunk {
    my ($self, $chunk) = @_;
    puke 600, "from_chunk should be overridden by class ", ref($self);
}

#------------------------------------------------------------
# icon 
#------------------------------------------------------------
# Return the appropriate icon for this stage of processing.
#
# Returns the HTML for the tiny checkmark, clock, or slashed-circle 
# icon, based on how we're doing.  Does NOT print it!

sub icon {
    my $self = shift;
    my $mark;

    # Get the basic icon:
    if ($self->still_running()) {     # still running
	return $self->mark('RUN');
    }
    elsif ($self->finished_ok()) {    # no longer running, and success
	return $self->mark('OK'); 
    }
    else {                            # no longer running, and failure
        my $status = $self->status;
        # The following are error statuses:
        # 8  Pertinent inventory system unavailable; try again later
        # 9  Bad message; message contains syntax error(s)
        # 11 System error; please try again later
        # 17 Bad message; protocol error
        # 18 System busy; try again later
        # 19 Data center not responding to query, may be experiencing computer
        #    or network problems; try search again later
        # 29 Network connection unexpectedly closed; try search again later
        # 30 Error servicing request; contact User Services
        if ($status == 8 || $status == 9 || $status == 11 ||
            $status == 17 || $status == 18 || $status == 19 ||
            $status == 29 || $status == 30) {
	    return $self->mark('FAIL');
	} else {
	    return $self->mark('NO');
	}
    }
}

#------------------------------------------------------------
# mark STATUS
#------------------------------------------------------------
# I<Class/instance method:> return the HTML for the image.
# Status is one of: OK, RUN, FAIL, NO.

sub mark {
    my ($self, $status) = @_;

    unless ($Mark) {
	$Mark = {};
	my $tag;
	foreach $tag (qw(ok run fail no)) {
	    $Mark->{uc($tag)} = 
		html_tag('IMG', 
			 SRC=>doc_url("icons/status-$tag-sm.gif"), 
			 ALT=>uc($tag));
	}
    }
    $Mark->{uc($status)}
}

#------------------------------------------------------------
# print_icon 
#------------------------------------------------------------
# Print the summary icon.

sub print_icon {
    my $self = shift;
    print $self->icon(@_);
}

#------------------------------------------------------------
# print_pot 
#------------------------------------------------------------
# Print the "LED display" showing our status.

sub print_pot {
    my $self = shift;
    puke 600, "print_pot() should be overriden by ", ref($self);
}

#------------------------------------------------------------
# still_running
#------------------------------------------------------------
# Is the activity still running?

sub still_running {
    my $self = shift;
    puke 600, "still_running should be overridden by class ", ref($self);
}

#------------------------------------------------------------
1;



