package IW::SummaryList;

=head1 NAME

IW::SummaryList - object list for presenting tabulated information

=head1 OUTPUT FORMAT

Each possibility uses a different number of rows and columns,
as based on the user's display preference for summary style.

HORIZONTALLY (1 row, n columns):

         +-------+-------+-------+-------+
         |   a   |   b   |   c   |   d   |
         |       |       |       |       |
         | a-val | b-val | c-val | d-val |
         +-------+-------+-------+-------+

HORIZONTALLY (2 rows, n columns):

         +-------+-------+-------+-------+
         |   a   |   b   |   c   |   d   |
         +-------+-------+-------+-------+
         | a-val | b-val | c-val | d-val |
         +-------+-------+-------+-------+


VERTICALLY (1 row, 1 column):

         +--------------------------+
         | a:  a-val                |
         | b:  b-val                |
         | c:  c-val                |
         | d:  d-val                |
         +--------------------------+

TWO-COLUMN (1 row, 2 columns):

         +-------------+------------+
         | a:  a-val   | c: c-val   |
         | b:  b-val   | d: d-val   |
         +-------------+------------+


=head1 NOTES

This is just a thin abstract extension of IW::ObjectList.
Subclasses can use these methds for setting up tabulated data.

This class absorbs the old "minitable" class.

=head1 WARNING

It is currently expected that all data areas have the same number of
rows and columns!  If you need to override that for a particular entry,
we may need to put in a ROWSPAN/COLSPAN override mechanism.

=head1 AUTHOR

Eryq, Huighes STX Corporation.

=head1 RCS INFO

$Id: SummaryList.pm,v 1.4 1997/09/29 22:17:31 lrosen Exp $

=cut

use IW::ObjectList;
use IW::User;
use IW::App;
use HTML::Stream qw(:funcs);

@ISA = qw(IW::ObjectList);

use strict;


#------------------------------
#
# OVERRIDES
#
#------------------------------

#------------------------------------------------------------
# new PARAMHASH
#------------------------------------------------------------
# I<Override>.  Create a new summary-list object.
# User defaults are examined!

sub new {
    my $class = shift;
    my %params = @_;

    # Construct:
    my $self = $class->SUPER::new(@_);

    # Get display preferences:
    my $display = eval{$SESSION->state->{User}->display}|| new IW::UserDisplay;
    $self->{Display} = $display;       # if subclasses want it!

    # If unspecified, pick good values for the presentation:
    $self->{SummaryStyle}  = $display->summary_style;
    $self->{SummaryMenu}   = $display->summary_menu;
    $self->{PrimaryActSiz} = $display->primary_act_siz;

    # Set up defaults, etc:    
    $self->{MenuStyle} = ($self->{SummaryMenu} eq 'LARGE') ? 'HMENU' : 'VMENU';

    $self;
}

#------------------------------------------------------------
# data_dimensions
#------------------------------------------------------------
# I<Override>.  Returns the rows and columns in the data to be printed.

sub data_dimensions {
    my $self = shift;

    if ($self->{SummaryStyle} eq 'HORIZONTALLY') {
	return(1, int(@{$self->{SummaryCols}}));
    }
    elsif ($self->{SummaryStyle} eq 'VERTICALLY') {
	return(1, 1);
    }
    elsif ($self->{SummaryStyle} eq 'TWO-COLUMN') {
	return(1, 2);
    }
    return(1, 1);
}

#------------------------------------------------------------
# print_data I
#------------------------------------------------------------
# I<Override>.  Print the data for the i'th element.
# Assumes that data have already been set up!

sub print_data {
    my ($self, $i) = @_;
    my $title;
     

    # VERTICAL TABLE...       
    if ($self->{SummaryStyle} eq 'VERTICALLY') {

	# Keys and values (single cell, one line for each key/value pair):
	print "<TD NOWRAP>";
	$HTML -> small;
	foreach $title (@{$self->{SummaryCols}}) {
	    print '<B>', $self->{SummaryHdrs}{$title}, ': </B>';
	    $HTML -> hspace(2);
	    print $self->{SummaryData}{$title}, "<BR>\n";
	}
	$HTML -> _small;
    }

    # TWO-COLUMN TABLE...
    elsif ($self->{SummaryStyle} eq 'TWO-COLUMN') {
	my $col;
	my $nhdrs = int(@{$self->{SummaryCols}});
	my $ncols = 2;
	foreach $col (0..($ncols-1)) {
	    my $start = $col*($nhdrs/$ncols);
	    my $end   = ((1+$col)*($nhdrs/$ncols))-1;

	    print "<TD NOWRAP>";
	    $HTML -> small;
	    my $hdrno;
	    foreach $hdrno ($start..$end) {
		my $title = $self->{SummaryCols}[$hdrno] || last;
		print '<B>', $self->{SummaryHdrs}{$title}, ': </B>';
		$HTML -> hspace(2);
		print $self->{SummaryData}{$title};
		$HTML -> hspace(4) -> BR;
	    }
	    $HTML -> _small;
	}
    }

    # HORIZONTAL TABLE...
    else {

	# Keys & values (two lines in one cell for each key/value pair):
	foreach $title (@{$self->{SummaryCols}}) {
	    print '<td>';
	    $HTML -> small;
	    print '<b>', $self->{SummaryHdrs}{$title}, '</b><br>';
	    print $self->{SummaryData}{$title}, "\n";
	    $HTML -> _small;
	}
    }
    1;
}



#------------------------------
#
# NEW METHODS
#
#------------------------------

#------------------------------------------------------------
# clear_data_headers
#------------------------------------------------------------
sub clear_data_headers {
    my $self = shift;
    $self->{SummaryCols} = [];    
}

#------------------------------------------------------------
# add_data_header TITLE[,HREF]
#------------------------------------------------------------
# Add a the header to use for all elements.
# Call this when setting up to print.

sub add_data_header {
    my ($self, $title, $href) = @_;

    # Set up data structures, if needed:
    $self->{SummaryCols} or $self->{SummaryCols} = [];   # sequence of titles
    $self->{SummaryHdrs} or $self->{SummaryHdrs} = {}; # indexed by title

    # Add tag, and store title as HTML (for fast printing);
    push @{$self->{SummaryCols}}, $title;
    $self->{SummaryHdrs}{$title} = 
	($href ? (qq{<a href="$href">} . html_escape($title) . qq{</a>})
	       : html_escape($title));
    1;
}

#------------------------------------------------------------
# clear_data_values
#------------------------------------------------------------
sub clear_data_values {
    my $self = shift;
    $self->{SummaryData} = {};   # indexed by title
    
}

#------------------------------------------------------------
# set_data_value TITLE,VALUE,[ISHTML]
#------------------------------------------------------------
# Set up the data for this Summary.
# Call this when setting up to print a particular element.
# If "ISHTML", no need to convert it!

sub set_data_value {
    my ($self, $title, $value, $ishtml) = @_;
    defined($value) or $value = '';

    # Convert value to HTML, if it isn't already:
    $value = html_escape($value) unless ($ishtml);    
    
    # Save it:
    $self->{SummaryData}{$title} = $value;    
}

#------------------------------------------------------------
1;


