package IW::User;

=head1 NAME

IW::User - user preferences/profile information

=head1 DESCRIPTION

This class sits on top of various component classes, and models
the information found in a user preferences file.

=head2 ODL INTERFACE

The basic methods across all component classes are:

    from_CGI        Constructor: build new object from CGI input
    print           Class/instance method: output FORM fields
    from_chunk      Constructor: build new object from IK/ODL input

=head2 CGI INTERFACE

Each component class is able to initialize itself from CGI form input,
and write itself as an HTML form.  It is because of the dependency
on $CGI and $HTML objects that these classes must go in IW::.

Each CGI tag that initializes such an object is expected to have
a name of the form:

    {prefix}{method}

For example, the first name in a shipping address could be called
either of these:

    ship.firstname          (prefix = "ship.")
    firstname               (prefix = "")

The two basic methods across all component classes are:

    from_CGI        Constructor: build new object from CGI input
    print           Class/instance method: output FORM fields

=head1 MAPPING

    INSTANCE    INSTANCE          ODL                CLASS
    METHOD      VARIABLE          GROUP              (IW::User...)
    ---------------------------------------------------------------
    req_addr    REQUEST           REQUEST            Addr
    ship_addr   SHIPPING          SHIPPING           Addr
    bill_addr   BILLING           BILLING            Addr
    affil       USER_AFFILIATION  USER_AFFILIATION   Affil
    keys        KEYS              KEYS               Keys
    display     WWW_DISPLAY       WWW_DISPLAY        Display
    devel       WWW_DEVELOPER     WWW_DEVELOPER      Devel
    
=head1 REDESIGN CONSEQUENCES

Most of the access to components of User have been internalized
so that the various modes no longer get, manipulate, and print
components directly.  Instead User handles the specific forms.
The purpose was to provide a safer, consistent behavior.  The
consequence is that the print() and from_CGI() methods are very long.

=head1 SYNOPSIS

=head2 Getting/setting components

    # Extract components from an IW::User object:
    my $reqaddr   = $user->req_addr;   
    my $shipaddr  = $user->ship_addr;  
    my $billaddr  = $user->bill_addr;  
    my $affil     = $user->affil;      
    my $keys      = $user->keys;       
    my $display   = $user->display;    
    my $devel     = $user->devel; 
  
    # Replace a component:
    $user->req_addr($new_req_object);

=head2 ODL interface

    # Read the user preferences from an ODL file:
    open PROFILE, "profile.dat" or die "open: $!";
    my $odl = parse IU::ODL \*PROFILE;
    close PROFILE;
   
    # Use that ODL to create a user preferences object:
    my $user = IW::User->from_ODL($odl);

=head2 HTML interface

    # Print appropriate parts of User object for an Order form.
    $user->print('FORM'=>'ORDER');

    # Print appropriate parts of User object for Preferences form.
    $user->print('FORM'=>'PREFS');

=head2 CGI interface

    # Read User information from a CGI form.  Note external program
    # only needs to specify which form to retrieve.
    $user->from_CGI('FORM'=>'ALL');

=head2 Installing in the session state

    # Install the given IW::User object in the session state and
    # "prefs.dat" file:
    $user->install;
    
    # Update selected information in the user object in the session
    # state, and then install:
    update IW::User 'ship_addr'=>$newshipaddr,
                    'bill_addr'=>$newbilladdr;

=head1 AUTHOR

Eryq, Hughes STX Corporation

Larry P. Rosen, Raytheon Systems Corp.
=> Major Redesign (version 1.24-1.27)

=head1 VERSION

$Id: User.pm,v 1.78 2003/02/14 16:33:03 solomon Exp $

=cut

#=====================================================================

use IU;
use IU::ODL ();
use IU::Chunk;
use IU::Errors;
use IU::Event;
use IU::Log;
use strict;

use IW::App;
use IW::Util qw(:url);
require IW::UserAddr;
require IW::UserAffil;
require IW::UserKeys;
require IW::UserDevel;
require IW::UserDisplay;

#---------------------------------------------------------------------
# Addr
# Affil
# Keys
# Display
# Devel
#---------------------------------------------------------------------
# I<Class methods>.  # Fake subclasses.  Return the appropriate
# classes to use when creating components.  These allow caller to
# say: IW::User::Addr when really the class is IW::UserAddr.

sub Addr    { 'IW::UserAddr' }
sub Affil   { 'IW::UserAffil' }
sub Keys    { 'IW::UserKeys' }
sub Display { 'IW::UserDisplay' }
sub Devel   { 'IW::UserDevel' }

# Backwards compatibility: old-style classes are empty subclasses
# of new:
@IW::User::Addr::ISA    = qw(IW::UserAddr);
@IW::User::Affil::ISA   = qw(IW::UserAffil);
@IW::User::Keys::ISA    = qw(IW::UserKeys);
@IW::User::Display::ISA = qw(IW::UserDisplay);
@IW::User::Devel::ISA   = qw(IW::UserDevel);

# Map ODL tag to pretty name.
my %Pretty = (REQUEST  => 'Contact Address',
              SHIPPING => 'Shipping Address',
              BILLING  => 'Billing Address',
              KEYS     => 'Data Access Keys');


# These file globals are hashes of required fields for various
# address printings.  They are defined here since they are used
# in both print and from_CGI methods.

# Two types of required fields for complete addresses:
# None required:
my %No_Req_Fields = (title   =>0, firstname=>0,  initial=>0,
                     lastname=>0, organization=>0, email=>0,
                     street  =>0, city=>0, state=>0, zip=>0,
                     country =>0, phone=>0,          fax=>0);

# Standard required:
my %Std_Req_Fields = (title   =>0, firstname=>1,  initial=>0,
                      lastname=>1, organization=>0, email=>1,
                      street  =>1, city=>1, state=>1, zip=>0,
                      country =>1, phone=>1,          fax=>0);

# Required fields for just name:
my %Name_Fields = (title =>0, firstname=>1, initial=>0, lastname=>1);

# Required fields for comments forms:
my %Comment_Fields = (title=>0, firstname=>1, initial=>0,
                      lastname=>1, organization=>0, email=>1,
                      phone=>1, fax=>0);

# Required fields for creating an account:
my %New_Account_Fields = (title   =>0, firstname=>1,  initial=>0,
                          lastname=>1, organization=>0, email=>1,
                          street  =>0, city=>0, state=>0, zip=>0,
                          country =>0, phone=>1,          fax=>0);
#---------------------------------------------------------------------
# new
#---------------------------------------------------------------------
# Constructor.  Creates a new, empty user.

sub new {
   my $class = shift;                   # Get class.
   my $self = {};                       # Start with empty hash.
   $self->{ERRORS} = new IU::Errors;    # Add empty errors object.
   bless $self, $class;                 # Create the object.
   return ($self);                      # Return object.
}

#---------------------------------------------------------------------
# _ivar - PRIVATE! Get/set a named instance variable
#---------------------------------------------------------------------
# Returns the current value of the variable if getting,
# and the I<new> value if setting.

sub _ivar {
   my ($varname, $self, $value) = @_;
   (@_ > 2) and $self->{$varname} = $value;
   $self->{$varname};
}

#---------------------------------------------------------------------
# Get/set methods: req_addr, ship_addr, etc.
#---------------------------------------------------------------------

sub req_addr     { _ivar('REQUEST', @_) }
sub ship_addr    { _ivar('SHIPPING', @_) }
sub bill_addr    { _ivar('BILLING', @_) }
sub affil        { _ivar('USER_AFFILIATION', @_) }
sub keys         { _ivar('KEYS', @_) }
sub display      { _ivar('WWW_DISPLAY', @_) }
sub devel        { _ivar('WWW_DEVELOPER', @_) }
sub archive_id   { _ivar('ArchiveId', @_) }
sub status_code  { _ivar('StatusCode', @_) }
sub status_code_comment  { _ivar('StatusCodeComment', @_) || '' }

# Get/set methods for setting package information in the User object
sub data_format { _ivar('DataFormat', @_) }
sub media_type { _ivar('MediaType', @_) }
sub media_format { _ivar('MediaFormat', @_) }
sub ftp_host { _ivar('FtpHost', @_) }
sub ftp_password { _ivar('FtpPassword', @_) }
sub ftp_pushdest { _ivar('FtpPushDest', @_) }
sub ftp_user { _ivar('FtpUser', @_) }
sub ftp_userstring { _ivar('FtpUserString', @_) }

#---------------------------------------------------------------------
# errors
#---------------------------------------------------------------------
# Get/Set method for Errors object stored in User object.
# This could have been: sub errors { _ivar('ERRORS',@_)};
# EXCEPT that for backwards compatibility with persistent sessions,
# we need to see if the Errors object exists, and if not create one.

sub errors {
   my $self = shift;
   my $value = shift;
   # If setting, set value:
   if (defined ($value) and $value) {
      $self->{ERRORS} = $value;
   } else {
      # If there is no current Errors object, create one.
      $self->{ERRORS} = new IU::Errors   unless ($self->{ERRORS});
   }
   return ($self->{ERRORS});
}

#---------------------------------------------------------------------
# authenticator
#---------------------------------------------------------------------
# Return the authenticator... built from the user's first and last
# names, and the authentication_key.  Returns undef if information
# not available.

sub authenticator {
   my $self = shift;
   require IU::Auth;

   my ($addr, $keys);
   return '' if (!($addr = $self->req_addr) ||
                 !($keys = $self->keys));

   my $firstname = $addr->firstname;
   my $lastname =  $addr->lastname;
   my $authentication_key = $keys->authentication_key;

   return '' if (!$firstname || !$lastname || !$authentication_key);

   IU::Auth::make_authenticator ($firstname, $lastname,
                                 $authentication_key);
}

#---------------------------------------------------------------------
# nasda_authenticator
#---------------------------------------------------------------------
# Return the NASDA authenticator... built from the user's first and last
# names, and the nasda_authentication_key.  Returns undef if information
# not available.

sub nasda_authenticator {
   my $self = shift;
   require IU::Auth;

   my ($addr, $keys);
   return '' if (!($addr = $self->req_addr) ||
                 !($keys = $self->keys));

   my $firstname = $addr->firstname;
   my $lastname =  $addr->lastname;
   my $nasda_authentication_key = $keys->nasda_authentication_key;

   return '' if (!$firstname || !$lastname || !$nasda_authentication_key);

   IU::Auth::make_authenticator ($firstname, $lastname,
                                 $nasda_authentication_key);
}

#---------------------------------------------------------------------
# ecs_authenticator
#---------------------------------------------------------------------
# Return the ecs authenticator... built from the user name, and the
# ecs_authentication_key.  Returns undef if information not available.

sub ecs_authenticator {
   my $self = shift;
   require IU::Auth;

   my $keys;
   # Get keys object or return '' if not defined yet.
   return '' unless ($keys = $self->keys);

   my $ecs_username = $keys->ecs_username;
   my $ecs_authentication_key = $keys->ecs_authentication_key;

   return '' if (!$ecs_username || !$ecs_authentication_key);

   IU::Auth::make_ecs_authenticator($ecs_username,
                                    $ecs_authentication_key);
}

#---------------------------------------------------------------------
# old_ecs_authenticator
#---------------------------------------------------------------------
# Return the old_ecs authenticator... built from the user name, and the
# old_ecs_authentication_key.  Returns undef if information not available.

sub old_ecs_authenticator {
   my $self = shift;
   require IU::Auth;

   my $keys;
   # Get keys object or return '' if not defined yet.
   return '' unless ($keys = $self->keys);

   my $ecs_username = $keys->ecs_username;
   my $old_ecs_authentication_key = $keys->old_ecs_authentication_key;

   return '' if (!$ecs_username || !$old_ecs_authentication_key);

   IU::Auth::make_ecs_authenticator($ecs_username,
                                    $old_ecs_authentication_key);
}

# ecs_username
#---------------------------------------------------------------------
# Return the ecs username.
# Returns undef if information not available.

sub ecs_username {
   my $self = shift;

   my $keys;
   # Get keys object or return '' if not defined yet.
   return '' unless ($keys = $self->keys);

   my $ecs_username = $keys->ecs_username;

   return '' if (!$ecs_username);

   $ecs_username;
}

#---------------------------------------------------------------------
# home_daac
#---------------------------------------------------------------------
# Return the home DAAC.
# Returns undef if information not available.

sub home_daac {
   my $self = shift;

   my $keys;
   # Get keys object or return '' if not defined yet.
   return '' unless ($keys = $self->keys);

   my $home_daac = $keys->home_daac;

   return '' if (!$home_daac);
   return $home_daac;
}

#---------------------------------------------------------------------
# install
#---------------------------------------------------------------------
# I<Instance method.>
# Install this object as the current user prefs.  Updates the session
# state I<and> the session "prefs.dat" file.  Returns the self object
# on success, false (or fatal exception) on error.

sub install {
   my $self = shift;
   my %params = @_;
   my $dont_install = $params{dont_install};

   # Start by activating the preferences:
   $self->display->install       if $self->display;
   $self->devel->install         if $self->devel;
    
   # Instead of creating a COPY function, I decided to allow install
   # to accept a destination directory (right now it's used in the
   # persistent directory).  We also need to save the OLD SESSION
   # directory.  Creating a persistent directory doesn't mean they
   # are actually going to use it right now. JSA 2/4/98

   my $save_old;
   if ($params{To}) {
      $save_old = $SESSION->dir;
      $SESSION->{Dir} = $params{To};
   }
#   debug "Installing to = ".$SESSION->dir;

   # Get the names of the existing and new user prefs files:
   my $old_prefs = $SESSION->path_to('prefs.dat');
   my $new_prefs = $SESSION->path_to('prefs.dat.new');

   # Save the user file data to the session:
   open PREFS, ">$new_prefs" or puke "couldn't open $new_prefs: $!";
   my $odl = $self->to_ODL;
   $odl->output(\*PREFS);
   $odl->end(\*PREFS);      # needed to make IK happy
   close PREFS;

   # Install it as the new user preferences:
   $SESSION->state->{User} = $self unless($dont_install);
   rename($new_prefs, $old_prefs) or
      puke "couldn't rename $new_prefs: $!";

   # Add symlink to file that IK will recognize:
   my $ik_prefs = $SESSION->path_to('profile2.odl');
   unlink($ik_prefs);
   symlink("prefs.dat",$ik_prefs) or
      puke "couldn't make $ik_prefs link: $!";

   # Set the SESSION directory BACK to the original value.
   if ($params{To}) {
      $SESSION->{Dir} = $save_old;
   }
   # Done!
   $self;
}

#---------------------------------------------------------------------
# update PARAMHASH
#---------------------------------------------------------------------
# I<Class and Instance method.>
# Get the current user prefs object, modify the selected fields, and
# then re-install it.  Update params are the same as the method names:
#
#      req_addr   => the new request/comment address
#      ship_addr  => the new shipping address
#      bill_addr  => the new billing address
#      affil      => the new affiliation
#      keys       => the new user keys
#      display    => the new display prefs
#      devel      => the new developer prefs
#      errors     => the new errors object
#
# Creates a new user object automatically if needed.  Returns
# the new object on success, false (or fatal exception) on error.
# NOTE THAT YOU DO NOT NEED TO "install" AFTER OR BEFORE AN "update".
# IT IS DONE FOR YOU.

sub update {
   my $class = shift;
   my %params = @_;
   my $dont_install = $params{dont_install};

   my $user;
   if (ref ($class)) {  # Given a User object; use it.
      $user = $class;
   } elsif(defined($dont_install) && $dont_install) {
      $user = $class->new;
   } else {
      # Get current session User object or create a new one:
      $user = $SESSION->state->{User} || $class->new;
   }
   # Process "attribute update" args:
   my $attr;
   foreach $attr (qw(req_addr ship_addr bill_addr affil keys 
                     display devel errors)){
       $user->$attr($params{$attr}) if exists($params{$attr});
   }
   # Install and return it!
   $user->install unless($dont_install);
   $user;
}

#---------------------------------------------------------------------
# from_ODL ODLOBJECT
#---------------------------------------------------------------------
# I<Class or instance method.>
# Initialize the prefs from a standard IMS ODL object.

sub from_ODL {
   my $self = shift;
   my $odl = shift() || return undef;
   my $dont_install = shift;

   # Hash of ODL groups.
   my %fields = $odl->as_array;

   # Check if file has ANY of the correct groups:
   my $bad_ODL = 1;
   foreach (qw (REQUEST SHIPPING BILLING USER_AFFILIATION KEYS 
                WWW_DISPLAY WWW_DEVELOPER)) {
      if (exists ($fields{$_})) {
         $bad_ODL = 0;
         last;            # Found a correct group; break loop.
      }
   }
   return undef if ($bad_ODL);     # Return undef if no groups.

   # Get the pieces of a user object from the ODL.  If any of the
   # fields are missing, the methods return undef.

   my $req   = IW::UserAddr->from_ODL    ($fields{REQUEST});
   my $ship  = IW::UserAddr->from_ODL    ($fields{SHIPPING});
   my $bill  = IW::UserAddr->from_ODL    ($fields{BILLING});
   my $keys  = IW::UserKeys->from_ODL    ($fields{KEYS});
   my $affil = IW::UserAffil->from_ODL   ($fields{USER_AFFILIATION});
   my $disp  = IW::UserDisplay->from_ODL ($fields{WWW_DISPLAY});
   my $devel = IW::UserDevel->from_ODL   ($fields{WWW_DEVELOPER});

   # Gather any errors found.  Ignores undef fields.
   my $errors = collect_errors ('contact' => $req,
                                'ship'    => $ship,
                                'bill'    => $bill,
                                'keys'    => $keys,
                                'disp'    => $disp,
                                'affil'   => $affil);

   # For each of the pieces above, if it is defined, add it to the
   # list of pieces to update in the User object.
   my @ulist;      # List of items to update.

   push (@ulist, 'req_addr'  => $req)   if ($req);
   push (@ulist, 'ship_addr' => $ship)  if ($ship);
   push (@ulist, 'bill_addr' => $bill)  if ($bill);
   push (@ulist, 'keys'      => $keys)  if ($keys);
   push (@ulist, 'disp'      => $disp)  if ($disp);
   push (@ulist, 'affil'     => $affil) if ($affil);
   push (@ulist, 'display'   => $disp)  if ($disp);
   push (@ulist, 'devel'     => $devel) if ($devel);
   push (@ulist, 'dont_install' => $dont_install) if($dont_install);

   # Return undef if something went wrong and nothing in list.
   return undef  unless (@ulist);

   # Add errors object (even if there were no errors).
   push (@ulist, 'errors' => $errors);

   # Update the current session state user object.  If it
   # doesn't exist, a new one is created and installed.
   update IW::User @ulist;

}

#---------------------------------------------------------------------
# to_ODL 
#---------------------------------------------------------------------
# I<Instance method.>  Convert entire object to ODL, and return it.
# Only those components which exist will be converted!
# NOTE: yes, we now store the keys.  Because the GUI does, too.

sub to_ODL {
   my $self = shift;
   my $part;

   # Create group:
   my $odl = IU::ODL->new_group;

   # Convert components to ODL, ONLY if they're there!:
   $part = $self->req_addr  ? $self->req_addr->to_ODL : undef;
   $odl->add_param('REQUEST', $part) if $part;
   $part = $self->ship_addr ? $self->ship_addr->to_ODL : undef;
   $odl->add_param('SHIPPING', $part) if $part;
   $part = $self->bill_addr ? $self->bill_addr->to_ODL : undef;
   $odl->add_param('BILLING', $part) if $part;
   $part = $self->affil     ? $self->affil->to_ODL : undef;
   $odl->add_param('USER_AFFILIATION', $part) if $part;

   # Since the AUTHENTICATOR is built from some REQUEST info, we need
   # to edit it a little...
   if ($part = $self->keys ? $self->keys->to_ODL : undef) {

      # Add AUTHENTICATOR to keys... Even if it's empty !
      # The IK code needs both AUTHENTICATION_KEY and AUTHENTICATOR
      # for it to be 'Happy' with our Prefs.

      $part->add_param('AUTHENTICATOR', $self->authenticator);
      $part->add_param('ECS_AUTHENTICATOR',
                       $self->ecs_authenticator);
      # Now, add the part!
      $odl->add_param('KEYS', $part);
   }
   $part = $self->display   ? $self->display->to_ODL : undef;
   $odl->add_param('WWW_DISPLAY', $part) if $part;
   $part = $self->devel ? $self->devel->to_ODL : undef;
   $odl->add_param('WWW_DEVELOPER', $part) if $part;

   # Return it:
   $odl;
}

#---------------------------------------------------------------------
# from_chunk CHUNK
#---------------------------------------------------------------------
# I<Class method, constructor.>
# Read a PROFILE_RETRIEVAL_RESULT message from a chunk.  A chunk
# containing "x-error" is considered null and void.  Returns the User
# object for success, false for failure (incluing EOF).  This method
# currently only used for ECS profile retrieval.

sub from_chunk {
   my ($self, $chunk, $newfile) = @_;

   # Input chunk, checking for errors:
   defined ($chunk) or return undef;
   my $errmsg;
   defined ($errmsg = $chunk->value ('x-error')) and
      return error 454, $errmsg;

   # If no User object given, create one.
   ref ($self) or $self = $self->new;

   # Get fields from chunk.

   $self->{ArchiveId}         = $chunk->value('DATA_CENTER_ID');
   $self->{StatusCode}        = $chunk->value('STATUS_CODE');
   $self->{StatusCodeComment} = $chunk->value('STATUS_CODE_COMMENT');

   # Contact info:
   my $req_chunk = $chunk->value ('CONTACT_ADDRESS');
   if ($req_chunk) {
      $self->{REQUEST} = IW::UserAddr->from_chunk ($req_chunk);
   }
   # Shipping info:
   my $ship_chunk = $chunk->value ('SHIPPING_ADDRESS');
   if ($ship_chunk) {
      $self->{SHIPPING} = IW::UserAddr->from_chunk ($ship_chunk);
   }
   # Billing info:
   my $bill_chunk = $chunk->value ('BILLING_ADDRESS');
   if ($bill_chunk) {
      $self->{BILLING} = IW::UserAddr->from_chunk ($bill_chunk);
   }
   # User Affiliation info:
   my $affil_chunk = $chunk->value ('USER_AFFILIATION');
   if ($affil_chunk) {
      $self->{USER_AFFILIATION} =
         IW::UserAffil->from_chunk ($affil_chunk);
   }
   # Gather any errors found.  Creates IU::Errors object.  Function
   # ignores any undefined pieces.
   $self->{ERRORS} = collect_errors ('contact' => $self->{REQUEST},
                                     'ship'    => $self->{SHIPPING},
                                     'bill'    => $self->{BILLING},
                              'affil'   => $self->{USER_AFFILIATION});

   # Save the user data in a file in the session.
   my $newpath = $SESSION->path_to ($newfile) if $newfile;
   if ($newpath) {
      open PREFS, ">$newpath" or puke "couldn't open $newpath: $!";
      my $odl = $self->to_ODL;
      $odl->output (\*PREFS);
      $odl->end (\*PREFS);      # needed to make IK happy
      close (PREFS);
   }
   # Done.  Return User object.
   return ($self);
}

#---------------------------------------------------------------------
# print %PARAMHASH
#---------------------------------------------------------------------
# New method for printing HTML forms with/for User object information
# or its components. This replaces the many occurrences of these
# form generators in the various modes. The hash parameter %paramhash
# tells the method what kind of form to print. It can also be used to
# tell the User object which specific fields to print.  print
# actually calls the print routines of the components as necessary.
#
# %paramhash can have the following key/value pairs:
#    FORM => value
#       Values: PREFS  - Most of the user preferences form.
#               ECS    - ECS fields and buttons part of a form.
#               - ECS form has additional key:
#                 ORIGIN - has mode to return to after ECS operation.
#               ORDER  - Order Summary form.
#               ACCOUNT - account status form.
#               EMAIL   - Required e-mail for Order status form.
#               COMMENT - Part of address for comment form and FTP.
#               CREATEACCT - Create account form.
#               RECEIPT - order receipt. (not to $HTML!)
#               - RECEIPT form has additional key:
#                 FILEHANDLE - has value where to print.
#    REQUIRED => 1 or 0. Optional. Tells whether fields printed are
#                        required (printed with red labels).
#    AUTHREQUIRED => 1 or 0. Optional. Tells whether fields for V0
#                            authenticator are required.
#
# Note: ECS keys are never required. Some data is available to guests.
#---------------------------------------------------------------------

sub print {
   my $self = shift;
   my %params = @_;

   # Make sure we got a proper parameter.
   defined ($params{FORM}) or
           puke 600, 'Coding error.  User::print must be called ',
                     'with a hash of the form to print.';

   # If calling function didn't set required flag, default to false.
   my $required = $params{REQUIRED} || 0;

   # Set flag if V0 auth. required.
   my $auth_required = $params{AUTHREQUIRED} || 0;

   # Set flag if NASDA auth. required.
   my $nasda_required = $params{NASDAREQUIRED} || 0;

   # Set flag if ECS auth. required.
   my $ecs_required = $params{ECSREQUIRED} || 0;

   # Use instance, or get from session, or make new.
   ref ($self) or $self = $SESSION->state->{User} or
       $self = new IW::User;

   my $print_fields;     # Reference to hash of fields to print.
   my $type;
   # Handle form based printing.  Use = to assign to $type.
   if ($type = $params{FORM}) {

      if ($type eq 'PREFS') {

         # Print major part of a user preferences form.
         $HTML -> H3 -> t('Interface Preferences:') -> _H3;

         $HTML -> A(NAME=>'CUSTOM');
         $HTML->HR
              ->TABLE(SIZE=>"75%",BORDER=>0,CELLSPACING=>3)
              ->TR->TD(VALIGN=>'TOP',ALIGN=>'RIGHT');
         $HTML -> msubmit_button(
           SRC => "${IU::DocsUrl}btns/customize_order.gif",
           ALT =>'Customize Order Options',
           QUERY => [ 'mode' => 'PREFSPKG',
                      'origin' => 'PREFS' ]);
         $HTML -> _TD->TD(VALIGN=>'TOP');
                        
         $HTML -> t("to set order options automatically for future orders ");
         $HTML -> t("from this guest session or your registered user account.");
         $HTML->_TD->_TR->_TABLE;

         # Get current display object or create a new one.
         my $display = defined ($self->{WWW_DISPLAY}) ?
             $self->display : IW::UserDisplay->new;

         # Print the fields of the display object.
         $display->print_fields (Prefix=>'display.');

         # Only print developer prefs if known:
         my $devel = $self->{WWW_DEVELOPER};
         if ($devel) {
            $HTML -> HR (ALIGN=>'CENTER', WIDTH=>'50%');
            $HTML -> H3 -> t('Developers Only:') -> _H3;
            $devel->print_fields (Prefix=>'devel.');
         }
         # Submit button for User Display Prefs Only
         my $fix_type = ['mode' => 'PREFS', 'submode' => 'DISPLAY' ];
         my $newtype = $REQ->param ('type');
         push @$fix_type, 'type' => $newtype if defined ($newtype);
         $HTML -> CENTER
               -> msubmit_button (CAPTION =>
                                  'Ok! Accept Display Options Only!',
                                  QUERY   => $fix_type)
               -> _CENTER -> BR;
         $HTML -> HR (ALIGN=>'CENTER', WIDTH=>'50%');

         # Set reference to Hash of fields to print. Defined at top
         # of module.
         if ($required) {  # Selected fields required.
            $print_fields = \%Std_Req_Fields;
         } else {  # No fields required.
            $print_fields = \%No_Req_Fields;
         }
         # If told to, show all 3 addresses.
         my $all3 = $REQ->param ('ShowAddress');
         if (defined ($all3) and $all3) {
            # Print long form.
            $self->print_all_address ($print_fields);
         } else {
            # Else print contact and unique addresses.
            # Print short form.
            $self->print_address ('FIELDS'   => $print_fields,
                                  'MODE'     => 'PREFS',
                                  'SUBMODE'  => 'DISPLAY');
         }
         # User affiliation:
         $HTML -> HR (ALIGN=>'CENTER', WIDTH=>'50%');
         $HTML -> H3 -> t('Your Affiliation:') -> _H3;
         my $affil = defined ($self->{USER_AFFILIATION}) ?
             $self->{USER_AFFILIATION} : IW::UserAffil->new;
         $affil->print ('TYPE'=>$required, 'CATEGORY'=>$required);

         my $keys = defined ($self->{KEYS}) ?
             $self->{KEYS} : IW::UserKeys->new;

         # V0 data access key.
         $HTML -> HR (ALIGN=>'CENTER', WIDTH=>'50%');
         $keys->print ('V0'=>$auth_required,'NASDA'=>$nasda_required);
         $HTML -> HR (ALIGN=>'CENTER', WIDTH=>'50%');

      } # EndIf PREFS.
      elsif ($type eq 'ECS') {
         # Calling function should have set origin tag. Default PREFS.
         my $origin = $params{ORIGIN} || 'PREFS';

         # Print the ecs fields.
         $self->print_ecs_fields ($origin);
      }
      elsif ($type eq 'WARN_V0') {
         # Order Status warning (need V0 auth).

         my $keys = defined ($self->{KEYS}) ?
             $self->{KEYS} : IW::UserKeys->new;
         $keys->print ('V0'=>1);

      } # EndIf WARN_V0.
      elsif ($type eq 'WARN_NASDA') {
         # Order Status warning (need NASDA auth).

         my $keys = defined ($self->{KEYS}) ?
             $self->{KEYS} : IW::UserKeys->new;
         $keys->print ('NASDA'=>1);

      } # EndIf WARN_NASDA.

      elsif ($type eq 'ORDER') {
         # Set reference to Hash of fields to print. Defined at top
         # of module.
         if ($required) {  # Selected fields required.
            $print_fields = \%Std_Req_Fields;
         } else {  # No fields required.
            $print_fields = \%No_Req_Fields;
         }
         # If told to, show all 3 addresses.
         my $all3 = $REQ->param ('ShowAddress');
         if (defined ($all3) and $all3) {
            # Print long form.
            $self->print_all_address ($print_fields);
         } else {
            # Else print contact and unique addresses.
            # Print short form.
            $self->print_address ('FIELDS'   => $print_fields,
                                  'MODE'     => 'ORDER',
                                  'SUBMODE'  => 'FORM');
         }
         # User affiliation:
         $HTML -> HR (ALIGN=>'CENTER', WIDTH=>'50%');
         $HTML -> H3 -> t('Your Affiliation:') -> _H3;
         my $affil = defined ($self->{USER_AFFILIATION}) ?
             $self->{USER_AFFILIATION} : IW::UserAffil->new;
         $affil->print ('TYPE'=>$required, 'CATEGORY'=>$required);

         my $keys = defined ($self->{KEYS}) ?
             $self->{KEYS} : IW::UserKeys->new;

         # V0 data access key. (Not required).
         $HTML -> HR (ALIGN=>'CENTER', WIDTH=>'50%');
         $keys->print ('V0'=>$auth_required,'NASDA'=>$nasda_required);
         $HTML -> HR (ALIGN=>'CENTER', WIDTH=>'50%');

      } # EndIf ORDER.

      elsif ($type eq 'EMAIL') {
         # Print contact info e-mail box.
         # Get current contact address object or create a new one.
         my $req = defined ($self->{REQUEST}) ?
             $self->{REQUEST} : IW::UserAddr->new;
         # Print current e-mail in box.  (may be empty string).
         $req->print_email;
      }
      elsif ($type eq 'ORDERSTATUS') {
         # print order status form

         # Email address:
         my $req = defined ($self->{REQUEST}) ?
             $self->{REQUEST} : IW::UserAddr->new;
         # Print current e-mail in box.  (may be empty string).
         $req->print_email;
         $req->print_names;

         # Data access keys:
         my $keys = defined ($self->{KEYS}) ?
             $self->{KEYS} : IW::UserKeys->new;
         $HTML -> HR (ALIGN=>'CENTER', WIDTH=>'50%');
         $keys->print ('V0'=>0,'NASDA'=>0);
      }
      elsif ($type eq 'ACCOUNT') {
         # Account Status form
    
         # Contact Address 
         my $req = defined ($self->{REQUEST}) ? 
            $self->{REQUEST} : IW::UserAddr->new;
         $HTML -> H3 -> t('Your Contact Address:') -> _H3;
         $req->print ('Prefix'=>'req.', 'FieldsRequired'=>\%Std_Req_Fields);

         my $keys = defined ($self->{KEYS}) ?
             $self->{KEYS} : IW::UserKeys->new;

         # Data access keys.
         $HTML -> HR (ALIGN=>'CENTER', WIDTH=>'50%');
         $keys->print ('V0'=>0,'NASDA'=>0);

         # Note ECS fields are output as hidden tags: 
         $keys->print_hidden('ECS' => 0);

      } # EndIf ACCOUNT.
      elsif ($type eq 'COMMENT') {
         # Comment form or FTP browse.
         # Get current contact address object or create a new one.
         my $req = defined ($self->{REQUEST}) ?
             $self->{REQUEST} : IW::UserAddr->new;
         $req->print ('Prefix'=>'req.',
                      'FieldsRequired'=>\%Comment_Fields);
      }
      elsif ($type eq 'CREATEACCT') {
         $print_fields = \%Std_Req_Fields;
         # Create account form.  Similar to Order. 
         # ECS access key.
         $HTML -> HR (ALIGN=>'CENTER', WIDTH=>'50%');
         my $keys = defined ($self->{KEYS}) ?
             $self->{KEYS} : IW::UserKeys->new;
         #$HTML -> H3 -> t('ECS Access Key:') -> _H3;
         #$keys->print ('ECS' => 0);

         # Print addresses.
         # If told to, show all 3 addresses.
         my $all3 = $REQ->param ('ShowAddress');
         if (defined ($all3) and $all3) {
            # Print long form.
            $self->print_all_address ($print_fields);
         } else {
            # Else print contact and unique addresses.
            # Print short form.
            $self->print_address ('FIELDS'   => $print_fields,
                                  'MODE'     => 'CREATEACCOUNT',
                                  'SUBMODE'  => 'CREATE');
         }
         # User affiliation:
         $HTML -> HR (ALIGN=>'CENTER', WIDTH=>'50%');
         $HTML -> H3 -> t('Your Affiliation:') -> _H3;
         my $affil = defined ($self->{USER_AFFILIATION}) ?
             $self->{USER_AFFILIATION} : IW::UserAffil->new;
         $affil->print ('TYPE'=>1, 'CATEGORY'=>1);

         # V0 data access key:
         $HTML -> HR (ALIGN=>'CENTER', WIDTH=>'50%');
         $keys->print ('V0'=>$auth_required,'NASDA'=>0);
         $HTML -> HR (ALIGN=>'CENTER', WIDTH=>'50%');

      } # EndIf CREATEACCT.
      elsif ($type eq 'RECEIPT') {
         # To print a receipt, we need a filehandle.  Get it.
         my $FH = $params{FILEHANDLE} or
             puke 'IW::User cannot print RECEIPT without FILEHANDLE';
         # Print contact address:
         print $FH "USER ADDRESS\n";
         my $req = $self->req_addr || new IW::UserAddr;
         $req->print_receipt_addr ($FH);

         # Print affiliation.
         my $affil = $self->affil;
         if ($affil) {
            print $FH "     Affiliation:  ", $affil->category,
                      " / ", $affil->type, "\n";
         }
         print $FH "\n";

         # Shipping address:
         print $FH "SHIPPING ADDRESS\n";
         my $ship = $self->ship_addr || new IW::UserAddr;
         $ship->print_receipt_addr ($FH);
         print $FH "\n";

         # Billing address:
         print $FH "BILLING ADDRESS\n";
         my $bill = $self->bill_addr;
         $bill->print_receipt_addr ($FH);
         print $FH "\n";
      }
   }
   1;
}

#--------------------------------------------------------------------
# print_all_address  \%REF_HASH_FIELDS
#--------------------------------------------------------------------
# Print all three, contact, billing, and shipping address forms.
# Input: Reference to a hash of address fields to print.  Keys
#        are field names.  Values indicate whether fields are
#        required or not.  1 = print labels red for selected fields.
#        0 = black.  The hashes are defined at the top.  The one
#        passed in is set in print().
# Output: prints to $HTML.

sub print_all_address {
   my $self = shift;
   my $print_fields = shift;

   # Print Address Directions:
   $HTML -> directions;
   print <<EOF;
      <P>Note that you do <B>not</B> have to manually fill in
      the shipping/billing addresses if they are the same as
      your contact address: just select the "Same as..." option
      next to each.
EOF
   $HTML -> _directions;

   # Get current address objects or create a new ones.
   my $req = defined ($self->{REQUEST}) ?
             $self->{REQUEST} : IW::UserAddr->new;
   my $ship = defined ($self->{SHIPPING}) ?
             $self->{SHIPPING} : IW::UserAddr->new;
   my $bill = defined ($self->{BILLING}) ?
             $self->{BILLING} : IW::UserAddr->new;

   # Contact address:
   $HTML -> H3 -> t('Your Contact Address:') -> _H3;
   $req->print ('Prefix'=>'req.', 'FieldsRequired'=>$print_fields);

   # Shipping address:
   $HTML -> HR (ALIGN=>'CENTER', WIDTH=>'50%');
   $HTML -> tag ('NOBR') -> H3 -> t('Shipping Address:') -> _H3
         -> select_one (Name=>'ship.dup',
                        Choices=>['Given below...',
                                  'Same as contact address (above)'])
         -> tag ('_NOBR');
   $ship->print ('Prefix'=>'ship.', 'FieldsRequired'=>$print_fields);

   # Billing address:
   $HTML -> HR (ALIGN=>'CENTER', WIDTH=>'50%');
   $HTML -> tag ('NOBR') -> H3 -> t('Billing Address:') -> _H3
         -> select_one (Name=>'bill.dup',
                        Choices=>['Given below...',
                                  'Same as contact address (above)',
                                  'Same as shipping address (above)'])
         -> tag ('_NOBR');
   $bill->print ('Prefix'=>'bill.', 'FieldsRequired'=>$print_fields);
   1;
}

#--------------------------------------------------------------------
# print_address  %PARAMHASH
#--------------------------------------------------------------------
# Print contact address.  Print billing, and shipping addresses only
# if they are different from contact address.  "Short form".  Prints
# button to expand to long form.
# Input: Parameter hash.  Key/Values:
#    FIELDS => Reference to a hash of address fields to print.  Keys
#              are field names.  Values indicate whether fields are
#              required or not.  1 = print labels red for selected
#              fields.  0 = black.  The hashes are defined at the top.
#              The one passed in is set in print().
#    MODE => 'ORDER' or 'PREFS' - Mode to reload if change ship/bill.
#    SUBMODE => 'FORM' or 'DISPLAY' - Submode if change ship/bill.
# Output: prints to $HTML.

sub print_address {
   my $self = shift;
   my %paramhash = @_;
   my $num_addresses = 1;

   # Reference to Hash of fields to print. Defined at top of module.
   my $print_fields = $paramhash{FIELDS};

   # Get current address objects or create a new ones.
   my $req = defined ($self->{REQUEST}) ?
             $self->{REQUEST} : IW::UserAddr->new;
   my $ship = defined ($self->{SHIPPING}) ?
             $self->{SHIPPING} : IW::UserAddr->new;
   my $bill = defined ($self->{BILLING}) ?
             $self->{BILLING} : IW::UserAddr->new;

   # Contact address:
   $HTML -> H3 -> t('Your Contact Address:') -> _H3;
   $req->print ('Prefix'=>'req.', 'FieldsRequired'=>$print_fields);

   # Shipping address. If same as contact, print label and hidden tag.
   # Otherwise, print shipping address.
   if ($req->equal ($ship)) {
      $HTML -> BR -> tag ('NOBR') -> B; 
      $HTML -> t('Shipping Address:  Same as contact address (above)')
            -> _B -> tag ('_NOBR')
            -> hidden ('ship.dup'=>'contact');
   } else {
      $num_addresses++;
      $HTML -> HR (ALIGN=>'CENTER', WIDTH=>'50%');
      $HTML -> tag ('NOBR') -> B;
      $HTML -> t('Shipping Address:')
            -> select_one (Name=>'ship.dup',
                           Choices=>['Given below...',
                                   'Same as contact address (above)'])
            -> _B -> tag ('_NOBR');
      $ship->print ('Prefix'=>'ship.',
                    'FieldsRequired'=>$print_fields);
   }
   # Billing address. If same as contact or shipping, print label and
   # hidden tag.  Otherwise, print billing address.
   if ($bill->equal ($req)) {
      $HTML -> BR -> tag ('NOBR') -> B;
      $HTML -> t('Billing Address:  Same as contact address (above)')
            -> _B -> tag ('_NOBR')
            -> hidden ('bill.dup'=>'contact');
   } elsif ($bill->equal ($ship)) {
      $HTML -> BR -> tag ('NOBR') -> B;
      $HTML -> t('Billing Address:  Same as shipping address (above)')
            -> _B -> tag ('_NOBR')
            -> hidden ('bill.dup'=>'shipping');
   } else {
      $num_addresses++;
      $HTML -> HR (ALIGN=>'CENTER', WIDTH=>'50%');
      $HTML -> tag ('NOBR') -> B;
      $HTML -> t('Billing Address:')
            -> select_one (Name=>'bill.dup',
                           Choices=>['Given below...',
                                    'Same as contact address (above)',
                                  'Same as shipping address (above)'])
            -> _B -> tag('_NOBR');
      $bill->print ('Prefix'=>'bill.',
                    'FieldsRequired'=>$print_fields);
   }
   # Print change address button unless we printed all 3.
   unless ($num_addresses == 3) {
      $HTML -> msubmit_flag;
      $HTML -> CENTER -> FONT (SIZE=>'+1') -> B
            -> msubmit_button (CAPTION =>
                               'Change Shipping/Billing Address',
                           QUERY =>['ShowAddress' => 'YES',
                                    'mode'   => $paramhash{MODE},
                                    'submode'=> $paramhash{SUBMODE}]);
      $HTML -> _B -> _FONT -> _CENTER;
   }
   1;
}

#---------------------------------------------------------------------
# print_ecs_fields $ORIGIN
#---------------------------------------------------------------------
# Internal Instance Method - must be called with a User object.
# Print the ECS Access fields.
# Example: $user->print('FORM'=>'ECS');
#          This calls $self->print_ecs_fields;
#          You could also cheat: $user->print_ecs_fields;
# Note that the origin is the mode we are coming from (either
# PREFS or ORDER).

sub print_ecs_fields {
   my $self = shift;
   my $origin = shift;

   # Allow for Multiple Submit buttons
   $HTML -> msubmit_flag;

   # Get the current keys, if any.
   my $keys = defined ($self->{KEYS}) ?
       $self->{KEYS} : IW::UserKeys->new;

   # Have the keys print the ECS fields.  0 -> not required.
   $keys->print ('ECS' => 0);

   # If flag EcsProfileSmc is null, we don't have an ECS Profile
   # Smc to talk to, return without displaying the buttons.

   if ($IU::EcsProfileSmc eq '') {
      return;
   }
   $HTML -> TABLE -> TR -> TD;
   my $gateway = $IU::EcsProfileSmc;
   my @selection = split(/;/, $gateway);
   my $daac = $selection[0];
   $HTML -> msubmit_button(CAPTION => 'Retrieve User Profile',
                           QUERY   => ['mode'        => $origin,
                                       'submode'     => 'ECSPROFILE',
                                       'nextsubmode' => 'RETRIEVE'] );
   $HTML -> _TD;
   $HTML -> TD
         -> t('Press this button to retrieve your user profile. ');
   if($ENV{REMOTE_USER} eq 'guest' || !$ENV{REMOTE_USER}) {
      $HTML -> t('You must enter your ')
            -> t('User Name')
            -> t(' and ')
            -> t('Password')
            -> t(' to access your user profile.');
   }
   $HTML -> _TD -> _TR;

   $HTML -> TR -> TD
         -> msubmit_button (CAPTION => 'Update User Profile',
                            QUERY   => ['mode'        => $origin,
                                        'submode'     => 'ECSPROFILE',
                                        'nextsubmode' => 'UPDATE'] );
   $HTML -> _TD;
   $HTML -> TD
         -> t('Press this button to update your user profile. ');
   if($ENV{REMOTE_USER} eq 'guest' || !$ENV{REMOTE_USER}) {
      $HTML -> t('You must enter your ')
            -> t('User Name')
            -> t(' and ')
            -> t('Password')
            -> t(' to access your user profile.');
   }

   $HTML -> _TD -> _TR -> _TABLE;

   $HTML->BR;
   1;
}

#---------------------------------------------------------------------
# from_CGI %PARAMHASH
#---------------------------------------------------------------------
# Class and instance method for reading user information from a CGI
# form and storing in the User object.  The session user object will
# be updated and installed.  The form it reads should be generated
# using the print method above.  %paramhash tells the type of form to
# parse.  It can have the following key/value pairs:
#    FORM => value
#       Values: ALL     - all User info from User Preferences page.
#               DISPLAY - just display info from Preferences page.
#               CUSTOMIZE - just results list from the Customize page.
#               WARNING - warning page from Search & Order Status.
#               ORDER   - Order Summary page.
#               ACCOUNT - Account Status form.
#               EMAIL   - required e-mail for Order status form.
#               COMMENT - parts of address from Comments form or FTP.
#               CREATEACCT - Create Account page.
#    REQUIRED => 1 or 0.  Tells whether address fields are required.
#    AUTHREQUIRED => 1 or 0.  Tell whether fields for V0
#                              authenticator are required.
#---------------------------------------------------------------------

sub from_CGI {
   my $self = shift;
   my %params = @_;

   # Puke if called as an instance method.
   puke 600, "Coding error. Don't call User::from_CGI as an ",
      "instance method.  Use IW::User->from_CGI." if (ref ($self));

   # Make sure we got a proper parameter.
   defined ($params{FORM}) or
       puke 600, 'Coding error.  User::from_CGI must be called ',
                 'with a hash of the form to read.';

   # If calling function didn't set required flag, default to false.
   my $required = $params{REQUIRED} || 0;

   my @ulist;        # List of User items to update.
   my $type;         # Type of form to read.
   # Handle form based reading.
   if ($type = $params{FORM}) {

      if ($type eq 'ALL') {
         # Get the whole preferences form - (entire User object).

         # Set reference to Hash of fields to read, depending on
         # wheter a REQUIRED flag is set. Defined at top of module.
         my $req_fields;
         if ($required) {
            $req_fields = \%Std_Req_Fields;
         } else {
            $req_fields = \%No_Req_Fields;
         }
         # Get the addresses.  Note: only required if doing ECS
         # profile update.  See ModePREFS::set_prefs_from_edit().
         my ($req,$ship,$bill) = get_addresses_from_CGI($req_fields);

         # Affiliation:
         my $affil = IW::UserAffil->from_CGI ('type'    => 0,
                                              'category'=> 0);

         # Keys:
         my $keys = IW::UserKeys->from_CGI ('ECS' => 0,
                                            'V0'  => 0,
                                            'NASDA' => 0);

         # User's WWW Display Preferences:
         my $disp = IW::UserDisplay->from_CGI (Prefix => 'display.');

         # Developer Tools (if given! returns undef if no CGI tags!):
         my $devel = IW::UserDevel->from_CGI (Prefix => 'devel.');

         # Gather up any errors found. Note we don't need to check
         # WWW_DISPLAY, or WWW_DEVELOPER because those have only
         # pull-down menus, no type-in fields that could be wrong.
         # Also don't check $affil because it is not required so can't
         # be in error.  It is a pull-down menu.)

         my $errors = collect_errors ('contact' => $req,
                                      'ship'    => $ship,
                                      'bill'    => $bill,
                                      'disp'    => $disp,
                                      'keys'    => $keys);

         # List of items to update:
         @ulist = ('req_addr'  => $req,
                   'ship_addr' => $ship,
                   'bill_addr' => $bill,
                   'affil'     => $affil,
                   'keys'      => $keys,
                   'display'   => $disp,
                   'errors'    => $errors,   # Clears previous err.
                   'devel'     => $devel);
      } # EndIf ALL
      elsif ($type eq 'DISPLAY') {
         # Get just display preferences from User Prefs page.

         # User's WWW Display Preferences:
         my $disp = IW::UserDisplay->from_CGI (Prefix => 'display.');

         # Developer Tools (if given! returns undef if no CGI tags!):
         my $devel = IW::UserDevel->from_CGI (Prefix => 'devel.');

         my $errors = collect_errors ('disp' => $disp);
         # List of items to update:
         @ulist = ('display'   => $disp,
                   'devel'     => $devel);
      } # EndIf DISPLAY
      elsif ($type eq 'CUSTOMIZE') {
         # Get just results list display prefs from the customize page.

#  debug "User:from_CGI: ready to call UserDisplay->from_CGI_customize";
         # User's WWW Display Preferences:
         my $disp = IW::UserDisplay->from_CGI_customize (Prefix => 'display.');

         # List of items to update:
         @ulist = ('display'   => $disp);

         ims_event 'CUSTOMIZE',
                   Show_empty_gran => $disp->show_empty_columns_gran(),
                   Show_empty_dset => $disp->show_empty_columns_dset(),
                   List_length     => $disp->list_length(),
                   Date_type       => $SESSION->state->{DateType},
                   Dset_columns    => $disp->dataset_summary(),
                   Gran_columns    => $disp->granule_summary();
                   

      } # EndIf CUSTOMIZE 

      elsif ($type eq 'METADATA') {
        my $disp = IW::UserDisplay->from_CGI_metadata (Prefix => 'display.');
        @ulist = ('display'   => $disp);
      } # EndIf METADATA

      elsif ($type eq 'WARNING') {
         # Get info from a warning page.

         # Get any keys from CGI.
         my $keys = $self->get_merge_keys ($required);

         @ulist = ('keys' => $keys);

         my $errors;
         # If an address is present, get names and merge.
         if (defined ($REQ->param('req.lastname'))){

            # Get the contact address names.  Required.
            my $req = get_merge_addr_from_CGI (\%Name_Fields);

            # Add addresses and affil to list of User items to update.
            push (@ulist, 'req_addr'  => $req);

            # Gather up error messages.
            $errors = collect_errors ('contact' => $req,
                                      'keys'    => $keys);
         } else {
            # Gather up error messages.
            $errors = collect_errors ('keys' => $keys);
         }
         # Add errors object to list of User items to update.
         # If no errors, this will clear previous errors reported.
         push (@ulist, 'errors' => $errors);

      } # EndIf WARNING
      elsif ($type eq 'ORDER') {
         # Get info from the Order Summary and Create Account page.

	 # Set flag if Order Summary and V0 auth. required.
         my $auth_required = $params{AUTHREQUIRED} || 0;

         # Set flag if NASDA auth. required.
         my $nasda_required = $params{NASDAREQUIRED} || 0;

         # Set reference to Hash of fields to read, depending on
         # wheter a REQUIRED flag is set. Defined at top of module.
         my $req_fields;
         if ($required) {
            $req_fields = \%Std_Req_Fields;
         } else {
            $req_fields = \%No_Req_Fields;
         }
         # Get the addresses:
         my ($req,$ship,$bill) = get_addresses_from_CGI($req_fields);

         # Affiliation:
         my $affil = IW::UserAffil->from_CGI ('type'    => $required,
                                              'category'=> $required);

         # Keys.  V0 might be required.
         my $keys = IW::UserKeys->from_CGI ('ECS'=>0,
                                            'V0'=>$auth_required,
                                            'NASDA'=>$nasda_required);

         # Gather up error messages.
         my $errors = collect_errors ('contact' => $req,
                                      'ship'    => $ship,
                                      'bill'    => $bill,
                                      'keys'    => $keys,
                                      'affil'   => $affil);

         # List of items to update:
         @ulist = ('req_addr'  => $req,
                   'ship_addr' => $ship,
                   'bill_addr' => $bill,
                   'affil'     => $affil,
                   'keys'      => $keys,
                   'errors'    => $errors);
      } # EndIf ORDER
      elsif ($type eq 'CREATEACCT') {
         my $print_fields = \%Std_Req_Fields;
         # Get info from the Order Summary and Create Account page.

	 # Set flag if Order Summary and V0 auth. required.
         my $auth_required = $params{AUTHREQUIRED} || 0;

         # Set flag if NASDA auth. required.
         my $nasda_required = $params{NASDAREQUIRED} || 0;

         # Get the addresses:
         my ($req, $ship, $bill) =
            get_addresses_from_CGI ($print_fields);

         # Affiliation:
         my $affil = IW::UserAffil->from_CGI ('type'    => 1,
                                              'category'=> 1);

         # Keys.  V0 might be required.
         my $keys = IW::UserKeys->from_CGI ('ECS'=>0,
                                            'V0'=>$auth_required,
                                            'NASDA'=>0);

         # Gather up error messages.
         my $errors = collect_errors ('contact' => $req,
                                      'ship'    => $ship,
                                      'bill'    => $bill,
                                      'keys'    => $keys,
                                      'affil'   => $affil);

         # List of items to update:
         @ulist = ('req_addr'  => $req,
                   'ship_addr' => $ship,
                   'bill_addr' => $bill,
                   'affil'     => $affil,
                   'keys'      => $keys,
                   'errors'    => $errors);
      } # EndIf CREATEACCT
      elsif ($type eq 'EMAIL') {
         # Get required contact e-mail from form (order status page).
         # Note this will merge with any existing contact address.
         my $req = IW::UserAddr->email_from_CGI;

         # Gather up error messages.
         my $errors = collect_errors ('contact' => $req);

         # List of items to update:
         @ulist = ('req_addr' => $req, 'errors' => $errors);

      } # EndIf EMAIL

      elsif ($type eq 'ORDERSTATUS') {
         # Get required contact e-mail from form (order status page).
         # Note this will merge with any existing contact address.
         my $req = IW::UserAddr->email_from_CGI;
         $req->names_from_CGI;

         # Get any keys from CGI.
         my $keys = $self->get_merge_keys ($required);

         # Gather up error messages.
         my $errors = collect_errors ('contact' => $req,'keys' => $keys);

         # List of items to update:
         @ulist = ('keys'=> $keys, 'req_addr' => $req, 'errors' => $errors);

      } # EndIf ORDERSTATUS. 

      elsif ($type eq 'ACCOUNT') {
         # Get info from an account status page.

         # Get keys from CGI.  None are required.
         my $keys = IW::UserKeys->from_CGI (V0=>0,NASDA=>0,ECS=>0);

         # Get the contact addresses.  Required.

         # Get a new contact address from CGI.
         my $req = IW::UserAddr->from_CGI (Prefix => 'req.',
                                     FieldsRequired =>\%Std_Req_Fields);

         # Gather up error messages.
         my $errors = collect_errors ('contact' => $req,
                                      'keys'    => $keys
                                      );

         # List of items to update:
         @ulist = ('req_addr'  => $req,
                   'keys'      => $keys,
                   'errors'    => $errors);


      } # EndIf ACCOUNT
      elsif ($type eq 'COMMENT') {
         # Get address info from a Comment form.  Merge with existing.
         my $req = get_merge_addr_from_CGI (\%Comment_Fields);

         # Gather up any error messages.
         my $errors = collect_errors ('contact' => $req);

         # List of items to update:
         @ulist = ('req_addr' => $req, 'errors' => $errors);

      } # EndIf COMMENT
   } # EndIf FORM

   # Update the current session state user object.  If it
   # doesn't exist, a new one is created and installed.
   update IW::User @ulist;

   # Return the user object.
   return $SESSION->state->{User};
}

#---------------------------------------------------------------------
# get_addresses_from_CGI  \%REF_HASH_FIELDS
#---------------------------------------------------------------------
# Get the 3 addresses from the CGI form.  Creates the 3 UserAddr
# objects and fills in data from the CGI forms.
# Input: Reference to a hash of address fields to get.  Keys
#        are field names.  Values indicate whether fields are
#        required or not.  1 = field is required, 0 = not.
#        The hashes are defined at the top.  The one passed in is set
#        in from_CGI().
# Output: (contact, shipping, billing address objects).
# Usage: my ($req, $ship, $bill) = get_addresses_from_CGI (\%hash);

sub get_addresses_from_CGI {
   my $req_fields = shift;

   # Get a new contact address from CGI.
   my $req = IW::UserAddr->from_CGI (Prefix => 'req.',
                                     FieldsRequired => $req_fields);

   # Get a new shipping address from the CGI or copy the
   # contact address if a duplicate is desired.
   my $ship;
   my $dup = $REQ->param ('ship.dup');
   if (defined ($dup) and $dup =~ /contact/i) {
      $ship = new IW::UserAddr;
      $ship->copy ($req);
   } else {
      $ship = IW::UserAddr->from_CGI (Prefix => 'ship.',
                                      FieldsRequired => $req_fields);
   }
   # Get a new billing address from the CGI or copy the
   # contact or shipping address if a duplicate is desired.
   my $bill;
   $dup = $REQ->param ('bill.dup');
   if (defined ($dup) and $dup =~ /contact/i) {
      $bill = new IW::UserAddr;
      $bill->copy ($req);
   } elsif (defined ($dup) and $dup =~ /shipping/i) {
      $bill = new IW::UserAddr;
      $bill->copy ($ship);
   } else {
      $bill = IW::UserAddr->from_CGI (Prefix => 'bill.',
                                      FieldsRequired => $req_fields);
   }
   # Return a list of the 3 address objects.
   return (($req, $ship, $bill));
}

#---------------------------------------------------------------------
# get_merge_addr_from_CGI   (\%REQUIRED_FIELDS)
#---------------------------------------------------------------------
# Get just the specified parts of the contact address from the CGI
# form.  Merges with current contact address or creates a new
# UserAddr object.
# Input: A reference to a hash of the fields to retrieve and if they
#        are required. The keys of the hash are the fields to
#        retrieve. They should match the list of fields printed on
#        the form. These are the names of get/set methods in
#        IW::Address. The values are 1 or 0. If a field has the value
#        1, that field is "required". If a required field is missing
#        or blank, then Error strings are written to the ERRORS
#        object.
# Output: contact address object.  May be Errors in the object.
# Usage: my $req = get_merge_addr_from_CGI (\%Comment_Fields);

sub get_merge_addr_from_CGI {
   my $req_fields = shift;

   # Get current session User object or create new.
   my $user = $SESSION->state->{User} || new IW::User;

   # Get the current contact address or create new.
   my $req = $user->req_addr || new IW::UserAddr;

   # Get fields from CGI into this local address object.
   $req->from_CGI (Prefix => 'req.', FieldsRequired => $req_fields);

   # Return the address object.
   return ($req);
}

#---------------------------------------------------------------------
# get_names_from_CGI
#---------------------------------------------------------------------
# Get just the contact address names from the CGI form.  Merge names
# with current contact address or create a new UserAddr object.
# Input: none.  We assume the fields are required.
# Output: contact address object.  May be Errors in the object.
# Usage: my $req = get_names_from_CGI ();

sub get_names_from_CGI {
   my $user = $SESSION->state->{User} || new IW::User;
   my $req = $user->req_addr || new IW::UserAddr;

   my $req_fields = {firstname=>1, lastname=>1};

   # Get names from CGI into this local address object.
   $req->from_CGI (Prefix => 'req.', FieldsRequired => $req_fields);

   # Return the address object.
   return ($req);
}

#---------------------------------------------------------------------
# get_merge_keys    [$required]
#---------------------------------------------------------------------
# Get any keys from the CGI form.  Merges keys if only getting one
# of the keys.  Creates new keys object if getting both.  Clears
# any errors from keys object.
# Input: scalar - if true then V0 key is required or error.
# Output: Returns modified or new keys object.

sub get_merge_keys {
   my $self = shift;
   my $required = shift;
   defined ($required) or $required = 0;

   my $keys;

   # Determine which keys to get based on CGI tags.
   my $get;
   if (defined ($REQ->param('_dak.ecs_username'))) {
      if (defined ($REQ->param('_dak.authentication_key'))) {
         if (defined ($REQ->param('_dak.nasda_authentication_key'))) {
            $get = 'ALL';
         } else {
            $get = 'BOTH';
         }
      } elsif (defined ($REQ->param('_dak.nasda_authentication_key'))) {
         $get = 'NASDAECS'
      } else {
         $get = 'ECS';
      }
   } elsif (defined ($REQ->param('_dak.authentication_key'))) {
      if (defined ($REQ->param('_dak.nasda_authentication_key'))) {
         $get = 'V0NASDA';
      } else {
         $get = 'V0';
      }
   } elsif (defined ($REQ->param('_dak.nasda_authentication_key'))) {
      $get = 'NASDA';
   } else {
      # Neither V0, NASDA, nor ECS keys in form.  Return current keys or
      # create new.  This case should not happen.
      ref ($self) or ($self = $SESSION->state->{User}) or
          ($self = new IW::User);
      ($keys = $self->keys) or ($keys = new IW::UserKeys);
      return ($keys);
   }
   # Get the keys.
   if ($get eq 'BOTH') {
      # Creates new keys object.
      $keys = IW::UserKeys->from_CGI ('V0' => $required, 'ECS' => 0);
   } elsif ($get eq 'ALL') {
      # Creates new keys object.
      $keys = IW::UserKeys->from_CGI ('V0' => $required, 
                                      'ECS' => 0, 'NASDA' => $required);
   } elsif ($get eq 'V0NASDA') {
      # Creates new keys object.
      $keys = IW::UserKeys->from_CGI ('V0' => $required, 'NASDA' => $required);
   } elsif ($get eq 'NASDAECS') {
      # Creates new keys object.
      $keys = IW::UserKeys->from_CGI ('NASDA' => $required, 'ECS' => 0);
   } else {
      # Only getting one key so merge. Use given keys, session keys,
      # or make new.
      ref ($self) or ($self = $SESSION->state->{User}) or
          ($self = new IW::User);
      ($keys = $self->keys) or ($keys = new IW::UserKeys);

      # Clear any pre-existing error messages. ("if" for old sessions)
      $keys->errors->clear   if ($keys->errors);

      # Get the keys.
      if ($get eq 'V0') {
         $keys->from_CGI ('V0' => $required);
      } elsif ($get eq 'NASDA') {
         $keys->from_CGI ('NASDA' => $required);
      } elsif ($get eq 'ECS') {
         $keys->from_CGI ('ECS' => 0);
      }
   }
   return ($keys);
}

#---------------------------------------------------------------------
# collect_errors  (%Component_hash)
#---------------------------------------------------------------------
# Collect errors from the addresses and keys.  Return one error object
# with just unique messages.  Clear errors in the components.
# Input: Hash of components to check for errors:
#           Keys: 'contact', 'ship', 'bill', 'keys', and/or 'affil'.
#           Values: address objects, key object, affiliation object.
# Output: New IU::Errors object with combined error messages.
# SideEffect: Clears error messages from the components.

sub collect_errors {
   my %components = @_;

   my $errors = new IU::Errors;            # Always create new object.

   # Handle contact address.
   my $err_contact;                 # Error object in contact_address.
   if (exists ($components{'contact'}) and
       defined ($components{'contact'})) {
      $err_contact = $components{'contact'}->errors;
      if ($err_contact->any_errors) {
         $errors->set (map {"Error in $Pretty{'REQUEST'}: $_"}
                       $err_contact->get);
      }
   }
   # Handle shipping address.
   my $err_ship;                   # Error object in shipping_address.
   if (exists ($components{'ship'}) and
       defined ($components{'ship'})) {
      $err_ship = $components{'ship'}->errors;
      if ($err_ship->any_errors) {
         # Report shipping errors unless same as contact.
         unless ($err_contact and $err_ship->equals($err_contact)) {
            $errors->set (map {"Error in $Pretty{'SHIPPING'}: $_"}
                          $err_ship->get);
         }
      }
   }
   # Handle billing address.
   my $err_bill;
   if (exists ($components{'bill'}) and
       defined ($components{'bill'})) {
      $err_bill = $components{'bill'}->errors;
      if ($err_bill->any_errors) {
         # Only mention billing errors if different from contact
         # and shipping.
         unless (($err_contact and $err_bill->equals($err_contact)) or
                 ($err_ship and $err_bill->equals($err_ship))) {
            $errors->set (map {"Error in $Pretty{'BILLING'}: $_"}
                          $err_bill->get);
         }
      }
   }
   # Clear collected errors (to save storage space).
   $err_contact->clear   if ($err_contact);
   $err_ship->clear      if ($err_ship);
   $err_bill->clear      if ($err_bill);

   # Handle keys.
   if (exists ($components{'keys'}) and
       defined ($components{'keys'})) {
      my $err_keys = $components{'keys'}->errors;
      if ($err_keys->any_errors) {
         map {$errors->set ($_)} $err_keys->get();
         $err_keys->clear;
      }
   }
   # Add display errors.
   if (exists ($components{'disp'}) and
       defined ($components{'disp'})) {
      my $err_disp = $components{'disp'}->errors;
      if ($err_disp->any_errors) {
         map {$errors->set ($_)} $err_disp->get();
         $err_disp->clear;
      }
   }
   # Add affiliation errors.
   if (exists ($components{'affil'}) and
       defined ($components{'affil'})) {
      my $err_affil = $components{'affil'}->errors;
      if ($err_affil->any_errors) {
         map {$errors->set ($_)} $err_affil->get();
         $err_affil->clear;
      }
   }
   return ($errors);
}
#---------------------------------------------------------------------
1;
