package IW::UserAddr;

=head1 NAME

IW::UserAddr - represents a user address - contact, billing, or
shipping.

=head1 DESCRIPTION

This class represents a user address, used for contact, billing or
shipping.  This is a subclass of IW::Address and uses most of the
default data and methods of it.

=head2 PARENT/BASE CLASS

IW::Address

=head2 DATA

=begin list

=item

%MaxLen

This is just the inherited IW::Address::MaxLength.  Needed for print.

=item

%Label

This is a map from field name to printed label.  Pretty versions.

=end list

=head2 METHODS

=begin list

=item

Uses inherited constructor new ( ).

=item

bogus ( )

This is a constructor that sets default values for the address.  It
is used for browse requests.

=item

from_CGI (%PARAMHASH)

This retrieves address values from a CGI form.  It will set values
into the user address object.  If no object is given, it will create
a new one.  If there are any errors, such as strings being too long
or required fields being missing, it will set strings in the Errors
object in $self, but return as much of the address as it was able to
get.  from_CGI reads from the form generated by the print method.

=item

print (%PARAMHASH)

Prints an HTML CGI form containing the address.  This is the reverse
of from_CGI.

=head1 AUTHOR

Larry P. Rosen, EOS Data Gateway Team, Raytheon Systems Corp.
Based on old IW::UserAddr.pm by Eryq (Erik Dorfman).

=head1 RCS INFO

$Id: UserAddr.pm,v 1.38 2002/09/17 20:03:28 joe Exp $

=cut

#=====================================================================

use IU;                      # For puke() 
use IU::Log;                 # For debug()
use IU::Session;             # For $SESSION->id()
use IU::FileReader;          # For state and country lists. 
use IW::App;                 # For $HTML, $REQ, and $SESSION.
use IW::Address;
use IW::Util qw(:string);    # For trim() and blank();

# Set up inheritance.
@ISA = qw (IW::Address);

use strict;

# Inherit %MaxLength from parent for print().
my %MaxLen = %{IW::Address::MaxLength};

# Hash for USA states with corresponding abreviations.
my %UsaStates;

# Array for list of countries.
my @Countries;

# Dashed line as separator on country and state lists.
my $dashes = "-- None --";

# Map from field name to printed label.
my %Label = (title        => 'Title',
             firstname    => 'First Name',
             initial      => 'Initial',
             lastname     => 'Last Name',
             street       => 'Street Address',
             organization => 'Organization',
             email        => 'Internet E-Mail Address',
             city         => 'City',
             state        => 'State/Province',
             selstate     => 'Selectc USA State',
             enterstate   => 'Or Enter State/Province',
             zip          => 'Zip/Postal Code',
             country      => 'Country',
             selcountry   => 'Select Country',
             entercountry => 'Or Enter Country',
             phone        => 'Telephone',
             fax          => 'Fax',
            );

#---------------------------------------------------------------------
# new 
#---------------------------------------------------------------------
# Use base class constructor method, then initialize 
# class specific countries and states listings.

sub new {
   my $self = shift->SUPER::new();

   # Hash for USA states with corresponding abreviations.
   # Example:   MARYLAND => "MD"
   %UsaStates = IU::FileReader->new("${IU::LibDir}/${IU::StateList}")
                              ->get_hash;

   # Array for list of countries.
   my @countries = IU::FileReader->new("${IU::LibDir}/${IU::CountryList}")
                                 ->get_array;

   # Make sure we upcase, just to be sure.
   # IU::UserAdmin needs this spelled out (Why? Beats me, but it does.)
   @Countries = map {IW::Util::trim(uc($_))} (sort @countries);

   return ($self);
}

#---------------------------------------------------------------------
# bogus 
#---------------------------------------------------------------------
# This is a constructor that sets default values for the address.  It
# is used for browse requests.

sub bogus {
   my $class = shift;                   # Get class.
   my $self = $class->new;              # Start with a new UserAddr.
   $self->firstname('Anonymous');
   $self->lastname($SESSION->id);
   $self->street(['UNKNOWN STREET1', 'UNKNOWN STREET2']);
   $self->city('UNKNOWN CITY');
   $self->state('UNKNOWN STATE');
   $self->country('UNKNOWN COUNTRY');
   $self->email('UNKNOWN@'.($ENV{REMOTE_HOST}||$ENV{REMOTE_ADDR}));
   $self->phone('UNKNOWN PHONE');
   return ($self);                      # Return object.
}

#---------------------------------------------------------------------
# v0_ok_to_go
#---------------------------------------------------------------------
# Check if we have everything from the Address object needed for
# the v0 authenticator.  Currently that means first and last names.
# Return 1 if yes, 0 if not.

sub v0_ok_to_go {
   my $self = shift;
   if ($self->firstname && $self->lastname) {
      return 1;                            # ready to go
   } else {
      return 0;
   }
}

#---------------------------------------------------------------------
# from_CGI (%PARAMHASH)
#---------------------------------------------------------------------
# This retrieves address values from a CGI form.  It will set values
# into the user address object.  If no object is given, it will create
# a new one.  If there are any errors, such as strings being too long
# or required fields being missing, it will set strings in the Errors
# object in $self, but return as much of the address as it was able to
# get.  from_CGI reads from the form generated by the print method.
# Input:  A parameter hash with the following key/values:
#    FieldsRequired => \%HASH
#       (Required) This is a reference to a hash.  The keys of the
#       hash are the fields to retrieve.  They should match the list
#       of fields printed on the form.  These are lowercase versions
#       of the simple_vars and 'street' (for 'ADDRESS').  They are
#       the names of get/set methods.  The values are 1 or 0.  If
#       a field has the value 1, that field is "required".  If a
#       required field is missing or blank, then Error strings are
#       written to the ERRORS object.
#    Prefix => value
#       (Optional) This scalar string can have one of these values:
#       "req.", "ship.", or "bill.".  It indicates the type of address
#       to get and is the prefix part of the CGI form tag names.
# Output: Filled in UserAddr object, with possible error strings
#         stored in its Errors object.

sub from_CGI {
   my $self = shift;
   my %params = @_;

   # Make new, if invoked as a class method:
   ref ($self) or $self = $self->new;

   # Extract parameters passed to function.
   my $prefix = defined ($params{Prefix}) ? $params{Prefix} : '';
   defined ($params{FieldsRequired}) or
      puke 600, 'Coding error.  UserAddr::from_CGI must be called ',
                'with a hash of the fields to retrieve.';

   my %fields = %{$params{FieldsRequired}};
   my @missing;          # List of required fields that are missing.
   my $longstate;

   # Get each field in requested list.
   my ($tag, $value);
   foreach $tag (keys (%fields)) {
      # Tag "street" needs special treatment.
      if ($tag eq 'street') {
         # street can be three lines and is passed from form
         # using 3 CGI tags: street1, street2, and street3.
         my ($line, @lines);
         foreach ('street1', 'street2', 'street3') {
            $line = $REQ->param ("${prefix}$_");
            push (@lines, trim ($line)) if (defined ($line) and
                                            !blank ($line));
         }
         # If there were any non-blank lines, store them.
         if (@lines) {
            $self->street (\@lines);
         } elsif ($fields{street}) {
            # street requested and required but not found. Note error.
            push (@missing, 'Street Address');
         } else {
            # street requested but not required and blank.  Set blank.
            $self->street (['','','']);
         }

      # Tag "state" needs special treatment.
      } elsif ($tag eq 'state') {
         # state could be choosen from a picklist or entered 
         # from a type in field, picklist takes precedence.
         $value = $REQ->param ("${prefix}selstate");
         $value = trim ($value);
         if (!defined ($value) or blank ($value)
                    or ($value eq $dashes)) {
             # Not selected from picklist, try type in field
             $value = $REQ->param ("${prefix}enterstate");
             $value = trim ($value);
             $value = uc ($value);
         } else {
             # Selected from picklist, we want to store 
             # state abreviation, not long name.
             $longstate = $value;
             $value = $UsaStates{$value};
             $value = trim ($value);
             $value = uc ($value);
         }
         if (defined ($value) and !blank ($value)) {
            $self->state($value);
         } else {
            $self->state ('');
         }

      # Tag "country" needs special treatment.
      } elsif ($tag eq 'country') {
         # country could be choosen from a picklist or entered 
         # from a type in field, picklist takes precedence.
         $value = $REQ->param ("${prefix}selcountry");
         $value = trim ($value);
         if (!defined ($value) or blank ($value) 
                    or ($value eq $dashes)) {
             # Not selected from picklist, try type in field
             $value = $REQ->param ("${prefix}entercountry");
             $value = trim ($value);
             $value = uc ($value);
         }
         if (defined ($value) and !blank ($value)) {
            $self->country($value);
         } elsif ($fields{country}) {
            push (@missing, 'Country');
         } else {
            $self->country ('');
         }
      } else {
         # Tag other than 'street', 'state' or 'country'.  
         # If CGI tag defined:
         if (defined ($value = $REQ->param ("${prefix}$tag"))) {
            # Tag is there.  Handle special case of "title".
            if ($tag eq 'title') {
               $value or $value = '';
               $value = '' if ($value =~ /none/i);  # fake empty value
            }
            # Trim leading/ending blanks.
            $value = trim ($value);

            # If value non-blank, set it and process next tag.
            # Note that if string too long, Errors will be set.
            # Note tags are function names for Address objects.
            if (!blank ($value)) {
               # Set the value using the set function.
               my $cmd = '$self->' . $tag . '($value)';
               eval ($cmd);
               next;
            }
         }
         # Requested field was not filled in.  If required, add error.
         if ($fields{$tag}) {
            push (@missing, $Label{$tag});
         } else {
            # Tag requested but not required and blank.  Set blank.
            my $cmd = '$self->' . $tag . "('')";
            eval ($cmd);
         }
      }
   }  # foreach tag loop.

   # ------------------------------------------------------------------------
   # Country needs the following special checking when it is a required field:
   # - If country is US, a state is required.
   # - If country is Canada, a provice is required.
   # - If country is non US or Canada, no state is required.
   # - If country is non US, cannot have a US state selected from list.
   # ------------------------------------------------------------------------
   if ($fields{country}) {
      if ( ($self->country eq 'UNITED STATES') and ($self->state eq '') ) {
         push (@missing, 'State');
      } elsif ( ($self->country eq 'CANADA') and ($self->state eq '') ) {
         push (@missing, 'Enter State/Province');
      } elsif ( ($UsaStates{$longstate}) and 
                ($self->country ne 'UNITED STATES')) {
         push (@missing, 
             'Must select United States as country when selecting a US state');
      }
   }

   # If any required fields are missing, store error message.
   if (@missing) {
      $self->{ERRORS}->set ('Fields required but not entered: ' .
                            join (', ', @missing) . '.');
   }
   # Return the UserAddr object.
   return ($self);
}

#---------------------------------------------------------------------
# print (%PARAMHASH)
#---------------------------------------------------------------------
# This prints the HTML CGI form for address input.  It will initialize
# fields with current values in the user address object, if any.
# from_CGI() reads from the form generated.
# Input:  A parameter hash with the following key/values:
#    FieldsRequired => \%HASH
#       (Required) This is a reference to a hash.  The keys of the
#       hash are the fields to print.  They should match the list of
#       fields read from the form.  These are lowercase versions of
#       the simple_vars and 'street' (for 'ADDRESS').  They are the
#       names of get/set methods.  The values are 1 or 0.  If a field
#       has the value 1, that field is "required".  If a field is
#       required, its label is printed in red instead of black.
#    Prefix => value
#       (Required) This scalar string can have one of these values:
#       "req.", "ship.", or "bill.".  It indicates the type of address
#       to print and is the prefix part of the CGI form tag names.
# Output: prints table to $HTML stream and returns 1.
       
sub print {
   my $self = shift;
   my %params = @_;

   # Extract parameters passed to function.
   my $prefix = defined ($params{Prefix}) ? $params{Prefix} : '';
   defined ($params{FieldsRequired}) or
      puke 600, 'Coding error.  UserAddr::print must be called ',
                'with a hash of the fields to retrieve.';
   my %fields = %{$params{FieldsRequired}};

   # Make new, if invoked as a class method.
   ref ($self) or $self = $self->new;

   my ($field, $value);   # A field name and current value.
   my $numchars;          # The number of chars for a field.

   # Name row.  Print a row if any part of name requested.
   if (exists ($fields{title}) or exists ($fields{firstname}) or
       exists ($fields{initial}) or exists ($fields{lastname})) {
      $HTML -> TABLE (CELLPADDING => 1, CELLSPACING => 1) -> TR;
      if (exists ($fields{title})) {
         $HTML -> TD (NOWRAP => undef)
               -> field_cell (Title => 'Title:', Stack => 1,
                              Req => $fields{title})
               -> select_one (Name    => "${prefix}title",
                              Choices => ['(none)','Dr.','Mr.','Ms.'],
                              Current => $self->title)
               -> _field_cell -> _TD;
      }
      if (exists ($fields{firstname})) {
         $HTML -> TD (NOWRAP => undef)
               -> field_cell (Title => 'First Name:', Stack => 1,
                              Req => $fields{firstname})
               -> INPUT (SIZE      => $MaxLen{FIRST_NAME},
                         NAME      => "${prefix}firstname",
                         VALUE     => $self->firstname,
                         MAXLENGTH => $MaxLen{FIRST_NAME})
               -> _field_cell -> _TD;
      }
      if (exists ($fields{initial})) {
         # Note that initial field is 1 char wide, does not allow period.
         $HTML -> TD (NOWRAP => undef)
               -> field_cell (Title => 'Initial:', Stack => 1,
                              Req => $fields{initial})
               -> INPUT (SIZE      =>  1,
                         NAME      => "${prefix}initial",
                         VALUE     => $self->initial,
                         MAXLENGTH => 1)
               -> _field_cell -> _TD;
      }
      if (exists ($fields{lastname})) {
         $HTML -> TD (NOWRAP => undef)
               -> field_cell (Title => 'Last Name:', Stack => 1,
                              Req => $fields{lastname})
               -> INPUT (SIZE      => $MaxLen{LAST_NAME},
                         NAME      => "${prefix}lastname",
                         VALUE     => $self->lastname,
                         MAXLENGTH => $MaxLen{LAST_NAME})
               -> _field_cell -> _TD;
      }
      $HTML -> _TR -> _TABLE;
   }  # end of name row.
   # Organization and email row.  These have smaller input field
   # sizes so that the form isn't too wide.
   if (exists ($fields{organization}) or exists ($fields{email})) {
      $HTML -> TABLE (CELLPADDING => 1, CELLSPACING => 1) -> TR;
      foreach $field ('organization', 'email') {
         if (exists ($fields{$field})) {
            # Get the current value.
            eval ("\$value = \$self->$field");
            # Get the field length. Use uppercase field as key.
            $numchars = $MaxLen{"\U$field"};
            # Print the table cell.
            $HTML -> TD (NOWRAP => undef)
                  -> field_cell (Title => "$Label{$field}:",
                                 Stack => 1,
                                 Req => $fields{$field});
            $HTML -> INPUT (SIZE      => 40,
                            NAME      => "${prefix}$field",
                            VALUE     => $value,
                            MAXLENGTH => $numchars)
                  -> _field_cell -> _TD;
         }
      }
      $HTML -> _TR -> _TABLE;
   }  # end of organization and e-mail row.
   # Street?
   if (exists ($fields{street})) {
      $HTML -> TABLE (CELLPADDING => 1, CELLSPACING => 1) -> TR;
      # Get the current array of street lines.
      my @street_lines = $self->street;
      $HTML -> TD (NOWRAP => undef)
            -> field_cell (Title => 'Street Address:', Stack => 1,
                           Req => $fields{street})
            -> INPUT (SIZE      => $MaxLen{STREET},
                      NAME      => "${prefix}street1",
                      VALUE     => $street_lines[0] || '',
                      MAXLENGTH => $MaxLen{STREET})
            -> BR;
      $HTML -> INPUT (SIZE      => $MaxLen{STREET},
                      NAME      => "${prefix}street2",
                      VALUE     => $street_lines[1] || '',
                      MAXLENGTH => $MaxLen{STREET})
            -> BR;
      $HTML -> INPUT (SIZE      => $MaxLen{STREET},
                      NAME      => "${prefix}street3",
                      VALUE     => $street_lines[2] || '',
                      MAXLENGTH => $MaxLen{STREET})
            -> _field_cell -> _TD -> _TR -> _TABLE;
   }  # end street row.
   # City, state, zip row.  Print a row if any of these.
   if (exists ($fields{city}) or exists ($fields{state}) or
       exists ($fields{zip})) {
      $HTML -> TABLE (CELLPADDING => 1, CELLSPACING => 1) -> TR;
      if (exists ($fields{city})) {
         # Print the table cell.
         $HTML -> TD (NOWRAP => undef)
               -> field_cell (Title => "City:",
                              Stack => 1,
                              Req => $fields{city});
         $HTML -> INPUT (SIZE      => $MaxLen{CITY},
                         NAME      => "${prefix}city",
                         VALUE     => $self->city,
                         MAXLENGTH => $MaxLen{CITY})
               -> _field_cell -> _TD -> _TR;
      }
      if (exists ($fields{state})) {
         # Get the current value.
         $value = $self->state;
         # Get the field length. Use uppercase field as key.
         $numchars = $MaxLen{STATE};

         # -------------------------------------------------------------
         # State is a special case, there will be a picklist for USA
         # states selections, and a type in field for non-USA states.  
         # When using the picklist we store the abreviation value, but
         # the list displays the state long name, class hash %UsaStates 
         # has the list of states with corresponding abreviation.
         #
         # A call to state_selection is done to get the user selection 
         # and to get the list of states for the picklist.
         # -------------------------------------------------------------
         my ($entered, $selected, @choices) = $self->state_selection ();
         $HTML -> TD (NOWRAP => undef)
               -> field_cell (Title => "Select State (US only):", 
                              Stack => 1,
                              Req => $fields{state})
               -> select_one (Name => "${prefix}selstate",
                              Size => 5,
                              Current => $selected,
                              Choices=> [@choices])
               -> _field_cell -> _TD;

         # Print input field for state or province.
         $HTML -> TD (NOWRAP => undef)
               -> field_cell (Title => "Or Enter State/Province:", Stack => 1);
         $HTML -> INPUT (SIZE      => $MaxLen{STATE},
                         NAME      => "${prefix}enterstate",
                         VALUE     => $entered,
                         MAXLENGTH => $MaxLen{STATE})
               -> _field_cell;
         $HTML -> BR;
         print "To enter a state/province, select None from the";
         $HTML -> BR;
         print "selection menu and enter state/province above.";
         $HTML -> _TD -> _TR;
      }
      if (exists ($fields{zip})) {
         # Print the table cell.
         $HTML -> TD (NOWRAP => undef)
               -> field_cell (Title => "Zip/Postal Code:",
                              Stack => 1,
                              Req => $fields{zip});
         $HTML -> INPUT (SIZE      => $MaxLen{CITY},
                         NAME      => "${prefix}zip",
                         VALUE     => $self->zip,
                         MAXLENGTH => $MaxLen{ZIP})
               -> _field_cell -> _TD -> _TR;
      }
      $HTML -> _TR -> _TABLE;
   }  # end of city state zip row.
   # Country row.
   if (exists ($fields{country})) {
      $HTML -> TABLE (CELLPADDING => 1, CELLSPACING => 1) -> TR;
      if (exists ($fields{country})) {
         # Get the field length. Use uppercase field as key.
         $numchars = $MaxLen{COUNTRY};

         # -------------------------------------------------------------
         # Country is also a special case, there will be a picklist with
         # names of countries in alphabetical order, users can select the
         # country from the list, there will also be a type in field in 
         # case the country is not found in the list.
         # Class array @Countries has the list of countries.
         #
         # A call to country_selection is done to get the user selection 
         # and to get the list of countries for the picklist.
         # -------------------------------------------------------------
         my ($entered, $selected, @choices) = $self->country_selection ();
         # Print countries selection.
         $HTML -> TD (NOWRAP => undef)
               -> field_cell (Title => "Select Country:", 
                              Stack => 1,
                              Req => $fields{country})
               -> select_one (Name => "${prefix}selcountry", 
                              Size => 5,
                              Choices => [@choices],
                              Current => $selected)
               -> _field_cell -> _TD;

         # Print input field for country.
         $HTML -> TD (NOWRAP => undef)
               -> field_cell (Title => "Or Enter Country:", Stack => 1);
         $HTML -> INPUT (SIZE      => $numchars,
                         NAME      => "${prefix}entercountry",
                         VALUE     => $entered,
                         MAXLENGTH => $numchars)
               -> _field_cell;
         $HTML -> BR;
         print "To enter a country, select None from the";
         $HTML -> BR;
         print "selection menu and enter country above.";
         $HTML -> _TD;
      }
      $HTML -> _TR -> _TABLE;
   }  # end of country row.

   # Phone, fax row.
   if (exists ($fields{phone}) or exists ($fields{fax})) {
      $HTML -> TABLE (CELLPADDING => 1, CELLSPACING => 1) -> TR;
      foreach $field ('phone', 'fax') {
         if (exists ($fields{$field})) {
            # Get the current value.
            eval ("\$value = \$self->$field");
            # Get the field length. Use uppercase field as key.
            $numchars = $MaxLen{"\U$field"};
            # Print the table cell.
            $HTML -> TD (NOWRAP => undef)
                  -> field_cell (Title => "$Label{$field}:",
                                 Stack => 1,
                                 Req => $fields{$field});
            $HTML -> INPUT (SIZE      => $numchars,
                            NAME      => "${prefix}$field",
                            VALUE     => $value,
                            MAXLENGTH => $numchars)
                  -> _field_cell -> _TD;
         }
      }
      $HTML -> _TR -> _TABLE;
   }  # end of phone fax row.
   1;
}

#---------------------------------------------------------------------
# print_email
#---------------------------------------------------------------------
# This prints the HTML CGI form for e-mail input.  The field is
# required.  This is used for order status request form.

sub print_email {
   my $self = shift;
   # Get the field length. Use uppercase field as key.
   my $numchars = $MaxLen{EMAIL};
   $HTML -> TABLE (CELLPADDING => 1, CELLSPACING => 1) -> TR;
   $HTML -> TD (NOWRAP => undef)
         -> field_cell (Title => "$Label{email}:",
                        Stack => 1, Req => 1);
   $HTML -> INPUT (SIZE      => 40,
                   NAME      => 'req.email',
                   VALUE     => $self->{EMAIL},
                   MAXLENGTH => $numchars)
         -> _field_cell -> _TD -> _TR -> _TABLE;
}

#---------------------------------------------------------------------
# print_names
#---------------------------------------------------------------------
# This prints the HTML CGI form for first&last name input.  The field is
# required.  This is used for order status request form.

sub print_names {
   my $self = shift;

   $HTML -> TABLE (CELLPADDING => 1, CELLSPACING => 1) -> TR;
   $HTML -> TD (NOWRAP => undef)
         -> field_cell (Title => 'First Name:', Stack => 1, Req => 1)
         -> INPUT (SIZE      => $MaxLen{FIRST_NAME},
                   NAME      => "req.firstname",
                   VALUE     => $self->firstname,
                   MAXLENGTH => $MaxLen{FIRST_NAME})
         -> _field_cell -> _TD;
   $HTML -> TD (NOWRAP => undef)
         -> field_cell (Title => 'Last Name:', Stack => 1, Req => 1)
         -> INPUT (SIZE      => $MaxLen{LAST_NAME},
                   NAME      => "req.lastname",
                   VALUE     => $self->lastname,
                   MAXLENGTH => $MaxLen{LAST_NAME})
         -> _field_cell -> _TD;

   $HTML -> _TR -> _TABLE;
   1;
}

#---------------------------------------------------------------------
# names_from_CGI
#---------------------------------------------------------------------
# Gets first and last names from form (OrderStatus). Required
# so if it is not found it will set an error message.  If this is
# invoked without an existing address object, it will first try to
# get the contact address from the session state.  If there is none,
# it will create a new address object.

sub names_from_CGI {
   my $self = shift;

   # If not given an address, try to get contact address from session.
   # If that is missing, make a new address object.
   unless (ref ($self)) {
      my $user = $SESSION->state->{User};
      if ($user) {
         my $addr = $user->req_addr;
         if ($addr) {
            $self = $addr;
         } else {
            $self = $self->new;
         }
      } else {
         $self = $self->new;
      }
   }

   my $first;
   # If CGI tag defined:
   if (defined ($first = $REQ->param ('req.firstname'))) {
      # Trim leading/ending blanks.
      $first = trim ($first);

      # If value non-blank, set it.  Note if e-mail string is bad,
      # an Error will be set.
      if (!blank ($first)) {
         # Set the value using the set function.
         $self->firstname ($first);
      } else {
         # E-mail is required but missing, so set an error.
         $self->{ERRORS}->set ('First name is
            required for this function.  Please enter one.');
      }
   } else {
      # E-mail is required but missing, so set an error.
      $self->{ERRORS}->set ('First name is
         required for this function.  Please enter one.');
   }

   my $last;
   # If CGI tag defined:
   if (defined ($last = $REQ->param ('req.lastname'))) {
      # Trim leading/ending blanks.
      $last = trim ($last);

      # If value non-blank, set it.  Note if e-mail string is bad,
      # an Error will be set.
      if (!blank ($last)) {
         # Set the value using the set function.
         $self->lastname ($last);
      } else {
         # E-mail is required but missing, so set an error.
         $self->{ERRORS}->set ('Last name is
            required for this function.  Please enter one.');
      }
   } else {
      # E-mail is required but missing, so set an error.
      $self->{ERRORS}->set ('Last name is
         required for this function.  Please enter one.');
   }



   return ($self);
}


#---------------------------------------------------------------------
# email_from_CGI
#---------------------------------------------------------------------
# Gets e-mail address from form (OrderStatus).  E-mail is required
# so if it is not found it will set an error message.  If this is
# invoked without an existing address object, it will first try to
# get the contact address from the session state.  If there is none,
# it will create a new address object.

sub email_from_CGI {
   my $self = shift;

   # If not given an address, try to get contact address from session.
   # If that is missing, make a new address object.
   unless (ref ($self)) {
      my $user = $SESSION->state->{User};
      if ($user) {
         my $addr = $user->req_addr;
         if ($addr) {
            $self = $addr;
         } else {
            $self = $self->new;
         }
      } else {
         $self = $self->new;
      }
   }
   my $value;
   # If CGI tag defined:
   if (defined ($value = $REQ->param ('req.email'))) {
      # Trim leading/ending blanks.
      $value = trim ($value);

      # If value non-blank, set it.  Note if e-mail string is bad,
      # an Error will be set.
      if (!blank ($value)) {
         # Set the value using the set function.
         $self->email ($value);
      } else {
         # E-mail is required but missing, so set an error.
         $self->{ERRORS}->set ('A valid e-mail address for you is
            required for this function.  Please enter one.');
      }
   } else {
      # E-mail is required but missing, so set an error.
      $self->{ERRORS}->set ('A valid e-mail address for you is
         required for this function.  Please enter one.');
   }
   return ($self);
}

#---------------------------------------------------------------------
# country_selection
#---------------------------------------------------------------------

sub country_selection {
  my $self = shift;
  
  my $entered = "";
  my $selected = "";
  unless (defined $self->{Countries}) {
    $self->{Countries} = IU::FileReader->new("$IU::LibDir/$IU::CountryList")
      ->get_array;
  }
  @Countries = @{$self->{Countries}};
		 
  my @choices = ($dashes, "UNITED STATES");
  push (@choices, @Countries);
  
  my $value = $self->country;
  $value = uc ($value);
  $value = trim ($value);
  
  if (!defined ($value) or blank ($value)) {
    return ($entered, $selected, @choices);
  }

  # We have some country or another now.
  if (($value eq "UNITED STATES") || ($value eq "USA") || ($value eq "US")) {
    $selected = "UNITED STATES";
    @choices = ("UNITED STATES", $dashes);
    push (@choices, @Countries);
    return ($entered, $selected, @choices);
  }
  
  # Not the US.
  my $field;
  my $found = 0;
  foreach $field (@Countries) {
    if (trim(uc($field)) eq $value) {
      $selected = $value;
      $found = 1;
      last; 
    }
  }  
  
  if ($found) {
    @choices = ($selected, $dashes, "UNITED STATES");
    push (@choices, @Countries);
  } 
  else {
    $entered = $value;
  }
  
  return ($entered, $selected, @choices);
}

#---------------------------------------------------------------------
# state_selection
#---------------------------------------------------------------------

sub state_selection {
   my $self = shift;

   my $entered = "";
   my $selected = "";
   unless (defined $self->{States}) {
     %UsaStates = IU::FileReader->new("$IU::LibDir/$IU::StateList")->get_hash;
     $self->{States} = \%UsaStates;
   }
   %UsaStates = %{$self->{States}};
   my @States = sort keys (%UsaStates);
   my @choices = ($dashes, @States);

   my $value = $self->state;
   $value = uc ($value);
   $value =~ s/\.//g;
   $value = trim ($value);

   if (!defined ($value) or blank ($value)) {
      return ($entered, $selected, @choices);
   }

   # Otherwise, state is selected.
   my $key;
   my $short;
   my $found = 0;
   foreach $key (keys %UsaStates) {
      $short = $UsaStates{$key};
      $short = uc ($short);
      $short = trim ($short);
      if ($short eq $value) {
         $selected = $key;
         $found = 1;
         last; 
      }
   }  

   if ($found) {
      @choices = ($selected, $dashes);
      push (@choices, @States);
   } else {
      $entered = $value;
   }
      
   return ($entered, $selected, @choices);
}

#=====================================================================
1;
