package IW::UserAffil;

=head1 NAME

IW::UserAffil - represents a user affiliation

=head1 DESCRIPTION

Represents a user affiliation.  The associated CGI tags (without 
prefix) are identical to the simple access methods; they are:

   category
   type

=head1 SYNOPSIS

   # Create empty affil, and initiallize:
   my $affil = new IW::User->Affil;    
   $addr->category("USA");
   $addr->type("GOVERNMENT");
    
   # Create affil from ODL:
   my $affil_odl = parse IU::ODL \*STDIN;
   my $affil = IW::User->Affil->from_ODL($affil_odl);

   # Get some info:    
   my $cat   = $affil->category;
   my $type  = $affil->type;

=head2 Fixes for EDG 2.3 and beyond

   For TYPE, the internal representation should be in all uppercase
   letters.  The display should be mixed case.  However, we have been
   wrongly storing values in mixed case.  Persistent sessions may
   contain User objects with these values in mixed case.  Now,
   when writing to ODL, uppercase.  When setting, uppercase.  When
   writing as ODL, use all uppercase.  When printing CGI form, use
   map to get mixed case for pull-down menu.

   For CATEGORY it is even more difficult.  We had been using,
   internally and externally, the values 'USA' and 'NON_USA'.  These
   should have been 'USA' and 'NOT USA'.  So, we need to start storing
   these correctly and writing to ODL correctly.  We'll have to handle
   both values for non-usa users in existing accounts.  For now, the
   solution is in the get/set method.  If the value is 'NON_USA',
   we change it to 'NOT USA'.  Since this method is used by the
   to_ODL() method, we will write the correct thing.

=cut

use IU;
use IU::Chunk;
use IU::Errors;
use IU::Log;       # for debug
use IW::App;
use IW::Util qw(:output :string);

use strict;

# Mapping to check membership:
my %SimpleVar = (CATEGORY => 1,
                 TYPE     => 1);

# Required fields (for is_valid()):
my @RequiredFields = qw(CATEGORY TYPE);

# Valid lengths for data:
my %Max = (TYPE     => 32,   
           CATEGORY => 32);

# Map simple vars to printed labels.
my %Label = (TYPE     => 'Type',   
             CATEGORY => 'Category');

#------------------------------------------------------------
# new
#------------------------------------------------------------
# Constructor.  Creates a new, empty affil.

sub new {
   my $class = shift;                   # Get class.
   my $self = {};                       # Start with empty hash.
   $self->{ERRORS} = new IU::Errors;    # Add empty errors object.
   bless $self, $class;                 # Create the object.
   return ($self);                      # Return object.
}

#------------------------------------------------------------
# is_valid
#------------------------------------------------------------
# Require certain minimal information, as given by the @RequiredFields

sub is_valid {
    my $self = shift;
    foreach (@RequiredFields) {
        (defined($self->{$_}) && ($self->{$_} ne '')) or return undef;
    }
    1;
}

#------------------------------------------------------------
# _simple_ivar [NEWVALUE]
#------------------------------------------------------------
# Internal get/set function.  Not called as a method!

sub _simple_ivar {
   my ($ivar, $self, $value) = @_;
   if (@_ > 2) {
      # Setting.
      if (defined ($value)) {
         # If value string too long, set an error string and return.
         if (length ($value) > $Max{$ivar}) {
            $self->{ERRORS}->set ($Label{$ivar} .
               " ($value) is too long: " . length ($value) .
               ". Maximum is ". $Max{$ivar} . '.');
            return ($self->{ERRORS});
         }
      }
      # Uppercase the value and set it and return it.
      $self->{$ivar} = uc ($value);
      return ($value);
   } else {
      # Getting.  Return value if defined or empty string.
      return (defined ($self->{$ivar}) ? $self->{$ivar} : '');
   }
}

#------------------------------------------------------------
# Public Get/Set methods for category and type affiliations.
#------------------------------------------------------------
sub type     { _simple_ivar ('TYPE',     @_) }
sub category {
   my ($self, $value) = @_;

   # If value is given, store it and return it.
   if ($value) {
      # If value is 'NON_USA', replace with correct value: 'NOT USA'.
      $value = 'NOT USA'  if ($value =~ /^NON/i);
      # Set value in object, and return value.
      _simple_ivar ('CATEGORY', $self, $value);
   } else {
      # Else we are getting.
      # Get current stored value.
      $value = _simple_ivar ('CATEGORY', $self);
      # If value is 'NON_USA':
      if ($value =~ /^NON/i) {
	 # Replace with correct value: 'NOT USA'
	 $value = 'NOT USA';
	 # Set value in object, and return value.
	 _simple_ivar ('CATEGORY', $self, $value);
      } else {
	 # Else, just return the value.
	 $value;
      }
   }
}

#---------------------------------------------------------------------
# errors
#---------------------------------------------------------------------
# Returns Errors object stored in UserAffil object.
# This could have been: sub errors { return ($_[0]->{ERRORS}); }
# EXCEPT that for backwards compatibility with persistent sessions,
# we need to see if the Errors object exists, and if not create one.

sub errors {
   my $self = shift;
   unless ($self->{ERRORS}) {
      $self->{ERRORS} = new IU::Errors;    # Add empty errors object.
   }
   return $self->{ERRORS};
}

#------------------------------
#
# ODL interface
#
#------------------------------

#------------------------------------------------------------
# from_ODL ODLOBJECT
#------------------------------------------------------------
# I<Class or instance method.>
# Initiallize the affiliation from a standard IMS ODL object. The
# instance variables have the same names as the associated parameters.

sub from_ODL {
   my $self = shift;
   my $odl = shift || return (undef);
   my ($param, $value);

   # Be a class method, too!
   ref ($self) or $self = $self->new;
 
   # Read the ODL stream:
   $odl->rewind;
   while (($param, $value) = $odl->next) {
      if (($value->type eq 'SCALAR') && $SimpleVar{$param}) { 
         # Store the value using set method.  This can set Errors.
         # It also forces values to be uppercase (as of EDG 2.3).
         _simple_ivar ($param, $self, $value->text);
      }
      else {
         warn "skipping unrecognized address parameter: '$param'";
      }
   }
   # Return it:
   $self;
}

#------------------------------------------------------------
# to_ODL
#------------------------------------------------------------
# I<Instance method.>
# Convert the address into a standard IMS ODL object.
# Note that we force ODL representation to be uppercase.  This
# was NOT true prior to EDG 2.3.  Older persistent sessions may
# have mixed case stored.

sub to_ODL {
   my $self = shift;
   my $odl = IU::ODL->new_group;
   my $ivar; 
   foreach $ivar (sort keys %SimpleVar) {
      $odl->add_param($ivar, IU::ODL->quoted(uc($self->{$ivar})));
   }
   $odl;
}

#------------------------------
#
# CGI interface
#
#------------------------------

# Valids....
my %Valids = (Category => ['Choose one ...',
                           'USA',
                           'NOT USA'],
              Type     => ['Choose one ...',
                           'Government', 
                           'K-12', 
                           'University', 
                           'Commercial', 
                           'Other']
              );

my %MapInternalTypeToCGI = ('GOVERNMENT' => 'Government',
                            'K-12'       => 'K-12',
                            'UNIVERSITY' => 'University',
                            'COMMERCIAL' => 'Commercial',
                            'OTHER'      => 'Other');

#------------------------------------------------------------
# from_CGI  %FieldsRequired
#------------------------------------------------------------
# I<Class method.>
# Initialize the affiliation info from CGI tags.  It will set values
# into the UserAffil object.  If no object is given, it will create
# a new one.  If there are any errors, such as strings being too long
# or required fields being missing, it will set strings in the Errors
# object in $self, but return as many of the fields as it was able to
# get.  from_CGI reads from the form generated by the print method.
# If the user selected the '(none)' value, that value will NOT be
# stored in the object.  That field will be undef or its previous
# value.
# Input:  A hash of fields and whether they are required.  The hash
#         keys can be 'type' and 'category', indicating they should be
#         gotten.  The values are 1 or 0.  If a field has the value 1,
#         that field is "required".  If a required field is missing
#         or blank, then Error strings are written to the ERRORS
#         object.
# Output: Filled in UserAffil object, with possible error strings
#         stored in its Errors object.

sub from_CGI {
   my $self = shift;
   my %fields_req = @_;

   # Make new, if invoked as a class method:
   ref ($self) or $self = $self->new;

   # If no hash, puke.
   %fields_req or
      puke 600, 'Coding error.  UserAffil::from_CGI must be called ',
                'with a hash of the fields to retrieve.';

   my ($parm, $tag, $value);

   # Note that function name is lowercase of Field name.
   foreach $tag (keys (%fields_req)) {
      # Is tag defined?
      if (defined ($value = $REQ->param ("affil.$tag"))) {
         # Trim the value.
         $value = trim ($value);

         # If value non-blank and not "Choose one ...", set it and process
         # next tag.  Note that if string too long or bad format,
         # Errors will be set.
         if (!blank ($value) and $value ne 'Choose one ...') {
            my $cmd = '$self->' . $tag . '($value)';
            eval ($cmd);
            next;
         }
      }
      # Requested field was not filled in.  If required, add error.
      if ($fields_req{$tag}) {
         my $parm = uc ($tag);   # Parameter is uppercase of function.
         $self->{ERRORS}->set
                          ("$Label{$parm} required but not entered.");
      }
      # Else it is blank and not required, so skip.
   }
   $self;
}

#------------------------------------------------------------
# print  %FieldsRequired
#------------------------------------------------------------
# I<Class/instance methods.>
# Input: Hash of each field to print and whether it is required.
#        Hash keys are 'TYPE' and 'CATEGORY'.  Values are 1 if
#        required (and label is red) or 0 (and label is shown black).
# Output HTML form fields for inputing an affiliation.

sub print {
   my $self = shift;
   my %fields = @_;

   %fields or
      puke 600, 'Coding error.  UserAffil::print must be called ',
                'with a hash of the fields to print.';

   # Make new, if invoked as a class method:
   ref ($self) or $self = $self->new;

   # Type...
   $HTML -> TABLE (CELLPADDING=>1, CELLSPACING=>1) -> TR;
   if (exists ($fields{TYPE})) {
      $HTML -> TD (NOWRAP=>undef)
            -> field_cell (Title=>'Type: ', Req=>$fields{TYPE},
                           Stack=>1)
            -> select_one (Name    =>'affil.type',
                           Choices => $Valids{Type},
                           Current =>
                              $MapInternalTypeToCGI{$self->type})
            -> _field_cell;
      $HTML -> _TD;
   }
   if (exists ($fields{CATEGORY})) {
      $HTML -> TD (NOWRAP=>undef)
            -> field_cell (Title=>'Category: ', Req=>$fields{TYPE},
                           Stack=>1)
            -> select_one (Name    =>'affil.category',
                           Choices => $Valids{Category},
                           Current => $self->category)
            -> _field_cell;
      $HTML -> _TD;
   }
   $HTML -> _TR -> _TABLE;
   1;
}

#------------------------------
#
# CHUNK INTERFACE
#
#------------------------------

#------------------------------------------------------------
# from_chunk CHUNK
#------------------------------------------------------------
# Read a chunk from into the internal data structure.
# A chunk containing "x-error" is considered null and void.

sub from_chunk {
    my ($self, $chunk) = @_;
    my $errmsg;

    # Be a class method, too!
    ref($self) or $self = $self->new;
 
    # Input chunk, checking for errors:
    defined($chunk) or return undef;

    my $var;
    foreach $var (keys %SimpleVar) {
        $self->{$var} = $chunk->value($var);
    }
    $self;
}

#------------------------------------------------------------
1;
