package IW::UserDevel;

=head1 NAME

IW::UserDevel - represents a developer's special user preferences

=head1 DESCRIPTION

This module is used by the developers for their specific needs.


=head1 SYNOPSIS

    # Create empty key, and initiallize:
    my $devel = new IW::User->Devel;    
    
    # Create object from ODL:
    my $devel_odl = parse IU:    :ODL \*STDIN;
    my $devel = IW::User->Devel->from_ODL($devel_odl);

    # Get/set desired debugging mode ('ON', 'OFF', 'DEFAULT').
    #   Associated instance/ODL var: "DEBUG"
    #   WARNING: if defined, overrides the "ForceDebug" value for the site!
    $self->debug_mode('ON');
    if ($self->debug_mode eq 'OFF') { ...
     

=head1 ODL REPRESENTATION

Here's an example of a group called 'WWW_DEVELOPER':

    GROUP         = WWW_DEVELOPER
       DEBUG        = "DEFAULT"
    END_GROUP     = WWW_DEVELOPER


=head1 NOTES

Debugging has been given I<three> possible values:

    ON       Turn debugging on for this session, no matter what ForceDebug is
    OFF      Turn debugging off for this session, no matter what ForceDebug is
    DEFAULT  Do whatever you would normally do.

=cut


use IU;
use IW::App;
use IU::Log;
use IW::Util qw(:output);

use strict;


# Mapping to check membership:
my %SimpleVar = (DEBUG => 1);

# Required fields (for is_valid()):
my @RequiredFields = ();

# Valid lengths for various things:
my %Max = (DEBUG => 3);


#------------------------------------------------------------
# new
#------------------------------------------------------------
# Constructor.  Creates a new, empty affil.
sub new {
    bless {}, shift;
}

#------------------------------------------------------------
# is_valid
#------------------------------------------------------------
sub is_valid {
    my $self = shift;
    1;
}

#------------------------------------------------------------
# debug_mode
#------------------------------------------------------------
# Force debugging on/off?
# Returns 'ON', 'OFF', or 'DEFAULT'.

sub debug_mode { 
    my $self = shift;
    $self->{DEBUG} = $_[0] if (@_ > 1);
    return ($self->{DEBUG} || 'DEFAULT');
}

#------------------------------
#
# ODL interface
#
#------------------------------

#------------------------------------------------------------
# from_ODL ODLOBJECT
#------------------------------------------------------------
# I<Class or instance method.>
# Initiallize from a standard IMS ODL object.
# The instance variables have the same names as the associated parameters.

sub from_ODL {
    my $self = shift;
    # just return undef if no ODL!!!!
    my $odl = shift || return (undef);
    my ($param, $value);

    # Be a class method, too!
    ref($self) or $self = $self->new;
 
    # Read the ODL stream:
    $odl->rewind;
    while (($param, $value) = $odl->next) {
	if (($value->type eq 'SCALAR') && $SimpleVar{$param}) {
            $self->{$param} = $value->text;
	}
	else {
	    warn "skipping unrecognized developer parameter: '$param'";
	}
    }

    # Return it:
    $self;
}

#------------------------------------------------------------
# to_ODL
#------------------------------------------------------------
# I<Instance method.>
# Convert the to a standard IMS ODL object.

sub to_ODL {
    my $self = shift;
    my $odl = IU::ODL->new_group;

    # Debugging:
    $odl->add_param('DEBUG', IU::ODL->quoted($self->{DEBUG}));
    $odl;
}

#------------------------------------------------------------
# from_CGI PARAMHASH
#------------------------------------------------------------
# I<Class or instance method.>
# Initiallize the address from CGI tags.  The PARAMHASH contains
# the following:
#
#     Prefix   =>  The prefix (like "dev.").  Optional; defaults to "".
#
# NOTE: if NO relevant CGI tags are given, returns undef!!!!
# This means that it will ONLY create the object if it was specified!

sub from_CGI {
    my $self = shift;
    my %params = @_;
    my $prefix = $params{Prefix} || '';         # get prefix
    my $value;

    # No CGI tags?  Don't return anything:
    $REQ->param("${prefix}debug") or return undef;    

    # Make new, if needed:
    ref($self) or $self = $self->new;

    # Debug (CGI values are 'ON', 'OFF', and 'DEFAULT'):
    $self->{DEBUG} = uc($REQ->param("${prefix}debug") || 'DEFAULT');
    $self;
}

#------------------------------------------------------------
# print_fields
#------------------------------------------------------------
# I<Class/instance methods.>
# Output HTML form fields for inputting an affiliation.  
# The PARAMHASH contains the following:
#
#     Prefix =>  The prefix (like "ship.").  Optional; defaults to "".
#
# Required fields are shown in red.

sub print_fields {
    my $self = shift;
    my %params = @_;
    my $prefix = $params{Prefix} || '';         # get prefix

    # Make new, if invoked as a class method:
    ref($self) or $self = $self->new;

    $HTML -> TABLE(CELLPADDING=>1, CELLSPACING=>1);
    
    # Debugging:
    $HTML -> TR;
    $HTML -> TD 
	  -> t('Debugging for this session should be in the ')
          -> select_one(Name    =>"${prefix}debug",
			Choices =>['DEFAULT', 'ON', 'OFF'],
			Value   =>($self->{DEBUG} || 'DEFAULT'))
	  -> t(' setting.');
    $HTML -> _TD;
    $HTML -> _TR;

    $HTML -> _TABLE;
    1;
}

#------------------------------------------------------------
# install
#------------------------------------------------------------
# Activates the preferences set here IMMEDIATELY.
#                     "Just push the button, Frank..." - MST3K
#
# Also, updates the session state.

sub install {
    my $self = shift;

    # Debugging...
    if ($self->debug_mode eq 'DEFAULT') {  # don't override site value
	$SESSION->state->{Debug} = undef;      
    }
    else {                                   # override site value!
	my $onoff = (($self->debug_mode eq 'ON') ? 1 : 0);
	$SESSION->state->{Debug} = $onoff;
	$APP->debug_mode($onoff);                   # set it now!
    }
    1;
}

#------------------------------------------------------------
1;
