package IW::UserDisplay;

=head1 NAME

IW::UserDisplay - represents user's presentation preferences


=head1 ODL REPRESENTATION

Here's an example of a group called 'WWW_DISPLAY':

    GROUP            = WWW_DISPLAY
       SMALL_FONT       = Y
       SHOW_EMPTY_COLUMNS_DSET = Y
       SHOW_EMPTY_COLUMNS_GRAN = Y
       SHOW_EMPTY_COLUMNS_CART = Y
       BROWSER_SPEED    = "15"
       JAVA             = Y
       DYNAMIC          = N
       LIST_LENGTH      = "10"
       ADMINISTRATOR    = N
       
       GROUP            = DATASETS
         SORT_BY          = ("DATA_CENTER_ID", 
			     "GRANULE_ID")
         SUMMARY          = ("START_DATE", 
			     "STOP_DATE", 
			     "DATA_CENTER", 
			     "NUM_GRANULES")
       END_GROUP        = DATASETS
      
       GROUP            = GRANULES
         SORT_BY          = ("DATASET_ID", 
			     "GRANULE_ID")
         SUMMARY          = ("START_DATE", 
			     "STOP_DATE", 
			     "SENSOR_NAMES")
       END_GROUP        = GRANULES
    
    END_GROUP        = WWW_DISPLAY


=head1 ATTRIBUTES

This is a list of all standard names used by this module in ODL 
and some CGI input.

    DATASET_ID           Name of data set
    DATA_CENTER_ID       Name of data center
    GEO_CENTER           Geographic center
    GEO_COVERAGE         Geographic coverage
    GRANULE_ID           Name of granule
    NUM_GRANULES         Number of granules, in results
    SENSOR_NAMES         Sensor names
    SOURCE_NAMES         Source names
    START_DATE           Earliest start date, in results
    STOP_DATE            Latest stop date, in results


=head2 Dataset attributes

Possible values to sort on are:

    DATASET_ID           Name of data set
    DATA_CENTER_ID       Name of data center
    NUM_GRANULES         Number of granules, in results
    START_DATE           Earliest start date, in results
    STOP_DATE            Latest stop date, in results
   
Possible values to show in summary lists are these sort keys, plus:

    SENSOR_NAMES         Sensor names, across all results
    SOURCE_NAMES         Source names, across all results


=head2 Granule attributes

Possible values to sort on are:

    DATASET_ID           Name of data set
    DATA_CENTER_ID       Name of data center
    GRANULE_ID           Name of granule
    START_DATE           Start date
    STOP_DATE            Stop date

Possible values to show in summary lists are these sort keys, plus:    

    GEO_CENTER           Geographic center
    GEO_COVERAGE         Geographic coverage
    SENSOR_NAMES         Sensor names
    SOURCE_NAMES         Source names

=cut

use IU;
use IU::Errors;
use IU::Log;
use IU::Query;
use IU::WWWValids;
use IW::App;
use IW::Util qw(:url :output);
use IW::Babel;
use IW::Sort;

use strict;

# Simple variables:
my @SimpleVars = 
    qw(SMALL_FONT
       SHOW_EMPTY_COLUMNS_DSET
       SHOW_EMPTY_COLUMNS_GRAN
       SHOW_EMPTY_COLUMNS_CART
       LIST_LENGTH
       BROWSER_SPEED
       JAVA
       DYNAMIC
       ADMINISTRATOR
       GUIDE_LIMIT
       );

# Summary attributes, datasets:
my @DsetSummaryVars = 
    qw(AREA_LOC
       CAMPAIGN 
       DATA_CENTER_ID 
       MD_ENTRY 
       NUM_GRANULES  
       PROCESSING_LEVEL 
       SENSOR_NAMES  
       SOURCE_NAMES  
       START_DATE    
       STOP_DATE     
       );

# Summary attributes, granules:
my @GranSummaryVars = 
    qw(CAMPAIGN
       DATA_CENTER_ID
       GEO_CENTER
       GEO_COVERAGE
       PROCESSING_LEVEL
       SENSOR_NAMES
       SOURCE_NAMES
       START_DATE  
       STOP_DATE
       );

# Summary styles:
my @SummaryStyles = qw(ONE-LINE-PER-ITEM TWO-COLUMN HORIZONTALLY VERTICALLY);

# Summary menus:
my @SummaryMenus  = qw(COMPACT LARGE);

# Primary Action button Size:
my @PrimaryActSiz = qw(LARGE COMPACT);

# Max sort keys:
my $MaxSortKeys = 3;

# Map browser connection speed to comm. stat. refresh rate (in seconds).
my %SpeedForRefreshRate = (
                           '60' => 'Very Slow',
                           '30' => 'Slow',
                           '15' => 'Medium',
                           '10' => 'Fast',
                            '5' => 'Very Fast'
                          );

# Map comm. stat. refresh rate (in seconds) to browser connection speed.
my %RefreshRateForSpeed = (
                           'Very Slow' => '60',
                           'Slow'      => '30',
                           'Medium'    => '15',
                           'Fast'      => '10',
                           'Very Fast' =>  '5'
       	                  );

#------------------------------------------------------------
# new
#------------------------------------------------------------
# Create a new instance.  Some defaults are provided!

sub new {
    my $class = shift;
    my $self = bless {}, $class;

    # Init general:
    $self->{SMALL_FONT}      = 'Y';
    $self->{SHOW_EMPTY_COLUMNS_DSET} = 'N';
    $self->{SHOW_EMPTY_COLUMNS_GRAN} = 'N';
    $self->{SHOW_EMPTY_COLUMNS_CART} = 'N';
    $self->{BROWSER_SPEED} = ${IU::UpdateEvery};

    $self->{LIST_LENGTH} = $IU::ListLength ? $IU::ListLength : 20;

    $self->{METADATA_FLAG} = "DEFAULT";

    $self->{JAVA}          = $IU::AllowJava;
    $self->{JAVASCRIPT}    = "Y";

    $self->{DYNAMIC}       = 'N';
    $self->{GUIDE_LIMIT} = ${IU::GuideLimit} || 100;

    $WWWVALIDS = IU::WWWValids->new() if (!defined($WWWVALIDS));

    # Init dataset options:
    $self->{DSET_SORT_BY} = new IW::Sort::Dataset;
    $self->{DSET_SUMMARY} = $WWWVALIDS->dsetListDefaults();

    # Init granule options:
    $self->{GRAN_SORT_BY} = new IW::Sort::Granule;
    $self->{GRAN_SUMMARY} = $WWWVALIDS->granListDefaults(); 

    $self->{ERRORS} = new IU::Errors;    # Add empty errors object.

    $self->update_from_session() if defined $SESSION;

    $self;
}

sub update_from_session {
   my $self = shift;

   $self->{BROWSER_SPEED} = $SESSION->state->{RefreshRate}
      if exists $SESSION->state->{RefreshRate};

   $self->{LIST_LENGTH} = $SESSION->state->{ListLength}
      if exists $SESSION->state->{ListLength};

   $self->{METADATA_FLAG} = $SESSION->state->{MetadataFlag}
      if exists $SESSION->state->{MetadataFlag};

   $self->{JAVA} = $SESSION->state->{AllowJava}
      if exists $SESSION->state->{AllowJava};

   #$self->{GUIDE_LIMIT} = $SESSION->state->{Query}->max_guides
   #   if exists $SESSION->state->{Query};
   if ($SESSION->state->{Query}) {
      $self->{GUIDE_LIMIT} = $SESSION->state->{Query}->max_guides;
   }

   $self->{JAVASCRIPT} = $SESSION->state->{JavascriptDesired} &&
                         $SESSION->javascript_capable();
   $self;
}

#---------------------------------------------------------------------
# errors
#---------------------------------------------------------------------
# Returns Errors object stored in UserAffil object.
# This could have been: sub errors { return ($_[0]->{ERRORS}); }
# EXCEPT that for backwards compatibility with persistent sessions,
# we need to see if the Errors object exists, and if not create one.

sub errors {
   my $self = shift;
   unless ($self->{ERRORS}) {
      $self->{ERRORS} = new IU::Errors;    # Add empty errors object.
   }
   return $self->{ERRORS};
}

#------------------------------
#
# ACCESS METHODS
#
#------------------------------


#------------------------------------------------------------
# _generic_stash
#------------------------------------------------------------
# Stashes whatever you give it into the object (and, optionally, the session).
# If there is nothing given to stash, it returns the value in the object
# (and sets the appropriate field in the session state if you passed a field
# name).
#
# Note: excess dependencies on knowing the field names in the session state.
#       We would be better off with a copy of the User object in there!

sub _generic_stash {
    my ($self, $my_field, $session_field, $value) = @_;
    # current setting is the value passed in, if there is one.
    # if not, use the current value in this object.
    my $current_setting = (defined $value ? $value
                                         : $self->{$my_field});
    $self->{$my_field} = $current_setting;

    $SESSION->state->{$session_field} = $current_setting 
        if defined $session_field;

    $self->{$my_field};
}

#------------------------------------------------------------
# _generic_YN_stash
#------------------------------------------------------------
# Stashes Y or N data into the session. 

sub _generic_YN_stash {
    my ($self, $my_field, $session_field, $value) = @_;

    $value = uc(substr $value,0,1) if defined $value;

    $self->_generic_stash($my_field, $session_field, $value);
}

#------------------------------------------------------------
# _generic__YN_predicate
#------------------------------------------------------------
# 1 for Y, 0 for N.

sub _generic_YN_predicate {
    my ($self,$field) = @_;

    return 0 unless defined $self->{$field};

    ($self->{$field} eq 'Y') ? 1 : 0;
}
 
#------------------------------------------------------------
# small_font_ok
#------------------------------------------------------------
# Will they allow small fonts?  Returns true (the default) or false.

sub small_font_ok {
    my $self = shift;
    $self->_generic_YN_predicate('SMALL_FONT');
}

#------------------------------------------------------------
# show_empty_columns_dset_ok 
#------------------------------------------------------------
# Show empty columns in a dataset list.  Returns true (the default) or false.

sub show_empty_columns_dset_ok {
    my $self = shift;
    $self->_generic_YN_predicate('SHOW_EMPTY_COLUMNS_DSET');
}

#------------------------------------------------------------
# show_empty_columns_gran_ok
#------------------------------------------------------------
# Show empty columns in a granule list.  Returns true (the default) or false.

sub show_empty_columns_gran_ok {
    my $self = shift;
    $self->_generic_YN_predicate('SHOW_EMPTY_COLUMNS_GRAN');
}

#------------------------------------------------------------
# show_empty_columns_cart_ok
#------------------------------------------------------------
# Show empty columns in a cart list.  Returns true (the default) or false.

sub show_empty_columns_cart_ok {
    my $self = shift;
    $self->_generic_YN_predicate('SHOW_EMPTY_COLUMNS_CART');
}

#------------------------------------------------------------
# small_font [VALUE]
#------------------------------------------------------------
# Will they allow small fonts?  Returns 'Y' or 'N'.
# With arg, sets to that value.  Accepts 'Yes...' and 'No...', also.

sub small_font {
    my $self = shift;
    $self->_generic_YN_stash('SMALL_FONT', 'AllowSmallFont', @_);
}

#------------------------------------------------------------
# show_empty_columns_dset [VALUE]
#------------------------------------------------------------
# Show empty columns in a dataset results?  Returns 'Y' or 'N'.
# With arg, sets to that value.  Accepts 'Yes...' and 'No...', also.

sub show_empty_columns_dset {
    my $self = shift;
    $self->_generic_YN_stash('SHOW_EMPTY_COLUMNS_DSET', 'AllowSmallFont', @_);
}

#------------------------------------------------------------
# show_empty_columns_gran [VALUE]
#------------------------------------------------------------
# Show empty columns in a granule results?  Returns 'Y' or 'N'.
# With arg, sets to that value.  Accepts 'Yes...' and 'No...', also.

sub show_empty_columns_gran {
    my $self = shift;
    $self->_generic_YN_stash('SHOW_EMPTY_COLUMNS_GRAN', 'AllowSmallFont', @_);
}

#------------------------------------------------------------
# show_empty_columns_cart [VALUE]
#------------------------------------------------------------
# Show empty columns in a cart results?  Returns 'Y' or 'N'.
# With arg, sets to that value.  Accepts 'Yes...' and 'No...', also.

sub show_empty_columns_cart {
    my $self = shift;
    $self->_generic_YN_stash('SHOW_EMPTY_COLUMNS_CART', 'AllowSmallFont', @_);
}
#------------------------------------------------------------
# browser_speed [VALUE]
#------------------------------------------------------------
# Get/Set method for browser speed.
# VALUE can be 'Very Slow', 'Slow', 'Medium', 'Fast', or 'Very Fast'.
sub browser_speed {
   my $self = shift;
   $_[0] and $self->{BROWSER_SPEED} = $RefreshRateForSpeed{$_[0]};
   $self->{BROWSER_SPEED};
}

#------------------------------------------------------------
# java_ok
#------------------------------------------------------------
# Will they allow java?  Returns true (the default) or false.

sub java_ok {
    my $self = shift;
    $self->_generic_YN_predicate('JAVA');
}

#------------------------------------------------------------
# javascript_ok
#------------------------------------------------------------
# Will they allow Javascript?  Returns true (the default) or false.

sub javascript_ok {
    my $self = shift;
    $self->_generic_YN_predicate('JAVASCRIPT');
}

#------------------------------------------------------------
# java [VALUE]
#------------------------------------------------------------
# Will they allow java?  Returns 'Y' or 'N'.
# With arg, sets to that value.  Accepts 'Yes...' and 'No...', also.

sub java {
    my $self = shift;
    $self->_generic_YN_stash('JAVA', 'AllowJava', @_);
}

#------------------------------------------------------------
# javascript [VALUE]
#------------------------------------------------------------
# Do they want Javascript?  Returns 'Y' or 'N'.
# With arg, sets to that value.  Accepts 'Yes...' and 'No...', also.

sub javascript {
    my $self = shift;
    $self->_generic_YN_stash('JAVASCRIPT', 'JavascriptDesired', @_);
}

#------------------------------------------------------------
# dynamic_ok
#------------------------------------------------------------
# Will they allow dynamic valids?  Returns false (the default) or true.
# Does not use generic routine because of possible override from session state.

sub dynamic_ok {
    my $self = shift;
    $self->_generic_YN_predicate('DYNAMIC') or
            (($SESSION->state->{AllowDynamicValids} eq 'Y') ? 1 : 0);
}

#------------------------------------------------------------
# dynamic [VALUE]
#------------------------------------------------------------
# Will they allow dynamic?  Returns 'Y' or 'N'.
# With arg, sets to that value.  Accepts 'Yes...' and 'No...', also.

sub dynamic {
    my $self = shift;
    $self->_generic_YN_stash('DYNAMIC', 'AllowDynamicValids', @_);
}

#------------------------------------------------------------
# maxguides
#------------------------------------------------------------

sub maxguides {
   my ($self, $maxguides) = @_;

   if ($maxguides) {
      $maxguides =~ s/\s//g;
      if ($maxguides =~ /\D/) {
         $self->{ERRORS}->set 
                    ("Maximum number of guide documents must be a " .
                     "positive whole number, you entered: \"$maxguides\".");
      }
      else {
         $SESSION->state->{Query}->max_guides($maxguides);
         $SESSION->state->{Query}->install;
      }
   }
}

#------------------------------------------------------------
# list_length [VALUE]
#------------------------------------------------------------
# Returns the list_length  With arg, sets to that value.

sub list_length {
    my $self = shift;
    my $length = $self->_generic_stash('LIST_LENGTH','ListLength',@_);
    if (!(defined $length)){
       $length = ${IU::ListLength} ||  20;
    }
    return $length;
}

#------------------------------------------------------------
# summary_style [VALUE]
#------------------------------------------------------------
# Returns the summary style, in all caps.  With arg, sets to that value. 
# Note we do not set a corresponding field in the session state anymore.
#
#sub summary_style {
#    my $self = shift;
#    $self->_generic_stash('SUMMARY_STYLE', undef, @_);
#}

#------------------------------------------------------------
# summary_menu [VALUE]
#------------------------------------------------------------
# Returns the summary menu, in all caps.  With arg, sets to that value.
#
#sub summary_menu {
#    my $self = shift;
#    $self->_generic_stash('SUMMARY_MENU', undef, @_);
#}

#------------------------------------------------------------
# primary_act_siz [VALUE]
#------------------------------------------------------------
# Returns the primary action button size.  With arg, sets to that value.
#
#sub primary_act_siz {
#    my $self = shift;
#    $self->_generic_stash('PRIMARY_ACT_SIZ', undef, @_);
#}

#------------------------------------------------------------
# admin [VALUE]
#------------------------------------------------------------
# Are they an administrator?  Returns 'Y' or 'N', or false if undefined.
# With arg, sets to that value.  Accepts 'Yes...' and 'No...', also.

sub admin {
    my $self = shift;
    $self->_generic_YN_stash('ADMINISTRATOR', undef, @_);
}

#------------------------------------------------------------
# is_valid
#------------------------------------------------------------
sub is_valid {
    1;
}

#------------------------------------------------------------
# dataset_sort
# dataset_summary
#------------------------------------------------------------
sub dataset_sort    { $_[0]->{DSET_SORT_BY} }
sub dataset_summary { $_[0]->{DSET_SUMMARY} }

#------------------------------------------------------------
# granule_sort
# granule_summary
#------------------------------------------------------------
sub granule_sort    { $_[0]->{GRAN_SORT_BY} }
sub granule_summary { $_[0]->{GRAN_SUMMARY} }

#------------------------------
#
# ODL INTERFACE
#
#------------------------------

#------------------------------------------------------------
# from_ODL ODLOBJECT
#------------------------------------------------------------
# I<Class or instance method.>
# Initiallize the address from a standard IMS ODL object.
# The instance variables have the same names as the associated parameters.

sub from_ODL {
    my $self = shift;
    my $odl = shift || return (undef);
    my ($param, $value_odl);

    # Be a class method, too!
    ref($self) or $self = $self->new;

    # Read some of the basic params:
    my $ivar;
    foreach $ivar (@SimpleVars) {
	if ($value_odl = $odl->param($ivar)) {
	    $self->{$ivar} = $value_odl->text;
	}
    }

    # Check for guide limit, if we have it saved in a file,
    # we need to read it and set the query object with the value.
    my $maxguides;
    if (($maxguides = $self->{GUIDE_LIMIT}) && $SESSION) {
       unless ($SESSION->state->{Query}) {
          $SESSION->state->{Query} = IU::Query->new;
       }
       $SESSION->state->{Query}->max_guides($maxguides);
       $SESSION->state->{Query}->install;
    }

    # Get the dataset info, if any:
    my $dset_odl;
    if ($dset_odl = $odl->param('DATASETS')) {

	if ($value_odl = $dset_odl->param('SORT_BY')) {
	    $self->{DSET_SORT_BY} = new IW::Sort::Dataset;
	    $self->{DSET_SORT_BY}->sequence(map {var2short($_->text)}
					    $value_odl->as_array);
	}
	if ($value_odl = $dset_odl->param('SUMMARY')) {
	    $self->{DSET_SUMMARY} = [map {$_->text} $value_odl->as_array];
	}
    }
    
    # Get the granule info, if any:
    my $gran_odl;
    if ($gran_odl = $odl->param('GRANULES')) {

	if ($value_odl = $gran_odl->param('SORT_BY')) {
	    $self->{GRAN_SORT_BY} = new IW::Sort::Granule;
	    $self->{GRAN_SORT_BY}->sequence(map {var2short($_->text)}
					    $value_odl->as_array);	    
	}
	if ($value_odl = $gran_odl->param('SUMMARY')) {
	    $self->{GRAN_SUMMARY} = [map {$_->text} $value_odl->as_array];
	}
    }
    
    # Return it:
    $self;
}

#------------------------------------------------------------
# to_ODL
#------------------------------------------------------------
# I<Instance method.>
# Convert into a standard IMS ODL object.

sub to_ODL {
    my $self = shift;
    my $value;

    # Create group, and add stuff in standard order:
    my $odl = IU::ODL->new_group;


    # SIMPLE ATTRIBUTES:

    # ...not concealed...
    my $ivar;
    foreach $ivar (qw(SMALL_FONT 
                      SHOW_EMPTY_COLUMNS_DSET 
                      SHOW_EMPTY_COLUMNS_GRAN
                      SHOW_EMPTY_COLUMNS_CART
                      LIST_LENGTH 
                      BROWSER_SPEED 
                      JAVA 
                      DYNAMIC )) {
	$odl->add_param($ivar, IU::ODL->quoted($self->{$ivar}));
    }

    # ...concealed...
    if ($self->admin) {   
	$odl->add_param('ADMINISTRATOR',IU::ODL->quoted($self->admin));
    }

    # MAXIMUN GUIDE DOCUMENTS
    $odl->add_param('GUIDE_LIMIT', 
                    IU::ODL->quoted($SESSION->state->{Query}->max_guides));

    # DATASET ATTRIBUTES:
    my $dset_odl = IU::ODL->new_group;
    $odl->add_param('DATASETS', $dset_odl);

    # ...sort...
    if ($self->{DSET_SORT_BY} && ($self->{DSET_SORT_BY}->sequence)) {
	my @vars = map {short2var($_)} $self->{DSET_SORT_BY}->sequence;
	$dset_odl->add_param('SORT_BY', IU::ODL->conv(\@vars));
    }
    # ...summary...
    if (@{$self->{DSET_SUMMARY}}) {
	$dset_odl->add_param('SUMMARY',
			   IU::ODL->conv($self->{DSET_SUMMARY}));
    }

    # GRANULE ATTRIBUTES:
    my $gran_odl = IU::ODL->new_group;
    $odl->add_param('GRANULES', $gran_odl);

    # ...sort...
    if ($self->{GRAN_SORT_BY} && ($self->{GRAN_SORT_BY}->keys)) {
	my @vars = map {short2var($_)} $self->{GRAN_SORT_BY}->sequence;
	$gran_odl->add_param('SORT_BY', IU::ODL->conv(\@vars));
    }
    # ...summary...
    if (@{$self->{GRAN_SUMMARY}}) {
	$gran_odl->add_param('SUMMARY',
			   IU::ODL->conv($self->{GRAN_SUMMARY}));
    }
    # Done!
    $odl;
}

#------------------------------
#
# CGI INTERFACE
#
#------------------------------

#------------------------------------------------------------
# from_CGI PARAMHASH
#------------------------------------------------------------
# I<Class/instance method.>
# Initiallize the address from CGI tags.  The PARAMHASH contains
# the following:
#
#     Prefix =>  The prefix (like "ship.").  Optional; defaults to "".
#
# NOTE: since checkboxes are absent from CGI input when set "off",
# all information in the object MUST be cleared first!

sub from_CGI {
    my $class = shift;
    my %params = @_;
    my $prefix = $params{Prefix} || '';         # get prefix
    my ($i, $value);

    my $self;
    if (ref ($class)) {  # Given a User object; use it.
       $self = $class;
    } elsif (defined ($SESSION->state->{User})) {  
       # Get current session User Display object or create a new one:
       $self = $SESSION->state->{User}->display || $class->new;
    } else {
       $self = $class->new;
    }

    # Simple params.
    # Note that a missing param (like 'admin') sets to an empty/undef value! 

    $self->browser_speed($REQ->param("${prefix}browser_speed"));
    $self->small_font($REQ->param("${prefix}small_font"));
    $self->java($REQ->param("${prefix}java"));
    $self->dynamic($REQ->param("${prefix}dynamic"));
    $self->maxguides($REQ->param("${prefix}maxguides"));
    $self->admin($REQ->param("${prefix}admin"));

    # Done!
    $self;
}

#------------------------------------------------------------
# from_CGI_customize PARAMHASH
#------------------------------------------------------------
# I<Class/instance method.>

sub from_CGI_customize {
    my $class = shift;
    my %params = @_;
    my $prefix = $params{Prefix} || '';         # get prefix

   my $self;
   if (ref ($class)) {  # Given a User object; use it.
      $self = $class;
   } elsif (defined ($SESSION->state->{User})) {
      # Get current session User Display object or create a new one:
      $self = $SESSION->state->{User}->display || $class->new;
   } else {
      $self = $class->new;
   }
    # Get list length (pagination size);
    $self->list_length($REQ->param("list_length"));

    # get current columns from hidden CGI tags
    # For the JavaScript version, we will filter out hidden keys with
    # null.  This version requires all possible values set and then
    # user hits OK, a JavaScript function is called which sets the 
    # tags within the current columns list and sets all others to null.
    my $i=1;
    my $defindex=0;
    my @default_columns;
    while ($REQ -> param("curlist.$i")) {
       if (($REQ -> param("curlist.$i")) =~ "null") { }
       else {
          $default_columns[$defindex++] = $REQ->param("curlist.$i");
       }
       $i++;
    }

  # Need if cond on whether exec from DataSet or Granule results
  # Hardcode to granule for now.
   my $listype = ($REQ->param('type') || 'gkey');

   if ($listype eq 'dskey') {
      # Get the dataset sort option:
      $self->{DSET_SORT_BY} = IW::Sort::Dataset->from_CGI();
      # Set the dataset summary option:
      # Have to convert the pretty name from CGI tags to variable('ugly') name
      @default_columns = $WWWVALIDS->dsetListUgly(\@default_columns);
      #dataset_summary(@default_columns);
      $self->{DSET_SUMMARY} = [@default_columns];

      $self->show_empty_columns_dset((defined $REQ->param('showemptycols'))?'Y':'N');
      $self->javascript( $REQ->param('javascriptOK') );
      # Update the date range session state variable with CGI tag value
      $SESSION->state->{DateType} =  $REQ->param('datetype');
    }
    else {           
      # Get the granule sort option:
      $self->{GRAN_SORT_BY} = IW::Sort::Granule->from_CGI();

      # Set the granule summary option:
      # Have to convert the pretty name from CGI tags to variable('ugly') name
      @default_columns = $WWWVALIDS->granListUgly(\@default_columns);
      $self->{GRAN_SUMMARY} = [@default_columns];

      $self->show_empty_columns_gran((defined $REQ->param('showemptycols'))?'Y':'N');
      $self->javascript( $REQ->param('javascriptOK') );
      # Update the date range session state variable with CGI tag value
      $SESSION->state->{DateType} =  $REQ->param('datetype');
    }
    # Done!
    $self;
}

#------------------------------------------------------------
# from_CGI_metadata PARAMHASH
# This method is called when a user switches between cgi and
# javascript while customizing metadata in ModeMETADATA.
#------------------------------------------------------------
# I<Class/instance method.>

sub from_CGI_metadata {
    my $class = shift;
    my %params = @_;
    my $prefix = $params{Prefix} || '';         # get prefix

   my $self;
   if (ref ($class)) {  # Given a User object; use it.
      $self = $class;
   } elsif (defined ($SESSION->state->{User})) {
      # Get current session User Display object or create a new one:
      $self = $SESSION->state->{User}->display || $class->new;
   } else {
      $self = $class->new;
   }

   # Toggle between cgi and javascript, if flag is not set
   # switch to javascript because default is CGI.
   my $ java_script_flag = $self->javascript;
   if ($java_script_flag eq 'N') {
      $self->javascript ('Y');
   } elsif ($java_script_flag eq 'Y') {
      $self->javascript ('N');
   } else {  
      $self->javascript ('Y');
   }

   # Done!
    $self;
}
#------------------------------------------------------------
# print_fields
#------------------------------------------------------------
# Output HTML form fields for inputting.  The PARAMHASH contains
# the following:
#
#     Prefix =>  The prefix (like "ship.").  Optional; defaults to "".
#
# Required fields are shown in red.

sub print_fields {
    my $self = shift;
    my %params = @_;
    my $prefix = $params{Prefix} || '';
    my $var;

    my $query = $SESSION->state->{Query};
    my $max_guides = ($query ? 
                      $query->max_guides : ($IU::Query::MAX_GUIDES_DEFAULT));

    $HTML -> TABLE(CELLPADDING=>0, CELLSPACING=>0);

    # Use Java? 
    $HTML -> TR -> TD;
    print "Use Java where it is available?  ";
    $HTML -> select_one(Name   =>"${prefix}java",
                        Choices=>['YES', 'NO'],
                        Default=>($self->java_ok ? 'YES' : 'NO'));
    $HTML->hidden("${prefix}java" => 'NO');

    # Use Javascript? 
    if ($SESSION->javascript_capable()) {
       $HTML -> _TR -> _TD -> TR -> TD;
       print "Use Javascript where it is available?  ";
       $HTML -> select_one(Name   =>"${prefix}javascript",
                           Choices=>['YES', 'NO'],
                           Default=>($self->javascript_ok ? 'YES' : 'NO'));
       $HTML->hidden("${prefix}javascript" => 'NO');
    }

     $HTML -> _TD -> _TR ;

    # Use Dynamic Valids? 
    $HTML -> _TD -> _TR -> TR -> TD;
    print "Use Dynamic Valids in generating a query?  ";
    $HTML -> select_one(Name   =>"${prefix}dynamic",
                        Choices=>['YES', 'NO'],
                        Default=>($self->dynamic_ok ? 'YES' : 'NO'));
    $HTML -> _TD -> _TR;

    # Maximun guide documents
    $HTML -> TR -> TD;
    $HTML -> t("Return a maximum number of ")
          -> input_text(Size    => 5,
                        Name    => "${prefix}maxguides",
                        Default => $max_guides)
          -> t(' ') -> t("detailed documents.");

    # Small fonts?
    $HTML -> _TD -> _TR -> TR -> TD;
    print "Use <B>small fonts</B> <FONT SIZE=-1>(like this)</FONT> ",
          "when appropriate? ";
    $HTML -> select_one(Name   =>"${prefix}small_font",
			Choices=>['YES', 'NO'], 
			Default=>($self->small_font_ok ? 'YES' : 'NO'));
    $HTML -> _TD -> _TR;

    # Browser+ConnectionSpeed?
    # This determines refresh rate for comm stat screen.
    $HTML -> TR -> TD;
    print "My browser and internet connection speed seem to be: ";
    # If value is from SETUP it may not match a value in the
    # SpeedForRefreshRate hash, so let's pick the closest.
    my $speed;
    if (exists ($SpeedForRefreshRate{$self->{BROWSER_SPEED}})) {
       $speed = $SpeedForRefreshRate{$self->{BROWSER_SPEED}};
    } else {
       my @keys = keys (%SpeedForRefreshRate);
       # diff is the list of absolute differences between the rates
       # and the rate that was set in SETUP.
       my @diff = map {abs ($_ - $self->{BROWSER_SPEED}) } @keys;
       # Which key element has the lowest diff?
       my $lowest = 0;
       my $i;
       for ($i=1; $i < $#keys; $i++) {
          $lowest = $i if ($diff[$i] < $diff[$lowest]);
       }
       $speed = $SpeedForRefreshRate{$keys[$lowest]};
    }
    $HTML -> select_one(Name   =>"${prefix}browser_speed",
			Choices=>[sort by_rate keys(%RefreshRateForSpeed)],
			Default=>($speed));
    $HTML -> _TD -> _TR;

    # Administrator?  OUTPUT CHOICE ONLY IF CURRENTLY DEFINED!
    if ($self->admin) {          # defined... so they've been here
	$HTML -> TR -> TD;
	print "Display <B>Site Administrator</B> link? ";
	$HTML -> select_one(Name   =>"${prefix}admin",
			    Choices=>['YES', 'NO'], 
			    Default=>(($self->admin =~ /^Y/i) ? 'YES' : 'NO'));
	$HTML -> _TD -> _TR;
    }
    $HTML -> _TABLE;

}
#------------------------------------------------------------
# Sort routine comparator for refresh rates.
#------------------------------------------------------------
sub by_rate {
   $RefreshRateForSpeed{$a} <=> $RefreshRateForSpeed{$b};
}

#------------------------------------------------------------
# install
#------------------------------------------------------------
# Install (activate) these preferences in a user's session.

sub install {
    my $self = shift;

    # Set small font in session state, *and* activate it now!
    $SESSION->state->{AllowSmallFont} = $self->small_font;
    $HTML->allow_small($self->small_font);

    $SESSION->state->{RefreshRate}        = $self->browser_speed;
    $SESSION->state->{AllowJava}          = $self->java;
    $SESSION->state->{AllowJavascript}    = $self->javascript;
    $SESSION->state->{AllowDynamicValids} = $self->dynamic;
    $SESSION->state->{ListLength}         = $self->list_length;

    # Set dataset/granule sorts in session state!
    if ($self->{DSET_SORT_BY}) {
	$SESSION->state->{LastDatasetSort} = $self->{DSET_SORT_BY};
    }
    if ($self->{GRAN_SORT_BY}) {
	$SESSION->state->{LastGranuleSort} = $self->{GRAN_SORT_BY};
    }
    1;
}
#------------------------------------------------------------
1;
