package IW::UserKeys;

=head1 NAME

IW::UserKey - represents a user data access keys.

=head1 DESCRIPTION

This module is used for the ASF Authenticator Key, ECS
Authenticator Key, and the NASDA Authenticator Key.
The associated CGI tags (without prefix)
are identical to the simple access methods; they are:

   authentication_key
   ecs_username
   ecs_authentication_key
   old_ecs_authentication_key
   nasda_authentication_key

These are also lowercase versions of the hash keys in $self.

=head2 DATA 

=begin list

=item

%SimpleVar

Map of acceptable variables. Contains:
AUTHENTICATION_KEY, ECS_USERNAME, ECS_AUTHENTICATION_KEY, HOME_DAAC,
OLD_ECS_AUTHENTICATION_KEY,NASDA_AUTHENTICATION_KEY

=item

%Max 

Map maximum valid string lengths for the above simple variables.

=item

%Label

Map simple variables to printed labels for user interface.

=item

%Internal

Map external key names to internal get/set function calls.
   'V0'  => ['authentication_key']
   'ECS' => ['ecs_username', 'ecs_authentication_key']
   'NASDA' => ['nasda_authentication_key']

=item

$self

The usual hash reference will have the above simple variables and one
new one: ERRORS => IU::Errors object.  This errors object holds any
error strings generated when keys are read from CGI or being set.

=end list

=head2 METHODS

=begin list

=item

new ( )

Constructor. Now adds an empty IU::Errors object to the $self hash.

=item

Internal get/set method for the above simple variables. If setting and
length of $NEWVALUE is too long, set error strings in Errors object
and return that object.

=item

authentication_key ([$NEWVALUE])

Get/set method for AUTHENTICATION_KEY.  Calls _simple_ivar.

=item

nasda_authentication_key ([$NEWVALUE])

Get/set method for the NASDA_AUTHENTICATION_KEY.  Calls _simple_ivar.

=item

ecs_username ([$NEWVALUE])

Get/set method for ECS_USERNAME. If setting, checks for valid
characters. If invalid, sets error string in Errors object and returns
that object.  Otherwise, calls _simple_ivar.

=item

ecs_authentication_key ([$NEWVALUE])

Get/set method for ECS_AUTHENTICATION_KEY.  Calls _simple_ivar.

=item

errors ( )

New public method to access Errors object in the UserKeys object.

=item

v0_ok_to_go ( )

Returns true if the AUTHENTICATION_KEY is set to a true value.

=item

nasda_ok_to_go ( )

Returns true if the NASDA_AUTHENTICATION_KEY is set to a true value.

=item

ecs_ok_to_go ( )

Returns true if the ECS_USERNAME and ECS_AUTHENTICATION_KEY are set.

=item

from_ODL (IU::ODL)

Get UserKeys object from an IU::ODL object.  Now uses _simple_ivar to
do setting so that if a field is too long an Error string will be set.

=item

to_ODL ( )

Adds ODL parameter values for the simple variables to the ODL object.

=item

from_CGI (%PARAMHASH)

Method for reading keys from a CGI form.  It now uses the Errors
object instead of dying and puking, if there are errors.  It accepts
a hash of the keys to read and whether they are required.  The
keys are "ECS" and/or "V0" and/or "NASDA".  The values are 1 or 0. 
If a value is 1, that key is required and an error string will be
generated if the required field is left blank.

=endlist

=item

print (%KEYSREQUIRED)

Modify to accept a hash of keys to print. The acceptable keys are:
'ECS' and/or 'V0' and/or 'NASDA'. The values are 1 or 0. If the value is 1, the
labels are printed red, indicating they are required.

=endlist

=head1 SYNOPSIS

   # Create empty key, and initiallize:
   my $key = new IW::User->Key;    
   $addr->authentication_key("");
    
   # Create from ODL:
   my $keys_odl = parse IU::ODL \*STDIN;
   my $key = IW::User->Keys->from_ODL($keys_odl);

   # Get some info:    
   my $key_info   = $key->authentication_key;

=head1 NOTES

   Note that the prefix for keys is '_dak.'.  The leading underscore
   prevents the keys's values from being shown in debug and in dying
   gasps.

   Simple variable names are uppercase versions of get/set methods.

=cut

#=====================================================================
use IU;
use IU::Log;
use IU::Errors;
use IW::App;
use IW::Util qw(:string);

use strict;

# Mapping to check membership:
my %SimpleVar = map {($_ => 1)}
   qw (AUTHENTICATION_KEY ECS_USERNAME ECS_AUTHENTICATION_KEY HOME_DAAC
       OLD_ECS_AUTHENTICATION_KEY NASDA_AUTHENTICATION_KEY);

# Valid lengths for various things:
my %Max = (AUTHENTICATION_KEY     => 16,
           ECS_USERNAME           => 14,
           ECS_AUTHENTICATION_KEY => 20,
           HOME_DAAC => 20,
           OLD_ECS_AUTHENTICATION_KEY => 20,
           NASDA_AUTHENTICATION_KEY     => 16);

# Map simple vars to printed labels.
my %Label = (AUTHENTICATION_KEY     => 'ASF Data Access Key',
             ECS_USERNAME           => 'User Name',
             ECS_AUTHENTICATION_KEY => 'Password',
             HOME_DAAC              => 'Home Data Center',
             OLD_ECS_AUTHENTICATION_KEY => 'Old Password',
             NASDA_AUTHENTICATION_KEY => 'NASDA Data Access Key');

# Map key names to internal get/set methods.  Allows calling functions
# to ask for ('V0'), ('ECS'), ('NASDA'), ('V0', 'ECS') or 
# ('V0', 'ECS', 'NASDA').
my %Internal = ('V0'  => ['authentication_key'],
                'ECS' => ['ecs_username', 'ecs_authentication_key','home_daac',
                         'old_ecs_authentication_key'],
                'NASDA' => ['nasda_authentication_key']);

#---------------------------------------------------------------------
# new
#---------------------------------------------------------------------
# Constructor.  Creates a new, empty keys object.

sub new {
   my $class = shift;                   # Get class.
   my $self = {};                       # Start with empty hash.
   $self->{ERRORS} = new IU::Errors;    # Add empty errors object.
   bless $self, $class;                 # Create the object.
   return ($self);                      # Return object.
}

#---------------------------------------------------------------------
# _simple_ivar [$NEWVALUE]
#---------------------------------------------------------------------
sub _simple_ivar {
   my ($ivar, $self, $value) = @_;
   if (@_ > 2) {
      # Setting.
      if (defined ($value)) {
         # If value string too long, set an error string and return.
         if (length ($value) > $Max{$ivar}) {
            $self->{ERRORS}->set ("Error. $Label{$ivar}" .
               " ($value) is too long: " . length ($value) .
               ". Maximum is ". $Max{$ivar} . '.');
            return ($self->{ERRORS});
         }
      }
      # Set new value and return it.
      $self->{$ivar} = $value;
      return ($value);
   } else {
      # Getting.  Return value if defined or empty string.
      return (defined ($self->{$ivar}) ? $self->{$ivar} : '');
   }
}

#---------------------------------------------------------------------
# authentication_key 
#---------------------------------------------------------------------
# The V0 data access key.

sub authentication_key {
   _simple_ivar ('AUTHENTICATION_KEY', @_)
}

#---------------------------------------------------------------------
# nasda_authentication_key 
#---------------------------------------------------------------------
# The NASDA data access key.

sub nasda_authentication_key {
   _simple_ivar ('NASDA_AUTHENTICATION_KEY', @_)
}

#---------------------------------------------------------------------
# ecs_username 
#---------------------------------------------------------------------
# The ECS username.

sub ecs_username {
   my ($self, $value) = @_;

   # If setting and value has bad chars, set an error string.
   # Note that we allow blanking out a value if $value is ''.
   if ((@_ > 1) && ($value ne '') && ($value !~ /^[a-z0-9_]+$/i)) {
      $self->{ERRORS}->set
         ("Error. $Label{ECS_USERNAME} contains invalids characters.".
          " An ECS user name can only have these characters: " .
          "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ" .
          "0123456789 or underscores \"_\". No spaces or tabs.");
      return ($self->{ERRORS});
   }
   _simple_ivar ('ECS_USERNAME', @_);
}

#---------------------------------------------------------------------
# ecs_authentication_key
#---------------------------------------------------------------------
# The ECS data access key.

sub ecs_authentication_key {
   _simple_ivar ('ECS_AUTHENTICATION_KEY', @_)
}

#---------------------------------------------------------------------
# old_ecs_authentication_key
#---------------------------------------------------------------------
# The old ECS data access key.

sub old_ecs_authentication_key {
   _simple_ivar ('OLD_ECS_AUTHENTICATION_KEY', @_)
}

#---------------------------------------------------------------------
# home_daac
#---------------------------------------------------------------------
# The home DAAC where the ECS data access key is valid.

sub home_daac {
   _simple_ivar ('HOME_DAAC', @_)
}

#---------------------------------------------------------------------
# errors
#---------------------------------------------------------------------
# Returns Errors object stored in UserKeys.
# This could have been: sub errors { return ($_[0]->{ERRORS}); }
# EXCEPT that for backwards compatibility with persistent sessions,
# we need to see if the Errors object exists, and if not create one.

sub errors {
   my $self = shift;
   unless ($self->{ERRORS}) {
      $self->{ERRORS} = new IU::Errors;    # Add empty errors object.
   }
   return $self->{ERRORS};
}

#---------------------------------------------------------------------
# v0_ok_to_go
#---------------------------------------------------------------------
# Check if everything ready for the V0 authenticator.  1 yes, 0 no.

sub v0_ok_to_go {
   my $self = shift; 
   if ($self->authentication_key) { 
      return 1; # ready to go 
   } else {
      return 0;
   }
}

#---------------------------------------------------------------------
# nasda_ok_to_go
#---------------------------------------------------------------------
# Check if everything ready for the NASDA authenticator.  1 yes, 0 no.

sub nasda_ok_to_go {
   my $self = shift; 
   if ($self->nasda_authentication_key) { 
      return 1; # ready to go 
   } else {
      return 0;
   }
}

#---------------------------------------------------------------------
# ecs_ok_to_go
#---------------------------------------------------------------------
# Check if everything ready for the ECS authenticator.  1 yes, 0 no.

sub ecs_ok_to_go {
   my $self = shift; 
   if ($self->ecs_authentication_key && $self->ecs_username) { 
      return 1;    # ready to go
   } else {
      return 0;
   }
}

#---------------------------------------------------------------------
# from_ODL ODLOBJECT
#---------------------------------------------------------------------
# I<Class or instance method.>
# Initiallize the affiliation from a standard IMS ODL object.
# The instance variables have the same names as the associated
# parameters.

sub from_ODL {
   my $self = shift;
   my $ecs_password;
   my $old_user;
   if($SESSION && $SESSION->state->{User}) {
      $old_user = $SESSION->state->{User};
   }
   if(ref($old_user) && ($old_user->keys) && $ENV{REMOTE_USER} ne 'guest') {
      $ecs_password = $old_user->keys->ecs_authentication_key;
   }
   my $odl = shift || return (undef);

   # Be a class method, too!
   ref ($self) or $self = $self->new;

   # Read the ODL stream:
   $odl->rewind;
   my $clean;
   my ($param, $value);
   while (($param, $value) = $odl->next) {
      if($ENV{REMOTE_USER} ne 'guest' && $param eq 'ECS_USERNAME') {
         next;
      } elsif($SESSION && $SESSION->state->{Upload} && ref($old_user) && 
              ($old_user->keys) &&
              $ENV{REMOTE_USER} ne 'guest' && 
	      $param eq 'ECS_AUTHENTICATION_KEY') {
         next;
      } elsif (($value->type eq 'SCALAR') && $SimpleVar{$param}) {
         $clean = $value->text;
         $clean =~ s/\s+$//g;
         # Store the value using set method.  This can set Errors.
         _simple_ivar ($param, $self, $clean);
      } 
   }
   if($ENV{REMOTE_USER} ne 'guest') {
      $param = 'ECS_USERNAME';
      $clean = $ENV{REMOTE_USER};
      # Store the value using set method.  This can set Errors.
      _simple_ivar ($param, $self, $clean);
      if($SESSION && $SESSION->state->{Upload} && ref($old_user)) {
         $param = 'ECS_AUTHENTICATION_KEY';
         # Store the value using set method.  This can set Errors.
         _simple_ivar ($param, $self, $ecs_password);
      }
   }
   $SESSION->state->{Upload} = 0 if ($SESSION);
   # Return it:
   return ($self);
}

#---------------------------------------------------------------------
# to_ODL
#---------------------------------------------------------------------
# I<Instance method.>
# Convert to a standard IMS ODL object.

sub to_ODL {
   my $self = shift;
   my $odl = IU::ODL->new_group;
    
   # Print it even though it may be Blank.

   $odl->add_param ('AUTHENTICATION_KEY',
                    IU::ODL->quoted ($self->authentication_key));

   $odl->add_param ('ECS_AUTHENTICATION_KEY',
                    IU::ODL->quoted ($self->ecs_authentication_key));

   $odl->add_param ('ECS_USERNAME',
                    IU::ODL->quoted ($self->ecs_username));

   if($self->home_daac) {$odl->add_param ('HOME_DAAC',
                    IU::ODL->quoted ($self->home_daac));}

   if($self->nasda_authentication_key) {
      $odl->add_param ('NASDA_AUTHENTICATION_KEY', 
                       IU::ODL->quoted ($self->nasda_authentication_key));
   }

   $odl;
}

#---------------------------------------------------------------------
# from_CGI %KeysRequired
#---------------------------------------------------------------------
# This retrieves key values from a CGI form.  It will set values
# into the UserKeys object.  If no object is given, it will create
# a new one.  If there are any errors, such as strings being too long
# or required fields being missing, it will set strings in the Errors
# object in $self, but return as many of the keys as it was able to
# get.  from_CGI reads from the form generated by the print method.
# Input:  A hash of keys and whether they are required.  The hash
#         keys can be 'V0' and 'ECS', indicating they should be
#         gotten.  The values are 1 or 0.  If a key has the value 1,
#         that field is "required".  If a required field is missing
#         or blank, then Error strings are written to the ERRORS
#         object.
# Output: Filled in UserKeys object, with possible error strings
#         stored in its Errors object.

sub from_CGI {
   my $self = shift;
   my %keys_req = @_;

   # Make new, if invoked as a class method:
   ref ($self) or $self = $self->new;

   # If no hash, puke.
   %keys_req or
      puke 600, 'Coding error.  UserKeys::from_CGI must be called ',
                'with a hash of the keys to retrieve.';

   # Map keys to internal get/set methods and create hash "%fields"
   # using the internal names and the values (1's and 0's) from the
   # KeysRequired hash.  Note in the following:
   #    %keys_req is the hash of keys (like ('V0'=>1,'ECS'=>0,'NASDA'=>0))
   #    @{$Internal{$keyname}} is a list of internal get/set methods.
   #    Lower case of internal hash key is set function and tag.
   my ($keyname, $value, %fields);
   while (($keyname, $value) = each %keys_req) {
      map {$fields{$_} = $value} @{$Internal{$keyname}};
   }
   # Get each field in requested list.
   my ($tag);
   foreach $tag (keys (%fields)) {

      # Is tag defined?
      if ((defined ($value = $REQ->param ("_dak.$tag"))) || 
          (defined ($value = $REQ->param ("$tag")))) {

        # Trim the value.
         $value = trim ($value);

         # If value non-blank, set it and process next tag.  Note
         # that if string too long or bad format, Errors will be set.
         if (!blank ($value)) {
            my $cmd = '$self->' . $tag . '($value)';
            eval ($cmd);
            next;
         }
      }
      # Requested field was not filled in.  If required, add error.
      if ($fields{$tag}) {
         my $parm = uc ($tag);   # Parameter is uppercase of function.
         $self->{ERRORS}->set
             ("$Label{$parm} required but not entered.");
      } else {
         # Tag requested but not required and blank.  Set blank.
         my $cmd = '$self->' . $tag . "('')";
         eval ($cmd);
      }
   }
   return ($self);
}

#---------------------------------------------------------------------
# print %KeysRequired
#---------------------------------------------------------------------
# Input: (Required) A hash of keys to print.
#    The hash keys can be 'V0', 'ECS' and 'NASDA', indicating they should be
#    printed.  Values are 1 or 0.  If a key has the value 1, that
#    field is "required".  Required fields are shown in red.
# Output: Prints to HTML stream, a CGI form with fields for
#    inputing keys.

sub print {
   my ($self, %keys_req) = @_;

   # Extract KeysRequired passed to function.
   %keys_req or
      puke 600, 'Coding error.  UserKeys::print must be called ',
                'with a hash of the keys to print.';

   # Make new, if invoked as a class method:
   ref ($self) or $self = $self->new;

   $HTML -> TABLE (CELLPADDING=>1, CELLSPACING=>1) -> TR;
   # If hash requested 'V0' key, print it.
   if (exists ($keys_req{V0})) {
      $HTML -> TD (NOWRAP=>undef)
            -> field_cell (Title => "$Label{AUTHENTICATION_KEY}: ",
                           Subtitle=>'(Alaska SAR Facility customers only)',
                           Stack => 1, Req => $keys_req{V0})
            -> INPUT (TYPE      => 'password',
                      SIZE      => $Max{AUTHENTICATION_KEY},
                      NAME      => '_dak.authentication_key',
                      VALUE     => $self->authentication_key,
                      MAXLENGTH => $Max{AUTHENTICATION_KEY})
            -> _field_cell; 
      $HTML -> _TD;
   }

   # If hash requested 'NASDA' key, print it.
   if (exists ($keys_req{NASDA})) {
      $HTML -> TD (NOWRAP=>undef)
            -> field_cell (Title => "$Label{NASDA_AUTHENTICATION_KEY}: ",
                           Subtitle=>'(NASDA/HEOC/Japan customers only)',
                           Stack => 1, Req => $keys_req{NASDA})
            -> INPUT (TYPE      => 'password',
                      SIZE      => $Max{NASDA_AUTHENTICATION_KEY},
                      NAME      => '_dak.nasda_authentication_key',
                      VALUE     => $self->nasda_authentication_key,
                      MAXLENGTH => $Max{NASDA_AUTHENTICATION_KEY})
            -> _field_cell; 
      $HTML -> _TD;
   }
   $HTML -> _TR ->_TABLE;

   # If hash requested 'ECS' key, print ECS Name and Password fields.
   if (exists ($keys_req{ECS})) {
      if($ENV{REMOTE_USER} eq 'guest' || !$ENV{REMOTE_USER}) {
         $HTML -> TABLE (CELLPADDING=>1, CELLSPACING=>1) -> TR;
         $HTML -> TD (NOWRAP=>undef)
               -> field_cell (Title => "$Label{ECS_USERNAME}: ",
                              Small=>1, Stack => 1,
                              Req => $keys_req{ECS_USERNAME})
               -> INPUT (TYPE      => 'text',
                         SIZE      => $Max{ECS_USERNAME},
                         NAME      => '_dak.ecs_username',
                         VALUE     => $self->ecs_username,
                         MAXLENGTH => $Max{ECS_USERNAME})
               -> _field_cell;
         $HTML -> _TD;
         $HTML -> TD (NOWRAP=>undef)
               -> field_cell (Title =>"$Label{ECS_AUTHENTICATION_KEY}: ",
                              Small=>1, Stack => 1,
                              Req => $keys_req{ECS_AUTHENTICATION_KEY})
               -> INPUT (TYPE      => 'password',
                         SIZE      => $Max{ECS_AUTHENTICATION_KEY},
                         NAME      => '_dak.ecs_authentication_key',
                         VALUE     => $self->ecs_authentication_key,
                         MAXLENGTH => $Max{ECS_AUTHENTICATION_KEY})
               -> _field_cell; 
         $HTML -> _TD;
         $HTML -> _TR -> _TABLE;
      } else {
	 # This user is a persistent user
	 my $name = $self->ecs_username;
	 my $pass = $self->ecs_authentication_key;
         $HTML -> hidden('_dak.ecs_username' => "$name");
	 $HTML -> hidden('_dak.ecs_authentication_key' => "$pass");
      }
   }
   1;
}

#---------------------------------------------------------------------
# print_hidden %KeysRequested
#---------------------------------------------------------------------

sub print_hidden {

   my ($self, %keys_req) = @_;

   # Extract KeysRequested passed to function.
   %keys_req or
      puke 600, 'Coding error.  UserKeys::print must be called ',
                'with a hash of the keys to print.';

   # Make new, if invoked as a class method:
   ref ($self) or $self = $self->new;

   # If hash requested 'V0' key, print V0 Name and Password fields.
   if (exists ($keys_req{VO})) {
      $HTML -> hidden('_dak.authentication_key' => $self->authentication_key);
   }

   # If hash requested 'NASDA' key, print NASDA Name and Password fields.
   if (exists ($keys_req{NASDA})) {
      $HTML -> hidden('_dak.nasda_authentication_key' => 
                      $self->nasda_authentication_key);
   }


   # If hash requested 'ECS' key, print ECS Name and Password fields.
   if (exists ($keys_req{ECS})) {
      $HTML -> hidden('_dak.ecs_username' => $self->ecs_username);
      $HTML -> hidden('_dak.ecs_authentication_key' => 
                      $self->ecs_authentication_key);
   }
}
#---------------------------------------------------------------------
# print_login %KeysRequired
#---------------------------------------------------------------------
# Input: (Required) A hash of keys to print.
#    The hash keys can be 'V0' and 'ECS', indicating they should be
#    printed.  Values are 1 or 0.  If a key has the value 1, that
#    field is "required".  Required fields are shown in red.
# Output: Prints to HTML stream, a CGI form with fields for
#    inputing keys.

sub print_login {
   my ($self, %keys_req) = @_;

   # Extract KeysRequired passed to function.
   %keys_req or
      puke 600, 'Coding error.  UserKeys::print must be called ',
                'with a hash of the keys to print.';


   # Make new, if invoked as a class method:
   ref ($self) or $self = $self->new;

   # print ECS Name and Password fields.
   if (exists ($keys_req{ECS})) {
     $HTML -> TABLE (CELLPADDING=>1, CELLSPACING=>1) -> TR;
     $HTML -> TD (NOWRAP=>undef)
           -> field_cell (Title => "Account Username:",
                          Small=>0, Stack => 0,
                          Req => $keys_req{ECS_USERNAME})
           -> INPUT (TYPE      => 'text',
                     SIZE      => $Max{ECS_USERNAME},
                     NAME      => '_dak.ecs_username',
                     VALUE     => $self->ecs_username,
                     MAXLENGTH => $Max{ECS_USERNAME})
           -> _field_cell;
     $HTML -> _TD;
     $HTML -> _TR;
     $HTML -> TR;
     $HTML -> TD (NOWRAP=>undef)
           -> field_cell (Title =>"Password:",
                          Small=>0, Stack => 0,
                          Req => $keys_req{ECS_AUTHENTICATION_KEY})
           -> INPUT (TYPE      => 'password',
                     SIZE      => $Max{ECS_AUTHENTICATION_KEY},
                     NAME      => '_dak.ecs_authentication_key',
                     VALUE     => $self->ecs_authentication_key,
                     MAXLENGTH => $Max{ECS_AUTHENTICATION_KEY})
           -> _field_cell;
     $HTML -> _TD;
     $HTML -> _TR;
   }
   1;
}
#---------------------------------------------------------------------
1;
