package IW::Util;

=head1 NAME

IW::Util - utility functions for the IMS/WWW CGI scripts


=head1 NOTES

In many of the URL-generating routines, you will be asked to supply
the kind of entity you want guide/DIF/definition information on.  
To make this module easier to use, all these routines use the
following "standard entity names" from the data dictionary.
Here is how they map into the most-similar-in-spirit entity names 
in other systems:

    Data Dictionary    MD Supp.      Guide        Defn
                       Info Query    Search       File
    ----------------------------------------------------------------------
    CAMPAIGN           Campaign      Campaign     campaign
    DATASET_ID                       Dataset      dataset
    DATA_CENTER_ID     Data Center   Data_Center  data_center
    PARAMETER                                     parameter
    SENSOR_NAME        Sensor        Sensor       sensor
    SOURCE_NAME        Source        Source       source
    PROCESSING_LEVEL                              processing_level


=head1 AUTHOR

Eryq, Hughes STX Corporation

=head1 VERSION

$Id: Util.pm,v 1.54 2002/02/25 23:30:12 seufert Exp $

=cut

use URI::Escape;
use IU; 
use IU::Log;
use IU::URL;
use LWP::Simple;

# Need to do it this way, since two modules can't just "use" each other:
require IW::App;

use Exporter ();
@ISA = qw(Exporter);
%EXPORT_TAGS = ( 
        'output' => [qw(interpolate 
                        nobreak
			trim)],
	'taint'  => ['untaint'],
	'string' => [qw(blank
                        nobreak
			trim)],
	'url'    => [qw(defn_url 
			dif_url  
			doc_url 
			guide_url 
                        gcmd_url
			gloss_url 
			help_url
			logged_url
			query_url 
                        secured_url
			script_url 
			script_action
			supp_url)],
	);
Exporter::export_ok_tags qw(output taint url string);

use strict;


#------------------------------------------------------------
#
# GLOBALS
#
#------------------------------------------------------------

# Mapping of entity names...
my $SuppMap = undef;
my $GuideMap = undef;
my $DefnMap = undef;
my $ShortName = undef;
# Fix for broken URI::Escape 3.6 (grrrrr):
my $UESC = '^A-Za-z0-9';

# Current uniquifier:
my $Uniquifier;


#------------------------------------------------------------
#
# Private utilities...
#
#------------------------------------------------------------

#------------------------------------------------------------
# uniq
#------------------------------------------------------------
# Return the current uniqifier.
# All hyperlinks generated on this page will have this embedded
# as a u= tag.  So in a way, it identifies the page you came FROM.

sub uniq {
    defined($Uniquifier) or $Uniquifier = (int(time) % 1000000);
    $Uniquifier;
}



#------------------------------------------------------------
#
# The string/output utilities..
#
#------------------------------------------------------------

#------------------------------------------------------------
# interpolate OPENEXPR [,OUTFH]
#------------------------------------------------------------
# Open OPENEXPR for reading, and write contents line-by-line to OUTFH.
# Returns true on success, false on error.

sub interpolate {
    my ($openexpr, $outfh) = @_;
    no strict 'refs';
    $outfh or $outfh = select;
    open IN, $openexpr or return error "open '$openexpr': $!";
    while (<IN>) {
	print $outfh $_;
    }
    close IN;
    1;
}

#------------------------------------------------------------
# blank STRING
#------------------------------------------------------------
# Is this string blank?  It is if it's either undefined, or
# empty, or all-whitespace:

sub blank {
    my $str = shift;
    !defined($str) or ($str eq '') or ($str !~ /\S/);
}

#------------------------------------------------------------
# trim STRING
#------------------------------------------------------------
# Remove leading/trailing whitespace from STRING.
# Since this is largely used for cleaning up CGI input or input from a
# foreign source, undefined values are silently coerced to the empty string.

sub trim {
    my $str = shift;
    defined($str) or return '';
    $str =~ s/^\s+|\s+$//g;
    $str;
}


#------------------------------------------------------------
# nobreak STRING
#------------------------------------------------------------
# Change all spaces in the string to nonbreaking ones.

sub nobreak {
   my $str = shift;
   defined $str or return "";
   $str =~ s/ /\xA0/g;
   $str;
}
#------------------------------------------------------------
#
# The "url" utilities..
#
#------------------------------------------------------------

#------------------------------------------------------------
# _build_query SCRIPT, PARAMS
#------------------------------------------------------------
# PRIVATE subroutine
#
# The SCRIPT is the URL of the script to query (everything
# before the '?').   The PARAMLIST is of the form:
#
#      param=>value, ..., param=>value
#
sub _build_query {
    my $href = shift;
    $href .= '?';   
    while (@_) {                   # add params...
	$href .= uri_escape(shift @_, $UESC) . '=' 
               . uri_escape(shift @_, $UESC) . '&';
    }
    chop $href;                    # remove '?' or '&' at end

    $href;
}

#------------------------------------------------------------
# defn_url TYPE,ITEM
#------------------------------------------------------------
# Return the URL for a definition document.
# The TYPE is the type of definition being sought; one of the standard set:
#
#        * Campaign          * Parameter   
#        * Data_Center	     * Sensor      
#        * Dataset	     * Source      
#
# The ITEM is the item whose definition you want.

sub defn_url {
  #  my ($type, $item) = @_;
  my ($type, $item, $page_layout) = @_;
  defined($type) or return undef;
    # Ensure that mapping exists:
    unless ($DefnMap) {
	$DefnMap = {CAMPAIGN       => 'campaign',
		    DATASET_ID     => 'dataset',
		    DATA_CENTER_ID => 'data_center',
		    PARAMETER      => 'parameter',
		    SENSOR_NAME    => 'sensor',
		    SOURCE_NAME    => 'source',
		    PROCESSING_LEVEL => 'processing_level'};
    }

    # Rewritten URLs are required, since we are pointing back
    # to the EDG,

    #  if true, Navigator turned off
    if ((defined($page_layout)) && ($page_layout eq "OUT_OF_LINE")) {
	debug "OUT_OF_LINE defn";
       script_url('mode' => 'DEFN',
		'submode' => 'OUT_OF_LINE',
               'type' => $DefnMap->{$type},
               'item' => $item);
    } else {
       # Return the URL:
#	debug "Same page defn";
      
       script_url('mode' => 'DEFN',  
	       'type' => $DefnMap->{$type},
	       'item' => $item);
     }
  
    

}

#------------------------------------------------------------
# dif_url MDENTRY
#------------------------------------------------------------
# Given a MD entry id, compose the URL for the DIF.
# Returns undef if the argument is undefined.
# We DO NOT rewrite the URL, because the DIF server typically
# is running under a different web server.

sub dif_url {
    my ($mdentry) = @_;
    defined($mdentry) or return undef;
    _build_query("$IU::GcmdGetdifUrl?$mdentry");
}

#------------------------------------------------------------
# doc_url RELATIVE
#------------------------------------------------------------
# Return the URL for the given document, specified by the path
# RELATIVE to the docs directory.

sub doc_url {
    "${IU::DocsUrl}$_[0]";
}

#------------------------------------------------------------
# gloss_url TERM
#------------------------------------------------------------
# Given a term, compose the URL for the glossary entry.
# Returns undef if the argument is undefined or empty.
#
# NOTE: DO NOT URI-ESCAPE THE TERM!!!! 
# Netscape is idiotic, and won't recognize "#DATA+CENTER" 
# or "#DATA%20CENTER"... it will only take "#DATA CENTER".
# This produces questionable URLs, since they contain whitespace.

sub gloss_url {
    my $term = (shift || return undef);    
    "${IU::ImsGlossUrl}#$term";
}

#------------------------------------------------------------
# guide_url TYPE,ITEM
#------------------------------------------------------------
# Return the URL for this guide document, or undef if none.
# TYPE is one of the standard set (described at top).

sub guide_url {
    my ($type, $item,$page_layout) = @_;
    $item = trim($item);        # remove leading/trailing whitespace

    # Make sure map is available:
    # LEAVE Dataset ONE WORD!!!!
    unless ($GuideMap) {
	$GuideMap = {  CAMPAIGN       => 'Campaign',
                       PARAMETER      => 'Parameter',
		       DATASET_ID     => 'Dataset',
		       DATA_CENTER_ID => 'Data_Center',
		       SENSOR_NAME    => 'Sensor',
		       SOURCE_NAME    => 'Source'  }
    }
    $GuideMap->{$type} or return undef;
# since there are no guide documents for parameter, return undef if type
# equal parameter
    if($type ne 'PARAMETER'){
       #  if true, Navigator turned off
       if ((defined($page_layout)) && ($page_layout eq "OUT_OF_LINE")) {
            script_url('mode'    => 'GUIDE', 
		    'submode'    => 'OUT_OF_LINE', 
                   'doctype' => $GuideMap->{$type},
                   'item'    => $item);
        }
        else {
            # Return the URL:
            script_url('mode'    => 'GUIDE',  
	           'doctype' => $GuideMap->{$type},
	           'item'    => $item);
        }
    }
    else {
       return undef;
    }
}
#------------------------------------------------------------
# help_url CONTEXT
#------------------------------------------------------------
# Return the "help" URL for this topic, given by its CONTEXT.
# Return undef if no context.

sub help_url {
    my $context = shift || (return undef);

    script_url('mode'=>'HELP', 'page'=>"imsuser_$context");
}
#------------------------------------------------------------
# logged_url URL, PARAMHASH
#------------------------------------------------------------
# We have a URL which leads outside the system.  Construct a URL which
# goes back to a mode which logs the URL and then redirects the browse.
# All subsequent args are turned directly into CGI params. 

sub logged_url {
    my $url = shift; 

    script_url('mode'=>'ENERGIZE', 'url'=>$url, @_);
}

#------------------------------------------------------------
# query_url SCRIPT,PARAMLIST
#------------------------------------------------------------
# Rewrite the URL so that it points into the EDG and build
# the necessary query string.

sub query_url {
    # Get the in/complete url.
    my $original_url = shift @_ ;

    # Break the url into components.
    my ($protocol, $server, $port, $server_stripped) =
      IU::URL::decompose_url($original_url);

    # Is it a in/complete URL? Incomplete, build into a full Url.
    my $href = undef;
    if ( $original_url =~ /^http:/i   ) {
       $href = "$protocol://$server:$port/$server_stripped";
    }
    else {
       $href = "${IU::CgiUrl}$server_stripped";
    }

    # Build up the query.
    _build_query($href, @_);
}

#------------------------------------------------------------
# script_action [JUMPTO]
#------------------------------------------------------------
# Action suitable for the ACTION tag of a form submitted to this script. 
# If JUMPTO is given, an attempt is made to jump to an internal tag.
# 
# Uniquifier is embedded in the script name to force uniqueness.

sub script_action {
    my $jumpto = shift;
    defined($jumpto) ? ($jumpto = "#$jumpto") : ($jumpto = '');
   my ($protocol, $server, $port, $server_stripped) = 
      IU::URL::decompose_url($ENV{SCRIPT_NAME});
    "$protocol://$server:$port$server_stripped/u" . uniq() . $jumpto;
}

#------------------------------------------------------------
# script_url PARAMLIST
#------------------------------------------------------------
# Return a URL back to this IMS/WWW script.  THIS IS NOT FOR CONSTRUCTING
# GENERALIZED HYPERLINKS!  TAGS LIKE THE sid AND u ARE AUTO-EMBEDDED
# IN THE LINK!!!!  USE query_url() INSTEAD!
#
# The PARAMLIST is of the form
#
#      param=>value, ..., param=>value
#
# and is used create the query portion of the script, except for several
# special params:
#
#       SCRIPT  => supply an alternative script URL
#       JUMPTO  => if present, causes jump to internal tag (if supported)
#
# If a session is active, then an "sid=(sid)" parameter automatically
# becomes part of the URL.  Otherwise, it's just left off.
# 
# Specifying 'UNIQUE' is no longer necessary... it is the default
# case.  This is because an accidental jump back to a CGI document in cache
# means that all the documents off of it are likewise possibly in cache.
# Better to just sidestep the whole issue (note that we can't use 
# a "no-cache" pragma due to the way it screws up the "back" button!).

sub script_url {
    my @params = @_;

    # Start with default script URL:
    my $script = $ENV{SCRIPT_NAME};
    my $jumpto = '';

    # Add sid info, if any:
    unshift(@params, 'sid', $IW::App::SESSION->id)  if $IW::App::SESSION;

    # Add uniquifier:
    unshift(@params, 'u', uniq());

    # Add endform :
    unshift(@params, 'endform', '1');

    my $secured_dir = $IU::SecuredDir;
#     debug "$secured_dir";

    # Add other params:
    my $href = '';
    my $secure = '';
    while (@params) {
	my $k = shift @params ;
	my $v = shift @params || ''  ;
	if    ($k eq 'UNIQUE') { 1 }
	elsif ($k eq 'SCRIPT') { $script = $v }
	elsif ($k eq 'SECURE') { $secure = $secured_dir }
	elsif ($k eq 'JUMPTO') { $jumpto = "#$v" }
	else { $href .= uri_escape($k, $UESC) . '=' 
		      . uri_escape($v, $UESC) . '&' }
    }      

    # Use FULLY-QUALIFIED script name, in case our document has a BASE tag!
    my $s = ($IU::SecureServer ? "s" : "");
    my $port = ($IU::SecureServer ? ($IU::SecurePort) : $ENV{SERVER_PORT});
    ($script =~ /^http$s:/i) or
	$script = "http$s://$ENV{SERVER_NAME}:$port$script";

    # Secure it, if necessary ...
    $script =~ s{(.*)/(.*)$}{$1/$secure/$2} if ($secure && 
                (!($script =~ /$secure/)));

    # Complete it...
    $href = "$script$jumpto" . '?' . $href;    # prefix with script;
    chop $href;                                # remove '?' or '&' at end
    $href;
}

#------------------------------------------------------------
# supp_url TYPE,SHORTNAME, MDENTRY, FLAG
#------------------------------------------------------------
# Return the URL for supplementary DIF info, or undef if none.
#
# The portion after the "?" takes the following (unencoded) forms,
# based on what is being sought:
#
#     suppmenu=TYPE:SHORTNAME

sub supp_url {
    my ($type, $shortname, $entry, $flag) = @_;

    # Make sure mapping is available:
    unless ($SuppMap) {

	$SuppMap = {CAMPAIGN       => 'campaign',
		    DATA_CENTER_ID => 'center',
		    SENSOR_NAME    => 'sensor',
		    SOURCE_NAME    => 'source'};
    }
    $SuppMap->{$type} or return undef;
    my $shtname;

    # if type is data center id we want to get the gcmd name from
    # wwwvalids.conf
    if ($type eq "DATA_CENTER_ID") {
        # Open static valids database:
        my %WWWVALIDS;
#
        if (-e "${IU::LibDir}/conf/wwwvalids.gdb") {
            eval {tie %WWWVALIDS, 'IU::Store',
                  "${IU::LibDir}/conf/wwwvalids.gdb", 'READER'};
            $@ and return error "tie failed: $@";
        } 

        # Get list of daacs to check.
        my @daacs;
        @daacs = map {substr ($_, 5)}
             grep { $_ =~ /^daac:/} keys (%WWWVALIDS);

       # For each data center we're considering...
       my $archid;       # Each daac.

       foreach $archid (@daacs) {
          # Does this daac match the shortname. If so, is there an entry
          # for it in the static valids?
          my $valids;
          if ($archid eq $shortname){
              if ($valids = $WWWVALIDS{"daac:$archid"}) {

                  $shtname = $valids->{GCMDName};
 
#debug "archid = $archid";
#debug "valids gcmd name = $valids->{GCMDName}";

	 
              }
           }
       }                                    # end: foreach daac 

       # Close static valids database:
       untie %WWWVALIDS; 

    }                                       # end: if type eq DATA_CENTER_ID
    else {
       $shtname = $shortname;
    }
     
 

    # Return the URL:
    my $url = undef;
    my $len = length($shortname);
    if (($len != 0) && (defined($shtname)) ) {
       if ($flag == 0) {

      $url =    _build_query($IU::GcmdGetsuppUrl,
 	             'type' => ($SuppMap->{$type}),
                     'interface' => 'imswww',
                     'short_names' =>$shtname );
        }
        else {
            if ((defined $entry) && ($entry ne "Dummy")) {
                $url =    _build_query($IU::GcmdGetdifUrl,
 	                 'format' => 'Display DIF',
                         'entry' => $entry . '#SUPPLEMENTARY',
	                 'form'   => 'imswww');
            }else {
                  $url = undef;
            }
             
        }

    }     

   $url;
}
#------------------------------------------------------------
# gcmd_url  TYPE, SHORTNAME, MDENTRY
#------------------------------------------------------------
# Return the  URL for this shortname.  If shortname is not a 
# valid keyword for GCMD, return url of supplementary information
# in DIF.

sub gcmd_url{
   my ($type, $shortname, $entry) = @_;
     
     my $url; 
    # Use the shortname to construct the url for the gcmd
    $url = supp_url($type, $shortname, $entry,0);

    # Try and get the page.  If can't get page set url to undef so button
    # will be greyed out.
    my $page;
    if (defined $url) {
        unless (defined ($page = LWP::Simple::get($url))) {
           $url = undef;
        }
        if (defined $page) {
          my @page = split(/\n/,$page);
          my @list;
          my $errmsg = 0;
          # go through each line in the page looking for the error message
          foreach my $line (@page){

              if ($line =~ m(<HTML><HEAD><TITLE>Supplemental Query Error</TITLE>)) {
                  $errmsg = 1;

              }
          }
          if ($errmsg == 1) {
                 $url = supp_url($type, $shortname, $entry, 1);
          
          }
        }
    }


     $url;
}

#------------------------------------------------------------
#
# The "taint" utilities..
#
#------------------------------------------------------------

#------------------------------------------------------------
# untaint SCALAR
#------------------------------------------------------------
# Return an untainted equivalent to the given scalar.
#
# B<DO NOT USE THIS UNLESS YOU ABSOLUTELY POSITIVELY UNDERSTAND
# WHAT YOU ARE DOING!! I MEAN IT!!!> You can compromise the security
# of both the system I<and the host itself> if you blindly use this
# "to get rid of an error message".
#
# Next to I<every> call of this, you I<must> sign your name and
# swear that you know that what you're doing is safe.  Like this:
#
#      my $sid = untaint($sid);    # I KNOW what I'm doing is safe - Joe.
#
# That way, if it looks like you DON'T know what you're doing, we can
# find you and you can explain.  You might consider a comment that 
# explains why you are being safe:
#
#      # The sid is scanned by IU::Session for illegal characters,
#      # an a fatal exception is raised if there are any.

sub untaint {
    my $tainted = shift;
    $tainted =~ /^(.*)$/;
    return $1;
}


#------------------------------------------------------------
#
# The "nav" utilities...
#
#------------------------------------------------------------

# URLs for the various navigator buttons.
# These all take into account whether or not a session is needed.

# nav_welcome_url
sub nav_welcome_url {
    $IU::WelcomeUrl;
}

# nav_startmenu_url
sub nav_startmenu_url {
    $IW::App::SESSION or return undef;
    script_url('mode'=>'WELCOME');
}

# nav_prefs_url
sub nav_prefs_url {
    $IW::App::SESSION or return undef;
    script_url('mode'=>'PREFS');
}

# nav_search_url
sub nav_search_url {
    $IW::App::SESSION or return undef;
    script_url('mode'=>'MAINSRCH');
}

# nav_datasets_url
sub nav_datasets_url {
    ($IW::App::SESSION && (-e $IW::App::SESSION->path_to("results.gdb")))
	or return undef;
    script_url('mode'=>'DSETLIST');
}

# nav_bench_url
sub nav_bench_url {
    ($IW::App::SESSION && (-e $IW::App::SESSION->path_to("workbench.gdb")))
        or return undef;
    script_url('mode'=>'WORKBENCH','submode'=>'LIST','select'=>'NONE');
}


# nav_cart_url
sub nav_cart_url {
    ($IW::App::SESSION && (-e $IW::App::SESSION->path_to("cart.gdb")))
	or return undef;
    script_url('mode'=>'CART');
}

# nav_comments_url
sub nav_comments_url {
    $IW::App::SESSION or return undef;
    script_url('mode'=>'CMNTFORM');
}

# nav_help_url CONTEXT
sub nav_help_url {
    my $context = shift || '';
    return undef if ($context eq 'helphelp');
    help_url($context);
}

# nav_admin_url
sub nav_admin_url {
    script_url('mode'=>'ADMIN');
}

# logout
sub logout {
    # Check for Session State
    $IW::App::SESSION or return undef;
    script_url('mode'=>'LOGIN','submode'=>'LOGOUT');
}


#------------------------------------------------------------
#
# Misc utilities (not exported)
#
#------------------------------------------------------------

#------------------------------------------------------------
# msubmit_tag QUERY
#------------------------------------------------------------
# I<Class/instance method>.  Return the name for a multisubmit
# parameter that encodes the given QUERY, which should be an
# arrayref of the form C<[param=>value, ..., param=>value]>.
#
# The name will be C<msubmit?param=value&...&param=value&>.
# THAT'S RIGHT: there will be a trailing '&'!!!

sub msubmit_tag {
    my $query = shift || [];

    my $name = 'msubmit?';
    my @query = map { URI::Escape::uri_escape($_, $UESC) } @$query;
    while (@query) {
	$name .= shift(@query) . '=' . shift(@query) . '&';
    }
    $name;
}

#------------------------------------------------------------
1;
