package IW::Widget::CombinedShowGeo;

=head1 NAME

IW::Widget::CombinedShowGeo - widget to show a geoselection category

=head1 DESCRIPTION

An instance of this class displays the single query category for
a geographic region.

=head1 NOTES

Should attempt to take more advantage of its superclass' methods.

=cut

use IU;
use IU::Query;
use IU::Category;
use IU::Log;

use IW::App;
use IW::Util qw(:url);

use IW::GeoSelect::NewBase;
use IW::GeoSelect::Mapfile;
use IW::GranMap;
use HTML::Stream qw(:funcs);


use IW::Widget::CombinedWidget;


use strict;
# Descriptions of geographic selection:

my @geo_types = (
                  "Orthographic Map (Java)",
                  "Stereographic S-pole Map",
                  "Type in Lat/Lon Range",
                  "Equatorial Map",
                  "Stereographic N-pole Map",
                  "Type in Path/Row Range",
                  "Global Search",
                  "Global granules only",
                  "Type in Lat/Lon Point");

use vars qw(@ISA);

@ISA = qw(IW::Widget::CombinedWidget);

#------------------------------------------------------------
# new
#------------------------------------------------------------
# I<Constructor.>

sub new {
    my ($class, %params) = @_;
    $params{Category} ||= 'GEOGRAPHIC_REGION';
    $params{Mode}     ||= 'NOMODE';
    $params{BGCOLOR}  ||= '#D0D0D0';
    $params{OptionType}||= "Type in Lat/Lon Range";

    bless \%params, $class;
}
#------------------------------------------------------------
# map_src
#------------------------------------------------------------
# Return the src of the summary map.
# Returns undef if map couldn't be obtained, and 'NONE' if none
# should be displayed.

sub map_src {
    my $self = shift;

    # Get criterion, and underlying GEO object:
    my $crit = $self->{Crit} = ($QUERY->criteria('GEOGRAPHIC_REGION'))[0];
    my $geo  = $crit->geo;
    my $mapfile;
    my $display_geo = $REQ->param('p.mapname-next'); 

    if ($geo->ODL_type eq 'RANGE_LOC' &&
        $display_geo eq 'Type in Lat/Lon Range') {

        my ($w,$n,$e,$s) = ($geo->west, $geo->north, $geo->east, $geo->south);
        $mapfile = lookup IW::GeoSelect::Mapfile 'Equatorial Map (small)';
        return query_url($mapfile->script,
                         'map'   => $mapfile->filename,
                         'grid'  => 'full',
                         'srect' => "$w,$n,$e,$s");
    }         
    elsif ($geo->ODL_type eq 'POINT_LOC' &&
        $display_geo eq 'Type in Lat/Lon Point') {

        my ($lat,$lon) = ($geo->lat, $geo->lon);
        $mapfile = lookup IW::GeoSelect::Mapfile 'Equatorial Map (small)';
        return query_url($mapfile->script,
                         'map'   => $mapfile->filename,
                         'grid'  => 'full',
                         'gmark' => "$lon,$lat");
    }
    elsif ($geo->ODL_type eq 'PATH_ROW_LOC' &&
           $display_geo eq 'Type in Path/Row Range') {
 
        # Get bounding rectangle:
        my $br = $geo->bound_rect;
        $self->{ApproxGeo} = $br;
 
        # Get source:
        my ($w,$n,$e,$s) = ($br->west, $br->north, $br->east, $br->south);
        $mapfile = lookup IW::GeoSelect::Mapfile 'Equatorial Map (small)';
        return query_url($mapfile->script,
                         'map'   => $mapfile->filename,
                         'grid'  => 'full',
                         'srect' => "$w,$n,$e,$s");
    } 
    elsif ($geo->ODL_type eq 'POLYGON_LOC' && 
           $geo->projection eq 'ORTHOGRAPHIC') {
 
       my $orthogif = undef;
       my $zoom = $geo->zoom || 1;

       # hack to handle zooms greater than 16x
       $zoom = 16 if ($zoom > 16);
                
       my $centerLat = $geo->tangent_lat || 0;
       my $centerLon = $geo->tangent_lon || 0;

       my @dummy = ();
       my $some_granules = \@dummy;
       my $newstamps = 0;
       my $xsize = 200;
       my $ysize = 150;

       # Generate the GIF:
       my @filenames = ();
       IW::GranMap::MakeMapGifs(Granules  => $some_granules,
                                Zoom      => $zoom,
                                CenterLat => $centerLat,
                                CenterLon => $centerLon,
                                NewStamps => $newstamps,
                                XSize     => $xsize,
                                YSize     => $ysize,
                                Gifnames  => \@filenames) || return undef;

       # Rename the main output file:
       my $path = $filenames[0];
       my $session_dir = $SESSION->dir;
       rename($path, "$session_dir/pub/geo_query.gif")
          or die "couldn't rename $path: $!";

       my $url = $SESSION->url("pub/geo_query.gif?u=".scalar(time));
       return ($url);
   } 
   undef;
}

#------------------------------------------------------------
# map_alt
#------------------------------------------------------------
# Return the ALT text of the summary map.
# Returns "unknown" if map couldn't be obtained, and 'no map displayed' if none
# should be displayed.

sub map_alt {
    my $self = shift;

    # Get criterion, and underlying GEO object:
    my $crit = $self->{Crit} = ($QUERY->criteria('GEOGRAPHIC_REGION'))[0];
    my $geo  = $crit->geo;
    my $mapfile;
    my $display_geo = $REQ->param('p.mapname-next'); 

    if ($geo->ODL_type eq 'RANGE_LOC' &&
        $display_geo eq 'Type in Lat/Lon Range') {
       return "small equatorial map, showing lat/lon range (if any)";
    }         
    elsif ($geo->ODL_type eq 'POINT_LOC' &&
        $display_geo eq 'Type in Lat/Lon Point') {
       return "small equatorial map, showing point (if any)";
    }
    elsif ($geo->ODL_type eq 'PATH_ROW_LOC' &&
           $display_geo eq 'Type in Path/Row Range') {
       return "small equatorial map, showing path/row range (if any)";
    } 
    elsif ($geo->ODL_type eq 'POLYGON_LOC' && 
           $geo->projection eq 'ORTHOGRAPHIC') {
       return "orthographic map, showing current selection (if any)";
   } 
   return "No map to be shown";
}

#------------------------------------------------------------
# print
#------------------------------------------------------------
sub print {

    my $self = shift;

    # Listing:
    $HTML -> TABLE(WIDTH=>'100%', BORDER=>0, CELLPADDING=>4, CELLSPACING=>1,
		   BGCOLOR=>$self->{BGCOLOR});

    # Get the current choices:
    my $cat = 'GEOGRAPHIC_REGION';
    my $crit = $self->{Crit} = ($QUERY->criteria('GEOGRAPHIC_REGION'))[0];
    my $geo = $QUERY->geo_region;   
    ###
    ### ROW 1
    ###
    $HTML -> TR ->TD(WIDTH=>'100%', COLSPAN =>3);

    $self->print_section_title("Choose Search Area");  

    if ($IU::GazetteerUrl) {
       $HTML->small;
       my $gazetteerlink = script_url('mode'=>'GAZETTEER');
   
       $HTML->hypertext('Find location using Gazetteer',
                     $gazetteerlink, TARGET=>'gazetteer');
       $HTML->_small -> t(" (opens a new window)");
    }
    $HTML ->_TD->_TR;
    ###
    ### ROW 2, COLUMN 1...
    ###
    $HTML -> TR(VALIGN=>'TOP');
    $HTML -> TD(WIDTH=>'5%');
 
     if (!$self->req_opttype("p.mapname-next")) {       # new page

	$self->print_form_init;

    }
  


    ###
    ### ROW 2-3, COLUMN 2...
    ###
    # Map:


    $self->print_form_for_widget;
   
 

    # Arguments to print_toggle_buttons are:
    # outer increment - the increment of the outerloop that prints the 
    #    toggle buttons in rows. Should increment by number of rows
    #    needed.
    # inner limit - if more than one row is needed for all of the 
    #    options, inner limit is number of items in the row.
    # time types - array of tags for the buttons
    # parameter name - the current selected option
    # mode - return mode to set the url for the toggle buttons
    # submode - return submode to set the url for the toggle buttons
    # nextsubmode - next submode to set the url for the toggle buttons
   $self->print_toggle_buttons(3,3,\@geo_types,'p.mapname-next',
            'SSRCHFORM', 'READINPUT', 'SWITCHMAP');
    $HTML -> _TABLE;    
}
#------------------------------------------------------------
# print_form_init
#------------------------------------------------------------
# I<Instance method, override.>
# 
#
sub print_form_init {
    my $self = shift;

    # Get map to use, and set as default:

    # "Java enabled Map" changed to "Orthographic Map (Java)", 
    # fix up LastMapname from old session states, so we
    # are backwards compatible.


    $SESSION->state->{LastMapname} = 'Orthographic Map (Java)' 
       if ($SESSION->state->{LastMapname} eq 'Java enabled Map');

    # "I want GLOBAL GRANULES ONLY" changed to "Global Granules Only", 
    # "I want to do a GLOBAL SEARCH" changed to "Global Search",
    # "Type in Path/Row" changed to "Type in Path/Row Range",
    # fix up LastMapname from old session states, so we
    # are backwards compatible.

   $SESSION->state->{LastMapname} = 'Global granules only' 
       if ($SESSION->state->{LastMapname} eq 'I want GLOBAL GRANULES ONLY');

   $SESSION->state->{LastMapname} = 'Global Search' 
       if ($SESSION->state->{LastMapname} eq 'I want to do a GLOBAL SEARCH');

   $SESSION->state->{LastMapname} = 'Type in Path/Row Range' 
       if ($SESSION->state->{LastMapname} eq 'Type in Path/Row');

    my $draw_mapname;

    $draw_mapname = $self->req_opttype('p.mapname') || # explicit selection
 	            $SESSION->state->{LastMapname}  || # last map used 
                    "Type in Lat/Lon Range";         # fallback
    
    $REQ->param('p.mapname-next'=>$draw_mapname);

}
#------------------------------------------------------------
# print_form_for_widget
#------------------------------------------------------------
# I<Instance method, override.>
# Output the main body of the page.
#
sub print_form_for_widget {
   my $self = shift;

   my $draw_mapname;
   my $javascriptOK = $SESSION->javascript_capable();
   #--------------------
   # PROCESS INPUT FROM LAST MAP....
   #--------------------

   # Get the "last" map (the one the user would have clicked on).
   # We'll need this to decode their map click:
   my $mapname = $self->req_opttype('p.mapname') ||   # explicit selection
                 $SESSION->state->{LastMapname}  ||   # last map used
                 "Type in Lat/Lon Range";           # 1st time: default map 

   my $crit = $self->{Crit} = ($QUERY->criteria('GEOGRAPHIC_REGION'))[0];

   #--------------------
   # OUTPUT USING NEXT MAP....
   #--------------------
   my $q_uploaded = $SESSION->state->{QueryUpload};
    
   if ($q_uploaded eq 'yes') {
      # figure out the mapname
      $draw_mapname = $self->upload_mapname($crit);
      $mapname = $draw_mapname;
      # reset session state
      $SESSION->state->{QueryUpload} = 'no';
   } else {
      # Get the "next" map (the one the user wants to see next, from pulldown).
      # Usually, this will be identical to the "last" map.
      # If none selected, default to current map.
      $draw_mapname = $self->req_opttype('p.mapname-next') || $mapname;
   }
   # Name:
   $HTML -> B -> t($draw_mapname) -> _B -> BR;

   $HTML-> _TD;

   my $geoslct = lookup IW::GeoSelect::NewBase $mapname;
   $geoslct->state($REQ->param('p.state'));          # set state

   # Get input from user's last action (tweaking x,y clicks, etc.):
   $geoslct->get_input($q_uploaded);

   my $draw_mapslct = $geoslct->switch_to($draw_mapname);

   $REQ->param('p.mapname-next' => $draw_mapname);      # init widget
   $self->{OptionType} = $draw_mapname;
   # Remember this:
   $SESSION->state->{LastMapname} = $draw_mapname;

   # The map/form:

   $HTML -> hidden('p.mapname' => $draw_mapname);
   $HTML -> hidden( 'p.browse_location'  => $REQ->param('p.browse_location'))
     if $REQ->param('p.browse_location');
   $crit = $self->{Crit} = ($QUERY->criteria('GEOGRAPHIC_REGION'))[0];

   if (($draw_mapname eq "Global granules only") || 
       ($draw_mapname eq "No geographic region") ) {
      
      $HTML -> TD(WIDTH=>'95%',ROWSPAN=>1, COLSPAN=>2, 
                  ALIGN=>'LEFT',VALIGN=>'MIDDLE', BGCOLOR=>'#FFFFFF');
      $HTML -> t("MAP NOT AVAILABLE");
  
   } elsif ($draw_mapname eq 'Type in Lat/Lon Range' or
            $draw_mapname eq 'Type in Lat/Lon Point') {

      $HTML -> TD(WIDTH=>'95%',ROWSPAN=>2, COLSPAN=>2,
                  ALIGN=>'LEFT',VALIGN=>'MIDDLE', BGCOLOR=>'#FFFFFF');
      my $browse_location = $REQ->param('p.browse_location');
      $draw_mapslct->draw(Upload=>$q_uploaded);
      $HTML -> _TD->_TR;
      $HTML -> TR(VALIGN=>'TOP') ;
      $HTML -> TD;
      $HTML -> small;
      # Only show summary text and map if geo type stored in query matches 
      # current geo widget.
      my $map;
      if ( ($crit->geo->ODL_type eq 'RANGE_LOC' and 
            $draw_mapname eq 'Type in Lat/Lon Range')
          or
           ($crit->geo->ODL_type eq 'POINT_LOC' and 
            $draw_mapname eq 'Type in Lat/Lon Point')
         ) {  
         print join("<BR>",
            split /\n/, html_escape($crit->as_string(Verbose=>2)))
               unless ($javascriptOK); 
         $map = $self->map_src();
      } else {
         print 'No selection to show on map:<BR>'
            unless ($javascriptOK);
         my $mapfile = lookup IW::GeoSelect::Mapfile 'Equatorial Map (small)';
         $map = query_url($mapfile->script,
                         'map'   => $mapfile->filename,
                         'grid'  => 'full'); 
      }
      print qq(<IMG NAME=map SRC=$map ALT="),$self->map_alt,qq(">);
   } elsif ($draw_mapname eq 'Type in Path/Row Range') {

      $HTML -> TD(WIDTH=>'95%',ROWSPAN=>2, COLSPAN=>2,
                  ALIGN=>'LEFT',VALIGN=>'MIDDLE', BGCOLOR=>'#FFFFFF');
      my $browse_location = $REQ->param('p.browse_location');
      $draw_mapslct->draw();
      $HTML -> _TD->_TR;
      $HTML -> TR(VALIGN=>'TOP') ;
      $HTML -> TD;
      $HTML -> small;
      # Only show summary text and map if geo type stored in query matches
      # current geo widget.
      my $map;
      if ($crit->geo->ODL_type eq 'PATH_ROW_LOC') {
         print join("<BR>",
            split /\n/, html_escape($crit->as_string(Verbose=>2)));
         $map = $self->map_src();
      } else {
         print 'No Path/Row Range to show on map:<BR>';
         my $mapfile = lookup IW::GeoSelect::Mapfile 'Equatorial Map (small)';
         $map = query_url($mapfile->script,
                         'map'   => $mapfile->filename,
                         'grid'  => 'full');
      }
      print qq(<IMG NAME=map SRC=$map ALT="),$self->map_alt,qq(">);
   } elsif ($draw_mapname eq 'Orthographic Map (Java)') { 

      if ($self->print_summary($crit, $draw_mapname)) {

         my $map = $self->map_src;
         if ($map) {
                  
            $HTML -> TD(WIDTH=>'95%',ROWSPAN=>2, COLSPAN=>2, 
                        ALIGN=>'LEFT',VALIGN=>'MIDDLE', BGCOLOR=>'#FFFFFF');
            $HTML -> button(SRC=>$map, ALT=>$self->map_alt);
            $HTML -> _TD->_TR;
            $HTML -> TR(VALIGN=>'TOP') ;
            $HTML -> TD(WIDTH=>'5%');
            $HTML -> small; 
            print join("<BR>", 
                  split /\n/, html_escape($crit->as_string(Verbose=>2)));
                    
         }# end if map
            
 
      } else {
 
         #geo from ortho map was deleted, default to global type-in
         $HTML -> TD(WIDTH=>'95%',ROWSPAN=>1, COLSPAN=>2, 
                     ALIGN=>'LEFT',VALIGN=>'MIDDLE', BGCOLOR=>'#FFFFFF');
          
         $mapname = "Type in Lat/Lon Range";           # default map 
         $geoslct = lookup IW::GeoSelect::NewBase $mapname;
         $geoslct->state($REQ->param('p.state'));      # set state
         my $geo = IU::Geo::Range->new(90.0, -180.0, -90.0, 180.0);
         $geo->fixup->validate if ($geo);         # fix up and validate geo
         $QUERY->geo_region($geo);            # store in query object
         $QUERY->install;
         $crit = $self->{Crit} = ($QUERY->criteria('GEOGRAPHIC_REGION'))[0];
         $draw_mapname = $mapname;
         my $draw_mapslct = $geoslct->switch_to($draw_mapname);
         $REQ->param('p.mapname-next' => $draw_mapname);      # init widget
         $self->{OptionType} = $draw_mapname;
         # Remember this:
         $SESSION->state->{LastMapname} = $draw_mapname;
         $HTML -> hidden('p.mapname' => $draw_mapname);
      }

   } else {
      # not global granules, not java
      if ($self->print_summary($crit, $draw_mapname)) {

         $HTML -> TD(WIDTH=>'95%',ROWSPAN=>2, COLSPAN=>2, 
                     ALIGN=>'LEFT',VALIGN=>'MIDDLE', BGCOLOR=>'#FFFFFF');
         my $browse_location = $REQ->param('p.browse_location');
         $draw_mapslct->draw(Upload=>$q_uploaded);
         $HTML -> _TD->_TR;
         $HTML -> TR(VALIGN=>'TOP') ;
         $HTML -> TD;
         $HTML -> small; 
         print join("<BR>", 
               split /\n/, html_escape($crit->as_string(Verbose=>2)));
#         my $map = $self->map_src();  
#         print "<IMG NAME=map SRC=$map ALT=\'MAP\'>"; 

      } else {

         $HTML -> TD(WIDTH=>'95%',ROWSPAN=>1, COLSPAN=>2, 
                     ALIGN=>'LEFT',VALIGN=>'MIDDLE', BGCOLOR=>'#FFFFFF');
         my $browse_location = $REQ->param('p.browse_location');
         $draw_mapslct->draw(Upload=>$q_uploaded);
         $HTML->_TD->_TR;
      }
   }
   $HTML -> _TD -> _TR;
   1;
}
#--------------------------------------------------
# print_summary
#--------------------------------------------------
# print the map summary
sub print_summary {
  my ($self, $crit, $draw_mapname) = @_;
  my $print_it = 0;
  if($crit) {  
      my $geo = $crit->geo;
       
      #Check to see that there are no conflicts between the mapname
      # and the odl type. If there are, don't print the summary
      if (($draw_mapname eq "Orthographic Map (Java)") 
           && ($geo->ODL_type eq 'POLYGON_LOC')) {
           $print_it = 1;
      }elsif (($draw_mapname eq "Equatorial Map") 
             && ($geo->ODL_type eq 'RANGE_LOC' )) {
           $print_it = 1;
      }elsif (($draw_mapname eq "Stereographic N-pole Map") 
             && ($geo->ODL_type eq 'POLYGON_LOC')
             && ($geo->projection eq 'NORTH_POLAR_STEREOGRAPHIC')) {
           $print_it = 1;
      }elsif (($draw_mapname eq "Stereographic S-pole Map")
             && ($geo->ODL_type eq 'POLYGON_LOC')
             && ($geo->projection eq 'SOUTH_POLAR_STEREOGRAPHIC')) {
            $print_it = 1;
      }elsif (($draw_mapname eq "Type in Lat/Lon Range")
             && ($geo->ODL_type eq 'RANGE_LOC' )) {
            $print_it = 1;
      }elsif (($draw_mapname eq "Type in Lat/Lon Point")
             && ($geo->ODL_type eq 'POINT_LOC' )) {
            $print_it = 1;
      }elsif (($draw_mapname eq "Type in Path/Row Range")
             && ($geo->ODL_type eq 'PATH_ROW_LOC' )) {
            $print_it = 1;
      }elsif (($draw_mapname eq  "Global Search")
             && ($geo->ODL_type eq 'RANGE_LOC')) {
            $print_it = 1;
      }
  }
  $print_it;
}                  
#--------------------------------------------------
# upload_mapname
#--------------------------------------------------
# When the user has uploaded a query, the map name is
# not saved.  We need to determine the correct map 
# name based on the projection and odl type.
# 
sub upload_mapname {
  my ($self, $crit) = @_;
  my $draw_mapname;
  if ($crit) {
      my $geo = $crit->geo;
      my $odl_type = $geo->ODL_type;

      if( $odl_type eq 'POLYGON_LOC') {
          if($geo->projection eq 'ORTHOGRAPHIC'){
              $draw_mapname = "Orthographic Map (Java)";
          }elsif($geo->projection eq 'NORTH_POLAR_STEREOGRAPHIC') {
              $draw_mapname = "Stereographic N-pole Map";
          }elsif($geo->projection eq 'SOUTH_POLAR_STEREOGRAPHIC') {
              $draw_mapname = "Stereographic S-pole Map";
          }
     }elsif($odl_type eq 'RANGE_LOC') {
         $draw_mapname = "Type in Lat/Lon Range";                
     }elsif($odl_type eq 'POINT_LOC') {
         $draw_mapname = "Type in Lat/Lon Point";
     }elsif($odl_type eq 'PATH_ROW_LOC') {
         $draw_mapname = "Type in Path/Row Range";
     }elsif($odl_type eq 'GLOBAL_GRANULES_ONLY') {
         $draw_mapname = "Global granules only";
     }else {
         $draw_mapname = "No geographic region";
     }
    }
   $draw_mapname ;
}
#------------------------------------------------------------
# print_toggle_buttons
# Arguments to print_toggle_buttons are:
# outerincrement - the increment of the outerloop that prints the 
#    toggle buttons in rows. Should increment by number of rows
#    needed.
# innerlimit - if more than one row is needed for all of the 
#    options, inner limit is number of items in the row.
# opt_types - array of tags for the buttons
# paramname - the current selected option
# mode - return mode to set the url for the toggle buttons
# submode - return submode to set url for toggle buttons
# nextsubmode - next submode to set url for toggle buttons
# Overrides class method
#
#------------------------------------------------------------
sub print_toggle_buttons {
    my ($self, $outerincrement, $innerlimit, $opt_types, 
        $paramname, $mode, $submode, $nextsubmode) = @_;

  for (my $i = 0; $i < int(@$opt_types); $i += $outerincrement) {

        $HTML -> TR;
        for (my $j = 0; $j < $innerlimit; $j++) {
           $HTML -> TD(NOWRAP=>undef, WIDTH=>'1%');

	   # Get info for this geo type:
	   my $typeinfo = $$opt_types[$i+$j];
           
	   # Get URL for switching option type (if any):
	   my $src = undef;

           # Set the source for the toggle button, if it is
           # on we get a toggle-on, otherwise a toggle-off.
           if ($$opt_types[$i+$j] ne $self->{OptionType}) {
              $src = "${IU::DocsUrl}icons/toggle-off.gif";
           }
           else {
              $src = "${IU::DocsUrl}icons/toggle-on.gif";
           }

           # Create the toggle buttons, the java button is special.              
           if($$opt_types[$i+$j] eq "Orthographic Map (Java)") {
              $HTML-> msubmit_button(QUERY => ['mode'=>'SSRCHFORM',
                                               'submode' =>'READINPUT',
                                               'nextmode' =>'SSLCTGEO',
		   	                       'nextsubmode' =>'SWITCH',
                                               'param' => $paramname,
                                               'value' =>$$opt_types[$i+$j]],
                                     ALT   => "Switch to $$opt_types[$i+$j]",
                                     SRC   => $src);

           } else {
             $HTML-> msubmit_button(QUERY => ['mode'=>$mode,
                                           'submode'=>$submode,
                                           'nextsubmode'=>$nextsubmode,
                                           'param'=> $paramname,
                                           'value' =>$$opt_types[$i+$j] ],
                                     ALT   => "Switch to $$opt_types[$i+$j]",
                                     SRC   => $src);
	   }
           # Print a space between toggle button and label
	   print " ";
	   $HTML -> B -> t($$opt_types[$i+$j]) -> _B;
           $HTML -> _TD;
       }

       $HTML -> _TR;
    }
}

#------------------------------------------------------------
1;



