package IW::Widget::DepValids;

=head1 NAME

IW::Widget::DepValids - manage dependent valids selection (mode SLCTDEPVAL)

=head1 DESCRIPTION

Each instance represents one multiple-selection box in the HTML
form on the dependent valids page... an instance knows how to render 
itself, and how to read the CGI input when the form is submitted.

=head1 NOTES

This module introduces the concept of "clean" versus "annotated"
valids.  A "clean" valid is a valid in its natural form; e.g,

     EDC

An "annotated" valid has a leading "*" if it is incompatible:

     *EDC

The clean_valid() function can be used to clean up an arbitrary valid. 

=cut

use IW::App;
use IU::Log;
use strict;
use vars qw(@ISA @EXPORT);

@ISA = qw(Exporter);
@EXPORT = qw(clean_valid);

#------------------------------------------------------------
# clean_valid
#------------------------------------------------------------
# I<Function, exported.>
# Remove leading "*" (any maybe other stuff in the future?) from string.

sub clean_valid {
    my $str = shift;
    $str =~ s/^\*\s?//;
    $str;
}

#------------------------------------------------------------
# show_valid CLEAN
#------------------------------------------------------------
# I<Instance method.>
# Show valid for display.  Because of comments, the result may 
# not be "cleanable!"

sub show_valid {
    my ($self, $clean) = @_;
    my $inval  = $self->{Invalid}{$clean} ? '*' : '';
    my $cmnt   = $self->{Comments}{$clean};
    $inval . $clean . ($cmnt ? " ($cmnt)" : '');
}

#------------------------------------------------------------
# new PARAMHASH
#------------------------------------------------------------
# I<Class method, constructor.>
# PARAMHASH contains...
#
#     All      => All menu entries, as an array of strings.
#                 Invalid ones should have a leading "*".   
#
#     Selected => Any valids currently selected.
#                 They will be automatically cleaned (i.e., any leading "*"s
#                 will be ignored; only the "*"s in All are heeded).
#                 Any ones not present in All will be silently added as 
#                 "incompatible" valids.
#
#     Comments => Optional ref to hash mapping clean valids to comments.
#                 Intended for acronym expansion.
#                 Comments are displayed in the menu after the valid itself.

sub new {
    my ($class, %params) = @_;

    # Get arguments:
    my @a_all      = @{$params{All} || []};
    my @a_selected = @{$params{Selected} || []};

    # Get CLEAN list of ALL valids (clean == without stars/comments).
    # Be sure to remove duplicates:
    my %in_full = map {clean_valid($_) => 1} @a_all, @a_selected;
    my @full = keys %in_full;

    # Get CLEAN lookup table of SELECTED valids:
    my %is_selected   = map {clean_valid($_) => 1
			     } @a_selected; 

    # Get CLEAN lookup table of UNSELECTED valids:
    my %is_unselected = map {$_ => 1
			     } (grep {!$is_selected{$_}} @full);

    # Get CLEAN lookup table of INVALID valids (ones in All with leading "*"):
    my %is_invalid    = map {clean_valid($_) => 1
			     } (grep /^\*/, @a_all);  

    # Create info:
    my $self = {
	Selected   => \%is_selected,
	Unselected => \%is_unselected,
	Invalid    => \%is_invalid,
	Comments   => ($params{Comments} || {}),
    };
    bless $self, $class;
}

#------------------------------------------------------------
# items
#------------------------------------------------------------
# I<Instance method.>
# Return the menu items, as an array of strings in the following
# form:
#                selected item
#                ...
#                selected item
#                ----------
#                unselected item
#                ...
#                unselected item
#
# The items you get back will NOT have leading "*"s or comments in them!
#
# The dividing line is only put in if there are both selected and
# unselected items.
#
# ----------------------------------------------------------------------
# NOTE:  Change for 2.3
# Because some browsers show the list of items with the first value 
# selected as a default, we are adding an extra line of dashes only 
# no items are selected.  The list should look like this:
#                ----------
#                selected item
#                ...
#                selected item
# ----------------------------------------------------------------------
#
#
sub items {
    my $self = shift;
    my ($inv, $cmnt);

    my @selected = (sort keys %{$self->{Selected}});    
    my @unselected = (sort keys %{$self->{Unselected}});
                             # Add dash line only if none selected.
    return ( ((@selected) ? () : "----------" ), 
            @selected,
	    ((@selected && @unselected) ? "----------" : ()),
	    @unselected);
}

#------------------------------------------------------------
# selected
#------------------------------------------------------------
# I<Instance method.>
# Return the selected items, as a ref to a hash whose keys
# are the clean items and whose values are true.
#
sub selected {
    shift->{Selected};
}

#------------------------------------------------------------
# print PARAMHASH
#------------------------------------------------------------
# I<Instance method.>
# Output the widget.  PARAMHASH contains:
#
#     Name  => widget name
#     Size  => number of rows displayed
#     Title => String to be used in titling multi-section tables:
#              "Whatsit list" translates to "Whatsit list 1",
#              "Whatsit list 2", etc.
#
sub print {
    my ($self, %params) = @_;
    my $name = $params{Name};
    my $size = $params{Size} || 5;
    my $title = $params{Title} || "";

    # Pare size down:
    my @items = $self->items;
    $size = (@items || 1) if (@items < $size);

    # Create choice map:
    my @choicemap = map {
	(/^-+$/ 
	 ? ($_ => '')                         # separator
	 : ($self->show_valid($_) => $_))     # valid
    } @items;

    ### KLUDGE: Mac Netscape breaks when the number of items goes 
    ###         over a limit of around 500 items. We split the list
    ###         into segments to get around this.
    ###
    ###         Since the items in choicemap are key-value pairs, we
    ###         need to double the breakage limit. For example, if
    ###         we want 200-item lists, $BREAKAGE_LIMIT = 400;

    my $BREAKAGE_LIMIT = 1000;

    # Output:
#    $HTML -> P;
    my $count = 1;
    while (@choicemap) {
        my @segment = splice(@choicemap,0,$BREAKAGE_LIMIT);
        $HTML ->BR-> B -> t($title . " " . $count++.':') -> _B -> BR if $title;
        $HTML -> select(Name  => $name,
                        Multi => 1,
                        Size  => $size,
                        ChoiceMap => \@segment,
                        Values    => [keys %{$self->{Selected}}],
                        None => 1);
#        $HTML -> P; 
    }
}

#------------------------------------------------------------
# count
#------------------------------------------------------------
# I<Instance method.>
# Return a hashref to some counts of items.  The keys to the hashref
# are as follows:
#
#      Total       => Total number of items 
#      Selected    => Number of items which are SELECTED
#      Unselected  => Number of items which are UNSELECTED
#      Valid       => Number of items WITHOUT a leading "*"
#      Inalid      => Number of items WITH a leading "*"
#
sub count {
    my $self = shift;
    my $info = {};
    $info->{Selected}   = keys %{$self->{Selected}};
    $info->{Unselected} = keys %{$self->{Unselected}};
    $info->{Total}      = $info->{Selected} + $info->{Unselected};
    $info->{Invalid}    = keys %{$self->{Invalid}};
    $info->{Valid}      = $info->{Total} - $info->{Invalid};
    $info;
}

#------------------------------------------------------------
1;
