package IW::Widget::QueryEdit;

=head1 NAME

IW::Widget::QueryEdit - manage the "grey navigator" for editting queries

=head1 DESCRIPTION

This displays and manages the widget that appears throughout the
search forms, both on the main page and in the "popups".

=cut

use strict;
use vars qw(@ISA @EXPORT);
use Exporter;

use IU;
use IU::Log;
use IU::Query;
use IU::Category;

use IW::App;

@ISA = qw(Exporter);



#------------------------------
#
# Methods...
#
#------------------------------

#------------------------------------------------------------
# new PARAMHASH
#------------------------------------------------------------
# Create a new instance.  Params are:
#
#      Popup => Category if created for a popup (e.g., SOURCE_NAME).
#               If missing, assumes that this is not for a popup.
#      Directions => alternate HTML text for the directions.

sub new {
    my ($class, %params) = @_;
    my $self = bless \%params, $class;
    if (!$params{Popup}) {
	$self->{Directions} = <<EOF;
	    <B>To edit your query,</B> select the query attribute you wish to
	    add/modify, and hit the EDIT button.  
	    <B>To submit your query,</B> use the "START SEARCH" button at 
	    lower-left (the button will appear only if your query is ready).
EOF
    }
    else {
	$self->{Directions} = <<EOF;
	    <B>Optional:</B> 
            Normally when you're done with this screen (e.g., by clicking OK,
            Cancel, or Return), 
            you'd go back to the main search form.  However, you can use 
            this menu to change your destination.
EOF
    }

    $self->{HomeMode} ||= $SESSION->state->{SearchFormMode};
  

    $self->{ReturnTo} = 'MAIN_SEARCH_FORM';
    $self;
}

#------------------------------------------------------------
# set_return CATEGORY
#------------------------------------------------------------
# Instance method. Change the default setting of the pop-up.
#------------------------------------------------------------
sub set_return {
   my ($self,%target) = @_;
   $self->{ReturnTo} = $target{ReturnTo} if defined $target{ReturnTo};
   $self;
}

# sig CATEGORY
#------------------------------------------------------------
# Class/instance method.  Return the signature of the given CATEGORY.
#
sub sig {
    my ($self, $cat) = @_;
    IU::Category->lookup($cat)->{Sig};
}

#------------------------------------------------------------
# context CATEGORY
#------------------------------------------------------------
# Class/instance method.  Return the signature of the given CATEGORY.
#
sub context {
    my ($self, $cat) = @_;
    IU::Category->lookup($cat)->{Context};
}

#------------------------------------------------------------
# get_menu_map
#------------------------------------------------------------
# Instance methods, private.
# Get information we need to build the menus.

sub get_menu_map {
    my @items;
    my %menu;
    my $item;
    my %map;

    # Geo, prettified

    $menu{Geo} = {};
    my @Core;

    # Put the new core attributes in the jump menu for inventory form only
    if ($QUERY->search_type eq 'INVENTORY_SEARCH') {
       foreach $item (IU::Category->get_core_in_display_order) {
          IU::Category->is_core($item) or next;
          $QUERY->uses($item) or next;
          push(@Core , category2pretty($item),$item );
       }
    }
    # Put the guide core attributes in the jump menu for guide and
    # directory forms only
    elsif (($QUERY->search_type eq 'GUIDE_SEARCH') || 
           ($QUERY->search_type eq 'DIRECTORY_SEARCH')){
       foreach $item (IU::Category->get_core_in_display_order) {
          IU::Category->is_guidesumm_core($item) or next;
          $QUERY->uses($item) or next;
          push(@Core , category2pretty($item),$item );
       }
    }

    else {
       foreach $item ('GEOGRAPHIC_REGION') {
           $QUERY->uses($item) or next;
           $menu{Geo}{category2pretty($item)} = $item;
       }

       # Non-dependent categories, prettified and sorted:
       $menu{ID} = {};
       foreach $item (qw(GRANULE_ID_REQ)) {
           $QUERY->uses($item) or next;
           $menu{ID}{category2pretty($item)} = $item;
       }

       # Non-dependent categories, prettified and sorted:
       # Added Request ID for Product Status Request
       $menu{RID} = {};
       foreach $item (qw(REQUEST_ID)) {
           $QUERY->uses($item) or next;
           $menu{RID}{category2pretty($item)} = $item;
       }

       # Non-dependent categories, prettified and sorted:
       $menu{NonDep} = {};
       foreach $item (qw(TEXT_STRING        
   		      TIME_RANGE)) {
          $QUERY->uses($item) or next;
          $menu{NonDep}{category2pretty($item)} = $item;
       }

       # Core dependent categories, prettified and sorted by display order.
       foreach $item (IU::Category->get_core_in_display_order) {
          $QUERY->uses($item) or next;
          push(@Core , category2pretty($item),$item );
       }

       # Extended categories, prettified and sorted:
       $menu{Ext} = {};
       foreach $item ('EXTENDED_SEARCH') {
  	  $QUERY->uses($item) or next;
	  $menu{Ext}{"Other criteria..."} = $item;
       }

       # Other stuff:
       $menu{Misc} = {"Search Options" => 'SEARCH_OPTIONS'};
    }


    # Build the map:
    my @Geo     = map {$_ => $menu{Geo}{$_}} sort keys %{$menu{Geo}};
    my @ID      = map {$_ => $menu{ID}{$_}} sort keys %{$menu{ID}};
    my @RID      = map {$_ => $menu{RID}{$_}} sort keys %{$menu{RID}};
    my @NonDep  = map {$_ => $menu{NonDep}{$_}} sort keys %{$menu{NonDep}};
    my @Ext     = map {$_ => $menu{Ext}{$_}   } sort keys %{$menu{Ext}};
    my @Misc    = map {$_ => $menu{Misc}{$_}  } sort keys %{$menu{Misc}};  
    my @CoreExt = (@Core, @Ext);
    return ("(SELECT ONE...)"=>'',
            (@Geo     ? ("----------"=>'', @Geo)  : ()),
            (@ID      ? ("----------"=>'', @ID)  : ()),
            (@RID      ? ("----------"=>'', @RID)  : ()),
	    (@CoreExt ? ("----------"=>'', @CoreExt) : ()),
            (@NonDep  ? ("----------"=>'', @NonDep)  : ()),
	    (@Misc    ? ("----------"=>'', @Misc)    : ()));
}

#------------------------------------------------------------
# go
#------------------------------------------------------------
# Class method.  Create a new instance of this class, then
# immediately do a "goto" as dictated by the "goto" CGI param.
#
sub go {
    my $class = shift;
    return $class->new->goto($REQ->param('goto'));
}

#------------------------------------------------------------
# goto DESTINATION
#------------------------------------------------------------
# Instance method.
# Given a particular DESTINATION to go to (core category, extended category, 
# or special): load the appropriate "mode" module, create an instance of 
# that mode, and send it a "goto" message.
#
sub goto {
    my ($self, $goto) = @_;
    my $gkey = $REQ->param('gkey');

    # KLUDGE: clear out CGI params:
    my $p;
    foreach $p (grep { /^p\./ } $REQ->params) {
	### debug "deleteing CGI param $p";
	$REQ->delete($p);
    }
    # Get the destination:
    my $cat = $goto;
    
    # Get the mode we need:
    my $modename = $self->mode_for($cat);
    if (!$modename) {       # default is to go to main screen
	warn "no support for going to destination $cat";
	$modename = $self->{HomeMode};
    }

    # Load module for mode (NOTE: we need the eval to make Perl happy):
    my $modemod = "IW::Mode$modename";
    eval "require $modemod" or puke "can't load mode $modemod!";

    # Go!
    
    my $mode = $modemod->new(Category=>$cat);
    return $mode->goto($cat);    
}

#------------------------------------------------------------
# mode_for CATEGORY
#------------------------------------------------------------
# Instance method.
# Map a category destination to a mode.

sub mode_for {
    my ($self, $cat) = @_;
    my $modename;

    my $java = $IW::App::SESSION->state->{AllowJava};
    my $search_type = $QUERY->search_type;

    $cat =~ s/^-+$//;     # KLUDGE to get rid of separators

    if (!$cat or ($cat eq 'MAIN_SEARCH_FORM')) {
	$modename = $self->{HomeMode};
    }
    elsif (IU::Category->is_core($cat)) {
        $modename = 'SLCTDEPVAL';
    }
    elsif ($cat eq 'EXTENDED_SEARCH') {
	$modename = 'SLCTALL';
    }
    elsif ($cat eq 'GEOGRAPHIC_REGION') {
        $modename = 'MAPPLET';
    }
    else {
	$modename = 'SLCTALL';
    }
    return $modename;       
}

#------------------------------------------------------------
# print
#------------------------------------------------------------
# Instance method.
# Render yourself to the default HTML output stream.
#
sub print {
    my $self = shift;
    $self->{Popup} ?  $self->_print_popup : $self->_print_main;
}

#------------------------------------------------------------
# _print_main
#------------------------------------------------------------
# Instance method, private.  Print self as a main menu.
#
sub _print_main {
    my $self = shift;

    my $gkey = $REQ->param('gkey');
    
    # Get menu items:
    my @menumap = get_menu_map();

    # Directions:
    $HTML -> P -> small;
    print $self->{Directions};
    $HTML -> _small;

    # Start table:
    $HTML -> CENTER;
    $HTML -> TABLE(CELLPADDING=>1, CELLSPACING=>1, BORDER=>0);
    $HTML -> TR;

    # Submit button and pulldown:
    $HTML -> TD(NOWRAP=>undef, VALIGN=>'MIDDLE', ALIGN=>'CENTER');
    $HTML -> msubmit_flag;
    $HTML -> msubmit_button(CAPTION =>'Edit...',
			    QUERY   => ["mode"    => $self->{HomeMode},
					"submode" => "GOTO",
                                        "gkey" => $gkey]);
    $HTML -> select_one(Name      => 'goto',
			Value     => 'MAIN_SEARCH_FORM',
			ChoiceMap => \@menumap);

    $HTML -> _TABLE;
    $HTML -> _CENTER;
}

#------------------------------------------------------------
# _print_popup
#------------------------------------------------------------
# Instance method, private.  Print self as a popup menu.

sub _print_popup {
    my $self = shift;

    # Get menu menumap:
    my @menumap = get_menu_map();
    ($menumap[0], $menumap[1]) = ('Main Search Form' => 'MAIN_SEARCH_FORM');
    
    my $gkey = $REQ->param('gkey');
    

    # Start table:
    $HTML -> P -> CENTER;
    $HTML -> TABLE(CELLPADDING=>1, CELLSPACING=>0, BORDER=>0, WIDTH=>'100%');
    $HTML -> TR;

    # Row 1-2, Col 1...
    # Icon:
    $HTML -> TD(VALIGN=>'MIDDLE', ROWSPAN=>2);
    $HTML -> IMG(SRC=>"${IU::DocsUrl}icons/goto.gif");
    $HTML -> _TD;

    # Row 1, Col 2...
    # Directions:
    $HTML -> TD;
    $HTML -> small;
    print $self->{Directions};
    $HTML -> _small;
    $HTML -> _TD -> _TR;

    # Row 2, Col 2...
    # Submit button and pulldown:
    $HTML -> TR;
    $HTML -> TD(NOWRAP=>undef, VALIGN=>'MIDDLE');
    $HTML -> msubmit_flag;
    $HTML -> t("Go to ");
    $HTML -> select_one(Name    => 'goto',
			Value   => $self->{ReturnTo},
			ChoiceMap => \@menumap);
    $HTML -> t("when I'm done with this screen.");
    $HTML -> _TD -> _TR;
    $HTML -> _TABLE;
    $HTML -> _CENTER;
}

#------------------------------------------------------------
1;
