package IW::Widget::ShowCategory;

=head1 NAME

IW::Widget::ShowCategory - widget on search form for displaying category

=head1 DESCRIPTION

An instance of this class displays a single query category on the
main search form.  Other, more-tailored classes are derived from this class.
The following are the possible scenarios:

    CATEGORY is...                  CLASS to use...
    ----------------------------------------------------------
    Traditional dependent valids    ShowDepval
    Extended dependent valids       ShowExt
    Sole geographic region          ShowGeo
    Other (with single-criterion)   ShowCategory (**there should be a ShowOne)

=cut

use IU;
use IU::Log;
use IU::Query;
use IU::Category;

use IW::App;
use IW::Util qw(:url);

use HTML::Stream qw(:funcs);

use strict;


#------------------------------------------------------------
# new PARAMHASH
#------------------------------------------------------------
# I<Constructor.>  PARAMHASH can contain:
#
#     BGCOLOR  => background color of table cells (default is light grey)
#     Category => the query category... mandatory!
#     Mode     => the main search-form mode to perform edit/delete with
#     ShowCritTitles => show titles of individual criteria? default is false
#     Title    => widget title (used in place of pretty category name)

sub new {
    my ($class, %params) = @_;
    $params{Mode}    ||= 'NOMODE';
    $params{BGCOLOR} ||= '#D0D0D0';
    $params{Category} || puke 600, "can't have widget without category!";
    bless \%params, $class;
}

#------------------------------------------------------------
# category
#------------------------------------------------------------
# I<Instance method.>  Return my category.

sub category {
    shift->{Category};
}

#------------------------------------------------------------
# title
#------------------------------------------------------------
# I<Instance method.>  Return my title. Note: overriden for
# extended search criteria (ShowExt.pm).

sub title {
   my $self = shift;
   $self->{Title} || category2pretty($self->category);
}

#------------------------------------------------------------
# icon
#------------------------------------------------------------
# I<Instance method.>
# Return the SRC for the icon.
#
sub icon {
    my $self = shift;
    my $cat = $self->category;

    # Clean up category name, to create a valid filename:
    $cat = uc($cat);
    $cat =~ tr/A-Z0-9_/_/c;
    $cat =~ s/_+/_/g;
    $cat =~ s/^_|_$//g;
#    IU::Log::debug("cat icon for ", $self->category, " = $cat");

    # Get icon, or fallback icon if none:
    my $icon_name = "icons/category/$cat.gif";
    return ((-e "$IU::DocsDir/$icon_name") 
	    ? "${IU::DocsUrl}$icon_name"
	    : "${IU::DocsUrl}icons/category/misc.gif");
}

#------------------------------------------------------------
# print PARAMHASH
#------------------------------------------------------------
# I<Instance method.>
# Each widget is three table cells in a row in a three-column table.
# PARAMHASH may contain:
#
#      Anchor => put a SCROLL anchor around the widget text

sub print {
    my ($self, %params) = @_;

    # Get category:
    my $cat = $self->category || die "can't print without category!";
    my $catinfo = IU::Category->lookup($cat);
    my $defn    = $catinfo->{Defn};
    my $context = $catinfo->{Context};
    my $delete  = defined($self->{Delete}) ? $self->{Delete} : 1; # can delete
    my $hasicon = defined($self->{Icon})   ? $self->{Icon}   : 1; # has icon
    
    my $gkey = $REQ->param('gkey');
    

    #-------------
    # COLUMN 1...
    #-------------
    # Name:
    $HTML -> TD;
    print "<NOBR>";
    $HTML -> A(NAME=>'SCROLL')      if $params{Anchor};
    $HTML -> B -> t($self->title) -> _B;
    $HTML -> _A                     if $params{Anchor};
    $HTML -> BR;
    
    #-------------
    # Check for search form type to display gif buttons differently.
    # The inventory search, guide search, data granule id, and local 
    # granule id forms need to have submit buttons to be able to read 
    # any input fields in the form.
    #
    # Other search forms also need tiny help buttons in each category
    # while the inventory search, guide search, data granule id, and local 
    # granule id form does not require them.
    #-------------
    if ($QUERY->search_type ne 'ACCOUNT_STATUS_REQUEST') {
       # Edit button:
       $HTML -> msubmit_button(SRC   => "${IU::DocsUrl}btns/tiny-edit.gif",
                               ALT   => "Edit",
                               QUERY => ['mode'        =>'SSRCHFORM',
                                         'submode'     =>'READINPUT',
                                         'nextsubmode' =>'GOTO',
                                         'goto'        =>$cat,
                                         'gkey'        =>$gkey]);
       # Delete button:
       $HTML -> msubmit_button(SRC   => "${IU::DocsUrl}btns/tiny-delete.gif",
                                ALT   => "Del",
                                QUERY => ['mode'        =>'SSRCHFORM',
                                          'submode'     =>'READINPUT',
                                          'nextsubmode' =>'DELETECAT',
                                          'category'    =>$cat,
                                          'gkey'        =>$gkey]);
    }
    else {  # in old search form
       # URL for edit button:
       my $edit_url   = script_url('mode'        =>$self->{Mode},
                                   'submode'     =>'READINPUT',
                                   'nextmode'    =>$self->{Mode},
                                   'nextsubmode' =>'GOTO',
                                   'goto'        =>$cat,
                                   'gkey'        =>$gkey);
       # Edit button:
       $HTML -> button(SRC  =>"${IU::DocsUrl}btns/tiny-edit.gif",
                       ALT  =>"Edit",
                       HREF =>$edit_url);

       # URL for delete button:
       my $delete_url = script_url('mode'        =>$self->{Mode},
                                   'submode'     => 'READINPUT',
                                   'nextmode'    =>$self->{Mode},
                                   'nextsubmode' =>'DELETECAT',
                                   'category'    =>$cat,
                                   'gkey'        =>$gkey);
       # Delete button:
       $HTML -> button(SRC  =>"${IU::DocsUrl}btns/tiny-delete.gif",
                       ALT  =>"Del",
                       HREF =>$delete_url);
       
       # Help button:
       $HTML -> help_btn($context)     if $context;
    }

    print "</NOBR>";
    $HTML -> _TD;

    #-------------
    # COLUMN 2...
    #-------------
    my @criteria = $QUERY->criteria($cat);
    $HTML -> TD(($self->{NOWRAP} ? (NOWRAP=>undef) : ()),
		(@criteria ? (BGCOLOR=>'#FFFFFF') : ()));

    # Criterion:
    if (@criteria) {
	$HTML -> small;
	$self->print_criteria(\@criteria);
	$HTML -> _small;
    }
    else {
	# Brief definition:
	if ($defn and !($self->{NoDefn})) {
	    $HTML -> t("($defn)");
	}
    }
    $HTML -> _TD;
}

#------------------------------------------------------------
# print_criterion_title CRIT
#------------------------------------------------------------
# I<Instance method, private.>
# Output the criterion title.  This is desirable for extended search
# categories, or any categories which may take more than one variant.
# Privately invoked during print_criteria().
#
# Normally disabled unless ShowCritTitles is set true, since "normal"
# query criteria have exactly one variant.

sub print_criterion_title {
    my ($self, $crit) = @_;
    $self->{ShowCritTitles} or return undef;

    # Get name, preceding it with a "*" if it was judged invariant:
    my $name = ($self->{InvalidVariant}{$crit->variant}
		? "*" . $crit->name     # invalid
		: $crit->name);         # still valid
    
    # Output:
    $HTML -> t_B($name) -> BR;
}

#------------------------------------------------------------
# print_criteria CRITS
#------------------------------------------------------------
# I<Instance method.>
# Output all the criteria in this category.
# Subclasses may wish to override this.

sub print_criteria {
    my ($self, $crits) = @_;       
    my $crit;

    # Output criteria:
    foreach $crit (@$crits) {
	# Separator?
	$HTML -> P  if ($crit ne $crits->[0]);

	# Print title:
	$self->print_criterion_title($crit);
	
	# Print criteria:
	my $hspace = ($self->{NoDefn} ? "" : "&#160;&#160;&#160;");
	print join("<BR>",
		   split /\n/, html_escape($crit->as_string(Verbose=>1)));
    }
}

#------------------------------------------------------------
1;


