package IW::Widget::ShowExt;

=head1 NAME

IW::Widget::ShowExt - widget for "extended" category

=head1 DESCRIPTION

Each of these objects display a single query category which is an
"extended search criteria" (or "extended dependent valids")
category... e.g., "Cloud Cover".

Such categories have one or more criteria or any type, where each
criterion is also known as a "variant" (e.g., "Percent Cloud Cover").  
Like the valids of traditional dependent valids, the varient names
are run through the dependent-valids program, and each variant name is 
annotated with a "*" if that variant has become invalid.

=cut

use IU::Query;
use IU::Category;

use IW::App;
use IW::Util qw(:url);
use IW::Widget::ShowCategory;


use strict;
use vars qw(@ISA);

@ISA = qw(IW::Widget::ShowCategory);

#------------------------------------------------------------
# new
#------------------------------------------------------------
# Constructor, override.
#
sub new {
    my $self = shift->SUPER::new(@_);
    $self->{ShowCritTitles} = 1;   # force on
### $self->{NOWRAP} = 1;            ### screws up display
    $self;
}

#------------------------------------------------------------
# title
#------------------------------------------------------------
# I<Instance method.>  Return my title.
# Note: extended search criteria not pretty printed.
#
sub title {
   my $self = shift;
   $self->{Title} || extended2pretty($self->category);
}

#------------------------------------------------------------
# print_criteria
#------------------------------------------------------------
# Instance method, override.
#
sub print_criteria {
    my ($self, $crits) = @_;

    # Get current variants:
    my @variants = map {$_->variant} $QUERY->criteria($self->category);
    IU::Log::debug("Variants for ",$self->category," = ",join("|", @variants));
    @variants or return;         # don't print if nothing was selected!

    # Get invalid variants, stripped of leading "*", as a lookup table:
    my @all = $QUERY->valids_for($self->category);      ### runs subprocess
    IU::Log::debug("Valids for ",$self->category," = ",join("|", @all));
    my %is_invalid = map {$_=>1} IW::DepValid::invalid_in(\@all);
    $self->{InvalidVariant} = \%is_invalid;

    # Call inherited routine:
    $self->SUPER::print_criteria($crits);
}

#------------------------------------------------------------
1;
