package IW::Widget::ShowGeo;

=head1 NAME

IW::Widget::ShowGeo - widget to show a geoselection category

=head1 DESCRIPTION

An instance of this class displays the single query category for
a geographic region.

=head1 NOTES

Should attempt to take more advantage of its superclass' methods.

=cut

use IU::Query;
use IU::Category;
use IU::Log;

use IW::App;
use IW::Util qw(:url);
use IW::Widget::ShowCategory;
use IW::GeoSelect::Mapfile;
use IW::GranMap;
use HTML::Stream qw(:funcs);

use strict;
use vars qw(@ISA);

@ISA = qw(IW::Widget::ShowCategory);


#------------------------------------------------------------
# new
#------------------------------------------------------------
# I<Constructor.>

sub new {
    my ($class, %params) = @_;
    $params{Category} ||= 'GEOGRAPHIC_REGION';
    $params{Mode}     ||= 'NOMODE';
    $params{BGCOLOR}  ||= '#D0D0D0';
    bless \%params, $class;
}

#------------------------------------------------------------
# map_src
#------------------------------------------------------------
# Return the src of the summary map.
# Returns undef if map couldn't be obtained, and 'NONE' if none
# should be displayed.

sub map_src {
    my $self = shift;

    # Get criterion, and underlying GEO object:
    my $crit = $self->{Crit};
    my $geo  = $crit->geo;
    my $mapfile;

    # UGH! The switch statement of death!
    if ($geo->ODL_type eq 'GLOBAL_GRANULES_ONLY') {
        return undef;
    }
    elsif ($geo->ODL_type eq 'RANGE_LOC') {
	my ($w,$n,$e,$s) = ($geo->west, $geo->north, $geo->east, $geo->south);
	$mapfile = lookup IW::GeoSelect::Mapfile 'Equatorial Map (small)';
	return query_url($mapfile->script,
			 'map'   => $mapfile->filename,
			 'grid'  => 'full',
			 'srect' => "$w,$n,$e,$s");
    }
    elsif ($geo->ODL_type eq 'POLYGON_LOC') {

	if ($geo->projection eq 'ORTHOGRAPHIC') {

	    my $orthogif = undef;

	    my $zoom = $geo->zoom || 1;

            # hack to handle zooms greater than 16x
            if ($zoom > 16) {
               $zoom = 16;
                
            }
            my $centerLat = $geo->tangent_lat || 0;
            my $centerLon = $geo->tangent_lon || 0;

            my @dummy = ();
	    my $some_granules = \@dummy;
	    my $newstamps = 0;
            my $xsize = 200;
            my $ysize = 150;

            # Generate the GIF:
    	    my @filenames = ();
            IW::GranMap::MakeMapGifs(Granules  => $some_granules,
                 Zoom      => $zoom,
                 CenterLat => $centerLat,
                 CenterLon => $centerLon,
                 NewStamps => $newstamps,
                 XSize => $xsize,
                 YSize => $ysize,
                 Gifnames => \@filenames) || return undef;

    	    # Rename the main output file:
            my $path = $filenames[0];
            my $session_dir = $SESSION->dir;
            rename($path, "$session_dir/pub/geo_query.gif")
                     or die "couldn't rename $path: $!";

            my $url = $SESSION->url("pub/geo_query.gif?u=".scalar(time));
	
	    return ($url);

	} else {

	    # Determine mapfile:
	    my $mapname = (($geo->projection =~ /north/i)
		       ? 'Stereographic N-pole (small)'
		       : 'Stereographic S-pole (small)');
	    $mapfile = lookup IW::GeoSelect::Mapfile $mapname;     

	    # Get corners:
	    my $p0 = $geo->points->[0];      # top-left
	    my $p2 = $geo->points->[2];      # bottom-right

	    # Get source:
	    my ($w,$n,$e,$s) = ($p0->lon, $p0->lat, $p2->lon, $p2->lat);
	    return query_url($mapfile->script,
			 'map'        => $mapfile->filename,
			 'srect'      => "$w,$n,$e,$s",
			 'projection' => lc($mapfile->projection),
			 'height'     => $mapfile->height,
			 'width'      => $mapfile->width,
			 'radius'     => $mapfile->radius,
			 'centerlat'  => $mapfile->center_lat,
			 'centerlon'  => $mapfile->center_lon,
			 'sess_dir'   => $SESSION->dir);
        }
    }
    elsif ($geo->ODL_type eq 'PATH_ROW_LOC') {

	# Get bounding rectangle:
# Remove the fixup for now
#	my $br = $geo->bound_rect->fixup(2);
        my $br = $geo->bound_rect;
	$self->{ApproxGeo} = $br;

	# Get source:
	my ($w,$n,$e,$s) = ($br->west, $br->north, $br->east, $br->south);
	$mapfile = lookup IW::GeoSelect::Mapfile 'Equatorial Map (small)';
	return query_url($mapfile->script,
			 'map'   => $mapfile->filename,
			 'grid'  => 'full',
			 'srect' => "$w,$n,$e,$s");
    }
    undef;
}


#------------------------------------------------------------
# print
#------------------------------------------------------------
sub print {
    my $self = shift;

    # Listing:
    $HTML -> TABLE(WIDTH=>'100%', BORDER=>4, CELLPADDING=>4, CELLSPACING=>1,
		   BGCOLOR=>$self->{BGCOLOR});
    
    # Get list of categories, sorted on the PRETTY name:
    my @cats = sort {category2pretty($a) cmp category2pretty($b)
		     } IU::Category->get_core;

    # Get the current choices:
    my $cat = 'GEOGRAPHIC_REGION';
    my $crit = $self->{Crit} = ($QUERY->criteria('GEOGRAPHIC_REGION'))[0];
    my $geo = $QUERY->geo_region;   
    
    # Get URLs, etc:
    my $edit_url   = script_url('mode'    =>$self->{Mode},
				'submode' =>'GOTO',
				'goto'    =>$cat);
    my $delete_url = script_url('mode'    =>$self->{Mode},
				'submode' =>'DELETECAT',
				'category'=>$cat);
    
    ###
    ### ROW 1, COLUMN 1...
    ###
    $HTML -> TR(VALIGN=>'TOP');
    $HTML -> TD(WIDTH=>'50%');

    # Icon:
    #$HTML -> A(HREF=>$edit_url);
    #$HTML -> IMG(SRC=>$self->icon, ALT=>"Edit", BORDER=>0, 
    #             ALIGN=>'LEFT', HSPACE=>4);
    #$HTML -> _A;
    
    # Name:
    $HTML -> B -> t(category2pretty($cat)) -> _B -> BR;
    
    # Start nonbreaking region (for buttons):
    $HTML -> tag('NOBR');
    
    # Edit:
    $HTML -> button(SRC=>"${IU::DocsUrl}btns/tiny-edit.gif",
		    ALT=>"Edit",
		    HREF=>$edit_url);
    # Delete:
    $HTML -> button(SRC=>"${IU::DocsUrl}btns/tiny-delete.gif",
		    ALT=>"Del",
		    HREF=>$delete_url);
    # Help:
    my $context = IU::Category->lookup($cat)->{Context};
    $HTML -> help_btn($context)     if $context;
    
    # End nonbreaking region (for buttons):
    $HTML -> tag('_NOBR');
    $HTML-> _TD;

    # Are there criteria? 
    if ($crit) {      # Yes, there are criteria!

	###
	### ROW 1-2, COLUMN 2...
	###
	# Map:
	$HTML -> TD(WIDTH=>'50%',ROWSPAN=>2, ALIGN=>'CENTER',VALIGN=>'MIDDLE',
		    BGCOLOR=>'#FFFFFF');
	my $map = $self->map_src;
	if ($map) {
	    $HTML -> button(SRC=>$map, 
                            ALT=>'Map showing current geographic selection', 
                            HREF=>$edit_url);
	}
	else {
	    $HTML -> t("MAP NOT AVAILABLE");
	}
	$HTML -> _TD -> _TR;

	###
	### ROW 2, COLUMN 1...
	###
	$HTML -> TR(VALIGN=>'TOP') -> TD(WIDTH=>'50%');
	$HTML -> small;       

	# Actual criterion:
	print join("<BR>", 
		   split /\n/, html_escape($crit->as_string(Verbose=>2)));

	# Approximation, if any:
	my $app = $self->{ApproxGeo};
	if ($app) {
	    $HTML -> P -> t_B("If necessary, it will be approximated by...") 
		  -> BR;
	    print join("<BR>", 
		       split /\n/, html_escape($app->as_string(Verbose=>2)));
	}
	$HTML -> _small;
	$HTML -> _TD;
    }
    else {            # No, no criteria!
	
	###
	### ROW 1, COLUMN 2...
	###

	$HTML -> TD(WIDTH=>'50%');
	$HTML -> t("(Geographic region of interest, default is no selection.)"); 
    }

    $HTML -> _TR -> _TABLE;    
}

#------------------------------------------------------------
1;



