package IW::Widget::ViewBrowse;

=head1 NAME

IW::Widget::ViewBrowse - show a browse product which is in cache

=head1 NOTES

This has been made into a widget so that it may be shared between
modes.

=cut

use IU;
use IU::Log;
use IU::FileCache;
use IU::Util qw(touch path_filename);

use IW::App;
use IW::Util qw(:url);
use strict;

#------------------------------------------------------------
# cached_files GKEY
#------------------------------------------------------------
# I<Class/instance method, utility.>
# Is the given granule in cache?
# If so, returns array of file objects.  Otherwise, returns empty array.
#
# Does an actual cache operation.

sub cached_files {
    my ($class, $gkey) = @_;

    # Make session scratch area, if needed:
    my $scratch;
    (-d ($scratch = $SESSION->path_to('browse')))
	or mkdir $scratch, 0755 
	    or puke "can't mkdir $scratch: $!";
    
    # Open cache:
    my $cache = IU::FileCache->open(Dir=>"${IU::DocsDir}/imswww-usr/cache")
	|| puke "can't open data granule file cache!";

    # Get files, cached under this granule key:
    my $cacheentry = $cache->lookup($gkey);
    if ($cacheentry) {
	inform "got ", $gkey, " from cache!";
	return $cacheentry->paths;
    }

    # Nope!
    return ();
}

#------------------------------------------------------------
# browse_url CACHEFILE
#------------------------------------------------------------
# I<Function.>  Return URL for a browse file.
#
sub browse_url {
    my $file = shift;

    # Extract the filename:
    my $filename = path_filename($file);
    doc_url("imswww-usr/cache/$filename");
}

#------------------------------------------------------------
# new 
#------------------------------------------------------------
# Create a new widget.  Params are:
#
#      Granule => a granule object, whose browse files should be
#                 in cache if you want them displayed. 
#
sub new {
    my ($class, %params) = @_;

    # Create the object:
    my $self = bless \%params, $class;
    $self->{Granule} or die "missing data granule!";

    # Load the browse files (if any):
    $self->{Files} = [$self->cached_files($self->{Granule}->key)];

    # Done! 
    $self;
}

#------------------------------------------------------------
# files 
#------------------------------------------------------------
# Return the array of loaded cache files, or empty if none.

sub files {
    @{$_[0]->{Files} || []}
}

#------------------------------------------------------------
# print_inline
#------------------------------------------------------------
# Output a presentation of a browse image to the standard HTML stream.
# The object will remember which file it "displayed".
#
sub print_inline {
    my $self = shift;
    my $gran = $self->{Granule};
    
    # The first browsefile is the HDF.
    my ($HDF, @GIFs) = $self->files;

    # The last one is the labels, unless this is hanging around from before.
    # If so, we have no labels file, and we just skip label pocessing.
    my $giflabels = pop @GIFs;
   
    my @labels = ();
    if ($giflabels =~ /\.lbl$/) { 
        debug "Assuming label file: $giflabels";
        open LABELS, "<$giflabels" 
            or puke "Can't open labels file $giflabels: $!";
        while (<LABELS>) {
            push @labels,$_;
        }
    }
   
    my $browsefile;
    my @new_list = grep  { /\.jpg$/ } @GIFs;
    my @new_palettes = grep  { /\.pal\./ } @GIFs;

    # As of 3.6, EDG now displays ALL JPG images and their palettes. JSA
    # Output:
    $HTML -> CENTER;
    if (int(@new_list) > 0 ) {
        foreach my $single_file (@new_list) {
        my $label = (shift @labels) || "Legend unavailable for this graphic";
   
        # Use the label as the ALT tag, lists the name of the DSET the image
        # is from.
	$HTML -> IMG(SRC=>browse_url($single_file), ALT=>"$label");
        my $single_palette = shift @new_palettes;

        # IE needs these breaks so they can display the images verticially.
        print '<br>';
        $HTML -> IMG(SRC=>browse_url($single_palette), ALT=>"Palette from HDF file") if (defined($single_palette));
        print '<br>';
        }
    }
    else {
	$HTML -> TABLE(BORDER=>2, BGCOLOR=>'#FFFFFF') -> TR -> TD;
	print <<EOF;
              <I>There are no images available in the
	      browse data. (This may be completely 
	      normal for the type of data you've selected.)
EOF
        $HTML -> _TD -> _TR -> _TABLE;
    }
    $HTML -> _CENTER;
    1;
}

#------------------------------------------------------------
# print_palette
#------------------------------------------------------------
# Output a presentation of color palette used for  browse image to the 
# standard HTML stream.
# The object will remember which file "palette".
#
sub print_palette {
    my $self = shift;

    # The first browsefile is the HDF: pop it:
    my ($HDF, @GIFs) = $self->files;

    # Pick the first palette file, and output the image in-line:
    my $palettefile;
    foreach $palettefile (@GIFs) {
        # Make sure we're NOT using a Metadata file!
        if ($palettefile =~ /pal/) {
          (-s $palettefile) and ($self->{Palette} = $palettefile) and last;
        }
    }

    # Output:
    if ($self->{Palette}) {
        $HTML -> BR;
        $HTML -> CENTER;
        $HTML -> IMG(SRC=>browse_url($self->{Palette}), 
                     ALT=>"Palette from HDF file");
        $HTML -> _CENTER;
    }
    1;
}

#------------------------------------------------------------
# print_menu
#------------------------------------------------------------
# Output the table of all browse files available.
# Now includes Metadata as well -JSA 6-98

sub print_menu {
   my $self = shift;
   my ($HDF, @image_types) = $self->files;

   # Each will hold a list of filenames :
   my (@no_gifs , @no_jpegs, @no_palette, @no_metadata, $no_columns);

   # Holds the file size ( in bytes)
   my $size_of_file;

   # Throwaways :
   my $i;

   # Subset the array image_types into jpegs, gifs and metadata.
   # Store'em in their own seperate arrays.
   #
   # all gifs include gif images and gif palettes.
   my @all_gifs    = grep { $_ =~ /gif/ } @image_types;
   # weed out the gif palettes and only keep gif images.
   @no_gifs     = grep !/\.pal/, @all_gifs;
   @no_jpegs    = grep { $_ =~ /jpg/ } @image_types;
   @no_palette  = grep { $_ =~ /pal/ } @image_types;
   @no_metadata = grep { $_ =~ /des/ } @image_types;

   # @image_types can be nulled out.
   @image_types = (); 
 
   # So what do we have? Give us a column per entry type.
   $no_columns += 1 if ( int(@no_gifs) > 0);
   $no_columns += 1 if ( int(@no_jpegs) > 0);
   $no_columns += 1 if ( int(@no_palette) > 0);
   $no_columns += 1 if ( int(@no_metadata) > 0);

   # Display the compressed HDF file:
   $HTML -> TABLE;
   $HTML -> TR -> TD;
   $HTML -> hypertext("Download browse HDF_file: ", browse_url($HDF));
   $HTML -> TD;
      my $compressed = (($HDF =~ /\.Z$/) ? 'Unix-compressed' : 'non-compressed');
   $HTML -> t("A $compressed ");
   $HTML -> hypertext("HDF", $IU::HdfInfoUrl);
   $HTML -> t(" file containing all the browse products. "); 
   $HTML -> t(" Have your browser ");
   $HTML -> t_EM("save this directly to a file");
   $HTML -> t(" - you won't be able to display it; it's not a simple image - ");
   $HTML -> t("and then use your favorite HDF toolkit to extract "); 
   $HTML -> t("information from it.");
   $HTML -> _TABLE;

   # Order of browse data being displayed: Jpeg, Gifs, and Metadata
   # Layout of an entry :
   # an Icon (labeled JPEG/GIF) , the Image number x+1, the file size, and if 
   # it's the same image as being displayed at the top.

   # Create a table, 2/3x2
   $HTML -> TABLE(WIDTH=>'100%', BORDER=>'2', COLS=>"$no_columns");

   # First Row, Header : GIF/Palette/Metadata
   # JPEG now displayed in v3.6
   $HTML -> TR;
   #   $HTML ->TD ->t_B('JPEGs') -> _TD if ( int(@no_jpegs) > 0);
   $HTML ->TD -> t_B('GIFs') -> _TD if ( int(@no_gifs) > 0);
   $HTML ->TD -> t_B('Palette') -> _TD if ( int(@no_palette) > 0);
   $HTML -> TD -> t_B ('Metadata') -> _TD if ( int(@no_metadata) > 0);

   # Second row, the actual data.
   $HTML -> _TR;

   # Change the color of the table to 'Gray'
   $HTML -> TR (BGCOLOR=>'#D0D0D0');

# As of 3.6 we display all Jpegs, below is the code that would display the
# Jpegs in the table as well

=head1 NAME

    # Display the JPEGS if you've got'em :
    if ( int(@no_jpegs) > 0) {
      # Make sure everything is "aligned" to the TOP of the table.
      $HTML -> TD(VALIGN=>'TOP');
      for ($i = 0; $i < @no_jpegs; $i++) {
         $HTML-> hypertext("Display JPEG image ", browse_url($no_jpegs[$i]));

        # Get the file size
        $size_of_file = ( -s $no_jpegs[$i] );

        # Display the number and file size
        $HTML->BR->t("Image #",($i+1))
             -> BR
             ->t_EM("  Size ",$size_of_file) -> t_B(" bytes")->BR;

        # Display a Note if this is the same image as the one being displayed
        #  already
        if ($self->{Displayed} && ($no_jpegs[$i] eq $self->{Displayed})) {
              print (" <i>(same as image shown above)</i>");
        }
        print ("<BR> \n");
      }
      $HTML -> _TD;
    }

=cut  

    # Display the GIFS if you've got'em :
    if ( int(@no_gifs) > 0) {
      # Make sure everything is "aligned" to the TOP of the table.
      $HTML -> TD(VALIGN=>'TOP');
      for ($i = 0; $i < @no_gifs; $i++) {
         $HTML -> hypertext("Display GIF image ", browse_url($no_gifs[$i])); 

        # Get the file size
        $size_of_file = ( -s $no_gifs[$i] );

        # Display the number and file size
        $HTML->BR->t("Image #",($i+1))
             -> BR
             ->t_EM("  Size ",$size_of_file) -> t_B(" bytes")->BR;

        # Display a Note if this is the same image as the one being displayed
        #  already
        if ($self->{Displayed} && ($no_gifs[$i] eq $self->{Displayed})) {
            print " <i>(same as image shown above)</i>";
        }
        print " <BR> \n";
      }
      $HTML -> _TD;
    }

    # Display the PALETTEs if you've got'em :
    if ( int(@no_palette) > 0) {
       $HTML -> TD (VALIGN=>'TOP');
       for ($i = 0; $i < @no_palette; $i++) {
       $HTML -> hypertext("Display Palette ", browse_url($no_palette[$i])); 

         # Get the file size 
         $size_of_file = ( -s $no_palette[$i] );

         # Display the number and file size
         $HTML->BR->t("Palette ",($i+1))
              -> BR
              ->t_EM("  Size ",$size_of_file) -> t_B(" bytes");

         print " <BR> \n";
       }
       $HTML -> _TD;
    } 

    # Display the METADATA if you've got'em :
    if ( int(@no_metadata) > 0) {
      $HTML -> TD (VALIGN=>'TOP');
       for ($i = 0; $i < @no_metadata; $i++) {
          $HTML-> hypertext("Display Metadata ", browse_url($no_metadata[$i]));

         # Get the file size
         $size_of_file = ( -s $no_metadata[$i] );

         # Display the number and file size
         $HTML->BR->t("Metadata ",($i+1))
              -> BR
              ->t_EM("  Size ",$size_of_file) -> t_B(" bytes");
         print " <BR> \n";
       }
       $HTML -> _TD;
    }

   $HTML -> _TR;
   $HTML -> _TABLE;

   # Force an access on the inline GIF, even if WWW browser doesn't get it:
   if ($self->{Displayed}) {
	touch($self->{Displayed}, "-a") or error("touch: $!");   # okay to fail
   }

   1;
}

#------------------------------------------------------------
# print_selection
#------------------------------------------------------------
# Output a table with a single selection Radio button, a single browse image
# and its metadata. The user selects one image via the radio button, and
# submits the form to proceed to geoselection.
sub print_selection {
   my $self = shift;
   my %params = @_;
   my $display_geo_selection = $params{DisplayGeoSelect};

   my ($HDF, @image_types) = $self->files;

   # Each will hold a list of filenames :
   my (@no_gifs , @no_jpegs, @no_metadata, $no_columns);

   # Holds the file size ( in bytes)
   my $size_of_file;

   # Throwaways :
   my $i;

   # Subset the array image_types into jpegs, gifs and metadata.
   # Store'em in their own seperate arrays.
   @no_gifs     = grep { $_ =~ /gif/ } @image_types;
   @no_jpegs    = grep { $_ =~ /jpg/ } @image_types;
   @no_metadata = grep { $_ =~ /des/ } @image_types;

   # @image_types can be nulled out.
   @image_types = ();

   # So what do we have? Give us a column per entry type.
   $no_columns += 1;
   $no_columns += 1 if ( int(@no_gifs) > 0);
   $no_columns += 1 if ( int(@no_metadata) > 0);

   $HTML -> TABLE(WIDTH=>'30%', BORDER=>'2', COLS=>"$no_columns");
     # First Row, Header : JPEG/GIF/Metadata
     $HTML -> TR;
     $HTML ->TD ->t_B('Select One') -> _TD;
     $HTML ->TD -> t_B('GIFs') -> _TD if ( int(@no_gifs) > 0);
     $HTML -> TD -> t_B ('Metadata') -> _TD if ( int(@no_metadata) > 0);

     # Second row, the actual data.
     $HTML -> _TR;

     # Change the color of the table to 'Gray'
     $HTML -> script_FORM;
     $HTML -> msubmit_flag;

     # Display the GIFS if you've got'em :
     if ( int(@no_gifs) > 0) {
       my $href;
       # Make sure everything is "aligned" to the TOP of the table.
       $HTML -> TD(VALIGN=>'TOP');
       for ($i = 1; $i < @no_gifs; $i = $i+2) {

          $HTML -> TR  (BGCOLOR=>'#D0D0D0');
          $HTML -> TD(VALIGN=>'CENTER');
          $HTML->radio_button(Name=>"p.browse_location", # "pk-choice.1"
                                  Choice=>$no_gifs[$i] # "1.2"
                                  );

          $HTML -> _TD;
          $HTML -> TD(VALIGN=>'CENTER');
          my @browse_list = (split ('/',$no_gifs[$i]));
          my $browse_filename = pop @browse_list;

          my @metadata = ();
          if ( int(@no_metadata) >= 0) {
             my $single_meta_filename = shift @no_metadata;
             open (METAFILE,$single_meta_filename);
             while (<METAFILE>) {
                chomp;
                push @metadata, $_;
             }
             close (METAFILE);
          }
          my $counter = $i;
          $HTML -> IMG(SRC=>browse_url($no_gifs[$i]), ALT=>"Image $counter; Description: @metadata");
          $HTML -> BR;
          $HTML -> IMG(SRC=>browse_url($no_gifs[$i-1]), ALT=>"Palette $counter");
          print " <BR> \n";
          $HTML -> _TD;

          # Display metadata if any.
          if ( int(@no_metadata) >= 0) {
             $HTML-> TD (VALIGN=>'TOP') ->PRE;
             map  {$HTML->t("$_\n")} @metadata;
             $HTML -> _PRE -> _TD;
          }
       }
    }
    # Display the METADATA if you've got'em :
    $HTML -> _TR;
    $HTML -> _TABLE;
    $HTML -> FONT(SIZE=>'+2') -> B
            -> msubmit_button(CAPTION => 'Accept My Selection!',
                              ALT => "Accept my Selection!",
                     QUERY => ['mode' => 'AIRS',
                                        'p.geoopt'=>'Equatorial Map',
                                        'category' => 'GEOGRAPHIC_REGION',
                                        'displayGeoSelect' => $display_geo_selection,
                                        'p.geotype' => 'RANGE_LOC',
                                        'p.state' => 'START',
                                    'p.mapname-next'=>'Equatorial Map']);
   $HTML -> _FONT;
   $HTML -> _B;
   $HTML -> _script_FORM;
}

#------------------------------------------------------------
1;
