#!/usr/bin/perl -w
package OHTTPD;

=head1 NAME

OHTTPD - object-oriented HTTPD base

=head1 SYNOPSIS

    # Here's a simple invocation:
    use OHTTPD;
    use ...;             # all your modules
    
    OHTTPD->become_a_daemon;             # always!
    my $HTTPD = new OHTTPD Port=>8123    or die "couldn't start!"; 
    $HTTPD->open_log                     or die "couldn't log!";
    $HTTPD->wipe_environ;
    close(STDIN);
    close(STDOUT);
    $HTTPD->set_action(\&myaction);
    $HTTPD->serve;      

=cut

require 5.002;

use IU::FileHandle;
use Carp;
use Socket;
use Exporter;
use POSIX qw(:sys_wait_h setsid);

use sigtrap;                    # safe way to trap misc fatal signals
use strict qw(refs subs);
use vars qw($VERSION);


#------------------------------------------------------------
#
# Globals
#
#------------------------------------------------------------

# RCS info:
$Revision = '$Id: OHTTPD.pm,v 1.14 2001/04/17 16:55:22 joe Exp $';
($VERSION = $Revision) =~ s/.*(\d+\.\d+).*/$1/;
my $Version  = $Revision; $Version =~ s/.*(\d+\.\d+).*/$1/;

# How hard to try to fork:
my $MaxForkTries = 5;

# Define low default timeout (in fractional minutes) and sensible range limits
my $DefaultTimeout = 2;
my $MinTimeout     = 1;
my $MaxTimeout     = 60 * 2;

# Server protocol:
my $Protocol = 'HTTP/1.0';

# Map numbers to messages:
my %ExplainMsg = 
    (
     200 => 'OK',
     201 => 'Created',
     202 => 'Accepted',
     204 => 'No Content',
     
     301 => 'Moved Permanently',
     302 => 'Moved Temporarily',
     304 => 'Not Modified',
     
     400 => 'Bad Request',
     401 => 'Unauthorized',
     403 => 'Forbidden',
     404 => 'Not Found',
     405 => 'Method Not Allowed',      # really for CGI scripts only!
     
     500 => 'Server Error',
     501 => 'Not Implemented',
     502 => 'Bad Gateway',
     503 => 'Service Unavailable'
     );

# Log entry mapping:
my %LogGlyph = 
    (
     DEBUG => 'D   ',
     INFO  => 'II  ',
     WARN  => 'WWW ',
     ERROR => 'EEEE',
     );

#------------------------------------------------------------
#
# Utilities
#
#------------------------------------------------------------

#------------------------------------------------------------
# accept_timeout
#------------------------------------------------------------
# Signal handler.
# WARNING: don't stdio from signal handlers (but malloc's not much better)!

sub accept_timeout {
   die "Interrupted\n";	# see await_connect()
}

#------------------------------------------------------------
# _new_inet_socket SOCKET[,PORT]
#------------------------------------------------------------
sub _new_inet_socket {
   my ($FH, $port) = @_;
   $port or $port = 0;       # if no port, system will assign one
   
   # Create the socket:
   socket($FH, PF_INET, SOCK_STREAM, (getprotobyname('tcp'))[2])
       or confess "socket(tcp): $!";
   
   # This allows us to rebind immediately without error:
   setsockopt($FH, SOL_SOCKET, SO_REUSEADDR, 1) 
       or confess "setsockopt REUSEADDR: $!";
   
   # Apache does this... maybe to detect a client that's gone bye-bye?
   setsockopt($FH, SOL_SOCKET, SO_KEEPALIVE, 1) 
       or confess "setsockopt KEEPALIVE: $!";
   
   # Bind it to the port:
   my $inaddr = pack('S n C4 x8', AF_INET, $port, 0,0,0,0);
   unless (bind($FH, $inaddr)) {
      my $err = "bind: port $port: $!";
      close($FH);
      confess $err;
   }
   
   # Retrieve the assigned port, and return it:
   return ((unpack('S n C4 x8',getsockname($FH)))[1]);
}

#------------------------------------------------------------
# _new_unix_socket SOCKET,PORTPATH
#------------------------------------------------------------
sub _new_unix_socket {
   my ($FH, $portpath) = @_;
   
   # Create the socket:
   socket($FH, AF_UNIX, SOCK_STREAM, 0) or confess "socket($portpath): $!";
   my $inaddr = pack('S a', AF_INET, $portpath);
   
   # Bind it to the port path:
   unless (bind($FH, $inaddr)) {
      my $err = "bind: $!";
      close($FH);
      confess $err;
   }
   $portpath;
}



#------------------------------------------------------------
#
# OO interface, private
#
#------------------------------------------------------------

#------------------------------------------------------------
# DESTROY
#------------------------------------------------------------
sub DESTROY {
   my $self = shift;
   close $self->{ACCEPT} if $self->{ACCEPT};
   $self->debug(1, ref($self). " terminating");
}

#------------------------------------------------------------
# await_connect
#------------------------------------------------------------
# Await the next connection
# If timeout > 0, is used: else, no timeout
# Set up the client info.

sub await_connect {
   my($self, $timeout) = @_;
   my $CLIENTaddr;
   $timeout or $timeout = 0;
   
   # Accept connection, catching any timeouts:
   $self->debug(1,"Awaiting connect on port $self->{Port}, timeout $timeout");
   if ($timeout) {
      local($SIG{'ALRM'}) = \&accept_timeout; # local for auto reset
      alarm($timeout * 60);
      $CLIENTaddr = accept(CLIENT, $self->{ACCEPT}); 
      alarm(0);
   }
   else {
      $CLIENTaddr = accept(CLIENT, $self->{ACCEPT}); 
   }
   
   # Handle failure:
   unless ($CLIENTaddr){
      my $err = "accept() failed: ". ( $@ ? $@ : $!);
      $err .= ' (timed out?)' if ($timeout && ($err =~ m/Interrupted/i));
      $self->error($err);
      return undef;
   }
   
   # Success! Client now contains an accepted connection!
   $self->{CLIENT} = \*CLIENT;
   $self->{CLIENTaddr} = $CLIENTaddr;
   1;
}

#------------------------------------------------------------
# disconnect
#------------------------------------------------------------
sub disconnect {
   my $self = shift;
   close $self->{CLIENT} if $self->{CLIENT};
   $self->{CLIENT} = undef;
   $self->{CLIENTaddr} = undef;
   1;
}

#------------------------------------------------------------
# done [YESNO]
#------------------------------------------------------------
# Are we done?  Can also be used to set it.
sub done {
   my ($self, $yesno) = shift;
   $self->{Done} = $yesno if defined($yesno);
   $self->{Done};
}

#------------------------------------------------------------
# fork
#------------------------------------------------------------
# Try hard to fork.  
# If we can't, don't die: just log it and return undef
sub fork {
   my $self = shift;
   
   my $pid;
   my $tries = 0;
   while (!defined($pid = fork) || ($pid < 0)) {
      $self->error("fork (#", ++$tries, ") failed: $!");
      return undef if $tries > $MaxForkTries;
      sleep 1;
   }
   $pid;
}

#------------------------------------------------------------
# log MESSAGE
#------------------------------------------------------------
# Log message.  Put CR after it.

sub log {
   my $self = shift;
   my $class = shift;
   
   my $time = scalar(localtime);
   $time =~ s|^(\S+)\s+(\S+)\s+(\d+)\s+(\S+)\s+(\S+)$|$3/$2/$5:$4|;
   $time = '' if $self->{NoTime};
   seek(STDERR,0,2);    
   my $glyph = $LogGlyph{uc($class)} || '????';
   print STDERR "$glyph [$time $$] ", @_, "\n";
}

#------------------------------------------------------------
# read_header 
#------------------------------------------------------------
# Read header from STDIN (the client).
# Sets up the Header and URI instance variables, and the envirnoment.
# Also makes note of the fact that we did it... or at least, tried to!

sub read_header {
   my $self = shift;
   my $uri;
   
   # Make note (mostly for 'sorry'):
   $self->{TriedReadingHeader} = 1;
   
   # Read first line, skipping blank lines:  KLUDGE! ONLY SKIP 2 THEN GIVE UP!
   my $line;
   my $skipped = 0;
   while (defined($line = <STDIN>)) {
      last if ($line !~ m/^\s+$/);
      die "400: too many leading blank lines" if (++$skipped > 2);
   }
   
   # Log it, if debugging:
   $self->debug(2,"Request-Line: $line");
   
   # Read the request line.  This might be old (HTTP/0.9) or new (HTTP/1.0)!
   my $prot;
   ($ENV{REQUEST_METHOD}, $uri, $prot) = split(/\s+/, $line);
   defined($uri) or $uri = '';
   $self->{RequestProtocol} = uc($prot || '');
   
   # Extract information from the URI:
   my ($script, $query);
   ($script, $query) = split(/\?/, $uri, 2);
   $ENV{SCRIPT_NAME}  = $script if defined($script);
   $ENV{QUERY_STRING} = $query  if defined($query);
   $self->{URI} = $uri;    
   
   # If *implicit* HTTP/0.9, don't read header... else, read it!
   if (!$self->{RequestProtocol}) {
      $self->{Header} = {};	
   }
   else {                # read RFC-822-style header...
      my %head;
      my ($key, $value);
      while (<STDIN>) {
         s/\s+$//;		# remove trailing whitespace
         last if $_ eq '';	# end of headers
         $self->debug(3,"Header: <$_>");
         
         # Handle:
         if (/^(\S+):\s*/) {     # header line...
            ($key, $value) = (lc($1), $');
            $head{$key} or $head{$key} = [];
            push @{$head{$key}}, $value;
         }
         elsif (/^(\s)/) {       # continuation line...
            next if (!$key);
            $head{$key}[-1] .= "$1$'";
         }
         else {                  # log error, but KEEP GOING... 
            $self->error("Header syntax: <$_>");
         }
      }
      
      # Save it!
      $self->{Header} = \%head; 
   }
   
   # Setup envirnoment:
   $self->setup_env;
   1;
}

#------------------------------------------------------------
# setup_env
#------------------------------------------------------------
# Setup environment from header and other information.
#
# UNFINISHED: if option is set, could turn all unrecognized client
# headers into env. vars.  However, since they're nonstandard, this
# would be ill advised.  Do that AFTER all other env vars have been set, 
# and ONLY assign to env vars that haven't been set yet!!!!

sub setup_env {
   my $self = shift;
   
   # Basic stuff:
   $ENV{QUERY_STRING}      = ''         unless $ENV{QUERY_STRING};
   $ENV{GATEWAY_INTERFACE} = 'CGI/1.1';
   $ENV{SCRIPT_FILENAME}   = $0;
   
   # Who we are:
   $ENV{SERVER_NAME}       = $self->{Hostname};
   $ENV{SERVER_PORT}       = $self->{Port};
   $ENV{SERVER_PROTOCOL}   = ($self->{RequestProtocol} || 'HTTP/0.9');
   $ENV{SERVER_SOFTWARE}   = $self->software;
   
   # Who we're talking to:
   my ($r_family, $r_port, @r_addr)= unpack("S n C4 x8", $self->{CLIENTaddr});
   my $r_dotquad = join('.', @r_addr);
   my $r_host    = (gethostbyaddr(pack("C4",@r_addr), AF_INET))[0];
   $ENV{REMOTE_ADDR}  = $r_dotquad;
   $ENV{REMOTE_HOST}  = $r_host if $r_host;
   $self->debug(1,"Connection from ", $r_host||$r_dotquad,
                " ($r_dotquad:$r_port)");
   
   # Extract information from header:
   my $val;
   if ($self->header('Accept')) {
      $ENV{HTTP_ACCEPT} = join(', ', $self->header('Accept'));
   }
   if ($val = $self->header('User-Agent')) {
      $ENV{HTTP_USER_AGENT} = $val;
   }
   if ($val = $self->header('Referer')) {
      $ENV{HTTP_REFERER} = $val;
   } 
   if ($val = $self->header('Content-length')) {
      $ENV{CONTENT_LENGTH} = $val;
   } 
   if ($val = $self->header('Content-type')) {
      $ENV{CONTENT_TYPE} = $val;
   }
   if ($val = $self->header('Authorization')) {
      $self->authinfo($val);
      $ENV{AUTH_TYPE} = (split($val))[1];
   }
   
   # Set up special vars, if debugging:
   if ($self->{Debug}) {
      $ENV{OHTTPD_STARTED} = localtime($self->{Started}) . ' (local time)';
      my $uptime = time - $self->{Started};
      $ENV{OHTTPD_UPTIME} = int($uptime/60) . " mins " . 
          ($uptime%60) . " secs";
      $ENV{OHTTPD_STARTED} = localtime($self->{Started});
   }
   1;
}



#------------------------------------------------------------
#
# OO interface, basic
#
#------------------------------------------------------------

#------------------------------------------------------------
# become_a_daemon
#------------------------------------------------------------
# I<Class method>.  Does all the setup necessary for becoming a 
# generic daemon.  Thus:
#
#   * Signal handlers are set here to catch anything that might kill us.
#   * The CWD is changed to "/".
#   * The daemon is disassociated from a process group.
#   * The umask is cleared.
#   * The environment is cleared.
#
# Algorithm from Steven's "UNIX Network Programming" and "Advanced Programming
# in the UNIX Environment".
#
# WARNING: notice that it's useless to log our pid to a file until this
# method completes, because of the fork!!!

sub become_a_daemon {
   
   # Ignore terminal stop signals (BSD):
   $SIG{TTOU} = 'IGNORE';
   $SIG{TTIN} = 'IGNORE';
   $SIG{TSTP} = 'IGNORE';
   
   # If we were not started in the background, fork and let the parent exit.
   # This also guarantees that first child isn't a process group leader.
   # Note: it is normal for the parent to *temporarily* become a zombie!
   my $have_child = CORE::fork();
   defined($have_child) or die "fork: $!";
   exit 0 if ($have_child);             # parent goes away
   
   # First child process.  Become a session leader!  This does 3 things:
   #   * Disassociates from controlling terminal
   #   * Process becomes group leader of new process group
   setsid();               # new in POSIX.1 ?   
   
   # Ignore SIGHUP:
   $SIG{HUP} = 'IGNORE';
   
   # Change to root directory, to be nice for system shutdown:
   chdir("/") or die "chdir: $!";
   
   # Clear inherited umask:
   umask(0);
   
   # *** DO NOT UNCOMMENT THESE LINES! It will cause zombies!
   # Ignore SIGCLD portably (waitpid is in POSIX.1):
   # $SIG{CLD} = sub {
   #   while (waitpid(-1,&WNOHANG) > 0) { };
   # };
   1;
}

#------------------------------------------------------------
# new PARAMHASH
#------------------------------------------------------------
# Params are...
#
#     Hostname  => if given, the name of this host
#     Port      => if given, the port to listen at
#     PortPath  => if given, use local Unix socket
#     MaxListen => listen queue... defaults to 5
#     LogDir    => logging directory (default is /tmp)
#     Force     => kill any existing ohttpd if true
#
# WARNING: the creation of the pid file is done here, since it
# seemed the best way to avoid accidents without resorting to an 
# elaborate flock-based system.
#
# Multiple HTTPDs with the same log directory are illegal, but this
# is currently not detected automatically.  :-(

sub new {			# should only ever be called once
   my ($class, %params) = @_;
   my $port      = $params{Port} || 0;
   my $portpath  = $params{PortPath};
   my $maxlisten = defined($params{MaxListen}) ? $params{MaxListen} : 5;
   my $force     = $params{Force};
   my $hostname  = $params{Hostname};
   
   # Get an new object:
   my $self = {};
   bless $self, $class;
   
   # Set up basic stuff:
   $self->{LogDir} = ($params{LogDir} || '/tmp');
   $self->{Debug} = 0;
   
   # Create a socket for accepting connections:
   my $ACCEPT = IU::FileHandle->new;	# new unopened file handle ref
   
   # Create a new socket, based on the port path:
   $self->{Port} = '';
   my $retrying = 0;
 RETRY:
   eval {
      if ($portpath) {   # file system / non-network / private socket:
         _new_unix_socket($ACCEPT, $portpath) or confess $!;
      } 
      else {              # internet / public socket:
         $self->{Port} = _new_inet_socket($ACCEPT, $port) or confess $!;
      }
   };
   if ($@) {         # failed!
      # Are any other HTTPDs signed in?
      my $other = $self->signed_in;
      
      # Muscle in, if we failed because of a failed "bind":
      if (($@ =~ /^bind: /i) && !$retrying && $other && $force) {
         warn "killing existing ohttpd (pid $other)";
         kill 'KILL', $other;
         sleep 3;                  # give it time to die
         $retrying = 1; goto RETRY;
      }
      die "$@ (and can't retry)";       # give up.
   }
   
   # We're here!
   $self->sign_in or die "couldn't write pid!";
   
   # Set the listen queue:
   listen($ACCEPT, $maxlisten) or confess "listen($maxlisten): $!";
   $self->{Timeout} = $DefaultTimeout;
   $self->{ACCEPT}  = $ACCEPT;
   
   # What is our FULL hostname?
   unless ($self->{Hostname} = $hostname) {      # if not given explicitly...
      $self->{Hostname} = (gethostent())[0] 
          or die "can't get my own hostname: $?: $!";
      endhostent();
   }
   
   # Other info:
   $self->{Started} = time;
   $self;
}

#------------------------------------------------------------
# action
#------------------------------------------------------------
# Run the installed action, sending it the client as its first arg,
# and any installed args as its remaining args.  Does nothing if
# no action was installed.
#
# You can override this method in a subclass, if you like.

sub action {
   my $self = shift;
   $self->{Action} or die "500: no action installed!";
   my $args = $self->{ActionArgs} || [];
   &{$self->{Action}}($self, @$args);
}

#------------------------------------------------------------
# header [FIELD]
#------------------------------------------------------------
# In array context, returns array of values for that field, or () if none.
# In scalar context, returns the first value, or undef if none.
# If no FIELD is given, returns the known header fields.

sub header {
   my ($self, $field) = @_;
   
   if (@_ > 1) {      # FIELD given
      my $valsref = ($self->{Header} ? $self->{Header}{lc($field)} : undef);
      $valsref or $valsref = [];
      return (wantarray ? @{$valsref} : $valsref->[0]);
   }
   else {
      keys %{$self->{Header}};
   }
}

#------------------------------------------------------------
# software
#------------------------------------------------------------
# Identify yourself, no?  This MUST be of the form "NAME/VERSION"!
# Note that this subroutine should remain callable as either a
# method or a straight subroutine; it is used in both ways by
# different parts of IMS/WWW.
sub software {
   "OHTTPD/$Version";
}

#------------------------------------------------------------
# date
#------------------------------------------------------------
# The current date in HTTP standard format, i.e., RFC 1123
# "Sun, 06 Nov 1994 08:49:37 GMT".
# Note that this subroutine should remain callable as either a
# method or a straight subroutine; it is used in both ways by
# different parts of IMS/WWW.
sub date {
   ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = gmtime(time);
   join(" ", (qw(Sun Mon Tue Wed Thu Fri Sat))[$wday] . ",",
        ($mday < 10 ? "0$mday" : $mday) ,
        (qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec))[$mon],
        $year+1900,
        ($hour < 10 ? "0$hour" : $hour),
        ($min  < 10 ? "0$min"  : $min),
        ($sec  < 10 ? "0$sec"  : $sec),
        "GMT"
        );
}

#------------------------------------------------------------
# authinfo
#------------------------------------------------------------
# Get/set information from AUTHORIZATION headers. Used to 
# make this info available to subprocesses, but not to leave
# it lying around in (easily-visible) environment variables.
sub authinfo {
   my ($self,$authinfo) = @_;
   my $oldinfo = (defined($self->{AuthInfo}) ? $self->{AuthInfo} : undef);
   $self->{AuthInfo} = $authinfo if $authinfo;
   return $oldinfo;
}

#------------------------------------------------------------
# host
#------------------------------------------------------------
# Get our fully-qualified host.domain name.

sub host {
   $_[0]->{Hostname};
}

#------------------------------------------------------------
# port
#------------------------------------------------------------
# Get the port we listen at.

sub port {
   $_[0]->{Port};
}

#------------------------------------------------------------
# pid
#------------------------------------------------------------
# Get the process id of the HTTPD instance that's currently managing
# serve() loop (note that we don't set this value until then).  
# This is intended for child processes.

sub pid {
   $_[0]->{Pid};
}


#------------------------------------------------------------
# set_action SUBREF,[@ARGS]
#------------------------------------------------------------
# Set the action to perform on a connect.  
# Unless overridden in a subclass, the action() method will
# call the SUBREF, passing it the HTTPD and then the ARGS.
# 
#          THIS:                      DOES THIS:
#
#      $self->action      --->   &$SUBREF($self, @args)
#
# If you don't like callbacks, you can also inherit from this class 
# and override action() to do what you need.  

sub set_action {
   my ($self, $action, @args) = @_;
   $self->{Action}     = $action if $action;
   $self->{ActionArgs} = \@args;
   $self->{Action}
}

#------------------------------------------------------------
# sorry MSGNO,MESSAGE...
#------------------------------------------------------------
# Read client input, then a quick apology to the client, before a disconnect.
# Logs it as an error as well.
#
# WARNING: DON'T EVER EVER EVER CALL THIS FROM THE PARENT HTTPD! 
# For one thing, it won't work: i/o goes through STDIN/STDOUT.
#
# A MSGNO of zero means to try and extract it from MESSAGE, if MESSAGE
# begins with /\d{3}: /.
# 
# We have to gobble the client's input; otherwise, they get a "Broken pipe" 
# error trying to send us a command.
#
# Returns undef in scalar context, () in array context.

sub sorry {
   my ($self, $msgno, @msg) = @_;
   my $msg = join('',@msg);
   my $hmsg;
   my %map = ('"' =>"quot", '<'=>'lt', '>'=>'gt', '&'=>'amp');
   
   # Log:
   $self->error($msgno, ' ', $msg); 
   $self->debug(2, 'about to apologize');
   
   # Extract msgno, if needed:
   if (!$msgno && ($msg =~ /^(\d{3}): /)) {
      ($msgno, $msg) = ($1, $');
   }
   $msgno = sprintf("%03d", ($msgno || 500));        # default and clean up
   
   # Fix up message, and convert to HTML:
   ($hmsg = $msg) =~ s/([\<\>\&\"])/\&$map{$1};/g;   # convert ", etc. 
   $hmsg =~ s/(\s+at)( \S+ line \d+.*)/<BR><B>Stopped at:<\/B> $2/;
   
   # KLUDGE: suck up their header, if needed:
   unless ($self->{TriedReadingHeader}) {
      $self->debug(3,'eating header');
      eval {
         local $SIG{ALRM} = { 1 };
         alarm(3);
         while (<STDIN>) { last if /\S/ };         # skip until nonblank
         while (<STDIN>) { last if /^\s*/};        # skip until blank
      };
      alarm(0);
      $self->debug(1,'ate header');
   }
   $self->{TriedReadingHeader} = 1;
   
   
   # Output reply header, *except* if request is an *implicit* HTTP/0.9!:
   if ($self->{RequestProtocol}) {         # HTTP/1.0 style!
      print "$Protocol $msgno $ExplainMsg{$msgno}\r\n"; 
      print "Server: ", $self->software, "\r\n";
      print "Content-type: text/html\r\n";
      print "\r\n";
   }
   
   # Output reply document:          UNFINISHED! THIS SHOULD BE OVERRIDABLE!
   my $explain = ($ExplainMsg{$msgno} || 'Application Error');
   print "<TITLE>", $self->software, " Error $msgno</TITLE>\n";
   print "<H1>", $self->software, " Error</H1>";
   print "<H2>$msgno $explain</H2>";
   print "<P><B>Details:</B>  ", $hmsg;
   print <<EOF;
   <P>The HTTP server suffered an error or a misconfiguration,
   and was unable to complete your request.  For 400-series errors, you 
       might want to wait a few seconds, and hit 'reload'; if the problem 
           persists, please contact the system administrator at this site 
               immediately (<CODE>root\@$self->{Hostname}</CODE>).
EOF
                       STDOUT->flush;
   
   # Close connection now!
   open STDIN, "</dev/null";    
   open STDOUT, ">/dev/null";
   wantarray ? () : undef;
}

#------------------------------------------------------------
# serve_client
#------------------------------------------------------------
# I<Internal>.  We're the child of the httpd (after the first fork).
# Spawn another child (the grandchild) to serve the client, and then die.

sub serve_client {
   my $self = shift;
   
   $self->debug(2, "Serving client");
   
   # STDIN/STDOUT should talk through socket to client:
   open STDIN, "<&CLIENT" or die "open STDIN: $!";
   open STDOUT, ">&CLIENT" or die "open STDOUT: $!";
   
   # Fork again!
   $self->debug(2, "Forking grandchild:...");
   my $cpid = $self->fork;
   defined($cpid) or die "503: slave fork failed: $!";
   exit(0) if $cpid;               # exit immediately if not the grandchild
   
   # At this point, we're the grandchild:
   $self->ps_display("slave");
   
   # Turn signals back on:
   $SIG{CLD} = 'DEFAULT';     # or else shell cmds won't work!
   
   # Read the header:
   $self->read_header or die "400: I can't understand your HTTP header";
   
   # Do the action, passing it any args we gave it!
   $self->debug(2, "Grandchild: about to do action");
   my $status = $self->action(@{$self->{ActionArgs} || []});
   
   # Cleanup and exit:
   $self->debug(2,"Grandchild: about to flush STDIO and exit");
   STDOUT->flush;
   exit ($status ? 0 : -1);
}

#------------------------------------------------------------
# serve
#------------------------------------------------------------
# Serve the socket.  Every time a connection comes in,
# a grandchild is forked and the action() method is invoked.
#
# NOTE: NEVER EVER EVER DO ANYTHING ELABORATE WITH THE CLIENT 
# UNTIL AFTER YOU FORK!  DON'T READ, DON'T WRITE, DON'T PLAY GAMES!
# THAT ALSO MEANS, DON'T CALL sorry()!!!!!

sub serve {
   my ($self) = @_;
   my $conn_count = 0;
   
   # Keep going until someone tells us to stop:
   $self->{Pid} = $$;           # forked children get access to info!
   $self->inform("Entering service loop...");
   $self->done(0);
   until ($self->done) {
      
      # Accept next connection:
      ++$conn_count;
      $self->ps_display("awaiting client $conn_count");
      eval { $self->await_connect } or next;
      my $client = $self->{CLIENT};
      
      # We've got a client!
      $self->ps_display("connected to client $conn_count");
      $self->{TriedReadingHeader} = 0;          # important: must reset!
      
      # Fork:
      my $pid = $self->fork;
      if (!defined($pid)) {   # can't fork; fall back into waiting
         $self->error("fork failed: $!");
         goto DISCONNECT;                      # do NOT apologize!!!!
      }
      if (!$pid) {            # we're the child...
         eval { $self->serve_client };         # serve client
         $@ and $self->sorry(0, $@);           # apologize if we failed
         exit 0;
      }	
      
      # Wait for child to die:
      $self->ps_display("spawning slave for client $conn_count");
      wait(); 
      
      # Child is dead, so we can close client on our end: 
    DISCONNECT:
      $self->disconnect;
   }
   
   # Done!
   $self->inform("Exiting!");
   1;
}

#------------------------------------------------------------
# wipe_environ
#------------------------------------------------------------
# Wipe out the entire envirnoment, and substitute a simple one.
# We have found that we can't wipe out the timezone without
# forcing all of the localtime() calls to suddenly become GMT,
# so we I<don't> wipe TZ.
sub wipe_environ {
   # Save the current timezone setting to keep localtime() sane.
   my $oldtz = $ENV{TZ};
   
   %ENV = ();
   
   $ENV{PATH} = "/bin:/usr/bin";
   $ENV{SHELL} = '/bin/sh'; 
   
   # Restore the timezone.
   $ENV{TZ} = $oldtz;
}



#------------------------------------------------------------
#
# Logging...
#
#------------------------------------------------------------

#------------------------------------------------------------
# log_dir
#------------------------------------------------------------
sub log_dir {
   $_[0]->{LogDir};
}

#------------------------------------------------------------
# log_file NAME
#------------------------------------------------------------
# Return path to the given file.
sub log_file {
   $_[0]->{LogDir} . ($_[1] ? "/$_[1]" : '');
}



#------------------------------------------------------------
# open_log [LOGFILE]
#------------------------------------------------------------
# Redirect STDERR to the given file, which is opened for appending.
# If no logfile given, defaults to "ohttpd.err" in the LogDir.

sub open_log {
   my $self = shift;
   my $logfile = shift || "$self->{LogDir}/ohttpd.err";
   STDERR->autoflush(1);
   open STDERR, ">>$logfile" or die "open: $!";
   $self->{LogFile} = $logfile;
   $self->{Logging} = 1;
}

#------------------------------------------------------------
# debugging
#------------------------------------------------------------
sub debugging {
   my ($self, $level) = @_;
   $self->{Debug} = $level || 0;
}

#------------------------------------------------------------
# debug  LEVEL, MESSAGE
#------------------------------------------------------------
# Debug levels: 1=suspect, 2=normal, 3=FMTEYEWTK.
# If debug level is set to 3+, all messages are echoed in "ps" area

sub debug { 
   my $self = shift;
   $self->{Debug} >= shift or return undef;
   my $msg = join('',@_);
   
   $self->log('debug',$msg);                    # log 
   $0 = "(ohttpd $self->{Port}: debug: $msg)"   # ps-display it
       if ($self->{Debug} >= 3);
}

#------------------------------------------------------------
# inform  MESSAGE
# warning MESSAGE
# error   MESSAGE
#------------------------------------------------------------
sub inform  { shift->log('info', @_) }
sub warning { shift->log('warn', @_) }
sub error   { shift->log('error', @_) }

#------------------------------------------------------------
# is_logging 
#------------------------------------------------------------
sub is_logging {
   $_[0]->{Logging};
}

#------------------------------------------------------------
# ps_display MESSAGE
#------------------------------------------------------------
# Advertise process state in 'ps' area.  Also logged as a debug message.
# This is how you advertise "normal" changes in state.

sub ps_display {
   my ($self, $message) = @_;
   $self->debug(2, "state: $message");
   $0 = "(ohttpd $self->{Port}: $message)";
}


#------------------------------------------------------------
#
# PID stuff... 
#
#------------------------------------------------------------

#------------------------------------------------------------
# sign_in
#------------------------------------------------------------
# Write our pid to "LogDir/ohttpd.pid".
# If the bind fails, and "Force" is set true, this will be used to 
# kill the process.

sub sign_in {
   my $self = shift;
   my $pidfile = "$self->{LogDir}/ohttpd.pid";
   open PID, ">$pidfile" or die "open $pidfile: $!";
   print PID $$, "\n";
   close PID;
   1;
}

#------------------------------------------------------------
# signed_in
#------------------------------------------------------------
# Is someone currently signed in?  Returns their pid if yes, false if not.

sub signed_in {
   my $self = shift;
   my $pidfile = shift || "$self->{LogDir}/ohttpd.pid";
   
   open PID, "<$pidfile" or return undef;
   my $pid = <PID>;
   close PID;
   $pid =~ s/\s//g;    # get rid of all whitespace
   $pid;
}

#------------------------------------------------------------
1;

