#!/usr/bin/perl -w
package PHTTPD;

=head1 NAME

PHTTPD - object-oriented HTTPD base

=head1 SYNOPSIS

    # Here's a simple invocation:
    use PHTTPD;

     

=cut

require 5.002;

use FileHandle;
use Carp;

use strict qw(refs subs);
use vars qw($VERSION);


#------------------------------------------------------------
#
# Globals
#
#------------------------------------------------------------

# RCS info:
$Revision = '$Id: PHTTPD.pm,v 1.2 1997/10/06 14:57:35 eryq Exp $';
($VERSION = $Revision) =~ s/.*(\d+\.\d+).*/$1/;
my $Version  = $Revision; $Version =~ s/.*(\d+\.\d+).*/$1/;

# How hard to try to fork:
my $MaxForkTries = 5;

# Define low default timeout (in fractional minutes) and sensible range limits
my $DefaultTimeout = 2;
my $MinTimeout     = 1;
my $MaxTimeout     = 60 * 2;

# Server protocol:
my $Protocol = 'HTTP/1.0';

# Map numbers to messages:
my %ExplainMsg = 
    (
     200 => 'OK',
     201 => 'Created',
     202 => 'Accepted',
     204 => 'No Content',

     301 => 'Moved Permanently',
     302 => 'Moved Temporarily',
     304 => 'Not Modified',

     400 => 'Bad Request',
     401 => 'Unauthorized',
     403 => 'Forbidden',
     404 => 'Not Found',
     405 => 'Method Not Allowed',      # really for CGI scripts only!

     500 => 'Server Error',
     501 => 'Not Implemented',
     502 => 'Bad Gateway',
     503 => 'Service Unavailable'
);

# Log entry mapping:
my %LogGlyph = 
    (
     DEBUG => 'D   ',
     INFO  => 'II  ',
     WARN  => 'WWW ',
     ERROR => 'EEEE',
     );


#------------------------------------------------------------
#
# OO interface, private
#
#------------------------------------------------------------

#------------------------------------------------------------
# DESTROY
#------------------------------------------------------------
sub DESTROY {
    my $self = shift;
    $self->debug(1, ref($self). " terminating");
}


#------------------------------
#
# Logging...
#
#------------------------------

#------------------------------------------------------------
# log MESSAGE
#------------------------------------------------------------
# Log message.  Put CR after it.

sub log {
    my $self = shift;
    my $class = shift;

    my $time = scalar(localtime);
    $time =~ s|^(\S+)\s+(\S+)\s+(\d+)\s+(\S+)\s+(\S+)$|$3/$2/$5:$4|;
    $time = '' if $self->{NoTime};
    seek(STDERR,0,2);    
    my $glyph = $LogGlyph{uc($class)} || '????';
    print STDERR "$glyph [$time $$] ", @_, "\n";
}
sub inform  { shift->log('info',  @_) }
sub warning { shift->log('warn',  @_) }
sub error   { shift->log('error', @_) }

#------------------------------------------------------------
# debugging
#------------------------------------------------------------
sub debugging {
    my ($self, $level) = @_;
    $self->{Debug} = $level || 0;
}

#------------------------------------------------------------
# debug  LEVEL, MESSAGE
#------------------------------------------------------------
# Debug levels: 1=suspect, 2=normal, 3=FMTEYEWTK.
# If debug level is set to 3+, all messages are echoed in "ps" area

sub debug { 
    my $self = shift;
    $self->{Debug} >= shift or return undef;
    my $msg = join('',@_);

    $self->log('debug',$msg);                    # log 
    $0 = "(ohttpd $self->{Port}: debug: $msg)"   # ps-display it
	if ($self->{Debug} >= 3);
}

#------------------------------------------------------------
# read_header 
#------------------------------------------------------------
# Read header from STDIN (the client).
# Sets up the Header and URI instance variables, and the envirnoment.
# Also makes note of the fact that we did it... or at least, tried to!

sub read_header {
    my $self = shift;
    my $uri;

    # Make note (mostly for 'sorry'):
    $self->{TriedReadingHeader} = 1;

    # Read first line, skipping blank lines:  KLUDGE! ONLY SKIP 2 THEN GIVE UP!
    my $line;
    my $skipped = 0;
    while (defined($line = <STDIN>)) {
	last if ($line !~ m/^\s+$/);
	die "400: too many leading blank lines" if (++$skipped > 2);
    }

    # Log it, if debugging:
    $self->debug(2,"Request-Line: $line");

    # Read the request line.  This might be old (HTTP/0.9) or new (HTTP/1.0)!
    my $prot;
    ($ENV{REQUEST_METHOD}, $uri, $prot) = split(/\s+/, $line);
    defined($uri) or $uri = '';
    $self->{RequestProtocol} = uc($prot || '');

    # Extract information from the URI:
    my ($script, $query);
    ($script, $query) = split(/\?/, $uri, 2);
    $ENV{SCRIPT_NAME}  = $script if defined($script);
    $ENV{QUERY_STRING} = $query  if defined($query);
    $self->{URI} = $uri;    

    # If *implicit* HTTP/0.9, don't read header... else, read it!
    if (!$self->{RequestProtocol}) {
	$self->{Header} = {};	
    }
    else {                # read RFC-822-style header...
	my %head;
	my ($key, $value);
	while (<STDIN>) {
	    s/\s+$//;		# remove trailing whitespace
	    last if $_ eq '';	# end of headers
	    $self->debug(3,"Header: <$_>");
	    
	    # Handle:
	    if (/^(\S+):\s*/) {     # header line...
		($key, $value) = (lc($1), $');
		$head{$key} or $head{$key} = [];
		push @{$head{$key}}, $value;
	    }
	    elsif (/^(\s)/) {       # continuation line...
		next if (!$key);
		$head{$key}[-1] .= "$1$'";
	    }
	    else {                  # log error, but KEEP GOING... 
		$self->error("Header syntax: <$_>");
	    }
	}

	# Save it!
	$self->{Header} = \%head; 
    }
    
    # Setup envirnoment:
    $self->setup_env;
    1;
}

#------------------------------------------------------------
# setup_env
#------------------------------------------------------------
# Setup environment from header and other information.
#
# UNFINISHED: if option is set, could turn all unrecognized client
# headers into env. vars.  However, since they're nonstandard, this
# would be ill advised.  Do that AFTER all other env vars have been set, 
# and ONLY assign to env vars that haven't been set yet!!!!

sub setup_env {
    my $self = shift;

    # Basic stuff:
    $ENV{QUERY_STRING}      = ''         unless $ENV{QUERY_STRING};
    $ENV{GATEWAY_INTERFACE} = 'CGI/1.1';
    $ENV{SCRIPT_FILENAME}   = $0;

    # Who we are:
    $ENV{SERVER_NAME}       = $self->{Hostname};
    $ENV{SERVER_PORT}       = $self->{Port};
    $ENV{SERVER_PROTOCOL}   = ($self->{RequestProtocol} || 'HTTP/0.9');
    $ENV{SERVER_SOFTWARE}   = $self->software;

    # Extract information from header:
    my $val;
    if ($self->header('Accept')) {
	$ENV{HTTP_ACCEPT} = join(', ', $self->header('Accept'));
    }
    if ($val = $self->header('User-Agent')) {
	$ENV{HTTP_USER_AGENT} = $val;
    }
    if ($val = $self->header('Referer')) {
	$ENV{HTTP_REFERER} = $val;
    } 
    if ($val = $self->header('Content-length')) {
	$ENV{CONTENT_LENGTH} = $val;
    } 
    if ($val = $self->header('Content-type')) {
	$ENV{CONTENT_TYPE} = $val;
    }

    # Set up special vars, if debugging:
    if ($self->{Debug}) {
	$ENV{OHTTPD_STARTED} = localtime($self->{Started}) . ' (local time)';
	my $uptime = time - $self->{Started};
	$ENV{OHTTPD_UPTIME} = int($uptime/60) . " mins " . 
	                         ($uptime%60) . " secs";
	$ENV{OHTTPD_STARTED} = localtime($self->{Started});
    }
    1;
}



#------------------------------------------------------------
#
# OO interface, basic
#
#------------------------------------------------------------

#------------------------------------------------------------
# new PARAMHASH
#------------------------------------------------------------
# Constructor.

sub new {			# should only ever be called once
    my ($class, %params) = @_;

    # Get an new object:
    my $self = \%params;
    bless $self, $class;

    # What is our FULL hostname?
    unless ($self->{Hostname}) {      # if not given explicitly...
	while ($_ = gethostent()) {
	    next if /^localhost$/;      # must be a better one
	    $self->{Hostname} = $_; last;
	}
	endhostent();
    }
    $self->{Hostname} or die "can't get my own hostname: $?: $!";

    # Other info:
    $self->{Started} = time;
    $self;
}

#------------------------------------------------------------
# action
#------------------------------------------------------------
# Run the installed action, sending it the client as its first arg,
# and any installed args as its remaining args.  Does nothing if
# no action was installed.
#
# You can override this method in a subclass, if you like.

sub action {
    my $self = shift;
    $self->{Action} or die "500: no action installed!";
    my $args = $self->{ActionArgs} || [];
    &{$self->{Action}}($self, @$args);
}

#------------------------------------------------------------
# header [FIELD]
#------------------------------------------------------------
# In array context, returns array of values for that field, or () if none.
# In scalar context, returns the first value, or undef if none.
# If no FIELD is given, returns the known header fields.

sub header {
    my ($self, $field) = @_;
    
    if (@_ > 1) {      # FIELD given
	my $valsref = ($self->{Header} ? $self->{Header}{lc($field)} : undef);
	$valsref or $valsref = [];
	return (wantarray ? @{$valsref} : $valsref->[0]);
    }
    else {
	keys %{$self->{Header}};
    }
}

#------------------------------------------------------------
# software
#------------------------------------------------------------
# Identify yourself, no?  This MUST be of the form "NAME/VERSION"!
sub software {
    "PHTTPD/$Version";
}

#------------------------------------------------------------
# host
#------------------------------------------------------------
# Get our fully-qualified host.domain name.

sub host {
    $_[0]->{Hostname};
}

#------------------------------------------------------------
# port
#------------------------------------------------------------
# Get the port we listen at.

sub port {
    $_[0]->{Port};
}

#------------------------------------------------------------
# sorry MSGNO,MESSAGE...
#------------------------------------------------------------
# Read client input, then a quick apology to the client, before a disconnect.
# Logs it as an error as well.
#
# WARNING: DON'T EVER EVER EVER CALL THIS FROM THE PARENT HTTPD! 
# For one thing, it won't work: i/o goes through STDIN/STDOUT.
#
# A MSGNO of zero means to try and extract it from MESSAGE, if MESSAGE
# begins with /\d{3}: /.
# 
# We have to gobble the client's input; otherwise, they get a "Broken pipe" 
# error trying to send us a command.
#
# Returns undef in scalar context, () in array context.

sub sorry {
    my ($self, $msgno, @msg) = @_;
    my $msg = join('',@msg);
    my $hmsg;
    my %map = ('"' =>"quot", '<'=>'lt', '>'=>'gt', '&'=>'amp');

    # Log:
    $self->error($msgno, ' ', $msg); 
    $self->debug(2, 'about to apologize');

    # Extract msgno, if needed:
    if (!$msgno && ($msg =~ /^(\d{3}): /)) {
      ($msgno, $msg) = ($1, $');
    }
    $msgno = sprintf("%03d", ($msgno || 500));        # default and clean up

    # Fix up message, and convert to HTML:
    ($hmsg = $msg) =~ s/([\<\>\&\"])/\&$map{$1};/g;   # convert ", etc. 
    $hmsg =~ s/(\s+at)( \S+ line \d+.*)/<BR><B>Stopped at:<\/B> $2/;

    # KLUDGE: suck up their header, if needed:
    unless ($self->{TriedReadingHeader}) {
	$self->debug(3,'eating header');
	eval {
	    local $SIG{ALRM} = { 1 };
	    alarm(3);
	    while (<STDIN>) { last if /\S/ };         # skip until nonblank
	    while (<STDIN>) { last if /^\s*/};        # skip until blank
	};
	alarm(0);
	$self->debug(1,'ate header');
    }
    $self->{TriedReadingHeader} = 1;

    # Output reply header, *except* if request is an *implicit* HTTP/0.9!:
    if ($self->{RequestProtocol}) {         # HTTP/1.0 style!
	print "$Protocol $msgno $ExplainMsg{$msgno}\r\n"; 
	print "Server: ", $self->software, "\r\n";
	print "Content-type: text/html\r\n";
	print "\r\n";
    }

    # Output reply document:          UNFINISHED! THIS SHOULD BE OVERRIDABLE!
    my $explain = ($ExplainMsg{$msgno} || 'Application Error');
    print "<TITLE>", $self->software, " Error $msgno</TITLE>\n";
    print "<H1>", $self->software, " Error</H1>";
    print "<H2>$msgno $explain</H2>";
    print "<P><B>Details:</B>  ", $hmsg;
    print <<EOF;
	<P>The HTTP server suffered an error or a misconfiguration,
        and was unable to complete your request.  
         
        <P>For '404 Not Found' errors, please double-check the URL
        you requested.  For other 400-series errors, you 
        might want to wait a few seconds, and hit 'reload'; if the problem 
        persists, please contact the system administrator at this site 
        immediately (<CODE>root\@$self->{Hostname}</CODE>).
EOF
    STDOUT->flush;

    # Close connection now!
    open STDIN, "</dev/null";    
    open STDOUT, ">/dev/null";
    wantarray ? () : undef;
}

#------------------------------------------------------------
# wipe_environ
#------------------------------------------------------------
# Wipe out the entire envirnoment, and substitute a simple one.
# We leave certain environment values alone, since they may be
# coming from the HTTPD code.

sub wipe_environ {
    my $self = shift;

    foreach (keys %ENV) {
	next if /^(REMOTE_HOST|REMOTE_ADDR)$/;
	delete $ENV{$_};
    }
    $ENV{PATH} = "/bin:/usr/bin";
    $ENV{SHELL} = '/bin/sh';    
}

#------------------------------------------------------------
1;

