
=head1 NAME

imsrunclient - run the "imsclient" program

=head1 SYNOPSIS

Command-line usage is identical to imsclient; all arguments 
are passed in verbatim.

=head1 DESCRIPTION

Run the C<imsclient> program.  If necessary, covert the output to a
form more readily-usable by the IMS/WWW system.

Set up an INT and TERM signal handler that will kill the "imsclient" 
child before dying itself.

=head2 The "results.gdb" file

This creates the F<results.gdb> file in the session directory, 
which basically contains the information from the F<dataset.gdb> 
and F<granule.gdb> files.  Keys are of the following form:

      dskey:DSKEY      A dataset object, with dataset key DSKEY
      gkey:GKEY        A granule object, with granule key GKEY
      pk2gkeys:PKKEY   A package-key-to-granule-keys mapping
      ds2gkeys:DSKEY   A dataset-key-to-granule-keys mapping

=head2 The "runclient.stat" file

As search results are processed, the current status is written to
the file F<runclient.stat> in the session directory.  The file
is updated atomically.

=head2 The "runclient.log" file

STDERR is redirected to the log file, F<runclient.log>, in the
session directory.

=head2 Environment variables

The following environment variables are used explicitly by this code:

    IW_SESSION_ID      the session id

=head1 NOTES

The ingest of the packages is the least straightforward, and
is worthy of some explanation:

Basically, in the results, a pkkey comes in one of three forms: 

    CASE "GRAN": archtag/dstag/*/gid
        The package equal to THIS SINGLE granule.

    CASE "DSET": archtag/dstag/*
        The pseudo-package of ALL granules of this dataset.
        What this says is that all granules in this dataset may be
        ordered individually.        

    CASE "PKID": archtag/dstag/pkid
        This package.

(WG 2.1):
Use grandchild for imsclient so that grandparent can check
whether it is still running by sending the 0 signal to
the grandchild. (Using kill 0)(Couldn't use child because
its process exists as a zombie until a wait is done.  The
grandchild doesn't because its parent is unix init.)

Assumes the maximum process ID has less than 15 characters.

Status now (WG 2.1) contains the pid of imsrunclient so that if
user starts a new search and kills the running search,
the pid can be used to kill imsrunclient (instead of
just imsclient).  The pid can also be used (by ModeSEARCH
for instance) to determine if the whole client process is
really done or running.

We now (WG 2.1) delete granule.gdb and dataset.gdb when client
finishes.  All WG code now gets info from results.gdb.  Because
of this, we also no longer need "merged" versions.

WG 2.1 does granule, inventory, directory searches in a
subdirectory called "search".  As the results.gdb is generated,
it is renamed to the session directory.  In future versions,
if there are multiple searches, you will want to change this
to leave it in the subdirectory, but then you'll also have to
change about 10 other modules where results.gdb is referenced!

=head1 AUTHOR

Eryq (Erik Dorfman), Hughes STX Corporation
Modified for partial results viewing: Larry Rosen, Raytheon STX.

=head1 RCS INFO

$Id: imsrunclient.pm,v 1.29 2002/11/07 20:21:07 lorena Exp $

=cut

use IU;
use IU::Log;
use IU::FileLogger;
use IU::Session;
use IU::ClientChunks;
use IU::Dataset;
use IU::Event;
use IU::Granule;
use IU::Chunk;
use IU::InvStore;
use IU::ClientChunks;
use IU::WWWValids;

use File::Copy 'cp';
use FileHandle ();
use GDBM_File;

use strict;

#------------------------------------------------------------
#
# Globals
#
#------------------------------------------------------------

# Current action:
my $Action;

# Subdirectory to run client in.  May be blank.
my $RunDir = '';

# Pid of grandchild imsclient process.  Global for signal handler.
my $GrandChildPid = undef;

# Session:
my $SESSION;

# Stop (kill) our client if any of these signals are delivered:
my @StopSigs = qw( INT TERM );

# Current status:
my %Status = ();

# Last number of data sets received (in the trimmed client.obj).
# It's -1 to ensure that we trigger a database creation even when no
# granules show up.
my $NumDataSets = -1;

# Global of full-path filename of client.obj (results) file.
# Set in init() so it only gets set once.
my $Client_obj;

# Environment setup for ims_event.
my %environment;

#------------------------------------------------------------
# sig_stop_client 
#------------------------------------------------------------
# Signal handler for SIGINT, SIGTERM.
# We are about to die.  Stop the imsclient child process!

sub sig_stop_client {

   # Handle:
   kill 9, $GrandChildPid if defined ($GrandChildPid);

   inform 'Interrupted imsclient operation.';
   ims_event('SEARCH_KILLED',\%environment);

   exit 1;

   # Reset:  (not needed since exiting!)
   #    foreach (@StopSigs) { $SIG{$_} = \&sig_stop_client }
}

#------------------------------------------------------------
# print_status PARAMHASH
#------------------------------------------------------------
# Output the "runclient.stat" status document which describes 
# the post-processing phases, and is used in generating the
# comm status screen.  
#
# The file is similar in syntax to the "client.stat"
# file that the imsclient generates.

sub print_status {
   my %params = @_;
   my $key;

   # Tweak status:
   $Status{LED_STAGE} ||= 0;
   $Status{NUMBER_OF_GRANULES} ||= 0;

   # Create new status file:
   my $current = $SESSION->path_to("$RunDir".'runclient.stat');
   my $new     = $SESSION->path_to("$RunDir".'runclient.stat.new');
   open(STATUS, ">$new") or return undef;
   foreach $key (sort keys %Status) {
      next if $key =~ /^_/;          # skip private info ("_xxx")
      print STATUS "$key: $Status{$key}\n";
   }
   print STATUS "\n";
   close(STATUS);

   # Rename:
   rename $new => $current;
}

#------------------------------------------------------------
# update_status PARAMHASH
#------------------------------------------------------------
# Update the "runclient.stat" status data structure which describes 
# the post-processing phases, and is used in generating the
# comm status screen.  
#
# Params are:
#
#     LED_STAGE           New value changes
#     LED_STAGE_FORMAT    Format for the comment, where %d means num 
#                         datasets and %g means num granules, and %r
#                         means runtime.
#     NUMBER_OF_DATASETS  '++' to increment, new value changes.
#     NUMBER_OF_GRANULES  '++' to increment, new value changes.
#     Reset               Reset all counts to 0.  Other params ignored.
#     Flush               Force a flush to file.

sub update_status {
   my %params = @_;
   $params{LED_STAGE} ||= 0;
   my $ngrans = $params{NUMBER_OF_GRANULES}; 
   my $ndsets = $params{NUMBER_OF_DATASETS};

   # Reset, if desired:
   if ($params{Reset}) {
      %Status = (PID                =>$$,
                 LED_STAGE          =>0,
                 NUMBER_OF_DATASETS =>0,
                 NUMBER_OF_GRANULES =>0,
                 _INGEST_START      =>time
                 );
      return 1;
   }
   # Update status:
   foreach (qw(LED_STAGE NUMBER_OF_GRANULES NUMBER_OF_DATASETS)) {
      defined($params{$_}) or next;
      if   ($params{$_} eq '++') { ++$Status{$_} }
      else                       { $Status{$_} = $params{$_} }
   }
   # Make sure there's a comment:
   $params{LED_STAGE_FORMAT} or $params{LED_STAGE_FORMAT} =
       "Converted %d data sets, %g granules";
   $Status{LED_STAGE_COMMENT} = $params{LED_STAGE_FORMAT};
   $Status{LED_STAGE_COMMENT} =~ s/%d/$Status{NUMBER_OF_DATASETS}/g;
   $Status{LED_STAGE_COMMENT} =~ s/%g/$Status{NUMBER_OF_GRANULES}/g;
   $Status{LED_STAGE_COMMENT} =~ 
       s/%r/time - $Status{'_INGEST_START'}/eg;

   # Flush changes to disk?
   if ($params{Flush} or
       (!(($Status{NUMBER_OF_DATASETS} +
           $Status{NUMBER_OF_GRANULES}) % 20))){
      print_status();
   }
}

#------------------------------------------------------------
# add_gran_info_to_dataset 
#------------------------------------------------------------
# Build up dataset summary info, using one of the dataset's granules.
#
# In many cases, there is info we want to have at the dataset level,
# but it doesn't exist there: for example, start and stop times.
# What we do is to take all granules returned under that dataset,
# and build up the info from those granules.

sub add_gran_info_to_dataset {
   my ($gran, $dset) = @_;
   my $value;

   # Tally granule:
   $dset->n_granules($dset->n_granules + 1);

   # Start and stop dates:
   if (!($dset->start_date) ||
       ($gran->start_date lt $dset->start_date)) {
      $dset->start_date($gran->start_date);
   }
   if (!($dset->stop_date) ||
       ($gran->stop_date gt $dset->stop_date)) {
      $dset->stop_date($gran->stop_date);
   }
   # Add strings or string sets from granule into dataset string sets:
   $dset->campaigns(  [@{$dset->campaigns},    $gran->campaign]);
   $dset->parameters( [@{$dset->parameters}, @{$gran->parameters}]);
   $dset->proc_levels([@{$dset->proc_levels},  $gran->proc_level]);
   $dset->sensors(    [@{$dset->sensors},    @{$gran->sensors}]);
   $dset->sources(    [@{$dset->sources},    @{$gran->sources}]);
   1;
}

#------------------------------------------------------------
# read_objects OBJPATH
#------------------------------------------------------------
# Read in the .obj file OBJPATH of dataset/granule information,
# and convert it to IU::Dataset/IU::Granule objects.
#
# Returns a reference to a hash of the objects.
# A granule with key GKEY is hashed under "gkey:GKEY".
# A dataset with key DSKEY is hashed under "dskey:DSKEY".

sub read_objects {
   my $objpath = shift;
   my $chunk;   # the current chunk
   my @chunks;  # chunks we got
   my %objs;    # dataset/granule objects we got, hashed appropriately
   my $exp;     # hash of expanded chunks:
   my $key;

   # Read and queue chunks:    
   debug "reading raw chunks from $objpath";
   open(OBJ, "<$objpath") or
       return error "couldn't read $objpath: $!";
   ims_event('reading_chunks',\%environment);
   while ($chunk = input IU::Chunk \*OBJ) {
      push @chunks, $chunk;
   }
   close OBJ;
   ims_event('chunks_read',\%environment,Num=>(scalar @chunks));

   # Expand them:
   debug "expanding ", int(@chunks), " raw chunks";
   $exp = IU::ClientChunks->expand_chunks(\@chunks);
   ims_event('chunks_expanded',\%environment,Num=>(scalar keys %$exp));

   # Convert them:
   debug "converting ", int(keys %$exp), " expanded chunks";
   foreach $key (keys %$exp) {
      $chunk = $exp->{$key};

      # Convert it:
      if ($chunk->value('X_TYPE') eq 'DATASET') {
         my $dset = IU::Dataset->from_chunk($chunk);
         $objs{'dskey:' . $dset->key} = $dset;
      }
      elsif ($chunk->value('X_TYPE') eq 'GRANULE') {
         my $gran = IU::Granule->from_chunk($chunk);
         $objs{'gkey:' . $gran->key} = $gran;
      }
      elsif ($chunk->value('X_TYPE') eq 'MEDIA_HANDLING_FEE') {
         my $media = IU::MediaHandlingFee->from_chunk($chunk);
         $objs{'media:' . $media->type} = $media;
      }
      elsif ($chunk->value('X_TYPE') eq 'MEDIA_MAXIMUM_CAPACITY') {
         my $mediasize = IU::MediaMaximumCapacity->from_chunk($chunk);
         $objs{'mediasize:' . $mediasize->type} = $mediasize;
      }
      else {
         # no-op
      }
      # Get rid of the chunk, for efficiency:
      delete $exp->{$key};
   }
   # Done!
   ims_event('chunks_converted',\%environment,Num=>(scalar keys %objs));
   return \%objs;
}

#------------------------------------------------------------
# trim_results
#------------------------------------------------------------
# Trim the results file to after last data set group.
# Parameter: Full-path/filename to trim.
# Returns: Number of data set groups found.
# Side Effect: Replaces original file with a trimmed one.
# Method: Read line at a time into a list.
#         Check each line for begin with "END DATASET".
#         If found, write list to temp file.
#         When in last group, there's no END DATASET,
#         so the partial group isn't written.
#         Append "END DIRECTORY_RESULT" or "END INVENTORY_RESULT".
#         Rename temp file back to input file.
#         Return the number of data set groups found.

sub trim_results {
   my $file = shift;
   my $tempfile = "$file" . ".temp";
   open (IN, "<$file") or return undef;
   unless (open (TEMP, ">$tempfile") ) {
      close IN;
      return undef;
   }
   my @list;    # List of lines read, including \n's.
   my $type;    # Result type: INVENTORY or DIRECTORY.

   # Read into list until first BEGIN group to find RESULT type.
   while (<IN>) {
      # Add input line to list.
      push @list, $_;

      # Check for RESULT type (in first BEGIN).
      if (/^BEGIN (\w+)/) {
         $type = $1;          # $type is like "INVENTORY_RESULT\n".
         last;
      }
   }
   # Now have type and 1st list or not in input file yet.
   # If have type, write the list.
   if (defined ($type)) {
      print TEMP @list;      # Print the 1st list.
      @list = ();            # Empty the list.
      my $num_groups = 0;    # Number of data set groups found.

      # Loop over all the input lines.
      while  (<IN>) {
         # Add input line to list.
         push @list, $_;

         # If a data set group ended, print the group.
         if (/^\s*END DATASET\n$/) {
            print TEMP @list;
            @list = ();
            $num_groups++;
         }
      }
      # Done.  If we got any complete groups:
      if ($num_groups) {
         # Append the END *_RESULT line.
         print TEMP "END $type\n\n";
         close TEMP;
         close IN;

         # Rename the trimmed file back to the original.
         rename $tempfile, $file;

         # Return the number of groups found, indicating
         # that the file should be processed.
         return $num_groups;
      }
   }
   close TEMP;
   close IN;

   # Input had no type (client may have crashed), or had no
   # complete groups (yet).  Delete the temp file.
   unlink $tempfile;

   # Return 0 to indicate NO processing of file is needed.
   return 0;
}

#------------------------------------------------------------
# ingest_search_results
#------------------------------------------------------------
# We just did an inventory/directory search, and we need to
# convert the results from the "dataset.gdb/granule.gdb" files
# to the "results.gdb" database.

sub ingest_search_results {
   my $dskey;         # dataset key
   my $dset;          # dataset object
   my $gkey;          # granule key
   my $gran;          # granule object
   my $okey;          # object key
   my %pk2gkeys;      # $pk2gkeys{$pkkey} = array of gkeys
   my %ds2gkeys;      # $ds2gkeys{$dskey} = array of gkeys
   my $pkkey;

   my $trimfile = $SESSION->path_to ("$RunDir".'trimmed_client.obj');
   # For partial results, copy the input file so we can edit it.
   cp $Client_obj, $trimfile or return 1;

   # For partial results, trim the copied file to last DATASET group.
   my $num_groups = trim_results ($trimfile);

   # Unless the number of groups in the trimmed file is greater than
   # the last time we attempted to ingest, remove the file and return.
   unless ($num_groups > $NumDataSets) {
      unlink $trimfile;
      return 1;
   }
   # Store the number of data set groups.
   $NumDataSets = $num_groups;

   # Start conversion:
   debug "ingesting search results";
   ims_event('BEGIN_INGEST',\%environment);
   update_status(Reset=>1);

   # KLUDGE: Open static valids file for email info:
   my %WWWVALIDS;
   if (-e "${IU::LibDir}/conf/wwwvalids.gdb") {
      eval {tie %WWWVALIDS, 'IU::Store', 
            "${IU::LibDir}/conf/wwwvalids.gdb", 'READER'};
      $@ and error "tie failed: $@";
   }
   # Open database where we will store the results.
   # For partial results, we call it new_results.gdb and rename later.
   my $new_results = $SESSION->path_to ("$RunDir".'new_results.gdb');
   my $RESULTS = IU::InvStore->open ($new_results, 'WRCREAT')
       or return error 452, "couldn't open storage on $new_results";

   # Read in the trimmed .obj file of results objects.
   # This gives us back a hash of IU::Dataset/IU::Granule objects:
   ims_event('READ_OBJECTS',\%environment);
   my $objs = read_objects ($trimfile);
   ims_event('MERGE_OBJECTS',\%environment);
   #
   # STEP 1: PROCESS ALL THE GRANULE OBJECTS...
   #
   foreach $okey (keys %$objs) {

      # If this is the key to a granule object, get it:
      (($gkey) = ($okey =~ /^gkey:(.*)$/)) or next;
      $gran = $objs->{$okey};
      delete $objs->{$okey};          # extract it for efficiency
      #
      # STORE THE GRANULE...
      #
      # Figure out the dataset key:
      my ($archtag, $dstag) = &IU::Granule::SplitGkey($gkey);

      # Store it to the database:
      #debug "Storing granule $gkey";
      $RESULTS->add_granules([$gran]);
      update_status(NUMBER_OF_GRANULES => '++');
      #
      # LINK THE GRANULE WITH ITS DATASET AND PACKAGES...
      #
      # Link this granule's dataset to the granule:
      $dskey = "$archtag/$dstag";
      defined($ds2gkeys{$dskey}) || ($ds2gkeys{$dskey} = []);
      push(@{$ds2gkeys{$dskey}}, $gkey);

      # Link all packages in this granule to the granule:
      my $pkid;
      foreach $pkid (@{$gran->package_ids}) {
         next if $pkid eq '*';

         # Generic package key (1/M/PKID):
         $pkkey = "$archtag/M/$pkid";
         defined($pk2gkeys{$pkkey}) || ($pk2gkeys{$pkkey} = []);
         push(@{$pk2gkeys{$pkkey}}, $gkey);
      }
      # Add granule info to the dataset:
      add_gran_info_to_dataset($gran,
                               $objs->{'dskey:'. $gran->dataset_key});
   }
   # Force status update now that granules are done:
   update_status(LED_STAGE=>1, Flush=>1);

   #
   # STEP 1.5: PROCESS ALL THE MEDIA HANDLING FEE OBJECTS...
   #
   my %medias;

   foreach $okey (keys %$objs) {
      # If this is the key to a media object, get it:
      ($okey =~ /^media:(.*)$/) or next;
      my $mediaobj = $objs->{$okey};
      # Store by archive id.
      push @{$medias{$mediaobj->archive_id}}, $objs->{$okey};
      delete $objs->{$okey};          # extract it for efficiency
   }

   #
   # STEP 1.7: PROCESS ALL THE MEDIA MAXIMUM CAPACITY OBJECTS...
   #
   my %mediasizes;

   # First acquire any media capacities from WWWValids. This will
   # be an array of MediaMaximumCapacity objects. We'll extract the
   # archive ID from each of these and store it as appropriate.
   my $WWWVALIDS = IU::WWWValids->new();
   debug "WWWValids " . (defined $WWWVALIDS ? "is" : "isn't") . " okay";
   my @capacities = $WWWVALIDS->capacities();
   debug "Capacities are" . (int @capacities ? "" : "n't") . " okay";
   foreach my $mediasizeobj (@capacities) {
      # These media sizes are defaults; if the incoming ODL has
      # any by the same name, these will get overridden by the new
      # ones in the incoming ODL.
      push @{$mediasizes{$mediasizeobj->archive_id}}, $mediasizeobj;
      debug "Adding a capacity for ".$mediasizeobj->archive_id();
   }

   foreach $okey (keys %$objs) {
      # If this is the key to a media object, get it:
      ($okey =~ /^mediasize:(.*)$/) or next;
      my $mediasizeobj = $objs->{$okey};
      # Store by archive id.
      push @{$mediasizes{$mediasizeobj->archive_id}}, $objs->{$okey};
      delete $objs->{$okey};          # extract it for efficiency
   }

   #
   # STEP 2: PROCESS ALL THE DATASET OBJECTS...
   #
   foreach $okey (keys %$objs) {

      # If this is the key to a dataset object, get it:
      (($dskey) = ($okey =~ /^dskey:(.*)$/)) or next;
      $dset = $objs->{$okey};
      delete $objs->{$okey};          # extract it for efficiency
      my $archive_id = $dset->archive_id;      
      # Store array of media handling fee objects in each data set.
      $dset->mediahandlingfee($medias{$archive_id});
      # Store array of media maximum capacity objects in each data set.
      $dset->mediamaximumcapacity($mediasizes{$archive_id});

      # KLUDGE!
      # Grab contact info from static valids, for now... ultimately,
      # we'll want this info to come from imsclient!
      my $valids;
      my ($archid, $dsid) = ($dset->archive_id, $dset->id);

      # Are there valids for this dataset specifically?
      if ($valids = $WWWVALIDS{"dset:$archid:$dsid"}) {
         #debug "got valids for dataset $archid:$dsid";
         $dset->contact_addr($valids->{Addr});
         $dset->allows_email_order($valids->{Allows});
      }
      # Are there valids for this dataset's DAAC?
      elsif ($valids = $WWWVALIDS{"daac:$archid"}) {
         #debug "got fallback valids for daac $archid";
         $dset->contact_addr($valids->{Addr});
         $dset->allows_email_order($valids->{Allows});
      }
      # I guess there are NO relevant valids for this dataset...
      else {
         my %warned;
         unless ($warned{"$archid:$dsid"}) {
            warning "no 'wwwvalids.gdb' valids for $archid:$dsid";
            $warned{"$archid:$dsid"} = 1;
         }
      }
      # Store it:
      #debug "Storing dataset $dskey";
      $RESULTS->add_datasets([$dset]);
      update_status(NUMBER_OF_DATASETS => '++');
   }
   #
   # STEP 3: STORE PACKAGE-TO-GRANULE MAPPING...
   #
   # Store mapping from package key to granule keys:
   foreach $pkkey (keys %pk2gkeys) {
      #debug "Storing pk2gkeys mapping for $pkkey";
      $RESULTS->{STORE}{"pk2gkeys:$pkkey"} = $pk2gkeys{$pkkey};
   }
   #
   # STEP 4: STORE DATASET-TO-GRANULE MAPPING...
   #
   # Store mapping from package key to granule keys:
   foreach $dskey (keys %ds2gkeys) {
      #debug "Storing ds2gkeys mapping for $dskey";
      $RESULTS->{STORE}{"ds2gkeys:$dskey"} = $ds2gkeys{$dskey}
   }    
   # Close and untie:
   $RESULTS->close;

   # Remove the lock file.
   unlink $SESSION->path_to ("$RunDir".'new_results.gdb..LCK');

   # Rename the new results file to the old one that is read.
   rename $new_results, $SESSION->path_to('results.gdb');

   update_status(Flush=>1,
                 LED_STAGE => 2, 
                 LED_STAGE_FORMAT=>("Converted %d datasets, ".
                                    "%g granules in %r seconds"));
   # Remove the trimmed results file.
   unlink $trimfile;
   ims_event('RESULTS_CONVERTED',\%environment);
   1;
}

#------------------------------------------------------------
# init
#------------------------------------------------------------
# Initiallize this program.

sub init {
   require IU::FileLogger;

   # Set up stop signal handlers:
   foreach (@StopSigs) { $SIG{$_} = \&sig_stop_client }

   # Resume the session (must do this first!):
   exists($ENV{'IW_SESSION_ID'}) or die "no session id";

   $SESSION = IU::Session->resume(Id=>$ENV{'IW_SESSION_ID'}) ;

   $SESSION or die "couldn't resume session $ENV{'IW_SESSION_ID'}";

   # Set the global full-path filename of results.
   $Client_obj = $SESSION->path_to ("$RunDir".'client.obj');

   # Assume Client->preflight deleted files from a previous run.

   # Status now contains the pid of imsrunclient.
   $Status{PID} = $$;

   # Create a file with the pid number in the name.
   my $pidfile = $SESSION->path_to("$RunDir"."runclient.pid.$$");
   open PIDFILE, ">$pidfile";
   close PIDFILE;

   # Open a per-invocation log, and start logging:
   &IU::Log::init() or puke "yow! couldn't init logger!";
   my $logfile = $SESSION->path_to("$RunDir".'runclient.log');
   open STDERR, ">$logfile" or puke "couldn't open logfile: $!";    
   STDERR->autoflush(1);
   my $logger = IU::FileLogger->open(\*STDERR)
       or puke "yow! couldn't open $logfile: " .
               IU::Log::last_error();
   &IU::Log::set_handler('TEMPLOG', $logger);
   inform "Logging begun";
   1;
}

#------------------------------------------------------------
# parse_args
#------------------------------------------------------------
# Parse the command-line arguments.
# We now assume that the third argument is the remote user
# environment variable in the user's session.  We remove
# it from the ARGV list.

sub parse_args {

   if (@ARGV == 3 && defined($ARGV[2])) {
      $ENV{REMOTE_USER} = pop(@ARGV);
   }
   my @argv = @ARGV;
   if ($argv[0] =~ /^-/) {     # new style: -aACTION -iINFILE, etc.
      while (@argv) {
         shift @argv;
         /^-a/ and $Action = lc($');
      }
   } else {                      # old style: ACTION INFILE 
      $Action = lc($argv[0]) || die "usage error!";
   }
   # Inventory and Directory searches run in a subdirectory 'search'.
   if ($Action eq 'is' || $Action eq 'ds') {
      $RunDir = 'search/';
   } elsif ($Action eq 'as') {
      $RunDir = 'account_status/';
   }
   1;
}

#------------------------------------------------------------
# do_partial
#------------------------------------------------------------
# Do polling/processing of results to handle partial results
# capability.
# Use global $GrandChildPid for checking client process.

sub do_partial {

   # Define time in seconds to wait for more data to arrive.
   my $MORE_DATA_WAIT = 5;
   my $client_running = 1;     # Flag true if client still running.
   my $last_results_size = 0;  # Previous size of client.obj file.
   my $results_size = 0;       # Current size of client.obj file.
   my $never_did_anything = 1; # Flag; on if we never run an ingest.

   # Go into wait/process loop until client process dead.
   # Note that we test for a live client at the end of the
   # loop and not the beginning, so that we can process any
   # results that have arrived before the client stopped.
   do {
      # Sleep to allow more data to arrive before processing.
      sleep $MORE_DATA_WAIT;

      # Set flag if client done.
      # Perl note: kill 0 returns true if pid running.
      $client_running = kill 0 => $GrandChildPid;

      # Get current size of client.obj file.
      $results_size = -s $Client_obj;

      # If more results have arrived:
      if ($results_size > $last_results_size) {

         # Store the new size.
         $last_results_size = $results_size;

         # Get and post-process current results:
         ingest_search_results();

         # We did something at least once.
         $never_did_anything = 0;
      }
      # Break the wait/processing loop only when the client
      # process goes away.
   } while ($client_running);

   # If we got no results, we still need an empty database.
   if ($never_did_anything) {
     debug "Ack! We got nothin'!";
     ingest_search_results();
   }
   1;
}

#------------------------------------------------------------
# main
#------------------------------------------------------------
# Main entry point.

sub main {

   # Get arguments.
   parse_args() or die "can't parse arguments!";

   # Init:
   init() or die "can't init!";
   debug "Action = $Action";
   ($environment{SESSID}) = ($ENV{GAEATMP_DIR} =~ m|.*/(.*)$|);

   # Report our status:
   print_status();
   #
   # RUN THE IMSCLIENT (ASYNCHRONOUSLY as a Grandchild)...
   #
   # Convert the arguments to new-style form:
   if ((@ARGV == 2) and !(grep /^-/, @ARGV)) {     # old-style args:
      my @args = ("-a$ARGV[0]", "-i$ARGV[1]");
      @ARGV = @args;
   }
   # KLUDGE: Force fast-directory search on:
   push @ARGV, "-gn";

   # If going to run client in a subdirectory, set environment
   # variable now. Note insert '/' and chop final '/' from subdir.
   if ($RunDir) {
      $ENV{GAEATMP_DIR} .= '/' . $RunDir;
      chop $ENV{GAEATMP_DIR};
   }
   # Get filehandles for pipe between grandchild and grandparent.
   my $READER = new FileHandle;
   my $WRITER = new FileHandle;

   # Create pipes.
   pipe $READER, $WRITER or die "pipe: $!";
   $WRITER->autoflush(1);

   # Fork the child process.
   inform "Beginning imsclient operation."; 
   ims_event('CLIENT_STARTUP',\%environment);
   my $imsclient = "${IU::LibDir}/bin/imsclient";
   my $kidpid = fork;
   if (!defined($kidpid) or $kidpid < 0) {
      puke "imsrunclient child fork failed: $!";
   }
   unless ($kidpid) {         # We're the child. ($kidpid == 0)
      close $READER;         # Child doesn't need to read.
      my $gc_pid = fork();   # Fork the grandchild.
      if (!defined($gc_pid) or ($gc_pid < 0)) {
         puke "imsrunclient grandchild fork failed: $!";
      }
      unless ($gc_pid) {     # We're the grandchild.
         # Send the grandchild pid to the grandparent process.
         print $WRITER sprintf("%15d", $$);
         close $WRITER;

         # Execute the imsclient.  Don't come back.
         exec $imsclient, @ARGV;
         exit 1;            # exec shouldn't return here!
      }                      # We're the child, again.
      exit 0;                # Child exits ok.
   }
   #------------------------------------
   # We're the parent.
   waitpid ($kidpid, 0);   # Reap the dead child.

   # Close write end of pipe.
   close $WRITER; 

   # Synchronize: read until data (grandchild) or EOF (dead):
   my $buffer;
   my $nread = read ($READER, $buffer, 15);
   ($nread == 15) or
       puke "imsrunclient didn't get grandchild pid: $!";
   (int ($buffer) == -1) and
       puke "Can't spawn process: ". <$READER> ."\n";
   close $READER;
   ims_event('CLIENT_FORKED',\%environment);

   $GrandChildPid = int ($buffer);

   inform "Client pid = $GrandChildPid";

   # If this is an inventory or directory search then do partial
   # results handling code.

   if ($Action eq 'is' || $Action eq 'ds') {
      # Clear selections in resultslist by deleting the database file.
      unlink $SESSION->path_to('resultslist.gdb');
      do_partial ();
      # Delete the big client.obj, granule.gdb, and dataset.gdb
      # files since not used after this script ends.
#      unlink $Client_obj;
      unlink $SESSION->path_to("$RunDir".'dataset.gdb');
      unlink $SESSION->path_to("$RunDir".'granule.gdb');
   } else {
      # Not inventory or directory search.  Wait until client
      # process exits.  (Inv/Dir searches do this in do_partial).
      my $WAIT_FOR_GRANDCHILD = 2;
      while (kill 0 => $GrandChildPid) {
         sleep $WAIT_FOR_GRANDCHILD;
      }
   }
   $GrandChildPid = undef;

   # Set done flag in status and print it.
   $Status{DONE} = 'TRUE';
   print_status();
   ims_event('CLIENT_TERMINATED',\%environment);

   # Create a done flag file for this pid.
   my $donefile = $SESSION->path_to("$RunDir"."runclient.$$.done");
   open DONEFILE, ">$donefile";
   close DONEFILE;
   ims_event('FLAGGED_DONE',\%environment);

   # Done!
   inform 'imsrunclient Completed!';
   exit 0;
}
#------------------------------------------------------------
1;
