#!/u/joe/perl561/bin/perl -w

=head1 NAME

C<makewelcome> - creates static HTML pages for EDG "home" pages

=head1 SYNOPSIS

    # Quick and easy:
    perl makewelcome.pm
    
    # If the executable bit is set, and the first line points to your perl:
    makewelcome.pm  
    
=head1 TEMPLATES

C<makewelcome> has been radically redesigned to split up presentation and
code. Most of the code in the old vesion was there only to print out large
volumes of text which needed minor changes for each installation.

C<Text::Template> allows you to set up files which contain snippets of Perl
code that are evaluated and then replaced with the value corresponding to 
their execution. For example, if we wanted to have a field on a page that
gave the last update date and time, we could simply write this in a template
file:

   Last updated: {scalar localtime}

C<Text:Template> would find the C<scalar localtime>, execute it, and replace
the block with the current date and time. You are not limited to simple
functions; you can in fact execute code of whatever complexity you desire.

Out templates live in a subdirectory of docs, called -- oddly enough --
c<templates>; they are pointed to by the $IU::Templates setup variable.

For C<makewelcome>, the templates are rather simple; generally speaking, most
of the code is involved in determining what the base URL is; after that, it's
mostly text substitution. The most complex template code is that which builds 
the "policy" pages; this code does a lot of file checking and open checking.

Links are build using one function: C<html_link()>. This function is 
parameterized at the time it is called to see whether or not the fancy 
Javascript links are to be created, or if we just want simple ones. Since 
all the information is present in the template, we simply expand the 
template using two different values for the link type to get the plain and 
fancy pages.

=head1 SETUPS

At the moment, we have two different ways that an EDG can be installed: as
a regular EDG site, or as a hidden client.

Regular EDG sites have the regular welcome page,  plain or fancy as required,
and the necessary C<.htaccess> files to keep people from snooping in places
they shouldn't and to enable SSI.

Hidden EDG sites add a "wrapper" page around the whole site; you have to 
either have a persistent account for this site, or you have to know the
guest account's password. (it's possible, if you already know the raw URL
you'd need to get into the hidden client, to just enter that and go in without
any checking, but this isn't a real problem - hidden clients are just obscured,
not secured).

We deal with having to put many different files in many different places by 
using a data structure called a Setup. These are stored in the package
global C<%setups>. This was done to make it simpler to add any new files or
locations, should we need to. The old code required a lot of diddling to 
figure out what was being put where and when; the new code simply has a 
list of locations, along with a list of code references for each location
that are to be called when the location is to be created/updated. The output 
from each of the routines is written to the file; if the routine outputs
nothing, nothing is written.

Each setup contains an _ORDER_ entry; this entry lists the keys in the
order they should be processed. This is necessary because sometimes the
expansion processing will set up variables that will be needed by other
expansions (in particular, the .htaccess files used to keep users from 
poking around in the log and sessions directories need to know where the
denied.html file is, but we don't know that until after the callbacks are
done for denied.html.

=head1 CLOSURES

Closures are used extensively to create code to be called for the setups.
A closure is a subroutine that captures the relevant portions of the lexical
context (all variables, including C<my> variables) inside a code reference that
it creates and returns. This allows us to create a "code factory" which can be
called multiple times, creating code which "remembers" what the relevant state
of the world was when it was called.

The C<mark_xeq> routine in this code is a simple example. mark_xeq is called
with the name of a file to be chmod'ed. The file is not chmod'ed when mark_xeq
is initially called; instead, mark_xeq returns a piece of code (the closure)
which, when it is called, will remember what mark_xeq was called with and will
operate on that value.

If this is all too confusing, here's an example (from the Camel book):

   sub newprint {
      my $x = shift;
      return sub {my $y = shift; print "$x, $y!\n"); };
   }

   # Create a couple of closures:
   $hello = newprint("Hello");
   $greet = newprint("Greetings");

   # more code executes ...

   &$hello("world");    # prints "Hello, world!\n"
   &$greet("earthling") # prints "Greetings, earthling!\n"

Note how the value of $x (the parameter passed in to newprint()) is captured
inside -- closed up in -- the code returned by newprint(), even though $x
seemingly has gone out of scope when the anonymous subroutine runs.

=head1 TYPEGLOBS

We use typeglob assignment to repoint certain of the subroutines in this
program (package c<main>) into package c<makewelcome>. This makes it simpler
for C<Text::Template> to deal with them. 

The C<target_dir()> routine was misbehaving and couldn't be called without
full qualification (at least under Perl 5.005); the templates which refer to
it call it as C<makewelcome::target_dir> and that works OK.

=head1 JAVASCRIPT

We're probably going to end up stripping out the Javascript code later for
accessibility reasons; right now, the templates do not try to load in the 
Javascript and style sheets (other than the most basic one that changes the
font to sans-serif).

The welcome page has been changed radically to allow us to use Javascript
"popups". This entails the use of cascading style sheets for both the page
and the various DIVs that are used as popups, as well as the Javascript 
substrate that shows, hides, and positions them.

Basically, we generate a DIV for each popup. The popups are positioned by
Javascript code relative to the position of the cursor when the user mouses
over a link that has an attached popup. CSS is used to format the DIVs, make 
them separate from the page itself, and to show/hide them. 

=head1 NETSCAPE 4.74 AND UP ODDITIES

Someone at Netscape decided to get really, really anal-retentive about <p> tags
in Netscape 4.74 and up. This evidenced itself as popups after the first one
containing a <p> no longer popping up.

You I<must> open and close paragraph tags in popups! Yes, I know you never had
to do it before, but now you do. Deal. 

Anything that was 

   text<p>more text 

is now 

   <p>text</p><p>more text</p>

I blame over-strict potty training and XML.

=head1 GENERAL FLOW

C<makewelcome> chooses a setup based on the value of C<IU::ReleaseType>.

It runs through each of those (in the order specified by the _ORDER_ entry), 
calling the closures that were created when the program was initially compiled 
by Perl.

Each one of those either expands a template, or sets up global information to
be used later.

During the expansion of the template, some routines are "called back" into the
main program.

=begin list

=item 

C<html_link()>  stuffs the information necessary back into the %balloons
hash to be embedded at the bottom of a page to include the DIVs.

=item

C<target_dir> determines where the user should be taken if s/he tries to go
somewhere s/he shouldn't.

=item

C<calendar_url> figures out how many levels up the CGI directory is from
the page where the URL is to be embedded.

=item 

C<getpass> prompts for the hidden password, encrypts it, and returns it.

=cut

#------------------------------------------------------------
# PHASE 0: determine our top-level directory, and set up include path
#------------------------------------------------------------
use Cwd;
my $top;
use vars qw($system_architecture);


# Figure out the system_architecture to include (IRIX, Solaris, etc.)
# We need to add this to the @INC so that Perl can find the C modules 
# associated with the CPAN modules.

BEGIN {
    # Determine architecture
    eval "use Config";
    $system_architecture = $Config{archname};
}

BEGIN {
    # Figure out top directory:
    my @here = split(/\//,cwd());
    my @path = split(/\//,$0);
    my $path = $0;
    unless($0 =~ /^\//) {
        foreach my $component (@path) {
            if ($component eq "..") {
                pop @here;                   # move up one
            }
            elsif ($component eq ".") {
                next;                        # discard
            }
            else {
                push @here,@component;       # add to path
            }
        }
        $path = join("/",@here);
    }
    ($path =~  m|/lib/perl/[^/]+$|) or die "can't extract info from $path!";
    $top = $`;

    # Set up include path:
    unshift @INC, "$top/lib/perl/first", "$top/lib/perl",
                  "$top/lib/perl/CPAN",  "$top/lib/perl/plus", 
                  "$top/lib/perl/CPAN/$system_architecture";

    push @INC, "$top/lib/perl/fallback";
}

# Okay, now that we know where everything is, load 'em up.

use strict;

use IU;
use IU::URL;
use LWP::Simple;
use File::Copy;
use Text::Template;
use Getopt::Long;

#------------------------------------------------------------
# Setups
#
# Each setup is a hash, with each entry being a location to set up
# and a list of subroutines to call to complete the setup.
#
# We might want to add a layer here: setups could reference a COMMON
# set of files, and add on a custom set as well ... maybe later.
my %setups = (
	      PUBLIC  => {
   _ORDER_ => 
               [
		"$IU::DocsDir/imswelcome/denied.html",
		"$IU::DocsDir/imswelcome/index.html",
		"$IU::DocsDir/imswelcome/plain.html",
		"$IU::DocsDir/imswelcome/privacy_policy.html",
		"$IU::DocsDir/imswelcome/security_policy.html",
		"$IU::DocsDir/imswelcome/access_policy.html",
		"$IU::DocsDir/imswelcome/.htaccess",
		"$IU::TopDir/log/.htaccess",
                "$IU::TopDir/cgi-bin/.htaccess",
	       ],
   "$IU::DocsDir/imswelcome/.htaccess"   =>
      [expand("ssi")],
   "$IU::DocsDir/imswelcome/denied.html" =>
      [expand("denied"), denied("$IU::DocsDir/imswelcome/denied.html")],
   "$IU::DocsDir/imswelcome/index.html"  =>
      [expand("homepage","linktype"=>"popup"), mark_xeq()],
   "$IU::DocsDir/imswelcome/plain.html"  =>
      [expand("homepage","linktype"=>"plain"), mark_xeq()],
   "$IU::DocsDir/imswelcome/privacy_policy.html"  => 
      [expand("privacy_policy")],
   "$IU::DocsDir/imswelcome/security_policy.html" => 
      [expand("security_policy")],
   "$IU::DocsDir/imswelcome/access_policy.html" => 
      [expand("access_policy")],
   "$IU::TopDir/log/.htaccess" =>
      [expand("block_all")],
   "$IU::TopDir/cgi-bin/.htaccess" =>
      [expand("allow_cgi")],
                        },
              HIDDEN => {
   _ORDER_ => [
	       "$IU::DocsDir/protected/denied.html",
	       "$IU::DocsDir/protected/.htaccess",
	       "$IU::DocsDir/protected/index.html",
	       "$IU::DocsDir/protected/plain.html",
	       "$IU::DocsDir/imswelcome/index.html",
	       "$IU::DocsDir/imswelcome/privacy_policy.html",
	       "$IU::DocsDir/imswelcome/security_policy.html",
		"$IU::DocsDir/imswelcome/access_policy.html",
	       "$IU::TopDir/.htpasswd",
	       "$IU::TopDir/log/.htaccess",
               "$IU::TopDir/cgi-bin/.htaccess",
	      ],
   "$IU::DocsDir/protected/.htaccess"    => 
      [expand("ssi-pwd")],
   "$IU::DocsDir/protected/denied.html"  => 
      [expand("denied"), denied("$IU::DocsDir/protected/denied.html")],
   "$IU::DocsDir/protected/index.html"  => 
      [expand("homepage","linktype"=>"popup"),mark_xeq()],
   "$IU::DocsDir/protected/plain.html"  => 
      [expand("homepage","linktype"=>"plain"), mark_xeq()],
   "$IU::DocsDir/imswelcome/index.html"  => 
      [expand("hidden","linktype"=>"popup"), mark_xeq()],
   "$IU::DocsDir/imswelcome/privacy_policy.html"  => 
      [expand("privacy_policy")],
   "$IU::DocsDir/imswelcome/security_policy.html" => 
      [expand("security_policy")],
   "$IU::DocsDir/imswelcome/access_policy.html" => 
      [expand("access_policy")],
   "$IU::TopDir/.htpasswd" =>
      [expand("htpasswd")],
   "$IU::TopDir/log/.htaccess" =>
      [expand("block_all")],
   "$IU::TopDir/cgi-bin/.htaccess" =>
      [expand("allow_cgi")],
                     },
              PREOPS => {
   _ORDER_ => [
	       "$IU::DocsDir/protected/denied.html",
	       "$IU::DocsDir/protected/.htaccess",
	       "$IU::DocsDir/protected/index.html",
	       "$IU::DocsDir/protected/plain.html",
	       "$IU::DocsDir/imswelcome/index.html",
	       "$IU::DocsDir/imswelcome/privacy_policy.html",
	       "$IU::DocsDir/imswelcome/security_policy.html",
		"$IU::DocsDir/imswelcome/access_policy.html",
	       "$IU::TopDir/.htpasswd",
	       "$IU::TopDir/log/.htaccess",
               "$IU::TopDir/cgi-bin/.htaccess",
	      ],
   "$IU::DocsDir/protected/.htaccess"    => 
      [expand("ssi-pwd")],
   "$IU::DocsDir/protected/denied.html"  => 
      [expand("denied"), denied("$IU::DocsDir/protected/denied.html")],
   "$IU::DocsDir/protected/index.html"  => 
      [expand("homepage","linktype"=>"popup"),mark_xeq()],
   "$IU::DocsDir/protected/plain.html"  => 
      [expand("homepage","linktype"=>"plain"), mark_xeq()],
   "$IU::DocsDir/imswelcome/index.html"  => 
      [expand("hidden","linktype"=>"popup"), mark_xeq()],
   "$IU::DocsDir/imswelcome/privacy_policy.html"  => 
      [expand("privacy_policy")],
   "$IU::DocsDir/imswelcome/security_policy.html" => 
      [expand("security_policy")],
   "$IU::DocsDir/imswelcome/access_policy.html" => 
      [expand("access_policy")],
   "$IU::TopDir/.htpasswd" =>
      [expand("htpasswd")],
   "$IU::TopDir/log/.htaccess" =>
      [expand("block_all")],
   "$IU::TopDir/cgi-bin/.htaccess" =>
      [expand("allow_cgi")],
                     },
              'HIDDEN-PREOPS' => {
   _ORDER_ => [
	       "$IU::DocsDir/protected/denied.html",
	       "$IU::DocsDir/protected/.htaccess",
	       "$IU::DocsDir/protected/index.html",
	       "$IU::DocsDir/protected/plain.html",
	       "$IU::DocsDir/imswelcome/index.html",
	       "$IU::DocsDir/imswelcome/privacy_policy.html",
	       "$IU::DocsDir/imswelcome/security_policy.html",
		"$IU::DocsDir/imswelcome/access_policy.html",
	       "$IU::TopDir/.htpasswd",
	       "$IU::TopDir/log/.htaccess",
               "$IU::TopDir/cgi-bin/.htaccess",
	      ],
   "$IU::DocsDir/protected/.htaccess"    => 
      [expand("ssi-pwd")],
   "$IU::DocsDir/protected/denied.html"  => 
      [expand("denied"), denied("$IU::DocsDir/protected/denied.html")],
   "$IU::DocsDir/protected/index.html"  => 
      [expand("homepage","linktype"=>"popup"),mark_xeq()],
   "$IU::DocsDir/protected/plain.html"  => 
      [expand("homepage","linktype"=>"plain"), mark_xeq()],
   "$IU::DocsDir/imswelcome/index.html"  => 
      [expand("hidden","linktype"=>"popup"), mark_xeq()],
   "$IU::DocsDir/imswelcome/privacy_policy.html"  => 
      [expand("privacy_policy")],
   "$IU::DocsDir/imswelcome/security_policy.html" => 
      [expand("security_policy")],
   "$IU::DocsDir/imswelcome/access_policy.html" => 
      [expand("access_policy")],
   "$IU::TopDir/.htpasswd" =>
      [expand("htpasswd")],
   "$IU::TopDir/log/.htaccess" =>
      [expand("block_all")],
   "$IU::TopDir/cgi-bin/.htaccess" =>
      [expand("allow_cgi")],
                     },
             );
#------------------------------------------------------------
# Check arguments ...
#------------------------------------------------------------
my $debug;
my $result = GetOptions(debug => \$debug);

#------------------------------------------------------------
# Set up configuration variables...
#------------------------------------------------------------

my ($http,$server,$secured,$port,$relative_url,$relative_doc) = 
   validate_URLs();

my $GATEWAY = "$http://$server:$port/$relative_url/nph-ims.cgi";
my $SECURED = "$http://$server:$port/$relative_url/$secured/nph-ims.cgi";
my $DOCS    = "$http://$server:$port/$relative_doc";
my $HOSTNAME    = ${IU::HttpdHost};

# Get the information for the what's new.
my ($fetched_date,@fetched_urls) = get_whats_new();

# Configuration variables to be passed to the expander closures when they
# are called. Note that linktype is NOT in here; the expanders insert their
# own value for this if it's needed.
my %config_variables =
 (
  GATEWAY  => $GATEWAY,
  SECURED  => $SECURED,
  DOCS     => $DOCS,
  HOSTNAME => $HOSTNAME,
  almost_white => "#EEEEEE",
  nice_aqua    => "#CCFFFF",
  bgcolor      => "white",
  fetched_date => $fetched_date,
  fetched_urls => [@fetched_urls],
  base_font_size => "14pt",
 );

# the popups hash, written by html_link().
use vars(qw(%balloons));
%balloons = ();

# the denied.html path, written by denied().
my $deniedpath;

#---------------------------------------------------------------------
# Text::Template helpers

# These typeglob assignments make it simpler to code templates.
# include()          calls Text::Template::_load_text   directly
# calendar_url       calls main::calendar_url           directly
# html_link          calls main::html_link              directly
#
# Now the templates can simply say {include("foo.template")} and not
# have to say {Text::Template::_load_text("foo.template")}.
#
# Why are they in package "makewelcome"? Because we tell Text::Template
# to do everything in that package (with the PACKAGE option).
#
# These are ASSIGNMENTS. They have to be up here, before the rest of
# the code is executed. If you relocate them down below the main code,
# they won't get executed in time, and Text::Template will complain
# bitterly about not being able to find them. LEAVE THEM AT THE TOP.
#
# Sometimes, Text::Template gets confused and thinks it should be 
# looking for these subroutines in itself -- this may be a Perl bug.
# Adding the aliases inside Text::Template as well is a backup to
# resolve the names properly.

*makewelcome::include             = \&Text::Template::_load_text;
*Text::Template::include          = \&Text::Template::_load_text;

# This front-ends Text::Template's include_and_fill with our version,
# to have the proper hash passed along automatically. See the 
# include_and_fill() code further down in this module for details.
*makewelcome::include_and_fill    = \&main::include_and_fill;
*Text::Template::include_and_fill = \&makewelcome::include_and_fill;

# If we want Text::Template to be able to see subroutines from this program,
# we have to do a similar trick by putting them into package "makewelcome" 
# as well.
*makewelcome::target_dir       = \&main::target_dir;
*makewelcome::html_link        = \&main::html_link;
*makewelcome::calendar_url     = \&main::calendar_url;
*makewelcome::getpass          = \&main::getpass;

#------------------------------------------------------------
# Determine which setup is to be used.
# The value for each key in %setup is a hash reference;
# we dereference it again as a hash so we can assign it to %files.
my %files = %{$setups{uc($IU::ReleaseType)}};

#------------------------------------------------------------
# Create the directories that we may need.
mkdir("$IU::DocsDir/imswelcome",0775);
mkdir("$IU::DocsDir/protected",0775);

#------------------------------------------------------------
# Do all the real work.
#
# Be ready in case something blows up. As we work with pages, we
# push them onto the page registry. If something goes kablooie,
# we rename them back again. If nothing does, we will remove all 
# the ".old" files for anything in the registry at the end of the
# program.

my @page_registry = ();
$SIG{__DIE__} = \&we_died;

# Expansion code.
# Pull out the order in which we should do things from the _ORDER_
# key in the hash. For each file (key), run through the list of associated 
# closures. If the closure returns a value, print that into the file.
print "Creating files ...";
foreach my $file (@{$files{_ORDER_}}) {
    push @page_registry,$file;
    print "Processing $file ... " if $debug;
    backup_file($file);
    # Write contents into file.
    if (open(MWFILE,">$file")) {
        print "opened OK ... " if $debug;
        # Select MWFILE, turn buffering off, select STDOUT again.
	select((select(MWFILE),$|++)[0]);
        # extract the anonymous array which is the value of the key in %files.
        # For each item (closure), call it with %config_variables, which it
        # needs to do the actual expansion. Print any result into the file.
        #
        # Note that we add the current file name to the hash as well;
        # calendar_url needs it to know how far back up to go to generate
        # the proper relative URL so the calendar will embed.
        foreach my $code (@{$files{$file}}) {
	    my $result = $code->((%config_variables),"current_file"=>$file);
	    print MWFILE $result if $result;
            print ".";
	}
        close MWFILE;
    }
    else {
	print "[Can't open $file: $!]";
    }
}
print " done\n";

# Cleanup code.
# Create the necessary symlinks to the denied.html file. We want the contents
# of the directories to be accessible via the Web server, so .htaccess is
# out, but we don't want just anyone poking around in there, so we add a
# symlink (as index.html) to the "denied.html" file.

# General cases. Link index.html in the directory to denied.html.
print STDOUT "Creating links to denied.html ... ";
my @docs_subdirs = qw(btns dict icons imswww-usr ipod java navigator 
                      persistent pics test-scripts
                      . .. ../bin ../cgi-bin ../lib ../persistent ../src
                      ../support_files ../tmp);

foreach my $subdir (@docs_subdirs) {
   print "... $subdir "; 
   next unless -e "$IU::DocsDir/$subdir";
   unlink "$IU::DocsDir/$subdir/index.html";
   symlink($deniedpath, "$IU::DocsDir/$subdir/index.html")
      or print "(skipped: $!)";
}
print " ... done\n";

# Special case: javadoc dir gets a symlink from the existing packages.html
# in the directory to index.html.
print "Fixing Javadoc ...";
unlink  "${IU::DocsDir}/javadoc/index.html";
symlink("${IU::DocsDir}/javadoc/packages.html",
        "${IU::DocsDir}/javadoc/index.html")
      or die "Can't build symbolic link for ".
             "${IU::DocsDir}/javadoc/index.html: $!";
print " done\n";

# Done, get out and let the END block clean up.
exit(0);

#-----------------------------------------------------------------------
# Utility routines.

# Set the global 'we died" flag; report the error, move on.
my $we_died;

sub we_died { 
  $we_died=1; 
  print "\nError: " . shift;
}

# The finalization routine. Remove ".old" files that are left around.
# Called by Perl itself when the program terminates.
END {
   print "Cleaning up ...";
   if ($we_died) {
      fixup_pages();
   }
   else {   
      foreach (@page_registry) {
         unlink("$_.old") ? print "." 
                          : print ("[$_ backup not cleaned up ($!)]");
      }
   }
   print " done\n";
}

# The cleanup routine. Rename ".old" files back.
sub fixup_pages {
    undef $SIG{__DIE__};
    print "*** Restoring old versions\n";
    # We don't want to foreach, because that would leave all the file
    # names in the registry, and then the END block would try to process
    # them again.
    while (@page_registry) {
       $_ = pop @page_registry;
       print "Restoring $_ ... " if $debug;
       rename("$_.old",$_) or 
          print "[Can't fix $_: $!]";
       print  "\n" if $debug;
    }
}

# Back up files by copying them to a ".old" file.
sub backup_file {
   my $filename = shift;
   copy($filename, "$filename.old");
}

sub passwd {
   my $pass1 = shift;
    # Generate salt:
    srand(time);
    my $salt = (('0'..'9','a'..'z','A'..'Z')[int(62 * rand)] .
                ('0'..'9','a'..'z','A'..'Z')[int(62 * rand)]);

    # Encrypt and return:
    return crypt($pass1, $salt);
}
sub validate_URLs {
    # We first want to guarantee that the proper port number is associated 
    # with the host and that all the URLs are properly qualified so they
    # are secure (https) URLS as needed. We do this by taking the CGiUrl
    # specified by the user and decomposing it. 
    #
    # If it's a fully-qualified URL, we strip all of the pertinent info out 
    # of it and ignore the setup variables that would override it (with a 
    # warning, of course!). If it's a relative URL, we use the appropriate 
    # SETUP variables to figure out the info we need.
    
    my ($http,  $server,  $Port,  $relative_url) = 
      IU::URL::decompose_url($IU::CgiUrl);
    my ($http2, $server2, $Port2, $relative_doc) = 
      IU::URL::decompose_url($IU::DocsUrl);
    
    # Protocols must be both "http" or "https".
    if ($http ne $http2) {
	print <<EOF;
Protocol conflict between CgiUrl and DocsUrl:
CgiUrl  uses $http protocol
DocsUrl uses $http2 protocol

Change the protocols to match and rerun makewelcome.
EOF
	exit(1);
    }
    
    # Docs and scripts should be on the same server.
    if ($server ne $server2) {
	print <<EOF;
Servers do not match for CgiUrl and DocsUrl:
   CgiUrl  uses the server $server
   DocsUrl uses the server $server2

Change the servers to match and rerun makewelcome.
EOF
	exit(1);
    }
    
    # Both URLS should use the same port.
    if ($Port ne $Port2) {
	print <<EOF;
Server ports do not match for CgiUrl and DocsUrl:
   CgiUrl  uses the port $Port
   DocsUrl uses the port $Port2

Change the ports to match and rerun makewelcome.
EOF
	exit(1);
    }
    
    # Add slashes as needed to the relative URLs.
    $relative_url =~ s{/$}{} if $relative_url =~ m{/$};
    $relative_url =~ s{^/}{} if $relative_url =~ m{^/};
    $relative_doc =~ s{/$}{} if $relative_doc =~ m{/$};
    $relative_doc =~ s{^/}{} if $relative_doc =~ m{^/};

    # Where the password-protected directory is.
    my $secured = $IU::SecuredDir;
    $secured =~ s{/$}{} if $secured =~ m{/$};
    $secured =~ s{^/}{} if $secured =~ m{^/};

    ($http, $server, $secured, $Port, $relative_url, $relative_doc);
}

# Figure out whether the denied link and similar links should go to the
# imswelcome directory or the protected one.
#
# If the release is a hidden one, the denied page should point into the
# protected area (thus requiring a password).
sub target_dir {
   if (uc($IU::ReleaseType) ne "PUBLIC") {
      return "protected";
   }
   # If it's not hidden, go to the (open) home page.
   else {
      return "imswelcome";
   }
}

# get_whats_new uses LWP to read the current entries on the "What's New" page.
sub get_whats_new {
    my $page = LWP::Simple::get($IU::WhatsNewUrl);
    my @list =();
    my $when = "news server unavailable";
    if (defined $page) {
	$when = 0;
        my @page = split(/\n/,$page);
        foreach my $line (@page) {
	    if ($line =~ /<li>/i) {
		# This is a list item. Make sure the URL is fully-qualified.
		unless ($line =~ /href="http:/i) {
                    $line =~ s/href="/class="isTip" href="$IU::WhatsNewUrl/i;
                    $line =~ s/news.html//;
                    $line =~ s/<li>/<br>/i;
		}
            push @list, $line;
	    }
            ($when) = ($line =~ /Last .*: (.*, \d{4})/) unless $when;
	}
    }
    ($when,@list);
}

# calendar_url checks to see where the current page is relative to the
# cgi_bin directory, and emits a URL with the proper number of ".."s.
# Needed because we want the hidden home pages to be able to show the
# calendar too.
sub calendar_url{
   my $current_location = shift;
   my @components = split(/\//,$current_location);
   my $cal_url = $IU::CalendarEmbed;
   pop @components;   # drop the last item (the file name)
   pop @components;   # drop the last item again (the containing directory)
   while (grep /^docs$/,@components and $components[-1] ne 'docs') {
       pop @components;
       $cal_url = "../$cal_url";
   }
   $cal_url;
}

# html_link creates a link of the proper type.
sub html_link {
    my $link = new Popup_Link(@_);
    my ($div,$text) = $link->div;
    %balloons = (%balloons,$div=>$text) if defined $div;
    $link->output;
}

# getpass returns an encrypted password.
sub getpass {
    # Prompt for, get, and create the password.
    print "==> Enter the password for the " . lc($IU::ReleaseType) . " client: ";
    my $answer = <STDIN>;
    chomp $answer;
    # Put the password into the incoming parameters.
    passwd($answer);
}

#---------------------------------------------------------------------
# Closures.

# expand takes a file name and an optional hash.
# it returns a closure that, when called, does the following:
#   - merge any other hash passed in with any hash it was created with
#   - creates a Text::Template object from the file it was created with
#   - expands the template using the values in the merged hash.
# expands returns the contents of the expanded template;
#
# Note that anything passed in the variables hash overrides anything
# passed in when the expansion actually occurs.
sub expand {
    # Get the name of the file to be expanded later and any overrides.
    my ($file,%variables) = @_;
    return sub {
        # Capture the parameters passed in when this closure is called later.
        my %hash = @_;
        # Override the values passed in now with the ones
        # passed in when the closure was created.
        local $_;
        foreach (keys %variables) {
           $hash{$_} = $variables{$_};
	}
        # Load and expand the template. Value of fill_in() is the text
        # to be written into the file.
        my $template = 
           Text::Template->new(TYPE   => 'FILE', 
                               SOURCE => "$IU::Templates/$file.template");
           $template->fill_in(PACKAGE=> "makewelcome",
                              HASH   => \%hash) 
           or print "Template error: $Text::Template::ERROR\n";
    };
}

# include_and_fill front-ends Text::Template's fill_in_file, passing along
# the configuration variables from makewelcome automatically. 
#
# Text::Template::fill_in_file shifts off the first parameter (the file
# name), and then passes all the leftover parameters to 
# Text::Template::fill_in, so we can just tack on the extra parameters here
# and then call the thin we really wanted in the first place.
sub include_and_fill {
  # %config_variables already exists outside, so we can pass it along.
  Text::Template::fill_in_file(shift, @_, PACKAGE=>"makewelcome",
                                          HASH   =>\%config_variables); 
}

# mark_xeq returns a closure which will mark the current file name as 
# readable and executable by all. The closure specifically returns undef 
# to ensure we don't try to write out any value from this
# closure.
sub mark_xeq {
   return sub {
       my %args = @_;
       chmod 0764, $args{"current_file"}; 
       return;
   };
}

# denied simply sets a variable to record the filename of the "denied.html"
# file. Returns undef to keep extraneous values out of the file.
sub denied {
   # This is in "use vars", so it doesn't need to be my'ed.
   return sub {
       my %args = @_;
       $deniedpath = $args{"current_file"};
       return;
   };
}

#------------------------------------------------------------

package Popup_Link;

# Encapsulates management and output behavior for popup links.
# automatically handles differentiating between popup and plain,
# outputs necessary Javascript and stylesheet information.

# new
#
# Create new instance of a Popup_Link.
# Accepts a hash as parameter; the hash contains any or all of the
# following:
#     Field name        Accessor      Type       Comments
#     ------------------------------------------------------------------
#     BaseUrl           base_url()    string     Base URL for HREF
#     IsExternal        is_external() T/F        true if an external URL
#                                                (not a link into the EDG)
#     Popup             popup()       reference  Reference to a Popup object
#     LinkText          link_text()   string     Text to appear as the target
#                                                for the HREF
#     OptParams         opt_params()  reference  Reference to a hash containing
#                                                key/value pairs for other CGI
#                                                parameters

sub new {
    my $class = shift;
    my %params = @_;
    my $self = {};
    bless $self, $class;

    foreach my $p (keys %params) {
	$self->{$p} = $params{$p};
    }
    $self->{OptParams} = undef unless defined $self->{OptParams};
    $self;
}

sub _accessor {
    my ($self,$field,$value) = @_;
    $self->{$field} = $value if defined $value;
    $self->{$field};
}

sub image {
    my ($self,$value) = @_;
    _accessor($self,"Image",$value) or "";
}

sub alt {
    my ($self,$value) = @_;
    _accessor($self,"Alt",$value) or "";
}

sub base_url {
    my ($self,$value) = @_;
    _accessor($self,"URL",$value);
}

sub is_external {
    my ($self,$value) = @_;
    _accessor($self,"IsExternal",$value);
}

sub opt_params {
    my ($self,$value) = @_;
    _accessor($self,"OptParams",$value);
}

sub popup {
    my ($self,$value) = @_;
    _accessor($self,"Popup",$value);
}

sub link_text {
    my ($self,$value) = @_;
    _accessor($self,"LinkText",$value) or "";
}

sub type {
    my ($self,$value) = @_;
    _accessor($self,"Type",$value) or "";
}

sub output {
    my ($self) = @_;
    # Do the common actions in setting up the opt_params.
    $self->common_actions();

    # Add the HTML to the popup page.
    if ($self->type() eq 'popup') {
	my $link  = ($self->popup_link() or '');
	my $name  = ($self->popup()->name() or '');
	my $text  = ($self->link_text() or '');
	my $image = ($self->image() or '');
	my $pop = qq( <a href="$link" class="isTip">$image$text</a>);
       return $pop;
    }

    # Add the HTML to the plain page.
    if ($self->type() eq 'plain') {
	my $link  = ($self->plain_link() or '');
	my $text  = ($self->link_text() or '');
	my $image = ($self->image() or '');
        my $alt   = ($self->alt() or '');
     
        $image  = qq(<IMG SRC="$image" alt="$alt" BORDER=0>) if $image;
	my $pln = qq(<a href="$link">$image$text</a>);
	return $pln;
    }
}

sub div {
    my $self = shift;
    return if $self->{Type} eq 'plain';
    ($self->popup->name,$self->popup->text);
}

sub popup_link {
    my $self = shift;
    $self->opt_params()->{JS} = "1" if defined $self->opt_params;
    $self->decoded_url();
}

sub plain_link {
    my $self = shift;
    $self->opt_params()->{JS} = "0" if defined $self->opt_params;
    $self->decoded_url();
}

sub common_actions {
    my $self = shift;
    # Insert ModePREPARE as frontend.
    # If no mode was specified, we're going to MAINSRCH.
    # If no submode was specified, ignore it.
    # Swap the current mode into nextmode, and change mode to PREPARE.
    if (defined $self->opt_params()) {
	$self->opt_params()->{"nextmode"} = $self->opt_params()->{"mode"}
	  if defined $self->opt_params()->{"mode"};
	$self->opt_params()->{"nextsub"}  = $self->opt_params()->{"submode"}
	  if defined $self->opt_params()->{"submode"};
	$self->opt_params()->{"nextmode"} = "MAINSRCH" 
	  unless defined $self->opt_params()->{"nextmode"};
	$self->opt_params()->{"mode"}     = "PREPARE";
	$self->opt_params()->{"submode"}  = "DEFAULT";
    }
}
  
sub decoded_url {
    my $self = shift;
    my $base_link = $self->base_url();
    unless ($self -> is_external()) {
	my $opt_params = $self->opt_params();
	
	# If there are CGI params for this link, add them on.
	# Note that we are currently NOT URL ENCODING.
	if (defined $opt_params) {
	    my $sep = "?";
	    foreach my $k (keys %$opt_params) {
		my $v = $opt_params->{$k};
		$base_link .= "$sep$k=$v";
		$sep = "&";
	    }
	}
    }
    $base_link;
}
  
package Popup;

sub new {
   my $class = shift;
   my %params = @{$_[0]};
   my $self = {};
   bless $self, $class;

   $self->{Width} = 180;
   foreach my $p (keys %params) {
      $self->{$p} = $params{$p};
   }
   $self;
}

sub _accessor {
   my ($self,$field,$value) = @_;
   $self->{$field} = $value if defined $value;
   $self->{$field};
}

sub name {
   my ($self,$value) = @_;
   _accessor($self,"Name",$value);
}

sub text {
   my ($self,$value) = @_;
   _accessor($self,"Text",$value);
}

1;
