#!/u/joe/perl561/bin/perl

=head1 NAME

setup - manage the setup of the operational environment

=head1 DESCRIPTION

=begin desc

=item setup -make

Make a new SETUP file, initialized by making intelligent
guesses at your site configuration.

=item setup -makefrom I<file>

Make a new SETUP file, initialized from an existing one.
Useful for upgrades, where you want to "inherit" as many values
from an existing SETUP file as you can.

=item setup -apply

Apply the contents of the SETUP file, generating 
sh, Perl, etc. source files containing the predigested information.

=end desc

Invoke with no arguments for the full usage.

=head1 ABOUT THE SETUP PROCESS

The following variables are absolutely, positively, 100%-hard-coded
for safety reasons:

    $Version = (the version number)
    $SystemName = (the current official system name)

The following variable is extrapolated from $0:

    $TopDir  = (the top-level directory)

From this is derived the following:

    $CgiDir    = "$TopDir/cgi";
    $DocsDir   = "$TopDir/docs";
    $Templates = "$TopDir/docs/templates";
    $LibDir    = "$TopDir/lib";
    $LogDir    = "$TopDir/log";

All other variables are set up via the F<SETUP> file.

=begin construction

The GaeaDataDir can (and should) possibly be removed from the
F<configure> process.

=end construction


=head1 AUTHOR

Eryq / Hughes STX Corporation

=cut

# Set up include path:
use Cwd;
my $Top;
BEGIN {
    # Figure out top directory:
    my $path = cwd() . "/$0";
    ($path =~ m|/lib/perl/[^/]+$|) or
	die "can't extract info from $path!";   
    $::Top = $`;

    # Set up include path:
    unshift @INC, "$::Top/lib/perl/first", 
                  "$::Top/lib/perl", 
                  "$::Top/lib/perl/plus";
    push @INC,    "$::Top/lib/perl/fallback";
}

require 5.002;          # make sure they have the right Perl version 

use IU;
use IU::Log;
use IU::Util qw(which);
require IU::Ask;
require IU::SetupFile;
require IU::Version;
require IU::SystemName;
require IU::Escape;
use Cwd;


#------------------------------------------------------------
#
# Globals...
#
#------------------------------------------------------------

# Fallback values, for anything we couldn't "guess at".
# KEEP THIS LIST IN ALPHABETICAL ORDER!!!!
%Fallback = 
    (
     AccountMax      => '20mb',
     AcronymsUrl     => 'http://gcmd.gsfc.nasa.gov/acronyms.html',
     ActiveSessionPort  => '6600',
     ActiveTimeCheck => '15',
     AdminCrypt      => 'NO PASSWORD',
     AirsBrowseSearchTitle => 'AIRS Browse',
     AllAccountsMax  => '1 gb',
     AllowJava       => 'N',
     AlternatePort   => '',
     CalendarUrl     => 'http://redhook.gsfc.nasa.gov/ims-bin/uptime-cal.cgi',
     CalendarEmbed   => "../../cgi-bin/$IU::Version::VERSION/grab-cal.cgi",
     CalendarTimeout => 10,
     CacheComfort    => '5mb',
     CacheLinger     => '2m',
     CacheMax        => '10mb',
     CacheShelfLife  => '30m',
     CgiUrl          => '/ims-bin/1.7/',
     CountryList     => '/conf/countries.txt',
     CriticalAllAccountsMax => '3 gb',
     CustomizeMetadata => 'NO',
     DaacDocsUrl => 'http://www-v0ims.gsfc.nasa.gov/v0ims/DOCUMENTATION/documentation.html',
     DocsUrl         => '/~imswww/1.7/',
     DictFile        => '/www/imsdev/docs/dict/knuth_words',
     DisableByDiscipline => 'NO',
     DisplayBrowseSearch => 'NO',
     EcsProfileGateway => '',
     EcsProfileSmc => '',
     EcsHomeDaac => 'GSFC;EDC;LARC;NSIDC',
     EcsRegistration  =>
'http://lyta.gsfc.nasa.gov/~imswww/pub/imswelcome/ecsNotice.html',
     EmailGasp     => 'YES',
     EOSHomeUrl    => 'http://eos.nasa.gov',
     EosDisHomeUrl => 'http://eos.nasa.gov/eosdis',
     EosRelatedSitesUrl => 'http://www-v0ims.gsfc.nasa.gov/v0ims/RELSITES/relsites.html',
     EosOtherSourcesUrl => 'http://www-v0ims.gsfc.nasa.gov/v0ims/RELSITES/other_source.html',
     ExpectLocal   => '/usr/bin/expect',
     ExplorerUrl   => 'http://www.microsoft.com/ie',
     JavaFAQUrl    => 
'http://harp.gsfc.nasa.gov/~imswww/pub/imswelcome/imswww-javafaq.html',
     FAQUrl        => 
'http://harp.gsfc.nasa.gov/~imswww/pub/imswelcome/imswww.faq.html',
     ForceDebug    => 'NO',
     GaeaDataDir   => "$ENV{HOME}/support_files/sup_files",
     GaeaDistribDir  => "$ENV{HOME}/support_files/dist_files",
     GazetteerUrl => 'http://fat-albert.alexandria.ucsb.edu:8827/gazetteer/',
     GcmdGetdifUrl => 'http://gcmd.nasa.gov/getdif.htm',
     GcmdGetsuppUrl => 'http://gcmd.gsfc.nasa.gov/cgi-bin/md/getsupp.pl',
     GcmdHomeUrl   => 'http://gcmd.gsfc.nasa.gov/',
     GcmdHostAddr  => '192.107.190.77',
     GcmdPort      => '5502',
     GnatsEmail    => 'bugs@killians.gsfc.nasa.gov',
     GranColumnsMax => '20',
     GuideKillTime => '15 m',
     GuideLimit    => '100',
     GuideType     => 'TRANSITIONAL',
     HavePersistentSessions => 'NO',
     HdfInfoUrl    => 
'http://www.ncsa.uiuc.edu/SDG/Homepage/hdf.html',
     HdfTools      =>'http://hdfeos.gsfc.nasa.gov/hdfeos/viewingHDFEOS.html',
     HomeEmail     => 'imswww-dev@harp.gsfc.nasa.gov',
     HomeUrl       => 
'http://www-v0ims.gsfc.nasa.gov/v0ims/eosdis_home.html',
     HttpdHost     => '',
     ImsDefnUrl    => 
'http://www-v0ims.gsfc.nasa.gov/v0ims/definitions/',
     ImsGUIUrl     => 
'http://www-v0ims.gsfc.nasa.gov/v0ims/install_0.html',
     ImsGlossUrl   => 
'http://www-v0ims.gsfc.nasa.gov/v0ims/glossary.of.terms.html',
     ImsHomeUrl    => 
'http://www-v0ims.gsfc.nasa.gov/v0ims/eosdis_home.html',
     ImsSitesUrl   => 
'http://harp.gsfc.nasa.gov/~imswww/pub/imswelcome/imswwwsites.html',
     ImsUserUrl    => 
'http://www-v0ims.gsfc.nasa.gov/v0ims/DAACS.html',
     LdLibraryPath => '/usr/lib:/usr/local/lib:/usr/local/hdf5/lib',
     ListLength => '10',
     ListMax => '1000',
     LocalStatsDir => '',
     LogLevel      => 'normal',
     LogoutTime    => '1 d',
     MapCgiScript  => 'imswdbtile.cgi',
     MapResolution => 'MEDIUM',
     MaxGransDefault => '100',
     MaxSearchTime => '90',
     MaxFreetextDatasets => '20',
     MisrLimit     => '10',
     MTPEHomeUrl   => 'http://www.earth.nasa.gov',
     MorePaths     => 
'/bin:/usr/bin:/usr/local/bin:/usr/local/wais/bin',
     NASAHomeUrl   => 'http://www.nasa.gov',
     NASAOrgCode   => '423',
     NetscapeUrl   => 'http://home.netscape.com/computing/download/index.html',
     OfficialName   => 'Medora Macie',
     OfficialEmail  => 'Mail Code 423, NASA/GSFC, Greenbelt, MD 20771',
     OfficialTag => 'Medora.B.Macie.1',
     OutreachUrl => 'http://www-v0ims.gsfc.nasa.gov/v0ims/OUTREACH/outreach.html',
     PerlHttpdPort => '',
     PerlPath      => '/usr/bin/perl',
     PersistentDir => '/usr/people/imswww/persistentUsers/',
     PolicySecurity => 'US Government',
     PolicyPrivacy => 'US Government',
     PolicyAccess => 'US Government',
     PrimaryPort   => '80',
     PSAMax => '2',
     QueryPreviewUrl => '',
     RandomContext => '6',
     RandomLength  => '9',
     ReleaseType   => 'Public',
     ResultsLifetime => '14 d',
     SampleDataUrl => 'http://www-v0ims.gsfc.nasa.gov/v0ims/DATAPRODSERV/sample_data.html',
     SecuredDir    => 'secured',
     SecureServer  => '0',
     SecurePort    => '',
     SendmailCmd   => '/usr/lib/sendmail -t -i -oeq',
     SessionLife   => '4h',
     SiteAdmin     => 'imswww',
     SiteName      => 'SiteName',
     SiteSpecDir   => "$ENV{HOME}/site_specific",
     SitesWithKey  => "ASF,asf\@eos.gsfc.nasa.gov",
     Slogin        => '/usr/local/bin/slogin',
     SpidexUrl     => "http://guide.gsfc.nasa.gov:81/guide/cgi-bin/guide/search/nph-search.cgi", 
     SpidexConf    => "http://guide.gsfc.nasa.gov:81/guide/conf/spidex.conf",
     SpidexTimeout => "300",
     StateList     => '/conf/states.txt',
     StatsEmail    => 'imswww-stats@harp.gsfc.nasa.gov',
     StatsId       => 'v0ieww00',
     StatsUrl      => 
'http://harp.gsfc.nasa.gov/~imswww/stats/summary/web/global-imswww-stats.html',
     SyslogLevel   => '6',
     TerraProductsUrl => 
'http://eosdatainfo.gsfc.nasa.gov/eosdata/terra/terra_dataprod.html',
     TutorialUrl   => 'http://edcwww.cr.usgs.gov/landdaac/tutorial/',
     UpdateEvery   => '15',
     UseNonSpatial =>  '',
     UseNonTemporal =>  '',
     UserSurveyUrl   => '',
     ValidsPort    => 'ops',
     WebmasterName => 'Chao-Hsi Chang',
     WebmasterEmail => 'chao-hsi_chang@sesda.com',
     WebmasterTag => 'Chao-Hsi.Chang.1',
     WhatsNewUrl   => 'http://www-v0ims.gsfc.nasa.gov/v0ims/NEWS/news.html',
     WgHttpdPort   => '',
     WwwUser       => 'wwwuser.wwwuser',
     );
#------------------------------------------------------------
#
# For "setup [-make|-makefrom]"...
#
#------------------------------------------------------------

#------------------------------------------------------------
# write_setup -- write the SETUP file
#------------------------------------------------------------

sub write_setup {
    my ($setup) = @_;
    my $now = localtime;

    # Save old setup file (okay if we can't):
    rename("SETUP", "SETUP.back") or
	warn "couldn't save old SETUP file: $!";

    # Do it:
    open(SETUP, ">SETUP") or
	puke("Couldn't open SETUP for writing: $!");

    print SETUP <<EOF;
#------------------------------------------------------------
# SETUP
#------------------------------------------------------------
#
# DESCRIPTION
#    Setup for the current IMS/www site.
#    Generated $now.
#
#    Lines beginning with #, and blank lines, are ignored.
#    The general form is 'VarName "var value"', all on one line.
#    Within the value string, quotes (") are forbidden, and \-escapes
#    are not recognized.
#
# ORIGINAL AUTHOR
#    Eryq (Erik Dorfman)
#    Hughes STX Corporation

##--------------------------------------
## SITE-SPECIFIC CRITICAL SETUP...
##--------------------------------------
## You may need to change some of these to get the system to run.

#              The crypt'ed password for the administrators to use,
#              exactly as you'd see it in an /etc/passwd file... but
#              PLEASE, DO NOT USE AN ACTUAL SYSTEM PASSWORD FOR THIS!
#              Note: the special value "NO PASSWORD" allows ANYONE in!!
AdminCrypt     "$setup->{AdminCrypt}"

#              Where the IMS/wwww CGIs; a URL with trailing '/'.
#              Note that specifiying a fully-qualified URL here will
#              OVERRIDE any PrimaryPort, SecureServer, or SecurePort
#              settings; setup extracts the server type (secure or not),
#              server name, and server port from the URL if it is
#              fully-qualified. A relative URL is recommended.

CgiUrl         "$setup->{CgiUrl}"

#              Where the IMS/www HTML docs are; URL with trailing '/'.
#              If this is specified as a fully-qualified URL, you must 
#              make sure that the protocol, server, and port match those
#              specified in CgiUrl. A relative URL is recommended.
DocsUrl        "$setup->{DocsUrl}"

#              Which daac in archive.odl is for ECS profile submit requests.
#              If multiple ECS profile gateways are available, they
#              should all be listed here and separated by semi-colons(;).
EcsProfileGateway   "$setup->{EcsProfileGateway}"

#              Which daac in archive.odl is for ECS profile requests
#              other than ECS profile submit requests.
#              If multiple ECS profile gateways are available, they
#              should all be listed here and separated by semi-colons(;).
EcsProfileSmc   "$setup->{EcsProfileSmc}"

#              This is the list of Home DAACs that a user can pick from.
#              The default will be the first in the list which should be
#              set to the current DAAC if it applies.
#              If multiple Home DAACS are available, they
#              should all be listed here and separated by semi-colons(;).
EcsHomeDaac   "$setup->{EcsHomeDaac}"

#              Where is the  ECS User Registration tool?
EcsRegistration    "$setup->{EcsRegistration}"

#              Path to the directory that holds the IMS valids:
GaeaDataDir    "$setup->{GaeaDataDir}"

#              Path to directory that holds the IMS distribution files:
GaeaDistribDir "$setup->{GaeaDistribDir}"

#              Maximum number of columns for a Data granule listing.
#              This is only an advisement, not a hard limit.
GranColumnsMax "$setup->{GranColumnsMax}"

#              This is the hostname that will be used in URLs pointing
#              back into the system, and also for email sent to/from 
#              this site; leave it empty ("") for the default hostname.
HttpdHost      "$setup->{HttpdHost}"

#              Default number of rows per page for dataset/granule/cart list.

ListLength     "$setup->{ListLength}"

#              Default maximum number of items total which can be in a 
#              granule/cart list.

ListMax        "$setup->{ListMax}"

#              Default for whether or not to use java.

AllowJava      "$setup->{AllowJava}"

#              Map server script used by the Geo selection java map.
#              "imswdbtile.cgi" - lower resolution, fewer features,
#                                 smaller database
#              "drawglobe.cgi" - higher resolution, more features,
#                                larger data base.
#              "imswdbtile.cgi" is the default.

MapCgiScript   "$setup->{MapCgiScript}"

#              Used when MapCgiScript = "drawglobe.cgi", must install correct
#              map database in support_files/support_dist/map. See installation
#              instructions for more details. 
#              MEDIUM - requires lowres and medres databases (25MB), map detail 
#                       increases up to zoom 64x.
#              HIGH   - requires lowres, medres and highres databases (237MB), 
#                       map detail increases up to zoom 500x.
#              ULTRA  - requires lowres, medres, highres, and ultrares 
#                     databases (1331MB), map detail increases up to zoom 4000x.

MapResolution  "$setup->{MapResolution}"

#              *PATH* to the CGI script which embeds calendar info.
#              (You'll need to update this to reflect where your cgi-bin
#              is relative to your docs/imswelcome. It has to be a relative
#              path! The sample value is probably right if you're following
#              our install directions.)
CalendarEmbed  "$setup->{CalendarEmbed}"

#              The calendar-embed script will wait this many seconds
#              before deciding the calendar is unavailable.
CalendarTimeout "$setup->{CalendarTimeout}"

#              Default for whether or not we allow the user to customize
#              metadata coming back as results.

CustomizeMetadata "$setup->{CustomizeMetadata}"

#              Test versions of the EDG may not have the valids necessary
#              to drive the "By Discipline" feature.  To disable it,
#              set this to "YES". The "By Discipline" will be shown if
#              this is left blank or any other value.
#

DisableByDiscipline "$setup->{DisableByDiscipline}"


#		Should we display Browse Search -> YES/NO ???

DisplayBrowseSearch "$setup->{DisplayBrowseSearch}"

#              This is added to the \$PATH variable, to pick up...
#                 compress (for unpacking hdf.Z files)
#                 gunzip   (for receiving compressed trees, and HDF.gz)
#                 gzip     (ditto)
#                 waisq    (for guide searches)
#                 zcat     (for hdf2gif)

MorePaths      "$setup->{MorePaths}"

#              (IRIX only) Set this variable if you see dying gasps 
#              saying that "no GIFs were recovered, but no errors 
#              were reported!" It should contain the path to the 
#              directory containing libhdf5.so. Non-IRIX users may
#              ignore this setup variable.
LdLibraryPath  "$setup->{LdLibraryPath}"

#              Port number for the IMS/WWW *custom* Perl HTTPD... NOT
#              the normal one!  UNDER NO CIRCUMSTANCES SHOULD YOU PUT
#              THE NUMBER '80' HERE!  If no number is supplied here,
#              only the regular HTTPD will be used.
#
#              Note that you cannot have a PerlHttpdPort and a 
#              SecureServer simultaneously. You must choose one or
#              the other. 

PerlHttpdPort  "$setup->{PerlHttpdPort}"

#              Location of the Persistent User Profiles
PersistentDir  "$setup->{PersistentDir}"

#              Which security policy to link to from the welcome page:
#
#              Default is "US Government" for US Government Security Policy 
#              (lib/conf/input_security.html)
#
#              Set to "local" if you want to define a local security policy
#              by putting it in $SiteSpecDir/input_security.html
#
#              Set to blank "" if you don't want a security policy link on the
#              Welcome page. 
#
PolicySecurity "$setup->{PolicySecurity}"

#              Which privacy policy to link to from the welcome page:
#
#              Default is "US Government" for US Government Privacy Policy
#              (lib/conf/input_privacy.html)
#
#              Set to "local" if you want to define a local privacy policy
#              by putting it in $SiteSpecDir/input_security.html
#
#              Set to blank "" if you don't want a privacy policy link on the
#              Welcome page.
#
PolicyPrivacy  "$setup->{PolicyPrivacy}"

#              Which accessibility policy to link to from the welcome page:
#
#              Default is "US Government" for US Government Accessibility Policy
#              (lib/conf/input_access.html)
#
#              Set to "local" if you want to define a local privacy policy
#              by putting it in $SiteSpecDir/input_access.html
#
#              Set to blank "" if you don't want an accessibility policy link 
#              on the  Welcome page.
#
PolicyAccess  "$setup->{PolicyAccess}"

#              Port where the primary HTTPD on your system listens.
#              Should only be changed if you run your primary httpd
#              on a port other than 80.
PrimaryPort    "$setup->{PrimaryPort}"

#              Maximum number of Product Specific Attributes (PSAs) allowed
#              in a search.
PSAMax         "$setup->{PSAMax}"

#              Which type of release:
#                 Public, Hidden, Preops, or Hidden-Preops ? 
#              Leave this as Public unless you know why you'd want
#              to use one of the others.
ReleaseType    "$setup->{ReleaseType}"

#              Location of the secured subdirectory of cgi-bin (for
#              persistent logins).
SecuredDir     "secured"

#              Whether or not you ar using a secure server (SSL support).
#              Set to any non-zero value if you are doing so.
SecureServer   "$setup->{SecureServer}"

#              The port at which the secure server listens on your machine.
#              Leave empty if you do not have a secure server.
SecurePort     "$setup->{SecurePort}"

#              Full path to SSH's slogin
Slogin         "$setup->{Slogin}"

#              Full path to expect
ExpectLocation "$setup->{ExpectLocal}"

#              
#              How to invoke sendmail on this system, so that (1) the
#              message is scanned for recipients, and (2) a dot on a
#              line by itself doesn't end input.
#              E.g.: "/usr/lib/sendmail -t -i".
SendmailCmd    "$setup->{SendmailCmd}"

#              Name for your Web Gateway.  Appears on the Welcome page.
#              "Compose a Search using the IMS Web Gateway
#               at <SiteName>."
SiteName       "$setup->{SiteName}"

#              Directory to hold site specific info and documents
#              for the gateway, such as welome_header.html.
SiteSpecDir    "$setup->{SiteSpecDir}"

#              Our local signature, for statistics.   DO NOT FILL
#              THIS IS WITH A RANDOM NUMBER!  CONTACT THE IMS/WWW
#              DEVELOPERS FOR A LEGITIMATE NUMBER FOR YOUR SITE!!!
StatsId        "$setup->{StatsId}"

#              Port number for the *custom* Web Gateway HTTPD.
#              Use (BASE + (100 * MAJORVERSION) + MINORVERSION).
# 
#              The BASE is as follows:
#
#              8000    "Public" or "operational" installs.  
#                      The most typical case.
#              9000    "Beta" or "test" installs.  
#                      Usually for developer site only.
#              10000   "Development" installs.  
#                      Usually for developer site only.
#
#              The MAJORVERSION and MINORVERSION are the Web Gateway 
#              version: "MAJ.MIN".  So for example, a public (or soon 
#              to be public) v.1.8 would be at port:
#              8000 + (100*1) + 8 = 8108.
WgHttpdPort    "$setup->{WgHttpdPort}"

#              What user.group does/should HTTP scripts run as?
#              This must be identical to your httpd's configuration!
#              UNDER ***NO*** CIRCUMSTANCES SHOULD THIS BE ROOT!!!
WwwUser        "$setup->{WwwUser}"

#              Location to copy event logs to on the local machine
#              for further processing into the local statistics page.
#              Leave blank if you don't want to generate a local 
LocalStatsDir  "$setup->{LocalStatsDir}"

##--------------------------------------
## SITE-SPECIFIC CACHE/SESSION OPTIONS...
##--------------------------------------
## These affect how the various global resources are managed.
## Adjust these based on your site's capabilities.

#	       Allow the users to have persistent directories on the 
#	       server. "YES" or "NO".  Recommend "NO" until you talk
#              with the Web Gateway team about it.
HavePersistentSessions "$setup->{HavePersistentSessions}"

#              What's a *comfortable* size for the browse cache?
#              Used by imscleanup.  Cleanups continue until the cache
#              shrinks below this size.  Specify integer, optional
#              spaces, and units.  e.g., "500kb", or "1mb", or "4 mb".
CacheComfort   "$setup->{CacheComfort}"

#              What's the *maximum* size of the browse cache?
#              Used by imscleanup.  E.g., "5mb".  Cleanups are forced
#              when the cache exceeds this size.  CURRENTLY, the cache
#              MAY exceed this size BETWEEN cleanups.
CacheMax       "$setup->{CacheMax}"

#              How long can browse products remain unrequested in cache
#              until they're "stale" and ready for deletion?  Specify a
#              time with units: e.g., "40s", or "20m", or "1h".
#              Recommend: "30m".
CacheShelfLife "$setup->{CacheShelfLife}"

#              How long after the logical deletion of an item from
#              cache until the item can be physically deleted?
#              Recommend: "2m".
CacheLinger    "$setup->{CacheLinger}"

#              How long can sessions remain unused until they're
#              deemed "stale" and are deleted?  Specify a time with
#              units:  e.g., "40m", "2h", etc.  Recommend: "4h".
SessionLife    "$setup->{SessionLife}"

#              How long should a logged in account be inactive before
#              we log the user out?  Used by imscleanaccounts.pl.
#              Specify integer time, optional blanks, and units.
#              Recommend:  "1 d" for one day.
LogoutTime     "$setup->{LogoutTime}"

#              How long can results be kept in a user account, that 
#              is not accessed, before they are deleted (to free up
#              disk space and keep info from becoming out-of-date)?
#              Used by imscleanaccounts.pl.  Specify integer time,
#              optional blanks, and units: e.g., "10d" for 10 days.
#              Recommend: "14 d" for fourteen days, the valids cycle.
ResultsLifetime  "$setup->{ResultsLifetime}"

#              How much disk space should be allowed for each user
#              account?  Used by imscleanaccounts.pl.  This maximum
#              limit should be enough to handle the results of a large
#              search.  If a user exceeds this size and they are not
#              signed-in, their results will be purged.  It they are
#              signed-in they will be warned.
#              Specify as an integer, optional spaces, and units (mb).
#              e.g., "20 mb" for 20 megabytes.  Recommend: "20 mb".
AccountMax     "$setup->{AccountMax}"

#              How much disk space should be allowed for ALL user 
#              accounts?  Used by imscleanaccounts.pl.  This maximum
#              limit should be quite large, since each account can be
#              typically 3-8 MB (but could be much smaller or bigger).
#              Specify as an integer, optional spaces, and units (mb
#              or gb).  e.g., "500 mb" or "2gb".  Recommend: "1 gb".
AllAccountsMax "$setup->{AllAccountsMax}"

#              How much disk space should be allowed for ALL user
#              accounts before more drastic measures are invoked to
#              clear space?  Used by imscleanaccounts.pl.  This maximum
#              limit should be very large, since each account can be
#              typically 3-8 MB (but could be much smaller or bigger).
#              It should be larger than AllAccountsMax.
#              Specify as an integer, optional spaces, and units (mb
#              or gb).  e.g., "500mb" or "2 gb".  Recommend: "3 gb".
CriticalAllAccountsMax   "$setup->{CriticalAllAccountsMax}"

#              Under CRITICAL disk space conditions (account usage
#              exceedingCriticalAllAccountsMax), if guide search
#              results are not accessed for this period of time, then
#              they will be deleted - even from accounts where the user
#              is signed in.  Specify a time, optional spaces,
#              and units:  e.g. "5m" for 5 minutes.  Recommend: "15 m".
GuideKillTime  "$setup->{GuideKillTime}"

##--------------------------------------
## LOAD BALANCING OPTIONS...
##--------------------------------------
## Used by cgi-bin/ActiveSession.pl to decide how many sessions are presently
## 'active.'

#                What port should ActiveSession.pl listen on?
#                Should NOT be 80 or any other port that currently used
#                in the EDG.
ActiveSessionPort  "$setup->{ActiveSessionPort}"

#                How long (in minutes) is a session considered 'active'
#
ActiveTimeCheck "$setup->{ActiveTimeCheck}"

##--------------------------------------
## SITE-SPECIFIC MISC. OPTIONS...
##--------------------------------------
## You can change these if you want, but you probably don't need to.

#              The dictionary file used as a corpus for generating 
#              random passwords. Assumes a path realtive to the "dict"
#              directory in your "docs" directory unless you specify a
#              fully-qualified pathname.
DictFile       "knuth_words"

#              Should copies of serious "dying gasps" be emailed to the
#              SiteAdmin address (below)?  YES or NO.
EmailGasp      "$setup->{EmailGasp}"

#              Turn debugging on/off for entire site (usually for
#              debugging or development only!).  
#              YES = force debugging on, NO = don't force debugging on.
ForceDebug     "$setup->{ForceDebug}"

#              The maximum number of documents a guide search will return.
#              This just sets a default value; the users can override this.
GuideLimit     "100"

#              Set the type of guide search query strings.
#              Prior to April 1997 all daacs used style 'OLD'.
#              All daacs should be using 'NEW' by May 1997.
#              OLD: "Data_Center=NSIDC and Parameter=SNOW_DEPTH"
#              NEW: "arch7 and parm1006"
#              BOTH: "arch7 or Data_Center=NSIDC
#                                and parm1006 or Parameter=SNOW_DEPTH"
#              XML: "datacenter/nsicd and parameter/snow_depth"
#              TRANSITIONAL: run both XML and WAIS searches, automatically
#                            dispatching the search to the right server(s)
GuideType      "$setup->{GuideType}"

#              Logging level for IMS event logging.
#
#              The IMS developers will inform you as to any other 
#              settings required; for now, please leave this as-is.
LogLevel       "$setup->{LogLevel}"

#              Context used to determine "next letter" for random
#              passwords. "6" seems to provide passwords which are
#              memorable, but not actual English.
RandomContext  "6"

#              Length of randomly-generated passwords.
RandomLength   "9"

#              What user account on this host should site-specific
#              mail go to?  The "Web Gateway site administrator".
SiteAdmin      "$setup->{SiteAdmin}"

#              How long should we wait for Spidex before giving up?
#              (in seconds)
SpidexTimeout  "$setup->{SpidexTimeout}"

#              What level should the IMS libraries log messages at?
#                   4 = up to WARNING       
#                   5 = up to NOTICE
#                   6 = up to INFO
#                   7 = up to DEBUG
SyslogLevel    "$setup->{SyslogLevel}"

#              How often to update search status, in seconds:
UpdateEvery    "$setup->{UpdateEvery}"

#              The URL of the welcome page.
#              If you set this to the empty string, it will point at
#              the default location, which is
#              "{DocsUrl}imswelcome/index.html"
WelcomeUrl     "$setup->{WelcomeUrl}"

#              How many granules to return per dataset:
MaxGransDefault  "$setup->{MaxGransDefault}"

#              How long a search should run, max (in minutes):
MaxSearchTime  "$setup->{MaxSearchTime}"

#              Add checkbox for Non-Spatial granules in Search Options.
#              To allow public versions to switch off till ECS finish tests
UseNonSpatial   "$setup->{UseNonSpatial}"

#              Add checkbox for Non-Temporal granules in Search Options.
#              To allow public versions to switch off till ECS finish tests
UseNonTemporal   "$setup->{UseNonTemporal}"

#              Maximun number of datasets allow in a freetext search
#              without sending a large results warning.
MaxFreetextDatasets "$setup->{MaxFreetextDatasets}"

#              How many MISR granules are allowed to be render on covermap
#              at ONE time
MisrLimit      "$setup->{MisrLimit}"

#              Which version of valids should autoxfer get?  You can
#              set this to 'ops' for operational, 'preops' for
#              testing preops valids, etc.  You can also enter a
#              specific valids port number if you know it.
ValidsPort     "$setup->{ValidsPort}"


#              The name of the webmaster
WebmasterName  "$setup->{WebmasterName}"

#              The email address of the webmaster
WebmasterEmail "$setup->{WebmasterEmail}" 

#              The meta tag of the webmaster
WebmasterTag   "$setup->{WebmasterTag}" 

##--------------------------------------
## Page Titles
##--------------------------------------

#              The title of the AIRS Browse Search page in the Navigator
#              and on the page.
AirsBrowseSearchTitle "$setup->{AirsBrowseSearchTitle}"

##--------------------------------------
## IMS/WWW EXTERNAL FILES...
##--------------------------------------
## You probably won't need to change these.

#              List Of Countries for address:
CountryList    "$setup->{CountryList}"

#              List Of States for address:
StateList      "$setup->{StateList}"

##--------------------------------------
## IMS/WWW EXTERNAL EMAIL ADDRESSES...
##--------------------------------------
## You probably won't need to change these.

#              Email Address for GNATS bug-reporting database:
GnatsEmail     "$setup->{GnatsEmail}"

#              Who to email with questions/comments about the IMS/www
#              system:
HomeEmail      "$setup->{HomeEmail}"

#              Place event logs are mailed to by the imsstats program:
StatsEmail     "$setup->{StatsEmail}"

##--------------------------------------
## GLOBAL SYSTEM OPTIONS...
##--------------------------------------
## You should NOT change these without permission from the NASA Official.

#              NASA Official responsible for V0 IMS Web Gateway
#              Do not change this without first getting permission 
#              from the NASA Official.
OfficialName   "$setup->{OfficialName}"

#              E-mail address for NASA Official responsible.
OfficialEmail  "$setup->{OfficialEmail}"

#              meta tag for NASA Official responsible.
OfficialTag    "$setup->{OfficialTag}"

#              meta tag for the NASA Organization Code
NASAOrgCode    "$setup->{NASAOrgCode}"

#              User Services Offices for sites using Data Access Keys.
SitesWithKey   "ASF,asf\@eos.gsfc.nasa.gov"

##--------------------------------------
## IMS/WWW EXTERNAL URLS...
##--------------------------------------
## You probably won't need to change these.

#              URL for the EOSDIS home page:
AcronymsUrl    "$setup->{AcronymsUrl}"

#              URL for the uptime calendar:
CalendarUrl    "$setup->{CalendarUrl}"

#              URL for the DAAC-related documentation:
DaacDocsUrl    "$setup->{DaacDocsUrl}"

#              URL for the EOS home page:
EOSHomeUrl     "$setup->{EOSHomeUrl}"

#              URL for the EOSDIS home page:
EosDisHomeUrl  "$setup->{EosDisHomeUrl}"

#              URL for the EOS Related Sites:
EosRelatedSitesUrl  "$setup->{EosRelatedSitesUrl}"

#              URL for the other sources:
EosOtherSourcesUrl  "$setup->{EosOtherSourcesUrl}"

#              URL for the "Microsoft Internet Explorer" button:
ExplorerUrl    "$setup->{ExplorerUrl}"

#              URL for Frequently Asked Questions (FAQ)
FAQUrl         "$setup->{FAQUrl}"

#              URL for Java Frequently Asked Questions (FAQ)
JavaFAQUrl     "$setup->{JavaFAQUrl}"

#              URL for Gazetteer link.
GazetteerUrl   "$setup->{GazetteerUrl}"

#              URL for the GCMD "getdif" script:
GcmdGetdifUrl  "$setup->{GcmdGetdifUrl}"

#              URL for the GCMD supplementary information:
GcmdGetsuppUrl  "$setup->{GcmdGetsuppUrl}"

#              URL for the GCMD home page:
GcmdHomeUrl    "$setup->{GcmdHomeUrl}"

#              URL for the GCMD Directory Searches:
GcmdHostAddr   "$setup->{GcmdHostAddr}"

#              URL for the GCMD Directory Port:
GcmdPort       "$setup->{GcmdPort}"

#              URL pointing to info on the HDF format, and available
#              tools:
HdfInfoUrl     "$setup->{HdfInfoUrl}"

#              URL pointing to HDF viewing tools
HdfTools       "$setup->{HdfTools}"

#              URL of the IMS/WWW home page:
HomeUrl        "$setup->{HomeUrl}"

#              URL for the IMS definitions directory, with trailing '/'
ImsDefnUrl     "$setup->{ImsDefnUrl}"

#              URL for the IMS Glossary of Terms...
ImsGlossUrl    "$setup->{ImsGlossUrl}"

#              URL for GUI Client Access
ImsGUIUrl      "$setup->{ImsGUIUrl}"

#              URL for the IMS home page:
ImsHomeUrl     "$setup->{ImsHomeUrl}"

#              URL for Active Web Gateway Sites
ImsSitesUrl    "$setup->{ImsSitesUrl}"

#              URL for User Support Contact Information
ImsUserUrl      "$setup->{ImsUserUrl}"

#              URL for Mission to Planet Earth Home Page
MTPEHomeUrl    "$setup->{MTPEHomeUrl}"

#              URL for the NASA Home Page
NASAHomeUrl    "$setup->{NASAHomeUrl}"

#              URL for the "Netscape Now!" button:
NetscapeUrl    "$setup->{NetscapeUrl}"

#              URL for Outreach Info:
OutreachUrl    "$setup->{OutreachUrl}"

#              URL for the Query Preview system.
#              Leave blank '' if you don't want the Query Preview link on
#              the navigator.
QueryPreviewUrl "$setup->{QueryPreviewUrl}"

#              URL for the Sample Data Products:
SampleDataUrl    "$setup->{SampleDataUrl}"

#              URL for Statistics Page
StatsUrl       "$setup->{StatsUrl}"

#              URL for Spidex (the XML guide "spider")
SpidexUrl      "$setup->{SpidexUrl}"

#              URL for Spidex config file (has guide catalog docs)
SpidexConf     "$setup->{SpidexConf}"

#              URL for Terra Products Page
TerraProductsUrl "$setup->{TerraProductsUrl}"

#              URL for Tutorial Page
TutorialUrl    "$setup->{TutorialUrl}"

#              URL for the "What's new" page
WhatsNewUrl    "$setup->{WhatsNewUrl}"

#              URL for the "User Survey" page
UserSurveyUrl    "$setup->{UserSurveyUrl}"


#-------------------- end of "SETUP" --------------------
EOF
    
    # Done!
    inform "Wrote $TOP/SETUP.";
    &IU::Util::banner("Now edit the SETUP file, and run `setup -apply'.");
    1;    
}

#------------------------------------------------------------
#
# try_to...
#
#------------------------------------------------------------

#------------------------------------------------------------
# try_to_find_ncsa_httpd
#------------------------------------------------------------
sub try_to_find_ncsa_httpd {
    foreach $home ('/usr/local/etc',		   
		   '/usr/local/lib',
		   '/usr/local',
		   '/usr/etc',
		   '/etc',
		   '/') {
	next if (! -d $home);    
	debug "scanning for an HTTPD in $home...";

	# Prefer "httpd":
	if (-d ($path = "$home/httpd")) {
	    $SuspectNcsaHttpd = $path;
	    inform "Looks like NCSA HTTPD, in...   $SuspectNcsaHttpd";
	    return 1;
	}			
	else {    # take highest-numbered version:
	    opendir(HOME, $home) || next;
	    @files = readdir(HOME);
	    closedir(HOME);
	    @files = sort {$b cmp $a} @files;             	    
	    foreach $file (@files) {          
		if ($file =~ /^httpd/) {
		    $SuspectNcsaHttpd = "$home/$file";
		    inform "Looks like NCSA HTTPD, in...   $SuspectNcsaHttpd";
		    return 1;
		}
	    }
	}			
    }
    warn "Couldn't verify NCSA HTTPD";
    0;   # nothing found
}

#------------------------------------------------------------
# try_to_scan_httpd_conf -- if there's an httpd.conf, scan it for info
#------------------------------------------------------------
sub try_to_scan_httpd_conf {

    # Make sure we have an HTTPD:
    return undef unless ($SuspectNcsaHttpd);

    open(CONF, "$SuspectNcsaHttpd/conf/httpd.conf") or return undef;
    debug "Scanning httpd.conf for info...";
    while (defined($line = <CONF>)) {
	if ($line =~ m|^\s*User\s+(\S+)|i) {
	    $SuspectCgiUser = $1;
	    inform "[httpd.conf]: CgiUser is...   $SuspectCgiUser";
	}
	elsif ($line =~ m|^\s*Group\s+(\S+)|i) {
	    $SuspectCgiGroup = $1;
	    inform "[httpd.conf]: CgiGroup is...  $SuspectCgiGroup";
	}
    }
    close(CONF);
    1;
}

#------------------------------------------------------------
# try_to_scan_srm_conf -- if there's an srm.conf, scan it for info
#------------------------------------------------------------
sub try_to_scan_srm_conf {

    # Make sure we have an HTTPD:
    return undef unless ($SuspectNcsaHttpd);

    open(CONF, "$SuspectNcsaHttpd/conf/srm.conf") or return undef;
    debug "Scanning srm.conf for info...";
    while (defined($line = <CONF>)) {
	if ($line =~ m|^\s*ScriptAlias\s*/ims-bin/\s*|) {
	    $SuspectCgiUrl = '/ims-bin/1.5/';
	    inform "[srm.conf]: CgiUrl is...   $SuspectCgiUrl";
	}
	elsif ($line =~ m|^\s*ScriptAlias\s*/cgi-bin/\s*(\S+)|) {
	    $SuspectServerCgiBinDir = $1;
	    chop $SuspectServerCgiBinDir;
	    inform "[srm.conf]: HTTPD cgi-bin is...   $SuspectServerCgiBinDir";
	}
	elsif ($line =~ m|^\s*Alias\s*/imswww/\s*|) {
	    $SuspectDocsUrl = '/~imswww/1.5/'; 
	    inform "[srm.conf]: DocsUrl is...   $SuspectDocsUrl";
	}	    
    }
    close(CONF);
    1;
}

#------------------------------------------------------------
# try_to_find_server_cgi_bin
#------------------------------------------------------------
sub try_to_find_server_cgi_bin {
    if ($SuspectServerCgiBinDir) {
	return 1;     # no-op
    }
    elsif ($SuspectNcsaHttpd) {
	$SuspectServerCgiBinDir = "$SuspectNcsaHttpd/cgi-bin";
	return 1;
    }
    else {
	# no suspects!
	return 0;
    }
}

#------------------------------------------------------------
# try_to_find_server_docs
#------------------------------------------------------------
sub try_to_find_server_docs {
    if ($SuspectServerDocsDir) {
	return 1;     # no-op
    }
    elsif ($SuspectNcsaHttpd) {
	$SuspectServerDocsDir = "$SuspectNcsaHttpd/htdocs";
	return 1;
    }
    else {
	# no suspects!
	return 0;
    }
}

#------------------------------------------------------------
#
# guess_at...
#
#------------------------------------------------------------

#------------------------------------------------------------
# guess_at_CgiUrl
#------------------------------------------------------------
sub guess_at_CgiUrl {
    my $guess = '/ims-bin/1.5/';     # default

    # Got it already from srm.conf?
    if (defined($SuspectCgiUrl)) {
	$guess = $SuspectCgiUrl;
    }
    elsif (defined($SuspectServerCgiBinDir)) {
	foreach $symlink ('ims-www', 
			  'imswww', 
			  'ims-bin', 
			  'imsbin', 
			  'ims') {
	    if (-l ($path = "$SuspectServerCgiBinDir/$symlink")) {
		$guess = "/cgi-bin/$symlink/";
		last;
	    }	
	}
    }
    debug "Guessed at CgiUrl: $guess";
    return $guess;
}

#------------------------------------------------------------
# guess_at_DocsUrl
#------------------------------------------------------------
sub guess_at_DocsUrl {
    my $guess = '/~imswww/1.5/';


    # Got it already from srm.conf?
#    if (defined($SuspectDocsUrl)) {
#	$guess = $SuspectDocsUrl;
#    }
#    elsif (defined($SuspectServerDocsDir)) {
#	foreach $symlink ('ims-www', 
#			  'imswww', 
#			  'ims-docs', 
#			  'imsdocs', 
#			  'ims') {
#	    if (-l ($path = "$SuspectServerDocsDir/$symlink")) {
#		$guess = "/$symlink/";
#		last;
#	    }	
#	}
#    }
    $guess = '/~imswww/1.5/';
    debug "Guessed at DocsUrl: $guess";
    return $guess;
}

#------------------------------------------------------------
# guess_at_GaeaDataDir
#------------------------------------------------------------
# Guesses are...
#      ~/support_files/sup_files           
#      $TopDir/lib/support_files
#      ~/lib/support_files
#      ~/support_files
#      ~ims/lib/support_files

sub guess_at_GaeaDataDir {
    my $guess;
    my @pwent;

    # Get directories to try:
    my @guesses = ("$ENV{HOME}/support_files/sup_files",
		   "$::Top/support_files",
		   "$::Top/lib/support_files",
		   "$ENV{HOME}/support_files",
		   "$ENV{HOME}/lib/support_files");
    push @guesses, "$pwent[7]/lib/support_files" if (@pwent = getpwnam('ims'));

    # Try each:
    foreach $guess (@guesses) {
	if (-d $guess) {
	    debug "Guessed at GaeaDataDir: $guess";
	    return $guess;
	}
    }
    undef;
}

#------------------------------------------------------------
# guess_at_HttpdHost
#------------------------------------------------------------
sub guess_at_HttpdHost {
    my $host;
    while ($host = (gethostent())[0]) {
	last if ($host =~ /\S+\.\S+/);
    }
    endhostent();
    $host;
}

#------------------------------------------------------------
# guess_at_MorePaths
#------------------------------------------------------------
sub guess_at_MorePaths {
    my $path;
    my %paths;
    my $guess;
    
    foreach $prog ('compress', 
		   'gunzip',
		   'gzip',
		   'waisq',
		   'zcat',
		   ) {
	if ($path = which($prog)) {
	    $path =~ s|/[^/]+$||g;	
	    $paths{$path} = 1;
	}
    }
    $guess = join(':', (sort keys %paths),'/usr/local/wais/bin');
    debug "Guessed at MorePaths: $guess";
    return $guess;
}

#------------------------------------------------------------
# guess_at_PerlHttpdPort
#------------------------------------------------------------
# In theory we *could* base this off our version number;
# namely, 8000 + 100*major + minor (e.g., 1.5 => 8105).  HOWEVER:
# this can allow an admin to mindlessly have multiple daemons running...
# and those guys are BIG!  To, to prevent accidents, NO GUESSES!
# ALWAYS USE THE SAME FALLBACK VALUE!

sub guess_at_PerlHttpdPort {
    undef;    # ALWAYS
}

#------------------------------------------------------------
# guess_at_PersistentDir
#------------------------------------------------------------
# Assume that the persistent directory will be under ~imswww
# but outside ANY specific installation.
#
# For the default value to be propagated to SETUP, this routine
# needs to be added.

sub guess_at_PersistentDir {
    undef;    # ALWAYS
}

#------------------------------------------------------------
# guess_at_SecuredDir
#------------------------------------------------------------
# Assume that the secured directory is named secured and is
# located directly under cgi-bin.

sub guess_at_SecuredDir {
    "secured";	# recommended in install doc
}


#------------------------------------------------------------
# guess_at_PerlPath
#------------------------------------------------------------
sub guess_at_PerlPath {
    return which('perl') || '/usr/local/bin/perl';
}

#------------------------------------------------------------
# guess_at_SendmailCmd
#------------------------------------------------------------
sub guess_at_SendmailCmd {
    my ($dir, $name);
    my $guess;

    foreach $name (qw( sendmail smail )) {
	foreach $dir (qw( /usr/lib /usr/bin /sbin /usr/sbin )) {
	    if (-x "$dir/$name") {
		$guess = "$dir/$name";
		goto found;
	    }
	}
    }
    return undef;

  found:
    $guess .= " -t -i -oeq";
    debug "Guessed at SendmailCmd: $guess";
    return $guess;
}

#------------------------------------------------------------
# guess_at_WwwUser
#------------------------------------------------------------
sub guess_at_WwwUser {
    if ($SuspectCgiUser && $SuspectCgiGroup) {
	return "$SuspectCgiUser.$SuspectCgiGroup";
    }
}

#------------------------------------------------------------
# guess_at
#------------------------------------------------------------
sub guess_at {
    my $var = shift;
    my $val;

    # Try to call "guess" routine:
    local(*stab) = *{"main::"};
    local(*funcglob) = $stab{"guess_at_$var"};
    if (defined(&funcglob)) {	
	debug "Calling 'guess' function for $var...";
	$val = &funcglob();
    }
    return $val if defined($val);
    
    # Try to use "fallback" values:
    debug "Trying 'fallback' for $var...";
    $val = $Fallback{$var};    
    return $val if defined($val);

    # Warn:
    warning "Couldn't make a good guess at $var";
    return ''
}

#------------------------------------------------------------
# guess_setup -- guess the entire setup
#------------------------------------------------------------
sub guess_setup {
    my ($setup) = @_;

    # Try to get info from NCSA HTTPD's srm.conf:
    try_to_find_ncsa_httpd();
    try_to_scan_srm_conf();
    try_to_scan_httpd_conf();
    try_to_find_server_cgi_bin();
    try_to_find_server_docs();

    # Go through all the vars:
    my $var;
    foreach $var (sort keys %Fallback) {
	if (!defined($setup->{$var})) {
	    debug "have to guess $var";
	    $setup->{$var} = guess_at($var);
	} 
    }
    1;
}

#------------------------------------------------------------
#
# For "setup -apply"...
#
#------------------------------------------------------------

#------------------------------------------------------------
# write_setup_sh -- write the .sh setup file
#------------------------------------------------------------
sub write_setup_sh
{
    my ($setup) = @_;

    my $targetfile = "lib/bin/SETUP.sh";
    unlink $targetfile;
    open(SETUP, ">$targetfile") or puke "couldn't create $targetfile: $!";
    print SETUP <<EOF;
#------------------------------------------------------------
# SETUP.sh
#------------------------------------------------------------
#
# DESCRIPTION
#    Current IMS/www setup, predigested for Bourne shell scripts.
#    Contains mostly the info in SETUP.
#
# WARNING
#    DO NOT EDIT THIS FILE!  Your changes will be lost when 
#    setup is re-run.  Edit the SETUP file instead,
#    and re-run "setup.
#

EOF

    foreach $varname (sort keys %$setup) {
	my $upvarname = ('IW_' . IU::Util::mixed_to_caps($varname));
	print SETUP $upvarname, '=';
	print SETUP '"', IU::Escape::sh($setup->{$varname}), '"', "\n";
    }
    print SETUP <<EOF;

#-------------------- end of file --------------------
EOF
    close(SETUP);
    chmod $targetfile;
    inform "Wrote $targetfile.";
    1;
}

#------------------------------------------------------------
# write_setup_pl -- write the .pm setup file
#------------------------------------------------------------
sub write_setup_pl
{
    my ($setup) = @_;

    my $targetfile = "lib/perl/IU/SETUP.pm";
    unlink $targetfile;
    open(SETUP, ">$targetfile") or puke("couldn't create $targetfile: $!");
    print SETUP <<EOF;
#------------------------------------------------------------
package IU::SETUP;

=head1 NAME

IU::SETUP - IMS/WWW setup parameters, in Perl form

=head1 DESCRIPTION

Current IWS/www setup, predigested for Perl.
Contains mostly the info in SETUP.

=head1 WARNINGS

If you are writing an IW module, do I<not> "use" this file directly...  
It will corrupt your namespace!  It is loaded into the IU:: package; 
you should access these values through \$IU::VarName.

B<DO NOT EDIT THIS FILE!>  Your changes will be lost when setup is re-run.  
Edit the top-level SETUP file instead, and re-run "setup.

=cut

require Exporter;
\@ISA = qw(Exporter);

EOF

    # The @EXPORT list:
    print SETUP '@EXPORT = qw($',
        join(' $', sort keys %$setup), ");\n\n";

    # The variables themselves:
    foreach $varname (sort keys %$setup) {
	printf SETUP "\$%s = \n      ", $varname;
	print  SETUP "'", $setup->{$varname}, "';\n"; 
    }

    print SETUP <<EOF;

#-------------------- end of file --------------------
EOF
    close(SETUP);
    inform "Wrote $targetfile.";
    1;
}

#------------------------------------------------------------
# write_setup_dat -- write the SETUP.dat file
#------------------------------------------------------------
sub write_setup_dat
{
    my ($setup) = @_;

    my $targetfile = "lib/conf/SETUP.dat";
    unlink $targetfile;
    open(SETUP, ">$targetfile") or puke "couldn't create $targetfile: $!";
    print SETUP <<EOF;
#------------------------------------------------------------
# SETUP.dat
#------------------------------------------------------------
#
# DESCRIPTION
#    Current IWS/www setup, predigested to be parseable by C programs.
#    Contains mostly the info in SETUP.
#
# FORMAT
#    Lines beginning with # and blank lines are ignored.
#    Other lines should be in the form
#       <varname><one space><varvalue><return>
#
# WARNINGS
#    DO NOT EDIT THIS FILE!  Your changes will be lost when 
#    setup is re-run.  Edit the SETUP file instead,
#    and re-run "setup".
#

EOF
    foreach $varname (sort keys %$setup) {
	print SETUP $varname, " ", $setup->{$varname}, "\n";
    }
    print SETUP <<EOF;

#-------------------- end of file --------------------
EOF
    close(SETUP);
    inform "Wrote $targetfile.";
    1;
}

#------------------------------------------------------------
#
# Main...
#
#------------------------------------------------------------

#------------------------------------------------------------
# do_makefrom
#------------------------------------------------------------
sub do_makefrom {
    my ($file) = @_;

    # Warn them:
    if (-e "SETUP") {
	unless (IU::Ask::yes_no(
	    "This will overwrite your SETUP.  Proceed?")) {
	    inform "Okay, bye...";
	    exit(-1);
	}
    }

    # Load the existing setup:
    my $setup = IU::SetupFile->read($file);
    $setup or puke("couldn't parse $file");
    
    # Add values:
    guess_setup($setup);

    # Output:
    write_setup($setup) or puke "couldn't output";
    1;
}

#------------------------------------------------------------
# do_make
#------------------------------------------------------------
sub do_make {
    my %setup = ();

    # Warn them:
    if (-e "SETUP") {
	unless (IU::Ask::yes_no(
	    "This will overwrite your SETUP.  Proceed?")) {
	    inform "Okay, bye...";
	    exit(-1);
	}
    }

    # Create a new setup:
    my $setup = IU::SetupFile->new;

    # Guess values:
    guess_setup($setup);

    # Output:
    write_setup($setup) or puke "couldn't output";
    1;
}

#------------------------------------------------------------
# do_apply
#------------------------------------------------------------
sub do_apply {
    
    # Warn them:
    unless (IU::Ask::yes_no(
        "This will overwrite some files.  Proceed?")) {
	inform "Okay, bye...";
	exit(-1);
    }
    
    # Load the setup:
    my $setup = IU::SetupFile->read('SETUP');
    $setup or puke("couldn't parse SETUP");

    ### DERIVE SOME INFO...

    # Add directories...
    $setup->set('TopDir',   $::Top);
    $setup->set('DocsDir',  "$::Top/docs");
    $setup->set('LibDir',   "$::Top/lib");
    $setup->set('LogDir',   "$::Top/log");
    $setup->set('Templates',"$::Top/docs/templates"); 

    # Force the hostname...
    if (! $setup->{HttpdHost}) {
	$setup->{HttpdHost} = guess_at_HttpdHost ||
	    die "I can't figure out my HttpdHost: please, " .
		"fill in a value!";
    }

    # Add the welcome page, ONLY if it wasn't defined... 
    if (! $setup->{WelcomeUrl}) {
	$setup->set('WelcomeUrl',
		    "$setup->{DocsUrl}imswelcome/index.html");
    }

    # Add how to get in touch with the site administrator...
    my $email = $setup->{SiteAdmin};
    $email .= '@'. $setup->{HttpdHost}   unless ($email =~ /\@/);
    $setup->set('SiteAdminEmail', $email);

    # Add the version number.  It's too dangerous to put this in the
    # normal SETUP file; someone might obliterate it by blindly copying
    # over an existing SETUP file:    
    $setup->set('Version', $IU::Version::VERSION);

    # Add the system name. We protect this is a manner similar to
    # the version number, for much the same reasons.
    $setup->set('SystemName', $IU::SystemName::SYSTEM_NAME);

    # Generate the program-usable setup files...
    write_setup_pl($setup);     # for Perl programs
    write_setup_dat($setup);    # for C programs (read in at run time)
    write_setup_sh($setup);     # for shell scripts

    inform "Running makewelcome.";
    system ("$::Top/lib/perl/makewelcome.pm");

    inform "Running makePersistent.";
    system ("$::Top/lib/perl/makePersistent.pm");

    if($setup->{MapCgiScript} eq 'drawglobe.cgi') {
	unlink $::Top."/cgi-bin/map" if (-l $::Top."/cgi-bin/map");
	unlink $::Top."/cgi-bin/map_symbols" 
	 if (-l $::Top."/cgi-bin/map_symbols");
        unless(-e $::Top."/cgi-bin/map") {
            inform "Creating link to glismapper map.";
            symlink($setup->{GaeaDistribDir}."/map",$::Top."/cgi-bin/map");
        }

        unless(-e $::Top."/cgi-bin/map_symbols") {
            inform "Creating link to glismapper map_symbols.";
            symlink($setup->{GaeaDistribDir}."/map_symbols",
                    $::Top."/cgi-bin/map_symbols");
        }
    }

    inform "Compiling wwwvalids.conf. ";
    system("$::Top/lib/perl/compwwwvalids.pm");

    # update time stamp on lib/conf/navigator.conf so it will recompile 
    # with any new setup variable changes:

    my $now = time;
    utime $now, $now, "$::Top/lib/conf/navigator.conf";


    # Done!
    inform "Done!";
    1;    
}

#------------------------------------------------------------
# usage
#------------------------------------------------------------
sub usage {
    print STDERR <<EOF;
Usage: 
    setup -make
        Make a new SETUP file, initialized by making intelligent
        guesses at your site configuration.

    setup -makefrom <file>
        Make a new SETUP file, initialized from an existing one.
        Useful for upgrades, where you want to "inherit" as many values
        from an existing SETUP file as you can.

    setup -apply
        Apply the contents of the SETUP file.
        This will generate needed files.

EOF
    1;
}

#------------------------------------------------------------
# main
#------------------------------------------------------------
sub main {

    # Hello:
    IU::Log::init_term('STDOUT');
    IU::Util::banner("WELCOME TO 'SETUP', FOR IMS/WWW.");

    # Make sure we're where we should be:
    $TOP = &getcwd();
    if (! -e "$TOP/setup") {
	puke(
	"You must be in the top-level IMS/www directory to do this!");
    }

    # Handle args:
    if ($ARGV[0] eq '-make') {
	do_make() || puke("couldn't make from setup file");
    }
    elsif ($ARGV[0] eq '-makefrom') {
	do_makefrom($ARGV[1]) ||
	    puke("couldn't make from file $ARGV[1]");
    }
    elsif ($ARGV[0] eq '-apply') {
	do_apply() || puke("couldn't make from setup file");
    }
    else {
	usage();
	exit(-2);
    }
    1;
}
exit(&main ? 0 : -1);

1;
#-------------------------- end of file ------------------------------
