/****************************************************************************/
/*                                 JULDAT.C                                 */ 
/* Author : Erich Franz Stocker                                             */
/* Date:    09/14/87                                                        */
/* System:  Master directory                                                */
/*--------------------------------------------------------------------------*/
/* This library contains date oriented routines. These include julian date  */
/* conversion  and reconversion routines.  Routines to check the validity   */
/* of dates and routines which deal with the calculations of days of the    */
/* week.                                                                    */
/****************************************************************************/

#include <ctype.h>
#include <string.h>
#include "juldat.h"          /* contains DATEINT structure type & */
                             /* DATESIZE constant.                */

/*****************************************************************************/
/* Julian is a float (because of julian date size) function which returns a  */
/* specially normalized julian date calculated from 3 integers representing  */
/* month, day, and year.                                                     */
/*                                                                           */
/* This normalized julian date is calculated starting on Jan 1, 0001         */
/* instead of the usual January 1, 4713 B.C.   Also, instead of ending on    */
/* December 31, 3267, this function is not valid after November 6, 2738.     */
/*                                                                           */
/* To obtain the actual julian date one need add only 1,721,409 to the date  */
/* this routine returns.                                                     */
/*****************************************************************************/
float julian(IntDate)
DATEINT *IntDate; 
{
   float calcjulian;  
   int month,day,year;
 
   month=IntDate->mo;
   day=IntDate->day;  
   year=IntDate->year;
   calcjulian=((int)(30.57 * month)) + ((int)(365.25 * year - 395.25)) + day;
   if (month > 2){ 
      if ((year % 4) == 0 && year % 100 !=0 ||year % 400 == 0)
         calcjulian--; 
      else
         calcjulian=calcjulian-2;
   }
   return calcjulian;  
}
                    
/**************************************************************************/
/* jul_to_date takes a normalized julian date and converts it to a date   */
/* string.  Each element of the date is separated by a hypen. The string  */
/* passed to the routine must of sufficient size to hold all characters   */
/* including the null (i.e. 11 characters) or incorrect string is returned*/
/* Routine also ensures that the date string returned is always 10 char.  */
/* When days or months are less than 10 the zero will have leading zeros  */
/* inserted before being returned.                                        */
/**************************************************************************/
void jul_to_date(julday,str_date)
float julday;
char *str_date;
{
   int month,day,year;
   int d1;

   year =  (int)(julday / 365.25) + 1;
   day = julday + (int)(395.25 - 365.25 * year);
   if ((year % 4) == 0 && year % 100 != 0 || year % 400 == 0) 
      d1 = 1;
   else
      d1 = 2;
   if (day > (91-d1))
      day = day+d1;
   month=(int)(day / 30.57);
   day = day - (int)(30.57*month);
   if(month > 12 ) {
      month=1;
      year++;
   }       
    sprintf(str_date,"%4d-%2d-%2d",year,month,day);   
    if(str_date[5]==' ')
       str_date[5]='0';
    if(str_date[8]==' ')
       str_date[8]='0';    
    if(str_date[0]==' ')
       str_date[0]='0';
    if(str_date[1]==' ')
       str_date[1]='0';
    if(str_date[2]==' ')
       str_date[2]='0';
   
}

/*************************************************************************/
/* day_of_week takes the month, day and year sent to it and returns the  */
/* name of the day of the week.  Zeller's congruence algorithm is used   */
/* in calculating the day of the week.  Numbers equate to the following  */
/*                                                                       */
/*                      0=Sunday                                         */
/*                      1=Monday                                         */
/*                      2=Tuesday                                        */
/*                      3=Wednesday                                      */
/*                      4=Thursday                                       */
/*                      5=Friday                                         */
/*                      6=Saturday                                       */
/*                                                                       */
/* Routine actually returns a pointer to the pointer in the day name     */
/* vector declared in this environment as a static data structure.       */
/*************************************************************************/
char *day_of_week(IntDate)
DATEINT *IntDate;
{      
   static char *name[]= {"Sunday","Monday","Tuesday",
                         "Wednesday","Thursday","Friday","Saturday",
                         "Invalid"};
                 
   int month,day,year,century,dow;
   int n;
     
   month=IntDate->mo;
   day=IntDate->day;
   year=IntDate->year;      
   if (month > 2) 
      month=month-2;
   else {
      month=month+10;
      --year;
   }
   century=year / 100;
   year=year % 100;
   n = (int)((13*month - 1) / 5) + day + year + (int)(year / 4) +
       (int)(century / 4) - 2 * century +77;
   dow = n - 7 * (int)(n / 7);
   return ((dow<0 ||dow>6) ? name[7] : name[dow]);
}

/*************************************************************************/
/* ConvertToNum takes a date which contains the month in the standard 3  */
/* character abbreviation and converts the month into the numeric equi-  */
/* valent with the necessary leading zeros.                              */
/*************************************************************************/
int ConvertToNum(month)
char *month;
{ 
   char temp[6];
   int ch;
   
   strcpy(temp,month);
   for(ch=0;ch<=strlen(temp);ch++){
      if(isalpha(temp[ch]))
         if(islower(temp[ch]))
           temp[ch]= toupper(temp[ch]);
           /* temp[ch]=_toupper(temp[ch]);   */
   }
   if(strcmp(temp,"JAN")==0)
      strcpy(month,"01");
   else if(strcmp(temp,"FEB")==0)
      strcpy(month,"02");
   else if(strcmp(temp,"MAR")==0)
      strcpy(month,"03");
   else if(strcmp(temp,"APR")==0)
      strcpy(month,"04"); 
   else if(strcmp(temp,"MAY")==0)
      strcpy(month,"05");
   else if(strcmp(temp,"JUN")==0)
      strcpy(month,"06");
   else if(strcmp(temp,"JUL")==0)
      strcpy(month,"07");
   else if(strcmp(temp,"AUG")==0)
      strcpy(month,"08");
   else if(strcmp(temp,"SEP")==0)
      strcpy(month,"09");
   else if(strcmp(temp,"OCT")==0)
      strcpy(month,"10");
   else if(strcmp(temp,"NOV")==0)
      strcpy(month,"11");
   else if(strcmp(temp,"DEC")==0)
      strcpy(month,"12");
   else
      return(0);         
   return(1);
}

/************************************************************************/
/* ISO_date returns a 0 when the date sent into the routine can not be  */
/* converted into an ISO date format (YYYY-MM-DD). Otherwise, it takes  */
/* any valid date including ones with 3 character month and converts to */
/* the ISO format.  All ISO date strings will be padded with leading    */
/* zeros when the day or the month equate to a value less than 10.      */
/* The following date formats can be converted by the routine:          */
/*                                                                      */
/* 1. All numeric values in the string (except for separators):         */
/*    a.  YYYY-MM-DD                                                    */
/*    b.  MM-DD-YYYY                                                    */
/*        (note: DD-MM-YYYY is not possible since routine will not      */
/*              know which numbers are month or day. So form b always   */
/*              assumed.)                                               */
/*                                                                      */
/* 2. Any form where the month is sent in as 3 char alpha. Any lower    */
/*    case is converted to upper before processing.                     */
/*    a.  MMM DD YYYY                                                   */
/*    b.  MMM DD YYYY                                                   */
/*    c.  DD MMM YYYY                                                   */
/*    d.  DD YYYY MMM                                                   */
/*    e.  YYYY MMM DD                                                   */
/*    f.  YYYY DD MMM                                                   */
/*                                                                      */
/* The routine accepts blank,'/', '-' as valid separators between date  */
/* date parts. However, all are converted to the ISO standard '-'.      */
/************************************************************************/
int ISO_date(datein,dateout)
char *datein,*dateout;
{ 
   int digits=0,ch,place=0;
   char temp[6],temp1[6],temp2[6],temp3[6];  
   char tempmo[6],tempyr[6],tempday[6];
 
   for(ch=0;ch<=strlen(datein)&&ch<=DATESIZE;ch++){  
      if(datein[ch]=='/'||datein[ch]=='-'||datein[ch]==' '||
                                         datein[ch]=='\0') { 
         temp[digits]='\0';
         digits=0;
         switch (place){
            case 0: strcpy(temp1,temp);
                    break;
            case 1: strcpy(temp2,temp);
                    break;
            case 2: strcpy(temp3,temp);
                    break;
            default:return (0); 
         }    
         place++;
      }  
      else
         temp[digits++]=datein[ch];
   }
   if(strlen(temp1)==4){ 
      strcpy(tempyr,temp1);
      if(strlen(temp2)==3){ 
         strcpy(tempmo,temp2);
         strcpy(tempday,temp3);
      }
      else if(strlen(temp3)==3){  
         strcpy(tempmo,temp3);
         strcpy(tempday,temp2);
      }
      else{                      /* no alpha char so ISO format */
         strcpy(tempmo,temp2);
         strcpy(tempday,temp3);
      }
   }
   else if(strlen(temp2)==4){   
      strcpy(tempyr,temp2);
      if(strlen(temp1)==3){   
         strcpy(tempmo,temp1);
         strcpy(tempday,temp3);
      }
      else if(strlen(temp3)==3){ 
         strcpy(tempmo,temp3);
         strcpy(tempday,temp1);
      }
      else                      /*if no alpha year can't be second */
         return(0);
   }
   else if(strlen(temp3)==4){ 
      strcpy(tempyr,temp3);
      if(strlen(temp1)==3){  
         strcpy(tempmo,temp1);
         strcpy(tempday,temp2);
      }
      else if(strlen(temp2)==3){
         strcpy(tempmo,temp2);
         strcpy(tempday,temp1);
      }
      else{                     /* no alpha so in MM-DD-YYYY style */  
         strcpy(tempmo,temp1);
         strcpy(tempday,temp2);
      }
   }
   else                        /* no year entered or year < 4 */
      return(0);               
   if(strlen(tempmo)==3)
      if(!ConvertToNum(tempmo)) 
         return(0);
   strcpy(dateout,tempyr);
   strcat(dateout,"-");
   if(strlen(tempmo)<2)
      strcat(dateout,"0");
   strcat(dateout,tempmo);
   strcat(dateout,"-");
   if(strlen(tempday)<2)
      strcat(dateout,"0");
   strcat(dateout,tempday);
   return(1);
}

/***************************************************************************/
/* valid_date checks a  character string sent to it to determine whether a */
/* valid month, day and year are included.  The string must contain only   */
/* the date and not the time. The year must contain the full four digits & */
/* the elements of the date must be separated by a blank, /, or - .        */
/* The date sent in must have a null character appended at the end.        */
/* If the date is valid the routine returns 1 else it returns 0.           */
/*                                                                         */
/* The following invalid types of dates are detected:                      */
/*          1) Date contains alpha characters not standard                 */
/*             abbreviation characters for month.                          */
/*          2) Month is less than 1 or greater than 12.                    */
/*          3) Day exceeds the valid number of days for the month          */
/*             in question.                                                */
/*          4) 02/29 listed for a year other than a leap year.             */
/*                                                                         */
/*          5) A Day comes in with zero for the day.                       */
/* As a side-effect of the check for validity, the function returns the    */
/* month, day, and year as integers in the DATEINT structure. If the date  */
/* is invalid the structure value is undefined.                            */
/*                                                                         */
/* An additional side-effect if the date is valid is the conversion of the */
/* input datestring into an ISO standard date string.  Users should always */
/* use the ISO standard date string for further processing.                */
/***************************************************************************/
int valid_date(datein,DateOut,isodat)
char *datein; 
DATEINT *DateOut;
char *isodat; 
{                    
   int ch,place=0,tempno,digits=0,tempmo,tempday,tempyr;        
   static int days[]={0,31,29,31,30,31,30,31,31,30,31,30,31}; 
   static char temp[6]=" ";
   char hdate[DATESIZE],indate[DATESIZE]; 
                                                    
   if(!ISO_date(datein,hdate))
      return(0);
   else
      strcpy(isodat,hdate);

   for(ch=0;place<3&&ch<=strlen(isodat);ch++){  
      if (isodat[ch]==' '||isodat[ch]=='-'||isodat[ch]=='/'||
                               isodat[ch]=='\0') { 
         temp[digits]='\0';
         digits=0;
         tempno=atoi(temp); 
         if (place==0)
            tempyr=tempno;
         else if (place==1)
            tempmo=tempno;
         else if(place==2)
            tempday=tempno;             
         else
            return 0;
         place++;
      }          
      else {                                   
         if (isdigit(isodat[ch]))
            temp[digits++]=isodat[ch];   
         else
            return 0;
      }
   }                       
   if(tempmo>12||tempmo<1)
      return 0;
   else if (tempday == 0)
      return 0;
   else if(tempmo==2) {
      if((tempyr % 4) != 0){   /*year is not a potential for leap year */
         if(tempday>28)
            return(0);
      } 
      else{                    /*year is candidate for leap year */
         if ((tempyr % 100) == 0 && (tempyr % 400) != 0) {
            if(tempday>28)     /* years divisible by 100 but not 400*/
               return 0;       /* are not leap years */
         }
         else{                 /* years divisible by 100 are leap years */
           if(tempday> days[tempmo])
              return 0;
         }
      }
   }                  
   else {
      if (tempday>days[tempmo])
         return 0;  
   }             
   if(tempyr<1||tempyr>2738) 
       return 0;     
   DateOut->mo=tempmo;
   DateOut->day=tempday;
   DateOut->year=tempyr;
   return 1;
}

/************************************************************************/
/* juldif returns the absolute days difference between two normalized   */
/* julian dates. The difference returned is always the absolute value of*/
/* the calculated difference of julday1 and julday2.                    */
/************************************************************************/
float juldif(julday1,julday2)
float julday1,julday2;
{   
   if (julday1>julday2)
      return (julday1-julday2);
   else
      return (julday2-julday1);
}                              

/*************************************************************************/
/* stopGTstart checks to determine whether its first argument, the start */
/* date is less than or equal to the stop date. Results are the following*/
/*                                                                       */
/*                      0: startdate = stopdate                          */
/*                      1: startdate < stopdate                          */ 
/*                     -1: startdate > stopdate                          */
/*                   -998: startdate invalid                             */
/*                   -999: stopdate  invalid                             */
/*************************************************************************/
int stopGTstart(startdate,stopdate)
char *startdate,*stopdate;
{   
   DATEINT start,stop;
   float fltstart,fltstop;                   
   char isodat[DATESIZE];       /* not needed for the actual check */
                                /* but necessary to capture the    */
                                /* side-effect of valid_date.      */
   
   if (valid_date(startdate,&start,isodat)) 
      if (valid_date(stopdate,&stop,isodat)) {  
         fltstart=julian(&start);
         fltstop=julian(&stop);
         if (fltstart==fltstop)
            return (0);
         else if (fltstart>fltstop)
            return (-1);
         else
            return (1);   
      }    
      else
         return(-999);
   else
      return (-998);
}                      

/***********************************************************************/
/* datedif compares two date strings to determine the absolute number  */
/* of days separating them.  It first checks for the validity of the   */
/* which also has the sideeffects of filling the integer date structure*/
/* which is then sent to the julian routine to get a normalized julian */
/* date for each of the string dates. These julian dates are then used */
/* to call juldif which returns the absolute difference.               */
/*                                                                     */
/* In the event that a string date is invalid, datedif returns a -1.   */
/***********************************************************************/
float datedif(date1,date2)
char *date1,*date2;
{        
   DATEINT d1,d2;
   char iso1[DATESIZE],iso2[DATESIZE];
   float j1,j2,dif;
        
   if(valid_date(date1,&d1,iso1)){  
      j1=julian(&d1);  
      if(valid_date(date2,&d2,iso2)){ 
      j2=julian(&d2); 
      dif=juldif(j1,j2);
      return dif;
      }
      else
         return(-1);   
   }
   else
      return(-1);
}


void StripChar(instr,outstr,len)
char *instr,*outstr;
int len;
{
   strncpy(outstr,instr,len);
   outstr[len] = 0;
} 

/*******************************************************************/
/* timediff takes two time strings in the required format HH:MM:SS */
/* and calculates the difference between the two times in hours,min*/
/* and sec. It then returns this difference in the format:         */
/*                         99H 99M 99S                             */
/*******************************************************************/
int timediff(starttime,endtime,diffstr,hr,min,sec)
char *starttime,*endtime,*diffstr;               
int *hr,*min,*sec;
{ 
   char stmphr[5],stmpmin[5],stmpsec[5];
   char etmphr[5],etmpmin[5],etmpsec[5];
   char tempstr[255];
   int shr,smin,ssec,stime;
   int ehr,emin,esec,etime;
   int diff,i;
                    
   if (strlen(starttime) < 8 || strlen(endtime) < 8)
      return(0);  
   if(starttime[2] != ':' && starttime[5] != ':')
      return(0);
   if(endtime[2] != ':' && endtime[5] != ':')
      return(0);              

   StripChar(starttime,stmphr,2);
   StripChar(starttime+3,stmpmin,2);
   StripChar(starttime+6,stmpsec,2);
   StripChar(endtime,etmphr,2);
   StripChar(endtime+3,etmpmin,2);
   StripChar(endtime+6,etmpsec,2); 
   shr = atoi(stmphr) * 3600;
   smin = atoi(stmpmin) * 60;
   ssec = atoi(stmpsec);
   stime = shr+smin+ssec;
   ehr = atoi(etmphr) * 3600;
   emin = atoi(etmpmin) * 60;
   esec = atoi(etmpsec);
   etime = ehr+emin+esec;  
   diff = etime - stime;
   if(diff < 0)
      return(0);
   shr = diff / 3600;
   ehr = diff % 3600;
   smin = ehr / 60;
   emin = ehr % 60;
   ssec = emin;
   strcpy(diffstr,"\0\0\0\0\0\0\0\0\0"); 
   if(shr == 0)
      strcpy(stmphr,"   ");
   else
      sprintf(stmphr,"%2.2dH",shr);
   if(smin == 0)
      strcpy(stmpmin,"   ");
   else
      sprintf(stmpmin,"%2.2dM",smin);
   if(ssec == 0) 
      strcpy(stmpsec,"   ");
   else
      sprintf(stmpsec,"%2.2dS",ssec);

   sprintf(diffstr,"%s %s %s",stmphr,stmpmin,stmpsec);
   *hr = shr;
   *min = smin;
   *sec = ssec;
   return(1);
}
