/*********************************************************************/
/*  Module Name: md_connect.c                                        */
/*  System: NASA Master Directory                                    */
/*  Author: Connie Y. Li                                             */
/*  Creation Date:  08/01/91                                         */
/*-------------------------------------------------------------------*/
/*  Description:                                                     */
/*   Network_Request - This function is used to interact with a MD   */
/*   server using network sockets.                                   */
/*                                                                   */
/*   Direct_Request - This function is used to interact with a MD    */
/*   server using the traditional way.                               */
/*                                                                   */
/*-------------------------------------------------------------------*/
/*  Imports:                                                         */
/*    inet.h                                                         */
/*    mdcodes.h                                                      */
/*    mddefs.h                                                       */
/*    ParseFetch()                                                   */
/*-------------------------------------------------------------------*/
/*  Exports: Network_Request                                         */
/*-------------------------------------------------------------------*/
/*  Restrictions:                                                    */
/*  Global symbols needed for this module:                           */
/*     SERVHOSTADDR -- server host IP address to connect to.         */
/*     SERVTCPPORT  -- database port number that will be listening.  */
/*     LOGTCPPORT  -- log port number that will be listening.        */
/*     LINKTCPPORT  -- link port number that will be listening.      */
/* ------------------------------------------------------------------*/   
/*  Return values:                                                   */
/*    Nework_Request - MD database return code that is sent back     */
/*                     by the server after the query is issued.      */
/*                     Or, the network error code if a connection,   */
/*                     read or write operation fails.                */
/*-------------------------------------------------------------------*/
/* Comments:                                                         */
/* last modified date:  7/20/92                                      */
/*********************************************************************/
#include <stdio.h>
#include <signal.h>
#include "inet.h"
#include "mddefs.h"
#include "mdcodes.h"
#include "strutils.h"

/*--------------------------------------------------------------------*/
static int dbserver_init=FALSE; 
static int logserver_init=FALSE; 
static int dbsock;
static int logsock;
static int linksock;
static int mode;
static int batchwrite=FALSE;
static char openstmt[DBLINESIZE];
/*--------------------------------------------------------------------*/

int Network_Request (qrybuf, resultbuf, bufsize, server)
char *qrybuf;
char *resultbuf;
int  bufsize;      /* This flag is ignored by this function */
int  server;       /* The server number we are requesting service */
{

   switch (server) {
      case DBSERVER:
         return(Run_DBServer(qrybuf, resultbuf, server));
         break;
      case LOGSERVER:
         return(Run_LogServer(qrybuf, resultbuf, server));
         break;
      case LINKSERVER:
         return(Run_LinkServer(qrybuf, resultbuf, server));
         break;
   }
}


/*-----------------------------------------------------------------*/
/*  Close the socket relating to a server and reset flags...       */
/*-----------------------------------------------------------------*/
void CloseMDSocket(server)
int  server;       /* The server number we are requesting service */
{

   switch (server) {
      case DBSERVER:
         close_connection(dbsock);
         dbserver_init = FALSE;
         break;
      case LOGSERVER:
         close_connection(logsock);
         logserver_init = FALSE;
         break;
   }
}

/*-----------------------------------------------------------------*/
/* Return Codes: network return codes on network errors, otherwise */
/*               database return codes.  See mdcodes.h             */
/*-----------------------------------------------------------------*/
int Run_DBServer(qrybuf, resultbuf, server) 
char *qrybuf;
char *resultbuf;
int server;
{
   int lenrtn,   /* length of the return buffer read from server */
       status;   /* status of a write to the server.             */
   char *sptr;
   char errormsg[DBWORDSIZE];
   int md_error; /* captures the error returned by the server */

   if (!dbserver_init)   
      return(init_connect(qrybuf, resultbuf, server));
   else {
      /*----------------------------------------------------------*/
      /* Network connection established... Now perform write/read */
      /* based on the mode the server was opened initially.       */
      /*----------------------------------------------------------*/
      /* batch mode: write query statement once after the server  */
      /* has been up.  Only read from the server after the query  */
      /* is written to the server in the subsequent call.         */
      /*----------------------------------------------------------*/
      if (mode == BATCH) {
         if (!batchwrite) {
            if ((status=write_message(dbsock,qrybuf,strlen(qrybuf))) == -1) {
               close_connection(dbsock);
               return(MD_WRITERR);
            }
            batchwrite = TRUE;
         }
      }
      /*----------------------------------------------------------*/
      /* interactive mode: always start with a write to server.   */ 
      /*----------------------------------------------------------*/ 
      else { 
         if ((status=write_message(dbsock,qrybuf,strlen(qrybuf))) == -1) { 
            close_connection(dbsock); 
            return(MD_WRITERR);
         }
      }

      lenrtn = get_message(dbsock, resultbuf, CLIENT);
      if (!valid_read(lenrtn)) {
         close_connection(dbsock);
         return(lenrtn);
      }
      else {
         /*-----------------------------------------------------*/
         /* Read and write operations done successfully.        */
         /* Now just return the server database return code.    */
         /*-----------------------------------------------------*/
         resultbuf[lenrtn] = 0;
         sptr = ParseString(resultbuf, errormsg, "~");
         md_error = atoi(errormsg);

         /*-----------------------------------------------------*/
         /* read finishes... reset batchwrite flag for next     */
         /* batch SQL call.                                     */
         /*-----------------------------------------------------*/
         if (md_error != MD_NORMAL)
            batchwrite = FALSE;

         return(md_error);
      }
   }
}

/*-----------------------------------------------------------------*/
int Run_LogServer(qrybuf, resultbuf, server)
char *qrybuf;
char *resultbuf;
int server;
{
   int  status;   /* status of a write to the server.             */
   char *sptr, token[DBLINESIZE];

   if (!logserver_init)
      return(init_connect(qrybuf, resultbuf, server));
   else {
      /*----------------------------------------------------------*/
      /* Network connection established...                        */
      /* Client will write to the log server without waiting for  */
      /* additional information sent back by server.....          */
      /*----------------------------------------------------------*/
      if ((status=write_message(logsock,qrybuf,strlen(qrybuf))) == -1) {
         close_connection(logsock);
         return(MD_WRITERR);
      }
      /* reset flag for reconnecting to log server if a close was issued */
      sptr = ParseString(qrybuf, token, "~");
      chgupper(token);
      if (!strcmp(token, "CLOSE"))
         logserver_init=FALSE;
   }
   return(1);  /* no problem connecting and writing to socket */
}

/*-----------------------------------------------------------------*/
int Run_LinkServer(qrybuf, resultbuf, server)
char *qrybuf;
char *resultbuf;
int server;
{
   return(init_connect(qrybuf, resultbuf, server));
}

/*---------------------------------------------------------------------*/
/* Purpose: initiate a connect to the MD server with the previous open */
/* statement.                                                          */
/* return codes: MD_NORMAL, if connection to the MD server succeeded.  */
/*               Otherwise, network error code indicating the reason   */
/*                     for failure.  See mdcodes.h                     */
/*---------------------------------------------------------------------*/
int init_connect(qrybuf, resultbuf, server) 
char *qrybuf;
char *resultbuf;
int server;
{
   char addr[DBWORDSIZE], port[DBWORDSIZE], portname[DBWORDSIZE];
   int portno;
   int status;

   switch (server) {
      case 1: sprintf(portname, "SERVTCPPORT"); break;
      case 2: sprintf(portname, "LOGTCPPORT"); break;
      case 3: sprintf(portname, "LINKTCPPORT"); break;
   }

   trn_symbol("SERVHOSTADDR", addr);
   trn_symbol(portname, port);
   portno = atoi(port);

   
   switch (server) {
      case 1:  dbsock = call_MDserver(addr, portno); 
               if (dbsock == -1)
                  return(MD_CONNECTERR);
               else
                  return(SendDB_Query(qrybuf, resultbuf));
               break;
      case 2:  logsock = call_MDserver(addr, portno); 
               if (logsock == -1)
                  return(MD_CONNECTERR);
               else { 
                  /* just write the first statement to log server */
                  if ((status=write_message(logsock, qrybuf, 
                       strlen(qrybuf))) < 0) { 
                     close_connection(dbsock);
                     return(MD_WRITERR);
                  }
               }
               logserver_init = TRUE;
               break;
      case 3:  linksock = call_MDserver(addr, portno); 
               if (linksock == -1)
                  return(MD_CONNECTERR);
               break;
   }
   return(1);
}

/*---------------------------------------------------------------*/
SendDB_Query(qrybuf, resultbuf)
char *qrybuf, 
     *resultbuf;
{
   int lenrtn,   /* length of the return buffer read from server */
       status;   /* status of a write to the server.             */
   char *sptr;
   char errormsg[DBWORDSIZE];
   int md_error; /* captures the error returned by the server */
   char junk[MAXBUFSIZE];
   int junkchannel;

   /*------------------------------------------------------------*/
   /* Obtain the mode of the server when starting up the server  */
   /* and try to write/read once to the server.                  */
   /*------------------------------------------------------------*/
   ParseQuery(qrybuf, &junkchannel, junk, &mode);
   if ((status=write_message(dbsock, qrybuf, strlen(qrybuf))) < 0) { 
      printf("init_connect: error!!  Can't write to server..exiting\n");
      close_connection(dbsock);
      return(MD_WRITERR);
   }
   lenrtn = get_message(dbsock, resultbuf, CLIENT);
   if (!valid_read(lenrtn))
      return(lenrtn);
   else {
      resultbuf[lenrtn] = 0;
      sptr = ParseString(resultbuf, errormsg, "~");
      md_error = atoi(errormsg);

      dbserver_init = TRUE;
      return(md_error);
   }
}

/*----------------------------------------------------------------*/
/* Purpose: reissue the connect procedure to the server using the */
/*          open statement previously saved...                    */
/* return codes: MD_NORMAL, on success.                           */
/*               Otherwise, md network error codes on failure.    */
/*----------------------------------------------------------------*/
int reconnect_server()
{
   char outbuf[BUF_SIZE];
   int status;

   if ((status=init_connect(openstmt, outbuf)) != MD_NORMAL) { 
         printf("reconnect_server: Can't reconnect to MD Server");
         close_connection(dbsock);
   }
   return(status);
}

/*----------------------------------------------------------------*/
/* Purpose: called after get_message to check whether the read    */
/*          was successful.                                       */
/* return codes: 0, on read failure.                              */
/*               1, on success.                                   */
/*----------------------------------------------------------------*/
int valid_read(status)
int status;
{
   if (status == MD_READERR) { 
       /* printf("Error: in reading the result from server...exiting\n"); */
       close_connection(dbsock);
       return(0);
   }
   if (status == MD_TIMEDOUT) {
      /* printf("Error: server response time limit reached...exiting\n"); */
      close_connection(dbsock);
      return(0);
   }
   return(1);
}

