/* File: $RCSfile: IK_DiskCheck.c,v $
 *
 * Synopsis: Checks disk usage for a user.
 *
 * Notes: */

/*--------------------------------------------------
 | $Log: IK_DiskCheck.c,v $
 | Revision 1.6  1997/03/06 15:38:52  amrhein
 | Updated to GUI's 6_1_12 IK
 |
 * Revision 6.1  1996/10/18  14:15:52  ims
 * Promoted for 6.1 J. Katen
 *
 * Revision 6.0  1996/02/29  15:37:34  ims
 * COPIED FROM 5.1 FOR GUI_PROMO_6_0_960229.
 *
 * Revision 5.1  1995/12/28  19:55:06  ims
 * COPIED FROM 5.0 FOR GUI_PROMO_5_1_951228.
 *
 * Revision 5.0  1995/11/06  13:03:35  ims
 * PROMOTION FROM GUI_4_5_8_951030 FOR GUI_PROMO_5_0_951106
 *
 * Revision 4.5  1995/07/27  18:42:04  ims
 * PROMOTION FROM GUI_4_4_8_950501 FOR GUI_PROMO_4_5_950726
 *
 * Revision 4.4  1995/02/13  22:19:59  ims
 * PROMOTION FROM GUI_4_4_950106 FOR GUI_PROMO_4_4_950213
 *
 * Revision 4.3  1994/08/26  10:49:20  ims
 * COPIED FROM GUI_4_3_940824.
 *
 * Revision 4.0  1994/06/08  14:34:56  ims
 * PROMOTION FROM REV4
 *
 * Revision 3.0.1.6  1994/05/12  14:23:47  kerbel
 * Change default for source path to GaeaTmp direcotory instead of UserData
 * directory.
 *
 * This is part of overhaul to move all transient files to GaeaTmp directory
 *
 * Revision 3.0.1.5  1994/04/20  19:45:58  kerbel
 * Make IK_Ims.h the first include file
 *
 * Revision 3.0.1.4  1994/04/20  15:49:13  kerbel
 * Initialize IK_ImsErrno to IK_ENOERROR, and add pre/post condition comments
 *
 * Revision 3.0.1.3  1994/04/20  15:40:23  kerbel
 * Add code to set IK_ImsErrno when system call fails and IK_CheckDiskAlloc()
 * returns -1.  If unknown error occurs set IK_ImsErrno to -1.
 *
 * Revision 3.0.1.2  1994/04/12  19:26:46  kerbel
 * changed function name to IK_CheckDiskAlloc()
 *
 * Revision 3.0.1.1  1994/04/12  19:09:58  winter
 * First checkin of real code to determine disk availability.
 *
 |
 +------------------------------------------------*/

static const char *rcsid =
"$Id: IK_DiskCheck.c,v 1.6 1997/03/06 15:38:52 amrhein Exp $";

#include "IK_Ims.h"             /* IMS configuration file                   */
                                /* NOTE: IK_Ims.h must be the first include */
                                /*       to ensure all config options are   */
                                /*       set properly                       */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dirent.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "IK_Errno.h"           /* IMS error number definitions             */
#include "IK_DiskCheck.h"       /* defs associated with IK_DiskCheck module */

/*
 * Name:
 *   IK_CheckDiskAlloc
 *
 * Description:
 *   Check bytes used in a users data directory and determine if there
 *   is enough room under the quota specified by the environment variable
 *   DISK_ALLOC to save the files specified by the file_list parameter
 *
 * Parameters:
 *   char *file_list[] - list of files to be saved
 *   char *source_path - directory where files to be saved are located
 *   char *dest_path   - directory where saved files are to be stored
 *                       this is the directory that must stay below
 *                       DISK_ALLOC bytes
 *
 * Return Values:
 *
 *   type: short
 *
 *         IK_TRUE  - there is enough room to save files
 *         IK_FALSE - there is not enough room to save files
 *         -1       - an error occurred while making a system call
 *
 * Global Variables Used:
 *
 * Pre- and Post-Conditions
 *
 * Precondition  1: There is enough room in destination directory to
 *                  save files specified in file_list
 *
 * Postcondition 1: IK_ImsErrno is set to 0, and IK_TRUE is returned
 *
 *
 * Precondition  2: There is NOT enough room in destination directory to
 *                  save files specified in file_list
 *
 * Postcondition 2: IK_ImsErrno is set to 0, and IK_FALSE is returned
 *
 *
 * Precondition  3: Some error condition exists that will cause a 
 *                  system call to fail
 *
 * Postcondition 3: IK_ImsErrno is set based on unix errno, and -1 is
 *                  returned
 *
 * Notes:
 *
 * 1) If neither source_path is not specified default to GAEATMP_DIR.
 *    If dest_path is not specified default to USRDATA_DIR
 *
 * 2) This function must be called once for each source/dest pair.  That
 *    is there is no way to specify a set of files that come from more than
 *    one directory.
 *
 * 3) THis function will only check disk usage for the dest_path driectory
 *    or its default.  It is not capable of checking disk usage across
 *    multiple directories, nor will it remember how much space was used in
 *    directory A when directory B is checked later.  The effect of this is
 *    that the DISK_ALLOC environment variable has the effect of limiting
 *    disk usage only within a single directory, not across a users entire
 *    environment.  In the current IMS application this is all that is
 *    needed, but if in the future users are allowed to specify where
 *    they want their data saved this routine will not be sufficient to 
 *    enforce disk quotas.
 *
 * 4) This function assumes files with the same name in source_path and
 *    dest_path will be overwritten, and therefor does not count the size 
 *    of a duplicate file name in the dest_path direcotry if that file
 *    is in file_list.
 * 
 * 5) This function does not protect against oversized temporary result
 *    files.
 *
 */

short IK_CheckDiskAlloc(char **file_list, char *source_path, char *dest_path)
{
   long quota;             /* num of bytes user is allowed in dest_path dir */
   unsigned short i;       /* loop control index                            */
   char fullpath[255];     /* full path to file being examined              */
   long new_sz = 0;        /* num of bytes needed to save new files         */
   long curr_sz = 0;       /* num of bytes currently used in dest_path dir  */
   struct stat stat_buf;   /* status info for file being examined           */
   DIR *directory;         /* info on save directory                        */
   struct dirent *file_info;  /* info on files already in dest_path dir     */
   short skip_flag;        /* flag is true if file being looked at by       */
                           /* file_info is either in the ignore_list or     */
                           /* is in file_list                               */

   /* list of files that don't count against byte total.              */
   /* First two entries denote the dest_path directory and its parent */
   static char *ignore_list[] =  { ".", "..", 0 };

   /* initialize IK_ImsErrno to no error */
   IK_ImsErrno = IK_ENOERROR;

   /* get the disk quota alloted to this user. */
   if ((quota = atol(IK_SysConfig("DiskAlloc"))) < 0)
       return (IK_TRUE);

   /* if no source path specified assume new files are in UserData */
   if (source_path == (char *)0)
    {
       source_path = (char *)malloc(strlen(IK_SysConfig("GaeaTmp")) + 1);
       strcpy(source_path, IK_SysConfig("GaeaTmp"));
    }

   /* if no destination path specified save files in UserData */
   if (dest_path == (char *)0)
    {
       dest_path = (char *)malloc(strlen(IK_SysConfig("UserData")) + 1);
       strcpy(dest_path, IK_SysConfig("UserData"));
    }

   /* loop through all the new files to be saved.  Get their status          */
   /* and add up their sizes to get the total space needed for the new files */
   for (i = 0; file_list[i]; i++)
    {
       sprintf(fullpath, "%s/%s", source_path, file_list[i]);
       if (stat(fullpath, &stat_buf) < 0)
	{
	   switch (errno)
	    {
	    case ENOTDIR:      
	       IK_ImsErrno = IK_ENOTDIR;
	       break;
	    case ENOENT:
	       IK_ImsErrno = IK_ENOENT;
	       break;
	    case EACCES:
	       IK_ImsErrno = IK_EPERM;
	       break;
	    case ENAMETOOLONG:
	       IK_ImsErrno = IK_EDOM;
	       break;
	    case ELOOP:
	       IK_ImsErrno = IK_EMLINK;
	       break;
	    case EFAULT:
	       IK_ImsErrno = IK_EFAULT;
	       break;
	    default:
	       IK_ImsErrno = -1;
	       break;
	    }

	   return (-1);
	}

       new_sz += stat_buf.st_size;
    }

   /* if new files exceed the quota by themselves return IK_FALSE now */
   if (new_sz > quota)
       return (IK_FALSE);

   /* open the destination directory and check for errors */
   if ((directory = opendir(dest_path)) == (DIR *)0)
    {
       switch (errno)
	{
	    case ENOTDIR:      
	       IK_ImsErrno = IK_ENOTDIR;
	       break;
	    case EACCES:
	       IK_ImsErrno = IK_EPERM;
	       break;
	    case ENAMETOOLONG:
	       IK_ImsErrno = IK_EDOM;
	       break;
	    case ENOENT:
	       IK_ImsErrno = IK_ENOENT;
	       break;
	    case EMFILE:
	       IK_ImsErrno = IK_EMFILE;
	       break;
	    case EFAULT:
	       IK_ImsErrno = IK_EFAULT;
	       break;
	    case EAGAIN:
	       IK_ImsErrno = IK_EAGAIN;
	       break;
	    case ENOMEM:
	       IK_ImsErrno = IK_ENOMEM;
	       break;
	    default:
	       IK_ImsErrno = -1;
	       break;
	}
       
       return (-1);
    }

   /* loop through all the files found in the destinaion directory */
   while ((file_info = readdir(directory)))
    {
       sprintf(fullpath, "%s/%s", dest_path, file_info->d_name);

       skip_flag = IK_FALSE; /* initialize skip flag to indicate that  */
                             /* current file still counts toward total */

       /* loop through all the files in the ignore list.  If the current */
       /* file is in this list set the skip_flag and exit the loop       */
       for (i = 0; (ignore_list[i] && (!skip_flag)); i++)
	{
	   if (strcmp(ignore_list[i], file_info->d_name) == 0)
	       skip_flag = IK_TRUE;
	}

       /* loop through all the files in file_list.  If the current */
       /* file is in this list set the skip_flag and exit the loop */
       for (i = 0; (file_list[i] && (!skip_flag)); i++)
	{
	   if (strcmp(file_list[i], file_info->d_name) == 0)
	       skip_flag = IK_TRUE;
	}

       /* if this file doesn't count go to the next file */
       if (skip_flag)
	   continue;

       /* get stat info for current file */
       if (stat(fullpath, &stat_buf) < 0)
	{
	   switch (errno)
	    {
	    case ENOTDIR:      
	       IK_ImsErrno = IK_ENOTDIR;
	       break;
	    case ENOENT:
	       IK_ImsErrno = IK_ENOENT;
	       break;
	    case EACCES:
	       IK_ImsErrno = IK_EPERM;
	       break;
	    case ENAMETOOLONG:
	       IK_ImsErrno = IK_EDOM;
	       break;
	    case ELOOP:
	       IK_ImsErrno = IK_EMLINK;
	       break;
	    case EFAULT:
	       IK_ImsErrno = IK_EFAULT;
	       break;
	    default:
	       IK_ImsErrno = -1;
	       break;
	    }

	   return (-1);
	}
      
       /* add current file size to total */
       curr_sz += stat_buf.st_size;

       /* if total exceeds the quota than return IK_FALSE now */
       if ((curr_sz + new_sz) > quota)
	{
	   closedir(directory);
	   return (IK_FALSE);
	}
    }

   /* close the dest_path directory */
   closedir(directory);

   /* there is room for the new files so return IK_TRUE */
   return (IK_TRUE);
}
