/*
 * Name: IK_InGran.c
 *
 * Description: This file contains code to ingest granules from an
 * Inventory Result message. This code is internal to
 * IK_InvResultIngest(). Basically, this file contains one main
 * function, IK_IngestGranules(), which ingests the contents of a
 * DATASET group from an inventory results message. The information
 * for each granule within the dataset group is extracted, packed into
 * a data structure, and written to the local gdbm database file which
 * holds granule information.
 *
 * Notes:
 *
 *-------------------------------------------------------------------------
 *
 * RCS information:
 *
 * $RCSfile: IK_InGran.c,v $
 *
 * $Id: IK_InGran.c,v 1.7 1997/03/06 15:38:52 amrhein Exp $
 *
 * $Log: IK_InGran.c,v $
 * Revision 1.7  1997/03/06 15:38:52  amrhein
 * Updated to GUI's 6_1_12 IK
 *
 * Revision 6.1.1.2  1996/12/29  02:51:58  winter
 * Added #include "IK_ODLetc.h".
 *
 * Revision 6.1.1.1  1996/10/21  14:23:05  winter
 * Copied from 6.0.1.6 after 6.1 promotion.
 *
 * Revision 6.0.1.6  1996/09/25  18:55:51  winter
 * Added #include <sys/time.h> to avoid compiler warning on select()
 * prototype in <unistd.h>.
 *
 * Revision 6.0.1.5  1996/04/01  14:41:55  wright
 * removed call to IK_InDsetStatCodeComment.  The call now resides in IK_IngestGranules
 *
 * Revision 6.0.1.4  1996/03/23  21:29:30  walbert
 * In IK_IngestGranule, under SENSOR_NAME ingest area, changed
 * the statement if(param == NULL) to if(param != NULL) to
 * allow the sensor name to be ingested properly.  This fixed
 * SMRs 3513 and 3515 and an ASF crash.
 *
 * Revision 6.0.1.3  1996/03/16  13:52:56  winter
 * Cleaned up and improved commenting.
 * ,
 *
 * Revision 6.0.1.2  1996/03/14  21:00:20  winter
 * Cleaned up and fixed crash which occurred when a DATASET group contained
 * a dataset name not in the valids.
 *
 * Revision 6.0.1.1  1996/03/14  16:32:16  winter
 * Placeholder for new branch.
 *
 * Revision 6.0  1996/02/29  15:39:39  ims
 * COPIED FROM 5.1.2.5 FOR GUI_PROMO_6_0_960229.
 *
 * Revision 5.1.2.5  1996/02/15  20:20:01  winter
 * Fixed bugs arising from changing the tag type again.
 *
 * Revision 5.1.2.4  1996/02/14  20:11:14  winter
 * Changed call to IK_MapDatasetIdent() to IK_MapDatasetNameToTag().
 * Changed call to IK_MapArchiveIdent() to IK_MapArchiveNameToTag().
 * Changed memcpy() calls to account for new IK_DatasetTag and IK_ArchiveTag
 * data types.
 *
 * Revision 5.1.2.3  1996/02/01  17:34:35  winter
 * MERGED 5.1.2.2 AND 5.1.1.2.
 *
 * Revision 5.1.2.2  1996/01/18  16:00:09  ying
 * Changes are made to ingest dataset level STATUS_CODE and STATUS_CODE_COMMENT
 * if any.
 *
 * Revision 5.1.2.1  1996/01/18  15:05:38  ying
 * Placeholder for new branch.
 *
 * Revision 5.1  1995/12/28  19:58:25  ims
 * COPIED FROM 5.0.1.1 FOR GUI_PROMO_5_1_951228.
 *
 * Revision 5.0.1.1  1995/11/15  20:10:06  iredell
 * allow PROCESSING_LEVEL, DAY_NIGHT, and COMMENT to be grabbed from the dat
 * group.
 * null terminated the end of CAMPAIGN, IMAGE_ID, and
 * PROCESSING_LEVEL.
 *
 * Revision 5.0  1995/11/06  13:03:49  ims
 * PROMOTION FROM GUI_4_5_8_951030 FOR GUI_PROMO_5_0_951106
 *
 * Revision 4.5  1995/07/27  18:42:24  ims
 * PROMOTION FROM GUI_4_4_8_950501 FOR GUI_PROMO_4_5_950726
 *
 * Revision 4.4.1.2  1995/04/03  19:21:43  shiva
 * Added the capability to grap the CAMPAIGN name even if it
 * is defined at the dataset level.  It used to not be set if it
 * could not find it at the granule level.
 * Now if it's defined at either places, the campaign name will be
 * ingested.  Now when the user sorts based on campaign, the name
 * will come up instead of blank!
 *
 * Revision 4.4  1995/02/13  22:20:29  ims
 * PROMOTION FROM GUI_4_4_950106 FOR GUI_PROMO_4_4_950213
 *
 * Revision 4.3.1.3  1994/11/01  16:51:52  kerbel
 * In call to IK_IngestDataset fixed in 4.3.1.2 changed archive_tag
 * and dataset tag actual parameters to pass contents of pointer instead
 * of pointer as called for in IK_IngestDatset() prototype.
 *
 * Revision 4.3.1.2  1994/11/01  16:29:47  kerbel
 * Changed actual parameters for IK_IngestDataset() call from goodODL_tree
 * section of IK_IngestGranules to use archive_tag and dataset_tag locl
 * variables, instead of values pointed to by the granule ptr grP.  This
 * was done to correctly handle the last chunk of results from MSFC
 * which only contains the number of granule hits field without any actual
 * granules returned in the last chunk.  Without this fix in place the
 * archive and datset tags passed to the IK_IngestDataset() function
 * had garbage values, and created an eroneous entry in the dataset.gdb
 * results file.  This was happening because without any granules in
 * the last chunk the grP granule pointer never had any values placed
 * in the data area it pointed to.
 *
 * Revision 4.3.1.1  1994/11/01  16:25:57  kerbel
 * new branch
 *
 * Revision 4.3  1994/08/26  10:49:39  ims
 * COPIED FROM GUI_4_3_940824.
 *
 * Revision 4.0.1.4  1994/07/08  18:09:05  vaffajou
 * Fixing the problem with Mark for Browse when Browse is not available for
 * a granule (Marking through the menu).  Added a line which copies the
 * granule's browse_type to the granule's Marked.avail_browse_type
 *
 * Revision 4.0.1.3  1994/06/24  17:44:15  erabelli
 * Fixed the memory allocation for package_selected when more than on package_id
 * is ingested into the gdbm. And also, all package_selected variables are
 * initialized to IK_FALSE.
 *
 * Revision 4.0.1.2  1994/06/20  16:11:46  kerbel
 * reset errno to 0 in clean_up section of IK_IngestGranules routine.
 * This was done to avoid problems with free() and should no longer
 * be needed now that ims_free is being used
 *
 * Revision 4.0.1.1  1994/06/20  16:09:15  kerbel
 * new branch
 *
 * Revision 4.0  1994/06/08  14:50:40  ims
 * PROMOTION FROM REV4
 *
 * Revision 3.2.1.3  1994/06/03  18:21:36  dandi
 * Added temporary fix to the problem with memory allocation
 * for package_selected when more than on package_id is
 * returned in the result.
 *
 * Revision 3.2.1.2  1994/05/25  19:07:28  kerbel
 * moved IK_Ims.h to top of include list
 *
 * Revision 3.2.1.1  1994/05/25  18:53:40  kerbel
 * new branch
 *
 * Revision 3.2  1994/05/03  20:11:33  ims
 * Promotion from BUILD_94MAY03.
 *
 * Revision 3.1.1.2  1994/04/26  13:41:40  winter
 * Merged deltas from 3.0.2.1 (Bridge stuff) with 3.1.1.1.
 *
 * Revision 3.1.1.1  1994/04/19  13:07:15  winter
 * Copy of revision 3.0.1.4. This revision contains browse product description
 * changes.
 *
 * Revision 3.0.1.4  1994/04/12  15:50:41  winter
 * Changes made by Leonid: reversed order of #include for IK_DataDic.h
 * and IK_Coverag.h.
 *
 * Revision 3.0.1.3  1994/03/23  12:39:29  winter
 * Added code to process BROWSE_PRODUCT_DESCRIPTION keyword, and to put the
 * strings composing this group into the browse_desc member of the IK_Granule
 * structure.
 *
 * Revision 3.0.1.2  1994/03/22  17:25:54  winter
 * Just did code cleanup - new comments, more braces, etc..
 *
 * Revision 3.0.1.1  1994/01/27  19:30:49  winter
 * Placeholder revision to begin code cleanup.
 *
 * Revision 3.0  1993/12/29  15:18:49  winter
 * ALPHA-1
 *
 * Revision 1.25  1993/08/13  19:35:07  sudarsan
 * Modified the code to reflect changes in IK_Granule.h ie replaced
 * image_avail with brs_avail
 *
 * Revision 1.24  1993/06/28  20:49:07  sylvain
 * The prototype for IK_Sequence2List was finually updated, so the call had
 * 	to reflect the new change. (the type of the first parameter was
 * 	change from size_t to int)
 *
 * Revision 1.23  1993/06/25  19:48:31  sylvain
 * added the ingestion of the image_id into the granule database in
 * 	IK_IngestGranules so that IK_MapImageFilename wont crash gaea
 * 	when it looks in the granule that is passed to it to map it to a
 * 	filename from the hdfnames.gdb.
 *
 * Revision 1.22  1993/06/17  17:36:27  sylvain
 * Fixed a corruption in the granule database that was being cause because the
 * 	address weren't properly being aligned to byte offsets.
 * Marks are now saved before a granule is overwritten.
 * Better error check is performed when IK_PutGranule fails.
 *
 * Revision 1.21  1993/04/22  13:48:06  heather
 * set start date and stop date to offset before memory realloc and then reset
 * the dates backe to pointers
 *
 * Revision 1.20  1993/01/06  19:18:47  honce
 * IK_Granule.h broken up into IK_Granule.h and IK_GranulP.h.
 * IK_GranulP.h is for private data.
 * IK_Dataset.h broken up into IK_Dataset.h and IK_DatastP.h.
 * IK_DatastP.h is for private data.
 * grP->granule.marked.visible now set to IK_VISIBLE instead of IK_TRUE
 *
 * Revision 1.19  1992/11/24  21:04:23  honce
 * Forgot to allocate space for granule.package_selected.
 *
 * Revision 1.18  1992/11/20  22:33:40  honce
 * Commenting out all the code for browse_type was a little too much.
 * Therefore, the code is restored and defaults to IK_NO_BROWSE.
 * Changed FTP_Only to FTP_ONLY, to match data dictionary.
 *
 * Revision 1.17  1992/11/18  02:21:03  sylvain
 * Don't check for browse_type in the granule group anymore, it's not
 * required.
 *
 * Revision 1.16  1992/11/16  03:56:56  honce
 * Missed POLE_INCLUDED in POLYGON_LOCs
 *
 * Revision 1.15  1992/11/10  22:20:22  honce
 * Added ingestion of PROCESSING_LEVEL which had been forgotten. :-(
 *
 * Revision 1.14  1992/11/09  22:36:05  honce
 * Because of a "feature" of ODL, it is not safe to allow groups to be
 * defaulted to the dataset group if they are not in the granule group.
 * The "feature" is that ODL, when looking for a group, will do a recursive
 * descent throughout the whole tree until it finds the required member!
 *
 * Revision 1.13  1992/11/09  20:06:09  honce
 * centroid_long was used instead of CENTROID_LON.
 *
 * Revision 1.12  1992/11/02  23:49:34  honce
 * granule.package was renamed granule.package_id
 *
 * Revision 1.11  1992/11/02  23:17:55  honce
 * source, sensor, and parameter no longer required fields.
 *
 * Revision 1.10  1992/10/22  21:34:56  honce
 * Added code to check for duplicate granule id's and not increase the granule
 * count if the duplicate is overwritting a previous granule.  Added code to
 * "back out" the granules already created if the ODL tree is found to be
 * corrupt.  This process currently does not free the disk space used by
 * the deleted granules.
 *
 * Revision 1.9  1992/10/19  22:14:45  honce
 * now set parm_name for every parameter while processing.
 *
 * Revision 1.8  1992/10/14  20:51:07  honce
 * Now need to count the NULL for Ident mapping on strings.
 *
 * Revision 1.7  1992/10/08  02:14:13  honce
 * Can't "FixUP" pointers until done realloc()'ing the buffer.  Realloc()
 * if needed could move the buffer somewhere else in memory when making it
 * grow, this would screw up any pointers pointing into the buffer.
 * After all the realloc()'s are done, then we make the offsets pointers.
 *
 * Revision 1.6  1992/10/07  18:00:20  honce
 * Instead of Remove[Parameter|Group]'ing every item as it is done, I now
 * wait and destory the whole dataset record at once.  This improves
 * preformance a little bit and since the whole operation is atomic the
 * user will never notice. :-)
 *
 * Revision 1.5  1992/10/06  21:04:54  honce
 * Missed one empty statement.
 *
 * Revision 1.4  1992/10/06  21:02:02  honce
 * Removed extra #include <stddef.h>, printf's and empty statements.
 *
 * Revision 1.3  1992/10/05  23:20:01  honce
 * Was removing granule level information that just happened to live at the
 * dataset level. All granule information lives until whole message ingested.
 * Now using ODLFormatReal() to get "reals" instead of strncpy(), ouch!
 *
 * Revision 1.2  1992/10/02  18:57:20  honce
 * Added call to IK_GranuleCount() to increase number of ingested granules.
 * set visible flag for granules.
 *
 * Revision 1.1  1992/10/01  18:41:34  honce
 * Initial revision
 *
*/

/*****************************************************************************/

/* #include directives */

/* This header must be #included first in order to use
   autoconf-generated information. */
#include "IK_Ims.h"

/* Standard C header files */
#include <sys/time.h>
#include <assert.h>
#include <stdio.h>  /* Needed for odldef.h */
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifdef BRIDGE_CONFIG
#include <signal.h>
#endif

/* Headers for third-party code */
#include "odldef.h"
#include "odlinter.h"

/* IMS headers */
#include "IK_Coverag.h"
#include "IK_DataDic.h"
#include "IK_Errno.h"
#include "IK_GranulP.h"
#include "IK_Granule.h"
#include "IK_Fixup.h"
#include "IK_Ingest.h"
#include "IK_Map.h"
#include "IK_Mark.h"
#include "IK_Msg.h"
#include "IK_ODLetc.h"
#include "IK_Syslog.h"

/*****************************************************************************/

/* Define the RCS identifier string. */
static const char *rcsid = "$Id: IK_InGran.c,v 1.7 1997/03/06 15:38:52 amrhein Exp $";

/*****************************************************************************/

/* Local function prototypes */

#ifdef BRIDGE_CONFIG
void (*IK_IngestTimeout)() = 0;
#endif

static IK_GranuleP *granule_list2c(IK_GranuleP *pGranuleP, size_t *psz_granule,
				   char **ppc_stringList, size_t sz_list);

/****************************************************************************/

/*
 * Name: IK_IngestGranules()
 *
 * Description: This function ingests the granules in a DATASET group
 * from an inventory result message.
 *
 * Arguments:
 * AGGREGATE agg_datasetGroup: Dataset containing granules to ingest
 *
 * Return Values:
 * Upon success, or NULL input, return 0.
 * On failure, return -1.
 *
 * Warnings:
 * Why does this function return 0 when a NULL pointer is passed in?
 *
 * Global Variables Used:
 * IK_ImsErrno
 *
 * Pre- and Post-Conditions:
 *
 * Revision History:
 * 22 March 1994: Eric Winter modified for code cleanup and to add extraction
 * of browse product description from dataset group.
 *
 * Wednesday 14 February 1996 (Eric Winter) - Changed call to
 * IK_MapArchiveIdent() to IK_MapArchiveNameToTag(). Changed call to
 * IK_MapDatasetIdent() to call to IK_MapDatasetNameToTag().
 *
 * Thursday 14 March 1996 (Eric Winter) - Extensive overhaul during
 * bug investigation.
*/

int
IK_IngestGranules(AGGREGATE agg_datasetGroup)
{

    /* SortKey_granuleList is the head of the linked list of ingested
       granules from this DATASET group. */
    IK_SortKey SortKey_granuleList;

    /* pSortKey_granule points to the sort key for the current
       granule. */
    IK_SortKey *pSortKey_granule = NULL;

    /* pGranuleP points to the current granule buffer (a private data
       structure). */
    IK_GranuleP *pGranuleP = NULL;

    /* sz_granuleBuffer is the current allocated size of the granule
       buffer. */
    size_t sz_granuleBuffer = 0;

    /* pc_paramName is the name of the current ODL parameter. */
    char *pc_paramName = NULL;

    /* param is used as a temporary ODL parameter variable. */
    PARAMETER param = NULL;

    /* value is the value of the current ODL parameter. */
    VALUE value = NULL;

    /* pDatasetTag is a pointer to the current dataset tag. */
    IK_DatasetTag *pDatasetTag = NULL;

    /* agg_parentGroup points to the parent ODL group for this dataset
       group. */
    AGGREGATE agg_parentGroup = NULL;

    /* pArchiveTag is a pointer to the current archive tag. */
    IK_ArchiveTag *pArchiveTag = NULL;

    /* agg_granuleGroup is the ODL aggregate for the current GRANULE
       group being processed. */
    AGGREGATE agg_granuleGroup = NULL;

    /* i_numGranules is the number of granules ingested so far. */
    int i_numGranules = 0;

    /* i_numBytes is the number of bytes required by the previous
       operation. */
    int i_numBytes = 0;

    /* sz_alignedBytes is the number of aligned bytes required by the
       previous operation. */
    size_t sz_alignedBytes = 0;

    /* ppc_valueList points to the list of parameter values. */
    char **ppc_valueList = NULL;

    /* Number of order packages */
    int i_packageCount = 0;

    /* pc_groupName is the name of the current ODL group. */
    char *pc_groupName = NULL;

    /* agg_coverageArea is the ODL aggregate for geographic coverage
       from the granule. */
    AGGREGATE agg_coverageArea = NULL;

    /* i_numPolygonSides is the number of sides in a coverage polygon. */
    int i_numPolygonSides = 0;

    /* This is a simple loop counter. */
    int i = 0;

#ifdef BRIDGE_CONFIG
    sigset_t sig_set;           /* set of signals to block */
    struct sigaction act;       /* new signal handling info for SIGALRM */
    struct sigaction oact;      /* old singal handling info for SIGALRM */
#endif /* BRIDGE_CONFIG */

    /*----------------------------------------------------------------------*/

#ifdef BRIDGE_CONFIG
    /* create empty signal set */
    sigemptyset(&sig_set);

    /* add SIGALRM to empty signal set, so signal set has just SIGALRM */
    sigaddset(&sig_set, SIGALRM);
#endif /* BRIDGE_CONFIG */

    /*----------------------------------------------------------------------*/

    /* If we were passed a NULL pointer, abort this function. */
    if (agg_datasetGroup == NULL) {
	return(0);
    }

    /*----------------------------------------------------------------------*/

    /* The head of the granule key list will just be a place to start
       and won't really have a granule key in it. */
    SortKey_granuleList.usrapp1 = NULL;
    pSortKey_granule = &SortKey_granuleList;

    /*----------------------------------------------------------------------*/

    /* Allocate the beginning of the granule buffer.  This buffer will
       not be deallocated until all the granules for a dataset have
       been ingested. This buffer will grow as needed but will not
       shrink. This buffer is cleared between granules. */
    pGranuleP = malloc(sizeof(IK_GranuleP) + IK_MEMBUFSIZ);
    assert(pGranuleP != NULL);

    /* Calculate the number of allocated bytes. This is the size of
       the area pointed to be pGranuleP. */
    sz_granuleBuffer = sizeof(IK_GranuleP) + IK_MEMBUFSIZ;

    /* Map the ODL "DATASET_ID" keyword to the pDatasetTag */
    pc_paramName = "DATASET_ID";
    param = FindParameter(agg_datasetGroup, pc_paramName);
    if (param == NULL) {
	goto bad_ODLtree;
    }
    value = FirstValue(param);
    if (value == NULL) {
	goto bad_ODLtree;
    }
    if (value->item.value.string == NULL) {
	goto bad_ODLtree;
    }
    pDatasetTag = IK_MapDatasetNameToTag(value->item.value.string);
    if (pDatasetTag == NULL) {
	goto bad_ODLtree;
    }

    /*----------------------------------------------------------------------*/

    /* Map the ODL keyword "DATA_CENTER_ID" from the parent of this
       ODL group to an pArchiveTag. */
    pc_paramName = "DATA_CENTER_ID";
    agg_parentGroup = ParentGroup(agg_datasetGroup);
    if (agg_parentGroup == NULL) {
	goto bad_ODLtree;
    }
    param = FindParameter(agg_parentGroup, pc_paramName);
    if (param == NULL) {
	goto bad_ODLtree;
    }
    value = FirstValue(param);
    if (value == NULL) {
	goto bad_ODLtree;
    }
    if (value->item.value.string == NULL) {
	goto bad_ODLtree;
    }
    pArchiveTag = IK_MapArchiveNameToTag(value->item.value.string);
    if (pArchiveTag == NULL) {
	goto bad_ODLtree;
    }

    /*----------------------------------------------------------------------*/

#ifdef BRIDGE_CONFIG
    if (IK_IngestTimeout) {
        act.sa_handler = IK_IngestTimeout;
        sigemptyset(&act.sa_mask);
        act.sa_flags = 0;
        sigaction(SIGALRM, &act, &oact);
        alarm (BRIDGE_INGEST_TIMEOUT);
    }
#endif /* BRIDGE_CONFIG */

    /*----------------------------------------------------------------------*/

    /* Walk through all the granules in the given dataset. */
    agg_granuleGroup = FindGroup(agg_datasetGroup, "GRANULE");
    for (i_numGranules = 0; agg_granuleGroup != NULL;
	 agg_granuleGroup = FindNextGroup(agg_datasetGroup, agg_granuleGroup,
					  "GRANULE")) {

#ifdef BRIDGE_CONFIG
	/* block all signals while processing a granule */
	sigprocmask(SIG_BLOCK, &sig_set, 0);
#endif

	/* Clear the granule buffer. */
	(void) memset(pGranuleP, '\000', sz_granuleBuffer);

	/* Aligned_data is going to point to the unused "heap" at the
	   end of the granule buffer. */
	pGranuleP->aligned_data = (char *) &pGranuleP->aligned_data +
	  sizeof(void *);

	/* Reset the number of used bytes used in the granule buffer
           to be the space used by the IK_GranulP header structure. */
	pGranuleP->nbytes = sizeof(IK_GranuleP);

	/* Copy the dataset and archive tags to the granule. */
	(void) memcpy(pGranuleP->granule.dataset_tag, pDatasetTag,
		      IK_DATASET_TAG_SZ);
	(void) memcpy(pGranuleP->granule.archive_tag, pArchiveTag,
		      IK_ARCHIVE_TAG_SZ);

	/*------------------------------------------------------------------*/

	/* Get parameter values and stuff them into the granule
	   structure again and again and again ... */

	/*------------------------------------------------------------------*/

	/* (Required) Granule ID */
	pc_paramName = "GRANULE_ID";
	param = FindParameter(agg_granuleGroup, pc_paramName);
	if (param == NULL) {
	    goto bad_ODLtree;
	}
	value = FirstValue(param);
	if (value == NULL) {
	    goto bad_ODLtree;
	}
	if (value->item.value.string == NULL) {
	    goto bad_ODLtree;
	}
	(void) strncpy(pGranuleP->granule.granule_id, value->item.value.string,
		       value->item.length);
	pGranuleP->granule.granule_id[value->item.length] = '\0';

	/*------------------------------------------------------------------*/

	/* (Required) Start date for data in this granule */
	pc_paramName = "START_DATE";
	param = FindParameter(agg_granuleGroup, pc_paramName);
	if (param == NULL) {
	    goto bad_ODLtree;
	}
	value = FirstValue(param);
	if (value == NULL) {
	    goto bad_ODLtree;
	}
	pGranuleP->granule.start_date = (char *) pGranuleP->aligned_data;
	i_numBytes = ODLFormatDateTime(pGranuleP->granule.start_date,
				       &value->item);
	pGranuleP->granule.start_date =
	  (char *) IK_RELOC(&pGranuleP->aligned_data, pGranuleP->aligned_data);
	sz_alignedBytes = IK_ALIGN(i_numBytes);
	pGranuleP->aligned_data = (char *) pGranuleP->aligned_data +
	  sz_alignedBytes;
	pGranuleP->nbytes += sz_alignedBytes;

	/*------------------------------------------------------------------*/

	/* (Required) Stop date for data in this granule */
	pc_paramName = "STOP_DATE";
	param = FindParameter(agg_granuleGroup, pc_paramName);
	if (param == NULL) {
	    goto bad_ODLtree;
	}
	value = FirstValue(param);
	if (value == NULL) {
	    goto bad_ODLtree;
	}
	pGranuleP->granule.stop_date = (char *) pGranuleP->aligned_data;
	i_numBytes = ODLFormatDateTime(pGranuleP->granule.stop_date,
				       &value->item);
	pGranuleP->granule.stop_date =
	  (char *) IK_RELOC(&pGranuleP->aligned_data, pGranuleP->aligned_data);
	sz_alignedBytes = IK_ALIGN(i_numBytes);
	pGranuleP->aligned_data = (char *) pGranuleP->aligned_data +
	  sz_alignedBytes;
	pGranuleP->nbytes += sz_alignedBytes;

	/*------------------------------------------------------------------*/

	/* (Optional) Type of browse available for this granule. */
	pc_paramName = "BROWSE_TYPE";
	pGranuleP->granule.browse_type = IK_NO_BROWSE;
	param = FindParameter(agg_granuleGroup, pc_paramName);
	if (param != NULL) {
	    value = FirstValue(param);
	    if (value != NULL) {
		if (value->item.value.string != NULL) {
		    if (*value->item.value.string == 'N') {
			pGranuleP->granule.browse_type = IK_NO_BROWSE;
		    } else if (*value->item.value.string == 'Y') {
			pGranuleP->granule.browse_type = IK_INTEGRATED;
		    } else if (strcmp(value->item.value.string, "FTP_ONLY")
			       == 0) {
			pGranuleP->granule.browse_type = IK_FTP;
		    }
		}
	    }
	}
	pGranuleP->granule.marked.avail_browse_type =
	  pGranuleP->granule.browse_type;

	/*------------------------------------------------------------------*/

	/* (Optional) Browse product description (at granule
           (preferred) or dataset level). */
	pc_paramName = "BROWSE_PRODUCT_DESCRIPTION";
	param = FindParameter(agg_granuleGroup, pc_paramName);
	if (param == NULL) {
	    param = FindParameter(agg_datasetGroup, pc_paramName);
	}
	if (param != NULL) {
	    ppc_valueList = IK_Sequence2List(&i_numBytes, param,
					     IK_BROWSE_PRODUCT_DESCRIPTION_SZ);
	    assert(ppc_valueList != NULL);
	    pGranuleP->granule.browse_desc =
	      (char **) IK_RELOC(&pGranuleP->aligned_data,
				 pGranuleP->aligned_data);
	    pGranuleP = granule_list2c(pGranuleP, &sz_granuleBuffer,
				       ppc_valueList, i_numBytes);
	    if (pGranuleP == NULL) {
		goto bad_ODLtree;
	    }
	}

	/*------------------------------------------------------------------*/

	/* (Optional) Name of collection campaign for this dataset (or
           granule). */
	pc_paramName = "CAMPAIGN";
	param = FindParameter(agg_datasetGroup, pc_paramName);
	if (param != NULL) {
	    value = FirstValue(param);
	} else {
	    param = FindParameter(agg_granuleGroup, pc_paramName);
	    if (param != NULL) {
		value = FirstValue(param);
	    } else {
		value = NULL;
	    }
	}
	if (value != NULL) {
	    if (value->item.value.string != NULL) {
		(void) strncpy(pGranuleP->granule.campaign,
			       value->item.value.string, IK_CAMPAIGN_ID_SZ);
		pGranuleP->granule.campaign[IK_CAMPAIGN_ID_SZ - 1] = '\0';  
	    }
	}

	/*------------------------------------------------------------------*/

	/* (Optional) Image ID */
	pc_paramName = "IMAGE_ID";
	param = FindParameter(agg_granuleGroup, pc_paramName);
	if (param != NULL) {
	    value = FirstValue(param);
	    if (value != NULL) {
		if (value->item.value.string != NULL) {
		    (void) strncpy(pGranuleP->granule.image_id,
				   value->item.value.string, IK_IMAGE_ID_SZ);
		    pGranuleP->granule.image_id[IK_IMAGE_ID_SZ - 1] = '\0';
		}
	    }
	}

	/*------------------------------------------------------------------*/

	/* (Optional) Data processing level for this granule */
	pc_paramName = "PROCESSING_LEVEL";
	param = FindParameter(agg_granuleGroup, pc_paramName);
	if (param != NULL) {
	    value = FirstValue(param);
	} else {
	    param = FindParameter(agg_datasetGroup, pc_paramName);
	    if (param != NULL) {
		value = FirstValue(param);
	    } else {
		value = NULL;
	    }
	}
	if (value != NULL) {
	    if (value->item.value.string != NULL) {
		(void) strncpy(pGranuleP->granule.processing_level,
			       value->item.value.string, IK_PROCESS_LEVEL_SZ);
		pGranuleP->granule.processing_level[IK_PROCESS_LEVEL_SZ - 1]
		  = '\0';
	    }
	}

	/*------------------------------------------------------------------*/

	/* (Optional) Day/night flag for data in this granule. */
	pc_paramName = "DAY_NIGHT";
	pGranuleP->granule.day_night = IK_BOTH;
	param = FindParameter(agg_granuleGroup, pc_paramName);
	if (param != NULL) {
	    value = FirstValue(param);
	} else {
	    param = FindParameter(agg_datasetGroup, pc_paramName);
	    if (param != NULL) {
		value = FirstValue(param);
	    } else {
		value = NULL;
	    }
	}
	if (value != NULL) {
	    if (value->item.value.string != NULL) {
		if (*value->item.value.string == 'D') {
		    pGranuleP->granule.day_night = IK_DAY;
		} else if (*value->item.value.string == 'N') {
		    pGranuleP->granule.day_night = IK_NIGHT;
		}
	    }
	}

	/*------------------------------------------------------------------*/

	/* (Optional) Granule comment */
	pc_paramName = "COMMENT";
	param = FindParameter(agg_granuleGroup, pc_paramName);
	if (param == NULL) {
	    param = FindParameter(agg_datasetGroup, pc_paramName);
	}
	if (param != NULL) {
	    ppc_valueList = IK_Sequence2List(&i_numBytes, param, IK_COMMENT_SZ);
	    assert(ppc_valueList != NULL);
	    pGranuleP->granule.comment =
	      (char **) IK_RELOC(&pGranuleP->aligned_data,
				 pGranuleP->aligned_data);
	    pGranuleP = granule_list2c(pGranuleP, &sz_granuleBuffer,
				       ppc_valueList, i_numBytes);
	    if (pGranuleP == NULL) {
		goto bad_ODLtree;
	    }
	}

	/*------------------------------------------------------------------*/

	/* (Optional) Name of sensor which collected this granule */
	pc_paramName = "SENSOR_NAME";
	param = FindParameter(agg_granuleGroup, pc_paramName);
	if (param == NULL) {
	    param = FindParameter(agg_datasetGroup, pc_paramName);
	}
	if (param != NULL) {
	    ppc_valueList = IK_Sequence2List(&i_numBytes, param,
					     IK_SENSOR_ID_SZ);
	    assert(ppc_valueList != NULL);
	    pGranuleP->granule.sensor_name =
	      (char **) IK_RELOC(&pGranuleP->aligned_data,
				 pGranuleP->aligned_data);
	    pGranuleP = granule_list2c(pGranuleP, &sz_granuleBuffer,
				       ppc_valueList, i_numBytes);
	    if (pGranuleP == NULL) {
		goto bad_ODLtree;
	    }
	}

	/*------------------------------------------------------------------*/

	/* (Optional) Name of data source */
	pc_paramName = "SOURCE_NAME";
	param = FindParameter(agg_granuleGroup, pc_paramName);
	if (param == NULL) {
	    param = FindParameter(agg_datasetGroup, pc_paramName);
	}
	if (param != NULL) {
	    ppc_valueList = IK_Sequence2List(&i_numBytes, param,
					     IK_SOURCE_ID_SZ);
	    assert(ppc_valueList);
	    pGranuleP->granule.source_name =
	      (char **) IK_RELOC(&pGranuleP->aligned_data,
				 pGranuleP->aligned_data);
	    pGranuleP = granule_list2c(pGranuleP, &sz_granuleBuffer,
				       ppc_valueList, i_numBytes);
	    if (pGranuleP == NULL) {
		goto bad_ODLtree;
	    }
	}

	/*------------------------------------------------------------------*/

	/* (Optional) Name of data parameters for this granule */
	pc_paramName = "PARAMETER";
	param = FindParameter(agg_granuleGroup, pc_paramName);
	if (param == NULL) {
	    param = FindParameter(agg_datasetGroup, pc_paramName);
	}
	if (param != NULL) {
	    ppc_valueList = IK_Sequence2List(&i_numBytes, param,
					     IK_PARAMETER_ID_SZ);
	    assert(ppc_valueList != NULL);
	    pGranuleP->granule.parameter =
	      (char **) IK_RELOC(&pGranuleP->aligned_data,
				 pGranuleP->aligned_data);
	    pGranuleP = granule_list2c(pGranuleP, &sz_granuleBuffer,
				       ppc_valueList, i_numBytes);
	    if (pGranuleP == NULL) {
		goto bad_ODLtree;
	    }
	}

	/*------------------------------------------------------------------*/

	/* (Optional) ID of packages containing this granule */
	pc_paramName = "PACKAGE_ID";
	param = FindParameter(agg_granuleGroup, pc_paramName);
	if (param == NULL) {
	    param = FindParameter(agg_datasetGroup, pc_paramName);
	}
	if (param != NULL) {
	    ppc_valueList = IK_Sequence2List(&i_numBytes, param,
					     IK_PACKAGE_ID_SZ);
	    assert(ppc_valueList != NULL);
	    pGranuleP->granule.package_id =
	      (char **) IK_RELOC(&pGranuleP->aligned_data,
				 pGranuleP->aligned_data);
	    pGranuleP = granule_list2c(pGranuleP, &sz_granuleBuffer,
				       ppc_valueList, i_numBytes);
	    if (pGranuleP == NULL) {
		goto bad_ODLtree;
	    }
	    pGranuleP->granule.package_selected = (IK_BOOLEAN *)
	      IK_RELOC(&pGranuleP->aligned_data, pGranuleP->aligned_data);
	    i_packageCount = param->value_count;
	    sz_alignedBytes = i_packageCount * sizeof(IK_BOOLEAN);
	    pGranuleP->aligned_data = (char *) pGranuleP->aligned_data +
	      sz_alignedBytes;
	    pGranuleP->nbytes += sz_alignedBytes;
	}

	/*------------------------------------------------------------------*/

	/* (Optional) Global coverage indicator */
	pc_paramName = "GLOBAL_GRANULE";
	param = FindParameter(agg_granuleGroup, pc_paramName);
	if (param == NULL) {
	    param = FindParameter(agg_datasetGroup, pc_paramName);
	}
	if (param != NULL) {
	    pGranuleP->granule.area.type = IK_GLOBAL;
	}

	/*------------------------------------------------------------------*/

	/* (Optional) Point location of granule */
	pc_groupName = "POINT_LOC";
	agg_coverageArea = FindGroup(agg_granuleGroup, pc_groupName);
	if (agg_coverageArea != NULL) {
	    pc_paramName = "LATITUDE";
	    param = FindParameter(agg_coverageArea, pc_paramName);
	    if (param != NULL) {
		value = FirstValue(param);
	    } else {
		goto bad_ODLtree;
	    }
	    if (value != NULL) {
		(void) ODLFormatReal(pGranuleP->granule.area.area.point.
				     latitude, &value->item);
	    } else {
		goto bad_ODLtree;
	    }
	    pc_paramName = "LONGITUDE";
	    param = FindParameter(agg_coverageArea, pc_paramName);
	    if (param != NULL) {
		value = FirstValue(param);
	    } else {
		goto bad_ODLtree;
	    }
	    if (value != NULL) {
		(void) ODLFormatReal(pGranuleP->granule.area.area.point.
				     longitude, &value->item);
	    } else {
		goto bad_ODLtree;
	    }
	    pGranuleP->granule.area.type = IK_POINT;
	}

	/*------------------------------------------------------------------*/

	/* (Optional) Polygon-delineated granule coverage */
	pc_groupName = "POLYGON_LOC";
	agg_coverageArea = FindGroup(agg_granuleGroup, pc_groupName);
	if (agg_coverageArea != NULL) {
	    pc_paramName = "LATITUDE";
	    param = FindParameter(agg_coverageArea, pc_paramName);
	    if (param != NULL) {
		value = FirstValue(param);
		for (i_numPolygonSides = 0; value &&
		     i_numPolygonSides < IK_NPOLYSIDES; i_numPolygonSides++) {
		    (void) ODLFormatReal(pGranuleP->granule.area.area.polygon.
					 latitude[i_numPolygonSides],
					 &value->item);
		    value = NextValue(value);
		}
	    } else {
		goto bad_ODLtree;
	    }
	    pc_paramName = "LONGITUDE";
	    param = FindParameter(agg_coverageArea, pc_paramName);
	    if (param != NULL) {
		value = FirstValue(param);
		for (i_numPolygonSides = 0; value &&
		     i_numPolygonSides < IK_NPOLYSIDES; i_numPolygonSides++) {
		    (void) ODLFormatReal(pGranuleP->granule.area.area.polygon.
					 longitude[i_numPolygonSides],
					 &value->item);
		    value = NextValue(value);
		}
	    } else {
		goto bad_ODLtree;
	    }
	    pc_paramName = "CENTROID_LAT";
	    param = FindParameter(agg_coverageArea, pc_paramName);
	    if (param != NULL) {
		value = FirstValue(param);
		if (value != NULL) {
		    (void) ODLFormatReal(pGranuleP->granule.area.area.polygon.
					 centroid_lat, &value->item);
		} else {
		    goto bad_ODLtree;
		}
	    } else {
		goto bad_ODLtree;
	    }
	    pc_paramName = "CENTROID_LON";
	    param = FindParameter(agg_coverageArea, pc_paramName);
	    if (param != NULL) {
		value = FirstValue(param);
		if (value != NULL) {
		    (void) ODLFormatReal(pGranuleP->granule.area.area.polygon.
					 centroid_long, &value->item);
		} else {
		    goto bad_ODLtree;
		}
	    } else {
		goto bad_ODLtree;
	    }
	    pGranuleP->granule.area.area.polygon.pole_inside = '\000';
	    pc_paramName = "POLE_INCLUDED";
	    param = FindParameter(agg_coverageArea, pc_paramName);
	    if (param != NULL) {
		value = FirstValue(param);
		if (value != NULL) {
		    if (value->item.value.string != NULL) {
			pGranuleP->granule.area.area.polygon.pole_inside =
			  *value->item.value.string;
		    }
		}
	    }
	    pGranuleP->granule.area.type = IK_POLYGON;
	}

	/*------------------------------------------------------------------*/

	/* Latitude and longitude ranges for the granule */
	pc_groupName = "RANGE_LOC";
	agg_coverageArea = FindGroup(agg_granuleGroup, pc_groupName);
	if (agg_coverageArea != NULL) {
	    pc_paramName = "NORTH_LATITUDE";
	    param = FindParameter(agg_coverageArea, pc_paramName);
	    if (param != NULL) {
		value = FirstValue(param);
		if (value != NULL) {
		    (void) ODLFormatReal(pGranuleP->granule.area.area.range.
					 north_lat, &value->item);
		} else {
		    goto bad_ODLtree;
		}
	    } else {
		goto bad_ODLtree;
	    }
	    pc_paramName = "SOUTH_LATITUDE";
	    param = FindParameter(agg_coverageArea, pc_paramName);
	    if (param != NULL) {
		value = FirstValue(param);
		if (value != NULL) {
		    (void) ODLFormatReal(pGranuleP->granule.area.area.range.
					 south_lat, &value->item);
		} else {
		    goto bad_ODLtree;
		}
	    } else {
		goto bad_ODLtree;
	    }
	    pc_paramName = "EAST_LONGITUDE";
	    param = FindParameter(agg_coverageArea, pc_paramName);
	    if (param != NULL) {
		value = FirstValue(param);
		if (value != NULL) {
		    (void) ODLFormatReal(pGranuleP->granule.area.area.range.
					 east_long, &value->item);
		} else {
		    goto bad_ODLtree;
		}
	    } else {
		goto bad_ODLtree;
	    }
	    pc_paramName = "WEST_LONGITUDE";
	    param = FindParameter(agg_coverageArea, pc_paramName);
	    if (param != NULL) {
		value = FirstValue(param);
		if (value != NULL) {
		    (void) ODLFormatReal(pGranuleP->granule.area.area.range.
					 west_long, &value->item);
		} else {
		    goto bad_ODLtree;
		}
	    } else {
		goto bad_ODLtree;
	    }
	    pGranuleP->granule.area.type = IK_RANGE;
	}

	/*------------------------------------------------------------------*/

	/* Since this is a new granule, make sure it is marked
           visible. */
	pGranuleP->granule.marked.visible = IK_VISIBLE;

	/* Make sure the aligned_data contains the offset to the next
           free byte in the heap area. */
	pGranuleP->aligned_data = (void *)
	  IK_RELOC(&pGranuleP->aligned_data, pGranuleP->aligned_data);

	/* Now that all ingest and memory management has been done,
           convert all of the offsets into pointers. */
	IK_PTR_FIXUP_LIST(pGranuleP, pGranuleP->granule.comment);
	IK_PTR_FIXUP_LIST(pGranuleP, pGranuleP->granule.sensor_name);
	IK_PTR_FIXUP_LIST(pGranuleP, pGranuleP->granule.source_name);
	IK_PTR_FIXUP_LIST(pGranuleP, pGranuleP->granule.parameter);
	IK_PTR_FIXUP_LIST(pGranuleP, pGranuleP->granule.browse_desc);
	IK_PTR_FIXUP_LIST(pGranuleP, pGranuleP->granule.package_id);
	pGranuleP->granule.start_date =
	  (char *) IK_FIXUP(&pGranuleP->aligned_data,
			    pGranuleP->granule.start_date);
	pGranuleP->granule.stop_date =
	  (char *) IK_FIXUP(&pGranuleP->aligned_data,
			    pGranuleP->granule.stop_date);
	if (pGranuleP->granule.package_selected != NULL) {
	    pGranuleP->granule.package_selected = (IK_BOOLEAN *)
	      IK_FIXUP(&pGranuleP->aligned_data,
		       pGranuleP->granule.package_selected);

	    /* Initialize all the package_selected variables to IK_FALSE */
	    for (i = 0; i < i_packageCount; i++) {
		pGranuleP->granule.package_selected[i] = IK_FALSE;
	    }

	}

	/* Build the sort key for this granule. */
	pSortKey_granule->usrapp1 =
	  IK_BuildGranuleKey(pGranuleP->granule.archive_tag,
			     pGranuleP->granule.dataset_tag,
			     pGranuleP->granule.granule_id);

#ifdef BRIDGE_CONFIG
	if (IK_bridge_already_found(pSortKey_granule->usrapp1)) {
	    continue;
	} else {
#endif

	    if (pSortKey_granule->usrapp1) {

		/* pGranule holds the current granule being fetched
                   from the database. */
		IK_Granule *pGranule = NULL;

		/* Add this sort key to the end of the granule key
		   linked list and make that granule key current. */
		pSortKey_granule = pSortKey_granule->usrapp1;

		/* Query the granule database for an old copy of the
		   granule we just ingested.  If a copy is not found,
		   update the granule counter, otherwise don't update
		   the counter. */
		if (IK_GetGranule(pSortKey_granule, &pGranule)) {
		    if (IK_ImsErrno == IK_ENOENT) {
			i_numGranules++;
			IK_ImsErrno = 0;
		    } else {
			pc_paramName = "IK_GetGranule()";
			goto bad_ODLtree;
		    }
		} else {

		    /* If we found it, save all of the marks, but the
                       information may have changed, so re-write
                       it. */
		    pGranuleP->granule.marked.brs_avail =
		      pGranule->marked.brs_avail;
		    pGranuleP->granule.marked.for_delete =
		      pGranule->marked.for_delete;
		    pGranuleP->granule.marked.for_coverage =
		      pGranule->marked.for_coverage;
		    pGranuleP->granule.marked.for_detail =
		      pGranule->marked.for_detail;
		    pGranuleP->granule.marked.for_order =
		      pGranule->marked.for_order;
		    pGranuleP->granule.marked.for_browse =
		      pGranule->marked.for_browse;
		    pGranuleP->granule.marked.selected =
		      pGranule->marked.selected;
		    pGranuleP->granule.marked.avail_browse_type =
		      pGranule->marked.avail_browse_type;

		    /* Release the granule, since we are about to
                       write over it. */
		    free(pGranule);

		}

		/* Store the new or updated granule to the
                   database. */
		if (IK_PutGranule((IK_Granule *) pGranuleP) < 0) {
		    IK_vSyslog(LOG_ERR, ".IK_IngestGranule: unable to "
			       "write the granule to disk, IK_ImsErrno = %s",
			       IK_GetMessage(IK_ImsErrno));
		}

	    } else {

		/* The key for this granule could not be created. */
		pc_paramName = "IK_BuildGranuleKey()";
		goto bad_ODLtree;

	    }

#ifdef BRIDGE_CONFIG
	    sigprocmask(SIG_UNBLOCK, &sig_set, 0);
	}
#endif
    }

#ifdef BRIDGE_CONFIG
    if (IK_IngestTimeout) {
        alarm(0);
        sigaction(SIGALRM, &oact, 0);
    }
#endif

    /*----------------------------------------------------------------------*/

    /* If the granules were ingested properly, ingest the DATASET
       group-level information. */
  good_ODLtree:
    {
	int err;

	/* Ingest the DATASET group information. */
	err = IK_IngestDataset(*pArchiveTag, *pDatasetTag, agg_datasetGroup,
			       i_numGranules);
	if (err) {

	    /* An error arose during the DATASET ingest, so delete all
               of the granules from this DATASET group from the
               granule database. Then log an error message. */
	    for (pSortKey_granule = &SortKey_granuleList;
		 pSortKey_granule->usrapp1;
		 pSortKey_granule = pSortKey_granule->usrapp1) {
		(void) IK_DeleteGranule((IK_SortKey *)
					pSortKey_granule->usrapp1);
	    }
	    IK_vSyslog(LOG_ERR, "IK_InvResultIngest.IK_IngestDataset %d",
		       IK_ImsErrno);

	} else {

	    /* The DATASET was properly ingested. */
	    (void) IK_GranuleCount('+', i_numGranules);

	    /* Log the granule count. */
	    pc_paramName = "DATASET_ID";
	    value = FirstValue(FindParameter(agg_datasetGroup, pc_paramName));
	    IK_vSyslog(LOG_NOTICE, "Ingested %d granules in %s dataset",
		       i_numGranules, value->item.value.string);

	}
	goto clean_up;
    }

    /*----------------------------------------------------------------------*/

    /* If an error arose during ingest, delete any ingested granules
       from this DATASET. */
  bad_ODLtree:
    {
	for (pSortKey_granule = &SortKey_granuleList;
	     pSortKey_granule->usrapp1;
	     pSortKey_granule = pSortKey_granule->usrapp1) {
	    (void) IK_DeleteGranule((IK_SortKey *) pSortKey_granule->usrapp1);
	}
	if (pc_groupName != NULL) {
	    IK_vSyslog(LOG_ERR,
		       "IK_IngestGranules(): Missing/invalid field %s in "
		       "group %s", pc_paramName, pc_groupName);
	} else {
	    IK_vSyslog(LOG_ERR,
		       "IK_IngestGranules(): Missing/invalid field %s",
		       pc_paramName);
	}
    }

    /*----------------------------------------------------------------------*/

    /* Perform general cleanup - memory deallocation, etc.. */
  clean_up:
    {
	IK_SortKey *last;

	/* Delete this DATASET group (WHY?!) */
	(void) RemoveAggregate(agg_datasetGroup);

	/* Reset errno to 0 to avoid problems with free() in
           RemoveAggregate(). NOTE: this should no longer be needed.
           It was added as a work around, and should now be taken care
           of by the use of ims_free() */
	errno = 0;

	/* Free allocated memory. */
	if (pArchiveTag) {
	    free(pArchiveTag);
	}
	if (pDatasetTag) {
	    free(pDatasetTag);
	}
	if (pGranuleP) {
	    free(pGranuleP);
	}

	/* Free each node in the linked list of granule keys. */
	last = NULL;
	for (pSortKey_granule = SortKey_granuleList.usrapp1;
	     pSortKey_granule; pSortKey_granule = last->usrapp1) {
	    if (last) {
		free(last);
	    }
	    last = pSortKey_granule;
	}
	if (last) {
	    free(last);
	}

    }

    /*----------------------------------------------------------------------*/

    /* Return the execution status. */
    return IK_ImsErrno ? -1 : 0;

}

/*****************************************************************************/

/*
 * Name: granule_list2c()
 *
 * Description: This function moves a list into a C structure. It
 * copies a NULL terminated list of null-terminated strings into the
 * granule structure. More memory is allocated if it is needed.
 *
 * Arguments:
 * IK_GranuleP *pGranuleP: Granule buffer
 * size_t *psz_granule: Size of granule buffer
 * char **ppc_stringList: List of strings
 * size_t sz_list: Size of list
 *
 * Return Values:
 * Upon success, a pointer to a granule. On failure, NULL.
 *
 * Warnings:
 *
 * Global Variables Used:
 * None
 *
 * Pre- and Post-Conditions:
 *
 * Revision History:
 *
*/

static IK_GranuleP *
granule_list2c(IK_GranuleP *pGranuleP, size_t *psz_granule,
	       char **ppc_stringList, size_t sz_list)
{

    /* ui_alignedSize is the aligned size of the list. */
    unsigned int ui_alignedSize;

    /*----------------------------------------------------------------------*/

    /* Adjust the list size to an aligned value. */
    ui_alignedSize = IK_ALIGN(sz_list);

    /* Do we need more buffer space to contain this list? If so,
       reallocate the buffer to be IK_MEMBUFSZ bytes larger. */
    if (pGranuleP->nbytes + ui_alignedSize > *psz_granule) {
	IK_GranuleP *pGranuleP_temp;
	pGranuleP_temp = realloc(pGranuleP, *psz_granule += IK_MEMBUFSIZ);
	if (pGranuleP_temp != NULL) {
	    pGranuleP = pGranuleP_temp;
	} else {
	    IK_ImsErrno = IK_ENOMEM;
	    free(pGranuleP);
	    pGranuleP = NULL;
	}
    }

    /* If we still have a buffer, then copy the list into the granule
       buffer while playing the Great Pointer Game. */
    if (pGranuleP != NULL) {

	/* Convert each pointer in the list to an offset from the
           first byte in the list. */
	IK_RELOC_LIST(ppc_stringList);

	/* Copy the entire list into the heap area at the end of the
           granule buffer. */
	(void) memcpy(pGranuleP->aligned_data, ppc_stringList, sz_list);

	/* Adjust the heap pointer to point to the next free aligned
           byte. */
	pGranuleP->aligned_data = (char *) pGranuleP->aligned_data +
	  ui_alignedSize;

	/* Update the byte usage count in the granule buffer. */
	pGranuleP->nbytes += ui_alignedSize;

	/* Free the original string list. */
	free(ppc_stringList);

    }

    /* Return a pointer to the modified granule buffer. */
    return(pGranuleP);

}
