/*
 * Name: IK_ListNode.c
 *
 * Description: This file contains functions for the IK_ListNode data
 * structure, which is used as the basis for linked lists.
 *
 * Notes:
 *
 *-------------------------------------------------------------------------
 *
 * RCS information:
 *
 * $RCSfile: IK_ListNode.c,v $
 *
 * $Id: IK_ListNode.c,v 1.1 1997/03/06 15:38:52 amrhein Exp $
 *
 * $Log: IK_ListNode.c,v $
 * Revision 1.1  1997/03/06 15:38:52  amrhein
 * Initial revision
 *
 * Revision 6.1.1.7  1996/11/23  20:15:05  winter
 * Made internal code more robust and self-consistent.
 *
 * Revision 6.1.1.6  1996/11/23  11:31:42  winter
 * Added code for IK_ListNode_Destroy().
 *
 * Revision 6.1.1.5  1996/11/19  16:45:08  winter
 * Added #include <stdio.h> for prototypes.
 *
 * Revision 6.1.1.4  1996/11/19  14:34:08  winter
 * Added code for printing NULL pointers in IK_ListNode_Dump().
 *
 * Revision 6.1.1.3  1996/11/19  14:28:42  winter
 * Added code for IK_ListNode_Dump(), a debugging function.
 *
 * Revision 6.1.1.2  1996/11/19  13:42:52  winter
 * Added IK_ListNode creation and deletion functions.
 *
 * Revision 6.1.1.1  1996/11/19  13:06:12  winter
 * Placeholder for new branch during Task Manager development.
 *
 * Revision 6.1  1996/11/19  13:04:57  winter
 * Initial checkin of stub file.
 *
 * */

/*****************************************************************************/

/* Define the RCS identifier string. */
static const char *rcsid = "$Id: IK_ListNode.c,v 1.1 1997/03/06 15:38:52 amrhein Exp $";

/*****************************************************************************/

/* #include directives */

/* This header must be #included first in order to use
   autoconf-generated information. */
#include "IK_Ims.h"

/* System headers */

/* Standard headers */
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

/* Third-party library headers */

/* IMS headers */
#include "IK_ListNode.h"

/*****************************************************************************/

/* Local #define directives */

/*****************************************************************************/

/* Data type definitions */

/*****************************************************************************/

/* External variable declarations */

/*****************************************************************************/

/* Local variable definitions */

/*****************************************************************************/

/* External function prototypes */

/*****************************************************************************/

/* Local function prototypes */

/* Creation and deletion functions */
static IK_ListNode *Allocate(void);
static void Free(IK_ListNode *plistnode);

/*****************************************************************************/

/*
 * Name: IK_ListNode_New()
 *
 * Description: This function creates a new IK_ListNode structure for
 * the supplied item, and returns a pointer to the new structure.
 *
 * Parameters:
 * void *pv_item - Pointer to data to be held by the node
 *
 * Return Value:
 * IK_ListNode *plistnode_new - Pointer to new IK_ListNode structure
 * for the supplied datum
 * NULL - If an error occurs
 *
 * Notes:
 * (1) NULL is a _valid_ value for pv_item!
 *
 * Revision History:
 *
 * Tuesday 19 November 1996 (Eric Winter) - Started initial version.
 *
 * Saturday 23 November 1996 (Eric Winter) - Changed to use call to
 * Allocate().
 * */

IK_ListNode *
IK_ListNode_New(void *pv_item)
{

    /* plistnode_new points to the new IK_ListNode structure. */
    IK_ListNode *plistnode_new;

    /*-----------------------------------------------------------------------*/

    /* Allocate the new list node. */
    plistnode_new = Allocate();
    if (plistnode_new == NULL) {
	return(NULL);
    }

    /* Store the supplied item in the list node. */
    plistnode_new->pv_item = pv_item;

    /* Initialize the next and previous node pointers to NULL, since
       this node is not yet part of a list. */
    plistnode_new->plistnode_next = NULL;
    plistnode_new->plistnode_prev = NULL;

    /* Return a pointer to the new node. */
    return(plistnode_new);

}

/*****************************************************************************/

/*
 * Name: IK_ListNode_Delete()
 *
 * Description: This function deletes a IK_ListNode structure. The
 * data stored in the structure is not affected.
 *
 * Parameters:
 * IK_ListNode *plistnode - Pointer to IK_ListNode to delete
 *
 * Return Value:
 * None
 *
 * Notes:
 * (1) If the data held by the node is allocated space, but is no
 * longer needed, it must be freed prior to this call.
 *
 * Revision History:
 *
 * Tuesday 19 November 1996 (Eric Winter) - Started initial version.
 *
 * Saturday 23 November 1996 (Eric Winter) - Changed to use call to
 * Free().
 * */

void
IK_ListNode_Delete(IK_ListNode *plistnode)
{

    /* Validate input. */
    assert(plistnode != NULL);

    /*-----------------------------------------------------------------------*/

    /* Delete the node. */
    Free(plistnode);

}

/*****************************************************************************/

/*
 * Name: IK_ListNode_Destroy()
 *
 * Description: This function deletes a IK_ListNode structure. The
 * data stored in the structure is also freed.
 *
 * Parameters:
 * IK_ListNode *plistnode - Pointer to IK_ListNode to destroy
 *
 * Return Value:
 * None
 *
 * Notes:
 * (1) If the data held by the node itself refers to allocated space,
 * but is no longer needed, it must be freed prior to this call.
 *
 * Revision History:
 *
 * Saturday 23 November 1996 (Eric Winter) - Started initial version.
 * */

void
IK_ListNode_Destroy(IK_ListNode *plistnode)
{

    /* Validate input. */
    assert(plistnode != NULL);

    /*-----------------------------------------------------------------------*/

    /* Free the memory stored in the list node (if any). */
    if (plistnode->pv_item != NULL) {
	free(plistnode->pv_item);
    }

    /* Delete the node. */
    Free(plistnode);

}

/*****************************************************************************/

/*
 * Name: IK_ListNode_Dump()
 *
 * Description: This debugging function dumps the contents of the
 * specified IK_ListNode to the specified stream.
 *
 * Parameters:
 * FILE *pFILE - Pointer to FILE for output
 * IK_ListNode *plistnode - Pointer to IK_ListNode to dump
 *
 * Return Value:
 * None
 *
 * Notes:
 *
 * Revision History:
 *
 * Tuesday 19 November 1996 (Eric Winter) - Started initial version.
 * */

void
IK_ListNode_Dump(FILE *pFILE, IK_ListNode *plistnode)
{

    /* Validate input. */
    assert(pFILE != NULL);

    /*-----------------------------------------------------------------------*/

    /* Print the node contents to the stream. */
    (void) fprintf(pFILE, "Dumping IK_ListNode at address %p:\n", plistnode);
    if (plistnode != NULL) {
	(void) fprintf(pFILE, "pv_item = %p\n", plistnode->pv_item);
	(void) fprintf(pFILE, "plistnode_next = %p\n",
		       plistnode->plistnode_next);
	(void) fprintf(pFILE, "plistnode_prev = %p\n",
		       plistnode->plistnode_prev);
    } else {
	(void) fprintf(pFILE, "(NULL)\n");
    }

}

/*****************************************************************************/

/* FUNCTIONS BELOW THIS POINT ARE PRIVATE TO THIS SOURCE FILE ONLY! */

/*****************************************************************************/

/*
 * Name: Allocate()
 *
 * Description: This function allocates space for a new IK_ListNode
 * structure.
 *
 * Parameters:
 * None
 *
 * Return Value:
 * IK_ListNode *plistnode_new - Pointer to new IK_ListNode structure
 * NULL - If an error occurs
 *
 * Notes:
 * (1) This function is used to provide flexibility in the node
 * allocation mechanism. For example, a list of nodes can be
 * pre-allocated and doled out from this function. For now, it simply
 * acts as a front-end to malloc().
 *
 * Revision History:
 *
 * Tuesday 19 November 1996 (Eric Winter) - Started initial version.
 *
 * Saturday 23 November 1996 (Eric Winter) - Changed name to
 * Allocate().
 * */

static IK_ListNode *
Allocate(void)
{

    /* plistnode_new points to the new IK_ListNode structure. */
    IK_ListNode *plistnode_new;

    /*-----------------------------------------------------------------------*/

    /* Allocate and return the new list node. */
    return(malloc(sizeof(IK_ListNode)));

}

/*****************************************************************************/

/*
 * Name: Free()
 *
 * Description: This function frees the memory allocated to a
 * IK_ListNode structure. The data stored in the structure is not
 * affected.
 *
 * Parameters:
 * IK_ListNode *plistnode - Pointer to IK_ListNode to free
 *
 * Return Value:
 * None
 *
 * Notes:
 * (1) See note (1) for IK_ListNode_Allocate().
 *
 * Revision History:
 *
 * Tuesday 19 November 1996 (Eric Winter) - Started initial version.
 *
 * Saturday 23 November 1996 (Eric Winter) - Changed name to Free().
 * */

static void
Free(IK_ListNode *plistnode)
{

    /* Validate input. */
    assert(plistnode != NULL);

    /*-----------------------------------------------------------------------*/

    /* Delete the node. */
    free(plistnode);

}
