/*
 * Name: IK_Map.c
 *
 * Description: This file contains routines to map identifiers to
 * other identifiers, such as DAAC names to tags and vice-versa.
 *
 * Notes:
 *
 *-----------------------------------------------------------------------------
 *
 * RCS information:
 *
 * $RCSfile: IK_Map.c,v $
 *
 * $Id: IK_Map.c,v 1.6 1997/03/06 15:38:52 amrhein Exp $
 *
 * $Log: IK_Map.c,v $
 * Revision 1.6  1997/03/06 15:38:52  amrhein
 * Updated to GUI's 6_1_12 IK
 *
 * Revision 6.1.1.3  1997/02/11  20:32:54  winter
 * Moved IK_ValidStringToIndex() to IK_NewValids.[ch].
 *
 * Revision 6.1.1.2  1997/02/11  12:02:09  winter
 * Added code for IK_MapArchiveTagToIndex(), IK_MapArchiveIndexToTag(),
 * IK_MapDatasetTagToIndex() and IK_MapDatasetIndexToTag().
 *
 * Revision 6.1.1.1  1997/02/11  01:44:11  winter
 * Placeholder for new branch for addition of utility functions.
 *
 * Revision 6.1  1996/10/18  14:16:09  ims
 * Promoted for 6.1 J. Katen
 *
 * Revision 6.0  1996/02/29  15:39:58  ims
 * COPIED FROM 5.1.1.6 FOR GUI_PROMO_6_0_960229.
 *
 * Revision 5.1.1.6  1996/02/23  00:09:19  winter
 * Fixed SMR 3415 and 3419. Fixed functions which map names to tags so that
 * they return NULL if the strings are not found.
 *
 * Revision 5.1.1.5  1996/02/15  20:21:58  winter
 * Fixed all functions to account for new tag type (again).
 *
 * Revision 5.1.1.4  1996/02/14  20:27:35  winter
 * Replaced IK_MapArchiveIdent() with IK_MapArchiveNameToTag() and
 * IK_MapArchiveTagToName(). Replaced IK_MapDatasetIdent() with
 * IK_MapDatasetNameToTag() and IK_MapDatasetTagToName().
 *
 * Revision 5.1.1.3  1996/01/17  20:52:30  winter
 * Added #include "IK_Map.h".
 *
 * Revision 5.1.1.2  1995/12/28  20:00:57  ims
 * COPIED FROM 5.0.1.2 FOR GUI_PROMO_5_1_951228.
 *
 * Revision 5.0.1.2  1995/12/22  14:38:23  winter
 * Temporary revision - added hack for using a 2-byte DAAC tag.
 *
 * Revision 5.0.1.1  1995/12/21  00:52:39  winter
 * Placeholder for new branch.
 *
 * Revision 5.0  1995/11/06  13:03:52  ims
 * PROMOTION FROM GUI_4_5_8_951030 FOR GUI_PROMO_5_0_951106
 *
 * Revision 4.5.1.2  1995/09/27  18:42:48  winter
 * Cleaned up. Removed unused header files.
 *
 * Revision 4.5.1.1  1995/07/27  19:02:34  ims
 * COPIED FROM 4.4.1.1 FOR GUI_PROMO_4_5_950726
 *
 * Revision 4.4.1.1  1995/05/24  20:17:51  ims
 * The source code uses FILENAME_MAX to define size for strings that
 * will hold pathnames. Since HP/UX still supports filesystems with
 * short filenames (15 characters), FILENAME_MAX is set to 15 for HP/UX.
 *
 * Revision 4.4  1995/02/13  22:20:36  ims
 * PROMOTION FROM GUI_4_4_950106 FOR GUI_PROMO_4_4_950213
 *
 * Revision 4.3  1994/08/26  10:49:45  ims
 * COPIED FROM GUI_4_3_940824.
 *
 * Revision 4.0  1994/06/08  14:57:30  ims
 * PROMOTION FROM REV4
 *
 * Revision 3.1  1994/04/05  09:49:52  ims
 * Promotion from MAR28_RELEASE
 *
 * Revision 3.0.1.1  1994/01/27  19:36:33  winter
 * Placeholder revision to begin code cleanup.
 *
 * Revision 3.0  1993/12/29  15:18:49  winter
 * ALPHA-1
 *
 * Revision 1.3  1992/10/14  20:35:01  honce
 * Added comment about the importance of counting the NULL.
 *
 * Revision 1.2  1992/10/05  16:42:39  honce
 * Added prototype for IK_MapGdbmErrno();
 *
 * Revision 1.1  1992/10/01  18:33:08  honce
 * Initial revision
 *
*/

/*****************************************************************************/

/* Define the RCS identifier string. */
static const char *rcsid = "$Id: IK_Map.c,v 1.6 1997/03/06 15:38:52 amrhein Exp $";

/*****************************************************************************/

/* #include directives */

/* This header must be #included first in order to use
   autoconf-generated information. */
#include "IK_Ims.h"

/* Standard headers */
#include <assert.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Third-party library headers */

/* IMS headers */
#include "IK_Map.h"
#include "IK_Misc.h"
#include "IK_NewValids.h"
#include "IK_ValNames.h"

/*****************************************************************************/

/* Local #define directives */

/*****************************************************************************/

/*
 * Name: IK_MapArchiveNameToTag()
 *
 * Description: This function maps an archive name to an archive
 * tag. The tags are just the indices of the DAAC names in the
 * dependent valids string list. A newly-allocated copy of the tag
 * value is returned, to maintain compatibility with existing code. If
 * the archive name does not exist in the valids list, return NULL.
 *
 * Parameters:
 * const char *pc_archiveName - Name of archive to map; must be
 * non-NULL
 *
 * Return Values:
 * IK_ArchiveTag *pArchiveTag - New copy of the tag for archive named
 * in pc_archiveName
 * NULL- If an error occurs
 *
 * Global Variables Used:
 * None
 *
 * Notes:
 *
 * This function assumes that the IK_ArchiveTag type is defined as a
 * 4-byte value (an int). If this is not the case, the results are
 * unpredictable.
 *
 * Revision History:
 *
 * Wednesday 14 February 1996 (Eric Winter) - Initial version.
 *
 * Thursday 22 February 1996 (Eric Winter) - Modified to return NULL
 * if the archive name is not found, instead of aborting on a failed
 * assert() call.
 *
 * Monday 10 February 1997 (Eric Winter) - Modified to use
 * IK_MapArchiveIndexToTag().
 *
 * Tuesday 11 February 1997 (Eric Winter) - Changed to use
 * VALIDS_DAAC.
 * */

IK_ArchiveTag *
IK_MapArchiveNameToTag(const char *pc_archiveName)
{

    /* i_index is the index value returned from
       IK_ValidStringToIndex(). */
    int i_index;

    /* pArchiveTag is a pointer to a new copy of the archive tag for
       this archive name. */
    IK_ArchiveTag *pArchiveTag;

    /*-----------------------------------------------------------------------*/

    /* Validate input. */
    assert(pc_archiveName != NULL);

    /*-----------------------------------------------------------------------*/

    /* Map the archive name to the tag value. */
    i_index = IK_ValidStringToIndex(pc_archiveName, VALIDS_DAAC);

    /* If a valid tag was found, make a copy of it. */
    if (i_index != -1) {

	/* Map the index to a tag. */
	pArchiveTag = IK_MapArchiveIndexToTag(i_index);
	assert(pArchiveTag != NULL);

    } else {
	pArchiveTag = NULL;
    }

    /*-----------------------------------------------------------------------*/

    /* Return a pointer to the new copy of the archive tag, or NULL if
       a match was not found. */
    return(pArchiveTag);

}

/*****************************************************************************/

/*
 * Name: IK_MapArchiveTagToName()
 *
 * Description: This function maps an archive tag to an archive
 * name. The tags are just the indices of the DAAC names in the
 * dependent valids string list. A newly-allocated copy of the archive
 * name is returned.
 *
 * Parameters:
 * IK_ArchiveTag archiveTag - Tag for archive to map to a name; must
 * be >= 0
 *
 * Return Values:
 * char *pc_archiveName - New copy of archive name
 * NULL - If an error occurs
 *
 * Global Variables Used:
 * None
 *
 * Notes:
 *
 * This function assumes that the IK_ArchiveTag type is defined as a
 * 4-byte value (an int). If this is not the case, the results are
 * unpredictable.
 *
 * Revision History:
 *
 * Wednesday 14 February 1996 (Eric Winter) - Initial version.
 *
 * Monday 10 February 1997 (Eric Winter) - Modified to use
 * IK_MapArchiveTagToIndex().
 * */

char *
IK_MapArchiveTagToName(IK_ArchiveTag archiveTag)
{

    /* i_index holds the integer version of the tag. */
    int i_index;

    /* pc_tempArchiveName holds the return value from
       IK_ValidName(). */
    const char *pc_tempArchiveName;

    /* pc_archiveName is the newly-allocated copy of the archive name
       for this archive tag. */
    char *pc_archiveName;

    /*-----------------------------------------------------------------------*/

    /* Validate input. */
    assert(archiveTag != NULL);

    /*-----------------------------------------------------------------------*/

    /* Copy the tag to an integer. */
    i_index = IK_MapArchiveTagToIndex(archiveTag);
    assert(i_index != -1);

    /* Map the archive tag to the name. */
    pc_tempArchiveName = IK_ValidName(i_index);
    assert(pc_tempArchiveName != NULL);

    /* Duplicate the archive name. */
    pc_archiveName = IK_DuplicateString(pc_tempArchiveName);
    assert(pc_archiveName != NULL);

    /*-----------------------------------------------------------------------*/

    /* Return the archive name. */
    return(pc_archiveName);

}

/*****************************************************************************/

/*
 * Name: IK_MapArchiveTagToIndex()
 *
 * Description: This function maps an archive tag to an index in the
 * valids string table. The tags are just the indices of the archive
 * names in the dependent valids string list.
 *
 * Parameters:
 * IK_ArchiveTag archiveTag - Tag for archive to map to an index; must
 * be >= 0
 *
 * Return Values:
 * int i_index - Integer equivalent of supplied tag
 *
 * Global Variables Used:
 * None
 *
 * Notes:
 *
 * This function assumes that the IK_ArchiveTag type is defined as a
 * 4-byte value (an int). If this is not the case, the results are
 * unpredictable.
 *
 * This function cannot fail if the tag is non-NULL.
 *
 * Revision History:
 *
 * Monday 10 February 1997 (Eric Winter) - Initial version.
 *
 * */

int
IK_MapArchiveTagToIndex(IK_ArchiveTag archiveTag)
{

    /* i_index holds the integer version of the tag. */
    int i_index;

    /*-----------------------------------------------------------------------*/

    /* Validate input. */
    assert(archiveTag != NULL);

    /*-----------------------------------------------------------------------*/

    /* Copy the tag to an integer. */
    (void) memcpy(&i_index, archiveTag, sizeof(IK_ArchiveTag));

    /*-----------------------------------------------------------------------*/

    /* Return the archive index. */
    return(i_index);

}

/*****************************************************************************/

/*
 * Name: IK_MapArchiveIndexToTag()
 *
 * Description: This function maps an archive index to an archive tag.
 * The tags are just the indices of the archive names in the dependent
 * valids string list.
 *
 * Parameters:
 * int i_index - Index to map; must be >= 0
 *
 * Return Values:
 * IK_ArchiveTag *parchiveTag - Pointer to new IK_ArchiveTag
 *
 * Global Variables Used:
 * None
 *
 * Notes:
 *
 * This function assumes that the IK_ArchiveTag type is defined as a
 * 4-byte value (an int). If this is not the case, the results are
 * unpredictable.
 *
 * Given a non-negative integer value, and barring a memory allocation
 * error, this function cannot fail.
 *
 * Revision History:
 *
 * Monday 10 February 1997 (Eric Winter) - Initial version.
 *
 * */

IK_ArchiveTag *
IK_MapArchiveIndexToTag(int i_index)
{

    /* pArchiveTag points to the new tag. */
    IK_ArchiveTag *pArchiveTag;

    /*-----------------------------------------------------------------------*/

    /* Validate input. */
    if (i_index < 0) {
	return(NULL);
    }

    /*-----------------------------------------------------------------------*/

    /* Allocate space for a copy of the archive tag. */
    pArchiveTag = (IK_ArchiveTag *) calloc(1, sizeof(IK_ArchiveTag));
    assert(pArchiveTag != NULL);

    /* Copy the archive tag. */
    (void) memcpy(pArchiveTag, &i_index, sizeof(IK_ArchiveTag));

    /*-----------------------------------------------------------------------*/

    /* Return the pointer to the new tag. */
    return(pArchiveTag);

}

/*****************************************************************************/

/*
 * Name: IK_MapDatasetNameToTag()
 *
 * Description: This function maps a dataset name to a dataset
 * tag. The tags are just the indices of the dataset names in the
 * dependent valids string list. A newly-allocated copy of the tag
 * value is returned, to maintain compatibilty with existing code. If
 * the data set name does not exist in the valids list, return NULL.
 *
 * Parameters:
 * const char *pc_datasetName - Name of dataset to map; must be
 * non-NULL
 *
 * Return Values:
 * IK_DatasetTag *pDatasetTag - Pointer to new copy of the tag for
 * dataset named in pc_datasetName
 * NULL - If an error occurs
 *
 * Global Variables Used:
 * None
 *
 * Notes:
 *
 * This function assumes that the IK_DatasetTag type is defined as a
 * 4-byte value (an int). If this is not the case, the results are
 * unpredictable.
 *
 * Revision History:
 *
 * Wednesday 14 February 1996 (Eric Winter) - Initial version.
 *
 * Thursday 22 February 1996 (Eric Winter) - Modified to return NULL
 * if the data set name is not found, instead of aborting on a failed
 * assert() call.
 *
 * Monday 10 February 1997 (Eric Winter) - Modified to use
 * IK_MapDatasetIndexToTag().
 *
 * Tuesday 11 February 1997 (Eric Winter) - Changed to use
 * VALIDS_DATASET.
 * */

IK_DatasetTag *
IK_MapDatasetNameToTag(const char *pc_datasetName)
{

    /* i_index is a copy of the value returned from
       IK_ValidStringToIndex(). */
    int i_index;

    /* pDatasetTag points to a new copy of the dataset tag for this
       dataset name. */
    IK_DatasetTag *pDatasetTag;

    /*-----------------------------------------------------------------------*/

    /* Validate input. */
    assert(pc_datasetName != NULL);

    /*-----------------------------------------------------------------------*/

    /* Map the dataset name to the tag value. */
    i_index = IK_ValidStringToIndex(pc_datasetName, VALIDS_DATASET);

    /* If a valid tag was found, make a copy of it. */
    if (i_index != -1) {

	/* Map the index to a tag. */
	pDatasetTag = IK_MapDatasetIndexToTag(i_index);
	assert(pDatasetTag != NULL);

    } else {
	pDatasetTag = NULL;
    }

    /*-----------------------------------------------------------------------*/

    /* Return a pointer to the new dataset tag, or NULL if a match was
       not found. */
    return(pDatasetTag);

}

/*****************************************************************************/

/*
 * Name: IK_MapDatasetTagToName()
 *
 * Description: This function maps a dataset tag to a dataset
 * name. The tags are just the indices of the dataset names in the
 * dependent valids string list. A newly-allocated copy of the dataset
 * name is returned.
 *
 * Parameters:
 * IK_DatasetTag datasetTag - Tag for dataset to map to a name; must
 * be >= 0
 *
 * Return Values:
 * char *pc_datasetName - New copy of dataset name
 * NULL - If an error occurs
 *
 * Global Variables Used:
 * None
 *
 * Notes:
 *
 * This function assumes that the IK_DatasetTag type is defined as a
 * 4-byte value (an int). If this is not the case, the results are
 * unpredictable.
 *
 * Revision History:
 *
 * Wednesday 14 February 1996 (Eric Winter) - Initial version.
 *
 * Monday 10 February 1997 (Eric Winter) - Modified to use
 * IK_MapDatasetTagToIndex().
 * */

char *
IK_MapDatasetTagToName(IK_DatasetTag datasetTag)
{

    /* i_index holds the integer version of the dataset tag. */
    int i_index;

    /* pc_tempDatasetName holds the return value from
       IK_ValidName(). */
    const char *pc_tempDatasetName;

    /* pc_datasetName is the newly-allocated copy of the dataset name
       for this dataset tag. */
    char *pc_datasetName;

    /*-----------------------------------------------------------------------*/

    /* Validate input. */
    assert(datasetTag != NULL);

    /*-----------------------------------------------------------------------*/

    /* Copy the tag to an integer. */
    i_index = IK_MapDatasetTagToIndex(datasetTag);
    assert(i_index != -1);

    /* Map the dataset tag to the name. */
    pc_tempDatasetName = IK_ValidName(i_index);
    assert(pc_tempDatasetName != NULL);

    /* Duplicate the dataset name. */
    pc_datasetName = IK_DuplicateString(pc_tempDatasetName);
    assert(pc_datasetName != NULL);

    /*-----------------------------------------------------------------------*/

    /* Return the dataset name. */
    return(pc_datasetName);

}

/*****************************************************************************/

/*
 * Name: IK_MapDatasetTagToIndex()
 *
 * Description: This function maps a dataset tag to an index in the
 * valids string table. The tags are just the indices of the dataset
 * names in the dependent valids string list.
 *
 * Parameters:
 * IK_DatasetTag datasetTag - Tag for dataset to map to an index; must
 * be >= 0
 *
 * Return Values:
 * int i_index - Integer equivalent of supplied tag
 *
 * Global Variables Used:
 * None
 *
 * Notes:
 *
 * This function assumes that the IK_DatasetTag type is defined as a
 * 4-byte value (an int). If this is not the case, the results are
 * unpredictable.
 *
 * This function cannot fail if the tag is non-NULL.
 *
 * Revision History:
 *
 * Monday 10 February 1997 (Eric Winter) - Initial version.
 *
 * */

int
IK_MapDatasetTagToIndex(IK_DatasetTag datasetTag)
{

    /* i_index holds the integer version of the tag. */
    int i_index;

    /*-----------------------------------------------------------------------*/

    /* Validate input. */
    assert(datasetTag != NULL);

    /*-----------------------------------------------------------------------*/

    /* Copy the tag to an integer. */
    (void) memcpy(&i_index, datasetTag, sizeof(IK_DatasetTag));

    /*-----------------------------------------------------------------------*/

    /* Return the dataset index. */
    return(i_index);

}

/*****************************************************************************/

/*
 * Name: IK_MapDatasetIndexToTag()
 *
 * Description: This function maps a dataset index to a dataset tag.
 * The tags are just the indices of the dataset names in the dependent
 * valids string list.
 *
 * Parameters:
 * int i_index - Index to map; must be >= 0
 *
 * Return Values:
 * IK_DatasetTag *pDatasetTag - Pointer to new IK_DatasetTag
 *
 * Global Variables Used:
 * None
 *
 * Notes:
 *
 * This function assumes that the IK_DatasetTag type is defined as a
 * 4-byte value (an int). If this is not the case, the results are
 * unpredictable.
 *
 * Given a non-negative integer value, and barring a memory allocation
 * error, this function cannot fail.
 *
 * Revision History:
 *
 * Monday 10 February 1997 (Eric Winter) - Initial version.
 *
 * */

IK_DatasetTag *
IK_MapDatasetIndexToTag(int i_index)
{

    /* pDatasetTag points to the new tag. */
    IK_DatasetTag *pDatasetTag;

    /*-----------------------------------------------------------------------*/

    /* Validate input. */
    if (i_index < 0) {
	return(NULL);
    }

    /*-----------------------------------------------------------------------*/

    /* Allocate space for a copy of the dataset tag. */
    pDatasetTag = (IK_DatasetTag *) calloc(1, sizeof(IK_DatasetTag));
    assert(pDatasetTag != NULL);

    /* Copy the dataset tag. */
    (void) memcpy(pDatasetTag, &i_index, sizeof(IK_DatasetTag));

    /*-----------------------------------------------------------------------*/

    /* Return the pointer to the new tag. */
    return(pDatasetTag);

}
