/*
 * Name: IK_RingList.c
 *
 * Description: This file contains functions for the IK_RingList data
 * structure, which implements a ring-shaped doubly-linked list. The
 * user never sees the list node structures - all access is based on
 * the items stored in the nodes in the list.
 *
 * Notes:
 *
 *-------------------------------------------------------------------------
 *
 * RCS information:
 *
 * $RCSfile: IK_RingList.c,v $
 *
 * $Id: IK_RingList.c,v 1.1 1997/03/06 15:38:52 amrhein Exp $
 *
 * $Log: IK_RingList.c,v $
 * Revision 1.1  1997/03/06 15:38:52  amrhein
 * Initial revision
 *
 * Revision 6.1.1.8  1996/11/25  12:33:05  winter
 * Added IK_RingList_GetNextItem() and IK_RingList_GetPrevItem(). Changed
 * static function FindItem() to FindNodeByItem().
 *
 * Revision 6.1.1.7  1996/11/25  12:23:41  winter
 * Added IK_RingList_GetHeadItem().
 *
 * Revision 6.1.1.6  1996/11/23  20:14:35  winter
 * Changed to use pv_item instead of pv_nodeData for list item internals.
 *
 * Revision 6.1.1.5  1996/11/23  19:48:44  winter
 * Completely overhauled to remove use of list nodes by caller. Made internals
 * more robust.
 *
 * Revision 6.1.1.5  1996/11/23  19:48:44  winter
 * Completely overhauled to remove use of list nodes by caller. Made internals
 * more robust.
 *
 * Revision 6.1.1.4  1996/11/21  19:47:57  winter
 * Added code for IK_RingList_CountNodes() and IK_RingList_RemoveNode().
 *
 * Revision 6.1.1.3  1996/11/19  19:32:22  winter
 * Added code for IK_RingList_AddNode().
 *
 * Revision 6.1.1.2  1996/11/19  16:45:47  winter
 * Added creation, deletion and dump functions.
 *
 * Revision 6.1.1.1  1996/11/19  15:08:58  winter
 * Placeholder for new branch during Task Manager development.
 *
 * Revision 6.1  1996/11/19  15:07:34  winter
 * Initial checkin of stub file.
 *
 * */

/*****************************************************************************/

/* Define the RCS identifier string. */
static const char *rcsid = "$Id: IK_RingList.c,v 1.1 1997/03/06 15:38:52 amrhein Exp $";

/*****************************************************************************/

/* #include directives */

/* This header must be #included first in order to use
   autoconf-generated information. */
#include "IK_Ims.h"

/* System headers */

/* Standard headers */
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

/* Third-party library headers */

/* IMS headers */
#include "IK_ListNode.h"
#include "IK_RingList.h"

/*****************************************************************************/

/* Local #define directives */

/*****************************************************************************/

/* Data type definitions */

/*****************************************************************************/

/* External variable declarations */

/*****************************************************************************/

/* Local variable definitions */

/*****************************************************************************/

/* External function prototypes */

/*****************************************************************************/

/* Local function prototypes */

/* Creation and deletion functions */
static IK_RingList *Allocate(void);
static void Free(IK_RingList *pringlist);

/* Node addition and removal functions */
static void AddNode(IK_RingList *pringlist, IK_ListNode *plistnode);
static void RemoveNode(IK_RingList *pringlist, IK_ListNode *plistnode);

/* Data retrieval and lookup functions */
static IK_ListNode *FindNodeByItem(IK_RingList *pringlist, void *pv_item);

/*****************************************************************************/

/*
 * Name: IK_RingList_New()
 *
 * Description: This function creates and initializes a new, empty
 * IK_RingList structure.
 *
 * Parameters:
 * None
 *
 * Return Value:
 * IK_RingList *pringlist_new - Pointer to new IK_RingList structure
 * NULL - If an error occurs
 *
 * Notes:
 *
 * Revision History:
 *
 * Tuesday 19 November 1996 (Eric Winter) - Started initial version.
 *
 * Saturday 23 November 1996 (Eric Winter) - Removed argument. List is
 * now created empty automatically.
 * */

IK_RingList *
IK_RingList_New(void)
{

    /* pringlist_new points to the new IK_RingList structure. */
    IK_RingList *pringlist_new;

    /*-----------------------------------------------------------------------*/

    /* Allocate the new ring list. */
    pringlist_new = Allocate();
    if (pringlist_new == NULL) {
	return(NULL);
    }

    /* Clear the head node pointer. */
    pringlist_new->plistnode_head = NULL;

    /* Return a pointer to the new list. */
    return(pringlist_new);

}

/*****************************************************************************/

/*
 * Name: IK_RingList_Delete()
 *
 * Description: This function deletes a IK_RingList structure. The
 * data stored in the list nodes is not affected.
 *
 * Parameters:
 * IK_RingList *pringlist - Pointer to IK_RingList to delete
 *
 * Return Value:
 * None
 *
 * Notes:
 *
 * Revision History:
 *
 * Tuesday 19 November 1996 (Eric Winter) - Started initial version.
 *
 * Saturday 23 November 1996 (Eric Winter) - Added call to
 * IK_RingList_RemoveAllItems().
 * */

void
IK_RingList_Delete(IK_RingList *pringlist)
{

    /* Validate input. */
    assert(pringlist != NULL);

    /*-----------------------------------------------------------------------*/

    /* Remove each node from the list. */
    IK_RingList_RemoveAllItems(pringlist);

    /* Delete the list. */
    Free(pringlist);

}

/*****************************************************************************/

/*
 * Name: IK_RingList_AddItem()
 *
 * Description: This function creates a new list node for the data
 * referred to by pv_item, and adds the list node to the ring list.
 *
 * Parameters:
 * IK_RingList *pringlist - Pointer to ring list add item to
 * void *pv_item - Pointer to item data to store (may be NULL)
 *
 * Return Value:
 * IK_BOOLEAN IK_TRUE - If the item was added
 * IK_BOOLEAN IK_FALSE - If not
 *
 * Notes:
 *
 * Revision History:
 *
 * Saturday 23 November 1996 (Eric Winter) - Started initial version.
 * */

IK_BOOLEAN
IK_RingList_AddItem(IK_RingList *pringlist, void *pv_item)
{

    /* plistnode points to the list node for the item to add. */
    IK_ListNode *plistnode;

    /*-----------------------------------------------------------------------*/

    /* Validate input. */
    assert(pringlist != NULL);

    /*-----------------------------------------------------------------------*/

    /* Create a list node for the item. */
    plistnode = IK_ListNode_New(pv_item);
    if (plistnode == NULL) {
	return(IK_FALSE);
    }

    /* Add the node to the list. */
    AddNode(pringlist, plistnode);

    /* Return normally. */
    return(IK_TRUE);

}

/*****************************************************************************/

/*
 * Name: IK_RingList_RemoveItem()
 *
 * Description: This function removes the list node for the specified
 * item from the specified list. The list node is deleted, but the
 * item itself is not affected.
 *
 * Parameters:
 * IK_RingList *pringlist - Pointer to ring list to remove item from
 * void *pv_item - Pointer to item data to remove (may be NULL)
 *
 * Return Value:
 * IK_BOOLEAN IK_TRUE - If the item was removed
 * IK_BOOLEAN IK_FALSE - If not
 *
 * Notes:
 *
 * Revision History:
 *
 * Saturday 23 November 1996 (Eric Winter) - Started initial version.
 * */

IK_BOOLEAN
IK_RingList_RemoveItem(IK_RingList *pringlist, void *pv_item)
{

    /* plistnode points to the list node for the item to remove. */
    IK_ListNode *plistnode;

    /*-----------------------------------------------------------------------*/

    /* Validate input. */
    assert(pringlist != NULL);

    /*-----------------------------------------------------------------------*/

    /* Find the item on the list. */
    plistnode = FindNodeByItem(pringlist, pv_item);
    if (plistnode == NULL) {
	return(IK_FALSE);
    }

    /* Remove the node from the list and delete the node. */
    RemoveNode(pringlist, plistnode);
    IK_ListNode_Delete(plistnode);

    /* Return normally. */
    return(IK_TRUE);

}

/*****************************************************************************/

/*
 * Name: IK_RingList_RemoveAllItems()
 *
 * Description: This function removes all the nodes from the specified
 * list. The data contained within the nodes is not affected.
 *
 * Parameters:
 * IK_RingList *pringlist - Pointer to ring list to empty
 *
 * Return Value:
 * None
 *
 * Notes:
 *
 * Revision History:
 *
 * Saturday 23 November 1996 (Eric Winter) - Started initial version.
 * */

void
IK_RingList_RemoveAllItems(IK_RingList *pringlist)
{

    /* This is a temporary pointer to the current list node. */
    IK_ListNode *plistnode;

    /*-----------------------------------------------------------------------*/

    /* Validate input. */
    assert(pringlist != NULL);

    /*-----------------------------------------------------------------------*/

    /* Delete each node, from the head. Note that if the head node is
       NULL, the list is already empty, and this loop is not
       executed. */
    while (pringlist->plistnode_head != NULL) {
	plistnode = pringlist->plistnode_head;
	RemoveNode(pringlist, plistnode);
	IK_ListNode_Delete(plistnode);
    }

}

/*****************************************************************************/

/*
 * Name: IK_RingList_CountItems()
 *
 * Description: This function returns a count of the number of nodes
 * in a list. Note that nodes containing the same item are counted
 * once per occurrence.
 *
 * Parameters:
 * IK_RingList *pringlist - Pointer to IK_RingList to count items in
 *
 * Return Value:
 * unsigned int ui_numItems - Item count
 *
 * Notes:
 *
 * Revision History:
 *
 * Tuesday 19 November 1996 (Eric Winter) - Started initial version.
 *
 * Saturday 23 November 1996 (Eric Winter) - Changed name to
 * IK_RingList_CountItems().
 *
 * */

unsigned int
IK_RingList_CountItems(IK_RingList *pringlist)
{

    /* ui_numNodes is the number of nodes in the list. The number of
       nodes is identical to the number of items in the list, since
       one item is stored in each node (even if that item is NULL). */
    unsigned int ui_numNodes;

    /* plistnode is a temporary pointer to the current list node. */
    IK_ListNode *plistnode;

    /*-----------------------------------------------------------------------*/

    /* Validate input. */
    assert(pringlist != NULL);

    /*-----------------------------------------------------------------------*/

    /* If the list is empty, there are no nodes. If not, count the
       nodes, starting at the head. */
    ui_numNodes = 0;
    if (pringlist->plistnode_head != NULL) {
	plistnode = pringlist->plistnode_head;
	do {
	    assert(plistnode != NULL);
	    ui_numNodes++;
	    plistnode = plistnode->plistnode_next;
	} while (plistnode != pringlist->plistnode_head);
    }

    /* Return the node count. */
    return(ui_numNodes);

}

/*****************************************************************************/

/*
 * Name: IK_RingList_GetHeadItem()
 *
 * Description: This function returns a pointer to the item stored in
 * the head list node.
 *
 * Parameters:
 * IK_RingList *pringlist - Pointer to IK_RingList to fetch the head
 * item from
 *
 * Return Value:
 * void * - Pointer to item stored at head node (may be NULL)
 *
 * Notes:
 *
 * Revision History:
 *
 * Monday 25 November 1996 (Eric Winter) - Started initial version.
 *
 * */

void *
IK_RingList_GetHeadItem(IK_RingList *pringlist)
{

    /* Validate input. */
    assert(pringlist != NULL);

    /*-----------------------------------------------------------------------*/

    /* Return the data item pointer from the head node. */
    assert(pringlist->plistnode_head != NULL);
    return(pringlist->plistnode_head->pv_item);

}

/*****************************************************************************/

/*
 * Name: IK_RingList_GetNextItem()
 *
 * Description: This function returns a pointer to the item stored in
 * the list node which comes after the list node containing the
 * specified item.
 *
 * Parameters:
 * IK_RingList *pringlist - Pointer to IK_RingList to fetch the next
 * item from
 * void *pv_item - Pointer to item to use as the basis for the search
 *
 * Return Value:
 * void * - Pointer to item stored at next node (may be NULL)
 *
 * Notes:
 *
 * Revision History:
 *
 * Monday 25 November 1996 (Eric Winter) - Started initial version.
 *
 * */

void *
IK_RingList_GetNextItem(IK_RingList *pringlist, void *pv_item)
{

    /* This points to the node containing the current item. */
    IK_ListNode *plistnode;

    /*-----------------------------------------------------------------------*/

    /* Validate input. */
    assert(pringlist != NULL);

    /*-----------------------------------------------------------------------*/

    /* Find the node containing the current item. */
    plistnode = FindNodeByItem(pringlist, pv_item);
    if (plistnode == NULL) {
	return(NULL);
    }

    /* Return the pointer to the item from the next node. */
    assert(plistnode->plistnode_next != NULL);
    return(plistnode->plistnode_next->pv_item);

}

/*****************************************************************************/

/*
 * Name: IK_RingList_GetPrevItem()
 *
 * Description: This function returns a pointer to the item stored in
 * the list node which comes before the list node containing the
 * specified item.
 *
 * Parameters:
 * IK_RingList *pringlist - Pointer to IK_RingList to fetch the
 * previous item from
 * void *pv_item - Pointer to item to use as the basis for the search
 *
 * Return Value:
 * void * - Pointer to item stored at previous node (may be NULL)
 *
 * Notes:
 *
 * Revision History:
 *
 * Monday 25 November 1996 (Eric Winter) - Started initial version.
 *
 * */

void *
IK_RingList_GetPrevItem(IK_RingList *pringlist, void *pv_item)
{

    /* This points to the node containing the current item. */
    IK_ListNode *plistnode;

    /*-----------------------------------------------------------------------*/

    /* Validate input. */
    assert(pringlist != NULL);

    /*-----------------------------------------------------------------------*/

    /* Find the node containing the current item. */
    plistnode = FindNodeByItem(pringlist, pv_item);
    if (plistnode == NULL) {
	return(NULL);
    }

    /* Return the pointer to the item from the previous node. */
    assert(plistnode->plistnode_prev != NULL);
    return(plistnode->plistnode_prev->pv_item);

}

/*****************************************************************************/

/*
 * Name: IK_RingList_Dump()
 *
 * Description: This debugging function dumps the contents of the
 * specified IK_RingList to the specified stream.
 *
 * Parameters:
 * FILE *pFILE - Pointer to FILE for output
 * IK_RingList *pringlist - Pointer to IK_RingList to dump
 *
 * Return Value:
 * None
 *
 * Notes:
 * Need to make this function call the Dump functions for the list
 * nodes the list contains.
 *
 * Revision History:
 *
 * Tuesday 19 November 1996 (Eric Winter) - Started initial version.
 * */

void
IK_RingList_Dump(FILE *pFILE, IK_RingList *pringlist)
{

    /* Validate input. */
    assert(pFILE != NULL);

    /*-----------------------------------------------------------------------*/

    /* Print the ring list contents to the stream. */
    (void) fprintf(pFILE, "Dumping IK_RingList at address %p:\n", pringlist);
    if (pringlist != NULL) {
	(void) fprintf(pFILE, "plistnode_head = %p\n",
		       pringlist->plistnode_head);
    } else {
	(void) fprintf(pFILE, "(NULL)\n");
    }

}

/*****************************************************************************/

/* FUNCTIONS BELOW THIS POINT ARE PRIVATE TO THIS SOURCE FILE ONLY! */

/*****************************************************************************/

/*
 * Name: Allocate()
 *
 * Description: This function allocates space for a new IK_RingList
 * structure.
 *
 * Parameters:
 * None
 *
 * Return Value:
 * IK_RingList *pringlist_new - Pointer to new IK_RingList structure
 * NULL - If an error occurs
 *
 * Notes:
 * (1) This function is used to provide flexibility in the allocation
 * mechanism. For example, a group of lists can be pre-allocated and
 * doled out from this function. For now, it simply acts as a
 * front-end to malloc().
 *
 * Revision History:
 *
 * Tuesday 19 November 1996 (Eric Winter) - Started initial version.
 *
 * Saturday 23 November 1996 (Eric Winter) - Changed name to Allocate.
 * */

static IK_RingList *
Allocate(void)
{

    /* Allocate the new ring list and return it. */
    return(malloc(sizeof(IK_RingList)));

}

/*****************************************************************************/

/*
 * Name: Free()
 *
 * Description: This function frees the memory allocated to a
 * IK_RingList structure.
 *
 * Parameters:
 * IK_RingList *pringlist - Pointer to IK_RingList to free
 *
 * Return Value:
 * None
 *
 * Notes:
 * (1) See note (1) for Allocate().
 *
 * Revision History:
 *
 * Tuesday 19 November 1996 (Eric Winter) - Started initial version.
 *
 * Saturday 23 November 1996 (Eric Winter) - Changed name to Free().
 * */

static void
Free(IK_RingList *pringlist)
{

    /* Validate input. */
    assert(pringlist != NULL);

    /*-----------------------------------------------------------------------*/

    /* Delete the node. */
    free(pringlist);

}

/*****************************************************************************/

/*
 * Name: AddNode()
 *
 * Description: This function adds a node at the "end" of a ring
 * list. In code terms, the new node is inserted before the "head"
 * node, since this is a ring list.
 *
 * Parameters:
 * IK_RingList *pringlist - Pointer to ring list to add node to
 * IK_ListNode *plistnode_new - Node to add to the ring list
 *
 * Return Value:
 * None
 *
 * Notes:
 *
 * Revision History:
 *
 * Tuesday 19 November 1996 (Eric Winter) - Started initial version.
 *
 * Saturday 23 November 1996 (Eric Winter) - Renamed to AddNode().
 * */

static void
AddNode(IK_RingList *pringlist, IK_ListNode *plistnode_new)
{

    /* plistnode_oldPrev is a temporary pointer to the previous list
       node in the ring. */
    IK_ListNode *plistnode_oldPrev;

    /*-----------------------------------------------------------------------*/

    /* Validate input. */
    assert(pringlist != NULL);
    assert(plistnode_new != NULL);

    /*-----------------------------------------------------------------------*/

    /* If the list is currently empty, add the node at the head and
       attach it to itself. If the list is not empty, insert the node
       prior to the "head" node. */
    if (pringlist->plistnode_head == NULL) {

	/* The list is empty. Add the new node at the head and attach
           it to itself. */
	pringlist->plistnode_head = plistnode_new;
	plistnode_new->plistnode_next = plistnode_new;
	plistnode_new->plistnode_prev = plistnode_new;

    } else {

	/* The list is not empty. */

	/* Save a pointer to the current previous node from the head
           node. */
	plistnode_oldPrev = pringlist->plistnode_head->plistnode_prev;
	assert(plistnode_oldPrev != NULL);

	/* Insert the new node before the head node, and after the old
           previous node. */
	pringlist->plistnode_head->plistnode_prev = plistnode_new;
	plistnode_new->plistnode_next = pringlist->plistnode_head;
	plistnode_oldPrev->plistnode_next = plistnode_new;
	plistnode_new->plistnode_prev = plistnode_oldPrev;

    }

}

/*****************************************************************************/

/*
 * Name: RemoveNode()
 *
 * Description: This function removes a node from a ring list.
 *
 * Parameters:
 * IK_RingList *pringlist - Pointer to ring list remove node from
 * IK_ListNode *plistnode_remove - Node to remove from the ring list
 *
 * Return Value:
 * None
 *
 * Notes:
 * The results of this function are undefined (and probably bad) if
 * this function is passed a IK_ListItem that is not a member of the
 * specified IK_RingList.
 *
 * Revision History:
 *
 * Tuesday 19 November 1996 (Eric Winter) - Started initial version.
 *
 * Saturday 23 November 1996 (Eric Winter) - Renamed to RemoveNode().
 * */

static void
RemoveNode(IK_RingList *pringlist, IK_ListNode *plistnode_remove)
{

    /* Validate input. */
    assert(pringlist != NULL);
    assert(pringlist->plistnode_head != NULL);
    assert(plistnode_remove != NULL);
    assert(plistnode_remove->plistnode_next != NULL);
    assert(plistnode_remove->plistnode_prev != NULL);

    /*-----------------------------------------------------------------------*/

    /* Remove the node. If the last node is removed, set the head node
       pointer to NULL. */
    if (plistnode_remove->plistnode_next == plistnode_remove) {

	/* Remove the last (head) node. */
	assert(pringlist->plistnode_head == plistnode_remove);
	pringlist->plistnode_head = NULL;

    } else {

	/* If removing the head node and nodes remain, move the next
           node to the head position. */
	if (plistnode_remove == pringlist->plistnode_head) {
	    pringlist->plistnode_head = plistnode_remove->plistnode_next;
	}

	/* Sever the connections from the list to the node. */
	plistnode_remove->plistnode_prev->plistnode_next =
	    plistnode_remove->plistnode_next;
	plistnode_remove->plistnode_next->plistnode_prev =
	    plistnode_remove->plistnode_prev;

    }

    /* Sever the connections from the node into the list. */
    plistnode_remove->plistnode_next = NULL;
    plistnode_remove->plistnode_prev = NULL;

}

/*****************************************************************************/

/*
 * Name: FindNodeByItem()
 *
 * Description: This function finds the list node containing the
 * specified item and returns a pointer to the node.
 *
 * Parameters:
 * IK_RingList *pringlist - Pointer to IK_RingList to count nodes in
 * void *pv_item - Pointer to item to find
 *
 * Return Value:
 * IK_ListNode *plistnode - Pointer to list node for item to find
 * NULL - If the list does not contain the item
 *
 * Notes:
 *
 * Revision History:
 *
 * Saturday 23 November 1996 (Eric Winter) - Started initial version.
 * */

static IK_ListNode *
FindNodeByItem(IK_RingList *pringlist, void *pv_item)
{

    /* plistnode is a temporary pointer to the current list node. */
    IK_ListNode *plistnode;

    /*-----------------------------------------------------------------------*/

    /* Validate input. */
    assert(pringlist != NULL);

    /*-----------------------------------------------------------------------*/

    /* If the list is empty, it cannot contain the item. */
    if (pringlist->plistnode_head == NULL) {
	return(NULL);
    }

    /* The list contains at least one item, so search through all
       items until the desired item is found. */
    plistnode = pringlist->plistnode_head;
    do {

	/* Check to see if the current node holds the desired item. If
           so, break out of the loop. If not, get the next list
           item. */
	assert(plistnode != NULL);
	if (plistnode->pv_item == pv_item) {
	    return(plistnode);
	}
	plistnode = plistnode->plistnode_next;

    } while (plistnode != pringlist->plistnode_head);

    /* If the code reaches this point, the item was not found. */
    return(NULL);

}
