/*
 * Name: IK_Socket.c
 *
 * Description: This file contains the low level functions used to
 * open socket connections, and write data to and read data from
 * sockets.
 *
 * Notes:
 *
 *-------------------------------------------------------------------------
 *
 * RCS information:
 *
 * $Id: IK_Socket.c,v 1.2 1997/10/20 17:26:29 solomon Exp $
 *
 * $Log: IK_Socket.c,v $
 * Revision 1.2  1997/10/20 17:26:29  solomon
 * Updated to GUI IK layer 7_0_0_13.
 *
 * Revision 7.0.1.6  1997/10/07  17:55:48  jarva
 * Modified IK_IsSocketConnected for operation under linux. See SMR 4207.
 *
 * Revision 7.0.1.5  1997/07/22  12:14:26  winter
 * SMR 4131: Changed 4th argument to setsockopt() to be a (const char *) to
 * eliminate compiler warnings on Solaris. The prototype for this function
 * shoud really use a (void *) here, so that any pointer type may be used,
 * without a cast required. Also added NULL as second argument to gettimeofday()
 * for the timezone. On SGI, this argument is optional. On Solaris 2.5, it is
 * required.
 *
 * Revision 7.0.1.4  1997/07/11  18:42:07  winter
 * SMR 4063: Added a wrapper around the select() call in
 * IK_WaitForSocketToBeReadable() to trap for the EINTR error, and force
 * a re-issue of the select() call.
 *
 * Revision 7.0.1.3  1997/05/22  14:55:30  winter
 * SMR 4063: Added a lot of error checking and logging.
 *
 * Revision 7.0.1.2  1997/05/22  10:45:33  winter
 * SMR 4024,4057: Added new function IK_WaitForSocketToBeReadable().
 *
 * Revision 7.0.1.1  1997/05/21  18:50:27  winter
 * Placeholder for new branch during SMR 4057 work.
 *
 * Revision 7.0  1997/04/11  17:54:45  ims
 * 6.1.1.10 Promoted to 7.0 J. Katen
 *
 * Revision 6.1.1.10  1997/03/17  16:29:09  winter
 * SMR 3943 - IK_IsSocketConnected() now clears errno and IK_ImsErrno after
 * the connect() call, since errno was being set as a (beneficial) side effect
 * of the call.
 *
 * Revision 6.1.1.9  1997/03/14  19:48:22  winter
 * SMR 3858 (yet again) - In order to maintain source-code compatibility,
 * several functions were renamed to their previous names:
 * IK_CloseSocket()->IK_Close()
 * IK_ShutdownSocket()->IK_Shutdown()
 * IK_HandleSocketSIGALRM()->IK_AlarmHandler()
 *
 * Revision 6.1.1.8  1997/03/14  19:25:15  winter
 * SMR 3858 (again) - Added IK_IsSocketReadable().
 *
 * Revision 6.1.1.7  1997/03/14  15:00:43  winter
 * SMR 3858 - Moved IK_Signal_Brk declaration to IK_Network.h.
 *
 * Revision 6.1.1.6  1997/03/11  20:06:02  winter
 * SMR 3879 - Added a single  line to IK_ShutdownSocket() to ensure that is
 * shutdown() fails the socket is still closed, so that the socket does not
 * take up a slot in the descriptor table.
 *
 * Revision 6.1.1.5  1997/02/27  18:35:08  winter
 * Added code for new function, IK_IsSocketConnected().
 *
 * Revision 6.1.1.4  1997/01/08  16:34:51  ims
 * Added stdio.h to include directives for SunOS 4.1.3 compile. J. Katen
 *
 * Revision 6.1.1.3  1996/10/31  20:08:00  winter
 * Fixed bug in IK_CreateSelectableSocket() that could leave unselectable
 * sockets open.
 *
 * Revision 6.1.1.2  1996/10/29  17:24:49  winter
 * Moved several functions around for readability. Added IK_ShutdownSocket(),
 * socket option and signal handling functions.
 *
 * Revision 6.1.1.1  1996/10/28  18:11:22  winter
 * First real version.
 *
 * */

/*****************************************************************************/

/* Define the RCS identifier string. */
static const char *rcsid = "$Id: IK_Socket.c,v 1.2 1997/10/20 17:26:29 solomon Exp $";

/*****************************************************************************/

/* #include directives */

/* This header must be #included first in order to use
   autoconf-generated information. */
#include "IK_Ims.h"

/* System headers */
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>

/* Network headers */
#include <netinet/in.h>

/* Standard headers */
#include <assert.h>
#ifdef HAVE_BSTRING_H
#include <bstring.h>
#endif /* HAVE_BSTRING_H */
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

/* Third-party library headers */

/* IMS headers */
#include "IK_Errno.h"
#include "IK_Misc.h"
#include "IK_Network.h"
#include "IK_Socket.h"
#include "IK_Syslog.h"

/*****************************************************************************/

/* Local #define directives */

/* This constant is used to indicate that the default protocol should
   be used when creating a socket. */
#define USE_DEFAULT_SOCKET_PROTOCOL (0)

/* Define the number of bytes in a binary IP address. */
#define IP_ADDRESS_BYTES (4)

/* These constants are used to set up the timeout for the select()
   call used to check a single socket for readability. */
#define SELECT_TIMEOUT_SEC (1)
#define SELECT_TIMEOUT_USEC (0)

/*****************************************************************************/

/* Local variable definitions */

/*****************************************************************************/

/* Local function prototypes */

/*****************************************************************************/

/*
 * Name: IK_CreateSelectableSocket()
 *
 * Description: This function creates a new socket using the socket()
 * system call, then examines the socket descriptor to ensure that it
 * satisfies the restriction of selectability. If so, return the
 * socket descriptor in *pi_socketDesc, and IK_TRUE as the function
 * value. If not, return IK_FALSE.
 *
 * Parameters:
 * None
 *
 * Return Values:
 * IK_BOOLEAN IK_TRUE - If the socket is created
 * IK_BOOLEAN IK_FALSE - If not
 *
 * Warnings:
 *
 * Global Variables Used:
 *
 * Revision History:
 *
 * Friday 25 October 1996 (Eric Winter) - Initial version.
 *
 * Friday 14 March 1997 (Eric Winter) - Changed to use IK_Close()
 * instead of IK_CloseSocket(); this was done to maintain
 * compatibility with legacy DAAC code.
 *
 * Thursday 22 May 1997 (Eric Winter) - Added error logging code.
 * */

IK_BOOLEAN
IK_CreateSelectableSocket(int *pi_socketDesc)
{

    /* This is the descriptor returned from socket(). */
    int i_socketDesc;

    /*-----------------------------------------------------------------------*/

    /* Validate input. */
    assert(pi_socketDesc != NULL);

    /*-----------------------------------------------------------------------*/

    /* Create the socket. */
    errno = 0;
    i_socketDesc = socket(PF_INET, SOCK_STREAM, USE_DEFAULT_SOCKET_PROTOCOL);
    if (i_socketDesc == -1) {
	IK_vSyslog(LOG_ERR, "IK_CreateSelectableSocket(): socket() failed;; "
		   "errno = %d (%s)", errno, strerror(errno));
	(void) IK_Close(i_socketDesc);
	return(IK_FALSE);
    }

    /* Verify the socket descriptor can be used in the select()
       call. If not, close the socket and return an error code. */
    if (i_socketDesc >= FD_SETSIZE) {
	IK_vSyslog(LOG_NOTICE, "IK_CreateSelectableSocket(): Socket number %d "
		   "too high to be used in select()", i_socketDesc);
	(void) IK_Close(i_socketDesc);
	return(IK_FALSE);
    }

    /* Save the socket descriptor. */
    *pi_socketDesc = i_socketDesc;

    /* Return the execution status. */
    return(IK_TRUE);

}

/*****************************************************************************/

/*
 * Name: IK_BeginBackgroundConnection()
 *
 * Description: This function initiates a non-blocking connect() call
 * for the specified socket.
 *
 * Parameters:
 * int i_socketDesc - Socket descriptor to connect
 * const char *pch_destinationAddress - IP address or FQDN of host to
 * connect to
 * int i_portNumber - TCP port number to connect to on destination host
 *
 * Return Values:
 * IK_BOOLEAN IK_TRUE - If the connection was successfully started
 * IK_BOOLEAN IK_FALSE - If not
 *
 * Warnings:
 *
 * Global Variables Used:
 *
 * Revision History:
 *
 * Friday 25 October 1996 (Eric Winter) - Initial version.
 *
 * Thursday 22 May 1997 (Eric Winter) - Added a lot of debugging and
 * logging code.
 *
 * Tuesday 7 Oct. 1997 (Ken Jarva) - added action in the case where
 * connect() returns 0 to mollify lint. Else clause now issues
 * return( IK_TRUE ).
 * */

IK_BOOLEAN
IK_BeginBackgroundConnection(int i_socketDesc,
			     const char *pch_destinationAddress,
			     int i_portNumber)
{

    /* phostent_server points to information about the host
       address. */
    struct hostent *phostent_server;

    /* sockaddr_server contains the address of the server. */
    struct sockaddr_in sockaddr_server;

    /* i_connectStatus holds the return value from connect(). */
    int i_connectStatus;

    /*-----------------------------------------------------------------------*/

    /* Validate input. */
    assert(i_socketDesc >= 0);
    assert(pch_destinationAddress != NULL);
    assert(i_portNumber >= 0);

    /*-----------------------------------------------------------------------*/

    /* Fill in the structure containing information on the destination
       host. */

    /* Clear the server address structure. */
    (void) memset(&sockaddr_server, 0, sizeof(sockaddr_server));

    /* The host was specified by a FQDN or dotted-decimal string. Map
       it to a host data structure. */
    errno = 0;
    phostent_server = gethostbyname(pch_destinationAddress);
    if (phostent_server == NULL) {
	IK_vSyslog(LOG_ERR, "IK_BeginBackgroundConnection(): gethostbyname() "
		   "failed; errno = %d (%s)", errno, strerror(errno));
	return(IK_FALSE);
    }

    /* Set the socket address type to the server address type (always
       AF_INET). */
    assert(phostent_server->h_addrtype == AF_INET);
    sockaddr_server.sin_family = phostent_server->h_addrtype;

    /* Copy the IP address (a 32-bit number in network byte order) of
       the server to the IP address of the socket to connect to. Note
       that the address (actually, the first element of the array)
       pointed to by phostent_server->h_addr (an alias for
       h_addr_list[0]) is not a char *; it is actually a in_addr
       *. Therefore, the h_length member is always 4 (for now). */
    assert(phostent_server->h_addr != NULL);
    assert(phostent_server->h_length == IP_ADDRESS_BYTES);
    (void) memcpy(&sockaddr_server.sin_addr, phostent_server->h_addr,
		  phostent_server->h_length);

    /* Set the port number to connect to at the server end, in network
       byte order. */
    sockaddr_server.sin_port = htons((short) i_portNumber);

    /*-----------------------------------------------------------------------*/

    /* Issue the non-blocking connect() call. */
    errno = 0;
    i_connectStatus = connect(i_socketDesc,
			      (const struct sockaddr *) &sockaddr_server,
			      sizeof(sockaddr_server));
    assert(i_connectStatus == -1 || i_connectStatus == 0);

    /* Check the results. */
    if (i_connectStatus == -1) {

	/* The connection did not complete (the usual case, since the
           socket should be non-blocking), or an error occurred. Check
           the global error code to verify that the connection process
           has begun. If not, return an error code. */
	if (errno != EINPROGRESS) {
	    IK_vSyslog(LOG_ERR, "IK_BeginBackgroundConnection(): connect() "
		       "failed; errno = %d (%s)", errno, strerror(errno));
	    return(IK_FALSE);
	}

    } else {

	/* The connection completed! This should happen rarely, if
           ever, for the non-blocking connect() call. */
           return(IK_TRUE);

    }

    /*-----------------------------------------------------------------------*/

    /* Return status. */
    return(IK_TRUE);

}

/*****************************************************************************/

/*
 * Name: IK_IsSocketConnected()
 *
 * Description: This function checks to see if the specified socket is
 * connected to the server defined by the destination address and port
 * number.
 *
 * Parameters:
 * int i_socketDesc - Socket descriptor to check for connection
 * const char *pch_destinationAddress - IP address or FQDN of host to
 * connect to
 * int i_portNumber - TCP port number to connect to on destination host
 *
 * Return Values:
 * IK_BOOLEAN IK_TRUE - If the socket is connected to the server
 * IK_BOOLEAN IK_FALSE - If not
 *
 * Warnings:
 *
 *  Different operating systems may respond differently to connect
 *  calls issued multiple times against a socket open with a
 *  non-blocking connect.  In some cases (IRIX & solaris) the connect()
 *  call will never return anything but -1 and the system variable
 *  errno must be checked to determine whether the connect has all
 *  ready succeeded (errno = EISCONN) or not. In others (e.g. linux)
 *  connect() will return 0 the first time it's called after successfully
 *  making the non-blocking connection.
 *
 * Global Variables Used:
 *
 * Revision History:
 *
 * Thursday 27 February 1997 (Eric Winter) - Initial version
 *
 * Monday 17 March 1997 (Eric Winter) - Added code to clear errno and
 * IK_ImsErrno after connect() call.
 *
 * Thursday 22 May 1997 (Eric Winter) - Added debugging and error
 * logging code.
 *
 * Tuesday 7 October 1997 (Ken Jarva) - Changed reaction of this
 * function to the event that connect() returns 0, which it will do
 * under some conditions under linux.  Now IK_IsSocketConnected will
 * return IK_TRUE if connect() returns 0. (See SMR 4207)
 *
 * */

IK_BOOLEAN
IK_IsSocketConnected(int i_socketDesc, const char *pch_destinationAddress,
		     int i_portNumber)
{

    /* phostent_server points to information about the host
       address. */
    struct hostent *phostent_server;

    /* sockaddr_server contains the address of the server. */
    struct sockaddr_in sockaddr_server;

    /* i_connectStatus holds the return value from connect(). */
    int i_connectStatus;

    /* This flag holds the execution status of this function. */
    IK_BOOLEAN bool_isConnected;

    /*-----------------------------------------------------------------------*/

    /* Validate input. */
    assert(i_socketDesc >= 0);
    assert(pch_destinationAddress != NULL);
    assert(i_portNumber >= 0);

    /*-----------------------------------------------------------------------*/

    /* Fill in the structure containing information on the destination
       host. */

    /* Clear the server address structure. */
    (void) memset(&sockaddr_server, 0, sizeof(sockaddr_server));

    /* The host was specified by a FQDN or dotted-decimal string. Map
       it to a host data structure. */
    errno = 0;
    phostent_server = gethostbyname(pch_destinationAddress);
    if (phostent_server == NULL) {
	IK_vSyslog(LOG_ERR, "IK_IsSocketConnected(): gethostbyname() failed; "
		   "errno = %d (%s)", errno, strerror(errno));
	return(IK_FALSE);
    }

    /* Set the socket address type to the server address type (always
       AF_INET). */
    assert(phostent_server->h_addrtype == AF_INET);
    sockaddr_server.sin_family = phostent_server->h_addrtype;

    /* Copy the IP address (a 32-bit number in network byte order) of
       the server to the IP address of the socket to connect to. Note
       that the address (actually, the first element of the array)
       pointed to by phostent_server->h_addr (an alias for
       h_addr_list[0]) is not a char *; it is actually a in_addr
       *. Therefore, the h_length member is always 4 (for now). */
    assert(phostent_server->h_addr != NULL);
    assert(phostent_server->h_length == IP_ADDRESS_BYTES);
    (void) memcpy(&sockaddr_server.sin_addr, phostent_server->h_addr,
		  phostent_server->h_length);

    /* Set the port number to connect to at the server end, in network
       byte order. */
    sockaddr_server.sin_port = htons((short) i_portNumber);

    /*-----------------------------------------------------------------------*/

    /* Issue the connect() call to see if the connection is
       complete. */
    errno = 0;
    i_connectStatus = connect(i_socketDesc,
			      (const struct sockaddr *) &sockaddr_server,
			      sizeof(sockaddr_server));
    assert(i_connectStatus == -1 || i_connectStatus == 0);

    /* Check the results. */
    if (i_connectStatus == -1) {

	/* The connect() call failed. Now check to see if errno is
           EISCONN. If so, the socket is connected. Clear error codes
           and return. */
	if (errno == EISCONN) {
	    bool_isConnected = IK_TRUE;
	} else {
	    bool_isConnected = IK_FALSE;
	}

    } else {
	bool_isConnected = IK_TRUE; /* changed for linux */
    }

    /* Clear the error codes and return the connection status. The
       error codes are cleared because errno is set by connect() to
       EISCONN as a side-effect, even though this is not really an
       error condition. */
    errno = 0;
    IK_ImsErrno = IK_ENOERROR;
    return(bool_isConnected);

}

/*****************************************************************************/

/*
 * Name: IK_Shutdown()
 *
 * Description: This function shuts down a socket connection.
 *
 * Parameters:
 * int i_socketDesc - Socket descriptor to shut dow
 *
 * Return Values:
 * IK_BOOLEAN IK_TRUE - If the connection was successfully shut down
 * IK_BOOLEAN IK_FALSE - If not
 *
 * Warnings:
 *
 * Global Variables Used:
 *
 * Revision History:
 *
 * Friday 25 October 1996 (Eric Winter) - Initial version.
 *
 * Tuesday 11 March 1997 (Eric Winter) - Added code to close the
 * socket even when shutdown() fails, to ensure that the dead socket
 * does not take up a slot in the descriptor table.
 *
 * Friday 14 March 1997 (Eric Winter) - Changed name back to
 * IK_Shutdown() to maintain compatibility with legacy DAAC code.
 *
 * Thursday 22 May 1997 (Eric Winter) - Added debugging and error
 * logging code.
 * */

IK_BOOLEAN
IK_Shutdown(int i_socketDesc)
{

    /* Validate input. */
    assert(i_socketDesc >= 0);

    /*-----------------------------------------------------------------------*/

    /* Shut down the socket in both directions. */
    errno = 0;
    if (shutdown(i_socketDesc, SOCKET_SHUTDOWN_READ_WRITE) == -1) {
	IK_vSyslog(LOG_ERR, "IK_Shutdown(): shutdown() failed; "
		   "errno = %d (%s)", errno, strerror(errno));
	(void) IK_Close(i_socketDesc);
	return(IK_FALSE);
    }

    /* Now close the socket. */
    if (IK_Close(i_socketDesc) == IK_FALSE) {
	IK_Syslog(LOG_ERR, "IK_Shutdown(): IK_Close() failed");
	return(IK_FALSE);
    }

    /* Return normally. */
    return(IK_TRUE);

}

/*****************************************************************************/

/*
 * Name: IK_Close()
 *
 * Description: This function closes a socket connection.
 *
 * Parameters:
 * int i_socketDesc - Socket descriptor to close
 *
 * Return Values:
 * IK_BOOLEAN IK_TRUE - If the connection was successfully closed
 * IK_BOOLEAN IK_FALSE - If not
 *
 * Warnings:
 *
 * Global Variables Used:
 *
 * Revision History:
 *
 * Friday 25 October 1996 (Eric Winter) - Initial version.
 *
 * Friday 14 March 1997 (Eric Winter) - Changed name back to
 * IK_Close() to maintain compatibilty with legacy DAAC code.
 *
 * Thursday 22 May 1997 (Eric Winter) - Added debugging and error
 * logging code.
 * */

IK_BOOLEAN
IK_Close(int i_socketDesc)
{

    /*-----------------------------------------------------------------------*/

    /* Validate input. */
    assert(i_socketDesc >= 0);

    /*-----------------------------------------------------------------------*/

    /* Close the socket. */
    errno = 0;
    if (close(i_socketDesc) == -1) {
	IK_vSyslog(LOG_ERR, "IK_Close(): close() failed; errno = %d (%s)",
		   errno, strerror(errno));
	return(IK_FALSE);
    }

    /* Return normally. */
    return(IK_TRUE);

}

/*****************************************************************************/

/*
 * Name: IK_MakeSocketNonblocking()
 *
 * Description: This function sets a socket to use non-blocking
 * I/O. This function is typically used to set a socket up to use a
 * non-blocking connect() call. The socket is then reverted to
 * blocking I/O once the connection is established.
 *
 * Parameters:
 * None
 *
 * Return Values:
 * IK_BOOLEAN IK_TRUE - If the socket was made non-blocking
 * IK_BOOLEAN IK_FALSE - If not
 *
 * Warnings:
 *
 * Global Variables Used:
 *
 * Revision History:
 *
 * Friday 25 October 1996 (Eric Winter) - Initial version.
 *
 * Thursday 22 May 1997 (Eric Winter) - Added debugging and error
 * logging code.
 * */

IK_BOOLEAN
IK_MakeSocketNonblocking(int i_socketDesc)
{

    /* i_socketFlags holds the flags for the current socket, as
       returned from fcntl(). */
    int i_socketFlags;

    /*-----------------------------------------------------------------------*/

    /* Validate input. */
    assert(i_socketDesc >= 0);

    /*-----------------------------------------------------------------------*/

    /* Fetch the current socket flags. Note that this call cannot
       fail. */
    errno = 0;
    i_socketFlags = fcntl(i_socketDesc, F_GETFL, 0);
    if (i_socketFlags == -1) {
	IK_vSyslog(LOG_ERR, "IK_MakeSocketNonblocking(): fcntl() failed for "
		   "fetch; errno = %d (%s)", errno, strerror(errno));
	return(IK_FALSE);
    }

    /* Set the POSIX.1 non-blocking I/O bit. */
    i_socketFlags |= O_NONBLOCK;

    /* Make the socket non-blocking. */
    errno = 0;
    if (fcntl(i_socketDesc, F_SETFL, i_socketFlags) == -1) {
	IK_vSyslog(LOG_ERR, "IK_MakeSocketNonblocking(): fcntl() failed for "
		   "set; errno = %d (%s)", errno, strerror(errno));
	return(IK_FALSE);
    } else {
	return(IK_TRUE);
    }

}

/*****************************************************************************/

/*
 * Name: IK_MakeSocketBlocking()
 *
 * Description: This function sets a socket to use blocking I/O. This
 * function is typically used to restore a socket to blocking status
 * after a successful non-blocking connect() call.
 *
 * Parameters:
 * None
 *
 * Return Values:
 * IK_BOOLEAN IK_TRUE - If the socket was made blocking
 * IK_BOOLEAN IK_FALSE - If not
 *
 * Warnings:
 *
 * Global Variables Used:
 *
 * Revision History:
 *
 * Friday 25 October 1996 (Eric Winter) - Initial version.
 *
 * Thursday 22 May 1997 (Eric Winter) - Added debugging and error
 * logging code.
 * */

IK_BOOLEAN
IK_MakeSocketBlocking(int i_socketDesc)
{

    /* i_socketFlags holds the flags for the current socket, as
       returned from fcntl(). */
    int i_socketFlags;

    /*-----------------------------------------------------------------------*/

    /* Validate input. */
    assert(i_socketDesc >= 0);

    /*-----------------------------------------------------------------------*/

    /* Fetch the current socket flags. Note that this call cannot
       fail. */
    errno = 0;
    i_socketFlags = fcntl(i_socketDesc, F_GETFL, 0);
    if (i_socketFlags == -1) {
	IK_vSyslog(LOG_ERR, "IK_MakeSocketBlocking(): fcntl() failed for "
		   "fetch; errno = %d (%s)", errno, strerror(errno));
	return(IK_FALSE);
    }

    /* Clear the POSIX.1 non-blocking I/O bit. */
    i_socketFlags &= ~O_NONBLOCK;

    /* Make the socket blocking. */
    errno = 0;
    if (fcntl(i_socketDesc, F_SETFL, i_socketFlags) == -1) {
	IK_vSyslog(LOG_ERR, "IK_MakeSocketBlocking(): fcntl() failed for "
		   "set; errrno = %d (%s)", errno, strerror(errno));
	return(IK_FALSE);
    } else {
	return(IK_TRUE);
    }

}

/*****************************************************************************/

/*
 * Name: IK_ActivateSocketLingering()
 *
 * Description: This function turns on the lingering option on a
 * socket.
 *
 * Parameters:
 * int i_socketDesc - Socket descriptor to activate lingering for
 *
 * Return Values:
 * IK_BOOLEAN IK_TRUE - If the lingering option was successfully activated
 * IK_BOOLEAN IK_FALSE - If not
 *
 * Warnings:
 *
 * Global Variables Used:
 *
 * Revision History:
 *
 * Tuesday 29 October 1996 (Eric Winter) - Initial version.
 *
 * Thursday 22 May 1997 (Eric Winter) - Added debugging and error
 * logging code.
 * */

IK_BOOLEAN
IK_ActivateSocketLingering(int i_socketDesc)
{

    /* This structure is used to specify how long to continue to try
       to transmit unsent data on the socket after shutdown. */
    struct linger linger;

    /*-----------------------------------------------------------------------*/

    /* Validate input. */
    assert(i_socketDesc >= 0);

    /*-----------------------------------------------------------------------*/

    /* Set the lingering options - turn on lingering for 10
       seconds. Note that in current implementations, the l_linger
       value is ignored. The effect is to try to send any unsent data
       when the socket is closed if l_linger is non-zero. */
    linger.l_onoff = TURN_LINGERING_ON;
    linger.l_linger = LINGER_TIME;

    /* Turn on socket lingering. */
    errno = 0;
    if (setsockopt(i_socketDesc, SOL_SOCKET, SO_LINGER,
		   (const char *) &linger, sizeof(linger)) == -1) {
	IK_vSyslog(LOG_ERR, "IK_ActivateSocketLingering(): setsockopt() "
		   "failed; errno = %d (%s)", errno, strerror(errno));
	return(IK_FALSE);
    }

    /*-----------------------------------------------------------------------*/

    /* Return normally. */
    return(IK_TRUE);

}

/*****************************************************************************/

/*
 * Name: IK_DeactivateSocketLingering()
 *
 * Description: This function turns off the lingering option on a
 * socket.
 *
 * Parameters:
 * int i_socketDesc - Socket descriptor to deactivate lingering for
 *
 * Return Values:
 * IK_BOOLEAN IK_TRUE - If the lingering option was successfully deactivated
 * IK_BOOLEAN IK_FALSE - If not
 *
 * Warnings:
 *
 * Global Variables Used:
 *
 * Revision History:
 *
 * Tuesday 29 October 1996 (Eric Winter) - Initial version.
 *
 * Thursday 22 May 1997 (Eric Winter) - Added debugging and error
 * logging code.
 * */

IK_BOOLEAN
IK_DeactivateSocketLingering(int i_socketDesc)
{

    /* This structure is used to specify how long to continue to try
       to transmit unsent data on the socket after shutdown. */
    struct linger linger;

    /*-----------------------------------------------------------------------*/

    /* Validate input. */
    assert(i_socketDesc >= 0);

    /*-----------------------------------------------------------------------*/

    /* Set the lingering options - turn off lingering. */
    linger.l_onoff = TURN_LINGERING_OFF;

    /* Turn off socket lingering. */
    errno = 0;
    if (setsockopt(i_socketDesc, SOL_SOCKET, SO_LINGER,
		   (const char *) &linger, sizeof(linger)) == -1) {
	IK_vSyslog(LOG_ERR, "IK_DeactivateSocketLingering(): setsockopt() "
		   "failed; errno = %d (%s)", errno, strerror(errno));
	return(IK_FALSE);
    }

    /*-----------------------------------------------------------------------*/

    /* Return normally. */
    return(IK_TRUE);

}

/*****************************************************************************/

/*
 * Name: IK_ActivateSocketKeepAlive()
 *
 * Description: This function turns on the "keep-alive" option on a
 * socket.
 *
 * Parameters:
 * int i_socketDesc - Socket descriptor to activate "keep-alive" for
 *
 * Return Values:
 * IK_BOOLEAN IK_TRUE - If the "keep-alive" option was successfully activated
 * IK_BOOLEAN IK_FALSE - If not
 *
 * Warnings:
 *
 * Global Variables Used:
 *
 * Revision History:
 *
 * Tuesday 29 October 1996 (Eric Winter) - Initial version.
 *
 * Thursday 22 May 1997 (Eric Winter) - Added debugging and error
 * logging code.
 * */

IK_BOOLEAN
IK_ActivateSocketKeepAlive(int i_socketDesc)
{

    /* This flag is used to specify whether or not to use the
       "keep-alive" option on the socket. */
    int i_keepAlive;

    /*-----------------------------------------------------------------------*/

    /* Validate input. */
    assert(i_socketDesc >= 0);

    /*-----------------------------------------------------------------------*/

    /* Use the "keep-alive" option. */
    i_keepAlive = TURN_KEEPALIVE_ON;

    /* Turn on socket "keep-alive". */
    errno = 0;
    if (setsockopt(i_socketDesc, SOL_SOCKET, SO_KEEPALIVE,
		   (const char *) &i_keepAlive, sizeof(i_keepAlive)) == -1) {
	IK_vSyslog(LOG_ERR, "IK_ActivateSocketKeepAlive(): setsockopt() "
		   "failed; errno = %d (%s)", errno, strerror(errno));
	return(IK_FALSE);
    }

    /*-----------------------------------------------------------------------*/

    /* Return normally. */
    return(IK_TRUE);

}

/*****************************************************************************/

/*
 * Name: IK_DeactivateSocketKeepAlive()
 *
 * Description: This function turns off the "keep-alive" option on a
 * socket.
 *
 * Parameters:
 * int i_socketDesc - Socket descriptor to deactivate "keep-alive" for
 *
 * Return Values:
 * IK_BOOLEAN IK_TRUE - If the "keep-alive" option was successfully deactivated
 * IK_BOOLEAN IK_FALSE - If not
 *
 * Warnings:
 *
 * Global Variables Used:
 *
 * Revision History:
 *
 * Tuesday 29 October 1996 (Eric Winter) - Initial version.
 *
 * Thursday 22 May 1997 (Eric Winter) - Added debugging and error
 * logging code.
 * */

IK_BOOLEAN
IK_DeactivateSocketKeepAlive(int i_socketDesc)
{

    /* This flag is used to specify whether or not to use the
       "keep-alive" option on the socket. */
    int i_keepAlive;

    /*-----------------------------------------------------------------------*/

    /* Validate input. */
    assert(i_socketDesc >= 0);

    /*-----------------------------------------------------------------------*/

    /* Do not use the "keep-alive" option. */
    i_keepAlive = TURN_KEEPALIVE_OFF;

    /* Turn on socket "keep-alive". */
    errno = 0;
    if (setsockopt(i_socketDesc, SOL_SOCKET, SO_KEEPALIVE,
		   (const char *) &i_keepAlive, sizeof(i_keepAlive)) == -1) {
	IK_vSyslog(LOG_ERR, "IK_DeactivateSocketKeepAlive(): setsockopt() "
		   "failed; errno = %d (%s)", errno, strerror(errno));
	return(IK_FALSE);
    }

    /*-----------------------------------------------------------------------*/

    /* Return normally. */
    return(IK_TRUE);

}

/*****************************************************************************/

/*
 * Name: IK_InstallSocketSIGALRMHandler()
 *
 * Description: This function installs a handler for the SIGALRM signal.
 *
 * Parameters:
 * None
 *
 * Return Values:
 * IK_BOOLEAN IK_TRUE - If the handler was properly installed
 * IK_BOOLEAN IK_FALSE - If not
 *
 * Warnings:
 *
 * Global Variables Used:
 *
 * Revision History:
 *
 * Tuesday 29 October 1996 (Eric Winter) - Initial version.
 *
 * Friday 14 March 1997 (Eric Winter) - Changed to use
 * IK_AlarmHandler() to maintain compatibility with legavy DAAC code.
 *
 * Thursday 22 May 1997 (Eric Winter) - Added debugging and error
 * logging code.
 * */

IK_BOOLEAN
IK_InstallSocketSIGALRMHandler(void)
{

    /* Catch all SIGALRM signals (used for timeouts). */
    if (signal(SIGALRM, IK_AlarmHandler) == SIG_ERR) {
	IK_vSyslog(LOG_ERR, "IK_InstallSocketSIGALRMHandler(): signal() "
		   "failed; errno = %d (%s)", errno, strerror(errno));
	return(IK_FALSE);
    }

    /*-----------------------------------------------------------------------*/

    /* Return normally. */
    return(IK_TRUE);

}

/*****************************************************************************/

/*
 * Name: IK_InstallSocketSIGPIPEHandler()
 *
 * Description: This function installs a handler for the SIGPIPE
 * signal.
 *
 * Parameters:
 * None
 *
 * Return Values:
 * IK_BOOLEAN IK_TRUE - If the handler was properly installed
 * IK_BOOLEAN IK_FALSE - If not
 *
 * Warnings:
 *
 * Global Variables Used:
 *
 * Revision History:
 *
 * Tuesday 29 October 1996 (Eric Winter) - Initial version.
 *
 * Thursday 22 May 1997 (Eric Winter) - Added debugging and error
 * logging code.
 * */

IK_BOOLEAN
IK_InstallSocketSIGPIPEHandler(void)
{

    /* Catch all SIGPIPE signals (they arise when the receiving end of
       a connection has died). */
    if (signal(SIGPIPE, IK_HandleSocketSIGPIPE) == SIG_ERR) {
	IK_vSyslog(LOG_ERR, "IK_InstallSocketSIGPIPEHandler(): socket() "
		   "failed; errno = %d (%s)", errno, strerror(errno));
	return(IK_FALSE);
    }

    /*-----------------------------------------------------------------------*/

    /* Return normally. */
    return(IK_TRUE);

}

/*****************************************************************************/

/*
 * Name: IK_AlarmHandler()
 *
 * Description: This function is the SIGALRM handler for sockets.
 *
 * Parameters:
 * int i_signalNumber - Number for signal which occurred
 *
 * Return Values:
 * None
 *
 * Warnings:
 *
 * Global Variables Used:
 *
 * Revision History:
 *
 * Tuesday 29 October 1996 (Eric Winter) - Initial version.
 *
 * Friday 14 March 1997 (Eric Winter) - Changed to use
 * IK_AlarmHandler() to maintain compatibility with legavy DAAC code.
 * */

void
IK_AlarmHandler(int i_signalNumber)
{

    /* Validate input. */
    assert(i_signalNumber == SIGALRM);

    /*-----------------------------------------------------------------------*/

    /* Re-install this handler. */
    (void) IK_InstallSocketSIGALRMHandler();

    /* If this signal occurred, then an alarm went off, usually on a
       socket read or write operation. */
    IK_ImsErrno = IK_ETIMEDOUT;
    IK_Signal_Brk = IK_TRUE;

}

/*****************************************************************************/

/*
 * Name: IK_HandleSocketSIGPIPE()
 *
 * Description: This function is the SIGPIPE handler for sockets.
 *
 * Parameters:
 * int i_signalNumber - Number for signal which occurred
 *
 * Return Values:
 * None
 *
 * Warnings:
 *
 * Global Variables Used:
 *
 * Revision History:
 *
 * Tuesday 29 October 1996 (Eric Winter) - Initial version.
 * */

void
IK_HandleSocketSIGPIPE(int i_signalNumber)
{

    /* Validate input. */
    assert(i_signalNumber == SIGPIPE);

    /*-----------------------------------------------------------------------*/

    /* Re-install this handler. */
    (void) IK_InstallSocketSIGPIPEHandler();

    /* If this signal occurred, then the receiving end of a connection
       was lost unexpectedly. Set the error global accordingly. */
    IK_ImsErrno = IK_EHOSTUNREACH;

}

/*****************************************************************************/

/*
 * Name: IK_IsSocketReadable()
 *
 * Description: This function uses the select() system call to
 * determine if a socket has data waiting to be read.
 *
 * Parameters:
 * int i_socketDesc - Descriptor for socket to check
 *
 * Return Values:
 * IK_TRUE - If data is ready
 * IK_FALSE - If not
 *
 * Warnings:
 * This function uses a 1 second timeout on the select() call.
 *
 * Global Variables Used:
 *
 * Revision History:
 *
 * Friday 14 March 1997 (Eric Winter) - Initial version.
 *
 * Thursday 22 May 1997 (Eric Winter) - Added debugging and error
 * logging code.
 * */

IK_BOOLEAN
IK_IsSocketReadable(int i_socketDesc)
{

    /* This mask is used to keep track of the descriptor being
       checked. */
    fd_set fdset_readable;

    /* i_numDescriptors is the number of writable or readable
       descriptors detected by select(). */
    int i_numDescriptors;

    /* This is used to specify the timeout for the select() call. */
    struct timeval tv_selectTimeout;

    /*-----------------------------------------------------------------------*/

    /* Validate input. */
    assert(i_socketDesc >= 0);

    /*-----------------------------------------------------------------------*/

    /* Initialize the mask for readability. */
    FD_ZERO(&fdset_readable);

    /* Set the mask bit for the socket to check. */
    FD_SET(i_socketDesc, &fdset_readable);

    /* Initialize the timeout structure for the select() call. */
    tv_selectTimeout.tv_sec = SELECT_TIMEOUT_SEC;
    tv_selectTimeout.tv_usec = SELECT_TIMEOUT_USEC;

    /* Issue the select() call. */
    errno = 0;
    i_numDescriptors = select(i_socketDesc + 1, &fdset_readable,
			      NULL, NULL, &tv_selectTimeout);

    /* Examine the result. If -1, an error occurred. If 0, the socket
       does not have data ready, and the timeout expired. If > 0, the
       socket has data ready. */
    if (i_numDescriptors == -1) {
	IK_vSyslog(LOG_ERR, "IK_IsSocketReadable(): select() failed; "
		   "errno = %d (%s)", errno, strerror(errno));
	return(IK_FALSE);
    } else if (i_numDescriptors == 0) {

	/* The timeout occurred. */
	IK_vSyslog(LOG_NOTICE, "IK_IsSocketReadable(): select() timed out");
	return(IK_FALSE);

    } else {

	/* The socket should have data available now, but
           double-check. */
	if (FD_ISSET(i_socketDesc, &fdset_readable)) {
	    return(IK_TRUE);
	} else {
	    IK_vSyslog(LOG_ERR, "IK_IsSocketReadable(): error in select() "
		       "result set!");
	    return(IK_FALSE);
	}

    }

}

/*****************************************************************************/

/*
 * Name: IK_WaitForSocketToBeReadable()
 *
 * Description: This function uses the select() system call to wait
 * for a socket to have readable data on it. If the maximum specified
 * time elapses, the function returns IK_FALSE.
 *
 * Parameters:
 * int i_socketDesc - Descriptor for socket to check
 * time_t time_maxWait - Maximum time to wait (seconds)
 *
 * Return Values:
 * IK_TRUE - If data is ready
 * IK_FALSE - If not
 *
 * Global Variables Used:
 *
 * Revision History:
 *
 * Wednesday 21 May 1997 (Eric Winter) - Initial version.
 *
 * Thursday 22 May 1997 (Eric Winter) - Added debugging and error
 * logging code.
 *
 * Friday 11 July 1997 (Eric Winter) - Added a code wrapper around the
 * select() call to trap for EINTR, which causes problems on IRIX 6.3.
 * */

IK_BOOLEAN
IK_WaitForSocketToBeReadable(int i_socketDesc, struct timeval *ptv_maxWait)
{

    /* This mask is used to keep track of the descriptor being
       checked. */
    fd_set fdset_readable;

    /* i_numDescriptors is the number of writable or readable
       descriptors detected by select(). */
    int i_numDescriptors;

    /* This is the starting time for the select() call. */
    struct timeval tv_start;

    /* This is used to specify the timeout for the select() call. */
    struct timeval tv_selectTimeout;

    /* This is the number of seconds remaining for the next select()
       call. */
    double d_secondsLeft;

    /* This is a temporary time holder. */
    struct timeval tv;

    /*-----------------------------------------------------------------------*/

    /* Validate input. */
    assert(i_socketDesc >= 0);
    assert(ptv_maxWait != NULL);
    assert(ptv_maxWait->tv_sec >= 0);
    assert(ptv_maxWait->tv_usec >= 0);

    /*-----------------------------------------------------------------------*/

    /* Initialize the mask for readability. */
    FD_ZERO(&fdset_readable);

    /* Set the mask bit for the socket to check. */
    FD_SET(i_socketDesc, &fdset_readable);

    /* Fetch the current time to use as the start time. */
    errno = 0;
    if (gettimeofday(&tv_start, NULL) == -1) {
	IK_vSyslog(LOG_ERR, "IK_WaitForSocketToBeReadable(): Error in "
		   "gettimeofday(): errno = %d (%s)", errno,
		   strerror(errno));
	return(IK_FALSE);
    }

    /* Set the time remaining to be the desired timeout value. */
    d_secondsLeft = IK_TimevalToSecs(ptv_maxWait);

    /* While there is time left, issue the select() call. */
    while (d_secondsLeft > 0.0) {

	/* Fill in the timeout value structure. */
	IK_SecsToTimeval(d_secondsLeft, &tv_selectTimeout);

	/* Issue the select() call. */
	errno = 0;
	i_numDescriptors = select(i_socketDesc + 1, &fdset_readable,
				  NULL, NULL, &tv_selectTimeout);

	/* Examine the result. If -1, an error occurred. If 0, the
	   socket does not have data ready, and the timeout
	   expired. If > 0, the socket has data ready. */
	if (i_numDescriptors == -1) {

	    /* If an error other than EINTR (interrupted system call)
               occurred, abort this function. If EINTR occurred,
               calculate the remaining time and issue the select()
               call again. */
	    if (errno == EINTR) {

		/* Log a message indicating the error was noticed. */
		IK_vSyslog(LOG_NOTICE, "IK_WaitForSocketToBeReadable(): "
			   "select() failed with EINTR; errno = %d (%s)",
			   errno, strerror(errno));

		/* Fetch the current time. */
		errno = 0;
		if (gettimeofday(&tv, NULL) == -1) {
		    IK_vSyslog(LOG_ERR, "IK_WaitForSocketToBeReadable(): "
			       "Error in gettimeofday(): errno = %d (%s)",
			       errno, strerror(errno));
		    return(IK_FALSE);
		}

		/* Calculate the elapsed time. Check and adjust if
                   needed to ensure that the microseconds part is
                   positive. */
		tv.tv_sec -= tv_start.tv_sec;
		tv.tv_usec -= tv_start.tv_usec;
		if (tv.tv_usec < 0) {
		    tv.tv_sec--;
		    tv.tv_usec += ONE_MILLION;
		}
		d_secondsLeft = IK_TimevalToSecs(&tv);

		/* Go back to the start of the loop. */
		continue;

	    } else {

		/* Some other error occurred during the select()
                   call. */
		IK_vSyslog(LOG_ERR, "IK_WaitForSocketToBeReadable(): select() "
			   "failed; errno = %d (%s)", errno, strerror(errno));
		return(IK_FALSE);

	    }

	} else if (i_numDescriptors == 0) {

	    /* The timeout occurred. */
	    IK_vSyslog(LOG_NOTICE, "IK_WaitForSocketToBeReadable(): select() "
		       "timed out");
	    return(IK_FALSE);

	} else {

	    /* The socket should have data available now, but
	       double-check. */
	    if (FD_ISSET(i_socketDesc, &fdset_readable)) {
		return(IK_TRUE);
	    } else {
		IK_vSyslog(LOG_ERR, "IK_WaitForSocketToBeReadable(): error in "
			   "select() result set!");
		return(IK_FALSE);
	    }

	}
	
    }

    /* If we got this far (and we really should never get here), no
       data is ready for reading. */
    return(IK_FALSE);

}
