/*
 * Name: IK_TaskManager.c
 *
 * Description: This file contains functions for the IK_TaskManager
 * data structure, which is used as the controlling mechanism for
 * collections of communication connections.
 *
 * Notes:
 *
 *-------------------------------------------------------------------------
 *
 * RCS information:
 *
 * $RCSfile: IK_TaskManager.c,v $
 *
 * $Id: IK_TaskManager.c,v 1.1 1997/03/06 15:38:52 amrhein Exp $
 *
 * $Log: IK_TaskManager.c,v $
 * Revision 1.1  1997/03/06 15:38:52  amrhein
 * Initial revision
 *
 * Revision 6.1.1.4  1996/12/11  18:21:11  winter
 * Added new function IK_TaskManager_ResetLastExecutedTask().
 *
 * Revision 6.1.1.3  1996/11/25  20:21:23  winter
 * Filled out stub for IK_TaskManager_DispatchTask().
 *
 * Revision 6.1.1.2  1996/11/25  19:54:46  winter
 * Added code for IK_TaskManager_ManageConnection(),
 * IK_TaskManager_UnmanageConnection(), and MapConnectionToTask().
 *
 * Revision 6.1.1.1  1996/11/25  16:41:16  winter
 * Placeholder for new branch during Task Manager development.
 *
 * Revision 6.1  1996/11/25  16:40:18  winter
 * Initial checkin of basic functions.
 *
 * */

/*****************************************************************************/

/* Define the RCS identifier string. */
static const char *rcsid = "$Id: IK_TaskManager.c,v 1.1 1997/03/06 15:38:52 amrhein Exp $";

/*****************************************************************************/

/* #include directives */

/* This header must be #included first in order to use
   autoconf-generated information. */
#include "IK_Ims.h"

/* System headers */

/* Standard headers */
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

/* Third-party library headers */

/* IMS headers */
#include "IK_Task.h"
#include "IK_TaskManager.h"

/*****************************************************************************/

/* Local #define directives */

/*****************************************************************************/

/* Data type definitions */

/*****************************************************************************/

/* External variable declarations */

/*****************************************************************************/

/* Local variable definitions */

/*****************************************************************************/

/* External function prototypes */

/*****************************************************************************/

/* Local function prototypes */

/* Task Manager creation and deletion functions */
static IK_TaskManager *Allocate(void);
static void Free(IK_TaskManager *ptaskmanager);

/* Utility functions */
static IK_Task *MapConnectionToTask(IK_TaskManager *ptaskmanager,
				    CONNECTION_INFO *pCI);

/*****************************************************************************/

/*
 * Name: IK_TaskManager_New()
 *
 * Description: This function creates and initializes a new
 * IK_TaskManager data structure for a collection of connections.
 *
 * Parameters:
 * None
 *
 * Return Value:
 * IK_TaskManager *ptaskmanager_new - Pointer to new IK_TaskManager structure
 * NULL - If an error occurs
 *
 * Notes:
 *
 * Revision History:
 *
 * Monday 25 November 1996 (Eric Winter) - Started initial version.
 * */

IK_TaskManager *
IK_TaskManager_New(void)
{

    /* This points to the new IK_TaskManager structure. */
    IK_TaskManager *ptaskmanager_new;

    /*-----------------------------------------------------------------------*/

    /* Allocate the new IK_TaskManager structure. */
    ptaskmanager_new = Allocate();
    if (ptaskmanager_new == NULL) {
	return(NULL);
    }

    /* Create the ring list to hold all tasks. */
    ptaskmanager_new->pringlist_tasks = IK_RingList_New();
    if (ptaskmanager_new->pringlist_tasks == NULL) {
	IK_TaskManager_Delete(ptaskmanager_new);
	return(NULL);
    }

    /* No tasks have been executed yet. */
    ptaskmanager_new->ptask_lastExecuted = NULL;

    /* Return a pointer to the new task manager. */
    return(ptaskmanager_new);

}

/*****************************************************************************/

/*
 * Name: IK_TaskManager_Delete()
 *
 * Description: This function delete a IK_TaskManager structure.
 *
 * Parameters:
 * IK_TaskManager *ptaskmanager - Pointer to task manager to delete
 *
 * Return Value:
 * None
 *
 * Notes:
 *
 * Revision History:
 *
 * Monday 25 November 1996 (Eric Winter) - Started initial version.
 * */

void
IK_TaskManager_Delete(IK_TaskManager *ptaskmanager)
{

    /* Validate input. */
    assert(ptaskmanager != NULL);

    /*-----------------------------------------------------------------------*/

    /* Delete any tasks which may currently be under management. */
    if (ptaskmanager->pringlist_tasks != NULL) {
	IK_RingList_Delete(ptaskmanager->pringlist_tasks);
    }

    /* Free the task manager. */
    Free(ptaskmanager);

}

/*****************************************************************************/

/*
 * Name: IK_TaskManager_ManageConnection()
 *
 * Description: This function adds a new connection to the list of
 * tasks managed by this task manager.
 *
 * Parameters:
 * IK_TaskManager *ptaskmanager - Task manager to add connection to
 * CONNECTION_INFO *pCI - Pointer to connection data for new connection
 *
 * Return Value:
 * IK_BOOLEAN IK_TRUE - If connection is now managed
 * IK_BOOLEAN IK_FALSE - If an error occurs
 *
 * Notes:
 *
 * Revision History:
 *
 * Monday 25 November 1996 (Eric Winter) - Started initial version.
 * */

IK_BOOLEAN
IK_TaskManager_ManageConnection(IK_TaskManager *ptaskmanager,
				CONNECTION_INFO *pCI)
{

    /* This new task will hold the new connection. */
    IK_Task *ptask_new;

    /*-----------------------------------------------------------------------*/

    /* Validate input. */
    assert(ptaskmanager != NULL);
    assert(pCI != NULL);

    /*-----------------------------------------------------------------------*/

    /* Create a new task for this connection. */
    ptask_new = IK_Task_New(pCI);
    if (ptask_new == NULL) {
	return(IK_FALSE);
    }

    /* Add the new task to the list of managed tasks. */
    if (IK_RingList_AddItem(ptaskmanager->pringlist_tasks, ptask_new)
	== IK_FALSE) {
	IK_Task_Delete(ptask_new);
	return(IK_FALSE);
    }

    /* Return normally. */
    return(IK_TRUE);

}

/*****************************************************************************/

/*
 * Name: IK_TaskManager_UnmanageConnection()
 *
 * Description: This function removes a connection from the list of
 * tasks managed by this task manager.
 *
 * Parameters:
 * IK_TaskManager *ptaskmanager - Task manager to remove connection
 * from
 * CONNECTION_INFO *pCI - Pointer to connection data for connection to
 * remove
 *
 * Return Value:
 * IK_BOOLEAN IK_TRUE - If connection is removed from management
 * IK_BOOLEAN IK_FALSE - If an error occurs
 *
 * Notes:
 *
 * Revision History:
 *
 * Monday 25 November 1996 (Eric Winter) - Started initial version.
 * */

IK_BOOLEAN
IK_TaskManager_UnmanageConnection(IK_TaskManager *ptaskmanager,
				  CONNECTION_INFO *pCI)
{

    /* This task holds the connection to remove. */
    IK_Task *ptask;

    /*-----------------------------------------------------------------------*/

    /* Validate input. */
    assert(ptaskmanager != NULL);
    assert(pCI != NULL);

    /*-----------------------------------------------------------------------*/

    /* Find the task for this connection. */
    ptask = MapConnectionToTask(ptaskmanager, pCI);
    if (ptask == NULL) {
	return(IK_FALSE);
    }

    /* If this task was last executed, set the pointer to the last
       executed task to the previous task. */
    if (ptaskmanager->ptask_lastExecuted == ptask) {
	ptaskmanager->ptask_lastExecuted =
	    IK_RingList_GetPrevItem(ptaskmanager->pringlist_tasks,
				    ptask);
	assert(ptaskmanager->ptask_lastExecuted != NULL);
    }

    /* Remove the task from the list of tasks. */
    IK_RingList_RemoveItem(ptaskmanager->pringlist_tasks, ptask);

    /* Delete the task. */
    IK_Task_Delete(ptask);

    /* Return normally. */
    return(IK_TRUE);

}

/*****************************************************************************/

/*
 * Name: IK_TaskManager_DispatchTask()
 *
 * Description: This function determines which task to execute, and
 * does so.
 *
 * Parameters:
 * IK_TaskManager *ptaskmanager - Task manager to execute task for
 *
 * Return Value:
 * IK_BOOLEAN IK_TRUE - If execution goes normally (whether or not a
 * task is executed)
 * IK_BOOLEAN IK_FALSE - If an error occurs
 *
 * Notes:
 *
 * Revision History:
 *
 * Monday 25 November 1996 (Eric Winter) - Started initial version.
 * */

IK_BOOLEAN
IK_TaskManager_DispatchTask(IK_TaskManager *ptaskmanager)
{

    /* This points to the original value of the last executed task for
       this task manager. */
    IK_Task *ptask_lastExecuted;

    /* This is used as a cursor to loop through all the tasks until
       one is executed. */
    IK_Task *ptask;

    /*-----------------------------------------------------------------------*/

    /* Validate input. */
    assert(ptaskmanager != NULL);

    /*-----------------------------------------------------------------------*/

    /* Save the last executed task. */
    ptask_lastExecuted = ptaskmanager->ptask_lastExecuted;
    assert(ptask_lastExecuted != NULL);

    /* Starting with the task _after_ the last executed task, try to
       find a task which can do something useful. If none can be found
       after checking all tasks, return. */
    ptask = IK_RingList_GetNextItem(ptaskmanager->pringlist_tasks,
				    ptask_lastExecuted);
    do {
	assert(ptask != NULL);
	if (IK_Task_CanExecute(ptask) == IK_TRUE) {
	    if (IK_Task_DispatchTaskSegment(ptask) == IK_FALSE) {
		return(IK_FALSE);
	    }
	    ptaskmanager->ptask_lastExecuted = ptask;
	    break;
	}
	ptask = IK_RingList_GetNextItem(ptaskmanager->pringlist_tasks,
					ptask);
    } while (ptask != ptask_lastExecuted);

    /* Return normally. */
    return(IK_TRUE);

}

/*****************************************************************************/

/*
 * Name: IK_TaskManager_ResetLastExecutedTask()
 *
 * Description: This resets the "last executed task" to be the "last"
 * task in the ring list. The net effect is that the next time the
 * Task Manager is invoked, the head Task will be tapped for
 * execution.
 *
 * Parameters:
 * IK_TaskManager *ptaskmanager - Pointer to IK_TaskManager to reset
 * last executed task for
 *
 * Return Value:
 * IK_BOOLEAN IK_TRUE - If the last executed task is reset
 * IK_BOOLEAN IK_FALSE - If an error occurs
 *
 * Notes:
 *
 * Revision History:
 *
 * Wednesday 11 December 1996 (Eric Winter) - Started initial version.
 * */

IK_BOOLEAN
IK_TaskManager_ResetLastExecutedTask(IK_TaskManager *ptaskmanager)
{

    /* Validate input. */
    assert(ptaskmanager != NULL);

    /*-----------------------------------------------------------------------*/

    /* Set the last executed task to be the task previous to the head
       task. */
    assert(ptaskmanager->pringlist_tasks->plistnode_head != NULL);
    ptaskmanager->ptask_lastExecuted =
	ptaskmanager->pringlist_tasks->plistnode_head->plistnode_prev->pv_item;

    /* Return normally. */
    return(IK_TRUE);

}

/*****************************************************************************/

/*
 * Name: IK_TaskManager_Dump()
 *
 * Description: This debugging function dumps the contents of the
 * specified IK_TaskManager to the specified stream.
 *
 * Parameters:
 * FILE *pFILE - Pointer to FILE for output
 * IK_TaskManager *ptaskmanager - Pointer to IK_TaskManager to dump
 *
 * Return Value:
 * None
 *
 * Notes:
 *
 * Revision History:
 *
 * Monday 25 November 1996 (Eric Winter) - Started initial version.
 * */

void
IK_TaskManager_Dump(FILE *pFILE, IK_TaskManager *ptaskmanager)
{

    /* Validate input. */
    assert(pFILE != NULL);

    /*-----------------------------------------------------------------------*/

    /* Print the task manager contents to the stream. */
    (void) fprintf(pFILE, "Dumping IK_TaskManager at address %p:\n",
		   ptaskmanager);
    if (ptaskmanager != NULL) {
	(void) fprintf(pFILE, "pringlist_tasks = %p\n",
		       ptaskmanager->pringlist_tasks);
	(void) fprintf(pFILE, "ptask_lastExecuted = %p\n",
		       ptaskmanager->ptask_lastExecuted);
    } else {
	(void) fprintf(pFILE, "(NULL)\n");
    }

}

/*****************************************************************************/

/* NOTE - PRIVATE MODULE FUNCTIONS ONLY BELOW THIS POINT! */

/*****************************************************************************/

/*
 * Name: Allocate()
 *
 * Description: This function allocates space for a new IK_TaskManager
 * structure.
 *
 * Parameters:
 * None
 *
 * Return Value:
 * IK_TaskManager * - Pointer to new IK_TaskManager structure
 * NULL - If an error occurs
 *
 * Notes:
 * (1) This function is used to provide flexibility in the allocation
 * mechanism. For example, a group of task managers can be
 * pre-allocated and doled out from this function. For now, it simply
 * acts as a front-end to malloc().
 *
 * Revision History:
 *
 * Monday 25 November 1996 (Eric Winter) - Started initial version.
 *
 * */

static IK_TaskManager *
Allocate(void)
{

    /* Allocate the new task manager and return it. */
    return(malloc(sizeof(IK_TaskManager)));

}

/*****************************************************************************/

/*
 * Name: Free()
 *
 * Description: This function frees the memory allocated to a
 * IK_TaskManager structure.
 *
 * Parameters:
 * IK_TaskManager *ptaskmanager - Pointer to IK_TaskManager to free
 *
 * Return Value:
 * None
 *
 * Notes:
 * (1) See note (1) for Allocate().
 *
 * Revision History:
 *
 * Monday 25 November 1996 (Eric Winter) - Started initial version.
 *
 * */

static void
Free(IK_TaskManager *ptaskmanager)
{

    /* Validate input. */
    assert(ptaskmanager != NULL);

    /*-----------------------------------------------------------------------*/

    /* Delete the task manager. */
    free(ptaskmanager);

}

/*****************************************************************************/

/*
 * Name: MapConnectionToTak()
 *
 * Description: This function finds the IK_Task which contains the
 * specified connection pointer (if any).
 *
 * Parameters:
 * IK_TaskManager *ptaskmanager - Pointer to IK_TaskManager to search
 * for task
 * CONNECTION_INFO *pCI - Connection to search for task for
 *
 * Return Value:
 * IK_Task *ptask - Pointer to task which contains the connection
 * NULL - If an error occurs, or connection not found
 *
 * Notes:
 *
 * Revision History:
 *
 * Monday 25 November 1996 (Eric Winter) - Started initial version.
 *
 * */

static IK_Task *
MapConnectionToTask(IK_TaskManager *ptaskmanager, CONNECTION_INFO *pCI)
{

    /* This points to the current task. */
    IK_Task *ptask;

    /* This points to the original head task. */
    IK_Task *ptask_head;

    /*-----------------------------------------------------------------------*/

    /* Validate input. */
    assert(ptaskmanager != NULL);
    assert(pCI != NULL);

    /*-----------------------------------------------------------------------*/

    /* Starting with the first task, search each task in the ring list
       until the task for the current connection is found. */
    assert(ptaskmanager->pringlist_tasks != NULL);
    ptask_head = (IK_Task *)
	IK_RingList_GetHeadItem(ptaskmanager->pringlist_tasks);
    assert(ptask_head != NULL);
    ptask = ptask_head;
    do {
	if (ptask->pCI == pCI) {
	    return(ptask);
	}
	ptask = IK_RingList_GetNextItem(ptaskmanager->pringlist_tasks, ptask);
	assert(ptask != NULL);
    } while (ptask != ptask_head);

    /* A match was not found. */
    return(NULL);

}
