/********* IK_ValNames.c  - Randy Barth - November 2, 1995 ********/
#include "IK_Ims.h"

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <assert.h>
#include <string.h>
#include <limits.h>
#include "IK_ValNames.h"
/* maximum number of valids read */

typedef struct HashNode {
	struct HashNode* next;
		/* Next hash node in list of values that hash to this value. */
	short index;
		/* Indexes into validsNames and alCnt arrays.  Valid that hashes to value n 
			will have it's index stored in some HashNode in the list headed by 
			validsTable[n].  If there are multiple valids with the same name (but
			different indexes), they will all be stored in the same list.  */
	} HashNode;

extern IK_BITVECTOR *IK_fieldMask; /* indicates which field a valid belongs */

static int offset; /* used to calculate the row in the IK_fieldMask */
static int numcols; /* number of bytes in a bitvector; computed in IK_ValidsTableInit
                        from validsNum. */
static int validsNum = 0;
		/* Number of valids values, as set by IK_ValidsTableInit. 
			Used as size of validsNames, alCnt, and validsTable 
			arrays. */

static HashNode* validsTable = NULL;
		/* Array of HashNode lists.  validsNum is used as the size of the array 
			since that makes a good estimate, though it wouldn't have to be.  
			The index field of each node in the list headed at validsTable[n] 
			points at index within validsNames and alCnt for a valid that 
			hashes to n, or is -1 if that entry is not in use. */

extern char** validsNames;
		/* Array of char* to valids names.  validsNames[i] points to
			null-terminated name of ith valid, followed immediately by
			zero or more null-terminated alias names.  The number
			of aliases is given in alCnt[i].  For example, validsNames[i]
			might point at:    carbon dioxide\0CO2\0.   */

short* alCnt = NULL;
		/* Array of integers; alCnt[i] is the number of aliases that follow 
			*validsNames[i].	*/ 




int hashStr (char const *s, const int limit)

		/*  Hashes string s to value between 0 and limit-1.  Each 
		    alphnumeric character (with alphas downshifted)  is XORed
		    into running 24-bit hash after left circular shift of bits. Result is 
			taken modulo limit to get function result.  No more than 
			the first 200 characters of s are processed.  */

{   int i;				/* Position within s. */
    int hash = 0;		/* Accumulating hash value. */
	char ch;			/* Character of s being examined. */

    assert (limit > 0);
    for (i = 0;  i < 200 && (ch = s[i]) != '\0'; i++) {
							    /* For each character of string: */
		if (isalnum(ch)) {		/* Process only significant characters. */
			if (isupper(ch)) ch = tolower(ch);
			hash = ((hash << 1 | hash >> 23) ^ ch) & 0xffffff;
                                /* Rotate hash byte left one bit,
                                    XOR new character, mask to three bytes. */
			}
		}										 
    return (hash % limit);
}



void setHash (const char* s, const int x)

	/* Put entry for valid/alias s whose bitmap index is x into the validsTable
		at the place where it hashes.  Not that duplicate entries may exist.  This
		routine does not check to insure IK_ValidsTableInit has been called,
		but assumes it's caller has assured this. */

{	int h;					/* Hash value for s. */
	HashNode* n;			/* Node in the list for hash h. */

	h = hashStr(s, validsNum);
	n = validsTable + h;
			/* Get first node in list for h. */
	if (n->index >= 0) {
			/* Some things have previously hashed here. */
		while (n->next != NULL) n = n->next;
			/* Scan down list to last node. */
		n->next = (HashNode *) malloc(sizeof(HashNode));
			/* Add new node at end. */
		assert (n->next != NULL);
			/* Abort if out of memory. */
		n = n->next;
			/* Point to empty node. */
		}
	n->index = x;
	n->next = NULL;
			/* Put new entry here. */
}


IK_BOOLEAN signifCompare (const char* pattern, const char* target,
		const IK_BOOLEAN prefixOK)

	/* This routine is used for all searches of the valids and controls the definition
		of a match.  Returns true if the string pattern matches target.  If prefixOK,
		pattern can be a prefix of target, otherwise it must consume the significant
		part.  Only alphameric characters are normally significant (and alphabetics
		are compared case blind), but if other characters are present in pattern
		they must match those in target.  So "A-b" matches "a - /b" but not "a/b". */

{	char pch, tch;		/* Current pattern and target character, downshifted. */

	while ((pch = *pattern) != '\0') {	
				/* For each character of pattern. */
		if (isupper(pch)) pch = tolower(pch);
		tch = *target;
		if (isupper(tch)) tch = tolower(tch);
		if (!(isalnum(pch) && isalnum(tch))) {
				/* Unless both are alphanumeric: */
			while (pch != tch) {
					/* Advance target through any insignificant characters until
						you find one that matches pattern;  fail if you consume
						target or bump into nonmatching alphanumerics. */
				if (isalnum(tch) || tch == '\0') return IK_FALSE;
				tch = *(++target);
				if (isupper(tch)) tch = tolower(tch);
			}
		}
				/* pch and tch are both alphanumeric or are equal. */
		if (pch != tch) return IK_FALSE;
		pattern++;	target++;
	}
	if (prefixOK) return IK_TRUE;
			/* All matched through end of pattern. */
	while (*target != '\0') {
				/* Any remaining target must be insignificant. */
		if (isalnum(*target)) return IK_FALSE;
		target++;
	}
	return IK_TRUE;
}


IK_BOOLEAN valMatch (const char* pattern, const short x,
		const IK_BOOLEAN prefixMatch, const IK_BOOLEAN includeAliases)

	/* Return true if string pattern matches the valid name or an alias (if 
		includeAliases) at validsNames[x].  If prefixMatch, pattern need only
		be a prefix of valid to match.  Matching is defined by signifCompare. */

{	int namesLeft;			/* Number of names and aliases yet to test. */
	char* s;				/* Pointer to name/alias to match.*/

	assert (0 <= x && x < validsNum);
			/* Validate x. */
	s = validsNames[x];
			/* Point to main name. */
	for (namesLeft = includeAliases ? alCnt[x] + 1 : 1; namesLeft > 0; namesLeft--) {
				/* For each name: */
		if (signifCompare(pattern, s, prefixMatch)) return IK_TRUE;
				/* If that matches, we're done. */
		while (*s != '\0') s++;
		s++;
				/* Advance s past this name. */
	}
	return IK_FALSE;
}
	

/*******************************************************************************
 * 
 * Name: IK_ValidsTableInit 
 *
 * Description: Sets up data structures to keep track on the names of the valids
 *     and their aliases: it allocates memory and intializes the data
 *     structures.
 *
 * Arguments:  n is the maximum number of aliases that have to be accomodated.
 *
 * Return Values: none.
 *
 * Warnings: this function should only be called once, and, since it sets the
 *    static variables validsNum & numcols, it must be called before any of the
 *    other functions in this module.
 *
 * Global Variables Used:  validsNames[]; the following static variables
 *     are used: validsNum, numcols, & validsTable[].
 *
 * Pre- and Post-Conditions:
 *	pre-condition: the Client has just been started.  The names of the 
 *          various valids are about to be read from disk.
 *      post-condition: the hash table (validsTable) and the array that will
 *          contain pointers to the stings of valids and alias strings will
 *          have been allocated and initialized.
 *
 * Revision History:
 *    12/26/96 (kvj) -- added calculation of numcols and the doc header.
 *    
 * */
void IK_ValidsTableInit (const int n)
	/*	Initialize IK_ValidsTable.  n is the number of valids to be processed. */

{	int i;					/* Work variable. */
	
	assert(validsNum == 0);
			/* Can't be called more than once. */  
        /* set values for the variables with module span: */
	    validsNum = n;
            numcols = (validsNum + CHAR_BIT -1) / CHAR_BIT; /* number of bytes in a
                                                             * bitvector */
	validsTable = (HashNode *) calloc(n, sizeof(HashNode));
			/* Make n headers for hash lists. */
	assert (validsTable != NULL);
			/* Assure no memory failure. */
	validsNames = (char * *) calloc(n, sizeof(char*));
	assert (validsNames != NULL);
	alCnt = (short *) calloc(n, sizeof(short));
	assert (alCnt != NULL);
	for (i = 0; i < n; i++) {
			/* Initialize all entries. */
		validsTable[i].next = NULL;
				/* Nothing follows in this list. */
		validsTable[i].index = -1;
				/* This entry isn't in use either. */
		validsNames[i] = NULL;
				/* No name for this index yet. */
		alCnt[i] = 0;
				/* Name has no aliases. */
	}
	return;
}
/******************************************************************************/
void IK_SetValid (const char* v, const int x)
	
	/* Store a copy of valid v (which may include aliases) into validsTable at 
		index x in such a way it can be found later by IK_FindValids 
		and IK_ValidName.  Format of v is 
			"valid-name {alias-name}{alias-name} terminator"
		That is, alias(es) will be surrounded by braces and follow
		the name; terminator is either null character or new-line. */

{	const char* s;		/* Source string. */ 
	char* r;			/* Receiving strings for copy. */
	const char* prev;	/* Beginning of previous name/alias. */
	int i;				/* Temporary. */

	assert (0 <= x && x < validsNum);
			/* Assure IK_ValidsTableInit was called and 0 <= x < validsNum.*/
	for (i = 1, s = v; *s != '\0' && *s != '\n'; s++)  if (*s != '}') i++;
			/* Scan v counting all but "}"; i starts at 1 for terminator. */
	prev = r = validsNames[x] = (char *) malloc(i);
			/* Allocate space for copy of name and aliases. */
	assert (r != NULL);
			/* Assure we have memory. */
	for (s = v;  *s != '\0' && *s != '\n'; s++)
			/* For each character of s: */
		if (*s != '}')  
					/* Discard any "}" during copy. */
			if (*s != '{')
				*(r++) = *s;
					/* Simply copy all but "{". */
			else {
				*(r++) = '\0';
					/* Replace "{" by nulls to terminate previous name/alias. */
				setHash(prev, x);
					/* Install previous name/alias into table. */
				prev = r;
				alCnt[x]++;
					/* We've got another alias starting now. */
			}
	*r = '\0';
			/* Last name/alias is ended. */
	setHash	(prev, x);
			/* Install it. */
}
				

const char* IK_ValidName (const int x)

	/* Return a const char * to  the valid at index x.  Note that valid will be terminated
		by null and will not include aliases.  "<unknown>" returned for out-of-range
		values.  */

{	static  char *unknown_valid = "<unknown>";

	assert(validsNames != NULL);
		/* Assure table was built. */
	if (0 <= x && x < validsNum) return validsNames[x];
		/* Return good valid. */
	return unknown_valid;	
}


int IK_FindValidsBeginning (const char* pattern, const int fieldType,
	IK_BOOLEAN includeAliases, IK_BITVECTOR* result, int* numHits)

	/* Turn on bits in result corresponding to fields from validsTable
		of type fieldType whose name or alias (if includeAliases) begin with
		pattern. Only alphanumeric characters of pattern are considered 
		and match is case-blind.  Return the number of matches
		(number of bits set in result) in *numHits. The bit index
		of the first (or only) hit is returned as function value 
		(or is -1 if *numHits==0).  */

{ 	int firstHit = -1;		/* Index of first match, or -1 if none. */
	short x;				/* Index in validsTable being examined. */

	assert (validsTable != NULL);
			/* Assure we've been initialized. */
	*numHits = 0;
			/* Nothing found yet. */
	for (x = 0; x < validsNum; x++) {
				/* For each valid in the table: */
		offset = CALC_OFFSET(fieldType,0,numcols);
		if (TESTBIT(IK_fieldMask+offset, x) && 
					valMatch(pattern, x, IK_TRUE, includeAliases)) {
				/* If this valid is of proper field and it is a prefix of the pattern: */
			SETBIT(result, x);
			(*numHits)++;
			if (firstHit < 0) firstHit = x;
				/* Set it's bit, count it, and note if it's first. */
		}
	}
	return firstHit;
}


int IK_FindValidsContaining (const char* pattern, const int fieldType,
	IK_BOOLEAN includeAliases, IK_BITVECTOR* result, int* numHits)

	/* Turn on bits in result corresponding to fields from validsTable
		of type fieldType whose name or alias (if includeAliases) contain
		pattern. Only alphanumeric characters of pattern are considered 
		and match is case-blind.  Return the number of matches
		(number of bits set in result) in *numHits. The bit index
		of the first (or only) hit is returned as function value 
		(or is -1 if *numHits==0).  */

{ 	int firstHit = -1;		/* Index of first match, or -1 if none. */
	short x;				/* Index in validsTable being examined. */
	IK_BOOLEAN noMatchOnX;	/* True until some name or alias at index x matches. */
	int namesLeft;			/* Number of names and aliases yet to be tested at x. */
	char* s;				/* Starting point for target of comparison. */

	assert (validsTable != NULL);
			/* Assure we've been initialized. */
	*numHits = 0;
			/* Nothing found yet. */
	for (x = 0; x < validsNum; x++) {
				/* For each valid in the table: */
		offset = CALC_OFFSET(fieldType,0,numcols);
		if (TESTBIT(IK_fieldMask+offset, x)) {
					/* If this valid is of proper field: */
			noMatchOnX = IK_TRUE;
					/* True until valid or alias for x matches. */
			s = validsNames[x];
					/* Start with real name of valid x. */ 
			for (namesLeft = includeAliases ? alCnt[x] + 1 : 1;
					namesLeft > 0 && noMatchOnX;  namesLeft--) {
				for ( ; *s != '\0' && noMatchOnX; s++) {
						/* Starting at each character of name/alias: */
					if (signifCompare(pattern, s, IK_TRUE)) {
							/* If you can match starting here: */ 
						noMatchOnX = IK_FALSE;
								/* No need to try further on this x. */
						SETBIT(result, x);
						(*numHits)++;
						if (firstHit < 0) firstHit = x;
							/* Set it's bit, count it, and note if it's first. */
					}
				}
				s++;
						/* Step over trailing null to prepare for next name. */
			}
			
		}
	}
	return firstHit;
}


int IK_FindValids (const char* pattern, const int fieldType,
	IK_BITVECTOR* result, int* numHits)

	/* Turn on bits in result corresponding to fields from validsTable
		of type fieldType whose name or alias match pattern.  Only 
		alphanumeric characters of pattern are considered and
		match is case-blind, but match must otherwise be complete.
		Return the number of matches (number of bits set in result) 
		in *numHits. The bit index of the first (or only) hit is returned 
		as function value (or is -1 if *numHits==0). 
		
		Bonus feature: if pattern ends with an "*" then it will be treated as
		an assumed wild card;  an optional leading "*" will also be 
		processed in that case to call in turn IK_FindValidsContaining or 
		IK_FindValidsBeginning respectively. */

{	int firstHit = -1;	/* Index of first match, or -1 if none. */
	HashNode* n;		/* Node being examined
									of chain to which pattern hashes. */
	short x;			/* Index in validsTable being examined. */
	short lastPat;		/* Index of last pattern character. */
	char* newPat;		/* Copy of pattern without wild cards. */
	short newLen;		/* Legnth of newPat. */
	short i, j;				/* Indexes. */
	
	assert (validsTable != NULL);
			/* Assure we've been initialized. */
        numcols = (validsNum + CHAR_BIT -1) / CHAR_BIT;
	*numHits = 0;
			/* Nothing found yet. */
	lastPat = strlen(pattern) - 1;
	if (lastPat < 0) lastPat = 0;
			/* Get index of last pattern character or 0 if null. */
	if (pattern[lastPat] == '*')  {
		if (pattern[0] == '*')	{
					/* Wild card begin and end */
			newLen = lastPat > 1 ? lastPat - 1: 0;
					/* New string without wildcards normally needs length = 
						lastPat - 1 == strlen(pattern) - 2 wildcard, so *xxx*
						is newLen of 3;  either "**" or "*" will produce null string
						of length 0. */
			i = 1;	
					/* Start copy at pattern[1]. */
		} else {
					/* Wild card at end only. */
			newLen = lastPat > 0 ? lastPat : 0;
					/* Normal newLen is lastPat == strlen(pattern) - 1 wildcard. */
			i = 0;
		}
		newPat = (char *) malloc(newLen + 1);
		assert (newPat != NULL);
		for (j = 0; i < lastPat; i++, j++)
			newPat[j] = pattern[i];
		newPat[j] = '\0';
				/* Make copy of pattern without wild cards. */
		if (pattern[0] == '*')
			firstHit = IK_FindValidsContaining(newPat, fieldType, IK_TRUE, result, numHits);
		else
			firstHit = IK_FindValidsBeginning(newPat, fieldType, IK_TRUE, result, numHits);
				/* Call corresponding routine and return it's results after freeing copy. */
		free(newPat);
		return firstHit;
	}			
	n = validsTable + hashStr(pattern, validsNum);
			/* Start n at the node directly in the validsTable for the hash index. */
	if (n->index == -1) return firstHit;
			/* Return if this hash bucket not in use. */
	for ( ; n != NULL; n = n-> next) {
			/* For each valid that hashed to this bucket: */
		x = n->index;
				/* Get index of this valid and it's aliases. */
		offset = CALC_OFFSET(fieldType,0,numcols);
		if (TESTBIT(IK_fieldMask+offset, x) && 
					valMatch(pattern, x, IK_FALSE, IK_TRUE)) {
				/* If this valid is of proper field and it (or its alias) is a 
						complete match of the pattern: */
			SETBIT(result, x);
			(*numHits)++;
			if (firstHit < 0) firstHit = x;
				/* Set it's bit, count it, and note if it's first. */
		}
	}
	return firstHit;
}

/*
 * Name: IK_ValStringToIndex()
 *
 * Description:
 *   This function returns the valid code number corresponding to the given
 *   valid string name and field type.
 *
 * Version: 6.1.1.3  Date: 01/14/97  Programmer: Lloyd M. Terrell
 *   Guide search modifications: New Function
 *
 * Return Value:
 *
 * Global Variables Used:
 *
 * Pre- and Post-Conditions:
 *
 * Warnings:
 *
 */
int IK_ValStringToIndex(const char* pattern, const int fieldType)
{ 
int        firstHit = -1;    /* Index of first match, or -1 if none. */
HashNode  *n;	             /* Node being examined */
short      x;	             /* Index in validsTable being examined. */

    n = validsTable + hashStr(pattern, validsNum);

    /* Start n at the node directly in the validsTable for the hash index. */
    if (n->index == -1) return firstHit;

    /* Return if this hash bucket not in use. */
    for ( ; n != NULL; n = n-> next)
    {
        /* For each valid that hashed to this bucket: */
	x = n->index;

	/* Get index of this valid and it's aliases. */
	offset = CALC_OFFSET(fieldType,0,numcols);

	if (TESTBIT(IK_fieldMask+offset, x) && (!strcasecmp(pattern, validsNames[x])))
        {
            firstHit = x;
            break;
        }
    }
return firstHit;
}
