/*
 * $Header: /imsdev/beta-3.5/src/IK_lib/RCS/strtokenize.c,v 1.6 1997/03/06 15:38:52 amrhein Exp $
 */

#include "IK_Ims.h"

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>

static char *rcsid =
"$Id: strtokenize.c,v 1.6 1997/03/06 15:38:52 amrhein Exp $";

char *strtokenize(const char *s, const char *sep);

/*

   strtokenize() is (imho) a better version of the Unix library
   function, strtok().  It accepts the same arguments but behaves in a
   subtly different way.  Before you use this function, make sure you
   understand what strtok() does.

   strtok() has the feature (or bug, depending on your point of view)
   that it separates tokens which are separated by "one or more" of
   the characters in the separator.  In a Perl regular expression, the
   separator would look like /[${sep}]+/.  While this is useful for
   breaking up strings based on whitespace, it can have unintended
   effects when used on more tightly formatted strings.

   For instance, you cannot safely use strtok() to scan the
   /etc/passwd file.  Consider the following passwd entry where the
   user has no gecos (real name) field and no shell:

   uucp:*:4:8::/var/spool/uucppublic:

   Calling strtok iteratively would yield the following strings:

   'uucp', '*', '4', '8', '/var/spool/uucppublic'

   when what you really want is

   'uucp', '*', '4', '8', '', '/var/spool/uucppublic', ''


   strtok() completely misses the zero length fields.

   strtokenize() works like strtok() except that it grabs tokens which
   are separated by exactly one character from the separator string.
   Here is an example which parses a line from the passwd file:

   char s[BUFSIZE], *token;
   FILE *fp;

   fp=fopen("/etc/passwd","r");
   fgets(s,BUFSIZE,fp);
   for (token=strtokenize(s,":"); token!=NULL; token=strtokenize(NULL,":"))
          printf("token = %s\n",token);
	  

   strtokenize(), unlike strtok(), does *not* modify the input string
   at all.  strtokenize() returns a pointer to a static buffer which
   contains the current token.
   
   

   P. Ryan, HSTX, 26.5.94, ryan@odouls.stx.com

   **/




char *strtokenize(const char *s, const char *sep)
{

    static char const
	*p1,
	*p2,
	*c;
    static char
	*tok;
    int
	i, got_one;

    if (s!=NULL) {
	/* new string */
	p1=s;
	p2=NULL;
    }

    if (p1==NULL)
	return NULL;

    /* Traverse the string. */
    got_one=0;
    for (p2=p1; *p2!='\0'; ++p2) {
    /* Try each element of the separator array. */
	for (c=sep; *c!='\0'; ++c) {
	    if (*p2 == *c) { /* Got one */
		got_one=1;
		break;
	    }
	}
	if (got_one) /* No need to go any further */
	    break;
    }

    if (got_one) {
	/*  Copy the new token. */
	tok=realloc(tok,(size_t) (p2-p1+1));
	for (i=0; p1<p2; ++i, ++p1)
	    tok[i]=*p1;
	tok[i]='\0';
	p1=p2+1;
    } else {
	/* Didn't find a token.  Just copt the rest of the string. */
	tok=realloc(tok,(size_t) (strlen(p1)+1));
	strcpy(tok,p1);
	p1=NULL;
    }


    return tok;
    
}
