/************************************************************
 * IW.c
 ************************************************************
 *
 * DESCRIPTION
 *    Core utilities for the WebLib C utilities.
 *
 * AUTHOR
 *    Eryq (Erik Dorfman)
 *    Hughes STX Corporation
 *    30 Dec 1994
 *
 * NAMESPACE
 *    "IW_" is the Weblib C Utilities namespace.  All callable
 *     C code provided by WebLib is in this namespace.
 *
 * RETURN-VALUES
 *    All functions which return integer status return 0 for
 *    success and negative for failure, unless otherwise noted.
 *
 *    All functions which return pointers return NULL to 
 *    indicate failure, unless otherwise noted.
 */

/*
 * BEGIN TERMS AND CONDITIONS
 * 
 *  h 
 * 
 *                        NO WARRANTY
 * 
 * BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
 * FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT
 * WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER
 * PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND,
 * EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE
 * PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME
 * THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
 * 
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
 * REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH HOLDER
 * OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 * 
 * END TERMS AND CONDITIONS
 */

#ifndef lint
static char RCSID[] =
   "$Id: IW.c,v 1.5 1997/10/02 21:57:00 eryq Exp $";
#endif /* lint */

#include "IW.h"
#include "IW_Str.h"
#include "IW_Hasht.h"
#include "IW_New.h"
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>


/************************************************************
 *
 * Globals
 *
 ************************************************************/

/* Function to call on failure: */
void (*gDieHandler)();

/* Last error: */
int  IW_Errno;
char IW_Errstr[1024];



/* 
 * Setup variables.
 * See the system documentation for the gory details: 
 */
static IW_Hasht *IW_GSetup;


/************************************************************
 *
 * Library setup
 *
 ************************************************************/


/************************************************************
 * IW_ParseConfig
 ************************************************************
 *
 * DESCRIPTION
 *    Parse in the lib/conf/SETUP.dat file.
 *
 * NOTES
 *    We read the config at run-time, so that admins can change minor 
 *    aspects of the configuration without having to do a full system
 *    recompile.
 */ 

int IW_ParseConfig()
{
  FILE *fp;
  char path[256];
  char line[2048];
  char *pos;
  char varname[64];
  register int i;
  char *value;


  /* Create setup: */
  IW_GSetup = IW_HashtAllocCS(83, free);

  /* Open predigested SETUP file: */
  sprintf(path, "%s/lib/conf/SETUP.dat", IW_TOP_DIR);
  if (!(fp = fopen(path, "r"))) {
    return IW_Err(0, "couldn't open %s", path);
  }

  /* Slurp contents into hash: */
  while (fgets(line, 2048, fp)) {
    if ((line[0] == '#') || line[0] == '\n') {
      continue;   /* comment or empty line */
    }
    line[strlen(line)-1] = '\0';         /* remove terminating newline */

    /* Find and extract varname: */
    if (!(pos = strchr(line, ' '))) {
      continue;   /* syntax error; okay for now */
    }
    strncpy(varname, line, pos-line);
    varname[pos-line] = '\0';

    /* Duplicate value, and store: */
    value = IW_Strdup(pos+1);
    IW_HashtAtPut(IW_GSetup, varname, value);
  }
  close(fp);

  /* Add missing info: */
  if (!*IW_Var("TmpDir")) {
    sprintf(line, "%s/tmp", IW_Var("TopDir"));
    IW_SetVar("TmpDir", line);
  }

  return 0;
}

/************************************************************
 * IW_Var
 ************************************************************
 * Return a setup variable, or empty if missing (which should
 * suffice for most tests).
 */

char *IW_Var(char *name)
{
  char *s;
  if (!(s = IW_HashtAt(IW_GSetup, name))) s = "";
  return s;
} 

/************************************************************
 * IW_SetVar
 ************************************************************
 * Return a setup variable, or empty if missing (which should
 * suffice for most tests).
 */

char *IW_SetVar(char *name, char *value)
{
  value = value ? value : ""; 
  IW_HashtAtPut(IW_GSetup, name, IW_Strdup(value));
  return value;
}

/************************************************************
 * IW_Init
 ************************************************************
 *
 * DESCRIPTION
 *    Init the IW library.
 * 
 *    This sets up the environment (in case the IK routines need 
 *    to be called), and possibly will do other stuff.
 *
 * NOTES
 *    Note that we cannot depend on the environment being 
 *    set up, since we may be running as a CGI script under an
 *    HTTPD!
 */ 

int IW_Init()
{
  char path[2048];


  /* READ IN THE CONFIG FILE... */

  if (IW_ParseConfig() < 0) {
    return -1;
  } 


  /* SET UP THE NECESSARY ENVIRONMENT VARS... */

  /* Our basic environment vars: */
  (void)IW_Putenv("IW_TMP_DIR",  IW_TMP_DIR);
  (void)IW_Putenv("IW_LIB_DIR",  IW_LIB_DIR);
  (void)IW_Putenv("IW_DOCS_DIR", IW_DOCS_DIR);
  (void)IW_Putenv("IW_CGI_DIR",  IW_CGI_DIR);

  /* GAEA environment vars: */
  if (!(getenv("GAEADATA_DIR"))) {
    (void)IW_Putenv("GAEADATA_DIR", IW_GAEA_DATA_DIR);
  }
  if (!(getenv("SYSLOG_LEVEL"))) {
    (void)IW_Putenv("SYSLOG_LEVEL", "7");
  }
  if (!(getenv("IMS_STAFF"))) {
    (void)IW_Putenv("IMS_STAFF", "1");
  }

  /* Search path: */
  sprintf(path, "%s:%s", IW_Getenv("PATH"), IW_MORE_PATHS);
  (void)IW_Putenv("PATH", path);

  return 0;
}



/************************************************************
 *
 * Error handling
 *
 ************************************************************/


/************************************************************
 * IW_Err
 ************************************************************
 *
 * DESCRIPTION
 *    Make note of an error.  Returns -1 always.
 *
 * NOTES
 *    Use of error number is yet to be specified.  Use 0 for now.
 *
 *    Every call overwrites the information from the last call, so
 *    this really should just be used for lowest-level errors only:
 *    file not opened, etc.  If an IW routine fails, it is assumed
 *    that it called IW_Err, so don't make ANOTHER call to IW_Err
 *    in response.
 */ 

int IW_Err(int eno, char *format, ...)
{
  va_list args;

  /* Save the error number: */
  IW_Errno = eno;

  /* Format the error message: */
  va_start(args, format);
  vsprintf(IW_Errstr, format, args);
  va_end(args);
  return -1;
}


/************************************************************
 * IW_Die
 ************************************************************
 *
 * DESCRIPTION
 *    Die, calling the failure handler.
 *
 * ARGUMENTS
 *    If "eno" < 0, it indicates that further details already
 *    exist, in the IW_Errno/IW_Errstr.  Otherwise, these arguments
 *    are used to make the call to IW_Err() first.
 */ 

void IW_Die(int eno, char *format, ...)
{
  va_list args;
  char reason[1024];

  /* Format the string: */
  va_start(args, format);
  vsprintf(reason, format, args);
  va_end(args);

  /* Set errno, if needed: */
  if (eno >= 0) {
    IW_Err(eno, "%s", reason);
  }

  /* Output it: */
  if (gDieHandler) {
    (*gDieHandler)(reason);
  }
  else {
    fprintf(stderr, "IW_Die: %s\n", reason);
  }
  exit(-1);
}


/************************************************************
 * IW_SetDieHandler
 ************************************************************
 *
 * DESCRIPTION
 *    Set the failure-handling routine.  
 *
 *    This can do an exit if it wants; if it doesn't,
 *    the failure wrapper will do it instead.
 */ 

void IW_SetDieHandler(handler)
     void (*handler)(char *);
{
  gDieHandler = handler;
}




/************************************************************
 *
 * Basic stuff
 *
 ************************************************************/



/************************************************************
 * IW_Malloc
 ************************************************************
 *
 * DESCRIPTION
 *    Safe malloc.
 */ 

char *IW_Malloc(len)
     int len;
{
  char *p;

  if (!(p = (char *)malloc(len)))
    IW_Die(0, "malloc of %d bytes failed", len);
  return p;
}


/************************************************************
 * IW_Realloc
 ************************************************************
 *
 * DESCRIPTION
 *    Safe realloc.
 */ 

char *IW_Realloc(p, len)
     char *p;
     int len;
{
  if (!(p = (char *)realloc(p, len)))
    IW_Die(0, "realloc to %d bytes failed", len);
  return p;
}


/************************************************************
 * IW_Getenv
 ************************************************************
 *
 * DESCRIPTION
 *    Nicer getenv.  Returns empty string instead of
 *    NULL if variable is absent, which suffices for most
 *    tests.
 */ 

char *IW_Getenv(var)
     char *var;
{
  char *value;

  return ((value = getenv(var)) ? value : "");
}


/************************************************************
 * IW_Putenv
 ************************************************************
 *
 * DESCRIPTION
 *    Nicer putenv.  Formats and allocates string for you.
 *
 * RETURNS
 *    0 on success, negative on failure.
 *
 * NOTES
 *    The SunOS manpage for putenv() says that we need to malloc
 *    the environment var space.  The Linux page is vague.  So
 *    we play it safe, and malloc...
 */ 

int IW_Putenv(var, value)
     char *var;
     char *value;
{
  char s[2048];

  value = value ? value : "";      /* coerce to "" */
  if ((strlen(var) + 1 + strlen(value) + 1) >= 2048) {
    return IW_Err(0, "environment var/value too big!");
  }
  sprintf(s, "%s=%s", var, value);
  return putenv(IW_Strdup(s));
}

/******************** end of file "IW.c" ********************/
