/************************************************************
 * IW_Array.c                           -- variable-sized arrays
 ************************************************************
 *
 * DESCRIPTION
 *    Pseudoclass for variable-sized arrays.  See IW_Array.h for
 *    details.
 *
 * AUTHOR
 *    Eryq (Erik Dorfman)
 *    Hughes STX
 *    Mon, 21 Sep 1992
 */

/*
* BEGIN TERMS AND CONDITIONS
* 
*  h 
* 
*                        NO WARRANTY
* 
* BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
* FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT
* WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER
* PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND,
* EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
* PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE
* PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME
* THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
* 
* IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
* WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
* REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR
* DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL
* DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM
* (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
* INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF
* THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH HOLDER
* OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
* 
* END TERMS AND CONDITIONS
 */

#ifndef lint
static char RCSID[] =
"$Id: IW_Array.c,v 1.3 1996/04/17 21:16:12 imswww Exp $";
#endif /* lint */

/* Library headers: */
#include "IW_Array.h"       /* Pseudoclass header */
#include "IW_New.h"        /* memory allocation */

/* System headers: */
#include <stdio.h>
#include <stdlib.h>


/************************************************************
 *
 * Globals, private
 *
 ************************************************************/

/* External element comparison function: */
static int (*gElemCompareFunc)();



/************************************************************
 *
 * Private utilities
 *
 ************************************************************/


/************************************************************
 * IW_ArrayDiscardAt (private)
 ************************************************************
 *
 * DESCRIPTION
 *    Given a IW_Array and an index (zero based), discard the
 *    element stored at that index via the IW_Array's discard
 *    function.  This sets the element to NULL, for good measure.
 *
 * NOTES
 *    No bounds checking is done.
 */

static void IW_ArrayDiscardAt(this, i)
     IW_Array *this;     /* the array object */
     int i;
{
  IW_Elem elem;

  if ((this->discard_f) && ((elem = this->elems[i]) != NULL)) {
    (*(this->discard_f))(elem);      /* discard it */
    this->elems[i] = NULL;    /* erase it */
  }
}


/************************************************************
 * _IW_ArrayNewBufSize (private)
 ************************************************************
 *
 * DESCRIPTION
 *    Given a new size for the ARRAY, and a current buffer
 *    size, compute a new buffer size which is just large enough
 *    to hold the new number of elements, and which is arrived
 *    at by repeated doubling (or halving) of the current buffer
 *    size.
 *
 * ARGUMENTS
 *    'oldA' is the original array size (number of elements).
 *    'oldB' is the original buffer size.
 *    'newA' is the new array size (number of elements) which the
 *        new buffer size must be adjusted to.
 *
 * RETURNS
 *    The new buffer size, silently limited to a minimum of 2.
 *
 * NOTES
 *    Seems weird, you say.  Why not just make the buffer the exact
 *    same size as the new number if elements when resizing the array?
 *
 *    The general idea is to protect users from serious performance
 *    problems when they do very silly things, like growing/shrinking
 *    arrays by one element at a time, over and over... if the buffer
 *    size were to change *exactly* to the new size on a grow or shrink,
 *    then 'n' resizes could easily take time O(n^2) (assuming 'n'
 *    reallocs).  In this scheme, the buffer size changes in quantum
 *    leaps: from a given size the buffer may only:
 *
 *        * Double in size repeatedly until it reaches a size which
 *          is big enough to hold the new number of elements, when the
 *          size of the array grows.  Or...
 *
 *        * Halve in size repeatedly until the next size down would be
 *          too small to hold the new number of elements, when the
 *          size of the array shrinks.
 *
 *    It can be seen by looking at the code, that:
 *
 *        * If the new number of elements is greater than the old
 *          number of elements, then the new buffer size will
 *          either be the same as the old, or it will be larger.
 *
 *        * If the new number of elements is less than the old
 *          number of elements, then the new buffer size will
 *          either be the same as the old, or it will be smaller.
 *
 *        * If the number of elements iwll remain the same, so will
 *          the buffer size.
 */

static int _IW_ArrayNewBufSize(oldA, oldB, newA)
     int oldA;
     int oldB;
     int newA;
{
  register int newB = oldB;

  if (newA > oldA) {                  /* GROW! */
    while (newB < newA) {
      newB <<= 1;
    }
  }
  else if (newA < oldA) {             /* SHRINK! */
    while (newB && ((newB >> 1) >= newA)) {
      newB >>= 1;
    }
  }
  return ((newB < 2) ? 2 : newB);
}


/************************************************************
 * _IW_ArrayGrow (private)
 ************************************************************
 *
 * DESCRIPTION
 *    Given a IW_Array and a new number of elements which is
 *    STRICTLY GREATER THAN the current number of elements,
 *    grow the IW_Array.  Pad all new slots with the NULL
 *    element.
 *
 *    If the request cannot be honored, the IW_Array is left in
 *    the original state and an error code is returned.
 *
 * ARGUMENTS:
 *    'this' is a pointer to the IW_Array.
 *    'newnelems' is the new number of elements, which MUST be strictly
 *        greater than the existing number of elements... no check will
 *        be made in this routine!
 *
 * RETURNS
 *    IW_ArrayOK on success, <IW_ArrayOK on error.
 *
 * NOTES
 *    See _IW_ArrayNewBufSize() for details.
 */

static int _IW_ArrayGrow(this, newnelems)
     IW_Array *this;     /* the array object */
     int newnelems;
{
  int newbufsize;
  IW_Elem *newelems;
  register int i;


  /* Get new buffer size: */
  newbufsize = _IW_ArrayNewBufSize(this->numelems, this->bufsize, newnelems);

  /* Determine whether the buffer must grow or not, and handle
     accordingly... */
  if (newbufsize != this->bufsize) {         /* GROW BUFFER: */

    /* Reallocate buffer: */
    newelems = IW_RenewArray(this->elems, IW_Elem, newbufsize);
    if (!newelems) {
      IW_Err(0, "couldn't reallocate buffer");
      return IW_ArrayESYS;
    }

    /* Install the new buffer! */
    this->elems = newelems;
    this->bufsize = newbufsize;
  }

  /* At this point, 'newelems' holds a buffer of size 'newbufsize' which
     holds all the elements of the original buffer (and may even be
     the original buffer, although if it is not, the original buffer
     has already be freed). */

  /* Adjust the number of ELEMENTS in the array: */
  for (i = this->numelems; i < newnelems; i++) {
    this->elems[i] = NULL;                 /* pad */
  }
  this->numelems = newnelems;

  /* Done! */
  return IW_ArrayOK;
}


/************************************************************
 * _IW_ArrayShrink (private)
 ************************************************************
 *
 * DESCRIPTION
 *    Given a IW_Array and a new number of elements which is
 *    STRICTLY LESS THAN the current number of elements,
 *    shrink the IW_Array.  Destroy the elements in slots from
 *    'oldnelems-1' down through 'newnelems'.
 *
 *    If the request cannot be honored, the IW_Array is left in
 *    the original state and an error code is returned.
 *
 * ARGUMENTS:
 *    'this' is a pointer to the IW_Array.
 *    'newnelems' is the new number of elements, which MUST be strictly
 *        less than the existing number of elements... no check will
 *        be made in this routine!
 *
 * RETURNS
 *    IW_ArrayOK on success, <IW_ArrayOK on error.
 *
 * NOTES
 *    See _IW_ArrayNewBufSize() for details.
 */

static int _IW_ArrayShrink(this, newnelems)
     IW_Array *this;     /* the array object */
     int newnelems;
{
  int newbufsize;
  IW_Elem *newelems;
  register int i;
  IW_Elem elem;


  /* Get new buffer size: */
  newbufsize = _IW_ArrayNewBufSize(this->numelems, this->bufsize, newnelems);

  /* Determine if the buffer must shrink or not, and handle accordingly...
     note that, whether or not the BUFFER grows or shrinks, some
     discarding will be necessary since the ARRAY is shrinking: */
  if (newbufsize < this->bufsize) {         /* SHRINK BUFFER: */

    /* Allocate smaller buffer: */
    newelems = IW_NewArray(IW_Elem, newbufsize);
    if (!newelems) {
      IW_Err(0, "couldn't allocate buffer");
      return IW_ArrayESYS;
    }

    /* Copy those elements to new buffer which will not be discarded: */
    for (i = 0; i < newnelems; i++) {
      newelems[i] = (this->elems)[i];
    }

    /* Discard those elements in old buffer which are going away: */
    if (this->discard_f) {                          /* there's a discard fn */
      for (i = newnelems; i < this->numelems; i++) {
	if (elem = this->elems[i]) {                   /* elem is not null */
	  (*(this->discard_f))(elem);                     /* chuck it! */
	}
      }
    }

    /* Free up the old buffer: */
    IW_Free(this->elems);

    /* Set the new values: */
    this->elems = newelems;
    this->bufsize = newbufsize;
    this->numelems = newnelems;
  }

  else {                                   /* KEEP BUFFER */

    /* Discard those elements in old buffer which are going away: */
    if (this->discard_f) {                          /* there's a discard fn */
      for (i = newnelems; i < this->numelems; i++) {
	if (elem = this->elems[i]) {                   /* elem is not null */
	  (*(this->discard_f))(elem);                     /* chuck it! */
	}
      }
    }

    /* Set the new values: */
    this->numelems = newnelems;
  }

  /* Done! */
  return IW_ArrayOK;
}



/************************************************************
 *
 * Creation/destruction
 *
 ************************************************************/


/************************************************************
 * IW_ArrayCreate                                 -- constructor
 ************************************************************
 *
 * DESCRIPTION
 *    Create an array with the supplied buffer size, but with
 *    no elements currently stored in it (i.e., IW_ArrayGetSize() will
 *    return 0).
 *
 * ARGUMENTS
 *    The 'this' is a pointer to an uninitialized IW_Array.
 *
 *    The 'bufsize' is the size of the IW_Array's internal buffer.
 *    You should (but need not) make this large enough to hold
 *    all the elements that you will add to the IW_Array.  The
 *    value is silently limited to a minimum of 2.
 *
 *    The 'discard_f' is a function (possibly NULL) which will be called
 *    whenever one of the elements stored in the array is overwritten
 *    or destroyed (including when the array as a whole is destroyed).
 *    This function should take one argument: an object which is stored
 *    in the array, which will typically be a pointer to a malloc'ed
 *    data structure.  The function should perform the necessary deallocation
 *    of the data structure (e.g., call free() to reclaim the malloc'ed
 *    memory).
 *
 * RETURNS
 *    IW_ArrayOK on success, <IW_ArrayOK on failure.
 *
 * WARNINGS
 *    Obsolete.  Use IW_ArrayAlloc() instead.
 */

int IW_ArrayCreate(this, bufsize, discard_f)
     IW_Array *this;             /* space in which to create the object */
     int bufsize;               /* initial size of internal buffer */
     IW_FnDiscard discard_f;     /* the function which frees an element */
{
  int realbufsize = ((bufsize < 2) ? 2 : bufsize);

  /* Allocate elements: */
  if (!(this->elems = IW_NewArray(IW_Elem, realbufsize))) {
    IW_Err(0, "couldn't allocate buffer");
    return IW_ArrayESYS;
  }

  /* Other initialization: */
  this->bufsize = realbufsize;
  this->numelems = 0;
  this->discard_f = discard_f;
  return IW_ArrayOK;
}


/************************************************************
 * IW_ArrayCreateSize                             -- constructor
 ************************************************************
 *
 * DESCRIPTION
 *    See IW_ArrayCreate().  This constructor actually sets the
 *    size of the array instantly to the buffer size, and
 *    puts NULL in all the slots.
 *
 * RETURNS
 *    IW_ArrayOK on success, <IW_ArrayOK on failure.
 *
 * WARNINGS
 *    Obsolete.  Use IW_ArrayAllocSize() instead.
 */

int IW_ArrayCreateSize(this, bufsize, discard_f)
     IW_Array *this;          /* space in which to create the object */
     int bufsize;            /* initial size of internal buffer */
     IW_FnDiscard discard_f;  /* the function which frees an element */
{
  int result;

  if ((result = IW_ArrayCreate(this, bufsize, discard_f)) < IW_ArrayOK) {
    return result;
  }
  if ((result = IW_ArraySetSize(this, bufsize) < IW_ArrayOK)) {
    return result;
  }
  return IW_ArrayOK;
}


/************************************************************
 * IW_ArrayDestroy                                 -- destructor
 ************************************************************
 *
 * DESCRIPTION
 *    Destructor for IW_Array.  The calls the discard function
 *    for every non-NULL element in the IW_Array (if there is
 *    a discard function), and then frees the allocated array.
 *
 * ARGUMENTS
 *    'this' is a pointer to the IW_Array to be destroyed.
 */

void IW_ArrayDestroy(this)
     IW_Array *this;     /* the array object */
{
  register int i;
  IW_Elem elem;

  /* If array has a discard function, then discard the elements: */
  if (this->discard_f) {
    for (i = 0; i < this->numelems; i++) {
      if ((elem = this->elems[i]) != NULL) {
	(*(this->discard_f))(elem);      /* discard it */
	this->elems[i] = NULL;    /* erase it, for good measure */
      }
    }
  }

  /* Free the array: */
  IW_Free(this->elems);
}


/************************************************************
 * IW_ArrayAlloc                           -- allocate and construct
 ************************************************************
 *
 * DESCRIPTION
 *    Allocate space for a IW_Array, then create it using
 *    IW_ArrayCreate(), and return it.
 *
 * ARGUMENTS
 *    Identical to IW_ArrayCreate(), without the 'this' pointer.
 *
 * RETURNS
 *    A new IW_Array on success, and NULL on error.
 */

IW_Array *IW_ArrayAlloc(bufsize, discard_f)
     int bufsize;               /* initial size of internal buffer */
     IW_FnDiscard discard_f;     /* the function which frees an element */
{
  IW_Array *new;

  if (!(new = IW_New(IW_Array))) {    /* allocate array */
    return (IW_Array *)NULL;
  }
  if (IW_ArrayCreate(new, bufsize, discard_f) < IW_ArrayOK) {  /* create */
    IW_Free(new);
    return (IW_Array *)NULL;
  }
  return new;              /* done! */
}


/************************************************************
 * IW_ArrayAllocSize                    -- allocate and construct
 ************************************************************
 *
 * DESCRIPTION
 *    Allocate space for a IW_Array, then create it using
 *    IW_ArrayCreateSize(), and return it.
 *
 * ARGUMENTS
 *    Identical to IW_ArraySize(), without the 'this' pointer.
 *
 * RETURNS
 *    A new IW_Array on success, and NULL on error.
 */

IW_Array *IW_ArrayAllocSize(size, discard_f)
     int size;                  /* the size of the array to create */
     IW_FnDiscard discard_f;     /* the function which frees an element */
{
  IW_Array *new;

  if (!(new = IW_New(IW_Array))) {    /* allocate array */
    return (IW_Array *)NULL;
  }
  if (IW_ArrayCreateSize(new, size, discard_f) < IW_ArrayOK) {  /* create */
    IW_Free(new);
    return (IW_Array *)NULL;
  }
  return new;              /* done! */
}


/************************************************************
 * IW_ArrayFree                              -- destroy and free
 ************************************************************
 *
 * DESCRIPTION
 *    Given a pointer to a malloc'ed IW_Array, destroy the array
 *    and then free the pointer.
 *
 * ARGUMENTS
 *    'this' is a poiner to a malloc'ed IW_Array.
 *
 * RETURNS
 *    IW_ArrayOK always.
 */

int IW_ArrayFree(this)
     IW_Array *this;     /* the array object */
{
  if (this) {
    (void)IW_ArrayDestroy(this);
    IW_Free(this);
  }
  return IW_ArrayOK;
}



/************************************************************
 *
 * Setting/getting attributes
 *
 ************************************************************/


/************************************************************
 * IW_ArraySetSize                       -- change size of array
 ************************************************************
 *
 * DESCRIPTION
 *    Change the size of the array.
 *
 *    If increasing the size, new elements are filled with NULL
 *    automatically: this is so a call to IW_ArrayDestroy() doesn't puke
 *    while trying to call the discard function on garbage values
 *    in the array.
 *
 *    If decreasing the size, elements past the new horizon are freed
 *    using the discard_f.
 *
 * NOTES
 *    Uses _IW_ArrayGrow() or _IW_ArrayShrink() as appropriate.  If a
 *    resizing request cannot be honored, an error is returned and the
 *    array is left in the same state as before entry into the function.
 *
 * RETURNS
 *    IW_ArrayOK on success, <IW_ArrayOK on failure.  Returns IW_ArrayESYS
 *    if a "malloc" routine fails.
 */

int IW_ArraySetSize(this, newnelems)
     IW_Array *this;     /* the array object */
     int newnelems;     /* the new size */
{
  /* Check for negative size: */
  if (newnelems < 0) {
    IW_Err(0, "negative array size");
    return IW_ArrayEBOUNDS;
  }

  /* Resize appropriately: */
  if (newnelems < this->numelems) {
    return _IW_ArrayShrink(this, newnelems);
  }
  else if (newnelems > this->numelems) {
    return _IW_ArrayGrow(this, newnelems);
  }
  else {
    return IW_ArrayOK;
  }
}


/************************************************************
 * IW_ArrayGetSize                     -- get number of elements
 ************************************************************
 *
 * DESCRIPTION
 *    Get the size of the IW_Array.  Note that this is not the
 *    same as the internal buffer size!  It is the number of
 *    elements currently "stored" in the IW_Array.
 *
 * ARGUMENTS
 *    The 'this' is a pointer to a IW_Array.
 *
 *    The 'sizep' is a pointer to an integer which will be set
 *    to the size of the IW_Array on exit.
 *
 * RETURNS
 *    IW_ArrayOK always.
 *
 * NOTES
 *    Try IW_ArraySize() for convenience.
 */

int IW_ArrayGetSize(this, sizep)
     IW_Array *this;     /* the array object */
     int *sizep;        /* where to put the size */
{
  *sizep = this->numelems;
  return IW_ArrayOK;
}


/************************************************************
 * IW_ArrayElems -- get underlying array of elements
 ************************************************************
 *
 * DESCRIPTION
 *    Return the underlying C array of elements in the IW_Array,
 *    or NULL on failure.
 *
 * WARNINGS
 *    The pointer returned by this function is invalidated
 *    by any operations which add elements to the array
 *    (like IW_ArrayAdd) or which resize it (like IW_ArraySetSize).
 *    You must call this function fresh after each such operation.
 *
 *    Try and avoid using this if you can... this kind
 *    of "popping the hood" is provided only so that you
 *    can send the IW_Array (or part of it, anyway) into
 *    routines which demand a C array (e.g., qsort, bsearch).
 *
 * RETURNS
 *    A pointer to the underlying array.
 */ 

IW_Elem *IW_ArrayElems(this)
     IW_Array *this;        /* the array object */
{
  return (this ? this->elems : NULL);
}


/************************************************************
 *
 * Storing/retrieving elements
 *
 ************************************************************/


/************************************************************
 * IW_ArrayAdd                    -- add element to end of array
 ************************************************************
 *
 * DESCRIPTION
 *    Add (a pointer to) an element into the array, at the end.
 *    You must allocate the space for the element, if there is danger
 *    of it disappearing!
 *
 * NOTES
 *    If more buffer space is needed, this expands the buffer size
 *    to double its present size.  Perhaps this should expand it to
 *    a power of 2, like ArraySetSize(), or perhaps I should just
 *    use ArraySetSize(); I don't know...
 *
 * RETURNS
 *    IW_ArrayOK on success, <IW_ArrayOK on failure.
 */

int IW_ArrayAdd(this, newelem)
     IW_Array *this;     /* the array object */
     IW_Elem newelem;    /* the element to add */
{
  int newbufsize;
  IW_Elem *newelems;


  /* Expand buffer space (by a factor of two), if necessary: */
  if (this->numelems == this->bufsize) {
    newbufsize = this->bufsize << 1;
    if (!(newelems = IW_RenewArray(this->elems, IW_Elem, newbufsize))) {
      IW_Err(0, "couldn't allocate buffer");
      return IW_ArrayESYS;
    }
    this->bufsize = newbufsize;
    this->elems = newelems;
  }

  /* We now have enough buffer space... copy element: */
  this->elems[this->numelems] = newelem;
  ++(this->numelems);

  /* Ok! */
  return IW_ArrayOK;
}


/************************************************************
 * IW_ArrayAtGet                           -- retrieve element at index
 ************************************************************
 *
 * DESCRIPTION
 *    Get the element of the array indexed by 'i', and set '*elemp'
 *    to that element.
 *
 * NOTES
 *    IW_ArrayAt() can be more convenient; however, there is no way
 *    to distinguish between a NULL element returned, and a NULL
 *    returned because of an error.  That's why IW_ArrayAtGet() is
 *    provided.
 *
 * RETURNS
 *    IW_ArrayOK on success, IW_ArrayEBOUNDS on a bounds error.  In the
 *    event of a failure, '*elemp' is set to NULL.
 */

int IW_ArrayAtGet(this, i, elemp)
     IW_Array *this;     /* the array object */
     int i;             /* the index into the array */
     IW_Elem *elemp;     /* where to put the element at that index */
{
  if ((i < 0) || (i >= this->numelems)) {
    *elemp =  NULL;
    IW_Err(0, "index out of array bounds");
    return IW_ArrayEBOUNDS;
  }
  else {
    *elemp = this->elems[i];
    return IW_ArrayOK;
  }
}


/************************************************************
 * IW_ArrayAt                           -- retrieve element at index
 ************************************************************
 *
 * DESCRIPTION
 *    Return the element of the array indexed by 'i', or NULL
 *    in the event of an error.
 *
 * NOTES
 *    There is no way to distinguish between the NULL return values
 *    which indicates a NULL element, and a NULL returned because of
 *    an error.  If you need to do so, use IW_ArrayAtGet().
 *
 * RETURNS
 *    The element on success, and NULL on failure.
 */

IW_Elem IW_ArrayAt(this, i)
     IW_Array *this;     /* the array object */
     int i;             /* the zero-based index */
{
  if ((i < 0) || (i >= this->numelems))
    return NULL;
  else
    return this->elems[i];
}


/************************************************************
 * IW_ArrayAtPut                         -- store element at index
 ************************************************************
 *
 * DESCRIPTION
 *    Store the supplied element at the given index of the array.
 *    if there is already an elem stored at that index (i.e., the
 *    slot is not NULL), then this function calls the discard function
 *    to get rid of the existing elem.
 *
 * RETURNS
 *    IW_ArrayOK on success, <IW_ArrayOK on failure.
 */

int IW_ArrayAtPut(this, i, elem)
     IW_Array *this;     /* the array object */
     int i;             /* the index into the array */
     IW_Elem elem;       /* the element to store at that index */
{
  /* If out of bounds, return immediately: */
  if ((i < 0) || (i >= this->numelems)) {      /* check bounds... */
    IW_Err(0, "index out of array bounds");
    return IW_ArrayEBOUNDS;
  }

  /* If slot presently occupied, free contents: */
  if (this->elems[i] != NULL) {
    IW_ArrayDiscardAt(this, i);                   /* free old */
  }

  /* Insert! */
  this->elems[i] = elem;                   /* insert new */
  return IW_ArrayOK;
}


/************************************************************
 * IW_ArrayLocAt                -- get address of array element (obsolete)
 ************************************************************
 *
 * DESCRIPTION
 *    Return the l-value of the indexed element of the array.
 *    Returns NULL on out-of-bounds index.
 *
 * WARNINGS
 *    Obsolete.  Use IW_ArrayElems() to get at the underlying C 
 *    array of elements.
 *
 *    Danger!  Like all pointers into black boxes, the return
 *    value must be used with care.  Don't put garbage in it,
 *    or you may cause the IW_Array to break.  If you discard
 *    the element that's stored there, be sure to also set the
 *    element to NULL.
 *
 *    In short, you probably don't need this function, so you
 *    probably should avoid it.
 *
 * RETURNS
 *    A pointer to an element in the underlying array.
 */

IW_Elem *IW_ArrayLocAt(this, i)
     IW_Array *this;     /* the array object */
     int i;             /* an index into the array */
{
  if ((i < 0) || (i >= this->numelems))
    return (IW_Elem *)NULL;
  else
    return &(this->elems[i]);
}



/************************************************************
 *
 * Iterator functions
 *
 ************************************************************
 *
 * All container Pseudoclasses try to support container-specific
 * "iterator" types which have the same general syntax and usage.
 * For example, if 'IW_Bag' is a container Pseudoclass type and you
 * wanted to print out the elements of 'mybag', you would do so
 * as follows:
 *
 *|   1.  Declare 'bi' as a IW_BagIter.
 *|   2.  Call IW_BagIRewind() to rewind 'bi' to the beginning of 'mybag'.
 *|   3.  Keep calling IW_BagINext() to get the next element out of the
 *|       bag... it will return:
 *|         - IW_BagOGO if an element was actually retrieved
 *|         - IW_BagOSTOP if no element was retrieved because the end of
 *|           the collection was reached.
 *|         - values less than IW_BagOK on error.
 *
 * Repeated calls to IW_BagINext() are okay when OSTOP is returned;
 * the subsequent calls will just return OSTOP (much like getting
 * characters from a stream which is at "end of file").
 *
 * The above example describes a "detached iterator", in that the iterator
 * variable is separate from the container it is iterating over.
 * In the case of a IW_Array, the "detached iterator" code looks like:
 *
 *|   IW_Array      A;             \* an array of objects *\
 *|   IW_ArrayIter  ai;            \* an iterator *\
 *|   int          index;         \* index into array *\
 *|   Object      *elem;          \* element stored in array *\
 *|   ...
 *|   IW_ArrayIRewind(&A, &ai);        \* rewind to begin *\
 *|   while (IW_ArrayINext(&A, &ai, &index, &elem) == IW_ArrayGO) {
 *|       \* do stuff with index and/or elem *\
 *|   }
 *
 * Some container Pseudoclasses may support "attached iterators",
 * where every container has its own internal iterator that you can
 * use (much like a FILE's internal file pointer in stdio routines).
 * The notation is simpler, and does not require the declaration of
 * an additional varible.  In the case of a IW_Array, the "attached
 * iterator" code looks like:
 *
 *|   IW_Array      A;             \* an array of objects *\
 *|   int          index;         \* index into array *\
 *|   Object      *elem;          \* element stored in array *\
 *|   ...
 *|   IW_ArrayRewind(&A);        \* rewind to begin *\
 *|   while (IW_ArrayNext(&A, &index, &elem) == IW_ArrayGO) {
 *|       \* do stuff with index and/or elem *\
 *|   }
 *
 */


/************************************************************
 * IW_ArrayIRewind                                  -- iteration
 ************************************************************
 *
 * DESCRIPTION
 *    Rewind a IW_ArrayIter to the beginning of the array.
 *
 * ARGUMENTS
 *    this -- a pointer to a IW_Array.
 *
 *    iter -- a pointer to the iterator to be rewound.
 *
 * RETURNS
 *    IW_ArrayOK always.
 *
 * ARGSUSED
 */

int IW_ArrayIRewind(this, iter)
     IW_Array *this;      /* the array object */
     IW_ArrayIter *iter;  /* the black-box "iterator" variable */
{
  *iter = 0;
  return IW_ArrayOK;
}


/************************************************************
 * IW_ArrayINext                                    -- iteration
 ************************************************************
 *
 * DESCRIPTION
 *    Given a IW_ArrayIter which points into a IW_Array, set *elemp to
 *    point to the element, and advance the iterator to the next element.
 *
 * ARGUMENTS
 *    'this' is a pointer to a IW_Array.
 *
 *    'iter' is a pointer to the iterator which indexes into the array.
 *
 *    'indexp' is a pointer to an intIW_r which, on successful exit, will
 *     be set to index [0..] into the array of the 'elem' returned, and
 *     which on failure will be set to -1.  If you are not interested
 *     in this information, you may supply a NULL 'indexp'.
 *
 *    'elemp' is a pointer to a IW_Elem  which, on successful exit, will
 *     be set to the next array element to be retrieved, and which on
 *     failure will be set to NULL.  If you are not interested
 *     in this information, you may supply a NULL 'elemp'.
 *
 * RETURNS
 *    IW_ArrayOGO if *elemp was set successfully.
 *
 *    IW_ArrayOSTOP if the end of the array was reached, and sets *elemp
 *    to NULL.  Successive calls to this function will have the same 
 *    effect.
 */

int IW_ArrayINext(this, iter, indexp, elemp)
     IW_Array *this;      /* the array object */
     IW_ArrayIter *iter;  /* the black-box "iterator" variable */
     int *indexp;        /* where to put the index of the current element */
     IW_Elem *elemp;      /* where to put the current element */
{
  if (*iter < IW_ArraySize(this)) {
    if (indexp)                              /* set index */
      *indexp = *iter;
    if (elemp)                               /* set element */
      *elemp = IW_ArrayAt(this, *iter);
    ++(*iter);
    return IW_ArrayOGO;
  }
  else {
    if (indexp)                              /* set index to bogus value */
      *indexp = -1;
    if (elemp)                               /* set element to bogus value */
      *elemp = NULL;
    return IW_ArrayOSTOP;
  }
}



/************************************************************
 *
 * Additional stuff:
 *
 ************************************************************/


/************************************************************
 * IW_ArrayDo                           -- iteration over array
 ************************************************************
 *
 * DESCRIPTION
 *    Perform a function on each element.  The function should be
 *    declared as:
 *
 *        int action(IW_Elem elem, int index, IW_Arg arg);
 *
 *    This action will be invoked for every element in the IW_Array;
 *    when it is called, 'elem' will be the current element, 'index'
 *    will be the current index (of that element), and 'arg' will
 *    be the same 'arg' passed into the call to IW_ArrayDo().
 *
 *    The action should return 0; any other action will cause the
 *    iteration to be halted, and IW_ArrayEHALTED to be returned.
 *
 * ARGUMENTS
 *    The 'this' is a pointer to a IW_Array.
 *
 *    The 'action' is a function, as described above.
 *
 *    The 'arg' is a "callback argument", which is sent on to each
 *    invocation of 'action'.  This allows you to get extra data
 *    from the outside world into the action without having to
 *    use global variables.
 *
 * RETURNS
 *    IW_ArrayOK if iterated through all elements, IW_ArrayEHALTED
 *    if halted.
 */

int IW_ArrayDo(this, action, arg)
     IW_Array *this;   /* the array object */
     int (*action)(int, IW_Elem, IW_Arg);   /* what to do for each element */
     IW_Arg arg;       /* the callback argument */
{
  register int i;
  int halted = 0;

  for (i = 0; ((i < this->numelems) && !halted); i++) {
    halted |= (*action)(i, this->elems[i], arg);
  }
  if (halted) {
    IW_Err(0, "halted by callback");
    return IW_ArrayEHALTED;
  }
  return IW_ArrayOK;
}


/************************************************************
 * IW_ArrayQsort -- quicksort array
 ************************************************************
 *
 * DESCRIPTION
 *    Do a quicksort on the array's elements.
 *
 * ARGUMENTS
 *    The 'this' is a pointer to a IW_Array.
 *
 *    The 'compare_f' is a function which compares two elements
 *    stored in the IW_Array, THROUGH POINTERS TO THOSE ELEMENTS.  
 *    It should be declared as:
 *
 *|       int mycompare(IW_Elem *ap, IW_Elem *bp)
 *
 *    It should return less than, equal to, or greater than zero if
 *    the first element is (respectively) less than, equal to, or
 *    greater than the second element.
 *
 * WARNINGS
 *    Note that the comparison function takes two POINTERS to array
 *    elements, which are themselves usually pointers to structures.
 */

void IW_ArrayQsort(this, compare_f)
     IW_Array *this;        /* the array object */
     int (*compare_f)();   /* comparison function */
{
  qsort((IW_Elem)(this->elems),      /* base */
	this->numelems,             /* number of elements */
	sizeof(IW_Elem),             /* width */
	compare_f);                 /* comparison function */
}


/************************************************************
 * IW_ArrayQsortCompare (private) 
 ************************************************************
 *
 * DESCRIPTION
 *    Private comparison function for sorting.  
 *    Used by IW_ArrayQsortElems.
 */ 

static int IW_ArrayQsortCompare(ap, bp)
     const void *ap;
     const void *bp;
{
  return (*gElemCompareFunc)(*((IW_Elem *)ap), *((IW_Elem *)bp));
}


/************************************************************
 * IW_ArrayQsortElems                        -- quicksort array
 ************************************************************
 *
 * DESCRIPTION
 *    Do a quicksort on the array's elements.
 *
 * ARGUMENTS
 *    The 'this' is a pointer to a IW_Array.
 *
 *    The 'compare_f' is a function which compares two elements
 *    stored in the IW_Array.  It should be declared as:
 *
 *|       int mycompare(IW_Elem a, IW_Elem b)
 *
 *    It should return less than, equal to, or greater than zero if
 *    the first element is (respectively) less than, equal to, or
 *    greater than the second element.
 */

void IW_ArrayQsortElems(this, compare_f)
     IW_Array *this;        /* the array object */
     int (*compare_f)();   /* comparison function */
{
  gElemCompareFunc = compare_f;
  qsort((IW_Elem)(this->elems),      /* base */
	this->numelems,             /* number of elements */
	sizeof(IW_Elem),             /* width */
	IW_ArrayQsortCompare);       /* comparison function */
}

/******************** end of file "IW_Array.c" ********************/
