/************************************************************
 * IW_Array.h
 ************************************************************
 *
 * DESCRIPTION
 *    "Container" Pseudoclass for variable-sized arrays. 
 * 
 *    Like all container Pseudoclasses, IW_Arrays can contain "elements"
 *    which are the same size as a pointer; thus, the elements
 *    that you will store will most likely be POINTERS to other objects.
 *
 *    You can create a IW_Array that is initially of size 0 and add
 *    elements to the end of it (increasing its size), or you can create 
 *    one which has a bunch of initial slots.  After creation, you can:
 *
 *       - Add elements to the end of the IW_Array.
 *       - Retrieve the element stored at a given (slot) index.
 *       - Store an element at an existing (slot) index, possibly
 *         discarding whatever element is currently stored there.
 *       - Sort the elements, iterate through the elements, etc.
 *
 *    Like all container Pseudoclasses, you can specify a "discard" 
 *    function for a particular IW_Array which will be called whenever
 *    an element stored in the IW_Array is overwritten, or when the
 *    IW_Array is destroyed.  Typically, this is a function to gracefully
 *    free any malloc'ed entities which are pointed to by elements of
 *    the IW_Array, so you don't get accidental memory leaks.
 *
 * USAGE
 *    When you create a new IW_Array through the Create() method, the
 *    IW_Array has no elements in it, and so has a size of 0.  However,
 *    you may specify the internal BUFFER size, in anticipation of elements
 *    to be added.  
 * 
 *    When you create a new IW_Array through the CreateSize() method, 
 *    the IW_Array has the given number of elements (all initially NULL),
 *    and a buffer of exactly that size.
 *
 *    With either method of creation, you may specify a discard function
 *    for gracefully destroying elements stored in the IW_Array.  This
 *    function is of type IW_eDiscardFunc.  It is never called on a NULL
 *    element, so you don't need to make that check within the function.
 *
 *    Indexing of the IW_Array is zero-based, just as it is with normal
 *    arrays.
 *
 *    Remember: usually if you wish to "store" an object in a IW_Array, you
 *    must malloc() space for the object and store a *pointer* to the
 *    malloc'ed space (of course, you need not do this if the object is the
 *    size of a pointer or smaller).  Actually, *any* pointers into stable
 *    space will do; they don't have to come from calls to malloc().
 *
 * IMPLEMENTATION
 *    Internally, a IW_Array keeps all of the pointers in an actual
 *    array.  This means that useful routines like qsort() (and, ultimately,
 *    bsearch() are provided and work just as efficiently.  Also, 
 *    IW_Arrays are more compact than linked lists, and do not require
 *    the overhead of an invisible malloc() for each object to hold
 *    the list node.  However, IW_Arrays seem to be best employed when you know
 *    the approximate number of elements that you will need to store; e.g.,
 *    if you need a container which will hold 20 plus-or-minus 10 elements.
 *    However, if you are anticipating that a given container might need 
 *    to hold anywhere from 2 to 200000 elements, you are better off using a 
 *    IW_List.
 *
 * ANALYSIS
 *    Arrays are created with an initial buffer, which is silently
 *    realloc'ed when the array grows beyond it.  Every such realloc doubles
 *    the size of the buffer, so adding n elements to an array with an
 *    initial buffer size of 1 only results in log(n) realloc's.  Although
 *    realloc is not a primitive operation, we estimate that each realloc
 *    takes the following time:
 *    
 *        realloc(i) -->
 *             To malloc/realloc size i/2 buffer to size i:  i
 *             To copy i/2 elements from old to new buffer:  i/2
 *   
 *    So, assuming n = 2^k, we have the following sequence:
 *    
 *        realloc(2)  -->  2+1 time steps
 *        realloc(4)  -->  4+2 time steps
 *        realloc(8)  -->  8+4 time steps
 *        realloc(16) -->  16+8 time steps
 *        ...
 *        realloc(n)  -->  n+(n/2) time steps
 *   
 *    This means that the total time spent doing realloc is about:
 *         
 *             .- (log n/2)
 *            /
 *        3 * | 2^x dx           or REALLY (3n - 3), which is O(n).
 *            /
 *        0 -'
 *   
 *    (Yes, that's meant to be an integral sign).  So, even starting with a 
 *    1-item buffer, n inserts take time O(n).
 *   
 *    As you would expect from an array, we have the following time 
 *    complexities for the more popular operations:
 *   
 *        indexed retrieval:  O(1)
 *        indexed storage:    O(1)
 *        sorting:            O(n log n)  (uses qsort)                
 *
 * NAMESPACE
 *    Constants, types, functions and macros beginning with "IW_Array".
 *
 * MODIFICATIONS
 *    Mon, 21 Sep 1992 [eryq]
 *       Release 1.0 under old Pseudoclass standard. 
 *
 *    Tue, 16 Feb 1993 [eryq]
 *       Release 2.0 under new "Pc" naming convention.  Fixed realloc'ing 
 *       bug, and completely rewrote SetSize() algorithm, which under old
 *       version was buggy, and which caused confusing behavior of a buffer
 *       enlargement after a size reduction (in some cases).
 *
 *    Tue, 23 Feb 1993 [eryq]
 *       Release 2.1.  Fixed SetSize() bug (it seems I hadn't done Grow()
 *       properly, nor had I finished Shrink()!).
 *
 *    Mon, 04 Oct 1993 [eryq]
 *       Fixed documentation bug in Qsort() method as to how the callback
 *       should be declared.  Added new sorting method, QsortElems(), which
 *       takes a callback that compares two elems.  Non-reentrant, but
 *       this shouldn't be a problem.
 *
 * AUTHOR
 *    Eryq (Erik Dorfman)
 *    Hughes STX
 *    Mon, 21 Sep 1992
 */

#ifndef __IW_Array_h__
#define __IW_Array_h__


#include "IW.h"    /* conditional-compilation macros, etc. */
#include "IW_Types.h"    /* Pseudoclass environment types, macros, etc. */



/************************************************************
 * 
 * Constants:
 *
 ************************************************************/

/* "Ok" return values: */
#define IW_ArrayOSTOP       2   /* iterator hit end of array; stop */
#define IW_ArrayOGO         1   /* iterator got next element; keep going */
#define IW_ArrayOK          0   /* ok */

/* "Error" return values (by definition, error is signaled by being
   less than the generic OK value): */
#define IW_ArrayEMISC      -1   /* misc. error */
#define IW_ArrayENOTFOUND  -2   /* elem not found by test-driven search */
#define IW_ArrayEBOUNDS    -3   /* index out of bounds */
#define IW_ArrayESYS       -4   /* system call err; check errno for details */
#define IW_ArrayEHALTED    -5   /* iteration halted by action */


/************************************************************
 * 
 * Types:
 *
 ************************************************************/

/* An "iterator" (a black-box index into an array): */
typedef int IW_ArrayIter;

/* A variable-sized array: */
typedef struct {
  int  bufsize;        /* number of elements that we have buffer space for */
  int  numelems;       /* number of elements "currently" in array */
  IW_Elem *elems;      /* an array of pointers to the actual elements */
  IW_FnDiscard discard_f;  /* function to discard an element, or NULL */
  IW_ArrayIter iter;    /* "attached iterator": a built in index into array */
} IW_Array;


/************************************************************
 * 
 * Public interface:
 *
 ************************************************************/

/* 
 * Constructors and destructor: 
 */
extern int IW_ArrayCreate       (IW_Array *, 
					  int bufsize, 
					  IW_FnDiscard discard_f);
extern int IW_ArrayCreateSize   (IW_Array *, 
					  int size,
					  IW_FnDiscard discard_f);
extern void IW_ArrayDestroy     (IW_Array *);


/* 
 * Allocators and freer:
 */
extern IW_Array *IW_ArrayAlloc      (int bufsize, 
					   IW_FnDiscard discard_f);
extern IW_Array *IW_ArrayAllocSize  (int size, 
					   IW_FnDiscard discard_f);
extern int       IW_ArrayFree       (IW_Array *);


/* 
 * Attribute setting and getting: 
 */
extern int IW_ArraySetSize  (IW_Array *, int size);
extern int IW_ArrayGetSize  (IW_Array *, int *sizep);


/* 
 * Storage and retrieval: 
 */
extern int      IW_ArrayAdd    (IW_Array *, IW_Elem);
extern IW_Elem  IW_ArrayAt     (IW_Array *, int);
extern int      IW_ArrayAtGet  (IW_Array *, int, IW_Elem *);
extern int      IW_ArrayAtPut  (IW_Array *, int, IW_Elem);
extern IW_Elem *IW_ArrayLocAt  (IW_Array *, int);  /* obsolete */
extern IW_Elem *IW_ArrayElems  (IW_Array *);


/* 
 * Iteration: 
 */
extern int IW_ArrayDo        (IW_Array *, 
				       int (*action)(int, IW_Elem, IW_Arg),
				       IW_Arg arg);
extern int IW_ArrayIRewind   (IW_Array *, IW_ArrayIter *);
extern int IW_ArrayINext     (IW_Array *, IW_ArrayIter *, 
				       int *, IW_Elem *);
extern int IW_ArrayISeek     (IW_Array *, IW_ArrayIter *, 
				       int);

     
/* 
 * Utilities: 
 */
extern void IW_ArrayQsort      (IW_Array *, 
					 int (*compare_f)());
extern void IW_ArrayQsortElems (IW_Array *, 
					 int (*compare_f)());



/************************************************************
 * IW_ArraySize
 ************************************************************
 *
 * DESCRIPTION
 *    Return the number of elements in the IW_Array.  IW_Arrays may only
 *    increase in size as the result of adding elements to the
 *    end with IW_ArrayAdd(), or by an explicit call to IW_ArraySetSize().
 *
 * NOTES
 *    Notice that this is NOT the same as the IW_Array's buffer size! 
 *
 * RETURNS
 *    The size of the array, or -1 on error.
 */

#define IW_ArraySize(this) \
  ((this) ? ((this)->numelems) : -1)


/************************************************************
  IW_ArrayAddAfter
  ************************************************************/
/*
 * DESCRIPTION
 *    Add element to the end of the IW_Array.  Notice that there is
 *    no add_before().
 *
 * RETURNS
 *    Whatever IW_ArrayAdd returns.
 */
#define IW_ArrayAddAfter IW_ArrayAdd


/************************************************************
  IW_ArrayRewind
  ************************************************************/
/* 
 * DESCRIPTION
 *    Rewind the IW_Array's built-in index.
 *
 * RETURNS
 *    IW_ArrayOK on success, <IW_ArrayOK on failure.
 */

#define IW_ArrayRewind(this) \
  IW_ArrayIRewind((this), &((this)->iter))


/************************************************************
  IW_ArrayNext
  ************************************************************/
/* 
 * DESCRIPTION
 *    Get the next element indexed by the IW_Array's built-in iterator.
 *
 * RETURNS
 *    IW_ArrayOK on success, <IW_ArrayOK on failure.
 */
#define IW_ArrayNext(this, indexp, objectp) \
  IW_ArrayINext((this), &((this)->iter), (indexp), (objectp))


/************************************************************
  IW_ArraySeek
  ************************************************************/
/* 
 * DESCRIPTION
 *    Seek the IW_Array's built-in iterator to the indicated element.
 *
 * RETURNS
 *    IW_ArrayOK on success, <IW_ArrayOK on failure. 
 */
#define IW_ArraySeek(this, index) \
  IW_ArrayISeek((this), &((this)->iter), (index))



#endif   /* __IW_Array_h__ */

/******************** end of file "IW_Array.h" ********************/
