/************************************************************
 * IW_Cgi.c
 ************************************************************
 *
 * DESCRIPTION
 *    Some basic utilities for writing DBIs in C.
 *    Using the following functions...
 *
 *        IW_CgiGetInput -- get DBI tags from the outside world
 *
 *        IW_CgiValue -- get the value of an input tag
 *        IW_CgiRewind -- prepare to iterate through tags
 *        IW_CgiNext -- get next tag/value pair
 *        IW_CgiNextValue -- get next value of given tag
 *
 *        IW_CgiStartOutput -- prepare to output to the outside world
 *
 *    will simplify writing your DBIs.
 *
 * IMPORTANT-NOTE!!!
 *    The whole notion of "local" DBIs vs. "CGI" DBIs went away in 
 *    WebLib v.1.7.  In an effort to make everything simpler, ALL
 *    DBIs take their input -- and generate their output -- exactly
 *    as CGI scripts do.
 *
 * AUTHOR
 *    Eryq (Erik Dorfman)
 *    Hughes STX Corporation
 *    30 Dec 1994
 *
 * NAMESPACE
 *    "IW_" is the Weblib C Utilities namespace.  All callable
 *     C code provided by WebLib is in this namespace.
 *
 *    "Dbi" indicates routines geared towards the creation of 
 *     database interfaces (DBIs).
 *
 * RETURN-VALUES
 *    All functions which return integer status return 0 for
 *    success and negative for failure, unless otherwise noted.
 *
 *    All functions which return pointers return NULL to 
 *    indicate failure, unless otherwise noted.
 */

/*
 * BEGIN TERMS AND CONDITIONS
 * 
 *  h 
 * 
 *                        NO WARRANTY
 * 
 * BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
 * FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT
 * WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER
 * PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND,
 * EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE
 * PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME
 * THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
 * 
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
 * REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH HOLDER
 * OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 * 
 * END TERMS AND CONDITIONS
 */

#ifndef lint
static char RCSID[] =
   "$Id: IW_Cgi.c,v 1.5 2001/03/20 15:21:23 amrhein Exp $";
#endif /* lint */



/************************************************************
 *
 * External declarations
 *
 ************************************************************/

#include "IW.h"
#include "IW_Cgi.h"
#include "IW_Str.h"
#include <stdio.h>
#include <string.h>

extern char *getenv();


/************************************************************
 *
 * Constants
 *
 ************************************************************/

/* Maximum number of tags which can be read in: */
#define IW_CgiMAXTAGS 1024

/* Linefeed and carriage-return: */
#define LF 10
#define CR 13



/************************************************************
 *
 * Types
 *
 ************************************************************/

/* A tag-value pair.  The tag will be all lowercase */
typedef struct IW_CgiEntry {
  char *tag;
  char *value;
} IW_CgiEntry;



/************************************************************
 *
 * Globals
 *
 ************************************************************/

/* The tags: */
IW_CgiEntry *gEntries[IW_CgiMAXTAGS+1];
int gNumEntries;

/* For iterating over the values: */
int gIter;



/************************************************************
 * IW_CgiAddEntry
 ************************************************************
 *
 * DESCRIPTION
 *    Add a tag to the internal data structure of tag-value
 *    pairs.  Private.
 */ 

static int IW_CgiAddEntry(tag, value)
     char *tag;
     char *value;
{
  IW_CgiEntry *new;

  /* Create new entry: */
  new = (IW_CgiEntry *)IW_Malloc(sizeof(IW_CgiEntry));

  /* Copy tag as all lowercase: */
  new->tag = IW_Strdup(tag);
  IW_Strdowncase(new->tag);
   
  /* Copy value: */
  new->value = IW_Strdup(value);  

  /* Store it: */
  if (gNumEntries == IW_CgiMAXTAGS) {
    IW_Die(0, "IW_Cgi: reached maximum number of allowed entries (%d)", 
	   IW_CgiMAXTAGS);
  }
  gEntries[gNumEntries] = new;
  ++gNumEntries;  
  
  /* Done! */
  return 0;
}



/************************************************************
 *
 * Populating list of entries via CGI tags
 *
 ************************************************************/


/************************************************************
 * IW_CgiGetword
 ************************************************************
 *
 * NOTES
 *    Taken directly from the "cgi-src" distribution.
 */ 

static void IW_CgiGetword(word, line, stop)
     char *word;
     char *line;
     char stop;
{
  int x = 0;
  int y;
  
  for (x=0; ((line[x]) && (line[x] != stop)); x++)
    word[x] = line[x];
  
  word[x] = '\0';
  if (line[x]) 
    ++x;
  y=0;
  
  while (line[y++] = line[x++])
    ;
}


/************************************************************
 * IW_CgiMakeword
 ************************************************************
 * 
 * NOTES
 *    Taken directly from the "cgi-src" distribution,
 *    with minor modifications.
 */ 

static char *IW_CgiMakeword(line, stop)
     char *line;
     char stop;
{
  int x = 0,y;
  char *word = (char *)IW_Malloc(strlen(line) + 1);
  
  for (x=0; ((line[x]) && (line[x] != stop)); x++)
    word[x] = line[x];  
  word[x] = '\0';
  if (line[x]) 
    ++x;
  y=0;
  
  while(line[y++] = line[x++])
    ;
  return word;
}


/************************************************************
 * IW_CgiFMakeword
 ************************************************************
 * 
 * NOTES
 *    Taken directly from the "cgi-src" distribution,
 *    with minor modifications.
 */ 

static char *IW_CgiFMakeword(fp, stop, len)
     FILE *fp;
     char stop;
     int *len; 
{
  int wsize;
  char *word;
  int i;
  
  wsize = 32;
  i = 0;
  word = (char *)IW_Malloc(sizeof(char) * (wsize + 1));
  
  while(1) {
    word[i] = (char)fgetc(fp);
    if (i == wsize) {
      word[i+1] = '\0';
      wsize *= 2;
      word = (char *)IW_Realloc(word,sizeof(char)*(wsize+1));
    }
    --(*len);
    if ((word[i] == stop) || (feof(fp)) || (!(*len))) {
      if (word[i] != stop) 
	i++;
      word[i] = '\0';
      return word;
    }
    ++i;
  }
}


/************************************************************
 * IW_CgiDecodeChar -- decode a hex-encoded char 
 ************************************************************
 * 
 * NOTES
 *    Taken directly from the "cgi-src" distribution,
 *    with minor modifications.
 */ 

static char IW_CgiDecodeChar(what)
     char *what;
{
  register char digit;
  
  digit = (what[0] >= 'A' ? ((what[0] & 0xdf) - 'A') + 10 : (what[0] - '0'));
  digit *= 16;
  digit += (what[1] >= 'A' ? ((what[1] & 0xdf) - 'A') + 10 : (what[1] - '0'));
  return (digit);
}


/************************************************************
 * IW_CgiDecode -- decode an encoded string
 ************************************************************
 * 
 * NOTES
 *    Taken directly from the "cgi-src" distribution,
 *    with minor modifications.
 */ 

static void IW_CgiDecode(str)
     char *str;
{
  register int x,y;
  
  for(x = 0, y = 0; str[y]; ++x, ++y) {
    if ((str[x] = str[y]) == '%') {
      str[x] = IW_CgiDecodeChar(&str[y+1]);
      y += 2;
    }
  }
  str[x] = '\0';
}


/************************************************************
 * IW_CgiPlusToSpace -- convert pluses to spaces in string
 ************************************************************
 * 
 * NOTES
 *    Taken directly from the "cgi-src" distribution,
 *    with minor modifications.
 */ 

static void IW_CgiPlusToSpace(str)
     char *str;
{
  register int x;
  
  for (x = 0; str[x]; x++) {
    if (str[x] == '+') 
      str[x] = ' ';
  }
}


/************************************************************
 * IW_CgiInit -- init the tags from CGI arguments
 ************************************************************
 *
 * DESCRIPTION
 *    We are running as a CGI script.
 *    Load the form's contents from the standard input into
 *    the private structure of entries.
 */ 

static int IW_CgiInit()
{
  int i;
  int contentLength;
  char *tag;
  char *value;
  char querystring[1024];
  register char *qs, *qe;  
  int count = 0;


  /* If this was not accessed by filling out a form, don't panic:
   * simply act as though a blank form were submitted: 
   */
  if (IW_Strequal(IW_Getenv("REQUEST_METHOD"),"POST")) {

    /* Get content length: */
    if (IW_Strequal(IW_Getenv("CONTENT_TYPE"),
		    "application/x-www-form-urlencoded")) {
      contentLength = atoi(IW_Getenv("CONTENT_LENGTH"));
    }
    else {                     /* unknown POST encoding */
      contentLength = 0;
    }

    /* Get input from stdin: */
    for (i = 0; contentLength && (!feof(stdin)); i++) {
      
      /* Get entry info: */
      value = IW_CgiFMakeword(stdin, '&', &contentLength);
      IW_CgiPlusToSpace(value);
      IW_CgiDecode(value);
      tag = IW_CgiMakeword(value, '=');
      
      /* Store entry, and free the info: */
      (void)IW_CgiAddEntry(tag, value);
      free(tag);
      free(value);
    }
  }
  else if (IW_Strequal(IW_Getenv("REQUEST_METHOD"),"GET")) {

    /* Grab query string: */
    strcpy(querystring, IW_Getenv("QUERY_STRING"));
    qs = querystring;
    while (*qs) {

      /* Extract tag: */
      tag = qs;
      if (!(qe = strchr(qs, '='))) {
	break;    /* bad string */
      }
      *qe = '\0';
      IW_CgiPlusToSpace(tag);
      IW_CgiDecode(tag);
      qs = qe + 1;

      /* Extract value: */
      value = qs;
      if (!(qe = strchr(qs, '&'))) {
	qe = qs + strlen(qs);        /* default to EOS */
	qs = qe;
      }
      else {
	qs = qe + 1;
      }
      *qe = '\0';
      IW_CgiPlusToSpace(value);
      IW_CgiDecode(value);
      
      /* Save! */
      (void)IW_CgiAddEntry(tag, value);
      ++count;
      if (count > 100) {
	IW_Err(0, "too many CGI entries");
	return -1;
      }
    }
  }    
  else {
    IW_Err(0, "unknown request method: '%s'", IW_Getenv("REQUEST_METHOD"));
    return -1;
  }

  /* Done! */
  return 0;
}




/************************************************************
 *
 * Initialization
 *
 ************************************************************/


/************************************************************
 * IW_CgiDieHandler
 ************************************************************
 *
 * DESCRIPTION
 *    Install this as the IW fail handler, if writing a CGI.
 *    It outputs an HTTP failure header.
 */ 

void IW_CgiDieHandler(reason)
     char *reason;
{
  printf("Status: 500 Failed.\n");
  printf("Content-type: text/plain\n\n");
  printf("IW_Cgi failure.\n\n");
  printf("Dying gasp:      %s\n", reason);
  printf("Low-level error: (%d) %s\n", IW_Errno, IW_Errstr);
  printf("\n");
  printf("IW_CGI_DIR  = %s\n", IW_CGI_DIR);
  printf("IW_DOCS_DIR = %s\n", IW_DOCS_DIR);
  printf("IW_LIB_DIR  = %s\n", IW_LIB_DIR);
  printf("IW_TMP_DIR  = %s\n", IW_TMP_DIR);
  exit(-1);
}


/************************************************************
 * IW_CgiGetInput -- init the package and load the values
 ************************************************************
 *
 * DESCRIPTION
 *    This inits the module, and loads the values from
 *    the input as appropriate.
 */ 

int IW_CgiGetInput()
{
  return IW_CgiInit();
}



/************************************************************
 *
 * Accessing the values
 *
 ************************************************************/



/************************************************************
 * IW_CgiValue -- get the value of the named entry
 ************************************************************
 *
 * DESCRIPTION
 *    Return the first value of the named entry, or empty if
 *    it was not provided.
 *
 *    The comparison of tags is case-insensitive.
 *
 * RETURNS
 *    A read-only string.
 */ 

char *IW_CgiValue(tag)
     char *tag;   /* the name of the entry whose value we want */
{
  register int i;
  IW_CgiEntry *entry;

  for (i = 0; i < gNumEntries; i++) {
    entry = gEntries[i];
    if (IW_Strcasecmp(entry->tag, tag) == 0) {
      return entry->value;
    }
  }
  return "";
}

char *IW_CgiDecodedValue(tag)
char *tag;        /* the name of the entry whose value we want */
{
  register int i;
  IW_CgiEntry *entry;
  char *decoded_value;

  for (i = 0; i < gNumEntries; i++) {
    entry = gEntries[i];
    if (IW_Strcasecmp(entry->tag, tag) == 0) {
          sprintf(decoded_value,"%s" , entry->value);
          IW_CgiDecode(decoded_value); 
      return (decoded_value);
    }
  }
  return "";
}




/************************************************************
 * IW_CgiRewind
 ************************************************************
 *
 * DESCRIPTION
 *    Rewind the form to the first entry.
 */ 

void IW_CgiRewind()
{
  gIter = 0;
}


/************************************************************
 * IW_CgiNext
 ************************************************************
 *
 * DESCRIPTION
 *    Get the next tag=value pair.  
 *
 * RETURNS
 *    The entry name, or NULL at end.
 */ 

char *IW_CgiNext(tagp, valuep)
     char **tagp;   /* where to put a pointer to the tag of the entry */
     char **valuep;  /* where to put a pointer to the value of the entry */
{
  IW_CgiEntry *entry;

  if (gIter >= gNumEntries) {
    *tagp = NULL;
    *valuep = NULL;
    return NULL;
  }
  else {
    entry = gEntries[gIter];
    *tagp = entry->tag;
    *valuep = entry->value;
    ++gIter;
    return *tagp;
  }
}


/************************************************************
 * IW_CgiNextValue
 ************************************************************
 *
 * DESCRIPTION
 *    Get the next tag=value pair for the given tag.  
 *    The comparison of tags is case-insensitive.
 *
 * RETURNS
 *    The entry tag, or NULL at end.
 */ 

char *IW_CgiNextValue(tag, valuep)
     char *tag;     /* the tag of the next entry we are interested in */
     char **valuep;  /* where to put a pointer to the value of the entry */
{
  char *thistag;

  while (IW_CgiNext(&thistag, valuep)) {
    if (IW_Strcasecmp(thistag, tag) == 0) {
      return thistag;
    }
  }
  return NULL;        /* reached the end! */
}


/******************** end of file "IW_Cgi.c" ********************/
