/*************************************************************
 *
 * NAME
 *    IW_Chunk - routines for outputting RFC822-style chunks
 *
 * AUTHOR
 *    Eryq (Erik Dorfman)
 *    Hughes STX Corporation
 *
 *************************************************************/

#include <string.h>
#include <ctype.h>
#include <stdarg.h>

#include "IW_New.h"
#include "IW_Chunk.h"

/* RCS id: */
static char *RCSID = 
   "$Id: IW_Chunk.c,v 1.5 1997/03/14 17:20:10 eryq Exp $";


/************************************************************
 *
 * Constants, private
 *
 ************************************************************/

/* For each level of indentation... */
#ifndef IW_ChunkINDENT
#define IW_ChunkINDENT "  "
#endif



/************************************************************
 *
 * Globals, private
 *
 ************************************************************/

/* The default chunk stream, for backwards compatibility: */
static IW_Chunksp *IW_ChunkspGDefault;




/************************************************************
 * IW_ChunkspNew
 ************************************************************
 *
 * DESCRIPTION
 *    Establish a chunk stream on a stdio stream.
 *    Use it like this:
 *
 *         IW_Chunksp *sp;
 *         ...
 *         sp = IW_ChunkspNew(stdout);
 *         ...
 *         IW_ChunkspField(sp, "FIELD", "%s", value);
 *
 * RETURNS
 *    The new chunk stream, or NULL on error.
 */

IW_Chunksp *IW_ChunkspNew(FILE *fp) 
{
  IW_Chunksp *this = IW_New(IW_Chunksp);
  memset(this, '\0', sizeof(*this));
  this->fp = fp;
  this->opts.syntax = IW_ChunkINDENTED;
  return this;
}

/************************************************************
 * IW_ChunkspFree
 ************************************************************
 *
 * DESCRIPTION
 *    Free a chunk stream.
 *    This does not close the stdio pointer!
 */

void IW_ChunkspFree(IW_Chunksp *this)
{
  IW_Free(this);
}



/************************************************************
 * IW_ChunkspIndent
 ************************************************************
 *
 * DESCRIPTION
 *    Private.  Output the indentation prefix on fields, etc.
 */

static void IW_ChunkspIndent(IW_Chunksp *this, int level)
{
  if (this->opts.syntax == IW_ChunkINDENTED) {
    register int i;
    for (i = 0; i < level; i++) {
      fputs(IW_ChunkINDENT, this->fp);
    }
  }
}

/************************************************************
 * IW_ChunkspContinue
 ************************************************************
 *
 * DESCRIPTION
 *    Private.  Output the continuation prefix on field values.
 *    This is the string that begins a continuation line.
 */

static void IW_ChunkspContinue(IW_Chunksp *this)
{
  if (this->opts.syntax == IW_ChunkINDENTED) {
    IW_ChunkspIndent(this, this->level+1); 
    fputs(": ", this->fp);
  }
  else {
    fputc(' ', this->fp);
  }
}

/************************************************************
 * IW_ChunkspFieldV
 ************************************************************
 *
 * DESCRIPTION
 *    The varargs backend to IW_ChunkspField() and IW_ChunkField().
 *    Outputs a formatted field, with proper carriage-return escapes.
 *
 * NOTES
 *    Currently suppresses a final "\n" in the value, to patch coding 
 *    accidents.
 *
 *    Must be able to support the output of DIFs... which can be huge! 
 *    So, the format of "%s" is specially recognized... it does not
 *    attempt to copy to a raw buffer.  This enables us to handle 
 *    the simplest case much more efficiently.
 *
 * WARNINGS
 *    Any format other than '%s' limits the output length to 2047 characters.
 */ 

int IW_ChunkspFieldV(IW_Chunksp *sp, char *field, char *fmt, va_list args) 
{
  static char raw[2048];
  char *raw_p = raw;
  char *s;
  FILE *fp = sp->fp;           /* get underlying file pointer */

  /* Indent: */
  IW_ChunkspIndent(sp, sp->level);  

  /* Generate raw message buffer: */
  if (strcmp(fmt, "%s") == 0) {         /* we use use data as-is */
    raw_p = va_arg(args, char *);
  }
  else {                                /* we have to use the buffer */
    vsprintf(raw, fmt, args);
  }

  /* Output field: */
  fprintf(fp, "%s: ", field);
  
  /* Output value: */
  for (s = raw_p; *s; s++) {
    if (*s == '\n') {      /* newline... */
      if ((s == raw_p) || (!*(s+1))) {   /* ignore nl if first/last char */ 
	continue;   
      }
      else {                             /* output continuation sequence */
	fputc('\n', fp);
	IW_ChunkspContinue(sp);
      }
    }
    else {                 /* not a newline... */
      fputc(*s, fp);
    }
  }
  fputc('\n', fp);
}


/************************************************************
 * IW_ChunkspField
 ************************************************************
 *
 * DESCRIPTION
 *    Output a formatted field, with proper carriage-return escapes.
 *    Use it like this:
 *
 *         IW_Chunksp sp = IW_ChunkspNew(stdout);
 *         ...
 *         IW_ChunkspField(sp, "fieldname", "%s %d", var1, var2);
 *
 *    So this:        IW_ChunkspField(sp, "Foo", "%s%d", "bar", 42);
 *    Outputs this:   Foo: bar42
 *
 * NOTES
 *    See IW_ChunkspFieldV for details.
 */ 

int IW_ChunkspField(IW_Chunksp *sp, char *field, char *fmt, ...)     
{
  va_list args;
  int result;

  va_start(args, fmt);
  result = IW_ChunkspFieldV(sp, field, fmt, args);
  va_end(args);
  return result;
}

/************************************************************
 * IW_ChunkspBeginGroup
 ************************************************************
 * Begin a chunk "group".
 */

int IW_ChunkspBeginGroup(IW_Chunksp *sp, char *groupname)
{
  FILE *fp = sp->fp;
  register int i;

  /* Indent: */
  IW_ChunkspIndent(sp, (sp->level)++);

  /* Output: */
  fprintf(fp, "BEGIN %s\n", groupname);
}

/************************************************************
 * IW_ChunkspEndGroup
 ************************************************************
 * End a chunk "group".
 */

int IW_ChunkspEndGroup(IW_Chunksp *sp, char *groupname) 
{
  FILE *fp = sp->fp;           /* get underlying file pointer */

  /* Indent: */
  IW_ChunkspIndent(sp, --(sp->level));

  /* Output: */
  fprintf(fp, "END %s\n", groupname);
}

/************************************************************
 * IW_ChunkspBegin
 ************************************************************
 * Begin a chunk. 
 * If type is given, outputs the "X-TYPE: TYPE" line.
 */

int IW_ChunkspBegin(IW_Chunksp *sp, char *type) 
{
  sp->level = 0;

  /* Special syntax flag for indentation: */
  if (sp->opts.syntax == IW_ChunkINDENTED) {
      fprintf(sp->fp, "--\n");
  }

  /* Output special type field: */
  if (type) {
    IW_ChunkspField(sp, "X-TYPE", "%s", type);
  }
}

/************************************************************
 * IW_ChunkspEnd
 ************************************************************
 * End a chunk.
 */

int IW_ChunkspEnd(IW_Chunksp *sp, char *type) 
{
  /* End chunk: */
  fprintf(sp->fp, "\n");
}




/******************************
 *
 * BACKWARDS-COMPATIBILITY...
 *
 ******************************
 * All of the following are deprecated.  They are provided for 
 * convenience and backwards-compatibility only!
 */


/************************************************************
 * IW_ChunkspDefault
 ************************************************************
 * Assert a default stream on the given file pointer.
 */

static IW_Chunksp* IW_ChunkspDefault(FILE *fp)
{
  if (!(IW_ChunkspGDefault)) {
    IW_ChunkspGDefault = IW_ChunkspNew(fp);
  }
  IW_ChunkspGDefault->fp = fp;
  IW_ChunkspGDefault->opts.syntax = IW_ChunkFLUSHLEFT;
  return IW_ChunkspGDefault;
}

/************************************************************
 * IW_ChunkBegin
 ************************************************************
 * Begin a chunk.
 */

int IW_ChunkBegin(FILE *fp, char *type) {
  IW_ChunkspBegin(IW_ChunkspDefault(fp), type);
}

/************************************************************
 * IW_ChunkEnd
 ************************************************************
 * End a chunk.
 */

int IW_ChunkEnd(FILE *fp) {
  IW_ChunkspEnd(IW_ChunkspDefault(fp), NULL);
}

/************************************************************
 * IW_ChunkBeginGroup
 ************************************************************
 * Begin a chunk "group".
 */

int IW_ChunkBeginGroup(FILE *fp, char *groupname) {
  IW_ChunkspBeginGroup(IW_ChunkspDefault(fp), groupname);
}

/************************************************************
 * IW_ChunkEndGroup
 ************************************************************
 * End a chunk "group".
 */

int IW_ChunkEndGroup(FILE *fp, char *groupname) {
  IW_ChunkspEndGroup(IW_ChunkspDefault(fp), groupname);
}

/************************************************************
 * IW_ChunkField
 ************************************************************
 * Output a chunk field.
 */

int IW_ChunkField(FILE *fp, char *field, char *fmt, ...)     
{
  va_list args;
  int result;

  va_start(args, fmt);
  result = IW_ChunkspFieldV(IW_ChunkspDefault(fp), field, fmt, args);
  va_end(args);
  return result;
}






/******************************
 *
 * UTILITIES...
 *
 ******************************/


/************************************************************
 * IW_ChunkUnfold
 ************************************************************
 *
 * DESCRIPTION
 *    Collapse a NULL-terminated string vector into a single string.
 *
 *        char *s = IWM_Multiline(argv);
 *
 * RETURNS
 *    The unfolded string.
 *
 * WARNINGS
 *    The string is limited to 4095 characters.
 */ 

char *IW_ChunkUnfold(char **strv)
{
  static char buffer[4096];
  char **sp;
  char *tp = buffer;

  *tp = '\0';
  for (sp = strv; *sp; sp++) {
    strcpy(tp, *sp);
    tp += strlen(tp);
    *tp++ = '\n';
  }
  *tp = '\0';
  return buffer;
}

/*************************** end of file ***************************/

