/************************************************************
 * IW_Chunk.h
 ************************************************************
 *
 * NAME
 *    IW_Chunk - routines for creating and outputting to "chunk streams"
 *
 * DESCRIPTION
 *    RFC-822 style chunks are the kind you see at the tops
 *    of email messages, like:
 *
 *         From: yakko@wb.com <Yakko Warner>
 *         To: wakko@wb.com <Wakko Warner>
 *         Subject: Hey, did someone leave the 
 *          fridge unplugged last night?
 * 
 *    Noitce how continuation lines are flagged by a leading space.
 *
 *    Chunks are terminated by -- and ONLY by -- a completely
 *    blank line (nothing before the "\n") or an EOF:
 *
 *         From: yakko@wb.com <Yakko Warner>
 *         To: wakko@wb.com <Wakko Warner>
 *         Subject: Hey, did someone leave the 
 *          fridge unplugged last night?
 *
 *         From: dot@wb.com <Dot Warner>
 *         To: yakko@wb.com <Yakko Warner>
 *         Subject: Wasn't me!
 *
 *    In IMS/WWW 1.6 at least, this syntax was extended to allow
 *    subgroups (notice the BEGIN and END):
 *
 *         NAME: John 
 *         AGE: 21
 *         WEIGHT: 17.5
 *         BEGIN ADDRESS
 *         STREET: Yoyodyne Propulsion Labs
 *          123 Johnstown Road
 *          Suite 1
 *         CITY: Grover's Mill
 *         STATE: NJ
 *         END ADDRESS
 *
 *    In IMS/WWW 1.7, we've extended the syntax to handle subgroups 
 *    and indenting (notice how continuation lines now have a leading ':'):
 *
 *         --
 *         NAME: John 
 *         AGE: 21
 *         WEIGHT: 17.5
 *         BEGIN ADDRESS
 *            STREET: Yoyodyne Propulsion Labs
 *                  : 123 Johnstown Road
 *                  : Suite 1
 *            CITY: Grover's Mill
 *            STATE: NJ
 *         END ADDRESS
 *
 *    The leading "--" allows the parser to detect and handle both the 
 *    older "flush-left" and the newer "indented" styles.
 *    Like a blank line or a comment, /^--/ also ends a chunk.
 *
 *    (The "--" is intended to be invocative of old-style multipart mail
 *    separators, which are still used above signatures in messages:
 *    also, it was a syntax error in previous releases.)
 *
 *    Do not put anything after the '--' on the line; that syntax
 *    is reserved for future expansions.
 *
 * AUTHOR
 *    Eryq (Erik Dorfman)
 *    Hughes STX Corporation
 *
 ************************************************************/

#ifndef __IW_Chunk_h__
#define __IW_Chunk_h__

#include <stdio.h>


/************************************************************
 *
 * Constants
 *
 ************************************************************/

/* Syntaxes: */
#define IW_ChunkFLUSHLEFT 0
#define IW_ChunkINDENTED  1


/************************************************************
 *
 * Types
 *
 ************************************************************/

/* A chunk input/output stream: */
typedef struct IW_Chunksp {
  FILE *fp;          /* file pointer */
  short level;       /* current indentation level on stream */
  struct {
    unsigned syntax : 2;    /* Output syntax: FLUSHLEFT or INDENTED */
  } opts;
} IW_Chunksp;


/************************************************************
 *
 * Macros
 *
 ************************************************************/

/* Get the underlying file pointer for the chunk stream: */
#define IW_ChunkspFp(this) \
    ((this)->fp)

/* Set the syntax for the stream: */
#define IW_ChunkspSetSyntax(this, syn) \
    ((this)->opts.syntax = (syn))



/************************************************************
 *
 * Prototypes
 *
 ************************************************************/

/* Opening/closing chunk streams: */
extern IW_Chunksp *IW_ChunkspNew(FILE *fp);
extern void        IW_ChunkspFree(IW_Chunksp *);

/* Outputting chunks to chunk streams: */
extern int IW_ChunkspBegin      (IW_Chunksp *, char *);
extern int IW_ChunkspEnd        (IW_Chunksp *, char *);
extern int IW_ChunkspBeginGroup (IW_Chunksp *, char *);
extern int IW_ChunkspEndGroup   (IW_Chunksp *, char *);
extern int IW_ChunkspField      (IW_Chunksp *, char *, char *, ...);



/* Outputting chunks to file pointers (deprecated; backwards compatibility): */
extern int IW_ChunkBegin      (FILE *, char *);
extern int IW_ChunkEnd        (FILE *);
extern int IW_ChunkBeginGroup (FILE *, char *);
extern int IW_ChunkEndGroup   (FILE *, char *);
extern int IW_ChunkField      (FILE *, char *, char *, ...);

/* Utilities: */
extern char *IW_ChunkUnfold(char **);


#endif /* __IW_Chunk_h__ */

/******************** end of file "IW_Chunk.h" ********************/

