/*************************************************************
 *
 * NAME
 *    IW_Conn - structure for a communication status
 *
 * DESCRIPTION
 *    See IW_Conn.h for details.
 *
 * AUTHOR
 *    Eryq (Erik Dorfman)
 *    Hughes STX Corporation
 *
 *************************************************************/

/*
 * BEGIN TERMS AND CONDITIONS
 * END TERMS AND CONDITIONS
 */

#ifndef lint
static char RCSID[] =
   "$Id: IW_Conn.c,v 1.1 1997/03/14 17:20:10 eryq Exp $";
#endif /* lint */

#include "IW_New.h"
#include "IW_Conn.h"
#include "IW_Str.h"

#include "IK_InDsetStatCodeCmt.h"
#include "IK_Nmsts.h"


/************************************************************
 *
 * IW_Connstage methods...
 *
 ************************************************************/

/************************************************************
 * IW_ConnstageName
 ***********************************************************
 * Map an IW_ConnS_XXX constant to a printable name.
 */

char *IW_ConnstageName(IW_Connstage stage) 
{
  switch (stage) {
  case IW_ConnS_PENDING:   return "PENDING";    
  case IW_ConnS_CONNECTED: return "CONNECTED";
  case IW_ConnS_SENDING:   return "SENDING";
  case IW_ConnS_RECEIVING: return "RECEIVING";
  case IW_ConnS_OK:        return "OK";
  case IW_ConnS_EABORTED:  return "EABORTED";
  case IW_ConnS_EDAACDOWN: return "EDAACDOWN";
  case IW_ConnS_EFAILED:   return "EFAILED";
  case IW_ConnS_ELOOKSBAD: return "ELOOKSBAD";
  default:                 return "(UNKNOWN STAGE)";
  }
}



/************************************************************
 *
 * IW_Conn methods...
 *
 ************************************************************/


/******************************
 *
 * Construction/destruction...
 *
 ******************************/

/************************************************************
 * IW_ConnNew
 ***********************************************************
 * Return a new connection status object.
 */
IW_Conn *IW_ConnNew(char *arch_id)
{
  IW_Conn *this;

  /* Allocate space, and zero: */
  this = IW_New(IW_Conn);
  memset(this, '\0', sizeof(IW_Conn));

  /* Setup basic values: */
  strcpy(this->arch_id, arch_id);
  this->LED_stage = IW_ConnS_PENDING;
  this->status_code = NM_ABNORMAL;

  /* Allocate hashtables: */
  this->dataset_keys = IW_HashtAllocCS(41, NULL);
  this->granule_keys = IW_HashtAllocCS(41, NULL);
	 
  /* Done! */
  return this;
}

/************************************************************
 * IW_ConnFree
 ***********************************************************
 * Delete a connection status object.
 */
void IW_ConnFree(IW_Conn *this)
{
  if (this) {
    IW_HashtFree(this->dataset_keys);
    IW_HashtFree(this->granule_keys);
    IW_Free(this);
  }
}



/******************************
 *
 * Status code comment...
 *
 ******************************/


/************************************************************
 * IW_ConnSetStatusCmnt
 ***********************************************************
 * Set the status code comment.  The string is automatically duplicated,
 * and any existing string is freed.
 *
 * You may set it to NULL, if you like.
 */
void IW_ConnSetStatusCmnt(IW_Conn *this, char *cmnt)
{
  /* Free old: */
  if (this->status_code_cmnt) {
    IW_Free(this->status_code_cmnt); 
    this->status_code_cmnt = NULL;
  }

  /* Set new: */
  if (cmnt) {
    this->status_code_cmnt = IW_Strdup(cmnt);
  }
}

/************************************************************
 * IW_ConnStatusCmnt
 ***********************************************************
 * Return the literal status code comment, as a string.
 */
char *IW_ConnStatusCmnt(IW_Conn *this)
{
  return (this->status_code_cmnt ? this->status_code_cmnt : "");
}

/************************************************************
 * IW_ConnAnyStatusCmnt
 ***********************************************************
 * Return the literal status code comment, or a fake one
 * if none is available.  The string you get back may exist 
 * in status memory; use it quick!
 */
char *IW_ConnAnyStatusCmnt(IW_Conn *this)
{
  char *cmnt;
  static char cmnt_buf[IK_MAX_DATA_CENTER_COMMENT_SZ+1];

  if ((cmnt = this->status_code_cmnt) && (*cmnt)) {
    return cmnt;
  }
  else if (this->status_code == 0) {       /* no real info */
    return "";
  }
  else {
    IK_GetCommunicationComment(this->status_code, cmnt_buf);
    cmnt_buf[IK_MAX_DATA_CENTER_COMMENT_SZ] = '\0';
    return cmnt_buf;
  }
}




/******************************
 *
 * Datasets... 
 *
 ******************************/

/************************************************************
 * IW_ConnAddDataset
 ***********************************************************
 * Add the given dataset web key (e.g., "1/234/GID") to the list
 * of datasets we've seen.
 */
int IW_ConnAddDataset(IW_Conn *this, char *web_gkey)
{
  return IW_HashtAtPut(this->dataset_keys, web_gkey, (IW_Elem)1);
}

/************************************************************
 * IW_ConnHasDataset
 ***********************************************************
 * Has the given dataset web key (e.g., "1/234/GID") been seen?
 */
int IW_ConnHasDataset(IW_Conn *this, char *web_gkey)
{
  return (int)(IW_HashtAt(this->dataset_keys, web_gkey));
}

/************************************************************
 * IW_ConnNumDatasets
 ***********************************************************
 * Return the total number of datasets we've seen.
 */
int IW_ConnNumDatasets(IW_Conn *this)
{
  return (IW_HashtNumElems(this->dataset_keys));
}



/******************************
 *
 * Granules... 
 *
 ******************************/

/************************************************************
 * IW_ConnAddGranule
 ***********************************************************
 * Add the given granule web key (e.g., "1/234/GID") to the list
 * of granules we've seen.
 */
int IW_ConnAddGranule(IW_Conn *this, char *web_gkey)
{
  return IW_HashtAtPut(this->granule_keys, web_gkey, (IW_Elem)1);
}

/************************************************************
 * IW_ConnHasGranule
 ***********************************************************
 * Has the given granule web key (e.g., "1/234/GID") been seen?
 */
int IW_ConnHasGranule(IW_Conn *this, char *web_gkey)
{
  return (int)(IW_HashtAt(this->granule_keys, web_gkey));
}

/************************************************************
 * IW_ConnNumGranules
 ***********************************************************
 * Return the total number of granules we've seen.
 */
int IW_ConnNumGranules(IW_Conn *this)
{
  return (IW_HashtNumElems(this->granule_keys));
}


/******************** end of file "IW_Conn.c" ********************/

