/*************************************************************
 *
 * NAME
 *    IW_Conn - structure for a communication status
 *
 * DESCRIPTION
 *    This abstract datatype represents a single connection
 *    to a DAAC in the imsclient application.  It contains
 *    information more oriented to the WWW interface than
 *    the corresponding CONNECTION_INFO structure maintained
 *    by the IK layer.
 *
 * AUTHOR
 *    Eryq (Erik Dorfman)
 *    Hughes STX Corporation
 *
 *************************************************************/

#ifndef __IW_Conn_h__
#define __IW_Conn_h__

#include "IW_Hasht.h"



/************************************************************
 *
 * Constants
 *
 ************************************************************/

/* 
 * Stage of the comm status screen LED display: 
 * WARNING: THESE CONSTANTS MUST AGREE WITH THE VALUES IN THE PERL IW::Conn! 
 */

/*   The stage will creep up through these, in order: */
#define IW_ConnS_PENDING      0  
#define IW_ConnS_CONNECTED    1
#define IW_ConnS_SENDING      2
#define IW_ConnS_RECEIVING    3
/*   The stage will end at one (and only one) of these: */
#define IW_ConnS_END_MIN      4  /* min final stage */
#define IW_ConnS_OK           4
#define IW_ConnS_EABORTED     5
#define IW_ConnS_EDAACDOWN    6
#define IW_ConnS_EFAILED      7
#define IW_ConnS_ELOOKSBAD    8  /* we never have to deal with this here */



/************************************************************
 *
 * Types
 *
 ************************************************************/

/* An LED stage: */
typedef int IW_Connstage;

/* A single connection: */
typedef struct IW_Conn {

  /*                        Basic info... */
  char arch_id[32];             /* data center id */      

  /*                        Hashes of keys we've seen... */
  IW_Hasht *dataset_keys;       /* web keys of datasets we've tallied */
  IW_Hasht *granule_keys;       /* web keys of granules we've tallied */

  /*                        Status info... */
  int  status_code;             /* comm. status code */
  char *status_code_cmnt;       /* comm. status code comment */
  IW_Connstage LED_stage;       /* stage of the comm status LED */

} IW_Conn;



/************************************************************
 *
 * Macros
 *
 ************************************************************/

/* Get a pointer to the DAAC name: */
#define IW_ConnDAAC(this) \
    ((this)->arch_id)

/* Get/set the current stage (no sanity checking): */
#define IW_ConnStage(this) \
    ((this)->LED_stage)
#define IW_ConnSetStage(this, new) \
    ((this)->LED_stage = (new))

/* Get/set the status code: */
#define IW_ConnStatusCode(this) \
    ((this)->status_code)
#define IW_ConnSetStatusCode(this, code) \
    ((this)->status_code = (code))




/************************************************************
 *
 * Prototypes
 *
 ************************************************************/

/* Stages: */
char    *IW_ConnstageName (IW_Connstage);

/* Constructing: */
IW_Conn *IW_ConnNew  (char *);
void     IW_ConnFree (IW_Conn *);

/* Status comment: */
void  IW_ConnSetStatusCmnt (IW_Conn *, char *);
char *IW_ConnStatusCmnt    (IW_Conn *);
char *IW_ConnAnyStatusCmnt (IW_Conn *);

/* Datasets: */
int IW_ConnAddDataset(IW_Conn *, char *);
int IW_ConnHasDataset(IW_Conn *, char *);
int IW_ConnNumDatasets(IW_Conn *);

/* Granules: */
int IW_ConnAddGranule(IW_Conn *, char *);
int IW_ConnHasGranule(IW_Conn *, char *);
int IW_ConnNumGranules(IW_Conn *);


#endif /* __IW_Conn_h__ */

/******************** end of file "IW_Conn.h" ********************/
