/************************************************************
 * IW_Hasht.c -- general-purpose hashtable
 ************************************************************
 *
 * DESCRIPTION
 *    General-purpose hashtable Pseudoclass.
 *
 * AUTHOR
 *    Eryq (Erik Dorfman)
 *    Hughes STX Corporation
 */

/*
* BEGIN TERMS AND CONDITIONS
* 
*  h 
* 
*                        NO WARRANTY
* 
* BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
* FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT
* WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER
* PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND,
* EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
* PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE
* PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME
* THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
* 
* IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
* WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
* REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR
* DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL
* DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM
* (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
* INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF
* THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH HOLDER
* OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
* 
* END TERMS AND CONDITIONS
 */


#ifndef lint
static char RCSID[] = 
   "$Id: IW_Hasht.c,v 1.3 1996/04/17 21:16:15 imswww Exp $";
#endif /* lint */


#include "IW_Hasht.h"
#include "IW_New.h"

#include <stdio.h>
#include <ctype.h>
#include <string.h>

#define IW_Hasht_GOOD_PRIME      211
#define IW_Hasht_TOP_FOUR_BITS   0xF0000000
#define IW_Hasht_THREE_BYTES     24


/************************************************************
 * 
 * Private types
 *
 ************************************************************/

/* 
 * Private: a hashtable node: 
 */
typedef struct IW_HashtNode {
  char *key;                 /* the hash key */
  IW_Elem value;             /* (pointer to) the actual value being stored */
  struct IW_HashtNode *next;  /* next node in the bucket */
} IW_HashtNode;



/************************************************************
 *
 * Private macros
 *
 ************************************************************/

/************************************************************
 * IW_HashtAsLower
 ************************************************************
 *
 * Convert character to lowercase.
 */
#define IW_HashtAsLower(c) \
  (islower((c)) ? (c) : ((c)+'a'-'A'))



/************************************************************
 *
 * String functions
 *
 ************************************************************/

/************************************************************
 * IW_HashtStrEqual
 ************************************************************
 * 
 * DESCRIPTION
 *    Return nonzero (true) if the two strings are equal (in a
 *    case-sensitive comparison), and zero (false) if they are not.
 */

static int IW_HashtStrEqual(s, t) 
     char *s;
     char *t;
{
  return (!strcmp(s,t));
}


/************************************************************
 * IW_HashtStrEqualCI
 ************************************************************
 * 
 * DESCRIPTION
 *    Return nonzero (true) if the two strings are equal (in a
 *    case-insensitive comparison), and zero (false) if they are not.
 */

static int IW_HashtStrEqualCI(s, t) 
     char *s;
     char *t;
{
  if (s && t) {
    for ( ; (IW_HashtAsLower(*s) == IW_HashtAsLower(*t)); s++, t++)
      if (*s == '\0')
	return 1;
    return 0;
  }
  else {
    return(s == t);
  }
}


/************************************************************
 *
 * Hashing functions
 *
 ************************************************************/


/************************************************************
 * IW_HashtHashPJW -- hash function
 ************************************************************
 * 
 * DESCRIPTION
 *    A good hash function, from p.436 of the 1986 edition of Aho, 
 *    Sethi & Ullman's book "Compilers" (also known as the "Dragon" book.  
 *
 * ARGUMENTS
 *    A string to hash.
 *
 * RETURNS
 *    An integer (the hash value).  Note that this value must ultimately 
 *    be taken modulo the number of buckets in the hashtable.
 */
int IW_HashtHashPJW (str)
     char *str;   /* the string to hash */
{
  register char *s;
  register unsigned hashval = 0;
  register unsigned g;

  for (s = str; *s; s++) {
    hashval = (hashval << 4) + *s;
    if (g = (hashval & IW_Hasht_TOP_FOUR_BITS)) {
      hashval ^= (g >> IW_Hasht_THREE_BYTES);
      hashval ^= g;
    }
  }
  return (hashval);
}


/************************************************************
 * IW_HashtHashPJWCI  -- case-insensitive hash function
 ************************************************************
 * 
 * DESCRIPTION
 *    A good hash function, from p.436 of the 1986 edition of Aho, 
 *    Sethi & Ullman's book "Compilers" (also known as the "Dragon" book. 
 *
 * ARGUMENTS
 *    A string to hash.
 *
 * RETURNS
 *    An integer (the hash value).  Note that this value must ultimately 
 *    be taken modulo the number of buckets in the hashtable.
 */

int IW_HashtHashPJWCI (str)
     char *str;   /* the string to hash */
{
  char *s;
  unsigned hashval = 0;
  unsigned g;

  for (s = str; *s; s++) {
    hashval = (hashval << 4) + (isupper(*s) ? tolower(*s) : *s);
    if (g = (hashval & IW_Hasht_TOP_FOUR_BITS)) {
      hashval ^= (g >> IW_Hasht_THREE_BYTES);
      hashval ^= g;
    }
  }
  return (hashval);
}



/************************************************************
 *
 * Internal utilities
 *
 ************************************************************/


/************************************************************
 * IW_HashtAllocateBuckets (private)
 ************************************************************
 *
 * DESCRIPTION
 *    Allocate memory for the buckets.  Leaves original configuration
 *    undisturbed if request could not be granted.  If success, sets
 *    this->buckets to the new array of buckets and this->num_buckets
 *    to the new number of buckets (as given).
 *
 * RETURNS
 *    IW_HashtOK on success, <IW_HashtOK on error.
 */

static int IW_HashtAllocateBuckets(this, newnumbuckets)
     IW_Hasht *this; /* the hashtable */
     int newnumbuckets;
{
  register int i;
  IW_HashtNode **newbuckets;

  /* Allocate array of pointers to nodes: */
  newbuckets = IW_NewArray(IW_HashtNode *, newnumbuckets);
  if (!newbuckets) {
    IW_Err(0, "couldn't allocate buckets");
    return IW_HashtESYS;
  }

  /* Init all pointers to NULL: */
  for (i = 0; i < newnumbuckets; i++) {
    newbuckets[i] = NULL;
  }

  /* Install new buckets! */
  this->buckets = newbuckets;
  this->num_buckets = newnumbuckets;

  return IW_HashtOK;
}


/************************************************************
 * IW_HashtNodeAt (private)
 ************************************************************
 * 
 * DESCRIPTION
 *    Given a key, find the *node* which correpsonds to that key.
 *
 * RETURNS
 *    A pointer to that node, or NULL if no such node exists.
 */

static IW_HashtNode *IW_HashtNodeAt(this, key) 
     IW_Hasht *this; /* the hashtable */
     char *key;   /* the lookup key */
{
  IW_HashtNode *node;
  int bucket_number;

  bucket_number = (*(this->hash_f))(key) % this->num_buckets; 

  for (node = this->buckets[bucket_number]; node != NULL; node = node->next) {
    if ((*(this->compare_f))(key, node->key)) 
      return node;  /* found a match */
  }
  return (IW_HashtNode *)NULL;  /* no match */
}


/************************************************************
 * IW_HashtInstall (private)
 ************************************************************
 * 
 * DESCRIPTION
 *    Install a new element in the hashtable at the given key.  
 *
 * NOTES
 *    This assumes that the entry does not already exist, and as
 *    such, this function should NOT be called directly...use
 *    IW_HashtAtPut() instead!!!!  
 *
 * RETURNS
 *    A pointer to the node where the entry was stored, or NULL
 *    if the installation failed.
 */

static IW_HashtNode *IW_HashtInstall(this, key, value)
     IW_Hasht *this; /* the hashtable */
     char *key;     /* the lookup key */
     IW_Elem value;  /* the value to store */
{
  IW_HashtNode *node;
  int hashval;
  int i;
  IW_HashtNode *old_node;
  IW_HashtNode **old_buckets;
  int old_num_buckets;
  int new_num_buckets;
  

  /* Attempt to allocate space for a node: */
  node = IW_New(IW_HashtNode);
  if (node == NULL) {
    return NULL;
  }  

  /* Store a fresh copy of the key, and store the given value, in the 
     new hashtable node: */  
  node->key = IW_NewArray(char, 1+strlen(key));
  if (!node->key) {
    IW_Free(node);
    return NULL;
  }
  strcpy(node->key, key);
  node->value = value;
  
  /* Get the bucket number, and install the new hashtable node: */
  hashval = (*(this->hash_f))(key) % this->num_buckets;
  node->next = this->buckets[hashval];
  this->buckets[hashval] = node;
  ++this->num_entries;

  /* Now check to see if we should grow the hash table... if we cannot, 
     then don't fail; simply return the node.  This hash table will be
     less efficient, that's all...*/
  if (this->num_entries > (this->num_buckets * this->max_avg_bucket_size)){
    
    /* Save the old setup: */
    old_buckets = this->buckets;
    old_num_buckets = this->num_buckets;
    
    /* Create space for the new buckets (forcing new size to be odd): */
    new_num_buckets = (this->num_entries | 1);
    if (IW_HashtAllocateBuckets(this, new_num_buckets) < IW_HashtOK) {
      return node;
    }
    
    /* Rehash: */
    for (i = 0; i < old_num_buckets; i++) {    /* for each (old) bucket: */
      while (old_buckets[i] != NULL) {             /* for each node in list: */

	/* Save pointer to node: */
	old_node = old_buckets[i];
	
	/* Remove node from old bucket list: */
	old_buckets[i] = old_node->next;
	
	/* Install in new bucket list, right before the first element: */
	hashval = (*(this->hash_f))(old_node->key) % (this->num_buckets);
	old_node->next = this->buckets[hashval];
	this->buckets[hashval] = old_node;
      }
    }
    
    /* Kill old space: */
    IW_Free((char *)(old_buckets));
  }
  
  /* Return the installed node: */
  return(node);
}



/************************************************************
 *
 * Constructors/destructors
 *
 ************************************************************/


/************************************************************
 * IW_HashtCreate -- constructor
 ************************************************************
 * 
 * DESCRIPTION
 *    Create a hash table. 
 *
 * RETURNS
 *    IW_HashtOK on success, <IW_HashtOK on error.
 *
 * WARNINGS
 *    Obsolete.  Use IW_HashtAlloc() instead.
 */

int IW_HashtCreate(this, num_buckets, 
		  hash_f, compare_f, discard_f)
     IW_Hasht *this;               /* the hashtable */
     int num_buckets;             /* initial number of buckets to allocate */
     IW_HashtFnHash hash_f;        /* used to hash a key */
     IW_HashtFnCompare compare_f;  /* used to compare two keys */
     IW_FnDiscard discard_f;       /* used to discard an element */
{
  int status;

  /* Set the number of buckets: */
  if (num_buckets <= 0) {
    num_buckets = IW_Hasht_GOOD_PRIME;
  }
  
  /* Create the buckets: */
  if ((status = IW_HashtAllocateBuckets(this, num_buckets)) < IW_HashtOK) {
    return status; 
  }
 
  /* Initial number of entries: */
  this->num_entries = 0;
  
  /* The supporting functions: */
  this->hash_f = hash_f;
  this->compare_f = compare_f;
  this->discard_f = discard_f;  
  
  /* Internal addendum: */
  this->max_avg_bucket_size = 2;
  return IW_HashtOK;
}


/************************************************************
 * IW_HashtCreateCS -- constructor
 ************************************************************
 *
 * DESCRIPTION
 *    Create a standard case-sensitive hash table.
 *
 * WARNINGS
 *    Obsolete.  Use IW_HashtAllocCS() instead.
 *
 * RETURNS
 *    IW_HashtOK on success, <IW_HashtOK on failure.
 */

int IW_HashtCreateCS(this, num_buckets, discard_f)
     IW_Hasht *this;      /* the hashtable */
     int num_buckets;    /* initial number of buckets to allocate */
     IW_FnDiscard discard_f;  /* used to discard an element */
{
  return IW_HashtCreate(this, 
		       num_buckets, 
		       IW_HashtHashPJW, 
		       IW_HashtStrEqual,
		       discard_f);
}


/************************************************************
 * IW_HashtCreateCI -- constructor
 ************************************************************
 *
 * DESCRIPTION
 *    Create a standard case-insensitive hash table.
 *
 * RETURNS
 *    IW_HashtOK on success, <IW_HashtOK on failure.
 *
 * WARNINGS
 *    Obsolete.  Use IW_HashtAllocCI() instead.
 */

int IW_HashtCreateCI(this, num_buckets, discard_f)
     IW_Hasht *this;         /* the hashtable */
     int num_buckets;       /* initial number of buckets to allocate */
     IW_FnDiscard discard_f; /* used to discard an element */
{
  return IW_HashtCreate(this, 
		       num_buckets, 
		       IW_HashtHashPJWCI, 
		       IW_HashtStrEqualCI, 
		       discard_f);
}


/************************************************************
 * IW_HashtAlloc -- allocator
 ************************************************************
 *
 * DESCRIPTION
 *    Allocate a general-purpose hashtable.  The keys
 *    are strings, but how they are hashed and how they
 *    are compared is up the user.
 *
 * RETURNS
 *    The allocated hashtable, or NULL on failure.
 */ 

IW_Hasht *IW_HashtAlloc(num_buckets, hash_f, compare_f, discard_f)
     int num_buckets;             /* initial number of buckets to allocate */
     IW_HashtFnHash hash_f;        /* used to hash a key */
     IW_HashtFnCompare compare_f;  /* used to compare two keys */
     IW_FnDiscard discard_f;       /* used to discard an element */
{
  IW_Hasht *new;

  /* Allocate space for the hashtable, and then create it: */
  if (!(new = IW_New(IW_Hasht))) {
    return NULL;
  }
  if (IW_HashtCreate(new, num_buckets, 
		    hash_f, compare_f, discard_f) < IW_HashtOK) {
    IW_Free(new);
    return NULL;
  }
  return new;
}


/************************************************************
 * IW_HashtAllocCS
 ************************************************************
 *
 * DESCRIPTION
 *    Allocate a hashtable whose keys are case-sensitive strings.
 *
 * RETURNS
 *    The allocated hashtable, or NULL on failure.
 */ 

IW_Hasht *IW_HashtAllocCS(num_buckets, discard_f)
     int num_buckets;             /* initial number of buckets to allocate */
     IW_FnDiscard discard_f;       /* used to discard an element */
{
  return IW_HashtAlloc(num_buckets, IW_HashtHashPJW, IW_HashtStrEqual, discard_f);
}


/************************************************************
 * IW_HashtAllocCI
 ************************************************************
 *
 * DESCRIPTION
 *    Allocate a hashtable whose keys are case-insensitive strings.
 *
 * RETURNS
 *    The allocated hashtable, or NULL on failure.
 */ 

IW_Hasht *IW_HashtAllocCI(num_buckets, discard_f)
     int num_buckets;             /* initial number of buckets to allocate */
     IW_FnDiscard discard_f;       /* used to discard an element */
{
  return IW_HashtAlloc(num_buckets, IW_HashtHashPJWCI, IW_HashtStrEqualCI, 
		      discard_f);
}


/************************************************************
 * IW_HashtDestroy -- destructor
 ************************************************************
 * 
 * DESCRIPTION
 *    Destroy a hashtable.  This frees the space allocated for each
 *    key, and calls the supplied discard-function to free each value
 *    (if the function is not NULL).
 *
 * WARNINGS
 *    Obsolete.  Use the IW_HashtAlloc...() functions, and destroy the
 *    hashtable with IW_HashtFree().
 */

void IW_HashtDestroy(this)
     IW_Hasht *this; /* the hashtable */
{
  int i;
  IW_HashtNode *node, *next;

  /* Free each bucket: */
  for (i = 0; i < this->num_buckets; i++) {
    node = this->buckets[i];
    while (node != NULL) {

      /* Save pointer to next node: */
      next = node->next;

      /* Free up the key and value: */
      IW_Free(node->key);
      if (this->discard_f)
	(*(this->discard_f))(node->value);
      
      /* Free the node, go on to the next one: */
      IW_Free(node);
      node = next;
    }
  }

  /* Free the buckets themselves: */
  IW_Free(this->buckets);
}


/************************************************************
 * IW_HashtFree -- deallocator
 ************************************************************
 *
 * DESCRIPTION
 *    Free up a hashtable, and destroy its contents.
 */ 

void IW_HashtFree(this)
     IW_Hasht *this; /* the hashtable */
{
  if (this) {
    IW_HashtDestroy(this);
    IW_Free(this);
  }
}


/************************************************************
 *
 * Element storage/retrieval
 *
 ************************************************************/


/************************************************************
 * IW_HashtAtGet -- lookup an element
 ************************************************************
 * 
 * DESCRIPTION
 *    Given the address of a pointer, change the pointer so that it points
 *    to the "value" in the hashtable indexed by the "key".  If the key is 
 *    not found in the hashtable, set the pointer to NULL.
 *
 * RETURNS
 *    IW_HashtOK if key was found in table, IW_HashtENOTFOUND if it was not.
 *
 * NOTES
 *    If you will never store a NULL object in the hashtable, IW_HashtAt() 
 *    will usually be more convenient.
 */

int IW_HashtAtGet(this, key, valuep) 
     IW_Hasht *this;    /* the hashtable */
     char *key;        /* the lookup key */
     IW_Elem *valuep;   /* where to store the retrieved element */
{
  IW_HashtNode *node;

  /* Get the node in the hash table, or NULL: */
  node = IW_HashtNodeAt(this, key);   
  
  /* Return whether the value was missing or not: */
  if (node != NULL) {        /* key in table: */
    *valuep = node->value;
    return IW_HashtOK;
  }
  else {                     /* key not in table: */
    *valuep = NULL;
    IW_Err(0, "key not in table");
    return IW_HashtENOTFOUND;
  }
}


/************************************************************
 * IW_HashtAt -- lookup an element
 ************************************************************
 * 
 * DESCRIPTION
 *    Given a key, return the value stored at that key, or NULL
 *    if no such value exists.  Note that there is no way to
 *    distinguish between a NULL value and a nonexistent key;
 *    use IW_HashtAtGet() if you need to make that distinction.
 *
 * RETURNS
 *    The value stored at the given key, or NULL if the key is not
 *    in the table.
 */

IW_Elem IW_HashtAt(this, key)
     IW_Hasht *this; /* the hashtable */
     char *key;     /* the lookup key */
{
  IW_Elem value;

  if (IW_HashtAtGet(this, key, &value) == IW_HashtOK) {
    return value;
  }
  else {
    return NULL;
  }
}


/************************************************************
 * IW_HashtAtPut -- store an element
 ************************************************************
 *
 * DESCRIPTION
 *    Store an element under the given key.  If an element under
 *    that key already exists, discard the old element in favor 
 *    of the new one.
 *
 * RETURNS
 *    IW_HashtOK on success, <IW_HashtOK on failure.
 */

int IW_HashtAtPut(this, key, value)
     IW_Hasht *this; /* the hashtable */
     char *key;     /* the lookup key */
     IW_Elem value;  /* the value to store */
{
  IW_HashtNode *node;
                                                  /* Key not found: */
  if ((node = IW_HashtNodeAt(this, key)) == NULL) {
    if (IW_HashtInstall(this, key, value) == NULL) {  /* install failed! */
      IW_Err(0, "couldn't install node");
      return IW_HashtEMISC;
    }
  }
                                                  /* Key found: */
  else {                                                
    if (this->discard_f) {
      (*(this->discard_f))(node->value);               /* free old value */
    }
    node->value = value;                               /* store new value */
  }

  /* Return success: */
  return IW_HashtOK;
}


/************************************************************
 * IW_HashtAtDiscard -- remove and destroy an element
 ************************************************************
 * 
 * DESCRIPTION
 *    Discard an entry in the hashtable at the given key, according
 *    to the hashtable's discard function.
 *
 * RETURNS
 *    IW_HashtOK on success, <IW_HashtOK on failure.
 */

int IW_HashtAtDiscard(this, key)
     IW_Hasht *this; /* the hashtable */
     char *key;     /* the lookup key */
{
  IW_HashtNode *node;
  int bucket_number;
  IW_HashtNode *pred;

  /* Find the backet number: */
  bucket_number = (*(this->hash_f))(key) % this->num_buckets; 

  /* Find the PREDECESSOR of the node to be deleted: */
  pred = NULL;
  node = this->buckets[bucket_number];  
  while (node) {

    /* Is this it? */
    if ((*(this->compare_f))(key, node->key)) {
      break;
    }

    /* Next! */
    pred = node;
    node = node->next;
  }

  /* If the node is NULL, then we failed to find the given node.
     Return an error: */
  if (!node) {            /* no such node! */
    IW_Err(0, "no such hashtable node");
    return IW_HashtEMISC;
  }

  /* Pred is either the node's predecessor node, or NULL (meaning that
     the node is the first node of the bucket).  Excise the node from
     the table: */  
  if (!pred)             /* node is first in bucket */
    this->buckets[bucket_number] = node->next;
  else                   /* pred is node befgore node to be deleted: */
    pred->next = node->next;
  
  /* Free the node: */
  IW_Free(node->key);
  if (this->discard_f != NULL)
    (*(this->discard_f))(node->value);
  IW_Free(node);

  /* Done! */
  return IW_HashtOK;
}



/************************************************************
 *
 * Iterator functions
 *
 ************************************************************
 *
 * All container Pseudoclasses try to support container-specific
 * "iterator" types which have the same general syntax and usage.
 * For example, if 'EgBag' is a container Pseudoclass type and you
 * wanted to print out the elements of 'mybag', you would do so
 * as follows:
 *
 *    1.  Declare 'bi' as a EgBagIter.
 *    2.  Call EgBagIRewind() to rewind 'bi' to the beginning of 'mybag'.
 *    3.  Keep calling EgBagINext() to get the next element out of the
 *        bag... it will return:
 *          - EgBagOGO if an element was actually retrieved
 *          - EgBagOSTOP if no element was retrieved because the end of
 *            the collection was reached.
 *          - values less than EgBagOK on error.
 *
 * Repeated calls to EgBagINext() are okay when OSTOP is returned;
 * the subsequent calls will just return OSTOP (much like getting
 * characters from a stream which is at "end of file").
 * 
 * The above example describes a "detached iterator", in that the iterator 
 * variable is separate from the container it is iterating over. 
 * In the case of a IW_Hasht, the "detached iterator" code looks like:
 *
 *    IW_Hasht      H;             \* a hashtable of objects *\
 *    IW_HashtIter  hi;            \* an iterator *\
 *    char        *key;           \* a key *\
 *    Object      *value;         \* an associated value *\
 *    ...
 *    IW_HashtIRewind(&H, &hi);        \* rewind to begin *\
 *    while (IW_HashtINext(&H, &hi, &key, &value) == IW_HashtOGO) {
 *        \* do stuff with key and/or value *\
 *    }
 *
 * Some container Pseudoclasses may support "attached iterators",
 * where every container has its own internal iterator that you can
 * use (much like a FILE's internal file pointer in stdio routines).
 * The notation is simpler, and does not require the declaration of
 * an additional varible.  In the case of a IW_Hasht, the "attached
 * iterator" code looks like:
 *
 *    IW_Hasht      H;             \* a hashtable of objects *\
 *    char        *key;           \* a key *\
 *    Object      *value;         \* an associated value *\
 *    ...
 *    IW_HashtRewind(&H);        \* rewind to begin *\
 *    while (IW_HashtNext(&H, &key, &value) == IW_HashtOGO) {
 *        \* do stuff with key and/or value *\
 *    }
 *
 */



/************************************************************
 * IW_HashtIRewind -- rewind a hashtable element "stream"
 ************************************************************
 *
 * DESCRIPTION
 *    Rewind a detached iterator to the beginning of a hashtable.
 *
 * RETURNS
 *    IW_HashtOK always.
 */

int IW_HashtIRewind(this, i)
     IW_Hasht *this;     /* the hashtable */
     IW_HashtIter *i;    /* pointer to the "iterator" over the table */
{
  /* Set the iterator to point at the first element of the first bucket
     of the hashtable, and keep moving forwards through the buckets (and
     the nodes in the buckets) until a node with an element is found: */
  for (i->bucket = 0; i->bucket < this->num_buckets; i->bucket++) {
    if (this->buckets[i->bucket] != NULL) {
      i->node = this->buckets[i->bucket];
      return IW_HashtOK;
    }
  }

  /* If this point is reached, then the hashtable is empty.  The iterator
     already points to the "end of table", since this condition is signalled
     by a bucket number which is greater than the last bucket in the 
     table.  All that is left is to return appropriately: */
  return IW_HashtOK;
}


/************************************************************
 * IW_HashtINext -- get next element from hashtable "stream"
 ************************************************************
 *
 * DESCRIPTION
 *    Given a hashtable and in iterator pointing into it, get
 *    the next element stored in the hashtable.
 *
 * ARGUMENTS
 *    'this' is a pointer to a IW_Hasht.
 *
 *    'i' is a pointer to a IW_HashtIter, pointing into 'this'.
 *
 *    'keyp' is a pointer to a character-pointer, which on successful
 *      exit will be set to the next key in the table... you may
 *      give this argument as NULL, in which case it will be ignored.
 *
 *    'valuep' is a pointer to an element-pointer, which on successful
 *      exit will be set to the next value in the table... you may
 *      give this argument as NULL, in which case it will be ignored.
 *
 * ALGORITHM
 *    At the entry to the function, the iterator either points to an
 *    element of the hashtable, or else it signals an end-of-table
 *    condition:
 *
 *       If it points to an element, set the object pointer to that 
 *       element and move forward to the next element (if one exists), 
 *       returning IW_HashtOGO.
 * 
 *       If it signals end-of-table, set the object pointer to NULL
 *       and return IW_HashtOSTOP.
 *
 *
 * RETURNS
 *    IW_HashtOGO if a "next" element was retrieved, and IW_HashtOSTOP
 *    if none remain.
 */

int IW_HashtINext(this, i, keyp, valuep) 
     IW_Hasht *this;   /* the hashtable */
     IW_HashtIter *i;  /* pointer to the "iterator" over the table */
     char **keyp;     /* where to put the pointer to the "next" key */
     IW_Elem *valuep;  /* where to put the "next" value */
{
  /* If the iterator is at the end of the hashtable, set the object
     to NULL and return IW_HashtOSTOP: */
  if (i->bucket >= this->num_buckets) {   /* bad bucket signals "end": */
    if (keyp) 
      *keyp = NULL;
    if (valuep) 
      *valuep = NULL;
    return IW_HashtOSTOP;
  }
  
  /* At this point, the iterator points to a valid entry.  Set the object
     pointer: */
  if (keyp) 
    *keyp = i->node->key;
  if (valuep) 
    *valuep = i->node->value;


  /* 
   * Now, move forward to the next object in the table (in preparation for 
   * the next call to this function)... 
   */
  

  /* First, try to go to the next object in this bucket: */
  if ((i->node = i->node->next) != NULL) {  /* next element in bucket exists */
    return IW_HashtOGO;
  }
  /* No next element: go through buckets until a non-empty bucket is
     found, or the table end is reached... this non-empty bucket's
     first element will be the next element that the iterator returns: */
  else {
    for (++i->bucket; i->bucket < this->num_buckets; ++(i->bucket)) {
      if (this->buckets[i->bucket] != NULL) {
	i->node = this->buckets[i->bucket];
	break;    /* found the next element in the table! */
      }
    }
  }
  
  /* At this point, the iterator either points to the next element, or
     is past the last bucket (an end-of-table condition): */
  return IW_HashtOGO;
}


/************************************************************
 *
 * Other stuff:
 *
 ************************************************************/


/************************************************************
 * IW_HashtDo -- do action for each element in hashtable
 ************************************************************
 *
 * DESCRIPTION
 *    Perform an action with/on each element stored in the hashtable. 
 *    The action should be a function declared as:
 *
 *|       int action(char *key, IW_Elem value, IW_Arg arg);
 *  
 *    This action will be invoked for every element in the hashtable;
 *    when it is called, 'key' will be the current key, 'value' will
 *    be the element stored at that key, and 'arg' will be the same
 *    'arg' passed into the call to IW_HashtDo(). 
 *
 *    The action should return 0; any other action will cause the
 *    iteration to be halted, and IW_HashtEHALTED to be returned.
 * 
 * ARGUMENTS
 *    The 'this' is a pointer to a IW_Hasht.
 *
 *    The 'action' is a function, as described above.
 *
 *    The 'arg' is a "callback argument", which is sent on to each
 *       invocation of 'action'.  This allows you to get extra data
 *       from the outside world into the action without having to 
 *       use global variables.
 *
 * RETURNS
 *    IW_HashtOK if iterated through all elements, IW_HashtEHALTED 
 *    if halted.
 */

int IW_HashtDo(this, action, arg)
     IW_Hasht *this;      /* the hashtable */
     int (*action)(char *, IW_Elem, IW_Arg);   /* the action */
     IW_Arg arg;          /* callback argument */
{
  int i;
  IW_HashtNode *node;
  int halted = 0;

  /* For each bucket: */
  for (i = 0; ((i < this->num_buckets) && (!halted)); i++) {
    node = this->buckets[i];
    while ((node != NULL) && (!halted)) {

      /* Execute function: */
      halted |= (*action)(node->key, node->value, arg);

      /* Go to next node: */
      node = node->next;
    }
  }

  /* Return whether or not computation was halted: */
  if (halted) {
    IW_Err(0, "halted by callback");
    return IW_HashtEHALTED;
  }
  return IW_HashtOK;
}


/******************** end of file "IW_Hasht.c" ********************/
