/************************************************************
 * IW_Hasht.h
 ************************************************************
 *
 * DESCRIPTION
 *    General-purpose hashtable Pseudoclass.
 *
 * USAGE
 *    Suppose we have a type 'Thing' defined as follows:
 *   
 *         typedef struct {
 *            char   *name;
 *            int    age;
 *            double weight;
 *         } Thing;
 *   
 *    And suppose we wish to hash a bunch of 'Thing's on their 'name' field 
 *    in a case-insensitive manner.  Here's how to do it using 'IW_Hasht'...
 *   
 *    First, define a free function for Things, returning int.  It must 
 *    destroy any space malloc'ed inside the Thing, and finally destroy 
 *    the Thing itself:
 *   
 *        int thing_free(Thing *   doomed) 
 *        {     
 *           free(doomed->name);
 *           free(doomed);
 *        }
 *   
 *    Now, you may create a hashtable in the following manner:
 *   
 *        IW_Hasht thingTable;
 *   
 *        IW_HashtCreate (
 *            &thingTable,                \* the table *\
 *            211,                        \* number of buckets *\
 *            IW_HashtHashPJWCI,          \* case-insensitive hash func *\
 *            c_i_strequal,               \* case-insensitive strequal *\
 *            thing_free);                \* how to kill a Thing, or NULL *\
 *   
 *    or, more simply:
 *   
 *        IW_Hasht thingTable;
 *   
 *        IW_HashtCreateCI (
 *            &thingTable,                \* the table *\
 *            211,                        \* number of buckets *\
 *            thing_free);                \* how to kill a Thing, or NULL *\
 *   
 *    To add a Thing (after you have malloc'ed space for the Thing and its 
 *    name, and initiallized it):
 *   
 *         IW_HashtAtPut(
 *              &thingTable,          \* the table *\
 *              newThing->name,       \* key (a fresh copy will be made) *\
 *              &newThing);           \* value (pointer to malloc'ed Thing) *\
 *   
 *    Find a thing by name:
 *   
 *         Thing *aThingPtr; 
 *   
 *         IW_HashtAtGet(
 *              &thingTable,              \* the table *\
 *              name,                     \* the name *\
 *              &aThingPtr);              \* pointer to pointer to Thing *\
 *    or:
 *         aThingPtr = IW_HashtAt(
 *              &thingTable,              \* the table *\
 *              name);                    \* the name *\
 *   
 *    Destroy a hashtable, and all Things inside it (calls the destructor
 *    for each Thing, and for each key, if necessary):
 *   
 *        IW_HashtDestroy(&thingTable);
 *   
 *    Important note: the hash function should just return any old number:
 *    it is up to the internal hash table functions to map this into a 
 *    bucket, which it does simply by taking (hashvalue % num_buckets).
 *    This setup is necessary to allow for a changing number of buckets
 *    as the hashtable grows. 
 *    
 * NAMESPACE
 *    Constants, types, functions and macros beginning with "IW_Hasht".
 *   
 * AUTHOR
 *    Eryq (Erik Dorfman)
 *    Hughes STX Corporation
 *    Thu, 24 Sep 1992
 *   
 * MODIFICATIONS
 *    Thu, 24 Sep 1992 [eryq]
 *    Release 1.0, under old Pseudoclass standard.
 *    
 *    Tue, 16 Feb 1993 [eryq]
 *    Release 2.0, under new "Pc" Pseudoclass standard for reuse library.
 *    
 *    Tue, 23 Feb 1993 [eryq]
 *    Release 2.1.  Fixed bug in AllocateBuckets().
 *
 *    Thu, 15 Apr 1993
 *    Release 2.2.  Added IW_HashtNumElems() macro, and removed (bad)
 *    semicolons from Rewind() and Next() macros.
 *
 * RCS INFO
 *    $Id: IW_Hasht.h,v 1.3 1996/04/17 21:16:16 imswww Exp $
 */

#ifndef __IW_Hasht_h__
#define __IW_Hasht_h__

#include "IW.h"       /* conditional-compilation macros, etc. */
#include "IW_Types.h"  



/************************************************************
 *
 * Constants
 *
 ************************************************************/

/* Return values: */
#define IW_HashtOSTOP        2  /* end of collection */
#define IW_HashtOGO          1  /* got elem while iterating: continue */
#define IW_HashtOK           0
#define IW_HashtEMISC       -1  /* misc error */
#define IW_HashtENOTFOUND   -2  /* element not in collection */
#define IW_HashtESYS        -3  /* error in system call: see errno */
#define IW_HashtEHALTED     -4  /* iteration halted */


/************************************************************
 *
 * Types
 *
 ************************************************************/

/* 
 * An "iterator" for traversing a hashtable... you can declare and use
 * your own iterators, or you can use the built-in one: 
 */
typedef struct {
  int bucket;
  struct IW_HashtNode *node;
} IW_HashtIter;

/*
 * Function which hashes a key: 
 */
typedef int (*IW_HashtFnHash) (char *);

/*
 * Function which compares two keys:
 */
typedef int (*IW_HashtFnCompare) (char *, char *);

/* 
 * A hashtable: 
 */
typedef struct {
  int num_entries;        /* number of entries in the hash table */
  int num_buckets;        /* number of buckets in the hash table */
  struct IW_HashtNode **buckets;  /* the buckets */
  IW_HashtFnHash    hash_f;       /* the hash function */
  IW_HashtFnCompare compare_f;    /* the function which compares two keys */
  IW_FnDiscard discard_f;         /* the function which frees an element */
  int max_avg_bucket_size;        /* signals when to grow table */
  IW_HashtIter i;                 /* for iteration over elements in table */
} IW_Hasht;


/************************************************************
 *
 * Functions
 *
 ************************************************************/

/* Hashing functions: */
extern int IW_HashtHashPJW   (char *);
extern int IW_HashtHashPJWCI (char *); 


/* Creation/destruction: */
extern int IW_HashtCreate       (IW_Hasht *,
				 int num_buckets, 
				 IW_HashtFnHash hash_f,
				 IW_HashtFnCompare compare_f,
				 IW_FnDiscard discard_f);
extern int IW_HashtCreateCS     (IW_Hasht *,
				 int num_buckets,
				 IW_FnDiscard discard_f);
extern int IW_HashtCreateCI     (IW_Hasht *,
				 int num_buckets,
				 IW_FnDiscard discard_f);
extern void IW_HashtDestroy     (IW_Hasht *);


/* Allocation/deallocation: */
extern IW_Hasht *IW_HashtAlloc   (int num_buckets, 
				  IW_HashtFnHash hash_f,
				  IW_HashtFnCompare compare_f,
				  IW_FnDiscard discard_f);
extern IW_Hasht *IW_HashtAllocCS (int num_buckets,
				  IW_FnDiscard discard_f);
extern IW_Hasht *IW_HashtAllocCI (int num_buckets,
				  IW_FnDiscard discard_f);
extern void     IW_HashtFree    (IW_Hasht *);


/* Storage and retrieval: */
extern int     IW_HashtAtPut      (IW_Hasht *, char *, IW_Elem);
extern int     IW_HashtAtGet      (IW_Hasht *, char *, IW_Elem *);
extern IW_Elem IW_HashtAt         (IW_Hasht *, char *);
extern int     IW_HashtAtRemove   (IW_Hasht *, char *, IW_Elem *);
extern int     IW_HashtAtDiscard  (IW_Hasht *, char *);


/* Iteration: */
extern int IW_HashtDo    (IW_Hasht *,
			  int (*func)(char *, IW_Elem, IW_Arg),
			  IW_Arg arg);


/* Iterator-driven iteration: */
extern int IW_HashtIRewind (IW_Hasht *, IW_HashtIter *);
extern int IW_HashtINext   (IW_Hasht *, IW_HashtIter *,
			    char **keyp,
			    IW_Elem *valuep);


/************************************************************
 * IW_HashtRewind
 * IW_HashtNext
 ************************************************************
 *
 * These iteration macros don't require you to declare an 
 * iterator; they use the Hashtable's built-in iterator.  You
 * can use them to write code like this:
 *
 *    char *keyp;
 *    IW_Elem valuep;
 *    ...
 *    IW_HashtRewind(&mytable);        \* rewind to begin *\
 *    while (IW_HashtNext(&mytable, &keyp, &valuep) == IW_HashtOGO) {
 *        \* do stuff with keyp and valuep *\
 *    }
 *
 */
#define IW_HashtRewind(this) \
  IW_HashtIRewind((this), &((this)->i))

#define IW_HashtNext(this, keyp, valuep) \
  IW_HashtINext((this), &((this)->i), (keyp), (valuep))


/************************************************************
 * IW_HashtNumElems
 ************************************************************
 * 
 * Return the number of elements stored in the hashtable.
 * This is purposely NOT called 'IW_HashtSize()', because of
 * the ambiguity of the meaning of "size" in the context
 * of a hashtable (e.g., num elems? num buckets? num nodes?).
 */ 
#define IW_HashtNumElems(this) \
  ((this)->num_entries)



#endif     /* __IW_Hasht_h__ */

/******************** end of file "IW_Hasht.h" ********************/
