/************************************************************
 * IW_Key.c
 ************************************************************
 *
 * DESCRIPTION
 *    See IW_Key.h for details.
 */

#ifndef lint
static char RCSID[] =
   "$Id: IW_Key.c,v 1.2 1997/03/14 17:20:10 eryq Exp $";
#endif /* lint */

#include <stdio.h>
#include <stdlib.h>

#include "IK_Dataset.h"
#include "IK_Granule.h"
#include "IK_Map.h"

#include "IW_Key.h"
#include "IW_Log.h"


/************************************************************
 * IW_ArchiveTagToWeb
 ************************************************************
 * Convert archive tag to text, and return a pointer to the
 * buffer containing the text.
 *
 * If 'web' is not NULL, it is assumed to be the buffer that is
 * copied into and then returned.  Else, a static areas is used.
 */ 

char *IW_ArchiveTagToWeb(IK_ArchiveTag atag,
			 char *web)
{
  static char web_s[IW_Key_MAX_ATAG+1];
  web = (web ? web : web_s);
  sprintf(web, "%d", *((int *)atag));
  return(web);
}

/************************************************************
 * IW_ArchiveTagFromWeb
 ************************************************************
 * Create archive tag from text.
 */ 

void IW_ArchiveTagFromWeb(IK_ArchiveTag atag, 
			  char *web)
{
  int atag_i;
  sscanf(web, "%d", &atag_i);
  memcpy((void *)atag, (void *)&atag_i, sizeof(IK_ArchiveTag));
}

/************************************************************
 * IW_DatasetTagToWeb
 ************************************************************
 * Convert dataset tag to text, and return a pointer to the
 * buffer containing the text.
 *
 * If 'web' is not NULL, it is assumed to be the buffer that is
 * copied into and then returned.  Else, a static areas is used.
 */ 

char *IW_DatasetTagToWeb(IK_DatasetTag dtag,
			 char *web)
{
  static char web_s[IW_Key_MAX_DSTAG+1];
  web = (web ? web : web_s);
  sprintf(web, "%d", *((int *)dtag));
  return(web);
}

/************************************************************
 * IW_DatasetTagFromWeb
 ************************************************************
 * Create dataset tag from text.
 */ 

void IW_DatasetTagFromWeb(IK_DatasetTag dtag, 
			  char *web)
{
  int dtag_i;
  sscanf(web, "%d", &dtag_i);
  memcpy((void *)dtag, (void *)&dtag_i, sizeof(IK_DatasetTag));
}



/************************************************************
 *
 * DATASET WEB KEYS...
 *
 ************************************************************/



/************************************************************
 * IW_DatasetIdPartsToWeb
 ************************************************************
 * Convert archive and dataset ids to "1/234" text, 
 * and return a pointer to the buffer containing the text.
 * Returns NULL on failure.
 *
 * If 'web' is not NULL, it is assumed to be the buffer that is
 * copied into and then returned.  Else, a static area is used.
 */ 

char *IW_DatasetIdPartsToWeb(char *archid, /* archive id, like "GSFC" */
			     char *dsetid, /* dataset id, like "COOL STUFF" */
			     char *web) 
{
  IK_ArchiveTag *atagp;
  IK_DatasetTag *dtagp;

  /* Get archive and dataset tags: */
  if (!(atagp = IK_MapArchiveNameToTag(archid))) {
    IW_LogErr("can't map archive %s to tag!", archid); return NULL;
  }
  if (!(dtagp = IK_MapDatasetNameToTag(dsetid))) {
    IW_LogErr("can't map dataset %s to tag!", dsetid); return NULL;
  }

  /* Map, cleanup, and return REAL location: */
  web = IW_DatasetKeyPartsToWeb(*atagp, *dtagp, web);
  free(atagp);
  free(dtagp);
  return(web);
}

/************************************************************
 * IW_DatasetKeyPartsToWeb
 ************************************************************
 * Convert an archive and dataset tags to "1/234" text, 
 * and return a pointer to the buffer containing the text.
 *
 * If 'web' is not NULL, it is assumed to be the buffer that is
 * copied into and then returned.  Else, a static area is used.
 */ 

char *IW_DatasetKeyPartsToWeb(IK_ArchiveTag atag,
			      IK_DatasetTag dtag,
			      char *web) 
{
  static char web_s[IW_Key_MAX_DSKEY+1];
  web = (web ? web : web_s);
  sprintf(web, "%s/%s", 
	  IW_ArchiveTagToWeb(atag, NULL),
	  IW_DatasetTagToWeb(dtag, NULL));
  return(web);
}

/************************************************************
 * IW_DatasetKeyToWeb
 ************************************************************
 * Convert dataset sort key to "1/234" text, 
 * and return a pointer to the buffer containing the text.
 *
 * If 'web' is not NULL, it is assumed to be the buffer that is
 * copied into and then returned.  Else, a static area is used.
 */

char *IW_DatasetKeyToWeb(IK_SortKey *key,
			 char *web)
{
  IK_ArchiveTag atag;
  IK_DatasetTag dtag;

  IK_StripDatasetKey(atag, dtag, key);
  return(IW_DatasetKeyPartsToWeb(atag, dtag, web));
}

/************************************************************
 * IW_DatasetKeyFromWeb
 ************************************************************
 * Create dataset sort key from "1/234" text.
 * Returns a malloc'ed sort key: it's up to the caller to free it!
 */ 

IK_SortKey *IW_DatasetKeyFromWeb(char *web)
{
  IK_ArchiveTag atag;
  IK_DatasetTag dtag;
  int atag_i, dtag_i;
  char atag_s[IW_Key_MAX_ATAG+1], dtag_s[IW_Key_MAX_DSTAG+1]; 

  /* Extract components: */
  sscanf(web, "%d/%d", &atag_i, &dtag_i);

  /* Create archive tag: */
  sprintf(atag_s, "%d", atag_i);
  IW_ArchiveTagFromWeb(atag, atag_s);

  /* Create dataset tag: */
  sprintf(dtag_s, "%d", dtag_i);
  IW_DatasetTagFromWeb(dtag, dtag_s);

  /* Build sort key, and return: */
  return IK_BuildDatasetKey(atag, dtag);
}



/************************************************************
 *
 * GRANULE WEB KEYS...
 *
 ************************************************************/



/************************************************************
 * IW_GranuleIdPartsToWeb
 ************************************************************
 * Convert archive, dataset, and granule ids to "1/234/GID" text, 
 * and return a pointer to the buffer containing the text.
 * Returns NULL if no such mapping could be done.
 *
 * If 'web' is not NULL, it is assumed to be the buffer that is
 * copied into and then returned.  Else, a static area is used.
 */ 

char *IW_GranuleIdPartsToWeb(char *archid, /* archive id, like "GSFC" */
			     char *dsetid, /* dataset id, like "COOL STUFF" */
			     char *gid,    /* granule id */
			     char *web) 
{
  IK_ArchiveTag *atagp;
  IK_DatasetTag *dtagp;

  /* Get archive and dataset tags: */
  if (!(atagp = IK_MapArchiveNameToTag(archid))) {
    IW_LogErr("can't map archive %s to tag!", archid); return NULL;
  }
  if (!(dtagp = IK_MapDatasetNameToTag(dsetid))) {
    IW_LogErr("can't map dataset %s to tag!", dsetid); return NULL;
  }

  /* Map, cleanup, and return REAL location: */
  web = IW_GranuleKeyPartsToWeb(*atagp, *dtagp, gid, web);
  free(atagp);
  free(dtagp);
  return(web);
}

/************************************************************
 * IW_GranuleKeyPartsToWeb
 ************************************************************
 * Convert an archive tag, dataset tag, and gid to "1/234/GID" text, 
 * and return a pointer to the buffer containing the text.
 *
 * If 'web' is not NULL, it is assumed to be the buffer that is
 * copied into and then returned.  Else, a static area is used.
 */ 

char *IW_GranuleKeyPartsToWeb(IK_ArchiveTag atag,
			      IK_DatasetTag dtag,
			      char *gid,
			      char *web) 
{
  static char web_s[IW_Key_MAX_GKEY+1];
  web = (web ? web : web_s);
  sprintf(web, "%s/%s/%s",
	  IW_ArchiveTagToWeb(atag,NULL), IW_DatasetTagToWeb(dtag,NULL), gid);
  return(web);
}

/************************************************************
 * IW_GranuleKeyToWeb
 ************************************************************
 * Convert granule sort key to "1/234/GID" text,
 * and return a pointer to the buffer containing the text.
 *
 * If 'web' is not NULL, it is assumed to be the buffer that is
 * copied into and then returned.  Else, a static area is used.

 */ 

char *IW_GranuleKeyToWeb(IK_SortKey *key,
			 char *web)
{
  IK_ArchiveTag atag;
  IK_DatasetTag dtag;
  char gid[IK_GRANULE_ID_SZ+1];

  IK_StripGranuleKey(atag, dtag, gid, key);
  return(IW_GranuleKeyPartsToWeb(atag, dtag, gid, web));
}

/************************************************************
 * IW_GranuleKeyFromWeb
 ************************************************************
 * Create granule sort key from "1/234/GID" text.
 * Returns a malloc'ed sort key: it's up to the caller to free it!
 */ 

IK_SortKey *IW_GranuleKeyFromWeb(char *web)
{
  IK_ArchiveTag atag;
  IK_DatasetTag dtag;
  int atag_i, dtag_i;
  char atag_s[IW_Key_MAX_ATAG+1], dtag_s[IW_Key_MAX_DSTAG+1]; 
  char gid[IK_GRANULE_ID_SZ+1];

  /* Extract components: */
  sscanf(web, "%d/%d/%s", &atag_i, &dtag_i, gid);

  /* Create archive tag: */
  sprintf(atag_s, "%d", atag_i);
  IW_ArchiveTagFromWeb(atag, atag_s);

  /* Create granule tag: */
  sprintf(dtag_s, "%d", dtag_i);
  IW_DatasetTagFromWeb(dtag, dtag_s);

  /* Build sort key, and return: */
  return IK_BuildGranuleKey(atag, dtag, gid);
}

/******************** end of file "IW_Key.c" ********************/
