/************************************************************
 * IW_Log.c
 ************************************************************
 * See IW_Log.h for details.
 */

#ifndef lint
static char RCSID[] =
   "$Id: IW_Log.c,v 1.2 1996/12/10 04:16:12 eryq Exp $";
#endif /* lint */


#include <stdio.h>
#include <stdarg.h>
#include <string.h>

#include "IK_Syslog.h"

#include "IW_Log.h"


/************************************************************
 *
 * Globals
 *
 ************************************************************/

/* The message verbosity.
 * Messages of 'level' and above will be logged to the stderr.
 * Default causes nothing to be logged.
 * Notice that syslog logging is controlled by a separate mechanism.
 */
static int Verbosity = -1; 

/* The program name: */
static char *ProgName = NULL;


/************************************************************
 * priority_glyph
 ************************************************************
 * Return the leading glyph for the given priority.
 */ 
static char *priority_glyph(int priority)
{
  switch (priority) {
  case LOG_DEBUG:   return "D   ";
  case LOG_INFO:    return "II  " ;
  case LOG_NOTICE:  return "NN  " ;
  case LOG_WARNING: return "WWW " ;
  case LOG_ERR:     return "EEEE" ;
  case LOG_EMERG:	
  case LOG_ALERT:
  case LOG_CRIT:    return "!!!!";
  default:          return "????" ;
  }
}

/************************************************************
 * priority_name
 ************************************************************
 * Return the name of the given priority.
 */ 
static char *priority_name(int priority)
{
  switch (priority) {
  case LOG_DEBUG:   return "debug";
  case LOG_INFO:    return "info" ;
  case LOG_NOTICE:  return "note" ;
  case LOG_WARNING: return "warn" ;
  case LOG_ERR:     return "error";
  case LOG_EMERG:	
  case LOG_ALERT:
  case LOG_CRIT:    return "serious error";
  default:          return "ERROR";
  }
}

/************************************************************
 * vmessage
 ************************************************************
 * Private.  Output a message, both to the IK_Syslog and
 * to the given file pointer.
 */ 
static void vmessage(FILE *fp, int priority, char *fmt, va_list args)
{
  char msg[2048];
 
  /* Format for the syslog: */
  vsprintf(msg, fmt, args);
  IK_Syslog(priority, msg);

  /* Output to stream: trap against the logging level: */
  if ((Verbosity == -1) || (priority > Verbosity)) return;    
  if (ProgName) fprintf(fp, "%s: ", ProgName);  
  fprintf(fp, "%s %s: ", priority_glyph(priority), priority_name(priority));
  vfprintf(fp, fmt, args);
  fprintf(fp, "\n");
}

/************************************************************
 * IW_LogDebug
 ************************************************************
 * Output debugging info.  Use like printf.
 */ 
void IW_LogDebug(char *fmt, ...)
{
  va_list args;
  va_start(args, fmt); 
  vmessage(stderr, LOG_DEBUG, fmt, args); 
  va_end(args);
}

/************************************************************
 * IW_LogInfo
 ************************************************************
 * Output an informational message.  Use like printf.
 */ 
void IW_LogInfo(char *fmt, ...)
{
  va_list args;
  va_start(args, fmt); 
  vmessage(stderr, LOG_INFO, fmt, args); 
  va_end(args);
}

/************************************************************
 * IW_LogNotice
 ************************************************************
 * Output a notice message.  Use like printf.
 */ 
void IW_LogNotice(char *fmt, ...)
{
  va_list args;
  va_start(args, fmt); 
  vmessage(stderr, LOG_NOTICE, fmt, args); 
  va_end(args);
}

/************************************************************
 * IW_LogWarning
 ************************************************************
 * Output a warn message.  Use like printf.
 */ 
void IW_LogWarning(char *fmt, ...)
{
  va_list args;
  va_start(args, fmt); 
  vmessage(stderr, LOG_WARNING, fmt, args); 
  va_end(args);
}

/************************************************************
 * IW_LogErr
 ************************************************************
 * Output a error message.  Use like printf.
 * Note that it returns a convenient "error" value, so you can say:
 *
 *     return IW_LogErr("can't open %s", filename);
 */ 
int IW_LogErr(char *fmt, ...)
{
  va_list args;
  va_start(args, fmt); 
  vmessage(stderr, LOG_ERR, fmt, args); 
  va_end(args);
  return -1;
}



/************************************************************
 * IW_LogSetProgramName
 ************************************************************
 * Set the program name.  The string must persist!
 */ 
void IW_LogSetProgramName(char *name)
{
  ProgName = name;
}

/************************************************************
 * IW_LogSetVerbosity
 ************************************************************
 * Set the verbosity.
 * Messages of priority 'level' and lower will be logged.
 * 0 logs everything, -1 logs nothing.
 */ 
void IW_LogSetVerbosity(int level)
{
  Verbosity = level;
}



/******************** end of file "IW_Log.c" ********************/




