/*************************************************************
 *
 * NAME
 *    IW_ODL - additional routines for dealing with ODL
 *
 * AUTHOR
 *    Eryq (Erik Dorfman)
 *    Hughes STX Corporation
 *
 *************************************************************/

#ifndef lint
static char RCSID[] =
   "$Id: IW_ODL.c,v 1.4 1997/09/03 18:44:51 solomon Exp $";
#endif /* lint */

#include <stdio.h>
#include <string.h>

#include "IW_ODL.h"
#include "IW_Str.h"
#include "IW_New.h"
#include "IW_Chunk.h"

/* Set the following true if ODL is plagued with this... */
#define CORRECT_ODL_SLASH_N_BUG 1



/*************************************************************
 * IW_ODLReadFile
 *************************************************************
 *
 * DESCRIPTION
 *    This function will read in the ODL label file specified in
 *    parameter, and build an ODL tree from it.   If the "root"
 *    argument is non-NULL, then that wil be used; else, a new
 *    aggregate named "ROOT" will be created and returned.
 *
 * PARAMETERS
 *    AGGREGATE root   the empty root odl tree
 *    char *filename   filename for the odl label file
 * 
 * RETURN VALUES
 *    Return the AGGEGATE, or NULL on failure.
 */ 

AGGREGATE IW_ODLReadFile(char *path, AGGREGATE root)
{
   FILE *fp;

   /* Open the file: */
   if (!(fp = fopen(path, "rb"))) {
     return NULL;
   }
   /* Get the root: */
   if (!root) {
     if (!(root = NewAggregate(NULL, KA_GROUP, "ROOT", ""))) {
       fclose(fp);
       return NULL;
     }
   }
   ReadLabel(fp, root);
   fclose(fp);
   return (root);
}


/************************************************************
 * IW_ODLAddStringParam
 ************************************************************
 * Add a string-valued parameter to the aggregate.
 */ 

int IW_ODLAddStringParam(AGGREGATE agg, char *parmname, char *value_s)
{
  PARAMETER parm;
  VALUE_DATA vdata;

  /* Sanity: */
  if (!agg)      
    return IW_LogErr("can't add string param to NULL aggregate");
  if (!parmname || !*parmname) 
    return IW_LogErr("empty param name");
  if (!value_s)
    return IW_LogErr("NULL value");

  /* Add it: */
  parm = NewParameter(agg, KP_ATTRIBUTE, parmname);
  vdata = ODLConvertString(value_s, strlen(value_s));
  NewValue(parm, &vdata);
  return 0;
}


/************************************************************
 *
 * The next two functions allow us to say...
 *
 *     for (group = IW_ODLFirstChildGroup(parent, "groupname");
 *          group;
 *          group = IW_ODLNextChildGroup(parent, "groupname", group)) {
 *          ...
 *     }
 *
 ************************************************************/      


/************************************************************
 * IW_ODLFirstChildGroup
 ************************************************************
 * Find the first child group of the "parent" aggregate with 
 * the name "groupname".
 *
 * This does NOT recurse!
 */

AGGREGATE IW_ODLFirstChildGroup(AGGREGATE parent, char *groupname)
{
  AGGREGATE node;
  
  /* Sanity: */
  if (!parent || !groupname) 
    return NULL;
  
  /* Find first group of that name: */
  for (node = parent->first_child; node; node = node->right_sibling) {
    if (node->kind == KA_GROUP && strcmp(node->name, groupname) == 0) {
     return node;
    }
  }
  return NULL;
}

/************************************************************
 * IW_ODLNextChildGroup
 ************************************************************
 * Find the next child group of the "parent" aggregate after "previous"
 * with name "groupname".  The parent argument makes the call
 * identical to that of FindNextGroup.
 *
 * This does NOT recurse!
 */

AGGREGATE IW_ODLNextChildGroup(AGGREGATE parent, char *groupname,
			       AGGREGATE previous)
{
  AGGREGATE node;

  /* Sanity: */
  if (!previous || !groupname) 
    return NULL;
  
  /* Return next sibling with that name: */
  for (node = previous->right_sibling; node; node = node->right_sibling) {
    if (node->kind == KA_GROUP && strcmp(node->name, groupname) == 0) {
      return node;
    }
  }
  return NULL;
}

/************************************************************
 * IW_ODLJoinSequence
 ************************************************************
 * Composite a sequence string into a simple string.   
 *
 * Returns an allocated string, or NULL on error.
 */

char *IW_ODLJoinSequence(PARAMETER parm,
			 char *join)   /* join char, or 0 for none */
{
  char *buffer;
  char *s = buffer;
  char **list;
  char **listp;
  int length = 0;
  int n;


  /* Coerce NULL join to empty string: */
  if (!join) {
    join = "";
  }

  /* Convert to list: */
  if (!(list = IW_ODLVectorFromSequence(parm))) {
    return NULL;
  }

  /* Get total length: */
  for (listp = list; *listp; listp++) {
    length += strlen(*listp) + strlen(join);
  }
  
  /* Reallocate: */
  if (!(buffer = (char *)malloc(1 + (length * sizeof(char))))) {
    IW_Die(0, "IW_ODLJoinSequence: out of memory (malloc of %d bytes)!", 
	   length);
  }

  /* Copy: */
  s = buffer;
  *s = '\0';
  for (listp = list; *listp; listp++) {
    /* Copy: */
    sprintf(s, "%s%s", *listp, (listp[1] ? join : ""));
    s += strlen(s);   /* skip to char past end */
  }

  /* Cleanup and return: */
  IW_ODLVectorFree(list);
  return buffer;
}

/************************************************************
 * IW_ODLParamStringified
 ************************************************************
 * Return the parameter's value as a newly-allocated string.
 *
 * Multiple values are joined into a single string, with individual values
 # separated by "join".
 *
 * Single values are dup'ed.
 *
 * Returns NULL on error, or if parameter is NULL.
 */

char *IW_ODLParamStringified(PARAMETER parm,
			     char *join)   /* join char, or 0 for none */
{
  VALUE value;
  char *s;

  /* Trap for null parameter: */
  if (!parm) {
    return NULL;
  }

  switch (parm->value_kind) {

  case KV_SEQUENCE:
    return IW_ODLJoinSequence(parm, join);
    break;

  case KV_SCALAR:
    value = FirstValue(parm);
    if (!(s = IW_ODLValueStringified(value))) return NULL;  
    return IW_Strdup(s);
    break;

  default: 
    IW_Err(0, "Unrecognized VALUE node type: %d", value->item.type);
    return NULL;
  }
}

/************************************************************
 * IW_ODLVectorFromSequence
 ************************************************************
 * A simpler, saner way to convert a sequence-string into a
 * C vector (an array of char *).  
 * Returns an allocated vector of strings, or NULL on error.
 *
 * Call IW_ODLVectorFree() to dispose of the vector.
 */

char **IW_ODLVectorFromSequence(PARAMETER parm)
{
  char **strv = NULL;    /* the vector */
  char **strp = NULL;    /* the current entry */
  int strc = 0;          /* the number of entries */
  int i;
  VALUE value;           /* a string value in the parameter's list */

  /* Sanity: */
  if (!parm) return NULL;

  /* How many entries will we have? */
  if (!(strc = parm->value_count)) return NULL;
  
  /* Allocate space for vector itself, and zero it: */
  if (!(strv = IW_NewArray(char *, (strc+1)))) {
    IW_Die(0, "IW_ODL can't allocate string vector!");
  }
  for (i = 0; i <= strc; i++) { strv[i] = NULL; }   
  
  /* For each entry: */
  for (value = FirstValue(parm), strp = strv; 
       value;
       value = NextValue(value), strp++) {
    if (!(*strp = IW_Strdup(IW_ODLValueStringified(value)))) {
      IW_Die(0, "IW_ODL can't duplicate vector element!");
    }
  }
  *strp = NULL;     /* finish it! */
  
  /* Done! */
  return(strv);
}

/************************************************************
 * IW_ODLVectorFree
 ************************************************************
 * Free the string vector returned by IW_ODLVectorFromSequence.
 */

void IW_ODLVectorFree(char **strv) 
{
  char **strp = NULL;    /* the current entry */

  for (strp = strv; *strp; strp++) {
    IW_Free(*strp);     /* free the entry */
  }
  IW_Free(strv);       /* free the vector! */
}

/************************************************************
 * IW_ODLValueString
 ************************************************************
 * Like value->item.value.string, but safer since it skips the \n.
 * Returns a pointer right into the data structure!
 *
 * NOTE: This does NOT handle sequence strings.
 */

char *IW_ODLValueString(VALUE value)
{
#if CORRECT_ODL_SLASH_N_BUG
  char *start = value->item.value.string;

  if      (strncmp(start, "\\n\n", 3) == 0)  start += 3;
  else if (strncmp(start, "\\n", 2) == 0)    start += 2;
  return start;
#else
  return value->item.value.string;
#endif
}

/************************************************************
 * IW_ODLValueStringified
 ************************************************************
 * Format and return a string for the given scalar value.
 *
 * NOTE: This does NOT handle sequence strings.
 *
 * The return value is either a pointer into a static buffer or into
 * the data structure; copy it quick!
 */

char *IW_ODLValueStringified(VALUE value)
{
  static char buf[1024];
  char *bufp = buf;
  
  switch (value->item.type){
  case TV_INTEGER:
    sprintf(buf, "%d", value->item.value.integer.number);
    break;
    
  case TV_REAL:
    sprintf(buf, "%f", value->item.value.real.number);
    break;
    
  case TV_SYMBOL:
    bufp = IW_ODLValueString(value);
    break;
    
  case TV_STRING:
    bufp = IW_ODLValueString(value);
    break;
	
  case TV_DATE:
    ODLFormatDate(buf, &value->item);
    break;
	
  case TV_TIME:
    ODLFormatTime(buf, &value->item);
    break;
	
  case TV_DATE_TIME:
    ODLFormatDateTime(buf, &value->item);
    break;
    
  default:
    return NULL;
  }
  return bufp;
}

/************************************************************
 * IW_ODLWriteOptsClear
 ************************************************************
 * Dump the ODL to the given file as "chunks".
 */

void IW_ODLWriteOptsClear(IW_ODLWriteOpts *this)
{
  memset(this, '\0', sizeof(*this));
}


/************************************************************
 * IW_ODLWriteChunkFields
 ************************************************************
 * Dump the ODL to the given chunk stream as a "chunk".
 */

int IW_ODLWriteChunkFields(IW_Chunksp *sp,        /* where to output to */
			   AGGREGATE node_agg,    /* top-level group */
			   IW_ODLWriteOpts *opts) /* output opts or NULL */
{
  AGGREGATE  group;    /* Pointer to current node */
  PARAMETER  parm;     /* pointer to current parameter node */
  VALUE      value;    /* pointer to current parameter value */
  char  *vstr;         /* value as string */
  char  *except_params;   /* skip these params */
  char  *except_groups;   /* skip these groups */

  /* Init opts: */
  except_params = (opts ? opts->except_params : "");
  if (!except_params) { except_params = ""; }
  except_groups = (opts ? opts->except_groups : "");
  if (!except_groups) { except_groups = ""; }
  
  
  /* Do nothing if no root: */
  if (!node_agg) return 0;
  
  /* Output all parameters in this aggregate: */
  for (parm = FirstParameter(node_agg);
       parm;
       parm = NextParameter(parm)) {

    /* Skip if desired: */
    if (strstr(except_params, parm->name)) 
      continue;

    /* Output all values of this parameter: */
    if (parm->value_count == 1) {        /* single-valued parameter... */
      if (value = FirstValue(parm)) {
	if (vstr = IW_ODLValueStringified(value)) {
	  IW_ChunkspField(sp, parm->name, "%s", vstr);
	}
      }   
    }
    else if (parm->value_count > 1) {    /* multi-valued parameter... */
      if (0) {                                /* output as params.. */
	for (value = FirstValue(parm); value; value = NextValue(value)) {
	  if (vstr = IW_ODLValueStringified(value)) {
	    IW_ChunkspField(sp, parm->name, "%s", vstr);
	  }
	}
      }
      else {                                  /* auto-fold into string... */
	if (vstr = IW_ODLParamStringified(parm, "\001\n")) {
	  IW_ChunkspField(sp, parm->name, "%s", vstr);
	  IW_Free(vstr);
	}
      }
    }
  }     /* for each param */

  /* Output all child groups of this aggregate: */
  for (group = node_agg->first_child; 
       group; 
       group = group->right_sibling) {

    /* Skip if desired: */
    if (strstr(except_groups, group->name)) 
      continue;

    /* Output: */
    IW_ChunkspBeginGroup(sp, group->name);
    IW_ODLWriteChunkFields(sp, group, opts);
    IW_ChunkspEndGroup(sp, group->name);
  }     /* for each child group */

  /* Done! */
  return 1;
}

/************************************************************
 * IW_ODLWriteChunk
 ************************************************************
 * Dump the ODL to the given chunk stream as "chunks".
 */

int IW_ODLWriteChunk(IW_Chunksp *sp,        /* where to output to */
		     AGGREGATE node_agg,    /* top-level group */
		     IW_ODLWriteOpts *opts) /* output opts or NULL */
{
  IW_ODLWriteOpts opts_buf;


  /* No-op if no node: */
  if (!node_agg) return 0;

  /* Compile the options: */
  if (!opts) {
    IW_ODLWriteOptsClear(&opts_buf);
    opts = &opts_buf;
  }

  /* Output: */
  IW_ChunkspBegin(sp, node_agg->name);
  IW_ODLWriteChunkFields(sp, node_agg, opts);
  IW_ChunkspEnd(sp, node_agg->name);
}

/************************************************************
 * IW_ODLWriteChunkGroup
 ************************************************************
 * Dump the ODL to the given file as a "chunk group".
 */

int IW_ODLWriteChunkGroup(IW_Chunksp *sp,        /* where to output to */
			  AGGREGATE node_agg,    /* top-level group */
			  IW_ODLWriteOpts *opts) /* output opts or NULL */
{
  IW_ODLWriteOpts opts_buf;


  /* No-op if no node: */
  if (!node_agg) return 0;

  /* Compile the options: */
  if (!opts) {
    IW_ODLWriteOptsClear(&opts_buf);
    opts = &opts_buf;
  }

  /* Output: */
  IW_ChunkspBeginGroup(sp, node_agg->name);
  IW_ODLWriteChunkFields(sp, node_agg, opts);
  IW_ChunkspEndGroup(sp, node_agg->name);
}

/******************** end of file "IW_ODL.c" ********************/
