/************************************************************
 * IW_Str.c
 ************************************************************
 *
 * DESCRIPTION
 *    Some string utilities for the WebLib C utilities.
 *
 * AUTHOR
 *    Eryq (Erik Dorfman)
 *    Hughes STX Corporation
 *    30 Dec 1994
 *
 * NAMESPACE
 *    "IW_" is the Weblib C Utilities namespace.  All callable
 *     C code provided by WebLib is in this namespace.
 *
 *    "Str" indicates string-processing routines.
 *
 * RETURN-VALUES
 *    All functions which return integer status return 0 for
 *    success and negative for failure, unless otherwise noted.
 *
 *    All functions which return pointers return NULL to 
 *    indicate failure, unless otherwise noted.
 */

/*
 * BEGIN TERMS AND CONDITIONS
 * 
 *  h 
 * 
 *                        NO WARRANTY
 * 
 * BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
 * FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT
 * WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER
 * PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND,
 * EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE
 * PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME
 * THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
 * 
 * IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
 * WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
 * REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR
 * DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL
 * DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM
 * (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
 * INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF
 * THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH HOLDER
 * OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 * 
 * END TERMS AND CONDITIONS
 */

#ifndef lint
static char RCSID[] =
   "$Id: IW_Str.c,v 1.3 1996/04/17 21:16:16 imswww Exp $";
#endif /* lint */


#include "IW_Str.h"
#include <stdio.h>
#include <string.h>



/************************************************************
 *
 * String utilities
 *
 ************************************************************/


/************************************************************
 * IW_Strdup
 ************************************************************
 *
 * DESCRIPTION
 *    Implements strdup, in case we don't have it.
 */

char *IW_Strdup(s)
     char *s;
{
  char *new;

  new = IW_Malloc(strlen(s)+1);
  strcpy(new, s);
  return new;
}


/************************************************************
 * IW_Stradd
 ************************************************************
 *
 * DESCRIPTION
 *    Add to malloc'ed string.  Reallocs if necessary.
 * 
 * RETURNS
 *    Pointer to new string, which will be different from
 *    old if realloc was done.
 */ 

char *IW_Stradd(old, more)
     char *old;
     char *more;
{
  char *new;
  register int oldlen;

  if (old) {     /* it exists */
    oldlen = strlen(old);
    new = IW_Realloc(old, oldlen + strlen(more) + 1);
    strcpy(new+oldlen, more);
    return new;
  }
  else {         /* it don't! */
    return IW_Strdup(more);    
  }
}


/************************************************************
 * IW_Strequal
 ************************************************************
 *
 * DESCRIPTION
 *    Are the two strings equal?
 */ 

int IW_Strequal(s, t)
     char *s;
     char *t;
{
  return !strcmp(s, t);
}


/************************************************************
 * IW_Strcasecmp
 ************************************************************
 *
 * DESCRIPTION
 *    Case-insensitive strncmp.  In case we don't have strcasecmp.
 */ 

int IW_Strcasecmp(s, t)
     register char *s;
     register char *t;     
{
  int diff;
  
  for (; 
       *s && *t && (!(diff = (tolower(*s) - tolower(*t)))); 
       s++, t++)
    ;
  return (tolower(*s) - tolower(*t));
}


/************************************************************
 * IW_Strncasecmp
 ************************************************************
 *
 * DESCRIPTION
 *    Case-insensitive strncmp.  In case we don't have
 *    strncasecmp.
 */ 

int IW_Strncasecmp(s, t, n)
     register char *s;
     register char *t;     
     int n;
{
  int diff;
  
  for (; 
       *s && *t && n && (!(diff = (tolower(*s) - tolower(*t)))); 
       s++, t++, n--)
    ;
  return (n ? (tolower(*s) - tolower(*t)) : 0);
}


/************************************************************
 * IW_Strsplit -- split a string 
 ************************************************************
 *
 * DESCRIPTION
 *    Split string 's', using any of the characters of 'split'
 *    as delimiters (a la "strtok").  Pointers to the components
 *    are placed in 'argv', which must be large enough.
 *
 * WARNINGS
 *    Destructively modifies the string.
 */ 

void IW_Strsplit(s, split, argv)
     char *s;
     char *split;
     char *argv[];
{
  char **argp = argv;

  if (*argp++ = strtok(s, split)) {
    while (*argp++ = strtok(NULL, split))
      ;
  }
}
     

/************************************************************
 * IW_Strdowncase -- convert string to lowercase
 ************************************************************/

char *IW_Strdowncase(str)
     char *str;
{
  register char *s;
  for (s = str; *s; s++) {
    if ((*s >= 'A') && (*s <= 'Z')) 
      *s = *s - 'A' + 'a';
  }
  return str;
}


/************************************************************
 * IW_Strupcase -- convert string to uppercase
 ************************************************************/

char *IW_Strupcase(str)
     char *str;
{
  register char *s;
  for (s = str; *s; s++) {
    if ((*s >= 'a') && (*s <= 'z')) 
      *s = *s - 'a' + 'A';
  }
  return str;
}




/************************************************************
 *
 * IW_Str methods
 *
 ************************************************************/


/************************************************************
 * IW_StrNew
 ************************************************************
 *
 * DESCRIPTION
 *    Make a new string from a (char *).
 */ 

IW_Str *IW_StrNew(s)
     char *s;
{
  IW_Str *new;
  new = (IW_Str *)IW_Malloc(sizeof(IW_Str));
  if (s) {
    new->len = strlen(s);
    new->buf = IW_Strdup(s);
  }
  else {
    new->len = 0;
    new->buf = IW_Strdup("");
  }
  return new;
}


/************************************************************
 * IW_StrFree
 ************************************************************
 *
 * DESCRIPTION
 *    Free a string
 */ 

void IW_StrFree(this)
     IW_Str *this;
{
  if (this) {
    free(this->buf);
    free(this);
  }
}


/************************************************************
 * IW_StrCat
 ************************************************************
 *
 * DESCRIPTION
 *    Concat with a (char *), and return self.
 */ 

IW_Str *IW_StrCat(this, s)
     IW_Str *this;
     char *s;
{
  int len = strlen(s);
  this->buf = IW_Realloc(this->buf, this->len+len+1);
  strcpy(this->buf + this->len, s);
  this->len += len;
  return this;  
}


/************************************************************
 * IW_StrBuf -- return the string buffer
 ************************************************************/

char *IW_StrBuf(this)
     IW_Str *this;
{
  return this->buf;
}


/************************************************************
 * IW_StrLen -- return the length of the string
 ************************************************************/

int IW_StrLen(this)
     IW_Str *this;
{
  return this->len;
}


/******************** end of file "IW_Str.c" ********************/
