/*
 * giflabel
 *
 * Adapted from the original "webgif" program, distributed with GD.
 *
 * Usage: 
 *    giflabel file 'working dir' 'label1' 'label2' ... 'labeln' 
 *
 * giflabel adds label(s) to the graphic, formatted as follows:
 *   label1
 *   label2
 *   ...
 *   labeln
 *
 * A "blank" (either white, or as close to white as possible) area
 * is added at the bottom of the graphic, and the lettering is in black
 * (or the closest color available).
 *
 * If the graphic is not big enough to contain the whole label, 
 * giflabel expands it to the right.
 *
 * Author: Original by Tom Boutell
 *         Adapted by Joe McMahon
 *
 * updated 5/22/01 Joe McMahon
 * - changed command-line arguments to aloow an arbitrary number of labels
 */

/* Turn debugging off */
#define DEBUG 0

/* Bring in the gd library functions */
#include "gd.h"
#include "gdfontg.h"

/* Bring in standard I/O and string manipulation functions */
#include <stdio.h>
#include <string.h>

/* Define a constant margin size. */
#define MARGIN 2

int main(int argc, char **argv)
{
	FILE *in;
	FILE *out;
        int  best_white;
        int  best_black;
        int  wide,high;
        int  wider, higher;
        int  maxlen;
        int label_count;
	int i;

	/* Declare our image pointer */
	gdImagePtr im = 0;

        /* The output image (with caption) */
        gdImagePtr resized_image = 0;

        /* We'll set 'no' if the user hasn't made a proper request. */
	int no = 0;

	/* C programs always get at least one argument; we want at
		least three more (the image, the working directory, and 
                at least one label). 
        */
	if (argc < 4) {
		no = 1;	
		goto usage;
	}

	/* The first argument should be the image. Open the file. */
	in = fopen(argv[1], "rb");	
	if (!in) {
		fprintf(stderr,
			"Error: can't open file %s.\n", argv[argc-1]);
	}
	/* Now load the image. */	
	im = gdImageCreateFromGif(in);
	fclose(in);
	/* If the load failed, it must not be a GIF file. */
	if (!im) {
		printf("Error: %s is not a valid gif file.\n", argv[1]);
		exit(1);	
	}

        /* If the invocation was bad, print the usage message. */
usage:
	if (no) {
		/* If the command failed, output an explanation. */
		fprintf(stderr, 
	"Usage: giflabel file 'working directory' 'label' 'label' ...\n");
	} 

       /* Things look good. */
       /* Figure out how wide the widest label string is, and how high
          the output graphics will need to be. */
       wide = gdImageSX(im);
       high = gdImageSY(im);

       maxlen = 0; 
       label_count = 0;
       higher = high + MARGIN;

       for (i = 3; i < argc; i++) {
           label_count++;
           higher = higher + MARGIN + gdFontGiant->h;
           maxlen = ( strlen(argv[i]) >= maxlen ) 
                        ? strlen(argv[i]) : maxlen;
       }
       maxlen = maxlen*gdFontGiant->w;
#ifdef DEBUG
       printf("%d labels\n",label_count);
#endif

       if (maxlen == 0) {
		fprintf(stderr,
			"Error: All labels are blank.\n");
                exit(1);
       }

       /* Height of caption is (number of labels) lines of font plus margin 
          above, between, and below. */
       wider  = (wide <  2*MARGIN + maxlen ? 2*MARGIN + maxlen : wide);
#ifdef DEBUG
       printf("wide: %d, high: %d\n",wide,high);
       printf("wider: %d, higher: %d, maxlen: %d\n",wider,higher,maxlen);
#endif

       /* Copy the old GIF into a new, bigger one. */
       resized_image = gdImageCreate(wider,higher);
       best_white = gdImageColorAllocate(resized_image,255,255,255);
       best_black = gdImageColorAllocate(resized_image,0,0,0);

#ifdef DEBUG
       printf("New image: %d by %d\n", gdImageSX(resized_image),
                                       gdImageSY(resized_image));
#endif
       gdImageCopy(resized_image, im, 
                   0, 0, 
                   0, 0, 
                   wide, high);
       /* Insert the lettering. */
       for(i=3; i<argc; i++) {
           gdImageString(resized_image,gdFontGiant,
                         MARGIN,high+MARGIN,
                         argv[i], best_black);
           high = high + MARGIN + gdFontGiant->h;
#ifdef DEBUG
           printf("label at %d %d\n", MARGIN, high+MARGIN);
           printf("i: %d high: %d\n",i,high);
#endif
       }

	/* Ready to save. Open a temporary file. */
#ifdef DEBUG
        printf("Ready to save ... ");
#endif
        chdir(argv[2]);
	out = fopen("temp.tmp", "wb");
	if (!out) {
		fprintf(stderr,
			"Unable to write to temp.tmp -- exiting\n");
		exit(1);
	}

	/* Write the new gif. */
	gdImageGif(resized_image, out);
	fclose(out);
#ifdef DEBUG
        printf("saved.\n");
#endif

	/* Erase the old gif. */
	unlink(argv[1]);
	/* Rename the new to the old. */
	rename("temp.tmp", argv[1]);

	/* Delete the image from memory. */
	if (im) {
		gdImageDestroy(im);
	}
        if (resized_image) {
            gdImageDestroy(resized_image);
        }
	/* All's well that ends well. */
	return 0;
}

/*
 * Note: This function was back-ported from gd 1.8.3 to GD 1.2
 * to make it easier and faster to find an approximate black and
 * white as needed.
 */
int gdImageColorResolve(gdImagePtr im, int r, int g, int b) {
    int c;
    int ct = -1;
    int op = -1;
    long rd, gd, bd, dist;
    long mindist = 3*255*255;       /* init to max poss dist */

    for (c = 0; c < im->colorsTotal; c++) {
        if (im->open[c]) {
           op = c;                  /* Save open slot */
           continue;                /* Color not in use */
        }
        rd = (long)(im->red  [c] - r);
        gd = (long)(im->green[c] - g);
        bd = (long)(im->blue [c] - b);
        dist = rd * rd + gd * gd + bd * bd;
        if (dist < mindist) {
            if (dist == 0) {
                printf("Exact match!\n");
                return c;           /* Return exact match color */
            }
            mindist = dist;
            ct = c;
        }
    }
    /* no exact match.  We now know closest, but first try to allocate exact */
    if (op == -1) {
        op = im->colorsTotal;
        if (op == gdMaxColors) {    /* No room for more colors */
            printf("No exact match\n");
            return ct;              /* Return closest available color */
        }
        im->colorsTotal++;
    }
    im->red  [op] = r;
    im->green[op] = g;
    im->blue [op] = b;
    im->open [op] = 0;
    printf("new color inserted\n");
    return op;                      /* Return newly allocated color */
}

