/*
 * Copyright (c) 1996 U.S. Geological Survey (USGS)
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of USGS not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  USGS makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * USGS DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL USGS
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */
/*
 * glismapper.c - draws reference aid into a gif image
 *
 *	Usage: glismapper [-mapdir mapdir] [-symboldir symboldir]
 *			[-invcoords x y] [-clat degrees] [-clon degrees]
 *			[-zoom percent] [-width pixels] [-height pixels]
 *			[-gridinc degrees] [-water] [-political]
 *			[-roads] [-rail] [-places] [-napp] [-nhap_bw] 
 *                      [-nhap_cir] [-wrs1] [-wrs2]
 *			[-zoomreportonly]
 *
 * Default Behavior:
 *
 *	Reads additional features to plot from stdin where each feature is
 *	a pipe delimited record with this format:
 *
 *	key|type|label|color|width|style|lat1|lon1|lat2|lon2|	...
 *
 *	where:
 *	- key is a unique integer
 *	- type is one of: point, range, polygon, polyline, spherical_polygon,
 *		or spherical_polyline
 *	- label is a text string to label the feature (use "" if no label)
 *	- color is a valid colorname from the X11R4 color database
 *	- width is the line width in pixels
 *	- style is one of: solid, dashed, dotted, large_dot, small_dot, med_dot,
 *		target,	star, or cross
 *	- lat and lon are floating point latitudes and longitudes
 *
 *	Outputs a GIF image to stdout
 *
 * Behavior if -invcoords is used:
 *
 *	Returns latitude|longitude on stdout for the x,y point desired
 */
#define GLOBAL
#include "glismapper.h"

int
main(argc, argv)
int argc;
char **argv;
{
    int i, rtn;
    int invcoords = 0;
    int image_x, image_y;
    POINT rtn_coords;

    /* initialize the options */
    strcpy(mapdir, "map");
    strcpy(symboldir, "map_symbols");
    center.latitude = DEFAULT_CENTER_LAT;
    center.longitude = DEFAULT_CENTER_LON;
    width = DEFAULT_WIDTH;
    height = DEFAULT_HEIGHT;
    zoom_percent = DEFAULT_ZOOM;
    grid_increment = DEFAULT_GRID;
    water_layer = FALSE;
    political_layer = FALSE;
    rail_layer = FALSE;
    roads_layer = FALSE;
    populated_places_layer = FALSE;
    napp_layer = FALSE;
    nhap_bw_layer = FALSE;
    nhap_cir_layer = FALSE;
    wrs1_layer = FALSE;
    wrs2_layer = FALSE;
    zoom_report_only = FALSE;

    /* read in the arguments */
    for(i = 1; i < argc; i++)
    {
	if(!strcmp(argv[i], "-mapdir"))
	{
	    if(i != argc - 1)
	    {
	        strcpy(mapdir, argv[i + 1]);
	        if(mapdir[strlen(mapdir) - 1] == '/')
		    mapdir[strlen(mapdir) - 1] = '\0';
	    }
	    i++;
	}
	else if(!strcmp(argv[i], "-symboldir"))
	{
	    if(i != argc - 1)
	    {
	        strcpy(symboldir, argv[i + 1]);
	        if(symboldir[strlen(symboldir) - 1] == '/')
		    symboldir[strlen(symboldir) - 1] = '\0';
	    }
	    i++;
	}
	else if(!strcmp(argv[i], "-proj"))
	{
	    if(i != argc - 1)
	    {
	        if(!strcmp("gnomonic", argv[i + 1])) {
		    projection = 1;
                } else {
		    projection = 0;
                }
	    }
	    i++;
	}
	else if(!strcmp(argv[i], "-invcoords"))
	{
	    if(i < argc - 2)
	    {
	        invcoords = 1;
		image_x = atoi(argv[i + 1]);
		image_y = atoi(argv[i + 2]);
	    }
	    i += 2;
	}
	else if(!strcmp(argv[i], "-clat"))
	{
	    if(i != argc - 1)
		center.latitude = atof(argv[i + 1]);
	    if(fabs(center.latitude) > 90)
		center.latitude = DEFAULT_CENTER_LAT;
	    i++;
	}
	else if(!strcmp(argv[i], "-clon"))
	{
	    if(i != argc - 1)
		center.longitude = atof(argv[i + 1]);
	    if(fabs(center.longitude) > 180)
		center.latitude = DEFAULT_CENTER_LON;
	    i++;
	}
	else if(!strcmp(argv[i], "-zoom"))
	{
	    if(i != argc - 1)
		zoom_percent = atoi(argv[i + 1]);
	    if(zoom_percent < 100)
		zoom_percent = DEFAULT_ZOOM;
	    if(zoom_percent > 500000)
		zoom_percent = 500000;
	    i++;
	}
	else if(!strcmp(argv[i], "-width"))
	{
	    if(i != argc - 1)
		width = atoi(argv[i + 1]);
	    if(width > 1280)
		width = DEFAULT_WIDTH;
	    i++;
	}
	else if(!strcmp(argv[i], "-height"))
	{
	    if(i != argc - 1)
		height = atoi(argv[i + 1]);
	    if(height > 1024)
		height = DEFAULT_HEIGHT;
	    i++;
	}
	else if(!strcmp(argv[i], "-gridinc"))
	{
	    if(i != argc - 1)
	    {
		grid_increment = atof(argv[i + 1]);
		if(grid_increment != 0 &&
			(grid_increment < 0.0166 || grid_increment > 45))
		    grid_increment = DEFAULT_GRID;
	    }
	    i++;
	}
	else if(!strcmp(argv[i], "-water"))
	    water_layer = TRUE;
	else if(!strcmp(argv[i], "-political"))
	    political_layer = TRUE;
	else if(!strcmp(argv[i], "-roads"))
	    roads_layer = TRUE;
	else if(!strcmp(argv[i], "-rail"))
	    rail_layer = TRUE;
	else if(!strcmp(argv[i], "-places"))
	    populated_places_layer = TRUE;
	else if(!strcmp(argv[i], "-napp"))
	    napp_layer = TRUE;
	else if(!strcmp(argv[i], "-nhap_bw"))
	    nhap_bw_layer = TRUE;
	else if(!strcmp(argv[i], "-nhap_cir"))
	    nhap_cir_layer = TRUE;
	else if(!strcmp(argv[i], "-wrs1"))
	    wrs1_layer = TRUE;
	else if(!strcmp(argv[i], "-wrs2"))
	    wrs2_layer = TRUE;
	else if(!strcmp(argv[i], "-zoomreportonly"))
	    zoom_report_only = TRUE;
	else if(!strcmp(argv[i], "-help"))
	{
	    fprintf(stderr, "usage: glismapper [-mapdir mapdir] [-symboldir symboldir]");
	    fprintf(stderr, "\n\t[-invcoords x y] [-clat degrees] [-clon degrees]");
	    fprintf(stderr, "\n\t[-zoom percent] [-width pixels] [-height pixels]");
	    fprintf(stderr, "\n\t[-gridinc degrees] [-water] [-political]");
	    fprintf(stderr, "\n\t[-roads] [-rail] [-places] [-napp] [-nhap_bw] [-nhap_cir] [-wrs1] [-wrs2]");
	    fprintf(stderr, "\n\t[-zoomreportonly]\n");
	    exit(0);
	}
    }

    rtn = setup_mapper();
    if(rtn)
	exit(rtn);

    /* process invcoord requests */
    if(invcoords)
    {
	if(projection) { /* projection is gnomonic */
            rtn = invgnomonic(radius, center.latitude / RAD,
                              center.longitude / RAD,
                              image_x - width / 2, height / 2 - image_y,
			      &rtn_coords.latitude, &rtn_coords.longitude);
	} else {  /* projection is ortho */
            rtn = invortho(radius, center.latitude / RAD,
                              center.longitude / RAD,
                              image_x - width / 2, height / 2 - image_y,
			      &rtn_coords.latitude, &rtn_coords.longitude);
	}
	if(!rtn)
	    fprintf(stdout, "%7.4f|%7.4f\n", rtn_coords.latitude * RAD,
					rtn_coords.longitude * RAD); 
	exit(rtn);
    }

    /* load the entities */
    if(!zoom_report_only)
        LoadEntities(stdin);

    /* draw the map */
    DrawRefaid();

    /* interlace it for effect */
    gdImageInterlace(image, TRUE);

    /* write it out to stdout */
    gdImageGif(image, stdout);

    /* free the memory */
    gdImageDestroy(small_dot);
    gdImageDestroy(med_dot);
    gdImageDestroy(large_dot);
    gdImageDestroy(star);
    gdImageDestroy(target);
    gdImageDestroy(box2);
    gdImageDestroy(cross);
    gdImageDestroy(image);

    exit(0);
}
