/*
 * Copyright (c) 1996 U.S. Geological Survey (USGS)
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of USGS not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  USGS makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * USGS DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL USGS
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */
/*
 * mapper.c
 */
#include <netinet/in.h>
#include <ctype.h>
#include "glismapper.h"
#include "rgb.h"

extern double rint(double);

static void DrawGrid(EXTENT *);
static void LabelGrid(EXTENT *, SCREENLABEL **);
static void plot_entity(Entity *);
static void plot_route(GC *, POINT *, int, int);
static int calcRoute(POINT *, POINT *, POINT *);
static int calcGreatCircleRoute(POINT *, POINT *, POINT *);
static char lonquad(float, int);
static char latquad(float, int);
static int latlontile(float, int);
static int latlonsubtile(float, int);
int draw_layer(int, int, EXTENT *, SCREENLABEL **);
static int plot_tile(char *, int, int, SCREENLABEL **);
static int find_crossing(POINT *, POINT *, POINT *);
static void addlabeltolist(SCREENLABEL **, SCREENLABEL *);

#ifndef HAVESINCOS
static void
sincos(double x, double *s, double *c)
{
    *s = sin(x);
    *c = cos(x);
}
#endif

int
setup_mapper(void)
{
    FILE *fp;
    char path[256];

    if(width <= height)
        radius = width * zoom_percent / 200;
    else
        radius = height * zoom_percent / 200;

    /* create the image */
    if(!(image = gdImageCreate(width, height)))
    {
	fprintf(stderr, "Could not create image of size %d x %d\n",
								width, height);
	return(-1);
    }

    /* allocate the colors, background first */
    background = allocate_color(image, "gray85"); /* for land */
    black = allocate_color(image, "gray60");      /* for political lines */
    white = allocate_color(image, "white");       /* for grids */
    red = allocate_color(image, "red");
    blue = allocate_color(image, "ocean");        /* for ocean and rivers */
    cyan = allocate_color(image, "cyan");
    magenta = allocate_color(image, "magenta");
    orange = allocate_color(image, "orange2");
    gray25 = allocate_color(image, "gray25");     /* for lat and lon labels */

    land_gc.color = background; /* Color for land */
    label_gc.color = gray25; /* Color for degree labels */
    grid_gc.color = white;   /* Color for grids */
    entity_gc.color = black; /* Color for political lines */
    entity_gc.line_width = grid_gc.line_width = 1;
    entity_gc.line_style = grid_gc.line_style = SOLID;

    /* load large_dot symbol */
    sprintf(path, "%s/large_dot.xbm", symboldir);
    if(!(fp = fopen(path, "r")))
    {
	fprintf(stderr, "Could not open %s\n", path);
	return(-1);
    }
    large_dot = gdImageCreateFromXbm(fp);
    gdImageColorTransparent(large_dot, 0);
    gdImageColorDeallocate(large_dot, 1);
    gdImageColorAllocate(large_dot, 255, 0, 0);
    fclose(fp);

    /* load med_dot symbol */
    sprintf(path, "%s/med_dot.xbm", symboldir);
    if(!(fp = fopen(path, "r")))
    {
	fprintf(stderr, "Could not open %s\n", path);
	return(-1);
    }
    med_dot = gdImageCreateFromXbm(fp);
    gdImageColorTransparent(med_dot, 0);
    gdImageColorDeallocate(med_dot, 1);
    gdImageColorAllocate(med_dot, 255, 0, 0);
    fclose(fp);

    /* load small_dot symbol */
    sprintf(path, "%s/small_dot.xbm", symboldir);
    if(!(fp = fopen(path, "r")))
    {
	fprintf(stderr, "Could not open %s\n", path);
	return(-1);
    }
    small_dot = gdImageCreateFromXbm(fp);
    gdImageColorTransparent(small_dot, 0);
    gdImageColorDeallocate(small_dot, 1);
    gdImageColorAllocate(small_dot, 255, 0, 0);
    fclose(fp);

    /* load small_black_dot symbol */
    sprintf(path, "%s/small_dot.xbm", symboldir);
    if(!(fp = fopen(path, "r")))
    {
	fprintf(stderr, "Could not open %s\n", path);
	return(-1);
    }
    small_black_dot = gdImageCreateFromXbm(fp);
    gdImageColorTransparent(small_black_dot, 0);
    gdImageColorDeallocate(small_black_dot, 1);
    gdImageColorAllocate(small_black_dot, 0, 0, 0);
    fclose(fp);

    /* load star symbol */
    sprintf(path, "%s/star.xbm", symboldir);
    if(!(fp = fopen(path, "r")))
    {
	fprintf(stderr, "Could not open %s\n", path);
	return(-1);
    }
    star = gdImageCreateFromXbm(fp);
    gdImageColorTransparent(star, 0);
    gdImageColorDeallocate(star, 1);
    gdImageColorAllocate(star, 255, 0, 0);
    fclose(fp);

    /* load target symbol */
    sprintf(path, "%s/target.xbm", symboldir);
    if(!(fp = fopen(path, "r")))
    {
	fprintf(stderr, "Could not open %s\n", path);
	return(-1);
    }
    target = gdImageCreateFromXbm(fp);
    gdImageColorTransparent(target, 0);
    fclose(fp);

    /* load box2 symbol */
    sprintf(path, "%s/box2.xbm", symboldir);
    if(!(fp = fopen(path, "r")))
    {
	fprintf(stderr, "Could not open %s\n", path);
	return(-1);
    }
    box2 = gdImageCreateFromXbm(fp);
    fclose(fp);

    /* load cross symbol */
    sprintf(path, "%s/cross.xbm", symboldir);
    if(!(fp = fopen(path, "r")))
    {
	fprintf(stderr, "Could not open %s\n", path);
	return(-1);
    }
    cross = gdImageCreateFromXbm(fp);
    gdImageColorTransparent(cross, 0);
    gdImageColorDeallocate(cross, 1);
    gdImageColorAllocate(cross, 255, 0, 0);
    fclose(fp);

    return(0);
}

/*      PURPOSE:        Orthographically projects geographic coordinates of a
 *                      point in radians to X-Y coordinates in pixels.
 */
int
ortho(int radius, float center_lat, float center_lon, float inlat, float inlon,
							int *outx, int *outy)
/* radius - radius of the earth in pixels */
/* center_lat - latitude of center of projection in radians */
/* center_lon - longitude of center of projection in radians */
/* inlat - latitude of the input point in radians */
/* inlon - longitude of the input point in radians */
/* *outx - x distance from center of window in pixels */
/* *outy - y distance from center of window in pixels */
{
    double delta;
    double cosL;
    double cosC;
    double sin_inlat;
    double cos_inlat;
    static float last_center_lat = 0;
    static double sin_center_lat = 0;
    static double cos_center_lat = 1;

    delta = inlon - center_lon;
    if (delta < -M_PI)
       delta = delta + M_2PI;
    else if (delta > M_PI)
       delta = delta - M_2PI;

    if(center_lat != last_center_lat)
    {
	last_center_lat = center_lat;
	sincos((double)center_lat, &sin_center_lat, &cos_center_lat);

    }
    sincos((double)inlat, &sin_inlat, &cos_inlat);

    cosL = cos(delta) * cos_inlat;
    cosC = sin_center_lat * sin_inlat + cos_center_lat * cosL;

    if (cosC >= 0)
    {
        *outx = cos_inlat * sin(delta) * radius;
        *outy = (cos_center_lat * sin_inlat - sin_center_lat * cosL) * radius;
        return(0);  /* OK */
    }
    else
        return(-1); /* point cannot be projected */
}

/*      PURPOSE:        Projects using gnomonic geographic coordinates of a
 *                      point in radians to X-Y coordinates in pixels.
 */
int
gnomonic(int radius, float center_lat, float center_lon, float inlat, 
         float inlon, int *outx, int *outy)
/* radius - radius of the earth in pixels */
/* center_lat - latitude of center of projection in radians */
/* center_lon - longitude of center of projection in radians */
/* inlat - latitude of the input point in radians */
/* inlon - longitude of the input point in radians */
/* *outx - x distance from center of window in pixels */
/* *outy - y distance from center of window in pixels */
{
    double delta;
    double cosL;
    double cosC;
    double sin_inlat;
    double cos_inlat;
    double distance;
    static float last_center_lat = 0;
    static double sin_center_lat = 0;
    static double cos_center_lat = 1;

    delta = inlon - center_lon;
    if (delta < -M_PI)
       delta = delta + M_2PI;
    else if (delta > M_PI)
       delta = delta - M_2PI;

    if(center_lat != last_center_lat)
    {
	last_center_lat = center_lat;
	sincos((double)center_lat, &sin_center_lat, &cos_center_lat);

    }
    sincos((double)inlat, &sin_inlat, &cos_inlat);

    cosL = cos(delta) * cos_inlat;
    cosC = sin_center_lat * sin_inlat + cos_center_lat * cosL;

    if (cosC > 0)
    {
        *outx = cos_inlat * sin(delta) * radius / cosC;
        *outy = (cos_center_lat * sin_inlat - sin_center_lat * cosL) * radius /                 cosC;
        distance = sqrt((double)(*outx * *outx + *outy * *outy));
	if ((*outx > radius) || (*outx < -radius) || (*outy > radius) ||
            (*outy < -radius)) {
	    return(-1);  /* point cannot be projected */
        } else {
            return(0);  /* OK */
	}
    }
    else
        return(-1); /* point cannot be projected */
}

/*	PURPOSE:	Converts X-Y image coordinates of a point in pixels
 *			to geographic point in radians by inverting the
 *			orthographic projection.
 */
int
invortho(int radius, float center_lat, float center_lon, int inx, int iny,
						float *outlat, float *outlon)
/* radius - radius of the earth in pixels */
/* center_lat - latitude of center of projection in radians */
/* center_lon - longitude of center of projection in radians */
/* inx - x distance from center of window in pixels */
/* iny - y distance from center of window in pixels */
/* *outlat - latitude of the selected point in radians */
/* *outlon - longitude of the selected point in radians */
{
    double distance;
    double CON;
    double sin_outlat;
    double sin_Z;
    double cos_Z;
    static float last_center_lat = 0;
    static double sin_center_lat = 0;
    static double cos_center_lat = 1;

    distance = sqrt((double)(inx * inx + iny * iny));

    if (distance > radius)
        return(-1);  /* out of this world */
    if (distance == 0)
    {
        *outlat = center_lat;
        *outlon = center_lon;
    }
    else
    {
	if(center_lat != last_center_lat)
	{
	    last_center_lat = center_lat;
	    if(fabs((double)center_lat) < 89.5 / RAD)
	        sincos((double)center_lat, &sin_center_lat, &cos_center_lat);
	    /* make sure that asin2(0,0) doesn't occur */
	    else if(center_lat > 0)
	        sincos((double) M_PI_2 - 0.00000001, &sin_center_lat,
							&cos_center_lat);
	    else
	        sincos((double) -M_PI_2 + 0.00000001, &sin_center_lat,
							&cos_center_lat);
	}

        sin_Z = distance / radius;
        cos_Z = sqrt(1 - (sin_Z * sin_Z));

        sin_outlat = (cos_Z * sin_center_lat) +
                        	(iny * sin_Z * cos_center_lat / distance);
        *outlat = asin(sin_outlat);
        CON = cos_Z - (sin_center_lat * sin_outlat);
        *outlon = center_lon + atan2((inx * sin_Z * cos_center_lat),
                    (CON * distance));
        if (*outlon < -M_PI)
            *outlon += M_2PI;
        else if (*outlon > M_PI)
            *outlon -= M_2PI;
    }
    return(0);
}


/*	PURPOSE:	Converts X-Y image coordinates of a point in pixels
 *			to geographic point in radians by inverting the
 *			gnomonic projection.
 */
int
invgnomonic(int radius, float center_lat, float center_lon, int inx, int iny,
						float *outlat, float *outlon)
/* radius - radius of the earth in pixels */
/* center_lat - latitude of center of projection in radians */
/* center_lon - longitude of center of projection in radians */
/* inx - x distance from center of window in pixels */
/* iny - y distance from center of window in pixels */
/* *outlat - latitude of the selected point in radians */
/* *outlon - longitude of the selected point in radians */
{
    double distance;
    double CON;
    double sin_outlat;
    double normalized_distance;
    double theta;
    static double sin_theta = 0;
    static double cos_theta = 1;
    static float last_center_lat = 0;
    static double sin_center_lat = 0;
    static double cos_center_lat = 1;

    distance = sqrt((double)(inx * inx + iny * iny));

    if (distance > radius)
        return(-1);  /* out of this world */
    if (distance == 0)
    {
        *outlat = center_lat;
        *outlon = center_lon;
    }
    else
    {
	if(center_lat != last_center_lat)
	{
	    last_center_lat = center_lat;
	    if(fabs((double)center_lat) < 89.5 / RAD)
	        sincos((double)center_lat, &sin_center_lat, &cos_center_lat);
	    /* make sure that asin2(0,0) doesn't occur */
	    else if(center_lat > 0)
	        sincos((double) M_PI_2 - 0.00000001, &sin_center_lat,
							&cos_center_lat);
	    else
	        sincos((double) -M_PI_2 + 0.00000001, &sin_center_lat,
							&cos_center_lat);
	}

        normalized_distance = distance / radius;
	theta = atan(normalized_distance);

        sincos(theta, &sin_theta, &cos_theta);
        sin_outlat = cos_theta * sin_center_lat + 
	  ((iny * sin_theta * cos_theta * cos_center_lat) / distance);
        *outlat = asin(sin_outlat);
        CON = distance * cos_center_lat * cos_theta -
	  (iny * sin_center_lat * sin_theta);
        *outlon = center_lon + atan2((inx * sin_theta), CON);
        if (*outlon < -M_PI)
            *outlon += M_2PI;
        else if (*outlon > M_PI)
            *outlon -= M_2PI;
    }
    return(0);
}

static int
calcRoute(POINT *start, POINT *end, POINT *route)
{
    register int i;
    float end_lon = end->longitude;
    float start_lon = start->longitude;
    float xd, yd, xinc, yinc;
    int numsegs;

    xd = end->longitude - start->longitude;
    yd = end->latitude - start->latitude;

    /* check for greater than 180 degrees of longitude */
    if(fabs((double)xd) > 180)
    {
        if(start_lon < end_lon)
            start_lon += 360;
        else
            end_lon += 360;

        xd = end_lon - start_lon;
    }

    /* split larger dimension into 1 degree segments */
    if(fabs((double)xd) > fabs((double)yd))
    {
        if(!(numsegs = fabs((double)xd)))
            numsegs = 1;
        xinc = xd / numsegs;
        yinc = yd / xd * xinc;
    }
    else
    {
        if(!(numsegs = fabs((double)yd)))
            numsegs = 1;
        yinc = yd / numsegs;
        xinc = xd / yd * yinc;
    }

    route[0].latitude = start->latitude;
    route[0].longitude = start->longitude;
    for(i = 1; i <= numsegs; i++)
    {
	route[i].latitude = route[i - 1].latitude + yinc;
	route[i].longitude = route[i - 1].longitude + xinc;
    }
    return(numsegs + 1);
}

static int
calcGreatCircleRoute(POINT *start, POINT *end, POINT *route)
{
    int intval_no,num_intvals;
    double central_angle;
    double angle_increment;
    double cos_central_angle;
    double sin_central_angle;
    double cos_start_lat;
    double sin_start_lat;
    double cos_start_lng;
    double sin_start_lng;
    double cos_end_lat;
    double sin_end_lat;
    double cos_end_lng;
    double sin_end_lng;
    double cos_route_angle;
    double sin_route_angle;
    double route_angle;
    double prev_route_lat;
    double prev_route_lng;
    double route_vector_x,route_vector_y,route_vector_z;
    double temp1,temp2;

    cos_start_lat = cos(start->latitude / RAD);
    sin_start_lat = sin(start->latitude / RAD);
    cos_start_lng = cos(start->longitude / RAD);
    sin_start_lng = sin(start->longitude / RAD);
    cos_end_lat = cos(end->latitude / RAD);
    sin_end_lat = sin(end->latitude / RAD);
    cos_end_lng = cos(end->longitude / RAD);
    sin_end_lng = sin(end->longitude / RAD);

    cos_central_angle = cos_start_lat * cos_end_lat *
			cos((start->longitude - end->longitude) / RAD) +
			sin_start_lat * sin_end_lat;
    if(cos_central_angle <= -1.)
	return(-1);	/* failure - don't know which way to draw */
    else if(cos_central_angle >=  1.)
        return(-1);	/* no interval - ending points are equal */

    sin_central_angle = sqrt(fabs(1-cos_central_angle*cos_central_angle));
    if(cos_central_angle >= 1.0)
	central_angle = 0.0;
    else if(cos_central_angle <= -1.0)
	central_angle = M_PI;
    else
	central_angle = acos(cos_central_angle);

    /* If we think that the angle should be > 180, make it think that it is */
/*
    if (fabs((double)end->longitude-start->longitude)>180)
    {
	central_angle=M_2PI-central_angle;
	sin_central_angle=-sin_central_angle;
    }
*/

    num_intvals = ceil(central_angle/2*RAD);
    angle_increment = central_angle/num_intvals;

    prev_route_lat=start->latitude;
    prev_route_lng=start->longitude;
    for(intval_no=0;intval_no<=num_intvals;intval_no++)
    {
	route_angle = intval_no*angle_increment;
	cos_route_angle = cos(route_angle);
	sin_route_angle = sin(route_angle);

	temp1 = sin_route_angle/sin_central_angle;
	temp2 = cos_route_angle - temp1*cos_central_angle;

	route_vector_x = temp2*cos_start_lat*cos_start_lng
                   + temp1*cos_end_lat*cos_end_lng;

	route_vector_y = -(temp2*cos_start_lat*sin_start_lng
                       + temp1*cos_end_lat*sin_end_lng);

	route_vector_z = temp2*sin_start_lat + temp1*sin_end_lat;

	route[intval_no].latitude = asin(route_vector_z) * RAD;

	route[intval_no].longitude = -atan2(route_vector_y,route_vector_x) * RAD;
    }
    return(intval_no);
}

static void
plot_entity(Entity *entity)
{
    int i, x, y, length, maxchars, maxlines;
    int center_x = width / 2;
    int center_y = height / 2;
    int status;
    POINT route[512];

    /* Set the entity_gc if necessary */
    if(entity->area.type != SINGLE_PT)
    {
	entity_gc.color = entity->color;
	entity_gc.line_width = entity->line_width;
	entity_gc.line_style = entity->line_style;
	entity_gc.dash_style = entity->dash_style;
    }

    switch(entity->area.type)
    {
	case SINGLE_PT:
	if(projection) { /* projection is gnomonic */
            status = gnomonic(radius, center.latitude / RAD,
			center.longitude / RAD,
			entity->area.points[0].latitude / RAD,
			entity->area.points[0].longitude / RAD, &x, &y);
	} else { /* projection is orthographic */
            status = ortho(radius, center.latitude / RAD,
			center.longitude / RAD,
			entity->area.points[0].latitude / RAD,
			entity->area.points[0].longitude / RAD, &x, &y);
	}
        x += center_x;
        y = center_y - y;
	switch(entity->symbol)
	{
	    case SMALL_DOT:
	    gdImageColorDeallocate(small_dot, 1);
	    gdImageColorAllocate(small_dot,
			gdImageRed(image, entity->color),
			gdImageGreen(image, entity->color),
			gdImageBlue(image, entity->color));
            gdImageCopy(image, small_dot,
		    x - gdImageSX(small_dot) / 2, y - gdImageSY(small_dot) / 2,
		    0, 0, gdImageSX(small_dot), gdImageSY(small_dot));
	    if(entity->label[0] != '\0')
                gdImageStringFilled(image, gdFontMediumBold,
		    x - gdFontW(gdFontMediumBold) * (int)strlen(entity->label) / 2,
		    y - gdImageSY(small_dot) / 2 - gdFontH(gdFontMediumBold)+2,
		    entity->label, entity->color, background);
	    break;

	    case MED_DOT:
	    gdImageColorDeallocate(med_dot, 1);
	    gdImageColorAllocate(med_dot,
			gdImageRed(image, entity->color),
			gdImageGreen(image, entity->color),
			gdImageBlue(image, entity->color));
            gdImageCopy(image, med_dot,
		    x - gdImageSX(med_dot) / 2, y - gdImageSY(med_dot) / 2,
		    0, 0, gdImageSX(med_dot), gdImageSY(med_dot));
	    if(entity->label[0] != '\0')
                gdImageStringFilled(image, gdFontMediumBold,
		    x - gdFontW(gdFontMediumBold) * (int)strlen(entity->label) / 2,
		    y - gdImageSY(med_dot) / 2 - gdFontH(gdFontMediumBold)+2,
		    entity->label, entity->color, background);
	    break;

	    case LARGE_DOT:
	    gdImageColorDeallocate(large_dot, 1);
	    gdImageColorAllocate(large_dot,
			gdImageRed(image, entity->color),
			gdImageGreen(image, entity->color),
			gdImageBlue(image, entity->color));
            gdImageCopy(image, large_dot,
		    x - gdImageSX(large_dot) / 2, y - gdImageSY(large_dot) / 2,
		    0, 0, gdImageSX(large_dot), gdImageSY(large_dot));
	    if(entity->label[0] != '\0')
                gdImageStringFilled(image, gdFontMediumBold,
		    x - gdFontW(gdFontMediumBold) * (int)strlen(entity->label) / 2,
		    y - gdImageSY(large_dot) / 2 - gdFontH(gdFontMediumBold)+2,
		    entity->label, entity->color, background);
	    break;

	    case TARGET:
	    gdImageColorDeallocate(target, 1);
	    gdImageColorAllocate(target,
			gdImageRed(image, entity->color),
			gdImageGreen(image, entity->color),
			gdImageBlue(image, entity->color));
            gdImageCopy(image, target,
		    x - gdImageSX(target) / 2, y - gdImageSY(target) / 2,
		    0, 0, gdImageSX(target), gdImageSY(target));
	    if(entity->label[0] != '\0')
                gdImageStringFilled(image, gdFontMediumBold,
		    x - gdFontW(gdFontMediumBold) * (int)strlen(entity->label) / 2,
		    y - gdImageSY(target) / 2 - gdFontH(gdFontMediumBold)+2,
		    entity->label, entity->color, background);
	    break;

	    case STAR:
	    gdImageColorDeallocate(star, 1);
	    gdImageColorAllocate(star,
			gdImageRed(image, entity->color),
			gdImageGreen(image, entity->color),
			gdImageBlue(image, entity->color));
            gdImageCopy(image, star,
		    x - gdImageSX(star) / 2, y - gdImageSY(star) / 2,
		    0, 0, gdImageSX(star), gdImageSY(star));
	    if(entity->label[0] != '\0')
                gdImageStringFilled(image, gdFontMediumBold,
		    x - gdFontW(gdFontMediumBold) * (int)strlen(entity->label) / 2,
		    y - gdImageSY(star) / 2 - gdFontH(gdFontMediumBold)+2,
		    entity->label, entity->color, background);
	    break;

	    case CROSS:
	    gdImageColorDeallocate(cross, 1);
	    gdImageColorAllocate(cross,
			gdImageRed(image, entity->color),
			gdImageGreen(image, entity->color),
			gdImageBlue(image, entity->color));
            gdImageCopy(image, cross,
		    x - gdImageSX(cross) / 2, y - gdImageSY(cross) / 2,
		    0, 0, gdImageSX(cross), gdImageSY(cross));
	    if(entity->label[0] != '\0')
                gdImageStringFilled(image, gdFontMediumBold,
		    x - gdFontW(gdFontMediumBold) * (int)strlen(entity->label) / 2,
		    y - gdImageSY(cross) / 2 - gdFontH(gdFontMediumBold)+2,
		    entity->label, entity->color, background);
	    break;
	}
	break;

	case RANGE:
	if(entity->area.npts != 2)
	    break;

	/* add intermediate corner points */
	entity->area.points[2].latitude = entity->area.points[0].latitude;
	entity->area.points[2].longitude = entity->area.points[1].longitude;
	entity->area.points[3].latitude = entity->area.points[1].latitude;
	entity->area.points[3].longitude = entity->area.points[0].longitude;

	/* plot the sides */
	length = calcRoute(&entity->area.points[0],
				&entity->area.points[2], route);
	plot_route(&entity_gc, route, length, POLYLINE);
	length = calcRoute(&entity->area.points[2],
				&entity->area.points[1], route);
	plot_route(&entity_gc, route, length, POLYLINE);
	length = calcRoute(&entity->area.points[1],
				&entity->area.points[3], route);
	plot_route(&entity_gc, route, length, POLYLINE);
	length = calcRoute(&entity->area.points[3],
				&entity->area.points[0], route);
	plot_route(&entity_gc, route, length, POLYLINE);

        /* add the label */
	if(entity->label[0] != '\0')
	{
	    int status0, status1, x0, x1, y0, y1;

	    /* find the center of the range */
            if(projection) { /* projection is gnomonic */
                status0 = gnomonic(radius, center.latitude / RAD,
                                   center.longitude / RAD,
                                   entity->area.points[0].latitude / RAD,
                                   entity->area.points[0].longitude / RAD, 
                                   &x0, &y0);
            } else { /* projection is orthographic */
                status0 = ortho(radius, center.latitude / RAD,
                                center.longitude / RAD,
                                entity->area.points[0].latitude / RAD,
                                entity->area.points[0].longitude / RAD, 
                                &x0, &y0);
	    }
            x0 += center_x;
            y0 = center_y - y0;
            if(projection) { /* projection is gnomonic */
                status1 = gnomonic(radius, center.latitude / RAD,
                                   center.longitude / RAD,
                                   entity->area.points[1].latitude / RAD,
                                   entity->area.points[1].longitude / RAD, 
                                   &x1, &y1);
            } else { /* projection is orthographic */
                status1 = ortho(radius, center.latitude / RAD,
                                center.longitude / RAD,
                                entity->area.points[1].latitude / RAD,
                                entity->area.points[1].longitude / RAD, 
                                &x1, &y1);
	    }
            x1 += center_x;
            y1 = center_y - y1;

	    /* now add the label if one will fit at all */
	    maxchars = (abs(x1 - x0) - 4) / gdFontW(gdFontMediumBold);
	    maxlines = (abs(y1 - y0) - 4) / (gdFontH(gdFontMediumBold) + 2);
	    if(!status0 && !status1 && maxchars > 4 && maxlines > 0)
	    {
		if(strlen(entity->label) <= maxchars)
                    gdImageStringFilled(image, gdFontMediumBold,
				(x0 + x1 - gdFontW(gdFontMediumBold) *
				(int)strlen(entity->label)) / 2,
				(y0 + y1 - gdFontH(gdFontMediumBold)) / 2,
				entity->label, entity->color, background);
		else
		{
		    /* the label will have to be broken onto several lines */
		    char *lines[100];
		    int spaces[100];
		    int i, numspaces = 0, start = 0, numlines = 0;
		    int ymin, zoneheight;

		    /* find all of the spaces */
		    for(i = 0; i < strlen(entity->label); i++)
		    {
			if(entity->label[i] == ' ')
			    spaces[numspaces++] = i;
		    }
		    spaces[numspaces++] = i;

		    /* now form the lines */
		    start = 0;
		    lines[numlines++] = entity->label + start;
		    for(i = 1; i < numspaces; i++)
		    {
			if(spaces[i] - start >= maxchars)
			{
			    /* split the line on the previous word */
			    entity->label[spaces[i - 1]] = '\0';
			    start = spaces[i - 1] + 1;
			    lines[numlines++] = entity->label + start;
			}
		    }

		    /* trim long lines to fit if necessary */
		    for(i = 0; i < numlines; i++)
			if(strlen(lines[i]) > maxchars)
			    lines[i][maxchars] = '\0';

		    /* now draw the lines */
		    if(numlines > maxlines)
			numlines = maxlines;
		    if(y0 < y1)
			ymin = y0;
		    else
			ymin = y1;
		    zoneheight = abs(y1 - y0) / numlines;

		    for(i = 0; i < numlines; i++)
		    {
                        gdImageStringFilled(image, gdFontMediumBold,
				(x0 + x1 - gdFontW(gdFontMediumBold) *
				(int)strlen(lines[i])) / 2,
				ymin + zoneheight * i +
				(zoneheight - gdFontH(gdFontMediumBold)) / 2,
				lines[i], entity->color, background);
		    }
		}
	    }
	}
	break;

	case POLYGON:
	if(entity->area.npts < 3)
	    break;
	for(i = 0; i < entity->area.npts - 1; i++)
	{
	    length = calcRoute(&entity->area.points[i],
				&entity->area.points[i + 1], route);
	    plot_route(&entity_gc, route, length, POLYLINE);
	}
	length = calcRoute(&entity->area.points[entity->area.npts - 1],
				&entity->area.points[0], route);
	plot_route(&entity_gc, route, length, POLYLINE);
	break;

	case POLYLINE:
	if(entity->area.npts < 2)
	    break;
	for(i = 0; i < entity->area.npts - 1; i++)
	{
	    length = calcRoute(&entity->area.points[i],
				&entity->area.points[i + 1], route);
	    plot_route(&entity_gc, route, length, POLYLINE);
	}
	break;

	case SPHERICAL_POLYGON:
        for(i = 0; i < entity->area.npts - 1; i++)
        {
	    length = calcGreatCircleRoute(&entity->area.points[i],
				&entity->area.points[i + 1], route);
	    plot_route(&entity_gc, route, length, POLYLINE);
        }
        length = calcGreatCircleRoute(
			&entity->area.points[entity->area.npts - 1],
			&entity->area.points[0], route);
        plot_route(&entity_gc, route, length, POLYLINE);
	break;

	case SPHERICAL_POLYLINE:
        for(i = 0; i < entity->area.npts - 1; i++)
        {
	    length = calcGreatCircleRoute(&entity->area.points[i],
				&entity->area.points[i + 1], route);
	    plot_route(&entity_gc, route, length, POLYLINE);
        }
	break;
    }
}

void
LoadEntities(FILE *fp)
{
    char line[BUFSIZ];
    char *field;
    Entity *hit;
    int i, failed;

    entities = 0;

    /* read records from fp */
    while(fgets(line, BUFSIZ, fp) != 0)
    {
        /* allocate new node */
        hit = (Entity *)malloc(sizeof(Entity));

	/* key */
	field = strtok(line, "|");
	if(field)
	    hit->key = atoi(field);
	else
	{
	    free(hit);
	    continue;
	}

	/* type */
	field = strtok((char *)0, "|");
	if(!strcmp(field, "point"))
	    hit->area.type = SINGLE_PT;
	else if(!strcmp(field, "range"))
	    hit->area.type = RANGE;
	else if(!strcmp(field, "polygon"))
	    hit->area.type = POLYGON;
	else if(!strcmp(field, "polyline"))
	    hit->area.type = POLYLINE;
	else if(!strcmp(field, "spherical_polygon"))
	    hit->area.type = SPHERICAL_POLYGON;
	else if(!strcmp(field, "sherical_polyline"))
	    hit->area.type = SPHERICAL_POLYLINE;
	else
	{
	    free(hit);
	    continue;
	}

	/* label */
	strncpy(hit->label, strtok((char *)0, "|"), 79);
	hit->label[79] = '\0';
	if(!strcmp(hit->label, "\"\""))
	    hit->label[0] = '\0';

	/* color */
	field = strtok((char *)0, "|");
	hit->color = allocate_color(image, field);

	/* width */
	field = strtok((char *)0, "|");
	if(field)
	    hit->line_width = atoi(field);
	else
	{
	    free(hit);
	    continue;
	}

	/* style */
	field = strtok((char *)0, "|");
	if(hit->area.type == SINGLE_PT)
	{
	    if(!strcmp(field, "large_dot"))
		hit->symbol = LARGE_DOT;
	    else if(!strcmp(field, "med_dot"))
		hit->symbol = MED_DOT;
	    else if(!strcmp(field, "small_dot"))
		hit->symbol = SMALL_DOT;
	    else if(!strcmp(field, "target"))
		hit->symbol = TARGET;
	    else if(!strcmp(field, "star"))
		hit->symbol = STAR;
	    else if(!strcmp(field, "cross"))
		hit->symbol = CROSS;
	    else
	    {
	        free(hit);
	        continue;
	    }
	}
	else
	{
	    if(!strcmp(field, "solid"))
		hit->line_style = SOLID;
	    else if(!strcmp(field, "dashed"))
	    {
		hit->line_style = BROKEN;
		hit->dash_style = DASHED;
	    }
	    else if(!strcmp(field, "dotted"))
	    {
		hit->line_style = BROKEN;
		hit->dash_style = DOTTED;
	    }
	    else
	    {
	        free(hit);
	        continue;
	    }
	}

	/* read in the coords */
	failed = 0;
	for(i = 0; (i < MAX_POLY_PTS) && (field = strtok((char *)0, "|")); i++)
	{
	    hit->area.points[i].latitude = atof(field);
	    field = strtok((char *)0, "|");
	    if(field)
		hit->area.points[i].longitude = atof(field);
	    else
	    {
		failed = 1;
		break;
	    }
	}
	if(failed)
	{
	    free(hit);
	    continue;
	}
	else
	    hit->area.npts = i;

        /* link into front of the list */
        hit->next = entities;
        entities = hit;
    }
}

void
DrawRefaid(void)
{
    register int i;
    int signbit1, signbit2;
    gdPoint gdpoints[256];
    Entity *hit;
    double tmp;
    double step;
    int resolution;
    int newres;
    int status;
    int center_x = width / 2;
    int center_y = height / 2;
    int min_dimension;
    int pseudoscale;
    int rtn, found_tiles = 0;
    POINT corner[6];
    float last_lon;
    int crossings;
    EXTENT extent;
    SCREENLABEL *listpnt, *listscan, *lastlistpnt;
    SCREENLABEL *screenlabellist = NULL;

    if(width <= height)
	min_dimension = width;
    else
	min_dimension = height;
    pseudoscale = zoom_percent * min_dimension / 500;

    /*  239 allowed a zoom of 4; 700 allowed a zoom of 16 */
    if(pseudoscale <= 479) /* changed to allow zoom of 8 */
        resolution = LOW_RES;
    else if(pseudoscale <= 6000)
        resolution = MED_RES;
    else if(pseudoscale <= 20000)
        resolution = HIGH_RES;
    else
        resolution = ULTRA_RES;

retry:

    /* find coordinates of image corners */
    /* upper left */
    if(projection) { /* projection is gnomonic */
        status = invgnomonic(radius, center.latitude / RAD,
                             center.longitude / RAD,
                             -center_x, center_y,
                             &corner[0].latitude, &corner[0].longitude);
    } else { /* projection is orthographic */
        status = invortho(radius, center.latitude / RAD,
                          center.longitude / RAD,
                          -center_x, center_y,
                          &corner[0].latitude, &corner[0].longitude);
    }
    /* center top */
    if(!status)
	if(projection) { /* projection is gnomonic */
            status = invgnomonic(radius, center.latitude / RAD,
                                 center.longitude / RAD,
                                 0, center_y,
                                 &corner[1].latitude, &corner[1].longitude);
	} else { /* projection is orthographic */
            status = invortho(radius, center.latitude / RAD,
                              center.longitude / RAD,
                              0, center_y,
                              &corner[1].latitude, &corner[1].longitude);
	}
    /* upper right */
    if(!status)
	if(projection) { /* projection is gnomonic */
            status = invgnomonic(radius, center.latitude / RAD,
                                 center.longitude / RAD,
                                 center_x, center_y,
                                 &corner[2].latitude, &corner[2].longitude);
        } else { /* projection is orthographic */
            status = invortho(radius, center.latitude / RAD,
                              center.longitude / RAD,
                              center_x, center_y,
                              &corner[2].latitude, &corner[2].longitude);
        }
    /* lower right */
    if(!status)
	if(projection) { /* projection is gnomonic */
            status = invgnomonic(radius, center.latitude / RAD,
                                 center.longitude / RAD,
                                 center_x, -center_y,
                                 &corner[3].latitude, &corner[3].longitude);
        } else { /* projection is orthographic */
            status = invortho(radius, center.latitude / RAD,
                              center.longitude / RAD,
                              center_x, -center_y,
                              &corner[3].latitude, &corner[3].longitude);
        }
    /* center bottom */
    if(!status)
	if(projection) { /* projection is gnomonic */
            status = invgnomonic(radius, center.latitude / RAD,
                                 center.longitude / RAD,
                                 0, -center_y,
                                 &corner[4].latitude, &corner[4].longitude);
        } else { /* projection is orthographic */
            status = invortho(radius, center.latitude / RAD,
                              center.longitude / RAD,
                              0, -center_y,
                              &corner[4].latitude, &corner[4].longitude);
        }
    /* lower left */
    if(!status)
        if(projection) { /* projection is gnomonic */
            status = invgnomonic(radius, center.latitude / RAD,
                                 center.longitude / RAD,
                                 -center_x, -center_y,
                                 &corner[5].latitude, &corner[5].longitude);
        } else { /* projection is orthographic */
            status = invortho(radius, center.latitude / RAD,
                              center.longitude / RAD,
                              -center_x, -center_y,
                              &corner[5].latitude, &corner[5].longitude);
        }

    if(!status)
    {
	/* check number of date line crossings */
	crossings = 0;
	last_lon = corner[5].longitude;
	for(i = 0; i < 6; i++)
	{
	    if(corner[i].longitude < 0)
		signbit1 = 1;
	    else
		signbit1 = 0;
	    if(last_lon < 0)
		signbit2 = 1;
	    else
		signbit2 = 0;
	    if((signbit1 != signbit2) &&
			(fabs((double)corner[i].longitude - last_lon) >= M_PI))
		crossings++;
	    last_lon = corner[i].longitude;
	}

	/* find the bounding box of the viewing area */
	if(center.latitude >= 0)
	{
	    extent.minlat = corner[5].latitude * RAD;
	    extent.maxlat = corner[1].latitude * RAD;
	}
	else
	{
	    extent.minlat = corner[4].latitude * RAD;
	    extent.maxlat = corner[0].latitude * RAD;
	}
	if(crossings % 2)	/* pole is visible */
	{
	    if(center.latitude >= 0)
		extent.maxlat = 90;
	    else
		extent.minlat = -90;
	    extent.minlon = -180;
	    extent.maxlon = 180;
	}
	else
	{
	    if(center.latitude >= 0)
	    {
		extent.minlon = corner[0].longitude * RAD;
		extent.maxlon = corner[2].longitude * RAD;
	    }
	    else
	    {
		extent.minlon = corner[5].longitude * RAD;
		extent.maxlon = corner[3].longitude * RAD;
	    }
	}

        /* determine what tiles we need */
	if(resolution != LOW_RES)
	{
            extent.minlonquad = lonquad(extent.minlon, TRUE);
            extent.maxlonquad = lonquad(extent.maxlon, FALSE);
            extent.minlatquad = latquad(extent.minlat, TRUE);
            extent.maxlatquad = latquad(extent.maxlat, FALSE);
            extent.minlontile = latlontile(extent.minlon, TRUE);
            extent.maxlontile = latlontile(extent.maxlon, FALSE);
            extent.minlattile = latlontile(extent.minlat, TRUE);
            extent.maxlattile = latlontile(extent.maxlat, FALSE);
            extent.minlonsubtile = latlonsubtile(extent.minlon, TRUE);
            extent.maxlonsubtile = latlonsubtile(extent.maxlon, FALSE);
            extent.minlatsubtile = latlonsubtile(extent.minlat, TRUE);
            extent.maxlatsubtile = latlonsubtile(extent.maxlat, FALSE);
	}
    }
    else
    {
	extent.minlat = -90;
	extent.maxlat = 90;
	extent.minlon = -180;
	extent.maxlon = 180;
    }

    /* draw the outer circle */
    if(resolution == LOW_RES)
    {
        step = 3 / RAD;
        for(tmp = 0, i = 0; tmp <= M_2PI; tmp += step, i++)
        {
            gdpoints[i].x = (cos(tmp) * radius) + center_x;
            gdpoints[i].y = (sin(tmp) * radius) + center_y;
        }
        gdpoints[i].x = gdpoints[0].x;
        gdpoints[i].y = gdpoints[0].y;
        gdImagePolyline(image, gdpoints, i + 1, grid_gc.color);
    }

    /* draw map data */
    rtn = draw_layer(resolution, OCPOLY, &extent, &screenlabellist);
    if(rtn != -2)
        found_tiles = 1;
    if(water_layer)
    {
	rtn = draw_layer(resolution, DNPOLY, &extent, &screenlabellist);
	if(rtn != -2)
            found_tiles = 1;
	rtn = draw_layer(resolution, DNLINE, &extent, &screenlabellist);
	if(rtn != -2)
            found_tiles = 1;
    }
    if(political_layer)
    {
	rtn = draw_layer(resolution, POLINE, &extent, &screenlabellist);
	if(rtn != -2)
            found_tiles = 1;
    }
    if(rail_layer)
    {
	rtn = draw_layer(resolution, RRLINE, &extent, &screenlabellist);
	if(rtn != -2)
            found_tiles = 1;
    }
    if(roads_layer)
    {
	rtn = draw_layer(resolution, RDLINE, &extent, &screenlabellist);
	if(rtn != -2)
            found_tiles = 1;
    }
    /* don't bother plotting napp_layer if not on ULTRARES layer */
    if(resolution == ULTRA_RES && napp_layer)
    {
	rtn = draw_layer(resolution, NAPPPOINT, &extent, &screenlabellist);
	if(rtn != -2)
            found_tiles = 1;
    }
    /* don't bother plotting nhap_bw_layer if not on ULTRARES layer */
    if(resolution == ULTRA_RES && nhap_bw_layer)
    {
	rtn = draw_layer(resolution, NHAPBWPOINT, &extent, &screenlabellist);
	if(rtn != -2)
            found_tiles = 1;
    }
    /* don't bother plotting nhap_cir_layer if not on ULTRARES layer */
    if(resolution == ULTRA_RES && nhap_cir_layer)
    {
	rtn = draw_layer(resolution, NHAPCIRPOINT, &extent, &screenlabellist);
	if(rtn != -2)
            found_tiles = 1;
    }
    /* don't bother plotting wrs1_layer if not on MEDRES layer */
    if(resolution == MED_RES && wrs1_layer)
    {
	rtn = draw_layer(resolution, WRS1POINT, &extent, &screenlabellist);
	if(rtn != -2)
            found_tiles = 1;
    }
    /* don't bother plotting wrs2_layer if not on MEDRES layer */
    if(resolution == MED_RES && wrs2_layer)
    {
	rtn = draw_layer(resolution, WRS2POINT, &extent, &screenlabellist);
	if(rtn != -2)
            found_tiles = 1;
    }
    /* don't bother plotting populated places if no other layers are there */
    if(found_tiles && populated_places_layer)
    {
	if(resolution == HIGH_RES)
	    rtn = draw_layer(resolution, PPPOLY, &extent, &screenlabellist);

	/* if the pseudoscale indicates ULTRA_RES but no linework was available
           try to use the ULTRA_RES pppoint data with the HIGH_RES linework */
        if(resolution == HIGH_RES && pseudoscale > 20000)
        {
            rtn = draw_layer(ULTRA_RES, PPPOINT, &extent, &screenlabellist);
            if(rtn == -2)
                rtn = draw_layer(HIGH_RES, PPPOINT, &extent, &screenlabellist);
        }
        else
	    rtn = draw_layer(resolution, PPPOINT, &extent, &screenlabellist);
    }

    /* if no tiles exist in this area, adjust zoom level and try again */
    if(!found_tiles)
    {
        switch(resolution)
        {
            case ULTRA_RES:
            if(pseudoscale > 40000)
                zoom_percent = 40000 * 500 / min_dimension;
            resolution = HIGH_RES;
	    break;

            case HIGH_RES:
            if(pseudoscale > 20000)
                zoom_percent = 20000 * 500 / min_dimension;
            resolution = MED_RES;
	    break;

            case MED_RES:
            if(pseudoscale > 1000)
                zoom_percent = 1000 * 500 / min_dimension;
            resolution = LOW_RES;
	    break;

            case LOW_RES:
            if(zoom_percent != DEFAULT_ZOOM)
                zoom_percent = DEFAULT_ZOOM;
	    break;
        }
        radius = min_dimension * zoom_percent / 200;
        pseudoscale = zoom_percent * min_dimension / 500;

        /* free the screenlabellist */
        while(screenlabellist)
        {
            listpnt = screenlabellist->next;
            free(screenlabellist);
            screenlabellist = listpnt;
        }

        goto retry;
    }

    /* write the actual zoom_percent to stdout if zoomreportonly specified */
    if(zoom_report_only)
    {
	printf("%d", zoom_percent);
	exit(0);
    }

    /* add the grid lines */
    DrawGrid(&extent);
    LabelGrid(&extent, &screenlabellist);

    /* now draw all of the screen labels without clobber */
    for(listpnt = screenlabellist; listpnt; listpnt = listpnt->next)
    {
        for(listscan = screenlabellist; listscan != listpnt;
                                                listscan = listscan->next)
        {
            if(listscan->xmin <= listpnt->xmax &&
                        listscan->xmax >= listpnt->xmin &&
                        listscan->ymin <= listpnt->ymax &&
                        listscan->ymax >= listpnt->ymin)
                break;
        }
        if(listscan == listpnt)
        {
            /* no clobber, so draw this one */
            gdImageString(image, listpnt->font, listpnt->x, listpnt->y,
			listpnt->label, label_gc.color);
        }
        else
        {
            /* delete this one to avoid clobbering an earlier or later one */
            lastlistpnt->next = listpnt->next;
            free(listpnt);
            listpnt = lastlistpnt;
        }
        lastlistpnt = listpnt;
    }

    /* now free the screenlabellist */
    while(screenlabellist)
    {
        listpnt = screenlabellist->next;
        free(screenlabellist);
        screenlabellist = listpnt;
    }

    /* Draw the entities */
    for(hit = entities; hit; hit = hit->next)
	plot_entity(hit);
}

static void
DrawGrid(EXTENT *extent)
{
    int npts;
    float i, j, increment, maxlon;
    POINT route[32768];

    if((extent->minlat < -90) || (extent->maxlat > 90) ||
				(extent->maxlat < extent->minlat))
        return;

    /* wrap around at date line */
    maxlon = extent->maxlon;
    while(maxlon < extent->minlon)
	maxlon += 360;

    if(grid_increment >= 0)
	increment = grid_increment;
    else
    {
	/* auto grid */
	if((extent->maxlat - extent->minlat) < (maxlon - extent->minlon))
	    increment = (extent->maxlat - extent->minlat) / 3;
	else
	    increment = (maxlon - extent->minlon) / 3;
	if(increment > 45)
	    increment = 45;
	else if(increment > 30)
	    increment = 30;
	else if(increment > 20)
	    increment = 20;
	else if(increment > 15)
	    increment = 15;
	else if(increment > 10)
	    increment = 10;
	else if(increment > 5)
	    increment = 5;
	else if(increment > 2.5)
	    increment = 2.5;
	else if(increment > 1)
	    increment = 1;
	else if(increment > 0.5)
	    increment = 0.5;
        else if(increment > 0.25)
            increment = 0.25;
        else if(increment > 0.125)
            increment = 0.125;
        else if(increment > 0.0625)
            increment = 0.0625;
        else
            increment = (float)1 / 60;
    }

    /* draw grid lines */
    if(increment > 0)
    {
        /* draw parallels */
        for(i = (int)(extent->maxlat / increment) * increment;
			i > extent->minlat; i -= increment)
        {
	    if(fabs((double)i) == 90)
		continue;

	    npts = 0;
            for(j = extent->minlon; j < maxlon + 1; j++)
            {
		route[npts].latitude = i;
		route[npts].longitude = j;
		npts++;
	    }
	    plot_route(&grid_gc, route, npts, POLYLINE);
        }

        /* draw meridians */
        for(i = (int)(maxlon / increment) * increment;
			i > extent->minlon; i -= increment)
        {
	    npts = 0;
            for(j = extent->minlat; j < extent->maxlat + 1; j++)
            {
		route[npts].latitude = j;
		route[npts].longitude = i;
		npts++;
	    }
	    plot_route(&grid_gc, route, npts, POLYLINE);
        }
    }
}

static void
LabelGrid(EXTENT *extent, SCREENLABEL **screenlabellist)
{
    int status, inside, last_inside;
    unsigned char label[32];
    float i, j, increment, maxlon, step;
    int x, y;
    int center_x = width / 2;
    int center_y = height / 2;
    double center_lat = center.latitude / RAD;
    double center_lon = center.longitude / RAD;
    SCREENLABEL *labelpnt;

    /* wrap around at date line */
    maxlon = extent->maxlon;
    while(maxlon < extent->minlon)
	maxlon += 360;

    if(grid_increment >= 0)
	increment = grid_increment;
    else
    {
	/* auto grid */
	if((extent->maxlat - extent->minlat) < (maxlon - extent->minlon))
	    increment = (extent->maxlat - extent->minlat) / 3;
	else
	    increment = (maxlon - extent->minlon) / 3;
	if(increment > 45)
	    increment = 45;
	else if(increment > 30)
	    increment = 30;
	else if(increment > 20)
	    increment = 20;
	else if(increment > 15)
	    increment = 15;
	else if(increment > 10)
	    increment = 10;
	else if(increment > 5)
	    increment = 5;
	else if(increment > 2.5)
	    increment = 2.5;
	else if(increment > 1)
	    increment = 1;
	else if(increment > 0.5)
	    increment = 0.5;
        else if(increment > 0.25)
            increment = 0.25;
        else if(increment > 0.125)
            increment = 0.125;
        else if(increment > 0.0625)
            increment = 0.0625;
        else
            increment = (float)1 / 60;
    }

    if((increment == 0) || ((extent->maxlat == 90) && (extent->minlat == -90)))
	return;

    /* determine a step to make sure labels appear accurately */
    step = (float)200 / zoom_percent;

    /* label parallels */
    for(i = (int)(extent->maxlat / increment) * increment;
			i > extent->minlat; i -= increment)
    {
	if(fabs((double)i) == 90)
	    continue;

	if(projection) { /* projection is gnomonic */
            status = gnomonic(radius, center_lat, center_lon, i / RAD,
                              extent->minlon / RAD, &x, &y);
        } else { /* projection is orthographic */
            status = ortho(radius, center_lat, center_lon, i / RAD,
                           extent->minlon / RAD, &x, &y);
        }
	if(!status && (abs(x) < center_x) && (abs(y) < center_y))
	    last_inside = inside = TRUE;
	else
	    last_inside = inside = FALSE;
        for(j = extent->minlon + step; j <= maxlon; j += step)
        {
            if(projection) { /* projection is gnomonic */
                status = gnomonic(radius, center_lat, center_lon, i / RAD,
                                  j / RAD, &x, &y);
            } else { /* projection is orthographic */
                status = ortho(radius, center_lat, center_lon, i / RAD,
                               j / RAD, &x, &y);
            }
	    if(!status && (abs(x) < center_x) && (abs(y) < center_y))
		inside = TRUE;
	    else
		inside = FALSE;
	    if(inside != last_inside)
	    {
		x += center_x;
		y = center_y - y;
		if((x < center_x) && (y <= height - 20))
		{
                    char dir;
                    int deg = fabs(i);
                    int min = (fabs(i) - deg) * 60;
                    int sec = rint((fabs(i) - deg - (double)min / 60) * 3600);
                    if(sec == 60)
                    {
                        sec = 0;
                        min++;
                        if(min == 60)
                        {
                            min = 0;
                            deg++;
                        }
                    }
                    if(i < 0)
                        dir = 'S';
                    else
                        dir = 'N';

                    /* add this label to the list */
		    labelpnt = (SCREENLABEL *)malloc(sizeof(SCREENLABEL));
                    if(increment == (int)increment)
                        sprintf(labelpnt->label, "%c%d\260", dir, deg);
                    else if((increment * 60) == (int)(increment * 60))
                        sprintf(labelpnt->label, "%c%d\260%d'", dir, deg, min);
                    else
                        sprintf(labelpnt->label, "%c%d\260%d'%d\"",
							dir, deg, min, sec);
		    labelpnt->font = gdFontMediumBold;
                    labelpnt->xmin = labelpnt->x = 2;
                    labelpnt->xmax = labelpnt->xmin +
			   gdFontW(gdFontMediumBold) * (int)strlen(labelpnt->label);
                    labelpnt->y = labelpnt->ymin = y -
					gdFontH(gdFontMediumBold) - 2;
                    labelpnt->ymax = labelpnt->ymin + gdFontH(gdFontMediumBold);
                    labelpnt->next = *screenlabellist;
                    *screenlabellist = labelpnt;
		}
	        last_inside = inside;
	    }
	}
    }

    /* label meridians */
    for(i = (int)(maxlon / increment) * increment;
			i > extent->minlon; i -= increment)
    {
        if(projection) { /* projection is gnomonic */
            status = gnomonic(radius, center_lat, center_lon, 
                              extent->minlat / RAD, i / RAD, &x, &y);
        } else { /* projection is orthographic */
            status = ortho(radius, center_lat, center_lon, 
                           extent->minlat / RAD, i / RAD, &x, &y);
        }
	if(!status && (abs(x) < center_x) && (abs(y) < center_y))
	    last_inside = inside = TRUE;
	else
	    last_inside = inside = FALSE;
        for(j = extent->minlat + step; j <= extent->maxlat; j += step)
        {
            if(projection) { /* projection is gnomonic */
                status = gnomonic(radius, center_lat, center_lon, j / RAD,
                                  i / RAD, &x, &y);
            } else { /* projection is orthographic */
                status = ortho(radius, center_lat, center_lon, j / RAD,
                               i / RAD, &x, &y);
            }
	    if(!status && (abs(x) < center_x) && (abs(y) < center_y))
		inside = TRUE;
	    else
		inside = FALSE;
	    if(inside != last_inside)
	    {
		x += center_x;
		y = center_y - y;
		if(y > center_y)
		{
                    char dir;
                    int deg = fabs(i);
                    int min = (fabs(i) - deg) * 60;
                    int sec = rint((fabs(i) - deg - (double)min / 60) * 3600);
                    if(sec == 60)
                    {
                        sec = 0;
                        min++;
                        if(min == 60)
                        {
                            min = 0;
                            deg++;
                        }
                    }
                    if(deg > 180)
                    {
                        deg = (int)fabs(i - 360);
                        if(i - 360 < 0)
                            dir = 'W';
                        else
                            dir = 'E';
                    }
                    else
                    {
                        if(i < 0)
                            dir = 'W';
                        else
                            dir = 'E';
                    }

                    /* add this label to the list */
                    labelpnt = (SCREENLABEL *)malloc(sizeof(SCREENLABEL));
                    if(increment == (int)increment)
                        sprintf(labelpnt->label, "%c%d\260", dir, deg);
                    else if((increment * 60) == (int)(increment * 60))
                        sprintf(labelpnt->label, "%c%d\260%d'", dir, deg, min);
                    else
                        sprintf(labelpnt->label, "%c%d\260%d'%d\"",
							dir, deg, min, sec);
		    labelpnt->font = gdFontMediumBold;
                    labelpnt->xmin = labelpnt->x = x + 2;
                    labelpnt->xmax = labelpnt->xmin +
			   gdFontW(gdFontMediumBold) * (int)strlen(labelpnt->label);
                    labelpnt->y = labelpnt->ymin = height -
					gdFontH(gdFontMediumBold);
                    labelpnt->ymax = labelpnt->ymin +
					gdFontH(gdFontMediumBold);
                    labelpnt->next = *screenlabellist;
                    *screenlabellist = labelpnt;
		    break;
		}
		last_inside = inside;
	    }
	}
    }
}

static char
latquad(float lat, int minflag)
{
    char latquad;

    if(lat >= 30) /* greater than 'i' */
        latquad = 'j' + ((int)(lat -  30)) / 15;
    else
	latquad = 'a' + ((int)(lat + 90)) / 15;

    if((!minflag) && (fmod(lat, 15) == 0))
    {
	latquad--;
	if(latquad == 'i')
	    latquad--;
    }
    return(latquad);
}

static char
lonquad(float lon, int minflag)
{
    char lonquad;

    if(lon >= 15) /* greater than 'o' */
        lonquad = 'p' + ((int)(lon - 15)) / 15;
    else if(lon >= -60) /* greater than 'i' */
        lonquad = 'j' + ((int)(lon + 60)) / 15;
    else
	lonquad = 'a' + ((int)(lon + 180)) / 15;

    if((!minflag) && (fmod(lon, 15) == 0))
    {
	lonquad--;
	if((lonquad == 'i') || (lonquad == 'o'))
	    lonquad--;
    }
    return(lonquad);
}

static int
latlontile(float coord, int minflag)
{
    int rtn;
    float modval;

    modval = fabs(fmod((double)coord, (double)15));

    if(modval == 0)
    {
	/* right on the quad boundary */
	if(minflag)
	    rtn = 1;
	else
	    rtn = 6;
    }
    else
    {
	/* inside of a quad */
	if(coord >= 0)
	    rtn = 1 + (int)(modval / 2.5);
        else
	    rtn = 6 - (int)(modval / 2.5);

	if(fmod((double)modval, (double)2.5) == 0)
	{
	    /* right on a tile boundary inside of a quad */
	    if(!minflag && (coord > 0))
		rtn--;
	    else if(minflag && (coord < 0))
		rtn++;
	}
    }
    return(rtn);
}

static int
latlonsubtile(float coord, int minflag)
{
    int rtn;
    float modval;

    modval = fabs(fmod((double)coord, (double)2.5));

    if(modval == 0)
    {
        /* right on the tile boundary */
        if(minflag)
            rtn = 1;
        else
            rtn = 5;
    }
    else
    {
        /* inside of a tile */
        if(coord >= 0)
            rtn = 1 + (int)(modval / 0.5);
        else
            rtn = 5 - (int)(modval / 0.5);

        if(fmod((double)modval, (double)0.5) == 0)
        {
            /* right on a subtile boundary inside of a tile */
            if(!minflag && (coord > 0))
                rtn--;
            else if(minflag && (coord < 0))
                rtn++;
        }
    }
    return(rtn);
}

int
draw_layer(int resolution, int layer, EXTENT *extent,
				SCREENLABEL **screenlabellist)
{
    register int k, m, n, o;
    char i, j, max;
    char fname[1024];
    int rtn;
    int found = 0;
    static char *names[] = { "poline", "ocpoly", "dnline", "dnpoly", \
	"pppoint", "pppoly", "rdline", "rrline", "nappflightstations",
	"nhap_bw", "nhap_cir", "landsatwrs1", "landsatwrs2" };

    /* correct for date line crossing */
    if(resolution != LOW_RES)
    {
        if(extent->maxlonquad < extent->minlonquad)
	    max = 'z' + 1;
        else
	    max = extent->maxlonquad;
    }

    /* set up the GCs */
    if(resolution == ULTRA_RES)
    {
      switch(layer)
      {
        case RDLINE:
	map_gc[1].color = magenta;
	map_gc[1].line_width = 3;
	map_gc[1].line_style = SOLID;
	map_gc[2].color = magenta;
	map_gc[2].line_width = 2;
	map_gc[2].line_style = SOLID;
	map_gc[3].color = magenta;
	map_gc[3].line_width = 2;
	map_gc[3].line_style = SOLID;
	map_gc[4].color = magenta;
	map_gc[4].line_width = 2;
	map_gc[4].line_style = SOLID;
	map_gc[5].color = magenta;
	map_gc[5].line_width = 2;
	map_gc[5].line_style = SOLID;
	map_gc[6].color = magenta;
	map_gc[6].line_width = 1;
	map_gc[6].line_style = SOLID;
	map_gc[7].color = magenta;
	map_gc[7].line_width = 1;
	map_gc[7].line_style = SOLID;
	map_gc[10].color = magenta;
	map_gc[10].line_width = 1;
	map_gc[10].line_style = SOLID;
	map_gc[11].color = magenta;
	map_gc[11].line_width = 1;
	map_gc[11].line_style = SOLID;
	map_gc[12].color = magenta;
	map_gc[12].line_width = 1;
	map_gc[12].line_style = SOLID;
	map_gc[15].color = magenta;
	map_gc[15].line_width = 1;
	map_gc[15].line_style = SOLID;
	map_gc[8].color = magenta;
	map_gc[8].line_width = 1;
	map_gc[8].line_style = BROKEN;
	map_gc[8].dash_style = DASHED;
	map_gc[13].color = magenta;
	map_gc[13].line_width = 1;
	map_gc[13].line_style = BROKEN;
	map_gc[13].dash_style = DASHED;
	map_gc[14].color = magenta;
	map_gc[14].line_width = 1;
	map_gc[14].line_style = BROKEN;
	map_gc[14].dash_style = DASHED;
	map_gc[9].color = magenta;
	map_gc[9].line_width = 1;
	map_gc[9].line_style = BROKEN;
	map_gc[9].dash_style = DOTTED;
        break;

        case DNLINE:
	map_gc[1].color = blue;
	map_gc[1].line_width = 1;
	map_gc[1].line_style = SOLID;
	map_gc[2].color = blue;
	map_gc[2].line_width = 1;
	map_gc[2].line_style = BROKEN;
	map_gc[2].dash_style = DASHED;
	map_gc[3].color = blue;
	map_gc[3].line_width = 1;
	map_gc[3].line_style = BROKEN;
	map_gc[3].dash_style = DASHED;
	map_gc[4].color = blue;
	map_gc[4].line_width = 1;
	map_gc[4].line_style = BROKEN;
	map_gc[4].dash_style = DASHED;
	map_gc[5].color = blue;
	map_gc[5].line_width = 1;
	map_gc[5].line_style = BROKEN;
	map_gc[5].dash_style = DOTTED;
        break;

        case DNPOLY:
	map_gc[1].color = blue;
	map_gc[2].color = blue;
	map_gc[3].color = blue;
	map_gc[4].color = white;
	map_gc[5].color = cyan;
	map_gc[6].color = cyan;
	map_gc[7].color = cyan;
	map_gc[8].color = blue;
        break;

        case POLINE:
	map_gc[1].color = black;
	map_gc[1].line_width = 3;
	map_gc[1].line_style = SOLID;
	map_gc[2].color = black;
	map_gc[2].line_width = 2;
	map_gc[2].line_style = SOLID;
	map_gc[3].color = black;
	map_gc[3].line_width = 1;
	map_gc[3].line_style = BROKEN;
	map_gc[3].dash_style = DASHED;
        break;

	case RRLINE:
	map_gc[1].color = orange;
	map_gc[1].line_width = 1;
	map_gc[1].line_style = SOLID;
	map_gc[10].color = orange;
	map_gc[10].line_width = 1;
	map_gc[10].line_style = SOLID;
	map_gc[2].color = orange;
	map_gc[2].line_width = 1;
	map_gc[2].line_style = BROKEN;
	map_gc[2].dash_style = DASHED;
	map_gc[5].color = orange;
	map_gc[5].line_width = 1;
	map_gc[5].line_style = BROKEN;
	map_gc[5].dash_style = DASHED;
	map_gc[7].color = orange;
	map_gc[7].line_width = 1;
	map_gc[7].line_style = BROKEN;
	map_gc[7].dash_style = DOTTED;
	map_gc[8].color = orange;
	map_gc[8].line_width = 1;
	map_gc[8].line_style = BROKEN;
	map_gc[8].dash_style = DASHED;
	break;

        case OCPOLY:
	map_gc[1].color = blue;
	map_gc[2].color = black;
      }
    }
    else
    {
      switch(layer)
      {
	case POLINE:
	map_gc[1].color = black;
	if(resolution == HIGH_RES)
	    map_gc[1].line_width = 3;
	else
	    map_gc[1].line_width = 2;
	map_gc[1].line_style = SOLID;

	map_gc[2].color = black;
	if(resolution == HIGH_RES)
	    map_gc[2].line_width = 2;
	else
	    map_gc[2].line_width = 1;
	map_gc[2].line_style = SOLID;

	map_gc[3].color = black;
	map_gc[3].line_width = 1;
	map_gc[3].line_style = BROKEN;
	map_gc[3].dash_style = DASHED;

	map_gc[4].color = black;
	map_gc[4].line_width = 1;
	map_gc[4].line_style = BROKEN;
	map_gc[4].dash_style = DOTTED;

	map_gc[5].color = black;
	map_gc[5].line_width = 1;
	map_gc[5].line_style = BROKEN;
	map_gc[5].dash_style = DOTTED;

	map_gc[6].color = black;
	map_gc[6].line_width = 3;
	map_gc[6].line_style = BROKEN;
	map_gc[6].dash_style = DASHED;
	break;

	case OCPOLY:
	map_gc[1].color = blue;
	map_gc[2].color = blue;
	map_gc[3].color = blue;
	map_gc[4].color = blue;
	map_gc[5].color = white;
	break;

	case DNLINE:
	map_gc[1].color = blue;
	map_gc[1].line_width = 1;
	map_gc[1].line_style = SOLID;

	map_gc[2].color = blue;
	map_gc[2].line_width = 1;
	map_gc[2].line_style = SOLID;

	map_gc[3].color = blue;
	map_gc[3].line_width = 1;
	map_gc[3].line_style = SOLID;

	map_gc[4].color = blue;
	map_gc[4].line_width = 1;
	if(resolution == HIGH_RES)
	{
	    map_gc[4].line_style = BROKEN;
	    map_gc[4].dash_style = DASHED;
	}
	else
	    map_gc[4].line_style = SOLID;

	map_gc[9].color = blue;
	map_gc[9].line_width = 1;
	map_gc[9].line_style = SOLID;
	break;

	case DNPOLY:
	map_gc[1].color = blue;
	if(resolution == HIGH_RES)
	{
	    map_gc[2].color = cyan;
	    map_gc[3].color = cyan;
	    map_gc[4].color = white;
	    map_gc[5].color = white;
	}
	else
	{
	    map_gc[2].color = blue;
	    map_gc[3].color = blue;
	    map_gc[4].color = cyan;
	    map_gc[5].color = cyan;
	}
	map_gc[8].color = white;
	break;

	case PPPOLY:
	map_gc[1].color = red;
	map_gc[2].color = red;
	break;

	case RDLINE:
	map_gc[1].color = magenta;
	map_gc[1].line_width = 2;
	map_gc[1].line_style = SOLID;

	map_gc[2].color = magenta;
	map_gc[2].line_width = 1;
	map_gc[2].line_style = SOLID;

	map_gc[3].color = magenta;
	map_gc[3].line_width = 1;
	map_gc[3].line_style = BROKEN;
	map_gc[3].dash_style = DASHED;

	map_gc[4].color = magenta;
	map_gc[4].line_width = 1;
	map_gc[4].line_style = BROKEN;
	map_gc[4].dash_style = DASHED;
	break;

	case RRLINE:
	map_gc[1].color = orange;
	if(resolution == HIGH_RES)
	    map_gc[1].line_width = 1;
	else
	    map_gc[1].line_width = 2;
	map_gc[1].line_style = SOLID;

	map_gc[2].color = orange;
	if(resolution == HIGH_RES)
	    map_gc[2].line_width = 2;
	else
	    map_gc[2].line_width = 1;
	map_gc[2].line_style = SOLID;

	map_gc[3].color = orange;
	map_gc[3].line_width = 1;
	map_gc[3].line_style = BROKEN;
	map_gc[3].dash_style = DASHED;

	map_gc[4].color = orange;
	map_gc[4].line_width = 1;
	map_gc[4].line_style = BROKEN;
	map_gc[4].dash_style = DASHED;
	break;
      }
    }

    if(resolution == LOW_RES)
    {
      sprintf(fname, "%s/lowres/%s", mapdir, names[layer]);
      rtn = plot_tile(fname, layer, resolution, screenlabellist);
      if(!rtn)
	  found = 1;
    }
    else
    {
      /* for each 15 degree longitude */
      for(i = extent->minlonquad; i <= max; i++)
      {
	/* skip unused values */
	if((i == 'i') || (i == 'o'))
	  continue;

	/* wrap around at date line */
	if(i > 'z')
	{
	    i = 'a';
	    max = extent->maxlonquad;
	}

	/* for each 15 degree latitude */
	for(j = extent->maxlatquad; j >= extent->minlatquad; j--)
	{
	  /* skip unused value */
	  if(j == 'i')
	    continue;

	  if(resolution == MED_RES)
	  {
	    /* build the path name to this quad */
	    sprintf(fname, "%s/medres/%c/%c/%s", mapdir, i, j, names[layer]);

	    /* plot the tile */
	    rtn = plot_tile(fname, layer, resolution, screenlabellist);
            if(!rtn)
	        found = 1;
	  }
	  else
	  {
	    /* for each 2.5 degree longitude */
	    for(k = 1; k <= 6; k++)
	    {
	      if(((i == extent->minlonquad) && (k < extent->minlontile)) ||
			((i == extent->maxlonquad) && (k > extent->maxlontile)))
	        continue;

	      /* for each 2.5 degree latitude */
	      for(m = 6; m >= 1; m--)
	      {
	        if(((j == extent->maxlatquad) && (m > extent->maxlattile)) ||
			((j == extent->minlatquad) && (m < extent->minlattile)))
	          continue;

		if(resolution == HIGH_RES)
		{
	          /* build the path name to this tile */
	          sprintf(fname, "%s/highres/%c/%c/%d%d/%s", mapdir, i, j, k, m,
								names[layer]);

	          /* plot the tile */
	          rtn = plot_tile(fname, layer, resolution, screenlabellist);
                  if(!rtn)
	              found = 1;
		}
		else
		{
		  /* for each 0.5 degree longitude */
		  for(n = 1; n <= 5; n++)
		  {
		    if(((k == extent->minlontile) &&
			(n < extent->minlonsubtile)) ||
			((k == extent->maxlontile) &&
			(n > extent->maxlonsubtile)))
		      continue;

		    /* for each 0.5 degree latitude */
		    for(o = 5; o >= 1; o--)
		    {
		      if(((m == extent->maxlattile) &&
                          (o > extent->maxlatsubtile)) ||
                          ((m == extent->minlattile) &&
                          (o < extent->minlatsubtile)))
                        continue;

                      if(resolution == ULTRA_RES)
                      {
                        /* build the path name to this subtile */
                        sprintf(fname, "%s/ultrares/%c/%c/%d%d/%d%d/%s",
					mapdir, i, j, k, m, n, o, names[layer]);

	                /* plot the tile */
	                rtn = plot_tile(fname, layer, resolution,
							screenlabellist);
                        if(!rtn)
	                    found = 1;


		      }
		    }
		  }
		}
	      }
	    }
	  }
	}
      }
    }
    if(found)
        return(0);
    else
	return(-2);
}

static int
plot_tile(char *fname, int layer, int resolution, SCREENLABEL **screenlabellist)
{
    register int npts;
    FILE *fp;
    POINTHEADER pointheader;
    char label[30];
    LINEHEADER lineheader;
    LONG_NODE long_node;
    SHORT_NODE short_node;
    POINT route[32768];
    int center_x = width / 2;
    int center_y = height / 2;
    float center_lat = center.latitude / RAD;
    float center_lon = center.longitude / RAD;
    int status, x, y, pseudoscale, min_dimension;
    int skip_streets = 0;
    int skip_streams = 0;
    int skip_ponds = 0;
    int skip_states = 0;
    SCREENLABEL *labelpnt;

    if(!(fp = fopen(fname, "r")))
        return(-1);

    if(width <= height)
        min_dimension = width;
    else
        min_dimension = height;
    pseudoscale = zoom_percent * min_dimension / 500;

    if(!zoom_report_only)
    {
      switch(layer)
      {
	case PPPOINT:
        while(fread((char *)&pointheader, sizeof(POINTHEADER), 1, fp))
        {
	    if(!fread((char *)&label, (int)pointheader.label_length, 1, fp))
		break;
	
	    pointheader.lat = (long)ntohl((unsigned long)pointheader.lat);
	    pointheader.lon = (long)ntohl((unsigned long)pointheader.lon);
            if(projection) { /* projection is gnomonic */
                status = gnomonic(radius, center_lat, center_lon,
                                  (float)pointheader.lat * 0.00001 / RAD,
                                  (float)pointheader.lon * 0.00001 / RAD, 
                                  &x, &y);
            } else { /* projection is orthographic */
                status = ortho(radius, center_lat, center_lon,
                               (float)pointheader.lat * 0.00001 / RAD,
                               (float)pointheader.lon * 0.00001 / RAD, &x, &y);
            }

	    /* check if point is visible */
            if(status != -1 && abs(x) <= width / 2 && abs(y) <= height / 2)
            {
		x += center_x;
                y = center_y - y;
		switch(resolution)
		{
		    case LOW_RES:
		    case MED_RES:
                    if(zoom_percent < 200)
                        break;
                    if(resolution == MED_RES)
                    {
                        /* correct error in attribution */
                        if(pointheader.attribute_level == 2)
                            pointheader.attribute_level = 3;
                        else if(pointheader.attribute_level == 3)
                            pointheader.attribute_level = 2;
                    }
		    switch(pointheader.attribute_level)
		    {
		        case 1:
		        gdImageCopy(image, star,
				x - gdImageSX(star) / 2,
				y - gdImageSY(star) / 2,
				0, 0, gdImageSX(star), gdImageSY(star));
                        /* add this string to the list */
                        labelpnt = (SCREENLABEL *)malloc(sizeof(SCREENLABEL));
                        strcpy(labelpnt->label, label);
			labelpnt->font = gdFontMediumBold;
                        labelpnt->xmin = labelpnt->x = x + gdImageSX(star) / 2;
                        labelpnt->xmax = labelpnt->xmin +
				gdFontW(gdFontMediumBold) *
				(pointheader.label_length - 1);
                        labelpnt->y = y - gdFontH(gdFontMediumBold) / 2;
                        labelpnt->ymin = labelpnt->y -
				gdFontH(gdFontMediumBold) / 2;
                        labelpnt->ymax = labelpnt->y +
				gdFontH(gdFontMediumBold) / 2;
                        labelpnt->attribute = pointheader.attribute_level;
                        addlabeltolist(screenlabellist, labelpnt);
		        break;

		        case 2:
		        gdImageCopy(image, large_dot,
				x - gdImageSX(large_dot) / 2,
				y - gdImageSY(large_dot) / 2,
				0, 0, gdImageSX(large_dot),
				gdImageSY(large_dot));
                        /* add this string to the list */
                        labelpnt = (SCREENLABEL *)malloc(sizeof(SCREENLABEL));
                        strcpy(labelpnt->label, label);
			labelpnt->font = gdFontMediumBold;
                        labelpnt->xmin = labelpnt->x = x +
				gdImageSX(large_dot) / 2;
                        labelpnt->xmax = labelpnt->xmin +
				gdFontW(gdFontMediumBold) *
				(pointheader.label_length - 1);
                        labelpnt->y = y - gdFontH(gdFontMediumBold) / 2;
                        labelpnt->ymin = labelpnt->y -
				gdFontH(gdFontMediumBold) / 2;
                        labelpnt->ymax = labelpnt->y +
				gdFontH(gdFontMediumBold) / 2;
                        labelpnt->attribute = pointheader.attribute_level;
                        addlabeltolist(screenlabellist, labelpnt);
		        break;

		        case 3:
			if(pseudoscale < 400)
			    break;
		        gdImageCopy(image, med_dot,
				x - gdImageSX(med_dot) / 2,
				y - gdImageSY(med_dot) / 2,
				0, 0, gdImageSX(med_dot),
				gdImageSY(med_dot));
                        /* add this string to the list */
                        labelpnt = (SCREENLABEL *)malloc(sizeof(SCREENLABEL));
                        strcpy(labelpnt->label, label);
			labelpnt->font = gdFontMediumBold;
                        labelpnt->xmin = labelpnt->x = x +
				gdImageSX(med_dot) / 2;
                        labelpnt->xmax = labelpnt->xmin +
				gdFontW(gdFontMediumBold) *
				(pointheader.label_length - 1);
                        labelpnt->y = y - gdFontH(gdFontMediumBold) / 2;
                        labelpnt->ymin = labelpnt->y -
				gdFontH(gdFontMediumBold) / 2;
                        labelpnt->ymax = labelpnt->y +
				gdFontH(gdFontMediumBold) / 2;
                        labelpnt->attribute = pointheader.attribute_level;
                        addlabeltolist(screenlabellist, labelpnt);
			break;

		        case 5:
			if(pseudoscale < 2000)
			    break;
		        gdImageCopy(image, small_dot,
				x - gdImageSX(small_dot) / 2,
				y - gdImageSY(small_dot) / 2,
				0, 0, gdImageSX(small_dot),
				gdImageSY(small_dot));
                        /* add this string to the list */
                        labelpnt = (SCREENLABEL *)malloc(sizeof(SCREENLABEL));
                        strcpy(labelpnt->label, label);
			labelpnt->font = gdFontMediumBold;
                        labelpnt->xmin = labelpnt->x = x +
				gdImageSX(small_dot) / 2;
                        labelpnt->xmax = labelpnt->xmin +
				gdFontW(gdFontMediumBold) *
				(pointheader.label_length - 1);
                        labelpnt->y = y - gdFontH(gdFontMediumBold) / 2;
                        labelpnt->ymin = labelpnt->y -
				gdFontH(gdFontMediumBold) / 2;
                        labelpnt->ymax = labelpnt->y +
				gdFontH(gdFontMediumBold) / 2;
                        labelpnt->attribute = pointheader.attribute_level;
                        addlabeltolist(screenlabellist, labelpnt);
			break;
		    }
		    break;

		    case HIGH_RES:
		    if(pointheader.label_length <= 1)
		    {
			if(pseudoscale > 10000)
		            gdImageCopy(image, small_dot,
				x - gdImageSX(small_dot) / 2,
				y - gdImageSY(small_dot) / 2,
				0, 0, gdImageSX(small_dot),
				gdImageSY(small_dot));
		    }
		    else if(islower(label[1]))
		    {
		        gdImageCopy(image, med_dot,
				x - gdImageSX(med_dot) / 2,
				y - gdImageSY(med_dot) / 2,
				0, 0, gdImageSX(med_dot),
				gdImageSY(med_dot));
			if(pseudoscale > 5000)
			{
                          /* add this string to the list */
                          labelpnt = (SCREENLABEL *)malloc(sizeof(SCREENLABEL));
                          strcpy(labelpnt->label, label);
			  labelpnt->font = gdFontMediumBold;
                          labelpnt->xmin = labelpnt->x = x +
				  gdImageSX(med_dot) / 2;
                          labelpnt->xmax = labelpnt->xmin +
				  gdFontW(gdFontMediumBold) *
				  (pointheader.label_length - 1);
                          labelpnt->y = y - gdFontH(gdFontMediumBold) / 2;
                          labelpnt->ymin = labelpnt->y -
				  gdFontH(gdFontMediumBold) / 2;
                          labelpnt->ymax = labelpnt->y +
				  gdFontH(gdFontMediumBold) / 2;
                          labelpnt->attribute = pointheader.attribute_level;
                          addlabeltolist(screenlabellist, labelpnt);
			}
		    }
		    else
		    {
		        gdImageCopy(image, large_dot,
				x - gdImageSX(large_dot) / 2,
				y - gdImageSY(large_dot) / 2,
				0, 0, gdImageSX(large_dot),
				gdImageSY(large_dot));
                        /* add this string to the list */
                        labelpnt = (SCREENLABEL *)malloc(sizeof(SCREENLABEL));
                        strcpy(labelpnt->label, label);
			labelpnt->font = gdFontMediumBold;
                        labelpnt->xmin = labelpnt->x = x +
				gdImageSX(large_dot) / 2;
                        labelpnt->xmax = labelpnt->xmin +
				gdFontW(gdFontMediumBold) *
				(pointheader.label_length - 1);
                        labelpnt->y = y - gdFontH(gdFontMediumBold) / 2;
                        labelpnt->ymin = labelpnt->y -
				gdFontH(gdFontMediumBold) / 2;
                        labelpnt->ymax = labelpnt->y +
				gdFontH(gdFontMediumBold) / 2;
                        labelpnt->attribute = pointheader.attribute_level;
                        addlabeltolist(screenlabellist, labelpnt);
		    }
		    break;

		    case ULTRA_RES:
		    switch(pointheader.attribute_level)
		    {
		        case 1:
		        gdImageCopy(image, large_dot,
				x - gdImageSX(large_dot) / 2,
				y - gdImageSY(large_dot) / 2,
				0, 0, gdImageSX(large_dot),
				gdImageSY(large_dot));
                        /* add this string to the list */
                        labelpnt = (SCREENLABEL *)malloc(sizeof(SCREENLABEL));
                        strcpy(labelpnt->label, label);
			labelpnt->font = gdFontMediumBold;
                        labelpnt->xmin = labelpnt->x = x +
				gdImageSX(large_dot) / 2;
                        labelpnt->xmax = labelpnt->xmin +
				gdFontW(gdFontMediumBold) *
				(pointheader.label_length - 1);
                        labelpnt->y = y - gdFontH(gdFontMediumBold) / 2;
                        labelpnt->ymin = labelpnt->y -
				gdFontH(gdFontMediumBold) / 2;
                        labelpnt->ymax = labelpnt->y +
				gdFontH(gdFontMediumBold) / 2;
                        labelpnt->attribute = pointheader.attribute_level;
                        addlabeltolist(screenlabellist, labelpnt);
		        break;

		        case 2:
		        gdImageCopy(image, med_dot,
				x - gdImageSX(med_dot) / 2,
				y - gdImageSY(med_dot) / 2,
				0, 0, gdImageSX(med_dot),
				gdImageSY(med_dot));
                        /* add this string to the list */
                        labelpnt = (SCREENLABEL *)malloc(sizeof(SCREENLABEL));
                        strcpy(labelpnt->label, label);
			labelpnt->font = gdFontMediumBold;
                        labelpnt->xmin = labelpnt->x = x +
				gdImageSX(med_dot) / 2;
                        labelpnt->xmax = labelpnt->xmin +
				gdFontW(gdFontMediumBold) *
				(pointheader.label_length - 1);
                        labelpnt->y = y - gdFontH(gdFontMediumBold) / 2;
                        labelpnt->ymin = labelpnt->y -
				gdFontH(gdFontMediumBold) / 2;
                        labelpnt->ymax = labelpnt->y +
				gdFontH(gdFontMediumBold) / 2;
                        labelpnt->attribute = pointheader.attribute_level;
                        addlabeltolist(screenlabellist, labelpnt);
		        break;

		        case 3:
		        gdImageCopy(image, small_dot,
				x - gdImageSX(small_dot) / 2,
				y - gdImageSY(small_dot) / 2,
				0, 0, gdImageSX(small_dot),
				gdImageSY(small_dot));
			if(pseudoscale >= 20000)
			{
                          /* add this string to the list */
                          labelpnt = (SCREENLABEL *)malloc(sizeof(SCREENLABEL));
                          strcpy(labelpnt->label, label);
			  labelpnt->font = gdFontMediumBold;
                          labelpnt->xmin = labelpnt->x = x +
				gdImageSX(small_dot) / 2;
                          labelpnt->xmax = labelpnt->xmin +
				gdFontW(gdFontMediumBold) *
				(pointheader.label_length - 1);
                          labelpnt->y = y - gdFontH(gdFontMediumBold) / 2;
                          labelpnt->ymin = labelpnt->y -
				gdFontH(gdFontMediumBold) / 2;
                          labelpnt->ymax = labelpnt->y +
				gdFontH(gdFontMediumBold) / 2;
                          labelpnt->attribute = pointheader.attribute_level;
                          addlabeltolist(screenlabellist, labelpnt);
			}
		        break;

		        case 4:
		        case 5:
			if(pseudoscale >= 150000)
			{
		          gdImageCopy(image, small_dot,
				x - gdImageSX(small_dot) / 2,
				y - gdImageSY(small_dot) / 2,
				0, 0, gdImageSX(small_dot),
				gdImageSY(small_dot));
                          /* add this string to the list */
                          labelpnt = (SCREENLABEL *)malloc(sizeof(SCREENLABEL));
                          strcpy(labelpnt->label, label);
			  labelpnt->font = gdFontTiny;
                          labelpnt->xmin = labelpnt->x = x +
				gdImageSX(small_dot) / 2;
                          labelpnt->xmax = labelpnt->xmin +
				gdFontW(gdFontTiny) *
				(pointheader.label_length - 1);
                          labelpnt->y = y - gdFontH(gdFontTiny) / 2;
                          labelpnt->ymin = labelpnt->y -
				gdFontH(gdFontTiny) / 2;
                          labelpnt->ymax = labelpnt->y +
				gdFontH(gdFontTiny) / 2;
                          labelpnt->attribute = pointheader.attribute_level;
                          addlabeltolist(screenlabellist, labelpnt);
			}
		        break;

		        case 6:
			if(pseudoscale >= 30000)
			{
		          gdImageCopy(image, small_dot,
				x - gdImageSX(small_dot) / 2,
				y - gdImageSY(small_dot) / 2,
				0, 0, gdImageSX(small_dot),
				gdImageSY(small_dot));
                          /* add this string to the list */
                          labelpnt = (SCREENLABEL *)malloc(sizeof(SCREENLABEL));
                          strcpy(labelpnt->label, label);
			  labelpnt->font = gdFontTiny;
                          labelpnt->xmin = labelpnt->x = x +
				gdImageSX(small_dot) / 2;
                          labelpnt->xmax = labelpnt->xmin +
				gdFontW(gdFontTiny) *
				(pointheader.label_length - 1);
                          labelpnt->y = y - gdFontH(gdFontTiny) / 2;
                          labelpnt->ymin = labelpnt->y -
				gdFontH(gdFontTiny) / 2;
                          labelpnt->ymax = labelpnt->y +
				gdFontH(gdFontTiny) / 2;
                          labelpnt->attribute = pointheader.attribute_level;
                          addlabeltolist(screenlabellist, labelpnt);
			}
		        break;
		    }
		    break;
		}
	    }
	}
	break;

	case NHAPBWPOINT:
	case NHAPCIRPOINT:
	case NAPPPOINT:
        while(fread((char *)&pointheader, sizeof(POINTHEADER), 1, fp))
        {
	    if(!fread((char *)&label, (int)pointheader.label_length, 1, fp))
		break;

	    pointheader.lat = (long)ntohl((unsigned long)pointheader.lat);
	    pointheader.lon = (long)ntohl((unsigned long)pointheader.lon);
            if(projection) { /* projection is gnomonic */
                status = gnomonic(radius, center_lat, center_lon,
                                  (float)pointheader.lat * 0.00001 / RAD,
                                  (float)pointheader.lon * 0.00001 / RAD, 
                                  &x, &y);
            } else { /* projection is orthographic */
                status = ortho(radius, center_lat, center_lon,
                               (float)pointheader.lat * 0.00001 / RAD,
                               (float)pointheader.lon * 0.00001 / RAD, 
                               &x, &y);
            }

	    /* check if point is visible */
            if(status != -1 && abs(x) <= width / 2 && abs(y) <= height / 2)
            {
		x += center_x;
                y = center_y - y;
		if(resolution == ULTRA_RES)
		{
		    gdImageCopy(image, small_black_dot,
				x - gdImageSX(small_black_dot) / 2,
				y - gdImageSY(small_black_dot) / 2,
				0, 0, gdImageSX(small_black_dot),
				gdImageSY(small_black_dot));
                    /* add this string to the list */
                    if( label[0] != '\0' )
                    {
                       labelpnt = (SCREENLABEL *)malloc(sizeof(SCREENLABEL));
                       strcpy(labelpnt->label, label);
		       labelpnt->font = gdFontMediumBold;
                       labelpnt->xmin = labelpnt->x = x -
				gdFontW(gdFontMediumBold) *
				(pointheader.label_length - 1) / 2;
                       labelpnt->xmax = labelpnt->xmin +
				gdFontW(gdFontMediumBold) *
				(pointheader.label_length - 1);
                       labelpnt->y = labelpnt->ymin = y -
				gdImageSY(small_black_dot) / 2 - 
				gdFontH(gdFontMediumBold) + 2;
                       labelpnt->ymax = labelpnt->ymin + gdFontH(gdFontMediumBold);
                       labelpnt->attribute = pointheader.attribute_level;
                       addlabeltolist(screenlabellist, labelpnt);
                    }
		}
	    }
	}

	case WRS1POINT:
	case WRS2POINT:
        while(fread((char *)&pointheader, sizeof(POINTHEADER), 1, fp))
        {
	    if(!fread((char *)&label, (int)pointheader.label_length, 1, fp))
		break;

	    pointheader.lat = (long)ntohl((unsigned long)pointheader.lat);
	    pointheader.lon = (long)ntohl((unsigned long)pointheader.lon);
            if(projection) { /* projection is gnomonic */
                status = gnomonic(radius, center_lat, center_lon,
                                  (float)pointheader.lat * 0.00001 / RAD,
                                  (float)pointheader.lon * 0.00001 / RAD, 
                                  &x, &y);
            } else { /* projection is orthographic */
                status = ortho(radius, center_lat, center_lon,
                               (float)pointheader.lat * 0.00001 / RAD,
                               (float)pointheader.lon * 0.00001 / RAD, 
                               &x, &y);
            }

	    /* check if point is visible */
            if(status != -1 && abs(x) <= width / 2 && abs(y) <= height / 2)
            {
		x += center_x;
                y = center_y - y;
		if(resolution == MED_RES)
		{
		    gdImageCopy(image, small_black_dot,
				x - gdImageSX(small_black_dot) / 2,
				y - gdImageSY(small_black_dot) / 2,
				0, 0, gdImageSX(small_black_dot),
				gdImageSY(small_black_dot));
                    /* add this string to the list */
                    labelpnt = (SCREENLABEL *)malloc(sizeof(SCREENLABEL));
                    strcpy(labelpnt->label, label);
		    labelpnt->font = gdFontMediumBold;
                    labelpnt->xmin = labelpnt->x = x -
				gdFontW(gdFontMediumBold) *
				(pointheader.label_length - 1) / 2;
                    labelpnt->xmax = labelpnt->xmin +
				gdFontW(gdFontMediumBold) *
				(pointheader.label_length - 1);
                    labelpnt->y = labelpnt->ymin = y -
				gdImageSY(small_black_dot) / 2 - 
				gdFontH(gdFontMediumBold) + 2;
                    labelpnt->ymax = labelpnt->ymin + gdFontH(gdFontMediumBold);
                    labelpnt->attribute = pointheader.attribute_level;
                    addlabeltolist(screenlabellist, labelpnt);
		}
	    }
	}
	break;
	
        case DNLINE:
        case POLINE:
        case RDLINE:
        case RRLINE:
        if(resolution == LOW_RES && zoom_percent < 200 && layer == POLINE)
            skip_states = 1;
        if(resolution == ULTRA_RES && pseudoscale < 80000 && layer == RDLINE)
            skip_streets = 1;
        if(resolution == ULTRA_RES && pseudoscale < 80000 && layer == DNLINE)
            skip_streams = 1;
        while(fread((char *)&lineheader, sizeof(LINEHEADER), 1, fp))
        {
	    lineheader.node_count = ntohs(lineheader.node_count);
            if(lineheader.short_long_flag == 's')
            {
                fread((char *)&long_node, sizeof(LONG_NODE), 1, fp);
	        long_node.lat = (long)ntohl((unsigned long)long_node.lat);
	        long_node.lon = (long)ntohl((unsigned long)long_node.lon);
                route[0].latitude = (float)long_node.lat * 0.00001;
                route[0].longitude = (float)long_node.lon * 0.00001;
                for(npts = 1; npts < lineheader.node_count; npts++)
                {
                    fread((char *)&short_node, sizeof(SHORT_NODE), 1, fp);
	            short_node.delta_lat =
			(short)ntohs((unsigned short)short_node.delta_lat);
	            short_node.delta_lon =
			(short)ntohs((unsigned short)short_node.delta_lon);
		    long_node.lat += short_node.delta_lat;
		    long_node.lon += short_node.delta_lon;
                    route[npts].latitude = (float)long_node.lat * 0.00001;
                    route[npts].longitude = (float)long_node.lon * 0.00001;
                }
            }
            else
            {
                for(npts = 0; npts < lineheader.node_count; npts++)
                {
                    fread((char *)&long_node, sizeof(LONG_NODE), 1, fp);
	            long_node.lat = (long)ntohl((unsigned long)long_node.lat);
	            long_node.lon = (long)ntohl((unsigned long)long_node.lon);
                    route[npts].latitude = (float)long_node.lat * 0.00001;
                    route[npts].longitude = (float)long_node.lon * 0.00001;
                }
            }
            if(skip_streets && lineheader.attribute_level > 5)
              continue;
            if(skip_streams && lineheader.attribute_level != 1)
              continue;
	    if(skip_states)
	    {
	      if(lineheader.attribute_level != 1)
	        continue;
	      else
		lineheader.attribute_level++;	/* reduce to one pixel line */
	    }
            plot_route(&map_gc[lineheader.attribute_level],
			route, npts, POLYLINE);
        }
	break;

        case PPPOLY:
        if(resolution == ULTRA_RES)
            break;
        case OCPOLY:
        case DNPOLY:
        if(resolution == ULTRA_RES && pseudoscale < 100000 && layer == DNPOLY)
            skip_ponds = 1;
        while(fread((char *)&lineheader, sizeof(LINEHEADER), 1, fp))
        {
	    lineheader.node_count = ntohs(lineheader.node_count);
            fread((char *)&long_node, sizeof(LONG_NODE), 1, fp);
	    long_node.lat = (long)ntohl((unsigned long)long_node.lat);
	    long_node.lon = (long)ntohl((unsigned long)long_node.lon);
            route[0].latitude = (float)long_node.lat * 0.00001;
            route[0].longitude = (float)long_node.lon * 0.00001;
            for(npts = 1; npts < lineheader.node_count; npts++)
            {
		/* get the next node */
                if(lineheader.short_long_flag == 's')
		{
                    fread((char *)&short_node, sizeof(SHORT_NODE), 1, fp);
	            short_node.delta_lat =
			(short)ntohs((unsigned short)short_node.delta_lat);
	            short_node.delta_lon =
			(short)ntohs((unsigned short)short_node.delta_lon);
		    long_node.lat += short_node.delta_lat;
		    long_node.lon += short_node.delta_lon;
		}
		else
		{
                    fread((char *)&long_node, sizeof(LONG_NODE), 1, fp);
	            long_node.lat = (long)ntohl((unsigned long)long_node.lat);
	            long_node.lon = (long)ntohl((unsigned long)long_node.lon);
		}

                route[npts].latitude = (float)long_node.lat * 0.00001;
                route[npts].longitude = (float)long_node.lon * 0.00001;
            }
            if(skip_ponds && lineheader.attribute_level > 4)
                continue;
	    else if(lineheader.attribute_level != -1)
                plot_route(&map_gc[lineheader.attribute_level],
			route, npts, POLYGON);
	    else /* island */
                plot_route(&land_gc, route, npts, POLYGON);
        }
	break;
      }
    }
    fclose(fp);
    return(0);
}

static void
plot_route(GC *gc, POINT *route, int npts, int type)
{
    register int i;
    int x, y, last_x, last_y, status, last_status, rtn;
    int minx, miny, maxx, maxy;
    int inside = 0;
    int crossings = 0;
    int center_x, center_y;
    double center_lat = center.latitude / RAD;
    double center_lon = center.longitude / RAD;
    POLYLIST *polylisthead, *polypnt, *polyscan;
    POINT crossing;
    double angle1, angle2, angle, step;

    center_x = width / 2;
    center_y = height / 2;

    polypnt = polylisthead = (POLYLIST *)malloc(sizeof(POLYLIST));
    polylisthead->gdpoints = (gdPoint *)malloc((npts + 360) * sizeof(gdPoint));
    if(!polylisthead || !polylisthead->gdpoints)
    {
	fprintf(stderr, "malloc failed\n");
	return;
    }
    polylisthead->npts = 0;
    polylisthead->next = 0;
    polylisthead->type = 0;

    if(projection) { /* projection is gnomonic */
        last_status = status = gnomonic(radius, center_lat, center_lon,
		                        route[0].latitude / RAD, 
                                        route[0].longitude / RAD, &x, &y);
    } else { /* projection is orthographic */
        last_status = status = ortho(radius, center_lat, center_lon,
		                     route[0].latitude / RAD, 
                                     route[0].longitude / RAD, &x, &y);
    }
    if(status != -1)	/* visible */
    {
	inside = 1;
        polypnt->gdpoints[polypnt->npts].x = center_x + x;
        polypnt->gdpoints[polypnt->npts].y = center_y - y;
        polypnt->npts++;
        last_x = x;
        last_y = y;
    }

    for(i = 1; i < npts; i++)
    {
        if(projection) { /* projection is gnomonic */
            status = gnomonic(radius, center_lat, center_lon,
		              route[i].latitude / RAD, 
                              route[i].longitude / RAD, &x, &y);
	} else { /* projection is orthographic */
            status = ortho(radius, center_lat, center_lon,
		           route[i].latitude / RAD, 
                           route[i].longitude / RAD, &x, &y);
        }
	if(status != last_status)
	{
	    crossings++;
	    if(status == -1)
	        rtn = find_crossing(&route[i - 1], &route[i], &crossing);
	    else
	        rtn = find_crossing(&route[i], &route[i - 1], &crossing);
	    if(rtn)
	    {
		fprintf(stderr, "Convergence failure\n");
		return;
	    }
            if(projection) { /* projection is gnomonic */
                gnomonic(radius, center_lat, center_lon, 
                         route[i].latitude / RAD,
                         route[i].longitude / RAD, &x, &y);
            } else { /* projection is orthographic */
                ortho(radius, center_lat, center_lon, 
                      route[i].latitude / RAD,
                      route[i].longitude / RAD, &x, &y);
            }
            polypnt->gdpoints[polypnt->npts].x = center_x + x;
            polypnt->gdpoints[polypnt->npts].y = center_y - y;
            polypnt->npts++;
	    last_x = x;
	    last_y = y;

	    if(status == -1)
	    {
		polypnt->next = (POLYLIST *)malloc(sizeof(POLYLIST));
		polypnt = polypnt->next;
		polypnt->gdpoints = (gdPoint *)malloc((npts + 360) * sizeof(gdPoint));
		if(!polypnt || !polypnt->gdpoints)
		{
		    fprintf(stderr, "malloc failed\n");
		    return;
		}
		polypnt->npts = 0;
		polypnt->next = 0;
		polypnt->type = 0;
	    }
	}
	if(!status)
	{
	    if((x != last_x) || (y != last_y))
	    {
                polypnt->gdpoints[polypnt->npts].x = center_x + x;
                polypnt->gdpoints[polypnt->npts].y = center_y - y;
                polypnt->npts++;
		last_x = x;
		last_y = y;
	    }
	}
	last_status = status;
    }

    /* if we started inside, merge the first and last polygons */
    if((type == POLYGON) && inside && (crossings > 0))
    {
	for(i = 0; i < polylisthead->npts; i++)
	{
	    polypnt->gdpoints[polypnt->npts].x = polylisthead->gdpoints[i].x;
	    polypnt->gdpoints[polypnt->npts].y = polylisthead->gdpoints[i].y;
            polypnt->npts++;
	}
	polyscan = polylisthead->next;
	free(polylisthead->gdpoints);
	free(polylisthead);
	polylisthead = polyscan;
    }

    /* connect the endpoints along the horizon */
    if((type == POLYGON) && (crossings > 0))
    {
        for(polypnt = polylisthead; polypnt; polypnt = polypnt->next)
        {
	    /* skip degenerate polygons */
	    if(polypnt->npts < 2)
		continue;

	    angle1 = atan2(
		(double)polypnt->gdpoints[polypnt->npts - 1].y - center_y,
		(double)polypnt->gdpoints[polypnt->npts - 1].x - center_x);
	    angle2 = atan2((double)polypnt->gdpoints[0].y - center_y,
				(double)polypnt->gdpoints[0].x - center_x);

	    if(fabs(angle2 - angle1) > M_PI_2)
	    {
		if(angle1 < angle2)
		    angle1 += M_2PI;
		else
		    angle2 += M_2PI;
	    }

	    if(angle1 < angle2)
	        step = 3 / RAD;
	    else
	        step = -3 / RAD;
	    angle = (int)(angle1 / step) * step + step;
	    for(i = (angle2 - angle1) / step; i > 0; i--)
	    {
		polypnt->gdpoints[polypnt->npts].x = cos(angle) * radius + center_x;
		polypnt->gdpoints[polypnt->npts].y = sin(angle) * radius + center_y;
		polypnt->npts++;
		angle += step;
	    }
        }
    }

    /* plot 'em and free them */
    for(polypnt = polylisthead; polypnt;)
    {
	if(polypnt->npts > 0)
	{
            /* check if this polygon is viewable at all */
            minx = maxx = polypnt->gdpoints[0].x;
            miny = maxy = polypnt->gdpoints[0].y;
            for(i = 1; i < polypnt->npts; i++)
            {
	        if(polypnt->gdpoints[i].x < minx)
	            minx = polypnt->gdpoints[i].x;
	        else if(polypnt->gdpoints[i].x > maxx)
	            maxx = polypnt->gdpoints[i].x;
    
	        if(polypnt->gdpoints[i].y < miny)
	            miny = polypnt->gdpoints[i].y;
	        else if(polypnt->gdpoints[i].y > maxy)
	            maxy = polypnt->gdpoints[i].y;
            }
    
            if((maxx > 0) && (minx < width) && (maxy > 0) && (miny < height))
            {
	        if(polypnt->npts == 1)
	            gdImageSetPixel(image, polypnt->gdpoints[0].x,
				    polypnt->gdpoints[0].y, gc->color);
	        else if(type == POLYLINE)
	        {
	            if((gc->line_style == BROKEN) && (gc->line_width > 1))
	            {
		        if(gc->dash_style == DASHED)
		        {
		            int dashed[8];
    
		            dashed[0]=dashed[1]=dashed[2]=dashed[3]=1;
		            dashed[4]=dashed[5]=dashed[6]=dashed[7]=0;
		            gdImageSetStyle(image, dashed, 8);
		        }
		        else if(gc->dash_style == DOTTED)
		        {
		            int dotted[4];
    
		            dotted[0] = 1;
		            dotted[1] = dotted[2] = dotted[3] = 0;
		            gdImageSetStyle(image, dotted, 4);
		        }
		        gdImageColorDeallocate(box2, 1);
		        gdImageColorAllocate(box2,
				    gdImageRed(image, gc->color),
				    gdImageGreen(image, gc->color),
				    gdImageBlue(image, gc->color));
		        gdImageSetBrush(image, box2);
		        gdImagePolyline(image, polypnt->gdpoints, polypnt->npts,
					    gdStyledBrushed);
	            }
	            else if(gc->line_style == BROKEN)
	            {
		        if(gc->dash_style == DASHED)
		        {
		            int dashed[8];
    
		            dashed[0]=dashed[1]=dashed[2]=dashed[3]=gc->color;
		            dashed[4]=dashed[5]=dashed[6]=dashed[7]=gdTransparent;
		            gdImageSetStyle(image, dashed, 8);
		        }
		        else if(gc->dash_style == DOTTED)
		        {
		            int dotted[4];
    
		            dotted[0] = gc->color;
		            dotted[1] = dotted[2] = dotted[3] = gdTransparent;
		            gdImageSetStyle(image, dotted, 4);
		        }
		        gdImagePolyline(image, polypnt->gdpoints, polypnt->npts,
					    gdStyled);
	            }
	            else if(gc->line_width > 1)
	            {
		        gdImageColorDeallocate(box2, 1);
		        gdImageColorAllocate(box2,
				    gdImageRed(image, gc->color),
				    gdImageGreen(image, gc->color),
				    gdImageBlue(image, gc->color));
		        gdImageSetBrush(image, box2);
		        gdImagePolyline(image, polypnt->gdpoints, polypnt->npts,
					    gdBrushed);
	            }
	            else
		        gdImagePolyline(image, polypnt->gdpoints, polypnt->npts,
					    gc->color);
	        }
                else
	            miFilledPolygon(image, (DDXPointPtr)polypnt->gdpoints,
				    polypnt->npts, gc->color);
            }
	}
	polyscan = polypnt->next;
        free(polypnt->gdpoints);
        free(polypnt);
	polypnt = polyscan;
    }
}

static int
find_crossing(POINT *from, POINT *to, POINT *rtn)
/* *from - point in projection */
/* *to - point over the horizon */
{
    register int i;
    float err, x, y, yinc, xinc, sin_ctr_lat, cos_ctr_lat, ctr_lon;

    sin_ctr_lat = sin(center.latitude / RAD);
    cos_ctr_lat = cos(center.latitude / RAD);
    ctr_lon = center.longitude / RAD;
    y = (float)from->latitude / RAD;
    x = (float)from->longitude / RAD;
    xinc = (float)(to->longitude - from->longitude) / 2 / RAD;
    yinc = (float)(to->latitude - from->latitude) / 2 / RAD;

    /* check if from point is close enough */
    err = sin_ctr_lat * sin(y) + cos_ctr_lat * cos(x - ctr_lon) * cos(y);
    if(fabs(err) < 0.0001)
    {
        rtn->latitude = from->latitude;
        rtn->longitude = from->longitude;
	return(0);
    }

    for(i = 0; ((err < 0) || (err > 0.0001)) && (i < 50); i++)
    {
        err = sin_ctr_lat * sin(y + yinc) + cos_ctr_lat *
				cos(x + xinc - ctr_lon) * cos(y + yinc);
	if(err > 0)
	{
	    x += xinc;
	    y += yinc;
	}
	xinc /= 2;
	yinc /= 2;
    }

    if(i < 50)
    {
        rtn->latitude = y * RAD;
        rtn->longitude = x * RAD;
	return(0);
    }
    else
	return(-1);
}

int
allocate_color(gdImagePtr image, char *colorname)
{
    register int i;
    int rtn;

    for(i = 0; i < NCOLORS; i++)
	if(!strcmp(colorname, colors[i].colorname))
	    break;

    if(i < NCOLORS)
    {
	rtn = gdImageColorExact(image,
			colors[i].red, colors[i].green, colors[i].blue);
	if(rtn == -1)
	    rtn = gdImageColorAllocate(image,
			colors[i].red, colors[i].green, colors[i].blue);
    }
    else
	rtn = 1;	/* black */

    return(rtn);
}

static void
addlabeltolist(SCREENLABEL **screenlabellist, SCREENLABEL *labelpnt)
{
    SCREENLABEL *listscan = NULL, *lastscan = NULL;

    if(!*screenlabellist)
    {
        /* this is the first one */
        *screenlabellist = labelpnt;
        labelpnt->next = NULL;
    }
    else if(labelpnt->attribute < (*screenlabellist)->attribute)
    {
        /* found a new first one */
        labelpnt->next = *screenlabellist;
        *screenlabellist = labelpnt;
    }
    else
    {
        /* insert after the last one with the same or lower attribute level */
        lastscan = *screenlabellist;
        for(listscan = lastscan->next; listscan &&
                        labelpnt->attribute >= listscan->attribute;
                        listscan = listscan->next)
            lastscan = listscan;
        lastscan->next = labelpnt;
        labelpnt->next = listscan;
    }
}
