/****************************************************************************
 * Name:   hdf5process.c                                                    *
 *                                                                          *
 * Description:                                                             *
 *    Given an HDF file, check if the file is in HDF5 format.  If it is,    *
 *    open the file, read the needed information to generate a browse       *
 *    image, get the palette information for the image colormap, then       *
 *    use the HDF5 libraries to generate the browse image.  Notify the      *
 *    server using the program interface standars from the main program.    *
 *                                                                          *
 *  Program Interface Standards:                                            *
 *    The subroutines in this module use the same program interface         *
 *    standars used by the main program.  The outputs of this routines      *
 *    are in a VERY specific format.  Besides writing GIF files, this       *
 *    program communicates specific messages to the Web Gateway by          *
 *    printing them to standard output.  So,                                *
 *          DO NOT ADD PRINT STATEMENTS UNLESS YOU KNOW THE WEB             *
 *          GATEWAY KNOWS HOW TO READ WHAT YOU ARE PRINTING!                *
 *    This code only prints two things to standard output:                  *
 *    1: For each GIF file produced it prints:                              *
 *          "Image-pathname: filename.gif"                                  *
 *    without the quotes, with the actual filename, and followed by         *
 *    a blank line.                                                         *
 *    2: If there were any errors in processing, the code prints an         *
 *    informative error message preceeded by "X-Error: ".                   *
 *                                                                          *
 * Notes:                                                                   *
 *    When this program was originally written, the HDF5 utilities were     *
 *    in beta mode.  This was the first attempt to combine both the HDF4    *
 *    and the HDF5 utilities.  Since there is no standars available for     *
 *    HDF5 files and utilities usage, this program uses very basic code.    *
 *    As the HDF5 utilites evolve, this program will need to change to      *
 *    accomodate the HDF5 libraries changes.                                *
 *         BEFORE MODIFYING THIS CODE, YOU WILL PROBABLY NEED TO            *
 *         HAVE A COPY OF THE HDF5 REFERENCE MANUAL FROM NCSA               *
 *         (http://hdf.ncsa.uiuc.edu/HDF5).  I (Lorena) used                *
 *         Version HDF5-1.4 (1.4.0 and 1.4.1) from April 2001.              *
 *                                                                          *
 * Revision History:                                                        *
 *    May 2001, Lorena Marsans, original prototype with help from Jim       *
 *       Amrhein.                                                           *
 *                                                                          *
 ****************************************************************************/

#include <stdio.h>
#include <hdf5.h>
#include <malloc.h>
#include <unistd.h>
#include "odldef.h"
#include "odlext.h"

#ifdef HAVE_STRINGS_H
#include <strings.h>
#else
#include <string.h>
#endif /*HAVE_STRINGS_H*/

/* Global Variable used to store the output filename base. */
static char *BASE;

/* typedef's from NCSA: */
typedef unsigned char UCHAR;

/* External functions: */
/* WriteGIF writes the image to the specified file as a GIF. */
extern int WriteGIF (char *outFilename, UCHAR *image, 
                     int xdim, int ydim, char *colormap);

/* WritePalette creates and writes the color palette used for the image. */
extern void WritePalette (char *palFilename, char *colormap, 
                                             int image, int inputFD);

/* Local functions: */
/* dump_all gets the information out of the hdf5 file, it also controls *
 * what to do based on the type of information.                         */
static herr_t dump_all(hid_t group, const char *name, void *op_data);

/* dump_dataset reads the image dataset information needed to generate  *
 * an hdf5 browse image.                                                */
static void dump_dataset (hid_t did, const char *name, int *pal);


/* --------------------------------------------------------------------- 
 * Function: Hdf5Process
 * Description:
 *     Main routine that controls processing for HDF5 files.
 *     Check that the file is in HDF5 format, open the file,
 *     open the image group, then process the dataset image
 *     information to generate the browse image.
 * Parameters: 
 *     char *inputFilename - name of input HDF5 file.
 *     char *base          - output filename base.
 * Return: 
 *        int 0  - success.
 *     or int 1  - error.
 *
 * --------------------------------------------------------------------- */
int Hdf5Process (char *inputFilename, char *base) {

   int   i;
   char  *gname = "images";
   hid_t gid;
   hid_t file=-1, plist=-1;
   H5G_stat_t  sb;

   BASE = base;

   /* Use HDF function to check if the file is an HDF5 file. */
   if (H5Fis_hdf5 (inputFilename)) {
      plist = H5Pcreate(H5P_FILE_ACCESS);

      /* Try to open the file */
      file = H5Fopen(inputFilename, H5F_ACC_RDONLY, plist);
      if (file < 0) {
         printf ("X-Error: Problem opening %s \n",inputFilename);
         return (1);
      }
      H5Pclose(plist);

      if (H5Gget_objinfo(file, "/", TRUE, &sb)>=0 ) {
         if ((gid = H5Gopen (file, gname)) >=0 ) {
            H5Giterate (gid, ".", NULL, dump_all, NULL);
         }
         H5Gclose(gid);
         H5Fclose(file);
      }
      return (0);
   }
   else {  printf ("X-Error: It's NOT an HDF file!\n",inputFilename);
      return (1);
   }
}

/* ---------------------------------------------------------------------
 * Function: dump_all
 * Description:
 *     Get information for the given group, take proper action based
 *     on the type of object obtained from the group.
 *     The calling of this function is done with the following command:
 *        H5Giterate (gid, ".", NULL, dump_all, NULL)
 *     The HDF5 functions take care of passing the correct information.
 * Parameters:
 *     hid_t     group - identifier of group to access dataset within
 *     char      *name - name of dataset to access
 *     void      *op_data - used by the HDF5 utilities.
 * Return:
 *        herr_t 0  - success.
 *     or herr_t 1  - error.
 *
 * --------------------------------------------------------------------- */
static herr_t dump_all (hid_t group, const char *name, 
                                     void  *op_data) {
   H5G_stat_t  statbuf;
   char        *buf, *tmp;
   char        name_palette[128];
   hid_t       obj;
   hid_t       obj_palette;
   hid_t       filespace;                   
   hid_t       memspace;                  
   hsize_t     dims[2];           /* dataset and chunk dimensions*/ 
   hsize_t     i, j;
   herr_t      status, status_n;                        
   int         data_out[256][3];  /* buffer for dataset to be read */
   int         rank;

   H5Gget_objinfo(group, name, FALSE, &statbuf);
   switch (statbuf.type) {
      case H5G_LINK:
         break;

      case H5G_DATASET:
         if ((obj=H5Dopen (group, name))>=0) {
            H5Gget_objinfo(obj, ".", TRUE, &statbuf);
         }
         if (!strcmp (name, name_palette)) { 
            break; 
         }
         sprintf (name_palette, "%s_palette",name);
         if ((obj_palette=H5Dopen (group, name_palette))>=0) {
            H5Gget_objinfo(obj_palette, ".", TRUE, &statbuf); 

            /* Get filespace handle first. */
            filespace = H5Dget_space(obj_palette);    
            rank      = H5Sget_simple_extent_ndims(filespace);
            status_n  = H5Sget_simple_extent_dims(filespace, dims, NULL);
            memspace  = H5Screate_simple(rank, dims, NULL);
            status    = H5Dread(obj_palette, H5T_NATIVE_INT, memspace, 
                                        filespace, H5P_DEFAULT, data_out);
            dump_dataset (obj, name, *data_out);
         }
         break;

      case H5G_TYPE:
         break;

      case H5G_GROUP:
         break;

      default:
         printf ("X-Error: h5dump error: unknown object %s\n", name);
         return 1;
   }
   return 0;
}

/* ---------------------------------------------------------------------
 * Function: dump_dataset
 * Description:
 *     Generate the image and the palette files.
 * Parameters:
 *     hid_t     did   - dataset identifier
 *     char      *name - dataset name
 *     int       *pal  - palette data
 *
 * --------------------------------------------------------------------- */
static void dump_dataset (hid_t did, const char *name, int *pal) {

   hsize_t  dims_out[2];       /* dataset dimensions */      
   hid_t    type;
   hid_t    space;
   hid_t    f_space;           /* file data space */
   hid_t    memspace;                 
   int      ndims;
   int      rank;
   UCHAR    *buffer;
   int      n;                 /* Local counter */
   UCHAR    colormap[768];     /* Output colormap */
   char     outFilename[1024]; /* Output file name. */
   char     palFilename[1024]; /* Palette file name. */

   sprintf (outFilename, "%s%s.gif", BASE, name);
   sprintf (palFilename, "%s%s.pal.gif", BASE, name);

   type    = H5Dget_type (did);
   space   = H5Dget_space (did);
   f_space = H5Dget_space (did);
   rank    = H5Sget_simple_extent_ndims (f_space);
   ndims   = H5Sget_simple_extent_dims (space, dims_out, NULL);

   /* Define the memory space to read dataset.  */
   memspace = H5Screate_simple (rank, dims_out, NULL);
 
   /* Allocate space for image */
   if ((buffer = (UCHAR *) malloc (dims_out[0] * dims_out[1])) == NULL) {
      printf ("X-Error: dump_dataset: Unable to allocate memory for buffer.\n");
      exit(1);
   }

   if (H5Dread (did,  type, memspace, space, H5P_DEFAULT, buffer) < 0) {
      printf ("X-Error: Problem reading data. \n");
   }

   /* Dummy up a grayscale palette if there wasn't one. */
   if (!pal) {
      for (n = 0; n < 256; ++n) {
         pal[3*n] = pal[3*n+1] = pal[3*n+2] = n;
      }
   }

   /* Convert the palette into a colormap. */
   for (n = 0; n < 256; ++n) {
      colormap[n] = pal[3*n];
      colormap[n+256] = pal[3*n+1];
      colormap[n+512] = pal[3*n+2];
   }

   /* Create and write out the palette file */
   WritePalette (palFilename, (char *) colormap, 0, 0);

   /* Write out the image. */
   if (WriteGIF (outFilename, (UCHAR *) buffer, 
           (int) (dims_out[0]), (int) (dims_out[1]), (char *) colormap)) {
      printf ("X-Error: Process8BitImages: Converter program error,\
Don't know why!\n");
      exit (1);
   }

   /* Report the filename */
   printf ("Image-pathname: %s\n\n", outFilename);
}
