/*****************************************************************************/
/*  Name:	odlext.c - ODL function extensions			     */
/*									     */
/*  Description:							     */
/*	This file contains extension routines to the ODL library.	     */
/*									     */
/*  Bugs:								     */
/*	None.								     */
/*									     */
/*  Revision history:							     */
/*	94-03-09 BEB Written						     */
/*****************************************************************************/

#include <Gaea_config.h>
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "odldef.h"
#include "odlext.h"

/*---------------------------------------------------------------------------*/
/*  Name:	ReadString - Parse a string into an ODL tree		     */
/*									     */
/*  Usage:	int result = ReadString (char *str, AGGREGATE root);	     */
/*		str - String to parse					     */
/*		root - Root node to put tree into			     */
/*		result - Zero if unsuccessful, one if successful	     */
/*									     */
/*  Description:							     */
/*	This function is analogous to the ODL ReadLabel function, except     */
/*	that it parses the contents of the "str" into the ODL tree.  One     */
/*	is returned on success; zero is returned on failure.		     */
/*									     */
/*	The string must consist of ODL-formatted lines delimited by new      */
/*	line characters and terminated by a single NUL character.	     */
/*									     */
/*  Bugs:								     */
/*	The implementation is kind of a kludge in that it creates a temp     */
/*	file to write the string, uses ReadLabel to read it in, and then     */
/*	deletes the file.						     */
/*									     */
/*  Revision history:							     */
/*	94-03-09 BEB Written						     */
/*---------------------------------------------------------------------------*/
 
int
ReadString (char *str, AGGREGATE root) {

    FILE	       *ifd;		/* Temporary file */
    char	       *nam;		/* Temporary file name */
    int			len;		/* Length of string */
    int			ofd;		/* Temporary file */
    int			result;		/* Result of ODL call */

/*  Compute the length of the string, but don't do empty strings  */

    if (str == NULL || root == NULL || (len = strlen (str)) <= 0)
	return (0);

/*  Make a temporary file name  */

    if ((nam = tmpnam (NULL)) == NULL)
	return (0);

/*  Open the file  */

    if ((ofd = open (nam, O_WRONLY | O_CREAT | O_EXCL, 0400)) <= 0) {
	free (nam);
	return (0);
    }

/*  Write the data  */

    if (write (ofd, str, len) != len) {
	close (ofd);
	unlink (nam);
	return (0);
    }

/*  Close, then reopen the file  */

    close (ofd);

    if ((ifd = fopen (nam, "r")) == NULL) {
	unlink (nam);
	return (0);
    }

/*  Attempt to parse the file  */

    result = ReadLabel (ifd, root);

/*  Clean up before returning  */

    fclose (ifd);
    unlink (nam);
    return (result);
}

/*---------------------------------------------------------------------------*/
/*  Name:	WriteString - Write an ODL tree to a string		     */
/*									     */
/*  Usage:	char *result = WriteString (AGGREGATE root);		     */
/*		root - Root of ODL tree					     */
/*		result - Returned string or NULL on errors		     */
/*									     */
/*  Description:							     */
/*	Convert the ODL tree into text and store it in a string.	     */
/*									     */
/*  Bugs:								     */
/*	Returns NULL if memory cannot be allocated.			     */
/*									     */
/*  Revision history:							     */
/*	94-05-25 BEB Written						     */
/*---------------------------------------------------------------------------*/

char *
WriteString (AGGREGATE root) {

    FILE	       *ofd;		/* Temporary file */
    int			ifd;		/* Temporary file */
    char	       *nam;		/* Temporary file name */
    char	       *result;		/* Result of operation */
    struct stat		stats;		/* File stats */

/*  Make a temporary file name  */

    if ((nam = tmpnam (NULL)) == NULL)
	return (NULL);

/*  Open the file  */

    if ((ofd = fopen (nam, "w")) == NULL) {
	free (nam);
	return (NULL);
    }

/*  Write the ODL to the file  */

    WriteLabel (ofd, root);

/*  See how big the file is  */

    fclose (ofd);

    if (stat (nam, &stats)) {
	free (nam);
	return (NULL);
    }

/*  Allocate space for the result  */

    if ((result = (char *) malloc (stats.st_size + 1)) == NULL) {
	free (nam);
	return (NULL);
    }

/*  Open and read in the data */

    if ((ifd = open (nam, O_RDONLY, 0)) <= 0) {
	free (result);
	free (nam);
	return (NULL);
    }

    if (read (ifd, result, stats.st_size) != stats.st_size) {
	free (result);
	free (nam);
	return (NULL);
    }

/*  Terminate the string  */

    result[stats.st_size] = '\0';

/*  Close files and go home  */

    close (ifd);
    unlink (nam);
    free (nam);
    return (result);
}
