/* $Id: sendbrowse.c,v 1.22 2002/05/01 13:06:09 lorena Exp $ */
/**********************************************************************
 *  Name:   sendbrowse.c
 *          - (Because it is used to send a browse image to the user.)
 *          - (Compiles to "hdf2gif".)
 *
 *  Usage:	hdf2gif filename
 *		filename - Name of browse file
 *	or:
 *		hdf2gif filename outputdirectory
 *
 *  Description:
 *	Extract images from HDF file and convert to GIFs.
 *      This is the main code for the hdf2gif program.
 *      This code is used by the Web Gateway when a user requests to
 *      view a browse image.  The images are sent from the data
 *      centers in HDF files.  We cache them and use this program to
 *      extract the images and make GIF files.  The GIFs are displayed
 *      to the user.  When this program is run by the Web Gateway, it
 *      is run by user wwwuser and is invoked with the output directory
 *      specified.  The HDF filename is actually generated in the IK
 *      layer software when the client receives the HDF file.  This
 *      program can also be run standalone to extract GIF images from
 *      HDF files.
 *
 *      The hdf2gif program is composed of this file, sendbrowse.c, 
 *      a function in this file taken from NCSA : r24r8, convert8to24,
 *	an external program, a converted version of NCSA's hdf2jpeg, hdf2jpg.c,
 *      an external program, writeGIF.c,
 *      an ODL extensions file, odlext.c,
 *      and the HDF library (linked in).
 *      The HDF routines are declared externally in df.h.
 *      See the Makefile for details (hah!).
 *
 *  Program Interface Standards:
 *      Besides the program invocation (see Usage above), the outputs
 *      of this program are in a VERY specific format.  Besides writing
 *      GIF files, this program communicates specific messages to the
 *      Web Gateway by printing them to standard output.  So,
 *      DO NOT ADD PRINT STATEMENTS UNLESS YOU KNOW THE WEB GATEWAY
 *      KNOWS HOW TO READ WHAT YOU ARE PRINTING!
 *      This code only prints three things to standard output:
 *      1: If there was any textual description information in the HDF
 *      file (metadata), this code writes it to a file and prints:
 *         "Description-Filename: filename.des"
 *      without the quotes, with the actual filename, and followed by
 *      a blank line.
 *      2: For each GIF/JPEG file produced it prints:
 *         "Image-pathname: filename.gif" or 
 *	   "Image-pathname: filename.jpg"
 *      without the quotes, with the actual filename, and followed by
 *      a blank line.
 *      3: If there were any errors in processing, the code prints an
 *      informative error message preceeded by "X-Error: ".
 *
 *      For debugging old and new features, write the print statement
 *      like this:
 *         DEBUG(printf ("Debug - The value is %d\n", value));
 *      and set the switch: debug_on = TRUE.  Then you can run hdf2gif
 *      manually and see all the debug messages.  But when you check
 *      the code back into RCS it should have debug_on = FALSE.
 *
 *  Notes:
 *      BEFORE MODIFYING THIS CODE, YOU WILL PROBABLY NEED TO HAVE A
 *      COPY OF THE HDF REFERENCE MANUAL FROM NCSA
 *      (http://hdf.ncsa.uiuc.edu).  I (Larry) used Version 4.1r2
 *      from March 1998.  (I used the pdf version and viewed it using
 *      the free Adobe Reader.  Don't print it! It's 370 pages!!!)
 *
 *      HDF Intro:
 *      HDF functions used, or refered to, in this module begin with:
 *         H, AN, DFAN, SD, DFR8, and DF24.
 *      H functions provide basic HDF access to files.
 *      AN and DFAN functions provide access to textual information
 *      stored in the file as HDF "annotations".  This information can
 *      retrieved with either AN or DFAN, regardless of which was used
 *      to store it.  AN and DFAN are called "interfaces" because they
 *      are different views into the file for the same data.
 *      Annotation data comes in four types: file labels, file
 *      descriptions, data labels and data descriptions.
 *      SD functions provide access to structured information stored in
 *      the file using SD.  This is generally parameter/value groups of
 *      textual information (in ODL or PVL format).  SD is also called
 *      an interface.  It may stand for "Science Data".
 *      DFR8 functions provide access to 8-bit raster images.
 *      DF24 functions provide access to 24-bit raster images.
 *
 *      Chaos (and avoiding mind reading software):
 *      FOR THE BROWSE IMAGES RECEIVED BY THE WEB GATEWAY, data centers
 *      store textual information (often called "metadata") in the HDF
 *      files.  Unfortunately, THERE IS NO STANDARD/COMMON METHOD OF
 *      STORING THE METADATA AMONG OUR DATA PROVIDERS.  We've seen HDF
 *      files with metadata in SD format, annotation file labels and/or
 *      file descriptions and/or annotation data labels and/or data
 *      descriptions.  Since ther is no commonality, we currently seek
 *      all the information we can and throw it all into a single
 *      metadata file, whose name ends in ".des".
 *
 *      As in the B0SOT (V0 X-Client), we set the maximum number of
 *      metadata groups called COREMETADATA.* to 10.  This value can
 *      be adjusted.  It is defined below as MAX_METADATA_GROUPS.
 *
 *  Revision history:
 *	95-05-09 BEB Written
 *      May-June 1998, Larry P. Rosen, major restructuring.  With
 *         help from Jim Amrhein and NCSA.  Added 24-bit image
 *         handling, AN annotations, and SD metatdata.  All the
 *         comments you see in this version are by me.
 *
 *      December 1998, Jim Amrhein, JPEG support. Code used was obtained
 *	   from NCSA's hdf8to24 and hdf2jpeg. I've added the methods
 *	   convert8to24(),   
 *
 *
 *********************************************************************/

#include <stdio.h>
#include <malloc.h>
#include <unistd.h>
#include "gd.h"                 /* Bring in the gd library functions */
#include "gdfontg.h"
#include "df.h"                 /* Gets HDF functions here. */
#include "odldef.h"
#include "odlext.h"
#include "strutils.h"

#ifdef HAVE_STRINGS_H
#include <strings.h>
#else
#include <string.h>
#endif /*HAVE_STRINGS_H*/

/* I can't believe I need to include this. BEB */
/* Don't need to; it's just convenient. Used for gif filenames. LPR */
#define MY_FILENAME_MAX 1024
#define DISCRETE_MAX 30

/* I want a boolean type. */
#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

/* If debug is turned on, print debug messages to stdout. */
/*static int debug_on = FALSE;*/
static int debug_on = FALSE;
#define DEBUG(a) debug_on ? (void)(a) : (void)NULL

/* The maximum number of metadata groups called "COREMETADATA.*".  */
#define MAX_METADATA_GROUPS 10

/* The number of bytes in a kilobyte. */
#define BYTES_PER_KB 1024

/* the quality of the JPEG compression.  */
#define JPEG_COMPRESSION_QUALITY 80;

/* typedef's from NCSA: */
typedef unsigned char UCHAR;
typedef unsigned int UINT;

/* Global Variables */
uint8       red_comp[256], green_comp[256], blue_comp[256];
char        DISCRETE[DISCRETE_MAX][DISCRETE_MAX];

/* External functions: */
/* WriteGIF writes the image to the specified file as a GIF. */
extern int WriteGIF (char *outFilename, UCHAR *image, int xdim,
		     int ydim, char *colormap);

/* hdf2jpg - given a  24 JPEG compressed hdf filename extracts them into  
images */
extern int hdf2jpg ( char* infile, char *outfile, int multijpg, int process_metadata);

extern int Hdf5Process (char *inputFilename, char *base);

/* Local functions: */

/* FileExists returns TRUE if the specified file can be opened. */
int FileExists (char *origFilename);

/* Decompress uncompresses the specified file if it is compressed. */
char  *Decompress (char *origFilename);

/* GetFilenameBase gets a base for output filenames. */
char  *GetFilenameBase (char *inputFilename, char *outputDir);

/* PIDToString converts the process ID to a string. */
char  *PIDToString (pid_t pid);

/* GetAllAnnotations gets all the textual information stored in the 
   HDF file as HDF "annotations". Uses AN interface. */
int   GetAllAnnotations (int32 inputFD, char *metaFilename);

int   GetSpecificAnnotations (int32 inputFD, char *metaFilename, int32 imageNo);

/* GetAnn uses HDF AN functions to get one annotation type of info from
   the file. */
void  GetAnn (int annotNumber, int32 ANID, ann_type annType,
	      FILE *outFD, char *printTitle, int32 inputFD);

/* GetSDMetadata gets any structured textual information stored in the
   HDF file using the SD interface. */
int   GetSDMetadata (char *inputFilename, char *metaFilename,
		     int isMetadata, int32 inputFD);

/* GetV0StyleMetadata gets structured textual information stored in the
   HDF file by the same method found in version 7 of the B0SOT (the V0
   X-Client).  This uses the SD interface. */
char  *GetV0StyleMetadata (int32 nglobals, int32 SDID,
			   char *metaFilename, int32 *size_ptr,
			   int32 inputFD);

/* WriteMetadata writes the SD info found to the specified file. */
void  WriteMetadata (char *metaFilename, char *metadata, int32 size);

/* decodeAnnot decodes the type of annotation to create the label */
int decodeAnnot (char *buffer, char *units, int *colorbar); 

/* getPaletteLabels gets the labels for the palette from the annotations. */
int getPaletteLabels (int32 inputFD, int n8Images, char *units, int *colorbar);

/* WritePalette creates and writes the palette of colors used for the image. */
void WritePalette (char *palFilename, char *colormap, int n8Images, 
                                                       int32 inputFD);

/* ReadSDS tries to read SD info for the specified group name. */
char  *ReadSDS (int32 sds_id, char *metaname, int32 *size_ptr,
		int32 inputFD);

/* ConvertSemiColonsToSpace does what it is called. */
static void ConvertSemiColonsToSpace (char *str);

/* Process8BitImages extracts all 8-bit images, converts them to GIFs and 24 
   bit JPEGs, and writes them. */
void  Process8BitImages (char *inputFilename, int n8Images,
			 char *base, int32 inputFD);

/* Process24BitImages extracts all 24-bit images, converts them to
   24-bit JPEGs, and writes them. */
void  Process24BitImages (char *inputFilename, int n24Images,
			  char *base, int32 inputFD, int process_metadata);

void  ProcessMetadata (char *inputFilename, int n24Images,
                          char *base, int32 inputFD);

/* r24r8() converts 24-bit raster image to 8-bit raster image. NCSA. */
int   r24r8 (int32 xres, int32 yres, UCHAR *dat24, UCHAR *dat8, 
             int cres, UCHAR *cdat);

/* convert8to24() converts 8-bit raster image to 24-bit raster image. NCSA. */
static intn
convert8to24(uint8 *img8_buf, uint8 *img24_buf, int32 img_xdim,
             int32 img_ydim);

/* Compresses a 24 bit image using the JPEG compression then saves it as a HDF file */
intn
writeJPEG_HDF (char  *hdfFilename, uint8 *img24_buf, int32  xdim, int32 ydim);


/* = * = * = * = * = * = * = * = * = * = * = * = * = * = * = * = * = *
 *  Name:	main - Main program for sendbrowse
 *
 *  Usage:	main (int argc, char *argv[]);
 *		argc - Count of arguments
 *		argv - Array of arguments
 *
 *  Description:
 *	Figure out how to send argv[1] to the user as a browse image.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

main (int argc, char *argv[]) {

   char  *origFilename = NULL;  /* Pointer to original filename. */
   char  *outputDir = NULL;     /* Pointer to output directory name. */
   char  *inputFilename = NULL; /* Decompressed input filename. */
   int32 inputFD;               /* Input file descriptor (HDF). */
   int   n8Images;              /* Number of raster 8-bit images. */
   int   n24Images;             /* Number of raster 24-bit images. */
   char  *base = NULL;          /* Base of output file names. */
   char  *metaFilename = NULL;  /* Filename for all metadata. */
   int   isMetadata = FALSE;    /* Found any metadata? (AN or SD). */
   int   process_metadata = FALSE; /* Does ProcessMetadata need to be ran? */
   int   status;

   /* Check the calling sequence. */
   if (argc < 2) {
     printf ("X-Error: main: Usage: hdf2gif  hdf_filename  \
output_directory.\n");
      exit (1);
   }
   origFilename = argv[1];
   if (argc >= 3) {
      outputDir = argv[2];
   }
   /* Test if file exists. */
   if (! FileExists (origFilename)) {
      printf ("X-Error: main: Could not locate the file you \
specified: %s.\n", origFilename);
      exit (1);
   }
   /* If the file is compressed, decompress it. */
   inputFilename = Decompress (origFilename);

   /* Get the base of the output filenames. */
   base = GetFilenameBase (inputFilename, outputDir);
/*   printf ("inputFilename is %s, outputDir is %s, base is %s\n", inputFilename, outputDir, base); */

   /* Use HDF 5 function to check if the file is an HDF 5 file. */
   if (! Hishdf (inputFilename)) {
      DEBUG(printf ("Debug -  base: >%s< with stringlen: >%d<\n",
         base, strlen(base)));
      DEBUG(printf ("Debug - : File %s is hdf5, location is %s\n", inputFilename, outputDir));
      status = Hdf5Process (inputFilename, base);
   }
   else {
      /* Open the input file for HDF. */
      if ((inputFD = Hopen (inputFilename, DFACC_READ, 0)) < 0) {
         printf ("X-Error: main: Could not open the image file %s\n",
	         inputFilename);
         exit (1);
      }

      /* We passed the HDF4 test, it is an HDF 4 file. */
      DEBUG(printf ("Debug - : File %s is hdf4\n", inputFilename));

      /*  Count the number of 8-bit and 24 bit images in the file. */
      n8Images = DFR8nimages (inputFilename);
      n24Images = DF24nimages (inputFilename);

      DEBUG(printf ("Debug - : There are %d eight bit images\n", n8Images));
      DEBUG(printf ("Debug - : There are %d twentyfour bit images\n", n24Images));

      /* If there are no images, quit. */
      if (n8Images <= 0 && n24Images <= 0) {
         Hclose (inputFD);
         printf ("X-Error: main: There are no images in the HDF file.\n");
         exit (1);
      }
      if (n8Images > 0) {
         /* Process the 8-bit images. */
         Process8BitImages (inputFilename, n8Images, base, inputFD);
      }
      if (n24Images > 0) {
         if ( n8Images <= 0  ) {
            process_metadata = TRUE;
         }
         /* Process the 24-bit images. */
         Process24BitImages (inputFilename, n24Images, base, 
                                        inputFD, process_metadata);
      }
      /* Close the HDF access. */
      Hclose (inputFD);

      /* Done. */
      status = 0;
   }
   exit (status);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Function: FileExists
 * Parameters: char *origFilename   - Specified input filename.
 * Return: int FALSE  - false if file doesn't exist.
 *     or  int TRUE   - true if file exists.
 */
int FileExists (char *origFilename) {

   FILE *fileStream;   /* Input file stream */
   char *ptr;          /* Pointer to position in filename */

   /* See if the file actually exists. */
   if ((fileStream = fopen (origFilename, "r")) == NULL) {
      return (FALSE);
   }
   fclose (fileStream);
   return (TRUE);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Function: Decompress
 * Parameters: char *origFilename - original fullpath filename.
 * Return: char *decompressedFilename  - name of decompressed file.
 * Notes: Allocates space for decompressed file name and sets it.
 *       Assumes that file ends in ".Z" if compressed 
 *           (case sensitive!).
 *       For example, an input file could be: myimages.hdf.Z
 *       Returns pointer to original filename if it was not compressed.
 *       Exits on failure (not return).
 *       Uses "zcat".
 */
char *Decompress (char *origFilename) {

   char *ptr;                  /* Pointer to position in filename. */
   int  newLength;             /* Length of new filename (no path). */
   char *decompressedFilename; /* New filename of decompressed file. */
   int  origLength;            /* Length of input fullpath filename. */
   char *cmd;                  /* Unix command to uncompress. */

   /* Original full pathname length. */
   origLength = strlen (origFilename);

   /* Find last . in filename. */
   ptr = strrchr (origFilename, '.');

   /* If filename doesn't end in .Z, assume it is not compressed.
      Return it. */
   if (strcmp (ptr, ".Z") != 0) {

      return (origFilename);

   } else {
      /* File ended in .Z. Assume compressed. Try to decompress. */

      /* Extract just the filename (skipping path). */
      if ((ptr = strrchr (origFilename, '/')) == NULL) {
         ptr = origFilename;
      } else {
         ++ptr;
      }
      /* Length of new filename (without path) is length of ptr 
         (filename skipping path) + 1 for \0 and -2 for .Z we are
         removing. */
      newLength = strlen (ptr) - 1;

      /* Allocate memory for new filename string. */
      decompressedFilename = (char *) malloc (newLength);
      if (decompressedFilename == NULL) {
         printf ("X-Error: Decompress: System could not allocate \
memory for decompressed filename.\n");
         exit (1);
      }
      /* Copy the input filename without the ".Z".  Number of
         characters = (newLength - 1) = strlen(ptr) - 2 */
      (void) strncpy (decompressedFilename, ptr, newLength - 1);

      /* Add a final null to string. */
      decompressedFilename[newLength - 1] = '\0';

      /* Now form the decompress command. */
      cmd = (char *) malloc (9 + origLength + newLength);
      if (cmd == NULL) {
         printf ("X-Error: Decompress: System could not allocate \
memory for decompress command.\n");
         exit (1);
      }
      sprintf (cmd, "zcat %s > %s",  origFilename,
	       decompressedFilename);

      /* Now execute the decompress command. */
      if (system (cmd)) {
         printf ("X-Error: Decompress: Unable to uncompress the \
file %s\n", origFilename);
         exit (1);
      }
      free (cmd);
      return (decompressedFilename);
   }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Function: GetFilenameBase
 * Description: Construct base of output filenames based on input
 *    filename, output directory, and pid.  Allocates memory here.
 * Parameters: char *inputFilename   - Input filename.
 *             char *outputDir       - Output directory.
 * Return: char *base     - Output filename base.
 */
char *GetFilenameBase (char *inputFilename, char *outputDir) {

   char  *justInputName;        /* The input filename without path. */
   pid_t pid;                   /* Process ID. */
   char  *pidString;            /* pid as a string. */
   char  *base = NULL;          /* Base of output file names. */
   int   len;

   /* Extract just the input filename (without path info). */
   if ((justInputName = strrchr (inputFilename, '/')) == NULL) {
      justInputName = inputFilename;
   } else {
      ++justInputName;
   }
   /* Get the process ID. */
   pid = getpid ();
   pidString = PIDToString (pid);

   /* If output directory is specified, set the base for output
      filenames base =  outputdir/ + inputfilename - ".hdf" +'\0'. */
   if (outputDir != NULL) {
      base = (char *) malloc (strlen (outputDir) +
			      strlen (justInputName) +
                              strlen (pidString) - 2);
      if (base == NULL) {
         printf ("X-Error: GetFilenameBase: System could not allocate \
memory for base of output file names.\n");
         exit (1);
      }
      (void) strcpy (base, outputDir);
      /* If output directory doesn't end in "/", add a "/". */
      if (outputDir[strlen (outputDir) - 1] != '/') {
         (void) strcat (base, "/");
      }
      /* The following strips out the ".hdf" ending. */
      len = strlen (justInputName) - 4;
      (void) strncat (base, justInputName, len);
      base[strlen(outputDir) + len + 1] = '\0';
      (void) strcat (base, pidString);
   } else {
   /* Else no output directory specified; set the base for output
      filenames base =  inputfilename - ".hdf" +'\0'. */
      base = (char *) malloc (strlen (justInputName) + \
                              strlen (pidString) - 3);
      if (base == NULL) {
         printf ("X-Error: GetFilenameBase: System could not allocate \
memory for base of output file names.\n");
         exit (1);
      }
      len = strlen (justInputName) - 4;
      (void) strncpy (base, justInputName, len);
      base[len] = '\0';
      (void) strcat (base, pidString);
   }
   DEBUG(printf ("Debug - base: >%s< with stringlen: >%d<\n",
         base, strlen(base)));
   free (pidString);    /* Was allocated in PIDToString() */
   return (base);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Function: PIDToString
 * Description: Convert a pid to a string.  Allocates memory for it.
 * Parameters: pid_t pid   - process ID
 * Return: char  *pid_string   - process ID as string.
 * Notes: Assumes pid is less than 100 characters long. 
 */
char *PIDToString (pid_t pid) {

   char buffer[100];
   int  length;
   char *pidString = NULL;

   /* Print pid into a char buffer and append null. */
   sprintf (buffer, "%d%c", pid, '\0');

   /* Could now return buffer, but it is 100 chars long. Instead,
      allocate just enough space to hold the string. */
   length = strlen (buffer);
   pidString = (char *) malloc (length + 1);
   if (pidString == NULL) {
      printf ("X-Error: PIDToString: System could not allocate memory \
for pid string.\n");
      exit (1);
   }
   (void) strcpy (pidString, buffer);
   return (pidString);
}   

int GetSpecificAnnotations (int32 inputFD, char *metaFilename, int32 imageNo) {

   int32  ANID;                   /* ID for AN interface. */
   int32  numFileLabels;          /* Number of file labels. */
   int32  numFileDescs;           /* Number of file descriptions. */
   int32  numDataLabels;          /* Number of data labels. */
   int32  numDataDescs;           /* Number of data descriptions. */
   int32  status;                 /* Status of AN function call. */
   FILE   *outFS;                 /* File stream for output file. */

   /* Initialize the AN interface and obtain an interface id. */
   ANID = ANstart (inputFD);

   /* Get the annotation information. */
   status = ANfileinfo (ANID, &numFileLabels, &numFileDescs, \
                        &numDataLabels, &numDataDescs);

   DEBUG(printf ("Debug - AN interface found the following:\n"));
   DEBUG(printf ("Debug - Number of file labels:       %d\n",
                 numFileLabels));
   DEBUG(printf ("Debug - Number of file descriptions: %d\n",
                 numFileDescs));

   DEBUG(printf ("Debug - Number of data labels:       %d\n",
                 numDataLabels));
   DEBUG(printf ("Debug - Number of data descriptions: %d\n\n",
                 numDataDescs));

   /* If there is no AN information, end AN interface and return. */
   if (numFileLabels == 0 && numFileDescs == 0 && numDataLabels == 0 &&
       numDataDescs == 0) {

      status = ANend (ANID);
      return (FALSE);
   }
   /* Else go on ... */

   /* Open the file for writing. */
   outFS = fopen (metaFilename, "w");
   if (outFS == NULL) {
      printf ("X-Error: GetAllAnnotations: Could not open file for \
writing annotation.");
      status = ANend (ANID);
      return (FALSE);
   }
   /* Get the file labels. */
   GetAnn (imageNo, ANID, AN_FILE_LABEL, outFS, "File Label",
              inputFD);
   /* Get the file descriptions. */
   GetAnn (imageNo, ANID, AN_FILE_DESC, outFS, "File Description",
              inputFD);
   /* Get the data label. */
   GetAnn (imageNo, ANID, AN_DATA_LABEL, outFS, "Data Label",
              inputFD);
   /* Get the data description. */
   GetAnn (imageNo, ANID, AN_DATA_DESC, outFS, "Data Description",
              inputFD);
   /* Terminate access to the AN interface. */
   status = ANend (ANID);

   /* Close the output file. */
   fprintf (outFS, "\n");
   fclose (outFS);
   DEBUG(printf ("Debug - annotation info found. Put in %s.\n",
                 metaFilename));
   return (TRUE);
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Function: GetAllAnnotations
 * Description: Get all of the textual information stored in the HDF
 *    file as HDF annotations.
 *    Uses AN interface because it is more straight forward than the
 *    DFAN interface.  Either could work.  (See notes in file header.)
 *    (Either HDF interface can get access to textual information stored
 *    as annotations data.)
 *    If found, writes to specified metadata file.
 * Parameters: int32 inputFD       - Input file stream.
 *             char  *metaFilename - Output filename.
 * Return: int TRUE if annotation info found.  FALSE otherwise.
 */
int GetAllAnnotations (int32 inputFD, char *metaFilename) {

   int32  ANID;                   /* ID for AN interface. */
   int32  numFileLabels;          /* Number of file labels. */
   int32  numFileDescs;           /* Number of file descriptions. */
   int32  numDataLabels;          /* Number of data labels. */
   int32  numDataDescs;           /* Number of data descriptions. */
   int32  status;                 /* Status of AN function call. */
   FILE   *outFS;                 /* File stream for output file. */
   int32  i;                      /* Counter */

   /* Initialize the AN interface and obtain an interface id. */
   ANID = ANstart (inputFD);

   /* Get the annotation information. */
   status = ANfileinfo (ANID, &numFileLabels, &numFileDescs, \
                        &numDataLabels, &numDataDescs);

   DEBUG(printf ("Debug - AN interface found the following:\n"));
   DEBUG(printf ("Debug - Number of file labels:       %d\n", 
		 numFileLabels));
   DEBUG(printf ("Debug - Number of file descriptions: %d\n",
		 numFileDescs));
   DEBUG(printf ("Debug - Number of data labels:       %d\n",
		 numDataLabels));
   DEBUG(printf ("Debug - Number of data descriptions: %d\n\n",
		 numDataDescs));

   /* If there is no AN information, end AN interface and return. */
   if (numFileLabels == 0 && numFileDescs == 0 && numDataLabels == 0 &&
       numDataDescs == 0) {

      status = ANend (ANID);
      return (FALSE);
   }
   /* Else go on ... */

   /* Open the file for writing. */
   outFS = fopen (metaFilename, "w");
   if (outFS == NULL) {
      printf ("X-Error: GetAllAnnotations: Could not open file for \
writing annotation.");
      status = ANend (ANID);
      return (FALSE);
   }

   /* Get the file labels. */
   for (i = 0; i < numFileLabels; ++i) {
      GetAnn (i, ANID, AN_FILE_LABEL, outFS, "File Label",
	      inputFD);
   }
   /* Get the file descriptions. */
   for (i = 0; i < numFileDescs; i++) {
      GetAnn (i, ANID, AN_FILE_DESC, outFS, "File Description",
	      inputFD);
   }
   /* Get the data label. */
   for (i = 0; i < numDataLabels; ++i) {
      GetAnn (i, ANID, AN_DATA_LABEL, outFS, "Data Label",
	      inputFD);
   }
   /* Get the data description. */
   for (i = 0; i < numDataDescs; ++i) {
      GetAnn (i, ANID, AN_DATA_DESC, outFS, "Data Description",
	      inputFD);
   }
   /* Terminate access to the AN interface. */
   status = ANend (ANID);

   /* Close the output file. */
   fprintf (outFS, "\n");
   fclose (outFS);

   DEBUG(printf ("Debug - annotation info found. Put in %s.\n",
		 metaFilename));
   return (TRUE);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Function: GetAnn
 * Description: Extract the specific file annotation if any.
 *    Uses AN interface because it is more straight forward than DFAN.
 *    See notes in file header.  (Either HDF interface can get access
 *    to textual information stored as annotations data.)
 *    If found, write to file.
 * Parameters: int      annotNumber - Number of annotation in file.
 *             int32    ANID        - AN interface to input file.
 *             ann_type annType     - Annotation type.
 *             FILE     *outFS      - Output file stream.
 *             char     *printTitle - Title to print for annotation.
 *             int32    inputFD     - Input HDF file descriptor.
 * Return: none.
 */
void GetAnn (int annotNumber, int32 ANID, ann_type annType,
	     FILE *outFS, char *printTitle, int32 inputFD) {

   int32  itemID;           /* Item (label, data) ID. */
   int32  length;           /* Length of string to read. */
   char   *buffer = NULL;   /* Buffer for holding annotation read. */
   int32  status;           /* Status of AN function call. */
   size_t numWrote;         /* Number of chars just written to file. */

   /* Get the identifier for the current annotation. Return if fail. */
   itemID = ANselect (ANID, annotNumber, annType);
   if (itemID < 0) {
      return;
   }
   /* Get the length of the annotation. Return if fail. */
   length = ANannlen (itemID);
   if (length < 0) {
      return;
   }
   DEBUG(printf ("Debug - Length of %s annotation %d: %d\n",
                 printTitle, annotNumber + 1, length));

   /* Allocate buffer memory for annotation. Exit if fail. */
   buffer = (char *) malloc (length + 1);
   if (buffer == NULL) {
      Hclose (inputFD);
      printf ("X-Error: GetAnn: System could not allocate memory for \
annotation: size %d.\n", length);
      exit (1);
   }
   /* Read the annotation into buffer. Return if fail. */
   status = ANreadann (itemID, buffer, length);
   if (status < 0) {
      printf ("X-Error: GetAnn: Failed to read file label %d.\n",
	      annotNumber + 1);
      free (buffer);
      return;
   }
   /* Write buffer to output file. */
   fprintf (outFS, "WG_BEGIN %s : \
---------------------------------------\n",
            printTitle);
   numWrote = fwrite (buffer, sizeof (char), (size_t) length, outFS);
   fprintf (outFS, "\nWG_END %s : \
---------------------------------------\n",
            printTitle);

   free (buffer);

   /* Terminate access to the annotation. */
   status = ANendaccess (itemID);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Function: GetSDMetadata
 * Description: Get any structured textual information stored using
 *              the SD interface.  If description information was
 *              previously found using AN interface, then this function
 *              will just append any SD info.
 * Parameters: char  *inputFilename  - Input filename (HDF).
 *             char  *metaFilename   - Output metadata file.
 *             int   isMetadata      - TRUE if already found AN info.
 *                                     FALSE if not.
 *             int32 inputFD         - Input HDF file descriptor.
 * Return: int TRUE - if found SD information.
 * Return: int isMetadata - (previous value) if no SD info.
 */
int GetSDMetadata (char *inputFilename, char *metaFilename,
		   int isMetadata, int32 inputFD) {

   int32  SDID;              /* Input file desc (SD - HDF) */
   int32  ndatasets;         /* Number of datasets in HDF. */
   int32  nglobals;          /* Number of Metadata groups. */
   int32  status;            /* Status of function call.   */
   char   *metadata = NULL;  /* Metadata read or NULL if none found. */
   int32  size = 0;          /* Size of metadata buffer in chars. */

   /* Start the SD interface to the HDF file. */
   if ((SDID = SDstart (inputFilename, DFACC_READ)) < 0) {
      printf ("X-Error: GetSDMetadata: Could not open the SDS \
interface for the HDF file.\n");
      return (isMetadata);
   }
   DEBUG(printf ("Debug - SD started for metadata.\n"));
   DEBUG(printf ("Debug - get SD file info\n"));

   /* Get number of global attributes. */
   if ((status = SDfileinfo (SDID, &ndatasets, &nglobals)) 
       != SUCCEED) {
      DEBUG(printf ("Debug - didn't get SD file info\n"));
      /* Close the SDS connection. */
      SDend (SDID);
      return (isMetadata);
   }
   DEBUG(printf ("Debug - got SD file info!\n\tnum datasets: %d, \
num globals: %d\n", ndatasets, nglobals));

   /**************************************
    * First we try the V0 X-Client method.
    **************************************/
   if (nglobals > 0) {
      DEBUG(printf ("Debug - First try V0 X-client method.\n"));
      metadata = GetV0StyleMetadata (nglobals, SDID, metaFilename,
				     &size, inputFD);
   }
   /************************************************************
    * If we didn't get V0 style metadata, try for BROWSE_HEADER.
    ************************************************************/
   if (!metadata) {
      DEBUG(printf ("Debug - V0 failed, try BROWSE_HEADER.\n")); 
      metadata = ReadSDS (SDID, "BROWSE_HEADER", &size, inputFD);
   }

   /********************************************************
    * If we didn't get BROWSE_HEADER, try for COREMETADATA.0
    ********************************************************/
   if (!metadata) {
      DEBUG(printf ("Debug - BROWSE_HEADER failed, try COREMETADATA.0\n"));
      metadata = ReadSDS (SDID, "COREMETADATA.0", &size, inputFD);
   }

   /*****************************************************
    * If there was metadata, write it to the output file.
    *****************************************************/
   if (metadata) {
      /* Write the metadata to the output file. */
      WriteMetadata (metaFilename, metadata, size);

      /* Free the allocated buffer. */
      free (metadata);

      /* Set flag indicating metadata was found in the HDF file. */
      isMetadata = TRUE;
   }
   /* ELSE we didn't find any SD metadata. That's OK. */

   /* Either way, close the SD interface and return. */
   SDend (SDID);
   return (isMetadata);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Function: GetV0StyleMetadata
 * Description: Get V0 style file metadata if any.  This gets
 * structured textual information stored in the HDF file by the
 * same method found in version 7 of the B0SOT (the V0 X-Client).
 * This uses the SD interface.
 *
 * Parameters: int32 nglobals       - Number of globals in metadata.
 *             int32 SDID           - Input file desc SD - HDF.
 *             char  *metaFilename  - Output file name.
 *             int32 *size_ptr    - Pointer to size of metadata buffer.
 *             int32 inputFD        - Input HDF file descriptor.
 * Return: char *metadata  - A buffer of metadata read or NULL if none.
 * Global Definition: MAX_METADATA_GROUPS = 10
 */
char *GetV0StyleMetadata (int32 nglobals, int32 SDID,
			  char *metaFilename, int32 *size_ptr,
			  int32 inputFD) {

   int32  i;                       /* Counter. */
   int32  status;                  /* Status of function call. */
   int32  type;                    /* Type of objects (char). */
   int32  attr_size;               /* Size of attribute. */
   int32  size = 0;                /* Size of all SD metadata read. */
   char   metaName[80];            /* Name of metadata to read. */
   char   str[80];
   char   *metadata = NULL;        /* Array of all metadata read. */
   int    j, ret_status;

   /*  Determine the size of all global attributes */
   for (i = 0; i < nglobals && i < MAX_METADATA_GROUPS; ++i) {
      DEBUG(printf ("Debug - get V0 style attr info\n"));
      if ((status = SDattrinfo (SDID, i, metaName, &type, &attr_size))
           != SUCCEED) {
         DEBUG(printf ("Debug - failed\n"));
         break;
      } else {
         DEBUG(printf ("Debug - got it!\n  metaName: %s, type: %d, \
attrsize: %d\n", metaName, type, attr_size));
      }
      sprintf (str, "COREMETADATA.%d", i);
      for (j=0; j < strlen (metaName); ++j) {
         metaName[j] = toupper (metaName[j]);
      }
      if ((strcmp (metaName, str) == 0) && (type == DFNT_CHAR8)) {
         size += attr_size;
      }
   }
   /* Allocate space for metadata. */
   if (size) {
      metadata = (char *) malloc (DFKNTsize (DFNT_CHAR8) * size);
      if (metadata == NULL) {
	 Hclose (inputFD);
         printf ("X-Error: GetV0StyleMetadata: System could not \
allocate memory for metadata.\n");
         exit (1);
      }
      metadata[0] = 0;
   }
   /* Read metadata in. */
   for (i = 0, size = 0; i < nglobals && i < MAX_METADATA_GROUPS; ++i){
      if ((status = SDattrinfo (SDID, i, metaName, &type, &attr_size)) 
          != SUCCEED){
         metadata = NULL;
         break;
      }
      sprintf (str, "COREMETADATA.%d", i);
      for (j=0; j < strlen (metaName); ++j) {
         metaName[j] = toupper (metaName[j]);
      }
      if ((strcmp (metaName, str) == 0) && (type == DFNT_CHAR8)) {
         size += attr_size;
         SDreadattr (SDID, 0, &metadata[strlen (metadata)]);
         metadata[size-1] = 0;
      }
   }
   /* If metadata, set size of metadata buffer and return buffer. */
   if (size > 0) {
      DEBUG(printf ("Debug - We got V0 style metadata, size: %d.\n",
		    size));
      *size_ptr = DFKNTsize (DFNT_CHAR8) * size;
      return (metadata);
   } else {
      DEBUG(printf ("Debug - We didn't find any V0 style metadata.\n"));
      *size_ptr = 0;
      return (NULL);
   }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Function: WriteMetadata
 * Description: Write metadata to file.
 * Parameters: char  *metaFilename  - Output file name.
 *             char  *metadata    - Buffer of all metadata matching V0.
 *             int32 size         - Size of buffer to write.
 */
void WriteMetadata (char *metaFilename, char *metadata, int32 size) {

   FILE *metaFD;
   int numWrote;

   /* Open the metadata file into which we will dump the metadata. */
   /* "append" in case we wrote some AN description stuff before. */
   /* metaFD - output metadata file descriptor. */
   metaFD = fopen (metaFilename, "a");
   if (metaFD == NULL) {
      printf ("X-Error: WriteMetadata: Could not open output file \
for metadata: %s\n", metaFilename);
      return;
   }
   /* Put the data into the File */
   fprintf (metaFD, "\nWG_BEGIN SD Metadata: \
--------------------------------\n");
   numWrote = fwrite (metadata, sizeof (char), (size_t) size, metaFD);
   fprintf (metaFD, "\nWG_END SD Metadata: \
----------------------------------\n");
   fprintf (metaFD, "\n");
   fclose (metaFD);
   return;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Function: ReadSDS
 * Description: Trys to read metadata for specified name.
 * Parameters: int32  SDID       - SD interface descriptor.
 *             char*  metaname   - Name of metadata in HDF file.
 *             int32  *size_ptr  - Pointer to size of metadata buffer.
 *             int32  inputFD    - Input HDF file descriptor.
 * Returns: char  *metadata  - Buffer of all metadata matching
 *                             specified name.
 */
char *ReadSDS (int32 sds_id, char *metaname, int32 *size_ptr,
	       int32 inputFD) {

   int32  attr_index;
   int32  attr_info;
   /* attr_name - Name assigned to the attribute/dataset/dimension */
   char  attr_name[MY_FILENAME_MAX];
   int32 number_type;  /* Data type of the attribute values */
   int32 counter;      /* Total number values in specified attribute */
   int32 buffer_size;
   char  *buffer;    /* Buffer for metadata */

   DEBUG(printf ("Debug - Inside ReadSDS\n"));
   DEBUG(printf ("Debug - sds_id = %d, metaname: %s\n",
		 sds_id, metaname));
   if ((attr_index = SDfindattr (sds_id, metaname)) >= 0) {

      attr_info = SDattrinfo (sds_id, attr_index, attr_name,
			      &number_type, &counter);

      DEBUG(printf ("Debug - attr_info=%d\n", attr_info));
      DEBUG(printf ("Debug - attr_name=%s\n", attr_name));
      DEBUG(printf ("Debug - Counter : %d\n", counter  ));
      DEBUG(printf ("Debug - number_type : %d\n",
		    DFKNTsize (number_type)));

      /* How much memory do we allocate? */
      buffer_size = counter * DFKNTsize (number_type);
 
      if ((buffer = (char *) malloc (buffer_size + 1)) == NULL) {
	 Hclose (inputFD);
         printf ("X-Error: ReadSDS: Insufficient Memory for \
Buffer\n");
         exit (1);
      }
      /* Get the Data */
      attr_index = SDreadattr (sds_id, attr_index, buffer);

      /* DEBUG(printf ("Debug - Buffer %s \n", buffer)); */
      /* Strip all the ";" from the variable "buffer" before it is
	 wriiten to the file */

      ConvertSemiColonsToSpace (buffer);

      /* Set the buffer size and return it. */
      *size_ptr = buffer_size;
      return (buffer);

   } else {
      *size_ptr = 0;
      return (NULL);
   }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * ConvertSemiColonsToSpace
 */
static void ConvertSemiColonsToSpace (char *str) {

   register int x;

   for (x = 0; str[x]; ++x) {
      if (str[x] == ';')
         str[x] = ' ';
   }
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Function: decodeAnnot
 * Description: Decode the annotation to create correct label
 * Parameters:  
 */
int decodeAnnot (char *buffer, char *units, int *colorbar) {
   
   char *p;
   char *tempbuf;
   int status;
   int i, j;

   status = 0;

   tempbuf = (char *) malloc (strlen (buffer));
   strcpy (tempbuf, buffer);
   StripLeadingBlanks(tempbuf);
   StripTrailingBlanks(tempbuf);

   p = NULL;
   p = strtok (tempbuf, "=");
   StripLeadingBlanks(p);
   StripTrailingBlanks(p);

   if (!strcmp(p, "UNITS")) { 
      DEBUG(printf ("Debug - in UNITS value of token is ...%s...\n", p));
      p = strtok ('\0', "\0");
      StripLeadingBlanks(p);
      StripTrailingBlanks(p);
      strcpy (units, p);
      status = 1;
   }

   if (!strcmp(p, "COLORBAR")) { 
      DEBUG(printf ("Debug - in COLORBAR value of token is ...%s...\n", p));
      for (i = 0; i <= 1; i++) {
         p = strtok ('\0', ",");
         if(p) colorbar[i] = atoi (p); 
      }
      status = 2;
   }

   if (!strcmp(p, "DISCRETE")) { 
      DEBUG(printf ("Debug - in DISCRETE value of token is ...%s...\n", p));
      for (i = 0; i < DISCRETE_MAX; i++) {
         p = strtok ('\0', ",");
         if(p) {
            strcpy (DISCRETE[i], p); 
         } else {
            strcpy (DISCRETE[i], "");
            i = DISCRETE_MAX;
         }
      }
      status = 3;
   }

   free (tempbuf);
   return (status);
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Function: getPaletteLabels
 * Description: Get palette labels from annotations
 * Parameters: 
 */
int getPaletteLabels (int32 inputFD, int n8Images, 
                        char *units, int *colorbar) {
   int found = 0;
   int i;
   int32 ANID;
   int32 numFileLab, numFileDesc, numDataLab, numDataDesc;
   int32 itemID;
   int32 length;
   char *buffer = NULL;
   int32 status;

   ANID = ANstart (inputFD);
   ANfileinfo (ANID, &numFileLab, &numFileDesc,
                              &numDataLab, &numDataDesc);

   if (numDataDesc == 0) {
      ANend (ANID);
      return (found);
   }

   /* Get identifier for current annotation */
   itemID = ANselect (ANID, n8Images, AN_DATA_DESC);
   if (itemID < 0 ) {
      ANend (ANID);
      return (found);
   }

   /* Get length of annotation */
   length = ANannlen (itemID);
   if (length < 0 ) {
      ANend (ANID);
      return (found);
   }

   /* Allocate memory for annotation */
   buffer = (char *) malloc (length +1);
   if (buffer == NULL) {
      printf ("X-Error: getPaletteLabels: System could not allocate \
memory for annotation: size %d.\n", length);
      exit (1);
   }

   /* Read the annotation into buffer */
   status = ANreadann (itemID, buffer, length);
   if (status < 0 ) {
      printf ("X-Error: getPaletteLabels: System could not read file \
label %d.\n", n8Images+1);
      free (buffer);
      return (found);
   }

   /* Decode value from annotation, return 1 if it is a palette label, */
   /* otherwise return zero. */
   status = decodeAnnot (buffer, units, colorbar);
   found =  status;

   free (buffer);
   ANend (ANID);
   return (found);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Function: WritePalette
 * Description: Write image palette to file.
 * Parameters:  char  *colormap     - Buffer of all metadata matching V0.
 */
void WritePalette (char *palFilename, char *colormap, 
                                      int n8Images, int32 inputFD) {

   /* Declare the image */
   gdImagePtr im;

   FILE *palFD;
   int  color, red, green, blue, black;
   int  label, n, i, j;
   int  margin = 2;
   int  high = 20;
   int  wide = 600;
   int  top = 0;
   int  valuestop;
   int  higher;

   char units[80];
   int  colorbar[2];
   int  discrete[DISCRETE_MAX]; 
   int  discrete_range[DISCRETE_MAX]; 
   char values[DISCRETE_MAX][DISCRETE_MAX];

   strcpy (units, "");
   colorbar[0] = 0;
   colorbar[1] = 255;
   discrete[0] = 0;
   discrete[1] = 255;
   discrete_range[0] = 0;
   discrete_range[1] = 255;
   strcpy (values[0], "1");
   strcpy (values[1], "256");

   for (i = 0; i < DISCRETE_MAX; i++) {
      strcpy (DISCRETE[i], "");
   }

   /* Open the palette file into which we will write the colorbar. */
   /* palFD - output metadata file descriptor. */
   palFD = fopen (palFilename, "wb");
   if (palFD == NULL) {
      printf ("X-Error: WritePalette: Could not open output file \
for palette: %s\n", palFilename);
      return;
   }

   if (inputFD == 0) {
      label = 0;
   }
   else {
      label = getPaletteLabels (inputFD, n8Images, units, colorbar);
      if (label == 2) {
         char value[DISCRETE_MAX] = "\0";
         sprintf (value, "%d", colorbar[0]);
         strcpy (values[0], value);
         sprintf (value, "%d", colorbar[1]);
         strcpy (values[1], value);
      }
      if (label == 3) {
         j = 0;
         for (i = 0; i < DISCRETE_MAX; i++) {
            discrete_range[i] = atoi (DISCRETE[j]);
            j++;
            strcpy (values[i], DISCRETE[j]);
            j++;
            if (!strcmp (DISCRETE[i], "")) {
               i = DISCRETE_MAX;
            }
         }
      }
   }

   /* Create image area, set background color to white */
   higher = margin+gdFontGiant->h+margin+high+2*gdFontGiant->h+margin;
   top = margin+gdFontGiant->h+margin;
   valuestop = top + high + gdFontGiant->h;
      
   im = gdImageCreate (wide, higher);

   gdImageColorAllocate(im, 255, 255, 255);
   black = gdImageColorAllocate(im, 0, 0, 0);

   /* Create the palette with the colorbar, create labels with the colorbar
      numbers, try to put ticks to associate numbers with colors. */
   gdImageString (im, gdFontGiant, margin, margin, units, black);

   /* Get the combination of colors from the colormap
      to get the real color. */
   j = 0;
   for (n = 0; n < 256; ++n) {
      red   = colormap[n];
      green = colormap[n+256];
      blue  = colormap[n+512];

      color = gdImageColorAllocate (im, red, green, blue);
      /* upper left first (x1, y1), then lower right (x2, y2) */
      /* gdImagePtr im, int x1, int y1, int x2, int y2, int color */
      gdImageFilledRectangle(im, n*2+margin,top, n*2+1+margin, high+top, color);

      if (discrete_range[j] == n) {
         gdImageString (im, gdFontGiant, n*2, top+high, "|", black);
         gdImageString (im, gdFontGiant, n*2, valuestop, values[j], black);
         j++;
      }
   }

   /* Put the palette into the File */
   gdImageGif(im, palFD);

   /* Report the filename */
   printf ("Image-pathname: %s\n\n", palFilename);

   /* Close the file, free image area, and return */
   fclose (palFD);
   gdImageDestroy(im);
   return;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Function: Process8BitImages
 * Description: Extract and convert the 8-bit images to GIF files.
 *    Prints the names of the GIF files.
 * Parameters: char  *inputFilename - HDF input filename.
 *             int   n8Images       - Number of 8-bit images in file.
 *             char  *base          - Output filename base.
 *             int32 inputFD        - Input HDF file descriptor.
 * Return: none.
 */
void Process8BitImages (char *inputFilename, int n8Images, char *base,
			int32 inputFD){

   int32 xdim;            /* X-dimension */
   int32 ydim;            /* Y-dimension */
   int   ispal;           /* Non-zero if palette */
   int   n;               /* Local counter */
   int   color;           /* Color number */
   char  p;               /* Local pointer */
   uint8 palette[768];    /* Input palette */
   uint8 colormap[768];   /* Output colormap */
   UCHAR *image;          /* Actual image */
   char  outFilename[MY_FILENAME_MAX]; /* Output file name. */
   char  jpegFilename[MY_FILENAME_MAX]; /* Output file name. */
   char  hdfFilename[MY_FILENAME_MAX]; /* Output file name. */
   char  palFilename[MY_FILENAME_MAX]; /* Output file name. */
   intn  i;               /* local counting variable */

   uint8      *img24_buf;      /* buffer to store the 24-bit image in */
   comp_info   cinfo;              /* compression structure */

   while (n8Images--) {

      /* Output the reference. */
      sprintf (outFilename, "%s%d.gif", base, n8Images);
      sprintf (jpegFilename, "%s%d.jpg", base, n8Images);
      sprintf (hdfFilename, "%s%d.hdf", base, n8Images);
      sprintf (palFilename, "%s%d.pal.gif", base, n8Images);

      /* Runs on every Rastar 8 image. */
      if ( n8Images > 0 ) {
         /* printf(" Images number %d \n", n8Images); */
         ProcessMetadata (inputFilename, n8Images, base, inputFD);
      }


      /*    printf(" running DFR8getdims \n");
       Get the image dimensions. */
      if (DFR8getdims (inputFilename, &xdim, &ydim, &ispal) == -1) {
         printf ("X-Error: Process8BitImages: Could not get the image \
dimensions for %s\n", outFilename);
	 Hclose (inputFD);
         exit (1);
      }

      /* Read in the image */
      if ((image = (UCHAR *) malloc (xdim * ydim)) == NULL) {
	 Hclose (inputFD);
         printf ("X-Error: Process8BitImages: Insufficient memory to \
load image: %dkb\n", (xdim * ydim) / BYTES_PER_KB);
         exit (1);
      }
      if (DFR8getimage (inputFilename, image, xdim, ydim, palette) < 0) {
	 Hclose (inputFD);
	 free (image);
         printf ("X-Error: Process8BitImages: Failed to get 8-Bit \
image from file.\n");
         exit (1);
      }

      if (ispal) {
         uint8   *pal_ptr = palette;   /* temporary pointer into the palette */
         for (i = 0; i < 256; i++) {
             red_comp[i] = *pal_ptr++;
             green_comp[i] = *pal_ptr++;
             blue_comp[i] = *pal_ptr++;
         }   /* end for */
      }  /* end if */
      else
      {     /* no palette, use a greyscale palette */
         for (i = 0; i < 256; i++)
              red_comp[i] = green_comp[i] = blue_comp[i] = (uint8) i;
      }     /* end else */


      if ((img24_buf = (uint8 *) HDmalloc((size_t)(xdim * ydim * 3))) == NULL)
      {
         printf("Error, cannot allocate space for %ldx%ld 24-bit image\n", xdim, ydim);
          exit (1);
      }     /* end if */

      /* convert the image from a 8 bit image to a 24 bit image */
      if (!convert8to24(image, img24_buf, (int) xdim, (int) ydim))
      {
          printf("Error converting 8-bit image to 24-bit image\n");
          exit (1);
      }     /* end if */


      /* For right now, SAVE the 24 bit image into a HDF */
      /* so that we can convert it into a JPEG Image */
      if (writeJPEG_HDF ( hdfFilename, img24_buf, (int) xdim, (int) ydim) == FALSE ) {
         free (img24_buf);
         Hclose (inputFD);
         printf ("X-Error: writeJPEG_HDF: Converter program error, Don't know why!\n");
         exit (1);
      }

      /* Open the 24 Bit JPEG hdf and save it AS a JPEG Image */
      /* Cumbersome , but it works */
      if ( hdf2jpg( hdfFilename, jpegFilename, 0,0) == FAIL ) {
         free (img24_buf);
         Hclose (inputFD);
         printf ("X-Error: DF24putimage: Converter program error, Don't know why!\n");
         exit (1);
      }

      /* Dummy up a grayscale palette if there wasn't one. */
      if (!ispal) {
         for (n = 0; n < 256; ++n) {
            palette[3*n] = palette[3*n+1] = palette[3*n+2] = n;
         }
      }
      /* Convert the palette into a colormap. */
      for (n = 0; n < 256; ++n) {
         colormap[n] = palette[3*n];
         colormap[n+256] = palette[3*n+1];
         colormap[n+512] = palette[3*n+2];
      }

      /* Create and write out the palette with legend */
      WritePalette (palFilename, (char *) colormap, n8Images, inputFD);

      /* Write out the image. */
      if (WriteGIF (outFilename, image, (int) xdim, (int) ydim,
                    (char *) colormap)) {
	 free (image);
	 Hclose (inputFD);
         printf ("X-Error: Process8BitImages: Converter program error,\
Don't know why!\n");
         exit (1);
      }
      /* Report the filename */
      printf ("Image-pathname: %s\n\n", outFilename);

      /* Free the image memory. */
      free (image);
   }
}

void ProcessMetadata (char *inputFilename, int n24Images,
                         char *base, int32 inputFD) {

   char  outFilename[MY_FILENAME_MAX]; /* Output file name. */
   int   isMetadata = FALSE;    /* Found any metadata? (AN or SD). */

      sprintf (outFilename, "%s%d.des", base, n24Images);

   if (outFilename == NULL) {
      printf ("X-Error: main: System could not allocate memory for \
metadata filename.\n");
      Hclose (inputFD);
      exit (1);
   }

   /* Get all file annotations from HDF into an output file.
      Returns TRUE if found any.  Returns FALSE if none found.*/
   isMetadata = GetSpecificAnnotations (inputFD, outFilename, n24Images );

   /* Get SD metadata if any, from HDF into the same output file.
      Appends if found. Returns TRUE if found; previous isMetadata
      value if none found. */
   isMetadata = GetSDMetadata (inputFilename, outFilename,isMetadata, inputFD);

   /* If there were any annotations or metadata, print the filename. */
   if (isMetadata) {
      printf ("Description-Filename: %s\n\n", outFilename);
   }
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Function: Process24BitImages
 * Description: Extract and convert the 24-bit images to GIF files.
 *    Note that we must first convert the images to 8-Bit.
 *    Prints the names of the GIF files.
 * Parameters: char  *inputFilename - HDF input filename.
 *             int   n24Images      - Number of 24-bit images in file.
 *             char  *base          - Output filename base.
 *             int32 inputFD        - Input HDF file descriptor.
 * Return: none.
 */
void Process24BitImages (char *inputFilename, int n24Images,
			 char *base, int32 inputFD, int process_metadata) {

   int32 xdim;            /* X-dimension */
   int32 ydim;            /* Y-dimension */
   int32 size;            /* Size of image = xdim * ydim */
   int   ispal;           /* Non-zero if palette */
   int   n;               /* Local counter */
   int   color;           /* Color number */
   char  p;               /* Local pointer */
   uint8 palette[768];    /* Input palette */
   uint8 colormap[768];   /* Output colormap */
   UCHAR *image24;        /* Actual 24-Bit image */
   UCHAR *image8;         /* Converted 8-Bit image */
   char  outFilename[MY_FILENAME_MAX]; /* Output file name. */
   char  jpegFilename[MY_FILENAME_MAX]; /* Output file name. */
   char  hdfFilename[MY_FILENAME_MAX]; /* Output file name. */
   char  baseFilename[MY_FILENAME_MAX]; /* Output file name. */
   int   interlace;       /* 24-bit image interlace (not used). */

   comp_info   cinfo;              /* compression structure */

      /* Output the reference. */
      sprintf (outFilename, "%s%d.gif", base, n24Images);
      sprintf (jpegFilename, "%s%d.jpg", base, n24Images);

      sprintf (hdfFilename, "%s%d.hdf", base, n24Images);
      sprintf (baseFilename, "%s.hdf", base);

      /* Get the image dimensions. */
      if (DF24getdims (inputFilename, &xdim, &ydim, &interlace) == -1){
	 Hclose (inputFD);
         printf ("X-Error: Process24BitImages: Could not get the \
image dimensions for %s\n", outFilename);
         exit (1);
      }
      size = xdim * ydim;
      /* Allocate memory for image. Note factor of 3 for 3 colors. */
      if ((image24 = (UCHAR *) malloc (size * 3)) == NULL) {
	 Hclose (inputFD);
         printf ("X-Error: Process24BitImages: Insufficient memory to \
load image: %dkb\n", (size * 3) / BYTES_PER_KB);
         exit (1);
      }
      /* Read in the image. */
      if (DF24getimage (inputFilename, image24, xdim, ydim) < 0) {
	 free (image24);
	 Hclose (inputFD);
         printf ("X-Error: Process24BitImages: Failed to get 24-Bit \
image from file.\n");
         exit (1);
      }

      /* Open the 24 Bit JPEG hdf and save it AS a JPEG Image */
      /* Cumbersome , but it works */
      if ( hdf2jpg( inputFilename, jpegFilename,1,process_metadata) == FAIL ) { 
         free (image24);
         Hclose (inputFD);
         printf ("X-Error: DF24putimage: Converter program error, Don't know why!\n");
         exit (1);
      }

      /* Free the image memory. */
      free (image24);
      free (image8);
}

/**********************************************************************
 * Function: r24r8()
 * Description: converts 24-Bit raster image to 8-Bit raster image.
 * Taken from hdf24hdf8 program from NCSA.  I didn't modify this code
 * or add any inline comments so don't blame me!
 *
 * NCSA HDF
 * Software Development Group
 * National Center for Supercomputing Applications
 * University of Illinois at Urbana-Champaign
 * 605 E. Springfield, Champaign IL 61820
 *********************************************************************/
int r24r8 (int32 xres, int32 yres, UCHAR* dat24, UCHAR* dat8, 
           int cres, UCHAR *cdat) {

    int         ct, xct, yct;
    int         rres, rd, rr, rn, rct;
    int         gres, gd, gr, gn, gct;
    int         bres, bd, br, bn, bct;
    int         coff;
    UINT       *idat[2];
    UINT       *cp, *np;
    UCHAR      *dip, *dop, *rp, *gp, *bp;

    if ((idat[0] = (UINT *) HDmalloc ((size_t)(6 * xres) * 
                                      sizeof(UINT))) == NULL)
      {
          fprintf (stderr, "error: Memory allocation fault\n");
          return -1;
      }
    idat[1] = idat[0] + (3 * xres);
    rres = 6;
    gres = 7;
    bres = 6;
    coff = 2;
    rr = gr = br = 255;
    rn = rres - 1;
    gn = gres - 1;
    bn = bres - 1;
    rp = cdat + coff;
    gp = rp + cres;
    bp = gp + cres;

    for (rct = 0; rct < rres; rct++)
      {
          for (gct = 0; gct < gres; gct++)
            {
                for (bct = 0; bct < bres; bct++)
                  {
                      *rp++ = (UCHAR) (rr * rct / rn);
                      *gp++ = (UCHAR) (gr * gct / gn);
                      *bp++ = (UCHAR) (br * bct / bn);
                  }
            }
      }

    rp = cdat;
    gp = rp + cres;
    bp = gp + cres;
    cp = idat[0];
    np = idat[1];
    dip = dat24;
    dop = dat8;

    for (xct = (intn) (3 * xres); --xct >= 0;)
        *cp++ = (UINT)*dip++;

    for (yct = 0; yct < (yres - 1); yct++)
      {
          np = idat[(yct + 1) % 2];
          for (xct = (intn) (3 * xres); --xct >= 0;)
              *np++ = (UINT)*dip++;

          cp = idat[yct % 2];
          np = idat[(yct + 1) % 2];

          if ((rct = ((int)cp[0] * rn / rr)) > rn)
              rct = rn;
          if ((gct = ((int)cp[1] * gn / gr)) > gn)
              gct = gn;
          if ((bct = ((int)cp[2] * bn / br)) > bn)
              bct = bn;

          *dop++ = (UCHAR) (ct = (rct * gres + gct) * bres + bct +
			    coff);

          rd = (int)cp[0] - rp[ct];
          gd = (int)cp[1] - gp[ct];
          bd = (int)cp[2] - bp[ct];
          cp += 3;
          np += 3;
          cp[0] += (UINT)(rd * 7 / 16);
          cp[1] += (UINT)(gd * 7 / 16);
          cp[2] += (UINT)(bd * 7 / 16);
          np[-3] += (UINT)(rd * 5 / 16);
          np[-2] += (UINT)(gd * 5 / 16);
          np[-1] += (UINT)(bd * 5 / 16);
          np[0] += (UINT)(rd / 16);
          np[1] += (UINT)(gd / 16);
          np[2] += (UINT)(bd / 16);

          for (xct = 2; xct < xres; xct++)
            {
                if ((rct = ((int)cp[0] * rn / rr)) > rn)
                    rct = rn;
                if ((gct = ((int)cp[1] * gn / gr)) > gn)
                    gct = gn;
                if ((bct = ((int)cp[2] * bn / br)) > bn)
                    bct = bn;

                *dop++ = (UCHAR) (ct = (rct * gres + gct) * bres + 
				  bct + coff);

                rd = (int)cp[0] - rp[ct];
                gd = (int)cp[1] - gp[ct];
                bd = (int)cp[2] - bp[ct];
                cp += 3;
                np += 3;
                cp[0] += (UINT)(rd * 7 / 16);
                cp[1] += (UINT)(gd * 7 / 16);
                cp[2] += (UINT)(bd * 7 / 16);
                np[-6] += (UINT)(rd * 3 / 16);
                np[-5] += (UINT)(gd * 3 / 16);
                np[-4] += (UINT)(bd * 3 / 16);
                np[-3] += (UINT)(rd * 5 / 16);
                np[-2] += (UINT)(gd * 5 / 16);
                np[-1] += (UINT)(bd * 5 / 16);
                np[0] += (UINT)(rd / 16);
                np[1] += (UINT)(gd / 16);
                np[2] += (UINT)(bd / 16);

            }

          if ((rct = ((int)cp[0] * rn / rr)) > rn)
              rct = rn;
          if ((gct = ((int)cp[1] * gn / gr)) > gn)
              gct = gn;
          if ((bct = ((int)cp[2] * bn / br)) > bn)
              bct = bn;

          *dop++ = (UCHAR) (ct = (rct * gres + gct) * bres + bct + 
			    coff);

          rd = (int)cp[0] - rp[ct];
          gd = (int)cp[1] - gp[ct];
          bd = (int)cp[2] - bp[ct];
          cp += 3;
          np += 3;
          np[-6] += (UINT)(rd * 3 / 16);
          np[-5] += (UINT)(gd * 3 / 16);
          np[-4] += (UINT)(bd * 3 / 16);
          np[-3] += (UINT)(rd * 5 / 16);
          np[-2] += (UINT)(gd * 5 / 16);
          np[-1] += (UINT)(bd * 5 / 16);
      }

    cp = idat[yct % 2];

    if ((rct = ((int)cp[0] * rn / rr)) > rn)
        rct = rn;
    if ((gct = ((int)cp[1] * gn / gr)) > gn)
        gct = gn;
    if ((bct = ((int)cp[2] * bn / br)) > bn)
        bct = bn;

    *dop++ = (UCHAR) (ct = (rct * gres + gct) * bres + bct + coff);

    rd = (int)cp[0] - rp[ct];
    gd = (int)cp[1] - gp[ct];
    bd = (int)cp[2] - bp[ct];
    cp += 3;
    cp[0] += (UINT)(rd * 7 / 16);
    cp[1] += (UINT)(gd * 7 / 16);
    cp[2] += (UINT)(bd * 7 / 16);

    for (xct = 2; xct < xres; xct++)
      {
          if ((rct = ((int)cp[0] * rn / rr)) > rn)
              rct = rn;
          if ((gct = ((int)cp[1] * gn / gr)) > gn)
              gct = gn;
          if ((bct = ((int)cp[2] * bn / br)) > bn)
              bct = bn;

          *dop++ = (UCHAR) (ct = (rct * gres + gct) * bres + bct + 
			    coff);

          rd = (int)cp[0] - rp[ct];
          gd = (int)cp[1] - gp[ct];
          bd = (int)cp[2] - bp[ct];
          cp += 3;
          cp[0] += (UINT)(rd * 7 / 16);
          cp[1] += (UINT)(gd * 7 / 16);
          cp[2] += (UINT)(bd * 7 / 16);
      }

    if ((rct = ((int)cp[0] * rn / rr)) > rn)
        rct = rn;
    if ((gct = ((int)cp[1] * gn / gr)) > gn)
        gct = gn;
    if ((bct = ((int)cp[2] * bn / br)) > bn)
        bct = bn;

    *dop++ = (UCHAR) ((rct * gres + gct) * bres + bct + coff);

    HDfree (idat[0]);
    return (0);
}


/**********************************************************************
*  Function :   convert8to24
*  Purpose  :   Convert an 8-bit image to a 24-bit image, using the
*                   palette components already set up.
*  Parameters   :
*           img_8_buf - pointer to the buffer which contains the 8-bit image
*           img24_buf - pointer to the buffer in which to place the
*                           24-bit image
*           img_xdim, img_ydim - the width and height of the images
*  Returns  :   (TRUE) for success, (FALSE) for error
*  Calls    :
*  Called by    :
**********************************************************************/
static intn
convert8to24(uint8 *img8_buf, uint8 *img24_buf, int32 img_xdim,
             int32 img_ydim)
{
    uint32      pixels;         /* local counting variable */


    if (img_xdim <= 0 || img_ydim <= 0)     /* check for bad image dimensions */
        return (FALSE);
    if (img8_buf == NULL || img24_buf == NULL)  /* check for invalid images */
        return (FALSE);

    /* get the number of pixels to process */
    pixels = (uint32)(img_xdim * img_ydim);   
    while (pixels > 0)
      {     /* do all the pixels */
          *img24_buf++ = red_comp[*img8_buf];
          *img24_buf++ = green_comp[*img8_buf];
          *img24_buf++ = blue_comp[*img8_buf];
          img8_buf++;
          pixels--;
      }     /* end while */
    return (TRUE);
}   /* end convert8to24() */

intn
writeJPEG_HDF (char  *hdfFilename, uint8 *img24_buf, int32  xdim, int32 ydim) {

    comp_info   cinfo;              /* compression structure */

    cinfo.jpeg.quality = JPEG_COMPRESSION_QUALITY;
    cinfo.jpeg.force_baseline = TRUE;
    DF24setcompress(COMP_JPEG, &cinfo);   /* set compression parameters */

    if (DF24putimage( hdfFilename, img24_buf, (int) xdim, (int) ydim) == FAIL ) {
/*     free (img24_buf);
       Hclose (inputFD); */
       printf ("X-Error: DF24putimage: Converter program error, Don't know why!\n");
       exit (1);
    }
    return (TRUE);
}

/* END OF FILE */
