/************************************************************
 * The following documentation is in POD format...
  

=head1 NAME

imsclient - standalone IMS client program



=head1 DESCRIPTION

This is a standalone program that can take a correctly formatted IMS
ODL query tree and executed a search to the DAACs.  This version is 
customized for use with the IMS/WWW gateway.  It will currently support
all major search types.


=head1 SYNOPSIS

Invoke this with no arguments to get the usage.  The forms are:

    imsclient <action> <infile>           pre-1.6; deprecated
    imsclient [options]                   1.6+

where the options are...

=over

=item -aACTION

Preferred form of the I<action> argument above.

    ds     DIRECTORY_SEARCH (introduced in 1.6)
    is     INVENTORY_SEARCH          
    fb     BROWSE_REQUEST (FTP)
    ib     BROWSE_REQUEST (Integrated)
    pr     PRODUCT_REQUEST           
    os     PRODUCT_STATUS_REQUEST
    as     ACCOUNT_STATUS_REQUEST
    rr     PROFILE_RETRIEVAL_REQUEST
    ur     PROFILE_UPDATE_REQUEST
    er     PRICE_ESTIMATE_REQUEST
    sr     PROFILE_SUBMIT_REQUEST
    cr     USER_PASSWORD_CHANGE_REQUEST

=item -gYESNO

Perform GCMD searches?  The default is yes.  
C<-gy> turns this on, C<-gn> turns it off.  
Really only used with C<-ads>.

=item -iINFILE  

The F<infile.odl> contains the odl tree. 

=item -mMODE    

Set the mode of operation.  Default is C<g>.
Possibilities are:

    n      Normal (default).  Env. vars are not touched.
    d      Developer test.  Environment variables
           are set up to point to the cwd and to the standard
           valids directory, and logging is set to max.

=item -v

Verbose (or, if you like, "vocal").  Debugs are output to stderr.  
However, the logging level is I<not> currently changed.

=back



=head1 OVERVIEW

    1. Based on the desired action (from the command line), the input
       file is parsed in one of a number of possible ways.  Each action
       has its own setup routine.

    2. The dataset/granule databases are not deleted; however, all the
       current entries are flagged as "invisible".

    3. The search commences.  This repeatedly sleeps for 2 seconds and
       invokes the poller to get the updated statuses.   Searching
       stops when the search is done, or when a client.stop file is placed
       in the directory.  Else, FOR EACH POLL, we do this:

         3.1.  FOR EACH CONNECTION: information relevant to the
               connection (e.g., num datasets/granules received) 
               is gathered.  This is written out to a temp status file.

         3.2.  The file is closed and renamed to the real file, client.stat.
               This ensures that updates to client.stat are atomic.



=head1 FILES CREATED

=head2 The "client.stat" file

During the search, the current status of each connection is output
to the special F<client.stat> file every time a state transition
is caused (about every 2 seconds).  It looks like this:

    X-TYPE: COMMSTATUS
    DATA_CENTER_ID: MSFC
    LED_STAGE: 4
    NUMBER_OF_DATASETS: 1
    NUMBER_OF_GRANULES: 10
    STATUS_CODE: 19
    STATUS_CODE_COMMENT: Success; results being returned.
   
    X-TYPE: COMMSTATUS
    DATA_CENTER_ID: GSFC
    LED_STAGE: 2
    NUMBER_OF_DATASETS: 4
    NUMBER_OF_GRANULES: 60
    STATUS_CODE: 19
    STATUS_CODE_COMMENT: Still going...
    ...

To ensure that this file is updated atomically, each new version is
written to a temp file, F<client.stat.new>.  When the new file is
complete, it is closed and renamed to F<client.stat>.


=head2 The "client.obj" file

As of IMS/WWW 1.6, we try to extract information directly from
the ODL of an INVENTORY SEARCH, instead of relying on complex ODL
extensions being implemented down at the IK level.  Thus, we output
information in an imsgdbdump-style format to the F<client.obj> file,
like this:

     X-Type: DATASET
     DATASET_WEB_KEY: 1/23
     DATASET_COMMENT: Dude! This is a gnarly mechanism
        : for extending the functionality of the system.  Basically,
        : we grab info from as many different sources as we need,
        : jam it all together, and that's our object.
     BEGIN DATA_URL
        URL: http://some.host/datasetname
        URL_COMMENT: Here's the documentation!
     END 
      
     X-Type: DATASET
     DATASET_WEB_KEY: 1/24
     ...


=head2 The "client.list" file

As of IMS/WWW 1.6, during searches, the existing F<dataset.gdb> and
F<granule.gdb> files are B<not> deleted.  This is to allow users to continue
to browse granules obtained from previous searches.  

Although it is not needed by the system as of IMS/WWW 1.7, the 
special F<client.list> file is still output: this file contains a 
simple list of all datasets/granules returned by the B<most recent
directory/inventory search.>  Like this:

     DATASET_WEB_KEY: 1/23
     DATASET_WEB_KEY: 2/345
     DATASET_WEB_KEY: 2/357
     GRANULE_WEB_KEY: 1/23/ABCD001
     GRANULE_WEB_KEY: 1/23/ABCD002
     ...




=head1 NOTES

=head2 Status processing

The basic algorithm is this, for each tree that comes in:

     ### Status code comment:
     IF there is a STATUS_CODE_COMMENT
         save that value in the connection-info
     ELSE (there is *no* STATUS_CODE_COMMENT)
         IF the STATUS_CODE has changed
             blank out the STATUS_CODE_COMMENT in the connection info
         ELSE
             leave the old STATUS_CODE_COMMENT value
         ENDIF
     ENDIF
      
     ### Status code:
     Save the new STATUS_CODE in the connection info:

The idea is to prefer explicit status code comments over implicit ones,
and to never overwrite a status code comment with an implicit comment
unless the status code has changed.


=head2 Miscellaneous

=over

=item *

This program assumes the following environment variable has  
been set to the correct value, otherwise, the program will   
use default value which may cause the program to crash or    
cause files to be wrriten in the incorrect directory.        
Env. Variables:                                              
 
    GAEATMP_DIR:  contains the path to the temp user data    
    GAEADATA_DIR: contains the path to valids and other      
                  support files, very important              
    USRDATA_DIR:  not used currently                          
    SYSLOG_LEVEL: how detailed the syslog should be, default is ok. 
 
=item *
                                                                  
In order to do a integrated browse, we need to first do a    
inventory search, because the integrated browse needs the     
information from the dataset.gdb.

=item *

When gathering connection statuses there's a nasty problem:
it is possible for a badly-behaved DAAC "XYZ" to send back 
granules with archive id "PDQ".  Yuck. 

=item *

The term "status" is WAY overloaded in IK.  We define...

    STATUS_CODE            numeric code in the ODL result tree.
                           one of the NM_XXX values, from IK_Nmsts.h
    STATUS_CODE_COMMENT    string in the ODL result tree
    LED_STAGE              numeric code meant for LED stage in Gateway

=back



=head1 RCS INFO

$Id: imsclient.c,v 1.56 2002/05/08 13:54:45 solomon Exp $



=head1 AUTHORS

The design is based on a test driver program created  
by S. Sudarsean and modifed by B. Kerbel. The first operational
version for IMS/WWW (July 95) was written by Sze Hong.       
                                                              
Mark Solomon, Jim Amrhein, and Eryq have modified it 
for the 1.5 and 1.6 releases (Summer/Fall 96).

Eryq gutted the object-output code for the 1.7 release (Spring 97),
and modified it to use the new IK message hook function.

=cut



************************************************************/

/* IK include file */
#include "Gaea_config.h"
/* System files include files */
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <unistd.h>
#include <sys/file.h>  /* for flock */
#include <stdarg.h>
#include <time.h>
#include <signal.h>


/* IK and other public include files */
#include "IK_Ims.h"
#include "IK_DataDic.h"
#include "IK_Errno.h"
#include "IK_Syslog.h"
#include "IK_Granule.h"
#include "IK_Msg.h"
#include "IK_InDsetStatCodeCmt.h"

#include "gdbm.h"
#include "odldef.h"
#include "odlinter.h"

#include "IK_Nm.h"
#include "IK_Nmpro.h"
#include "IK_Session.h"
#include "IK_Ingest.h"

/* The following include was added for the new valids format */
#include "IK_NewValids.h"

#include "IK_Network.h"

#include "IW_Array.h"
#include "IW_Chunk.h"
#include "IW_Conn.h"
#include "IW_Hasht.h"
#include "IW_Key.h"
#include "IW_New.h"
#include "IW_ODL.h"
#include "IW_Str.h"

/* Get definitions for debug, info, etc.: */
#define IW_Log_NICENAMES 1
#include "IW_Log.h"



/************************************************************
 *
 * Constants...
 *
 ************************************************************/

/* Some files... */
#define SYSLOG_FILE "syslog"
#define G_GDB_FILE  "granule.gdb"
#define D_GDB_FILE  "dataset.gdb"
#define ARCHIVE_ODL "archive.odl"
#define DYNAMIC_FILE "dynamic.odl" /* name of the dynamic valids file */

#define PATH_LEN   256

/* added by Mark Solomon */
#define MAX_COMMENT (2048) /* Maximum size of comment */
#define MAX_PUKE (2048) /* Maximum size of puke message */
#define MAXLEN (81) /* the maximum number of bytes in a valid + 1 */
#define MAX_CORE_FIELD (7) /* the maximum number of core attributes */
/* Modes of operation: */
#define MODE_NULL      0   /* NONE */
#define MODE_NORMAL    1   /* standard mode for IMS/WWW gateway */
#define MODE_DEVELOPER 2   /* IMS/developer testing mode */

/* Actions to perform: */
#define ACTION_NULL 0              /* NONE */
#define ACTION_DIR_SEARCH   1      /* directory search */
#define ACTION_INV_SEARCH   2      /* inventory search */
#define ACTION_INT_BROWSE   3      /* integrated browse request */
#define ACTION_FTP_BROWSE   4      /* FTP browse request */
#define ACTION_PROD_REQUEST 5      /* product request */
#define ACTION_PROD_STAT_REQUEST 6 /* product status request */
#define ACTION_ACCT_STAT_REQUEST 7 /* account status request */
#define ACTION_PROF_RETRIEVE_REQUEST 8 /* profile retrieval request */
#define ACTION_PROF_UPDATE_REQUEST 9   /* profile update request */
#define ACTION_PRICE_ESTIMATE_REQUEST 10  /* price estimate request */
#define ACTION_PROF_SUBMIT_REQUEST 11  /* profile submit request */
#define ACTION_PASSWORD_CHANGE_REQUEST 12  /* user password change request */

#define ACTION_NUM 13 /* Total number of actions available */ 
                      /* This should be one more than the last action */
/* This is a safety feature added by Eryq for the IMS/WWW gateway,
 * to minimize risks.  Searches are automatically aborted after 
 * this many minutes.  IN THE FUTURE, this should be read from
 * the SETUP.dat file.
 */
#define MAX_RUNTIME_MINS 60

/* Patch a bug that causes IK to crash? */
#define IMSCLIENT_PATCH_UNMAPPED_FIELD_BUG 1

/* Little macro: */
#define MAX_NUM(a,b) ((a) > (b) ? (a) : (b))



/************************************************************
 * 
 * Declarations and prototypes...
 *
 ************************************************************/

/* Variables... */
extern GDBM_FILE dep_valids_dbf, tag_dbf;
extern char **fieldname; /* names of the fields */
static int ValidsCount; /* number of valids in the list */
static int totalDaacCount; /* number of DAACS in archive.odl */
/* External functions... */
static AGGREGATE read_label(AGGREGATE,char*);

/* Internal functions... */
static void   usage();
static void   puke(char *, ...);
static void   get_opts(int, char **);
AGGREGATE     lookup_data_center (char *);
static int    count_data_centers();
static int    invoke_poller(int);

static int    setup_inv_search  (AGGREGATE, AGGREGATE *);
static int    setup_dir_search  (AGGREGATE, AGGREGATE *);
static int    setup_browse      (AGGREGATE, AGGREGATE *);
static int    setup_prod_request(AGGREGATE, AGGREGATE *);
static int    setup_prod_stat_request(AGGREGATE, AGGREGATE *);
static int    setup_acct_stat_request(AGGREGATE, AGGREGATE *);
static int    setup_prof_retrieve_request(AGGREGATE, AGGREGATE *);
static int    setup_prof_update_request(AGGREGATE, AGGREGATE *);
static int    setup_price_estimate_request(AGGREGATE, AGGREGATE *);
static int    setup_prof_submit_request(AGGREGATE, AGGREGATE *);
static int    setup_password_change_request(AGGREGATE, AGGREGATE *);

static IW_Conn *lookup_conn(char *);
static void   reply_received(CONNECTION_INFO *, AGGREGATE, int);
static int    report_tree(AGGREGATE, IW_Chunksp *);
int handle_value (char ***, AGGREGATE, int *, char *);


/************************************************************
 *
 * Global variables...
 *
 ************************************************************/

/* RCS id... */
static char *RCSID = 
     "$Id: imsclient.c,v 1.56 2002/05/08 13:54:45 solomon Exp $";

char GaeaTmpDir[PATH_LEN];        /* Pathname of gaea temp directory */
char GaeaDataDir[PATH_LEN];       /* Pathname of gaea data directory */
char GranuleGdbPath[PATH_LEN];    /* Full name for granule.gdb */
char SyslogPath[PATH_LEN];        /* Full name for the syslog file */
char ArchivePath[PATH_LEN];       /* Full path for the archive.odl */
char StopPath[PATH_LEN];         /* Full path for client abort file */
char dynamic_file[PATH_LEN];     /* name of dynamic valids file to open */

int DebugLevel; 
int inv_search = 0;

/* IK-specific errno: */
IK_Errno IK_ImsErrno;   

/* Mode of operation: */
static int Mode = MODE_NULL;

/* Action to perform: */
static int Action = ACTION_NULL;

/* Input ODL file: */
static char Infile[PATH_LEN];

/* Are we just browsing? */
static int BrowseFlag = 0;        /* DO NOT SET THIS TRUE, EVER! [eryq] */ 

/* Clear database files from last activity? */
static int ClearLast = 0;

/* Should we output debug info to stderr, for interactive debugging? */
static int Verbose = 0;

/* Do GCMD directory search? */
static int GCMDSearch = 1;

/* List file, containing what we got from the last search: */
static char ListPath[PATH_LEN];    /* where last reported search results go */
FILE *ListFp;                      /* where we write them */

/* Object file, containing chunks of additional information: */
static char ObjPath[PATH_LEN];    /* where last reported search results go */
FILE *ObjFp;                      /* where we write them */

/* Object file, containing chunks of additional information: */
static char OdlPath[PATH_LEN];    /* where last reported search results go */
FILE *OdlFp;                      /* where we write them */

/* Map actions to ids: */
static struct {
  char *id;
  char *name;
  int (*setup)();
} ActionInfo[] = {
  {"", "", NULL},
  {"ds", "directory search",  setup_dir_search},
  {"is", "inventory search",  setup_inv_search},
  {"ib", "integrated browse", setup_browse},
  {"fb", "FTP browse",        setup_browse},
  {"pr", "product request",   setup_prod_request},
  {"os", "product status request",   setup_prod_stat_request},
  {"as", "account status request",   setup_acct_stat_request},
  {"rr", "profile retrieval request", setup_prof_retrieve_request},
  {"ur", "profile update request",   setup_prof_update_request},
  {"er", "price estimate request",   setup_price_estimate_request},
  {"sr", "profile submit request",   setup_prof_submit_request},
  {"cr", "user password change request",   setup_password_change_request}
};

/* mySearchType is based on the input action */
static char mySearchType = NULL;

/* searchArray is the one character type of action */
static char searchArray[ACTION_NUM] = 
{NULL, 'd', 'i', 'b', 'b', 'p', 'o', 'a', 'r', 'u', 'e', 's', 'c'};

/* Callback-gathered info, by DAAC.
 * This is a hashtable of Conn structs, hashed (case-insensitively)
 * by DATA_CENTER_ID. 
 */
static IW_Hasht *Conns;

/* Should we be queueing messages? */
static int QueueMessages = 1;

/* Queue of trees to be processed: */
static IW_Array *MsgQueue;

/* The "archive.odl": */
static AGGREGATE ArchiveAgg;

char *ptr; /* a tmp pointer used to upcase the field name */

static char **table; /* table to hold dependent valids */

/************************************************************
 *
 * Functions related to logging, usage, args, etc.
 *
 ************************************************************/

/************************************************************
 * main
 ************************************************************
 *
 * DESCRIPTION 
 *    This is the main function of the program.  It will first
 *    check to if the arguments supply by the user is valid or not. 
 *
 *    If it is a valid inventory search, it will use the dependent valids 
 *    information to add the correct archive information into the odl tree.
 *
 *    If it is a valid integrated browse search, it will carry out the search 
 *    and put the results in the directory specified in the environment 
 *    variables.
 *
 * PARAMETERS
 *    Standard argc and argv.
 *
 * RETURN VALUES
 *    Return 0 upon success and nonzero on failure.`
 */

int main(int argc, char **argv, char **envp)
{
  int total_conn;
  int num, ret;           /* from CauseStateTransition */
  FILE *fp;
  PARAMETER temp_para, tempdatacenter;
  VALUE temp_val;
  AGGREGATE temp_agg, pr_agg, data_center_agg;
  int conn;
  int (*setup)() = NULL;
  AGGREGATE root;     /* root for the odl search tree */
  AGGREGATE dynamic_root; /* root AGGREGATE of the dynamic valids file */
  AGGREGATE *toproot; /* array to hold order trees from different daacs.
			    it is also used to hold the regular odl search 
			    tree */
  static char *required_env_vars[] = {
    "GAEADATA_DIR",
    "GAEATMP_DIR",
    "SERVHOSTADDR",
    "SERVTCPPORT",
    NULL,
  };
  char **sp;
  time_t start_time;
  time_t now_time;
  double run_d;
  char **fields; /* requested valids of a field */
  int j; /* counter for valids found in the input odl */
  int k; /* counter for fields */
  int i; /* counter */
  IK_BITVECTOR *fieldValidsVector = NULL; /* vector of selected valids */
  int numHits; /* number of valids selected by IK_FindValids */
  int found; /* return code for IK_FindValids */
  int dyIndex; /* the index of the new dynamic valid */

  /******************************
   * SETUP DATA STRUCTURES, GET COMMAND LINE OPTIONS...
   ******************************/

  /* Get command-line options: */
  get_opts(argc, argv);

  /* Tweak settings, according to mode: */
  switch (Mode) {
  case MODE_DEVELOPER:
    fprintf(stderr, "WARNING: Missing environment vars will be filled in!\n");
    Verbose = 1;
    putenv("SYSLOG_LEVEL=9");
    putenv("IMS_STAFF=1");
    if (!getenv("GAEATMP_DIR"))  { putenv("GAEATMP_DIR=."); }
    if (!getenv("USRDATA_DIR"))  { putenv("USRDATA_DIR=."); }
    if (!getenv("SERVHOSTADDR")) { putenv("SERVHOSTADDR=192.107.190.77"); }
    if (!getenv("SERVTCPPORT"))  { putenv("SERVTCPPORT=5502"); }
    break;
  case MODE_NORMAL:           /* no-op */
  default:
    break;    
  }
  putenv("IMS_LOG_TX=1");
  putenv("IMS_LOG_RX=1");

  /* Check environment: */
  for (sp = required_env_vars; *sp; sp++) {
    if (!getenv(*sp)) { 
      puke("please set %s in your environment", *sp); 
    }
  }

  /* Initialize the major directories: */
  strcpy(GaeaDataDir, IK_SysConfig("GaeaData"));
  strcpy(GaeaTmpDir,  IK_SysConfig("GaeaTmp"));

  /* Patch an IK bug which causes the temp dir to end in "//"; make it
   * end in "/" instead: */
  if (strstr(GaeaTmpDir, "//")) {
    GaeaTmpDir[strlen(GaeaTmpDir)-1] = '\0';     /* chop */
  }
  if (strstr(GaeaDataDir, "//")) {
    GaeaDataDir[strlen(GaeaDataDir)-1] = '\0';     /* chop */
  }

  /* Init all the paths: */
  sprintf(ArchivePath,   "%s%s", GaeaDataDir, ARCHIVE_ODL);
  sprintf(SyslogPath,    "%s%s", GaeaTmpDir, SYSLOG_FILE);
  sprintf(GranuleGdbPath,"%s%s", GaeaTmpDir, G_GDB_FILE);
  sprintf(dynamic_file,  "%s%s", GaeaTmpDir, DYNAMIC_FILE);
  sprintf(StopPath,      "%s%s", GaeaTmpDir, "client.stop");
  sprintf(ListPath,      "%s%s", GaeaTmpDir, "client.list");
  sprintf(ObjPath,       "%s%s", GaeaTmpDir, "client.obj");
  sprintf(OdlPath,       "%s%s", GaeaTmpDir, "client.odl");

  /******************************
   * OPEN LOGS AND OTHER FILES, AND SETUP IK STUFF...
   ******************************/

  /* Open the syslog file, and output some useful info to it: */
  IK_NameSyslog(SyslogPath); 
  IW_LogSetVerbosity(Verbose ? LOG_DEBUG : LOG_WARNING);

  /* Log our starting environment: */
  for (sp = envp; *sp; sp++) {        /* the environment... */
    debug("ENVIRON: %s", *sp);
  }
  debug("GranuleGdbPath = %s", GranuleGdbPath);
  info("About to do %s (%s).", ActionInfo[Action].id, ActionInfo[Action].name);
  info("Input file is %s.", Infile);
  mySearchType = searchArray[Action];

  /* Initialize the session data. Update for 5.0.3: */
  if (IK_InitSessionData() == -1) {
    puke("main(): Error in IK_InitSessionData().");
  }

  /* Initialize the dependent valids. 
     IK_DepValidsInit (); old valids format.
     The following call was added for the new dependent valids code -Solomon 
     */
  if (IK_ReadDepValids() == -1) {
    IK_ImsErrno = IK_EIO;
    err("Unable to read Valids files. See support staff.");
  } /* end of new dependent valids code */
  ValidsCount = IK_ValidsCnt();

  /* Allocate memory for dependent valids table to maximum number of valids */
  table = (char **)calloc(ValidsCount, sizeof (char *));
  /* Load the "archive.odl": */
  ArchiveAgg = NewAggregate(NULL, KA_GROUP, "ARCHIVE", NULL);
  if (!IW_ODLReadFile(ArchivePath, ArchiveAgg)) {
    puke("couldn't read archive.odl!");
  }
  totalDaacCount = count_data_centers();
  toproot = (AGGREGATE *) calloc(totalDaacCount+1, sizeof (AGGREGATE));

  /* Create hash of connections: */
  if (!(Conns = IW_HashtAllocCI(totalDaacCount+1, NULL))) {
    puke("Yow! Couldn't allocate hashtable!");
  }

  /* Create queue of result trees: */
  if (!(MsgQueue = IW_ArrayAlloc(totalDaacCount, 
       ((void (*)())RemoveAggregate)))) {
    puke("Yow! Couldn't allocate tree queue!");
  }

  /* If desired, for GCMD searches to fail immediately: */
  if (!GCMDSearch) {
    warning("Forcing GCMD searches off!");
    putenv("SERVHOSTADDR=127.0.0.1");
    putenv("SERVTCPPORT=0"); 
  }




  /******************************
   * START THE REQUEST...
   ******************************/

  /* Open dataset.gdb and granule.gdb files (do this before setup):  */
  if (IK_DirInit(GaeaTmpDir) != 0) {
    puke("IK_DirInit error [%d]\t", IK_ImsErrno);
  }
  if (IK_GranuleInit(GranuleGdbPath, ClearLast) != 0) {
    puke("IK_GranuleInit error [%d]\t", IK_ImsErrno);
  }

  /* Create the root to hold the search tree: */
  root = NewAggregate(NULL,KA_GROUP,"ROOT",NULL); 

  /* Create the root to hold the dynamic valids file */
  dynamic_root = NewAggregate(NULL,KA_GROUP,"ROOT",NULL);

  /* Open the object file: */  
  if (!(ObjFp = fopen(ObjPath, "w"))) {
    puke("couldn't open object file %s", ObjPath);
  }

  /* Open the odl file: */  
#if OUTPUT_CLIENT_ODL
  if (!(OdlFp = fopen(OdlPath, "w"))) {
    puke("couldn't open odlect file %s", OdlPath);
  }
#endif

  /* Setup for the intended action: */
  if (setup = ActionInfo[Action].setup) {
    if ((*setup)(root, toproot) < 0) {
      puke("couldn't set up for %s", ActionInfo[Action].name);
    }
  }
/* This code is added for dynamic valids */
  /* Read in the label: */
  if ((IW_ODLReadFile(dynamic_file, dynamic_root))) {
    fields = (char **)calloc(ValidsCount, sizeof (char *));
    fieldValidsVector = IK_callocBV();       /* allocates bit vector */
  
    for (k=0; k < MAX_CORE_FIELD; k++) {
      j = 0;
      if (handle_value (&fields, dynamic_root->first_child, &j, fieldname[k])){
        for (i=0; i < j; i++) {
	  for (ptr = fields[i];*ptr;ptr++)       /* upcase the valid name */
	     *ptr = toupper (*ptr);
          fprintf (stderr, "item in fields %s \n", fields[i]);
  	  found = IK_FindValids(fields[i],k,fieldValidsVector,&numHits);
          if(found < 0) {
  	    if(IK_dynamicSlotAvailable()) {
    	      dyIndex = IK_addDynamicValid(fields[i],k);
    	      IK_SelectValid(dyIndex);
  	    }
	  }
	  free (fields[i]);
	  fields[i] = NULL;
	}
      }
    }
    RemoveAggregate(dynamic_root);
  }
  /* end of code for dynamic valids */

  IK_CreateQuitTree(&IK_QuitTree);
  IK_CreateAbortTree(&IK_AbortTree);
  IK_CreateAckTree(&IK_AckTree);

  /* Install the callback: */
  IK_SetODLRxCallback(reply_received);

  /* Create the search connections [Q: IS THIS RIGHT????]: */
  num = IK_GetNMSlots();
  conn = 0; 
  total_conn = 0;
  while (toproot[conn]) {

    /* WARNING: ONLY WRITE THE FINAL LABEL TO THE SYSLOG/STDERR IF DEBUGGING
     * INTERACTIVELY... IT CAN CONTAIN USER-SENSITIVE INFORMATION, LIKE
     * DATA ACCESS KEYS!  THAT MEANS THAT THIS INFO WILL GET EMAILED
     * OUT IN A DYING GASP!
     *
     * DO NOT UNDER *ANY* CIRCUMSTANCES ALTER THIS, EVEN IF IT'S KNOWN THAT
     * DEBUGGING IS TURNED ON AT THE SITE!  THE RISK OF LETTING OUT THE
     * PASSWORD OF A CUSTOMER TESTING THE SYSTEM IS TOO GREAT.
     *
     * YEAH, IT'S A PAIN TO HAVE TO GO TO THE SESSION DIRECTORY FOR THE 
     * INFO.  LIFE'S ROUGH.  SO ARE SECURITY BREACHES.
     */
    if (Verbose && !getenv("REQUEST_METHOD")) {    /* not under an HTTPD */
      WriteLabel(stderr, toproot[conn]);
    }
    if ((num = IK_CreateSearchConnections(&toproot[conn])) == -1) {
      err("Couldn't create search connection #%d", conn);
    }
    else {
      total_conn += num;
    }
    conn++;
  }
  if (total_conn <= 0){
    WriteLabel(stderr,toproot[0]); 
    puke("Faulty Search Tree");
  }


  /******************************
   * POLL UNTIL DONE!
   ******************************/

  /* Start calling the loop to poll the search: */
  ret = 1;
  debug("About to start operation...");
  start_time = time(NULL);
  while (ret != 0){ 

    /* Clear any outstanding errors: */
    IK_ImsErrno = errno = 0;

    /* Invoke poller, and remember result: */
    debug("About to invoke poller...");
    if ((ret = invoke_poller(num)) == -1) {
      puke("Error from poller");
    }
    debug("Poller gave us %d", ret);

    /* Check for abort file: */
    if (fp = fopen(StopPath, "r")) {
      fclose(fp);
      IK_AbortSearch();
      break;
    }
    
    /* Check for excessive run-time: */
    now_time = time(NULL);
    if ((run_d = difftime(now_time, start_time)) > (MAX_RUNTIME_MINS * 60)) {
      err("Exceeded %d minute maximum; aborting", MAX_RUNTIME_MINS);
      IK_AbortSearch();
      break;
    }

    /* Get ready for the next round! */
    debug("Sleeping until next round...");
    sleep(2);
  }
  debug("Done! (final return = %d)", ret);

  /* Done! */
  terminate_status();
  write_status();
  exit(0);
}

/************************************************************
 * usage
 ************************************************************
 * Output usage info to the stderr.
 */

static void usage() 
{
  fprintf(stderr, "\nUsage:\n");
  fprintf(stderr, "    imsclient action infile.odl      (deprecated)\n");
  fprintf(stderr, "    imsclient [opts...]\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "Opts are:\n");
  fprintf(stderr, "    -a<action>  one of...\n");
  fprintf(stderr, "                     ds = directory search\n");
  fprintf(stderr, "                     fb = FTP browse\n");
  fprintf(stderr, "                     ib = integrated browse\n");
  fprintf(stderr, "                     is = inventory search\n");
  fprintf(stderr, "                     pr = product request\n");
  fprintf(stderr, "                     os = product status request\n");
  fprintf(stderr, "                     as = account status request\n");
  fprintf(stderr, "                     rr = profile retrieval request\n");
  fprintf(stderr, "                     ur = profile update request\n");
  fprintf(stderr, "                     er = price estimate request\n");
  fprintf(stderr, "                     sr = profile submit request\n");
  fprintf(stderr, "                     cr = user password change request\n");
  fprintf(stderr, "    -g[yn]      do GCMD searches? default is yes (-gy)\n");
  fprintf(stderr, "    -i<infile>  the ODL input file\n");
  fprintf(stderr, "    -m<mode>    one of...\n");
  fprintf(stderr, "                     n = normal (default)\n");
  fprintf(stderr, "                     d = developer test\n");
  fprintf(stderr, "    -v          verbose output (debugs go to stderr)\n");
  fprintf(stderr, "\n");
}

/************************************************************
 * puke
 ************************************************************
 * Output and log error, and die.  
 * Invoke this like printf().
 */

static void puke(char *fmt, ...)
{
  char message[MAX_PUKE];
  va_list args;
  
  /* Format message to buffer: */
  va_start(args, fmt);
  vsprintf(message, fmt, args);
  va_end(args);
  
  /* Output to syslog AND to stderr: */
  err("Fatal error: %s\n", message);
  exit(-1);
}

/************************************************************
 * get_opts
 ************************************************************
 * Get command line options, and set global variables from them.
 */

static void get_opts(int argc, char **argv) {
  char **argp;
  char *arg;
  int barecount = 0;
  char buf[PATH_LEN];
  int action;    


  /* Setup defaults: */
  Mode = MODE_NORMAL;
  Action = 0;
  Infile[0] = '\0';

  /* Output usage right away if no args: */
  ++argv;
  if (!*argv) {
    usage(); 
    exit(-1);
  }

  /* Parse: */
  for (argp = argv; *argp; argp++) {
    arg = *argp;

  RETRY_ARG:
    if (*arg == '-') {     /* option... */
      switch (arg[1]) {  

      case 'a':                  /* -a<action> */
	for (action = 0; action < sizeof(ActionInfo); action++) {
	  if (strcmp(arg+2, ActionInfo[action].id) == 0) { 
	    Action = action;
	    break;
	  }
	}
	if (!action) {              /* not found! */
	  usage(); puke("bad action: %s", arg+2);
	}
	break;

      case 'g':                  /* -g<yesno> */
	GCMDSearch = (arg[2] != 'n');
	break;

      case 'i':                  /* -i<infile> */
	strcpy(Infile, arg+2);
	break;

      case 'm':                  /* -m<mode> */
	switch (arg[2]) {  
	case 'd':                  /* -md = developer */
	  Mode = MODE_DEVELOPER;
	  break;
	case 'n':                  /* -mn = normal */
	  Mode = MODE_NORMAL;
	  break;
	default:
	  usage(); puke("bad mode: %s", arg+2);
	}
	break;

      case 'v':                  /* verbose */
	Verbose = 1;
	break;
	
      default:
	usage(); puke("bad option: %s", arg);
      }
    }

    else {         /* bareword: */
      ++barecount;
      switch (barecount) {
      case 1:   /* first bareword is the action */
	sprintf(buf, "-a%s", arg);
	arg = buf;
	goto RETRY_ARG;
	break;
      case 2:   /* second bareword is the infile */
	sprintf(buf, "-i%s", arg);
	arg = buf;
	goto RETRY_ARG;
	break;
      default:
	usage(); puke("too many barewords! %s", arg);
      }
    }
  }
  
  /* Done! Do sanity check: */
  if (!Infile[0]) {
    usage(); puke("missing input file");
  }
  if (!Action) {
    usage(); puke("missing action");
  }
}



/************************************************************
 *
 * Functions related to our global information...
 *
 ************************************************************/


/*************************************************************
 * lookup_conn
 *************************************************************
 * Look up a connection by the DAAC name, creating it if need be.
 */
static IW_Conn *lookup_conn(char *daac) 
{
  IW_Conn *conn;

  /* Lookup; and if not there... */
  if (!(conn = IW_HashtAt(Conns, daac))) {
    info("Creating connection object for %s", daac);
    conn = IW_ConnNew(daac);                     /* create new object */
    IW_HashtAtPut(Conns, daac, (void *)conn);    /* store it */
  }
  return conn;
}

/*************************************************************
 * count_data_centers
 *************************************************************
 * This function will take the pre-loaded "archive.odl" aggregate
 * and return the number of data centers found + 1.
 */

static int count_data_centers()
{
  char *group_name;
  AGGREGATE dci_agg;
  AGGREGATE temp_agg;
  VALUE     temp_val;
  int total_daacs = 0;

  /* Sanity: */
  if (!ArchiveAgg) puke("%s not yet loaded",  ArchivePath);

  /* Find the data center info: */
  if (!(dci_agg = FindGroup(ArchiveAgg, "DATA_CENTER_INFO"))) {
    puke("%s is a bad archive.odl file");
  }

  /* For each DATA_CENTER group in the "archive.odl" tree... */
  group_name = "DATA_CENTER";  
  for (temp_agg = FindGroup(dci_agg, group_name);
       temp_agg;
       temp_agg = FindNextGroup(dci_agg, temp_agg, group_name)) {

    if ((temp_val = FirstValue(FindParameter(temp_agg, "DATA_CENTER_ID")))) {
      total_daacs++;
    } 
  }
  total_daacs++;
  return total_daacs;

}
/*************************************************************
 * lookup_data_center
 *************************************************************
 * This function will take a DAAC name, look at the "archive.odl", and 
 * return the archive information to the calling function.  
 *
 * Returns the DATA_CENTER aggregate on success, NULL 
 * on error or if not found.
 *
 * For efficiency, it uses the pre-loaded "archive.odl" aggregate.
 */

AGGREGATE lookup_data_center(char *arch_id)
{
  char *group_name;
  AGGREGATE dci_agg;
  AGGREGATE temp_agg;
  VALUE     temp_val;

  /* Sanity: */
  if (!ArchiveAgg) puke("%s not yet loaded",  ArchivePath);

  /* Find the data center info: */
  if (!(dci_agg = FindGroup(ArchiveAgg, "DATA_CENTER_INFO"))) {
    puke("%s is a bad archive.odl file");
  }

  /* For each DATA_CENTER group in the "archive.odl" tree... */
  group_name = "DATA_CENTER";  
  for (temp_agg = FindGroup(dci_agg, group_name);
       temp_agg;
       temp_agg = FindNextGroup(dci_agg, temp_agg, group_name)) {

    if (!(temp_val = FirstValue(FindParameter(temp_agg, "DATA_CENTER_ID")))) {
      warning("DATA_CENTER in %s without ID!", ArchivePath);
      continue;
    } 
    if (IW_Strequal(IW_ODLValueString(temp_val), arch_id)) {
      return temp_agg;       /* got it! */
    }
  }

  /* If we reached this point, we failed to find it: */
  err("data center '%s' not found in %s", arch_id, ArchivePath);  
  return NULL;
}





/************************************************************
 *
 * Functions related to setting up actions...
 *
 ************************************************************/

/************************************************************
 * reset_daac_counts
 ************************************************************
 *
 * DESCRIPTION
 *    Reset DAAC counts by making all existing datasets invisible,
 *    and setting their granule counts to 0.
 *    This must be done before an inventory/directory search!
 *
 *    After an inventory/directory search, the ones returned should
 *    become visible again, so it'll be easy to figure out what we got.
 *
 * NOTES
 *    This routine does an IK_GetDataset() and an IK_PutDataset() for each
 *    dataset currently in storage when called, and likewise for each
 *    granule.  Yuck.
 */

int reset_daac_counts()
{
  IK_Dataset *dset;     /* pointer to dataset struct */
  IK_Granule *gran; 	/* pointer to granule struct */
  IK_SortKey *key;	/* pointer to dataset/granule key */

  info("About to reset DAAC counts");

  /* Clear the datasets, so that tallying works properly. 
   * Note that we also have to clear the granule counts.
   * This might fail if there is no dataset.gdb yet; that's okay. */
  for (key = IK_FirstDatasetKey(); key; key = IK_NextDatasetKey()) {

    /* Get the dataset: */
    if (IK_GetDataset(key, &dset) != 0) {
      err("reset_daac_counts: couldn't get dataset!");
      IK_ImsErrno = errno = 0;
      free(key); 
      continue;
    }
      
    /* Reset the granule count, and make it invisible: */
    dset->marked.is_visible = IK_FALSE;
    dset->ngranules = 0;
    if (IK_PutDataset(dset) < 0) {
      err("reset_daac_counts: couldn't re-store dataset!");
      IK_ImsErrno = errno = 0;
    }

    /* Cleanup: */
    free(dset);		
    free(key);		
  }
  
  /* Done! */
  IK_ImsErrno = errno = 0;
  return 0;
}

/************************************************************
 * setup_any_search
 ************************************************************
 * Common set up for a directory/inventory searches.
 */
static int setup_any_search(AGGREGATE root,
			    AGGREGATE toproot[])
{
  /* Open the list file, and install the hooks to it: */  
  if (!(ListFp = fopen(ListPath, "w"))) {
    puke("couldn't open list file %s", ListPath);
  }

  /* Reset all the counts: */
  reset_daac_counts();
  IK_ImsErrno = errno = 0;

  /* Okay! */
  return 0;
}

/************************************************************
 * setup_dir_search
 ************************************************************
 * Set up for a directory search.
 */
static int setup_dir_search(AGGREGATE root,
			    AGGREGATE toproot[])
{
  int i, j = 0;
  AGGREGATE temp_agg, data_center_agg;
  PARAMETER temp_para;
  VALUE     temp_val;
 

  /* Read in the label: */
  if (!(IW_ODLReadFile(Infile, root))) {
    puke("Couldn't read %s", Infile);
  }

  /* Build the search tree from the odl file: */
  if (!(temp_agg = FindGroup (root, "DIRECTORY_SEARCH"))) {
    puke("Cannot find DIRECTORY_SEARCH group in input file");
  }

  /* Locate the data center id: */
  if (!(temp_para = FindParameter(temp_agg, "DATA_CENTER_ID"))) {
    puke("Cannot find DATA_CENTER_ID parameter in input file");
  }
  temp_val = FirstValue(temp_para);
  while (temp_val) {
    
    /* Get the parameter string and put the string into the fields array: */
    table[j] = (char *)calloc(strlen(temp_val->item.value.string)+1,
			      sizeof(char));
    strcpy (table[j], temp_val->item.value.string);
    j++;
    temp_val = NextValue(temp_val);
  }
  table[j] = (char *)NULL;

  for (i=0; table[i]!=(char*)NULL; ++i) {
    /* this loop will find the correct datacenter information from the 
       archive.odl file and put it into the search tree */
    if (!(data_center_agg = lookup_data_center(table[i]))) {
      puke("No info for data center %s.", table[i]);
    }
    PasteAggregate(temp_agg, CopyAggregate(data_center_agg));
  }

  /* Set up the vector: */
  toproot[0] = root;
  toproot[1] = NULL;

  /* Finally, do the common setup for searches: */
  setup_any_search(root, toproot);
  return 0;
}

/************************************************************
 * setup_inv_search
 ************************************************************
 * Set up for an inventory search.
 *
 * Exactly 1 aggregate -- the root -- is placed in the toproot vector.
 */
static int setup_inv_search(AGGREGATE root,
			    AGGREGATE toproot[])
{
  AGGREGATE temp_agg, data_center_agg;
  PARAMETER temp_para;
  VALUE     temp_val;
  int i = 0;
  int j = 0;

  
  /* Read in the label: */
  if (!(IW_ODLReadFile(Infile, root))) {
    puke("Couldn't read %s", Infile);
  }
  
  /* find the parameters specified in the odl tree and generate the 
   * dependent valids information, by first putting the parameter
   * values in the fields array. */
  if (!(temp_agg = FindGroup(root, "INVENTORY_SEARCH"))) {
    puke("Cannot find INVENTORY_SEARCH group in input file");
  }
  if (!(temp_para = FindParameter(temp_agg, "DATA_CENTER_ID"))) {
    puke("Cannot find DATA_CENTER_ID parameter in input file");
  }

  temp_val = FirstValue(temp_para);
  while (temp_val) {
    
    /* get the parameter string and put the string into the fields array */
    table[j] = (char *)calloc(strlen(temp_val->item.value.string)+1, 
			      sizeof(char));
    strcpy (table[j], temp_val->item.value.string);
    j++;
    temp_val = NextValue (temp_val);
  }
  table[j] = (char *) NULL;
  
  for (i=0; table[i]!=(char*)NULL; ++i) {
    
    /* this loop will find the correct datacenter infromation from the 
     * archive.odl file and put it into the search tree */
    if (!(data_center_agg = lookup_data_center(table[i]))) {
      puke("No info for data center %s", table[i]);
    }
    PasteAggregate(temp_agg, CopyAggregate(data_center_agg));
  } 

  /* Setup the toproot vector: */
  toproot[0] = root;
  toproot[1] = NULL;

  /* Finally, do the common setup for searches: */
  setup_any_search(root, toproot);
  return 0;
}

/************************************************************
 * setup_browse
 ************************************************************
 * Set up for an integrated/FTP browse.
 *
 * Exactly 1 aggregate -- the root -- is placed in the toproot vector.
 */
static int setup_browse(AGGREGATE root,
			AGGREGATE toproot[])
{
  AGGREGATE temp_agg, data_center_agg;
  PARAMETER temp_para;
  VALUE     temp_val;

  /* Read in the label: */
  if (!(IW_ODLReadFile(Infile, root))) {
    puke("Couldn't read %s", Infile);
  }

  /* Extract info: */
  if (!(temp_agg = FindGroup(root, "BROWSE_REQUEST"))) {
    puke("Cannot find BROWSE_REQUEST group in input file");
  }

  /* Locate the data center id: */
  if (!(temp_para = FindParameter(temp_agg, "DATA_CENTER_ID"))) { 
    puke("Cannot find DATA_CENTER_ID parameter in input file");
  }

  /* Add the data center information from archive.odl to the search tree: */
  temp_val = FirstValue(temp_para);
  if (!(data_center_agg = lookup_data_center(temp_val->item.value.string))) {
    puke("no such data center");
  }
  PasteAggregate(temp_agg, CopyAggregate(data_center_agg));

  /* Put the odl tree to the toproot array: */
  toproot[0] = root;
  toproot[1] = NULL;

  /* Okay: */
  return 0;
}

/************************************************************
 * setup_prod_request
 ************************************************************
 * Set up for a product request.
 *
 * This places in the toproot[] array a NULL-terminated vector of 
 * n ODL trees, where each tree holds exactly one of the n
 * DAAC requests, and looks like:
 *
 *      GROUP         = root
 *        GROUP         = PRODUCT_REQUEST
 *          GROUP          = DATA_CENTER
 *            ...
 *          END_GROUP      = DATA_CENTER
 *          ...
 *        END_GROUP     = PRODUCT_REQUEST
 *      END_GROUP     = root
 *
 * The DATA_CENTER group is pasted in from the archive.odl.
 * That top-level "root" aggregate is important!
 *
 * Returns 0 on success, negative or fatal exception on failure.
 */
static int setup_prod_request(AGGREGATE root,
			      AGGREGATE toproot[])
{
  AGGREGATE pr_agg, pr_copy_agg, data_center_agg;
  PARAMETER temp_para;
  VALUE     temp_val;
  int i;


  /* Read in the label: */
  if (!(IW_ODLReadFile(Infile, root))) {
    puke("Couldn't read %s", Infile);
  }

  /* Extract the product request groups: */
  toproot[0] = NULL;
  for (pr_agg = FindAggregate(root, "PRODUCT_REQUEST"), i = 0;
       pr_agg;
       pr_agg = FindNextGroup(root, pr_agg, "PRODUCT_REQUEST"), ++i) {
    char *arch_id;

    /* Get the DATA_CENTER_ID: */
    if (!(temp_para = FindParameter(pr_agg, "DATA_CENTER_ID"))) {
      puke("Cannot find DATA_CENTER_ID parameter in input file");
    }
    arch_id = IW_ODLValueString(FirstValue(temp_para));
       
    /* Look up the data center info: */
    if (!(data_center_agg = lookup_data_center(arch_id))) {
      puke("No info for %s", arch_id);
    }

    /* Copy the PRODUCT_REQUEST, and paste in the data center info: */
    pr_copy_agg = CopyAggregate(pr_agg);
    PasteAggregate(pr_copy_agg, CopyAggregate(data_center_agg));    
    
    /* Create a new root node: */
    toproot[i] = NewAggregate(NULL, KA_GROUP, "root", "");
    PasteAggregate(toproot[i], pr_copy_agg);
    toproot[i+1] = NULL;
  }

  /* Did we get anything? */
  if (!toproot[0]) {
    puke("No PRODUCT_REQUEST groups!");
  }
  return 0;
}

/************************************************************
 * setup_prod_stat_request
 ************************************************************
 * Set up for a product status request.
 *
 * This places in the toproot[] array a NULL-terminated vector of 
 * n ODL trees, where each tree holds exactly one of the n
 * DAAC requests, and looks like:
 *
 *      GROUP         = root
 *        GROUP         = PRODUCT_STATUS_REQUEST
 *          GROUP          = DATA_CENTER
 *            ...
 *          END_GROUP      = DATA_CENTER
 *          ...
 *        END_GROUP     = PRODUCT_STATUS_REQUEST
 *      END_GROUP     = root
 *
 * The DATA_CENTER group is pasted in from the archive.odl.
 * That top-level "root" aggregate is important!
 *
 * Returns 0 on success, negative or fatal exception on failure.
 */
static int setup_prod_stat_request(AGGREGATE root,
			      AGGREGATE toproot[])
{
  AGGREGATE pr_agg, data_center_agg;
  PARAMETER temp_para;
  VALUE     temp_val;
  int i;
  int k = 0;
  int j = 0;


  /* Read in the label: */
  if (!(IW_ODLReadFile(Infile, root))) {
    puke("Couldn't read %s", Infile);
  }

  /* Extract the product request groups: */
  toproot[0] = NULL;
  for (pr_agg = FindAggregate(root, "PRODUCT_STATUS_REQUEST"), i = 0;
       pr_agg;
       pr_agg = FindNextGroup(root, pr_agg, "PRODUCT_STATUS_REQUEST"), ++i) {
    /* Create a new root node: */
    toproot[i] = root;

    /* Get the DATA_CENTER_ID: */
    if (!(temp_para = FindParameter(pr_agg, "DATA_CENTER_ID"))) {
      puke("Cannot find DATA_CENTER_ID parameter in input file");
    }
    temp_val = FirstValue(temp_para);
    while (temp_val) {
    
      /* get the parameter string and put the string into the fields array */
      table[j] = (char *)calloc(strlen(temp_val->item.value.string)+1, 
			      sizeof(char));
      strcpy (table[j], temp_val->item.value.string);
      j++;
      temp_val = NextValue (temp_val);
    }
    table[j] = (char *) NULL;
  
    for (k=0; table[k]!=(char*)NULL; ++k) {
    
      /* this loop will find the correct datacenter infromation from the 
       * archive.odl file and put it into the search tree */
      if (!(data_center_agg = lookup_data_center(table[k]))) {
        puke("No info for data center %s", table[k]);
      }
      /* Copy the PRODUCT_STATUS_REQUEST, and paste in the data center info: */
      PasteAggregate(pr_agg, CopyAggregate(data_center_agg));
    } 
    
    toproot[i+1] = NULL;
  }

  /* Did we get anything? */
  if (!toproot[0]) {
    puke("No PRODUCT_STATUS_REQUEST groups!");
  }
  return 0;
} /* end of setup_prod_stat_request */

/************************************************************
 * setup_acct_stat_request
 ************************************************************
 * Set up for a account status request.
 *
 * This places in the toproot[] array a NULL-terminated vector of 
 * n ODL trees, where each tree holds exactly one of the n
 * DAAC requests, and looks like:
 *
 *      GROUP         = root
 *        GROUP         = ACCOUNT_STATUS_REQUEST
 *          GROUP          = DATA_CENTER
 *            ...
 *          END_GROUP      = DATA_CENTER
 *          ...
 *        END_GROUP     = ACCOUNT_STATUS_REQUEST
 *      END_GROUP     = root
 *
 * The DATA_CENTER group is pasted in from the archive.odl.
 * That top-level "root" aggregate is important!
 *
 * Returns 0 on success, negative or fatal exception on failure.
 */
static int setup_acct_stat_request(AGGREGATE root,
			      AGGREGATE toproot[])
{
  AGGREGATE pr_agg, data_center_agg;
  PARAMETER temp_para;
  VALUE     temp_val;
  int i;
  int k = 0;
  int j = 0;


  /* Read in the label: */
  if (!(IW_ODLReadFile(Infile, root))) {
    puke("Couldn't read %s", Infile);
  }

  /* Extract the product request groups: */
  toproot[0] = NULL;
  for (pr_agg = FindAggregate(root, "ACCOUNT_STATUS_REQUEST"), i = 0;
       pr_agg;
       pr_agg = FindNextGroup(root, pr_agg, "ACCOUNT_STATUS_REQUEST"), ++i) {
    toproot[i] = root;
    /* Get the DATA_CENTER_ID: */
    if (!(temp_para = FindParameter(pr_agg, "DATA_CENTER_ID"))) {
      puke("Cannot find DATA_CENTER_ID parameter in input file");
    }
    temp_val = FirstValue(temp_para);
    while (temp_val) {

      /* get the parameter string and put the string into the fields array */
      table[j] = (char *)calloc(strlen(temp_val->item.value.string)+1,
                              sizeof(char));
      strcpy (table[j], temp_val->item.value.string);
      j++;
      temp_val = NextValue (temp_val);
    }
    table[j] = (char *) NULL;
 
    for (k=0; table[k]!=(char*)NULL; ++k) {

      /* this loop will find the correct datacenter infromation from the
       * archive.odl file and put it into the search tree */
      if (!(data_center_agg = lookup_data_center(table[k]))) {
        puke("No info for data center %s", table[k]);
      }
      /* Copy the ACCOUNT_STATUS_REQUEST, and paste in the data center info: */
      PasteAggregate(pr_agg, CopyAggregate(data_center_agg));
    }

    toproot[i+1] = NULL;
  }

  /* Did we get anything? */
  if (!toproot[0]) {
    puke("No ACCOUNT_STATUS_REQUEST groups!");
  }
  return 0;
} /* end of setup_acct_stat_request */

/************************************************************
 * setup_prof_retrieve_request
 ************************************************************
 * Set up for a profile retrieval request.
 * Generated from setup_prod_stat_request.
 * This places in the toproot[] array a NULL-terminated vector of 
 * n ODL trees, where each tree holds exactly one of the n
 * DAAC requests, and looks like:
 *
 *      GROUP         = root
 *        GROUP         = PROFILE_RETRIEVAL_REQUEST
 *          GROUP          = DATA_CENTER
 *            ...
 *          END_GROUP      = DATA_CENTER
 *          ...
 *        END_GROUP     = PROFILE_RETRIEVAL_REQUEST
 *      END_GROUP     = root
 *
 * The DATA_CENTER group is pasted in from the archive.odl.
 * That top-level "root" aggregate is important!
 *
 * Returns 0 on success, negative or fatal exception on failure.
 */
static int setup_prof_retrieve_request(AGGREGATE root,
			      AGGREGATE toproot[])
{
  AGGREGATE pr_agg, data_center_agg;
  PARAMETER temp_para;
  VALUE     temp_val;
  int i;
  int k = 0;
  int j = 0;

  /* Read in the label: */
  if (!(IW_ODLReadFile(Infile, root))) {
    puke("Couldn't read %s", Infile);
  }

  /* Extract the product request groups: */
  toproot[0] = NULL;
  for (pr_agg = FindAggregate(root, "PROFILE_RETRIEVAL_REQUEST"), i = 0;
       pr_agg;
       pr_agg = FindNextGroup(root, pr_agg, "PROFILE_RETRIEVAL_REQUEST"), ++i) {
    /* Create a new root node: */
    toproot[i] = root;

    /* Get the DATA_CENTER_ID: */
    if (!(temp_para = FindParameter(pr_agg, "DATA_CENTER_ID"))) {
      puke("Cannot find DATA_CENTER_ID parameter in input file");
    }
    temp_val = FirstValue(temp_para);
    while (temp_val) {
    
      /* get the parameter string and put the string into the fields array */
      table[j] = (char *)calloc(strlen(temp_val->item.value.string)+1, 
			      sizeof(char));
      strcpy (table[j], temp_val->item.value.string);
      j++;
      temp_val = NextValue (temp_val);
    }
    table[j] = (char *) NULL;
  
    for (k=0; table[k]!=(char*)NULL; ++k) {
    
      /* this loop will find the correct datacenter infromation from the 
       * archive.odl file and put it into the search tree */
      if (!(data_center_agg = lookup_data_center(table[k]))) {
        puke("No info for data center %s", table[k]);
      }
      /* Copy the PROFILE_RETRIEVAL_REQUEST, and paste in the data center 
       * info: */
      PasteAggregate(pr_agg, CopyAggregate(data_center_agg));
    } 
    
    toproot[i+1] = NULL;
  }

  /* Did we get anything? */
  if (!toproot[0]) {
    puke("No PROFILE_RETRIEVAL_REQUEST groups!");
  }
  return 0;
} /* end of setup_prof_retrieve_request */

/************************************************************
 * setup_prof_update_request
 ************************************************************
 * Set up for a profile update request.
 * Generated from account status request.
 *
 * This places in the toproot[] array a NULL-terminated vector of 
 * n ODL trees, where each tree holds exactly one of the n
 * DAAC requests, and looks like:
 *
 *      GROUP         = root
 *        GROUP         = PROFILE_UPDATE_REQUEST
 *          GROUP          = DATA_CENTER
 *            ...
 *          END_GROUP      = DATA_CENTER
 *          ...
 *        END_GROUP     = PROFILE_UPDATE_REQUEST
 *      END_GROUP     = root
 *
 * The DATA_CENTER group is pasted in from the archive.odl.
 * That top-level "root" aggregate is important!
 *
 * Returns 0 on success, negative or fatal exception on failure.
 */
static int setup_prof_update_request(AGGREGATE root,
			      AGGREGATE toproot[])
{
  AGGREGATE pr_agg, data_center_agg;
  PARAMETER temp_para;
  VALUE     temp_val;
  int i;
  int k = 0;
  int j = 0;

  /* Read in the label: */
  if (!(IW_ODLReadFile(Infile, root))) {
    puke("Couldn't read %s", Infile);
  }

  /* Extract the product request groups: */
  toproot[0] = NULL;
  for (pr_agg = FindAggregate(root, "PROFILE_UPDATE_REQUEST"), i = 0;
       pr_agg;
       pr_agg = FindNextGroup(root, pr_agg, "PROFILE_UPDATE_REQUEST"), ++i) {
    toproot[i] = root;
    /* Get the DATA_CENTER_ID: */
    if (!(temp_para = FindParameter(pr_agg, "DATA_CENTER_ID"))) {
      puke("Cannot find DATA_CENTER_ID parameter in input file");
    }
    temp_val = FirstValue(temp_para);
    while (temp_val) {

      /* get the parameter string and put the string into the fields array */
      table[j] = (char *)calloc(strlen(temp_val->item.value.string)+1,
                              sizeof(char));
      strcpy (table[j], temp_val->item.value.string);
      j++;
      temp_val = NextValue (temp_val);
    }
    table[j] = (char *) NULL;
 
    for (k=0; table[k]!=(char*)NULL; ++k) {

      /* this loop will find the correct datacenter infromation from the
       * archive.odl file and put it into the search tree */
      if (!(data_center_agg = lookup_data_center(table[k]))) {
        puke("No info for data center %s", table[k]);
      }
      /* Copy the PROFILE_UPDATE_REQUEST, and paste in the data center info: */
      PasteAggregate(pr_agg, CopyAggregate(data_center_agg));
    }

    toproot[i+1] = NULL;
  }

  /* Did we get anything? */
  if (!toproot[0]) {
    puke("No PROFILE_UPDATE_REQUEST groups!");
  }
  return 0;
} /* end of setup_prof_update_request */


/************************************************************
 * setup_price_estimate_request
 ************************************************************
 * Set up for a price estimate request.
 * Generated from profile update request.
 *
 * This places in the toproot[] array a NULL-terminated vector of 
 * n ODL trees, where each tree holds exactly one of the n
 * DAAC requests, and looks like:
 *
 *      GROUP         = root
 *        GROUP         = PRICE_ESTIMATE_REQUEST
 *          GROUP          = DATA_CENTER
 *            ...
 *          END_GROUP      = DATA_CENTER
 *          ...
 *        END_GROUP     = PRICE_ESTIMATE_REQUEST
 *      END_GROUP     = root
 *
 * The DATA_CENTER group is pasted in from the archive.odl.
 * That top-level "root" aggregate is important!
 *
 * Returns 0 on success, negative or fatal exception on failure.
 */
static int setup_price_estimate_request(AGGREGATE root,
			      AGGREGATE toproot[])
{
  AGGREGATE pr_agg, data_center_agg;
  PARAMETER temp_para;
  VALUE     temp_val;
  int i;
  int k = 0;
  int j = 0;

  /* Read in the label: */
  if (!(IW_ODLReadFile(Infile, root))) {
    puke("Couldn't read %s", Infile);
  }

  /* Extract the product request groups: */
  toproot[0] = NULL;
  for (pr_agg = FindAggregate(root, "PRICE_ESTIMATE_REQUEST"), i = 0;
       pr_agg;
       pr_agg = FindNextGroup(root, pr_agg, "PRICE_ESTIMATE_REQUEST"), ++i) {
    toproot[i] = root;
    /* Get the DATA_CENTER_ID: */
    if (!(temp_para = FindParameter(pr_agg, "DATA_CENTER_ID"))) {
      puke("Cannot find DATA_CENTER_ID parameter in input file");
    }
    temp_val = FirstValue(temp_para);
    while (temp_val) {

      /* get the parameter string and put the string into the fields array */
      table[j] = (char *)calloc(strlen(temp_val->item.value.string)+1,
                              sizeof(char));
      strcpy (table[j], temp_val->item.value.string);
      j++;
      temp_val = NextValue (temp_val);
    }
    table[j] = (char *) NULL;
 
    for (k=0; table[k]!=(char*)NULL; ++k) {

      /* this loop will find the correct datacenter infromation from the
       * archive.odl file and put it into the search tree */
      if (!(data_center_agg = lookup_data_center(table[k]))) {
        puke("No info for data center %s", table[k]);
      }
      /* Copy the PRICE_ESTIMATE_REQUEST, and paste in the data center info: */
      PasteAggregate(pr_agg, CopyAggregate(data_center_agg));
    }

    toproot[i+1] = NULL;
  }

  /* Did we get anything? */
  if (!toproot[0]) {
    puke("No PRICE_ESTIMATE_REQUEST groups!");
  }
  return 0;
} /* end of setup_price_estimate_request */

/************************************************************
 * setup_prof_submit_request
 ************************************************************
 * Set up for a profile submit request.
 * Generated from price estimate request.
 *
 * This places in the toproot[] array a NULL-terminated vector of 
 * n ODL trees, where each tree holds exactly one of the n
 * DAAC requests, and looks like:
 *
 *      GROUP         = root
 *        GROUP         = PROFILE_SUBMIT_REQUEST
 *          GROUP          = DATA_CENTER
 *            ...
 *          END_GROUP      = DATA_CENTER
 *          ...
 *        END_GROUP     = PROFILE_SUBMIT_REQUEST
 *      END_GROUP     = root
 *
 * The DATA_CENTER group is pasted in from the archive.odl.
 * That top-level "root" aggregate is important!
 *
 * Returns 0 on success, negative or fatal exception on failure.
 */
static int setup_prof_submit_request(AGGREGATE root,
			      AGGREGATE toproot[])
{
  AGGREGATE pr_agg, data_center_agg;
  PARAMETER temp_para;
  VALUE     temp_val;
  int i;
  int k = 0;
  int j = 0;

  /* Read in the label: */
  if (!(IW_ODLReadFile(Infile, root))) {
    puke("Couldn't read %s", Infile);
  }

  /* Extract the product request groups: */
  toproot[0] = NULL;
  for (pr_agg = FindAggregate(root, "PROFILE_SUBMIT_REQUEST"), i = 0;
       pr_agg;
       pr_agg = FindNextGroup(root, pr_agg, "PROFILE_SUBMIT_REQUEST"), ++i) {
    toproot[i] = root;
    /* Get the DATA_CENTER_ID: */
    if (!(temp_para = FindParameter(pr_agg, "DATA_CENTER_ID"))) {
      puke("Cannot find DATA_CENTER_ID parameter in input file");
    }
    temp_val = FirstValue(temp_para);
    while (temp_val) {

      /* get the parameter string and put the string into the fields array */
      table[j] = (char *)calloc(strlen(temp_val->item.value.string)+1,
                              sizeof(char));
      strcpy (table[j], temp_val->item.value.string);
      j++;
      temp_val = NextValue (temp_val);
    }
    table[j] = (char *) NULL;
 
    for (k=0; table[k]!=(char*)NULL; ++k) {

      /* this loop will find the correct datacenter infromation from the
       * archive.odl file and put it into the search tree */
      if (!(data_center_agg = lookup_data_center(table[k]))) {
        puke("No info for data center %s", table[k]);
      }
      /* Copy the PROFILE_SUBMIT_REQUEST, and paste in the data center info: */
      PasteAggregate(pr_agg, CopyAggregate(data_center_agg));
    }

    toproot[i+1] = NULL;
  }

  /* Did we get anything? */
  if (!toproot[0]) {
    puke("No PROFILE_SUBMIT_REQUEST groups!");
  }
  return 0;
} /* end of setup_prof_submit_request */

/************************************************************
 * setup_password_change_request
 ************************************************************
 * Set up for a user password change request.
 * Generated from price estimate request.
 *
 * This places in the toproot[] array a NULL-terminated vector of 
 * n ODL trees, where each tree holds exactly one of the n
 * DAAC requests, and looks like:
 *
 *      GROUP         = root
 *        GROUP         = USER_PASSWORD_CHANGE_REQUEST
 *          GROUP          = DATA_CENTER
 *            ...
 *          END_GROUP      = DATA_CENTER
 *          ...
 *        END_GROUP     = USER_PASSWORD_CHANGE_REQUEST
 *      END_GROUP     = root
 *
 * The DATA_CENTER group is pasted in from the archive.odl.
 * That top-level "root" aggregate is important!
 *
 * Returns 0 on success, negative or fatal exception on failure.
 */
static int setup_password_change_request(AGGREGATE root,
			      AGGREGATE toproot[])
{
  AGGREGATE pr_agg, data_center_agg;
  PARAMETER temp_para;
  VALUE     temp_val;
  int i;
  int k = 0;
  int j = 0;

  /* Read in the label: */
  if (!(IW_ODLReadFile(Infile, root))) {
    puke("Couldn't read %s", Infile);
  }

  /* Extract the product request groups: */
  toproot[0] = NULL;
  for (pr_agg = FindAggregate(root, "USER_PASSWORD_CHANGE_REQUEST"), i = 0;
       pr_agg;
       pr_agg = FindNextGroup(root, pr_agg, "USER_PASSWORD_CHANGE_REQUEST"), 
       ++i) {
    toproot[i] = root;
    /* Get the DATA_CENTER_ID: */
    if (!(temp_para = FindParameter(pr_agg, "DATA_CENTER_ID"))) {
      puke("Cannot find DATA_CENTER_ID parameter in input file");
    }
    temp_val = FirstValue(temp_para);
    while (temp_val) {

      /* get the parameter string and put the string into the fields array */
      table[j] = (char *)calloc(strlen(temp_val->item.value.string)+1,
                              sizeof(char));
      strcpy (table[j], temp_val->item.value.string);
      j++;
      temp_val = NextValue (temp_val);
    }
    table[j] = (char *) NULL;
 
    for (k=0; table[k]!=(char*)NULL; ++k) {

      /* this loop will find the correct datacenter infromation from the
       * archive.odl file and put it into the search tree */
      if (!(data_center_agg = lookup_data_center(table[k]))) {
        puke("No info for data center %s", table[k]);
      }
      /* Copy the USER_PASSWORD_CHANGE_REQUEST, and paste in the data 
         center info: */
      PasteAggregate(pr_agg, CopyAggregate(data_center_agg));
    }

    toproot[i+1] = NULL;
  }

  /* Did we get anything? */
  if (!toproot[0]) {
    puke("No USER_PASSWORD_CHANGE_REQUEST groups!");
  }
  return 0;
} /* end of setup_password_change_request */

#if 0
/************************************************************
 * conn_event
 ***********************************************************
 *
 * DESCRIPTION
 *    Map the event code to a string, and use it to update the LED stage.
 *
 *    The event code is one of the IK_* constants in IK_Nm.h.
 *    They represent events returned by all the functions used by the
 *    IK Network Manager state machine.
 *
 * PARAMETERS
 *    'conn' is the current connection, 'event' is the last event
 *    that happened on the connection, and 'event_string' is a buffer.
 *
 *    On output, IW_ConnStage(conn) holds the new LED stage, guaranteed 
 *    never to go to a silly next stage.
 *
 * NOTES
 *    Basically, a simple state-machine.
 *
 *    In some transitions, we explicitly check against the last stage. In 
 *    others, we do not.  This is intentional: we check against the last stage
 *    when we KNOW that certain transitions should normally be coerced to
 *    "no-ops".  The idea is that where we DON'T know whether the transition 
 *    should legally be a no-op, we attempt it anyway, knowing that a 
 *    warning will be output by the "sanity clause" section at the bottom.
 *    We can then examine the syslog for these warnings, which begin
 *    with "LED: REFUSED".
 */ 

static void conn_event(IW_Conn *conn, int event, char event_string[])
{
  IW_Connstage old_stage;        /* safe copy of old stage */
  IW_Connstage stage;            /* the new stage */


  /* Set up stages: */
  old_stage = IW_ConnStage(conn);  /* make safe copy of old stage */
  stage = old_stage;               /* new stage: default is current stage */

  /* Set up event string: */
  event_string[0] = '\0';
  debug("Event %d, old LED stage = %d (%s)", event, stage, 
	IW_ConnstageName(stage));



  /* 
   * Get the correct event string and stage code based on event...
   */
  switch (event) {

    /******************************
     * Null or ambiguous transitions...
     ******************************/
    
  case IK_CONNECT_FAILED:          /* Logical connection failed... must
				    * try again (so not a final state). */
    strcpy(event_string, "Connection failed");
    /*                                ELIDED: stage = IW_ConnS_EFAILED; */
    /* STAGE DOES NOT CHANGE */
    break;

  case IK_CONNECT_CLOSED:          /* Normal at end */
    strcpy(event_string, "Connection closed");
    /* STAGE DOES NOT CHANGE */
    break;

  case IK_GOT_NODATA:              /* Ok, server just hasn't sent stuff yet */
    strcpy(event_string, "Got No Data");
    /* STAGE DOES NOT CHANGE */
    break;

  case IK_GOT_ACK: 
    strcpy(event_string, "Got ack tree");
    /* STAGE DOES NOT CHANGE */
    break;

  case IK_GCMD_AMBIGOUS:
    strcpy(event_string, "Need to retry MD search");
    /* STAGE DOES NOT CHANGE */
    break;

  case IK_GOT_QUIT:                /* Got a "quit" tree */
    debug("Got quit tree.");
    strcpy(event_string, "Got quit tree");    
    if (stage < IW_ConnS_RECEIVING) {    /* quitting with nothing back: */ 
      stage = IW_ConnS_EFAILED;          /*   assume that we failed */
    }    
    else if (stage == IW_ConnS_RECEIVING) { /* quit after receiving stuff: */
      switch (Action) {
      case ACTION_INV_SEARCH:            /* did we get inv search results? */
	debug("Got quit tree for INV SEARCH while receiving results.");
	stage = (IW_ConnNumGranules(conn) ? IW_ConnS_OK : IW_ConnS_EFAILED);
	break;
      case ACTION_DIR_SEARCH:            /* did we get dir search results? */
	debug("Got quit tree for DIR SEARCH while receiving results.");
	stage = (IW_ConnNumDatasets(conn) ? IW_ConnS_OK : IW_ConnS_EFAILED);
	break;
      default:
	debug("Got quit tree for misc operation.");
	stage = IW_ConnS_OK;            /* assume okay */
      }
    }
    else {
      /* STAGE DOES NOT CHANGE */
    }
    break;

    /******************************
     * Transitions to IW_ConnS_EFAILED...
     ******************************/
    
  case IK_CLIENT_CRASH:            /* Eric W says this is a bogus event */
    strcpy(event_string, "Client has crashed");
    stage = IW_ConnS_EFAILED;
    break;

  case IK_CLIENTSYS_ERROR:         /* Yow! Something wrong on our end! */
    strcpy(event_string, "Client system error occur");
    stage = IW_ConnS_EFAILED;
    break;

    /******************************
     * Transitions to IW_ConnS_EABORTED...
     ******************************/

  case IK_ABORT_SEARCH:            /* Search being aborted */
    strcpy(event_string, "Got abort search command");
    stage = IW_ConnS_EABORTED;
    break;

  case IK_GOT_ABORT:
    strcpy(event_string, "Got abort tree");
    stage = IW_ConnS_EABORTED;
    break;

    /******************************
     * Transitions to IW_ConnS_EDAACDOWN...
     ******************************/

  case IK_ARCHIVE_DOWN:            /* Note: Eric W says that this can be sent
				    * in conditions other than the DAAC being
				    * down.  I wonder... is it because this
				    * constant has value '0', and someone
				    * is misusing it? */
    strcpy(event_string, "Archive is down");
    if (stage >= IW_ConnS_END_MIN){/* If we've ALREADY reached an end stage: */
      break;                             /* don't change the stage */
    }
    else {                         /* If we haven't reached an end stage... */
      switch (Action) {
      case ACTION_INV_SEARCH:            /* not down if we got granules: */
	if (!IW_ConnNumGranules(conn)) { stage = IW_ConnS_EDAACDOWN; }
	break;
      case ACTION_DIR_SEARCH:            /* not down if we got datasets: */
	if (!IW_ConnNumDatasets(conn)) { stage = IW_ConnS_EDAACDOWN; }
	break;
      default:                           /* no way to verify; assume down */ 
	stage = IW_ConnS_EDAACDOWN; 
      }
    }
    break;

    /******************************
     * Transitions to IW_ConnS_CONNECTED...
     ******************************/

  case IK_CONNECT_OPEN:            /* Normal at the start */
    strcpy(event_string, "Connection opened");
    stage = IW_ConnS_CONNECTED;
    break;

    /******************************
     * Transitions to IW_ConnS_SENDING...
     ******************************/

  case IK_TRANSMITTED:             /* We sent a tree; a good thing */
    strcpy(event_string, "Tree transmitted");
    if (stage < IW_ConnS_SENDING)     /* If weren't already sending stuff, */
      stage = IW_ConnS_SENDING;       /*    we are now. */
    break;

    /******************************
     * Transitions to IW_ConnS_RECEIVING...
     ******************************/

  case IK_GOT_INVRESULT:           /* Get inv. result *message* */
    strcpy(event_string, "Got inventory result");
    if (stage < IW_ConnS_RECEIVING)   /* If weren't already receiving stuff, */
      stage = IW_ConnS_RECEIVING;     /*    we are now. */
    break;

  case IK_GOT_PRODRESULT: 
    strcpy(event_string, "Got product result");
    stage = IW_ConnS_RECEIVING;
    break;

  case IK_GOT_FTPBROWSE:
    strcpy(event_string, "Got FTP browse");
    stage = IW_ConnS_RECEIVING;
    break;

  case IK_GOT_MDDIFS:
    strcpy(event_string, "Got Browse Label Tree");
    stage = IW_ConnS_RECEIVING;
    break;

  case IK_GOT_INTBROWSE:     /* just read i.b. HEADER ODL MESSAGE, not data */
    strcpy(event_string, "Got integrated browse");
    stage = IW_ConnS_RECEIVING;
    break;

  case IK_IMAGE_RECEIVED:
    strcpy(event_string, "Got a fragment of an image");
    if (stage < IW_ConnS_RECEIVING)   /* If weren't already receiving stuff, */
      stage = IW_ConnS_RECEIVING;     /*    we are now. */
    break;

  case IK_GCMD_SUCCESS:
    strcpy(event_string, "The MD search worked");
    stage = IW_ConnS_RECEIVING;
    break;

  case IK_INGESTIMG_SUCCESS:       /* We got a *portion* of the browse 
				      image... but not the whole thing! */
    strcpy(event_string, "Portion of image successfuly ingested");
    if (stage < IW_ConnS_RECEIVING)   /* If weren't already receiving stuff, */
      stage = IW_ConnS_RECEIVING;     /*    we are now. */
    break;

  case IK_INGESTINV_SUCCESS:       /* Ingested a chunk okay... */
    strcpy(event_string, "Inventory ingest to the Data Manager was succesful");
    if (stage < IW_ConnS_RECEIVING)   /* If weren't already receiving stuff, */
      stage = IW_ConnS_RECEIVING;     /*    we are now. */
    break;

  case IK_INGESTGCMD_SUCCESS:
    strcpy(event_string, "Data from GCMD got ingested to Data Manager");
    if (stage < IW_ConnS_RECEIVING)   /* If weren't already receiving stuff, */
      stage = IW_ConnS_RECEIVING;     /*    we are now. */
    break;

    /******************************
     * Transitions to IW_ConnS_OK...
     ******************************/

  case IK_IMAGE_COMPLETE:
    strcpy(event_string, "Whole image ingested");
    stage = IW_ConnS_OK;
    break;

  case IK_INGESTFTPB_SUCCESS:
    strcpy(event_string, "FTP Browse results ingested");
    stage = IW_ConnS_OK;
    break;

  case IK_INGESTPR_SUCCESS:
    strcpy(event_string, "Product Request results ingested");
    stage = IW_ConnS_OK;
    break;

  case IK_TERMINATED:              /* Always the last stage. */
    if (stage == IW_ConnS_RECEIVING){ /* if we were receiving results, */
      stage = IW_ConnS_OK;        /*    assume we got stuff */
      strcpy(event_string, "Search finished");
    }
    break;


  default:
    /* event message does not change */
    /* stage does not change */
    break;
  }


  /* Sanity clause!  
   * (Hey, you canna foola me... there's so such ting assa Sanity Claus...)
   *
   * We want to make SURE that stages don't go "backwards", and
   * we make sure that the final stage is "sticky":
   */
  if (stage == old_stage) {             /* nothing changed */
    /* no-op */
  }
  else if (old_stage >= IW_ConnS_END_MIN){ /* old stage was a final stage */
    warning("LED: REFUSED attempt to unfinalize %s to %s", 
	    IW_ConnstageName(old_stage), IW_ConnstageName(stage));
    stage = old_stage;                      /* sticky */  
  }
  else if (stage < old_stage) {         /* old stage nonfinal, new is lower! */
    warning("LED: REFUSED attempt to downgrade %s to %s", 
	    IW_ConnstageName(old_stage), IW_ConnstageName(stage));
    stage = old_stage;                      /* hold at old stage */
  }
  else {                                /* old stage nonfinal, new is higher */
    notice("LED: Transition from %s to %s",
	   IW_ConnstageName(old_stage), IW_ConnstageName(stage));
  }

  /* Done! */
  IW_ConnSetStage(conn, stage);
}
#endif




/************************************************************
 *
 * Functions related to communication...
 *
 ************************************************************/

/*************************************************************
 * terminate_status
 *************************************************************
 * Force all statuses to have some final LED stage.
 * This is done before a graceful exit.
 */

int terminate_status()
{
  char *arch_id;
  IW_Conn *conn;

  /* For each connection in the hash... */
  IW_HashtRewind(Conns);
  while (IW_HashtNext(Conns, &arch_id, (IW_Elem *)(&conn)) == IW_HashtOGO) {
    if (IW_ConnStage(conn) < IW_ConnS_END_MIN) {
      IW_ConnSetStage(conn, IW_ConnS_EFAILED);
    }
  }
  return 0;
}

/*************************************************************
 * write_status
 *************************************************************
 * Write out the current communication status.
 * We get this from the global hash of connections.
 *
 * Returns 0 on success, negative on error.
 */

int write_status()
{
  char stat_tmpname[PATH_LEN];   /* "client.stat.new" */
  char stat_realname[PATH_LEN];  /* "client.stat" */
  FILE *fp;                      /* file ptr to the "client.stat.new" file */
  char *arch_id;  
  IW_Conn *conn;

  /* Create the full path name for the search stat files: */
  sprintf(stat_realname, "%s%s", GaeaTmpDir, "client.stat");
  sprintf(stat_tmpname,  "%s%s", GaeaTmpDir, "client.stat.tmp");
  
  /* Open the search stat file (open a temp file, write, then rename it): */
  if (!(fp = fopen(stat_tmpname, "w"))) {
    return err("can't open %s for writing", stat_tmpname);
  }

  /* For each connection in the hash... */
  IW_HashtRewind(Conns);
  while (IW_HashtNext(Conns, &arch_id, (IW_Elem *)(&conn)) == IW_HashtOGO) {
    
    /* Write this record to the search stat file: */
    IW_ChunkBegin(fp,"COMMSTATUS");
    IW_ChunkField(fp,"DATA_CENTER_ID",     "%s", arch_id);
    IW_ChunkField(fp,"LED_STAGE",          "%d", IW_ConnStage(conn));
    IW_ChunkField(fp,"NUMBER_OF_DATASETS", "%d", IW_ConnNumDatasets(conn));
    IW_ChunkField(fp,"NUMBER_OF_GRANULES", "%d", IW_ConnNumGranules(conn));
    IW_ChunkField(fp,"STATUS_CODE",        "%d", IW_ConnStatusCode(conn));
    if (IW_ConnStage(conn) < 3) {
      /* no real status known */
    }
    else {
      IW_ChunkField(fp,"STATUS_CODE_COMMENT","%s", IW_ConnAnyStatusCmnt(conn));
    }
    IW_ChunkEnd(fp);
  }

  /* Done with statuses! Close file... */
  fclose(fp);

  /* ...and rename temp file to permanent: */
  if (rename(stat_tmpname, stat_realname) != 0) {
    return err("can't rename %s to %s", stat_tmpname, stat_realname);
  }
  
  /* Done! */
  return 0;  
}


/************************************************************
 * update_status
 ************************************************************
 * 
 * DESCRIPTION
 *    This function updates the IK data structure that holds the
 *    search status.  From this info, it tries to determine the
 *    current connection stage.
 *
 * RETURNS
 *    The new connection stage, or negative on error.
 *
 * REVISION HISTORY
 *    Eryq made an attempt to adapt this to imsclient.
 *
 *    October 21, 1996 (Sean Wright) - Major overhaul of data struct
 *    which contains the status of each search connection.
 *
 *    Friday 21 February 1997 (Eric Winter) - Cleaned up. Removed cases
 *    of "fall through" in switch statements.  */

static IW_Connstage
update_status(int i_connNum,        /* 0-based connection index */
		    char *ac_archiveName, /* DAAC name */
		    int status,           /* raw 0-7, from IC_GetStatus */
		    char *search,         /* INVENTORY_SEARCH, etc. */
		    int count,            /* result count? */
		    char *ac_messageId)   /* message id */
{
  /* This string holds the comment returned on the current connection: */
  char ac_comment[IK_MAX_DATA_CENTER_COMMENT_SZ + 1];
  
  /* This is a 1-character code representing the current search type: */
  char c_searchType;
  
  /* Misc. return value: */
  int ret = 0;
  
  /* Guess stage: */
  IW_Connstage guess_stage;
  
  /* Was the poller was previously invoked? */
  static int invoked = 0;




  /*-----------------------------------------------------------------------*/
  
  /* Validate input. */
  assert(ac_archiveName != NULL);
  assert(search != NULL);
  assert(count >= 0);
  assert(ac_messageId != NULL);
  
  /*-----------------------------------------------------------------------*/
  
  /* Convert the search type and DAAC name strings to lower-case. */
  IW_Strdowncase(search);
  IW_Strupcase(ac_archiveName);

  /* Save the first character of the search type string; use it as a
     character code representing the search type. */
/*   c_searchType = *search; */
  /* This was modified to determine search type based on the input action. */
  c_searchType = mySearchType;
  /* If this is the first pass, initialize arrays for bookkeeping: */
  if (!invoked) {
    (void) IK_StatusInitialize(c_searchType, IK_GetTotalConnections());
    invoked = 1;
  }
  
  
  /* Clear the comment string, to guarantee a non-garbage value: */
  ac_comment[0] = '\0';
  
  /*-----------------------------------------------------------------------*/
  
  /* Update the result count display for the current connection. */
  switch (c_searchType) {
  case 'i':      /* INVENTORY_SEARCH */
    IK_StatusPutDaac(c_searchType, i_connNum, ac_archiveName);
    IK_StatusPutCount(c_searchType, i_connNum, count);
    if (count > 0 && !IK_StatusGetResultsFlag(c_searchType)) {
      IK_StatusSetResultsFlag(c_searchType);
    }
    break;
  case 'd':      /* DIRECTORY_SEARCH */
    IK_StatusPutDaac(c_searchType, i_connNum, ac_archiveName);
    IK_StatusPutCount(c_searchType, i_connNum, count);
    if (count > 0 && !IK_StatusGetResultsFlag(c_searchType)) {
      IK_StatusSetResultsFlag(c_searchType);
    }
    break;
  case 'b':      /* BROWSE_REQUEST */
    if (count > 0 && !IK_StatusGetResultsFlag(c_searchType)) {
      IK_StatusSetResultsFlag(c_searchType);
    }
    IK_StatusPutId(c_searchType, i_connNum, ac_messageId);
    IK_StatusPutCount(c_searchType, i_connNum, count);
    break;
  case 'p':      /* PRODUCT_REQUEST */
    IK_StatusPutId(c_searchType, i_connNum, ac_messageId);
    IK_StatusPutCount(c_searchType, i_connNum, count);
    break;
  case 'o':      /* PRODUCT_STATUS_REQUEST */
    IK_StatusPutId(c_searchType, i_connNum, ac_messageId);
    IK_StatusPutCount(c_searchType, i_connNum, count);
    break;
  case 'a':      /* ACCOUNT_STATUS_REQUEST */
    IK_StatusPutId(c_searchType, i_connNum, ac_messageId);
    IK_StatusPutCount(c_searchType, i_connNum, count);
    break;
  case 'r':      /* PROFILE_RETRIEVAL_REQUEST */
    IK_StatusPutId(c_searchType, i_connNum, ac_messageId);
    IK_StatusPutCount(c_searchType, i_connNum, count);
    break;
  case 'u':      /* PROFILE_UPDATE_REQUEST */
    IK_StatusPutId(c_searchType, i_connNum, ac_messageId);
    IK_StatusPutCount(c_searchType, i_connNum, count);
    break;
  case 'e':      /* PRICE_ESTIMATE_REQUEST */
    IK_StatusPutId(c_searchType, i_connNum, ac_messageId);
    IK_StatusPutCount(c_searchType, i_connNum, count);
    break;
  case 's':      /* PROFILE_SUBMIT_REQUEST */
    IK_StatusPutId(c_searchType, i_connNum, ac_messageId);
    IK_StatusPutCount(c_searchType, i_connNum, count);
    break;
  case 'c':      /* USER_PASSWORD_CHANGE_REQUEST */
    IK_StatusPutId(c_searchType, i_connNum, ac_messageId);
    IK_StatusPutCount(c_searchType, i_connNum, count);
    break;
  }
  
  /*-----------------------------------------------------------------------*/

    /* Get the network manager error code. */
    ret = ((c_searchType == 'b') ? IK_Nmstatus(0) : IK_Nmstatus(i_connNum));

    /*-----------------------------------------------------------------------*/

    /* Update the status lights for the current connection. In
       general, check to see if the flag for a given light is set. If
       not, do nothing. If so, set the flag. If visible, and turn on
       the corresponding light. */

    switch (status) {

    case 1:          /* "connection open" */

	/* Event codes which give this status code: IK_CONNECT_OPEN */

	/* This connection has opened. If the open flag is not set,
           set it and turn on the Open Connection light. Otherwise, do
           nothing; since the flag is already set, the light should
           already be on. */
	IK_StatusSetOpen(c_searchType, i_connNum);
	break;

    case 2:          /* "sent query" */

	/* Event codes which give this status code: IK_TRANSMITTED */

	/* A request message has been sent over this connection. */

	/* Since we have sent data, the connection must be
           open. Therefore, check to see if the open flag is set. If
           not, set it and turn on the open light. */
	IK_StatusSetOpen(c_searchType, i_connNum);

	/* If the send flag is not set, set it and turn on the send light. */
	IK_StatusSetSend(c_searchType, i_connNum);
	break;

    case 3:          /* "receiving results" */

	/* Event codes which give this status code: IK_GOT_INVRESULT,
           IK_GOT_PRODRESULT, IK_GOT_FTPBROWSE, IK_GOT_MDDIFS,
           IK_GOT_INTBROWSE, IK_IMAGE_RECEIVED, IK_GCMD_SUCCESS,
           IK_INGESTINV_SUCCESS, IK_INGESTIMG_SUCCESS */

	/* A message has been received on this connection, or results
           have been ingested. */

	/* Since we have received data, the connection must be
           open. Therefore, check to see if the open flag is set. If
           not, set it and turn on the open light. */
	IK_StatusSetOpen(c_searchType, i_connNum);

	/* Since we have received data, we must have sent a search
           request. If the send flag is not set, set it and turn on
           the send light. */
	IK_StatusSetSend(c_searchType, i_connNum);

	/* If the receive flag is not set, set it and turn on the
           receive light. */
	IK_StatusSetReceive(c_searchType, i_connNum);

	/* Check to see of the problem flag is set. If so, turn on the
           problem light. */
	if (IK_StatusGetProblem(c_searchType, i_connNum)) {
	  /* ??? */
	}
	break;

    case 4:          /* "success" */

	/* Event codes which give this status code: IK_GOT_QUIT,
           IK_IMAGE_COMPLETE, IK_INGESTFTPB_SUCCESS,
           IK_INGESTPR_SUCCESS, IK_INGESTGCMD_SUCCESS */

	/* This search connection is complete. This case arises when
           the client receives a QUIT, or an integrated browse image
           is complete, or an FTP browse, order response, or GCMD
           response has been ingested. */

	/* Since the search has ended successfully, the connection
           must be open. Therefore, check to see if the open flag is
           set. If not, set it and turn on the open light. */
	IK_StatusSetOpen(c_searchType, i_connNum);

	/* Since the search has ended successfully, we must have sent
           a search request. If the send flag is not set, set it and
           turn on the send light. */
	IK_StatusSetSend(c_searchType, i_connNum);

	/* Since the search has ended successfully, we must have
           received results. If the receive flag is not set, set it
           and turn on the receive light. */
	IK_StatusSetReceive(c_searchType, i_connNum);

	/* Check to see if the target archive is down. If not, fill in
           the result counts and set the complete flag. */
	if (!IK_StatusGetDown(c_searchType, i_connNum)) {
	  
	  /* Inventory or directory search needs results to be ok;
	   * others do not: */
	  if (c_searchType == 'i' || c_searchType == 'd' ) {
	    if (IK_StatusGetCount(c_searchType, i_connNum) >= 1) {
	      IK_StatusSetComplete(c_searchType, i_connNum);
	    }
	  }
	  else {      /* not inv/dir search; just ok! */
	    IK_StatusSetComplete(c_searchType,i_connNum);
	  }
	}
	break;

    case 5:          /* "aborted" */

	/* Event codes which give this status code: IK_ABORT_SEARCH */

	/* The search has been aborted. */

	/* If the abort flag is not set, set it and turn on the abort
           light. Otherwise, do nothing; if the abort flag is already
           set, the abort light should be on. */
	IK_StatusSetAbort(c_searchType, i_connNum);
	break;

    case 6:          /* "archive down" */

	/* Event codes which give this status code: IK_ARCHIVE_DOWN */

	/* The connection has failed. */
	ret = NM_NOSRVR_CONN;

	/* Set the terminate flag. */
	IK_StatusSetTerminate(c_searchType, i_connNum);

	/* If the DAAC down flag is not set, set it and turn on the
           DAAC down light. Otherwise, do nothin; if the flag is set,
           the light is already on. */
	IK_StatusSetDown(c_searchType, i_connNum);
	break;

    case 7:          /* "client error" */

	/* Event codes which give this status code: IK_CLIENT_CRASH,
           IK_CLIENTSYS_ERROR, IK_GOT_ABORT */

	/* There is a problem with the connection. */

	/* Match the Network Manager result code to a comment string
           for the user to view. */
	IK_GetCommunicationComment(ret, ac_comment);
/*	IK_StatusPutComment(c_searchType, i_connNum, ac_comment); */
	IK_StatusPutDataCenterComment(c_searchType, i_connNum, ac_comment);

	/* If the problem flag is not set, set it and turn on the
           problem light. Otherwise, do nothing; if the flag is set,
           the light is on. Note that the "problem" light really
           should be called the "comment" light. */
	IK_StatusSetProblem(c_searchType, i_connNum);
	break;

    case 8:          /* "connection terminated" */

	/* Event codes which give this status code: IK_TERMINATED */

	/* The connection has terminated. */

	/* Check to see if the connection never actually got
           underway. If it did, check if the terminate flag is not
           set, set it, and activate any appropriate lights. If not,
           light up the Connection Failed light. */
	if (ret == NM_NOSRVR_CONN) {
	    if (!IK_StatusGetDown(c_searchType, i_connNum)) {
		IK_StatusSetDown(c_searchType, i_connNum);
	    }
	}

	/* If the terminate flag is not set, set it, and activate any
           appropriate lights. If not, light up the Connection Failed
           light. */
	if (!IK_StatusGetTerminate(c_searchType, i_connNum)) {
	    IK_StatusSetTerminate(c_searchType, i_connNum);

	    /* If the search has not been aborted, and any good
               results were returned, then it completed successfully,
               so set the completed flag and turn on the completed
               light. */
	    if (!IK_StatusGetAbort(c_searchType, i_connNum)) {
		IK_GetCommunicationComment(ret, ac_comment);
		IK_StatusPutDataCenterComment(c_searchType, i_connNum, ac_comment);
		if (ret == NM_MATCH || (ret == 0)) {
		    if (!IK_StatusGetReceive(c_searchType, i_connNum)) {
			IK_StatusSetReceive(c_searchType, i_connNum);
		    }

		    /* If we do not see a DAAC is Down status set as complete */
                    if(!IK_StatusGetDown(c_searchType, i_connNum)) {
		       IK_StatusSetComplete(c_searchType, i_connNum);
                    }
		}
		else {
		    IK_StatusSetProblem(c_searchType, i_connNum);
		}
	    }
	}
	break;

    default:

	/* Event codes which give this status code:
           IK_CONNECT_IN_PROGRESS, IK_CONNECT_FAILED,
           IK_CONNECT_CLOSED, IK_GOT_NODATA, IK_GOT_ACK,
           IK_GCMD_AMBIGUOUS */

	/* Nothing has changed. */
	break;
    }



    /* 
     * Based on the activated lights, bump the stage forward...
     */

    guess_stage = IW_ConnS_PENDING;

    /* Bump forward through non-terminal stages... */
    if (IK_StatusGetOpen(c_searchType, i_connNum)) {
      guess_stage = MAX_NUM(guess_stage, IW_ConnS_CONNECTED);
    }
    if (IK_StatusGetSend(c_searchType, i_connNum)) {
      guess_stage = MAX_NUM(guess_stage, IW_ConnS_SENDING);
    }
    if (IK_StatusGetReceive(c_searchType, i_connNum)) {
      guess_stage = MAX_NUM(guess_stage, IW_ConnS_RECEIVING);
    }

    /* Decide on at most one terminal stage... */
    if (IK_StatusGetComplete(c_searchType, i_connNum)) {
      guess_stage = MAX_NUM(guess_stage, IW_ConnS_OK);
    }
    else if (IK_StatusGetDown(c_searchType, i_connNum)) {
      guess_stage = MAX_NUM(guess_stage, IW_ConnS_EDAACDOWN);
    }
    else if (IK_StatusGetTerminate(c_searchType, i_connNum)) {
      guess_stage = MAX_NUM(guess_stage, IW_ConnS_EFAILED);
    }
    else {
      /* we haven't reached a terminal stage */
    }

    /* Done! */
    return guess_stage;
}
   
/************************************************************
 * update_conn_stage
 ************************************************************
 *
 * DESCRIPTION
 *    Given a connection structure and its connection number, update and
 *    return the new current stage for the connection, one of the 
 *    IW_ConnS* constants.
 *
 * RETURNS
 *    Non-negative (the stage) on success, negative on error.
 */ 

static IW_Connstage 
update_conn_stage(IW_Conn *conn,    /* the connection */
		  int i_connNum)    /* its index in the connection array */
{
    /* This buffer is used to construct error messages. */
    char ac_messageId[IK_MESSAGE_ID_SZ];

    /* This holds the most recent event code for the current connection: */
    IK_EventCode eventCode;

    /* This holds the code for the current search type. */
    IK_SEARCH_TYPE searchType;

    /* This string holds the string name of the current archive. */
    char ac_archiveName[IK_ARCHIVE_ACRONYM_SZ];

    /* The count of current "results" received - granules, datasets, etc.: */
    int i_resultCount = 0;

    /* This is the current "status code" for the current connection;
       it is used to determine which light(s) to activate on the
       communication status screen. */
    int i_connectionStatus;

    /* A representation of the current search type. */
    char ac_searchTypeName[IK_MESSAGE_ID_SZ];

    /* This is the current status of the Network Manager. */
    int i_networkManagerStatus;

    /* The old and new stages: */
    IW_Connstage old_stage = -1;
    IW_Connstage new_stage = -1;


    /*------------------------------------------------------------*/

    /* Fetch the most recent event code for the current connection: */
    eventCode = IK_GetConnectionStatus((i_connNum + 1),
				       &searchType, 
				       ac_archiveName,
				       ac_messageId);
    if (eventCode == -1) {
      return -1;
    }

    /* Get the current count of results returned: */
    if (searchType == IK_MSD_SEARCH || searchType == IK_GCMD_SEARCH) {
      i_resultCount = IW_ConnNumDatasets(conn);
    } 
    else if (searchType == IK_PR_SEARCH) {
      i_resultCount = 0;
    } 
    else {
      i_resultCount = IW_ConnNumGranules(conn);
    }

    /* Calculate the status code for the current connection, so it
     * can be used to determine which lights to turn on for this
     * connection on the communication status screen.  This is 
     * a raw version of our "LED stage".
     */
    i_connectionStatus = IC_GetStatus(searchType, eventCode);
    
    /* Get the name of the current search type. */
    IC_GetSearch(searchType, ac_searchTypeName);
    assert(tolower(ac_searchTypeName[0]));

    /* Update the status display for this connection. */
    old_stage = IW_ConnStage(conn);
    new_stage = update_status(i_connNum, 
			      ac_archiveName, 
			      i_connectionStatus,
			      ac_searchTypeName, 
			      i_resultCount,
			      ac_messageId);
    if (new_stage >= 0) {
      notice("LED: Transition from %s to %s",
	     IW_ConnstageName(old_stage), IW_ConnstageName(new_stage));
      IW_ConnSetStage(conn, new_stage);
    }
    return new_stage;
}

/*************************************************************
 * invoke_poller
 *************************************************************
 *
 * DESCRIPTION
 *    This function is called by the main program to poll all the open
 *    connections.  It will check the current status of the each connection
 *    and update the status of every connection. It will also output the
 *    current search status into a file name "search.stat" so that the
 *    the Web Gateway can do an update.  Based on the status of all the 
 *    connections, it will also determine whether the whole search is 
 *    completed or not.
 *
 * PARAMETERS
 *    int  num_conn         total number of connections
 *
 * NOTES
 *    Notice that all entries in the "Conns" hash are created before 
 *    the "client.stat" file is output.  This is *very * important.
 *
 * RETURN VALUES
 *    type: int
 *    Return: Return 0 while search has completed
 *            Return 1 while search has not yet completed
 *            Return -1 while search ended with error 
 *
 * GLOBAL VARIABLES USED
 *    GaeaTmpDir
 *  
 */
static int invoke_poller(int num_conn)
{
  /*** Per-connection info: ***/
  IW_Conn *conn;               /* the current connection to a DAAC */
  char arch_id[PATH_LEN];      /* data center id */
  char message_id[PATH_LEN];   /* message id */
  int  event;                  /* event code: an IK_XXX constant */
  char status_string[PATH_LEN];/* status string */
  char status_code_cmnt[MAX_COMMENT]; 
                               /* search status string (mapped from NM_XXX) */
  
  /*** Misc: ***/
  int i;                       /* connection index */
  int ret;                     /* return value */
  IK_SEARCH_TYPE search_type;  /* search context */
  int n_slots;



  /* 
   * DO THE STATE TRANSITION... 
   *
   *    As a side-effect, reply_received() will be called whenever
   *    we get a tree.  This will cause the STATUS_CODE_COMMENT of
   *    each connection to be updated.
   */

  /* Get and log the number of slots: */
  n_slots = IK_GetNMSlots();
  debug("Invoked poller, num_conn = %d, %d slots in Network Manager", 
	num_conn, n_slots);
  
  /* Cause state transition function to advance states of all connections: */
  if ((ret = IK_CauseStateTransition()) == -1) { 
    return err("State Tables corrupted\n");
  }
   
 

  /* 
   * OUTPUT ANY QUEUED TREES...
   *
   *    During the state transition, we may have queued up trees that
   *    we received, for processing *after* normal state-machine activities
   *    (like ingest).  Handle them now by invoking report_tree().
   */
  if (QueueMessages) {
    AGGREGATE msg_agg;    /* a queued message */
    IW_Chunksp *sp;       /* the chunk-output stream */
    int i;

    /* Setup the chunk-output stream: */
    sp = IW_ChunkspNew(ObjFp);

    /* For each message tree, output it: */
    for (i = 0; i < IW_ArraySize(MsgQueue); i++) {
      if (!(msg_agg = IW_ArrayAt(MsgQueue, i))) {
	err("WHOA: queued a NULL tree! continuing...");
	continue;
      }
      debug("reporting queued tree %d", i);
      if (report_tree(msg_agg, sp) < 0) {
	err("couldn't report queued tree; continuing...");
      }
    }

    /* Close the chunk-output stream: */
    IW_ChunkspFree(sp);    

    /* Empty the queue, discarding all message trees: */
    IW_ArraySetSize(MsgQueue, 0);
  }



  /* 
   * UPDATE STATUS INFO FOR EACH CONNECTION... 
   */

  for (i = 1; i < (num_conn + 1); i++) {
    int new_status_code;
    
    /* Get "event" status for the connection (including the DATA_CENTER_ID): */
    event = IK_GetConnectionStatus(i, &search_type, arch_id, message_id);
    debug("%-9s: ConnNum    %d", arch_id, i);
    debug("      SearchType %d", search_type);
    debug("      MessageId  %s", message_id);
    debug("      Event/ConnStatus %d (an IK_Nm.h constant)", event);
     
    /* Get the object for this connection (creating it if necessary): */
    if (!(conn = lookup_conn(arch_id))) {
      puke("invoke_poller: unable to lookup %s!", arch_id);
    }

    /* Get/save the STATUS_CODE (an IK_Nmsts.h NM_XXX constant):
     * If the code matches the last one that we extracted from a
     * tree, then keep the status comment (which might be detailed);
     * else, discard the status comment.
     */
    new_status_code = IK_Nmstatus(i-1);  /* assign NEW conn status */
    debug("%-9s: StatusCode %d (an IK_Nmsts.h NM_XXX constant)", 
	  arch_id, new_status_code);
    if(new_status_code == 0) new_status_code = 1;
    if (new_status_code != IW_ConnStatusCode(conn)) {
/*      IW_ConnSetStatusCmnt(conn, NULL); */
    }
    if (new_status_code > IW_ConnStatusCode(conn)) {
      IW_ConnSetStatusCode(conn, new_status_code);
    }

    /* Get/save the LED_STAGE code based on everything we know so far!!! */
    status_string[0] = '\0';
    debug("%-9s:   OldLEDStage    %d ",  arch_id, IW_ConnStage(conn));
    update_conn_stage(conn, (i-1));
    debug("%-9s:   NewLEDStage    %d %s",arch_id, IW_ConnStage(conn), 
	  status_string);    
  }

  
  /* 
   * OUTPUT STATUS INFO AND CHECK SEARCH STATUS...
   */

  /* Output the status info: */
  (void)write_status();

  /* Check the status of the whole search, and return appropriately: */
  ret = IK_EvaluateNetmgrStatus();   
  debug("netmgr val = %d", ret);
  switch (ret) {
  case 0:     /* search completed */
    return 0;
  case 1:     /* search not yet completed */
    return 1;
  case -1:    /* errors occur */
  default:
    return err("State tables corrupted: netmgr status = %d", ret);
  }
  /* NOTREACHED */
}




/******************************
 *
 * The following deal with tallying 
 * DATASET groups in the callback...
 *
 ******************************/


/************************************************************
 * dataset_is_countable
 ************************************************************
 *
 * DESCRIPTION
 *    We just got a result tree which contains a DATASET
 *    group.  Does it look like a dataset we should tally?
 */

static int
dataset_is_countable(AGGREGATE dset_agg,      /* the dataset aggregate */
		     int tree_type)           /* what kind of tree was this? */
{
  AGGREGATE gran_agg;

  /* Is it there? */
  if (!dset_agg) return 0;

/** Return true even if data set doesn't have any granules
 *
 * * Look for meaningful information, based on tree type: *
 * switch (tree_type) {
 * case IK_INVRESULT:        * an inventory result... *
 *
 *   * It had better have at least one granule! *
 *   if (!(gran_agg = FindGroup(dset_agg, "GRANULE"))) {
 *     return 0;
 *   }
 *   break;
 * }
 */

  /* Good! */
  return 1;
}


/************************************************************
 * GCMD_info
 ************************************************************
 *
 * DESCRIPTION
 *    Get the GCMD info for the given dataset, as a new aggregate.
 *    Return the AGGREGATE, or NULL on error.
 */

static AGGREGATE GCMD_info(char *web_dskey)      /* the dataset web key */ 
{
  IK_Dataset *dataset;
  IK_SortKey *dskey;
  AGGREGATE gcmd_agg;

  /* Get sort key: */
  if (!(dskey = IW_DatasetKeyFromWeb(web_dskey))) {
    err("no sort key for %s", web_dskey); return NULL;
  }

  /* Get dataset: */
  if (IK_GetDataset(dskey, &dataset) != 0) {
    free(dskey);
    err("couldn't get dataset %s (not ingested yet?)", web_dskey); return NULL;
  }
  free(dskey);

  /* Create aggregate node: */
  gcmd_agg = NewAggregate(NULL, KA_GROUP, "GCMD_INFO", NULL);

  /* Add info to the aggregate: */
  if (dataset->md_start_time) {
    IW_ODLAddStringParam(gcmd_agg, "MD_START_TIME", dataset->md_start_time);   
  }
  if (dataset->md_stop_time) {
    IW_ODLAddStringParam(gcmd_agg, "MD_STOP_TIME", dataset->md_stop_time);
  }
  if (dataset->md_title) {
    IW_ODLAddStringParam(gcmd_agg, "MD_TITLE", dataset->md_title);
  }
  if (dataset->md_attribute) {
    IW_ODLAddStringParam(gcmd_agg, "MD_ATTRIBUTE", dataset->md_attribute);
  }
  if (dataset->md_reference) {
    IW_ODLAddStringParam(gcmd_agg, "MD_REFERENCE", dataset->md_reference); 
  }
  if (dataset->md_personnel) {
    IW_ODLAddStringParam(gcmd_agg, "MD_PERSONNEL", dataset->md_personnel); 
  }
  if (dataset->md_archive_info) {
    IW_ODLAddStringParam(gcmd_agg, "MD_ARCHIVE_INFO",dataset->md_archive_info);
  }
  
  /* Cleanup and return: */
  free(dataset);
  return gcmd_agg;
}


/************************************************************
 * fixup_dataset
 ************************************************************
 *
 * DESCRIPTION
 *    We just got a result tree which contains a DATASET
 *    group, and we want to diddle with it a little and
 *    also log some information.
 */

static int
fixup_dataset(AGGREGATE root_agg,       /* copy of the tree we just got */
	       int tree_type,            /* what kind of tree was this? */
	       AGGREGATE dset_agg)       /* the dataset aggregate */
{
  AGGREGATE msg_agg;      /* archive message aggregate */
  VALUE value;            /* temp value */
  char *arch_id;                      /* archive id */
  char *dset_id;                      /* dataset id */
  char web_dskey[IW_Key_MAX_DSKEY+1];   /* dataset/granule web key */
  IW_Conn *conn;          /* connection status */



  /*
   * ADD BASIC INFO...
   */

  /* Get related aggregates: */
  if (!(msg_agg = ParentAggregate(dset_agg))) { 
    return err("DATASET has no parent agg!"); 
  }
  
  /* Get the archive id: */    
  if (!(value = FirstValue(FindParameter(msg_agg, "DATA_CENTER_ID")))) {
    return err("DATASET's data center has no data center id!");
  }
  arch_id = IW_ODLValueString(value);
  
  /* Get the dataset id: */
  if (!(value = FirstValue(FindParameter(dset_agg, "DATASET_ID")))) {
    return err("DATASET has no data set id!"); 
  }
  dset_id = IW_ODLValueString(value);
  
  /* Map the ids to the dataset web key: */
  if (!(IW_DatasetIdPartsToWeb(arch_id, dset_id, web_dskey))) {
    return err("couldn't map %s/%s to web key", arch_id, dset_id); 
  }
  debug("Mapped %s/%s to %s", arch_id, dset_id, web_dskey);

  /* Add the web key parameter: */
  (void)IW_ODLAddStringParam(dset_agg, "DATASET_WEB_KEY", web_dskey);

  /* Log the web key to the list: */
  if (ListFp) {
    fprintf(ListFp, "DATASET_WEB_KEY: %s\n", web_dskey);
  }


  /*
   * ADD GCMD INFO...
   */
  if ((Action == ACTION_DIR_SEARCH) && GCMDSearch) {
    AGGREGATE gcmd_agg;     /* GCMD info (if any) */

    /* Is there any? */
    debug("Trying to get GCMD info for %s", web_dskey);
    if (gcmd_agg = GCMD_info(web_dskey)) {
      PasteAggregate(dset_agg, gcmd_agg);
    }
  }


  /*
   * TALLY THE DATASET...
   */

  /* Get/create the connection status object for this tree: */
  conn = lookup_conn(arch_id);

  /* If dataset looks legit (it is non-empty), 
   * and if we haven't seen it yet, then tally it: */
  if (dataset_is_countable(dset_agg, tree_type)) {
    IW_ConnAddDataset(conn, web_dskey);
    debug("Tallied dataset %s (now have %d datasets, %d granules)",
	  web_dskey, IW_ConnNumDatasets(conn), IW_ConnNumGranules(conn));
  }
  else {
    warning("Not counting dataset %s/%s (probably empty)", arch_id, dset_id);
  }

  /* Done! */
  return 0;
}



/******************************
 *
 * The following deal with tallying 
 * GRANULE groups in the callback...
 *
 ******************************/


/************************************************************
 * package_exists
 ************************************************************
 * Does the given package exist our valids (static or dynamic)?
 *
 * Note: I just lifted this algorithm from imsgdbdump; as I look at
 * it, it seems very inefficient: it might open a file and read an
 * ODL label for every granule!  We need an IW_PkgExists() which
 * is fast, easy, and does the same thing; there's no need to read
 * in the ODL label... we just want to know that it's there.  [Eryq]
 */

int package_exists(char *arch_id, char *dset_id, char *pkg_id)
{
  int RETURN = 0;
  AGGREGATE pkg_agg = NULL;
  IK_ArchiveTag *atag_p = NULL;
  IK_DatasetTag *dtag_p = NULL;


  debug("Checking package (arch=%s, dset=%s, pkid=%s", 
	arch_id, dset_id, pkg_id);

  /* Map ids to tags: */
  if (!(atag_p = IK_MapArchiveNameToTag(arch_id))) {
    err("Can't map %s to an atag", arch_id); goto CLEANUP;
  }
  if (!(dtag_p = IK_MapDatasetNameToTag(dset_id))) {
    err("Can't map %s to a dstag", dset_id); goto CLEANUP;
  }

  /* Package of '*' means dataset: */
  pkg_id = (IW_Strequal(pkg_id, "*") ? dset_id : pkg_id);  
  
  /* Create the root node: */
  if (!(pkg_agg = NewAggregate(NULL, KA_GROUP, "PACKAGE", NULL))) {
    err("Can't allocate aggregate"); goto CLEANUP;
  }
  
  /* Get the package: */
  if (IK_GetPackage(*atag_p, *dtag_p, pkg_id, &pkg_agg) < 0) {
    err("No such package: arch %s, dset %s, pkg %s", arch_id, dset_id, pkg_id);
    goto CLEANUP;
  }
  
  /* Done! */
  debug("Package verified");
  RETURN = 1;


  /*** Cleanup and return: ******************************/
CLEANUP:
  if (atag_p) free(atag_p);
  if (dtag_p) free(dtag_p);
  if (pkg_agg) RemoveAggregate(pkg_agg);  
  return (RETURN);
}


/************************************************************
 * fixup_granule
 ************************************************************
 *
 * DESCRIPTION
 *    We just got a result tree which contains a GRANULE
 *    group, and we want to diddle with it a little and
 *    also log some information.
 *
 * NOTES
 *    When looking at an output granule, be aware that the 
 *    following parameters default from the dataset level:
 *
 *        BROWSE_PRODUCT_DESCRIPTION
 *        COMMENT
 *        DAY_NIGHT
 *        GLOBAL_GRANULE
 *        MD_ENTRY_ID
 *        PACKAGE_ID
 *
 *        CAMPAIGN
 *        PARAMETER
 *        PROCESSING_LEVEL
 *        SENSOR_NAME
 *        SOURCE_NAME
 */

static int
fixup_granule(AGGREGATE root_agg,    /* copy of the tree we just got */
	       int tree_type,         /* what kind of tree was this? */
	       AGGREGATE gran_agg)    /* the granule aggregate */
{
  AGGREGATE msg_agg;      /* archive message aggregate */
  AGGREGATE dset_agg;     /* dataset aggregate */
  PARAMETER parm;         /* valid package parameter */
  VALUE value;            /* temp value */
  char *arch_id;                     /* archive id */
  char *dset_id;                     /* dataset id */
  char *gran_id;                     /* granule id */
  char web_gkey[IW_Key_MAX_GKEY+1];   /* granule web key */
  IW_Conn *conn;          /* connection status */


  /*
   * GET RELATED AGGREGATES...
   */

  /* Get related aggregates: */
  if (!(dset_agg = ParentAggregate(gran_agg))) { 
    return err("GRANULE has no parent agg!"); 
  }
  if (!(msg_agg = ParentAggregate(dset_agg))) { 
    return err("GRANULE's dataset has no parent agg!"); 
  }


  /*
   * ADD THE GRANULE WEB KEY...
   */
  
  /* Get the archive id: */    
  if (!(value = FirstValue(FindParameter(msg_agg, "DATA_CENTER_ID")))) {
    return err("GRANULE's data center has no data center id!");
  }
  arch_id = IW_ODLValueString(value);
  
  /* Get the dataset id: */
  if (!(value = FirstValue(FindParameter(dset_agg, "DATASET_ID")))) {
    return err("GRANULE's dataset has no data set id!"); 
  }
  dset_id = IW_ODLValueString(value);

  /* Get the granule id: */
  if (!(value = FirstValue(FindParameter(gran_agg, "GRANULE_ID")))) {
    return err("GRANULE has no granule id!"); 
  }
  gran_id = IW_ODLValueString(value);
 
  /* Map the ids to the dataset web key: */
  if (!(IW_GranuleIdPartsToWeb(arch_id, dset_id, gran_id, web_gkey))) {
    return err("couldn't map %s/%s to web key", arch_id, dset_id); 
  }
  
  /* Add the web key parameter: */
  (void)IW_ODLAddStringParam(gran_agg, "GRANULE_WEB_KEY", web_gkey);

  /* Log the web key to the list: */
  if (ListFp) {
    fprintf(ListFp, "GRANULE_WEB_KEY: %s\n", web_gkey);
  }



  /*
   * ADD THE SEQUENCE OF *VALID* PACKAGE IDS (ones we have pkg info for):
   */

  /* Valid package ids (needed for ordering capability): */
  if (parm = FindParameter(gran_agg, "PACKAGE_ID")) {
    char *pkg_id;           /* unfiltered package id */
    AGGREGATE pkg_agg;      /* aggregate holds the package */
    PARAMETER vid_parm;     /* valid package ids */
    VALUE     id_val;

    /* Get a place to put the valid package ids: */
    vid_parm = NewParameter(gran_agg, KP_ATTRIBUTE, "VALID_PACKAGE_ID");
    vid_parm->value_kind = KV_SEQUENCE;

    /* For each package id: */
    for (id_val = FirstValue(parm); id_val; id_val = NextValue(id_val)) {
      pkg_id = IW_ODLValueStringified(id_val);

      /* If it exists, copy it into the new parameter: */
      if (package_exists(arch_id, dset_id, pkg_id)) {
	PasteValue(vid_parm, CopyValue(id_val));
	vid_parm->rows = 1;
      }
    }
  }


  /*
   * TALLY THE GRANULE...
   */

  /* Get/create the connection status object for this tree: */
  conn = lookup_conn(arch_id);

  /* If we haven't seen it yet, then tally it: */
  IW_ConnAddGranule(conn, web_gkey);
  debug("Tallied granule %s (now have %d datasets, %d granules)",
	web_gkey, IW_ConnNumDatasets(conn), IW_ConnNumGranules(conn));

  /* Done! */
  return 0;
}



/******************************
 *
 * The following deal with other 
 * callback activities...
 *
 ******************************/


/************************************************************
 * report_tree
 ************************************************************
 *
 * DESCRIPTION
 *    Output the tree as chunks to the object file, after first 
 *    performing some fixup and bookkeeping.
 */

static int
report_tree(AGGREGATE root_agg,  /* copy of tree just received */	
	    IW_Chunksp *sp)      /* where to output chunks to */
{
  int RESULT = 0;         /* what to return */

  /*** Processing ODL... ***/
  IK_ArgCode tree_type;   /* the type of the tree we received */
  AGGREGATE msg_agg = NULL;  /* archive message aggregate */
  AGGREGATE dset_agg = NULL;  /* dataset aggregate */
  AGGREGATE gran_agg = NULL;  /* granule aggregate */
  VALUE value;                /* temp value */
  char *group_name;           /* current group */
  char *arch_id;

  /*** Outputting chunks... ***/
  IW_ODLWriteOpts opts;   /* output options */

  /*** Misc... ***/
  IW_Conn *conn;          /* connection status */


  
  debug("Reporting tree");


  /*
   * SETUP A COPY OF THE TREE, AND OUR CONNECTION-STATUS INFO...
   */
  
  /* Get the tree type: */
  tree_type = IK_IdentifyTree(root_agg);

  /* Get message aggregate: */
  msg_agg = NextGroup(root_agg);

  /* Get the archive id: */    
  if (!(value = FirstValue(FindParameter(msg_agg, "DATA_CENTER_ID")))) {
    return err("msg %s has no data center id!", msg_agg->name);
  }
  arch_id = IW_ODLValueString(value); 
  
  /* Get the connection status object for this archive: */
  conn = lookup_conn(arch_id);



  /* 
   * TALLY (AND TWEAK) ALL DATASET GROUPS IN THE TREE...
   */

  /* For each DATASET group in this tree... */
  group_name = "DATASET";
  for (dset_agg = FindGroup(root_agg, group_name);
       dset_agg;
       dset_agg = FindNextGroup(root_agg, dset_agg, group_name)) {   

    /* Tally and tweak it: */
    if (fixup_dataset(root_agg, tree_type, dset_agg) < 0) {
      warning("couldn't process dataset");
    }    
  }


  /* 
   * TALLY (AND TWEAK) ALL GRANULE GROUPS IN THE TREE...
   */

  /* For each GRANULE group in this tree... */
  group_name = "GRANULE";
  for (gran_agg = FindGroup(root_agg, group_name);
       gran_agg;
       gran_agg = FindNextGroup(root_agg, gran_agg, group_name)) {   
    PARAMETER parm;         /* valid package parameter */

    /* Tally and tweak it: */
    if (fixup_granule(root_agg, tree_type, gran_agg) < 0) {
      warning("couldn't process granule");
    }    
  }


  /*
   * OUTPUT THE ODL TREE TO THE "OBJECT" CHUNK STREAM...
   */
  
  /* Set options, based on message type: */
  IW_ODLWriteOptsClear(&opts);
  switch (tree_type) {
  case IK_INVRESULT:        /* an inventory result... */
  case IK_MDDIF:            /* a DIF tree... */
    opts.except_groups = "MONITOR|VERSION|PACKAGE";
    break;
  default:
    opts.except_groups = "MONITOR|VERSION";  
    break;
  }

  /* Output the ODL tree, as chunks: */
  debug("OUTPUT!");
  IW_ODLWriteChunk(sp, root_agg, &opts);
#if OUTPUT_CLIENT_ODL
  WriteLabel(OdlFp, root_agg);
#endif

  /*** Done! ******************************/
CLEANUP:
  /* no-op */
  return RESULT;
}


/************************************************************
 * reply_received
 ************************************************************
 *
 * DESCRIPTION
 *    Perform this action when we receive an ODL tree on a
 *    connection.  
 */

static void
reply_received(CONNECTION_INFO *conn_info,   /* connection info structure */
	       AGGREGATE origroot_agg,       /* tree just received */
	       int conn_i)                   /* connection index, 0-based */ 
{
  AGGREGATE root_agg = NULL;  /* copy of top-level aggregate */
  AGGREGATE msg_agg;     /* message aggregate */
  AGGREGATE agg = NULL;  /* temp aggregate */
  AGGREGATE package_grp; /* package_grp is the current "PACKAGE" group. */
  PARAMETER parm;        /* temp parameter */
  VALUE value;           /* temp value */
  char *arch_id;         /* archive id */
  char *str;             /* scratch string */
  IW_Conn *conn;         /* current connection info */

  /* Status stuff: */
  int tree_type;         /* type of this tree */
  int quit_tree = 0;     /* QUIT or ABORT tree */
  int new_status_code;
  char new_status_cmnt[MAX_COMMENT];

  /*
   * SETUP A COPY OF THE TREE, AND OUR CONNECTION-STATUS INFO...
   */

  /* Get the archive: */
  arch_id = conn_info->archive_name;
  
  /* Sanity check: */
  if (!(origroot_agg)) {   err("No tree!"); goto CLEANUP;   }
  if (!(conn_info))    {   err("No connection!"); goto CLEANUP;   }
  if (origroot_agg && (agg = NextAggregate(origroot_agg))) {
    debug("Received '%s' tree from %s", agg->name, arch_id);
  }

  /* Is this a tree we care about?  If not, just return: */
  tree_type = IK_IdentifyTree(origroot_agg);
  switch (tree_type) {
  case IK_INVRESULT:        /* INVENTORY_RESULT */
  case IK_MDDIF:            /* DIRECTORY_RESULT */
  case IK_FTPBROWSE:        /* FTP_BROWSE_RESULT */
  case IK_INTBROWSE:        /* INTEGRATED_BROWSE_RESULT */
  case IK_PRODRESULT:       /* PRODUCT_RESULT */    
  case IK_POSTATUS:         /* PRODUCT_STATUS_INFO */
  case IK_ACCTSTATUS:       /* ACCOUNT_STATUS_RESULT */
  case IK_PROFRETRIEVE:     /* PROFILE_RETRIEVAL_RESULT */
  case IK_PROFUPDATE:       /* PROFILE_UPDATE_RESULT */
  case IK_PRICEESTIMATE:    /* PRICE_ESTIMATE_RESULT */
  case IK_PROFILESUBMIT:    /* PROFILE_SUBMIT_RESULT */
  case IK_PASSWORDCHANGE:   /* USER_PASSWORD_CHANGE_RESULT */
    break;                      /* no-op */
  case IK_QUIT:
  case IK_ABORT:
    quit_tree = 1;
    break;
  case IK_XX:               /* unknown */
  default:                  /* other */
    goto CLEANUP;               /* so long, farewell... */
  }

  /*
   * DO SOME BASIC BOOKKEEPING...
   */
  
  /* Get the connection status object for this archive: */
  conn = lookup_conn(arch_id);

  /* If there's no message aggregate, quit now: */
  if (!(msg_agg = NextAggregate(origroot_agg))) {
    warning("No message aggregate; quitting"); return;
  }

  switch (tree_type) {
  case IK_MDDIF:            /* DIRECTORY_RESULT */
    for (IK_ImsErrno = 0, package_grp = FindGroup(msg_agg, "PACKAGE");
     package_grp && !IK_ImsErrno;
     package_grp = FindGroup(msg_agg, "PACKAGE")) {
      IK_IngestPackage(package_grp);
      (void) RemoveAggregate(package_grp);
    }
  }
  


  /* Try to get the STATUS_CODE and STATUS_CODE_COMMENT from this tree, 
   * and store them in our connection info.
   * 
   * Careful! STATUS_CODE and STATUS_CODE_COMMENT are sent back in
   * QUIT trees as well... make sure we don't have bogus information overwrite
   * the information we got previously! 
   */

  /* Get STATUS_CODE (or NM_ABNORMAL if value is missing/suspicious): */
  new_status_code = NM_ABNORMAL;
  if (value = FirstValue(FindParameter(msg_agg, "STATUS_CODE"))) {
    
    /* Our value is SUPPOSED to be an integer, but we tolerate strings: */
    switch (value->item.type) {
    case TV_INTEGER: new_status_code = value->item.value.integer.number; break;
    case TV_STRING:  new_status_code = atoi(value->item.value.string); break;
    }
    
    /* Can we trust this value?  If not, reset it to abnormal: */
    if (quit_tree &&                        /* a QUIT or ABORT tree... */
	(new_status_code == NM_NOERR) &&    /* which claims "no error"... */
	(IW_ConnStatusCode(conn) != NM_NOERR)) { /* but we remember an err! */
      new_status_code = NM_ABNORMAL;
    }
  }                                          /* get STATUS_CODE */
  if(new_status_code == 0) new_status_code = 1;
  /* Get STATUS_CODE_COMMENT: */
  new_status_cmnt[0] = '\0';
  /* This line was added to convert a STATUS_CODE to a STATUS_CODE_COMMENT */
  if(new_status_code > 0) 
   IK_GetCommunicationComment(new_status_code, new_status_cmnt);
  if ((parm = FindParameter(msg_agg, "STATUS_CODE_COMMENT")) &&
      (str = IW_ODLParamStringified(parm, "\n"))) {
    
    /* Copy status code comment: */
    strncpy(new_status_cmnt, str, MAX_COMMENT); 
    new_status_cmnt[MAX_COMMENT-1] = '\0';
    free(str);
  }                                          /* get STATUS_CODE_COMMENT */



  /* Update STATUS_CODE_COMMENT in connection information: */
  if (new_status_cmnt[0]) {                       /* if we got a new cmnt: */
    IW_ConnSetStatusCmnt(conn, new_status_cmnt);  /*    save it! */
  }
  else if ((new_status_code != NM_ABNORMAL) &&   /* if status code changed */
	   (new_status_code != IW_ConnStatusCode(conn))) {
    IW_ConnSetStatusCmnt(conn, NULL);
  }

  /* Update STATUS_CODE in connection information: */
  if (new_status_code != NM_ABNORMAL) {
    IW_ConnSetStatusCode(conn, new_status_code);
  }


  /* Stop right here is a QUIT tree: */
  switch (tree_type) {
  case IK_QUIT:
  case IK_ABORT:
    return;
  }


  /* Copy the tree, so we can muck with it without hurting anything: */
  if (!(root_agg = CopyAggregate(origroot_agg))) {
    err("Couldn't copy tree!"); goto CLEANUP;
  }
  IK_ScrubODL(root_agg);

#if IMSCLIENT_PATCH_UNMAPPED_FIELD_BUG
  /* UNFINISHED */
  /* Remove UNMAPPED_FIELD from actual message ODL, to prevent IK crash */  
#endif



  /*
   * EITHER REPORT THE INFORMATION NOW, OR QUEUE IT FOR REPORTING...
   */

  if (QueueMessages) {       /* Report later... */

    /* Add to the queue, and remove it from our cleanup list: */
    debug("Callback: adding %s %s tree to queue", arch_id, msg_agg->name);
    IW_ArrayAdd(MsgQueue, (IW_Elem)root_agg);
    root_agg = NULL;

  }
  else {                     /* Report now... */
    IW_Chunksp * sp;    

    /* Setup the output: */
    sp = IW_ChunkspNew(ObjFp);
    
    /* Output the tree (possibly mucking with it in the process): */
    debug("Callback: reporting %s %s tree", arch_id, msg_agg->name);
    if (report_tree(root_agg, sp) < 0) {
      err("couldn't report tree; continuing");
    }

    /* Close output: */
    IW_ChunkspFree(sp);    
  }


  /*** Done! ******************************/
CLEANUP:
  if (root_agg) { RemoveAggregate(root_agg); }
  return;
}


/************************************************************
 * handle_value
 ************************************************************
 *
 * DESCRIPTION
 *    This function looks for the requestField in the group, tgroup.
 *    If it finds the group it returns the valids in the field_ary along
 *    with *num of valids found in the group.
 */
int handle_value (char ***field_ary, AGGREGATE tGroup, int *num, 
		  char *requestField)
{
  VALUE tempValue;
  VALUE fieldNameValue;
  VALUE fieldValue;
  PARAMETER tempPara;
  PARAMETER fieldPara;
  PARAMETER fieldNamePara;
  AGGREGATE fieldGroup;
  char category_id[MAXLEN];
  char requestValid[MAXLEN];

  for (ptr = requestField;*ptr;ptr++)       /* upcase the field name */
   *ptr = toupper (*ptr);
  tempPara = FindParameter (tGroup, requestField);
  if (tempPara) {
    tempValue = FirstValue (tempPara);
    while (tempValue) {
      (*field_ary)[*num] = 
       (char *)calloc(strlen(tempValue->item.value.string)+1, sizeof(char));
      strcpy ((*field_ary)[*num], tempValue->item.value.string);
      (*num)++;
      tempValue = NextValue (tempValue);
    }
    (*field_ary)[*num] = (char *) NULL;
    return (1);
  }
  else {
    for (fieldGroup = FindGroup(tGroup, "FIELD");
     fieldGroup;
     fieldGroup = FindNextGroup(tGroup, fieldGroup, "FIELD")){
      fieldNamePara = FindParameter (fieldGroup, "NAME");
      if (fieldNamePara) {
        fieldNameValue = FirstValue (fieldNamePara);
        if(fieldNameValue && fieldNameValue->item.value.string)
         strncpy (category_id, fieldNameValue->item.value.string,MAXLEN-1);
        category_id[MAXLEN-1] = '\0';
	for (ptr = category_id;*ptr;ptr++)       /* upcase the field name */
	 *ptr = toupper (*ptr);
        if(requestField && category_id && 
         (strcmp(requestField, category_id) == 0)) {
	  fieldPara = FindParameter (fieldGroup, "VALUES");
	  fieldValue = FirstValue (fieldPara);
          while (fieldValue) {
            if(fieldValue && fieldValue->item.value.string)
              strncpy (requestValid, fieldValue->item.value.string,MAXLEN-1);
            requestValid[MAXLEN-1] = '\0';
            (*field_ary)[*num] = 
	     (char *)calloc(strlen(requestValid)+1, sizeof(char));
            strcpy ((*field_ary)[*num], requestValid);
            (*num)++;
            fieldValue = NextValue (fieldValue);
          }
          (*field_ary)[*num] = (char *) NULL;
          return (1);
        }
      }
    }
  }
    return (0);
}

/****************************** end of file ******************************/
