/* --------------------------------------------------------------------------
  File: gdmap.c

  This file contains map projection code and used the gd1.1+ library for
  drawing.


  Author: Leonid Kokourine,
          AGNET Systems (Greenbelt, MD)
          Tel: 301-441-3857, Fax: 301-441-2351
          E-mail: agnet@cais.com
  Modified: Gyorgy Fekete
          All my mods are bracketed with regular C comments
	  that contain "<><>"  so they are easy to find.

--------------------------------------------------------------------------- */

#include <math.h>
#include <Gaea_config.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <sys/types.h>
				/* <><> */
#undef	DEBUG_THIS
#ifdef	DEBUG_THIS
#include	<signal.h>
#endif

#include "gdfontg.h"
#include "gdfonts.h"
#include "gdfontt.h"
#include "gdmap.h"
#include	"const.h"

#ifndef SEEK_SET
#define SEEK_SET 0
#endif

/* Local definitions */

#define USE_FAST_TRIG

#if defined (USE_FAST_TRIG)
#define sine(x) isin(x)
#define cosine(x) icos(x)
#else
#define sine(x) sin(x)
#define cosine(x) cos(x)
#endif

#define LINE_STEP_DEG  5
#define LINE_STEP_RAD  (LINE_STEP_DEG*DEG2RAD)
#define TILE_STEP_DEG  20
#define V_TILES (180/TILE_STEP_DEG)
#define H_TILES (360/TILE_STEP_DEG)

#define MERC_SCALE (.5/(90.*DEG2RAD))

#define COORD_ORIGIN   0	/* points rel. to origin */
#define COORD_PREVIOUS 1	/* points rel. to previous */

typedef struct {
  short res;
  float lat, lon;
} MapPoint;

typedef struct {
  int np[N_MAP_LAYERS];		/* number of points in tile per layer */
  MapPoint *p[N_MAP_LAYERS];	/* map points for each layer */
} Tile;


/* Global Variable for trace flag */

int  DoGyTrace = 0;

/* Local variables used in tracing for debug */
static int gy_x = -11, gy_y = -11;

static MapGCStr map_gc[(N_MAP_LAYERS + 1) * 2];
static Tile tile[V_TILES][H_TILES];


/* Local functions prototypes */

#if defined (USE_FAST_TRIG)
static double isin(double arg);
static double icos(double arg);
#endif

static int
    calcGreatCir(double start_lat,	/* latitude of the starting point */
		 double start_lng,	/* longitude of the starting point */
		 double end_lat,/* latitude of the ending point */
		 double end_lng,/* longitude of the ending point */
		 double step,	/* angle segmemt size */
		 LatLon route[],/* resulting intermediate points */
		 int *num_points);	/* the number of resulting points */

static void
     calcLine(double lat1,	/* lat coord of starting point */
	      double lon1,	/* lon coord of starting point */
	      double lat2,	/* lat coord of ending point */
	      double lon2,	/* lon coord of ending point */
	      double step,	/* step */
	      LatLon * route,	/* resulting intermediate points */
	      int *npoints);	/* number of resulting points */

static int
    drawMercSegm(
		  LatLon route[],	/* point latitudes/longitudes (rad) */
		  int npoints,	/* number of points */
		  double max_dist,	/* hint-max distance betw. points (rad) */
		  gdPoint xpnt[],	/* points buffer */
		  double center_lat,	/* Latitude of center of projection */
		  double center_lon,	/* Longitude of center of projection */
		  double radius,/* earth radius in pixels */
		  gdImagePtr drawable,	/* drawable */
		  MapGC gc,	/* graphics context */
		  int center_x,	/* x coord of the Drawable center */
		  int center_y);/* y coord of the Drawable center */

static int
    drawOrthoSegm(
		   LatLon route[],	/* input lat/lon points */
		   int npoints,	/* number of points */
		   gdPoint xpnt[],	/* buffer for points */
		   double center_lat,	/* lat of center of projection */
		   double center_lon,	/* lon of center of projection */
		   double radius,	/* earth radius in pixels */
		   gdImagePtr drawable,	/* drawable */
		   MapGC front_gc,	/* GC for visible part of the line */
		   MapGC back_gc,	/* GC for invisible part of the line */
		   int back,	/* back face drawing flag */
		   int center_x,/* window center x */
		   int center_y);	/* window center y */

static int outside(int x, int y,
		   int minx, int miny,
		   int maxx, int maxy);

static void gyTrace(
		    int x0,
		    int y0,
		    int x1,
		    int y1,
		    int color,
		    int mode,
		    int npoints);

static void drawLines(
		       gdImagePtr drawable,
		       MapGC gc,
		       gdPoint * points,
		       int npoints,
		       int mode);

static void drawPoints(
			gdImagePtr drawable,
			MapGC gc,
			gdPoint * points,
			int npoints,
			int mode);


/*==========================================================================*/
/*=================== Local function definitions ===========================*/
/*==========================================================================*/
#if defined (USE_FAST_TRIG)

#define GRADE 50
#define DGRADE 50.0

static double isin(double arg)
{
  static int first = 1;
  static double sintab[360 * GRADE + 1];
  int i;

  if (first) {
    for (i = 0; i <= 360 * GRADE; i++)
      sintab[i] = sin((double) i * DEG2RAD / DGRADE);
    first = 0;
  }
  if (arg < 0)
    arg += M_2PI;
  else if (arg > M_2PI)
    arg -= M_2PI;
  return (sintab[(int) (arg * RAD2DEG * DGRADE + .5)]);
}

static double icos(double arg)
{
  return (isin(M_PI_2 - arg));
}

#endif

/* ----------------------- segmGreatCir () ---------------------------------
   Name:
     int segmGreatCir()

   Description:
     Calculates intermadiate Great Circle segment lat/lon point with
     the given distance ratio. All input and output angle arguments are given
     in radians.

   Returns:
     Returns 0 on success, -1 if could not decide which way to go.
     This happens when the angle between two points is 180 degrees.

   Warning:
     If the abs(end_longitude - start_longitude) > PI, function
     calculates greater of two possible segments.
	    Example: start_lng=170; end_lng=-170; ratio=0.5.
		     The computed point would have longitude=0.
		     In order to calculate shortest circle segment,
		     you must add 2*PI to negative longitude or subtract
		     2*PI from positive one:
		     start_lng=170; end_lng=190; ratio=0.5.
		     The computed point would have longitude=180.
		     Remember! All values are IN RADIANS.

   Global Variables used:
     None.

   Pre/Post Conditions:

   Revision History:

--------------------------------------------------------------------------- */

int segmGreatCir(double start_lat,	/* latitude of the starting point */
		 double start_lng,	/* longitude of the starting point */
		 double end_lat,/* latitude of the ending point */
		 double end_lng,/* longitude of the ending point */
		 double ratio,	/* angle segmemt ratio */
		 LatLon * route)
{				/* resulting intermediate point */
  double central_angle;
  double cos_central_angle;
  double sin_central_angle;
  double cos_start_lat;
  double sin_start_lat;
  double cos_start_lng;
  double sin_start_lng;
  double cos_end_lat;
  double sin_end_lat;
  double cos_end_lng;
  double sin_end_lng;
  double cos_route_angle;
  double sin_route_angle;
  double route_angle;
  double route_vector_x, route_vector_y, route_vector_z;
  double temp1, temp2;

  cos_start_lat = cos(start_lat);
  sin_start_lat = sin(start_lat);
  cos_start_lng = cos(start_lng);
  sin_start_lng = sin(start_lng);
  cos_end_lat = cos(end_lat);
  sin_end_lat = sin(end_lat);
  cos_end_lng = cos(end_lng);
  sin_end_lng = sin(end_lng);

  cos_central_angle = cos_start_lat * cos_end_lat * cos(start_lng - end_lng)
    + sin_start_lat * sin_end_lat;
  if (cos_central_angle <= -1.) {
    return -1;			/* failure - don't know which way to draw */
  }
  if (cos_central_angle >= 1. || fabs(ratio) < 0.001) {
    route->lat = start_lat;
    route->lon = start_lng;
    return 0;			/* no interval - ending points are equal */
  }

  sin_central_angle = sqrt(fabs(1 - cos_central_angle * cos_central_angle));
  if (cos_central_angle >= 1.0)
    central_angle = 0.0;
  else if (cos_central_angle <= -1.0)
    central_angle = M_PI;
  else
    central_angle = acos(cos_central_angle);

  /* If we think that the angle should be > 180, make it think that it is */

  if (fabs(end_lng - start_lng) > M_PI) {
    central_angle = M_2PI - central_angle;
    sin_central_angle = -sin_central_angle;
  }

  route_angle = central_angle * ratio;

  cos_route_angle = cos(route_angle);
  sin_route_angle = sin(route_angle);

  temp1 = sin_route_angle / sin_central_angle;
  temp2 = cos_route_angle - temp1 * cos_central_angle;

  route_vector_x = temp2 * cos_start_lat * cos_start_lng
    + temp1 * cos_end_lat * cos_end_lng;

  route_vector_y = -(temp2 * cos_start_lat * sin_start_lng
		     + temp1 * cos_end_lat * sin_end_lng);

  route_vector_z = temp2 * sin_start_lat + temp1 * sin_end_lat;

  route->lat = asin(route_vector_z);
  route->lon = -atan2(route_vector_y, route_vector_x);
  return 0;
}


/* ---------------------------------------------------------------------------
   Name:
    void  segmLine()

   Description:
     Calculates intermadiate line segment point with the given ratio.

   Global Variables used:
     None.

   Pre/Post Conditions:

   Revision History:

--------------------------------------------------------------------------- */
void segmLine(double lat1,	/* lat coord of starting point */
	      double lon1,	/* lon coord of starting point */
	      double lat2,	/* lat coord of ending point */
	      double lon2,	/* lon coord of ending point */
	      double ratio,	/* ratio */
	      LatLon * route)
{				/* resulting intermediate point */
  double dx, dy, step_x, step_y;

  /* Calc the number of segments and step */

  dx = lat2 - lat1;
  dy = lon2 - lon1;

  step_x = dx * ratio;
  step_y = dy * ratio;

  route->lat = lat1 + step_x;
  route->lon = lon1 + step_y;

  return;
}

/* ---------------------------------------------------------------------------
   Name:
    void  calcSegm()

   Description:
     Calculates intermadiate line segment point with the given ratio.

   Global Variables used:
     None.

   Pre/Post Conditions:

   Revision History:

--------------------------------------------------------------------------- */

int calcSegm(double lat1,	/* lat coord of starting point */
	     double lon1,	/* lon coord of starting point */
	     double lat2,	/* lat coord of ending point */
	     double lon2,	/* lon coord of ending point */
	     double ratio,	/* ratio */
	     int line_type,	/* line type */
	     LatLon * route)
{				/* resulting intermediate point */
  if (line_type == STRAIGHTLINE) {
    segmLine(lat1, lon1, lat2, lon2, ratio, route);
    return 0;
  }
  else {
    return segmGreatCir(lat1, lon1, lat2, lon2, ratio, route);
  }
}

/* ----------------------- calcGreatCir () -----------------------------------
   Name:
     int calcGreatCir()

   Description:
     Calculates intermadiate Great Circle segment lat/lon points with
     the given step. All input and output angle arguments are given in
     radians.

   Returns:
     Returns 0 on success, -1 if could not decide which way to go.
     This happens when the angle between two points is 180 degrees.

   Warning:
     If the abs(end_longitude - start_longitude) > PI, function
     calculates greater of two possible segments.
	    Example: start_lng=170; end_lng=-170; step=10.
		     The computed points would have longitudes:
		     170, 160, 150,..., -160, -170.
		     In order to calculate shortest circle segment,
		     you must add 2*PI to negative longitude or subtract
		     2*PI from positive one:
		     start_lng=170; end_lng=190; step=10.
		     The computed points would have longitudes:
		     170, 180, -170.
		     Remember! All values should be IN RADIANS.

   Global Variables used:
     None.

   Pre/Post Conditions:

   Revision History:

--------------------------------------------------------------------------- */


static int calcGreatCir(double start_lat,	/* latitude of the starting point */
			double start_lng,	/* long of the starting point */
			double end_lat,	/* latitude of the ending point */
			double end_lng,	/* longitude of the ending point */
			double step,	/* angle segmemt size */
			LatLon route[],	/* resulting intermediate points */
			int *num_points)
{				/* the number of resulting points */
  int intval_no, num_intvals;
  double central_angle;
  double angle_increment;
  double cos_central_angle;
  double sin_central_angle;
  double cos_start_lat;
  double sin_start_lat;
  double cos_start_lng;
  double sin_start_lng;
  double cos_end_lat;
  double sin_end_lat;
  double cos_end_lng;
  double sin_end_lng;
  double cos_route_angle;
  double sin_route_angle;
  double route_angle;
  double prev_route_lat;
  double prev_route_lng;
  double route_vector_x, route_vector_y, route_vector_z;
  double temp1, temp2;

  /* <><> 
  fprintf(stderr, "calcCircle %f %f --> %f %f\n",
	  start_lat * RAD2DEG,
	  start_lng * RAD2DEG,
	  end_lat * RAD2DEG,
	  end_lng * RAD2DEG);
  */

  cos_start_lat = cos(start_lat);
  sin_start_lat = sin(start_lat);
  cos_start_lng = cos(start_lng);
  sin_start_lng = sin(start_lng);
  cos_end_lat = cos(end_lat);
  sin_end_lat = sin(end_lat);
  cos_end_lng = cos(end_lng);
  sin_end_lng = sin(end_lng);

  cos_central_angle = cos_start_lat * cos_end_lat * cos(start_lng - end_lng)
    + sin_start_lat * sin_end_lat;
  if (cos_central_angle <= -1.) {
    *num_points = 0;
    return -1;			/* failure - don't know which way to draw */
  }
  if (cos_central_angle >= 1.) {
    *num_points = 1;
    route[0].lat = start_lat;
    route[0].lon = start_lng;
    return 0;			/* no interval - ending points are equal */
  }

  sin_central_angle = sqrt(fabs(1 - cos_central_angle * cos_central_angle));
  if (cos_central_angle >= 1.0)
    central_angle = 0.0;
  else if (cos_central_angle <= -1.0)
    central_angle = M_PI;
  else
    central_angle = acos(cos_central_angle);

  /* If we think that the angle should be > 180, make it think that it is */

  if (fabs(end_lng - start_lng) > M_PI) {
    central_angle = M_2PI - central_angle;
    sin_central_angle = -sin_central_angle;
  }

  num_intvals = ceil(central_angle / step);
  *num_points = num_intvals + 1;
  angle_increment = central_angle / num_intvals;

  prev_route_lat = start_lat;
  prev_route_lng = start_lng;
  for (intval_no = 0; intval_no <= num_intvals; intval_no++) {
    route_angle = intval_no * angle_increment;
    cos_route_angle = cos(route_angle);
    sin_route_angle = sin(route_angle);

    temp1 = sin_route_angle / sin_central_angle;
    temp2 = cos_route_angle - temp1 * cos_central_angle;

    route_vector_x = temp2 * cos_start_lat * cos_start_lng
      + temp1 * cos_end_lat * cos_end_lng;

    route_vector_y = -(temp2 * cos_start_lat * sin_start_lng
		       + temp1 * cos_end_lat * sin_end_lng);

    route_vector_z = temp2 * sin_start_lat + temp1 * sin_end_lat;

    route[intval_no].lat = asin(route_vector_z);

    route[intval_no].lon = -atan2(route_vector_y, route_vector_x);
  }
  return 0;
}


/* ---------------------------------------------------------------------------
   Name:
    void  calcLine()

   Description:
     Calculates intermadiate line segment points with the given step.

   Global Variables used:
     None.

   Pre/Post Conditions:

   Revision History:

--------------------------------------------------------------------------- */
static void calcLine(double lat1,	/* lat coord of starting point */
		     double lon1,	/* lon coord of starting point */
		     double lat2,	/* lat coord of ending point */
		     double lon2,	/* lon coord of ending point */
		     double step,	/* step */
		     LatLon * route,	/* resulting intermediate points */
		     int *npoints)
{				/* number of resulting points */
  int i, n_seg;
  double dx, dy, step_x, step_y, n_seg2;

  /* Calc the number of segments and step */

  route[0].lat = lat1;
  route[0].lon = lon1;
  dx = lat2 - lat1;
  dy = lon2 - lon1;

  if ((n_seg2 = (dx * dx + dy * dy) / (step * step)) <= 1.) {
    route[1].lat = lat2;
    route[1].lon = lon2;
    *npoints = 2;
    return;
  }

  n_seg = (int) ceil(sqrt(n_seg2));
  if (n_seg == 0) {
    *npoints = 0;
    return;
  }
  *npoints = n_seg + 1;
  step_x = dx / n_seg;
  step_y = dy / n_seg;

  for (i = 1; i < *npoints; i++) {
    route[i].lat = route[i - 1].lat + step_x;
    route[i].lon = route[i - 1].lon + step_y;
  }

  return;
}

/* --------------------- int drawMercSegm () --------------------------------
   Name:
     int drawMercSegm()

   Description:
     Draws a segmented line on the mercator projection.
     Takes care of the boundary conditions.

   Returns:
     0 on success, nonzero on failure.

   Global Variables used:
     None.

   Pre/Post Conditions:

   Revision History:

--------------------------------------------------------------------------- */
static int drawMercSegm(
			 LatLon route[],	/* points (rad) */
			 int npoints,	/* number of points */
			 double max_dist,	/* max distance betw. points (rad) */
			 gdPoint xpnt[],	/* point buffer */
			 double center_lat,	/* Latitude of center of projection */
			 double center_lon,	/* Longitude of center of projection */
			 double radius,	/* earth radius in pixels */
			 gdImagePtr drawable,	/* drawable */
			 MapGC gc,	/* graphics context */
			 int center_x,	/* x coord of the Drawable center */
			 int center_y)
{				/* y coord of the Drawable center */
  int idx, i;
  double dlat, dlon, edge_lat, edge_lon, lat1, lon1, lat2, lon2;

  if (npoints < 2)
    return 0;
  /*
  if (center_lon>M_PI) center_lon-=M_2PI;
  else if (center_lon<-M_PI) center_lon+=M_2PI;
*/
  for (i = 0, idx = 0; i < npoints; i++) {
    lon2 = route[i].lon - center_lon;
    lat2 = route[i].lat - center_lat;
    if (lon2 > M_PI)
      lon2 -= M_2PI;
    else if (lon2 < -M_PI)
      lon2 += M_2PI;

    /* check if 2 points are on the opposite sides of an edge */

    if (idx > 0) {
      if (fabs(lon2 - lon1) > M_PI) {	/* if we jumped over the window edge */
	if (lon2 > 0) {		/* if 2nd point was below -PI */
	  edge_lon = -M_PI;	/* restore its original value */
	  lon2 -= M_2PI;
	}
	else {			/* if 2nd point was above PI */
	  edge_lon = M_PI;
	  lon2 += M_2PI;	/* restore its original value */
	}

	/* calc the edge latitude so we can draw 2 segments instead of 1 */

	dlat = lat2 - lat1;
	dlon = lon2 - lon1;
	edge_lat = dlat / dlon * (edge_lon - lon1) + lat1;	/* dlon is always != 0 */

	/* create the fake point on the edge */

	xpnt[idx].x = center_x + (short) (edge_lon * radius * MERC_SCALE);
	xpnt[idx].y = center_y - (short) (edge_lat * radius * MERC_SCALE);
	idx++;

	/* draw all prevoius segments including fake one */

	drawLines(drawable, gc, xpnt, idx, COORD_ORIGIN);

	/* create another fake point on opposite edge for next segment set */

	lon2 = -edge_lon;
	lat2 = edge_lat;
	idx = 0;
	i--;			/* force to process ending point again */
      }
    }
    xpnt[idx].x = center_x + (short) (lon2 * radius * MERC_SCALE);
    xpnt[idx].y = center_y - (short) (lat2 * radius * MERC_SCALE);
    idx++;
    lon1 = lon2;
    lat1 = lat2;
  }

  /* finally put all segments out */

  drawLines(drawable, gc, xpnt, idx, COORD_ORIGIN);

  return 0;
}

/* ------------------------------------------------------------------------
   Name:         drawOrthoSegm();

   Description:  Draws a set of connected line segments on the orthographic
		 map projection.

   Returns:      0 on success, !=0 on failure

   Global Variables used:  None.

--------------------------------------------------------------------------- */
static int drawOrthoSegm(LatLon route[],	/* input lat/lon points */
			 int npoints,	/* number of points */
			 gdPoint xpnt[],	/* buffer for points */
			 double center_lat,	/* lat of center of projection */
			 double center_lon,	/* lon of center of projection */
			 double radius,	/* earth radius in pixels */
			 gdImagePtr drawable,	/* drawable */
			 MapGC front_gc,	/* GC for visible part of the line */
			 MapGC back_gc,	/* GC for invisible part of the line */
			 int back,	/* back face drawing flag */
			 int center_x,	/* window center x */
			 int center_y)
{				/* window center y */
  int idx, i, istat, old_stat;
  double x, y;
  int print_label = 0;


  idx = 0;
  old_stat = 0;


  for (i = 0; i <= npoints; i++) {
    if (i < npoints)
      istat = ortho(radius, center_lat, center_lon, route[i].lat, route[i].lon, &x, &y);

    if ((istat != old_stat && idx > 1) || i == npoints) {
      if (old_stat && back) {
	drawLines(drawable, back_gc, xpnt, idx, COORD_ORIGIN); }
      else if (!old_stat)
	drawLines(drawable, front_gc, xpnt, idx, COORD_ORIGIN);
     
      idx = 0;
    }


    xpnt[idx].x = center_x + (short) x;
    xpnt[idx].y = center_y - (short) y;

    idx++;
    old_stat = istat;

}


  return 0;
}

/*==========================================================================*/
/*==================== Global functions definitions ========================*/
/*==========================================================================*/

/* ----------------------------- ortho() -------------------------------------
   Name: int ortho();

   Description:
     Computes the (x,y) coordinates in pixels for orthographic globe
     projection given the latitude/longitude and projection parameters.

   Returns:
     0 if point can be projected successfully, nonzero if point
     projects on the opposite side of the globe (invisible)

   Global Variables used: None

   Pre/Post Conditions:

   Revision History:

--------------------------------------------------------------------------- */

int ortho(double radius,	/* radius of the earth in pixels */
	  double center_lat,	/* latitude of center of projection (rad) */
	  double center_lon,	/* longitude of center of projection (rad) */
	  double inlat,		/* latitude of input point (rad) */
	  double inlon,		/* longitude of input point (rad) */
	  double *outx,		/* x offset from center of window in pixels */
	  double *outy)
{				/* y offset from center of window in pixels */
  double delta;
  double cosL;
  double cosC;
  double sin_inlat;
  double cos_inlat;
  static double last_center_lat = 0;
  static double sin_center_lat = 0;
  static double cos_center_lat = 1;

  delta = inlon - center_lon;
  if (delta < -M_PI)
    delta = delta + M_2PI;
  else if (delta > M_PI)
    delta = delta - M_2PI;

  if (center_lat != last_center_lat) {
    last_center_lat = center_lat;
    sin_center_lat = sine(center_lat);
    cos_center_lat = cosine(center_lat);
  }
  sin_inlat = sine(inlat);
  cos_inlat = cosine(inlat);

  cosL = cosine(delta) * cos_inlat;
  cosC = sin_center_lat * sin_inlat + cos_center_lat * cosL;

  *outx = cos_inlat * sine(delta) * radius;
  *outy = (cos_center_lat * sin_inlat - sin_center_lat * cosL) * radius;
  if (cosC >= 0)
    return (0);			/* OK */
  else
    return (-1);		/* point is on the "dark" side */
}


/* -------------------------- invortho() -------------------------------------
   Name: int invortho()

   Description:
     Converts X-Y screen coordinates of a point in pixels
     to geographic point in radians by inverting orthographic projection.

   Returns: 0 on success, nonzero if point is outside of the globe.

   Global Variables used:

   Pre/Post Conditions:

   Revision History:

--------------------------------------------------------------------------- */

int invortho(double radius,	/* radius of the Earth in pixels */
	     double center_lat,	/* latitude of center of projection (rad) */
	     double center_lon,	/* longitude of center of projection (rad) */
	     double inx,	/* x offset from center of window in pixles */
	     double iny,	/* y offset from center of window in pixles */
	     double *outlat,	/* latitude of selected point in radians */
	     double *outlon)
{				/* longitude of selected point in radians */
  double distance;
  double CON;
  double sin_outlat;
  double sin_Z;
  double cos_Z;
  static double last_center_lat = 0;
  static double sin_center_lat = 0;
  static double cos_center_lat = 1;

  distance = sqrt(inx * inx + iny * iny);

  if (distance > radius)
    return (-1);		/* out of this world */
  if (distance == 0) {
    *outlat = center_lat;
    *outlon = center_lon;
  }
  else {
    if (center_lat != last_center_lat) {
      last_center_lat = center_lat;
      if (fabs(center_lat) < (double) 89.5 * DEG2RAD) {
	sin_center_lat = sin(center_lat);
	cos_center_lat = cos(center_lat);
      }
      /* make sure that asin2(0,0) doesn't occur */
      else if (center_lat > 0) {
	sin_center_lat = sin((double) M_PI_2 - 0.00000001);
	cos_center_lat = cos((double) M_PI_2 - 0.00000001);
      }
      else {
	sin_center_lat = sin((double) -M_PI_2 + 0.00000001);
	cos_center_lat = cos((double) -M_PI_2 + 0.00000001);
      }
    }

    sin_Z = distance / radius;
    cos_Z = sqrt(1 - (sin_Z * sin_Z));

    sin_outlat = (cos_Z * sin_center_lat) +
      (iny * sin_Z * cos_center_lat / distance);
    *outlat = asin(sin_outlat);
    CON = cos_Z - (sin_center_lat * sin_outlat);
    *outlon = center_lon + atan2((inx * sin_Z * cos_center_lat),
				 (CON * distance));
    if (*outlon < -M_PI)
      *outlon = *outlon + M_2PI;
    else if (*outlon > M_PI)
      *outlon = *outlon - M_2PI;
  }
  return (0);
}

/* --------------------------------------------------------------------------
   Name: int mercate()

   Description:
     Lat/Lon in radians to X-Y screen coordinates of a point in pixels
     for Mercator projection. Radius=number of pixels in 90-degree swath

   Returns: 0 on success, nonzero if point is outside of the screen.

   Global Variables used: None

   Pre/Post Conditions:

   Revision History:

--------------------------------------------------------------------------- */
int mercate(double radius,	/* radius of the earth in pixels */
	    double center_lat,	/* latitude of center of projection (rad) */
	    double center_lon,	/* longitude of center of projection (rad) */
	    double inlat,	/* latitude of input point (rad) */
	    double inlon,	/* longitude of input point (rad) */
	    double *outx,	/* x offset from center of window in pixels */
	    double *outy)
{				/* y offset from center of window in pixels */


  inlon -= center_lon;
  inlat -= center_lat;
  if (inlon > M_PI)
    inlon -= M_2PI;
  else if (inlon < -M_PI)
    inlon += M_2PI;
  *outx = inlon * radius * MERC_SCALE;
  *outy = inlat * radius * MERC_SCALE;
  return 0;
}



/* --------------------------------------------------------------------------
   Name: int invmercate()

   Description:
     Converts X-Y screen coordinates of a point in pixels
     to geographic point in radians by inverting mercator projection.

   Returns: 0 on success, nonzero if point is outside of the map.

--------------------------------------------------------------------------- */
int invmercate(double radius,	/* radius of the Earth in pixels */
	       double center_lat,	/* latitude of center of projection (rad) */
	       double center_lon,	/* longitude of center of projection (rad) */
	       double inx,	/* x offset from center of window in pixles */
	       double iny,	/* y offset from center of window in pixles */
	       double *outlat,	/* latitude of selected point in radians */
	       double *outlon)
{				/* longitude of selected point in radians */
  *outlat = iny / (radius * MERC_SCALE) + center_lat;
  if (*outlat > M_PI_2 || *outlat < -M_PI_2)
    return -1;
  *outlon = inx / (radius * MERC_SCALE);
  if (*outlon > M_PI || *outlon < -M_PI)
    return -1;
  *outlon += center_lon;
  if (*outlon < -M_PI)
    *outlon += M_2PI;
  else if (*outlon > M_PI)
    *outlon -= M_2PI;

  return 0;
}


/* --------------------------------------------------------------------------
   Name: int Map()

   Description:
     Converts X-Y screen coordinates of a point in pixels
     to geographic point in radians.

   Returns: 0 on success, nonzero if point is outside of the map.

--------------------------------------------------------------------------- */
extern int Map(
		int projection,	/* proj type */
		double radius,	/* radius of the earth in pixels */
		double center_lat,	/* latitude of center of projection (rad) */
		double center_lon,	/* longitude of center of projection (rad) */
		int center_x,	/* x coord of the Drawable center */
		int center_y,	/* y coord of the Drawable center */
		double inlat,	/* latitude of input point (rad) */
		double inlon,	/* longitude of input point (rad) */
		double *outx,	/* x coordinate in the window in pixels */
		double *outy)
{				/* y coordinate in the window in pixels */
  int status;
  double x, y;

  if (projection == MERCATOR) {
    status = mercate(radius, center_lat, center_lon,
		     inlat, inlon, &x, &y);
  }
  else {
    status = ortho(radius, center_lat, center_lon,
		   inlat, inlon, &x, &y);
  }
  *outx = x + center_x;
  *outy = center_y - y;
  return status;
}

/* --------------------------------------------------------------------------
   Name: int invMap()

   Description:
     Converts Lat/Lon to X-Y screen coordinates of a point in pixels

   Returns: 0 on success, nonzero if point is outside of the map.

--------------------------------------------------------------------------- */
extern int invMap(
		   int projection,	/* proj type */
		   double radius,	/* radius of the Earth in pixels */
		   double center_lat,	/* latitude of center of projection (rad) */
		   double center_lon,	/* longitude of center of projection (rad) */
		   int center_x,/* x coord of the Drawable center */
		   int center_y,/* y coord of the Drawable center */
		   double inx,	/* x coordinate in the window in pixles */
		   double iny,	/* y coordinate in the window in pixles */
		   double *outlat,	/* latitude of selected point in radians */
		   double *outlon)
{				/* longitude of selected point in radians */
  int status;

  if (projection == MERCATOR) {
    status = invmercate(radius, center_lat, center_lon,
			inx - center_x, center_y - iny,
			outlat, outlon);
  }
  else {
    status = invortho(radius, center_lat, center_lon,
		      inx - center_x, center_y - iny,
		      outlat, outlon);
  }
  return status;
}

/* --------------------- int drawMapLine () --------------------------------
   Name:
     int drawMapLine()

   Description:
     Draws a projected line on the map projection. Four map projections
     supported: mercator, orthographic, north and south polar. Two types of
     lines are supported: straight line and great circle segment, depending on
     the "type" argument value (STRAIGHTLINE or GREATCIRCLE respectively). If
     the "back" flag is nonzero, it indicates that we want to draw continuation
     of the line on the opposite (invisible) side of an orthographic projection.
     The "back_gc" graphics context is used in this case.

   Returns: 0 on success, nonzero on failure.

   Warning: All coordinate arguments are in radians.

   Global Variables used: None.

   Pre/Post Conditions:

   Revision History:

--------------------------------------------------------------------------- */
int drawMapLine(
		 double lat1,	/* 1st point latitude (rad) */
		 double lon1,	/* 1st point longitude (rad) */
		 double lat2,	/* 2nd point latitude (rad) */
		 double lon2,	/* 2nd point longitude (rad) */
		 double center_lat,	/* Latitude of center of projection */
		 double center_lon,	/* Longitude of center of projection */
		 double radius,	/* earth radius in pixels */
		 gdImagePtr drawable,	/* drawable */
		 MapGC front_gc,/* graphics context */
		 MapGC back_gc,	/* graphics context for back face */
		 int center_x,	/* x coord of the Drawable center */
		 int center_y,	/* y coord of the Drawable center */
		 int linetype,	/* line type */
		 int projection,/* projection type */
		 int back)
{				/* back face drawing flag */
  LatLon route[360 / LINE_STEP_DEG + 4];
  gdPoint xpnt[360 / LINE_STEP_DEG + 4];
  int npoints, istat;

  if (fabs(lon1 - lon2) < 0.001)
    linetype = STRAIGHTLINE;
  switch (linetype) {
  case STRAIGHTLINE:
    calcLine(lat1, lon1, lat2, lon2, LINE_STEP_RAD, route, &npoints);
    break;
  case GREATCIRCLE:

    /* if we are unlucky enough 1st time and it did not know which direction to
take (central angle == 180) then change latitude a bit and try again. */
    for (istat = 1; istat; lat1 += 0.001)
      istat = calcGreatCir(lat1, lon1, lat2, lon2, LINE_STEP_RAD, route, &npoints);
    break;
  default:
    return -1;
  }

  if (npoints < 2)
    return 0;

  if (projection == MERCATOR)
    return
      drawMercSegm(route, npoints, LINE_STEP_RAD, xpnt,
		   center_lat, center_lon, radius,
		   drawable, front_gc, center_x, center_y);
  else
    return
      drawOrthoSegm(route, npoints, xpnt,
		    center_lat, center_lon, radius,
		    drawable, front_gc, back_gc,
		    back, center_x, center_y);
}


/* ------------------------------------------------------------------------
   Name:   int drawMapPoint()

   Description:   Draws a "thick" point on the World Map (~5 pixels wide)

   Returns:       0 if point is visible, !=0 otherwise

   Global Variables used:  None

   Pre/Post Conditions:

   Revision History:

--------------------------------------------------------------------------- */
int drawMapPoint(
		  double lat,	/* latitude of the point */
		  double lon,	/* longitude of the point */
		  double center_lat,	/* Latitude of center of projection */
		  double center_lon,	/* Longitude of center of projection */
		  double radius,/* earth radius in pixels */
		  gdImagePtr drawable,	/* drawable */
		  MapGC front_gc,	/* graphics context */
		  MapGC back_gc,/* graphics context for back face */
		  int center_x,	/* x coord of the Drawable center */
		  int center_y,	/* y coord of the Drawable center */
		  int projection,	/* projection type */
		  int back)
{				/* back face drawing flag */
  static gdPoint bullseye_pts[] =
  {
    {0, 0},
    {-1, -2},
    {1, 0},
    {1, 0},
    {-3, 1},
    {1, 0},
    {1, 0},
    {1, 0},
    {1, 0},
    {-4, 1},
    {1, 0},
    {2, 0},
    {1, 0},
    {-4, 1},
    {1, 0},
    {1, 0},
    {1, 0},
    {1, 0},
    {-3, 1},
    {1, 0},
    {1, 0}};
  /* bulls eye points for IK_POINT granule type */
#define BULLSEYE_NPTS 21	/* number of bulls eye points */

  double x, y;
  MapGC gc;

  int status = Map(projection, radius, center_lat, center_lon,
		   center_x, center_y, lat, lon, &x, &y);

  bullseye_pts[0].x = x;
  bullseye_pts[0].y = y;

  if (projection == MERCATOR) {
    gc = front_gc;
  }
  else {
    if (status != 0) {
      if (back == 0)
	return -1;
      gc = back_gc;
    }
    else
      gc = front_gc;
  }
  drawPoints(drawable, gc, bullseye_pts, BULLSEYE_NPTS, COORD_PREVIOUS);
  return 0;
}


/* ------------------------------------------------------------------------
   Name: drawArea()

   Description: Draws polygonal area on the map. All coordinates in rad
                The line_type can be GREATCIRCLE or STRAIGHTLINE. First
                is used in polygons, however if the longitudial distance
                between two points is > 180 degrees, the line is forced to
                be STRAIGHTLINE to avoid incorrect area drawing.
                The 'join' flag should be !=0 if the area must be a closed
                polygon, otherwise first and last points are not connected.
                If pole_inside==1 then all points are connected with
                shortest lines NOT taking into account the "clockwiseness'.
                This is necessary to draw areas covering poles
                unambiguously.


   Returns:       void

------------------------------------------------------------------------ */

void drawArea(double inlat[],	/* lat values */
	      double inlon[],	/* lon values */
	      int n_sides,	/* number of sides */
              char label[],
	      int line_type,	/* GREATCIRCLE or STRAIGHTLINE */
	      int pole_inside,	/* if a pole is inside the area */
	      gdImagePtr Workspace,	/* gd image pointer */
	      MapGC front_gc,	/* GC for drawing */
	      MapGC back_gc,	/* GC for drawing "invisible" side */
	      double center_lat,/* center lat of the projection */
	      double center_lon,/* center lon of the projection */
	      double radius,	/* earth radius in pixels */
	      int center_x,	/* x coord of map center */
	      int center_y,	/* y coord of map center */
              double zfact,
              int drawLabels,
	      int projection,	/* ORTHOGRAPHIC or MERCATOR */
	      int backface,	/* !=0 to draw invisible side in Ortho*/
	      int join)
{				/* see description above */
  double lat1, lat2, lat3, lon1, lon2, lon3, lo;
  int i, j, k, istat;
  int seq = 0;
  int yellow = gdImageColorClosest(Workspace, 0 , 0, 0);
  double x, y;
 int position = 0;
 double center_pos_lat, center_pos_lon; 

  /* Always set the pole included flag */
  /* It appears to work correctly for all cases this way */
  pole_inside = 1;

  for (i = 0; i < n_sides; i += 2) {
    j = i + 1;
    if (i < n_sides - 2 || (!join && i < n_sides - 1))
      k = i + 2;
    else
      k = 0;

/* <><> */
    /*
   fprintf(stderr, "%6.3d %6.3d %6.3d (%d)\n", i, j, k, n_sides);
   */
/* It's always: 0 1 2, then 2 3 0 n_sides == 4 */
/* <><> */

    lon1 = inlon[i];
    lon2 = inlon[j];
    lon3 = inlon[k];
    lat1 = inlat[i];
    lat2 = inlat[j];
    lat3 = inlat[k];

    if (fabs(lon1 - lon2) > M_PI) {
      if (lon2 > lon1)
	lon1 += M_2PI;
      else
	lon1 -= M_2PI;
    }

    if (fabs(lon3 - lon2) > M_PI) {
      if (lon2 > lon3)
	lon3 += M_2PI;
      else
	lon3 -= M_2PI;
    }

    if (!points_clockwise(lat1, lon1, lat2, lon2, lat3, lon3)
	&& !pole_inside) {

      lo = lon1 + M_2PI;
      if (points_clockwise(lat1, lo, lat2, lon2, lat3, lon3))
	lon1 = lo;
      else {
	lo = lon1 - M_2PI;
	if (points_clockwise(lat1, lo, lat2, lon2, lat3, lon3))
	  lon1 = lo;
      }
    }

				/* <><> 

    fprintf(stderr, "[%d] 1-2 drawMapLine %6.3f %6.3f -> %6.3f %6.3f (p %d)\n",
	    seq,
	    lat1, lon1, lat2, lon2,
	    pole_inside);
    gy_x = gy_y = -1;
                                  <><> */
    istat = drawMapLine(lat1, lon1, lat2, lon2, center_lat, center_lon,
			radius, Workspace,
			front_gc, back_gc,
			center_x, center_y,
			fabs(lon1 - lon2) >= M_PI ? STRAIGHTLINE : line_type,
			projection, backface);
				/* <><> 
    seq++;
                                   <><> */

    if (j < n_sides || join) {
				/* <><> 
       fprintf(stderr, "[%d] drawMapLine %6.3f %6.3f -> %6.3f %6.3f \
       (p %d, join)\n",
	       seq,
	       lat3, lon3, lat2, lon2,
	       pole_inside);
       gy_x = gy_y = -1;
				  <><> */
       istat = drawMapLine(lat3, lon3, lat2, lon2, center_lat, center_lon,
			   radius, Workspace,
			   front_gc, back_gc,
			   center_x, center_y,
			  fabs(lon1 - lon2) >= M_PI ? STRAIGHTLINE : line_type,
			   projection, backface);
				/* <><> 
       seq++;
				 <><> */

    }

  } /* for i */


  /* For MISR
   Label the Polygons with their BLOCK number */

  /* Ok to draw the Labels? */
  if (drawLabels) {

     /* Returns:
     0 if point can be projected successfully, nonzero if point
     projects on the opposite side of the globe (invisible)
     */


/*     fprintf(stderr, "Pt 1  %6.3f %6.3f , Pt 2  %6.3f %6.3f , Pt 3 %6.3f %6.3f \n",lat1 , lon1, lat2, lon2, lat3, lon3 );
*/

/* ((lon1 - lon2)/2) + lon2; */

     center_pos_lon = ((lon2 - lon1)/2) + lon1;

/* works, don't touch ((lat2 - lat3)/2) +  lat2 
     center_pos_lat  = ((lat2 - lat3)/2) +  lat3; */

    center_pos_lat  = ((inlat[1] - inlat[3])/2) +  inlat[3];



     istat = ortho(radius, center_lat, center_lon,  center_pos_lat, center_pos_lon ,&x, &y);

     if (!istat || backface ) {
      if ( zfact == 4 ) 
        gdImageString(Workspace,
                      gdFontTiny,
                        center_x  + (short) x -2   ,
                        center_y -  (short) y - 2  ,
                      label , yellow );

      else if ( zfact > 4 ) 
        gdImageString(Workspace,
                      gdFontSmall,
                        center_x + (short) x - 5 ,
                        center_y -  (short) y - 5   ,
                      label , yellow );

    }
  }
}


/* ------------------- int points_clockwise() --------------------------------

   Name: int points_clockwise()

   Description:
     Determines if 3 points are in clockwise direction.

   Returns: Returns zero if not, nonzero if yes.

   Global Variables used: None.

--------------------------------------------------------------------------- */

int points_clockwise(double lat1,
		     double lon1,
		     double lat2,
		     double lon2,
		     double lat3,
		     double lon3)
{
  double ax = lon1 - lon2;
  double ay = lat1 - lat2;
  double bx = lon3 - lon2;
  double by = lat3 - lat2;

  return (ax * by - bx * ay > 0);
}


/* ---------------------------------------------------------------------------
   Name: outside();

   Description:
     Computes outcodes (flags) for clipping algorithm.
     Checks to see if ending points are inside/outside of clipping region.

   Returns: true if line can be trivially rejected.

   Global Variables used: None.

--------------------------------------------------------------------------- */
static int outside(int x, int y,
		   int minx, int miny,
		   int maxx, int maxy)
{

  int c = (x < minx ? 1 : (x > maxx ? 2 : 0));
  if (y < miny)
    c |= 4;
  else if (y > maxy)
    c |= 8;
  return c;
}



/* ---------------------------------------------------------------------------
   Name: line_check();

   Description:
     Computes outcodes (flags) for clipping algorithm.
     Checks to see if ending points are inside/outside of clipping region.

   Returns: true if line can be trivially rejected.

   Global Variables used: None.

--------------------------------------------------------------------------- */

int line_check(int xmin,
	       int xmax,
	       int ymin,
	       int ymax,
	       int x1,
	       int y1,
	       int x2,
	       int y2)
{
  int code = 0;

  code |= (ymax - y1) < 0 ? 0x01 : 0;
  code |= (y1 - ymin) < 0 ? 0x02 : 0;
  code |= (xmax - x1) < 0 ? 0x04 : 0;
  code |= (x1 - xmin) < 0 ? 0x08 : 0;
  code &= (ymax - y2) < 0 ? 0x01 : 0;
  code &= (y2 - ymin) < 0 ? 0x02 : 0;
  code &= (xmax - x2) < 0 ? 0x04 : 0;
  code &= (x2 - xmin) < 0 ? 0x08 : 0;

  return (code);
}


/* ------------------------------------------------------------------------
   Name: readpw()

   Description: Read partial word from stream
                Should work on any system

   Returns:     word cast to long

   Global vars: None

------------------------------------------------------------------------ */

static long readpw(size_t size, FILE * f)
{
  long v = getc(f);
  if (v & 0x80)
    v = (v & 0xff) - 256;
  while (--size)
    v = (v << 8) + (getc(f) & 0xff);
  return v;
}

/* ------------------------------------------------------------------------
   Name: readPoint()

   Description: Read WDB entry from stream
                Should work on any system

   Returns:     0 on failure, !=0 on success

   Global vars: None

------------------------------------------------------------------------ */

static int readPoint(short *point, FILE * f)
{
  int i;
  for (i = 0; i < 3; i++)
    point[i] = readpw(sizeof(short), f);
  return (!ferror(f));
}



/* ------------------------------------------------------------------------
   Name: tileIsVisible()

   Description: Determines if the specified tile is inside the screen

   Returns:     0 if not visible, !=0 if visible.
                For stereographic projections returns != 0 also
                if within viewable area on the back face and "back" flag is
                true

   Global vars: None

------------------------------------------------------------------------ */
int tileIsVisible(int ilat,
		  int ilon,
		  int projection,
		  double center_lat,
		  double center_lon,
		  double radius,
		  int backface,
		  int ctr_x,
		  int ctr_y)
{
  int istat, i;
  int x[4], y[4], minx, miny, maxx, maxy;
  double xx, yy, lat, lon;

  lat = (ilat * TILE_STEP_DEG - 90) * DEG2RAD;
  lon = (ilon * TILE_STEP_DEG - 180) * DEG2RAD;

  if (projection != MERCATOR) {
    istat = ortho(radius, center_lat, center_lon, lat, lon, &xx, &yy);
    x[0] = xx;
    y[0] = yy;
    if ((istat == 0 || backface == BACK_FACE_OFF) &&
	!outside(x[0], y[0], -ctr_x, -ctr_y, ctr_x, ctr_y))
      return 1;

    lon += TILE_STEP_DEG * DEG2RAD;
    lat += TILE_STEP_DEG * DEG2RAD;
    istat = ortho(radius, center_lat, center_lon, lat, lon, &xx, &yy);
    x[1] = xx;
    y[1] = yy;
    if ((istat == 0 || backface == BACK_FACE_OFF) &&
	!outside(x[1], y[1], -ctr_x, -ctr_y, ctr_x, ctr_y))
      return 1;

    lat -= TILE_STEP_DEG * DEG2RAD;
    istat = ortho(radius, center_lat, center_lon, lat, lon, &xx, &yy);
    x[2] = xx;
    y[2] = yy;
    if ((istat == 0 || backface == BACK_FACE_OFF) &&
	!outside(x[2], y[2], -ctr_x, -ctr_y, ctr_x, ctr_y))
      return 1;

    lon -= TILE_STEP_DEG * DEG2RAD;
    lat += TILE_STEP_DEG * DEG2RAD;
    istat = ortho(radius, center_lat, center_lon, lat, lon, &xx, &yy);
    x[3] = xx;
    y[3] = yy;
    if ((istat == 0 || backface == BACK_FACE_OFF) &&
	!outside(x[3], y[3], -ctr_x, -ctr_y, ctr_x, ctr_y))
      return 1;
  }
  else {
    istat = mercate(radius, center_lat, center_lon, lat, lon, &xx, &yy);
    x[0] = xx;
    y[0] = yy;
    if (istat == 0 && !outside(x[0], y[0], -ctr_x, -ctr_y, ctr_x, ctr_y))
      return 1;

    lon += TILE_STEP_DEG * DEG2RAD;
    lat += TILE_STEP_DEG * DEG2RAD;
    istat = mercate(radius, center_lat, center_lon, lat, lon, &xx, &yy);
    x[1] = xx;
    y[1] = yy;
    if (istat == 0 && !outside(x[1], y[1], -ctr_x, -ctr_y, ctr_x, ctr_y))
      return 1;

    lat -= TILE_STEP_DEG * DEG2RAD;
    istat = mercate(radius, center_lat, center_lon, lat, lon, &xx, &yy);
    x[2] = xx;
    y[2] = yy;
    if (istat == 0 && !outside(x[2], y[2], -ctr_x, -ctr_y, ctr_x, ctr_y))
      return 1;

    lon -= TILE_STEP_DEG * DEG2RAD;
    lat += TILE_STEP_DEG * DEG2RAD;
    istat = mercate(radius, center_lat, center_lon, lat, lon, &xx, &yy);
    x[3] = xx;
    y[3] = yy;
    if (istat == 0 && !outside(x[3], y[3], -ctr_x, -ctr_y, ctr_x, ctr_y))
      return 1;
  }
  minx = INT_MAX;
  miny = INT_MAX;
  maxx = INT_MIN;
  maxy = INT_MIN;
  for (i = 0; i < 4; i++) {
    if (x[i] > maxx)
      maxx = x[i];
    if (y[i] > maxy)
      maxy = y[i];
    if (x[i] < minx)
      minx = x[i];
    if (y[i] < miny)
      miny = y[i];
  }

  if (((minx < ctr_x && maxx > -ctr_x) || (minx > ctr_x && maxx < -ctr_x)) &&
      ((miny < ctr_y && maxy > -ctr_y) || (miny > ctr_y && maxy < -ctr_y)))
    return 1;
  else
    return 0;
}


/* ------------------------------------------------------------------------
   Name:         int drawWorldMap();

   Description:  Draw the World Map on the specified Drawable in specified
                 projection with specified resolution. Projections supported:
		 MERCATOR, ORTHOGRAPHIC, SOUTH_- and NORTH_POLAR_STEREOGRAPHIC.
		 The "resolution" ranges from WDB_RES_LOW to WDB_RES_HI.
		
		 The map database file is searched for in the
                 following order:
                 1 - file pointed to by WDBTILE environment variable
                 2 - ./WDBTILE file in the current directory

   Returns:      0 on success, !=0 on failure.

   Warning:      This function allocates memory for WDB data first time it is
		 called. It never deallocates this memory.

   Global Vars:  map_gc

   Pre/Post Conditions:

   Revision History:

--------------------------------------------------------------------------- */
#ifdef	DEBUG_THIS
				/* <><> I used this to find why it
				 crashes under -n32 */
struct {
  int read;
  int i;
  int ilat, ilon;
  int layer;
} Debug = { -2, -2, -2, -2, -2};
void sighandler()
{
  
  debug(" SIGSEGV: read was %d, i = %d, lon/lat = (%d, %d) layer %d", 
	Debug.read,
	Debug.i, 
	Debug.ilon, 
	Debug.ilat,
	Debug.layer);
  fatal("Oops!");
}
#endif

int drawWorldMap(gdImagePtr drawable,	/* drawable */
		 int desired[],	/* desired map layers */
		 int resolution,/* desired map resolution */
		 int projection,/* projection type */
		 int backface,	/* backface flag for ortho proj. */
		 int grid_step,	/* grid step (deg); 0-no grid */
		 double center_lat,	/* latitude of map center */
		 double center_lon,	/* longitude of map center */
		 double radius,	/* earth radius in pixels */
		 int center_x,	/* window center x */
		 int center_y,
                 int display_LatLons) {/* window center y */
  FILE *fp;
  static char *mappath;
  int istat, i, layer, npoints;
  long seg_len, max_seg_len;
  short point[3], res;
  double lat, lon;
  int ilat, ilon, counter = 0;
  int minlon, maxlon, minlat, maxlat;

  static int first = 1;
  static gdPoint *xpoint;
  static LatLon *segment;
  /* was: static off_t tileoffs[V_TILES][H_TILES]; */
  static long tileoffs[V_TILES][H_TILES];

  int color_black = gdImageColorClosest(drawable, 0, 0, 0);
  char path_label[7] = " " ;
  double x_x, y_y;
  double long_x, long_y;
  div_t add_to;
  char positionNS[1] = " ", positionEW[1]  = " ";  /* Holds either N,E,W,S */

  /* adjust resolution */

  if (resolution <= 0)
    resolution = WDB_RES_HI;

  /* if this is first call - read WDB in memory and setup pointers */

  if (first) {
    if (!(mappath = getenv("WDBTILE")))
      mappath = "./WDBTILE";

    if ((fp = fopen(mappath, "r")) == NULL) {
      fatal("couldn't open map `%s'", mappath);
    }

    /* first read tile offset table */

    for (ilat = 0; ilat < V_TILES; ilat++) {
      for (ilon = 0; ilon < H_TILES; ilon++) {
	tileoffs[ilat][ilon] = readpw(sizeof(tileoffs[ilat][ilon]), fp);

      }
    }
    if (ferror(fp)) {
      fclose(fp);
      return -2;
    }

#ifdef	DEBUG_THIS
    signal(SIGSEGV, sighandler);
#endif

    /* read WDB and setup tiles */

    max_seg_len = 0;
    for (ilat = 0; ilat < V_TILES; ilat++) {
      for (ilon = 0; ilon < H_TILES; ilon++) {

	/* init tiles */

	for (layer = 0; layer < N_MAP_LAYERS; layer++) {
	  tile[ilat][ilon].np[layer] = 0;
	  tile[ilat][ilon].p[layer] = NULL;
	}

	/* read the tile from WDB if data for it exists */

	if (tileoffs[ilat][ilon] != 0) {
	  fseek(fp, tileoffs[ilat][ilon], SEEK_SET);

	  /* count points for this tile */

	  seg_len = 1;
	  do {
	    if (readPoint(point, fp) == 0) {
	      fclose(fp);
	      return -3;
	    }
	    if ((point[0] & 0xff) != 0) {	/* if not end of tile */
	      layer = ((point[0] >> 4) & 0xf) - 1;
	      tile[ilat][ilon].np[layer]++;

	    }
	    if (point[0] & 0xf) {	/* if not the end of segment or tile */
	      seg_len++;
	    }
	    else {		/* last seg point - store max segment length */
	      if (seg_len > max_seg_len)
		max_seg_len = seg_len;
	      seg_len = 1;
	    }
	  }
	  while (!feof(fp) && (point[0] & 0xff) != 0);

	  /* allocate memory for tile data */

	  for (layer = 0; layer < N_MAP_LAYERS; layer++) {
	    if (tile[ilat][ilon].np[layer] > 0) {
	      tile[ilat][ilon].p[layer] = (MapPoint *)
		malloc(tile[ilat][ilon].np[layer] * sizeof(MapPoint));
	    }
	    tile[ilat][ilon].np[layer] = 0;	/* reset point count */
	  }

	  /* rewind the file to tile starting point and read it in */

	  fseek(fp, tileoffs[ilat][ilon], SEEK_SET);
	  do {
#ifdef	DEBUG_THIS
	    Debug.ilat = ilat;
	    Debug.ilon = ilon;
	    Debug.read = 1;
	    Debug.layer = -1;
	    Debug.i = -1;
#endif
	    
	    if (readPoint(point, fp) == 0) {
	      fclose(fp);
	      return -4;
	    }

	    if ((point[0] & 0xff) != 0) {	/* not the end of tile */
	      layer = ((point[0] >> 4) & 0xf) - 1;
#ifdef	DEBUG_THIS       
	      Debug.layer = layer;
#endif
	      i = tile[ilat][ilon].np[layer];
#ifdef	DEBUG_THIS
	      Debug.i = i;
#endif
	      tile[ilat][ilon].p[layer][i].lat = (double) point[1] * MIN2RAD;
	      tile[ilat][ilon].p[layer][i].lon = (double) point[2] * MIN2RAD;
	      tile[ilat][ilon].p[layer][i].res = point[0] & 0xf;
	      tile[ilat][ilon].np[layer]++;
	    }
	  }
	  while (!feof(fp) && (point[0] & 0xff) != 0);
	}
      }
    }


    xpoint = (gdPoint *) malloc(max_seg_len * sizeof(gdPoint));
    segment = (LatLon *) malloc(max_seg_len * sizeof(LatLon));

    first = 0;
  }

  /* Draw the outer circle of the globe if it fits in the window */

  if (projection != MERCATOR &&
      (radius < (center_x > center_y ? center_x : center_y) * 3. / 2.)) {
    gdImageArc(drawable,
	       center_x, center_y,
	       (int) radius * 2, (int) radius * 2,
	       0, 360,
	       map_gc[GRID_GC * 2].color);
  }
				/* <><> [ed:debug] just the circle */
				/* return here if you want just the circle */
				/* <><> [ed:debug] remove when done */
  /* lay the grid over */

  if (grid_step > 0) {
    /* make sure the back face grid lines are drawn first */

    /* draw horizontal grid lines */

    minlon = (int) ((center_lon + M_PI_2) * RAD2DEG + .5) / grid_step * grid_step;

    if (minlon > 180)
      minlon -= 360;
    maxlon = minlon + 360;

    if (center_lat > 0.)
      minlat = -90 + (grid_step - 90 % grid_step);
    else {
      minlat = 90 - (grid_step - 90 % grid_step);
      grid_step = -grid_step;
    }

    for (i = minlat; i > -90 && i < 90; i += grid_step) {
      lat = (double) i *DEG2RAD;

       /* Label the Latitudes  */
       if ( display_LatLons) {
          /* Change 'counter' into text then based on the position add N/S, so
              the user knows where they are */
          sprintf(path_label, "%i",  i);
          if ( i > 0 ) {
             strcpy (positionNS ,  "N");
          }
          else if ( i == 0 ) {
                strcpy (positionNS," ");
          }
          else {
            strcpy (positionNS, "S");
          }
          strcat ( path_label, positionNS);

          /* Start Labeling */
          istat = ortho(radius,  center_lat  , center_lon  , lat, center_lon  , &x_x, &y_y);

          if ( istat == 0 ) {
               gdImageString(drawable,
                      gdFontSmall,
                            ( center_x  + x_x )  ,
                           ( center_y  - y_y - 5 ) ,
                         path_label , color_black );
          }
      } 
      drawMapLine(lat, minlon * DEG2RAD, lat, maxlon * DEG2RAD,
		  center_lat, center_lon, radius,
		  drawable, &map_gc[GRID_GC * 2], &map_gc[GRID_GC * 2 + 1],
		  center_x, center_y, STRAIGHTLINE, projection, backface);
    }

    /* draw vertical grid lines */

    grid_step = abs(grid_step);

    if (center_lat < 0.)
      minlat = 90;
    else
      minlat = -90;
    maxlat = -minlat;

    
    /* Set counter to the present longitude, then move it to
       the closest 'grid step'
    */
    counter = (center_lon * RAD2DEG);
    add_to = div ( counter , grid_step);
    counter +=  (  grid_step -  add_to.rem)  ;

   for (i = minlon; i < maxlon; i += grid_step) { 
/*      for (i = counter ; i <   counter + 360 ; i += grid_step) {  */

       lon = (double) i *DEG2RAD;


       /* Label the Longitudes */
       if ( display_LatLons) {

           /* Allows the user to rotate the map 360 degrees */
           if ( counter >= 180 ) {
              counter -= 360;
           }

            strcpy (positionEW, " ");
  /*          Change counter into text then based on the position add W/E, so
              the user knows where they are */
           if ( counter > 0 ) { 
              strcpy (positionEW, "E");
           }
           else if ( counter == 0 ) {
           }
           else {
              strcpy (positionEW, "W");
           }
           /* strcat ( path_label, positionEW);
           fprintf(stderr , " positionEW = %s \n ", positionEW); */
           sprintf(path_label, "%i%s",  counter, positionEW );

           /* Start Labeling */
           istat = ortho(radius, center_lat, center_lon, center_lat, (double) (counter *DEG2RAD), &long_x, &long_y);


            if ( istat == 0 ) {
              gdImageString(drawable,
                      gdFontSmall,
                           ( center_x  + long_x -5   )  ,
                           ( center_y  - long_y +10  ) ,
                           path_label , color_black );
            }
       }

       drawMapLine(
		   minlat * DEG2RAD, lon, maxlat * DEG2RAD, lon,
		   center_lat, center_lon, radius,
		   drawable, &map_gc[GRID_GC * 2], &map_gc[GRID_GC * 2 + 1],
		   center_x, center_y, STRAIGHTLINE, projection, backface);

       /* Only label every OTHER longtitude */
      counter += ( grid_step * 2 ) ; 
    }
  }



				/* <><> [ed:debug] just the grid */
				/*  return here if you don't want map */
				/* <><> [ed:debug] remove when done */


  /* Draw map on the Drawable layer by layer, segment by segment */
  /* loop over all tiles and see which are visible */

  for (ilat = 0; ilat < V_TILES; ilat++) {
    for (ilon = 0; ilon < H_TILES; ilon++) {
      if (tileIsVisible(ilat, ilon, projection,
			center_lat, center_lon, radius, backface,
			center_x, center_y)) {
	for (layer = 0; layer < N_MAP_LAYERS; layer++) {
	  if (desired[layer]) {

	    /* If this layer is desired, loop over all points */

	    npoints = 0;
	    for (i = 0; i < tile[ilat][ilon].np[layer]; i++) {
	      res = tile[ilat][ilon].p[layer][i].res;
	      if (res <= resolution) {
		segment[npoints].lat = tile[ilat][ilon].p[layer][i].lat;
		segment[npoints].lon = tile[ilat][ilon].p[layer][i].lon;
		npoints++;
	      }
	      if (res == 0 && npoints > 1) {	/* end of segment */
		if (projection == MERCATOR)
		  drawMercSegm(segment, npoints, LINE_STEP_RAD, xpoint,
			       center_lat, center_lon, radius,
			       drawable, &map_gc[layer * 2],
			       center_x, center_y);
		else
		  drawOrthoSegm(segment, npoints, xpoint,
				center_lat, center_lon, radius,
		       drawable, &map_gc[layer * 2], &map_gc[layer * 2 + 1],
				backface, center_x, center_y);
		npoints = 0;
	      }
	    }			/* for all layer points */
	  }			/* if layer desired */
	}			/* for layer */
      }				/* if visible */
    }				/* for horiz tiles */
  }				/* for vert tiles */





  return 0;
}


/* ------------------------------------------------------------------------
   Name:   void setMapGC();

   Description:  Assigns GCs to varoius WDB map lines (for COAST, STATE, etc)
		 The first GC is for visible part of a line and for MERCATOR
		 The second one is for invisible part of aline on a
		   ORTHOGRAPHIC projection.

   Returns: Nothing

   Global Variables used:  map_gc array

   Pre/Post Conditions:

   Revision History:

--------------------------------------------------------------------------- */
void setMapGC(
	       int idx,		/* the GC index */
	       MapGC front,	/* GC for visible part */
	       MapGC back)
{				/* GC for invisible part */
  memcpy(&map_gc[idx * 2], front, sizeof(*front));
  memcpy(&map_gc[idx * 2 + 1], back, sizeof(*back));
}


/*------------------------------------------------------------------------*/
/*------------------------------------------------------------------------
 Name: gyTrace()

 Description: Traces short drawLines calls when global flag DoGyTrace is set
 ------------------------------------------------------------------------- */
static void gyTrace(
		    int x0,
		    int y0,
		    int x1,
		    int y1,
		    int color,
		    int mode,
		    int npoints)

{
  int brk;
  if (!DoGyTrace)
      return;
  if (gy_x < 0 || gy_y < 0){
     gy_x = x0;
     gy_y = y0;
  }

  brk = (gy_x != x0 || gy_y != y0);
  
  if (mode)
      fprintf(stderr, "\tdL %4d %4d -> %4d %4d (origin, color %2d) %2d %s\n",
	      x0, y0, x1, y1, color, npoints,
	      brk ? "*******" : "");
  else
      fprintf(stderr, "\tdL %4d %4d -> %4d %4d (relati, color %2d) %2d %s\n",
	      x0, y0, x1, y1, color, npoints,
	      brk ? "*******" : "");
  gy_x = x1;
  gy_y = y1;

}


				/* <><> */

/* -------------------------------------------------------------------------
  Name: drawLines()

  Description: Draws joint straight line segments on the gd image.
               If mode=COORD_ORIGIN then coordinates are absolute,
               otherwise each subsequent point coordinates are relative to
               previous point. First point coordinates are always absolute

  Returns:     void
------------------------------------------------------------------------- */
static void drawLines(
		       gdImagePtr drawable,
		       MapGC gc,
		       gdPoint * points,
		       int npoints,
		       int mode)
{
  int i;
  gdPoint *p = points, *p1 = points + 1;

  if (npoints < 2)
    return;

/* fprintf(stderr, "Color set %d \n", gc->color); */

  if (mode == COORD_ORIGIN) {
    for (i = 0; i < npoints - 1; i++) {
      gdImageLine(drawable, p->x, p->y, p1->x, p1->y, gc->color);
      if (gc->line_width == 2){
	int x0,y0,x1,y1;	/* so x0 <- x1 and y0 <= y1 */
	int dx, dy;
	if(p->x<=p1->x){x0=p->x;x1=p1->x;} else {x0=p1->x;x1=p->x;}
	if(p->y<=p1->y){y0=p->y;y1=p1->y;} else {y0=p1->y;y1=p->y;}
	dx=x1-x0;dy=y1-y0;
	if(dx > dy)		/* more horizontal */ {
	  gdImageLine(drawable, p->x, p->y+1, p1->x, p1->y+1, gc->color);
	}
	else			/* more vertical */ {
	  gdImageLine(drawable, p->x+1, p->y, p1->x+1, p1->y, gc->color);
	}
      }
				/* <><> */
      /* gyTrace( p->x, p->y, p1->x, p1->y, gc->color, 1, npoints); */

      p = p1;
      p1++;
    }
  }
  else {
    int x = p->x, y = p->y;
    for (i = 0; i < npoints - 1; i++) {
       /* 
	  gyTrace(x, y, x+p1->x, y+p1->y, gc->color, 0, npoints);
	*/
       gdImageLine(drawable, x, y, x += p1->x, y += p1->y, gc->color);
       p1++;
    }
  }
}

/* -------------------------------------------------------------------------
  Name: drawPoints()

  Description: Draws multiple points on the gd image.
               If mode=COORD_ORIGIN then coordinates are absolute,
               otherwise each subsequent point coordinates are relative to
               previous point. First point coordinates are always absolute.

  Returns:     void

------------------------------------------------------------------------- */
static void drawPoints(
			gdImagePtr drawable,
			MapGC gc,
			gdPoint * points,
			int npoints,
			int mode)
{
  int i;
  gdPoint *p = points;

  if (npoints < 2)
    return;

  if (mode == COORD_ORIGIN) {
    for (i = 0; i < npoints; i++, p++)
      gdImageSetPixel(drawable, p->x, p->y, gc->color);
  }
  else {
    int x = 0, y = 0;
    for (i = 0; i < npoints; i++, p++) {
      x += p->x, y += p->y;
      gdImageSetPixel(drawable, x, y, gc->color);
    }
  }
}




/* ------------------------ end of file gdmap.c ------------------------ */
