/* -------------------------------------------------------------------------
  imscovermap.c

  EOS IMS "map giffer"
  Produces GIF image of the globe in orthographic or Plate-carree projection
  with a given zoom factor, center lat/lon and other parameters.
  Also plots arbitrary number of coverage areas of polygonal or range type.

  Author: Leonid Kokourine,
          AGNET Systems (Greenbelt, MD)
          Tel: 301-441-3857, Fax: 301-441-2351
          E-mail: agnet@cais.com

  Notes:
          Added support for Point: (eryq)
	  Added support for 
	      filled spherical polygon (gyuri)
	      filled ranges
	      translucent spherical polygons
	      too many to list

  Modification History:
          Added support for filled spherical polygon (gyuri)
	  Added color command (gyuri)
------------------------------------------------------------------------- */
#define	FUNKY

#include <Gaea_config.h>
#include <unistd.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "gdmap.h"
#include "view.h"
#include "gring.h"

#include "gdfontt.h"
#include "gdfonts.h"

#include <stdarg.h>		/* eryq */

#include	<sqtTypes.h>	/* Gyuri */
#include	<sqtFuncs.h>
#include	<sqtQTypes.h>
#include       	<sqtQFuncs.h>
#include	"trixels.h"

/* Error levels */

#define FATAL    -3
#define WARN     -2
#define INFO     -1
#define OK        0

/* Coverage types */

/* 
 * [gyf]: 
 *	Now that we are dealing with 4 not 2 kinds of areas, let's use
 * 	some smarter macros for COV_ stuff. 
 *	Let bit 0 (value 1) represent POINT
 *	Let bit 1 (value 2) represent POLYGON 
 *	Let bit 2 (value 4) represent RANGE
 *      Let bit 3 (value 8) represent GRING
 *	Let bit 4 (value 16) represent Trixel (qmesh) 
 */

#define	BIT_POINT	1
#define	BIT_POLY	2
#define	BIT_RANGE	4
#define	BIT_GRING	8	/* [ed:gyuri:new] December 1997 */
#define	BIT_QMESH	16
#define	BIT_GLOBAL	32


#define COV_POINT     (BIT_POINT)

#define COV_POLYGON   (BIT_POLY)
#define COV_RANGE     (BIT_RANGE)

#define	COV_QMESH     (BIT_POLY  | BIT_QMESH)
#define	COV_QRANGE    (BIT_RANGE | BIT_QMESH)
#define	COV_QGRING    (BIT_GRING | BIT_QMESH)
#define	COV_GLOBAL    (BIT_GLOBAL| BIT_QMESH)

#ifdef	DARK
#ifdef  FUNKY
Cant_Have_Both_Dark_and_Funky();:;
#endif
#endif

#ifndef DARK
#ifndef FUNKY

#define COLOR_INIT    255, 255, 255
#define COLOR_CIL       0,   0,   0
#define COLOR_CIL_BK  214, 214, 214
#define COLOR_GRID    197, 197, 197
#define COLOR_GRID_BK 224, 224, 224
#define COLOR_RIV       0,   0, 255
#define COLOR_RIV_BK  127, 127, 255
#define COLOR_NBD       0, 200,   0
#define COLOR_NBD_BK  127, 200, 127
#define COLOR_COV     255,   0,   0
#define COLOR_COV_BK  255, 127, 127
#endif
#endif

#ifdef DARK
#define COLOR_INIT      0,   0,   0
#define COLOR_CIL     255, 255, 255
#define COLOR_CIL_BK  127, 127, 127
#define COLOR_GRID    224, 224, 224
#define COLOR_GRID_BK 127, 127, 127
#define COLOR_RIV       0,   0, 255
#define COLOR_RIV_BK    0,   0, 127
#define COLOR_NBD       0, 255,   0
#define COLOR_NBD_BK    0, 127,   0
#define COLOR_COV     255,   0,   0
#define COLOR_COV_BK  127,   0,   0
#endif

#ifdef	FUNKY
#define COLOR_INIT     64,  64,  64
#define COLOR_CIL     192,   0,  142
#define COLOR_CIL_BK   63,   0,  63
#define COLOR_GRID     92, 129,  92
#define COLOR_GRID_BK 127, 127, 127
#define COLOR_RIV       0,   0, 255
#define COLOR_RIV_BK    0,   0, 127
#define COLOR_NBD       0, 142, 102
#define COLOR_NBD_BK    0, 127,   0
#define COLOR_COV     255, 180,   0
#define COLOR_COV_BK  127, 90,   0
#endif

/* Instead of drawing the cov using COLOR_COV colors, let user define the 
   color */

/* Input arguments */

#define MAX_KW_LEN 40

typedef enum {
  KW_PROJECTION,
  KW_DIMENSIONS,
  KW_CENTER,
  KW_ZOOM,
  KW_TRANSPARENT,
  KW_POLITICAL,
  KW_LAKESRIVERS,
  KW_STAMPPOLITICAL,
  KW_STAMPLAKESRIVERS,
  KW_POINT,
  KW_RANGE,
  KW_COLOR,			/* [ed:gyuri:new] */
  KW_POLYGON,
  KW_QMESH,
  KW_QRANGE,
  KW_QRING,			/* [ed:gyuri Dec 16, 1997] */
  KW_QRESOLUTION,
  KW_STAMPS,
  KW_LABELS,			/* [ed:jsa Feb 2000] */
  KW_GRID,			/* [ed:jsa Feb 2000] */
  KW_LABEL_GRIDS,		/* [ed:jsa Feb 2000] */
  KW_GIFIMAGE,			/* [ed:jsa Feb 2000] */
  KW_TEST
} KwordCode;

typedef struct CmdParm {
  char keyword[MAX_KW_LEN];	/* keyword */
  KwordCode code;		/* keyword code */
  int mandatory;		/* 1 if this parameter is mandatory */
  int given;			/* number of times it was given */
}

CmdParm;

CmdParm cmd_parm[] =
{
  {"PROJECTION: ", KW_PROJECTION, 1, 0},
  {"DIMENSIONS: ", KW_DIMENSIONS, 1, 0},
  {"CENTER: ", KW_CENTER, 1, 0},
  {"ZOOM: ", KW_ZOOM, 0, 0},
  {"TRANSPARENT: ", KW_TRANSPARENT, 0, 0},
  {"POLITICAL: ", KW_POLITICAL, 0, 0},
  {"LAKESRIVERS: ", KW_LAKESRIVERS, 0, 0},
  {"STAMPPOLITICAL: ", KW_STAMPPOLITICAL, 0, 0},
  {"STAMPLAKESRIVERS: ", KW_STAMPLAKESRIVERS, 0, 0},
  {"POINT: ", KW_POINT, 0, 0},
  {"RANGE: ", KW_RANGE, 0, 0},
  {"POLYGON: ", KW_POLYGON, 0, 0},
  {"QCOLOR: ", KW_COLOR, 0, 0},
  {"QMESH: ", KW_QMESH, 0, 0},
  {"QRESOLUTION: ", KW_QRESOLUTION, 0, 0},
  {"QRANGE: ", KW_QRANGE, 0, 0},
  {"QRING: ", KW_QRING, 0, 0},
  {"STAMPS: ", KW_STAMPS, 0, 0},
  {"LABELS: ", KW_LABELS, 0, 0},
  {"GRID: ", KW_GRID, 0, 0},
  {"GIFIMAGE: ", KW_GIFIMAGE, 0, 0},
  {"LABELGRIDS: ", KW_LABEL_GRIDS, 0, 0},
  {"TEST: ", KW_TEST, 0, 0},
};

int num_cmd_parm = sizeof(cmd_parm) / sizeof(CmdParm);

View *view = NULL;
int verbose = 0;

char *wdir = ".";


/* A NULL-terminated array of pointers to views: */
#define MAXVIEWS 31
View *gViews[MAXVIEWS + 1];

/* Centers of postage stamps, in order: */
typedef struct CenterLatLon {
  int clat;
  int clon;
}

CenterLatLon;

static CenterLatLon gStampCenters[] =
{
  {90, 0},			/* north pole */
  {0, -90},			/* america */
  {0, 0},			/* africa */
  {0, 90},			/* asia */
  {0, 180},			/* pacific */
  {-90, 0}			/* south pole */
};

typedef struct {
  int r;
  int g;
  int b;
} ColorType;

static ColorType CurrentColor = {COLOR_COV};

/* -------------------------------------------------------------------------
  Name: debug()

  Description: Output a debugging message.

  Returns: void

  Author: Eryq

------------------------------------------------------------------------- */
void debug(char *fmt,...)
{
  va_list args;
  va_start(args, fmt);
  fflush(stdout);
  fprintf(stderr, "DEBUG: ");
  vfprintf(stderr, fmt, args);
  va_end(args);
  fprintf(stderr, "\n");
  fflush(stderr);
}

/* -------------------------------------------------------------------------
  Name: fatal()

  Description: Die immediately with a fatal error.

  Returns: does not return

  Author: Eryq

------------------------------------------------------------------------- */
int fatal(char *fmt,...)
{
  va_list args;
  va_start(args, fmt);
  printf("X-Error: ");
  vfprintf(stdout, fmt, args);
  va_end(args);
  printf("\n\n");
  exit(-1);
  return 1;			/* to keep C happy */
}



/* -------------------------------------------------------------------------
  Name: errOut()

  Description: Outputs error, warning and informational messages to stdout.
               Warnings are output only if the verbose level > 0.

  Returns: void

------------------------------------------------------------------------- */
static void errOut(int sev, char *txt, char *txt1)
{
  switch (sev) {
  case WARN:
    if (verbose > 0)
      printf("X-Warn: %s %s\n", txt, txt1);
    break;
  case FATAL:
    printf("X-Error: %s %s\n", txt, txt1);
    break;
  default:
    printf("%s %s\n", txt, txt1);
  }
}

/* -------------------------------------------------------------------------
  Name: initGC()

  Description: Allocates colors for all map and coverage drawing for a given
               View.

  Returns: OK on success, FATAL if a color could not be allocated

------------------------------------------------------------------------- */
static int initGC(View * v)
{
  int color;

  if ((color = gdImageColorAllocate(v->image, COLOR_INIT)) < 0)
    return FATAL;
  v->gc_init.color = color;

  if ((color = gdImageColorAllocate(v->image, COLOR_CIL)) < 0)
    return FATAL;
  v->gc_cil.color = color;

  if ((color = gdImageColorAllocate(v->image, COLOR_CIL_BK)) < 0)
    return FATAL;
  v->gc_cil_bk.color = color;

  if ((color = gdImageColorAllocate(v->image, COLOR_GRID)) < 0)
    return FATAL;
  v->gc_grid.color = color;

  if ((color = gdImageColorAllocate(v->image, COLOR_GRID_BK)) < 0)
    return FATAL;
  v->gc_grid_bk.color = color;

  if ((color = gdImageColorAllocate(v->image, COLOR_RIV)) < 0)
    return FATAL;
  v->gc_riv.color = color;

  if ((color = gdImageColorAllocate(v->image, COLOR_RIV_BK)) < 0)
    return FATAL;
  v->gc_riv_bk.color = color;

  if ((color = gdImageColorAllocate(v->image, COLOR_NBD)) < 0)
    return FATAL;
  v->gc_nbd.color = v->gc_ibd.color = color;

  if ((color = gdImageColorAllocate(v->image, COLOR_NBD_BK)) < 0)
    return FATAL;
  v->gc_nbd_bk.color = v->gc_ibd_bk.color = color;

  if ((color = gdImageColorAllocate(v->image, COLOR_COV)) < 0)
    return FATAL;
  v->gc_cov.color = color;

  if ((color = gdImageColorAllocate(v->image, COLOR_COV_BK)) < 0)
    return FATAL;
  v->gc_cov_bk.color = color;

  v->gc_init.line_width =
    v->gc_cil.line_width = v->gc_cil_bk.line_width = v->gc_riv.line_width =
    v->gc_riv_bk.line_width = v->gc_nbd.line_width = v->gc_nbd_bk.line_width =
    v->gc_ibd.line_width = v->gc_ibd_bk.line_width = 1;

  v->gc_cov.line_width = v->gc_cov_bk.line_width = 2;/* [ed:line_width]<><> */
  return OK;
}

/* -------------------------------------------------------------------------
  Name: initView()

  Description: Allocates and initializes a new View structure

  Returns: View* pointer, or NULL on memory or gd error

------------------------------------------------------------------------- */
static View *initView(
		       int w,
		       int h,
		       int proj,
		       double zfact,
		       int transp,
		       int lat,
		       int lon,
                       int grid_in_degrees,
                       char *image_filename)
{
  View *v = malloc(sizeof(View));
  int i;
  FILE *in;


  if (!v)
    return NULL;

  in = fopen(image_filename, "rb");
  if (!in) {
     errOut(WARN, "Creating new image from scratch", "");
     if (!(v->image = gdImageCreate(w, h))) {
        free(v);
        errOut(FATAL, "Can't create image!", "");
        return NULL;
     }
  }
  else {
          errOut(WARN, "Reading from file : ", image_filename);
          v->image = gdImageCreateFromGif(in);
          w =  gdImageSX(v->image);
          h = gdImageSY(v->image);

          fclose(in);
  }

/********************
Original Code :

  if (!(v->image = gdImageCreate(w, h))) {
    free(v);
    errOut(FATAL, "Can't create image!", "");
    return NULL;
  }
*******************/


  if (!(v->scratch = gdImageCreate(w, h))) {
    free(v);
    errOut(FATAL, "Can't create scratch image!", "");
    return NULL;
  }

  if (initGC(v) != OK) {
    free(v);
    errOut(FATAL, "Can't initialize GC", "");
    return NULL;
  }
				/* copy new image into scratch */
  gdImageCopy(v->scratch, v->image, 0, 0, 0, 0, w-1, h-1);

  for (i = 0; i < N_MAP_LAYERS; i++)
    v->layers[i] = 1;
  v->projection = proj;
  v->resolution = WDB_RES_MID + (zfact - 1.) / 2.;
  v->backface = transp;
  v->gridinc = grid_in_degrees;
  v->center_lat = lat;
  v->center_lon = lon;
  v->width = w;
  v->height = h;
  v->center_x = w / 2;
  v->center_y = h / 2;
  v->zfact = zfact;
  v->radius = zfact * ((proj == ORTHOGRAPHIC)
		       ? (w > h ? h / 2 - 1 : w / 2 - 1)
		       : (w > h * 2 ? h - 1 : w / 2 - 1));
  v->cover = NULL;
  v->next = NULL;
  return v;
}

/* -------------------------------------------------------------------------
  Name: freeView()

  Description: Deallocates resources of the given View

  Returns: void

------------------------------------------------------------------------- */
static void freeView(View * v)
{
  Cover *c, *c1;
  if (!v)
    return;
  if (v->image)
    gdImageDestroy(v->image);
  if (v->scratch)
    gdImageDestroy(v->scratch);
  for (c = v->cover; c;) {
    c1 = c->next;
    free(c);
    c = c1;
  }
  free(v);
}


/* -------------------------------------------------------------------------
  Name: initCover()

  Description: Initializes Cover structure

  Returns: Cover* struct pointer, or NULL on memory error

------------------------------------------------------------------------- */
static Cover *initCover(int type, int pole)
{
  Cover *c = malloc(sizeof(Cover));
  if (!c)
    return NULL;
  c->npoints = MAX_POLYPOINTS;
  c->type = type;

  /* */
  /* was:  c->pole_inside = type == COV_RANGE ? 0 : pole; [gyf] */
  /* */
  c->pole_inside = (type & BIT_RANGE) ? 0 : pole;

  c->next = NULL;
  c->label[0] = 'X';
  c->label[1] = '\0';
  c->zerosize = 0;
  c->path[0] = '\0';

  return c;
}

static void gyPolyTraceXXX(View *v)
{
				/* test this by puting a grid on the GIF */
  int i, j;
  int w, h;
  gdPoint Rect[4];
  gdImagePtr drawable;
  gdPointPtr pp = Rect;
  MapGC gc;

  w = v->width;
  h = v->height;
  drawable = v->image;
  gc = &v->gc_cov;

  for(i=0; i<w; i++){
     for(j=0; j<h; j++){
				/* Is this pixel inside a polygon? */
	if (i<32)     continue;
	if (j<32)     continue;
	if (i>(w-32)) continue;
	if (j>(w-32)) continue;
	if(((i+j)%100) == 0){
	   gdImageSetPixel(drawable, i, j, gc->color);
	 }
      }
   }
  return;
}
/* -------------------------------------------------------------------------
  Name: drawCover()

  Description: Draws single coverage area/point specified by Cover* struct
               for the given View.
  Returns: void

------------------------------------------------------------------------- */
static void drawCover(View *v, Cover *c)
{
int position = 0;



  if (!v || !c)
    return;
				/* Drawarea takes care of polygon and mesh */
				/* I will not change that */
				/* drawQmesh will insert trixels */
				/* into Sphere quadtree as either great */
				/* circles, or lines of equal latitude */


  if (c->type == COV_POINT) {
    drawMapPoint(c->lat[0], c->lon[0],
		 v->center_lat * DEG2RAD, v->center_lon * DEG2RAD,
		 v->radius, v->image, &v->gc_cov, &v->gc_cov_bk,
		 v->center_x, v->center_y, v->projection, v->backface);
  }
  else if (c->zerosize) {
    

				/* reduces to RANGE. Want so support color */
				/* do it here, by providing a custom gc    */
				/* c->r,g,b are significant */

				/* get custom color */
      int color;
      int c_save;
      color = gdImageColorExact(v->image, c->r, c->g, c->b);
      if (color < 0){
	color = gdImageColorAllocate(v->image, c->r, c->g, c->b);
	if (color < 0){
	  color = gdImageColorClosest(v->image, c->r, c->g, c->b);
	}
      }
      c_save = v->gc_cov.color;
      v->gc_cov.color = color;

      drawArea(c->lat, c->lon, c->npoints, c->label,
	     STRAIGHTLINE,
	     0,
	     v->image, &v->gc_cov, &v->gc_cov_bk,
	     v->center_lat * DEG2RAD, v->center_lon * DEG2RAD,
	     v->radius, v->center_x, v->center_y, v->zfact, v->displayLabels,
	     v->projection, v->backface, 1);
      v->gc_cov.color = c_save;
  }
  else if (c->type & BIT_QMESH) {
    if (c->type & BIT_GLOBAL) {
      drawQglobal(v, c->r, c->g, c->b);
    } else if (c->type & BIT_POLY){
      drawQmesh(v,
		c->lat,
		c->lon,
		c->npoints,
		c->label,
		c->r, c->g, c->b,
		1,		/* was:c->type & BIT_POLY, */
		c->pole_inside);
    } else if (c->type & BIT_RANGE) {
      drawQmesh(v,
		c->lat,
		c->lon,
		c->npoints,
		c->label,
		c->r, c->g, c->b,
		0,		/* was: c->type & BIT_POLY, */
		c->pole_inside);
    } else if (c->type & BIT_GRING) {

				/* [ed:gringtest] */
      int i;
      /*
      fprintf(stderr, " >>> drawQRing with label \"%s\" (0x%x)\n", 
	      c->label, v);
      fflush(stderr);
      */
      drawQGring(v, c->label, c->r, c->g, c->b);

#ifdef	NEVER
      c->lat[0] = 63.0000;
      c->lon[0] = -2.0000;
      c->lat[1] = 56.0000;
      c->lon[1] = 37.0000;
      c->lat[2] = 35.0000;
      c->lon[2] = 20.0000;
      c->lat[3] = 38.0000;
      c->lon[3] = -1.0000;
      c->pole_inside = POLE_NONE;
      c->npoints = 4;
      for (i = 0; i < c->npoints; i++) {
	c->lat[i] *= DEG2RAD;
	c->lon[i] *= DEG2RAD;
      }
      
      drawQPoly(v, c->lat, c->lon, 3, c->label, c->r, c->g, c->b,
		56.0, 
		34.0);
#endif
    }
    if (0 & c->type & BIT_POLY) {
      drawArea(c->lat, c->lon, c->npoints, c->label,
	       (c->type & BIT_RANGE) ? STRAIGHTLINE : GREATCIRCLE,
	       (c->type & BIT_RANGE) ? 0 : c->pole_inside,
	       v->image, &v->gc_cov, &v->gc_cov_bk,
	       v->center_lat * DEG2RAD, v->center_lon * DEG2RAD,
	       v->radius, v->center_x, v->center_y, v->zfact, v->displayLabels,
	       v->projection,  v->backface, 1);
    }
  }
  else {
     /* [gyf] was:
	     c->type == COV_RANGE ? STRAIGHTLINE : GREATCIRCLE,
	     c->type == COV_RANGE ? 0 : c->pole_inside,
	     ...
      */
       int color;
       int yellow = gdImageColorClosest(v->image, 0 , 0, 0);


       color = gdImageColorExact(v->image, c->r, c->g, c->b);
/*      fprintf(stderr, "Color %d R %d G %d B %d \n ",  color, c->r, c->g, c->b);
*/ 
       if (color < 0){
         color = gdImageColorAllocate(v->image, c->r, c->g, c->b);
         if (color < 0){
           color = gdImageColorClosest(v->image, c->r, c->g, c->b);
         }
       }
 
       v->gc_cov.color = color;
/*      fprintf(stderr, "Color %d R %d G %d B %d \n ",  color, c->r, c->g, c->b);
*/ 

    drawArea(c->lat, c->lon, c->npoints, c->label,
	     (c->type & BIT_RANGE) ? STRAIGHTLINE : GREATCIRCLE,
	     (c->type & BIT_RANGE) ? 0 : c->pole_inside,
	     v->image, &v->gc_cov, &v->gc_cov_bk,
	     v->center_lat * DEG2RAD, v->center_lon * DEG2RAD,
	     v->radius, v->center_x, v->center_y, v->zfact, v->displayLabels,
	     v->projection, v->backface, 1);


  }

}

/* -------------------------------------------------------------------------
  Name: drawView()

  Description: Draws the World Map and coverage areas specified for the
               given View on the gd image.

  Returns: OK on success, FATAL on error (if map could not be drawn)

------------------------------------------------------------------------- */
static int drawView(View * v)
{
  extern int DoGyTrace;		/* no other function need to see this */
  int ret;
  Cover *c;
int yellow;
int chosen_color;
int position = 2;
char legend[50] ;
char path_label[50] ;
/*int display_main_only = 0; */

  if (!v)
    return FATAL;


 yellow = gdImageColorClosest(v->image, 0 , 0, 0);
      chosen_color = gdImageColorExact(v->image, v->cover->r, v->cover->g, v->cover->b);
      if (chosen_color < 0){
        chosen_color = gdImageColorAllocate(v->image, v->cover->r, v->cover->g, v->cover->b);
        if (chosen_color < 0){
          chosen_color = gdImageColorClosest(v->image, v->cover->r, v->cover->g, v->cover->b);
        }
      }


				/* Erase to background color */
  gdImageFilledRectangle(
			  v->image,
			  0, 0, v->width - 1, v->height - 1,
			  v->gc_init.color);



  gdImageColorTransparent(v->image, v->gc_init.color); /* make background
							"transparent" */





  gdImageColorTransparent(v->scratch, v->gc_init.color); /* make background
							"transparent" */

  setMapGC(COAST, &v->gc_cil, &v->gc_cil_bk);
  setMapGC(RIVER, &v->gc_riv, &v->gc_riv_bk);
  setMapGC(IBOUND, &v->gc_ibd, &v->gc_ibd_bk);
  setMapGC(NBOUND, &v->gc_nbd, &v->gc_nbd_bk);
  setMapGC(GRID_GC, &v->gc_grid, &v->gc_grid_bk);

  gdImageCopy(v->scratch, v->image, 0, 0, 0, 0, v->width - 1, v->height - 1);
				/* erase scratch */

/*  JSA */

 if (!( strlen(v->baseImage) > 1 )) {
  gdImageFilledRectangle(
			  v->scratch,
			  0, 0, v->width - 1, v->height - 1,
			  v->gc_init.color);
 }

				/* make another square in scratch */
				/* put a test square in top left */

  sqt_icosinit();
  sqtQ_init();

  DoGyTrace = 1;
 

/*    strcpy(legend, v->cover->path); */
strcpy(legend, "Legend"); 

/*     fprintf(stderr, " Legend reset \n" ); */

 for (c = v->cover; c; c = c->next) {


/*     gdImageLine( v->image, 0,position , 3, position , chosen_color); */




/* fprintf(stderr, " Legend %s, path %s \n", legend, c->path );
*/

/* if ( (strcmp(c->path,legend )) && ( display_main_only == 0 ) ) {
*/

if (v->displayLabels == 1 ) {

if  (strcmp(c->path,legend ))  {
      chosen_color = gdImageColorExact(v->image, c->r, c->g, c->b);
      if (chosen_color < 0){
        chosen_color = gdImageColorAllocate(v->image, c->r, c->g, c->b);
        if (chosen_color < 0){
          chosen_color = gdImageColorClosest(v->image, c->r, c->g, c->b);
        }
      }

if ( strcmp(c->path,"\0") && strcmp(c->path, "0") ) {
    gdImageFilledRectangle(
                          v->image,
                          0, position, 5,  position +5 ,
                           chosen_color);



     strcpy(path_label, " Path:");
     strcat(path_label,  c->path);

      gdImageString(v->image,
                      gdFontSmall,
                        5 ,
                        position  - 2,
                      path_label , yellow );

      position = position +8;
      strcpy(legend, c->path);
}
/* fprintf(stderr, " Should be the same Legend %s, path %s Position  %i\n", legend, c->path, position );
*/
/* display_main_only = 1;
*/

}
}

    drawCover(v, c);		/* [ed:gyuri renders into scratch */
    if (1){
				/* [ed:gyuri] blend above cover into image */
      gdSetMode(GD_BLEND, 128);
      gdImageCopy(v->image, v->scratch, 0, 0, 0, 0, v->width-1, v->height-1);
				/* restore copy mode */
      gdSetMode(GD_COPY, 128);
				/* erase scratch to background color */
/*      gdImageFilledRectangle(
			  v->scratch,
			  0, 0, v->width - 1, v->height - 1,
			  v->gc_init.color);
*/

    }
  }
  DoGyTrace = 0;
  gdSetMode(GD_BLEND, 128);	/* Blend worldmap and grid into image */
  ret = 0;

  ret = drawWorldMap(v->image,
		     v->layers,
		     v->resolution,
		     v->projection,
		     v->backface,
		     v->gridinc,
		     v->center_lat * DEG2RAD,
		     v->center_lon * DEG2RAD,
		     v->radius,
		     v->center_x,
		     v->center_y,
                     v->labelLatLong );


  DoGyTrace = 1;

  if (ret != 0) {
    fatal("Error %d reading WDBTILE map datafile", ret);
  }
				/* [ed:gyuri: drawCover loop was here] */
/*   testDown(); */
  return OK;
}

/* -------------------------------------------------------------------------
  Name: createGif()

  Description: Creates a .gif file from a gd image. A unique file name
               assigned to this file is reported on stdout in RFC-822 style
               in case of success. Otherwise Error is reported


  Returns: OK on success, FATAL on error

------------------------------------------------------------------------- */
static int createGif(View * v)
{
  int ret = OK;
  FILE *fout;
  static char pathname[1024];
  static int seq = 0;
  int the_time;
  int the_pid;

  if (!v || !v->image)
    return FATAL;

  /* This was changed since it is a highly unportable use of mktemp [eryq] */
  /* sprintf (pathname, "%s/imscovermapXXXXXX", wdir);
   *  mktemp (pathname);
   *  strcat (pathname, ".gif");
   */
  the_time = time(NULL);
  the_pid = getpid();
  sprintf(pathname, "%s/imscovermap%x%x%d.gif", wdir, the_time, the_pid, seq);
  ++seq;

  if (!(fout = fopen(pathname, "wb"))) {
    errOut(FATAL, "Can't open output file:", pathname);
    ret = FATAL;
  }
  else {
    gdImageGif(v->image, fout);
    if (ferror(fout)) {
      errOut(FATAL, "Error writing gif file: ", pathname);
      ret = FATAL;
    }
  }

  if (ret == OK)
    errOut(OK, "Filename:", pathname);


  fclose(fout);

  if (0) {			/*  look at scratch too */
    fout = fopen("s.gif", "wb");
    if (fout)
      gdImageGif(v->scratch, fout);
    fclose(fout);
  }
  return ret;
}

/* -------------------------------------------------------------------------
  Name: parseCmd()

  Description: Parses the command line and sets global variables
               for verbosity and working dir

  Returns: OK if default mode was used ('-i') or INFO if -i was not given

------------------------------------------------------------------------- */
static int parseCmd(int argc, char **argv)
{
  int ret = OK;
  int def_mode = 0;
  int i;


  /* Get current working dir */

  if (!(wdir = getcwd(NULL, 2048))) {
    errOut(WARN, "Can't get fully qualified CWD name. Using ./", "");
    wdir = ".";
  }

  /* Parse command line */

  for (i = 1; i < argc; i++) {
    if (strncmp(argv[i], "-d", 2) == 0) {	/* work dir */
      int len = (int) strcspn(&argv[i][2], " \t");
      if (!len) {
	errOut(WARN, "Bad working dir specification. Using current dir:",
	       wdir);
      }
      else {
	wdir = (char *) malloc(len + 1);
	strncpy(wdir, &argv[i][2], len);
	wdir[len] = 0;
      }
    }
    else if (strncmp(argv[i], "-v", 2) == 0) {	/* verbosity */
      if (!isdigit(argv[i][2])) {
	errOut(WARN, "Bad verbosity value. Staying silent (-v0).", "");
      }
      else {
	verbose = atoi(&argv[i][2]);
      }
    }
    else if (strncmp(argv[i], "-i", 2) == 0) {	/* default mode */
      def_mode = 1;
    }
  }

  if (!def_mode) {
    errOut(INFO, "imscovermap usage:\n"
	   "  imscovermap [-i] [-v<verbosity>] [-d<dir>]\n"
	   "  -i - take input parameters from standard input\n"
	   "       (normal mode of operation)\n"
	   "  -v0 .. -vn - verbosity level (-v0 - silent)\n"
	   "  -d<dir> - use working directory dir. Default - ./\n", "");
    ret = INFO;
  }
  return ret;
}


/* -------------------------------------------------------------------------
  Name: initArg()

  Description: Initializes input parameters - meaning nullifies the 'given'
               counter so later we can see if mandatory parameters are given

  Returns: void

------------------------------------------------------------------------- */
static void initCmdParm(void)
{
  int i;
  for (i = 0; i < num_cmd_parm; i++)
    cmd_parm[i].given = 0;
}

/* -------------------------------------------------------------------------
  Name: parseInput()

  Description: Processes keyword/value pairs in RFC-822 standard
               Creates the view according to input parameters
               Reports all errors/warnings encountered
               Stops on empty line or EOF

  Returns:     OK on success, even if there were warnings
               FATAL on unrecoverable error

------------------------------------------------------------------------- */
static int parseInput(void)
{
  int i, ret = OK;
  static char str[2048];
  int proj, width, height, clat, clon, bface = 0, polit = 1, lakriv = 1, latlon_label = 0;
  int stamp_polit = 0, stamp_lakriv = 0;
  double zoom = 1.0;
  Cover *cov = NULL;
  View *view;
  View **viewp;
  int wantPostageStamps = 0;
  int wantBlocksLabeled = 0;
  int grid = 20;
  char image_filename[80] = "\n";
  initCmdParm();

  while (!feof(stdin) && !ferror(stdin)) {


    /* Get the keyword/value pair */

    if (fgets(str, sizeof(str), stdin)) {
      int par = -1;
      char *nl;

      /* Capitalize the string and remove trailing \n */

      for (i = 0; i < strlen(str); i++)
	str[i] = toupper(str[i]);
      if ((nl = strrchr(str, '\n')))
	*nl = 0;

      /* If it was an empty string - just quit */

      if (!strlen(str))
	break;

      /* Check if the keyword is valid */

      for (i = 0; i < num_cmd_parm; i++) {
	if (strncmp(cmd_parm[i].keyword, str, 
		    strlen(cmd_parm[i].keyword)) == 0) {
	  par = i;
	  cmd_parm[i].given++;
	  break;
	}
      }

      if (par < 0) {		/* invalid keyword */
	errOut(WARN, "Invalid keyword/value: ", str);
      }
      else {

	/* extract pure value from the string */

	char *val = &str[strlen(cmd_parm[par].keyword)];
	val += strspn(val, " \t");	/* remove leading whitespaces */

	/* Process the value */

	switch (cmd_parm[par].code) {
	case KW_PROJECTION:

	  if (strncmp(val, "ORTHO", 5) == 0)
	    proj = ORTHOGRAPHIC;
	  else if (strncmp(val, "PLATE", 5) == 0)
	    proj = MERCATOR;
	  else {
	    errOut(FATAL, "Wrong projection type: ", str);
	    ret = FATAL;
	  }
	  break;

        case KW_GIFIMAGE:

           if (sscanf(val, "%s", &image_filename) != 1) {
            errOut(FATAL, "Couldn't parse gifImage : ", str);
            ret = FATAL;
          }
          break;


	case KW_DIMENSIONS:

	  if (sscanf(val, "%d %d", &width, &height) != 2 ||
	      width < 1 || height < 1) {
	    errOut(FATAL, "Invalid image dimensions: ", str);
	    ret = FATAL;
	  }
	  break;

	case KW_CENTER:

	  if (sscanf(val, "%d %d", &clat, &clon) != 2 ||
	      clat < -90 || clat > 90 || clon < -180 || clon > 180) {
	    errOut(FATAL, "Invalid map center coordinate(s): ", str);
	    ret = FATAL;
	  }
	  break;

	case KW_ZOOM:
	  {
	    double z;

	    if (sscanf(val, "%lf", &z) != 1 || z < 1. || z > 20.) {
	      errOut(WARN, "Invalid zoom factor value: ", str);
	    }
	    else
	      zoom = z;
	  }
	  break;

        case KW_GRID:
          {
            int g;

            if (sscanf(val, "%i", &g) != 1 || g < 0 || g > 20) {
              errOut(WARN, "Invalid grid value: ", str);
            }
            else
              grid = g;
              fprintf(stderr, "X-Warn: %i \n ", grid);
          }
          break;

          case KW_LABEL_GRIDS:

          if (strncmp(val, "ON", 2) == 0)
            latlon_label = 1;
          else if (strncmp(val, "OFF", 3) == 0)
             latlon_label = 0;
          else {
            errOut(WARN, "Bad Label Grids flag value: ", str);
          }
          break;

	case KW_TRANSPARENT:

	  if (strncmp(val, "YES", 3) == 0)
	    bface = 1;
	  else if (strncmp(val, "NO", 2) == 0)
	    bface = 0;
	  else {
	    errOut(WARN, "Invalid transparency value: ", str);
	  }
	  break;

	case KW_STAMPS:

	  if (strncmp(val, "YES", 3) == 0) {
	    wantPostageStamps = 1;
	    errOut(WARN, "Will do stamps", "");
	  }
	  else if (strncmp(val, "NO", 2) == 0)
	    wantPostageStamps = 0;
	  else {
	    errOut(WARN, "Invalid 'stamps' value: ", str);
	  }
	  break;

        case KW_LABELS:

          if (strncmp(val, "YES", 3) == 0) {
             wantBlocksLabeled = 1;
            errOut(WARN, "Will Label Blocks", "");
          }
          else if (strncmp(val, "NO", 2) == 0)
              wantBlocksLabeled = 0;
          else {
            errOut(WARN, "Invalid 'labels' value: ", str);
          }
          break;

	case KW_POLITICAL:

	  if (strncmp(val, "YES", 3) == 0)
	    polit = 1;
	  else if (strncmp(val, "NO", 2) == 0)
	    polit = 0;
	  else {
	    errOut(WARN, "Bad political boundaries layer flag value: ", str);
	  }
	  break;

	case KW_LAKESRIVERS:

	  if (strncmp(val, "YES", 3) == 0)
	    lakriv = 1;
	  else if (strncmp(val, "NO", 2) == 0)
	    lakriv = 0;
	  else {
	    errOut(WARN, "Bad lakes/rivers layer flag value: ", str);
	  }
	  break;

	case KW_STAMPPOLITICAL:

	  if (strncmp(val, "YES", 3) == 0)
	    stamp_polit = 1;
	  else if (strncmp(val, "NO", 2) == 0)
	    stamp_polit = 0;
	  else {
	    errOut(WARN, "Bad (stamp) political boundaries layer flag value: ", str);
	  }
	  break;

	case KW_STAMPLAKESRIVERS:

	  if (strncmp(val, "YES", 3) == 0)
	    stamp_lakriv = 1;
	  else if (strncmp(val, "NO", 2) == 0)
	    stamp_lakriv = 0;
	  else {
	    errOut(WARN, "Bad (stamp) lakes/rivers layer flag value: ", str);
	  }
	  break;

	case KW_POINT:		/* eryq */
	  {
	    Cover *c = initCover(COV_POINT, 0);
	    if (sscanf(val, "%lf %lf",
		       &c->lat[0], &c->lon[0]) != 2 ||
		c->lat[0] < -90. || c->lat[0] > 90. ||
		c->lon[0] < -180. || c->lon[0] > 180.) {
	      errOut(WARN, "Bad range lat/lon value(s): ", str);
	    }
	    else {
	      Cover *cp;
	      int i;
	      c->lon[1] = c->lon[2];
	      c->lat[1] = c->lat[0];
	      c->lon[3] = c->lon[0];
	      c->lat[3] = c->lat[2];

	      /* Normalize lat/lon */
	      c->npoints = 1;
	      for (i = 0; i < c->npoints; i++) {
		c->lat[i] *= DEG2RAD;
		c->lon[i] *= DEG2RAD;
	      }

	      /* Attach new coverage to coverage list */
	      if (!cov) {
		cov = c;
	      }
	      else {
		for (cp = cov; cp->next; cp = cp->next);
		cp->next = c;
	      }
	    }
	    break;
	  }

	case KW_RANGE:
	  {
	    Cover *c = initCover(COV_RANGE, 0);
	    if (sscanf(val, "%lf %lf %lf %lf",
		     &c->lat[0], &c->lon[0], &c->lat[2], &c->lon[2]) != 4 ||
		c->lat[0] < -90. || c->lat[0] > 90. ||
		c->lat[2] < -90. || c->lat[2] > 90. ||
		c->lon[0] < -180. || c->lon[0] > 180. ||
		c->lon[2] < -180. || c->lon[2] > 180.) {
	      errOut(WARN, "Bad range lat/lon value(s): ", str);
	    }
	    else {
	      Cover *cp;
	      int i;

	      c->lon[1] = c->lon[2];
	      c->lat[1] = c->lat[0];
	      c->lon[3] = c->lon[0];
	      c->lat[3] = c->lat[2];

                               /* Record current color in effect */
              c->r = CurrentColor.r;
              c->g = CurrentColor.g;
              c->b = CurrentColor.b;

				/* Convert degrees to radians */

	      /* Normalize lat/lon */
	      for (i = 0; i < c->npoints; i++) {
		c->lat[i] *= DEG2RAD;
		c->lon[i] *= DEG2RAD;
	      }

	      /* Attach new coverage to coverage list */
	      if (!cov) {
		cov = c;
	      }
	      else {
		for (cp = cov; cp->next; cp = cp->next)
		  ;
		cp->next = c;
	      }
	    }
	    break;
	  }

	case KW_QRANGE:
	  {
	    Cover *c = initCover(COV_QRANGE, 0);
	    char cov_label[256];
	    if (sscanf(val, "%s %lf %lf %lf %lf", 
				/* WARNING! If you change the order here
				 also make sure that trixels.c knows about 
				 it. Look for "imscovermap.c" in the comments*/

		       /* Switch from SW-NE to NW-SE ( JSA )
		  &c->lat[0], &c->lon[0], &c->lat[2], &c->lon[2]) != 4 || */
		      cov_label,
		     &c->lat[2], &c->lon[0], &c->lat[0], &c->lon[2]) != 5 ||
		c->lat[0] < -90. || c->lat[0] > 90. ||
		c->lat[2] < -90. || c->lat[2] > 90. ||
		c->lon[0] < -180. || c->lon[0] > 180. ||
		c->lon[2] < -180. || c->lon[2] > 180.) {
	      errOut(WARN, "Bad qrange lat/lon value(s): ", str);
	    }
	    else {
	      Cover *cp;
	      int i;
	      extern double fabs();
				/* WARNING! Depends on the fact that the
				 user supplied data is lat[0] and lat[2]
				 GyF */


	      if (fabs(c->lat[2] - 90.0) <= 0.01){
		c->pole_inside = POLE_NORTH;
	      }

	      c->lon[1] = c->lon[2];
	      c->lat[1] = c->lat[0];
	      c->lon[3] = c->lon[0];
	      c->lat[3] = c->lat[2];

	      if (strlen(cov_label) > 6){
		cov_label[6] = '\0';
	      }
	      strcpy(c->label, cov_label);

				/* Record current color in effect */
	      c->r = CurrentColor.r;
	      c->g = CurrentColor.g;
	      c->b = CurrentColor.b;
				/* [ed:gyf Check for zero-width feature */
	      {
		double area;
		extern double fabs();
		area = fabs((c->lat[2] - c->lat[0]) * (c->lon[2] - c->lon[0]));
		if (area < 0.000001){
		  c->zerosize = 1;
		}
	      }

	      /* Normalize lat/lon */
	      for (i = 0; i < c->npoints; i++) {
		c->lat[i] *= DEG2RAD;
		c->lon[i] *= DEG2RAD;
	      }

	      /* Attach new coverage to coverage list */
	      if (!cov) {
		cov = c;
	      }
	      else {
		for (cp = cov; cp->next; cp = cp->next)
		  ;
		cp->next = c;
	      }
	    }
	    break;
	  }

	case KW_COLOR:		/* [ed:gyuri:new] */
	  {
	    int r, g, b;
				/* [ed:gyuri:todo] */
	    /*
	      Polygon specs should also include the following:
	      1: color			R,G,B
	      2: writing rule
	      Writing rule is one of:	GX_COPY, GX_BLEND,
	      Default is GX_COPY, so old stuff does not break.
	      Color won't actually be allocated until the polygon
	      is drawn to the gis image.
	      */
	    if (sscanf(val, "%d %d %d", &r, &g, &b) != 3) {
	      errOut(WARN, "Bad color specs: ", str);
	    } else {
	      ColorType *ct = &CurrentColor;
	      ct->r = r;
	      ct->g = g;
	      ct->b = b;
	    }
	  }
	break;

	case KW_POLYGON:
	  {
	    Cover *c = initCover(COV_POLYGON, 0);
	    char p;
            char cov_label[256];
            char path[50];
	    if (sscanf(val, "%s %s %c %lf %lf %lf %lf %lf %lf %lf %lf",
		       &c->path, cov_label, &p, &c->lat[0], &c->lon[0],
		       &c->lat[1], &c->lon[1],
		       &c->lat[2], &c->lon[2],
		       &c->lat[3], &c->lon[3]) != 11 ||
		(p != 'Y' && p != 'N') ||
		c->lat[0] < -90. || c->lat[0] > 90. ||
		c->lat[1] < -90. || c->lat[1] > 90. ||
		c->lat[2] < -90. || c->lat[2] > 90. ||
		c->lat[3] < -90. || c->lat[3] > 90. ||
		c->lon[0] < -180. || c->lon[0] > 180. ||
		c->lon[1] < -180. || c->lon[1] > 180. ||
		c->lon[2] < -180. || c->lon[2] > 180. ||
		c->lon[3] < -180. || c->lon[3] > 180.) {
	      errOut(WARN, "Bad polygon specification: ", str);
	    } else {
	      Cover *cp;
	      int i;

              if (strlen(cov_label) > 7) {
                cov_label[7] = '\0';
              }
              strcpy(c->label, cov_label);
                                /* Record current color in effect */
	      c->r = CurrentColor.r;
	      c->g = CurrentColor.g;
	      c->b = CurrentColor.b;
				/* Normalize lat/lon */
	      for (i = 0; i < c->npoints; i++) {
		c->lat[i] *= DEG2RAD;
		c->lon[i] *= DEG2RAD;
	      }
	      c->pole_inside = (p == 'Y');

	      /* Attach new coverage to coverage list */
	      if (!cov)
		cov = c;
	      else {
		for (cp = cov; cp->next; cp = cp->next);
		cp->next = c;
	      }
	    }
	    break;
	  }
	case KW_QMESH:
	  {
	    Cover *c = initCover(COV_QMESH, 0);	/* [ed:WARNING:memory leak] */
	    char p;
	    char cov_label[256];
	    int ns;
	    ns = sscanf(val, "%s %c %lf %lf %lf %lf %lf %lf %lf %lf",
		       cov_label, &p, 
		       &c->lat[0], &c->lon[0],
		       &c->lat[1], &c->lon[1],
		       &c->lat[2], &c->lon[2],
		       &c->lat[3], &c->lon[3]);
	    if (ns != 10 ||
		(p != 'G' && p != 'S' && p != 'X' && p != 'N' && p != 'B') ||
		c->lat[0] < -90. || c->lat[0] > 90. ||
		c->lat[1] < -90. || c->lat[1] > 90. ||
		c->lat[2] < -90. || c->lat[2] > 90. ||
		c->lat[3] < -90. || c->lat[3] > 90. ||
		c->lon[0] < -180. || c->lon[0] > 180. ||
		c->lon[1] < -180. || c->lon[1] > 180. ||
		c->lon[2] < -180. || c->lon[2] > 180. ||
		c->lon[3] < -180. || c->lon[3] > 180.) {
	      errOut(WARN, "Bad qmesh specification: ", str);
	    } else {
	      Cover *cp;
	      int i;
				/* Check cov_label length */
				/* truncate to 6 if necessary */
	      if (strlen(cov_label) > 6) {
		cov_label[6] = '\0';
	      }
	      strcpy(c->label, cov_label);
				/* Record current color in effect */
	      c->r = CurrentColor.r;
	      c->g = CurrentColor.g;
	      c->b = CurrentColor.b;
				/* Convert degrees to radians */
	      for (i = 0; i < c->npoints; i++) {
		c->lat[i] *= DEG2RAD;
		c->lon[i] *= DEG2RAD;
	      }
	      c->pole_inside = POLE_NONE; /* default assumption */
	      switch (p) {
	      case 'S':
		c->pole_inside = POLE_SOUTH;
		break;
	      case 'N':
		c->pole_inside = POLE_NORTH;
		break;
	      case 'B':
		c->pole_inside = POLE_BOTH;
		break;
	      case 'G':
		c->type = COV_GLOBAL;
		break;
	      default:
		break;
	      }

	      /* Attach new coverage to coverage list */
	      if (!cov)
		cov = c;
	      else {

		for (cp = cov; cp->next; cp = cp->next);
		cp->next = c;
	      }
	    }
	    break;
	  }
	case KW_QRING:
	  {
				/* GRings are treated differently */
				/* Each line can spcify the same gring */
				/* Only add covermap once. */

	    char cov_label[256];
	    int parseStatus;
	    parseStatus = GRingParse(str, cov_label);
	    if (parseStatus < 0){
	      errOut(WARN, "Syntax error: ", str);
	    } else if (parseStatus == 0){ /* first time with this label */
	      Cover *c = initCover(COV_QGRING, 0);
	      Cover *cp;
	      c->r = CurrentColor.r;
	      c->g = CurrentColor.g;
	      c->b = CurrentColor.b;
	      strcpy(c->label, cov_label);


				/* Grings are kept separate */
				/* Symbolic names only are entered */

				/* Attach new coverage to coverage list */
	      if (!cov)
		cov = c;
	      else {
		for (cp = cov; cp->next; cp = cp->next);
		cp->next = c;
	      }
	    } 
	  }
	  break;
	case KW_QRESOLUTION:
	   {
	     int res;
	     if (sscanf(val, "%d", &res) != 1){
		errOut(WARN, "Bad Qresolution spec: ", str);
	     } else {
		setQresolution(res);
	     }
	     break;
	   }
	case KW_TEST:
	  errOut(WARN, "Reserved for testing", "");
	  break;
	default:
	  errOut(WARN, "Invalid keyword code (should not happen)", "");
	}
      }
    }
    if (ret != OK)
      break;
  }
  if (ferror(stdin)) {
    fatal("Error reading standard input!");
  }

  /* Check if all mandatory parameters were given */
  for (i = 0; i < num_cmd_parm; i++) {
    if (cmd_parm[i].mandatory && !cmd_parm[i].given) {
      fatal("Mandatory parameter not given:", cmd_parm[i].keyword);
    }
  }


  /*
   * Create new primary view:
   */



    /* Bug: Traditionally, Characters are converted to uppercase, but we're 
    dealing with subdirectories. So we'll have a problem if the directory
    is suppose to have UPPERCASE/Mixed directories because we're lowercasing
    the directory */
    for (i = 0; i < strlen(image_filename); i++)
        image_filename[i] = tolower(image_filename[i]);

  /* Create view: */
  if (!(view = initView(width, height, proj, zoom, bface, clat, clon,grid, image_filename)))
    fatal("malloc() error in initView()");

  /* Set layers and attach coverage to the view structure */
  view->layers[RIVER] = lakriv;
  view->layers[IBOUND] = view->layers[NBOUND] = polit;
  view->cover = cov;

  view->displayLabels = wantBlocksLabeled;
  view->labelLatLong = latlon_label; 

  if ( strlen(image_filename) > 1 ) {
      strcpy ( view->baseImage, (image_filename));
  }



  /* Store it at the end of the view array: */
  viewp = gViews;
  *viewp++ = view;
  *viewp = NULL;


  /*
   * If desired, create 6 postage stamps:
   */

  if (wantPostageStamps) {
    int c;
    for (c = 0; c < 6; c++) {	/* for each center point */

      /* Create view: */
      if (!(view = initView(60, 60, ORTHOGRAPHIC, 1, 0, gStampCenters[c].clat, 
                           gStampCenters[c].clon,20, image_filename)))
	fatal("malloc() error in initView() for postage stamp %d", c);

      /* Set layers and attach coverage to the view structure */
      view->layers[RIVER] = stamp_lakriv;
      view->layers[IBOUND] = view->layers[NBOUND] = stamp_polit;
      view->cover = cov;
      view->gridinc = 0;


      /* Store it at the end of the view array: */
      *viewp++ = view;
      *viewp = NULL;
    }
  }

  return OK;
}

/* -------------------------------------------------------------------------
  main()
------------------------------------------------------------------------- */
int main(int argc, char **argv)
{
  View **viewp;
  int err;


  /* Parse command line: */
  if ((err = parseCmd(argc, argv)) != OK)
    fatal("couldn't parse command");

  /* Parse input stream: */
  if ((err = parseInput()) != OK)
    fatal("couldn't parse input");

  /* At this point, the global array "gViews" is initiallized: */
				/* but colors for polygons have not yet
				 been allocated */
  for (viewp = gViews; *viewp; viewp++) {
    if ((err = drawView(*viewp)) != OK)
      fatal("couldn't draw view");

    if ((err = createGif(*viewp)) != OK)
      fatal("couldn't create GIF");

  }


  printf("\n");			/* Final \n */
  return err;
}
