/************************************************************
 * imscovertime.c
 ************************************************************
 *
 * DESCRIPTION
 *    Generate the GIF for the temporal coverage map in the
 *    IMS/WWW interface.
 *
 * AUTHOR
 *    Sze Hong (HSTX) wrote the original code.
 *    Eryq (HSTX) adapted it for more comprehensive graphing.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>

#include "gd.h"
#include "gdfontt.h"

#include "IW_Str.h"
#include "IW_New.h"


/************************************************************
 *
 * Contants
 *
 ************************************************************/

/* The GIF files: */
#define IA_FILENAME "covertime.gif"

/* Limits: */
#define IA_MAXRANGE 1000      /* Maxium granule in graph */
#define IA_MAXLABEL 31       /* Maximum length of a label */

/* Modes: */
#define IA_MODE_BOGUS 0
#define IA_MODE_INPUT 1

/* Graphing offsets: */
#define IA_LABELMARGIN 10       /* from left edge to left edge of labels */
                                /* from left edge to left edge of graph: */
#define IA_LMARGIN (gMaxLabelLength + (2 * IA_LABELMARGIN))
#define IA_RMARGIN 30           /* from right edge of graph to right edge */
#define IA_TMARGIN 30           /* from top edge to top edge of graph */
#define IA_BMARGIN 30           /* from bottom edge of graph to bottom edge */
#define IA_BARSEP   4
#define IA_BARHGT  (gInput.barHeight)
#define IA_MINORTICKLEN 3
#define IA_MAJORTICKLEN 6
#define IA_ARROWLEN     6

/* Conversion: */
#define IA_DAYS_IN_A_YEAR 365
#define IA_DAYS_IN_A_MONTH 30
#define IA_SECS_PER_DAY (24 * 60 * 60)

/* Tick units: */
#define IA_UNIT_YEARS   1
#define IA_UNIT_MONTHS  2
#define IA_UNIT_DAYS    3
#define IA_UNIT_HOURS   4
#define IA_UNIT_MINUTES 5

/* Alignment: */
#define IA_ALIGN_LEFT   1
#define IA_ALIGN_CNTR   2


/************************************************************
 *
 * Types
 *
 ************************************************************/

typedef long IA_Julian;

typedef struct IA_Abstime {   /* an absolute time */
  IA_Julian jd;             /* julian day */
  int secs;                 /* seconds after start of day */
} IA_Abstime;

typedef struct IA_Time { /* structure to hold the date-time info */
   int yr;                  /* the year */
   int mon;                 /* the month */
   int day;                 /* the day */
   int hr;                  /* the hour value */
   int min;                 /* the min value */
   int sec;                 /* the sec value */   
   IA_Abstime abs;          /* absolute time */
} IA_Time;

typedef struct IA_Range { /* structure to hold the range */
   char *label;             /* label for the granule */
   IA_Time start;           /* start date */
   IA_Time stop;            /* stop time */
} IA_Range;

typedef struct IA_Input { /* struct to hold the whole time coverage info */
   IA_Time start;           /* the earliest start time for all granules */
   IA_Time stop;            /* the latest end time for all granules */
   int totalWidth;          /* the total width of the gif image in pixels */
   int barHeight;           /* the height of each bar */
   int nranges;             /* total no. of granule */
   IA_Range *ranges[IA_MAXRANGE]; /* array to hold the coverage ranges */
} IA_Input;


/************************************************************
 *
 * Globals
 *
 ************************************************************/

/* Mode: */
static int gMode = IA_MODE_BOGUS;

/* Verbosity: 
 *    0  just errors and main info
 *    1  all of the above, plus warnings
 *    2  all of the above, plus debug messages
 */
static int gVerbosity = 1;

/* The coverage: */
static IA_Input gInput;

/*
 * The following defines the first day that the Gregorian calendar was used
 * in the British Empire (Sep 14, 1752).  The previous day was Sep 2, 1752
 * by the Julian Calendar.  The year began at March 25th before this date.
 */
IA_Julian gBritishEpoch = 2361222;

/* The borders of the *selected viewable range* within the graph: */
static IA_Time gSlctStart;    /* left edge of selection */
static IA_Time gSlctStop;     /* right edge of selection */
static IA_Abstime gSlctLen;   /* length of selection */

/* The borders of the graph: */
static IA_Time gGraphStart;    /* left edge of graphed area */
static IA_Time gGraphStop;     /* right edge of graphed area */
static IA_Abstime gGraphLen;   /* length of graphed area */

/* The major ticks: */
static int gMajorTick;
static int gMajorUnits;
static int gMajorDivs;        /* number of minor divisions */

/* Graphing and conversion: */
static int gGraphWidth = 0;      /* width of the graph rectangle */
static int gGraphHeight = 0;     /* height of the graph rectangle */
static int gGifWidth = 0;        /* width of the entire GIF */
static int gGifHeight = 0;       /* height of the entire GIF */
static double gPixelsPerDay = 0;
static double gPixelsPerSec = 0;
static int gMaxLabelLength = 0;  /* max label length, in pixels */

/* Colors: */
static int gColorTrans;  /* transparent */
static int gColorBlack;  /* foreground (text, tickmarks, axes) color */
static int gColorGrey;   /* light-foreground color */
static int gColorWhite;  /* background color, of graph */
static int gColorData;   /* color of data */




/*********************************************************************
 *
 * Function Prototypes
 *
 *********************************************************************/




/************************************************************
 *
 * Output...
 *
 ************************************************************/

/************************************************************
 * IA_VMesg -- back-end for message output
 ************************************************************/
void IA_VMesg(char *kwd, char *fmt, va_list args) {
  fprintf(stdout, "%s: ", kwd);
  vfprintf(stdout, fmt, args);
  fprintf(stdout, "\n");
} 

/************************************************************
 * IA_Debug -- output a debug message to stderr
 ************************************************************/
int IA_Debug(char *fmt, ...)
{
  va_list args;

  if (gVerbosity < 2) return 0;

  fflush(stdout);
  va_start(args, fmt);
  fflush(stdout);
  fprintf(stdout, " ** ");
  vfprintf(stdout, fmt, args);
  fprintf(stdout, "\n");
  va_end(args);
  fflush(stdout);
  return 0;
}

/************************************************************
 * IA_Warn -- output a warning message
 ************************************************************/
 IA_Warn(char *fmt, ...) 
{
  va_list args;

  if (gVerbosity < 1) return 0;

  va_start(args, fmt);
  IA_VMesg("X-warn", fmt, args);
  va_end(args);
  return 0;
}

/************************************************************
 * IA_Error -- output an error message
 ************************************************************/
int IA_Error(char *fmt, ...) 
{
  va_list args;
  va_start(args, fmt);
  IA_VMesg("X-error", fmt, args);
  va_end(args);
  return -1;
}

/************************************************************
 * IA_Fatal -- output an error message, and die
 ************************************************************/
void IA_Fatal(char *fmt, ...) 
{
  va_list args;
  va_start(args, fmt);
  IA_VMesg("X-error", fmt, args);
  va_end(args);
  printf("\n");
  exit(-1);
}



/************************************************************
 *
 * IA_Julian...
 *
 ************************************************************/


/************************************************************
 * IA_JulianFromYMD
 ************************************************************
 *
 * DESCRIPTION
 *    Calculate the julian day, given year, month (1-12), and 
 *    day of month.
 *
 * AUTHOR
 *    Perl algorithm by David Muir Sharnoff <muir@idiom.com>.
 *
 * NOTES
 *    No idea how this works, or what the "zero day" is, except that
 *    it appears to be fairly B.C.   If it's any help, 1/1/1970
 *    has a Julian day of 2440588, so I think we're looking at 
 *    "epoch" being approximately 4700 B.C.
 */
IA_Julian IA_JulianFromYMD(int year, int month, int day)
{
  IA_Julian tmp;

  tmp = day - 32075
    + (1461 * (year + 4800 - ((14 - month)/12))/4)
      + (367 * (month - 2 + ((14 - month)/12)*12)/12)
	- (3 * ((year + 4900 - (14 - month)/12)/100)/4);
  return tmp;
}			 

/************************************************************
 * IA_JulianToYMD
 ************************************************************
 *
 * DESCRIPTION
 *    Convert a julian day to year, month, and day.
 *
 * WARNINGS
 *    The first day that the Gregorian calendar was used in the British 
 *    Empire was Sep 14, 1752.  The previous day was Sep 2, 1752
 *    by the Julian Calendar.  The year began at March 25th before this date.
 *
 *    If the Julian day pre-dates the British use of the Gregorian
 *    calendar (September 1752), the conversion will not be accurate.
 *    I suppose this could be corrected.
 */ 

IA_Julian IA_JulianToYMD(IA_Julian jd,
			 int *yearp, int *monp, int *dayp)
{
  IA_Julian jd_tmp;
  
  if (jd < gBritishEpoch) {
    /* place warning here */
  }
  
  jd_tmp = jd - 1721119;
  *yearp = ((4 * jd_tmp - 1)/146097);
  jd_tmp = 4 * jd_tmp - 1 - 146097 * *yearp;
  
  *dayp = (jd_tmp/4);
  jd_tmp = ((4 * *dayp + 3)/1461);
  *dayp = 4 * *dayp + 3 - 1461 * jd_tmp;
  *dayp = ((*dayp + 4)/4);
  *monp = ((5 * *dayp - 3)/153);
  *dayp = 5 * *dayp - 3 - 153 * *monp;
  *dayp = ((*dayp + 5) / 5);
  *yearp = 100 * *yearp + jd_tmp;
  if (*monp < 10) {
    *monp += 3;
  } 
  else {
    *monp -= 9;
    ++*yearp;
  }
  return jd;
}




/************************************************************
 *
 * IA_Time...
 *
 ************************************************************/

/************************************************************
 * IA_TimeFmt -- format a time
 ************************************************************/ 

char *IA_TimeFmt(IA_Time *this)
{
  static char buf[32];

  sprintf(buf, "%04d-%02d-%02d %02d:%02d:%02d", 
	  this->yr, this->mon, this->day, this->hr, this->min, this->sec);
  return buf;
}

/************************************************************
 * IA_TimeCopy -- copy a time
 ************************************************************/ 

void IA_TimeCopy(IA_Time *this, IA_Time *other)
{
  memcpy(this, other, sizeof(IA_Time));  
}

/************************************************************
 * IA_TimeCmp -- compare two times
 ************************************************************/ 

int IA_TimeCmp(IA_Time *this, IA_Time *other)
{
  if (this->abs.jd != other->abs.jd) {
    return (this->abs.jd - other->abs.jd);
  }
  return (this->abs.secs - other->abs.secs);
}

/************************************************************
 * IA_TimeParse -- parse in Time from a string
 ************************************************************
 *
 * DESCRIPTION
 *    The string must be of the ODL GMT form of "yyyy-mm-ddThh:mm:ssZ"; 
 *    e.g., "1992-09-23T15:57:00Z". 
 */ 

int IA_TimeParse(IA_Time *this, char *str)
{
  /* Extract the time parts: */
  if (sscanf(str, "%d-%d-%dT%d:%d:%dZ", 
	     &(this->yr), &(this->mon), &(this->day),
	     &(this->hr), &(this->min), &(this->sec)) != 6) {
    return IA_Error("Bad date: <%s>", str);
  }

  /* Convert to an absolute time: */
  IA_TimeRecalcAbstime(this);

  IA_Debug("Parsed <%s> into year %d, month %d, ..., sec %d",
	   str, this->yr, this->mon, this->sec);
  return 0;
}

/************************************************************
 * IA_TimeRecalcAbstime -- the timeparts have changed; recompute the abstime
 ************************************************************/

int IA_TimeRecalcAbstime(IA_Time *this)
{
  IA_AbstimeFromTime(&(this->abs), this);
}

/************************************************************
 * IA_TimeBelow -- compute the time under this time, aligned on the given unit
 ************************************************************/

int IA_TimeBelow(IA_Time *this, int nunits, int units, IA_Time *below)
{
  /* Copy the time: */
  IA_TimeCopy(below, this);

  /* Diddle with the new time: */
  switch (units) {

  case IA_UNIT_YEARS:     /* 1992-09-12T12:34:56Z => 1992-01-01T00:00:00Z */

    /* Go to beginning of year: */
    below->mon = below->day = 1;
    below->hr = below->min = below->sec = 0;

    /* Go to first year <= this one which is a multiple of nunits: */
    below->yr = ((below->yr / nunits) * nunits);
    break;

  case IA_UNIT_MONTHS:    /* 1992-09-12T12:34:56Z => 1992-09-01T00:00:00Z */

    /* Go to beginning of month: */
    below->day = 1;
    below->hr = below->min = below->sec = 0;

    /* Go to first month <= this one where (mon-1) is a multiple of nunits: */
    below->mon = (((below->mon - 1) / nunits) * nunits) + 1;
    break;

  case IA_UNIT_DAYS:      /* 1992-09-12T12:34:56Z => 1992-09-12T00:00:00Z */

    /* Go to beginning of day: */
    below->hr = below->min = below->sec = 0;

    /* Go to first day <= this one where day is a multiple of nunits: */
    below->day = ((below->day / nunits) * nunits);    
    break;

  case IA_UNIT_HOURS:     /* 1992-09-12T12:34:56Z => 1992-09-12T12:00:00Z */

    /* Go to beginning of hour: */
    below->min = below->sec = 0;

    /* Go to first hour <= this one where hour is a multiple of nunits: */
    below->hr = ((below->hr / nunits) * nunits);    
    break;

  case IA_UNIT_MINUTES:   /* 1992-09-12T12:34:56Z => 1992-09-12T12:34:00Z */

    /* Go to beginning of minute: */
    below->sec = 0;

    /* Go to first minute <= this one where minute is a multiple of nunits: */
    below->min = ((below->min / nunits) * nunits);    
    break;

  default:
    IA_Fatal("Unsupported case in IA_TimeBelow");
  }

  /* Recalc the absolute time: */
  IA_TimeRecalcAbstime(below);
  IA_Debug("Time below = <%s>", IA_TimeFmt(below));
  return 0;
}     


/************************************************************
 * IA_TimeAddUnit -- add a conceptual unit of time
 ************************************************************/

void IA_TimeAddUnit(IA_Time *this, int n, int units) 
{
  int nmon;
  int carry, rem;   


  /* Diddle with the new time: */
  switch (units) {
  case IA_UNIT_YEARS:
    this->yr += n;
    break;

  case IA_UNIT_MONTHS:
    IA_Debug("++ Adding %d mons to [%ld,%d]", n, this->abs.jd, this->abs.secs);
    nmon = (this->mon - 1 + n);
    this->mon = (nmon % 12) + 1;
    this->yr += (nmon / 12);
    break;

  case IA_UNIT_DAYS:
    IA_Debug("++ Adding %d days to [%ld,%d]", n, this->abs.jd, this->abs.secs);
    this->abs.jd += (n);
    IA_JulianToYMD(this->abs.jd, &(this->yr), &(this->mon), &(this->day));
    break;

  case IA_UNIT_HOURS:  
    IA_Debug("++ Adding %d hours to [%ld,%d]", n,this->abs.jd, this->abs.secs);
    carry = (this->hr + n) / 24;
    rem   = (this->hr + n) % 24;
    if (carry > 0) {
      IA_TimeAddUnit(this, carry, IA_UNIT_DAYS);
    }
    this->hr = rem;
    break;

  case IA_UNIT_MINUTES:
    IA_Debug("++ Adding %d mins to [%ld,%d]", n,this->abs.jd, this->abs.secs);
    carry = (this->min + n) / 60;
    rem   = (this->min + n) % 60;
    if (carry > 0) {
      IA_TimeAddUnit(this, carry, IA_UNIT_HOURS);
    }
    this->min = rem;
    break;

  default:
    IA_Fatal("Unsupported case in IA_TimeAddUnit");
  }

  /* Recalc the absolute time: */
  IA_TimeRecalcAbstime(this);
  IA_Debug("== Result is [%ld,%d]",this->abs.jd, this->abs.secs);
  return;
}


/************************************************************
 * IA_TimeLabel -- return a brief major-tick-label for this time
 ************************************************************/

char *IA_TimeLabel(IA_Time *this, int nunits, int units) 
{
  static char buf[32];
  static char *mabbrev2[] = {
    "Ja","Fe","Mr","Ap","My","Jn",
    "Jl","Au","Se","Oc","No","De"
  };
  int nmon;


  switch (units) {

  case IA_UNIT_YEARS:
    sprintf(buf, "%04d", this->yr);
    break;

  case IA_UNIT_MONTHS:
    sprintf(buf, "%s\n%02d", mabbrev2[this->mon - 1], (this->yr % 100));
    break;

  case IA_UNIT_DAYS: 
    sprintf(buf, "%d/%d", this->mon, this->day);
    break;

  case IA_UNIT_HOURS: 
    if (this->hr == 0) {
      sprintf(buf, "%d:00\n%d/%d", this->hr, this->mon, this->day);
    }
    else {
      sprintf(buf, "%d:00", this->hr);
    }
    break;

  case IA_UNIT_MINUTES:
    sprintf(buf, "%02d:%02d", this->hr, this->min);
    break;

  default:
    IA_Fatal("Unsupported case in IA_TimeLabel");
  }

  return buf;
}


/************************************************************
 *
 * IA_Abstime...
 *
 ************************************************************/


/************************************************************
 * IA_AbstimeSet -- set the absolute time
 ************************************************************/
IA_Abstime *IA_AbstimeSet(IA_Abstime *this, 
			  IA_Julian jd, int secs)
{
  this->jd = jd;
  this->secs = secs;
  return 0;
}

/************************************************************
 * IA_AbstimeParse -- parse in the absolute time
 ************************************************************/
int IA_AbstimeParse(IA_Abstime *this, char *str)
{
  IA_Time tp;

  /* Parse the string into its components: */
  if (IA_TimeParse(&tp, str) < 0) {
    return -1;
  }

  /* Use the Time to create the time: */
  if (IA_AbstimeFromTime(this, &tp) < 0) {
    return -1;
  }

  IA_Debug("Parsed <%s> into Julian %ld, secs %d", str, this->jd, this->secs);
  return 0;
}

/************************************************************
 * IA_AbstimeFromTime -- convert a date to an absolute time
 ************************************************************
 *
 * DESCRIPTION
 *    Given a set of numbers describing a GMT datetime, convert 
 *    to an absolute time.
 */ 

int IA_AbstimeFromTime(IA_Abstime *this, IA_Time *tp)
{
  this->jd = IA_JulianFromYMD(tp->yr, tp->mon, tp->day);
  this->secs = (tp->hr * 3600) + (tp->min * 60) + tp->sec;
  return 0;
}

/************************************************************
 * IA_AbstimeDiff -- diff = a - b
 ************************************************************/

void IA_AbstimeDiff(IA_Abstime *a, IA_Abstime *b, IA_Abstime *diff)
{
  int carry = 0;

  if ((diff->secs = a->secs - b->secs) < 0) {
    diff->secs += IA_SECS_PER_DAY;
    carry = 1;
  }
  diff->jd = a->jd - b->jd - carry;
}

/************************************************************
 * IA_AbstimeToGXCoord -- return graph x coord for this abs time
 ************************************************************
 *
 * NOTES
 *    Uses global conversion info.
 */

int IA_AbstimeToGXCoord(IA_Abstime *this)
{
  int fromDays, fromSecs = 0;
  IA_Julian jd_diff   = (this->jd   - gGraphStart.abs.jd);
  int       secs_diff = (this->secs - gGraphStart.abs.secs);

  IA_Debug("Abstime: [%ld, %d]", this->jd, this->secs);
  fromDays = (int)(jd_diff * gPixelsPerDay);
  fromSecs = (int)(secs_diff * gPixelsPerSec);
  IA_Debug("GXCoord: [%ld, %d] = %d from days, %d from secs = %d", 
	   jd_diff, secs_diff, fromDays, fromSecs, (fromDays + fromSecs));
  return (fromDays + fromSecs);
}


/************************************************************
 *
 * IA_Range...
 *
 ************************************************************/


/************************************************************
 * IA_RangeParseNew
 ************************************************************
 *
 * DESCRIPTION
 *    Parse a new range, and return it.
 */ 

IA_Range *IA_RangeParseNew(char *str)
{
  IA_Range *new;
  char label[256];
  char start[128];
  char stop[128];

  /* Allocate: */
  if (!(new = IW_New(IA_Range))) {
    IA_Fatal("no memory");
  }

  /* Scan parts: */
  if (sscanf(str, "%s %s %s", label, start, stop) != 3) {
    IA_Error("Bad range spec: <%s>", str);
    free(new);
    return NULL;
  }
  
  /* Initiallize: */
  new->label = IW_Strdup(label);
  IA_TimeParse(&(new->start), start);
  IA_TimeParse(&(new->stop), stop);
  
  return new;
}


/************************************************************
 *
 * Main 
 *
 ************************************************************/

/************************************************************
 * IA_Usage -- output usage error message to stderr
 ************************************************************/

void IA_Usage() 
{
  fflush(stdout);
  fprintf(stderr, "Usage:\n");
  fprintf(stderr, "   imscovertime -i [-v<verbosity>] [-d<directory>]\n");
  fprintf(stderr, "      <verbosity> is 0-3.\n");
  fprintf(stderr, "      <directory> is the working directory (if not cwd)\n");
  fprintf(stderr, "\n");
}

/************************************************************
 * IA_ParseArgs -- parse command-line args
 ************************************************************/

int IA_ParseArgs(int argc, char *argv[])
{
  char **argp;

  for (argp = argv+1; *argp; argp++) {
    if ((*argp)[0] == '-') {
      switch ((*argp)[1]) {

      case 'd':    /* directory */
	if (chdir((*argp)+2) < 0) {
	  IA_Fatal("Couldn't change directory to '%s'", ((*argp)+2));
	}
	break;

      case 'i':    /* input */
	gMode = IA_MODE_INPUT;
	break;

      case 'v':    /* verbosity */
	gVerbosity = atoi((*argp)+2);
	break;
	
      default:
	IA_Usage();
	IA_Fatal("Bad option, <%s>", *argp);
      }
    }
    else {
      IA_Usage();
      IA_Fatal("Bad argument, <%s>", *argp);
    }
  }

  /* Check final mode: */
  if (gMode != IA_MODE_INPUT) {
    IA_Fatal("Missing usage mode");
  }

  return 0;
}

/************************************************************
 * IA_ParseInput -- parse stdin input
 ************************************************************
 */

int IA_ParseInput(IA_Input *input)
{
  char buf[1024];
  char *pos;
  char *arg;

  /* Init: */
  memset(input, sizeof(IA_Input), 0);
  input->totalWidth = 300;
  input->barHeight = 8;

  /* Read lines until done: */
  while (fgets(buf, 1024, stdin)) {
    buf[strlen(buf)-1] = '\0';       /* chop */

    /* Stop at blank line: */
    if (!*buf) {
      break;
    }

    /* Find keyword end, and partition: */
    if (!(pos = strstr(buf, ": "))) {
      IA_Error("Malformed input line: <%s>", buf);
      continue;
    }
    *pos = '\0';
    arg = pos + 2;
    IW_Strdowncase(buf);

    /* Handle: */
    IA_Debug("Got input keyword: <%s>", buf);
    if (IW_Strequal(buf, "range")) {
      IA_Range *range;

      /* Bounds check: */
      if (input->nranges >= IA_MAXRANGE) {
	IA_Error("Range overflow: %d: ignoring", IA_MAXRANGE);
	continue;
      }

      /* Parse the range: */
      if (!(range = IA_RangeParseNew(arg))) {
	IA_Error("Bad range: <%s>", arg);
	continue;
      }

      /* Add it: */
      input->ranges[(input->nranges)++] = range;
    }

    else if (IW_Strequal(buf, "bar-height")) {
      input->barHeight = atoi(arg);
    }

    else if (IW_Strequal(buf, "total-width")) {
      input->totalWidth = atoi(arg);
    }

    else if (IW_Strequal(buf, "start-time")) {
      if (IA_TimeParse(&(input->start), arg) < 0) {
	IA_Fatal("Bad start time: <%s>", arg);
      }
    }

    else if (IW_Strequal(buf, "stop-time")) {
      if (IA_TimeParse(&(input->stop), arg) < 0) {
	IA_Fatal("Bad stop time: <%s>", arg);
      }
    }    

    else {
      IA_Error("Unrecognized keyword: <%s>", buf);
      continue;
    }
  }

  /* If no start date given, use earliest range date: */
  if (!(input->start.abs.jd)) {
    int i;
    IA_TimeCopy(&(input->start), &(input->ranges[0]->start));
    for (i = 1; i < input->nranges; i++) {
      if (IA_TimeCmp(&(input->start), &(input->ranges[i]->start)) > 0) {
	IA_TimeCopy(&(input->start), &(input->ranges[i]->start));
      }
    }
    IA_Debug("Filled in start date: <%s>", IA_TimeFmt(&(input->start)));
  }

  /* If no stop date given, use earliest range date: */
  if (!(input->stop.abs.jd)) {
    int i;
    IA_TimeCopy(&(input->stop), &(input->ranges[0]->stop));
    for (i = 1; i < input->nranges; i++) {
      if (IA_TimeCmp(&(input->stop), &(input->ranges[i]->stop)) < 0) {
	IA_TimeCopy(&(input->stop), &(input->ranges[i]->stop));
      }
    }
    IA_Debug("Filled in stop date: <%s>", IA_TimeFmt(&(input->stop)));
  }

  return 0;
}

/************************************************************
 * IA_gdImageStrings -- draw strings with newlines
 ************************************************************/

void IA_gdImageStrings(gdImagePtr im, gdFontPtr font,
		       int x, int y, char *strings, int color, 
		       int align)
{
  char sbuf[1024];
  char *sv[32];
  char **sp = sv;
  int xi = x, yi = y;

  /* Break strings up into array of strings, as delimited by newlines: */
  strcpy(sbuf, strings);
  if (*sp++ = strtok(sbuf, "\n")) 
    while (*sp++ = strtok(NULL, "\n"))
      ;

  /* For each string... */
  for (sp = sv; *sp; sp++) {
    
    /* Adjust x coord, based on text alignment: */
    switch (align) {
    case IA_ALIGN_LEFT:      
      break;
    case IA_ALIGN_CNTR:
      xi -= ((strlen(*sp) * font->w) / 2);
      break;
    default:
      IA_Fatal("Bad text alignment.");
    }

    /* Draw: */
    gdImageString(im, font, xi, yi, *sp, color);

    /* Next! */
    xi = x;
    yi += font->h + 1;
  }
}


/************************************************************
 * IA_DrawLegend -- draw the title
 ************************************************************/

void IA_DrawLegend(gdImagePtr im)
{
  char adate[32];
  int x, y = 10;

  /* Start time: */
  sprintf(adate, "%04d-%02d-%02d\n%02d:%02d:%02d", 
	  gGraphStart.yr, gGraphStart.mon, gGraphStart.day,
	  gGraphStart.hr, gGraphStart.min, gGraphStart.sec);
  IA_gdImageStrings(im, gdFontTiny, IA_LMARGIN, y, 
		    adate, gColorBlack, IA_ALIGN_CNTR);

  /* Stop time: */
  sprintf(adate, "%04d-%02d-%02d\n%02d:%02d:%02d", 
	  gGraphStop.yr, gGraphStop.mon, gGraphStop.day,
	  gGraphStop.hr, gGraphStop.min, gGraphStop.sec);
  IA_gdImageStrings(im, gdFontTiny, IA_LMARGIN + gGraphWidth, y, 
		    adate, gColorBlack, IA_ALIGN_CNTR);
}

/************************************************************
 * IA_DrawImage -- create the image
 ************************************************************
 *
 * WARNINGS
 *    The order in which things are drawn is VERY IMPORTANT.
 *    It ensures that less-important things don't cover more-important things.
 */

int IA_DrawImage(gdImagePtr im)
{
  IA_Time i_time;
  int x, y;
  int xMinor, xMaj, xLastMaj;
  int xLabel, yLabel;
  char *label;
  char labelbuf[IA_MAXLABEL+1];
  int i;
  
  /* Set basic characteristics: */
  gdImageInterlace(im, 1);
  
  /* Allocate the colors: */
  gColorBlack = gdImageColorAllocate(im,   0,   0,   0);
  gColorWhite = gdImageColorAllocate(im, 255, 255, 255);
  gColorGrey  = gdImageColorAllocate(im, 220, 220, 220);
  gColorTrans = gdImageColorAllocate(im, 222, 222, 222);
  gColorData  = gdImageColorAllocate(im, 255,   0,   0);
  gdImageColorTransparent(im, gColorTrans); 
  
  /* Wipe everything to be transparent: */
  gdImageFilledRectangle(im, 0, 0, gGifWidth, gGifHeight, gColorTrans);

  /* Give the graph a white background: */
  gdImageFilledRectangle(im, 
			 IA_LMARGIN, IA_TMARGIN, 
			 IA_LMARGIN+gGraphWidth-1, IA_TMARGIN+gGraphHeight-1, 
			 gColorWhite);  

  /* The major and minor x-axis ticks, and the grid lines: */
  IA_Debug("DRAW X-AXIS TICKS/LABELS...");
  xLastMaj = -1;
  y = IA_TMARGIN + gGraphHeight;
  IA_TimeCopy(&i_time, &gGraphStart);
  while (IA_TimeCmp(&i_time, &gGraphStop) <= 0) {

    /* Compute the x offset of the next major tick: */
    xMaj = IA_LMARGIN + IA_AbstimeToGXCoord(&(i_time.abs));

    /* If there was a previous major tick, draw the minor ticks from it: */
    if ((xLastMaj > 0) && (gMajorDivs > 1)) {      
      int i;
      IA_Debug("Dividing into %d divisions", gMajorDivs);
      for (i = 1; i < gMajorDivs; i++) {

	/* Compute the x offset, and draw the minor tick: */
	xMinor = xLastMaj + (int)((i / (double)gMajorDivs) * (xMaj-xLastMaj));
	gdImageLine(im, 
		    xMinor, y, xMinor, (y + IA_MINORTICKLEN), 
		    gColorBlack);
      }
    }

    /* Draw the next major line and tick: */
    xMaj = IA_LMARGIN + IA_AbstimeToGXCoord(&(i_time.abs));
    IA_Debug("Major Tick: <%s> (x = %d, y = %d)", IA_TimeFmt(&i_time), 
	     xMaj, y);
    gdImageLine(im,                    /* the gridline... */
		xMaj, IA_TMARGIN, 
		xMaj, (IA_TMARGIN + gGraphHeight),
		gColorGrey);
    gdImageLine(im,                    /* the tick... */
		xMaj, y, 
		xMaj, (y + IA_MAJORTICKLEN), 
		gColorBlack);
  
    /* Draw the tick label: */
    label = IA_TimeLabel(&i_time, gMajorTick, gMajorUnits);
    xLabel = xMaj;
    yLabel = y + IA_MAJORTICKLEN + 1;
    IA_gdImageStrings(im, gdFontTiny, xLabel, yLabel, label, 
		      gColorBlack, IA_ALIGN_CNTR);
    
    /* Compute the next offset! */
    IA_TimeAddUnit(&i_time, gMajorTick, gMajorUnits);    
    xLastMaj = xMaj;
  }
  
  /* The y-axis labels: */
  IA_Debug("DRAW Y-AXIS LABELS...");
  for (i = 0; i < gInput.nranges; i++) {
    IA_Range *range = gInput.ranges[i];
    int yCenter;

    /* Get a safe tick label: */
    if (strlen(range->label) <= IA_MAXLABEL) {
      strcpy(labelbuf, range->label);
    }
    else {
      strncpy(labelbuf, range->label, IA_MAXLABEL);
      strcpy(labelbuf + IA_MAXLABEL - 3, "...");
    }

    /* Get the label position: */
    xLabel = IA_LMARGIN - IA_LABELMARGIN - (strlen(labelbuf) * gdFontTiny->w);
    yCenter = IA_TMARGIN + IA_BARSEP +     /* top margin and top bar sep */
      (i * (IA_BARSEP + IA_BARHGT)) +      /* bars above us */
	(IA_BARHGT / 2);                   /* to center of this bar */
    yLabel = yCenter - (gdFontTiny->h / 2);   /* to top of label */

    /* Draw the label: */
    gdImageString(im, gdFontTiny, xLabel, yLabel, labelbuf, gColorBlack);

    /* Draw a line: */
    gdImageLine(im, 
		IA_LMARGIN, yCenter, IA_LMARGIN+gGraphWidth, yCenter, 
		gColorGrey);
    
  }

  /* Outline the graph in black: */
  gdImageRectangle(im, 
			 IA_LMARGIN, IA_TMARGIN, 
			 IA_LMARGIN+gGraphWidth-1, IA_TMARGIN+gGraphHeight, 
			 gColorBlack);  

  /* The ranges: */
  IA_Debug("DRAW RANGES...");
  for (i = 0; i < gInput.nranges; i++) {
    int yTop, yBot, gxLeft, gxRight, xLeft, xRight;
    IA_Range *range = gInput.ranges[i];

    /* Compute the borders: */
    yTop = IA_TMARGIN + IA_BARSEP +     /* top margin and top bar sep */
      (i * (IA_BARSEP + IA_BARHGT));    /* bars above us */
    yBot = yTop + IA_BARHGT;
    gxLeft  = IA_AbstimeToGXCoord(&(range->start.abs));
    gxRight = IA_AbstimeToGXCoord(&(range->stop.abs));
        
    /* Draw indicators of out-of-range values: */
    if ((gxLeft < 0) || (gxRight < 0)) {

      /* Left arrow: */
      y = yTop + (IA_BARHGT / 2);
      xLeft  = IA_LMARGIN - IA_ARROWLEN;
      xRight = IA_LMARGIN + IA_ARROWLEN - 2;
      gdImageLine(im, xLeft, y, xRight, y, gColorData);      
      gdImageLine(im, xLeft+2, y-2, xLeft, y, gColorData); 
      gdImageLine(im, xLeft+2, y+2, xLeft, y, gColorData); 
    }
    if ((gxLeft > gGraphWidth) || (gxRight > gGraphWidth)) {  

      /* Right arrow: */
      xLeft  = IA_LMARGIN + gGraphWidth - IA_ARROWLEN + 2;
      xRight = IA_LMARGIN + gGraphWidth + IA_ARROWLEN;
      y = yTop + (IA_BARHGT / 2);
      gdImageLine(im, xLeft, y, xRight, y, gColorData);      
      gdImageLine(im, xRight-2, y-2, xRight, y, gColorData); 
      gdImageLine(im, xRight-2, y+2, xRight, y, gColorData); 
    }

    /* If any part in range, draw: */
    if (!(((gxLeft < 0) && (gxRight < 0)) ||
	  ((gxLeft > gGraphWidth) && (gxRight > gGraphWidth)))) {

      /* Clip: */
      xLeft  = IA_LMARGIN + ((gxLeft < 0) ? 0 : gxLeft);
      xRight = IA_LMARGIN + ((gxRight > gGraphWidth) ? gGraphWidth : gxRight);
      
      /* Draw: */
      gdImageFilledRectangle(im, xLeft, yTop, xRight, yBot, gColorData);
    }
  }


  /* Legend: */
  IA_DrawLegend(im);

  return 0;
}

/************************************************************
 * IA_CreateGif -- create the GIF file, or die
 ************************************************************/

int IA_CreateGif(char *filename)
{
  gdImagePtr im;
  FILE *fp;
  char dirbuf[256];

  /* Open the file: */
  if (!(fp = fopen(filename, "wb"))) {
    IA_Fatal("couldn't create <%s>", filename);
  }

  /* Create the image object: */
  if (!(im = gdImageCreate(gGifWidth, gGifHeight))) {
    fclose(fp);
    IA_Fatal("couldn't create image (width %d, height %d)",
	     gGifWidth, gGifHeight);
  }

  /* Draw the image: */
  if (IA_DrawImage(im) < 0) {
    fclose(fp);
    gdImageDestroy(im);
    IA_Fatal("couldn't draw the image");
  }

  /* Output it to file: */
  gdImageGif(im, fp);
  fclose(fp);
  gdImageDestroy(im);  
 
  /* Report: */
  getcwd(dirbuf, 256);
  printf("Filename: %s/%s\n", dirbuf, filename);
  return 0;  
}


/************************************************************
 * IA_MakeOutput
 ************************************************************
 *
 * DESCRIPTION
 *    Generate the GIF(s).
 */ 

int IA_MakeOutput()
{
  int i;
  int labelLen;
  int maxLabelLen = 1;


  /* 
   * Compute the boundaries and units to be used in the graph... 
   */

  /* Get the length of the longest label, in pixels (must do this first!): */
  for (i = 0; i < gInput.nranges; i++) {
    IA_Range *range = gInput.ranges[i];

    labelLen = strlen(range->label);
    maxLabelLen = ((labelLen > maxLabelLen) ? labelLen : maxLabelLen);    
  }
  gMaxLabelLength = (maxLabelLen * gdFontTiny->w);

  /* Calculate size of gap, height and width of graph: */
  gGifWidth   = gInput.totalWidth;
  gGraphWidth = gGifWidth - IA_LMARGIN - IA_RMARGIN;
  gGraphHeight = (gInput.nranges * (IA_BARHGT + IA_BARSEP)) + IA_BARSEP;
  gGifHeight   = gGraphHeight + IA_TMARGIN + IA_BMARGIN;

  /* Get extents of *desired selection* (not of entire graph): */
  memcpy(&gSlctStart, &(gInput.start), sizeof(IA_Time));
  memcpy(&gSlctStop,  &(gInput.stop), sizeof(IA_Time));
  IA_AbstimeDiff(&(gSlctStop.abs),  &(gSlctStart.abs),  &gSlctLen);
  IA_Debug("Total start-to-stop coverage, %ld days, %d secs",
	   gSlctLen.jd, gSlctLen.secs);


  /*
   * Compute units for tick marks...
   */

  /* Based on the time period, what major/minor ticks will we use? */
  if (gSlctLen.jd > (100 * IA_DAYS_IN_A_YEAR)) {      /* > 100 y */
    gMajorUnits = IA_UNIT_YEARS;
    gMajorTick  = 10;     /* 10 yrs per major tick */
    gMajorDivs  = 2;      /* 5 yrs per minor tick */
  }
  else if (gSlctLen.jd > (50 * IA_DAYS_IN_A_YEAR)) {  /* 50-100 y */
    gMajorUnits = IA_UNIT_YEARS;
    gMajorTick  = 5;     /* 5 yrs per major tick */
    gMajorDivs  = 1;     /* no minor ticks */
  }
  else if (gSlctLen.jd > (10 * IA_DAYS_IN_A_YEAR)) {  /* 10-50 y */
    gMajorUnits = IA_UNIT_YEARS;
    gMajorTick  = 5;     /* 5 yrs per major tick */
    gMajorDivs  = 5;     /* 1 yr per minor tick */
  }
  else if (gSlctLen.jd > (5 * IA_DAYS_IN_A_YEAR)) {   /* 5-10 y */
    gMajorUnits = IA_UNIT_YEARS;
    gMajorTick  = 1;     /* 1 yr per major tick */
    gMajorDivs  = 2;     /* 6 mon per minor tick */
  }
  else if (gSlctLen.jd > (1 * IA_DAYS_IN_A_YEAR)) {   /* 1-5y */
    gMajorUnits = IA_UNIT_YEARS;
    gMajorTick  = 1;     /* 1 yr per major tick */
    gMajorDivs  = 6;     /* 2 mon per minor tick */
  }
  else if (gSlctLen.jd > (6 * IA_DAYS_IN_A_MONTH)) {  /* 6 mos - 1y */
    gMajorUnits = IA_UNIT_MONTHS;
    gMajorTick  = 1;     /* 1 mon per major tick */
    gMajorDivs  = 1;     /* no minor ticks */
  }
  else if (gSlctLen.jd > (1 * IA_DAYS_IN_A_MONTH)) {  /* 1 - 6 mos */
    gMajorUnits = IA_UNIT_MONTHS;
    gMajorTick  = 1;     /* 1 mon per major tick */
    gMajorDivs  = 4;     /* ~1 wk per minor tick */
  }
  else if (gSlctLen.jd > 5) {                         /* 5 - 30 days */
    gMajorUnits = IA_UNIT_DAYS;
    gMajorTick  = 5;     /* 5 days per major tick */
    gMajorDivs  = 5;     /* 1 day per minor tick */
  }
  else if (gSlctLen.jd > 1) {                         /* 1 - 5 days */
    gMajorUnits = IA_UNIT_DAYS;
    gMajorTick  = 1;     /* 1 day per major tick */
    gMajorDivs  = 4;     /* 6 hrs per minor tick */
  }
  else {                                              /* < 1 day */
    gMajorUnits = IA_UNIT_HOURS;
    gMajorTick  = 6;     /* 6 hrs per major tick */
    gMajorDivs  = 4;     /* 1 hr per minor tick */
  }


  /*
   * Compute various graph measurements and conversion values:
   */

  /* gGraphStart -- figure out nearest major ticks below the data: */
  IA_TimeBelow(&gSlctStart, gMajorTick, gMajorUnits, &gGraphStart);

  /* gGraphStop -- keep adding major ticks to get 1st one above the data: */
  IA_TimeCopy(&gGraphStop, &gGraphStart);
  while (IA_TimeCmp(&gGraphStop, &gSlctStop) < 0) {
    IA_TimeAddUnit(&gGraphStop, gMajorTick, gMajorUnits);  
  }
  /* ...gGraphStop is now the first major tick at-or-above the data */
  IA_Debug("Time above = <%s>", IA_TimeFmt(&gGraphStop));

  /* gGraphLen: */
  IA_AbstimeDiff(&(gGraphStop.abs), &(gGraphStart.abs), &gGraphLen);

  /* gGraphWidth -- get width of graph in pixels: */
  gGraphWidth = gInput.totalWidth - IA_LMARGIN - IA_RMARGIN;
  if (gGraphWidth < 100) {
    IA_Fatal("Graph is too small: it won't look good");
  }

  /* gPixelsPerDay, gPixelsPerSec: */
  IA_Debug("graph len = %ld days, %d secs", gGraphLen.jd, gGraphLen.secs);
  if (gGraphLen.jd > 0) {
    gPixelsPerDay = 
      ((double)gGraphWidth / 
       (gGraphLen.jd + ((double)gGraphLen.secs / IA_SECS_PER_DAY)));
    gPixelsPerSec = gPixelsPerDay / IA_SECS_PER_DAY;
    IA_Debug("PPD = %lf, PPS = %lf", gPixelsPerDay, gPixelsPerSec);
  }
  else if (gGraphLen.secs > 0) {
    gPixelsPerSec = 
      gGraphWidth /
	((double)(IA_SECS_PER_DAY * gGraphLen.jd) + gGraphLen.secs);
    gPixelsPerDay = gPixelsPerSec * IA_SECS_PER_DAY;
    IA_Debug("PPD = %lf, PPS = %lf", gPixelsPerDay, gPixelsPerSec);
  }
  else {
    /* IA_Fatal("Graph has zero time coverage: please alert developers.");
     *  (Graph had zero time coverage.  Force it to be 1 second wide.)
     */
    gPixelsPerSec = gGraphWidth;       /* Force 1 second width. */
    gPixelsPerDay = gPixelsPerSec * IA_SECS_PER_DAY;
    IA_Debug("PPD = %lf, PPS = %lf", gPixelsPerDay, gPixelsPerSec);
  }

  /* Create the GIF: */
  if (IA_CreateGif(IA_FILENAME) < 0) {
    return -1;
  }

  return 0;
}

/************************************************************
 * main
 ************************************************************/
  
int main(argc, argv) 
     int argc;
     char *argv[];
{
  if (IA_ParseArgs(argc, argv) < 0) {
    IA_Fatal("couldn't parse arguments");
  }
  if (IA_ParseInput(&gInput) < 0) {
    IA_Fatal("couldn't parse input");
  }
  if (IA_MakeOutput() < 0) {
    IA_Fatal("couldn't generate output");
  }
  printf("\n");
  exit(0);
}
/******************** end of file "imscovertime.c" ********************/
