/*************************************************************
 *
 * NAME
 *    imsdepval - get the dependent valids based on input ODL requests
 * DESCRIPTION
 *    Based on the input ODL imsdepval sets the selections and returns
 *    valids information based on the type of request
 *
 * USAGE
 *       imsdepval [options] [action]
 *    The options are:
 *       -l	 	Use chunking output instead of ODL output.
 *       -i <filename> 	The input request ODL file.
 *                      If this option is missing it will take input from stdin.
 *       -o <filename> 	The output file.
 *                      If this option is missing it will output to stdout.
 *    The actions are:
 *	 -mf	Return the filters that are compatible with the REQUEST_FIELD.
 *	 -mv	Return the valids that are compatible with the REQUEST_FIELD
 *              and the optional filter or optional pattern supplied.
 *	 -ms	Return the list of selected valids based on the REQUEST_FIELD.
 *		Incompatibles are prefixed by an asterisk "*".
 *	 -mc	Return the list of compatible fields based on the input ODL.
 *	 -mi	Return the list of fields that have become incompatible as
 *		a result of selections made in the input ODL file.
 *	 -mo	Return the SPECIALIZED_CRITERIA ODL for the REQUEST_FIELD.
 *	 -ma	Return the valids that are compatible with the REQUEST_FIELD
 *          	that are listed with the keyword "ATTRIBUTES", where
 *          	the strings that follow the keyword can be any of the 7
 *          	core attributes.
 *	 -mb	Return the all of the valids for DATA CENTER, DATA SET,
 *         	SOURCE, SENSOR and PARAMETER
 *	 -md    Return the matching valids that are listed with the
 *		keyword "ATTRIBUTES".
 *	 -mg    Return the fields associated with the keyword "ATTRIBUTES".
 *       -mq    Return a query of selected valids.
 *		Incompatibles are prefixed by an asterisk "*".
 *       -mr    Return a list of CRITERIA_NAMEs in the EXTENDED.valids file
 *              that have the flag RESULTS_SELECTABLE = "Y".
 *       -me    Return a list of CRITERIA_NAMEs in the EXTENDED.valids file
 *              that have the flag DEFAULT = "Y".
 *	 -mz	Return the SPECIALIZED_CRITERIA ODL for the REQUEST_FIELD
 *              where the REQUEST_FIELD is actually a CRITERIA_NAME.
 *
 *        	The default action if no action is supplied is to return
 *        	the dependent valids that are compatible with the REQUEST_FIELD
 *        	based on the selections in the rest of the input file.
 *    Any other options will return a usage page.
 */
#include <stdio.h>
#include <assert.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>              /* needed for toupper() */
#include <sys/types.h>          /* needed for stat */
#include <sys/stat.h>           /* needed for stat */

#include "Gaea_config.h"
#include "IK_Ims.h"
#include "IK_DataDic.h"
#include "IK_Errno.h"
#include "IK_Syslog.h"
#include "IK_Granule.h"
#include "IK_Msg.h"
#include "IK_ValNames.h"
#include "IK_DepValid.h"
#include "IC_Commun.h"

#include "gdbm.h"
#include "IK_Nm.h"
#include "IK_Nmpro.h"
#include "odldef.h"
#include "odlinter.h"
/* The following include was added for the new valids format */
#include "IK_NewValids.h"

#include "IW_ODL.h"
#include "IW_Str.h"
#include "IW_New.h"
#include "IW_Chunk.h"

#define CHAR_BIT 8 /* the number of bits in a byte */
#define NOTCOMP "*" /* the leading asterisk indicates not compatible */
#define MISSING (-1) /* if the field is missing set to this */
#define MAXLEN (90) /* was 81 the maximum number of bytes in a valid + 1 */
#define MAXFIELD (10) /* the number of dependent valid fields */
#define DAAC     "Data Center"
#define DATASET  "Data Set"
#define SOURCE   "Source"
#define SENSOR   "Sensor"
#define PARAM    "Parameter"
#define CAMPAIGN "Campaign"
#define PLEVEL   "Processing Level"

#define MAX_CORE_FIELD (7) /* the maximum number of core attributes */
#define MAX_SIMPLE_FIELD (6) /* the maximum number of simple attributes */

extern char **fieldname; /* names of the fields */
char **newfieldname; /* output names of the fields */
#define SYSLOG_FILE "depsyslog"  /* the name of the log file */
#define DYNAMIC_FILE "dynamic.odl" /* name of the dynamic valids file */
#define EXTENDED_ODL_FILE "EXTENDED.valids"  /* the name of the ODL file for
                                                extended valids */

extern GDBM_FILE dep_valids_dbf, tag_dbf; /* used for old dependent valids */
extern char *optarg; /* used to parse the command line arguments */

static AGGREGATE IK_BuildSearchTree(AGGREGATE, char *);
static int dump_tree(AGGREGATE root);
static int handle_all(AGGREGATE tempGroup);
static int handle_attributes(AGGREGATE tempGroup);
static int ReturnFieldNumber(char *request_id);
static char* ReturnFieldCategory(char *request_id);
static void usage();

/* The following added for filters */
static int bytesPerRow; /* number of bytes in a row of the valids bitmap */
static int ValidsCount; /* number of valids in the list */
static int FiltersCount; /* number of filters in the list */
static int *filter_indexes = NULL; /* array of indexes for the list of filters*/
static int *indexes = NULL; /* array of indexes for the list of valids */
static int *filter_saved = NULL; /* array of indexes for filters that remain */
static int j; /* counter for valids found in the input odl */
static int k; /* counter for fields */
static IK_BITVECTOR *FieldValidsVector = NULL; /* vector of selected valids */
static IK_BITVECTOR *FiltersVector = NULL; /* vector of selected filters */
extern IK_BITVECTOR *IK_fieldMask; /* indicates which field a valid belongs */
static char **table = NULL;   /* return list of valids. */
static char **tableTemp = NULL;   /* return list of valids. */
static char **tableSecond = NULL;   /* return list of valids. */
static char request_id[MAXLEN]; /* string that represents REQUESTED_FIELD */ 
static char *request_cat; /* represent REQUESTED_FIELD as stored in fieldname */
                          /* used for extended valids mixed case values.      */
static char result_id[MAXLEN]; /* string that represents RESULT */
static char category_id[MAXLEN]; /* Extended search category name */
static char *filterflag = NULL; /* indicates which mode to operate */
static FILE *outfile = NULL; /* output file */
static char **fields; /* requested valids of a field */
static int line_out=0; /* if set to 1 use chunking instead of ODL */
static int tableflag=0; /* if set to 1 table output exists */
static int i; /* counter */
static long fieldCnt; /* total number of fields */
static int ifield; /* number of the dependent valids field */
static int rtn_code; /* return code from IK_FieldFilter */
void AllocateBitvectors();
void FreeAllocated();
void OutputValids();
AGGREGATE root; /* root for input ODL file */
IK_Errno IK_ImsErrno; /* holds IK error status */
int handle_value (char ***, AGGREGATE, int *, char *);
void build_out_odl (char **, char *, FILE *);
char *ptr; /* a tmp pointer used to upcase the field name */
extern void VectorAnd(const IK_BITVECTOR *vector, const IK_BITVECTOR *vector2,
 IK_BITVECTOR *vectornew, int len);
extern void VectorZero(IK_BITVECTOR *vector, int len);
/* The output channel for the chunks: */
IW_Chunksp *COUT; /* pointer for chunk stream structure */
static char *attributes = "ATTRIBUTES";
static char *pattern = "PATTERN";

int main(int argc, char ** argv)
{
  int filterFieldCnt; /* number of filters for the current field */
  int patternFlag=0; /* when set to 1 indicates a pattern or filter */
  char *infilename = NULL; /* this is the input ODL file */
  FILE *odlfp; /* file pointer for extended valids ODL file */
  PARAMETER tempPara; /* used to find a PARAMETER within a group */
  VALUE tempValue; /* used to store a value of the tempPara */
  AGGREGATE tempGroup; /* used to store a particular group */
  PARAMETER extPara; /* used to store a PARAMETER of extended ODL */
  VALUE extValue; /* used to store a VALUE of extPara */
  AGGREGATE extGroup; /* used to store a group  of extended ODL */
  AGGREGATE odl_root; /* root AGGREGATE of the extended ODL file */
  AGGREGATE dynamic_root; /* root AGGREGATE of the dynamic valids file */
  int count; /* count of the number of strings returned in tableTemp */
  int c; /* character in the input arguments */
  int numHits; /* number of valids selected by IK_FindValids */
  int found; /* return code for IK_FindValids */
  int dyIndex; /* the index of the new dynamic valid */
  char gd_dir[255];  /* Gaea data directory */
  char gaea_tmp_dir[255];  /* Gaea temp directory */
  char syslog_file[255];  /* Syslog file to open */
  char odl_file[255];  /* extended valids ODL file */
  char requestField[MAXLEN]; /* input requestField for handle_value */
  char dynamic_file[255];  /* name of dynamic valids file to open */
  struct stat statbuf; /* needed for stat */

  while ((c=getopt(argc, argv, "li:o:m:")) != EOF){
    switch (c) {
    case 'l':
      line_out = 1;
      break;
    case 'i':
      infilename = optarg;
      break;
    case 'o':
      outfile = fopen(optarg, "w");
      break;
    case 'm':
      filterflag = optarg;
      break;
    case '?':
      fprintf (stderr, "Unrecognized option!\n");
      usage();
      exit(0);
    }
  }

  if(getenv("GAEADATA_DIR")!=NULL) {
    strcpy (gd_dir, getenv("GAEADATA_DIR"));
  } else { 
    (void) fprintf(stderr, "Unable to find the Gaea Data Directory.\n");
    exit(0);
  }
  if(getenv("GAEATMP_DIR")!=NULL) {
    strcpy (gaea_tmp_dir, getenv("GAEATMP_DIR"));
    sprintf (syslog_file, "%s/%s", gaea_tmp_dir, SYSLOG_FILE);
    sprintf (dynamic_file, "%s/%s", gaea_tmp_dir, DYNAMIC_FILE);
  } else { 
    (void) fprintf(stderr, "Unable to find the Gaea Temp Directory.\n");
    exit(0);
  }
#ifdef DEBUGME
  printf ("gaea_dir %s \n", gd_dir);
#endif

  root = NewAggregate(NULL,KA_GROUP,"ROOT",NULL); 
  dynamic_root = NewAggregate(NULL,KA_GROUP,"ROOT",NULL); 
  IK_NameSyslog(syslog_file); 
/*  IK_DepValidsInit (); old valids format */
  /* The following call was added for the new dependent valids code -Solomon */
  if (IK_ReadDepValids() == -1) {
    IK_ImsErrno = IK_EIO;
    (void) fprintf(stderr, "Unable to read Valids files. See support staff.");
    exit(0);
  } /* end of new dependent valids code */
  ValidsCount = IK_ValidsCnt();
  FiltersCount = IK_FilterCnt();
  fieldCnt = IK_FieldCnt(); 
  newfieldname = (char **)calloc(fieldCnt, sizeof(char *));
  fields = (char **)calloc(ValidsCount, sizeof (char *));

  /* Initialize the static field names. */
  newfieldname[VALIDS_DAAC]    = calloc(strlen(DAAC)+1, sizeof(char));
  (void) strcpy(newfieldname[VALIDS_DAAC],DAAC);
  newfieldname[VALIDS_DATASET] = calloc(strlen(DATASET)+1, sizeof(char));
  (void) strcpy(newfieldname[VALIDS_DATASET],DATASET);
  newfieldname[VALIDS_SOURCE]  = calloc(strlen(SOURCE)+1, sizeof(char));
  (void) strcpy(newfieldname[VALIDS_SOURCE],SOURCE);
  newfieldname[VALIDS_SENSOR]  = calloc(strlen(SENSOR)+1, sizeof(char));
  (void) strcpy(newfieldname[VALIDS_SENSOR],SENSOR);
  newfieldname[VALIDS_PARAMETERS] = calloc(strlen(PARAM)+1, sizeof(char));
  (void) strcpy(newfieldname[VALIDS_PARAMETERS],PARAM);
  newfieldname[VALIDS_CAMPAIGN] = calloc(strlen(CAMPAIGN)+1, sizeof(char));
  (void) strcpy(newfieldname[VALIDS_CAMPAIGN],CAMPAIGN);
  newfieldname[VALIDS_PLEVEL]   = calloc(strlen(PLEVEL)+1, sizeof(char));
  (void) strcpy(newfieldname[VALIDS_PLEVEL],PLEVEL);
  for(i = VALIDS_PLEVEL+1; i<fieldCnt; i++) {
    newfieldname[i] = calloc(strlen(fieldname[i])+1, sizeof(char));
    (void) strcpy(newfieldname[i],fieldname[i]);
  }
  AllocateBitvectors();
  /* read the search tree from the odl file */
  root = IK_BuildSearchTree(root, infilename);
  if (stat(dynamic_file,&statbuf) != -1) {
    dynamic_root = IK_BuildSearchTree(dynamic_root, dynamic_file);
    VectorZero(FieldValidsVector, ValidsCount);
    for (k=0; k < MAX_CORE_FIELD; k++) {
      j = 0;
      if (handle_value (&fields, dynamic_root->first_child, &j, fieldname[k])){
        /* fprintf (stderr, "item in fields %s \n", fields[0]); */
        for (i=0; i < j; i++) {
          found = IK_FindValids(fields[i],k,FieldValidsVector,&numHits);
          if(found < 0) {
            if(IK_dynamicSlotAvailable()) {
              dyIndex = IK_addDynamicValid(fields[i],k);
              IK_SelectValid(dyIndex);
            }
          }
          free (fields[i]);
          fields[i] = NULL;
        }
      }
    }
  }

  if(filterflag !=NULL && filterflag[0] == 'f') {
  /* This code gets the list of filters given an attribute name in REQUEST_FIELD
   */
    /* (void) fprintf(stdout, "Get Filters.\n"); */
    /* following code tries to extract the parameter values from */
    /* the odl tree */
    tempGroup = FindGroup(root, "FILTERS");
    tempPara = FindParameter (tempGroup, "REQUEST_FIELD");
    tempValue = FirstValue (tempPara);
    /* fprintf (stdout, "req-field %s \n", tempValue->item.value.string); */
    if(tempValue && tempValue->item.value.string) 
     strcpy (request_id, tempValue->item.value.string);
    ifield = ReturnFieldNumber(request_id);
    /* (void) fprintf(stderr, "ifield = %d\n",ifield); */
    if(ifield == MISSING) {
      (void) fprintf(stderr, "Field Invalid!\n");
      exit(0);
    }
    rtn_code = handle_all(tempGroup);
    /* Get the Filters Vector given the field number */
    rtn_code = IK_FieldFilter(ifield, FiltersVector);
    /* Get the list of Filter Strings given the Filters Vector */
    tableTemp = IK_FilterString(FiltersVector, filter_indexes);
    if(tableTemp != NULL) {
      count = 0;
      while(tableTemp && *tableTemp && tableTemp[count]) {
        count++;
      } /* end while tableTemp */
      j = 0;
      table = calloc(count+1, sizeof(char *));
      for (i=0; i < count; ++i) {
        /* Get the Valids Vector for the filter desired */
	VectorZero(FieldValidsVector, ValidsCount);
	rtn_code = 
	  IK_FilterVectorComp(filter_indexes[i], ifield, FieldValidsVector);

	if(IK_ValidIndexesCnt(FieldValidsVector) > 0) {
          table[j] =
            (char *) calloc(strlen(tableTemp[i])+1,sizeof(char));
	  (void) strcpy(table[j],tableTemp[i]);
	  j++;
	}
      }
    }
    /* (void) fprintf(stdout, "Filters are Okay!\n"); */
    OutputValids();
  } else if(filterflag !=NULL && filterflag[0] == 'v') {
  /* This section returns the compatible valids for a given filter name */
    /* (void) fprintf(stderr, "Get Valids.\n"); */
    /* following code tries to extract the parameter values from */
    /* the odl tree */
    tempGroup = FindGroup(root, "VALIDS");
    tempPara = FindParameter (tempGroup, "REQUEST_FIELD");
    tempValue = FirstValue (tempPara);
    /*fprintf (stderr, "req-field %s \n", tempValue->item.value.string);*/
    if(tempValue && tempValue->item.value.string) 
     strcpy (request_id, tempValue->item.value.string);

    /* ---------------------------------------- */ 
    /* Get string representation as it is       */
    /* stored in fieldname[index].              */ 
    /*                                          */
    /* Lorena Marsans - 7/17/98                 */
    /* ---------------------------------------- */ 
    request_cat = ReturnFieldCategory(request_id);

    ifield = ReturnFieldNumber(request_id);
    if(ifield == MISSING) {
      (void) fprintf(stderr, "Field Invalid!\n");
      exit(0);
    }
    rtn_code = handle_all(tempGroup);
    rtn_code = IK_NotCompSelect(ifield,FieldValidsVector);
    tableTemp = IK_ValidString(FieldValidsVector, indexes);
    if(tableTemp != NULL) {
      j = 0;
      while(tableTemp && *tableTemp && tableTemp[j]) {
        j++;
      } /* end while tableTemp */
      table = calloc(j+1, sizeof(char *));
      j = 0;
      for (i=0; tableTemp[i]!=(char*)NULL; ++i) {
        table[j] =
         (char *) calloc(strlen(tableTemp[j])+2,sizeof(char));
        /* start the string with the NOTCOMP (not-compatible) character */
        (void) strcpy(table[j],NOTCOMP);
        (void) strcat(table[j],tableTemp[j]);
        j++;
      }
    }
    OutputValids();
    j = 0;
    (void) strcpy(requestField, "FILTER");
    if (handle_value (&fields, tempGroup, &j, requestField)) {
      patternFlag = 1;
      /* Get the Filters Vector given the field number */
      rtn_code = IK_FieldFilter(ifield, FiltersVector);
      /* Get the number of filters for the current field */
      filterFieldCnt = IK_FilterFieldCnt(ifield);
      /* Get the Filter Strings for the given Filters Vector */
      table = IK_FilterString(FiltersVector, filter_indexes);
      count = 0;
      for(k=0; k < filterFieldCnt; k++) {
        for (ptr = table[k];*ptr;ptr++)       /* upcase the filter name */
          *ptr = toupper (*ptr);
        for(i=0; i < j; i++) {
	  /* Compare the list of filter names with the requested filter name */
	  if(strcmp(table[k], fields[i]) == 0) {
	    filter_saved[count] = filter_indexes[k];
	    count++;
	  }
	}
      }

      /* --------------------------------------------------- */
      /* Modified to use request_cat instead of request_id.  */
      /* Changed was mainly done for the extended search     */
      /* valids, because extended categories are in mix      */
      /* case in the valids but are in caps in the odl tree. */
      /*                                                     */
      /* Lorena Marsans - 7/17/98                            */
      /* --------------------------------------------------- */
      IK_GetDepValids (&table, request_cat, NULL);
      if(table) {
	IK_FreeStringList(table);
	table = NULL;
      }
      if(count > 0) {
	for (i=0; i < count; i++) {
	  /* Get the Valids Vector for the filter desired */
          VectorZero(FieldValidsVector, ValidsCount);
	  rtn_code = IK_FilterVectorComp(filter_saved[i], ifield, FieldValidsVector);
	  /* Get the list of valids for selected filter. */
	  table = IK_ValidString(FieldValidsVector, indexes);
	  OutputValids();
	}
      }

      for (i=0; i < j; i++){
        free (fields[i]);
	fields[i] = NULL;
      }
    }
    j = 0;
    (void) strcpy(requestField, "PATTERN");
    if (handle_value (&fields, tempGroup, &j, requestField)) {
      patternFlag = 1;
      for (i=0; i < j; i++){
        /* Get the Vector that matches the requested string */
        VectorZero(FieldValidsVector, ValidsCount);
        rtn_code = IK_FindValids(fields[i],ifield,FieldValidsVector,&numHits);
        /* Get the current compatibles for the field ifield */
        rtn_code = IK_Compatible(ifield,FiltersVector);
        /* And the FiltersVector with the FieldValidsVector to get the
         * compatible valids with the filter desired
         */
        VectorAnd(FieldValidsVector,FiltersVector,FieldValidsVector,ValidsCount);
        /* Get the list of valids that are compatible with the text entries.*/
        table = IK_ValidString(FieldValidsVector, indexes);
	OutputValids();
        free (fields[i]);
        fields[i] = NULL;
      }
      j = 0;
    }
    if(!patternFlag) {
      /* --------------------------------------------------- */
      /* Modified to use request_cat instead of request_id.  */
      /* Changed was mainly done for the extended search     */
      /* valids, because extended categories are in mix      */
      /* case in the valids but are in caps in the odl tree. */
      /*                                                     */
      /* Lorena Marsans - 7/17/98                            */
      /* --------------------------------------------------- */
      IK_GetDepValids (&table, request_cat, NULL);
      OutputValids();
    }



  } else if(filterflag !=NULL && filterflag[0] == 's') {
  /* This section returns the selected valids for a given field name */
    /* following code tries to extract the parameter values from */
    /* the odl tree */
    tempGroup = FindGroup(root, "VALIDS");
    tempPara = FindParameter (tempGroup, "REQUEST_FIELD");
    tempValue = FirstValue (tempPara);
    if(tempValue && tempValue->item.value.string) 
     strcpy (request_id, tempValue->item.value.string);
    ifield = ReturnFieldNumber(request_id);
    if(ifield == MISSING) {
      (void) fprintf(stderr, "Field Invalid!\n");
      exit(0);
    }
    rtn_code = handle_all(tempGroup);
    rtn_code = IK_NotCompSelect(ifield,FieldValidsVector);
    tableTemp = IK_ValidString(FieldValidsVector, indexes);
    if(tableTemp != NULL) {
      j = 0;
      while(tableTemp && *tableTemp && tableTemp[j]) {
        j++;
      } /* end while tableTemp */
      table = calloc(j+1, sizeof(char *));
      j = 0;
      for (i=0; tableTemp[i]!=(char*)NULL; ++i) {
        table[j] =
         (char *) calloc(strlen(tableTemp[j])+2,sizeof(char));
        /* start the string with the NOTCOMP (not-compatible) character */
        (void) strcpy(table[j],NOTCOMP);
        (void) strcat(table[j],tableTemp[j]);
        j++;
      }
    }
    OutputValids();
    if(table) {
	IK_FreeStringList(table);
	table = NULL;
    }
    VectorZero(FieldValidsVector, ValidsCount);
    rtn_code = IK_CompSelect(ifield,FieldValidsVector);
    table = IK_ValidString(FieldValidsVector, indexes);
    OutputValids();
  } else if(filterflag !=NULL && filterflag[0] == 'q') {
  /* This section returns the selected valids from an input query */
  /* Incompatibles have a leading '*'. */
    /* following code tries to extract the parameter values from */
    /* the odl tree */
    tempGroup = FindGroup(root, "VALIDS");
    rtn_code = handle_all(tempGroup);
    for (ifield=0; ifield < MAX_CORE_FIELD; ifield++) {
      strcpy (result_id, fieldname[ifield]);
      rtn_code = IK_NotCompSelect(ifield,FieldValidsVector);
      tableTemp = IK_ValidString(FieldValidsVector, indexes);
      k = 0;
      if(tableTemp != NULL) {
        while(tableTemp && *tableTemp && tableTemp[k]) {
          k++;
        } /* end while tableTemp */
      }
      VectorZero(FieldValidsVector, ValidsCount);
      rtn_code = IK_CompSelect(ifield,FieldValidsVector);
      tableSecond = IK_ValidString(FieldValidsVector, indexes);
      j = 0;
      while(tableSecond && *tableSecond && tableSecond[j]) {
        k++;
        j++;
      } /* end while tableTemp */
      table = calloc(k+1, sizeof(char *));
      j = 0;
      if(tableTemp != NULL) {
        for (i=0; tableTemp[i]!=(char*)NULL; ++i) {
          table[j] =
           (char *) calloc(strlen(tableTemp[j])+2,sizeof(char));
          /* start the string with the NOTCOMP (not-compatible) character */
          (void) strcpy(table[j],NOTCOMP);
          (void) strcat(table[j],tableTemp[j]);
          j++;
        }
      }
      k = 0;
      if(tableSecond != NULL) {
        for (i=0; tableSecond[i]!=(char*)NULL; ++i) {
          table[j] =
           (char *) calloc(strlen(tableSecond[k])+1,sizeof(char));
          (void) strcat(table[j],tableSecond[k]);
          j++;
          k++;
        }
      }
      OutputValids();
      if(table) {
        IK_FreeStringList(table);
    	table = NULL;
      }
      if(tableTemp) {
        IK_FreeStringList(tableTemp);
    	tableTemp = NULL;
      }
      if(tableSecond) {
        IK_FreeStringList(tableSecond);
    	tableSecond = NULL;
      }
    } /* end for ifield */
  } else if(filterflag !=NULL && filterflag[0] == 'c') {
  /* This section returns the compatible fields for a given filter name */
    /* (void) fprintf(stderr, "Get fields.\n"); */
    /* following code tries to extract the parameter values from */
    /* the odl tree */
    tempGroup = FindGroup(root, "VALIDS");
    strcpy(request_id, "RESULT");
    rtn_code = handle_all(tempGroup);
    rtn_code = IK_GetSelected(FieldValidsVector);
    tableTemp = IK_GetCompFieldStrings(FieldValidsVector);
    table = calloc(fieldCnt+1, sizeof(char *));
    if(tableTemp != NULL) {
      j = 0;
      /* Entries from tableTemp had to be copied to table because there were */
      /* blank lines in the tableTemp array. */
      for(i=0; i<fieldCnt; i++) {
        if(tableTemp[i]) {
	  table[j] = (char *) calloc(strlen(tableTemp[i])+1,sizeof(char));
	  strcpy(table[j],tableTemp[i]);
	  j++;
        }
      }
    }
    OutputValids();
    if(table) {
      IK_FreeStringList(table);
      table = NULL;
    }
    if(tableTemp) {
      IK_FreeStringList(tableTemp);
      tableTemp = NULL;
    }
  } else if(filterflag !=NULL && filterflag[0] == 'i') {
  /* This section returns the incompatible fields for given selections */
    /* (void) fprintf(stderr, "Get fields.\n"); */
    VectorZero(FieldValidsVector, ValidsCount);
    /* following code tries to extract the parameter values from */
    /* the odl tree */
    tempGroup = FindGroup(root, "VALIDS");
    strcpy(request_id, "RESULT");
    rtn_code = handle_all(tempGroup);
    table = IK_ReturnIncompatibleFields();
    OutputValids();
  } else if(filterflag !=NULL && filterflag[0] == 'o') {
    /* This section returns the ODL for an extended valids field */
    sprintf (odl_file, "%s/%s", gd_dir, EXTENDED_ODL_FILE);
    /* (void) fprintf(stderr, "Get ODL from %s.\n",odl_file); */
    /* read the search tree from the odl file */
    odl_root = NewAggregate(NULL,KA_GROUP,"EXTENDED_SEARCH",NULL);
    odlfp = fopen (odl_file, "rb");
    if (odlfp == NULL) {
       IK_Syslog(LOG_ERR,"imsdepval:  Couldn't find EXTENDED.valids file.");
       return (-1);
    }
    ReadLabel (odlfp, odl_root);
    fclose (odlfp);

    /* (void) fprintf(stderr, "Got ODL from %s.\n",odl_file); */
    /* following code tries to extract the parameter values from */
    /* the odl tree */
    tempGroup = FindGroup(root, "VALIDS");
    tempPara = FindParameter (tempGroup, "REQUEST_FIELD");
    tempValue = FirstValue (tempPara);
    if(tempValue && tempValue->item.value.string) 
     strcpy (request_id, tempValue->item.value.string);
    for (ptr = request_id;*ptr;ptr++)       /* upcase the field name */
     *ptr = toupper (*ptr);
    if (!(COUT = IW_ChunkspNew(stdout))) {
      (void) fprintf(stderr, "no chunk output stream!");
    }

    for (extGroup = FindGroup(odl_root, "SPECIALIZED_CRITERIA");
     extGroup;
     extGroup = FindNextGroup(odl_root, extGroup, "SPECIALIZED_CRITERIA")){
      extPara = FindParameter (extGroup, "CATEGORY_NAME");
      extValue = FirstValue (extPara);
      if(extValue && extValue->item.value.string) 
       strcpy (category_id, extValue->item.value.string);
      for (ptr = category_id;*ptr;ptr++)       /* upcase the field name */
       *ptr = toupper (*ptr);
      if(request_id && category_id && (strcmp(request_id, category_id) == 0)) {
	rtn_code = dump_tree(extGroup);
      }
    }
  } else if(filterflag !=NULL && filterflag[0] == 'z') {
    /* This section returns the ODL for an extended valids field */
    sprintf (odl_file, "%s/%s", gd_dir, EXTENDED_ODL_FILE);
    /* (void) fprintf(stderr, "Get ODL from %s.\n",odl_file); */
    /* read the search tree from the odl file */
    odl_root = NewAggregate(NULL,KA_GROUP,"EXTENDED_SEARCH",NULL);
    odlfp = fopen (odl_file, "rb");
    if (odlfp == NULL) {
       IK_Syslog(LOG_ERR,"imsdepval:  Couldn't find EXTENDED.valids file.");
       return (-1);
    }
    ReadLabel (odlfp, odl_root);
    fclose (odlfp);

    /* (void) fprintf(stderr, "Got ODL from %s.\n",odl_file); */
    /* following code tries to extract the parameter values from */
    /* the odl tree */
    tempGroup = FindGroup(root, "VALIDS");
    tempPara = FindParameter (tempGroup, "REQUEST_FIELD");
    tempValue = FirstValue (tempPara);
    if(tempValue && tempValue->item.value.string) 
     strcpy (request_id, tempValue->item.value.string);
    for (ptr = request_id;*ptr;ptr++)       /* upcase the field name */
     *ptr = toupper (*ptr);
    if (!(COUT = IW_ChunkspNew(stdout))) {
      (void) fprintf(stderr, "no chunk output stream!");
    }

    for (extGroup = FindGroup(odl_root, "SPECIALIZED_CRITERIA");
     extGroup;
     extGroup = FindNextGroup(odl_root, extGroup, "SPECIALIZED_CRITERIA")){
      extPara = FindParameter (extGroup, "CRITERIA_NAME");
      extValue = FirstValue (extPara);
      if(extValue && extValue->item.value.string) 
       strcpy (category_id, extValue->item.value.string);
      for (ptr = category_id;*ptr;ptr++)       /* upcase the field name */
       *ptr = toupper (*ptr);
      if(request_id && category_id && (strcmp(request_id, category_id) == 0)) {
	rtn_code = dump_tree(extGroup);
        exit(1);
      }
    }
  } else if(filterflag !=NULL && filterflag[0] == 'r') {
    /* This section returns the ODL for an extended valids field */
    sprintf (odl_file, "%s/%s", gd_dir, EXTENDED_ODL_FILE);
    /* (void) fprintf(stderr, "Get ODL from %s.\n",odl_file); */
    /* read the search tree from the odl file */
    odl_root = NewAggregate(NULL,KA_GROUP,"EXTENDED_SEARCH",NULL);
    odlfp = fopen (odl_file, "rb");
    if (odlfp == NULL) {
       IK_Syslog(LOG_ERR,"imsdepval:  Couldn't find EXTENDED.valids file.");
       return (-1);
    }
    ReadLabel (odlfp, odl_root);
    fclose (odlfp);

    /* (void) fprintf(stderr, "Got ODL from %s.\n",odl_file); */
    /* following code tries to extract the parameter values from */
    /* the odl tree */
    tempGroup = FindGroup(root, "VALIDS");
    strcpy (request_id, "Y");
    if (!(COUT = IW_ChunkspNew(stdout))) {
      (void) fprintf(stderr, "no chunk output stream!");
    }

    count = 0;
    /* Count the number of "SPECIALIZED_CRITERIA" groups in the file */
    for (extGroup = FindGroup(odl_root, "SPECIALIZED_CRITERIA");
     extGroup;
     extGroup = FindNextGroup(odl_root, extGroup, "SPECIALIZED_CRITERIA")){
     count++;
    }
    table = calloc(count+1, sizeof(char *));
    
    count = 0;
    for (extGroup = FindGroup(odl_root, "SPECIALIZED_CRITERIA");
      extGroup;
      extGroup = FindNextGroup(odl_root, extGroup, "SPECIALIZED_CRITERIA")){
      extPara = FindParameter (extGroup, "RESULTS_SELECTABLE");
      extValue = FirstValue (extPara);
      if(extValue && extValue->item.value.string) 
       strcpy (category_id, extValue->item.value.string);
      for (ptr = category_id;*ptr;ptr++)       /* upcase the field name */
      *ptr = toupper (*ptr);
      if(request_id && category_id && (strcmp(request_id, category_id) == 0)) {
        extPara = FindParameter (extGroup, "CRITERIA_NAME");
        extValue = FirstValue (extPara);
        if(extValue && extValue->item.value.string) 
         strcpy (result_id, extValue->item.value.string);
        for (ptr = result_id;*ptr;ptr++)       /* upcase the field name */
         *ptr = toupper (*ptr);
        table[count] = calloc(strlen(result_id)+1, sizeof(char));
        strcpy(table[count],result_id);
        count++;
      }
      category_id[0] = NULL;
    }
    table[count] = NULL;
    OutputValids();
    if(table) {
      IK_FreeStringList(table);
      table = NULL;
    }
  } else if(filterflag !=NULL && filterflag[0] == 'e') {
    /* This section returns the ODL for an extended valids field */
    sprintf (odl_file, "%s/%s", gd_dir, EXTENDED_ODL_FILE);
    /* (void) fprintf(stderr, "Get ODL from %s.\n",odl_file); */
    /* read the search tree from the odl file */
    odl_root = NewAggregate(NULL,KA_GROUP,"EXTENDED_SEARCH",NULL);
    odlfp = fopen (odl_file, "rb");
    if (odlfp == NULL) {
       IK_Syslog(LOG_ERR,"imsdepval:  Couldn't find EXTENDED.valids file.");
       return (-1);
    }
    ReadLabel (odlfp, odl_root);
    fclose (odlfp);

    /* (void) fprintf(stderr, "Got ODL from %s.\n",odl_file); */
    /* following code tries to extract the parameter values from */
    /* the odl tree */
    tempGroup = FindGroup(root, "VALIDS");
    strcpy (request_id, "Y");
    if (!(COUT = IW_ChunkspNew(stdout))) {
      (void) fprintf(stderr, "no chunk output stream!");
    }

    count = 0;
    for (extGroup = FindGroup(odl_root, "SPECIALIZED_CRITERIA");
     extGroup;
     extGroup = FindNextGroup(odl_root, extGroup, "SPECIALIZED_CRITERIA")){
     count++;
    }
    table = calloc(count+1, sizeof(char *));
    
    count = 0;
    for (extGroup = FindGroup(odl_root, "SPECIALIZED_CRITERIA");
     extGroup;
     extGroup = FindNextGroup(odl_root, extGroup, "SPECIALIZED_CRITERIA")){
     extPara = FindParameter (extGroup, "DEFAULT");
     extValue = FirstValue (extPara);
     if(extValue && extValue->item.value.string) 
      strcpy (category_id, extValue->item.value.string);
     for (ptr = category_id;*ptr;ptr++)       /* upcase the field name */
      *ptr = toupper (*ptr);
     if(request_id && category_id && (strcmp(request_id, category_id) == 0)) {
        extPara = FindParameter (extGroup, "CRITERIA_NAME");
        extValue = FirstValue (extPara);
        if(extValue && extValue->item.value.string) 
         strcpy (result_id, extValue->item.value.string);
        for (ptr = result_id;*ptr;ptr++)       /* upcase the field name */
         *ptr = toupper (*ptr);
        table[count] = calloc(strlen(result_id)+1, sizeof(char));
        strcpy(table[count],result_id);
        count++;
      }
      category_id[0] = NULL;
    }
    table[count] = NULL;
    OutputValids();
    if(table) {
      IK_FreeStringList(table);
      table = NULL;
    }
  } else if(filterflag !=NULL && filterflag[0] == 'a') {
    j=0;
    /* following code tries to extract the parameter values from */
    /* the odl tree */
    tempGroup = FindGroup(root, "VALIDS");
    tempPara = FindParameter (tempGroup, "REQUEST_FIELD");
    tempValue = FirstValue (tempPara);
    /*fprintf (stderr, "req-field %s \n", tempValue->item.value.string);*/
    if(tempValue && tempValue->item.value.string) 
     strcpy (request_id, tempValue->item.value.string);
    rtn_code = handle_attributes(tempGroup);
  } else if(filterflag !=NULL && filterflag[0] == 'b') {
    for (k=0; k < MAX_SIMPLE_FIELD; k++) {
      if(k == VALIDS_DATASET) continue;
      /* initialize dependent valids table */
      IK_GetDepValids (&table, NULL, NULL);
      IK_GetDepValids (&table, fieldname[k], NULL);
      OutputValids();
      if(table) {
        IK_FreeStringList(table);
        table = NULL;
      }
    }
  } else if(filterflag !=NULL && filterflag[0] == 'd') {
    /* following code tries to extract the parameter values from */
    /* the odl tree */
    tempGroup = FindGroup(root, "VALIDS");
    j = 0;
    if (handle_value (&fields, tempGroup, &j, attributes)) {
      for (k=0; k < MAX_SIMPLE_FIELD; k++) {
        if(k == VALIDS_DATASET) continue;
        /* initialize dependent valids table */
        IK_GetDepValids (&table, NULL, NULL);
        /* call GetDepValids to get updated dependent valids table */
        IK_GetDepValids (&table, fieldname[k], &fields[0]);
        VectorZero(FieldValidsVector, ValidsCount);
        rtn_code = IK_CompSelect(k,FieldValidsVector);
        table = IK_ValidString(FieldValidsVector, indexes);
        if(table && *table && table[0]) {
          OutputValids();
        }
        if(table) {
          IK_FreeStringList(table);
          table = NULL;
        }
      }
    }
  } else if(filterflag !=NULL && filterflag[0] == 'g') {
    /* following code tries to extract the parameter values from */
    /* the odl tree */
    tempGroup = FindGroup(root, "VALIDS");
    j = 0;
    if (handle_value (&fields, tempGroup, &j, pattern)) {
      for (k=0; k < fieldCnt; k++) {
        /* initialize dependent valids table */
        IK_GetDepValids (&table, NULL, NULL);
        /* call GetDepValids to get updated dependent valids table */
        IK_GetDepValids (&table, fieldname[k], &fields[0]);
        VectorZero(FieldValidsVector, ValidsCount);
        rtn_code = IK_CompSelect(k,FieldValidsVector);
        table = IK_ValidString(FieldValidsVector, indexes);
        if(table && *table && table[0]) {
          IK_FreeStringList(table);
          table = NULL;
          table = calloc(2, sizeof(char *));
          table[0] = calloc(strlen(fieldname[k])+1, sizeof(char));
          strcpy(table[0],fieldname[k]);
          table[1] = NULL;
          OutputValids();
        }
        if(table) {
          IK_FreeStringList(table);
          table = NULL;
        }
      }
    }
  } else {

    j=0;
    /* initialize dependent valids table */
    IK_GetDepValids (&table, NULL, NULL); 
    /* following code tries to extract the parameter values from */
    /* the odl tree */
    tempGroup = FindGroup(root, "VALIDS");
    tempPara = FindParameter (tempGroup, "REQUEST_FIELD");
    tempValue = FirstValue (tempPara);
    /*fprintf (stderr, "req-field %s \n", tempValue->item.value.string);*/
    if(tempValue && tempValue->item.value.string) 
     strcpy (request_id, tempValue->item.value.string);
    rtn_code = handle_all(tempGroup);
    IK_GetDepValids (&table, request_id, NULL);
    OutputValids();
  }
  if(tableflag && line_out) {
    if (outfile) { 
     fprintf(outfile, "\n");     /* needed to end record [eryq] */
     fclose (outfile);
    } else {
     fprintf(stdout, "\n");     /* needed to end record [eryq] */
    }
  }
  if(table) {
    IK_FreeStringList(table);
	table = NULL;
  }
  FreeAllocated();
  return 1;
}
void OutputValids() {
  int i;
  if(filterflag !=NULL && filterflag[0] != 'q') strcpy(result_id, "RESULT");
  if(filterflag !=NULL && filterflag[0] == 'v') strcpy(result_id, "VALIDS");
  if(filterflag !=NULL && filterflag[0] == 'f') strcpy(result_id, "FILTERS");
  if (table) tableflag = 1;
  if (!line_out) {
    if (outfile) {
      build_out_odl (table, result_id, outfile);
      fclose (outfile);
    }
    else
      build_out_odl (table, result_id, stdout);
  }
  else {
    if (outfile) { 
      if(filterflag !=NULL && filterflag[0] == 'f' &&
         table != NULL && *table != NULL) {
	fprintf (outfile, "%s: >All Compatible %ss\n", 
		 result_id, newfieldname[ifield]);
      }
      if(table != NULL) {
        for (i=0; table[i]!=(char*)NULL; ++i) {
          for (ptr = table[i];*ptr;ptr++)       /* upcase the field name */
            *ptr = toupper (*ptr);
          if(strcmp(table[i],"<UNSPECIFIED>") != 0) {
            fprintf (outfile, "%s: %s\n", result_id, table[i]);
          }
        }
      }
    }
    else
      if(filterflag !=NULL && filterflag[0] == 'f' && 
         table != NULL && *table != NULL) {
	fprintf (stdout, "%s: >All Compatible %ss\n", 
		 result_id, newfieldname[ifield]);
      }
      if(table != NULL) {
        for (i=0; table[i]!=(char*)NULL; ++i) {
          for (ptr = table[i];*ptr;ptr++)       /* upcase the field name */
            *ptr = toupper (*ptr);
          if(strcmp(table[i],"<UNSPECIFIED>") != 0) {
            fprintf (stdout, "%s: %s\n", result_id, table[i]);
          }
        }
      }
  }
  return;
}

void build_out_odl (char **table_ary, char *request, FILE *fp)
{
  AGGREGATE root;
  AGGREGATE tree;
  PARAMETER para;
  VALUE     new;
  VALUE_DATA node;
  int i;

  root = NewAggregate(NULL, KA_GROUP,"root",NULL);
  tree = NewAggregate(root, KA_GROUP,"VALIDS", NULL);
  para = NewParameter(tree, KP_ATTRIBUTE, request);
  para->value_kind = KV_SEQUENCE;
  if(table_ary && table_ary[0]) {
    for (i=0; table_ary[i]!=(char *)NULL; ++i) {
      for (ptr = table_ary[i];*ptr;ptr++)       /* upcase the field name */
        *ptr = toupper (*ptr);
      if(strcmp(table_ary[i],"<UNSPECIFIED>") != 0) {
        node = ODLConvertString(table_ary[i], strlen(table_ary[i]));
        new = NewValue(para, &node);
        para->rows = 1;
        para->columns = para->value_count;
      }
    }
    WriteLabel (fp, root);
  }
}

/************************************************************
 * handle_value
 ************************************************************
 *
 * DESCRIPTION
 *    This function looks for the requestField in the group, tgroup.
 *    If it finds the group it returns the valids in the field_ary along
 *    with *num of valids found in the group.
 */
int handle_value (char ***field_ary, AGGREGATE tGroup, int *num, 
		  char *requestField)
{
  VALUE tempValue;
  VALUE fieldNameValue;
  VALUE fieldValue;
  PARAMETER tempPara;
  PARAMETER fieldPara;
  PARAMETER fieldNamePara;
  AGGREGATE fieldGroup;
  char category_id[MAXLEN];
  char requestValid[MAXLEN];
  char requestFieldCopy[MAXLEN];

  
  strcpy(requestFieldCopy, requestField);
  for (ptr = requestFieldCopy;*ptr;ptr++)       /* upcase the field name */
   *ptr = toupper (*ptr);

  tempPara = FindParameter (tGroup, requestFieldCopy);
  if (tempPara) {
    /* fprintf (stderr, "found Para\n"); */
    tempValue = FirstValue (tempPara);
    while (tempValue) {
      /* fprintf (stderr, "%s is %s \n", requestFieldCopy, tempValue->item.value.string); */
      (*field_ary)[*num] = 
       (char *)calloc(strlen(tempValue->item.value.string)+1, sizeof(char));
      strcpy ((*field_ary)[*num], tempValue->item.value.string);
      (*num)++;
      tempValue = NextValue (tempValue);
    }
    (*field_ary)[*num] = (char *) NULL;
    return (1);
  }
  else {
    for (fieldGroup = FindGroup(tGroup, "FIELD");
     fieldGroup;
     fieldGroup = FindNextGroup(tGroup, fieldGroup, "FIELD")){
      fieldNamePara = FindParameter (fieldGroup, "NAME");
      if (fieldNamePara) {
        fieldNameValue = FirstValue (fieldNamePara);
        if(fieldNameValue && fieldNameValue->item.value.string)
         strncpy (category_id, fieldNameValue->item.value.string,MAXLEN-1);
        category_id[MAXLEN-1] = '\0';
	for (ptr = category_id;*ptr;ptr++)       /* upcase the field name */
	 *ptr = toupper (*ptr);
        if(requestFieldCopy && category_id && 
         (strcmp(requestFieldCopy, category_id) == 0)) {
	  fieldPara = FindParameter (fieldGroup, "VALUES");
	  fieldValue = FirstValue (fieldPara);
          while (fieldValue) {
            if(fieldValue && fieldValue->item.value.string)
              strncpy (requestValid, fieldValue->item.value.string,MAXLEN-1);
            requestValid[MAXLEN-1] = '\0';
            /* fprintf (stderr, "%s is %s \n", requestFieldCopy, requestValid); */
            (*field_ary)[*num] = 
	     (char *)calloc(strlen(requestValid)+1, sizeof(char));
            strcpy ((*field_ary)[*num], requestValid);
            (*num)++;
            fieldValue = NextValue (fieldValue);
          }
          (*field_ary)[*num] = (char *) NULL;
          return (1);
        }
      }
    }
  }
    return (0);
}
static AGGREGATE IK_BuildSearchTree(AGGREGATE root, char *filename)
 
/*
 * Name IK_BuildSearchTree
 *
 * Description:  This function will read in the odl label file specified in
 * parameter, and build an odl tree from it.  If a filename is not supplied
 * read from stdin.
 *
 * Parameters:
 *   AGGREGATE  root   the empty root odl tree
 *   char *filename    filename for the odl label file
 *
 * Return Values:
 *   Return AGGEGATE that contains the odl tree. Exit the program if fail.
 *
 * Global Variable Used:
 *   none
 *
 */
{
  FILE *fp;
   if (filename) {
     fp = fopen (filename, "r");
     ReadLabel(fp, root);
     fclose (fp);
   }
   else
     ReadLabel (stdin, root);
   return (root);
}
/************************************************************
 * AllocateBitvectors
 ************************************************************
 *
 * DESCRIPTION
 *    Allocate 2 bitvectors and 3 index arrays.
 *
 *    bitvectors:
 *       FieldValidsVector
 *       FiltersVector
 *    index arrays:
 *       indexes
 *       filter_indexes
 *       filter_saved 
 *
 */

void AllocateBitvectors()
{


   indexes = (int *) calloc(ValidsCount, sizeof(int));
   filter_indexes = (int *) calloc(FiltersCount, sizeof(int));
   filter_saved = (int *) calloc(FiltersCount, sizeof(int));

   bytesPerRow = (ValidsCount+CHAR_BIT-1)/CHAR_BIT;
   FieldValidsVector = (IK_BITVECTOR *) calloc(bytesPerRow, sizeof(char));
   FiltersVector = (IK_BITVECTOR *) calloc(bytesPerRow, sizeof(char));

   VectorZero(FieldValidsVector, ValidsCount);
   VectorZero(FiltersVector, FiltersCount);

   return;
}

/************************************************************
 * FreeAllocated
 ************************************************************
 *
 * DESCRIPTION
 *    Free the structures allocated in AllocateBitvectors
 */

void FreeAllocated()
{
   /* Free allocated bitvectors. */
   if (FieldValidsVector != NULL) {
      free(FieldValidsVector);
      FieldValidsVector = NULL;
   }
   if (FiltersVector != NULL) {
      free(FiltersVector);
      FiltersVector = NULL;
   }
   if (filter_indexes != NULL) {
      free(filter_indexes);
      filter_indexes = NULL;
   }
   if (indexes != NULL) {
      free(indexes);
      indexes = NULL;
   }
   if (filter_saved != NULL) {
      free(filter_saved);
      filter_saved = NULL;
   }
   return;
}

/************************************************************
 * dump_tree
 ************************************************************
 *
 * DESCRIPTION
 *    Dump the chunk structure of an aggregate, as a tree.
 */

static int dump_tree(AGGREGATE root)
{
  IW_ODLWriteChunk(COUT, root, NULL);
  return 0;
}
/************************************************************
 * ReturnFieldNumber
 ************************************************************
 *
 * DESCRIPTION
 *    Given a string that represents a field (request_id) return
 *    the field number.
 */

static int ReturnFieldNumber(char *request_id)
{
    int ifield;
    for (ptr = request_id;*ptr;ptr++)       /* upcase the field name */
     *ptr = toupper (*ptr);
    /* Get the field number given the field name */
    ifield = MISSING;
    for(i=0; i<fieldCnt && ifield == MISSING; i++) {
	strcpy(category_id, fieldname[i]);
        for (ptr = category_id;*ptr;ptr++)       /* upcase the field name */
         *ptr = toupper (*ptr);
        if((strlen(request_id) > (int) 1) &&
         (strcmp(request_id, category_id)==0)) {
            ifield = i;
        } /* end if strlen */
    } /* end for */
    return ifield;
}

/************************************************************
 * ReturnFieldCategory
 ************************************************************
 *
 * DESCRIPTION
 *    Given a string that represents a field (request_id) return
 *    the field name.
 *
 * This function mas primarily added to solve the problem of the extended
 * valids.  The extended valids categories are stored in the valids files
 * as mixed case.  The odl writes them as CAPS, that created a problem
 * because there was not a match.  This function returns the field name
 * for the category as it is stored in fieldname[index].
 *
 * Lorena Marsans - 7/17/98
 *
 */

static char* ReturnFieldCategory(char *request_id)
{
    char* category;
    int ifield;
    for (ptr = request_id;*ptr;ptr++)       /* upcase the field name */
     *ptr = toupper (*ptr);
    /* Get the field number given the field name */
    ifield = MISSING;
    for(i=0; i<fieldCnt && ifield == MISSING; i++) {
        strcpy(category_id, fieldname[i]);
	category = calloc(strlen(category_id)+1, sizeof(char));
	strcpy(category, fieldname[i]);
        for (ptr = category_id;*ptr;ptr++)       /* upcase the field name */
         *ptr = toupper (*ptr);
        if((strlen(request_id) > (int) 1) &&
         (strcmp(request_id, category_id)==0)) {
            return category;
        } /* end if strlen */
        free (category);
    } /* end for */
    return NULL;
}
/************************************************************
 * handle_all
 ************************************************************
 *
 * DESCRIPTION
 *    This process selects all of the dependent valids before the query
 *    action is performed.  It takes a group from the input ODL file and 
 *    loops through all of the fields.  If the field is found, the valids 
 *    of that field are selected in dependent valids with the call to 
 *    IK_GetDepValids.
 */
static int handle_all(AGGREGATE tempGroup) {
    /* This loops through all of the dependent valids */
    j = 0;
    for (k=0; k < fieldCnt; k++) {
      if (handle_value (&fields, tempGroup, &j, fieldname[k])){
        /* call GetDepValids to get updated dependent valids table */
        IK_GetDepValids (&table, fieldname[k], &fields[0]);
        /* fprintf (stderr, "item in table %s \n", table[0]); */
        for (i=0; i < j; i++) free (fields[i]);
        j = 0;
        if(table) {
          IK_FreeStringList(table);
	  table = NULL;
        }
      }
    }
  return 0;
}

/************************************************************
 * handle_attributes
 ************************************************************
 *
 * DESCRIPTION
 *    This process takes a list of attributes and selects all of the 
 *    dependent valids before the query
 *    action is performed.  It takes a group from the input ODL file and 
 *    loops through all of the fields.  If the field is found, the valids 
 *    of that field are selected in dependent valids with the call to 
 *    IK_GetDepValids.
 */
static int handle_attributes(AGGREGATE tempGroup) {
    /* This loops through all of the dependent valids */
    j = 0;
    if (handle_value (&fields, tempGroup, &j, attributes)) {
      for (k=0; k < fieldCnt; k++) {
          /* initialize dependent valids table */
          IK_GetDepValids (&table, NULL, NULL);
          /* call GetDepValids to get updated dependent valids table */
          IK_GetDepValids (&table, fieldname[k], &fields[0]);
          /* fprintf (stderr, "item in table %s \n", table[0]); */
          j = 0;
          VectorZero(FieldValidsVector, ValidsCount);
          rtn_code = IK_CompSelect(k,FieldValidsVector);
          table = IK_ValidString(FieldValidsVector, indexes);
          if(table && *table && table[0]) {
            if(strcmp(fieldname[k],"DATASET_ID") != 0) {
              IK_FreeStringList(table);
              table = NULL;
              IK_GetDepValids (&table, request_id, NULL);
            }
            OutputValids();
          }
          if(table) {
            IK_FreeStringList(table);
            table = NULL;
          }
      }
      for (i=0; i < j; i++) free (fields[i]);
    }
  return 0;
}

static void usage() {
 fprintf(stderr, " USAGE:\n");
 fprintf(stderr, "    imsdepval [options] [action]\n");
 fprintf(stderr, " The options are:\n");
 fprintf(stderr, 
  "  -l            Use chunking output instead of ODL output.\n");
 fprintf(stderr, "  -i <filename> The input request ODL file.\n");
 fprintf(stderr, 
  "                If this option is missing it will take input");
 fprintf(stderr, " from stdin.\n");
 fprintf(stderr, "  -o <filename> The output file.\n");
 fprintf(stderr, 
  "                If this option is missing it will output to stdout.\n");
 fprintf(stderr, " The actions are:\n");
 fprintf(stderr, 
  " -mf Return the filters that are compatible with the REQUEST_FIELD.\n");
 fprintf(stderr, 
  " -mv Return the valids that are compatible with the REQUEST_FIELD\n");
 fprintf(stderr, 
  "     and the optional filter or optional pattern supplied.\n");
 fprintf(stderr, 
  " -ms Return the list of selected valids based on the REQUEST_FIELD.\n");
 fprintf(stderr, 
  "     Incompatibles are prefixed by an asterisk '*'.\n");
 fprintf(stderr, 
  " -mc Return the list of compatible fields based on the input ODL.\n");
 fprintf(stderr, 
  " -mi Return the list of fields that have become incompatible as\n");
 fprintf(stderr, 
  "     a result of selections made in the input ODL file.\n");
 fprintf(stderr, 
  " -mo Return the SPECIALIZED_CRITERIA ODL for the REQUEST_FIELD.\n");
 fprintf(stderr, 
  " -ma Return the valids that are compatible with the REQUEST_FIELD\n");
 fprintf(stderr, 
  "     that are listed with the keyword 'ATTRIBUTES', where\n");
 fprintf(stderr, 
  "     the strings that follow the keyword can be any of the 7\n");
 fprintf(stderr, 
  "     core attributes.\n");
 fprintf(stderr, 
  " -mb Return the all of the valids for DATA CENTER, DATA SET,\n");
 fprintf(stderr, 
  "     SOURCE, SENSOR and PARAMETER.\n");
 fprintf(stderr, 
  " -md Return the matching valids that are listed with the\n");
 fprintf(stderr, 
  "     keyword 'ATTRIBUTES'.\n");
 fprintf(stderr, 
  " -mg Return the field(s) associated with the keyword 'ATTRIBUTES'.\n");
 fprintf(stderr,
  " -mq Return a query of selected valids.\n");
 fprintf(stderr,
  "     Incompatibles are prefixed by an asterisk '*'.\n");
 fprintf(stderr,
  " -mr Return a list of CRITERIA_NAMEs in the EXTENDED.valids file\n");
 fprintf(stderr,
  "     that have the flag RESULTS_SELECTABLE = ''Y''.\n");
 fprintf(stderr,
  " -me Return a list of CRITERIA_NAMEs in the EXTENDED.valids file\n");
 fprintf(stderr,
  "     that have the flag DEFAULT = ''Y''.\n");
 fprintf(stderr, "\n");
 fprintf(stderr,
  " -mz Return the SPECIALIZED_CRITERIA ODL for the REQUEST_FIELD\n");
 fprintf(stderr,
  "     where the REQUEST_FIELD is actually a CRITERIA_NAME.\n");
 fprintf(stderr, "\n");
 fprintf(stderr, 
  "     The default action if no action is supplied is to return\n");
 fprintf(stderr, 
  "     the dependent valids that are compatible with the REQUEST_FIELD\n");
 fprintf(stderr, 
  "     based on the selections in the rest of the input file.\n");
 fprintf(stderr, 
  "\n     Any other options will return this usage page.\n");
}
