/*************************************************************
 *
 * NAME
 *    imsgdbdump - dump information from the IMS GDB files.
 *
 * DESCRIPTION
 *    Dumps out the information lurking in the dataset.gdb and
 *    granule.gdb files after inventory/directory searches.
 *
 *    The output format is an RFC-822-like syntax, consisting
 *    of chunks like you see at the tops of mail messages:
 *
 *        X-Type: GRANULE
 *        Archive-Id: GSFC
 *        Dataset-Id: XJHS00 TOP 10 AVHRR READINGS
 *        Granule-Id: 18323
 *        Start-date: 1992-09-12T12:34:56Z
 *        Stop-date: 1992-09-12T12:34:58Z
 *
 *        X-Type: GRANULE
 *        Archive-Id: GSFC
 *        Dataset-Id: XJHS00 TOP 10 AVHRR READINGS
 *        Granule-Id: 18324
 *        ...
 *
 *    This output is easy to generate and very efficiently parsed by
 *    both C and Perl.
 *
 * USAGE
 *    No arguments gives a summary of the usage.  The usual invocation 
 *    form is:
 *        
 *        imsgdbdump [options] <action...>
 *
 *    The options are:
 *
 *        -v      Verbose: output debug messages to the stderr (normally,
 *                only warnings, errors, and above are echoed to stderr).
 *        -s      Summary output only (good for a quick peek)
 *
 *    The following actions are still supported:
 *
 *        all
 *            Dump all dataset and granule info.
 *
 *        list
 *            Dump dataset and granule info from client.list.
 *
 *        ps <gkey>
 *            Dump all packages in gkey.
 *
 *        pt <gkey>
 *            Dump all packages in gkey, using "tree" format.
 *
 *    The following actions *may* still work, but are unsupported and 
 *    generally untested:
 *
 *        ds
 *            Dump info on all datasets.
 *
 *        d <dskey>
 *            Dump info on a single dataset <dskey>.
 *
 *        gd <dskey>
 *            Dump info for all granules in dataset <dskey>.
 *
 *        g <gkey>
 *            Dump info for a single granule <gkey>.
 *
 *   The following actions have been permanently removed:
 *
 *        gp <pkey>
 *            Dump brief info for all granules in package <pkey>.
 *            The <pkey> is of any of these forms (* shown as #):
 *
 *                <archtag>/<dstag>/#           all granules in dataset
 *                <archtag>/<dstag>/#/<gid>     just one granule
 *                <archtag>/<dstag>/<pkid>      all granules in package
 *
 *        gdf <dskey-file>
 *            Dump brief info for all granules across all datasets
 *            listed in the given <dskey-file>.
 * 
 * NOTES
 *    In the code, we distinguish between the BINARY representation
 *    of dataset/granule sort keys, and the ASCII representation
 *    used by the WWW software, as follows:
 *
 *         dskey      A binary dataset sort key (IK_SortKey)
 *         gkey       A binary granule sort key (IK_SortKey)
 *
 *         web_dskey  An ASCII dataset sort key, like "1/23"
 *         web_gkey   An ASCII granule sort key, like "1/23/ABCDEFG"
 *
 * AUTHOR
 *    Eryq wrote the original version, in Jul 96.
 */


/************************************************************
 *
 * Includes...
 *
 ************************************************************/

#include "Gaea_config.h"

#include <stdio.h>
#include <assert.h>
#include <unistd.h>

#include <string.h>
#include <stdarg.h>
#include <ctype.h>              /* needed for toupper() */

#include "IK_Ims.h"
#include "IK_DataDic.h"
#include "IK_Errno.h"
#include "IK_Syslog.h"
#include "IK_Granule.h"
#include "IK_Msg.h"

/* The following include added for new tags */
#include "IK_Map.h"

/* The following include was added for the new valids format */
#include "IK_NewValids.h"

/* These are general tools: */
#include "IW_New.h"
#include "IW_Hasht.h"
#include "IW_Str.h"
#include "IW_Chunk.h"
#include "IW_Key.h"
#include "IW_ODL.h"

/* Get definitions for debug, info, etc.: */
#define IW_Log_NICENAMES 1
#include "IW_Log.h"

/************************************************************
 *
 * Constants...
 *
 ************************************************************/

/* Endian-ness: */
#if defined(LINUX386) || defined(__linux__)
#define BIG_ENDIAN 0 
#else
#define BIG_ENDIAN 1
#endif

/* GDBM-related constants: */
#define READER    1
#define WRCREATE  0

/* How detailed we want our output: */
#define ONELINE  0
#define SUMMARY  1
#define DETAILED 2

/* What do we want to dump? */
#define DUMP_ALL    0
#define DUMP_LIST   1

#define DYNAMIC_FILE "dynamic.odl" /* name of the dynamic valids file */
#define PACKAGE_IDS "package_ids" /* the file that contains package ids */
#define MAX_CORE_FIELD (7) /* maximum # of core attributes for dynamic valids */
#define MAXLEN (81) /* the maximum number of bytes in a valid + 1 */

/* maximum number of characters in input_string and line buffer */
#define MAXLINE (82)

/************************************************************
 *
 * Globals...
 *
 ************************************************************/

/* The errno variable: */
IK_Errno IK_ImsErrno;

int GNumDatasets = -1;

/* Hash of timespans: */
IW_Hasht *GTimespans;

/* The output channel for the chunks: */
IW_Chunksp *COUT;

/* RCS id: */
static char *RCSID = "$Id: imsgdbdump.c,v 1.34 2003/10/02 17:13:56 amrhein Exp $";

/* Hash of keys we got the last time through: */
IW_Hasht *ListHash = NULL;

/* Verbose output (for debugging)? */
int Verbose = 0;

/* Level of detail: */
int Detail = DETAILED;

/* The following added for dynamic valids */
extern char **fieldname; /* names of the fields */
static int ValidsCount; /* number of valids in the list */
char *ptr; /* a tmp pointer used to upcase the field name */
char package_ids_file[1024];  /* name of package ids file to open */

/************************************************************
 *
 * Prototypes
 *
 ************************************************************/

static int calc_dataset_timespans(IK_SortKey *, IK_Granule *, void *);

/* Internal functions... */
static void   usage();
static void   puke(char *, ...);
int handle_value (char ***, AGGREGATE, int *, char *);

/*------------------------------ begin timespan submodule -- */

/************************************************************
 *
 * IWM_Timespan
 *
 ************************************************************
 *
 * The following section of code deals with spans of time.
 */

typedef struct IWM_Timespan {
    char start[IK_TIME_STAMP_SZ+1];
    char stop[IK_TIME_STAMP_SZ+1];
} IWM_Timespan;


/************************************************************
 * IWM_TimespanAlloc
 ************************************************************/

IWM_Timespan *IWM_TimespanAlloc(char *start, 
				char *stop)
{
  IWM_Timespan *new = IW_New(IWM_Timespan);
  if (new) {
    strcpy(new->start, start);
    strcpy(new->stop, stop);
  }
  return new;
}

/************************************************************
 * IWM_TimespanFree
 ************************************************************/

void IWM_TimespanFree(this)
     IWM_Timespan *this;
{
  IW_Free(this);
}

/*------------------------------ end timespan submodule -- */

/************************************************************
 *
 * Miscellaneous utilities...
 *
 ************************************************************/

/************************************************************
 * dying_gasp
 ************************************************************
 * Install this as the IW fail handler... it will get called automatically
 * if something fatal happens inside libIW.
 * It outputs appropriate failure info, and then exits.
 */ 

static void dying_gasp(char *reason)
{
  err("Fatal error: %s", reason);
  fprintf(stdout, "\nX-Error: %s\n\n", reason);
  exit(-1);
}

/************************************************************
 * puke
 ************************************************************
 * Output and log error, and die.  
 * Invoke this like printf().
 */

static void puke(char *fmt, ...)
{
  char message[2048];
  va_list args;
    
  va_start(args, fmt);
  vsprintf(message, fmt, args);
  va_end(args);
  dying_gasp(message);
}

/************************************************************
 * calc_dataset_timespans
 ************************************************************
 * Calculate the timespan of a dataset.
 */ 

static int calc_dataset_timespans(IK_SortKey *key,
				  IK_Granule *granule,
				  void * arg)
{
  IWM_Timespan *timespan;
  char keybuf[20];

  debug("Converting dataset parts to web key");

  /* Construct the web key for this granule's dataset: */
  IW_DatasetKeyPartsToWeb(granule->archive_tag, granule->dataset_tag, keybuf);
  debug("Got: %s", keybuf);


  /* Lookup any existing timespan info: */
  if (timespan = IW_HashtAt(GTimespans, keybuf)) {   /* existing dataset */
    
    /* If new start date is earlier than old, replace old with new: */
    if (strcmp(granule->start_date, timespan->start) < 0) {
      strcpy(timespan->start, granule->start_date);
    }

    /* If new stop date is later than old, replace old with new: */
    if (strcmp(granule->stop_date, timespan->stop) > 0) {
      strcpy(timespan->stop, granule->stop_date);
    }
    
  }
  else {     /* new dataset */

    /* Create and store this timespan: */
    if (!(timespan = IWM_TimespanAlloc(granule->start_date, 
				   granule->stop_date))){
      return err("IWM_TimespanAlloc");
    }
    if (IW_HashtAtPut(GTimespans, keybuf, (void *)timespan) < IW_HashtOK) {
      return err("IW_HashtAtPut");
    }    
  }

  /* Done! */
  return 0;
}

/************************************************************
 *
 * good_if_*
 *
 ************************************************************
 * Tests for the various "do" iterator functions.
 */

/************************************************************
 * good_if_match_one_dataset
 ************************************************************
 * Does this granule match the given dataset key?
 */ 

int good_if_match_one_dataset(IK_SortKey *gkey, 
			      void *dskey)
{
  IK_ArchiveTag g_archTag, d_archTag;
  IK_DatasetTag g_dsTag, d_dsTag;
  static char g_gid[IK_GRANULE_ID_SZ+1];    /* static for efficiency */

  IK_StripGranuleKey(g_archTag, g_dsTag, g_gid, gkey);
  IK_StripDatasetKey(d_archTag, d_dsTag, (IK_SortKey *)dskey);
  return ((memcmp(g_archTag, d_archTag, IK_ARCHIVE_TAG_SZ) == 0) &&
	  (memcmp(g_dsTag, d_dsTag, IK_DATASET_TAG_SZ) == 0));
}

/************************************************************
 * good_if_in_list
 ************************************************************
 * Is this granule in the list of most-recent results?
 */ 

int good_if_in_list(IK_SortKey *gkey, void *arg)
{
  char web_gkey[IK_GRANULE_ID_SZ+32];    /* gkey, in web */

  /* Format web key: */
  IW_GranuleKeyToWeb(gkey, web_gkey);
  
  /* In the list? */
  (int)(IW_HashtAt(ListHash, web_gkey));
}

/************************************************************
 *
 * do_*
 *
 ************************************************************
 * The following are for iterating over one or many datasets/granules.
 */

/************************************************************
 * do_granules
 ************************************************************
 * Perform an "action" for all granules where the granule key "isgood".
 *
 * If no "isgood" test was specified, does action for all granules.
 */ 

int do_granules(int (*isgood)(IK_SortKey *, void *),
		void * isgood_arg,
		int (*action)(IK_SortKey *, IK_Granule *, void *),
		void * action_arg)
{
  IK_Granule *granule;
  IK_SortKey *key;
  int result;

  /* Step through: */
  key = IK_FirstGranuleKey();
  while (key) {

    /* Get the next granule key... if we only want particular
     * ones, and this doesn't match, then discard it and continue
     * with next granule:
     */
    if (isgood) {                           /* there is a test... */
      if (!((*isgood)(key, isgood_arg))) {     /* ...and we failed: */
	free(key);
	key = IK_NextGranuleKey();	
	continue;
      }
    }

    /* Get the granule! */
    if (IK_GetGranule(key, &granule) != 0) {
      free(key);
      return err("couldn't get granule (maybe it doesn't exist?)");
    }

    /* Call the callback: */
    if (action) {
      debug("Calling action");
      if ((result = (*action)(key, granule, action_arg)) != 0) {
	free(key);
	free(granule);
	return err("do_granules callback failed: %d", result);
      }
    }

    /* Next: */
    free(granule);
    free(key);
    key = IK_NextGranuleKey();
  }

  return 0;
}

/************************************************************
 * do_dataset
 ************************************************************
 * Perform an action for a single dataset.
 * 
 * The dskey specifies the dataset.  The action is then invoked as:
 *
 *     (*action)(dskey, dataset, arg);
 */ 

int do_dataset(IK_SortKey *dskey, 
	       int (*action)(), void *arg)
{
  IK_Dataset *dataset;
  int result;


  /* Calculate the timespan for each dataset: */
  if (do_granules(NULL, NULL, calc_dataset_timespans, NULL) != 0) {
    return err("do_granules");
  }

  /* Get dataset: */
  if (IK_GetDataset(dskey, &dataset) != 0) {
    return err("couldn't get dataset (maybe it doesn't exist?)");
  }

  /* Call the callback: */
  result = (*action)(dskey, dataset, arg);
  free(dataset);
  return result;
}

/************************************************************
 * do_granule
 ************************************************************
 * Perform an action for a single granule.
 * 
 * The gkey specifies the granule.  The action is invoked as:
 *
 *     (*action)(gkey, granule, arg);
 */ 

static int do_granule(IK_SortKey *gkey,
		      int (*action)(), void *arg)
{
  IK_Granule *granule;
  int result;


  /* Get granule: */
  if (IK_GetGranule(gkey, &granule) != 0) {
    return err("couldn't get granule");
  }

  /* Call the callback: */
  result = (*action)(gkey, granule, arg);
  free(granule);
  return result;
}

/************************************************************
 *
 * dump_*
 *
 ************************************************************
 * The following dump out various kinds of entities.
 */

/************************************************************
 * dump_area_fields
 ************************************************************
 * Output info for an area, as *part* of a chunk.  All areas 
 * are output in the same format.
 *
 * If points are involved, the general idea is to output each point
 * as "LATITUDE LONGITUDE", where LATITUDE ranges from -90.0 to 90.0, 
 * and LONGITUDE from -180.0 to 180.0.
 */ 

void dump_area_fields(IK_Coverage *area) {

  /* Report type of area, and compute center point (if there is one): */  
  switch (area->type) {
  case IK_POINT:     /* report as...   Area: lat lon */
    {
      double centerLat, centerLon;

      /* Get coords: */
      centerLat = atof(area->area.point.latitude);
      centerLon = atof(area->area.point.longitude);

      /* Report area and center: */
      IW_ChunkspBeginGroup(COUT, "POINT_LOC");
      IW_ChunkspField(COUT, "LATITUDE",  "%.6f", centerLat);
      IW_ChunkspField(COUT, "LONGITUDE", "%.6f", centerLon);
      IW_ChunkspEndGroup(COUT, "POINT_LOC");
    }
    break;

  case IK_RANGE:    /* report as...   Area: Nlat Wlon Slat Elon */
    {
      double northLat, southLat, eastLon, westLon;
      double centerLat, centerLon;

      /* Get boundaries: */
      northLat = atof(area->area.range.north_lat);
      southLat = atof(area->area.range.south_lat);
      eastLon  = atof(area->area.range.east_long);
      westLon  = atof(area->area.range.west_long);

      /* Compute center: */
      centerLat = (northLat + southLat) / 2.0;
      centerLon = (eastLon + westLon) / 2.0;

      /* Report area and center: */
      IW_ChunkspBeginGroup(COUT, "RANGE_LOC");
      IW_ChunkspField(COUT, "NORTH_LATITUDE", "%.6f", northLat);
      IW_ChunkspField(COUT, "SOUTH_LATITUDE", "%.6f", southLat);
      IW_ChunkspField(COUT, "EAST_LONGITUDE", "%.6f", eastLon);
      IW_ChunkspField(COUT, "WEST_LONGITUDE", "%.6f", westLon);
      IW_ChunkspEndGroup(COUT, "RANGE_LOC");
    }
    break;

  case IK_POLYGON:  /* report as...   Area: Y/N lat1 lon1 .. lat4 lon4  */
    {
      int i;
      char pole_inside;
      double lat, lon;
      double centerLat, centerLon;

      /* Report area type: */
      IW_ChunkspBeginGroup(COUT, "POLYGON_LOC");
      
      /* Pole inside: */
      IW_ChunkspField(COUT, "POLE_INSIDE", "%c", 
		      (pole_inside ? pole_inside : 'X'));


#if 0
      /******* UNFINISHED *****/
      /* Coordinates: */
      pole_inside = area->area.polygon.pole_inside;
      printf("Area: %c ", 
      for (i = 0; i < IK_NPOLYSIDES; i++) {
	lat = atof(area->area.polygon.latitude[i]);
	lon = atof(area->area.polygon.longitude[i]);
	printf("%.2f %.2f ", lat, lon);
      }
      printf("\n");
#endif

      /* Report center: */
      centerLat = atof(area->area.polygon.centroid_lat);
      centerLon = atof(area->area.polygon.centroid_long);
      IW_ChunkspField(COUT, "CENTROID_LAT", "%.6f", centerLat);
      IW_ChunkspField(COUT, "CENTROID_LON", "%.6f", centerLon);
      
      IW_ChunkspEndGroup(COUT, "POLYGON_LOC");
    }
    break;
      
  case IK_GLOBAL:
    {
      /* Report area: */
      IW_ChunkspField(COUT, "GLOBAL_GRANULE", "%s", "Y");
    }
    break;
    
  default:  /* none, or one we don't handle */    
    /* no-op */
    0;
  }
}

/************************************************************
 * dump_dataset
 ************************************************************
 * Output info for a single dataset, as an RFC-822-style chunk.
 */ 

static int dump_dataset(IK_SortKey *key,
			IK_Dataset *dataset,
			void *arg  /* generic callback arg... */
			)
{
  int count;
  int i;
  char keybuf[64];
  char archid[64];
  char dsid[IK_DATASET_ID_SZ+1];
  char *str;
  char **strp;
  IWM_Timespan *timespan;

  /*
   * GET SOME BASIC INFO, ENOUGH FOR THE ONE-LINER... 
   */

  /* Key (WWW format): */
  IW_DatasetKeyPartsToWeb(dataset->archive_tag, dataset->dataset_tag, keybuf);

  /* Archive name: */
  archid[0] = '\0';
  if (str = IK_MapArchiveTagToName(dataset->archive_tag)) {
    strcpy(archid, str);
  }

  /* Dataset id: */
  dsid[0] = '\0';
  if (str = IK_MapDatasetTagToName(dataset->dataset_tag)) {
    strcpy(dsid, str);
    IW_Strupcase(dsid);    /* covert to uppercase */
  }

  /* If we just want a one-liner, output it and return: */
  if (Detail == ONELINE) {
    fprintf(stdout, "%-8s %-8s %s %4d\n", keybuf, 
	    archid, dsid, dataset->ngranules);
    return 0;
  }

  /*
   * OUTPUT CHUNK (SUMMARY INFO FIRST, DETAILED INFO NEXT)...
   */

  /*
   * Start with the summary info:
   */

  /* Type: */
  IW_ChunkspBegin(COUT, "DATASET");

  /* Key: */
  IW_ChunkspField(COUT, "DATASET_WEB_KEY", "%s", keybuf);
  
  /* Archive name: */
  IW_ChunkspField(COUT, "DATA_CENTER_ID", "%s", archid);
  
  /* Dataset name: */
  if (dsid[0]) {
    IW_ChunkspField(COUT, "DATASET_ID", "%s", dsid);
  }
  else {
    IW_ChunkspField(COUT, "DATASET_ID", "%s (name unavailable)", keybuf);
  }

  /* Num granules: */
  IW_ChunkspField(COUT, "NUMBER_OF_GRANULES", "%d", dataset->ngranules);

  /*
   * Everything below this point is DETAILED info; we can skip it if
   * we just want SUMMARY info:
   */
  if (Detail == SUMMARY) {
    IW_ChunkspEnd(COUT, "DATASET");
    return 0;
  }
    
  /* Timespan: */
  if (timespan = (IWM_Timespan *)IW_HashtAt(GTimespans, keybuf)) {
    IW_ChunkspField(COUT, "START_DATE", "%s", timespan->start);
    IW_ChunkspField(COUT, "STOP_DATE",  "%s", timespan->stop);
  }

  /* Area: */
  dump_area_fields(&(dataset->area));

  /* MD entry name: */
  if (dataset->md_entry[0]) {
    IW_ChunkspField(COUT, "MD_ENTRY_ID", "%s", dataset->md_entry);    
  }
  
  /* The following code was added to print account information */
  if (dataset->ValAcctCount > 0) {
    for (i = 0; (i < dataset->ValAcctCount) && (i < VALID_ACCOUNTS_MAX_NUM);
	 i++){
      IW_ChunkspBeginGroup(COUT, "VALID_ACCOUNTS");
      if (dataset->val_accts[i].accountNumber[0] != '\0') {
	IW_ChunkspField(COUT, "ACCOUNT_NUMBER", "%s", 
		      dataset->val_accts[i].accountNumber);
      }
      if (dataset->val_accts[i].accountBalance[0] != '\0') {
        IW_ChunkspField(COUT, "BALANCE", "%s", 
		      dataset->val_accts[i].accountBalance);
      }
      if (dataset->val_accts[i].errorMsg) {
        IW_ChunkspField(COUT, "ERROR", "%s", 
		       IW_ChunkUnfold(dataset->val_accts[i].errorMsg));
      }
      IW_ChunkspEndGroup(COUT, "VALID_ACCOUNTS");
    }
  }

  /* Originating center: */
  if (dataset->org_center[0]) {
    IW_ChunkspField(COUT, "ORG_CENTER", "%s", dataset->org_center);    
  }

  /* Restrictions: */
  if (dataset->restriction) {
    IW_ChunkspField(COUT, "RESTRICTION", "%s", 
		  IW_ChunkUnfold(dataset->restriction));
  }
    
  /* Comment: */
  if (dataset->comment) {
    IW_ChunkspField(COUT, "COMMENT", "%s", IW_ChunkUnfold(dataset->comment));
  }

  /* Browse description: */
  if (dataset->browse_desc) {
    IW_ChunkspField(COUT, "BROWSE_PRODUCT_DESCRIPTION", "%s", 
		  IW_ChunkUnfold(dataset->browse_desc));
  }
  
  /* MD info: */
  if (dataset->md_start_time) {
    IW_ChunkspField(COUT, "MD_START_TIME", "%s", dataset->md_start_time);    
  }
  if (dataset->md_stop_time) {
    IW_ChunkspField(COUT, "MD_STOP_TIME", "%s", dataset->md_stop_time);    
  }
  if (dataset->md_title) {
    IW_ChunkspField(COUT, "MD_TITLE", "%s", dataset->md_title);    
  }
  if (dataset->md_attribute) {
    IW_ChunkspField(COUT, "MD_ATTRIBUTE", "%s", dataset->md_attribute);    
  }
  if (dataset->md_reference) {
    IW_ChunkspField(COUT, "MD_REFERENCE", "%s", dataset->md_reference);    
  }
  if (dataset->md_personnel) {
    IW_ChunkspField(COUT, "MD_PERSONNEL", "%s", dataset->md_personnel);
  }
  if (dataset->md_archive_info) {
    IW_ChunkspField(COUT, "MD_ARCHIVE_INFO", "%s", dataset->md_archive_info); 
  }
  

  /* Done! */
  IW_ChunkspEnd(COUT, "DATASET");
  return 0;
}

/************************************************************
 * dump_datasets
 ************************************************************
 * Output info for all datasets returned from the last search.
 */ 

static int dump_datasets(int which)       /* dump ALL, or only ones in LIST? */
{
  IK_SortKey *key;
  IK_Dataset *dataset;
  int count;
  IK_Dataset **datasets;
  int di, dn;  
  char keybuf[64];
  IWM_Timespan *timespan;
  char web_dskey[256];
  int (*is_good_gran)();

  debug("Doing granules");

  /* Calculate the timespan for each dataset: */
  is_good_gran = ((which == DUMP_LIST) ? good_if_in_list : NULL);
  if (do_granules(is_good_gran, NULL, 
		  calc_dataset_timespans, NULL) != 0) {
    return err("do_granules");
  }

  debug("Dumping dataset objects");

  /* Dump the dataset objects: */
  key = IK_FirstDatasetKey();
  while (key) {
    char *str;
    char **strp;

    /* If LIST ONLY, is this key in the list? */
    if (which == DUMP_LIST) {
      IW_DatasetKeyToWeb(key, web_dskey);
      if (!(IW_HashtAt(ListHash, web_dskey))) {  /* not on the list! */
	goto NEXT_KEY;
      }
    }

    /* Get dataset: */
    if (IK_GetDataset(key, &dataset) != 0) {
      free(key);
      return err("couldn't get dataset");
    }
   
    /* Dump it: */
    if (dump_dataset(key, dataset, NULL) != 0) {
      free(key);
      free(dataset);
      return err("couldn't dump dataset");
    }
    free(dataset);

    /* Next: */
  NEXT_KEY:
    free(key);
    key = IK_NextDatasetKey();
  }
  
  return 0;
}

/************************************************************
 * dump_granule
 ************************************************************
 * Dump a granule as an RFC-822-tyle chunk.
 */ 

static int dump_granule(IK_SortKey *key,
			IK_Granule *granule,
			void *arg    /* generic callback arg */
			)
{
  char *str;
  char *pid;
  char **strp;
  char keybuf[256];
  IK_Dataset *dataset;
  IK_SortKey *dataset_key;
  extern double atof();
  char archid[64];
  char dsid[IK_DATASET_ID_SZ+1];

  /*
   * GET SOME BASIC INFO, ENOUGH FOR THE ONE-LINER... 
   */

  /* Key (WWW format): */
  IW_GranuleKeyPartsToWeb(granule->archive_tag, granule->dataset_tag, 
			  granule->granule_id, keybuf);

  /* Archive name: */
  archid[0] = '\0';
  if (str = IK_MapArchiveTagToName(granule->archive_tag)) {
    strcpy(archid, str);
  }

  /* Dataset id: */
  dsid[0] = '\0';
  if (str = IK_MapDatasetTagToName(granule->dataset_tag)) {
    strcpy(dsid, str);
    IW_Strupcase(dsid);    /* covert to uppercase */
  }
  
  /* If we just want a one-liner, output it and return: */
  if (Detail == ONELINE) {
    fprintf(stdout, "%-8s %-8s %s %s\n", 
	    keybuf, archid, dsid, granule->granule_id);
    return 0;
  }

  /*
   * OUTPUT CHUNK (SUMMARY INFO FIRST, DETAILED INFO NEXT)...
   */

  /*
   * Start with the summary info...
   */

  /* Type: */
  IW_ChunkspBegin(COUT, "GRANULE");

  /* Key: */
  IW_ChunkspField(COUT, "GRANULE_WEB_KEY", "%s", keybuf);
  
  /* Archive name: */
  IW_ChunkspField(COUT, "DATA_CENTER_ID", "%s", archid);
  
  /* Dataset name (this might be unavailable if valids are lagging): */
  IW_ChunkspField(COUT, "DATASET_ID", "%s", (dsid[0] ? dsid : keybuf));
  
  /* Granule id: */
  IW_ChunkspField(COUT, "GRANULE_ID", "%s", granule->granule_id);   

  /*
   * Everything below this point is DETAILED info; we can skip it if
   * we just want SUMMARY info:
   */
  if (Detail == SUMMARY) {
    IW_ChunkspEnd(COUT, "GRANULE");
    return 0;
  }
  
  /* Start and stop time: */
  if (granule->start_date) {
    IW_ChunkspField(COUT, "START_DATE", "%s", granule->start_date);    
  }
  if (granule->stop_date) {
    IW_ChunkspField(COUT, "STOP_DATE", "%s", granule->stop_date);    
  }
  
  /* Browse type: */
  switch (granule->browse_type) {
  case IK_INTEGRATED:    /* means BOTH integrated & FTP... sheesh. */
    IW_ChunkspField(COUT, "BROWSE_TYPE", "Y"); 
    break;
  case IK_FTP:           /* means FTP only */
    IW_ChunkspField(COUT, "BROWSE_TYPE", "FTP_ONLY"); 
    break;
  case IK_NO_BROWSE:     /* means no browse images */
  default:            
    IW_ChunkspField(COUT, "BROWSE_TYPE", "N");
  }

  /* Package ids (needed for ordering capability): */
  if (granule->package_id) {
    char **pkid_p;

    for (pkid_p = granule->package_id; *pkid_p; pkid_p++) {
      AGGREGATE root;
      IW_ChunkspField(COUT, "PACKAGE_ID", "%s", *pkid_p);      

      pid = (IW_Strequal(*pkid_p, "*") ? dsid : *pkid_p);

      /* Create the root node: */
      root = NewAggregate(NULL, KA_GROUP, "PACKAGE", NULL);

      /* Get the package.  If we can't, just continue on...: */
      if (IK_GetPackage(granule->archive_tag,  granule->dataset_tag,
                      pid, &root) >= 0) {
        IW_ChunkspField(COUT, "VALID_PACKAGE_ID", "%s", *pkid_p);
      }
      RemoveAggregate(root);
    }    
  }
  /* Sources: */
  if (granule->source_name) {
    for (strp = granule->source_name; *strp; strp++) {
      IW_ChunkspField(COUT, "SOURCE_NAME", "%s", *strp);      
    }    
  }
  /* Sensors: */
  if (granule->sensor_name) {
    for (strp = granule->sensor_name; *strp; strp++) {
      IW_ChunkspField(COUT, "SENSOR_NAME", "%s", *strp);      
    }
  }    

  /* Area: */
  dump_area_fields(&(granule->area));

  /* 
   * Dataset-specific info: 
   */

  /* Build key: */
  if (!(dataset_key = IK_BuildDatasetKey(granule->archive_tag, 
					 granule->dataset_tag))) {
    return err("couldn't build dataset key");
  }  

  /* Get dataset: */
  if (IK_GetDataset(dataset_key, &dataset) != 0) {
    free(dataset_key);
    return err("couldn't get dataset");
  }
  free(dataset_key);

  /* MD entry name (needed for brief info): */
  if (dataset->md_entry[0]) {
    IW_ChunkspField(COUT, "MD_ENTRY_ID", "%s", dataset->md_entry);    
  }

  free(dataset);

  /* Day/night flag: */
  switch (granule->day_night) {
  case IK_DAY:   IW_ChunkspField(COUT, "DAY_NIGHT", "D"); break;
  case IK_NIGHT: IW_ChunkspField(COUT, "DAY_NIGHT", "N"); break;
  case IK_BOTH: 
  default:       IW_ChunkspField(COUT, "DAY_NIGHT", "BOTH");
  }

  /* Processing level: */
  if (granule->processing_level[0]) {
    IW_ChunkspField(COUT, "PROCESSING_LEVEL", "%s", granule->processing_level);
  }

  /* Image id: */
  if (granule->image_id[0]) {
    IW_ChunkspField(COUT, "IMAGE_ID", "%s", granule->image_id);    
  }

  /* Campaign: */
  if (granule->campaign[0]) {
    IW_ChunkspField(COUT, "CAMPAIGN", "%s", granule->campaign);
  }

  /* Comment: */
  if (granule->comment) {
    IW_ChunkspField(COUT, "COMMENT", "%s", IW_ChunkUnfold(granule->comment));
  }

  /* Browse description: */
  if (granule->browse_desc) {
    IW_ChunkspField(COUT, "BROWSE_PRODUCT_DESCRIPTION", "%s", 
		  IW_ChunkUnfold(granule->browse_desc));
  }
  
  /* Parameters: */
  if (granule->parameter) {
    for (strp = granule->parameter; *strp; strp++) {
      IW_ChunkspField(COUT, "PARAMETER", "%s", *strp);
    }    
  }

  /* Done! */
  IW_ChunkspEnd(COUT, "GRANULE");
  return 0;
}

/************************************************************
 * dump_package_tree
 ************************************************************
 *
 * DESCRIPTION
 *    Dump the chunk structure of a package, as a tree.
 */

static int dump_package_tree(AGGREGATE root)
{  
  IW_ODLWriteChunk(COUT, root, NULL);
  return 0;
}

/************************************************************
 * dump_package
 ************************************************************
 *
 * DESCRIPTION
 *    Dump all combinations of options for a package.  
 *    Output looks like this:
 *
 *        --
 *        X-Type: PACKAGE
 *        PACKAGE_ID: XXXYYYYZZZZ
 *        OPTION_ID: Level 1
 *        TYPE_ID: CD-ROM
 *        FORMAT_ID: XVS44
 *        APPROX_COST: 10
 *        BEGIN URL
 *           URL: ...
 *           URL_COMMENT: ...
 *        END URL 
 *        BEGIN URL
 *           URL: ...
 *           URL_COMMENT: ...
 *        END URL
 *
 *        --
 *        X-Type: PACKAGE
 *        PACKAGE_ID: XXXYYYYZZZZ
 *        OPTION_ID: Level 1
 *        TYPE_ID: CD-ROM
 *        FORMAT_ID: XVS88
 *        ...
 */

static int dump_package(AGGREGATE root)
{
  AGGREGATE package;
  AGGREGATE procOpt;
  AGGREGATE mediaType;
  AGGREGATE mediaFmt;
  AGGREGATE miscURL;
  PARAMETER param;
  VALUE value;  
  char *infoPrompt_str = NULL;
  char *comment_str = NULL;
  char *url = NULL;
  char *urlComment = NULL;
  char packageId[256];    /* package id */
  char procOptId[256];    /* processing option id */
  char mediaTypeId[256];  /* media type id */
  char mediaFmtId[256];   /* media format id */
  double approxCost; 
  long numGrans;

  /* Get package group: */
  if (!(package = NextSubGroup(root, root))) {
    return err("empty package!");
  }
  /* Get package id: */
  if (!(value = FirstValue(FindParameter(package, "PACKAGE_ID")))) {    
    return err("no PACKAGE_ID!");
  }
  strcpy(packageId, value->item.value.string);  

  /* Get info prompt: */
  if (param = FindParameter(package, "INFO_PROMPT")) {
    infoPrompt_str = IW_ODLParamStringified(param, NULL);
  }
  /* Get comment: */
  if (param = FindParameter(package, "COMMENT")) {
    comment_str = IW_ODLParamStringified(param, NULL);
  }
  /* Get Miscellaneous URL if present. */
  if (miscURL = FindGroup(package, "MISC_URL")) {
    if (param = FindParameter(miscURL, "URL")) {
      url = IW_ODLParamStringified(param, NULL);
    }
    if (param = FindParameter(miscURL, "URL_COMMENT")) {
      urlComment = IW_ODLParamStringified(param, NULL);
    }
  }
  /* Get number of granules: */
  numGrans = 0;
  if (value = FirstValue(FindParameter(package, "NUMBER_OF_GRANULES"))) {
    numGrans = value->item.value.integer.number;
  }
  /* For each processing option... */
  for (procOpt = FindGroup    (package,          "PROCESSING_OPTION");
       procOpt;
       procOpt = FindNextGroup(package, procOpt, "PROCESSING_OPTION")) {
    /* Get option id: */
    value = FirstValue(FindParameter(procOpt, "OPTION_ID"));
    strcpy(procOptId, value->item.value.string);
    
    /* For each media type... */
    for (mediaType = FindGroup    (procOpt,            "MEDIA_TYPE");
	 mediaType;
	 mediaType = FindNextGroup(procOpt, mediaType, "MEDIA_TYPE")) {
      /* Get media type id: */
      value = FirstValue(FindParameter(mediaType, "TYPE_ID"));
      strcpy(mediaTypeId, value->item.value.string);

      /* For each media format... */
      for (mediaFmt = FindGroup    (mediaType,           "MEDIA_FORMAT");
	   mediaFmt;
	   mediaFmt = FindNextGroup(mediaType, mediaFmt, "MEDIA_FORMAT")) {
	int have_cost = 0;

	/* Get media format id: */ 
	value = FirstValue(FindParameter(mediaFmt, "FORMAT_ID"));
	strcpy(mediaFmtId, value->item.value.string);
	
	/* Get approximate cost: */ 
	if (value = FirstValue(FindParameter(mediaFmt, "APPROX_COST"))) {
	  approxCost = value->item.value.real.number;
	  have_cost = 1;
	}

	/* OUTPUT! Output as either one-liner or chunk: */	
	if (Detail == ONELINE) {          /* ONELINE... */
	  fprintf(stdout, "%s %s %s %s %02lf\n",
		  packageId, procOptId, mediaTypeId, mediaFmtId,
		  (have_cost ? approxCost : 0));
	}
	else {                            /* SUMMARY/DETAILED... */

	  /* Start the chunk: */
	  IW_ChunkspBegin(COUT, "PACKAGE");

	  /* Output the "inherited" attributes: */
	  IW_ChunkspField(COUT, "PACKAGE_ID",  "%s",  packageId);
	  IW_ChunkspField(COUT, "OPTION_ID",   "%s",  procOptId);
	  IW_ChunkspField(COUT, "TYPE_ID",     "%s",  mediaTypeId); 
	  if (infoPrompt_str) {
	    IW_ChunkspField(COUT, "INFO_PROMPT", "%s",  infoPrompt_str);
	  }
	  if (comment_str) {
	    IW_ChunkspField(COUT, "COMMENT",     "%s",  comment_str);
	  }
	  if (miscURL) {
	    IW_ChunkspBeginGroup(COUT, "MISC_URL");
	    IW_ChunkspField(COUT, "URL",  "%s", url);
	    IW_ChunkspField(COUT, "URL_COMMENT", "%s", urlComment);
	    IW_ChunkspEndGroup(COUT, "MISC_URL");
	  }
	  IW_ChunkspField(COUT, "NUMBER_OF_GRANULES", "%ld", numGrans);

	  /* Output all the info in the MEDIA_FORMAT group... */
          IW_ODLWriteChunkFields(COUT, mediaFmt, NULL);

	  /* End the chunk: */
	  IW_ChunkspEnd(COUT, "PACKAGE");
	}	

      } /* next media format */      
    } /* next media type */
  } /* next processing option */

  /* Cleanup: */
CLEANUP:
  IW_Free(comment_str);
  IW_Free(infoPrompt_str);
  return 0;
}

/************************************************************
 * read_list_file
 ************************************************************
 * Parse in the list file of datasets/granules into a hash.
 * Sets up the global ListHash.
 */

static int read_list_file(char *path)
{
    FILE *list_fp;
    char line[256];
    char key[256];
    char type[64];

    /* Create the hash: */
    ListHash = IW_HashtAllocCS(211, NULL);

    /* Open the file: */
    if (!(list_fp = fopen(path, "r"))) {
      return err("couldn't open list file: %s", path);
    }
    
    /* Parse it: */
    while (fgets(line, 255, list_fp)) {
      if (sscanf(line, "%s %s", type, key) == 2) {
	debug("Adding list item %s '%s'", type, key);
	IW_HashtAtPut(ListHash, key, (void *)1);
      }
    }
    fclose(list_fp);
    return(0);
}

/************************************************************
 *
 * cmd_*
 *
 ************************************************************
 * The following handle the top-level commands.
 */

/************************************************************
 * cmd_d -- output the given [D]ataset
 ************************************************************/
  
static int cmd_d (char *web_dskey) 
{  
  IK_SortKey *dskey;
  char test[100];

  /* Extract dataset key from web representation: */
  if (!web_dskey || !*web_dskey) puke("missing web key!");
  dskey = IW_DatasetKeyFromWeb(web_dskey);
  if (do_dataset(dskey, dump_dataset, NULL) != 0) {
    puke("couldn't do dataset %s", web_dskey);
  }

  /* Done! */
  free(dskey);
  return 0;
}  

/************************************************************
 * cmd_ds -- output summary of all [D]atasets
 ************************************************************/

static int cmd_ds ()
{  
  debug("Dumping datasets");
  return dump_datasets(DUMP_ALL);
}  

/************************************************************
 * cmd_g -- output the given [G]ranule
 ************************************************************/

static int cmd_g (char *web_gkey) 
{  
  IK_SortKey *gkey;

  /* Extract parts of granule key: */
  if (!web_gkey || !*web_gkey) puke("missing web key!");
  gkey = IW_GranuleKeyFromWeb(web_gkey);

  /* Dump the granule: */
  if (do_granule(gkey, dump_granule, NULL) != 0) {
    puke("couldn't do granule %s", web_gkey);
  }  

  /* Done! */
  free(gkey);
  return 0;
}  

/************************************************************
 * cmd_gd -- output all [G]ranules in the given [D]ataset
 ************************************************************/

static int cmd_gd (char *web_dskey) 
{  
  if (!web_dskey || !*web_dskey) {    /* LIST ALL GRANULES */
    if (do_granules(NULL, NULL, dump_granule, NULL) != 0) {
      puke("couldn't do granules in %s", web_dskey);
    }
  }
  else {                              /* LIST GRANULES FROM SINGLE DATASET */
    IK_SortKey *dskey;

    /* Create sort key from web key: */
    dskey = IW_DatasetKeyFromWeb(web_dskey);
    if (do_granules(good_if_match_one_dataset, (void *)dskey,
		    dump_granule, NULL) != 0) {
      puke("couldn't do granules in %s", web_dskey);
    }
  } 
  return 0;
}  

/************************************************************
 * cmd_ps -- output packages
 ************************************************************
 *
 * DESCRIPTION
 *    Given a granule key, output package info for all packages 
 *    in that granule.
 */ 

static int cmd_ps(char *web_gkey)
{
  IK_ArchiveTag archTag;
  IK_DatasetTag dsTag;
  char gid[IK_GRANULE_ID_SZ+1];
  IK_SortKey *key;
  IK_Granule *granule;
  char *pkid;
  char **pkids;
  char x[2560];
  char *str;
  char dsid[IK_DATASET_ID_SZ+1];


  /* Convert gkey string to an actual sort key, and get the parts: */
  if (!web_gkey || !*web_gkey) puke("missing web key!");
  debug("key = %s", web_gkey);
  key = IW_GranuleKeyFromWeb(web_gkey);
  IK_StripGranuleKey(archTag, dsTag, gid, key);  

  /* Get dataset id, and upcase it: */
  if (!(str = IK_MapDatasetTagToName(dsTag)) || !*str) {
    return err("couldn't get dataset id for granule %s", web_gkey);
  }
  strcpy(dsid, str);
  IW_Strupcase(dsid);
  debug("Got dataset id: '%s'", dsid);
  
  /* Get granule: */
  if (IK_GetGranule(key, &granule) != 0) {
    free(key);
    return err("couldn't get granule %s", web_gkey);
  }
  
  /* For each package in the granule: */
  for (pkids = granule->package_id; *pkids; pkids++) {
    AGGREGATE root;

    /* Translate '*' to dataset id: */
    pkid = (IW_Strequal(*pkids, "*") ? dsid : *pkids);

    /* Create the root node: */
    root = NewAggregate(NULL, KA_GROUP, "PACKAGE", NULL);

    /* Get the package.  If we can't, just continue on...: */
    info("package: %s", pkid);
    if (IK_GetPackage(granule->archive_tag,  granule->dataset_tag,
		      pkid, &root) < 0) {
      RemoveAggregate(root);
      err("couldn't get package %s for %s", pkid, web_gkey);
      continue;
    }
    
    /* Dump it out, and clean it up: */
    dump_package(root); 
    RemoveAggregate(root);
  }
 
  /* Done! */
  return 0;
}

/************************************************************
 * cmd_pt -- output package tree
 ************************************************************
 *
 * DESCRIPTION
 *    Given a granule key, output package info for all packages 
 *    in that granule, as a bunch of trees.
 */ 

static int cmd_pt(char *web_gkey)
{
  IK_ArchiveTag archTag;
  IK_DatasetTag dsTag;
  char gid[IK_GRANULE_ID_SZ+1];
  IK_SortKey *key;
  IK_Granule *granule;
  char *pkid;
  char **pkids;
  char x[2560];
  char *str;
  char dsid[IK_DATASET_ID_SZ+1];

  /* Convert gkey string to an actual sort key, and get the parts: */
  if (!web_gkey || !*web_gkey) puke("missing web key!");
  debug("key = %s", web_gkey);
  key = IW_GranuleKeyFromWeb(web_gkey);
  IK_StripGranuleKey(archTag, dsTag, gid, key);  

  /* Get dataset id, and upcase it: */
  if (!(str = IK_MapDatasetTagToName(dsTag)) || !*str) {
    return err("couldn't get dataset id for granule %s", web_gkey);
  }
  strcpy(dsid, str);
  IW_Strupcase(dsid);
  debug("Got dataset id: '%s'", dsid);
  
  /* Get granule: */
  if (IK_GetGranule(key, &granule) != 0) {
    free(key);
    return err("couldn't get granule %s", web_gkey);
  }
  
  /* For each package in the granule: */
  for (pkids = granule->package_id; *pkids; pkids++) {
    AGGREGATE root;

    /* Translate '*' to dataset id: */
    pkid = (IW_Strequal(*pkids, "*") ? dsid : *pkids);

    /* Create the root node: */
    root = NewAggregate(NULL, KA_GROUP, "PACKAGE", NULL);

    /* Get the package.  If we can't, just continue on...: */
    info("package: %s", pkid);
    if (IK_GetPackage(granule->archive_tag,  granule->dataset_tag,
		      pkid, &root) < 0) {
      RemoveAggregate(root);
      err("couldn't get package %s for %s", pkid, web_gkey);
      continue;
    }
    
    /* Dump it out, and clean it up: */
    dump_package_tree(root); 
    RemoveAggregate(root);
  }
 
  /* Done! */
  return 0;
}

/************************************************************
 * cmd_ptg -- output package tree
 ************************************************************
 *
 * DESCRIPTION
 *    Given a granule key, output package info for all packages 
 *    in that granule, as a bunch of trees.
 *    It reads the package ids from a file in the session.
 */ 

static int cmd_ptg(char *web_gkey)
{
  IK_ArchiveTag archTag;
  IK_DatasetTag dsTag;
  char gid[IK_GRANULE_ID_SZ+1];
  IK_SortKey *key;
  IK_Granule *granule;
  char *pkid;
  char **pkids;
/*  char x[2560]; */
  char *str;
  char dsid[IK_DATASET_ID_SZ+1];
  FILE *fp=NULL;      /* File pointer to file structure. */
  char lineBuffer[MAXLINE]; /* Buffer to read one line of filter names */
  int count=0;  /* number of package ids in the file */
  int i=0;  /* counter for pkids */
  int j=0;  /* counter for pkids */
  int len; /* length of the lineBuffer */

  /* Convert gkey string to an actual sort key, and get the parts: */
  if (!web_gkey || !*web_gkey) puke("missing web key!");
  debug("key = %s", web_gkey);
  key = IW_GranuleKeyFromWeb(web_gkey);
  IK_StripGranuleKey(archTag, dsTag, gid, key);  

  /* Get dataset id, and upcase it: */
  if (!(str = IK_MapDatasetTagToName(dsTag)) || !*str) {
    return err("couldn't get dataset id for granule %s", web_gkey);
  }
  strcpy(dsid, str);
  IW_Strupcase(dsid);
  debug("Got dataset id: '%s'", dsid);
  
  /* Get granule: */
  fp = fopen(package_ids_file, "r");
  while (fgets(lineBuffer,MAXLINE,fp)) {
    count++;
  }
  (void) fclose(fp);
  pkids = calloc(count+1, sizeof(char *));
  fp = fopen(package_ids_file, "r");
  while (fgets(lineBuffer,MAXLINE,fp)) {
    len = strlen(lineBuffer);
    pkids[i] = calloc(len, sizeof(char));
    (void) strncpy(pkids[i],lineBuffer,len-1); /* to remove the newline */
    pkids[i][len-1] = '\0';
    i++;
  }
  (void) fclose(fp);
  
  /* For each package in the granule: */
  for (j = 0; j < count; j++) {
    AGGREGATE root;

    /* Translate '*' to dataset id: */
    pkid = (IW_Strequal(pkids[j], "*") ? dsid : pkids[j]);

    /* Create the root node: */
    root = NewAggregate(NULL, KA_GROUP, "PACKAGE", NULL);

    /* Get the package.  If we can't, just continue on...: */
    info("package: %s", pkid);
    if (IK_GetPackage(archTag, dsTag, pkid, &root) < 0) {
      RemoveAggregate(root);
      err("couldn't get package %s for %s", pkid, web_gkey);
      continue;
    }
    /* Dump it out, and clean it up: */
    dump_package_tree(root); 
    RemoveAggregate(root);
  }
  /* Done! */
  return 0;
}

/************************************************************
 * cmd_ptd -- output package tree for data set ordering
 ************************************************************
 *
 * DESCRIPTION
 *    Given a data set key, output package info for package id pkid
 *    in that data set, as a bunch of trees.
 */ 

static int cmd_ptd(char *web_dskey, char *pkid)
{
  IK_ArchiveTag archTag;
  IK_DatasetTag dsTag;
  IK_SortKey *key;
  char *str;
  char dsid[IK_DATASET_ID_SZ+1];
  AGGREGATE root;

  /* Convert dskey string to an actual sort key, and get the parts: */
  if (!web_dskey || !*web_dskey) puke("missing web key!");
  debug("key = %s", web_dskey);
  key = IW_DatasetKeyFromWeb(web_dskey);
  IK_StripDatasetKey(archTag, dsTag, key);  

  /* Get dataset id, and upcase it: */
  if (!(str = IK_MapDatasetTagToName(dsTag)) || !*str) {
    return err("couldn't get dataset id for dataset %s", web_dskey);
  }
  strcpy(dsid, str);
  IW_Strupcase(dsid);
  debug("Got dataset id: '%s'", dsid);

  /* Translate '*' to dataset id: */
/*    pkid = (IW_Strequal(pkid, "*") ? dsid : pkid); */

  /* Create the root node: */
  root = NewAggregate(NULL, KA_GROUP, "PACKAGE", NULL);

  /* Get the package.  If we can't, just continue on...: */
  info("package: %s", pkid);
  if (IK_GetPackage(archTag, dsTag, pkid, &root) < 0) {
    RemoveAggregate(root);
    err("couldn't get package %s for %s", pkid, web_dskey);
    return -1;
  }
    
  /* Dump it out, and clean it up: */
  dump_package_tree(root); 
  RemoveAggregate(root);
 
  /* Done! */
  return 0;
}

/************************************************************
 * cmd_all -- output everything, in detail
 ************************************************************/ 

static int cmd_all()
{
  /* Dump all datasets... */
  if (dump_datasets(DUMP_ALL) < 0) {
    return err("couldn't dump datasets");
  }
  
  /* Dump all granules... */
  if (do_granules(NULL, NULL, dump_granule, NULL) < 0) {
    return err("couldn't dump granules");
  }
  return 0;  
}

/************************************************************
 * cmd_list -- output last search results, in detail
 ************************************************************/ 

static int cmd_list(char *list_file)
{
  /* Suck in the keys from the last search, and hash them: */
  debug("Reading list file");
  read_list_file(list_file);
 
  /* Dump all datasets in the list... */
  debug("Dumping datasets");
  if (dump_datasets(DUMP_LIST) < 0) {
    return err("couldn't dump datasets");
  }
  
  /* Dump all granules in the list... */
  debug("Dumping granules");
  if (do_granules(good_if_in_list, NULL, dump_granule, NULL) < 0) {
    return err("couldn't dump granules");
  }
  /* Done! */
  return 0;  
}

/************************************************************
 * usage
 ************************************************************
 * Output usage info to the stderr.
 */

static void usage() 
{
  fprintf(stderr, "\nUsage:\n");
  fprintf(stderr, "    imsgdbdump [opts] <action [args...]>\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "Options are:\n");
  fprintf(stderr, "    -f<format>   Output format; one of...\n");
  fprintf(stderr, "                     d   detailed chunks (default)\n");
  fprintf(stderr, "                     s   summary chunks (shorter)\n");
  fprintf(stderr, "                     1   one-liners (very short)\n");
  fprintf(stderr, "    -s           Same as -fs\n");
  fprintf(stderr, "    -v           Verbose; debugs go to stderr\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "Supported actions are:\n");
  fprintf(stderr, "    all          Dump *all* objects\n");
  fprintf(stderr, "    list         Dump all objects in list file\n");
  fprintf(stderr, "    ps <gkey>    Dump all packages for this granule\n");
  fprintf(stderr, "    pt <gkey>    Same as `ps', but in `tree' format\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "Unsupported actions are:\n");
  fprintf(stderr, "    ds           Dump *all* datasets\n");
  fprintf(stderr, "    d <dskey>    Dump this dataset\n");
  fprintf(stderr, "    gd <dskey>   Dump granules in dataset\n");
  fprintf(stderr, "    g <gkey>     Dump this granule\n");
  fprintf(stderr, "\n");
}

/************************************************************
 * get_opts
 ************************************************************
 * Get command line options, and set global variables from them.
 * Returns an argument pointer pointing at the first non-option.
 *
 * The argument vector argv is expected to still be pointing to the
 * 0th arg (the program name).
 */

char **get_opts(int argc, char **argv) {
  char **argp;
  char *arg;
  int barecount = 0;
  char buf[256];
  int action;    

  /* Parse: */
  for (argp = argv+1; *argp; argp++) {
    arg = *argp;

  RETRY_ARG:
    if (*arg == '-') {     /* option... */
      switch (arg[1]) {  

      case 'f':                  /* output format */
	switch (arg[2]) {
	case 'd':  Detail = DETAILED; break;	  
	case 's':  Detail = SUMMARY;  break;	  
	case '1':  Detail = ONELINE;  break;
	case 'l':  Detail = ONELINE;  break;
	default:   usage(); puke("bad option: %s", arg);
	}
	break;

      case 's':                  /* summarize output (deprecated) */
	Detail = SUMMARY;
	break;

      case 'v':                  /* verbose */
	Verbose = 1;
	break;
	
      default:
	usage(); puke("bad option: %s", arg);
      }
    }

    else {         /* bareword: */
      return argp;
    }
  }
}

/************************************************************
 * main
 ************************************************************
 * Main entry point.
 */ 

int main(int argc, char **argv)
{
  char *gaeaTmpDir;
  char *newGaeaTmpDir;
  char syslog_file[512];
  char dynamic_file[512];  /* name of dynamic valids file to open */
  char list_file[512];
  char granule_gdb[512];
  char dataset_gdb[512];
  char **rest;
  static char *required_env_vars[] = {
    "GAEADATA_DIR",
    "GAEATMP_DIR",
    NULL,
  };
  char **sp;
  int result;

  /* The following added for dynamic valids */
  AGGREGATE dynamic_root; /* root AGGREGATE of the dynamic valids file */
  IK_BITVECTOR *fieldValidsVector = NULL; /* vector of selected valids */
  int numHits; /* number of valids selected by IK_FindValids */
  int found; /* return code for IK_FindValids */
  int dyIndex; /* the index of the new dynamic valid */
  char **fields; /* requested valids of a field */
  int j; /* counter for valids found in the input odl */
  int k; /* counter for fields */
  int i; /* counter */

  newGaeaTmpDir = (char *) calloc(1024, sizeof(char)); 
  /******************************
   * PREFLIGHT...
   ******************************/

  /* Output usage right away if no args: */
  if (!argv[1]) {
    usage(); 
    exit(-1);
  }

  /* Load options ASAP (so we can start debugging output, etc.): */
  rest = get_opts(argc, argv);
  /* Check environment: */
  for (sp = required_env_vars; *sp; sp++) {
    if (!getenv(*sp)) puke("please set %s in your environment", *sp);
  }

  /******************************
   * IK LIBRARY INIT...
   ******************************/

  (void) strcpy(newGaeaTmpDir, getenv(required_env_vars[1]));
  if (IW_Strequal(rest[0], "ptd") && argc == 5 && rest[3]) {
    (void) strcat(newGaeaTmpDir, "/");
    (void) strcat(newGaeaTmpDir, rest[3]);
    IW_Putenv(required_env_vars[1], newGaeaTmpDir);
  }
  if (IW_Strequal(rest[0], "ptg") && argc == 4 && rest[2]) {
    (void) strcat(newGaeaTmpDir, "/");
    (void) strcat(newGaeaTmpDir, rest[2]);
    IW_Putenv(required_env_vars[1], newGaeaTmpDir);
  }
  /* Setup the temp directory (and patch an IK bug that ends it with "//"): */
  gaeaTmpDir = IK_SysConfig("GaeaTmp");
  if (strstr(gaeaTmpDir, "//")) {
    gaeaTmpDir[strlen(gaeaTmpDir)-1] = '\0';     /* patch IK bug */
  }
  debug("GaeaTmpDir = <%s>", gaeaTmpDir);

  /* Setup files we need (many of them in the temp directory): */
  sprintf(dataset_gdb, "%sdataset.gdb",    gaeaTmpDir);
  sprintf(granule_gdb, "%sgranule.gdb",    gaeaTmpDir);
  sprintf(list_file,   "%sclient.list",    gaeaTmpDir);
  sprintf(syslog_file, "%sgdbdump.syslog", gaeaTmpDir);
  sprintf(dynamic_file, "%s%s", gaeaTmpDir, DYNAMIC_FILE);
  sprintf(package_ids_file, "%s%s", gaeaTmpDir, PACKAGE_IDS);

  /* The following call was added for the new dependent valids code -Solomon */
  if (IK_ReadDepValids() == -1) {
    IK_ImsErrno = IK_EIO;
    warning("Unable to read Valids files. See support staff.");
  } /* end of new dependent valids code */
  /* The following line was added because IK_ReadDepValids reset the GaeaTmp */
  gaeaTmpDir = IK_SysConfig("GaeaTmp");
  if (strstr(gaeaTmpDir, "//")) {
    gaeaTmpDir[strlen(gaeaTmpDir)-1] = '\0';     /* patch IK bug */
  }

  /*  Call to IK_DirInit was here and it is moved below */
  /*  Call to IK_GranuleInit was here and it is moved below */

  /******************************
   * IW LIBRARY INIT...
   ******************************/

  /* Set fail handler, and init library: */
  IW_SetDieHandler(dying_gasp);
  if (IW_Init() < 0) {                /* read in SETUP file, etc. */
    puke("couldn't init IW library");
  }

  /* Set up other data structures: */
  if (!(GTimespans = IW_HashtAllocCS(7, (void (*)())IWM_TimespanFree))) {
    puke("couldn't allocate hashtable");
  }
  if (!(COUT = IW_ChunkspNew(stdout))) {
    puke("no chunk output stream!");
  }

  /******************************
   * APPLICATION INIT...
   ******************************/

  /* Start logging: */
  if (IK_NameSyslog(syslog_file) == -1) {
    puke("couldn't open syslog");
  }

  if (!IW_Strequal(rest[0], "ptd") && !IW_Strequal(rest[0], "ptg")) {
  /* Open databases: */
    if (IK_DirInit(gaeaTmpDir) != 0) {
      puke("IK_DirInit(%s): %d\t", gaeaTmpDir, IK_ImsErrno);
    }
    if (IK_GranuleInit(granule_gdb, 0) != 0) {
      puke("IK_GranuleInit: %d\t", IK_ImsErrno);
    }
  }
  /* Set the logging level: */
  IW_LogSetVerbosity(Verbose ? LOG_DEBUG : LOG_WARNING);
  info("Logging has begun");

  /* Create the root to hold the dynamic valids file */
  dynamic_root = NewAggregate(NULL,KA_GROUP,"ROOT",NULL);

/* This code is added for dynamic valids */
  /* Read in the label: */
  if ((IW_ODLReadFile(dynamic_file, dynamic_root))) {
    ValidsCount = IK_ValidsCnt();
    fields = (char **)calloc(ValidsCount, sizeof (char *));
    fieldValidsVector = IK_callocBV();       /* allocates bit vector */
 
    for (k=0; k < MAX_CORE_FIELD; k++) {
      j = 0;
      if (handle_value (&fields, dynamic_root->first_child, &j, fieldname[k])){
        for (i=0; i < j; i++) {
	  for (ptr = fields[i];*ptr;ptr++)
	   *ptr = toupper (*ptr);
          fprintf (stderr, "item in fields %s \n", fields[i]);
          found = IK_FindValids(fields[i],k,fieldValidsVector,&numHits);
          if(found < 0) {
            if(IK_dynamicSlotAvailable()) {
              dyIndex = IK_addDynamicValid(fields[i],k);
              IK_SelectValid(dyIndex);
            }
          }
          free (fields[i]);
          fields[i] = NULL;
        }
      }
    }
    RemoveAggregate(dynamic_root);
  }
  /* end of code for dynamic valids */

  /******************************
   * GO!
   ******************************/

  /* Do action: */
  if (!rest[0]) {
    usage(); puke("missing command option");
  }
  else if (IW_Strequal(rest[0], "ds")) {  /* output Datasets */
    result = cmd_ds();
  }
  else if (IW_Strequal(rest[0], "ps")) {  /* output Packages */
    result = cmd_ps(rest[1]);
  }
  else if (IW_Strequal(rest[0], "pt")) {  /* output Package Trees */
    result = cmd_pt(rest[1]);
  }
  else if (IW_Strequal(rest[0], "ptg")) { /* output Package Trees for granule */
    result = cmd_ptg(rest[1]);
  }
  else if (IW_Strequal(rest[0], "ptd")) { /* output Package Trees for Dataset */
    result = cmd_ptd(rest[1],rest[2]);
  }
  else if (IW_Strequal(rest[0], "d"))  {  /* output Dataset */
    result = cmd_d(rest[1]);
  }
  else if (IW_Strequal(rest[0], "g"))  {  /* output Granule */
    result = cmd_g(rest[1]);
  }
  else if (IW_Strequal(rest[0], "gd")) {    /* output Granules in Dataset */
    result = cmd_gd(rest[1]);
  }
  else if (IW_Strequal(rest[0], "all")) {   /* output all data, detailed */
    result = cmd_all();
  }
  else if (IW_Strequal(rest[0], "list")) {  /* output last search */
    result = cmd_list(list_file);
  }
  else {
    usage(); puke("bad usage");
  }
  
  /* How did we do? */
  if (result != 0) {
    puke("command failed");
  }

  /* Ok! */
  exit(0);
}

/************************************************************
 * handle_value
 ************************************************************
 *
 * DESCRIPTION
 *    This function looks for the requestField in the group, tgroup.
 *    If it finds the group it returns the valids in the field_ary along
 *    with *num of valids found in the group.
 */
int handle_value (char ***field_ary, AGGREGATE tGroup, int *num, 
		  char *requestField)
{
  VALUE tempValue;
  VALUE fieldNameValue;
  VALUE fieldValue;
  PARAMETER tempPara;
  PARAMETER fieldPara;
  PARAMETER fieldNamePara;
  AGGREGATE fieldGroup;
  char category_id[MAXLEN];
  char requestValid[MAXLEN];

  for (ptr = requestField;*ptr;ptr++)       /* upcase the field name */
   *ptr = toupper (*ptr);
  tempPara = FindParameter (tGroup, requestField);
  if (tempPara) {
    fprintf (stderr, "found Para\n");
    tempValue = FirstValue (tempPara);
    while (tempValue) {
      fprintf (stderr, "%s is %s \n", requestField, tempValue->item.value.string);
      (*field_ary)[*num] = 
       (char *)calloc(strlen(tempValue->item.value.string)+1, sizeof(char));
      strcpy ((*field_ary)[*num], tempValue->item.value.string);
      (*num)++;
      tempValue = NextValue (tempValue);
    }
    (*field_ary)[*num] = (char *) NULL;
    return (1);
  }
  else {
    for (fieldGroup = FindGroup(tGroup, "FIELD");
     fieldGroup;
     fieldGroup = FindNextGroup(tGroup, fieldGroup, "FIELD")){
      fieldNamePara = FindParameter (fieldGroup, "NAME");
      if (fieldNamePara) {
        fieldNameValue = FirstValue (fieldNamePara);
        if(fieldNameValue && fieldNameValue->item.value.string)
         strncpy (category_id, fieldNameValue->item.value.string,MAXLEN-1);
        category_id[MAXLEN-1] = '\0';
	for (ptr = category_id;*ptr;ptr++)       /* upcase the field name */
	 *ptr = toupper (*ptr);
        if(requestField && category_id && 
         (strcmp(requestField, category_id) == 0)) {
	  fieldPara = FindParameter (fieldGroup, "VALUES");
	  fieldValue = FirstValue (fieldPara);
          while (fieldValue) {
            if(fieldValue && fieldValue->item.value.string)
              strncpy (requestValid, fieldValue->item.value.string,MAXLEN-1);
            requestValid[MAXLEN-1] = '\0';
            fprintf (stderr, "%s is %s \n", requestField, requestValid);
            (*field_ary)[*num] = 
	     (char *)calloc(strlen(requestValid)+1, sizeof(char));
            strcpy ((*field_ary)[*num], requestValid);
            (*num)++;
            fieldValue = NextValue (fieldValue);
          }
          (*field_ary)[*num] = (char *) NULL;
          return (1);
        }
      }
    }
  }
    return (0);
}

/******************** end of file "imsgdbdump.c" ********************/
