/************************************************************
 * imsmap.c
 ************************************************************
 *
 * DESCRIPTION
 *    Perform GIF drawing for basic map selection.
 *
 * NAMESPACE
 *    Uses the IWM_ namespace, for the "main" module of Iw programs.
 */

#ifndef lint
static char RCSID[] =
   "$Id: imsmap.c,v 1.8 2003/05/02 20:10:24 seufert Exp $";
#endif /* lint */

#include <stdio.h>
#include "gd.h"
#include "gdfontl.h"
#include "gdfonts.h"
#include <varargs.h>

#include "IW.h"
#include "IW_Cgi.h"
#include "IW_Str.h"


/************************************************************
 *
 * Constants
 *
 ************************************************************/


/* The border to leave around the bare GIF when turning it into a map: */
#define IWM_XBORDER 26
#define IWM_YBORDER 17

/* How long a tick mark should be: */
#define IWM_BIGTICK 5
#define IWM_LILTICK 2
#define IWM_MARKLEN 3

/* How often should the tick marks occur? */
#define IWM_XGRIDEVERY 60
#define IWM_YGRIDEVERY 30

/* Mode: */
#define IWM_MPOINT 1
#define IWM_MRECT  2


/************************************************************
 *
 * Globals
 *
 ************************************************************/

/* Some colors we'll need: */
int IWM_ColorBlack;
int IWM_ColorLtGrey;
int IWM_ColorRed;
int IWM_ColorGranule;
int IWM_ColorClear;

/* Total width and height of the full output image: */
int IWM_TotalWidth;
int IWM_TotalHeight;

/* Width and height of just the map portion of the output image: */
int IWM_MapWidth;
int IWM_MapHeight;

/* Useful coordinates: */
int IWM_MapTop, IWM_MapBot;
int IWM_MapLeft, IWM_MapRight;

/* The map file: */
char *IWM_MapFile;

/* Provides a different image then the one we're normally using */
char *IWM_MapFullPath;

/* Gridlines wanted? */
int IWM_WantGrid = 0;

/* Great circles wanted? */
int IWM_WantGreatCircles = 0;



/************************************************************
 * IWM_LonToX
 ************************************************************
 *
 * DESCRIPTION
 *    Convert a -180...+180 longitude to an x coordinate
 *    for a rectangular projection of the given width.
 */ 

int IWM_LonToX(double lon)
{
  int x = (int)((lon + 180.0) * (IWM_MapWidth / 360.0));
  if (x < 0) 
    return 0 + IWM_XBORDER;
  else if (x > IWM_MapWidth)
    return IWM_MapWidth + IWM_XBORDER;
  else
    return x + IWM_XBORDER;
}

/************************************************************
 * IWM_LatToY
 ************************************************************
 *
 * DESCRIPTION
 *    Convert a 90...-90 latitude to a y coordinate
 *    for a rectangular projection of the given height.
 */ 

int IWM_LatToY(double lat)
{
  int y = (int)((90.0 - lat) * (IWM_MapHeight / 180.0));
  if (y < 0) 
    return 0 + IWM_YBORDER;
  else if (y > IWM_MapHeight)
    return IWM_MapHeight + IWM_YBORDER;
  else
    return y + IWM_YBORDER;
}

/************************************************************
 * IWM_Color
 ************************************************************
 *
 * DESCRIPTION
 *    Attempts to reuse an existing color; otherwise,
 *    it allocates a new color entry.
 *
 * WARNINGS
 *    Time consuming on a regular basis.  Allocate the
 *    colors you need ahead of time instead.
 */ 

int IWM_Color(im, r, g, b)
     gdImagePtr im;
     int r, g, b;
{
  int color;

  if ((color = gdImageColorExact(im, r, g, b)) != -1) {
    return color;
  }
  if ((color = gdImageColorAllocate(im, r, g, b)) != -1) {
    return color;
  }
  return gdImageColorClosest(im, r, g, b);
}



/************************************************************
 * IWM_Axes
 ************************************************************
 *
 * DESCRIPTION
 *    Put in the axes.
 */ 

IWM_Axes(im)
     gdImagePtr im;
{
  int lon, lat;
  int x, xl, xr, y, yt, yb, tick;
  char ticklabel[8];
  int dotted[4];

  /* Make style for dotted grid lines: */
  dotted[0] = IWM_ColorLtGrey;
  dotted[1] = gdTransparent;  

  /* Fill in background: */
  gdImageFilledRectangle(im, 0, 0, IWM_TotalWidth, IWM_MapTop, 
			 IWM_ColorClear);
  gdImageFilledRectangle(im, 0, 0, IWM_MapLeft, IWM_TotalHeight, 
			 IWM_ColorClear);
  gdImageFilledRectangle(im, 0, IWM_MapBot, IWM_TotalWidth, IWM_TotalHeight, 
			 IWM_ColorClear);
  gdImageFilledRectangle(im, IWM_MapRight, 0, IWM_TotalWidth, IWM_TotalHeight, 
			 IWM_ColorClear);

  /* X axis... */
  for (lon = -180; lon <= 180; lon += 20) {
    x = IWM_LonToX(lon);

    /* Gridlines (every major tick mark): */
    if ((IWM_WantGrid) && (!((lon+180) % IWM_XGRIDEVERY))) {  
      gdImageSetStyle(im, dotted, 2);
      gdImageLine(im, x, IWM_MapBot, x, IWM_MapTop, gdStyled);
    }

    /* Tick marks: */
    tick = !((lon+180) % IWM_XGRIDEVERY) ? IWM_BIGTICK : IWM_LILTICK;
    gdImageLine(im, x, IWM_MapBot, x, IWM_MapBot + tick, IWM_ColorBlack);  
    gdImageLine(im, x, IWM_MapTop, x, IWM_MapTop - tick, IWM_ColorBlack);  

    /* Tick labels: */
    if (tick == IWM_BIGTICK) {
      sprintf(ticklabel, "%d", lon);
      x -= (strlen(ticklabel) * gdFontSmall->w / 2),
      yt = (IWM_MapTop-IWM_BIGTICK) - gdFontSmall->h;
      yb = (IWM_MapBot+IWM_BIGTICK+1);
      gdImageString(im, gdFontSmall, x, yt, ticklabel, IWM_ColorBlack);
      gdImageString(im, gdFontSmall, x, yb, ticklabel, IWM_ColorBlack);
    }
  }

  /* Y axis... */
  for (lat = -90; lat <= 90; lat += 10) {
    y = IWM_LatToY(lat);

    /* Gridlines (every major tick mark): */
    if ((IWM_WantGrid) && (!((lat+90) % IWM_YGRIDEVERY))) {  
      gdImageSetStyle(im, dotted, 2);
      gdImageLine(im, IWM_MapLeft, y, IWM_MapRight, y, gdStyled);
    }

    /* Tick mark: */
    tick = !((lat+90) % IWM_YGRIDEVERY) ? IWM_BIGTICK : IWM_LILTICK;
    gdImageLine(im, IWM_MapRight, y, IWM_MapRight + tick, y, IWM_ColorBlack);
    gdImageLine(im, IWM_MapLeft,  y, IWM_MapLeft - tick,  y, IWM_ColorBlack);

    /* Tick labels: */
    if (tick == IWM_BIGTICK) {
      sprintf(ticklabel, "%d", lat);
      y -= ((gdFontSmall->h / 2) + 1),
      xr = (IWM_MapRight+IWM_BIGTICK+10) - 
 	((strlen(ticklabel) * gdFontSmall->w / 2));
      xl = (IWM_MapLeft-IWM_BIGTICK-10) - 
 	((strlen(ticklabel) * gdFontSmall->w / 2));
      gdImageString(im, gdFontSmall, xr, y, ticklabel, IWM_ColorBlack);
      gdImageString(im, gdFontSmall, xl, y, ticklabel, IWM_ColorBlack);
    }
  }

  /* Draw the great circles: */
  if (IWM_WantGreatCircles) {
    (void)IWM_GreatCircles(im);
  }

  /* Frame the map: */
  gdImageRectangle(im, IWM_MapLeft, IWM_MapTop, IWM_MapRight, IWM_MapBot,
		   IWM_ColorBlack);

  /* Done! */
  return 0;
}


/************************************************************
 * IWM_GreatCircles
 ************************************************************
 *
 * DESCRIPTION
 *    Draw dotted lines for the equator and the prime meridian.
 */ 

IWM_GreatCircles(im)
     gdImagePtr im;
{
  int dotted[4];
  int eq, pm;   /* equator and prime meridian */

  /* Get color: */
  dotted[0] = IWM_ColorLtGrey;
  dotted[1] = gdTransparent;  

  /* Draw lines: */
  gdImageSetStyle(im, dotted, 2);
  eq = IWM_LatToY(0.0);
  pm = IWM_LonToX(0.0);
  gdImageLine(im, IWM_MapLeft, eq, IWM_MapRight, eq, gdStyled);
  gdImageLine(im, pm, IWM_MapTop, pm, IWM_MapBot, gdStyled);

  return 0;
}


/************************************************************
 * IWM_DrawSRect
 ************************************************************
 *
 * DESCRIPTION
 *    Draw the selection rectangle.
 *
 *    The points are in the [-180,90...180,-90] coordinate
 *    system.
 */ 

int IWM_DrawSRect(im, lon1, lat1, lon2, lat2)
     gdImagePtr im;                  /* output image */
     double lon1, lat1, lon2, lat2;  /* rectangle, in globe coords */
{
  int x1, y1, x2, y2, newx2;
  int dotted[4];
  int x, y;

  /* Create a style: */
  dotted[0] = IWM_ColorBlack;
  dotted[1] = gdTransparent;
  gdImageSetStyle(im, dotted, 2);

  /* Convert to real image coords: */
  x1 = IWM_LonToX(lon1);
  y1 = IWM_LatToY(lat1);
  x2 = IWM_LonToX(lon2);
  y2 = IWM_LatToY(lat2);

  /* Make sure we know which is top, left, bottom, and right: */
  if (y1 > y2) {
    y = y2; y2 = y1; y1 = y;
  }
  if (lon1 > lon2) {
    newx2 = IWM_LonToX(180.0); 
    /* Draw the rectangle: */
    gdImageRectangle(im, x1+1, y1+1, newx2-1, y2-1, IWM_ColorBlack);
    gdImageRectangle(im, x1-1, y1-1, newx2+1, y2+1, IWM_ColorBlack);
    gdImageRectangle(im, x1, y1, newx2, y2, IWM_ColorRed);
    x1 = IWM_LonToX(-180.0);
  }
     
  /* Draw the rectangle: */
  gdImageRectangle(im, x1+1, y1+1, x2-1, y2-1, IWM_ColorBlack);
  gdImageRectangle(im, x1-1, y1-1, x2+1, y2+1, IWM_ColorBlack);
  gdImageRectangle(im, x1, y1, x2, y2, IWM_ColorRed);

  return 0;
}



/************************************************************
 * IWM_DrawSPoint
 ************************************************************
 *
 * DESCRIPTION
 *    Draw the selection point.
 *
 *    The points are in the [-180,90...180,-90] coordinate
 *    system.
 */ 

int IWM_DrawSPoint(im, lon, lat)
     gdImagePtr im;     /* output image */
     double lon, lat;   /* point, in globe coords */
{
  int dotted[4];
  int x, y;

  /* Create a style: */
  dotted[0] = IWM_ColorBlack;
  dotted[1] = gdTransparent;
  gdImageSetStyle(im, dotted, 2);

  /* Convert to real image coords: */
  x = IWM_LonToX(lon);
  y = IWM_LatToY(lat);

  /* Draw the cross hairs: */
  gdImageLine(im, x-1, IWM_MapTop, x-1, IWM_MapBot, IWM_ColorBlack);
  gdImageLine(im, x+1, IWM_MapTop, x+1, IWM_MapBot, IWM_ColorBlack);
  gdImageLine(im, IWM_MapLeft, y-1, IWM_MapRight, y-1, IWM_ColorBlack);
  gdImageLine(im, IWM_MapLeft, y+1, IWM_MapRight, y+1, IWM_ColorBlack);

  gdImageLine(im, x, IWM_MapTop, x, IWM_MapBot, IWM_ColorRed);
  gdImageLine(im, IWM_MapLeft, y, IWM_MapRight, y, IWM_ColorRed);

  return 0;
}


/************************************************************
 * IWM_DrawGMark
 ************************************************************
 *
 * DESCRIPTION
 *    Draw a granule marker
 *
 *    The points are in the [-180,90...180,-90] coordinate
 *    system.
 */ 

int IWM_DrawGMark(im, lon, lat)
     gdImagePtr im;     /* output image */
     double lon, lat;   /* point, in globe coords */
{
  int dotted[4];
  int x, y;

  /* Convert to real image coords: */
  x = IWM_LonToX(lon);
  y = IWM_LatToY(lat);

  /* Draw the cross hairs: */
  gdImageLine(im, x-1, y-IWM_MARKLEN, x-1, y+IWM_MARKLEN, IWM_ColorBlack);
  gdImageLine(im, x+1, y-IWM_MARKLEN, x+1, y+IWM_MARKLEN, IWM_ColorBlack);
  gdImageLine(im, x-IWM_MARKLEN, y-1, x+IWM_MARKLEN, y-1, IWM_ColorBlack);
  gdImageLine(im, x-IWM_MARKLEN, y+1, x+IWM_MARKLEN, y+1, IWM_ColorBlack);

  gdImageLine(im, x, y-IWM_MARKLEN, x, y+IWM_MARKLEN, IWM_ColorGranule);
  gdImageLine(im, x-IWM_MARKLEN, y, x+IWM_MARKLEN, y, IWM_ColorGranule);

  return 0;
}


/************************************************************
 * IWM_Usage
 ************************************************************
 *
 * DESCRIPTION
 *    Die, with a usage message.
 */ 

static void IWM_Usage(msg)
     char *msg;
{  
  IW_Die(0, "bad invocation: %s", msg);
}


/************************************************************
 * IWM_GetTags
 ************************************************************
 *
 * DESCRIPTION
 *    Get the CGI input.
 */ 

IWM_GetTags()
{
  char *value;
  int mode;


  /* Get CGI tags: */
  if (IW_CgiGetInput() < 0) {
    IW_Die(-1, "couldn't get CGI tags");
  }
  
  /* Get the map file to use: */
  IWM_MapFile = IW_CgiValue("map");
  IWM_MapFullPath = IW_CgiValue("mapfull");

  /* No Base Map provided */
  if (!*IWM_MapFile && !*IWM_MapFullPath ) {
    IW_Die(0, "missing `map' directive");
  }

  /* Get the gid display option: */
  if (*IW_CgiValue("grid")) {
    if (IW_Strequal(IW_CgiValue("grid"), "full")) {
      IWM_WantGrid = 1;
      IWM_WantGreatCircles = 0;
    }
    else if (IW_Strequal(IW_CgiValue("grid"), "gc")) {
      IWM_WantGrid = 0;
      IWM_WantGreatCircles = 1;
    }    
  }  

  return 0;
}


/************************************************************
 * IWM_Draw
 ************************************************************
 *
 * DESCRIPTION
 *    Do all the commands
 */ 

int IWM_Draw(im)
     gdImagePtr im;
{
  char *key;
  char *value;
  double lon1, lat1, lon2, lat2;


  /* For each "key = value" pair... */
  IW_CgiRewind();
  while (IW_CgiNext(&key, &value)) {

    if (IW_Strequal(key, "spoint")) {        /* selection point */

      if (sscanf(value, "%lf,%lf", &lon1, &lat1) != 2) {
	return IW_Err(0, "Bad point coordinates: %s", value);
      }
      if (IWM_DrawSPoint(im, lon1, lat1) < 0) {
	return IW_Err(0, "couldn't draw selection point");
      }
    }
    else if (IW_Strequal(key, "srect")) {    /* selection rectangle */
      if (sscanf(value, "%lf,%lf,%lf,%lf", &lon1, &lat1, &lon2, &lat2) != 4) {
	return IW_Err(0, "Bad rectangle coordinates: %s", value);	
      }
      if (IWM_DrawSRect(im, lon1, lat1, lon2, lat2) < 0) {
	return IW_Err(0, "couldn't draw selection rectangle");
      }
    }
    else if (IW_Strequal(key, "gmark")) {    /* granule mark */
      if (sscanf(value, "%lf,%lf", &lon1, &lat1) != 2) {
	return IW_Err(0, "Bad granule coordinates: %s", value);	
      }
      if (IWM_DrawGMark(im, lon1, lat1) < 0) {
	return IW_Err(0, "couldn't draw granule mark");
      }
    }
    else {
      /* ignore */
    }
  }

  return 0;
}


/************************************************************
 * IWM_Main
 ************************************************************
 *
 * DESCRIPTION
 *    Main.
 */ 

int IWM_Main()
{  
  char **argp;
  FILE *in, *out;              /* input and output files */
  gdImagePtr im_in, im_out;    /* input and output images */  
  gdImagePtr brush;            /* brush image */
  char inpath[256];
  int mode;


  /* Set fail handler, and init library: */
  IW_SetDieHandler(IW_CgiDieHandler);
  if (IW_Init() < 0) {
    IW_Die(-1, "couldn't init library");
  }

  /* Get CGI tags, etc. */
  if (IWM_GetTags() < 0) {
    IW_Die(-1, "couldn't get arguments");
  }

  /* First, try to load compiled GIF image, ".gif.gd": *
  sprintf(inpath, "%s/pics/%s.gd", IW_DOCS_DIR, IWM_MapFile);
  if (in = fopen(inpath, "rb")) {     * yep! *
    im_in = gdImageCreateFromGd(in);
    fclose(in);    
  }   */

  if (*IWM_MapFullPath) {
     sprintf(inpath, "/%s", IWM_MapFullPath);
  }
  else {
     sprintf(inpath, "%s/pics/%s", IW_DOCS_DIR, IWM_MapFile);
  }

    /* Load GIF input image, ".gif": */
    if (!(in = fopen(inpath, "rb"))) {    /* nope! */
      IW_Die(0, "can't open %s", inpath);
    } 
    im_in = gdImageCreateFromGif(in);
    fclose(in);    
  
  /* Get size: */
  IWM_MapWidth = gdImageSX(im_in);
  IWM_MapHeight = gdImageSY(im_in);
  IWM_TotalWidth = IWM_MapWidth + (2 * IWM_XBORDER);
  IWM_TotalHeight = IWM_MapHeight + (2 * IWM_YBORDER);
  IWM_MapTop = IWM_YBORDER;
  IWM_MapBot = IWM_TotalHeight - IWM_YBORDER;
  IWM_MapLeft  = IWM_XBORDER;
  IWM_MapRight = IWM_TotalWidth - IWM_XBORDER;
  
  /* Create output image, with a little border: */
  im_out = gdImageCreate(IWM_TotalWidth, IWM_TotalHeight);
    
  /* Copy input image to output, leaving a border on outside: */
  gdImageCopy(im_out, im_in, 
	      IWM_XBORDER, IWM_YBORDER, 0, 0, IWM_MapWidth, IWM_MapHeight);

  /* Allocate the colors we'll need: */
  IWM_ColorBlack   = IWM_Color(im_out, 0, 0, 0);
  IWM_ColorRed     = IWM_Color(im_out, 255, 102, 102); /* was 255,64,64 */
  IWM_ColorLtGrey  = IWM_Color(im_out, 192, 192, 192);
  IWM_ColorGranule = IWM_Color(im_out, 255, 255, 0);
  IWM_ColorClear   = gdImageColorAllocate(im_out, 192, 192, 192);
  gdImageColorTransparent(im_out, IWM_ColorClear);

  /* Draw the axes: */
  if (IWM_Axes(im_out) < 0) {
    return IW_Err(0, "couldn't draw axes");
  }

  /* Draw the stuff on the map: */
  if (IWM_Draw(im_out) < 0) {
    return IW_Err(0, "couldn't draw map content");
  }

  /* Make output image interlaced: */
  gdImageInterlace(im_out, 1);

  /* Write output GIF: */
  printf("Content-type: image/gif\n");
  /* printf("Expires: Friday, 09-Dec-1994 17:30:00 GMT\n"); */
  printf("\n");
  gdImageGif(im_out, stdout);

  gdImageDestroy(im_in);
  gdImageDestroy(im_out);

  return 0;
}


/************************************************************
 * main
 ************************************************************
 *
 * DESCRIPTION
 *    Main.
 */ 

int main()
{
  if (IWM_Main() < 0) {
    IW_Die(-1, "general failure");
  }
  else {
    exit(0);
  }
}


/******************** end of file "imsmap.c" ********************/
