/* imsmapguide.c
 *
 * Purpose:
 *    Program for getting the new guide valid codes for a given array
 *    of attribute/valid pairs.
 *    The old guide search string would be like:
 *       "Parameter=CARBON or Parameter=OZONE"
 *    The new guide search string will be like:
 *       "parm928 or parm2039"
 *    This code will convert (map) each Attribute/value pair into the new code.
 *
 * Usages:
 *
 *    The environment variable GAEADATA_DIR must be set to the
 *    valids directory.
 *
 *    There are two different ways to use this code.
 *    1. To get the new guide valid code for a single attribute/valid
 *       pair, you can supply the atttribute and value on the command
 *       line. Examples:
 *       You type:   imsmapguide Parameter OZONE
 *       Output:     parm2039
 *       You type:   imsmapguide Parameter ORGANIC MATTER
 *       Output:     parm2061
 *
 *    2. To get the new guide valid codes for more than one attribute/valid
 *       pair, you must create a file containing the pairs. You then supply
 *       the name of the file on the command line:
 *       You type:   imsmapguide  queryfile
 *
 *       The format of the file is:
 *       - The first line contains an integer which indicates the number
 *         of attribute/valid pairs.
 *       - The next line contains one attribute, a space, and a valid in
 *         which any spaces have been replaced with question marks '?'.
 *       - Each subsequent line begins with the boolean separating this
 *         pair from the previous. It will be "or" or "and".
 *       - The boolean is followed by a space ' '.
 *       - The space is followed by the attribute.
 *       - The attribute is followed by a space ' '.
 *       - That space is followed by the value for the attribute, but with
 *         any spaces in the value replaced with question marks '?'.
 *
 *       For instance, a search for:
 *          "Parameter=CARBON or Parameter=OZONE and Data_Center=GSFC"
 *       Would be in the file as:
 *          3
 *          Parameter CARBON
 *          or Parameter ORGANIC?MATTER
 *          and Data_Center GSFC
 *
 *       Note that no boolean precedes the first pair. 
 *
 *       The output would be: "parm928 or parm2061 and arch2"
 *
 * Output
 *    Output is a string to standard out (stdout), suitable for
 *    doing a new guide search.
 *
 * Author:
 *    Larry Rosen, Hughes STX Corp, 2-25-97.
 * Notes:
 *    Calls IK codes written be various members of the V0 team to read the
 *    entire set of valids and to get the index of a valid.
 */
#include "Gaea_config.h"
#include <stdio.h>
#include <assert.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>              /* needed for toupper() */
#include <sys/types.h>
#include <sys/stat.h>

#include "IK_Ims.h"
#include "IK_Errno.h"
#include "IK_Syslog.h"
#include "IK_ValNames.h"        /* Macros manipulate valids bit vectors */
#include "IK_DepValid.h"        /* Dependent valids functions */
#include "IK_NewValids.h"       /* The new valids format */

/* Definitions */
/* Maximum length of valids directory and max length of valid. */
#define MAX_LENGTH 256
/* Maximum length of next "boolean newcode", eg. " and sensor2231" */
#define MAX_NEXT 17

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
char *main(int argc, char *argv[])
{
   /* valid type definitions ought to match those in the new IC_GdCommon.c. */
   const char *validType_ptr[] = {
                                 "arch",    /* Data Centers */
                                 "ds",      /* Data Sets */
                                 "source",  /* Platforms/sources */
                                 "sensor",  /* Instruments/Sensors */
                                 "parm",    /* Parameters */
                                 "camp",    /* Campaigns/Projects */
                                 };

   char         *gd_env_ptr;          /* Pointer to environment variable */
   char         gd_dir[MAX_LENGTH];   /* Valids directory */
   int          i;                    /* Counter */
   char         valid[MAX_LENGTH];    /* Value - like ORGANIC MATTER\0 */
   int          attrType;             /* attribute as a number */
   int          num;                  /* Valids index number we want */
   struct stat  file_status;          /* File status */
   FILE         *fp_input;            /* Input file pointer */
   int          read_status;          /* Status for a read data function */
   int          num_valids;           /* Number of valids to map. */
   char         attribute[12];        /* Attribute - like Data_Center\0 */
   char         *pos;                 /* Position within a string */
   char         *return_str;          /* Return string */
   char         bool[4];              /* Boolean - or\0 or and\0 */
   char         next_str[MAX_NEXT];   /* Next string to add to return string */
   int          max_ret_length;       /* Maximum length of return string */
   /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
   /* Get the valids directory */  
   gd_env_ptr = getenv("GAEADATA_DIR");
   if (gd_env_ptr == NULL) {
      (void) printf ("GAEADATA_DIR not specified in environment.");
      exit(1);
   } else {
      strcpy (gd_dir, gd_env_ptr);
   }

   /* Read the dependent valids into bit vector. */
   if (IK_ReadDepValids() == -1) {
      /* Failed to read valids.  exit. */
      (void) printf("Unable to read Valids files. See support staff.");
      exit(1);
   }
   /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
   /* Check the command line. */
   if (argc > 2) {
      /* The number of arguments is two or more (remember argc includes
       * the command name). The user has presumably specified an attribute
       * and a value. Call IK_ValStringToIndex() once to get the new guide
       * index.
       */
      /* The rest of the arguments compose the value. Concatenate them. */
      (void) strcpy(valid, argv[2]);
      for (i=3; i<argc; ++i) {
         (void) strcat(valid, " ");
         (void) strcat(valid, argv[i]);
      }
      /* Convert the attribute name (argv[1]) into a numeric value. */
      if (strcmp(argv[1], "Dataset") == 0)
         attrType = VALIDS_DATASET;
      else if (strcmp(argv[1], "Data_Center") == 0)
         attrType = VALIDS_DAAC;
      else if (strcmp(argv[1], "Parameter") == 0)
         attrType = VALIDS_PARAMETERS;
      else if (strcmp(argv[1], "Sensor") == 0)
         attrType = VALIDS_SENSOR;
      else if (strcmp(argv[1], "Source") == 0)
         attrType = VALIDS_SOURCE;
      else if (strcmp(argv[1], "Campaign") == 0)
         attrType = VALIDS_CAMPAIGN;
      else {
         (void) printf ("Attribute type %s doesn't match any of Dataset, Data_Center, Parameter, Sensor, Source, or Campaign", argv[1]);
         exit(1);
      }
      /* Call IK code to get index of valid. */
      num = IK_ValStringToIndex(valid, attrType);
      (void) printf ("%s%d", validType_ptr[attrType], num);

   /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
   } else if (argc == 2) {
      /* The number of arguments is one (remember argc includes command name).
       * The user has presumably specified a query file. Open and parse file.
       * Call IK_ValStringToIndex() to get the new guide index for each pair.
       */
      /* Check status of file. */
      if (!stat(argv[1], &file_status) == 0) {
         (void) printf("Received non-existent file (%s). See support staff.", argv[1]);
         exit(1);
      } else {
         /* Make sure it is a regular file. */
         if (!S_ISREG(file_status.st_mode)) {
            (void) printf("Received non-regular file (%s). See support staff.", argv[1]);
            exit(1);
         } else {
            /* Make sure file size is > 0 */
            if (file_status.st_size <= 0) {
               (void) printf("Received a file size 0 (%s). See support staff.", argv[1]);
               exit(1);
            }
         }
      }
      /* Passed filename is an ok file.  Open file. */
      fp_input = fopen (argv[1], "r");
      if (fp_input == NULL) {
         (void) printf("Couldn't open query file (%s). See support staff.", argv[1]);
         exit(1);
      }
      /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
      /* First line of file gives number of attribute value pairs. */
      read_status = fscanf(fp_input, "%d", &num_valids);
      if (read_status != 1) {
         (void) printf("Couldn't read query file count (%s). See support staff.", argv[1]);
         exit(1);
      }
      /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
      /* Allocate memory for string. */
      max_ret_length = num_valids * MAX_NEXT;
      return_str = (char *) malloc(max_ret_length);
      if (!return_str) {
         (void) printf("Couldn't allocate enough memory. Needs %d bytes. Query may be too large. See support staff or try again but select fewer items.", max_ret_length);
         exit(1);
      }
      /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
      /* Read the first attribute value pair. This line has no boolean. */
      read_status = fscanf(fp_input, "%s%s", &attribute, &valid);
      if (read_status != 2) {
         (void) printf("Couldn't read query file pair 1 (%s). See support staff.", argv[1]);
         exit(1);
      }
      /* Unencode spaces of value (replace any '?' with ' '). */
      while (pos = strchr(valid, '?')) {
         *pos = ' ';
      }
      /* Convert the attribute name (attribute) into a numeric value. */
      if (strcmp(attribute, "Dataset") == 0)
         attrType = VALIDS_DATASET;
      else if (strcmp(attribute, "Data_Center") == 0)
         attrType = VALIDS_DAAC;
      else if (strcmp(attribute, "Parameter") == 0)
         attrType = VALIDS_PARAMETERS;
      else if (strcmp(attribute, "Sensor") == 0)
         attrType = VALIDS_SENSOR;
      else if (strcmp(attribute, "Source") == 0)
         attrType = VALIDS_SOURCE;
      else if (strcmp(attribute, "Campaign") == 0)
         attrType = VALIDS_CAMPAIGN;
      else {
         (void) printf ("Attribute type %s doesn't match any of Dataset, Data_Center, Parameter, Sensor, Source, or Campaign", attribute);
         exit(1);
      }
      /* Call IK code to get index of valid. */
      num = IK_ValStringToIndex(valid, attrType);
      /* Begin writing codes into the return string. */
      (void) sprintf (return_str, "%s%d", validType_ptr[attrType], num);
      /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
      /* Now do the rest of the attribute value pairs.
      /* Read each line containing boolean attribute value. */
      for (i=1; i < num_valids; ++i) {
         /* Read the next boolean attribute value line. */
         read_status = fscanf(fp_input, "%s%s%s", &bool, &attribute, &valid);
         if (read_status != 3) {
            (void) printf("Couldn't read query file pair %d (%s). See support staff.", i+1, argv[1]);
            exit(1);
         }
         /* Unencode spaces of value (replace any '?' with ' '). */
         while (pos = strchr(valid, '?')) {
            *pos = ' ';
         }
         /* Convert the attribute name (attribute) into a numeric value. */
         if (strcmp(attribute, "Dataset") == 0)
            attrType = VALIDS_DATASET;
         else if (strcmp(attribute, "Data_Center") == 0)
            attrType = VALIDS_DAAC;
         else if (strcmp(attribute, "Parameter") == 0)
            attrType = VALIDS_PARAMETERS;
         else if (strcmp(attribute, "Sensor") == 0)
            attrType = VALIDS_SENSOR;
         else if (strcmp(attribute, "Source") == 0)
            attrType = VALIDS_SOURCE;
         else if (strcmp(attribute, "Campaign") == 0)
            attrType = VALIDS_CAMPAIGN;
         else {
            (void) printf ("Attribute type %s doesn't match any of Dataset, Data_Center, Parameter, Sensor, Source, or Campaign", attribute);
            exit(1);
         }
         /* Call IK code to get index of valid. */
         num = IK_ValStringToIndex(valid, attrType);
         /* Generate the next string to concatenate to the return string. */
         (void) sprintf (next_str, " %s %s%d",
                         bool, validType_ptr[attrType], num);
         /* Check if concatenating the new code will exceed the length. */
         if (strlen(return_str) + strlen(next_str) + 1 > max_ret_length) {
            (void) printf("String length too long. Select fewer items and try again. There may be a software error: Max length: %d, Current length: %d, Next item length: %d", max_ret_length, strlen(return_str), strlen(next_str));
            exit(1);
         }
         /* Append new code to the return string, precedid by boolean. */
         (void) strcat (return_str, next_str);
      }
      /* Print the composed string. */
      (void) printf ("%s", return_str);

   /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
   } else {
      /* Error. There must be at least one or two arguments. see usage. */
      (void) printf ("Error. There must be at least one or two arguments. See usage.");
      exit(1);
   }
   exit(0);
}
