/*
*	PROGRAM:	imspathrow_WRS2.c
*
*	DATE:		18 March 1997
*
*	PURPOSE:	The purpose of this program is to calculate the
*			latitude and longitude coordinates for the four
*			corner-points for a given WRS-2 path and row.
*
*       USAGE:          imspathrow_WRS2 -p PATH -r ROW
*                       where:
*                           -p PATH  is the path, an integer from 1 to 233
*                           -r ROW   is the row, an integer from 1 to 248
*
*	NOTES:		This program is based on 2 functions written at EDC:
*			  wrsys.c (J. Storey) - calculates the center-point
*			          of a landsat scene for a given path/row
*			  calcorn.c (D. Jackson, C. Wivell) - calculates the 
*				  corner-points for a landsat scene given the 
*				  center-point
*/

#include <stdio.h>
#include <math.h>

#define REQ 6378.135		/* Radius of the earth at the equator in km */
#define E .006694541854858764	/* Eccentricity of the earth */
#ifndef PI
#define PI 3.14159265358979	/* Ratio of circumference to diameter */
#endif
#define RAD 57.295779505601047	/* (Circumference / 2) / PI */
#define WE 0.00007292115642	/* Rotation of the earth */

void wrsys();
void calcorn();

main( argc, argv )
  int argc;				/* argument count */
  char *argv[];				/* argument list */
{
  char path_ind[] = "-p";		/* path number indicator */
  char row_ind[] = "-r";		/* row number indicator */

  double  wrs_lat;			/* WRS SC geodetic latitude */
  double  wrs_long;			/* WRS SC longitude */

  int	path;				/* path number */
  int	row;				/* row number */
  int 	usage = 0;              	/* flag to display program usage */


  if ( argc < 4 )
    usage = 1;

  else if (strcmp(argv[1], path_ind) == 0 )
    {
    sscanf( argv[2], "%d", &path );
    if (strcmp(argv[3], row_ind) == 0 )
      sscanf( argv[4], "%d", &row );
    else
      usage = 1;
    }
  else if (strcmp(argv[1], row_ind) == 0 )
    {
    sscanf( argv[2], "%d", &row );
    if (strcmp(argv[3], path_ind) == 0 )
      sscanf( argv[4], "%d", &path );
    else
      usage = 1;
    }
  else
    usage = 1;

  if (usage == 1 )
    {
    printf( "\nUsage:  %s -p <path> -r <row.ff>\n", argv[0] );
    return 1;
    }

  if (path < 1 || path > 233 || row < 1 || row > 248)
    {
    printf("/* Invalid path and/or row value */\n");
    printf("/* Path value must be in the range 1 to 233 */\n");
    printf("/* Row value must be in the range 1 to 248 */\n");
    printf("\n");
    printf("GROUP = PATH_ROW_USAGE\n");
    printf("  PATH = (1, 233)\n");
    printf("  ROW  = (1, 248)\n");
    printf("END_GROUP = PATH_ROW_USAGE\n");
    printf("\n");
    return 1;
    }

  wrsys (path, row, &wrs_lat, &wrs_long);
  calcorn(wrs_lat, wrs_long, row);

}

void wrsys ( 
  int 		path, 		/* Input path number */
  int         	row, 		/* Input row number */
  double 	*sc_lat,	/* Output scene center latitude (radians) */
  double 	*sc_lon)	/* Output scene center longitude (radians) */

{
  double  central_angle;		/* central travel angle for row */
  double  delta_long;			/* long offset due to central angle */
  double  earth_spin_rate;		/* Earth solar rotation rate */
  double  inclination_angle;		/* CPF orbital inclination */
  double  long_origin;			/* desc node long for this path */
  double  long_path1_row60;		/* CPF Longitude of WRS 001/060 */
  double  sc_ang_rate;			/* angular velocity of spacecraft */
  double  semi_major_axis = 6378137.0;	/* CPF Earth major axis */
  double  semi_minor_axis = 6356752.314;/* CPF Earth minor axis */
  double  wrs_gclat;			/* WRS SC geocentric latitude */
  double  wrs_lat;			/* WRS SC geodetic latitude */
  double  wrs_long;			/* WRS SC longitude */

  int	descending_node_row = 60;  	/* CPF row number at desc node */
  int	scenes_per_orbit = 248;		/* CPF number of WRS rows */
  int	wrs_cycle_days = 16;		/* CPF Number of days per WRS cycle */
  int	wrs_cycle_orbits = 233;		/* CPF number of WRS paths */

/* Get the CPF Parameters (hard code for now) */
  inclination_angle = 98.2 * PI / 180.0;
  long_path1_row60 = -64.6 * PI / 180.0; 

/* Compute Earth angular rate (solar to account for orbital precession) */
  earth_spin_rate = 2.0 * PI / (24.0*3600.0);   /* Solar Rate */

/* Compute the spacecraft angular rate */
  sc_ang_rate = 2.0 * PI * (double)wrs_cycle_orbits
              / (double)(wrs_cycle_days * 24 * 3600);

/* Compute the central travel angle for this row */
  central_angle = 2.0 * PI * ((double)row - (double)descending_node_row)
                / (double)scenes_per_orbit;

/* Compute the WRS geocentric latitude */
  wrs_gclat = asin( -sin(central_angle)*sin(inclination_angle) );

/* Compute the longitude of row 60 for this path */
  long_origin = long_path1_row60 - 2.0 * PI * (double)(path-1)
                                 / (double)wrs_cycle_orbits;

/* Compute the WRS longitude */
  delta_long = atan2( tan(wrs_gclat)/tan(inclination_angle),
                      cos(central_angle)/cos(wrs_gclat) );
  wrs_long = long_origin - delta_long
           - central_angle*earth_spin_rate/sc_ang_rate;

/* Make sure the longitude is in +/- PI range */
  while ( wrs_long > PI )
    wrs_long -= 2.0 * PI;
  while ( wrs_long < -PI )
    wrs_long += 2.0 * PI;

/* Convert the WRS geocentric latitude to geodetic latitude */
  wrs_lat = atan( tan(wrs_gclat)*(semi_major_axis/semi_minor_axis)
                                *(semi_major_axis/semi_minor_axis) );

/* Convert angles to degrees */
  wrs_lat *= 180.0 / PI;
  wrs_long *= 180.0 / PI;

  *sc_lat = wrs_lat;
  *sc_lon = wrs_long;
}


void calcorn (lats, longs, row)
    double lats, longs;
    int row;
    {

    int	   x;		/* counter for loop */

    double   maxlatsat;   /* Maximum center latitude for satellite */

    double A1,		/* Heading of beginning of a scene */
	   A2,		/* Heading of ending of a scene */
	   B1,		/* Scan azimuth of corner 1 */
	   B2,		/* Scan azimuth of corner 2 */
	   B3,		/* Scan azimuth of corner 3 */
	   B4,		/* Scan azimuth of corner 4 */
	   ANODE,	/* Longitude of the Ascending node */
	   L,		/* Length of scene (earth central angles) */
	   M,		/* Mean anomaly past the equator */
	   M1,		/* Mean anomaly for the beginning of the scene */
	   M2,		/* Mean anomaly for the ending of the scene */
	   MDOT,	/* Mean motion */
	   P,		/* Orbital Period */
	   Re,		/* Radius of the earth */
	   W,		/* Width of scene (earth central angles) */
	   clats,	/* cosine of lats (latitude of scene center */
	   clats1,	/* cosine of lats (latitude of beginning of scene */
	   clats2,	/* cosine of lats (latitude of ending of scene */
	   ck,		/* cosine of k */
	   cW2,		/* cosine of W divided by 2 */
	   dlam,	/* rotation of earth * dt */
	   dlong,	/* longitude at beginning of the scene */
	   dlong1,	/* longitude at beginning of the scene */
	   dlong2,	/* longitude at beginning of the scene */
	   dM,		/* MDOT * dt */
	   dt,		/* time required to create half of the scene */
	   i,		/* inclanation angle */
	   k,		/* PI - i */
	   l,		/* length of scene (km) */
	   lats1,	/* latitude of beginning of scene */
	   lats2,	/* latitude of ending of scene */
	   longs1,	/* longitude of beginning of scene */
	   longs2,	/* longitude of ending of scene */
	   newlat[4],	/* latitudes calulated for corner points in radians */
	   newlon[4],	/* longitudes calulated for corner points in radians */
	   q,		/* Temp calculation to be used later */
	   sinA,	/* sin of Angle A */
	   sk,		/* sin of k */
	   slats,	/* sin of lats */
	   slats1,	/* sin of lats1 */
	   slats2,	/* sin of lats2 */
	   sW2,		/* sin of (W divided by 2) */
	   tmplat[4],	/* calculated latitudes before order is decided */
	   tmplon[4],	/* calculated longitudes before order is decided */
	   w = 185.0,	/* width of scene (km) */
	   xc,		/* Temp calculation to be used later */
	   zc;		/* Temp calculation to be used later */


	    i = 98.22;		/* inclanation angle */
	    l = 170.0;		/* length of scene */
	    P = 98.9;		/* Orbital Period */

/*      if center latitude is greater than the max, adjust the i to the max */
    maxlatsat = 180.0 - i;
    if (fabs(lats) >= maxlatsat)
        i = 180.0 - (fabs(lats)) - .1;

    i = i / RAD;	/* inclanation angle  in radians */

	/* convert center latitude to radians */
    lats = lats / RAD;

	/* convert center longitude to radians */
    longs = longs / RAD;
    if (longs < 0)
	longs = (2 * PI) + longs;

    /* Calculate the radius of the earth at scene center */

    slats = sin(lats);
    q = sqrt(1 - E * E * slats * slats);

/* REQ is the radius of the earth at the equator
 * E is the eccentricity of the earth
 */

    clats = cos (lats);
    xc = (REQ * clats) / q;

    zc = (REQ * (1 - E * E) * slats) / q;

    Re = sqrt(xc * xc + zc * zc);

    /* Calculate the mean anomaly past the equator (M),
       the longitude of the ascending node (ANODE),
       mean motion (MDOT) and change the swath lengths
       to earth central angles (W,L) */

    k = PI - i;
    ck = cos(k);
    sk = sin(k);

    if (fabs(lats) > k)
	lats = (fabs(lats)/lats) * k;
    M = asin(slats/sk);

    if (row < 120)	/* Descending */
	M = PI - M;
    if (M < 0)
	M = (2 * PI) + M;

    dlong = acos(cos(M) / clats);

    if (lats < 0)
	dlong = -dlong;

    ANODE = longs + dlong;

    if (ANODE > (2 * PI))
	ANODE = ANODE - (2 * PI);

    if (ANODE < 0)
	ANODE = (2 * PI) + ANODE;

    MDOT = ((2 * PI) / (P * 60));

    W = w / Re;

    L = l / Re;

/* Calculate the time required to create half of the swath (dt).
 * First calculate the sine of the angle A
 */

    sinA = ck / clats;
    dt = sqrt((1 - cos(L / 2)) / ((MDOT * MDOT) / 2 +
	       MDOT * WE * sinA + (WE * WE) / 2));
    dM = MDOT * dt;
    M1 = M - dM;
    M2 = M + dM;
    lats1 = asin(sk * sin(M1));
    lats2 = asin(sk * sin(M2));
    clats1 = cos(lats1);
    clats2 = cos(lats2);
    dlam = WE * dt;
    dlong1 =acos(cos(M1) / clats1);
    if (lats1 < 0)
	dlong1 = (2 * PI) - dlong1;
    dlong2 =acos(cos(M2) / clats2);
    if (lats2 < 0)
	dlong2 = (2 * PI) - dlong2;
    longs1 = ANODE + dlam - dlong1;
    longs2 = ANODE - dlam - dlong2;
    A1 = asin (ck/clats1);
    A2 = asin (ck/clats2);
    B1 = PI/2 - A1;
    B2 = PI/2 + A1;
    B3 = PI/2 - A2;
    B4 = PI/2 + A2;
    slats1 = sin (lats1);
    slats2 = sin (lats2);
    cW2 = cos (W/2);
    sW2 = sin (W/2);
    tmplat[0] = asin(slats1 * cW2 + clats1 * sW2 * cos(B1));
    tmplat[1] = asin(slats1 * cW2 + clats1 * sW2 * cos(B2));
    tmplat[2] = asin(slats2 * cW2 + clats2 * sW2 * cos(B3));
    tmplat[3] = asin(slats2 * cW2 + clats2 * sW2 * cos(B4));

    tmplon[0]=acos((cW2 - slats1 * sin(tmplat[0])) / (clats1 * cos(tmplat[0])));
    tmplon[1]=acos((cW2 - slats1 * sin(tmplat[1])) / (clats1 * cos(tmplat[1])));
    tmplon[2]=acos((cW2 - slats2 * sin(tmplat[2])) / (clats2 * cos(tmplat[2])));
    tmplon[3]=acos((cW2 - slats2 * sin(tmplat[3])) / (clats2 * cos(tmplat[3])));

    if (row < 120)	/* Descending */
	{
	newlat[0] = tmplat[0];			/* North West Corner */
	newlat[3] = tmplat[1];			/* North East Corner */
	newlat[1] = tmplat[2];			/* South West Corner */
	newlat[2] = tmplat[3];			/* South East Corner */

	newlon[0] = longs1 - tmplon[0];		/* North West Corner */
	newlon[3] = longs1 + tmplon[1];		/* North East Corner */
	newlon[1] = longs2 - tmplon[2];		/* South West Corner */
	newlon[2] = longs2 + tmplon[3];		/* South East Corner */
	}
    else				/* Ascending */
	{
	newlat[2] = tmplat[0];			/* South East Corner */
	newlat[1] = tmplat[1];			/* South West Corner */
	newlat[3] = tmplat[2];			/* North East Corner */
	newlat[0] = tmplat[3];			/* North West Corner */

	newlon[2] = longs1 + tmplon[0];		/* South East Corner */
	newlon[1] = longs1 - tmplon[1];		/* South West Corner */
	newlon[3] = longs2 + tmplon[2];		/* North East Corner */
	newlon[0] = longs2 - tmplon[3];		/* North West Corner */
	}
    for (x = 0; x < 4; x++)
	{
	newlat[x] = newlat[x] * RAD;
	if (newlon[x] > PI)
	    newlon[x] = newlon[x] - (2 * PI);
	if (newlon[x] < -PI)
	    newlon[x] = (2 * PI) + newlon[x];
	newlon[x] = newlon[x] * RAD;
	}

        printf("GROUP = POLYGON_LOC\n");
        printf("  LATITUDE  = (%7.4f, %7.4f, %7.4f, %7.4f)\n",
          newlat[0], newlat[1], newlat[2], newlat[3]);
        printf("  LONGITUDE = (%7.4f, %7.4f, %7.4f, %7.4f)\n",
          newlon[0], newlon[1], newlon[2], newlon[3]);
	printf("  MAP_PROJECTION_TYPE = \"PLATE_CARREE\"\n");
	printf("  TANGENT_LATITUDE  = 0.0\n");
	printf("  TANGENT_LONGITUDE = 0.0\n");
        printf("END_GROUP = POLYGON_LOC\n");
	printf("\n");
    }



