/* ---------------------------------------------------------------------------
   File:
   ~~~~
     gdmap.h

   Description:
   ~~~~~~~~~~~
     Contains function prototypes and necessary definitions for
     functions in gdmap.c - gd-based mapping code

--------------------------------------------------------------------------- */
#ifndef __MAP_H
#define __MAP_H

#include "gd.h"
#include "const.h"

#define STRAIGHTLINE   0
#define GREATCIRCLE    1
#define BACK_FACE_ON   1  /* for drawing the invisible side of the globe */
#define BACK_FACE_OFF  0

/* Map projection types */

#define MERCATOR                 0
#define ORTHOGRAPHIC             1
#define NORTH_POLAR_ORTHOGRAPHIC 2
#define SOUTH_POLAR_ORTHOGRAPHIC 3

/* WDB map definitions */

#define N_MAP_LAYERS      4
#define COAST             0
#define RIVER             1
#define IBOUND            2
#define NBOUND            3

/* GC index for grid */

#define GRID_GC           N_MAP_LAYERS

/* Map resolution constants */

#define WDB_RES_LOW       1
#define WDB_RES_MID       2
#define WDB_RES_HI        5

typedef struct
{
  double lat;
  double lon;
}
LatLon;


typedef struct
{
  int line_width;
  int color;
}
MapGCStr, *MapGC;

/* Functions prototypes */

extern int
ortho     (double radius,     /* radius of the earth in pixels */
	   double center_lat, /* latitude of center of projection (rad) */
	   double center_lon, /* longitude of center of projection (rad) */
	   double inlat,      /* latitude of input point (rad) */
	   double inlon,      /* longitude of input point (rad) */
	   double *outx,      /* x offset from center of window in pixels */
	   double *outy);     /* y offset from center of window in pixels */

extern int
    invortho(double radius,     /* radius of the Earth in pixels */
	     double center_lat, /* latitude of center of projection (rad) */
	     double center_lon, /* longitude of center of projection (rad) */
	     double inx,        /* x offset from center of window in pixles */
	     double iny,        /* y offset from center of window in pixles */
	     double *outlat,    /* latitude of selected point in radians */
	     double *outlon);   /* longitude of selected point in radians */

extern int mercate(double radius,     /* radius of the earth in pixels */
	    double center_lat, /* latitude of center of projection (rad) */
	    double center_lon, /* longitude of center of projection (rad) */
	    double inlat,      /* latitude of input point (rad) */
	    double inlon,      /* longitude of input point (rad) */
	    double *outx,      /* x offset from center of window in pixels */
	    double *outy);     /* y offset from center of window in pixels */

extern int invmercate(double radius,     /* radius of the Earth in pixels */
	       double center_lat, /* latitude of center of projection (rad) */
	       double center_lon, /* longitude of center of projection (rad) */
	       double inx,        /* x offset from center of window in pixles */
	       double iny,        /* y offset from center of window in pixles */
	       double *outlat,    /* latitude of selected point in radians */
	       double *outlon);   /* longitude of selected point in radians */


extern int Map(
     int projection,    /* proj type */
     double radius,     /* radius of the earth in pixels */
     double center_lat, /* latitude of center of projection (rad) */
     double center_lon, /* longitude of center of projection (rad) */
     int    center_x,   /* x coord of the Drawable center */
     int    center_y,   /* y coord of the Drawable center */
     double inlat,      /* latitude of input point (rad) */
     double inlon,      /* longitude of input point (rad) */
     double *outx,      /* x coordinate in the window in pixels */
     double *outy);     /* y coordinate in the window in pixels */


extern int invMap(
         int    projection, /* proj type */ 
         double radius,     /* radius of the Earth in pixels */
         double center_lat, /* latitude of center of projection (rad) */
         double center_lon, /* longitude of center of projection (rad) */
         int    center_x,   /* x coord of the Drawable center */
         int    center_y,   /* y coord of the Drawable center */
         double inx,        /* x coordinate in the window in pixles */
         double iny,        /* y coordinate in the window in pixles */
         double *outlat,    /* latitude of selected point in radians */
         double *outlon);   /* longitude of selected point in radians */


extern int
    segmGreatCir(
    double start_lat,   /* latitude of the starting point */
    double start_lng,   /* longitude of the starting point */
    double end_lat,     /* latitude of the ending point */
    double end_lng,     /* longitude of the ending point */
    double ratio,       /* angle segmemt ratio */
    LatLon *route);     /* resulting intermediate point */

extern void
    segmLine(double lat1,      /* lat coord of starting point */
	     double lon1,      /* lon coord of starting point */
	     double lat2,      /* lat coord of ending point */
	     double lon2,      /* lon coord of ending point */
	     double ratio,     /* ratio */
	     LatLon *route);   /* resulting intermediate point */

extern int
    calcSegm(double lat1,      /* lat coord of starting point */
	     double lon1,      /* lon coord of starting point */
	     double lat2,      /* lat coord of ending point */
	     double lon2,      /* lon coord of ending point */
	     double ratio,     /* ratio */
	     int    line_type, /* line type */
	     LatLon *route);   /* resulting intermediate point */

extern int
    drawMapLine(
       double   lat1,       /* 1st point latitude (rad) */
       double   lon1,       /* 1st point longitude (rad) */
       double   lat2,       /* 2nd point latitude (rad) */
       double   lon2,       /* 2nd point longitude (rad) */
       double   center_lat, /* Latitude of center of projection */
       double   center_lon, /* Longitude of center of projection */
       double   radius,     /* earth radius in pixels */
       gdImagePtr drawable, /* gd image */
       MapGC    front_gc,   /* graphics context */
       MapGC    back_gc,    /* graphics context for back face */
       int      center_x,   /* x coord of the Drawable center */
       int      center_y,   /* y coord of the Drawable center */
       int      linetype,   /* line type */
       int      projection, /* projection type */
       int      back);      /* back face drawing flag */
	
extern int
   drawMapPoint(
       double   lat,        /* latitude of the point */
       double   lon,        /* longitude of the point */
       double   center_lat, /* Latitude of center of projection */
       double   center_lon, /* Longitude of center of projection */
       double   radius,     /* earth radius in pixels */
       gdImagePtr drawable, /* drawable */
       MapGC    front_gc,   /* graphics context */
       MapGC    back_gc,    /* graphics context for back face */
       int      center_x,   /* x coord of the Drawable center */
       int      center_y,   /* y coord of the Drawable center */
       int      projection, /* projection type */
       int      back);      /* back face drawing flag */

extern void 
       drawArea(double inlat[],      /* latitudes of all points */
                double inlon[],      /* longitudes of all points */
                int    n_sides,      /* number of points */
                int    line_type,    /* STRAIGHTLINE or GREATCIRCLE */
                int    pole_inside,  /* True if pole is inside the area */
                gdImagePtr Workspace,/* gd image */ 
                MapGC  front_gc,     /* GC for visible part */
                MapGC  back_gc,      /* GC for invisible part */
                double center_lat,   /* map center lat */
                double center_lon,   /* map center lon */
                double radius,       /* map radius */
                int    center_x,     /* x coord of drawable center */
                int    center_y,     /* y coord of drawable center */
                int    projection,   /* projection type */
                int    backface,     /* True if backface should be drawn */
                int    join);        /* 0 - leave end open */


extern int
    drawWorldMap(gdImagePtr  drawable,    /* gd image */
  		 int         desired[],   /* desired map layers */
  		 int         resolution,  /* desired map resolution */
  		 int         projection,  /* projection type */
  		 int         backface,    /* backface flag for ortho proj. */
  		 int         grid_step,   /* grid step (deg); 0-no grid */
  		 double      center_lat,  /* latitude of map center */
  		 double      center_lon,  /* longitude of map center */
  		 double      radius,      /* earth radius in pixels */
  		 int         center_x,    /* window center x */
  		 int         center_y);   /* window center y */

extern void
  setMapGC(
     int    idx,      /* the GC index */
     MapGC  front,    /* GC for visible part */
     MapGC  back);    /* GC for invisible part */

extern int
  points_clockwise(
     double lat1,
     double lon1,
     double lat2,
     double lon2,
     double lat3,
     double lon3);

extern int
    line_check(
       int xmin, 
       int xmax, 
       int ymin, 
       int ymax, 
       int x1, 
       int y1, 
       int x2, 
       int y2);

#endif  /* __MAP_H */
