package IMSWG.IU;

import External.Acme.Widgets.Slider;
import java.awt.*;

/**
  * This class defines a lightweight Sliderbar class based on the 
  * External.Acme.Widgets.Slider class.
  */

public class Sliderbar extends External.Acme.Widgets.Slider {

  /**
    * Constants to define Slider orientation.
    */
  public static final int HORIZONTAL = 0;
  public static final int VERTICAL   = 1;

  // Constants used to set up standard slider proportions and sizes.
  // These had to be copied from External.Acme.Widgets.Slider because
  // they were not public.

    private static final int THUMB_SIZE = 14;
    private static final int BUFFER = 2;

    private static final int V_TEXT_HEIGHT = 24;
    private static final int H_TEXT_HEIGHT = 15;
    private static final int TEXT_BUFFER = 3;

    private static final int DEFAULT_WIDTH = 100;
    private static final int DEFAULT_HEIGHT = 15;

    private static final int MIN_HEIGHT = 2 * ( THUMB_SIZE + BUFFER + 1 );
    private static final int MIN_WIDTH  = 2 * ( BUFFER + 1 );

    private static final int DEFAULT_MIN = 1;
    private static final int DEFAULT_MAX = 100;

    private int orientation;

  // Other fields inherited from External.Acme.Widgets.Slider, so they don't 
  // need to be duplicated here.

  /** Base constructor, orientation only.
    */
  public Sliderbar(int which) {
    super();
    validateOrientation(which);
    orientation = which;
    setDefaults();
  }

  /** Constructor, orientation plus some values.
    * Call the default constructor and then set minimum and maximum.
    */
  public Sliderbar(int which, int valueMin, int valueMax) {
    this(which);
    if (orientation == Sliderbar.HORIZONTAL) {
      setMinimum(valueMin);
      setMaximum(valueMax);
    }
    else {
      setMinimum(valueMax);
      setMaximum(valueMin);
    }
  }

  /** Constructor, orientation plus some values.
    * Call the minimum/maximum constructor and then set the width.
    */
  public Sliderbar(int which, int valueMin, int valueMax, int width) {
    this(which, valueMin, valueMax);
    setWidth(width);
  }

  // Validate the orientation supplied. Should be either HORIZONTAL or
  // VERTICAL; throw an exception if not.

  void validateOrientation(int theOrientation) {
            switch (theOrientation) {
          case Sliderbar.HORIZONTAL:
          case Sliderbar.VERTICAL:
            this.orientation = orientation;
            break;

          default:
            throw new IllegalArgumentException("illegal scrollbar orientation");
        }

  }

  // Set the default colors and sizes.
    private void setDefaults() {
      if (orientation == Sliderbar.HORIZONTAL) {
        valueMin = DEFAULT_MIN;
        valueMax = DEFAULT_MAX;
      }
      else {
        valueMax = DEFAULT_MIN;
        valueMin = DEFAULT_MAX;
      }
      thumbColor = getBackground();
      if ( thumbColor == null ) {
        thumbColor = Color.lightGray;
      }
      barColor = thumbColor.darker();
      textColor = getForeground();
      if ( textColor == null )
          textColor = Color.black;
      lineColor = textColor;
      if (orientation == Sliderbar.HORIZONTAL) {
         resize( DEFAULT_WIDTH, DEFAULT_HEIGHT+H_TEXT_HEIGHT);
      }
      else {
        resize(DEFAULT_HEIGHT+V_TEXT_HEIGHT,DEFAULT_WIDTH);
      }
      setValue( valueMin );
    }

  /** Sets the slider "height" - that is, the number of pixels required for
      the current value and the slider itself.
    */
  public void setHeight(int num) {
    if (orientation == Sliderbar.HORIZONTAL) {
      super.setHeight(num);
    }
    else {
      if (num < MIN_WIDTH) {
        num = MIN_WIDTH;
      }
      resize(num, size().height);
      repaint();
    }
  }

  /** Sets the slider width - the total number of pixels over which the
      slider will slide.
    */
  public void setWidth(int num) {
    if (orientation == Sliderbar.HORIZONTAL) {
      super.setWidth(num);
    }
    else {
      if (num < MIN_HEIGHT) {
        num = MIN_HEIGHT;
      }
      resize(size().width, num);
      repaint();
    }
  }

    /// Sets the maximum value for the slider.
    // @param num The new maximum.
    public void setMaximum( int num ) {
        valueMax = num;
        setValue( value );
        }

    /// Sets the minimum value for the slider.
    // @param num The new minimum.
    public void setMinimum( int num ) {
        valueMin = num;
        setValue( value );
        }

    /// Sets the current value for the slider. The thumb will move to
    // reflect the new setting.
    // @param num The new setting for the slider.
    public void setValue( int num ) {

        if (orientation == Sliderbar.HORIZONTAL) {
           super.setValue(num);
           return;
        }

        value = num;

        if ( value < valueMax )
            value = valueMax;
        else if ( value > valueMin )
            value = valueMin;

        if ( value != valueMax )
            pixel = (int) ( Math.round(
                Math.abs(
                    (double) ( value - valueMax ) /
                    (double) ( valueMin - valueMax ) ) *
                (double) ( pixelMax - pixelMin ) ) + pixelMin );
        else
            pixel = pixelMin;

        repaint();
        }

  /** The paint method just does a quick check to see if the Slider is 
    * vertical or horizontal. If it's Horizontal, we just have the parent
    * class handle it. Otherwise, we have to do it ourselves.
    */
  public void paint(Graphics g) {
    if (this.orientation == Sliderbar.HORIZONTAL) {
      super.paint(g);
      return;
    }

    // Else we have to do it.
    int height = size().height;
    int width = size().width;

    // Determine how wide the text is. Slight kludge to make sure that the
    // actual area where the sliders are don't get out of sync and look bad.
    int txtHeight;
    if (orientation == Sliderbar.HORIZONTAL) {
      txtHeight = H_TEXT_HEIGHT;
    }
    else {
      txtHeight = V_TEXT_HEIGHT;
    }

    // Figure out how wide the string is.
    String str = String.valueOf(value);
    int strWidth = getFontMetrics(g.getFont()).stringWidth(str);

    // Paint the bar.
    g.setColor(barColor);
    g.fill3DRect(txtHeight, 0, width - txtHeight, height, false);

    // Paint the thumb.
    g.setColor(thumbColor);
    g.fill3DRect(txtHeight + TEXT_BUFFER, pixel - THUMB_SIZE, 
	 width - 2 * BUFFER - txtHeight, THUMB_SIZE * 2 + 1, true);

    // Paint the "index mark" on the thumb.
    g.setColor(lineColor);
    g.drawLine(txtHeight + BUFFER + 1, pixel, width - 2 * BUFFER, pixel);

    // Show the current value.
    g.setColor(textColor);
    g.drawString(str, 0, pixel);
  }

  /** Reshape the slider as required.
    */
  public void reshape(int x, int y, int height, int width) {

    // KLUDGE: We need to make the Canvas (the parent of 
    // External.Acme.Widgets.Slider reshape, but it's bad karma to actually
    // assume anything about the parent class. We'll tell it to do its 
    // reshape anyway, knowing it will draw things wrong, then we'll 
    // redraw them to fix them up here.

    super.reshape(x, y, height, width);

    // If we're horizontal, everything's done OK already.
    if (this.orientation == Sliderbar.HORIZONTAL) {
      return;
    }

    // Calculate new range of pixels in this reshaped slider.
    pixelMax = THUMB_SIZE + BUFFER;
    pixelMin = width - THUMB_SIZE - BUFFER - 1;

    // Calculate where to put the slider. Note that this code reverses
    // the sense of the same code done in the parent class.
    if (value != valueMax) {
      pixel = (int)(Math.round(
        Math.abs(
          (double)(value - valueMax) /
          (double)(valueMin - valueMax)) *
        (double)(pixelMax - pixelMin)) + pixelMin);
    }
    else {
      pixel = pixelMin;
    }
    repaint();
  }

  /** Reposition the thumb based on the current click.
    * Note that this slider does absolute thumb location - that is,
    * the thumb is moved directly to wherever the user clicks, instead
    * of "paging" toward that point.
    */
  public boolean mouseDown(Event e, int x, int y) {
    if (this.orientation == Sliderbar.HORIZONTAL) {
      return super.mouseDown(e, x, y);
    }

    // Handle the vertical case.

    // If the user clicks inside the thumb, move one step toward the
    // side he/she clicked on.
    if (y >= pixel - THUMB_SIZE && y <= pixel + THUMB_SIZE) {
      if (y < pixel) {
	setValue(value + 1);
      }
      else if (y > pixel) {
	setValue(value - 1);
      }
    }
    // Otherwise, use mouseAction to "jump" the thumb to the new point.
    else {
      mouseAction(y);
    }
    motion();
    return true;
  }

  /** Handle mouse drags.
    * We simply call mouseAction to move the thumb appropriately,
    * and also call the user hook to allow subclasses to do appropriate
    * drag actions.
    */
  public boolean mouseDrag(Event e, int x, int y) {
    if (this.orientation == Sliderbar.HORIZONTAL) {
      return super.mouseDrag(e, x, y);
    }
    mouseAction(y);
    motion();
    return true;
  }

  public void mouseAction(int rawPixel) {
    double percent;
    if (orientation == Sliderbar.HORIZONTAL) {
      super.mouseAction(rawPixel);
      return;
    }

    // Force rawPixel into the pixelMin-pixelMax range.
    pixel = Math.min(Math.max(rawPixel, pixelMax), pixelMin);

    // Calculate new percentage.
    if (pixel != pixelMin) {
      percent = (double)(pixel - pixelMin) / (double)(pixelMax - pixelMin);
    }
    else {
      percent = 0;
    }

    // Calculate new value.
    value = (int)
       (Math.round(percent * (double) (valueMin-valueMax))) + valueMax;
    repaint();
  }
}

