package IMSWG.IW;

import java.awt.*;
import java.util.*;
import IMSWG.IW.Parameters;

/**
 * The JumpMenu class sets up a "jump" control item, used to allow the user 
 * to select an alternate destination once he or she leaves the current page. 
 */

public class JumpMenu {

    // jumpButton holds a customised control associated with this instance.
    private JumpChoice   jumpButton;

    // theLastChoice holds the last selection made in the pop-up. 
    private String     theLastChoice;

    // Holds a copy of the Parameters passed in.
    private Parameters myParams;

    /**
     * This method constructs the custom IU pop-up to be associated with
     * this menu and fills it with the "pretty names" passed in.
     * @param myParams The Parameters object constructed by Modelet.
     */

    public JumpMenu (Parameters incomingParams) {

        // Save the Parameters.
        myParams = incomingParams;

        // Construct the UI button.
        jumpButton = new JumpChoice(this);

        // Fill in the button. 
        jumpButton.fill(myParams.getJumpNames());
    }

    /**
     * This routine gives the caller a reference to the associated pop-up.
     * It is cast to Choice because no one outside of JumpMenu is allowed 
     * to know about JumpChoice objects.
     */

    public Choice getUIComponent() {
        return (Choice) jumpButton; 
    }

    /**
     * Returns the String corresponding to the user's last choice; this 
     * will be the default if no choice was made. 
     */
    public String getJumpMenuSelection() {
        return theLastChoice;
    }

    /**
     * Accepts a String and saves it in the "last choice" 
     * field. Invoked by the UI component only.
     * @param theChoice A string containing the last choice made.
     */
    public void setChoice(String theChoice) {
        theLastChoice = myParams.getUgly(theChoice);
    }

/**
 * The JumpChoice class is used to provide the appropriate
 * action routine for the Choice used to implement the UI pop-up.
 * This routine uses a reference to a JumpMenu class to handle
 * the reflection of changes to the UI back to whatever program is
 * using it. This class is declared private because it is not used
 * by anything other than JumpMenu.
 */
}

class JumpChoice extends Choice {

    // Flags whether or not anything has been added yet.
    boolean hasItems = false;

    // Holds a reference back to the JumpMenu this pop-up is
    // associated with.
    JumpMenu theOriginal;

    // Nothing-added constant.
    static final String nowhere = "No alternate locations available";

    // Where-to-go-when-nothing-added constant.
    static final String goNowhere = "Main Search Form";

    /**
     * This method builds a Choice and saves a reference back to the 
     * associated JumpMenu.
     */

    public JumpChoice(JumpMenu theMenu) {
        super();
        theOriginal = theMenu;
    }

    /**
     * Fill up the button with the proper labels. If none were supplied,
     * add a harmless default.
     * @param names Vector containing location names passed in from the
     *        Perl code.
     */

    public void fill(Vector names) {

        String firstEntry = null;
        String nextEntry = null;

        // Extract each entry and add it to the Choice.
        for (Enumeration e = names.elements(); e.hasMoreElements();) {
            nextEntry = (String) e.nextElement();
            this.addItem(nextEntry);

	    // Save the first entry for setting the "current selection" later.
            if (firstEntry == null) {
                firstEntry = nextEntry;
            }

        }

        // If no elements were supplied, build a dummy menu.

        if (firstEntry == null) {
           this.addItem(nowhere);
           firstEntry = goNowhere;
        }

        // Set the current selection to the default (first) selection.
        theOriginal.setChoice(firstEntry); 
    }

    /**
     * Handle an action in this pop-up by extracting the name 
     * and saving that back to the 
     * JumpMenu item associated with this JumpChoice. 
     * @param e Event reference to the event triggering this call
     * @param what If this is our event, the item selected in the Choice.
     */

    public boolean action(Event e, Object what) {
        String s;
        if (e.target instanceof JumpChoice) {
            s = (String) what;
            if (s.equals(nowhere)) {
                theOriginal.setChoice(goNowhere);
            }
            else {
                theOriginal.setChoice(s);
            }
            return true;
        }
        else {
            return false;
        }
    }
}
