/** DisplayOptions.java
 *
 * Description:
 *    Display Options component for Web Gateway's Mapplet.
 *    This file contains the classes for opening a pop-up window when the
 *    user selects Display Options menu item from the Mapplet Map Options menu.
 *    The pop-up enables the user to do the following:
 *       1. Setting the map projection to ortho or plate.
 *       2. Turn map grid overlay on/off.
 *       3. Set the grid spacing in degrees.
 *       4. Turn rivers & streams overlay on/off.
 *       5. Turn political boundaries overlay on/off.
 *       6. At medium resolution (zoom 8 or higher)
 *          - turn places names overlay on/off
            - turn wrs overlay none/wrs1/wrs2
            - turn rail road overlay on/off
            - turn major roads overlay on/off
 *       7. Call the globe object to update the appearance of the globe map
 *          which will be based on these values.
 *
 * Classes:
 *    DisplayOptions - a menu item
 *    DisplayOptionsFrame - a frame that pops up a dialog box to set options.
 *
 * Author: Larry P. Rosen, Hughes STX Corp., September 1997.
 */
package IMSWG.IW.Mapplet;

import java.awt.*;

import IMSWG.IU.ObjectRegistry;
import IMSWG.IW.ModeletTools;
import IMSWG.IW.Mapplet.GeoPolygon;

// = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
/** DisplayOptions class, the main class of this file.
 *    Essentially a menu item with a pop-up window function.
 * Public Access Methods:
 *    DisplayOptions() - constructor.
 *    String getProjection() - returns projection string.
 *    String getGridSpacing() - returns spacing of grid lines in degrees.
 *    String getRivers() - "yes" or "no" to display lakes & rivers.
 *    String getPolitical() - "yes" or "no" to display political boundaries.
 *    String getRail() - "yes" or "no" to display rail roads.
 *    String getRoads() - "yes" or "no" to display roads.
 *    String getPlaces() - "yes" or "no" to display place names.
 *    String getWRS() - "wrs1" , "wrs2" or "none" to display wrs scenes.
 *    void die() - dispose of the frame window - when mapplet ends.
 *    boolean action() - pops up frame dialog box when button clicked.
 */
public class DisplayOptions extends MenuItem{
  private DisplayOptionsFrame window;

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  /**   DisplayOptions() - constructor.  Creates Menu item labeled
   *    "Display Options", creates the pop-up window.
   */
  public DisplayOptions(GeoPolygon geoPoly,boolean alloptions) {
    super("Map Display Options");
    boolean all = false;
    all = alloptions;
    window = new DisplayOptionsFrame("Map Display Options", geoPoly, all);
    ObjectRegistry.putObject("DisplayOptions", this);
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  /** String getProjection() - returns projection string set by dialog box.
   *  Currently, it can be "ortho" or "plate".
   */
  public String getProjection() {
    return (window.getProjection());
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  /** String getGridSpacing() - returns spacing of grid lines in degrees.
   *  Can be "none", "1", "2", "5", "10", or "20".
   */
  public String getGridSpacing() {
    return (window.getGridSpacing());
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  /** String getRail() - returns "yes" or "no" to display rail roads.
   */
  public String getRail() {
    return (window.getRail());
  } 
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  /** String getRoads() - returns "yes" or "no" to display major roads.
   */
  public String getRoads() {
    return (window.getRoads());
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  /** String getPlaces() - returns "yes" or "no" to display place names.
   */
  public String getPlaces() {
    return (window.getPlaces());
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  /** String getWRS() - returns "yes" or "no" to display WRS scenes.
   */
  public String getWRS() {
    return (window.getWRS());
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  /** String getRivers() - returns "yes" or "no" to display lakes & rivers.
   */
  public String getRivers() {
    return (window.getRivers());
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  /** String getPolitical() - returns "yes" or "no" to display political
   *  boundaries.
   */
  public String getPolitical() {
    return (window.getPolitical());
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  /** boolean fast() - returns true if mapSpeed is "fast" or false if
   *  mapSpeed is "detailed".
   */
  public boolean fast() {
    return (window.fast());
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  /** void die() - dispose of the frame - called when mapplet ends. 
   */
  public void die() {
    window.die();
    return;
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  /** boolean displayMenuAction() - handles Display Options menu item click by showing
   * pop-up window.
   */
  public boolean displayMenuAction(Event e) {   
//    System.out.println("e.target = "+e.target);
    if (e.target instanceof MenuItem) {
      if (!window.isShowing())
	window.show();
      return true;
    }
    else {
      System.out.println("Non-MenuItem event");
      return false;
    }
  }
}
// = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
/** DisplayOptionsFrame class, a class for the frame containing the
 *  pop-up dialog box (Frame). Frame has checkboxes and choice menus.
 *  Public Access Methods:
 *     DisplayOptionsFrame() - constructor.
 *     void show() - Set the dialog components to the most recently applied
 *        values and display the frame.
 *    String getProjection() - returns projection string.
 *    String getGridSpacing() - returns spacing of grid lines in degrees.
 *    String getRivers() - "yes" or "no" to display lakes & rivers.
 *    String getPolitical() - "yes" or "no" to display political boundaries.
 *    String getRail() - "yes" or "no" to display rail roads.
 *    String getRoads() - "yes" or "no" to display roads.
 *    String getPlaces() - "yes" or "no" to display place names.
 *    String getWRS() - "wrs1" , "wrs2" or "none" to display wrs scenes.
 *    boolean fast() - true if speed is fast, false if speed is "detailed".
 *    void enableComponents() - enable grid, river, political components.
 *    void disableComponents() - disable grid, river, political components.
 *    void die() - dispose of frame - called when mapplet ends.
 *    boolean action() - respond to Apply, OK, and Cancel button presses.
 *    boolean handleEvent() - handle Window destroy event (clicking X).
 */
class DisplayOptionsFrame extends Frame {
  // UI components in frame.
  private Choice projection;            // "ortho" or "plate"
  private Choice gridSpacing;           // "none", "1", "2", "5", "10", or "20"
  private Choice wrsChoice;             // "none", "wrs1", "wrs2"
  private Checkbox riverChkBx;          // getState() = true or false
  private Checkbox politicalChkBx;      // getState() = true or false
  private Checkbox railChkBx;          // getState() = true or false
  private Checkbox roadsChkBx;          // getState() = true or false
  private Checkbox placesChkBx;          // getState() = true or false
  private Choice mapSpeedChoice;        // "detailed" or "fast"
  // State variables. Initialize these here.
  private String proj;
  private String grid = "20";
  private String wrs = "none";
  private boolean river = true;
  private boolean politic = true;
  private boolean rail = false;
  private boolean roads = false;
  private boolean places = true;
  private String speed = "detailed";
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  /** DisplayOptionsFrame() - constructor.  Creates a Frame with the
   *    given title.  Creates a dialog box pop-up with checkboxes,
   *    choice menus, and Close button.
   */
  DisplayOptionsFrame(String title, GeoPolygon geoPoly,boolean alloptions) {
    super(title);
    boolean all = false;
    all = alloptions; 
    if (all) {
       resize(270, 400);
       setLayout(new GridLayout(12,1));
    } else {
       resize(270, 200);
       setLayout(new GridLayout(7,1));
    }
    add(new Label("Select any of these options:"));
    Panel p1 = new Panel();
    p1.setLayout(new BorderLayout());
    projection = new Choice();
    projection.addItem("ortho");
    // comment out plate carre option for now.
    //    projection.addItem("plate");
    //    proj = (geoPoly.getProjection().equals("ORTHOGRAPHIC") ? "ortho" :"plate");
    proj = "ortho";
    projection.select(proj);
    p1.add("East", projection);
    p1.add("West", new Label("Projection:"));
    add(p1);

    Panel p2 = new Panel();
    p2.setLayout(new BorderLayout());
    gridSpacing = new Choice();
    gridSpacing.addItem("None");
    gridSpacing.addItem("1");
    gridSpacing.addItem("2");
    gridSpacing.addItem("5");
    gridSpacing.addItem("10");
    gridSpacing.addItem("20");
    gridSpacing.select(grid);
    p2.add("East", gridSpacing);
    p2.add("West", new Label("Grid spacing (in degrees):"));
    add(p2);

    Panel p3 = new Panel();
    p3.setLayout(new BorderLayout());
    riverChkBx = new Checkbox("Rivers & Lakes Overlay (Blue)");
    riverChkBx.setState(river);
    p3.add("West", riverChkBx);
    add(p3);

    Panel p4 = new Panel();
    p4.setLayout(new BorderLayout());
    politicalChkBx = new Checkbox("Political Boundaries Overlay (Black)");
    politicalChkBx.setState(politic);
    p4.add("West", politicalChkBx);
    add(p4);

    Panel p5 = new Panel();
    p5.setLayout(new BorderLayout());
    mapSpeedChoice = new Choice();
    mapSpeedChoice.addItem("detailed");
    mapSpeedChoice.addItem("fast");
    mapSpeedChoice.select(speed);
    // Until there are two or more choices, just print a blank label.
    //    p5.add("East", mapSpeedChoice);
    //    p5.add("West", new Label("Map detail/speed:"));
    p5.add("Center", new Label(" "));
    add(p5);

    railChkBx = new Checkbox("Railroad Overlay (Yellow)");
    railChkBx.setState(rail);

    roadsChkBx = new Checkbox("Major roads Overlay (Pink)");
    roadsChkBx.setState(roads);

    placesChkBx = new Checkbox("Place name Overlay (Red)");
    placesChkBx.setState(places);

    wrsChoice = new Choice();
    wrsChoice.addItem("none");
    wrsChoice.addItem("wrs1");
    wrsChoice.addItem("wrs2");
    wrsChoice.select(wrs);

    // Show these options only if all options are asked for
    if (all) {

       Panel p6 = new Panel();
       p6.setLayout(new BorderLayout());
       p6.add("Center", new Label("These overlays only visible at zoom 8 or higher: "));
       add(p6);

       Panel p7 = new Panel();
       p7.setLayout(new BorderLayout());
       p7.add("West", railChkBx);
       add(p7);

       Panel p8 = new Panel();
       p8.setLayout(new BorderLayout());
       p8.add("West", roadsChkBx);
       add(p8);

       Panel p9 = new Panel();
       p9.setLayout(new BorderLayout());
       p9.add("West", placesChkBx);
       add(p9);

       Panel p10 = new Panel();
       p10.setLayout(new BorderLayout());
       p10.add("East", wrsChoice);
       p10.add("West", new Label("WRS scenes Overlay (path-row):"));
       add(p10);

    }

    Panel p11 = new Panel();
    p11.setLayout(new GridLayout(1,3)); 
    p11.add(new Label(" "));
    p11.add(new Button("Close"));
    p11.add(new Label(" "));
    add(p11);
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  /** void show() - display the dialog box with the check boxes and choice
   *     menus in the current state.
   */
  public void show() {
    projection.select(proj);
    gridSpacing.select(grid);
    riverChkBx.setState(river);
    politicalChkBx.setState(politic);
    mapSpeedChoice.select(speed);
    railChkBx.setState(rail);
    roadsChkBx.setState(roads);
    placesChkBx.setState(places);
    wrsChoice.select(wrs);
    // Print instructional info to status box.
    ModeletTools.help("Display Options: Currently only orthographic projection is supported in this applet.\n");
    // ModeletTools.help("Display Options: Choosing fast map disables grid, rivers, political boundaries.\n");
    super.show();
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  /** String getProjection() - returns projection string set by dialog box.
   *  Currently, it can be "ortho" or "plate".
   */
  public String getProjection() {
    return (proj);
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  /** String getGridSpacing() - returns spacing of grid lines in degrees.
   *  Can be "none", "1", "2", "5", "10", or "20".
   */
  public String getGridSpacing() {
    return (grid.equals("none") ? "0" : grid );
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  /** String getRivers() - returns "yes" or "no" to display lakes & rivers.
   */
  public String getRivers() {
    return (river ? "yes" : "no");
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  /** String getPolitical() - returns "yes" or "no" to display political
   *  boundaries.
   */
  public String getPolitical() {
    return (politic ? "yes" : "no");
  }
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  /** String getWRS() - returns wrs overlay type. 
   *  Can be "none", "wrs1", "wrs2".
   */
  public String getWRS() {
    return (wrs);
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  /** String getRail() - returns "yes" or "no" to display rail roads.
   */
  public String getRail() {
    return (rail ? "yes" : "no");
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  /** String getRoads() - returns "yes" or "no" to display major roads.
   */
  public String getRoads() {
    return (roads ? "yes" : "no");
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  /** String getPlaces() - returns "yes" or "no" to display place names.
   */
  public String getPlaces() {
    return (places ? "yes" : "no");
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  /** boolean fast() - returns true if mapSpeed is "fast" or false if
   *  mapSpeed is "detailed".
   */
  public boolean fast() {
    return (speed.equals("fast") ? true : false);
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  /** void enableComponents() - enable grid, river, political components.
   */
  void enableComponents() {
    gridSpacing.enable();
    riverChkBx.enable();
    politicalChkBx.enable();   
    railChkBx.enable();
    roadsChkBx.enable();
    placesChkBx.enable();
    wrsChoice.enable(); 
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  /** void disableComponents() - disable grid, river, political components.
   */
  void disableComponents() {
    gridSpacing.disable();
    riverChkBx.disable();
    politicalChkBx.disable();    
    railChkBx.disable();
    roadsChkBx.disable();
    placesChkBx.disable();
    wrsChoice.disable();
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  /** void die() - dispose of the frame - called when mapplet ends. 
   */
  public void die() {
    dispose();
    return;
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  /** boolean action() - React to the button clicks in the dialog box.
   *    Close - Close dialog.
   *    riverChkBx - River check box was changed. Get value and update map.
   *    politicalChkBx - Political boundaries check box was changed. Get ...
   *    gridSpacing - Grid spacing was changed. Get value and update map.
   */
  public boolean action(Event evt, Object obj) {
    if (obj.equals("Close")) {
      dispose();
      return true;
    }
    else if (evt.target.equals(riverChkBx)) {
      river = riverChkBx.getState();
      // Get the Master globe and tell it to redraw itself and reference.
      MasterGlobe localMasterGlobe;
      localMasterGlobe = (MasterGlobe)ObjectRegistry.getObject("MasterGlobe");
      localMasterGlobe.updateDrawing();
      return true;
    }
    else if (evt.target.equals(politicalChkBx)) {
      politic = politicalChkBx.getState();
      // Get the Master globe and tell it to redraw itself and reference.
      MasterGlobe localMasterGlobe;
      localMasterGlobe = (MasterGlobe)ObjectRegistry.getObject("MasterGlobe");
      localMasterGlobe.updateDrawing();
      return true;
    }
    else if (evt.target.equals(gridSpacing)) {
      grid = gridSpacing.getSelectedItem();
      // Get the Master globe and tell it to redraw itself and reference.
      MasterGlobe localMasterGlobe;
      localMasterGlobe = (MasterGlobe)ObjectRegistry.getObject("MasterGlobe");
      localMasterGlobe.updateDrawing();
      return true;
    }
    else if (evt.target.equals(railChkBx)) {
      rail = railChkBx.getState();
      // Get the Master globe and tell it to redraw itself and reference.
      MasterGlobe localMasterGlobe;
      localMasterGlobe = (MasterGlobe)ObjectRegistry.getObject("MasterGlobe");
      localMasterGlobe.updateDrawing();
      return true;
    }
    else if (evt.target.equals(roadsChkBx)) {
      roads = roadsChkBx.getState();
      // Get the Master globe and tell it to redraw itself and reference.
      MasterGlobe localMasterGlobe;
      localMasterGlobe = (MasterGlobe)ObjectRegistry.getObject("MasterGlobe");
      localMasterGlobe.updateDrawing();
      return true;
    }
    else if (evt.target.equals(placesChkBx)) {
      places= placesChkBx.getState();
      // Get the Master globe and tell it to redraw itself and reference.
      MasterGlobe localMasterGlobe;
      localMasterGlobe = (MasterGlobe)ObjectRegistry.getObject("MasterGlobe");
      localMasterGlobe.updateDrawing();
      return true;
    }
    else if (evt.target.equals(wrsChoice)) {
      wrs = wrsChoice.getSelectedItem();
      // Get the Master globe and tell it to redraw itself and reference.
      MasterGlobe localMasterGlobe;
      localMasterGlobe = (MasterGlobe)ObjectRegistry.getObject("MasterGlobe");
      localMasterGlobe.updateDrawing();
      return true;
    }
    else if (evt.target.equals(mapSpeedChoice)) {
      speed = mapSpeedChoice.getSelectedItem();
      // If chosen map speed is fast disable some options, else enable.
      if (obj.equals("fast")) {
	disableComponents();
      } else {
	enableComponents();
      }
      // Get the Master globe and tell it to redraw itself and reference.
      MasterGlobe localMasterGlobe;
      localMasterGlobe = (MasterGlobe)ObjectRegistry.getObject("MasterGlobe");
      localMasterGlobe.updateDrawing();
      return true;
    }
    return false;
  }
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  /** boolean handleEvent() - handle the Window destroy event when the user
   *   selects "close" from the window's standard menu.  If we didn't,
   *    "close" would also close the Mapplet.  "kill" does kill both.
   */
  public boolean handleEvent(Event evt) {
    if (evt.id == Event.WINDOW_DESTROY && evt.target == this) {
      dispose();
    }
    return super.handleEvent(evt);
  }
}
