package IMSWG.IW.SubSet;

import java.lang.*;
import IMSWG.IW.LatLonPoint;
import IMSWG.IW.SubSet.GeoPolygon;

/**
 * Utility to compute the number of scanlines in a subinterval from its
 * four corner lat/lon points. 
 *
 * This Java code was adapted from a C++ utility DmGwScanSizeUtility.cxx which 
 * was written for the ECS project.  The C++ code has COPYRIGHT[@] 1999
 * by Raytheon Systems Company.  The algorithm used in the C++ code and
 * thus reproduced here in Java was provide by Dr. Brij Gambhir of SAII.
 * 
 * @author Ed Seufert - Raytheon ITSS 
 */

public class ScanSizeUtility {

   // Globals

   double wsSatOrbRate = 14.56282;
   double orbitInclination = 0.1431169986635;
   double scansPerOrbitAngle = 13140.2546212;

  /**
   * The constructor.
   */

   public ScanSizeUtility() {
  
   } 

  /**
   * Public methods 
   */ 

  /**
   * Calculate number of scan lines in the region from GeoPolygon object.
   */
   public int LatLonToScanSize (GeoPolygon gp) {

      return LatLonToScanSize (DegToRads(gp.getPointDouble("p.0.lon")),
                               DegToRads(gp.getPointDouble("p.0.lat")),
                               DegToRads(gp.getPointDouble("p.1.lon")),
                               DegToRads(gp.getPointDouble("p.1.lat")),
                               DegToRads(gp.getPointDouble("p.2.lon")),
                               DegToRads(gp.getPointDouble("p.2.lat")),
                               DegToRads(gp.getPointDouble("p.3.lon")),
                               DegToRads(gp.getPointDouble("p.3.lat")) );
   }

  /**
   * Calculate number of scan lines in the region from coverage object. 
   */
   public int LatLonToScanSize (LatLonPoint[] coverage) {

      return LatLonToScanSize (DegToRads(coverage[0].getLon()),
                               DegToRads(coverage[0].getLat()),
                               DegToRads(coverage[1].getLon()),
                               DegToRads(coverage[1].getLat()),
                               DegToRads(coverage[2].getLon()),
                               DegToRads(coverage[2].getLat()),
                               DegToRads(coverage[3].getLon()),
                               DegToRads(coverage[3].getLat()) );
   }

  /**
   * Calculate number of scan lines in the region from its 4 corner points.
   */
   public int LatLonToScanSize (double beginRightLon, double beginRightLat, 
                                double beginLeftLon, double beginLeftLat, 
                                double endLeftLon, double endLeftLat,
                                double endRightLon, double endRightLat) {

      double beginOrbPos = OrbitFit(beginRightLon,beginRightLat,
                                    beginLeftLon ,beginLeftLat );
// System.out.println("BeginOrbPos is"+beginOrbPos);
      double endOrbPos   = OrbitFit(endRightLon,endRightLat, 
                                    endLeftLon, endLeftLat );
// System.out.println("EndOrbPos is"+endOrbPos);
      // number of scan lines in the subinterval region:
//      if (endOrbPos < 0) { endOrbPos = endOrbPos + 2*Math.PI;}
//      if (beginOrbPos < 0) {beginOrbPos = beginOrbPos + 2*Math.PI;}
      double delta = Math.abs(endOrbPos-beginOrbPos);
      return (int) (scansPerOrbitAngle*delta)+1;
   }

  /**
   * Private methods
   */

  /**
   * Convert to from Decimal degrees to radians.
   */
   private double DegToRads (double deg) {

      return (double) (deg*Math.PI)/180.0;
   }

  /**
   * Fit the 4 corners to a nominal orbit Model 
   */
   private double OrbitFit (double rightLon, double rightLat,
                            double leftLon,  double leftLat ) {


      double descendingNodePhi;
      double descendingNodeLamda;
      double absDescendingNodeLamda;
      double absAngleTheta;
      double longDiff;
      int    ascendDescFlag;
      double xComp;
      double yComp;
      double zComp;
      double normalizer;
      double xn,yn,zn;
      double tempPhi;
      double angleTheta;

      descendingNodePhi = leftLat - rightLat;
      descendingNodeLamda = leftLon - rightLon;

      longDiff = descendingNodeLamda;
      absDescendingNodeLamda = descendingNodeLamda;
      if (descendingNodeLamda < 0) {

         absDescendingNodeLamda = -descendingNodeLamda;

      }
      /* Adjust for longitudes stradling 180 degrees */
      if (absDescendingNodeLamda > Math.PI) {

         longDiff = descendingNodeLamda - 
                   (double)2.0*Math.PI*GetTheSign(descendingNodeLamda);     
      }

      ascendDescFlag = -1;
      if (longDiff/descendingNodePhi > 0) ascendDescFlag = 1;

      xComp = 0.5 * (Math.cos(rightLat) * Math.cos(rightLon) +
                     Math.cos(leftLat)  * Math.cos(leftLon));
      yComp = 0.5 * (Math.cos(rightLat) * Math.sin(rightLon) +
                     Math.cos(leftLat)  * Math.sin(leftLon));
      zComp = 0.5 * (Math.sin(rightLat) + Math.sin(leftLat));

      normalizer = Math.sqrt(xComp*xComp + yComp*yComp + zComp*zComp);

      /*Normalize it */
      xn = xComp / normalizer;
      yn = yComp / normalizer;
      zn = zComp / normalizer;

      tempPhi   = Math.asin(zn);

      angleTheta = - Math.asin(Math.sin(tempPhi)/Math.cos(orbitInclination));
      absAngleTheta = angleTheta;
      if (angleTheta < 0.0) absAngleTheta = -angleTheta;
      if (ascendDescFlag > 0) {
         angleTheta = GetTheSign(tempPhi) * (absAngleTheta - Math.PI);
      }

      return angleTheta; // Orbit position of Descending Node
   } 

   private int GetTheSign(double theValue) {

      if (theValue > 0) {
         return (int)1;
      } 
      else if (theValue < 0) {
         return (int)-1;
      }
      else { 
         return (int)0; 
      }
   }

} // end of class ScanSizeUtility
