/*========================================================================
 * map projections - convert geographic map coordinates
 *
 * 2-July-1991 K.Knowles knowles@kryos.colorado.edu 303-492-0644
 *     (Original version in C)
 * 10-Dec-1992 R.Swick swick@krusty.colorado.edu 303-492-1395 
 *     (Added ellipsoid projections)
 * 22-May-1997 R.Swick swick@chukchi.colorado.edu 303-492-6069
 *     (Java class)
 * National Snow & Ice Data Center, University of Colorado, Boulder
 *========================================================================*/
package IMSWG.IW.SubSet;
import java.io.*;
import java.lang.*;
import java.util.StringTokenizer;


public class mapx
{
  public String map_projection_name;
  public float lat0, lon0, lat1, lon1;
  public float rotation;
  public float scale;
  public float center_lat, center_lon;
  public float south, north;
  public float east, west;
  public float lat_interval, lon_interval;
  public float label_lat, label_lon;
  public int cil_detail, bdy_detail, riv_detail;
  public float equatorial_radius;
  public float eccentricity;
  
  private final float default_eccentricity = (float)0.082271673;
  private final float default_equatorial_radius = (float)6371.228; /* in Km. */
 /* private final boolean TRUE = true;
  private final boolean FALSE = false;
*/private final int TRUE = 0;
  private final int FALSE = -1;

  private double T00, T01, T10, T11, u0, v0;
  private boolean map_stradles_180;
  private double e2, e4, e6, e8, qp, Rq, q1, Rg;
  private double sin_phi1, cos_phi1, sin_phi0, cos_phi0, sin_lam1, cos_lam1;
  private double beta1, sin_beta1, cos_beta1, D, phis, kz;
  private double rho0, n, F, m0, m1, t0, t1;
 
  private String projection_name;
  /*private FILE *mpp_file; *****/
  /*private char *mpp_filename;*****/
  
  private mapx current;
  
/*
 * function prototypes for map transformations
 *
 *
static int init_orthographic(void);
static int orthographic(float,float,float*,float*);
static int gnomonic(float,float,float*,float*);
static int inverse_orthographic(float,float,float*,float*);
static int inverse_gnomonic(float,float,float*,float*);
static int init_polar_stereographic(void);
static int polar_stereographic(float,float,float*,float*);
static int inverse_polar_stereographic(float,float,float*,float*);
static char *standard_name(char *);
*/
/*----------------------------------------------------------------------
 * mapx - initialize map projection 
 *
 *	input :          Map_Projection_Name
 *			 lat0 lon0 [lat1 lon1] (decimal degrees)
 *			 rotation (portrait=0, landscape=90)
 *			 scale (kilometers per map unit)
 *			 center_lat center_lon (for map)
 *			 south north (decimal degrees)
 *			 west east (decimal degrees)
 *			 lat_interval lon_interval (for graticule)
 *			 label_lat label_lon (for graticule)
 *			 cil_detail bdy_detail riv_detail (see database_info)
 *	    [optional]   equatorial_radius (kilometers) 
 *	    [optional]	 eccentricity 
 *			                  
 *
 *			valid projection names are:
 *			 Orthographic
 *			 Polar_Stereographic
 *			or anything reasonably similar
 *
 *	The parameter lat1 is currently used for the Polar_Stereographic
 *	projection to define the "true at" parallel (default pole) and
 *	the Cylindrical_Equal_Area projection to define the latitude of
 *	1:1 aspect ratio (default 30.00). In the Lambert Conic Conformal
 *	lat0 and lat1 are the standard parallels.
 *
 *	result: new mapx class instance initialized for this map
 *		or NULL if an error occurs during initialization
 *
 *	note  : if unable to open .mpp file on first attempt then the
 *		value of the search path environment variable is prepended
 *		to the filename and a second attempt is made
 *
 *		Some important notes on specifying longitudes:
 *		All longitudes should be >= -180 and <= 360.
 *		west to east should not span more than 360 degrees.
 *		West specifies the left side of the map and east the right,
 *		not necessarily the minimum and maximum longitudes.
 *		For purposes of bounds checking all longitudes are 
 *		normalized -180 to 180.
 *
 *----------------------------------------------------------------------*/
  public mapx mapx() {
    return this;
  }

  public mapx init_mapx( String given_map_projection_name,
		       float given_lat0, float given_lon0, float given_lat1, float given_lon1,
		       float given_rotation,
		       float given_scale,
		       float given_center_lat, float given_center_lon,
		       float given_south, float given_north,
		       float given_west, float given_east,
		       float given_lat_interval, float given_lon_interval,
		       float given_label_lat, float given_label_lon,
		       int given_cil_detail, int given_bdy_detail, int given_riv_detail,
		       float given_equatorial_radius,
		       float given_eccentricity)
  {
    this.map_projection_name = this.standard_name(given_map_projection_name);
    this.lat0 = given_lat0;
    this.lon0 = given_lon0;
    this.lat1 = given_lat1;
    this.lon1 = given_lon1;
    this.rotation = given_rotation;
    this.scale = given_scale;
    this.center_lat = given_center_lat;
    this.center_lon = given_center_lon;
    this.south = given_south;
    this.north = given_north;
    this.east = given_east;
    this.west = given_west;
    this.lat_interval = given_lat_interval;
    this.lon_interval = given_lon_interval;
    this.label_lat = given_label_lat;
    this.label_lon = given_label_lon;
    this.cil_detail = given_cil_detail;
    this.bdy_detail = given_bdy_detail;
    this.riv_detail = given_riv_detail;
    this.equatorial_radius = given_equatorial_radius;
    this.eccentricity = given_eccentricity;
    
/*
 *	initialize map projection constants
 */
  if (0 != reinit_mapx())
  { 
    close_mapx();
   return null;
  }
  
  return this;
}

/*
 *  mapx without the optional eccentricity.
 */
public void init_mapx(String given_map_projection_name,
		     float given_lat0, float given_lon0, float given_lat1, float given_lon1,
		     float given_rotation,
		     float given_scale,
		     float given_center_lat, float given_center_lon,
		     float given_south, float given_north,
		     float given_west, float given_east,
		     float given_lat_interval, float given_lon_interval,
		     float given_label_lat, float given_label_lon,
		     int given_cil_detail, int given_bdy_detail, int given_riv_detail,
		     float given_equatorial_radius)
{
  
  init_mapx(given_map_projection_name,
	   given_lat0,  given_lon0,  given_lat1,  given_lon1,
	   given_rotation,
	   given_scale,
	   given_center_lat,  given_center_lon,
	   given_south,  given_north,
	   given_west,  given_east,
	   given_lat_interval, given_lon_interval,
	   given_label_lat, given_label_lon,
	   given_cil_detail, given_bdy_detail, given_riv_detail,
	   given_equatorial_radius,
	   default_eccentricity);
  
}



/*
 *  mapx without the optional eccentricity or equatorial_radius.
 */
public void init_mapx(String given_map_projection_name,
		     float given_lat0, float given_lon0, float given_lat1, float given_lon1,
		     float given_rotation,
		     float given_scale,
		     float given_center_lat, float given_center_lon,
		     float given_south, float given_north,
		     float given_west, float given_east,
		     float given_lat_interval, float given_lon_interval,
		     float given_label_lat, float given_label_lon,
		     int given_cil_detail, int given_bdy_detail, int given_riv_detail)
{
  
  init_mapx(given_map_projection_name,
	   given_lat0,  given_lon0,  given_lat1,  given_lon1,
	   given_rotation,
	   given_scale,
	   given_center_lat,  given_center_lon,
	   given_south,  given_north,
	   given_west,  given_east,
	   given_lat_interval, given_lon_interval,
	   given_label_lat, given_label_lon,
	   given_cil_detail, given_bdy_detail, given_riv_detail,
	   default_equatorial_radius,
	   default_eccentricity);
  
}
 
/* init_mapx with only Projection, lat, lon, scale and radius 
 *
 */
public void init_mapx(String given_map_projection_name,
   float given_lat0, float given_lon0, float scale, float radius) {
  
  init_mapx(given_map_projection_name,
	   given_lat0,  given_lon0,  (float) 0.0,  (float) 0.0,
	   (float) 0.0,
	   scale,
	   given_lat0,  given_lon0,
	   (float) -90.0,  (float) 90,
	   (float) -180.0,  (float) 180.0,
	   (float) 180.0, (float) 360.0,
	   given_lat0, given_lon0,
	   1, 1, 1,
	   radius,
	   default_eccentricity);
  
}
/**** switching methods - depend on the map projection ****/  

private int initialize()
{
  if((this.map_projection_name.equals("ORTHOGRAPHIC")) ||
     (this.map_projection_name.equals("GNOMONIC"))) 
    return init_orthographic();
  
  else if(this.map_projection_name.equals("POLARSTEREOGRAPHIC"))
    return init_polar_stereographic();
  
  else
  { System.out.println("initialize: unknown projection "+ this.map_projection_name);
    System.out.println("valid types are:");
    System.out.println(" Orthographic");
    System.out.println(" Polar Stereographic");
    System.out.println(" Or anything reasonably similar.");
    close_mapx();
    return -1;
  }
  
}



public int geo_to_map(float lat, float lon, float[] u, float[] v)
{
  if(this.map_projection_name.equals("ORTHOGRAPHIC"))
    return orthographic(lat, lon, u, v);
  
  else if(this.map_projection_name.equals("POLARSTEREOGRAPHIC"))
    return polar_stereographic(lat, lon, u, v);
  
  else if(this.map_projection_name.equals("GNOMONIC"))
    return gnomonic(lat, lon, u, v);
  
  
  else
  { System.out.println("geo_to_map: unknown projection " + this.map_projection_name);
    System.out.println("valid types are:");
    System.out.println(" Orthographic");
    System.out.println(" Polar Stereographic");
    System.out.println(" Or anything reasonably similar.");
    close_mapx();
    return -1;
  }
  
}


public int map_to_geo(float x, float y, float[] lat, float[] lon)
{
  
  if(this.map_projection_name.equals("ORTHOGRAPHIC"))
    return inverse_orthographic(x, y, lat, lon);
  
  else if(this.map_projection_name.equals("POLARSTEREOGRAPHIC"))
    return inverse_polar_stereographic(x, y, lat, lon);
  
  else if(this.map_projection_name.equals("GNOMONIC"))
    return inverse_gnomonic(x, y, lat, lon);
  
  else
  { System.out.println("map_to_geo: unknown projection " + this.map_projection_name);
    System.out.println("valid types are:");
    System.out.println(" Orthographic");
    System.out.println(" Polar Stereographic");
    System.out.println(" Or anything reasonably similar.");
    close_mapx();
    return -1;
  }
  
}

/*------------------------------------------------------------------------
 * close_mapx - free resources associated with active mapx struct
 *
 *	input : this - pointer to map data structure (returned by mapx)
 *
 *      result: do nothing - garbage collection handles it.
 *------------------------------------------------------------------------*/
public void close_mapx()
{
/*
  this = null;
*/
    return;
}

/*------------------------------------------------------------------------
 * reinit_mapx - re-initialize map projection constants
 *
 *	input : this - pointer to map data structure (returned by mapx)
 *
 *		The client may set user specified constants in the 
 *		mapx_class struct and this routine will re-calculate
 *		the appropriate private constants for the projection
 *
 *	result: 0 = success, -1 = error return
 *
 *------------------------------------------------------------------------*/
public int reinit_mapx()
{
  float theta;
  float[] u = {999}, v = {999};
  
  /*
   *	check map bounds
   */
  if (this.east < -180 || this.east > 360 
      || this.west < -180 || this.west > 360)
  { System.out.println("reinit_mapx: illegal bounds: west="+this.west+"east="+this.east);
    System.out.println("           should be >= -180 and <= 360 degrees.");
    return -1;
  }
  
  if (Math.abs(this.east - this.west) > 360)
  { System.out.println("reinit_mapx: illegal bounds: west="+this.west+"east="+this.east);
    System.out.println("           bounds cannot span > 360 degrees.");
    return -1;
  }
  
  if (this.east > 180 && this.west > 180)
  { this.east -=360;
    this.west -=360;
  }
  
  /*
   *	set flag for bounds checking
   */
  if (this.east < this.west || this.east > 180)
    this.map_stradles_180 = true;
  else
    this.map_stradles_180 = false;
  
  this.east = normalize(this.east);
  this.west = normalize(this.west);
  
  /*
   *	set series expansion constants
   */
  this.e2 = (this.eccentricity) * (this.eccentricity);
  this.e4 = (this.e2) * (this.e2);
  this.e6 = (this.e4) * (this.e2); 
  this.e8 = (this.e4) * (this.e4);
  
  /*
   *	set projection constants
   */
  current = this;
  if (TRUE != this.initialize()) return -1;
  
  /*
   *	set scaled radius for spherical projections
   */
  this.Rg = this.equatorial_radius / this.scale;
  
  /*
   *	create rotation matrix
   */
  theta = radians(this.rotation);
  
  this.T00 =  Math.cos(theta);
  this.T01 =  Math.sin(theta);
  this.T10 = -Math.sin(theta);
  this.T11 =  Math.cos(theta);
  
  /*
   *	get the offset from the projection origin (lat0,lon0)
   *	to this map's origin (center_lat, center_lon)
   */
  this.u0 = this.v0 = 0.0;
  forward_mapx (this.center_lat, this.center_lon, u, v);
  this.u0 = (double) u[0];
  this.v0 = (double) v[0];
  
  return 0;
}

/*------------------------------------------------------------------------
 * These were macros in the C version - Java has no macros.
 *
 *	radians(degrees) - convert degrees to radians.
 *      degrees(radians) - convert radians to degrees.
 *	normalize(degrees) - normalize a longitude.
 *
 *------------------------------------------------------------------------*/
public float radians(double degrees) 
{
  return((float) (degrees * Math.PI/180));
}

public float degrees(double radians) 
{
  return((float) (radians * 180/Math.PI));
}

public void normalize (float[] lon) 
{ 
  while (lon[0] < -180) lon[0] += 360; 
  while (lon[0] >  180) lon[0] -= 360; 
}
public float normalize (float lon) 
{ 
  while (lon < -180) lon += 360; 
  while (lon >  180) lon -= 360;
  return(lon);
}
public float normalize (double lon) 
{ 
  while (lon < -180) lon += 360; 
  while (lon >  180) lon -= 360;
  return((float) lon);
}



/*------------------------------------------------------------------------
 * within_mapx - test if lat,lon are within map transformation bounds
 *
 *	input : lat,lon - geographic coordinates in decimal degrees
 *
 *	result: TRUE iff lat,lon are within specified mapx min,max
 *
 *------------------------------------------------------------------------*/
public int within_mapx (float lat, float lon)
{
  if (lat < this.south || lat > this.north) return FALSE;
  
  normalize(lon);
  
  if (this.map_stradles_180)
  { if (lon > this.east && lon < this.west)
      return FALSE;
  }
  else
  { if (lon < this.west || lon > this.east)
      return FALSE;
  }
  
  return TRUE;
}

/*------------------------------------------------------------------------
 * forward_mapx - forward map transformation
 *
 *	input : lat,lon - geographic coordinates in decimal degrees
 *
 *	output: u,v - map coordinates in map units
 *
 *------------------------------------------------------------------------*/
public int forward_mapx (float lat, float lon, float[] u, float[] v)
{
  int status;
  current = this;
  status = this.geo_to_map(lat, lon, u, v);
  return status;
}

/*------------------------------------------------------------------------
 * inverse_mapx - inverse map transformation
 *
 *	input : u,v - map coordinates in map units
 *
 *	output: lat,lon - geographic coordinates in decimal degrees
 *
 *------------------------------------------------------------------------*/
public int inverse_mapx (float u, float v, float[] lat, float[] lon)
{
  int status;
  current = this;
  status = this.map_to_geo(u, v, lat, lon);
  return status;
}

/*--------------------------------------------------------------------------
 * standard_name - standardize projection name
 *
 *	input : s - original projection name string
 *
 *	result: a valid projection name or ""
 *
 *-------------------------------------------------------------------------*/
public String standard_name(String s)
{
/****
  static char new_name[80];
  char *p = new_name;
****/
  String new_name = s;
  StringTokenizer name_tokenizer = null;
  int i, name_token_count = 0;
  String tmp_name = "";
 
// System.out.println("New Name is: "+new_name);
  
name_tokenizer = new StringTokenizer(new_name, "_");
  name_token_count = name_tokenizer.countTokens();
  for(i=0; i<name_token_count; i++)
  {
    tmp_name = tmp_name + name_tokenizer.nextToken();
  }
  new_name = tmp_name;
  tmp_name = "";
//     System.out.println("New Name is: "+new_name);
  name_tokenizer = new StringTokenizer(new_name, "-");
  name_token_count = name_tokenizer.countTokens();
  for(i=0; i<name_token_count; i++)
  {
    tmp_name = tmp_name + name_tokenizer.nextToken();
  }
  new_name = tmp_name;
  tmp_name = "";
     
// System.out.println("New Name is: "+new_name);
   
name_tokenizer = new StringTokenizer(new_name, " ");
  name_token_count = name_tokenizer.countTokens();
  for(i=0; i<name_token_count; i++)
  {
    tmp_name = tmp_name + name_tokenizer.nextToken();
  }
  new_name = tmp_name;
  tmp_name = "";
     
// System.out.println("New Name is: "+new_name);
  
name_tokenizer = new StringTokenizer(new_name, "(");
  name_token_count = name_tokenizer.countTokens();
  for(i=0; i<name_token_count; i++)
  {
    tmp_name = tmp_name + name_tokenizer.nextToken();
  }
  new_name = tmp_name;
  tmp_name = "";
    
// System.out.println("New Name is: "+new_name); 
  
name_tokenizer = new StringTokenizer(new_name, ")");
  name_token_count = name_tokenizer.countTokens();
  for(i=0; i<name_token_count; i++)
  {
    tmp_name = tmp_name + name_tokenizer.nextToken();
  }
  new_name = tmp_name.toUpperCase();
  
// System.out.println("New Name is: "+new_name);
  
  if (new_name.equals("POLARSTEREOGRAPHIC") || 
	   new_name.equals("STEREOGRAPHICPOLAR"))
  { new_name = "POLARSTEREOGRAPHIC";
  }
  
  return new_name;
}

/*::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
 * projections 
 *
 *	To add a new projection the projection names should be added
 *	to the standard_name function, the standard name is added to
 *	the if-else-if construct in init_mapx and three private functions
 *	must be defined below and declared in the prototypes section.
 *
 *	The initialization function sets all pre-computed projection
 *	constants (this function takes no arguments).
 *
 *	The forward function converts geographic to map coordinates.
 *
 *	input : lat,lon - geographic coords. (decimal degrees)
 *
 *	output: u,v - map coordinates (map units)
 *
 *	result: 0 = valid coordinates
 *		-1 = invalid point
 *
 *	The inverse function converts map to geographic coordinates.
 *
 *	input : u,v - map coordinates (map units)
 *
 *	output: lat,lon - geographic coords. (decimal degrees);
 *
 *	result: 0 = valid coordinates
 *		-1 = invalid point
 *
 *	These functions use the projection parameters specified in the
 *	static global variable "current".
 *
 *::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/

/*------------------------------------------------------------------------
 * polar_stereographic
 *------------------------------------------------------------------------*/

public int init_polar_stereographic()
{
  if (current.lat1 == 999) current.lat1 = current.lat0;
  current.sin_phi1 = Math.sin(radians(current.lat1));
  if (current.lat0 != 90.00 && current.lat0 != -90.00)
  { System.out.println("mapx: only polar aspects allowed: lat0 = "+current.lat0);
    return -1;
  }
  return 0;
}

public int polar_stereographic (float lat, float lon, float[] u, float[] v)
{
  double x = 999, y = 999;
  double phi, lam, rho;
  
  phi = radians(lat);
  lam = radians(lon - current.lon0);
  
  if (90.0 == current.lat0)
  { rho = current.Rg * Math.cos(phi) 
      * (1 + current.sin_phi1) / (1 + Math.sin(phi));
    x =  rho * Math.sin(lam);
    y = -rho * Math.cos(lam);
  }
  else if (-90.0 == current.lat0)
  { rho = current.Rg * Math.cos(phi) 
      * (1 - current.sin_phi1) / (1 - Math.sin(phi));
    x = rho * Math.sin(lam);
    y = rho * Math.cos(lam);
  }
  
  u[0] = (float) (current.T00*x + current.T01*y - current.u0);
  v[0] = (float) (current.T10*x + current.T11*y - current.v0);
  
  return 0;
}

public int inverse_polar_stereographic (float u, float v, float[] lat, float[] lon)
{
  double phi = 999, lam = 999, rho, c, q, x, y;
  
  x =  current.T00*(u+current.u0) - current.T01*(v + current.v0);
  y = -current.T10*(u+current.u0) + current.T11*(v+current.v0);
  
  rho = Math.sqrt(x*x + y*y);
  q = current.Rg*(1 + current.sin_phi1);
  c = 2*Math.atan2(rho, q);
  
  if (90.0 == current.lat0)
  { phi = Math.asin(Math.cos(c));
    lam = Math.atan2(x, -y);
  }
  else if (-90.0 == current.lat0)
  { phi = Math.asin(-Math.cos(c));
    lam = Math.atan2(x, y);
  }
  
  lat[0] = degrees(phi);
  lon[0] = degrees(lam) + current.lon0;
  normalize(lon);
  
  return 0;
}

/*------------------------------------------------------------------------
 * orthographic
 *------------------------------------------------------------------------*/

public int init_orthographic()
{ 
  current.cos_phi1 = Math.cos(radians(current.lat0));
  current.sin_phi1 = Math.sin(radians(current.lat0));

  return 0;
}

public int orthographic (float lat, float lon, float[] u, float[] v)
{
  double x, y;
  double phi, lam, sin_phi, cos_phi, sin_lam, cos_lam, cos_beta;
  
  phi = radians(lat);
  lam = radians(lon - current.lon0);
  
  sin_phi = Math.sin(phi);
  cos_phi = Math.cos(phi);
  cos_lam = Math.cos(lam);
  
  cos_beta = current.sin_phi1 * sin_phi
    + current.cos_phi1 * cos_phi * cos_lam;
  
  if (cos_beta < 0.0) return(-1);
  
  sin_lam = Math.sin(lam);
  x = current.Rg * cos_phi * sin_lam;
  y = current.Rg * (current.cos_phi1*sin_phi 
		     - current.sin_phi1*cos_phi*cos_lam);
  
  u[0] = (float) (current.T00*x + current.T01*y - current.u0);
  v[0] = (float) (current.T10*x + current.T11*y - current.v0);
  
  return 0;
}

public int inverse_orthographic (float u, float v, float[] lat, float[] lon)
{
  double phi, lam, rho, x, y, cos_beta, sin_beta;
  
  x =  current.T00*(u+current.u0) - current.T01*(v+current.v0);
  y = -current.T10*(u+current.u0) + current.T11*(v+current.v0);
  
//  System.out.println("x = "+ x);
//  System.out.println("y = "+ y);
  // System.out.println("current.Rg = "+ current.Rg);
  rho = Math.sqrt(x*x + y*y);
  if (rho == 0.0)
  { phi = radians(current.lat0);
    lam = 0.0;
  }
  else
  { sin_beta = rho/current.Rg;
    cos_beta = Math.sqrt(1 - sin_beta*sin_beta);
    phi = Math.asin(cos_beta*current.sin_phi1 
	       + y*sin_beta*current.cos_phi1/rho);
    if (current.lat0 == 90)
    { lam = Math.atan2(x, -y);
    }
    else if (current.lat0 == -90)
    { lam = Math.atan2(x, y);
    }
    else
    { lam = Math.atan2(x*sin_beta, rho*current.cos_phi1*cos_beta
		  - y*current.sin_phi1*sin_beta);
    }
  }
  
  lat[0] = degrees(phi);
  lon[0] = degrees(lam) + current.lon0;
  normalize(lon);
  /* System.out.println("lon[0] = "+ lon[0]);
  System.out.println("lat[0] = "+ lat[0]); */
  
  return 0;
}

public int gnomonic (float lat, float lon, float[] u, float[] v)
{
  double x, y;
  double phi, lam, sin_phi, cos_phi, sin_lam, cos_lam, cos_beta;
  
  phi = radians(lat);
  lam = radians(lon - current.lon0);
  
  sin_phi = Math.sin(phi);
  cos_phi = Math.cos(phi);
  cos_lam = Math.cos(lam);
  
  cos_beta = current.sin_phi1 * sin_phi
    + current.cos_phi1 * cos_phi * cos_lam;
  
  if (cos_beta <= 0.0) return(-1);
  
  sin_lam = Math.sin(lam);
  x = current.Rg * cos_phi * sin_lam / cos_beta;
  y = (current.Rg * (current.cos_phi1*sin_phi 
		     - current.sin_phi1*cos_phi*cos_lam)) / cos_beta;

  if ((x > current.Rg) || (x < -current.Rg) || (y > current.Rg) || 
      (y < -current.Rg)) {
    return(-1);  /* point cannot be projected */
  }

  u[0] = (float) (current.T00*x + current.T01*y - current.u0);
  v[0] = (float) (current.T10*x + current.T11*y - current.v0);
  return 0;
}

public int inverse_gnomonic (float u, float v, float[] lat, float[] lon)
{
  double phi, lam, rho, x, y, cos_beta, theta, sin_theta, cos_theta, con;
  
  x =  current.T00*(u+current.u0) - current.T01*(v+current.v0);
  y = -current.T10*(u+current.u0) + current.T11*(v+current.v0);
  
//  System.out.println("x = "+ x);
//  System.out.println("y = "+ y);
  // System.out.println("current.Rg = "+ current.Rg);
  rho = Math.sqrt(x*x + y*y);
  if (rho > current.Rg) {
    return (-1);  /* out of this world */
  }
  if (rho == 0.0)
  { phi = radians(current.lat0);
    lam = 0.0;
  }
  else
  { theta = Math.atan(rho/current.Rg);
    sin_theta = Math.sin(theta);
    cos_theta = Math.cos(theta);

    phi = Math.asin((cos_theta * current.sin_phi1) + 
		    (y * sin_theta * cos_theta * current.cos_phi1 / rho));
    con = rho * current.cos_phi1 * cos_theta - 
	  (y * current.sin_phi1 * sin_theta);

    lam = Math.atan2((x * sin_theta), con); 
  }
  
  lat[0] = degrees(phi);
  lon[0] = degrees(lam) + current.lon0;
  normalize(lon);
  /* System.out.println("lon[0] = "+ lon[0]);
  System.out.println("lat[0] = "+ lat[0]); */
  
  return 0;
}


} /*** END MAPX CLASS ****/


/**** For the java stuff we don't need the rest of this. 
 **** It might be handy but for now ... (RSS) *****************************


#ifdef MTEST
/*------------------------------------------------------------------------
 * mtest - interactive test for mapx routines
 *------------------------------------------------------------------------*//*
main(int argc, char *argv[])
{
  float lat, lon, u, v;
  char readln[80];
  mapx_class *the_map;
  int status;
  
  for (;;)
  { printf("\nenter .mpp file name - ");
    gets(readln);
    if (feof(stdin)) { printf("\n"); exit(0);}
    if (*readln == '\0') break;
    the_map = init_mapx(readln);
    if (the_map == NULL) continue;
    
    printf("\nforward_mapx:\n");
    for (;;)
    { printf("enter lat lon - ");
      gets(readln);
      if (feof(stdin)) { printf("\n"); exit(0);}
      if (*readln == '\0') break;
      sscanf(readln, "%f %f", &lat, &lon);
      status = forward_mapx(the_map, lat, lon, &u, &v);
      printf("u,v = %f %f    %s\n", u, v, 
	     status == 0 ? "valid" : "invalid");
      status = inverse_mapx(the_map, u, v, &lat, &lon);
      printf("lat,lon = %f %f     %s\n", lat, lon, 
	     status == 0 ? "valid" : "invalid");
    }
    
    printf("\ninverse_mapx:\n");
    for (;;)
    { int temp;
      printf("enter u v - ");
      gets(readln);
      if (feof(stdin)) { printf("\n"); exit(0);}
      if (*readln == '\0') break;
      sscanf(readln, "%f %f", &u, &v);
      status = inverse_mapx(the_map, u, v, &lat, &lon);
      printf("lat,lon = %f %f    %s\n", lat, lon, 
	     status == 0 ? "valid" : "invalid");
      status = forward_mapx(the_map, lat, lon, &u, &v);
      printf("u,v = %f %f    %s\n", u, v, 
	     status == 0 ? "valid" : "invalid");
    }
    printf("\nwithin_mapx:\n");
    for (;;)
    { printf("enter lat lon - ");
      gets(readln);
      if (feof(stdin)) { printf("\n"); exit(0);}
      if (*readln == '\0') break;
      sscanf(readln, "%f %f", &lat, &lon);
      printf("%s\n", within_mapx(the_map, lat, lon) ? "INSIDE" : "OUTSIDE");
    }
    
    close_mapx();
  }
}
#endif

#ifdef MACCT
#define usage "usage: macct mpp_file"
/*------------------------------------------------------------------------
 * macct - accuracy test mapx routines
 *------------------------------------------------------------------------*//*
#define MPMON
static float dist_km(float lat1, float lon1, float lat2, float lon2)
{ register float phi1, lam1, phi2, lam2, beta;
  phi1 = radians(lat1);
  lam1 = radians(lon1);
  phi2 = radians(lat2);
  lam2 = radians(lon2);
  beta = acos(Math.cos(phi1)*Math.cos(phi2)*Math.cos(lam1-lam2) + Math.sin(phi1)*Math.sin(phi2));
  return beta*mapx_Re_km;
}
#endif

#ifdef MPMON
#ifndef usage
#define usage "usage: mpmon mpp_file [num_its]"
#endif
/*------------------------------------------------------------------------
 * mpmon - performance test mapx routines
 *------------------------------------------------------------------------*//*
main(int argc, char *argv[])
{
  register int i_lat, i_lon, npts = 0, bad_pts = 0;
  int status1, status2;
  int ii, its = 1, pts_lat = 319, pts_lon = 319;
  register float lat, lon, dlat, dlon;
  float latx, lonx, u, v;
  mapx_class *the_map;
#ifdef MACCT
  float err=0, sum=0, sum2=0, stdv=0, max_err=0, lat_max=0, lon_max=0;
#endif
  
  if (argc < 2)
#ifdef MACCT
  { System.out.println("#\tmacct can be used to test the accuracy\n");
    System.out.println("#\tof the mapx routines. It runs the forward and\n");
    System.out.println("#\tinverse transforms at ~100K points over the whole\n");
    System.out.println("#\tmap. Error statistics are accumulated in kilometers.\n");
    System.out.println("#\tTo run the test type:\n");
    System.out.println("#\t\tmacct test.mpp\n");
    System.out.println("\n");
    error_exit(usage);
  }
#else
  { System.out.println("#\tmpmon can be used to monitor the performance\n");
    System.out.println("#\tof the mapx routines. It runs the forward and\n");
    System.out.println("#\tinverse transforms at ~100K points over the whole\n");
    System.out.println("#\tmap. The optional parameter num_its specifies how\n");
    System.out.println("#\tmany times to run through the entire map, (the\n");
    System.out.println("#\tdefault is 1). To run the test type:\n");
    System.out.println("#\t\tmpmon test.mpp\n");
    System.out.println("#\t\tprof mpmon\n");
    System.out.println("\n");
    error_exit(usage);
  }
#endif
  
  the_map = init_mapx(argv[1]);
  if (the_map == NULL) error_exit(usage);
  if (argc < 3 || sscanf(argv[2],"%d",&its) != 1) its = 1;
  
  dlat = the_map.north - the_map.south;
  dlon = the_map.east - the_map.west;
  
  for (ii = 1; ii <= its; ii++)
  { for (i_lat = 0; i_lat <= pts_lat; i_lat++)
    { lat = (float)i_lat/pts_lat * dlat + the_map.south;
      for (i_lon = 0; i_lon <= pts_lon; i_lon++)
      { lon = (float)i_lon/pts_lon * dlon + the_map.west;
	status1 = forward_mapx(the_map, lat, lon, &u, &v);
	status2 = inverse_mapx(the_map, u, v, &latx, &lonx);
	if ((status1 | status2) != 0) ++bad_pts;
	++npts;
#ifdef MACCT
	if ((status1 | status2) == 0)
	{ err = dist_km(lat, lon, latx, lonx);
	  if (err > 0) { sum += err; sum2 += err*err; }
	  if (err > max_err) { max_err=err; lat_max=lat; lon_max=lon; }
	}
#endif
      }
    }
  }
  System.out.println("%d points,  %d bad points\n", npts, bad_pts);
#ifdef MACCT
  npts -= bad_pts;
  if (npts > 0)
  { err = sum/npts;
    stdv = Math.sqrt((sum2 - npts*err*err)/(npts-1));
  }
  System.out.println("average error = %10.4e km\n", err);
  System.out.println("std dev error = %10.4e km\n", stdv);
  System.out.println("maximum error = %10.4e km\n", max_err);
  System.out.println("max error was at %4.2f%c %4.2f%c\n",
	  Math.abs(lat_max), lat_max >= 0 ? 'N' : 'S',
	  Math.abs(lon_max), lon_max >= 0 ? 'E' : 'W');
#endif
}
#endif

**************************************************************/
