/*
 * 3D routines
 */

#include <Gaea_config.h>
#include	<math.h>
#include	<sqt3D.h>

/*
 * The model:  Camera is at some location 'pos', pointing at some 'direction'.
 *	       The focal length of the camera is 'f'.
 *             The final matrix is what you use for projecting an xyz
 *	       coordinate into screen space (UV).
 *
 * How to use: Get position and direction.
 * 	       Apply the next two functions to get FinalMatrix.
 *	       From a point XYZ to UV point ProjectXYZ 
 *
 * World2Cam(pos, direction, RigidMatrix);
 * Cam2Perspective(RigidMatrix, FinalMatrix, f);
 * ProjectXYZ (FinalMatrix, xyz, uv);
 * TransformXYZ (RigidMatrix, xyz_in, xyz_out);
 */

static void trans _ANSI_ARGS_ ((double mat[4][4], double dir[3]));
static void flip  _ANSI_ARGS_ ((double mat[4][4]));
static void yerot _ANSI_ARGS_ ((double mat[4][4], double dir[3]));
static void xerot _ANSI_ARGS_ ((double mat[4][4], double dir[3]));
static void mat_mul _ANSI_ARGS_ ((double u[4][4], double v[4][4], 
				  double w[4][4]));
static void makeiden _ANSI_ARGS_ ((double mat[4][4]));
static int  xmult    _ANSI_ARGS_ ((double *a, int ra, int ca, 
				   double *b, int rb, int cb,
				   double *c, int *rc, int *cc));
static double vec_mag _ANSI_ARGS_ ((double vector[3]));

void TransformXYZ (rigid, xyz_in, xyz_out)
     double rigid[4][4];
     double xyz_in[3];
     double xyz_out[3];
{
  register int i, j;

  for (i=0; i<3; i++) {
    xyz_out[i] = 0.0;
    for (j=0; j<3; j++) {
      xyz_out[i] += rigid[i][j] * xyz_in[j];
    }
    xyz_out[i] += rigid[i][3];	/* xyz[4] is implied 1 */
  }
}

void ProjectXYZ (perspect, xyz, uv)
     double perspect[3][4];
     double xyz[3];
     double uv[3];
{
  register int i, j;		/* assumes that *
				   fourth coordinate of xyz is 1 */
  for (i = 0; i < 3; i++) {
    uv[i] = 0.;
    for (j = 0; j < 3; j++) {
      uv[i] += perspect[i][j] * xyz[j];
    }
    uv[i] += perspect[i][3];	/* xyz[4] is implied 1 */
  }
  uv[0] /= uv[2];
  uv[1] /= uv[2];
  uv[2] = 1.0;
}

void World2Cam (pos, direction, mat)
     double pos[3];
     double direction[3];
     double mat[4][4];
     

	/* mat is the transformation matrix that transforms 
	   to the eye-coordinate system where

		pos		location of the viewer
		direction	direction of view
		mat	 	transformation matrix

	   for now, the viewer's x-axis is assumed to be
	    parallel to the world's X-Y plane
        */


{
  double I[4][4], II[4][4], III[4][4], IV[4][4], T[4][4];
  trans(I, pos);
  flip(II);		/* convert to left-handed */
  yerot(III, direction);  /* rotate eye's y-axis */
  xerot(IV,  direction);   /* rotate eye's x-axis */
	
  mat_mul(IV, III, T);	/* multiply together to get  */
  mat_mul(T,  II, IV);    /* transformation matrix v   */
  mat_mul(IV,  I,  mat);
}

/*
 * Compute the 3 by 4 perspective transformation matrix from the 
 * rigid transformation matrix and the camera parameters
 */

void Cam2Perspective (rigid, perspect, focus)
     double rigid[4][4];
     double perspect[3][4];
     double focus;
{
  register int i,j;
  double mtmp[4][4];
  int pr, pc;
  for(i=0; i<3; i++){
    for(j=0; j<4; j++)
      mtmp[i][j] = 0.0;
  }
  mtmp[0][0] = focus;
  mtmp[1][1] = focus;
  
  mtmp[0][2] = 0.0;
  mtmp[1][2] = 0.0;
  mtmp[2][2] = 1.0;
  xmult((double *) mtmp, 3, 4, (double *) rigid, 4, 4, (double *) perspect,
	&pr, &pc);
}

static void flip(mat)
     double mat[4][4]; /* to convert to left-handed system. */
{
  register i, j;
  for(i = 0; i < 4; i++){
    for(j = 0; j < 4; j++){
      
      mat[i][j] = 0;
      
    }
  }
  mat[0][0] = mat[2][1] = -1;
  mat[1][2] = mat[3][3] = 1;
  
}

#define	ZERO	(0.0)

static int xmult(a, ra, ca, b, rb, cb, c, rc, cc)
     double *a, *b, *c;
     int ca, cb, *cc;
     int ra, rb, *rc;
{
  double vala, valb;
  double  temp;
  register i, j, k;
  if (ca != rb)
    return (1);
  *rc = ra;
  *cc = cb;
  for (i = 0; i < ra; i++) {
    for (j = 0; j < cb; j++) {
      temp = ZERO;
      for (k = 0; k < ca; k++) {
	vala = a[i * ca + k];
	valb = b[k * cb + j];
	temp += vala * valb;
      }
      c[i * cb + j] = temp;
    }
  }
  return (0);
}

static void trans(mat, pos)
     double mat[4][4];
     double pos[3];
{
  register i;
  makeiden(mat);
  for(i = 0; i < 3; i++)
    mat[i][3] = - pos[i];
}
static void yerot(mat, dir)
     double mat[4][4];
     double dir[3];
{
  /* see create_v */
  
  double theta, v, pow(), sin_th, cos_th;

  v = sqrt((dir[0] * dir[0]) + (dir[1]*dir[1]));
  if(v > 0.000001){
    sin_th = - dir[0]/v;
    cos_th = - dir[1]/v;
  }
  else {
    sin_th = 0.;
    cos_th = 1.;
  }
  makeiden(mat);

  mat[0][0] = (mat[2][2] = cos_th);
  mat[0][2] = -(mat[2][0] = sin_th);
}
static void xerot(mat, dir)
     double mat[4][4];
     double dir[3];
{
  double sin_phi, cos_phi, sqrt(), v, w;


  v = sqrt((dir[0]*dir[0]) + (dir[1]*dir[1]));
  w = vec_mag(dir);

  sin_phi = - dir[2]/w;
  cos_phi = v/w;

  makeiden(mat);
  mat[1][1] = (mat[2][2] = cos_phi);
  mat[2][1] = -(mat[1][2] = sin_phi);
}

static void mat_mul(u, v, w)
     double u[4][4];
     double v[4][4];
     double w[4][4];
{
  register i,j,k;
  for(i=0; i<4; i++){
    for(j=0; j<4; j++)
      w[i][j] = 0.0;
  }
  for(i= 0; i<4; i++){
    for(j=0; j<4; j++){
      for(k=0; k<4; k++)
	w[i][j] = (w[i][j]) + ((u[i][k]) * (v[k][j]));
    }
  }
}
static void makeiden(mat)
     double mat[4][4]; 	/* make identity matrix */
{
  int i, j;
  for(i = 0; i < 4; i++){
    for(j = 0; j < 4; j++){
      mat[i][j] = (i == j);
    }
  }
}
static double vec_mag(vector)
     double vector[3];
{
  return(sqrt((vector[0]*vector[0]) +
	      (vector[1]*vector[1]) +
	      (vector[2]*vector[2])));
}
/***********************************************************************/
void Sqt3D_mat_vec(mat, vec, res) /* res = mat +.# vec (inner product) */
     Matrix mat;
     double vec[3], res[3];
{
    register i,j;		/* similar to mat_mul, but assumes that */
				/* fourth coordinate of vec is 1 */
    for(i=0; i<3; i++){
	res[i] = 0.;
	for(j=0; j<3; j++){
	    res[i] = res[i] + mat[i][j]*vec[j];
	}
	res[i] = res[i] + mat[i][3]; /* vec[4] is implied 1 */
    }
}
void Sqt3D_about(a, costh, sinth, m)
    double a[];
    double costh, sinth;
    Matrix m;
{
    Matrix I_cross_a;
    double aaprime;
    register i,j;
  
				/* Make identity matrix */
    for(i=0; i<4; i++){
	for(j=0; j<4; j++){
	    m[i][j] = (i==j);
	}
    }

    I_cross_a[0][0] = I_cross_a[1][1] = I_cross_a[2][2] = 0.0;
    I_cross_a[1][0] = -(I_cross_a[0][1] = a[2]);
    I_cross_a[0][2] = -(I_cross_a[2][0] = a[1]);
    I_cross_a[2][1] = -(I_cross_a[1][2] = a[0]);
  
    for(i=0;i<3;i++){
	for(j=0;j<3;j++){
	    aaprime=a[i] * a[j];
	    m[i][j] = aaprime + costh *((i==j)-aaprime) + 
		sinth * I_cross_a[i][j];
	}
    }
    return;
}

/************************************************************************/
void SqtViewportAttributes(image, width, height)
				/* w, h are width and height */
				/* of the X window or canvas */
				/* scrollable region. */
    Imaging *image;
    int width, height;
{
    ViewPort *vp;
    int dy = 0;
    vp = &(image->viewport);
    vp->x0 = 0,
    vp->y0 = dy;
    vp->x1 = width;
    vp->y1 = height;
    vp->w = (vp->x1 - vp->x0);
    vp->h = (vp->y1 - vp->y0);
}

void Sqt3DWindow(image)
    Imaging *image;
{
  image->datawindow.x0 = -1.0;
  image->datawindow.y0 = -1.0;
  image->datawindow.x1 = 1.0;
  image->datawindow.y1 = 1.0;
  image->datawindow.dx = image->datawindow.x1 - image->datawindow.x0;
  image->datawindow.dy = image->datawindow.y1 - image->datawindow.y0;
}

void SqtUV2XY (image, u, v, x, y)
    Imaging *image;
    double u, v;
    int *x, *y;
{
    long ltmp;
    double dtmp;
    dtmp = u - image->datawindow.x0;
    dtmp *= image->viewport.w;
    ltmp = dtmp / image->datawindow.dx;
    *x = ltmp + image->viewport.x0;

    dtmp = v - image->datawindow.y0;
    dtmp *= image->viewport.h;
    ltmp = dtmp / image->datawindow.dy;
    *y = image->viewport.h - ltmp + image->viewport.y0;
}


void SqtAspect (image)
    Imaging *image;
{
  double dtmp;
  double half;

  /*
   * increase datawindow so that aspect ratio matches viewport
   */

  dtmp = (image->datawindow.dy * image->viewport.w) / image->viewport.h;
  if (image->datawindow.dx < dtmp) {
    half = (dtmp - image->datawindow.dx) / 2.0;
    image->datawindow.x0 -= half;
    image->datawindow.x1 += half;
    image->datawindow.dx = dtmp;
  }
  else {
    dtmp = (image->datawindow.dx * image->viewport.h) / image->viewport.w;
    half = (dtmp - image->datawindow.dy ) / 2.0;
    image->datawindow.y0 -= half;
    image->datawindow.y1 += half;
    image->datawindow.dy = dtmp;
  }
}
