/***************************************************************************
 *
 * NAME: SF_client.c 
 *
 * DESCRIPTION:  This task is the client side of the support file
 *               transfer client/server implementation.  The client
 *               task communicates with the server task to obtain the 
 *               latest support files.
 * 
 *               The client task attempts to connect to the support file
 *               server task based on the command line entries or default
 *               values for the server host name and server service port 
 *               number.
 *
 *               The client task passes the server the platform type
 *               and operating system release as returned by the uname -sr
 *               command or from values passed in on the command line.
 *               These values are needed by the server to determine the
 *               support file type to send.
 *
 *           
 *               In addition, the client task tries to access the current
 *               support files to determine the serial number that is 
 *               embedded in the header if no command line entry for 
 *               the serial number is received.  If no command line entry 
 *               is received and the support files can't be accessed, the 
 *               default value of zero is sent to the server to force it to
 *               send the latest support file set.  The serial number is 
 *               needed so the server can determine if a more up-to-date
 *               version of the support files exists.
 *
 *               In response to the client task's request, the server will
 *               respond with a no update message, an error message, or an
 *               update message.  Following the update message is the number
 *               of bytes that will be transferred, then the new support
 *               files. 
 *
 *               The new support files are written to a single 
 *               file that is in the format of a compressed tar file.
 *               This file is written in the directory as specified
 *               by the support file direcory command line entry.
 *
 * INVOCATION:
 *
 *       status = SF_client [-d] [-f support file name] [-h <host>]
 *                          [-l support file location] [-n serial number]
 *                          [-o os release] [-p platform] [-s service port]
 *                          [-t tar file name] [-v]
 * 
 *           where    
 *                 -d - turns debug messages on, errors and trace messages
 *                 -f - support file name to obtain current serial number from
 *                 -h - server host name or address in dotted decimal notation
 *                 -l - location of the current support files
 *                 -n - serial number of the curren support files
 *                 -o - operating system release as returned from uname -r
 *                 -p - platform name as returned from uname -s
 *                 -s - server service name or port number as in /etc/services
 *                 -t - file name to write file in received from server
 *                 -v - turns verbose messages on, general status messages
 *
 *                 status -  0 = executed properly, no transfer required
 *                         100 = executed properly with transfer
 *                         < 0 = error condition
 *                 Note: The value returned for a good transfer is 
 *                       arbitrarily set to 100.  It could not be set
 *                       to 1 since that conflictss with the command
 *                       failed value of 1 when run from within a script.
 *
 **************************************************************************
 *
 * RCS INFORMATION
 *
 * $RCSfile: SF_client.c,v $
 *
 * $Id: SF_client.c,v 1.2 1997/10/23 19:06:33 solomon Exp $
 *
 * $Log: SF_client.c,v $
 * Revision 1.2  1997/10/23 19:06:33  solomon
 * Added: #include <Gaea_config.h>
 *
 * Revision 1.1  1996/11/19  23:20:24  lrosen
 * Initial revision
 *
 * Revision 1.0.1.6  1996/09/13  15:07:27  ims
 * Changed client server name to alias.  J. Katen
 *
 * Revision 1.0.1.5  1996/08/07  15:34:37  lrosen
 * Changed connect() to SF_connect() to put alarm around connect() in
 * SF_sockIOLib.c.  Added command line option -c # to specify the number
 * of seconds for timeout for connect().
 *
 * Revision 1.0.1.4  1996/07/11  21:52:45  ims
 * Replace references to FILENAME_MAX with MAXPATHLEN and added
 * include file sys/param.h. Also added static in front of callServer
 * to remove compiler warning on HP.
 *
 * Revision 1.0.1.3  1996/06/28  14:05:37  iredell
 * Corrected comment that was entered wrong in previous version
 *
 * Revision 1.0.1.2  1996/06/20  21:11:46  ims
 * Added define for FILENAME_MAX if not already defined to resolve a
 * problem seen at EDC.
 *
 * Revision 1.0.1.1  1996/06/20  21:08:16  ims
 * place holder for new branch
 *
 * Revision 1.0  1996/05/21  18:36:02  ims
 * First offical version, promoted from 0.1.1.5
 *
 * Revision 0.1.1.5  1996/04/11  12:00:01  walbert
 * added ifdef for HPUX systems to reset the size of the
 * FILENAME_MAX value to 255.
 *
 * Revision 0.1.1.4  1996/03/08  04:30:40  walbert
 * Modified the default port number to 4000 to correspond
 * to the port assignment for the operational default
 * support file server as setup by the sysadmin team.
 *
 * Revision 0.1.1.3  1996/02/12  16:42:28  walbert
 * Made changes to routine getSerialNum to check for the
 * support file header version prior to extracting the serial
 * number.  If file header version doesnt match SF_CURRENT_HDR_VERSION
 * it uses the default version.
 *
 * Revision 0.1.1.2  1996/01/09  14:34:17  walbert
 * Incorporated walkthru comments and added additional comments
 *
 * Revision 0.1.1.1  1995/12/19  18:57:59  walbert
 * Updated prologs and comments
 *
 * Revision 0.1  1995/12/18  21:16:20  walbert
 * First pre-release version of the client
 *
 * 
 *************************************************************************/

 /* Define the RCS identification string */
static char rcsid[] = "$Id: SF_client.c,v 1.2 1997/10/23 19:06:33 solomon Exp $"; 

#include <Gaea_config.h>
#include <stdio.h>      
#include <stdlib.h>      
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>  
#include <sys/param.h>    /* MAXPATHLEN for path/filename max length */ 
#include <sys/socket.h>  
#include <netdb.h>        /* hostent, servent, protoent */
#include <netinet/in.h>   /* INADDR_NONE */
#include <arpa/inet.h>    /* inet_addr */
#include <errno.h>        /* errno */
#include <string.h>       /* strlen, memcpy, memset */
#include <signal.h>       /* for sigaction, sigemptyset */
#include <time.h>         /* time_t */
#include <sys/utsname.h>  /* uname information */
#include <assert.h>       /* assert macro for debug purposes */

#include "SF_msg_char.h"  /* client/server message structures */ 

/*************************************************************************
 *
 * declare functions that are external to this file   
 *  
 ************************************************************************/

#include "SF_sockIOLib.h"

/*************************************************************************
 *
 * declare variables that are external to this file   
 *  
 ************************************************************************/

extern char *optarg;

/*************************************************************************
 *
 * declare functions that are internal to this file   
 *  
 ************************************************************************/
 
static void openLogFile (char *, char *);
static void getPfInfo (char **, char **);
static void getSerialNum(char *, char *, char **);
static int  callServer(char *, char *, unsigned int);
static int  rqstSupFile(char *, char *, char *, int);
static int  getSvrReply(char *, char *, int);
static int  getSupFile(char *, char *, int);
static void closeServer(int *);
static void closeLogFile(void);

/***********************************************************************
 *
 * declare variables global to this file 
 *
 ***********************************************************************/

static int debug_on  = 0;
static int verbose_on = 0;
static FILE *log_fp = (FILE *)NULL;

/***********************************************************************
 *
 * define values that are needed throughout the routines in the support
 * file client task.
 *
 ***********************************************************************/

#define DBG_IT(a) debug_on ? (void)(a) : (void)NULL   /* if debug is turned on print debugs */
#define SAY_IT(a) verbose_on ? (void)(a) : (void)NULL /* if verbose is turned on print info */
#define LOG_IT(a) (log_fp != (FILE *)NULL) ? (void)(a) : (void)NULL  /* if log open, write to file */

#define SF_CLIENT_LOG_FILE    "sf_client.log"
#define SF_CURRENT_HDR_VERSION 1

#define DEFAULT_HOST          "imssfp.gsfc.nasa.gov"
#define DEFAULT_PORT          "4000"
#define DEFAULT_SF_PATH       "/usr/local/lib/gaea_data"
#define DEFAULT_SF_NAME       "VALIDSMAP.valids"
#define DEFAULT_NEW_NAME      "new_sf.tar.Z"
#define DEFAULT_SERIAL_NUMBER "0"
#define DEFAULT_PLATFORM      "default"
#define DEFAULT_RELEASE       "all"
#define DEFAULT_CONNECT_ALARM  20

#define TRANSFER_ERROR          -1
#define TRANSFER_NOT_NEEDED      0
#define TRANSFER_SUCCESS         1


/***********************************************************************
 *
 *  MAIN
 *
 ***********************************************************************/
main(int argc, char **argv)
{

  int exit_status= 0;               /* status assigned when exiting */
  int errflg = 0;                   /* err flag for getopt processing */
  int c = 0;                        /* return value for getopt call */
  int client_socket = -1;           /* socket number for connection to server */
  int rqst_status = -1;             /* status returned from server rqst call */
  int xfer_status = TRANSFER_NOT_NEEDED; /* status returned from server reply call */

  struct sigaction sigalrm_act;     /* structure for holding SIGALRM signal information */
  struct sigaction sigpipe_act;     /* structure for holding SIGPIPE signal information */

/***********************************************************************************
 *
 * Command line variable declarations with appropriate defaults
 *
 **********************************************************************************/

  char *servHost = DEFAULT_HOST;        /* servers host name or dotted decimal address */
  char *servPort = DEFAULT_PORT;        /* servers port name or number */
  char *sf_path =  DEFAULT_SF_PATH;     /* location for support file directory  */
  char *sf_name =  DEFAULT_SF_NAME;     /* support file name to extract current serial number  */
  char *new_sf_name = DEFAULT_NEW_NAME; /* name of file to store new  support file in */
  char *platform = (char *)NULL;        /* platform that support files are needed for */
  char *os_release = (char *)NULL;      /* os release that support files are needed for */
  char *serial_number = (char *)NULL;   /* version to send to server to request for update */
  unsigned int conn_alarm = DEFAULT_CONNECT_ALARM;  /* connection alarm time in seconds */


/*********************************
 *
 * Process command line arguments
 *
 *********************************/

  while((c=getopt(argc, argv, "c:df:h:l:n:o:p:s:t:v")) != EOF)
  {
    switch (c) 
    {
      case 'c' :
        SAY_IT(printf("*** User set connect timeout to %s secs.\n", optarg));
        conn_alarm = (unsigned) atoi(optarg);
        break;

      case 'd' :
        debug_on = 1;
        SAY_IT(printf("*** The debug flag has been turned on.\n"));
        break;

      case 'f':
        sf_name = optarg;
        SAY_IT(printf("*** The support file name entered is %s.\n", sf_name));
        break;

      case 'h':                
        servHost = optarg;
        SAY_IT(printf("*** The server host name entered is %s.\n", servHost));
        break;

      case 'l':
        sf_path = optarg;
        SAY_IT(printf("*** The support file path entered is %s.\n", sf_path));
        break;

      case 'n':
        serial_number = optarg;
        SAY_IT(printf("*** The serial number entered is %s.\n", serial_number));
        break;

      case 'o':                 
        os_release = optarg;
        SAY_IT(printf("*** The operating system release entered is %s.\n", os_release));
        break;

      case 'p':          
        platform = optarg;
        SAY_IT(printf("*** The platform name  entered is %s.\n",platform));
        break;

      case 's':            
        servPort = optarg;
        SAY_IT(printf("*** The server service port entered is %s.\n", servPort));
        break;

      case 't':
        new_sf_name = optarg;
        SAY_IT(printf("*** The file name entered for saving the tar file is %s.\n", new_sf_name));
        break;

      case 'v' :
        verbose_on = 1;
        SAY_IT(printf("*** The verbose flag has been turned on.\n"));
        break;

      case '?':
      default :
        errflg++;
    }
  }
  if(errflg)
  {
    (void)printf("Usage: sf_client [-d] [-f support file name] [-h server host name/address]\n");
    (void)printf("                 [-l path of support files] [-n serial number]\n");
    (void)printf("                 [-o os release from uname -r] [-p platform from uname -s]\n");
    (void)printf("                 [-s server service name/port] [-t tar file name] [-v]\n");
    exit(-1);
  }

/*********************************
 *
 * Setup signal handlers 
 *
 *********************************/

  /* setup signal handler for handling read timeout on socket */
  sigalrm_act.sa_handler = SF_sigHandler; /* specialized signal handler routine */
  sigemptyset(&sigalrm_act.sa_mask);      /* no other signals blocked */
  sigalrm_act.sa_flags = 0;               /* no special options */
  sigaction(SIGALRM, &sigalrm_act, NULL);

  /* setup signal handler for handling writing to a broken socket */
  sigpipe_act.sa_handler = SF_sigHandler; /* specialized signal handler routine */
  sigemptyset(&sigpipe_act.sa_mask);      /* no other signals blocked */
  sigpipe_act.sa_flags = 0;               /* no special options */
  sigaction(SIGPIPE, &sigpipe_act, NULL);


/*********************************
 *
 * Start client processing here 
 *
 *********************************/

  openLogFile(sf_path, SF_CLIENT_LOG_FILE); 

  getPfInfo(&platform, &os_release);

  getSerialNum(sf_path, sf_name, &serial_number); 

  client_socket = callServer(servHost, servPort, conn_alarm);
  if(client_socket != -1)
  {
    rqst_status = rqstSupFile(serial_number, platform, os_release, client_socket);
    if(rqst_status != -1)
    {
      xfer_status = getSvrReply(sf_path, new_sf_name, client_socket);
      switch(xfer_status)
      {
        case TRANSFER_ERROR :
           LOG_IT(fprintf(log_fp, "!!! Unable to complete the support file transfer process, exiting\n"));
           SAY_IT(printf("!!! Unable to complete the support file transfer process, exiting.\n"));
           exit_status = -4; /* general_failure */
           break;

        case TRANSFER_SUCCESS  :
           LOG_IT(fprintf(log_fp, "*** New support files have been transferred successfully.\n"));
           SAY_IT(printf("*** New support files have been transferred successfully.\n"));
           exit_status = 100; /* files_transferred */
           break;

        case TRANSFER_NOT_NEEDED  :
        default :
           LOG_IT(fprintf(log_fp, "*** Support files are up to date, no transfer required.\n"));
           SAY_IT(printf("*** Support files are up to date, no transfer required.\n"));
           exit_status = 0; /* no_transfer_required */
           break;
      }
    } 
    else
    {
      LOG_IT(fprintf(log_fp, "!!! Unable to send support file request to server, exiting.\n"));
      SAY_IT(printf("!!! Unable to send support file request to server, exiting.\n"));
      exit_status = -3;  /* write_error */
    }
    closeServer(&client_socket);
  }
  else  /* connect to server failed */
  {
    LOG_IT(fprintf(log_fp, "!!! Unable to connect to support file server host, exiting.\n"));
    SAY_IT(printf("!!! Unable to connect to support file server host, exiting.\n"));
    exit_status = -2;  /* Connect_error */
  }
  closeLogFile();
  
  SAY_IT(printf("*** Exiting the support file client transfer task.\n"));

  return(exit_status);
}


/*******************************************************************************
 *
 *  NAME: openLogFile
 *
 *  DESCRIPTION:
 *
 *  INVOCATION:
 *
 *
 *******************************************************************************/


static void openLogFile(char *path, char *name )
{

  char file_name[MAXPATHLEN];
  char *time_ptr=(char *)NULL;
  time_t longtime;
   
  DBG_IT(printf("*** Opening log file to record support file transfer events.\n"));
  assert(path != (char *)NULL);
  assert(name != (char *)NULL);

  (void)sprintf(file_name, "%s/%s", path, name);
  if((log_fp = fopen(file_name, "wb")) == (FILE *) NULL)
  {
    SAY_IT(printf("!!! Unable to open log file.\n"));
    DBG_IT(printf("!!! Error opening file %s (%s).\n",file_name,strerror(errno)));
  }
  else
  {
    SAY_IT(printf("*** Log file is %s.\n", file_name)); 
    longtime = time((time_t *) NULL);
    time_ptr =  asctime(localtime(&longtime));
    (void)fprintf(log_fp, "\n********************************************************************\n\n");
    (void)fprintf(log_fp, "SUPPORT FILE CLIENT - ACTIVITY LOG FOR : %s\n",time_ptr);
    (void)fprintf(log_fp, "   *** - indicates normal activity\n");
    (void)fprintf(log_fp, "   !!! - indicates abnormal activity\n"); 
    (void)fprintf(log_fp, "\n********************************************************************\n\n");
  }
}


/*******************************************************************************
 *
 *  NAME: getPfInfo
 *
 *  DESCRIPTION:
 *
 *  INVOCATION:
 *
 *
 *  Note:  If the platform type or release value is NULL, these pointers are assigned
 *         to static data whose values could be overwritten the next time this function
 *         is called.  Therefore, if there is some reason that this routine would need
 *         to be called again, the previous return values should be copied prior to the
 *         calling of this routine.
 *
 *******************************************************************************/


static void getPfInfo(char **platform, char **release )
{
   static struct utsname sysInfo;
   int uname_status = 0;

   DBG_IT(printf("*** Verifying support file client system parameters.\n"));
   assert(platform != (char **)NULL);
   assert(release  != (char **)NULL);

   uname_status = uname(&sysInfo);

   if(*platform == (char *)NULL) 
   {
      *platform = (uname_status != -1) ? sysInfo.sysname : DEFAULT_PLATFORM ;  
   }

/* Now check operating system release, if platform equals the current system platform,
   and the os release is NULL, set it to the current system's os release; otherwise
   set it to the default. 
*/
   if(*release == (char *)NULL)
   {
     if(!strcmp(*platform, sysInfo.sysname) && (uname_status != -1)) 
     {
        *release  = sysInfo.release;
     }
     else
     {
        *release = DEFAULT_RELEASE;
     }
   }
   DBG_IT(printf("*** Using platform %s and release %s.\n", *platform, *release));
}



/*******************************************************************************
 *
 *  NAME: getSerialNum
 *
 *
 *  DESCRIPTION: This routine sets the serial number if it was not entered
 *               as a command line argument. It first tries to open the current
 *               support file to obtain the serial number from the header.  If
 *               that fails, it sets the serial number to the default value.
 *               The default value requests that the server send the latest 
 *               version of the support files. 
 *
 *  INVOCATION:
 *
 *  Note:  If the serial number is received from a current support file, the value
 *         returned in sn_wnated is a pointer to a static character string file_sn. 
 *         Therefore, if there is some reason that this routine would need
 *         to be called again, the previous serial number should be copied prior to the
 *         calling of this routine.
 *
 *******************************************************************************/
 
static void getSerialNum(char *path, char *name, char **sn_wanted)
{ 

  FILE *fp = (FILE *)NULL;
  char file_name[MAXPATHLEN];
  long  hdr_version = 0;
  static char file_sn[80]; /* 80 is an arbitrary length, the routine that sends this
                              to the server will trim to required length if necessary */

  if (*sn_wanted == (char *)NULL)  /* no serial number entered, get from current sf */
  {
    (void)sprintf(file_name, "%s/%s", path, name); 
    fp = fopen(file_name, "r");
    if(fp != (FILE *)NULL)
    {

      DBG_IT(printf("!!! Support file used to get serial number is %s.\n", file_name));
      (void)fscanf(fp, "%ld", &hdr_version);
      DBG_IT(printf("!!! The support file header version is %ld.\n", hdr_version));
      if(hdr_version == SF_CURRENT_HDR_VERSION)
      {
        (void)fscanf(fp, "%79s", file_sn);
        DBG_IT(printf("*** Serial number from current file is (%s).\n",file_sn));
        *sn_wanted = file_sn; 
      }
      else
      {
        DBG_IT(printf("!!! Invalid header version (%ld) in current support file.\n",hdr_version));
        LOG_IT(fprintf(log_fp, "!!! Invalid header version in support file, requesting latest files.\n"));
        SAY_IT(printf("!!! Invalid header version in support file, requesting latest files.\n"));
        *sn_wanted = DEFAULT_SERIAL_NUMBER;
      }
      if(fclose(fp))
      {
        DBG_IT(printf("!!! Error closing current support file (%s)\n", strerror(errno)));
      }
    }
    else  /* fp == NULL */
    {
      DBG_IT(printf("!!! Error opening file %s (%s).\n",file_name,strerror(errno)));
      LOG_IT(fprintf(log_fp, "!!! Unable to get current serial number, requesting latest files.\n"));
      SAY_IT(printf("!!! Unable to get current serial number, requesting latest files.\n"));
      *sn_wanted = DEFAULT_SERIAL_NUMBER;
    }
  }

  DBG_IT(printf("*** Using serial number %s.\n", *sn_wanted));

}




/***************************************************************************
 *
 * NAME: callServer
 *
 * DESCRIPTION:  This routine makes a TCP socket connection to a specified
 *               host and port number.  If the connection fails, a -1 is
 *               returned as the socket descriptor.  If the connection
 *               succeeds, the socket descriptor assigned to the socket is
 *               returned.
 *
 * INVOCATION:
 *
 *               socket_number = callServer(host, service, conn_alarm)
 *
 *     where     host - is a char * to either the name such as
 *                      imssfp.gsfc.nasa.gov (killians) or the dotted
 *                      decimal notation such as 192.107.190.142
 *
 *               service - is a char * to either the service name such
 *                         as support_file_server which would be in the
 *                         /etc/services file or the actual port number
 *                         that corresponds to the service name in the
 *                         services file. This port number is the port
 *                         that the support file server "listens" for
 *                         connection requests
 *               conn_alarm - timeout in seconds for alarm around connect().
 *
 **************************************************************************/

static int callServer( char *host, char *service, unsigned int conn_alarm )
{

#ifndef INADDR_NONE
#define INADDR_NONE  0xffffffff
#endif  /* INADDR_NONE */

  struct hostent  *hostInfo =  (struct hostent *)NULL; /* pointer to host information entry */
  struct servent  *servInfo =  (struct servent *)NULL; /* pointer to server information entry */
  struct sockaddr_in sockInfo; /* an internet endpoint address structure */
  int socket_number = -1;      /* socket descriptor number */
  int retstat;                 /* return status */

  LOG_IT(fprintf(log_fp, "*** Issueing connect to host %s at port %s.\n", host, service));
  SAY_IT(printf("*** Issueing connect to host %s at port %s.\n", host, service));
  assert(host != (char*)NULL);
  assert(service != (char*)NULL);

  (void)memset((void *)&sockInfo, 0, sizeof(sockInfo));
  sockInfo.sin_family = AF_INET;

/*********************************************************************************
 *
 * Get the port number from the service name. If that fails, the service name was
 * entered in in the format of an ascii port number so convert to network byte
 * ordered integer.
 *
 ********************************************************************************/
  if(servInfo = getservbyname(service, "tcp"))
  {
    sockInfo.sin_port = servInfo->s_port;
  }
  else if((sockInfo.sin_port = htons((u_short)atoi(service))) == 0)
  {
    DBG_IT(printf("!!! Unable to resolve service name (%s).\n", service));
    return(-1);
  }

  DBG_IT(printf("*** Service port number is (%d).\n", sockInfo.sin_port));


/*********************************************************************************
 *
 * Get the IP address from the host name.  If that fails, the host name was input in
 * the in dotted decimal notation so call inet_addr to convert to a network byte
 * ordered integer.
 *
 ********************************************************************************/
  if(hostInfo = gethostbyname(host))
  {
    (void)memcpy((void *)&sockInfo.sin_addr, (void *)hostInfo->h_addr, hostInfo ->h_length);
  }
  else if((sockInfo.sin_addr.s_addr = inet_addr(host)) == INADDR_NONE)
  {
    DBG_IT(printf("!!! Unable to resolve host name (%s).\n", host));
    return(-1);
  }

  DBG_IT(printf("*** Host address is (%lx).\n", sockInfo.sin_addr.s_addr));

/*********************************************************************************
 *
 * Allocate a socket, then make the connection and return to the calling
 * routine the socket descriptor.
 *
 ********************************************************************************/
  socket_number = socket(PF_INET, SOCK_STREAM, 0);
  DBG_IT(printf("*** Socket number is (%d).\n", socket_number));
  if(socket_number < 0)
  {
    DBG_IT(printf("!!! Unable to create a socket (%s)\n", strerror(errno)));
    return(-1);
  }

  retstat = SF_connect(socket_number, (struct sockaddr *)&sockInfo, conn_alarm);

  if(retstat != 0)
  {
    DBG_IT(printf("!!! Unable to connect to %s:%s (%s).\n", host, service,
                                                              strerror(errno)));
    (void)close(socket_number); /* close since socket was allocated */
    return(-1);
  }

  SAY_IT(printf("*** Connected to %s:%s.\n", host, service));

  return(socket_number);
}



/*******************************************************************************
 *
 *  NAME: rqstSupFile
 *
 *  DESCRIPTION: This routine builds the client request message and writes it
 *               to the support file server.  The serial number, hardware platform,
 *               and operating system version are received as parameters and are used
 *               to build the client request message.
 *
 *  INVOCATION:
 *
 *               status = rqstSupFile(serial_num, platform, release, sock_num)
 * 
 *               where
 *                       serial_num - is the serial number of the current support
 *                                    files that reside on the client system. Parameter
 *                                    format is char *.
 *             
 *                       platform - is the hardware platform that the support files
 *                                  are being requested for. Parameter format is char *.
 *
 *                       release - is the operating system release number of the platform
 *                                 that the support files are being requested for.  Parameter
 *                                 format is char *.
 *
 *                       sock_num - is the file descriptor for the socket that has
 *                                  been opened for communication with the server.  Parameter
 *                                  format is int.
 *
 *                       status - status of writing the request to the server. 
 *                                -1 for failure.
 *                                 0 for success.
 *
 *******************************************************************************/

static int rqstSupFile(char *serial_num, char *platform, char *release, int sock_num)
{

  int status = 0; 
  int num_written = 0;
  char *why = (char *)NULL;

  CL_RQST_MSG rqst;

  assert(serial_num != (char *)NULL);
  assert(platform   != (char *)NULL);
  assert(release    != (char *)NULL);
  assert(sock_num   != -1);
 
/***********************************
 * 
 * Build the client request message
 *
 ***********************************/

  (void)strncpy(rqst.type, CL_RQST_TYPE, sizeof(rqst.type));   /* set message type */
  rqst.type[MT_SIZE - 1] = '\0';

  (void)strncpy(rqst.msg.pf_type, platform, sizeof(rqst.msg.pf_type)); /* set platform type */
  rqst.msg.pf_type[PF_SIZE -1] = '\0'; /* in case platform length > PF_SIZE */ 
  
  (void)strncpy(rqst.msg.pf_release, release, sizeof(rqst.msg.pf_release)); /* set os release number */
  rqst.msg.pf_release[PF_SIZE -1] = '\0'; /* in case release length > PF_SIZE */ 
  
  (void)strncpy(rqst.msg.curr_sn, serial_num, sizeof(rqst.msg.curr_sn)); /* set current serial number */
  rqst.msg.curr_sn[SN_SIZE - 1] = '\0';

/***********************************
 * 
 * Send the client request message
 *
 ***********************************/

  LOG_IT(fprintf(log_fp, "*** Sending request to server for system %s:%s, serial number %s.\n",
                               rqst.msg.pf_type, rqst.msg.pf_release, rqst.msg.curr_sn));
  SAY_IT(printf("*** Sending request to server for system %s:%s, serial number %s.\n",
                               rqst.msg.pf_type, rqst.msg.pf_release, rqst.msg.curr_sn));

  DBG_IT(printf("*** Writing %d bytes to server.\n", sizeof(rqst)));
  num_written = SF_writeStream(sock_num,(char *)&rqst, sizeof (rqst));
  DBG_IT(printf("*** Wrote %d bytes to server.\n", num_written));
  if((num_written  < 0) || (num_written != sizeof(rqst))) 
  {
     why = (num_written == -2) ?  "timeout" : "error";
     LOG_IT(fprintf(log_fp,"!!! A write %s occurred while sending the request.\n", why));
     SAY_IT(printf("!!! A write %s occurred while sending the request.\n", why));
     status = -1; /* error writing request to server */
  }
  return(status);
}


/*******************************************************************************
 *
 *  NAME: getSvrReply
 *
 *  DESCRIPTION:  This routine reads in the response from the support file server
 *                after a request has been sent.  The server can respond with a 
 *                SV_NO_UPDATE_TYPE message, a SV_ERROR_TYPE message or a SV_UPDATE_TYPE
 *                message. Preceding each message is a message identifier. The SV_UPDATE_TYPE
 *                message is the only message that the client receives that contains 
 *                additional information.  The update message contains the number of bytes
 *                that are contained in the compressed support file that will be transferred.
 *                Once the byte count has been read, a routine is called to read in and 
 *                write to disk the entire file.  
 *
 *  INVOCATION:
 *
 *              status = getSvrReply(path, name, sock_num)
 *             
 *              where
 *
 *                       path - is the directory path to write the new support file
 *                              should the server send us one.  Parameter format is
 *                              char*.
 *
 *                       name - is the name of the file to save the new support file
 *                              in.  Parameter format is char*.
 *
 *                       sock_num - is the file descriptor for the socket that has been
 *                                  established for communication with the server.  Parameter
 *                                  format is int.
 *
 *                       status - is the status of the transfer process
 *                                TRANSFER_NOT_NEEDED
 *                                TRANSFER_SUCCESS
 *                                TRANSFER_ERROR
 * 
 *******************************************************************************/
static int getSvrReply(char *path, char *name, int sock_num)
{

  int num_read = 0;
  char *why = (char *)NULL;
  char full_name[MAXPATHLEN];
  int status = TRANSFER_NOT_NEEDED;

  SF_MSG_TYPE msg_type;
  SV_UPDATE   update_resp;

  assert(path != (char *)NULL);
  assert(name != (char *)NULL);
  assert(sock_num != -1);

  DBG_IT(printf("*** Reading in %d bytes from server.\n",sizeof(msg_type)));
  num_read = SF_readStream(sock_num, (char *)&msg_type, sizeof(msg_type));
  if((num_read < 0) || (num_read != sizeof(msg_type)))  /* uh oh, had a read error */
  {
     why = (num_read == -2) ?  "timeout" : "error";
     LOG_IT(fprintf(log_fp,"!!! A read %s occurred while reading the message type.\n", why));
     SAY_IT(printf("*** A read %s occurred while reading the message type.\n", why));
     status = TRANSFER_ERROR; /* error reading file from server */
  }
  else
  {
     DBG_IT(printf("*** Read in %d bytes from server.\n",num_read));
     DBG_IT(printf("*** Received msg_type of %s.\n",msg_type));

     if(!strncmp(msg_type, SV_NO_UPDATE_TYPE, sizeof(SF_MSG_TYPE)))
     {
        LOG_IT(fprintf(log_fp, "*** Received an SV_NO_UPDATE_TYPE message from the server.\n"));
        SAY_IT(printf("*** Received an SV_NO_UPDATE_TYPE message from the server.\n"));
        status = TRANSFER_NOT_NEEDED ;
     }
     else if (!strncmp(msg_type, SV_UPDATE_TYPE,sizeof(SF_MSG_TYPE)))
     {
        LOG_IT(fprintf(log_fp, "*** Received an SV_UPDATE_TYPE message from the server.\n"));
        SAY_IT(printf("*** Received an SV_UPDATE_TYPE message from the server.\n"));

        num_read = SF_readStream(sock_num, update_resp.num_bytes, sizeof(update_resp.num_bytes));
        DBG_IT(printf("*** Read in %d bytes from server.\n",num_read));
        if((num_read < 0) || (num_read != sizeof(update_resp.num_bytes))) 
        {
           why = (num_read == -2) ?  "timeout" : "error";
           LOG_IT(fprintf(log_fp,"!!! A read %s occurred while reading support file size.\n", why));
           SAY_IT(printf("*** A read %s occurred while reading supportfile size.\n", why));
           status = TRANSFER_ERROR; /* error reading file from server */
        }
        else
        {
            (void)sprintf(full_name, "%s/%s", path, name);
            status = getSupFile(update_resp.num_bytes, full_name,  sock_num);
        }
     }
     else if(!strcmp(msg_type, SV_ERROR_TYPE))
     {
        LOG_IT(fprintf(log_fp, "*** Received an SV_ERROR_TYPE message from the server.\n"));
        SAY_IT(printf("*** Received an SV_ERROR_TYPE  message type from server.\n"));
        status = TRANSFER_ERROR;
     }
     else
     {
        LOG_IT(fprintf(log_fp, "!!! Received an unknown message from the server.\n"));
        SAY_IT(printf("!!! Received an unknown message type from server.\n"));
        status = TRANSFER_ERROR;
     }
  }
  return(status);
}


/*******************************************************************************
 *
 *  NAME: getSupFile
 *
 *  DESCRIPTION: This routine opens up a file to write the new support file to 
 *               then calls the appropriate routine to read in and save the file.
 *               reads in the support file from the server.
 *
 *  INVOCATION:
 *
 *           status = getSupFile(file_size, file_name, sock_num)
 *         
 *              where
 *                     file_size - is the number of bytes to receive. Parameter
 *                                 format is char*.
 *
 *                     file_name - complete path and name of the file to store the
 *                                 new support file in.  Parameter format is char*.
 *
 *                      sock_num - is the file descriptor for the socket that has been
 *                                  established for communication with the server.  Parameter
 *                                  format is int.
 *
 *                       status - is the status of the transfer process
 *                                TRANSFER_NOT_NEEDED
 *                                TRANSFER_SUCCESS
 *                                TRANSFER_ERROR
 *
 *  Note:  If the number of bytes read in from the server matches the size that
 *         the server indicated was in the file, then check to see if more data
 *         is available on the socket.  If so, then the server transmitted more
 *         than it was supposed to. Since we don't know if this data is part of 
 *         the file, consider this an error and abort the support file transfer 
 *         process.  
 *
 *******************************************************************************/
static int getSupFile(char *file_size, char *file_name, int sock_num)
{
  int status = TRANSFER_SUCCESS; /* lets be optimistic about this */
  int num_read = 0;              /* number of bytes from additional read */
  long file_len, xfer_len;       /* the length of file from server, # of bytes we read in */
  int sf_fd = (int)NULL;

  char extra_buf[MT_SIZE];   /* buffer for read to determine if more data is available */
  char *why;                 /* points to "error' or "timeout"  for error messages */

  assert(file_size != (char *)NULL);
  assert(file_name != (char *)NULL);
  assert(sock_num != -1);

  if((sf_fd = open(file_name, O_WRONLY|O_CREAT, 0755)) != -1)
  {
    file_len = atol(file_size);
    if(file_len > 0)
    {
       LOG_IT(fprintf(log_fp, "*** Receiving %ld bytes in file %s.\n", 
                                                                 file_len, file_name));
       SAY_IT(printf("*** Receiving %ld bytes in file %s.\n",file_len, file_name));

       xfer_len = SF_moveStream(sock_num, sf_fd, file_len);
       if(xfer_len > 0)
       {
          DBG_IT(printf("*** A total of %ld bytes were received from the server.\n", xfer_len)); 

          if(xfer_len == file_len) /* bytes read  match server's size, check for more data */
          {
             num_read = SF_readStream(sock_num, extra_buf, sizeof(extra_buf));
             DBG_IT(printf("*** Value returned after extra read is %d.\n",num_read ));
             if(num_read <= 0) /* no more data, transfer was a success */
             {
                status = TRANSFER_SUCCESS; 
             }
             else /* num_read > 0, more data which we don't know what it is */
             {
                LOG_IT(fprintf(log_fp,"!!! More data available than server indicated, aborting transfer.\n"));
                SAY_IT(printf("!!! More data available than server indicated, aborting transfer.\n")); 
                status = TRANSFER_ERROR; 
             }
          }
          else /* xfer_len != file_len */
          {
             LOG_IT(fprintf(log_fp,"!!! Support file size does not match size indicated by server.\n"));
             SAY_IT(printf("!!! Support file size does not match size indicated by server.\n")); 
             status = TRANSFER_ERROR; 
          }
       }
       else  /* xfer_len <= 0 */
       {
         if(xfer_len == -3)
         {
            LOG_IT(fprintf(log_fp,"!!! A write error occurred while saving the support files.\n"));
            SAY_IT(printf("!!! A write error occurred while saving the support files.\n"));
         }
         else
         {  
             why = (xfer_len == -2) ?  "timeout" : "error";
             LOG_IT(fprintf(log_fp,"!!! A read %s occurred while receiving the support files.\n",why));
             SAY_IT(printf("!!! A read %s occurred while receiving the new support files.\n", why));
         }
         status = TRANSFER_ERROR; 
       }
    }
    else  /* file_len <= 0 */
    {
       DBG_IT(printf("*** File size received from server is %s.\n", file_size));
       DBG_IT(printf("!!! Error converting file size value to integer (%s).\n",
                                                                    strerror(errno)));
       LOG_IT(fprintf(log_fp, "!!! Invalid file size value received from server.\n"));
       SAY_IT(printf("!!! Invalid file size value received from server.\n"));
       status = TRANSFER_ERROR; 
    }
    if(close(sf_fd))
    {
      DBG_IT(printf("!!! Error closing file %s (%s).\n",file_name,strerror(errno)));
    }
    if (status == TRANSFER_ERROR)
    {
       if(remove(file_name))
       {
          DBG_IT(printf("!!! Error removing file %s (%s).\n",file_name,strerror(errno)));
       }
    }
  }
  else  /* sf_fd == NULL */ 
  {
     DBG_IT(printf("!!! Error opening file %s (%s).\n",file_name,strerror(errno)));
     LOG_IT(fprintf(log_fp, "!!! Unable to open file %s to store support files.\n",file_name));
     SAY_IT(printf("!!! Unable to open file %s to store support files.\n", file_name));
     status = TRANSFER_ERROR; /* general failure status */
  }
  return(status);
}


/***************************************************************************
 *
 * NAME: closeServer
 *
 * DESCRIPTION:  This routine closes a TCP socket connection.  A shutdown
 *               is done first to terminate any I/O on the socket, then a
 *               close is done to close the socket.
 *
 * INVOCATION:
 *
 *               closeServer(&socket)
 *
 *     where    socket - is a int * to the socket number to close
 *
 **************************************************************************/

static void closeServer(int *sock_num)
{

  assert(sock_num != (int *)NULL);
  assert(*sock_num >= -1);

  DBG_IT(printf("*** Closing socket number %d.\n",*sock_num));

  if(*sock_num != -1)
  {
    if(shutdown(*sock_num, 2))     /* shudown I/O on socket both ways(2) */
    {
       DBG_IT(printf("!!! Error shutting down socket number %d (%s).\n",*sock_num, strerror(errno)));
    }
    if(close(*sock_num))
    {
       DBG_IT(printf("!!! Error closing socket number %d (%s).\n",*sock_num, strerror(errno)));
    }
    *sock_num = -1;
  }

}


/*******************************************************************************
 *
 *  NAME: closeLogFile
 *
 *  DESCRIPTION: This routine closes the log file if it is open.  Prior to closing
 *               a format line is written to the log file.  
 *
 *  INVOCATION:
 *
 *                closeLogFile()
 *
 *
 *  uses the global variables : log_fp, verbose_on through the LOG_IT and SAY_IT macros
 *
 *******************************************************************************/

static void closeLogFile(void)
{

  LOG_IT(fprintf(log_fp, "\n********************************************************************\n\n"));
  if(log_fp != NULL)
  {
    DBG_IT(printf("*** Closing the log file.\n"));
    if(fclose(log_fp))
    {
       DBG_IT(printf("!!! Error closing the log file(%s).\n",strerror(errno)));
    }
  }
}
