/**************************************************************
 *
 *  NAME: SF_sockIOLib.c
 *
 *  DESCRIPTION: This is a family of functions that are used to
 *               handle the socket communication between the
 *               support file client and server tasks.
 *
 *  ROUTINES:
 *
 *              SF_sigHandler  - signal handler for SIGALRM/SIGPIPE
 *              SF_readStream  - reads from a stream socket
 *              SF_writeStream - writes to a stream socket
 *              SF_moveStream  - reads from one file descriptor
 *                               (file or socket) and writes to
 *                               another 
 *
 **************************************************************
 *
 * RCS INFORMATION
 *
 * $RCSfile: SF_sockIOLib.c,v $
 *
 * $Id: SF_sockIOLib.c,v 1.2 1997/10/23 19:07:14 solomon Exp $
 *
 * $Log: SF_sockIOLib.c,v $
 * Revision 1.2  1997/10/23 19:07:14  solomon
 * Added: #include <Gaea_config.h>
 *
 * Revision 1.1  1996/11/19  23:37:38  lrosen
 * Initial revision
 *
 * Revision 1.0.1.2  1996/08/07  15:40:35  lrosen
 * Added function SF_connect which puts an alarm around the call to
 * connect().
 *
 * Revision 1.0.1.1  1996/08/07  13:46:49  lrosen
 * Place holder for new branch.
 *
 * Revision 1.0  1996/05/22  02:43:25  ims
 * First official version, promoted from 0.1.1.3
 *
 * Revision 0.1.1.3  1996/03/08  04:35:27  walbert
 * Increased the read/write timeout periods from 10 seconds
 * to 60 seconds.  These were previously set to a small number
 * for testing purposes.  There is no magic to the 60 seconds,
 * this was chosen as an arbitrary number.
 *
 * Revision 0.1.1.2  1996/01/09  14:36:48  walbert
 * Incorporated walkthru comments and added additional comments.
 *
 * Revision 0.1.1.1  1995/12/19  19:06:01  walbert
 * Updated prologs
 *
 * Revision 0.1  1995/12/19  14:53:27  walbert
 * First pre-release version
 *
 *
 *************************************************************/
 
#include <Gaea_config.h>
#include <stdio.h>        
#include <stdlib.h>       /* atoi */
#include <unistd.h>       /* read,write...*/
#include <signal.h>       /* SIGALRM */
#include <errno.h>        /* errno */
#include <assert.h>       /* assert */
#include <sys/socket.h>   /* socket stuff */

void SF_sigHandler(int);
int  SF_readStream(int, char*, int);
long SF_moveStream(int, int, long);
int  SF_writeStream(int, char*, int);
int  SF_connect(int, struct sockaddr *, unsigned int);

#define SF_READ_TIMEOUT     60
#define SF_WRITE_TIMEOUT    60
#define READ_FILE_SIZE    2048    /* used as the read size when receiving file */

enum SF_Sig_Val
{
   SIG_RESET_VAL,
   SIG_ALRM_VAL,
   SIG_PIPE_VAL
};
typedef enum SF_Sig_Val SF_SIG_VAL;

static SF_SIG_VAL sig_value = SIG_RESET_VAL;
 


/**************************************************************************
 *
 *  NAME: SF_sigHandler
 *
 *  This routine is used as a general purpose signal handler.  It is currently
 *  used to catch the SIGALRM and SIGPIPE signals.  The SIGALRM signal is received
 *  when the timer (alarm()) expires.  The SIGPIPE alarm is received if a 
 *  write is performed on a closed socket.  
 * 
 *  INVOCATION: (when approoriate signal raises)
 *          
 *              SF_sigHandler(sig_num)
 *
 *              where     sig_num - is the number of the signal that was raised
 *
 *  Note:
 *        This routine assumes that you have already set up the signal handler 
 *        facility with the sigaction function.
 *
 **************************************************************************/

void SF_sigHandler( int sig_num)
{
  if(sig_num == SIGALRM)
  {
    sig_value = SIG_ALRM_VAL;
  }
  else if(sig_num == SIGPIPE)
  {
    sig_value = SIG_PIPE_VAL;
  }
}


/***************************************************************************
 * 
 * NAME: SF_readStream
 *
 * DESCRIPTION:  This routine reads a stream of bytes from a socket.  A read
 *               on a socket might input fewer bytes than requested, but this
 *               is not an error condition.  The reason is that buffer limits
 *               might be reached for the socket in the kernel and all that is 
 *               required is for the caller to invoke the read system call again,
 *               to input the remaining bytes.  Therefore, this routine loops 
 *               until the number of bytes requested has been read, an error
 *               occurs, or an EOF is found.  It also uses the alarm routine to
 *               setup a timer  to interrupt the read call after a specified 
 *               amount of time.  The alarm routine raises the SIGALRM signal
 *               upon timer expiration.
 *
 *
 * INVOCATION:
 *
 *               status = SF_readStream(fd, ptr, nbytes)
 *
 *     where        fd - is a int file descriptor for the socket
 *                  ptr - is a char * to the data buffer
 *                  nbytes - an int specifying the number of bytes to read.
 *                  status - number of bytes read/error condition(<0)
 *
 * Note: Inside the do loop the conditions for nread >0 and nread <0 are explicitly tested
 *       for since additional actions are required.  The condition of nread == 0
 *       is not tested , but will be caught by the while portion of the do loop with
 *       the test (nread > 0). 
 * 
 **************************************************************************/

int SF_readStream(int fd, char* ptr, int nbytes)
{
  int  nleft;      /* number of bytes left to read */
  int  nread;      /* number of bytes read */

  assert(fd >= 0);
  assert(ptr != (char *)NULL);
  assert(nbytes > 0);

  sig_value = SIG_RESET_VAL;
  nleft = nbytes;
  do
  {
    (void)alarm(SF_READ_TIMEOUT);  /* set timer to prevent read from blocking forever */
    nread = read(fd, ptr, nleft);
    (void)alarm(0); 
    if(nread > 0)     /* good read */
    {
      nleft -= nread;
      ptr += nread;
    }
    else if(nread < 0)     /* ERROR - either a read timeout or a read error */
    {
      if(sig_value == SIG_ALRM_VAL) /* set by SF_sigHandler if SIGALRM signal */
      {
        sig_value = SIG_RESET_VAL;
        nread = -2;   /* change from -1 so calling rtn can tell a timeout occurred */
      }
    }
  }while ((nleft > 0) && (nread > 0));

  return( (nread < 0) ? nread : (nbytes - nleft));   

}


/***************************************************************************
 * 
 * NAME: SF_moveStream
 *
 * DESCRIPTION:  This routine reads a stream of bytes from a socket/file and writes
 *               them to a socket/file. This routine is passed in the file 
 *               descriptors for reading/writing to and the number of bytes to 
 *               transfer.    
 *
 *               It also uses the alarm routine to to setup a timer to interrupt
 *               the read call after a specified amount of time.  The alarm 
 *               routine raises the SIGALRM signal upon timer expiratation.
 *
 *
 * INVOCATION:
 *
 *               status = SF_moveStream(fd, ptr, nbytes)
 *
 *     where        in_fd  - is an int file descriptor for theinput file/socket 
 *                  out_fd - is an int file descriptor for the output file/socket.
 *                  nbytes - an long  specifying the number of bytes to read.
 *                  status - number of bytes read/error condition(<0)
 *
 * Note: Inside the do loop the conditions for nread >0 and nread <0 are explicitly tested
 *       for since additional actions are required.  The condition of nread == 0
 *       is not tested , but will be caught by the while portion of the do loop with
 *       the test (nread > 0). 
 * 
 **************************************************************************/

long SF_moveStream(int in_fd, int out_fd, long nbytes)
{
  long  nleft;    /* number of bytes left to read */
  long  nread;    /* number of bytes read */
  int num_written; 
  
  char buf[READ_FILE_SIZE]; 

  assert(in_fd > 0);
  assert(out_fd > 0);
  assert(nbytes > 0);
 

  sig_value = SIG_RESET_VAL;
  nleft = nbytes;

/************************************************************************
 *
 * Perform do loop until read/write error (nread < 0), EOF(nread =0), or 
 * all bytes have been read (nleft == 0)
 *
 ************************************************************************/
  do
  {
    (void)alarm(SF_READ_TIMEOUT);   /* set timer to prevent read from blocking forever */
    nread = (nleft > READ_FILE_SIZE) ? read(in_fd, buf, READ_FILE_SIZE) : read(in_fd, buf, nleft); 
    (void)alarm(0);
    if(nread > 0)     /*good read */ 
    {
      num_written = SF_writeStream(out_fd, buf, (int)nread);
      if(num_written != nread)
      {
        /* set nread to failure state so we will break out of loop */
        nread = -3;
      }
      else
      {
        nleft -= nread;
      } 

    }
    else if(nread < 0) /* bad read */
    {
      if(sig_value == SIG_ALRM_VAL) /* set by SF_sigHandler if SIGALRM signal */
      {
        sig_value = SIG_RESET_VAL;
        nread = -2;   /* change from -1 so calling rtn can tell a timeout occurred */
      }
    }
  }while ((nleft > 0) && (nread > 0));

  return( (nread < 0) ? nread : (nbytes - nleft));   
}


/***************************************************************************
 * 
 * NAME: SF_writeStream
 *
 * DESCRIPTION:  This routine writes a stream of bytes to a socket.  A write
 *               on a socket might output fewer bytes than requested, but this
 *               is not an error condition.  The reason is that buffer limits
 *               might be reached for the socket in the kernel and all that is 
 *               required is for the caller to invoke the write system call again,
 *               to output the remaining bytes.  Therefore, this routine loops 
 *               until the number of bytes requested has been writei or an error
 *               occurs.
 *
 *               This routine also uses the alarm function to setup a timer to 
 *               interrupt the write call after a specified amount of time.  The
 *               alarm routine raises the SIGALRM signal upon timer expiration.
 *
 *
 * INVOCATION:
 *
 *               status = SF_writeStream(fd, ptr, nbytes)
 *
 *     where        fd - is a int file descriptor for the socket
 *                  ptr - is a char * to the data buffer
 *                  nbytes - an int specifying the number of bytes to write.
 *                  status - number of bytes written/error condition(<0)
 *
 * Note: Inside the do loop the conditions for nwritten >0 and nwritten <0 are explicitly 
 *       tested for since additional actions are required.  The condition of nwritten == 0
 *       is not tested , but will be caught by the while portion of the do loop with
 *       the test (nwritten > 0). 
 * 
 **************************************************************************/


int SF_writeStream(int fd, char *ptr, int nbytes)
{

  int  nleft;      /* number of bytes left to write */
  int  nwritten;   /* number of bytes written */

  assert(fd >= 0);
  assert(ptr != (char *)NULL);
  assert(nbytes > 0);

  sig_value = SIG_RESET_VAL;
  nleft = nbytes;
  do 
  {
    (void)alarm(SF_WRITE_TIMEOUT);
    nwritten = write(fd, ptr, nleft);
    (void)alarm(0);
    if(nwritten > 0)     /* good write */
    {
      nleft -= nwritten;
      ptr   += nwritten;
    }
    else if(nwritten < 0)     /* ERROR */
    {
      if(sig_value == SIG_ALRM_VAL) /* set by SF_sigHandler if SIGALRM signal */
      {
        sig_value = SIG_RESET_VAL;
        nwritten = -2; /* change from -1 so calling rtn can tell a timeout occurred */
      }
    }
  }while ((nleft > 0) && (nwritten > 0));

  return( (nwritten < 0) ? nwritten : (nbytes - nleft));   
}

/***************************************************************************
 * 
 * NAME: SF_connect
 *
 * DESCRIPTION:  This routine attempts to establish a connection to server
 *               at the host.
 *
 *               This routine also uses the alarm function to setup a timer to 
 *               interrupt the connect call after a specified amount of time.
 *               The alarm raises the SIGALRM signal upon timer expiration.
 *
 * INVOCATION:
 *
 *               status = SF_connect(socket_number, psockInfo, sf_timeout)
 *
 *     where        socket_number - is an int file descriptor for the socket.
 *                  psockInfo - is a pointer to the struct sockaddr socket
 *                              address to connect to.
 *                  sf_timeout - is an unsigned int time for alarm() to wait
 *                               for connect in seconds.
 *                  status - 0 - all ok; anything else - error.
 *
 **************************************************************************/

int SF_connect(int socket_number, struct sockaddr *psockInfo,
               unsigned int sf_timeout)
{
  int constat;       /* return status of connect() function */

  assert(socket_number >= 0);
  assert(psockInfo != (struct sockaddr *)NULL);
  assert(sf_timeout > 0);

  /* reset signal value */
  sig_value = SIG_RESET_VAL;
  /* set alarm with timeout */
  (void)alarm(sf_timeout);
  /* attempt connection */
  constat = connect(socket_number, psockInfo, sizeof(struct sockaddr));
  /* turn alarm off */
  (void)alarm(0);
  /* check if SIGALARM or SIGPIPE signal was raised. */
  if(sig_value != SIG_RESET_VAL)
  {
    if(sig_value == SIG_ALRM_VAL) /* set by SF_sigHandler if SIGALRM signal */
    {
      sig_value = SIG_RESET_VAL;
      return(SIGALRM);
    }
    else if(sig_value == SIG_PIPE_VAL)  /* set if SIGPIPE */
    {
      sig_value = SIG_RESET_VAL;
      return(SIGPIPE);
    }
  }
  else
  {
    /* return the return status of connect. 0 = ok, -1 = error. */
    return(constat);
  }
}
